/* Generated by Cython 3.2.4 */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
/* InitLimitedAPI */
#if defined(Py_LIMITED_API)
  #if !defined(CYTHON_LIMITED_API)
  #define CYTHON_LIMITED_API 1
  #endif
#elif defined(CYTHON_LIMITED_API)
  #ifdef _MSC_VER
  #pragma message ("Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.")
  #else
  #warning Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x03080000
    #error Cython requires Python 3.8+.
#else
#define __PYX_ABI_VERSION "3_2_4"
#define CYTHON_HEX_VERSION 0x030204F0
#define CYTHON_FUTURE_DIVISION 1
/* CModulePreamble */
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 1
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 1
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PYPY_VERSION_NUM >= 0x07030C00)
  #endif
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_NUM >= 0x07031100)
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND (__PYX_LIMITED_VERSION_HEX >= 0x030A0000)
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #ifndef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 1
  #endif
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #ifdef Py_GIL_DISABLED
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 1
  #else
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #endif
  #if PY_VERSION_HEX < 0x030A0000
    #undef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #elif !defined(CYTHON_USE_TYPE_SLOTS)
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLIST_INTERNALS)
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #elif !defined(CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL 0
  #elif !defined(CYTHON_FAST_GIL)
    #define CYTHON_FAST_GIL (PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_SYS_MONITORING
    #define CYTHON_USE_SYS_MONITORING (PY_VERSION_HEX >= 0x030d00B1)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5 && !CYTHON_USE_MODULE_STATE)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS (!CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
  #if defined(CYTHON_IMMORTAL_CONSTANTS) && PY_VERSION_HEX < 0x030C0000
    #undef CYTHON_IMMORTAL_CONSTANTS
    #define CYTHON_IMMORTAL_CONSTANTS 0  // definitely won't work
  #elif !defined(CYTHON_IMMORTAL_CONSTANTS)
    #define CYTHON_IMMORTAL_CONSTANTS (PY_VERSION_HEX >= 0x030C0000 && !CYTHON_USE_MODULE_STATE && CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
#endif
#ifndef CYTHON_COMPRESS_STRINGS
  #define CYTHON_COMPRESS_STRINGS 1
#endif
#ifndef CYTHON_FAST_PYCCALL
#define CYTHON_FAST_PYCCALL  CYTHON_FAST_PYCALL
#endif
#ifndef CYTHON_VECTORCALL
#if CYTHON_COMPILING_IN_LIMITED_API
#define CYTHON_VECTORCALL  (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
#else
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL)
#endif
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#include <stdint.h>
typedef uintptr_t  __pyx_uintptr_t;
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifndef Py_UNREACHABLE
  #define Py_UNREACHABLE()  assert(0); abort()
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

/* CInitCode */
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

/* PythonCompatibility */
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#define __Pyx_BUILTIN_MODULE_NAME "builtins"
#define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    #ifndef CO_OPTIMIZED
    static int CO_OPTIMIZED;
    #endif
    #ifndef CO_NEWLOCALS
    static int CO_NEWLOCALS;
    #endif
    #ifndef CO_VARARGS
    static int CO_VARARGS;
    #endif
    #ifndef CO_VARKEYWORDS
    static int CO_VARKEYWORDS;
    #endif
    #ifndef CO_ASYNC_GENERATOR
    static int CO_ASYNC_GENERATOR;
    #endif
    #ifndef CO_GENERATOR
    static int CO_GENERATOR;
    #endif
    #ifndef CO_COROUTINE
    static int CO_COROUTINE;
    #endif
#else
    #ifndef CO_COROUTINE
      #define CO_COROUTINE 0x80
    #endif
    #ifndef CO_ASYNC_GENERATOR
      #define CO_ASYNC_GENERATOR 0x200
    #endif
#endif
static int __Pyx_init_co_variables(void);
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef Py_TPFLAGS_IMMUTABLETYPE
  #define Py_TPFLAGS_IMMUTABLETYPE (1UL << 8)
#endif
#ifndef Py_TPFLAGS_DISALLOW_INSTANTIATION
  #define Py_TPFLAGS_DISALLOW_INSTANTIATION (1UL << 7)
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#ifndef METH_FASTCALL
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void (*cfunc)(void)) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if PY_VERSION_HEX < 0x03090000 || (CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000)
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#elif CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) GraalPyFrame_SetLineNumber((frame), (lineno))
#elif CYTHON_COMPILING_IN_GRAAL
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) _PyFrame_SetLineNumber((frame), (lineno))
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#endif
#if CYTHON_USE_MODULE_STATE
static CYTHON_INLINE void *__Pyx__PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#define __Pyx_PyModule_GetState(o) (__pyx_mstatetype *)__Pyx__PyModule_GetState(o)
#else
#define __Pyx_PyModule_GetState(op) ((void)op,__pyx_mstate_global)
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE((PyObject *) obj), name, func_ctype)
#define __Pyx_PyObject_TryGetSlot(obj, name, func_ctype) __Pyx_PyType_TryGetSlot(Py_TYPE(obj), name, func_ctype)
#define __Pyx_PyObject_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#define __Pyx_PyObject_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype) __Pyx_PyType_GetSlot(type, name, func_ctype)
  #define __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype) (((type)->sub) ? ((type)->sub->name) : NULL)
  #define __Pyx_PyType_TryGetSubSlot(type, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype)
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype)\
    ((__PYX_LIMITED_VERSION_HEX >= 0x030A0000 ||\
     (PyType_GetFlags(type) & Py_TPFLAGS_HEAPTYPE) || __Pyx_get_runtime_version() >= 0x030A0000) ?\
     __Pyx_PyType_GetSlot(type, name, func_ctype) : NULL)
  #define __Pyx_PyType_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSlot(obj, name, func_ctype)
  #define __Pyx_PyType_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSlot(obj, name, func_ctype)
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
#define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
#endif
#define __Pyx_PyObject_GetIterNextFunc(iterator)  __Pyx_PyObject_GetSlot(iterator, tp_iternext, iternextfunc)
#if CYTHON_USE_TYPE_SPECS
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#else
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains)
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && PyUnstable_Object_IsUniquelyReferenced(obj)) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#elif CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
enum __Pyx_ReferenceSharing {
  __Pyx_ReferenceSharing_DefinitelyUnique, // We created it so we know it's unshared - no need to check
  __Pyx_ReferenceSharing_OwnStrongReference,
  __Pyx_ReferenceSharing_FunctionArgument,
  __Pyx_ReferenceSharing_SharedReference, // Never trust it to be unshared because it's a global or similar
};
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && PY_VERSION_HEX >= 0x030E0000
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing)\
    (sharing == __Pyx_ReferenceSharing_DefinitelyUnique ? 1 :\
      (sharing == __Pyx_ReferenceSharing_FunctionArgument ? PyUnstable_Object_IsUniqueReferencedTemporary(o) :\
      (sharing == __Pyx_ReferenceSharing_OwnStrongReference ? PyUnstable_Object_IsUniquelyReferenced(o) : 0)))
#elif (CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING) || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)sharing), Py_REFCNT(o) == 1)
#else
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)o), ((void)sharing), 0)
#endif
#if CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyList_GetItemRef(o, i) (likely((i) >= 0) ? PySequence_GetItem(o, i) : (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
  #else
    #define __Pyx_PyList_GetItemRef(o, i) PySequence_ITEM(o, i)
  #endif
#elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #else
    #define __Pyx_PyList_GetItemRef(o, i) __Pyx_XNewRef(PyList_GetItem(o, i))
  #endif
#else
  #define __Pyx_PyList_GetItemRef(o, i) __Pyx_NewRef(PyList_GET_ITEM(o, i))
#endif
#if CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS && !CYTHON_COMPILING_IN_LIMITED_API && CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) (__Pyx_IS_UNIQUELY_REFERENCED(o, unsafe_shared) ?\
    __Pyx_NewRef(PyList_GET_ITEM(o, i)) : __Pyx_PyList_GetItemRef(o, i))
#else
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) __Pyx_PyList_GetItemRef(o, i)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyDict_GetItemRef(dict, key, result) PyDict_GetItemRef(dict, key, result)
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyObject_GetItem(dict, key);
  if (*result == NULL) {
    if (PyErr_ExceptionMatches(PyExc_KeyError)) {
      PyErr_Clear();
      return 0;
    }
    return -1;
  }
  return 1;
}
#else
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyDict_GetItemWithError(dict, key);
  if (*result == NULL) {
    return PyErr_Occurred() ? -1 : 0;
  }
  Py_INCREF(*result);
  return 1;
}
#endif
#if defined(CYTHON_DEBUG_VISIT_CONST) && CYTHON_DEBUG_VISIT_CONST
  #define __Pyx_VISIT_CONST(obj)  Py_VISIT(obj)
#else
  #define __Pyx_VISIT_CONST(obj)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GET_ITEM(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GET_ITEM(o, i)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GetItem(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GetItem(o, i)
#endif
#if CYTHON_ASSUME_SAFE_SIZE
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GET_LENGTH(o)
#else
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GetLength(o)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_InternFromString)
  #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
#endif
#define __Pyx_PyLong_FromHash_t PyLong_FromSsize_t
#define __Pyx_PyLong_AsHash_t   __Pyx_PyIndex_AsSsize_t
#if __PYX_LIMITED_VERSION_HEX >= 0x030A0000
    #define __Pyx_PySendResult PySendResult
#else
    typedef enum {
        PYGEN_RETURN = 0,
        PYGEN_ERROR = -1,
        PYGEN_NEXT = 1,
    } __Pyx_PySendResult;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030A00A3
  typedef __Pyx_PySendResult (*__Pyx_pyiter_sendfunc)(PyObject *iter, PyObject *value, PyObject **result);
#else
  #define __Pyx_pyiter_sendfunc sendfunc
#endif
#if !CYTHON_USE_AM_SEND
#define __PYX_HAS_PY_AM_SEND 0
#elif __PYX_LIMITED_VERSION_HEX >= 0x030A0000
#define __PYX_HAS_PY_AM_SEND 1
#else
#define __PYX_HAS_PY_AM_SEND 2  // our own backported implementation
#endif
#if __PYX_HAS_PY_AM_SEND < 2
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
        __Pyx_pyiter_sendfunc am_send;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_SlotTpAsAsync(s) ((PyAsyncMethods*)(s))
#endif
#if CYTHON_USE_AM_SEND && PY_VERSION_HEX < 0x030A00F0
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (1UL << 21)
#else
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (0)
#endif
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyInterpreterState_Get() PyInterpreterState_Get()
#else
#define __Pyx_PyInterpreterState_Get() PyThreadState_Get()->interp
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030A0000
#ifdef __cplusplus
extern "C"
#endif
PyAPI_FUNC(void *) PyMem_Calloc(size_t nelem, size_t elsize);
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_init_co_variable(PyObject *inspect, const char* name, int *write_to) {
    int value;
    PyObject *py_value = PyObject_GetAttrString(inspect, name);
    if (!py_value) return 0;
    value = (int) PyLong_AsLong(py_value);
    Py_DECREF(py_value);
    *write_to = value;
    return value != -1 || !PyErr_Occurred();
}
static int __Pyx_init_co_variables(void) {
    PyObject *inspect;
    int result;
    inspect = PyImport_ImportModule("inspect");
    result =
#if !defined(CO_OPTIMIZED)
        __Pyx_init_co_variable(inspect, "CO_OPTIMIZED", &CO_OPTIMIZED) &&
#endif
#if !defined(CO_NEWLOCALS)
        __Pyx_init_co_variable(inspect, "CO_NEWLOCALS", &CO_NEWLOCALS) &&
#endif
#if !defined(CO_VARARGS)
        __Pyx_init_co_variable(inspect, "CO_VARARGS", &CO_VARARGS) &&
#endif
#if !defined(CO_VARKEYWORDS)
        __Pyx_init_co_variable(inspect, "CO_VARKEYWORDS", &CO_VARKEYWORDS) &&
#endif
#if !defined(CO_ASYNC_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_ASYNC_GENERATOR", &CO_ASYNC_GENERATOR) &&
#endif
#if !defined(CO_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_GENERATOR", &CO_GENERATOR) &&
#endif
#if !defined(CO_COROUTINE)
        __Pyx_init_co_variable(inspect, "CO_COROUTINE", &CO_COROUTINE) &&
#endif
        1;
    Py_DECREF(inspect);
    return result ? 0 : -1;
}
#else
static int __Pyx_init_co_variables(void) {
    return 0;  // It's a limited API-only feature
}
#endif

/* MathInitCode */
#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #ifndef _USE_MATH_DEFINES
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#ifndef CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#define CYTHON_CLINE_IN_TRACEBACK_RUNTIME 0
#endif
#ifndef CYTHON_CLINE_IN_TRACEBACK
#define CYTHON_CLINE_IN_TRACEBACK CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#endif
#if CYTHON_CLINE_IN_TRACEBACK
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; __pyx_clineno = __LINE__; (void) __pyx_clineno; }
#else
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; (void) __pyx_clineno; }
#endif
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__pyverbs__qp
#define __PYX_HAVE_API__pyverbs__qp
/* Early includes */
#include <stdint.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/time.h>
#include <infiniband/verbs.h>
#include <infiniband/verbs_api.h>
#include <infiniband/driver.h>
#include <infiniband/tm_types.h>
#include "infiniband/verbs.h"
#include "infiniband/driver.h"
#include <string.h>
#include <stdlib.h>
#include <rdma/rdma_cma.h>
#include <infiniband/ib.h>
#include <rdma/rdma_verbs.h>
#include <stdio.h>
#include "endian.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifdef CYTHON_FREETHREADING_COMPATIBLE
#if CYTHON_FREETHREADING_COMPATIBLE
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_NOT_USED
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AS_STRING(s)
#else
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AsString(s)
#endif
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
static CYTHON_INLINE PyObject *__Pyx_NewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_NewRef)
    return Py_NewRef(obj);
#else
    Py_INCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_XNewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_XNewRef)
    return Py_XNewRef(obj);
#else
    Py_XINCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b);
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __Pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AS_DOUBLE(x)
#else
#define __Pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AsDouble(x)
#endif
#define __Pyx_PyFloat_AsFloat(x) ((float) __Pyx_PyFloat_AsDouble(x))
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#elif __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeASCII(c_str, size, NULL)
#else
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
/* PretendToInitialize */
#ifdef __cplusplus
#if __cplusplus > 201103L
#include <type_traits>
#endif
template <typename T>
static void __Pyx_pretend_to_initialize(T* ptr) {
#if __cplusplus > 201103L
    if ((std::is_trivially_default_constructible<T>::value))
#endif
        *ptr = T();
    (void)ptr;
}
#else
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }
#endif


#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * const __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char* const __pyx_f[] = {
  "pyverbs/qp.pyx",
  "EnumTypeToPy",
  "pyverbs/device.pxd",
  "pyverbs/cq.pxd",
  "pyverbs/pd.pxd",
  "pyverbs/wq.pxd",
  "pyverbs/mr.pxd",
  "pyverbs/wr.pxd",
  "cmid.pxd",
  "pyverbs/addr.pxd",
  "pyverbs/flow.pxd",
  "cpython/type.pxd",
  "pyverbs/xrcd.pxd",
  "pyverbs/srq.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* Atomics.proto (used by UnpackUnboundCMethod) */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __PYX_GET_CYTHON_COMPILING_IN_CPYTHON_FREETHREADING() CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __pyx_atomic_int_type int
#define __pyx_nonatomic_int_type int
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__))
    #include <stdatomic.h>
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)))
    #include <atomic>
#endif
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__) &&\
                       ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type atomic_int
    #define __pyx_atomic_ptr_type atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) atomic_fetch_add_explicit(value, 1, memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) atomic_fetch_add_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) atomic_fetch_sub_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) atomic_load(value)
    #define __pyx_atomic_store(value, new_value) atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) atomic_load_explicit(value, memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) atomic_load_explicit(value, memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C atomics"
    #endif
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)) &&\
                    ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type std::atomic_int
    #define __pyx_atomic_ptr_type std::atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) std::atomic_fetch_sub_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) std::atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) std::atomic_load(value)
    #define __pyx_atomic_store(value, new_value) std::atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) std::atomic_load_explicit(value, std::memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) std::atomic_load_explicit(value, std::memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) std::atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C++ atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C++ atomics"
    #endif
#elif CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_ptr_type void*
    #define __pyx_nonatomic_ptr_type void*
    #define __pyx_atomic_incr_relaxed(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) __sync_fetch_and_sub(value, 1)
    #define __pyx_atomic_sub(value, arg) __sync_fetch_and_sub(value, arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_store(value, new_value) __sync_lock_test_and_set(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_load_acquire(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) __sync_lock_test_and_set(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_nonatomic_ptr_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER)
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #define __pyx_atomic_ptr_type void*
    #undef __pyx_nonatomic_int_type
    #define __pyx_nonatomic_int_type long
    #define __pyx_nonatomic_ptr_type void*
    #pragma intrinsic (_InterlockedExchangeAdd, _InterlockedExchange, _InterlockedCompareExchange, _InterlockedCompareExchangePointer, _InterlockedExchangePointer)
    #define __pyx_atomic_incr_relaxed(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) _InterlockedExchangeAdd(value, -1)
    #define __pyx_atomic_sub(value, arg) _InterlockedExchangeAdd(value, -arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = _InterlockedCompareExchange(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) _InterlockedExchangeAdd(value, 0)
    #define __pyx_atomic_store(value, new_value) _InterlockedExchange(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) *(void * volatile *)value
    #define __pyx_atomic_pointer_load_acquire(value) _InterlockedCompareExchangePointer(value, 0, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) _InterlockedExchangePointer(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_atomic_ptr_type old = _InterlockedCompareExchangePointer(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif

/* CriticalSectionsDefinition.proto (used by CriticalSections) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection void*
#define __Pyx_PyCriticalSection2 void*
#define __Pyx_PyCriticalSection_End(cs)
#define __Pyx_PyCriticalSection2_End(cs)
#else
#define __Pyx_PyCriticalSection PyCriticalSection
#define __Pyx_PyCriticalSection2 PyCriticalSection2
#define __Pyx_PyCriticalSection_End PyCriticalSection_End
#define __Pyx_PyCriticalSection2_End PyCriticalSection2_End
#endif

/* CriticalSections.proto (used by ParseKeywordsImpl) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection_Begin(cs, arg) (void)(cs)
#define __Pyx_PyCriticalSection2_Begin(cs, arg1, arg2) (void)(cs)
#else
#define __Pyx_PyCriticalSection_Begin PyCriticalSection_Begin
#define __Pyx_PyCriticalSection2_Begin PyCriticalSection2_Begin
#endif
#if PY_VERSION_HEX < 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_BEGIN_CRITICAL_SECTION(o) {
#define __Pyx_END_CRITICAL_SECTION() }
#else
#define __Pyx_BEGIN_CRITICAL_SECTION Py_BEGIN_CRITICAL_SECTION
#define __Pyx_END_CRITICAL_SECTION Py_END_CRITICAL_SECTION
#endif

/* IncludeStructmemberH.proto (used by FixUpExtensionType) */
#include <structmember.h>

/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_7pyverbs_4base_PyverbsObject;
struct __pyx_obj_7pyverbs_4base_PyverbsCM;
struct __pyx_obj_7pyverbs_6device_Context;
struct __pyx_obj_7pyverbs_6device_DeviceAttr;
struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput;
struct __pyx_obj_7pyverbs_6device_ODPCaps;
struct __pyx_obj_7pyverbs_6device_RSSCaps;
struct __pyx_obj_7pyverbs_6device_PacketPacingCaps;
struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps;
struct __pyx_obj_7pyverbs_6device_TMCaps;
struct __pyx_obj_7pyverbs_6device_CQModerationCaps;
struct __pyx_obj_7pyverbs_6device_TSOCaps;
struct __pyx_obj_7pyverbs_6device_DeviceAttrEx;
struct __pyx_obj_7pyverbs_6device_AllocDmAttr;
struct __pyx_obj_7pyverbs_6device_DM;
struct __pyx_obj_7pyverbs_6device_PortAttr;
struct __pyx_obj_7pyverbs_6device_GIDEntry;
struct __pyx_obj_7pyverbs_6device_AsyncEvent;
struct __pyx_obj_7pyverbs_6device_FdArr;
struct __pyx_obj_7pyverbs_2cq_CompChannel;
struct __pyx_obj_7pyverbs_2cq_CQ;
struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx;
struct __pyx_obj_7pyverbs_2cq_CQEX;
struct __pyx_obj_7pyverbs_2cq_WC;
struct __pyx_obj_7pyverbs_2cq_PollCqAttr;
struct __pyx_obj_7pyverbs_2cq_WcTmInfo;
struct __pyx_obj_7pyverbs_2pd_PD;
struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr;
struct __pyx_obj_7pyverbs_2pd_ParentDomain;
struct __pyx_obj_7pyverbs_2pd_ParentDomainContext;
struct __pyx_obj_7pyverbs_2wq_WQInitAttr;
struct __pyx_obj_7pyverbs_2wq_WQAttr;
struct __pyx_obj_7pyverbs_2wq_WQ;
struct __pyx_obj_7pyverbs_2wq_RwqIndTableInitAttr;
struct __pyx_obj_7pyverbs_2wq_RwqIndTable;
struct __pyx_obj_7pyverbs_2wq_RxHashConf;
struct __pyx_obj_7pyverbs_2mr_MR;
struct __pyx_obj_7pyverbs_2mr_MREx;
struct __pyx_obj_7pyverbs_2mr_MWBindInfo;
struct __pyx_obj_7pyverbs_2mr_MWBind;
struct __pyx_obj_7pyverbs_2mr_MW;
struct __pyx_obj_7pyverbs_2mr_DMMR;
struct __pyx_obj_7pyverbs_2mr_DmaBufMR;
struct __pyx_obj_7pyverbs_2mr_DmaHandleInitAttr;
struct __pyx_obj_7pyverbs_2mr_DMAHandle;
struct __pyx_obj_7pyverbs_2wr_SGE;
struct __pyx_obj_7pyverbs_2wr_RecvWR;
struct __pyx_obj_7pyverbs_2wr_SendWR;
struct __pyx_obj_7pyverbs_4cmid_CMID;
struct __pyx_obj_7pyverbs_4cmid_CMEventChannel;
struct __pyx_obj_7pyverbs_4cmid_CMEvent;
struct __pyx_obj_7pyverbs_4cmid_AddrInfo;
struct __pyx_obj_7pyverbs_4cmid_ConnParam;
struct __pyx_obj_7pyverbs_4cmid_UDParam;
struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx;
struct __pyx_obj_7pyverbs_4addr_GID;
struct __pyx_obj_7pyverbs_4addr_GRH;
struct __pyx_obj_7pyverbs_4addr_GlobalRoute;
struct __pyx_obj_7pyverbs_4addr_AHAttr;
struct __pyx_obj_7pyverbs_4addr_AH;
struct __pyx_obj_7pyverbs_4flow_FlowAttr;
struct __pyx_obj_7pyverbs_4flow_Flow;
struct __pyx_obj_7pyverbs_4flow_FlowAction;
struct __pyx_obj_7pyverbs_4xrcd_XRCDInitAttr;
struct __pyx_obj_7pyverbs_4xrcd_XRCD;
struct __pyx_obj_7pyverbs_3srq_SrqAttr;
struct __pyx_obj_7pyverbs_3srq_SrqInitAttr;
struct __pyx_obj_7pyverbs_3srq_SrqInitAttrEx;
struct __pyx_obj_7pyverbs_3srq_OpsWr;
struct __pyx_obj_7pyverbs_3srq_SRQ;
struct __pyx_obj_7pyverbs_2qp_QPCap;
struct __pyx_obj_7pyverbs_2qp_QPInitAttr;
struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx;
struct __pyx_obj_7pyverbs_2qp_QPAttr;
struct __pyx_obj_7pyverbs_2qp_QP;
struct __pyx_obj_7pyverbs_2qp_DataBuffer;
struct __pyx_obj_7pyverbs_2qp_QPEx;
struct __pyx_obj_7pyverbs_2qp_ECE;

/* "pyverbs/base.pxd":6
 * #cython: language_level=3
 * 
 * cdef class PyverbsObject(object):             # <<<<<<<<<<<<<<
 *     cdef object __weakref__
 *     cdef object logger
*/
struct __pyx_obj_7pyverbs_4base_PyverbsObject {
  PyObject_HEAD
  PyObject *__weakref__;
  PyObject *logger;
};


/* "pyverbs/base.pxd":10
 *     cdef object logger
 * 
 * cdef class PyverbsCM(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cpdef close(self)
 * 
*/
struct __pyx_obj_7pyverbs_4base_PyverbsCM {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM *__pyx_vtab;
};


/* "pyverbs/device.pxd":10
 * 
 * 
 * cdef class Context(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_context *context
 *     cdef v.ibv_device *device
*/
struct __pyx_obj_7pyverbs_6device_Context {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_context *context;
  struct ibv_device *device;
  PyObject *name;
  PyObject *pds;
  PyObject *dms;
  PyObject *ccs;
  PyObject *cqs;
  PyObject *qps;
  PyObject *xrcds;
  PyObject *vars;
  PyObject *uars;
  PyObject *pps;
  PyObject *sched_nodes;
  PyObject *sched_leafs;
  PyObject *dr_domains;
  PyObject *wqs;
  PyObject *rwq_ind_tbls;
  PyObject *crypto_logins;
  PyObject *event_channels;
  PyObject *dmahs;
};


/* "pyverbs/device.pxd":33
 *     cdef object dmahs
 * 
 * cdef class DeviceAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_device_attr dev_attr
 * 
*/
struct __pyx_obj_7pyverbs_6device_DeviceAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_device_attr dev_attr;
};


/* "pyverbs/device.pxd":36
 *     cdef v.ibv_device_attr dev_attr
 * 
 * cdef class QueryDeviceExInput(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_query_device_ex_input input
 * 
*/
struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_query_device_ex_input input;
};


/* "pyverbs/device.pxd":39
 *     cdef v.ibv_query_device_ex_input input
 * 
 * cdef class ODPCaps(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_odp_caps odp_caps
 *     cdef object xrc_odp_caps
*/
struct __pyx_obj_7pyverbs_6device_ODPCaps {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_odp_caps odp_caps;
  PyObject *xrc_odp_caps;
};


/* "pyverbs/device.pxd":43
 *     cdef object xrc_odp_caps
 * 
 * cdef class RSSCaps(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_rss_caps rss_caps
 * 
*/
struct __pyx_obj_7pyverbs_6device_RSSCaps {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_rss_caps rss_caps;
};


/* "pyverbs/device.pxd":46
 *     cdef v.ibv_rss_caps rss_caps
 * 
 * cdef class PacketPacingCaps(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_packet_pacing_caps packet_pacing_caps
 * 
*/
struct __pyx_obj_7pyverbs_6device_PacketPacingCaps {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_packet_pacing_caps packet_pacing_caps;
};


/* "pyverbs/device.pxd":49
 *     cdef v.ibv_packet_pacing_caps packet_pacing_caps
 * 
 * cdef class PCIAtomicCaps(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_pci_atomic_caps caps
 * 
*/
struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_pci_atomic_caps caps;
};


/* "pyverbs/device.pxd":52
 *     cdef v.ibv_pci_atomic_caps caps
 * 
 * cdef class TMCaps(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_tm_caps tm_caps
 * 
*/
struct __pyx_obj_7pyverbs_6device_TMCaps {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_tm_caps tm_caps;
};


/* "pyverbs/device.pxd":55
 *     cdef v.ibv_tm_caps tm_caps
 * 
 * cdef class CQModerationCaps(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_cq_moderation_caps cq_mod_caps
 * 
*/
struct __pyx_obj_7pyverbs_6device_CQModerationCaps {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_cq_moderation_caps cq_mod_caps;
};


/* "pyverbs/device.pxd":58
 *     cdef v.ibv_cq_moderation_caps cq_mod_caps
 * 
 * cdef class TSOCaps(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_tso_caps tso_caps
 * 
*/
struct __pyx_obj_7pyverbs_6device_TSOCaps {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_tso_caps tso_caps;
};


/* "pyverbs/device.pxd":61
 *     cdef v.ibv_tso_caps tso_caps
 * 
 * cdef class DeviceAttrEx(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_device_attr_ex dev_attr
 * 
*/
struct __pyx_obj_7pyverbs_6device_DeviceAttrEx {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_device_attr_ex dev_attr;
};


/* "pyverbs/device.pxd":64
 *     cdef v.ibv_device_attr_ex dev_attr
 * 
 * cdef class AllocDmAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_alloc_dm_attr alloc_dm_attr
 * 
*/
struct __pyx_obj_7pyverbs_6device_AllocDmAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_alloc_dm_attr alloc_dm_attr;
};


/* "pyverbs/device.pxd":67
 *     cdef v.ibv_alloc_dm_attr alloc_dm_attr
 * 
 * cdef class DM(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_dm *dm
 *     cdef object dm_mrs
*/
struct __pyx_obj_7pyverbs_6device_DM {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_dm *dm;
  PyObject *dm_mrs;
  PyObject *context;
  PyObject *_is_imported;
};


/* "pyverbs/device.pxd":74
 *     cdef add_ref(self, obj)
 * 
 * cdef class PortAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_port_attr attr
 * 
*/
struct __pyx_obj_7pyverbs_6device_PortAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_port_attr attr;
};


/* "pyverbs/device.pxd":77
 *     cdef v.ibv_port_attr attr
 * 
 * cdef class GIDEntry(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_gid_entry entry
 * 
*/
struct __pyx_obj_7pyverbs_6device_GIDEntry {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_gid_entry entry;
};


/* "pyverbs/device.pxd":80
 *     cdef v.ibv_gid_entry entry
 * 
 * cdef class AsyncEvent(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_async_event event
 * 
*/
struct __pyx_obj_7pyverbs_6device_AsyncEvent {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_async_event event;
};


/* "pyverbs/device.pxd":83
 *     cdef v.ibv_async_event event
 * 
 * cdef class FdArr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_fd_arr attr
 *     cpdef close(self)
*/
struct __pyx_obj_7pyverbs_6device_FdArr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct __pyx_vtabstruct_7pyverbs_6device_FdArr *__pyx_vtab;
  struct ibv_fd_arr attr;
};


/* "pyverbs/cq.pxd":9
 * cimport pyverbs.libibverbs as v
 * 
 * cdef class CompChannel(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_comp_channel *cc
 *     cpdef close(self)
*/
struct __pyx_obj_7pyverbs_2cq_CompChannel {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_comp_channel *cc;
  PyObject *context;
  PyObject *cqs;
};


/* "pyverbs/cq.pxd":16
 *     cdef object cqs
 * 
 * cdef class CQ(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_cq *cq
 *     cpdef close(self)
*/
struct __pyx_obj_7pyverbs_2cq_CQ {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_cq *cq;
  PyObject *context;
  PyObject *qps;
  PyObject *srqs;
  PyObject *channel;
  PyObject *num_events;
  PyObject *wqs;
};


/* "pyverbs/cq.pxd":27
 *     cdef object wqs
 * 
 * cdef class CqInitAttrEx(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_cq_init_attr_ex attr
 *     cdef object channel
*/
struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_cq_init_attr_ex attr;
  PyObject *channel;
  PyObject *parent_domain;
};


/* "pyverbs/cq.pxd":32
 *     cdef object parent_domain
 * 
 * cdef class CQEX(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_cq_ex *cq
 *     cdef v.ibv_cq *ibv_cq
*/
struct __pyx_obj_7pyverbs_2cq_CQEX {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_cq_ex *cq;
  struct ibv_cq *ibv_cq;
  PyObject *context;
  PyObject *qps;
  PyObject *srqs;
  PyObject *wqs;
};


/* "pyverbs/cq.pxd":42
 *     cdef object wqs
 * 
 * cdef class WC(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_wc wc
 * 
*/
struct __pyx_obj_7pyverbs_2cq_WC {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_wc wc;
};


/* "pyverbs/cq.pxd":45
 *     cdef v.ibv_wc wc
 * 
 * cdef class PollCqAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_poll_cq_attr attr
 * 
*/
struct __pyx_obj_7pyverbs_2cq_PollCqAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_poll_cq_attr attr;
};


/* "pyverbs/cq.pxd":48
 *     cdef v.ibv_poll_cq_attr attr
 * 
 * cdef class WcTmInfo(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_wc_tm_info info
*/
struct __pyx_obj_7pyverbs_2cq_WcTmInfo {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_wc_tm_info info;
};


/* "pyverbs/pd.pxd":12
 * 
 * 
 * cdef class PD(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_pd *pd
 *     cdef Context ctx
*/
struct __pyx_obj_7pyverbs_2pd_PD {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_pd *pd;
  struct __pyx_obj_7pyverbs_6device_Context *ctx;
  PyObject *srqs;
  PyObject *mrs;
  PyObject *mws;
  PyObject *ahs;
  PyObject *qps;
  PyObject *parent_domains;
  PyObject *wqs;
  PyObject *mkeys;
  PyObject *deks;
  PyObject *_is_imported;
};


/* "pyverbs/pd.pxd":28
 *     cdef object _is_imported
 * 
 * cdef class ParentDomainInitAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_parent_domain_init_attr init_attr
 *     cdef object pd
*/
struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_parent_domain_init_attr init_attr;
  PyObject *pd;
  PyObject *alloc;
  PyObject *dealloc;
};


/* "pyverbs/pd.pxd":34
 *     cdef object dealloc
 * 
 * cdef class ParentDomain(PD):             # <<<<<<<<<<<<<<
 *     cdef add_ref(self, obj)
 *     cdef object protection_domain
*/
struct __pyx_obj_7pyverbs_2pd_ParentDomain {
  struct __pyx_obj_7pyverbs_2pd_PD __pyx_base;
  PyObject *protection_domain;
  PyObject *cqs;
};


/* "pyverbs/pd.pxd":39
 *     cdef object cqs
 * 
 * cdef class ParentDomainContext(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef object p_alloc
 *     cdef object p_free
*/
struct __pyx_obj_7pyverbs_2pd_ParentDomainContext {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  PyObject *p_alloc;
  PyObject *p_free;
  PyObject *pd;
  PyObject *user_data;
};


/* "pyverbs/wq.pxd":13
 * 
 * 
 * cdef class WQInitAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_wq_init_attr attr
 *     cdef PD pd
*/
struct __pyx_obj_7pyverbs_2wq_WQInitAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_wq_init_attr attr;
  struct __pyx_obj_7pyverbs_2pd_PD *pd;
  PyObject *cq;
};


/* "pyverbs/wq.pxd":18
 *     cdef object cq
 * 
 * cdef class WQAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_wq_attr attr
 * 
*/
struct __pyx_obj_7pyverbs_2wq_WQAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_wq_attr attr;
};


/* "pyverbs/wq.pxd":21
 *     cdef v.ibv_wq_attr attr
 * 
 * cdef class WQ(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_wq *wq
 *     cdef Context context
*/
struct __pyx_obj_7pyverbs_2wq_WQ {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_wq *wq;
  struct __pyx_obj_7pyverbs_6device_Context *context;
  struct __pyx_obj_7pyverbs_2pd_PD *pd;
  PyObject *cq;
  PyObject *rwq_ind_tables;
};


/* "pyverbs/wq.pxd":29
 *     cpdef add_ref(self, obj)
 * 
 * cdef class RwqIndTableInitAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_rwq_ind_table_init_attr attr
 *     cdef object wqs_list
*/
struct __pyx_obj_7pyverbs_2wq_RwqIndTableInitAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_rwq_ind_table_init_attr attr;
  PyObject *wqs_list;
};


/* "pyverbs/wq.pxd":33
 *     cdef object wqs_list
 * 
 * cdef class RwqIndTable(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_rwq_ind_table *rwq_ind_table
 *     cdef Context context
*/
struct __pyx_obj_7pyverbs_2wq_RwqIndTable {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_rwq_ind_table *rwq_ind_table;
  struct __pyx_obj_7pyverbs_6device_Context *context;
  PyObject *wqs;
  PyObject *qps;
};


/* "pyverbs/wq.pxd":40
 *     cpdef add_ref(self, obj)
 * 
 * cdef class RxHashConf(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_rx_hash_conf rx_hash_conf
*/
struct __pyx_obj_7pyverbs_2wq_RxHashConf {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_rx_hash_conf rx_hash_conf;
};


/* "pyverbs/mr.pxd":12
 * 
 * 
 * cdef class MR(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef object pd
 *     cdef object cmid
*/
struct __pyx_obj_7pyverbs_2mr_MR {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *pd;
  PyObject *cmid;
  struct ibv_mr *mr;
  int mmap_length;
  PyObject *is_huge;
  PyObject *is_user_addr;
  void *buf;
  PyObject *_is_imported;
};


/* "pyverbs/mr.pxd":25
 *     cpdef read(self, length, offset)
 * 
 * cdef class MREx(MR):             # <<<<<<<<<<<<<<
 *     cdef object dmah
 * 
*/
struct __pyx_obj_7pyverbs_2mr_MREx {
  struct __pyx_obj_7pyverbs_2mr_MR __pyx_base;
  PyObject *dmah;
};


/* "pyverbs/mr.pxd":28
 *     cdef object dmah
 * 
 * cdef class MWBindInfo(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_mw_bind_info info
 *     cdef object mr
*/
struct __pyx_obj_7pyverbs_2mr_MWBindInfo {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_mw_bind_info info;
  PyObject *mr;
};


/* "pyverbs/mr.pxd":32
 *     cdef object mr
 * 
 * cdef class MWBind(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_mw_bind mw_bind
 *     cdef object mr
*/
struct __pyx_obj_7pyverbs_2mr_MWBind {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_mw_bind mw_bind;
  PyObject *mr;
};


/* "pyverbs/mr.pxd":36
 *     cdef object mr
 * 
 * cdef class MW(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef object pd
 *     cdef v.ibv_mw *mw
*/
struct __pyx_obj_7pyverbs_2mr_MW {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *pd;
  struct ibv_mw *mw;
};


/* "pyverbs/mr.pxd":40
 *     cdef v.ibv_mw *mw
 * 
 * cdef class DMMR(MR):             # <<<<<<<<<<<<<<
 *     cdef object dm
 * 
*/
struct __pyx_obj_7pyverbs_2mr_DMMR {
  struct __pyx_obj_7pyverbs_2mr_MR __pyx_base;
  PyObject *dm;
};


/* "pyverbs/mr.pxd":43
 *     cdef object dm
 * 
 * cdef class DmaBufMR(MR):             # <<<<<<<<<<<<<<
 *     cdef object dmabuf
 *     cdef unsigned long offset
*/
struct __pyx_obj_7pyverbs_2mr_DmaBufMR {
  struct __pyx_obj_7pyverbs_2mr_MR __pyx_base;
  PyObject *dmabuf;
  unsigned long offset;
  PyObject *is_dmabuf_internal;
};


/* "pyverbs/mr.pxd":48
 *     cdef object is_dmabuf_internal
 * 
 * cdef class DmaHandleInitAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_dmah_init_attr init_attr
 * 
*/
struct __pyx_obj_7pyverbs_2mr_DmaHandleInitAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_dmah_init_attr init_attr;
};


/* "pyverbs/mr.pxd":51
 *     cdef v.ibv_dmah_init_attr init_attr
 * 
 * cdef class DMAHandle(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_dmah *dmah
 *     cdef object mrs
*/
struct __pyx_obj_7pyverbs_2mr_DMAHandle {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_dmah *dmah;
  PyObject *mrs;
  PyObject *ctx;
};


/* "pyverbs/wr.pxd":10
 * 
 * 
 * cdef class SGE(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_sge *sge
 *     cpdef read(self, length, offset)
*/
struct __pyx_obj_7pyverbs_2wr_SGE {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_sge *sge;
};


/* "pyverbs/wr.pxd":14
 *     cpdef read(self, length, offset)
 * 
 * cdef class RecvWR(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_recv_wr recv_wr
 * 
*/
struct __pyx_obj_7pyverbs_2wr_RecvWR {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_recv_wr recv_wr;
};


/* "pyverbs/wr.pxd":17
 *     cdef v.ibv_recv_wr recv_wr
 * 
 * cdef class SendWR(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_send_wr send_wr
 *     cdef object ah
*/
struct __pyx_obj_7pyverbs_2wr_SendWR {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_send_wr send_wr;
  PyObject *ah;
};


/* "cmid.pxd":10
 * 
 * 
 * cdef class CMID(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef cm.rdma_cm_id *id
 *     cdef object event_channel
*/
struct __pyx_obj_7pyverbs_4cmid_CMID {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct rdma_cm_id *id;
  PyObject *event_channel;
  PyObject *ctx;
  PyObject *pd;
  PyObject *mrs;
};


/* "cmid.pxd":20
 * 
 * 
 * cdef class CMEventChannel(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef cm.rdma_event_channel *event_channel
 *     cpdef close(self)
*/
struct __pyx_obj_7pyverbs_4cmid_CMEventChannel {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct __pyx_vtabstruct_7pyverbs_4cmid_CMEventChannel *__pyx_vtab;
  struct rdma_event_channel *event_channel;
};


/* "cmid.pxd":25
 * 
 * 
 * cdef class CMEvent(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef cm.rdma_cm_event *event
 *     cpdef close(self)
*/
struct __pyx_obj_7pyverbs_4cmid_CMEvent {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct __pyx_vtabstruct_7pyverbs_4cmid_CMEvent *__pyx_vtab;
  struct rdma_cm_event *event;
};


/* "cmid.pxd":30
 * 
 * 
 * cdef class AddrInfo(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef cm.rdma_addrinfo *addr_info
 *     cpdef close(self)
*/
struct __pyx_obj_7pyverbs_4cmid_AddrInfo {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct __pyx_vtabstruct_7pyverbs_4cmid_AddrInfo *__pyx_vtab;
  struct rdma_addrinfo *addr_info;
};


/* "cmid.pxd":35
 * 
 * 
 * cdef class ConnParam(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef cm.rdma_conn_param conn_param
 *     cdef object data
*/
struct __pyx_obj_7pyverbs_4cmid_ConnParam {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct rdma_conn_param conn_param;
  PyObject *data;
};


/* "cmid.pxd":40
 * 
 * 
 * cdef class UDParam(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef cm.rdma_ud_param ud_param
 * 
*/
struct __pyx_obj_7pyverbs_4cmid_UDParam {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct rdma_ud_param ud_param;
};


/* "cmid.pxd":43
 *     cdef cm.rdma_ud_param ud_param
 * 
 * cdef class JoinMCAttrEx(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef cm.rdma_cm_join_mc_attr_ex join_mc_attr_ex
*/
struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct rdma_cm_join_mc_attr_ex join_mc_attr_ex;
};


/* "pyverbs/addr.pxd":11
 * 
 * 
 * cdef class GID(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_gid gid
 * 
*/
struct __pyx_obj_7pyverbs_4addr_GID {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  union ibv_gid gid;
};


/* "pyverbs/addr.pxd":14
 *     cdef v.ibv_gid gid
 * 
 * cdef class GRH(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_grh grh
 * 
*/
struct __pyx_obj_7pyverbs_4addr_GRH {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_grh grh;
};


/* "pyverbs/addr.pxd":17
 *     cdef v.ibv_grh grh
 * 
 * cdef class GlobalRoute(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_global_route gr
 * 
*/
struct __pyx_obj_7pyverbs_4addr_GlobalRoute {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_global_route gr;
};


/* "pyverbs/addr.pxd":20
 *     cdef v.ibv_global_route gr
 * 
 * cdef class AHAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_ah_attr ah_attr
 *     cdef init_from_ud_param(self, UDParam udparam)
*/
struct __pyx_obj_7pyverbs_4addr_AHAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct __pyx_vtabstruct_7pyverbs_4addr_AHAttr *__pyx_vtab;
  struct ibv_ah_attr ah_attr;
};


/* "pyverbs/addr.pxd":24
 *     cdef init_from_ud_param(self, UDParam udparam)
 * 
 * cdef class AH(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_ah *ah
 *     cdef object pd
*/
struct __pyx_obj_7pyverbs_4addr_AH {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_ah *ah;
  PyObject *pd;
};


/* "pyverbs/flow.pxd":9
 * cimport pyverbs.libibverbs as v
 * 
 * cdef class FlowAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_flow_attr attr
 *     cdef object specs
*/
struct __pyx_obj_7pyverbs_4flow_FlowAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_flow_attr attr;
  PyObject *specs;
};


/* "pyverbs/flow.pxd":13
 *     cdef object specs
 * 
 * cdef class Flow(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_flow *flow
 *     cdef object qp
*/
struct __pyx_obj_7pyverbs_4flow_Flow {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_flow *flow;
  PyObject *qp;
};


/* "pyverbs/flow.pxd":18
 *     cpdef close(self)
 * 
 * cdef class FlowAction(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_flow_action *action
*/
struct __pyx_obj_7pyverbs_4flow_FlowAction {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_flow_action *action;
};


/* "pyverbs/xrcd.pxd":11
 * 
 * 
 * cdef class XRCDInitAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_xrcd_init_attr attr
 * 
*/
struct __pyx_obj_7pyverbs_4xrcd_XRCDInitAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_xrcd_init_attr attr;
};


/* "pyverbs/xrcd.pxd":15
 * 
 * 
 * cdef class XRCD(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_xrcd *xrcd
 *     cdef Context ctx
*/
struct __pyx_obj_7pyverbs_4xrcd_XRCD {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_xrcd *xrcd;
  struct __pyx_obj_7pyverbs_6device_Context *ctx;
  PyObject *srqs;
  PyObject *qps;
};


/* "pyverbs/srq.pxd":9
 * from . cimport libibverbs as v
 * 
 * cdef class SrqAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_srq_attr attr
 * 
*/
struct __pyx_obj_7pyverbs_3srq_SrqAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_srq_attr attr;
};


/* "pyverbs/srq.pxd":12
 *     cdef v.ibv_srq_attr attr
 * 
 * cdef class SrqInitAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_srq_init_attr    attr
 * 
*/
struct __pyx_obj_7pyverbs_3srq_SrqInitAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_srq_init_attr attr;
};


/* "pyverbs/srq.pxd":15
 *     cdef v.ibv_srq_init_attr    attr
 * 
 * cdef class SrqInitAttrEx(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_srq_init_attr_ex attr
 *     cdef object _cq
*/
struct __pyx_obj_7pyverbs_3srq_SrqInitAttrEx {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_srq_init_attr_ex attr;
  PyObject *_cq;
  PyObject *_pd;
  PyObject *_xrcd;
};


/* "pyverbs/srq.pxd":21
 *     cdef object _xrcd
 * 
 * cdef class OpsWr(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_ops_wr ops_wr
 * 
*/
struct __pyx_obj_7pyverbs_3srq_OpsWr {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_ops_wr ops_wr;
};


/* "pyverbs/srq.pxd":24
 *     cdef v.ibv_ops_wr ops_wr
 * 
 * cdef class SRQ(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_srq *srq
 *     cdef object cq
*/
struct __pyx_obj_7pyverbs_3srq_SRQ {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_srq *srq;
  PyObject *cq;
  PyObject *qps;
};


/* "pyverbs/qp.pxd":9
 * cimport pyverbs.libibverbs as v
 * 
 * cdef class QPCap(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_qp_cap cap
 * 
*/
struct __pyx_obj_7pyverbs_2qp_QPCap {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_qp_cap cap;
};


/* "pyverbs/qp.pxd":12
 *     cdef v.ibv_qp_cap cap
 * 
 * cdef class QPInitAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_qp_init_attr attr
 *     cdef object scq
*/
struct __pyx_obj_7pyverbs_2qp_QPInitAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_qp_init_attr attr;
  PyObject *scq;
  PyObject *rcq;
  PyObject *srq;
};


/* "pyverbs/qp.pxd":18
 *     cdef object srq
 * 
 * cdef class QPInitAttrEx(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_qp_init_attr_ex attr
 *     cdef object scq
*/
struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_qp_init_attr_ex attr;
  PyObject *scq;
  PyObject *rcq;
  PyObject *_pd;
  PyObject *xrcd;
  PyObject *srq;
  PyObject *ind_table;
};


/* "pyverbs/qp.pxd":27
 *     cdef object ind_table
 * 
 * cdef class QPAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_qp_attr attr
 * 
*/
struct __pyx_obj_7pyverbs_2qp_QPAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_qp_attr attr;
};


/* "pyverbs/qp.pxd":30
 *     cdef v.ibv_qp_attr attr
 * 
 * cdef class QP(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_qp *qp
 *     cdef int type
*/
struct __pyx_obj_7pyverbs_2qp_QP {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_qp *qp;
  int type;
  int state;
  PyObject *pd;
  PyObject *context;
  PyObject *xrcd;
  PyObject *scq;
  PyObject *rcq;
  PyObject *mws;
  PyObject *srq;
  PyObject *flows;
  PyObject *dr_actions;
};


/* "pyverbs/qp.pxd":47
 *     cdef add_ref(self, obj)
 * 
 * cdef class DataBuffer(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_data_buf data
 * 
*/
struct __pyx_obj_7pyverbs_2qp_DataBuffer {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_data_buf data;
};


/* "pyverbs/qp.pxd":50
 *     cdef v.ibv_data_buf data
 * 
 * cdef class QPEx(QP):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_qp_ex *qp_ex
 *     cdef object ind_table
*/
struct __pyx_obj_7pyverbs_2qp_QPEx {
  struct __pyx_obj_7pyverbs_2qp_QP __pyx_base;
  struct ibv_qp_ex *qp_ex;
  PyObject *ind_table;
};


/* "pyverbs/qp.pxd":54
 *     cdef object ind_table
 * 
 * cdef class ECE(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_ece ece
*/
struct __pyx_obj_7pyverbs_2qp_ECE {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_ece ece;
};



/* "pyverbs/base.pxd":10
 *     cdef object logger
 * 
 * cdef class PyverbsCM(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cpdef close(self)
 * 
*/

struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM {
  PyObject *(*close)(struct __pyx_obj_7pyverbs_4base_PyverbsCM *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM *__pyx_vtabptr_7pyverbs_4base_PyverbsCM;


/* "pyverbs/device.pxd":10
 * 
 * 
 * cdef class Context(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_context *context
 *     cdef v.ibv_device *device
*/

struct __pyx_vtabstruct_7pyverbs_6device_Context {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_6device_Context *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_6device_Context *__pyx_vtabptr_7pyverbs_6device_Context;


/* "pyverbs/device.pxd":67
 *     cdef v.ibv_alloc_dm_attr alloc_dm_attr
 * 
 * cdef class DM(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_dm *dm
 *     cdef object dm_mrs
*/

struct __pyx_vtabstruct_7pyverbs_6device_DM {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_6device_DM *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_6device_DM *__pyx_vtabptr_7pyverbs_6device_DM;


/* "pyverbs/device.pxd":83
 *     cdef v.ibv_async_event event
 * 
 * cdef class FdArr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_fd_arr attr
 *     cpdef close(self)
*/

struct __pyx_vtabstruct_7pyverbs_6device_FdArr {
  PyObject *(*close)(struct __pyx_obj_7pyverbs_6device_FdArr *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyverbs_6device_FdArr *__pyx_vtabptr_7pyverbs_6device_FdArr;


/* "pyverbs/cq.pxd":9
 * cimport pyverbs.libibverbs as v
 * 
 * cdef class CompChannel(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_comp_channel *cc
 *     cpdef close(self)
*/

struct __pyx_vtabstruct_7pyverbs_2cq_CompChannel {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_2cq_CompChannel *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_2cq_CompChannel *__pyx_vtabptr_7pyverbs_2cq_CompChannel;


/* "pyverbs/cq.pxd":16
 *     cdef object cqs
 * 
 * cdef class CQ(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_cq *cq
 *     cpdef close(self)
*/

struct __pyx_vtabstruct_7pyverbs_2cq_CQ {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_2cq_CQ *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_2cq_CQ *__pyx_vtabptr_7pyverbs_2cq_CQ;


/* "pyverbs/cq.pxd":32
 *     cdef object parent_domain
 * 
 * cdef class CQEX(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_cq_ex *cq
 *     cdef v.ibv_cq *ibv_cq
*/

struct __pyx_vtabstruct_7pyverbs_2cq_CQEX {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_2cq_CQEX *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_2cq_CQEX *__pyx_vtabptr_7pyverbs_2cq_CQEX;


/* "pyverbs/pd.pxd":12
 * 
 * 
 * cdef class PD(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_pd *pd
 *     cdef Context ctx
*/

struct __pyx_vtabstruct_7pyverbs_2pd_PD {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_2pd_PD *, PyObject *);
  PyObject *(*remove_ref)(struct __pyx_obj_7pyverbs_2pd_PD *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_2pd_PD *__pyx_vtabptr_7pyverbs_2pd_PD;


/* "pyverbs/pd.pxd":34
 *     cdef object dealloc
 * 
 * cdef class ParentDomain(PD):             # <<<<<<<<<<<<<<
 *     cdef add_ref(self, obj)
 *     cdef object protection_domain
*/

struct __pyx_vtabstruct_7pyverbs_2pd_ParentDomain {
  struct __pyx_vtabstruct_7pyverbs_2pd_PD __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2pd_ParentDomain *__pyx_vtabptr_7pyverbs_2pd_ParentDomain;


/* "pyverbs/wq.pxd":21
 *     cdef v.ibv_wq_attr attr
 * 
 * cdef class WQ(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_wq *wq
 *     cdef Context context
*/

struct __pyx_vtabstruct_7pyverbs_2wq_WQ {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_2wq_WQ *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyverbs_2wq_WQ *__pyx_vtabptr_7pyverbs_2wq_WQ;


/* "pyverbs/wq.pxd":33
 *     cdef object wqs_list
 * 
 * cdef class RwqIndTable(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_rwq_ind_table *rwq_ind_table
 *     cdef Context context
*/

struct __pyx_vtabstruct_7pyverbs_2wq_RwqIndTable {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_2wq_RwqIndTable *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyverbs_2wq_RwqIndTable *__pyx_vtabptr_7pyverbs_2wq_RwqIndTable;


/* "pyverbs/mr.pxd":12
 * 
 * 
 * cdef class MR(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef object pd
 *     cdef object cmid
*/

struct __pyx_vtabstruct_7pyverbs_2mr_MR {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  void (*_allocate_buffer)(struct __pyx_obj_7pyverbs_2mr_MR *, size_t, int, int *);
  void (*_free_buffer)(struct __pyx_obj_7pyverbs_2mr_MR *, int, int);
  PyObject *(*read)(struct __pyx_obj_7pyverbs_2mr_MR *, PyObject *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyverbs_2mr_MR *__pyx_vtabptr_7pyverbs_2mr_MR;


/* "pyverbs/mr.pxd":25
 *     cpdef read(self, length, offset)
 * 
 * cdef class MREx(MR):             # <<<<<<<<<<<<<<
 *     cdef object dmah
 * 
*/

struct __pyx_vtabstruct_7pyverbs_2mr_MREx {
  struct __pyx_vtabstruct_7pyverbs_2mr_MR __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2mr_MREx *__pyx_vtabptr_7pyverbs_2mr_MREx;


/* "pyverbs/mr.pxd":28
 *     cdef object dmah
 * 
 * cdef class MWBindInfo(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_mw_bind_info info
 *     cdef object mr
*/

struct __pyx_vtabstruct_7pyverbs_2mr_MWBindInfo {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2mr_MWBindInfo *__pyx_vtabptr_7pyverbs_2mr_MWBindInfo;


/* "pyverbs/mr.pxd":32
 *     cdef object mr
 * 
 * cdef class MWBind(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_mw_bind mw_bind
 *     cdef object mr
*/

struct __pyx_vtabstruct_7pyverbs_2mr_MWBind {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2mr_MWBind *__pyx_vtabptr_7pyverbs_2mr_MWBind;


/* "pyverbs/mr.pxd":36
 *     cdef object mr
 * 
 * cdef class MW(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef object pd
 *     cdef v.ibv_mw *mw
*/

struct __pyx_vtabstruct_7pyverbs_2mr_MW {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2mr_MW *__pyx_vtabptr_7pyverbs_2mr_MW;


/* "pyverbs/mr.pxd":40
 *     cdef v.ibv_mw *mw
 * 
 * cdef class DMMR(MR):             # <<<<<<<<<<<<<<
 *     cdef object dm
 * 
*/

struct __pyx_vtabstruct_7pyverbs_2mr_DMMR {
  struct __pyx_vtabstruct_7pyverbs_2mr_MR __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2mr_DMMR *__pyx_vtabptr_7pyverbs_2mr_DMMR;


/* "pyverbs/mr.pxd":43
 *     cdef object dm
 * 
 * cdef class DmaBufMR(MR):             # <<<<<<<<<<<<<<
 *     cdef object dmabuf
 *     cdef unsigned long offset
*/

struct __pyx_vtabstruct_7pyverbs_2mr_DmaBufMR {
  struct __pyx_vtabstruct_7pyverbs_2mr_MR __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2mr_DmaBufMR *__pyx_vtabptr_7pyverbs_2mr_DmaBufMR;


/* "pyverbs/mr.pxd":51
 *     cdef v.ibv_dmah_init_attr init_attr
 * 
 * cdef class DMAHandle(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_dmah *dmah
 *     cdef object mrs
*/

struct __pyx_vtabstruct_7pyverbs_2mr_DMAHandle {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_2mr_DMAHandle *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_2mr_DMAHandle *__pyx_vtabptr_7pyverbs_2mr_DMAHandle;


/* "pyverbs/wr.pxd":10
 * 
 * 
 * cdef class SGE(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_sge *sge
 *     cpdef read(self, length, offset)
*/

struct __pyx_vtabstruct_7pyverbs_2wr_SGE {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*read)(struct __pyx_obj_7pyverbs_2wr_SGE *, PyObject *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyverbs_2wr_SGE *__pyx_vtabptr_7pyverbs_2wr_SGE;


/* "pyverbs/wr.pxd":14
 *     cpdef read(self, length, offset)
 * 
 * cdef class RecvWR(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_recv_wr recv_wr
 * 
*/

struct __pyx_vtabstruct_7pyverbs_2wr_RecvWR {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2wr_RecvWR *__pyx_vtabptr_7pyverbs_2wr_RecvWR;


/* "pyverbs/wr.pxd":17
 *     cdef v.ibv_recv_wr recv_wr
 * 
 * cdef class SendWR(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_send_wr send_wr
 *     cdef object ah
*/

struct __pyx_vtabstruct_7pyverbs_2wr_SendWR {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2wr_SendWR *__pyx_vtabptr_7pyverbs_2wr_SendWR;


/* "cmid.pxd":10
 * 
 * 
 * cdef class CMID(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef cm.rdma_cm_id *id
 *     cdef object event_channel
*/

struct __pyx_vtabstruct_7pyverbs_4cmid_CMID {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_4cmid_CMID *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_4cmid_CMID *__pyx_vtabptr_7pyverbs_4cmid_CMID;


/* "cmid.pxd":20
 * 
 * 
 * cdef class CMEventChannel(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef cm.rdma_event_channel *event_channel
 *     cpdef close(self)
*/

struct __pyx_vtabstruct_7pyverbs_4cmid_CMEventChannel {
  PyObject *(*close)(struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyverbs_4cmid_CMEventChannel *__pyx_vtabptr_7pyverbs_4cmid_CMEventChannel;


/* "cmid.pxd":25
 * 
 * 
 * cdef class CMEvent(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef cm.rdma_cm_event *event
 *     cpdef close(self)
*/

struct __pyx_vtabstruct_7pyverbs_4cmid_CMEvent {
  PyObject *(*close)(struct __pyx_obj_7pyverbs_4cmid_CMEvent *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyverbs_4cmid_CMEvent *__pyx_vtabptr_7pyverbs_4cmid_CMEvent;


/* "cmid.pxd":30
 * 
 * 
 * cdef class AddrInfo(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef cm.rdma_addrinfo *addr_info
 *     cpdef close(self)
*/

struct __pyx_vtabstruct_7pyverbs_4cmid_AddrInfo {
  PyObject *(*close)(struct __pyx_obj_7pyverbs_4cmid_AddrInfo *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyverbs_4cmid_AddrInfo *__pyx_vtabptr_7pyverbs_4cmid_AddrInfo;


/* "pyverbs/addr.pxd":20
 *     cdef v.ibv_global_route gr
 * 
 * cdef class AHAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_ah_attr ah_attr
 *     cdef init_from_ud_param(self, UDParam udparam)
*/

struct __pyx_vtabstruct_7pyverbs_4addr_AHAttr {
  PyObject *(*init_from_ud_param)(struct __pyx_obj_7pyverbs_4addr_AHAttr *, struct __pyx_obj_7pyverbs_4cmid_UDParam *);
};
static struct __pyx_vtabstruct_7pyverbs_4addr_AHAttr *__pyx_vtabptr_7pyverbs_4addr_AHAttr;


/* "pyverbs/addr.pxd":24
 *     cdef init_from_ud_param(self, UDParam udparam)
 * 
 * cdef class AH(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_ah *ah
 *     cdef object pd
*/

struct __pyx_vtabstruct_7pyverbs_4addr_AH {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_4addr_AH *__pyx_vtabptr_7pyverbs_4addr_AH;


/* "pyverbs/flow.pxd":13
 *     cdef object specs
 * 
 * cdef class Flow(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_flow *flow
 *     cdef object qp
*/

struct __pyx_vtabstruct_7pyverbs_4flow_Flow {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_4flow_Flow *__pyx_vtabptr_7pyverbs_4flow_Flow;


/* "pyverbs/xrcd.pxd":15
 * 
 * 
 * cdef class XRCD(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_xrcd *xrcd
 *     cdef Context ctx
*/

struct __pyx_vtabstruct_7pyverbs_4xrcd_XRCD {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_4xrcd_XRCD *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_4xrcd_XRCD *__pyx_vtabptr_7pyverbs_4xrcd_XRCD;


/* "pyverbs/srq.pxd":21
 *     cdef object _xrcd
 * 
 * cdef class OpsWr(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_ops_wr ops_wr
 * 
*/

struct __pyx_vtabstruct_7pyverbs_3srq_OpsWr {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_3srq_OpsWr *__pyx_vtabptr_7pyverbs_3srq_OpsWr;


/* "pyverbs/srq.pxd":24
 *     cdef v.ibv_ops_wr ops_wr
 * 
 * cdef class SRQ(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_srq *srq
 *     cdef object cq
*/

struct __pyx_vtabstruct_7pyverbs_3srq_SRQ {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_3srq_SRQ *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_3srq_SRQ *__pyx_vtabptr_7pyverbs_3srq_SRQ;


/* "pyverbs/qp.pyx":927
 * 
 * 
 * cdef class QP(PyverbsCM):             # <<<<<<<<<<<<<<
 *     def __init__(self, object creator not None, object init_attr not None,
 *                  QPAttr qp_attr=None):
*/

struct __pyx_vtabstruct_7pyverbs_2qp_QP {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*update_cqs)(struct __pyx_obj_7pyverbs_2qp_QP *, PyObject *);
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_2qp_QP *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_2qp_QP *__pyx_vtabptr_7pyverbs_2qp_QP;


/* "pyverbs/qp.pyx":1280
 * 
 * 
 * cdef class DataBuffer(PyverbsCM):             # <<<<<<<<<<<<<<
 *     def __init__(self, addr, length):
 *         super().__init__()
*/

struct __pyx_vtabstruct_7pyverbs_2qp_DataBuffer {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2qp_DataBuffer *__pyx_vtabptr_7pyverbs_2qp_DataBuffer;


/* "pyverbs/qp.pyx":1287
 * 
 * 
 * cdef class QPEx(QP):             # <<<<<<<<<<<<<<
 *     def __init__(self, object creator not None, object init_attr not None,
 *                  QPAttr qp_attr=None):
*/

struct __pyx_vtabstruct_7pyverbs_2qp_QPEx {
  struct __pyx_vtabstruct_7pyverbs_2qp_QP __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2qp_QPEx *__pyx_vtabptr_7pyverbs_2qp_QPEx;


/* "pyverbs/qp.pyx":884
 * 
 * 
 * cdef class ECE(PyverbsCM):             # <<<<<<<<<<<<<<
 *     def __init__(self, vendor_id=0, options=0, comp_mask=0):
 *         """
*/

struct __pyx_vtabstruct_7pyverbs_2qp_ECE {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2qp_ECE *__pyx_vtabptr_7pyverbs_2qp_ECE;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto (used by PyErrFetchRestore) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto (used by GetBuiltinName) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* HasAttr.proto (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_HasAttr(o, n)  PyObject_HasAttrWithError(o, n)
#else
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);
#endif

/* PyObjectCall.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto (used by PyObjectCallOneArg) */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs);

/* PyObjectCallOneArg.proto (used by ImportImpl) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* TupleAndListFromArray.proto (used by ImportImpl) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
#endif
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto (used by ImportImpl) */
#include <string.h>

/* ImportImpl.export */
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level);

/* Import.proto */
static CYTHON_INLINE PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* GetTopmostException.proto (used by SaveResetException) */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyImportError_Check.proto */
#define __Pyx_PyExc_ImportError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_ImportError)

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectFastCallMethod.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyObject_FastCallMethod(name, args, nargsf) PyObject_VectorcallMethod(name, args, nargsf, NULL)
#else
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf);
#endif

/* BytesEquals.proto (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_PySequence_ITEM(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_NewRef(__Pyx_PyTuple_GET_ITEM(args, i))
#else
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_XNewRef(PyTuple_GetItem(args, i))
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_ArgRef_FASTCALL(args, i) __Pyx_NewRef(args[i])
    #define __Pyx_NumKwargs_FASTCALL(kwds) __Pyx_PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
#else
    #define __Pyx_ArgRef_FASTCALL __Pyx_ArgRef_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
#endif
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#if CYTHON_METH_FASTCALL || (CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(args + start, stop - start)
#else
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* py_dict_items.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d);

/* CallCFunction.proto (used by CallUnboundCMethod0) */
#define __Pyx_CallCFunction(cfunc, self, args)\
    ((PyCFunction)(void(*)(void))(cfunc)->func)(self, args)
#define __Pyx_CallCFunctionWithKeywords(cfunc, self, args, kwargs)\
    ((PyCFunctionWithKeywords)(void(*)(void))(cfunc)->func)(self, args, kwargs)
#define __Pyx_CallCFunctionFast(cfunc, self, args, nargs)\
    ((__Pyx_PyCFunctionFast)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs)
#define __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, nargs, kwnames)\
    ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs, kwnames)

/* UnpackUnboundCMethod.proto (used by CallUnboundCMethod0) */
typedef struct {
    PyObject *type;
    PyObject **method_name;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && CYTHON_ATOMICS
    __pyx_atomic_int_type initialized;
#endif
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
static CYTHON_INLINE int __Pyx_CachedCFunction_GetAndSetInitializing(__Pyx_CachedCFunction *cfunc) {
#if !CYTHON_ATOMICS
    return 1;
#else
    __pyx_nonatomic_int_type expected = 0;
    if (__pyx_atomic_int_cmp_exchange(&cfunc->initialized, &expected, 1)) {
        return 0;
    }
    return expected;
#endif
}
static CYTHON_INLINE void __Pyx_CachedCFunction_SetFinishedInitializing(__Pyx_CachedCFunction *cfunc) {
#if CYTHON_ATOMICS
    __pyx_atomic_store(&cfunc->initialized, 2);
#endif
}
#else
#define __Pyx_CachedCFunction_GetAndSetInitializing(cfunc) 2
#define __Pyx_CachedCFunction_SetFinishedInitializing(cfunc)
#endif

/* CallUnboundCMethod0.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* py_dict_values.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d);

/* OwnedDictNext.proto (used by ParseKeywordsImpl) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue);
#else
CYTHON_INLINE
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue);
#endif

/* RaiseDoubleKeywords.proto (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywordsImpl.export */
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name
);
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* CallUnboundCMethod2.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* ParseKeywords.proto */
static CYTHON_INLINE int __Pyx_ParseKeywords(
    PyObject *kwds, PyObject *const *kwvalues, PyObject ** const argnames[],
    PyObject *kwds2, PyObject *values[],
    Py_ssize_t num_pos_args, Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* UnicodeConcatInPlace.proto */
# if CYTHON_COMPILING_IN_CPYTHON
    #if CYTHON_REFNANNY
        #define __Pyx_PyUnicode_ConcatInPlace(left, right, unsafe_shared) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right, unsafe_shared, __pyx_refnanny)
    #else
        #define __Pyx_PyUnicode_ConcatInPlace(left, right, unsafe_shared) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right, unsafe_shared)
    #endif
    #define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_DefinitelyUniqueInPlace(left, right) __Pyx_PyUnicode_ConcatInPlace(left, right, __Pyx_ReferenceSharing_DefinitelyUnique)
    #define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(left, right) __Pyx_PyUnicode_ConcatInPlace(left, right, __Pyx_ReferenceSharing_OwnStrongReference)
    #define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_FunctionArgumentInPlace(left, right) __Pyx_PyUnicode_ConcatInPlace(left, right, __Pyx_ReferenceSharing_DefinitelyUnique)
    #define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_SharedReferenceInPlace(left, right) __Pyx_PyUnicode_ConcatInPlace(left, right, __Pyx_ReferenceSharing_SharedReference)
    static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right, int unsafe_shared
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    );
#else
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_DefinitelyUniqueInPlace __Pyx_PyUnicode_Concat
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace __Pyx_PyUnicode_Concat
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_FunctionArgumentInPlace __Pyx_PyUnicode_Concat
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_SharedReferenceInPlace __Pyx_PyUnicode_Concat
#endif
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_DefinitelyUniqueInPlaceSafe(left, right)\
    ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_DefinitelyUniqueInPlace(left, right))
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlaceSafe(left, right)\
    ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(left, right))
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_FunctionArgumentInPlaceSafe(left, right)\
    ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_FunctionArgumentInPlace(left, right))
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_SharedReferenceInPlaceSafe(left, right)\
    ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_SharedReferenceInPlace(left, right))

/* RejectKeywords.export */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds);

/* PyTypeError_Check.proto */
#define __Pyx_PyExc_TypeError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_TypeError)

/* RaiseException.export */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* ArgTypeTestFunc.export */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))

/* PyDictVersioning.proto (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __Pyx_XNewRef(__pyx_dict_cached_value);\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_mstate_global->__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectVectorCallKwBuilder.proto (used by PyObjectVectorCallMethodKwBuilder) */
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#if CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_Vectorcall_CallFromBuilder PyObject_Vectorcall
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder _PyObject_Vectorcall
#endif
#define __Pyx_MakeVectorcallBuilderKwds(n) PyTuple_New(n)
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder __Pyx_PyObject_FastCallDict
#define __Pyx_MakeVectorcallBuilderKwds(n) __Pyx_PyDict_NewPresized(n)
#define __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n) PyDict_SetItem(builder, key, value)
#define __Pyx_VectorcallBuilder_AddArgStr(key, value, builder, args, n) PyDict_SetItemString(builder, key, value)
#endif

/* PyObjectVectorCallMethodKwBuilder.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_VectorcallMethod_CallFromBuilder PyObject_VectorcallMethod
#else
static PyObject *__Pyx_Object_VectorcallMethod_CallFromBuilder(PyObject *name, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#endif

/* IterFinish.proto (used by dict_iter) */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyObjectCallNoArg.proto (used by PyObjectCallMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectGetMethod.proto (used by PyObjectCallMethod0) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);
#endif

/* PyObjectCallMethod0.proto (used by dict_iter) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto (used by UnpackTuple2) */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto (used by UnpackItemEndCheck) */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto (used by UnpackTuple2) */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto (used by UnpackTupleError) */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto (used by UnpackTuple2) */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto (used by dict_iter) */
static CYTHON_INLINE int __Pyx_unpack_tuple2(
    PyObject* tuple, PyObject** value1, PyObject** value2, int is_tuple, int has_known_size, int decref_tuple);
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* DictGetItem.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck, unsafe_shared) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck, int unsafe_shared);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* PyObjectDelAttr.proto (used by PyObjectSetAttrStr) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
#define __Pyx_PyObject_DelAttr(o, n) PyObject_SetAttr(o, n, NULL)
#else
#define __Pyx_PyObject_DelAttr(o, n) PyObject_DelAttr(o, n)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   __Pyx_PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* PyRange_Check.proto */
#if CYTHON_COMPILING_IN_PYPY && !defined(PyRange_Check)
  #define PyRange_Check(obj)  __Pyx_TypeCheck((obj), &PyRange_Type)
#endif

/* CallNextTpDealloc.proto */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallTypeTraverse.proto */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#define __Pyx_call_type_traverse(o, always_call, visit, arg) 0
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg);
#endif

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_2_4
#define __PYX_HAVE_RT_ImportType_proto_3_2_4
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_2_4 {
   __Pyx_ImportType_CheckSize_Error_3_2_4 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_2_4 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_2_4 = 2
};
static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size);
#endif

/* LimitedApiGetTypeDict.proto (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp);
#endif

/* SetItemOnTypeDict.proto (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v);
#define __Pyx_SetItemOnTypeDict(tp, k, v) __Pyx__SetItemOnTypeDict((PyTypeObject*)tp, k, v)

/* FixUpExtensionType.proto */
static CYTHON_INLINE int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);

/* ValidateBasesTuple.proto (used by PyType_Ready) */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* DelItemOnTypeDict.proto (used by SetupReduce) */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k);
#define __Pyx_DelItemOnTypeDict(tp, k) __Pyx__DelItemOnTypeDict((PyTypeObject*)tp, k)

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* MergeVTables.proto */
static int __Pyx_MergeVtables(PyTypeObject *type);

/* FunctionImport.proto */
static int __Pyx_ImportFunction_3_2_4(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* dict_setdefault.proto (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value);

/* AddModuleRef.proto (used by FetchSharedCythonModule) */
#if ((CYTHON_COMPILING_IN_CPYTHON_FREETHREADING ) ||\
     __PYX_LIMITED_VERSION_HEX < 0x030d0000)
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name);
#else
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#endif

/* FetchSharedCythonModule.proto (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto (used by CommonTypesMetaclass) */
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases);

/* CommonTypesMetaclass.proto (used by CythonFunctionShared) */
static int __pyx_CommonTypesMetaclass_init(PyObject *module);
#define __Pyx_CommonTypesMetaclass_USED

/* PyMethodNew.proto (used by CythonFunctionShared) */
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ);

/* PyVectorcallFastCallDict.proto (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto (used by CythonFunction) */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && CYTHON_METH_FASTCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_dict;
#endif
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    PyObject *defaults;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_mstate_global->__pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_mstate_global->__pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_mstate_global->__pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void));
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func,
                                                         PyTypeObject *defaults_type);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* CLineInTraceback.proto (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#else
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#endif

/* CodeObjectCache.proto (used by AddTraceback) */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject __Pyx_CachedCodeObjectType;
#else
typedef PyCodeObject __Pyx_CachedCodeObjectType;
#endif
typedef struct {
    __Pyx_CachedCodeObjectType* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_int_type accessor_count;
  #endif
};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned int __Pyx_PyLong_As_unsigned_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_int(unsigned int value);

/* CIntFromPy.proto */
static CYTHON_INLINE enum ibv_qp_type __Pyx_PyLong_As_enum__ibv_qp_type(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned short __Pyx_PyLong_As_unsigned_short(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned long __Pyx_PyLong_As_unsigned_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE enum ibv_qp_state __Pyx_PyLong_As_enum__ibv_qp_state(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_short(unsigned short value);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned char __Pyx_PyLong_As_unsigned_char(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE enum ibv_mtu __Pyx_PyLong_As_enum__ibv_mtu(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE enum ibv_mig_state __Pyx_PyLong_As_enum__ibv_mig_state(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_long(unsigned long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_char(unsigned char value);

/* CIntFromPy.proto */
static CYTHON_INLINE uint32_t __Pyx_PyLong_As_uint32_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_uint32_t(uint32_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE enum ibv_wr_opcode __Pyx_PyLong_As_enum__ibv_wr_opcode(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyLong_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint64_t __Pyx_PyLong_As_uint64_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_uint64_t(uint64_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE uint8_t __Pyx_PyLong_As_uint8_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint16_t __Pyx_PyLong_As_uint16_t(PyObject *);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#else
static __Pyx_TypeName __Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp);
#endif
#else  // !LIMITED_API
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetFullyQualifiedName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2) {
    return PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2);
}
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#ifdef PyExceptionInstance_Check
  #define __Pyx_PyBaseException_Check(obj) PyExceptionInstance_Check(obj)
#else
  #define __Pyx_PyBaseException_Check(obj) __Pyx_TypeCheck(obj, PyExc_BaseException)
#endif

/* GetRuntimeVersion.proto */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
static unsigned long __Pyx_cached_runtime_version = 0;
static void __Pyx_init_runtime_version(void);
#else
#define __Pyx_init_runtime_version()
#endif
static unsigned long __Pyx_get_runtime_version(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* DecompressString.proto */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo);

/* MultiPhaseInitModuleState.proto */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
static PyObject *__Pyx_State_FindModule(void*);
static int __Pyx_State_AddModule(PyObject* module, void*);
static int __Pyx_State_RemoveModule(void*);
#elif CYTHON_USE_MODULE_STATE
#define __Pyx_State_FindModule PyState_FindModule
#define __Pyx_State_AddModule PyState_AddModule
#define __Pyx_State_RemoveModule PyState_RemoveModule
#endif

/* #### Code section: module_declarations ### */
/* CythonABIVersion.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
    #if CYTHON_METH_FASTCALL
        #define __PYX_FASTCALL_ABI_SUFFIX  "_fastcall"
    #else
        #define __PYX_FASTCALL_ABI_SUFFIX
    #endif
    #define __PYX_LIMITED_ABI_SUFFIX "limited" __PYX_FASTCALL_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#else
    #define __PYX_LIMITED_ABI_SUFFIX
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    #define __PYX_AM_SEND_ABI_SUFFIX
#elif __PYX_HAS_PY_AM_SEND == 2
    #define __PYX_AM_SEND_ABI_SUFFIX "amsendbackport"
#else
    #define __PYX_AM_SEND_ABI_SUFFIX "noamsend"
#endif
#ifndef __PYX_MONITORING_ABI_SUFFIX
    #define __PYX_MONITORING_ABI_SUFFIX
#endif
#if CYTHON_USE_TP_FINALIZE
    #define __PYX_TP_FINALIZE_ABI_SUFFIX
#else
    #define __PYX_TP_FINALIZE_ABI_SUFFIX "nofinalize"
#endif
#if CYTHON_USE_FREELISTS || !defined(__Pyx_AsyncGen_USED)
    #define __PYX_FREELISTS_ABI_SUFFIX
#else
    #define __PYX_FREELISTS_ABI_SUFFIX "nofreelists"
#endif
#define CYTHON_ABI  __PYX_ABI_VERSION __PYX_LIMITED_ABI_SUFFIX __PYX_MONITORING_ABI_SUFFIX __PYX_TP_FINALIZE_ABI_SUFFIX __PYX_FREELISTS_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."

static PyObject *__pyx_f_7pyverbs_2qp_2QP_update_cqs(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, PyObject *__pyx_v_init_attr); /* proto*/
static PyObject *__pyx_f_7pyverbs_2qp_2QP_add_ref(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, PyObject *__pyx_v_obj); /* proto*/
static PyObject *__pyx_f_7pyverbs_2qp_2QP_close(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/

/* Module declarations from "pyverbs.base" */
static PyObject *(*__pyx_f_7pyverbs_4base_close_weakrefs)(PyObject *); /*proto*/

/* Module declarations from "libc.stdint" */

/* Module declarations from "posix.types" */

/* Module declarations from "posix.signal" */

/* Module declarations from "posix.time" */

/* Module declarations from "pyverbs.libibverbs_enums" */

/* Module declarations from "pyverbs.libibverbs" */

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdlib" */

/* Module declarations from "pyverbs.device" */

/* Module declarations from "pyverbs.cq" */

/* Module declarations from "pyverbs.pd" */

/* Module declarations from "pyverbs.wq" */

/* Module declarations from "pyverbs.librdmacm_enums" */

/* Module declarations from "pyverbs.librdmacm" */

/* Module declarations from "pyverbs" */

/* Module declarations from "pyverbs.mr" */

/* Module declarations from "pyverbs.wr" */

/* Module declarations from "pyverbs.cmid" */

/* Module declarations from "pyverbs.addr" */

/* Module declarations from "pyverbs.flow" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "cpython" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.ref" */

/* Module declarations from "pyverbs.xrcd" */

/* Module declarations from "pyverbs.srq" */

/* Module declarations from "pyverbs.qp" */
static PyObject *__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py = 0;
static PyObject *__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py = 0;
static PyObject *__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py = 0;
static PyObject *__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py = 0;
static PyObject *__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py = 0;
static PyObject *__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py = 0;
static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(enum ibv_qp_type); /*proto*/
static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py(enum ibv_qp_state); /*proto*/
static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py(enum ibv_mtu); /*proto*/
static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py(enum ibv_mig_state); /*proto*/
static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py(enum ibv_qp_attr_mask); /*proto*/
static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py(enum ibv_qp_init_attr_mask); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "pyverbs.qp"
extern int __pyx_module_is_main_pyverbs__qp;
int __pyx_module_is_main_pyverbs__qp = 0;

/* Implementation of "pyverbs.qp" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_super;
/* #### Code section: string_decls ### */
/* #### Code section: decls ### */
static int __pyx_pf_7pyverbs_2qp_5QPCap___init__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self, PyObject *__pyx_v_max_send_wr, PyObject *__pyx_v_max_recv_wr, PyObject *__pyx_v_max_send_sge, PyObject *__pyx_v_max_recv_sge, PyObject *__pyx_v_max_inline_data); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_5QPCap_11max_send_wr___get__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_5QPCap_11max_send_wr_2__set__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_5QPCap_11max_recv_wr___get__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_5QPCap_11max_recv_wr_2__set__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_5QPCap_12max_send_sge___get__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_5QPCap_12max_send_sge_2__set__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_5QPCap_12max_recv_sge___get__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_5QPCap_12max_recv_sge_2__set__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_5QPCap_15max_inline_data___get__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_5QPCap_15max_inline_data_2__set__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_5QPCap_2__str__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_5QPCap_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_5QPCap_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7pyverbs_2qp_10QPInitAttr___init__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, PyObject *__pyx_v_qp_type, PyObject *__pyx_v_qp_context, struct __pyx_obj_7pyverbs_4base_PyverbsObject *__pyx_v_scq, struct __pyx_obj_7pyverbs_4base_PyverbsObject *__pyx_v_rcq, struct __pyx_obj_7pyverbs_3srq_SRQ *__pyx_v_srq, struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_cap, PyObject *__pyx_v_sq_sig_all); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_7send_cq___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_10QPInitAttr_7send_cq_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_3srq___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_10QPInitAttr_3srq_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, struct __pyx_obj_7pyverbs_3srq_SRQ *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_7recv_cq___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_10QPInitAttr_7recv_cq_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_3cap___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_10QPInitAttr_3cap_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_7qp_type___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_10QPInitAttr_7qp_type_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_10sq_sig_all___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_10QPInitAttr_10sq_sig_all_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_11max_send_wr___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_10QPInitAttr_11max_send_wr_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_11max_recv_wr___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_10QPInitAttr_11max_recv_wr_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_12max_send_sge___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_10QPInitAttr_12max_send_sge_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_12max_recv_sge___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_10QPInitAttr_12max_recv_sge_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_15max_inline_data___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_10QPInitAttr_15max_inline_data_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_2__str__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx___init__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_qp_type, CYTHON_UNUSED PyObject *__pyx_v_qp_context, struct __pyx_obj_7pyverbs_4base_PyverbsObject *__pyx_v_scq, struct __pyx_obj_7pyverbs_4base_PyverbsObject *__pyx_v_rcq, struct __pyx_obj_7pyverbs_3srq_SRQ *__pyx_v_srq, struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_cap, PyObject *__pyx_v_sq_sig_all, PyObject *__pyx_v_comp_mask, struct __pyx_obj_7pyverbs_2pd_PD *__pyx_v_pd, struct __pyx_obj_7pyverbs_4xrcd_XRCD *__pyx_v_xrcd, PyObject *__pyx_v_create_flags, PyObject *__pyx_v_max_tso_header, PyObject *__pyx_v_source_qpn, struct __pyx_obj_7pyverbs_2wq_RxHashConf *__pyx_v_hash_conf, struct __pyx_obj_7pyverbs_2wq_RwqIndTable *__pyx_v_ind_table, PyObject *__pyx_v_send_ops_flags); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_7send_cq___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_7send_cq_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_7recv_cq___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_7recv_cq_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_3cap___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_3cap_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_7qp_type___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_7qp_type_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_10sq_sig_all___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_10sq_sig_all_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_9comp_mask___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_9comp_mask_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_2pd___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_2pd_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, struct __pyx_obj_7pyverbs_2pd_PD *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_4xrcd___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_4xrcd_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, struct __pyx_obj_7pyverbs_4xrcd_XRCD *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_3srq___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_3srq_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, struct __pyx_obj_7pyverbs_3srq_SRQ *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_12create_flags___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_12create_flags_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_14max_tso_header___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_14max_tso_header_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_10source_qpn___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_10source_qpn_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_11max_send_wr___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_11max_send_wr_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_11max_recv_wr___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_11max_recv_wr_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_12max_send_sge___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_12max_send_sge_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_12max_recv_sge___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_12max_recv_sge_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_15max_inline_data___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_15max_inline_data_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_9ind_table___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_9ind_table_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, struct __pyx_obj_7pyverbs_2wq_RwqIndTable *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_2mask_to_str(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_mask); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_4flags_to_str(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_flags); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_6__str__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr___init__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_qp_state, PyObject *__pyx_v_cur_qp_state, PyObject *__pyx_v_port_num, PyObject *__pyx_v_path_mtu); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_8qp_state___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_8qp_state_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_12cur_qp_state___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_12cur_qp_state_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_8path_mtu___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_8path_mtu_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_14path_mig_state___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_14path_mig_state_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_4qkey___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_4qkey_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_6rq_psn___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_6rq_psn_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_6sq_psn___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_6sq_psn_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_11dest_qp_num___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_11dest_qp_num_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_15qp_access_flags___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_15qp_access_flags_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_3cap___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_3cap_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_7ah_attr___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_7ah_attr_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_11alt_ah_attr___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_11alt_ah_attr_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_2_copy_ah(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, struct __pyx_obj_7pyverbs_4addr_AHAttr *__pyx_v_ah_attr, PyObject *__pyx_v_is_alt); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_10pkey_index___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_10pkey_index_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_14alt_pkey_index___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_14alt_pkey_index_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_19en_sqd_async_notify___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_19en_sqd_async_notify_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_11sq_draining___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_11sq_draining_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_13max_rd_atomic___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_13max_rd_atomic_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_18max_dest_rd_atomic___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_18max_dest_rd_atomic_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_13min_rnr_timer___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_13min_rnr_timer_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_8port_num___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_8port_num_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_7timeout___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_7timeout_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_9retry_cnt___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_9retry_cnt_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_9rnr_retry___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_9rnr_retry_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_12alt_port_num___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_12alt_port_num_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_11alt_timeout___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_11alt_timeout_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_10rate_limit___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_6QPAttr_10rate_limit_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_4__str__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7pyverbs_2qp_3ECE___init__(struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_self, PyObject *__pyx_v_vendor_id, PyObject *__pyx_v_options, PyObject *__pyx_v_comp_mask); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_3ECE_9vendor_id___get__(struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_3ECE_9vendor_id_2__set__(struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_3ECE_7options___get__(struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_3ECE_7options_2__set__(struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_3ECE_9comp_mask___get__(struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_3ECE_9comp_mask_2__set__(struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_3ECE_2__str__(struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_3ECE_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_3ECE_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7pyverbs_2qp_2QP___init__(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, PyObject *__pyx_v_creator, PyObject *__pyx_v_init_attr, struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_qp_attr); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_2QP_2_create_qp(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, struct __pyx_obj_7pyverbs_2pd_PD *__pyx_v_pd, struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_attr); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_2QP_4_create_qp_ex(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, struct __pyx_obj_7pyverbs_6device_Context *__pyx_v_ctx, struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_attr); /* proto */
static void __pyx_pf_7pyverbs_2qp_2QP_6__dealloc__(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_2QP_8close(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_2QP_10_get_comp_mask(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, PyObject *__pyx_v_dst); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_2QP_12to_init(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_qp_attr); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_2QP_14to_rtr(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_qp_attr); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_2QP_16to_rts(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_qp_attr); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_2QP_18query(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, PyObject *__pyx_v_attr_mask); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_2QP_20modify(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_qp_attr, PyObject *__pyx_v_comp_mask); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_2QP_22post_recv(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, struct __pyx_obj_7pyverbs_2wr_RecvWR *__pyx_v_wr, struct __pyx_obj_7pyverbs_2wr_RecvWR *__pyx_v_bad_wr); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_2QP_24post_send(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, struct __pyx_obj_7pyverbs_2wr_SendWR *__pyx_v_wr, struct __pyx_obj_7pyverbs_2wr_SendWR *__pyx_v_bad_wr); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_2QP_26set_ece(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_ece); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_2QP_28query_ece(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_2QP_30bind_mw(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, struct __pyx_obj_7pyverbs_2mr_MW *__pyx_v_mw, struct __pyx_obj_7pyverbs_2mr_MWBind *__pyx_v_mw_bind); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_2QP_32query_data_in_order(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, PyObject *__pyx_v_op, PyObject *__pyx_v_flags); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_2QP_7qp_type___get__(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_2QP_8qp_state___get__(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_2QP_6qp_num___get__(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_2QP_2qp___get__(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_2QP_34__str__(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_2QP_36__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_2QP_38__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7pyverbs_2qp_10DataBuffer___init__(struct __pyx_obj_7pyverbs_2qp_DataBuffer *__pyx_v_self, PyObject *__pyx_v_addr, PyObject *__pyx_v_length); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_10DataBuffer_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_DataBuffer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_10DataBuffer_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_DataBuffer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7pyverbs_2qp_4QPEx___init__(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_creator, PyObject *__pyx_v_init_attr, struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_qp_attr); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_9comp_mask___get__(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_4QPEx_9comp_mask_2__set__(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_5wr_id___get__(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_4QPEx_5wr_id_2__set__(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_8wr_flags___get__(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2qp_4QPEx_8wr_flags_2__set__(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_9ind_table___get__(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_2wr_atomic_cmp_swp(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_rkey, PyObject *__pyx_v_remote_addr, PyObject *__pyx_v_compare, PyObject *__pyx_v_swap); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_4wr_atomic_fetch_add(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_rkey, PyObject *__pyx_v_remote_addr, PyObject *__pyx_v_add); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_6wr_bind_mw(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, struct __pyx_obj_7pyverbs_2mr_MW *__pyx_v_mw, PyObject *__pyx_v_rkey, struct __pyx_obj_7pyverbs_2mr_MWBindInfo *__pyx_v_bind_info); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_8wr_local_inv(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_invalidate_rkey); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_10wr_atomic_write(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_rkey, PyObject *__pyx_v_remote_addr, PyObject *__pyx_v_atomic_wr); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_12wr_rdma_read(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_rkey, PyObject *__pyx_v_remote_addr); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_14wr_rdma_write(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_rkey, PyObject *__pyx_v_remote_addr); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_16wr_rdma_write_imm(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_rkey, PyObject *__pyx_v_remote_addr, PyObject *__pyx_v_data); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_18wr_flush(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_rkey, PyObject *__pyx_v_remote_addr, PyObject *__pyx_v_length, PyObject *__pyx_v_ptype, PyObject *__pyx_v_level); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_20wr_send(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_22wr_send_imm(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_data); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_24wr_send_inv(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_invalidate_rkey); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_26wr_send_tso(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_hdr, PyObject *__pyx_v_hdr_sz, PyObject *__pyx_v_mss); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_28wr_set_ud_addr(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, struct __pyx_obj_7pyverbs_4addr_AH *__pyx_v_ah, PyObject *__pyx_v_remote_qpn, PyObject *__pyx_v_remote_rkey); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_30wr_set_xrc_srqn(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_remote_srqn); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_32wr_set_inline_data(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_addr, PyObject *__pyx_v_length); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_34wr_set_inline_data_list(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_num_buf, PyObject *__pyx_v_buf_list); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_36wr_set_sge(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, struct __pyx_obj_7pyverbs_2wr_SGE *__pyx_v_sge); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_38wr_set_sge_list(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_num_sge, PyObject *__pyx_v_sg_list); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_40wr_start(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_42wr_complete(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_44wr_abort(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_46__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_48__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7pyverbs_2qp__copy_caps(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_src, PyObject *__pyx_v_dst); /* proto */
static PyObject *__pyx_tp_new_7pyverbs_2qp_QPCap(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7pyverbs_2qp_QPInitAttr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7pyverbs_2qp_QPInitAttrEx(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7pyverbs_2qp_QPAttr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7pyverbs_2qp_QP(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7pyverbs_2qp_DataBuffer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7pyverbs_2qp_QPEx(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7pyverbs_2qp_ECE(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
/* SmallCodeConfig */
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  PyTypeObject *__pyx_ptype_7pyverbs_4base_PyverbsObject;
  PyTypeObject *__pyx_ptype_7pyverbs_4base_PyverbsCM;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_Context;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_DeviceAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_QueryDeviceExInput;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_ODPCaps;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_RSSCaps;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_PacketPacingCaps;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_PCIAtomicCaps;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_TMCaps;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_CQModerationCaps;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_TSOCaps;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_DeviceAttrEx;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_AllocDmAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_DM;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_PortAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_GIDEntry;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_AsyncEvent;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_FdArr;
  PyTypeObject *__pyx_ptype_7pyverbs_2cq_CompChannel;
  PyTypeObject *__pyx_ptype_7pyverbs_2cq_CQ;
  PyTypeObject *__pyx_ptype_7pyverbs_2cq_CqInitAttrEx;
  PyTypeObject *__pyx_ptype_7pyverbs_2cq_CQEX;
  PyTypeObject *__pyx_ptype_7pyverbs_2cq_WC;
  PyTypeObject *__pyx_ptype_7pyverbs_2cq_PollCqAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_2cq_WcTmInfo;
  PyTypeObject *__pyx_ptype_7pyverbs_2pd_PD;
  PyTypeObject *__pyx_ptype_7pyverbs_2pd_ParentDomainInitAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_2pd_ParentDomain;
  PyTypeObject *__pyx_ptype_7pyverbs_2pd_ParentDomainContext;
  PyTypeObject *__pyx_ptype_7pyverbs_2wq_WQInitAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_2wq_WQAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_2wq_WQ;
  PyTypeObject *__pyx_ptype_7pyverbs_2wq_RwqIndTableInitAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_2wq_RwqIndTable;
  PyTypeObject *__pyx_ptype_7pyverbs_2wq_RxHashConf;
  PyTypeObject *__pyx_ptype_7pyverbs_2mr_MR;
  PyTypeObject *__pyx_ptype_7pyverbs_2mr_MREx;
  PyTypeObject *__pyx_ptype_7pyverbs_2mr_MWBindInfo;
  PyTypeObject *__pyx_ptype_7pyverbs_2mr_MWBind;
  PyTypeObject *__pyx_ptype_7pyverbs_2mr_MW;
  PyTypeObject *__pyx_ptype_7pyverbs_2mr_DMMR;
  PyTypeObject *__pyx_ptype_7pyverbs_2mr_DmaBufMR;
  PyTypeObject *__pyx_ptype_7pyverbs_2mr_DmaHandleInitAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_2mr_DMAHandle;
  PyTypeObject *__pyx_ptype_7pyverbs_2wr_SGE;
  PyTypeObject *__pyx_ptype_7pyverbs_2wr_RecvWR;
  PyTypeObject *__pyx_ptype_7pyverbs_2wr_SendWR;
  PyTypeObject *__pyx_ptype_7pyverbs_4cmid_CMID;
  PyTypeObject *__pyx_ptype_7pyverbs_4cmid_CMEventChannel;
  PyTypeObject *__pyx_ptype_7pyverbs_4cmid_CMEvent;
  PyTypeObject *__pyx_ptype_7pyverbs_4cmid_AddrInfo;
  PyTypeObject *__pyx_ptype_7pyverbs_4cmid_ConnParam;
  PyTypeObject *__pyx_ptype_7pyverbs_4cmid_UDParam;
  PyTypeObject *__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx;
  PyTypeObject *__pyx_ptype_7pyverbs_4addr_GID;
  PyTypeObject *__pyx_ptype_7pyverbs_4addr_GRH;
  PyTypeObject *__pyx_ptype_7pyverbs_4addr_GlobalRoute;
  PyTypeObject *__pyx_ptype_7pyverbs_4addr_AHAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_4addr_AH;
  PyTypeObject *__pyx_ptype_7pyverbs_4flow_FlowAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_4flow_Flow;
  PyTypeObject *__pyx_ptype_7pyverbs_4flow_FlowAction;
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  PyTypeObject *__pyx_ptype_7pyverbs_4xrcd_XRCDInitAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_4xrcd_XRCD;
  PyTypeObject *__pyx_ptype_7pyverbs_3srq_SrqAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_3srq_SrqInitAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_3srq_SrqInitAttrEx;
  PyTypeObject *__pyx_ptype_7pyverbs_3srq_OpsWr;
  PyTypeObject *__pyx_ptype_7pyverbs_3srq_SRQ;
  PyObject *__pyx_type_7pyverbs_2qp_QPCap;
  PyObject *__pyx_type_7pyverbs_2qp_QPInitAttr;
  PyObject *__pyx_type_7pyverbs_2qp_QPInitAttrEx;
  PyObject *__pyx_type_7pyverbs_2qp_QPAttr;
  PyObject *__pyx_type_7pyverbs_2qp_QP;
  PyObject *__pyx_type_7pyverbs_2qp_DataBuffer;
  PyObject *__pyx_type_7pyverbs_2qp_QPEx;
  PyObject *__pyx_type_7pyverbs_2qp_ECE;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_QPCap;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_QPInitAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_QPInitAttrEx;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_QPAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_QP;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_DataBuffer;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_QPEx;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_ECE;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_items;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_values;
  __Pyx_CachedCFunction __pyx_umethod_PyUnicode_Type__format;
  PyObject *__pyx_k__2;
  PyObject *__pyx_k__3;
  PyObject *__pyx_k__6;
  PyObject *__pyx_k__7;
  PyObject *__pyx_k__8;
  PyObject *__pyx_tuple[3];
  PyObject *__pyx_codeobj_tab[57];
  PyObject *__pyx_string_tab[477];
  PyObject *__pyx_number_tab[8];
/* #### Code section: module_state_contents ### */
/* CommonTypesMetaclass.module_state_decls */
PyTypeObject *__pyx_CommonTypesMetaclassType;

/* CachedMethodType.module_state_decls */
#if CYTHON_COMPILING_IN_LIMITED_API
PyObject *__Pyx_CachedMethodType;
#endif

/* CythonFunctionShared.module_state_decls */
PyTypeObject *__pyx_CyFunctionType;

/* CodeObjectCache.module_state_decls */
struct __Pyx_CodeObjectCache __pyx_code_cache;

/* #### Code section: module_state_end ### */
} __pyx_mstatetype;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate_global (__Pyx_PyModule_GetState(__Pyx_State_FindModule(&__pyx_moduledef)))

#define __pyx_m (__Pyx_State_FindModule(&__pyx_moduledef))
#else
static __pyx_mstatetype __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstatetype * const __pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: constant_name_defines ### */
#define __pyx_kp_u_ __pyx_string_tab[0]
#define __pyx_kp_u_20_20 __pyx_string_tab[1]
#define __pyx_kp_u_20_20_2 __pyx_string_tab[2]
#define __pyx_kp_u_22_0x_20x __pyx_string_tab[3]
#define __pyx_kp_u_22_20 __pyx_string_tab[4]
#define __pyx_kp_u_22_20_2 __pyx_string_tab[5]
#define __pyx_kp_u_22_20_3 __pyx_string_tab[6]
#define __pyx_kp_u_AH_Attr __pyx_string_tab[7]
#define __pyx_kp_u_Alt_AH_Attr __pyx_string_tab[8]
#define __pyx_kp_u_Alt_PKey_index __pyx_string_tab[9]
#define __pyx_kp_u_Alt_port_number __pyx_string_tab[10]
#define __pyx_kp_u_Alt_timeout __pyx_string_tab[11]
#define __pyx_kp_u_Block_self_mcast_loopback __pyx_string_tab[12]
#define __pyx_kp_u_CVLAN_stripping __pyx_string_tab[13]
#define __pyx_kp_u_Closing_QP __pyx_string_tab[14]
#define __pyx_kp_u_Comp_Mask __pyx_string_tab[15]
#define __pyx_kp_u_Create_Flags __pyx_string_tab[16]
#define __pyx_kp_u_Dest_QP_number __pyx_string_tab[17]
#define __pyx_kp_u_ECE __pyx_string_tab[18]
#define __pyx_kp_u_En_SQD_async_notify __pyx_string_tab[19]
#define __pyx_kp_u_Expected_CQ_CQEX_got_t __pyx_string_tab[20]
#define __pyx_kp_u_Failed_to_Bind_MW __pyx_string_tab[21]
#define __pyx_kp_u_Failed_to_allocate_SGE_buffer __pyx_string_tab[22]
#define __pyx_kp_u_Failed_to_allocate_data_buffer __pyx_string_tab[23]
#define __pyx_kp_u_Failed_to_create_QP __pyx_string_tab[24]
#define __pyx_kp_u_Failed_to_create_extended_QP __pyx_string_tab[25]
#define __pyx_kp_u_Failed_to_destroy_QP __pyx_string_tab[26]
#define __pyx_kp_u_Failed_to_modify_QP __pyx_string_tab[27]
#define __pyx_kp_u_Failed_to_modify_QP_state_to_RTR __pyx_string_tab[28]
#define __pyx_kp_u_Failed_to_modify_QP_state_to_RTS __pyx_string_tab[29]
#define __pyx_kp_u_Failed_to_modify_QP_state_to_ini __pyx_string_tab[30]
#define __pyx_kp_u_Failed_to_post_recv __pyx_string_tab[31]
#define __pyx_kp_u_Failed_to_post_send __pyx_string_tab[32]
#define __pyx_kp_u_Failed_to_query_ECE __pyx_string_tab[33]
#define __pyx_kp_u_Failed_to_query_QP __pyx_string_tab[34]
#define __pyx_kp_u_Failed_to_set_ECE __pyx_string_tab[35]
#define __pyx_kp_u_Indirection_Table __pyx_string_tab[36]
#define __pyx_kp_u_Max_RD_atomic __pyx_string_tab[37]
#define __pyx_kp_u_Max_TSO_header __pyx_string_tab[38]
#define __pyx_kp_u_Max_dest_RD_atomic __pyx_string_tab[39]
#define __pyx_kp_u_Min_RNR_timer __pyx_string_tab[40]
#define __pyx_kp_u_Note_that_Cython_is_deliberately __pyx_string_tab[41]
#define __pyx_kp_u_PCI_write_end_padding __pyx_string_tab[42]
#define __pyx_kp_u_PKey_index __pyx_string_tab[43]
#define __pyx_kp_u_Path_MTU __pyx_string_tab[44]
#define __pyx_kp_u_Path_mig_state __pyx_string_tab[45]
#define __pyx_kp_u_Pickling_of_struct_members_such __pyx_string_tab[46]
#define __pyx_kp_u_Pickling_of_struct_members_such_2 __pyx_string_tab[47]
#define __pyx_kp_u_Pickling_of_struct_members_such_3 __pyx_string_tab[48]
#define __pyx_kp_u_Port_number __pyx_string_tab[49]
#define __pyx_kp_u_QP_access_flags __pyx_string_tab[50]
#define __pyx_kp_u_QP_caps __pyx_string_tab[51]
#define __pyx_kp_u_QP_current_state __pyx_string_tab[52]
#define __pyx_kp_u_QP_state __pyx_string_tab[53]
#define __pyx_kp_u_QP_type __pyx_string_tab[54]
#define __pyx_kp_u_RNR_retry __pyx_string_tab[55]
#define __pyx_kp_u_RQ_PSN __pyx_string_tab[56]
#define __pyx_kp_u_RX_hash __pyx_string_tab[57]
#define __pyx_kp_u_Rate_limit __pyx_string_tab[58]
#define __pyx_kp_u_Retry_counter __pyx_string_tab[59]
#define __pyx_kp_u_SQ_PSN __pyx_string_tab[60]
#define __pyx_kp_u_SQ_draining __pyx_string_tab[61]
#define __pyx_kp_u_SQ_sig_all __pyx_string_tab[62]
#define __pyx_kp_u_Scatter_FCS __pyx_string_tab[63]
#define __pyx_kp_u_Source_QPN __pyx_string_tab[64]
#define __pyx_kp_u_Unrecognized_object_type __pyx_string_tab[65]
#define __pyx_kp_u_Vendor_ID __pyx_string_tab[66]
#define __pyx_kp_u__4 __pyx_string_tab[67]
#define __pyx_kp_u__5 __pyx_string_tab[68]
#define __pyx_kp_u__9 __pyx_string_tab[69]
#define __pyx_kp_u_add_note __pyx_string_tab[70]
#define __pyx_kp_u_comp_mask_2 __pyx_string_tab[71]
#define __pyx_kp_u_create_flags_2 __pyx_string_tab[72]
#define __pyx_kp_u_disable __pyx_string_tab[73]
#define __pyx_kp_u_enable __pyx_string_tab[74]
#define __pyx_kp_u_enum_class_ibv_mig_state_not_imp __pyx_string_tab[75]
#define __pyx_kp_u_enum_class_ibv_mtu_not_importabl __pyx_string_tab[76]
#define __pyx_kp_u_enum_class_ibv_qp_attr_mask_not __pyx_string_tab[77]
#define __pyx_kp_u_enum_class_ibv_qp_init_attr_mask __pyx_string_tab[78]
#define __pyx_kp_u_enum_class_ibv_qp_state_not_impo __pyx_string_tab[79]
#define __pyx_kp_u_enum_class_ibv_qp_type_not_impor __pyx_string_tab[80]
#define __pyx_kp_u_flow_label_2 __pyx_string_tab[81]
#define __pyx_kp_u_gc __pyx_string_tab[82]
#define __pyx_kp_u_hop_limit_2 __pyx_string_tab[83]
#define __pyx_kp_u_ibv_wr_complete_failed __pyx_string_tab[84]
#define __pyx_kp_u_is_global_2 __pyx_string_tab[85]
#define __pyx_kp_u_isenabled __pyx_string_tab[86]
#define __pyx_kp_u_max_TSO_header __pyx_string_tab[87]
#define __pyx_kp_u_max_inline_data_2 __pyx_string_tab[88]
#define __pyx_kp_u_max_inline_data_3 __pyx_string_tab[89]
#define __pyx_kp_u_max_recv_SGE __pyx_string_tab[90]
#define __pyx_kp_u_max_recv_SGE_2 __pyx_string_tab[91]
#define __pyx_kp_u_max_recv_WR __pyx_string_tab[92]
#define __pyx_kp_u_max_recv_WR_2 __pyx_string_tab[93]
#define __pyx_kp_u_max_recv_sges __pyx_string_tab[94]
#define __pyx_kp_u_max_recv_wrs __pyx_string_tab[95]
#define __pyx_kp_u_max_send_SGE __pyx_string_tab[96]
#define __pyx_kp_u_max_send_SGE_2 __pyx_string_tab[97]
#define __pyx_kp_u_max_send_WR __pyx_string_tab[98]
#define __pyx_kp_u_max_send_WR_2 __pyx_string_tab[99]
#define __pyx_kp_u_max_send_sges __pyx_string_tab[100]
#define __pyx_kp_u_max_send_wrs __pyx_string_tab[101]
#define __pyx_kp_u_number __pyx_string_tab[102]
#define __pyx_kp_u_port_num_2 __pyx_string_tab[103]
#define __pyx_kp_u_pyverbs_qp_pyx __pyx_string_tab[104]
#define __pyx_kp_u_qp_ex_is_not_accessible_since_IB __pyx_string_tab[105]
#define __pyx_kp_u_self_attr_cannot_be_converted_to __pyx_string_tab[106]
#define __pyx_kp_u_self_data_cannot_be_converted_to __pyx_string_tab[107]
#define __pyx_kp_u_self_qp_cannot_be_converted_to_a __pyx_string_tab[108]
#define __pyx_kp_u_self_qp_self_qp_ex_cannot_be_con __pyx_string_tab[109]
#define __pyx_kp_u_sgid_index_2 __pyx_string_tab[110]
#define __pyx_kp_u_source_path_bits __pyx_string_tab[111]
#define __pyx_kp_u_state __pyx_string_tab[112]
#define __pyx_kp_u_static_rate_2 __pyx_string_tab[113]
#define __pyx_kp_u_tree_fragment __pyx_string_tab[114]
#define __pyx_n_u_DataBuffer __pyx_string_tab[115]
#define __pyx_n_u_DataBuffer___reduce_cython __pyx_string_tab[116]
#define __pyx_n_u_DataBuffer___setstate_cython __pyx_string_tab[117]
#define __pyx_n_u_ECE_2 __pyx_string_tab[118]
#define __pyx_n_u_ECE___reduce_cython __pyx_string_tab[119]
#define __pyx_n_u_ECE___setstate_cython __pyx_string_tab[120]
#define __pyx_n_u_IBV_MIG_ARMED __pyx_string_tab[121]
#define __pyx_n_u_IBV_MIG_MIGRATED __pyx_string_tab[122]
#define __pyx_n_u_IBV_MIG_REARM __pyx_string_tab[123]
#define __pyx_n_u_IBV_MTU_1024 __pyx_string_tab[124]
#define __pyx_n_u_IBV_MTU_2048 __pyx_string_tab[125]
#define __pyx_n_u_IBV_MTU_256 __pyx_string_tab[126]
#define __pyx_n_u_IBV_MTU_4096 __pyx_string_tab[127]
#define __pyx_n_u_IBV_MTU_512 __pyx_string_tab[128]
#define __pyx_n_u_IBV_QPS_ERR __pyx_string_tab[129]
#define __pyx_n_u_IBV_QPS_INIT __pyx_string_tab[130]
#define __pyx_n_u_IBV_QPS_RESET __pyx_string_tab[131]
#define __pyx_n_u_IBV_QPS_RTR __pyx_string_tab[132]
#define __pyx_n_u_IBV_QPS_RTS __pyx_string_tab[133]
#define __pyx_n_u_IBV_QPS_SQD __pyx_string_tab[134]
#define __pyx_n_u_IBV_QPS_SQE __pyx_string_tab[135]
#define __pyx_n_u_IBV_QPS_UNKNOWN __pyx_string_tab[136]
#define __pyx_n_u_IBV_QPT_DRIVER __pyx_string_tab[137]
#define __pyx_n_u_IBV_QPT_RAW_PACKET __pyx_string_tab[138]
#define __pyx_n_u_IBV_QPT_RC __pyx_string_tab[139]
#define __pyx_n_u_IBV_QPT_UC __pyx_string_tab[140]
#define __pyx_n_u_IBV_QPT_UD __pyx_string_tab[141]
#define __pyx_n_u_IBV_QPT_XRC_RECV __pyx_string_tab[142]
#define __pyx_n_u_IBV_QPT_XRC_SEND __pyx_string_tab[143]
#define __pyx_n_u_IBV_QP_ACCESS_FLAGS __pyx_string_tab[144]
#define __pyx_n_u_IBV_QP_ALT_PATH __pyx_string_tab[145]
#define __pyx_n_u_IBV_QP_AV __pyx_string_tab[146]
#define __pyx_n_u_IBV_QP_CAP __pyx_string_tab[147]
#define __pyx_n_u_IBV_QP_CUR_STATE __pyx_string_tab[148]
#define __pyx_n_u_IBV_QP_DEST_QPN __pyx_string_tab[149]
#define __pyx_n_u_IBV_QP_EN_SQD_ASYNC_NOTIFY __pyx_string_tab[150]
#define __pyx_n_u_IBV_QP_INIT_ATTR_CREATE_FLAGS __pyx_string_tab[151]
#define __pyx_n_u_IBV_QP_INIT_ATTR_IND_TABLE __pyx_string_tab[152]
#define __pyx_n_u_IBV_QP_INIT_ATTR_MAX_TSO_HEADER __pyx_string_tab[153]
#define __pyx_n_u_IBV_QP_INIT_ATTR_PD __pyx_string_tab[154]
#define __pyx_n_u_IBV_QP_INIT_ATTR_RX_HASH __pyx_string_tab[155]
#define __pyx_n_u_IBV_QP_INIT_ATTR_SEND_OPS_FLAGS __pyx_string_tab[156]
#define __pyx_n_u_IBV_QP_INIT_ATTR_XRCD __pyx_string_tab[157]
#define __pyx_n_u_IBV_QP_MAX_DEST_RD_ATOMIC __pyx_string_tab[158]
#define __pyx_n_u_IBV_QP_MAX_QP_RD_ATOMIC __pyx_string_tab[159]
#define __pyx_n_u_IBV_QP_MIN_RNR_TIMER __pyx_string_tab[160]
#define __pyx_n_u_IBV_QP_PATH_MIG_STATE __pyx_string_tab[161]
#define __pyx_n_u_IBV_QP_PATH_MTU __pyx_string_tab[162]
#define __pyx_n_u_IBV_QP_PKEY_INDEX __pyx_string_tab[163]
#define __pyx_n_u_IBV_QP_PORT __pyx_string_tab[164]
#define __pyx_n_u_IBV_QP_QKEY __pyx_string_tab[165]
#define __pyx_n_u_IBV_QP_RATE_LIMIT __pyx_string_tab[166]
#define __pyx_n_u_IBV_QP_RETRY_CNT __pyx_string_tab[167]
#define __pyx_n_u_IBV_QP_RNR_RETRY __pyx_string_tab[168]
#define __pyx_n_u_IBV_QP_RQ_PSN __pyx_string_tab[169]
#define __pyx_n_u_IBV_QP_SQ_PSN __pyx_string_tab[170]
#define __pyx_n_u_IBV_QP_STATE __pyx_string_tab[171]
#define __pyx_n_u_IBV_QP_TIMEOUT __pyx_string_tab[172]
#define __pyx_n_u_INIT __pyx_string_tab[173]
#define __pyx_n_u_Options __pyx_string_tab[174]
#define __pyx_n_u_PD __pyx_string_tab[175]
#define __pyx_n_u_PyverbsError __pyx_string_tab[176]
#define __pyx_n_u_PyverbsRDMAErrno __pyx_string_tab[177]
#define __pyx_n_u_PyverbsRDMAError __pyx_string_tab[178]
#define __pyx_n_u_PyverbsUserError __pyx_string_tab[179]
#define __pyx_n_u_Pyx_PyDict_NextRef __pyx_string_tab[180]
#define __pyx_n_u_QKey __pyx_string_tab[181]
#define __pyx_n_u_QP __pyx_string_tab[182]
#define __pyx_n_u_QPAttr __pyx_string_tab[183]
#define __pyx_n_u_QPAttr___reduce_cython __pyx_string_tab[184]
#define __pyx_n_u_QPAttr___setstate_cython __pyx_string_tab[185]
#define __pyx_n_u_QPAttr__copy_ah __pyx_string_tab[186]
#define __pyx_n_u_QPCap __pyx_string_tab[187]
#define __pyx_n_u_QPCap___reduce_cython __pyx_string_tab[188]
#define __pyx_n_u_QPCap___setstate_cython __pyx_string_tab[189]
#define __pyx_n_u_QPEx __pyx_string_tab[190]
#define __pyx_n_u_QPEx___reduce_cython __pyx_string_tab[191]
#define __pyx_n_u_QPEx___setstate_cython __pyx_string_tab[192]
#define __pyx_n_u_QPEx_wr_abort __pyx_string_tab[193]
#define __pyx_n_u_QPEx_wr_atomic_cmp_swp __pyx_string_tab[194]
#define __pyx_n_u_QPEx_wr_atomic_fetch_add __pyx_string_tab[195]
#define __pyx_n_u_QPEx_wr_atomic_write __pyx_string_tab[196]
#define __pyx_n_u_QPEx_wr_bind_mw __pyx_string_tab[197]
#define __pyx_n_u_QPEx_wr_complete __pyx_string_tab[198]
#define __pyx_n_u_QPEx_wr_flush __pyx_string_tab[199]
#define __pyx_n_u_QPEx_wr_local_inv __pyx_string_tab[200]
#define __pyx_n_u_QPEx_wr_rdma_read __pyx_string_tab[201]
#define __pyx_n_u_QPEx_wr_rdma_write __pyx_string_tab[202]
#define __pyx_n_u_QPEx_wr_rdma_write_imm __pyx_string_tab[203]
#define __pyx_n_u_QPEx_wr_send __pyx_string_tab[204]
#define __pyx_n_u_QPEx_wr_send_imm __pyx_string_tab[205]
#define __pyx_n_u_QPEx_wr_send_inv __pyx_string_tab[206]
#define __pyx_n_u_QPEx_wr_send_tso __pyx_string_tab[207]
#define __pyx_n_u_QPEx_wr_set_inline_data __pyx_string_tab[208]
#define __pyx_n_u_QPEx_wr_set_inline_data_list __pyx_string_tab[209]
#define __pyx_n_u_QPEx_wr_set_sge __pyx_string_tab[210]
#define __pyx_n_u_QPEx_wr_set_sge_list __pyx_string_tab[211]
#define __pyx_n_u_QPEx_wr_set_ud_addr __pyx_string_tab[212]
#define __pyx_n_u_QPEx_wr_set_xrc_srqn __pyx_string_tab[213]
#define __pyx_n_u_QPEx_wr_start __pyx_string_tab[214]
#define __pyx_n_u_QPInitAttr __pyx_string_tab[215]
#define __pyx_n_u_QPInitAttrEx __pyx_string_tab[216]
#define __pyx_n_u_QPInitAttrEx___reduce_cython __pyx_string_tab[217]
#define __pyx_n_u_QPInitAttrEx___setstate_cython __pyx_string_tab[218]
#define __pyx_n_u_QPInitAttrEx_flags_to_str __pyx_string_tab[219]
#define __pyx_n_u_QPInitAttrEx_mask_to_str __pyx_string_tab[220]
#define __pyx_n_u_QPInitAttr___reduce_cython __pyx_string_tab[221]
#define __pyx_n_u_QPInitAttr___setstate_cython __pyx_string_tab[222]
#define __pyx_n_u_QP___reduce_cython __pyx_string_tab[223]
#define __pyx_n_u_QP___setstate_cython __pyx_string_tab[224]
#define __pyx_n_u_QP__create_qp __pyx_string_tab[225]
#define __pyx_n_u_QP__create_qp_ex __pyx_string_tab[226]
#define __pyx_n_u_QP__get_comp_mask __pyx_string_tab[227]
#define __pyx_n_u_QP_bind_mw __pyx_string_tab[228]
#define __pyx_n_u_QP_close __pyx_string_tab[229]
#define __pyx_n_u_QP_modify __pyx_string_tab[230]
#define __pyx_n_u_QP_post_recv __pyx_string_tab[231]
#define __pyx_n_u_QP_post_send __pyx_string_tab[232]
#define __pyx_n_u_QP_query __pyx_string_tab[233]
#define __pyx_n_u_QP_query_data_in_order __pyx_string_tab[234]
#define __pyx_n_u_QP_query_ece __pyx_string_tab[235]
#define __pyx_n_u_QP_set_ece __pyx_string_tab[236]
#define __pyx_n_u_QP_to_init __pyx_string_tab[237]
#define __pyx_n_u_QP_to_rtr __pyx_string_tab[238]
#define __pyx_n_u_QP_to_rts __pyx_string_tab[239]
#define __pyx_n_u_RTR __pyx_string_tab[240]
#define __pyx_n_u_RTS __pyx_string_tab[241]
#define __pyx_n_u_Timeout __pyx_string_tab[242]
#define __pyx_n_u_WeakSet __pyx_string_tab[243]
#define __pyx_n_u_XRCD __pyx_string_tab[244]
#define __pyx_n_u_access_flags_to_str __pyx_string_tab[245]
#define __pyx_n_u_add __pyx_string_tab[246]
#define __pyx_n_u_addr __pyx_string_tab[247]
#define __pyx_n_u_ah __pyx_string_tab[248]
#define __pyx_n_u_ah_attr __pyx_string_tab[249]
#define __pyx_n_u_asyncio_coroutines __pyx_string_tab[250]
#define __pyx_n_u_atomic_wr __pyx_string_tab[251]
#define __pyx_n_u_atomic_wr_c __pyx_string_tab[252]
#define __pyx_n_u_attr __pyx_string_tab[253]
#define __pyx_n_u_attr_mask __pyx_string_tab[254]
#define __pyx_n_u_bad_wr __pyx_string_tab[255]
#define __pyx_n_u_bind_info __pyx_string_tab[256]
#define __pyx_n_u_bind_mw __pyx_string_tab[257]
#define __pyx_n_u_buf_list __pyx_string_tab[258]
#define __pyx_n_u_cap __pyx_string_tab[259]
#define __pyx_n_u_class_getitem __pyx_string_tab[260]
#define __pyx_n_u_cline_in_traceback __pyx_string_tab[261]
#define __pyx_n_u_close __pyx_string_tab[262]
#define __pyx_n_u_comp_mask __pyx_string_tab[263]
#define __pyx_n_u_comp_masks __pyx_string_tab[264]
#define __pyx_n_u_compare __pyx_string_tab[265]
#define __pyx_n_u_copy_ah __pyx_string_tab[266]
#define __pyx_n_u_copy_caps __pyx_string_tab[267]
#define __pyx_n_u_create_flags __pyx_string_tab[268]
#define __pyx_n_u_create_qp __pyx_string_tab[269]
#define __pyx_n_u_create_qp_ex __pyx_string_tab[270]
#define __pyx_n_u_create_str __pyx_string_tab[271]
#define __pyx_n_u_creator __pyx_string_tab[272]
#define __pyx_n_u_ctx __pyx_string_tab[273]
#define __pyx_n_u_cur_qp_state __pyx_string_tab[274]
#define __pyx_n_u_data __pyx_string_tab[275]
#define __pyx_n_u_data_buf __pyx_string_tab[276]
#define __pyx_n_u_debug __pyx_string_tab[277]
#define __pyx_n_u_dgid __pyx_string_tab[278]
#define __pyx_n_u_dlid __pyx_string_tab[279]
#define __pyx_n_u_dst __pyx_string_tab[280]
#define __pyx_n_u_ece __pyx_string_tab[281]
#define __pyx_n_u_f __pyx_string_tab[282]
#define __pyx_n_u_flags __pyx_string_tab[283]
#define __pyx_n_u_flags_to_str __pyx_string_tab[284]
#define __pyx_n_u_flow_label __pyx_string_tab[285]
#define __pyx_n_u_format __pyx_string_tab[286]
#define __pyx_n_u_func __pyx_string_tab[287]
#define __pyx_n_u_get_comp_mask __pyx_string_tab[288]
#define __pyx_n_u_getstate __pyx_string_tab[289]
#define __pyx_n_u_gid_str __pyx_string_tab[290]
#define __pyx_n_u_gr __pyx_string_tab[291]
#define __pyx_n_u_hash_conf __pyx_string_tab[292]
#define __pyx_n_u_hdr __pyx_string_tab[293]
#define __pyx_n_u_hdr_sz __pyx_string_tab[294]
#define __pyx_n_u_hop_limit __pyx_string_tab[295]
#define __pyx_n_u_i __pyx_string_tab[296]
#define __pyx_n_u_ibv_mig_state __pyx_string_tab[297]
#define __pyx_n_u_ibv_mtu __pyx_string_tab[298]
#define __pyx_n_u_ibv_qp_attr_mask __pyx_string_tab[299]
#define __pyx_n_u_ibv_qp_init_attr_mask __pyx_string_tab[300]
#define __pyx_n_u_ibv_qp_state __pyx_string_tab[301]
#define __pyx_n_u_ibv_qp_type __pyx_string_tab[302]
#define __pyx_n_u_imm_data __pyx_string_tab[303]
#define __pyx_n_u_ind_table __pyx_string_tab[304]
#define __pyx_n_u_info __pyx_string_tab[305]
#define __pyx_n_u_init __pyx_string_tab[306]
#define __pyx_n_u_init_attr __pyx_string_tab[307]
#define __pyx_n_u_invalidate_rkey __pyx_string_tab[308]
#define __pyx_n_u_is_alt __pyx_string_tab[309]
#define __pyx_n_u_is_coroutine __pyx_string_tab[310]
#define __pyx_n_u_is_global __pyx_string_tab[311]
#define __pyx_n_u_items __pyx_string_tab[312]
#define __pyx_n_u_length __pyx_string_tab[313]
#define __pyx_n_u_level __pyx_string_tab[314]
#define __pyx_n_u_lkey __pyx_string_tab[315]
#define __pyx_n_u_main __pyx_string_tab[316]
#define __pyx_n_u_mask __pyx_string_tab[317]
#define __pyx_n_u_mask_str __pyx_string_tab[318]
#define __pyx_n_u_mask_to_str __pyx_string_tab[319]
#define __pyx_n_u_masks __pyx_string_tab[320]
#define __pyx_n_u_max_inline_data __pyx_string_tab[321]
#define __pyx_n_u_max_recv_sge __pyx_string_tab[322]
#define __pyx_n_u_max_recv_wr __pyx_string_tab[323]
#define __pyx_n_u_max_send_sge __pyx_string_tab[324]
#define __pyx_n_u_max_send_wr __pyx_string_tab[325]
#define __pyx_n_u_max_tso_header __pyx_string_tab[326]
#define __pyx_n_u_mig_state_to_str __pyx_string_tab[327]
#define __pyx_n_u_modify __pyx_string_tab[328]
#define __pyx_n_u_module __pyx_string_tab[329]
#define __pyx_n_u_mss __pyx_string_tab[330]
#define __pyx_n_u_mtu_to_str __pyx_string_tab[331]
#define __pyx_n_u_mw __pyx_string_tab[332]
#define __pyx_n_u_mw_bind __pyx_string_tab[333]
#define __pyx_n_u_my_bad_wr __pyx_string_tab[334]
#define __pyx_n_u_name __pyx_string_tab[335]
#define __pyx_n_u_num_buf __pyx_string_tab[336]
#define __pyx_n_u_num_sge __pyx_string_tab[337]
#define __pyx_n_u_op __pyx_string_tab[338]
#define __pyx_n_u_options __pyx_string_tab[339]
#define __pyx_n_u_path_mtu __pyx_string_tab[340]
#define __pyx_n_u_pd __pyx_string_tab[341]
#define __pyx_n_u_pop __pyx_string_tab[342]
#define __pyx_n_u_port_num __pyx_string_tab[343]
#define __pyx_n_u_post_recv __pyx_string_tab[344]
#define __pyx_n_u_post_send __pyx_string_tab[345]
#define __pyx_n_u_ptr __pyx_string_tab[346]
#define __pyx_n_u_ptype __pyx_string_tab[347]
#define __pyx_n_u_pyverbs_base __pyx_string_tab[348]
#define __pyx_n_u_pyverbs_libibverbs_enums __pyx_string_tab[349]
#define __pyx_n_u_pyverbs_pyverbs_error __pyx_string_tab[350]
#define __pyx_n_u_pyverbs_qp __pyx_string_tab[351]
#define __pyx_n_u_pyverbs_utils __pyx_string_tab[352]
#define __pyx_n_u_pyx_state __pyx_string_tab[353]
#define __pyx_n_u_pyx_vtable __pyx_string_tab[354]
#define __pyx_n_u_qa __pyx_string_tab[355]
#define __pyx_n_u_qia __pyx_string_tab[356]
#define __pyx_n_u_qiae __pyx_string_tab[357]
#define __pyx_n_u_qp_attr __pyx_string_tab[358]
#define __pyx_n_u_qp_context __pyx_string_tab[359]
#define __pyx_n_u_qp_num __pyx_string_tab[360]
#define __pyx_n_u_qp_state __pyx_string_tab[361]
#define __pyx_n_u_qp_state_to_str __pyx_string_tab[362]
#define __pyx_n_u_qp_type __pyx_string_tab[363]
#define __pyx_n_u_qp_type_to_str __pyx_string_tab[364]
#define __pyx_n_u_qualname __pyx_string_tab[365]
#define __pyx_n_u_query __pyx_string_tab[366]
#define __pyx_n_u_query_data_in_order __pyx_string_tab[367]
#define __pyx_n_u_query_ece __pyx_string_tab[368]
#define __pyx_n_u_rc __pyx_string_tab[369]
#define __pyx_n_u_rcq __pyx_string_tab[370]
#define __pyx_n_u_recv_cq __pyx_string_tab[371]
#define __pyx_n_u_reduce __pyx_string_tab[372]
#define __pyx_n_u_reduce_cython __pyx_string_tab[373]
#define __pyx_n_u_reduce_ex __pyx_string_tab[374]
#define __pyx_n_u_remote_addr __pyx_string_tab[375]
#define __pyx_n_u_remote_qpn __pyx_string_tab[376]
#define __pyx_n_u_remote_rkey __pyx_string_tab[377]
#define __pyx_n_u_remote_srqn __pyx_string_tab[378]
#define __pyx_n_u_rkey __pyx_string_tab[379]
#define __pyx_n_u_scq __pyx_string_tab[380]
#define __pyx_n_u_self __pyx_string_tab[381]
#define __pyx_n_u_send_cq __pyx_string_tab[382]
#define __pyx_n_u_send_ops_flags __pyx_string_tab[383]
#define __pyx_n_u_set_ece __pyx_string_tab[384]
#define __pyx_n_u_set_name __pyx_string_tab[385]
#define __pyx_n_u_setdefault __pyx_string_tab[386]
#define __pyx_n_u_setstate __pyx_string_tab[387]
#define __pyx_n_u_setstate_cython __pyx_string_tab[388]
#define __pyx_n_u_sg_list __pyx_string_tab[389]
#define __pyx_n_u_sge __pyx_string_tab[390]
#define __pyx_n_u_sgid_index __pyx_string_tab[391]
#define __pyx_n_u_sl __pyx_string_tab[392]
#define __pyx_n_u_source_qpn __pyx_string_tab[393]
#define __pyx_n_u_sq_sig_all __pyx_string_tab[394]
#define __pyx_n_u_src __pyx_string_tab[395]
#define __pyx_n_u_src_path_bits __pyx_string_tab[396]
#define __pyx_n_u_srq __pyx_string_tab[397]
#define __pyx_n_u_static_rate __pyx_string_tab[398]
#define __pyx_n_u_super __pyx_string_tab[399]
#define __pyx_n_u_swap __pyx_string_tab[400]
#define __pyx_n_u_t __pyx_string_tab[401]
#define __pyx_n_u_test __pyx_string_tab[402]
#define __pyx_n_u_to_init __pyx_string_tab[403]
#define __pyx_n_u_to_rtr __pyx_string_tab[404]
#define __pyx_n_u_to_rts __pyx_string_tab[405]
#define __pyx_n_u_traffic_class __pyx_string_tab[406]
#define __pyx_n_u_values __pyx_string_tab[407]
#define __pyx_n_u_vendor_id __pyx_string_tab[408]
#define __pyx_n_u_warn __pyx_string_tab[409]
#define __pyx_n_u_warnings __pyx_string_tab[410]
#define __pyx_n_u_weakref __pyx_string_tab[411]
#define __pyx_n_u_wr __pyx_string_tab[412]
#define __pyx_n_u_wr_abort __pyx_string_tab[413]
#define __pyx_n_u_wr_atomic_cmp_swp __pyx_string_tab[414]
#define __pyx_n_u_wr_atomic_fetch_add __pyx_string_tab[415]
#define __pyx_n_u_wr_atomic_write __pyx_string_tab[416]
#define __pyx_n_u_wr_bind_mw __pyx_string_tab[417]
#define __pyx_n_u_wr_complete __pyx_string_tab[418]
#define __pyx_n_u_wr_flush __pyx_string_tab[419]
#define __pyx_n_u_wr_local_inv __pyx_string_tab[420]
#define __pyx_n_u_wr_rdma_read __pyx_string_tab[421]
#define __pyx_n_u_wr_rdma_write __pyx_string_tab[422]
#define __pyx_n_u_wr_rdma_write_imm __pyx_string_tab[423]
#define __pyx_n_u_wr_send __pyx_string_tab[424]
#define __pyx_n_u_wr_send_imm __pyx_string_tab[425]
#define __pyx_n_u_wr_send_inv __pyx_string_tab[426]
#define __pyx_n_u_wr_send_tso __pyx_string_tab[427]
#define __pyx_n_u_wr_set_inline_data __pyx_string_tab[428]
#define __pyx_n_u_wr_set_inline_data_list __pyx_string_tab[429]
#define __pyx_n_u_wr_set_sge __pyx_string_tab[430]
#define __pyx_n_u_wr_set_sge_list __pyx_string_tab[431]
#define __pyx_n_u_wr_set_ud_addr __pyx_string_tab[432]
#define __pyx_n_u_wr_set_xrc_srqn __pyx_string_tab[433]
#define __pyx_n_u_wr_start __pyx_string_tab[434]
#define __pyx_n_u_xrcd __pyx_string_tab[435]
#define __pyx_kp_b_PyObject_PyObject_close_weakrefs __pyx_string_tab[436]
#define __pyx_kp_b_iso88591_A_1D_T __pyx_string_tab[437]
#define __pyx_kp_b_iso88591_A_1D_a __pyx_string_tab[438]
#define __pyx_kp_b_iso88591_A_3d_S_l_4uAS_3c_8 __pyx_string_tab[439]
#define __pyx_kp_b_iso88591_A_4t3a_t1_G6_1D_d_a_a_q_a_q_a_a __pyx_string_tab[440]
#define __pyx_kp_b_iso88591_A_4xq __pyx_string_tab[441]
#define __pyx_kp_b_iso88591_A_4xvQ __pyx_string_tab[442]
#define __pyx_kp_b_iso88591_A_4z_Q_q_t_1_A_nAT_awgQ_3c_Gq __pyx_string_tab[443]
#define __pyx_kp_b_iso88591_A_4z_Q_t_1_A_nAT_awgQ_3c_Gq __pyx_string_tab[444]
#define __pyx_kp_b_iso88591_A_8_1_c_A_Q_E_vRq_l_1_a_q __pyx_string_tab[445]
#define __pyx_kp_b_iso88591_A_AT_5_A __pyx_string_tab[446]
#define __pyx_kp_b_iso88591_A_A_axr_5_1_aq_E_aq_81A_885_XU_Q __pyx_string_tab[447]
#define __pyx_kp_b_iso88591_A_C_A_7x_2_5Fa_0_A_8_336H_0_1_y __pyx_string_tab[448]
#define __pyx_kp_b_iso88591_A_G_AS_4q __pyx_string_tab[449]
#define __pyx_kp_b_iso88591_A_G_E_a __pyx_string_tab[450]
#define __pyx_kp_b_iso88591_A_Qd __pyx_string_tab[451]
#define __pyx_kp_b_iso88591_A_Qd_2 __pyx_string_tab[452]
#define __pyx_kp_b_iso88591_A_Qd_Q __pyx_string_tab[453]
#define __pyx_kp_b_iso88591_A_WAQ_XQ __pyx_string_tab[454]
#define __pyx_kp_b_iso88591_A_WAQ_q_HF_q __pyx_string_tab[455]
#define __pyx_kp_b_iso88591_A_XS_s_A __pyx_string_tab[456]
#define __pyx_kp_b_iso88591_A_a_l_1_4s_aq_E_aq_q_A_q_G1Ba_q __pyx_string_tab[457]
#define __pyx_kp_b_iso88591_A_aq_XU __pyx_string_tab[458]
#define __pyx_kp_b_iso88591_A_aq_he81 __pyx_string_tab[459]
#define __pyx_kp_b_iso88591_A_at1 __pyx_string_tab[460]
#define __pyx_kp_b_iso88591_A_at1_3c_a __pyx_string_tab[461]
#define __pyx_kp_b_iso88591_A_at86_hgQ __pyx_string_tab[462]
#define __pyx_kp_b_iso88591_A_c_nAT_as_3c_q __pyx_string_tab[463]
#define __pyx_kp_b_iso88591_A_c_s_Q_4q_1_E_uBa_Jaq_A_q __pyx_string_tab[464]
#define __pyx_kp_b_iso88591_A_ha __pyx_string_tab[465]
#define __pyx_kp_b_iso88591_A_hfM __pyx_string_tab[466]
#define __pyx_kp_b_iso88591_A_l_4uBe1G1_3c_8 __pyx_string_tab[467]
#define __pyx_kp_b_iso88591_A_nAT_awgQ_3c __pyx_string_tab[468]
#define __pyx_kp_b_iso88591_A_q_HF_y __pyx_string_tab[469]
#define __pyx_kp_b_iso88591_A_q_X_q_a_HAQ __pyx_string_tab[470]
#define __pyx_kp_b_iso88591_A_t_1_A_Qd_q_wa_3c_H __pyx_string_tab[471]
#define __pyx_kp_b_iso88591_A_vQ_Ja_m1D_Qd_AYa_3c_9_vQ __pyx_string_tab[472]
#define __pyx_kp_b_iso88591_Q __pyx_string_tab[473]
#define __pyx_kp_b_iso88591_a_83a_4q_U_1_E_U_aq_84uD_XT_whd __pyx_string_tab[474]
#define __pyx_kp_b_iso88591_nAT_ar_1A_3c_aq_j_7 __pyx_string_tab[475]
#define __pyx_kp_b_iso88591_t3a_t1E_A_l_5_O3d_5_O3d_5_Ct1_5 __pyx_string_tab[476]
#define __pyx_int_0 __pyx_number_tab[0]
#define __pyx_int_1 __pyx_number_tab[1]
#define __pyx_int_2 __pyx_number_tab[2]
#define __pyx_int_4 __pyx_number_tab[3]
#define __pyx_int_8 __pyx_number_tab[4]
#define __pyx_int_10 __pyx_number_tab[5]
#define __pyx_int_16 __pyx_number_tab[6]
#define __pyx_int_32 __pyx_number_tab[7]
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_clear(PyObject *m) {
  __pyx_mstatetype *clear_module_state = __Pyx_PyModule_GetState(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __Pyx_State_RemoveModule(NULL);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4base_PyverbsObject);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4base_PyverbsCM);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_Context);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_DeviceAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_QueryDeviceExInput);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_ODPCaps);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_RSSCaps);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_PacketPacingCaps);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_PCIAtomicCaps);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_TMCaps);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_CQModerationCaps);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_TSOCaps);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_DeviceAttrEx);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_AllocDmAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_DM);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_PortAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_GIDEntry);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_AsyncEvent);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_FdArr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2cq_CompChannel);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2cq_CQ);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2cq_CQEX);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2cq_WC);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2cq_PollCqAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2cq_WcTmInfo);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2pd_PD);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2pd_ParentDomainInitAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2pd_ParentDomain);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2pd_ParentDomainContext);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2wq_WQInitAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2wq_WQAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2wq_WQ);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2wq_RwqIndTableInitAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2wq_RwqIndTable);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2wq_RxHashConf);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2mr_MR);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2mr_MREx);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2mr_MWBindInfo);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2mr_MWBind);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2mr_MW);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2mr_DMMR);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2mr_DmaBufMR);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2mr_DmaHandleInitAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2mr_DMAHandle);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2wr_SGE);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2wr_RecvWR);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2wr_SendWR);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4cmid_CMID);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4cmid_CMEventChannel);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4cmid_CMEvent);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4cmid_AddrInfo);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4cmid_ConnParam);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4cmid_UDParam);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4addr_GID);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4addr_GRH);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4addr_GlobalRoute);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4addr_AHAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4addr_AH);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4flow_FlowAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4flow_Flow);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4flow_FlowAction);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4xrcd_XRCDInitAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4xrcd_XRCD);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_3srq_SrqAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_3srq_SrqInitAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_3srq_SrqInitAttrEx);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_3srq_OpsWr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_3srq_SRQ);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_QPCap);
  Py_CLEAR(clear_module_state->__pyx_type_7pyverbs_2qp_QPCap);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_QPInitAttr);
  Py_CLEAR(clear_module_state->__pyx_type_7pyverbs_2qp_QPInitAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx);
  Py_CLEAR(clear_module_state->__pyx_type_7pyverbs_2qp_QPInitAttrEx);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_QPAttr);
  Py_CLEAR(clear_module_state->__pyx_type_7pyverbs_2qp_QPAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_QP);
  Py_CLEAR(clear_module_state->__pyx_type_7pyverbs_2qp_QP);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_DataBuffer);
  Py_CLEAR(clear_module_state->__pyx_type_7pyverbs_2qp_DataBuffer);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_QPEx);
  Py_CLEAR(clear_module_state->__pyx_type_7pyverbs_2qp_QPEx);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_ECE);
  Py_CLEAR(clear_module_state->__pyx_type_7pyverbs_2qp_ECE);
  Py_CLEAR(clear_module_state->__pyx_k__2);
  Py_CLEAR(clear_module_state->__pyx_k__3);
  Py_CLEAR(clear_module_state->__pyx_k__6);
  Py_CLEAR(clear_module_state->__pyx_k__7);
  Py_CLEAR(clear_module_state->__pyx_k__8);
  for (int i=0; i<3; ++i) { Py_CLEAR(clear_module_state->__pyx_tuple[i]); }
  for (int i=0; i<57; ++i) { Py_CLEAR(clear_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<477; ++i) { Py_CLEAR(clear_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<8; ++i) { Py_CLEAR(clear_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_clear_contents ### */
/* CommonTypesMetaclass.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CyFunctionType);

/* #### Code section: module_state_clear_end ### */
return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstatetype *traverse_module_state = __Pyx_PyModule_GetState(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_tuple);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_bytes);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_unicode);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4base_PyverbsObject);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4base_PyverbsCM);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_Context);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_DeviceAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_QueryDeviceExInput);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_ODPCaps);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_RSSCaps);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_PacketPacingCaps);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_PCIAtomicCaps);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_TMCaps);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_CQModerationCaps);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_TSOCaps);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_DeviceAttrEx);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_AllocDmAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_DM);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_PortAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_GIDEntry);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_AsyncEvent);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_FdArr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2cq_CompChannel);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2cq_CQ);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2cq_CQEX);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2cq_WC);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2cq_PollCqAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2cq_WcTmInfo);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2pd_PD);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2pd_ParentDomainInitAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2pd_ParentDomain);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2pd_ParentDomainContext);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2wq_WQInitAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2wq_WQAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2wq_WQ);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2wq_RwqIndTableInitAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2wq_RwqIndTable);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2wq_RxHashConf);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2mr_MR);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2mr_MREx);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2mr_MWBindInfo);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2mr_MWBind);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2mr_MW);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2mr_DMMR);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2mr_DmaBufMR);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2mr_DmaHandleInitAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2mr_DMAHandle);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2wr_SGE);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2wr_RecvWR);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2wr_SendWR);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4cmid_CMID);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4cmid_CMEventChannel);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4cmid_CMEvent);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4cmid_AddrInfo);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4cmid_ConnParam);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4cmid_UDParam);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4addr_GID);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4addr_GRH);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4addr_GlobalRoute);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4addr_AHAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4addr_AH);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4flow_FlowAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4flow_Flow);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4flow_FlowAction);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4xrcd_XRCDInitAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4xrcd_XRCD);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_3srq_SrqAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_3srq_SrqInitAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_3srq_SrqInitAttrEx);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_3srq_OpsWr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_3srq_SRQ);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_QPCap);
  Py_VISIT(traverse_module_state->__pyx_type_7pyverbs_2qp_QPCap);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_QPInitAttr);
  Py_VISIT(traverse_module_state->__pyx_type_7pyverbs_2qp_QPInitAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx);
  Py_VISIT(traverse_module_state->__pyx_type_7pyverbs_2qp_QPInitAttrEx);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_QPAttr);
  Py_VISIT(traverse_module_state->__pyx_type_7pyverbs_2qp_QPAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_QP);
  Py_VISIT(traverse_module_state->__pyx_type_7pyverbs_2qp_QP);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_DataBuffer);
  Py_VISIT(traverse_module_state->__pyx_type_7pyverbs_2qp_DataBuffer);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_QPEx);
  Py_VISIT(traverse_module_state->__pyx_type_7pyverbs_2qp_QPEx);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_ECE);
  Py_VISIT(traverse_module_state->__pyx_type_7pyverbs_2qp_ECE);
  Py_VISIT(traverse_module_state->__pyx_k__2);
  Py_VISIT(traverse_module_state->__pyx_k__3);
  Py_VISIT(traverse_module_state->__pyx_k__6);
  Py_VISIT(traverse_module_state->__pyx_k__7);
  Py_VISIT(traverse_module_state->__pyx_k__8);
  for (int i=0; i<3; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_tuple[i]); }
  for (int i=0; i<57; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<477; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<8; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_traverse_contents ### */
/* CommonTypesMetaclass.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CyFunctionType);

/* #### Code section: module_state_traverse_end ### */
return 0;
}
#endif
/* #### Code section: module_code ### */

/* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(ibv_qp_type c_val):
 *     cdef object __pyx_enum
*/

static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(enum ibv_qp_type __pyx_v_c_val) {
  PyObject *__pyx_v___pyx_enum = 0;
  PyObject *__pyx_v_warnings = NULL;
  int __pyx_v_underlying_c_val;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  size_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py", 0);

  /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_qp_type as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py == Py_None);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_qp_type as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py
 *         except ImportError:
*/
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "EnumTypeToPy":13
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py is None:
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_qp_type as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py             # <<<<<<<<<<<<<<
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py = False  # False indicates "don't try again"
*/
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_qp_type};
          __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyverbs_libibverbs_enums, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L4_error)
        }
        __pyx_t_5 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_5);
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_qp_type};
          __pyx_t_7 = 0; {
            __pyx_t_8 = __Pyx_ImportFrom(__pyx_t_5, __pyx_imported_names[__pyx_t_7]); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 13, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_8);
            switch (__pyx_t_7) {
              case 0:
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py);
              __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py, __pyx_t_8);
              __Pyx_GIVEREF(__pyx_t_8);
              break;
              default:;
            }
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_qp_type as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py
 *         except ImportError:
*/
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "EnumTypeToPy":14
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_qp_type as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py
 *         except ImportError:             # <<<<<<<<<<<<<<
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py = False  # False indicates "don't try again"
 *             import warnings
*/
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_ImportError))));
      if (__pyx_t_9) {
        __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_8, &__pyx_t_10) < 0) __PYX_ERR(1, 14, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_10);

        /* "EnumTypeToPy":15
 *             from pyverbs.libibverbs_enums import ibv_qp_type as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py = False  # False indicates "don't try again"             # <<<<<<<<<<<<<<
 *             import warnings
 *             warnings.warn(
*/
        __Pyx_INCREF(Py_False);
        __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py);
        __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py, Py_False);
        __Pyx_GIVEREF(Py_False);

        /* "EnumTypeToPy":16
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py = False  # False indicates "don't try again"
 *             import warnings             # <<<<<<<<<<<<<<
 *             warnings.warn(
 *                 f"enum class ibv_qp_type not importable from pyverbs.libibverbs_enums. "
*/
        __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_warnings, 0, 0, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 16, __pyx_L6_except_error)
        __pyx_t_11 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_v_warnings = __pyx_t_11;
        __pyx_t_11 = 0;

        /* "EnumTypeToPy":17
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py = False  # False indicates "don't try again"
 *             import warnings
 *             warnings.warn(             # <<<<<<<<<<<<<<
 *                 f"enum class ibv_qp_type not importable from pyverbs.libibverbs_enums. "
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
*/
        __pyx_t_12 = __pyx_v_warnings;
        __Pyx_INCREF(__pyx_t_12);
        __pyx_t_13 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_mstate_global->__pyx_kp_u_enum_class_ibv_qp_type_not_impor};
          __pyx_t_11 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_warn, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 17, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_11);
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L5_exception_handled;
      }
      goto __pyx_L6_except_error;

      /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_qp_type as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py
 *         except ImportError:
*/
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L5_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_L9_try_end:;
    }

    /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_qp_type as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py
*/
  }

  /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py == Py_False);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":24
 * 
 * 
 *         return <int>c_val             # <<<<<<<<<<<<<<
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyLong_From_int(((int)__pyx_v_c_val)); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumTypeToPy":25
 * 
 *         return <int>c_val
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py             # <<<<<<<<<<<<<<
 * 
 *     if 0:
*/
  __Pyx_INCREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py);
  __pyx_v___pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py;

  /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_qp_type.IBV_QPT_RC:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPT_RC
 *     elif c_val == ibv_qp_type.IBV_QPT_UC:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QPT_RC);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":30
 *         pass
 *     elif c_val == ibv_qp_type.IBV_QPT_RC:
 *         return __pyx_enum.IBV_QPT_RC             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_type.IBV_QPT_UC:
 *         return __pyx_enum.IBV_QPT_UC
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QPT_RC); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_qp_type.IBV_QPT_RC:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPT_RC
 *     elif c_val == ibv_qp_type.IBV_QPT_UC:
*/
  }

  /* "EnumTypeToPy":31
 *     elif c_val == ibv_qp_type.IBV_QPT_RC:
 *         return __pyx_enum.IBV_QPT_RC
 *     elif c_val == ibv_qp_type.IBV_QPT_UC:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPT_UC
 *     elif c_val == ibv_qp_type.IBV_QPT_UD:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QPT_UC);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":32
 *         return __pyx_enum.IBV_QPT_RC
 *     elif c_val == ibv_qp_type.IBV_QPT_UC:
 *         return __pyx_enum.IBV_QPT_UC             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_type.IBV_QPT_UD:
 *         return __pyx_enum.IBV_QPT_UD
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QPT_UC); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 32, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":31
 *     elif c_val == ibv_qp_type.IBV_QPT_RC:
 *         return __pyx_enum.IBV_QPT_RC
 *     elif c_val == ibv_qp_type.IBV_QPT_UC:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPT_UC
 *     elif c_val == ibv_qp_type.IBV_QPT_UD:
*/
  }

  /* "EnumTypeToPy":33
 *     elif c_val == ibv_qp_type.IBV_QPT_UC:
 *         return __pyx_enum.IBV_QPT_UC
 *     elif c_val == ibv_qp_type.IBV_QPT_UD:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPT_UD
 *     elif c_val == ibv_qp_type.IBV_QPT_RAW_PACKET:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QPT_UD);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":34
 *         return __pyx_enum.IBV_QPT_UC
 *     elif c_val == ibv_qp_type.IBV_QPT_UD:
 *         return __pyx_enum.IBV_QPT_UD             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_type.IBV_QPT_RAW_PACKET:
 *         return __pyx_enum.IBV_QPT_RAW_PACKET
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QPT_UD); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":33
 *     elif c_val == ibv_qp_type.IBV_QPT_UC:
 *         return __pyx_enum.IBV_QPT_UC
 *     elif c_val == ibv_qp_type.IBV_QPT_UD:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPT_UD
 *     elif c_val == ibv_qp_type.IBV_QPT_RAW_PACKET:
*/
  }

  /* "EnumTypeToPy":35
 *     elif c_val == ibv_qp_type.IBV_QPT_UD:
 *         return __pyx_enum.IBV_QPT_UD
 *     elif c_val == ibv_qp_type.IBV_QPT_RAW_PACKET:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPT_RAW_PACKET
 *     elif c_val == ibv_qp_type.IBV_QPT_XRC_SEND:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QPT_RAW_PACKET);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":36
 *         return __pyx_enum.IBV_QPT_UD
 *     elif c_val == ibv_qp_type.IBV_QPT_RAW_PACKET:
 *         return __pyx_enum.IBV_QPT_RAW_PACKET             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_type.IBV_QPT_XRC_SEND:
 *         return __pyx_enum.IBV_QPT_XRC_SEND
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QPT_RAW_PACKET); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":35
 *     elif c_val == ibv_qp_type.IBV_QPT_UD:
 *         return __pyx_enum.IBV_QPT_UD
 *     elif c_val == ibv_qp_type.IBV_QPT_RAW_PACKET:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPT_RAW_PACKET
 *     elif c_val == ibv_qp_type.IBV_QPT_XRC_SEND:
*/
  }

  /* "EnumTypeToPy":37
 *     elif c_val == ibv_qp_type.IBV_QPT_RAW_PACKET:
 *         return __pyx_enum.IBV_QPT_RAW_PACKET
 *     elif c_val == ibv_qp_type.IBV_QPT_XRC_SEND:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPT_XRC_SEND
 *     elif c_val == ibv_qp_type.IBV_QPT_XRC_RECV:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QPT_XRC_SEND);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":38
 *         return __pyx_enum.IBV_QPT_RAW_PACKET
 *     elif c_val == ibv_qp_type.IBV_QPT_XRC_SEND:
 *         return __pyx_enum.IBV_QPT_XRC_SEND             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_type.IBV_QPT_XRC_RECV:
 *         return __pyx_enum.IBV_QPT_XRC_RECV
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QPT_XRC_SEND); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":37
 *     elif c_val == ibv_qp_type.IBV_QPT_RAW_PACKET:
 *         return __pyx_enum.IBV_QPT_RAW_PACKET
 *     elif c_val == ibv_qp_type.IBV_QPT_XRC_SEND:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPT_XRC_SEND
 *     elif c_val == ibv_qp_type.IBV_QPT_XRC_RECV:
*/
  }

  /* "EnumTypeToPy":39
 *     elif c_val == ibv_qp_type.IBV_QPT_XRC_SEND:
 *         return __pyx_enum.IBV_QPT_XRC_SEND
 *     elif c_val == ibv_qp_type.IBV_QPT_XRC_RECV:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPT_XRC_RECV
 *     elif c_val == ibv_qp_type.IBV_QPT_DRIVER:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QPT_XRC_RECV);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":40
 *         return __pyx_enum.IBV_QPT_XRC_SEND
 *     elif c_val == ibv_qp_type.IBV_QPT_XRC_RECV:
 *         return __pyx_enum.IBV_QPT_XRC_RECV             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_type.IBV_QPT_DRIVER:
 *         return __pyx_enum.IBV_QPT_DRIVER
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QPT_XRC_RECV); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":39
 *     elif c_val == ibv_qp_type.IBV_QPT_XRC_SEND:
 *         return __pyx_enum.IBV_QPT_XRC_SEND
 *     elif c_val == ibv_qp_type.IBV_QPT_XRC_RECV:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPT_XRC_RECV
 *     elif c_val == ibv_qp_type.IBV_QPT_DRIVER:
*/
  }

  /* "EnumTypeToPy":41
 *     elif c_val == ibv_qp_type.IBV_QPT_XRC_RECV:
 *         return __pyx_enum.IBV_QPT_XRC_RECV
 *     elif c_val == ibv_qp_type.IBV_QPT_DRIVER:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPT_DRIVER
 *     else:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QPT_DRIVER);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":42
 *         return __pyx_enum.IBV_QPT_XRC_RECV
 *     elif c_val == ibv_qp_type.IBV_QPT_DRIVER:
 *         return __pyx_enum.IBV_QPT_DRIVER             # <<<<<<<<<<<<<<
 *     else:
 *         underlying_c_val = <int>c_val
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QPT_DRIVER); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":41
 *     elif c_val == ibv_qp_type.IBV_QPT_XRC_RECV:
 *         return __pyx_enum.IBV_QPT_XRC_RECV
 *     elif c_val == ibv_qp_type.IBV_QPT_DRIVER:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPT_DRIVER
 *     else:
*/
  }

  /* "EnumTypeToPy":44
 *         return __pyx_enum.IBV_QPT_DRIVER
 *     else:
 *         underlying_c_val = <int>c_val             # <<<<<<<<<<<<<<
 *         return __pyx_enum(underlying_c_val)
*/
  /*else*/ {
    __pyx_v_underlying_c_val = ((int)__pyx_v_c_val);

    /* "EnumTypeToPy":45
 *     else:
 *         underlying_c_val = <int>c_val
 *         return __pyx_enum(underlying_c_val)             # <<<<<<<<<<<<<<
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = NULL;
    __Pyx_INCREF(__pyx_v___pyx_enum);
    __pyx_t_5 = __pyx_v___pyx_enum; 
    __pyx_t_11 = __Pyx_PyLong_From_int(__pyx_v_underlying_c_val); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_13 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_8);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_13 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_11};
      __pyx_t_10 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 45, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
    }
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;
  }

  /* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(ibv_qp_type c_val):
 *     cdef object __pyx_enum
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_enum);
  __Pyx_XDECREF(__pyx_v_warnings);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py(enum ibv_qp_state __pyx_v_c_val) {
  PyObject *__pyx_v___pyx_enum = 0;
  PyObject *__pyx_v_warnings = NULL;
  int __pyx_v_underlying_c_val;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  size_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py", 0);

  /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_qp_state as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py == Py_None);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_qp_state as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py
 *         except ImportError:
*/
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "EnumTypeToPy":13
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py is None:
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_qp_state as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py             # <<<<<<<<<<<<<<
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py = False  # False indicates "don't try again"
*/
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_qp_state};
          __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyverbs_libibverbs_enums, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L4_error)
        }
        __pyx_t_5 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_5);
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_qp_state};
          __pyx_t_7 = 0; {
            __pyx_t_8 = __Pyx_ImportFrom(__pyx_t_5, __pyx_imported_names[__pyx_t_7]); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 13, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_8);
            switch (__pyx_t_7) {
              case 0:
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py);
              __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py, __pyx_t_8);
              __Pyx_GIVEREF(__pyx_t_8);
              break;
              default:;
            }
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_qp_state as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py
 *         except ImportError:
*/
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "EnumTypeToPy":14
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_qp_state as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py
 *         except ImportError:             # <<<<<<<<<<<<<<
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py = False  # False indicates "don't try again"
 *             import warnings
*/
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_ImportError))));
      if (__pyx_t_9) {
        __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_8, &__pyx_t_10) < 0) __PYX_ERR(1, 14, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_10);

        /* "EnumTypeToPy":15
 *             from pyverbs.libibverbs_enums import ibv_qp_state as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py = False  # False indicates "don't try again"             # <<<<<<<<<<<<<<
 *             import warnings
 *             warnings.warn(
*/
        __Pyx_INCREF(Py_False);
        __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py);
        __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py, Py_False);
        __Pyx_GIVEREF(Py_False);

        /* "EnumTypeToPy":16
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py = False  # False indicates "don't try again"
 *             import warnings             # <<<<<<<<<<<<<<
 *             warnings.warn(
 *                 f"enum class ibv_qp_state not importable from pyverbs.libibverbs_enums. "
*/
        __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_warnings, 0, 0, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 16, __pyx_L6_except_error)
        __pyx_t_11 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_v_warnings = __pyx_t_11;
        __pyx_t_11 = 0;

        /* "EnumTypeToPy":17
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py = False  # False indicates "don't try again"
 *             import warnings
 *             warnings.warn(             # <<<<<<<<<<<<<<
 *                 f"enum class ibv_qp_state not importable from pyverbs.libibverbs_enums. "
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
*/
        __pyx_t_12 = __pyx_v_warnings;
        __Pyx_INCREF(__pyx_t_12);
        __pyx_t_13 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_mstate_global->__pyx_kp_u_enum_class_ibv_qp_state_not_impo};
          __pyx_t_11 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_warn, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 17, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_11);
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L5_exception_handled;
      }
      goto __pyx_L6_except_error;

      /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_qp_state as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py
 *         except ImportError:
*/
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L5_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_L9_try_end:;
    }

    /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_qp_state as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py
*/
  }

  /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py == Py_False);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":24
 * 
 * 
 *         return <int>c_val             # <<<<<<<<<<<<<<
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyLong_From_int(((int)__pyx_v_c_val)); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumTypeToPy":25
 * 
 *         return <int>c_val
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py             # <<<<<<<<<<<<<<
 * 
 *     if 0:
*/
  __Pyx_INCREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py);
  __pyx_v___pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py;

  /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_qp_state.IBV_QPS_RESET:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPS_RESET
 *     elif c_val == ibv_qp_state.IBV_QPS_INIT:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QPS_RESET);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":30
 *         pass
 *     elif c_val == ibv_qp_state.IBV_QPS_RESET:
 *         return __pyx_enum.IBV_QPS_RESET             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_state.IBV_QPS_INIT:
 *         return __pyx_enum.IBV_QPS_INIT
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QPS_RESET); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_qp_state.IBV_QPS_RESET:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPS_RESET
 *     elif c_val == ibv_qp_state.IBV_QPS_INIT:
*/
  }

  /* "EnumTypeToPy":31
 *     elif c_val == ibv_qp_state.IBV_QPS_RESET:
 *         return __pyx_enum.IBV_QPS_RESET
 *     elif c_val == ibv_qp_state.IBV_QPS_INIT:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPS_INIT
 *     elif c_val == ibv_qp_state.IBV_QPS_RTR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QPS_INIT);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":32
 *         return __pyx_enum.IBV_QPS_RESET
 *     elif c_val == ibv_qp_state.IBV_QPS_INIT:
 *         return __pyx_enum.IBV_QPS_INIT             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_state.IBV_QPS_RTR:
 *         return __pyx_enum.IBV_QPS_RTR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QPS_INIT); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 32, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":31
 *     elif c_val == ibv_qp_state.IBV_QPS_RESET:
 *         return __pyx_enum.IBV_QPS_RESET
 *     elif c_val == ibv_qp_state.IBV_QPS_INIT:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPS_INIT
 *     elif c_val == ibv_qp_state.IBV_QPS_RTR:
*/
  }

  /* "EnumTypeToPy":33
 *     elif c_val == ibv_qp_state.IBV_QPS_INIT:
 *         return __pyx_enum.IBV_QPS_INIT
 *     elif c_val == ibv_qp_state.IBV_QPS_RTR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPS_RTR
 *     elif c_val == ibv_qp_state.IBV_QPS_RTS:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QPS_RTR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":34
 *         return __pyx_enum.IBV_QPS_INIT
 *     elif c_val == ibv_qp_state.IBV_QPS_RTR:
 *         return __pyx_enum.IBV_QPS_RTR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_state.IBV_QPS_RTS:
 *         return __pyx_enum.IBV_QPS_RTS
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QPS_RTR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":33
 *     elif c_val == ibv_qp_state.IBV_QPS_INIT:
 *         return __pyx_enum.IBV_QPS_INIT
 *     elif c_val == ibv_qp_state.IBV_QPS_RTR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPS_RTR
 *     elif c_val == ibv_qp_state.IBV_QPS_RTS:
*/
  }

  /* "EnumTypeToPy":35
 *     elif c_val == ibv_qp_state.IBV_QPS_RTR:
 *         return __pyx_enum.IBV_QPS_RTR
 *     elif c_val == ibv_qp_state.IBV_QPS_RTS:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPS_RTS
 *     elif c_val == ibv_qp_state.IBV_QPS_SQD:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QPS_RTS);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":36
 *         return __pyx_enum.IBV_QPS_RTR
 *     elif c_val == ibv_qp_state.IBV_QPS_RTS:
 *         return __pyx_enum.IBV_QPS_RTS             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_state.IBV_QPS_SQD:
 *         return __pyx_enum.IBV_QPS_SQD
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QPS_RTS); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":35
 *     elif c_val == ibv_qp_state.IBV_QPS_RTR:
 *         return __pyx_enum.IBV_QPS_RTR
 *     elif c_val == ibv_qp_state.IBV_QPS_RTS:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPS_RTS
 *     elif c_val == ibv_qp_state.IBV_QPS_SQD:
*/
  }

  /* "EnumTypeToPy":37
 *     elif c_val == ibv_qp_state.IBV_QPS_RTS:
 *         return __pyx_enum.IBV_QPS_RTS
 *     elif c_val == ibv_qp_state.IBV_QPS_SQD:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPS_SQD
 *     elif c_val == ibv_qp_state.IBV_QPS_SQE:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QPS_SQD);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":38
 *         return __pyx_enum.IBV_QPS_RTS
 *     elif c_val == ibv_qp_state.IBV_QPS_SQD:
 *         return __pyx_enum.IBV_QPS_SQD             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_state.IBV_QPS_SQE:
 *         return __pyx_enum.IBV_QPS_SQE
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QPS_SQD); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":37
 *     elif c_val == ibv_qp_state.IBV_QPS_RTS:
 *         return __pyx_enum.IBV_QPS_RTS
 *     elif c_val == ibv_qp_state.IBV_QPS_SQD:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPS_SQD
 *     elif c_val == ibv_qp_state.IBV_QPS_SQE:
*/
  }

  /* "EnumTypeToPy":39
 *     elif c_val == ibv_qp_state.IBV_QPS_SQD:
 *         return __pyx_enum.IBV_QPS_SQD
 *     elif c_val == ibv_qp_state.IBV_QPS_SQE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPS_SQE
 *     elif c_val == ibv_qp_state.IBV_QPS_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QPS_SQE);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":40
 *         return __pyx_enum.IBV_QPS_SQD
 *     elif c_val == ibv_qp_state.IBV_QPS_SQE:
 *         return __pyx_enum.IBV_QPS_SQE             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_state.IBV_QPS_ERR:
 *         return __pyx_enum.IBV_QPS_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QPS_SQE); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":39
 *     elif c_val == ibv_qp_state.IBV_QPS_SQD:
 *         return __pyx_enum.IBV_QPS_SQD
 *     elif c_val == ibv_qp_state.IBV_QPS_SQE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPS_SQE
 *     elif c_val == ibv_qp_state.IBV_QPS_ERR:
*/
  }

  /* "EnumTypeToPy":41
 *     elif c_val == ibv_qp_state.IBV_QPS_SQE:
 *         return __pyx_enum.IBV_QPS_SQE
 *     elif c_val == ibv_qp_state.IBV_QPS_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPS_ERR
 *     elif c_val == ibv_qp_state.IBV_QPS_UNKNOWN:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QPS_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":42
 *         return __pyx_enum.IBV_QPS_SQE
 *     elif c_val == ibv_qp_state.IBV_QPS_ERR:
 *         return __pyx_enum.IBV_QPS_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_state.IBV_QPS_UNKNOWN:
 *         return __pyx_enum.IBV_QPS_UNKNOWN
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QPS_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":41
 *     elif c_val == ibv_qp_state.IBV_QPS_SQE:
 *         return __pyx_enum.IBV_QPS_SQE
 *     elif c_val == ibv_qp_state.IBV_QPS_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPS_ERR
 *     elif c_val == ibv_qp_state.IBV_QPS_UNKNOWN:
*/
  }

  /* "EnumTypeToPy":43
 *     elif c_val == ibv_qp_state.IBV_QPS_ERR:
 *         return __pyx_enum.IBV_QPS_ERR
 *     elif c_val == ibv_qp_state.IBV_QPS_UNKNOWN:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPS_UNKNOWN
 *     else:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QPS_UNKNOWN);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":44
 *         return __pyx_enum.IBV_QPS_ERR
 *     elif c_val == ibv_qp_state.IBV_QPS_UNKNOWN:
 *         return __pyx_enum.IBV_QPS_UNKNOWN             # <<<<<<<<<<<<<<
 *     else:
 *         underlying_c_val = <int>c_val
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QPS_UNKNOWN); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":43
 *     elif c_val == ibv_qp_state.IBV_QPS_ERR:
 *         return __pyx_enum.IBV_QPS_ERR
 *     elif c_val == ibv_qp_state.IBV_QPS_UNKNOWN:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QPS_UNKNOWN
 *     else:
*/
  }

  /* "EnumTypeToPy":46
 *         return __pyx_enum.IBV_QPS_UNKNOWN
 *     else:
 *         underlying_c_val = <int>c_val             # <<<<<<<<<<<<<<
 *         return __pyx_enum(underlying_c_val)
*/
  /*else*/ {
    __pyx_v_underlying_c_val = ((int)__pyx_v_c_val);

    /* "EnumTypeToPy":47
 *     else:
 *         underlying_c_val = <int>c_val
 *         return __pyx_enum(underlying_c_val)             # <<<<<<<<<<<<<<
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = NULL;
    __Pyx_INCREF(__pyx_v___pyx_enum);
    __pyx_t_5 = __pyx_v___pyx_enum; 
    __pyx_t_11 = __Pyx_PyLong_From_int(__pyx_v_underlying_c_val); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_13 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_8);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_13 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_11};
      __pyx_t_10 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 47, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
    }
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;
  }

  /* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py(ibv_qp_state c_val):
 *     cdef object __pyx_enum
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_enum);
  __Pyx_XDECREF(__pyx_v_warnings);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py(enum ibv_mtu __pyx_v_c_val) {
  PyObject *__pyx_v___pyx_enum = 0;
  PyObject *__pyx_v_warnings = NULL;
  int __pyx_v_underlying_c_val;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  size_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py", 0);

  /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_mtu as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py == Py_None);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_mtu as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py
 *         except ImportError:
*/
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "EnumTypeToPy":13
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py is None:
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_mtu as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py             # <<<<<<<<<<<<<<
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py = False  # False indicates "don't try again"
*/
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_mtu};
          __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyverbs_libibverbs_enums, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L4_error)
        }
        __pyx_t_5 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_5);
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_mtu};
          __pyx_t_7 = 0; {
            __pyx_t_8 = __Pyx_ImportFrom(__pyx_t_5, __pyx_imported_names[__pyx_t_7]); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 13, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_8);
            switch (__pyx_t_7) {
              case 0:
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py);
              __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py, __pyx_t_8);
              __Pyx_GIVEREF(__pyx_t_8);
              break;
              default:;
            }
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_mtu as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py
 *         except ImportError:
*/
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "EnumTypeToPy":14
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_mtu as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py
 *         except ImportError:             # <<<<<<<<<<<<<<
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py = False  # False indicates "don't try again"
 *             import warnings
*/
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_ImportError))));
      if (__pyx_t_9) {
        __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_8, &__pyx_t_10) < 0) __PYX_ERR(1, 14, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_10);

        /* "EnumTypeToPy":15
 *             from pyverbs.libibverbs_enums import ibv_mtu as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py = False  # False indicates "don't try again"             # <<<<<<<<<<<<<<
 *             import warnings
 *             warnings.warn(
*/
        __Pyx_INCREF(Py_False);
        __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py);
        __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py, Py_False);
        __Pyx_GIVEREF(Py_False);

        /* "EnumTypeToPy":16
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py = False  # False indicates "don't try again"
 *             import warnings             # <<<<<<<<<<<<<<
 *             warnings.warn(
 *                 f"enum class ibv_mtu not importable from pyverbs.libibverbs_enums. "
*/
        __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_warnings, 0, 0, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 16, __pyx_L6_except_error)
        __pyx_t_11 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_v_warnings = __pyx_t_11;
        __pyx_t_11 = 0;

        /* "EnumTypeToPy":17
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py = False  # False indicates "don't try again"
 *             import warnings
 *             warnings.warn(             # <<<<<<<<<<<<<<
 *                 f"enum class ibv_mtu not importable from pyverbs.libibverbs_enums. "
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
*/
        __pyx_t_12 = __pyx_v_warnings;
        __Pyx_INCREF(__pyx_t_12);
        __pyx_t_13 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_mstate_global->__pyx_kp_u_enum_class_ibv_mtu_not_importabl};
          __pyx_t_11 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_warn, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 17, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_11);
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L5_exception_handled;
      }
      goto __pyx_L6_except_error;

      /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_mtu as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py
 *         except ImportError:
*/
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L5_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_L9_try_end:;
    }

    /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_mtu as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py
*/
  }

  /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py == Py_False);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":24
 * 
 * 
 *         return <int>c_val             # <<<<<<<<<<<<<<
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyLong_From_int(((int)__pyx_v_c_val)); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumTypeToPy":25
 * 
 *         return <int>c_val
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py             # <<<<<<<<<<<<<<
 * 
 *     if 0:
*/
  __Pyx_INCREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py);
  __pyx_v___pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py;

  /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_mtu.IBV_MTU_256:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_MTU_256
 *     elif c_val == ibv_mtu.IBV_MTU_512:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_MTU_256);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":30
 *         pass
 *     elif c_val == ibv_mtu.IBV_MTU_256:
 *         return __pyx_enum.IBV_MTU_256             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_mtu.IBV_MTU_512:
 *         return __pyx_enum.IBV_MTU_512
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_MTU_256); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_mtu.IBV_MTU_256:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_MTU_256
 *     elif c_val == ibv_mtu.IBV_MTU_512:
*/
  }

  /* "EnumTypeToPy":31
 *     elif c_val == ibv_mtu.IBV_MTU_256:
 *         return __pyx_enum.IBV_MTU_256
 *     elif c_val == ibv_mtu.IBV_MTU_512:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_MTU_512
 *     elif c_val == ibv_mtu.IBV_MTU_1024:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_MTU_512);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":32
 *         return __pyx_enum.IBV_MTU_256
 *     elif c_val == ibv_mtu.IBV_MTU_512:
 *         return __pyx_enum.IBV_MTU_512             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_mtu.IBV_MTU_1024:
 *         return __pyx_enum.IBV_MTU_1024
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_MTU_512); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 32, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":31
 *     elif c_val == ibv_mtu.IBV_MTU_256:
 *         return __pyx_enum.IBV_MTU_256
 *     elif c_val == ibv_mtu.IBV_MTU_512:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_MTU_512
 *     elif c_val == ibv_mtu.IBV_MTU_1024:
*/
  }

  /* "EnumTypeToPy":33
 *     elif c_val == ibv_mtu.IBV_MTU_512:
 *         return __pyx_enum.IBV_MTU_512
 *     elif c_val == ibv_mtu.IBV_MTU_1024:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_MTU_1024
 *     elif c_val == ibv_mtu.IBV_MTU_2048:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_MTU_1024);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":34
 *         return __pyx_enum.IBV_MTU_512
 *     elif c_val == ibv_mtu.IBV_MTU_1024:
 *         return __pyx_enum.IBV_MTU_1024             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_mtu.IBV_MTU_2048:
 *         return __pyx_enum.IBV_MTU_2048
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_MTU_1024); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":33
 *     elif c_val == ibv_mtu.IBV_MTU_512:
 *         return __pyx_enum.IBV_MTU_512
 *     elif c_val == ibv_mtu.IBV_MTU_1024:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_MTU_1024
 *     elif c_val == ibv_mtu.IBV_MTU_2048:
*/
  }

  /* "EnumTypeToPy":35
 *     elif c_val == ibv_mtu.IBV_MTU_1024:
 *         return __pyx_enum.IBV_MTU_1024
 *     elif c_val == ibv_mtu.IBV_MTU_2048:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_MTU_2048
 *     elif c_val == ibv_mtu.IBV_MTU_4096:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_MTU_2048);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":36
 *         return __pyx_enum.IBV_MTU_1024
 *     elif c_val == ibv_mtu.IBV_MTU_2048:
 *         return __pyx_enum.IBV_MTU_2048             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_mtu.IBV_MTU_4096:
 *         return __pyx_enum.IBV_MTU_4096
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_MTU_2048); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":35
 *     elif c_val == ibv_mtu.IBV_MTU_1024:
 *         return __pyx_enum.IBV_MTU_1024
 *     elif c_val == ibv_mtu.IBV_MTU_2048:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_MTU_2048
 *     elif c_val == ibv_mtu.IBV_MTU_4096:
*/
  }

  /* "EnumTypeToPy":37
 *     elif c_val == ibv_mtu.IBV_MTU_2048:
 *         return __pyx_enum.IBV_MTU_2048
 *     elif c_val == ibv_mtu.IBV_MTU_4096:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_MTU_4096
 *     else:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_MTU_4096);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":38
 *         return __pyx_enum.IBV_MTU_2048
 *     elif c_val == ibv_mtu.IBV_MTU_4096:
 *         return __pyx_enum.IBV_MTU_4096             # <<<<<<<<<<<<<<
 *     else:
 *         underlying_c_val = <int>c_val
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_MTU_4096); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":37
 *     elif c_val == ibv_mtu.IBV_MTU_2048:
 *         return __pyx_enum.IBV_MTU_2048
 *     elif c_val == ibv_mtu.IBV_MTU_4096:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_MTU_4096
 *     else:
*/
  }

  /* "EnumTypeToPy":40
 *         return __pyx_enum.IBV_MTU_4096
 *     else:
 *         underlying_c_val = <int>c_val             # <<<<<<<<<<<<<<
 *         return __pyx_enum(underlying_c_val)
*/
  /*else*/ {
    __pyx_v_underlying_c_val = ((int)__pyx_v_c_val);

    /* "EnumTypeToPy":41
 *     else:
 *         underlying_c_val = <int>c_val
 *         return __pyx_enum(underlying_c_val)             # <<<<<<<<<<<<<<
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = NULL;
    __Pyx_INCREF(__pyx_v___pyx_enum);
    __pyx_t_5 = __pyx_v___pyx_enum; 
    __pyx_t_11 = __Pyx_PyLong_From_int(__pyx_v_underlying_c_val); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_13 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_8);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_13 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_11};
      __pyx_t_10 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 41, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
    }
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;
  }

  /* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py(ibv_mtu c_val):
 *     cdef object __pyx_enum
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_enum);
  __Pyx_XDECREF(__pyx_v_warnings);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py(enum ibv_mig_state __pyx_v_c_val) {
  PyObject *__pyx_v___pyx_enum = 0;
  PyObject *__pyx_v_warnings = NULL;
  int __pyx_v_underlying_c_val;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  size_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py", 0);

  /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_mig_state as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py == Py_None);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_mig_state as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py
 *         except ImportError:
*/
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "EnumTypeToPy":13
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py is None:
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_mig_state as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py             # <<<<<<<<<<<<<<
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py = False  # False indicates "don't try again"
*/
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_mig_state};
          __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyverbs_libibverbs_enums, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L4_error)
        }
        __pyx_t_5 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_5);
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_mig_state};
          __pyx_t_7 = 0; {
            __pyx_t_8 = __Pyx_ImportFrom(__pyx_t_5, __pyx_imported_names[__pyx_t_7]); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 13, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_8);
            switch (__pyx_t_7) {
              case 0:
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py);
              __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py, __pyx_t_8);
              __Pyx_GIVEREF(__pyx_t_8);
              break;
              default:;
            }
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_mig_state as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py
 *         except ImportError:
*/
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "EnumTypeToPy":14
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_mig_state as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py
 *         except ImportError:             # <<<<<<<<<<<<<<
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py = False  # False indicates "don't try again"
 *             import warnings
*/
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_ImportError))));
      if (__pyx_t_9) {
        __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_8, &__pyx_t_10) < 0) __PYX_ERR(1, 14, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_10);

        /* "EnumTypeToPy":15
 *             from pyverbs.libibverbs_enums import ibv_mig_state as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py = False  # False indicates "don't try again"             # <<<<<<<<<<<<<<
 *             import warnings
 *             warnings.warn(
*/
        __Pyx_INCREF(Py_False);
        __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py);
        __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py, Py_False);
        __Pyx_GIVEREF(Py_False);

        /* "EnumTypeToPy":16
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py = False  # False indicates "don't try again"
 *             import warnings             # <<<<<<<<<<<<<<
 *             warnings.warn(
 *                 f"enum class ibv_mig_state not importable from pyverbs.libibverbs_enums. "
*/
        __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_warnings, 0, 0, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 16, __pyx_L6_except_error)
        __pyx_t_11 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_v_warnings = __pyx_t_11;
        __pyx_t_11 = 0;

        /* "EnumTypeToPy":17
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py = False  # False indicates "don't try again"
 *             import warnings
 *             warnings.warn(             # <<<<<<<<<<<<<<
 *                 f"enum class ibv_mig_state not importable from pyverbs.libibverbs_enums. "
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
*/
        __pyx_t_12 = __pyx_v_warnings;
        __Pyx_INCREF(__pyx_t_12);
        __pyx_t_13 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_mstate_global->__pyx_kp_u_enum_class_ibv_mig_state_not_imp};
          __pyx_t_11 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_warn, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 17, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_11);
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L5_exception_handled;
      }
      goto __pyx_L6_except_error;

      /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_mig_state as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py
 *         except ImportError:
*/
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L5_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_L9_try_end:;
    }

    /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_mig_state as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py
*/
  }

  /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py == Py_False);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":24
 * 
 * 
 *         return <int>c_val             # <<<<<<<<<<<<<<
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyLong_From_int(((int)__pyx_v_c_val)); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumTypeToPy":25
 * 
 *         return <int>c_val
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py             # <<<<<<<<<<<<<<
 * 
 *     if 0:
*/
  __Pyx_INCREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py);
  __pyx_v___pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py;

  /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_mig_state.IBV_MIG_MIGRATED:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_MIG_MIGRATED
 *     elif c_val == ibv_mig_state.IBV_MIG_REARM:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_MIG_MIGRATED);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":30
 *         pass
 *     elif c_val == ibv_mig_state.IBV_MIG_MIGRATED:
 *         return __pyx_enum.IBV_MIG_MIGRATED             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_mig_state.IBV_MIG_REARM:
 *         return __pyx_enum.IBV_MIG_REARM
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_MIG_MIGRATED); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_mig_state.IBV_MIG_MIGRATED:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_MIG_MIGRATED
 *     elif c_val == ibv_mig_state.IBV_MIG_REARM:
*/
  }

  /* "EnumTypeToPy":31
 *     elif c_val == ibv_mig_state.IBV_MIG_MIGRATED:
 *         return __pyx_enum.IBV_MIG_MIGRATED
 *     elif c_val == ibv_mig_state.IBV_MIG_REARM:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_MIG_REARM
 *     elif c_val == ibv_mig_state.IBV_MIG_ARMED:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_MIG_REARM);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":32
 *         return __pyx_enum.IBV_MIG_MIGRATED
 *     elif c_val == ibv_mig_state.IBV_MIG_REARM:
 *         return __pyx_enum.IBV_MIG_REARM             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_mig_state.IBV_MIG_ARMED:
 *         return __pyx_enum.IBV_MIG_ARMED
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_MIG_REARM); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 32, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":31
 *     elif c_val == ibv_mig_state.IBV_MIG_MIGRATED:
 *         return __pyx_enum.IBV_MIG_MIGRATED
 *     elif c_val == ibv_mig_state.IBV_MIG_REARM:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_MIG_REARM
 *     elif c_val == ibv_mig_state.IBV_MIG_ARMED:
*/
  }

  /* "EnumTypeToPy":33
 *     elif c_val == ibv_mig_state.IBV_MIG_REARM:
 *         return __pyx_enum.IBV_MIG_REARM
 *     elif c_val == ibv_mig_state.IBV_MIG_ARMED:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_MIG_ARMED
 *     else:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_MIG_ARMED);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":34
 *         return __pyx_enum.IBV_MIG_REARM
 *     elif c_val == ibv_mig_state.IBV_MIG_ARMED:
 *         return __pyx_enum.IBV_MIG_ARMED             # <<<<<<<<<<<<<<
 *     else:
 *         underlying_c_val = <int>c_val
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_MIG_ARMED); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":33
 *     elif c_val == ibv_mig_state.IBV_MIG_REARM:
 *         return __pyx_enum.IBV_MIG_REARM
 *     elif c_val == ibv_mig_state.IBV_MIG_ARMED:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_MIG_ARMED
 *     else:
*/
  }

  /* "EnumTypeToPy":36
 *         return __pyx_enum.IBV_MIG_ARMED
 *     else:
 *         underlying_c_val = <int>c_val             # <<<<<<<<<<<<<<
 *         return __pyx_enum(underlying_c_val)
*/
  /*else*/ {
    __pyx_v_underlying_c_val = ((int)__pyx_v_c_val);

    /* "EnumTypeToPy":37
 *     else:
 *         underlying_c_val = <int>c_val
 *         return __pyx_enum(underlying_c_val)             # <<<<<<<<<<<<<<
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = NULL;
    __Pyx_INCREF(__pyx_v___pyx_enum);
    __pyx_t_5 = __pyx_v___pyx_enum; 
    __pyx_t_11 = __Pyx_PyLong_From_int(__pyx_v_underlying_c_val); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_13 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_8);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_13 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_11};
      __pyx_t_10 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 37, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
    }
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;
  }

  /* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py(ibv_mig_state c_val):
 *     cdef object __pyx_enum
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_enum);
  __Pyx_XDECREF(__pyx_v_warnings);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py(enum ibv_qp_attr_mask __pyx_v_c_val) {
  PyObject *__pyx_v___pyx_enum = 0;
  PyObject *__pyx_v_warnings = NULL;
  int __pyx_v_underlying_c_val;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  size_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py", 0);

  /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_qp_attr_mask as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py == Py_None);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_qp_attr_mask as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py
 *         except ImportError:
*/
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "EnumTypeToPy":13
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py is None:
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_qp_attr_mask as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py             # <<<<<<<<<<<<<<
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py = False  # False indicates "don't try again"
*/
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_qp_attr_mask};
          __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyverbs_libibverbs_enums, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L4_error)
        }
        __pyx_t_5 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_5);
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_qp_attr_mask};
          __pyx_t_7 = 0; {
            __pyx_t_8 = __Pyx_ImportFrom(__pyx_t_5, __pyx_imported_names[__pyx_t_7]); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 13, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_8);
            switch (__pyx_t_7) {
              case 0:
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py);
              __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py, __pyx_t_8);
              __Pyx_GIVEREF(__pyx_t_8);
              break;
              default:;
            }
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_qp_attr_mask as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py
 *         except ImportError:
*/
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "EnumTypeToPy":14
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_qp_attr_mask as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py
 *         except ImportError:             # <<<<<<<<<<<<<<
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py = False  # False indicates "don't try again"
 *             import warnings
*/
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_ImportError))));
      if (__pyx_t_9) {
        __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_8, &__pyx_t_10) < 0) __PYX_ERR(1, 14, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_10);

        /* "EnumTypeToPy":15
 *             from pyverbs.libibverbs_enums import ibv_qp_attr_mask as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py = False  # False indicates "don't try again"             # <<<<<<<<<<<<<<
 *             import warnings
 *             warnings.warn(
*/
        __Pyx_INCREF(Py_False);
        __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py);
        __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py, Py_False);
        __Pyx_GIVEREF(Py_False);

        /* "EnumTypeToPy":16
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py = False  # False indicates "don't try again"
 *             import warnings             # <<<<<<<<<<<<<<
 *             warnings.warn(
 *                 f"enum class ibv_qp_attr_mask not importable from pyverbs.libibverbs_enums. "
*/
        __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_warnings, 0, 0, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 16, __pyx_L6_except_error)
        __pyx_t_11 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_v_warnings = __pyx_t_11;
        __pyx_t_11 = 0;

        /* "EnumTypeToPy":17
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py = False  # False indicates "don't try again"
 *             import warnings
 *             warnings.warn(             # <<<<<<<<<<<<<<
 *                 f"enum class ibv_qp_attr_mask not importable from pyverbs.libibverbs_enums. "
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
*/
        __pyx_t_12 = __pyx_v_warnings;
        __Pyx_INCREF(__pyx_t_12);
        __pyx_t_13 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_mstate_global->__pyx_kp_u_enum_class_ibv_qp_attr_mask_not};
          __pyx_t_11 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_warn, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 17, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_11);
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L5_exception_handled;
      }
      goto __pyx_L6_except_error;

      /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_qp_attr_mask as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py
 *         except ImportError:
*/
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L5_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_L9_try_end:;
    }

    /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_qp_attr_mask as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py
*/
  }

  /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py == Py_False);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":24
 * 
 * 
 *         return <int>c_val             # <<<<<<<<<<<<<<
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyLong_From_int(((int)__pyx_v_c_val)); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumTypeToPy":25
 * 
 *         return <int>c_val
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py             # <<<<<<<<<<<<<<
 * 
 *     if 0:
*/
  __Pyx_INCREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py);
  __pyx_v___pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py;

  /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_STATE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_STATE
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_CUR_STATE:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_STATE);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":30
 *         pass
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_STATE:
 *         return __pyx_enum.IBV_QP_STATE             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_CUR_STATE:
 *         return __pyx_enum.IBV_QP_CUR_STATE
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_STATE); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_STATE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_STATE
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_CUR_STATE:
*/
  }

  /* "EnumTypeToPy":31
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_STATE:
 *         return __pyx_enum.IBV_QP_STATE
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_CUR_STATE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_CUR_STATE
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_EN_SQD_ASYNC_NOTIFY:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_CUR_STATE);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":32
 *         return __pyx_enum.IBV_QP_STATE
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_CUR_STATE:
 *         return __pyx_enum.IBV_QP_CUR_STATE             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_EN_SQD_ASYNC_NOTIFY:
 *         return __pyx_enum.IBV_QP_EN_SQD_ASYNC_NOTIFY
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_CUR_STATE); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 32, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":31
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_STATE:
 *         return __pyx_enum.IBV_QP_STATE
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_CUR_STATE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_CUR_STATE
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_EN_SQD_ASYNC_NOTIFY:
*/
  }

  /* "EnumTypeToPy":33
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_CUR_STATE:
 *         return __pyx_enum.IBV_QP_CUR_STATE
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_EN_SQD_ASYNC_NOTIFY:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_EN_SQD_ASYNC_NOTIFY
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_ACCESS_FLAGS:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_EN_SQD_ASYNC_NOTIFY);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":34
 *         return __pyx_enum.IBV_QP_CUR_STATE
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_EN_SQD_ASYNC_NOTIFY:
 *         return __pyx_enum.IBV_QP_EN_SQD_ASYNC_NOTIFY             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_ACCESS_FLAGS:
 *         return __pyx_enum.IBV_QP_ACCESS_FLAGS
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_EN_SQD_ASYNC_NOTIFY); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":33
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_CUR_STATE:
 *         return __pyx_enum.IBV_QP_CUR_STATE
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_EN_SQD_ASYNC_NOTIFY:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_EN_SQD_ASYNC_NOTIFY
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_ACCESS_FLAGS:
*/
  }

  /* "EnumTypeToPy":35
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_EN_SQD_ASYNC_NOTIFY:
 *         return __pyx_enum.IBV_QP_EN_SQD_ASYNC_NOTIFY
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_ACCESS_FLAGS:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_ACCESS_FLAGS
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PKEY_INDEX:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_ACCESS_FLAGS);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":36
 *         return __pyx_enum.IBV_QP_EN_SQD_ASYNC_NOTIFY
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_ACCESS_FLAGS:
 *         return __pyx_enum.IBV_QP_ACCESS_FLAGS             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PKEY_INDEX:
 *         return __pyx_enum.IBV_QP_PKEY_INDEX
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_ACCESS_FLAGS); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":35
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_EN_SQD_ASYNC_NOTIFY:
 *         return __pyx_enum.IBV_QP_EN_SQD_ASYNC_NOTIFY
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_ACCESS_FLAGS:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_ACCESS_FLAGS
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PKEY_INDEX:
*/
  }

  /* "EnumTypeToPy":37
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_ACCESS_FLAGS:
 *         return __pyx_enum.IBV_QP_ACCESS_FLAGS
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PKEY_INDEX:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_PKEY_INDEX
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PORT:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_PKEY_INDEX);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":38
 *         return __pyx_enum.IBV_QP_ACCESS_FLAGS
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PKEY_INDEX:
 *         return __pyx_enum.IBV_QP_PKEY_INDEX             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PORT:
 *         return __pyx_enum.IBV_QP_PORT
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_PKEY_INDEX); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":37
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_ACCESS_FLAGS:
 *         return __pyx_enum.IBV_QP_ACCESS_FLAGS
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PKEY_INDEX:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_PKEY_INDEX
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PORT:
*/
  }

  /* "EnumTypeToPy":39
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PKEY_INDEX:
 *         return __pyx_enum.IBV_QP_PKEY_INDEX
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PORT:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_PORT
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_QKEY:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_PORT);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":40
 *         return __pyx_enum.IBV_QP_PKEY_INDEX
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PORT:
 *         return __pyx_enum.IBV_QP_PORT             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_QKEY:
 *         return __pyx_enum.IBV_QP_QKEY
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_PORT); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":39
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PKEY_INDEX:
 *         return __pyx_enum.IBV_QP_PKEY_INDEX
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PORT:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_PORT
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_QKEY:
*/
  }

  /* "EnumTypeToPy":41
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PORT:
 *         return __pyx_enum.IBV_QP_PORT
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_QKEY:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_QKEY
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_AV:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_QKEY);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":42
 *         return __pyx_enum.IBV_QP_PORT
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_QKEY:
 *         return __pyx_enum.IBV_QP_QKEY             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_AV:
 *         return __pyx_enum.IBV_QP_AV
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_QKEY); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":41
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PORT:
 *         return __pyx_enum.IBV_QP_PORT
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_QKEY:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_QKEY
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_AV:
*/
  }

  /* "EnumTypeToPy":43
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_QKEY:
 *         return __pyx_enum.IBV_QP_QKEY
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_AV:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_AV
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PATH_MTU:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_AV);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":44
 *         return __pyx_enum.IBV_QP_QKEY
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_AV:
 *         return __pyx_enum.IBV_QP_AV             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PATH_MTU:
 *         return __pyx_enum.IBV_QP_PATH_MTU
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_AV); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":43
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_QKEY:
 *         return __pyx_enum.IBV_QP_QKEY
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_AV:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_AV
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PATH_MTU:
*/
  }

  /* "EnumTypeToPy":45
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_AV:
 *         return __pyx_enum.IBV_QP_AV
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PATH_MTU:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_PATH_MTU
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_TIMEOUT:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_PATH_MTU);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":46
 *         return __pyx_enum.IBV_QP_AV
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PATH_MTU:
 *         return __pyx_enum.IBV_QP_PATH_MTU             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_TIMEOUT:
 *         return __pyx_enum.IBV_QP_TIMEOUT
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_PATH_MTU); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 46, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":45
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_AV:
 *         return __pyx_enum.IBV_QP_AV
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PATH_MTU:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_PATH_MTU
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_TIMEOUT:
*/
  }

  /* "EnumTypeToPy":47
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PATH_MTU:
 *         return __pyx_enum.IBV_QP_PATH_MTU
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_TIMEOUT:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_TIMEOUT
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RETRY_CNT:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_TIMEOUT);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":48
 *         return __pyx_enum.IBV_QP_PATH_MTU
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_TIMEOUT:
 *         return __pyx_enum.IBV_QP_TIMEOUT             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RETRY_CNT:
 *         return __pyx_enum.IBV_QP_RETRY_CNT
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_TIMEOUT); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":47
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PATH_MTU:
 *         return __pyx_enum.IBV_QP_PATH_MTU
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_TIMEOUT:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_TIMEOUT
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RETRY_CNT:
*/
  }

  /* "EnumTypeToPy":49
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_TIMEOUT:
 *         return __pyx_enum.IBV_QP_TIMEOUT
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RETRY_CNT:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_RETRY_CNT
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RNR_RETRY:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_RETRY_CNT);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":50
 *         return __pyx_enum.IBV_QP_TIMEOUT
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RETRY_CNT:
 *         return __pyx_enum.IBV_QP_RETRY_CNT             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RNR_RETRY:
 *         return __pyx_enum.IBV_QP_RNR_RETRY
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_RETRY_CNT); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":49
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_TIMEOUT:
 *         return __pyx_enum.IBV_QP_TIMEOUT
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RETRY_CNT:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_RETRY_CNT
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RNR_RETRY:
*/
  }

  /* "EnumTypeToPy":51
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RETRY_CNT:
 *         return __pyx_enum.IBV_QP_RETRY_CNT
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RNR_RETRY:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_RNR_RETRY
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RQ_PSN:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_RNR_RETRY);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":52
 *         return __pyx_enum.IBV_QP_RETRY_CNT
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RNR_RETRY:
 *         return __pyx_enum.IBV_QP_RNR_RETRY             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RQ_PSN:
 *         return __pyx_enum.IBV_QP_RQ_PSN
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_RNR_RETRY); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":51
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RETRY_CNT:
 *         return __pyx_enum.IBV_QP_RETRY_CNT
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RNR_RETRY:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_RNR_RETRY
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RQ_PSN:
*/
  }

  /* "EnumTypeToPy":53
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RNR_RETRY:
 *         return __pyx_enum.IBV_QP_RNR_RETRY
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RQ_PSN:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_RQ_PSN
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MAX_QP_RD_ATOMIC:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_RQ_PSN);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":54
 *         return __pyx_enum.IBV_QP_RNR_RETRY
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RQ_PSN:
 *         return __pyx_enum.IBV_QP_RQ_PSN             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MAX_QP_RD_ATOMIC:
 *         return __pyx_enum.IBV_QP_MAX_QP_RD_ATOMIC
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_RQ_PSN); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":53
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RNR_RETRY:
 *         return __pyx_enum.IBV_QP_RNR_RETRY
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RQ_PSN:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_RQ_PSN
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MAX_QP_RD_ATOMIC:
*/
  }

  /* "EnumTypeToPy":55
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RQ_PSN:
 *         return __pyx_enum.IBV_QP_RQ_PSN
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MAX_QP_RD_ATOMIC:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_MAX_QP_RD_ATOMIC
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_ALT_PATH:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_MAX_QP_RD_ATOMIC);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":56
 *         return __pyx_enum.IBV_QP_RQ_PSN
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MAX_QP_RD_ATOMIC:
 *         return __pyx_enum.IBV_QP_MAX_QP_RD_ATOMIC             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_ALT_PATH:
 *         return __pyx_enum.IBV_QP_ALT_PATH
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_MAX_QP_RD_ATOMIC); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":55
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RQ_PSN:
 *         return __pyx_enum.IBV_QP_RQ_PSN
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MAX_QP_RD_ATOMIC:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_MAX_QP_RD_ATOMIC
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_ALT_PATH:
*/
  }

  /* "EnumTypeToPy":57
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MAX_QP_RD_ATOMIC:
 *         return __pyx_enum.IBV_QP_MAX_QP_RD_ATOMIC
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_ALT_PATH:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_ALT_PATH
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MIN_RNR_TIMER:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_ALT_PATH);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":58
 *         return __pyx_enum.IBV_QP_MAX_QP_RD_ATOMIC
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_ALT_PATH:
 *         return __pyx_enum.IBV_QP_ALT_PATH             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MIN_RNR_TIMER:
 *         return __pyx_enum.IBV_QP_MIN_RNR_TIMER
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_ALT_PATH); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":57
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MAX_QP_RD_ATOMIC:
 *         return __pyx_enum.IBV_QP_MAX_QP_RD_ATOMIC
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_ALT_PATH:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_ALT_PATH
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MIN_RNR_TIMER:
*/
  }

  /* "EnumTypeToPy":59
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_ALT_PATH:
 *         return __pyx_enum.IBV_QP_ALT_PATH
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MIN_RNR_TIMER:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_MIN_RNR_TIMER
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_SQ_PSN:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_MIN_RNR_TIMER);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":60
 *         return __pyx_enum.IBV_QP_ALT_PATH
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MIN_RNR_TIMER:
 *         return __pyx_enum.IBV_QP_MIN_RNR_TIMER             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_SQ_PSN:
 *         return __pyx_enum.IBV_QP_SQ_PSN
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_MIN_RNR_TIMER); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":59
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_ALT_PATH:
 *         return __pyx_enum.IBV_QP_ALT_PATH
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MIN_RNR_TIMER:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_MIN_RNR_TIMER
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_SQ_PSN:
*/
  }

  /* "EnumTypeToPy":61
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MIN_RNR_TIMER:
 *         return __pyx_enum.IBV_QP_MIN_RNR_TIMER
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_SQ_PSN:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_SQ_PSN
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MAX_DEST_RD_ATOMIC:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_SQ_PSN);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":62
 *         return __pyx_enum.IBV_QP_MIN_RNR_TIMER
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_SQ_PSN:
 *         return __pyx_enum.IBV_QP_SQ_PSN             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MAX_DEST_RD_ATOMIC:
 *         return __pyx_enum.IBV_QP_MAX_DEST_RD_ATOMIC
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_SQ_PSN); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":61
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MIN_RNR_TIMER:
 *         return __pyx_enum.IBV_QP_MIN_RNR_TIMER
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_SQ_PSN:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_SQ_PSN
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MAX_DEST_RD_ATOMIC:
*/
  }

  /* "EnumTypeToPy":63
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_SQ_PSN:
 *         return __pyx_enum.IBV_QP_SQ_PSN
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MAX_DEST_RD_ATOMIC:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_MAX_DEST_RD_ATOMIC
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PATH_MIG_STATE:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_MAX_DEST_RD_ATOMIC);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":64
 *         return __pyx_enum.IBV_QP_SQ_PSN
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MAX_DEST_RD_ATOMIC:
 *         return __pyx_enum.IBV_QP_MAX_DEST_RD_ATOMIC             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PATH_MIG_STATE:
 *         return __pyx_enum.IBV_QP_PATH_MIG_STATE
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_MAX_DEST_RD_ATOMIC); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":63
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_SQ_PSN:
 *         return __pyx_enum.IBV_QP_SQ_PSN
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MAX_DEST_RD_ATOMIC:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_MAX_DEST_RD_ATOMIC
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PATH_MIG_STATE:
*/
  }

  /* "EnumTypeToPy":65
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MAX_DEST_RD_ATOMIC:
 *         return __pyx_enum.IBV_QP_MAX_DEST_RD_ATOMIC
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PATH_MIG_STATE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_PATH_MIG_STATE
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_CAP:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_PATH_MIG_STATE);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":66
 *         return __pyx_enum.IBV_QP_MAX_DEST_RD_ATOMIC
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PATH_MIG_STATE:
 *         return __pyx_enum.IBV_QP_PATH_MIG_STATE             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_CAP:
 *         return __pyx_enum.IBV_QP_CAP
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_PATH_MIG_STATE); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":65
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_MAX_DEST_RD_ATOMIC:
 *         return __pyx_enum.IBV_QP_MAX_DEST_RD_ATOMIC
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PATH_MIG_STATE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_PATH_MIG_STATE
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_CAP:
*/
  }

  /* "EnumTypeToPy":67
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PATH_MIG_STATE:
 *         return __pyx_enum.IBV_QP_PATH_MIG_STATE
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_CAP:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_CAP
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_DEST_QPN:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_CAP);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":68
 *         return __pyx_enum.IBV_QP_PATH_MIG_STATE
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_CAP:
 *         return __pyx_enum.IBV_QP_CAP             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_DEST_QPN:
 *         return __pyx_enum.IBV_QP_DEST_QPN
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_CAP); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":67
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_PATH_MIG_STATE:
 *         return __pyx_enum.IBV_QP_PATH_MIG_STATE
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_CAP:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_CAP
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_DEST_QPN:
*/
  }

  /* "EnumTypeToPy":69
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_CAP:
 *         return __pyx_enum.IBV_QP_CAP
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_DEST_QPN:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_DEST_QPN
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RATE_LIMIT:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_DEST_QPN);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":70
 *         return __pyx_enum.IBV_QP_CAP
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_DEST_QPN:
 *         return __pyx_enum.IBV_QP_DEST_QPN             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RATE_LIMIT:
 *         return __pyx_enum.IBV_QP_RATE_LIMIT
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_DEST_QPN); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":69
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_CAP:
 *         return __pyx_enum.IBV_QP_CAP
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_DEST_QPN:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_DEST_QPN
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RATE_LIMIT:
*/
  }

  /* "EnumTypeToPy":71
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_DEST_QPN:
 *         return __pyx_enum.IBV_QP_DEST_QPN
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RATE_LIMIT:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_RATE_LIMIT
 *     else:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_RATE_LIMIT);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":72
 *         return __pyx_enum.IBV_QP_DEST_QPN
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RATE_LIMIT:
 *         return __pyx_enum.IBV_QP_RATE_LIMIT             # <<<<<<<<<<<<<<
 *     else:
 *         underlying_c_val = <int>c_val
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_RATE_LIMIT); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":71
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_DEST_QPN:
 *         return __pyx_enum.IBV_QP_DEST_QPN
 *     elif c_val == ibv_qp_attr_mask.IBV_QP_RATE_LIMIT:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_RATE_LIMIT
 *     else:
*/
  }

  /* "EnumTypeToPy":74
 *         return __pyx_enum.IBV_QP_RATE_LIMIT
 *     else:
 *         underlying_c_val = <int>c_val             # <<<<<<<<<<<<<<
 *         return __pyx_enum(underlying_c_val)
*/
  /*else*/ {
    __pyx_v_underlying_c_val = ((int)__pyx_v_c_val);

    /* "EnumTypeToPy":75
 *     else:
 *         underlying_c_val = <int>c_val
 *         return __pyx_enum(underlying_c_val)             # <<<<<<<<<<<<<<
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = NULL;
    __Pyx_INCREF(__pyx_v___pyx_enum);
    __pyx_t_5 = __pyx_v___pyx_enum; 
    __pyx_t_11 = __Pyx_PyLong_From_int(__pyx_v_underlying_c_val); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_13 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_8);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_13 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_11};
      __pyx_t_10 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 75, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
    }
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;
  }

  /* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py(ibv_qp_attr_mask c_val):
 *     cdef object __pyx_enum
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_enum);
  __Pyx_XDECREF(__pyx_v_warnings);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py(enum ibv_qp_init_attr_mask __pyx_v_c_val) {
  PyObject *__pyx_v___pyx_enum = 0;
  PyObject *__pyx_v_warnings = NULL;
  int __pyx_v_underlying_c_val;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  size_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py", 0);

  /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_qp_init_attr_mask as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py == Py_None);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_qp_init_attr_mask as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py
 *         except ImportError:
*/
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "EnumTypeToPy":13
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py is None:
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_qp_init_attr_mask as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py             # <<<<<<<<<<<<<<
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py = False  # False indicates "don't try again"
*/
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_qp_init_attr_mask};
          __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyverbs_libibverbs_enums, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L4_error)
        }
        __pyx_t_5 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_5);
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_qp_init_attr_mask};
          __pyx_t_7 = 0; {
            __pyx_t_8 = __Pyx_ImportFrom(__pyx_t_5, __pyx_imported_names[__pyx_t_7]); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 13, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_8);
            switch (__pyx_t_7) {
              case 0:
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py);
              __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py, __pyx_t_8);
              __Pyx_GIVEREF(__pyx_t_8);
              break;
              default:;
            }
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_qp_init_attr_mask as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py
 *         except ImportError:
*/
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "EnumTypeToPy":14
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_qp_init_attr_mask as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py
 *         except ImportError:             # <<<<<<<<<<<<<<
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py = False  # False indicates "don't try again"
 *             import warnings
*/
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_ImportError))));
      if (__pyx_t_9) {
        __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_8, &__pyx_t_10) < 0) __PYX_ERR(1, 14, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_10);

        /* "EnumTypeToPy":15
 *             from pyverbs.libibverbs_enums import ibv_qp_init_attr_mask as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py = False  # False indicates "don't try again"             # <<<<<<<<<<<<<<
 *             import warnings
 *             warnings.warn(
*/
        __Pyx_INCREF(Py_False);
        __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py);
        __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py, Py_False);
        __Pyx_GIVEREF(Py_False);

        /* "EnumTypeToPy":16
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py = False  # False indicates "don't try again"
 *             import warnings             # <<<<<<<<<<<<<<
 *             warnings.warn(
 *                 f"enum class ibv_qp_init_attr_mask not importable from pyverbs.libibverbs_enums. "
*/
        __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_warnings, 0, 0, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 16, __pyx_L6_except_error)
        __pyx_t_11 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_v_warnings = __pyx_t_11;
        __pyx_t_11 = 0;

        /* "EnumTypeToPy":17
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py = False  # False indicates "don't try again"
 *             import warnings
 *             warnings.warn(             # <<<<<<<<<<<<<<
 *                 f"enum class ibv_qp_init_attr_mask not importable from pyverbs.libibverbs_enums. "
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
*/
        __pyx_t_12 = __pyx_v_warnings;
        __Pyx_INCREF(__pyx_t_12);
        __pyx_t_13 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_mstate_global->__pyx_kp_u_enum_class_ibv_qp_init_attr_mask};
          __pyx_t_11 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_warn, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 17, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_11);
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L5_exception_handled;
      }
      goto __pyx_L6_except_error;

      /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_qp_init_attr_mask as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py
 *         except ImportError:
*/
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L5_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_L9_try_end:;
    }

    /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_qp_init_attr_mask as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py
*/
  }

  /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py == Py_False);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":24
 * 
 * 
 *         return <int>c_val             # <<<<<<<<<<<<<<
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyLong_From_int(((int)__pyx_v_c_val)); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumTypeToPy":25
 * 
 *         return <int>c_val
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py             # <<<<<<<<<<<<<<
 * 
 *     if 0:
*/
  __Pyx_INCREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py);
  __pyx_v___pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py;

  /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_PD:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_INIT_ATTR_PD
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_XRCD:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_INIT_ATTR_PD);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":30
 *         pass
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_PD:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_PD             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_XRCD:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_XRCD
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_INIT_ATTR_PD); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_PD:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_INIT_ATTR_PD
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_XRCD:
*/
  }

  /* "EnumTypeToPy":31
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_PD:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_PD
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_XRCD:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_INIT_ATTR_XRCD
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_CREATE_FLAGS:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_INIT_ATTR_XRCD);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":32
 *         return __pyx_enum.IBV_QP_INIT_ATTR_PD
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_XRCD:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_XRCD             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_CREATE_FLAGS:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_CREATE_FLAGS
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_INIT_ATTR_XRCD); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 32, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":31
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_PD:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_PD
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_XRCD:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_INIT_ATTR_XRCD
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_CREATE_FLAGS:
*/
  }

  /* "EnumTypeToPy":33
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_XRCD:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_XRCD
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_CREATE_FLAGS:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_INIT_ATTR_CREATE_FLAGS
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_MAX_TSO_HEADER:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_INIT_ATTR_CREATE_FLAGS);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":34
 *         return __pyx_enum.IBV_QP_INIT_ATTR_XRCD
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_CREATE_FLAGS:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_CREATE_FLAGS             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_MAX_TSO_HEADER:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_MAX_TSO_HEADER
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_INIT_ATTR_CREATE_FLAGS); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":33
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_XRCD:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_XRCD
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_CREATE_FLAGS:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_INIT_ATTR_CREATE_FLAGS
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_MAX_TSO_HEADER:
*/
  }

  /* "EnumTypeToPy":35
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_CREATE_FLAGS:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_CREATE_FLAGS
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_MAX_TSO_HEADER:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_INIT_ATTR_MAX_TSO_HEADER
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_IND_TABLE:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_INIT_ATTR_MAX_TSO_HEADER);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":36
 *         return __pyx_enum.IBV_QP_INIT_ATTR_CREATE_FLAGS
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_MAX_TSO_HEADER:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_MAX_TSO_HEADER             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_IND_TABLE:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_IND_TABLE
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_INIT_ATTR_MAX_TSO_HEADER); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":35
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_CREATE_FLAGS:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_CREATE_FLAGS
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_MAX_TSO_HEADER:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_INIT_ATTR_MAX_TSO_HEADER
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_IND_TABLE:
*/
  }

  /* "EnumTypeToPy":37
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_MAX_TSO_HEADER:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_MAX_TSO_HEADER
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_IND_TABLE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_INIT_ATTR_IND_TABLE
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_RX_HASH:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_INIT_ATTR_IND_TABLE);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":38
 *         return __pyx_enum.IBV_QP_INIT_ATTR_MAX_TSO_HEADER
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_IND_TABLE:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_IND_TABLE             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_RX_HASH:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_RX_HASH
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_INIT_ATTR_IND_TABLE); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":37
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_MAX_TSO_HEADER:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_MAX_TSO_HEADER
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_IND_TABLE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_INIT_ATTR_IND_TABLE
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_RX_HASH:
*/
  }

  /* "EnumTypeToPy":39
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_IND_TABLE:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_IND_TABLE
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_RX_HASH:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_INIT_ATTR_RX_HASH
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_SEND_OPS_FLAGS:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_INIT_ATTR_RX_HASH);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":40
 *         return __pyx_enum.IBV_QP_INIT_ATTR_IND_TABLE
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_RX_HASH:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_RX_HASH             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_SEND_OPS_FLAGS:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_SEND_OPS_FLAGS
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_INIT_ATTR_RX_HASH); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":39
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_IND_TABLE:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_IND_TABLE
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_RX_HASH:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_INIT_ATTR_RX_HASH
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_SEND_OPS_FLAGS:
*/
  }

  /* "EnumTypeToPy":41
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_RX_HASH:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_RX_HASH
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_SEND_OPS_FLAGS:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_INIT_ATTR_SEND_OPS_FLAGS
 *     else:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_QP_INIT_ATTR_SEND_OPS_FLAGS);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":42
 *         return __pyx_enum.IBV_QP_INIT_ATTR_RX_HASH
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_SEND_OPS_FLAGS:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_SEND_OPS_FLAGS             # <<<<<<<<<<<<<<
 *     else:
 *         underlying_c_val = <int>c_val
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_QP_INIT_ATTR_SEND_OPS_FLAGS); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":41
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_RX_HASH:
 *         return __pyx_enum.IBV_QP_INIT_ATTR_RX_HASH
 *     elif c_val == ibv_qp_init_attr_mask.IBV_QP_INIT_ATTR_SEND_OPS_FLAGS:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_QP_INIT_ATTR_SEND_OPS_FLAGS
 *     else:
*/
  }

  /* "EnumTypeToPy":44
 *         return __pyx_enum.IBV_QP_INIT_ATTR_SEND_OPS_FLAGS
 *     else:
 *         underlying_c_val = <int>c_val             # <<<<<<<<<<<<<<
 *         return __pyx_enum(underlying_c_val)
*/
  /*else*/ {
    __pyx_v_underlying_c_val = ((int)__pyx_v_c_val);

    /* "EnumTypeToPy":45
 *     else:
 *         underlying_c_val = <int>c_val
 *         return __pyx_enum(underlying_c_val)             # <<<<<<<<<<<<<<
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = NULL;
    __Pyx_INCREF(__pyx_v___pyx_enum);
    __pyx_t_5 = __pyx_v___pyx_enum; 
    __pyx_t_11 = __Pyx_PyLong_From_int(__pyx_v_underlying_c_val); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_13 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_8);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_13 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_11};
      __pyx_t_10 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 45, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
    }
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;
  }

  /* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py(ibv_qp_init_attr_mask c_val):
 *     cdef object __pyx_enum
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_enum);
  __Pyx_XDECREF(__pyx_v_warnings);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":36
 * 
 * cdef class QPCap(PyverbsObject):
 *     def __init__(self, max_send_wr=1, max_recv_wr=10, max_send_sge=1,             # <<<<<<<<<<<<<<
 *                  max_recv_sge=1, max_inline_data=0):
 *         """
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_5QPCap_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2qp_5QPCap___init__, "\n        Initializes a QPCap object with user-provided or default values.\n        :param max_send_wr: max number of outstanding WRs in the SQ\n        :param max_recv_wr: max number of outstanding WRs in the RQ\n        :param max_send_sge: Requested max number of scatter-gather elements in\n                             a WR in the SQ\n        :param max_recv_sge: Requested max number of scatter-gather elements in\n                             a WR in the RQ\n        :param max_inline_data: max number of data (bytes) that can be posted\n                                inline to the SQ, otherwise 0\n        :return:\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_7pyverbs_2qp_5QPCap___init__;
#endif
static int __pyx_pw_7pyverbs_2qp_5QPCap_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_max_send_wr = 0;
  PyObject *__pyx_v_max_recv_wr = 0;
  PyObject *__pyx_v_max_send_sge = 0;
  PyObject *__pyx_v_max_recv_sge = 0;
  PyObject *__pyx_v_max_inline_data = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_max_send_wr,&__pyx_mstate_global->__pyx_n_u_max_recv_wr,&__pyx_mstate_global->__pyx_n_u_max_send_sge,&__pyx_mstate_global->__pyx_n_u_max_recv_sge,&__pyx_mstate_global->__pyx_n_u_max_inline_data,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 36, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 36, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 36, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 36, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 36, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 36, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 36, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_1));
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_10));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_1));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_1));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    } else {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 36, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 36, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 36, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 36, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 36, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_1));
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_10));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_1));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_1));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    }
    __pyx_v_max_send_wr = values[0];
    __pyx_v_max_recv_wr = values[1];
    __pyx_v_max_send_sge = values[2];
    __pyx_v_max_recv_sge = values[3];
    __pyx_v_max_inline_data = values[4];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 5, __pyx_nargs); __PYX_ERR(0, 36, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPCap.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_5QPCap___init__(((struct __pyx_obj_7pyverbs_2qp_QPCap *)__pyx_v_self), __pyx_v_max_send_wr, __pyx_v_max_recv_wr, __pyx_v_max_send_sge, __pyx_v_max_recv_sge, __pyx_v_max_inline_data);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_5QPCap___init__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self, PyObject *__pyx_v_max_send_wr, PyObject *__pyx_v_max_recv_wr, PyObject *__pyx_v_max_send_sge, PyObject *__pyx_v_max_recv_sge, PyObject *__pyx_v_max_inline_data) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyverbs/qp.pyx":50
 *         :return:
 *         """
 *         super().__init__()             # <<<<<<<<<<<<<<
 *         self.cap.max_send_wr = max_send_wr
 *         self.cap.max_recv_wr = max_recv_wr
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPCap), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":51
 *         """
 *         super().__init__()
 *         self.cap.max_send_wr = max_send_wr             # <<<<<<<<<<<<<<
 *         self.cap.max_recv_wr = max_recv_wr
 *         self.cap.max_send_sge = max_send_sge
*/
  __pyx_t_6 = __Pyx_PyLong_As_unsigned_int(__pyx_v_max_send_wr); if (unlikely((__pyx_t_6 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 51, __pyx_L1_error)
  __pyx_v_self->cap.max_send_wr = __pyx_t_6;

  /* "pyverbs/qp.pyx":52
 *         super().__init__()
 *         self.cap.max_send_wr = max_send_wr
 *         self.cap.max_recv_wr = max_recv_wr             # <<<<<<<<<<<<<<
 *         self.cap.max_send_sge = max_send_sge
 *         self.cap.max_recv_sge = max_recv_sge
*/
  __pyx_t_6 = __Pyx_PyLong_As_unsigned_int(__pyx_v_max_recv_wr); if (unlikely((__pyx_t_6 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 52, __pyx_L1_error)
  __pyx_v_self->cap.max_recv_wr = __pyx_t_6;

  /* "pyverbs/qp.pyx":53
 *         self.cap.max_send_wr = max_send_wr
 *         self.cap.max_recv_wr = max_recv_wr
 *         self.cap.max_send_sge = max_send_sge             # <<<<<<<<<<<<<<
 *         self.cap.max_recv_sge = max_recv_sge
 *         self.cap.max_inline_data = max_inline_data
*/
  __pyx_t_6 = __Pyx_PyLong_As_unsigned_int(__pyx_v_max_send_sge); if (unlikely((__pyx_t_6 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 53, __pyx_L1_error)
  __pyx_v_self->cap.max_send_sge = __pyx_t_6;

  /* "pyverbs/qp.pyx":54
 *         self.cap.max_recv_wr = max_recv_wr
 *         self.cap.max_send_sge = max_send_sge
 *         self.cap.max_recv_sge = max_recv_sge             # <<<<<<<<<<<<<<
 *         self.cap.max_inline_data = max_inline_data
 * 
*/
  __pyx_t_6 = __Pyx_PyLong_As_unsigned_int(__pyx_v_max_recv_sge); if (unlikely((__pyx_t_6 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 54, __pyx_L1_error)
  __pyx_v_self->cap.max_recv_sge = __pyx_t_6;

  /* "pyverbs/qp.pyx":55
 *         self.cap.max_send_sge = max_send_sge
 *         self.cap.max_recv_sge = max_recv_sge
 *         self.cap.max_inline_data = max_inline_data             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_6 = __Pyx_PyLong_As_unsigned_int(__pyx_v_max_inline_data); if (unlikely((__pyx_t_6 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 55, __pyx_L1_error)
  __pyx_v_self->cap.max_inline_data = __pyx_t_6;

  /* "pyverbs/qp.pyx":36
 * 
 * cdef class QPCap(PyverbsObject):
 *     def __init__(self, max_send_wr=1, max_recv_wr=10, max_send_sge=1,             # <<<<<<<<<<<<<<
 *                  max_recv_sge=1, max_inline_data=0):
 *         """
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.qp.QPCap.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":57
 *         self.cap.max_inline_data = max_inline_data
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_send_wr(self):
 *         return self.cap.max_send_wr
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_5QPCap_11max_send_wr_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_5QPCap_11max_send_wr_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_5QPCap_11max_send_wr___get__(((struct __pyx_obj_7pyverbs_2qp_QPCap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_5QPCap_11max_send_wr___get__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":59
 *     @property
 *     def max_send_wr(self):
 *         return self.cap.max_send_wr             # <<<<<<<<<<<<<<
 *     @max_send_wr.setter
 *     def max_send_wr(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->cap.max_send_wr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":57
 *         self.cap.max_inline_data = max_inline_data
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_send_wr(self):
 *         return self.cap.max_send_wr
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPCap.max_send_wr.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":60
 *     def max_send_wr(self):
 *         return self.cap.max_send_wr
 *     @max_send_wr.setter             # <<<<<<<<<<<<<<
 *     def max_send_wr(self, val):
 *         self.cap.max_send_wr = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_5QPCap_11max_send_wr_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_5QPCap_11max_send_wr_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_5QPCap_11max_send_wr_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPCap *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_5QPCap_11max_send_wr_2__set__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":62
 *     @max_send_wr.setter
 *     def max_send_wr(self, val):
 *         self.cap.max_send_wr = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 62, __pyx_L1_error)
  __pyx_v_self->cap.max_send_wr = __pyx_t_1;

  /* "pyverbs/qp.pyx":60
 *     def max_send_wr(self):
 *         return self.cap.max_send_wr
 *     @max_send_wr.setter             # <<<<<<<<<<<<<<
 *     def max_send_wr(self, val):
 *         self.cap.max_send_wr = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPCap.max_send_wr.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":64
 *         self.cap.max_send_wr = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_recv_wr(self):
 *         return self.cap.max_recv_wr
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_5QPCap_11max_recv_wr_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_5QPCap_11max_recv_wr_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_5QPCap_11max_recv_wr___get__(((struct __pyx_obj_7pyverbs_2qp_QPCap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_5QPCap_11max_recv_wr___get__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":66
 *     @property
 *     def max_recv_wr(self):
 *         return self.cap.max_recv_wr             # <<<<<<<<<<<<<<
 *     @max_recv_wr.setter
 *     def max_recv_wr(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->cap.max_recv_wr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":64
 *         self.cap.max_send_wr = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_recv_wr(self):
 *         return self.cap.max_recv_wr
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPCap.max_recv_wr.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":67
 *     def max_recv_wr(self):
 *         return self.cap.max_recv_wr
 *     @max_recv_wr.setter             # <<<<<<<<<<<<<<
 *     def max_recv_wr(self, val):
 *         self.cap.max_recv_wr = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_5QPCap_11max_recv_wr_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_5QPCap_11max_recv_wr_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_5QPCap_11max_recv_wr_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPCap *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_5QPCap_11max_recv_wr_2__set__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":69
 *     @max_recv_wr.setter
 *     def max_recv_wr(self, val):
 *         self.cap.max_recv_wr = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 69, __pyx_L1_error)
  __pyx_v_self->cap.max_recv_wr = __pyx_t_1;

  /* "pyverbs/qp.pyx":67
 *     def max_recv_wr(self):
 *         return self.cap.max_recv_wr
 *     @max_recv_wr.setter             # <<<<<<<<<<<<<<
 *     def max_recv_wr(self, val):
 *         self.cap.max_recv_wr = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPCap.max_recv_wr.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":71
 *         self.cap.max_recv_wr = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_send_sge(self):
 *         return self.cap.max_send_sge
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_5QPCap_12max_send_sge_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_5QPCap_12max_send_sge_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_5QPCap_12max_send_sge___get__(((struct __pyx_obj_7pyverbs_2qp_QPCap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_5QPCap_12max_send_sge___get__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":73
 *     @property
 *     def max_send_sge(self):
 *         return self.cap.max_send_sge             # <<<<<<<<<<<<<<
 *     @max_send_sge.setter
 *     def max_send_sge(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->cap.max_send_sge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":71
 *         self.cap.max_recv_wr = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_send_sge(self):
 *         return self.cap.max_send_sge
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPCap.max_send_sge.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":74
 *     def max_send_sge(self):
 *         return self.cap.max_send_sge
 *     @max_send_sge.setter             # <<<<<<<<<<<<<<
 *     def max_send_sge(self, val):
 *         self.cap.max_send_sge = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_5QPCap_12max_send_sge_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_5QPCap_12max_send_sge_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_5QPCap_12max_send_sge_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPCap *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_5QPCap_12max_send_sge_2__set__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":76
 *     @max_send_sge.setter
 *     def max_send_sge(self, val):
 *         self.cap.max_send_sge = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 76, __pyx_L1_error)
  __pyx_v_self->cap.max_send_sge = __pyx_t_1;

  /* "pyverbs/qp.pyx":74
 *     def max_send_sge(self):
 *         return self.cap.max_send_sge
 *     @max_send_sge.setter             # <<<<<<<<<<<<<<
 *     def max_send_sge(self, val):
 *         self.cap.max_send_sge = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPCap.max_send_sge.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":78
 *         self.cap.max_send_sge = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_recv_sge(self):
 *         return self.cap.max_recv_sge
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_5QPCap_12max_recv_sge_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_5QPCap_12max_recv_sge_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_5QPCap_12max_recv_sge___get__(((struct __pyx_obj_7pyverbs_2qp_QPCap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_5QPCap_12max_recv_sge___get__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":80
 *     @property
 *     def max_recv_sge(self):
 *         return self.cap.max_recv_sge             # <<<<<<<<<<<<<<
 *     @max_recv_sge.setter
 *     def max_recv_sge(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->cap.max_recv_sge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":78
 *         self.cap.max_send_sge = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_recv_sge(self):
 *         return self.cap.max_recv_sge
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPCap.max_recv_sge.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":81
 *     def max_recv_sge(self):
 *         return self.cap.max_recv_sge
 *     @max_recv_sge.setter             # <<<<<<<<<<<<<<
 *     def max_recv_sge(self, val):
 *         self.cap.max_recv_sge = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_5QPCap_12max_recv_sge_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_5QPCap_12max_recv_sge_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_5QPCap_12max_recv_sge_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPCap *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_5QPCap_12max_recv_sge_2__set__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":83
 *     @max_recv_sge.setter
 *     def max_recv_sge(self, val):
 *         self.cap.max_recv_sge = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 83, __pyx_L1_error)
  __pyx_v_self->cap.max_recv_sge = __pyx_t_1;

  /* "pyverbs/qp.pyx":81
 *     def max_recv_sge(self):
 *         return self.cap.max_recv_sge
 *     @max_recv_sge.setter             # <<<<<<<<<<<<<<
 *     def max_recv_sge(self, val):
 *         self.cap.max_recv_sge = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPCap.max_recv_sge.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":85
 *         self.cap.max_recv_sge = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_inline_data(self):
 *         return self.cap.max_inline_data
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_5QPCap_15max_inline_data_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_5QPCap_15max_inline_data_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_5QPCap_15max_inline_data___get__(((struct __pyx_obj_7pyverbs_2qp_QPCap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_5QPCap_15max_inline_data___get__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":87
 *     @property
 *     def max_inline_data(self):
 *         return self.cap.max_inline_data             # <<<<<<<<<<<<<<
 *     @max_inline_data.setter
 *     def max_inline_data(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->cap.max_inline_data); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":85
 *         self.cap.max_recv_sge = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_inline_data(self):
 *         return self.cap.max_inline_data
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPCap.max_inline_data.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":88
 *     def max_inline_data(self):
 *         return self.cap.max_inline_data
 *     @max_inline_data.setter             # <<<<<<<<<<<<<<
 *     def max_inline_data(self, val):
 *         self.cap.max_inline_data = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_5QPCap_15max_inline_data_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_5QPCap_15max_inline_data_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_5QPCap_15max_inline_data_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPCap *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_5QPCap_15max_inline_data_2__set__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":90
 *     @max_inline_data.setter
 *     def max_inline_data(self, val):
 *         self.cap.max_inline_data = val             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 90, __pyx_L1_error)
  __pyx_v_self->cap.max_inline_data = __pyx_t_1;

  /* "pyverbs/qp.pyx":88
 *     def max_inline_data(self):
 *         return self.cap.max_inline_data
 *     @max_inline_data.setter             # <<<<<<<<<<<<<<
 *     def max_inline_data(self, val):
 *         self.cap.max_inline_data = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPCap.max_inline_data.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":92
 *         self.cap.max_inline_data = val
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         print_format = '{:20}: {:<20}\n'
 *         return print_format.format('max send wrs', self.cap.max_send_wr) +\
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_5QPCap_3__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_5QPCap_3__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_5QPCap_2__str__(((struct __pyx_obj_7pyverbs_2qp_QPCap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_5QPCap_2__str__(struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self) {
  PyObject *__pyx_v_print_format = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "pyverbs/qp.pyx":93
 * 
 *     def __str__(self):
 *         print_format = '{:20}: {:<20}\n'             # <<<<<<<<<<<<<<
 *         return print_format.format('max send wrs', self.cap.max_send_wr) +\
 *                print_format.format('max recv wrs', self.cap.max_recv_wr) +\
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_20_20);
  __pyx_v_print_format = __pyx_mstate_global->__pyx_kp_u_20_20;

  /* "pyverbs/qp.pyx":94
 *     def __str__(self):
 *         print_format = '{:20}: {:<20}\n'
 *         return print_format.format('max send wrs', self.cap.max_send_wr) +\             # <<<<<<<<<<<<<<
 *                print_format.format('max recv wrs', self.cap.max_recv_wr) +\
 *                print_format.format('max send sges', self.cap.max_send_sge) +\
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->cap.max_send_wr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_max_send_wrs, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":95
 *         print_format = '{:20}: {:<20}\n'
 *         return print_format.format('max send wrs', self.cap.max_send_wr) +\
 *                print_format.format('max recv wrs', self.cap.max_recv_wr) +\             # <<<<<<<<<<<<<<
 *                print_format.format('max send sges', self.cap.max_send_sge) +\
 *                print_format.format('max recv sges', self.cap.max_recv_sge) +\
*/
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->cap.max_recv_wr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_max_recv_wrs, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":94
 *     def __str__(self):
 *         print_format = '{:20}: {:<20}\n'
 *         return print_format.format('max send wrs', self.cap.max_send_wr) +\             # <<<<<<<<<<<<<<
 *                print_format.format('max recv wrs', self.cap.max_recv_wr) +\
 *                print_format.format('max send sges', self.cap.max_send_sge) +\
*/
  __pyx_t_1 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":96
 *         return print_format.format('max send wrs', self.cap.max_send_wr) +\
 *                print_format.format('max recv wrs', self.cap.max_recv_wr) +\
 *                print_format.format('max send sges', self.cap.max_send_sge) +\             # <<<<<<<<<<<<<<
 *                print_format.format('max recv sges', self.cap.max_recv_sge) +\
 *                print_format.format('max inline data', self.cap.max_inline_data)
*/
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->cap.max_send_sge); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_max_send_sges, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":95
 *         print_format = '{:20}: {:<20}\n'
 *         return print_format.format('max send wrs', self.cap.max_send_wr) +\
 *                print_format.format('max recv wrs', self.cap.max_recv_wr) +\             # <<<<<<<<<<<<<<
 *                print_format.format('max send sges', self.cap.max_send_sge) +\
 *                print_format.format('max recv sges', self.cap.max_recv_sge) +\
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":97
 *                print_format.format('max recv wrs', self.cap.max_recv_wr) +\
 *                print_format.format('max send sges', self.cap.max_send_sge) +\
 *                print_format.format('max recv sges', self.cap.max_recv_sge) +\             # <<<<<<<<<<<<<<
 *                print_format.format('max inline data', self.cap.max_inline_data)
 * 
*/
  __pyx_t_2 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->cap.max_recv_sge); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_max_recv_sges, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":96
 *         return print_format.format('max send wrs', self.cap.max_send_wr) +\
 *                print_format.format('max recv wrs', self.cap.max_recv_wr) +\
 *                print_format.format('max send sges', self.cap.max_send_sge) +\             # <<<<<<<<<<<<<<
 *                print_format.format('max recv sges', self.cap.max_recv_sge) +\
 *                print_format.format('max inline data', self.cap.max_inline_data)
*/
  __pyx_t_2 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":98
 *                print_format.format('max send sges', self.cap.max_send_sge) +\
 *                print_format.format('max recv sges', self.cap.max_recv_sge) +\
 *                print_format.format('max inline data', self.cap.max_inline_data)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->cap.max_inline_data); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_max_inline_data_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":97
 *                print_format.format('max recv wrs', self.cap.max_recv_wr) +\
 *                print_format.format('max send sges', self.cap.max_send_sge) +\
 *                print_format.format('max recv sges', self.cap.max_recv_sge) +\             # <<<<<<<<<<<<<<
 *                print_format.format('max inline data', self.cap.max_inline_data)
 * 
*/
  __pyx_t_1 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":92
 *         self.cap.max_inline_data = val
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         print_format = '{:20}: {:<20}\n'
 *         return print_format.format('max send wrs', self.cap.max_send_wr) +\
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyverbs.qp.QPCap.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_print_format);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.cap must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_5QPCap_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_5QPCap_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_5QPCap_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_5QPCap_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2qp_5QPCap_4__reduce_cython__(((struct __pyx_obj_7pyverbs_2qp_QPCap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_5QPCap_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.cap must be explicitly requested with @auto_pickle(True)"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "Pickling of struct members such as self.cap must be explicitly requested with @auto_pickle(True)"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_Pickling_of_struct_members_such, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.cap must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPCap.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.cap must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.cap must be explicitly requested with @auto_pickle(True)"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_5QPCap_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_5QPCap_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_5QPCap_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_5QPCap_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPCap.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_5QPCap_6__setstate_cython__(((struct __pyx_obj_7pyverbs_2qp_QPCap *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_5QPCap_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "Pickling of struct members such as self.cap must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "Pickling of struct members such as self.cap must be explicitly requested with @auto_pickle(True)"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_Pickling_of_struct_members_such, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.cap must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.cap must be explicitly requested with @auto_pickle(True)"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPCap.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":102
 * 
 * cdef class QPInitAttr(PyverbsObject):
 *     def __init__(self, qp_type=e.IBV_QPT_UD, qp_context=None,             # <<<<<<<<<<<<<<
 *                  PyverbsObject scq=None, PyverbsObject rcq=None,
 *                  SRQ srq=None, QPCap cap=None, sq_sig_all=1):
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2qp_10QPInitAttr___init__, "\n        Initializes a QpInitAttr object representing ibv_qp_init_attr struct.\n        :param qp_type: The desired QP type (see enum ibv_qp_type)\n        :param qp_context: Associated QP context\n        :param scq: Send CQ to be used for this QP\n        :param rcq: Receive CQ to be used for this QP\n        :param srq: Shared receive queue to be used as RQ in QP\n        :param cap: A QPCap object\n        :param sq_sig_all: If set, each send WR will generate a completion\n                           entry\n        :return: A QpInitAttr object\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_7pyverbs_2qp_10QPInitAttr___init__;
#endif
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_qp_type = 0;
  PyObject *__pyx_v_qp_context = 0;
  struct __pyx_obj_7pyverbs_4base_PyverbsObject *__pyx_v_scq = 0;
  struct __pyx_obj_7pyverbs_4base_PyverbsObject *__pyx_v_rcq = 0;
  struct __pyx_obj_7pyverbs_3srq_SRQ *__pyx_v_srq = 0;
  struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_cap = 0;
  PyObject *__pyx_v_sq_sig_all = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[7] = {0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_qp_type,&__pyx_mstate_global->__pyx_n_u_qp_context,&__pyx_mstate_global->__pyx_n_u_scq,&__pyx_mstate_global->__pyx_n_u_rcq,&__pyx_mstate_global->__pyx_n_u_srq,&__pyx_mstate_global->__pyx_n_u_cap,&__pyx_mstate_global->__pyx_n_u_sq_sig_all,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 102, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  7:
        values[6] = __Pyx_ArgRef_VARARGS(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 102, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 102, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 102, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 102, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 102, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 102, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 102, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 102, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__2);
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "pyverbs/qp.pyx":103
 * cdef class QPInitAttr(PyverbsObject):
 *     def __init__(self, qp_type=e.IBV_QPT_UD, qp_context=None,
 *                  PyverbsObject scq=None, PyverbsObject rcq=None,             # <<<<<<<<<<<<<<
 *                  SRQ srq=None, QPCap cap=None, sq_sig_all=1):
 *         """
*/
      if (!values[2]) values[2] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_4base_PyverbsObject *)Py_None));
      if (!values[3]) values[3] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_4base_PyverbsObject *)Py_None));

      /* "pyverbs/qp.pyx":104
 *     def __init__(self, qp_type=e.IBV_QPT_UD, qp_context=None,
 *                  PyverbsObject scq=None, PyverbsObject rcq=None,
 *                  SRQ srq=None, QPCap cap=None, sq_sig_all=1):             # <<<<<<<<<<<<<<
 *         """
 *         Initializes a QpInitAttr object representing ibv_qp_init_attr struct.
*/
      if (!values[4]) values[4] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_3srq_SRQ *)Py_None));
      if (!values[5]) values[5] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2qp_QPCap *)Py_None));
      if (!values[6]) values[6] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_1));
    } else {
      switch (__pyx_nargs) {
        case  7:
        values[6] = __Pyx_ArgRef_VARARGS(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 102, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 102, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 102, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 102, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 102, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 102, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 102, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__2);

      /* "pyverbs/qp.pyx":102
 * 
 * cdef class QPInitAttr(PyverbsObject):
 *     def __init__(self, qp_type=e.IBV_QPT_UD, qp_context=None,             # <<<<<<<<<<<<<<
 *                  PyverbsObject scq=None, PyverbsObject rcq=None,
 *                  SRQ srq=None, QPCap cap=None, sq_sig_all=1):
*/
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "pyverbs/qp.pyx":103
 * cdef class QPInitAttr(PyverbsObject):
 *     def __init__(self, qp_type=e.IBV_QPT_UD, qp_context=None,
 *                  PyverbsObject scq=None, PyverbsObject rcq=None,             # <<<<<<<<<<<<<<
 *                  SRQ srq=None, QPCap cap=None, sq_sig_all=1):
 *         """
*/
      if (!values[2]) values[2] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_4base_PyverbsObject *)Py_None));
      if (!values[3]) values[3] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_4base_PyverbsObject *)Py_None));

      /* "pyverbs/qp.pyx":104
 *     def __init__(self, qp_type=e.IBV_QPT_UD, qp_context=None,
 *                  PyverbsObject scq=None, PyverbsObject rcq=None,
 *                  SRQ srq=None, QPCap cap=None, sq_sig_all=1):             # <<<<<<<<<<<<<<
 *         """
 *         Initializes a QpInitAttr object representing ibv_qp_init_attr struct.
*/
      if (!values[4]) values[4] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_3srq_SRQ *)Py_None));
      if (!values[5]) values[5] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2qp_QPCap *)Py_None));
      if (!values[6]) values[6] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_1));
    }
    __pyx_v_qp_type = values[0];
    __pyx_v_qp_context = values[1];
    __pyx_v_scq = ((struct __pyx_obj_7pyverbs_4base_PyverbsObject *)values[2]);
    __pyx_v_rcq = ((struct __pyx_obj_7pyverbs_4base_PyverbsObject *)values[3]);
    __pyx_v_srq = ((struct __pyx_obj_7pyverbs_3srq_SRQ *)values[4]);
    __pyx_v_cap = ((struct __pyx_obj_7pyverbs_2qp_QPCap *)values[5]);
    __pyx_v_sq_sig_all = values[6];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 7, __pyx_nargs); __PYX_ERR(0, 102, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttr.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_scq), __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, 1, "scq", 0))) __PYX_ERR(0, 103, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_rcq), __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, 1, "rcq", 0))) __PYX_ERR(0, 103, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_srq), __pyx_mstate_global->__pyx_ptype_7pyverbs_3srq_SRQ, 1, "srq", 0))) __PYX_ERR(0, 104, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_cap), __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPCap, 1, "cap", 0))) __PYX_ERR(0, 104, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr___init__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self), __pyx_v_qp_type, __pyx_v_qp_context, __pyx_v_scq, __pyx_v_rcq, __pyx_v_srq, __pyx_v_cap, __pyx_v_sq_sig_all);

  /* "pyverbs/qp.pyx":102
 * 
 * cdef class QPInitAttr(PyverbsObject):
 *     def __init__(self, qp_type=e.IBV_QPT_UD, qp_context=None,             # <<<<<<<<<<<<<<
 *                  PyverbsObject scq=None, PyverbsObject rcq=None,
 *                  SRQ srq=None, QPCap cap=None, sq_sig_all=1):
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_10QPInitAttr___init__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, PyObject *__pyx_v_qp_type, PyObject *__pyx_v_qp_context, struct __pyx_obj_7pyverbs_4base_PyverbsObject *__pyx_v_scq, struct __pyx_obj_7pyverbs_4base_PyverbsObject *__pyx_v_rcq, struct __pyx_obj_7pyverbs_3srq_SRQ *__pyx_v_srq, struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_cap, PyObject *__pyx_v_sq_sig_all) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  struct ibv_cq *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  enum ibv_qp_type __pyx_t_10;
  int __pyx_t_11;
  struct ibv_srq *__pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyverbs/qp.pyx":117
 *         :return: A QpInitAttr object
 *         """
 *         super().__init__()             # <<<<<<<<<<<<<<
 *         _copy_caps(cap, self)
 *         self.attr.qp_context = <void*>qp_context
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPInitAttr), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":118
 *         """
 *         super().__init__()
 *         _copy_caps(cap, self)             # <<<<<<<<<<<<<<
 *         self.attr.qp_context = <void*>qp_context
 *         if scq is not None:
*/
  __pyx_t_3 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_copy_caps); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    assert(__pyx_t_3);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_3, ((PyObject *)__pyx_v_cap), ((PyObject *)__pyx_v_self)};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":119
 *         super().__init__()
 *         _copy_caps(cap, self)
 *         self.attr.qp_context = <void*>qp_context             # <<<<<<<<<<<<<<
 *         if scq is not None:
 *             if type(scq) is CQ:
*/
  __pyx_v_self->attr.qp_context = ((void *)__pyx_v_qp_context);

  /* "pyverbs/qp.pyx":120
 *         _copy_caps(cap, self)
 *         self.attr.qp_context = <void*>qp_context
 *         if scq is not None:             # <<<<<<<<<<<<<<
 *             if type(scq) is CQ:
 *                 self.attr.send_cq = (<CQ>scq).cq
*/
  __pyx_t_6 = (((PyObject *)__pyx_v_scq) != Py_None);
  if (__pyx_t_6) {

    /* "pyverbs/qp.pyx":121
 *         self.attr.qp_context = <void*>qp_context
 *         if scq is not None:
 *             if type(scq) is CQ:             # <<<<<<<<<<<<<<
 *                 self.attr.send_cq = (<CQ>scq).cq
 *             elif isinstance(scq, CQEX):
*/
    __pyx_t_6 = (((PyObject *)Py_TYPE(((PyObject *)__pyx_v_scq))) == ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQ));
    if (__pyx_t_6) {

      /* "pyverbs/qp.pyx":122
 *         if scq is not None:
 *             if type(scq) is CQ:
 *                 self.attr.send_cq = (<CQ>scq).cq             # <<<<<<<<<<<<<<
 *             elif isinstance(scq, CQEX):
 *                 self.attr.send_cq = (<CQEX>scq).ibv_cq
*/
      __pyx_t_7 = ((struct __pyx_obj_7pyverbs_2cq_CQ *)__pyx_v_scq)->cq;
      __pyx_v_self->attr.send_cq = __pyx_t_7;

      /* "pyverbs/qp.pyx":121
 *         self.attr.qp_context = <void*>qp_context
 *         if scq is not None:
 *             if type(scq) is CQ:             # <<<<<<<<<<<<<<
 *                 self.attr.send_cq = (<CQ>scq).cq
 *             elif isinstance(scq, CQEX):
*/
      goto __pyx_L4;
    }

    /* "pyverbs/qp.pyx":123
 *             if type(scq) is CQ:
 *                 self.attr.send_cq = (<CQ>scq).cq
 *             elif isinstance(scq, CQEX):             # <<<<<<<<<<<<<<
 *                 self.attr.send_cq = (<CQEX>scq).ibv_cq
 *             else:
*/
    __pyx_t_6 = __Pyx_TypeCheck(((PyObject *)__pyx_v_scq), __pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX); 
    if (likely(__pyx_t_6)) {

      /* "pyverbs/qp.pyx":124
 *                 self.attr.send_cq = (<CQ>scq).cq
 *             elif isinstance(scq, CQEX):
 *                 self.attr.send_cq = (<CQEX>scq).ibv_cq             # <<<<<<<<<<<<<<
 *             else:
 *                 raise PyverbsUserError('Expected CQ/CQEX, got {t}'.\
*/
      __pyx_t_7 = ((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_scq)->ibv_cq;
      __pyx_v_self->attr.send_cq = __pyx_t_7;

      /* "pyverbs/qp.pyx":123
 *             if type(scq) is CQ:
 *                 self.attr.send_cq = (<CQ>scq).cq
 *             elif isinstance(scq, CQEX):             # <<<<<<<<<<<<<<
 *                 self.attr.send_cq = (<CQEX>scq).ibv_cq
 *             else:
*/
      goto __pyx_L4;
    }

    /* "pyverbs/qp.pyx":126
 *                 self.attr.send_cq = (<CQEX>scq).ibv_cq
 *             else:
 *                 raise PyverbsUserError('Expected CQ/CQEX, got {t}'.\             # <<<<<<<<<<<<<<
 *                                        format(t=type(scq)))
 *         self.scq = scq
*/
    /*else*/ {
      __pyx_t_2 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_PyverbsUserError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 126, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "pyverbs/qp.pyx":127
 *             else:
 *                 raise PyverbsUserError('Expected CQ/CQEX, got {t}'.\
 *                                        format(t=type(scq)))             # <<<<<<<<<<<<<<
 *         self.scq = scq
 * 
*/
      __pyx_t_8 = __pyx_mstate_global->__pyx_kp_u_Expected_CQ_CQEX_got_t;
      __Pyx_INCREF(__pyx_t_8);
      __pyx_t_5 = 0;
      {
        PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_8, NULL};
        __pyx_t_9 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 127, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_t, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_scq))), __pyx_t_9, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 127, __pyx_L1_error)
        __pyx_t_4 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_format, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_9);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 127, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
      }
      __pyx_t_5 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
        assert(__pyx_t_2);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
        __pyx_t_5 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_4};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 126, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 126, __pyx_L1_error)
    }
    __pyx_L4:;

    /* "pyverbs/qp.pyx":120
 *         _copy_caps(cap, self)
 *         self.attr.qp_context = <void*>qp_context
 *         if scq is not None:             # <<<<<<<<<<<<<<
 *             if type(scq) is CQ:
 *                 self.attr.send_cq = (<CQ>scq).cq
*/
  }

  /* "pyverbs/qp.pyx":128
 *                 raise PyverbsUserError('Expected CQ/CQEX, got {t}'.\
 *                                        format(t=type(scq)))
 *         self.scq = scq             # <<<<<<<<<<<<<<
 * 
 *         if rcq is not None:
*/
  __Pyx_INCREF((PyObject *)__pyx_v_scq);
  __Pyx_GIVEREF((PyObject *)__pyx_v_scq);
  __Pyx_GOTREF(__pyx_v_self->scq);
  __Pyx_DECREF(__pyx_v_self->scq);
  __pyx_v_self->scq = ((PyObject *)__pyx_v_scq);

  /* "pyverbs/qp.pyx":130
 *         self.scq = scq
 * 
 *         if rcq is not None:             # <<<<<<<<<<<<<<
 *             if type(rcq) is CQ:
 *                 self.attr.recv_cq = (<CQ>rcq).cq
*/
  __pyx_t_6 = (((PyObject *)__pyx_v_rcq) != Py_None);
  if (__pyx_t_6) {

    /* "pyverbs/qp.pyx":131
 * 
 *         if rcq is not None:
 *             if type(rcq) is CQ:             # <<<<<<<<<<<<<<
 *                 self.attr.recv_cq = (<CQ>rcq).cq
 *             elif isinstance(rcq, CQEX):
*/
    __pyx_t_6 = (((PyObject *)Py_TYPE(((PyObject *)__pyx_v_rcq))) == ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQ));
    if (__pyx_t_6) {

      /* "pyverbs/qp.pyx":132
 *         if rcq is not None:
 *             if type(rcq) is CQ:
 *                 self.attr.recv_cq = (<CQ>rcq).cq             # <<<<<<<<<<<<<<
 *             elif isinstance(rcq, CQEX):
 *                 self.attr.recv_cq = (<CQEX>rcq).ibv_cq
*/
      __pyx_t_7 = ((struct __pyx_obj_7pyverbs_2cq_CQ *)__pyx_v_rcq)->cq;
      __pyx_v_self->attr.recv_cq = __pyx_t_7;

      /* "pyverbs/qp.pyx":131
 * 
 *         if rcq is not None:
 *             if type(rcq) is CQ:             # <<<<<<<<<<<<<<
 *                 self.attr.recv_cq = (<CQ>rcq).cq
 *             elif isinstance(rcq, CQEX):
*/
      goto __pyx_L6;
    }

    /* "pyverbs/qp.pyx":133
 *             if type(rcq) is CQ:
 *                 self.attr.recv_cq = (<CQ>rcq).cq
 *             elif isinstance(rcq, CQEX):             # <<<<<<<<<<<<<<
 *                 self.attr.recv_cq = (<CQEX>rcq).ibv_cq
 *             else:
*/
    __pyx_t_6 = __Pyx_TypeCheck(((PyObject *)__pyx_v_rcq), __pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX); 
    if (likely(__pyx_t_6)) {

      /* "pyverbs/qp.pyx":134
 *                 self.attr.recv_cq = (<CQ>rcq).cq
 *             elif isinstance(rcq, CQEX):
 *                 self.attr.recv_cq = (<CQEX>rcq).ibv_cq             # <<<<<<<<<<<<<<
 *             else:
 *                 raise PyverbsUserError('Expected CQ/CQEX, got {t}'.\
*/
      __pyx_t_7 = ((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_rcq)->ibv_cq;
      __pyx_v_self->attr.recv_cq = __pyx_t_7;

      /* "pyverbs/qp.pyx":133
 *             if type(rcq) is CQ:
 *                 self.attr.recv_cq = (<CQ>rcq).cq
 *             elif isinstance(rcq, CQEX):             # <<<<<<<<<<<<<<
 *                 self.attr.recv_cq = (<CQEX>rcq).ibv_cq
 *             else:
*/
      goto __pyx_L6;
    }

    /* "pyverbs/qp.pyx":136
 *                 self.attr.recv_cq = (<CQEX>rcq).ibv_cq
 *             else:
 *                 raise PyverbsUserError('Expected CQ/CQEX, got {t}'.\             # <<<<<<<<<<<<<<
 *                                        format(t=type(rcq)))
 *         self.rcq = rcq
*/
    /*else*/ {
      __pyx_t_3 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsUserError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 136, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "pyverbs/qp.pyx":137
 *             else:
 *                 raise PyverbsUserError('Expected CQ/CQEX, got {t}'.\
 *                                        format(t=type(rcq)))             # <<<<<<<<<<<<<<
 *         self.rcq = rcq
 *         self.attr.qp_type = qp_type
*/
      __pyx_t_9 = __pyx_mstate_global->__pyx_kp_u_Expected_CQ_CQEX_got_t;
      __Pyx_INCREF(__pyx_t_9);
      __pyx_t_5 = 0;
      {
        PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_9, NULL};
        __pyx_t_8 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 137, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_t, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_rcq))), __pyx_t_8, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 137, __pyx_L1_error)
        __pyx_t_2 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_format, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_8);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 137, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __pyx_t_5 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
        assert(__pyx_t_3);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
        __pyx_t_5 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_2};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 136, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 136, __pyx_L1_error)
    }
    __pyx_L6:;

    /* "pyverbs/qp.pyx":130
 *         self.scq = scq
 * 
 *         if rcq is not None:             # <<<<<<<<<<<<<<
 *             if type(rcq) is CQ:
 *                 self.attr.recv_cq = (<CQ>rcq).cq
*/
  }

  /* "pyverbs/qp.pyx":138
 *                 raise PyverbsUserError('Expected CQ/CQEX, got {t}'.\
 *                                        format(t=type(rcq)))
 *         self.rcq = rcq             # <<<<<<<<<<<<<<
 *         self.attr.qp_type = qp_type
 *         self.attr.sq_sig_all = sq_sig_all
*/
  __Pyx_INCREF((PyObject *)__pyx_v_rcq);
  __Pyx_GIVEREF((PyObject *)__pyx_v_rcq);
  __Pyx_GOTREF(__pyx_v_self->rcq);
  __Pyx_DECREF(__pyx_v_self->rcq);
  __pyx_v_self->rcq = ((PyObject *)__pyx_v_rcq);

  /* "pyverbs/qp.pyx":139
 *                                        format(t=type(rcq)))
 *         self.rcq = rcq
 *         self.attr.qp_type = qp_type             # <<<<<<<<<<<<<<
 *         self.attr.sq_sig_all = sq_sig_all
 *         self.srq = srq
*/
  __pyx_t_10 = ((enum ibv_qp_type)__Pyx_PyLong_As_enum__ibv_qp_type(__pyx_v_qp_type)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 139, __pyx_L1_error)
  __pyx_v_self->attr.qp_type = __pyx_t_10;

  /* "pyverbs/qp.pyx":140
 *         self.rcq = rcq
 *         self.attr.qp_type = qp_type
 *         self.attr.sq_sig_all = sq_sig_all             # <<<<<<<<<<<<<<
 *         self.srq = srq
 *         self.attr.srq = srq.srq if srq else NULL
*/
  __pyx_t_11 = __Pyx_PyLong_As_int(__pyx_v_sq_sig_all); if (unlikely((__pyx_t_11 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 140, __pyx_L1_error)
  __pyx_v_self->attr.sq_sig_all = __pyx_t_11;

  /* "pyverbs/qp.pyx":141
 *         self.attr.qp_type = qp_type
 *         self.attr.sq_sig_all = sq_sig_all
 *         self.srq = srq             # <<<<<<<<<<<<<<
 *         self.attr.srq = srq.srq if srq else NULL
 * 
*/
  __Pyx_INCREF((PyObject *)__pyx_v_srq);
  __Pyx_GIVEREF((PyObject *)__pyx_v_srq);
  __Pyx_GOTREF(__pyx_v_self->srq);
  __Pyx_DECREF(__pyx_v_self->srq);
  __pyx_v_self->srq = ((PyObject *)__pyx_v_srq);

  /* "pyverbs/qp.pyx":142
 *         self.attr.sq_sig_all = sq_sig_all
 *         self.srq = srq
 *         self.attr.srq = srq.srq if srq else NULL             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_6 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_srq)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 142, __pyx_L1_error)
  if (__pyx_t_6) {
    __pyx_t_12 = __pyx_v_srq->srq;
  } else {
    __pyx_t_12 = NULL;
  }
  __pyx_v_self->attr.srq = __pyx_t_12;

  /* "pyverbs/qp.pyx":102
 * 
 * cdef class QPInitAttr(PyverbsObject):
 *     def __init__(self, qp_type=e.IBV_QPT_UD, qp_context=None,             # <<<<<<<<<<<<<<
 *                  PyverbsObject scq=None, PyverbsObject rcq=None,
 *                  SRQ srq=None, QPCap cap=None, sq_sig_all=1):
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttr.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":144
 *         self.attr.srq = srq.srq if srq else NULL
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def send_cq(self):
 *         return self.scq
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_7send_cq_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_7send_cq_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_7send_cq___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_7send_cq___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":146
 *     @property
 *     def send_cq(self):
 *         return self.scq             # <<<<<<<<<<<<<<
 *     @send_cq.setter
 *     def send_cq(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->scq);
  __pyx_r = __pyx_v_self->scq;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":144
 *         self.attr.srq = srq.srq if srq else NULL
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def send_cq(self):
 *         return self.scq
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":147
 *     def send_cq(self):
 *         return self.scq
 *     @send_cq.setter             # <<<<<<<<<<<<<<
 *     def send_cq(self, val):
 *         if type(val) is CQ:
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_7send_cq_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_7send_cq_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_7send_cq_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_10QPInitAttr_7send_cq_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  struct ibv_cq *__pyx_t_2;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "pyverbs/qp.pyx":149
 *     @send_cq.setter
 *     def send_cq(self, val):
 *         if type(val) is CQ:             # <<<<<<<<<<<<<<
 *             self.attr.send_cq = (<CQ>val).cq
 *         elif type(val) is CQEX:
*/
  __pyx_t_1 = (((PyObject *)Py_TYPE(__pyx_v_val)) == ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQ));
  if (__pyx_t_1) {

    /* "pyverbs/qp.pyx":150
 *     def send_cq(self, val):
 *         if type(val) is CQ:
 *             self.attr.send_cq = (<CQ>val).cq             # <<<<<<<<<<<<<<
 *         elif type(val) is CQEX:
 *             self.attr.send_cq = (<CQEX>val).ibv_cq
*/
    __pyx_t_2 = ((struct __pyx_obj_7pyverbs_2cq_CQ *)__pyx_v_val)->cq;
    __pyx_v_self->attr.send_cq = __pyx_t_2;

    /* "pyverbs/qp.pyx":149
 *     @send_cq.setter
 *     def send_cq(self, val):
 *         if type(val) is CQ:             # <<<<<<<<<<<<<<
 *             self.attr.send_cq = (<CQ>val).cq
 *         elif type(val) is CQEX:
*/
    goto __pyx_L3;
  }

  /* "pyverbs/qp.pyx":151
 *         if type(val) is CQ:
 *             self.attr.send_cq = (<CQ>val).cq
 *         elif type(val) is CQEX:             # <<<<<<<<<<<<<<
 *             self.attr.send_cq = (<CQEX>val).ibv_cq
 *         self.scq = val
*/
  __pyx_t_1 = (((PyObject *)Py_TYPE(__pyx_v_val)) == ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX));
  if (__pyx_t_1) {

    /* "pyverbs/qp.pyx":152
 *             self.attr.send_cq = (<CQ>val).cq
 *         elif type(val) is CQEX:
 *             self.attr.send_cq = (<CQEX>val).ibv_cq             # <<<<<<<<<<<<<<
 *         self.scq = val
 * 
*/
    __pyx_t_2 = ((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_val)->ibv_cq;
    __pyx_v_self->attr.send_cq = __pyx_t_2;

    /* "pyverbs/qp.pyx":151
 *         if type(val) is CQ:
 *             self.attr.send_cq = (<CQ>val).cq
 *         elif type(val) is CQEX:             # <<<<<<<<<<<<<<
 *             self.attr.send_cq = (<CQEX>val).ibv_cq
 *         self.scq = val
*/
  }
  __pyx_L3:;

  /* "pyverbs/qp.pyx":153
 *         elif type(val) is CQEX:
 *             self.attr.send_cq = (<CQEX>val).ibv_cq
 *         self.scq = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_INCREF(__pyx_v_val);
  __Pyx_GIVEREF(__pyx_v_val);
  __Pyx_GOTREF(__pyx_v_self->scq);
  __Pyx_DECREF(__pyx_v_self->scq);
  __pyx_v_self->scq = __pyx_v_val;

  /* "pyverbs/qp.pyx":147
 *     def send_cq(self):
 *         return self.scq
 *     @send_cq.setter             # <<<<<<<<<<<<<<
 *     def send_cq(self, val):
 *         if type(val) is CQ:
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":155
 *         self.scq = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def srq(self):
 *         return self.srq
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_3srq_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_3srq_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_3srq___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_3srq___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":157
 *     @property
 *     def srq(self):
 *         return self.srq             # <<<<<<<<<<<<<<
 *     @srq.setter
 *     def srq(self, SRQ val):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->srq);
  __pyx_r = __pyx_v_self->srq;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":155
 *         self.scq = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def srq(self):
 *         return self.srq
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":158
 *     def srq(self):
 *         return self.srq
 *     @srq.setter             # <<<<<<<<<<<<<<
 *     def srq(self, SRQ val):
 *         self.attr.srq = <v.ibv_srq*>val.srq
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_3srq_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_3srq_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_val), __pyx_mstate_global->__pyx_ptype_7pyverbs_3srq_SRQ, 1, "val", 0))) __PYX_ERR(0, 159, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_3srq_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self), ((struct __pyx_obj_7pyverbs_3srq_SRQ *)__pyx_v_val));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  goto __pyx_L5_cleaned_up;
  __pyx_L0:;
  __pyx_L5_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_10QPInitAttr_3srq_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, struct __pyx_obj_7pyverbs_3srq_SRQ *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "pyverbs/qp.pyx":160
 *     @srq.setter
 *     def srq(self, SRQ val):
 *         self.attr.srq = <v.ibv_srq*>val.srq             # <<<<<<<<<<<<<<
 *         self.srq = val
 * 
*/
  __pyx_v_self->attr.srq = ((struct ibv_srq *)__pyx_v_val->srq);

  /* "pyverbs/qp.pyx":161
 *     def srq(self, SRQ val):
 *         self.attr.srq = <v.ibv_srq*>val.srq
 *         self.srq = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_INCREF((PyObject *)__pyx_v_val);
  __Pyx_GIVEREF((PyObject *)__pyx_v_val);
  __Pyx_GOTREF(__pyx_v_self->srq);
  __Pyx_DECREF(__pyx_v_self->srq);
  __pyx_v_self->srq = ((PyObject *)__pyx_v_val);

  /* "pyverbs/qp.pyx":158
 *     def srq(self):
 *         return self.srq
 *     @srq.setter             # <<<<<<<<<<<<<<
 *     def srq(self, SRQ val):
 *         self.attr.srq = <v.ibv_srq*>val.srq
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":163
 *         self.srq = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def recv_cq(self):
 *         return self.rcq
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_7recv_cq_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_7recv_cq_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_7recv_cq___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_7recv_cq___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":165
 *     @property
 *     def recv_cq(self):
 *         return self.rcq             # <<<<<<<<<<<<<<
 *     @recv_cq.setter
 *     def recv_cq(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->rcq);
  __pyx_r = __pyx_v_self->rcq;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":163
 *         self.srq = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def recv_cq(self):
 *         return self.rcq
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":166
 *     def recv_cq(self):
 *         return self.rcq
 *     @recv_cq.setter             # <<<<<<<<<<<<<<
 *     def recv_cq(self, val):
 *         if type(val) is CQ:
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_7recv_cq_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_7recv_cq_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_7recv_cq_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_10QPInitAttr_7recv_cq_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  struct ibv_cq *__pyx_t_2;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "pyverbs/qp.pyx":168
 *     @recv_cq.setter
 *     def recv_cq(self, val):
 *         if type(val) is CQ:             # <<<<<<<<<<<<<<
 *             self.attr.recv_cq = (<CQ>val).cq
 *         elif type(val) is CQEX:
*/
  __pyx_t_1 = (((PyObject *)Py_TYPE(__pyx_v_val)) == ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQ));
  if (__pyx_t_1) {

    /* "pyverbs/qp.pyx":169
 *     def recv_cq(self, val):
 *         if type(val) is CQ:
 *             self.attr.recv_cq = (<CQ>val).cq             # <<<<<<<<<<<<<<
 *         elif type(val) is CQEX:
 *             self.attr.recv_cq = (<CQEX>val).ibv_cq
*/
    __pyx_t_2 = ((struct __pyx_obj_7pyverbs_2cq_CQ *)__pyx_v_val)->cq;
    __pyx_v_self->attr.recv_cq = __pyx_t_2;

    /* "pyverbs/qp.pyx":168
 *     @recv_cq.setter
 *     def recv_cq(self, val):
 *         if type(val) is CQ:             # <<<<<<<<<<<<<<
 *             self.attr.recv_cq = (<CQ>val).cq
 *         elif type(val) is CQEX:
*/
    goto __pyx_L3;
  }

  /* "pyverbs/qp.pyx":170
 *         if type(val) is CQ:
 *             self.attr.recv_cq = (<CQ>val).cq
 *         elif type(val) is CQEX:             # <<<<<<<<<<<<<<
 *             self.attr.recv_cq = (<CQEX>val).ibv_cq
 *         self.rcq = val
*/
  __pyx_t_1 = (((PyObject *)Py_TYPE(__pyx_v_val)) == ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX));
  if (__pyx_t_1) {

    /* "pyverbs/qp.pyx":171
 *             self.attr.recv_cq = (<CQ>val).cq
 *         elif type(val) is CQEX:
 *             self.attr.recv_cq = (<CQEX>val).ibv_cq             # <<<<<<<<<<<<<<
 *         self.rcq = val
 * 
*/
    __pyx_t_2 = ((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_val)->ibv_cq;
    __pyx_v_self->attr.recv_cq = __pyx_t_2;

    /* "pyverbs/qp.pyx":170
 *         if type(val) is CQ:
 *             self.attr.recv_cq = (<CQ>val).cq
 *         elif type(val) is CQEX:             # <<<<<<<<<<<<<<
 *             self.attr.recv_cq = (<CQEX>val).ibv_cq
 *         self.rcq = val
*/
  }
  __pyx_L3:;

  /* "pyverbs/qp.pyx":172
 *         elif type(val) is CQEX:
 *             self.attr.recv_cq = (<CQEX>val).ibv_cq
 *         self.rcq = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_INCREF(__pyx_v_val);
  __Pyx_GIVEREF(__pyx_v_val);
  __Pyx_GOTREF(__pyx_v_self->rcq);
  __Pyx_DECREF(__pyx_v_self->rcq);
  __pyx_v_self->rcq = __pyx_v_val;

  /* "pyverbs/qp.pyx":166
 *     def recv_cq(self):
 *         return self.rcq
 *     @recv_cq.setter             # <<<<<<<<<<<<<<
 *     def recv_cq(self, val):
 *         if type(val) is CQ:
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":174
 *         self.rcq = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def cap(self):
 *         return QPCap(max_send_wr=self.attr.cap.max_send_wr,
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_3cap_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_3cap_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_3cap___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_3cap___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":176
 *     @property
 *     def cap(self):
 *         return QPCap(max_send_wr=self.attr.cap.max_send_wr,             # <<<<<<<<<<<<<<
 *                      max_recv_wr=self.attr.cap.max_recv_wr,
 *                      max_send_sge=self.attr.cap.max_send_sge,
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = NULL;
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_send_wr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "pyverbs/qp.pyx":177
 *     def cap(self):
 *         return QPCap(max_send_wr=self.attr.cap.max_send_wr,
 *                      max_recv_wr=self.attr.cap.max_recv_wr,             # <<<<<<<<<<<<<<
 *                      max_send_sge=self.attr.cap.max_send_sge,
 *                      max_recv_sge=self.attr.cap.max_recv_sge,
*/
  __pyx_t_4 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_recv_wr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "pyverbs/qp.pyx":178
 *         return QPCap(max_send_wr=self.attr.cap.max_send_wr,
 *                      max_recv_wr=self.attr.cap.max_recv_wr,
 *                      max_send_sge=self.attr.cap.max_send_sge,             # <<<<<<<<<<<<<<
 *                      max_recv_sge=self.attr.cap.max_recv_sge,
 *                      max_inline_data=self.attr.cap.max_inline_data)
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_send_sge); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "pyverbs/qp.pyx":179
 *                      max_recv_wr=self.attr.cap.max_recv_wr,
 *                      max_send_sge=self.attr.cap.max_send_sge,
 *                      max_recv_sge=self.attr.cap.max_recv_sge,             # <<<<<<<<<<<<<<
 *                      max_inline_data=self.attr.cap.max_inline_data)
 *     @cap.setter
*/
  __pyx_t_6 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_recv_sge); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "pyverbs/qp.pyx":180
 *                      max_send_sge=self.attr.cap.max_send_sge,
 *                      max_recv_sge=self.attr.cap.max_recv_sge,
 *                      max_inline_data=self.attr.cap.max_inline_data)             # <<<<<<<<<<<<<<
 *     @cap.setter
 *     def cap(self, val):
*/
  __pyx_t_7 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_inline_data); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = 1;
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 5 : 0)] = {__pyx_t_2, NULL};
    __pyx_t_9 = __Pyx_MakeVectorcallBuilderKwds(5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_max_send_wr, __pyx_t_3, __pyx_t_9, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 176, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_max_recv_wr, __pyx_t_4, __pyx_t_9, __pyx_callargs+1, 1) < (0)) __PYX_ERR(0, 176, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_max_send_sge, __pyx_t_5, __pyx_t_9, __pyx_callargs+1, 2) < (0)) __PYX_ERR(0, 176, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_max_recv_sge, __pyx_t_6, __pyx_t_9, __pyx_callargs+1, 3) < (0)) __PYX_ERR(0, 176, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_max_inline_data, __pyx_t_7, __pyx_t_9, __pyx_callargs+1, 4) < (0)) __PYX_ERR(0, 176, __pyx_L1_error)
    __pyx_t_1 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPCap, __pyx_callargs+__pyx_t_8, (1-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 176, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_r = ((PyObject *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":174
 *         self.rcq = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def cap(self):
 *         return QPCap(max_send_wr=self.attr.cap.max_send_wr,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttr.cap.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":181
 *                      max_recv_sge=self.attr.cap.max_recv_sge,
 *                      max_inline_data=self.attr.cap.max_inline_data)
 *     @cap.setter             # <<<<<<<<<<<<<<
 *     def cap(self, val):
 *         _copy_caps(val, self)
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_3cap_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_3cap_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_3cap_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_10QPInitAttr_3cap_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "pyverbs/qp.pyx":183
 *     @cap.setter
 *     def cap(self, val):
 *         _copy_caps(val, self)             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_copy_caps); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
    __pyx_t_4 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_v_val, ((PyObject *)__pyx_v_self)};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_4, (3-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":181
 *                      max_recv_sge=self.attr.cap.max_recv_sge,
 *                      max_inline_data=self.attr.cap.max_inline_data)
 *     @cap.setter             # <<<<<<<<<<<<<<
 *     def cap(self, val):
 *         _copy_caps(val, self)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttr.cap.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":185
 *         _copy_caps(val, self)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qp_type(self):
 *         return self.attr.qp_type
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_7qp_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_7qp_type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_7qp_type___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_7qp_type___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":187
 *     @property
 *     def qp_type(self):
 *         return self.attr.qp_type             # <<<<<<<<<<<<<<
 *     @qp_type.setter
 *     def qp_type(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(__pyx_v_self->attr.qp_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":185
 *         _copy_caps(val, self)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qp_type(self):
 *         return self.attr.qp_type
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttr.qp_type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":188
 *     def qp_type(self):
 *         return self.attr.qp_type
 *     @qp_type.setter             # <<<<<<<<<<<<<<
 *     def qp_type(self, val):
 *         self.attr.qp_type = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_7qp_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_7qp_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_7qp_type_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_10QPInitAttr_7qp_type_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  enum ibv_qp_type __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":190
 *     @qp_type.setter
 *     def qp_type(self, val):
 *         self.attr.qp_type = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = ((enum ibv_qp_type)__Pyx_PyLong_As_enum__ibv_qp_type(__pyx_v_val)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 190, __pyx_L1_error)
  __pyx_v_self->attr.qp_type = __pyx_t_1;

  /* "pyverbs/qp.pyx":188
 *     def qp_type(self):
 *         return self.attr.qp_type
 *     @qp_type.setter             # <<<<<<<<<<<<<<
 *     def qp_type(self, val):
 *         self.attr.qp_type = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttr.qp_type.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":192
 *         self.attr.qp_type = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def sq_sig_all(self):
 *         return self.attr.sq_sig_all
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_10sq_sig_all_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_10sq_sig_all_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_10sq_sig_all___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_10sq_sig_all___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":194
 *     @property
 *     def sq_sig_all(self):
 *         return self.attr.sq_sig_all             # <<<<<<<<<<<<<<
 *     @sq_sig_all.setter
 *     def sq_sig_all(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_self->attr.sq_sig_all); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":192
 *         self.attr.qp_type = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def sq_sig_all(self):
 *         return self.attr.sq_sig_all
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttr.sq_sig_all.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":195
 *     def sq_sig_all(self):
 *         return self.attr.sq_sig_all
 *     @sq_sig_all.setter             # <<<<<<<<<<<<<<
 *     def sq_sig_all(self, val):
 *         self.attr.sq_sig_all = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_10sq_sig_all_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_10sq_sig_all_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_10sq_sig_all_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_10QPInitAttr_10sq_sig_all_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":197
 *     @sq_sig_all.setter
 *     def sq_sig_all(self, val):
 *         self.attr.sq_sig_all = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 197, __pyx_L1_error)
  __pyx_v_self->attr.sq_sig_all = __pyx_t_1;

  /* "pyverbs/qp.pyx":195
 *     def sq_sig_all(self):
 *         return self.attr.sq_sig_all
 *     @sq_sig_all.setter             # <<<<<<<<<<<<<<
 *     def sq_sig_all(self, val):
 *         self.attr.sq_sig_all = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttr.sq_sig_all.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":199
 *         self.attr.sq_sig_all = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_send_wr(self):
 *         return self.attr.cap.max_send_wr
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_11max_send_wr_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_11max_send_wr_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_11max_send_wr___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_11max_send_wr___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":201
 *     @property
 *     def max_send_wr(self):
 *         return self.attr.cap.max_send_wr             # <<<<<<<<<<<<<<
 *     @max_send_wr.setter
 *     def max_send_wr(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_send_wr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":199
 *         self.attr.sq_sig_all = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_send_wr(self):
 *         return self.attr.cap.max_send_wr
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttr.max_send_wr.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":202
 *     def max_send_wr(self):
 *         return self.attr.cap.max_send_wr
 *     @max_send_wr.setter             # <<<<<<<<<<<<<<
 *     def max_send_wr(self, val):
 *         self.attr.cap.max_send_wr = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_11max_send_wr_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_11max_send_wr_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_11max_send_wr_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_10QPInitAttr_11max_send_wr_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":204
 *     @max_send_wr.setter
 *     def max_send_wr(self, val):
 *         self.attr.cap.max_send_wr = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 204, __pyx_L1_error)
  __pyx_v_self->attr.cap.max_send_wr = __pyx_t_1;

  /* "pyverbs/qp.pyx":202
 *     def max_send_wr(self):
 *         return self.attr.cap.max_send_wr
 *     @max_send_wr.setter             # <<<<<<<<<<<<<<
 *     def max_send_wr(self, val):
 *         self.attr.cap.max_send_wr = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttr.max_send_wr.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":206
 *         self.attr.cap.max_send_wr = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_recv_wr(self):
 *         return self.attr.cap.max_recv_wr
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_11max_recv_wr_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_11max_recv_wr_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_11max_recv_wr___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_11max_recv_wr___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":208
 *     @property
 *     def max_recv_wr(self):
 *         return self.attr.cap.max_recv_wr             # <<<<<<<<<<<<<<
 *     @max_recv_wr.setter
 *     def max_recv_wr(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_recv_wr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":206
 *         self.attr.cap.max_send_wr = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_recv_wr(self):
 *         return self.attr.cap.max_recv_wr
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttr.max_recv_wr.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":209
 *     def max_recv_wr(self):
 *         return self.attr.cap.max_recv_wr
 *     @max_recv_wr.setter             # <<<<<<<<<<<<<<
 *     def max_recv_wr(self, val):
 *         self.attr.cap.max_recv_wr = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_11max_recv_wr_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_11max_recv_wr_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_11max_recv_wr_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_10QPInitAttr_11max_recv_wr_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":211
 *     @max_recv_wr.setter
 *     def max_recv_wr(self, val):
 *         self.attr.cap.max_recv_wr = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 211, __pyx_L1_error)
  __pyx_v_self->attr.cap.max_recv_wr = __pyx_t_1;

  /* "pyverbs/qp.pyx":209
 *     def max_recv_wr(self):
 *         return self.attr.cap.max_recv_wr
 *     @max_recv_wr.setter             # <<<<<<<<<<<<<<
 *     def max_recv_wr(self, val):
 *         self.attr.cap.max_recv_wr = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttr.max_recv_wr.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":213
 *         self.attr.cap.max_recv_wr = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_send_sge(self):
 *         return self.attr.cap.max_send_sge
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_12max_send_sge_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_12max_send_sge_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_12max_send_sge___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_12max_send_sge___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":215
 *     @property
 *     def max_send_sge(self):
 *         return self.attr.cap.max_send_sge             # <<<<<<<<<<<<<<
 *     @max_send_sge.setter
 *     def max_send_sge(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_send_sge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":213
 *         self.attr.cap.max_recv_wr = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_send_sge(self):
 *         return self.attr.cap.max_send_sge
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttr.max_send_sge.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":216
 *     def max_send_sge(self):
 *         return self.attr.cap.max_send_sge
 *     @max_send_sge.setter             # <<<<<<<<<<<<<<
 *     def max_send_sge(self, val):
 *         self.attr.cap.max_send_sge = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_12max_send_sge_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_12max_send_sge_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_12max_send_sge_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_10QPInitAttr_12max_send_sge_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":218
 *     @max_send_sge.setter
 *     def max_send_sge(self, val):
 *         self.attr.cap.max_send_sge = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 218, __pyx_L1_error)
  __pyx_v_self->attr.cap.max_send_sge = __pyx_t_1;

  /* "pyverbs/qp.pyx":216
 *     def max_send_sge(self):
 *         return self.attr.cap.max_send_sge
 *     @max_send_sge.setter             # <<<<<<<<<<<<<<
 *     def max_send_sge(self, val):
 *         self.attr.cap.max_send_sge = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttr.max_send_sge.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":220
 *         self.attr.cap.max_send_sge = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_recv_sge(self):
 *         return self.attr.cap.max_recv_sge
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_12max_recv_sge_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_12max_recv_sge_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_12max_recv_sge___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_12max_recv_sge___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":222
 *     @property
 *     def max_recv_sge(self):
 *         return self.attr.cap.max_recv_sge             # <<<<<<<<<<<<<<
 *     @max_recv_sge.setter
 *     def max_recv_sge(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_recv_sge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":220
 *         self.attr.cap.max_send_sge = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_recv_sge(self):
 *         return self.attr.cap.max_recv_sge
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttr.max_recv_sge.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":223
 *     def max_recv_sge(self):
 *         return self.attr.cap.max_recv_sge
 *     @max_recv_sge.setter             # <<<<<<<<<<<<<<
 *     def max_recv_sge(self, val):
 *         self.attr.cap.max_recv_sge = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_12max_recv_sge_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_12max_recv_sge_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_12max_recv_sge_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_10QPInitAttr_12max_recv_sge_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":225
 *     @max_recv_sge.setter
 *     def max_recv_sge(self, val):
 *         self.attr.cap.max_recv_sge = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 225, __pyx_L1_error)
  __pyx_v_self->attr.cap.max_recv_sge = __pyx_t_1;

  /* "pyverbs/qp.pyx":223
 *     def max_recv_sge(self):
 *         return self.attr.cap.max_recv_sge
 *     @max_recv_sge.setter             # <<<<<<<<<<<<<<
 *     def max_recv_sge(self, val):
 *         self.attr.cap.max_recv_sge = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttr.max_recv_sge.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":227
 *         self.attr.cap.max_recv_sge = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_inline_data(self):
 *         return self.attr.cap.max_inline_data
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_15max_inline_data_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_15max_inline_data_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_15max_inline_data___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_15max_inline_data___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":229
 *     @property
 *     def max_inline_data(self):
 *         return self.attr.cap.max_inline_data             # <<<<<<<<<<<<<<
 *     @max_inline_data.setter
 *     def max_inline_data(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_inline_data); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":227
 *         self.attr.cap.max_recv_sge = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_inline_data(self):
 *         return self.attr.cap.max_inline_data
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttr.max_inline_data.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":230
 *     def max_inline_data(self):
 *         return self.attr.cap.max_inline_data
 *     @max_inline_data.setter             # <<<<<<<<<<<<<<
 *     def max_inline_data(self, val):
 *         self.attr.cap.max_inline_data = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_15max_inline_data_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_10QPInitAttr_15max_inline_data_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_15max_inline_data_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_10QPInitAttr_15max_inline_data_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":232
 *     @max_inline_data.setter
 *     def max_inline_data(self, val):
 *         self.attr.cap.max_inline_data = val             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 232, __pyx_L1_error)
  __pyx_v_self->attr.cap.max_inline_data = __pyx_t_1;

  /* "pyverbs/qp.pyx":230
 *     def max_inline_data(self):
 *         return self.attr.cap.max_inline_data
 *     @max_inline_data.setter             # <<<<<<<<<<<<<<
 *     def max_inline_data(self, val):
 *         self.attr.cap.max_inline_data = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttr.max_inline_data.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":234
 *         self.attr.cap.max_inline_data = val
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         print_format = '{:20}: {:<20}\n'
 *         ident_format = '    {:20}: {:<20}\n'
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_3__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_3__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_2__str__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_2__str__(struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self) {
  PyObject *__pyx_v_print_format = NULL;
  PyObject *__pyx_v_ident_format = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "pyverbs/qp.pyx":235
 * 
 *     def __str__(self):
 *         print_format = '{:20}: {:<20}\n'             # <<<<<<<<<<<<<<
 *         ident_format = '    {:20}: {:<20}\n'
 *         return print_format.format('QP type', qp_type_to_str(self.qp_type)) +\
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_20_20);
  __pyx_v_print_format = __pyx_mstate_global->__pyx_kp_u_20_20;

  /* "pyverbs/qp.pyx":236
 *     def __str__(self):
 *         print_format = '{:20}: {:<20}\n'
 *         ident_format = '    {:20}: {:<20}\n'             # <<<<<<<<<<<<<<
 *         return print_format.format('QP type', qp_type_to_str(self.qp_type)) +\
 *                print_format.format('SQ sig. all', self.sq_sig_all) +\
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_20_20_2);
  __pyx_v_ident_format = __pyx_mstate_global->__pyx_kp_u_20_20_2;

  /* "pyverbs/qp.pyx":237
 *         print_format = '{:20}: {:<20}\n'
 *         ident_format = '    {:20}: {:<20}\n'
 *         return print_format.format('QP type', qp_type_to_str(self.qp_type)) +\             # <<<<<<<<<<<<<<
 *                print_format.format('SQ sig. all', self.sq_sig_all) +\
 *                'QP caps:\n' +\
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_qp_type_to_str); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_qp_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 237, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_QP_type, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":238
 *         ident_format = '    {:20}: {:<20}\n'
 *         return print_format.format('QP type', qp_type_to_str(self.qp_type)) +\
 *                print_format.format('SQ sig. all', self.sq_sig_all) +\             # <<<<<<<<<<<<<<
 *                'QP caps:\n' +\
 *                ident_format.format('max send WR', self.attr.cap.max_send_wr) +\
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_sq_sig_all); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_SQ_sig_all, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":237
 *         print_format = '{:20}: {:<20}\n'
 *         ident_format = '    {:20}: {:<20}\n'
 *         return print_format.format('QP type', qp_type_to_str(self.qp_type)) +\             # <<<<<<<<<<<<<<
 *                print_format.format('SQ sig. all', self.sq_sig_all) +\
 *                'QP caps:\n' +\
*/
  __pyx_t_1 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":238
 *         ident_format = '    {:20}: {:<20}\n'
 *         return print_format.format('QP type', qp_type_to_str(self.qp_type)) +\
 *                print_format.format('SQ sig. all', self.sq_sig_all) +\             # <<<<<<<<<<<<<<
 *                'QP caps:\n' +\
 *                ident_format.format('max send WR', self.attr.cap.max_send_wr) +\
*/
  __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_1, __pyx_mstate_global->__pyx_kp_u_QP_caps); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":240
 *                print_format.format('SQ sig. all', self.sq_sig_all) +\
 *                'QP caps:\n' +\
 *                ident_format.format('max send WR', self.attr.cap.max_send_wr) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('max recv WR', self.attr.cap.max_recv_wr) +\
 *                ident_format.format('max send SGE',
*/
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_send_wr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_kp_u_max_send_WR, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":239
 *         return print_format.format('QP type', qp_type_to_str(self.qp_type)) +\
 *                print_format.format('SQ sig. all', self.sq_sig_all) +\
 *                'QP caps:\n' +\             # <<<<<<<<<<<<<<
 *                ident_format.format('max send WR', self.attr.cap.max_send_wr) +\
 *                ident_format.format('max recv WR', self.attr.cap.max_recv_wr) +\
*/
  __pyx_t_1 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":241
 *                'QP caps:\n' +\
 *                ident_format.format('max send WR', self.attr.cap.max_send_wr) +\
 *                ident_format.format('max recv WR', self.attr.cap.max_recv_wr) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('max send SGE',
 *                                    self.attr.cap.max_send_sge) +\
*/
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_recv_wr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_kp_u_max_recv_WR, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":240
 *                print_format.format('SQ sig. all', self.sq_sig_all) +\
 *                'QP caps:\n' +\
 *                ident_format.format('max send WR', self.attr.cap.max_send_wr) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('max recv WR', self.attr.cap.max_recv_wr) +\
 *                ident_format.format('max send SGE',
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":243
 *                ident_format.format('max recv WR', self.attr.cap.max_recv_wr) +\
 *                ident_format.format('max send SGE',
 *                                    self.attr.cap.max_send_sge) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('max recv SGE',
 *                                    self.attr.cap.max_recv_sge) +\
*/
  __pyx_t_4 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_send_sge); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "pyverbs/qp.pyx":242
 *                ident_format.format('max send WR', self.attr.cap.max_send_wr) +\
 *                ident_format.format('max recv WR', self.attr.cap.max_recv_wr) +\
 *                ident_format.format('max send SGE',             # <<<<<<<<<<<<<<
 *                                    self.attr.cap.max_send_sge) +\
 *                ident_format.format('max recv SGE',
*/
  __pyx_t_1 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_kp_u_max_send_SGE, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":241
 *                'QP caps:\n' +\
 *                ident_format.format('max send WR', self.attr.cap.max_send_wr) +\
 *                ident_format.format('max recv WR', self.attr.cap.max_recv_wr) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('max send SGE',
 *                                    self.attr.cap.max_send_sge) +\
*/
  __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":245
 *                                    self.attr.cap.max_send_sge) +\
 *                ident_format.format('max recv SGE',
 *                                    self.attr.cap.max_recv_sge) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('max inline data',
 *                                    self.attr.cap.max_inline_data)
*/
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_recv_sge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pyverbs/qp.pyx":244
 *                ident_format.format('max send SGE',
 *                                    self.attr.cap.max_send_sge) +\
 *                ident_format.format('max recv SGE',             # <<<<<<<<<<<<<<
 *                                    self.attr.cap.max_recv_sge) +\
 *                ident_format.format('max inline data',
*/
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_kp_u_max_recv_SGE, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":243
 *                ident_format.format('max recv WR', self.attr.cap.max_recv_wr) +\
 *                ident_format.format('max send SGE',
 *                                    self.attr.cap.max_send_sge) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('max recv SGE',
 *                                    self.attr.cap.max_recv_sge) +\
*/
  __pyx_t_1 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":247
 *                                    self.attr.cap.max_recv_sge) +\
 *                ident_format.format('max inline data',
 *                                    self.attr.cap.max_inline_data)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_inline_data); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "pyverbs/qp.pyx":246
 *                ident_format.format('max recv SGE',
 *                                    self.attr.cap.max_recv_sge) +\
 *                ident_format.format('max inline data',             # <<<<<<<<<<<<<<
 *                                    self.attr.cap.max_inline_data)
 * 
*/
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_kp_u_max_inline_data_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":245
 *                                    self.attr.cap.max_send_sge) +\
 *                ident_format.format('max recv SGE',
 *                                    self.attr.cap.max_recv_sge) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('max inline data',
 *                                    self.attr.cap.max_inline_data)
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":234
 *         self.attr.cap.max_inline_data = val
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         print_format = '{:20}: {:<20}\n'
 *         ident_format = '    {:20}: {:<20}\n'
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttr.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_print_format);
  __Pyx_XDECREF(__pyx_v_ident_format);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_10QPInitAttr_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_10QPInitAttr_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_4__reduce_cython__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_attr_cannot_be_converted_to, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttr.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_10QPInitAttr_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_10QPInitAttr_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_10QPInitAttr_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttr.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_10QPInitAttr_6__setstate_cython__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_10QPInitAttr_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_attr_cannot_be_converted_to, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttr.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":251
 * 
 * cdef class QPInitAttrEx(PyverbsObject):
 *     def __init__(self, qp_type=e.IBV_QPT_UD, qp_context=None,             # <<<<<<<<<<<<<<
 *                  PyverbsObject scq=None, PyverbsObject rcq=None,
 *                  SRQ srq=None, QPCap cap=None, sq_sig_all=0, comp_mask=0,
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2qp_12QPInitAttrEx___init__, "\n        Initialize a QPInitAttrEx object with user-defined or default values.\n        :param qp_type: QP type to be created\n        :param qp_context: Associated user context\n        :param scq: Send CQ to be used for this QP\n        :param rcq: Recv CQ to be used for this QP\n        :param srq: Shared receive queue to be used as RQ in QP\n        :param cap: A QPCap object\n        :param sq_sig_all: If set, each send WR will generate a completion\n                           entry\n        :param comp_mask: bit mask to determine which of the following fields\n                          are valid\n        :param pd: A PD object to be associated with this QP\n        :param xrcd: XRC domain to be used for XRC QPs\n        :param create_flags: Creation flags for this QP\n        :param max_tso_header: Maximum TSO header size\n        :param source_qpn: Source QP number (requires IBV_QP_CREATE_SOURCE_QPN\n                           set in create_flags)\n        :param hash_conf: A RxHashConf object, config of RX hash key.\n        :param ind_table: A RwqIndTable object, indirection table of RWQs.\n        :param send_ops_flags: Send opcodes to be supported by the extended QP.\n                               Use ibv_qp_create_send_ops_flags enum\n        :return: An initialized QPInitAttrEx object\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_7pyverbs_2qp_12QPInitAttrEx___init__;
#endif
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_qp_type = 0;
  CYTHON_UNUSED PyObject *__pyx_v_qp_context = 0;
  struct __pyx_obj_7pyverbs_4base_PyverbsObject *__pyx_v_scq = 0;
  struct __pyx_obj_7pyverbs_4base_PyverbsObject *__pyx_v_rcq = 0;
  struct __pyx_obj_7pyverbs_3srq_SRQ *__pyx_v_srq = 0;
  struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_cap = 0;
  PyObject *__pyx_v_sq_sig_all = 0;
  PyObject *__pyx_v_comp_mask = 0;
  struct __pyx_obj_7pyverbs_2pd_PD *__pyx_v_pd = 0;
  struct __pyx_obj_7pyverbs_4xrcd_XRCD *__pyx_v_xrcd = 0;
  PyObject *__pyx_v_create_flags = 0;
  PyObject *__pyx_v_max_tso_header = 0;
  PyObject *__pyx_v_source_qpn = 0;
  struct __pyx_obj_7pyverbs_2wq_RxHashConf *__pyx_v_hash_conf = 0;
  struct __pyx_obj_7pyverbs_2wq_RwqIndTable *__pyx_v_ind_table = 0;
  PyObject *__pyx_v_send_ops_flags = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_qp_type,&__pyx_mstate_global->__pyx_n_u_qp_context,&__pyx_mstate_global->__pyx_n_u_scq,&__pyx_mstate_global->__pyx_n_u_rcq,&__pyx_mstate_global->__pyx_n_u_srq,&__pyx_mstate_global->__pyx_n_u_cap,&__pyx_mstate_global->__pyx_n_u_sq_sig_all,&__pyx_mstate_global->__pyx_n_u_comp_mask,&__pyx_mstate_global->__pyx_n_u_pd,&__pyx_mstate_global->__pyx_n_u_xrcd,&__pyx_mstate_global->__pyx_n_u_create_flags,&__pyx_mstate_global->__pyx_n_u_max_tso_header,&__pyx_mstate_global->__pyx_n_u_source_qpn,&__pyx_mstate_global->__pyx_n_u_hash_conf,&__pyx_mstate_global->__pyx_n_u_ind_table,&__pyx_mstate_global->__pyx_n_u_send_ops_flags,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 251, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case 16:
        values[15] = __Pyx_ArgRef_VARARGS(__pyx_args, 15);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[15])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 15:
        values[14] = __Pyx_ArgRef_VARARGS(__pyx_args, 14);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[14])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 14:
        values[13] = __Pyx_ArgRef_VARARGS(__pyx_args, 13);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[13])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 13:
        values[12] = __Pyx_ArgRef_VARARGS(__pyx_args, 12);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[12])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 12:
        values[11] = __Pyx_ArgRef_VARARGS(__pyx_args, 11);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[11])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 11:
        values[10] = __Pyx_ArgRef_VARARGS(__pyx_args, 10);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[10])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 10:
        values[9] = __Pyx_ArgRef_VARARGS(__pyx_args, 9);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[9])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  9:
        values[8] = __Pyx_ArgRef_VARARGS(__pyx_args, 8);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  8:
        values[7] = __Pyx_ArgRef_VARARGS(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_VARARGS(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 251, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__3);
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "pyverbs/qp.pyx":252
 * cdef class QPInitAttrEx(PyverbsObject):
 *     def __init__(self, qp_type=e.IBV_QPT_UD, qp_context=None,
 *                  PyverbsObject scq=None, PyverbsObject rcq=None,             # <<<<<<<<<<<<<<
 *                  SRQ srq=None, QPCap cap=None, sq_sig_all=0, comp_mask=0,
 *                  PD pd=None, XRCD xrcd=None, create_flags=0,
*/
      if (!values[2]) values[2] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_4base_PyverbsObject *)Py_None));
      if (!values[3]) values[3] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_4base_PyverbsObject *)Py_None));

      /* "pyverbs/qp.pyx":253
 *     def __init__(self, qp_type=e.IBV_QPT_UD, qp_context=None,
 *                  PyverbsObject scq=None, PyverbsObject rcq=None,
 *                  SRQ srq=None, QPCap cap=None, sq_sig_all=0, comp_mask=0,             # <<<<<<<<<<<<<<
 *                  PD pd=None, XRCD xrcd=None, create_flags=0,
 *                  max_tso_header=0, source_qpn=0, RxHashConf hash_conf=None,
*/
      if (!values[4]) values[4] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_3srq_SRQ *)Py_None));
      if (!values[5]) values[5] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2qp_QPCap *)Py_None));
      if (!values[6]) values[6] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[7]) values[7] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));

      /* "pyverbs/qp.pyx":254
 *                  PyverbsObject scq=None, PyverbsObject rcq=None,
 *                  SRQ srq=None, QPCap cap=None, sq_sig_all=0, comp_mask=0,
 *                  PD pd=None, XRCD xrcd=None, create_flags=0,             # <<<<<<<<<<<<<<
 *                  max_tso_header=0, source_qpn=0, RxHashConf hash_conf=None,
 *                  RwqIndTable ind_table=None, send_ops_flags=0):
*/
      if (!values[8]) values[8] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2pd_PD *)Py_None));
      if (!values[9]) values[9] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_4xrcd_XRCD *)Py_None));
      if (!values[10]) values[10] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[11]) values[11] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[12]) values[12] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));

      /* "pyverbs/qp.pyx":255
 *                  SRQ srq=None, QPCap cap=None, sq_sig_all=0, comp_mask=0,
 *                  PD pd=None, XRCD xrcd=None, create_flags=0,
 *                  max_tso_header=0, source_qpn=0, RxHashConf hash_conf=None,             # <<<<<<<<<<<<<<
 *                  RwqIndTable ind_table=None, send_ops_flags=0):
 *         """
*/
      if (!values[13]) values[13] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2wq_RxHashConf *)Py_None));

      /* "pyverbs/qp.pyx":256
 *                  PD pd=None, XRCD xrcd=None, create_flags=0,
 *                  max_tso_header=0, source_qpn=0, RxHashConf hash_conf=None,
 *                  RwqIndTable ind_table=None, send_ops_flags=0):             # <<<<<<<<<<<<<<
 *         """
 *         Initialize a QPInitAttrEx object with user-defined or default values.
*/
      if (!values[14]) values[14] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2wq_RwqIndTable *)Py_None));
      if (!values[15]) values[15] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    } else {
      switch (__pyx_nargs) {
        case 16:
        values[15] = __Pyx_ArgRef_VARARGS(__pyx_args, 15);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[15])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 15:
        values[14] = __Pyx_ArgRef_VARARGS(__pyx_args, 14);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[14])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 14:
        values[13] = __Pyx_ArgRef_VARARGS(__pyx_args, 13);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[13])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 13:
        values[12] = __Pyx_ArgRef_VARARGS(__pyx_args, 12);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[12])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 12:
        values[11] = __Pyx_ArgRef_VARARGS(__pyx_args, 11);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[11])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 11:
        values[10] = __Pyx_ArgRef_VARARGS(__pyx_args, 10);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[10])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 10:
        values[9] = __Pyx_ArgRef_VARARGS(__pyx_args, 9);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[9])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  9:
        values[8] = __Pyx_ArgRef_VARARGS(__pyx_args, 8);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  8:
        values[7] = __Pyx_ArgRef_VARARGS(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_VARARGS(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__3);

      /* "pyverbs/qp.pyx":251
 * 
 * cdef class QPInitAttrEx(PyverbsObject):
 *     def __init__(self, qp_type=e.IBV_QPT_UD, qp_context=None,             # <<<<<<<<<<<<<<
 *                  PyverbsObject scq=None, PyverbsObject rcq=None,
 *                  SRQ srq=None, QPCap cap=None, sq_sig_all=0, comp_mask=0,
*/
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "pyverbs/qp.pyx":252
 * cdef class QPInitAttrEx(PyverbsObject):
 *     def __init__(self, qp_type=e.IBV_QPT_UD, qp_context=None,
 *                  PyverbsObject scq=None, PyverbsObject rcq=None,             # <<<<<<<<<<<<<<
 *                  SRQ srq=None, QPCap cap=None, sq_sig_all=0, comp_mask=0,
 *                  PD pd=None, XRCD xrcd=None, create_flags=0,
*/
      if (!values[2]) values[2] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_4base_PyverbsObject *)Py_None));
      if (!values[3]) values[3] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_4base_PyverbsObject *)Py_None));

      /* "pyverbs/qp.pyx":253
 *     def __init__(self, qp_type=e.IBV_QPT_UD, qp_context=None,
 *                  PyverbsObject scq=None, PyverbsObject rcq=None,
 *                  SRQ srq=None, QPCap cap=None, sq_sig_all=0, comp_mask=0,             # <<<<<<<<<<<<<<
 *                  PD pd=None, XRCD xrcd=None, create_flags=0,
 *                  max_tso_header=0, source_qpn=0, RxHashConf hash_conf=None,
*/
      if (!values[4]) values[4] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_3srq_SRQ *)Py_None));
      if (!values[5]) values[5] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2qp_QPCap *)Py_None));
      if (!values[6]) values[6] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[7]) values[7] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));

      /* "pyverbs/qp.pyx":254
 *                  PyverbsObject scq=None, PyverbsObject rcq=None,
 *                  SRQ srq=None, QPCap cap=None, sq_sig_all=0, comp_mask=0,
 *                  PD pd=None, XRCD xrcd=None, create_flags=0,             # <<<<<<<<<<<<<<
 *                  max_tso_header=0, source_qpn=0, RxHashConf hash_conf=None,
 *                  RwqIndTable ind_table=None, send_ops_flags=0):
*/
      if (!values[8]) values[8] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2pd_PD *)Py_None));
      if (!values[9]) values[9] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_4xrcd_XRCD *)Py_None));
      if (!values[10]) values[10] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[11]) values[11] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[12]) values[12] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));

      /* "pyverbs/qp.pyx":255
 *                  SRQ srq=None, QPCap cap=None, sq_sig_all=0, comp_mask=0,
 *                  PD pd=None, XRCD xrcd=None, create_flags=0,
 *                  max_tso_header=0, source_qpn=0, RxHashConf hash_conf=None,             # <<<<<<<<<<<<<<
 *                  RwqIndTable ind_table=None, send_ops_flags=0):
 *         """
*/
      if (!values[13]) values[13] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2wq_RxHashConf *)Py_None));

      /* "pyverbs/qp.pyx":256
 *                  PD pd=None, XRCD xrcd=None, create_flags=0,
 *                  max_tso_header=0, source_qpn=0, RxHashConf hash_conf=None,
 *                  RwqIndTable ind_table=None, send_ops_flags=0):             # <<<<<<<<<<<<<<
 *         """
 *         Initialize a QPInitAttrEx object with user-defined or default values.
*/
      if (!values[14]) values[14] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2wq_RwqIndTable *)Py_None));
      if (!values[15]) values[15] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    }
    __pyx_v_qp_type = values[0];
    __pyx_v_qp_context = values[1];
    __pyx_v_scq = ((struct __pyx_obj_7pyverbs_4base_PyverbsObject *)values[2]);
    __pyx_v_rcq = ((struct __pyx_obj_7pyverbs_4base_PyverbsObject *)values[3]);
    __pyx_v_srq = ((struct __pyx_obj_7pyverbs_3srq_SRQ *)values[4]);
    __pyx_v_cap = ((struct __pyx_obj_7pyverbs_2qp_QPCap *)values[5]);
    __pyx_v_sq_sig_all = values[6];
    __pyx_v_comp_mask = values[7];
    __pyx_v_pd = ((struct __pyx_obj_7pyverbs_2pd_PD *)values[8]);
    __pyx_v_xrcd = ((struct __pyx_obj_7pyverbs_4xrcd_XRCD *)values[9]);
    __pyx_v_create_flags = values[10];
    __pyx_v_max_tso_header = values[11];
    __pyx_v_source_qpn = values[12];
    __pyx_v_hash_conf = ((struct __pyx_obj_7pyverbs_2wq_RxHashConf *)values[13]);
    __pyx_v_ind_table = ((struct __pyx_obj_7pyverbs_2wq_RwqIndTable *)values[14]);
    __pyx_v_send_ops_flags = values[15];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 16, __pyx_nargs); __PYX_ERR(0, 251, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_scq), __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, 1, "scq", 0))) __PYX_ERR(0, 252, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_rcq), __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, 1, "rcq", 0))) __PYX_ERR(0, 252, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_srq), __pyx_mstate_global->__pyx_ptype_7pyverbs_3srq_SRQ, 1, "srq", 0))) __PYX_ERR(0, 253, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_cap), __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPCap, 1, "cap", 0))) __PYX_ERR(0, 253, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_pd), __pyx_mstate_global->__pyx_ptype_7pyverbs_2pd_PD, 1, "pd", 0))) __PYX_ERR(0, 254, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_xrcd), __pyx_mstate_global->__pyx_ptype_7pyverbs_4xrcd_XRCD, 1, "xrcd", 0))) __PYX_ERR(0, 254, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hash_conf), __pyx_mstate_global->__pyx_ptype_7pyverbs_2wq_RxHashConf, 1, "hash_conf", 0))) __PYX_ERR(0, 255, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_ind_table), __pyx_mstate_global->__pyx_ptype_7pyverbs_2wq_RwqIndTable, 1, "ind_table", 0))) __PYX_ERR(0, 256, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx___init__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self), __pyx_v_qp_type, __pyx_v_qp_context, __pyx_v_scq, __pyx_v_rcq, __pyx_v_srq, __pyx_v_cap, __pyx_v_sq_sig_all, __pyx_v_comp_mask, __pyx_v_pd, __pyx_v_xrcd, __pyx_v_create_flags, __pyx_v_max_tso_header, __pyx_v_source_qpn, __pyx_v_hash_conf, __pyx_v_ind_table, __pyx_v_send_ops_flags);

  /* "pyverbs/qp.pyx":251
 * 
 * cdef class QPInitAttrEx(PyverbsObject):
 *     def __init__(self, qp_type=e.IBV_QPT_UD, qp_context=None,             # <<<<<<<<<<<<<<
 *                  PyverbsObject scq=None, PyverbsObject rcq=None,
 *                  SRQ srq=None, QPCap cap=None, sq_sig_all=0, comp_mask=0,
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx___init__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_qp_type, CYTHON_UNUSED PyObject *__pyx_v_qp_context, struct __pyx_obj_7pyverbs_4base_PyverbsObject *__pyx_v_scq, struct __pyx_obj_7pyverbs_4base_PyverbsObject *__pyx_v_rcq, struct __pyx_obj_7pyverbs_3srq_SRQ *__pyx_v_srq, struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_cap, PyObject *__pyx_v_sq_sig_all, PyObject *__pyx_v_comp_mask, struct __pyx_obj_7pyverbs_2pd_PD *__pyx_v_pd, struct __pyx_obj_7pyverbs_4xrcd_XRCD *__pyx_v_xrcd, PyObject *__pyx_v_create_flags, PyObject *__pyx_v_max_tso_header, PyObject *__pyx_v_source_qpn, struct __pyx_obj_7pyverbs_2wq_RxHashConf *__pyx_v_hash_conf, struct __pyx_obj_7pyverbs_2wq_RwqIndTable *__pyx_v_ind_table, PyObject *__pyx_v_send_ops_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  struct ibv_cq *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  struct ibv_srq *__pyx_t_10;
  struct ibv_xrcd *__pyx_t_11;
  struct ibv_rx_hash_conf __pyx_t_12;
  struct ibv_rwq_ind_table *__pyx_t_13;
  enum ibv_qp_type __pyx_t_14;
  int __pyx_t_15;
  unsigned int __pyx_t_16;
  struct ibv_pd *__pyx_t_17;
  unsigned short __pyx_t_18;
  unsigned long __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyverbs/qp.pyx":281
 *         :return: An initialized QPInitAttrEx object
 *         """
 *         super().__init__()             # <<<<<<<<<<<<<<
 *         _copy_caps(cap, self)
 *         if scq is not None:
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":282
 *         """
 *         super().__init__()
 *         _copy_caps(cap, self)             # <<<<<<<<<<<<<<
 *         if scq is not None:
 *             if type(scq) is CQ:
*/
  __pyx_t_3 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_copy_caps); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    assert(__pyx_t_3);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_3, ((PyObject *)__pyx_v_cap), ((PyObject *)__pyx_v_self)};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":283
 *         super().__init__()
 *         _copy_caps(cap, self)
 *         if scq is not None:             # <<<<<<<<<<<<<<
 *             if type(scq) is CQ:
 *                 self.attr.send_cq = (<CQ>scq).cq
*/
  __pyx_t_6 = (((PyObject *)__pyx_v_scq) != Py_None);
  if (__pyx_t_6) {

    /* "pyverbs/qp.pyx":284
 *         _copy_caps(cap, self)
 *         if scq is not None:
 *             if type(scq) is CQ:             # <<<<<<<<<<<<<<
 *                 self.attr.send_cq = (<CQ>scq).cq
 *             elif isinstance(scq, CQEX):
*/
    __pyx_t_6 = (((PyObject *)Py_TYPE(((PyObject *)__pyx_v_scq))) == ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQ));
    if (__pyx_t_6) {

      /* "pyverbs/qp.pyx":285
 *         if scq is not None:
 *             if type(scq) is CQ:
 *                 self.attr.send_cq = (<CQ>scq).cq             # <<<<<<<<<<<<<<
 *             elif isinstance(scq, CQEX):
 *                 self.attr.send_cq = (<CQEX>scq).ibv_cq
*/
      __pyx_t_7 = ((struct __pyx_obj_7pyverbs_2cq_CQ *)__pyx_v_scq)->cq;
      __pyx_v_self->attr.send_cq = __pyx_t_7;

      /* "pyverbs/qp.pyx":284
 *         _copy_caps(cap, self)
 *         if scq is not None:
 *             if type(scq) is CQ:             # <<<<<<<<<<<<<<
 *                 self.attr.send_cq = (<CQ>scq).cq
 *             elif isinstance(scq, CQEX):
*/
      goto __pyx_L4;
    }

    /* "pyverbs/qp.pyx":286
 *             if type(scq) is CQ:
 *                 self.attr.send_cq = (<CQ>scq).cq
 *             elif isinstance(scq, CQEX):             # <<<<<<<<<<<<<<
 *                 self.attr.send_cq = (<CQEX>scq).ibv_cq
 *             else:
*/
    __pyx_t_6 = __Pyx_TypeCheck(((PyObject *)__pyx_v_scq), __pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX); 
    if (likely(__pyx_t_6)) {

      /* "pyverbs/qp.pyx":287
 *                 self.attr.send_cq = (<CQ>scq).cq
 *             elif isinstance(scq, CQEX):
 *                 self.attr.send_cq = (<CQEX>scq).ibv_cq             # <<<<<<<<<<<<<<
 *             else:
 *                 raise PyverbsUserError('Expected CQ/CQEX, got {t}'.\
*/
      __pyx_t_7 = ((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_scq)->ibv_cq;
      __pyx_v_self->attr.send_cq = __pyx_t_7;

      /* "pyverbs/qp.pyx":286
 *             if type(scq) is CQ:
 *                 self.attr.send_cq = (<CQ>scq).cq
 *             elif isinstance(scq, CQEX):             # <<<<<<<<<<<<<<
 *                 self.attr.send_cq = (<CQEX>scq).ibv_cq
 *             else:
*/
      goto __pyx_L4;
    }

    /* "pyverbs/qp.pyx":289
 *                 self.attr.send_cq = (<CQEX>scq).ibv_cq
 *             else:
 *                 raise PyverbsUserError('Expected CQ/CQEX, got {t}'.\             # <<<<<<<<<<<<<<
 *                                        format(t=type(scq)))
 *         self.scq = scq
*/
    /*else*/ {
      __pyx_t_2 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_PyverbsUserError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "pyverbs/qp.pyx":290
 *             else:
 *                 raise PyverbsUserError('Expected CQ/CQEX, got {t}'.\
 *                                        format(t=type(scq)))             # <<<<<<<<<<<<<<
 *         self.scq = scq
 * 
*/
      __pyx_t_8 = __pyx_mstate_global->__pyx_kp_u_Expected_CQ_CQEX_got_t;
      __Pyx_INCREF(__pyx_t_8);
      __pyx_t_5 = 0;
      {
        PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_8, NULL};
        __pyx_t_9 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 290, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_t, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_scq))), __pyx_t_9, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 290, __pyx_L1_error)
        __pyx_t_4 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_format, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_9);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 290, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
      }
      __pyx_t_5 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
        assert(__pyx_t_2);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
        __pyx_t_5 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_4};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 289, __pyx_L1_error)
    }
    __pyx_L4:;

    /* "pyverbs/qp.pyx":283
 *         super().__init__()
 *         _copy_caps(cap, self)
 *         if scq is not None:             # <<<<<<<<<<<<<<
 *             if type(scq) is CQ:
 *                 self.attr.send_cq = (<CQ>scq).cq
*/
  }

  /* "pyverbs/qp.pyx":291
 *                 raise PyverbsUserError('Expected CQ/CQEX, got {t}'.\
 *                                        format(t=type(scq)))
 *         self.scq = scq             # <<<<<<<<<<<<<<
 * 
 *         if rcq is not None:
*/
  __Pyx_INCREF((PyObject *)__pyx_v_scq);
  __Pyx_GIVEREF((PyObject *)__pyx_v_scq);
  __Pyx_GOTREF(__pyx_v_self->scq);
  __Pyx_DECREF(__pyx_v_self->scq);
  __pyx_v_self->scq = ((PyObject *)__pyx_v_scq);

  /* "pyverbs/qp.pyx":293
 *         self.scq = scq
 * 
 *         if rcq is not None:             # <<<<<<<<<<<<<<
 *             if type(rcq) is CQ:
 *                 self.attr.recv_cq = (<CQ>rcq).cq
*/
  __pyx_t_6 = (((PyObject *)__pyx_v_rcq) != Py_None);
  if (__pyx_t_6) {

    /* "pyverbs/qp.pyx":294
 * 
 *         if rcq is not None:
 *             if type(rcq) is CQ:             # <<<<<<<<<<<<<<
 *                 self.attr.recv_cq = (<CQ>rcq).cq
 *             elif isinstance(rcq, CQEX):
*/
    __pyx_t_6 = (((PyObject *)Py_TYPE(((PyObject *)__pyx_v_rcq))) == ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQ));
    if (__pyx_t_6) {

      /* "pyverbs/qp.pyx":295
 *         if rcq is not None:
 *             if type(rcq) is CQ:
 *                 self.attr.recv_cq = (<CQ>rcq).cq             # <<<<<<<<<<<<<<
 *             elif isinstance(rcq, CQEX):
 *                 self.attr.recv_cq = (<CQEX>rcq).ibv_cq
*/
      __pyx_t_7 = ((struct __pyx_obj_7pyverbs_2cq_CQ *)__pyx_v_rcq)->cq;
      __pyx_v_self->attr.recv_cq = __pyx_t_7;

      /* "pyverbs/qp.pyx":294
 * 
 *         if rcq is not None:
 *             if type(rcq) is CQ:             # <<<<<<<<<<<<<<
 *                 self.attr.recv_cq = (<CQ>rcq).cq
 *             elif isinstance(rcq, CQEX):
*/
      goto __pyx_L6;
    }

    /* "pyverbs/qp.pyx":296
 *             if type(rcq) is CQ:
 *                 self.attr.recv_cq = (<CQ>rcq).cq
 *             elif isinstance(rcq, CQEX):             # <<<<<<<<<<<<<<
 *                 self.attr.recv_cq = (<CQEX>rcq).ibv_cq
 *             else:
*/
    __pyx_t_6 = __Pyx_TypeCheck(((PyObject *)__pyx_v_rcq), __pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX); 
    if (likely(__pyx_t_6)) {

      /* "pyverbs/qp.pyx":297
 *                 self.attr.recv_cq = (<CQ>rcq).cq
 *             elif isinstance(rcq, CQEX):
 *                 self.attr.recv_cq = (<CQEX>rcq).ibv_cq             # <<<<<<<<<<<<<<
 *             else:
 *                 raise PyverbsUserError('Expected CQ/CQEX, got {t}'.\
*/
      __pyx_t_7 = ((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_rcq)->ibv_cq;
      __pyx_v_self->attr.recv_cq = __pyx_t_7;

      /* "pyverbs/qp.pyx":296
 *             if type(rcq) is CQ:
 *                 self.attr.recv_cq = (<CQ>rcq).cq
 *             elif isinstance(rcq, CQEX):             # <<<<<<<<<<<<<<
 *                 self.attr.recv_cq = (<CQEX>rcq).ibv_cq
 *             else:
*/
      goto __pyx_L6;
    }

    /* "pyverbs/qp.pyx":299
 *                 self.attr.recv_cq = (<CQEX>rcq).ibv_cq
 *             else:
 *                 raise PyverbsUserError('Expected CQ/CQEX, got {t}'.\             # <<<<<<<<<<<<<<
 *                                        format(t=type(rcq)))
 *         self.rcq = rcq
*/
    /*else*/ {
      __pyx_t_3 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsUserError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 299, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "pyverbs/qp.pyx":300
 *             else:
 *                 raise PyverbsUserError('Expected CQ/CQEX, got {t}'.\
 *                                        format(t=type(rcq)))             # <<<<<<<<<<<<<<
 *         self.rcq = rcq
 * 
*/
      __pyx_t_9 = __pyx_mstate_global->__pyx_kp_u_Expected_CQ_CQEX_got_t;
      __Pyx_INCREF(__pyx_t_9);
      __pyx_t_5 = 0;
      {
        PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_9, NULL};
        __pyx_t_8 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 300, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_t, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_rcq))), __pyx_t_8, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 300, __pyx_L1_error)
        __pyx_t_2 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_format, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_8);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 300, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __pyx_t_5 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
        assert(__pyx_t_3);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
        __pyx_t_5 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_2};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 299, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 299, __pyx_L1_error)
    }
    __pyx_L6:;

    /* "pyverbs/qp.pyx":293
 *         self.scq = scq
 * 
 *         if rcq is not None:             # <<<<<<<<<<<<<<
 *             if type(rcq) is CQ:
 *                 self.attr.recv_cq = (<CQ>rcq).cq
*/
  }

  /* "pyverbs/qp.pyx":301
 *                 raise PyverbsUserError('Expected CQ/CQEX, got {t}'.\
 *                                        format(t=type(rcq)))
 *         self.rcq = rcq             # <<<<<<<<<<<<<<
 * 
 *         self.srq = srq
*/
  __Pyx_INCREF((PyObject *)__pyx_v_rcq);
  __Pyx_GIVEREF((PyObject *)__pyx_v_rcq);
  __Pyx_GOTREF(__pyx_v_self->rcq);
  __Pyx_DECREF(__pyx_v_self->rcq);
  __pyx_v_self->rcq = ((PyObject *)__pyx_v_rcq);

  /* "pyverbs/qp.pyx":303
 *         self.rcq = rcq
 * 
 *         self.srq = srq             # <<<<<<<<<<<<<<
 *         self.attr.srq = srq.srq if srq else NULL
 *         self.xrcd = xrcd
*/
  __Pyx_INCREF((PyObject *)__pyx_v_srq);
  __Pyx_GIVEREF((PyObject *)__pyx_v_srq);
  __Pyx_GOTREF(__pyx_v_self->srq);
  __Pyx_DECREF(__pyx_v_self->srq);
  __pyx_v_self->srq = ((PyObject *)__pyx_v_srq);

  /* "pyverbs/qp.pyx":304
 * 
 *         self.srq = srq
 *         self.attr.srq = srq.srq if srq else NULL             # <<<<<<<<<<<<<<
 *         self.xrcd = xrcd
 *         self.attr.xrcd = xrcd.xrcd if xrcd else NULL
*/
  __pyx_t_6 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_srq)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 304, __pyx_L1_error)
  if (__pyx_t_6) {
    __pyx_t_10 = __pyx_v_srq->srq;
  } else {
    __pyx_t_10 = NULL;
  }
  __pyx_v_self->attr.srq = __pyx_t_10;

  /* "pyverbs/qp.pyx":305
 *         self.srq = srq
 *         self.attr.srq = srq.srq if srq else NULL
 *         self.xrcd = xrcd             # <<<<<<<<<<<<<<
 *         self.attr.xrcd = xrcd.xrcd if xrcd else NULL
 *         if hash_conf:
*/
  __Pyx_INCREF((PyObject *)__pyx_v_xrcd);
  __Pyx_GIVEREF((PyObject *)__pyx_v_xrcd);
  __Pyx_GOTREF(__pyx_v_self->xrcd);
  __Pyx_DECREF(__pyx_v_self->xrcd);
  __pyx_v_self->xrcd = ((PyObject *)__pyx_v_xrcd);

  /* "pyverbs/qp.pyx":306
 *         self.attr.srq = srq.srq if srq else NULL
 *         self.xrcd = xrcd
 *         self.attr.xrcd = xrcd.xrcd if xrcd else NULL             # <<<<<<<<<<<<<<
 *         if hash_conf:
 *             self.attr.rx_hash_conf = hash_conf.rx_hash_conf
*/
  __pyx_t_6 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_xrcd)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 306, __pyx_L1_error)
  if (__pyx_t_6) {
    __pyx_t_11 = __pyx_v_xrcd->xrcd;
  } else {
    __pyx_t_11 = NULL;
  }
  __pyx_v_self->attr.xrcd = __pyx_t_11;

  /* "pyverbs/qp.pyx":307
 *         self.xrcd = xrcd
 *         self.attr.xrcd = xrcd.xrcd if xrcd else NULL
 *         if hash_conf:             # <<<<<<<<<<<<<<
 *             self.attr.rx_hash_conf = hash_conf.rx_hash_conf
 *         self.ind_table = ind_table
*/
  __pyx_t_6 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_hash_conf)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 307, __pyx_L1_error)
  if (__pyx_t_6) {

    /* "pyverbs/qp.pyx":308
 *         self.attr.xrcd = xrcd.xrcd if xrcd else NULL
 *         if hash_conf:
 *             self.attr.rx_hash_conf = hash_conf.rx_hash_conf             # <<<<<<<<<<<<<<
 *         self.ind_table = ind_table
 *         self.attr.rwq_ind_tbl = ind_table.rwq_ind_table if ind_table else NULL
*/
    __pyx_t_12 = __pyx_v_hash_conf->rx_hash_conf;
    __pyx_v_self->attr.rx_hash_conf = __pyx_t_12;

    /* "pyverbs/qp.pyx":307
 *         self.xrcd = xrcd
 *         self.attr.xrcd = xrcd.xrcd if xrcd else NULL
 *         if hash_conf:             # <<<<<<<<<<<<<<
 *             self.attr.rx_hash_conf = hash_conf.rx_hash_conf
 *         self.ind_table = ind_table
*/
  }

  /* "pyverbs/qp.pyx":309
 *         if hash_conf:
 *             self.attr.rx_hash_conf = hash_conf.rx_hash_conf
 *         self.ind_table = ind_table             # <<<<<<<<<<<<<<
 *         self.attr.rwq_ind_tbl = ind_table.rwq_ind_table if ind_table else NULL
 *         self.attr.qp_type = qp_type
*/
  __Pyx_INCREF((PyObject *)__pyx_v_ind_table);
  __Pyx_GIVEREF((PyObject *)__pyx_v_ind_table);
  __Pyx_GOTREF(__pyx_v_self->ind_table);
  __Pyx_DECREF(__pyx_v_self->ind_table);
  __pyx_v_self->ind_table = ((PyObject *)__pyx_v_ind_table);

  /* "pyverbs/qp.pyx":310
 *             self.attr.rx_hash_conf = hash_conf.rx_hash_conf
 *         self.ind_table = ind_table
 *         self.attr.rwq_ind_tbl = ind_table.rwq_ind_table if ind_table else NULL             # <<<<<<<<<<<<<<
 *         self.attr.qp_type = qp_type
 *         self.attr.sq_sig_all = sq_sig_all
*/
  __pyx_t_6 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_ind_table)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 310, __pyx_L1_error)
  if (__pyx_t_6) {
    __pyx_t_13 = __pyx_v_ind_table->rwq_ind_table;
  } else {
    __pyx_t_13 = NULL;
  }
  __pyx_v_self->attr.rwq_ind_tbl = __pyx_t_13;

  /* "pyverbs/qp.pyx":311
 *         self.ind_table = ind_table
 *         self.attr.rwq_ind_tbl = ind_table.rwq_ind_table if ind_table else NULL
 *         self.attr.qp_type = qp_type             # <<<<<<<<<<<<<<
 *         self.attr.sq_sig_all = sq_sig_all
 *         self.attr.comp_mask = comp_mask
*/
  __pyx_t_14 = ((enum ibv_qp_type)__Pyx_PyLong_As_enum__ibv_qp_type(__pyx_v_qp_type)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 311, __pyx_L1_error)
  __pyx_v_self->attr.qp_type = __pyx_t_14;

  /* "pyverbs/qp.pyx":312
 *         self.attr.rwq_ind_tbl = ind_table.rwq_ind_table if ind_table else NULL
 *         self.attr.qp_type = qp_type
 *         self.attr.sq_sig_all = sq_sig_all             # <<<<<<<<<<<<<<
 *         self.attr.comp_mask = comp_mask
 *         if pd is not None:
*/
  __pyx_t_15 = __Pyx_PyLong_As_int(__pyx_v_sq_sig_all); if (unlikely((__pyx_t_15 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 312, __pyx_L1_error)
  __pyx_v_self->attr.sq_sig_all = __pyx_t_15;

  /* "pyverbs/qp.pyx":313
 *         self.attr.qp_type = qp_type
 *         self.attr.sq_sig_all = sq_sig_all
 *         self.attr.comp_mask = comp_mask             # <<<<<<<<<<<<<<
 *         if pd is not None:
 *             self._pd = pd
*/
  __pyx_t_16 = __Pyx_PyLong_As_unsigned_int(__pyx_v_comp_mask); if (unlikely((__pyx_t_16 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 313, __pyx_L1_error)
  __pyx_v_self->attr.comp_mask = __pyx_t_16;

  /* "pyverbs/qp.pyx":314
 *         self.attr.sq_sig_all = sq_sig_all
 *         self.attr.comp_mask = comp_mask
 *         if pd is not None:             # <<<<<<<<<<<<<<
 *             self._pd = pd
 *             self.attr.pd = pd.pd
*/
  __pyx_t_6 = (((PyObject *)__pyx_v_pd) != Py_None);
  if (__pyx_t_6) {

    /* "pyverbs/qp.pyx":315
 *         self.attr.comp_mask = comp_mask
 *         if pd is not None:
 *             self._pd = pd             # <<<<<<<<<<<<<<
 *             self.attr.pd = pd.pd
 *         self.attr.create_flags = create_flags
*/
    __Pyx_INCREF((PyObject *)__pyx_v_pd);
    __Pyx_GIVEREF((PyObject *)__pyx_v_pd);
    __Pyx_GOTREF(__pyx_v_self->_pd);
    __Pyx_DECREF(__pyx_v_self->_pd);
    __pyx_v_self->_pd = ((PyObject *)__pyx_v_pd);

    /* "pyverbs/qp.pyx":316
 *         if pd is not None:
 *             self._pd = pd
 *             self.attr.pd = pd.pd             # <<<<<<<<<<<<<<
 *         self.attr.create_flags = create_flags
 *         self.attr.max_tso_header = max_tso_header
*/
    __pyx_t_17 = __pyx_v_pd->pd;
    __pyx_v_self->attr.pd = __pyx_t_17;

    /* "pyverbs/qp.pyx":314
 *         self.attr.sq_sig_all = sq_sig_all
 *         self.attr.comp_mask = comp_mask
 *         if pd is not None:             # <<<<<<<<<<<<<<
 *             self._pd = pd
 *             self.attr.pd = pd.pd
*/
  }

  /* "pyverbs/qp.pyx":317
 *             self._pd = pd
 *             self.attr.pd = pd.pd
 *         self.attr.create_flags = create_flags             # <<<<<<<<<<<<<<
 *         self.attr.max_tso_header = max_tso_header
 *         self.attr.source_qpn = source_qpn
*/
  __pyx_t_16 = __Pyx_PyLong_As_unsigned_int(__pyx_v_create_flags); if (unlikely((__pyx_t_16 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 317, __pyx_L1_error)
  __pyx_v_self->attr.create_flags = __pyx_t_16;

  /* "pyverbs/qp.pyx":318
 *             self.attr.pd = pd.pd
 *         self.attr.create_flags = create_flags
 *         self.attr.max_tso_header = max_tso_header             # <<<<<<<<<<<<<<
 *         self.attr.source_qpn = source_qpn
 *         self.attr.send_ops_flags = send_ops_flags
*/
  __pyx_t_18 = __Pyx_PyLong_As_unsigned_short(__pyx_v_max_tso_header); if (unlikely((__pyx_t_18 == (unsigned short)-1) && PyErr_Occurred())) __PYX_ERR(0, 318, __pyx_L1_error)
  __pyx_v_self->attr.max_tso_header = __pyx_t_18;

  /* "pyverbs/qp.pyx":319
 *         self.attr.create_flags = create_flags
 *         self.attr.max_tso_header = max_tso_header
 *         self.attr.source_qpn = source_qpn             # <<<<<<<<<<<<<<
 *         self.attr.send_ops_flags = send_ops_flags
 * 
*/
  __pyx_t_16 = __Pyx_PyLong_As_unsigned_int(__pyx_v_source_qpn); if (unlikely((__pyx_t_16 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 319, __pyx_L1_error)
  __pyx_v_self->attr.source_qpn = __pyx_t_16;

  /* "pyverbs/qp.pyx":320
 *         self.attr.max_tso_header = max_tso_header
 *         self.attr.source_qpn = source_qpn
 *         self.attr.send_ops_flags = send_ops_flags             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_19 = __Pyx_PyLong_As_unsigned_long(__pyx_v_send_ops_flags); if (unlikely((__pyx_t_19 == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 320, __pyx_L1_error)
  __pyx_v_self->attr.send_ops_flags = __pyx_t_19;

  /* "pyverbs/qp.pyx":251
 * 
 * cdef class QPInitAttrEx(PyverbsObject):
 *     def __init__(self, qp_type=e.IBV_QPT_UD, qp_context=None,             # <<<<<<<<<<<<<<
 *                  PyverbsObject scq=None, PyverbsObject rcq=None,
 *                  SRQ srq=None, QPCap cap=None, sq_sig_all=0, comp_mask=0,
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":322
 *         self.attr.send_ops_flags = send_ops_flags
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def send_cq(self):
 *         return self.scq
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_7send_cq_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_7send_cq_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_7send_cq___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_7send_cq___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":324
 *     @property
 *     def send_cq(self):
 *         return self.scq             # <<<<<<<<<<<<<<
 *     @send_cq.setter
 *     def send_cq(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->scq);
  __pyx_r = __pyx_v_self->scq;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":322
 *         self.attr.send_ops_flags = send_ops_flags
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def send_cq(self):
 *         return self.scq
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":325
 *     def send_cq(self):
 *         return self.scq
 *     @send_cq.setter             # <<<<<<<<<<<<<<
 *     def send_cq(self, val):
 *         if type(val) is CQ:
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_7send_cq_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_7send_cq_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_7send_cq_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_7send_cq_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  struct ibv_cq *__pyx_t_2;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "pyverbs/qp.pyx":327
 *     @send_cq.setter
 *     def send_cq(self, val):
 *         if type(val) is CQ:             # <<<<<<<<<<<<<<
 *             self.attr.send_cq = (<CQ>val).cq
 *         elif type(val) is CQEX:
*/
  __pyx_t_1 = (((PyObject *)Py_TYPE(__pyx_v_val)) == ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQ));
  if (__pyx_t_1) {

    /* "pyverbs/qp.pyx":328
 *     def send_cq(self, val):
 *         if type(val) is CQ:
 *             self.attr.send_cq = (<CQ>val).cq             # <<<<<<<<<<<<<<
 *         elif type(val) is CQEX:
 *             self.attr.send_cq = (<CQEX>val).ibv_cq
*/
    __pyx_t_2 = ((struct __pyx_obj_7pyverbs_2cq_CQ *)__pyx_v_val)->cq;
    __pyx_v_self->attr.send_cq = __pyx_t_2;

    /* "pyverbs/qp.pyx":327
 *     @send_cq.setter
 *     def send_cq(self, val):
 *         if type(val) is CQ:             # <<<<<<<<<<<<<<
 *             self.attr.send_cq = (<CQ>val).cq
 *         elif type(val) is CQEX:
*/
    goto __pyx_L3;
  }

  /* "pyverbs/qp.pyx":329
 *         if type(val) is CQ:
 *             self.attr.send_cq = (<CQ>val).cq
 *         elif type(val) is CQEX:             # <<<<<<<<<<<<<<
 *             self.attr.send_cq = (<CQEX>val).ibv_cq
 *         self.scq = val
*/
  __pyx_t_1 = (((PyObject *)Py_TYPE(__pyx_v_val)) == ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX));
  if (__pyx_t_1) {

    /* "pyverbs/qp.pyx":330
 *             self.attr.send_cq = (<CQ>val).cq
 *         elif type(val) is CQEX:
 *             self.attr.send_cq = (<CQEX>val).ibv_cq             # <<<<<<<<<<<<<<
 *         self.scq = val
 * 
*/
    __pyx_t_2 = ((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_val)->ibv_cq;
    __pyx_v_self->attr.send_cq = __pyx_t_2;

    /* "pyverbs/qp.pyx":329
 *         if type(val) is CQ:
 *             self.attr.send_cq = (<CQ>val).cq
 *         elif type(val) is CQEX:             # <<<<<<<<<<<<<<
 *             self.attr.send_cq = (<CQEX>val).ibv_cq
 *         self.scq = val
*/
  }
  __pyx_L3:;

  /* "pyverbs/qp.pyx":331
 *         elif type(val) is CQEX:
 *             self.attr.send_cq = (<CQEX>val).ibv_cq
 *         self.scq = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_INCREF(__pyx_v_val);
  __Pyx_GIVEREF(__pyx_v_val);
  __Pyx_GOTREF(__pyx_v_self->scq);
  __Pyx_DECREF(__pyx_v_self->scq);
  __pyx_v_self->scq = __pyx_v_val;

  /* "pyverbs/qp.pyx":325
 *     def send_cq(self):
 *         return self.scq
 *     @send_cq.setter             # <<<<<<<<<<<<<<
 *     def send_cq(self, val):
 *         if type(val) is CQ:
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":333
 *         self.scq = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def recv_cq(self):
 *         return self.rcq
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_7recv_cq_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_7recv_cq_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_7recv_cq___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_7recv_cq___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":335
 *     @property
 *     def recv_cq(self):
 *         return self.rcq             # <<<<<<<<<<<<<<
 *     @recv_cq.setter
 *     def recv_cq(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->rcq);
  __pyx_r = __pyx_v_self->rcq;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":333
 *         self.scq = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def recv_cq(self):
 *         return self.rcq
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":336
 *     def recv_cq(self):
 *         return self.rcq
 *     @recv_cq.setter             # <<<<<<<<<<<<<<
 *     def recv_cq(self, val):
 *         if type(val) is CQ:
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_7recv_cq_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_7recv_cq_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_7recv_cq_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_7recv_cq_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  struct ibv_cq *__pyx_t_2;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "pyverbs/qp.pyx":338
 *     @recv_cq.setter
 *     def recv_cq(self, val):
 *         if type(val) is CQ:             # <<<<<<<<<<<<<<
 *             self.attr.recv_cq = (<CQ>val).cq
 *         elif type(val) is CQEX:
*/
  __pyx_t_1 = (((PyObject *)Py_TYPE(__pyx_v_val)) == ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQ));
  if (__pyx_t_1) {

    /* "pyverbs/qp.pyx":339
 *     def recv_cq(self, val):
 *         if type(val) is CQ:
 *             self.attr.recv_cq = (<CQ>val).cq             # <<<<<<<<<<<<<<
 *         elif type(val) is CQEX:
 *             self.attr.recv_cq = (<CQEX>val).ibv_cq
*/
    __pyx_t_2 = ((struct __pyx_obj_7pyverbs_2cq_CQ *)__pyx_v_val)->cq;
    __pyx_v_self->attr.recv_cq = __pyx_t_2;

    /* "pyverbs/qp.pyx":338
 *     @recv_cq.setter
 *     def recv_cq(self, val):
 *         if type(val) is CQ:             # <<<<<<<<<<<<<<
 *             self.attr.recv_cq = (<CQ>val).cq
 *         elif type(val) is CQEX:
*/
    goto __pyx_L3;
  }

  /* "pyverbs/qp.pyx":340
 *         if type(val) is CQ:
 *             self.attr.recv_cq = (<CQ>val).cq
 *         elif type(val) is CQEX:             # <<<<<<<<<<<<<<
 *             self.attr.recv_cq = (<CQEX>val).ibv_cq
 *         self.rcq = val
*/
  __pyx_t_1 = (((PyObject *)Py_TYPE(__pyx_v_val)) == ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX));
  if (__pyx_t_1) {

    /* "pyverbs/qp.pyx":341
 *             self.attr.recv_cq = (<CQ>val).cq
 *         elif type(val) is CQEX:
 *             self.attr.recv_cq = (<CQEX>val).ibv_cq             # <<<<<<<<<<<<<<
 *         self.rcq = val
 * 
*/
    __pyx_t_2 = ((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_val)->ibv_cq;
    __pyx_v_self->attr.recv_cq = __pyx_t_2;

    /* "pyverbs/qp.pyx":340
 *         if type(val) is CQ:
 *             self.attr.recv_cq = (<CQ>val).cq
 *         elif type(val) is CQEX:             # <<<<<<<<<<<<<<
 *             self.attr.recv_cq = (<CQEX>val).ibv_cq
 *         self.rcq = val
*/
  }
  __pyx_L3:;

  /* "pyverbs/qp.pyx":342
 *         elif type(val) is CQEX:
 *             self.attr.recv_cq = (<CQEX>val).ibv_cq
 *         self.rcq = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_INCREF(__pyx_v_val);
  __Pyx_GIVEREF(__pyx_v_val);
  __Pyx_GOTREF(__pyx_v_self->rcq);
  __Pyx_DECREF(__pyx_v_self->rcq);
  __pyx_v_self->rcq = __pyx_v_val;

  /* "pyverbs/qp.pyx":336
 *     def recv_cq(self):
 *         return self.rcq
 *     @recv_cq.setter             # <<<<<<<<<<<<<<
 *     def recv_cq(self, val):
 *         if type(val) is CQ:
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":344
 *         self.rcq = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def cap(self):
 *         return QPCap(max_send_wr=self.attr.cap.max_send_wr,
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_3cap_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_3cap_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_3cap___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_3cap___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":346
 *     @property
 *     def cap(self):
 *         return QPCap(max_send_wr=self.attr.cap.max_send_wr,             # <<<<<<<<<<<<<<
 *                      max_recv_wr=self.attr.cap.max_recv_wr,
 *                      max_send_sge=self.attr.cap.max_send_sge,
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = NULL;
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_send_wr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "pyverbs/qp.pyx":347
 *     def cap(self):
 *         return QPCap(max_send_wr=self.attr.cap.max_send_wr,
 *                      max_recv_wr=self.attr.cap.max_recv_wr,             # <<<<<<<<<<<<<<
 *                      max_send_sge=self.attr.cap.max_send_sge,
 *                      max_recv_sge=self.attr.cap.max_recv_sge,
*/
  __pyx_t_4 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_recv_wr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "pyverbs/qp.pyx":348
 *         return QPCap(max_send_wr=self.attr.cap.max_send_wr,
 *                      max_recv_wr=self.attr.cap.max_recv_wr,
 *                      max_send_sge=self.attr.cap.max_send_sge,             # <<<<<<<<<<<<<<
 *                      max_recv_sge=self.attr.cap.max_recv_sge,
 *                      max_inline_data=self.attr.cap.max_inline_data)
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_send_sge); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "pyverbs/qp.pyx":349
 *                      max_recv_wr=self.attr.cap.max_recv_wr,
 *                      max_send_sge=self.attr.cap.max_send_sge,
 *                      max_recv_sge=self.attr.cap.max_recv_sge,             # <<<<<<<<<<<<<<
 *                      max_inline_data=self.attr.cap.max_inline_data)
 *     @cap.setter
*/
  __pyx_t_6 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_recv_sge); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "pyverbs/qp.pyx":350
 *                      max_send_sge=self.attr.cap.max_send_sge,
 *                      max_recv_sge=self.attr.cap.max_recv_sge,
 *                      max_inline_data=self.attr.cap.max_inline_data)             # <<<<<<<<<<<<<<
 *     @cap.setter
 *     def cap(self, val):
*/
  __pyx_t_7 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_inline_data); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = 1;
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 5 : 0)] = {__pyx_t_2, NULL};
    __pyx_t_9 = __Pyx_MakeVectorcallBuilderKwds(5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_max_send_wr, __pyx_t_3, __pyx_t_9, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 346, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_max_recv_wr, __pyx_t_4, __pyx_t_9, __pyx_callargs+1, 1) < (0)) __PYX_ERR(0, 346, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_max_send_sge, __pyx_t_5, __pyx_t_9, __pyx_callargs+1, 2) < (0)) __PYX_ERR(0, 346, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_max_recv_sge, __pyx_t_6, __pyx_t_9, __pyx_callargs+1, 3) < (0)) __PYX_ERR(0, 346, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_max_inline_data, __pyx_t_7, __pyx_t_9, __pyx_callargs+1, 4) < (0)) __PYX_ERR(0, 346, __pyx_L1_error)
    __pyx_t_1 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPCap, __pyx_callargs+__pyx_t_8, (1-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 346, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_r = ((PyObject *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":344
 *         self.rcq = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def cap(self):
 *         return QPCap(max_send_wr=self.attr.cap.max_send_wr,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.cap.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":351
 *                      max_recv_sge=self.attr.cap.max_recv_sge,
 *                      max_inline_data=self.attr.cap.max_inline_data)
 *     @cap.setter             # <<<<<<<<<<<<<<
 *     def cap(self, val):
 *         _copy_caps(val, self)
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_3cap_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_3cap_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_3cap_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_3cap_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "pyverbs/qp.pyx":353
 *     @cap.setter
 *     def cap(self, val):
 *         _copy_caps(val, self)             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_copy_caps); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
    __pyx_t_4 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_v_val, ((PyObject *)__pyx_v_self)};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_4, (3-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":351
 *                      max_recv_sge=self.attr.cap.max_recv_sge,
 *                      max_inline_data=self.attr.cap.max_inline_data)
 *     @cap.setter             # <<<<<<<<<<<<<<
 *     def cap(self, val):
 *         _copy_caps(val, self)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.cap.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":355
 *         _copy_caps(val, self)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qp_type(self):
 *         return self.attr.qp_type
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_7qp_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_7qp_type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_7qp_type___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_7qp_type___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":357
 *     @property
 *     def qp_type(self):
 *         return self.attr.qp_type             # <<<<<<<<<<<<<<
 *     @qp_type.setter
 *     def qp_type(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(__pyx_v_self->attr.qp_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":355
 *         _copy_caps(val, self)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qp_type(self):
 *         return self.attr.qp_type
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.qp_type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":358
 *     def qp_type(self):
 *         return self.attr.qp_type
 *     @qp_type.setter             # <<<<<<<<<<<<<<
 *     def qp_type(self, val):
 *         self.attr.qp_type = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_7qp_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_7qp_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_7qp_type_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_7qp_type_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  enum ibv_qp_type __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":360
 *     @qp_type.setter
 *     def qp_type(self, val):
 *         self.attr.qp_type = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = ((enum ibv_qp_type)__Pyx_PyLong_As_enum__ibv_qp_type(__pyx_v_val)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 360, __pyx_L1_error)
  __pyx_v_self->attr.qp_type = __pyx_t_1;

  /* "pyverbs/qp.pyx":358
 *     def qp_type(self):
 *         return self.attr.qp_type
 *     @qp_type.setter             # <<<<<<<<<<<<<<
 *     def qp_type(self, val):
 *         self.attr.qp_type = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.qp_type.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":362
 *         self.attr.qp_type = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def sq_sig_all(self):
 *         return self.attr.sq_sig_all
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_10sq_sig_all_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_10sq_sig_all_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_10sq_sig_all___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_10sq_sig_all___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":364
 *     @property
 *     def sq_sig_all(self):
 *         return self.attr.sq_sig_all             # <<<<<<<<<<<<<<
 *     @sq_sig_all.setter
 *     def sq_sig_all(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_self->attr.sq_sig_all); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":362
 *         self.attr.qp_type = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def sq_sig_all(self):
 *         return self.attr.sq_sig_all
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.sq_sig_all.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":365
 *     def sq_sig_all(self):
 *         return self.attr.sq_sig_all
 *     @sq_sig_all.setter             # <<<<<<<<<<<<<<
 *     def sq_sig_all(self, val):
 *         self.attr.sq_sig_all = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_10sq_sig_all_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_10sq_sig_all_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_10sq_sig_all_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_10sq_sig_all_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":367
 *     @sq_sig_all.setter
 *     def sq_sig_all(self, val):
 *         self.attr.sq_sig_all = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 367, __pyx_L1_error)
  __pyx_v_self->attr.sq_sig_all = __pyx_t_1;

  /* "pyverbs/qp.pyx":365
 *     def sq_sig_all(self):
 *         return self.attr.sq_sig_all
 *     @sq_sig_all.setter             # <<<<<<<<<<<<<<
 *     def sq_sig_all(self, val):
 *         self.attr.sq_sig_all = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.sq_sig_all.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":369
 *         self.attr.sq_sig_all = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def comp_mask(self):
 *         return self.attr.comp_mask
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_9comp_mask_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_9comp_mask_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_9comp_mask___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_9comp_mask___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":371
 *     @property
 *     def comp_mask(self):
 *         return self.attr.comp_mask             # <<<<<<<<<<<<<<
 *     @comp_mask.setter
 *     def comp_mask(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.comp_mask); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":369
 *         self.attr.sq_sig_all = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def comp_mask(self):
 *         return self.attr.comp_mask
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.comp_mask.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":372
 *     def comp_mask(self):
 *         return self.attr.comp_mask
 *     @comp_mask.setter             # <<<<<<<<<<<<<<
 *     def comp_mask(self, val):
 *         self.attr.comp_mask = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_9comp_mask_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_9comp_mask_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_9comp_mask_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_9comp_mask_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":374
 *     @comp_mask.setter
 *     def comp_mask(self, val):
 *         self.attr.comp_mask = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 374, __pyx_L1_error)
  __pyx_v_self->attr.comp_mask = __pyx_t_1;

  /* "pyverbs/qp.pyx":372
 *     def comp_mask(self):
 *         return self.attr.comp_mask
 *     @comp_mask.setter             # <<<<<<<<<<<<<<
 *     def comp_mask(self, val):
 *         self.attr.comp_mask = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.comp_mask.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":376
 *         self.attr.comp_mask = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def pd(self):
 *         return self._pd
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_2pd_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_2pd_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_2pd___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_2pd___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":378
 *     @property
 *     def pd(self):
 *         return self._pd             # <<<<<<<<<<<<<<
 *     @pd.setter
 *     def pd(self, PD val):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_pd);
  __pyx_r = __pyx_v_self->_pd;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":376
 *         self.attr.comp_mask = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def pd(self):
 *         return self._pd
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":379
 *     def pd(self):
 *         return self._pd
 *     @pd.setter             # <<<<<<<<<<<<<<
 *     def pd(self, PD val):
 *         self.attr.pd = <v.ibv_pd*>val.pd
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_2pd_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_2pd_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_val), __pyx_mstate_global->__pyx_ptype_7pyverbs_2pd_PD, 1, "val", 0))) __PYX_ERR(0, 380, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_2pd_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self), ((struct __pyx_obj_7pyverbs_2pd_PD *)__pyx_v_val));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  goto __pyx_L5_cleaned_up;
  __pyx_L0:;
  __pyx_L5_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_2pd_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, struct __pyx_obj_7pyverbs_2pd_PD *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "pyverbs/qp.pyx":381
 *     @pd.setter
 *     def pd(self, PD val):
 *         self.attr.pd = <v.ibv_pd*>val.pd             # <<<<<<<<<<<<<<
 *         self._pd = val
 * 
*/
  __pyx_v_self->attr.pd = ((struct ibv_pd *)__pyx_v_val->pd);

  /* "pyverbs/qp.pyx":382
 *     def pd(self, PD val):
 *         self.attr.pd = <v.ibv_pd*>val.pd
 *         self._pd = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_INCREF((PyObject *)__pyx_v_val);
  __Pyx_GIVEREF((PyObject *)__pyx_v_val);
  __Pyx_GOTREF(__pyx_v_self->_pd);
  __Pyx_DECREF(__pyx_v_self->_pd);
  __pyx_v_self->_pd = ((PyObject *)__pyx_v_val);

  /* "pyverbs/qp.pyx":379
 *     def pd(self):
 *         return self._pd
 *     @pd.setter             # <<<<<<<<<<<<<<
 *     def pd(self, PD val):
 *         self.attr.pd = <v.ibv_pd*>val.pd
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":384
 *         self._pd = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def xrcd(self):
 *         return self.xrcd
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_4xrcd_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_4xrcd_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_4xrcd___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_4xrcd___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":386
 *     @property
 *     def xrcd(self):
 *         return self.xrcd             # <<<<<<<<<<<<<<
 *     @xrcd.setter
 *     def xrcd(self, XRCD val):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->xrcd);
  __pyx_r = __pyx_v_self->xrcd;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":384
 *         self._pd = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def xrcd(self):
 *         return self.xrcd
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":387
 *     def xrcd(self):
 *         return self.xrcd
 *     @xrcd.setter             # <<<<<<<<<<<<<<
 *     def xrcd(self, XRCD val):
 *         self.attr.xrcd = <v.ibv_xrcd*>val.xrcd
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_4xrcd_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_4xrcd_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_val), __pyx_mstate_global->__pyx_ptype_7pyverbs_4xrcd_XRCD, 1, "val", 0))) __PYX_ERR(0, 388, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_4xrcd_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self), ((struct __pyx_obj_7pyverbs_4xrcd_XRCD *)__pyx_v_val));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  goto __pyx_L5_cleaned_up;
  __pyx_L0:;
  __pyx_L5_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_4xrcd_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, struct __pyx_obj_7pyverbs_4xrcd_XRCD *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "pyverbs/qp.pyx":389
 *     @xrcd.setter
 *     def xrcd(self, XRCD val):
 *         self.attr.xrcd = <v.ibv_xrcd*>val.xrcd             # <<<<<<<<<<<<<<
 *         self.xrcd = val
 * 
*/
  __pyx_v_self->attr.xrcd = ((struct ibv_xrcd *)__pyx_v_val->xrcd);

  /* "pyverbs/qp.pyx":390
 *     def xrcd(self, XRCD val):
 *         self.attr.xrcd = <v.ibv_xrcd*>val.xrcd
 *         self.xrcd = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_INCREF((PyObject *)__pyx_v_val);
  __Pyx_GIVEREF((PyObject *)__pyx_v_val);
  __Pyx_GOTREF(__pyx_v_self->xrcd);
  __Pyx_DECREF(__pyx_v_self->xrcd);
  __pyx_v_self->xrcd = ((PyObject *)__pyx_v_val);

  /* "pyverbs/qp.pyx":387
 *     def xrcd(self):
 *         return self.xrcd
 *     @xrcd.setter             # <<<<<<<<<<<<<<
 *     def xrcd(self, XRCD val):
 *         self.attr.xrcd = <v.ibv_xrcd*>val.xrcd
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":392
 *         self.xrcd = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def srq(self):
 *         return self.srq
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_3srq_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_3srq_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_3srq___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_3srq___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":394
 *     @property
 *     def srq(self):
 *         return self.srq             # <<<<<<<<<<<<<<
 *     @srq.setter
 *     def srq(self, SRQ val):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->srq);
  __pyx_r = __pyx_v_self->srq;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":392
 *         self.xrcd = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def srq(self):
 *         return self.srq
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":395
 *     def srq(self):
 *         return self.srq
 *     @srq.setter             # <<<<<<<<<<<<<<
 *     def srq(self, SRQ val):
 *         self.attr.srq = <v.ibv_srq*>val.srq
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_3srq_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_3srq_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_val), __pyx_mstate_global->__pyx_ptype_7pyverbs_3srq_SRQ, 1, "val", 0))) __PYX_ERR(0, 396, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_3srq_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self), ((struct __pyx_obj_7pyverbs_3srq_SRQ *)__pyx_v_val));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  goto __pyx_L5_cleaned_up;
  __pyx_L0:;
  __pyx_L5_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_3srq_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, struct __pyx_obj_7pyverbs_3srq_SRQ *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "pyverbs/qp.pyx":397
 *     @srq.setter
 *     def srq(self, SRQ val):
 *         self.attr.srq = <v.ibv_srq*>val.srq             # <<<<<<<<<<<<<<
 *         self.srq = val
 * 
*/
  __pyx_v_self->attr.srq = ((struct ibv_srq *)__pyx_v_val->srq);

  /* "pyverbs/qp.pyx":398
 *     def srq(self, SRQ val):
 *         self.attr.srq = <v.ibv_srq*>val.srq
 *         self.srq = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_INCREF((PyObject *)__pyx_v_val);
  __Pyx_GIVEREF((PyObject *)__pyx_v_val);
  __Pyx_GOTREF(__pyx_v_self->srq);
  __Pyx_DECREF(__pyx_v_self->srq);
  __pyx_v_self->srq = ((PyObject *)__pyx_v_val);

  /* "pyverbs/qp.pyx":395
 *     def srq(self):
 *         return self.srq
 *     @srq.setter             # <<<<<<<<<<<<<<
 *     def srq(self, SRQ val):
 *         self.attr.srq = <v.ibv_srq*>val.srq
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":400
 *         self.srq = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def create_flags(self):
 *         return self.attr.create_flags
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_12create_flags_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_12create_flags_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_12create_flags___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_12create_flags___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":402
 *     @property
 *     def create_flags(self):
 *         return self.attr.create_flags             # <<<<<<<<<<<<<<
 *     @create_flags.setter
 *     def create_flags(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.create_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":400
 *         self.srq = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def create_flags(self):
 *         return self.attr.create_flags
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.create_flags.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":403
 *     def create_flags(self):
 *         return self.attr.create_flags
 *     @create_flags.setter             # <<<<<<<<<<<<<<
 *     def create_flags(self, val):
 *         self.attr.create_flags = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_12create_flags_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_12create_flags_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_12create_flags_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_12create_flags_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":405
 *     @create_flags.setter
 *     def create_flags(self, val):
 *         self.attr.create_flags = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 405, __pyx_L1_error)
  __pyx_v_self->attr.create_flags = __pyx_t_1;

  /* "pyverbs/qp.pyx":403
 *     def create_flags(self):
 *         return self.attr.create_flags
 *     @create_flags.setter             # <<<<<<<<<<<<<<
 *     def create_flags(self, val):
 *         self.attr.create_flags = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.create_flags.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":407
 *         self.attr.create_flags = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_tso_header(self):
 *         return self.attr.max_tso_header
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_14max_tso_header_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_14max_tso_header_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_14max_tso_header___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_14max_tso_header___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":409
 *     @property
 *     def max_tso_header(self):
 *         return self.attr.max_tso_header             # <<<<<<<<<<<<<<
 *     @max_tso_header.setter
 *     def max_tso_header(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_short(__pyx_v_self->attr.max_tso_header); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":407
 *         self.attr.create_flags = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_tso_header(self):
 *         return self.attr.max_tso_header
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.max_tso_header.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":410
 *     def max_tso_header(self):
 *         return self.attr.max_tso_header
 *     @max_tso_header.setter             # <<<<<<<<<<<<<<
 *     def max_tso_header(self, val):
 *         self.attr.max_tso_header = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_14max_tso_header_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_14max_tso_header_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_14max_tso_header_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_14max_tso_header_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned short __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":412
 *     @max_tso_header.setter
 *     def max_tso_header(self, val):
 *         self.attr.max_tso_header = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_short(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned short)-1) && PyErr_Occurred())) __PYX_ERR(0, 412, __pyx_L1_error)
  __pyx_v_self->attr.max_tso_header = __pyx_t_1;

  /* "pyverbs/qp.pyx":410
 *     def max_tso_header(self):
 *         return self.attr.max_tso_header
 *     @max_tso_header.setter             # <<<<<<<<<<<<<<
 *     def max_tso_header(self, val):
 *         self.attr.max_tso_header = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.max_tso_header.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":414
 *         self.attr.max_tso_header = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def source_qpn(self):
 *         return self.attr.source_qpn
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_10source_qpn_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_10source_qpn_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_10source_qpn___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_10source_qpn___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":416
 *     @property
 *     def source_qpn(self):
 *         return self.attr.source_qpn             # <<<<<<<<<<<<<<
 *     @source_qpn.setter
 *     def source_qpn(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.source_qpn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":414
 *         self.attr.max_tso_header = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def source_qpn(self):
 *         return self.attr.source_qpn
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.source_qpn.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":417
 *     def source_qpn(self):
 *         return self.attr.source_qpn
 *     @source_qpn.setter             # <<<<<<<<<<<<<<
 *     def source_qpn(self, val):
 *         self.attr.source_qpn = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_10source_qpn_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_10source_qpn_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_10source_qpn_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_10source_qpn_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":419
 *     @source_qpn.setter
 *     def source_qpn(self, val):
 *         self.attr.source_qpn = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 419, __pyx_L1_error)
  __pyx_v_self->attr.source_qpn = __pyx_t_1;

  /* "pyverbs/qp.pyx":417
 *     def source_qpn(self):
 *         return self.attr.source_qpn
 *     @source_qpn.setter             # <<<<<<<<<<<<<<
 *     def source_qpn(self, val):
 *         self.attr.source_qpn = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.source_qpn.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":421
 *         self.attr.source_qpn = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_send_wr(self):
 *         return self.attr.cap.max_send_wr
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_11max_send_wr_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_11max_send_wr_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_11max_send_wr___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_11max_send_wr___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":423
 *     @property
 *     def max_send_wr(self):
 *         return self.attr.cap.max_send_wr             # <<<<<<<<<<<<<<
 *     @max_send_wr.setter
 *     def max_send_wr(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_send_wr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":421
 *         self.attr.source_qpn = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_send_wr(self):
 *         return self.attr.cap.max_send_wr
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.max_send_wr.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":424
 *     def max_send_wr(self):
 *         return self.attr.cap.max_send_wr
 *     @max_send_wr.setter             # <<<<<<<<<<<<<<
 *     def max_send_wr(self, val):
 *         self.attr.cap.max_send_wr = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_11max_send_wr_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_11max_send_wr_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_11max_send_wr_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_11max_send_wr_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":426
 *     @max_send_wr.setter
 *     def max_send_wr(self, val):
 *         self.attr.cap.max_send_wr = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 426, __pyx_L1_error)
  __pyx_v_self->attr.cap.max_send_wr = __pyx_t_1;

  /* "pyverbs/qp.pyx":424
 *     def max_send_wr(self):
 *         return self.attr.cap.max_send_wr
 *     @max_send_wr.setter             # <<<<<<<<<<<<<<
 *     def max_send_wr(self, val):
 *         self.attr.cap.max_send_wr = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.max_send_wr.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":428
 *         self.attr.cap.max_send_wr = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_recv_wr(self):
 *         return self.attr.cap.max_recv_wr
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_11max_recv_wr_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_11max_recv_wr_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_11max_recv_wr___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_11max_recv_wr___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":430
 *     @property
 *     def max_recv_wr(self):
 *         return self.attr.cap.max_recv_wr             # <<<<<<<<<<<<<<
 *     @max_recv_wr.setter
 *     def max_recv_wr(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_recv_wr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":428
 *         self.attr.cap.max_send_wr = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_recv_wr(self):
 *         return self.attr.cap.max_recv_wr
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.max_recv_wr.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":431
 *     def max_recv_wr(self):
 *         return self.attr.cap.max_recv_wr
 *     @max_recv_wr.setter             # <<<<<<<<<<<<<<
 *     def max_recv_wr(self, val):
 *         self.attr.cap.max_recv_wr = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_11max_recv_wr_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_11max_recv_wr_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_11max_recv_wr_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_11max_recv_wr_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":433
 *     @max_recv_wr.setter
 *     def max_recv_wr(self, val):
 *         self.attr.cap.max_recv_wr = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 433, __pyx_L1_error)
  __pyx_v_self->attr.cap.max_recv_wr = __pyx_t_1;

  /* "pyverbs/qp.pyx":431
 *     def max_recv_wr(self):
 *         return self.attr.cap.max_recv_wr
 *     @max_recv_wr.setter             # <<<<<<<<<<<<<<
 *     def max_recv_wr(self, val):
 *         self.attr.cap.max_recv_wr = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.max_recv_wr.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":435
 *         self.attr.cap.max_recv_wr = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_send_sge(self):
 *         return self.attr.cap.max_send_sge
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_12max_send_sge_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_12max_send_sge_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_12max_send_sge___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_12max_send_sge___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":437
 *     @property
 *     def max_send_sge(self):
 *         return self.attr.cap.max_send_sge             # <<<<<<<<<<<<<<
 *     @max_send_sge.setter
 *     def max_send_sge(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_send_sge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":435
 *         self.attr.cap.max_recv_wr = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_send_sge(self):
 *         return self.attr.cap.max_send_sge
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.max_send_sge.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":438
 *     def max_send_sge(self):
 *         return self.attr.cap.max_send_sge
 *     @max_send_sge.setter             # <<<<<<<<<<<<<<
 *     def max_send_sge(self, val):
 *         self.attr.cap.max_send_sge = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_12max_send_sge_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_12max_send_sge_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_12max_send_sge_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_12max_send_sge_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":440
 *     @max_send_sge.setter
 *     def max_send_sge(self, val):
 *         self.attr.cap.max_send_sge = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 440, __pyx_L1_error)
  __pyx_v_self->attr.cap.max_send_sge = __pyx_t_1;

  /* "pyverbs/qp.pyx":438
 *     def max_send_sge(self):
 *         return self.attr.cap.max_send_sge
 *     @max_send_sge.setter             # <<<<<<<<<<<<<<
 *     def max_send_sge(self, val):
 *         self.attr.cap.max_send_sge = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.max_send_sge.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":442
 *         self.attr.cap.max_send_sge = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_recv_sge(self):
 *         return self.attr.cap.max_recv_sge
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_12max_recv_sge_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_12max_recv_sge_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_12max_recv_sge___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_12max_recv_sge___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":444
 *     @property
 *     def max_recv_sge(self):
 *         return self.attr.cap.max_recv_sge             # <<<<<<<<<<<<<<
 *     @max_recv_sge.setter
 *     def max_recv_sge(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_recv_sge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":442
 *         self.attr.cap.max_send_sge = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_recv_sge(self):
 *         return self.attr.cap.max_recv_sge
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.max_recv_sge.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":445
 *     def max_recv_sge(self):
 *         return self.attr.cap.max_recv_sge
 *     @max_recv_sge.setter             # <<<<<<<<<<<<<<
 *     def max_recv_sge(self, val):
 *         self.attr.cap.max_recv_sge = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_12max_recv_sge_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_12max_recv_sge_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_12max_recv_sge_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_12max_recv_sge_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":447
 *     @max_recv_sge.setter
 *     def max_recv_sge(self, val):
 *         self.attr.cap.max_recv_sge = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 447, __pyx_L1_error)
  __pyx_v_self->attr.cap.max_recv_sge = __pyx_t_1;

  /* "pyverbs/qp.pyx":445
 *     def max_recv_sge(self):
 *         return self.attr.cap.max_recv_sge
 *     @max_recv_sge.setter             # <<<<<<<<<<<<<<
 *     def max_recv_sge(self, val):
 *         self.attr.cap.max_recv_sge = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.max_recv_sge.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":449
 *         self.attr.cap.max_recv_sge = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_inline_data(self):
 *         return self.attr.cap.max_inline_data
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_15max_inline_data_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_15max_inline_data_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_15max_inline_data___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_15max_inline_data___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":451
 *     @property
 *     def max_inline_data(self):
 *         return self.attr.cap.max_inline_data             # <<<<<<<<<<<<<<
 *     @max_inline_data.setter
 *     def max_inline_data(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_inline_data); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":449
 *         self.attr.cap.max_recv_sge = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_inline_data(self):
 *         return self.attr.cap.max_inline_data
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.max_inline_data.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":452
 *     def max_inline_data(self):
 *         return self.attr.cap.max_inline_data
 *     @max_inline_data.setter             # <<<<<<<<<<<<<<
 *     def max_inline_data(self, val):
 *         self.attr.cap.max_inline_data = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_15max_inline_data_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_15max_inline_data_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_15max_inline_data_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_15max_inline_data_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":454
 *     @max_inline_data.setter
 *     def max_inline_data(self, val):
 *         self.attr.cap.max_inline_data = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 454, __pyx_L1_error)
  __pyx_v_self->attr.cap.max_inline_data = __pyx_t_1;

  /* "pyverbs/qp.pyx":452
 *     def max_inline_data(self):
 *         return self.attr.cap.max_inline_data
 *     @max_inline_data.setter             # <<<<<<<<<<<<<<
 *     def max_inline_data(self, val):
 *         self.attr.cap.max_inline_data = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.max_inline_data.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":456
 *         self.attr.cap.max_inline_data = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ind_table(self):
 *         return self.ind_table
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_9ind_table_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_9ind_table_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_9ind_table___get__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_9ind_table___get__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":458
 *     @property
 *     def ind_table(self):
 *         return self.ind_table             # <<<<<<<<<<<<<<
 *     @ind_table.setter
 *     def ind_table(self, RwqIndTable val):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->ind_table);
  __pyx_r = __pyx_v_self->ind_table;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":456
 *         self.attr.cap.max_inline_data = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ind_table(self):
 *         return self.ind_table
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":459
 *     def ind_table(self):
 *         return self.ind_table
 *     @ind_table.setter             # <<<<<<<<<<<<<<
 *     def ind_table(self, RwqIndTable val):
 *         self.attr.rwq_ind_tbl = <v.ibv_rwq_ind_table*>val.rwq_ind_table
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_9ind_table_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_9ind_table_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_val), __pyx_mstate_global->__pyx_ptype_7pyverbs_2wq_RwqIndTable, 1, "val", 0))) __PYX_ERR(0, 460, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_9ind_table_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self), ((struct __pyx_obj_7pyverbs_2wq_RwqIndTable *)__pyx_v_val));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  goto __pyx_L5_cleaned_up;
  __pyx_L0:;
  __pyx_L5_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_9ind_table_2__set__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, struct __pyx_obj_7pyverbs_2wq_RwqIndTable *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "pyverbs/qp.pyx":461
 *     @ind_table.setter
 *     def ind_table(self, RwqIndTable val):
 *         self.attr.rwq_ind_tbl = <v.ibv_rwq_ind_table*>val.rwq_ind_table             # <<<<<<<<<<<<<<
 *         self.ind_table = val
 * 
*/
  __pyx_v_self->attr.rwq_ind_tbl = ((struct ibv_rwq_ind_table *)__pyx_v_val->rwq_ind_table);

  /* "pyverbs/qp.pyx":462
 *     def ind_table(self, RwqIndTable val):
 *         self.attr.rwq_ind_tbl = <v.ibv_rwq_ind_table*>val.rwq_ind_table
 *         self.ind_table = val             # <<<<<<<<<<<<<<
 * 
 *     def mask_to_str(self, mask):
*/
  __Pyx_INCREF((PyObject *)__pyx_v_val);
  __Pyx_GIVEREF((PyObject *)__pyx_v_val);
  __Pyx_GOTREF(__pyx_v_self->ind_table);
  __Pyx_DECREF(__pyx_v_self->ind_table);
  __pyx_v_self->ind_table = ((PyObject *)__pyx_v_val);

  /* "pyverbs/qp.pyx":459
 *     def ind_table(self):
 *         return self.ind_table
 *     @ind_table.setter             # <<<<<<<<<<<<<<
 *     def ind_table(self, RwqIndTable val):
 *         self.attr.rwq_ind_tbl = <v.ibv_rwq_ind_table*>val.rwq_ind_table
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":464
 *         self.ind_table = val
 * 
 *     def mask_to_str(self, mask):             # <<<<<<<<<<<<<<
 *         comp_masks = {1: 'PD', 2: 'XRCD', 4: 'Create Flags',
 *                       8: 'Max TSO header', 16: 'Indirection Table',
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_3mask_to_str(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_12QPInitAttrEx_3mask_to_str = {"mask_to_str", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_3mask_to_str, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_3mask_to_str(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_mask = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mask_to_str (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_mask,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 464, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 464, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "mask_to_str", 0) < (0)) __PYX_ERR(0, 464, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("mask_to_str", 1, 1, 1, i); __PYX_ERR(0, 464, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 464, __pyx_L3_error)
    }
    __pyx_v_mask = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("mask_to_str", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 464, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.mask_to_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_2mask_to_str(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self), __pyx_v_mask);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_2mask_to_str(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_mask) {
  PyObject *__pyx_v_comp_masks = NULL;
  PyObject *__pyx_v_mask_str = NULL;
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mask_to_str", 0);

  /* "pyverbs/qp.pyx":465
 * 
 *     def mask_to_str(self, mask):
 *         comp_masks = {1: 'PD', 2: 'XRCD', 4: 'Create Flags',             # <<<<<<<<<<<<<<
 *                       8: 'Max TSO header', 16: 'Indirection Table',
 *                       32: 'RX hash'}
*/
  __pyx_t_1 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_int_1, __pyx_mstate_global->__pyx_n_u_PD) < (0)) __PYX_ERR(0, 465, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_int_2, __pyx_mstate_global->__pyx_n_u_XRCD) < (0)) __PYX_ERR(0, 465, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_int_4, __pyx_mstate_global->__pyx_kp_u_Create_Flags) < (0)) __PYX_ERR(0, 465, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_int_8, __pyx_mstate_global->__pyx_kp_u_Max_TSO_header) < (0)) __PYX_ERR(0, 465, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_int_16, __pyx_mstate_global->__pyx_kp_u_Indirection_Table) < (0)) __PYX_ERR(0, 465, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_int_32, __pyx_mstate_global->__pyx_kp_u_RX_hash) < (0)) __PYX_ERR(0, 465, __pyx_L1_error)
  __pyx_v_comp_masks = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":468
 *                       8: 'Max TSO header', 16: 'Indirection Table',
 *                       32: 'RX hash'}
 *         mask_str = ''             # <<<<<<<<<<<<<<
 *         for f in comp_masks:
 *             if mask & f:
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u__4);
  __pyx_v_mask_str = __pyx_mstate_global->__pyx_kp_u__4;

  /* "pyverbs/qp.pyx":469
 *                       32: 'RX hash'}
 *         mask_str = ''
 *         for f in comp_masks:             # <<<<<<<<<<<<<<
 *             if mask & f:
 *                 mask_str += comp_masks[f]
*/
  __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_v_comp_masks, 1, ((PyObject *)NULL), (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_6 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, NULL, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_6 == 0)) break;
    if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 469, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_f, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "pyverbs/qp.pyx":470
 *         mask_str = ''
 *         for f in comp_masks:
 *             if mask & f:             # <<<<<<<<<<<<<<
 *                 mask_str += comp_masks[f]
 *                 mask_str += ' '
*/
    __pyx_t_5 = PyNumber_And(__pyx_v_mask, __pyx_v_f); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 470, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 470, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_7) {

      /* "pyverbs/qp.pyx":471
 *         for f in comp_masks:
 *             if mask & f:
 *                 mask_str += comp_masks[f]             # <<<<<<<<<<<<<<
 *                 mask_str += ' '
 *         return mask_str
*/
      __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_comp_masks, __pyx_v_f); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 471, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_v_mask_str, __pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 471, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF_SET(__pyx_v_mask_str, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "pyverbs/qp.pyx":472
 *             if mask & f:
 *                 mask_str += comp_masks[f]
 *                 mask_str += ' '             # <<<<<<<<<<<<<<
 *         return mask_str
 * 
*/
      __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_v_mask_str, __pyx_mstate_global->__pyx_kp_u__5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 472, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF_SET(__pyx_v_mask_str, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "pyverbs/qp.pyx":470
 *         mask_str = ''
 *         for f in comp_masks:
 *             if mask & f:             # <<<<<<<<<<<<<<
 *                 mask_str += comp_masks[f]
 *                 mask_str += ' '
*/
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":473
 *                 mask_str += comp_masks[f]
 *                 mask_str += ' '
 *         return mask_str             # <<<<<<<<<<<<<<
 * 
 *     def flags_to_str(self, flags):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_mask_str);
  __pyx_r = __pyx_v_mask_str;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":464
 *         self.ind_table = val
 * 
 *     def mask_to_str(self, mask):             # <<<<<<<<<<<<<<
 *         comp_masks = {1: 'PD', 2: 'XRCD', 4: 'Create Flags',
 *                       8: 'Max TSO header', 16: 'Indirection Table',
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.mask_to_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_comp_masks);
  __Pyx_XDECREF(__pyx_v_mask_str);
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":475
 *         return mask_str
 * 
 *     def flags_to_str(self, flags):             # <<<<<<<<<<<<<<
 *         create_flags = {1: 'Block self mcast loopback', 2: 'Scatter FCS',
 *                         4: 'CVLAN stripping', 8: 'Source QPN',
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_5flags_to_str(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_12QPInitAttrEx_5flags_to_str = {"flags_to_str", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_5flags_to_str, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_5flags_to_str(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_flags = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("flags_to_str (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_flags,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 475, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 475, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "flags_to_str", 0) < (0)) __PYX_ERR(0, 475, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("flags_to_str", 1, 1, 1, i); __PYX_ERR(0, 475, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 475, __pyx_L3_error)
    }
    __pyx_v_flags = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("flags_to_str", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 475, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.flags_to_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_4flags_to_str(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self), __pyx_v_flags);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_4flags_to_str(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, PyObject *__pyx_v_flags) {
  PyObject *__pyx_v_create_flags = NULL;
  PyObject *__pyx_v_create_str = NULL;
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("flags_to_str", 0);

  /* "pyverbs/qp.pyx":476
 * 
 *     def flags_to_str(self, flags):
 *         create_flags = {1: 'Block self mcast loopback', 2: 'Scatter FCS',             # <<<<<<<<<<<<<<
 *                         4: 'CVLAN stripping', 8: 'Source QPN',
 *                         16: 'PCI write end padding'}
*/
  __pyx_t_1 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_int_1, __pyx_mstate_global->__pyx_kp_u_Block_self_mcast_loopback) < (0)) __PYX_ERR(0, 476, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_int_2, __pyx_mstate_global->__pyx_kp_u_Scatter_FCS) < (0)) __PYX_ERR(0, 476, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_int_4, __pyx_mstate_global->__pyx_kp_u_CVLAN_stripping) < (0)) __PYX_ERR(0, 476, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_int_8, __pyx_mstate_global->__pyx_kp_u_Source_QPN) < (0)) __PYX_ERR(0, 476, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_int_16, __pyx_mstate_global->__pyx_kp_u_PCI_write_end_padding) < (0)) __PYX_ERR(0, 476, __pyx_L1_error)
  __pyx_v_create_flags = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":479
 *                         4: 'CVLAN stripping', 8: 'Source QPN',
 *                         16: 'PCI write end padding'}
 *         create_str = ''             # <<<<<<<<<<<<<<
 *         for f in create_flags:
 *             if flags & f:
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u__4);
  __pyx_v_create_str = __pyx_mstate_global->__pyx_kp_u__4;

  /* "pyverbs/qp.pyx":480
 *                         16: 'PCI write end padding'}
 *         create_str = ''
 *         for f in create_flags:             # <<<<<<<<<<<<<<
 *             if flags & f:
 *                 create_str += create_flags[f]
*/
  __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_v_create_flags, 1, ((PyObject *)NULL), (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_6 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, NULL, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_6 == 0)) break;
    if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 480, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_f, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "pyverbs/qp.pyx":481
 *         create_str = ''
 *         for f in create_flags:
 *             if flags & f:             # <<<<<<<<<<<<<<
 *                 create_str += create_flags[f]
 *                 create_str += ' '
*/
    __pyx_t_5 = PyNumber_And(__pyx_v_flags, __pyx_v_f); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_7) {

      /* "pyverbs/qp.pyx":482
 *         for f in create_flags:
 *             if flags & f:
 *                 create_str += create_flags[f]             # <<<<<<<<<<<<<<
 *                 create_str += ' '
 *         return create_str
*/
      __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_create_flags, __pyx_v_f); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 482, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_v_create_str, __pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 482, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF_SET(__pyx_v_create_str, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "pyverbs/qp.pyx":483
 *             if flags & f:
 *                 create_str += create_flags[f]
 *                 create_str += ' '             # <<<<<<<<<<<<<<
 *         return create_str
 * 
*/
      __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_v_create_str, __pyx_mstate_global->__pyx_kp_u__5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 483, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF_SET(__pyx_v_create_str, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "pyverbs/qp.pyx":481
 *         create_str = ''
 *         for f in create_flags:
 *             if flags & f:             # <<<<<<<<<<<<<<
 *                 create_str += create_flags[f]
 *                 create_str += ' '
*/
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":484
 *                 create_str += create_flags[f]
 *                 create_str += ' '
 *         return create_str             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_create_str);
  __pyx_r = __pyx_v_create_str;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":475
 *         return mask_str
 * 
 *     def flags_to_str(self, flags):             # <<<<<<<<<<<<<<
 *         create_flags = {1: 'Block self mcast loopback', 2: 'Scatter FCS',
 *                         4: 'CVLAN stripping', 8: 'Source QPN',
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.flags_to_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_create_flags);
  __Pyx_XDECREF(__pyx_v_create_str);
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":486
 *         return create_str
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         print_format = '{:20}: {:<20}\n'
 *         return print_format.format('QP type', qp_type_to_str(self.qp_type)) +\
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_7__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_7__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_6__str__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_6__str__(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_v_print_format = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "pyverbs/qp.pyx":487
 * 
 *     def __str__(self):
 *         print_format = '{:20}: {:<20}\n'             # <<<<<<<<<<<<<<
 *         return print_format.format('QP type', qp_type_to_str(self.qp_type)) +\
 *                print_format.format('SQ sig. all', self.sq_sig_all) +\
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_20_20);
  __pyx_v_print_format = __pyx_mstate_global->__pyx_kp_u_20_20;

  /* "pyverbs/qp.pyx":488
 *     def __str__(self):
 *         print_format = '{:20}: {:<20}\n'
 *         return print_format.format('QP type', qp_type_to_str(self.qp_type)) +\             # <<<<<<<<<<<<<<
 *                print_format.format('SQ sig. all', self.sq_sig_all) +\
 *                'QP caps:\n' +\
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_qp_type_to_str); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_qp_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 488, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_QP_type, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":489
 *         print_format = '{:20}: {:<20}\n'
 *         return print_format.format('QP type', qp_type_to_str(self.qp_type)) +\
 *                print_format.format('SQ sig. all', self.sq_sig_all) +\             # <<<<<<<<<<<<<<
 *                'QP caps:\n' +\
 *                print_format.format('  max send WR',
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_sq_sig_all); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_SQ_sig_all, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":488
 *     def __str__(self):
 *         print_format = '{:20}: {:<20}\n'
 *         return print_format.format('QP type', qp_type_to_str(self.qp_type)) +\             # <<<<<<<<<<<<<<
 *                print_format.format('SQ sig. all', self.sq_sig_all) +\
 *                'QP caps:\n' +\
*/
  __pyx_t_1 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":489
 *         print_format = '{:20}: {:<20}\n'
 *         return print_format.format('QP type', qp_type_to_str(self.qp_type)) +\
 *                print_format.format('SQ sig. all', self.sq_sig_all) +\             # <<<<<<<<<<<<<<
 *                'QP caps:\n' +\
 *                print_format.format('  max send WR',
*/
  __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_1, __pyx_mstate_global->__pyx_kp_u_QP_caps); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":492
 *                'QP caps:\n' +\
 *                print_format.format('  max send WR',
 *                                    self.attr.cap.max_send_wr) +\             # <<<<<<<<<<<<<<
 *                print_format.format('  max recv WR',
 *                                    self.attr.cap.max_recv_wr) +\
*/
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_send_wr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pyverbs/qp.pyx":491
 *                print_format.format('SQ sig. all', self.sq_sig_all) +\
 *                'QP caps:\n' +\
 *                print_format.format('  max send WR',             # <<<<<<<<<<<<<<
 *                                    self.attr.cap.max_send_wr) +\
 *                print_format.format('  max recv WR',
*/
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_max_send_WR_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":490
 *         return print_format.format('QP type', qp_type_to_str(self.qp_type)) +\
 *                print_format.format('SQ sig. all', self.sq_sig_all) +\
 *                'QP caps:\n' +\             # <<<<<<<<<<<<<<
 *                print_format.format('  max send WR',
 *                                    self.attr.cap.max_send_wr) +\
*/
  __pyx_t_1 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":494
 *                                    self.attr.cap.max_send_wr) +\
 *                print_format.format('  max recv WR',
 *                                    self.attr.cap.max_recv_wr) +\             # <<<<<<<<<<<<<<
 *                print_format.format('  max send SGE',
 *                                    self.attr.cap.max_send_sge) +\
*/
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_recv_wr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "pyverbs/qp.pyx":493
 *                print_format.format('  max send WR',
 *                                    self.attr.cap.max_send_wr) +\
 *                print_format.format('  max recv WR',             # <<<<<<<<<<<<<<
 *                                    self.attr.cap.max_recv_wr) +\
 *                print_format.format('  max send SGE',
*/
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_max_recv_WR_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":492
 *                'QP caps:\n' +\
 *                print_format.format('  max send WR',
 *                                    self.attr.cap.max_send_wr) +\             # <<<<<<<<<<<<<<
 *                print_format.format('  max recv WR',
 *                                    self.attr.cap.max_recv_wr) +\
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":496
 *                                    self.attr.cap.max_recv_wr) +\
 *                print_format.format('  max send SGE',
 *                                    self.attr.cap.max_send_sge) +\             # <<<<<<<<<<<<<<
 *                print_format.format('  max recv SGE',
 *                                    self.attr.cap.max_recv_sge) +\
*/
  __pyx_t_4 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_send_sge); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "pyverbs/qp.pyx":495
 *                print_format.format('  max recv WR',
 *                                    self.attr.cap.max_recv_wr) +\
 *                print_format.format('  max send SGE',             # <<<<<<<<<<<<<<
 *                                    self.attr.cap.max_send_sge) +\
 *                print_format.format('  max recv SGE',
*/
  __pyx_t_1 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_max_send_SGE_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":494
 *                                    self.attr.cap.max_send_wr) +\
 *                print_format.format('  max recv WR',
 *                                    self.attr.cap.max_recv_wr) +\             # <<<<<<<<<<<<<<
 *                print_format.format('  max send SGE',
 *                                    self.attr.cap.max_send_sge) +\
*/
  __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":498
 *                                    self.attr.cap.max_send_sge) +\
 *                print_format.format('  max recv SGE',
 *                                    self.attr.cap.max_recv_sge) +\             # <<<<<<<<<<<<<<
 *                print_format.format('  max inline data',
 *                                    self.attr.cap.max_inline_data) +\
*/
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_recv_sge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pyverbs/qp.pyx":497
 *                print_format.format('  max send SGE',
 *                                    self.attr.cap.max_send_sge) +\
 *                print_format.format('  max recv SGE',             # <<<<<<<<<<<<<<
 *                                    self.attr.cap.max_recv_sge) +\
 *                print_format.format('  max inline data',
*/
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_max_recv_SGE_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":496
 *                                    self.attr.cap.max_recv_wr) +\
 *                print_format.format('  max send SGE',
 *                                    self.attr.cap.max_send_sge) +\             # <<<<<<<<<<<<<<
 *                print_format.format('  max recv SGE',
 *                                    self.attr.cap.max_recv_sge) +\
*/
  __pyx_t_1 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":500
 *                                    self.attr.cap.max_recv_sge) +\
 *                print_format.format('  max inline data',
 *                                    self.attr.cap.max_inline_data) +\             # <<<<<<<<<<<<<<
 *                print_format.format('comp mask',
 *                                    self.mask_to_str(self.attr.comp_mask)) +\
*/
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_inline_data); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "pyverbs/qp.pyx":499
 *                print_format.format('  max recv SGE',
 *                                    self.attr.cap.max_recv_sge) +\
 *                print_format.format('  max inline data',             # <<<<<<<<<<<<<<
 *                                    self.attr.cap.max_inline_data) +\
 *                print_format.format('comp mask',
*/
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_max_inline_data_3, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":498
 *                                    self.attr.cap.max_send_sge) +\
 *                print_format.format('  max recv SGE',
 *                                    self.attr.cap.max_recv_sge) +\             # <<<<<<<<<<<<<<
 *                print_format.format('  max inline data',
 *                                    self.attr.cap.max_inline_data) +\
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":502
 *                                    self.attr.cap.max_inline_data) +\
 *                print_format.format('comp mask',
 *                                    self.mask_to_str(self.attr.comp_mask)) +\             # <<<<<<<<<<<<<<
 *                print_format.format('create flags',
 *                                    self.flags_to_str(self.attr.create_flags)) +\
*/
  __pyx_t_1 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.comp_mask); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_t_2};
    __pyx_t_4 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_mask_to_str, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 502, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }

  /* "pyverbs/qp.pyx":501
 *                print_format.format('  max inline data',
 *                                    self.attr.cap.max_inline_data) +\
 *                print_format.format('comp mask',             # <<<<<<<<<<<<<<
 *                                    self.mask_to_str(self.attr.comp_mask)) +\
 *                print_format.format('create flags',
*/
  __pyx_t_2 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_comp_mask_2, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":500
 *                                    self.attr.cap.max_recv_sge) +\
 *                print_format.format('  max inline data',
 *                                    self.attr.cap.max_inline_data) +\             # <<<<<<<<<<<<<<
 *                print_format.format('comp mask',
 *                                    self.mask_to_str(self.attr.comp_mask)) +\
*/
  __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":504
 *                                    self.mask_to_str(self.attr.comp_mask)) +\
 *                print_format.format('create flags',
 *                                    self.flags_to_str(self.attr.create_flags)) +\             # <<<<<<<<<<<<<<
 *                print_format.format('max TSO header',
 *                                    self.attr.max_tso_header) +\
*/
  __pyx_t_3 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.create_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_1};
    __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_flags_to_str, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }

  /* "pyverbs/qp.pyx":503
 *                print_format.format('comp mask',
 *                                    self.mask_to_str(self.attr.comp_mask)) +\
 *                print_format.format('create flags',             # <<<<<<<<<<<<<<
 *                                    self.flags_to_str(self.attr.create_flags)) +\
 *                print_format.format('max TSO header',
*/
  __pyx_t_1 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_create_flags_2, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":502
 *                                    self.attr.cap.max_inline_data) +\
 *                print_format.format('comp mask',
 *                                    self.mask_to_str(self.attr.comp_mask)) +\             # <<<<<<<<<<<<<<
 *                print_format.format('create flags',
 *                                    self.flags_to_str(self.attr.create_flags)) +\
*/
  __pyx_t_2 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":506
 *                                    self.flags_to_str(self.attr.create_flags)) +\
 *                print_format.format('max TSO header',
 *                                    self.attr.max_tso_header) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Source QPN', self.attr.source_qpn)
 * 
*/
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_short(__pyx_v_self->attr.max_tso_header); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pyverbs/qp.pyx":505
 *                print_format.format('create flags',
 *                                    self.flags_to_str(self.attr.create_flags)) +\
 *                print_format.format('max TSO header',             # <<<<<<<<<<<<<<
 *                                    self.attr.max_tso_header) +\
 *                print_format.format('Source QPN', self.attr.source_qpn)
*/
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_max_TSO_header, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 505, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":504
 *                                    self.mask_to_str(self.attr.comp_mask)) +\
 *                print_format.format('create flags',
 *                                    self.flags_to_str(self.attr.create_flags)) +\             # <<<<<<<<<<<<<<
 *                print_format.format('max TSO header',
 *                                    self.attr.max_tso_header) +\
*/
  __pyx_t_1 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":507
 *                print_format.format('max TSO header',
 *                                    self.attr.max_tso_header) +\
 *                print_format.format('Source QPN', self.attr.source_qpn)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_4 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.source_qpn); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_Source_QPN, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":506
 *                                    self.flags_to_str(self.attr.create_flags)) +\
 *                print_format.format('max TSO header',
 *                                    self.attr.max_tso_header) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Source QPN', self.attr.source_qpn)
 * 
*/
  __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":486
 *         return create_str
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         print_format = '{:20}: {:<20}\n'
 *         return print_format.format('QP type', qp_type_to_str(self.qp_type)) +\
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_print_format);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_12QPInitAttrEx_9__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_8__reduce_cython__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_attr_cannot_be_converted_to, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_12QPInitAttrEx_11__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_12QPInitAttrEx_10__setstate_cython__(((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_12QPInitAttrEx_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_attr_cannot_be_converted_to, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPInitAttrEx.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":511
 * 
 * cdef class QPAttr(PyverbsObject):
 *     def __init__(self, qp_state=e.IBV_QPS_INIT, cur_qp_state=e.IBV_QPS_RESET,             # <<<<<<<<<<<<<<
 *                  port_num=1, path_mtu=e.IBV_MTU_1024):
 *         """
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2qp_6QPAttr___init__, "\n        Initializes a QPQttr object which represents ibv_qp_attr structs. It\n        can be used to modify a QP.\n        This function initializes default values for reset-to-init transition.\n        :param qp_state: Desired QP state\n        :param cur_qp_state: Current QP state\n        :return: An initialized QpAttr object\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_7pyverbs_2qp_6QPAttr___init__;
#endif
static int __pyx_pw_7pyverbs_2qp_6QPAttr_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_qp_state = 0;
  PyObject *__pyx_v_cur_qp_state = 0;
  PyObject *__pyx_v_port_num = 0;
  PyObject *__pyx_v_path_mtu = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_qp_state,&__pyx_mstate_global->__pyx_n_u_cur_qp_state,&__pyx_mstate_global->__pyx_n_u_port_num,&__pyx_mstate_global->__pyx_n_u_path_mtu,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 511, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 511, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 511, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 511, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 511, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 511, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__6);
      if (!values[1]) values[1] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__7);
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_1));
      if (!values[3]) values[3] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__8);
    } else {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 511, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 511, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 511, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 511, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__6);
      if (!values[1]) values[1] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__7);
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_1));
      if (!values[3]) values[3] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__8);
    }
    __pyx_v_qp_state = values[0];
    __pyx_v_cur_qp_state = values[1];
    __pyx_v_port_num = values[2];
    __pyx_v_path_mtu = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 4, __pyx_nargs); __PYX_ERR(0, 511, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr___init__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), __pyx_v_qp_state, __pyx_v_cur_qp_state, __pyx_v_port_num, __pyx_v_path_mtu);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr___init__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_qp_state, PyObject *__pyx_v_cur_qp_state, PyObject *__pyx_v_port_num, PyObject *__pyx_v_path_mtu) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  enum ibv_qp_state __pyx_t_6;
  unsigned char __pyx_t_7;
  enum ibv_mtu __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyverbs/qp.pyx":521
 *         :return: An initialized QpAttr object
 *         """
 *         super().__init__()             # <<<<<<<<<<<<<<
 *         self.attr.qp_state = qp_state
 *         self.attr.cur_qp_state = cur_qp_state
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPAttr), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":522
 *         """
 *         super().__init__()
 *         self.attr.qp_state = qp_state             # <<<<<<<<<<<<<<
 *         self.attr.cur_qp_state = cur_qp_state
 *         self.attr.port_num = port_num
*/
  __pyx_t_6 = ((enum ibv_qp_state)__Pyx_PyLong_As_enum__ibv_qp_state(__pyx_v_qp_state)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 522, __pyx_L1_error)
  __pyx_v_self->attr.qp_state = __pyx_t_6;

  /* "pyverbs/qp.pyx":523
 *         super().__init__()
 *         self.attr.qp_state = qp_state
 *         self.attr.cur_qp_state = cur_qp_state             # <<<<<<<<<<<<<<
 *         self.attr.port_num = port_num
 *         self.attr.path_mtu = path_mtu
*/
  __pyx_t_6 = ((enum ibv_qp_state)__Pyx_PyLong_As_enum__ibv_qp_state(__pyx_v_cur_qp_state)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 523, __pyx_L1_error)
  __pyx_v_self->attr.cur_qp_state = __pyx_t_6;

  /* "pyverbs/qp.pyx":524
 *         self.attr.qp_state = qp_state
 *         self.attr.cur_qp_state = cur_qp_state
 *         self.attr.port_num = port_num             # <<<<<<<<<<<<<<
 *         self.attr.path_mtu = path_mtu
 * 
*/
  __pyx_t_7 = __Pyx_PyLong_As_unsigned_char(__pyx_v_port_num); if (unlikely((__pyx_t_7 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 524, __pyx_L1_error)
  __pyx_v_self->attr.port_num = __pyx_t_7;

  /* "pyverbs/qp.pyx":525
 *         self.attr.cur_qp_state = cur_qp_state
 *         self.attr.port_num = port_num
 *         self.attr.path_mtu = path_mtu             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_8 = ((enum ibv_mtu)__Pyx_PyLong_As_enum__ibv_mtu(__pyx_v_path_mtu)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 525, __pyx_L1_error)
  __pyx_v_self->attr.path_mtu = __pyx_t_8;

  /* "pyverbs/qp.pyx":511
 * 
 * cdef class QPAttr(PyverbsObject):
 *     def __init__(self, qp_state=e.IBV_QPS_INIT, cur_qp_state=e.IBV_QPS_RESET,             # <<<<<<<<<<<<<<
 *                  port_num=1, path_mtu=e.IBV_MTU_1024):
 *         """
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":527
 *         self.attr.path_mtu = path_mtu
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qp_state(self):
 *         return self.attr.qp_state
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_8qp_state_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_8qp_state_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_8qp_state___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_8qp_state___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":529
 *     @property
 *     def qp_state(self):
 *         return self.attr.qp_state             # <<<<<<<<<<<<<<
 *     @qp_state.setter
 *     def qp_state(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py(__pyx_v_self->attr.qp_state); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":527
 *         self.attr.path_mtu = path_mtu
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qp_state(self):
 *         return self.attr.qp_state
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.qp_state.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":530
 *     def qp_state(self):
 *         return self.attr.qp_state
 *     @qp_state.setter             # <<<<<<<<<<<<<<
 *     def qp_state(self, val):
 *         self.attr.qp_state = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_8qp_state_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_8qp_state_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_8qp_state_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_8qp_state_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  enum ibv_qp_state __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":532
 *     @qp_state.setter
 *     def qp_state(self, val):
 *         self.attr.qp_state = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = ((enum ibv_qp_state)__Pyx_PyLong_As_enum__ibv_qp_state(__pyx_v_val)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 532, __pyx_L1_error)
  __pyx_v_self->attr.qp_state = __pyx_t_1;

  /* "pyverbs/qp.pyx":530
 *     def qp_state(self):
 *         return self.attr.qp_state
 *     @qp_state.setter             # <<<<<<<<<<<<<<
 *     def qp_state(self, val):
 *         self.attr.qp_state = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.qp_state.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":534
 *         self.attr.qp_state = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def cur_qp_state(self):
 *         return self.attr.cur_qp_state
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_12cur_qp_state_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_12cur_qp_state_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_12cur_qp_state___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_12cur_qp_state___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":536
 *     @property
 *     def cur_qp_state(self):
 *         return self.attr.cur_qp_state             # <<<<<<<<<<<<<<
 *     @cur_qp_state.setter
 *     def cur_qp_state(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py(__pyx_v_self->attr.cur_qp_state); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":534
 *         self.attr.qp_state = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def cur_qp_state(self):
 *         return self.attr.cur_qp_state
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.cur_qp_state.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":537
 *     def cur_qp_state(self):
 *         return self.attr.cur_qp_state
 *     @cur_qp_state.setter             # <<<<<<<<<<<<<<
 *     def cur_qp_state(self, val):
 *         self.attr.cur_qp_state = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_12cur_qp_state_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_12cur_qp_state_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_12cur_qp_state_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_12cur_qp_state_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  enum ibv_qp_state __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":539
 *     @cur_qp_state.setter
 *     def cur_qp_state(self, val):
 *         self.attr.cur_qp_state = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = ((enum ibv_qp_state)__Pyx_PyLong_As_enum__ibv_qp_state(__pyx_v_val)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 539, __pyx_L1_error)
  __pyx_v_self->attr.cur_qp_state = __pyx_t_1;

  /* "pyverbs/qp.pyx":537
 *     def cur_qp_state(self):
 *         return self.attr.cur_qp_state
 *     @cur_qp_state.setter             # <<<<<<<<<<<<<<
 *     def cur_qp_state(self, val):
 *         self.attr.cur_qp_state = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.cur_qp_state.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":541
 *         self.attr.cur_qp_state = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def path_mtu(self):
 *         return self.attr.path_mtu
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_8path_mtu_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_8path_mtu_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_8path_mtu___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_8path_mtu___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":543
 *     @property
 *     def path_mtu(self):
 *         return self.attr.path_mtu             # <<<<<<<<<<<<<<
 *     @path_mtu.setter
 *     def path_mtu(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py(__pyx_v_self->attr.path_mtu); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":541
 *         self.attr.cur_qp_state = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def path_mtu(self):
 *         return self.attr.path_mtu
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.path_mtu.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":544
 *     def path_mtu(self):
 *         return self.attr.path_mtu
 *     @path_mtu.setter             # <<<<<<<<<<<<<<
 *     def path_mtu(self, val):
 *         self.attr.path_mtu = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_8path_mtu_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_8path_mtu_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_8path_mtu_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_8path_mtu_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  enum ibv_mtu __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":546
 *     @path_mtu.setter
 *     def path_mtu(self, val):
 *         self.attr.path_mtu = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = ((enum ibv_mtu)__Pyx_PyLong_As_enum__ibv_mtu(__pyx_v_val)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 546, __pyx_L1_error)
  __pyx_v_self->attr.path_mtu = __pyx_t_1;

  /* "pyverbs/qp.pyx":544
 *     def path_mtu(self):
 *         return self.attr.path_mtu
 *     @path_mtu.setter             # <<<<<<<<<<<<<<
 *     def path_mtu(self, val):
 *         self.attr.path_mtu = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.path_mtu.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":548
 *         self.attr.path_mtu = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def path_mig_state(self):
 *         return self.attr.path_mig_state
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_14path_mig_state_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_14path_mig_state_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_14path_mig_state___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_14path_mig_state___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":550
 *     @property
 *     def path_mig_state(self):
 *         return self.attr.path_mig_state             # <<<<<<<<<<<<<<
 *     @path_mig_state.setter
 *     def path_mig_state(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py(__pyx_v_self->attr.path_mig_state); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":548
 *         self.attr.path_mtu = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def path_mig_state(self):
 *         return self.attr.path_mig_state
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.path_mig_state.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":551
 *     def path_mig_state(self):
 *         return self.attr.path_mig_state
 *     @path_mig_state.setter             # <<<<<<<<<<<<<<
 *     def path_mig_state(self, val):
 *         self.attr.path_mig_state = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_14path_mig_state_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_14path_mig_state_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_14path_mig_state_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_14path_mig_state_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  enum ibv_mig_state __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":553
 *     @path_mig_state.setter
 *     def path_mig_state(self, val):
 *         self.attr.path_mig_state = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = ((enum ibv_mig_state)__Pyx_PyLong_As_enum__ibv_mig_state(__pyx_v_val)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 553, __pyx_L1_error)
  __pyx_v_self->attr.path_mig_state = __pyx_t_1;

  /* "pyverbs/qp.pyx":551
 *     def path_mig_state(self):
 *         return self.attr.path_mig_state
 *     @path_mig_state.setter             # <<<<<<<<<<<<<<
 *     def path_mig_state(self, val):
 *         self.attr.path_mig_state = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.path_mig_state.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":555
 *         self.attr.path_mig_state = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qkey(self):
 *         return self.attr.qkey
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_4qkey_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_4qkey_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_4qkey___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_4qkey___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":557
 *     @property
 *     def qkey(self):
 *         return self.attr.qkey             # <<<<<<<<<<<<<<
 *     @qkey.setter
 *     def qkey(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.qkey); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":555
 *         self.attr.path_mig_state = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qkey(self):
 *         return self.attr.qkey
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.qkey.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":558
 *     def qkey(self):
 *         return self.attr.qkey
 *     @qkey.setter             # <<<<<<<<<<<<<<
 *     def qkey(self, val):
 *         self.attr.qkey = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_4qkey_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_4qkey_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_4qkey_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_4qkey_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":560
 *     @qkey.setter
 *     def qkey(self, val):
 *         self.attr.qkey = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 560, __pyx_L1_error)
  __pyx_v_self->attr.qkey = __pyx_t_1;

  /* "pyverbs/qp.pyx":558
 *     def qkey(self):
 *         return self.attr.qkey
 *     @qkey.setter             # <<<<<<<<<<<<<<
 *     def qkey(self, val):
 *         self.attr.qkey = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.qkey.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":562
 *         self.attr.qkey = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def rq_psn(self):
 *         return self.attr.rq_psn
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_6rq_psn_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_6rq_psn_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_6rq_psn___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_6rq_psn___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":564
 *     @property
 *     def rq_psn(self):
 *         return self.attr.rq_psn             # <<<<<<<<<<<<<<
 *     @rq_psn.setter
 *     def rq_psn(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.rq_psn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":562
 *         self.attr.qkey = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def rq_psn(self):
 *         return self.attr.rq_psn
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.rq_psn.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":565
 *     def rq_psn(self):
 *         return self.attr.rq_psn
 *     @rq_psn.setter             # <<<<<<<<<<<<<<
 *     def rq_psn(self, val):
 *         self.attr.rq_psn = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_6rq_psn_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_6rq_psn_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_6rq_psn_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_6rq_psn_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":567
 *     @rq_psn.setter
 *     def rq_psn(self, val):
 *         self.attr.rq_psn = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 567, __pyx_L1_error)
  __pyx_v_self->attr.rq_psn = __pyx_t_1;

  /* "pyverbs/qp.pyx":565
 *     def rq_psn(self):
 *         return self.attr.rq_psn
 *     @rq_psn.setter             # <<<<<<<<<<<<<<
 *     def rq_psn(self, val):
 *         self.attr.rq_psn = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.rq_psn.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":569
 *         self.attr.rq_psn = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def sq_psn(self):
 *         return self.attr.sq_psn
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_6sq_psn_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_6sq_psn_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_6sq_psn___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_6sq_psn___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":571
 *     @property
 *     def sq_psn(self):
 *         return self.attr.sq_psn             # <<<<<<<<<<<<<<
 *     @sq_psn.setter
 *     def sq_psn(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.sq_psn); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":569
 *         self.attr.rq_psn = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def sq_psn(self):
 *         return self.attr.sq_psn
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.sq_psn.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":572
 *     def sq_psn(self):
 *         return self.attr.sq_psn
 *     @sq_psn.setter             # <<<<<<<<<<<<<<
 *     def sq_psn(self, val):
 *         self.attr.sq_psn = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_6sq_psn_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_6sq_psn_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_6sq_psn_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_6sq_psn_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":574
 *     @sq_psn.setter
 *     def sq_psn(self, val):
 *         self.attr.sq_psn = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 574, __pyx_L1_error)
  __pyx_v_self->attr.sq_psn = __pyx_t_1;

  /* "pyverbs/qp.pyx":572
 *     def sq_psn(self):
 *         return self.attr.sq_psn
 *     @sq_psn.setter             # <<<<<<<<<<<<<<
 *     def sq_psn(self, val):
 *         self.attr.sq_psn = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.sq_psn.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":576
 *         self.attr.sq_psn = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dest_qp_num(self):
 *         return self.attr.dest_qp_num
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_11dest_qp_num_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_11dest_qp_num_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_11dest_qp_num___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_11dest_qp_num___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":578
 *     @property
 *     def dest_qp_num(self):
 *         return self.attr.dest_qp_num             # <<<<<<<<<<<<<<
 *     @dest_qp_num.setter
 *     def dest_qp_num(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.dest_qp_num); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":576
 *         self.attr.sq_psn = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dest_qp_num(self):
 *         return self.attr.dest_qp_num
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.dest_qp_num.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":579
 *     def dest_qp_num(self):
 *         return self.attr.dest_qp_num
 *     @dest_qp_num.setter             # <<<<<<<<<<<<<<
 *     def dest_qp_num(self, val):
 *         self.attr.dest_qp_num = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_11dest_qp_num_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_11dest_qp_num_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_11dest_qp_num_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_11dest_qp_num_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":581
 *     @dest_qp_num.setter
 *     def dest_qp_num(self, val):
 *         self.attr.dest_qp_num = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 581, __pyx_L1_error)
  __pyx_v_self->attr.dest_qp_num = __pyx_t_1;

  /* "pyverbs/qp.pyx":579
 *     def dest_qp_num(self):
 *         return self.attr.dest_qp_num
 *     @dest_qp_num.setter             # <<<<<<<<<<<<<<
 *     def dest_qp_num(self, val):
 *         self.attr.dest_qp_num = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.dest_qp_num.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":583
 *         self.attr.dest_qp_num = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qp_access_flags(self):
 *         return self.attr.qp_access_flags
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_15qp_access_flags_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_15qp_access_flags_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_15qp_access_flags___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_15qp_access_flags___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":585
 *     @property
 *     def qp_access_flags(self):
 *         return self.attr.qp_access_flags             # <<<<<<<<<<<<<<
 *     @qp_access_flags.setter
 *     def qp_access_flags(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.qp_access_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":583
 *         self.attr.dest_qp_num = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qp_access_flags(self):
 *         return self.attr.qp_access_flags
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.qp_access_flags.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":586
 *     def qp_access_flags(self):
 *         return self.attr.qp_access_flags
 *     @qp_access_flags.setter             # <<<<<<<<<<<<<<
 *     def qp_access_flags(self, val):
 *         self.attr.qp_access_flags = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_15qp_access_flags_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_15qp_access_flags_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_15qp_access_flags_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_15qp_access_flags_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":588
 *     @qp_access_flags.setter
 *     def qp_access_flags(self, val):
 *         self.attr.qp_access_flags = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 588, __pyx_L1_error)
  __pyx_v_self->attr.qp_access_flags = __pyx_t_1;

  /* "pyverbs/qp.pyx":586
 *     def qp_access_flags(self):
 *         return self.attr.qp_access_flags
 *     @qp_access_flags.setter             # <<<<<<<<<<<<<<
 *     def qp_access_flags(self, val):
 *         self.attr.qp_access_flags = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.qp_access_flags.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":590
 *         self.attr.qp_access_flags = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def cap(self):
 *         return QPCap(max_send_wr=self.attr.cap.max_send_wr,
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_3cap_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_3cap_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_3cap___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_3cap___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":592
 *     @property
 *     def cap(self):
 *         return QPCap(max_send_wr=self.attr.cap.max_send_wr,             # <<<<<<<<<<<<<<
 *                      max_recv_wr=self.attr.cap.max_recv_wr,
 *                      max_send_sge=self.attr.cap.max_send_sge,
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = NULL;
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_send_wr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "pyverbs/qp.pyx":593
 *     def cap(self):
 *         return QPCap(max_send_wr=self.attr.cap.max_send_wr,
 *                      max_recv_wr=self.attr.cap.max_recv_wr,             # <<<<<<<<<<<<<<
 *                      max_send_sge=self.attr.cap.max_send_sge,
 *                      max_recv_sge=self.attr.cap.max_recv_sge,
*/
  __pyx_t_4 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_recv_wr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "pyverbs/qp.pyx":594
 *         return QPCap(max_send_wr=self.attr.cap.max_send_wr,
 *                      max_recv_wr=self.attr.cap.max_recv_wr,
 *                      max_send_sge=self.attr.cap.max_send_sge,             # <<<<<<<<<<<<<<
 *                      max_recv_sge=self.attr.cap.max_recv_sge,
 *                      max_inline_data=self.attr.cap.max_inline_data)
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_send_sge); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "pyverbs/qp.pyx":595
 *                      max_recv_wr=self.attr.cap.max_recv_wr,
 *                      max_send_sge=self.attr.cap.max_send_sge,
 *                      max_recv_sge=self.attr.cap.max_recv_sge,             # <<<<<<<<<<<<<<
 *                      max_inline_data=self.attr.cap.max_inline_data)
 *     @cap.setter
*/
  __pyx_t_6 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_recv_sge); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "pyverbs/qp.pyx":596
 *                      max_send_sge=self.attr.cap.max_send_sge,
 *                      max_recv_sge=self.attr.cap.max_recv_sge,
 *                      max_inline_data=self.attr.cap.max_inline_data)             # <<<<<<<<<<<<<<
 *     @cap.setter
 *     def cap(self, val):
*/
  __pyx_t_7 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_inline_data); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = 1;
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 5 : 0)] = {__pyx_t_2, NULL};
    __pyx_t_9 = __Pyx_MakeVectorcallBuilderKwds(5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 592, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_max_send_wr, __pyx_t_3, __pyx_t_9, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 592, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_max_recv_wr, __pyx_t_4, __pyx_t_9, __pyx_callargs+1, 1) < (0)) __PYX_ERR(0, 592, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_max_send_sge, __pyx_t_5, __pyx_t_9, __pyx_callargs+1, 2) < (0)) __PYX_ERR(0, 592, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_max_recv_sge, __pyx_t_6, __pyx_t_9, __pyx_callargs+1, 3) < (0)) __PYX_ERR(0, 592, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_max_inline_data, __pyx_t_7, __pyx_t_9, __pyx_callargs+1, 4) < (0)) __PYX_ERR(0, 592, __pyx_L1_error)
    __pyx_t_1 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPCap, __pyx_callargs+__pyx_t_8, (1-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 592, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_r = ((PyObject *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":590
 *         self.attr.qp_access_flags = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def cap(self):
 *         return QPCap(max_send_wr=self.attr.cap.max_send_wr,
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.cap.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":597
 *                      max_recv_sge=self.attr.cap.max_recv_sge,
 *                      max_inline_data=self.attr.cap.max_inline_data)
 *     @cap.setter             # <<<<<<<<<<<<<<
 *     def cap(self, val):
 *         _copy_caps(val, self)
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_3cap_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_3cap_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_3cap_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_3cap_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "pyverbs/qp.pyx":599
 *     @cap.setter
 *     def cap(self, val):
 *         _copy_caps(val, self)             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_copy_caps); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
    __pyx_t_4 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_v_val, ((PyObject *)__pyx_v_self)};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_4, (3-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":597
 *                      max_recv_sge=self.attr.cap.max_recv_sge,
 *                      max_inline_data=self.attr.cap.max_inline_data)
 *     @cap.setter             # <<<<<<<<<<<<<<
 *     def cap(self, val):
 *         _copy_caps(val, self)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.cap.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":601
 *         _copy_caps(val, self)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ah_attr(self):
 *         if self.attr.ah_attr.is_global:
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_7ah_attr_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_7ah_attr_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_7ah_attr___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_7ah_attr___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_v_gid = NULL;
  struct __pyx_obj_7pyverbs_4addr_GID *__pyx_v_g = NULL;
  struct __pyx_obj_7pyverbs_4addr_GlobalRoute *__pyx_v_gr = NULL;
  struct __pyx_obj_7pyverbs_4addr_AHAttr *__pyx_v_ah = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":603
 *     @property
 *     def ah_attr(self):
 *         if self.attr.ah_attr.is_global:             # <<<<<<<<<<<<<<
 *             gid = gid_str(self.attr.ah_attr.grh.dgid._global.subnet_prefix,
 *                           self.attr.ah_attr.grh.dgid._global.interface_id)
*/
  __pyx_t_1 = (__pyx_v_self->attr.ah_attr.is_global != 0);
  if (__pyx_t_1) {

    /* "pyverbs/qp.pyx":604
 *     def ah_attr(self):
 *         if self.attr.ah_attr.is_global:
 *             gid = gid_str(self.attr.ah_attr.grh.dgid._global.subnet_prefix,             # <<<<<<<<<<<<<<
 *                           self.attr.ah_attr.grh.dgid._global.interface_id)
 *             g = GID(gid)
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_gid_str); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 604, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyLong_From_unsigned_long(__pyx_v_self->attr.ah_attr.grh.dgid.global.subnet_prefix); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 604, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "pyverbs/qp.pyx":605
 *         if self.attr.ah_attr.is_global:
 *             gid = gid_str(self.attr.ah_attr.grh.dgid._global.subnet_prefix,
 *                           self.attr.ah_attr.grh.dgid._global.interface_id)             # <<<<<<<<<<<<<<
 *             g = GID(gid)
 *             gr = GlobalRoute(flow_label=self.attr.ah_attr.grh.flow_label,
*/
    __pyx_t_6 = __Pyx_PyLong_From_unsigned_long(__pyx_v_self->attr.ah_attr.grh.dgid.global.interface_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 605, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_t_5, __pyx_t_6};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_7, (3-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 604, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __pyx_v_gid = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":606
 *             gid = gid_str(self.attr.ah_attr.grh.dgid._global.subnet_prefix,
 *                           self.attr.ah_attr.grh.dgid._global.interface_id)
 *             g = GID(gid)             # <<<<<<<<<<<<<<
 *             gr = GlobalRoute(flow_label=self.attr.ah_attr.grh.flow_label,
 *                              sgid_index=self.attr.ah_attr.grh.sgid_index,
*/
    __pyx_t_4 = NULL;
    __pyx_t_7 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_gid};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_4addr_GID, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 606, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_2);
    }
    __pyx_v_g = ((struct __pyx_obj_7pyverbs_4addr_GID *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":607
 *                           self.attr.ah_attr.grh.dgid._global.interface_id)
 *             g = GID(gid)
 *             gr = GlobalRoute(flow_label=self.attr.ah_attr.grh.flow_label,             # <<<<<<<<<<<<<<
 *                              sgid_index=self.attr.ah_attr.grh.sgid_index,
 *                              hop_limit=self.attr.ah_attr.grh.hop_limit, dgid=g,
*/
    __pyx_t_4 = NULL;
    __pyx_t_6 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.ah_attr.grh.flow_label); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 607, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "pyverbs/qp.pyx":608
 *             g = GID(gid)
 *             gr = GlobalRoute(flow_label=self.attr.ah_attr.grh.flow_label,
 *                              sgid_index=self.attr.ah_attr.grh.sgid_index,             # <<<<<<<<<<<<<<
 *                              hop_limit=self.attr.ah_attr.grh.hop_limit, dgid=g,
 *                              traffic_class=self.attr.ah_attr.grh.traffic_class)
*/
    __pyx_t_5 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.ah_attr.grh.sgid_index); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 608, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "pyverbs/qp.pyx":609
 *             gr = GlobalRoute(flow_label=self.attr.ah_attr.grh.flow_label,
 *                              sgid_index=self.attr.ah_attr.grh.sgid_index,
 *                              hop_limit=self.attr.ah_attr.grh.hop_limit, dgid=g,             # <<<<<<<<<<<<<<
 *                              traffic_class=self.attr.ah_attr.grh.traffic_class)
 *         else:
*/
    __pyx_t_3 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.ah_attr.grh.hop_limit); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 609, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "pyverbs/qp.pyx":610
 *                              sgid_index=self.attr.ah_attr.grh.sgid_index,
 *                              hop_limit=self.attr.ah_attr.grh.hop_limit, dgid=g,
 *                              traffic_class=self.attr.ah_attr.grh.traffic_class)             # <<<<<<<<<<<<<<
 *         else:
 *             gr = None
*/
    __pyx_t_8 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.ah_attr.grh.traffic_class); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 610, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = 1;
    {
      PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 5 : 0)] = {__pyx_t_4, NULL};
      __pyx_t_9 = __Pyx_MakeVectorcallBuilderKwds(5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 607, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_flow_label, __pyx_t_6, __pyx_t_9, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 607, __pyx_L1_error)
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_sgid_index, __pyx_t_5, __pyx_t_9, __pyx_callargs+1, 1) < (0)) __PYX_ERR(0, 607, __pyx_L1_error)
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_hop_limit, __pyx_t_3, __pyx_t_9, __pyx_callargs+1, 2) < (0)) __PYX_ERR(0, 607, __pyx_L1_error)
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_dgid, ((PyObject *)__pyx_v_g), __pyx_t_9, __pyx_callargs+1, 3) < (0)) __PYX_ERR(0, 607, __pyx_L1_error)
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_traffic_class, __pyx_t_8, __pyx_t_9, __pyx_callargs+1, 4) < (0)) __PYX_ERR(0, 607, __pyx_L1_error)
      __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_4addr_GlobalRoute, __pyx_callargs+__pyx_t_7, (1-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_9);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 607, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_2);
    }
    __pyx_v_gr = ((struct __pyx_obj_7pyverbs_4addr_GlobalRoute *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":603
 *     @property
 *     def ah_attr(self):
 *         if self.attr.ah_attr.is_global:             # <<<<<<<<<<<<<<
 *             gid = gid_str(self.attr.ah_attr.grh.dgid._global.subnet_prefix,
 *                           self.attr.ah_attr.grh.dgid._global.interface_id)
*/
    goto __pyx_L3;
  }

  /* "pyverbs/qp.pyx":612
 *                              traffic_class=self.attr.ah_attr.grh.traffic_class)
 *         else:
 *             gr = None             # <<<<<<<<<<<<<<
 *         ah = AHAttr(dlid=self.attr.ah_attr.dlid, sl=self.attr.ah_attr.sl,
 *                     port_num=self.attr.ah_attr.port_num,
*/
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __pyx_v_gr = ((struct __pyx_obj_7pyverbs_4addr_GlobalRoute *)Py_None);
  }
  __pyx_L3:;

  /* "pyverbs/qp.pyx":613
 *         else:
 *             gr = None
 *         ah = AHAttr(dlid=self.attr.ah_attr.dlid, sl=self.attr.ah_attr.sl,             # <<<<<<<<<<<<<<
 *                     port_num=self.attr.ah_attr.port_num,
 *                     src_path_bits=self.attr.ah_attr.src_path_bits,
*/
  __pyx_t_9 = NULL;
  __pyx_t_8 = __Pyx_PyLong_From_unsigned_short(__pyx_v_self->attr.ah_attr.dlid); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.ah_attr.sl); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "pyverbs/qp.pyx":614
 *             gr = None
 *         ah = AHAttr(dlid=self.attr.ah_attr.dlid, sl=self.attr.ah_attr.sl,
 *                     port_num=self.attr.ah_attr.port_num,             # <<<<<<<<<<<<<<
 *                     src_path_bits=self.attr.ah_attr.src_path_bits,
 *                     static_rate=self.attr.ah_attr.static_rate,
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.ah_attr.port_num); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "pyverbs/qp.pyx":615
 *         ah = AHAttr(dlid=self.attr.ah_attr.dlid, sl=self.attr.ah_attr.sl,
 *                     port_num=self.attr.ah_attr.port_num,
 *                     src_path_bits=self.attr.ah_attr.src_path_bits,             # <<<<<<<<<<<<<<
 *                     static_rate=self.attr.ah_attr.static_rate,
 *                     is_global=self.attr.ah_attr.is_global, gr=gr)
*/
  __pyx_t_6 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.ah_attr.src_path_bits); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "pyverbs/qp.pyx":616
 *                     port_num=self.attr.ah_attr.port_num,
 *                     src_path_bits=self.attr.ah_attr.src_path_bits,
 *                     static_rate=self.attr.ah_attr.static_rate,             # <<<<<<<<<<<<<<
 *                     is_global=self.attr.ah_attr.is_global, gr=gr)
 *         return ah
*/
  __pyx_t_4 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.ah_attr.static_rate); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "pyverbs/qp.pyx":617
 *                     src_path_bits=self.attr.ah_attr.src_path_bits,
 *                     static_rate=self.attr.ah_attr.static_rate,
 *                     is_global=self.attr.ah_attr.is_global, gr=gr)             # <<<<<<<<<<<<<<
 *         return ah
 * 
*/
  __pyx_t_10 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.ah_attr.is_global); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_7 = 1;
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 7 : 0)] = {__pyx_t_9, NULL};
    __pyx_t_11 = __Pyx_MakeVectorcallBuilderKwds(7); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_dlid, __pyx_t_8, __pyx_t_11, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 613, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_sl, __pyx_t_3, __pyx_t_11, __pyx_callargs+1, 1) < (0)) __PYX_ERR(0, 613, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_port_num, __pyx_t_5, __pyx_t_11, __pyx_callargs+1, 2) < (0)) __PYX_ERR(0, 613, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_src_path_bits, __pyx_t_6, __pyx_t_11, __pyx_callargs+1, 3) < (0)) __PYX_ERR(0, 613, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_static_rate, __pyx_t_4, __pyx_t_11, __pyx_callargs+1, 4) < (0)) __PYX_ERR(0, 613, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_is_global, __pyx_t_10, __pyx_t_11, __pyx_callargs+1, 5) < (0)) __PYX_ERR(0, 613, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_gr, ((PyObject *)__pyx_v_gr), __pyx_t_11, __pyx_callargs+1, 6) < (0)) __PYX_ERR(0, 613, __pyx_L1_error)
    __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_4addr_AHAttr, __pyx_callargs+__pyx_t_7, (1-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_11);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 613, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_2);
  }
  __pyx_v_ah = ((struct __pyx_obj_7pyverbs_4addr_AHAttr *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":618
 *                     static_rate=self.attr.ah_attr.static_rate,
 *                     is_global=self.attr.ah_attr.is_global, gr=gr)
 *         return ah             # <<<<<<<<<<<<<<
 * 
 *     @ah_attr.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_ah);
  __pyx_r = ((PyObject *)__pyx_v_ah);
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":601
 *         _copy_caps(val, self)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ah_attr(self):
 *         if self.attr.ah_attr.is_global:
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.ah_attr.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_gid);
  __Pyx_XDECREF((PyObject *)__pyx_v_g);
  __Pyx_XDECREF((PyObject *)__pyx_v_gr);
  __Pyx_XDECREF((PyObject *)__pyx_v_ah);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":620
 *         return ah
 * 
 *     @ah_attr.setter             # <<<<<<<<<<<<<<
 *     def ah_attr(self, val):
 *         self._copy_ah(val)
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_7ah_attr_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_7ah_attr_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_7ah_attr_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_7ah_attr_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "pyverbs/qp.pyx":622
 *     @ah_attr.setter
 *     def ah_attr(self, val):
 *         self._copy_ah(val)             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_val};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_copy_ah, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":620
 *         return ah
 * 
 *     @ah_attr.setter             # <<<<<<<<<<<<<<
 *     def ah_attr(self, val):
 *         self._copy_ah(val)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.ah_attr.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":624
 *         self._copy_ah(val)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def alt_ah_attr(self):
 *         if self.attr.alt_ah_attr.is_global:
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_11alt_ah_attr_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_11alt_ah_attr_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_11alt_ah_attr___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_11alt_ah_attr___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_v_gid = NULL;
  struct __pyx_obj_7pyverbs_4addr_GID *__pyx_v_g = NULL;
  struct __pyx_obj_7pyverbs_4addr_GlobalRoute *__pyx_v_gr = NULL;
  struct __pyx_obj_7pyverbs_4addr_AHAttr *__pyx_v_ah = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":626
 *     @property
 *     def alt_ah_attr(self):
 *         if self.attr.alt_ah_attr.is_global:             # <<<<<<<<<<<<<<
 *             gid = gid_str(self.attr.alt_ah_attr.grh.dgid._global.subnet_prefix,
 *                           self.attr.alt_ah_attr.grh.dgid._global.interface_id)
*/
  __pyx_t_1 = (__pyx_v_self->attr.alt_ah_attr.is_global != 0);
  if (__pyx_t_1) {

    /* "pyverbs/qp.pyx":627
 *     def alt_ah_attr(self):
 *         if self.attr.alt_ah_attr.is_global:
 *             gid = gid_str(self.attr.alt_ah_attr.grh.dgid._global.subnet_prefix,             # <<<<<<<<<<<<<<
 *                           self.attr.alt_ah_attr.grh.dgid._global.interface_id)
 *             g = GID(gid)
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_gid_str); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyLong_From_unsigned_long(__pyx_v_self->attr.alt_ah_attr.grh.dgid.global.subnet_prefix); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "pyverbs/qp.pyx":628
 *         if self.attr.alt_ah_attr.is_global:
 *             gid = gid_str(self.attr.alt_ah_attr.grh.dgid._global.subnet_prefix,
 *                           self.attr.alt_ah_attr.grh.dgid._global.interface_id)             # <<<<<<<<<<<<<<
 *             g = GID(gid)
 *             gr = GlobalRoute(flow_label=self.attr.alt_ah_attr.grh.flow_label,
*/
    __pyx_t_6 = __Pyx_PyLong_From_unsigned_long(__pyx_v_self->attr.alt_ah_attr.grh.dgid.global.interface_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 628, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_t_5, __pyx_t_6};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_7, (3-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 627, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __pyx_v_gid = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":629
 *             gid = gid_str(self.attr.alt_ah_attr.grh.dgid._global.subnet_prefix,
 *                           self.attr.alt_ah_attr.grh.dgid._global.interface_id)
 *             g = GID(gid)             # <<<<<<<<<<<<<<
 *             gr = GlobalRoute(flow_label=self.attr.alt_ah_attr.grh.flow_label,
 *                              sgid_index=self.attr.alt_ah_attr.grh.sgid_index,
*/
    __pyx_t_4 = NULL;
    __pyx_t_7 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_gid};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_4addr_GID, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 629, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_2);
    }
    __pyx_v_g = ((struct __pyx_obj_7pyverbs_4addr_GID *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":630
 *                           self.attr.alt_ah_attr.grh.dgid._global.interface_id)
 *             g = GID(gid)
 *             gr = GlobalRoute(flow_label=self.attr.alt_ah_attr.grh.flow_label,             # <<<<<<<<<<<<<<
 *                              sgid_index=self.attr.alt_ah_attr.grh.sgid_index,
 *                              hop_limit=self.attr.alt_ah_attr.grh.hop_limit,
*/
    __pyx_t_4 = NULL;
    __pyx_t_6 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.alt_ah_attr.grh.flow_label); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "pyverbs/qp.pyx":631
 *             g = GID(gid)
 *             gr = GlobalRoute(flow_label=self.attr.alt_ah_attr.grh.flow_label,
 *                              sgid_index=self.attr.alt_ah_attr.grh.sgid_index,             # <<<<<<<<<<<<<<
 *                              hop_limit=self.attr.alt_ah_attr.grh.hop_limit,
 *                              dgid=g,
*/
    __pyx_t_5 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.alt_ah_attr.grh.sgid_index); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 631, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "pyverbs/qp.pyx":632
 *             gr = GlobalRoute(flow_label=self.attr.alt_ah_attr.grh.flow_label,
 *                              sgid_index=self.attr.alt_ah_attr.grh.sgid_index,
 *                              hop_limit=self.attr.alt_ah_attr.grh.hop_limit,             # <<<<<<<<<<<<<<
 *                              dgid=g,
 *                              traffic_class=self.attr.alt_ah_attr.grh.traffic_class)
*/
    __pyx_t_3 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.alt_ah_attr.grh.hop_limit); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 632, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "pyverbs/qp.pyx":634
 *                              hop_limit=self.attr.alt_ah_attr.grh.hop_limit,
 *                              dgid=g,
 *                              traffic_class=self.attr.alt_ah_attr.grh.traffic_class)             # <<<<<<<<<<<<<<
 *         else:
 *             gr = None
*/
    __pyx_t_8 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.alt_ah_attr.grh.traffic_class); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 634, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = 1;
    {
      PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 5 : 0)] = {__pyx_t_4, NULL};
      __pyx_t_9 = __Pyx_MakeVectorcallBuilderKwds(5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 630, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_flow_label, __pyx_t_6, __pyx_t_9, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 630, __pyx_L1_error)
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_sgid_index, __pyx_t_5, __pyx_t_9, __pyx_callargs+1, 1) < (0)) __PYX_ERR(0, 630, __pyx_L1_error)
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_hop_limit, __pyx_t_3, __pyx_t_9, __pyx_callargs+1, 2) < (0)) __PYX_ERR(0, 630, __pyx_L1_error)
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_dgid, ((PyObject *)__pyx_v_g), __pyx_t_9, __pyx_callargs+1, 3) < (0)) __PYX_ERR(0, 630, __pyx_L1_error)
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_traffic_class, __pyx_t_8, __pyx_t_9, __pyx_callargs+1, 4) < (0)) __PYX_ERR(0, 630, __pyx_L1_error)
      __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_4addr_GlobalRoute, __pyx_callargs+__pyx_t_7, (1-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_9);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 630, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_2);
    }
    __pyx_v_gr = ((struct __pyx_obj_7pyverbs_4addr_GlobalRoute *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":626
 *     @property
 *     def alt_ah_attr(self):
 *         if self.attr.alt_ah_attr.is_global:             # <<<<<<<<<<<<<<
 *             gid = gid_str(self.attr.alt_ah_attr.grh.dgid._global.subnet_prefix,
 *                           self.attr.alt_ah_attr.grh.dgid._global.interface_id)
*/
    goto __pyx_L3;
  }

  /* "pyverbs/qp.pyx":636
 *                              traffic_class=self.attr.alt_ah_attr.grh.traffic_class)
 *         else:
 *             gr = None             # <<<<<<<<<<<<<<
 *         ah = AHAttr(dlid=self.attr.alt_ah_attr.dlid,
 *                     port_num=self.attr.ah_attr.port_num,
*/
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __pyx_v_gr = ((struct __pyx_obj_7pyverbs_4addr_GlobalRoute *)Py_None);
  }
  __pyx_L3:;

  /* "pyverbs/qp.pyx":637
 *         else:
 *             gr = None
 *         ah = AHAttr(dlid=self.attr.alt_ah_attr.dlid,             # <<<<<<<<<<<<<<
 *                     port_num=self.attr.ah_attr.port_num,
 *                     sl=self.attr.alt_ah_attr.sl,
*/
  __pyx_t_9 = NULL;
  __pyx_t_8 = __Pyx_PyLong_From_unsigned_short(__pyx_v_self->attr.alt_ah_attr.dlid); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "pyverbs/qp.pyx":638
 *             gr = None
 *         ah = AHAttr(dlid=self.attr.alt_ah_attr.dlid,
 *                     port_num=self.attr.ah_attr.port_num,             # <<<<<<<<<<<<<<
 *                     sl=self.attr.alt_ah_attr.sl,
 *                     src_path_bits=self.attr.alt_ah_attr.src_path_bits,
*/
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.ah_attr.port_num); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "pyverbs/qp.pyx":639
 *         ah = AHAttr(dlid=self.attr.alt_ah_attr.dlid,
 *                     port_num=self.attr.ah_attr.port_num,
 *                     sl=self.attr.alt_ah_attr.sl,             # <<<<<<<<<<<<<<
 *                     src_path_bits=self.attr.alt_ah_attr.src_path_bits,
 *                     static_rate=self.attr.alt_ah_attr.static_rate,
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.alt_ah_attr.sl); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "pyverbs/qp.pyx":640
 *                     port_num=self.attr.ah_attr.port_num,
 *                     sl=self.attr.alt_ah_attr.sl,
 *                     src_path_bits=self.attr.alt_ah_attr.src_path_bits,             # <<<<<<<<<<<<<<
 *                     static_rate=self.attr.alt_ah_attr.static_rate,
 *                     is_global=self.attr.alt_ah_attr.is_global, gr=gr)
*/
  __pyx_t_6 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.alt_ah_attr.src_path_bits); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "pyverbs/qp.pyx":641
 *                     sl=self.attr.alt_ah_attr.sl,
 *                     src_path_bits=self.attr.alt_ah_attr.src_path_bits,
 *                     static_rate=self.attr.alt_ah_attr.static_rate,             # <<<<<<<<<<<<<<
 *                     is_global=self.attr.alt_ah_attr.is_global, gr=gr)
 *         return ah
*/
  __pyx_t_4 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.alt_ah_attr.static_rate); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "pyverbs/qp.pyx":642
 *                     src_path_bits=self.attr.alt_ah_attr.src_path_bits,
 *                     static_rate=self.attr.alt_ah_attr.static_rate,
 *                     is_global=self.attr.alt_ah_attr.is_global, gr=gr)             # <<<<<<<<<<<<<<
 *         return ah
 * 
*/
  __pyx_t_10 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.alt_ah_attr.is_global); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_7 = 1;
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 7 : 0)] = {__pyx_t_9, NULL};
    __pyx_t_11 = __Pyx_MakeVectorcallBuilderKwds(7); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 637, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_dlid, __pyx_t_8, __pyx_t_11, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 637, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_port_num, __pyx_t_3, __pyx_t_11, __pyx_callargs+1, 1) < (0)) __PYX_ERR(0, 637, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_sl, __pyx_t_5, __pyx_t_11, __pyx_callargs+1, 2) < (0)) __PYX_ERR(0, 637, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_src_path_bits, __pyx_t_6, __pyx_t_11, __pyx_callargs+1, 3) < (0)) __PYX_ERR(0, 637, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_static_rate, __pyx_t_4, __pyx_t_11, __pyx_callargs+1, 4) < (0)) __PYX_ERR(0, 637, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_is_global, __pyx_t_10, __pyx_t_11, __pyx_callargs+1, 5) < (0)) __PYX_ERR(0, 637, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_gr, ((PyObject *)__pyx_v_gr), __pyx_t_11, __pyx_callargs+1, 6) < (0)) __PYX_ERR(0, 637, __pyx_L1_error)
    __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_4addr_AHAttr, __pyx_callargs+__pyx_t_7, (1-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_11);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 637, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_2);
  }
  __pyx_v_ah = ((struct __pyx_obj_7pyverbs_4addr_AHAttr *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":643
 *                     static_rate=self.attr.alt_ah_attr.static_rate,
 *                     is_global=self.attr.alt_ah_attr.is_global, gr=gr)
 *         return ah             # <<<<<<<<<<<<<<
 * 
 *     @alt_ah_attr.setter
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_ah);
  __pyx_r = ((PyObject *)__pyx_v_ah);
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":624
 *         self._copy_ah(val)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def alt_ah_attr(self):
 *         if self.attr.alt_ah_attr.is_global:
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.alt_ah_attr.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_gid);
  __Pyx_XDECREF((PyObject *)__pyx_v_g);
  __Pyx_XDECREF((PyObject *)__pyx_v_gr);
  __Pyx_XDECREF((PyObject *)__pyx_v_ah);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":645
 *         return ah
 * 
 *     @alt_ah_attr.setter             # <<<<<<<<<<<<<<
 *     def alt_ah_attr(self, val):
 *         self._copy_ah(val, True)
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_11alt_ah_attr_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_11alt_ah_attr_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_11alt_ah_attr_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_11alt_ah_attr_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "pyverbs/qp.pyx":647
 *     @alt_ah_attr.setter
 *     def alt_ah_attr(self, val):
 *         self._copy_ah(val, True)             # <<<<<<<<<<<<<<
 * 
 *     def _copy_ah(self, AHAttr ah_attr, is_alt=False):
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_v_val, Py_True};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_copy_ah, __pyx_callargs+__pyx_t_3, (3-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":645
 *         return ah
 * 
 *     @alt_ah_attr.setter             # <<<<<<<<<<<<<<
 *     def alt_ah_attr(self, val):
 *         self._copy_ah(val, True)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.alt_ah_attr.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":649
 *         self._copy_ah(val, True)
 * 
 *     def _copy_ah(self, AHAttr ah_attr, is_alt=False):             # <<<<<<<<<<<<<<
 *         if ah_attr is None:
 *             return
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_3_copy_ah(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_6QPAttr_3_copy_ah = {"_copy_ah", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_6QPAttr_3_copy_ah, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_3_copy_ah(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_4addr_AHAttr *__pyx_v_ah_attr = 0;
  PyObject *__pyx_v_is_alt = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_copy_ah (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_ah_attr,&__pyx_mstate_global->__pyx_n_u_is_alt,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 649, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 649, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 649, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_copy_ah", 0) < (0)) __PYX_ERR(0, 649, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_False));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_copy_ah", 0, 1, 2, i); __PYX_ERR(0, 649, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 649, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 649, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_False));
    }
    __pyx_v_ah_attr = ((struct __pyx_obj_7pyverbs_4addr_AHAttr *)values[0]);
    __pyx_v_is_alt = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_copy_ah", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 649, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPAttr._copy_ah", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_ah_attr), __pyx_mstate_global->__pyx_ptype_7pyverbs_4addr_AHAttr, 1, "ah_attr", 0))) __PYX_ERR(0, 649, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_2_copy_ah(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), __pyx_v_ah_attr, __pyx_v_is_alt);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_2_copy_ah(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, struct __pyx_obj_7pyverbs_4addr_AHAttr *__pyx_v_ah_attr, PyObject *__pyx_v_is_alt) {
  long __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  long __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned char __pyx_t_5;
  unsigned short __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_copy_ah", 0);

  /* "pyverbs/qp.pyx":650
 * 
 *     def _copy_ah(self, AHAttr ah_attr, is_alt=False):
 *         if ah_attr is None:             # <<<<<<<<<<<<<<
 *             return
 *         if not is_alt:
*/
  __pyx_t_1 = (((PyObject *)__pyx_v_ah_attr) == Py_None);
  if (__pyx_t_1) {

    /* "pyverbs/qp.pyx":651
 *     def _copy_ah(self, AHAttr ah_attr, is_alt=False):
 *         if ah_attr is None:
 *             return             # <<<<<<<<<<<<<<
 *         if not is_alt:
 *             for i in range(16):
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "pyverbs/qp.pyx":650
 * 
 *     def _copy_ah(self, AHAttr ah_attr, is_alt=False):
 *         if ah_attr is None:             # <<<<<<<<<<<<<<
 *             return
 *         if not is_alt:
*/
  }

  /* "pyverbs/qp.pyx":652
 *         if ah_attr is None:
 *             return
 *         if not is_alt:             # <<<<<<<<<<<<<<
 *             for i in range(16):
 *                 self.attr.ah_attr.grh.dgid.raw[i] = \
*/
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_is_alt); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 652, __pyx_L1_error)
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "pyverbs/qp.pyx":653
 *             return
 *         if not is_alt:
 *             for i in range(16):             # <<<<<<<<<<<<<<
 *                 self.attr.ah_attr.grh.dgid.raw[i] = \
 *                     ah_attr.ah_attr.grh.dgid.raw[i]
*/
    for (__pyx_t_3 = 0; __pyx_t_3 < 16; __pyx_t_3+=1) {
      __pyx_v_i = __pyx_t_3;

      /* "pyverbs/qp.pyx":654
 *         if not is_alt:
 *             for i in range(16):
 *                 self.attr.ah_attr.grh.dgid.raw[i] = \             # <<<<<<<<<<<<<<
 *                     ah_attr.ah_attr.grh.dgid.raw[i]
 *             self.attr.ah_attr.grh.flow_label = ah_attr.ah_attr.grh.flow_label
*/
      (__pyx_v_self->attr.ah_attr.grh.dgid.raw[__pyx_v_i]) = (__pyx_v_ah_attr->ah_attr.grh.dgid.raw[__pyx_v_i]);
    }

    /* "pyverbs/qp.pyx":656
 *                 self.attr.ah_attr.grh.dgid.raw[i] = \
 *                     ah_attr.ah_attr.grh.dgid.raw[i]
 *             self.attr.ah_attr.grh.flow_label = ah_attr.ah_attr.grh.flow_label             # <<<<<<<<<<<<<<
 *             self.attr.ah_attr.grh.sgid_index = ah_attr.ah_attr.grh.sgid_index
 *             self.attr.ah_attr.grh.hop_limit = ah_attr.ah_attr.grh.hop_limit
*/
    __pyx_t_4 = __pyx_v_ah_attr->ah_attr.grh.flow_label;
    __pyx_v_self->attr.ah_attr.grh.flow_label = __pyx_t_4;

    /* "pyverbs/qp.pyx":657
 *                     ah_attr.ah_attr.grh.dgid.raw[i]
 *             self.attr.ah_attr.grh.flow_label = ah_attr.ah_attr.grh.flow_label
 *             self.attr.ah_attr.grh.sgid_index = ah_attr.ah_attr.grh.sgid_index             # <<<<<<<<<<<<<<
 *             self.attr.ah_attr.grh.hop_limit = ah_attr.ah_attr.grh.hop_limit
 *             self.attr.ah_attr.grh.traffic_class = \
*/
    __pyx_t_5 = __pyx_v_ah_attr->ah_attr.grh.sgid_index;
    __pyx_v_self->attr.ah_attr.grh.sgid_index = __pyx_t_5;

    /* "pyverbs/qp.pyx":658
 *             self.attr.ah_attr.grh.flow_label = ah_attr.ah_attr.grh.flow_label
 *             self.attr.ah_attr.grh.sgid_index = ah_attr.ah_attr.grh.sgid_index
 *             self.attr.ah_attr.grh.hop_limit = ah_attr.ah_attr.grh.hop_limit             # <<<<<<<<<<<<<<
 *             self.attr.ah_attr.grh.traffic_class = \
 *                 ah_attr.ah_attr.grh.traffic_class
*/
    __pyx_t_5 = __pyx_v_ah_attr->ah_attr.grh.hop_limit;
    __pyx_v_self->attr.ah_attr.grh.hop_limit = __pyx_t_5;

    /* "pyverbs/qp.pyx":660
 *             self.attr.ah_attr.grh.hop_limit = ah_attr.ah_attr.grh.hop_limit
 *             self.attr.ah_attr.grh.traffic_class = \
 *                 ah_attr.ah_attr.grh.traffic_class             # <<<<<<<<<<<<<<
 *             self.attr.ah_attr.dlid = ah_attr.ah_attr.dlid
 *             self.attr.ah_attr.sl = ah_attr.ah_attr.sl
*/
    __pyx_t_5 = __pyx_v_ah_attr->ah_attr.grh.traffic_class;

    /* "pyverbs/qp.pyx":659
 *             self.attr.ah_attr.grh.sgid_index = ah_attr.ah_attr.grh.sgid_index
 *             self.attr.ah_attr.grh.hop_limit = ah_attr.ah_attr.grh.hop_limit
 *             self.attr.ah_attr.grh.traffic_class = \             # <<<<<<<<<<<<<<
 *                 ah_attr.ah_attr.grh.traffic_class
 *             self.attr.ah_attr.dlid = ah_attr.ah_attr.dlid
*/
    __pyx_v_self->attr.ah_attr.grh.traffic_class = __pyx_t_5;

    /* "pyverbs/qp.pyx":661
 *             self.attr.ah_attr.grh.traffic_class = \
 *                 ah_attr.ah_attr.grh.traffic_class
 *             self.attr.ah_attr.dlid = ah_attr.ah_attr.dlid             # <<<<<<<<<<<<<<
 *             self.attr.ah_attr.sl = ah_attr.ah_attr.sl
 *             self.attr.ah_attr.src_path_bits = ah_attr.ah_attr.src_path_bits
*/
    __pyx_t_6 = __pyx_v_ah_attr->ah_attr.dlid;
    __pyx_v_self->attr.ah_attr.dlid = __pyx_t_6;

    /* "pyverbs/qp.pyx":662
 *                 ah_attr.ah_attr.grh.traffic_class
 *             self.attr.ah_attr.dlid = ah_attr.ah_attr.dlid
 *             self.attr.ah_attr.sl = ah_attr.ah_attr.sl             # <<<<<<<<<<<<<<
 *             self.attr.ah_attr.src_path_bits = ah_attr.ah_attr.src_path_bits
 *             self.attr.ah_attr.static_rate = ah_attr.ah_attr.static_rate
*/
    __pyx_t_5 = __pyx_v_ah_attr->ah_attr.sl;
    __pyx_v_self->attr.ah_attr.sl = __pyx_t_5;

    /* "pyverbs/qp.pyx":663
 *             self.attr.ah_attr.dlid = ah_attr.ah_attr.dlid
 *             self.attr.ah_attr.sl = ah_attr.ah_attr.sl
 *             self.attr.ah_attr.src_path_bits = ah_attr.ah_attr.src_path_bits             # <<<<<<<<<<<<<<
 *             self.attr.ah_attr.static_rate = ah_attr.ah_attr.static_rate
 *             self.attr.ah_attr.is_global = ah_attr.ah_attr.is_global
*/
    __pyx_t_5 = __pyx_v_ah_attr->ah_attr.src_path_bits;
    __pyx_v_self->attr.ah_attr.src_path_bits = __pyx_t_5;

    /* "pyverbs/qp.pyx":664
 *             self.attr.ah_attr.sl = ah_attr.ah_attr.sl
 *             self.attr.ah_attr.src_path_bits = ah_attr.ah_attr.src_path_bits
 *             self.attr.ah_attr.static_rate = ah_attr.ah_attr.static_rate             # <<<<<<<<<<<<<<
 *             self.attr.ah_attr.is_global = ah_attr.ah_attr.is_global
 *             self.attr.ah_attr.port_num = ah_attr.ah_attr.port_num
*/
    __pyx_t_5 = __pyx_v_ah_attr->ah_attr.static_rate;
    __pyx_v_self->attr.ah_attr.static_rate = __pyx_t_5;

    /* "pyverbs/qp.pyx":665
 *             self.attr.ah_attr.src_path_bits = ah_attr.ah_attr.src_path_bits
 *             self.attr.ah_attr.static_rate = ah_attr.ah_attr.static_rate
 *             self.attr.ah_attr.is_global = ah_attr.ah_attr.is_global             # <<<<<<<<<<<<<<
 *             self.attr.ah_attr.port_num = ah_attr.ah_attr.port_num
 *         else:
*/
    __pyx_t_5 = __pyx_v_ah_attr->ah_attr.is_global;
    __pyx_v_self->attr.ah_attr.is_global = __pyx_t_5;

    /* "pyverbs/qp.pyx":666
 *             self.attr.ah_attr.static_rate = ah_attr.ah_attr.static_rate
 *             self.attr.ah_attr.is_global = ah_attr.ah_attr.is_global
 *             self.attr.ah_attr.port_num = ah_attr.ah_attr.port_num             # <<<<<<<<<<<<<<
 *         else:
 *             for i in range(16):
*/
    __pyx_t_5 = __pyx_v_ah_attr->ah_attr.port_num;
    __pyx_v_self->attr.ah_attr.port_num = __pyx_t_5;

    /* "pyverbs/qp.pyx":652
 *         if ah_attr is None:
 *             return
 *         if not is_alt:             # <<<<<<<<<<<<<<
 *             for i in range(16):
 *                 self.attr.ah_attr.grh.dgid.raw[i] = \
*/
    goto __pyx_L4;
  }

  /* "pyverbs/qp.pyx":668
 *             self.attr.ah_attr.port_num = ah_attr.ah_attr.port_num
 *         else:
 *             for i in range(16):             # <<<<<<<<<<<<<<
 *                 self.attr.alt_ah_attr.grh.dgid.raw[i] = \
 *                     ah_attr.ah_attr.grh.dgid.raw[i]
*/
  /*else*/ {
    for (__pyx_t_3 = 0; __pyx_t_3 < 16; __pyx_t_3+=1) {
      __pyx_v_i = __pyx_t_3;

      /* "pyverbs/qp.pyx":669
 *         else:
 *             for i in range(16):
 *                 self.attr.alt_ah_attr.grh.dgid.raw[i] = \             # <<<<<<<<<<<<<<
 *                     ah_attr.ah_attr.grh.dgid.raw[i]
 *             self.attr.alt_ah_attr.grh.flow_label = \
*/
      (__pyx_v_self->attr.alt_ah_attr.grh.dgid.raw[__pyx_v_i]) = (__pyx_v_ah_attr->ah_attr.grh.dgid.raw[__pyx_v_i]);
    }

    /* "pyverbs/qp.pyx":672
 *                     ah_attr.ah_attr.grh.dgid.raw[i]
 *             self.attr.alt_ah_attr.grh.flow_label = \
 *                 ah_attr.ah_attr.grh.flow_label             # <<<<<<<<<<<<<<
 *             self.attr.alt_ah_attr.grh.sgid_index = \
 *                 ah_attr.ah_attr.grh.sgid_index
*/
    __pyx_t_4 = __pyx_v_ah_attr->ah_attr.grh.flow_label;

    /* "pyverbs/qp.pyx":671
 *                 self.attr.alt_ah_attr.grh.dgid.raw[i] = \
 *                     ah_attr.ah_attr.grh.dgid.raw[i]
 *             self.attr.alt_ah_attr.grh.flow_label = \             # <<<<<<<<<<<<<<
 *                 ah_attr.ah_attr.grh.flow_label
 *             self.attr.alt_ah_attr.grh.sgid_index = \
*/
    __pyx_v_self->attr.alt_ah_attr.grh.flow_label = __pyx_t_4;

    /* "pyverbs/qp.pyx":674
 *                 ah_attr.ah_attr.grh.flow_label
 *             self.attr.alt_ah_attr.grh.sgid_index = \
 *                 ah_attr.ah_attr.grh.sgid_index             # <<<<<<<<<<<<<<
 *             self.attr.alt_ah_attr.grh.hop_limit = ah_attr.ah_attr.grh.hop_limit
 *             self.attr.alt_ah_attr.grh.traffic_class = \
*/
    __pyx_t_5 = __pyx_v_ah_attr->ah_attr.grh.sgid_index;

    /* "pyverbs/qp.pyx":673
 *             self.attr.alt_ah_attr.grh.flow_label = \
 *                 ah_attr.ah_attr.grh.flow_label
 *             self.attr.alt_ah_attr.grh.sgid_index = \             # <<<<<<<<<<<<<<
 *                 ah_attr.ah_attr.grh.sgid_index
 *             self.attr.alt_ah_attr.grh.hop_limit = ah_attr.ah_attr.grh.hop_limit
*/
    __pyx_v_self->attr.alt_ah_attr.grh.sgid_index = __pyx_t_5;

    /* "pyverbs/qp.pyx":675
 *             self.attr.alt_ah_attr.grh.sgid_index = \
 *                 ah_attr.ah_attr.grh.sgid_index
 *             self.attr.alt_ah_attr.grh.hop_limit = ah_attr.ah_attr.grh.hop_limit             # <<<<<<<<<<<<<<
 *             self.attr.alt_ah_attr.grh.traffic_class = \
 *                 ah_attr.ah_attr.grh.traffic_class
*/
    __pyx_t_5 = __pyx_v_ah_attr->ah_attr.grh.hop_limit;
    __pyx_v_self->attr.alt_ah_attr.grh.hop_limit = __pyx_t_5;

    /* "pyverbs/qp.pyx":677
 *             self.attr.alt_ah_attr.grh.hop_limit = ah_attr.ah_attr.grh.hop_limit
 *             self.attr.alt_ah_attr.grh.traffic_class = \
 *                 ah_attr.ah_attr.grh.traffic_class             # <<<<<<<<<<<<<<
 *             self.attr.alt_ah_attr.dlid = ah_attr.ah_attr.dlid
 *             self.attr.alt_ah_attr.sl = ah_attr.ah_attr.sl
*/
    __pyx_t_5 = __pyx_v_ah_attr->ah_attr.grh.traffic_class;

    /* "pyverbs/qp.pyx":676
 *                 ah_attr.ah_attr.grh.sgid_index
 *             self.attr.alt_ah_attr.grh.hop_limit = ah_attr.ah_attr.grh.hop_limit
 *             self.attr.alt_ah_attr.grh.traffic_class = \             # <<<<<<<<<<<<<<
 *                 ah_attr.ah_attr.grh.traffic_class
 *             self.attr.alt_ah_attr.dlid = ah_attr.ah_attr.dlid
*/
    __pyx_v_self->attr.alt_ah_attr.grh.traffic_class = __pyx_t_5;

    /* "pyverbs/qp.pyx":678
 *             self.attr.alt_ah_attr.grh.traffic_class = \
 *                 ah_attr.ah_attr.grh.traffic_class
 *             self.attr.alt_ah_attr.dlid = ah_attr.ah_attr.dlid             # <<<<<<<<<<<<<<
 *             self.attr.alt_ah_attr.sl = ah_attr.ah_attr.sl
 *             self.attr.alt_ah_attr.src_path_bits = ah_attr.ah_attr.src_path_bits
*/
    __pyx_t_6 = __pyx_v_ah_attr->ah_attr.dlid;
    __pyx_v_self->attr.alt_ah_attr.dlid = __pyx_t_6;

    /* "pyverbs/qp.pyx":679
 *                 ah_attr.ah_attr.grh.traffic_class
 *             self.attr.alt_ah_attr.dlid = ah_attr.ah_attr.dlid
 *             self.attr.alt_ah_attr.sl = ah_attr.ah_attr.sl             # <<<<<<<<<<<<<<
 *             self.attr.alt_ah_attr.src_path_bits = ah_attr.ah_attr.src_path_bits
 *             self.attr.alt_ah_attr.static_rate = ah_attr.ah_attr.static_rate
*/
    __pyx_t_5 = __pyx_v_ah_attr->ah_attr.sl;
    __pyx_v_self->attr.alt_ah_attr.sl = __pyx_t_5;

    /* "pyverbs/qp.pyx":680
 *             self.attr.alt_ah_attr.dlid = ah_attr.ah_attr.dlid
 *             self.attr.alt_ah_attr.sl = ah_attr.ah_attr.sl
 *             self.attr.alt_ah_attr.src_path_bits = ah_attr.ah_attr.src_path_bits             # <<<<<<<<<<<<<<
 *             self.attr.alt_ah_attr.static_rate = ah_attr.ah_attr.static_rate
 *             self.attr.alt_ah_attr.is_global = ah_attr.ah_attr.is_global
*/
    __pyx_t_5 = __pyx_v_ah_attr->ah_attr.src_path_bits;
    __pyx_v_self->attr.alt_ah_attr.src_path_bits = __pyx_t_5;

    /* "pyverbs/qp.pyx":681
 *             self.attr.alt_ah_attr.sl = ah_attr.ah_attr.sl
 *             self.attr.alt_ah_attr.src_path_bits = ah_attr.ah_attr.src_path_bits
 *             self.attr.alt_ah_attr.static_rate = ah_attr.ah_attr.static_rate             # <<<<<<<<<<<<<<
 *             self.attr.alt_ah_attr.is_global = ah_attr.ah_attr.is_global
 *             self.attr.alt_ah_attr.port_num = ah_attr.ah_attr.port_num
*/
    __pyx_t_5 = __pyx_v_ah_attr->ah_attr.static_rate;
    __pyx_v_self->attr.alt_ah_attr.static_rate = __pyx_t_5;

    /* "pyverbs/qp.pyx":682
 *             self.attr.alt_ah_attr.src_path_bits = ah_attr.ah_attr.src_path_bits
 *             self.attr.alt_ah_attr.static_rate = ah_attr.ah_attr.static_rate
 *             self.attr.alt_ah_attr.is_global = ah_attr.ah_attr.is_global             # <<<<<<<<<<<<<<
 *             self.attr.alt_ah_attr.port_num = ah_attr.ah_attr.port_num
 * 
*/
    __pyx_t_5 = __pyx_v_ah_attr->ah_attr.is_global;
    __pyx_v_self->attr.alt_ah_attr.is_global = __pyx_t_5;

    /* "pyverbs/qp.pyx":683
 *             self.attr.alt_ah_attr.static_rate = ah_attr.ah_attr.static_rate
 *             self.attr.alt_ah_attr.is_global = ah_attr.ah_attr.is_global
 *             self.attr.alt_ah_attr.port_num = ah_attr.ah_attr.port_num             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
    __pyx_t_5 = __pyx_v_ah_attr->ah_attr.port_num;
    __pyx_v_self->attr.alt_ah_attr.port_num = __pyx_t_5;
  }
  __pyx_L4:;

  /* "pyverbs/qp.pyx":649
 *         self._copy_ah(val, True)
 * 
 *     def _copy_ah(self, AHAttr ah_attr, is_alt=False):             # <<<<<<<<<<<<<<
 *         if ah_attr is None:
 *             return
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr._copy_ah", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":685
 *             self.attr.alt_ah_attr.port_num = ah_attr.ah_attr.port_num
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def pkey_index(self):
 *         return self.attr.pkey_index
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_10pkey_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_10pkey_index_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_10pkey_index___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_10pkey_index___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":687
 *     @property
 *     def pkey_index(self):
 *         return self.attr.pkey_index             # <<<<<<<<<<<<<<
 *     @pkey_index.setter
 *     def pkey_index(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_short(__pyx_v_self->attr.pkey_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":685
 *             self.attr.alt_ah_attr.port_num = ah_attr.ah_attr.port_num
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def pkey_index(self):
 *         return self.attr.pkey_index
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.pkey_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":688
 *     def pkey_index(self):
 *         return self.attr.pkey_index
 *     @pkey_index.setter             # <<<<<<<<<<<<<<
 *     def pkey_index(self, val):
 *         self.attr.pkey_index = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_10pkey_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_10pkey_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_10pkey_index_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_10pkey_index_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned short __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":690
 *     @pkey_index.setter
 *     def pkey_index(self, val):
 *         self.attr.pkey_index = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_short(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned short)-1) && PyErr_Occurred())) __PYX_ERR(0, 690, __pyx_L1_error)
  __pyx_v_self->attr.pkey_index = __pyx_t_1;

  /* "pyverbs/qp.pyx":688
 *     def pkey_index(self):
 *         return self.attr.pkey_index
 *     @pkey_index.setter             # <<<<<<<<<<<<<<
 *     def pkey_index(self, val):
 *         self.attr.pkey_index = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.pkey_index.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":692
 *         self.attr.pkey_index = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def alt_pkey_index(self):
 *         return self.attr.alt_pkey_index
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_14alt_pkey_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_14alt_pkey_index_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_14alt_pkey_index___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_14alt_pkey_index___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":694
 *     @property
 *     def alt_pkey_index(self):
 *         return self.attr.alt_pkey_index             # <<<<<<<<<<<<<<
 *     @alt_pkey_index.setter
 *     def alt_pkey_index(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_short(__pyx_v_self->attr.alt_pkey_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":692
 *         self.attr.pkey_index = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def alt_pkey_index(self):
 *         return self.attr.alt_pkey_index
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.alt_pkey_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":695
 *     def alt_pkey_index(self):
 *         return self.attr.alt_pkey_index
 *     @alt_pkey_index.setter             # <<<<<<<<<<<<<<
 *     def alt_pkey_index(self, val):
 *         self.attr.alt_pkey_index = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_14alt_pkey_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_14alt_pkey_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_14alt_pkey_index_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_14alt_pkey_index_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned short __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":697
 *     @alt_pkey_index.setter
 *     def alt_pkey_index(self, val):
 *         self.attr.alt_pkey_index = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_short(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned short)-1) && PyErr_Occurred())) __PYX_ERR(0, 697, __pyx_L1_error)
  __pyx_v_self->attr.alt_pkey_index = __pyx_t_1;

  /* "pyverbs/qp.pyx":695
 *     def alt_pkey_index(self):
 *         return self.attr.alt_pkey_index
 *     @alt_pkey_index.setter             # <<<<<<<<<<<<<<
 *     def alt_pkey_index(self, val):
 *         self.attr.alt_pkey_index = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.alt_pkey_index.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":699
 *         self.attr.alt_pkey_index = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def en_sqd_async_notify(self):
 *         return self.attr.en_sqd_async_notify
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_19en_sqd_async_notify_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_19en_sqd_async_notify_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_19en_sqd_async_notify___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_19en_sqd_async_notify___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":701
 *     @property
 *     def en_sqd_async_notify(self):
 *         return self.attr.en_sqd_async_notify             # <<<<<<<<<<<<<<
 *     @en_sqd_async_notify.setter
 *     def en_sqd_async_notify(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.en_sqd_async_notify); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":699
 *         self.attr.alt_pkey_index = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def en_sqd_async_notify(self):
 *         return self.attr.en_sqd_async_notify
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.en_sqd_async_notify.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":702
 *     def en_sqd_async_notify(self):
 *         return self.attr.en_sqd_async_notify
 *     @en_sqd_async_notify.setter             # <<<<<<<<<<<<<<
 *     def en_sqd_async_notify(self, val):
 *         self.attr.en_sqd_async_notify = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_19en_sqd_async_notify_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_19en_sqd_async_notify_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_19en_sqd_async_notify_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_19en_sqd_async_notify_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned char __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":704
 *     @en_sqd_async_notify.setter
 *     def en_sqd_async_notify(self, val):
 *         self.attr.en_sqd_async_notify = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_char(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 704, __pyx_L1_error)
  __pyx_v_self->attr.en_sqd_async_notify = __pyx_t_1;

  /* "pyverbs/qp.pyx":702
 *     def en_sqd_async_notify(self):
 *         return self.attr.en_sqd_async_notify
 *     @en_sqd_async_notify.setter             # <<<<<<<<<<<<<<
 *     def en_sqd_async_notify(self, val):
 *         self.attr.en_sqd_async_notify = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.en_sqd_async_notify.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":706
 *         self.attr.en_sqd_async_notify = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def sq_draining(self):
 *         return self.attr.sq_draining
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_11sq_draining_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_11sq_draining_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_11sq_draining___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_11sq_draining___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":708
 *     @property
 *     def sq_draining(self):
 *         return self.attr.sq_draining             # <<<<<<<<<<<<<<
 *     @sq_draining.setter
 *     def sq_draining(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.sq_draining); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":706
 *         self.attr.en_sqd_async_notify = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def sq_draining(self):
 *         return self.attr.sq_draining
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.sq_draining.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":709
 *     def sq_draining(self):
 *         return self.attr.sq_draining
 *     @sq_draining.setter             # <<<<<<<<<<<<<<
 *     def sq_draining(self, val):
 *         self.attr.sq_draining = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_11sq_draining_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_11sq_draining_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_11sq_draining_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_11sq_draining_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned char __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":711
 *     @sq_draining.setter
 *     def sq_draining(self, val):
 *         self.attr.sq_draining = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_char(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 711, __pyx_L1_error)
  __pyx_v_self->attr.sq_draining = __pyx_t_1;

  /* "pyverbs/qp.pyx":709
 *     def sq_draining(self):
 *         return self.attr.sq_draining
 *     @sq_draining.setter             # <<<<<<<<<<<<<<
 *     def sq_draining(self, val):
 *         self.attr.sq_draining = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.sq_draining.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":713
 *         self.attr.sq_draining = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_rd_atomic(self):
 *         return self.attr.max_rd_atomic
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_13max_rd_atomic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_13max_rd_atomic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_13max_rd_atomic___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_13max_rd_atomic___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":715
 *     @property
 *     def max_rd_atomic(self):
 *         return self.attr.max_rd_atomic             # <<<<<<<<<<<<<<
 *     @max_rd_atomic.setter
 *     def max_rd_atomic(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.max_rd_atomic); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":713
 *         self.attr.sq_draining = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_rd_atomic(self):
 *         return self.attr.max_rd_atomic
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.max_rd_atomic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":716
 *     def max_rd_atomic(self):
 *         return self.attr.max_rd_atomic
 *     @max_rd_atomic.setter             # <<<<<<<<<<<<<<
 *     def max_rd_atomic(self, val):
 *         self.attr.max_rd_atomic = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_13max_rd_atomic_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_13max_rd_atomic_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_13max_rd_atomic_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_13max_rd_atomic_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned char __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":718
 *     @max_rd_atomic.setter
 *     def max_rd_atomic(self, val):
 *         self.attr.max_rd_atomic = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_char(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 718, __pyx_L1_error)
  __pyx_v_self->attr.max_rd_atomic = __pyx_t_1;

  /* "pyverbs/qp.pyx":716
 *     def max_rd_atomic(self):
 *         return self.attr.max_rd_atomic
 *     @max_rd_atomic.setter             # <<<<<<<<<<<<<<
 *     def max_rd_atomic(self, val):
 *         self.attr.max_rd_atomic = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.max_rd_atomic.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":720
 *         self.attr.max_rd_atomic = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_dest_rd_atomic(self):
 *         return self.attr.max_dest_rd_atomic
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_18max_dest_rd_atomic_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_18max_dest_rd_atomic_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_18max_dest_rd_atomic___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_18max_dest_rd_atomic___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":722
 *     @property
 *     def max_dest_rd_atomic(self):
 *         return self.attr.max_dest_rd_atomic             # <<<<<<<<<<<<<<
 *     @max_dest_rd_atomic.setter
 *     def max_dest_rd_atomic(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.max_dest_rd_atomic); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":720
 *         self.attr.max_rd_atomic = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_dest_rd_atomic(self):
 *         return self.attr.max_dest_rd_atomic
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.max_dest_rd_atomic.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":723
 *     def max_dest_rd_atomic(self):
 *         return self.attr.max_dest_rd_atomic
 *     @max_dest_rd_atomic.setter             # <<<<<<<<<<<<<<
 *     def max_dest_rd_atomic(self, val):
 *         self.attr.max_dest_rd_atomic = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_18max_dest_rd_atomic_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_18max_dest_rd_atomic_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_18max_dest_rd_atomic_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_18max_dest_rd_atomic_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned char __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":725
 *     @max_dest_rd_atomic.setter
 *     def max_dest_rd_atomic(self, val):
 *         self.attr.max_dest_rd_atomic = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_char(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 725, __pyx_L1_error)
  __pyx_v_self->attr.max_dest_rd_atomic = __pyx_t_1;

  /* "pyverbs/qp.pyx":723
 *     def max_dest_rd_atomic(self):
 *         return self.attr.max_dest_rd_atomic
 *     @max_dest_rd_atomic.setter             # <<<<<<<<<<<<<<
 *     def max_dest_rd_atomic(self, val):
 *         self.attr.max_dest_rd_atomic = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.max_dest_rd_atomic.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":727
 *         self.attr.max_dest_rd_atomic = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def min_rnr_timer(self):
 *         return self.attr.min_rnr_timer
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_13min_rnr_timer_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_13min_rnr_timer_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_13min_rnr_timer___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_13min_rnr_timer___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":729
 *     @property
 *     def min_rnr_timer(self):
 *         return self.attr.min_rnr_timer             # <<<<<<<<<<<<<<
 *     @min_rnr_timer.setter
 *     def min_rnr_timer(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.min_rnr_timer); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":727
 *         self.attr.max_dest_rd_atomic = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def min_rnr_timer(self):
 *         return self.attr.min_rnr_timer
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.min_rnr_timer.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":730
 *     def min_rnr_timer(self):
 *         return self.attr.min_rnr_timer
 *     @min_rnr_timer.setter             # <<<<<<<<<<<<<<
 *     def min_rnr_timer(self, val):
 *         self.attr.min_rnr_timer = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_13min_rnr_timer_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_13min_rnr_timer_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_13min_rnr_timer_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_13min_rnr_timer_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned char __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":732
 *     @min_rnr_timer.setter
 *     def min_rnr_timer(self, val):
 *         self.attr.min_rnr_timer = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_char(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 732, __pyx_L1_error)
  __pyx_v_self->attr.min_rnr_timer = __pyx_t_1;

  /* "pyverbs/qp.pyx":730
 *     def min_rnr_timer(self):
 *         return self.attr.min_rnr_timer
 *     @min_rnr_timer.setter             # <<<<<<<<<<<<<<
 *     def min_rnr_timer(self, val):
 *         self.attr.min_rnr_timer = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.min_rnr_timer.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":734
 *         self.attr.min_rnr_timer = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def port_num(self):
 *         return self.attr.port_num
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_8port_num_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_8port_num_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_8port_num___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_8port_num___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":736
 *     @property
 *     def port_num(self):
 *         return self.attr.port_num             # <<<<<<<<<<<<<<
 *     @port_num.setter
 *     def port_num(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.port_num); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":734
 *         self.attr.min_rnr_timer = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def port_num(self):
 *         return self.attr.port_num
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.port_num.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":737
 *     def port_num(self):
 *         return self.attr.port_num
 *     @port_num.setter             # <<<<<<<<<<<<<<
 *     def port_num(self, val):
 *         self.attr.port_num = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_8port_num_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_8port_num_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_8port_num_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_8port_num_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned char __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":739
 *     @port_num.setter
 *     def port_num(self, val):
 *         self.attr.port_num = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_char(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 739, __pyx_L1_error)
  __pyx_v_self->attr.port_num = __pyx_t_1;

  /* "pyverbs/qp.pyx":737
 *     def port_num(self):
 *         return self.attr.port_num
 *     @port_num.setter             # <<<<<<<<<<<<<<
 *     def port_num(self, val):
 *         self.attr.port_num = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.port_num.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":741
 *         self.attr.port_num = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def timeout(self):
 *         return self.attr.timeout
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_7timeout_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_7timeout_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_7timeout___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_7timeout___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":743
 *     @property
 *     def timeout(self):
 *         return self.attr.timeout             # <<<<<<<<<<<<<<
 *     @timeout.setter
 *     def timeout(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.timeout); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":741
 *         self.attr.port_num = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def timeout(self):
 *         return self.attr.timeout
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.timeout.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":744
 *     def timeout(self):
 *         return self.attr.timeout
 *     @timeout.setter             # <<<<<<<<<<<<<<
 *     def timeout(self, val):
 *         self.attr.timeout = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_7timeout_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_7timeout_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_7timeout_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_7timeout_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned char __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":746
 *     @timeout.setter
 *     def timeout(self, val):
 *         self.attr.timeout = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_char(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 746, __pyx_L1_error)
  __pyx_v_self->attr.timeout = __pyx_t_1;

  /* "pyverbs/qp.pyx":744
 *     def timeout(self):
 *         return self.attr.timeout
 *     @timeout.setter             # <<<<<<<<<<<<<<
 *     def timeout(self, val):
 *         self.attr.timeout = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.timeout.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":748
 *         self.attr.timeout = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def retry_cnt(self):
 *         return self.attr.retry_cnt
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_9retry_cnt_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_9retry_cnt_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_9retry_cnt___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_9retry_cnt___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":750
 *     @property
 *     def retry_cnt(self):
 *         return self.attr.retry_cnt             # <<<<<<<<<<<<<<
 *     @retry_cnt.setter
 *     def retry_cnt(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.retry_cnt); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":748
 *         self.attr.timeout = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def retry_cnt(self):
 *         return self.attr.retry_cnt
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.retry_cnt.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":751
 *     def retry_cnt(self):
 *         return self.attr.retry_cnt
 *     @retry_cnt.setter             # <<<<<<<<<<<<<<
 *     def retry_cnt(self, val):
 *         self.attr.retry_cnt = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_9retry_cnt_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_9retry_cnt_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_9retry_cnt_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_9retry_cnt_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned char __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":753
 *     @retry_cnt.setter
 *     def retry_cnt(self, val):
 *         self.attr.retry_cnt = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_char(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 753, __pyx_L1_error)
  __pyx_v_self->attr.retry_cnt = __pyx_t_1;

  /* "pyverbs/qp.pyx":751
 *     def retry_cnt(self):
 *         return self.attr.retry_cnt
 *     @retry_cnt.setter             # <<<<<<<<<<<<<<
 *     def retry_cnt(self, val):
 *         self.attr.retry_cnt = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.retry_cnt.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":755
 *         self.attr.retry_cnt = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def rnr_retry(self):
 *         return self.attr.rnr_retry
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_9rnr_retry_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_9rnr_retry_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_9rnr_retry___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_9rnr_retry___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":757
 *     @property
 *     def rnr_retry(self):
 *         return self.attr.rnr_retry             # <<<<<<<<<<<<<<
 *     @rnr_retry.setter
 *     def rnr_retry(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.rnr_retry); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":755
 *         self.attr.retry_cnt = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def rnr_retry(self):
 *         return self.attr.rnr_retry
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.rnr_retry.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":758
 *     def rnr_retry(self):
 *         return self.attr.rnr_retry
 *     @rnr_retry.setter             # <<<<<<<<<<<<<<
 *     def rnr_retry(self, val):
 *         self.attr.rnr_retry = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_9rnr_retry_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_9rnr_retry_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_9rnr_retry_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_9rnr_retry_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned char __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":760
 *     @rnr_retry.setter
 *     def rnr_retry(self, val):
 *         self.attr.rnr_retry = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_char(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 760, __pyx_L1_error)
  __pyx_v_self->attr.rnr_retry = __pyx_t_1;

  /* "pyverbs/qp.pyx":758
 *     def rnr_retry(self):
 *         return self.attr.rnr_retry
 *     @rnr_retry.setter             # <<<<<<<<<<<<<<
 *     def rnr_retry(self, val):
 *         self.attr.rnr_retry = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.rnr_retry.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":762
 *         self.attr.rnr_retry = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def alt_port_num(self):
 *         return self.attr.alt_port_num
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_12alt_port_num_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_12alt_port_num_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_12alt_port_num___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_12alt_port_num___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":764
 *     @property
 *     def alt_port_num(self):
 *         return self.attr.alt_port_num             # <<<<<<<<<<<<<<
 *     @alt_port_num.setter
 *     def alt_port_num(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.alt_port_num); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":762
 *         self.attr.rnr_retry = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def alt_port_num(self):
 *         return self.attr.alt_port_num
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.alt_port_num.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":765
 *     def alt_port_num(self):
 *         return self.attr.alt_port_num
 *     @alt_port_num.setter             # <<<<<<<<<<<<<<
 *     def alt_port_num(self, val):
 *         self.attr.alt_port_num = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_12alt_port_num_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_12alt_port_num_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_12alt_port_num_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_12alt_port_num_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned char __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":767
 *     @alt_port_num.setter
 *     def alt_port_num(self, val):
 *         self.attr.alt_port_num = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_char(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 767, __pyx_L1_error)
  __pyx_v_self->attr.alt_port_num = __pyx_t_1;

  /* "pyverbs/qp.pyx":765
 *     def alt_port_num(self):
 *         return self.attr.alt_port_num
 *     @alt_port_num.setter             # <<<<<<<<<<<<<<
 *     def alt_port_num(self, val):
 *         self.attr.alt_port_num = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.alt_port_num.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":769
 *         self.attr.alt_port_num = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def alt_timeout(self):
 *         return self.attr.alt_timeout
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_11alt_timeout_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_11alt_timeout_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_11alt_timeout___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_11alt_timeout___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":771
 *     @property
 *     def alt_timeout(self):
 *         return self.attr.alt_timeout             # <<<<<<<<<<<<<<
 *     @alt_timeout.setter
 *     def alt_timeout(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.alt_timeout); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":769
 *         self.attr.alt_port_num = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def alt_timeout(self):
 *         return self.attr.alt_timeout
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.alt_timeout.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":772
 *     def alt_timeout(self):
 *         return self.attr.alt_timeout
 *     @alt_timeout.setter             # <<<<<<<<<<<<<<
 *     def alt_timeout(self, val):
 *         self.attr.alt_timeout = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_11alt_timeout_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_11alt_timeout_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_11alt_timeout_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_11alt_timeout_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned char __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":774
 *     @alt_timeout.setter
 *     def alt_timeout(self, val):
 *         self.attr.alt_timeout = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_char(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 774, __pyx_L1_error)
  __pyx_v_self->attr.alt_timeout = __pyx_t_1;

  /* "pyverbs/qp.pyx":772
 *     def alt_timeout(self):
 *         return self.attr.alt_timeout
 *     @alt_timeout.setter             # <<<<<<<<<<<<<<
 *     def alt_timeout(self, val):
 *         self.attr.alt_timeout = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.alt_timeout.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":776
 *         self.attr.alt_timeout = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def rate_limit(self):
 *         return self.attr.rate_limit
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_10rate_limit_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_10rate_limit_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_10rate_limit___get__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_10rate_limit___get__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":778
 *     @property
 *     def rate_limit(self):
 *         return self.attr.rate_limit             # <<<<<<<<<<<<<<
 *     @rate_limit.setter
 *     def rate_limit(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.rate_limit); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":776
 *         self.attr.alt_timeout = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def rate_limit(self):
 *         return self.attr.rate_limit
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.rate_limit.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":779
 *     def rate_limit(self):
 *         return self.attr.rate_limit
 *     @rate_limit.setter             # <<<<<<<<<<<<<<
 *     def rate_limit(self, val):
 *         self.attr.rate_limit = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_6QPAttr_10rate_limit_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_6QPAttr_10rate_limit_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_10rate_limit_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_6QPAttr_10rate_limit_2__set__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":781
 *     @rate_limit.setter
 *     def rate_limit(self, val):
 *         self.attr.rate_limit = val             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 781, __pyx_L1_error)
  __pyx_v_self->attr.rate_limit = __pyx_t_1;

  /* "pyverbs/qp.pyx":779
 *     def rate_limit(self):
 *         return self.attr.rate_limit
 *     @rate_limit.setter             # <<<<<<<<<<<<<<
 *     def rate_limit(self, val):
 *         self.attr.rate_limit = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.rate_limit.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":783
 *         self.attr.rate_limit = val
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         print_format = '{:22}: {:<20}\n'
 *         ah_format = '    {:22}: {:<20}\n'
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_5__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_5__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_4__str__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_4__str__(struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_v_print_format = NULL;
  PyObject *__pyx_v_ah_format = NULL;
  PyObject *__pyx_v_ident_format = NULL;
  PyObject *__pyx_v_global_ah = NULL;
  PyObject *__pyx_v_alt_global_ah = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "pyverbs/qp.pyx":784
 * 
 *     def __str__(self):
 *         print_format = '{:22}: {:<20}\n'             # <<<<<<<<<<<<<<
 *         ah_format = '    {:22}: {:<20}\n'
 *         ident_format = '  {:22}: {:<20}\n'
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_22_20);
  __pyx_v_print_format = __pyx_mstate_global->__pyx_kp_u_22_20;

  /* "pyverbs/qp.pyx":785
 *     def __str__(self):
 *         print_format = '{:22}: {:<20}\n'
 *         ah_format = '    {:22}: {:<20}\n'             # <<<<<<<<<<<<<<
 *         ident_format = '  {:22}: {:<20}\n'
 *         if self.attr.ah_attr.is_global:
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_22_20_2);
  __pyx_v_ah_format = __pyx_mstate_global->__pyx_kp_u_22_20_2;

  /* "pyverbs/qp.pyx":786
 *         print_format = '{:22}: {:<20}\n'
 *         ah_format = '    {:22}: {:<20}\n'
 *         ident_format = '  {:22}: {:<20}\n'             # <<<<<<<<<<<<<<
 *         if self.attr.ah_attr.is_global:
 *             global_ah = ah_format.format('dgid',
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_22_20_3);
  __pyx_v_ident_format = __pyx_mstate_global->__pyx_kp_u_22_20_3;

  /* "pyverbs/qp.pyx":787
 *         ah_format = '    {:22}: {:<20}\n'
 *         ident_format = '  {:22}: {:<20}\n'
 *         if self.attr.ah_attr.is_global:             # <<<<<<<<<<<<<<
 *             global_ah = ah_format.format('dgid',
 *                                          gid_str(self.attr.ah_attr.grh.dgid._global.subnet_prefix,
*/
  __pyx_t_1 = (__pyx_v_self->attr.ah_attr.is_global != 0);
  if (__pyx_t_1) {

    /* "pyverbs/qp.pyx":789
 *         if self.attr.ah_attr.is_global:
 *             global_ah = ah_format.format('dgid',
 *                                          gid_str(self.attr.ah_attr.grh.dgid._global.subnet_prefix,             # <<<<<<<<<<<<<<
 *                                                  self.attr.ah_attr.grh.dgid._global.interface_id)) +\
 *                         ah_format.format('flow label',
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_gid_str); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 789, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyLong_From_unsigned_long(__pyx_v_self->attr.ah_attr.grh.dgid.global.subnet_prefix); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 789, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "pyverbs/qp.pyx":790
 *             global_ah = ah_format.format('dgid',
 *                                          gid_str(self.attr.ah_attr.grh.dgid._global.subnet_prefix,
 *                                                  self.attr.ah_attr.grh.dgid._global.interface_id)) +\             # <<<<<<<<<<<<<<
 *                         ah_format.format('flow label',
 *                                          self.attr.ah_attr.grh.flow_label) +\
*/
    __pyx_t_6 = __Pyx_PyLong_From_unsigned_long(__pyx_v_self->attr.ah_attr.grh.dgid.global.interface_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_t_5, __pyx_t_6};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_7, (3-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 789, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }

    /* "pyverbs/qp.pyx":788
 *         ident_format = '  {:22}: {:<20}\n'
 *         if self.attr.ah_attr.is_global:
 *             global_ah = ah_format.format('dgid',             # <<<<<<<<<<<<<<
 *                                          gid_str(self.attr.ah_attr.grh.dgid._global.subnet_prefix,
 *                                                  self.attr.ah_attr.grh.dgid._global.interface_id)) +\
*/
    __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ah_format, __pyx_mstate_global->__pyx_n_u_dgid, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 788, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":792
 *                                                  self.attr.ah_attr.grh.dgid._global.interface_id)) +\
 *                         ah_format.format('flow label',
 *                                          self.attr.ah_attr.grh.flow_label) +\             # <<<<<<<<<<<<<<
 *                         ah_format.format('sgid index',
 *                                          self.attr.ah_attr.grh.sgid_index) +\
*/
    __pyx_t_2 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.ah_attr.grh.flow_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 792, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "pyverbs/qp.pyx":791
 *                                          gid_str(self.attr.ah_attr.grh.dgid._global.subnet_prefix,
 *                                                  self.attr.ah_attr.grh.dgid._global.interface_id)) +\
 *                         ah_format.format('flow label',             # <<<<<<<<<<<<<<
 *                                          self.attr.ah_attr.grh.flow_label) +\
 *                         ah_format.format('sgid index',
*/
    __pyx_t_6 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ah_format, __pyx_mstate_global->__pyx_kp_u_flow_label_2, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 791, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":790
 *             global_ah = ah_format.format('dgid',
 *                                          gid_str(self.attr.ah_attr.grh.dgid._global.subnet_prefix,
 *                                                  self.attr.ah_attr.grh.dgid._global.interface_id)) +\             # <<<<<<<<<<<<<<
 *                         ah_format.format('flow label',
 *                                          self.attr.ah_attr.grh.flow_label) +\
*/
    __pyx_t_2 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "pyverbs/qp.pyx":794
 *                                          self.attr.ah_attr.grh.flow_label) +\
 *                         ah_format.format('sgid index',
 *                                          self.attr.ah_attr.grh.sgid_index) +\             # <<<<<<<<<<<<<<
 *                         ah_format.format('hop limit',
 *                                          self.attr.ah_attr.grh.hop_limit) +\
*/
    __pyx_t_6 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.ah_attr.grh.sgid_index); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "pyverbs/qp.pyx":793
 *                         ah_format.format('flow label',
 *                                          self.attr.ah_attr.grh.flow_label) +\
 *                         ah_format.format('sgid index',             # <<<<<<<<<<<<<<
 *                                          self.attr.ah_attr.grh.sgid_index) +\
 *                         ah_format.format('hop limit',
*/
    __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ah_format, __pyx_mstate_global->__pyx_kp_u_sgid_index_2, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "pyverbs/qp.pyx":792
 *                                                  self.attr.ah_attr.grh.dgid._global.interface_id)) +\
 *                         ah_format.format('flow label',
 *                                          self.attr.ah_attr.grh.flow_label) +\             # <<<<<<<<<<<<<<
 *                         ah_format.format('sgid index',
 *                                          self.attr.ah_attr.grh.sgid_index) +\
*/
    __pyx_t_6 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 792, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "pyverbs/qp.pyx":796
 *                                          self.attr.ah_attr.grh.sgid_index) +\
 *                         ah_format.format('hop limit',
 *                                          self.attr.ah_attr.grh.hop_limit) +\             # <<<<<<<<<<<<<<
 *                         ah_format.format('traffic_class',
 *                                          self.attr.ah_attr.grh.traffic_class)
*/
    __pyx_t_4 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.ah_attr.grh.hop_limit); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "pyverbs/qp.pyx":795
 *                         ah_format.format('sgid index',
 *                                          self.attr.ah_attr.grh.sgid_index) +\
 *                         ah_format.format('hop limit',             # <<<<<<<<<<<<<<
 *                                          self.attr.ah_attr.grh.hop_limit) +\
 *                         ah_format.format('traffic_class',
*/
    __pyx_t_2 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ah_format, __pyx_mstate_global->__pyx_kp_u_hop_limit_2, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 795, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "pyverbs/qp.pyx":794
 *                                          self.attr.ah_attr.grh.flow_label) +\
 *                         ah_format.format('sgid index',
 *                                          self.attr.ah_attr.grh.sgid_index) +\             # <<<<<<<<<<<<<<
 *                         ah_format.format('hop limit',
 *                                          self.attr.ah_attr.grh.hop_limit) +\
*/
    __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_6, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":798
 *                                          self.attr.ah_attr.grh.hop_limit) +\
 *                         ah_format.format('traffic_class',
 *                                          self.attr.ah_attr.grh.traffic_class)             # <<<<<<<<<<<<<<
 *         else:
 *             global_ah = ''
*/
    __pyx_t_2 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.ah_attr.grh.traffic_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "pyverbs/qp.pyx":797
 *                         ah_format.format('hop limit',
 *                                          self.attr.ah_attr.grh.hop_limit) +\
 *                         ah_format.format('traffic_class',             # <<<<<<<<<<<<<<
 *                                          self.attr.ah_attr.grh.traffic_class)
 *         else:
*/
    __pyx_t_6 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ah_format, __pyx_mstate_global->__pyx_n_u_traffic_class, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 797, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":796
 *                                          self.attr.ah_attr.grh.sgid_index) +\
 *                         ah_format.format('hop limit',
 *                                          self.attr.ah_attr.grh.hop_limit) +\             # <<<<<<<<<<<<<<
 *                         ah_format.format('traffic_class',
 *                                          self.attr.ah_attr.grh.traffic_class)
*/
    __pyx_t_2 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_global_ah = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":787
 *         ah_format = '    {:22}: {:<20}\n'
 *         ident_format = '  {:22}: {:<20}\n'
 *         if self.attr.ah_attr.is_global:             # <<<<<<<<<<<<<<
 *             global_ah = ah_format.format('dgid',
 *                                          gid_str(self.attr.ah_attr.grh.dgid._global.subnet_prefix,
*/
    goto __pyx_L3;
  }

  /* "pyverbs/qp.pyx":800
 *                                          self.attr.ah_attr.grh.traffic_class)
 *         else:
 *             global_ah = ''             # <<<<<<<<<<<<<<
 *         if self.attr.alt_ah_attr.is_global:
 *             alt_global_ah = ah_format.format('dgid',
*/
  /*else*/ {
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u__4);
    __pyx_v_global_ah = __pyx_mstate_global->__pyx_kp_u__4;
  }
  __pyx_L3:;

  /* "pyverbs/qp.pyx":801
 *         else:
 *             global_ah = ''
 *         if self.attr.alt_ah_attr.is_global:             # <<<<<<<<<<<<<<
 *             alt_global_ah = ah_format.format('dgid',
 *                                              gid_str(self.attr.alt_ah_attr.grh.dgid._global.subnet_prefix,
*/
  __pyx_t_1 = (__pyx_v_self->attr.alt_ah_attr.is_global != 0);
  if (__pyx_t_1) {

    /* "pyverbs/qp.pyx":803
 *         if self.attr.alt_ah_attr.is_global:
 *             alt_global_ah = ah_format.format('dgid',
 *                                              gid_str(self.attr.alt_ah_attr.grh.dgid._global.subnet_prefix,             # <<<<<<<<<<<<<<
 *                                                      self.attr.alt_ah_attr.grh.dgid._global.interface_id)) +\
 *                             ah_format.format('flow label',
*/
    __pyx_t_6 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_gid_str); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyLong_From_unsigned_long(__pyx_v_self->attr.alt_ah_attr.grh.dgid.global.subnet_prefix); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "pyverbs/qp.pyx":804
 *             alt_global_ah = ah_format.format('dgid',
 *                                              gid_str(self.attr.alt_ah_attr.grh.dgid._global.subnet_prefix,
 *                                                      self.attr.alt_ah_attr.grh.dgid._global.interface_id)) +\             # <<<<<<<<<<<<<<
 *                             ah_format.format('flow label',
 *                                              self.attr.alt_ah_attr.grh.flow_label) +\
*/
    __pyx_t_3 = __Pyx_PyLong_From_unsigned_long(__pyx_v_self->attr.alt_ah_attr.grh.dgid.global.interface_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 804, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_6);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_t_5, __pyx_t_3};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_7, (3-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 803, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }

    /* "pyverbs/qp.pyx":802
 *             global_ah = ''
 *         if self.attr.alt_ah_attr.is_global:
 *             alt_global_ah = ah_format.format('dgid',             # <<<<<<<<<<<<<<
 *                                              gid_str(self.attr.alt_ah_attr.grh.dgid._global.subnet_prefix,
 *                                                      self.attr.alt_ah_attr.grh.dgid._global.interface_id)) +\
*/
    __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ah_format, __pyx_mstate_global->__pyx_n_u_dgid, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 802, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":806
 *                                                      self.attr.alt_ah_attr.grh.dgid._global.interface_id)) +\
 *                             ah_format.format('flow label',
 *                                              self.attr.alt_ah_attr.grh.flow_label) +\             # <<<<<<<<<<<<<<
 *                             ah_format.format('sgid index',
 *                                              self.attr.alt_ah_attr.grh.sgid_index) +\
*/
    __pyx_t_2 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.alt_ah_attr.grh.flow_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 806, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "pyverbs/qp.pyx":805
 *                                              gid_str(self.attr.alt_ah_attr.grh.dgid._global.subnet_prefix,
 *                                                      self.attr.alt_ah_attr.grh.dgid._global.interface_id)) +\
 *                             ah_format.format('flow label',             # <<<<<<<<<<<<<<
 *                                              self.attr.alt_ah_attr.grh.flow_label) +\
 *                             ah_format.format('sgid index',
*/
    __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ah_format, __pyx_mstate_global->__pyx_kp_u_flow_label_2, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 805, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":804
 *             alt_global_ah = ah_format.format('dgid',
 *                                              gid_str(self.attr.alt_ah_attr.grh.dgid._global.subnet_prefix,
 *                                                      self.attr.alt_ah_attr.grh.dgid._global.interface_id)) +\             # <<<<<<<<<<<<<<
 *                             ah_format.format('flow label',
 *                                              self.attr.alt_ah_attr.grh.flow_label) +\
*/
    __pyx_t_2 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 804, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyverbs/qp.pyx":808
 *                                              self.attr.alt_ah_attr.grh.flow_label) +\
 *                             ah_format.format('sgid index',
 *                                              self.attr.alt_ah_attr.grh.sgid_index) +\             # <<<<<<<<<<<<<<
 *                             ah_format.format('hop limit',
 *                                              self.attr.alt_ah_attr.grh.hop_limit) +\
*/
    __pyx_t_3 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.alt_ah_attr.grh.sgid_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 808, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "pyverbs/qp.pyx":807
 *                             ah_format.format('flow label',
 *                                              self.attr.alt_ah_attr.grh.flow_label) +\
 *                             ah_format.format('sgid index',             # <<<<<<<<<<<<<<
 *                                              self.attr.alt_ah_attr.grh.sgid_index) +\
 *                             ah_format.format('hop limit',
*/
    __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ah_format, __pyx_mstate_global->__pyx_kp_u_sgid_index_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 807, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyverbs/qp.pyx":806
 *                                                      self.attr.alt_ah_attr.grh.dgid._global.interface_id)) +\
 *                             ah_format.format('flow label',
 *                                              self.attr.alt_ah_attr.grh.flow_label) +\             # <<<<<<<<<<<<<<
 *                             ah_format.format('sgid index',
 *                                              self.attr.alt_ah_attr.grh.sgid_index) +\
*/
    __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 806, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "pyverbs/qp.pyx":810
 *                                              self.attr.alt_ah_attr.grh.sgid_index) +\
 *                             ah_format.format('hop limit',
 *                                              self.attr.alt_ah_attr.grh.hop_limit) +\             # <<<<<<<<<<<<<<
 *                             ah_format.format('traffic_class',
 *                                              self.attr.alt_ah_attr.grh.traffic_class)
*/
    __pyx_t_4 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.alt_ah_attr.grh.hop_limit); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "pyverbs/qp.pyx":809
 *                             ah_format.format('sgid index',
 *                                              self.attr.alt_ah_attr.grh.sgid_index) +\
 *                             ah_format.format('hop limit',             # <<<<<<<<<<<<<<
 *                                              self.attr.alt_ah_attr.grh.hop_limit) +\
 *                             ah_format.format('traffic_class',
*/
    __pyx_t_2 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ah_format, __pyx_mstate_global->__pyx_kp_u_hop_limit_2, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 809, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "pyverbs/qp.pyx":808
 *                                              self.attr.alt_ah_attr.grh.flow_label) +\
 *                             ah_format.format('sgid index',
 *                                              self.attr.alt_ah_attr.grh.sgid_index) +\             # <<<<<<<<<<<<<<
 *                             ah_format.format('hop limit',
 *                                              self.attr.alt_ah_attr.grh.hop_limit) +\
*/
    __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 808, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":812
 *                                              self.attr.alt_ah_attr.grh.hop_limit) +\
 *                             ah_format.format('traffic_class',
 *                                              self.attr.alt_ah_attr.grh.traffic_class)             # <<<<<<<<<<<<<<
 *         else:
 *             alt_global_ah = ''
*/
    __pyx_t_2 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.alt_ah_attr.grh.traffic_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 812, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "pyverbs/qp.pyx":811
 *                             ah_format.format('hop limit',
 *                                              self.attr.alt_ah_attr.grh.hop_limit) +\
 *                             ah_format.format('traffic_class',             # <<<<<<<<<<<<<<
 *                                              self.attr.alt_ah_attr.grh.traffic_class)
 *         else:
*/
    __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ah_format, __pyx_mstate_global->__pyx_n_u_traffic_class, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 811, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":810
 *                                              self.attr.alt_ah_attr.grh.sgid_index) +\
 *                             ah_format.format('hop limit',
 *                                              self.attr.alt_ah_attr.grh.hop_limit) +\             # <<<<<<<<<<<<<<
 *                             ah_format.format('traffic_class',
 *                                              self.attr.alt_ah_attr.grh.traffic_class)
*/
    __pyx_t_2 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_alt_global_ah = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":801
 *         else:
 *             global_ah = ''
 *         if self.attr.alt_ah_attr.is_global:             # <<<<<<<<<<<<<<
 *             alt_global_ah = ah_format.format('dgid',
 *                                              gid_str(self.attr.alt_ah_attr.grh.dgid._global.subnet_prefix,
*/
    goto __pyx_L4;
  }

  /* "pyverbs/qp.pyx":814
 *                                              self.attr.alt_ah_attr.grh.traffic_class)
 *         else:
 *             alt_global_ah = ''             # <<<<<<<<<<<<<<
 *         return print_format.format('QP state',
 *                                    qp_state_to_str(self.attr.qp_state)) +\
*/
  /*else*/ {
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u__4);
    __pyx_v_alt_global_ah = __pyx_mstate_global->__pyx_kp_u__4;
  }
  __pyx_L4:;

  /* "pyverbs/qp.pyx":815
 *         else:
 *             alt_global_ah = ''
 *         return print_format.format('QP state',             # <<<<<<<<<<<<<<
 *                                    qp_state_to_str(self.attr.qp_state)) +\
 *                print_format.format('QP current state',
*/
  __Pyx_XDECREF(__pyx_r);

  /* "pyverbs/qp.pyx":816
 *             alt_global_ah = ''
 *         return print_format.format('QP state',
 *                                    qp_state_to_str(self.attr.qp_state)) +\             # <<<<<<<<<<<<<<
 *                print_format.format('QP current state',
 *                             qp_state_to_str(self.attr.cur_qp_state)) +\
*/
  __pyx_t_3 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_qp_state_to_str); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py(__pyx_v_self->attr.qp_state); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_3);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_7 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 816, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }

  /* "pyverbs/qp.pyx":815
 *         else:
 *             alt_global_ah = ''
 *         return print_format.format('QP state',             # <<<<<<<<<<<<<<
 *                                    qp_state_to_str(self.attr.qp_state)) +\
 *                print_format.format('QP current state',
*/
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_QP_state, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":818
 *                                    qp_state_to_str(self.attr.qp_state)) +\
 *                print_format.format('QP current state',
 *                             qp_state_to_str(self.attr.cur_qp_state)) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Path MTU',
 *                                    mtu_to_str(self.attr.path_mtu)) +\
*/
  __pyx_t_5 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_qp_state_to_str); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 818, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py(__pyx_v_self->attr.cur_qp_state); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 818, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    assert(__pyx_t_5);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
    __pyx_t_7 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_6};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 818, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }

  /* "pyverbs/qp.pyx":817
 *         return print_format.format('QP state',
 *                                    qp_state_to_str(self.attr.qp_state)) +\
 *                print_format.format('QP current state',             # <<<<<<<<<<<<<<
 *                             qp_state_to_str(self.attr.cur_qp_state)) +\
 *                print_format.format('Path MTU',
*/
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_QP_current_state, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":816
 *             alt_global_ah = ''
 *         return print_format.format('QP state',
 *                                    qp_state_to_str(self.attr.qp_state)) +\             # <<<<<<<<<<<<<<
 *                print_format.format('QP current state',
 *                             qp_state_to_str(self.attr.cur_qp_state)) +\
*/
  __pyx_t_2 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":820
 *                             qp_state_to_str(self.attr.cur_qp_state)) +\
 *                print_format.format('Path MTU',
 *                                    mtu_to_str(self.attr.path_mtu)) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Path mig. state',
 *                             mig_state_to_str(self.attr.path_mig_state)) +\
*/
  __pyx_t_4 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_mtu_to_str); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py(__pyx_v_self->attr.path_mtu); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
    assert(__pyx_t_4);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
    __pyx_t_7 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_5};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 820, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }

  /* "pyverbs/qp.pyx":819
 *                print_format.format('QP current state',
 *                             qp_state_to_str(self.attr.cur_qp_state)) +\
 *                print_format.format('Path MTU',             # <<<<<<<<<<<<<<
 *                                    mtu_to_str(self.attr.path_mtu)) +\
 *                print_format.format('Path mig. state',
*/
  __pyx_t_6 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_Path_MTU, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":818
 *                                    qp_state_to_str(self.attr.qp_state)) +\
 *                print_format.format('QP current state',
 *                             qp_state_to_str(self.attr.cur_qp_state)) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Path MTU',
 *                                    mtu_to_str(self.attr.path_mtu)) +\
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_2, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 818, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "pyverbs/qp.pyx":822
 *                                    mtu_to_str(self.attr.path_mtu)) +\
 *                print_format.format('Path mig. state',
 *                             mig_state_to_str(self.attr.path_mig_state)) +\             # <<<<<<<<<<<<<<
 *                print_format.format('QKey', self.attr.qkey) +\
 *                print_format.format('RQ PSN', self.attr.rq_psn) +\
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_mig_state_to_str); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py(__pyx_v_self->attr.path_mig_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
    __pyx_t_7 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_4};
    __pyx_t_6 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 822, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
  }

  /* "pyverbs/qp.pyx":821
 *                print_format.format('Path MTU',
 *                                    mtu_to_str(self.attr.path_mtu)) +\
 *                print_format.format('Path mig. state',             # <<<<<<<<<<<<<<
 *                             mig_state_to_str(self.attr.path_mig_state)) +\
 *                print_format.format('QKey', self.attr.qkey) +\
*/
  __pyx_t_5 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_Path_mig_state, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "pyverbs/qp.pyx":820
 *                             qp_state_to_str(self.attr.cur_qp_state)) +\
 *                print_format.format('Path MTU',
 *                                    mtu_to_str(self.attr.path_mtu)) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Path mig. state',
 *                             mig_state_to_str(self.attr.path_mig_state)) +\
*/
  __pyx_t_6 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":823
 *                print_format.format('Path mig. state',
 *                             mig_state_to_str(self.attr.path_mig_state)) +\
 *                print_format.format('QKey', self.attr.qkey) +\             # <<<<<<<<<<<<<<
 *                print_format.format('RQ PSN', self.attr.rq_psn) +\
 *                print_format.format('SQ PSN', self.attr.sq_psn) +\
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.qkey); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_n_u_QKey, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":822
 *                                    mtu_to_str(self.attr.path_mtu)) +\
 *                print_format.format('Path mig. state',
 *                             mig_state_to_str(self.attr.path_mig_state)) +\             # <<<<<<<<<<<<<<
 *                print_format.format('QKey', self.attr.qkey) +\
 *                print_format.format('RQ PSN', self.attr.rq_psn) +\
*/
  __pyx_t_5 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_6, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":824
 *                             mig_state_to_str(self.attr.path_mig_state)) +\
 *                print_format.format('QKey', self.attr.qkey) +\
 *                print_format.format('RQ PSN', self.attr.rq_psn) +\             # <<<<<<<<<<<<<<
 *                print_format.format('SQ PSN', self.attr.sq_psn) +\
 *                print_format.format('Dest QP number', self.attr.dest_qp_num) +\
*/
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.rq_psn); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_RQ_PSN, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":823
 *                print_format.format('Path mig. state',
 *                             mig_state_to_str(self.attr.path_mig_state)) +\
 *                print_format.format('QKey', self.attr.qkey) +\             # <<<<<<<<<<<<<<
 *                print_format.format('RQ PSN', self.attr.rq_psn) +\
 *                print_format.format('SQ PSN', self.attr.sq_psn) +\
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "pyverbs/qp.pyx":825
 *                print_format.format('QKey', self.attr.qkey) +\
 *                print_format.format('RQ PSN', self.attr.rq_psn) +\
 *                print_format.format('SQ PSN', self.attr.sq_psn) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Dest QP number', self.attr.dest_qp_num) +\
 *                print_format.format('QP access flags',
*/
  __pyx_t_6 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.sq_psn); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_SQ_PSN, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "pyverbs/qp.pyx":824
 *                             mig_state_to_str(self.attr.path_mig_state)) +\
 *                print_format.format('QKey', self.attr.qkey) +\
 *                print_format.format('RQ PSN', self.attr.rq_psn) +\             # <<<<<<<<<<<<<<
 *                print_format.format('SQ PSN', self.attr.sq_psn) +\
 *                print_format.format('Dest QP number', self.attr.dest_qp_num) +\
*/
  __pyx_t_6 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":826
 *                print_format.format('RQ PSN', self.attr.rq_psn) +\
 *                print_format.format('SQ PSN', self.attr.sq_psn) +\
 *                print_format.format('Dest QP number', self.attr.dest_qp_num) +\             # <<<<<<<<<<<<<<
 *                print_format.format('QP access flags',
 *                                    access_flags_to_str(self.attr.qp_access_flags)) +\
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.dest_qp_num); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_Dest_QP_number, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":825
 *                print_format.format('QKey', self.attr.qkey) +\
 *                print_format.format('RQ PSN', self.attr.rq_psn) +\
 *                print_format.format('SQ PSN', self.attr.sq_psn) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Dest QP number', self.attr.dest_qp_num) +\
 *                print_format.format('QP access flags',
*/
  __pyx_t_5 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_6, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":828
 *                print_format.format('Dest QP number', self.attr.dest_qp_num) +\
 *                print_format.format('QP access flags',
 *                                    access_flags_to_str(self.attr.qp_access_flags)) +\             # <<<<<<<<<<<<<<
 *                'QP caps:\n' +\
 *                ident_format.format('max send WR',
*/
  __pyx_t_6 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_access_flags_to_str); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.qp_access_flags); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_6);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_7 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_2};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 828, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }

  /* "pyverbs/qp.pyx":827
 *                print_format.format('SQ PSN', self.attr.sq_psn) +\
 *                print_format.format('Dest QP number', self.attr.dest_qp_num) +\
 *                print_format.format('QP access flags',             # <<<<<<<<<<<<<<
 *                                    access_flags_to_str(self.attr.qp_access_flags)) +\
 *                'QP caps:\n' +\
*/
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_QP_access_flags, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":826
 *                print_format.format('RQ PSN', self.attr.rq_psn) +\
 *                print_format.format('SQ PSN', self.attr.sq_psn) +\
 *                print_format.format('Dest QP number', self.attr.dest_qp_num) +\             # <<<<<<<<<<<<<<
 *                print_format.format('QP access flags',
 *                                    access_flags_to_str(self.attr.qp_access_flags)) +\
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":828
 *                print_format.format('Dest QP number', self.attr.dest_qp_num) +\
 *                print_format.format('QP access flags',
 *                                    access_flags_to_str(self.attr.qp_access_flags)) +\             # <<<<<<<<<<<<<<
 *                'QP caps:\n' +\
 *                ident_format.format('max send WR',
*/
  __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_QP_caps); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":831
 *                'QP caps:\n' +\
 *                ident_format.format('max send WR',
 *                                    self.attr.cap.max_send_wr) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('max recv WR',
 *                                    self.attr.cap.max_recv_wr) +\
*/
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_send_wr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "pyverbs/qp.pyx":830
 *                                    access_flags_to_str(self.attr.qp_access_flags)) +\
 *                'QP caps:\n' +\
 *                ident_format.format('max send WR',             # <<<<<<<<<<<<<<
 *                                    self.attr.cap.max_send_wr) +\
 *                ident_format.format('max recv WR',
*/
  __pyx_t_5 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_kp_u_max_send_WR, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":829
 *                print_format.format('QP access flags',
 *                                    access_flags_to_str(self.attr.qp_access_flags)) +\
 *                'QP caps:\n' +\             # <<<<<<<<<<<<<<
 *                ident_format.format('max send WR',
 *                                    self.attr.cap.max_send_wr) +\
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":833
 *                                    self.attr.cap.max_send_wr) +\
 *                ident_format.format('max recv WR',
 *                                    self.attr.cap.max_recv_wr) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('max send SGE',
 *                                    self.attr.cap.max_send_sge) +\
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_recv_wr); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "pyverbs/qp.pyx":832
 *                ident_format.format('max send WR',
 *                                    self.attr.cap.max_send_wr) +\
 *                ident_format.format('max recv WR',             # <<<<<<<<<<<<<<
 *                                    self.attr.cap.max_recv_wr) +\
 *                ident_format.format('max send SGE',
*/
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_kp_u_max_recv_WR, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":831
 *                'QP caps:\n' +\
 *                ident_format.format('max send WR',
 *                                    self.attr.cap.max_send_wr) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('max recv WR',
 *                                    self.attr.cap.max_recv_wr) +\
*/
  __pyx_t_5 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":835
 *                                    self.attr.cap.max_recv_wr) +\
 *                ident_format.format('max send SGE',
 *                                    self.attr.cap.max_send_sge) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('max recv SGE',
 *                                    self.attr.cap.max_recv_sge) +\
*/
  __pyx_t_4 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_send_sge); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "pyverbs/qp.pyx":834
 *                ident_format.format('max recv WR',
 *                                    self.attr.cap.max_recv_wr) +\
 *                ident_format.format('max send SGE',             # <<<<<<<<<<<<<<
 *                                    self.attr.cap.max_send_sge) +\
 *                ident_format.format('max recv SGE',
*/
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_kp_u_max_send_SGE, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":833
 *                                    self.attr.cap.max_send_wr) +\
 *                ident_format.format('max recv WR',
 *                                    self.attr.cap.max_recv_wr) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('max send SGE',
 *                                    self.attr.cap.max_send_sge) +\
*/
  __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":837
 *                                    self.attr.cap.max_send_sge) +\
 *                ident_format.format('max recv SGE',
 *                                    self.attr.cap.max_recv_sge) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('max inline data',
 *                                    self.attr.cap.max_inline_data) +\
*/
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_recv_sge); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 837, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "pyverbs/qp.pyx":836
 *                ident_format.format('max send SGE',
 *                                    self.attr.cap.max_send_sge) +\
 *                ident_format.format('max recv SGE',             # <<<<<<<<<<<<<<
 *                                    self.attr.cap.max_recv_sge) +\
 *                ident_format.format('max inline data',
*/
  __pyx_t_5 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_kp_u_max_recv_SGE, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":835
 *                                    self.attr.cap.max_recv_wr) +\
 *                ident_format.format('max send SGE',
 *                                    self.attr.cap.max_send_sge) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('max recv SGE',
 *                                    self.attr.cap.max_recv_sge) +\
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":839
 *                                    self.attr.cap.max_recv_sge) +\
 *                ident_format.format('max inline data',
 *                                    self.attr.cap.max_inline_data) +\             # <<<<<<<<<<<<<<
 *                'AH Attr:\n' +\
 *                ident_format.format('port num', self.attr.ah_attr.port_num) +\
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cap.max_inline_data); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "pyverbs/qp.pyx":838
 *                ident_format.format('max recv SGE',
 *                                    self.attr.cap.max_recv_sge) +\
 *                ident_format.format('max inline data',             # <<<<<<<<<<<<<<
 *                                    self.attr.cap.max_inline_data) +\
 *                'AH Attr:\n' +\
*/
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_kp_u_max_inline_data_2, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":837
 *                                    self.attr.cap.max_send_sge) +\
 *                ident_format.format('max recv SGE',
 *                                    self.attr.cap.max_recv_sge) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('max inline data',
 *                                    self.attr.cap.max_inline_data) +\
*/
  __pyx_t_5 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 837, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":839
 *                                    self.attr.cap.max_recv_sge) +\
 *                ident_format.format('max inline data',
 *                                    self.attr.cap.max_inline_data) +\             # <<<<<<<<<<<<<<
 *                'AH Attr:\n' +\
 *                ident_format.format('port num', self.attr.ah_attr.port_num) +\
*/
  __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_5, __pyx_mstate_global->__pyx_kp_u_AH_Attr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":841
 *                                    self.attr.cap.max_inline_data) +\
 *                'AH Attr:\n' +\
 *                ident_format.format('port num', self.attr.ah_attr.port_num) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('sl', self.attr.ah_attr.sl) +\
 *                ident_format.format('source path bits',
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.ah_attr.port_num); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_kp_u_port_num_2, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":840
 *                ident_format.format('max inline data',
 *                                    self.attr.cap.max_inline_data) +\
 *                'AH Attr:\n' +\             # <<<<<<<<<<<<<<
 *                ident_format.format('port num', self.attr.ah_attr.port_num) +\
 *                ident_format.format('sl', self.attr.ah_attr.sl) +\
*/
  __pyx_t_5 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":842
 *                'AH Attr:\n' +\
 *                ident_format.format('port num', self.attr.ah_attr.port_num) +\
 *                ident_format.format('sl', self.attr.ah_attr.sl) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('source path bits',
 *                                    self.attr.ah_attr.src_path_bits) +\
*/
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.ah_attr.sl); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 842, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_n_u_sl, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 842, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":841
 *                                    self.attr.cap.max_inline_data) +\
 *                'AH Attr:\n' +\
 *                ident_format.format('port num', self.attr.ah_attr.port_num) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('sl', self.attr.ah_attr.sl) +\
 *                ident_format.format('source path bits',
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":844
 *                ident_format.format('sl', self.attr.ah_attr.sl) +\
 *                ident_format.format('source path bits',
 *                                    self.attr.ah_attr.src_path_bits) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('dlid', self.attr.ah_attr.dlid) +\
 *                ident_format.format('port num', self.attr.ah_attr.port_num) +\
*/
  __pyx_t_4 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.ah_attr.src_path_bits); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "pyverbs/qp.pyx":843
 *                ident_format.format('port num', self.attr.ah_attr.port_num) +\
 *                ident_format.format('sl', self.attr.ah_attr.sl) +\
 *                ident_format.format('source path bits',             # <<<<<<<<<<<<<<
 *                                    self.attr.ah_attr.src_path_bits) +\
 *                ident_format.format('dlid', self.attr.ah_attr.dlid) +\
*/
  __pyx_t_5 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_kp_u_source_path_bits, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":842
 *                'AH Attr:\n' +\
 *                ident_format.format('port num', self.attr.ah_attr.port_num) +\
 *                ident_format.format('sl', self.attr.ah_attr.sl) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('source path bits',
 *                                    self.attr.ah_attr.src_path_bits) +\
*/
  __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 842, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":845
 *                ident_format.format('source path bits',
 *                                    self.attr.ah_attr.src_path_bits) +\
 *                ident_format.format('dlid', self.attr.ah_attr.dlid) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('port num', self.attr.ah_attr.port_num) +\
 *                ident_format.format('static rate',
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_short(__pyx_v_self->attr.ah_attr.dlid); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_n_u_dlid, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":844
 *                ident_format.format('sl', self.attr.ah_attr.sl) +\
 *                ident_format.format('source path bits',
 *                                    self.attr.ah_attr.src_path_bits) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('dlid', self.attr.ah_attr.dlid) +\
 *                ident_format.format('port num', self.attr.ah_attr.port_num) +\
*/
  __pyx_t_5 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":846
 *                                    self.attr.ah_attr.src_path_bits) +\
 *                ident_format.format('dlid', self.attr.ah_attr.dlid) +\
 *                ident_format.format('port num', self.attr.ah_attr.port_num) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('static rate',
 *                                    self.attr.ah_attr.static_rate) +\
*/
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.ah_attr.port_num); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 846, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_kp_u_port_num_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 846, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":845
 *                ident_format.format('source path bits',
 *                                    self.attr.ah_attr.src_path_bits) +\
 *                ident_format.format('dlid', self.attr.ah_attr.dlid) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('port num', self.attr.ah_attr.port_num) +\
 *                ident_format.format('static rate',
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":848
 *                ident_format.format('port num', self.attr.ah_attr.port_num) +\
 *                ident_format.format('static rate',
 *                                    self.attr.ah_attr.static_rate) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('is global',
 *                                    self.attr.ah_attr.is_global) +\
*/
  __pyx_t_4 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.ah_attr.static_rate); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 848, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "pyverbs/qp.pyx":847
 *                ident_format.format('dlid', self.attr.ah_attr.dlid) +\
 *                ident_format.format('port num', self.attr.ah_attr.port_num) +\
 *                ident_format.format('static rate',             # <<<<<<<<<<<<<<
 *                                    self.attr.ah_attr.static_rate) +\
 *                ident_format.format('is global',
*/
  __pyx_t_5 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_kp_u_static_rate_2, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":846
 *                                    self.attr.ah_attr.src_path_bits) +\
 *                ident_format.format('dlid', self.attr.ah_attr.dlid) +\
 *                ident_format.format('port num', self.attr.ah_attr.port_num) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('static rate',
 *                                    self.attr.ah_attr.static_rate) +\
*/
  __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 846, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":850
 *                                    self.attr.ah_attr.static_rate) +\
 *                ident_format.format('is global',
 *                                    self.attr.ah_attr.is_global) +\             # <<<<<<<<<<<<<<
 *                global_ah +\
 *                'Alt. AH Attr:\n' +\
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.ah_attr.is_global); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "pyverbs/qp.pyx":849
 *                ident_format.format('static rate',
 *                                    self.attr.ah_attr.static_rate) +\
 *                ident_format.format('is global',             # <<<<<<<<<<<<<<
 *                                    self.attr.ah_attr.is_global) +\
 *                global_ah +\
*/
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_kp_u_is_global_2, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":848
 *                ident_format.format('port num', self.attr.ah_attr.port_num) +\
 *                ident_format.format('static rate',
 *                                    self.attr.ah_attr.static_rate) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('is global',
 *                                    self.attr.ah_attr.is_global) +\
*/
  __pyx_t_5 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 848, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":850
 *                                    self.attr.ah_attr.static_rate) +\
 *                ident_format.format('is global',
 *                                    self.attr.ah_attr.is_global) +\             # <<<<<<<<<<<<<<
 *                global_ah +\
 *                'Alt. AH Attr:\n' +\
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_5, __pyx_v_global_ah); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":851
 *                ident_format.format('is global',
 *                                    self.attr.ah_attr.is_global) +\
 *                global_ah +\             # <<<<<<<<<<<<<<
 *                'Alt. AH Attr:\n' +\
 *                ident_format.format('port num', self.attr.alt_ah_attr.port_num) +\
*/
  __pyx_t_5 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Alt_AH_Attr); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 851, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":853
 *                global_ah +\
 *                'Alt. AH Attr:\n' +\
 *                ident_format.format('port num', self.attr.alt_ah_attr.port_num) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('sl', self.attr.alt_ah_attr.sl) +\
 *                ident_format.format('source path bits',
*/
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.alt_ah_attr.port_num); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_kp_u_port_num_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":852
 *                                    self.attr.ah_attr.is_global) +\
 *                global_ah +\
 *                'Alt. AH Attr:\n' +\             # <<<<<<<<<<<<<<
 *                ident_format.format('port num', self.attr.alt_ah_attr.port_num) +\
 *                ident_format.format('sl', self.attr.alt_ah_attr.sl) +\
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 852, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":854
 *                'Alt. AH Attr:\n' +\
 *                ident_format.format('port num', self.attr.alt_ah_attr.port_num) +\
 *                ident_format.format('sl', self.attr.alt_ah_attr.sl) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('source path bits',
 *                                    self.attr.alt_ah_attr.src_path_bits) +\
*/
  __pyx_t_4 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.alt_ah_attr.sl); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_n_u_sl, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":853
 *                global_ah +\
 *                'Alt. AH Attr:\n' +\
 *                ident_format.format('port num', self.attr.alt_ah_attr.port_num) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('sl', self.attr.alt_ah_attr.sl) +\
 *                ident_format.format('source path bits',
*/
  __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":856
 *                ident_format.format('sl', self.attr.alt_ah_attr.sl) +\
 *                ident_format.format('source path bits',
 *                                    self.attr.alt_ah_attr.src_path_bits) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('dlid', self.attr.alt_ah_attr.dlid) +\
 *                ident_format.format('port num', self.attr.alt_ah_attr.port_num) +\
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.alt_ah_attr.src_path_bits); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "pyverbs/qp.pyx":855
 *                ident_format.format('port num', self.attr.alt_ah_attr.port_num) +\
 *                ident_format.format('sl', self.attr.alt_ah_attr.sl) +\
 *                ident_format.format('source path bits',             # <<<<<<<<<<<<<<
 *                                    self.attr.alt_ah_attr.src_path_bits) +\
 *                ident_format.format('dlid', self.attr.alt_ah_attr.dlid) +\
*/
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_kp_u_source_path_bits, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":854
 *                'Alt. AH Attr:\n' +\
 *                ident_format.format('port num', self.attr.alt_ah_attr.port_num) +\
 *                ident_format.format('sl', self.attr.alt_ah_attr.sl) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('source path bits',
 *                                    self.attr.alt_ah_attr.src_path_bits) +\
*/
  __pyx_t_5 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":857
 *                ident_format.format('source path bits',
 *                                    self.attr.alt_ah_attr.src_path_bits) +\
 *                ident_format.format('dlid', self.attr.alt_ah_attr.dlid) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('port num', self.attr.alt_ah_attr.port_num) +\
 *                ident_format.format('static rate',
*/
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_short(__pyx_v_self->attr.alt_ah_attr.dlid); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_n_u_dlid, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":856
 *                ident_format.format('sl', self.attr.alt_ah_attr.sl) +\
 *                ident_format.format('source path bits',
 *                                    self.attr.alt_ah_attr.src_path_bits) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('dlid', self.attr.alt_ah_attr.dlid) +\
 *                ident_format.format('port num', self.attr.alt_ah_attr.port_num) +\
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":858
 *                                    self.attr.alt_ah_attr.src_path_bits) +\
 *                ident_format.format('dlid', self.attr.alt_ah_attr.dlid) +\
 *                ident_format.format('port num', self.attr.alt_ah_attr.port_num) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('static rate',
 *                                    self.attr.alt_ah_attr.static_rate) +\
*/
  __pyx_t_4 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.alt_ah_attr.port_num); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_kp_u_port_num_2, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":857
 *                ident_format.format('source path bits',
 *                                    self.attr.alt_ah_attr.src_path_bits) +\
 *                ident_format.format('dlid', self.attr.alt_ah_attr.dlid) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('port num', self.attr.alt_ah_attr.port_num) +\
 *                ident_format.format('static rate',
*/
  __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":860
 *                ident_format.format('port num', self.attr.alt_ah_attr.port_num) +\
 *                ident_format.format('static rate',
 *                                    self.attr.alt_ah_attr.static_rate) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('is global',
 *                                    self.attr.alt_ah_attr.is_global) +\
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.alt_ah_attr.static_rate); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "pyverbs/qp.pyx":859
 *                ident_format.format('dlid', self.attr.alt_ah_attr.dlid) +\
 *                ident_format.format('port num', self.attr.alt_ah_attr.port_num) +\
 *                ident_format.format('static rate',             # <<<<<<<<<<<<<<
 *                                    self.attr.alt_ah_attr.static_rate) +\
 *                ident_format.format('is global',
*/
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_kp_u_static_rate_2, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":858
 *                                    self.attr.alt_ah_attr.src_path_bits) +\
 *                ident_format.format('dlid', self.attr.alt_ah_attr.dlid) +\
 *                ident_format.format('port num', self.attr.alt_ah_attr.port_num) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('static rate',
 *                                    self.attr.alt_ah_attr.static_rate) +\
*/
  __pyx_t_5 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":862
 *                                    self.attr.alt_ah_attr.static_rate) +\
 *                ident_format.format('is global',
 *                                    self.attr.alt_ah_attr.is_global) +\             # <<<<<<<<<<<<<<
 *                alt_global_ah +\
 *                print_format.format('PKey index', self.attr.pkey_index) +\
*/
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.alt_ah_attr.is_global); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "pyverbs/qp.pyx":861
 *                ident_format.format('static rate',
 *                                    self.attr.alt_ah_attr.static_rate) +\
 *                ident_format.format('is global',             # <<<<<<<<<<<<<<
 *                                    self.attr.alt_ah_attr.is_global) +\
 *                alt_global_ah +\
*/
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_ident_format, __pyx_mstate_global->__pyx_kp_u_is_global_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":860
 *                ident_format.format('port num', self.attr.alt_ah_attr.port_num) +\
 *                ident_format.format('static rate',
 *                                    self.attr.alt_ah_attr.static_rate) +\             # <<<<<<<<<<<<<<
 *                ident_format.format('is global',
 *                                    self.attr.alt_ah_attr.is_global) +\
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":862
 *                                    self.attr.alt_ah_attr.static_rate) +\
 *                ident_format.format('is global',
 *                                    self.attr.alt_ah_attr.is_global) +\             # <<<<<<<<<<<<<<
 *                alt_global_ah +\
 *                print_format.format('PKey index', self.attr.pkey_index) +\
*/
  __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_v_alt_global_ah); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":864
 *                                    self.attr.alt_ah_attr.is_global) +\
 *                alt_global_ah +\
 *                print_format.format('PKey index', self.attr.pkey_index) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Alt. PKey index',
 *                                    self.attr.alt_pkey_index) +\
*/
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_short(__pyx_v_self->attr.pkey_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_PKey_index, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":863
 *                ident_format.format('is global',
 *                                    self.attr.alt_ah_attr.is_global) +\
 *                alt_global_ah +\             # <<<<<<<<<<<<<<
 *                print_format.format('PKey index', self.attr.pkey_index) +\
 *                print_format.format('Alt. PKey index',
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":866
 *                print_format.format('PKey index', self.attr.pkey_index) +\
 *                print_format.format('Alt. PKey index',
 *                                    self.attr.alt_pkey_index) +\             # <<<<<<<<<<<<<<
 *                print_format.format('En. SQD async notify',
 *                                    self.attr.en_sqd_async_notify) +\
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_short(__pyx_v_self->attr.alt_pkey_index); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "pyverbs/qp.pyx":865
 *                alt_global_ah +\
 *                print_format.format('PKey index', self.attr.pkey_index) +\
 *                print_format.format('Alt. PKey index',             # <<<<<<<<<<<<<<
 *                                    self.attr.alt_pkey_index) +\
 *                print_format.format('En. SQD async notify',
*/
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_Alt_PKey_index, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":864
 *                                    self.attr.alt_ah_attr.is_global) +\
 *                alt_global_ah +\
 *                print_format.format('PKey index', self.attr.pkey_index) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Alt. PKey index',
 *                                    self.attr.alt_pkey_index) +\
*/
  __pyx_t_5 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":868
 *                                    self.attr.alt_pkey_index) +\
 *                print_format.format('En. SQD async notify',
 *                                    self.attr.en_sqd_async_notify) +\             # <<<<<<<<<<<<<<
 *                print_format.format('SQ draining', self.attr.sq_draining) +\
 *                print_format.format('Max RD atomic', self.attr.max_rd_atomic) +\
*/
  __pyx_t_4 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.en_sqd_async_notify); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "pyverbs/qp.pyx":867
 *                print_format.format('Alt. PKey index',
 *                                    self.attr.alt_pkey_index) +\
 *                print_format.format('En. SQD async notify',             # <<<<<<<<<<<<<<
 *                                    self.attr.en_sqd_async_notify) +\
 *                print_format.format('SQ draining', self.attr.sq_draining) +\
*/
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_En_SQD_async_notify, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":866
 *                print_format.format('PKey index', self.attr.pkey_index) +\
 *                print_format.format('Alt. PKey index',
 *                                    self.attr.alt_pkey_index) +\             # <<<<<<<<<<<<<<
 *                print_format.format('En. SQD async notify',
 *                                    self.attr.en_sqd_async_notify) +\
*/
  __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":869
 *                print_format.format('En. SQD async notify',
 *                                    self.attr.en_sqd_async_notify) +\
 *                print_format.format('SQ draining', self.attr.sq_draining) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Max RD atomic', self.attr.max_rd_atomic) +\
 *                print_format.format('Max dest. RD atomic',
*/
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.sq_draining); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_SQ_draining, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":868
 *                                    self.attr.alt_pkey_index) +\
 *                print_format.format('En. SQD async notify',
 *                                    self.attr.en_sqd_async_notify) +\             # <<<<<<<<<<<<<<
 *                print_format.format('SQ draining', self.attr.sq_draining) +\
 *                print_format.format('Max RD atomic', self.attr.max_rd_atomic) +\
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":870
 *                                    self.attr.en_sqd_async_notify) +\
 *                print_format.format('SQ draining', self.attr.sq_draining) +\
 *                print_format.format('Max RD atomic', self.attr.max_rd_atomic) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Max dest. RD atomic',
 *                                    self.attr.max_dest_rd_atomic) +\
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.max_rd_atomic); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 870, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_Max_RD_atomic, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 870, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":869
 *                print_format.format('En. SQD async notify',
 *                                    self.attr.en_sqd_async_notify) +\
 *                print_format.format('SQ draining', self.attr.sq_draining) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Max RD atomic', self.attr.max_rd_atomic) +\
 *                print_format.format('Max dest. RD atomic',
*/
  __pyx_t_5 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":872
 *                print_format.format('Max RD atomic', self.attr.max_rd_atomic) +\
 *                print_format.format('Max dest. RD atomic',
 *                                    self.attr.max_dest_rd_atomic) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Min RNR timer', self.attr.min_rnr_timer) +\
 *                print_format.format('Port number', self.attr.port_num) +\
*/
  __pyx_t_4 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.max_dest_rd_atomic); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "pyverbs/qp.pyx":871
 *                print_format.format('SQ draining', self.attr.sq_draining) +\
 *                print_format.format('Max RD atomic', self.attr.max_rd_atomic) +\
 *                print_format.format('Max dest. RD atomic',             # <<<<<<<<<<<<<<
 *                                    self.attr.max_dest_rd_atomic) +\
 *                print_format.format('Min RNR timer', self.attr.min_rnr_timer) +\
*/
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_Max_dest_RD_atomic, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":870
 *                                    self.attr.en_sqd_async_notify) +\
 *                print_format.format('SQ draining', self.attr.sq_draining) +\
 *                print_format.format('Max RD atomic', self.attr.max_rd_atomic) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Max dest. RD atomic',
 *                                    self.attr.max_dest_rd_atomic) +\
*/
  __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 870, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":873
 *                print_format.format('Max dest. RD atomic',
 *                                    self.attr.max_dest_rd_atomic) +\
 *                print_format.format('Min RNR timer', self.attr.min_rnr_timer) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Port number', self.attr.port_num) +\
 *                print_format.format('Timeout', self.attr.timeout) +\
*/
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.min_rnr_timer); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_Min_RNR_timer, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":872
 *                print_format.format('Max RD atomic', self.attr.max_rd_atomic) +\
 *                print_format.format('Max dest. RD atomic',
 *                                    self.attr.max_dest_rd_atomic) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Min RNR timer', self.attr.min_rnr_timer) +\
 *                print_format.format('Port number', self.attr.port_num) +\
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":874
 *                                    self.attr.max_dest_rd_atomic) +\
 *                print_format.format('Min RNR timer', self.attr.min_rnr_timer) +\
 *                print_format.format('Port number', self.attr.port_num) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Timeout', self.attr.timeout) +\
 *                print_format.format('Retry counter', self.attr.retry_cnt) +\
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.port_num); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 874, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_Port_number, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 874, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":873
 *                print_format.format('Max dest. RD atomic',
 *                                    self.attr.max_dest_rd_atomic) +\
 *                print_format.format('Min RNR timer', self.attr.min_rnr_timer) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Port number', self.attr.port_num) +\
 *                print_format.format('Timeout', self.attr.timeout) +\
*/
  __pyx_t_5 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":875
 *                print_format.format('Min RNR timer', self.attr.min_rnr_timer) +\
 *                print_format.format('Port number', self.attr.port_num) +\
 *                print_format.format('Timeout', self.attr.timeout) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Retry counter', self.attr.retry_cnt) +\
 *                print_format.format('RNR retry', self.attr.rnr_retry) +\
*/
  __pyx_t_4 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.timeout); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_n_u_Timeout, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":874
 *                                    self.attr.max_dest_rd_atomic) +\
 *                print_format.format('Min RNR timer', self.attr.min_rnr_timer) +\
 *                print_format.format('Port number', self.attr.port_num) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Timeout', self.attr.timeout) +\
 *                print_format.format('Retry counter', self.attr.retry_cnt) +\
*/
  __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 874, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":876
 *                print_format.format('Port number', self.attr.port_num) +\
 *                print_format.format('Timeout', self.attr.timeout) +\
 *                print_format.format('Retry counter', self.attr.retry_cnt) +\             # <<<<<<<<<<<<<<
 *                print_format.format('RNR retry', self.attr.rnr_retry) +\
 *                print_format.format('Alt. port number',
*/
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.retry_cnt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_Retry_counter, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":875
 *                print_format.format('Min RNR timer', self.attr.min_rnr_timer) +\
 *                print_format.format('Port number', self.attr.port_num) +\
 *                print_format.format('Timeout', self.attr.timeout) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Retry counter', self.attr.retry_cnt) +\
 *                print_format.format('RNR retry', self.attr.rnr_retry) +\
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":877
 *                print_format.format('Timeout', self.attr.timeout) +\
 *                print_format.format('Retry counter', self.attr.retry_cnt) +\
 *                print_format.format('RNR retry', self.attr.rnr_retry) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Alt. port number',
 *                                    self.attr.alt_port_num) +\
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.rnr_retry); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 877, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_RNR_retry, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 877, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":876
 *                print_format.format('Port number', self.attr.port_num) +\
 *                print_format.format('Timeout', self.attr.timeout) +\
 *                print_format.format('Retry counter', self.attr.retry_cnt) +\             # <<<<<<<<<<<<<<
 *                print_format.format('RNR retry', self.attr.rnr_retry) +\
 *                print_format.format('Alt. port number',
*/
  __pyx_t_5 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":879
 *                print_format.format('RNR retry', self.attr.rnr_retry) +\
 *                print_format.format('Alt. port number',
 *                                    self.attr.alt_port_num) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Alt. timeout', self.attr.alt_timeout) +\
 *                print_format.format('Rate limit', self.attr.rate_limit)
*/
  __pyx_t_4 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.alt_port_num); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "pyverbs/qp.pyx":878
 *                print_format.format('Retry counter', self.attr.retry_cnt) +\
 *                print_format.format('RNR retry', self.attr.rnr_retry) +\
 *                print_format.format('Alt. port number',             # <<<<<<<<<<<<<<
 *                                    self.attr.alt_port_num) +\
 *                print_format.format('Alt. timeout', self.attr.alt_timeout) +\
*/
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_Alt_port_number, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":877
 *                print_format.format('Timeout', self.attr.timeout) +\
 *                print_format.format('Retry counter', self.attr.retry_cnt) +\
 *                print_format.format('RNR retry', self.attr.rnr_retry) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Alt. port number',
 *                                    self.attr.alt_port_num) +\
*/
  __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 877, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":880
 *                print_format.format('Alt. port number',
 *                                    self.attr.alt_port_num) +\
 *                print_format.format('Alt. timeout', self.attr.alt_timeout) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Rate limit', self.attr.rate_limit)
 * 
*/
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_char(__pyx_v_self->attr.alt_timeout); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_Alt_timeout, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":879
 *                print_format.format('RNR retry', self.attr.rnr_retry) +\
 *                print_format.format('Alt. port number',
 *                                    self.attr.alt_port_num) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Alt. timeout', self.attr.alt_timeout) +\
 *                print_format.format('Rate limit', self.attr.rate_limit)
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":881
 *                                    self.attr.alt_port_num) +\
 *                print_format.format('Alt. timeout', self.attr.alt_timeout) +\
 *                print_format.format('Rate limit', self.attr.rate_limit)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.rate_limit); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_Rate_limit, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":880
 *                print_format.format('Alt. port number',
 *                                    self.attr.alt_port_num) +\
 *                print_format.format('Alt. timeout', self.attr.alt_timeout) +\             # <<<<<<<<<<<<<<
 *                print_format.format('Rate limit', self.attr.rate_limit)
 * 
*/
  __pyx_t_5 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":783
 *         self.attr.rate_limit = val
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         print_format = '{:22}: {:<20}\n'
 *         ah_format = '    {:22}: {:<20}\n'
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_print_format);
  __Pyx_XDECREF(__pyx_v_ah_format);
  __Pyx_XDECREF(__pyx_v_ident_format);
  __Pyx_XDECREF(__pyx_v_global_ah);
  __Pyx_XDECREF(__pyx_v_alt_global_ah);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_6QPAttr_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_6QPAttr_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_6__reduce_cython__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_Pickling_of_struct_members_such_2, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_6QPAttr_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_6QPAttr_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_6QPAttr_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_6QPAttr_8__setstate_cython__(((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_6QPAttr_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_Pickling_of_struct_members_such_2, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPAttr.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":885
 * 
 * cdef class ECE(PyverbsCM):
 *     def __init__(self, vendor_id=0, options=0, comp_mask=0):             # <<<<<<<<<<<<<<
 *         """
 *         :param vendor_id: Unique identifier of the provider vendor.
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_3ECE_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2qp_3ECE___init__, "\n        :param vendor_id: Unique identifier of the provider vendor.\n        :param options: Provider specific attributes which are supported or\n                        needed to be enabled by ECE users.\n        :param comp_mask: A bitmask specifying which ECE options should be\n                          valid.\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_7pyverbs_2qp_3ECE___init__;
#endif
static int __pyx_pw_7pyverbs_2qp_3ECE_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_vendor_id = 0;
  PyObject *__pyx_v_options = 0;
  PyObject *__pyx_v_comp_mask = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_vendor_id,&__pyx_mstate_global->__pyx_n_u_options,&__pyx_mstate_global->__pyx_n_u_comp_mask,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 885, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 885, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 885, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 885, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 885, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    } else {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 885, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 885, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 885, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    }
    __pyx_v_vendor_id = values[0];
    __pyx_v_options = values[1];
    __pyx_v_comp_mask = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 3, __pyx_nargs); __PYX_ERR(0, 885, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.ECE.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_3ECE___init__(((struct __pyx_obj_7pyverbs_2qp_ECE *)__pyx_v_self), __pyx_v_vendor_id, __pyx_v_options, __pyx_v_comp_mask);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_3ECE___init__(struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_self, PyObject *__pyx_v_vendor_id, PyObject *__pyx_v_options, PyObject *__pyx_v_comp_mask) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  uint32_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyverbs/qp.pyx":893
 *                           valid.
 *         """
 *         super().__init__()             # <<<<<<<<<<<<<<
 *         self.ece.vendor_id = vendor_id
 *         self.ece.options = options
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_ECE), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 893, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 893, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":894
 *         """
 *         super().__init__()
 *         self.ece.vendor_id = vendor_id             # <<<<<<<<<<<<<<
 *         self.ece.options = options
 *         self.ece.comp_mask = comp_mask
*/
  __pyx_t_6 = __Pyx_PyLong_As_uint32_t(__pyx_v_vendor_id); if (unlikely((__pyx_t_6 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 894, __pyx_L1_error)
  __pyx_v_self->ece.vendor_id = __pyx_t_6;

  /* "pyverbs/qp.pyx":895
 *         super().__init__()
 *         self.ece.vendor_id = vendor_id
 *         self.ece.options = options             # <<<<<<<<<<<<<<
 *         self.ece.comp_mask = comp_mask
 * 
*/
  __pyx_t_6 = __Pyx_PyLong_As_uint32_t(__pyx_v_options); if (unlikely((__pyx_t_6 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 895, __pyx_L1_error)
  __pyx_v_self->ece.options = __pyx_t_6;

  /* "pyverbs/qp.pyx":896
 *         self.ece.vendor_id = vendor_id
 *         self.ece.options = options
 *         self.ece.comp_mask = comp_mask             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_6 = __Pyx_PyLong_As_uint32_t(__pyx_v_comp_mask); if (unlikely((__pyx_t_6 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 896, __pyx_L1_error)
  __pyx_v_self->ece.comp_mask = __pyx_t_6;

  /* "pyverbs/qp.pyx":885
 * 
 * cdef class ECE(PyverbsCM):
 *     def __init__(self, vendor_id=0, options=0, comp_mask=0):             # <<<<<<<<<<<<<<
 *         """
 *         :param vendor_id: Unique identifier of the provider vendor.
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.qp.ECE.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":898
 *         self.ece.comp_mask = comp_mask
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def vendor_id(self):
 *         return self.ece.vendor_id
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_3ECE_9vendor_id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_3ECE_9vendor_id_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_3ECE_9vendor_id___get__(((struct __pyx_obj_7pyverbs_2qp_ECE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_3ECE_9vendor_id___get__(struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":900
 *     @property
 *     def vendor_id(self):
 *         return self.ece.vendor_id             # <<<<<<<<<<<<<<
 *     @vendor_id.setter
 *     def vendor_id(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_uint32_t(__pyx_v_self->ece.vendor_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 900, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":898
 *         self.ece.comp_mask = comp_mask
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def vendor_id(self):
 *         return self.ece.vendor_id
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.ECE.vendor_id.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":901
 *     def vendor_id(self):
 *         return self.ece.vendor_id
 *     @vendor_id.setter             # <<<<<<<<<<<<<<
 *     def vendor_id(self, val):
 *         self.ece.vendor_id = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_3ECE_9vendor_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_3ECE_9vendor_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_3ECE_9vendor_id_2__set__(((struct __pyx_obj_7pyverbs_2qp_ECE *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_3ECE_9vendor_id_2__set__(struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  uint32_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":903
 *     @vendor_id.setter
 *     def vendor_id(self, val):
 *         self.ece.vendor_id = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_uint32_t(__pyx_v_val); if (unlikely((__pyx_t_1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 903, __pyx_L1_error)
  __pyx_v_self->ece.vendor_id = __pyx_t_1;

  /* "pyverbs/qp.pyx":901
 *     def vendor_id(self):
 *         return self.ece.vendor_id
 *     @vendor_id.setter             # <<<<<<<<<<<<<<
 *     def vendor_id(self, val):
 *         self.ece.vendor_id = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.ECE.vendor_id.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":905
 *         self.ece.vendor_id = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def options(self):
 *         return self.ece.options
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_3ECE_7options_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_3ECE_7options_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_3ECE_7options___get__(((struct __pyx_obj_7pyverbs_2qp_ECE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_3ECE_7options___get__(struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":907
 *     @property
 *     def options(self):
 *         return self.ece.options             # <<<<<<<<<<<<<<
 *     @options.setter
 *     def options(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_uint32_t(__pyx_v_self->ece.options); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":905
 *         self.ece.vendor_id = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def options(self):
 *         return self.ece.options
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.ECE.options.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":908
 *     def options(self):
 *         return self.ece.options
 *     @options.setter             # <<<<<<<<<<<<<<
 *     def options(self, val):
 *         self.ece.options = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_3ECE_7options_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_3ECE_7options_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_3ECE_7options_2__set__(((struct __pyx_obj_7pyverbs_2qp_ECE *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_3ECE_7options_2__set__(struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  uint32_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":910
 *     @options.setter
 *     def options(self, val):
 *         self.ece.options = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_uint32_t(__pyx_v_val); if (unlikely((__pyx_t_1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 910, __pyx_L1_error)
  __pyx_v_self->ece.options = __pyx_t_1;

  /* "pyverbs/qp.pyx":908
 *     def options(self):
 *         return self.ece.options
 *     @options.setter             # <<<<<<<<<<<<<<
 *     def options(self, val):
 *         self.ece.options = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.ECE.options.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":912
 *         self.ece.options = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def comp_mask(self):
 *         return self.ece.comp_mask
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_3ECE_9comp_mask_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_3ECE_9comp_mask_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_3ECE_9comp_mask___get__(((struct __pyx_obj_7pyverbs_2qp_ECE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_3ECE_9comp_mask___get__(struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":914
 *     @property
 *     def comp_mask(self):
 *         return self.ece.comp_mask             # <<<<<<<<<<<<<<
 *     @comp_mask.setter
 *     def comp_mask(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_uint32_t(__pyx_v_self->ece.comp_mask); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 914, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":912
 *         self.ece.options = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def comp_mask(self):
 *         return self.ece.comp_mask
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.ECE.comp_mask.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":915
 *     def comp_mask(self):
 *         return self.ece.comp_mask
 *     @comp_mask.setter             # <<<<<<<<<<<<<<
 *     def comp_mask(self, val):
 *         self.ece.comp_mask = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_3ECE_9comp_mask_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_3ECE_9comp_mask_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_3ECE_9comp_mask_2__set__(((struct __pyx_obj_7pyverbs_2qp_ECE *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_3ECE_9comp_mask_2__set__(struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  uint32_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":917
 *     @comp_mask.setter
 *     def comp_mask(self, val):
 *         self.ece.comp_mask = val             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
*/
  __pyx_t_1 = __Pyx_PyLong_As_uint32_t(__pyx_v_val); if (unlikely((__pyx_t_1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 917, __pyx_L1_error)
  __pyx_v_self->ece.comp_mask = __pyx_t_1;

  /* "pyverbs/qp.pyx":915
 *     def comp_mask(self):
 *         return self.ece.comp_mask
 *     @comp_mask.setter             # <<<<<<<<<<<<<<
 *     def comp_mask(self, val):
 *         self.ece.comp_mask = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.ECE.comp_mask.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":919
 *         self.ece.comp_mask = val
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         print_format = '{:22}: 0x{:<20x}\n'
 *         return 'ECE:\n' +\
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_3ECE_3__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_3ECE_3__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_3ECE_2__str__(((struct __pyx_obj_7pyverbs_2qp_ECE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_3ECE_2__str__(struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_self) {
  PyObject *__pyx_v_print_format = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "pyverbs/qp.pyx":920
 * 
 *     def __str__(self):
 *         print_format = '{:22}: 0x{:<20x}\n'             # <<<<<<<<<<<<<<
 *         return 'ECE:\n' +\
 *             print_format.format('Vendor ID', self.ece.vendor_id) +\
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_22_0x_20x);
  __pyx_v_print_format = __pyx_mstate_global->__pyx_kp_u_22_0x_20x;

  /* "pyverbs/qp.pyx":921
 *     def __str__(self):
 *         print_format = '{:22}: 0x{:<20x}\n'
 *         return 'ECE:\n' +\             # <<<<<<<<<<<<<<
 *             print_format.format('Vendor ID', self.ece.vendor_id) +\
 *             print_format.format('Options', self.ece.options) +\
*/
  __Pyx_XDECREF(__pyx_r);

  /* "pyverbs/qp.pyx":922
 *         print_format = '{:22}: 0x{:<20x}\n'
 *         return 'ECE:\n' +\
 *             print_format.format('Vendor ID', self.ece.vendor_id) +\             # <<<<<<<<<<<<<<
 *             print_format.format('Options', self.ece.options) +\
 *             print_format.format('Comp Mask', self.ece.comp_mask)
*/
  __pyx_t_1 = __Pyx_PyLong_From_uint32_t(__pyx_v_self->ece.vendor_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 922, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_Vendor_ID, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 922, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":921
 *     def __str__(self):
 *         print_format = '{:22}: 0x{:<20x}\n'
 *         return 'ECE:\n' +\             # <<<<<<<<<<<<<<
 *             print_format.format('Vendor ID', self.ece.vendor_id) +\
 *             print_format.format('Options', self.ece.options) +\
*/
  __pyx_t_1 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_kp_u_ECE, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":923
 *         return 'ECE:\n' +\
 *             print_format.format('Vendor ID', self.ece.vendor_id) +\
 *             print_format.format('Options', self.ece.options) +\             # <<<<<<<<<<<<<<
 *             print_format.format('Comp Mask', self.ece.comp_mask)
 * 
*/
  __pyx_t_2 = __Pyx_PyLong_From_uint32_t(__pyx_v_self->ece.options); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 923, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_n_u_Options, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 923, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":922
 *         print_format = '{:22}: 0x{:<20x}\n'
 *         return 'ECE:\n' +\
 *             print_format.format('Vendor ID', self.ece.vendor_id) +\             # <<<<<<<<<<<<<<
 *             print_format.format('Options', self.ece.options) +\
 *             print_format.format('Comp Mask', self.ece.comp_mask)
*/
  __pyx_t_2 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 922, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":924
 *             print_format.format('Vendor ID', self.ece.vendor_id) +\
 *             print_format.format('Options', self.ece.options) +\
 *             print_format.format('Comp Mask', self.ece.comp_mask)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_3 = __Pyx_PyLong_From_uint32_t(__pyx_v_self->ece.comp_mask); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_Comp_Mask, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":923
 *         return 'ECE:\n' +\
 *             print_format.format('Vendor ID', self.ece.vendor_id) +\
 *             print_format.format('Options', self.ece.options) +\             # <<<<<<<<<<<<<<
 *             print_format.format('Comp Mask', self.ece.comp_mask)
 * 
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 923, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":919
 *         self.ece.comp_mask = val
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         print_format = '{:22}: 0x{:<20x}\n'
 *         return 'ECE:\n' +\
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyverbs.qp.ECE.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_print_format);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.ece must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_3ECE_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_3ECE_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_3ECE_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_3ECE_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2qp_3ECE_4__reduce_cython__(((struct __pyx_obj_7pyverbs_2qp_ECE *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_3ECE_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.ece must be explicitly requested with @auto_pickle(True)"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "Pickling of struct members such as self.ece must be explicitly requested with @auto_pickle(True)"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_Pickling_of_struct_members_such_3, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.ece must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.ECE.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.ece must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.ece must be explicitly requested with @auto_pickle(True)"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_3ECE_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_3ECE_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_3ECE_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_3ECE_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.ECE.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_3ECE_6__setstate_cython__(((struct __pyx_obj_7pyverbs_2qp_ECE *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_3ECE_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "Pickling of struct members such as self.ece must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "Pickling of struct members such as self.ece must be explicitly requested with @auto_pickle(True)"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_Pickling_of_struct_members_such_3, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.ece must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.ece must be explicitly requested with @auto_pickle(True)"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.ECE.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":928
 * 
 * cdef class QP(PyverbsCM):
 *     def __init__(self, object creator not None, object init_attr not None,             # <<<<<<<<<<<<<<
 *                  QPAttr qp_attr=None):
 *         """
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_2QP_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2qp_2QP___init__, "\n        Initializes a QP object and performs state transitions according to\n        user request.\n        A C ibv_qp object will be created using the provided init_attr.\n        If a qp_attr object is provided, pyverbs will consider this a hint to\n        transit the QP's state as far as possible towards RTS:\n        - In case of UD and Raw Packet QP types, if a qp_attr is provided the\n          QP will be returned in RTS state.\n        - In case of connected QPs (RC, UC), remote QPN is needed for INIT2RTR\n          transition, so if a qp_attr is provided, the QP will be returned in\n          INIT state.\n        :param creator: The object creating the QP. Can be of type PD so\n                        ibv_create_qp will be used or of type Context, so\n                        ibv_create_qp_ex will be used.\n        :param init_attr: QP initial attributes of type QPInitAttr (when\n                          created using PD) or QPInitAttrEx (when created\n                          using Context).\n        :param qp_attr: Optional QPAttr object. Will be used for QP state\n                        transitions after creation.\n        :return: An initialized QP object\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_7pyverbs_2qp_2QP___init__;
#endif
static int __pyx_pw_7pyverbs_2qp_2QP_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_creator = 0;
  PyObject *__pyx_v_init_attr = 0;
  struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_qp_attr = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_creator,&__pyx_mstate_global->__pyx_n_u_init_attr,&__pyx_mstate_global->__pyx_n_u_qp_attr,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 928, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 928, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 928, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 928, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 928, __pyx_L3_error)

      /* "pyverbs/qp.pyx":929
 * cdef class QP(PyverbsCM):
 *     def __init__(self, object creator not None, object init_attr not None,
 *                  QPAttr qp_attr=None):             # <<<<<<<<<<<<<<
 *         """
 *         Initializes a QP object and performs state transitions according to
*/
      if (!values[2]) values[2] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2qp_QPAttr *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, i); __PYX_ERR(0, 928, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 928, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 928, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 928, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[2]) values[2] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2qp_QPAttr *)Py_None));
    }
    __pyx_v_creator = values[0];
    __pyx_v_init_attr = values[1];
    __pyx_v_qp_attr = ((struct __pyx_obj_7pyverbs_2qp_QPAttr *)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, __pyx_nargs); __PYX_ERR(0, 928, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QP.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(((PyObject *)__pyx_v_creator) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "creator"); __PYX_ERR(0, 928, __pyx_L1_error)
  }
  if (unlikely(((PyObject *)__pyx_v_init_attr) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "init_attr"); __PYX_ERR(0, 928, __pyx_L1_error)
  }
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_qp_attr), __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPAttr, 1, "qp_attr", 0))) __PYX_ERR(0, 929, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp_2QP___init__(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self), __pyx_v_creator, __pyx_v_init_attr, __pyx_v_qp_attr);

  /* "pyverbs/qp.pyx":928
 * 
 * cdef class QP(PyverbsCM):
 *     def __init__(self, object creator not None, object init_attr not None,             # <<<<<<<<<<<<<<
 *                  QPAttr qp_attr=None):
 *         """
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_2QP___init__(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, PyObject *__pyx_v_creator, PyObject *__pyx_v_init_attr, struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_qp_attr) {
  struct __pyx_obj_7pyverbs_2pd_PD *__pyx_v_pd = 0;
  struct __pyx_obj_7pyverbs_6device_Context *__pyx_v_ctx = 0;
  struct __pyx_obj_7pyverbs_4xrcd_XRCD *__pyx_v_xrcd = NULL;
  struct __pyx_obj_7pyverbs_3srq_SRQ *__pyx_v_srq = NULL;
  PyObject *__pyx_v_funcs = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyverbs/qp.pyx":953
 *         cdef PD pd
 *         cdef Context ctx
 *         super().__init__()             # <<<<<<<<<<<<<<
 *         self.mws = weakref.WeakSet()
 *         self.flows = weakref.WeakSet()
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QP), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 953, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 953, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":954
 *         cdef Context ctx
 *         super().__init__()
 *         self.mws = weakref.WeakSet()             # <<<<<<<<<<<<<<
 *         self.flows = weakref.WeakSet()
 *         self.dr_actions = weakref.WeakSet()
*/
  __pyx_t_3 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_weakref); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_WeakSet); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 954, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_3);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 954, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->mws);
  __Pyx_DECREF(__pyx_v_self->mws);
  __pyx_v_self->mws = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":955
 *         super().__init__()
 *         self.mws = weakref.WeakSet()
 *         self.flows = weakref.WeakSet()             # <<<<<<<<<<<<<<
 *         self.dr_actions = weakref.WeakSet()
 *         self.update_cqs(init_attr)
*/
  __pyx_t_4 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_weakref); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_WeakSet); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    assert(__pyx_t_4);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 955, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->flows);
  __Pyx_DECREF(__pyx_v_self->flows);
  __pyx_v_self->flows = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":956
 *         self.mws = weakref.WeakSet()
 *         self.flows = weakref.WeakSet()
 *         self.dr_actions = weakref.WeakSet()             # <<<<<<<<<<<<<<
 *         self.update_cqs(init_attr)
 *         # QP initialization was not done by the provider, we should do it here
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_weakref); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 956, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_WeakSet); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 956, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 956, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->dr_actions);
  __Pyx_DECREF(__pyx_v_self->dr_actions);
  __pyx_v_self->dr_actions = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":957
 *         self.flows = weakref.WeakSet()
 *         self.dr_actions = weakref.WeakSet()
 *         self.update_cqs(init_attr)             # <<<<<<<<<<<<<<
 *         # QP initialization was not done by the provider, we should do it here
 *         if self.qp == NULL:
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_7pyverbs_2qp_QP *)__pyx_v_self->__pyx_base.__pyx_vtab)->update_cqs(__pyx_v_self, __pyx_v_init_attr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":959
 *         self.update_cqs(init_attr)
 *         # QP initialization was not done by the provider, we should do it here
 *         if self.qp == NULL:             # <<<<<<<<<<<<<<
 *             # In order to use cdef'd methods, a proper casting must be done,
 *             # let's infer the type.
*/
  __pyx_t_6 = (__pyx_v_self->qp == NULL);
  if (__pyx_t_6) {

    /* "pyverbs/qp.pyx":962
 *             # In order to use cdef'd methods, a proper casting must be done,
 *             # let's infer the type.
 *             if issubclass(type(creator), Context):             # <<<<<<<<<<<<<<
 *                 self._create_qp_ex(creator, init_attr)
 *                 if self.qp == NULL:
*/
    __pyx_t_6 = PyObject_IsSubclass(((PyObject *)Py_TYPE(__pyx_v_creator)), ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_6device_Context)); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 962, __pyx_L1_error)
    if (__pyx_t_6) {

      /* "pyverbs/qp.pyx":963
 *             # let's infer the type.
 *             if issubclass(type(creator), Context):
 *                 self._create_qp_ex(creator, init_attr)             # <<<<<<<<<<<<<<
 *                 if self.qp == NULL:
 *                     raise PyverbsRDMAErrno('Failed to create QP')
*/
      __pyx_t_3 = ((PyObject *)__pyx_v_self);
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_5 = 0;
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_creator, __pyx_v_init_attr};
        __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_create_qp_ex, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 963, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyverbs/qp.pyx":964
 *             if issubclass(type(creator), Context):
 *                 self._create_qp_ex(creator, init_attr)
 *                 if self.qp == NULL:             # <<<<<<<<<<<<<<
 *                     raise PyverbsRDMAErrno('Failed to create QP')
 *                 ctx = <Context>creator
*/
      __pyx_t_6 = (__pyx_v_self->qp == NULL);
      if (unlikely(__pyx_t_6)) {

        /* "pyverbs/qp.pyx":965
 *                 self._create_qp_ex(creator, init_attr)
 *                 if self.qp == NULL:
 *                     raise PyverbsRDMAErrno('Failed to create QP')             # <<<<<<<<<<<<<<
 *                 ctx = <Context>creator
 *                 self.context = ctx
*/
        __pyx_t_3 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 965, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Failed_to_create_QP};
          __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 965, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
        }
        __Pyx_Raise(__pyx_t_1, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __PYX_ERR(0, 965, __pyx_L1_error)

        /* "pyverbs/qp.pyx":964
 *             if issubclass(type(creator), Context):
 *                 self._create_qp_ex(creator, init_attr)
 *                 if self.qp == NULL:             # <<<<<<<<<<<<<<
 *                     raise PyverbsRDMAErrno('Failed to create QP')
 *                 ctx = <Context>creator
*/
      }

      /* "pyverbs/qp.pyx":966
 *                 if self.qp == NULL:
 *                     raise PyverbsRDMAErrno('Failed to create QP')
 *                 ctx = <Context>creator             # <<<<<<<<<<<<<<
 *                 self.context = ctx
 *                 ctx.add_ref(self)
*/
      __pyx_t_1 = __pyx_v_creator;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_v_ctx = ((struct __pyx_obj_7pyverbs_6device_Context *)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "pyverbs/qp.pyx":967
 *                     raise PyverbsRDMAErrno('Failed to create QP')
 *                 ctx = <Context>creator
 *                 self.context = ctx             # <<<<<<<<<<<<<<
 *                 ctx.add_ref(self)
 *                 if init_attr.pd is not None:
*/
      __Pyx_INCREF((PyObject *)__pyx_v_ctx);
      __Pyx_GIVEREF((PyObject *)__pyx_v_ctx);
      __Pyx_GOTREF(__pyx_v_self->context);
      __Pyx_DECREF(__pyx_v_self->context);
      __pyx_v_self->context = ((PyObject *)__pyx_v_ctx);

      /* "pyverbs/qp.pyx":968
 *                 ctx = <Context>creator
 *                 self.context = ctx
 *                 ctx.add_ref(self)             # <<<<<<<<<<<<<<
 *                 if init_attr.pd is not None:
 *                     pd = <PD>init_attr.pd
*/
      __pyx_t_1 = ((struct __pyx_vtabstruct_7pyverbs_6device_Context *)__pyx_v_ctx->__pyx_base.__pyx_vtab)->add_ref(__pyx_v_ctx, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 968, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyverbs/qp.pyx":969
 *                 self.context = ctx
 *                 ctx.add_ref(self)
 *                 if init_attr.pd is not None:             # <<<<<<<<<<<<<<
 *                     pd = <PD>init_attr.pd
 *                     pd.add_ref(self)
*/
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_init_attr, __pyx_mstate_global->__pyx_n_u_pd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 969, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = (__pyx_t_1 != Py_None);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_6) {

        /* "pyverbs/qp.pyx":970
 *                 ctx.add_ref(self)
 *                 if init_attr.pd is not None:
 *                     pd = <PD>init_attr.pd             # <<<<<<<<<<<<<<
 *                     pd.add_ref(self)
 *                     self.pd = pd
*/
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_init_attr, __pyx_mstate_global->__pyx_n_u_pd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 970, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __pyx_t_1;
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_v_pd = ((struct __pyx_obj_7pyverbs_2pd_PD *)__pyx_t_2);
        __pyx_t_2 = 0;

        /* "pyverbs/qp.pyx":971
 *                 if init_attr.pd is not None:
 *                     pd = <PD>init_attr.pd
 *                     pd.add_ref(self)             # <<<<<<<<<<<<<<
 *                     self.pd = pd
 *                 if init_attr.xrcd is not None:
*/
        __pyx_t_2 = ((struct __pyx_vtabstruct_7pyverbs_2pd_PD *)__pyx_v_pd->__pyx_base.__pyx_vtab)->add_ref(__pyx_v_pd, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 971, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "pyverbs/qp.pyx":972
 *                     pd = <PD>init_attr.pd
 *                     pd.add_ref(self)
 *                     self.pd = pd             # <<<<<<<<<<<<<<
 *                 if init_attr.xrcd is not None:
 *                     xrcd = <XRCD>init_attr.xrcd
*/
        __Pyx_INCREF((PyObject *)__pyx_v_pd);
        __Pyx_GIVEREF((PyObject *)__pyx_v_pd);
        __Pyx_GOTREF(__pyx_v_self->pd);
        __Pyx_DECREF(__pyx_v_self->pd);
        __pyx_v_self->pd = ((PyObject *)__pyx_v_pd);

        /* "pyverbs/qp.pyx":969
 *                 self.context = ctx
 *                 ctx.add_ref(self)
 *                 if init_attr.pd is not None:             # <<<<<<<<<<<<<<
 *                     pd = <PD>init_attr.pd
 *                     pd.add_ref(self)
*/
      }

      /* "pyverbs/qp.pyx":973
 *                     pd.add_ref(self)
 *                     self.pd = pd
 *                 if init_attr.xrcd is not None:             # <<<<<<<<<<<<<<
 *                     xrcd = <XRCD>init_attr.xrcd
 *                     xrcd.add_ref(self)
*/
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_init_attr, __pyx_mstate_global->__pyx_n_u_xrcd); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 973, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = (__pyx_t_2 != Py_None);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_6) {

        /* "pyverbs/qp.pyx":974
 *                     self.pd = pd
 *                 if init_attr.xrcd is not None:
 *                     xrcd = <XRCD>init_attr.xrcd             # <<<<<<<<<<<<<<
 *                     xrcd.add_ref(self)
 *                     self.xrcd = xrcd
*/
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_init_attr, __pyx_mstate_global->__pyx_n_u_xrcd); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 974, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = __pyx_t_2;
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_xrcd = ((struct __pyx_obj_7pyverbs_4xrcd_XRCD *)__pyx_t_1);
        __pyx_t_1 = 0;

        /* "pyverbs/qp.pyx":975
 *                 if init_attr.xrcd is not None:
 *                     xrcd = <XRCD>init_attr.xrcd
 *                     xrcd.add_ref(self)             # <<<<<<<<<<<<<<
 *                     self.xrcd = xrcd
 * 
*/
        __pyx_t_1 = ((struct __pyx_vtabstruct_7pyverbs_4xrcd_XRCD *)__pyx_v_xrcd->__pyx_base.__pyx_vtab)->add_ref(__pyx_v_xrcd, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 975, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pyverbs/qp.pyx":976
 *                     xrcd = <XRCD>init_attr.xrcd
 *                     xrcd.add_ref(self)
 *                     self.xrcd = xrcd             # <<<<<<<<<<<<<<
 * 
 *             else:
*/
        __Pyx_INCREF((PyObject *)__pyx_v_xrcd);
        __Pyx_GIVEREF((PyObject *)__pyx_v_xrcd);
        __Pyx_GOTREF(__pyx_v_self->xrcd);
        __Pyx_DECREF(__pyx_v_self->xrcd);
        __pyx_v_self->xrcd = ((PyObject *)__pyx_v_xrcd);

        /* "pyverbs/qp.pyx":973
 *                     pd.add_ref(self)
 *                     self.pd = pd
 *                 if init_attr.xrcd is not None:             # <<<<<<<<<<<<<<
 *                     xrcd = <XRCD>init_attr.xrcd
 *                     xrcd.add_ref(self)
*/
      }

      /* "pyverbs/qp.pyx":962
 *             # In order to use cdef'd methods, a proper casting must be done,
 *             # let's infer the type.
 *             if issubclass(type(creator), Context):             # <<<<<<<<<<<<<<
 *                 self._create_qp_ex(creator, init_attr)
 *                 if self.qp == NULL:
*/
      goto __pyx_L4;
    }

    /* "pyverbs/qp.pyx":979
 * 
 *             else:
 *                 self._create_qp(creator, init_attr)             # <<<<<<<<<<<<<<
 *                 if self.qp == NULL:
 *                     raise PyverbsRDMAErrno('Failed to create QP')
*/
    /*else*/ {
      __pyx_t_2 = ((PyObject *)__pyx_v_self);
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_5 = 0;
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_v_creator, __pyx_v_init_attr};
        __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_create_qp, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 979, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyverbs/qp.pyx":980
 *             else:
 *                 self._create_qp(creator, init_attr)
 *                 if self.qp == NULL:             # <<<<<<<<<<<<<<
 *                     raise PyverbsRDMAErrno('Failed to create QP')
 *                 pd = <PD>creator
*/
      __pyx_t_6 = (__pyx_v_self->qp == NULL);
      if (unlikely(__pyx_t_6)) {

        /* "pyverbs/qp.pyx":981
 *                 self._create_qp(creator, init_attr)
 *                 if self.qp == NULL:
 *                     raise PyverbsRDMAErrno('Failed to create QP')             # <<<<<<<<<<<<<<
 *                 pd = <PD>creator
 *                 self.pd = pd
*/
        __pyx_t_2 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 981, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
          assert(__pyx_t_2);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Failed_to_create_QP};
          __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 981, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
        }
        __Pyx_Raise(__pyx_t_1, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __PYX_ERR(0, 981, __pyx_L1_error)

        /* "pyverbs/qp.pyx":980
 *             else:
 *                 self._create_qp(creator, init_attr)
 *                 if self.qp == NULL:             # <<<<<<<<<<<<<<
 *                     raise PyverbsRDMAErrno('Failed to create QP')
 *                 pd = <PD>creator
*/
      }

      /* "pyverbs/qp.pyx":982
 *                 if self.qp == NULL:
 *                     raise PyverbsRDMAErrno('Failed to create QP')
 *                 pd = <PD>creator             # <<<<<<<<<<<<<<
 *                 self.pd = pd
 *                 pd.add_ref(self)
*/
      __pyx_t_1 = __pyx_v_creator;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_v_pd = ((struct __pyx_obj_7pyverbs_2pd_PD *)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "pyverbs/qp.pyx":983
 *                     raise PyverbsRDMAErrno('Failed to create QP')
 *                 pd = <PD>creator
 *                 self.pd = pd             # <<<<<<<<<<<<<<
 *                 pd.add_ref(self)
 *                 self.context = None
*/
      __Pyx_INCREF((PyObject *)__pyx_v_pd);
      __Pyx_GIVEREF((PyObject *)__pyx_v_pd);
      __Pyx_GOTREF(__pyx_v_self->pd);
      __Pyx_DECREF(__pyx_v_self->pd);
      __pyx_v_self->pd = ((PyObject *)__pyx_v_pd);

      /* "pyverbs/qp.pyx":984
 *                 pd = <PD>creator
 *                 self.pd = pd
 *                 pd.add_ref(self)             # <<<<<<<<<<<<<<
 *                 self.context = None
 *         if init_attr.srq is not None:
*/
      __pyx_t_1 = ((struct __pyx_vtabstruct_7pyverbs_2pd_PD *)__pyx_v_pd->__pyx_base.__pyx_vtab)->add_ref(__pyx_v_pd, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 984, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyverbs/qp.pyx":985
 *                 self.pd = pd
 *                 pd.add_ref(self)
 *                 self.context = None             # <<<<<<<<<<<<<<
 *         if init_attr.srq is not None:
 *             srq = <SRQ>init_attr.srq
*/
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      __Pyx_GOTREF(__pyx_v_self->context);
      __Pyx_DECREF(__pyx_v_self->context);
      __pyx_v_self->context = Py_None;
    }
    __pyx_L4:;

    /* "pyverbs/qp.pyx":959
 *         self.update_cqs(init_attr)
 *         # QP initialization was not done by the provider, we should do it here
 *         if self.qp == NULL:             # <<<<<<<<<<<<<<
 *             # In order to use cdef'd methods, a proper casting must be done,
 *             # let's infer the type.
*/
  }

  /* "pyverbs/qp.pyx":986
 *                 pd.add_ref(self)
 *                 self.context = None
 *         if init_attr.srq is not None:             # <<<<<<<<<<<<<<
 *             srq = <SRQ>init_attr.srq
 *             srq.add_ref(self)
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_init_attr, __pyx_mstate_global->__pyx_n_u_srq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = (__pyx_t_1 != Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "pyverbs/qp.pyx":987
 *                 self.context = None
 *         if init_attr.srq is not None:
 *             srq = <SRQ>init_attr.srq             # <<<<<<<<<<<<<<
 *             srq.add_ref(self)
 *             self.srq = srq
*/
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_init_attr, __pyx_mstate_global->__pyx_n_u_srq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 987, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __pyx_t_1;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_srq = ((struct __pyx_obj_7pyverbs_3srq_SRQ *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyverbs/qp.pyx":988
 *         if init_attr.srq is not None:
 *             srq = <SRQ>init_attr.srq
 *             srq.add_ref(self)             # <<<<<<<<<<<<<<
 *             self.srq = srq
 * 
*/
    __pyx_t_3 = ((struct __pyx_vtabstruct_7pyverbs_3srq_SRQ *)__pyx_v_srq->__pyx_base.__pyx_vtab)->add_ref(__pyx_v_srq, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 988, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyverbs/qp.pyx":989
 *             srq = <SRQ>init_attr.srq
 *             srq.add_ref(self)
 *             self.srq = srq             # <<<<<<<<<<<<<<
 * 
 *         if qp_attr is not None:
*/
    __Pyx_INCREF((PyObject *)__pyx_v_srq);
    __Pyx_GIVEREF((PyObject *)__pyx_v_srq);
    __Pyx_GOTREF(__pyx_v_self->srq);
    __Pyx_DECREF(__pyx_v_self->srq);
    __pyx_v_self->srq = ((PyObject *)__pyx_v_srq);

    /* "pyverbs/qp.pyx":986
 *                 pd.add_ref(self)
 *                 self.context = None
 *         if init_attr.srq is not None:             # <<<<<<<<<<<<<<
 *             srq = <SRQ>init_attr.srq
 *             srq.add_ref(self)
*/
  }

  /* "pyverbs/qp.pyx":991
 *             self.srq = srq
 * 
 *         if qp_attr is not None:             # <<<<<<<<<<<<<<
 *             funcs = {e.IBV_QPT_RC: self.to_init, e.IBV_QPT_UC: self.to_init,
 *                      e.IBV_QPT_UD: self.to_rts,
*/
  __pyx_t_6 = (((PyObject *)__pyx_v_qp_attr) != Py_None);
  if (__pyx_t_6) {

    /* "pyverbs/qp.pyx":992
 * 
 *         if qp_attr is not None:
 *             funcs = {e.IBV_QPT_RC: self.to_init, e.IBV_QPT_UC: self.to_init,             # <<<<<<<<<<<<<<
 *                      e.IBV_QPT_UD: self.to_rts,
 *                      e.IBV_QPT_XRC_RECV: self.to_init,
*/
    __pyx_t_3 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(IBV_QPT_RC); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_to_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_3, __pyx_t_1, __pyx_t_2) < (0)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(IBV_QPT_UC); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_to_init); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_3, __pyx_t_2, __pyx_t_1) < (0)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyverbs/qp.pyx":993
 *         if qp_attr is not None:
 *             funcs = {e.IBV_QPT_RC: self.to_init, e.IBV_QPT_UC: self.to_init,
 *                      e.IBV_QPT_UD: self.to_rts,             # <<<<<<<<<<<<<<
 *                      e.IBV_QPT_XRC_RECV: self.to_init,
 *                      e.IBV_QPT_XRC_SEND: self.to_init,
*/
    __pyx_t_1 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(IBV_QPT_UD); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_to_rts); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_3, __pyx_t_1, __pyx_t_2) < (0)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":994
 *             funcs = {e.IBV_QPT_RC: self.to_init, e.IBV_QPT_UC: self.to_init,
 *                      e.IBV_QPT_UD: self.to_rts,
 *                      e.IBV_QPT_XRC_RECV: self.to_init,             # <<<<<<<<<<<<<<
 *                      e.IBV_QPT_XRC_SEND: self.to_init,
 *                      e.IBV_QPT_RAW_PACKET: self.to_rts}
*/
    __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(IBV_QPT_XRC_RECV); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 994, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_to_init); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 994, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_3, __pyx_t_2, __pyx_t_1) < (0)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyverbs/qp.pyx":995
 *                      e.IBV_QPT_UD: self.to_rts,
 *                      e.IBV_QPT_XRC_RECV: self.to_init,
 *                      e.IBV_QPT_XRC_SEND: self.to_init,             # <<<<<<<<<<<<<<
 *                      e.IBV_QPT_RAW_PACKET: self.to_rts}
 *             funcs[self.qp.qp_type](qp_attr)
*/
    __pyx_t_1 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(IBV_QPT_XRC_SEND); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 995, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_to_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 995, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_3, __pyx_t_1, __pyx_t_2) < (0)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":996
 *                      e.IBV_QPT_XRC_RECV: self.to_init,
 *                      e.IBV_QPT_XRC_SEND: self.to_init,
 *                      e.IBV_QPT_RAW_PACKET: self.to_rts}             # <<<<<<<<<<<<<<
 *             funcs[self.qp.qp_type](qp_attr)
 * 
*/
    __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(IBV_QPT_RAW_PACKET); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 996, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_to_rts); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 996, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_3, __pyx_t_2, __pyx_t_1) < (0)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_funcs = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyverbs/qp.pyx":997
 *                      e.IBV_QPT_XRC_SEND: self.to_init,
 *                      e.IBV_QPT_RAW_PACKET: self.to_rts}
 *             funcs[self.qp.qp_type](qp_attr)             # <<<<<<<<<<<<<<
 * 
 *     cdef update_cqs(self, init_attr):
*/
    __pyx_t_1 = NULL;
    __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(__pyx_v_self->qp->qp_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyDict_GetItem(__pyx_v_funcs, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_1);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, ((PyObject *)__pyx_v_qp_attr)};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 997, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyverbs/qp.pyx":991
 *             self.srq = srq
 * 
 *         if qp_attr is not None:             # <<<<<<<<<<<<<<
 *             funcs = {e.IBV_QPT_RC: self.to_init, e.IBV_QPT_UC: self.to_init,
 *                      e.IBV_QPT_UD: self.to_rts,
*/
  }

  /* "pyverbs/qp.pyx":928
 * 
 * cdef class QP(PyverbsCM):
 *     def __init__(self, object creator not None, object init_attr not None,             # <<<<<<<<<<<<<<
 *                  QPAttr qp_attr=None):
 *         """
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.qp.QP.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_pd);
  __Pyx_XDECREF((PyObject *)__pyx_v_ctx);
  __Pyx_XDECREF((PyObject *)__pyx_v_xrcd);
  __Pyx_XDECREF((PyObject *)__pyx_v_srq);
  __Pyx_XDECREF(__pyx_v_funcs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":999
 *             funcs[self.qp.qp_type](qp_attr)
 * 
 *     cdef update_cqs(self, init_attr):             # <<<<<<<<<<<<<<
 *         cdef CQ cq
 *         cdef CQEX cqex
*/

static PyObject *__pyx_f_7pyverbs_2qp_2QP_update_cqs(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, PyObject *__pyx_v_init_attr) {
  struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_cq = 0;
  struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_cqex = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("update_cqs", 0);

  /* "pyverbs/qp.pyx":1002
 *         cdef CQ cq
 *         cdef CQEX cqex
 *         if init_attr.send_cq is not None:             # <<<<<<<<<<<<<<
 *             if type(init_attr.send_cq) == CQ:
 *                 cq = <CQ>init_attr.send_cq
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_init_attr, __pyx_mstate_global->__pyx_n_u_send_cq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_t_1 != Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "pyverbs/qp.pyx":1003
 *         cdef CQEX cqex
 *         if init_attr.send_cq is not None:
 *             if type(init_attr.send_cq) == CQ:             # <<<<<<<<<<<<<<
 *                 cq = <CQ>init_attr.send_cq
 *                 cq.add_ref(self)
*/
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_init_attr, __pyx_mstate_global->__pyx_n_u_send_cq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1003, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyObject_RichCompare(((PyObject *)Py_TYPE(__pyx_t_1)), ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQ), Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1003, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1003, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "pyverbs/qp.pyx":1004
 *         if init_attr.send_cq is not None:
 *             if type(init_attr.send_cq) == CQ:
 *                 cq = <CQ>init_attr.send_cq             # <<<<<<<<<<<<<<
 *                 cq.add_ref(self)
 *                 self.scq = cq
*/
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_init_attr, __pyx_mstate_global->__pyx_n_u_send_cq); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1004, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __pyx_t_3;
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_cq = ((struct __pyx_obj_7pyverbs_2cq_CQ *)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "pyverbs/qp.pyx":1005
 *             if type(init_attr.send_cq) == CQ:
 *                 cq = <CQ>init_attr.send_cq
 *                 cq.add_ref(self)             # <<<<<<<<<<<<<<
 *                 self.scq = cq
 *             else:
*/
      __pyx_t_1 = ((struct __pyx_vtabstruct_7pyverbs_2cq_CQ *)__pyx_v_cq->__pyx_base.__pyx_vtab)->add_ref(__pyx_v_cq, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1005, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyverbs/qp.pyx":1006
 *                 cq = <CQ>init_attr.send_cq
 *                 cq.add_ref(self)
 *                 self.scq = cq             # <<<<<<<<<<<<<<
 *             else:
 *                 cqex = <CQEX>init_attr.send_cq
*/
      __Pyx_INCREF((PyObject *)__pyx_v_cq);
      __Pyx_GIVEREF((PyObject *)__pyx_v_cq);
      __Pyx_GOTREF(__pyx_v_self->scq);
      __Pyx_DECREF(__pyx_v_self->scq);
      __pyx_v_self->scq = ((PyObject *)__pyx_v_cq);

      /* "pyverbs/qp.pyx":1003
 *         cdef CQEX cqex
 *         if init_attr.send_cq is not None:
 *             if type(init_attr.send_cq) == CQ:             # <<<<<<<<<<<<<<
 *                 cq = <CQ>init_attr.send_cq
 *                 cq.add_ref(self)
*/
      goto __pyx_L4;
    }

    /* "pyverbs/qp.pyx":1008
 *                 self.scq = cq
 *             else:
 *                 cqex = <CQEX>init_attr.send_cq             # <<<<<<<<<<<<<<
 *                 cqex.add_ref(self)
 *                 self.scq = cqex
*/
    /*else*/ {
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_init_attr, __pyx_mstate_global->__pyx_n_u_send_cq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1008, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1;
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_cqex = ((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "pyverbs/qp.pyx":1009
 *             else:
 *                 cqex = <CQEX>init_attr.send_cq
 *                 cqex.add_ref(self)             # <<<<<<<<<<<<<<
 *                 self.scq = cqex
 *         if init_attr.send_cq != init_attr.recv_cq and init_attr.recv_cq is not None:
*/
      __pyx_t_3 = ((struct __pyx_vtabstruct_7pyverbs_2cq_CQEX *)__pyx_v_cqex->__pyx_base.__pyx_vtab)->add_ref(__pyx_v_cqex, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1009, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "pyverbs/qp.pyx":1010
 *                 cqex = <CQEX>init_attr.send_cq
 *                 cqex.add_ref(self)
 *                 self.scq = cqex             # <<<<<<<<<<<<<<
 *         if init_attr.send_cq != init_attr.recv_cq and init_attr.recv_cq is not None:
 *             if type(init_attr.recv_cq) == CQ:
*/
      __Pyx_INCREF((PyObject *)__pyx_v_cqex);
      __Pyx_GIVEREF((PyObject *)__pyx_v_cqex);
      __Pyx_GOTREF(__pyx_v_self->scq);
      __Pyx_DECREF(__pyx_v_self->scq);
      __pyx_v_self->scq = ((PyObject *)__pyx_v_cqex);
    }
    __pyx_L4:;

    /* "pyverbs/qp.pyx":1002
 *         cdef CQ cq
 *         cdef CQEX cqex
 *         if init_attr.send_cq is not None:             # <<<<<<<<<<<<<<
 *             if type(init_attr.send_cq) == CQ:
 *                 cq = <CQ>init_attr.send_cq
*/
  }

  /* "pyverbs/qp.pyx":1011
 *                 cqex.add_ref(self)
 *                 self.scq = cqex
 *         if init_attr.send_cq != init_attr.recv_cq and init_attr.recv_cq is not None:             # <<<<<<<<<<<<<<
 *             if type(init_attr.recv_cq) == CQ:
 *                 cq = <CQ>init_attr.recv_cq
*/
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_init_attr, __pyx_mstate_global->__pyx_n_u_send_cq); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1011, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_init_attr, __pyx_mstate_global->__pyx_n_u_recv_cq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1011, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_t_1, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1011, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 1011, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_init_attr, __pyx_mstate_global->__pyx_n_u_recv_cq); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1011, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = (__pyx_t_4 != Py_None);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = __pyx_t_5;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_2) {

    /* "pyverbs/qp.pyx":1012
 *                 self.scq = cqex
 *         if init_attr.send_cq != init_attr.recv_cq and init_attr.recv_cq is not None:
 *             if type(init_attr.recv_cq) == CQ:             # <<<<<<<<<<<<<<
 *                 cq = <CQ>init_attr.recv_cq
 *                 cq.add_ref(self)
*/
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_init_attr, __pyx_mstate_global->__pyx_n_u_recv_cq); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1012, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyObject_RichCompare(((PyObject *)Py_TYPE(__pyx_t_4)), ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQ), Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1012, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1012, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_2) {

      /* "pyverbs/qp.pyx":1013
 *         if init_attr.send_cq != init_attr.recv_cq and init_attr.recv_cq is not None:
 *             if type(init_attr.recv_cq) == CQ:
 *                 cq = <CQ>init_attr.recv_cq             # <<<<<<<<<<<<<<
 *                 cq.add_ref(self)
 *                 self.rcq = cq
*/
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_init_attr, __pyx_mstate_global->__pyx_n_u_recv_cq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1013, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1;
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_cq, ((struct __pyx_obj_7pyverbs_2cq_CQ *)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "pyverbs/qp.pyx":1014
 *             if type(init_attr.recv_cq) == CQ:
 *                 cq = <CQ>init_attr.recv_cq
 *                 cq.add_ref(self)             # <<<<<<<<<<<<<<
 *                 self.rcq = cq
 *             else:
*/
      __pyx_t_4 = ((struct __pyx_vtabstruct_7pyverbs_2cq_CQ *)__pyx_v_cq->__pyx_base.__pyx_vtab)->add_ref(__pyx_v_cq, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1014, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "pyverbs/qp.pyx":1015
 *                 cq = <CQ>init_attr.recv_cq
 *                 cq.add_ref(self)
 *                 self.rcq = cq             # <<<<<<<<<<<<<<
 *             else:
 *                 cqex = <CQEX>init_attr.recv_cq
*/
      __Pyx_INCREF((PyObject *)__pyx_v_cq);
      __Pyx_GIVEREF((PyObject *)__pyx_v_cq);
      __Pyx_GOTREF(__pyx_v_self->rcq);
      __Pyx_DECREF(__pyx_v_self->rcq);
      __pyx_v_self->rcq = ((PyObject *)__pyx_v_cq);

      /* "pyverbs/qp.pyx":1012
 *                 self.scq = cqex
 *         if init_attr.send_cq != init_attr.recv_cq and init_attr.recv_cq is not None:
 *             if type(init_attr.recv_cq) == CQ:             # <<<<<<<<<<<<<<
 *                 cq = <CQ>init_attr.recv_cq
 *                 cq.add_ref(self)
*/
      goto __pyx_L8;
    }

    /* "pyverbs/qp.pyx":1017
 *                 self.rcq = cq
 *             else:
 *                 cqex = <CQEX>init_attr.recv_cq             # <<<<<<<<<<<<<<
 *                 cqex.add_ref(self)
 *                 self.rcq = cqex
*/
    /*else*/ {
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_init_attr, __pyx_mstate_global->__pyx_n_u_recv_cq); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1017, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = __pyx_t_4;
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF_SET(__pyx_v_cqex, ((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "pyverbs/qp.pyx":1018
 *             else:
 *                 cqex = <CQEX>init_attr.recv_cq
 *                 cqex.add_ref(self)             # <<<<<<<<<<<<<<
 *                 self.rcq = cqex
 * 
*/
      __pyx_t_1 = ((struct __pyx_vtabstruct_7pyverbs_2cq_CQEX *)__pyx_v_cqex->__pyx_base.__pyx_vtab)->add_ref(__pyx_v_cqex, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1018, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyverbs/qp.pyx":1019
 *                 cqex = <CQEX>init_attr.recv_cq
 *                 cqex.add_ref(self)
 *                 self.rcq = cqex             # <<<<<<<<<<<<<<
 * 
 *     def _create_qp(self, PD pd, QPInitAttr attr):
*/
      __Pyx_INCREF((PyObject *)__pyx_v_cqex);
      __Pyx_GIVEREF((PyObject *)__pyx_v_cqex);
      __Pyx_GOTREF(__pyx_v_self->rcq);
      __Pyx_DECREF(__pyx_v_self->rcq);
      __pyx_v_self->rcq = ((PyObject *)__pyx_v_cqex);
    }
    __pyx_L8:;

    /* "pyverbs/qp.pyx":1011
 *                 cqex.add_ref(self)
 *                 self.scq = cqex
 *         if init_attr.send_cq != init_attr.recv_cq and init_attr.recv_cq is not None:             # <<<<<<<<<<<<<<
 *             if type(init_attr.recv_cq) == CQ:
 *                 cq = <CQ>init_attr.recv_cq
*/
  }

  /* "pyverbs/qp.pyx":999
 *             funcs[self.qp.qp_type](qp_attr)
 * 
 *     cdef update_cqs(self, init_attr):             # <<<<<<<<<<<<<<
 *         cdef CQ cq
 *         cdef CQEX cqex
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.qp.QP.update_cqs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_cq);
  __Pyx_XDECREF((PyObject *)__pyx_v_cqex);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1021
 *                 self.rcq = cqex
 * 
 *     def _create_qp(self, PD pd, QPInitAttr attr):             # <<<<<<<<<<<<<<
 *         self.qp = v.ibv_create_qp(pd.pd, &attr.attr)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_3_create_qp(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_2QP_3_create_qp = {"_create_qp", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_3_create_qp, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_3_create_qp(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_2pd_PD *__pyx_v_pd = 0;
  struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_attr = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_create_qp (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pd,&__pyx_mstate_global->__pyx_n_u_attr,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1021, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1021, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1021, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_create_qp", 0) < (0)) __PYX_ERR(0, 1021, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_create_qp", 1, 2, 2, i); __PYX_ERR(0, 1021, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1021, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1021, __pyx_L3_error)
    }
    __pyx_v_pd = ((struct __pyx_obj_7pyverbs_2pd_PD *)values[0]);
    __pyx_v_attr = ((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_create_qp", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1021, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QP._create_qp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_pd), __pyx_mstate_global->__pyx_ptype_7pyverbs_2pd_PD, 1, "pd", 0))) __PYX_ERR(0, 1021, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_attr), __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPInitAttr, 1, "attr", 0))) __PYX_ERR(0, 1021, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp_2QP_2_create_qp(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self), __pyx_v_pd, __pyx_v_attr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_2QP_2_create_qp(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, struct __pyx_obj_7pyverbs_2pd_PD *__pyx_v_pd, struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_attr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_create_qp", 0);

  /* "pyverbs/qp.pyx":1022
 * 
 *     def _create_qp(self, PD pd, QPInitAttr attr):
 *         self.qp = v.ibv_create_qp(pd.pd, &attr.attr)             # <<<<<<<<<<<<<<
 * 
 *     def _create_qp_ex(self, Context ctx, QPInitAttrEx attr):
*/
  __pyx_v_self->qp = ibv_create_qp(__pyx_v_pd->pd, (&__pyx_v_attr->attr));

  /* "pyverbs/qp.pyx":1021
 *                 self.rcq = cqex
 * 
 *     def _create_qp(self, PD pd, QPInitAttr attr):             # <<<<<<<<<<<<<<
 *         self.qp = v.ibv_create_qp(pd.pd, &attr.attr)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1024
 *         self.qp = v.ibv_create_qp(pd.pd, &attr.attr)
 * 
 *     def _create_qp_ex(self, Context ctx, QPInitAttrEx attr):             # <<<<<<<<<<<<<<
 *         self.qp = v.ibv_create_qp_ex(ctx.context, &attr.attr)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_5_create_qp_ex(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_2QP_5_create_qp_ex = {"_create_qp_ex", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_5_create_qp_ex, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_5_create_qp_ex(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_6device_Context *__pyx_v_ctx = 0;
  struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_attr = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_create_qp_ex (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_ctx,&__pyx_mstate_global->__pyx_n_u_attr,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1024, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1024, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1024, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_create_qp_ex", 0) < (0)) __PYX_ERR(0, 1024, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_create_qp_ex", 1, 2, 2, i); __PYX_ERR(0, 1024, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1024, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1024, __pyx_L3_error)
    }
    __pyx_v_ctx = ((struct __pyx_obj_7pyverbs_6device_Context *)values[0]);
    __pyx_v_attr = ((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_create_qp_ex", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1024, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QP._create_qp_ex", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_ctx), __pyx_mstate_global->__pyx_ptype_7pyverbs_6device_Context, 1, "ctx", 0))) __PYX_ERR(0, 1024, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_attr), __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx, 1, "attr", 0))) __PYX_ERR(0, 1024, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp_2QP_4_create_qp_ex(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self), __pyx_v_ctx, __pyx_v_attr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_2QP_4_create_qp_ex(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, struct __pyx_obj_7pyverbs_6device_Context *__pyx_v_ctx, struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_attr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_create_qp_ex", 0);

  /* "pyverbs/qp.pyx":1025
 * 
 *     def _create_qp_ex(self, Context ctx, QPInitAttrEx attr):
 *         self.qp = v.ibv_create_qp_ex(ctx.context, &attr.attr)             # <<<<<<<<<<<<<<
 * 
 *     cdef add_ref(self, obj):
*/
  __pyx_v_self->qp = ibv_create_qp_ex(__pyx_v_ctx->context, (&__pyx_v_attr->attr));

  /* "pyverbs/qp.pyx":1024
 *         self.qp = v.ibv_create_qp(pd.pd, &attr.attr)
 * 
 *     def _create_qp_ex(self, Context ctx, QPInitAttrEx attr):             # <<<<<<<<<<<<<<
 *         self.qp = v.ibv_create_qp_ex(ctx.context, &attr.attr)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1027
 *         self.qp = v.ibv_create_qp_ex(ctx.context, &attr.attr)
 * 
 *     cdef add_ref(self, obj):             # <<<<<<<<<<<<<<
 *         if isinstance(obj, MW):
 *             self.mws.add(obj)
*/

static PyObject *__pyx_f_7pyverbs_2qp_2QP_add_ref(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_ref", 0);

  /* "pyverbs/qp.pyx":1028
 * 
 *     cdef add_ref(self, obj):
 *         if isinstance(obj, MW):             # <<<<<<<<<<<<<<
 *             self.mws.add(obj)
 *         elif isinstance(obj, Flow):
*/
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_mstate_global->__pyx_ptype_7pyverbs_2mr_MW); 
  if (__pyx_t_1) {

    /* "pyverbs/qp.pyx":1029
 *     cdef add_ref(self, obj):
 *         if isinstance(obj, MW):
 *             self.mws.add(obj)             # <<<<<<<<<<<<<<
 *         elif isinstance(obj, Flow):
 *             self.flows.add(obj)
*/
    __pyx_t_3 = __pyx_v_self->mws;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_obj};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_add, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1029, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":1028
 * 
 *     cdef add_ref(self, obj):
 *         if isinstance(obj, MW):             # <<<<<<<<<<<<<<
 *             self.mws.add(obj)
 *         elif isinstance(obj, Flow):
*/
    goto __pyx_L3;
  }

  /* "pyverbs/qp.pyx":1030
 *         if isinstance(obj, MW):
 *             self.mws.add(obj)
 *         elif isinstance(obj, Flow):             # <<<<<<<<<<<<<<
 *             self.flows.add(obj)
 *         else:
*/
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_mstate_global->__pyx_ptype_7pyverbs_4flow_Flow); 
  if (likely(__pyx_t_1)) {

    /* "pyverbs/qp.pyx":1031
 *             self.mws.add(obj)
 *         elif isinstance(obj, Flow):
 *             self.flows.add(obj)             # <<<<<<<<<<<<<<
 *         else:
 *             raise PyverbsError('Unrecognized object type')
*/
    __pyx_t_3 = __pyx_v_self->flows;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_obj};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_add, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1031, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":1030
 *         if isinstance(obj, MW):
 *             self.mws.add(obj)
 *         elif isinstance(obj, Flow):             # <<<<<<<<<<<<<<
 *             self.flows.add(obj)
 *         else:
*/
    goto __pyx_L3;
  }

  /* "pyverbs/qp.pyx":1033
 *             self.flows.add(obj)
 *         else:
 *             raise PyverbsError('Unrecognized object type')             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
*/
  /*else*/ {
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PyverbsError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1033, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_4 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Unrecognized_object_type};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1033, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1033, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "pyverbs/qp.pyx":1027
 *         self.qp = v.ibv_create_qp_ex(ctx.context, &attr.attr)
 * 
 *     cdef add_ref(self, obj):             # <<<<<<<<<<<<<<
 *         if isinstance(obj, MW):
 *             self.mws.add(obj)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyverbs.qp.QP.add_ref", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1035
 *             raise PyverbsError('Unrecognized object type')
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         self.close()
 * 
*/

/* Python wrapper */
static void __pyx_pw_7pyverbs_2qp_2QP_7__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7pyverbs_2qp_2QP_7__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_7pyverbs_2qp_2QP_6__dealloc__(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7pyverbs_2qp_2QP_6__dealloc__(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "pyverbs/qp.pyx":1036
 * 
 *     def __dealloc__(self):
 *         self.close()             # <<<<<<<<<<<<<<
 * 
 *     cpdef close(self):
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_7pyverbs_2qp_QP *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.close(((struct __pyx_obj_7pyverbs_4base_PyverbsCM *)__pyx_v_self), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":1035
 *             raise PyverbsError('Unrecognized object type')
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         self.close()
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("pyverbs.qp.QP.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "pyverbs/qp.pyx":1038
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.qp != NULL:
 *             if self.logger:
*/

static PyObject *__pyx_pw_7pyverbs_2qp_2QP_9close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7pyverbs_2qp_2QP_close(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, int __pyx_skip_dispatch) {
  int __pyx_v_rc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QP &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_close); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1038, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyverbs_2qp_2QP_9close)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1038, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyverbs/qp.pyx":1039
 * 
 *     cpdef close(self):
 *         if self.qp != NULL:             # <<<<<<<<<<<<<<
 *             if self.logger:
 *                 self.logger.debug('Closing QP')
*/
  __pyx_t_6 = (__pyx_v_self->qp != NULL);
  if (__pyx_t_6) {

    /* "pyverbs/qp.pyx":1040
 *     cpdef close(self):
 *         if self.qp != NULL:
 *             if self.logger:             # <<<<<<<<<<<<<<
 *                 self.logger.debug('Closing QP')
 *             close_weakrefs([self.mws, self.flows, self.dr_actions])
*/
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_self->__pyx_base.__pyx_base.logger); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 1040, __pyx_L1_error)
    if (__pyx_t_6) {

      /* "pyverbs/qp.pyx":1041
 *         if self.qp != NULL:
 *             if self.logger:
 *                 self.logger.debug('Closing QP')             # <<<<<<<<<<<<<<
 *             close_weakrefs([self.mws, self.flows, self.dr_actions])
 *             rc = v.ibv_destroy_qp(self.qp)
*/
      __pyx_t_2 = __pyx_v_self->__pyx_base.__pyx_base.logger;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_5 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Closing_QP};
        __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_debug, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1041, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyverbs/qp.pyx":1040
 *     cpdef close(self):
 *         if self.qp != NULL:
 *             if self.logger:             # <<<<<<<<<<<<<<
 *                 self.logger.debug('Closing QP')
 *             close_weakrefs([self.mws, self.flows, self.dr_actions])
*/
    }

    /* "pyverbs/qp.pyx":1042
 *             if self.logger:
 *                 self.logger.debug('Closing QP')
 *             close_weakrefs([self.mws, self.flows, self.dr_actions])             # <<<<<<<<<<<<<<
 *             rc = v.ibv_destroy_qp(self.qp)
 *             if rc:
*/
    __pyx_t_1 = PyList_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1042, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_self->mws);
    __Pyx_GIVEREF(__pyx_v_self->mws);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->mws) != (0)) __PYX_ERR(0, 1042, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_self->flows);
    __Pyx_GIVEREF(__pyx_v_self->flows);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->flows) != (0)) __PYX_ERR(0, 1042, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_self->dr_actions);
    __Pyx_GIVEREF(__pyx_v_self->dr_actions);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->dr_actions) != (0)) __PYX_ERR(0, 1042, __pyx_L1_error);
    __pyx_t_2 = __pyx_f_7pyverbs_4base_close_weakrefs(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1042, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":1043
 *                 self.logger.debug('Closing QP')
 *             close_weakrefs([self.mws, self.flows, self.dr_actions])
 *             rc = v.ibv_destroy_qp(self.qp)             # <<<<<<<<<<<<<<
 *             if rc:
 *                 raise PyverbsRDMAError('Failed to destroy QP', rc)
*/
    __pyx_v_rc = ibv_destroy_qp(__pyx_v_self->qp);

    /* "pyverbs/qp.pyx":1044
 *             close_weakrefs([self.mws, self.flows, self.dr_actions])
 *             rc = v.ibv_destroy_qp(self.qp)
 *             if rc:             # <<<<<<<<<<<<<<
 *                 raise PyverbsRDMAError('Failed to destroy QP', rc)
 *             self.qp = NULL
*/
    __pyx_t_6 = (__pyx_v_rc != 0);
    if (unlikely(__pyx_t_6)) {

      /* "pyverbs/qp.pyx":1045
 *             rc = v.ibv_destroy_qp(self.qp)
 *             if rc:
 *                 raise PyverbsRDMAError('Failed to destroy QP', rc)             # <<<<<<<<<<<<<<
 *             self.qp = NULL
 *             self.pd = None
*/
      __pyx_t_1 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1045, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyLong_From_int(__pyx_v_rc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1045, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
        assert(__pyx_t_1);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
        __pyx_t_5 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_mstate_global->__pyx_kp_u_Failed_to_destroy_QP, __pyx_t_3};
        __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1045, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 1045, __pyx_L1_error)

      /* "pyverbs/qp.pyx":1044
 *             close_weakrefs([self.mws, self.flows, self.dr_actions])
 *             rc = v.ibv_destroy_qp(self.qp)
 *             if rc:             # <<<<<<<<<<<<<<
 *                 raise PyverbsRDMAError('Failed to destroy QP', rc)
 *             self.qp = NULL
*/
    }

    /* "pyverbs/qp.pyx":1046
 *             if rc:
 *                 raise PyverbsRDMAError('Failed to destroy QP', rc)
 *             self.qp = NULL             # <<<<<<<<<<<<<<
 *             self.pd = None
 *             self.context = None
*/
    __pyx_v_self->qp = NULL;

    /* "pyverbs/qp.pyx":1047
 *                 raise PyverbsRDMAError('Failed to destroy QP', rc)
 *             self.qp = NULL
 *             self.pd = None             # <<<<<<<<<<<<<<
 *             self.context = None
 *             self.scq = None
*/
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->pd);
    __Pyx_DECREF(__pyx_v_self->pd);
    __pyx_v_self->pd = Py_None;

    /* "pyverbs/qp.pyx":1048
 *             self.qp = NULL
 *             self.pd = None
 *             self.context = None             # <<<<<<<<<<<<<<
 *             self.scq = None
 *             self.rcq = None
*/
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->context);
    __Pyx_DECREF(__pyx_v_self->context);
    __pyx_v_self->context = Py_None;

    /* "pyverbs/qp.pyx":1049
 *             self.pd = None
 *             self.context = None
 *             self.scq = None             # <<<<<<<<<<<<<<
 *             self.rcq = None
 * 
*/
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->scq);
    __Pyx_DECREF(__pyx_v_self->scq);
    __pyx_v_self->scq = Py_None;

    /* "pyverbs/qp.pyx":1050
 *             self.context = None
 *             self.scq = None
 *             self.rcq = None             # <<<<<<<<<<<<<<
 * 
 *     def _get_comp_mask(self, dst):
*/
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->rcq);
    __Pyx_DECREF(__pyx_v_self->rcq);
    __pyx_v_self->rcq = Py_None;

    /* "pyverbs/qp.pyx":1039
 * 
 *     cpdef close(self):
 *         if self.qp != NULL:             # <<<<<<<<<<<<<<
 *             if self.logger:
 *                 self.logger.debug('Closing QP')
*/
  }

  /* "pyverbs/qp.pyx":1038
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.qp != NULL:
 *             if self.logger:
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.qp.QP.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_9close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_2QP_9close = {"close", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_9close, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_9close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("close", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("close", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2qp_2QP_8close(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_2QP_8close(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7pyverbs_2qp_2QP_close(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QP.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1052
 *             self.rcq = None
 * 
 *     def _get_comp_mask(self, dst):             # <<<<<<<<<<<<<<
 *         masks = {e.IBV_QPT_RC: {'INIT': e.IBV_QP_PKEY_INDEX | e.IBV_QP_PORT |\
 *                                 e.IBV_QP_ACCESS_FLAGS, 'RTR': e.IBV_QP_AV |\
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_11_get_comp_mask(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_2QP_11_get_comp_mask = {"_get_comp_mask", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_11_get_comp_mask, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_11_get_comp_mask(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_dst = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_comp_mask (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_dst,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1052, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1052, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_get_comp_mask", 0) < (0)) __PYX_ERR(0, 1052, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_get_comp_mask", 1, 1, 1, i); __PYX_ERR(0, 1052, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1052, __pyx_L3_error)
    }
    __pyx_v_dst = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_comp_mask", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1052, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QP._get_comp_mask", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_2QP_10_get_comp_mask(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self), __pyx_v_dst);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_2QP_10_get_comp_mask(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, PyObject *__pyx_v_dst) {
  PyObject *__pyx_v_masks = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_comp_mask", 0);

  /* "pyverbs/qp.pyx":1053
 * 
 *     def _get_comp_mask(self, dst):
 *         masks = {e.IBV_QPT_RC: {'INIT': e.IBV_QP_PKEY_INDEX | e.IBV_QP_PORT |\             # <<<<<<<<<<<<<<
 *                                 e.IBV_QP_ACCESS_FLAGS, 'RTR': e.IBV_QP_AV |\
 *                                 e.IBV_QP_PATH_MTU | e.IBV_QP_DEST_QPN |\
*/
  __pyx_t_1 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(IBV_QPT_RC); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "pyverbs/qp.pyx":1054
 *     def _get_comp_mask(self, dst):
 *         masks = {e.IBV_QPT_RC: {'INIT': e.IBV_QP_PKEY_INDEX | e.IBV_QP_PORT |\
 *                                 e.IBV_QP_ACCESS_FLAGS, 'RTR': e.IBV_QP_AV |\             # <<<<<<<<<<<<<<
 *                                 e.IBV_QP_PATH_MTU | e.IBV_QP_DEST_QPN |\
 *                                 e.IBV_QP_RQ_PSN |\
*/
  __pyx_t_4 = __Pyx_PyLong_From_int(((IBV_QP_PKEY_INDEX | IBV_QP_PORT) | IBV_QP_ACCESS_FLAGS)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_INIT, __pyx_t_4) < (0)) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":1057
 *                                 e.IBV_QP_PATH_MTU | e.IBV_QP_DEST_QPN |\
 *                                 e.IBV_QP_RQ_PSN |\
 *                                 e.IBV_QP_MAX_DEST_RD_ATOMIC |\             # <<<<<<<<<<<<<<
 *                                 e.IBV_QP_MIN_RNR_TIMER,
 *                                 'RTS': e.IBV_QP_TIMEOUT |\
*/
  __pyx_t_4 = __Pyx_PyLong_From_int((((((IBV_QP_AV | IBV_QP_PATH_MTU) | IBV_QP_DEST_QPN) | IBV_QP_RQ_PSN) | IBV_QP_MAX_DEST_RD_ATOMIC) | IBV_QP_MIN_RNR_TIMER)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1057, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_RTR, __pyx_t_4) < (0)) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":1061
 *                                 'RTS': e.IBV_QP_TIMEOUT |\
 *                                 e.IBV_QP_RETRY_CNT | e.IBV_QP_RNR_RETRY |\
 *                                 e.IBV_QP_SQ_PSN | e.IBV_QP_MAX_QP_RD_ATOMIC},             # <<<<<<<<<<<<<<
 *                 e.IBV_QPT_UC: {'INIT': e.IBV_QP_PKEY_INDEX | e.IBV_QP_PORT |\
 *                                e.IBV_QP_ACCESS_FLAGS, 'RTR': e.IBV_QP_AV |\
*/
  __pyx_t_4 = __Pyx_PyLong_From_int(((((IBV_QP_TIMEOUT | IBV_QP_RETRY_CNT) | IBV_QP_RNR_RETRY) | IBV_QP_SQ_PSN) | IBV_QP_MAX_QP_RD_ATOMIC)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_RTS, __pyx_t_4) < (0)) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_t_3) < (0)) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":1062
 *                                 e.IBV_QP_RETRY_CNT | e.IBV_QP_RNR_RETRY |\
 *                                 e.IBV_QP_SQ_PSN | e.IBV_QP_MAX_QP_RD_ATOMIC},
 *                 e.IBV_QPT_UC: {'INIT': e.IBV_QP_PKEY_INDEX | e.IBV_QP_PORT |\             # <<<<<<<<<<<<<<
 *                                e.IBV_QP_ACCESS_FLAGS, 'RTR': e.IBV_QP_AV |\
 *                                e.IBV_QP_PATH_MTU | e.IBV_QP_DEST_QPN |\
*/
  __pyx_t_3 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(IBV_QPT_UC); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "pyverbs/qp.pyx":1063
 *                                 e.IBV_QP_SQ_PSN | e.IBV_QP_MAX_QP_RD_ATOMIC},
 *                 e.IBV_QPT_UC: {'INIT': e.IBV_QP_PKEY_INDEX | e.IBV_QP_PORT |\
 *                                e.IBV_QP_ACCESS_FLAGS, 'RTR': e.IBV_QP_AV |\             # <<<<<<<<<<<<<<
 *                                e.IBV_QP_PATH_MTU | e.IBV_QP_DEST_QPN |\
 *                                e.IBV_QP_RQ_PSN, 'RTS': e.IBV_QP_SQ_PSN},
*/
  __pyx_t_4 = __Pyx_PyLong_From_int(((IBV_QP_PKEY_INDEX | IBV_QP_PORT) | IBV_QP_ACCESS_FLAGS)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_INIT, __pyx_t_4) < (0)) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":1064
 *                 e.IBV_QPT_UC: {'INIT': e.IBV_QP_PKEY_INDEX | e.IBV_QP_PORT |\
 *                                e.IBV_QP_ACCESS_FLAGS, 'RTR': e.IBV_QP_AV |\
 *                                e.IBV_QP_PATH_MTU | e.IBV_QP_DEST_QPN |\             # <<<<<<<<<<<<<<
 *                                e.IBV_QP_RQ_PSN, 'RTS': e.IBV_QP_SQ_PSN},
 *                 e.IBV_QPT_UD: {'INIT': e.IBV_QP_PKEY_INDEX | e.IBV_QP_PORT |\
*/
  __pyx_t_4 = __Pyx_PyLong_From_int((((IBV_QP_AV | IBV_QP_PATH_MTU) | IBV_QP_DEST_QPN) | IBV_QP_RQ_PSN)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1064, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_RTR, __pyx_t_4) < (0)) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":1065
 *                                e.IBV_QP_ACCESS_FLAGS, 'RTR': e.IBV_QP_AV |\
 *                                e.IBV_QP_PATH_MTU | e.IBV_QP_DEST_QPN |\
 *                                e.IBV_QP_RQ_PSN, 'RTS': e.IBV_QP_SQ_PSN},             # <<<<<<<<<<<<<<
 *                 e.IBV_QPT_UD: {'INIT': e.IBV_QP_PKEY_INDEX | e.IBV_QP_PORT |\
 *                                e.IBV_QP_QKEY, 'RTR': 0,
*/
  __pyx_t_4 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py(IBV_QP_SQ_PSN); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1065, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_RTS, __pyx_t_4) < (0)) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_3, __pyx_t_2) < (0)) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1066
 *                                e.IBV_QP_PATH_MTU | e.IBV_QP_DEST_QPN |\
 *                                e.IBV_QP_RQ_PSN, 'RTS': e.IBV_QP_SQ_PSN},
 *                 e.IBV_QPT_UD: {'INIT': e.IBV_QP_PKEY_INDEX | e.IBV_QP_PORT |\             # <<<<<<<<<<<<<<
 *                                e.IBV_QP_QKEY, 'RTR': 0,
 *                                'RTS': e.IBV_QP_SQ_PSN},
*/
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(IBV_QPT_UD); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "pyverbs/qp.pyx":1067
 *                                e.IBV_QP_RQ_PSN, 'RTS': e.IBV_QP_SQ_PSN},
 *                 e.IBV_QPT_UD: {'INIT': e.IBV_QP_PKEY_INDEX | e.IBV_QP_PORT |\
 *                                e.IBV_QP_QKEY, 'RTR': 0,             # <<<<<<<<<<<<<<
 *                                'RTS': e.IBV_QP_SQ_PSN},
 *                 e.IBV_QPT_RAW_PACKET: {'INIT': e.IBV_QP_PORT, 'RTR': 0,
*/
  __pyx_t_4 = __Pyx_PyLong_From_int(((IBV_QP_PKEY_INDEX | IBV_QP_PORT) | IBV_QP_QKEY)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_INIT, __pyx_t_4) < (0)) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_RTR, __pyx_mstate_global->__pyx_int_0) < (0)) __PYX_ERR(0, 1066, __pyx_L1_error)

  /* "pyverbs/qp.pyx":1068
 *                 e.IBV_QPT_UD: {'INIT': e.IBV_QP_PKEY_INDEX | e.IBV_QP_PORT |\
 *                                e.IBV_QP_QKEY, 'RTR': 0,
 *                                'RTS': e.IBV_QP_SQ_PSN},             # <<<<<<<<<<<<<<
 *                 e.IBV_QPT_RAW_PACKET: {'INIT': e.IBV_QP_PORT, 'RTR': 0,
 *                                        'RTS': 0},
*/
  __pyx_t_4 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py(IBV_QP_SQ_PSN); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_RTS, __pyx_t_4) < (0)) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_t_3) < (0)) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":1069
 *                                e.IBV_QP_QKEY, 'RTR': 0,
 *                                'RTS': e.IBV_QP_SQ_PSN},
 *                 e.IBV_QPT_RAW_PACKET: {'INIT': e.IBV_QP_PORT, 'RTR': 0,             # <<<<<<<<<<<<<<
 *                                        'RTS': 0},
 *                 e.IBV_QPT_XRC_RECV: {'INIT': e.IBV_QP_PKEY_INDEX |\
*/
  __pyx_t_3 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(IBV_QPT_RAW_PACKET); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1069, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1069, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py(IBV_QP_PORT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1069, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_INIT, __pyx_t_4) < (0)) __PYX_ERR(0, 1069, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_RTR, __pyx_mstate_global->__pyx_int_0) < (0)) __PYX_ERR(0, 1069, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_RTS, __pyx_mstate_global->__pyx_int_0) < (0)) __PYX_ERR(0, 1069, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_3, __pyx_t_2) < (0)) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1071
 *                 e.IBV_QPT_RAW_PACKET: {'INIT': e.IBV_QP_PORT, 'RTR': 0,
 *                                        'RTS': 0},
 *                 e.IBV_QPT_XRC_RECV: {'INIT': e.IBV_QP_PKEY_INDEX |\             # <<<<<<<<<<<<<<
 *                                 e.IBV_QP_PORT | e.IBV_QP_ACCESS_FLAGS,
 *                                 'RTR': e.IBV_QP_AV | e.IBV_QP_PATH_MTU |\
*/
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(IBV_QPT_XRC_RECV); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "pyverbs/qp.pyx":1072
 *                                        'RTS': 0},
 *                 e.IBV_QPT_XRC_RECV: {'INIT': e.IBV_QP_PKEY_INDEX |\
 *                                 e.IBV_QP_PORT | e.IBV_QP_ACCESS_FLAGS,             # <<<<<<<<<<<<<<
 *                                 'RTR': e.IBV_QP_AV | e.IBV_QP_PATH_MTU |\
 *                                 e.IBV_QP_DEST_QPN | e.IBV_QP_RQ_PSN |   \
*/
  __pyx_t_4 = __Pyx_PyLong_From_int(((IBV_QP_PKEY_INDEX | IBV_QP_PORT) | IBV_QP_ACCESS_FLAGS)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_INIT, __pyx_t_4) < (0)) __PYX_ERR(0, 1071, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":1075
 *                                 'RTR': e.IBV_QP_AV | e.IBV_QP_PATH_MTU |\
 *                                 e.IBV_QP_DEST_QPN | e.IBV_QP_RQ_PSN |   \
 *                                 e.IBV_QP_MAX_DEST_RD_ATOMIC |\             # <<<<<<<<<<<<<<
 *                                 e.IBV_QP_MIN_RNR_TIMER,
 *                                 'RTS': e.IBV_QP_TIMEOUT | e.IBV_QP_SQ_PSN },
*/
  __pyx_t_4 = __Pyx_PyLong_From_int((((((IBV_QP_AV | IBV_QP_PATH_MTU) | IBV_QP_DEST_QPN) | IBV_QP_RQ_PSN) | IBV_QP_MAX_DEST_RD_ATOMIC) | IBV_QP_MIN_RNR_TIMER)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_RTR, __pyx_t_4) < (0)) __PYX_ERR(0, 1071, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":1077
 *                                 e.IBV_QP_MAX_DEST_RD_ATOMIC |\
 *                                 e.IBV_QP_MIN_RNR_TIMER,
 *                                 'RTS': e.IBV_QP_TIMEOUT | e.IBV_QP_SQ_PSN },             # <<<<<<<<<<<<<<
 *                 e.IBV_QPT_XRC_SEND: {'INIT': e.IBV_QP_PKEY_INDEX |\
 *                                 e.IBV_QP_PORT | e.IBV_QP_ACCESS_FLAGS,
*/
  __pyx_t_4 = __Pyx_PyLong_From_int((IBV_QP_TIMEOUT | IBV_QP_SQ_PSN)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1077, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_RTS, __pyx_t_4) < (0)) __PYX_ERR(0, 1071, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_t_3) < (0)) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":1078
 *                                 e.IBV_QP_MIN_RNR_TIMER,
 *                                 'RTS': e.IBV_QP_TIMEOUT | e.IBV_QP_SQ_PSN },
 *                 e.IBV_QPT_XRC_SEND: {'INIT': e.IBV_QP_PKEY_INDEX |\             # <<<<<<<<<<<<<<
 *                                 e.IBV_QP_PORT | e.IBV_QP_ACCESS_FLAGS,
 *                                 'RTR': e.IBV_QP_AV | e.IBV_QP_PATH_MTU |\
*/
  __pyx_t_3 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(IBV_QPT_XRC_SEND); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "pyverbs/qp.pyx":1079
 *                                 'RTS': e.IBV_QP_TIMEOUT | e.IBV_QP_SQ_PSN },
 *                 e.IBV_QPT_XRC_SEND: {'INIT': e.IBV_QP_PKEY_INDEX |\
 *                                 e.IBV_QP_PORT | e.IBV_QP_ACCESS_FLAGS,             # <<<<<<<<<<<<<<
 *                                 'RTR': e.IBV_QP_AV | e.IBV_QP_PATH_MTU |\
 *                                 e.IBV_QP_DEST_QPN | e.IBV_QP_RQ_PSN,
*/
  __pyx_t_4 = __Pyx_PyLong_From_int(((IBV_QP_PKEY_INDEX | IBV_QP_PORT) | IBV_QP_ACCESS_FLAGS)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1079, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_INIT, __pyx_t_4) < (0)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":1081
 *                                 e.IBV_QP_PORT | e.IBV_QP_ACCESS_FLAGS,
 *                                 'RTR': e.IBV_QP_AV | e.IBV_QP_PATH_MTU |\
 *                                 e.IBV_QP_DEST_QPN | e.IBV_QP_RQ_PSN,             # <<<<<<<<<<<<<<
 *                                 'RTS': e.IBV_QP_TIMEOUT |\
 *                                 e.IBV_QP_RETRY_CNT | e.IBV_QP_RNR_RETRY |\
*/
  __pyx_t_4 = __Pyx_PyLong_From_int((((IBV_QP_AV | IBV_QP_PATH_MTU) | IBV_QP_DEST_QPN) | IBV_QP_RQ_PSN)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_RTR, __pyx_t_4) < (0)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":1084
 *                                 'RTS': e.IBV_QP_TIMEOUT |\
 *                                 e.IBV_QP_RETRY_CNT | e.IBV_QP_RNR_RETRY |\
 *                                 e.IBV_QP_SQ_PSN | e.IBV_QP_MAX_QP_RD_ATOMIC}}             # <<<<<<<<<<<<<<
 * 
 *         return masks[self.qp.qp_type][dst] | e.IBV_QP_STATE
*/
  __pyx_t_4 = __Pyx_PyLong_From_int(((((IBV_QP_TIMEOUT | IBV_QP_RETRY_CNT) | IBV_QP_RNR_RETRY) | IBV_QP_SQ_PSN) | IBV_QP_MAX_QP_RD_ATOMIC)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_RTS, __pyx_t_4) < (0)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_3, __pyx_t_2) < (0)) __PYX_ERR(0, 1053, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_masks = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":1086
 *                                 e.IBV_QP_SQ_PSN | e.IBV_QP_MAX_QP_RD_ATOMIC}}
 * 
 *         return masks[self.qp.qp_type][dst] | e.IBV_QP_STATE             # <<<<<<<<<<<<<<
 * 
 *     def to_init(self, QPAttr qp_attr):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(__pyx_v_self->qp->qp_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_masks, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_v_dst); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py(IBV_QP_STATE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Or(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":1052
 *             self.rcq = None
 * 
 *     def _get_comp_mask(self, dst):             # <<<<<<<<<<<<<<
 *         masks = {e.IBV_QPT_RC: {'INIT': e.IBV_QP_PKEY_INDEX | e.IBV_QP_PORT |\
 *                                 e.IBV_QP_ACCESS_FLAGS, 'RTR': e.IBV_QP_AV |\
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.qp.QP._get_comp_mask", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_masks);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1088
 *         return masks[self.qp.qp_type][dst] | e.IBV_QP_STATE
 * 
 *     def to_init(self, QPAttr qp_attr):             # <<<<<<<<<<<<<<
 *         """
 *         Modify the current QP's state to INIT. If the current state doesn't
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_13to_init(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2qp_2QP_12to_init, "\n        Modify the current QP's state to INIT. If the current state doesn't\n        support transition to INIT, an exception will be raised.\n        The comp mask provided to the kernel includes the needed bits for 2INIT\n        transition for this QP type.\n        :param qp_attr: QPAttr object containing the needed attributes for\n                        2INIT transition\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_2qp_2QP_13to_init = {"to_init", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_13to_init, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2qp_2QP_12to_init};
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_13to_init(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_qp_attr = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("to_init (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_qp_attr,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1088, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1088, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "to_init", 0) < (0)) __PYX_ERR(0, 1088, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("to_init", 1, 1, 1, i); __PYX_ERR(0, 1088, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1088, __pyx_L3_error)
    }
    __pyx_v_qp_attr = ((struct __pyx_obj_7pyverbs_2qp_QPAttr *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("to_init", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1088, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QP.to_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_qp_attr), __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPAttr, 1, "qp_attr", 0))) __PYX_ERR(0, 1088, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp_2QP_12to_init(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self), __pyx_v_qp_attr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_2QP_12to_init(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_qp_attr) {
  PyObject *__pyx_v_mask = NULL;
  int __pyx_v_rc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("to_init", 0);

  /* "pyverbs/qp.pyx":1098
 *         :return: None
 *         """
 *         mask = self._get_comp_mask('INIT')             # <<<<<<<<<<<<<<
 *         qp_attr.qp_state = e.IBV_QPS_INIT
 *         rc =  v.ibv_modify_qp(self.qp, &qp_attr.attr, mask)
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_n_u_INIT};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_comp_mask, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1098, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_mask = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":1099
 *         """
 *         mask = self._get_comp_mask('INIT')
 *         qp_attr.qp_state = e.IBV_QPS_INIT             # <<<<<<<<<<<<<<
 *         rc =  v.ibv_modify_qp(self.qp, &qp_attr.attr, mask)
 *         if rc != 0:
*/
  __pyx_t_1 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py(IBV_QPS_INIT); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1099, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_qp_attr), __pyx_mstate_global->__pyx_n_u_qp_state, __pyx_t_1) < (0)) __PYX_ERR(0, 1099, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":1100
 *         mask = self._get_comp_mask('INIT')
 *         qp_attr.qp_state = e.IBV_QPS_INIT
 *         rc =  v.ibv_modify_qp(self.qp, &qp_attr.attr, mask)             # <<<<<<<<<<<<<<
 *         if rc != 0:
 *             raise PyverbsRDMAError('Failed to modify QP state to init', rc)
*/
  __pyx_t_4 = __Pyx_PyLong_As_int(__pyx_v_mask); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1100, __pyx_L1_error)
  __pyx_v_rc = ibv_modify_qp(__pyx_v_self->qp, (&__pyx_v_qp_attr->attr), __pyx_t_4);

  /* "pyverbs/qp.pyx":1101
 *         qp_attr.qp_state = e.IBV_QPS_INIT
 *         rc =  v.ibv_modify_qp(self.qp, &qp_attr.attr, mask)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Failed to modify QP state to init', rc)
 * 
*/
  __pyx_t_5 = (__pyx_v_rc != 0);
  if (unlikely(__pyx_t_5)) {

    /* "pyverbs/qp.pyx":1102
 *         rc =  v.ibv_modify_qp(self.qp, &qp_attr.attr, mask)
 *         if rc != 0:
 *             raise PyverbsRDMAError('Failed to modify QP state to init', rc)             # <<<<<<<<<<<<<<
 * 
 *     def to_rtr(self, QPAttr qp_attr):
*/
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyLong_From_int(__pyx_v_rc); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
      __pyx_t_3 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Failed_to_modify_QP_state_to_ini, __pyx_t_7};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_3, (3-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1102, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1102, __pyx_L1_error)

    /* "pyverbs/qp.pyx":1101
 *         qp_attr.qp_state = e.IBV_QPS_INIT
 *         rc =  v.ibv_modify_qp(self.qp, &qp_attr.attr, mask)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Failed to modify QP state to init', rc)
 * 
*/
  }

  /* "pyverbs/qp.pyx":1088
 *         return masks[self.qp.qp_type][dst] | e.IBV_QP_STATE
 * 
 *     def to_init(self, QPAttr qp_attr):             # <<<<<<<<<<<<<<
 *         """
 *         Modify the current QP's state to INIT. If the current state doesn't
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyverbs.qp.QP.to_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_mask);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1104
 *             raise PyverbsRDMAError('Failed to modify QP state to init', rc)
 * 
 *     def to_rtr(self, QPAttr qp_attr):             # <<<<<<<<<<<<<<
 *         """
 *         Modify the current QP's state to RTR. It assumes that its current
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_15to_rtr(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2qp_2QP_14to_rtr, "\n        Modify the current QP's state to RTR. It assumes that its current\n        state is INIT or RESET, in which case it will attempt a transition to\n        INIT prior to transition to RTR. As a result, if current state doesn't\n        support transition to INIT, an exception will be raised.\n        The comp mask provided to the kernel includes the needed bits for 2RTR\n        transition for this QP type.\n        :param qp_attr: QPAttr object containing the needed attributes for\n                        2RTR transition.\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_2qp_2QP_15to_rtr = {"to_rtr", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_15to_rtr, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2qp_2QP_14to_rtr};
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_15to_rtr(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_qp_attr = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("to_rtr (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_qp_attr,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1104, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1104, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "to_rtr", 0) < (0)) __PYX_ERR(0, 1104, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("to_rtr", 1, 1, 1, i); __PYX_ERR(0, 1104, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1104, __pyx_L3_error)
    }
    __pyx_v_qp_attr = ((struct __pyx_obj_7pyverbs_2qp_QPAttr *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("to_rtr", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1104, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QP.to_rtr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_qp_attr), __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPAttr, 1, "qp_attr", 0))) __PYX_ERR(0, 1104, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp_2QP_14to_rtr(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self), __pyx_v_qp_attr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_2QP_14to_rtr(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_qp_attr) {
  PyObject *__pyx_v_mask = NULL;
  int __pyx_v_rc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  size_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("to_rtr", 0);

  /* "pyverbs/qp.pyx":1116
 *         :return: None
 *         """
 *         if self.qp_state != e.IBV_QPS_INIT: #assume reset             # <<<<<<<<<<<<<<
 *             self.to_init(qp_attr)
 *         mask = self._get_comp_mask('RTR')
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_qp_state); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py(IBV_QPS_INIT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1116, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 1116, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pyverbs/qp.pyx":1117
 *         """
 *         if self.qp_state != e.IBV_QPS_INIT: #assume reset
 *             self.to_init(qp_attr)             # <<<<<<<<<<<<<<
 *         mask = self._get_comp_mask('RTR')
 *         qp_attr.qp_state = e.IBV_QPS_RTR
*/
    __pyx_t_2 = ((PyObject *)__pyx_v_self);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, ((PyObject *)__pyx_v_qp_attr)};
      __pyx_t_3 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_to_init, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1117, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyverbs/qp.pyx":1116
 *         :return: None
 *         """
 *         if self.qp_state != e.IBV_QPS_INIT: #assume reset             # <<<<<<<<<<<<<<
 *             self.to_init(qp_attr)
 *         mask = self._get_comp_mask('RTR')
*/
  }

  /* "pyverbs/qp.pyx":1118
 *         if self.qp_state != e.IBV_QPS_INIT: #assume reset
 *             self.to_init(qp_attr)
 *         mask = self._get_comp_mask('RTR')             # <<<<<<<<<<<<<<
 *         qp_attr.qp_state = e.IBV_QPS_RTR
 *         rc = v.ibv_modify_qp(self.qp, &qp_attr.attr, mask)
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_n_u_RTR};
    __pyx_t_3 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_comp_mask, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_v_mask = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":1119
 *             self.to_init(qp_attr)
 *         mask = self._get_comp_mask('RTR')
 *         qp_attr.qp_state = e.IBV_QPS_RTR             # <<<<<<<<<<<<<<
 *         rc = v.ibv_modify_qp(self.qp, &qp_attr.attr, mask)
 *         if rc != 0:
*/
  __pyx_t_3 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py(IBV_QPS_RTR); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_qp_attr), __pyx_mstate_global->__pyx_n_u_qp_state, __pyx_t_3) < (0)) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":1120
 *         mask = self._get_comp_mask('RTR')
 *         qp_attr.qp_state = e.IBV_QPS_RTR
 *         rc = v.ibv_modify_qp(self.qp, &qp_attr.attr, mask)             # <<<<<<<<<<<<<<
 *         if rc != 0:
 *             raise PyverbsRDMAError('Failed to modify QP state to RTR', rc)
*/
  __pyx_t_6 = __Pyx_PyLong_As_int(__pyx_v_mask); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1120, __pyx_L1_error)
  __pyx_v_rc = ibv_modify_qp(__pyx_v_self->qp, (&__pyx_v_qp_attr->attr), __pyx_t_6);

  /* "pyverbs/qp.pyx":1121
 *         qp_attr.qp_state = e.IBV_QPS_RTR
 *         rc = v.ibv_modify_qp(self.qp, &qp_attr.attr, mask)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Failed to modify QP state to RTR', rc)
 * 
*/
  __pyx_t_4 = (__pyx_v_rc != 0);
  if (unlikely(__pyx_t_4)) {

    /* "pyverbs/qp.pyx":1122
 *         rc = v.ibv_modify_qp(self.qp, &qp_attr.attr, mask)
 *         if rc != 0:
 *             raise PyverbsRDMAError('Failed to modify QP state to RTR', rc)             # <<<<<<<<<<<<<<
 * 
 *     def to_rts(self, QPAttr qp_attr):
*/
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PyLong_From_int(__pyx_v_rc); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Failed_to_modify_QP_state_to_RTR, __pyx_t_7};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_1, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1122, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1122, __pyx_L1_error)

    /* "pyverbs/qp.pyx":1121
 *         qp_attr.qp_state = e.IBV_QPS_RTR
 *         rc = v.ibv_modify_qp(self.qp, &qp_attr.attr, mask)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Failed to modify QP state to RTR', rc)
 * 
*/
  }

  /* "pyverbs/qp.pyx":1104
 *             raise PyverbsRDMAError('Failed to modify QP state to init', rc)
 * 
 *     def to_rtr(self, QPAttr qp_attr):             # <<<<<<<<<<<<<<
 *         """
 *         Modify the current QP's state to RTR. It assumes that its current
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyverbs.qp.QP.to_rtr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_mask);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1124
 *             raise PyverbsRDMAError('Failed to modify QP state to RTR', rc)
 * 
 *     def to_rts(self, QPAttr qp_attr):             # <<<<<<<<<<<<<<
 *         """
 *         Modify the current QP's state to RTS. It assumes that its current
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_17to_rts(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2qp_2QP_16to_rts, "\n        Modify the current QP's state to RTS. It assumes that its current\n        state is either RTR, INIT or RESET. If current state is not RTR, to_rtr()\n        will be called.\n        The comp mask provided to the kernel includes the needed bits for 2RTS\n        transition for this QP type.\n        :param qp_attr: QPAttr object containing the needed attributes for\n                        2RTS transition.\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_2qp_2QP_17to_rts = {"to_rts", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_17to_rts, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2qp_2QP_16to_rts};
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_17to_rts(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_qp_attr = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("to_rts (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_qp_attr,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1124, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1124, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "to_rts", 0) < (0)) __PYX_ERR(0, 1124, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("to_rts", 1, 1, 1, i); __PYX_ERR(0, 1124, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1124, __pyx_L3_error)
    }
    __pyx_v_qp_attr = ((struct __pyx_obj_7pyverbs_2qp_QPAttr *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("to_rts", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1124, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QP.to_rts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_qp_attr), __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPAttr, 1, "qp_attr", 0))) __PYX_ERR(0, 1124, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp_2QP_16to_rts(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self), __pyx_v_qp_attr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_2QP_16to_rts(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_qp_attr) {
  PyObject *__pyx_v_mask = NULL;
  int __pyx_v_rc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  size_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("to_rts", 0);

  /* "pyverbs/qp.pyx":1135
 *         :return: None
 *         """
 *         if self.qp_state != e.IBV_QPS_RTR: #assume reset/init             # <<<<<<<<<<<<<<
 *             self.to_rtr(qp_attr)
 *         mask = self._get_comp_mask('RTS')
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_qp_state); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py(IBV_QPS_RTR); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1135, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 1135, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "pyverbs/qp.pyx":1136
 *         """
 *         if self.qp_state != e.IBV_QPS_RTR: #assume reset/init
 *             self.to_rtr(qp_attr)             # <<<<<<<<<<<<<<
 *         mask = self._get_comp_mask('RTS')
 *         qp_attr.qp_state = e.IBV_QPS_RTS
*/
    __pyx_t_2 = ((PyObject *)__pyx_v_self);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, ((PyObject *)__pyx_v_qp_attr)};
      __pyx_t_3 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_to_rtr, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1136, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyverbs/qp.pyx":1135
 *         :return: None
 *         """
 *         if self.qp_state != e.IBV_QPS_RTR: #assume reset/init             # <<<<<<<<<<<<<<
 *             self.to_rtr(qp_attr)
 *         mask = self._get_comp_mask('RTS')
*/
  }

  /* "pyverbs/qp.pyx":1137
 *         if self.qp_state != e.IBV_QPS_RTR: #assume reset/init
 *             self.to_rtr(qp_attr)
 *         mask = self._get_comp_mask('RTS')             # <<<<<<<<<<<<<<
 *         qp_attr.qp_state = e.IBV_QPS_RTS
 *         rc = v.ibv_modify_qp(self.qp, &qp_attr.attr, mask)
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_n_u_RTS};
    __pyx_t_3 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_comp_mask, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_v_mask = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":1138
 *             self.to_rtr(qp_attr)
 *         mask = self._get_comp_mask('RTS')
 *         qp_attr.qp_state = e.IBV_QPS_RTS             # <<<<<<<<<<<<<<
 *         rc = v.ibv_modify_qp(self.qp, &qp_attr.attr, mask)
 *         if rc != 0:
*/
  __pyx_t_3 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py(IBV_QPS_RTS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_qp_attr), __pyx_mstate_global->__pyx_n_u_qp_state, __pyx_t_3) < (0)) __PYX_ERR(0, 1138, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/qp.pyx":1139
 *         mask = self._get_comp_mask('RTS')
 *         qp_attr.qp_state = e.IBV_QPS_RTS
 *         rc = v.ibv_modify_qp(self.qp, &qp_attr.attr, mask)             # <<<<<<<<<<<<<<
 *         if rc != 0:
 *             raise PyverbsRDMAError('Failed to modify QP state to RTS', rc)
*/
  __pyx_t_6 = __Pyx_PyLong_As_int(__pyx_v_mask); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1139, __pyx_L1_error)
  __pyx_v_rc = ibv_modify_qp(__pyx_v_self->qp, (&__pyx_v_qp_attr->attr), __pyx_t_6);

  /* "pyverbs/qp.pyx":1140
 *         qp_attr.qp_state = e.IBV_QPS_RTS
 *         rc = v.ibv_modify_qp(self.qp, &qp_attr.attr, mask)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Failed to modify QP state to RTS', rc)
 * 
*/
  __pyx_t_4 = (__pyx_v_rc != 0);
  if (unlikely(__pyx_t_4)) {

    /* "pyverbs/qp.pyx":1141
 *         rc = v.ibv_modify_qp(self.qp, &qp_attr.attr, mask)
 *         if rc != 0:
 *             raise PyverbsRDMAError('Failed to modify QP state to RTS', rc)             # <<<<<<<<<<<<<<
 * 
 *     def query(self, attr_mask):
*/
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PyLong_From_int(__pyx_v_rc); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Failed_to_modify_QP_state_to_RTS, __pyx_t_7};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_1, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1141, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1141, __pyx_L1_error)

    /* "pyverbs/qp.pyx":1140
 *         qp_attr.qp_state = e.IBV_QPS_RTS
 *         rc = v.ibv_modify_qp(self.qp, &qp_attr.attr, mask)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Failed to modify QP state to RTS', rc)
 * 
*/
  }

  /* "pyverbs/qp.pyx":1124
 *             raise PyverbsRDMAError('Failed to modify QP state to RTR', rc)
 * 
 *     def to_rts(self, QPAttr qp_attr):             # <<<<<<<<<<<<<<
 *         """
 *         Modify the current QP's state to RTS. It assumes that its current
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyverbs.qp.QP.to_rts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_mask);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1143
 *             raise PyverbsRDMAError('Failed to modify QP state to RTS', rc)
 * 
 *     def query(self, attr_mask):             # <<<<<<<<<<<<<<
 *         """
 *         Query the QP
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_19query(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2qp_2QP_18query, "\n        Query the QP\n        :param attr_mask: The minimum list of attributes to retrieve. Some\n                          devices may return additional attributes as well\n                          (see enum ibv_qp_attr_mask)\n        :return: (QPAttr, QPInitAttr) tuple containing the QP requested\n                 attributes\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_2qp_2QP_19query = {"query", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_19query, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2qp_2QP_18query};
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_19query(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_attr_mask = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("query (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_attr_mask,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1143, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1143, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "query", 0) < (0)) __PYX_ERR(0, 1143, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("query", 1, 1, 1, i); __PYX_ERR(0, 1143, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1143, __pyx_L3_error)
    }
    __pyx_v_attr_mask = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("query", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1143, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QP.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_2QP_18query(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self), __pyx_v_attr_mask);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_2QP_18query(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, PyObject *__pyx_v_attr_mask) {
  struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_attr = NULL;
  struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_init_attr = NULL;
  int __pyx_v_rc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("query", 0);

  /* "pyverbs/qp.pyx":1152
 *                  attributes
 *         """
 *         attr = QPAttr()             # <<<<<<<<<<<<<<
 *         init_attr = QPInitAttr()
 *         rc = v.ibv_query_qp(self.qp, &attr.attr, attr_mask, &init_attr.attr)
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPAttr, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1152, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_attr = ((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":1153
 *         """
 *         attr = QPAttr()
 *         init_attr = QPInitAttr()             # <<<<<<<<<<<<<<
 *         rc = v.ibv_query_qp(self.qp, &attr.attr, attr_mask, &init_attr.attr)
 *         if rc != 0:
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPInitAttr, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1153, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_init_attr = ((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":1154
 *         attr = QPAttr()
 *         init_attr = QPInitAttr()
 *         rc = v.ibv_query_qp(self.qp, &attr.attr, attr_mask, &init_attr.attr)             # <<<<<<<<<<<<<<
 *         if rc != 0:
 *             raise PyverbsRDMAError('Failed to query QP', rc)
*/
  __pyx_t_4 = __Pyx_PyLong_As_int(__pyx_v_attr_mask); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1154, __pyx_L1_error)
  __pyx_v_rc = ibv_query_qp(__pyx_v_self->qp, (&__pyx_v_attr->attr), __pyx_t_4, (&__pyx_v_init_attr->attr));

  /* "pyverbs/qp.pyx":1155
 *         init_attr = QPInitAttr()
 *         rc = v.ibv_query_qp(self.qp, &attr.attr, attr_mask, &init_attr.attr)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Failed to query QP', rc)
 *         return attr, init_attr
*/
  __pyx_t_5 = (__pyx_v_rc != 0);
  if (unlikely(__pyx_t_5)) {

    /* "pyverbs/qp.pyx":1156
 *         rc = v.ibv_query_qp(self.qp, &attr.attr, attr_mask, &init_attr.attr)
 *         if rc != 0:
 *             raise PyverbsRDMAError('Failed to query QP', rc)             # <<<<<<<<<<<<<<
 *         return attr, init_attr
 * 
*/
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyLong_From_int(__pyx_v_rc); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
      __pyx_t_3 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Failed_to_query_QP, __pyx_t_7};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_3, (3-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1156, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1156, __pyx_L1_error)

    /* "pyverbs/qp.pyx":1155
 *         init_attr = QPInitAttr()
 *         rc = v.ibv_query_qp(self.qp, &attr.attr, attr_mask, &init_attr.attr)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Failed to query QP', rc)
 *         return attr, init_attr
*/
  }

  /* "pyverbs/qp.pyx":1157
 *         if rc != 0:
 *             raise PyverbsRDMAError('Failed to query QP', rc)
 *         return attr, init_attr             # <<<<<<<<<<<<<<
 * 
 *     def modify(self, QPAttr qp_attr not None, comp_mask):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_attr);
  __Pyx_GIVEREF((PyObject *)__pyx_v_attr);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_attr)) != (0)) __PYX_ERR(0, 1157, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_init_attr);
  __Pyx_GIVEREF((PyObject *)__pyx_v_init_attr);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_init_attr)) != (0)) __PYX_ERR(0, 1157, __pyx_L1_error);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":1143
 *             raise PyverbsRDMAError('Failed to modify QP state to RTS', rc)
 * 
 *     def query(self, attr_mask):             # <<<<<<<<<<<<<<
 *         """
 *         Query the QP
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyverbs.qp.QP.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_attr);
  __Pyx_XDECREF((PyObject *)__pyx_v_init_attr);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1159
 *         return attr, init_attr
 * 
 *     def modify(self, QPAttr qp_attr not None, comp_mask):             # <<<<<<<<<<<<<<
 *         """
 *         Modify the QP
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_21modify(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2qp_2QP_20modify, "\n        Modify the QP\n        :param qp_attr: A QPAttr object with updated values to be applied to\n                        the QP\n        :param comp_mask: A bitmask specifying which QP attributes should be\n                          modified (see enum ibv_qp_attr_mask)\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_2qp_2QP_21modify = {"modify", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_21modify, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2qp_2QP_20modify};
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_21modify(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_qp_attr = 0;
  PyObject *__pyx_v_comp_mask = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("modify (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_qp_attr,&__pyx_mstate_global->__pyx_n_u_comp_mask,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1159, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1159, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1159, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "modify", 0) < (0)) __PYX_ERR(0, 1159, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("modify", 1, 2, 2, i); __PYX_ERR(0, 1159, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1159, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1159, __pyx_L3_error)
    }
    __pyx_v_qp_attr = ((struct __pyx_obj_7pyverbs_2qp_QPAttr *)values[0]);
    __pyx_v_comp_mask = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("modify", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1159, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QP.modify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_qp_attr), __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPAttr, 0, "qp_attr", 0))) __PYX_ERR(0, 1159, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp_2QP_20modify(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self), __pyx_v_qp_attr, __pyx_v_comp_mask);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_2QP_20modify(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_qp_attr, PyObject *__pyx_v_comp_mask) {
  int __pyx_v_rc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("modify", 0);

  /* "pyverbs/qp.pyx":1168
 *         :return: None
 *         """
 *         rc = v.ibv_modify_qp(self.qp, &qp_attr.attr, comp_mask)             # <<<<<<<<<<<<<<
 *         if rc != 0:
 *             raise PyverbsRDMAError('Failed to modify QP', rc)
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_comp_mask); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1168, __pyx_L1_error)
  __pyx_v_rc = ibv_modify_qp(__pyx_v_self->qp, (&__pyx_v_qp_attr->attr), __pyx_t_1);

  /* "pyverbs/qp.pyx":1169
 *         """
 *         rc = v.ibv_modify_qp(self.qp, &qp_attr.attr, comp_mask)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Failed to modify QP', rc)
 * 
*/
  __pyx_t_2 = (__pyx_v_rc != 0);
  if (unlikely(__pyx_t_2)) {

    /* "pyverbs/qp.pyx":1170
 *         rc = v.ibv_modify_qp(self.qp, &qp_attr.attr, comp_mask)
 *         if rc != 0:
 *             raise PyverbsRDMAError('Failed to modify QP', rc)             # <<<<<<<<<<<<<<
 * 
 *     def post_recv(self, RecvWR wr not None, RecvWR bad_wr=None):
*/
    __pyx_t_4 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1170, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyLong_From_int(__pyx_v_rc); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1170, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_Failed_to_modify_QP, __pyx_t_6};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_7, (3-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1170, __pyx_L1_error)

    /* "pyverbs/qp.pyx":1169
 *         """
 *         rc = v.ibv_modify_qp(self.qp, &qp_attr.attr, comp_mask)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Failed to modify QP', rc)
 * 
*/
  }

  /* "pyverbs/qp.pyx":1159
 *         return attr, init_attr
 * 
 *     def modify(self, QPAttr qp_attr not None, comp_mask):             # <<<<<<<<<<<<<<
 *         """
 *         Modify the QP
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyverbs.qp.QP.modify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1172
 *             raise PyverbsRDMAError('Failed to modify QP', rc)
 * 
 *     def post_recv(self, RecvWR wr not None, RecvWR bad_wr=None):             # <<<<<<<<<<<<<<
 *         """
 *         Post a receive WR on the QP.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_23post_recv(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2qp_2QP_22post_recv, "\n        Post a receive WR on the QP.\n        :param wr: The work request to post\n        :param bad_wr: A RecvWR object to hold the bad WR if it is available in\n                       case of a failure\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_2qp_2QP_23post_recv = {"post_recv", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_23post_recv, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2qp_2QP_22post_recv};
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_23post_recv(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_2wr_RecvWR *__pyx_v_wr = 0;
  struct __pyx_obj_7pyverbs_2wr_RecvWR *__pyx_v_bad_wr = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("post_recv (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_wr,&__pyx_mstate_global->__pyx_n_u_bad_wr,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1172, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1172, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1172, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "post_recv", 0) < (0)) __PYX_ERR(0, 1172, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2wr_RecvWR *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("post_recv", 0, 1, 2, i); __PYX_ERR(0, 1172, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1172, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1172, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[1]) values[1] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2wr_RecvWR *)Py_None));
    }
    __pyx_v_wr = ((struct __pyx_obj_7pyverbs_2wr_RecvWR *)values[0]);
    __pyx_v_bad_wr = ((struct __pyx_obj_7pyverbs_2wr_RecvWR *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("post_recv", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 1172, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QP.post_recv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_wr), __pyx_mstate_global->__pyx_ptype_7pyverbs_2wr_RecvWR, 0, "wr", 0))) __PYX_ERR(0, 1172, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_bad_wr), __pyx_mstate_global->__pyx_ptype_7pyverbs_2wr_RecvWR, 1, "bad_wr", 0))) __PYX_ERR(0, 1172, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp_2QP_22post_recv(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self), __pyx_v_wr, __pyx_v_bad_wr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_2QP_22post_recv(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, struct __pyx_obj_7pyverbs_2wr_RecvWR *__pyx_v_wr, struct __pyx_obj_7pyverbs_2wr_RecvWR *__pyx_v_bad_wr) {
  struct ibv_recv_wr *__pyx_v_my_bad_wr;
  int __pyx_v_rc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("post_recv", 0);

  /* "pyverbs/qp.pyx":1183
 *         # In order to provide a pointer to a pointer, use a temporary cdef'ed
 *         # variable.
 *         rc = v.ibv_post_recv(self.qp, &wr.recv_wr, &my_bad_wr)             # <<<<<<<<<<<<<<
 *         if rc != 0:
 *             if (bad_wr):
*/
  __pyx_v_rc = ibv_post_recv(__pyx_v_self->qp, (&__pyx_v_wr->recv_wr), (&__pyx_v_my_bad_wr));

  /* "pyverbs/qp.pyx":1184
 *         # variable.
 *         rc = v.ibv_post_recv(self.qp, &wr.recv_wr, &my_bad_wr)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             if (bad_wr):
 *                 memcpy(&bad_wr.recv_wr, my_bad_wr, sizeof(bad_wr.recv_wr))
*/
  __pyx_t_1 = (__pyx_v_rc != 0);
  if (__pyx_t_1) {

    /* "pyverbs/qp.pyx":1185
 *         rc = v.ibv_post_recv(self.qp, &wr.recv_wr, &my_bad_wr)
 *         if rc != 0:
 *             if (bad_wr):             # <<<<<<<<<<<<<<
 *                 memcpy(&bad_wr.recv_wr, my_bad_wr, sizeof(bad_wr.recv_wr))
 *             raise PyverbsRDMAError('Failed to post recv', rc)
*/
    __pyx_t_1 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_bad_wr)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1185, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "pyverbs/qp.pyx":1186
 *         if rc != 0:
 *             if (bad_wr):
 *                 memcpy(&bad_wr.recv_wr, my_bad_wr, sizeof(bad_wr.recv_wr))             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Failed to post recv', rc)
 * 
*/
      (void)(memcpy((&__pyx_v_bad_wr->recv_wr), __pyx_v_my_bad_wr, (sizeof(__pyx_v_bad_wr->recv_wr))));

      /* "pyverbs/qp.pyx":1185
 *         rc = v.ibv_post_recv(self.qp, &wr.recv_wr, &my_bad_wr)
 *         if rc != 0:
 *             if (bad_wr):             # <<<<<<<<<<<<<<
 *                 memcpy(&bad_wr.recv_wr, my_bad_wr, sizeof(bad_wr.recv_wr))
 *             raise PyverbsRDMAError('Failed to post recv', rc)
*/
    }

    /* "pyverbs/qp.pyx":1187
 *             if (bad_wr):
 *                 memcpy(&bad_wr.recv_wr, my_bad_wr, sizeof(bad_wr.recv_wr))
 *             raise PyverbsRDMAError('Failed to post recv', rc)             # <<<<<<<<<<<<<<
 * 
 *     def post_send(self, SendWR wr not None, SendWR bad_wr=None):
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyLong_From_int(__pyx_v_rc); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_6 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Failed_to_post_recv, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_6, (3-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1187, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1187, __pyx_L1_error)

    /* "pyverbs/qp.pyx":1184
 *         # variable.
 *         rc = v.ibv_post_recv(self.qp, &wr.recv_wr, &my_bad_wr)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             if (bad_wr):
 *                 memcpy(&bad_wr.recv_wr, my_bad_wr, sizeof(bad_wr.recv_wr))
*/
  }

  /* "pyverbs/qp.pyx":1172
 *             raise PyverbsRDMAError('Failed to modify QP', rc)
 * 
 *     def post_recv(self, RecvWR wr not None, RecvWR bad_wr=None):             # <<<<<<<<<<<<<<
 *         """
 *         Post a receive WR on the QP.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyverbs.qp.QP.post_recv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1189
 *             raise PyverbsRDMAError('Failed to post recv', rc)
 * 
 *     def post_send(self, SendWR wr not None, SendWR bad_wr=None):             # <<<<<<<<<<<<<<
 *         """
 *         Post a send WR on the QP.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_25post_send(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2qp_2QP_24post_send, "\n        Post a send WR on the QP.\n        :param wr: The work request to post\n        :param bad_wr: A SendWR object to hold the bad WR if it is available in\n                       case of a failure\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_2qp_2QP_25post_send = {"post_send", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_25post_send, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2qp_2QP_24post_send};
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_25post_send(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_2wr_SendWR *__pyx_v_wr = 0;
  struct __pyx_obj_7pyverbs_2wr_SendWR *__pyx_v_bad_wr = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("post_send (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_wr,&__pyx_mstate_global->__pyx_n_u_bad_wr,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1189, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1189, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1189, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "post_send", 0) < (0)) __PYX_ERR(0, 1189, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2wr_SendWR *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("post_send", 0, 1, 2, i); __PYX_ERR(0, 1189, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1189, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1189, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[1]) values[1] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2wr_SendWR *)Py_None));
    }
    __pyx_v_wr = ((struct __pyx_obj_7pyverbs_2wr_SendWR *)values[0]);
    __pyx_v_bad_wr = ((struct __pyx_obj_7pyverbs_2wr_SendWR *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("post_send", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 1189, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QP.post_send", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_wr), __pyx_mstate_global->__pyx_ptype_7pyverbs_2wr_SendWR, 0, "wr", 0))) __PYX_ERR(0, 1189, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_bad_wr), __pyx_mstate_global->__pyx_ptype_7pyverbs_2wr_SendWR, 1, "bad_wr", 0))) __PYX_ERR(0, 1189, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp_2QP_24post_send(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self), __pyx_v_wr, __pyx_v_bad_wr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_2QP_24post_send(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, struct __pyx_obj_7pyverbs_2wr_SendWR *__pyx_v_wr, struct __pyx_obj_7pyverbs_2wr_SendWR *__pyx_v_bad_wr) {
  struct ibv_send_wr *__pyx_v_my_bad_wr;
  int __pyx_v_rc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("post_send", 0);

  /* "pyverbs/qp.pyx":1200
 *         # variable.
 *         cdef v.ibv_send_wr *my_bad_wr
 *         rc = v.ibv_post_send(self.qp, &wr.send_wr, &my_bad_wr)             # <<<<<<<<<<<<<<
 *         if rc != 0:
 *             if (bad_wr):
*/
  __pyx_v_rc = ibv_post_send(__pyx_v_self->qp, (&__pyx_v_wr->send_wr), (&__pyx_v_my_bad_wr));

  /* "pyverbs/qp.pyx":1201
 *         cdef v.ibv_send_wr *my_bad_wr
 *         rc = v.ibv_post_send(self.qp, &wr.send_wr, &my_bad_wr)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             if (bad_wr):
 *                 memcpy(&bad_wr.send_wr, my_bad_wr, sizeof(bad_wr.send_wr))
*/
  __pyx_t_1 = (__pyx_v_rc != 0);
  if (__pyx_t_1) {

    /* "pyverbs/qp.pyx":1202
 *         rc = v.ibv_post_send(self.qp, &wr.send_wr, &my_bad_wr)
 *         if rc != 0:
 *             if (bad_wr):             # <<<<<<<<<<<<<<
 *                 memcpy(&bad_wr.send_wr, my_bad_wr, sizeof(bad_wr.send_wr))
 *             raise PyverbsRDMAError('Failed to post send', rc)
*/
    __pyx_t_1 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_bad_wr)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1202, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "pyverbs/qp.pyx":1203
 *         if rc != 0:
 *             if (bad_wr):
 *                 memcpy(&bad_wr.send_wr, my_bad_wr, sizeof(bad_wr.send_wr))             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Failed to post send', rc)
 * 
*/
      (void)(memcpy((&__pyx_v_bad_wr->send_wr), __pyx_v_my_bad_wr, (sizeof(__pyx_v_bad_wr->send_wr))));

      /* "pyverbs/qp.pyx":1202
 *         rc = v.ibv_post_send(self.qp, &wr.send_wr, &my_bad_wr)
 *         if rc != 0:
 *             if (bad_wr):             # <<<<<<<<<<<<<<
 *                 memcpy(&bad_wr.send_wr, my_bad_wr, sizeof(bad_wr.send_wr))
 *             raise PyverbsRDMAError('Failed to post send', rc)
*/
    }

    /* "pyverbs/qp.pyx":1204
 *             if (bad_wr):
 *                 memcpy(&bad_wr.send_wr, my_bad_wr, sizeof(bad_wr.send_wr))
 *             raise PyverbsRDMAError('Failed to post send', rc)             # <<<<<<<<<<<<<<
 * 
 *     def set_ece(self, ECE ece):
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyLong_From_int(__pyx_v_rc); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_6 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Failed_to_post_send, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_6, (3-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1204, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1204, __pyx_L1_error)

    /* "pyverbs/qp.pyx":1201
 *         cdef v.ibv_send_wr *my_bad_wr
 *         rc = v.ibv_post_send(self.qp, &wr.send_wr, &my_bad_wr)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             if (bad_wr):
 *                 memcpy(&bad_wr.send_wr, my_bad_wr, sizeof(bad_wr.send_wr))
*/
  }

  /* "pyverbs/qp.pyx":1189
 *             raise PyverbsRDMAError('Failed to post recv', rc)
 * 
 *     def post_send(self, SendWR wr not None, SendWR bad_wr=None):             # <<<<<<<<<<<<<<
 *         """
 *         Post a send WR on the QP.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyverbs.qp.QP.post_send", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1206
 *             raise PyverbsRDMAError('Failed to post send', rc)
 * 
 *     def set_ece(self, ECE ece):             # <<<<<<<<<<<<<<
 *         """
 *         Set ECE options and use them for QP configuration stage
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_27set_ece(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2qp_2QP_26set_ece, "\n        Set ECE options and use them for QP configuration stage\n        :param ece: The requested ECE values.\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_2qp_2QP_27set_ece = {"set_ece", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_27set_ece, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2qp_2QP_26set_ece};
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_27set_ece(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_ece = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_ece (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_ece,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1206, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1206, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "set_ece", 0) < (0)) __PYX_ERR(0, 1206, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("set_ece", 1, 1, 1, i); __PYX_ERR(0, 1206, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1206, __pyx_L3_error)
    }
    __pyx_v_ece = ((struct __pyx_obj_7pyverbs_2qp_ECE *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_ece", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1206, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QP.set_ece", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_ece), __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_ECE, 1, "ece", 0))) __PYX_ERR(0, 1206, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp_2QP_26set_ece(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self), __pyx_v_ece);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_2QP_26set_ece(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_ece) {
  int __pyx_v_rc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_ece", 0);

  /* "pyverbs/qp.pyx":1212
 *         :return: None
 *         """
 *         if ece.ece.vendor_id == 0:             # <<<<<<<<<<<<<<
 *             return
 * 
*/
  __pyx_t_1 = (__pyx_v_ece->ece.vendor_id == 0);
  if (__pyx_t_1) {

    /* "pyverbs/qp.pyx":1213
 *         """
 *         if ece.ece.vendor_id == 0:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         rc = v.ibv_set_ece(self.qp, &ece.ece)
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "pyverbs/qp.pyx":1212
 *         :return: None
 *         """
 *         if ece.ece.vendor_id == 0:             # <<<<<<<<<<<<<<
 *             return
 * 
*/
  }

  /* "pyverbs/qp.pyx":1215
 *             return
 * 
 *         rc = v.ibv_set_ece(self.qp, &ece.ece)             # <<<<<<<<<<<<<<
 *         if rc != 0:
 *             raise PyverbsRDMAError('Failed to set ECE', rc)
*/
  __pyx_v_rc = ibv_set_ece(__pyx_v_self->qp, (&__pyx_v_ece->ece));

  /* "pyverbs/qp.pyx":1216
 * 
 *         rc = v.ibv_set_ece(self.qp, &ece.ece)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Failed to set ECE', rc)
 * 
*/
  __pyx_t_1 = (__pyx_v_rc != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pyverbs/qp.pyx":1217
 *         rc = v.ibv_set_ece(self.qp, &ece.ece)
 *         if rc != 0:
 *             raise PyverbsRDMAError('Failed to set ECE', rc)             # <<<<<<<<<<<<<<
 * 
 *     def query_ece(self):
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyLong_From_int(__pyx_v_rc); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_6 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Failed_to_set_ECE, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_6, (3-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1217, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1217, __pyx_L1_error)

    /* "pyverbs/qp.pyx":1216
 * 
 *         rc = v.ibv_set_ece(self.qp, &ece.ece)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Failed to set ECE', rc)
 * 
*/
  }

  /* "pyverbs/qp.pyx":1206
 *             raise PyverbsRDMAError('Failed to post send', rc)
 * 
 *     def set_ece(self, ECE ece):             # <<<<<<<<<<<<<<
 *         """
 *         Set ECE options and use them for QP configuration stage
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyverbs.qp.QP.set_ece", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1219
 *             raise PyverbsRDMAError('Failed to set ECE', rc)
 * 
 *     def query_ece(self):             # <<<<<<<<<<<<<<
 *         """
 *         Query QPs ECE options
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_29query_ece(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2qp_2QP_28query_ece, "\n        Query QPs ECE options\n        :return: ECE object with this QP ece configuration.\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_2qp_2QP_29query_ece = {"query_ece", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_29query_ece, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2qp_2QP_28query_ece};
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_29query_ece(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("query_ece (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("query_ece", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("query_ece", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2qp_2QP_28query_ece(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_2QP_28query_ece(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self) {
  struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_ece = NULL;
  int __pyx_v_rc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("query_ece", 0);

  /* "pyverbs/qp.pyx":1224
 *         :return: ECE object with this QP ece configuration.
 *         """
 *         ece = ECE()             # <<<<<<<<<<<<<<
 *         rc = v.ibv_query_ece(self.qp, &ece.ece)
 *         if rc != 0:
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_ECE, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1224, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_ece = ((struct __pyx_obj_7pyverbs_2qp_ECE *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":1225
 *         """
 *         ece = ECE()
 *         rc = v.ibv_query_ece(self.qp, &ece.ece)             # <<<<<<<<<<<<<<
 *         if rc != 0:
 *             raise PyverbsRDMAError('Failed to query ECE', rc)
*/
  __pyx_v_rc = ibv_query_ece(__pyx_v_self->qp, (&__pyx_v_ece->ece));

  /* "pyverbs/qp.pyx":1226
 *         ece = ECE()
 *         rc = v.ibv_query_ece(self.qp, &ece.ece)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Failed to query ECE', rc)
 *         return ece
*/
  __pyx_t_4 = (__pyx_v_rc != 0);
  if (unlikely(__pyx_t_4)) {

    /* "pyverbs/qp.pyx":1227
 *         rc = v.ibv_query_ece(self.qp, &ece.ece)
 *         if rc != 0:
 *             raise PyverbsRDMAError('Failed to query ECE', rc)             # <<<<<<<<<<<<<<
 *         return ece
 * 
*/
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyLong_From_int(__pyx_v_rc); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_3 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Failed_to_query_ECE, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_3, (3-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1227, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1227, __pyx_L1_error)

    /* "pyverbs/qp.pyx":1226
 *         ece = ECE()
 *         rc = v.ibv_query_ece(self.qp, &ece.ece)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Failed to query ECE', rc)
 *         return ece
*/
  }

  /* "pyverbs/qp.pyx":1228
 *         if rc != 0:
 *             raise PyverbsRDMAError('Failed to query ECE', rc)
 *         return ece             # <<<<<<<<<<<<<<
 * 
 *     def bind_mw(self, MW mw not None, MWBind mw_bind):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_ece);
  __pyx_r = ((PyObject *)__pyx_v_ece);
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":1219
 *             raise PyverbsRDMAError('Failed to set ECE', rc)
 * 
 *     def query_ece(self):             # <<<<<<<<<<<<<<
 *         """
 *         Query QPs ECE options
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyverbs.qp.QP.query_ece", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_ece);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1230
 *         return ece
 * 
 *     def bind_mw(self, MW mw not None, MWBind mw_bind):             # <<<<<<<<<<<<<<
 *         """
 *         Bind Memory window type 1.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_31bind_mw(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2qp_2QP_30bind_mw, "\n        Bind Memory window type 1.\n        :param mw: The memory window to bind.\n        :param mw_bind: MWBind object, includes the bind attributes.\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_2qp_2QP_31bind_mw = {"bind_mw", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_31bind_mw, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2qp_2QP_30bind_mw};
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_31bind_mw(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_2mr_MW *__pyx_v_mw = 0;
  struct __pyx_obj_7pyverbs_2mr_MWBind *__pyx_v_mw_bind = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bind_mw (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_mw,&__pyx_mstate_global->__pyx_n_u_mw_bind,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1230, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1230, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1230, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "bind_mw", 0) < (0)) __PYX_ERR(0, 1230, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("bind_mw", 1, 2, 2, i); __PYX_ERR(0, 1230, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1230, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1230, __pyx_L3_error)
    }
    __pyx_v_mw = ((struct __pyx_obj_7pyverbs_2mr_MW *)values[0]);
    __pyx_v_mw_bind = ((struct __pyx_obj_7pyverbs_2mr_MWBind *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bind_mw", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1230, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QP.bind_mw", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mw), __pyx_mstate_global->__pyx_ptype_7pyverbs_2mr_MW, 0, "mw", 0))) __PYX_ERR(0, 1230, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mw_bind), __pyx_mstate_global->__pyx_ptype_7pyverbs_2mr_MWBind, 1, "mw_bind", 0))) __PYX_ERR(0, 1230, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp_2QP_30bind_mw(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self), __pyx_v_mw, __pyx_v_mw_bind);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_2QP_30bind_mw(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, struct __pyx_obj_7pyverbs_2mr_MW *__pyx_v_mw, struct __pyx_obj_7pyverbs_2mr_MWBind *__pyx_v_mw_bind) {
  int __pyx_v_rc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bind_mw", 0);

  /* "pyverbs/qp.pyx":1237
 *         :return: None
 *         """
 *         rc = v.ibv_bind_mw(self.qp, mw.mw, &mw_bind.mw_bind)             # <<<<<<<<<<<<<<
 *         if rc != 0:
 *             raise PyverbsRDMAError('Failed to Bind MW', rc)
*/
  __pyx_v_rc = ibv_bind_mw(__pyx_v_self->qp, __pyx_v_mw->mw, (&__pyx_v_mw_bind->mw_bind));

  /* "pyverbs/qp.pyx":1238
 *         """
 *         rc = v.ibv_bind_mw(self.qp, mw.mw, &mw_bind.mw_bind)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Failed to Bind MW', rc)
 * 
*/
  __pyx_t_1 = (__pyx_v_rc != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pyverbs/qp.pyx":1239
 *         rc = v.ibv_bind_mw(self.qp, mw.mw, &mw_bind.mw_bind)
 *         if rc != 0:
 *             raise PyverbsRDMAError('Failed to Bind MW', rc)             # <<<<<<<<<<<<<<
 * 
 *     def query_data_in_order(self, op, flags=0):
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1239, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyLong_From_int(__pyx_v_rc); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1239, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_6 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Failed_to_Bind_MW, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_6, (3-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1239, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1239, __pyx_L1_error)

    /* "pyverbs/qp.pyx":1238
 *         """
 *         rc = v.ibv_bind_mw(self.qp, mw.mw, &mw_bind.mw_bind)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Failed to Bind MW', rc)
 * 
*/
  }

  /* "pyverbs/qp.pyx":1230
 *         return ece
 * 
 *     def bind_mw(self, MW mw not None, MWBind mw_bind):             # <<<<<<<<<<<<<<
 *         """
 *         Bind Memory window type 1.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyverbs.qp.QP.bind_mw", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1241
 *             raise PyverbsRDMAError('Failed to Bind MW', rc)
 * 
 *     def query_data_in_order(self, op, flags=0):             # <<<<<<<<<<<<<<
 *         """
 *         Query if QP data is guaranteed to be in order.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_33query_data_in_order(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2qp_2QP_32query_data_in_order, "\n        Query if QP data is guaranteed to be in order.\n        :param op: Operation type.\n        :param flags: Flags are used to select a query type.\n                      For IBV_QUERY_QP_DATA_IN_ORDER_RETURN_CAPS, the function\n                      will return a capabilities vector. If 0, will query for\n                      IBV_QUERY_QP_DATA_IN_ORDER_WHOLE_MSG support and return\n                      0/1 result. (see enum ibv_query_qp_data_in_order_flags)\n        :return: Return value is determined by flags. For each capability bit,\n                 1 is returned if the data is guaranteed to be written in-order\n                 for selected operation and type, 0 otherwise.\n                 (see enum ibv_query_qp_data_in_order_caps)\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_2qp_2QP_33query_data_in_order = {"query_data_in_order", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_33query_data_in_order, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2qp_2QP_32query_data_in_order};
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_33query_data_in_order(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_op = 0;
  PyObject *__pyx_v_flags = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("query_data_in_order (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_op,&__pyx_mstate_global->__pyx_n_u_flags,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1241, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1241, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1241, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "query_data_in_order", 0) < (0)) __PYX_ERR(0, 1241, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("query_data_in_order", 0, 1, 2, i); __PYX_ERR(0, 1241, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1241, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1241, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    }
    __pyx_v_op = values[0];
    __pyx_v_flags = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("query_data_in_order", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 1241, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QP.query_data_in_order", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_2QP_32query_data_in_order(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self), __pyx_v_op, __pyx_v_flags);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_2QP_32query_data_in_order(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, PyObject *__pyx_v_op, PyObject *__pyx_v_flags) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  enum ibv_wr_opcode __pyx_t_1;
  uint32_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("query_data_in_order", 0);

  /* "pyverbs/qp.pyx":1255
 *                  (see enum ibv_query_qp_data_in_order_caps)
 *         """
 *         return v.ibv_query_qp_data_in_order(self.qp, op, flags)             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((enum ibv_wr_opcode)__Pyx_PyLong_As_enum__ibv_wr_opcode(__pyx_v_op)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1255, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_As_uint32_t(__pyx_v_flags); if (unlikely((__pyx_t_2 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1255, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyLong_From_int(ibv_query_qp_data_in_order(__pyx_v_self->qp, __pyx_t_1, __pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":1241
 *             raise PyverbsRDMAError('Failed to Bind MW', rc)
 * 
 *     def query_data_in_order(self, op, flags=0):             # <<<<<<<<<<<<<<
 *         """
 *         Query if QP data is guaranteed to be in order.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyverbs.qp.QP.query_data_in_order", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1257
 *         return v.ibv_query_qp_data_in_order(self.qp, op, flags)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qp_type(self):
 *         return self.qp.qp_type
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_7qp_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_7qp_type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_2QP_7qp_type___get__(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_2QP_7qp_type___get__(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":1259
 *     @property
 *     def qp_type(self):
 *         return self.qp.qp_type             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(__pyx_v_self->qp->qp_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":1257
 *         return v.ibv_query_qp_data_in_order(self.qp, op, flags)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qp_type(self):
 *         return self.qp.qp_type
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QP.qp_type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1261
 *         return self.qp.qp_type
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qp_state(self):
 *         return self.qp.state
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_8qp_state_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_8qp_state_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_2QP_8qp_state___get__(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_2QP_8qp_state___get__(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":1263
 *     @property
 *     def qp_state(self):
 *         return self.qp.state             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py(__pyx_v_self->qp->state); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":1261
 *         return self.qp.qp_type
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qp_state(self):
 *         return self.qp.state
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QP.qp_state.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1265
 *         return self.qp.state
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qp_num(self):
 *         return self.qp.qp_num
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_6qp_num_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_6qp_num_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_2QP_6qp_num___get__(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_2QP_6qp_num___get__(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":1267
 *     @property
 *     def qp_num(self):
 *         return self.qp.qp_num             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->qp->qp_num); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":1265
 *         return self.qp.state
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qp_num(self):
 *         return self.qp.qp_num
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QP.qp_num.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1269
 *         return self.qp.qp_num
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qp(self):
 *         return <uintptr_t>self.qp
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_2qp_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_2qp_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_2QP_2qp___get__(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_2QP_2qp___get__(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":1271
 *     @property
 *     def qp(self):
 *         return <uintptr_t>self.qp             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_FromSize_t(((uintptr_t)__pyx_v_self->qp)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":1269
 *         return self.qp.qp_num
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qp(self):
 *         return <uintptr_t>self.qp
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QP.qp.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1273
 *         return <uintptr_t>self.qp
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         print_format = '{:22}: {:<20}\n'
 *         return print_format.format('QP type', qp_type_to_str(self.qp_type)) +\
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_35__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_35__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_2QP_34__str__(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_2QP_34__str__(struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self) {
  PyObject *__pyx_v_print_format = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "pyverbs/qp.pyx":1274
 * 
 *     def __str__(self):
 *         print_format = '{:22}: {:<20}\n'             # <<<<<<<<<<<<<<
 *         return print_format.format('QP type', qp_type_to_str(self.qp_type)) +\
 *                print_format.format('  number', self.qp_num) +\
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_22_20);
  __pyx_v_print_format = __pyx_mstate_global->__pyx_kp_u_22_20;

  /* "pyverbs/qp.pyx":1275
 *     def __str__(self):
 *         print_format = '{:22}: {:<20}\n'
 *         return print_format.format('QP type', qp_type_to_str(self.qp_type)) +\             # <<<<<<<<<<<<<<
 *                print_format.format('  number', self.qp_num) +\
 *                print_format.format('  state', qp_state_to_str(self.qp_state))
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_qp_type_to_str); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_qp_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_QP_type, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":1276
 *         print_format = '{:22}: {:<20}\n'
 *         return print_format.format('QP type', qp_type_to_str(self.qp_type)) +\
 *                print_format.format('  number', self.qp_num) +\             # <<<<<<<<<<<<<<
 *                print_format.format('  state', qp_state_to_str(self.qp_state))
 * 
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_qp_num); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_number, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":1275
 *     def __str__(self):
 *         print_format = '{:22}: {:<20}\n'
 *         return print_format.format('QP type', qp_type_to_str(self.qp_type)) +\             # <<<<<<<<<<<<<<
 *                print_format.format('  number', self.qp_num) +\
 *                print_format.format('  state', qp_state_to_str(self.qp_state))
*/
  __pyx_t_1 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":1277
 *         return print_format.format('QP type', qp_type_to_str(self.qp_type)) +\
 *                print_format.format('  number', self.qp_num) +\
 *                print_format.format('  state', qp_state_to_str(self.qp_state))             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_3 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_qp_state_to_str); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_qp_state); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    assert(__pyx_t_3);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_6};
    __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __pyx_t_2 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_state, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/qp.pyx":1276
 *         print_format = '{:22}: {:<20}\n'
 *         return print_format.format('QP type', qp_type_to_str(self.qp_type)) +\
 *                print_format.format('  number', self.qp_num) +\             # <<<<<<<<<<<<<<
 *                print_format.format('  state', qp_state_to_str(self.qp_state))
 * 
*/
  __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":1273
 *         return <uintptr_t>self.qp
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         print_format = '{:22}: {:<20}\n'
 *         return print_format.format('QP type', qp_type_to_str(self.qp_type)) +\
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyverbs.qp.QP.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_print_format);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.qp cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_37__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_2QP_37__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_37__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_37__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2qp_2QP_36__reduce_cython__(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_2QP_36__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.qp cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.qp cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_qp_cannot_be_converted_to_a, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.qp cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QP.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.qp cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.qp cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_39__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_2QP_39__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_39__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_2QP_39__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QP.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_2QP_38__setstate_cython__(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_2QP_38__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QP *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.qp cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.qp cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_qp_cannot_be_converted_to_a, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.qp cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.qp cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QP.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1281
 * 
 * cdef class DataBuffer(PyverbsCM):
 *     def __init__(self, addr, length):             # <<<<<<<<<<<<<<
 *         super().__init__()
 *         self.data.addr = PyLong_AsVoidPtr(addr)
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_10DataBuffer_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7pyverbs_2qp_10DataBuffer_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_addr = 0;
  PyObject *__pyx_v_length = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_addr,&__pyx_mstate_global->__pyx_n_u_length,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1281, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1281, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1281, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 1281, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, i); __PYX_ERR(0, 1281, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1281, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1281, __pyx_L3_error)
    }
    __pyx_v_addr = values[0];
    __pyx_v_length = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1281, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.DataBuffer.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_10DataBuffer___init__(((struct __pyx_obj_7pyverbs_2qp_DataBuffer *)__pyx_v_self), __pyx_v_addr, __pyx_v_length);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_10DataBuffer___init__(struct __pyx_obj_7pyverbs_2qp_DataBuffer *__pyx_v_self, PyObject *__pyx_v_addr, PyObject *__pyx_v_length) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyverbs/qp.pyx":1282
 * cdef class DataBuffer(PyverbsCM):
 *     def __init__(self, addr, length):
 *         super().__init__()             # <<<<<<<<<<<<<<
 *         self.data.addr = PyLong_AsVoidPtr(addr)
 *         self.data.length = length
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_DataBuffer), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":1283
 *     def __init__(self, addr, length):
 *         super().__init__()
 *         self.data.addr = PyLong_AsVoidPtr(addr)             # <<<<<<<<<<<<<<
 *         self.data.length = length
 * 
*/
  __pyx_v_self->data.addr = PyLong_AsVoidPtr(__pyx_v_addr);

  /* "pyverbs/qp.pyx":1284
 *         super().__init__()
 *         self.data.addr = PyLong_AsVoidPtr(addr)
 *         self.data.length = length             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_5 = __Pyx_PyLong_As_size_t(__pyx_v_length); if (unlikely((__pyx_t_5 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1284, __pyx_L1_error)
  __pyx_v_self->data.length = __pyx_t_5;

  /* "pyverbs/qp.pyx":1281
 * 
 * cdef class DataBuffer(PyverbsCM):
 *     def __init__(self, addr, length):             # <<<<<<<<<<<<<<
 *         super().__init__()
 *         self.data.addr = PyLong_AsVoidPtr(addr)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.qp.DataBuffer.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.data cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_10DataBuffer_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_10DataBuffer_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_10DataBuffer_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_10DataBuffer_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2qp_10DataBuffer_2__reduce_cython__(((struct __pyx_obj_7pyverbs_2qp_DataBuffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_10DataBuffer_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_DataBuffer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.data cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.data cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_data_cannot_be_converted_to, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.data cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.DataBuffer.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.data cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.data cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_10DataBuffer_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_10DataBuffer_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_10DataBuffer_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_10DataBuffer_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.DataBuffer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_10DataBuffer_4__setstate_cython__(((struct __pyx_obj_7pyverbs_2qp_DataBuffer *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_10DataBuffer_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_DataBuffer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.data cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.data cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_data_cannot_be_converted_to, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.data cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.data cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.DataBuffer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1288
 * 
 * cdef class QPEx(QP):
 *     def __init__(self, object creator not None, object init_attr not None,             # <<<<<<<<<<<<<<
 *                  QPAttr qp_attr=None):
 *         """
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_4QPEx_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2qp_4QPEx___init__, "\n        Initializes a QPEx object. Since this is an extension of a QP, QP\n        creation is done in the parent class. The extended QP is retrieved by\n        casting the ibv_qp to ibv_qp_ex.\n        :return: An initialized QPEx object\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_7pyverbs_2qp_4QPEx___init__;
#endif
static int __pyx_pw_7pyverbs_2qp_4QPEx_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_creator = 0;
  PyObject *__pyx_v_init_attr = 0;
  struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_qp_attr = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_creator,&__pyx_mstate_global->__pyx_n_u_init_attr,&__pyx_mstate_global->__pyx_n_u_qp_attr,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1288, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1288, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1288, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1288, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 1288, __pyx_L3_error)

      /* "pyverbs/qp.pyx":1289
 * cdef class QPEx(QP):
 *     def __init__(self, object creator not None, object init_attr not None,
 *                  QPAttr qp_attr=None):             # <<<<<<<<<<<<<<
 *         """
 *         Initializes a QPEx object. Since this is an extension of a QP, QP
*/
      if (!values[2]) values[2] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2qp_QPAttr *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, i); __PYX_ERR(0, 1288, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1288, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1288, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1288, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[2]) values[2] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2qp_QPAttr *)Py_None));
    }
    __pyx_v_creator = values[0];
    __pyx_v_init_attr = values[1];
    __pyx_v_qp_attr = ((struct __pyx_obj_7pyverbs_2qp_QPAttr *)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 3, __pyx_nargs); __PYX_ERR(0, 1288, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPEx.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(((PyObject *)__pyx_v_creator) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "creator"); __PYX_ERR(0, 1288, __pyx_L1_error)
  }
  if (unlikely(((PyObject *)__pyx_v_init_attr) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "init_attr"); __PYX_ERR(0, 1288, __pyx_L1_error)
  }
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_qp_attr), __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPAttr, 1, "qp_attr", 0))) __PYX_ERR(0, 1289, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx___init__(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self), __pyx_v_creator, __pyx_v_init_attr, __pyx_v_qp_attr);

  /* "pyverbs/qp.pyx":1288
 * 
 * cdef class QPEx(QP):
 *     def __init__(self, object creator not None, object init_attr not None,             # <<<<<<<<<<<<<<
 *                  QPAttr qp_attr=None):
 *         """
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_4QPEx___init__(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_creator, PyObject *__pyx_v_init_attr, struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_qp_attr) {
  struct __pyx_obj_7pyverbs_2wq_RwqIndTable *__pyx_v_ind_table = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyverbs/qp.pyx":1296
 *         :return: An initialized QPEx object
 *         """
 *         super().__init__(creator, init_attr, qp_attr)             # <<<<<<<<<<<<<<
 *         if init_attr.ind_table is not None:
 *             ind_table = <RwqIndTable>init_attr.ind_table
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1296, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_2, __pyx_v_creator, __pyx_v_init_attr, ((PyObject *)__pyx_v_qp_attr)};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_5, (4-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1296, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/qp.pyx":1297
 *         """
 *         super().__init__(creator, init_attr, qp_attr)
 *         if init_attr.ind_table is not None:             # <<<<<<<<<<<<<<
 *             ind_table = <RwqIndTable>init_attr.ind_table
 *             ind_table.add_ref(self)
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_init_attr, __pyx_mstate_global->__pyx_n_u_ind_table); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = (__pyx_t_1 != Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "pyverbs/qp.pyx":1298
 *         super().__init__(creator, init_attr, qp_attr)
 *         if init_attr.ind_table is not None:
 *             ind_table = <RwqIndTable>init_attr.ind_table             # <<<<<<<<<<<<<<
 *             ind_table.add_ref(self)
 *             self.ind_table = ind_table
*/
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_init_attr, __pyx_mstate_global->__pyx_n_u_ind_table); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __pyx_t_1;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_ind_table = ((struct __pyx_obj_7pyverbs_2wq_RwqIndTable *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyverbs/qp.pyx":1299
 *         if init_attr.ind_table is not None:
 *             ind_table = <RwqIndTable>init_attr.ind_table
 *             ind_table.add_ref(self)             # <<<<<<<<<<<<<<
 *             self.ind_table = ind_table
 *         if init_attr.comp_mask & v.IBV_QP_INIT_ATTR_SEND_OPS_FLAGS:
*/
    __pyx_t_3 = ((struct __pyx_vtabstruct_7pyverbs_2wq_RwqIndTable *)__pyx_v_ind_table->__pyx_base.__pyx_vtab)->add_ref(__pyx_v_ind_table, ((PyObject *)__pyx_v_self), 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyverbs/qp.pyx":1300
 *             ind_table = <RwqIndTable>init_attr.ind_table
 *             ind_table.add_ref(self)
 *             self.ind_table = ind_table             # <<<<<<<<<<<<<<
 *         if init_attr.comp_mask & v.IBV_QP_INIT_ATTR_SEND_OPS_FLAGS:
 *             self.qp_ex = v.ibv_qp_to_qp_ex(self.qp)
*/
    __Pyx_INCREF((PyObject *)__pyx_v_ind_table);
    __Pyx_GIVEREF((PyObject *)__pyx_v_ind_table);
    __Pyx_GOTREF(__pyx_v_self->ind_table);
    __Pyx_DECREF(__pyx_v_self->ind_table);
    __pyx_v_self->ind_table = ((PyObject *)__pyx_v_ind_table);

    /* "pyverbs/qp.pyx":1297
 *         """
 *         super().__init__(creator, init_attr, qp_attr)
 *         if init_attr.ind_table is not None:             # <<<<<<<<<<<<<<
 *             ind_table = <RwqIndTable>init_attr.ind_table
 *             ind_table.add_ref(self)
*/
  }

  /* "pyverbs/qp.pyx":1301
 *             ind_table.add_ref(self)
 *             self.ind_table = ind_table
 *         if init_attr.comp_mask & v.IBV_QP_INIT_ATTR_SEND_OPS_FLAGS:             # <<<<<<<<<<<<<<
 *             self.qp_ex = v.ibv_qp_to_qp_ex(self.qp)
 *             if self.qp_ex == NULL:
*/
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_init_attr, __pyx_mstate_global->__pyx_n_u_comp_mask); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py(IBV_QP_INIT_ATTR_SEND_OPS_FLAGS); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_And(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 1301, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_6) {

    /* "pyverbs/qp.pyx":1302
 *             self.ind_table = ind_table
 *         if init_attr.comp_mask & v.IBV_QP_INIT_ATTR_SEND_OPS_FLAGS:
 *             self.qp_ex = v.ibv_qp_to_qp_ex(self.qp)             # <<<<<<<<<<<<<<
 *             if self.qp_ex == NULL:
 *                 raise PyverbsRDMAErrno('Failed to create extended QP')
*/
    __pyx_v_self->qp_ex = ibv_qp_to_qp_ex(__pyx_v_self->__pyx_base.qp);

    /* "pyverbs/qp.pyx":1303
 *         if init_attr.comp_mask & v.IBV_QP_INIT_ATTR_SEND_OPS_FLAGS:
 *             self.qp_ex = v.ibv_qp_to_qp_ex(self.qp)
 *             if self.qp_ex == NULL:             # <<<<<<<<<<<<<<
 *                 raise PyverbsRDMAErrno('Failed to create extended QP')
 *         else:
*/
    __pyx_t_6 = (__pyx_v_self->qp_ex == NULL);
    if (unlikely(__pyx_t_6)) {

      /* "pyverbs/qp.pyx":1304
 *             self.qp_ex = v.ibv_qp_to_qp_ex(self.qp)
 *             if self.qp_ex == NULL:
 *                 raise PyverbsRDMAErrno('Failed to create extended QP')             # <<<<<<<<<<<<<<
 *         else:
 *             self.logger.debug('qp_ex is not accessible since IBV_QP_INIT_ATTR_SEND_OPS_FLAGS was not passed.')
*/
      __pyx_t_1 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1304, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        assert(__pyx_t_1);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
        __pyx_t_5 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_mstate_global->__pyx_kp_u_Failed_to_create_extended_QP};
        __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1304, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 1304, __pyx_L1_error)

      /* "pyverbs/qp.pyx":1303
 *         if init_attr.comp_mask & v.IBV_QP_INIT_ATTR_SEND_OPS_FLAGS:
 *             self.qp_ex = v.ibv_qp_to_qp_ex(self.qp)
 *             if self.qp_ex == NULL:             # <<<<<<<<<<<<<<
 *                 raise PyverbsRDMAErrno('Failed to create extended QP')
 *         else:
*/
    }

    /* "pyverbs/qp.pyx":1301
 *             ind_table.add_ref(self)
 *             self.ind_table = ind_table
 *         if init_attr.comp_mask & v.IBV_QP_INIT_ATTR_SEND_OPS_FLAGS:             # <<<<<<<<<<<<<<
 *             self.qp_ex = v.ibv_qp_to_qp_ex(self.qp)
 *             if self.qp_ex == NULL:
*/
    goto __pyx_L4;
  }

  /* "pyverbs/qp.pyx":1306
 *                 raise PyverbsRDMAErrno('Failed to create extended QP')
 *         else:
 *             self.logger.debug('qp_ex is not accessible since IBV_QP_INIT_ATTR_SEND_OPS_FLAGS was not passed.')             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  /*else*/ {
    __pyx_t_3 = __pyx_v_self->__pyx_base.__pyx_base.__pyx_base.logger;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_qp_ex_is_not_accessible_since_IB};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_debug, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L4:;

  /* "pyverbs/qp.pyx":1288
 * 
 * cdef class QPEx(QP):
 *     def __init__(self, object creator not None, object init_attr not None,             # <<<<<<<<<<<<<<
 *                  QPAttr qp_attr=None):
 *         """
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.qp.QPEx.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_ind_table);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1308
 *             self.logger.debug('qp_ex is not accessible since IBV_QP_INIT_ATTR_SEND_OPS_FLAGS was not passed.')
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def comp_mask(self):
 *         return self.qp_ex.comp_mask
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_9comp_mask_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_9comp_mask_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_9comp_mask___get__(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_9comp_mask___get__(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":1310
 *     @property
 *     def comp_mask(self):
 *         return self.qp_ex.comp_mask             # <<<<<<<<<<<<<<
 *     @comp_mask.setter
 *     def comp_mask(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_uint64_t(__pyx_v_self->qp_ex->comp_mask); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":1308
 *             self.logger.debug('qp_ex is not accessible since IBV_QP_INIT_ATTR_SEND_OPS_FLAGS was not passed.')
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def comp_mask(self):
 *         return self.qp_ex.comp_mask
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPEx.comp_mask.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1311
 *     def comp_mask(self):
 *         return self.qp_ex.comp_mask
 *     @comp_mask.setter             # <<<<<<<<<<<<<<
 *     def comp_mask(self, val):
 *         self.qp_ex.comp_mask = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_4QPEx_9comp_mask_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_4QPEx_9comp_mask_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_9comp_mask_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_4QPEx_9comp_mask_2__set__(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  uint64_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":1313
 *     @comp_mask.setter
 *     def comp_mask(self, val):
 *         self.qp_ex.comp_mask = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_uint64_t(__pyx_v_val); if (unlikely((__pyx_t_1 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1313, __pyx_L1_error)
  __pyx_v_self->qp_ex->comp_mask = __pyx_t_1;

  /* "pyverbs/qp.pyx":1311
 *     def comp_mask(self):
 *         return self.qp_ex.comp_mask
 *     @comp_mask.setter             # <<<<<<<<<<<<<<
 *     def comp_mask(self, val):
 *         self.qp_ex.comp_mask = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPEx.comp_mask.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1315
 *         self.qp_ex.comp_mask = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def wr_id(self):
 *         return self.qp_ex.wr_id
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_5wr_id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_5wr_id_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_5wr_id___get__(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_5wr_id___get__(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":1317
 *     @property
 *     def wr_id(self):
 *         return self.qp_ex.wr_id             # <<<<<<<<<<<<<<
 *     @wr_id.setter
 *     def wr_id(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_uint64_t(__pyx_v_self->qp_ex->wr_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":1315
 *         self.qp_ex.comp_mask = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def wr_id(self):
 *         return self.qp_ex.wr_id
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_id.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1318
 *     def wr_id(self):
 *         return self.qp_ex.wr_id
 *     @wr_id.setter             # <<<<<<<<<<<<<<
 *     def wr_id(self, val):
 *         self.qp_ex.wr_id = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_4QPEx_5wr_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_4QPEx_5wr_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_5wr_id_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_4QPEx_5wr_id_2__set__(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  uint64_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":1320
 *     @wr_id.setter
 *     def wr_id(self, val):
 *         self.qp_ex.wr_id = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_uint64_t(__pyx_v_val); if (unlikely((__pyx_t_1 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1320, __pyx_L1_error)
  __pyx_v_self->qp_ex->wr_id = __pyx_t_1;

  /* "pyverbs/qp.pyx":1318
 *     def wr_id(self):
 *         return self.qp_ex.wr_id
 *     @wr_id.setter             # <<<<<<<<<<<<<<
 *     def wr_id(self, val):
 *         self.qp_ex.wr_id = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_id.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1322
 *         self.qp_ex.wr_id = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def wr_flags(self):
 *         return self.qp_ex.wr_flags
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_8wr_flags_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_8wr_flags_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_8wr_flags___get__(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_8wr_flags___get__(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":1324
 *     @property
 *     def wr_flags(self):
 *         return self.qp_ex.wr_flags             # <<<<<<<<<<<<<<
 *     @wr_flags.setter
 *     def wr_flags(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->qp_ex->wr_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":1322
 *         self.qp_ex.wr_id = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def wr_flags(self):
 *         return self.qp_ex.wr_flags
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_flags.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1325
 *     def wr_flags(self):
 *         return self.qp_ex.wr_flags
 *     @wr_flags.setter             # <<<<<<<<<<<<<<
 *     def wr_flags(self, val):
 *         self.qp_ex.wr_flags = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2qp_4QPEx_8wr_flags_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2qp_4QPEx_8wr_flags_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_8wr_flags_2__set__(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2qp_4QPEx_8wr_flags_2__set__(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/qp.pyx":1327
 *     @wr_flags.setter
 *     def wr_flags(self, val):
 *         self.qp_ex.wr_flags = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1327, __pyx_L1_error)
  __pyx_v_self->qp_ex->wr_flags = __pyx_t_1;

  /* "pyverbs/qp.pyx":1325
 *     def wr_flags(self):
 *         return self.qp_ex.wr_flags
 *     @wr_flags.setter             # <<<<<<<<<<<<<<
 *     def wr_flags(self, val):
 *         self.qp_ex.wr_flags = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_flags.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1329
 *         self.qp_ex.wr_flags = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ind_table(self):
 *         return self.ind_table
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_9ind_table_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_9ind_table_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_9ind_table___get__(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_9ind_table___get__(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/qp.pyx":1331
 *     @property
 *     def ind_table(self):
 *         return self.ind_table             # <<<<<<<<<<<<<<
 * 
 *     def wr_atomic_cmp_swp(self, rkey, remote_addr, compare, swap):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->ind_table);
  __pyx_r = __pyx_v_self->ind_table;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":1329
 *         self.qp_ex.wr_flags = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ind_table(self):
 *         return self.ind_table
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1333
 *         return self.ind_table
 * 
 *     def wr_atomic_cmp_swp(self, rkey, remote_addr, compare, swap):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_atomic_cmp_swp(self.qp_ex, rkey, remote_addr, compare, swap)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_3wr_atomic_cmp_swp(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_3wr_atomic_cmp_swp = {"wr_atomic_cmp_swp", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_3wr_atomic_cmp_swp, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_3wr_atomic_cmp_swp(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_rkey = 0;
  PyObject *__pyx_v_remote_addr = 0;
  PyObject *__pyx_v_compare = 0;
  PyObject *__pyx_v_swap = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_atomic_cmp_swp (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_rkey,&__pyx_mstate_global->__pyx_n_u_remote_addr,&__pyx_mstate_global->__pyx_n_u_compare,&__pyx_mstate_global->__pyx_n_u_swap,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1333, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1333, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1333, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1333, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1333, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "wr_atomic_cmp_swp", 0) < (0)) __PYX_ERR(0, 1333, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 4; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("wr_atomic_cmp_swp", 1, 4, 4, i); __PYX_ERR(0, 1333, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 4)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1333, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1333, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1333, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1333, __pyx_L3_error)
    }
    __pyx_v_rkey = values[0];
    __pyx_v_remote_addr = values[1];
    __pyx_v_compare = values[2];
    __pyx_v_swap = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wr_atomic_cmp_swp", 1, 4, 4, __pyx_nargs); __PYX_ERR(0, 1333, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_atomic_cmp_swp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_2wr_atomic_cmp_swp(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self), __pyx_v_rkey, __pyx_v_remote_addr, __pyx_v_compare, __pyx_v_swap);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_2wr_atomic_cmp_swp(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_rkey, PyObject *__pyx_v_remote_addr, PyObject *__pyx_v_compare, PyObject *__pyx_v_swap) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  uint64_t __pyx_t_2;
  uint64_t __pyx_t_3;
  uint64_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wr_atomic_cmp_swp", 0);

  /* "pyverbs/qp.pyx":1334
 * 
 *     def wr_atomic_cmp_swp(self, rkey, remote_addr, compare, swap):
 *         v.ibv_wr_atomic_cmp_swp(self.qp_ex, rkey, remote_addr, compare, swap)             # <<<<<<<<<<<<<<
 * 
 *     def wr_atomic_fetch_add(self, rkey, remote_addr, add):
*/
  __pyx_t_1 = __Pyx_PyLong_As_uint32_t(__pyx_v_rkey); if (unlikely((__pyx_t_1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1334, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_As_uint64_t(__pyx_v_remote_addr); if (unlikely((__pyx_t_2 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1334, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyLong_As_uint64_t(__pyx_v_compare); if (unlikely((__pyx_t_3 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1334, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyLong_As_uint64_t(__pyx_v_swap); if (unlikely((__pyx_t_4 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1334, __pyx_L1_error)
  ibv_wr_atomic_cmp_swp(__pyx_v_self->qp_ex, __pyx_t_1, __pyx_t_2, __pyx_t_3, __pyx_t_4);

  /* "pyverbs/qp.pyx":1333
 *         return self.ind_table
 * 
 *     def wr_atomic_cmp_swp(self, rkey, remote_addr, compare, swap):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_atomic_cmp_swp(self.qp_ex, rkey, remote_addr, compare, swap)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_atomic_cmp_swp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1336
 *         v.ibv_wr_atomic_cmp_swp(self.qp_ex, rkey, remote_addr, compare, swap)
 * 
 *     def wr_atomic_fetch_add(self, rkey, remote_addr, add):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_atomic_fetch_add(self.qp_ex, rkey, remote_addr, add)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_5wr_atomic_fetch_add(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_5wr_atomic_fetch_add = {"wr_atomic_fetch_add", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_5wr_atomic_fetch_add, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_5wr_atomic_fetch_add(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_rkey = 0;
  PyObject *__pyx_v_remote_addr = 0;
  PyObject *__pyx_v_add = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_atomic_fetch_add (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_rkey,&__pyx_mstate_global->__pyx_n_u_remote_addr,&__pyx_mstate_global->__pyx_n_u_add,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1336, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1336, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1336, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1336, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "wr_atomic_fetch_add", 0) < (0)) __PYX_ERR(0, 1336, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("wr_atomic_fetch_add", 1, 3, 3, i); __PYX_ERR(0, 1336, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1336, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1336, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1336, __pyx_L3_error)
    }
    __pyx_v_rkey = values[0];
    __pyx_v_remote_addr = values[1];
    __pyx_v_add = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wr_atomic_fetch_add", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1336, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_atomic_fetch_add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_4wr_atomic_fetch_add(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self), __pyx_v_rkey, __pyx_v_remote_addr, __pyx_v_add);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_4wr_atomic_fetch_add(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_rkey, PyObject *__pyx_v_remote_addr, PyObject *__pyx_v_add) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  uint64_t __pyx_t_2;
  uint64_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wr_atomic_fetch_add", 0);

  /* "pyverbs/qp.pyx":1337
 * 
 *     def wr_atomic_fetch_add(self, rkey, remote_addr, add):
 *         v.ibv_wr_atomic_fetch_add(self.qp_ex, rkey, remote_addr, add)             # <<<<<<<<<<<<<<
 * 
 *     def wr_bind_mw(self, MW mw, rkey, MWBindInfo bind_info):
*/
  __pyx_t_1 = __Pyx_PyLong_As_uint32_t(__pyx_v_rkey); if (unlikely((__pyx_t_1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1337, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_As_uint64_t(__pyx_v_remote_addr); if (unlikely((__pyx_t_2 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1337, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyLong_As_uint64_t(__pyx_v_add); if (unlikely((__pyx_t_3 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1337, __pyx_L1_error)
  ibv_wr_atomic_fetch_add(__pyx_v_self->qp_ex, __pyx_t_1, __pyx_t_2, __pyx_t_3);

  /* "pyverbs/qp.pyx":1336
 *         v.ibv_wr_atomic_cmp_swp(self.qp_ex, rkey, remote_addr, compare, swap)
 * 
 *     def wr_atomic_fetch_add(self, rkey, remote_addr, add):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_atomic_fetch_add(self.qp_ex, rkey, remote_addr, add)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_atomic_fetch_add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1339
 *         v.ibv_wr_atomic_fetch_add(self.qp_ex, rkey, remote_addr, add)
 * 
 *     def wr_bind_mw(self, MW mw, rkey, MWBindInfo bind_info):             # <<<<<<<<<<<<<<
 *         cdef v.ibv_mw_bind_info *info
 *         info = &bind_info.info
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_7wr_bind_mw(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_7wr_bind_mw = {"wr_bind_mw", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_7wr_bind_mw, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_7wr_bind_mw(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_2mr_MW *__pyx_v_mw = 0;
  PyObject *__pyx_v_rkey = 0;
  struct __pyx_obj_7pyverbs_2mr_MWBindInfo *__pyx_v_bind_info = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_bind_mw (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_mw,&__pyx_mstate_global->__pyx_n_u_rkey,&__pyx_mstate_global->__pyx_n_u_bind_info,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1339, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1339, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1339, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1339, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "wr_bind_mw", 0) < (0)) __PYX_ERR(0, 1339, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("wr_bind_mw", 1, 3, 3, i); __PYX_ERR(0, 1339, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1339, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1339, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1339, __pyx_L3_error)
    }
    __pyx_v_mw = ((struct __pyx_obj_7pyverbs_2mr_MW *)values[0]);
    __pyx_v_rkey = values[1];
    __pyx_v_bind_info = ((struct __pyx_obj_7pyverbs_2mr_MWBindInfo *)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wr_bind_mw", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1339, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_bind_mw", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mw), __pyx_mstate_global->__pyx_ptype_7pyverbs_2mr_MW, 1, "mw", 0))) __PYX_ERR(0, 1339, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_bind_info), __pyx_mstate_global->__pyx_ptype_7pyverbs_2mr_MWBindInfo, 1, "bind_info", 0))) __PYX_ERR(0, 1339, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_6wr_bind_mw(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self), __pyx_v_mw, __pyx_v_rkey, __pyx_v_bind_info);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_6wr_bind_mw(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, struct __pyx_obj_7pyverbs_2mr_MW *__pyx_v_mw, PyObject *__pyx_v_rkey, struct __pyx_obj_7pyverbs_2mr_MWBindInfo *__pyx_v_bind_info) {
  struct ibv_mw_bind_info *__pyx_v_info;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wr_bind_mw", 0);

  /* "pyverbs/qp.pyx":1341
 *     def wr_bind_mw(self, MW mw, rkey, MWBindInfo bind_info):
 *         cdef v.ibv_mw_bind_info *info
 *         info = &bind_info.info             # <<<<<<<<<<<<<<
 *         v.ibv_wr_bind_mw(self.qp_ex, <v.ibv_mw*>mw.mw, rkey,
 *                          <v.ibv_mw_bind_info*>info)
*/
  __pyx_v_info = (&__pyx_v_bind_info->info);

  /* "pyverbs/qp.pyx":1342
 *         cdef v.ibv_mw_bind_info *info
 *         info = &bind_info.info
 *         v.ibv_wr_bind_mw(self.qp_ex, <v.ibv_mw*>mw.mw, rkey,             # <<<<<<<<<<<<<<
 *                          <v.ibv_mw_bind_info*>info)
 *         self.add_ref(mw)
*/
  __pyx_t_1 = __Pyx_PyLong_As_uint32_t(__pyx_v_rkey); if (unlikely((__pyx_t_1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1342, __pyx_L1_error)

  /* "pyverbs/qp.pyx":1343
 *         info = &bind_info.info
 *         v.ibv_wr_bind_mw(self.qp_ex, <v.ibv_mw*>mw.mw, rkey,
 *                          <v.ibv_mw_bind_info*>info)             # <<<<<<<<<<<<<<
 *         self.add_ref(mw)
 * 
*/
  ibv_wr_bind_mw(__pyx_v_self->qp_ex, ((struct ibv_mw *)__pyx_v_mw->mw), __pyx_t_1, ((struct ibv_mw_bind_info *)__pyx_v_info));

  /* "pyverbs/qp.pyx":1344
 *         v.ibv_wr_bind_mw(self.qp_ex, <v.ibv_mw*>mw.mw, rkey,
 *                          <v.ibv_mw_bind_info*>info)
 *         self.add_ref(mw)             # <<<<<<<<<<<<<<
 * 
 *     def wr_local_inv(self, invalidate_rkey):
*/
  __pyx_t_2 = ((struct __pyx_vtabstruct_7pyverbs_2qp_QPEx *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.add_ref(((struct __pyx_obj_7pyverbs_2qp_QP *)__pyx_v_self), ((PyObject *)__pyx_v_mw)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1339
 *         v.ibv_wr_atomic_fetch_add(self.qp_ex, rkey, remote_addr, add)
 * 
 *     def wr_bind_mw(self, MW mw, rkey, MWBindInfo bind_info):             # <<<<<<<<<<<<<<
 *         cdef v.ibv_mw_bind_info *info
 *         info = &bind_info.info
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_bind_mw", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1346
 *         self.add_ref(mw)
 * 
 *     def wr_local_inv(self, invalidate_rkey):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_local_inv(self.qp_ex, invalidate_rkey)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_9wr_local_inv(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_9wr_local_inv = {"wr_local_inv", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_9wr_local_inv, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_9wr_local_inv(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_invalidate_rkey = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_local_inv (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_invalidate_rkey,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1346, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1346, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "wr_local_inv", 0) < (0)) __PYX_ERR(0, 1346, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("wr_local_inv", 1, 1, 1, i); __PYX_ERR(0, 1346, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1346, __pyx_L3_error)
    }
    __pyx_v_invalidate_rkey = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wr_local_inv", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1346, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_local_inv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_8wr_local_inv(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self), __pyx_v_invalidate_rkey);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_8wr_local_inv(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_invalidate_rkey) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wr_local_inv", 0);

  /* "pyverbs/qp.pyx":1347
 * 
 *     def wr_local_inv(self, invalidate_rkey):
 *         v.ibv_wr_local_inv(self.qp_ex, invalidate_rkey)             # <<<<<<<<<<<<<<
 * 
 *     def wr_atomic_write(self, rkey, remote_addr, atomic_wr):
*/
  __pyx_t_1 = __Pyx_PyLong_As_uint32_t(__pyx_v_invalidate_rkey); if (unlikely((__pyx_t_1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1347, __pyx_L1_error)
  ibv_wr_local_inv(__pyx_v_self->qp_ex, __pyx_t_1);

  /* "pyverbs/qp.pyx":1346
 *         self.add_ref(mw)
 * 
 *     def wr_local_inv(self, invalidate_rkey):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_local_inv(self.qp_ex, invalidate_rkey)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_local_inv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1349
 *         v.ibv_wr_local_inv(self.qp_ex, invalidate_rkey)
 * 
 *     def wr_atomic_write(self, rkey, remote_addr, atomic_wr):             # <<<<<<<<<<<<<<
 *         cdef char *atomic_wr_c = atomic_wr
 *         v.ibv_wr_atomic_write(self.qp_ex, rkey, remote_addr, atomic_wr_c)
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_11wr_atomic_write(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_11wr_atomic_write = {"wr_atomic_write", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_11wr_atomic_write, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_11wr_atomic_write(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_rkey = 0;
  PyObject *__pyx_v_remote_addr = 0;
  PyObject *__pyx_v_atomic_wr = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_atomic_write (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_rkey,&__pyx_mstate_global->__pyx_n_u_remote_addr,&__pyx_mstate_global->__pyx_n_u_atomic_wr,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1349, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1349, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1349, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1349, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "wr_atomic_write", 0) < (0)) __PYX_ERR(0, 1349, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("wr_atomic_write", 1, 3, 3, i); __PYX_ERR(0, 1349, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1349, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1349, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1349, __pyx_L3_error)
    }
    __pyx_v_rkey = values[0];
    __pyx_v_remote_addr = values[1];
    __pyx_v_atomic_wr = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wr_atomic_write", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1349, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_atomic_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_10wr_atomic_write(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self), __pyx_v_rkey, __pyx_v_remote_addr, __pyx_v_atomic_wr);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_10wr_atomic_write(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_rkey, PyObject *__pyx_v_remote_addr, PyObject *__pyx_v_atomic_wr) {
  char *__pyx_v_atomic_wr_c;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  uint32_t __pyx_t_2;
  uint64_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wr_atomic_write", 0);

  /* "pyverbs/qp.pyx":1350
 * 
 *     def wr_atomic_write(self, rkey, remote_addr, atomic_wr):
 *         cdef char *atomic_wr_c = atomic_wr             # <<<<<<<<<<<<<<
 *         v.ibv_wr_atomic_write(self.qp_ex, rkey, remote_addr, atomic_wr_c)
 * 
*/
  __pyx_t_1 = __Pyx_PyObject_AsWritableString(__pyx_v_atomic_wr); if (unlikely((!__pyx_t_1) && PyErr_Occurred())) __PYX_ERR(0, 1350, __pyx_L1_error)
  __pyx_v_atomic_wr_c = __pyx_t_1;

  /* "pyverbs/qp.pyx":1351
 *     def wr_atomic_write(self, rkey, remote_addr, atomic_wr):
 *         cdef char *atomic_wr_c = atomic_wr
 *         v.ibv_wr_atomic_write(self.qp_ex, rkey, remote_addr, atomic_wr_c)             # <<<<<<<<<<<<<<
 * 
 *     def wr_rdma_read(self, rkey, remote_addr):
*/
  __pyx_t_2 = __Pyx_PyLong_As_uint32_t(__pyx_v_rkey); if (unlikely((__pyx_t_2 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1351, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyLong_As_uint64_t(__pyx_v_remote_addr); if (unlikely((__pyx_t_3 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1351, __pyx_L1_error)
  ibv_wr_atomic_write(__pyx_v_self->qp_ex, __pyx_t_2, __pyx_t_3, __pyx_v_atomic_wr_c);

  /* "pyverbs/qp.pyx":1349
 *         v.ibv_wr_local_inv(self.qp_ex, invalidate_rkey)
 * 
 *     def wr_atomic_write(self, rkey, remote_addr, atomic_wr):             # <<<<<<<<<<<<<<
 *         cdef char *atomic_wr_c = atomic_wr
 *         v.ibv_wr_atomic_write(self.qp_ex, rkey, remote_addr, atomic_wr_c)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_atomic_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1353
 *         v.ibv_wr_atomic_write(self.qp_ex, rkey, remote_addr, atomic_wr_c)
 * 
 *     def wr_rdma_read(self, rkey, remote_addr):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_rdma_read(self.qp_ex, rkey, remote_addr)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_13wr_rdma_read(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_13wr_rdma_read = {"wr_rdma_read", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_13wr_rdma_read, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_13wr_rdma_read(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_rkey = 0;
  PyObject *__pyx_v_remote_addr = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_rdma_read (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_rkey,&__pyx_mstate_global->__pyx_n_u_remote_addr,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1353, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1353, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1353, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "wr_rdma_read", 0) < (0)) __PYX_ERR(0, 1353, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("wr_rdma_read", 1, 2, 2, i); __PYX_ERR(0, 1353, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1353, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1353, __pyx_L3_error)
    }
    __pyx_v_rkey = values[0];
    __pyx_v_remote_addr = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wr_rdma_read", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1353, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_rdma_read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_12wr_rdma_read(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self), __pyx_v_rkey, __pyx_v_remote_addr);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_12wr_rdma_read(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_rkey, PyObject *__pyx_v_remote_addr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  uint64_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wr_rdma_read", 0);

  /* "pyverbs/qp.pyx":1354
 * 
 *     def wr_rdma_read(self, rkey, remote_addr):
 *         v.ibv_wr_rdma_read(self.qp_ex, rkey, remote_addr)             # <<<<<<<<<<<<<<
 * 
 *     def wr_rdma_write(self, rkey, remote_addr):
*/
  __pyx_t_1 = __Pyx_PyLong_As_uint32_t(__pyx_v_rkey); if (unlikely((__pyx_t_1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1354, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_As_uint64_t(__pyx_v_remote_addr); if (unlikely((__pyx_t_2 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1354, __pyx_L1_error)
  ibv_wr_rdma_read(__pyx_v_self->qp_ex, __pyx_t_1, __pyx_t_2);

  /* "pyverbs/qp.pyx":1353
 *         v.ibv_wr_atomic_write(self.qp_ex, rkey, remote_addr, atomic_wr_c)
 * 
 *     def wr_rdma_read(self, rkey, remote_addr):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_rdma_read(self.qp_ex, rkey, remote_addr)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_rdma_read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1356
 *         v.ibv_wr_rdma_read(self.qp_ex, rkey, remote_addr)
 * 
 *     def wr_rdma_write(self, rkey, remote_addr):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_rdma_write(self.qp_ex, rkey, remote_addr)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_15wr_rdma_write(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_15wr_rdma_write = {"wr_rdma_write", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_15wr_rdma_write, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_15wr_rdma_write(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_rkey = 0;
  PyObject *__pyx_v_remote_addr = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_rdma_write (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_rkey,&__pyx_mstate_global->__pyx_n_u_remote_addr,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1356, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1356, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1356, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "wr_rdma_write", 0) < (0)) __PYX_ERR(0, 1356, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("wr_rdma_write", 1, 2, 2, i); __PYX_ERR(0, 1356, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1356, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1356, __pyx_L3_error)
    }
    __pyx_v_rkey = values[0];
    __pyx_v_remote_addr = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wr_rdma_write", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1356, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_rdma_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_14wr_rdma_write(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self), __pyx_v_rkey, __pyx_v_remote_addr);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_14wr_rdma_write(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_rkey, PyObject *__pyx_v_remote_addr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  uint64_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wr_rdma_write", 0);

  /* "pyverbs/qp.pyx":1357
 * 
 *     def wr_rdma_write(self, rkey, remote_addr):
 *         v.ibv_wr_rdma_write(self.qp_ex, rkey, remote_addr)             # <<<<<<<<<<<<<<
 * 
 *     def wr_rdma_write_imm(self, rkey, remote_addr, data):
*/
  __pyx_t_1 = __Pyx_PyLong_As_uint32_t(__pyx_v_rkey); if (unlikely((__pyx_t_1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1357, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_As_uint64_t(__pyx_v_remote_addr); if (unlikely((__pyx_t_2 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1357, __pyx_L1_error)
  ibv_wr_rdma_write(__pyx_v_self->qp_ex, __pyx_t_1, __pyx_t_2);

  /* "pyverbs/qp.pyx":1356
 *         v.ibv_wr_rdma_read(self.qp_ex, rkey, remote_addr)
 * 
 *     def wr_rdma_write(self, rkey, remote_addr):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_rdma_write(self.qp_ex, rkey, remote_addr)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_rdma_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1359
 *         v.ibv_wr_rdma_write(self.qp_ex, rkey, remote_addr)
 * 
 *     def wr_rdma_write_imm(self, rkey, remote_addr, data):             # <<<<<<<<<<<<<<
 *         cdef unsigned int imm_data = htobe32(data)
 *         v.ibv_wr_rdma_write_imm(self.qp_ex, rkey, remote_addr, imm_data)
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_17wr_rdma_write_imm(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_17wr_rdma_write_imm = {"wr_rdma_write_imm", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_17wr_rdma_write_imm, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_17wr_rdma_write_imm(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_rkey = 0;
  PyObject *__pyx_v_remote_addr = 0;
  PyObject *__pyx_v_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_rdma_write_imm (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_rkey,&__pyx_mstate_global->__pyx_n_u_remote_addr,&__pyx_mstate_global->__pyx_n_u_data,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1359, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1359, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1359, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1359, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "wr_rdma_write_imm", 0) < (0)) __PYX_ERR(0, 1359, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("wr_rdma_write_imm", 1, 3, 3, i); __PYX_ERR(0, 1359, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1359, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1359, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1359, __pyx_L3_error)
    }
    __pyx_v_rkey = values[0];
    __pyx_v_remote_addr = values[1];
    __pyx_v_data = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wr_rdma_write_imm", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1359, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_rdma_write_imm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_16wr_rdma_write_imm(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self), __pyx_v_rkey, __pyx_v_remote_addr, __pyx_v_data);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_16wr_rdma_write_imm(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_rkey, PyObject *__pyx_v_remote_addr, PyObject *__pyx_v_data) {
  unsigned int __pyx_v_imm_data;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  unsigned long __pyx_t_1;
  uint32_t __pyx_t_2;
  uint64_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wr_rdma_write_imm", 0);

  /* "pyverbs/qp.pyx":1360
 * 
 *     def wr_rdma_write_imm(self, rkey, remote_addr, data):
 *         cdef unsigned int imm_data = htobe32(data)             # <<<<<<<<<<<<<<
 *         v.ibv_wr_rdma_write_imm(self.qp_ex, rkey, remote_addr, imm_data)
 * 
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_long(__pyx_v_data); if (unlikely((__pyx_t_1 == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1360, __pyx_L1_error)
  __pyx_v_imm_data = htobe32(__pyx_t_1);

  /* "pyverbs/qp.pyx":1361
 *     def wr_rdma_write_imm(self, rkey, remote_addr, data):
 *         cdef unsigned int imm_data = htobe32(data)
 *         v.ibv_wr_rdma_write_imm(self.qp_ex, rkey, remote_addr, imm_data)             # <<<<<<<<<<<<<<
 * 
 *     def wr_flush(self, rkey, remote_addr, length, ptype, level):
*/
  __pyx_t_2 = __Pyx_PyLong_As_uint32_t(__pyx_v_rkey); if (unlikely((__pyx_t_2 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1361, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyLong_As_uint64_t(__pyx_v_remote_addr); if (unlikely((__pyx_t_3 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1361, __pyx_L1_error)
  ibv_wr_rdma_write_imm(__pyx_v_self->qp_ex, __pyx_t_2, __pyx_t_3, __pyx_v_imm_data);

  /* "pyverbs/qp.pyx":1359
 *         v.ibv_wr_rdma_write(self.qp_ex, rkey, remote_addr)
 * 
 *     def wr_rdma_write_imm(self, rkey, remote_addr, data):             # <<<<<<<<<<<<<<
 *         cdef unsigned int imm_data = htobe32(data)
 *         v.ibv_wr_rdma_write_imm(self.qp_ex, rkey, remote_addr, imm_data)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_rdma_write_imm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1363
 *         v.ibv_wr_rdma_write_imm(self.qp_ex, rkey, remote_addr, imm_data)
 * 
 *     def wr_flush(self, rkey, remote_addr, length, ptype, level):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_flush(self.qp_ex, rkey, remote_addr, length, ptype, level)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_19wr_flush(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_19wr_flush = {"wr_flush", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_19wr_flush, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_19wr_flush(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_rkey = 0;
  PyObject *__pyx_v_remote_addr = 0;
  PyObject *__pyx_v_length = 0;
  PyObject *__pyx_v_ptype = 0;
  PyObject *__pyx_v_level = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_flush (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_rkey,&__pyx_mstate_global->__pyx_n_u_remote_addr,&__pyx_mstate_global->__pyx_n_u_length,&__pyx_mstate_global->__pyx_n_u_ptype,&__pyx_mstate_global->__pyx_n_u_level,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1363, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1363, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1363, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1363, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1363, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1363, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "wr_flush", 0) < (0)) __PYX_ERR(0, 1363, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 5; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("wr_flush", 1, 5, 5, i); __PYX_ERR(0, 1363, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1363, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1363, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1363, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1363, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1363, __pyx_L3_error)
    }
    __pyx_v_rkey = values[0];
    __pyx_v_remote_addr = values[1];
    __pyx_v_length = values[2];
    __pyx_v_ptype = values[3];
    __pyx_v_level = values[4];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wr_flush", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1363, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_flush", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_18wr_flush(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self), __pyx_v_rkey, __pyx_v_remote_addr, __pyx_v_length, __pyx_v_ptype, __pyx_v_level);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_18wr_flush(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_rkey, PyObject *__pyx_v_remote_addr, PyObject *__pyx_v_length, PyObject *__pyx_v_ptype, PyObject *__pyx_v_level) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  uint64_t __pyx_t_2;
  size_t __pyx_t_3;
  uint8_t __pyx_t_4;
  uint8_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wr_flush", 0);

  /* "pyverbs/qp.pyx":1364
 * 
 *     def wr_flush(self, rkey, remote_addr, length, ptype, level):
 *         v.ibv_wr_flush(self.qp_ex, rkey, remote_addr, length, ptype, level)             # <<<<<<<<<<<<<<
 * 
 *     def wr_send(self):
*/
  __pyx_t_1 = __Pyx_PyLong_As_uint32_t(__pyx_v_rkey); if (unlikely((__pyx_t_1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1364, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_As_uint64_t(__pyx_v_remote_addr); if (unlikely((__pyx_t_2 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1364, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyLong_As_size_t(__pyx_v_length); if (unlikely((__pyx_t_3 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1364, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyLong_As_uint8_t(__pyx_v_ptype); if (unlikely((__pyx_t_4 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1364, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyLong_As_uint8_t(__pyx_v_level); if (unlikely((__pyx_t_5 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1364, __pyx_L1_error)
  ibv_wr_flush(__pyx_v_self->qp_ex, __pyx_t_1, __pyx_t_2, __pyx_t_3, __pyx_t_4, __pyx_t_5);

  /* "pyverbs/qp.pyx":1363
 *         v.ibv_wr_rdma_write_imm(self.qp_ex, rkey, remote_addr, imm_data)
 * 
 *     def wr_flush(self, rkey, remote_addr, length, ptype, level):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_flush(self.qp_ex, rkey, remote_addr, length, ptype, level)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_flush", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1366
 *         v.ibv_wr_flush(self.qp_ex, rkey, remote_addr, length, ptype, level)
 * 
 *     def wr_send(self):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_send(self.qp_ex)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_21wr_send(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_21wr_send = {"wr_send", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_21wr_send, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_21wr_send(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_send (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("wr_send", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("wr_send", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_20wr_send(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_20wr_send(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_send", 0);

  /* "pyverbs/qp.pyx":1367
 * 
 *     def wr_send(self):
 *         v.ibv_wr_send(self.qp_ex)             # <<<<<<<<<<<<<<
 * 
 *     def wr_send_imm(self, data):
*/
  ibv_wr_send(__pyx_v_self->qp_ex);

  /* "pyverbs/qp.pyx":1366
 *         v.ibv_wr_flush(self.qp_ex, rkey, remote_addr, length, ptype, level)
 * 
 *     def wr_send(self):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_send(self.qp_ex)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1369
 *         v.ibv_wr_send(self.qp_ex)
 * 
 *     def wr_send_imm(self, data):             # <<<<<<<<<<<<<<
 *         cdef unsigned int imm_data = htobe32(data)
 *         return v.ibv_wr_send_imm(self.qp_ex, imm_data)
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_23wr_send_imm(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_23wr_send_imm = {"wr_send_imm", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_23wr_send_imm, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_23wr_send_imm(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_send_imm (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_data,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1369, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1369, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "wr_send_imm", 0) < (0)) __PYX_ERR(0, 1369, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("wr_send_imm", 1, 1, 1, i); __PYX_ERR(0, 1369, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1369, __pyx_L3_error)
    }
    __pyx_v_data = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wr_send_imm", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1369, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_send_imm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_22wr_send_imm(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self), __pyx_v_data);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_22wr_send_imm(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_data) {
  unsigned int __pyx_v_imm_data;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  unsigned long __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wr_send_imm", 0);

  /* "pyverbs/qp.pyx":1370
 * 
 *     def wr_send_imm(self, data):
 *         cdef unsigned int imm_data = htobe32(data)             # <<<<<<<<<<<<<<
 *         return v.ibv_wr_send_imm(self.qp_ex, imm_data)
 * 
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_long(__pyx_v_data); if (unlikely((__pyx_t_1 == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1370, __pyx_L1_error)
  __pyx_v_imm_data = htobe32(__pyx_t_1);

  /* "pyverbs/qp.pyx":1371
 *     def wr_send_imm(self, data):
 *         cdef unsigned int imm_data = htobe32(data)
 *         return v.ibv_wr_send_imm(self.qp_ex, imm_data)             # <<<<<<<<<<<<<<
 * 
 *     def wr_send_inv(self, invalidate_rkey):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_void_to_None(ibv_wr_send_imm(__pyx_v_self->qp_ex, __pyx_v_imm_data)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyverbs/qp.pyx":1369
 *         v.ibv_wr_send(self.qp_ex)
 * 
 *     def wr_send_imm(self, data):             # <<<<<<<<<<<<<<
 *         cdef unsigned int imm_data = htobe32(data)
 *         return v.ibv_wr_send_imm(self.qp_ex, imm_data)
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_send_imm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1373
 *         return v.ibv_wr_send_imm(self.qp_ex, imm_data)
 * 
 *     def wr_send_inv(self, invalidate_rkey):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_send_inv(self.qp_ex, invalidate_rkey)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_25wr_send_inv(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_25wr_send_inv = {"wr_send_inv", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_25wr_send_inv, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_25wr_send_inv(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_invalidate_rkey = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_send_inv (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_invalidate_rkey,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1373, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1373, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "wr_send_inv", 0) < (0)) __PYX_ERR(0, 1373, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("wr_send_inv", 1, 1, 1, i); __PYX_ERR(0, 1373, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1373, __pyx_L3_error)
    }
    __pyx_v_invalidate_rkey = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wr_send_inv", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1373, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_send_inv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_24wr_send_inv(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self), __pyx_v_invalidate_rkey);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_24wr_send_inv(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_invalidate_rkey) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wr_send_inv", 0);

  /* "pyverbs/qp.pyx":1374
 * 
 *     def wr_send_inv(self, invalidate_rkey):
 *         v.ibv_wr_send_inv(self.qp_ex, invalidate_rkey)             # <<<<<<<<<<<<<<
 * 
 *     def wr_send_tso(self, hdr, hdr_sz, mss):
*/
  __pyx_t_1 = __Pyx_PyLong_As_uint32_t(__pyx_v_invalidate_rkey); if (unlikely((__pyx_t_1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1374, __pyx_L1_error)
  ibv_wr_send_inv(__pyx_v_self->qp_ex, __pyx_t_1);

  /* "pyverbs/qp.pyx":1373
 *         return v.ibv_wr_send_imm(self.qp_ex, imm_data)
 * 
 *     def wr_send_inv(self, invalidate_rkey):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_send_inv(self.qp_ex, invalidate_rkey)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_send_inv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1376
 *         v.ibv_wr_send_inv(self.qp_ex, invalidate_rkey)
 * 
 *     def wr_send_tso(self, hdr, hdr_sz, mss):             # <<<<<<<<<<<<<<
 *         ptr = PyLong_AsVoidPtr(hdr)
 *         v.ibv_wr_send_tso(self.qp_ex, ptr, hdr_sz, mss)
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_27wr_send_tso(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_27wr_send_tso = {"wr_send_tso", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_27wr_send_tso, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_27wr_send_tso(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_hdr = 0;
  PyObject *__pyx_v_hdr_sz = 0;
  PyObject *__pyx_v_mss = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_send_tso (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_hdr,&__pyx_mstate_global->__pyx_n_u_hdr_sz,&__pyx_mstate_global->__pyx_n_u_mss,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1376, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1376, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1376, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1376, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "wr_send_tso", 0) < (0)) __PYX_ERR(0, 1376, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("wr_send_tso", 1, 3, 3, i); __PYX_ERR(0, 1376, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1376, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1376, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1376, __pyx_L3_error)
    }
    __pyx_v_hdr = values[0];
    __pyx_v_hdr_sz = values[1];
    __pyx_v_mss = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wr_send_tso", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1376, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_send_tso", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_26wr_send_tso(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self), __pyx_v_hdr, __pyx_v_hdr_sz, __pyx_v_mss);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_26wr_send_tso(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_hdr, PyObject *__pyx_v_hdr_sz, PyObject *__pyx_v_mss) {
  void *__pyx_v_ptr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint16_t __pyx_t_1;
  uint16_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wr_send_tso", 0);

  /* "pyverbs/qp.pyx":1377
 * 
 *     def wr_send_tso(self, hdr, hdr_sz, mss):
 *         ptr = PyLong_AsVoidPtr(hdr)             # <<<<<<<<<<<<<<
 *         v.ibv_wr_send_tso(self.qp_ex, ptr, hdr_sz, mss)
 * 
*/
  __pyx_v_ptr = PyLong_AsVoidPtr(__pyx_v_hdr);

  /* "pyverbs/qp.pyx":1378
 *     def wr_send_tso(self, hdr, hdr_sz, mss):
 *         ptr = PyLong_AsVoidPtr(hdr)
 *         v.ibv_wr_send_tso(self.qp_ex, ptr, hdr_sz, mss)             # <<<<<<<<<<<<<<
 * 
 *     def wr_set_ud_addr(self, AH ah, remote_qpn, remote_rkey):
*/
  __pyx_t_1 = __Pyx_PyLong_As_uint16_t(__pyx_v_hdr_sz); if (unlikely((__pyx_t_1 == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1378, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_As_uint16_t(__pyx_v_mss); if (unlikely((__pyx_t_2 == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1378, __pyx_L1_error)
  ibv_wr_send_tso(__pyx_v_self->qp_ex, __pyx_v_ptr, __pyx_t_1, __pyx_t_2);

  /* "pyverbs/qp.pyx":1376
 *         v.ibv_wr_send_inv(self.qp_ex, invalidate_rkey)
 * 
 *     def wr_send_tso(self, hdr, hdr_sz, mss):             # <<<<<<<<<<<<<<
 *         ptr = PyLong_AsVoidPtr(hdr)
 *         v.ibv_wr_send_tso(self.qp_ex, ptr, hdr_sz, mss)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_send_tso", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1380
 *         v.ibv_wr_send_tso(self.qp_ex, ptr, hdr_sz, mss)
 * 
 *     def wr_set_ud_addr(self, AH ah, remote_qpn, remote_rkey):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_set_ud_addr(self.qp_ex, ah.ah, remote_qpn, remote_rkey)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_29wr_set_ud_addr(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_29wr_set_ud_addr = {"wr_set_ud_addr", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_29wr_set_ud_addr, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_29wr_set_ud_addr(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_4addr_AH *__pyx_v_ah = 0;
  PyObject *__pyx_v_remote_qpn = 0;
  PyObject *__pyx_v_remote_rkey = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_set_ud_addr (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_ah,&__pyx_mstate_global->__pyx_n_u_remote_qpn,&__pyx_mstate_global->__pyx_n_u_remote_rkey,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1380, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1380, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1380, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1380, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "wr_set_ud_addr", 0) < (0)) __PYX_ERR(0, 1380, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("wr_set_ud_addr", 1, 3, 3, i); __PYX_ERR(0, 1380, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1380, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1380, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1380, __pyx_L3_error)
    }
    __pyx_v_ah = ((struct __pyx_obj_7pyverbs_4addr_AH *)values[0]);
    __pyx_v_remote_qpn = values[1];
    __pyx_v_remote_rkey = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wr_set_ud_addr", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1380, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_set_ud_addr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_ah), __pyx_mstate_global->__pyx_ptype_7pyverbs_4addr_AH, 1, "ah", 0))) __PYX_ERR(0, 1380, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_28wr_set_ud_addr(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self), __pyx_v_ah, __pyx_v_remote_qpn, __pyx_v_remote_rkey);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_28wr_set_ud_addr(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, struct __pyx_obj_7pyverbs_4addr_AH *__pyx_v_ah, PyObject *__pyx_v_remote_qpn, PyObject *__pyx_v_remote_rkey) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  uint32_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wr_set_ud_addr", 0);

  /* "pyverbs/qp.pyx":1381
 * 
 *     def wr_set_ud_addr(self, AH ah, remote_qpn, remote_rkey):
 *         v.ibv_wr_set_ud_addr(self.qp_ex, ah.ah, remote_qpn, remote_rkey)             # <<<<<<<<<<<<<<
 * 
 *     def wr_set_xrc_srqn(self, remote_srqn):
*/
  __pyx_t_1 = __Pyx_PyLong_As_uint32_t(__pyx_v_remote_qpn); if (unlikely((__pyx_t_1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1381, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_As_uint32_t(__pyx_v_remote_rkey); if (unlikely((__pyx_t_2 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1381, __pyx_L1_error)
  ibv_wr_set_ud_addr(__pyx_v_self->qp_ex, __pyx_v_ah->ah, __pyx_t_1, __pyx_t_2);

  /* "pyverbs/qp.pyx":1380
 *         v.ibv_wr_send_tso(self.qp_ex, ptr, hdr_sz, mss)
 * 
 *     def wr_set_ud_addr(self, AH ah, remote_qpn, remote_rkey):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_set_ud_addr(self.qp_ex, ah.ah, remote_qpn, remote_rkey)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_set_ud_addr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1383
 *         v.ibv_wr_set_ud_addr(self.qp_ex, ah.ah, remote_qpn, remote_rkey)
 * 
 *     def wr_set_xrc_srqn(self, remote_srqn):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_set_xrc_srqn(self.qp_ex, remote_srqn)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_31wr_set_xrc_srqn(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_31wr_set_xrc_srqn = {"wr_set_xrc_srqn", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_31wr_set_xrc_srqn, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_31wr_set_xrc_srqn(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_remote_srqn = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_set_xrc_srqn (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_remote_srqn,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1383, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1383, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "wr_set_xrc_srqn", 0) < (0)) __PYX_ERR(0, 1383, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("wr_set_xrc_srqn", 1, 1, 1, i); __PYX_ERR(0, 1383, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1383, __pyx_L3_error)
    }
    __pyx_v_remote_srqn = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wr_set_xrc_srqn", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1383, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_set_xrc_srqn", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_30wr_set_xrc_srqn(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self), __pyx_v_remote_srqn);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_30wr_set_xrc_srqn(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_remote_srqn) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wr_set_xrc_srqn", 0);

  /* "pyverbs/qp.pyx":1384
 * 
 *     def wr_set_xrc_srqn(self, remote_srqn):
 *         v.ibv_wr_set_xrc_srqn(self.qp_ex, remote_srqn)             # <<<<<<<<<<<<<<
 * 
 *     def wr_set_inline_data(self, addr, length):
*/
  __pyx_t_1 = __Pyx_PyLong_As_uint32_t(__pyx_v_remote_srqn); if (unlikely((__pyx_t_1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1384, __pyx_L1_error)
  ibv_wr_set_xrc_srqn(__pyx_v_self->qp_ex, __pyx_t_1);

  /* "pyverbs/qp.pyx":1383
 *         v.ibv_wr_set_ud_addr(self.qp_ex, ah.ah, remote_qpn, remote_rkey)
 * 
 *     def wr_set_xrc_srqn(self, remote_srqn):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_set_xrc_srqn(self.qp_ex, remote_srqn)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_set_xrc_srqn", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1386
 *         v.ibv_wr_set_xrc_srqn(self.qp_ex, remote_srqn)
 * 
 *     def wr_set_inline_data(self, addr, length):             # <<<<<<<<<<<<<<
 *         ptr = PyLong_AsVoidPtr(addr)
 *         v.ibv_wr_set_inline_data(self.qp_ex, ptr, length)
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_33wr_set_inline_data(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_33wr_set_inline_data = {"wr_set_inline_data", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_33wr_set_inline_data, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_33wr_set_inline_data(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_addr = 0;
  PyObject *__pyx_v_length = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_set_inline_data (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_addr,&__pyx_mstate_global->__pyx_n_u_length,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1386, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1386, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1386, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "wr_set_inline_data", 0) < (0)) __PYX_ERR(0, 1386, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("wr_set_inline_data", 1, 2, 2, i); __PYX_ERR(0, 1386, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1386, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1386, __pyx_L3_error)
    }
    __pyx_v_addr = values[0];
    __pyx_v_length = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wr_set_inline_data", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1386, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_set_inline_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_32wr_set_inline_data(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self), __pyx_v_addr, __pyx_v_length);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_32wr_set_inline_data(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_addr, PyObject *__pyx_v_length) {
  void *__pyx_v_ptr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wr_set_inline_data", 0);

  /* "pyverbs/qp.pyx":1387
 * 
 *     def wr_set_inline_data(self, addr, length):
 *         ptr = PyLong_AsVoidPtr(addr)             # <<<<<<<<<<<<<<
 *         v.ibv_wr_set_inline_data(self.qp_ex, ptr, length)
 * 
*/
  __pyx_v_ptr = PyLong_AsVoidPtr(__pyx_v_addr);

  /* "pyverbs/qp.pyx":1388
 *     def wr_set_inline_data(self, addr, length):
 *         ptr = PyLong_AsVoidPtr(addr)
 *         v.ibv_wr_set_inline_data(self.qp_ex, ptr, length)             # <<<<<<<<<<<<<<
 * 
 *     def wr_set_inline_data_list(self, num_buf, buf_list):
*/
  __pyx_t_1 = __Pyx_PyLong_As_size_t(__pyx_v_length); if (unlikely((__pyx_t_1 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1388, __pyx_L1_error)
  ibv_wr_set_inline_data(__pyx_v_self->qp_ex, __pyx_v_ptr, __pyx_t_1);

  /* "pyverbs/qp.pyx":1386
 *         v.ibv_wr_set_xrc_srqn(self.qp_ex, remote_srqn)
 * 
 *     def wr_set_inline_data(self, addr, length):             # <<<<<<<<<<<<<<
 *         ptr = PyLong_AsVoidPtr(addr)
 *         v.ibv_wr_set_inline_data(self.qp_ex, ptr, length)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_set_inline_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1390
 *         v.ibv_wr_set_inline_data(self.qp_ex, ptr, length)
 * 
 *     def wr_set_inline_data_list(self, num_buf, buf_list):             # <<<<<<<<<<<<<<
 *         cdef v.ibv_data_buf *data = NULL
 *         data = <v.ibv_data_buf*>malloc(num_buf * sizeof(v.ibv_data_buf))
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_35wr_set_inline_data_list(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_35wr_set_inline_data_list = {"wr_set_inline_data_list", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_35wr_set_inline_data_list, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_35wr_set_inline_data_list(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_num_buf = 0;
  PyObject *__pyx_v_buf_list = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_set_inline_data_list (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_num_buf,&__pyx_mstate_global->__pyx_n_u_buf_list,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1390, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1390, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1390, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "wr_set_inline_data_list", 0) < (0)) __PYX_ERR(0, 1390, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("wr_set_inline_data_list", 1, 2, 2, i); __PYX_ERR(0, 1390, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1390, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1390, __pyx_L3_error)
    }
    __pyx_v_num_buf = values[0];
    __pyx_v_buf_list = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wr_set_inline_data_list", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1390, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_set_inline_data_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_34wr_set_inline_data_list(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self), __pyx_v_num_buf, __pyx_v_buf_list);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_34wr_set_inline_data_list(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_num_buf, PyObject *__pyx_v_buf_list) {
  struct ibv_data_buf *__pyx_v_data;
  PyObject *__pyx_v_i = NULL;
  struct __pyx_obj_7pyverbs_2qp_DataBuffer *__pyx_v_data_buf = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  void *__pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wr_set_inline_data_list", 0);

  /* "pyverbs/qp.pyx":1391
 * 
 *     def wr_set_inline_data_list(self, num_buf, buf_list):
 *         cdef v.ibv_data_buf *data = NULL             # <<<<<<<<<<<<<<
 *         data = <v.ibv_data_buf*>malloc(num_buf * sizeof(v.ibv_data_buf))
 *         if data == NULL:
*/
  __pyx_v_data = NULL;

  /* "pyverbs/qp.pyx":1392
 *     def wr_set_inline_data_list(self, num_buf, buf_list):
 *         cdef v.ibv_data_buf *data = NULL
 *         data = <v.ibv_data_buf*>malloc(num_buf * sizeof(v.ibv_data_buf))             # <<<<<<<<<<<<<<
 *         if data == NULL:
 *             raise PyverbsError('Failed to allocate data buffer')
*/
  __pyx_t_1 = __Pyx_PyLong_FromSize_t((sizeof(struct ibv_data_buf))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Multiply(__pyx_v_num_buf, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyLong_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_3 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1392, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_data = ((struct ibv_data_buf *)malloc(__pyx_t_3));

  /* "pyverbs/qp.pyx":1393
 *         cdef v.ibv_data_buf *data = NULL
 *         data = <v.ibv_data_buf*>malloc(num_buf * sizeof(v.ibv_data_buf))
 *         if data == NULL:             # <<<<<<<<<<<<<<
 *             raise PyverbsError('Failed to allocate data buffer')
 *         for i in range(num_buf):
*/
  __pyx_t_4 = (__pyx_v_data == NULL);
  if (unlikely(__pyx_t_4)) {

    /* "pyverbs/qp.pyx":1394
 *         data = <v.ibv_data_buf*>malloc(num_buf * sizeof(v.ibv_data_buf))
 *         if data == NULL:
 *             raise PyverbsError('Failed to allocate data buffer')             # <<<<<<<<<<<<<<
 *         for i in range(num_buf):
 *             data_buf = <DataBuffer>buf_list[i]
*/
    __pyx_t_1 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PyverbsError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_1);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_3 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_mstate_global->__pyx_kp_u_Failed_to_allocate_data_buffer};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1394, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1394, __pyx_L1_error)

    /* "pyverbs/qp.pyx":1393
 *         cdef v.ibv_data_buf *data = NULL
 *         data = <v.ibv_data_buf*>malloc(num_buf * sizeof(v.ibv_data_buf))
 *         if data == NULL:             # <<<<<<<<<<<<<<
 *             raise PyverbsError('Failed to allocate data buffer')
 *         for i in range(num_buf):
*/
  }

  /* "pyverbs/qp.pyx":1395
 *         if data == NULL:
 *             raise PyverbsError('Failed to allocate data buffer')
 *         for i in range(num_buf):             # <<<<<<<<<<<<<<
 *             data_buf = <DataBuffer>buf_list[i]
 *             data[i].addr = data_buf.data.addr
*/
  __pyx_t_5 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_num_buf};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(&PyRange_Type), __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1395, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_t_5 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1395, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    {
      __pyx_t_2 = __pyx_t_6(__pyx_t_5);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 1395, __pyx_L1_error)
          PyErr_Clear();
        }
        break;
      }
    }
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":1396
 *             raise PyverbsError('Failed to allocate data buffer')
 *         for i in range(num_buf):
 *             data_buf = <DataBuffer>buf_list[i]             # <<<<<<<<<<<<<<
 *             data[i].addr = data_buf.data.addr
 *             data[i].length = data_buf.data.length
*/
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_buf_list, __pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_data_buf, ((struct __pyx_obj_7pyverbs_2qp_DataBuffer *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "pyverbs/qp.pyx":1397
 *         for i in range(num_buf):
 *             data_buf = <DataBuffer>buf_list[i]
 *             data[i].addr = data_buf.data.addr             # <<<<<<<<<<<<<<
 *             data[i].length = data_buf.data.length
 *         v.ibv_wr_set_inline_data_list(self.qp_ex, num_buf, data)
*/
    __pyx_t_7 = __pyx_v_data_buf->data.addr;
    __pyx_t_8 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_8 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1397, __pyx_L1_error)
    (__pyx_v_data[__pyx_t_8]).addr = __pyx_t_7;

    /* "pyverbs/qp.pyx":1398
 *             data_buf = <DataBuffer>buf_list[i]
 *             data[i].addr = data_buf.data.addr
 *             data[i].length = data_buf.data.length             # <<<<<<<<<<<<<<
 *         v.ibv_wr_set_inline_data_list(self.qp_ex, num_buf, data)
 *         free(data)
*/
    __pyx_t_3 = __pyx_v_data_buf->data.length;
    __pyx_t_8 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_8 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1398, __pyx_L1_error)
    (__pyx_v_data[__pyx_t_8]).length = __pyx_t_3;

    /* "pyverbs/qp.pyx":1395
 *         if data == NULL:
 *             raise PyverbsError('Failed to allocate data buffer')
 *         for i in range(num_buf):             # <<<<<<<<<<<<<<
 *             data_buf = <DataBuffer>buf_list[i]
 *             data[i].addr = data_buf.data.addr
*/
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":1399
 *             data[i].addr = data_buf.data.addr
 *             data[i].length = data_buf.data.length
 *         v.ibv_wr_set_inline_data_list(self.qp_ex, num_buf, data)             # <<<<<<<<<<<<<<
 *         free(data)
 * 
*/
  __pyx_t_3 = __Pyx_PyLong_As_size_t(__pyx_v_num_buf); if (unlikely((__pyx_t_3 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1399, __pyx_L1_error)
  ibv_wr_set_inline_data_list(__pyx_v_self->qp_ex, __pyx_t_3, __pyx_v_data);

  /* "pyverbs/qp.pyx":1400
 *             data[i].length = data_buf.data.length
 *         v.ibv_wr_set_inline_data_list(self.qp_ex, num_buf, data)
 *         free(data)             # <<<<<<<<<<<<<<
 * 
 *     def wr_set_sge(self, SGE sge not None):
*/
  free(__pyx_v_data);

  /* "pyverbs/qp.pyx":1390
 *         v.ibv_wr_set_inline_data(self.qp_ex, ptr, length)
 * 
 *     def wr_set_inline_data_list(self, num_buf, buf_list):             # <<<<<<<<<<<<<<
 *         cdef v.ibv_data_buf *data = NULL
 *         data = <v.ibv_data_buf*>malloc(num_buf * sizeof(v.ibv_data_buf))
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_set_inline_data_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF((PyObject *)__pyx_v_data_buf);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1402
 *         free(data)
 * 
 *     def wr_set_sge(self, SGE sge not None):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_set_sge(self.qp_ex, sge.lkey, sge.addr, sge.length)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_37wr_set_sge(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_37wr_set_sge = {"wr_set_sge", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_37wr_set_sge, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_37wr_set_sge(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_2wr_SGE *__pyx_v_sge = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_set_sge (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_sge,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1402, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1402, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "wr_set_sge", 0) < (0)) __PYX_ERR(0, 1402, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("wr_set_sge", 1, 1, 1, i); __PYX_ERR(0, 1402, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1402, __pyx_L3_error)
    }
    __pyx_v_sge = ((struct __pyx_obj_7pyverbs_2wr_SGE *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wr_set_sge", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1402, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_set_sge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sge), __pyx_mstate_global->__pyx_ptype_7pyverbs_2wr_SGE, 0, "sge", 0))) __PYX_ERR(0, 1402, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_36wr_set_sge(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self), __pyx_v_sge);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_36wr_set_sge(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, struct __pyx_obj_7pyverbs_2wr_SGE *__pyx_v_sge) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  uint32_t __pyx_t_2;
  uint64_t __pyx_t_3;
  uint32_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wr_set_sge", 0);

  /* "pyverbs/qp.pyx":1403
 * 
 *     def wr_set_sge(self, SGE sge not None):
 *         v.ibv_wr_set_sge(self.qp_ex, sge.lkey, sge.addr, sge.length)             # <<<<<<<<<<<<<<
 * 
 *     def wr_set_sge_list(self, num_sge, sg_list):
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_sge), __pyx_mstate_global->__pyx_n_u_lkey); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_As_uint32_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1403, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_sge), __pyx_mstate_global->__pyx_n_u_addr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyLong_As_uint64_t(__pyx_t_1); if (unlikely((__pyx_t_3 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1403, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_sge), __pyx_mstate_global->__pyx_n_u_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyLong_As_uint32_t(__pyx_t_1); if (unlikely((__pyx_t_4 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1403, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  ibv_wr_set_sge(__pyx_v_self->qp_ex, __pyx_t_2, __pyx_t_3, __pyx_t_4);

  /* "pyverbs/qp.pyx":1402
 *         free(data)
 * 
 *     def wr_set_sge(self, SGE sge not None):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_set_sge(self.qp_ex, sge.lkey, sge.addr, sge.length)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_set_sge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1405
 *         v.ibv_wr_set_sge(self.qp_ex, sge.lkey, sge.addr, sge.length)
 * 
 *     def wr_set_sge_list(self, num_sge, sg_list):             # <<<<<<<<<<<<<<
 *         cdef v.ibv_sge *sge = NULL
 *         sge = <v.ibv_sge*>malloc(num_sge * sizeof(v.ibv_sge))
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_39wr_set_sge_list(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_39wr_set_sge_list = {"wr_set_sge_list", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_39wr_set_sge_list, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_39wr_set_sge_list(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_num_sge = 0;
  PyObject *__pyx_v_sg_list = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_set_sge_list (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_num_sge,&__pyx_mstate_global->__pyx_n_u_sg_list,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1405, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1405, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1405, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "wr_set_sge_list", 0) < (0)) __PYX_ERR(0, 1405, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("wr_set_sge_list", 1, 2, 2, i); __PYX_ERR(0, 1405, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1405, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1405, __pyx_L3_error)
    }
    __pyx_v_num_sge = values[0];
    __pyx_v_sg_list = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("wr_set_sge_list", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1405, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_set_sge_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_38wr_set_sge_list(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self), __pyx_v_num_sge, __pyx_v_sg_list);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_38wr_set_sge_list(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, PyObject *__pyx_v_num_sge, PyObject *__pyx_v_sg_list) {
  struct ibv_sge *__pyx_v_sge;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  unsigned long __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wr_set_sge_list", 0);

  /* "pyverbs/qp.pyx":1406
 * 
 *     def wr_set_sge_list(self, num_sge, sg_list):
 *         cdef v.ibv_sge *sge = NULL             # <<<<<<<<<<<<<<
 *         sge = <v.ibv_sge*>malloc(num_sge * sizeof(v.ibv_sge))
 *         if sge == NULL:
*/
  __pyx_v_sge = NULL;

  /* "pyverbs/qp.pyx":1407
 *     def wr_set_sge_list(self, num_sge, sg_list):
 *         cdef v.ibv_sge *sge = NULL
 *         sge = <v.ibv_sge*>malloc(num_sge * sizeof(v.ibv_sge))             # <<<<<<<<<<<<<<
 *         if sge == NULL:
 *             raise PyverbsError('Failed to allocate SGE buffer')
*/
  __pyx_t_1 = __Pyx_PyLong_FromSize_t((sizeof(struct ibv_sge))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Multiply(__pyx_v_num_sge, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyLong_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_3 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1407, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_sge = ((struct ibv_sge *)malloc(__pyx_t_3));

  /* "pyverbs/qp.pyx":1408
 *         cdef v.ibv_sge *sge = NULL
 *         sge = <v.ibv_sge*>malloc(num_sge * sizeof(v.ibv_sge))
 *         if sge == NULL:             # <<<<<<<<<<<<<<
 *             raise PyverbsError('Failed to allocate SGE buffer')
 *         for i in range(num_sge):
*/
  __pyx_t_4 = (__pyx_v_sge == NULL);
  if (unlikely(__pyx_t_4)) {

    /* "pyverbs/qp.pyx":1409
 *         sge = <v.ibv_sge*>malloc(num_sge * sizeof(v.ibv_sge))
 *         if sge == NULL:
 *             raise PyverbsError('Failed to allocate SGE buffer')             # <<<<<<<<<<<<<<
 *         for i in range(num_sge):
 *             sge[i].addr = sg_list[i].addr
*/
    __pyx_t_1 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PyverbsError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1409, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_1);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_3 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_mstate_global->__pyx_kp_u_Failed_to_allocate_SGE_buffer};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1409, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1409, __pyx_L1_error)

    /* "pyverbs/qp.pyx":1408
 *         cdef v.ibv_sge *sge = NULL
 *         sge = <v.ibv_sge*>malloc(num_sge * sizeof(v.ibv_sge))
 *         if sge == NULL:             # <<<<<<<<<<<<<<
 *             raise PyverbsError('Failed to allocate SGE buffer')
 *         for i in range(num_sge):
*/
  }

  /* "pyverbs/qp.pyx":1410
 *         if sge == NULL:
 *             raise PyverbsError('Failed to allocate SGE buffer')
 *         for i in range(num_sge):             # <<<<<<<<<<<<<<
 *             sge[i].addr = sg_list[i].addr
 *             sge[i].length = sg_list[i].length
*/
  __pyx_t_5 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_num_sge};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(&PyRange_Type), __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_t_5 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1410, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    {
      __pyx_t_2 = __pyx_t_6(__pyx_t_5);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 1410, __pyx_L1_error)
          PyErr_Clear();
        }
        break;
      }
    }
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":1411
 *             raise PyverbsError('Failed to allocate SGE buffer')
 *         for i in range(num_sge):
 *             sge[i].addr = sg_list[i].addr             # <<<<<<<<<<<<<<
 *             sge[i].length = sg_list[i].length
 *             sge[i].lkey = sg_list[i].lkey
*/
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_sg_list, __pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_addr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_7 = __Pyx_PyLong_As_unsigned_long(__pyx_t_1); if (unlikely((__pyx_t_7 == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1411, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_8 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1411, __pyx_L1_error)
    (__pyx_v_sge[__pyx_t_8]).addr = __pyx_t_7;

    /* "pyverbs/qp.pyx":1412
 *         for i in range(num_sge):
 *             sge[i].addr = sg_list[i].addr
 *             sge[i].length = sg_list[i].length             # <<<<<<<<<<<<<<
 *             sge[i].lkey = sg_list[i].lkey
 *         v.ibv_wr_set_sge_list(self.qp_ex, num_sge, sge)
*/
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_sg_list, __pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1412, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_length); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1412, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = __Pyx_PyLong_As_unsigned_int(__pyx_t_2); if (unlikely((__pyx_t_9 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1412, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_8 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_8 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1412, __pyx_L1_error)
    (__pyx_v_sge[__pyx_t_8]).length = __pyx_t_9;

    /* "pyverbs/qp.pyx":1413
 *             sge[i].addr = sg_list[i].addr
 *             sge[i].length = sg_list[i].length
 *             sge[i].lkey = sg_list[i].lkey             # <<<<<<<<<<<<<<
 *         v.ibv_wr_set_sge_list(self.qp_ex, num_sge, sge)
 *         free(sge)
*/
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_sg_list, __pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_lkey); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_9 = __Pyx_PyLong_As_unsigned_int(__pyx_t_1); if (unlikely((__pyx_t_9 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1413, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_8 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1413, __pyx_L1_error)
    (__pyx_v_sge[__pyx_t_8]).lkey = __pyx_t_9;

    /* "pyverbs/qp.pyx":1410
 *         if sge == NULL:
 *             raise PyverbsError('Failed to allocate SGE buffer')
 *         for i in range(num_sge):             # <<<<<<<<<<<<<<
 *             sge[i].addr = sg_list[i].addr
 *             sge[i].length = sg_list[i].length
*/
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/qp.pyx":1414
 *             sge[i].length = sg_list[i].length
 *             sge[i].lkey = sg_list[i].lkey
 *         v.ibv_wr_set_sge_list(self.qp_ex, num_sge, sge)             # <<<<<<<<<<<<<<
 *         free(sge)
 * 
*/
  __pyx_t_3 = __Pyx_PyLong_As_size_t(__pyx_v_num_sge); if (unlikely((__pyx_t_3 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1414, __pyx_L1_error)
  ibv_wr_set_sge_list(__pyx_v_self->qp_ex, __pyx_t_3, __pyx_v_sge);

  /* "pyverbs/qp.pyx":1415
 *             sge[i].lkey = sg_list[i].lkey
 *         v.ibv_wr_set_sge_list(self.qp_ex, num_sge, sge)
 *         free(sge)             # <<<<<<<<<<<<<<
 * 
 *     def wr_start(self):
*/
  free(__pyx_v_sge);

  /* "pyverbs/qp.pyx":1405
 *         v.ibv_wr_set_sge(self.qp_ex, sge.lkey, sge.addr, sge.length)
 * 
 *     def wr_set_sge_list(self, num_sge, sg_list):             # <<<<<<<<<<<<<<
 *         cdef v.ibv_sge *sge = NULL
 *         sge = <v.ibv_sge*>malloc(num_sge * sizeof(v.ibv_sge))
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_set_sge_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1417
 *         free(sge)
 * 
 *     def wr_start(self):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_start(self.qp_ex)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_41wr_start(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_41wr_start = {"wr_start", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_41wr_start, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_41wr_start(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_start (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("wr_start", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("wr_start", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_40wr_start(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_40wr_start(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_start", 0);

  /* "pyverbs/qp.pyx":1418
 * 
 *     def wr_start(self):
 *         v.ibv_wr_start(self.qp_ex)             # <<<<<<<<<<<<<<
 * 
 *     def wr_complete(self):
*/
  ibv_wr_start(__pyx_v_self->qp_ex);

  /* "pyverbs/qp.pyx":1417
 *         free(sge)
 * 
 *     def wr_start(self):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_start(self.qp_ex)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1420
 *         v.ibv_wr_start(self.qp_ex)
 * 
 *     def wr_complete(self):             # <<<<<<<<<<<<<<
 *         rc = v.ibv_wr_complete(self.qp_ex)
 *         if rc != 0:
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_43wr_complete(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_43wr_complete = {"wr_complete", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_43wr_complete, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_43wr_complete(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_complete (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("wr_complete", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("wr_complete", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_42wr_complete(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_42wr_complete(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self) {
  int __pyx_v_rc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wr_complete", 0);

  /* "pyverbs/qp.pyx":1421
 * 
 *     def wr_complete(self):
 *         rc = v.ibv_wr_complete(self.qp_ex)             # <<<<<<<<<<<<<<
 *         if rc != 0:
 *             raise PyverbsRDMAError('ibv_wr_complete failed.', rc)
*/
  __pyx_v_rc = ibv_wr_complete(__pyx_v_self->qp_ex);

  /* "pyverbs/qp.pyx":1422
 *     def wr_complete(self):
 *         rc = v.ibv_wr_complete(self.qp_ex)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('ibv_wr_complete failed.', rc)
 * 
*/
  __pyx_t_1 = (__pyx_v_rc != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pyverbs/qp.pyx":1423
 *         rc = v.ibv_wr_complete(self.qp_ex)
 *         if rc != 0:
 *             raise PyverbsRDMAError('ibv_wr_complete failed.', rc)             # <<<<<<<<<<<<<<
 * 
 *     def wr_abort(self):
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyLong_From_int(__pyx_v_rc); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_6 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_ibv_wr_complete_failed, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_6, (3-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1423, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1423, __pyx_L1_error)

    /* "pyverbs/qp.pyx":1422
 *     def wr_complete(self):
 *         rc = v.ibv_wr_complete(self.qp_ex)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('ibv_wr_complete failed.', rc)
 * 
*/
  }

  /* "pyverbs/qp.pyx":1420
 *         v.ibv_wr_start(self.qp_ex)
 * 
 *     def wr_complete(self):             # <<<<<<<<<<<<<<
 *         rc = v.ibv_wr_complete(self.qp_ex)
 *         if rc != 0:
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyverbs.qp.QPEx.wr_complete", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1425
 *             raise PyverbsRDMAError('ibv_wr_complete failed.', rc)
 * 
 *     def wr_abort(self):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_abort(self.qp_ex)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_45wr_abort(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_45wr_abort = {"wr_abort", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_45wr_abort, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_45wr_abort(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_abort (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("wr_abort", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("wr_abort", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_44wr_abort(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_44wr_abort(struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wr_abort", 0);

  /* "pyverbs/qp.pyx":1426
 * 
 *     def wr_abort(self):
 *         v.ibv_wr_abort(self.qp_ex)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  ibv_wr_abort(__pyx_v_self->qp_ex);

  /* "pyverbs/qp.pyx":1425
 *             raise PyverbsRDMAError('ibv_wr_complete failed.', rc)
 * 
 *     def wr_abort(self):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_abort(self.qp_ex)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.qp,self.qp_ex cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_47__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_47__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_47__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_47__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_46__reduce_cython__(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_46__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.qp,self.qp_ex cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.qp,self.qp_ex cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_qp_self_qp_ex_cannot_be_con, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.qp,self.qp_ex cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPEx.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.qp,self.qp_ex cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.qp,self.qp_ex cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_49__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2qp_4QPEx_49__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_49__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2qp_4QPEx_49__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp.QPEx.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2qp_4QPEx_48__setstate_cython__(((struct __pyx_obj_7pyverbs_2qp_QPEx *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp_4QPEx_48__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2qp_QPEx *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.qp,self.qp_ex cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.qp,self.qp_ex cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_qp_self_qp_ex_cannot_be_con, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.qp,self.qp_ex cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.qp,self.qp_ex cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.qp.QPEx.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/qp.pyx":1429
 * 
 * 
 * def _copy_caps(QPCap src, dst):             # <<<<<<<<<<<<<<
 *     """
 *     Copy the QPCaps values of src into the inner ibv_qp_cap struct of dst.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2qp_1_copy_caps(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2qp__copy_caps, "\n    Copy the QPCaps values of src into the inner ibv_qp_cap struct of dst.\n    Since both ibv_qp_init_attr and ibv_qp_attr have an inner ibv_qp_cap inner\n    struct, they can both be used.\n    :param src: A QPCap object\n    :param dst: A QPInitAttr / QPInitAttrEx / QPAttr object\n    :return: None\n    ");
static PyMethodDef __pyx_mdef_7pyverbs_2qp_1_copy_caps = {"_copy_caps", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_1_copy_caps, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2qp__copy_caps};
static PyObject *__pyx_pw_7pyverbs_2qp_1_copy_caps(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_src = 0;
  PyObject *__pyx_v_dst = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_copy_caps (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_dst,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1429, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1429, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1429, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "_copy_caps", 0) < (0)) __PYX_ERR(0, 1429, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("_copy_caps", 1, 2, 2, i); __PYX_ERR(0, 1429, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1429, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1429, __pyx_L3_error)
    }
    __pyx_v_src = ((struct __pyx_obj_7pyverbs_2qp_QPCap *)values[0]);
    __pyx_v_dst = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_copy_caps", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1429, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.qp._copy_caps", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_src), __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPCap, 1, "src", 0))) __PYX_ERR(0, 1429, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2qp__copy_caps(__pyx_self, __pyx_v_src, __pyx_v_dst);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2qp__copy_caps(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_7pyverbs_2qp_QPCap *__pyx_v_src, PyObject *__pyx_v_dst) {
  struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_qia = 0;
  struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *__pyx_v_qiae = 0;
  struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_qa = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  unsigned int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_copy_caps", 0);

  /* "pyverbs/qp.pyx":1444
 *     cdef QPInitAttrEx qiae
 *     cdef QPAttr qa
 *     if src is None:             # <<<<<<<<<<<<<<
 *         return
 *     if type(dst) == QPInitAttr:
*/
  __pyx_t_1 = (((PyObject *)__pyx_v_src) == Py_None);
  if (__pyx_t_1) {

    /* "pyverbs/qp.pyx":1445
 *     cdef QPAttr qa
 *     if src is None:
 *         return             # <<<<<<<<<<<<<<
 *     if type(dst) == QPInitAttr:
 *         qia = <QPInitAttr>dst
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "pyverbs/qp.pyx":1444
 *     cdef QPInitAttrEx qiae
 *     cdef QPAttr qa
 *     if src is None:             # <<<<<<<<<<<<<<
 *         return
 *     if type(dst) == QPInitAttr:
*/
  }

  /* "pyverbs/qp.pyx":1446
 *     if src is None:
 *         return
 *     if type(dst) == QPInitAttr:             # <<<<<<<<<<<<<<
 *         qia = <QPInitAttr>dst
 *         qia.attr.cap.max_send_wr = src.cap.max_send_wr
*/
  __pyx_t_2 = PyObject_RichCompare(((PyObject *)Py_TYPE(__pyx_v_dst)), ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPInitAttr), Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1446, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1446, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_1) {

    /* "pyverbs/qp.pyx":1447
 *         return
 *     if type(dst) == QPInitAttr:
 *         qia = <QPInitAttr>dst             # <<<<<<<<<<<<<<
 *         qia.attr.cap.max_send_wr = src.cap.max_send_wr
 *         qia.attr.cap.max_recv_wr = src.cap.max_recv_wr
*/
    __pyx_t_2 = __pyx_v_dst;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_qia = ((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":1448
 *     if type(dst) == QPInitAttr:
 *         qia = <QPInitAttr>dst
 *         qia.attr.cap.max_send_wr = src.cap.max_send_wr             # <<<<<<<<<<<<<<
 *         qia.attr.cap.max_recv_wr = src.cap.max_recv_wr
 *         qia.attr.cap.max_send_sge = src.cap.max_send_sge
*/
    __pyx_t_3 = __pyx_v_src->cap.max_send_wr;
    __pyx_v_qia->attr.cap.max_send_wr = __pyx_t_3;

    /* "pyverbs/qp.pyx":1449
 *         qia = <QPInitAttr>dst
 *         qia.attr.cap.max_send_wr = src.cap.max_send_wr
 *         qia.attr.cap.max_recv_wr = src.cap.max_recv_wr             # <<<<<<<<<<<<<<
 *         qia.attr.cap.max_send_sge = src.cap.max_send_sge
 *         qia.attr.cap.max_recv_sge = src.cap.max_recv_sge
*/
    __pyx_t_3 = __pyx_v_src->cap.max_recv_wr;
    __pyx_v_qia->attr.cap.max_recv_wr = __pyx_t_3;

    /* "pyverbs/qp.pyx":1450
 *         qia.attr.cap.max_send_wr = src.cap.max_send_wr
 *         qia.attr.cap.max_recv_wr = src.cap.max_recv_wr
 *         qia.attr.cap.max_send_sge = src.cap.max_send_sge             # <<<<<<<<<<<<<<
 *         qia.attr.cap.max_recv_sge = src.cap.max_recv_sge
 *         qia.attr.cap.max_inline_data = src.cap.max_inline_data
*/
    __pyx_t_3 = __pyx_v_src->cap.max_send_sge;
    __pyx_v_qia->attr.cap.max_send_sge = __pyx_t_3;

    /* "pyverbs/qp.pyx":1451
 *         qia.attr.cap.max_recv_wr = src.cap.max_recv_wr
 *         qia.attr.cap.max_send_sge = src.cap.max_send_sge
 *         qia.attr.cap.max_recv_sge = src.cap.max_recv_sge             # <<<<<<<<<<<<<<
 *         qia.attr.cap.max_inline_data = src.cap.max_inline_data
 *     elif type(dst) == QPInitAttrEx:
*/
    __pyx_t_3 = __pyx_v_src->cap.max_recv_sge;
    __pyx_v_qia->attr.cap.max_recv_sge = __pyx_t_3;

    /* "pyverbs/qp.pyx":1452
 *         qia.attr.cap.max_send_sge = src.cap.max_send_sge
 *         qia.attr.cap.max_recv_sge = src.cap.max_recv_sge
 *         qia.attr.cap.max_inline_data = src.cap.max_inline_data             # <<<<<<<<<<<<<<
 *     elif type(dst) == QPInitAttrEx:
 *         qiae = <QPInitAttrEx>dst
*/
    __pyx_t_3 = __pyx_v_src->cap.max_inline_data;
    __pyx_v_qia->attr.cap.max_inline_data = __pyx_t_3;

    /* "pyverbs/qp.pyx":1446
 *     if src is None:
 *         return
 *     if type(dst) == QPInitAttr:             # <<<<<<<<<<<<<<
 *         qia = <QPInitAttr>dst
 *         qia.attr.cap.max_send_wr = src.cap.max_send_wr
*/
    goto __pyx_L4;
  }

  /* "pyverbs/qp.pyx":1453
 *         qia.attr.cap.max_recv_sge = src.cap.max_recv_sge
 *         qia.attr.cap.max_inline_data = src.cap.max_inline_data
 *     elif type(dst) == QPInitAttrEx:             # <<<<<<<<<<<<<<
 *         qiae = <QPInitAttrEx>dst
 *         qiae.attr.cap.max_send_wr = src.cap.max_send_wr
*/
  __pyx_t_2 = PyObject_RichCompare(((PyObject *)Py_TYPE(__pyx_v_dst)), ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx), Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1453, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1453, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_1) {

    /* "pyverbs/qp.pyx":1454
 *         qia.attr.cap.max_inline_data = src.cap.max_inline_data
 *     elif type(dst) == QPInitAttrEx:
 *         qiae = <QPInitAttrEx>dst             # <<<<<<<<<<<<<<
 *         qiae.attr.cap.max_send_wr = src.cap.max_send_wr
 *         qiae.attr.cap.max_recv_wr = src.cap.max_recv_wr
*/
    __pyx_t_2 = __pyx_v_dst;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_qiae = ((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":1455
 *     elif type(dst) == QPInitAttrEx:
 *         qiae = <QPInitAttrEx>dst
 *         qiae.attr.cap.max_send_wr = src.cap.max_send_wr             # <<<<<<<<<<<<<<
 *         qiae.attr.cap.max_recv_wr = src.cap.max_recv_wr
 *         qiae.attr.cap.max_send_sge = src.cap.max_send_sge
*/
    __pyx_t_3 = __pyx_v_src->cap.max_send_wr;
    __pyx_v_qiae->attr.cap.max_send_wr = __pyx_t_3;

    /* "pyverbs/qp.pyx":1456
 *         qiae = <QPInitAttrEx>dst
 *         qiae.attr.cap.max_send_wr = src.cap.max_send_wr
 *         qiae.attr.cap.max_recv_wr = src.cap.max_recv_wr             # <<<<<<<<<<<<<<
 *         qiae.attr.cap.max_send_sge = src.cap.max_send_sge
 *         qiae.attr.cap.max_recv_sge = src.cap.max_recv_sge
*/
    __pyx_t_3 = __pyx_v_src->cap.max_recv_wr;
    __pyx_v_qiae->attr.cap.max_recv_wr = __pyx_t_3;

    /* "pyverbs/qp.pyx":1457
 *         qiae.attr.cap.max_send_wr = src.cap.max_send_wr
 *         qiae.attr.cap.max_recv_wr = src.cap.max_recv_wr
 *         qiae.attr.cap.max_send_sge = src.cap.max_send_sge             # <<<<<<<<<<<<<<
 *         qiae.attr.cap.max_recv_sge = src.cap.max_recv_sge
 *         qiae.attr.cap.max_inline_data = src.cap.max_inline_data
*/
    __pyx_t_3 = __pyx_v_src->cap.max_send_sge;
    __pyx_v_qiae->attr.cap.max_send_sge = __pyx_t_3;

    /* "pyverbs/qp.pyx":1458
 *         qiae.attr.cap.max_recv_wr = src.cap.max_recv_wr
 *         qiae.attr.cap.max_send_sge = src.cap.max_send_sge
 *         qiae.attr.cap.max_recv_sge = src.cap.max_recv_sge             # <<<<<<<<<<<<<<
 *         qiae.attr.cap.max_inline_data = src.cap.max_inline_data
 *     else:
*/
    __pyx_t_3 = __pyx_v_src->cap.max_recv_sge;
    __pyx_v_qiae->attr.cap.max_recv_sge = __pyx_t_3;

    /* "pyverbs/qp.pyx":1459
 *         qiae.attr.cap.max_send_sge = src.cap.max_send_sge
 *         qiae.attr.cap.max_recv_sge = src.cap.max_recv_sge
 *         qiae.attr.cap.max_inline_data = src.cap.max_inline_data             # <<<<<<<<<<<<<<
 *     else:
 *         qa = <QPAttr>dst
*/
    __pyx_t_3 = __pyx_v_src->cap.max_inline_data;
    __pyx_v_qiae->attr.cap.max_inline_data = __pyx_t_3;

    /* "pyverbs/qp.pyx":1453
 *         qia.attr.cap.max_recv_sge = src.cap.max_recv_sge
 *         qia.attr.cap.max_inline_data = src.cap.max_inline_data
 *     elif type(dst) == QPInitAttrEx:             # <<<<<<<<<<<<<<
 *         qiae = <QPInitAttrEx>dst
 *         qiae.attr.cap.max_send_wr = src.cap.max_send_wr
*/
    goto __pyx_L4;
  }

  /* "pyverbs/qp.pyx":1461
 *         qiae.attr.cap.max_inline_data = src.cap.max_inline_data
 *     else:
 *         qa = <QPAttr>dst             # <<<<<<<<<<<<<<
 *         qa.attr.cap.max_send_wr = src.cap.max_send_wr
 *         qa.attr.cap.max_recv_wr = src.cap.max_recv_wr
*/
  /*else*/ {
    __pyx_t_2 = __pyx_v_dst;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_qa = ((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyverbs/qp.pyx":1462
 *     else:
 *         qa = <QPAttr>dst
 *         qa.attr.cap.max_send_wr = src.cap.max_send_wr             # <<<<<<<<<<<<<<
 *         qa.attr.cap.max_recv_wr = src.cap.max_recv_wr
 *         qa.attr.cap.max_send_sge = src.cap.max_send_sge
*/
    __pyx_t_3 = __pyx_v_src->cap.max_send_wr;
    __pyx_v_qa->attr.cap.max_send_wr = __pyx_t_3;

    /* "pyverbs/qp.pyx":1463
 *         qa = <QPAttr>dst
 *         qa.attr.cap.max_send_wr = src.cap.max_send_wr
 *         qa.attr.cap.max_recv_wr = src.cap.max_recv_wr             # <<<<<<<<<<<<<<
 *         qa.attr.cap.max_send_sge = src.cap.max_send_sge
 *         qa.attr.cap.max_recv_sge = src.cap.max_recv_sge
*/
    __pyx_t_3 = __pyx_v_src->cap.max_recv_wr;
    __pyx_v_qa->attr.cap.max_recv_wr = __pyx_t_3;

    /* "pyverbs/qp.pyx":1464
 *         qa.attr.cap.max_send_wr = src.cap.max_send_wr
 *         qa.attr.cap.max_recv_wr = src.cap.max_recv_wr
 *         qa.attr.cap.max_send_sge = src.cap.max_send_sge             # <<<<<<<<<<<<<<
 *         qa.attr.cap.max_recv_sge = src.cap.max_recv_sge
 *         qa.attr.cap.max_inline_data = src.cap.max_inline_data
*/
    __pyx_t_3 = __pyx_v_src->cap.max_send_sge;
    __pyx_v_qa->attr.cap.max_send_sge = __pyx_t_3;

    /* "pyverbs/qp.pyx":1465
 *         qa.attr.cap.max_recv_wr = src.cap.max_recv_wr
 *         qa.attr.cap.max_send_sge = src.cap.max_send_sge
 *         qa.attr.cap.max_recv_sge = src.cap.max_recv_sge             # <<<<<<<<<<<<<<
 *         qa.attr.cap.max_inline_data = src.cap.max_inline_data
*/
    __pyx_t_3 = __pyx_v_src->cap.max_recv_sge;
    __pyx_v_qa->attr.cap.max_recv_sge = __pyx_t_3;

    /* "pyverbs/qp.pyx":1466
 *         qa.attr.cap.max_send_sge = src.cap.max_send_sge
 *         qa.attr.cap.max_recv_sge = src.cap.max_recv_sge
 *         qa.attr.cap.max_inline_data = src.cap.max_inline_data             # <<<<<<<<<<<<<<
*/
    __pyx_t_3 = __pyx_v_src->cap.max_inline_data;
    __pyx_v_qa->attr.cap.max_inline_data = __pyx_t_3;
  }
  __pyx_L4:;

  /* "pyverbs/qp.pyx":1429
 * 
 * 
 * def _copy_caps(QPCap src, dst):             # <<<<<<<<<<<<<<
 *     """
 *     Copy the QPCaps values of src into the inner ibv_qp_cap struct of dst.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyverbs.qp._copy_caps", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_qia);
  __Pyx_XDECREF((PyObject *)__pyx_v_qiae);
  __Pyx_XDECREF((PyObject *)__pyx_v_qa);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
/* #### Code section: module_exttypes ### */

static PyObject *__pyx_tp_new_7pyverbs_2qp_QPCap(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc_7pyverbs_2qp_QPCap(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyverbs_2qp_QPCap) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  if (PyType_IS_GC(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) PyObject_GC_Track(o);
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_dealloc, destructor)(o); else
  #endif
  __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_7pyverbs_2qp_QPCap);
}

static int __pyx_tp_traverse_7pyverbs_2qp_QPCap(PyObject *o, visitproc v, void *a) {
  int e;
  #if !CYTHON_USE_MODULE_STATE
  e = 0;
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    traverseproc traverse = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_traverse, traverseproc);
    if (traverse) { e = traverse(o, v, a); }
  } else
  #endif
  { e = __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7pyverbs_2qp_QPCap); }
  if (e) return e;
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyverbs_2qp_QPCap(PyObject *o) {
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    inquiry clear = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_clear, inquiry);
    if (clear) clear(o);
  } else
  #endif
  { __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7pyverbs_2qp_QPCap); }
  return 0;
}

static PyObject *__pyx_getprop_7pyverbs_2qp_5QPCap_max_send_wr(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_5QPCap_11max_send_wr_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_5QPCap_max_send_wr(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_5QPCap_11max_send_wr_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_5QPCap_max_recv_wr(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_5QPCap_11max_recv_wr_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_5QPCap_max_recv_wr(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_5QPCap_11max_recv_wr_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_5QPCap_max_send_sge(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_5QPCap_12max_send_sge_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_5QPCap_max_send_sge(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_5QPCap_12max_send_sge_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_5QPCap_max_recv_sge(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_5QPCap_12max_recv_sge_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_5QPCap_max_recv_sge(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_5QPCap_12max_recv_sge_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_5QPCap_max_inline_data(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_5QPCap_15max_inline_data_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_5QPCap_max_inline_data(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_5QPCap_15max_inline_data_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_7pyverbs_2qp_QPCap[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_5QPCap_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_5QPCap_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7pyverbs_2qp_QPCap[] = {
  {"max_send_wr", __pyx_getprop_7pyverbs_2qp_5QPCap_max_send_wr, __pyx_setprop_7pyverbs_2qp_5QPCap_max_send_wr, 0, 0},
  {"max_recv_wr", __pyx_getprop_7pyverbs_2qp_5QPCap_max_recv_wr, __pyx_setprop_7pyverbs_2qp_5QPCap_max_recv_wr, 0, 0},
  {"max_send_sge", __pyx_getprop_7pyverbs_2qp_5QPCap_max_send_sge, __pyx_setprop_7pyverbs_2qp_5QPCap_max_send_sge, 0, 0},
  {"max_recv_sge", __pyx_getprop_7pyverbs_2qp_5QPCap_max_recv_sge, __pyx_setprop_7pyverbs_2qp_5QPCap_max_recv_sge, 0, 0},
  {"max_inline_data", __pyx_getprop_7pyverbs_2qp_5QPCap_max_inline_data, __pyx_setprop_7pyverbs_2qp_5QPCap_max_inline_data, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyverbs_2qp_QPCap_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyverbs_2qp_QPCap},
  {Py_tp_str, (void *)__pyx_pw_7pyverbs_2qp_5QPCap_3__str__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyverbs_2qp_QPCap},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyverbs_2qp_QPCap},
  {Py_tp_methods, (void *)__pyx_methods_7pyverbs_2qp_QPCap},
  {Py_tp_getset, (void *)__pyx_getsets_7pyverbs_2qp_QPCap},
  {Py_tp_init, (void *)__pyx_pw_7pyverbs_2qp_5QPCap_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7pyverbs_2qp_QPCap},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyverbs_2qp_QPCap_spec = {
  "pyverbs.qp.QPCap",
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPCap),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyverbs_2qp_QPCap_slots,
};
#else

static PyTypeObject __pyx_type_7pyverbs_2qp_QPCap = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyverbs.qp.""QPCap", /*tp_name*/
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPCap), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyverbs_2qp_QPCap, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7pyverbs_2qp_5QPCap_3__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7pyverbs_2qp_QPCap, /*tp_traverse*/
  __pyx_tp_clear_7pyverbs_2qp_QPCap, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyverbs_2qp_QPCap, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7pyverbs_2qp_QPCap, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7pyverbs_2qp_5QPCap_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyverbs_2qp_QPCap, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7pyverbs_2qp_QPInitAttr(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7pyverbs_2qp_QPInitAttr *p;
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)o);
  p->scq = Py_None; Py_INCREF(Py_None);
  p->rcq = Py_None; Py_INCREF(Py_None);
  p->srq = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7pyverbs_2qp_QPInitAttr(PyObject *o) {
  struct __pyx_obj_7pyverbs_2qp_QPInitAttr *p = (struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyverbs_2qp_QPInitAttr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->scq);
  Py_CLEAR(p->rcq);
  Py_CLEAR(p->srq);
  if (PyType_IS_GC(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) PyObject_GC_Track(o);
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_dealloc, destructor)(o); else
  #endif
  __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_7pyverbs_2qp_QPInitAttr);
}

static int __pyx_tp_traverse_7pyverbs_2qp_QPInitAttr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7pyverbs_2qp_QPInitAttr *p = (struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)o;
  #if !CYTHON_USE_MODULE_STATE
  e = 0;
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    traverseproc traverse = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_traverse, traverseproc);
    if (traverse) { e = traverse(o, v, a); }
  } else
  #endif
  { e = __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7pyverbs_2qp_QPInitAttr); }
  if (e) return e;
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  if (p->scq) {
    e = (*v)(p->scq, a); if (e) return e;
  }
  if (p->rcq) {
    e = (*v)(p->rcq, a); if (e) return e;
  }
  if (p->srq) {
    e = (*v)(p->srq, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyverbs_2qp_QPInitAttr(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7pyverbs_2qp_QPInitAttr *p = (struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)o;
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    inquiry clear = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_clear, inquiry);
    if (clear) clear(o);
  } else
  #endif
  { __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7pyverbs_2qp_QPInitAttr); }
  tmp = ((PyObject*)p->scq);
  p->scq = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->rcq);
  p->rcq = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->srq);
  p->srq = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7pyverbs_2qp_10QPInitAttr_send_cq(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_10QPInitAttr_7send_cq_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_10QPInitAttr_send_cq(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_10QPInitAttr_7send_cq_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_10QPInitAttr_srq(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_10QPInitAttr_3srq_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_10QPInitAttr_srq(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_10QPInitAttr_3srq_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_10QPInitAttr_recv_cq(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_10QPInitAttr_7recv_cq_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_10QPInitAttr_recv_cq(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_10QPInitAttr_7recv_cq_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_10QPInitAttr_cap(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_10QPInitAttr_3cap_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_10QPInitAttr_cap(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_10QPInitAttr_3cap_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_10QPInitAttr_qp_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_10QPInitAttr_7qp_type_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_10QPInitAttr_qp_type(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_10QPInitAttr_7qp_type_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_10QPInitAttr_sq_sig_all(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_10QPInitAttr_10sq_sig_all_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_10QPInitAttr_sq_sig_all(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_10QPInitAttr_10sq_sig_all_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_10QPInitAttr_max_send_wr(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_10QPInitAttr_11max_send_wr_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_10QPInitAttr_max_send_wr(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_10QPInitAttr_11max_send_wr_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_10QPInitAttr_max_recv_wr(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_10QPInitAttr_11max_recv_wr_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_10QPInitAttr_max_recv_wr(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_10QPInitAttr_11max_recv_wr_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_10QPInitAttr_max_send_sge(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_10QPInitAttr_12max_send_sge_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_10QPInitAttr_max_send_sge(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_10QPInitAttr_12max_send_sge_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_10QPInitAttr_max_recv_sge(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_10QPInitAttr_12max_recv_sge_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_10QPInitAttr_max_recv_sge(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_10QPInitAttr_12max_recv_sge_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_10QPInitAttr_max_inline_data(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_10QPInitAttr_15max_inline_data_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_10QPInitAttr_max_inline_data(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_10QPInitAttr_15max_inline_data_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_7pyverbs_2qp_QPInitAttr[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_10QPInitAttr_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_10QPInitAttr_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7pyverbs_2qp_QPInitAttr[] = {
  {"send_cq", __pyx_getprop_7pyverbs_2qp_10QPInitAttr_send_cq, __pyx_setprop_7pyverbs_2qp_10QPInitAttr_send_cq, 0, 0},
  {"srq", __pyx_getprop_7pyverbs_2qp_10QPInitAttr_srq, __pyx_setprop_7pyverbs_2qp_10QPInitAttr_srq, 0, 0},
  {"recv_cq", __pyx_getprop_7pyverbs_2qp_10QPInitAttr_recv_cq, __pyx_setprop_7pyverbs_2qp_10QPInitAttr_recv_cq, 0, 0},
  {"cap", __pyx_getprop_7pyverbs_2qp_10QPInitAttr_cap, __pyx_setprop_7pyverbs_2qp_10QPInitAttr_cap, 0, 0},
  {"qp_type", __pyx_getprop_7pyverbs_2qp_10QPInitAttr_qp_type, __pyx_setprop_7pyverbs_2qp_10QPInitAttr_qp_type, 0, 0},
  {"sq_sig_all", __pyx_getprop_7pyverbs_2qp_10QPInitAttr_sq_sig_all, __pyx_setprop_7pyverbs_2qp_10QPInitAttr_sq_sig_all, 0, 0},
  {"max_send_wr", __pyx_getprop_7pyverbs_2qp_10QPInitAttr_max_send_wr, __pyx_setprop_7pyverbs_2qp_10QPInitAttr_max_send_wr, 0, 0},
  {"max_recv_wr", __pyx_getprop_7pyverbs_2qp_10QPInitAttr_max_recv_wr, __pyx_setprop_7pyverbs_2qp_10QPInitAttr_max_recv_wr, 0, 0},
  {"max_send_sge", __pyx_getprop_7pyverbs_2qp_10QPInitAttr_max_send_sge, __pyx_setprop_7pyverbs_2qp_10QPInitAttr_max_send_sge, 0, 0},
  {"max_recv_sge", __pyx_getprop_7pyverbs_2qp_10QPInitAttr_max_recv_sge, __pyx_setprop_7pyverbs_2qp_10QPInitAttr_max_recv_sge, 0, 0},
  {"max_inline_data", __pyx_getprop_7pyverbs_2qp_10QPInitAttr_max_inline_data, __pyx_setprop_7pyverbs_2qp_10QPInitAttr_max_inline_data, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyverbs_2qp_QPInitAttr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyverbs_2qp_QPInitAttr},
  {Py_tp_str, (void *)__pyx_pw_7pyverbs_2qp_10QPInitAttr_3__str__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyverbs_2qp_QPInitAttr},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyverbs_2qp_QPInitAttr},
  {Py_tp_methods, (void *)__pyx_methods_7pyverbs_2qp_QPInitAttr},
  {Py_tp_getset, (void *)__pyx_getsets_7pyverbs_2qp_QPInitAttr},
  {Py_tp_init, (void *)__pyx_pw_7pyverbs_2qp_10QPInitAttr_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7pyverbs_2qp_QPInitAttr},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyverbs_2qp_QPInitAttr_spec = {
  "pyverbs.qp.QPInitAttr",
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPInitAttr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyverbs_2qp_QPInitAttr_slots,
};
#else

static PyTypeObject __pyx_type_7pyverbs_2qp_QPInitAttr = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyverbs.qp.""QPInitAttr", /*tp_name*/
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPInitAttr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyverbs_2qp_QPInitAttr, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7pyverbs_2qp_10QPInitAttr_3__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7pyverbs_2qp_QPInitAttr, /*tp_traverse*/
  __pyx_tp_clear_7pyverbs_2qp_QPInitAttr, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyverbs_2qp_QPInitAttr, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7pyverbs_2qp_QPInitAttr, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7pyverbs_2qp_10QPInitAttr_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyverbs_2qp_QPInitAttr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7pyverbs_2qp_QPInitAttrEx(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *p;
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)o);
  p->scq = Py_None; Py_INCREF(Py_None);
  p->rcq = Py_None; Py_INCREF(Py_None);
  p->_pd = Py_None; Py_INCREF(Py_None);
  p->xrcd = Py_None; Py_INCREF(Py_None);
  p->srq = Py_None; Py_INCREF(Py_None);
  p->ind_table = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7pyverbs_2qp_QPInitAttrEx(PyObject *o) {
  struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *p = (struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyverbs_2qp_QPInitAttrEx) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->scq);
  Py_CLEAR(p->rcq);
  Py_CLEAR(p->_pd);
  Py_CLEAR(p->xrcd);
  Py_CLEAR(p->srq);
  Py_CLEAR(p->ind_table);
  if (PyType_IS_GC(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) PyObject_GC_Track(o);
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_dealloc, destructor)(o); else
  #endif
  __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_7pyverbs_2qp_QPInitAttrEx);
}

static int __pyx_tp_traverse_7pyverbs_2qp_QPInitAttrEx(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *p = (struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)o;
  #if !CYTHON_USE_MODULE_STATE
  e = 0;
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    traverseproc traverse = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_traverse, traverseproc);
    if (traverse) { e = traverse(o, v, a); }
  } else
  #endif
  { e = __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7pyverbs_2qp_QPInitAttrEx); }
  if (e) return e;
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  if (p->scq) {
    e = (*v)(p->scq, a); if (e) return e;
  }
  if (p->rcq) {
    e = (*v)(p->rcq, a); if (e) return e;
  }
  if (p->_pd) {
    e = (*v)(p->_pd, a); if (e) return e;
  }
  if (p->xrcd) {
    e = (*v)(p->xrcd, a); if (e) return e;
  }
  if (p->srq) {
    e = (*v)(p->srq, a); if (e) return e;
  }
  if (p->ind_table) {
    e = (*v)(p->ind_table, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyverbs_2qp_QPInitAttrEx(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *p = (struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx *)o;
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    inquiry clear = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_clear, inquiry);
    if (clear) clear(o);
  } else
  #endif
  { __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7pyverbs_2qp_QPInitAttrEx); }
  tmp = ((PyObject*)p->scq);
  p->scq = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->rcq);
  p->rcq = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_pd);
  p->_pd = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->xrcd);
  p->xrcd = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->srq);
  p->srq = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->ind_table);
  p->ind_table = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_send_cq(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_7send_cq_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_send_cq(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_7send_cq_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_recv_cq(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_7recv_cq_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_recv_cq(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_7recv_cq_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_cap(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_3cap_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_cap(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_3cap_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_qp_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_7qp_type_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_qp_type(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_7qp_type_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_sq_sig_all(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_10sq_sig_all_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_sq_sig_all(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_10sq_sig_all_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_comp_mask(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_9comp_mask_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_comp_mask(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_9comp_mask_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_pd(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_2pd_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_pd(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_2pd_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_xrcd(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_4xrcd_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_xrcd(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_4xrcd_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_srq(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_3srq_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_srq(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_3srq_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_create_flags(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_12create_flags_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_create_flags(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_12create_flags_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_max_tso_header(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_14max_tso_header_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_max_tso_header(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_14max_tso_header_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_source_qpn(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_10source_qpn_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_source_qpn(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_10source_qpn_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_max_send_wr(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_11max_send_wr_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_max_send_wr(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_11max_send_wr_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_max_recv_wr(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_11max_recv_wr_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_max_recv_wr(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_11max_recv_wr_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_max_send_sge(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_12max_send_sge_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_max_send_sge(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_12max_send_sge_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_max_recv_sge(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_12max_recv_sge_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_max_recv_sge(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_12max_recv_sge_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_max_inline_data(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_15max_inline_data_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_max_inline_data(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_15max_inline_data_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_ind_table(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_9ind_table_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_ind_table(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_9ind_table_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_7pyverbs_2qp_QPInitAttrEx[] = {
  {"mask_to_str", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_3mask_to_str, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"flags_to_str", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_5flags_to_str, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7pyverbs_2qp_QPInitAttrEx[] = {
  {"send_cq", __pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_send_cq, __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_send_cq, 0, 0},
  {"recv_cq", __pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_recv_cq, __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_recv_cq, 0, 0},
  {"cap", __pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_cap, __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_cap, 0, 0},
  {"qp_type", __pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_qp_type, __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_qp_type, 0, 0},
  {"sq_sig_all", __pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_sq_sig_all, __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_sq_sig_all, 0, 0},
  {"comp_mask", __pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_comp_mask, __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_comp_mask, 0, 0},
  {"pd", __pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_pd, __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_pd, 0, 0},
  {"xrcd", __pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_xrcd, __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_xrcd, 0, 0},
  {"srq", __pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_srq, __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_srq, 0, 0},
  {"create_flags", __pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_create_flags, __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_create_flags, 0, 0},
  {"max_tso_header", __pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_max_tso_header, __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_max_tso_header, 0, 0},
  {"source_qpn", __pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_source_qpn, __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_source_qpn, 0, 0},
  {"max_send_wr", __pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_max_send_wr, __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_max_send_wr, 0, 0},
  {"max_recv_wr", __pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_max_recv_wr, __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_max_recv_wr, 0, 0},
  {"max_send_sge", __pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_max_send_sge, __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_max_send_sge, 0, 0},
  {"max_recv_sge", __pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_max_recv_sge, __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_max_recv_sge, 0, 0},
  {"max_inline_data", __pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_max_inline_data, __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_max_inline_data, 0, 0},
  {"ind_table", __pyx_getprop_7pyverbs_2qp_12QPInitAttrEx_ind_table, __pyx_setprop_7pyverbs_2qp_12QPInitAttrEx_ind_table, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyverbs_2qp_QPInitAttrEx_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyverbs_2qp_QPInitAttrEx},
  {Py_tp_str, (void *)__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_7__str__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyverbs_2qp_QPInitAttrEx},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyverbs_2qp_QPInitAttrEx},
  {Py_tp_methods, (void *)__pyx_methods_7pyverbs_2qp_QPInitAttrEx},
  {Py_tp_getset, (void *)__pyx_getsets_7pyverbs_2qp_QPInitAttrEx},
  {Py_tp_init, (void *)__pyx_pw_7pyverbs_2qp_12QPInitAttrEx_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7pyverbs_2qp_QPInitAttrEx},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyverbs_2qp_QPInitAttrEx_spec = {
  "pyverbs.qp.QPInitAttrEx",
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyverbs_2qp_QPInitAttrEx_slots,
};
#else

static PyTypeObject __pyx_type_7pyverbs_2qp_QPInitAttrEx = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyverbs.qp.""QPInitAttrEx", /*tp_name*/
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyverbs_2qp_QPInitAttrEx, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_7__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7pyverbs_2qp_QPInitAttrEx, /*tp_traverse*/
  __pyx_tp_clear_7pyverbs_2qp_QPInitAttrEx, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyverbs_2qp_QPInitAttrEx, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7pyverbs_2qp_QPInitAttrEx, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7pyverbs_2qp_12QPInitAttrEx_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyverbs_2qp_QPInitAttrEx, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7pyverbs_2qp_QPAttr(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc_7pyverbs_2qp_QPAttr(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyverbs_2qp_QPAttr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  if (PyType_IS_GC(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) PyObject_GC_Track(o);
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_dealloc, destructor)(o); else
  #endif
  __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_7pyverbs_2qp_QPAttr);
}

static int __pyx_tp_traverse_7pyverbs_2qp_QPAttr(PyObject *o, visitproc v, void *a) {
  int e;
  #if !CYTHON_USE_MODULE_STATE
  e = 0;
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    traverseproc traverse = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_traverse, traverseproc);
    if (traverse) { e = traverse(o, v, a); }
  } else
  #endif
  { e = __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7pyverbs_2qp_QPAttr); }
  if (e) return e;
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyverbs_2qp_QPAttr(PyObject *o) {
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    inquiry clear = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_clear, inquiry);
    if (clear) clear(o);
  } else
  #endif
  { __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7pyverbs_2qp_QPAttr); }
  return 0;
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_qp_state(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_8qp_state_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_qp_state(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_8qp_state_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_cur_qp_state(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_12cur_qp_state_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_cur_qp_state(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_12cur_qp_state_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_path_mtu(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_8path_mtu_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_path_mtu(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_8path_mtu_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_path_mig_state(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_14path_mig_state_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_path_mig_state(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_14path_mig_state_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_qkey(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_4qkey_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_qkey(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_4qkey_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_rq_psn(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_6rq_psn_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_rq_psn(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_6rq_psn_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_sq_psn(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_6sq_psn_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_sq_psn(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_6sq_psn_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_dest_qp_num(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_11dest_qp_num_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_dest_qp_num(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_11dest_qp_num_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_qp_access_flags(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_15qp_access_flags_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_qp_access_flags(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_15qp_access_flags_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_cap(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_3cap_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_cap(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_3cap_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_ah_attr(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_7ah_attr_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_ah_attr(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_7ah_attr_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_alt_ah_attr(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_11alt_ah_attr_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_alt_ah_attr(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_11alt_ah_attr_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_pkey_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_10pkey_index_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_pkey_index(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_10pkey_index_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_alt_pkey_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_14alt_pkey_index_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_alt_pkey_index(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_14alt_pkey_index_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_en_sqd_async_notify(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_19en_sqd_async_notify_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_en_sqd_async_notify(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_19en_sqd_async_notify_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_sq_draining(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_11sq_draining_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_sq_draining(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_11sq_draining_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_max_rd_atomic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_13max_rd_atomic_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_max_rd_atomic(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_13max_rd_atomic_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_max_dest_rd_atomic(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_18max_dest_rd_atomic_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_max_dest_rd_atomic(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_18max_dest_rd_atomic_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_min_rnr_timer(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_13min_rnr_timer_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_min_rnr_timer(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_13min_rnr_timer_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_port_num(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_8port_num_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_port_num(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_8port_num_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_timeout(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_7timeout_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_timeout(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_7timeout_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_retry_cnt(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_9retry_cnt_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_retry_cnt(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_9retry_cnt_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_rnr_retry(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_9rnr_retry_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_rnr_retry(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_9rnr_retry_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_alt_port_num(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_12alt_port_num_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_alt_port_num(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_12alt_port_num_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_alt_timeout(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_11alt_timeout_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_alt_timeout(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_11alt_timeout_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_6QPAttr_rate_limit(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_6QPAttr_10rate_limit_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_6QPAttr_rate_limit(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_6QPAttr_10rate_limit_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_7pyverbs_2qp_QPAttr[] = {
  {"_copy_ah", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_6QPAttr_3_copy_ah, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_6QPAttr_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_6QPAttr_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7pyverbs_2qp_QPAttr[] = {
  {"qp_state", __pyx_getprop_7pyverbs_2qp_6QPAttr_qp_state, __pyx_setprop_7pyverbs_2qp_6QPAttr_qp_state, 0, 0},
  {"cur_qp_state", __pyx_getprop_7pyverbs_2qp_6QPAttr_cur_qp_state, __pyx_setprop_7pyverbs_2qp_6QPAttr_cur_qp_state, 0, 0},
  {"path_mtu", __pyx_getprop_7pyverbs_2qp_6QPAttr_path_mtu, __pyx_setprop_7pyverbs_2qp_6QPAttr_path_mtu, 0, 0},
  {"path_mig_state", __pyx_getprop_7pyverbs_2qp_6QPAttr_path_mig_state, __pyx_setprop_7pyverbs_2qp_6QPAttr_path_mig_state, 0, 0},
  {"qkey", __pyx_getprop_7pyverbs_2qp_6QPAttr_qkey, __pyx_setprop_7pyverbs_2qp_6QPAttr_qkey, 0, 0},
  {"rq_psn", __pyx_getprop_7pyverbs_2qp_6QPAttr_rq_psn, __pyx_setprop_7pyverbs_2qp_6QPAttr_rq_psn, 0, 0},
  {"sq_psn", __pyx_getprop_7pyverbs_2qp_6QPAttr_sq_psn, __pyx_setprop_7pyverbs_2qp_6QPAttr_sq_psn, 0, 0},
  {"dest_qp_num", __pyx_getprop_7pyverbs_2qp_6QPAttr_dest_qp_num, __pyx_setprop_7pyverbs_2qp_6QPAttr_dest_qp_num, 0, 0},
  {"qp_access_flags", __pyx_getprop_7pyverbs_2qp_6QPAttr_qp_access_flags, __pyx_setprop_7pyverbs_2qp_6QPAttr_qp_access_flags, 0, 0},
  {"cap", __pyx_getprop_7pyverbs_2qp_6QPAttr_cap, __pyx_setprop_7pyverbs_2qp_6QPAttr_cap, 0, 0},
  {"ah_attr", __pyx_getprop_7pyverbs_2qp_6QPAttr_ah_attr, __pyx_setprop_7pyverbs_2qp_6QPAttr_ah_attr, 0, 0},
  {"alt_ah_attr", __pyx_getprop_7pyverbs_2qp_6QPAttr_alt_ah_attr, __pyx_setprop_7pyverbs_2qp_6QPAttr_alt_ah_attr, 0, 0},
  {"pkey_index", __pyx_getprop_7pyverbs_2qp_6QPAttr_pkey_index, __pyx_setprop_7pyverbs_2qp_6QPAttr_pkey_index, 0, 0},
  {"alt_pkey_index", __pyx_getprop_7pyverbs_2qp_6QPAttr_alt_pkey_index, __pyx_setprop_7pyverbs_2qp_6QPAttr_alt_pkey_index, 0, 0},
  {"en_sqd_async_notify", __pyx_getprop_7pyverbs_2qp_6QPAttr_en_sqd_async_notify, __pyx_setprop_7pyverbs_2qp_6QPAttr_en_sqd_async_notify, 0, 0},
  {"sq_draining", __pyx_getprop_7pyverbs_2qp_6QPAttr_sq_draining, __pyx_setprop_7pyverbs_2qp_6QPAttr_sq_draining, 0, 0},
  {"max_rd_atomic", __pyx_getprop_7pyverbs_2qp_6QPAttr_max_rd_atomic, __pyx_setprop_7pyverbs_2qp_6QPAttr_max_rd_atomic, 0, 0},
  {"max_dest_rd_atomic", __pyx_getprop_7pyverbs_2qp_6QPAttr_max_dest_rd_atomic, __pyx_setprop_7pyverbs_2qp_6QPAttr_max_dest_rd_atomic, 0, 0},
  {"min_rnr_timer", __pyx_getprop_7pyverbs_2qp_6QPAttr_min_rnr_timer, __pyx_setprop_7pyverbs_2qp_6QPAttr_min_rnr_timer, 0, 0},
  {"port_num", __pyx_getprop_7pyverbs_2qp_6QPAttr_port_num, __pyx_setprop_7pyverbs_2qp_6QPAttr_port_num, 0, 0},
  {"timeout", __pyx_getprop_7pyverbs_2qp_6QPAttr_timeout, __pyx_setprop_7pyverbs_2qp_6QPAttr_timeout, 0, 0},
  {"retry_cnt", __pyx_getprop_7pyverbs_2qp_6QPAttr_retry_cnt, __pyx_setprop_7pyverbs_2qp_6QPAttr_retry_cnt, 0, 0},
  {"rnr_retry", __pyx_getprop_7pyverbs_2qp_6QPAttr_rnr_retry, __pyx_setprop_7pyverbs_2qp_6QPAttr_rnr_retry, 0, 0},
  {"alt_port_num", __pyx_getprop_7pyverbs_2qp_6QPAttr_alt_port_num, __pyx_setprop_7pyverbs_2qp_6QPAttr_alt_port_num, 0, 0},
  {"alt_timeout", __pyx_getprop_7pyverbs_2qp_6QPAttr_alt_timeout, __pyx_setprop_7pyverbs_2qp_6QPAttr_alt_timeout, 0, 0},
  {"rate_limit", __pyx_getprop_7pyverbs_2qp_6QPAttr_rate_limit, __pyx_setprop_7pyverbs_2qp_6QPAttr_rate_limit, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyverbs_2qp_QPAttr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyverbs_2qp_QPAttr},
  {Py_tp_str, (void *)__pyx_pw_7pyverbs_2qp_6QPAttr_5__str__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyverbs_2qp_QPAttr},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyverbs_2qp_QPAttr},
  {Py_tp_methods, (void *)__pyx_methods_7pyverbs_2qp_QPAttr},
  {Py_tp_getset, (void *)__pyx_getsets_7pyverbs_2qp_QPAttr},
  {Py_tp_init, (void *)__pyx_pw_7pyverbs_2qp_6QPAttr_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7pyverbs_2qp_QPAttr},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyverbs_2qp_QPAttr_spec = {
  "pyverbs.qp.QPAttr",
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPAttr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyverbs_2qp_QPAttr_slots,
};
#else

static PyTypeObject __pyx_type_7pyverbs_2qp_QPAttr = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyverbs.qp.""QPAttr", /*tp_name*/
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPAttr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyverbs_2qp_QPAttr, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7pyverbs_2qp_6QPAttr_5__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7pyverbs_2qp_QPAttr, /*tp_traverse*/
  __pyx_tp_clear_7pyverbs_2qp_QPAttr, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyverbs_2qp_QPAttr, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7pyverbs_2qp_QPAttr, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7pyverbs_2qp_6QPAttr_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyverbs_2qp_QPAttr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7pyverbs_2qp_QP __pyx_vtable_7pyverbs_2qp_QP;

static PyObject *__pyx_tp_new_7pyverbs_2qp_QP(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7pyverbs_2qp_QP *p;
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7pyverbs_2qp_QP *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM*)__pyx_vtabptr_7pyverbs_2qp_QP;
  p->pd = Py_None; Py_INCREF(Py_None);
  p->context = Py_None; Py_INCREF(Py_None);
  p->xrcd = Py_None; Py_INCREF(Py_None);
  p->scq = Py_None; Py_INCREF(Py_None);
  p->rcq = Py_None; Py_INCREF(Py_None);
  p->mws = Py_None; Py_INCREF(Py_None);
  p->srq = Py_None; Py_INCREF(Py_None);
  p->flows = Py_None; Py_INCREF(Py_None);
  p->dr_actions = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7pyverbs_2qp_QP(PyObject *o) {
  struct __pyx_obj_7pyverbs_2qp_QP *p = (struct __pyx_obj_7pyverbs_2qp_QP *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyverbs_2qp_QP) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_7pyverbs_2qp_2QP_7__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->pd);
  Py_CLEAR(p->context);
  Py_CLEAR(p->xrcd);
  Py_CLEAR(p->scq);
  Py_CLEAR(p->rcq);
  Py_CLEAR(p->mws);
  Py_CLEAR(p->srq);
  Py_CLEAR(p->flows);
  Py_CLEAR(p->dr_actions);
  if (PyType_IS_GC(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) PyObject_GC_Track(o);
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_dealloc, destructor)(o); else
  #endif
  __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_7pyverbs_2qp_QP);
}

static int __pyx_tp_traverse_7pyverbs_2qp_QP(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7pyverbs_2qp_QP *p = (struct __pyx_obj_7pyverbs_2qp_QP *)o;
  #if !CYTHON_USE_MODULE_STATE
  e = 0;
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) {
    traverseproc traverse = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_traverse, traverseproc);
    if (traverse) { e = traverse(o, v, a); }
  } else
  #endif
  { e = __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7pyverbs_2qp_QP); }
  if (e) return e;
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  if (p->pd) {
    e = (*v)(p->pd, a); if (e) return e;
  }
  if (p->context) {
    e = (*v)(p->context, a); if (e) return e;
  }
  if (p->xrcd) {
    e = (*v)(p->xrcd, a); if (e) return e;
  }
  if (p->scq) {
    e = (*v)(p->scq, a); if (e) return e;
  }
  if (p->rcq) {
    e = (*v)(p->rcq, a); if (e) return e;
  }
  if (p->mws) {
    e = (*v)(p->mws, a); if (e) return e;
  }
  if (p->srq) {
    e = (*v)(p->srq, a); if (e) return e;
  }
  if (p->flows) {
    e = (*v)(p->flows, a); if (e) return e;
  }
  if (p->dr_actions) {
    e = (*v)(p->dr_actions, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyverbs_2qp_QP(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7pyverbs_2qp_QP *p = (struct __pyx_obj_7pyverbs_2qp_QP *)o;
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) {
    inquiry clear = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_clear, inquiry);
    if (clear) clear(o);
  } else
  #endif
  { __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7pyverbs_2qp_QP); }
  tmp = ((PyObject*)p->pd);
  p->pd = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->context);
  p->context = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->xrcd);
  p->xrcd = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->scq);
  p->scq = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->rcq);
  p->rcq = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->mws);
  p->mws = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->srq);
  p->srq = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->flows);
  p->flows = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->dr_actions);
  p->dr_actions = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7pyverbs_2qp_2QP_qp_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_2QP_7qp_type_1__get__(o);
}

static PyObject *__pyx_getprop_7pyverbs_2qp_2QP_qp_state(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_2QP_8qp_state_1__get__(o);
}

static PyObject *__pyx_getprop_7pyverbs_2qp_2QP_qp_num(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_2QP_6qp_num_1__get__(o);
}

static PyObject *__pyx_getprop_7pyverbs_2qp_2QP_qp(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_2QP_2qp_1__get__(o);
}

static PyMethodDef __pyx_methods_7pyverbs_2qp_QP[] = {
  {"_create_qp", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_3_create_qp, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_create_qp_ex", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_5_create_qp_ex, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_get_comp_mask", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_11_get_comp_mask, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"to_init", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_13to_init, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2qp_2QP_12to_init},
  {"to_rtr", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_15to_rtr, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2qp_2QP_14to_rtr},
  {"to_rts", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_17to_rts, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2qp_2QP_16to_rts},
  {"query", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_19query, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2qp_2QP_18query},
  {"modify", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_21modify, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2qp_2QP_20modify},
  {"post_recv", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_23post_recv, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2qp_2QP_22post_recv},
  {"post_send", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_25post_send, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2qp_2QP_24post_send},
  {"set_ece", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_27set_ece, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2qp_2QP_26set_ece},
  {"query_ece", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_29query_ece, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2qp_2QP_28query_ece},
  {"bind_mw", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_31bind_mw, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2qp_2QP_30bind_mw},
  {"query_data_in_order", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_33query_data_in_order, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2qp_2QP_32query_data_in_order},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_37__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_2QP_39__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7pyverbs_2qp_QP[] = {
  {"qp_type", __pyx_getprop_7pyverbs_2qp_2QP_qp_type, 0, 0, 0},
  {"qp_state", __pyx_getprop_7pyverbs_2qp_2QP_qp_state, 0, 0, 0},
  {"qp_num", __pyx_getprop_7pyverbs_2qp_2QP_qp_num, 0, 0, 0},
  {"qp", __pyx_getprop_7pyverbs_2qp_2QP_qp, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyverbs_2qp_QP_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyverbs_2qp_QP},
  {Py_tp_str, (void *)__pyx_pw_7pyverbs_2qp_2QP_35__str__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyverbs_2qp_QP},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyverbs_2qp_QP},
  {Py_tp_methods, (void *)__pyx_methods_7pyverbs_2qp_QP},
  {Py_tp_getset, (void *)__pyx_getsets_7pyverbs_2qp_QP},
  {Py_tp_init, (void *)__pyx_pw_7pyverbs_2qp_2QP_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7pyverbs_2qp_QP},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyverbs_2qp_QP_spec = {
  "pyverbs.qp.QP",
  sizeof(struct __pyx_obj_7pyverbs_2qp_QP),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyverbs_2qp_QP_slots,
};
#else

static PyTypeObject __pyx_type_7pyverbs_2qp_QP = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyverbs.qp.""QP", /*tp_name*/
  sizeof(struct __pyx_obj_7pyverbs_2qp_QP), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyverbs_2qp_QP, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7pyverbs_2qp_2QP_35__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7pyverbs_2qp_QP, /*tp_traverse*/
  __pyx_tp_clear_7pyverbs_2qp_QP, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyverbs_2qp_QP, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7pyverbs_2qp_QP, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7pyverbs_2qp_2QP_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyverbs_2qp_QP, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7pyverbs_2qp_DataBuffer __pyx_vtable_7pyverbs_2qp_DataBuffer;

static PyObject *__pyx_tp_new_7pyverbs_2qp_DataBuffer(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7pyverbs_2qp_DataBuffer *p;
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7pyverbs_2qp_DataBuffer *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM*)__pyx_vtabptr_7pyverbs_2qp_DataBuffer;
  return o;
}

static void __pyx_tp_dealloc_7pyverbs_2qp_DataBuffer(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyverbs_2qp_DataBuffer) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  if (PyType_IS_GC(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) PyObject_GC_Track(o);
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_dealloc, destructor)(o); else
  #endif
  __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_7pyverbs_2qp_DataBuffer);
}

static int __pyx_tp_traverse_7pyverbs_2qp_DataBuffer(PyObject *o, visitproc v, void *a) {
  int e;
  #if !CYTHON_USE_MODULE_STATE
  e = 0;
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) {
    traverseproc traverse = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_traverse, traverseproc);
    if (traverse) { e = traverse(o, v, a); }
  } else
  #endif
  { e = __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7pyverbs_2qp_DataBuffer); }
  if (e) return e;
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyverbs_2qp_DataBuffer(PyObject *o) {
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) {
    inquiry clear = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_clear, inquiry);
    if (clear) clear(o);
  } else
  #endif
  { __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7pyverbs_2qp_DataBuffer); }
  return 0;
}

static PyMethodDef __pyx_methods_7pyverbs_2qp_DataBuffer[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_10DataBuffer_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_10DataBuffer_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyverbs_2qp_DataBuffer_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyverbs_2qp_DataBuffer},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyverbs_2qp_DataBuffer},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyverbs_2qp_DataBuffer},
  {Py_tp_methods, (void *)__pyx_methods_7pyverbs_2qp_DataBuffer},
  {Py_tp_init, (void *)__pyx_pw_7pyverbs_2qp_10DataBuffer_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7pyverbs_2qp_DataBuffer},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyverbs_2qp_DataBuffer_spec = {
  "pyverbs.qp.DataBuffer",
  sizeof(struct __pyx_obj_7pyverbs_2qp_DataBuffer),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyverbs_2qp_DataBuffer_slots,
};
#else

static PyTypeObject __pyx_type_7pyverbs_2qp_DataBuffer = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyverbs.qp.""DataBuffer", /*tp_name*/
  sizeof(struct __pyx_obj_7pyverbs_2qp_DataBuffer), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyverbs_2qp_DataBuffer, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7pyverbs_2qp_DataBuffer, /*tp_traverse*/
  __pyx_tp_clear_7pyverbs_2qp_DataBuffer, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyverbs_2qp_DataBuffer, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7pyverbs_2qp_10DataBuffer_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyverbs_2qp_DataBuffer, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7pyverbs_2qp_QPEx __pyx_vtable_7pyverbs_2qp_QPEx;

static PyObject *__pyx_tp_new_7pyverbs_2qp_QPEx(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7pyverbs_2qp_QPEx *p;
  PyObject *o = __pyx_tp_new_7pyverbs_2qp_QP(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7pyverbs_2qp_QPEx *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM*)__pyx_vtabptr_7pyverbs_2qp_QPEx;
  p->ind_table = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7pyverbs_2qp_QPEx(PyObject *o) {
  struct __pyx_obj_7pyverbs_2qp_QPEx *p = (struct __pyx_obj_7pyverbs_2qp_QPEx *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyverbs_2qp_QPEx) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->ind_table);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_7pyverbs_2qp_QP(o);
}

static int __pyx_tp_traverse_7pyverbs_2qp_QPEx(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7pyverbs_2qp_QPEx *p = (struct __pyx_obj_7pyverbs_2qp_QPEx *)o;
  e = __pyx_tp_traverse_7pyverbs_2qp_QP(o, v, a); if (e) return e;
  if (p->ind_table) {
    e = (*v)(p->ind_table, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyverbs_2qp_QPEx(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7pyverbs_2qp_QPEx *p = (struct __pyx_obj_7pyverbs_2qp_QPEx *)o;
  __pyx_tp_clear_7pyverbs_2qp_QP(o);
  tmp = ((PyObject*)p->ind_table);
  p->ind_table = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7pyverbs_2qp_4QPEx_comp_mask(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_4QPEx_9comp_mask_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_4QPEx_comp_mask(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_4QPEx_9comp_mask_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_4QPEx_wr_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_4QPEx_5wr_id_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_4QPEx_wr_id(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_4QPEx_5wr_id_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_4QPEx_wr_flags(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_4QPEx_8wr_flags_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_4QPEx_wr_flags(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_4QPEx_8wr_flags_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_4QPEx_ind_table(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_4QPEx_9ind_table_1__get__(o);
}

static PyMethodDef __pyx_methods_7pyverbs_2qp_QPEx[] = {
  {"wr_atomic_cmp_swp", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_3wr_atomic_cmp_swp, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"wr_atomic_fetch_add", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_5wr_atomic_fetch_add, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"wr_bind_mw", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_7wr_bind_mw, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"wr_local_inv", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_9wr_local_inv, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"wr_atomic_write", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_11wr_atomic_write, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"wr_rdma_read", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_13wr_rdma_read, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"wr_rdma_write", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_15wr_rdma_write, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"wr_rdma_write_imm", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_17wr_rdma_write_imm, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"wr_flush", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_19wr_flush, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"wr_send", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_21wr_send, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"wr_send_imm", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_23wr_send_imm, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"wr_send_inv", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_25wr_send_inv, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"wr_send_tso", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_27wr_send_tso, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"wr_set_ud_addr", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_29wr_set_ud_addr, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"wr_set_xrc_srqn", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_31wr_set_xrc_srqn, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"wr_set_inline_data", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_33wr_set_inline_data, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"wr_set_inline_data_list", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_35wr_set_inline_data_list, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"wr_set_sge", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_37wr_set_sge, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"wr_set_sge_list", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_39wr_set_sge_list, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"wr_start", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_41wr_start, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"wr_complete", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_43wr_complete, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"wr_abort", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_45wr_abort, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_47__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_4QPEx_49__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7pyverbs_2qp_QPEx[] = {
  {"comp_mask", __pyx_getprop_7pyverbs_2qp_4QPEx_comp_mask, __pyx_setprop_7pyverbs_2qp_4QPEx_comp_mask, 0, 0},
  {"wr_id", __pyx_getprop_7pyverbs_2qp_4QPEx_wr_id, __pyx_setprop_7pyverbs_2qp_4QPEx_wr_id, 0, 0},
  {"wr_flags", __pyx_getprop_7pyverbs_2qp_4QPEx_wr_flags, __pyx_setprop_7pyverbs_2qp_4QPEx_wr_flags, 0, 0},
  {"ind_table", __pyx_getprop_7pyverbs_2qp_4QPEx_ind_table, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyverbs_2qp_QPEx_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyverbs_2qp_QPEx},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyverbs_2qp_QPEx},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyverbs_2qp_QPEx},
  {Py_tp_methods, (void *)__pyx_methods_7pyverbs_2qp_QPEx},
  {Py_tp_getset, (void *)__pyx_getsets_7pyverbs_2qp_QPEx},
  {Py_tp_init, (void *)__pyx_pw_7pyverbs_2qp_4QPEx_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7pyverbs_2qp_QPEx},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyverbs_2qp_QPEx_spec = {
  "pyverbs.qp.QPEx",
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPEx),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyverbs_2qp_QPEx_slots,
};
#else

static PyTypeObject __pyx_type_7pyverbs_2qp_QPEx = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyverbs.qp.""QPEx", /*tp_name*/
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPEx), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyverbs_2qp_QPEx, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY || 0
  __pyx_pw_7pyverbs_2qp_2QP_35__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7pyverbs_2qp_QPEx, /*tp_traverse*/
  __pyx_tp_clear_7pyverbs_2qp_QPEx, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyverbs_2qp_QPEx, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7pyverbs_2qp_QPEx, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7pyverbs_2qp_4QPEx_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyverbs_2qp_QPEx, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7pyverbs_2qp_ECE __pyx_vtable_7pyverbs_2qp_ECE;

static PyObject *__pyx_tp_new_7pyverbs_2qp_ECE(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7pyverbs_2qp_ECE *p;
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7pyverbs_2qp_ECE *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM*)__pyx_vtabptr_7pyverbs_2qp_ECE;
  return o;
}

static void __pyx_tp_dealloc_7pyverbs_2qp_ECE(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyverbs_2qp_ECE) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  if (PyType_IS_GC(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) PyObject_GC_Track(o);
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_dealloc, destructor)(o); else
  #endif
  __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_7pyverbs_2qp_ECE);
}

static int __pyx_tp_traverse_7pyverbs_2qp_ECE(PyObject *o, visitproc v, void *a) {
  int e;
  #if !CYTHON_USE_MODULE_STATE
  e = 0;
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) {
    traverseproc traverse = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_traverse, traverseproc);
    if (traverse) { e = traverse(o, v, a); }
  } else
  #endif
  { e = __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7pyverbs_2qp_ECE); }
  if (e) return e;
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyverbs_2qp_ECE(PyObject *o) {
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) {
    inquiry clear = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_clear, inquiry);
    if (clear) clear(o);
  } else
  #endif
  { __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7pyverbs_2qp_ECE); }
  return 0;
}

static PyObject *__pyx_getprop_7pyverbs_2qp_3ECE_vendor_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_3ECE_9vendor_id_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_3ECE_vendor_id(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_3ECE_9vendor_id_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_3ECE_options(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_3ECE_7options_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_3ECE_options(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_3ECE_7options_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2qp_3ECE_comp_mask(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2qp_3ECE_9comp_mask_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2qp_3ECE_comp_mask(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2qp_3ECE_9comp_mask_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_7pyverbs_2qp_ECE[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_3ECE_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2qp_3ECE_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7pyverbs_2qp_ECE[] = {
  {"vendor_id", __pyx_getprop_7pyverbs_2qp_3ECE_vendor_id, __pyx_setprop_7pyverbs_2qp_3ECE_vendor_id, 0, 0},
  {"options", __pyx_getprop_7pyverbs_2qp_3ECE_options, __pyx_setprop_7pyverbs_2qp_3ECE_options, 0, 0},
  {"comp_mask", __pyx_getprop_7pyverbs_2qp_3ECE_comp_mask, __pyx_setprop_7pyverbs_2qp_3ECE_comp_mask, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyverbs_2qp_ECE_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyverbs_2qp_ECE},
  {Py_tp_str, (void *)__pyx_pw_7pyverbs_2qp_3ECE_3__str__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyverbs_2qp_ECE},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyverbs_2qp_ECE},
  {Py_tp_methods, (void *)__pyx_methods_7pyverbs_2qp_ECE},
  {Py_tp_getset, (void *)__pyx_getsets_7pyverbs_2qp_ECE},
  {Py_tp_init, (void *)__pyx_pw_7pyverbs_2qp_3ECE_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7pyverbs_2qp_ECE},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyverbs_2qp_ECE_spec = {
  "pyverbs.qp.ECE",
  sizeof(struct __pyx_obj_7pyverbs_2qp_ECE),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyverbs_2qp_ECE_slots,
};
#else

static PyTypeObject __pyx_type_7pyverbs_2qp_ECE = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyverbs.qp.""ECE", /*tp_name*/
  sizeof(struct __pyx_obj_7pyverbs_2qp_ECE), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyverbs_2qp_ECE, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7pyverbs_2qp_3ECE_3__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7pyverbs_2qp_ECE, /*tp_traverse*/
  __pyx_tp_clear_7pyverbs_2qp_ECE, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyverbs_2qp_ECE, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7pyverbs_2qp_ECE, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7pyverbs_2qp_3ECE_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyverbs_2qp_ECE, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
/* #### Code section: initfunc_declarations ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate); /*proto*/
/* #### Code section: init_module ### */

static int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py = Py_None; Py_INCREF(Py_None);
  __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py = Py_None; Py_INCREF(Py_None);
  __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py = Py_None; Py_INCREF(Py_None);
  __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py = Py_None; Py_INCREF(Py_None);
  __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py = Py_None; Py_INCREF(Py_None);
  __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_t_1 = PyImport_ImportModule("pyverbs.base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_4base_PyverbsObject = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.base", "PyverbsObject",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4base_PyverbsObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4base_PyverbsObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4base_PyverbsObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4base_PyverbsObject),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4base_PyverbsObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4base_PyverbsObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4base_PyverbsObject) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPCap = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyverbs_2qp_QPCap_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPCap)) __PYX_ERR(0, 35, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyverbs_2qp_QPCap_spec, __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPCap) < (0)) __PYX_ERR(0, 35, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPCap = &__pyx_type_7pyverbs_2qp_QPCap;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPCap->tp_base = __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPCap) < (0)) __PYX_ERR(0, 35, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPCap);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPCap->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPCap->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPCap->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPCap, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 35, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_7pyverbs_2qp_5QPCap___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_7pyverbs_2qp_5QPCap___init__.doc = __pyx_doc_7pyverbs_2qp_5QPCap___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_7pyverbs_2qp_5QPCap___init__;
    }
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_QPCap, (PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPCap) < (0)) __PYX_ERR(0, 35, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SLOTS
  if (__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPCap->tp_weaklistoffset == 0) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPCap->tp_weaklistoffset = offsetof(struct __pyx_obj_7pyverbs_2qp_QPCap, __pyx_base.__weakref__);
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPCap) < (0)) __PYX_ERR(0, 35, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyverbs_2qp_QPInitAttr_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttr)) __PYX_ERR(0, 101, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyverbs_2qp_QPInitAttr_spec, __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttr) < (0)) __PYX_ERR(0, 101, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttr = &__pyx_type_7pyverbs_2qp_QPInitAttr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPInitAttr->tp_base = __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttr) < (0)) __PYX_ERR(0, 101, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttr);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttr->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttr->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttr, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 101, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_7pyverbs_2qp_10QPInitAttr___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_7pyverbs_2qp_10QPInitAttr___init__.doc = __pyx_doc_7pyverbs_2qp_10QPInitAttr___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_7pyverbs_2qp_10QPInitAttr___init__;
    }
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_QPInitAttr, (PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttr) < (0)) __PYX_ERR(0, 101, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SLOTS
  if (__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttr->tp_weaklistoffset == 0) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttr->tp_weaklistoffset = offsetof(struct __pyx_obj_7pyverbs_2qp_QPInitAttr, __pyx_base.__weakref__);
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttr) < (0)) __PYX_ERR(0, 101, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyverbs_2qp_QPInitAttrEx_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx)) __PYX_ERR(0, 250, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyverbs_2qp_QPInitAttrEx_spec, __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx) < (0)) __PYX_ERR(0, 250, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx = &__pyx_type_7pyverbs_2qp_QPInitAttrEx;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx->tp_base = __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx) < (0)) __PYX_ERR(0, 250, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 250, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_7pyverbs_2qp_12QPInitAttrEx___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_7pyverbs_2qp_12QPInitAttrEx___init__.doc = __pyx_doc_7pyverbs_2qp_12QPInitAttrEx___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_7pyverbs_2qp_12QPInitAttrEx___init__;
    }
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_QPInitAttrEx, (PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx) < (0)) __PYX_ERR(0, 250, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SLOTS
  if (__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx->tp_weaklistoffset == 0) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx->tp_weaklistoffset = offsetof(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx, __pyx_base.__weakref__);
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx) < (0)) __PYX_ERR(0, 250, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPAttr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyverbs_2qp_QPAttr_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPAttr)) __PYX_ERR(0, 510, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyverbs_2qp_QPAttr_spec, __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPAttr) < (0)) __PYX_ERR(0, 510, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPAttr = &__pyx_type_7pyverbs_2qp_QPAttr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPAttr->tp_base = __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPAttr) < (0)) __PYX_ERR(0, 510, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPAttr);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPAttr->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPAttr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPAttr->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPAttr, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 510, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_7pyverbs_2qp_6QPAttr___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_7pyverbs_2qp_6QPAttr___init__.doc = __pyx_doc_7pyverbs_2qp_6QPAttr___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_7pyverbs_2qp_6QPAttr___init__;
    }
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_QPAttr, (PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPAttr) < (0)) __PYX_ERR(0, 510, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SLOTS
  if (__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPAttr->tp_weaklistoffset == 0) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPAttr->tp_weaklistoffset = offsetof(struct __pyx_obj_7pyverbs_2qp_QPAttr, __pyx_base.__weakref__);
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPAttr) < (0)) __PYX_ERR(0, 510, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_4base_PyverbsCM = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.base", "PyverbsCM",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4base_PyverbsCM), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4base_PyverbsCM),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4base_PyverbsCM), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4base_PyverbsCM),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4base_PyverbsCM), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4base_PyverbsCM),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4base_PyverbsCM) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_4base_PyverbsCM = (struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_4base_PyverbsCM); if (unlikely(!__pyx_vtabptr_7pyverbs_4base_PyverbsCM)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2qp_QP = &__pyx_vtable_7pyverbs_2qp_QP;
  __pyx_vtable_7pyverbs_2qp_QP.__pyx_base = *__pyx_vtabptr_7pyverbs_4base_PyverbsCM;
  __pyx_vtable_7pyverbs_2qp_QP.__pyx_base.close = (PyObject *(*)(struct __pyx_obj_7pyverbs_4base_PyverbsCM *, int __pyx_skip_dispatch))__pyx_f_7pyverbs_2qp_2QP_close;
  __pyx_vtable_7pyverbs_2qp_QP.update_cqs = (PyObject *(*)(struct __pyx_obj_7pyverbs_2qp_QP *, PyObject *))__pyx_f_7pyverbs_2qp_2QP_update_cqs;
  __pyx_vtable_7pyverbs_2qp_QP.add_ref = (PyObject *(*)(struct __pyx_obj_7pyverbs_2qp_QP *, PyObject *))__pyx_f_7pyverbs_2qp_2QP_add_ref;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 927, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QP = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyverbs_2qp_QP_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QP)) __PYX_ERR(0, 927, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyverbs_2qp_QP_spec, __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QP) < (0)) __PYX_ERR(0, 927, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QP = &__pyx_type_7pyverbs_2qp_QP;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QP->tp_base = __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QP) < (0)) __PYX_ERR(0, 927, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QP);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QP->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QP->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QP->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QP, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 927, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_7pyverbs_2qp_2QP___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_7pyverbs_2qp_2QP___init__.doc = __pyx_doc_7pyverbs_2qp_2QP___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_7pyverbs_2qp_2QP___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QP, __pyx_vtabptr_7pyverbs_2qp_QP) < (0)) __PYX_ERR(0, 927, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QP) < (0)) __PYX_ERR(0, 927, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_QP, (PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QP) < (0)) __PYX_ERR(0, 927, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SLOTS
  if (__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QP->tp_weaklistoffset == 0) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QP->tp_weaklistoffset = offsetof(struct __pyx_obj_7pyverbs_2qp_QP, __pyx_base.__pyx_base.__weakref__);
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QP) < (0)) __PYX_ERR(0, 927, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2qp_DataBuffer = &__pyx_vtable_7pyverbs_2qp_DataBuffer;
  __pyx_vtable_7pyverbs_2qp_DataBuffer.__pyx_base = *__pyx_vtabptr_7pyverbs_4base_PyverbsCM;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_DataBuffer = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyverbs_2qp_DataBuffer_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_DataBuffer)) __PYX_ERR(0, 1280, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyverbs_2qp_DataBuffer_spec, __pyx_mstate->__pyx_ptype_7pyverbs_2qp_DataBuffer) < (0)) __PYX_ERR(0, 1280, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_DataBuffer = &__pyx_type_7pyverbs_2qp_DataBuffer;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_DataBuffer->tp_base = __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_DataBuffer) < (0)) __PYX_ERR(0, 1280, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7pyverbs_2qp_DataBuffer);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_DataBuffer->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyverbs_2qp_DataBuffer->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyverbs_2qp_DataBuffer->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_DataBuffer, __pyx_vtabptr_7pyverbs_2qp_DataBuffer) < (0)) __PYX_ERR(0, 1280, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_DataBuffer) < (0)) __PYX_ERR(0, 1280, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_DataBuffer, (PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_DataBuffer) < (0)) __PYX_ERR(0, 1280, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SLOTS
  if (__pyx_mstate->__pyx_ptype_7pyverbs_2qp_DataBuffer->tp_weaklistoffset == 0) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_DataBuffer->tp_weaklistoffset = offsetof(struct __pyx_obj_7pyverbs_2qp_DataBuffer, __pyx_base.__pyx_base.__weakref__);
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_DataBuffer) < (0)) __PYX_ERR(0, 1280, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2qp_QPEx = &__pyx_vtable_7pyverbs_2qp_QPEx;
  __pyx_vtable_7pyverbs_2qp_QPEx.__pyx_base = *__pyx_vtabptr_7pyverbs_2qp_QP;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QP); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPEx = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyverbs_2qp_QPEx_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPEx)) __PYX_ERR(0, 1287, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyverbs_2qp_QPEx_spec, __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPEx) < (0)) __PYX_ERR(0, 1287, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPEx = &__pyx_type_7pyverbs_2qp_QPEx;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx->tp_base = __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QP;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPEx) < (0)) __PYX_ERR(0, 1287, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPEx);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPEx->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPEx->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPEx->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPEx, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 1287, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_7pyverbs_2qp_4QPEx___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_7pyverbs_2qp_4QPEx___init__.doc = __pyx_doc_7pyverbs_2qp_4QPEx___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_7pyverbs_2qp_4QPEx___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPEx, __pyx_vtabptr_7pyverbs_2qp_QPEx) < (0)) __PYX_ERR(0, 1287, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPEx) < (0)) __PYX_ERR(0, 1287, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_QPEx, (PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPEx) < (0)) __PYX_ERR(0, 1287, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SLOTS
  if (__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPEx->tp_weaklistoffset == 0) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPEx->tp_weaklistoffset = offsetof(struct __pyx_obj_7pyverbs_2qp_QPEx, __pyx_base.__pyx_base.__pyx_base.__weakref__);
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPEx) < (0)) __PYX_ERR(0, 1287, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2qp_ECE = &__pyx_vtable_7pyverbs_2qp_ECE;
  __pyx_vtable_7pyverbs_2qp_ECE.__pyx_base = *__pyx_vtabptr_7pyverbs_4base_PyverbsCM;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_ECE = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyverbs_2qp_ECE_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_ECE)) __PYX_ERR(0, 884, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyverbs_2qp_ECE_spec, __pyx_mstate->__pyx_ptype_7pyverbs_2qp_ECE) < (0)) __PYX_ERR(0, 884, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_ECE = &__pyx_type_7pyverbs_2qp_ECE;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_ECE->tp_base = __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_ECE) < (0)) __PYX_ERR(0, 884, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7pyverbs_2qp_ECE);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_ECE->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyverbs_2qp_ECE->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyverbs_2qp_ECE->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_mstate->__pyx_ptype_7pyverbs_2qp_ECE, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 884, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_7pyverbs_2qp_3ECE___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_7pyverbs_2qp_3ECE___init__.doc = __pyx_doc_7pyverbs_2qp_3ECE___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_7pyverbs_2qp_3ECE___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_ECE, __pyx_vtabptr_7pyverbs_2qp_ECE) < (0)) __PYX_ERR(0, 884, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_ECE) < (0)) __PYX_ERR(0, 884, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_ECE_2, (PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_ECE) < (0)) __PYX_ERR(0, 884, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SLOTS
  if (__pyx_mstate->__pyx_ptype_7pyverbs_2qp_ECE->tp_weaklistoffset == 0) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_ECE->tp_weaklistoffset = offsetof(struct __pyx_obj_7pyverbs_2qp_ECE, __pyx_base.__pyx_base.__weakref__);
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2qp_ECE) < (0)) __PYX_ERR(0, 884, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("pyverbs.device"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_Context = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "Context",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_Context), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_Context),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_Context), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_Context),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_Context), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_Context),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_Context) __PYX_ERR(2, 10, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_6device_Context = (struct __pyx_vtabstruct_7pyverbs_6device_Context*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_6device_Context); if (unlikely(!__pyx_vtabptr_7pyverbs_6device_Context)) __PYX_ERR(2, 10, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_DeviceAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "DeviceAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_DeviceAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DeviceAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_DeviceAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DeviceAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_DeviceAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DeviceAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_DeviceAttr) __PYX_ERR(2, 33, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_QueryDeviceExInput = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "QueryDeviceExInput",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_QueryDeviceExInput) __PYX_ERR(2, 36, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_ODPCaps = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "ODPCaps",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_ODPCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_ODPCaps),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_ODPCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_ODPCaps),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_ODPCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_ODPCaps),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_ODPCaps) __PYX_ERR(2, 39, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_RSSCaps = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "RSSCaps",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_RSSCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_RSSCaps),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_RSSCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_RSSCaps),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_RSSCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_RSSCaps),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_RSSCaps) __PYX_ERR(2, 43, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_PacketPacingCaps = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "PacketPacingCaps",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_PacketPacingCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PacketPacingCaps),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_PacketPacingCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PacketPacingCaps),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_PacketPacingCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PacketPacingCaps),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_PacketPacingCaps) __PYX_ERR(2, 46, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_PCIAtomicCaps = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "PCIAtomicCaps",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_PCIAtomicCaps) __PYX_ERR(2, 49, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_TMCaps = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "TMCaps",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_TMCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_TMCaps),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_TMCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_TMCaps),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_TMCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_TMCaps),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_TMCaps) __PYX_ERR(2, 52, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_CQModerationCaps = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "CQModerationCaps",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_CQModerationCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_CQModerationCaps),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_CQModerationCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_CQModerationCaps),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_CQModerationCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_CQModerationCaps),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_CQModerationCaps) __PYX_ERR(2, 55, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_TSOCaps = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "TSOCaps",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_TSOCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_TSOCaps),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_TSOCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_TSOCaps),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_TSOCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_TSOCaps),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_TSOCaps) __PYX_ERR(2, 58, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_DeviceAttrEx = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "DeviceAttrEx",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_DeviceAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DeviceAttrEx),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_DeviceAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DeviceAttrEx),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_DeviceAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DeviceAttrEx),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_DeviceAttrEx) __PYX_ERR(2, 61, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_AllocDmAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "AllocDmAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_AllocDmAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_AllocDmAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_AllocDmAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_AllocDmAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_AllocDmAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_AllocDmAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_AllocDmAttr) __PYX_ERR(2, 64, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_DM = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "DM",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_DM), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DM),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_DM), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DM),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_DM), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DM),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_DM) __PYX_ERR(2, 67, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_6device_DM = (struct __pyx_vtabstruct_7pyverbs_6device_DM*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_6device_DM); if (unlikely(!__pyx_vtabptr_7pyverbs_6device_DM)) __PYX_ERR(2, 67, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_PortAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "PortAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_PortAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PortAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_PortAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PortAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_PortAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PortAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_PortAttr) __PYX_ERR(2, 74, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_GIDEntry = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "GIDEntry",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_GIDEntry), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_GIDEntry),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_GIDEntry), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_GIDEntry),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_GIDEntry), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_GIDEntry),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_GIDEntry) __PYX_ERR(2, 77, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_AsyncEvent = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "AsyncEvent",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_AsyncEvent), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_AsyncEvent),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_AsyncEvent), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_AsyncEvent),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_AsyncEvent), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_AsyncEvent),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_AsyncEvent) __PYX_ERR(2, 80, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_FdArr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "FdArr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_FdArr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_FdArr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_FdArr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_FdArr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_FdArr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_FdArr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_FdArr) __PYX_ERR(2, 83, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_6device_FdArr = (struct __pyx_vtabstruct_7pyverbs_6device_FdArr*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_6device_FdArr); if (unlikely(!__pyx_vtabptr_7pyverbs_6device_FdArr)) __PYX_ERR(2, 83, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyverbs.cq"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CompChannel = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.cq", "CompChannel",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2cq_CompChannel), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CompChannel),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2cq_CompChannel), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CompChannel),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2cq_CompChannel), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CompChannel),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CompChannel) __PYX_ERR(3, 9, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2cq_CompChannel = (struct __pyx_vtabstruct_7pyverbs_2cq_CompChannel*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CompChannel); if (unlikely(!__pyx_vtabptr_7pyverbs_2cq_CompChannel)) __PYX_ERR(3, 9, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQ = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.cq", "CQ",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2cq_CQ), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CQ),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2cq_CQ), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CQ),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2cq_CQ), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CQ),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQ) __PYX_ERR(3, 16, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2cq_CQ = (struct __pyx_vtabstruct_7pyverbs_2cq_CQ*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQ); if (unlikely(!__pyx_vtabptr_7pyverbs_2cq_CQ)) __PYX_ERR(3, 16, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.cq", "CqInitAttrEx",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx) __PYX_ERR(3, 27, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQEX = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.cq", "CQEX",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2cq_CQEX), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CQEX),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2cq_CQEX), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CQEX),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2cq_CQEX), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CQEX),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQEX) __PYX_ERR(3, 32, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2cq_CQEX = (struct __pyx_vtabstruct_7pyverbs_2cq_CQEX*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQEX); if (unlikely(!__pyx_vtabptr_7pyverbs_2cq_CQEX)) __PYX_ERR(3, 32, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_WC = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.cq", "WC",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2cq_WC), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_WC),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2cq_WC), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_WC),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2cq_WC), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_WC),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_WC) __PYX_ERR(3, 42, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_PollCqAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.cq", "PollCqAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2cq_PollCqAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_PollCqAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2cq_PollCqAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_PollCqAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2cq_PollCqAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_PollCqAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_PollCqAttr) __PYX_ERR(3, 45, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_WcTmInfo = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.cq", "WcTmInfo",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2cq_WcTmInfo), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_WcTmInfo),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2cq_WcTmInfo), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_WcTmInfo),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2cq_WcTmInfo), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_WcTmInfo),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_WcTmInfo) __PYX_ERR(3, 48, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyverbs.pd"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_2pd_PD = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.pd", "PD",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2pd_PD), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_PD),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2pd_PD), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_PD),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2pd_PD), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_PD),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2pd_PD) __PYX_ERR(4, 12, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2pd_PD = (struct __pyx_vtabstruct_7pyverbs_2pd_PD*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2pd_PD); if (unlikely(!__pyx_vtabptr_7pyverbs_2pd_PD)) __PYX_ERR(4, 12, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2pd_ParentDomainInitAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.pd", "ParentDomainInitAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2pd_ParentDomainInitAttr) __PYX_ERR(4, 28, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2pd_ParentDomain = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.pd", "ParentDomain",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomain), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomain),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomain), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomain),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomain), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomain),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2pd_ParentDomain) __PYX_ERR(4, 34, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2pd_ParentDomain = (struct __pyx_vtabstruct_7pyverbs_2pd_ParentDomain*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2pd_ParentDomain); if (unlikely(!__pyx_vtabptr_7pyverbs_2pd_ParentDomain)) __PYX_ERR(4, 34, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2pd_ParentDomainContext = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.pd", "ParentDomainContext",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomainContext), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomainContext),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomainContext), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomainContext),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomainContext), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomainContext),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2pd_ParentDomainContext) __PYX_ERR(4, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyverbs.wq"); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_2wq_WQInitAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.wq", "WQInitAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2wq_WQInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_WQInitAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2wq_WQInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_WQInitAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2wq_WQInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_WQInitAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2wq_WQInitAttr) __PYX_ERR(5, 13, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2wq_WQAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.wq", "WQAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2wq_WQAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_WQAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2wq_WQAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_WQAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2wq_WQAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_WQAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2wq_WQAttr) __PYX_ERR(5, 18, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2wq_WQ = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.wq", "WQ",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2wq_WQ), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_WQ),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2wq_WQ), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_WQ),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2wq_WQ), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_WQ),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2wq_WQ) __PYX_ERR(5, 21, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2wq_WQ = (struct __pyx_vtabstruct_7pyverbs_2wq_WQ*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2wq_WQ); if (unlikely(!__pyx_vtabptr_7pyverbs_2wq_WQ)) __PYX_ERR(5, 21, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2wq_RwqIndTableInitAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.wq", "RwqIndTableInitAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2wq_RwqIndTableInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_RwqIndTableInitAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2wq_RwqIndTableInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_RwqIndTableInitAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2wq_RwqIndTableInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_RwqIndTableInitAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2wq_RwqIndTableInitAttr) __PYX_ERR(5, 29, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2wq_RwqIndTable = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.wq", "RwqIndTable",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2wq_RwqIndTable), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_RwqIndTable),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2wq_RwqIndTable), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_RwqIndTable),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2wq_RwqIndTable), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_RwqIndTable),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2wq_RwqIndTable) __PYX_ERR(5, 33, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2wq_RwqIndTable = (struct __pyx_vtabstruct_7pyverbs_2wq_RwqIndTable*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2wq_RwqIndTable); if (unlikely(!__pyx_vtabptr_7pyverbs_2wq_RwqIndTable)) __PYX_ERR(5, 33, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2wq_RxHashConf = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.wq", "RxHashConf",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2wq_RxHashConf), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_RxHashConf),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2wq_RxHashConf), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_RxHashConf),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2wq_RxHashConf), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_RxHashConf),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2wq_RxHashConf) __PYX_ERR(5, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyverbs.mr"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_2mr_MR = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.mr", "MR",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2mr_MR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MR),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2mr_MR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MR),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2mr_MR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MR),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2mr_MR) __PYX_ERR(6, 12, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2mr_MR = (struct __pyx_vtabstruct_7pyverbs_2mr_MR*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2mr_MR); if (unlikely(!__pyx_vtabptr_7pyverbs_2mr_MR)) __PYX_ERR(6, 12, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2mr_MREx = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.mr", "MREx",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2mr_MREx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MREx),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2mr_MREx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MREx),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2mr_MREx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MREx),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2mr_MREx) __PYX_ERR(6, 25, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2mr_MREx = (struct __pyx_vtabstruct_7pyverbs_2mr_MREx*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2mr_MREx); if (unlikely(!__pyx_vtabptr_7pyverbs_2mr_MREx)) __PYX_ERR(6, 25, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2mr_MWBindInfo = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.mr", "MWBindInfo",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2mr_MWBindInfo), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MWBindInfo),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2mr_MWBindInfo), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MWBindInfo),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2mr_MWBindInfo), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MWBindInfo),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2mr_MWBindInfo) __PYX_ERR(6, 28, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2mr_MWBindInfo = (struct __pyx_vtabstruct_7pyverbs_2mr_MWBindInfo*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2mr_MWBindInfo); if (unlikely(!__pyx_vtabptr_7pyverbs_2mr_MWBindInfo)) __PYX_ERR(6, 28, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2mr_MWBind = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.mr", "MWBind",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2mr_MWBind), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MWBind),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2mr_MWBind), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MWBind),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2mr_MWBind), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MWBind),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2mr_MWBind) __PYX_ERR(6, 32, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2mr_MWBind = (struct __pyx_vtabstruct_7pyverbs_2mr_MWBind*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2mr_MWBind); if (unlikely(!__pyx_vtabptr_7pyverbs_2mr_MWBind)) __PYX_ERR(6, 32, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2mr_MW = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.mr", "MW",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2mr_MW), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MW),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2mr_MW), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MW),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2mr_MW), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MW),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2mr_MW) __PYX_ERR(6, 36, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2mr_MW = (struct __pyx_vtabstruct_7pyverbs_2mr_MW*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2mr_MW); if (unlikely(!__pyx_vtabptr_7pyverbs_2mr_MW)) __PYX_ERR(6, 36, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2mr_DMMR = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.mr", "DMMR",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2mr_DMMR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DMMR),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2mr_DMMR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DMMR),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2mr_DMMR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DMMR),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2mr_DMMR) __PYX_ERR(6, 40, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2mr_DMMR = (struct __pyx_vtabstruct_7pyverbs_2mr_DMMR*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2mr_DMMR); if (unlikely(!__pyx_vtabptr_7pyverbs_2mr_DMMR)) __PYX_ERR(6, 40, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2mr_DmaBufMR = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.mr", "DmaBufMR",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2mr_DmaBufMR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DmaBufMR),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2mr_DmaBufMR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DmaBufMR),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2mr_DmaBufMR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DmaBufMR),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2mr_DmaBufMR) __PYX_ERR(6, 43, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2mr_DmaBufMR = (struct __pyx_vtabstruct_7pyverbs_2mr_DmaBufMR*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2mr_DmaBufMR); if (unlikely(!__pyx_vtabptr_7pyverbs_2mr_DmaBufMR)) __PYX_ERR(6, 43, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2mr_DmaHandleInitAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.mr", "DmaHandleInitAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2mr_DmaHandleInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DmaHandleInitAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2mr_DmaHandleInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DmaHandleInitAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2mr_DmaHandleInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DmaHandleInitAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2mr_DmaHandleInitAttr) __PYX_ERR(6, 48, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2mr_DMAHandle = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.mr", "DMAHandle",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2mr_DMAHandle), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DMAHandle),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2mr_DMAHandle), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DMAHandle),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2mr_DMAHandle), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DMAHandle),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2mr_DMAHandle) __PYX_ERR(6, 51, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2mr_DMAHandle = (struct __pyx_vtabstruct_7pyverbs_2mr_DMAHandle*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2mr_DMAHandle); if (unlikely(!__pyx_vtabptr_7pyverbs_2mr_DMAHandle)) __PYX_ERR(6, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyverbs.wr"); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_2wr_SGE = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.wr", "SGE",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2wr_SGE), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wr_SGE),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2wr_SGE), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wr_SGE),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2wr_SGE), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wr_SGE),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2wr_SGE) __PYX_ERR(7, 10, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2wr_SGE = (struct __pyx_vtabstruct_7pyverbs_2wr_SGE*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2wr_SGE); if (unlikely(!__pyx_vtabptr_7pyverbs_2wr_SGE)) __PYX_ERR(7, 10, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2wr_RecvWR = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.wr", "RecvWR",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2wr_RecvWR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wr_RecvWR),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2wr_RecvWR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wr_RecvWR),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2wr_RecvWR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wr_RecvWR),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2wr_RecvWR) __PYX_ERR(7, 14, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2wr_RecvWR = (struct __pyx_vtabstruct_7pyverbs_2wr_RecvWR*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2wr_RecvWR); if (unlikely(!__pyx_vtabptr_7pyverbs_2wr_RecvWR)) __PYX_ERR(7, 14, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2wr_SendWR = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.wr", "SendWR",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2wr_SendWR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wr_SendWR),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2wr_SendWR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wr_SendWR),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2wr_SendWR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wr_SendWR),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2wr_SendWR) __PYX_ERR(7, 17, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2wr_SendWR = (struct __pyx_vtabstruct_7pyverbs_2wr_SendWR*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2wr_SendWR); if (unlikely(!__pyx_vtabptr_7pyverbs_2wr_SendWR)) __PYX_ERR(7, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyverbs.cmid"); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMID = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.cmid", "CMID",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4cmid_CMID), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4cmid_CMID),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4cmid_CMID), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4cmid_CMID),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4cmid_CMID), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4cmid_CMID),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMID) __PYX_ERR(8, 10, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_4cmid_CMID = (struct __pyx_vtabstruct_7pyverbs_4cmid_CMID*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMID); if (unlikely(!__pyx_vtabptr_7pyverbs_4cmid_CMID)) __PYX_ERR(8, 10, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEventChannel = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.cmid", "CMEventChannel",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4cmid_CMEventChannel), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4cmid_CMEventChannel),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4cmid_CMEventChannel), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4cmid_CMEventChannel),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4cmid_CMEventChannel), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4cmid_CMEventChannel),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEventChannel) __PYX_ERR(8, 20, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_4cmid_CMEventChannel = (struct __pyx_vtabstruct_7pyverbs_4cmid_CMEventChannel*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEventChannel); if (unlikely(!__pyx_vtabptr_7pyverbs_4cmid_CMEventChannel)) __PYX_ERR(8, 20, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEvent = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.cmid", "CMEvent",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4cmid_CMEvent), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4cmid_CMEvent),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4cmid_CMEvent), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4cmid_CMEvent),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4cmid_CMEvent), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4cmid_CMEvent),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEvent) __PYX_ERR(8, 25, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_4cmid_CMEvent = (struct __pyx_vtabstruct_7pyverbs_4cmid_CMEvent*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEvent); if (unlikely(!__pyx_vtabptr_7pyverbs_4cmid_CMEvent)) __PYX_ERR(8, 25, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_AddrInfo = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.cmid", "AddrInfo",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4cmid_AddrInfo), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4cmid_AddrInfo),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4cmid_AddrInfo), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4cmid_AddrInfo),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4cmid_AddrInfo), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4cmid_AddrInfo),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_AddrInfo) __PYX_ERR(8, 30, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_4cmid_AddrInfo = (struct __pyx_vtabstruct_7pyverbs_4cmid_AddrInfo*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_AddrInfo); if (unlikely(!__pyx_vtabptr_7pyverbs_4cmid_AddrInfo)) __PYX_ERR(8, 30, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_ConnParam = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.cmid", "ConnParam",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4cmid_ConnParam), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4cmid_ConnParam),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4cmid_ConnParam), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4cmid_ConnParam),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4cmid_ConnParam), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4cmid_ConnParam),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_ConnParam) __PYX_ERR(8, 35, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_UDParam = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.cmid", "UDParam",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4cmid_UDParam), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4cmid_UDParam),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4cmid_UDParam), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4cmid_UDParam),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4cmid_UDParam), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4cmid_UDParam),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_UDParam) __PYX_ERR(8, 40, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.cmid", "JoinMCAttrEx",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx) __PYX_ERR(8, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyverbs.addr"); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_4addr_GID = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.addr", "GID",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4addr_GID), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_GID),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4addr_GID), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_GID),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4addr_GID), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_GID),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4addr_GID) __PYX_ERR(9, 11, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_4addr_GRH = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.addr", "GRH",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4addr_GRH), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_GRH),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4addr_GRH), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_GRH),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4addr_GRH), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_GRH),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4addr_GRH) __PYX_ERR(9, 14, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_4addr_GlobalRoute = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.addr", "GlobalRoute",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4addr_GlobalRoute), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_GlobalRoute),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4addr_GlobalRoute), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_GlobalRoute),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4addr_GlobalRoute), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_GlobalRoute),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4addr_GlobalRoute) __PYX_ERR(9, 17, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_4addr_AHAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.addr", "AHAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4addr_AHAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_AHAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4addr_AHAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_AHAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4addr_AHAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_AHAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4addr_AHAttr) __PYX_ERR(9, 20, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_4addr_AHAttr = (struct __pyx_vtabstruct_7pyverbs_4addr_AHAttr*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_4addr_AHAttr); if (unlikely(!__pyx_vtabptr_7pyverbs_4addr_AHAttr)) __PYX_ERR(9, 20, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_4addr_AH = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.addr", "AH",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4addr_AH), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_AH),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4addr_AH), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_AH),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4addr_AH), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_AH),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4addr_AH) __PYX_ERR(9, 24, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_4addr_AH = (struct __pyx_vtabstruct_7pyverbs_4addr_AH*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_4addr_AH); if (unlikely(!__pyx_vtabptr_7pyverbs_4addr_AH)) __PYX_ERR(9, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyverbs.flow"); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_4flow_FlowAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.flow", "FlowAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4flow_FlowAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4flow_FlowAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4flow_FlowAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4flow_FlowAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4flow_FlowAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4flow_FlowAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4flow_FlowAttr) __PYX_ERR(10, 9, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_4flow_Flow = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.flow", "Flow",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4flow_Flow), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4flow_Flow),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4flow_Flow), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4flow_Flow),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4flow_Flow), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4flow_Flow),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4flow_Flow) __PYX_ERR(10, 13, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_4flow_Flow = (struct __pyx_vtabstruct_7pyverbs_4flow_Flow*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_4flow_Flow); if (unlikely(!__pyx_vtabptr_7pyverbs_4flow_Flow)) __PYX_ERR(10, 13, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_4flow_FlowAction = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.flow", "FlowAction",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4flow_FlowAction), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4flow_FlowAction),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4flow_FlowAction), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4flow_FlowAction),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4flow_FlowAction), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4flow_FlowAction),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4flow_FlowAction) __PYX_ERR(10, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_2_4(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  0, 0,
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7cpython_4type_type) __PYX_ERR(11, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyverbs.xrcd"); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_4xrcd_XRCDInitAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.xrcd", "XRCDInitAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4xrcd_XRCDInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4xrcd_XRCDInitAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4xrcd_XRCDInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4xrcd_XRCDInitAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4xrcd_XRCDInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4xrcd_XRCDInitAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4xrcd_XRCDInitAttr) __PYX_ERR(12, 11, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_4xrcd_XRCD = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.xrcd", "XRCD",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4xrcd_XRCD), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4xrcd_XRCD),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4xrcd_XRCD), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4xrcd_XRCD),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4xrcd_XRCD), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4xrcd_XRCD),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4xrcd_XRCD) __PYX_ERR(12, 15, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_4xrcd_XRCD = (struct __pyx_vtabstruct_7pyverbs_4xrcd_XRCD*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_4xrcd_XRCD); if (unlikely(!__pyx_vtabptr_7pyverbs_4xrcd_XRCD)) __PYX_ERR(12, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyverbs.srq"); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_3srq_SrqAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.srq", "SrqAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_3srq_SrqAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SrqAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_3srq_SrqAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SrqAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_3srq_SrqAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SrqAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_3srq_SrqAttr) __PYX_ERR(13, 9, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_3srq_SrqInitAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.srq", "SrqInitAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_3srq_SrqInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SrqInitAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_3srq_SrqInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SrqInitAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_3srq_SrqInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SrqInitAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_3srq_SrqInitAttr) __PYX_ERR(13, 12, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_3srq_SrqInitAttrEx = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.srq", "SrqInitAttrEx",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_3srq_SrqInitAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SrqInitAttrEx),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_3srq_SrqInitAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SrqInitAttrEx),
  #else
  sizeof(struct __pyx_obj_7pyverbs_3srq_SrqInitAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SrqInitAttrEx),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_3srq_SrqInitAttrEx) __PYX_ERR(13, 15, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_3srq_OpsWr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.srq", "OpsWr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_3srq_OpsWr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_OpsWr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_3srq_OpsWr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_OpsWr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_3srq_OpsWr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_OpsWr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_3srq_OpsWr) __PYX_ERR(13, 21, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_3srq_OpsWr = (struct __pyx_vtabstruct_7pyverbs_3srq_OpsWr*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_3srq_OpsWr); if (unlikely(!__pyx_vtabptr_7pyverbs_3srq_OpsWr)) __PYX_ERR(13, 21, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_3srq_SRQ = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.srq", "SRQ",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_3srq_SRQ), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SRQ),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_3srq_SRQ), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SRQ),
  #else
  sizeof(struct __pyx_obj_7pyverbs_3srq_SRQ), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SRQ),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_3srq_SRQ) __PYX_ERR(13, 24, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_3srq_SRQ = (struct __pyx_vtabstruct_7pyverbs_3srq_SRQ*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_3srq_SRQ); if (unlikely(!__pyx_vtabptr_7pyverbs_3srq_SRQ)) __PYX_ERR(13, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  {
    __pyx_t_1 = PyImport_ImportModule("pyverbs.base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    const char * __pyx_import_signature = __Pyx_PyBytes_AsString(__pyx_mstate_global->__pyx_kp_b_PyObject_PyObject_close_weakrefs);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (unlikely(!__pyx_import_signature)) __PYX_ERR(0, 1, __pyx_L1_error)
    #endif
    const char * __pyx_import_name = __pyx_import_signature + 23;
    void (**const __pyx_import_pointers[])(void) = {(void (**)(void))&__pyx_f_7pyverbs_4base_close_weakrefs, (void (**)(void)) NULL};
    void (**const *__pyx_import_pointer)(void) = __pyx_import_pointers;
    const char *__pyx_import_current_signature = __pyx_import_signature;
    while (*__pyx_import_pointer) {
      if (__Pyx_ImportFunction_3_2_4(__pyx_t_1, __pyx_import_name, *__pyx_import_pointer, __pyx_import_current_signature) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
      ++__pyx_import_pointer;
      __pyx_import_name = strchr(__pyx_import_name, '\0') + 1;
      __pyx_import_signature = strchr(__pyx_import_signature, '\0') + 1;
      if (*__pyx_import_signature != '\0') __pyx_import_current_signature = __pyx_import_signature;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_qp(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_qp},
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  {Py_mod_gil, __Pyx_FREETHREADING_COMPATIBLE},
  #endif
  #if PY_VERSION_HEX >= 0x030C0000 && CYTHON_USE_MODULE_STATE
  {Py_mod_multiple_interpreters, Py_MOD_MULTIPLE_INTERPRETERS_NOT_SUPPORTED},
  #endif
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "qp",
      0, /* m_doc */
    #if CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstatetype), /* m_size */
    #else
      (CYTHON_PEP489_MULTI_PHASE_INIT) ? 0 : -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif

/* PyModInitFuncType */
#ifndef CYTHON_NO_PYINIT_EXPORT
  #define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#else
  #ifdef __cplusplus
  #define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
  #else
  #define __Pyx_PyMODINIT_FUNC PyObject *
  #endif
#endif

__Pyx_PyMODINIT_FUNC PyInit_qp(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_qp(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
/* ModuleCreationPEP489 */
#if CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
static PY_INT64_T __Pyx_GetCurrentInterpreterId(void) {
    {
        PyObject *module = PyImport_ImportModule("_interpreters"); // 3.13+ I think
        if (!module) {
            PyErr_Clear(); // just try the 3.8-3.12 version
            module = PyImport_ImportModule("_xxsubinterpreters");
            if (!module) goto bad;
        }
        PyObject *current = PyObject_CallMethod(module, "get_current", NULL);
        Py_DECREF(module);
        if (!current) goto bad;
        if (PyTuple_Check(current)) {
            PyObject *new_current = PySequence_GetItem(current, 0);
            Py_DECREF(current);
            current = new_current;
            if (!new_current) goto bad;
        }
        long long as_c_int = PyLong_AsLongLong(current);
        Py_DECREF(current);
        return as_c_int;
    }
  bad:
    PySys_WriteStderr("__Pyx_GetCurrentInterpreterId failed. Try setting the C define CYTHON_PEP489_MULTI_PHASE_INIT=0\n");
    return -1;
}
#endif
#if !CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    static PY_INT64_T main_interpreter_id = -1;
#if CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
    PY_INT64_T current_id = GraalPyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_GRAAL
    PY_INT64_T current_id = PyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
    PY_INT64_T current_id = __Pyx_GetCurrentInterpreterId();
#elif CYTHON_COMPILING_IN_LIMITED_API
    PY_INT64_T current_id = PyInterpreterState_GetID(PyInterpreterState_Get());
#else
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
#endif
    if (unlikely(current_id == -1)) {
        return -1;
    }
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return 0;
    } else if (unlikely(main_interpreter_id != current_id)) {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#endif
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    #if !CYTHON_USE_MODULE_STATE
    if (__Pyx_check_single_interpreter())
        return NULL;
    #endif
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_qp(PyObject *__pyx_pyinit_module)
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  __pyx_mstatetype *__pyx_mstate = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'qp' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #else
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_t_1 = __pyx_pyinit_module;
  Py_INCREF(__pyx_t_1);
  #else
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #if CYTHON_USE_MODULE_STATE
  {
    int add_module_result = __Pyx_State_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "qp" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = __pyx_t_1;
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  PyUnstable_Module_SetGIL(__pyx_m, Py_MOD_GIL_USED);
  #endif
  __pyx_mstate = __pyx_mstate_global;
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_mstate->__pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_mstate->__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_mstate->__pyx_d);
  __pyx_mstate->__pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_mstate->__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_cython_runtime = __Pyx_PyImport_AddModuleRef("cython_runtime"); if (unlikely(!__pyx_mstate->__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_mstate->__pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /* ImportRefnannyAPI */
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
    PyErr_Clear();
    __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
    if (!__Pyx_RefNanny)
        Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  
__Pyx_RefNannySetupContext("PyInit_qp", 0);
  __Pyx_init_runtime_version();
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_mstate->__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Library function declarations ---*/
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__pyx_module_is_main_pyverbs__qp) {
    if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_name, __pyx_mstate_global->__pyx_n_u_main) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "pyverbs.qp")) {
      if (unlikely((PyDict_SetItemString(modules, "pyverbs.qp", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_CreateCodeObjects(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_export_code(__pyx_mstate);
  (void)__Pyx_modinit_function_export_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_type_init_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_function_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/

  /* "EnumTypeToPy":1
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py = None             # <<<<<<<<<<<<<<
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py")
*/
  __Pyx_INCREF(Py_None);
  __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py);
  __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py, Py_None);
  __Pyx_GIVEREF(Py_None);

  /* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(ibv_qp_type c_val):
 *     cdef object __pyx_enum
*/
  __Pyx_INCREF(Py_None);
  __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py);
  __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py, Py_None);
  __Pyx_GIVEREF(Py_None);

  /* "EnumTypeToPy":1
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py = None             # <<<<<<<<<<<<<<
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py")
*/
  __Pyx_INCREF(Py_None);
  __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py);
  __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py, Py_None);
  __Pyx_GIVEREF(Py_None);

  /* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py(ibv_mtu c_val):
 *     cdef object __pyx_enum
*/
  __Pyx_INCREF(Py_None);
  __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py);
  __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mig_state_to_py, Py_None);
  __Pyx_GIVEREF(Py_None);

  /* "EnumTypeToPy":1
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py = None             # <<<<<<<<<<<<<<
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py")
*/
  __Pyx_INCREF(Py_None);
  __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py);
  __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py, Py_None);
  __Pyx_GIVEREF(Py_None);

  /* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_attr_mask_to_py(ibv_qp_attr_mask c_val):
 *     cdef object __pyx_enum
*/
  __Pyx_INCREF(Py_None);
  __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py);
  __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_init_attr_mask_to_py, Py_None);
  __Pyx_GIVEREF(Py_None);

  /* "pyverbs/qp.pyx":7
 * from libc.stdint cimport uintptr_t
 * from libc.string cimport memcpy
 * import weakref             # <<<<<<<<<<<<<<
 * 
 * from pyverbs.pyverbs_error import PyverbsUserError, PyverbsError, PyverbsRDMAError
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_weakref, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_weakref, __pyx_t_2) < (0)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":9
 * import weakref
 * 
 * from pyverbs.pyverbs_error import PyverbsUserError, PyverbsError, PyverbsRDMAError             # <<<<<<<<<<<<<<
 * from pyverbs.utils import gid_str, qp_type_to_str, qp_state_to_str, mtu_to_str
 * from pyverbs.utils import access_flags_to_str, mig_state_to_str
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_PyverbsUserError,__pyx_mstate_global->__pyx_n_u_PyverbsError,__pyx_mstate_global->__pyx_n_u_PyverbsRDMAError};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyverbs_pyverbs_error, __pyx_imported_names, 3, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_PyverbsUserError,__pyx_mstate_global->__pyx_n_u_PyverbsError,__pyx_mstate_global->__pyx_n_u_PyverbsRDMAError};
    for (__pyx_t_3=0; __pyx_t_3 < 3; __pyx_t_3++) {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 9, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 9, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":10
 * 
 * from pyverbs.pyverbs_error import PyverbsUserError, PyverbsError, PyverbsRDMAError
 * from pyverbs.utils import gid_str, qp_type_to_str, qp_state_to_str, mtu_to_str             # <<<<<<<<<<<<<<
 * from pyverbs.utils import access_flags_to_str, mig_state_to_str
 * from pyverbs.wq cimport RwqIndTable, RxHashConf
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_gid_str,__pyx_mstate_global->__pyx_n_u_qp_type_to_str,__pyx_mstate_global->__pyx_n_u_qp_state_to_str,__pyx_mstate_global->__pyx_n_u_mtu_to_str};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyverbs_utils, __pyx_imported_names, 4, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 10, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_gid_str,__pyx_mstate_global->__pyx_n_u_qp_type_to_str,__pyx_mstate_global->__pyx_n_u_qp_state_to_str,__pyx_mstate_global->__pyx_n_u_mtu_to_str};
    for (__pyx_t_3=0; __pyx_t_3 < 4; __pyx_t_3++) {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 10, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 10, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":11
 * from pyverbs.pyverbs_error import PyverbsUserError, PyverbsError, PyverbsRDMAError
 * from pyverbs.utils import gid_str, qp_type_to_str, qp_state_to_str, mtu_to_str
 * from pyverbs.utils import access_flags_to_str, mig_state_to_str             # <<<<<<<<<<<<<<
 * from pyverbs.wq cimport RwqIndTable, RxHashConf
 * from pyverbs.mr cimport MW, MWBindInfo, MWBind
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_access_flags_to_str,__pyx_mstate_global->__pyx_n_u_mig_state_to_str};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyverbs_utils, __pyx_imported_names, 2, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 11, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_access_flags_to_str,__pyx_mstate_global->__pyx_n_u_mig_state_to_str};
    for (__pyx_t_3=0; __pyx_t_3 < 2; __pyx_t_3++) {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 11, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 11, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":15
 * from pyverbs.mr cimport MW, MWBindInfo, MWBind
 * from pyverbs.wr cimport RecvWR, SendWR, SGE
 * from pyverbs.base import PyverbsRDMAErrno             # <<<<<<<<<<<<<<
 * from pyverbs.addr cimport AHAttr, GID, AH
 * from pyverbs.flow cimport FlowAttr, Flow
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyverbs_base, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno};
    __pyx_t_3 = 0; {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 15, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 15, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.cap must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_5QPCap_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPCap___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.cap must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.cap must be explicitly requested with @auto_pickle(True)"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_5QPCap_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPCap___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[1])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":102
 * 
 * cdef class QPInitAttr(PyverbsObject):
 *     def __init__(self, qp_type=e.IBV_QPT_UD, qp_context=None,             # <<<<<<<<<<<<<<
 *                  PyverbsObject scq=None, PyverbsObject rcq=None,
 *                  SRQ srq=None, QPCap cap=None, sq_sig_all=1):
*/
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(IBV_QPT_UD); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate_global->__pyx_k__2 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_10QPInitAttr_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPInitAttr___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[2])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_10QPInitAttr_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPInitAttr___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[3])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":251
 * 
 * cdef class QPInitAttrEx(PyverbsObject):
 *     def __init__(self, qp_type=e.IBV_QPT_UD, qp_context=None,             # <<<<<<<<<<<<<<
 *                  PyverbsObject scq=None, PyverbsObject rcq=None,
 *                  SRQ srq=None, QPCap cap=None, sq_sig_all=0, comp_mask=0,
*/
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_type_to_py(IBV_QPT_UD); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate_global->__pyx_k__3 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":464
 *         self.ind_table = val
 * 
 *     def mask_to_str(self, mask):             # <<<<<<<<<<<<<<
 *         comp_masks = {1: 'PD', 2: 'XRCD', 4: 'Create Flags',
 *                       8: 'Max TSO header', 16: 'Indirection Table',
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_12QPInitAttrEx_3mask_to_str, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPInitAttrEx_mask_to_str, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[4])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx, __pyx_mstate_global->__pyx_n_u_mask_to_str, __pyx_t_2) < (0)) __PYX_ERR(0, 464, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":475
 *         return mask_str
 * 
 *     def flags_to_str(self, flags):             # <<<<<<<<<<<<<<
 *         create_flags = {1: 'Block self mcast loopback', 2: 'Scatter FCS',
 *                         4: 'CVLAN stripping', 8: 'Source QPN',
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_12QPInitAttrEx_5flags_to_str, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPInitAttrEx_flags_to_str, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[5])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx, __pyx_mstate_global->__pyx_n_u_flags_to_str, __pyx_t_2) < (0)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_12QPInitAttrEx_9__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPInitAttrEx___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[6])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_12QPInitAttrEx_11__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPInitAttrEx___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[7])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":511
 * 
 * cdef class QPAttr(PyverbsObject):
 *     def __init__(self, qp_state=e.IBV_QPS_INIT, cur_qp_state=e.IBV_QPS_RESET,             # <<<<<<<<<<<<<<
 *                  port_num=1, path_mtu=e.IBV_MTU_1024):
 *         """
*/
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py(IBV_QPS_INIT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate_global->__pyx_k__6 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_qp_state_to_py(IBV_QPS_RESET); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate_global->__pyx_k__7 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":512
 * cdef class QPAttr(PyverbsObject):
 *     def __init__(self, qp_state=e.IBV_QPS_INIT, cur_qp_state=e.IBV_QPS_RESET,
 *                  port_num=1, path_mtu=e.IBV_MTU_1024):             # <<<<<<<<<<<<<<
 *         """
 *         Initializes a QPQttr object which represents ibv_qp_attr structs. It
*/
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_mtu_to_py(IBV_MTU_1024); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate_global->__pyx_k__8 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":649
 *         self._copy_ah(val, True)
 * 
 *     def _copy_ah(self, AHAttr ah_attr, is_alt=False):             # <<<<<<<<<<<<<<
 *         if ah_attr is None:
 *             return
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_6QPAttr_3_copy_ah, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPAttr__copy_ah, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[8])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[0]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPAttr, __pyx_mstate_global->__pyx_n_u_copy_ah, __pyx_t_2) < (0)) __PYX_ERR(0, 649, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_6QPAttr_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPAttr___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[9])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_6QPAttr_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPAttr___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[10])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.ece must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_3ECE_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ECE___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[11])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.ece must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.ece must be explicitly requested with @auto_pickle(True)"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_3ECE_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ECE___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[12])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1021
 *                 self.rcq = cqex
 * 
 *     def _create_qp(self, PD pd, QPInitAttr attr):             # <<<<<<<<<<<<<<
 *         self.qp = v.ibv_create_qp(pd.pd, &attr.attr)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_2QP_3_create_qp, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QP__create_qp, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[13])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QP, __pyx_mstate_global->__pyx_n_u_create_qp, __pyx_t_2) < (0)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1024
 *         self.qp = v.ibv_create_qp(pd.pd, &attr.attr)
 * 
 *     def _create_qp_ex(self, Context ctx, QPInitAttrEx attr):             # <<<<<<<<<<<<<<
 *         self.qp = v.ibv_create_qp_ex(ctx.context, &attr.attr)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_2QP_5_create_qp_ex, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QP__create_qp_ex, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[14])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1024, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QP, __pyx_mstate_global->__pyx_n_u_create_qp_ex, __pyx_t_2) < (0)) __PYX_ERR(0, 1024, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1038
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.qp != NULL:
 *             if self.logger:
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_2QP_9close, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QP_close, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[15])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QP, __pyx_mstate_global->__pyx_n_u_close, __pyx_t_2) < (0)) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1052
 *             self.rcq = None
 * 
 *     def _get_comp_mask(self, dst):             # <<<<<<<<<<<<<<
 *         masks = {e.IBV_QPT_RC: {'INIT': e.IBV_QP_PKEY_INDEX | e.IBV_QP_PORT |\
 *                                 e.IBV_QP_ACCESS_FLAGS, 'RTR': e.IBV_QP_AV |\
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_2QP_11_get_comp_mask, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QP__get_comp_mask, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[16])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QP, __pyx_mstate_global->__pyx_n_u_get_comp_mask, __pyx_t_2) < (0)) __PYX_ERR(0, 1052, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1088
 *         return masks[self.qp.qp_type][dst] | e.IBV_QP_STATE
 * 
 *     def to_init(self, QPAttr qp_attr):             # <<<<<<<<<<<<<<
 *         """
 *         Modify the current QP's state to INIT. If the current state doesn't
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_2QP_13to_init, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QP_to_init, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[17])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QP, __pyx_mstate_global->__pyx_n_u_to_init, __pyx_t_2) < (0)) __PYX_ERR(0, 1088, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1104
 *             raise PyverbsRDMAError('Failed to modify QP state to init', rc)
 * 
 *     def to_rtr(self, QPAttr qp_attr):             # <<<<<<<<<<<<<<
 *         """
 *         Modify the current QP's state to RTR. It assumes that its current
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_2QP_15to_rtr, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QP_to_rtr, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[18])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QP, __pyx_mstate_global->__pyx_n_u_to_rtr, __pyx_t_2) < (0)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1124
 *             raise PyverbsRDMAError('Failed to modify QP state to RTR', rc)
 * 
 *     def to_rts(self, QPAttr qp_attr):             # <<<<<<<<<<<<<<
 *         """
 *         Modify the current QP's state to RTS. It assumes that its current
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_2QP_17to_rts, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QP_to_rts, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[19])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QP, __pyx_mstate_global->__pyx_n_u_to_rts, __pyx_t_2) < (0)) __PYX_ERR(0, 1124, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1143
 *             raise PyverbsRDMAError('Failed to modify QP state to RTS', rc)
 * 
 *     def query(self, attr_mask):             # <<<<<<<<<<<<<<
 *         """
 *         Query the QP
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_2QP_19query, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QP_query, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[20])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QP, __pyx_mstate_global->__pyx_n_u_query, __pyx_t_2) < (0)) __PYX_ERR(0, 1143, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1159
 *         return attr, init_attr
 * 
 *     def modify(self, QPAttr qp_attr not None, comp_mask):             # <<<<<<<<<<<<<<
 *         """
 *         Modify the QP
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_2QP_21modify, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QP_modify, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[21])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QP, __pyx_mstate_global->__pyx_n_u_modify, __pyx_t_2) < (0)) __PYX_ERR(0, 1159, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1172
 *             raise PyverbsRDMAError('Failed to modify QP', rc)
 * 
 *     def post_recv(self, RecvWR wr not None, RecvWR bad_wr=None):             # <<<<<<<<<<<<<<
 *         """
 *         Post a receive WR on the QP.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_2QP_23post_recv, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QP_post_recv, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[22])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[1]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QP, __pyx_mstate_global->__pyx_n_u_post_recv, __pyx_t_2) < (0)) __PYX_ERR(0, 1172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1189
 *             raise PyverbsRDMAError('Failed to post recv', rc)
 * 
 *     def post_send(self, SendWR wr not None, SendWR bad_wr=None):             # <<<<<<<<<<<<<<
 *         """
 *         Post a send WR on the QP.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_2QP_25post_send, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QP_post_send, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[23])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[1]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QP, __pyx_mstate_global->__pyx_n_u_post_send, __pyx_t_2) < (0)) __PYX_ERR(0, 1189, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1206
 *             raise PyverbsRDMAError('Failed to post send', rc)
 * 
 *     def set_ece(self, ECE ece):             # <<<<<<<<<<<<<<
 *         """
 *         Set ECE options and use them for QP configuration stage
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_2QP_27set_ece, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QP_set_ece, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[24])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QP, __pyx_mstate_global->__pyx_n_u_set_ece, __pyx_t_2) < (0)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1219
 *             raise PyverbsRDMAError('Failed to set ECE', rc)
 * 
 *     def query_ece(self):             # <<<<<<<<<<<<<<
 *         """
 *         Query QPs ECE options
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_2QP_29query_ece, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QP_query_ece, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[25])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QP, __pyx_mstate_global->__pyx_n_u_query_ece, __pyx_t_2) < (0)) __PYX_ERR(0, 1219, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1230
 *         return ece
 * 
 *     def bind_mw(self, MW mw not None, MWBind mw_bind):             # <<<<<<<<<<<<<<
 *         """
 *         Bind Memory window type 1.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_2QP_31bind_mw, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QP_bind_mw, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[26])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QP, __pyx_mstate_global->__pyx_n_u_bind_mw, __pyx_t_2) < (0)) __PYX_ERR(0, 1230, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1241
 *             raise PyverbsRDMAError('Failed to Bind MW', rc)
 * 
 *     def query_data_in_order(self, op, flags=0):             # <<<<<<<<<<<<<<
 *         """
 *         Query if QP data is guaranteed to be in order.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_2QP_33query_data_in_order, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QP_query_data_in_order, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[27])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[2]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QP, __pyx_mstate_global->__pyx_n_u_query_data_in_order, __pyx_t_2) < (0)) __PYX_ERR(0, 1241, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.qp cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_2QP_37__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QP___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[28])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.qp cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.qp cannot be converted to a Python object for pickling"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_2QP_39__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QP___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[29])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.data cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_10DataBuffer_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_DataBuffer___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[30])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.data cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.data cannot be converted to a Python object for pickling"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_10DataBuffer_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_DataBuffer___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[31])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1333
 *         return self.ind_table
 * 
 *     def wr_atomic_cmp_swp(self, rkey, remote_addr, compare, swap):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_atomic_cmp_swp(self.qp_ex, rkey, remote_addr, compare, swap)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_3wr_atomic_cmp_swp, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx_wr_atomic_cmp_swp, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[32])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx, __pyx_mstate_global->__pyx_n_u_wr_atomic_cmp_swp, __pyx_t_2) < (0)) __PYX_ERR(0, 1333, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1336
 *         v.ibv_wr_atomic_cmp_swp(self.qp_ex, rkey, remote_addr, compare, swap)
 * 
 *     def wr_atomic_fetch_add(self, rkey, remote_addr, add):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_atomic_fetch_add(self.qp_ex, rkey, remote_addr, add)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_5wr_atomic_fetch_add, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx_wr_atomic_fetch_add, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[33])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx, __pyx_mstate_global->__pyx_n_u_wr_atomic_fetch_add, __pyx_t_2) < (0)) __PYX_ERR(0, 1336, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1339
 *         v.ibv_wr_atomic_fetch_add(self.qp_ex, rkey, remote_addr, add)
 * 
 *     def wr_bind_mw(self, MW mw, rkey, MWBindInfo bind_info):             # <<<<<<<<<<<<<<
 *         cdef v.ibv_mw_bind_info *info
 *         info = &bind_info.info
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_7wr_bind_mw, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx_wr_bind_mw, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[34])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx, __pyx_mstate_global->__pyx_n_u_wr_bind_mw, __pyx_t_2) < (0)) __PYX_ERR(0, 1339, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1346
 *         self.add_ref(mw)
 * 
 *     def wr_local_inv(self, invalidate_rkey):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_local_inv(self.qp_ex, invalidate_rkey)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_9wr_local_inv, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx_wr_local_inv, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[35])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx, __pyx_mstate_global->__pyx_n_u_wr_local_inv, __pyx_t_2) < (0)) __PYX_ERR(0, 1346, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1349
 *         v.ibv_wr_local_inv(self.qp_ex, invalidate_rkey)
 * 
 *     def wr_atomic_write(self, rkey, remote_addr, atomic_wr):             # <<<<<<<<<<<<<<
 *         cdef char *atomic_wr_c = atomic_wr
 *         v.ibv_wr_atomic_write(self.qp_ex, rkey, remote_addr, atomic_wr_c)
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_11wr_atomic_write, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx_wr_atomic_write, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[36])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx, __pyx_mstate_global->__pyx_n_u_wr_atomic_write, __pyx_t_2) < (0)) __PYX_ERR(0, 1349, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1353
 *         v.ibv_wr_atomic_write(self.qp_ex, rkey, remote_addr, atomic_wr_c)
 * 
 *     def wr_rdma_read(self, rkey, remote_addr):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_rdma_read(self.qp_ex, rkey, remote_addr)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_13wr_rdma_read, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx_wr_rdma_read, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[37])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx, __pyx_mstate_global->__pyx_n_u_wr_rdma_read, __pyx_t_2) < (0)) __PYX_ERR(0, 1353, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1356
 *         v.ibv_wr_rdma_read(self.qp_ex, rkey, remote_addr)
 * 
 *     def wr_rdma_write(self, rkey, remote_addr):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_rdma_write(self.qp_ex, rkey, remote_addr)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_15wr_rdma_write, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx_wr_rdma_write, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[38])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx, __pyx_mstate_global->__pyx_n_u_wr_rdma_write, __pyx_t_2) < (0)) __PYX_ERR(0, 1356, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1359
 *         v.ibv_wr_rdma_write(self.qp_ex, rkey, remote_addr)
 * 
 *     def wr_rdma_write_imm(self, rkey, remote_addr, data):             # <<<<<<<<<<<<<<
 *         cdef unsigned int imm_data = htobe32(data)
 *         v.ibv_wr_rdma_write_imm(self.qp_ex, rkey, remote_addr, imm_data)
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_17wr_rdma_write_imm, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx_wr_rdma_write_imm, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[39])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx, __pyx_mstate_global->__pyx_n_u_wr_rdma_write_imm, __pyx_t_2) < (0)) __PYX_ERR(0, 1359, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1363
 *         v.ibv_wr_rdma_write_imm(self.qp_ex, rkey, remote_addr, imm_data)
 * 
 *     def wr_flush(self, rkey, remote_addr, length, ptype, level):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_flush(self.qp_ex, rkey, remote_addr, length, ptype, level)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_19wr_flush, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx_wr_flush, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[40])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx, __pyx_mstate_global->__pyx_n_u_wr_flush, __pyx_t_2) < (0)) __PYX_ERR(0, 1363, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1366
 *         v.ibv_wr_flush(self.qp_ex, rkey, remote_addr, length, ptype, level)
 * 
 *     def wr_send(self):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_send(self.qp_ex)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_21wr_send, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx_wr_send, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[41])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx, __pyx_mstate_global->__pyx_n_u_wr_send, __pyx_t_2) < (0)) __PYX_ERR(0, 1366, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1369
 *         v.ibv_wr_send(self.qp_ex)
 * 
 *     def wr_send_imm(self, data):             # <<<<<<<<<<<<<<
 *         cdef unsigned int imm_data = htobe32(data)
 *         return v.ibv_wr_send_imm(self.qp_ex, imm_data)
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_23wr_send_imm, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx_wr_send_imm, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[42])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx, __pyx_mstate_global->__pyx_n_u_wr_send_imm, __pyx_t_2) < (0)) __PYX_ERR(0, 1369, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1373
 *         return v.ibv_wr_send_imm(self.qp_ex, imm_data)
 * 
 *     def wr_send_inv(self, invalidate_rkey):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_send_inv(self.qp_ex, invalidate_rkey)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_25wr_send_inv, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx_wr_send_inv, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[43])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx, __pyx_mstate_global->__pyx_n_u_wr_send_inv, __pyx_t_2) < (0)) __PYX_ERR(0, 1373, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1376
 *         v.ibv_wr_send_inv(self.qp_ex, invalidate_rkey)
 * 
 *     def wr_send_tso(self, hdr, hdr_sz, mss):             # <<<<<<<<<<<<<<
 *         ptr = PyLong_AsVoidPtr(hdr)
 *         v.ibv_wr_send_tso(self.qp_ex, ptr, hdr_sz, mss)
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_27wr_send_tso, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx_wr_send_tso, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[44])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx, __pyx_mstate_global->__pyx_n_u_wr_send_tso, __pyx_t_2) < (0)) __PYX_ERR(0, 1376, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1380
 *         v.ibv_wr_send_tso(self.qp_ex, ptr, hdr_sz, mss)
 * 
 *     def wr_set_ud_addr(self, AH ah, remote_qpn, remote_rkey):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_set_ud_addr(self.qp_ex, ah.ah, remote_qpn, remote_rkey)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_29wr_set_ud_addr, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx_wr_set_ud_addr, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[45])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx, __pyx_mstate_global->__pyx_n_u_wr_set_ud_addr, __pyx_t_2) < (0)) __PYX_ERR(0, 1380, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1383
 *         v.ibv_wr_set_ud_addr(self.qp_ex, ah.ah, remote_qpn, remote_rkey)
 * 
 *     def wr_set_xrc_srqn(self, remote_srqn):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_set_xrc_srqn(self.qp_ex, remote_srqn)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_31wr_set_xrc_srqn, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx_wr_set_xrc_srqn, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[46])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx, __pyx_mstate_global->__pyx_n_u_wr_set_xrc_srqn, __pyx_t_2) < (0)) __PYX_ERR(0, 1383, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1386
 *         v.ibv_wr_set_xrc_srqn(self.qp_ex, remote_srqn)
 * 
 *     def wr_set_inline_data(self, addr, length):             # <<<<<<<<<<<<<<
 *         ptr = PyLong_AsVoidPtr(addr)
 *         v.ibv_wr_set_inline_data(self.qp_ex, ptr, length)
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_33wr_set_inline_data, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx_wr_set_inline_data, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[47])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx, __pyx_mstate_global->__pyx_n_u_wr_set_inline_data, __pyx_t_2) < (0)) __PYX_ERR(0, 1386, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1390
 *         v.ibv_wr_set_inline_data(self.qp_ex, ptr, length)
 * 
 *     def wr_set_inline_data_list(self, num_buf, buf_list):             # <<<<<<<<<<<<<<
 *         cdef v.ibv_data_buf *data = NULL
 *         data = <v.ibv_data_buf*>malloc(num_buf * sizeof(v.ibv_data_buf))
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_35wr_set_inline_data_list, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx_wr_set_inline_data_list, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[48])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx, __pyx_mstate_global->__pyx_n_u_wr_set_inline_data_list, __pyx_t_2) < (0)) __PYX_ERR(0, 1390, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1402
 *         free(data)
 * 
 *     def wr_set_sge(self, SGE sge not None):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_set_sge(self.qp_ex, sge.lkey, sge.addr, sge.length)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_37wr_set_sge, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx_wr_set_sge, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[49])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx, __pyx_mstate_global->__pyx_n_u_wr_set_sge, __pyx_t_2) < (0)) __PYX_ERR(0, 1402, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1405
 *         v.ibv_wr_set_sge(self.qp_ex, sge.lkey, sge.addr, sge.length)
 * 
 *     def wr_set_sge_list(self, num_sge, sg_list):             # <<<<<<<<<<<<<<
 *         cdef v.ibv_sge *sge = NULL
 *         sge = <v.ibv_sge*>malloc(num_sge * sizeof(v.ibv_sge))
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_39wr_set_sge_list, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx_wr_set_sge_list, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[50])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx, __pyx_mstate_global->__pyx_n_u_wr_set_sge_list, __pyx_t_2) < (0)) __PYX_ERR(0, 1405, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1417
 *         free(sge)
 * 
 *     def wr_start(self):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_start(self.qp_ex)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_41wr_start, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx_wr_start, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[51])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx, __pyx_mstate_global->__pyx_n_u_wr_start, __pyx_t_2) < (0)) __PYX_ERR(0, 1417, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1420
 *         v.ibv_wr_start(self.qp_ex)
 * 
 *     def wr_complete(self):             # <<<<<<<<<<<<<<
 *         rc = v.ibv_wr_complete(self.qp_ex)
 *         if rc != 0:
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_43wr_complete, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx_wr_complete, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[52])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx, __pyx_mstate_global->__pyx_n_u_wr_complete, __pyx_t_2) < (0)) __PYX_ERR(0, 1420, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1425
 *             raise PyverbsRDMAError('ibv_wr_complete failed.', rc)
 * 
 *     def wr_abort(self):             # <<<<<<<<<<<<<<
 *         v.ibv_wr_abort(self.qp_ex)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_45wr_abort, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx_wr_abort, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[53])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPEx, __pyx_mstate_global->__pyx_n_u_wr_abort, __pyx_t_2) < (0)) __PYX_ERR(0, 1425, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.qp,self.qp_ex cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_47__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[54])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.qp,self.qp_ex cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.qp,self.qp_ex cannot be converted to a Python object for pickling"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_4QPEx_49__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_QPEx___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[55])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1429
 * 
 * 
 * def _copy_caps(QPCap src, dst):             # <<<<<<<<<<<<<<
 *     """
 *     Copy the QPCaps values of src into the inner ibv_qp_cap struct of dst.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2qp_1_copy_caps, 0, __pyx_mstate_global->__pyx_n_u_copy_caps, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_qp, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[56])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_copy_caps, __pyx_t_2) < (0)) __PYX_ERR(0, 1429, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/qp.pyx":1
 * # SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)             # <<<<<<<<<<<<<<
 * # Copyright (c) 2019 Mellanox Technologies, Inc. All rights reserved.
 * 
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_test, __pyx_t_2) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  if (__pyx_m) {
    if (__pyx_mstate->__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init pyverbs.qp", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init pyverbs.qp");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #else
  return __pyx_m;
  #endif
}
/* #### Code section: pystring_table ### */
/* #### Code section: cached_builtins ### */

static int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_super); if (!__pyx_builtin_super) __PYX_ERR(0, 50, __pyx_L1_error)

  /* Cached unbound methods */
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.method_name = &__pyx_mstate->__pyx_n_u_items;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.method_name = &__pyx_mstate->__pyx_n_u_pop;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.method_name = &__pyx_mstate->__pyx_n_u_values;
  __pyx_mstate->__pyx_umethod_PyUnicode_Type__format.type = (PyObject*)(&PyUnicode_Type);
  __pyx_mstate->__pyx_umethod_PyUnicode_Type__format.method_name = &__pyx_mstate->__pyx_n_u_format;
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "pyverbs/qp.pyx":649
 *         self._copy_ah(val, True)
 * 
 *     def _copy_ah(self, AHAttr ah_attr, is_alt=False):             # <<<<<<<<<<<<<<
 *         if ah_attr is None:
 *             return
*/
  __pyx_mstate_global->__pyx_tuple[0] = PyTuple_Pack(1, Py_False); if (unlikely(!__pyx_mstate_global->__pyx_tuple[0])) __PYX_ERR(0, 649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[0]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[0]);

  /* "pyverbs/qp.pyx":1172
 *             raise PyverbsRDMAError('Failed to modify QP', rc)
 * 
 *     def post_recv(self, RecvWR wr not None, RecvWR bad_wr=None):             # <<<<<<<<<<<<<<
 *         """
 *         Post a receive WR on the QP.
*/
  __pyx_mstate_global->__pyx_tuple[1] = PyTuple_Pack(1, Py_None); if (unlikely(!__pyx_mstate_global->__pyx_tuple[1])) __PYX_ERR(0, 1172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[1]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[1]);

  /* "pyverbs/qp.pyx":1241
 *             raise PyverbsRDMAError('Failed to Bind MW', rc)
 * 
 *     def query_data_in_order(self, op, flags=0):             # <<<<<<<<<<<<<<
 *         """
 *         Query if QP data is guaranteed to be in order.
*/
  __pyx_mstate_global->__pyx_tuple[2] = PyTuple_Pack(1, __pyx_mstate_global->__pyx_int_0); if (unlikely(!__pyx_mstate_global->__pyx_tuple[2])) __PYX_ERR(0, 1241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[2]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[2]);
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_tuple;
    for (Py_ssize_t i=0; i<3; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  {
    const struct { const unsigned int length: 9; } index[] = {{1},{14},{18},{17},{14},{18},{16},{9},{14},{15},{16},{12},{25},{15},{10},{9},{12},{14},{5},{20},{25},{17},{29},{30},{19},{28},{20},{19},{32},{32},{33},{19},{19},{19},{18},{17},{17},{13},{14},{19},{13},{179},{21},{10},{8},{15},{96},{97},{96},{11},{15},{9},{16},{8},{7},{9},{6},{7},{10},{13},{6},{11},{11},{11},{10},{24},{9},{0},{1},{1},{8},{9},{12},{7},{6},{170},{164},{173},{178},{169},{168},{10},{2},{9},{23},{9},{9},{14},{15},{17},{12},{14},{11},{13},{13},{12},{12},{14},{11},{13},{13},{12},{8},{8},{14},{77},{61},{61},{59},{70},{10},{16},{7},{11},{15},{10},{28},{30},{3},{21},{23},{13},{16},{13},{12},{12},{11},{12},{11},{11},{12},{13},{11},{11},{11},{11},{15},{14},{18},{10},{10},{10},{16},{16},{19},{15},{9},{10},{16},{15},{26},{29},{26},{31},{19},{24},{31},{21},{25},{23},{20},{21},{15},{17},{11},{11},{17},{16},{16},{13},{13},{12},{14},{4},{7},{2},{12},{16},{16},{16},{20},{4},{2},{6},{24},{26},{15},{5},{23},{25},{4},{22},{24},{13},{22},{24},{20},{15},{16},{13},{17},{17},{18},{22},{12},{16},{16},{16},{23},{28},{15},{20},{19},{20},{13},{10},{12},{30},{32},{25},{24},{28},{30},{20},{22},{13},{16},{17},{10},{8},{9},{12},{12},{8},{22},{12},{10},{10},{9},{9},{3},{3},{7},{7},{4},{19},{3},{4},{2},{7},{18},{9},{11},{4},{9},{6},{9},{7},{8},{3},{17},{18},{5},{9},{10},{7},{8},{10},{12},{10},{13},{10},{7},{3},{12},{4},{8},{5},{4},{4},{3},{3},{1},{5},{12},{10},{6},{8},{14},{12},{7},{2},{9},{3},{6},{9},{1},{13},{7},{16},{21},{12},{11},{8},{9},{4},{8},{9},{15},{6},{13},{9},{5},{6},{5},{4},{8},{4},{8},{11},{5},{15},{12},{11},{12},{11},{14},{16},{6},{10},{3},{10},{2},{7},{9},{8},{7},{7},{2},{7},{8},{2},{3},{8},{9},{9},{3},{5},{12},{24},{21},{10},{13},{11},{14},{2},{3},{4},{7},{10},{6},{8},{15},{7},{14},{12},{5},{19},{9},{2},{3},{7},{10},{17},{13},{11},{10},{11},{11},{4},{3},{4},{7},{14},{7},{12},{10},{12},{19},{7},{3},{10},{2},{10},{10},{3},{13},{3},{11},{5},{4},{1},{8},{7},{6},{6},{13},{6},{9},{4},{8},{7},{2},{8},{17},{19},{15},{10},{11},{8},{12},{12},{13},{17},{7},{11},{11},{11},{18},{23},{10},{15},{14},{15},{8},{4},{37},{23},{15},{59},{114},{13},{15},{83},{83},{69},{19},{113},{295},{22},{21},{11},{13},{22},{25},{26},{23},{122},{25},{27},{11},{36},{21},{55},{72},{13},{17},{47},{45},{19},{42},{63},{72},{9},{448},{74},{295}};
    #if (CYTHON_COMPRESS_STRINGS) == 3 && __PYX_LIMITED_VERSION_HEX >= 0x030e0000 /* compression: zstd (3841 bytes) */
const char* const cstring = "(\265/\375`@$\275w\000\272\215\244\035G\360\260\350\001\314\300v\204F\301\324\t\205\203)Fh\024L1B\343\322.\224\336>\")\371\033N\235\351\377\304$U\314\364ib\244\255\273\255P\326\000Y\231\177\016\373t\243\236w\236\nG=\326\343\335=\315}\377\203\005\244\001\263\001\314\001\242\005y\007\267\2757G\265\255\r\265\337\036\216\307\340\347\341%aI%\333y\n\025\177\261C\026\017mL\271\217\034\205\032\305\240|\270q\367P]\377p\344\331r\177\203\373\343Z\036\267l\2123\241Nv\343b\375\353\376C\335\373\370\361\342W:g\254\263\356\377\343\013z\354]\212\026/\377#.\261_>\366\375\235\367'\252\354\367#hE\234C\036\333V\326\240\375y\374p\362>\205X\033\264\351hl\353\343\223\347y\341\356/[\032\235B\332\1778&\307\242\304\376\305!F\261\241\366X\207v\3441\273*\331\220\327b\354\377o\257\270\365\332{\374e\325\350\255w\341\205\330\266\270\264l8i'\033\366\330\267?\277[?6\262\037\267\270\251s&\316\206;\324\265\207u\313\326\240\272\274A\3769\013\3034\372S\341\316\206|\377\355\205\245\257\275\215u=\272|N\326\235\213\177\207\033\305u\334j{\034\214q\355y\241\3760\235\361\307^\327\253\325\250\262\232\014\352\30374\026\306\362\377m\034#\343\303\260\237\263\254\206\224\266\2158l#\353\351\237\265\010\323b\253\375yhs\030\363\330\2226\224\r\205\332\307\2136\322\276\327$\317\302_\214\325\037\353>\302\361\230\306\212 \037\2356\353\3717b\205[\024w\377\336\2322\024)F8\031\253D\316?\373=\316\240]}\332[\234\314?\230?r\r\356\337\265\n\034C\373;\027a\332\013\267\215\351\023'\033\372_\366\361\322\377\317\354L\006\203\251\254\330\017\021/0\354\213\014\225\225\027\030\241!\276[\270\205\357\327(\034\254o:\337\177\347?\353\256\275\316o\270\357\367\\\324\236\1779\3645\253\363\177\377\016K\370\207\256\356o\356\260\356\215\337\357\357\336\373\271\377 \270\257\207{N\270\357\327\362\343\267z\201\327\013\267\312\363\360m\345\257W\342\313_\021\314\367w\256\302!\230&\342\026\2761\221\347\376\177v\332N\333\266\322\334\207i\323\375y\223`\261\323\275]{\033\036f4$\256\363\336\265\023\351\344\305\232\316\246\006\205:\335\3154\272*\270\216""\246\204\342n\035W\223\322Th'hJ\335\275\246\322\250DSC\342fP\331\314Mq\244\023\247\2212m\346\336TF\243\241\200\311\272K\322\270\221\212\230RWc\272#\223\nTvM\232\251Ci\244\316\346\224\215\356L\327\241@i\243QvS\214F\\\212d3:1\312N\367&\3438-E\211#\335\316tSc:a\262\321P\\\247\2418N\273\250\025Q\007\020*-\036:\270\264\250\254\264\254\250\210\256M\247q]\306\2214\032.\2732%\232\257\375\275\371\352<\336\033\202!]\370\241\006YX\367\237\322m\336\326\377\255\022\275\267Y\032D\227\270:\331\037\362\234\361\313\242d2\0241\027u\252\271$\223\206\272\\\327e\267+\231J\367\264\335\0240%\253\363\326\242\213\203\355/\257Y\236\273\2616\334_\270A\317\357\357cY\277C\372\273\375%\362Vf}Z\205?\373\342\337\352^\3215\376\020\254skc:\007\267\274\346\336\267W\326\203\211!\357\313\332\263+\355\211\331\342\321\306T\274\213Vw\254\342\206\274\252\353\234X\2405\253!#\035\245\".\250\0363D*\330t2Qza3O<+\350\2641S\345K\007\323\245t!\025\220\034\231\213\031\203d\200\311\005\256\022\227\032\004\220\035\372\t3\261M\240f\246I\226\202\240\035N\005g\t\004,:A\347WJ 92G\323\001!\027@(Y?h>C\017\324\007%\253(\025\220\272<P\005\224%\237\222\243\364(q\355\310l(\353|\231\330\\`veV\206kV\312\020\231 \223D\271\225#\345\013\222\355\3141a\016\\d\001Qi\200\000\031\002\262\272R\022(*6-\220\250$!\253\3134\341@b\324\004M9=\223\200\003\220\025d~L\017\245\0149\022\262\212\022\200\254\230d!\277N\013G\352RA\005\231\024\024\250\n@\326\017\231\rx\254\034'\312\314\314\261l)A\345\247\203\300\303\306\001\220\225\303D\225\031HO\207E\007\304\003\000\344FF\003\237\004N\017\223T\212\312\007\2365\243\361I;t\000J:Eg\206\020\365\343\001\000\341Fvj\350P\375\005p!\005CM\354\344\365\327J\367W\330\361\313\3337\307\353q\265\022\361\313w-\216\376\336\334\375\367a\177]\376c\371\377\215X\344\256\354C\321\245?\007a\372\272\262+\177x\3740\375\365vaW\016,&\336\353?\321\245=V\211\375\213\236\310E\016\216\032\304\341\2757\006\276\300\227\367\372\033\357\275\334\377\376\272\337\247s\364\327\372\254P\177\376\376\017_\277c\032\247\343{_\370\333\275\326*""\364\270\325Y\367\336\034\216\253{?\006\357\352\373\317\3636\313c\215*\355\3651\303~\356\334\353\216\275\316\340\275\"\316\253Q\305W\341\366\260*_\337y\016\3339rf\323B\005\021\365S\341D\347#\245s\300\303\206\2003%;;N\026\250\227\034\321:\233\316\001\317\352M\024\321\371\315\221i2\242\262\2305d\007)\"*\320\261z>\r\217\001\000$\334\340\330AI?\270(\t\224\031*B\035\234\275C\315\322\031\364\346\204\340*\200NX\001\251\030m\221\224l;#;%-6\026 \324;\027\234\235\t\301\006\203\223EZ\322\"u\271[(\202\372\301L\221\2413)\030\311 d\247KS\021@;\336\353\224\220\372\240\034\243\331\362`\304\246\244\324l\310\001\007\\\316P\207\272\221\031t>:V\216\234\024\304\216\2124K\037\240(\003@\023\344H\016F_.\363dF\210\007\300\310L\203\325\205\236\240\001\320\025\237\322S#\246\340\206\010\225\321\010>;4\350\352l\2012\227\211\022B\010\214\210\024)\035\220\254\256\263\345\314\320@m@R\301\227$e&I<@\351\013\350\364\t<\000>Zh\204\323\203\351\202\323\021\"\363\213\312O\"B]Z\245I\371)\033\200P\033\217\004\004*\002\364\0026aL\221)Sb\310P\312\0242\000\t\"\275\310\202H\212\202\372\263@\212\243,I7\2401\347\214\331a\272\312\t\2363e\216\230\034\246\rQC<\326\355\314\316\033\363$=\200Z\370\230 \320\231bz\345VF\000\225\036\35119?\247fR\030A\272\320\026\024\204~\230\210\2502,\247.\225\"eGf\344hc\313v.j\340\004\255\237\235,\244\2520O>+>C(\254\024\001\0323Q>\253\313D\231\020\362@U\354\272\355h-\252\010*\262\353\247\006N\332B=\330UT\003LY\322\020)j\254\234\360YE\250\000\216\0323\036T\026\224%uIQS\354J!\025\001Z9P\025\320\212wb:\250\025\320\025\320\n\252\001#\205\241\037\002\203\225\250R\223\226&\225!\032\021\021I\222\202\244\003b \010\214ha\232\347j>\362\251&CA\345\21012B2\022\210HI\nJJJ\031\306\343\0072QK\n\375-\272d\260\314?\005\3430\263}\341Y\322\006%\352n'\010]>\301\232\330S\317\204\210\000n`4\034\201\206\033\247\324b\014\267+\000\376\346#m=\"3\3461\241\202\243Q\245\\```\365\355\376\251T\242Q^\247d\031w\020\240\200YG\245uISC\256\206\241\202P\237\024\330\307\t\261c\307\2560\261\257\001\000X7\327\317 \306\351""\tU\016\270y\207Fp\035\272# x2U\340\000\377\237%B\351\315N\304\037\274\326\225\215\371\220\200\247#\3678\342\027\006\031\253\235\262\267\265\200\"\315\022\212\361\342\231\266\2548(!pB\024;\344\225\266E\304\027#=\211\316\251\323`\216\245\360eU)\335\315/\216\305l\267\236\344}?yN\240\266(\252\211\236:!\004\351\372\354\252s\260`\201\027Z\031\242@Z\226\213\031b\254\322gP8\t\333?L\320\024\213rH\224W\260/Y\351\262'\271\214\204\260P\302\213\016r\034\323S\311\222\336\022E\021\302\233\253\320m\357\352\023\307\223\317\235\273\320L\314\207\026\221\262R)\342,|q)\341h\216\350#\014\335('\020?\212_\300\347\266Gm\275+X\210+\014~f\227@\364!\306ap\246\213u\334\346`i\007'Vh\004V0\317\001\367\341v\017W_\002\333\3644\264X6\314\200z\272\250\312\020\262\357\311\215\306i\027\t\206r\316U\277\001 \026U\036C5\325\256\332\026~\216@\3240W\201\365t`\242\022\037\003\013\200\317\027\365\206\346\0013\026\261\025G}\315\233=\204\007\265(:\"\305\207\267Rc\244\376!sb\364q\002\031q*\257\335\305?\231p\004\000c\216\340Y\246xdZ\304I3\225\243\367s\341\26581\214e&\251Cu\016\350\307\362Su\025-\236Ap^}v\025\nX\0108\n!\323K\220\232\262\362\020R\225\2641\"\263>\203\3576L1\025B\253/x\341q\307\260+\0136nCp\336\302vw\360+\235\254\310\245\256\205\350\026\261\352~\260\244\022\361\327\304~\2370/)\"\245b\033\256\260\307,\002\212\213%\337\013\237\303\202\257u\300\216\231\333P2U{\234\371\364v\250\344]\274\005\270\\S\354\210;:\333!\335\224\354\020w*\024\300\003G,(\302+\"O\372@I\207\300\330\230\016A\017\242\300\206\001s\254\303K\357\251\243\027\331\261\025s\242\373\211\215\300<\213\357%\350\205x)\311\257\366e\236\352?\000>\252\343\261\310\0352\361\022J`\214>5\213R\303\250m\014\244xf\205K\364\200\323t\263\200\035\004k\331\000\010)\230:Gb*\337\001\010,\370\316\272\316\206Z1%\233-\265\265\031\361\037\246\026a\231\3367\3433\346Qd\000U\263i\235\314\210t\331\266\210\274E>\373\202\327b\254\247\210\361\312\035g\2012\\\0357 4\347\242<\246\237js\214\031!\343\206^\237\346i%\010\227\353\000\267\302c\351\347\n\">B\302\240\266\3034\323""\252'4l\341 \006V\210'\363\302]\030\312\357\177\367\2073\305\252\356\2779/\302\024\232\373\322\314C\031\225\027/D\320H\034zQ\002\3334n\026\345'p\367\206\305;VQ\240\t1?\250\265\nf\240\017o\246\006\210\234\"\363? \2211\033\207yO\220\027\377\276sd\312W#]^6\0104\337\363\234\230F\350.\017L\247\212IxgF\206\023V\314\271\214F\277\302\300\355\215\312+\262\370\021u\211~\254\307q\010L\360\2442\032\273\013\334ej\000^\367T(\004\233\2675\021Y\220LFI\364\377\237\016\247%U\336\303W\355B\276\025\214\\\250\247(;lw\272l\335\207;>z\004\250\030\0254'\272\303I\361\263\216/\313\310\243\2146`#\237(\274`0\3056G\2317,\212\203\306\347\223\264\321:\217\337\351\007[O\272\307\227\276\233\337\352\333w\237\004\253^WN[\023\016%g\274\032\311AU\374\357@\220\266\216\327H\223,1\010\356\004.G\007\n;7\006zf\027y\301N\227\274\331\325\330|v\000\025A7>\250I_\343\276\257Yg9\273T\215\202\344uC\202\356iJ\272\260\004\355\010c#\306zy\024\362cV\257\320g\261\364\371\373s\251\000\036\247s\275\366Z\234\355\005\215\354\203PA\201\000\214m\311\177\210\035j\375f;H \300g\021\266\251\205\217-*\215\242\027\332q\235\037&S&\310\245h\301KlN\310\301\001B\237\260\351v\363OS\007\254\223fS\026;\000\361M\361\003H\t\377\272\020O\300\302\000\235\263\035\354\324H\232\270\006)\270\236*\355<\347\265\312\033\367\321\243P)P\275\322\365\337a\275\210\010\\\226E\252\345\322\035\320\356\361\017\035\035\265\020\271\250\005\335\323\236\236M\300\225\254\322|\177o!^\226\262\030\000\0174\3169\316\235\370\033\314\237B\037\313\262\216\346\201UL\\\242e\300F\021I\207\370\364\r\2012T\372iI\030\274\315?\336\267\"\266&\232=O\004\032\201\376\303\347H\030V\266\005B\332v\204\374G\021\027\375\016\324\231R\372\253kW\276\007U\267\010\346\337\035\276\217\237\354\203\361\307\222.Q\240\262\371\205\2112()\225\270<\230\020\274\037r\222\222\001(\021\336=\257\204\330\335\346\007Z\345\307\300\\\334\264\255\361\247\362\275\022\335\336\037\313\313?\306\r\275\3125HN\352G\372\274 \247\330I\356\265\244c\355)\003\253\270\035v\021\200\336\237\327\007\314\034d\016N\373%Eh""\203\002A\034\351\345*\372\364\010\024\200Y\"\212p/\027\365\340\322dg\241\017tk$\351\335/\016\353\177\270\255M\303B\311\002\310\212\"\335b\301Lz\325\255U5\326li\014\301\250/4H\267cxNk\275u\261j\300\310\027T\273\363\006\257\307X\241\224\0364H\357\323\312}\223\344M\017\207\354 \365\r\364F\306r'P\r\304:\013\230\010gFE!\"\021\266L\312\177\tN\332c\004\022d\201\225Q\016\372\022\274\362\341\177\256?Zn0,~\302\216SH\004tcU\255\333\026*\302>1\364Uy\306\016\226\204|\353j\0036\271*\340'#\"\311%\230\350\n\241qe\005?\361\024\255\225\214\026\034;\336%\267\031\207\227H\226\037\356\020\266h\273\223b2\"q$\327\315\351n\254F8'0\333#\355\265\202\325D\266_\223\253Y\333\276UZ\267o\270\034tJ\273\307\336\027\372\227\351\337-3/\227\ta\347\256\"\321\033]\330\215Ut\355\376\276\363\024\252\252\333\024o\320\253\204\251A\r\342\014x2\237\200Po\233\203^\247d\031\267\341\362\265\031\026\273\003UB\"\2334\364\215;jO\026B\0237q|R\345\336\376Y\"\223=7\014K \"\224z\002_\266D\323b\304\304\366\346\246\350\222\311\300\227\360\310W[\265Ks\243(\376tY\265\244\223\322\353\363\232\031\372,\201\220\211\266~\354\001";
    PyObject *data = __Pyx_DecompressString(cstring, 3841, 3);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) == 2 /* compression: bz2 (4068 bytes) */
const char* const cstring = "BZh91AY&SY\0062\241X\000\002\315\377\377\377\377\377\377\377\377\377\375\377\377\377\356\277\377\377\377@@@@@@@@@@@@@\000@\000`\020\233\341\303Yw\267\211y\224%\331\250\357n)-=S\333y-\215$\326\205\330[T\352\264\264\354\360\003\347\034%\t\n\236\223F\231\032\000\032\032z\217&\231\211\244\364\220\006\215\036\230\243@\311\241\243#M4\006#&\322d\r\032\032\t$\000\201\014\211\251\232\247\350T\330\nh\364\2327\244!\351\000\032\006 \036\243j44\r\0004\014\020\0001\t\244\232\246\365F\231\244\321\240\036\242lQ\350\233M\000\004b`L\2004b0\000&\004\311\264\215\r\224\310\t\n \203HM5=C&OP\375SFOD\006\217)\204h\000i\352\000\000\031\250\003\324\032=@d\320\311\204*h\2310\230\206\230\323D1\0320\002\032\031\006\215\030A\220\300\t\350L\2304L\201\240\032\0324\002D\202\002\000\0053S\005M\220\332i&\323M\020\323\324\001\240\000\320\000\000\000\000\0004\311\246\272\354}V\227N\370\330\353[\227\3238\006\207\370\227\364?\263\355:#\333\243\020\325\225P\205U`\272\243\t\226{'\262\252\\XB\332\304\211[\005\222)\332\232\202\226\200\217\3740M\215\244\034\264\262\212\345F\223H'>ok\330\017T\031$c\214\002,\205[P\261b\021\324\252%\214Dq+\235M)\265\024\010TM\2156\330\252\2052d\300\t\315\032=\255&qck/\311K\035d\254\245\034I6)\311i\003\003\001\030\0305\013\207\010\200B\027]j`Mv\325(P0J\353\340\246\261\233\311\"V,N\324)jA\030T\255J\024\245\025\370\027Ng\013\234:\350\217\340S\237\010b\366\303\nE8\254\370\254\250D\306gp\1770\272\017_\353z\254\336\337\371\023G\366\334\217t\240/tO\351e\341\376+\347x\217o\330\266-h^:t\237\271\356C\3272\013\\\277\245\325zm\033/\\4+,\324\2211\t@.\r\001\220,\007\305l\300\3744\001\3038\246Iw[W\265\243\264\331\207\367\026\240\322B\310\214\024E\3138\372X\213\222x\t\221\270S\025\037PiG\255\374f/\330LP+\005\230x\334p~\276\032m}\010\313%\266\340\246\3137\345J#\375\220K\374\335\271\342\363\006z\361T\024\262V\354C\214\311\013\200\240\307P\254\267E\343\352\305HR\246\\\365\345\303b\312\216\327\217\261\310\261FuJ\217\256\273\342\231\0030\213\240\270\267\023Q\364\371-\316\347u\313B\332|\373<\321\010U""\td\216\233\225\023m\024(\374\003\0102>\313d\330,\330\267f\357$E\027\260`#\232P\307\211d+\227\305--V\365\221\241\264\\y\304\345o\005\326\367\332c\326\231\3665_B\345\235\021\352u\303\220\331y\254*\234\245>j(\332E\354e\253\373wy\3208\371W\206\027\304\206[\003\252t\260\306\036g\002\363af\262P\241<&s\222-fY\254,\"W\303$\274\234`\013\240oT\274\031@4\202Fn&%\375\377\337\222.\350Ym\356\277c\345\253\000,-\316\355ekv\014vX\rh\035\313|\0278\200\035\031\204\024R,\335\341\326\365\305\360E\353?^\240\324\031\002\275b\030@\300\274?5\275k\372\367\336\215ayyj\312s\024\244\360\317\363\326:6Wm-\256a+\275S\326\\\325\327\016\311\332\346\364\305\377\0007I\301\244\032\300q\007Bf\373\021\201\216~\257i\246\317\031\340}u\227\3644\375\215\311\264\303Za\"/W\235{\253\347\214\310B!\367NsA\316.s\234\247\252>\177\2531\226XBe\221\r\001\014s\245N\277\201T\024\375\336\252\020\263e\371\353&&\3336\233\205\332\327/\351y)\337\372\235\374\232\336\217\243b\345\340hW\020\335\207~\2067\227^G'\337\252\203\247'\221Q\341\036\307\221\350\370\313\254\277\232\242\354\\ek\233B\035~\305\rN\354\260\377IRY\373>\027\302\323\201?f\005\323\221y\234\363xmz36\323I\017o\230#\244\355\370\266\352\016\337\207\373^`\265\354\216\357\255\357\270\302\363\234\363\250\2144\006\014\363\351\231\345\036\301\031\232\034\326b\034\205<\232\324\350\2579\323\315\347\310\360r\3426T\344\355\3027\221\3544\373O\345\255}\271>\324\346\254_b2\356\372\030Y\206\270\216\302\370\271\014wtN`\037&d\026\224\362\003\233\203c\335\205\030wx4\366\260\347\320r\307\240o\020\361=\2652\030\364<=\266\351\266\372mM\334\331\300x\216\030w\265\021a\271':\317\031\034\204\212\357\253\247Y^C\264\202\255\001h0\364\330Gf\351\t-l\t\313\320\221&\311\021\300hH\367\033<<\214\263\263\335\334\035\275\277\035\327\265\265w\371 \034\3765u\370\260\014\211/\243\263\315\361G\306\352\321\327\300\357\270\261\323s\274X`\362q\265oj_\007\372\216g\356\360VM\233k.dRfd,\225\321\202H\354u{}\223\225\276\361\353+;\360\344(q\223\245\221\335c\231U:\224;\016\263\311\220\361\215\231\316\202K\251""\207f\227\330\273\235\363\222\353\r\316\266u\322xD_\226*(k\275\021\325\310i8j\265d\263\004g\212.\340B\343\217~\313\202\355\220\234N\3258\260\037\353\270\376<QW\251\240\362lnv[\346rL\214\224\314@q\250\262*\007\215+\000\317\351d\201\260\266<\013v\201\032\027GVS3{uB5\361\274\251\343\353\246m\324\204.\234\266\341\230O\236\345w&\363dX\224\030\354jAxT\332\237\346Z\233\331\330\013\222\212h*\314\353\32729N\216\272\346\001\347V\016;0\242~K&\216\350\322\221\230\023a\366\363\027\233\006r8o\304\354\3452\254u\036\367'\260(\331$\032\330\030r\256\302\323[\003\002p\360\203B\207\243\\\236\365\202]\205\2470\014\264\001c\261e\301\222;\270'\301\023\324\255\225\355\272\317Q\227\371\202\007\001\304\307\0011\210\010\252@\255\267#\267\200Z\250\315&t\240uG\313\331\030\354\315\230\250\365:\345\213\357\352\312/iI \270b\366\"\361\324\n)\274\010\221\201\215\316Lz\247L\007\224pN\267\0160Z_\212\255\306\213.\273\322\325)bb\312\215\262#I\0035`\022\302\")-f\234\352\252\373\206\232\275\250$Rb\300\2778\350\"\245\241*\242\214q\010\243\001\2449R\332?m\326j\031cs\2343\027\204\232\333\234\020'\014\010\3516\034\357\276\332\345\203\270%\224\235)\223;\207\2417)T\357\035\357\004\273\000\030\304\320l\027\002\005\312\362\351q\353\213YoK\3333K+_\361M\222\320\325\201\377heA>W/=%\333\323|*\267\234\231\032\225\271ntjf\260\235C\347\267q2\351\331\323!m\320Q.\032\013U\306g}\220\246*32\272zM)n[\323$n&1\330\305\024d\223\321\027\010\223\037\227\213d\226\006\320\247$\026_\345\022(\324\231\311\376\226q_\244\320\330!T\300\300k\324/Pr#\226^\214\327\207\322\006X\005yEW?<M\365\335\324\367\215\354\323z^a\241t\260ip\255\326\262\307\343$\213Ec\022\260\005\323[\007\214T\021\334i\264U\261\205\244\302\375\361n4;%\212\272\367\203fV\304D\344qUY\r{\263,\227\220E\342\300Y3\214\320\372\364\350Z1\315j\"\340alJ3\316\"E'0\354\241`\263\247\203mJv\310\025D\034R;\\\274\212\232[\013\305\005\215\3214\220\330\365\037q\2473\rZ\322\253#U\03504\263Q\205\316l\262`\3307\3251[\312-\362\246\2268(\033\367\213$\247\262U\026JPL\275\323\245\316!Q\"\334\222""\302\315\273y\303r[\272\014t\356m\367\205\312\030:LQ\303\312\005\245\014V\276\014B\006\306,VA\3249.ZC\034{\003\270\263\203$\243\246*\0216\333\030\306\336TH\245\304E\030\\\313\355\301?\027A\t\026n\343\305\\\344*J\357a,&C\005:@\275\234\363\006\203SD&\243\235\020Z\234L]\326\302$\213\2458Q\006\326jm\241\231\242\230?Jb\301\316\311(uU\346\024\2554EE\242W\304\325^8>\311\307I\255\232N5\322\016\213\263\000\265L\223\306\025\364\016\000t\013\002{3+\327^\030\227(\210\264\356XBK0\366)o\32394\023\265KAx\326\265\007\243\024\367\353\272\036WR\331\310,ay \007\347g;b1\034\216hrFHR\340\331\276\017\206\216+2AfI\253\026\355%\236\324\2104\3625m\364Ap4\3651L4)\226\211cM\266dK9\250\034\211\224\300\221\022\224Cv\333\t@\243D\244a\010\025\2356\367\r\230\365\207\234j\002\270k{\036\307\301\267,\275)>\243K\010az\351^ra\325\267\241Q\275\334\241s4^\332\226\000\270\000F\207\342w+\214\354\320/9\212\226\315a\344\324\211\324\243}@\231X\321(-\206\\^\rkLL2[\030\361\256\007\320\217Cw\347\204nK\001\032\330\\\213\347\007Z\241\212\316Yc\204\256\346\320Tfv\210f\030($!\314gE$e\317\350\357\002\320\014R\342.)\212;N\000\030c\271y\032\320\267\244\365\217&&Z%\216)J\002mmf\016\004\010\034\265\242\362\3636W]<\360\322\021i\035e\216\321v\303\226\200\273B\340I\356m\302\315\033w\360y\265\201\305\034\263?Q\365\354\"NR'\215[\300!Vi\256\032\326a\022\306\3270$\031';B\242\3360\260W\240\233\241\203.L\346wh#>\256c>9\030\2320\233\314\365@Q\264\352b\354\336\266\225\261\330\214L\010\306\014l\030p\334\006\006\206\\\320\211<\364\002\022\330\322\261\023\263\230\032\023\302<\235\031\261\"\363\027\270\024\370\320\034\314\364\257Q\006.\256h\342\002\206\033\356h\010&,\311\342\210F\344\267\nA\302\242\260G<\220a\322\222\0341\242\351\031\263\352\322\352\260\0269C\"\214\226\023\251K;\257\241\244\362\004\013H\203{\357\352\010<\202\345\303\301\270\324e1%\027p\331\034\262c\037;\204\211\314 La\272\032\331m1\027.P\3615\016\252\300\215w*\253\316\027\331\226\313\005#\246\037\267\343e\325\026y3\307\027]\262\300\240\264\240\2339c\214\322\3266r~\004""\324O\241\005X:\033\034\034Z,Y2\224&\322\357\204 _Q\022~<\033\354\226\336\3474 \034f\010@3\3200hg\341,\260\215\242\271\350\030A\003h`\320\302\006A\275\250\226\3710\010d#\214\002T\222\356A\032\254\317Y\271\000\270nW-\263\2550V\213\202L\362\354\021\325\314dJ\234\315X&\203\3230\354+>K\270\304\302\247w\\-iZ[#>\232\016FK\337@\266\212AaJ\270\326\240r\303g|%\203M\330\0304\230\013.\001\341\316q\233\325\020\222#Y\221tc\351R\005H6\244b(``\007\272\013f\216\321\316(\335\253\210<\013M\241\004\325\337\276Q\246j\233\372\354E\230\374\274/\202[\316F\023b\327\331yR\370\242H\343\204%\252\244KO\033\307mj\363LT\000\335\321M\355\272\327\221\321\333\350\334\360\305N\267$\314\031:\225\203\360N\304~\321\234\025\014\342\260\206\201\212'\256\274\262T\n\265Q\204\234*&\211b\364\344\251\200\335~\317\023\217.\204\355v\2642x\240\017\004\367+\323\336m\224`\375\244X\324\006\214R\003n|+[B\234*\365  \325\2008\t\002x\273\233\036\021I,$\265@\275\306\025iR\216*1\020a\033\032\254\307xQT,/v\331ch\022\263A\223\321\264F\336)\"\0200v\271\032\234]-e\317\232\230\013\177\205\352fzUA\215#F\301*\275!\266\213!\007\314j1a%\215`_\204!E\010\244B\317c\312\022d\034\002\023(\352d1s9\275?\246pK\3158\243\210InKI\315\224\346Hk*\304ji\034=\261\342L-v\353\355\202]\027\312\375\257McAnP81e\254\314\314\250\030\r\220&fu\363\027[\225\374@&\223H'\271\305\312\327\342\232\261\317Ti\027\247\270~\255\315\233r\265\033T\207w\034\237\205\364\262U\201\274\034\342\r\231\240\256\020\336/6&f\0037\337\240:\226+#\324\023#\005\335\202Q2\222\210\215\3144\305e\025r\026Q\036\261\300\016C\016\371x\264\346\rP\\q\021\350Q\216sn*is\t\366\215y\220Ng~\340\342 \264\024\003I\265\227\301\234\347\307\230\344\357|1\347\363\030\256\335\275\023\370\220\323\277i\2249\037W\227/b\361\370\250\033XN)E@:\252C\326Z\216\034S\321\334\305\006\341\351\021\270\323\032L2\261\030\033\247K\356i\031w(\323m\215/m\266\331\267\375\"(\3504E\242\275\361@J<\375\350\304oH\006\266\352\360\277\000\334\375uu\362\231\363\310\310d\365\246.xf\235.&\177\343-\236\036\264\"\033cI\361K\304""\326+;\2238\276\025r\024\327B\206\251\312U\32296\261\262\307\023f\227\277\\r\260\302\255A\274\005\355j\2725XI\316/\240\356!\257\240*\257=-Qv\326\252,\342\215\033\361\024\035\246\314\251\346>\301\262\267\301\371);\352P\nK\\:\240\361d\330\316\244|\036S\273kL\331\366\315)\377RY\006r\n\314\310\250\266\223\366!\0200\244\205u)_\333\242\206\271\240B\032\233X\234\242\340\211\3163\207\253\230\235\375:)\013\212MrH\347\337\031\206\031\275\320\331B\031\366.\232\310S}_Y\220C\270\212r\213\232\021\217\274\224\325\370\361\260\230\260\265\025\306\255jDF#4Y\330\023+DF\245w\244g\235\264\377\365GG\205\217\217(Us\336\027\335\256S%\303\2526@\2700GO\"RWK$2/XU\td\321G\026\006\212BO\021\\\254fm(\326|n\\\036\246\214v\tR\363\242GK2\037R.UGT{X\251\223\213\370\251*\344\223\2073\373\270\342\321\264\020\342\2624\352\024X\241\324\241b\207O~b\n*\230{\0203\306`\001\345g\225\207\360O\263H\254\"f\330\276,\2410\323\000\303\265@\202%|*$\224@F\230Ca\0169\036\356\233\200l\2227\357\331\013\017K\307(n\335\233,\004&\202CR\307(h\334\321i\240^\340\010G\336\025Aq`\333\3613\000\217\236\314\215@\322\014\212q\346h\354\0375\346\262\\Q\031\010Pu\302\020\220\226\"S\275\212LB\252\226`\003\240(\252\210e\t\325)U\020\302CCB\240QVU\207wR\232b\254\223\302yv\240\"\221\006^$\264\321\251\nP\225\222\311Hm\307YIy\013n./\205\215\335\\T\276\310\225\254\235qw$S\205\t\000c*\025\200";
    PyObject *data = __Pyx_DecompressString(cstring, 4068, 2);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) != 0 /* compression: zlib (3856 bytes) */
const char* const cstring = "x\332\325XKs\333V\262\266\306N\"9~H\266\254(\261\235\034\306v$O2\214i\313\216\306\223\232\014L\321\226&\226\304\227l\271n\246NA\000(b\014\022$\000Jdr3\225\245\226\\b\211%\226Xr\251\245\227\\r\231\237\340\2370_\037\000|\210Tr+S\267j\342\262x\272\373\274\372}\272\221\374\341\361\375{?>f?<\376\032\343y\206\177\243\024`\367\201\335k\n\274\331'\214n\030\241\214\342\322:\223\034\307z|^2\234$\033\305\262\337j-\246WU\255)\320\232i9\254\332\250\354i\226\300\035\275\242\231\r\347\211a*\257\231\255\031%VQd\333a\206i\326\366d\345u\372\305si\213\331\216\245\327jzu?m\2306\006\226\313\246\315J\215m\312\366\353\264\245\311\216\306\236\032\362\276\275\246ak.\033]\220Ig\036\237\317T\223\254\220[c\262\335\252*\254j:z\251\225i\3264\305\321T\226\316}\231\316ev\277`\373\246\303~p~|*\353\006\310\216\311\236\200e\266\371r@\220\r\260H\027\025\236e\330^\243T\322\254\t\223\252\354\310c\263J\310a.;F\322\232\216\006\325\250#s*\204\260\314\326\010\255b\252\340{2\t\332\241\263@\312\027\363\2772_\370\305y\275\252;\203\0055\023\312\2644\345\340\004\311\006\313\003R\275\241Y-\006U\237$\r\363jk\016-\331\250\252:\016tt\263\312\212\362\236\241m\312M\226\207m\034\263\242+\204\024\013\333\254\254\311\252f\021F\232H\016-\320\253,\277\225\027>cm\231\304sYvX\272\345\224q\240nc\275\241\303\356\220\306h\t\237\201\221-ZTe\331L\366O+\253+L\206Y-\355\237`\302fvcO1d\333\326lf\226`5\335pp\203\323\252iv\222m\224X\313l\260\252\026I\216u\303\033\234\262V\025b\001`Kr\025\216%\223\\\034\333\341\240K,\222\364@(\366\251l\330Z2\233\336`\207\226Nv\007\0275YU\261r\020\037Y\331)\263\315\342\216\030+\372~24LVW^\033\344\363\340\02125\024\207U4ro\342G)\303\261E\334$\025\271\306*\r\330g\217\374\252f\350\212\356@\r\226\006s\330\344\354\207:\316\375\233\334pL^\2433\265\345\242\325\320\356\376_\317\227\021\323\377\257\027h\212\366\333\316\037\244\0248\263\254(\032LU\242t\000\024Z\261\037\237'\240aYZ\325\tu\032;=F279\225\2459V+\237c\331\302V~\227\225e\273\234\247\2500\364\212\356\344i\216)f\243\n\177*\2105\370U-\031\001S\335\007h""\223\265\220\004\nH\002\344rO\323\205\202\331\260\024\n\372\255\235*<\301\334\257\352\337C\006s\217|O\334\372\002^`Zlc\215}\003W\340\360 ,CR\253 \251E\371AH\241\3526\305\212V\r\177\033\025&\234\220\351{\007\034n\302\303\360\305v\246W(\275\3222V\262\314\n\253\265\0164k\317N\"*\260\230@N\373\341\335\257\340\332\262\245\261\232e\356a}\2135DZ\225\231RS\265\022\023\267\250\032\356\261\3003\202Bf\311ZSe%\004t\030t\252\211 \240;\3132\\\234\246[\214A\032\214M\261,y\222S\247\361_\317c\275\306\311\3139Y\340\367\300,\345\353\337\027\307\277\017_\005\243\024\240\377}|\226\014\363\220\031\362\236f\354+e\263\026f'\342\370\320\342\224;\014\215\222\206xv\223x\017\367\rpb\350v\230:\324\312\310\003K\230^Eb\216\252\026v\202@(\275\376\242\340a\303X\037~\231\037\232x\231\357\203\366\276f\367\221CK\300T5\014\216\212\261>\034\037\025!}0>J 8\212E\231>\256##s|Y\257\221\226`8\255I\225\000i1|\tt\262\034\014\201\\\274\361\344\005\317e\371\306\326F\221K\305b\236\0272[k|;[\340O\237K\317\n\354P\0167\322[\017\005\016^=E\274\357\364.)f\025\3679Q\325\307\262a\351\021%\365\022lU\213\036;\261Y\350\365\267n\256\327\376\203\255_D#\251\343\267\234\262\257\253aQb\207\317X\215j\222=\035ES\370p\322\217\2560*\265\226\035K\243\330\220\367+x_\357\256A\346'\242\000\036@I\316\341\363\rE\343\212\270\222\363\2219\224Q\342\320\376,JE\374\037\337\025\022\307\226\223]77\236q)\277\231Y\213\021\374\345\245\342\000\317g0-\220\342\016O\335\273\277\022\303\367\357\255\254\366\341\207\217bp\345\336\237\373\360\303\324\375\320w\n<\223\317\307 \371Q\014\3473\205\314\000)\346\007`!\006\321\204\014\300L\014\356l}\273\265\375r+D\213|-\277\361\"\223\217\261\274\364\222g\245\364\267\361\321\240\244chg\000\255\305\320n>\rN\322/\206q\362\361\310\361\245t:S\210\234=&=/\342\206\342z\214F[yZ\312\306\320\016\342\244\010MF\370Z\246P\304\030q\3143[$\030\227\n\257\266\322|k\273\270\361\364\325X\230\245\241\373bf\344\336\301\344\006B\260(=y\236\031\233\331\224v9\322\025_\317Hk\261N\206\246\263kc\244\374._\227\n\353\277\022\347c\323PS|\026""\335)$\314C\246\342\366\346Fzh\002\303\030yc\213\243z\344\305\215\315>\213\244O\341r\303j\013\211\305\235\030\3756\363\212D\317\354\306\204\355|dd\236\303\\\004\222\007\363\347\033\233\261\243\301\272\305\374+\236\336\352\343\270[\320b<\307Q\234FHa\004\031b\206\270\335\336)\222\006\266k\324\265\330\331\265l\230H3\226eZ\021\234_\333\224\200W\315Q\274?\277c\243\307&\234\363l\253\211\27754\\|\013Mm^+\345\320\327\344\262\271,}\n\010\177\307\343\271O\037\013\351xF1k-.\227\321\356\3135\3613\351\220\220<\341\214L\223\376&m\021\324\211;\222xJ\345=\274.}D\364\236\\\251\240z9\254\235\240\2264G)s\224\357'\350\242\315\213i{\310\244\274r\030\243\361C\035\343%\243a\227c\204\276#\030(\354\016b\202\205g\033\374\313\352\010a\344\374\001\205\353\225JL\245\027s\030>97|\211\300\035\333\034\340\016\017K\001No\330)d\216\322\302\031\236\303c}\002\035[\322PI[\3260\251i)\334\266\352\325>\315\221I\373\033(nC\347\211!2\347\000\236d\326\221\331\t\346\035\232\027\255\025G+i\217^\221\244Rz\214\376K\227M\276j\322\216\323V\206\335\036\312\316\021\004o7\341\373P\021\271\214\250\361A\350{SR1L\033\nO\206_r\000\320\347\031NEW\014\207>\220\024\037d\34214\235^\345\246\245R\307\034Q\321}\003&{\204\0204@\355E\010Y\244\213\020\260\361\272\341U+\206\337\355^j\362\353\202\346P\006\r\313->\254WX\232\214-\227\345\262hS\30478\335L*\246\205\275p#\273\037.}\200+be\334\324\354\311*\210Bf\275Z2#\341\367\032%\341Y\350\3569\027u;\251\t\001P!\224\374\023\362\341\024E\243o\210BO}\025\366\001\233 \024\343q\222\tG\372`\020\231 \024e`\217a\313D0\204\024\220i)NSiX\375\036\207\224,\024\rVUm\257\261\257\242\244R\r\374\331h\363\265\2228{XWT\332sQ\332\243\020\253\310\016\347\245FU\341|\324\376\002\r\035\210\343@\332\270o\321\307\n,\251\226\312\252\205\377\334\376\036\275\001\017{\203\221\217\004Q\037~\262\325\235\330M\0167lC=\021\222\210p \262\203\350\212\310*<\334;8\001\251\005}\207JlZ\257\265\226ns\331@\352\2002c\313\003\216\272\023\030\3156\264\352\276S6\264\003\3150\260\236\203\003\030P\360!\244\206\230Ca)\214\207\246`8""\027\021J\236OI\247\017\037R\223\023&\267\210,\340\220\214t\307\243N(\326O|\276\210'\\o\252\r\003\212\256\3306\264\026O\036V\016EF\257\264x\350\234\234W\345\n\226\241\033!k\323\200\333\314\232\031>\255T:\223\326k*(\324\266\320\302~\244\366\303\264\346X5Rp\334_\356\311\266vZ\257\031\323\243\221k\364\010\307\304z-\206\240h\303\346\034}Q$\236\000\017\204\3318\257\313u\235\376k\221+\220\203\233U\007\2577 \334\022\033?\036#\371#?\210\206\210\210\323\032\262\021\252A\244\223\t\231\246\237f\0201J]\330G\251\367\223$\037K\227}\202\326$\260b\202\005\312&\021X\257U#\210<,\002\351\031!\324V\352\324\000\tk\023\210\301\254E\271)Jq\"\025G\226\003\204\016]n\300G\007\t\232OH\326\366\276H;\324\224R\364\205\035\222\021\366H\304\221]\3476\234I6\014\333R\360\237\013\343S\337\004\316\302\236\211S\317d7j\232e\037\3125\\H\315\024\347Q\272\rsm\230h\221\277J%\252<(\277!\236\032\232} >\031r]=\224\255*\375Q\257v\210\034li\245C+\256\\\306\212\226\t\365\312\211ReP\245\014\025(qm2\\\226\014W$#\245\307X\035\022\225\025C\325\307P\3411Ts\214\327\025\247T\032\203\252\342D}1ZZ\234\250*\342\202\002\270\232mm\207\275\356\037\227\007\340\3353\342}\340\221\032\355\356\271/|\351\355\21533s\335\331\317\375\224\277\026\274\023\024;s?I\275\231\353n\312]\363\246\375w}\371'\351\355\20533\027\216\036\034\251\355\317\335\2027\325\273p\361\347\351KG\206\233pW\334\206'y\005\177\2527\375>\026(\355\271\336\205+\335+\237z\211\356\255\325\316\024N\002}\345\310i?h\313\275\013\227\001\244z\263\363\355g\356#o\316K\364.|\2043\350\242w=\325_\016n\007Xt\245\375%\346n{b}\2757\273\320]\370\314\223\273K\337t\352\275\013\263\355w\351\240\376\360\276\233\242\341\275v=\032\210\367\217\004cMo\0049\360s\020d\201\004Y9\372\336\235ws\321\016\027\274\023c\337x\t/E\340\217.\270\276tTu%\267\350]\363d\357\320\337\017r\023\004|v\214\033\336.\2368r\332\235r\347\376\243#{\323\213\356\331\356\365\325\316\331N\252\267x\275{=\351+\301\\o\361\006\235r\255\215}\027\2162\264\353\362\321A;\017\321g?v\r?\341\247\010\220\351>!\371\rq\333\242\177\305""\177\030\\\350Ht\356m\017?\227\273\227\031\024.{M\337\302\261`\342a\373,\354\002\315\177\341\312n=<\376Z[&\245~\320\376_o\025\256!\221hS\355+\355Uw\325{\010{G\350cw\327\333\361\023\275\231;^N\330\360.x\306\001\322Q\216.\274\322Nz3\335\273_w\322\235\037\337H\275\304WA\263\363\303\233D/q?\270\325}\370\364X\356%\356\005S\275\304\327\035L~\332c\313\335e\270M/\361 x\320}\264\376f\212\246\317\006\251\336\334'n\253\273\374\227\316\203N\345M\252\307\036\005\345\316\3537s=\226\n\256vW2\307\271\036\273\027Lc\343\211uI\241\346O\226\274zon\311k\005\227:\373X\273\224\364\345\336\334\035\357U7)\035\317\365\022\241z\331\262\377>\335\n\023\010\366\002~,\235`\214\354\361\353\373r\247\212\3613L\323hK\355\242{\325\335\200\2764\377\001\205\027\351\353Y\367\352m\021I\347\203D\260\022\324#j\373\257\210\205O\275\214?\347\337\026Kg`\377\266\352&\010\274\351\346\\\325[\366\t\231Nxs\003\312g\301\305\216\320\377\035\357\245/\371\360\230\331\356,\363\246\274yo\327\037\231\230\371\304\255{\347\274u\377i\360\247\216p\233E\370\357<\254Z\360\337\363\355`\251s5t\035\341Y\024\364\237\341\030r\252T\030\332v;1\331s\340\205\355?\264\227\335%!\201\024\023\376\350>\363R\336\223(\274GV\364\331\277\033\014\373\320\245\356\245\217\305\3513\261\004;Abt\342Cw\316\275\355\226\241\320\325 EB,\200\007\307M\r\255r\274\324\204x\373kG\036,_\365\036\3717\341Y\360\021\304\365\37133\227\305\342\221\300\265\341\353\343\307<\356$\342\260\233^h+\356\202kC\214[\320\357\302\207\335\017?\367W\374zo\001\231\2507}\265\235\032\n\337F\373\t\362\330\354\r\367\3578\272N\200\324\017\337X&\301`B\244\304\262_\n6\217\221\237\337^\"\346\342\004\374\004r\247\374gA\352\264$\374\366\n-\377\365\354\0031\350\262a\227h\275\301~\362\333z{\006i\255\357\035\377\023\374!\270\023\324{\037R,A\240u\2705im\376\314\314\354X\346\273|\364/a\331;^\035Nu\210,?~\367\372\233\210\321Y$507\337\376{\350q\025\361@\274#\322\313Rp\275#u^\035O:\340\317\035\221u\261\371\247\251\336\271\363G\237\267s\335s\202\273\367\217V\303\367\347b\350\262\342\255x\247\275Co""\017\275F\031w\021n\265C\021\006;\315_\207#\254x\r\274\206S\010o\261t\027\212\373\030\214\227;\352q\342\027I7\241\332\335N\361xh\343\\o\366\203vS8`D\333E\360\255\007R\214\276\360\276B\342\214\017\351\336L\2123r\3614\207iW;\375\315\377\200\026\026;\375\363\277\363?\240\314\367\363I\221nxE\377\232\257\"\233\244&\212\364\035D\316\235\340\3544\332?\220\006\026\217\347\217s\277\264l1d$F\027\340e\345@\216\320.2#\235\321\277\374\223\340=<\005\365\030\2759,\362w\336\r\377e\260\336\221\272\347\340\221\364f\017;\257\345\177\024\244\002i\310\003\360\022\241L\240\254\343\276\353\376\023\331\370\253\316\322\361\t\267\236z\373\361\231w\246\177r\260\007\0361\323;\367\036\340\024\342\360j[\n\363Z\"|\t\317\265\267\275\007p\265S\320\356\374m/\3559a\366\233\210/!G\300\223z3\027\217rG\232p\2740\000\302\270\237o\363x\303$\264{\355\0162\3579_:\025_\366\317R>\371y\372\342\321.\005\312\371\243;\"\340n\301\340s\223\321\356\325[\261\020\247\340\237y\212?\357\347\376\rp\r\021\215";
    PyObject *data = __Pyx_DecompressString(cstring, 3856, 1);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #else /* compression: none (9536 bytes) */
const char* const bytes = ".{:20}: {:<20}\n    {:20}: {:<20}\n{:22}: 0x{:<20x}\n{:22}: {:<20}\n    {:22}: {:<20}\n  {:22}: {:<20}\nAH Attr:\nAlt. AH Attr:\nAlt. PKey indexAlt. port numberAlt. timeoutBlock self mcast loopbackCVLAN strippingClosing QPComp MaskCreate FlagsDest QP numberECE:\nEn. SQD async notifyExpected CQ/CQEX, got {t}Failed to Bind MWFailed to allocate SGE bufferFailed to allocate data bufferFailed to create QPFailed to create extended QPFailed to destroy QPFailed to modify QPFailed to modify QP state to RTRFailed to modify QP state to RTSFailed to modify QP state to initFailed to post recvFailed to post sendFailed to query ECEFailed to query QPFailed to set ECEIndirection TableMax RD atomicMax TSO headerMax dest. RD atomicMin RNR timerNote that Cython is deliberately stricter than PEP-484 and rejects subclasses of builtin types. If you need to pass subclasses then set the 'annotation_typing' directive to False.PCI write end paddingPKey indexPath MTUPath mig. statePickling of struct members such as self.cap must be explicitly requested with @auto_pickle(True)Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)Pickling of struct members such as self.ece must be explicitly requested with @auto_pickle(True)Port numberQP access flagsQP caps:\nQP current stateQP stateQP typeRNR retryRQ PSNRX hashRate limitRetry counterSQ PSNSQ drainingSQ sig. allScatter FCSSource QPNUnrecognized object typeVendor ID ?add_notecomp maskcreate flagsdisableenableenum class ibv_mig_state not importable from pyverbs.libibverbs_enums. You are probably using a cpdef enum declared in a .pxd file that does not have a .py  or .pyx file.enum class ibv_mtu not importable from pyverbs.libibverbs_enums. You are probably using a cpdef enum declared in a .pxd file that does not have a .py  or .pyx file.enum class ibv_qp_attr_mask not importable from pyverbs.libibverbs_enums. You are probably using a cpdef enum declared in a .pxd file that does not have a .py  or .pyx fil""e.enum class ibv_qp_init_attr_mask not importable from pyverbs.libibverbs_enums. You are probably using a cpdef enum declared in a .pxd file that does not have a .py  or .pyx file.enum class ibv_qp_state not importable from pyverbs.libibverbs_enums. You are probably using a cpdef enum declared in a .pxd file that does not have a .py  or .pyx file.enum class ibv_qp_type not importable from pyverbs.libibverbs_enums. You are probably using a cpdef enum declared in a .pxd file that does not have a .py  or .pyx file.flow labelgchop limitibv_wr_complete failed.is globalisenabledmax TSO headermax inline data  max inline datamax recv SGE  max recv SGEmax recv WR  max recv WRmax recv sgesmax recv wrsmax send SGE  max send SGEmax send WR  max send WRmax send sgesmax send wrs  numberport numpyverbs/qp.pyxqp_ex is not accessible since IBV_QP_INIT_ATTR_SEND_OPS_FLAGS was not passed.self.attr cannot be converted to a Python object for picklingself.data cannot be converted to a Python object for picklingself.qp cannot be converted to a Python object for picklingself.qp,self.qp_ex cannot be converted to a Python object for picklingsgid indexsource path bits  statestatic rate(tree fragment)DataBufferDataBuffer.__reduce_cython__DataBuffer.__setstate_cython__ECEECE.__reduce_cython__ECE.__setstate_cython__IBV_MIG_ARMEDIBV_MIG_MIGRATEDIBV_MIG_REARMIBV_MTU_1024IBV_MTU_2048IBV_MTU_256IBV_MTU_4096IBV_MTU_512IBV_QPS_ERRIBV_QPS_INITIBV_QPS_RESETIBV_QPS_RTRIBV_QPS_RTSIBV_QPS_SQDIBV_QPS_SQEIBV_QPS_UNKNOWNIBV_QPT_DRIVERIBV_QPT_RAW_PACKETIBV_QPT_RCIBV_QPT_UCIBV_QPT_UDIBV_QPT_XRC_RECVIBV_QPT_XRC_SENDIBV_QP_ACCESS_FLAGSIBV_QP_ALT_PATHIBV_QP_AVIBV_QP_CAPIBV_QP_CUR_STATEIBV_QP_DEST_QPNIBV_QP_EN_SQD_ASYNC_NOTIFYIBV_QP_INIT_ATTR_CREATE_FLAGSIBV_QP_INIT_ATTR_IND_TABLEIBV_QP_INIT_ATTR_MAX_TSO_HEADERIBV_QP_INIT_ATTR_PDIBV_QP_INIT_ATTR_RX_HASHIBV_QP_INIT_ATTR_SEND_OPS_FLAGSIBV_QP_INIT_ATTR_XRCDIBV_QP_MAX_DEST_RD_ATOMICIBV_QP_MAX_QP_RD_ATOMICIBV_QP_MIN_RNR_TIMERIBV_QP_PATH_MIG_STATEIBV_QP_PATH_MTUIBV_QP_PK""EY_INDEXIBV_QP_PORTIBV_QP_QKEYIBV_QP_RATE_LIMITIBV_QP_RETRY_CNTIBV_QP_RNR_RETRYIBV_QP_RQ_PSNIBV_QP_SQ_PSNIBV_QP_STATEIBV_QP_TIMEOUTINITOptionsPDPyverbsErrorPyverbsRDMAErrnoPyverbsRDMAErrorPyverbsUserError__Pyx_PyDict_NextRefQKeyQPQPAttrQPAttr.__reduce_cython__QPAttr.__setstate_cython__QPAttr._copy_ahQPCapQPCap.__reduce_cython__QPCap.__setstate_cython__QPExQPEx.__reduce_cython__QPEx.__setstate_cython__QPEx.wr_abortQPEx.wr_atomic_cmp_swpQPEx.wr_atomic_fetch_addQPEx.wr_atomic_writeQPEx.wr_bind_mwQPEx.wr_completeQPEx.wr_flushQPEx.wr_local_invQPEx.wr_rdma_readQPEx.wr_rdma_writeQPEx.wr_rdma_write_immQPEx.wr_sendQPEx.wr_send_immQPEx.wr_send_invQPEx.wr_send_tsoQPEx.wr_set_inline_dataQPEx.wr_set_inline_data_listQPEx.wr_set_sgeQPEx.wr_set_sge_listQPEx.wr_set_ud_addrQPEx.wr_set_xrc_srqnQPEx.wr_startQPInitAttrQPInitAttrExQPInitAttrEx.__reduce_cython__QPInitAttrEx.__setstate_cython__QPInitAttrEx.flags_to_strQPInitAttrEx.mask_to_strQPInitAttr.__reduce_cython__QPInitAttr.__setstate_cython__QP.__reduce_cython__QP.__setstate_cython__QP._create_qpQP._create_qp_exQP._get_comp_maskQP.bind_mwQP.closeQP.modifyQP.post_recvQP.post_sendQP.queryQP.query_data_in_orderQP.query_eceQP.set_eceQP.to_initQP.to_rtrQP.to_rtsRTRRTSTimeoutWeakSetXRCDaccess_flags_to_straddaddrahah_attrasyncio.coroutinesatomic_wratomic_wr_cattrattr_maskbad_wrbind_infobind_mwbuf_listcap__class_getitem__cline_in_tracebackclosecomp_maskcomp_maskscompare_copy_ah_copy_capscreate_flags_create_qp_create_qp_excreate_strcreatorctxcur_qp_statedatadata_bufdebugdgiddliddstecefflagsflags_to_strflow_labelformat__func___get_comp_mask__getstate__gid_strgrhash_confhdrhdr_szhop_limitiibv_mig_stateibv_mtuibv_qp_attr_maskibv_qp_init_attr_maskibv_qp_stateibv_qp_typeimm_dataind_tableinfo__init__init_attrinvalidate_rkeyis_alt_is_coroutineis_globalitemslengthlevellkey__main__maskmask_strmask_to_strmasksmax_inline_datamax_recv_sgemax_recv_wrmax_send_sgemax_send_wrmax_tso_headermig_state_to_strmodify__module__mssmtu_to_strmwmw_bindmy_bad_wr__name""__num_bufnum_sgeopoptionspath_mtupdpopport_numpost_recvpost_sendptrptypepyverbs.basepyverbs.libibverbs_enumspyverbs.pyverbs_errorpyverbs.qppyverbs.utils__pyx_state__pyx_vtable__qaqiaqiaeqp_attrqp_contextqp_numqp_stateqp_state_to_strqp_typeqp_type_to_str__qualname__queryquery_data_in_orderquery_ecercrcqrecv_cq__reduce____reduce_cython____reduce_ex__remote_addrremote_qpnremote_rkeyremote_srqnrkeyscqselfsend_cqsend_ops_flagsset_ece__set_name__setdefault__setstate____setstate_cython__sg_listsgesgid_indexslsource_qpnsq_sig_allsrcsrc_path_bitssrqstatic_ratesuperswapt__test__to_initto_rtrto_rtstraffic_classvaluesvendor_idwarnwarningsweakrefwrwr_abortwr_atomic_cmp_swpwr_atomic_fetch_addwr_atomic_writewr_bind_mwwr_completewr_flushwr_local_invwr_rdma_readwr_rdma_writewr_rdma_write_immwr_sendwr_send_immwr_send_invwr_send_tsowr_set_inline_datawr_set_inline_data_listwr_set_sgewr_set_sge_listwr_set_ud_addrwr_set_xrc_srqnwr_startxrcdPyObject *(PyObject *)\000close_weakrefs\320\004,\250A\360\034\000\t\021\320\020+\2501\250D\260\005\260T\270\021\200A\330\t\033\2301\230D\240\010\250\006\250a\200A\360\014\000\t\014\2103\210d\220+\230S\240\001\330\014\r\340\010\016\210l\230!\2304\230u\240A\240S\250\001\330\010\013\2103\210c\220\021\330\014\022\320\022\"\240!\320#8\270\001\200A\330\010\013\2104\210t\2203\220a\330\014\017\210t\2201\330\020\024\220G\2306\240\021\240!\330\014\032\230!\2301\230D\240\006\240d\250(\260$\260a\330\014\022\220/\240\021\240$\240a\330\014\017\210q\330\020\026\320\026&\240a\320'?\270q\330\014\020\220\006\220a\330\014\020\220\006\220a\330\014\020\220\013\2301\330\014\020\220\007\220q\330\014\020\220\007\220q\200A\330\t\032\230!\2304\230x\240q\200A\330\t\032\230!\2304\230x\240v\250Q\200A\360\026\000\t\014\2104\210z\230\024\230Q\330\014\020\220\007\220q\230\001\330\010\017\210t\220?\240!\2401\330\010\017\210}\230A\330\010\016\210n\230A\230T\240\025\240a\240w\250g\260Q\330\010\013\2103\210c\220\021\330\014\022\320\022\"\240!\320#G\300q\200A\360\030\000\t\014\2104\210z""\230\024\230Q\330\014\020\220\010\230\001\230\021\330\010\017\210t\220?\240!\2401\330\010\017\210}\230A\330\010\016\210n\230A\230T\240\025\240a\240w\250g\260Q\330\010\013\2103\210c\220\021\330\014\022\320\022\"\240!\320#G\300q\200A\330\010\030\230\003\320\0338\270\003\2701\330\030\033\320\033.\250c\260\021\330\030\034\230A\330\010\025\220Q\330\010\014\210E\220\021\330\014\017\210v\220R\220q\330\020\036\230l\250!\2501\330\020\036\230a\330\010\017\210q\200A\330\t\034\230A\230T\240\030\250\022\2505\260\014\270A\200A\330\010$\240A\330\010\017\320\017 \240\006\240a\240x\250r\260\021\330\010\013\2105\220\003\2201\330\014\022\220,\230a\230q\330\010\014\210E\220\025\220a\220q\330\014\027\220|\2408\2501\250A\330\014\020\220\001\220\022\2208\2308\2405\250\001\330\014\020\220\001\220\022\220:\230X\240U\250!\330\t%\240Q\240d\250(\260)\2701\330\010\014\210A\210Q\200A\330\010\022\220.\240\t\320)<\270C\270}\310A\330!7\260x\270{\310!\330!2\260#\3205F\300a\330!0\260\001\330!<\270A\330!\"\330 (\320(8\270\001\330!3\2603\3206H\310\001\330!0\260\003\2601\330\021\037\230y\320(;\2703\270m\3101\330 6\260h\270k\310\021\330 1\260\023\3204E\300Q\330 0\260\010\270\001\330\021\037\230y\320(;\2703\270m\3101\330 .\250g\260Q\330\037'\240q\330\021'\240y\260\016\270g\300Q\330'.\250a\330\021%\240Y\320.A\300\021\330!.\250c\260\021\330 (\250\013\2603\3206G\300q\330!2\260#\260_\300A\330!<\270A\330!\"\330 (\320(8\270\003\2701\330\021%\240Y\320.A\300\021\330!.\250c\260\021\330 (\250\013\2603\3206G\300q\330!2\260#\260Q\330 (\320(8\270\001\330!3\2603\3206H\310\001\330!0\260\003\2601\340\010\017\210u\220A\220T\230\023\230I\240Q\240e\2503\250a\200A\330\010\014\210G\320\023$\240A\240S\250\n\260!\2604\260q\200A\330\010\014\210G\220>\240\021\240\"\240E\250\021\250$\250a\200A\330\t\025\220Q\220d\230!\200A\330\t\035\230Q\230d\240(\250!\200A\330\010!\240\021\330\t\035\230Q\230d\240(\250&\260\r\270Q\200A\330\010%\240W\250A\250Q\330\010\020\320\020 \240\001\240\024\240X\250Q\200A\330\010%\240W\250A\250Q\330\t\037""\230q\240\004\240H\250F\260-\270q\200A\330\t\030\230\001\230\024\230X\240S\250\007\250s\260'\270\023\270A\200A\330\010\036\230a\330\010\016\210l\230&\240\001\240\030\250\022\2501\330\010\013\2104\210s\220!\330\014\022\220,\230a\230q\330\010\014\210E\220\025\220a\220q\330\014\017\210q\220\002\220(\230'\240\021\240\"\240A\330\014\017\210q\220\002\220*\230G\2401\240B\240a\330\014\017\210q\220\002\220(\230'\240\021\240\"\240A\330\t\035\230Q\230d\240(\250)\2601\330\010\014\210A\210Q\200A\330\010\016\320\016\036\230a\230q\330\t \240\001\240\024\240X\250U\260!\200A\330\010\016\320\016\036\230a\230q\330\t\031\230\021\230$\230h\240e\2508\2601\200A\330\t\026\220a\220t\2301\200A\330\010\016\320\016\036\230a\230t\2401\330\010\013\2103\210c\220\021\330\014\022\320\022\"\240!\320#>\270a\200A\330\t\026\220a\220t\2308\2406\250\035\260h\270g\300Q\200A\360\n\000\t\017\210c\220\021\330\010\016\210n\230A\230T\240\025\240a\240s\250!\330\010\013\2103\210c\220\021\330\014\022\320\022\"\240!\320#:\270!\330\010\017\210q\200A\330\010\026\220c\230\026\230s\240(\250#\250Q\330\026\031\320\031+\2504\250q\330\026\032\230!\330\010\023\2201\330\010\014\210E\220\021\330\014\017\210u\220B\220a\330\020\034\230J\240a\240q\330\020\034\230A\330\010\017\210q\200A\330\t\031\230\021\230$\230h\240a\200A\330\t!\240\021\240$\240h\250f\260M\300\021\200A\360\016\000\t\017\210l\230!\2304\230u\240B\240e\2501\250G\2601\330\010\013\2103\210c\220\021\330\014\022\320\022\"\240!\320#8\270\001\200A\360\022\000\t\017\210n\230A\230T\240\025\240a\240w\250g\260Q\330\010\013\2103\210c\220\021\330\014\022\320\022\"\240!\320#:\270!\200A\330\t\037\230q\240\004\240H\250F\260-\270y\310\001\200A\340\010\017\210q\220\t\230\021\330\t\030\230\001\230\024\230X\240[\260\002\260%\260q\330\031.\250a\330\010\014\210H\220A\220Q\200A\360\024\000\t\020\210t\220?\240!\2401\330\010\017\210}\230A\330\010\017\210~\230Q\230d\240%\240q\250\007\250w\260a\330\010\013\2103\210c\220\021\330\014\022\320\022\"\240!\320#H\310\001\200A\360\022\000\t\020""\210v\220Q\330\010\024\220J\230a\330\010\016\210m\2301\230D\240\005\240Q\240d\250'\260\033\270A\270Y\300a\330\010\013\2103\210c\220\021\330\014\022\320\022\"\240!\320#9\270\021\330\010\017\210v\220Q\200\001\330\004\n\210+\220Q\320\004.\250a\330\010\013\2108\2203\220a\330\014\r\330\010\013\2104\210q\330\014\020\220\005\220U\230!\2301\330\020\024\220E\230\030\240\024\240U\250$\250a\250q\330\024\033\2308\2404\240u\250D\260\001\260\021\330\014\020\220\005\220X\230T\240\036\250w\260h\270d\300!\330\014\020\220\005\220X\230T\240\036\250w\260h\270d\300!\330\014\020\220\005\220X\230T\240\035\250g\260X\270T\300\021\330\014\020\220\005\220X\230T\240\021\330\020\027\220x\230t\2401\330\014\020\220\005\220X\230X\240W\250H\260A\330\014\020\220\005\220X\230V\2407\250(\260!\330\014\020\220\005\220X\320\035.\250g\260X\270Q\330\014\020\220\005\220X\230_\250G\2608\2701\330\014\020\220\005\220X\230]\250'\260\030\270\021\330\014\020\220\005\220X\230\\\250\027\260\010\270\001\340\014\020\220\005\220U\230!\2301\330\020\024\220E\230\034\240T\250\025\250d\260!\2601\330\024\033\2308\2404\240u\250D\260\001\260\021\330\014\020\220\005\220\\\240\024\240Q\330\020\027\220x\230t\2401\330\014\020\220\005\220\\\240\024\240Q\330\020\027\220x\230t\2401\330\014\020\220\005\220\\\240\024\240]\260'\270\030\300\024\300Q\330\014\020\220\005\220\\\240\024\240Q\330\020\027\220x\230t\2401\330\014\020\220\005\220\\\240\030\250\027\260\010\270\001\330\014\020\220\005\220\\\240\026\240w\250h\260a\330\014\020\220\005\220\\\320!2\260'\270\030\300\021\330\014\020\220\005\220\\\240\037\260\007\260x\270q\330\014\020\220\005\220\\\240\035\250g\260X\270Q\330\014\020\220\005\220\\\240\034\250W\260H\270A\320\004:\270!\360\026\000\t\017\210n\230A\230T\240\025\240a\240r\250\032\2601\260A\330\010\013\2103\210c\220\021\330\014\020\220\001\330\020\026\220a\220q\230\006\230j\250\013\2607\270'\300\021\330\014\022\320\022\"\240!\320#:\270!\200\001\360\036\000\005\010\200t\2103\210a\330\010\t\330\004\007\200t\2101\210E\220\023""\220A\330\010\016\210l\230!\330\010\013\2105\220\004\220O\2403\240d\250!\330\010\013\2105\220\004\220O\2403\240d\250!\330\010\013\2105\220\004\320\024$\240C\240t\2501\330\010\013\2105\220\004\320\024$\240C\240t\2501\330\010\013\2105\220\004\320\024'\240s\250$\250a\330\t\r\210Q\210e\2203\220a\330\010\017\210~\230Q\330\010\014\210E\220\024\220_\240C\240t\2501\330\010\014\210E\220\024\220_\240C\240t\2501\330\010\014\210E\220\024\320\025%\240S\250\004\250A\330\010\014\210E\220\024\320\025%\240S\250\004\250A\330\010\014\210E\220\024\320\025(\250\003\2504\250q\340\010\r\210X\220Q\330\010\n\210%\210t\220?\240#\240T\250\021\330\010\n\210%\210t\220?\240#\240T\250\021\330\010\n\210%\210t\320\023#\2403\240d\250!\330\010\n\210%\210t\320\023#\2403\240d\250!\330\010\n\210%\210t\320\023&\240c\250\024\250Q";
    PyObject *data = NULL;
    CYTHON_UNUSED_VAR(__Pyx_DecompressString);
    #endif
    PyObject **stringtab = __pyx_mstate->__pyx_string_tab;
    Py_ssize_t pos = 0;
    for (int i = 0; i < 436; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyUnicode_DecodeUTF8(bytes + pos, bytes_length, NULL);
      if (likely(string) && i >= 115) PyUnicode_InternInPlace(&string);
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
      stringtab[i] = string;
      pos += bytes_length;
    }
    for (int i = 436; i < 477; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyBytes_FromStringAndSize(bytes + pos, bytes_length);
      stringtab[i] = string;
      pos += bytes_length;
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    Py_XDECREF(data);
    for (Py_ssize_t i = 0; i < 477; i++) {
      if (unlikely(PyObject_Hash(stringtab[i]) == -1)) {
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    #if CYTHON_IMMORTAL_CONSTANTS
    {
      PyObject **table = stringtab + 436;
      for (Py_ssize_t i=0; i<41; ++i) {
        #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
        #if PY_VERSION_HEX < 0x030E0000
        if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
        #else
        if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
        #endif
        {
          Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
        }
        #else
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
        #endif
      }
    }
    #endif
  }
  {
    PyObject **numbertab = __pyx_mstate->__pyx_number_tab + 0;
    int8_t const cint_constants_1[] = {0,1,2,4,8,10,16,32};
    for (int i = 0; i < 8; i++) {
      numbertab[i] = PyLong_FromLong(cint_constants_1[i - 0]);
      if (unlikely(!numbertab[i])) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_number_tab;
    for (Py_ssize_t i=0; i<8; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_codeobjects ### */
typedef struct {
    unsigned int argcount : 3;
    unsigned int num_posonly_args : 1;
    unsigned int num_kwonly_args : 1;
    unsigned int nlocals : 3;
    unsigned int flags : 10;
    unsigned int first_line : 11;
} __Pyx_PyCode_New_function_description;
/* NewCodeObj.proto */
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
);


static int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate) {
  PyObject* tuple_dedup_map = PyDict_New();
  if (unlikely(!tuple_dedup_map)) return -1;
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[0] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[0])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[1] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[1])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[2] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[2])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[3] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[3])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 464};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_mask, __pyx_mstate->__pyx_n_u_comp_masks, __pyx_mstate->__pyx_n_u_mask_str, __pyx_mstate->__pyx_n_u_f};
    __pyx_mstate_global->__pyx_codeobj_tab[4] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_mask_to_str, __pyx_mstate->__pyx_kp_b_iso88591_A_c_s_Q_4q_1_E_uBa_Jaq_A_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[4])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 475};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_flags, __pyx_mstate->__pyx_n_u_create_flags, __pyx_mstate->__pyx_n_u_create_str, __pyx_mstate->__pyx_n_u_f};
    __pyx_mstate_global->__pyx_codeobj_tab[5] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_flags_to_str, __pyx_mstate->__pyx_kp_b_iso88591_A_8_1_c_A_Q_E_vRq_l_1_a_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[5])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[6] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[6])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[7] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[7])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 649};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_ah_attr, __pyx_mstate->__pyx_n_u_is_alt, __pyx_mstate->__pyx_n_u_i};
    __pyx_mstate_global->__pyx_codeobj_tab[8] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_copy_ah, __pyx_mstate->__pyx_kp_b_iso88591_a_83a_4q_U_1_E_U_aq_84uD_XT_whd, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[8])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[9] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[9])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[10] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[10])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[11] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[11])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[12] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[12])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1021};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pd, __pyx_mstate->__pyx_n_u_attr};
    __pyx_mstate_global->__pyx_codeobj_tab[13] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_create_qp, __pyx_mstate->__pyx_kp_b_iso88591_A_G_E_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[13])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1024};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_ctx, __pyx_mstate->__pyx_n_u_attr};
    __pyx_mstate_global->__pyx_codeobj_tab[14] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_create_qp_ex, __pyx_mstate->__pyx_kp_b_iso88591_A_G_AS_4q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[14])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1038};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[15] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_close, __pyx_mstate->__pyx_kp_b_iso88591_A_4t3a_t1_G6_1D_d_a_a_q_a_q_a_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[15])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1052};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_dst, __pyx_mstate->__pyx_n_u_masks};
    __pyx_mstate_global->__pyx_codeobj_tab[16] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_get_comp_mask, __pyx_mstate->__pyx_kp_b_iso88591_A_C_A_7x_2_5Fa_0_A_8_336H_0_1_y, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[16])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1088};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_qp_attr, __pyx_mstate->__pyx_n_u_mask, __pyx_mstate->__pyx_n_u_rc};
    __pyx_mstate_global->__pyx_codeobj_tab[17] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_to_init, __pyx_mstate->__pyx_kp_b_iso88591_A_t_1_A_Qd_q_wa_3c_H, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[17])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1104};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_qp_attr, __pyx_mstate->__pyx_n_u_mask, __pyx_mstate->__pyx_n_u_rc};
    __pyx_mstate_global->__pyx_codeobj_tab[18] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_to_rtr, __pyx_mstate->__pyx_kp_b_iso88591_A_4z_Q_t_1_A_nAT_awgQ_3c_Gq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[18])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1124};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_qp_attr, __pyx_mstate->__pyx_n_u_mask, __pyx_mstate->__pyx_n_u_rc};
    __pyx_mstate_global->__pyx_codeobj_tab[19] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_to_rts, __pyx_mstate->__pyx_kp_b_iso88591_A_4z_Q_q_t_1_A_nAT_awgQ_3c_Gq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[19])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1143};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_attr_mask, __pyx_mstate->__pyx_n_u_attr, __pyx_mstate->__pyx_n_u_init_attr, __pyx_mstate->__pyx_n_u_rc};
    __pyx_mstate_global->__pyx_codeobj_tab[20] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_query, __pyx_mstate->__pyx_kp_b_iso88591_A_vQ_Ja_m1D_Qd_AYa_3c_9_vQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[20])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1159};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_qp_attr, __pyx_mstate->__pyx_n_u_comp_mask, __pyx_mstate->__pyx_n_u_rc};
    __pyx_mstate_global->__pyx_codeobj_tab[21] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_modify, __pyx_mstate->__pyx_kp_b_iso88591_A_nAT_awgQ_3c, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[21])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1172};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_wr, __pyx_mstate->__pyx_n_u_bad_wr, __pyx_mstate->__pyx_n_u_my_bad_wr, __pyx_mstate->__pyx_n_u_rc};
    __pyx_mstate_global->__pyx_codeobj_tab[22] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_post_recv, __pyx_mstate->__pyx_kp_b_iso88591_nAT_ar_1A_3c_aq_j_7, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[22])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1189};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_wr, __pyx_mstate->__pyx_n_u_bad_wr, __pyx_mstate->__pyx_n_u_my_bad_wr, __pyx_mstate->__pyx_n_u_rc};
    __pyx_mstate_global->__pyx_codeobj_tab[23] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_post_send, __pyx_mstate->__pyx_kp_b_iso88591_nAT_ar_1A_3c_aq_j_7, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[23])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1206};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_ece, __pyx_mstate->__pyx_n_u_rc};
    __pyx_mstate_global->__pyx_codeobj_tab[24] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_set_ece, __pyx_mstate->__pyx_kp_b_iso88591_A_3d_S_l_4uAS_3c_8, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[24])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1219};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_ece, __pyx_mstate->__pyx_n_u_rc};
    __pyx_mstate_global->__pyx_codeobj_tab[25] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_query_ece, __pyx_mstate->__pyx_kp_b_iso88591_A_c_nAT_as_3c_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[25])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1230};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_mw, __pyx_mstate->__pyx_n_u_mw_bind, __pyx_mstate->__pyx_n_u_rc};
    __pyx_mstate_global->__pyx_codeobj_tab[26] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_bind_mw, __pyx_mstate->__pyx_kp_b_iso88591_A_l_4uBe1G1_3c_8, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[26])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1241};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_op, __pyx_mstate->__pyx_n_u_flags};
    __pyx_mstate_global->__pyx_codeobj_tab[27] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_query_data_in_order, __pyx_mstate->__pyx_kp_b_iso88591_A_1D_T, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[27])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[28] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[28])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[29] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[29])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[30] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[30])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[31] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[31])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1333};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_rkey, __pyx_mstate->__pyx_n_u_remote_addr, __pyx_mstate->__pyx_n_u_compare, __pyx_mstate->__pyx_n_u_swap};
    __pyx_mstate_global->__pyx_codeobj_tab[32] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_wr_atomic_cmp_swp, __pyx_mstate->__pyx_kp_b_iso88591_A_q_HF_y, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[32])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1336};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_rkey, __pyx_mstate->__pyx_n_u_remote_addr, __pyx_mstate->__pyx_n_u_add};
    __pyx_mstate_global->__pyx_codeobj_tab[33] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_wr_atomic_fetch_add, __pyx_mstate->__pyx_kp_b_iso88591_A_hfM, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[33])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1339};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_mw, __pyx_mstate->__pyx_n_u_rkey, __pyx_mstate->__pyx_n_u_bind_info, __pyx_mstate->__pyx_n_u_info};
    __pyx_mstate_global->__pyx_codeobj_tab[34] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_wr_bind_mw, __pyx_mstate->__pyx_kp_b_iso88591_A_q_X_q_a_HAQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[34])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1346};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_invalidate_rkey};
    __pyx_mstate_global->__pyx_codeobj_tab[35] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_wr_local_inv, __pyx_mstate->__pyx_kp_b_iso88591_A_4xq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[35])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1349};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_rkey, __pyx_mstate->__pyx_n_u_remote_addr, __pyx_mstate->__pyx_n_u_atomic_wr, __pyx_mstate->__pyx_n_u_atomic_wr_c};
    __pyx_mstate_global->__pyx_codeobj_tab[36] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_wr_atomic_write, __pyx_mstate->__pyx_kp_b_iso88591_A_Qd_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[36])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1353};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_rkey, __pyx_mstate->__pyx_n_u_remote_addr};
    __pyx_mstate_global->__pyx_codeobj_tab[37] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_wr_rdma_read, __pyx_mstate->__pyx_kp_b_iso88591_A_4xvQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[37])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1356};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_rkey, __pyx_mstate->__pyx_n_u_remote_addr};
    __pyx_mstate_global->__pyx_codeobj_tab[38] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_wr_rdma_write, __pyx_mstate->__pyx_kp_b_iso88591_A_1D_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[38])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1359};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_rkey, __pyx_mstate->__pyx_n_u_remote_addr, __pyx_mstate->__pyx_n_u_data, __pyx_mstate->__pyx_n_u_imm_data};
    __pyx_mstate_global->__pyx_codeobj_tab[39] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_wr_rdma_write_imm, __pyx_mstate->__pyx_kp_b_iso88591_A_WAQ_q_HF_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[39])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1363};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_rkey, __pyx_mstate->__pyx_n_u_remote_addr, __pyx_mstate->__pyx_n_u_length, __pyx_mstate->__pyx_n_u_ptype, __pyx_mstate->__pyx_n_u_level};
    __pyx_mstate_global->__pyx_codeobj_tab[40] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_wr_flush, __pyx_mstate->__pyx_kp_b_iso88591_A_at86_hgQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[40])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1366};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[41] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_wr_send, __pyx_mstate->__pyx_kp_b_iso88591_A_Qd, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[41])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1369};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_data, __pyx_mstate->__pyx_n_u_imm_data};
    __pyx_mstate_global->__pyx_codeobj_tab[42] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_wr_send_imm, __pyx_mstate->__pyx_kp_b_iso88591_A_WAQ_XQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[42])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1373};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_invalidate_rkey};
    __pyx_mstate_global->__pyx_codeobj_tab[43] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_wr_send_inv, __pyx_mstate->__pyx_kp_b_iso88591_A_ha, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[43])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1376};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_hdr, __pyx_mstate->__pyx_n_u_hdr_sz, __pyx_mstate->__pyx_n_u_mss, __pyx_mstate->__pyx_n_u_ptr};
    __pyx_mstate_global->__pyx_codeobj_tab[44] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_wr_send_tso, __pyx_mstate->__pyx_kp_b_iso88591_A_aq_he81, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[44])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1380};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_ah, __pyx_mstate->__pyx_n_u_remote_qpn, __pyx_mstate->__pyx_n_u_remote_rkey};
    __pyx_mstate_global->__pyx_codeobj_tab[45] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_wr_set_ud_addr, __pyx_mstate->__pyx_kp_b_iso88591_A_AT_5_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[45])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1383};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_remote_srqn};
    __pyx_mstate_global->__pyx_codeobj_tab[46] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_wr_set_xrc_srqn, __pyx_mstate->__pyx_kp_b_iso88591_A_Qd_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[46])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1386};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_addr, __pyx_mstate->__pyx_n_u_length, __pyx_mstate->__pyx_n_u_ptr};
    __pyx_mstate_global->__pyx_codeobj_tab[47] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_wr_set_inline_data, __pyx_mstate->__pyx_kp_b_iso88591_A_aq_XU, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[47])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1390};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_num_buf, __pyx_mstate->__pyx_n_u_buf_list, __pyx_mstate->__pyx_n_u_data, __pyx_mstate->__pyx_n_u_i, __pyx_mstate->__pyx_n_u_data_buf};
    __pyx_mstate_global->__pyx_codeobj_tab[48] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_wr_set_inline_data_list, __pyx_mstate->__pyx_kp_b_iso88591_A_A_axr_5_1_aq_E_aq_81A_885_XU_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[48])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1402};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_sge};
    __pyx_mstate_global->__pyx_codeobj_tab[49] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_wr_set_sge, __pyx_mstate->__pyx_kp_b_iso88591_A_XS_s_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[49])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1405};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_num_sge, __pyx_mstate->__pyx_n_u_sg_list, __pyx_mstate->__pyx_n_u_sge, __pyx_mstate->__pyx_n_u_i};
    __pyx_mstate_global->__pyx_codeobj_tab[50] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_wr_set_sge_list, __pyx_mstate->__pyx_kp_b_iso88591_A_a_l_1_4s_aq_E_aq_q_A_q_G1Ba_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[50])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1417};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[51] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_wr_start, __pyx_mstate->__pyx_kp_b_iso88591_A_at1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[51])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1420};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_rc};
    __pyx_mstate_global->__pyx_codeobj_tab[52] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_wr_complete, __pyx_mstate->__pyx_kp_b_iso88591_A_at1_3c_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[52])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1425};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[53] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_wr_abort, __pyx_mstate->__pyx_kp_b_iso88591_A_at1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[53])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[54] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[54])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[55] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[55])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1429};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_dst, __pyx_mstate->__pyx_n_u_qia, __pyx_mstate->__pyx_n_u_qiae, __pyx_mstate->__pyx_n_u_qa};
    __pyx_mstate_global->__pyx_codeobj_tab[56] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_qp_pyx, __pyx_mstate->__pyx_n_u_copy_caps, __pyx_mstate->__pyx_kp_b_iso88591_t3a_t1E_A_l_5_O3d_5_O3d_5_Ct1_5, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[56])) goto bad;
  }
  Py_DECREF(tuple_dedup_map);
  return 0;
  bad:
  Py_DECREF(tuple_dedup_map);
  return -1;
}
/* #### Code section: init_globals ### */

static int __Pyx_InitGlobals(void) {
  /* PythonCompatibility.init */
  if (likely(__Pyx_init_co_variables() == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CommonTypesMetaclass.init */
  if (likely(__pyx_CommonTypesMetaclass_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CachedMethodType.init */
  #if CYTHON_COMPILING_IN_LIMITED_API
  {
      PyObject *typesModule=NULL;
      typesModule = PyImport_ImportModule("types");
      if (typesModule) {
          __pyx_mstate_global->__Pyx_CachedMethodType = PyObject_GetAttrString(typesModule, "MethodType");
          Py_DECREF(typesModule);
      }
  } // error handling follows
  #endif
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CythonFunctionShared.init */
  if (likely(__pyx_CyFunction_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError (used by GetBuiltinName) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_mstate_global->__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
            "name '%U' is not defined", name);
    }
    return result;
}

/* HasAttr (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!PyUnicode_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_PyObject_GetAttrStrNoError(o, n);
    if (!r) {
        return (unlikely(PyErr_Occurred())) ? -1 : 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}
#endif

/* PyObjectCall (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall (used by PyObjectCallOneArg) */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) != (0)) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
#if CYTHON_VECTORCALL && !CYTHON_COMPILING_IN_LIMITED_API
  #if PY_VERSION_HEX < 0x03090000
    #define __Pyx_PyVectorcall_Function(callable) _PyVectorcall_Function(callable)
  #elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE vectorcallfunc __Pyx_PyVectorcall_Function(PyObject *callable) {
    PyTypeObject *tp = Py_TYPE(callable);
    #if defined(__Pyx_CyFunction_USED)
    if (__Pyx_CyFunction_CheckExact(callable)) {
        return __Pyx_CyFunction_func_vectorcall(callable);
    }
    #endif
    if (!PyType_HasFeature(tp, Py_TPFLAGS_HAVE_VECTORCALL)) {
        return NULL;
    }
    assert(PyCallable_Check(callable));
    Py_ssize_t offset = tp->tp_vectorcall_offset;
    assert(offset > 0);
    vectorcallfunc ptr;
    memcpy(&ptr, (char *) callable + offset, sizeof(ptr));
    return ptr;
}
  #else
    #define __Pyx_PyVectorcall_Function(callable) PyVectorcall_Function(callable)
  #endif
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject *const *args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
          #if CYTHON_COMPILING_IN_LIMITED_API
            return PyObject_Vectorcall(func, args, _nargs, NULL);
          #else
            vectorcallfunc f = __Pyx_PyVectorcall_Function(func);
            if (f) {
                return f(func, args, _nargs, NULL);
            }
          #endif
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_mstate_global->__pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* PyObjectCallOneArg (used by ImportImpl) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* TupleAndListFromArray (used by ImportImpl) */
#if !CYTHON_COMPILING_IN_CPYTHON && CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    Py_ssize_t i;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    for (i = 0; i < n; i++) {
        if (unlikely(__Pyx_PyTuple_SET_ITEM(res, i, src[i]) < (0))) {
            Py_DECREF(res);
            return NULL;
        }
        Py_INCREF(src[i]);
    }
    return res;
}
#elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* ImportImpl (used by Import) */
static int __Pyx__Import_GetModule(PyObject *qualname, PyObject **module) {
    PyObject *imported_module = PyImport_GetModule(qualname);
    if (unlikely(!imported_module)) {
        *module = NULL;
        if (PyErr_Occurred()) {
            return -1;
        }
        return 0;
    }
    *module = imported_module;
    return 1;
}
static int __Pyx__Import_Lookup(PyObject *qualname, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject **module) {
    PyObject *imported_module;
    PyObject *top_level_package_name;
    Py_ssize_t i;
    int status, module_found;
    Py_ssize_t dot_index;
    module_found = __Pyx__Import_GetModule(qualname, &imported_module);
    if (unlikely(!module_found || module_found == -1)) {
        *module = NULL;
        return module_found;
    }
    if (imported_names) {
        for (i = 0; i < len_imported_names; i++) {
            PyObject *imported_name = imported_names[i];
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
            int has_imported_attribute = PyObject_HasAttr(imported_module, imported_name);
#else
            int has_imported_attribute = PyObject_HasAttrWithError(imported_module, imported_name);
            if (unlikely(has_imported_attribute == -1)) goto error;
#endif
            if (!has_imported_attribute) {
                goto not_found;
            }
        }
        *module = imported_module;
        return 1;
    }
    dot_index = PyUnicode_FindChar(qualname, '.', 0, PY_SSIZE_T_MAX, 1);
    if (dot_index == -1) {
        *module = imported_module;
        return 1;
    }
    if (unlikely(dot_index == -2)) goto error;
    top_level_package_name = PyUnicode_Substring(qualname, 0, dot_index);
    if (unlikely(!top_level_package_name)) goto error;
    Py_DECREF(imported_module);
    status = __Pyx__Import_GetModule(top_level_package_name, module);
    Py_DECREF(top_level_package_name);
    return status;
error:
    Py_DECREF(imported_module);
    *module = NULL;
    return -1;
not_found:
    Py_DECREF(imported_module);
    *module = NULL;
    return 0;
}
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *from_list = 0;
    int module_found;
    if (!qualname) {
        qualname = name;
    }
    module_found = __Pyx__Import_Lookup(qualname, imported_names, len_imported_names, &module);
    if (likely(module_found == 1)) {
        return module;
    } else if (unlikely(module_found == -1)) {
        return NULL;
    }
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    if (imported_names) {
#if CYTHON_COMPILING_IN_CPYTHON
        from_list = __Pyx_PyList_FromArray(imported_names, len_imported_names);
        if (unlikely(!from_list))
            goto bad;
#else
        from_list = PyList_New(len_imported_names);
        if (unlikely(!from_list)) goto bad;
        for (Py_ssize_t i=0; i<len_imported_names; ++i) {
            if (PyList_SetItem(from_list, i, __Pyx_NewRef(imported_names[i])) < 0) goto bad;
        }
#endif
    }
    if (level == -1) {
        const char* package_sep = strchr(__Pyx_MODULE_NAME, '.');
        if (package_sep != (0)) {
            module = PyImport_ImportModuleLevelObject(
                name, moddict, empty_dict, from_list, 1);
            if (unlikely(!module)) {
                if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                    goto bad;
                PyErr_Clear();
            }
        }
        level = 0;
    }
    if (!module) {
        module = PyImport_ImportModuleLevelObject(
            name, moddict, empty_dict, from_list, level);
    }
bad:
    Py_XDECREF(from_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level) {
    return __Pyx__Import(name, imported_names, len_imported_names, qualname, __pyx_mstate_global->__pyx_d, level);
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_mstate_global->__pyx_kp_u_);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400) ||\
                CYTHON_COMPILING_IN_GRAAL
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError, "cannot import name %S", name);
    }
    return value;
}

/* GetTopmostException (used by SaveResetException) */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C0000
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#elif __PYX_LIMITED_VERSION_HEX > 0x030C0000
    local_value = PyErr_GetRaisedException();
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
#if __PYX_LIMITED_VERSION_HEX > 0x030C0000
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
#else
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
#endif // __PYX_LIMITED_VERSION_HEX > 0x030C0000
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#elif __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    PyErr_SetHandledException(local_value);
    Py_XDECREF(local_value);
    Py_XDECREF(local_type);
    Py_XDECREF(local_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
#if __PYX_LIMITED_VERSION_HEX <= 0x030C0000
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
#endif
}

/* PyObjectFastCallMethod */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf) {
    PyObject *result;
    PyObject *attr = PyObject_GetAttr(args[0], name);
    if (unlikely(!attr))
        return NULL;
    result = __Pyx_PyObject_FastCall(attr, args+1, nargsf - 1);
    Py_DECREF(attr);
    return result;
}
#endif

/* BytesEquals (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL ||\
        !(CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS)
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length, length2;
        int kind;
        void *data1, *data2;
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        #endif
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length < 0)) return -1;
        #endif
        length2 = __Pyx_PyUnicode_GET_LENGTH(s2);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length2 < 0)) return -1;
        #endif
        if (length != length2) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    return (equals == Py_EQ);
return_ne:
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = __Pyx_PyTuple_GET_SIZE(kwnames);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(n == -1)) return NULL;
    #endif
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        if (s == namei) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        int eq = __Pyx_PyUnicode_Equals(s, namei, Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs;
    PyObject *dict;
#if !CYTHON_ASSUME_SAFE_SIZE
    nkwargs = PyTuple_Size(kwnames);
    if (unlikely(nkwargs < 0)) return NULL;
#else
    nkwargs = PyTuple_GET_SIZE(kwnames);
#endif
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
#if !CYTHON_ASSUME_SAFE_MACROS
        PyObject *key = PyTuple_GetItem(kwnames, i);
        if (!key) goto bad;
#else
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
#endif
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* UnpackUnboundCMethod (used by CallUnboundCMethod0) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
#elif CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject **args, Py_ssize_t nargs, PyObject *kwnames) {
        return _PyObject_Vectorcall
            (method, args ? args+1 : NULL, nargs ? nargs-1 : 0, kwnames);
}
#else
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *const *args, Py_ssize_t nargs, PyObject *kwnames) {
    return
#if PY_VERSION_HEX < 0x03090000
    _PyObject_Vectorcall
#else
    PyObject_Vectorcall
#endif
        (method, args ? args+1 : NULL, nargs ? (size_t) nargs-1 : 0, kwnames);
}
#endif
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
     METH_VARARGS | METH_KEYWORDS,
#else
     METH_FASTCALL | METH_KEYWORDS,
#endif
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method, *result=NULL;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    result = method;
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            result = unbound_method;
        }
    }
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    if (unlikely(target->method)) {
        Py_DECREF(result);
    } else
#endif
    target->method = result;
    return 0;
}

/* CallUnboundCMethod0 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        if (likely(cfunc->flag == METH_NOARGS))
            return __Pyx_CallCFunction(cfunc, self, NULL);
        if (likely(cfunc->flag == METH_FASTCALL))
            return __Pyx_CallCFunctionFast(cfunc, self, NULL, 0);
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, NULL, 0, NULL);
        if (likely(cfunc->flag == (METH_VARARGS | METH_KEYWORDS)))
            return __Pyx_CallCFunctionWithKeywords(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple, NULL);
        if (cfunc->flag == METH_VARARGS)
            return __Pyx_CallCFunction(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple);
        return __Pyx__CallUnboundCMethod0(cfunc, self);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod0(&tmp_cfunc, self);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod0(cfunc, self);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *result;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
    result = __Pyx_PyObject_CallOneArg(cfunc->method, self);
    return result;
}

/* py_dict_items (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_items, d);
}

/* py_dict_values (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_values, d);
}

/* OwnedDictNext (used by ParseKeywordsImpl) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue) {
    PyObject *next = NULL;
    if (!*ppos) {
        if (pvalue) {
            PyObject *dictview = pkey ? __Pyx_PyDict_Items(p) : __Pyx_PyDict_Values(p);
            if (unlikely(!dictview)) goto bad;
            *ppos = PyObject_GetIter(dictview);
            Py_DECREF(dictview);
        } else {
            *ppos = PyObject_GetIter(p);
        }
        if (unlikely(!*ppos)) goto bad;
    }
    next = PyIter_Next(*ppos);
    if (!next) {
        if (PyErr_Occurred()) goto bad;
        return 0;
    }
    if (pkey && pvalue) {
        *pkey = __Pyx_PySequence_ITEM(next, 0);
        if (unlikely(*pkey)) goto bad;
        *pvalue = __Pyx_PySequence_ITEM(next, 1);
        if (unlikely(*pvalue)) goto bad;
        Py_DECREF(next);
    } else if (pkey) {
        *pkey = next;
    } else {
        assert(pvalue);
        *pvalue = next;
    }
    return 1;
  bad:
    Py_XDECREF(next);
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
    PyErr_FormatUnraisable("Exception ignored in __Pyx_PyDict_NextRef");
#else
    PyErr_WriteUnraisable(__pyx_mstate_global->__pyx_n_u_Pyx_PyDict_NextRef);
#endif
    if (pkey) *pkey = NULL;
    if (pvalue) *pvalue = NULL;
    return 0;
}
#else // !CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue) {
    int result = PyDict_Next(p, ppos, pkey, pvalue);
    if (likely(result == 1)) {
        if (pkey) Py_INCREF(*pkey);
        if (pvalue) Py_INCREF(*pvalue);
    }
    return result;
}
#endif

/* RaiseDoubleKeywords (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            return __Pyx_CallCFunctionFast(cfunc, self, args, 2);
        }
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, 2, NULL);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod2(&tmp_cfunc, self, arg1, arg2);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        PyObject *result = NULL;
        PyObject *args = PyTuple_New(2);
        if (unlikely(!args)) return NULL;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = __Pyx_CallCFunctionWithKeywords(cfunc, self, args, NULL);
        else
            result = __Pyx_CallCFunction(cfunc, self, args);
        Py_DECREF(args);
        return result;
    }
#endif
    {
        PyObject *args[4] = {NULL, self, arg1, arg2};
        return __Pyx_PyObject_FastCall(cfunc->method, args+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
    }
}

/* ParseKeywordsImpl (used by ParseKeywords) */
static int __Pyx_ValidateDuplicatePosArgs(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char* function_name)
{
    PyObject ** const *name = argnames;
    while (name != first_kw_arg) {
        PyObject *key = **name;
        int found = PyDict_Contains(kwds, key);
        if (unlikely(found)) {
            if (found == 1) __Pyx_RaiseDoubleKeywordsError(function_name, key);
            goto bad;
        }
        name++;
    }
    return 0;
bad:
    return -1;
}
#if CYTHON_USE_UNICODE_INTERNALS
static CYTHON_INLINE int __Pyx_UnicodeKeywordsEqual(PyObject *s1, PyObject *s2) {
    int kind;
    Py_ssize_t len = PyUnicode_GET_LENGTH(s1);
    if (len != PyUnicode_GET_LENGTH(s2)) return 0;
    kind = PyUnicode_KIND(s1);
    if (kind != PyUnicode_KIND(s2)) return 0;
    const void *data1 = PyUnicode_DATA(s1);
    const void *data2 = PyUnicode_DATA(s2);
    return (memcmp(data1, data2, (size_t) len * (size_t) kind) == 0);
}
#endif
static int __Pyx_MatchKeywordArg_str(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    #if CYTHON_USE_UNICODE_INTERNALS
    Py_hash_t key_hash = ((PyASCIIObject*)key)->hash;
    if (unlikely(key_hash == -1)) {
        key_hash = PyObject_Hash(key);
        if (unlikely(key_hash == -1))
            goto bad;
    }
    #endif
    name = first_kw_arg;
    while (*name) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (key_hash == ((PyASCIIObject*)name_str)->hash && __Pyx_UnicodeKeywordsEqual(name_str, key)) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) {
                *index_found = (size_t) (name - argnames);
                return 1;
            }
        }
        #endif
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (unlikely(key_hash == ((PyASCIIObject*)name_str)->hash)) {
            if (__Pyx_UnicodeKeywordsEqual(name_str, key))
                goto arg_passed_twice;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            if (unlikely(name_str == key)) goto arg_passed_twice;
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) goto arg_passed_twice;
        }
        #endif
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
bad:
    return -1;
}
static int __Pyx_MatchKeywordArg_nostr(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    if (unlikely(!PyUnicode_Check(key))) goto invalid_keyword_type;
    name = first_kw_arg;
    while (*name) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (cmp == 1) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        if (unlikely(cmp == -1)) goto bad;
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (unlikely(cmp != 0)) {
            if (cmp == 1) goto arg_passed_twice;
            else goto bad;
        }
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
bad:
    return -1;
}
static CYTHON_INLINE int __Pyx_MatchKeywordArg(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    return likely(PyUnicode_CheckExact(key)) ?
        __Pyx_MatchKeywordArg_str(key, argnames, first_kw_arg, index_found, function_name) :
        __Pyx_MatchKeywordArg_nostr(key, argnames, first_kw_arg, index_found, function_name);
}
static void __Pyx_RejectUnknownKeyword(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char *function_name)
{
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos = NULL;
    #else
    Py_ssize_t pos = 0;
    #endif
    PyObject *key = NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(kwds);
    while (
        #if CYTHON_AVOID_BORROWED_REFS
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL)
        #else
        PyDict_Next(kwds, &pos, &key, NULL)
        #endif
    ) {
        PyObject** const *name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (!*name) {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp != 1) {
                if (cmp == 0) {
                    PyErr_Format(PyExc_TypeError,
                        "%s() got an unexpected keyword argument '%U'",
                        function_name, key);
                }
                #if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(key);
                #endif
                break;
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        #endif
    }
    __Pyx_END_CRITICAL_SECTION();
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(pos);
    #endif
    assert(PyErr_Occurred());
}
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t extracted = 0;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    name = first_kw_arg;
    while (*name && num_kwargs > extracted) {
        PyObject * key = **name;
        PyObject *value;
        int found = 0;
        #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        found = PyDict_GetItemRef(kwds, key, &value);
        #else
        value = PyDict_GetItemWithError(kwds, key);
        if (value) {
            Py_INCREF(value);
            found = 1;
        } else {
            if (unlikely(PyErr_Occurred())) goto bad;
        }
        #endif
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            extracted++;
        }
        name++;
    }
    if (num_kwargs > extracted) {
        if (ignore_unknown_kwargs) {
            if (unlikely(__Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name) == -1))
                goto bad;
        } else {
            __Pyx_RejectUnknownKeyword(kwds, argnames, first_kw_arg, function_name);
            goto bad;
        }
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t len;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    if (PyDict_Update(kwds2, kwds) < 0) goto bad;
    name = first_kw_arg;
    while (*name) {
        PyObject *key = **name;
        PyObject *value;
#if !CYTHON_COMPILING_IN_LIMITED_API && (PY_VERSION_HEX >= 0x030d00A2 || defined(PyDict_Pop))
        int found = PyDict_Pop(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
        }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        int found = PyDict_GetItemRef(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            if (unlikely(PyDict_DelItem(kwds2, key) < 0)) goto bad;
        }
#else
    #if CYTHON_COMPILING_IN_CPYTHON
        value = _PyDict_Pop(kwds2, key, kwds2);
    #else
        value = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_pop, kwds2, key, kwds2);
    #endif
        if (value == kwds2) {
            Py_DECREF(value);
        } else {
            if (unlikely(!value)) goto bad;
            values[name-argnames] = value;
        }
#endif
        name++;
    }
    len = PyDict_Size(kwds2);
    if (len > 0) {
        return __Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name);
    } else if (unlikely(len == -1)) {
        goto bad;
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject *key = NULL;
    PyObject** const * name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    for (Py_ssize_t pos = 0; pos < num_kwargs; pos++) {
#if CYTHON_AVOID_BORROWED_REFS
        key = __Pyx_PySequence_ITEM(kwds, pos);
#else
        key = __Pyx_PyTuple_GET_ITEM(kwds, pos);
#endif
#if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!key)) goto bad;
#endif
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            PyObject *value = kwvalues[pos];
            values[name-argnames] = __Pyx_NewRef(value);
        } else {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp == 1) {
                PyObject *value = kwvalues[pos];
                values[index_found] = __Pyx_NewRef(value);
            } else {
                if (unlikely(cmp == -1)) goto bad;
                if (kwds2) {
                    PyObject *value = kwvalues[pos];
                    if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
                } else if (!ignore_unknown_kwargs) {
                    goto invalid_keyword;
                }
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        key = NULL;
        #endif
    }
    return 0;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    goto bad;
bad:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(key);
    #endif
    return -1;
}

/* ParseKeywords */
static int __Pyx_ParseKeywords(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds)))
        return __Pyx_ParseKeywordsTuple(kwds, kwvalues, argnames, kwds2, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
    else if (kwds2)
        return __Pyx_ParseKeywordDictToDict(kwds, argnames, kwds2, values, num_pos_args, function_name);
    else
        return __Pyx_ParseKeywordDict(kwds, argnames, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* UnicodeConcatInPlace */
# if CYTHON_COMPILING_IN_CPYTHON
static int
__Pyx_unicode_modifiable(PyObject *unicode, int unsafe_shared)
{
    if (!__Pyx_IS_UNIQUELY_REFERENCED(unicode, unsafe_shared))
        return 0;
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030F0000
    if (PyUnstable_Unicode_GET_CACHED_HASH(unicode) != -1)
        return 0;
#endif
    if (!PyUnicode_CheckExact(unicode))
        return 0;
    if (PyUnicode_CHECK_INTERNED(unicode))
        return 0;
    return 1;
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right, int unsafe_shared
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    ) {
    PyObject *left = *p_left;
    Py_ssize_t left_len, right_len, new_len;
    if (unlikely(__Pyx_PyUnicode_READY(left) == -1))
        return NULL;
    if (unlikely(__Pyx_PyUnicode_READY(right) == -1))
        return NULL;
    left_len = PyUnicode_GET_LENGTH(left);
    if (left_len == 0) {
        Py_INCREF(right);
        return right;
    }
    right_len = PyUnicode_GET_LENGTH(right);
    if (right_len == 0) {
        Py_INCREF(left);
        return left;
    }
    if (unlikely(left_len > PY_SSIZE_T_MAX - right_len)) {
        PyErr_SetString(PyExc_OverflowError,
                        "strings are too large to concat");
        return NULL;
    }
    new_len = left_len + right_len;
    if (__Pyx_unicode_modifiable(left, unsafe_shared)
            && PyUnicode_CheckExact(right)
            && PyUnicode_KIND(right) <= PyUnicode_KIND(left)
            && !(PyUnicode_IS_ASCII(left) && !PyUnicode_IS_ASCII(right))) {
        int ret;
        __Pyx_GIVEREF(*p_left);
        ret = PyUnicode_Resize(p_left, new_len);
        __Pyx_GOTREF(*p_left);
        if (unlikely(ret != 0))
            return NULL;
        #if PY_VERSION_HEX >= 0x030d0000
        if (unlikely(PyUnicode_CopyCharacters(*p_left, left_len, right, 0, right_len) < 0)) return NULL;
        #else
        _PyUnicode_FastCopyCharacters(*p_left, left_len, right, 0, right_len);
        #endif
        __Pyx_INCREF(*p_left);
        __Pyx_GIVEREF(*p_left);
        return *p_left;
    } else {
        return __Pyx_PyUnicode_Concat(left, right);
    }
  }
#endif

/* RejectKeywords */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds) {
    PyObject *key = NULL;
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds))) {
        key = __Pyx_PySequence_ITEM(kwds, 0);
    } else {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *pos = NULL;
#else
        Py_ssize_t pos = 0;
#endif
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
        if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return;
#endif
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL);
#if CYTHON_AVOID_BORROWED_REFS
        Py_XDECREF(pos);
#endif
    }
    if (likely(key)) {
        PyErr_Format(PyExc_TypeError,
            "%s() got an unexpected keyword argument '%U'",
            function_name, key);
        Py_DECREF(key);
    }
}

/* RaiseException */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
#elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}

/* ArgTypeTestFunc (used by ArgTypeTest) */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    PyObject *extra_info = __pyx_mstate_global->__pyx_empty_unicode;
    int from_annotation_subclass = 0;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (!exact) {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    } else if (exact == 2) {
        if (__Pyx_TypeCheck(obj, type)) {
            from_annotation_subclass = 1;
            extra_info = __pyx_mstate_global->__pyx_kp_u_Note_that_Cython_is_deliberately;
        }
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(type);
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")"
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        "%s%U"
#endif
        , name, type_name, obj_type_name
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        , (from_annotation_subclass ? ". " : ""), extra_info
#endif
        );
#if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    if (exact == 2 && from_annotation_subclass) {
        PyObject *res;
        PyObject *vargs[2];
        vargs[0] = PyErr_GetRaisedException();
        vargs[1] = extra_info;
        res = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_kp_u_add_note, vargs, 2, NULL);
        Py_XDECREF(res);
        PyErr_SetRaisedException(vargs[0]);
    }
#endif
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* PyDictVersioning (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        if (!PyErr_Occurred())
            PyErr_SetNone(PyExc_NameError);
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
    PyErr_Clear();
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    if (unlikely(__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, name, &result) == -1)) PyErr_Clear();
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return result;
    }
#else
    result = _PyDict_GetItem_KnownHash(__pyx_mstate_global->__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectVectorCallKwBuilder (used by PyObjectVectorCallMethodKwBuilder) */
#if CYTHON_VECTORCALL
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_PyObject_FastCallDict;
    if (__Pyx_PyTuple_SET_ITEM(builder, n, key) != (0)) return -1;
    Py_INCREF(key);
    args[n] = value;
    return 0;
}
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_VectorcallBuilder_AddArgStr;
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n);
}
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    PyObject *pyKey = PyUnicode_FromString(key);
    if (!pyKey) return -1;
    return __Pyx_VectorcallBuilder_AddArg(pyKey, value, builder, args, n);
}
#else // CYTHON_VECTORCALL
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, CYTHON_UNUSED PyObject **args, CYTHON_UNUSED int n) {
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return PyDict_SetItem(builder, key, value);
}
#endif

/* PyObjectVectorCallMethodKwBuilder */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_Object_VectorcallMethod_CallFromBuilder(PyObject *name, PyObject *const *args, size_t nargsf, PyObject *kwnames) {
    PyObject *result;
    PyObject *obj = PyObject_GetAttr(args[0], name);
    if (unlikely(!obj))
        return NULL;
    result = __Pyx_Object_Vectorcall_CallFromBuilder(obj, args+1, nargsf-1, kwnames);
    Py_DECREF(obj);
    return result;
}
#endif

/* IterFinish (used by dict_iter) */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* PyObjectCallNoArg (used by PyObjectCallMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod (used by PyObjectCallMethod0) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(tp);
    PyErr_Format(PyExc_AttributeError,
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}
#endif

/* PyObjectCallMethod0 (used by dict_iter) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
#if CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000))
    PyObject *args[1] = {obj};
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_CallNoArg;
    return PyObject_VectorcallMethod(method_name, args, 1 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
#endif
}

/* RaiseNeedMoreValuesToUnpack (used by UnpackTuple2) */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack (used by UnpackItemEndCheck) */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck (used by UnpackTuple2) */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* RaiseNoneIterError (used by UnpackTupleError) */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError (used by UnpackTuple2) */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else {
      Py_ssize_t size = __Pyx_PyTuple_GET_SIZE(t);
 #if !CYTHON_ASSUME_SAFE_SIZE
      if (unlikely(size < 0)) return;
 #endif
      if (size < index) {
        __Pyx_RaiseNeedMoreValuesError(size);
      } else {
        __Pyx_RaiseTooManyValuesError(index);
      }
    }
}

/* UnpackTuple2 (used by dict_iter) */
static CYTHON_INLINE int __Pyx_unpack_tuple2(
        PyObject* tuple, PyObject** value1, PyObject** value2, int is_tuple, int has_known_size, int decref_tuple) {
    if (likely(is_tuple || PyTuple_Check(tuple))) {
        Py_ssize_t size;
        if (has_known_size) {
            return __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple);
        }
        size = __Pyx_PyTuple_GET_SIZE(tuple);
        if (likely(size == 2)) {
            return __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple);
        }
        if (size >= 0) {
            __Pyx_UnpackTupleError(tuple, 2);
        }
        return -1;
    } else {
        return __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple);
    }
}
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
    value1 = __Pyx_PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = __Pyx_PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = __Pyx_PyObject_GetIterNextFunc(iter);
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
#if CYTHON_COMPILING_IN_PYPY
#include <string.h>
#endif
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#else
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
#if !CYTHON_AVOID_BORROWED_REFS
static CYTHON_INLINE int __Pyx_dict_iter_next_source_is_dict(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem) {
    PyObject *key, *value;
    if (unlikely(orig_length != PyDict_Size(iter_obj))) {
        PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
        return -1;
    }
    if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
        return 0;
    }
    if (pitem) {
        PyObject* tuple = PyTuple_New(2);
        if (unlikely(!tuple)) {
            return -1;
        }
        Py_INCREF(key);
        Py_INCREF(value);
        #if CYTHON_ASSUME_SAFE_MACROS
        PyTuple_SET_ITEM(tuple, 0, key);
        PyTuple_SET_ITEM(tuple, 1, value);
        #else
        if (unlikely(PyTuple_SetItem(tuple, 0, key) < 0)) {
            Py_DECREF(value);
            Py_DECREF(tuple);
            return -1;
        }
        if (unlikely(PyTuple_SetItem(tuple, 1, value) < 0)) {
            Py_DECREF(tuple);
            return -1;
        }
        #endif
        *pitem = tuple;
    } else {
        if (pkey) {
            Py_INCREF(key);
            *pkey = key;
        }
        if (pvalue) {
            Py_INCREF(value);
            *pvalue = value;
        }
    }
    return 1;
}
#endif
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_AVOID_BORROWED_REFS
    if (source_is_dict) {
        int result;
#if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_BEGIN_CRITICAL_SECTION(iter_obj);
#endif
        result = __Pyx_dict_iter_next_source_is_dict(iter_obj, orig_length, ppos, pkey, pvalue, pitem);
#if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_END_CRITICAL_SECTION();
#endif
        return result;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        Py_ssize_t tuple_size = __Pyx_PyTuple_GET_SIZE(iter_obj);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(tuple_size < 0)) return -1;
        #endif
        if (unlikely(pos >= tuple_size)) return 0;
        *ppos = pos + 1;
        #if CYTHON_ASSUME_SAFE_MACROS
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        #else
        next_item = PyTuple_GetItem(iter_obj, pos);
        if (unlikely(!next_item)) return -1;
        #endif
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        Py_ssize_t list_size = __Pyx_PyList_GET_SIZE(iter_obj);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(list_size < 0)) return -1;
        #endif
        if (unlikely(pos >= list_size)) return 0;
        *ppos = pos + 1;
        next_item = __Pyx_PyList_GetItemRef(iter_obj, pos);
        if (unlikely(!next_item)) return -1;
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* DictGetItem */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    if (unlikely(__Pyx_PyDict_GetItemRef(d, key, &value) == 0)) { // no value, no error
        if (unlikely(PyTuple_Check(key))) {
            PyObject* args = PyTuple_Pack(1, key);
            if (likely(args)) {
                PyErr_SetObject(PyExc_KeyError, args);
                Py_DECREF(args);
            }
        } else {
            PyErr_SetObject(PyExc_KeyError, key);
        }
    }
    return value;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
    CYTHON_UNUSED_VAR(clineno);
    CYTHON_UNUSED_VAR(lineno);
    CYTHON_UNUSED_VAR(filename);
    CYTHON_MAYBE_UNUSED_VAR(nogil);
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(0);
    }
    ctx = PyUnicode_FromString(name);
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
    if (nogil)
        PyGILState_Release(state);
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS)) {
        return __Pyx_PyList_GetItemRefFast(o, wrapped_i, unsafe_shared);
    } else
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyList_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyTuple_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)) {
            return __Pyx_PyList_GetItemRefFast(o, n, unsafe_shared);
        } else if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            return __Pyx_NewRef(PyList_GET_ITEM(o, n));
        }
    } else
    #if !CYTHON_AVOID_BORROWED_REFS
    if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            return __Pyx_NewRef(PyTuple_GET_ITEM(o, n));
        }
    } else
    #endif
#endif
#if CYTHON_USE_TYPE_SLOTS && !CYTHON_COMPILING_IN_PYPY
    {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (!is_list && mm && mm->mp_subscript) {
            PyObject *r, *key = PyLong_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (is_list || likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    (void)wraparound;
    (void)boundscheck;
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_mstate_global->__pyx_n_u_class_getitem);
        if (!meth) {
            PyErr_Clear();
        } else {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* CallNextTpDealloc */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc) {
    PyTypeObject* type = Py_TYPE(obj);
    destructor tp_dealloc = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_dealloc, destructor) != current_tp_dealloc)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_dealloc = __Pyx_PyType_GetSlot(type, tp_dealloc, destructor)) == current_tp_dealloc)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type)
        tp_dealloc(obj);
}

/* CallNextTpTraverse */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    traverseproc tp_traverse = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc) != current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_traverse = __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc)) == current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_traverse)
        return tp_traverse(obj, v, a);
    return 0;
}

/* CallTypeTraverse */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg) {
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x03090000
    if (__Pyx_get_runtime_version() < 0x03090000) return 0;
    #endif
    if (!always_call) {
        PyTypeObject *base = __Pyx_PyObject_GetSlot(o, tp_base, PyTypeObject*);
        unsigned long flags = PyType_GetFlags(base);
        if (flags & Py_TPFLAGS_HEAPTYPE) {
            return 0;
        }
    }
    Py_VISIT((PyObject*)Py_TYPE(o));
    return 0;
}
#endif

/* CallNextTpClear */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    inquiry tp_clear = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_clear, inquiry) != current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_clear = __Pyx_PyType_GetSlot(type, tp_clear, inquiry)) == current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_clear)
        tp_clear(obj);
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_3_2_4
#define __PYX_HAVE_RT_ImportType_3_2_4
static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size)
{
    PyObject *result = 0;
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API)
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !( defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API) )
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    if (size == 0) {
        return (PyTypeObject *)result;
    }
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_2_4 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_2_4 && (size_t)basicsize > size) {
        if (PyErr_WarnFormat(NULL, 0,
                "%.200s.%.200s size changed, may indicate binary incompatibility. "
                "Expected %zd from C header, got %zd from PyObject",
                module_name, class_name, size, basicsize) < 0) {
            goto bad;
        }
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* LimitedApiGetTypeDict (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static Py_ssize_t __Pyx_GetTypeDictOffset(void) {
    PyObject *tp_dictoffset_o;
    Py_ssize_t tp_dictoffset;
    tp_dictoffset_o = PyObject_GetAttrString((PyObject*)(&PyType_Type), "__dictoffset__");
    if (unlikely(!tp_dictoffset_o)) return -1;
    tp_dictoffset = PyLong_AsSsize_t(tp_dictoffset_o);
    Py_DECREF(tp_dictoffset_o);
    if (unlikely(tp_dictoffset == 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' doesn't have a dictoffset");
        return -1;
    } else if (unlikely(tp_dictoffset < 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' has an unexpected negative dictoffset. "
            "Please report this as Cython bug");
        return -1;
    }
    return tp_dictoffset;
}
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp) {
    static Py_ssize_t tp_dictoffset = 0;
    if (unlikely(tp_dictoffset == 0)) {
        tp_dictoffset = __Pyx_GetTypeDictOffset();
        if (unlikely(tp_dictoffset == -1 && PyErr_Occurred())) {
            tp_dictoffset = 0; // try again next time?
            return NULL;
        }
    }
    return *(PyObject**)((char*)tp + tp_dictoffset);
}
#endif

/* SetItemOnTypeDict (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_SetItem(tp_dict, k, v);
    if (likely(!result)) {
        PyType_Modified(tp);
        if (unlikely(PyObject_HasAttr(v, __pyx_mstate_global->__pyx_n_u_set_name))) {
            PyObject *setNameResult = PyObject_CallMethodObjArgs(v, __pyx_mstate_global->__pyx_n_u_set_name,  (PyObject *) tp, k, NULL);
            if (!setNameResult) return -1;
            Py_DECREF(setNameResult);
        }
    }
    return result;
}

/* FixUpExtensionType */
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if __PYX_LIMITED_VERSION_HEX > 0x030900B1
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#else
    const PyType_Slot *slot = spec->slots;
    int changed = 0;
#if !CYTHON_COMPILING_IN_LIMITED_API
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
#if !CYTHON_COMPILING_IN_CPYTHON
        const
#endif  // !CYTHON_COMPILING_IN_CPYTHON)
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_vectorcall_offset = memb->offset;
                    changed = 1;
                }
#endif  // CYTHON_METH_FASTCALL
#if !CYTHON_COMPILING_IN_PYPY
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                    Py_DECREF(descr);
                    if (unlikely(set_item_result < 0)) {
                        return -1;
                    }
                    changed = 1;
                }
#endif  // !CYTHON_COMPILING_IN_PYPY
            }
            memb++;
        }
    }
#endif  // !CYTHON_COMPILING_IN_LIMITED_API
#if !CYTHON_COMPILING_IN_PYPY
    slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_getset)
        slot++;
    if (slot && slot->slot == Py_tp_getset) {
        PyGetSetDef *getset = (PyGetSetDef*) slot->pfunc;
        while (getset && getset->name) {
            if (getset->name[0] == '_' && getset->name[1] == '_' && strcmp(getset->name, "__module__") == 0) {
                PyObject *descr = PyDescr_NewGetSet(type, getset);
                if (unlikely(!descr))
                    return -1;
                #if CYTHON_COMPILING_IN_LIMITED_API
                PyObject *pyname = PyUnicode_FromString(getset->name);
                if (unlikely(!pyname)) {
                    Py_DECREF(descr);
                    return -1;
                }
                int set_item_result = __Pyx_SetItemOnTypeDict(type, pyname, descr);
                Py_DECREF(pyname);
                #else
                CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
                int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                #endif
                Py_DECREF(descr);
                if (unlikely(set_item_result < 0)) {
                    return -1;
                }
                changed = 1;
            }
            ++getset;
        }
    }
#else
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#endif  // !CYTHON_COMPILING_IN_PYPY
    if (changed)
        PyType_Modified(type);
#endif  // PY_VERSION_HEX > 0x030900B1
    return 0;
}

/* ValidateBasesTuple (used by PyType_Ready) */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_SIZE
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (unlikely(n < 0)) return -1;
#endif
    for (i = 1; i < n; i++)
    {
        PyTypeObject *b;
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !CYTHON_USE_TYPE_SLOTS
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
CYTHON_UNUSED static int __Pyx_PyType_HasMultipleInheritance(PyTypeObject *t) {
    while (t) {
        PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
        if (bases) {
            return 1;
        }
        t = __Pyx_PyType_GetSlot(t, tp_base, PyTypeObject*);
    }
    return 0;
}
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !CYTHON_COMPILING_IN_CPYTHON || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    if (!__Pyx_PyType_HasMultipleInheritance(t)) {
        return PyType_Ready(t);
    }
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400) &&\
                !CYTHON_COMPILING_IN_GRAAL
        gc = PyImport_GetModule(__pyx_mstate_global->__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_mstate_global->__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* DelItemOnTypeDict (used by SetupReduce) */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_DelItem(tp_dict, k);
    if (likely(!result)) PyType_Modified(tp);
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_mstate_global->__pyx_n_u_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_mstate_global->__pyx_n_u_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython);
            if (likely(reduce_cython)) {
                ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_mstate_global->__pyx_n_u_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetFullyQualifiedName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* GetVTable */
static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* SetVTable */
static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* MergeVTables */
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i=0;
    Py_ssize_t size;
    void** base_vtables;
    __Pyx_TypeName tp_base_name = NULL;
    __Pyx_TypeName base_name = NULL;
    void* unknown = (void*)-1;
    PyObject* bases = __Pyx_PyType_GetSlot(type, tp_bases, PyObject*);
    int base_depth = 0;
    {
        PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        while (base) {
            base_depth += 1;
            base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
        }
    }
    base_vtables = (void**) PyMem_Malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
#if CYTHON_COMPILING_IN_LIMITED_API
    size = PyTuple_Size(bases);
    if (size < 0) goto other_failure;
#else
    size = PyTuple_GET_SIZE(bases);
#endif
    for (i = 1; i < size; i++) {
        PyObject *basei;
        void* base_vtable;
#if CYTHON_AVOID_BORROWED_REFS
        basei = PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#else
        basei = PyTuple_GET_ITEM(bases, i);
#endif
        base_vtable = __Pyx_GetVtable((PyTypeObject*)basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
            }
        }
    }
    PyErr_Clear();
    PyMem_Free(base_vtables);
    return 0;
bad:
    {
        PyTypeObject* basei = NULL;
        PyTypeObject* tp_base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        tp_base_name = __Pyx_PyType_GetFullyQualifiedName(tp_base);
#if CYTHON_AVOID_BORROWED_REFS
        basei = (PyTypeObject*)PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = (PyTypeObject*)PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#else
        basei = (PyTypeObject*)PyTuple_GET_ITEM(bases, i);
#endif
        base_name = __Pyx_PyType_GetFullyQualifiedName(basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
    }
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
really_bad: // bad has failed!
#endif
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
other_failure:
#endif
    PyMem_Free(base_vtables);
    return -1;
}

/* PxdImportShared (used by FunctionImport) */
#ifndef __PYX_HAVE_RT_ImportFromPxd_3_2_4
#define __PYX_HAVE_RT_ImportFromPxd_3_2_4
static int __Pyx_ImportFromPxd_3_2_4(PyObject *module, const char *name, void **p, const char *sig, const char *what) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, "__pyx_capi__");
    if (!d)
        goto bad;
#if (defined(Py_LIMITED_API) && Py_LIMITED_API >= 0x030d0000) || (!defined(Py_LIMITED_API) && PY_VERSION_HEX >= 0x030d0000)
    PyDict_GetItemStringRef(d, name, &cobj);
#else
    cobj = PyDict_GetItemString(d, name);
    Py_XINCREF(cobj);
#endif
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C %.8s %.200s",
                PyModule_GetName(module), what, name);
        goto bad;
    }
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C %.8s %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             what, PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    Py_DECREF(cobj);
    return 0;
bad:
    Py_XDECREF(d);
    Py_XDECREF(cobj);
    return -1;
}
#endif

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction_3_2_4
#define __PYX_HAVE_RT_ImportFunction_3_2_4
static int __Pyx_ImportFunction_3_2_4(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    int result = __Pyx_ImportFromPxd_3_2_4(module, funcname, &tmp.p, sig, "function");
    if (result == 0) {
        *f = tmp.fp;
    }
    return result;
}
#endif

/* dict_setdefault (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value) {
    PyObject* value;
#if __PYX_LIMITED_VERSION_HEX >= 0x030F0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4)
    PyDict_SetDefaultRef(d, key, default_value, &value);
#elif CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    PyObject *args[] = {d, key, default_value};
    value = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_n_u_setdefault, args, 3 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#elif CYTHON_COMPILING_IN_LIMITED_API
    value = PyObject_CallMethodObjArgs(d, __pyx_mstate_global->__pyx_n_u_setdefault, key, default_value, NULL);
#else
    value = PyDict_SetDefault(d, key, default_value);
    if (unlikely(!value)) return NULL;
    Py_INCREF(value);
#endif
    return value;
}

/* AddModuleRef (used by FetchSharedCythonModule) */
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  static PyObject *__Pyx_PyImport_AddModuleObjectRef(PyObject *name) {
      PyObject *module_dict = PyImport_GetModuleDict();
      PyObject *m;
      if (PyMapping_GetOptionalItem(module_dict, name, &m) < 0) {
          return NULL;
      }
      if (m != NULL && PyModule_Check(m)) {
          return m;
      }
      Py_XDECREF(m);
      m = PyModule_NewObject(name);
      if (m == NULL)
          return NULL;
      if (PyDict_CheckExact(module_dict)) {
          PyObject *new_m;
          (void)PyDict_SetDefaultRef(module_dict, name, m, &new_m);
          Py_DECREF(m);
          return new_m;
      } else {
           if (PyObject_SetItem(module_dict, name, m) != 0) {
                Py_DECREF(m);
                return NULL;
            }
            return m;
      }
  }
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *py_name = PyUnicode_FromString(name);
      if (!py_name) return NULL;
      PyObject *module = __Pyx_PyImport_AddModuleObjectRef(py_name);
      Py_DECREF(py_name);
      return module;
  }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif

/* FetchSharedCythonModule (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef(__PYX_ABI_MODULE_NAME);
}

/* FetchCommonType (used by CommonTypesMetaclass) */
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject* __Pyx_PyType_FromMetaclass(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *result = __Pyx_PyType_FromModuleAndSpec(module, spec, bases);
    if (result && metaclass) {
        PyObject *old_tp = (PyObject*)Py_TYPE(result);
    Py_INCREF((PyObject*)metaclass);
#if __PYX_LIMITED_VERSION_HEX >= 0x03090000
        Py_SET_TYPE(result, metaclass);
#else
        result->ob_type = metaclass;
#endif
        Py_DECREF(old_tp);
    }
    return result;
}
#else
#define __Pyx_PyType_FromMetaclass(me, mo, s, b) PyType_FromMetaclass(me, mo, s, b)
#endif
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t expected_basicsize) {
    Py_ssize_t basicsize;
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (expected_basicsize == 0) {
        return 0; // size is inherited, nothing useful to check
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
    if (unlikely(!py_basicsize)) return -1;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = NULL;
    if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) return -1;
#else
    basicsize = ((PyTypeObject*) cached_type)->tp_basicsize;
#endif
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module = NULL, *cached_type = NULL, *abi_module_dict, *new_cached_type, *py_object_name;
    int get_item_ref_result;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    py_object_name = PyUnicode_FromString(object_name);
    if (!py_object_name) return NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) goto done;
    abi_module_dict = PyModule_GetDict(abi_module);
    if (!abi_module_dict) goto done;
    get_item_ref_result = __Pyx_PyDict_GetItemRef(abi_module_dict, py_object_name, &cached_type);
    if (get_item_ref_result == 1) {
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else if (unlikely(get_item_ref_result == -1)) {
        goto bad;
    }
    cached_type = __Pyx_PyType_FromMetaclass(
        metaclass,
        CYTHON_USE_MODULE_STATE ? module : abi_module,
        spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    new_cached_type = __Pyx_PyDict_SetDefault(abi_module_dict, py_object_name, cached_type);
    if (unlikely(new_cached_type != cached_type)) {
        if (unlikely(!new_cached_type)) goto bad;
        Py_DECREF(cached_type);
        cached_type = new_cached_type;
        if (__Pyx_VerifyCachedType(
                cached_type,
                object_name,
                spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else {
        Py_DECREF(new_cached_type);
    }
done:
    Py_XDECREF(abi_module);
    Py_DECREF(py_object_name);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CommonTypesMetaclass (used by CythonFunctionShared) */
static PyObject* __pyx_CommonTypesMetaclass_get_module(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED void* context) {
    return PyUnicode_FromString(__PYX_ABI_MODULE_NAME);
}
#if __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject* __pyx_CommonTypesMetaclass_call(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *args, CYTHON_UNUSED PyObject *kwds) {
    PyErr_SetString(PyExc_TypeError, "Cannot instantiate Cython internal types");
    return NULL;
}
static int __pyx_CommonTypesMetaclass_setattr(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *attr, CYTHON_UNUSED PyObject *value) {
    PyErr_SetString(PyExc_TypeError, "Cython internal types are immutable");
    return -1;
}
#endif
static PyGetSetDef __pyx_CommonTypesMetaclass_getset[] = {
    {"__module__", __pyx_CommonTypesMetaclass_get_module, NULL, NULL, NULL},
    {0, 0, 0, 0, 0}
};
static PyType_Slot __pyx_CommonTypesMetaclass_slots[] = {
    {Py_tp_getset, (void *)__pyx_CommonTypesMetaclass_getset},
    #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {Py_tp_call, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_new, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_setattro, (void*)__pyx_CommonTypesMetaclass_setattr},
    #endif
    {0, 0}
};
static PyType_Spec __pyx_CommonTypesMetaclass_spec = {
    __PYX_TYPE_MODULE_PREFIX "_common_types_metatype",
    0,
    0,
    Py_TPFLAGS_IMMUTABLETYPE |
    Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT,
    __pyx_CommonTypesMetaclass_slots
};
static int __pyx_CommonTypesMetaclass_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    PyObject *bases = PyTuple_Pack(1, &PyType_Type);
    if (unlikely(!bases)) {
        return -1;
    }
    mstate->__pyx_CommonTypesMetaclassType = __Pyx_FetchCommonTypeFromSpec(NULL, module, &__pyx_CommonTypesMetaclass_spec, bases);
    Py_DECREF(bases);
    if (unlikely(mstate->__pyx_CommonTypesMetaclassType == NULL)) {
        return -1;
    }
    return 0;
}

/* PyMethodNew (used by CythonFunctionShared) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *result;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    #if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    {
        PyObject *args[] = {func, self};
        result = PyObject_Vectorcall(__pyx_mstate_global->__Pyx_CachedMethodType, args, 2, NULL);
    }
    #else
    result = PyObject_CallFunctionObjArgs(__pyx_mstate_global->__Pyx_CachedMethodType, func, self, NULL);
    #endif
    return result;
}
#else
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#endif

/* PyVectorcallFastCallDict (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i;
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos;
    #else
    Py_ssize_t pos;
    #endif
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    #if !CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t nkw = PyDict_Size(kw);
    if (unlikely(nkw == -1)) return NULL;
    #else
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    #endif
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = 0;
    i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (__Pyx_PyDict_NextRef(kw, &pos, &key, &value)) {
        keys_are_strings &=
        #if CYTHON_COMPILING_IN_LIMITED_API
            PyType_GetFlags(Py_TYPE(key));
        #else
            Py_TYPE(key)->tp_flags;
        #endif
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(PyTuple_SetItem(kwnames, i, key) < 0)) goto cleanup;
        #else
        PyTuple_SET_ITEM(kwnames, i, key);
        #endif
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(pos);
    #endif
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    Py_ssize_t kw_size =
        likely(kw == NULL) ?
        0 :
#if !CYTHON_ASSUME_SAFE_SIZE
        PyDict_Size(kw);
#else
        PyDict_GET_SIZE(kw);
#endif
    if (kw_size == 0) {
        return vc(func, args, nargs, NULL);
    }
#if !CYTHON_ASSUME_SAFE_SIZE
    else if (unlikely(kw_size == -1)) {
        return NULL;
    }
#endif
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared (used by CythonFunction) */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunctionNoMethod(PyObject *func, void (*cfunc)(void)) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if ((PyObject*)Py_TYPE(func) == __pyx_mstate_global->__Pyx_CachedMethodType) {
        int result;
        PyObject *newFunc = PyObject_GetAttr(func, __pyx_mstate_global->__pyx_n_u_func);
        if (unlikely(!newFunc)) {
            PyErr_Clear(); // It's only an optimization, so don't throw an error
            return 0;
        }
        result = __Pyx__IsSameCyOrCFunctionNoMethod(newFunc, cfunc);
        Py_DECREF(newFunc);
        return result;
    }
    return __Pyx__IsSameCyOrCFunctionNoMethod(func, cfunc);
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if (PyMethod_Check(func)) {
        func = PyMethod_GET_FUNCTION(func);
    }
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure) {
    PyObject *result;
    CYTHON_UNUSED_VAR(closure);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_doc_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#else
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_name_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    PyObject *result;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    Py_INCREF(op->func_qualname);
    result = op->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
#endif
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_defaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_kwdefaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_annotations_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine_value(__pyx_CyFunctionObject *op) {
    int is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
    if (is_coroutine) {
        PyObject *is_coroutine_value, *module, *fromlist, *marker = __pyx_mstate_global->__pyx_n_u_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_mstate_global->__pyx_n_u_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        is_coroutine_value = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(is_coroutine_value)) {
            return is_coroutine_value;
        }
ignore:
        PyErr_Clear();
    }
    return __Pyx_PyBool_FromLong(is_coroutine);
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    result = __Pyx_CyFunction_get_is_coroutine_value(op);
    if (unlikely(!result))
        return NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    if (op->func_is_coroutine) {
        Py_DECREF(result);
        result = __Pyx_NewRef(op->func_is_coroutine);
    } else {
        op->func_is_coroutine = __Pyx_NewRef(result);
    }
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static void __Pyx_CyFunction_raise_argument_count_error(__pyx_CyFunctionObject *func, const char* message, Py_ssize_t size) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        py_name, message, size);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        name, message, size);
#endif
}
static void __Pyx_CyFunction_raise_type_error(__pyx_CyFunctionObject *func, const char* message) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s",
        py_name, message);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s",
        name, message);
#endif
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {"func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {"func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
#else
    {"func_dict", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
#endif
    {"func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {"__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {"_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    {"__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#endif
#if CYTHON_METH_FASTCALL
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {"__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(args);
    __Pyx_BEGIN_CRITICAL_SECTION(m);
    Py_INCREF(m->func_qualname);
    result = m->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_dict = NULL;
#endif
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func_dict);
#elif PY_VERSION_HEX < 0x030d0000
    _PyObject_ClearManagedDict((PyObject*)m);
#else
    PyObject_ClearManagedDict((PyObject*)m);
#endif
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    Py_CLEAR(m->defaults);
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    {
        int e = __Pyx_call_type_traverse((PyObject*)m, 1, visit, arg);
        if (e) return e;
    }
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func_dict);
#else
    {
        int e =
#if PY_VERSION_HEX < 0x030d0000
            _PyObject_VisitManagedDict
#else
            PyObject_VisitManagedDict
#endif
                ((PyObject*)m, visit, arg);
        if (e != 0) return e;
    }
#endif
    __Pyx_VISIT_CONST(m->func_name);
    __Pyx_VISIT_CONST(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    __Pyx_VISIT_CONST(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    Py_VISIT(m->defaults);
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
    PyObject *repr;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    repr = PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
    __Pyx_END_CRITICAL_SECTION();
    return repr;
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void(*)(void))meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes no arguments", size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes exactly one argument", size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
    __Pyx_CyFunction_raise_type_error(
        (__pyx_CyFunctionObject*)func, "takes no keyword arguments");
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_SIZE
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(argc < 0)) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            __Pyx_CyFunction_raise_type_error(
                cyfunc, "needs an argument");
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(__Pyx_PyTuple_GET_SIZE(kwnames))) {
        __Pyx_CyFunction_raise_type_error(
            cyfunc, "takes no keyword arguments");
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes no arguments", nargs);
        return NULL;
    }
    return meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes exactly one argument", nargs);
        return NULL;
    }
    return meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    #if PY_VERSION_HEX < 0x030e00A6
    size_t nargs_value = (size_t) nargs;
    #else
    Py_ssize_t nargs_value = nargs;
    #endif
    return ((__Pyx_PyCMethod)(void(*)(void))meth)(self, cls, args, nargs_value, kwnames);
}
#endif
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if CYTHON_METH_FASTCALL
#if defined(Py_TPFLAGS_HAVE_VECTORCALL)
    Py_TPFLAGS_HAVE_VECTORCALL |
#elif defined(_Py_TPFLAGS_HAVE_VECTORCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
#endif // CYTHON_METH_FASTCALL
#if PY_VERSION_HEX >= 0x030C0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_TPFLAGS_MANAGED_DICT |
#endif
    Py_TPFLAGS_IMMUTABLETYPE | Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
static int __pyx_CyFunction_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    mstate->__pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(
        mstate->__pyx_CommonTypesMetaclassType, module, &__pyx_CyFunctionType_spec, NULL);
    if (unlikely(mstate->__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func, PyTypeObject *defaults_type) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_CallObject((PyObject*)defaults_type, NULL); // _PyObject_New(defaults_type);
    if (unlikely(!m->defaults))
        return NULL;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_mstate_global->__pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* CLineInTraceback (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
#define __Pyx_PyProbablyModule_GetDict(o) __Pyx_XNewRef(PyModule_GetDict(o))
#elif !CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyProbablyModule_GetDict(o) PyObject_GenericGetDict(o, NULL);
#else
PyObject* __Pyx_PyProbablyModule_GetDict(PyObject *o) {
    PyObject **dict_ptr = _PyObject_GetDictPtr(o);
    return dict_ptr ? __Pyx_XNewRef(*dict_ptr) : NULL;
}
#endif
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline = NULL;
    PyObject *ptype, *pvalue, *ptraceback;
    PyObject *cython_runtime_dict;
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_mstate_global->__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
    cython_runtime_dict = __Pyx_PyProbablyModule_GetDict(__pyx_mstate_global->__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, cython_runtime_dict,
            __Pyx_PyDict_SetDefault(cython_runtime_dict, __pyx_mstate_global->__pyx_n_u_cline_in_traceback, Py_False))
    }
    if (use_cline == NULL || use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    Py_XDECREF(use_cline);
    Py_XDECREF(cython_runtime_dict);
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache (used by AddTraceback) */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static __Pyx_CachedCodeObjectType *__pyx__find_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line) {
    __Pyx_CachedCodeObjectType* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!code_cache->entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if (unlikely(pos >= code_cache->count) || unlikely(code_cache->entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = code_cache->entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__find_code_object;
    return NULL; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just miss.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type old_count = __pyx_atomic_incr_acq_rel(&code_cache->accessor_count);
    if (old_count < 0) {
        __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
        return NULL;
    }
#endif
    __Pyx_CachedCodeObjectType *result = __pyx__find_code_object(code_cache, code_line);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
#endif
    return result;
#endif
}
static void __pyx__insert_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line, __Pyx_CachedCodeObjectType* code_object)
{
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = code_cache->entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            code_cache->entries = entries;
            code_cache->max_count = 64;
            code_cache->count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if ((pos < code_cache->count) && unlikely(code_cache->entries[pos].code_line == code_line)) {
        __Pyx_CachedCodeObjectType* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_INCREF(code_object);
        Py_DECREF(tmp);
        return;
    }
    if (code_cache->count == code_cache->max_count) {
        int new_max = code_cache->max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            code_cache->entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        code_cache->entries = entries;
        code_cache->max_count = new_max;
    }
    for (i=code_cache->count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    code_cache->count++;
    Py_INCREF(code_object);
}
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__insert_code_object;
    return; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just fail.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type expected = 0;
    if (!__pyx_atomic_int_cmp_exchange(&code_cache->accessor_count, &expected, INT_MIN)) {
        return;
    }
#endif
    __pyx__insert_code_object(code_cache, code_line, code_object);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_sub(&code_cache->accessor_count, INT_MIN);
#endif
#endif
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result = PyObject_Call(replace, __pyx_mstate_global->__pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!code_object) {
        code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
        if (unlikely(!code_object)) goto bad;
        py_py_line = PyLong_FromLong(py_line);
        if (unlikely(!py_py_line)) goto bad;
        if (c_line) {
            py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        } else {
            py_funcname = PyUnicode_FromString(funcname);
        }
        if (unlikely(!py_funcname)) goto bad;
        dict = PyDict_New();
        if (unlikely(!dict)) goto bad;
        {
            PyObject *old_code_object = code_object;
            code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
            Py_DECREF(old_code_object);
        }
        if (unlikely(!code_object)) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, code_object);
    } else {
        dict = PyDict_New();
    }
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    if (c_line) {
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
    }
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_mstate_global->__pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE unsigned int __Pyx_PyLong_As_unsigned_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        unsigned int val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (unsigned int) -1;
        val = __Pyx_PyLong_As_unsigned_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(unsigned int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) >= 2 * PyLong_SHIFT)) {
                            return (unsigned int) (((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) >= 3 * PyLong_SHIFT)) {
                            return (unsigned int) (((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) >= 4 * PyLong_SHIFT)) {
                            return (unsigned int) (((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (unsigned int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(unsigned int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(unsigned int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned int) (((unsigned int)-1)*(((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(unsigned int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned int) ((((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned int) (((unsigned int)-1)*(((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned int) ((((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned int) (((unsigned int)-1)*(((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned int) ((((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(unsigned int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, long, PyLong_AsLong(x))
        } else if ((sizeof(unsigned int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        unsigned int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (unsigned int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (unsigned int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (unsigned int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (unsigned int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(unsigned int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((unsigned int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(unsigned int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((unsigned int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((unsigned int) 1) << (sizeof(unsigned int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (unsigned int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned int");
    return (unsigned int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned int");
    return (unsigned int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_int(unsigned int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned int) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned int) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(unsigned int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE enum ibv_qp_type __Pyx_PyLong_As_enum__ibv_qp_type(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_qp_type neg_one = (enum ibv_qp_type) -1, const_zero = (enum ibv_qp_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        enum ibv_qp_type val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (enum ibv_qp_type) -1;
        val = __Pyx_PyLong_As_enum__ibv_qp_type(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(enum ibv_qp_type, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(enum ibv_qp_type) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_qp_type, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_qp_type) >= 2 * PyLong_SHIFT)) {
                            return (enum ibv_qp_type) (((((enum ibv_qp_type)digits[1]) << PyLong_SHIFT) | (enum ibv_qp_type)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(enum ibv_qp_type) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_qp_type, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_qp_type) >= 3 * PyLong_SHIFT)) {
                            return (enum ibv_qp_type) (((((((enum ibv_qp_type)digits[2]) << PyLong_SHIFT) | (enum ibv_qp_type)digits[1]) << PyLong_SHIFT) | (enum ibv_qp_type)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(enum ibv_qp_type) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_qp_type, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_qp_type) >= 4 * PyLong_SHIFT)) {
                            return (enum ibv_qp_type) (((((((((enum ibv_qp_type)digits[3]) << PyLong_SHIFT) | (enum ibv_qp_type)digits[2]) << PyLong_SHIFT) | (enum ibv_qp_type)digits[1]) << PyLong_SHIFT) | (enum ibv_qp_type)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (enum ibv_qp_type) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(enum ibv_qp_type) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_qp_type, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(enum ibv_qp_type) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_qp_type, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(enum ibv_qp_type, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(enum ibv_qp_type) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_qp_type, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_qp_type) - 1 > 2 * PyLong_SHIFT)) {
                            return (enum ibv_qp_type) (((enum ibv_qp_type)-1)*(((((enum ibv_qp_type)digits[1]) << PyLong_SHIFT) | (enum ibv_qp_type)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(enum ibv_qp_type) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_qp_type, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_qp_type) - 1 > 2 * PyLong_SHIFT)) {
                            return (enum ibv_qp_type) ((((((enum ibv_qp_type)digits[1]) << PyLong_SHIFT) | (enum ibv_qp_type)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(enum ibv_qp_type) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_qp_type, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_qp_type) - 1 > 3 * PyLong_SHIFT)) {
                            return (enum ibv_qp_type) (((enum ibv_qp_type)-1)*(((((((enum ibv_qp_type)digits[2]) << PyLong_SHIFT) | (enum ibv_qp_type)digits[1]) << PyLong_SHIFT) | (enum ibv_qp_type)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(enum ibv_qp_type) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_qp_type, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_qp_type) - 1 > 3 * PyLong_SHIFT)) {
                            return (enum ibv_qp_type) ((((((((enum ibv_qp_type)digits[2]) << PyLong_SHIFT) | (enum ibv_qp_type)digits[1]) << PyLong_SHIFT) | (enum ibv_qp_type)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(enum ibv_qp_type) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_qp_type, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_qp_type) - 1 > 4 * PyLong_SHIFT)) {
                            return (enum ibv_qp_type) (((enum ibv_qp_type)-1)*(((((((((enum ibv_qp_type)digits[3]) << PyLong_SHIFT) | (enum ibv_qp_type)digits[2]) << PyLong_SHIFT) | (enum ibv_qp_type)digits[1]) << PyLong_SHIFT) | (enum ibv_qp_type)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(enum ibv_qp_type) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_qp_type, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_qp_type) - 1 > 4 * PyLong_SHIFT)) {
                            return (enum ibv_qp_type) ((((((((((enum ibv_qp_type)digits[3]) << PyLong_SHIFT) | (enum ibv_qp_type)digits[2]) << PyLong_SHIFT) | (enum ibv_qp_type)digits[1]) << PyLong_SHIFT) | (enum ibv_qp_type)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(enum ibv_qp_type) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_qp_type, long, PyLong_AsLong(x))
        } else if ((sizeof(enum ibv_qp_type) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_qp_type, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        enum ibv_qp_type val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() or PyLong_AsNativeBytes() not available, cannot convert large enums");
        val = (enum ibv_qp_type) -1;
#endif
        if (unlikely(ret))
            return (enum ibv_qp_type) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum ibv_qp_type");
    return (enum ibv_qp_type) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum ibv_qp_type");
    return (enum ibv_qp_type) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyLong_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE unsigned short __Pyx_PyLong_As_unsigned_short(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned short neg_one = (unsigned short) -1, const_zero = (unsigned short) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        unsigned short val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (unsigned short) -1;
        val = __Pyx_PyLong_As_unsigned_short(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned short, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(unsigned short) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned short, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned short) >= 2 * PyLong_SHIFT)) {
                            return (unsigned short) (((((unsigned short)digits[1]) << PyLong_SHIFT) | (unsigned short)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned short) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned short, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned short) >= 3 * PyLong_SHIFT)) {
                            return (unsigned short) (((((((unsigned short)digits[2]) << PyLong_SHIFT) | (unsigned short)digits[1]) << PyLong_SHIFT) | (unsigned short)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned short) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned short, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned short) >= 4 * PyLong_SHIFT)) {
                            return (unsigned short) (((((((((unsigned short)digits[3]) << PyLong_SHIFT) | (unsigned short)digits[2]) << PyLong_SHIFT) | (unsigned short)digits[1]) << PyLong_SHIFT) | (unsigned short)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (unsigned short) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(unsigned short) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned short, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(unsigned short) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned short, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned short, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(unsigned short) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned short, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned short) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned short) (((unsigned short)-1)*(((((unsigned short)digits[1]) << PyLong_SHIFT) | (unsigned short)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(unsigned short) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned short, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned short) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned short) ((((((unsigned short)digits[1]) << PyLong_SHIFT) | (unsigned short)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(unsigned short) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned short, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned short) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned short) (((unsigned short)-1)*(((((((unsigned short)digits[2]) << PyLong_SHIFT) | (unsigned short)digits[1]) << PyLong_SHIFT) | (unsigned short)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned short) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned short, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned short) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned short) ((((((((unsigned short)digits[2]) << PyLong_SHIFT) | (unsigned short)digits[1]) << PyLong_SHIFT) | (unsigned short)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(unsigned short) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned short, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned short) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned short) (((unsigned short)-1)*(((((((((unsigned short)digits[3]) << PyLong_SHIFT) | (unsigned short)digits[2]) << PyLong_SHIFT) | (unsigned short)digits[1]) << PyLong_SHIFT) | (unsigned short)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned short) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned short, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned short) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned short) ((((((((((unsigned short)digits[3]) << PyLong_SHIFT) | (unsigned short)digits[2]) << PyLong_SHIFT) | (unsigned short)digits[1]) << PyLong_SHIFT) | (unsigned short)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(unsigned short) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned short, long, PyLong_AsLong(x))
        } else if ((sizeof(unsigned short) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned short, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        unsigned short val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (unsigned short) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (unsigned short) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (unsigned short) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (unsigned short) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(unsigned short) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((unsigned short) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(unsigned short) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((unsigned short) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((unsigned short) 1) << (sizeof(unsigned short) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (unsigned short) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned short");
    return (unsigned short) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned short");
    return (unsigned short) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE unsigned long __Pyx_PyLong_As_unsigned_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned long neg_one = (unsigned long) -1, const_zero = (unsigned long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        unsigned long val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (unsigned long) -1;
        val = __Pyx_PyLong_As_unsigned_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(unsigned long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) >= 2 * PyLong_SHIFT)) {
                            return (unsigned long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) >= 3 * PyLong_SHIFT)) {
                            return (unsigned long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) >= 4 * PyLong_SHIFT)) {
                            return (unsigned long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (unsigned long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(unsigned long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned long, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(unsigned long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(unsigned long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned long) (((unsigned long)-1)*(((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(unsigned long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned long) ((((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned long) (((unsigned long)-1)*(((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned long) ((((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned long) (((unsigned long)-1)*(((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned long) ((((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(unsigned long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned long, long, PyLong_AsLong(x))
        } else if ((sizeof(unsigned long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned long, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        unsigned long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (unsigned long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (unsigned long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (unsigned long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (unsigned long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(unsigned long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((unsigned long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(unsigned long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((unsigned long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((unsigned long) 1) << (sizeof(unsigned long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (unsigned long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned long");
    return (unsigned long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned long");
    return (unsigned long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE enum ibv_qp_state __Pyx_PyLong_As_enum__ibv_qp_state(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_qp_state neg_one = (enum ibv_qp_state) -1, const_zero = (enum ibv_qp_state) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        enum ibv_qp_state val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (enum ibv_qp_state) -1;
        val = __Pyx_PyLong_As_enum__ibv_qp_state(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(enum ibv_qp_state, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(enum ibv_qp_state) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_qp_state, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_qp_state) >= 2 * PyLong_SHIFT)) {
                            return (enum ibv_qp_state) (((((enum ibv_qp_state)digits[1]) << PyLong_SHIFT) | (enum ibv_qp_state)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(enum ibv_qp_state) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_qp_state, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_qp_state) >= 3 * PyLong_SHIFT)) {
                            return (enum ibv_qp_state) (((((((enum ibv_qp_state)digits[2]) << PyLong_SHIFT) | (enum ibv_qp_state)digits[1]) << PyLong_SHIFT) | (enum ibv_qp_state)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(enum ibv_qp_state) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_qp_state, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_qp_state) >= 4 * PyLong_SHIFT)) {
                            return (enum ibv_qp_state) (((((((((enum ibv_qp_state)digits[3]) << PyLong_SHIFT) | (enum ibv_qp_state)digits[2]) << PyLong_SHIFT) | (enum ibv_qp_state)digits[1]) << PyLong_SHIFT) | (enum ibv_qp_state)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (enum ibv_qp_state) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(enum ibv_qp_state) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_qp_state, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(enum ibv_qp_state) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_qp_state, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(enum ibv_qp_state, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(enum ibv_qp_state) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_qp_state, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_qp_state) - 1 > 2 * PyLong_SHIFT)) {
                            return (enum ibv_qp_state) (((enum ibv_qp_state)-1)*(((((enum ibv_qp_state)digits[1]) << PyLong_SHIFT) | (enum ibv_qp_state)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(enum ibv_qp_state) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_qp_state, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_qp_state) - 1 > 2 * PyLong_SHIFT)) {
                            return (enum ibv_qp_state) ((((((enum ibv_qp_state)digits[1]) << PyLong_SHIFT) | (enum ibv_qp_state)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(enum ibv_qp_state) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_qp_state, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_qp_state) - 1 > 3 * PyLong_SHIFT)) {
                            return (enum ibv_qp_state) (((enum ibv_qp_state)-1)*(((((((enum ibv_qp_state)digits[2]) << PyLong_SHIFT) | (enum ibv_qp_state)digits[1]) << PyLong_SHIFT) | (enum ibv_qp_state)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(enum ibv_qp_state) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_qp_state, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_qp_state) - 1 > 3 * PyLong_SHIFT)) {
                            return (enum ibv_qp_state) ((((((((enum ibv_qp_state)digits[2]) << PyLong_SHIFT) | (enum ibv_qp_state)digits[1]) << PyLong_SHIFT) | (enum ibv_qp_state)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(enum ibv_qp_state) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_qp_state, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_qp_state) - 1 > 4 * PyLong_SHIFT)) {
                            return (enum ibv_qp_state) (((enum ibv_qp_state)-1)*(((((((((enum ibv_qp_state)digits[3]) << PyLong_SHIFT) | (enum ibv_qp_state)digits[2]) << PyLong_SHIFT) | (enum ibv_qp_state)digits[1]) << PyLong_SHIFT) | (enum ibv_qp_state)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(enum ibv_qp_state) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_qp_state, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_qp_state) - 1 > 4 * PyLong_SHIFT)) {
                            return (enum ibv_qp_state) ((((((((((enum ibv_qp_state)digits[3]) << PyLong_SHIFT) | (enum ibv_qp_state)digits[2]) << PyLong_SHIFT) | (enum ibv_qp_state)digits[1]) << PyLong_SHIFT) | (enum ibv_qp_state)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(enum ibv_qp_state) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_qp_state, long, PyLong_AsLong(x))
        } else if ((sizeof(enum ibv_qp_state) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_qp_state, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        enum ibv_qp_state val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() or PyLong_AsNativeBytes() not available, cannot convert large enums");
        val = (enum ibv_qp_state) -1;
#endif
        if (unlikely(ret))
            return (enum ibv_qp_state) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum ibv_qp_state");
    return (enum ibv_qp_state) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum ibv_qp_state");
    return (enum ibv_qp_state) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_short(unsigned short value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned short neg_one = (unsigned short) -1, const_zero = (unsigned short) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned short) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned short) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(unsigned short) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned short) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned short) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned short),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(unsigned short));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE unsigned char __Pyx_PyLong_As_unsigned_char(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned char neg_one = (unsigned char) -1, const_zero = (unsigned char) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        unsigned char val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (unsigned char) -1;
        val = __Pyx_PyLong_As_unsigned_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned char, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(unsigned char) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned char) >= 2 * PyLong_SHIFT)) {
                            return (unsigned char) (((((unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned char) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned char) >= 3 * PyLong_SHIFT)) {
                            return (unsigned char) (((((((unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned char) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned char) >= 4 * PyLong_SHIFT)) {
                            return (unsigned char) (((((((((unsigned char)digits[3]) << PyLong_SHIFT) | (unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (unsigned char) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(unsigned char) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned char, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(unsigned char) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned char, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(unsigned char) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned char) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned char) (((unsigned char)-1)*(((((unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(unsigned char) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned char) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned char) ((((((unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(unsigned char) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned char) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned char) (((unsigned char)-1)*(((((((unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned char) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned char) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned char) ((((((((unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(unsigned char) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned char) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned char) (((unsigned char)-1)*(((((((((unsigned char)digits[3]) << PyLong_SHIFT) | (unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned char) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned char) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned char) ((((((((((unsigned char)digits[3]) << PyLong_SHIFT) | (unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(unsigned char) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned char, long, PyLong_AsLong(x))
        } else if ((sizeof(unsigned char) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned char, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        unsigned char val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (unsigned char) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (unsigned char) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (unsigned char) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (unsigned char) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(unsigned char) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((unsigned char) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(unsigned char) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((unsigned char) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((unsigned char) 1) << (sizeof(unsigned char) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (unsigned char) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned char");
    return (unsigned char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned char");
    return (unsigned char) -1;
}

/* CIntFromPy */
static CYTHON_INLINE enum ibv_mtu __Pyx_PyLong_As_enum__ibv_mtu(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_mtu neg_one = (enum ibv_mtu) -1, const_zero = (enum ibv_mtu) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        enum ibv_mtu val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (enum ibv_mtu) -1;
        val = __Pyx_PyLong_As_enum__ibv_mtu(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(enum ibv_mtu, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(enum ibv_mtu) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_mtu, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_mtu) >= 2 * PyLong_SHIFT)) {
                            return (enum ibv_mtu) (((((enum ibv_mtu)digits[1]) << PyLong_SHIFT) | (enum ibv_mtu)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(enum ibv_mtu) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_mtu, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_mtu) >= 3 * PyLong_SHIFT)) {
                            return (enum ibv_mtu) (((((((enum ibv_mtu)digits[2]) << PyLong_SHIFT) | (enum ibv_mtu)digits[1]) << PyLong_SHIFT) | (enum ibv_mtu)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(enum ibv_mtu) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_mtu, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_mtu) >= 4 * PyLong_SHIFT)) {
                            return (enum ibv_mtu) (((((((((enum ibv_mtu)digits[3]) << PyLong_SHIFT) | (enum ibv_mtu)digits[2]) << PyLong_SHIFT) | (enum ibv_mtu)digits[1]) << PyLong_SHIFT) | (enum ibv_mtu)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (enum ibv_mtu) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(enum ibv_mtu) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_mtu, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(enum ibv_mtu) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_mtu, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(enum ibv_mtu, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(enum ibv_mtu) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_mtu, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_mtu) - 1 > 2 * PyLong_SHIFT)) {
                            return (enum ibv_mtu) (((enum ibv_mtu)-1)*(((((enum ibv_mtu)digits[1]) << PyLong_SHIFT) | (enum ibv_mtu)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(enum ibv_mtu) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_mtu, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_mtu) - 1 > 2 * PyLong_SHIFT)) {
                            return (enum ibv_mtu) ((((((enum ibv_mtu)digits[1]) << PyLong_SHIFT) | (enum ibv_mtu)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(enum ibv_mtu) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_mtu, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_mtu) - 1 > 3 * PyLong_SHIFT)) {
                            return (enum ibv_mtu) (((enum ibv_mtu)-1)*(((((((enum ibv_mtu)digits[2]) << PyLong_SHIFT) | (enum ibv_mtu)digits[1]) << PyLong_SHIFT) | (enum ibv_mtu)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(enum ibv_mtu) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_mtu, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_mtu) - 1 > 3 * PyLong_SHIFT)) {
                            return (enum ibv_mtu) ((((((((enum ibv_mtu)digits[2]) << PyLong_SHIFT) | (enum ibv_mtu)digits[1]) << PyLong_SHIFT) | (enum ibv_mtu)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(enum ibv_mtu) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_mtu, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_mtu) - 1 > 4 * PyLong_SHIFT)) {
                            return (enum ibv_mtu) (((enum ibv_mtu)-1)*(((((((((enum ibv_mtu)digits[3]) << PyLong_SHIFT) | (enum ibv_mtu)digits[2]) << PyLong_SHIFT) | (enum ibv_mtu)digits[1]) << PyLong_SHIFT) | (enum ibv_mtu)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(enum ibv_mtu) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_mtu, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_mtu) - 1 > 4 * PyLong_SHIFT)) {
                            return (enum ibv_mtu) ((((((((((enum ibv_mtu)digits[3]) << PyLong_SHIFT) | (enum ibv_mtu)digits[2]) << PyLong_SHIFT) | (enum ibv_mtu)digits[1]) << PyLong_SHIFT) | (enum ibv_mtu)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(enum ibv_mtu) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_mtu, long, PyLong_AsLong(x))
        } else if ((sizeof(enum ibv_mtu) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_mtu, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        enum ibv_mtu val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() or PyLong_AsNativeBytes() not available, cannot convert large enums");
        val = (enum ibv_mtu) -1;
#endif
        if (unlikely(ret))
            return (enum ibv_mtu) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum ibv_mtu");
    return (enum ibv_mtu) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum ibv_mtu");
    return (enum ibv_mtu) -1;
}

/* CIntFromPy */
static CYTHON_INLINE enum ibv_mig_state __Pyx_PyLong_As_enum__ibv_mig_state(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_mig_state neg_one = (enum ibv_mig_state) -1, const_zero = (enum ibv_mig_state) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        enum ibv_mig_state val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (enum ibv_mig_state) -1;
        val = __Pyx_PyLong_As_enum__ibv_mig_state(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(enum ibv_mig_state, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(enum ibv_mig_state) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_mig_state, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_mig_state) >= 2 * PyLong_SHIFT)) {
                            return (enum ibv_mig_state) (((((enum ibv_mig_state)digits[1]) << PyLong_SHIFT) | (enum ibv_mig_state)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(enum ibv_mig_state) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_mig_state, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_mig_state) >= 3 * PyLong_SHIFT)) {
                            return (enum ibv_mig_state) (((((((enum ibv_mig_state)digits[2]) << PyLong_SHIFT) | (enum ibv_mig_state)digits[1]) << PyLong_SHIFT) | (enum ibv_mig_state)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(enum ibv_mig_state) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_mig_state, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_mig_state) >= 4 * PyLong_SHIFT)) {
                            return (enum ibv_mig_state) (((((((((enum ibv_mig_state)digits[3]) << PyLong_SHIFT) | (enum ibv_mig_state)digits[2]) << PyLong_SHIFT) | (enum ibv_mig_state)digits[1]) << PyLong_SHIFT) | (enum ibv_mig_state)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (enum ibv_mig_state) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(enum ibv_mig_state) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_mig_state, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(enum ibv_mig_state) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_mig_state, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(enum ibv_mig_state, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(enum ibv_mig_state) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_mig_state, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_mig_state) - 1 > 2 * PyLong_SHIFT)) {
                            return (enum ibv_mig_state) (((enum ibv_mig_state)-1)*(((((enum ibv_mig_state)digits[1]) << PyLong_SHIFT) | (enum ibv_mig_state)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(enum ibv_mig_state) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_mig_state, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_mig_state) - 1 > 2 * PyLong_SHIFT)) {
                            return (enum ibv_mig_state) ((((((enum ibv_mig_state)digits[1]) << PyLong_SHIFT) | (enum ibv_mig_state)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(enum ibv_mig_state) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_mig_state, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_mig_state) - 1 > 3 * PyLong_SHIFT)) {
                            return (enum ibv_mig_state) (((enum ibv_mig_state)-1)*(((((((enum ibv_mig_state)digits[2]) << PyLong_SHIFT) | (enum ibv_mig_state)digits[1]) << PyLong_SHIFT) | (enum ibv_mig_state)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(enum ibv_mig_state) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_mig_state, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_mig_state) - 1 > 3 * PyLong_SHIFT)) {
                            return (enum ibv_mig_state) ((((((((enum ibv_mig_state)digits[2]) << PyLong_SHIFT) | (enum ibv_mig_state)digits[1]) << PyLong_SHIFT) | (enum ibv_mig_state)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(enum ibv_mig_state) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_mig_state, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_mig_state) - 1 > 4 * PyLong_SHIFT)) {
                            return (enum ibv_mig_state) (((enum ibv_mig_state)-1)*(((((((((enum ibv_mig_state)digits[3]) << PyLong_SHIFT) | (enum ibv_mig_state)digits[2]) << PyLong_SHIFT) | (enum ibv_mig_state)digits[1]) << PyLong_SHIFT) | (enum ibv_mig_state)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(enum ibv_mig_state) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_mig_state, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_mig_state) - 1 > 4 * PyLong_SHIFT)) {
                            return (enum ibv_mig_state) ((((((((((enum ibv_mig_state)digits[3]) << PyLong_SHIFT) | (enum ibv_mig_state)digits[2]) << PyLong_SHIFT) | (enum ibv_mig_state)digits[1]) << PyLong_SHIFT) | (enum ibv_mig_state)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(enum ibv_mig_state) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_mig_state, long, PyLong_AsLong(x))
        } else if ((sizeof(enum ibv_mig_state) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_mig_state, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        enum ibv_mig_state val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() or PyLong_AsNativeBytes() not available, cannot convert large enums");
        val = (enum ibv_mig_state) -1;
#endif
        if (unlikely(ret))
            return (enum ibv_mig_state) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum ibv_mig_state");
    return (enum ibv_mig_state) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum ibv_mig_state");
    return (enum ibv_mig_state) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_long(unsigned long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned long neg_one = (unsigned long) -1, const_zero = (unsigned long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(unsigned long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(unsigned long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyLong_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_char(unsigned char value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned char neg_one = (unsigned char) -1, const_zero = (unsigned char) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned char) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned char) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(unsigned char) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned char) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned char) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned char),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(unsigned char));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE uint32_t __Pyx_PyLong_As_uint32_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        uint32_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (uint32_t) -1;
        val = __Pyx_PyLong_As_uint32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint32_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(uint32_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) >= 2 * PyLong_SHIFT)) {
                            return (uint32_t) (((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint32_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) >= 3 * PyLong_SHIFT)) {
                            return (uint32_t) (((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint32_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) >= 4 * PyLong_SHIFT)) {
                            return (uint32_t) (((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (uint32_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(uint32_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint32_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(uint32_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint32_t) (((uint32_t)-1)*(((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(uint32_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint32_t) ((((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint32_t) (((uint32_t)-1)*(((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint32_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint32_t) ((((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint32_t) (((uint32_t)-1)*(((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint32_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint32_t) ((((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(uint32_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, long, PyLong_AsLong(x))
        } else if ((sizeof(uint32_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        uint32_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (uint32_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (uint32_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (uint32_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (uint32_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(uint32_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((uint32_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(uint32_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((uint32_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((uint32_t) 1) << (sizeof(uint32_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (uint32_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint32_t");
    return (uint32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint32_t");
    return (uint32_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_uint32_t(uint32_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint32_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(uint32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint32_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(uint32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(uint32_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint32_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE enum ibv_wr_opcode __Pyx_PyLong_As_enum__ibv_wr_opcode(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_wr_opcode neg_one = (enum ibv_wr_opcode) -1, const_zero = (enum ibv_wr_opcode) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        enum ibv_wr_opcode val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (enum ibv_wr_opcode) -1;
        val = __Pyx_PyLong_As_enum__ibv_wr_opcode(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(enum ibv_wr_opcode, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(enum ibv_wr_opcode) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wr_opcode, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wr_opcode) >= 2 * PyLong_SHIFT)) {
                            return (enum ibv_wr_opcode) (((((enum ibv_wr_opcode)digits[1]) << PyLong_SHIFT) | (enum ibv_wr_opcode)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(enum ibv_wr_opcode) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wr_opcode, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wr_opcode) >= 3 * PyLong_SHIFT)) {
                            return (enum ibv_wr_opcode) (((((((enum ibv_wr_opcode)digits[2]) << PyLong_SHIFT) | (enum ibv_wr_opcode)digits[1]) << PyLong_SHIFT) | (enum ibv_wr_opcode)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(enum ibv_wr_opcode) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wr_opcode, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wr_opcode) >= 4 * PyLong_SHIFT)) {
                            return (enum ibv_wr_opcode) (((((((((enum ibv_wr_opcode)digits[3]) << PyLong_SHIFT) | (enum ibv_wr_opcode)digits[2]) << PyLong_SHIFT) | (enum ibv_wr_opcode)digits[1]) << PyLong_SHIFT) | (enum ibv_wr_opcode)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (enum ibv_wr_opcode) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(enum ibv_wr_opcode) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_wr_opcode, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(enum ibv_wr_opcode) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_wr_opcode, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(enum ibv_wr_opcode, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(enum ibv_wr_opcode) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wr_opcode, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wr_opcode) - 1 > 2 * PyLong_SHIFT)) {
                            return (enum ibv_wr_opcode) (((enum ibv_wr_opcode)-1)*(((((enum ibv_wr_opcode)digits[1]) << PyLong_SHIFT) | (enum ibv_wr_opcode)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(enum ibv_wr_opcode) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wr_opcode, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wr_opcode) - 1 > 2 * PyLong_SHIFT)) {
                            return (enum ibv_wr_opcode) ((((((enum ibv_wr_opcode)digits[1]) << PyLong_SHIFT) | (enum ibv_wr_opcode)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(enum ibv_wr_opcode) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wr_opcode, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wr_opcode) - 1 > 3 * PyLong_SHIFT)) {
                            return (enum ibv_wr_opcode) (((enum ibv_wr_opcode)-1)*(((((((enum ibv_wr_opcode)digits[2]) << PyLong_SHIFT) | (enum ibv_wr_opcode)digits[1]) << PyLong_SHIFT) | (enum ibv_wr_opcode)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(enum ibv_wr_opcode) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wr_opcode, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wr_opcode) - 1 > 3 * PyLong_SHIFT)) {
                            return (enum ibv_wr_opcode) ((((((((enum ibv_wr_opcode)digits[2]) << PyLong_SHIFT) | (enum ibv_wr_opcode)digits[1]) << PyLong_SHIFT) | (enum ibv_wr_opcode)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(enum ibv_wr_opcode) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wr_opcode, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wr_opcode) - 1 > 4 * PyLong_SHIFT)) {
                            return (enum ibv_wr_opcode) (((enum ibv_wr_opcode)-1)*(((((((((enum ibv_wr_opcode)digits[3]) << PyLong_SHIFT) | (enum ibv_wr_opcode)digits[2]) << PyLong_SHIFT) | (enum ibv_wr_opcode)digits[1]) << PyLong_SHIFT) | (enum ibv_wr_opcode)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(enum ibv_wr_opcode) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wr_opcode, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wr_opcode) - 1 > 4 * PyLong_SHIFT)) {
                            return (enum ibv_wr_opcode) ((((((((((enum ibv_wr_opcode)digits[3]) << PyLong_SHIFT) | (enum ibv_wr_opcode)digits[2]) << PyLong_SHIFT) | (enum ibv_wr_opcode)digits[1]) << PyLong_SHIFT) | (enum ibv_wr_opcode)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(enum ibv_wr_opcode) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_wr_opcode, long, PyLong_AsLong(x))
        } else if ((sizeof(enum ibv_wr_opcode) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_wr_opcode, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        enum ibv_wr_opcode val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() or PyLong_AsNativeBytes() not available, cannot convert large enums");
        val = (enum ibv_wr_opcode) -1;
#endif
        if (unlikely(ret))
            return (enum ibv_wr_opcode) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum ibv_wr_opcode");
    return (enum ibv_wr_opcode) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum ibv_wr_opcode");
    return (enum ibv_wr_opcode) -1;
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyLong_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyLong_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 2 * PyLong_SHIFT)) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 3 * PyLong_SHIFT)) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 4 * PyLong_SHIFT)) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (size_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(size_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(size_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
        } else if ((sizeof(size_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        size_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (size_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (size_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (size_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (size_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(size_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((size_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(size_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((size_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((size_t) 1) << (sizeof(size_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (size_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE uint64_t __Pyx_PyLong_As_uint64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        uint64_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (uint64_t) -1;
        val = __Pyx_PyLong_As_uint64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint64_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(uint64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) >= 2 * PyLong_SHIFT)) {
                            return (uint64_t) (((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) >= 3 * PyLong_SHIFT)) {
                            return (uint64_t) (((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) >= 4 * PyLong_SHIFT)) {
                            return (uint64_t) (((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (uint64_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(uint64_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint64_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(uint64_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint64_t) (((uint64_t)-1)*(((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(uint64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint64_t) ((((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint64_t) (((uint64_t)-1)*(((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint64_t) ((((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint64_t) (((uint64_t)-1)*(((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint64_t) ((((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(uint64_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, long, PyLong_AsLong(x))
        } else if ((sizeof(uint64_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        uint64_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (uint64_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (uint64_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (uint64_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (uint64_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(uint64_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((uint64_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(uint64_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((uint64_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((uint64_t) 1) << (sizeof(uint64_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (uint64_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint64_t");
    return (uint64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint64_t");
    return (uint64_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_uint64_t(uint64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint64_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(uint64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint64_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(uint64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(uint64_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint64_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE uint8_t __Pyx_PyLong_As_uint8_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint8_t neg_one = (uint8_t) -1, const_zero = (uint8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        uint8_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (uint8_t) -1;
        val = __Pyx_PyLong_As_uint8_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint8_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(uint8_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) >= 2 * PyLong_SHIFT)) {
                            return (uint8_t) (((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint8_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) >= 3 * PyLong_SHIFT)) {
                            return (uint8_t) (((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint8_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) >= 4 * PyLong_SHIFT)) {
                            return (uint8_t) (((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (uint8_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(uint8_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint8_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(uint8_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint8_t) (((uint8_t)-1)*(((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(uint8_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint8_t) ((((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint8_t) (((uint8_t)-1)*(((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint8_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint8_t) ((((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint8_t) (((uint8_t)-1)*(((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint8_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint8_t) ((((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(uint8_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint8_t, long, PyLong_AsLong(x))
        } else if ((sizeof(uint8_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint8_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        uint8_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (uint8_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (uint8_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (uint8_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (uint8_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(uint8_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((uint8_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(uint8_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((uint8_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((uint8_t) 1) << (sizeof(uint8_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (uint8_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint8_t");
    return (uint8_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint8_t");
    return (uint8_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE uint16_t __Pyx_PyLong_As_uint16_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint16_t neg_one = (uint16_t) -1, const_zero = (uint16_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        uint16_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (uint16_t) -1;
        val = __Pyx_PyLong_As_uint16_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint16_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(uint16_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint16_t) >= 2 * PyLong_SHIFT)) {
                            return (uint16_t) (((((uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint16_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint16_t) >= 3 * PyLong_SHIFT)) {
                            return (uint16_t) (((((((uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint16_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint16_t) >= 4 * PyLong_SHIFT)) {
                            return (uint16_t) (((((((((uint16_t)digits[3]) << PyLong_SHIFT) | (uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (uint16_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(uint16_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint16_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(uint16_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint16_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint16_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(uint16_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint16_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint16_t) (((uint16_t)-1)*(((((uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(uint16_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint16_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint16_t) ((((((uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(uint16_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint16_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint16_t) (((uint16_t)-1)*(((((((uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint16_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint16_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint16_t) ((((((((uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(uint16_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint16_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint16_t) (((uint16_t)-1)*(((((((((uint16_t)digits[3]) << PyLong_SHIFT) | (uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint16_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint16_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint16_t) ((((((((((uint16_t)digits[3]) << PyLong_SHIFT) | (uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(uint16_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint16_t, long, PyLong_AsLong(x))
        } else if ((sizeof(uint16_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint16_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        uint16_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (uint16_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (uint16_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (uint16_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (uint16_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(uint16_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((uint16_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(uint16_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((uint16_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((uint16_t) 1) << (sizeof(uint16_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (uint16_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint16_t");
    return (uint16_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint16_t");
    return (uint16_t) -1;
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
static __Pyx_TypeName
__Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp)
{
    PyObject *module = NULL, *name = NULL, *result = NULL;
    #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
    name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_qualname);
    #else
    name = PyType_GetQualName(tp);
    #endif
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) goto bad;
    module = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_module);
    if (unlikely(module == NULL) || unlikely(!PyUnicode_Check(module))) goto bad;
    if (PyUnicode_CompareWithASCIIString(module, "builtins") == 0) {
        result = name;
        name = NULL;
        goto done;
    }
    result = PyUnicode_FromFormat("%U.%U", module, name);
    if (unlikely(result == NULL)) goto bad;
  done:
    Py_XDECREF(name);
    Py_XDECREF(module);
    return result;
  bad:
    PyErr_Clear();
    if (name) {
        result = name;
        name = NULL;
    } else {
        result = __Pyx_NewRef(__pyx_mstate_global->__pyx_kp_u__9);
    }
    goto done;
}
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* GetRuntimeVersion */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
void __Pyx_init_runtime_version(void) {
    if (__Pyx_cached_runtime_version == 0) {
        const char* rt_version = Py_GetVersion();
        unsigned long version = 0;
        unsigned long factor = 0x01000000UL;
        unsigned int digit = 0;
        int i = 0;
        while (factor) {
            while ('0' <= rt_version[i] && rt_version[i] <= '9') {
                digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
                ++i;
            }
            version += factor * digit;
            if (rt_version[i] != '.')
                break;
            digit = 0;
            factor >>= 8;
            ++i;
        }
        __Pyx_cached_runtime_version = version;
    }
}
#endif
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    return Py_Version & ~0xFFUL;
#else
    return __Pyx_cached_runtime_version;
#endif
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* NewCodeObj */
#if CYTHON_COMPILING_IN_LIMITED_API
    static PyObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                       PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                       PyObject *fv, PyObject *cell, PyObject* fn,
                                       PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
#elif PY_VERSION_HEX >= 0x030B0000
  static PyCodeObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                         PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                         PyObject *fv, PyObject *cell, PyObject* fn,
                                         PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, __pyx_mstate_global->__pyx_empty_bytes);
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030c00A1
    if (likely(result))
        result->_co_firsttraceable = 0;
    #endif
    return result;
  }
#elif !CYTHON_COMPILING_IN_PYPY
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
) {
    PyObject *code_obj = NULL, *varnames_tuple_dedup = NULL, *code_bytes = NULL;
    Py_ssize_t var_count = (Py_ssize_t) descr.nlocals;
    PyObject *varnames_tuple = PyTuple_New(var_count);
    if (unlikely(!varnames_tuple)) return NULL;
    for (Py_ssize_t i=0; i < var_count; i++) {
        Py_INCREF(varnames[i]);
        if (__Pyx_PyTuple_SET_ITEM(varnames_tuple, i, varnames[i]) != (0)) goto done;
    }
    #if CYTHON_COMPILING_IN_LIMITED_API
    varnames_tuple_dedup = PyDict_GetItem(tuple_dedup_map, varnames_tuple);
    if (!varnames_tuple_dedup) {
        if (unlikely(PyDict_SetItem(tuple_dedup_map, varnames_tuple, varnames_tuple) < 0)) goto done;
        varnames_tuple_dedup = varnames_tuple;
    }
    #else
    varnames_tuple_dedup = PyDict_SetDefault(tuple_dedup_map, varnames_tuple, varnames_tuple);
    if (unlikely(!varnames_tuple_dedup)) goto done;
    #endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(varnames_tuple_dedup);
    #endif
    if (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table != NULL && !CYTHON_COMPILING_IN_GRAAL) {
        Py_ssize_t line_table_length = __Pyx_PyBytes_GET_SIZE(line_table);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(line_table_length == -1)) goto done;
        #endif
        Py_ssize_t code_len = (line_table_length * 2 + 4) & ~3LL;
        code_bytes = PyBytes_FromStringAndSize(NULL, code_len);
        if (unlikely(!code_bytes)) goto done;
        char* c_code_bytes = PyBytes_AsString(code_bytes);
        if (unlikely(!c_code_bytes)) goto done;
        memset(c_code_bytes, 0, (size_t) code_len);
    }
    code_obj = (PyObject*) __Pyx__PyCode_New(
        (int) descr.argcount,
        (int) descr.num_posonly_args,
        (int) descr.num_kwonly_args,
        (int) descr.nlocals,
        0,
        (int) descr.flags,
        code_bytes ? code_bytes : __pyx_mstate_global->__pyx_empty_bytes,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        varnames_tuple_dedup,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        filename,
        funcname,
        (int) descr.first_line,
        (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table) ? line_table : __pyx_mstate_global->__pyx_empty_bytes
    );
done:
    Py_XDECREF(code_bytes);
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(varnames_tuple_dedup);
    #endif
    Py_DECREF(varnames_tuple);
    return code_obj;
}

/* DecompressString */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo) {
    PyObject *module = NULL, *decompress, *compressed_bytes, *decompressed;
    const char* module_name = algo == 3 ? "compression.zstd" : algo == 2 ? "bz2" : "zlib";
    PyObject *methodname = PyUnicode_FromString("decompress");
    if (unlikely(!methodname)) return NULL;
    #if __PYX_LIMITED_VERSION_HEX >= 0x030e0000
    if (algo == 3) {
        PyObject *fromlist = Py_BuildValue("[O]", methodname);
        if (unlikely(!fromlist)) goto bad;
        module = PyImport_ImportModuleLevel("compression.zstd", NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
    } else
    #endif
        module = PyImport_ImportModule(module_name);
    if (unlikely(!module)) goto import_failed;
    decompress = PyObject_GetAttr(module, methodname);
    if (unlikely(!decompress)) goto import_failed;
    {
        #ifdef __cplusplus
            char *memview_bytes = const_cast<char*>(s);
        #else
            #if defined(__clang__)
              #pragma clang diagnostic push
              #pragma clang diagnostic ignored "-Wcast-qual"
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic push
              #pragma GCC diagnostic ignored "-Wcast-qual"
            #endif
            char *memview_bytes = (char*) s;
            #if defined(__clang__)
              #pragma clang diagnostic pop
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic pop
            #endif
        #endif
        #if CYTHON_COMPILING_IN_LIMITED_API && !defined(PyBUF_READ)
        int memview_flags = 0x100;
        #else
        int memview_flags = PyBUF_READ;
        #endif
        compressed_bytes = PyMemoryView_FromMemory(memview_bytes, length, memview_flags);
    }
    if (unlikely(!compressed_bytes)) {
        Py_DECREF(decompress);
        goto bad;
    }
    decompressed = PyObject_CallFunctionObjArgs(decompress, compressed_bytes, NULL);
    Py_DECREF(compressed_bytes);
    Py_DECREF(decompress);
    Py_DECREF(module);
    Py_DECREF(methodname);
    return decompressed;
import_failed:
    PyErr_Format(PyExc_ImportError,
        "Failed to import '%.20s.decompress' - cannot initialise module strings. "
        "String compression was configured with the C macro 'CYTHON_COMPRESS_STRINGS=%d'.",
        module_name, algo);
bad:
    Py_XDECREF(module);
    Py_DECREF(methodname);
    return NULL;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    {
        const char* result;
        Py_ssize_t unicode_length;
        CYTHON_MAYBE_UNUSED_VAR(unicode_length); // only for __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (unlikely(PyArg_Parse(o, "s#", &result, length) < 0)) return NULL;
        #else
        result = PyUnicode_AsUTF8AndSize(o, length);
        #endif
        #if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        unicode_length = PyUnicode_GetLength(o);
        if (unlikely(unicode_length < 0)) return NULL;
        if (unlikely(unicode_length != *length)) {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
        #endif
        return result;
    }
#else
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
}
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
    if (PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
    if (PyByteArray_Check(o)) {
#if (CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS) || (CYTHON_COMPILING_IN_PYPY && (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE)))
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
#else
        *length = PyByteArray_Size(o);
        if (*length == -1) return NULL;
        return PyByteArray_AsString(o);
#endif
    } else
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_LongWrongResultType(PyObject* result) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(result));
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
    PyErr_Format(PyExc_TypeError,
                 "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ")",
                 result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  PyObject *res = NULL;
  if (likely(PyLong_Check(x)))
      return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  if (likely(m && m->nb_int)) {
      res = m->nb_int(x);
  }
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
      res = PyNumber_Long(x);
  }
#endif
  if (likely(res)) {
      if (unlikely(!PyLong_CheckExact(res))) {
          return __Pyx_PyNumber_LongWrongResultType(res);
      }
  }
  else if (!PyErr_Occurred()) {
      PyErr_SetString(PyExc_TypeError,
                      "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyLong_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyLong_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b) {
    CYTHON_UNUSED_VAR(b);
    return __Pyx_NewRef(Py_None);
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return __Pyx_NewRef(b ? Py_True: Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t ival) {
    return PyLong_FromSize_t(ival);
}


/* MultiPhaseInitModuleState */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
#ifndef CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#if (CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX >= 0x030C0000)
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 1
#else
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 0
#endif
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE && !CYTHON_ATOMICS
#error "Module state with PEP489 requires atomics. Currently that's one of\
 C11, C++11, gcc atomic intrinsics or MSVC atomic intrinsics"
#endif
#if !CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#define __Pyx_ModuleStateLookup_Lock()
#define __Pyx_ModuleStateLookup_Unlock()
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
static PyMutex __Pyx_ModuleStateLookup_mutex = {0};
#define __Pyx_ModuleStateLookup_Lock() PyMutex_Lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() PyMutex_Unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(__cplusplus) && __cplusplus >= 201103L
#include <mutex>
static std::mutex __Pyx_ModuleStateLookup_mutex;
#define __Pyx_ModuleStateLookup_Lock() __Pyx_ModuleStateLookup_mutex.lock()
#define __Pyx_ModuleStateLookup_Unlock() __Pyx_ModuleStateLookup_mutex.unlock()
#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ > 201112L) && !defined(__STDC_NO_THREADS__)
#include <threads.h>
static mtx_t __Pyx_ModuleStateLookup_mutex;
static once_flag __Pyx_ModuleStateLookup_mutex_once_flag = ONCE_FLAG_INIT;
static void __Pyx_ModuleStateLookup_initialize_mutex(void) {
    mtx_init(&__Pyx_ModuleStateLookup_mutex, mtx_plain);
}
#define __Pyx_ModuleStateLookup_Lock()\
  call_once(&__Pyx_ModuleStateLookup_mutex_once_flag, __Pyx_ModuleStateLookup_initialize_mutex);\
  mtx_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() mtx_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(HAVE_PTHREAD_H)
#include <pthread.h>
static pthread_mutex_t __Pyx_ModuleStateLookup_mutex = PTHREAD_MUTEX_INITIALIZER;
#define __Pyx_ModuleStateLookup_Lock() pthread_mutex_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() pthread_mutex_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(_WIN32)
#include <Windows.h>  // synchapi.h on its own doesn't work
static SRWLOCK __Pyx_ModuleStateLookup_mutex = SRWLOCK_INIT;
#define __Pyx_ModuleStateLookup_Lock() AcquireSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() ReleaseSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#else
#error "No suitable lock available for CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE.\
 Requires C standard >= C11, or C++ standard >= C++11,\
 or pthreads, or the Windows 32 API, or Python >= 3.13."
#endif
typedef struct {
    int64_t id;
    PyObject *module;
} __Pyx_InterpreterIdAndModule;
typedef struct {
    char interpreter_id_as_index;
    Py_ssize_t count;
    Py_ssize_t allocated;
    __Pyx_InterpreterIdAndModule table[1];
} __Pyx_ModuleStateLookupData;
#define __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE 32
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_int_type __Pyx_ModuleStateLookup_read_counter = 0;
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_ptr_type __Pyx_ModuleStateLookup_data = 0;
#else
static __Pyx_ModuleStateLookupData* __Pyx_ModuleStateLookup_data = NULL;
#endif
static __Pyx_InterpreterIdAndModule* __Pyx_State_FindModuleStateLookupTableLowerBound(
        __Pyx_InterpreterIdAndModule* table,
        Py_ssize_t count,
        int64_t interpreterId) {
    __Pyx_InterpreterIdAndModule* begin = table;
    __Pyx_InterpreterIdAndModule* end = begin + count;
    if (begin->id == interpreterId) {
        return begin;
    }
    while ((end - begin) > __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
        __Pyx_InterpreterIdAndModule* halfway = begin + (end - begin)/2;
        if (halfway->id == interpreterId) {
            return halfway;
        }
        if (halfway->id < interpreterId) {
            begin = halfway;
        } else {
            end = halfway;
        }
    }
    for (; begin < end; ++begin) {
        if (begin->id >= interpreterId) return begin;
    }
    return begin;
}
static PyObject *__Pyx_State_FindModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return NULL;
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData* data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
    {
        __pyx_atomic_incr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        if (likely(data)) {
            __Pyx_ModuleStateLookupData* new_data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_acquire(&__Pyx_ModuleStateLookup_data);
            if (likely(data == new_data)) {
                goto read_finished;
            }
        }
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        __Pyx_ModuleStateLookup_Lock();
        __pyx_atomic_incr_relaxed(&__Pyx_ModuleStateLookup_read_counter);
        data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
        __Pyx_ModuleStateLookup_Unlock();
    }
  read_finished:;
#else
    __Pyx_ModuleStateLookupData* data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_InterpreterIdAndModule* found = NULL;
    if (unlikely(!data)) goto end;
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            found = data->table+interpreter_id;
        }
    } else {
        found = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
    }
  end:
    {
        PyObject *result=NULL;
        if (found && found->id == interpreter_id) {
            result = found->module;
        }
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
#endif
        return result;
    }
}
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static void __Pyx_ModuleStateLookup_wait_until_no_readers(void) {
    while (__pyx_atomic_load(&__Pyx_ModuleStateLookup_read_counter) != 0);
}
#else
#define __Pyx_ModuleStateLookup_wait_until_no_readers()
#endif
static int __Pyx_State_AddModuleInterpIdAsIndex(__Pyx_ModuleStateLookupData **old_data, PyObject* module, int64_t interpreter_id) {
    Py_ssize_t to_allocate = (*old_data)->allocated;
    while (to_allocate <= interpreter_id) {
        if (to_allocate == 0) to_allocate = 1;
        else to_allocate *= 2;
    }
    __Pyx_ModuleStateLookupData *new_data = *old_data;
    if (to_allocate != (*old_data)->allocated) {
         new_data = (__Pyx_ModuleStateLookupData *)realloc(
            *old_data,
            sizeof(__Pyx_ModuleStateLookupData)+(to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
        if (!new_data) {
            PyErr_NoMemory();
            return -1;
        }
        for (Py_ssize_t i = new_data->allocated; i < to_allocate; ++i) {
            new_data->table[i].id = i;
            new_data->table[i].module = NULL;
        }
        new_data->allocated = to_allocate;
    }
    new_data->table[interpreter_id].module = module;
    if (new_data->count < interpreter_id+1) {
        new_data->count = interpreter_id+1;
    }
    *old_data = new_data;
    return 0;
}
static void __Pyx_State_ConvertFromInterpIdAsIndex(__Pyx_ModuleStateLookupData *data) {
    __Pyx_InterpreterIdAndModule *read = data->table;
    __Pyx_InterpreterIdAndModule *write = data->table;
    __Pyx_InterpreterIdAndModule *end = read + data->count;
    for (; read<end; ++read) {
        if (read->module) {
            write->id = read->id;
            write->module = read->module;
            ++write;
        }
    }
    data->count = write - data->table;
    for (; write<end; ++write) {
        write->id = 0;
        write->module = NULL;
    }
    data->interpreter_id_as_index = 0;
}
static int __Pyx_State_AddModule(PyObject* module, CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    int result = 0;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *old_data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *old_data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_ModuleStateLookupData *new_data = old_data;
    if (!new_data) {
        new_data = (__Pyx_ModuleStateLookupData *)calloc(1, sizeof(__Pyx_ModuleStateLookupData));
        if (!new_data) {
            result = -1;
            PyErr_NoMemory();
            goto end;
        }
        new_data->allocated = 1;
        new_data->interpreter_id_as_index = 1;
    }
    __Pyx_ModuleStateLookup_wait_until_no_readers();
    if (new_data->interpreter_id_as_index) {
        if (interpreter_id < __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
            result = __Pyx_State_AddModuleInterpIdAsIndex(&new_data, module, interpreter_id);
            goto end;
        }
        __Pyx_State_ConvertFromInterpIdAsIndex(new_data);
    }
    {
        Py_ssize_t insert_at = 0;
        {
            __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
                new_data->table, new_data->count, interpreter_id);
            assert(lower_bound);
            insert_at = lower_bound - new_data->table;
            if (unlikely(insert_at < new_data->count && lower_bound->id == interpreter_id)) {
                lower_bound->module = module;
                goto end;  // already in table, nothing more to do
            }
        }
        if (new_data->count+1 >= new_data->allocated) {
            Py_ssize_t to_allocate = (new_data->count+1)*2;
            new_data =
                (__Pyx_ModuleStateLookupData*)realloc(
                    new_data,
                    sizeof(__Pyx_ModuleStateLookupData) +
                    (to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
            if (!new_data) {
                result = -1;
                new_data = old_data;
                PyErr_NoMemory();
                goto end;
            }
            new_data->allocated = to_allocate;
        }
        ++new_data->count;
        int64_t last_id = interpreter_id;
        PyObject *last_module = module;
        for (Py_ssize_t i=insert_at; i<new_data->count; ++i) {
            int64_t current_id = new_data->table[i].id;
            new_data->table[i].id = last_id;
            last_id = current_id;
            PyObject *current_module = new_data->table[i].module;
            new_data->table[i].module = last_module;
            last_module = current_module;
        }
    }
  end:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, new_data);
#else
    __Pyx_ModuleStateLookup_data = new_data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return result;
}
static int __Pyx_State_RemoveModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *data = __Pyx_ModuleStateLookup_data;
#endif
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            data->table[interpreter_id].module = NULL;
        }
        goto done;
    }
    {
        __Pyx_ModuleStateLookup_wait_until_no_readers();
        __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
        if (!lower_bound) goto done;
        if (lower_bound->id != interpreter_id) goto done;
        __Pyx_InterpreterIdAndModule *end = data->table+data->count;
        for (;lower_bound<end-1; ++lower_bound) {
            lower_bound->id = (lower_bound+1)->id;
            lower_bound->module = (lower_bound+1)->module;
        }
    }
    --data->count;
    if (data->count == 0) {
        free(data);
        data = NULL;
    }
  done:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, data);
#else
    __Pyx_ModuleStateLookup_data = data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return 0;
}
#endif

/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
