/* Generated by Cython 3.2.4 */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
/* InitLimitedAPI */
#if defined(Py_LIMITED_API)
  #if !defined(CYTHON_LIMITED_API)
  #define CYTHON_LIMITED_API 1
  #endif
#elif defined(CYTHON_LIMITED_API)
  #ifdef _MSC_VER
  #pragma message ("Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.")
  #else
  #warning Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x03080000
    #error Cython requires Python 3.8+.
#else
#define __PYX_ABI_VERSION "3_2_4"
#define CYTHON_HEX_VERSION 0x030204F0
#define CYTHON_FUTURE_DIVISION 1
/* CModulePreamble */
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 1
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 1
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PYPY_VERSION_NUM >= 0x07030C00)
  #endif
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_NUM >= 0x07031100)
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND (__PYX_LIMITED_VERSION_HEX >= 0x030A0000)
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #ifndef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 1
  #endif
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #ifdef Py_GIL_DISABLED
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 1
  #else
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #endif
  #if PY_VERSION_HEX < 0x030A0000
    #undef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #elif !defined(CYTHON_USE_TYPE_SLOTS)
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLIST_INTERNALS)
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #elif !defined(CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL 0
  #elif !defined(CYTHON_FAST_GIL)
    #define CYTHON_FAST_GIL (PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_SYS_MONITORING
    #define CYTHON_USE_SYS_MONITORING (PY_VERSION_HEX >= 0x030d00B1)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5 && !CYTHON_USE_MODULE_STATE)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS (!CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
  #if defined(CYTHON_IMMORTAL_CONSTANTS) && PY_VERSION_HEX < 0x030C0000
    #undef CYTHON_IMMORTAL_CONSTANTS
    #define CYTHON_IMMORTAL_CONSTANTS 0  // definitely won't work
  #elif !defined(CYTHON_IMMORTAL_CONSTANTS)
    #define CYTHON_IMMORTAL_CONSTANTS (PY_VERSION_HEX >= 0x030C0000 && !CYTHON_USE_MODULE_STATE && CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
#endif
#ifndef CYTHON_COMPRESS_STRINGS
  #define CYTHON_COMPRESS_STRINGS 1
#endif
#ifndef CYTHON_FAST_PYCCALL
#define CYTHON_FAST_PYCCALL  CYTHON_FAST_PYCALL
#endif
#ifndef CYTHON_VECTORCALL
#if CYTHON_COMPILING_IN_LIMITED_API
#define CYTHON_VECTORCALL  (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
#else
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL)
#endif
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#include <stdint.h>
typedef uintptr_t  __pyx_uintptr_t;
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifndef Py_UNREACHABLE
  #define Py_UNREACHABLE()  assert(0); abort()
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

/* CInitCode */
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

/* PythonCompatibility */
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#define __Pyx_BUILTIN_MODULE_NAME "builtins"
#define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    #ifndef CO_OPTIMIZED
    static int CO_OPTIMIZED;
    #endif
    #ifndef CO_NEWLOCALS
    static int CO_NEWLOCALS;
    #endif
    #ifndef CO_VARARGS
    static int CO_VARARGS;
    #endif
    #ifndef CO_VARKEYWORDS
    static int CO_VARKEYWORDS;
    #endif
    #ifndef CO_ASYNC_GENERATOR
    static int CO_ASYNC_GENERATOR;
    #endif
    #ifndef CO_GENERATOR
    static int CO_GENERATOR;
    #endif
    #ifndef CO_COROUTINE
    static int CO_COROUTINE;
    #endif
#else
    #ifndef CO_COROUTINE
      #define CO_COROUTINE 0x80
    #endif
    #ifndef CO_ASYNC_GENERATOR
      #define CO_ASYNC_GENERATOR 0x200
    #endif
#endif
static int __Pyx_init_co_variables(void);
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef Py_TPFLAGS_IMMUTABLETYPE
  #define Py_TPFLAGS_IMMUTABLETYPE (1UL << 8)
#endif
#ifndef Py_TPFLAGS_DISALLOW_INSTANTIATION
  #define Py_TPFLAGS_DISALLOW_INSTANTIATION (1UL << 7)
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#ifndef METH_FASTCALL
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void (*cfunc)(void)) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if PY_VERSION_HEX < 0x03090000 || (CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000)
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#elif CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) GraalPyFrame_SetLineNumber((frame), (lineno))
#elif CYTHON_COMPILING_IN_GRAAL
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) _PyFrame_SetLineNumber((frame), (lineno))
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#endif
#if CYTHON_USE_MODULE_STATE
static CYTHON_INLINE void *__Pyx__PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#define __Pyx_PyModule_GetState(o) (__pyx_mstatetype *)__Pyx__PyModule_GetState(o)
#else
#define __Pyx_PyModule_GetState(op) ((void)op,__pyx_mstate_global)
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE((PyObject *) obj), name, func_ctype)
#define __Pyx_PyObject_TryGetSlot(obj, name, func_ctype) __Pyx_PyType_TryGetSlot(Py_TYPE(obj), name, func_ctype)
#define __Pyx_PyObject_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#define __Pyx_PyObject_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype) __Pyx_PyType_GetSlot(type, name, func_ctype)
  #define __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype) (((type)->sub) ? ((type)->sub->name) : NULL)
  #define __Pyx_PyType_TryGetSubSlot(type, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype)
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype)\
    ((__PYX_LIMITED_VERSION_HEX >= 0x030A0000 ||\
     (PyType_GetFlags(type) & Py_TPFLAGS_HEAPTYPE) || __Pyx_get_runtime_version() >= 0x030A0000) ?\
     __Pyx_PyType_GetSlot(type, name, func_ctype) : NULL)
  #define __Pyx_PyType_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSlot(obj, name, func_ctype)
  #define __Pyx_PyType_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSlot(obj, name, func_ctype)
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
#define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
#endif
#define __Pyx_PyObject_GetIterNextFunc(iterator)  __Pyx_PyObject_GetSlot(iterator, tp_iternext, iternextfunc)
#if CYTHON_USE_TYPE_SPECS
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#else
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains)
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && PyUnstable_Object_IsUniquelyReferenced(obj)) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#elif CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
enum __Pyx_ReferenceSharing {
  __Pyx_ReferenceSharing_DefinitelyUnique, // We created it so we know it's unshared - no need to check
  __Pyx_ReferenceSharing_OwnStrongReference,
  __Pyx_ReferenceSharing_FunctionArgument,
  __Pyx_ReferenceSharing_SharedReference, // Never trust it to be unshared because it's a global or similar
};
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && PY_VERSION_HEX >= 0x030E0000
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing)\
    (sharing == __Pyx_ReferenceSharing_DefinitelyUnique ? 1 :\
      (sharing == __Pyx_ReferenceSharing_FunctionArgument ? PyUnstable_Object_IsUniqueReferencedTemporary(o) :\
      (sharing == __Pyx_ReferenceSharing_OwnStrongReference ? PyUnstable_Object_IsUniquelyReferenced(o) : 0)))
#elif (CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING) || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)sharing), Py_REFCNT(o) == 1)
#else
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)o), ((void)sharing), 0)
#endif
#if CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyList_GetItemRef(o, i) (likely((i) >= 0) ? PySequence_GetItem(o, i) : (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
  #else
    #define __Pyx_PyList_GetItemRef(o, i) PySequence_ITEM(o, i)
  #endif
#elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #else
    #define __Pyx_PyList_GetItemRef(o, i) __Pyx_XNewRef(PyList_GetItem(o, i))
  #endif
#else
  #define __Pyx_PyList_GetItemRef(o, i) __Pyx_NewRef(PyList_GET_ITEM(o, i))
#endif
#if CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS && !CYTHON_COMPILING_IN_LIMITED_API && CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) (__Pyx_IS_UNIQUELY_REFERENCED(o, unsafe_shared) ?\
    __Pyx_NewRef(PyList_GET_ITEM(o, i)) : __Pyx_PyList_GetItemRef(o, i))
#else
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) __Pyx_PyList_GetItemRef(o, i)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyDict_GetItemRef(dict, key, result) PyDict_GetItemRef(dict, key, result)
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyObject_GetItem(dict, key);
  if (*result == NULL) {
    if (PyErr_ExceptionMatches(PyExc_KeyError)) {
      PyErr_Clear();
      return 0;
    }
    return -1;
  }
  return 1;
}
#else
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyDict_GetItemWithError(dict, key);
  if (*result == NULL) {
    return PyErr_Occurred() ? -1 : 0;
  }
  Py_INCREF(*result);
  return 1;
}
#endif
#if defined(CYTHON_DEBUG_VISIT_CONST) && CYTHON_DEBUG_VISIT_CONST
  #define __Pyx_VISIT_CONST(obj)  Py_VISIT(obj)
#else
  #define __Pyx_VISIT_CONST(obj)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GET_ITEM(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GET_ITEM(o, i)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GetItem(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GetItem(o, i)
#endif
#if CYTHON_ASSUME_SAFE_SIZE
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GET_LENGTH(o)
#else
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GetLength(o)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_InternFromString)
  #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
#endif
#define __Pyx_PyLong_FromHash_t PyLong_FromSsize_t
#define __Pyx_PyLong_AsHash_t   __Pyx_PyIndex_AsSsize_t
#if __PYX_LIMITED_VERSION_HEX >= 0x030A0000
    #define __Pyx_PySendResult PySendResult
#else
    typedef enum {
        PYGEN_RETURN = 0,
        PYGEN_ERROR = -1,
        PYGEN_NEXT = 1,
    } __Pyx_PySendResult;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030A00A3
  typedef __Pyx_PySendResult (*__Pyx_pyiter_sendfunc)(PyObject *iter, PyObject *value, PyObject **result);
#else
  #define __Pyx_pyiter_sendfunc sendfunc
#endif
#if !CYTHON_USE_AM_SEND
#define __PYX_HAS_PY_AM_SEND 0
#elif __PYX_LIMITED_VERSION_HEX >= 0x030A0000
#define __PYX_HAS_PY_AM_SEND 1
#else
#define __PYX_HAS_PY_AM_SEND 2  // our own backported implementation
#endif
#if __PYX_HAS_PY_AM_SEND < 2
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
        __Pyx_pyiter_sendfunc am_send;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_SlotTpAsAsync(s) ((PyAsyncMethods*)(s))
#endif
#if CYTHON_USE_AM_SEND && PY_VERSION_HEX < 0x030A00F0
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (1UL << 21)
#else
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (0)
#endif
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyInterpreterState_Get() PyInterpreterState_Get()
#else
#define __Pyx_PyInterpreterState_Get() PyThreadState_Get()->interp
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030A0000
#ifdef __cplusplus
extern "C"
#endif
PyAPI_FUNC(void *) PyMem_Calloc(size_t nelem, size_t elsize);
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_init_co_variable(PyObject *inspect, const char* name, int *write_to) {
    int value;
    PyObject *py_value = PyObject_GetAttrString(inspect, name);
    if (!py_value) return 0;
    value = (int) PyLong_AsLong(py_value);
    Py_DECREF(py_value);
    *write_to = value;
    return value != -1 || !PyErr_Occurred();
}
static int __Pyx_init_co_variables(void) {
    PyObject *inspect;
    int result;
    inspect = PyImport_ImportModule("inspect");
    result =
#if !defined(CO_OPTIMIZED)
        __Pyx_init_co_variable(inspect, "CO_OPTIMIZED", &CO_OPTIMIZED) &&
#endif
#if !defined(CO_NEWLOCALS)
        __Pyx_init_co_variable(inspect, "CO_NEWLOCALS", &CO_NEWLOCALS) &&
#endif
#if !defined(CO_VARARGS)
        __Pyx_init_co_variable(inspect, "CO_VARARGS", &CO_VARARGS) &&
#endif
#if !defined(CO_VARKEYWORDS)
        __Pyx_init_co_variable(inspect, "CO_VARKEYWORDS", &CO_VARKEYWORDS) &&
#endif
#if !defined(CO_ASYNC_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_ASYNC_GENERATOR", &CO_ASYNC_GENERATOR) &&
#endif
#if !defined(CO_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_GENERATOR", &CO_GENERATOR) &&
#endif
#if !defined(CO_COROUTINE)
        __Pyx_init_co_variable(inspect, "CO_COROUTINE", &CO_COROUTINE) &&
#endif
        1;
    Py_DECREF(inspect);
    return result ? 0 : -1;
}
#else
static int __Pyx_init_co_variables(void) {
    return 0;  // It's a limited API-only feature
}
#endif

/* MathInitCode */
#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #ifndef _USE_MATH_DEFINES
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#ifndef CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#define CYTHON_CLINE_IN_TRACEBACK_RUNTIME 0
#endif
#ifndef CYTHON_CLINE_IN_TRACEBACK
#define CYTHON_CLINE_IN_TRACEBACK CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#endif
#if CYTHON_CLINE_IN_TRACEBACK
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; __pyx_clineno = __LINE__; (void) __pyx_clineno; }
#else
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; (void) __pyx_clineno; }
#endif
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__pyverbs__libibverbs_enums
#define __PYX_HAVE_API__pyverbs__libibverbs_enums
/* Early includes */
#include <infiniband/verbs.h>
#include <infiniband/verbs_api.h>
#include <infiniband/driver.h>
#include <infiniband/tm_types.h>
#include <stddef.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifdef CYTHON_FREETHREADING_COMPATIBLE
#if CYTHON_FREETHREADING_COMPATIBLE
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_NOT_USED
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AS_STRING(s)
#else
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AsString(s)
#endif
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
static CYTHON_INLINE PyObject *__Pyx_NewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_NewRef)
    return Py_NewRef(obj);
#else
    Py_INCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_XNewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_XNewRef)
    return Py_XNewRef(obj);
#else
    Py_XINCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b);
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __Pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AS_DOUBLE(x)
#else
#define __Pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AsDouble(x)
#endif
#define __Pyx_PyFloat_AsFloat(x) ((float) __Pyx_PyFloat_AsDouble(x))
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#elif __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeASCII(c_str, size, NULL)
#else
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
/* PretendToInitialize */
#ifdef __cplusplus
#if __cplusplus > 201103L
#include <type_traits>
#endif
template <typename T>
static void __Pyx_pretend_to_initialize(T* ptr) {
#if __cplusplus > 201103L
    if ((std::is_trivially_default_constructible<T>::value))
#endif
        *ptr = T();
    (void)ptr;
}
#else
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }
#endif


#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * const __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char* const __pyx_f[] = {
  "pyverbs/libibverbs_enums.pyx",
  "EnumBase",
  "pyverbs/libibverbs_enums.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* Atomics.proto (used by CodeObjectCache) */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __PYX_GET_CYTHON_COMPILING_IN_CPYTHON_FREETHREADING() CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __pyx_atomic_int_type int
#define __pyx_nonatomic_int_type int
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__))
    #include <stdatomic.h>
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)))
    #include <atomic>
#endif
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__) &&\
                       ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type atomic_int
    #define __pyx_atomic_ptr_type atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) atomic_fetch_add_explicit(value, 1, memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) atomic_fetch_add_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) atomic_fetch_sub_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) atomic_load(value)
    #define __pyx_atomic_store(value, new_value) atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) atomic_load_explicit(value, memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) atomic_load_explicit(value, memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C atomics"
    #endif
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)) &&\
                    ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type std::atomic_int
    #define __pyx_atomic_ptr_type std::atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) std::atomic_fetch_sub_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) std::atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) std::atomic_load(value)
    #define __pyx_atomic_store(value, new_value) std::atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) std::atomic_load_explicit(value, std::memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) std::atomic_load_explicit(value, std::memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) std::atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C++ atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C++ atomics"
    #endif
#elif CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_ptr_type void*
    #define __pyx_nonatomic_ptr_type void*
    #define __pyx_atomic_incr_relaxed(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) __sync_fetch_and_sub(value, 1)
    #define __pyx_atomic_sub(value, arg) __sync_fetch_and_sub(value, arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_store(value, new_value) __sync_lock_test_and_set(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_load_acquire(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) __sync_lock_test_and_set(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_nonatomic_ptr_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER)
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #define __pyx_atomic_ptr_type void*
    #undef __pyx_nonatomic_int_type
    #define __pyx_nonatomic_int_type long
    #define __pyx_nonatomic_ptr_type void*
    #pragma intrinsic (_InterlockedExchangeAdd, _InterlockedExchange, _InterlockedCompareExchange, _InterlockedCompareExchangePointer, _InterlockedExchangePointer)
    #define __pyx_atomic_incr_relaxed(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) _InterlockedExchangeAdd(value, -1)
    #define __pyx_atomic_sub(value, arg) _InterlockedExchangeAdd(value, -arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = _InterlockedCompareExchange(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) _InterlockedExchangeAdd(value, 0)
    #define __pyx_atomic_store(value, new_value) _InterlockedExchange(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) *(void * volatile *)value
    #define __pyx_atomic_pointer_load_acquire(value) _InterlockedCompareExchangePointer(value, 0, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) _InterlockedExchangePointer(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_atomic_ptr_type old = _InterlockedCompareExchangePointer(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif

/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto (used by PyErrFetchRestore) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto (used by HasAttr) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* HasAttr.proto (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_HasAttr(o, n)  PyObject_HasAttrWithError(o, n)
#else
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);
#endif

/* PyObjectCall.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto (used by PyObjectCallOneArg) */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs);

/* PyObjectCallOneArg.proto (used by ImportImpl) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* TupleAndListFromArray.proto (used by ImportImpl) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
#endif
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto (used by ImportImpl) */
#include <string.h>

/* ImportImpl.export */
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level);

/* Import.proto */
static CYTHON_INLINE PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* ListPack.proto */
static PyObject *__Pyx_PyList_Pack(Py_ssize_t n, ...);

/* GetAttr.proto (used by Globals) */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* Globals.proto */
static PyObject* __Pyx_Globals(void);

/* PyObjectVectorCallKwBuilder.proto */
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#if CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_Vectorcall_CallFromBuilder PyObject_Vectorcall
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder _PyObject_Vectorcall
#endif
#define __Pyx_MakeVectorcallBuilderKwds(n) PyTuple_New(n)
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder __Pyx_PyObject_FastCallDict
#define __Pyx_MakeVectorcallBuilderKwds(n) __Pyx_PyDict_NewPresized(n)
#define __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n) PyDict_SetItem(builder, key, value)
#define __Pyx_VectorcallBuilder_AddArgStr(key, value, builder, args, n) PyDict_SetItemString(builder, key, value)
#endif

/* GetBuiltinName.proto (used by GetModuleGlobalName) */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyDictVersioning.proto (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __Pyx_XNewRef(__pyx_dict_cached_value);\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_mstate_global->__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectDelAttr.proto (used by PyObjectSetAttrStr) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
#define __Pyx_PyObject_DelAttr(o, n) PyObject_SetAttr(o, n, NULL)
#else
#define __Pyx_PyObject_DelAttr(o, n) PyObject_DelAttr(o, n)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   __Pyx_PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* dict_setdefault.proto (used by CLineInTraceback) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value);

/* CLineInTraceback.proto (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#else
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#endif

/* CodeObjectCache.proto (used by AddTraceback) */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject __Pyx_CachedCodeObjectType;
#else
typedef PyCodeObject __Pyx_CachedCodeObjectType;
#endif
typedef struct {
    __Pyx_CachedCodeObjectType* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_int_type accessor_count;
  #endif
};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_PY_LONG_LONG(unsigned PY_LONG_LONG value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_access_flags(enum ibv_access_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_atomic_cap(enum ibv_atomic_cap value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_counter_description(enum ibv_counter_description value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_cq_init_attr_mask(enum ibv_cq_init_attr_mask value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_create_cq_attr_flags(enum ibv_create_cq_attr_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(enum ibv_create_cq_wc_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_device_cap_flags(enum ibv_device_cap_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_dmah_init_attr_mask(enum ibv_dmah_init_attr_mask value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_event_type(enum ibv_event_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_flow_attr_type(enum ibv_flow_attr_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_flow_flags(enum ibv_flow_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_flow_spec_type(enum ibv_flow_spec_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_fork_status(enum ibv_fork_status value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_gid_type(enum ibv_gid_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_gid_type_sysfs(enum ibv_gid_type_sysfs value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_mig_state(enum ibv_mig_state value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(enum ibv_mr_init_attr_mask value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_mtu(enum ibv_mtu value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_mw_type(enum ibv_mw_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_node_type(enum ibv_node_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_odp_general_caps(enum ibv_odp_general_caps value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(enum ibv_odp_transport_cap_bits value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_ops_flags(enum ibv_ops_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_ops_wr_opcode(enum ibv_ops_wr_opcode value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_parent_domain_init_attr_mask(enum ibv_parent_domain_init_attr_mask value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_placement_type(enum ibv_placement_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_port_cap_flags(enum ibv_port_cap_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_port_cap_flags2(enum ibv_port_cap_flags2 value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_port_state(enum ibv_port_state value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_qp_attr_mask(enum ibv_qp_attr_mask value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_qp_create_flags(enum ibv_qp_create_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(enum ibv_qp_create_send_ops_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(enum ibv_qp_init_attr_mask value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_qp_state(enum ibv_qp_state value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_qp_type(enum ibv_qp_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_query_qp_data_in_order_caps(enum ibv_query_qp_data_in_order_caps value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_query_qp_data_in_order_flags(enum ibv_query_qp_data_in_order_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_raw_packet_caps(enum ibv_raw_packet_caps value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_read_counters_flags(enum ibv_read_counters_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(enum ibv_rereg_mr_err_code value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_rereg_mr_flags(enum ibv_rereg_mr_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_rx_hash_fields(enum ibv_rx_hash_fields value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_rx_hash_function_flags(enum ibv_rx_hash_function_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_selectivity_level(enum ibv_selectivity_level value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_send_flags(enum ibv_send_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_srq_attr_mask(enum ibv_srq_attr_mask value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(enum ibv_srq_init_attr_mask value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_srq_type(enum ibv_srq_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_tm_cap_flags(enum ibv_tm_cap_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_tmh_op(enum ibv_tmh_op value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_tph_mem_type(enum ibv_tph_mem_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_transport_type(enum ibv_transport_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_values_mask(enum ibv_values_mask value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_wc_flags(enum ibv_wc_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_wc_opcode(enum ibv_wc_opcode value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_wc_status(enum ibv_wc_status value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_wq_attr_mask(enum ibv_wq_attr_mask value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_wq_flags(enum ibv_wq_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_wq_init_attr_mask(enum ibv_wq_init_attr_mask value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_wq_state(enum ibv_wq_state value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_wq_type(enum ibv_wq_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_wr_opcode(enum ibv_wr_opcode value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_xrcd_init_attr_mask(enum ibv_xrcd_init_attr_mask value);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#else
static __Pyx_TypeName __Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp);
#endif
#else  // !LIMITED_API
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetFullyQualifiedName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2) {
    return PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2);
}
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#ifdef PyExceptionInstance_Check
  #define __Pyx_PyBaseException_Check(obj) PyExceptionInstance_Check(obj)
#else
  #define __Pyx_PyBaseException_Check(obj) __Pyx_TypeCheck(obj, PyExc_BaseException)
#endif

/* GetRuntimeVersion.proto */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
static unsigned long __Pyx_cached_runtime_version = 0;
static void __Pyx_init_runtime_version(void);
#else
#define __Pyx_init_runtime_version()
#endif
static unsigned long __Pyx_get_runtime_version(void);

/* AddModuleRef.proto */
#if ((CYTHON_COMPILING_IN_CPYTHON_FREETHREADING ) ||\
     __PYX_LIMITED_VERSION_HEX < 0x030d0000)
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name);
#else
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#endif

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From___pyx_anon_enum(int value);

/* DecompressString.proto */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo);

/* MultiPhaseInitModuleState.proto */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
static PyObject *__Pyx_State_FindModule(void*);
static int __Pyx_State_AddModule(PyObject* module, void*);
static int __Pyx_State_RemoveModule(void*);
#elif CYTHON_USE_MODULE_STATE
#define __Pyx_State_FindModule PyState_FindModule
#define __Pyx_State_AddModule PyState_AddModule
#define __Pyx_State_RemoveModule PyState_RemoveModule
#endif

/* #### Code section: module_declarations ### */
/* CythonABIVersion.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
    #if CYTHON_METH_FASTCALL
        #define __PYX_FASTCALL_ABI_SUFFIX  "_fastcall"
    #else
        #define __PYX_FASTCALL_ABI_SUFFIX
    #endif
    #define __PYX_LIMITED_ABI_SUFFIX "limited" __PYX_FASTCALL_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#else
    #define __PYX_LIMITED_ABI_SUFFIX
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    #define __PYX_AM_SEND_ABI_SUFFIX
#elif __PYX_HAS_PY_AM_SEND == 2
    #define __PYX_AM_SEND_ABI_SUFFIX "amsendbackport"
#else
    #define __PYX_AM_SEND_ABI_SUFFIX "noamsend"
#endif
#ifndef __PYX_MONITORING_ABI_SUFFIX
    #define __PYX_MONITORING_ABI_SUFFIX
#endif
#if CYTHON_USE_TP_FINALIZE
    #define __PYX_TP_FINALIZE_ABI_SUFFIX
#else
    #define __PYX_TP_FINALIZE_ABI_SUFFIX "nofinalize"
#endif
#if CYTHON_USE_FREELISTS || !defined(__Pyx_AsyncGen_USED)
    #define __PYX_FREELISTS_ABI_SUFFIX
#else
    #define __PYX_FREELISTS_ABI_SUFFIX "nofreelists"
#endif
#define CYTHON_ABI  __PYX_ABI_VERSION __PYX_LIMITED_ABI_SUFFIX __PYX_MONITORING_ABI_SUFFIX __PYX_TP_FINALIZE_ABI_SUFFIX __PYX_FREELISTS_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."


/* Module declarations from "libc.stddef" */

/* Module declarations from "pyverbs.libibverbs_enums" */
static PyObject *__Pyx_EnumBase = 0;
static PyObject *__Pyx_FlagBase = 0;
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "pyverbs.libibverbs_enums"
extern int __pyx_module_is_main_pyverbs__libibverbs_enums;
int __pyx_module_is_main_pyverbs__libibverbs_enums = 0;

/* Implementation of "pyverbs.libibverbs_enums" */
/* #### Code section: global_var ### */
/* #### Code section: string_decls ### */
/* #### Code section: decls ### */
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
/* SmallCodeConfig */
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  PyObject *__pyx_tuple[1];
  PyObject *__pyx_string_tab[470];
/* #### Code section: module_state_contents ### */
/* CodeObjectCache.module_state_decls */
struct __Pyx_CodeObjectCache __pyx_code_cache;

/* #### Code section: module_state_end ### */
} __pyx_mstatetype;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate_global (__Pyx_PyModule_GetState(__Pyx_State_FindModule(&__pyx_moduledef)))

#define __pyx_m (__Pyx_State_FindModule(&__pyx_moduledef))
#else
static __pyx_mstatetype __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstatetype * const __pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: constant_name_defines ### */
#define __pyx_kp_u_ __pyx_string_tab[0]
#define __pyx_kp_u__2 __pyx_string_tab[1]
#define __pyx_kp_u_pyverbs_libibverbs_enums __pyx_string_tab[2]
#define __pyx_n_u_IBV_ACCESS_FLUSH_GLOBAL __pyx_string_tab[3]
#define __pyx_n_u_IBV_ACCESS_FLUSH_PERSISTENT __pyx_string_tab[4]
#define __pyx_n_u_IBV_ACCESS_HUGETLB __pyx_string_tab[5]
#define __pyx_n_u_IBV_ACCESS_LOCAL_WRITE __pyx_string_tab[6]
#define __pyx_n_u_IBV_ACCESS_MW_BIND __pyx_string_tab[7]
#define __pyx_n_u_IBV_ACCESS_ON_DEMAND __pyx_string_tab[8]
#define __pyx_n_u_IBV_ACCESS_RELAXED_ORDERING __pyx_string_tab[9]
#define __pyx_n_u_IBV_ACCESS_REMOTE_ATOMIC __pyx_string_tab[10]
#define __pyx_n_u_IBV_ACCESS_REMOTE_READ __pyx_string_tab[11]
#define __pyx_n_u_IBV_ACCESS_REMOTE_WRITE __pyx_string_tab[12]
#define __pyx_n_u_IBV_ACCESS_ZERO_BASED __pyx_string_tab[13]
#define __pyx_n_u_IBV_ADVISE_MR_ADVICE_PREFETCH __pyx_string_tab[14]
#define __pyx_n_u_IBV_ADVISE_MR_ADVICE_PREFETCH_N __pyx_string_tab[15]
#define __pyx_n_u_IBV_ADVISE_MR_ADVICE_PREFETCH_W __pyx_string_tab[16]
#define __pyx_n_u_IBV_ADVISE_MR_FLAG_FLUSH __pyx_string_tab[17]
#define __pyx_n_u_IBV_ALLOCATOR_USE_DEFAULT __pyx_string_tab[18]
#define __pyx_n_u_IBV_ATOMIC_GLOB __pyx_string_tab[19]
#define __pyx_n_u_IBV_ATOMIC_HCA __pyx_string_tab[20]
#define __pyx_n_u_IBV_ATOMIC_NONE __pyx_string_tab[21]
#define __pyx_n_u_IBV_COUNTER_BYTES __pyx_string_tab[22]
#define __pyx_n_u_IBV_COUNTER_PACKETS __pyx_string_tab[23]
#define __pyx_n_u_IBV_CQ_INIT_ATTR_MASK_FLAGS __pyx_string_tab[24]
#define __pyx_n_u_IBV_CQ_INIT_ATTR_MASK_PD __pyx_string_tab[25]
#define __pyx_n_u_IBV_CREATE_CQ_ATTR_IGNORE_OVERRU __pyx_string_tab[26]
#define __pyx_n_u_IBV_CREATE_CQ_ATTR_SINGLE_THREAD __pyx_string_tab[27]
#define __pyx_n_u_IBV_DEVICE_AUTO_PATH_MIG __pyx_string_tab[28]
#define __pyx_n_u_IBV_DEVICE_BAD_PKEY_CNTR __pyx_string_tab[29]
#define __pyx_n_u_IBV_DEVICE_BAD_QKEY_CNTR __pyx_string_tab[30]
#define __pyx_n_u_IBV_DEVICE_CHANGE_PHY_PORT __pyx_string_tab[31]
#define __pyx_n_u_IBV_DEVICE_CURR_QP_STATE_MOD __pyx_string_tab[32]
#define __pyx_n_u_IBV_DEVICE_INIT_TYPE __pyx_string_tab[33]
#define __pyx_n_u_IBV_DEVICE_MANAGED_FLOW_STEERING __pyx_string_tab[34]
#define __pyx_n_u_IBV_DEVICE_MEM_MGT_EXTENSIONS __pyx_string_tab[35]
#define __pyx_n_u_IBV_DEVICE_MEM_WINDOW __pyx_string_tab[36]
#define __pyx_n_u_IBV_DEVICE_MEM_WINDOW_TYPE_2A __pyx_string_tab[37]
#define __pyx_n_u_IBV_DEVICE_MEM_WINDOW_TYPE_2B __pyx_string_tab[38]
#define __pyx_n_u_IBV_DEVICE_N_NOTIFY_CQ __pyx_string_tab[39]
#define __pyx_n_u_IBV_DEVICE_PCI_WRITE_END_PADDIN __pyx_string_tab[40]
#define __pyx_n_u_IBV_DEVICE_PORT_ACTIVE_EVENT __pyx_string_tab[41]
#define __pyx_n_u_IBV_DEVICE_RAW_IP_CSUM __pyx_string_tab[42]
#define __pyx_n_u_IBV_DEVICE_RAW_MULTI __pyx_string_tab[43]
#define __pyx_n_u_IBV_DEVICE_RAW_SCATTER_FCS __pyx_string_tab[44]
#define __pyx_n_u_IBV_DEVICE_RC_IP_CSUM __pyx_string_tab[45]
#define __pyx_n_u_IBV_DEVICE_RC_RNR_NAK_GEN __pyx_string_tab[46]
#define __pyx_n_u_IBV_DEVICE_RESIZE_MAX_WR __pyx_string_tab[47]
#define __pyx_n_u_IBV_DEVICE_SHUTDOWN_PORT __pyx_string_tab[48]
#define __pyx_n_u_IBV_DEVICE_SRQ_RESIZE __pyx_string_tab[49]
#define __pyx_n_u_IBV_DEVICE_SYS_IMAGE_GUID __pyx_string_tab[50]
#define __pyx_n_u_IBV_DEVICE_UD_AV_PORT_ENFORCE __pyx_string_tab[51]
#define __pyx_n_u_IBV_DEVICE_UD_IP_CSUM __pyx_string_tab[52]
#define __pyx_n_u_IBV_DEVICE_XRC __pyx_string_tab[53]
#define __pyx_n_u_IBV_DMAH_INIT_ATTR_MASK_CPU_ID __pyx_string_tab[54]
#define __pyx_n_u_IBV_DMAH_INIT_ATTR_MASK_PH __pyx_string_tab[55]
#define __pyx_n_u_IBV_DMAH_INIT_ATTR_MASK_TPH_MEM __pyx_string_tab[56]
#define __pyx_n_u_IBV_EVENT_CLIENT_REREGISTER __pyx_string_tab[57]
#define __pyx_n_u_IBV_EVENT_COMM_EST __pyx_string_tab[58]
#define __pyx_n_u_IBV_EVENT_CQ_ERR __pyx_string_tab[59]
#define __pyx_n_u_IBV_EVENT_DEVICE_FATAL __pyx_string_tab[60]
#define __pyx_n_u_IBV_EVENT_GID_CHANGE __pyx_string_tab[61]
#define __pyx_n_u_IBV_EVENT_LID_CHANGE __pyx_string_tab[62]
#define __pyx_n_u_IBV_EVENT_PATH_MIG __pyx_string_tab[63]
#define __pyx_n_u_IBV_EVENT_PATH_MIG_ERR __pyx_string_tab[64]
#define __pyx_n_u_IBV_EVENT_PKEY_CHANGE __pyx_string_tab[65]
#define __pyx_n_u_IBV_EVENT_PORT_ACTIVE __pyx_string_tab[66]
#define __pyx_n_u_IBV_EVENT_PORT_ERR __pyx_string_tab[67]
#define __pyx_n_u_IBV_EVENT_QP_ACCESS_ERR __pyx_string_tab[68]
#define __pyx_n_u_IBV_EVENT_QP_FATAL __pyx_string_tab[69]
#define __pyx_n_u_IBV_EVENT_QP_LAST_WQE_REACHED __pyx_string_tab[70]
#define __pyx_n_u_IBV_EVENT_QP_REQ_ERR __pyx_string_tab[71]
#define __pyx_n_u_IBV_EVENT_SM_CHANGE __pyx_string_tab[72]
#define __pyx_n_u_IBV_EVENT_SQ_DRAINED __pyx_string_tab[73]
#define __pyx_n_u_IBV_EVENT_SRQ_ERR __pyx_string_tab[74]
#define __pyx_n_u_IBV_EVENT_SRQ_LIMIT_REACHED __pyx_string_tab[75]
#define __pyx_n_u_IBV_EVENT_WQ_FATAL __pyx_string_tab[76]
#define __pyx_n_u_IBV_FLOW_ATTR_ALL_DEFAULT __pyx_string_tab[77]
#define __pyx_n_u_IBV_FLOW_ATTR_FLAGS_DONT_TRAP __pyx_string_tab[78]
#define __pyx_n_u_IBV_FLOW_ATTR_FLAGS_EGRESS __pyx_string_tab[79]
#define __pyx_n_u_IBV_FLOW_ATTR_MC_DEFAULT __pyx_string_tab[80]
#define __pyx_n_u_IBV_FLOW_ATTR_NORMAL __pyx_string_tab[81]
#define __pyx_n_u_IBV_FLOW_ATTR_SNIFFER __pyx_string_tab[82]
#define __pyx_n_u_IBV_FLOW_SPEC_ACTION_COUNT __pyx_string_tab[83]
#define __pyx_n_u_IBV_FLOW_SPEC_ACTION_DROP __pyx_string_tab[84]
#define __pyx_n_u_IBV_FLOW_SPEC_ACTION_HANDLE __pyx_string_tab[85]
#define __pyx_n_u_IBV_FLOW_SPEC_ACTION_TAG __pyx_string_tab[86]
#define __pyx_n_u_IBV_FLOW_SPEC_ESP __pyx_string_tab[87]
#define __pyx_n_u_IBV_FLOW_SPEC_ETH __pyx_string_tab[88]
#define __pyx_n_u_IBV_FLOW_SPEC_GRE __pyx_string_tab[89]
#define __pyx_n_u_IBV_FLOW_SPEC_INNER __pyx_string_tab[90]
#define __pyx_n_u_IBV_FLOW_SPEC_IPV4 __pyx_string_tab[91]
#define __pyx_n_u_IBV_FLOW_SPEC_IPV4_EXT __pyx_string_tab[92]
#define __pyx_n_u_IBV_FLOW_SPEC_IPV6 __pyx_string_tab[93]
#define __pyx_n_u_IBV_FLOW_SPEC_MPLS __pyx_string_tab[94]
#define __pyx_n_u_IBV_FLOW_SPEC_TCP __pyx_string_tab[95]
#define __pyx_n_u_IBV_FLOW_SPEC_UDP __pyx_string_tab[96]
#define __pyx_n_u_IBV_FLOW_SPEC_VXLAN_TUNNEL __pyx_string_tab[97]
#define __pyx_n_u_IBV_FLUSH_GLOBAL __pyx_string_tab[98]
#define __pyx_n_u_IBV_FLUSH_MR __pyx_string_tab[99]
#define __pyx_n_u_IBV_FLUSH_PERSISTENT __pyx_string_tab[100]
#define __pyx_n_u_IBV_FLUSH_RANGE __pyx_string_tab[101]
#define __pyx_n_u_IBV_FORK_DISABLED __pyx_string_tab[102]
#define __pyx_n_u_IBV_FORK_ENABLED __pyx_string_tab[103]
#define __pyx_n_u_IBV_FORK_UNNEEDED __pyx_string_tab[104]
#define __pyx_n_u_IBV_GID_TYPE_IB __pyx_string_tab[105]
#define __pyx_n_u_IBV_GID_TYPE_ROCE_V1 __pyx_string_tab[106]
#define __pyx_n_u_IBV_GID_TYPE_ROCE_V2 __pyx_string_tab[107]
#define __pyx_n_u_IBV_GID_TYPE_SYSFS_IB_ROCE_V1 __pyx_string_tab[108]
#define __pyx_n_u_IBV_GID_TYPE_SYSFS_ROCE_V2 __pyx_string_tab[109]
#define __pyx_n_u_IBV_MIG_ARMED __pyx_string_tab[110]
#define __pyx_n_u_IBV_MIG_MIGRATED __pyx_string_tab[111]
#define __pyx_n_u_IBV_MIG_REARM __pyx_string_tab[112]
#define __pyx_n_u_IBV_MTU_1024 __pyx_string_tab[113]
#define __pyx_n_u_IBV_MTU_2048 __pyx_string_tab[114]
#define __pyx_n_u_IBV_MTU_256 __pyx_string_tab[115]
#define __pyx_n_u_IBV_MTU_4096 __pyx_string_tab[116]
#define __pyx_n_u_IBV_MTU_512 __pyx_string_tab[117]
#define __pyx_n_u_IBV_MW_TYPE_1 __pyx_string_tab[118]
#define __pyx_n_u_IBV_MW_TYPE_2 __pyx_string_tab[119]
#define __pyx_n_u_IBV_NODE_CA __pyx_string_tab[120]
#define __pyx_n_u_IBV_NODE_RNIC __pyx_string_tab[121]
#define __pyx_n_u_IBV_NODE_ROUTER __pyx_string_tab[122]
#define __pyx_n_u_IBV_NODE_SWITCH __pyx_string_tab[123]
#define __pyx_n_u_IBV_NODE_UNKNOWN __pyx_string_tab[124]
#define __pyx_n_u_IBV_NODE_UNSPECIFIED __pyx_string_tab[125]
#define __pyx_n_u_IBV_NODE_USNIC __pyx_string_tab[126]
#define __pyx_n_u_IBV_NODE_USNIC_UDP __pyx_string_tab[127]
#define __pyx_n_u_IBV_ODP_SUPPORT __pyx_string_tab[128]
#define __pyx_n_u_IBV_ODP_SUPPORT_ATOMIC __pyx_string_tab[129]
#define __pyx_n_u_IBV_ODP_SUPPORT_ATOMIC_WRITE __pyx_string_tab[130]
#define __pyx_n_u_IBV_ODP_SUPPORT_FLUSH __pyx_string_tab[131]
#define __pyx_n_u_IBV_ODP_SUPPORT_IMPLICIT __pyx_string_tab[132]
#define __pyx_n_u_IBV_ODP_SUPPORT_READ __pyx_string_tab[133]
#define __pyx_n_u_IBV_ODP_SUPPORT_RECV __pyx_string_tab[134]
#define __pyx_n_u_IBV_ODP_SUPPORT_SEND __pyx_string_tab[135]
#define __pyx_n_u_IBV_ODP_SUPPORT_SRQ_RECV __pyx_string_tab[136]
#define __pyx_n_u_IBV_ODP_SUPPORT_WRITE __pyx_string_tab[137]
#define __pyx_n_u_IBV_OPS_SIGNALED __pyx_string_tab[138]
#define __pyx_n_u_IBV_OPS_TM_SYNC __pyx_string_tab[139]
#define __pyx_n_u_IBV_PARENT_DOMAIN_INIT_ATTR_ALLO __pyx_string_tab[140]
#define __pyx_n_u_IBV_PARENT_DOMAIN_INIT_ATTR_PD_C __pyx_string_tab[141]
#define __pyx_n_u_IBV_PORT_ACTIVE __pyx_string_tab[142]
#define __pyx_n_u_IBV_PORT_ACTIVE_DEFER __pyx_string_tab[143]
#define __pyx_n_u_IBV_PORT_ARMED __pyx_string_tab[144]
#define __pyx_n_u_IBV_PORT_AUTO_MIGR_SUP __pyx_string_tab[145]
#define __pyx_n_u_IBV_PORT_BOOT_MGMT_SUP __pyx_string_tab[146]
#define __pyx_n_u_IBV_PORT_CAP_MASK2_SUP __pyx_string_tab[147]
#define __pyx_n_u_IBV_PORT_CAP_MASK_NOTICE_SUP __pyx_string_tab[148]
#define __pyx_n_u_IBV_PORT_CLIENT_REG_SUP __pyx_string_tab[149]
#define __pyx_n_u_IBV_PORT_CM_SUP __pyx_string_tab[150]
#define __pyx_n_u_IBV_PORT_DEVICE_MGMT_SUP __pyx_string_tab[151]
#define __pyx_n_u_IBV_PORT_DOWN __pyx_string_tab[152]
#define __pyx_n_u_IBV_PORT_DR_NOTICE_SUP __pyx_string_tab[153]
#define __pyx_n_u_IBV_PORT_EXTENDED_SPEEDS_SUP __pyx_string_tab[154]
#define __pyx_n_u_IBV_PORT_INFO_EXT_SUP __pyx_string_tab[155]
#define __pyx_n_u_IBV_PORT_INIT __pyx_string_tab[156]
#define __pyx_n_u_IBV_PORT_IP_BASED_GIDS __pyx_string_tab[157]
#define __pyx_n_u_IBV_PORT_LED_INFO_SUP __pyx_string_tab[158]
#define __pyx_n_u_IBV_PORT_LINK_LATENCY_SUP __pyx_string_tab[159]
#define __pyx_n_u_IBV_PORT_LINK_SPEED_HDR_SUP __pyx_string_tab[160]
#define __pyx_n_u_IBV_PORT_LINK_SPEED_NDR_SUP __pyx_string_tab[161]
#define __pyx_n_u_IBV_PORT_LINK_WIDTH_2X_SUP __pyx_string_tab[162]
#define __pyx_n_u_IBV_PORT_MKEY_NVRAM __pyx_string_tab[163]
#define __pyx_n_u_IBV_PORT_NOP __pyx_string_tab[164]
#define __pyx_n_u_IBV_PORT_NOTICE_SUP __pyx_string_tab[165]
#define __pyx_n_u_IBV_PORT_OPT_IPD_SUP __pyx_string_tab[166]
#define __pyx_n_u_IBV_PORT_PKEY_NVRAM __pyx_string_tab[167]
#define __pyx_n_u_IBV_PORT_PKEY_SW_EXT_PORT_TRAP_S __pyx_string_tab[168]
#define __pyx_n_u_IBV_PORT_REINIT_SUP __pyx_string_tab[169]
#define __pyx_n_u_IBV_PORT_SET_NODE_DESC_SUP __pyx_string_tab[170]
#define __pyx_n_u_IBV_PORT_SL_MAP_SUP __pyx_string_tab[171]
#define __pyx_n_u_IBV_PORT_SM __pyx_string_tab[172]
#define __pyx_n_u_IBV_PORT_SNMP_TUNNEL_SUP __pyx_string_tab[173]
#define __pyx_n_u_IBV_PORT_SWITCH_PORT_STATE_TABLE __pyx_string_tab[174]
#define __pyx_n_u_IBV_PORT_SYS_IMAGE_GUID_SUP __pyx_string_tab[175]
#define __pyx_n_u_IBV_PORT_TRAP_SUP __pyx_string_tab[176]
#define __pyx_n_u_IBV_PORT_VENDOR_CLASS_SUP __pyx_string_tab[177]
#define __pyx_n_u_IBV_PORT_VIRT_SUP __pyx_string_tab[178]
#define __pyx_n_u_IBV_QPS_ERR __pyx_string_tab[179]
#define __pyx_n_u_IBV_QPS_INIT __pyx_string_tab[180]
#define __pyx_n_u_IBV_QPS_RESET __pyx_string_tab[181]
#define __pyx_n_u_IBV_QPS_RTR __pyx_string_tab[182]
#define __pyx_n_u_IBV_QPS_RTS __pyx_string_tab[183]
#define __pyx_n_u_IBV_QPS_SQD __pyx_string_tab[184]
#define __pyx_n_u_IBV_QPS_SQE __pyx_string_tab[185]
#define __pyx_n_u_IBV_QPS_UNKNOWN __pyx_string_tab[186]
#define __pyx_n_u_IBV_QPT_DRIVER __pyx_string_tab[187]
#define __pyx_n_u_IBV_QPT_RAW_PACKET __pyx_string_tab[188]
#define __pyx_n_u_IBV_QPT_RC __pyx_string_tab[189]
#define __pyx_n_u_IBV_QPT_UC __pyx_string_tab[190]
#define __pyx_n_u_IBV_QPT_UD __pyx_string_tab[191]
#define __pyx_n_u_IBV_QPT_XRC_RECV __pyx_string_tab[192]
#define __pyx_n_u_IBV_QPT_XRC_SEND __pyx_string_tab[193]
#define __pyx_n_u_IBV_QP_ACCESS_FLAGS __pyx_string_tab[194]
#define __pyx_n_u_IBV_QP_ALT_PATH __pyx_string_tab[195]
#define __pyx_n_u_IBV_QP_AV __pyx_string_tab[196]
#define __pyx_n_u_IBV_QP_CAP __pyx_string_tab[197]
#define __pyx_n_u_IBV_QP_CREATE_BLOCK_SELF_MCAST_L __pyx_string_tab[198]
#define __pyx_n_u_IBV_QP_CREATE_CVLAN_STRIPPING __pyx_string_tab[199]
#define __pyx_n_u_IBV_QP_CREATE_PCI_WRITE_END_PADD __pyx_string_tab[200]
#define __pyx_n_u_IBV_QP_CREATE_SCATTER_FCS __pyx_string_tab[201]
#define __pyx_n_u_IBV_QP_CREATE_SOURCE_QPN __pyx_string_tab[202]
#define __pyx_n_u_IBV_QP_CUR_STATE __pyx_string_tab[203]
#define __pyx_n_u_IBV_QP_DEST_QPN __pyx_string_tab[204]
#define __pyx_n_u_IBV_QP_EN_SQD_ASYNC_NOTIFY __pyx_string_tab[205]
#define __pyx_n_u_IBV_QP_EX_WITH_ATOMIC_CMP_AND_SW __pyx_string_tab[206]
#define __pyx_n_u_IBV_QP_EX_WITH_ATOMIC_FETCH_AND __pyx_string_tab[207]
#define __pyx_n_u_IBV_QP_EX_WITH_ATOMIC_WRITE __pyx_string_tab[208]
#define __pyx_n_u_IBV_QP_EX_WITH_BIND_MW __pyx_string_tab[209]
#define __pyx_n_u_IBV_QP_EX_WITH_FLUSH __pyx_string_tab[210]
#define __pyx_n_u_IBV_QP_EX_WITH_LOCAL_INV __pyx_string_tab[211]
#define __pyx_n_u_IBV_QP_EX_WITH_RDMA_READ __pyx_string_tab[212]
#define __pyx_n_u_IBV_QP_EX_WITH_RDMA_WRITE __pyx_string_tab[213]
#define __pyx_n_u_IBV_QP_EX_WITH_RDMA_WRITE_WITH_I __pyx_string_tab[214]
#define __pyx_n_u_IBV_QP_EX_WITH_SEND __pyx_string_tab[215]
#define __pyx_n_u_IBV_QP_EX_WITH_SEND_WITH_IMM __pyx_string_tab[216]
#define __pyx_n_u_IBV_QP_EX_WITH_SEND_WITH_INV __pyx_string_tab[217]
#define __pyx_n_u_IBV_QP_EX_WITH_TSO __pyx_string_tab[218]
#define __pyx_n_u_IBV_QP_INIT_ATTR_CREATE_FLAGS __pyx_string_tab[219]
#define __pyx_n_u_IBV_QP_INIT_ATTR_IND_TABLE __pyx_string_tab[220]
#define __pyx_n_u_IBV_QP_INIT_ATTR_MAX_TSO_HEADER __pyx_string_tab[221]
#define __pyx_n_u_IBV_QP_INIT_ATTR_PD __pyx_string_tab[222]
#define __pyx_n_u_IBV_QP_INIT_ATTR_RX_HASH __pyx_string_tab[223]
#define __pyx_n_u_IBV_QP_INIT_ATTR_SEND_OPS_FLAGS __pyx_string_tab[224]
#define __pyx_n_u_IBV_QP_INIT_ATTR_XRCD __pyx_string_tab[225]
#define __pyx_n_u_IBV_QP_MAX_DEST_RD_ATOMIC __pyx_string_tab[226]
#define __pyx_n_u_IBV_QP_MAX_QP_RD_ATOMIC __pyx_string_tab[227]
#define __pyx_n_u_IBV_QP_MIN_RNR_TIMER __pyx_string_tab[228]
#define __pyx_n_u_IBV_QP_PATH_MIG_STATE __pyx_string_tab[229]
#define __pyx_n_u_IBV_QP_PATH_MTU __pyx_string_tab[230]
#define __pyx_n_u_IBV_QP_PKEY_INDEX __pyx_string_tab[231]
#define __pyx_n_u_IBV_QP_PORT __pyx_string_tab[232]
#define __pyx_n_u_IBV_QP_QKEY __pyx_string_tab[233]
#define __pyx_n_u_IBV_QP_RATE_LIMIT __pyx_string_tab[234]
#define __pyx_n_u_IBV_QP_RETRY_CNT __pyx_string_tab[235]
#define __pyx_n_u_IBV_QP_RNR_RETRY __pyx_string_tab[236]
#define __pyx_n_u_IBV_QP_RQ_PSN __pyx_string_tab[237]
#define __pyx_n_u_IBV_QP_SQ_PSN __pyx_string_tab[238]
#define __pyx_n_u_IBV_QP_STATE __pyx_string_tab[239]
#define __pyx_n_u_IBV_QP_TIMEOUT __pyx_string_tab[240]
#define __pyx_n_u_IBV_QUERY_QP_DATA_IN_ORDER_ALIGN __pyx_string_tab[241]
#define __pyx_n_u_IBV_QUERY_QP_DATA_IN_ORDER_RETUR __pyx_string_tab[242]
#define __pyx_n_u_IBV_QUERY_QP_DATA_IN_ORDER_WHOLE __pyx_string_tab[243]
#define __pyx_n_u_IBV_RAW_PACKET_CAP_CVLAN_STRIPPI __pyx_string_tab[244]
#define __pyx_n_u_IBV_RAW_PACKET_CAP_DELAY_DROP __pyx_string_tab[245]
#define __pyx_n_u_IBV_RAW_PACKET_CAP_IP_CSUM __pyx_string_tab[246]
#define __pyx_n_u_IBV_RAW_PACKET_CAP_SCATTER_FCS __pyx_string_tab[247]
#define __pyx_n_u_IBV_READ_COUNTERS_ATTR_PREFER_CA __pyx_string_tab[248]
#define __pyx_n_u_IBV_REG_MR_MASK_ADDR __pyx_string_tab[249]
#define __pyx_n_u_IBV_REG_MR_MASK_DMAH __pyx_string_tab[250]
#define __pyx_n_u_IBV_REG_MR_MASK_FD __pyx_string_tab[251]
#define __pyx_n_u_IBV_REG_MR_MASK_FD_OFFSET __pyx_string_tab[252]
#define __pyx_n_u_IBV_REG_MR_MASK_IOVA __pyx_string_tab[253]
#define __pyx_n_u_IBV_REREG_MR_CHANGE_ACCESS __pyx_string_tab[254]
#define __pyx_n_u_IBV_REREG_MR_CHANGE_PD __pyx_string_tab[255]
#define __pyx_n_u_IBV_REREG_MR_CHANGE_TRANSLATION __pyx_string_tab[256]
#define __pyx_n_u_IBV_REREG_MR_ERR_CMD __pyx_string_tab[257]
#define __pyx_n_u_IBV_REREG_MR_ERR_CMD_AND_DO_FORK __pyx_string_tab[258]
#define __pyx_n_u_IBV_REREG_MR_ERR_DONT_FORK_NEW __pyx_string_tab[259]
#define __pyx_n_u_IBV_REREG_MR_ERR_DO_FORK_OLD __pyx_string_tab[260]
#define __pyx_n_u_IBV_REREG_MR_ERR_INPUT __pyx_string_tab[261]
#define __pyx_n_u_IBV_RX_HASH_DST_IPV4 __pyx_string_tab[262]
#define __pyx_n_u_IBV_RX_HASH_DST_IPV6 __pyx_string_tab[263]
#define __pyx_n_u_IBV_RX_HASH_DST_PORT_TCP __pyx_string_tab[264]
#define __pyx_n_u_IBV_RX_HASH_DST_PORT_UDP __pyx_string_tab[265]
#define __pyx_n_u_IBV_RX_HASH_FUNC_TOEPLITZ __pyx_string_tab[266]
#define __pyx_n_u_IBV_RX_HASH_SRC_IPV4 __pyx_string_tab[267]
#define __pyx_n_u_IBV_RX_HASH_SRC_IPV6 __pyx_string_tab[268]
#define __pyx_n_u_IBV_RX_HASH_SRC_PORT_TCP __pyx_string_tab[269]
#define __pyx_n_u_IBV_RX_HASH_SRC_PORT_UDP __pyx_string_tab[270]
#define __pyx_n_u_IBV_SEND_FENCE __pyx_string_tab[271]
#define __pyx_n_u_IBV_SEND_INLINE __pyx_string_tab[272]
#define __pyx_n_u_IBV_SEND_IP_CSUM __pyx_string_tab[273]
#define __pyx_n_u_IBV_SEND_SIGNALED __pyx_string_tab[274]
#define __pyx_n_u_IBV_SEND_SOLICITED __pyx_string_tab[275]
#define __pyx_n_u_IBV_SRQT_BASIC __pyx_string_tab[276]
#define __pyx_n_u_IBV_SRQT_TM __pyx_string_tab[277]
#define __pyx_n_u_IBV_SRQT_XRC __pyx_string_tab[278]
#define __pyx_n_u_IBV_SRQ_INIT_ATTR_CQ __pyx_string_tab[279]
#define __pyx_n_u_IBV_SRQ_INIT_ATTR_PD __pyx_string_tab[280]
#define __pyx_n_u_IBV_SRQ_INIT_ATTR_TM __pyx_string_tab[281]
#define __pyx_n_u_IBV_SRQ_INIT_ATTR_TYPE __pyx_string_tab[282]
#define __pyx_n_u_IBV_SRQ_INIT_ATTR_XRCD __pyx_string_tab[283]
#define __pyx_n_u_IBV_SRQ_LIMIT __pyx_string_tab[284]
#define __pyx_n_u_IBV_SRQ_MAX_WR __pyx_string_tab[285]
#define __pyx_n_u_IBV_TMH_EAGER __pyx_string_tab[286]
#define __pyx_n_u_IBV_TMH_FIN __pyx_string_tab[287]
#define __pyx_n_u_IBV_TMH_NO_TAG __pyx_string_tab[288]
#define __pyx_n_u_IBV_TMH_RNDV __pyx_string_tab[289]
#define __pyx_n_u_IBV_TM_CAP_RC __pyx_string_tab[290]
#define __pyx_n_u_IBV_TPH_MEM_TYPE_PM __pyx_string_tab[291]
#define __pyx_n_u_IBV_TPH_MEM_TYPE_VM __pyx_string_tab[292]
#define __pyx_n_u_IBV_TRANSPORT_IB __pyx_string_tab[293]
#define __pyx_n_u_IBV_TRANSPORT_IWARP __pyx_string_tab[294]
#define __pyx_n_u_IBV_TRANSPORT_UNKNOWN __pyx_string_tab[295]
#define __pyx_n_u_IBV_TRANSPORT_USNIC __pyx_string_tab[296]
#define __pyx_n_u_IBV_TRANSPORT_USNIC_UDP __pyx_string_tab[297]
#define __pyx_n_u_IBV_VALUES_MASK_RAW_CLOCK __pyx_string_tab[298]
#define __pyx_n_u_IBV_WC_ATOMIC_WRITE __pyx_string_tab[299]
#define __pyx_n_u_IBV_WC_BAD_RESP_ERR __pyx_string_tab[300]
#define __pyx_n_u_IBV_WC_BIND_MW __pyx_string_tab[301]
#define __pyx_n_u_IBV_WC_COMP_SWAP __pyx_string_tab[302]
#define __pyx_n_u_IBV_WC_DRIVER2 __pyx_string_tab[303]
#define __pyx_n_u_IBV_WC_DRIVER3 __pyx_string_tab[304]
#define __pyx_n_u_IBV_WC_EX_WITH_BYTE_LEN __pyx_string_tab[305]
#define __pyx_n_u_IBV_WC_EX_WITH_COMPLETION_TIMEST __pyx_string_tab[306]
#define __pyx_n_u_IBV_WC_EX_WITH_COMPLETION_TIMEST_2 __pyx_string_tab[307]
#define __pyx_n_u_IBV_WC_EX_WITH_CVLAN __pyx_string_tab[308]
#define __pyx_n_u_IBV_WC_EX_WITH_DLID_PATH_BITS __pyx_string_tab[309]
#define __pyx_n_u_IBV_WC_EX_WITH_FLOW_TAG __pyx_string_tab[310]
#define __pyx_n_u_IBV_WC_EX_WITH_IMM __pyx_string_tab[311]
#define __pyx_n_u_IBV_WC_EX_WITH_QP_NUM __pyx_string_tab[312]
#define __pyx_n_u_IBV_WC_EX_WITH_SL __pyx_string_tab[313]
#define __pyx_n_u_IBV_WC_EX_WITH_SLID __pyx_string_tab[314]
#define __pyx_n_u_IBV_WC_EX_WITH_SRC_QP __pyx_string_tab[315]
#define __pyx_n_u_IBV_WC_EX_WITH_TM_INFO __pyx_string_tab[316]
#define __pyx_n_u_IBV_WC_FATAL_ERR __pyx_string_tab[317]
#define __pyx_n_u_IBV_WC_FETCH_ADD __pyx_string_tab[318]
#define __pyx_n_u_IBV_WC_FLUSH __pyx_string_tab[319]
#define __pyx_n_u_IBV_WC_GENERAL_ERR __pyx_string_tab[320]
#define __pyx_n_u_IBV_WC_GRH __pyx_string_tab[321]
#define __pyx_n_u_IBV_WC_INV_EECN_ERR __pyx_string_tab[322]
#define __pyx_n_u_IBV_WC_INV_EEC_STATE_ERR __pyx_string_tab[323]
#define __pyx_n_u_IBV_WC_IP_CSUM_OK __pyx_string_tab[324]
#define __pyx_n_u_IBV_WC_LOCAL_INV __pyx_string_tab[325]
#define __pyx_n_u_IBV_WC_LOC_ACCESS_ERR __pyx_string_tab[326]
#define __pyx_n_u_IBV_WC_LOC_EEC_OP_ERR __pyx_string_tab[327]
#define __pyx_n_u_IBV_WC_LOC_LEN_ERR __pyx_string_tab[328]
#define __pyx_n_u_IBV_WC_LOC_PROT_ERR __pyx_string_tab[329]
#define __pyx_n_u_IBV_WC_LOC_QP_OP_ERR __pyx_string_tab[330]
#define __pyx_n_u_IBV_WC_LOC_RDD_VIOL_ERR __pyx_string_tab[331]
#define __pyx_n_u_IBV_WC_MW_BIND_ERR __pyx_string_tab[332]
#define __pyx_n_u_IBV_WC_RDMA_READ __pyx_string_tab[333]
#define __pyx_n_u_IBV_WC_RDMA_WRITE __pyx_string_tab[334]
#define __pyx_n_u_IBV_WC_RECV __pyx_string_tab[335]
#define __pyx_n_u_IBV_WC_RECV_RDMA_WITH_IMM __pyx_string_tab[336]
#define __pyx_n_u_IBV_WC_REM_ABORT_ERR __pyx_string_tab[337]
#define __pyx_n_u_IBV_WC_REM_ACCESS_ERR __pyx_string_tab[338]
#define __pyx_n_u_IBV_WC_REM_INV_RD_REQ_ERR __pyx_string_tab[339]
#define __pyx_n_u_IBV_WC_REM_INV_REQ_ERR __pyx_string_tab[340]
#define __pyx_n_u_IBV_WC_REM_OP_ERR __pyx_string_tab[341]
#define __pyx_n_u_IBV_WC_RESP_TIMEOUT_ERR __pyx_string_tab[342]
#define __pyx_n_u_IBV_WC_RETRY_EXC_ERR __pyx_string_tab[343]
#define __pyx_n_u_IBV_WC_RNR_RETRY_EXC_ERR __pyx_string_tab[344]
#define __pyx_n_u_IBV_WC_SEND __pyx_string_tab[345]
#define __pyx_n_u_IBV_WC_SUCCESS __pyx_string_tab[346]
#define __pyx_n_u_IBV_WC_TM_ADD __pyx_string_tab[347]
#define __pyx_n_u_IBV_WC_TM_DATA_VALID __pyx_string_tab[348]
#define __pyx_n_u_IBV_WC_TM_DEL __pyx_string_tab[349]
#define __pyx_n_u_IBV_WC_TM_ERR __pyx_string_tab[350]
#define __pyx_n_u_IBV_WC_TM_MATCH __pyx_string_tab[351]
#define __pyx_n_u_IBV_WC_TM_NO_TAG __pyx_string_tab[352]
#define __pyx_n_u_IBV_WC_TM_RECV __pyx_string_tab[353]
#define __pyx_n_u_IBV_WC_TM_RNDV_INCOMPLETE __pyx_string_tab[354]
#define __pyx_n_u_IBV_WC_TM_SYNC __pyx_string_tab[355]
#define __pyx_n_u_IBV_WC_TM_SYNC_REQ __pyx_string_tab[356]
#define __pyx_n_u_IBV_WC_TSO __pyx_string_tab[357]
#define __pyx_n_u_IBV_WC_WITH_IMM __pyx_string_tab[358]
#define __pyx_n_u_IBV_WC_WITH_INV __pyx_string_tab[359]
#define __pyx_n_u_IBV_WC_WR_FLUSH_ERR __pyx_string_tab[360]
#define __pyx_n_u_IBV_WQS_ERR __pyx_string_tab[361]
#define __pyx_n_u_IBV_WQS_RDY __pyx_string_tab[362]
#define __pyx_n_u_IBV_WQS_RESET __pyx_string_tab[363]
#define __pyx_n_u_IBV_WQS_UNKNOWN __pyx_string_tab[364]
#define __pyx_n_u_IBV_WQT_RQ __pyx_string_tab[365]
#define __pyx_n_u_IBV_WQ_ATTR_CURR_STATE __pyx_string_tab[366]
#define __pyx_n_u_IBV_WQ_ATTR_FLAGS __pyx_string_tab[367]
#define __pyx_n_u_IBV_WQ_ATTR_STATE __pyx_string_tab[368]
#define __pyx_n_u_IBV_WQ_FLAGS_CVLAN_STRIPPING __pyx_string_tab[369]
#define __pyx_n_u_IBV_WQ_FLAGS_DELAY_DROP __pyx_string_tab[370]
#define __pyx_n_u_IBV_WQ_FLAGS_PCI_WRITE_END_PADDI __pyx_string_tab[371]
#define __pyx_n_u_IBV_WQ_FLAGS_SCATTER_FCS __pyx_string_tab[372]
#define __pyx_n_u_IBV_WQ_INIT_ATTR_FLAGS __pyx_string_tab[373]
#define __pyx_n_u_IBV_WR_ATOMIC_CMP_AND_SWP __pyx_string_tab[374]
#define __pyx_n_u_IBV_WR_ATOMIC_FETCH_AND_ADD __pyx_string_tab[375]
#define __pyx_n_u_IBV_WR_ATOMIC_WRITE __pyx_string_tab[376]
#define __pyx_n_u_IBV_WR_BIND_MW __pyx_string_tab[377]
#define __pyx_n_u_IBV_WR_FLUSH __pyx_string_tab[378]
#define __pyx_n_u_IBV_WR_LOCAL_INV __pyx_string_tab[379]
#define __pyx_n_u_IBV_WR_RDMA_READ __pyx_string_tab[380]
#define __pyx_n_u_IBV_WR_RDMA_WRITE __pyx_string_tab[381]
#define __pyx_n_u_IBV_WR_RDMA_WRITE_WITH_IMM __pyx_string_tab[382]
#define __pyx_n_u_IBV_WR_SEND __pyx_string_tab[383]
#define __pyx_n_u_IBV_WR_SEND_WITH_IMM __pyx_string_tab[384]
#define __pyx_n_u_IBV_WR_SEND_WITH_INV __pyx_string_tab[385]
#define __pyx_n_u_IBV_WR_TAG_ADD __pyx_string_tab[386]
#define __pyx_n_u_IBV_WR_TAG_DEL __pyx_string_tab[387]
#define __pyx_n_u_IBV_WR_TAG_SYNC __pyx_string_tab[388]
#define __pyx_n_u_IBV_WR_TSO __pyx_string_tab[389]
#define __pyx_n_u_IBV_XRCD_INIT_ATTR_FD __pyx_string_tab[390]
#define __pyx_n_u_IBV_XRCD_INIT_ATTR_OFLAGS __pyx_string_tab[391]
#define __pyx_n_u_IBV_XRCD_INIT_ATTR_RESERVED __pyx_string_tab[392]
#define __pyx_n_u_IntEnum __pyx_string_tab[393]
#define __pyx_n_u_IntFlag __pyx_string_tab[394]
#define __pyx_n_u_cline_in_traceback __pyx_string_tab[395]
#define __pyx_n_u_enum __pyx_string_tab[396]
#define __pyx_n_u_get __pyx_string_tab[397]
#define __pyx_n_u_ibv_access_flags __pyx_string_tab[398]
#define __pyx_n_u_ibv_atomic_cap __pyx_string_tab[399]
#define __pyx_n_u_ibv_counter_description __pyx_string_tab[400]
#define __pyx_n_u_ibv_cq_init_attr_mask __pyx_string_tab[401]
#define __pyx_n_u_ibv_create_cq_attr_flags __pyx_string_tab[402]
#define __pyx_n_u_ibv_create_cq_wc_flags __pyx_string_tab[403]
#define __pyx_n_u_ibv_device_cap_flags __pyx_string_tab[404]
#define __pyx_n_u_ibv_dmah_init_attr_mask __pyx_string_tab[405]
#define __pyx_n_u_ibv_event_type __pyx_string_tab[406]
#define __pyx_n_u_ibv_flow_attr_type __pyx_string_tab[407]
#define __pyx_n_u_ibv_flow_flags __pyx_string_tab[408]
#define __pyx_n_u_ibv_flow_spec_type __pyx_string_tab[409]
#define __pyx_n_u_ibv_fork_status __pyx_string_tab[410]
#define __pyx_n_u_ibv_gid_type __pyx_string_tab[411]
#define __pyx_n_u_ibv_gid_type_sysfs __pyx_string_tab[412]
#define __pyx_n_u_ibv_mig_state __pyx_string_tab[413]
#define __pyx_n_u_ibv_mr_init_attr_mask __pyx_string_tab[414]
#define __pyx_n_u_ibv_mtu __pyx_string_tab[415]
#define __pyx_n_u_ibv_mw_type __pyx_string_tab[416]
#define __pyx_n_u_ibv_node_type __pyx_string_tab[417]
#define __pyx_n_u_ibv_odp_general_caps __pyx_string_tab[418]
#define __pyx_n_u_ibv_odp_transport_cap_bits __pyx_string_tab[419]
#define __pyx_n_u_ibv_ops_flags __pyx_string_tab[420]
#define __pyx_n_u_ibv_ops_wr_opcode __pyx_string_tab[421]
#define __pyx_n_u_ibv_parent_domain_init_attr_mask __pyx_string_tab[422]
#define __pyx_n_u_ibv_placement_type __pyx_string_tab[423]
#define __pyx_n_u_ibv_port_cap_flags __pyx_string_tab[424]
#define __pyx_n_u_ibv_port_cap_flags2 __pyx_string_tab[425]
#define __pyx_n_u_ibv_port_state __pyx_string_tab[426]
#define __pyx_n_u_ibv_qp_attr_mask __pyx_string_tab[427]
#define __pyx_n_u_ibv_qp_create_flags __pyx_string_tab[428]
#define __pyx_n_u_ibv_qp_create_send_ops_flags __pyx_string_tab[429]
#define __pyx_n_u_ibv_qp_init_attr_mask __pyx_string_tab[430]
#define __pyx_n_u_ibv_qp_state __pyx_string_tab[431]
#define __pyx_n_u_ibv_qp_type __pyx_string_tab[432]
#define __pyx_n_u_ibv_query_qp_data_in_order_caps __pyx_string_tab[433]
#define __pyx_n_u_ibv_query_qp_data_in_order_flags __pyx_string_tab[434]
#define __pyx_n_u_ibv_raw_packet_caps __pyx_string_tab[435]
#define __pyx_n_u_ibv_read_counters_flags __pyx_string_tab[436]
#define __pyx_n_u_ibv_rereg_mr_err_code __pyx_string_tab[437]
#define __pyx_n_u_ibv_rereg_mr_flags __pyx_string_tab[438]
#define __pyx_n_u_ibv_rx_hash_fields __pyx_string_tab[439]
#define __pyx_n_u_ibv_rx_hash_function_flags __pyx_string_tab[440]
#define __pyx_n_u_ibv_selectivity_level __pyx_string_tab[441]
#define __pyx_n_u_ibv_send_flags __pyx_string_tab[442]
#define __pyx_n_u_ibv_srq_attr_mask __pyx_string_tab[443]
#define __pyx_n_u_ibv_srq_init_attr_mask __pyx_string_tab[444]
#define __pyx_n_u_ibv_srq_type __pyx_string_tab[445]
#define __pyx_n_u_ibv_tm_cap_flags __pyx_string_tab[446]
#define __pyx_n_u_ibv_tmh_op __pyx_string_tab[447]
#define __pyx_n_u_ibv_tph_mem_type __pyx_string_tab[448]
#define __pyx_n_u_ibv_transport_type __pyx_string_tab[449]
#define __pyx_n_u_ibv_values_mask __pyx_string_tab[450]
#define __pyx_n_u_ibv_wc_flags __pyx_string_tab[451]
#define __pyx_n_u_ibv_wc_opcode __pyx_string_tab[452]
#define __pyx_n_u_ibv_wc_status __pyx_string_tab[453]
#define __pyx_n_u_ibv_wq_attr_mask __pyx_string_tab[454]
#define __pyx_n_u_ibv_wq_flags __pyx_string_tab[455]
#define __pyx_n_u_ibv_wq_init_attr_mask __pyx_string_tab[456]
#define __pyx_n_u_ibv_wq_state __pyx_string_tab[457]
#define __pyx_n_u_ibv_wq_type __pyx_string_tab[458]
#define __pyx_n_u_ibv_wr_opcode __pyx_string_tab[459]
#define __pyx_n_u_ibv_xrcd_init_attr_mask __pyx_string_tab[460]
#define __pyx_n_u_main __pyx_string_tab[461]
#define __pyx_n_u_member_names __pyx_string_tab[462]
#define __pyx_n_u_members __pyx_string_tab[463]
#define __pyx_n_u_module __pyx_string_tab[464]
#define __pyx_n_u_module_2 __pyx_string_tab[465]
#define __pyx_n_u_name __pyx_string_tab[466]
#define __pyx_n_u_qualname __pyx_string_tab[467]
#define __pyx_n_u_setdefault __pyx_string_tab[468]
#define __pyx_n_u_test __pyx_string_tab[469]
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_clear(PyObject *m) {
  __pyx_mstatetype *clear_module_state = __Pyx_PyModule_GetState(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __Pyx_State_RemoveModule(NULL);
  #endif
  for (int i=0; i<1; ++i) { Py_CLEAR(clear_module_state->__pyx_tuple[i]); }
  for (int i=0; i<470; ++i) { Py_CLEAR(clear_module_state->__pyx_string_tab[i]); }
/* #### Code section: module_state_clear_contents ### */
/* #### Code section: module_state_clear_end ### */
return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstatetype *traverse_module_state = __Pyx_PyModule_GetState(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_tuple);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_bytes);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_unicode);
  for (int i=0; i<1; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_tuple[i]); }
  for (int i=0; i<470; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_string_tab[i]); }
/* #### Code section: module_state_traverse_contents ### */
/* #### Code section: module_state_traverse_end ### */
return 0;
}
#endif
/* #### Code section: module_code ### */
/* #### Code section: module_exttypes ### */

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
/* #### Code section: initfunc_declarations ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate); /*proto*/
/* #### Code section: init_module ### */

static int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_EnumBase = Py_None; Py_INCREF(Py_None);
  __Pyx_FlagBase = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_libibverbs_enums(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_libibverbs_enums},
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  {Py_mod_gil, __Pyx_FREETHREADING_COMPATIBLE},
  #endif
  #if PY_VERSION_HEX >= 0x030C0000 && CYTHON_USE_MODULE_STATE
  {Py_mod_multiple_interpreters, Py_MOD_MULTIPLE_INTERPRETERS_NOT_SUPPORTED},
  #endif
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "libibverbs_enums",
      0, /* m_doc */
    #if CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstatetype), /* m_size */
    #else
      (CYTHON_PEP489_MULTI_PHASE_INIT) ? 0 : -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif

/* PyModInitFuncType */
#ifndef CYTHON_NO_PYINIT_EXPORT
  #define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#else
  #ifdef __cplusplus
  #define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
  #else
  #define __Pyx_PyMODINIT_FUNC PyObject *
  #endif
#endif

__Pyx_PyMODINIT_FUNC PyInit_libibverbs_enums(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_libibverbs_enums(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
/* ModuleCreationPEP489 */
#if CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
static PY_INT64_T __Pyx_GetCurrentInterpreterId(void) {
    {
        PyObject *module = PyImport_ImportModule("_interpreters"); // 3.13+ I think
        if (!module) {
            PyErr_Clear(); // just try the 3.8-3.12 version
            module = PyImport_ImportModule("_xxsubinterpreters");
            if (!module) goto bad;
        }
        PyObject *current = PyObject_CallMethod(module, "get_current", NULL);
        Py_DECREF(module);
        if (!current) goto bad;
        if (PyTuple_Check(current)) {
            PyObject *new_current = PySequence_GetItem(current, 0);
            Py_DECREF(current);
            current = new_current;
            if (!new_current) goto bad;
        }
        long long as_c_int = PyLong_AsLongLong(current);
        Py_DECREF(current);
        return as_c_int;
    }
  bad:
    PySys_WriteStderr("__Pyx_GetCurrentInterpreterId failed. Try setting the C define CYTHON_PEP489_MULTI_PHASE_INIT=0\n");
    return -1;
}
#endif
#if !CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    static PY_INT64_T main_interpreter_id = -1;
#if CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
    PY_INT64_T current_id = GraalPyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_GRAAL
    PY_INT64_T current_id = PyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
    PY_INT64_T current_id = __Pyx_GetCurrentInterpreterId();
#elif CYTHON_COMPILING_IN_LIMITED_API
    PY_INT64_T current_id = PyInterpreterState_GetID(PyInterpreterState_Get());
#else
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
#endif
    if (unlikely(current_id == -1)) {
        return -1;
    }
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return 0;
    } else if (unlikely(main_interpreter_id != current_id)) {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#endif
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    #if !CYTHON_USE_MODULE_STATE
    if (__Pyx_check_single_interpreter())
        return NULL;
    #endif
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_libibverbs_enums(PyObject *__pyx_pyinit_module)
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  __pyx_mstatetype *__pyx_mstate = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  size_t __pyx_t_18;
  int __pyx_t_19;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  PyObject *__pyx_t_28 = NULL;
  PyObject *__pyx_t_29 = NULL;
  PyObject *__pyx_t_30 = NULL;
  PyObject *__pyx_t_31 = NULL;
  PyObject *__pyx_t_32 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'libibverbs_enums' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #else
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_t_1 = __pyx_pyinit_module;
  Py_INCREF(__pyx_t_1);
  #else
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #if CYTHON_USE_MODULE_STATE
  {
    int add_module_result = __Pyx_State_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "libibverbs_enums" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = __pyx_t_1;
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  PyUnstable_Module_SetGIL(__pyx_m, Py_MOD_GIL_USED);
  #endif
  __pyx_mstate = __pyx_mstate_global;
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_mstate->__pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_mstate->__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_mstate->__pyx_d);
  __pyx_mstate->__pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_mstate->__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_cython_runtime = __Pyx_PyImport_AddModuleRef("cython_runtime"); if (unlikely(!__pyx_mstate->__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_mstate->__pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /* ImportRefnannyAPI */
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
    PyErr_Clear();
    __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
    if (!__Pyx_RefNanny)
        Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  
__Pyx_RefNannySetupContext("PyInit_libibverbs_enums", 0);
  __Pyx_init_runtime_version();
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_mstate->__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Library function declarations ---*/
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__pyx_module_is_main_pyverbs__libibverbs_enums) {
    if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_name, __pyx_mstate_global->__pyx_n_u_main) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "pyverbs.libibverbs_enums")) {
      if (unlikely((PyDict_SetItemString(modules, "pyverbs.libibverbs_enums", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_CreateCodeObjects(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_export_code(__pyx_mstate);
  (void)__Pyx_modinit_function_export_code(__pyx_mstate);
  (void)__Pyx_modinit_type_init_code(__pyx_mstate);
  (void)__Pyx_modinit_type_import_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_import_code(__pyx_mstate);
  (void)__Pyx_modinit_function_import_code(__pyx_mstate);
  /*--- Execution code ---*/

  /* "EnumBase":10
 * 
 * cdef object __Pyx_EnumBase
 * from enum import IntEnum as __Pyx_EnumBase             # <<<<<<<<<<<<<<
 * 
 * cdef object __Pyx_FlagBase
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_IntEnum};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_enum, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 10, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_IntEnum};
    __pyx_t_3 = 0; {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 10, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      switch (__pyx_t_3) {
        case 0:
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_XGOTREF(__Pyx_EnumBase);
        __Pyx_DECREF_SET(__Pyx_EnumBase, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        break;
        default:;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumBase":13
 * 
 * cdef object __Pyx_FlagBase
 * from enum import IntFlag as __Pyx_FlagBase             # <<<<<<<<<<<<<<
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_IntFlag};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_enum, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_IntFlag};
    __pyx_t_3 = 0; {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      switch (__pyx_t_3) {
        case 0:
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_XGOTREF(__Pyx_FlagBase);
        __Pyx_DECREF_SET(__Pyx_FlagBase, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        break;
        default:;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_access_flags = __Pyx_FlagBase('ibv_access_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_ACCESS_LOCAL_WRITE', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_LOCAL_WRITE)),
 *     ('IBV_ACCESS_REMOTE_WRITE', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_REMOTE_WRITE)),
*/
  __pyx_t_4 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_5 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_access_flags = __Pyx_FlagBase('ibv_access_flags', [
 *     ('IBV_ACCESS_LOCAL_WRITE', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_LOCAL_WRITE)),             # <<<<<<<<<<<<<<
 *     ('IBV_ACCESS_REMOTE_WRITE', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_REMOTE_WRITE)),
 *     ('IBV_ACCESS_REMOTE_READ', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_REMOTE_READ)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_LOCAL_WRITE); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ACCESS_LOCAL_WRITE, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":25
 * ibv_access_flags = __Pyx_FlagBase('ibv_access_flags', [
 *     ('IBV_ACCESS_LOCAL_WRITE', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_LOCAL_WRITE)),
 *     ('IBV_ACCESS_REMOTE_WRITE', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_REMOTE_WRITE)),             # <<<<<<<<<<<<<<
 *     ('IBV_ACCESS_REMOTE_READ', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_REMOTE_READ)),
 *     ('IBV_ACCESS_REMOTE_ATOMIC', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_REMOTE_ATOMIC)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_REMOTE_WRITE); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ACCESS_REMOTE_WRITE, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":26
 *     ('IBV_ACCESS_LOCAL_WRITE', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_LOCAL_WRITE)),
 *     ('IBV_ACCESS_REMOTE_WRITE', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_REMOTE_WRITE)),
 *     ('IBV_ACCESS_REMOTE_READ', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_REMOTE_READ)),             # <<<<<<<<<<<<<<
 *     ('IBV_ACCESS_REMOTE_ATOMIC', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_REMOTE_ATOMIC)),
 *     ('IBV_ACCESS_MW_BIND', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_MW_BIND)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_REMOTE_READ); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_9 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ACCESS_REMOTE_READ, __pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":27
 *     ('IBV_ACCESS_REMOTE_WRITE', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_REMOTE_WRITE)),
 *     ('IBV_ACCESS_REMOTE_READ', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_REMOTE_READ)),
 *     ('IBV_ACCESS_REMOTE_ATOMIC', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_REMOTE_ATOMIC)),             # <<<<<<<<<<<<<<
 *     ('IBV_ACCESS_MW_BIND', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_MW_BIND)),
 *     ('IBV_ACCESS_ZERO_BASED', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_ZERO_BASED)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_REMOTE_ATOMIC); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_10 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ACCESS_REMOTE_ATOMIC, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":28
 *     ('IBV_ACCESS_REMOTE_READ', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_REMOTE_READ)),
 *     ('IBV_ACCESS_REMOTE_ATOMIC', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_REMOTE_ATOMIC)),
 *     ('IBV_ACCESS_MW_BIND', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_MW_BIND)),             # <<<<<<<<<<<<<<
 *     ('IBV_ACCESS_ZERO_BASED', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_ZERO_BASED)),
 *     ('IBV_ACCESS_ON_DEMAND', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_ON_DEMAND)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_MW_BIND); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ACCESS_MW_BIND, __pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":29
 *     ('IBV_ACCESS_REMOTE_ATOMIC', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_REMOTE_ATOMIC)),
 *     ('IBV_ACCESS_MW_BIND', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_MW_BIND)),
 *     ('IBV_ACCESS_ZERO_BASED', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_ZERO_BASED)),             # <<<<<<<<<<<<<<
 *     ('IBV_ACCESS_ON_DEMAND', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_ON_DEMAND)),
 *     ('IBV_ACCESS_HUGETLB', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_HUGETLB)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_ZERO_BASED); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_12 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ACCESS_ZERO_BASED, __pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":30
 *     ('IBV_ACCESS_MW_BIND', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_MW_BIND)),
 *     ('IBV_ACCESS_ZERO_BASED', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_ZERO_BASED)),
 *     ('IBV_ACCESS_ON_DEMAND', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_ON_DEMAND)),             # <<<<<<<<<<<<<<
 *     ('IBV_ACCESS_HUGETLB', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_HUGETLB)),
 *     ('IBV_ACCESS_FLUSH_GLOBAL', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_FLUSH_GLOBAL)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_ON_DEMAND); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_13 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ACCESS_ON_DEMAND, __pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":31
 *     ('IBV_ACCESS_ZERO_BASED', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_ZERO_BASED)),
 *     ('IBV_ACCESS_ON_DEMAND', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_ON_DEMAND)),
 *     ('IBV_ACCESS_HUGETLB', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_HUGETLB)),             # <<<<<<<<<<<<<<
 *     ('IBV_ACCESS_FLUSH_GLOBAL', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_FLUSH_GLOBAL)),
 *     ('IBV_ACCESS_FLUSH_PERSISTENT', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_FLUSH_PERSISTENT)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_HUGETLB); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_14 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ACCESS_HUGETLB, __pyx_t_6); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":32
 *     ('IBV_ACCESS_ON_DEMAND', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_ON_DEMAND)),
 *     ('IBV_ACCESS_HUGETLB', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_HUGETLB)),
 *     ('IBV_ACCESS_FLUSH_GLOBAL', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_FLUSH_GLOBAL)),             # <<<<<<<<<<<<<<
 *     ('IBV_ACCESS_FLUSH_PERSISTENT', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_FLUSH_PERSISTENT)),
 *     ('IBV_ACCESS_RELAXED_ORDERING', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_RELAXED_ORDERING)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_FLUSH_GLOBAL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ACCESS_FLUSH_GLOBAL, __pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":33
 *     ('IBV_ACCESS_HUGETLB', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_HUGETLB)),
 *     ('IBV_ACCESS_FLUSH_GLOBAL', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_FLUSH_GLOBAL)),
 *     ('IBV_ACCESS_FLUSH_PERSISTENT', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_FLUSH_PERSISTENT)),             # <<<<<<<<<<<<<<
 *     ('IBV_ACCESS_RELAXED_ORDERING', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_RELAXED_ORDERING)),
 * 
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_FLUSH_PERSISTENT); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ACCESS_FLUSH_PERSISTENT, __pyx_t_6); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":34
 *     ('IBV_ACCESS_FLUSH_GLOBAL', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_FLUSH_GLOBAL)),
 *     ('IBV_ACCESS_FLUSH_PERSISTENT', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_FLUSH_PERSISTENT)),
 *     ('IBV_ACCESS_RELAXED_ORDERING', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_RELAXED_ORDERING)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_RELAXED_ORDERING); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_17 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ACCESS_RELAXED_ORDERING, __pyx_t_6); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_access_flags = __Pyx_FlagBase('ibv_access_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_ACCESS_LOCAL_WRITE', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_LOCAL_WRITE)),
 *     ('IBV_ACCESS_REMOTE_WRITE', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_REMOTE_WRITE)),
*/
  __pyx_t_6 = __Pyx_PyList_Pack(11, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11, __pyx_t_12, __pyx_t_13, __pyx_t_14, __pyx_t_15, __pyx_t_16, __pyx_t_17); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "EnumType":36
 *     ('IBV_ACCESS_RELAXED_ORDERING', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_RELAXED_ORDERING)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_17 = __Pyx_Globals(); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_17 = __Pyx_PyObject_Call(__pyx_t_16, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_4, __pyx_mstate_global->__pyx_n_u_ibv_access_flags, __pyx_t_6};
    __pyx_t_16 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_17, __pyx_t_16, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_16);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_access_flags, __pyx_t_2) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":38
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":43
 * 
 * 
 *     ibv_access_flags._member_names_ = list(ibv_access_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_ibv_access_flags); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 43, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 43, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PySequence_ListKeepNew(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 43, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_ibv_access_flags); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 43, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_2) < (0)) __PYX_ERR(1, 43, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "EnumType":38
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_atomic_cap = __Pyx_FlagBase('ibv_atomic_cap', [             # <<<<<<<<<<<<<<
 *     ('IBV_ATOMIC_NONE', __Pyx_PyLong_From_enum__ibv_atomic_cap(IBV_ATOMIC_NONE)),
 *     ('IBV_ATOMIC_HCA', __Pyx_PyLong_From_enum__ibv_atomic_cap(IBV_ATOMIC_HCA)),
*/
  __pyx_t_2 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_16 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_atomic_cap = __Pyx_FlagBase('ibv_atomic_cap', [
 *     ('IBV_ATOMIC_NONE', __Pyx_PyLong_From_enum__ibv_atomic_cap(IBV_ATOMIC_NONE)),             # <<<<<<<<<<<<<<
 *     ('IBV_ATOMIC_HCA', __Pyx_PyLong_From_enum__ibv_atomic_cap(IBV_ATOMIC_HCA)),
 *     ('IBV_ATOMIC_GLOB', __Pyx_PyLong_From_enum__ibv_atomic_cap(IBV_ATOMIC_GLOB)),
*/
  __pyx_t_17 = __Pyx_PyLong_From_enum__ibv_atomic_cap(IBV_ATOMIC_NONE); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_6 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ATOMIC_NONE, __pyx_t_17); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "EnumType":25
 * ibv_atomic_cap = __Pyx_FlagBase('ibv_atomic_cap', [
 *     ('IBV_ATOMIC_NONE', __Pyx_PyLong_From_enum__ibv_atomic_cap(IBV_ATOMIC_NONE)),
 *     ('IBV_ATOMIC_HCA', __Pyx_PyLong_From_enum__ibv_atomic_cap(IBV_ATOMIC_HCA)),             # <<<<<<<<<<<<<<
 *     ('IBV_ATOMIC_GLOB', __Pyx_PyLong_From_enum__ibv_atomic_cap(IBV_ATOMIC_GLOB)),
 * 
*/
  __pyx_t_17 = __Pyx_PyLong_From_enum__ibv_atomic_cap(IBV_ATOMIC_HCA); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ATOMIC_HCA, __pyx_t_17); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "EnumType":26
 *     ('IBV_ATOMIC_NONE', __Pyx_PyLong_From_enum__ibv_atomic_cap(IBV_ATOMIC_NONE)),
 *     ('IBV_ATOMIC_HCA', __Pyx_PyLong_From_enum__ibv_atomic_cap(IBV_ATOMIC_HCA)),
 *     ('IBV_ATOMIC_GLOB', __Pyx_PyLong_From_enum__ibv_atomic_cap(IBV_ATOMIC_GLOB)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_17 = __Pyx_PyLong_From_enum__ibv_atomic_cap(IBV_ATOMIC_GLOB); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ATOMIC_GLOB, __pyx_t_17); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_atomic_cap = __Pyx_FlagBase('ibv_atomic_cap', [             # <<<<<<<<<<<<<<
 *     ('IBV_ATOMIC_NONE', __Pyx_PyLong_From_enum__ibv_atomic_cap(IBV_ATOMIC_NONE)),
 *     ('IBV_ATOMIC_HCA', __Pyx_PyLong_From_enum__ibv_atomic_cap(IBV_ATOMIC_HCA)),
*/
  __pyx_t_17 = __Pyx_PyList_Pack(3, __pyx_t_6, __pyx_t_4, __pyx_t_15); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":28
 *     ('IBV_ATOMIC_GLOB', __Pyx_PyLong_From_enum__ibv_atomic_cap(IBV_ATOMIC_GLOB)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_15 = __Pyx_Globals(); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_2, __pyx_mstate_global->__pyx_n_u_ibv_atomic_cap, __pyx_t_17};
    __pyx_t_4 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_15, __pyx_t_4, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_5 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_16, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_atomic_cap, __pyx_t_5) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":35
 * 
 * 
 *     ibv_atomic_cap._member_names_ = list(ibv_atomic_cap.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_ibv_atomic_cap); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PySequence_ListKeepNew(__pyx_t_16); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_ibv_atomic_cap); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_5) < (0)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

    /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_counter_description = __Pyx_FlagBase('ibv_counter_description', [             # <<<<<<<<<<<<<<
 *     ('IBV_COUNTER_PACKETS', __Pyx_PyLong_From_enum__ibv_counter_description(IBV_COUNTER_PACKETS)),
 *     ('IBV_COUNTER_BYTES', __Pyx_PyLong_From_enum__ibv_counter_description(IBV_COUNTER_BYTES)),
*/
  __pyx_t_5 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_4 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_counter_description = __Pyx_FlagBase('ibv_counter_description', [
 *     ('IBV_COUNTER_PACKETS', __Pyx_PyLong_From_enum__ibv_counter_description(IBV_COUNTER_PACKETS)),             # <<<<<<<<<<<<<<
 *     ('IBV_COUNTER_BYTES', __Pyx_PyLong_From_enum__ibv_counter_description(IBV_COUNTER_BYTES)),
 * 
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__ibv_counter_description(IBV_COUNTER_PACKETS); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_17 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_COUNTER_PACKETS, __pyx_t_15); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":25
 * ibv_counter_description = __Pyx_FlagBase('ibv_counter_description', [
 *     ('IBV_COUNTER_PACKETS', __Pyx_PyLong_From_enum__ibv_counter_description(IBV_COUNTER_PACKETS)),
 *     ('IBV_COUNTER_BYTES', __Pyx_PyLong_From_enum__ibv_counter_description(IBV_COUNTER_BYTES)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__ibv_counter_description(IBV_COUNTER_BYTES); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_2 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_COUNTER_BYTES, __pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_counter_description = __Pyx_FlagBase('ibv_counter_description', [             # <<<<<<<<<<<<<<
 *     ('IBV_COUNTER_PACKETS', __Pyx_PyLong_From_enum__ibv_counter_description(IBV_COUNTER_PACKETS)),
 *     ('IBV_COUNTER_BYTES', __Pyx_PyLong_From_enum__ibv_counter_description(IBV_COUNTER_BYTES)),
*/
  __pyx_t_15 = __Pyx_PyList_Pack(2, __pyx_t_17, __pyx_t_2); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":27
 *     ('IBV_COUNTER_BYTES', __Pyx_PyLong_From_enum__ibv_counter_description(IBV_COUNTER_BYTES)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_2 = __Pyx_Globals(); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_5, __pyx_mstate_global->__pyx_n_u_ibv_counter_description, __pyx_t_15};
    __pyx_t_17 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_2, __pyx_t_17, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_16 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_17);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_counter_description, __pyx_t_16) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":34
 * 
 * 
 *     ibv_counter_description._member_names_ = list(ibv_counter_description.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_ibv_counter_description); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __pyx_t_16 = __Pyx_PySequence_ListKeepNew(__pyx_t_4); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_ibv_counter_description); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_16) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_cq_init_attr_mask = __Pyx_FlagBase('ibv_cq_init_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_CQ_INIT_ATTR_MASK_FLAGS', __Pyx_PyLong_From_enum__ibv_cq_init_attr_mask(IBV_CQ_INIT_ATTR_MASK_FLAGS)),
 *     ('IBV_CQ_INIT_ATTR_MASK_PD', __Pyx_PyLong_From_enum__ibv_cq_init_attr_mask(IBV_CQ_INIT_ATTR_MASK_PD)),
*/
  __pyx_t_16 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_17 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_cq_init_attr_mask = __Pyx_FlagBase('ibv_cq_init_attr_mask', [
 *     ('IBV_CQ_INIT_ATTR_MASK_FLAGS', __Pyx_PyLong_From_enum__ibv_cq_init_attr_mask(IBV_CQ_INIT_ATTR_MASK_FLAGS)),             # <<<<<<<<<<<<<<
 *     ('IBV_CQ_INIT_ATTR_MASK_PD', __Pyx_PyLong_From_enum__ibv_cq_init_attr_mask(IBV_CQ_INIT_ATTR_MASK_PD)),
 * 
*/
  __pyx_t_2 = __Pyx_PyLong_From_enum__ibv_cq_init_attr_mask(IBV_CQ_INIT_ATTR_MASK_FLAGS); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_CQ_INIT_ATTR_MASK_FLAGS, __pyx_t_2); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":25
 * ibv_cq_init_attr_mask = __Pyx_FlagBase('ibv_cq_init_attr_mask', [
 *     ('IBV_CQ_INIT_ATTR_MASK_FLAGS', __Pyx_PyLong_From_enum__ibv_cq_init_attr_mask(IBV_CQ_INIT_ATTR_MASK_FLAGS)),
 *     ('IBV_CQ_INIT_ATTR_MASK_PD', __Pyx_PyLong_From_enum__ibv_cq_init_attr_mask(IBV_CQ_INIT_ATTR_MASK_PD)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_2 = __Pyx_PyLong_From_enum__ibv_cq_init_attr_mask(IBV_CQ_INIT_ATTR_MASK_PD); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_CQ_INIT_ATTR_MASK_PD, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_cq_init_attr_mask = __Pyx_FlagBase('ibv_cq_init_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_CQ_INIT_ATTR_MASK_FLAGS', __Pyx_PyLong_From_enum__ibv_cq_init_attr_mask(IBV_CQ_INIT_ATTR_MASK_FLAGS)),
 *     ('IBV_CQ_INIT_ATTR_MASK_PD', __Pyx_PyLong_From_enum__ibv_cq_init_attr_mask(IBV_CQ_INIT_ATTR_MASK_PD)),
*/
  __pyx_t_2 = __Pyx_PyList_Pack(2, __pyx_t_15, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumType":27
 *     ('IBV_CQ_INIT_ATTR_MASK_PD', __Pyx_PyLong_From_enum__ibv_cq_init_attr_mask(IBV_CQ_INIT_ATTR_MASK_PD)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_5 = __Pyx_Globals(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_16, __pyx_mstate_global->__pyx_n_u_ibv_cq_init_attr_mask, __pyx_t_2};
    __pyx_t_15 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_5, __pyx_t_15, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_4 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_17, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_15);
    __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_cq_init_attr_mask, __pyx_t_4) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":34
 * 
 * 
 *     ibv_cq_init_attr_mask._member_names_ = list(ibv_cq_init_attr_mask.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_ibv_cq_init_attr_mask); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PySequence_ListKeepNew(__pyx_t_17); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_ibv_cq_init_attr_mask); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_4) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_create_cq_attr_flags = __Pyx_FlagBase('ibv_create_cq_attr_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_CREATE_CQ_ATTR_SINGLE_THREADED', __Pyx_PyLong_From_enum__ibv_create_cq_attr_flags(IBV_CREATE_CQ_ATTR_SINGLE_THREADED)),
 *     ('IBV_CREATE_CQ_ATTR_IGNORE_OVERRUN', __Pyx_PyLong_From_enum__ibv_create_cq_attr_flags(IBV_CREATE_CQ_ATTR_IGNORE_OVERRUN)),
*/
  __pyx_t_4 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_15 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_create_cq_attr_flags = __Pyx_FlagBase('ibv_create_cq_attr_flags', [
 *     ('IBV_CREATE_CQ_ATTR_SINGLE_THREADED', __Pyx_PyLong_From_enum__ibv_create_cq_attr_flags(IBV_CREATE_CQ_ATTR_SINGLE_THREADED)),             # <<<<<<<<<<<<<<
 *     ('IBV_CREATE_CQ_ATTR_IGNORE_OVERRUN', __Pyx_PyLong_From_enum__ibv_create_cq_attr_flags(IBV_CREATE_CQ_ATTR_IGNORE_OVERRUN)),
 * 
*/
  __pyx_t_5 = __Pyx_PyLong_From_enum__ibv_create_cq_attr_flags(IBV_CREATE_CQ_ATTR_SINGLE_THREADED); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_CREATE_CQ_ATTR_SINGLE_THREAD, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumType":25
 * ibv_create_cq_attr_flags = __Pyx_FlagBase('ibv_create_cq_attr_flags', [
 *     ('IBV_CREATE_CQ_ATTR_SINGLE_THREADED', __Pyx_PyLong_From_enum__ibv_create_cq_attr_flags(IBV_CREATE_CQ_ATTR_SINGLE_THREADED)),
 *     ('IBV_CREATE_CQ_ATTR_IGNORE_OVERRUN', __Pyx_PyLong_From_enum__ibv_create_cq_attr_flags(IBV_CREATE_CQ_ATTR_IGNORE_OVERRUN)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_5 = __Pyx_PyLong_From_enum__ibv_create_cq_attr_flags(IBV_CREATE_CQ_ATTR_IGNORE_OVERRUN); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_CREATE_CQ_ATTR_IGNORE_OVERRU, __pyx_t_5); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_create_cq_attr_flags = __Pyx_FlagBase('ibv_create_cq_attr_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_CREATE_CQ_ATTR_SINGLE_THREADED', __Pyx_PyLong_From_enum__ibv_create_cq_attr_flags(IBV_CREATE_CQ_ATTR_SINGLE_THREADED)),
 *     ('IBV_CREATE_CQ_ATTR_IGNORE_OVERRUN', __Pyx_PyLong_From_enum__ibv_create_cq_attr_flags(IBV_CREATE_CQ_ATTR_IGNORE_OVERRUN)),
*/
  __pyx_t_5 = __Pyx_PyList_Pack(2, __pyx_t_2, __pyx_t_16); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":27
 *     ('IBV_CREATE_CQ_ATTR_IGNORE_OVERRUN', __Pyx_PyLong_From_enum__ibv_create_cq_attr_flags(IBV_CREATE_CQ_ATTR_IGNORE_OVERRUN)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_16 = __Pyx_Globals(); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_4, __pyx_mstate_global->__pyx_n_u_ibv_create_cq_attr_flags, __pyx_t_5};
    __pyx_t_2 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_16, __pyx_t_2, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_17 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_15, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_create_cq_attr_flags, __pyx_t_17) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":34
 * 
 * 
 *     ibv_create_cq_attr_flags._member_names_ = list(ibv_create_cq_attr_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_ibv_create_cq_attr_flags); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __pyx_t_17 = __Pyx_PySequence_ListKeepNew(__pyx_t_15); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_ibv_create_cq_attr_flags); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_17) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_create_cq_wc_flags = __Pyx_FlagBase('ibv_create_cq_wc_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_WC_EX_WITH_BYTE_LEN', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_BYTE_LEN)),
 *     ('IBV_WC_EX_WITH_IMM', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_IMM)),
*/
  __pyx_t_17 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_2 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_create_cq_wc_flags = __Pyx_FlagBase('ibv_create_cq_wc_flags', [
 *     ('IBV_WC_EX_WITH_BYTE_LEN', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_BYTE_LEN)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_EX_WITH_IMM', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_IMM)),
 *     ('IBV_WC_EX_WITH_QP_NUM', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_QP_NUM)),
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_BYTE_LEN); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_BYTE_LEN, __pyx_t_16); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":25
 * ibv_create_cq_wc_flags = __Pyx_FlagBase('ibv_create_cq_wc_flags', [
 *     ('IBV_WC_EX_WITH_BYTE_LEN', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_BYTE_LEN)),
 *     ('IBV_WC_EX_WITH_IMM', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_IMM)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_EX_WITH_QP_NUM', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_QP_NUM)),
 *     ('IBV_WC_EX_WITH_SRC_QP', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_SRC_QP)),
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_IMM); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_IMM, __pyx_t_16); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":26
 *     ('IBV_WC_EX_WITH_BYTE_LEN', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_BYTE_LEN)),
 *     ('IBV_WC_EX_WITH_IMM', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_IMM)),
 *     ('IBV_WC_EX_WITH_QP_NUM', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_QP_NUM)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_EX_WITH_SRC_QP', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_SRC_QP)),
 *     ('IBV_WC_EX_WITH_SLID', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_SLID)),
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_QP_NUM); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_6 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_QP_NUM, __pyx_t_16); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":27
 *     ('IBV_WC_EX_WITH_IMM', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_IMM)),
 *     ('IBV_WC_EX_WITH_QP_NUM', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_QP_NUM)),
 *     ('IBV_WC_EX_WITH_SRC_QP', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_SRC_QP)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_EX_WITH_SLID', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_SLID)),
 *     ('IBV_WC_EX_WITH_SL', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_SL)),
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_SRC_QP); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_14 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_SRC_QP, __pyx_t_16); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":28
 *     ('IBV_WC_EX_WITH_QP_NUM', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_QP_NUM)),
 *     ('IBV_WC_EX_WITH_SRC_QP', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_SRC_QP)),
 *     ('IBV_WC_EX_WITH_SLID', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_SLID)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_EX_WITH_SL', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_SL)),
 *     ('IBV_WC_EX_WITH_DLID_PATH_BITS', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_DLID_PATH_BITS)),
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_SLID); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_13 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_SLID, __pyx_t_16); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":29
 *     ('IBV_WC_EX_WITH_SRC_QP', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_SRC_QP)),
 *     ('IBV_WC_EX_WITH_SLID', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_SLID)),
 *     ('IBV_WC_EX_WITH_SL', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_SL)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_EX_WITH_DLID_PATH_BITS', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_DLID_PATH_BITS)),
 *     ('IBV_WC_EX_WITH_COMPLETION_TIMESTAMP', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_COMPLETION_TIMESTAMP)),
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_SL); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_12 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_SL, __pyx_t_16); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":30
 *     ('IBV_WC_EX_WITH_SLID', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_SLID)),
 *     ('IBV_WC_EX_WITH_SL', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_SL)),
 *     ('IBV_WC_EX_WITH_DLID_PATH_BITS', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_DLID_PATH_BITS)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_EX_WITH_COMPLETION_TIMESTAMP', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_COMPLETION_TIMESTAMP)),
 *     ('IBV_WC_EX_WITH_CVLAN', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_CVLAN)),
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_DLID_PATH_BITS); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_DLID_PATH_BITS, __pyx_t_16); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":31
 *     ('IBV_WC_EX_WITH_SL', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_SL)),
 *     ('IBV_WC_EX_WITH_DLID_PATH_BITS', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_DLID_PATH_BITS)),
 *     ('IBV_WC_EX_WITH_COMPLETION_TIMESTAMP', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_COMPLETION_TIMESTAMP)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_EX_WITH_CVLAN', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_CVLAN)),
 *     ('IBV_WC_EX_WITH_FLOW_TAG', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_FLOW_TAG)),
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_COMPLETION_TIMESTAMP); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_10 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_COMPLETION_TIMEST, __pyx_t_16); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":32
 *     ('IBV_WC_EX_WITH_DLID_PATH_BITS', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_DLID_PATH_BITS)),
 *     ('IBV_WC_EX_WITH_COMPLETION_TIMESTAMP', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_COMPLETION_TIMESTAMP)),
 *     ('IBV_WC_EX_WITH_CVLAN', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_CVLAN)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_EX_WITH_FLOW_TAG', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_FLOW_TAG)),
 *     ('IBV_WC_EX_WITH_TM_INFO', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_TM_INFO)),
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_CVLAN); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_9 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_CVLAN, __pyx_t_16); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":33
 *     ('IBV_WC_EX_WITH_COMPLETION_TIMESTAMP', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_COMPLETION_TIMESTAMP)),
 *     ('IBV_WC_EX_WITH_CVLAN', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_CVLAN)),
 *     ('IBV_WC_EX_WITH_FLOW_TAG', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_FLOW_TAG)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_EX_WITH_TM_INFO', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_TM_INFO)),
 *     ('IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK)),
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_FLOW_TAG); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_8 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_FLOW_TAG, __pyx_t_16); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":34
 *     ('IBV_WC_EX_WITH_CVLAN', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_CVLAN)),
 *     ('IBV_WC_EX_WITH_FLOW_TAG', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_FLOW_TAG)),
 *     ('IBV_WC_EX_WITH_TM_INFO', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_TM_INFO)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK)),
 * 
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_TM_INFO); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_7 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_TM_INFO, __pyx_t_16); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":35
 *     ('IBV_WC_EX_WITH_FLOW_TAG', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_FLOW_TAG)),
 *     ('IBV_WC_EX_WITH_TM_INFO', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_TM_INFO)),
 *     ('IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_20 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_COMPLETION_TIMEST_2, __pyx_t_16); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_create_cq_wc_flags = __Pyx_FlagBase('ibv_create_cq_wc_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_WC_EX_WITH_BYTE_LEN', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_BYTE_LEN)),
 *     ('IBV_WC_EX_WITH_IMM', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_IMM)),
*/
  __pyx_t_16 = __Pyx_PyList_Pack(12, __pyx_t_5, __pyx_t_4, __pyx_t_6, __pyx_t_14, __pyx_t_13, __pyx_t_12, __pyx_t_11, __pyx_t_10, __pyx_t_9, __pyx_t_8, __pyx_t_7, __pyx_t_20); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":37
 *     ('IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK', __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_20 = __Pyx_Globals(); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_20, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_17, __pyx_mstate_global->__pyx_n_u_ibv_create_cq_wc_flags, __pyx_t_16};
    __pyx_t_7 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_20, __pyx_t_7, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_15 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_create_cq_wc_flags, __pyx_t_15) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":39
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":44
 * 
 * 
 *     ibv_create_cq_wc_flags._member_names_ = list(ibv_create_cq_wc_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_ibv_create_cq_wc_flags); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __pyx_t_15 = __Pyx_PySequence_ListKeepNew(__pyx_t_2); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_ibv_create_cq_wc_flags); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_15) < (0)) __PYX_ERR(1, 44, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":39
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_device_cap_flags = __Pyx_FlagBase('ibv_device_cap_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_RESIZE_MAX_WR', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RESIZE_MAX_WR)),
 *     ('IBV_DEVICE_BAD_PKEY_CNTR', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_BAD_PKEY_CNTR)),
*/
  __pyx_t_15 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_7 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_device_cap_flags = __Pyx_FlagBase('ibv_device_cap_flags', [
 *     ('IBV_DEVICE_RESIZE_MAX_WR', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RESIZE_MAX_WR)),             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_BAD_PKEY_CNTR', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_BAD_PKEY_CNTR)),
 *     ('IBV_DEVICE_BAD_QKEY_CNTR', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_BAD_QKEY_CNTR)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RESIZE_MAX_WR); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_RESIZE_MAX_WR, __pyx_t_20); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":25
 * ibv_device_cap_flags = __Pyx_FlagBase('ibv_device_cap_flags', [
 *     ('IBV_DEVICE_RESIZE_MAX_WR', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RESIZE_MAX_WR)),
 *     ('IBV_DEVICE_BAD_PKEY_CNTR', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_BAD_PKEY_CNTR)),             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_BAD_QKEY_CNTR', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_BAD_QKEY_CNTR)),
 *     ('IBV_DEVICE_RAW_MULTI', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RAW_MULTI)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_BAD_PKEY_CNTR); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_17 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_BAD_PKEY_CNTR, __pyx_t_20); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":26
 *     ('IBV_DEVICE_RESIZE_MAX_WR', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RESIZE_MAX_WR)),
 *     ('IBV_DEVICE_BAD_PKEY_CNTR', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_BAD_PKEY_CNTR)),
 *     ('IBV_DEVICE_BAD_QKEY_CNTR', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_BAD_QKEY_CNTR)),             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_RAW_MULTI', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RAW_MULTI)),
 *     ('IBV_DEVICE_AUTO_PATH_MIG', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_AUTO_PATH_MIG)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_BAD_QKEY_CNTR); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_8 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_BAD_QKEY_CNTR, __pyx_t_20); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":27
 *     ('IBV_DEVICE_BAD_PKEY_CNTR', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_BAD_PKEY_CNTR)),
 *     ('IBV_DEVICE_BAD_QKEY_CNTR', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_BAD_QKEY_CNTR)),
 *     ('IBV_DEVICE_RAW_MULTI', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RAW_MULTI)),             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_AUTO_PATH_MIG', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_AUTO_PATH_MIG)),
 *     ('IBV_DEVICE_CHANGE_PHY_PORT', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_CHANGE_PHY_PORT)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RAW_MULTI); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_9 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_RAW_MULTI, __pyx_t_20); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":28
 *     ('IBV_DEVICE_BAD_QKEY_CNTR', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_BAD_QKEY_CNTR)),
 *     ('IBV_DEVICE_RAW_MULTI', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RAW_MULTI)),
 *     ('IBV_DEVICE_AUTO_PATH_MIG', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_AUTO_PATH_MIG)),             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_CHANGE_PHY_PORT', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_CHANGE_PHY_PORT)),
 *     ('IBV_DEVICE_UD_AV_PORT_ENFORCE', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_UD_AV_PORT_ENFORCE)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_AUTO_PATH_MIG); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_10 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_AUTO_PATH_MIG, __pyx_t_20); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":29
 *     ('IBV_DEVICE_RAW_MULTI', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RAW_MULTI)),
 *     ('IBV_DEVICE_AUTO_PATH_MIG', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_AUTO_PATH_MIG)),
 *     ('IBV_DEVICE_CHANGE_PHY_PORT', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_CHANGE_PHY_PORT)),             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_UD_AV_PORT_ENFORCE', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_UD_AV_PORT_ENFORCE)),
 *     ('IBV_DEVICE_CURR_QP_STATE_MOD', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_CURR_QP_STATE_MOD)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_CHANGE_PHY_PORT); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_CHANGE_PHY_PORT, __pyx_t_20); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":30
 *     ('IBV_DEVICE_AUTO_PATH_MIG', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_AUTO_PATH_MIG)),
 *     ('IBV_DEVICE_CHANGE_PHY_PORT', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_CHANGE_PHY_PORT)),
 *     ('IBV_DEVICE_UD_AV_PORT_ENFORCE', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_UD_AV_PORT_ENFORCE)),             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_CURR_QP_STATE_MOD', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_CURR_QP_STATE_MOD)),
 *     ('IBV_DEVICE_SHUTDOWN_PORT', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_SHUTDOWN_PORT)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_UD_AV_PORT_ENFORCE); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_12 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_UD_AV_PORT_ENFORCE, __pyx_t_20); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":31
 *     ('IBV_DEVICE_CHANGE_PHY_PORT', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_CHANGE_PHY_PORT)),
 *     ('IBV_DEVICE_UD_AV_PORT_ENFORCE', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_UD_AV_PORT_ENFORCE)),
 *     ('IBV_DEVICE_CURR_QP_STATE_MOD', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_CURR_QP_STATE_MOD)),             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_SHUTDOWN_PORT', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_SHUTDOWN_PORT)),
 *     ('IBV_DEVICE_INIT_TYPE', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_INIT_TYPE)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_CURR_QP_STATE_MOD); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_13 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_CURR_QP_STATE_MOD, __pyx_t_20); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":32
 *     ('IBV_DEVICE_UD_AV_PORT_ENFORCE', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_UD_AV_PORT_ENFORCE)),
 *     ('IBV_DEVICE_CURR_QP_STATE_MOD', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_CURR_QP_STATE_MOD)),
 *     ('IBV_DEVICE_SHUTDOWN_PORT', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_SHUTDOWN_PORT)),             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_INIT_TYPE', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_INIT_TYPE)),
 *     ('IBV_DEVICE_PORT_ACTIVE_EVENT', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_PORT_ACTIVE_EVENT)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_SHUTDOWN_PORT); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_14 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_SHUTDOWN_PORT, __pyx_t_20); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":33
 *     ('IBV_DEVICE_CURR_QP_STATE_MOD', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_CURR_QP_STATE_MOD)),
 *     ('IBV_DEVICE_SHUTDOWN_PORT', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_SHUTDOWN_PORT)),
 *     ('IBV_DEVICE_INIT_TYPE', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_INIT_TYPE)),             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_PORT_ACTIVE_EVENT', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_PORT_ACTIVE_EVENT)),
 *     ('IBV_DEVICE_SYS_IMAGE_GUID', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_SYS_IMAGE_GUID)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_INIT_TYPE); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_6 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_INIT_TYPE, __pyx_t_20); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":34
 *     ('IBV_DEVICE_SHUTDOWN_PORT', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_SHUTDOWN_PORT)),
 *     ('IBV_DEVICE_INIT_TYPE', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_INIT_TYPE)),
 *     ('IBV_DEVICE_PORT_ACTIVE_EVENT', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_PORT_ACTIVE_EVENT)),             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_SYS_IMAGE_GUID', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_SYS_IMAGE_GUID)),
 *     ('IBV_DEVICE_RC_RNR_NAK_GEN', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RC_RNR_NAK_GEN)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_PORT_ACTIVE_EVENT); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_PORT_ACTIVE_EVENT, __pyx_t_20); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":35
 *     ('IBV_DEVICE_INIT_TYPE', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_INIT_TYPE)),
 *     ('IBV_DEVICE_PORT_ACTIVE_EVENT', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_PORT_ACTIVE_EVENT)),
 *     ('IBV_DEVICE_SYS_IMAGE_GUID', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_SYS_IMAGE_GUID)),             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_RC_RNR_NAK_GEN', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RC_RNR_NAK_GEN)),
 *     ('IBV_DEVICE_SRQ_RESIZE', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_SRQ_RESIZE)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_SYS_IMAGE_GUID); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_SYS_IMAGE_GUID, __pyx_t_20); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":36
 *     ('IBV_DEVICE_PORT_ACTIVE_EVENT', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_PORT_ACTIVE_EVENT)),
 *     ('IBV_DEVICE_SYS_IMAGE_GUID', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_SYS_IMAGE_GUID)),
 *     ('IBV_DEVICE_RC_RNR_NAK_GEN', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RC_RNR_NAK_GEN)),             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_SRQ_RESIZE', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_SRQ_RESIZE)),
 *     ('IBV_DEVICE_N_NOTIFY_CQ', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_N_NOTIFY_CQ)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RC_RNR_NAK_GEN); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_21 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_RC_RNR_NAK_GEN, __pyx_t_20); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":37
 *     ('IBV_DEVICE_SYS_IMAGE_GUID', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_SYS_IMAGE_GUID)),
 *     ('IBV_DEVICE_RC_RNR_NAK_GEN', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RC_RNR_NAK_GEN)),
 *     ('IBV_DEVICE_SRQ_RESIZE', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_SRQ_RESIZE)),             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_N_NOTIFY_CQ', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_N_NOTIFY_CQ)),
 *     ('IBV_DEVICE_MEM_WINDOW', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_WINDOW)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_SRQ_RESIZE); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_SRQ_RESIZE, __pyx_t_20); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":38
 *     ('IBV_DEVICE_RC_RNR_NAK_GEN', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RC_RNR_NAK_GEN)),
 *     ('IBV_DEVICE_SRQ_RESIZE', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_SRQ_RESIZE)),
 *     ('IBV_DEVICE_N_NOTIFY_CQ', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_N_NOTIFY_CQ)),             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_MEM_WINDOW', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_WINDOW)),
 *     ('IBV_DEVICE_UD_IP_CSUM', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_UD_IP_CSUM)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_N_NOTIFY_CQ); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_23 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_N_NOTIFY_CQ, __pyx_t_20); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":39
 *     ('IBV_DEVICE_SRQ_RESIZE', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_SRQ_RESIZE)),
 *     ('IBV_DEVICE_N_NOTIFY_CQ', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_N_NOTIFY_CQ)),
 *     ('IBV_DEVICE_MEM_WINDOW', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_WINDOW)),             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_UD_IP_CSUM', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_UD_IP_CSUM)),
 *     ('IBV_DEVICE_XRC', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_XRC)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_WINDOW); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_24 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_MEM_WINDOW, __pyx_t_20); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":40
 *     ('IBV_DEVICE_N_NOTIFY_CQ', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_N_NOTIFY_CQ)),
 *     ('IBV_DEVICE_MEM_WINDOW', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_WINDOW)),
 *     ('IBV_DEVICE_UD_IP_CSUM', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_UD_IP_CSUM)),             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_XRC', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_XRC)),
 *     ('IBV_DEVICE_MEM_MGT_EXTENSIONS', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_MGT_EXTENSIONS)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_UD_IP_CSUM); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_25 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_UD_IP_CSUM, __pyx_t_20); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":41
 *     ('IBV_DEVICE_MEM_WINDOW', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_WINDOW)),
 *     ('IBV_DEVICE_UD_IP_CSUM', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_UD_IP_CSUM)),
 *     ('IBV_DEVICE_XRC', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_XRC)),             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_MEM_MGT_EXTENSIONS', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_MGT_EXTENSIONS)),
 *     ('IBV_DEVICE_MEM_WINDOW_TYPE_2A', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_WINDOW_TYPE_2A)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_XRC); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_26 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_XRC, __pyx_t_20); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":42
 *     ('IBV_DEVICE_UD_IP_CSUM', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_UD_IP_CSUM)),
 *     ('IBV_DEVICE_XRC', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_XRC)),
 *     ('IBV_DEVICE_MEM_MGT_EXTENSIONS', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_MGT_EXTENSIONS)),             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_MEM_WINDOW_TYPE_2A', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_WINDOW_TYPE_2A)),
 *     ('IBV_DEVICE_MEM_WINDOW_TYPE_2B', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_WINDOW_TYPE_2B)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_MGT_EXTENSIONS); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_27 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_MEM_MGT_EXTENSIONS, __pyx_t_20); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":43
 *     ('IBV_DEVICE_XRC', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_XRC)),
 *     ('IBV_DEVICE_MEM_MGT_EXTENSIONS', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_MGT_EXTENSIONS)),
 *     ('IBV_DEVICE_MEM_WINDOW_TYPE_2A', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_WINDOW_TYPE_2A)),             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_MEM_WINDOW_TYPE_2B', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_WINDOW_TYPE_2B)),
 *     ('IBV_DEVICE_RC_IP_CSUM', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RC_IP_CSUM)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_WINDOW_TYPE_2A); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_28 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_MEM_WINDOW_TYPE_2A, __pyx_t_20); if (unlikely(!__pyx_t_28)) __PYX_ERR(1, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":44
 *     ('IBV_DEVICE_MEM_MGT_EXTENSIONS', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_MGT_EXTENSIONS)),
 *     ('IBV_DEVICE_MEM_WINDOW_TYPE_2A', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_WINDOW_TYPE_2A)),
 *     ('IBV_DEVICE_MEM_WINDOW_TYPE_2B', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_WINDOW_TYPE_2B)),             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_RC_IP_CSUM', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RC_IP_CSUM)),
 *     ('IBV_DEVICE_RAW_IP_CSUM', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RAW_IP_CSUM)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_WINDOW_TYPE_2B); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_29 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_MEM_WINDOW_TYPE_2B, __pyx_t_20); if (unlikely(!__pyx_t_29)) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":45
 *     ('IBV_DEVICE_MEM_WINDOW_TYPE_2A', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_WINDOW_TYPE_2A)),
 *     ('IBV_DEVICE_MEM_WINDOW_TYPE_2B', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_WINDOW_TYPE_2B)),
 *     ('IBV_DEVICE_RC_IP_CSUM', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RC_IP_CSUM)),             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_RAW_IP_CSUM', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RAW_IP_CSUM)),
 *     ('IBV_DEVICE_MANAGED_FLOW_STEERING', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MANAGED_FLOW_STEERING)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RC_IP_CSUM); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_30 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_RC_IP_CSUM, __pyx_t_20); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":46
 *     ('IBV_DEVICE_MEM_WINDOW_TYPE_2B', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MEM_WINDOW_TYPE_2B)),
 *     ('IBV_DEVICE_RC_IP_CSUM', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RC_IP_CSUM)),
 *     ('IBV_DEVICE_RAW_IP_CSUM', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RAW_IP_CSUM)),             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_MANAGED_FLOW_STEERING', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MANAGED_FLOW_STEERING)),
 * 
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RAW_IP_CSUM); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_31 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_RAW_IP_CSUM, __pyx_t_20); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":47
 *     ('IBV_DEVICE_RC_IP_CSUM', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RC_IP_CSUM)),
 *     ('IBV_DEVICE_RAW_IP_CSUM', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RAW_IP_CSUM)),
 *     ('IBV_DEVICE_MANAGED_FLOW_STEERING', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MANAGED_FLOW_STEERING)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MANAGED_FLOW_STEERING); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_32 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_MANAGED_FLOW_STEERING, __pyx_t_20); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_device_cap_flags = __Pyx_FlagBase('ibv_device_cap_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_DEVICE_RESIZE_MAX_WR', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_RESIZE_MAX_WR)),
 *     ('IBV_DEVICE_BAD_PKEY_CNTR', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_BAD_PKEY_CNTR)),
*/
  __pyx_t_20 = __Pyx_PyList_Pack(24, __pyx_t_16, __pyx_t_17, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11, __pyx_t_12, __pyx_t_13, __pyx_t_14, __pyx_t_6, __pyx_t_4, __pyx_t_5, __pyx_t_21, __pyx_t_22, __pyx_t_23, __pyx_t_24, __pyx_t_25, __pyx_t_26, __pyx_t_27, __pyx_t_28, __pyx_t_29, __pyx_t_30, __pyx_t_31, __pyx_t_32); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":49
 *     ('IBV_DEVICE_MANAGED_FLOW_STEERING', __Pyx_PyLong_From_enum__ibv_device_cap_flags(IBV_DEVICE_MANAGED_FLOW_STEERING)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_32 = __Pyx_Globals(); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_31 = __Pyx_PyObject_GetAttrStr(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __pyx_t_32 = __Pyx_PyObject_Call(__pyx_t_31, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_15, __pyx_mstate_global->__pyx_n_u_ibv_device_cap_flags, __pyx_t_20};
    __pyx_t_31 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_32, __pyx_t_31, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_31);
    __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_device_cap_flags, __pyx_t_2) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":51
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":56
 * 
 * 
 *     ibv_device_cap_flags._member_names_ = list(ibv_device_cap_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_ibv_device_cap_flags); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PySequence_ListKeepNew(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_ibv_device_cap_flags); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_2) < (0)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "EnumType":51
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_dmah_init_attr_mask = __Pyx_FlagBase('ibv_dmah_init_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_DMAH_INIT_ATTR_MASK_CPU_ID', __Pyx_PyLong_From_enum__ibv_dmah_init_attr_mask(IBV_DMAH_INIT_ATTR_MASK_CPU_ID)),
 *     ('IBV_DMAH_INIT_ATTR_MASK_PH', __Pyx_PyLong_From_enum__ibv_dmah_init_attr_mask(IBV_DMAH_INIT_ATTR_MASK_PH)),
*/
  __pyx_t_2 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_31 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_dmah_init_attr_mask = __Pyx_FlagBase('ibv_dmah_init_attr_mask', [
 *     ('IBV_DMAH_INIT_ATTR_MASK_CPU_ID', __Pyx_PyLong_From_enum__ibv_dmah_init_attr_mask(IBV_DMAH_INIT_ATTR_MASK_CPU_ID)),             # <<<<<<<<<<<<<<
 *     ('IBV_DMAH_INIT_ATTR_MASK_PH', __Pyx_PyLong_From_enum__ibv_dmah_init_attr_mask(IBV_DMAH_INIT_ATTR_MASK_PH)),
 *     ('IBV_DMAH_INIT_ATTR_MASK_TPH_MEM_TYPE', __Pyx_PyLong_From_enum__ibv_dmah_init_attr_mask(IBV_DMAH_INIT_ATTR_MASK_TPH_MEM_TYPE)),
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_dmah_init_attr_mask(IBV_DMAH_INIT_ATTR_MASK_CPU_ID); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_20 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DMAH_INIT_ATTR_MASK_CPU_ID, __pyx_t_32); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":25
 * ibv_dmah_init_attr_mask = __Pyx_FlagBase('ibv_dmah_init_attr_mask', [
 *     ('IBV_DMAH_INIT_ATTR_MASK_CPU_ID', __Pyx_PyLong_From_enum__ibv_dmah_init_attr_mask(IBV_DMAH_INIT_ATTR_MASK_CPU_ID)),
 *     ('IBV_DMAH_INIT_ATTR_MASK_PH', __Pyx_PyLong_From_enum__ibv_dmah_init_attr_mask(IBV_DMAH_INIT_ATTR_MASK_PH)),             # <<<<<<<<<<<<<<
 *     ('IBV_DMAH_INIT_ATTR_MASK_TPH_MEM_TYPE', __Pyx_PyLong_From_enum__ibv_dmah_init_attr_mask(IBV_DMAH_INIT_ATTR_MASK_TPH_MEM_TYPE)),
 * 
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_dmah_init_attr_mask(IBV_DMAH_INIT_ATTR_MASK_PH); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DMAH_INIT_ATTR_MASK_PH, __pyx_t_32); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":26
 *     ('IBV_DMAH_INIT_ATTR_MASK_CPU_ID', __Pyx_PyLong_From_enum__ibv_dmah_init_attr_mask(IBV_DMAH_INIT_ATTR_MASK_CPU_ID)),
 *     ('IBV_DMAH_INIT_ATTR_MASK_PH', __Pyx_PyLong_From_enum__ibv_dmah_init_attr_mask(IBV_DMAH_INIT_ATTR_MASK_PH)),
 *     ('IBV_DMAH_INIT_ATTR_MASK_TPH_MEM_TYPE', __Pyx_PyLong_From_enum__ibv_dmah_init_attr_mask(IBV_DMAH_INIT_ATTR_MASK_TPH_MEM_TYPE)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_dmah_init_attr_mask(IBV_DMAH_INIT_ATTR_MASK_TPH_MEM_TYPE); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_30 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_DMAH_INIT_ATTR_MASK_TPH_MEM, __pyx_t_32); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_dmah_init_attr_mask = __Pyx_FlagBase('ibv_dmah_init_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_DMAH_INIT_ATTR_MASK_CPU_ID', __Pyx_PyLong_From_enum__ibv_dmah_init_attr_mask(IBV_DMAH_INIT_ATTR_MASK_CPU_ID)),
 *     ('IBV_DMAH_INIT_ATTR_MASK_PH', __Pyx_PyLong_From_enum__ibv_dmah_init_attr_mask(IBV_DMAH_INIT_ATTR_MASK_PH)),
*/
  __pyx_t_32 = __Pyx_PyList_Pack(3, __pyx_t_20, __pyx_t_15, __pyx_t_30); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":28
 *     ('IBV_DMAH_INIT_ATTR_MASK_TPH_MEM_TYPE', __Pyx_PyLong_From_enum__ibv_dmah_init_attr_mask(IBV_DMAH_INIT_ATTR_MASK_TPH_MEM_TYPE)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_30 = __Pyx_Globals(); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_30, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __pyx_t_30 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_2, __pyx_mstate_global->__pyx_n_u_ibv_dmah_init_attr_mask, __pyx_t_32};
    __pyx_t_15 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_30, __pyx_t_15, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_7 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_31, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_15);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_dmah_init_attr_mask, __pyx_t_7) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":35
 * 
 * 
 *     ibv_dmah_init_attr_mask._member_names_ = list(ibv_dmah_init_attr_mask.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_ibv_dmah_init_attr_mask); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_31 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PySequence_ListKeepNew(__pyx_t_31); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_31, __pyx_mstate_global->__pyx_n_u_ibv_dmah_init_attr_mask); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_31, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_7) < (0)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

    /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_event_type = __Pyx_FlagBase('ibv_event_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_EVENT_CQ_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_CQ_ERR)),
 *     ('IBV_EVENT_QP_FATAL', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_FATAL)),
*/
  __pyx_t_7 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_15 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_event_type = __Pyx_FlagBase('ibv_event_type', [
 *     ('IBV_EVENT_CQ_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_CQ_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_EVENT_QP_FATAL', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_FATAL)),
 *     ('IBV_EVENT_QP_REQ_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_REQ_ERR)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_CQ_ERR); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_32 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_EVENT_CQ_ERR, __pyx_t_30); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":25
 * ibv_event_type = __Pyx_FlagBase('ibv_event_type', [
 *     ('IBV_EVENT_CQ_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_CQ_ERR)),
 *     ('IBV_EVENT_QP_FATAL', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_FATAL)),             # <<<<<<<<<<<<<<
 *     ('IBV_EVENT_QP_REQ_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_REQ_ERR)),
 *     ('IBV_EVENT_QP_ACCESS_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_ACCESS_ERR)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_FATAL); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_2 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_EVENT_QP_FATAL, __pyx_t_30); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":26
 *     ('IBV_EVENT_CQ_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_CQ_ERR)),
 *     ('IBV_EVENT_QP_FATAL', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_FATAL)),
 *     ('IBV_EVENT_QP_REQ_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_REQ_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_EVENT_QP_ACCESS_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_ACCESS_ERR)),
 *     ('IBV_EVENT_COMM_EST', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_COMM_EST)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_REQ_ERR); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_20 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_EVENT_QP_REQ_ERR, __pyx_t_30); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":27
 *     ('IBV_EVENT_QP_FATAL', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_FATAL)),
 *     ('IBV_EVENT_QP_REQ_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_REQ_ERR)),
 *     ('IBV_EVENT_QP_ACCESS_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_ACCESS_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_EVENT_COMM_EST', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_COMM_EST)),
 *     ('IBV_EVENT_SQ_DRAINED', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SQ_DRAINED)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_ACCESS_ERR); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_29 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_EVENT_QP_ACCESS_ERR, __pyx_t_30); if (unlikely(!__pyx_t_29)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":28
 *     ('IBV_EVENT_QP_REQ_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_REQ_ERR)),
 *     ('IBV_EVENT_QP_ACCESS_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_ACCESS_ERR)),
 *     ('IBV_EVENT_COMM_EST', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_COMM_EST)),             # <<<<<<<<<<<<<<
 *     ('IBV_EVENT_SQ_DRAINED', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SQ_DRAINED)),
 *     ('IBV_EVENT_PATH_MIG', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PATH_MIG)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_COMM_EST); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_28 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_EVENT_COMM_EST, __pyx_t_30); if (unlikely(!__pyx_t_28)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":29
 *     ('IBV_EVENT_QP_ACCESS_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_ACCESS_ERR)),
 *     ('IBV_EVENT_COMM_EST', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_COMM_EST)),
 *     ('IBV_EVENT_SQ_DRAINED', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SQ_DRAINED)),             # <<<<<<<<<<<<<<
 *     ('IBV_EVENT_PATH_MIG', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PATH_MIG)),
 *     ('IBV_EVENT_PATH_MIG_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PATH_MIG_ERR)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SQ_DRAINED); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_27 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_EVENT_SQ_DRAINED, __pyx_t_30); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":30
 *     ('IBV_EVENT_COMM_EST', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_COMM_EST)),
 *     ('IBV_EVENT_SQ_DRAINED', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SQ_DRAINED)),
 *     ('IBV_EVENT_PATH_MIG', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PATH_MIG)),             # <<<<<<<<<<<<<<
 *     ('IBV_EVENT_PATH_MIG_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PATH_MIG_ERR)),
 *     ('IBV_EVENT_DEVICE_FATAL', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_DEVICE_FATAL)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PATH_MIG); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_26 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_EVENT_PATH_MIG, __pyx_t_30); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":31
 *     ('IBV_EVENT_SQ_DRAINED', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SQ_DRAINED)),
 *     ('IBV_EVENT_PATH_MIG', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PATH_MIG)),
 *     ('IBV_EVENT_PATH_MIG_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PATH_MIG_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_EVENT_DEVICE_FATAL', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_DEVICE_FATAL)),
 *     ('IBV_EVENT_PORT_ACTIVE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PORT_ACTIVE)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PATH_MIG_ERR); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_25 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_EVENT_PATH_MIG_ERR, __pyx_t_30); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":32
 *     ('IBV_EVENT_PATH_MIG', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PATH_MIG)),
 *     ('IBV_EVENT_PATH_MIG_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PATH_MIG_ERR)),
 *     ('IBV_EVENT_DEVICE_FATAL', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_DEVICE_FATAL)),             # <<<<<<<<<<<<<<
 *     ('IBV_EVENT_PORT_ACTIVE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PORT_ACTIVE)),
 *     ('IBV_EVENT_PORT_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PORT_ERR)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_DEVICE_FATAL); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_24 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_EVENT_DEVICE_FATAL, __pyx_t_30); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":33
 *     ('IBV_EVENT_PATH_MIG_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PATH_MIG_ERR)),
 *     ('IBV_EVENT_DEVICE_FATAL', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_DEVICE_FATAL)),
 *     ('IBV_EVENT_PORT_ACTIVE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PORT_ACTIVE)),             # <<<<<<<<<<<<<<
 *     ('IBV_EVENT_PORT_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PORT_ERR)),
 *     ('IBV_EVENT_LID_CHANGE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_LID_CHANGE)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PORT_ACTIVE); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_23 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_EVENT_PORT_ACTIVE, __pyx_t_30); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":34
 *     ('IBV_EVENT_DEVICE_FATAL', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_DEVICE_FATAL)),
 *     ('IBV_EVENT_PORT_ACTIVE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PORT_ACTIVE)),
 *     ('IBV_EVENT_PORT_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PORT_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_EVENT_LID_CHANGE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_LID_CHANGE)),
 *     ('IBV_EVENT_PKEY_CHANGE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PKEY_CHANGE)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PORT_ERR); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_EVENT_PORT_ERR, __pyx_t_30); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":35
 *     ('IBV_EVENT_PORT_ACTIVE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PORT_ACTIVE)),
 *     ('IBV_EVENT_PORT_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PORT_ERR)),
 *     ('IBV_EVENT_LID_CHANGE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_LID_CHANGE)),             # <<<<<<<<<<<<<<
 *     ('IBV_EVENT_PKEY_CHANGE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PKEY_CHANGE)),
 *     ('IBV_EVENT_SM_CHANGE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SM_CHANGE)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_LID_CHANGE); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_21 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_EVENT_LID_CHANGE, __pyx_t_30); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":36
 *     ('IBV_EVENT_PORT_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PORT_ERR)),
 *     ('IBV_EVENT_LID_CHANGE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_LID_CHANGE)),
 *     ('IBV_EVENT_PKEY_CHANGE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PKEY_CHANGE)),             # <<<<<<<<<<<<<<
 *     ('IBV_EVENT_SM_CHANGE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SM_CHANGE)),
 *     ('IBV_EVENT_SRQ_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SRQ_ERR)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PKEY_CHANGE); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_EVENT_PKEY_CHANGE, __pyx_t_30); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":37
 *     ('IBV_EVENT_LID_CHANGE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_LID_CHANGE)),
 *     ('IBV_EVENT_PKEY_CHANGE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PKEY_CHANGE)),
 *     ('IBV_EVENT_SM_CHANGE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SM_CHANGE)),             # <<<<<<<<<<<<<<
 *     ('IBV_EVENT_SRQ_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SRQ_ERR)),
 *     ('IBV_EVENT_SRQ_LIMIT_REACHED', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SRQ_LIMIT_REACHED)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SM_CHANGE); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_EVENT_SM_CHANGE, __pyx_t_30); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":38
 *     ('IBV_EVENT_PKEY_CHANGE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_PKEY_CHANGE)),
 *     ('IBV_EVENT_SM_CHANGE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SM_CHANGE)),
 *     ('IBV_EVENT_SRQ_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SRQ_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_EVENT_SRQ_LIMIT_REACHED', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SRQ_LIMIT_REACHED)),
 *     ('IBV_EVENT_QP_LAST_WQE_REACHED', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_LAST_WQE_REACHED)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SRQ_ERR); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_6 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_EVENT_SRQ_ERR, __pyx_t_30); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":39
 *     ('IBV_EVENT_SM_CHANGE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SM_CHANGE)),
 *     ('IBV_EVENT_SRQ_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SRQ_ERR)),
 *     ('IBV_EVENT_SRQ_LIMIT_REACHED', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SRQ_LIMIT_REACHED)),             # <<<<<<<<<<<<<<
 *     ('IBV_EVENT_QP_LAST_WQE_REACHED', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_LAST_WQE_REACHED)),
 *     ('IBV_EVENT_CLIENT_REREGISTER', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_CLIENT_REREGISTER)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SRQ_LIMIT_REACHED); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_14 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_EVENT_SRQ_LIMIT_REACHED, __pyx_t_30); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":40
 *     ('IBV_EVENT_SRQ_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SRQ_ERR)),
 *     ('IBV_EVENT_SRQ_LIMIT_REACHED', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SRQ_LIMIT_REACHED)),
 *     ('IBV_EVENT_QP_LAST_WQE_REACHED', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_LAST_WQE_REACHED)),             # <<<<<<<<<<<<<<
 *     ('IBV_EVENT_CLIENT_REREGISTER', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_CLIENT_REREGISTER)),
 *     ('IBV_EVENT_GID_CHANGE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_GID_CHANGE)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_LAST_WQE_REACHED); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_13 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_EVENT_QP_LAST_WQE_REACHED, __pyx_t_30); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":41
 *     ('IBV_EVENT_SRQ_LIMIT_REACHED', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_SRQ_LIMIT_REACHED)),
 *     ('IBV_EVENT_QP_LAST_WQE_REACHED', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_LAST_WQE_REACHED)),
 *     ('IBV_EVENT_CLIENT_REREGISTER', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_CLIENT_REREGISTER)),             # <<<<<<<<<<<<<<
 *     ('IBV_EVENT_GID_CHANGE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_GID_CHANGE)),
 *     ('IBV_EVENT_WQ_FATAL', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_WQ_FATAL)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_CLIENT_REREGISTER); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_12 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_EVENT_CLIENT_REREGISTER, __pyx_t_30); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":42
 *     ('IBV_EVENT_QP_LAST_WQE_REACHED', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_LAST_WQE_REACHED)),
 *     ('IBV_EVENT_CLIENT_REREGISTER', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_CLIENT_REREGISTER)),
 *     ('IBV_EVENT_GID_CHANGE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_GID_CHANGE)),             # <<<<<<<<<<<<<<
 *     ('IBV_EVENT_WQ_FATAL', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_WQ_FATAL)),
 * 
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_GID_CHANGE); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_EVENT_GID_CHANGE, __pyx_t_30); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":43
 *     ('IBV_EVENT_CLIENT_REREGISTER', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_CLIENT_REREGISTER)),
 *     ('IBV_EVENT_GID_CHANGE', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_GID_CHANGE)),
 *     ('IBV_EVENT_WQ_FATAL', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_WQ_FATAL)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_WQ_FATAL); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_10 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_EVENT_WQ_FATAL, __pyx_t_30); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_event_type = __Pyx_FlagBase('ibv_event_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_EVENT_CQ_ERR', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_CQ_ERR)),
 *     ('IBV_EVENT_QP_FATAL', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_QP_FATAL)),
*/
  __pyx_t_30 = __Pyx_PyList_Pack(20, __pyx_t_32, __pyx_t_2, __pyx_t_20, __pyx_t_29, __pyx_t_28, __pyx_t_27, __pyx_t_26, __pyx_t_25, __pyx_t_24, __pyx_t_23, __pyx_t_22, __pyx_t_21, __pyx_t_5, __pyx_t_4, __pyx_t_6, __pyx_t_14, __pyx_t_13, __pyx_t_12, __pyx_t_11, __pyx_t_10); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":45
 *     ('IBV_EVENT_WQ_FATAL', __Pyx_PyLong_From_enum__ibv_event_type(IBV_EVENT_WQ_FATAL)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_10 = __Pyx_Globals(); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_7, __pyx_mstate_global->__pyx_n_u_ibv_event_type, __pyx_t_30};
    __pyx_t_11 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_10, __pyx_t_11, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_31 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_15, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_11);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_event_type, __pyx_t_31) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":47
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":52
 * 
 * 
 *     ibv_event_type._member_names_ = list(ibv_event_type.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_31, __pyx_mstate_global->__pyx_n_u_ibv_event_type); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
    __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_31, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __pyx_t_31 = __Pyx_PySequence_ListKeepNew(__pyx_t_15); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_ibv_event_type); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_31) < (0)) __PYX_ERR(1, 52, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":47
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_flow_attr_type = __Pyx_FlagBase('ibv_flow_attr_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_ATTR_NORMAL', __Pyx_PyLong_From_enum__ibv_flow_attr_type(IBV_FLOW_ATTR_NORMAL)),
 *     ('IBV_FLOW_ATTR_ALL_DEFAULT', __Pyx_PyLong_From_enum__ibv_flow_attr_type(IBV_FLOW_ATTR_ALL_DEFAULT)),
*/
  __pyx_t_31 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_11 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_flow_attr_type = __Pyx_FlagBase('ibv_flow_attr_type', [
 *     ('IBV_FLOW_ATTR_NORMAL', __Pyx_PyLong_From_enum__ibv_flow_attr_type(IBV_FLOW_ATTR_NORMAL)),             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_ATTR_ALL_DEFAULT', __Pyx_PyLong_From_enum__ibv_flow_attr_type(IBV_FLOW_ATTR_ALL_DEFAULT)),
 *     ('IBV_FLOW_ATTR_MC_DEFAULT', __Pyx_PyLong_From_enum__ibv_flow_attr_type(IBV_FLOW_ATTR_MC_DEFAULT)),
*/
  __pyx_t_10 = __Pyx_PyLong_From_enum__ibv_flow_attr_type(IBV_FLOW_ATTR_NORMAL); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_30 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLOW_ATTR_NORMAL, __pyx_t_10); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":25
 * ibv_flow_attr_type = __Pyx_FlagBase('ibv_flow_attr_type', [
 *     ('IBV_FLOW_ATTR_NORMAL', __Pyx_PyLong_From_enum__ibv_flow_attr_type(IBV_FLOW_ATTR_NORMAL)),
 *     ('IBV_FLOW_ATTR_ALL_DEFAULT', __Pyx_PyLong_From_enum__ibv_flow_attr_type(IBV_FLOW_ATTR_ALL_DEFAULT)),             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_ATTR_MC_DEFAULT', __Pyx_PyLong_From_enum__ibv_flow_attr_type(IBV_FLOW_ATTR_MC_DEFAULT)),
 *     ('IBV_FLOW_ATTR_SNIFFER', __Pyx_PyLong_From_enum__ibv_flow_attr_type(IBV_FLOW_ATTR_SNIFFER)),
*/
  __pyx_t_10 = __Pyx_PyLong_From_enum__ibv_flow_attr_type(IBV_FLOW_ATTR_ALL_DEFAULT); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_7 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLOW_ATTR_ALL_DEFAULT, __pyx_t_10); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":26
 *     ('IBV_FLOW_ATTR_NORMAL', __Pyx_PyLong_From_enum__ibv_flow_attr_type(IBV_FLOW_ATTR_NORMAL)),
 *     ('IBV_FLOW_ATTR_ALL_DEFAULT', __Pyx_PyLong_From_enum__ibv_flow_attr_type(IBV_FLOW_ATTR_ALL_DEFAULT)),
 *     ('IBV_FLOW_ATTR_MC_DEFAULT', __Pyx_PyLong_From_enum__ibv_flow_attr_type(IBV_FLOW_ATTR_MC_DEFAULT)),             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_ATTR_SNIFFER', __Pyx_PyLong_From_enum__ibv_flow_attr_type(IBV_FLOW_ATTR_SNIFFER)),
 * 
*/
  __pyx_t_10 = __Pyx_PyLong_From_enum__ibv_flow_attr_type(IBV_FLOW_ATTR_MC_DEFAULT); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_12 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLOW_ATTR_MC_DEFAULT, __pyx_t_10); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":27
 *     ('IBV_FLOW_ATTR_ALL_DEFAULT', __Pyx_PyLong_From_enum__ibv_flow_attr_type(IBV_FLOW_ATTR_ALL_DEFAULT)),
 *     ('IBV_FLOW_ATTR_MC_DEFAULT', __Pyx_PyLong_From_enum__ibv_flow_attr_type(IBV_FLOW_ATTR_MC_DEFAULT)),
 *     ('IBV_FLOW_ATTR_SNIFFER', __Pyx_PyLong_From_enum__ibv_flow_attr_type(IBV_FLOW_ATTR_SNIFFER)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_10 = __Pyx_PyLong_From_enum__ibv_flow_attr_type(IBV_FLOW_ATTR_SNIFFER); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_13 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLOW_ATTR_SNIFFER, __pyx_t_10); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_flow_attr_type = __Pyx_FlagBase('ibv_flow_attr_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_ATTR_NORMAL', __Pyx_PyLong_From_enum__ibv_flow_attr_type(IBV_FLOW_ATTR_NORMAL)),
 *     ('IBV_FLOW_ATTR_ALL_DEFAULT', __Pyx_PyLong_From_enum__ibv_flow_attr_type(IBV_FLOW_ATTR_ALL_DEFAULT)),
*/
  __pyx_t_10 = __Pyx_PyList_Pack(4, __pyx_t_30, __pyx_t_7, __pyx_t_12, __pyx_t_13); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "EnumType":29
 *     ('IBV_FLOW_ATTR_SNIFFER', __Pyx_PyLong_From_enum__ibv_flow_attr_type(IBV_FLOW_ATTR_SNIFFER)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_13 = __Pyx_Globals(); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_31, __pyx_mstate_global->__pyx_n_u_ibv_flow_attr_type, __pyx_t_10};
    __pyx_t_12 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_13, __pyx_t_12, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_15 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_11, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_12);
    __Pyx_XDECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_flow_attr_type, __pyx_t_15) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":31
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":36
 * 
 * 
 *     ibv_flow_attr_type._member_names_ = list(ibv_flow_attr_type.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_ibv_flow_attr_type); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __pyx_t_15 = __Pyx_PySequence_ListKeepNew(__pyx_t_11); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_ibv_flow_attr_type); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_15) < (0)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "EnumType":31
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_flow_flags = __Pyx_FlagBase('ibv_flow_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_ATTR_FLAGS_DONT_TRAP', __Pyx_PyLong_From_enum__ibv_flow_flags(IBV_FLOW_ATTR_FLAGS_DONT_TRAP)),
 *     ('IBV_FLOW_ATTR_FLAGS_EGRESS', __Pyx_PyLong_From_enum__ibv_flow_flags(IBV_FLOW_ATTR_FLAGS_EGRESS)),
*/
  __pyx_t_15 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_12 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_flow_flags = __Pyx_FlagBase('ibv_flow_flags', [
 *     ('IBV_FLOW_ATTR_FLAGS_DONT_TRAP', __Pyx_PyLong_From_enum__ibv_flow_flags(IBV_FLOW_ATTR_FLAGS_DONT_TRAP)),             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_ATTR_FLAGS_EGRESS', __Pyx_PyLong_From_enum__ibv_flow_flags(IBV_FLOW_ATTR_FLAGS_EGRESS)),
 * 
*/
  __pyx_t_13 = __Pyx_PyLong_From_enum__ibv_flow_flags(IBV_FLOW_ATTR_FLAGS_DONT_TRAP); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_10 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLOW_ATTR_FLAGS_DONT_TRAP, __pyx_t_13); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "EnumType":25
 * ibv_flow_flags = __Pyx_FlagBase('ibv_flow_flags', [
 *     ('IBV_FLOW_ATTR_FLAGS_DONT_TRAP', __Pyx_PyLong_From_enum__ibv_flow_flags(IBV_FLOW_ATTR_FLAGS_DONT_TRAP)),
 *     ('IBV_FLOW_ATTR_FLAGS_EGRESS', __Pyx_PyLong_From_enum__ibv_flow_flags(IBV_FLOW_ATTR_FLAGS_EGRESS)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_13 = __Pyx_PyLong_From_enum__ibv_flow_flags(IBV_FLOW_ATTR_FLAGS_EGRESS); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_31 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLOW_ATTR_FLAGS_EGRESS, __pyx_t_13); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_flow_flags = __Pyx_FlagBase('ibv_flow_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_ATTR_FLAGS_DONT_TRAP', __Pyx_PyLong_From_enum__ibv_flow_flags(IBV_FLOW_ATTR_FLAGS_DONT_TRAP)),
 *     ('IBV_FLOW_ATTR_FLAGS_EGRESS', __Pyx_PyLong_From_enum__ibv_flow_flags(IBV_FLOW_ATTR_FLAGS_EGRESS)),
*/
  __pyx_t_13 = __Pyx_PyList_Pack(2, __pyx_t_10, __pyx_t_31); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":27
 *     ('IBV_FLOW_ATTR_FLAGS_EGRESS', __Pyx_PyLong_From_enum__ibv_flow_flags(IBV_FLOW_ATTR_FLAGS_EGRESS)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_31 = __Pyx_Globals(); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_31, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __pyx_t_31 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_15, __pyx_mstate_global->__pyx_n_u_ibv_flow_flags, __pyx_t_13};
    __pyx_t_10 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_31, __pyx_t_10, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_11 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_12, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_flow_flags, __pyx_t_11) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":34
 * 
 * 
 *     ibv_flow_flags._member_names_ = list(ibv_flow_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_ibv_flow_flags); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PySequence_ListKeepNew(__pyx_t_12); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_ibv_flow_flags); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_11) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_flow_spec_type = __Pyx_FlagBase('ibv_flow_spec_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_SPEC_ETH', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ETH)),
 *     ('IBV_FLOW_SPEC_IPV4', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_IPV4)),
*/
  __pyx_t_11 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_10 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_flow_spec_type = __Pyx_FlagBase('ibv_flow_spec_type', [
 *     ('IBV_FLOW_SPEC_ETH', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ETH)),             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_SPEC_IPV4', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_IPV4)),
 *     ('IBV_FLOW_SPEC_IPV6', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_IPV6)),
*/
  __pyx_t_31 = __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ETH); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_13 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLOW_SPEC_ETH, __pyx_t_31); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":25
 * ibv_flow_spec_type = __Pyx_FlagBase('ibv_flow_spec_type', [
 *     ('IBV_FLOW_SPEC_ETH', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ETH)),
 *     ('IBV_FLOW_SPEC_IPV4', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_IPV4)),             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_SPEC_IPV6', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_IPV6)),
 *     ('IBV_FLOW_SPEC_IPV4_EXT', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_IPV4_EXT)),
*/
  __pyx_t_31 = __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_IPV4); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLOW_SPEC_IPV4, __pyx_t_31); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":26
 *     ('IBV_FLOW_SPEC_ETH', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ETH)),
 *     ('IBV_FLOW_SPEC_IPV4', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_IPV4)),
 *     ('IBV_FLOW_SPEC_IPV6', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_IPV6)),             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_SPEC_IPV4_EXT', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_IPV4_EXT)),
 *     ('IBV_FLOW_SPEC_ESP', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ESP)),
*/
  __pyx_t_31 = __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_IPV6); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_7 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLOW_SPEC_IPV6, __pyx_t_31); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":27
 *     ('IBV_FLOW_SPEC_IPV4', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_IPV4)),
 *     ('IBV_FLOW_SPEC_IPV6', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_IPV6)),
 *     ('IBV_FLOW_SPEC_IPV4_EXT', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_IPV4_EXT)),             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_SPEC_ESP', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ESP)),
 *     ('IBV_FLOW_SPEC_TCP', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_TCP)),
*/
  __pyx_t_31 = __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_IPV4_EXT); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_30 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLOW_SPEC_IPV4_EXT, __pyx_t_31); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":28
 *     ('IBV_FLOW_SPEC_IPV6', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_IPV6)),
 *     ('IBV_FLOW_SPEC_IPV4_EXT', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_IPV4_EXT)),
 *     ('IBV_FLOW_SPEC_ESP', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ESP)),             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_SPEC_TCP', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_TCP)),
 *     ('IBV_FLOW_SPEC_UDP', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_UDP)),
*/
  __pyx_t_31 = __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ESP); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_14 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLOW_SPEC_ESP, __pyx_t_31); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":29
 *     ('IBV_FLOW_SPEC_IPV4_EXT', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_IPV4_EXT)),
 *     ('IBV_FLOW_SPEC_ESP', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ESP)),
 *     ('IBV_FLOW_SPEC_TCP', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_TCP)),             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_SPEC_UDP', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_UDP)),
 *     ('IBV_FLOW_SPEC_VXLAN_TUNNEL', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_VXLAN_TUNNEL)),
*/
  __pyx_t_31 = __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_TCP); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_6 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLOW_SPEC_TCP, __pyx_t_31); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":30
 *     ('IBV_FLOW_SPEC_ESP', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ESP)),
 *     ('IBV_FLOW_SPEC_TCP', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_TCP)),
 *     ('IBV_FLOW_SPEC_UDP', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_UDP)),             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_SPEC_VXLAN_TUNNEL', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_VXLAN_TUNNEL)),
 *     ('IBV_FLOW_SPEC_GRE', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_GRE)),
*/
  __pyx_t_31 = __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_UDP); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLOW_SPEC_UDP, __pyx_t_31); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":31
 *     ('IBV_FLOW_SPEC_TCP', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_TCP)),
 *     ('IBV_FLOW_SPEC_UDP', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_UDP)),
 *     ('IBV_FLOW_SPEC_VXLAN_TUNNEL', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_VXLAN_TUNNEL)),             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_SPEC_GRE', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_GRE)),
 *     ('IBV_FLOW_SPEC_MPLS', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_MPLS)),
*/
  __pyx_t_31 = __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_VXLAN_TUNNEL); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLOW_SPEC_VXLAN_TUNNEL, __pyx_t_31); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":32
 *     ('IBV_FLOW_SPEC_UDP', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_UDP)),
 *     ('IBV_FLOW_SPEC_VXLAN_TUNNEL', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_VXLAN_TUNNEL)),
 *     ('IBV_FLOW_SPEC_GRE', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_GRE)),             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_SPEC_MPLS', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_MPLS)),
 *     ('IBV_FLOW_SPEC_INNER', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_INNER)),
*/
  __pyx_t_31 = __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_GRE); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_21 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLOW_SPEC_GRE, __pyx_t_31); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":33
 *     ('IBV_FLOW_SPEC_VXLAN_TUNNEL', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_VXLAN_TUNNEL)),
 *     ('IBV_FLOW_SPEC_GRE', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_GRE)),
 *     ('IBV_FLOW_SPEC_MPLS', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_MPLS)),             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_SPEC_INNER', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_INNER)),
 *     ('IBV_FLOW_SPEC_ACTION_TAG', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ACTION_TAG)),
*/
  __pyx_t_31 = __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_MPLS); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLOW_SPEC_MPLS, __pyx_t_31); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":34
 *     ('IBV_FLOW_SPEC_GRE', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_GRE)),
 *     ('IBV_FLOW_SPEC_MPLS', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_MPLS)),
 *     ('IBV_FLOW_SPEC_INNER', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_INNER)),             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_SPEC_ACTION_TAG', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ACTION_TAG)),
 *     ('IBV_FLOW_SPEC_ACTION_DROP', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ACTION_DROP)),
*/
  __pyx_t_31 = __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_INNER); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_23 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLOW_SPEC_INNER, __pyx_t_31); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":35
 *     ('IBV_FLOW_SPEC_MPLS', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_MPLS)),
 *     ('IBV_FLOW_SPEC_INNER', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_INNER)),
 *     ('IBV_FLOW_SPEC_ACTION_TAG', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ACTION_TAG)),             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_SPEC_ACTION_DROP', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ACTION_DROP)),
 *     ('IBV_FLOW_SPEC_ACTION_HANDLE', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ACTION_HANDLE)),
*/
  __pyx_t_31 = __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ACTION_TAG); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_24 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLOW_SPEC_ACTION_TAG, __pyx_t_31); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":36
 *     ('IBV_FLOW_SPEC_INNER', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_INNER)),
 *     ('IBV_FLOW_SPEC_ACTION_TAG', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ACTION_TAG)),
 *     ('IBV_FLOW_SPEC_ACTION_DROP', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ACTION_DROP)),             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_SPEC_ACTION_HANDLE', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ACTION_HANDLE)),
 *     ('IBV_FLOW_SPEC_ACTION_COUNT', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ACTION_COUNT)),
*/
  __pyx_t_31 = __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ACTION_DROP); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_25 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLOW_SPEC_ACTION_DROP, __pyx_t_31); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":37
 *     ('IBV_FLOW_SPEC_ACTION_TAG', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ACTION_TAG)),
 *     ('IBV_FLOW_SPEC_ACTION_DROP', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ACTION_DROP)),
 *     ('IBV_FLOW_SPEC_ACTION_HANDLE', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ACTION_HANDLE)),             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_SPEC_ACTION_COUNT', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ACTION_COUNT)),
 * 
*/
  __pyx_t_31 = __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ACTION_HANDLE); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_26 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLOW_SPEC_ACTION_HANDLE, __pyx_t_31); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":38
 *     ('IBV_FLOW_SPEC_ACTION_DROP', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ACTION_DROP)),
 *     ('IBV_FLOW_SPEC_ACTION_HANDLE', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ACTION_HANDLE)),
 *     ('IBV_FLOW_SPEC_ACTION_COUNT', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ACTION_COUNT)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_31 = __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ACTION_COUNT); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_27 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLOW_SPEC_ACTION_COUNT, __pyx_t_31); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_flow_spec_type = __Pyx_FlagBase('ibv_flow_spec_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_FLOW_SPEC_ETH', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ETH)),
 *     ('IBV_FLOW_SPEC_IPV4', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_IPV4)),
*/
  __pyx_t_31 = __Pyx_PyList_Pack(15, __pyx_t_13, __pyx_t_15, __pyx_t_7, __pyx_t_30, __pyx_t_14, __pyx_t_6, __pyx_t_4, __pyx_t_5, __pyx_t_21, __pyx_t_22, __pyx_t_23, __pyx_t_24, __pyx_t_25, __pyx_t_26, __pyx_t_27); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

  /* "EnumType":40
 *     ('IBV_FLOW_SPEC_ACTION_COUNT', __Pyx_PyLong_From_enum__ibv_flow_spec_type(IBV_FLOW_SPEC_ACTION_COUNT)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_27 = __Pyx_Globals(); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __pyx_t_26 = __Pyx_PyObject_GetAttrStr(__pyx_t_27, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __pyx_t_27 = __Pyx_PyObject_Call(__pyx_t_26, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_11, __pyx_mstate_global->__pyx_n_u_ibv_flow_spec_type, __pyx_t_31};
    __pyx_t_26 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_27, __pyx_t_26, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_12 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_10, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_26);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_flow_spec_type, __pyx_t_12) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "EnumType":42
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":47
 * 
 * 
 *     ibv_flow_spec_type._member_names_ = list(ibv_flow_spec_type.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_ibv_flow_spec_type); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = __Pyx_PySequence_ListKeepNew(__pyx_t_10); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_ibv_flow_spec_type); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_12) < (0)) __PYX_ERR(1, 47, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "EnumType":42
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_fork_status = __Pyx_FlagBase('ibv_fork_status', [             # <<<<<<<<<<<<<<
 *     ('IBV_FORK_DISABLED', __Pyx_PyLong_From_enum__ibv_fork_status(IBV_FORK_DISABLED)),
 *     ('IBV_FORK_ENABLED', __Pyx_PyLong_From_enum__ibv_fork_status(IBV_FORK_ENABLED)),
*/
  __pyx_t_12 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_26 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_fork_status = __Pyx_FlagBase('ibv_fork_status', [
 *     ('IBV_FORK_DISABLED', __Pyx_PyLong_From_enum__ibv_fork_status(IBV_FORK_DISABLED)),             # <<<<<<<<<<<<<<
 *     ('IBV_FORK_ENABLED', __Pyx_PyLong_From_enum__ibv_fork_status(IBV_FORK_ENABLED)),
 *     ('IBV_FORK_UNNEEDED', __Pyx_PyLong_From_enum__ibv_fork_status(IBV_FORK_UNNEEDED)),
*/
  __pyx_t_27 = __Pyx_PyLong_From_enum__ibv_fork_status(IBV_FORK_DISABLED); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __pyx_t_31 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FORK_DISABLED, __pyx_t_27); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

  /* "EnumType":25
 * ibv_fork_status = __Pyx_FlagBase('ibv_fork_status', [
 *     ('IBV_FORK_DISABLED', __Pyx_PyLong_From_enum__ibv_fork_status(IBV_FORK_DISABLED)),
 *     ('IBV_FORK_ENABLED', __Pyx_PyLong_From_enum__ibv_fork_status(IBV_FORK_ENABLED)),             # <<<<<<<<<<<<<<
 *     ('IBV_FORK_UNNEEDED', __Pyx_PyLong_From_enum__ibv_fork_status(IBV_FORK_UNNEEDED)),
 * 
*/
  __pyx_t_27 = __Pyx_PyLong_From_enum__ibv_fork_status(IBV_FORK_ENABLED); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FORK_ENABLED, __pyx_t_27); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

  /* "EnumType":26
 *     ('IBV_FORK_DISABLED', __Pyx_PyLong_From_enum__ibv_fork_status(IBV_FORK_DISABLED)),
 *     ('IBV_FORK_ENABLED', __Pyx_PyLong_From_enum__ibv_fork_status(IBV_FORK_ENABLED)),
 *     ('IBV_FORK_UNNEEDED', __Pyx_PyLong_From_enum__ibv_fork_status(IBV_FORK_UNNEEDED)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_27 = __Pyx_PyLong_From_enum__ibv_fork_status(IBV_FORK_UNNEEDED); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __pyx_t_25 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FORK_UNNEEDED, __pyx_t_27); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_fork_status = __Pyx_FlagBase('ibv_fork_status', [             # <<<<<<<<<<<<<<
 *     ('IBV_FORK_DISABLED', __Pyx_PyLong_From_enum__ibv_fork_status(IBV_FORK_DISABLED)),
 *     ('IBV_FORK_ENABLED', __Pyx_PyLong_From_enum__ibv_fork_status(IBV_FORK_ENABLED)),
*/
  __pyx_t_27 = __Pyx_PyList_Pack(3, __pyx_t_31, __pyx_t_11, __pyx_t_25); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":28
 *     ('IBV_FORK_UNNEEDED', __Pyx_PyLong_From_enum__ibv_fork_status(IBV_FORK_UNNEEDED)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_25 = __Pyx_Globals(); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_25, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __pyx_t_25 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_12, __pyx_mstate_global->__pyx_n_u_ibv_fork_status, __pyx_t_27};
    __pyx_t_11 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_25, __pyx_t_11, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_10 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_26, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_11);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_fork_status, __pyx_t_10) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":35
 * 
 * 
 *     ibv_fork_status._member_names_ = list(ibv_fork_status.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_ibv_fork_status); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_26 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PySequence_ListKeepNew(__pyx_t_26); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_26, __pyx_mstate_global->__pyx_n_u_ibv_fork_status); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_26, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_10) < (0)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

    /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_gid_type = __Pyx_FlagBase('ibv_gid_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_GID_TYPE_IB', __Pyx_PyLong_From_enum__ibv_gid_type(IBV_GID_TYPE_IB)),
 *     ('IBV_GID_TYPE_ROCE_V1', __Pyx_PyLong_From_enum__ibv_gid_type(IBV_GID_TYPE_ROCE_V1)),
*/
  __pyx_t_10 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_11 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_gid_type = __Pyx_FlagBase('ibv_gid_type', [
 *     ('IBV_GID_TYPE_IB', __Pyx_PyLong_From_enum__ibv_gid_type(IBV_GID_TYPE_IB)),             # <<<<<<<<<<<<<<
 *     ('IBV_GID_TYPE_ROCE_V1', __Pyx_PyLong_From_enum__ibv_gid_type(IBV_GID_TYPE_ROCE_V1)),
 *     ('IBV_GID_TYPE_ROCE_V2', __Pyx_PyLong_From_enum__ibv_gid_type(IBV_GID_TYPE_ROCE_V2)),
*/
  __pyx_t_25 = __Pyx_PyLong_From_enum__ibv_gid_type(IBV_GID_TYPE_IB); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_27 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_GID_TYPE_IB, __pyx_t_25); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":25
 * ibv_gid_type = __Pyx_FlagBase('ibv_gid_type', [
 *     ('IBV_GID_TYPE_IB', __Pyx_PyLong_From_enum__ibv_gid_type(IBV_GID_TYPE_IB)),
 *     ('IBV_GID_TYPE_ROCE_V1', __Pyx_PyLong_From_enum__ibv_gid_type(IBV_GID_TYPE_ROCE_V1)),             # <<<<<<<<<<<<<<
 *     ('IBV_GID_TYPE_ROCE_V2', __Pyx_PyLong_From_enum__ibv_gid_type(IBV_GID_TYPE_ROCE_V2)),
 * 
*/
  __pyx_t_25 = __Pyx_PyLong_From_enum__ibv_gid_type(IBV_GID_TYPE_ROCE_V1); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_12 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_GID_TYPE_ROCE_V1, __pyx_t_25); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":26
 *     ('IBV_GID_TYPE_IB', __Pyx_PyLong_From_enum__ibv_gid_type(IBV_GID_TYPE_IB)),
 *     ('IBV_GID_TYPE_ROCE_V1', __Pyx_PyLong_From_enum__ibv_gid_type(IBV_GID_TYPE_ROCE_V1)),
 *     ('IBV_GID_TYPE_ROCE_V2', __Pyx_PyLong_From_enum__ibv_gid_type(IBV_GID_TYPE_ROCE_V2)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_25 = __Pyx_PyLong_From_enum__ibv_gid_type(IBV_GID_TYPE_ROCE_V2); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_31 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_GID_TYPE_ROCE_V2, __pyx_t_25); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_gid_type = __Pyx_FlagBase('ibv_gid_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_GID_TYPE_IB', __Pyx_PyLong_From_enum__ibv_gid_type(IBV_GID_TYPE_IB)),
 *     ('IBV_GID_TYPE_ROCE_V1', __Pyx_PyLong_From_enum__ibv_gid_type(IBV_GID_TYPE_ROCE_V1)),
*/
  __pyx_t_25 = __Pyx_PyList_Pack(3, __pyx_t_27, __pyx_t_12, __pyx_t_31); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":28
 *     ('IBV_GID_TYPE_ROCE_V2', __Pyx_PyLong_From_enum__ibv_gid_type(IBV_GID_TYPE_ROCE_V2)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_31 = __Pyx_Globals(); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_31, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __pyx_t_31 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_10, __pyx_mstate_global->__pyx_n_u_ibv_gid_type, __pyx_t_25};
    __pyx_t_12 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_31, __pyx_t_12, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_26 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_11, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_12);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_gid_type, __pyx_t_26) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":35
 * 
 * 
 *     ibv_gid_type._member_names_ = list(ibv_gid_type.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_26, __pyx_mstate_global->__pyx_n_u_ibv_gid_type); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_26, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __pyx_t_26 = __Pyx_PySequence_ListKeepNew(__pyx_t_11); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_ibv_gid_type); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_26) < (0)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_gid_type_sysfs = __Pyx_FlagBase('ibv_gid_type_sysfs', [             # <<<<<<<<<<<<<<
 *     ('IBV_GID_TYPE_SYSFS_IB_ROCE_V1', __Pyx_PyLong_From_enum__ibv_gid_type_sysfs(IBV_GID_TYPE_SYSFS_IB_ROCE_V1)),
 *     ('IBV_GID_TYPE_SYSFS_ROCE_V2', __Pyx_PyLong_From_enum__ibv_gid_type_sysfs(IBV_GID_TYPE_SYSFS_ROCE_V2)),
*/
  __pyx_t_26 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_12 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_gid_type_sysfs = __Pyx_FlagBase('ibv_gid_type_sysfs', [
 *     ('IBV_GID_TYPE_SYSFS_IB_ROCE_V1', __Pyx_PyLong_From_enum__ibv_gid_type_sysfs(IBV_GID_TYPE_SYSFS_IB_ROCE_V1)),             # <<<<<<<<<<<<<<
 *     ('IBV_GID_TYPE_SYSFS_ROCE_V2', __Pyx_PyLong_From_enum__ibv_gid_type_sysfs(IBV_GID_TYPE_SYSFS_ROCE_V2)),
 * 
*/
  __pyx_t_31 = __Pyx_PyLong_From_enum__ibv_gid_type_sysfs(IBV_GID_TYPE_SYSFS_IB_ROCE_V1); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_25 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_GID_TYPE_SYSFS_IB_ROCE_V1, __pyx_t_31); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":25
 * ibv_gid_type_sysfs = __Pyx_FlagBase('ibv_gid_type_sysfs', [
 *     ('IBV_GID_TYPE_SYSFS_IB_ROCE_V1', __Pyx_PyLong_From_enum__ibv_gid_type_sysfs(IBV_GID_TYPE_SYSFS_IB_ROCE_V1)),
 *     ('IBV_GID_TYPE_SYSFS_ROCE_V2', __Pyx_PyLong_From_enum__ibv_gid_type_sysfs(IBV_GID_TYPE_SYSFS_ROCE_V2)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_31 = __Pyx_PyLong_From_enum__ibv_gid_type_sysfs(IBV_GID_TYPE_SYSFS_ROCE_V2); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_10 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_GID_TYPE_SYSFS_ROCE_V2, __pyx_t_31); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_gid_type_sysfs = __Pyx_FlagBase('ibv_gid_type_sysfs', [             # <<<<<<<<<<<<<<
 *     ('IBV_GID_TYPE_SYSFS_IB_ROCE_V1', __Pyx_PyLong_From_enum__ibv_gid_type_sysfs(IBV_GID_TYPE_SYSFS_IB_ROCE_V1)),
 *     ('IBV_GID_TYPE_SYSFS_ROCE_V2', __Pyx_PyLong_From_enum__ibv_gid_type_sysfs(IBV_GID_TYPE_SYSFS_ROCE_V2)),
*/
  __pyx_t_31 = __Pyx_PyList_Pack(2, __pyx_t_25, __pyx_t_10); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":27
 *     ('IBV_GID_TYPE_SYSFS_ROCE_V2', __Pyx_PyLong_From_enum__ibv_gid_type_sysfs(IBV_GID_TYPE_SYSFS_ROCE_V2)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_10 = __Pyx_Globals(); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_25 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_25, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_26, __pyx_mstate_global->__pyx_n_u_ibv_gid_type_sysfs, __pyx_t_31};
    __pyx_t_25 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_10, __pyx_t_25, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_11 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_12, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_25);
    __Pyx_XDECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_gid_type_sysfs, __pyx_t_11) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":34
 * 
 * 
 *     ibv_gid_type_sysfs._member_names_ = list(ibv_gid_type_sysfs.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_ibv_gid_type_sysfs); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PySequence_ListKeepNew(__pyx_t_12); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_ibv_gid_type_sysfs); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_11) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_mig_state = __Pyx_FlagBase('ibv_mig_state', [             # <<<<<<<<<<<<<<
 *     ('IBV_MIG_MIGRATED', __Pyx_PyLong_From_enum__ibv_mig_state(IBV_MIG_MIGRATED)),
 *     ('IBV_MIG_REARM', __Pyx_PyLong_From_enum__ibv_mig_state(IBV_MIG_REARM)),
*/
  __pyx_t_11 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_25 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_mig_state = __Pyx_FlagBase('ibv_mig_state', [
 *     ('IBV_MIG_MIGRATED', __Pyx_PyLong_From_enum__ibv_mig_state(IBV_MIG_MIGRATED)),             # <<<<<<<<<<<<<<
 *     ('IBV_MIG_REARM', __Pyx_PyLong_From_enum__ibv_mig_state(IBV_MIG_REARM)),
 *     ('IBV_MIG_ARMED', __Pyx_PyLong_From_enum__ibv_mig_state(IBV_MIG_ARMED)),
*/
  __pyx_t_10 = __Pyx_PyLong_From_enum__ibv_mig_state(IBV_MIG_MIGRATED); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_31 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_MIG_MIGRATED, __pyx_t_10); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":25
 * ibv_mig_state = __Pyx_FlagBase('ibv_mig_state', [
 *     ('IBV_MIG_MIGRATED', __Pyx_PyLong_From_enum__ibv_mig_state(IBV_MIG_MIGRATED)),
 *     ('IBV_MIG_REARM', __Pyx_PyLong_From_enum__ibv_mig_state(IBV_MIG_REARM)),             # <<<<<<<<<<<<<<
 *     ('IBV_MIG_ARMED', __Pyx_PyLong_From_enum__ibv_mig_state(IBV_MIG_ARMED)),
 * 
*/
  __pyx_t_10 = __Pyx_PyLong_From_enum__ibv_mig_state(IBV_MIG_REARM); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_26 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_MIG_REARM, __pyx_t_10); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":26
 *     ('IBV_MIG_MIGRATED', __Pyx_PyLong_From_enum__ibv_mig_state(IBV_MIG_MIGRATED)),
 *     ('IBV_MIG_REARM', __Pyx_PyLong_From_enum__ibv_mig_state(IBV_MIG_REARM)),
 *     ('IBV_MIG_ARMED', __Pyx_PyLong_From_enum__ibv_mig_state(IBV_MIG_ARMED)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_10 = __Pyx_PyLong_From_enum__ibv_mig_state(IBV_MIG_ARMED); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_27 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_MIG_ARMED, __pyx_t_10); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_mig_state = __Pyx_FlagBase('ibv_mig_state', [             # <<<<<<<<<<<<<<
 *     ('IBV_MIG_MIGRATED', __Pyx_PyLong_From_enum__ibv_mig_state(IBV_MIG_MIGRATED)),
 *     ('IBV_MIG_REARM', __Pyx_PyLong_From_enum__ibv_mig_state(IBV_MIG_REARM)),
*/
  __pyx_t_10 = __Pyx_PyList_Pack(3, __pyx_t_31, __pyx_t_26, __pyx_t_27); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

  /* "EnumType":28
 *     ('IBV_MIG_ARMED', __Pyx_PyLong_From_enum__ibv_mig_state(IBV_MIG_ARMED)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_27 = __Pyx_Globals(); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __pyx_t_26 = __Pyx_PyObject_GetAttrStr(__pyx_t_27, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __pyx_t_27 = __Pyx_PyObject_Call(__pyx_t_26, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_11, __pyx_mstate_global->__pyx_n_u_ibv_mig_state, __pyx_t_10};
    __pyx_t_26 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_27, __pyx_t_26, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_12 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_25, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_26);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_mig_state, __pyx_t_12) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":35
 * 
 * 
 *     ibv_mig_state._member_names_ = list(ibv_mig_state.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_ibv_mig_state); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_25 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = __Pyx_PySequence_ListKeepNew(__pyx_t_25); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_25, __pyx_mstate_global->__pyx_n_u_ibv_mig_state); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_25, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_12) < (0)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

    /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_mr_init_attr_mask = __Pyx_FlagBase('ibv_mr_init_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_REG_MR_MASK_IOVA', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_IOVA)),
 *     ('IBV_REG_MR_MASK_ADDR', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_ADDR)),
*/
  __pyx_t_12 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_26 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_mr_init_attr_mask = __Pyx_FlagBase('ibv_mr_init_attr_mask', [
 *     ('IBV_REG_MR_MASK_IOVA', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_IOVA)),             # <<<<<<<<<<<<<<
 *     ('IBV_REG_MR_MASK_ADDR', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_ADDR)),
 *     ('IBV_REG_MR_MASK_FD', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_FD)),
*/
  __pyx_t_27 = __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_IOVA); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __pyx_t_10 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_REG_MR_MASK_IOVA, __pyx_t_27); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

  /* "EnumType":25
 * ibv_mr_init_attr_mask = __Pyx_FlagBase('ibv_mr_init_attr_mask', [
 *     ('IBV_REG_MR_MASK_IOVA', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_IOVA)),
 *     ('IBV_REG_MR_MASK_ADDR', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_ADDR)),             # <<<<<<<<<<<<<<
 *     ('IBV_REG_MR_MASK_FD', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_FD)),
 *     ('IBV_REG_MR_MASK_FD_OFFSET', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_FD_OFFSET)),
*/
  __pyx_t_27 = __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_ADDR); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_REG_MR_MASK_ADDR, __pyx_t_27); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

  /* "EnumType":26
 *     ('IBV_REG_MR_MASK_IOVA', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_IOVA)),
 *     ('IBV_REG_MR_MASK_ADDR', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_ADDR)),
 *     ('IBV_REG_MR_MASK_FD', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_FD)),             # <<<<<<<<<<<<<<
 *     ('IBV_REG_MR_MASK_FD_OFFSET', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_FD_OFFSET)),
 *     ('IBV_REG_MR_MASK_DMAH', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_DMAH)),
*/
  __pyx_t_27 = __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_FD); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __pyx_t_31 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_REG_MR_MASK_FD, __pyx_t_27); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

  /* "EnumType":27
 *     ('IBV_REG_MR_MASK_ADDR', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_ADDR)),
 *     ('IBV_REG_MR_MASK_FD', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_FD)),
 *     ('IBV_REG_MR_MASK_FD_OFFSET', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_FD_OFFSET)),             # <<<<<<<<<<<<<<
 *     ('IBV_REG_MR_MASK_DMAH', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_DMAH)),
 * 
*/
  __pyx_t_27 = __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_FD_OFFSET); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __pyx_t_24 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_REG_MR_MASK_FD_OFFSET, __pyx_t_27); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

  /* "EnumType":28
 *     ('IBV_REG_MR_MASK_FD', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_FD)),
 *     ('IBV_REG_MR_MASK_FD_OFFSET', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_FD_OFFSET)),
 *     ('IBV_REG_MR_MASK_DMAH', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_DMAH)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_27 = __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_DMAH); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __pyx_t_23 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_REG_MR_MASK_DMAH, __pyx_t_27); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_mr_init_attr_mask = __Pyx_FlagBase('ibv_mr_init_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_REG_MR_MASK_IOVA', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_IOVA)),
 *     ('IBV_REG_MR_MASK_ADDR', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_ADDR)),
*/
  __pyx_t_27 = __Pyx_PyList_Pack(5, __pyx_t_10, __pyx_t_11, __pyx_t_31, __pyx_t_24, __pyx_t_23); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":30
 *     ('IBV_REG_MR_MASK_DMAH', __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(IBV_REG_MR_MASK_DMAH)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_23 = __Pyx_Globals(); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_12, __pyx_mstate_global->__pyx_n_u_ibv_mr_init_attr_mask, __pyx_t_27};
    __pyx_t_24 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_23, __pyx_t_24, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_25 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_26, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_24);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_mr_init_attr_mask, __pyx_t_25) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":32
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":37
 * 
 * 
 *     ibv_mr_init_attr_mask._member_names_ = list(ibv_mr_init_attr_mask.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_25, __pyx_mstate_global->__pyx_n_u_ibv_mr_init_attr_mask); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_26 = __Pyx_PyObject_GetAttrStr(__pyx_t_25, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __pyx_t_25 = __Pyx_PySequence_ListKeepNew(__pyx_t_26); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_26, __pyx_mstate_global->__pyx_n_u_ibv_mr_init_attr_mask); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_26, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_25) < (0)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

    /* "EnumType":32
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_mtu = __Pyx_FlagBase('ibv_mtu', [             # <<<<<<<<<<<<<<
 *     ('IBV_MTU_256', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_256)),
 *     ('IBV_MTU_512', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_512)),
*/
  __pyx_t_25 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_24 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_mtu = __Pyx_FlagBase('ibv_mtu', [
 *     ('IBV_MTU_256', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_256)),             # <<<<<<<<<<<<<<
 *     ('IBV_MTU_512', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_512)),
 *     ('IBV_MTU_1024', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_1024)),
*/
  __pyx_t_23 = __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_256); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_27 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_MTU_256, __pyx_t_23); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":25
 * ibv_mtu = __Pyx_FlagBase('ibv_mtu', [
 *     ('IBV_MTU_256', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_256)),
 *     ('IBV_MTU_512', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_512)),             # <<<<<<<<<<<<<<
 *     ('IBV_MTU_1024', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_1024)),
 *     ('IBV_MTU_2048', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_2048)),
*/
  __pyx_t_23 = __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_512); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_12 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_MTU_512, __pyx_t_23); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":26
 *     ('IBV_MTU_256', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_256)),
 *     ('IBV_MTU_512', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_512)),
 *     ('IBV_MTU_1024', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_1024)),             # <<<<<<<<<<<<<<
 *     ('IBV_MTU_2048', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_2048)),
 *     ('IBV_MTU_4096', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_4096)),
*/
  __pyx_t_23 = __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_1024); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_31 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_MTU_1024, __pyx_t_23); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":27
 *     ('IBV_MTU_512', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_512)),
 *     ('IBV_MTU_1024', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_1024)),
 *     ('IBV_MTU_2048', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_2048)),             # <<<<<<<<<<<<<<
 *     ('IBV_MTU_4096', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_4096)),
 * 
*/
  __pyx_t_23 = __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_2048); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_MTU_2048, __pyx_t_23); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":28
 *     ('IBV_MTU_1024', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_1024)),
 *     ('IBV_MTU_2048', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_2048)),
 *     ('IBV_MTU_4096', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_4096)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_23 = __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_4096); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_10 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_MTU_4096, __pyx_t_23); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_mtu = __Pyx_FlagBase('ibv_mtu', [             # <<<<<<<<<<<<<<
 *     ('IBV_MTU_256', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_256)),
 *     ('IBV_MTU_512', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_512)),
*/
  __pyx_t_23 = __Pyx_PyList_Pack(5, __pyx_t_27, __pyx_t_12, __pyx_t_31, __pyx_t_11, __pyx_t_10); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":30
 *     ('IBV_MTU_4096', __Pyx_PyLong_From_enum__ibv_mtu(IBV_MTU_4096)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_10 = __Pyx_Globals(); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_25, __pyx_mstate_global->__pyx_n_u_ibv_mtu, __pyx_t_23};
    __pyx_t_11 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_10, __pyx_t_11, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_26 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_24, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_11);
    __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_mtu, __pyx_t_26) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":32
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":37
 * 
 * 
 *     ibv_mtu._member_names_ = list(ibv_mtu.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_26, __pyx_mstate_global->__pyx_n_u_ibv_mtu); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_26, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __pyx_t_26 = __Pyx_PySequence_ListKeepNew(__pyx_t_24); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_mstate_global->__pyx_n_u_ibv_mtu); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_24, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_26) < (0)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":32
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_mw_type = __Pyx_FlagBase('ibv_mw_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_MW_TYPE_1', __Pyx_PyLong_From_enum__ibv_mw_type(IBV_MW_TYPE_1)),
 *     ('IBV_MW_TYPE_2', __Pyx_PyLong_From_enum__ibv_mw_type(IBV_MW_TYPE_2)),
*/
  __pyx_t_26 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_11 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_mw_type = __Pyx_FlagBase('ibv_mw_type', [
 *     ('IBV_MW_TYPE_1', __Pyx_PyLong_From_enum__ibv_mw_type(IBV_MW_TYPE_1)),             # <<<<<<<<<<<<<<
 *     ('IBV_MW_TYPE_2', __Pyx_PyLong_From_enum__ibv_mw_type(IBV_MW_TYPE_2)),
 * 
*/
  __pyx_t_10 = __Pyx_PyLong_From_enum__ibv_mw_type(IBV_MW_TYPE_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_23 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_MW_TYPE_1, __pyx_t_10); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":25
 * ibv_mw_type = __Pyx_FlagBase('ibv_mw_type', [
 *     ('IBV_MW_TYPE_1', __Pyx_PyLong_From_enum__ibv_mw_type(IBV_MW_TYPE_1)),
 *     ('IBV_MW_TYPE_2', __Pyx_PyLong_From_enum__ibv_mw_type(IBV_MW_TYPE_2)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_10 = __Pyx_PyLong_From_enum__ibv_mw_type(IBV_MW_TYPE_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_25 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_MW_TYPE_2, __pyx_t_10); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_mw_type = __Pyx_FlagBase('ibv_mw_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_MW_TYPE_1', __Pyx_PyLong_From_enum__ibv_mw_type(IBV_MW_TYPE_1)),
 *     ('IBV_MW_TYPE_2', __Pyx_PyLong_From_enum__ibv_mw_type(IBV_MW_TYPE_2)),
*/
  __pyx_t_10 = __Pyx_PyList_Pack(2, __pyx_t_23, __pyx_t_25); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":27
 *     ('IBV_MW_TYPE_2', __Pyx_PyLong_From_enum__ibv_mw_type(IBV_MW_TYPE_2)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_25 = __Pyx_Globals(); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_25, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __pyx_t_25 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_26, __pyx_mstate_global->__pyx_n_u_ibv_mw_type, __pyx_t_10};
    __pyx_t_23 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_25, __pyx_t_23, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_24 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_11, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_23);
    __Pyx_XDECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_mw_type, __pyx_t_24) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":34
 * 
 * 
 *     ibv_mw_type._member_names_ = list(ibv_mw_type.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_mstate_global->__pyx_n_u_ibv_mw_type); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PySequence_ListKeepNew(__pyx_t_11); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_ibv_mw_type); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_24) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_node_type = __Pyx_FlagBase('ibv_node_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_NODE_UNKNOWN', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_UNKNOWN)),
 *     ('IBV_NODE_CA', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_CA)),
*/
  __pyx_t_24 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_23 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_node_type = __Pyx_FlagBase('ibv_node_type', [
 *     ('IBV_NODE_UNKNOWN', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_UNKNOWN)),             # <<<<<<<<<<<<<<
 *     ('IBV_NODE_CA', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_CA)),
 *     ('IBV_NODE_SWITCH', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_SWITCH)),
*/
  __pyx_t_25 = __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_UNKNOWN); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_10 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_NODE_UNKNOWN, __pyx_t_25); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":25
 * ibv_node_type = __Pyx_FlagBase('ibv_node_type', [
 *     ('IBV_NODE_UNKNOWN', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_UNKNOWN)),
 *     ('IBV_NODE_CA', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_CA)),             # <<<<<<<<<<<<<<
 *     ('IBV_NODE_SWITCH', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_SWITCH)),
 *     ('IBV_NODE_ROUTER', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_ROUTER)),
*/
  __pyx_t_25 = __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_CA); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_26 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_NODE_CA, __pyx_t_25); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":26
 *     ('IBV_NODE_UNKNOWN', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_UNKNOWN)),
 *     ('IBV_NODE_CA', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_CA)),
 *     ('IBV_NODE_SWITCH', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_SWITCH)),             # <<<<<<<<<<<<<<
 *     ('IBV_NODE_ROUTER', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_ROUTER)),
 *     ('IBV_NODE_RNIC', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_RNIC)),
*/
  __pyx_t_25 = __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_SWITCH); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_31 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_NODE_SWITCH, __pyx_t_25); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":27
 *     ('IBV_NODE_CA', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_CA)),
 *     ('IBV_NODE_SWITCH', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_SWITCH)),
 *     ('IBV_NODE_ROUTER', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_ROUTER)),             # <<<<<<<<<<<<<<
 *     ('IBV_NODE_RNIC', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_RNIC)),
 *     ('IBV_NODE_USNIC', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_USNIC)),
*/
  __pyx_t_25 = __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_ROUTER); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_12 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_NODE_ROUTER, __pyx_t_25); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":28
 *     ('IBV_NODE_SWITCH', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_SWITCH)),
 *     ('IBV_NODE_ROUTER', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_ROUTER)),
 *     ('IBV_NODE_RNIC', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_RNIC)),             # <<<<<<<<<<<<<<
 *     ('IBV_NODE_USNIC', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_USNIC)),
 *     ('IBV_NODE_USNIC_UDP', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_USNIC_UDP)),
*/
  __pyx_t_25 = __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_RNIC); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_27 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_NODE_RNIC, __pyx_t_25); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":29
 *     ('IBV_NODE_ROUTER', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_ROUTER)),
 *     ('IBV_NODE_RNIC', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_RNIC)),
 *     ('IBV_NODE_USNIC', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_USNIC)),             # <<<<<<<<<<<<<<
 *     ('IBV_NODE_USNIC_UDP', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_USNIC_UDP)),
 *     ('IBV_NODE_UNSPECIFIED', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_UNSPECIFIED)),
*/
  __pyx_t_25 = __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_USNIC); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_NODE_USNIC, __pyx_t_25); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":30
 *     ('IBV_NODE_RNIC', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_RNIC)),
 *     ('IBV_NODE_USNIC', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_USNIC)),
 *     ('IBV_NODE_USNIC_UDP', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_USNIC_UDP)),             # <<<<<<<<<<<<<<
 *     ('IBV_NODE_UNSPECIFIED', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_UNSPECIFIED)),
 * 
*/
  __pyx_t_25 = __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_USNIC_UDP); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_21 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_NODE_USNIC_UDP, __pyx_t_25); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":31
 *     ('IBV_NODE_USNIC', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_USNIC)),
 *     ('IBV_NODE_USNIC_UDP', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_USNIC_UDP)),
 *     ('IBV_NODE_UNSPECIFIED', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_UNSPECIFIED)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_25 = __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_UNSPECIFIED); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_NODE_UNSPECIFIED, __pyx_t_25); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_node_type = __Pyx_FlagBase('ibv_node_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_NODE_UNKNOWN', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_UNKNOWN)),
 *     ('IBV_NODE_CA', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_CA)),
*/
  __pyx_t_25 = __Pyx_PyList_Pack(8, __pyx_t_10, __pyx_t_26, __pyx_t_31, __pyx_t_12, __pyx_t_27, __pyx_t_22, __pyx_t_21, __pyx_t_5); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumType":33
 *     ('IBV_NODE_UNSPECIFIED', __Pyx_PyLong_From_enum__ibv_node_type(IBV_NODE_UNSPECIFIED)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_5 = __Pyx_Globals(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_21, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_24, __pyx_mstate_global->__pyx_n_u_ibv_node_type, __pyx_t_25};
    __pyx_t_21 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_5, __pyx_t_21, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_11 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_23, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_21);
    __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_node_type, __pyx_t_11) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":35
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":40
 * 
 * 
 *     ibv_node_type._member_names_ = list(ibv_node_type.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_ibv_node_type); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PySequence_ListKeepNew(__pyx_t_23); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_ibv_node_type); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_11) < (0)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

    /* "EnumType":35
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_odp_general_caps = __Pyx_FlagBase('ibv_odp_general_caps', [             # <<<<<<<<<<<<<<
 *     ('IBV_ODP_SUPPORT', __Pyx_PyLong_From_enum__ibv_odp_general_caps(IBV_ODP_SUPPORT)),
 *     ('IBV_ODP_SUPPORT_IMPLICIT', __Pyx_PyLong_From_enum__ibv_odp_general_caps(IBV_ODP_SUPPORT_IMPLICIT)),
*/
  __pyx_t_11 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_21 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_odp_general_caps = __Pyx_FlagBase('ibv_odp_general_caps', [
 *     ('IBV_ODP_SUPPORT', __Pyx_PyLong_From_enum__ibv_odp_general_caps(IBV_ODP_SUPPORT)),             # <<<<<<<<<<<<<<
 *     ('IBV_ODP_SUPPORT_IMPLICIT', __Pyx_PyLong_From_enum__ibv_odp_general_caps(IBV_ODP_SUPPORT_IMPLICIT)),
 * 
*/
  __pyx_t_5 = __Pyx_PyLong_From_enum__ibv_odp_general_caps(IBV_ODP_SUPPORT); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_25 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ODP_SUPPORT, __pyx_t_5); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumType":25
 * ibv_odp_general_caps = __Pyx_FlagBase('ibv_odp_general_caps', [
 *     ('IBV_ODP_SUPPORT', __Pyx_PyLong_From_enum__ibv_odp_general_caps(IBV_ODP_SUPPORT)),
 *     ('IBV_ODP_SUPPORT_IMPLICIT', __Pyx_PyLong_From_enum__ibv_odp_general_caps(IBV_ODP_SUPPORT_IMPLICIT)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_5 = __Pyx_PyLong_From_enum__ibv_odp_general_caps(IBV_ODP_SUPPORT_IMPLICIT); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_24 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ODP_SUPPORT_IMPLICIT, __pyx_t_5); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_odp_general_caps = __Pyx_FlagBase('ibv_odp_general_caps', [             # <<<<<<<<<<<<<<
 *     ('IBV_ODP_SUPPORT', __Pyx_PyLong_From_enum__ibv_odp_general_caps(IBV_ODP_SUPPORT)),
 *     ('IBV_ODP_SUPPORT_IMPLICIT', __Pyx_PyLong_From_enum__ibv_odp_general_caps(IBV_ODP_SUPPORT_IMPLICIT)),
*/
  __pyx_t_5 = __Pyx_PyList_Pack(2, __pyx_t_25, __pyx_t_24); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

  /* "EnumType":27
 *     ('IBV_ODP_SUPPORT_IMPLICIT', __Pyx_PyLong_From_enum__ibv_odp_general_caps(IBV_ODP_SUPPORT_IMPLICIT)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_24 = __Pyx_Globals(); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_25 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_25, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_11, __pyx_mstate_global->__pyx_n_u_ibv_odp_general_caps, __pyx_t_5};
    __pyx_t_25 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_24, __pyx_t_25, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_23 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_21, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_25);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_odp_general_caps, __pyx_t_23) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":34
 * 
 * 
 *     ibv_odp_general_caps._member_names_ = list(ibv_odp_general_caps.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_ibv_odp_general_caps); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __pyx_t_23 = __Pyx_PySequence_ListKeepNew(__pyx_t_21); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_ibv_odp_general_caps); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_23) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_odp_transport_cap_bits = __Pyx_FlagBase('ibv_odp_transport_cap_bits', [             # <<<<<<<<<<<<<<
 *     ('IBV_ODP_SUPPORT_SEND', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_SEND)),
 *     ('IBV_ODP_SUPPORT_RECV', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_RECV)),
*/
  __pyx_t_23 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_25 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_odp_transport_cap_bits = __Pyx_FlagBase('ibv_odp_transport_cap_bits', [
 *     ('IBV_ODP_SUPPORT_SEND', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_SEND)),             # <<<<<<<<<<<<<<
 *     ('IBV_ODP_SUPPORT_RECV', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_RECV)),
 *     ('IBV_ODP_SUPPORT_WRITE', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_WRITE)),
*/
  __pyx_t_24 = __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_SEND); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ODP_SUPPORT_SEND, __pyx_t_24); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

  /* "EnumType":25
 * ibv_odp_transport_cap_bits = __Pyx_FlagBase('ibv_odp_transport_cap_bits', [
 *     ('IBV_ODP_SUPPORT_SEND', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_SEND)),
 *     ('IBV_ODP_SUPPORT_RECV', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_RECV)),             # <<<<<<<<<<<<<<
 *     ('IBV_ODP_SUPPORT_WRITE', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_WRITE)),
 *     ('IBV_ODP_SUPPORT_READ', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_READ)),
*/
  __pyx_t_24 = __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_RECV); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ODP_SUPPORT_RECV, __pyx_t_24); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

  /* "EnumType":26
 *     ('IBV_ODP_SUPPORT_SEND', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_SEND)),
 *     ('IBV_ODP_SUPPORT_RECV', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_RECV)),
 *     ('IBV_ODP_SUPPORT_WRITE', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_WRITE)),             # <<<<<<<<<<<<<<
 *     ('IBV_ODP_SUPPORT_READ', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_READ)),
 *     ('IBV_ODP_SUPPORT_ATOMIC', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_ATOMIC)),
*/
  __pyx_t_24 = __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_WRITE); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ODP_SUPPORT_WRITE, __pyx_t_24); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

  /* "EnumType":27
 *     ('IBV_ODP_SUPPORT_RECV', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_RECV)),
 *     ('IBV_ODP_SUPPORT_WRITE', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_WRITE)),
 *     ('IBV_ODP_SUPPORT_READ', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_READ)),             # <<<<<<<<<<<<<<
 *     ('IBV_ODP_SUPPORT_ATOMIC', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_ATOMIC)),
 *     ('IBV_ODP_SUPPORT_SRQ_RECV', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_SRQ_RECV)),
*/
  __pyx_t_24 = __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_READ); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_27 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ODP_SUPPORT_READ, __pyx_t_24); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

  /* "EnumType":28
 *     ('IBV_ODP_SUPPORT_WRITE', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_WRITE)),
 *     ('IBV_ODP_SUPPORT_READ', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_READ)),
 *     ('IBV_ODP_SUPPORT_ATOMIC', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_ATOMIC)),             # <<<<<<<<<<<<<<
 *     ('IBV_ODP_SUPPORT_SRQ_RECV', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_SRQ_RECV)),
 *     ('IBV_ODP_SUPPORT_FLUSH', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_FLUSH)),
*/
  __pyx_t_24 = __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_ATOMIC); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_12 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ODP_SUPPORT_ATOMIC, __pyx_t_24); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

  /* "EnumType":29
 *     ('IBV_ODP_SUPPORT_READ', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_READ)),
 *     ('IBV_ODP_SUPPORT_ATOMIC', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_ATOMIC)),
 *     ('IBV_ODP_SUPPORT_SRQ_RECV', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_SRQ_RECV)),             # <<<<<<<<<<<<<<
 *     ('IBV_ODP_SUPPORT_FLUSH', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_FLUSH)),
 *     ('IBV_ODP_SUPPORT_ATOMIC_WRITE', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_ATOMIC_WRITE)),
*/
  __pyx_t_24 = __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_SRQ_RECV); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_31 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ODP_SUPPORT_SRQ_RECV, __pyx_t_24); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

  /* "EnumType":30
 *     ('IBV_ODP_SUPPORT_ATOMIC', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_ATOMIC)),
 *     ('IBV_ODP_SUPPORT_SRQ_RECV', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_SRQ_RECV)),
 *     ('IBV_ODP_SUPPORT_FLUSH', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_FLUSH)),             # <<<<<<<<<<<<<<
 *     ('IBV_ODP_SUPPORT_ATOMIC_WRITE', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_ATOMIC_WRITE)),
 * 
*/
  __pyx_t_24 = __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_FLUSH); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_26 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ODP_SUPPORT_FLUSH, __pyx_t_24); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

  /* "EnumType":31
 *     ('IBV_ODP_SUPPORT_SRQ_RECV', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_SRQ_RECV)),
 *     ('IBV_ODP_SUPPORT_FLUSH', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_FLUSH)),
 *     ('IBV_ODP_SUPPORT_ATOMIC_WRITE', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_ATOMIC_WRITE)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_24 = __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_ATOMIC_WRITE); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __pyx_t_10 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_ODP_SUPPORT_ATOMIC_WRITE, __pyx_t_24); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_odp_transport_cap_bits = __Pyx_FlagBase('ibv_odp_transport_cap_bits', [             # <<<<<<<<<<<<<<
 *     ('IBV_ODP_SUPPORT_SEND', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_SEND)),
 *     ('IBV_ODP_SUPPORT_RECV', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_RECV)),
*/
  __pyx_t_24 = __Pyx_PyList_Pack(8, __pyx_t_5, __pyx_t_11, __pyx_t_22, __pyx_t_27, __pyx_t_12, __pyx_t_31, __pyx_t_26, __pyx_t_10); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":33
 *     ('IBV_ODP_SUPPORT_ATOMIC_WRITE', __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(IBV_ODP_SUPPORT_ATOMIC_WRITE)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_10 = __Pyx_Globals(); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_26 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_26, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_23, __pyx_mstate_global->__pyx_n_u_ibv_odp_transport_cap_bits, __pyx_t_24};
    __pyx_t_26 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_10, __pyx_t_26, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_21 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_25, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_26);
    __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_odp_transport_cap_bits, __pyx_t_21) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":35
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":40
 * 
 * 
 *     ibv_odp_transport_cap_bits._member_names_ = list(ibv_odp_transport_cap_bits.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_ibv_odp_transport_cap_bits); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __pyx_t_25 = __Pyx_PyObject_GetAttrStr(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __pyx_t_21 = __Pyx_PySequence_ListKeepNew(__pyx_t_25); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_25, __pyx_mstate_global->__pyx_n_u_ibv_odp_transport_cap_bits); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_25, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_21) < (0)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

    /* "EnumType":35
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_ops_flags = __Pyx_FlagBase('ibv_ops_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_OPS_SIGNALED', __Pyx_PyLong_From_enum__ibv_ops_flags(IBV_OPS_SIGNALED)),
 *     ('IBV_OPS_TM_SYNC', __Pyx_PyLong_From_enum__ibv_ops_flags(IBV_OPS_TM_SYNC)),
*/
  __pyx_t_21 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_26 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_ops_flags = __Pyx_FlagBase('ibv_ops_flags', [
 *     ('IBV_OPS_SIGNALED', __Pyx_PyLong_From_enum__ibv_ops_flags(IBV_OPS_SIGNALED)),             # <<<<<<<<<<<<<<
 *     ('IBV_OPS_TM_SYNC', __Pyx_PyLong_From_enum__ibv_ops_flags(IBV_OPS_TM_SYNC)),
 * 
*/
  __pyx_t_10 = __Pyx_PyLong_From_enum__ibv_ops_flags(IBV_OPS_SIGNALED); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_24 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_OPS_SIGNALED, __pyx_t_10); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":25
 * ibv_ops_flags = __Pyx_FlagBase('ibv_ops_flags', [
 *     ('IBV_OPS_SIGNALED', __Pyx_PyLong_From_enum__ibv_ops_flags(IBV_OPS_SIGNALED)),
 *     ('IBV_OPS_TM_SYNC', __Pyx_PyLong_From_enum__ibv_ops_flags(IBV_OPS_TM_SYNC)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_10 = __Pyx_PyLong_From_enum__ibv_ops_flags(IBV_OPS_TM_SYNC); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_23 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_OPS_TM_SYNC, __pyx_t_10); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_ops_flags = __Pyx_FlagBase('ibv_ops_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_OPS_SIGNALED', __Pyx_PyLong_From_enum__ibv_ops_flags(IBV_OPS_SIGNALED)),
 *     ('IBV_OPS_TM_SYNC', __Pyx_PyLong_From_enum__ibv_ops_flags(IBV_OPS_TM_SYNC)),
*/
  __pyx_t_10 = __Pyx_PyList_Pack(2, __pyx_t_24, __pyx_t_23); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":27
 *     ('IBV_OPS_TM_SYNC', __Pyx_PyLong_From_enum__ibv_ops_flags(IBV_OPS_TM_SYNC)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_23 = __Pyx_Globals(); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_21, __pyx_mstate_global->__pyx_n_u_ibv_ops_flags, __pyx_t_10};
    __pyx_t_24 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_23, __pyx_t_24, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_25 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_26, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_24);
    __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_ops_flags, __pyx_t_25) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":34
 * 
 * 
 *     ibv_ops_flags._member_names_ = list(ibv_ops_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_25, __pyx_mstate_global->__pyx_n_u_ibv_ops_flags); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_26 = __Pyx_PyObject_GetAttrStr(__pyx_t_25, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __pyx_t_25 = __Pyx_PySequence_ListKeepNew(__pyx_t_26); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_26, __pyx_mstate_global->__pyx_n_u_ibv_ops_flags); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_26, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_25) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_ops_wr_opcode = __Pyx_FlagBase('ibv_ops_wr_opcode', [             # <<<<<<<<<<<<<<
 *     ('IBV_WR_TAG_ADD', __Pyx_PyLong_From_enum__ibv_ops_wr_opcode(IBV_WR_TAG_ADD)),
 *     ('IBV_WR_TAG_DEL', __Pyx_PyLong_From_enum__ibv_ops_wr_opcode(IBV_WR_TAG_DEL)),
*/
  __pyx_t_25 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_24 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_ops_wr_opcode = __Pyx_FlagBase('ibv_ops_wr_opcode', [
 *     ('IBV_WR_TAG_ADD', __Pyx_PyLong_From_enum__ibv_ops_wr_opcode(IBV_WR_TAG_ADD)),             # <<<<<<<<<<<<<<
 *     ('IBV_WR_TAG_DEL', __Pyx_PyLong_From_enum__ibv_ops_wr_opcode(IBV_WR_TAG_DEL)),
 *     ('IBV_WR_TAG_SYNC', __Pyx_PyLong_From_enum__ibv_ops_wr_opcode(IBV_WR_TAG_SYNC)),
*/
  __pyx_t_23 = __Pyx_PyLong_From_enum__ibv_ops_wr_opcode(IBV_WR_TAG_ADD); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_10 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WR_TAG_ADD, __pyx_t_23); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":25
 * ibv_ops_wr_opcode = __Pyx_FlagBase('ibv_ops_wr_opcode', [
 *     ('IBV_WR_TAG_ADD', __Pyx_PyLong_From_enum__ibv_ops_wr_opcode(IBV_WR_TAG_ADD)),
 *     ('IBV_WR_TAG_DEL', __Pyx_PyLong_From_enum__ibv_ops_wr_opcode(IBV_WR_TAG_DEL)),             # <<<<<<<<<<<<<<
 *     ('IBV_WR_TAG_SYNC', __Pyx_PyLong_From_enum__ibv_ops_wr_opcode(IBV_WR_TAG_SYNC)),
 * 
*/
  __pyx_t_23 = __Pyx_PyLong_From_enum__ibv_ops_wr_opcode(IBV_WR_TAG_DEL); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_21 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WR_TAG_DEL, __pyx_t_23); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":26
 *     ('IBV_WR_TAG_ADD', __Pyx_PyLong_From_enum__ibv_ops_wr_opcode(IBV_WR_TAG_ADD)),
 *     ('IBV_WR_TAG_DEL', __Pyx_PyLong_From_enum__ibv_ops_wr_opcode(IBV_WR_TAG_DEL)),
 *     ('IBV_WR_TAG_SYNC', __Pyx_PyLong_From_enum__ibv_ops_wr_opcode(IBV_WR_TAG_SYNC)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_23 = __Pyx_PyLong_From_enum__ibv_ops_wr_opcode(IBV_WR_TAG_SYNC); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_31 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WR_TAG_SYNC, __pyx_t_23); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_ops_wr_opcode = __Pyx_FlagBase('ibv_ops_wr_opcode', [             # <<<<<<<<<<<<<<
 *     ('IBV_WR_TAG_ADD', __Pyx_PyLong_From_enum__ibv_ops_wr_opcode(IBV_WR_TAG_ADD)),
 *     ('IBV_WR_TAG_DEL', __Pyx_PyLong_From_enum__ibv_ops_wr_opcode(IBV_WR_TAG_DEL)),
*/
  __pyx_t_23 = __Pyx_PyList_Pack(3, __pyx_t_10, __pyx_t_21, __pyx_t_31); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":28
 *     ('IBV_WR_TAG_SYNC', __Pyx_PyLong_From_enum__ibv_ops_wr_opcode(IBV_WR_TAG_SYNC)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_31 = __Pyx_Globals(); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_31, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __pyx_t_31 = __Pyx_PyObject_Call(__pyx_t_21, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_25, __pyx_mstate_global->__pyx_n_u_ibv_ops_wr_opcode, __pyx_t_23};
    __pyx_t_21 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_31, __pyx_t_21, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_26 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_24, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_21);
    __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_ops_wr_opcode, __pyx_t_26) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":35
 * 
 * 
 *     ibv_ops_wr_opcode._member_names_ = list(ibv_ops_wr_opcode.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_26, __pyx_mstate_global->__pyx_n_u_ibv_ops_wr_opcode); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_26, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __pyx_t_26 = __Pyx_PySequence_ListKeepNew(__pyx_t_24); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_mstate_global->__pyx_n_u_ibv_ops_wr_opcode); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_24, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_26) < (0)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_parent_domain_init_attr_mask = __Pyx_FlagBase('ibv_parent_domain_init_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_PARENT_DOMAIN_INIT_ATTR_ALLOCATORS', __Pyx_PyLong_From_enum__ibv_parent_domain_init_attr_mask(IBV_PARENT_DOMAIN_INIT_ATTR_ALLOCATORS)),
 *     ('IBV_PARENT_DOMAIN_INIT_ATTR_PD_CONTEXT', __Pyx_PyLong_From_enum__ibv_parent_domain_init_attr_mask(IBV_PARENT_DOMAIN_INIT_ATTR_PD_CONTEXT)),
*/
  __pyx_t_26 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_21 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_parent_domain_init_attr_mask = __Pyx_FlagBase('ibv_parent_domain_init_attr_mask', [
 *     ('IBV_PARENT_DOMAIN_INIT_ATTR_ALLOCATORS', __Pyx_PyLong_From_enum__ibv_parent_domain_init_attr_mask(IBV_PARENT_DOMAIN_INIT_ATTR_ALLOCATORS)),             # <<<<<<<<<<<<<<
 *     ('IBV_PARENT_DOMAIN_INIT_ATTR_PD_CONTEXT', __Pyx_PyLong_From_enum__ibv_parent_domain_init_attr_mask(IBV_PARENT_DOMAIN_INIT_ATTR_PD_CONTEXT)),
 * 
*/
  __pyx_t_31 = __Pyx_PyLong_From_enum__ibv_parent_domain_init_attr_mask(IBV_PARENT_DOMAIN_INIT_ATTR_ALLOCATORS); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_23 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PARENT_DOMAIN_INIT_ATTR_ALLO, __pyx_t_31); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":25
 * ibv_parent_domain_init_attr_mask = __Pyx_FlagBase('ibv_parent_domain_init_attr_mask', [
 *     ('IBV_PARENT_DOMAIN_INIT_ATTR_ALLOCATORS', __Pyx_PyLong_From_enum__ibv_parent_domain_init_attr_mask(IBV_PARENT_DOMAIN_INIT_ATTR_ALLOCATORS)),
 *     ('IBV_PARENT_DOMAIN_INIT_ATTR_PD_CONTEXT', __Pyx_PyLong_From_enum__ibv_parent_domain_init_attr_mask(IBV_PARENT_DOMAIN_INIT_ATTR_PD_CONTEXT)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_31 = __Pyx_PyLong_From_enum__ibv_parent_domain_init_attr_mask(IBV_PARENT_DOMAIN_INIT_ATTR_PD_CONTEXT); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_25 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PARENT_DOMAIN_INIT_ATTR_PD_C, __pyx_t_31); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_parent_domain_init_attr_mask = __Pyx_FlagBase('ibv_parent_domain_init_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_PARENT_DOMAIN_INIT_ATTR_ALLOCATORS', __Pyx_PyLong_From_enum__ibv_parent_domain_init_attr_mask(IBV_PARENT_DOMAIN_INIT_ATTR_ALLOCATORS)),
 *     ('IBV_PARENT_DOMAIN_INIT_ATTR_PD_CONTEXT', __Pyx_PyLong_From_enum__ibv_parent_domain_init_attr_mask(IBV_PARENT_DOMAIN_INIT_ATTR_PD_CONTEXT)),
*/
  __pyx_t_31 = __Pyx_PyList_Pack(2, __pyx_t_23, __pyx_t_25); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":27
 *     ('IBV_PARENT_DOMAIN_INIT_ATTR_PD_CONTEXT', __Pyx_PyLong_From_enum__ibv_parent_domain_init_attr_mask(IBV_PARENT_DOMAIN_INIT_ATTR_PD_CONTEXT)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_25 = __Pyx_Globals(); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_25, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __pyx_t_25 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_26, __pyx_mstate_global->__pyx_n_u_ibv_parent_domain_init_attr_mask, __pyx_t_31};
    __pyx_t_23 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_25, __pyx_t_23, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_24 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_21, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_23);
    __Pyx_XDECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_parent_domain_init_attr_mask, __pyx_t_24) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":34
 * 
 * 
 *     ibv_parent_domain_init_attr_mask._member_names_ = list(ibv_parent_domain_init_attr_mask.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_mstate_global->__pyx_n_u_ibv_parent_domain_init_attr_mask); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PySequence_ListKeepNew(__pyx_t_21); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_ibv_parent_domain_init_attr_mask); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_24) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_placement_type = __Pyx_FlagBase('ibv_placement_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_FLUSH_GLOBAL', __Pyx_PyLong_From_enum__ibv_placement_type(IBV_FLUSH_GLOBAL)),
 *     ('IBV_FLUSH_PERSISTENT', __Pyx_PyLong_From_enum__ibv_placement_type(IBV_FLUSH_PERSISTENT)),
*/
  __pyx_t_24 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_23 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_placement_type = __Pyx_FlagBase('ibv_placement_type', [
 *     ('IBV_FLUSH_GLOBAL', __Pyx_PyLong_From_enum__ibv_placement_type(IBV_FLUSH_GLOBAL)),             # <<<<<<<<<<<<<<
 *     ('IBV_FLUSH_PERSISTENT', __Pyx_PyLong_From_enum__ibv_placement_type(IBV_FLUSH_PERSISTENT)),
 * 
*/
  __pyx_t_25 = __Pyx_PyLong_From_enum__ibv_placement_type(IBV_FLUSH_GLOBAL); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_31 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLUSH_GLOBAL, __pyx_t_25); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":25
 * ibv_placement_type = __Pyx_FlagBase('ibv_placement_type', [
 *     ('IBV_FLUSH_GLOBAL', __Pyx_PyLong_From_enum__ibv_placement_type(IBV_FLUSH_GLOBAL)),
 *     ('IBV_FLUSH_PERSISTENT', __Pyx_PyLong_From_enum__ibv_placement_type(IBV_FLUSH_PERSISTENT)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_25 = __Pyx_PyLong_From_enum__ibv_placement_type(IBV_FLUSH_PERSISTENT); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_26 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLUSH_PERSISTENT, __pyx_t_25); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_placement_type = __Pyx_FlagBase('ibv_placement_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_FLUSH_GLOBAL', __Pyx_PyLong_From_enum__ibv_placement_type(IBV_FLUSH_GLOBAL)),
 *     ('IBV_FLUSH_PERSISTENT', __Pyx_PyLong_From_enum__ibv_placement_type(IBV_FLUSH_PERSISTENT)),
*/
  __pyx_t_25 = __Pyx_PyList_Pack(2, __pyx_t_31, __pyx_t_26); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":27
 *     ('IBV_FLUSH_PERSISTENT', __Pyx_PyLong_From_enum__ibv_placement_type(IBV_FLUSH_PERSISTENT)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_26 = __Pyx_Globals(); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_31 = __Pyx_PyObject_GetAttrStr(__pyx_t_26, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  __pyx_t_26 = __Pyx_PyObject_Call(__pyx_t_31, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_24, __pyx_mstate_global->__pyx_n_u_ibv_placement_type, __pyx_t_25};
    __pyx_t_31 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_26, __pyx_t_31, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_21 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_23, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_31);
    __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_placement_type, __pyx_t_21) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":34
 * 
 * 
 *     ibv_placement_type._member_names_ = list(ibv_placement_type.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_ibv_placement_type); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __pyx_t_21 = __Pyx_PySequence_ListKeepNew(__pyx_t_23); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_ibv_placement_type); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_21) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_port_cap_flags = __Pyx_FlagBase('ibv_port_cap_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_SM', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SM)),
 *     ('IBV_PORT_NOTICE_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_NOTICE_SUP)),
*/
  __pyx_t_21 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_31 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_port_cap_flags = __Pyx_FlagBase('ibv_port_cap_flags', [
 *     ('IBV_PORT_SM', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SM)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_NOTICE_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_NOTICE_SUP)),
 *     ('IBV_PORT_TRAP_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_TRAP_SUP)),
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SM); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_25 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_SM, __pyx_t_26); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":25
 * ibv_port_cap_flags = __Pyx_FlagBase('ibv_port_cap_flags', [
 *     ('IBV_PORT_SM', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SM)),
 *     ('IBV_PORT_NOTICE_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_NOTICE_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_TRAP_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_TRAP_SUP)),
 *     ('IBV_PORT_OPT_IPD_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_OPT_IPD_SUP)),
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_NOTICE_SUP); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_24 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_NOTICE_SUP, __pyx_t_26); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":26
 *     ('IBV_PORT_SM', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SM)),
 *     ('IBV_PORT_NOTICE_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_NOTICE_SUP)),
 *     ('IBV_PORT_TRAP_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_TRAP_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_OPT_IPD_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_OPT_IPD_SUP)),
 *     ('IBV_PORT_AUTO_MIGR_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_AUTO_MIGR_SUP)),
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_TRAP_SUP); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_10 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_TRAP_SUP, __pyx_t_26); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":27
 *     ('IBV_PORT_NOTICE_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_NOTICE_SUP)),
 *     ('IBV_PORT_TRAP_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_TRAP_SUP)),
 *     ('IBV_PORT_OPT_IPD_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_OPT_IPD_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_AUTO_MIGR_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_AUTO_MIGR_SUP)),
 *     ('IBV_PORT_SL_MAP_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SL_MAP_SUP)),
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_OPT_IPD_SUP); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_12 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_OPT_IPD_SUP, __pyx_t_26); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":28
 *     ('IBV_PORT_TRAP_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_TRAP_SUP)),
 *     ('IBV_PORT_OPT_IPD_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_OPT_IPD_SUP)),
 *     ('IBV_PORT_AUTO_MIGR_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_AUTO_MIGR_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_SL_MAP_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SL_MAP_SUP)),
 *     ('IBV_PORT_MKEY_NVRAM', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_MKEY_NVRAM)),
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_AUTO_MIGR_SUP); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_27 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_AUTO_MIGR_SUP, __pyx_t_26); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":29
 *     ('IBV_PORT_OPT_IPD_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_OPT_IPD_SUP)),
 *     ('IBV_PORT_AUTO_MIGR_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_AUTO_MIGR_SUP)),
 *     ('IBV_PORT_SL_MAP_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SL_MAP_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_MKEY_NVRAM', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_MKEY_NVRAM)),
 *     ('IBV_PORT_PKEY_NVRAM', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_PKEY_NVRAM)),
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SL_MAP_SUP); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_SL_MAP_SUP, __pyx_t_26); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":30
 *     ('IBV_PORT_AUTO_MIGR_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_AUTO_MIGR_SUP)),
 *     ('IBV_PORT_SL_MAP_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SL_MAP_SUP)),
 *     ('IBV_PORT_MKEY_NVRAM', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_MKEY_NVRAM)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_PKEY_NVRAM', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_PKEY_NVRAM)),
 *     ('IBV_PORT_LED_INFO_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_LED_INFO_SUP)),
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_MKEY_NVRAM); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_MKEY_NVRAM, __pyx_t_26); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":31
 *     ('IBV_PORT_SL_MAP_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SL_MAP_SUP)),
 *     ('IBV_PORT_MKEY_NVRAM', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_MKEY_NVRAM)),
 *     ('IBV_PORT_PKEY_NVRAM', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_PKEY_NVRAM)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_LED_INFO_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_LED_INFO_SUP)),
 *     ('IBV_PORT_SYS_IMAGE_GUID_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SYS_IMAGE_GUID_SUP)),
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_PKEY_NVRAM); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_PKEY_NVRAM, __pyx_t_26); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":32
 *     ('IBV_PORT_MKEY_NVRAM', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_MKEY_NVRAM)),
 *     ('IBV_PORT_PKEY_NVRAM', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_PKEY_NVRAM)),
 *     ('IBV_PORT_LED_INFO_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_LED_INFO_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_SYS_IMAGE_GUID_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SYS_IMAGE_GUID_SUP)),
 *     ('IBV_PORT_PKEY_SW_EXT_PORT_TRAP_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_PKEY_SW_EXT_PORT_TRAP_SUP)),
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_LED_INFO_SUP); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_LED_INFO_SUP, __pyx_t_26); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":33
 *     ('IBV_PORT_PKEY_NVRAM', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_PKEY_NVRAM)),
 *     ('IBV_PORT_LED_INFO_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_LED_INFO_SUP)),
 *     ('IBV_PORT_SYS_IMAGE_GUID_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SYS_IMAGE_GUID_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_PKEY_SW_EXT_PORT_TRAP_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_PKEY_SW_EXT_PORT_TRAP_SUP)),
 *     ('IBV_PORT_EXTENDED_SPEEDS_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_EXTENDED_SPEEDS_SUP)),
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SYS_IMAGE_GUID_SUP); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_6 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_SYS_IMAGE_GUID_SUP, __pyx_t_26); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":34
 *     ('IBV_PORT_LED_INFO_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_LED_INFO_SUP)),
 *     ('IBV_PORT_SYS_IMAGE_GUID_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SYS_IMAGE_GUID_SUP)),
 *     ('IBV_PORT_PKEY_SW_EXT_PORT_TRAP_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_PKEY_SW_EXT_PORT_TRAP_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_EXTENDED_SPEEDS_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_EXTENDED_SPEEDS_SUP)),
 *     ('IBV_PORT_CAP_MASK2_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_CAP_MASK2_SUP)),
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_PKEY_SW_EXT_PORT_TRAP_SUP); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_14 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_PKEY_SW_EXT_PORT_TRAP_S, __pyx_t_26); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":35
 *     ('IBV_PORT_SYS_IMAGE_GUID_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SYS_IMAGE_GUID_SUP)),
 *     ('IBV_PORT_PKEY_SW_EXT_PORT_TRAP_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_PKEY_SW_EXT_PORT_TRAP_SUP)),
 *     ('IBV_PORT_EXTENDED_SPEEDS_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_EXTENDED_SPEEDS_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_CAP_MASK2_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_CAP_MASK2_SUP)),
 *     ('IBV_PORT_CM_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_CM_SUP)),
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_EXTENDED_SPEEDS_SUP); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_30 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_EXTENDED_SPEEDS_SUP, __pyx_t_26); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":36
 *     ('IBV_PORT_PKEY_SW_EXT_PORT_TRAP_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_PKEY_SW_EXT_PORT_TRAP_SUP)),
 *     ('IBV_PORT_EXTENDED_SPEEDS_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_EXTENDED_SPEEDS_SUP)),
 *     ('IBV_PORT_CAP_MASK2_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_CAP_MASK2_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_CM_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_CM_SUP)),
 *     ('IBV_PORT_SNMP_TUNNEL_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SNMP_TUNNEL_SUP)),
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_CAP_MASK2_SUP); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_7 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_CAP_MASK2_SUP, __pyx_t_26); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":37
 *     ('IBV_PORT_EXTENDED_SPEEDS_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_EXTENDED_SPEEDS_SUP)),
 *     ('IBV_PORT_CAP_MASK2_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_CAP_MASK2_SUP)),
 *     ('IBV_PORT_CM_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_CM_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_SNMP_TUNNEL_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SNMP_TUNNEL_SUP)),
 *     ('IBV_PORT_REINIT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_REINIT_SUP)),
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_CM_SUP); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_CM_SUP, __pyx_t_26); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":38
 *     ('IBV_PORT_CAP_MASK2_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_CAP_MASK2_SUP)),
 *     ('IBV_PORT_CM_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_CM_SUP)),
 *     ('IBV_PORT_SNMP_TUNNEL_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SNMP_TUNNEL_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_REINIT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_REINIT_SUP)),
 *     ('IBV_PORT_DEVICE_MGMT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_DEVICE_MGMT_SUP)),
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SNMP_TUNNEL_SUP); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_13 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_SNMP_TUNNEL_SUP, __pyx_t_26); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":39
 *     ('IBV_PORT_CM_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_CM_SUP)),
 *     ('IBV_PORT_SNMP_TUNNEL_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SNMP_TUNNEL_SUP)),
 *     ('IBV_PORT_REINIT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_REINIT_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_DEVICE_MGMT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_DEVICE_MGMT_SUP)),
 *     ('IBV_PORT_VENDOR_CLASS_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_VENDOR_CLASS_SUP)),
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_REINIT_SUP); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_28 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_REINIT_SUP, __pyx_t_26); if (unlikely(!__pyx_t_28)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":40
 *     ('IBV_PORT_SNMP_TUNNEL_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SNMP_TUNNEL_SUP)),
 *     ('IBV_PORT_REINIT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_REINIT_SUP)),
 *     ('IBV_PORT_DEVICE_MGMT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_DEVICE_MGMT_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_VENDOR_CLASS_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_VENDOR_CLASS_SUP)),
 *     ('IBV_PORT_DR_NOTICE_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_DR_NOTICE_SUP)),
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_DEVICE_MGMT_SUP); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_29 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_DEVICE_MGMT_SUP, __pyx_t_26); if (unlikely(!__pyx_t_29)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":41
 *     ('IBV_PORT_REINIT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_REINIT_SUP)),
 *     ('IBV_PORT_DEVICE_MGMT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_DEVICE_MGMT_SUP)),
 *     ('IBV_PORT_VENDOR_CLASS_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_VENDOR_CLASS_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_DR_NOTICE_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_DR_NOTICE_SUP)),
 *     ('IBV_PORT_CAP_MASK_NOTICE_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_CAP_MASK_NOTICE_SUP)),
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_VENDOR_CLASS_SUP); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_20 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_VENDOR_CLASS_SUP, __pyx_t_26); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":42
 *     ('IBV_PORT_DEVICE_MGMT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_DEVICE_MGMT_SUP)),
 *     ('IBV_PORT_VENDOR_CLASS_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_VENDOR_CLASS_SUP)),
 *     ('IBV_PORT_DR_NOTICE_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_DR_NOTICE_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_CAP_MASK_NOTICE_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_CAP_MASK_NOTICE_SUP)),
 *     ('IBV_PORT_BOOT_MGMT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_BOOT_MGMT_SUP)),
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_DR_NOTICE_SUP); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_2 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_DR_NOTICE_SUP, __pyx_t_26); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":43
 *     ('IBV_PORT_VENDOR_CLASS_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_VENDOR_CLASS_SUP)),
 *     ('IBV_PORT_DR_NOTICE_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_DR_NOTICE_SUP)),
 *     ('IBV_PORT_CAP_MASK_NOTICE_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_CAP_MASK_NOTICE_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_BOOT_MGMT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_BOOT_MGMT_SUP)),
 *     ('IBV_PORT_LINK_LATENCY_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_LINK_LATENCY_SUP)),
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_CAP_MASK_NOTICE_SUP); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_32 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_CAP_MASK_NOTICE_SUP, __pyx_t_26); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":44
 *     ('IBV_PORT_DR_NOTICE_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_DR_NOTICE_SUP)),
 *     ('IBV_PORT_CAP_MASK_NOTICE_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_CAP_MASK_NOTICE_SUP)),
 *     ('IBV_PORT_BOOT_MGMT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_BOOT_MGMT_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_LINK_LATENCY_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_LINK_LATENCY_SUP)),
 *     ('IBV_PORT_CLIENT_REG_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_CLIENT_REG_SUP)),
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_BOOT_MGMT_SUP); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_9 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_BOOT_MGMT_SUP, __pyx_t_26); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":45
 *     ('IBV_PORT_CAP_MASK_NOTICE_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_CAP_MASK_NOTICE_SUP)),
 *     ('IBV_PORT_BOOT_MGMT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_BOOT_MGMT_SUP)),
 *     ('IBV_PORT_LINK_LATENCY_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_LINK_LATENCY_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_CLIENT_REG_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_CLIENT_REG_SUP)),
 *     ('IBV_PORT_IP_BASED_GIDS', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_IP_BASED_GIDS)),
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_LINK_LATENCY_SUP); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_8 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_LINK_LATENCY_SUP, __pyx_t_26); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":46
 *     ('IBV_PORT_BOOT_MGMT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_BOOT_MGMT_SUP)),
 *     ('IBV_PORT_LINK_LATENCY_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_LINK_LATENCY_SUP)),
 *     ('IBV_PORT_CLIENT_REG_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_CLIENT_REG_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_IP_BASED_GIDS', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_IP_BASED_GIDS)),
 * 
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_CLIENT_REG_SUP); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_17 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_CLIENT_REG_SUP, __pyx_t_26); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":47
 *     ('IBV_PORT_LINK_LATENCY_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_LINK_LATENCY_SUP)),
 *     ('IBV_PORT_CLIENT_REG_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_CLIENT_REG_SUP)),
 *     ('IBV_PORT_IP_BASED_GIDS', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_IP_BASED_GIDS)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_26 = __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_IP_BASED_GIDS); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_IP_BASED_GIDS, __pyx_t_26); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_port_cap_flags = __Pyx_FlagBase('ibv_port_cap_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_SM', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_SM)),
 *     ('IBV_PORT_NOTICE_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_NOTICE_SUP)),
*/
  __pyx_t_26 = __Pyx_PyList_Pack(24, __pyx_t_25, __pyx_t_24, __pyx_t_10, __pyx_t_12, __pyx_t_27, __pyx_t_22, __pyx_t_11, __pyx_t_5, __pyx_t_4, __pyx_t_6, __pyx_t_14, __pyx_t_30, __pyx_t_7, __pyx_t_15, __pyx_t_13, __pyx_t_28, __pyx_t_29, __pyx_t_20, __pyx_t_2, __pyx_t_32, __pyx_t_9, __pyx_t_8, __pyx_t_17, __pyx_t_16); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":49
 *     ('IBV_PORT_IP_BASED_GIDS', __Pyx_PyLong_From_enum__ibv_port_cap_flags(IBV_PORT_IP_BASED_GIDS)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_16 = __Pyx_Globals(); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_21, __pyx_mstate_global->__pyx_n_u_ibv_port_cap_flags, __pyx_t_26};
    __pyx_t_17 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_16, __pyx_t_17, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_23 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_31, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_17);
    __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_port_cap_flags, __pyx_t_23) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":51
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":56
 * 
 * 
 *     ibv_port_cap_flags._member_names_ = list(ibv_port_cap_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_ibv_port_cap_flags); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_31 = __Pyx_PyObject_GetAttrStr(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __pyx_t_23 = __Pyx_PySequence_ListKeepNew(__pyx_t_31); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_31, __pyx_mstate_global->__pyx_n_u_ibv_port_cap_flags); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_31, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_23) < (0)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

    /* "EnumType":51
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_port_cap_flags2 = __Pyx_FlagBase('ibv_port_cap_flags2', [             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_SET_NODE_DESC_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_SET_NODE_DESC_SUP)),
 *     ('IBV_PORT_INFO_EXT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_INFO_EXT_SUP)),
*/
  __pyx_t_23 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_17 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_port_cap_flags2 = __Pyx_FlagBase('ibv_port_cap_flags2', [
 *     ('IBV_PORT_SET_NODE_DESC_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_SET_NODE_DESC_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_INFO_EXT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_INFO_EXT_SUP)),
 *     ('IBV_PORT_VIRT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_VIRT_SUP)),
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_SET_NODE_DESC_SUP); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_26 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_SET_NODE_DESC_SUP, __pyx_t_16); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":25
 * ibv_port_cap_flags2 = __Pyx_FlagBase('ibv_port_cap_flags2', [
 *     ('IBV_PORT_SET_NODE_DESC_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_SET_NODE_DESC_SUP)),
 *     ('IBV_PORT_INFO_EXT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_INFO_EXT_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_VIRT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_VIRT_SUP)),
 *     ('IBV_PORT_SWITCH_PORT_STATE_TABLE_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_SWITCH_PORT_STATE_TABLE_SUP)),
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_INFO_EXT_SUP); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_21 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_INFO_EXT_SUP, __pyx_t_16); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":26
 *     ('IBV_PORT_SET_NODE_DESC_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_SET_NODE_DESC_SUP)),
 *     ('IBV_PORT_INFO_EXT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_INFO_EXT_SUP)),
 *     ('IBV_PORT_VIRT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_VIRT_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_SWITCH_PORT_STATE_TABLE_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_SWITCH_PORT_STATE_TABLE_SUP)),
 *     ('IBV_PORT_LINK_WIDTH_2X_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_LINK_WIDTH_2X_SUP)),
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_VIRT_SUP); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_8 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_VIRT_SUP, __pyx_t_16); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":27
 *     ('IBV_PORT_INFO_EXT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_INFO_EXT_SUP)),
 *     ('IBV_PORT_VIRT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_VIRT_SUP)),
 *     ('IBV_PORT_SWITCH_PORT_STATE_TABLE_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_SWITCH_PORT_STATE_TABLE_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_LINK_WIDTH_2X_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_LINK_WIDTH_2X_SUP)),
 *     ('IBV_PORT_LINK_SPEED_HDR_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_LINK_SPEED_HDR_SUP)),
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_SWITCH_PORT_STATE_TABLE_SUP); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_9 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_SWITCH_PORT_STATE_TABLE, __pyx_t_16); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":28
 *     ('IBV_PORT_VIRT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_VIRT_SUP)),
 *     ('IBV_PORT_SWITCH_PORT_STATE_TABLE_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_SWITCH_PORT_STATE_TABLE_SUP)),
 *     ('IBV_PORT_LINK_WIDTH_2X_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_LINK_WIDTH_2X_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_LINK_SPEED_HDR_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_LINK_SPEED_HDR_SUP)),
 *     ('IBV_PORT_LINK_SPEED_NDR_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_LINK_SPEED_NDR_SUP)),
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_LINK_WIDTH_2X_SUP); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_32 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_LINK_WIDTH_2X_SUP, __pyx_t_16); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":29
 *     ('IBV_PORT_SWITCH_PORT_STATE_TABLE_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_SWITCH_PORT_STATE_TABLE_SUP)),
 *     ('IBV_PORT_LINK_WIDTH_2X_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_LINK_WIDTH_2X_SUP)),
 *     ('IBV_PORT_LINK_SPEED_HDR_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_LINK_SPEED_HDR_SUP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_LINK_SPEED_NDR_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_LINK_SPEED_NDR_SUP)),
 * 
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_LINK_SPEED_HDR_SUP); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_2 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_LINK_SPEED_HDR_SUP, __pyx_t_16); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":30
 *     ('IBV_PORT_LINK_WIDTH_2X_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_LINK_WIDTH_2X_SUP)),
 *     ('IBV_PORT_LINK_SPEED_HDR_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_LINK_SPEED_HDR_SUP)),
 *     ('IBV_PORT_LINK_SPEED_NDR_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_LINK_SPEED_NDR_SUP)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_LINK_SPEED_NDR_SUP); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_20 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_LINK_SPEED_NDR_SUP, __pyx_t_16); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_port_cap_flags2 = __Pyx_FlagBase('ibv_port_cap_flags2', [             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_SET_NODE_DESC_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_SET_NODE_DESC_SUP)),
 *     ('IBV_PORT_INFO_EXT_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_INFO_EXT_SUP)),
*/
  __pyx_t_16 = __Pyx_PyList_Pack(7, __pyx_t_26, __pyx_t_21, __pyx_t_8, __pyx_t_9, __pyx_t_32, __pyx_t_2, __pyx_t_20); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":32
 *     ('IBV_PORT_LINK_SPEED_NDR_SUP', __Pyx_PyLong_From_enum__ibv_port_cap_flags2(IBV_PORT_LINK_SPEED_NDR_SUP)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_20 = __Pyx_Globals(); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_20, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_23, __pyx_mstate_global->__pyx_n_u_ibv_port_cap_flags2, __pyx_t_16};
    __pyx_t_2 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_20, __pyx_t_2, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_31 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_17, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_port_cap_flags2, __pyx_t_31) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":34
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":39
 * 
 * 
 *     ibv_port_cap_flags2._member_names_ = list(ibv_port_cap_flags2.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_31, __pyx_mstate_global->__pyx_n_u_ibv_port_cap_flags2); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
    __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_31, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __pyx_t_31 = __Pyx_PySequence_ListKeepNew(__pyx_t_17); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_31);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_ibv_port_cap_flags2); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_31) < (0)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

    /* "EnumType":34
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_port_state = __Pyx_FlagBase('ibv_port_state', [             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_NOP', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_NOP)),
 *     ('IBV_PORT_DOWN', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_DOWN)),
*/
  __pyx_t_31 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_2 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_port_state = __Pyx_FlagBase('ibv_port_state', [
 *     ('IBV_PORT_NOP', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_NOP)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_DOWN', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_DOWN)),
 *     ('IBV_PORT_INIT', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_INIT)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_NOP); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_NOP, __pyx_t_20); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":25
 * ibv_port_state = __Pyx_FlagBase('ibv_port_state', [
 *     ('IBV_PORT_NOP', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_NOP)),
 *     ('IBV_PORT_DOWN', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_DOWN)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_INIT', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_INIT)),
 *     ('IBV_PORT_ARMED', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_ARMED)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_DOWN); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_23 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_DOWN, __pyx_t_20); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":26
 *     ('IBV_PORT_NOP', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_NOP)),
 *     ('IBV_PORT_DOWN', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_DOWN)),
 *     ('IBV_PORT_INIT', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_INIT)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_ARMED', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_ARMED)),
 *     ('IBV_PORT_ACTIVE', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_ACTIVE)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_INIT); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_32 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_INIT, __pyx_t_20); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":27
 *     ('IBV_PORT_DOWN', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_DOWN)),
 *     ('IBV_PORT_INIT', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_INIT)),
 *     ('IBV_PORT_ARMED', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_ARMED)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_ACTIVE', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_ACTIVE)),
 *     ('IBV_PORT_ACTIVE_DEFER', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_ACTIVE_DEFER)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_ARMED); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_9 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_ARMED, __pyx_t_20); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":28
 *     ('IBV_PORT_INIT', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_INIT)),
 *     ('IBV_PORT_ARMED', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_ARMED)),
 *     ('IBV_PORT_ACTIVE', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_ACTIVE)),             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_ACTIVE_DEFER', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_ACTIVE_DEFER)),
 * 
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_ACTIVE); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_8 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_ACTIVE, __pyx_t_20); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":29
 *     ('IBV_PORT_ARMED', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_ARMED)),
 *     ('IBV_PORT_ACTIVE', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_ACTIVE)),
 *     ('IBV_PORT_ACTIVE_DEFER', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_ACTIVE_DEFER)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_ACTIVE_DEFER); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_21 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_PORT_ACTIVE_DEFER, __pyx_t_20); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_port_state = __Pyx_FlagBase('ibv_port_state', [             # <<<<<<<<<<<<<<
 *     ('IBV_PORT_NOP', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_NOP)),
 *     ('IBV_PORT_DOWN', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_DOWN)),
*/
  __pyx_t_20 = __Pyx_PyList_Pack(6, __pyx_t_16, __pyx_t_23, __pyx_t_32, __pyx_t_9, __pyx_t_8, __pyx_t_21); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":31
 *     ('IBV_PORT_ACTIVE_DEFER', __Pyx_PyLong_From_enum__ibv_port_state(IBV_PORT_ACTIVE_DEFER)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_21 = __Pyx_Globals(); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_21 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_31, __pyx_mstate_global->__pyx_n_u_ibv_port_state, __pyx_t_20};
    __pyx_t_8 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_21, __pyx_t_8, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_17 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_port_state, __pyx_t_17) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "EnumType":33
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":38
 * 
 * 
 *     ibv_port_state._member_names_ = list(ibv_port_state.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_ibv_port_state); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __pyx_t_17 = __Pyx_PySequence_ListKeepNew(__pyx_t_2); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_ibv_port_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_17) < (0)) __PYX_ERR(1, 38, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":33
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_qp_attr_mask = __Pyx_FlagBase('ibv_qp_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_QP_STATE', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_STATE)),
 *     ('IBV_QP_CUR_STATE', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_CUR_STATE)),
*/
  __pyx_t_17 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_8 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_qp_attr_mask = __Pyx_FlagBase('ibv_qp_attr_mask', [
 *     ('IBV_QP_STATE', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_STATE)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_CUR_STATE', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_CUR_STATE)),
 *     ('IBV_QP_EN_SQD_ASYNC_NOTIFY', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_EN_SQD_ASYNC_NOTIFY)),
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_STATE); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_20 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_STATE, __pyx_t_21); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":25
 * ibv_qp_attr_mask = __Pyx_FlagBase('ibv_qp_attr_mask', [
 *     ('IBV_QP_STATE', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_STATE)),
 *     ('IBV_QP_CUR_STATE', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_CUR_STATE)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_EN_SQD_ASYNC_NOTIFY', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_EN_SQD_ASYNC_NOTIFY)),
 *     ('IBV_QP_ACCESS_FLAGS', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_ACCESS_FLAGS)),
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_CUR_STATE); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_31 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_CUR_STATE, __pyx_t_21); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":26
 *     ('IBV_QP_STATE', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_STATE)),
 *     ('IBV_QP_CUR_STATE', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_CUR_STATE)),
 *     ('IBV_QP_EN_SQD_ASYNC_NOTIFY', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_EN_SQD_ASYNC_NOTIFY)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_ACCESS_FLAGS', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_ACCESS_FLAGS)),
 *     ('IBV_QP_PKEY_INDEX', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PKEY_INDEX)),
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_EN_SQD_ASYNC_NOTIFY); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_9 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_EN_SQD_ASYNC_NOTIFY, __pyx_t_21); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":27
 *     ('IBV_QP_CUR_STATE', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_CUR_STATE)),
 *     ('IBV_QP_EN_SQD_ASYNC_NOTIFY', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_EN_SQD_ASYNC_NOTIFY)),
 *     ('IBV_QP_ACCESS_FLAGS', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_ACCESS_FLAGS)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_PKEY_INDEX', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PKEY_INDEX)),
 *     ('IBV_QP_PORT', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PORT)),
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_ACCESS_FLAGS); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_32 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_ACCESS_FLAGS, __pyx_t_21); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":28
 *     ('IBV_QP_EN_SQD_ASYNC_NOTIFY', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_EN_SQD_ASYNC_NOTIFY)),
 *     ('IBV_QP_ACCESS_FLAGS', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_ACCESS_FLAGS)),
 *     ('IBV_QP_PKEY_INDEX', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PKEY_INDEX)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_PORT', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PORT)),
 *     ('IBV_QP_QKEY', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_QKEY)),
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PKEY_INDEX); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_23 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_PKEY_INDEX, __pyx_t_21); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":29
 *     ('IBV_QP_ACCESS_FLAGS', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_ACCESS_FLAGS)),
 *     ('IBV_QP_PKEY_INDEX', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PKEY_INDEX)),
 *     ('IBV_QP_PORT', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PORT)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_QKEY', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_QKEY)),
 *     ('IBV_QP_AV', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_AV)),
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PORT); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_PORT, __pyx_t_21); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":30
 *     ('IBV_QP_PKEY_INDEX', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PKEY_INDEX)),
 *     ('IBV_QP_PORT', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PORT)),
 *     ('IBV_QP_QKEY', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_QKEY)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_AV', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_AV)),
 *     ('IBV_QP_PATH_MTU', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PATH_MTU)),
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_QKEY); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_26 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_QKEY, __pyx_t_21); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":31
 *     ('IBV_QP_PORT', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PORT)),
 *     ('IBV_QP_QKEY', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_QKEY)),
 *     ('IBV_QP_AV', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_AV)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_PATH_MTU', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PATH_MTU)),
 *     ('IBV_QP_TIMEOUT', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_TIMEOUT)),
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_AV); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_29 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_AV, __pyx_t_21); if (unlikely(!__pyx_t_29)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":32
 *     ('IBV_QP_QKEY', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_QKEY)),
 *     ('IBV_QP_AV', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_AV)),
 *     ('IBV_QP_PATH_MTU', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PATH_MTU)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_TIMEOUT', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_TIMEOUT)),
 *     ('IBV_QP_RETRY_CNT', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_RETRY_CNT)),
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PATH_MTU); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_28 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_PATH_MTU, __pyx_t_21); if (unlikely(!__pyx_t_28)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":33
 *     ('IBV_QP_AV', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_AV)),
 *     ('IBV_QP_PATH_MTU', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PATH_MTU)),
 *     ('IBV_QP_TIMEOUT', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_TIMEOUT)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_RETRY_CNT', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_RETRY_CNT)),
 *     ('IBV_QP_RNR_RETRY', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_RNR_RETRY)),
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_TIMEOUT); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_13 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_TIMEOUT, __pyx_t_21); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":34
 *     ('IBV_QP_PATH_MTU', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PATH_MTU)),
 *     ('IBV_QP_TIMEOUT', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_TIMEOUT)),
 *     ('IBV_QP_RETRY_CNT', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_RETRY_CNT)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_RNR_RETRY', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_RNR_RETRY)),
 *     ('IBV_QP_RQ_PSN', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_RQ_PSN)),
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_RETRY_CNT); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_RETRY_CNT, __pyx_t_21); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":35
 *     ('IBV_QP_TIMEOUT', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_TIMEOUT)),
 *     ('IBV_QP_RETRY_CNT', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_RETRY_CNT)),
 *     ('IBV_QP_RNR_RETRY', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_RNR_RETRY)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_RQ_PSN', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_RQ_PSN)),
 *     ('IBV_QP_MAX_QP_RD_ATOMIC', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_MAX_QP_RD_ATOMIC)),
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_RNR_RETRY); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_7 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_RNR_RETRY, __pyx_t_21); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":36
 *     ('IBV_QP_RETRY_CNT', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_RETRY_CNT)),
 *     ('IBV_QP_RNR_RETRY', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_RNR_RETRY)),
 *     ('IBV_QP_RQ_PSN', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_RQ_PSN)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_MAX_QP_RD_ATOMIC', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_MAX_QP_RD_ATOMIC)),
 *     ('IBV_QP_ALT_PATH', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_ALT_PATH)),
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_RQ_PSN); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_30 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_RQ_PSN, __pyx_t_21); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":37
 *     ('IBV_QP_RNR_RETRY', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_RNR_RETRY)),
 *     ('IBV_QP_RQ_PSN', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_RQ_PSN)),
 *     ('IBV_QP_MAX_QP_RD_ATOMIC', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_MAX_QP_RD_ATOMIC)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_ALT_PATH', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_ALT_PATH)),
 *     ('IBV_QP_MIN_RNR_TIMER', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_MIN_RNR_TIMER)),
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_MAX_QP_RD_ATOMIC); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_14 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_MAX_QP_RD_ATOMIC, __pyx_t_21); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":38
 *     ('IBV_QP_RQ_PSN', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_RQ_PSN)),
 *     ('IBV_QP_MAX_QP_RD_ATOMIC', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_MAX_QP_RD_ATOMIC)),
 *     ('IBV_QP_ALT_PATH', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_ALT_PATH)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_MIN_RNR_TIMER', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_MIN_RNR_TIMER)),
 *     ('IBV_QP_SQ_PSN', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_SQ_PSN)),
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_ALT_PATH); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_6 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_ALT_PATH, __pyx_t_21); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":39
 *     ('IBV_QP_MAX_QP_RD_ATOMIC', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_MAX_QP_RD_ATOMIC)),
 *     ('IBV_QP_ALT_PATH', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_ALT_PATH)),
 *     ('IBV_QP_MIN_RNR_TIMER', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_MIN_RNR_TIMER)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_SQ_PSN', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_SQ_PSN)),
 *     ('IBV_QP_MAX_DEST_RD_ATOMIC', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_MAX_DEST_RD_ATOMIC)),
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_MIN_RNR_TIMER); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_MIN_RNR_TIMER, __pyx_t_21); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":40
 *     ('IBV_QP_ALT_PATH', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_ALT_PATH)),
 *     ('IBV_QP_MIN_RNR_TIMER', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_MIN_RNR_TIMER)),
 *     ('IBV_QP_SQ_PSN', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_SQ_PSN)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_MAX_DEST_RD_ATOMIC', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_MAX_DEST_RD_ATOMIC)),
 *     ('IBV_QP_PATH_MIG_STATE', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PATH_MIG_STATE)),
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_SQ_PSN); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_SQ_PSN, __pyx_t_21); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":41
 *     ('IBV_QP_MIN_RNR_TIMER', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_MIN_RNR_TIMER)),
 *     ('IBV_QP_SQ_PSN', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_SQ_PSN)),
 *     ('IBV_QP_MAX_DEST_RD_ATOMIC', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_MAX_DEST_RD_ATOMIC)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_PATH_MIG_STATE', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PATH_MIG_STATE)),
 *     ('IBV_QP_CAP', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_CAP)),
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_MAX_DEST_RD_ATOMIC); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_MAX_DEST_RD_ATOMIC, __pyx_t_21); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":42
 *     ('IBV_QP_SQ_PSN', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_SQ_PSN)),
 *     ('IBV_QP_MAX_DEST_RD_ATOMIC', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_MAX_DEST_RD_ATOMIC)),
 *     ('IBV_QP_PATH_MIG_STATE', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PATH_MIG_STATE)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_CAP', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_CAP)),
 *     ('IBV_QP_DEST_QPN', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_DEST_QPN)),
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PATH_MIG_STATE); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_PATH_MIG_STATE, __pyx_t_21); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":43
 *     ('IBV_QP_MAX_DEST_RD_ATOMIC', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_MAX_DEST_RD_ATOMIC)),
 *     ('IBV_QP_PATH_MIG_STATE', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PATH_MIG_STATE)),
 *     ('IBV_QP_CAP', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_CAP)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_DEST_QPN', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_DEST_QPN)),
 *     ('IBV_QP_RATE_LIMIT', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_RATE_LIMIT)),
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_CAP); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_27 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_CAP, __pyx_t_21); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":44
 *     ('IBV_QP_PATH_MIG_STATE', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_PATH_MIG_STATE)),
 *     ('IBV_QP_CAP', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_CAP)),
 *     ('IBV_QP_DEST_QPN', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_DEST_QPN)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_RATE_LIMIT', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_RATE_LIMIT)),
 * 
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_DEST_QPN); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_12 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_DEST_QPN, __pyx_t_21); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":45
 *     ('IBV_QP_CAP', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_CAP)),
 *     ('IBV_QP_DEST_QPN', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_DEST_QPN)),
 *     ('IBV_QP_RATE_LIMIT', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_RATE_LIMIT)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_RATE_LIMIT); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_10 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_RATE_LIMIT, __pyx_t_21); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_qp_attr_mask = __Pyx_FlagBase('ibv_qp_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_QP_STATE', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_STATE)),
 *     ('IBV_QP_CUR_STATE', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_CUR_STATE)),
*/
  __pyx_t_21 = __Pyx_PyList_Pack(22, __pyx_t_20, __pyx_t_31, __pyx_t_9, __pyx_t_32, __pyx_t_23, __pyx_t_16, __pyx_t_26, __pyx_t_29, __pyx_t_28, __pyx_t_13, __pyx_t_15, __pyx_t_7, __pyx_t_30, __pyx_t_14, __pyx_t_6, __pyx_t_4, __pyx_t_5, __pyx_t_11, __pyx_t_22, __pyx_t_27, __pyx_t_12, __pyx_t_10); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":47
 *     ('IBV_QP_RATE_LIMIT', __Pyx_PyLong_From_enum__ibv_qp_attr_mask(IBV_QP_RATE_LIMIT)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_10 = __Pyx_Globals(); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_17, __pyx_mstate_global->__pyx_n_u_ibv_qp_attr_mask, __pyx_t_21};
    __pyx_t_12 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_10, __pyx_t_12, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_8, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_12);
    __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_qp_attr_mask, __pyx_t_2) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":49
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":54
 * 
 * 
 *     ibv_qp_attr_mask._member_names_ = list(ibv_qp_attr_mask.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_ibv_qp_attr_mask); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PySequence_ListKeepNew(__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_ibv_qp_attr_mask); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_2) < (0)) __PYX_ERR(1, 54, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "EnumType":49
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_qp_create_flags = __Pyx_FlagBase('ibv_qp_create_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_QP_CREATE_BLOCK_SELF_MCAST_LB', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_BLOCK_SELF_MCAST_LB)),
 *     ('IBV_QP_CREATE_SCATTER_FCS', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_SCATTER_FCS)),
*/
  __pyx_t_2 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_12 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_qp_create_flags = __Pyx_FlagBase('ibv_qp_create_flags', [
 *     ('IBV_QP_CREATE_BLOCK_SELF_MCAST_LB', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_BLOCK_SELF_MCAST_LB)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_CREATE_SCATTER_FCS', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_SCATTER_FCS)),
 *     ('IBV_QP_CREATE_CVLAN_STRIPPING', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_CVLAN_STRIPPING)),
*/
  __pyx_t_10 = __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_BLOCK_SELF_MCAST_LB); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_21 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_CREATE_BLOCK_SELF_MCAST_L, __pyx_t_10); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":25
 * ibv_qp_create_flags = __Pyx_FlagBase('ibv_qp_create_flags', [
 *     ('IBV_QP_CREATE_BLOCK_SELF_MCAST_LB', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_BLOCK_SELF_MCAST_LB)),
 *     ('IBV_QP_CREATE_SCATTER_FCS', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_SCATTER_FCS)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_CREATE_CVLAN_STRIPPING', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_CVLAN_STRIPPING)),
 *     ('IBV_QP_CREATE_SOURCE_QPN', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_SOURCE_QPN)),
*/
  __pyx_t_10 = __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_SCATTER_FCS); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_17 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_CREATE_SCATTER_FCS, __pyx_t_10); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":26
 *     ('IBV_QP_CREATE_BLOCK_SELF_MCAST_LB', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_BLOCK_SELF_MCAST_LB)),
 *     ('IBV_QP_CREATE_SCATTER_FCS', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_SCATTER_FCS)),
 *     ('IBV_QP_CREATE_CVLAN_STRIPPING', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_CVLAN_STRIPPING)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_CREATE_SOURCE_QPN', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_SOURCE_QPN)),
 *     ('IBV_QP_CREATE_PCI_WRITE_END_PADDING', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_PCI_WRITE_END_PADDING)),
*/
  __pyx_t_10 = __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_CVLAN_STRIPPING); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_27 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_CREATE_CVLAN_STRIPPING, __pyx_t_10); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":27
 *     ('IBV_QP_CREATE_SCATTER_FCS', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_SCATTER_FCS)),
 *     ('IBV_QP_CREATE_CVLAN_STRIPPING', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_CVLAN_STRIPPING)),
 *     ('IBV_QP_CREATE_SOURCE_QPN', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_SOURCE_QPN)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_CREATE_PCI_WRITE_END_PADDING', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_PCI_WRITE_END_PADDING)),
 * 
*/
  __pyx_t_10 = __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_SOURCE_QPN); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_CREATE_SOURCE_QPN, __pyx_t_10); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":28
 *     ('IBV_QP_CREATE_CVLAN_STRIPPING', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_CVLAN_STRIPPING)),
 *     ('IBV_QP_CREATE_SOURCE_QPN', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_SOURCE_QPN)),
 *     ('IBV_QP_CREATE_PCI_WRITE_END_PADDING', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_PCI_WRITE_END_PADDING)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_10 = __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_PCI_WRITE_END_PADDING); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_CREATE_PCI_WRITE_END_PADD, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_qp_create_flags = __Pyx_FlagBase('ibv_qp_create_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_QP_CREATE_BLOCK_SELF_MCAST_LB', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_BLOCK_SELF_MCAST_LB)),
 *     ('IBV_QP_CREATE_SCATTER_FCS', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_SCATTER_FCS)),
*/
  __pyx_t_10 = __Pyx_PyList_Pack(5, __pyx_t_21, __pyx_t_17, __pyx_t_27, __pyx_t_22, __pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":30
 *     ('IBV_QP_CREATE_PCI_WRITE_END_PADDING', __Pyx_PyLong_From_enum__ibv_qp_create_flags(IBV_QP_CREATE_PCI_WRITE_END_PADDING)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_11 = __Pyx_Globals(); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_22 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_22, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_2, __pyx_mstate_global->__pyx_n_u_ibv_qp_create_flags, __pyx_t_10};
    __pyx_t_22 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_11, __pyx_t_22, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_8 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_12, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_22);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_qp_create_flags, __pyx_t_8) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "EnumType":32
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":37
 * 
 * 
 *     ibv_qp_create_flags._member_names_ = list(ibv_qp_create_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_ibv_qp_create_flags); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PySequence_ListKeepNew(__pyx_t_12); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_ibv_qp_create_flags); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_8) < (0)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

    /* "EnumType":32
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_qp_create_send_ops_flags = __Pyx_FlagBase('ibv_qp_create_send_ops_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_QP_EX_WITH_RDMA_WRITE', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_RDMA_WRITE)),
 *     ('IBV_QP_EX_WITH_RDMA_WRITE_WITH_IMM', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_RDMA_WRITE_WITH_IMM)),
*/
  __pyx_t_8 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_22 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_qp_create_send_ops_flags = __Pyx_FlagBase('ibv_qp_create_send_ops_flags', [
 *     ('IBV_QP_EX_WITH_RDMA_WRITE', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_RDMA_WRITE)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_EX_WITH_RDMA_WRITE_WITH_IMM', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_RDMA_WRITE_WITH_IMM)),
 *     ('IBV_QP_EX_WITH_SEND', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_SEND)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_RDMA_WRITE); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_10 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_EX_WITH_RDMA_WRITE, __pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":25
 * ibv_qp_create_send_ops_flags = __Pyx_FlagBase('ibv_qp_create_send_ops_flags', [
 *     ('IBV_QP_EX_WITH_RDMA_WRITE', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_RDMA_WRITE)),
 *     ('IBV_QP_EX_WITH_RDMA_WRITE_WITH_IMM', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_RDMA_WRITE_WITH_IMM)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_EX_WITH_SEND', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_SEND)),
 *     ('IBV_QP_EX_WITH_SEND_WITH_IMM', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_SEND_WITH_IMM)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_RDMA_WRITE_WITH_IMM); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_2 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_EX_WITH_RDMA_WRITE_WITH_I, __pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":26
 *     ('IBV_QP_EX_WITH_RDMA_WRITE', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_RDMA_WRITE)),
 *     ('IBV_QP_EX_WITH_RDMA_WRITE_WITH_IMM', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_RDMA_WRITE_WITH_IMM)),
 *     ('IBV_QP_EX_WITH_SEND', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_SEND)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_EX_WITH_SEND_WITH_IMM', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_SEND_WITH_IMM)),
 *     ('IBV_QP_EX_WITH_RDMA_READ', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_RDMA_READ)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_SEND); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_27 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_EX_WITH_SEND, __pyx_t_11); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":27
 *     ('IBV_QP_EX_WITH_RDMA_WRITE_WITH_IMM', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_RDMA_WRITE_WITH_IMM)),
 *     ('IBV_QP_EX_WITH_SEND', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_SEND)),
 *     ('IBV_QP_EX_WITH_SEND_WITH_IMM', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_SEND_WITH_IMM)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_EX_WITH_RDMA_READ', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_RDMA_READ)),
 *     ('IBV_QP_EX_WITH_ATOMIC_CMP_AND_SWP', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_ATOMIC_CMP_AND_SWP)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_SEND_WITH_IMM); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_17 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_EX_WITH_SEND_WITH_IMM, __pyx_t_11); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":28
 *     ('IBV_QP_EX_WITH_SEND', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_SEND)),
 *     ('IBV_QP_EX_WITH_SEND_WITH_IMM', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_SEND_WITH_IMM)),
 *     ('IBV_QP_EX_WITH_RDMA_READ', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_RDMA_READ)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_EX_WITH_ATOMIC_CMP_AND_SWP', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_ATOMIC_CMP_AND_SWP)),
 *     ('IBV_QP_EX_WITH_ATOMIC_FETCH_AND_ADD', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_ATOMIC_FETCH_AND_ADD)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_RDMA_READ); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_21 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_EX_WITH_RDMA_READ, __pyx_t_11); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":29
 *     ('IBV_QP_EX_WITH_SEND_WITH_IMM', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_SEND_WITH_IMM)),
 *     ('IBV_QP_EX_WITH_RDMA_READ', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_RDMA_READ)),
 *     ('IBV_QP_EX_WITH_ATOMIC_CMP_AND_SWP', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_ATOMIC_CMP_AND_SWP)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_EX_WITH_ATOMIC_FETCH_AND_ADD', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_ATOMIC_FETCH_AND_ADD)),
 *     ('IBV_QP_EX_WITH_LOCAL_INV', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_LOCAL_INV)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_ATOMIC_CMP_AND_SWP); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_EX_WITH_ATOMIC_CMP_AND_SW, __pyx_t_11); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":30
 *     ('IBV_QP_EX_WITH_RDMA_READ', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_RDMA_READ)),
 *     ('IBV_QP_EX_WITH_ATOMIC_CMP_AND_SWP', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_ATOMIC_CMP_AND_SWP)),
 *     ('IBV_QP_EX_WITH_ATOMIC_FETCH_AND_ADD', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_ATOMIC_FETCH_AND_ADD)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_EX_WITH_LOCAL_INV', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_LOCAL_INV)),
 *     ('IBV_QP_EX_WITH_BIND_MW', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_BIND_MW)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_ATOMIC_FETCH_AND_ADD); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_EX_WITH_ATOMIC_FETCH_AND, __pyx_t_11); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":31
 *     ('IBV_QP_EX_WITH_ATOMIC_CMP_AND_SWP', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_ATOMIC_CMP_AND_SWP)),
 *     ('IBV_QP_EX_WITH_ATOMIC_FETCH_AND_ADD', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_ATOMIC_FETCH_AND_ADD)),
 *     ('IBV_QP_EX_WITH_LOCAL_INV', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_LOCAL_INV)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_EX_WITH_BIND_MW', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_BIND_MW)),
 *     ('IBV_QP_EX_WITH_SEND_WITH_INV', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_SEND_WITH_INV)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_LOCAL_INV); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_6 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_EX_WITH_LOCAL_INV, __pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":32
 *     ('IBV_QP_EX_WITH_ATOMIC_FETCH_AND_ADD', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_ATOMIC_FETCH_AND_ADD)),
 *     ('IBV_QP_EX_WITH_LOCAL_INV', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_LOCAL_INV)),
 *     ('IBV_QP_EX_WITH_BIND_MW', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_BIND_MW)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_EX_WITH_SEND_WITH_INV', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_SEND_WITH_INV)),
 *     ('IBV_QP_EX_WITH_TSO', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_TSO)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_BIND_MW); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_14 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_EX_WITH_BIND_MW, __pyx_t_11); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":33
 *     ('IBV_QP_EX_WITH_LOCAL_INV', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_LOCAL_INV)),
 *     ('IBV_QP_EX_WITH_BIND_MW', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_BIND_MW)),
 *     ('IBV_QP_EX_WITH_SEND_WITH_INV', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_SEND_WITH_INV)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_EX_WITH_TSO', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_TSO)),
 *     ('IBV_QP_EX_WITH_FLUSH', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_FLUSH)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_SEND_WITH_INV); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_30 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_EX_WITH_SEND_WITH_INV, __pyx_t_11); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":34
 *     ('IBV_QP_EX_WITH_BIND_MW', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_BIND_MW)),
 *     ('IBV_QP_EX_WITH_SEND_WITH_INV', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_SEND_WITH_INV)),
 *     ('IBV_QP_EX_WITH_TSO', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_TSO)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_EX_WITH_FLUSH', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_FLUSH)),
 *     ('IBV_QP_EX_WITH_ATOMIC_WRITE', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_ATOMIC_WRITE)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_TSO); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_7 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_EX_WITH_TSO, __pyx_t_11); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":35
 *     ('IBV_QP_EX_WITH_SEND_WITH_INV', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_SEND_WITH_INV)),
 *     ('IBV_QP_EX_WITH_TSO', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_TSO)),
 *     ('IBV_QP_EX_WITH_FLUSH', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_FLUSH)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_EX_WITH_ATOMIC_WRITE', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_ATOMIC_WRITE)),
 * 
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_FLUSH); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_EX_WITH_FLUSH, __pyx_t_11); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":36
 *     ('IBV_QP_EX_WITH_TSO', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_TSO)),
 *     ('IBV_QP_EX_WITH_FLUSH', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_FLUSH)),
 *     ('IBV_QP_EX_WITH_ATOMIC_WRITE', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_ATOMIC_WRITE)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_ATOMIC_WRITE); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_13 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_EX_WITH_ATOMIC_WRITE, __pyx_t_11); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_qp_create_send_ops_flags = __Pyx_FlagBase('ibv_qp_create_send_ops_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_QP_EX_WITH_RDMA_WRITE', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_RDMA_WRITE)),
 *     ('IBV_QP_EX_WITH_RDMA_WRITE_WITH_IMM', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_RDMA_WRITE_WITH_IMM)),
*/
  __pyx_t_11 = __Pyx_PyList_Pack(13, __pyx_t_10, __pyx_t_2, __pyx_t_27, __pyx_t_17, __pyx_t_21, __pyx_t_5, __pyx_t_4, __pyx_t_6, __pyx_t_14, __pyx_t_30, __pyx_t_7, __pyx_t_15, __pyx_t_13); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "EnumType":38
 *     ('IBV_QP_EX_WITH_ATOMIC_WRITE', __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(IBV_QP_EX_WITH_ATOMIC_WRITE)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_13 = __Pyx_Globals(); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_8, __pyx_mstate_global->__pyx_n_u_ibv_qp_create_send_ops_flags, __pyx_t_11};
    __pyx_t_15 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_13, __pyx_t_15, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_12 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_22, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_15);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_qp_create_send_ops_flags, __pyx_t_12) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "EnumType":40
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":45
 * 
 * 
 *     ibv_qp_create_send_ops_flags._member_names_ = list(ibv_qp_create_send_ops_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_ibv_qp_create_send_ops_flags); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_22 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = __Pyx_PySequence_ListKeepNew(__pyx_t_22); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_ibv_qp_create_send_ops_flags); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_12) < (0)) __PYX_ERR(1, 45, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

    /* "EnumType":40
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_qp_init_attr_mask = __Pyx_FlagBase('ibv_qp_init_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_QP_INIT_ATTR_PD', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_PD)),
 *     ('IBV_QP_INIT_ATTR_XRCD', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_XRCD)),
*/
  __pyx_t_12 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_15 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_qp_init_attr_mask = __Pyx_FlagBase('ibv_qp_init_attr_mask', [
 *     ('IBV_QP_INIT_ATTR_PD', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_PD)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_INIT_ATTR_XRCD', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_XRCD)),
 *     ('IBV_QP_INIT_ATTR_CREATE_FLAGS', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_CREATE_FLAGS)),
*/
  __pyx_t_13 = __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_PD); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_INIT_ATTR_PD, __pyx_t_13); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "EnumType":25
 * ibv_qp_init_attr_mask = __Pyx_FlagBase('ibv_qp_init_attr_mask', [
 *     ('IBV_QP_INIT_ATTR_PD', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_PD)),
 *     ('IBV_QP_INIT_ATTR_XRCD', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_XRCD)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_INIT_ATTR_CREATE_FLAGS', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_CREATE_FLAGS)),
 *     ('IBV_QP_INIT_ATTR_MAX_TSO_HEADER', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_MAX_TSO_HEADER)),
*/
  __pyx_t_13 = __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_XRCD); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_8 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_INIT_ATTR_XRCD, __pyx_t_13); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "EnumType":26
 *     ('IBV_QP_INIT_ATTR_PD', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_PD)),
 *     ('IBV_QP_INIT_ATTR_XRCD', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_XRCD)),
 *     ('IBV_QP_INIT_ATTR_CREATE_FLAGS', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_CREATE_FLAGS)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_INIT_ATTR_MAX_TSO_HEADER', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_MAX_TSO_HEADER)),
 *     ('IBV_QP_INIT_ATTR_IND_TABLE', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_IND_TABLE)),
*/
  __pyx_t_13 = __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_CREATE_FLAGS); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_7 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_INIT_ATTR_CREATE_FLAGS, __pyx_t_13); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "EnumType":27
 *     ('IBV_QP_INIT_ATTR_XRCD', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_XRCD)),
 *     ('IBV_QP_INIT_ATTR_CREATE_FLAGS', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_CREATE_FLAGS)),
 *     ('IBV_QP_INIT_ATTR_MAX_TSO_HEADER', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_MAX_TSO_HEADER)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_INIT_ATTR_IND_TABLE', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_IND_TABLE)),
 *     ('IBV_QP_INIT_ATTR_RX_HASH', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_RX_HASH)),
*/
  __pyx_t_13 = __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_MAX_TSO_HEADER); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_30 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_INIT_ATTR_MAX_TSO_HEADER, __pyx_t_13); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "EnumType":28
 *     ('IBV_QP_INIT_ATTR_CREATE_FLAGS', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_CREATE_FLAGS)),
 *     ('IBV_QP_INIT_ATTR_MAX_TSO_HEADER', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_MAX_TSO_HEADER)),
 *     ('IBV_QP_INIT_ATTR_IND_TABLE', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_IND_TABLE)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_INIT_ATTR_RX_HASH', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_RX_HASH)),
 *     ('IBV_QP_INIT_ATTR_SEND_OPS_FLAGS', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_SEND_OPS_FLAGS)),
*/
  __pyx_t_13 = __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_IND_TABLE); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_14 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_INIT_ATTR_IND_TABLE, __pyx_t_13); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "EnumType":29
 *     ('IBV_QP_INIT_ATTR_MAX_TSO_HEADER', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_MAX_TSO_HEADER)),
 *     ('IBV_QP_INIT_ATTR_IND_TABLE', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_IND_TABLE)),
 *     ('IBV_QP_INIT_ATTR_RX_HASH', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_RX_HASH)),             # <<<<<<<<<<<<<<
 *     ('IBV_QP_INIT_ATTR_SEND_OPS_FLAGS', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_SEND_OPS_FLAGS)),
 * 
*/
  __pyx_t_13 = __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_RX_HASH); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_6 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_INIT_ATTR_RX_HASH, __pyx_t_13); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "EnumType":30
 *     ('IBV_QP_INIT_ATTR_IND_TABLE', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_IND_TABLE)),
 *     ('IBV_QP_INIT_ATTR_RX_HASH', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_RX_HASH)),
 *     ('IBV_QP_INIT_ATTR_SEND_OPS_FLAGS', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_SEND_OPS_FLAGS)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_13 = __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_SEND_OPS_FLAGS); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QP_INIT_ATTR_SEND_OPS_FLAGS, __pyx_t_13); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_qp_init_attr_mask = __Pyx_FlagBase('ibv_qp_init_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_QP_INIT_ATTR_PD', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_PD)),
 *     ('IBV_QP_INIT_ATTR_XRCD', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_XRCD)),
*/
  __pyx_t_13 = __Pyx_PyList_Pack(7, __pyx_t_11, __pyx_t_8, __pyx_t_7, __pyx_t_30, __pyx_t_14, __pyx_t_6, __pyx_t_4); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":32
 *     ('IBV_QP_INIT_ATTR_SEND_OPS_FLAGS', __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(IBV_QP_INIT_ATTR_SEND_OPS_FLAGS)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_4 = __Pyx_Globals(); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_12, __pyx_mstate_global->__pyx_n_u_ibv_qp_init_attr_mask, __pyx_t_13};
    __pyx_t_6 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_4, __pyx_t_6, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_22 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_15, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_qp_init_attr_mask, __pyx_t_22) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":34
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":39
 * 
 * 
 *     ibv_qp_init_attr_mask._member_names_ = list(ibv_qp_init_attr_mask.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_ibv_qp_init_attr_mask); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __pyx_t_22 = __Pyx_PySequence_ListKeepNew(__pyx_t_15); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_ibv_qp_init_attr_mask); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_22) < (0)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":34
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_qp_state = __Pyx_FlagBase('ibv_qp_state', [             # <<<<<<<<<<<<<<
 *     ('IBV_QPS_RESET', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_RESET)),
 *     ('IBV_QPS_INIT', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_INIT)),
*/
  __pyx_t_22 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_6 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_qp_state = __Pyx_FlagBase('ibv_qp_state', [
 *     ('IBV_QPS_RESET', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_RESET)),             # <<<<<<<<<<<<<<
 *     ('IBV_QPS_INIT', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_INIT)),
 *     ('IBV_QPS_RTR', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_RTR)),
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_RESET); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_13 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QPS_RESET, __pyx_t_4); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":25
 * ibv_qp_state = __Pyx_FlagBase('ibv_qp_state', [
 *     ('IBV_QPS_RESET', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_RESET)),
 *     ('IBV_QPS_INIT', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_INIT)),             # <<<<<<<<<<<<<<
 *     ('IBV_QPS_RTR', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_RTR)),
 *     ('IBV_QPS_RTS', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_RTS)),
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_INIT); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_12 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QPS_INIT, __pyx_t_4); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":26
 *     ('IBV_QPS_RESET', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_RESET)),
 *     ('IBV_QPS_INIT', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_INIT)),
 *     ('IBV_QPS_RTR', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_RTR)),             # <<<<<<<<<<<<<<
 *     ('IBV_QPS_RTS', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_RTS)),
 *     ('IBV_QPS_SQD', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_SQD)),
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_RTR); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_14 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QPS_RTR, __pyx_t_4); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":27
 *     ('IBV_QPS_INIT', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_INIT)),
 *     ('IBV_QPS_RTR', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_RTR)),
 *     ('IBV_QPS_RTS', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_RTS)),             # <<<<<<<<<<<<<<
 *     ('IBV_QPS_SQD', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_SQD)),
 *     ('IBV_QPS_SQE', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_SQE)),
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_RTS); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_30 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QPS_RTS, __pyx_t_4); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":28
 *     ('IBV_QPS_RTR', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_RTR)),
 *     ('IBV_QPS_RTS', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_RTS)),
 *     ('IBV_QPS_SQD', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_SQD)),             # <<<<<<<<<<<<<<
 *     ('IBV_QPS_SQE', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_SQE)),
 *     ('IBV_QPS_ERR', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_ERR)),
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_SQD); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QPS_SQD, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":29
 *     ('IBV_QPS_RTS', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_RTS)),
 *     ('IBV_QPS_SQD', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_SQD)),
 *     ('IBV_QPS_SQE', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_SQE)),             # <<<<<<<<<<<<<<
 *     ('IBV_QPS_ERR', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_ERR)),
 *     ('IBV_QPS_UNKNOWN', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_UNKNOWN)),
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_SQE); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QPS_SQE, __pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":30
 *     ('IBV_QPS_SQD', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_SQD)),
 *     ('IBV_QPS_SQE', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_SQE)),
 *     ('IBV_QPS_ERR', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_QPS_UNKNOWN', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_UNKNOWN)),
 * 
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_ERR); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QPS_ERR, __pyx_t_4); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":31
 *     ('IBV_QPS_SQE', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_SQE)),
 *     ('IBV_QPS_ERR', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_ERR)),
 *     ('IBV_QPS_UNKNOWN', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_UNKNOWN)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_UNKNOWN); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QPS_UNKNOWN, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_qp_state = __Pyx_FlagBase('ibv_qp_state', [             # <<<<<<<<<<<<<<
 *     ('IBV_QPS_RESET', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_RESET)),
 *     ('IBV_QPS_INIT', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_INIT)),
*/
  __pyx_t_4 = __Pyx_PyList_Pack(8, __pyx_t_13, __pyx_t_12, __pyx_t_14, __pyx_t_30, __pyx_t_7, __pyx_t_8, __pyx_t_11, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumType":33
 *     ('IBV_QPS_UNKNOWN', __Pyx_PyLong_From_enum__ibv_qp_state(IBV_QPS_UNKNOWN)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_5 = __Pyx_Globals(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_22, __pyx_mstate_global->__pyx_n_u_ibv_qp_state, __pyx_t_4};
    __pyx_t_11 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_5, __pyx_t_11, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_15 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_11);
    __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_qp_state, __pyx_t_15) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":35
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":40
 * 
 * 
 *     ibv_qp_state._member_names_ = list(ibv_qp_state.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_ibv_qp_state); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __pyx_t_15 = __Pyx_PySequence_ListKeepNew(__pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_ibv_qp_state); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_15) < (0)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "EnumType":35
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_qp_type = __Pyx_FlagBase('ibv_qp_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_QPT_RC', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_RC)),
 *     ('IBV_QPT_UC', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_UC)),
*/
  __pyx_t_15 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_11 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_qp_type = __Pyx_FlagBase('ibv_qp_type', [
 *     ('IBV_QPT_RC', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_RC)),             # <<<<<<<<<<<<<<
 *     ('IBV_QPT_UC', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_UC)),
 *     ('IBV_QPT_UD', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_UD)),
*/
  __pyx_t_5 = __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_RC); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QPT_RC, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumType":25
 * ibv_qp_type = __Pyx_FlagBase('ibv_qp_type', [
 *     ('IBV_QPT_RC', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_RC)),
 *     ('IBV_QPT_UC', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_UC)),             # <<<<<<<<<<<<<<
 *     ('IBV_QPT_UD', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_UD)),
 *     ('IBV_QPT_RAW_PACKET', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_RAW_PACKET)),
*/
  __pyx_t_5 = __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_UC); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QPT_UC, __pyx_t_5); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumType":26
 *     ('IBV_QPT_RC', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_RC)),
 *     ('IBV_QPT_UC', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_UC)),
 *     ('IBV_QPT_UD', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_UD)),             # <<<<<<<<<<<<<<
 *     ('IBV_QPT_RAW_PACKET', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_RAW_PACKET)),
 *     ('IBV_QPT_XRC_SEND', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_XRC_SEND)),
*/
  __pyx_t_5 = __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_UD); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QPT_UD, __pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumType":27
 *     ('IBV_QPT_UC', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_UC)),
 *     ('IBV_QPT_UD', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_UD)),
 *     ('IBV_QPT_RAW_PACKET', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_RAW_PACKET)),             # <<<<<<<<<<<<<<
 *     ('IBV_QPT_XRC_SEND', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_XRC_SEND)),
 *     ('IBV_QPT_XRC_RECV', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_XRC_RECV)),
*/
  __pyx_t_5 = __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_RAW_PACKET); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QPT_RAW_PACKET, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumType":28
 *     ('IBV_QPT_UD', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_UD)),
 *     ('IBV_QPT_RAW_PACKET', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_RAW_PACKET)),
 *     ('IBV_QPT_XRC_SEND', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_XRC_SEND)),             # <<<<<<<<<<<<<<
 *     ('IBV_QPT_XRC_RECV', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_XRC_RECV)),
 *     ('IBV_QPT_DRIVER', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_DRIVER)),
*/
  __pyx_t_5 = __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_XRC_SEND); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_30 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QPT_XRC_SEND, __pyx_t_5); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumType":29
 *     ('IBV_QPT_RAW_PACKET', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_RAW_PACKET)),
 *     ('IBV_QPT_XRC_SEND', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_XRC_SEND)),
 *     ('IBV_QPT_XRC_RECV', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_XRC_RECV)),             # <<<<<<<<<<<<<<
 *     ('IBV_QPT_DRIVER', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_DRIVER)),
 * 
*/
  __pyx_t_5 = __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_XRC_RECV); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_14 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QPT_XRC_RECV, __pyx_t_5); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumType":30
 *     ('IBV_QPT_XRC_SEND', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_XRC_SEND)),
 *     ('IBV_QPT_XRC_RECV', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_XRC_RECV)),
 *     ('IBV_QPT_DRIVER', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_DRIVER)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_5 = __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_DRIVER); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_12 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QPT_DRIVER, __pyx_t_5); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_qp_type = __Pyx_FlagBase('ibv_qp_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_QPT_RC', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_RC)),
 *     ('IBV_QPT_UC', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_UC)),
*/
  __pyx_t_5 = __Pyx_PyList_Pack(7, __pyx_t_4, __pyx_t_22, __pyx_t_8, __pyx_t_7, __pyx_t_30, __pyx_t_14, __pyx_t_12); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "EnumType":32
 *     ('IBV_QPT_DRIVER', __Pyx_PyLong_From_enum__ibv_qp_type(IBV_QPT_DRIVER)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_12 = __Pyx_Globals(); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_15, __pyx_mstate_global->__pyx_n_u_ibv_qp_type, __pyx_t_5};
    __pyx_t_14 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_12, __pyx_t_14, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_6 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_11, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_14);
    __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_qp_type, __pyx_t_6) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":34
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":39
 * 
 * 
 *     ibv_qp_type._member_names_ = list(ibv_qp_type.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_ibv_qp_type); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PySequence_ListKeepNew(__pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_ibv_qp_type); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_6) < (0)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "EnumType":34
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_query_qp_data_in_order_caps = __Pyx_FlagBase('ibv_query_qp_data_in_order_caps', [             # <<<<<<<<<<<<<<
 *     ('IBV_QUERY_QP_DATA_IN_ORDER_WHOLE_MSG', __Pyx_PyLong_From_enum__ibv_query_qp_data_in_order_caps(IBV_QUERY_QP_DATA_IN_ORDER_WHOLE_MSG)),
 *     ('IBV_QUERY_QP_DATA_IN_ORDER_ALIGNED_128_BYTES', __Pyx_PyLong_From_enum__ibv_query_qp_data_in_order_caps(IBV_QUERY_QP_DATA_IN_ORDER_ALIGNED_128_BYTES)),
*/
  __pyx_t_6 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_14 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_query_qp_data_in_order_caps = __Pyx_FlagBase('ibv_query_qp_data_in_order_caps', [
 *     ('IBV_QUERY_QP_DATA_IN_ORDER_WHOLE_MSG', __Pyx_PyLong_From_enum__ibv_query_qp_data_in_order_caps(IBV_QUERY_QP_DATA_IN_ORDER_WHOLE_MSG)),             # <<<<<<<<<<<<<<
 *     ('IBV_QUERY_QP_DATA_IN_ORDER_ALIGNED_128_BYTES', __Pyx_PyLong_From_enum__ibv_query_qp_data_in_order_caps(IBV_QUERY_QP_DATA_IN_ORDER_ALIGNED_128_BYTES)),
 * 
*/
  __pyx_t_12 = __Pyx_PyLong_From_enum__ibv_query_qp_data_in_order_caps(IBV_QUERY_QP_DATA_IN_ORDER_WHOLE_MSG); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QUERY_QP_DATA_IN_ORDER_WHOLE, __pyx_t_12); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "EnumType":25
 * ibv_query_qp_data_in_order_caps = __Pyx_FlagBase('ibv_query_qp_data_in_order_caps', [
 *     ('IBV_QUERY_QP_DATA_IN_ORDER_WHOLE_MSG', __Pyx_PyLong_From_enum__ibv_query_qp_data_in_order_caps(IBV_QUERY_QP_DATA_IN_ORDER_WHOLE_MSG)),
 *     ('IBV_QUERY_QP_DATA_IN_ORDER_ALIGNED_128_BYTES', __Pyx_PyLong_From_enum__ibv_query_qp_data_in_order_caps(IBV_QUERY_QP_DATA_IN_ORDER_ALIGNED_128_BYTES)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_12 = __Pyx_PyLong_From_enum__ibv_query_qp_data_in_order_caps(IBV_QUERY_QP_DATA_IN_ORDER_ALIGNED_128_BYTES); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QUERY_QP_DATA_IN_ORDER_ALIGN, __pyx_t_12); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_query_qp_data_in_order_caps = __Pyx_FlagBase('ibv_query_qp_data_in_order_caps', [             # <<<<<<<<<<<<<<
 *     ('IBV_QUERY_QP_DATA_IN_ORDER_WHOLE_MSG', __Pyx_PyLong_From_enum__ibv_query_qp_data_in_order_caps(IBV_QUERY_QP_DATA_IN_ORDER_WHOLE_MSG)),
 *     ('IBV_QUERY_QP_DATA_IN_ORDER_ALIGNED_128_BYTES', __Pyx_PyLong_From_enum__ibv_query_qp_data_in_order_caps(IBV_QUERY_QP_DATA_IN_ORDER_ALIGNED_128_BYTES)),
*/
  __pyx_t_12 = __Pyx_PyList_Pack(2, __pyx_t_5, __pyx_t_15); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":27
 *     ('IBV_QUERY_QP_DATA_IN_ORDER_ALIGNED_128_BYTES', __Pyx_PyLong_From_enum__ibv_query_qp_data_in_order_caps(IBV_QUERY_QP_DATA_IN_ORDER_ALIGNED_128_BYTES)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_15 = __Pyx_Globals(); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_6, __pyx_mstate_global->__pyx_n_u_ibv_query_qp_data_in_order_caps, __pyx_t_12};
    __pyx_t_5 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_15, __pyx_t_5, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_11 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_14, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_query_qp_data_in_order_caps, __pyx_t_11) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":34
 * 
 * 
 *     ibv_query_qp_data_in_order_caps._member_names_ = list(ibv_query_qp_data_in_order_caps.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_ibv_query_qp_data_in_order_caps); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PySequence_ListKeepNew(__pyx_t_14); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_mstate_global->__pyx_n_u_ibv_query_qp_data_in_order_caps); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_14, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_11) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_query_qp_data_in_order_flags = __Pyx_FlagBase('ibv_query_qp_data_in_order_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_QUERY_QP_DATA_IN_ORDER_RETURN_CAPS', __Pyx_PyLong_From_enum__ibv_query_qp_data_in_order_flags(IBV_QUERY_QP_DATA_IN_ORDER_RETURN_CAPS)),
 * 
*/
  __pyx_t_11 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_5 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_query_qp_data_in_order_flags = __Pyx_FlagBase('ibv_query_qp_data_in_order_flags', [
 *     ('IBV_QUERY_QP_DATA_IN_ORDER_RETURN_CAPS', __Pyx_PyLong_From_enum__ibv_query_qp_data_in_order_flags(IBV_QUERY_QP_DATA_IN_ORDER_RETURN_CAPS)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__ibv_query_qp_data_in_order_flags(IBV_QUERY_QP_DATA_IN_ORDER_RETURN_CAPS); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_12 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_QUERY_QP_DATA_IN_ORDER_RETUR, __pyx_t_15); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_query_qp_data_in_order_flags = __Pyx_FlagBase('ibv_query_qp_data_in_order_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_QUERY_QP_DATA_IN_ORDER_RETURN_CAPS', __Pyx_PyLong_From_enum__ibv_query_qp_data_in_order_flags(IBV_QUERY_QP_DATA_IN_ORDER_RETURN_CAPS)),
 * 
*/
  __pyx_t_15 = __Pyx_PyList_Pack(1, __pyx_t_12); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "EnumType":26
 *     ('IBV_QUERY_QP_DATA_IN_ORDER_RETURN_CAPS', __Pyx_PyLong_From_enum__ibv_query_qp_data_in_order_flags(IBV_QUERY_QP_DATA_IN_ORDER_RETURN_CAPS)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_12 = __Pyx_Globals(); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_11, __pyx_mstate_global->__pyx_n_u_ibv_query_qp_data_in_order_flags, __pyx_t_15};
    __pyx_t_6 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_12, __pyx_t_6, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_14 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_query_qp_data_in_order_flags, __pyx_t_14) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

  /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":33
 * 
 * 
 *     ibv_query_qp_data_in_order_flags._member_names_ = list(ibv_query_qp_data_in_order_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_mstate_global->__pyx_n_u_ibv_query_qp_data_in_order_flags); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PySequence_ListKeepNew(__pyx_t_5); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_ibv_query_qp_data_in_order_flags); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_14) < (0)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_raw_packet_caps = __Pyx_FlagBase('ibv_raw_packet_caps', [             # <<<<<<<<<<<<<<
 *     ('IBV_RAW_PACKET_CAP_CVLAN_STRIPPING', __Pyx_PyLong_From_enum__ibv_raw_packet_caps(IBV_RAW_PACKET_CAP_CVLAN_STRIPPING)),
 *     ('IBV_RAW_PACKET_CAP_SCATTER_FCS', __Pyx_PyLong_From_enum__ibv_raw_packet_caps(IBV_RAW_PACKET_CAP_SCATTER_FCS)),
*/
  __pyx_t_14 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_6 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_raw_packet_caps = __Pyx_FlagBase('ibv_raw_packet_caps', [
 *     ('IBV_RAW_PACKET_CAP_CVLAN_STRIPPING', __Pyx_PyLong_From_enum__ibv_raw_packet_caps(IBV_RAW_PACKET_CAP_CVLAN_STRIPPING)),             # <<<<<<<<<<<<<<
 *     ('IBV_RAW_PACKET_CAP_SCATTER_FCS', __Pyx_PyLong_From_enum__ibv_raw_packet_caps(IBV_RAW_PACKET_CAP_SCATTER_FCS)),
 *     ('IBV_RAW_PACKET_CAP_IP_CSUM', __Pyx_PyLong_From_enum__ibv_raw_packet_caps(IBV_RAW_PACKET_CAP_IP_CSUM)),
*/
  __pyx_t_12 = __Pyx_PyLong_From_enum__ibv_raw_packet_caps(IBV_RAW_PACKET_CAP_CVLAN_STRIPPING); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_RAW_PACKET_CAP_CVLAN_STRIPPI, __pyx_t_12); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "EnumType":25
 * ibv_raw_packet_caps = __Pyx_FlagBase('ibv_raw_packet_caps', [
 *     ('IBV_RAW_PACKET_CAP_CVLAN_STRIPPING', __Pyx_PyLong_From_enum__ibv_raw_packet_caps(IBV_RAW_PACKET_CAP_CVLAN_STRIPPING)),
 *     ('IBV_RAW_PACKET_CAP_SCATTER_FCS', __Pyx_PyLong_From_enum__ibv_raw_packet_caps(IBV_RAW_PACKET_CAP_SCATTER_FCS)),             # <<<<<<<<<<<<<<
 *     ('IBV_RAW_PACKET_CAP_IP_CSUM', __Pyx_PyLong_From_enum__ibv_raw_packet_caps(IBV_RAW_PACKET_CAP_IP_CSUM)),
 *     ('IBV_RAW_PACKET_CAP_DELAY_DROP', __Pyx_PyLong_From_enum__ibv_raw_packet_caps(IBV_RAW_PACKET_CAP_DELAY_DROP)),
*/
  __pyx_t_12 = __Pyx_PyLong_From_enum__ibv_raw_packet_caps(IBV_RAW_PACKET_CAP_SCATTER_FCS); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_RAW_PACKET_CAP_SCATTER_FCS, __pyx_t_12); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "EnumType":26
 *     ('IBV_RAW_PACKET_CAP_CVLAN_STRIPPING', __Pyx_PyLong_From_enum__ibv_raw_packet_caps(IBV_RAW_PACKET_CAP_CVLAN_STRIPPING)),
 *     ('IBV_RAW_PACKET_CAP_SCATTER_FCS', __Pyx_PyLong_From_enum__ibv_raw_packet_caps(IBV_RAW_PACKET_CAP_SCATTER_FCS)),
 *     ('IBV_RAW_PACKET_CAP_IP_CSUM', __Pyx_PyLong_From_enum__ibv_raw_packet_caps(IBV_RAW_PACKET_CAP_IP_CSUM)),             # <<<<<<<<<<<<<<
 *     ('IBV_RAW_PACKET_CAP_DELAY_DROP', __Pyx_PyLong_From_enum__ibv_raw_packet_caps(IBV_RAW_PACKET_CAP_DELAY_DROP)),
 * 
*/
  __pyx_t_12 = __Pyx_PyLong_From_enum__ibv_raw_packet_caps(IBV_RAW_PACKET_CAP_IP_CSUM); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_30 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_RAW_PACKET_CAP_IP_CSUM, __pyx_t_12); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "EnumType":27
 *     ('IBV_RAW_PACKET_CAP_SCATTER_FCS', __Pyx_PyLong_From_enum__ibv_raw_packet_caps(IBV_RAW_PACKET_CAP_SCATTER_FCS)),
 *     ('IBV_RAW_PACKET_CAP_IP_CSUM', __Pyx_PyLong_From_enum__ibv_raw_packet_caps(IBV_RAW_PACKET_CAP_IP_CSUM)),
 *     ('IBV_RAW_PACKET_CAP_DELAY_DROP', __Pyx_PyLong_From_enum__ibv_raw_packet_caps(IBV_RAW_PACKET_CAP_DELAY_DROP)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_12 = __Pyx_PyLong_From_enum__ibv_raw_packet_caps(IBV_RAW_PACKET_CAP_DELAY_DROP); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_7 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_RAW_PACKET_CAP_DELAY_DROP, __pyx_t_12); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_raw_packet_caps = __Pyx_FlagBase('ibv_raw_packet_caps', [             # <<<<<<<<<<<<<<
 *     ('IBV_RAW_PACKET_CAP_CVLAN_STRIPPING', __Pyx_PyLong_From_enum__ibv_raw_packet_caps(IBV_RAW_PACKET_CAP_CVLAN_STRIPPING)),
 *     ('IBV_RAW_PACKET_CAP_SCATTER_FCS', __Pyx_PyLong_From_enum__ibv_raw_packet_caps(IBV_RAW_PACKET_CAP_SCATTER_FCS)),
*/
  __pyx_t_12 = __Pyx_PyList_Pack(4, __pyx_t_15, __pyx_t_11, __pyx_t_30, __pyx_t_7); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "EnumType":29
 *     ('IBV_RAW_PACKET_CAP_DELAY_DROP', __Pyx_PyLong_From_enum__ibv_raw_packet_caps(IBV_RAW_PACKET_CAP_DELAY_DROP)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_7 = __Pyx_Globals(); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_30 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_30, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_14, __pyx_mstate_global->__pyx_n_u_ibv_raw_packet_caps, __pyx_t_12};
    __pyx_t_30 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_7, __pyx_t_30, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_5 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_30);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_raw_packet_caps, __pyx_t_5) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumType":31
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":36
 * 
 * 
 *     ibv_raw_packet_caps._member_names_ = list(ibv_raw_packet_caps.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_ibv_raw_packet_caps); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PySequence_ListKeepNew(__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_ibv_raw_packet_caps); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_5) < (0)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "EnumType":31
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_read_counters_flags = __Pyx_FlagBase('ibv_read_counters_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_READ_COUNTERS_ATTR_PREFER_CACHED', __Pyx_PyLong_From_enum__ibv_read_counters_flags(IBV_READ_COUNTERS_ATTR_PREFER_CACHED)),
 * 
*/
  __pyx_t_5 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_30 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_read_counters_flags = __Pyx_FlagBase('ibv_read_counters_flags', [
 *     ('IBV_READ_COUNTERS_ATTR_PREFER_CACHED', __Pyx_PyLong_From_enum__ibv_read_counters_flags(IBV_READ_COUNTERS_ATTR_PREFER_CACHED)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_7 = __Pyx_PyLong_From_enum__ibv_read_counters_flags(IBV_READ_COUNTERS_ATTR_PREFER_CACHED); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_12 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_READ_COUNTERS_ATTR_PREFER_CA, __pyx_t_7); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_read_counters_flags = __Pyx_FlagBase('ibv_read_counters_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_READ_COUNTERS_ATTR_PREFER_CACHED', __Pyx_PyLong_From_enum__ibv_read_counters_flags(IBV_READ_COUNTERS_ATTR_PREFER_CACHED)),
 * 
*/
  __pyx_t_7 = __Pyx_PyList_Pack(1, __pyx_t_12); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "EnumType":26
 *     ('IBV_READ_COUNTERS_ATTR_PREFER_CACHED', __Pyx_PyLong_From_enum__ibv_read_counters_flags(IBV_READ_COUNTERS_ATTR_PREFER_CACHED)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_12 = __Pyx_Globals(); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_5, __pyx_mstate_global->__pyx_n_u_ibv_read_counters_flags, __pyx_t_7};
    __pyx_t_14 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_12, __pyx_t_14, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_6 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_30, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_14);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_read_counters_flags, __pyx_t_6) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":33
 * 
 * 
 *     ibv_read_counters_flags._member_names_ = list(ibv_read_counters_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_ibv_read_counters_flags); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_30 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PySequence_ListKeepNew(__pyx_t_30); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_30, __pyx_mstate_global->__pyx_n_u_ibv_read_counters_flags); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_30, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_6) < (0)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

    /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_rereg_mr_err_code = __Pyx_FlagBase('ibv_rereg_mr_err_code', [             # <<<<<<<<<<<<<<
 *     ('IBV_REREG_MR_ERR_INPUT', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_INPUT)),
 *     ('IBV_REREG_MR_ERR_DONT_FORK_NEW', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_DONT_FORK_NEW)),
*/
  __pyx_t_6 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_14 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_rereg_mr_err_code = __Pyx_FlagBase('ibv_rereg_mr_err_code', [
 *     ('IBV_REREG_MR_ERR_INPUT', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_INPUT)),             # <<<<<<<<<<<<<<
 *     ('IBV_REREG_MR_ERR_DONT_FORK_NEW', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_DONT_FORK_NEW)),
 *     ('IBV_REREG_MR_ERR_DO_FORK_OLD', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_DO_FORK_OLD)),
*/
  __pyx_t_12 = __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_INPUT); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_7 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_REREG_MR_ERR_INPUT, __pyx_t_12); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "EnumType":25
 * ibv_rereg_mr_err_code = __Pyx_FlagBase('ibv_rereg_mr_err_code', [
 *     ('IBV_REREG_MR_ERR_INPUT', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_INPUT)),
 *     ('IBV_REREG_MR_ERR_DONT_FORK_NEW', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_DONT_FORK_NEW)),             # <<<<<<<<<<<<<<
 *     ('IBV_REREG_MR_ERR_DO_FORK_OLD', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_DO_FORK_OLD)),
 *     ('IBV_REREG_MR_ERR_CMD', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_CMD)),
*/
  __pyx_t_12 = __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_DONT_FORK_NEW); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_REREG_MR_ERR_DONT_FORK_NEW, __pyx_t_12); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "EnumType":26
 *     ('IBV_REREG_MR_ERR_INPUT', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_INPUT)),
 *     ('IBV_REREG_MR_ERR_DONT_FORK_NEW', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_DONT_FORK_NEW)),
 *     ('IBV_REREG_MR_ERR_DO_FORK_OLD', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_DO_FORK_OLD)),             # <<<<<<<<<<<<<<
 *     ('IBV_REREG_MR_ERR_CMD', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_CMD)),
 *     ('IBV_REREG_MR_ERR_CMD_AND_DO_FORK_NEW', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_CMD_AND_DO_FORK_NEW)),
*/
  __pyx_t_12 = __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_DO_FORK_OLD); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_REREG_MR_ERR_DO_FORK_OLD, __pyx_t_12); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "EnumType":27
 *     ('IBV_REREG_MR_ERR_DONT_FORK_NEW', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_DONT_FORK_NEW)),
 *     ('IBV_REREG_MR_ERR_DO_FORK_OLD', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_DO_FORK_OLD)),
 *     ('IBV_REREG_MR_ERR_CMD', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_CMD)),             # <<<<<<<<<<<<<<
 *     ('IBV_REREG_MR_ERR_CMD_AND_DO_FORK_NEW', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_CMD_AND_DO_FORK_NEW)),
 * 
*/
  __pyx_t_12 = __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_CMD); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_REREG_MR_ERR_CMD, __pyx_t_12); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "EnumType":28
 *     ('IBV_REREG_MR_ERR_DO_FORK_OLD', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_DO_FORK_OLD)),
 *     ('IBV_REREG_MR_ERR_CMD', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_CMD)),
 *     ('IBV_REREG_MR_ERR_CMD_AND_DO_FORK_NEW', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_CMD_AND_DO_FORK_NEW)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_12 = __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_CMD_AND_DO_FORK_NEW); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_8 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_REREG_MR_ERR_CMD_AND_DO_FORK, __pyx_t_12); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_rereg_mr_err_code = __Pyx_FlagBase('ibv_rereg_mr_err_code', [             # <<<<<<<<<<<<<<
 *     ('IBV_REREG_MR_ERR_INPUT', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_INPUT)),
 *     ('IBV_REREG_MR_ERR_DONT_FORK_NEW', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_DONT_FORK_NEW)),
*/
  __pyx_t_12 = __Pyx_PyList_Pack(5, __pyx_t_7, __pyx_t_5, __pyx_t_11, __pyx_t_15, __pyx_t_8); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "EnumType":30
 *     ('IBV_REREG_MR_ERR_CMD_AND_DO_FORK_NEW', __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(IBV_REREG_MR_ERR_CMD_AND_DO_FORK_NEW)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_8 = __Pyx_Globals(); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_6, __pyx_mstate_global->__pyx_n_u_ibv_rereg_mr_err_code, __pyx_t_12};
    __pyx_t_15 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_8, __pyx_t_15, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_30 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_14, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_15);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_rereg_mr_err_code, __pyx_t_30) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":32
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":37
 * 
 * 
 *     ibv_rereg_mr_err_code._member_names_ = list(ibv_rereg_mr_err_code.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_30, __pyx_mstate_global->__pyx_n_u_ibv_rereg_mr_err_code); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_30, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __pyx_t_30 = __Pyx_PySequence_ListKeepNew(__pyx_t_14); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_30);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_mstate_global->__pyx_n_u_ibv_rereg_mr_err_code); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_14, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_30) < (0)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "EnumType":32
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_rereg_mr_flags = __Pyx_FlagBase('ibv_rereg_mr_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_REREG_MR_CHANGE_TRANSLATION', __Pyx_PyLong_From_enum__ibv_rereg_mr_flags(IBV_REREG_MR_CHANGE_TRANSLATION)),
 *     ('IBV_REREG_MR_CHANGE_PD', __Pyx_PyLong_From_enum__ibv_rereg_mr_flags(IBV_REREG_MR_CHANGE_PD)),
*/
  __pyx_t_30 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_15 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_rereg_mr_flags = __Pyx_FlagBase('ibv_rereg_mr_flags', [
 *     ('IBV_REREG_MR_CHANGE_TRANSLATION', __Pyx_PyLong_From_enum__ibv_rereg_mr_flags(IBV_REREG_MR_CHANGE_TRANSLATION)),             # <<<<<<<<<<<<<<
 *     ('IBV_REREG_MR_CHANGE_PD', __Pyx_PyLong_From_enum__ibv_rereg_mr_flags(IBV_REREG_MR_CHANGE_PD)),
 *     ('IBV_REREG_MR_CHANGE_ACCESS', __Pyx_PyLong_From_enum__ibv_rereg_mr_flags(IBV_REREG_MR_CHANGE_ACCESS)),
*/
  __pyx_t_8 = __Pyx_PyLong_From_enum__ibv_rereg_mr_flags(IBV_REREG_MR_CHANGE_TRANSLATION); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_12 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_REREG_MR_CHANGE_TRANSLATION, __pyx_t_8); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "EnumType":25
 * ibv_rereg_mr_flags = __Pyx_FlagBase('ibv_rereg_mr_flags', [
 *     ('IBV_REREG_MR_CHANGE_TRANSLATION', __Pyx_PyLong_From_enum__ibv_rereg_mr_flags(IBV_REREG_MR_CHANGE_TRANSLATION)),
 *     ('IBV_REREG_MR_CHANGE_PD', __Pyx_PyLong_From_enum__ibv_rereg_mr_flags(IBV_REREG_MR_CHANGE_PD)),             # <<<<<<<<<<<<<<
 *     ('IBV_REREG_MR_CHANGE_ACCESS', __Pyx_PyLong_From_enum__ibv_rereg_mr_flags(IBV_REREG_MR_CHANGE_ACCESS)),
 * 
*/
  __pyx_t_8 = __Pyx_PyLong_From_enum__ibv_rereg_mr_flags(IBV_REREG_MR_CHANGE_PD); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_6 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_REREG_MR_CHANGE_PD, __pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "EnumType":26
 *     ('IBV_REREG_MR_CHANGE_TRANSLATION', __Pyx_PyLong_From_enum__ibv_rereg_mr_flags(IBV_REREG_MR_CHANGE_TRANSLATION)),
 *     ('IBV_REREG_MR_CHANGE_PD', __Pyx_PyLong_From_enum__ibv_rereg_mr_flags(IBV_REREG_MR_CHANGE_PD)),
 *     ('IBV_REREG_MR_CHANGE_ACCESS', __Pyx_PyLong_From_enum__ibv_rereg_mr_flags(IBV_REREG_MR_CHANGE_ACCESS)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_8 = __Pyx_PyLong_From_enum__ibv_rereg_mr_flags(IBV_REREG_MR_CHANGE_ACCESS); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_REREG_MR_CHANGE_ACCESS, __pyx_t_8); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_rereg_mr_flags = __Pyx_FlagBase('ibv_rereg_mr_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_REREG_MR_CHANGE_TRANSLATION', __Pyx_PyLong_From_enum__ibv_rereg_mr_flags(IBV_REREG_MR_CHANGE_TRANSLATION)),
 *     ('IBV_REREG_MR_CHANGE_PD', __Pyx_PyLong_From_enum__ibv_rereg_mr_flags(IBV_REREG_MR_CHANGE_PD)),
*/
  __pyx_t_8 = __Pyx_PyList_Pack(3, __pyx_t_12, __pyx_t_6, __pyx_t_11); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":28
 *     ('IBV_REREG_MR_CHANGE_ACCESS', __Pyx_PyLong_From_enum__ibv_rereg_mr_flags(IBV_REREG_MR_CHANGE_ACCESS)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_11 = __Pyx_Globals(); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_30, __pyx_mstate_global->__pyx_n_u_ibv_rereg_mr_flags, __pyx_t_8};
    __pyx_t_6 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_11, __pyx_t_6, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_14 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_15, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_30); __pyx_t_30 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_rereg_mr_flags, __pyx_t_14) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

  /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":35
 * 
 * 
 *     ibv_rereg_mr_flags._member_names_ = list(ibv_rereg_mr_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_mstate_global->__pyx_n_u_ibv_rereg_mr_flags); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PySequence_ListKeepNew(__pyx_t_15); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_ibv_rereg_mr_flags); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_14) < (0)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_rx_hash_fields = __Pyx_FlagBase('ibv_rx_hash_fields', [             # <<<<<<<<<<<<<<
 *     ('IBV_RX_HASH_SRC_IPV4', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_SRC_IPV4)),
 *     ('IBV_RX_HASH_DST_IPV4', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_IPV4)),
*/
  __pyx_t_14 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_6 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_rx_hash_fields = __Pyx_FlagBase('ibv_rx_hash_fields', [
 *     ('IBV_RX_HASH_SRC_IPV4', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_SRC_IPV4)),             # <<<<<<<<<<<<<<
 *     ('IBV_RX_HASH_DST_IPV4', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_IPV4)),
 *     ('IBV_RX_HASH_SRC_IPV6', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_SRC_IPV6)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_SRC_IPV4); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_8 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_RX_HASH_SRC_IPV4, __pyx_t_11); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":25
 * ibv_rx_hash_fields = __Pyx_FlagBase('ibv_rx_hash_fields', [
 *     ('IBV_RX_HASH_SRC_IPV4', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_SRC_IPV4)),
 *     ('IBV_RX_HASH_DST_IPV4', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_IPV4)),             # <<<<<<<<<<<<<<
 *     ('IBV_RX_HASH_SRC_IPV6', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_SRC_IPV6)),
 *     ('IBV_RX_HASH_DST_IPV6', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_IPV6)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_IPV4); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_30 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_RX_HASH_DST_IPV4, __pyx_t_11); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":26
 *     ('IBV_RX_HASH_SRC_IPV4', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_SRC_IPV4)),
 *     ('IBV_RX_HASH_DST_IPV4', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_IPV4)),
 *     ('IBV_RX_HASH_SRC_IPV6', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_SRC_IPV6)),             # <<<<<<<<<<<<<<
 *     ('IBV_RX_HASH_DST_IPV6', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_IPV6)),
 *     ('IBV_RX_HASH_SRC_PORT_TCP', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_SRC_PORT_TCP)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_SRC_IPV6); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_RX_HASH_SRC_IPV6, __pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":27
 *     ('IBV_RX_HASH_DST_IPV4', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_IPV4)),
 *     ('IBV_RX_HASH_SRC_IPV6', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_SRC_IPV6)),
 *     ('IBV_RX_HASH_DST_IPV6', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_IPV6)),             # <<<<<<<<<<<<<<
 *     ('IBV_RX_HASH_SRC_PORT_TCP', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_SRC_PORT_TCP)),
 *     ('IBV_RX_HASH_DST_PORT_TCP', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_PORT_TCP)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_IPV6); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_RX_HASH_DST_IPV6, __pyx_t_11); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":28
 *     ('IBV_RX_HASH_SRC_IPV6', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_SRC_IPV6)),
 *     ('IBV_RX_HASH_DST_IPV6', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_IPV6)),
 *     ('IBV_RX_HASH_SRC_PORT_TCP', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_SRC_PORT_TCP)),             # <<<<<<<<<<<<<<
 *     ('IBV_RX_HASH_DST_PORT_TCP', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_PORT_TCP)),
 *     ('IBV_RX_HASH_SRC_PORT_UDP', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_SRC_PORT_UDP)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_SRC_PORT_TCP); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_7 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_RX_HASH_SRC_PORT_TCP, __pyx_t_11); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":29
 *     ('IBV_RX_HASH_DST_IPV6', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_IPV6)),
 *     ('IBV_RX_HASH_SRC_PORT_TCP', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_SRC_PORT_TCP)),
 *     ('IBV_RX_HASH_DST_PORT_TCP', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_PORT_TCP)),             # <<<<<<<<<<<<<<
 *     ('IBV_RX_HASH_SRC_PORT_UDP', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_SRC_PORT_UDP)),
 *     ('IBV_RX_HASH_DST_PORT_UDP', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_PORT_UDP)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_PORT_TCP); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_RX_HASH_DST_PORT_TCP, __pyx_t_11); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":30
 *     ('IBV_RX_HASH_SRC_PORT_TCP', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_SRC_PORT_TCP)),
 *     ('IBV_RX_HASH_DST_PORT_TCP', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_PORT_TCP)),
 *     ('IBV_RX_HASH_SRC_PORT_UDP', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_SRC_PORT_UDP)),             # <<<<<<<<<<<<<<
 *     ('IBV_RX_HASH_DST_PORT_UDP', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_PORT_UDP)),
 * 
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_SRC_PORT_UDP); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_RX_HASH_SRC_PORT_UDP, __pyx_t_11); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":31
 *     ('IBV_RX_HASH_DST_PORT_TCP', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_PORT_TCP)),
 *     ('IBV_RX_HASH_SRC_PORT_UDP', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_SRC_PORT_UDP)),
 *     ('IBV_RX_HASH_DST_PORT_UDP', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_PORT_UDP)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_PORT_UDP); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_13 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_RX_HASH_DST_PORT_UDP, __pyx_t_11); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_rx_hash_fields = __Pyx_FlagBase('ibv_rx_hash_fields', [             # <<<<<<<<<<<<<<
 *     ('IBV_RX_HASH_SRC_IPV4', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_SRC_IPV4)),
 *     ('IBV_RX_HASH_DST_IPV4', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_IPV4)),
*/
  __pyx_t_11 = __Pyx_PyList_Pack(8, __pyx_t_8, __pyx_t_30, __pyx_t_12, __pyx_t_5, __pyx_t_7, __pyx_t_22, __pyx_t_4, __pyx_t_13); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "EnumType":33
 *     ('IBV_RX_HASH_DST_PORT_UDP', __Pyx_PyLong_From_enum__ibv_rx_hash_fields(IBV_RX_HASH_DST_PORT_UDP)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_13 = __Pyx_Globals(); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_14, __pyx_mstate_global->__pyx_n_u_ibv_rx_hash_fields, __pyx_t_11};
    __pyx_t_4 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_13, __pyx_t_4, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_15 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_rx_hash_fields, __pyx_t_15) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":35
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":40
 * 
 * 
 *     ibv_rx_hash_fields._member_names_ = list(ibv_rx_hash_fields.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_ibv_rx_hash_fields); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __pyx_t_15 = __Pyx_PySequence_ListKeepNew(__pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_ibv_rx_hash_fields); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_15) < (0)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "EnumType":35
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_rx_hash_function_flags = __Pyx_FlagBase('ibv_rx_hash_function_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_RX_HASH_FUNC_TOEPLITZ', __Pyx_PyLong_From_enum__ibv_rx_hash_function_flags(IBV_RX_HASH_FUNC_TOEPLITZ)),
 * 
*/
  __pyx_t_15 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_4 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_rx_hash_function_flags = __Pyx_FlagBase('ibv_rx_hash_function_flags', [
 *     ('IBV_RX_HASH_FUNC_TOEPLITZ', __Pyx_PyLong_From_enum__ibv_rx_hash_function_flags(IBV_RX_HASH_FUNC_TOEPLITZ)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_13 = __Pyx_PyLong_From_enum__ibv_rx_hash_function_flags(IBV_RX_HASH_FUNC_TOEPLITZ); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_RX_HASH_FUNC_TOEPLITZ, __pyx_t_13); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_rx_hash_function_flags = __Pyx_FlagBase('ibv_rx_hash_function_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_RX_HASH_FUNC_TOEPLITZ', __Pyx_PyLong_From_enum__ibv_rx_hash_function_flags(IBV_RX_HASH_FUNC_TOEPLITZ)),
 * 
*/
  __pyx_t_13 = __Pyx_PyList_Pack(1, __pyx_t_11); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":26
 *     ('IBV_RX_HASH_FUNC_TOEPLITZ', __Pyx_PyLong_From_enum__ibv_rx_hash_function_flags(IBV_RX_HASH_FUNC_TOEPLITZ)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_11 = __Pyx_Globals(); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_15, __pyx_mstate_global->__pyx_n_u_ibv_rx_hash_function_flags, __pyx_t_13};
    __pyx_t_14 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_11, __pyx_t_14, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_6 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_14);
    __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_rx_hash_function_flags, __pyx_t_6) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":33
 * 
 * 
 *     ibv_rx_hash_function_flags._member_names_ = list(ibv_rx_hash_function_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_ibv_rx_hash_function_flags); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PySequence_ListKeepNew(__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_ibv_rx_hash_function_flags); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_6) < (0)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_selectivity_level = __Pyx_FlagBase('ibv_selectivity_level', [             # <<<<<<<<<<<<<<
 *     ('IBV_FLUSH_MR', __Pyx_PyLong_From_enum__ibv_selectivity_level(IBV_FLUSH_MR)),
 *     ('IBV_FLUSH_RANGE', __Pyx_PyLong_From_enum__ibv_selectivity_level(IBV_FLUSH_RANGE)),
*/
  __pyx_t_6 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_14 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_selectivity_level = __Pyx_FlagBase('ibv_selectivity_level', [
 *     ('IBV_FLUSH_MR', __Pyx_PyLong_From_enum__ibv_selectivity_level(IBV_FLUSH_MR)),             # <<<<<<<<<<<<<<
 *     ('IBV_FLUSH_RANGE', __Pyx_PyLong_From_enum__ibv_selectivity_level(IBV_FLUSH_RANGE)),
 * 
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_selectivity_level(IBV_FLUSH_MR); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_13 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLUSH_MR, __pyx_t_11); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":25
 * ibv_selectivity_level = __Pyx_FlagBase('ibv_selectivity_level', [
 *     ('IBV_FLUSH_MR', __Pyx_PyLong_From_enum__ibv_selectivity_level(IBV_FLUSH_MR)),
 *     ('IBV_FLUSH_RANGE', __Pyx_PyLong_From_enum__ibv_selectivity_level(IBV_FLUSH_RANGE)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_selectivity_level(IBV_FLUSH_RANGE); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_FLUSH_RANGE, __pyx_t_11); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_selectivity_level = __Pyx_FlagBase('ibv_selectivity_level', [             # <<<<<<<<<<<<<<
 *     ('IBV_FLUSH_MR', __Pyx_PyLong_From_enum__ibv_selectivity_level(IBV_FLUSH_MR)),
 *     ('IBV_FLUSH_RANGE', __Pyx_PyLong_From_enum__ibv_selectivity_level(IBV_FLUSH_RANGE)),
*/
  __pyx_t_11 = __Pyx_PyList_Pack(2, __pyx_t_13, __pyx_t_15); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":27
 *     ('IBV_FLUSH_RANGE', __Pyx_PyLong_From_enum__ibv_selectivity_level(IBV_FLUSH_RANGE)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_15 = __Pyx_Globals(); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_6, __pyx_mstate_global->__pyx_n_u_ibv_selectivity_level, __pyx_t_11};
    __pyx_t_13 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_15, __pyx_t_13, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_4 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_14, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_13);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_selectivity_level, __pyx_t_4) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":34
 * 
 * 
 *     ibv_selectivity_level._member_names_ = list(ibv_selectivity_level.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_ibv_selectivity_level); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PySequence_ListKeepNew(__pyx_t_14); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_mstate_global->__pyx_n_u_ibv_selectivity_level); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_14, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_4) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_send_flags = __Pyx_FlagBase('ibv_send_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_SEND_FENCE', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_FENCE)),
 *     ('IBV_SEND_SIGNALED', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_SIGNALED)),
*/
  __pyx_t_4 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_13 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_send_flags = __Pyx_FlagBase('ibv_send_flags', [
 *     ('IBV_SEND_FENCE', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_FENCE)),             # <<<<<<<<<<<<<<
 *     ('IBV_SEND_SIGNALED', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_SIGNALED)),
 *     ('IBV_SEND_SOLICITED', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_SOLICITED)),
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_FENCE); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_SEND_FENCE, __pyx_t_15); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":25
 * ibv_send_flags = __Pyx_FlagBase('ibv_send_flags', [
 *     ('IBV_SEND_FENCE', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_FENCE)),
 *     ('IBV_SEND_SIGNALED', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_SIGNALED)),             # <<<<<<<<<<<<<<
 *     ('IBV_SEND_SOLICITED', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_SOLICITED)),
 *     ('IBV_SEND_INLINE', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_INLINE)),
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_SIGNALED); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_6 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_SEND_SIGNALED, __pyx_t_15); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":26
 *     ('IBV_SEND_FENCE', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_FENCE)),
 *     ('IBV_SEND_SIGNALED', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_SIGNALED)),
 *     ('IBV_SEND_SOLICITED', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_SOLICITED)),             # <<<<<<<<<<<<<<
 *     ('IBV_SEND_INLINE', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_INLINE)),
 *     ('IBV_SEND_IP_CSUM', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_IP_CSUM)),
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_SOLICITED); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_SEND_SOLICITED, __pyx_t_15); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":27
 *     ('IBV_SEND_SIGNALED', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_SIGNALED)),
 *     ('IBV_SEND_SOLICITED', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_SOLICITED)),
 *     ('IBV_SEND_INLINE', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_INLINE)),             # <<<<<<<<<<<<<<
 *     ('IBV_SEND_IP_CSUM', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_IP_CSUM)),
 * 
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_INLINE); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_7 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_SEND_INLINE, __pyx_t_15); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":28
 *     ('IBV_SEND_SOLICITED', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_SOLICITED)),
 *     ('IBV_SEND_INLINE', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_INLINE)),
 *     ('IBV_SEND_IP_CSUM', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_IP_CSUM)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_IP_CSUM); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_SEND_IP_CSUM, __pyx_t_15); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_send_flags = __Pyx_FlagBase('ibv_send_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_SEND_FENCE', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_FENCE)),
 *     ('IBV_SEND_SIGNALED', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_SIGNALED)),
*/
  __pyx_t_15 = __Pyx_PyList_Pack(5, __pyx_t_11, __pyx_t_6, __pyx_t_22, __pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumType":30
 *     ('IBV_SEND_IP_CSUM', __Pyx_PyLong_From_enum__ibv_send_flags(IBV_SEND_IP_CSUM)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_5 = __Pyx_Globals(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_4, __pyx_mstate_global->__pyx_n_u_ibv_send_flags, __pyx_t_15};
    __pyx_t_7 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_5, __pyx_t_7, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_14 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_13, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_send_flags, __pyx_t_14) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

  /* "EnumType":32
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":37
 * 
 * 
 *     ibv_send_flags._member_names_ = list(ibv_send_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_mstate_global->__pyx_n_u_ibv_send_flags); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PySequence_ListKeepNew(__pyx_t_13); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_ibv_send_flags); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_14) < (0)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

    /* "EnumType":32
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_srq_attr_mask = __Pyx_FlagBase('ibv_srq_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_SRQ_MAX_WR', __Pyx_PyLong_From_enum__ibv_srq_attr_mask(IBV_SRQ_MAX_WR)),
 *     ('IBV_SRQ_LIMIT', __Pyx_PyLong_From_enum__ibv_srq_attr_mask(IBV_SRQ_LIMIT)),
*/
  __pyx_t_14 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_7 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_srq_attr_mask = __Pyx_FlagBase('ibv_srq_attr_mask', [
 *     ('IBV_SRQ_MAX_WR', __Pyx_PyLong_From_enum__ibv_srq_attr_mask(IBV_SRQ_MAX_WR)),             # <<<<<<<<<<<<<<
 *     ('IBV_SRQ_LIMIT', __Pyx_PyLong_From_enum__ibv_srq_attr_mask(IBV_SRQ_LIMIT)),
 * 
*/
  __pyx_t_5 = __Pyx_PyLong_From_enum__ibv_srq_attr_mask(IBV_SRQ_MAX_WR); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_SRQ_MAX_WR, __pyx_t_5); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumType":25
 * ibv_srq_attr_mask = __Pyx_FlagBase('ibv_srq_attr_mask', [
 *     ('IBV_SRQ_MAX_WR', __Pyx_PyLong_From_enum__ibv_srq_attr_mask(IBV_SRQ_MAX_WR)),
 *     ('IBV_SRQ_LIMIT', __Pyx_PyLong_From_enum__ibv_srq_attr_mask(IBV_SRQ_LIMIT)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_5 = __Pyx_PyLong_From_enum__ibv_srq_attr_mask(IBV_SRQ_LIMIT); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_SRQ_LIMIT, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_srq_attr_mask = __Pyx_FlagBase('ibv_srq_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_SRQ_MAX_WR', __Pyx_PyLong_From_enum__ibv_srq_attr_mask(IBV_SRQ_MAX_WR)),
 *     ('IBV_SRQ_LIMIT', __Pyx_PyLong_From_enum__ibv_srq_attr_mask(IBV_SRQ_LIMIT)),
*/
  __pyx_t_5 = __Pyx_PyList_Pack(2, __pyx_t_15, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":27
 *     ('IBV_SRQ_LIMIT', __Pyx_PyLong_From_enum__ibv_srq_attr_mask(IBV_SRQ_LIMIT)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_4 = __Pyx_Globals(); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_14, __pyx_mstate_global->__pyx_n_u_ibv_srq_attr_mask, __pyx_t_5};
    __pyx_t_15 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_4, __pyx_t_15, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_13 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_15);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_srq_attr_mask, __pyx_t_13) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":34
 * 
 * 
 *     ibv_srq_attr_mask._member_names_ = list(ibv_srq_attr_mask.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_ibv_srq_attr_mask); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_PySequence_ListKeepNew(__pyx_t_7); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_ibv_srq_attr_mask); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_13) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_srq_init_attr_mask = __Pyx_FlagBase('ibv_srq_init_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_SRQ_INIT_ATTR_TYPE', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_TYPE)),
 *     ('IBV_SRQ_INIT_ATTR_PD', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_PD)),
*/
  __pyx_t_13 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_15 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_srq_init_attr_mask = __Pyx_FlagBase('ibv_srq_init_attr_mask', [
 *     ('IBV_SRQ_INIT_ATTR_TYPE', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_TYPE)),             # <<<<<<<<<<<<<<
 *     ('IBV_SRQ_INIT_ATTR_PD', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_PD)),
 *     ('IBV_SRQ_INIT_ATTR_XRCD', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_XRCD)),
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_TYPE); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_SRQ_INIT_ATTR_TYPE, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":25
 * ibv_srq_init_attr_mask = __Pyx_FlagBase('ibv_srq_init_attr_mask', [
 *     ('IBV_SRQ_INIT_ATTR_TYPE', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_TYPE)),
 *     ('IBV_SRQ_INIT_ATTR_PD', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_PD)),             # <<<<<<<<<<<<<<
 *     ('IBV_SRQ_INIT_ATTR_XRCD', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_XRCD)),
 *     ('IBV_SRQ_INIT_ATTR_CQ', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_CQ)),
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_PD); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_14 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_SRQ_INIT_ATTR_PD, __pyx_t_4); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":26
 *     ('IBV_SRQ_INIT_ATTR_TYPE', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_TYPE)),
 *     ('IBV_SRQ_INIT_ATTR_PD', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_PD)),
 *     ('IBV_SRQ_INIT_ATTR_XRCD', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_XRCD)),             # <<<<<<<<<<<<<<
 *     ('IBV_SRQ_INIT_ATTR_CQ', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_CQ)),
 *     ('IBV_SRQ_INIT_ATTR_TM', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_TM)),
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_XRCD); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_SRQ_INIT_ATTR_XRCD, __pyx_t_4); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":27
 *     ('IBV_SRQ_INIT_ATTR_PD', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_PD)),
 *     ('IBV_SRQ_INIT_ATTR_XRCD', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_XRCD)),
 *     ('IBV_SRQ_INIT_ATTR_CQ', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_CQ)),             # <<<<<<<<<<<<<<
 *     ('IBV_SRQ_INIT_ATTR_TM', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_TM)),
 * 
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_CQ); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_SRQ_INIT_ATTR_CQ, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":28
 *     ('IBV_SRQ_INIT_ATTR_XRCD', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_XRCD)),
 *     ('IBV_SRQ_INIT_ATTR_CQ', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_CQ)),
 *     ('IBV_SRQ_INIT_ATTR_TM', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_TM)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_TM); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_SRQ_INIT_ATTR_TM, __pyx_t_4); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_srq_init_attr_mask = __Pyx_FlagBase('ibv_srq_init_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_SRQ_INIT_ATTR_TYPE', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_TYPE)),
 *     ('IBV_SRQ_INIT_ATTR_PD', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_PD)),
*/
  __pyx_t_4 = __Pyx_PyList_Pack(5, __pyx_t_5, __pyx_t_14, __pyx_t_22, __pyx_t_6, __pyx_t_11); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":30
 *     ('IBV_SRQ_INIT_ATTR_TM', __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(IBV_SRQ_INIT_ATTR_TM)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_11 = __Pyx_Globals(); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_13, __pyx_mstate_global->__pyx_n_u_ibv_srq_init_attr_mask, __pyx_t_4};
    __pyx_t_6 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_11, __pyx_t_6, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_7 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_15, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_srq_init_attr_mask, __pyx_t_7) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "EnumType":32
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":37
 * 
 * 
 *     ibv_srq_init_attr_mask._member_names_ = list(ibv_srq_init_attr_mask.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_ibv_srq_init_attr_mask); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PySequence_ListKeepNew(__pyx_t_15); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_ibv_srq_init_attr_mask); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_7) < (0)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":32
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_srq_type = __Pyx_FlagBase('ibv_srq_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_SRQT_BASIC', __Pyx_PyLong_From_enum__ibv_srq_type(IBV_SRQT_BASIC)),
 *     ('IBV_SRQT_XRC', __Pyx_PyLong_From_enum__ibv_srq_type(IBV_SRQT_XRC)),
*/
  __pyx_t_7 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_6 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_srq_type = __Pyx_FlagBase('ibv_srq_type', [
 *     ('IBV_SRQT_BASIC', __Pyx_PyLong_From_enum__ibv_srq_type(IBV_SRQT_BASIC)),             # <<<<<<<<<<<<<<
 *     ('IBV_SRQT_XRC', __Pyx_PyLong_From_enum__ibv_srq_type(IBV_SRQT_XRC)),
 *     ('IBV_SRQT_TM', __Pyx_PyLong_From_enum__ibv_srq_type(IBV_SRQT_TM)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_srq_type(IBV_SRQT_BASIC); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_SRQT_BASIC, __pyx_t_11); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":25
 * ibv_srq_type = __Pyx_FlagBase('ibv_srq_type', [
 *     ('IBV_SRQT_BASIC', __Pyx_PyLong_From_enum__ibv_srq_type(IBV_SRQT_BASIC)),
 *     ('IBV_SRQT_XRC', __Pyx_PyLong_From_enum__ibv_srq_type(IBV_SRQT_XRC)),             # <<<<<<<<<<<<<<
 *     ('IBV_SRQT_TM', __Pyx_PyLong_From_enum__ibv_srq_type(IBV_SRQT_TM)),
 * 
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_srq_type(IBV_SRQT_XRC); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_13 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_SRQT_XRC, __pyx_t_11); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":26
 *     ('IBV_SRQT_BASIC', __Pyx_PyLong_From_enum__ibv_srq_type(IBV_SRQT_BASIC)),
 *     ('IBV_SRQT_XRC', __Pyx_PyLong_From_enum__ibv_srq_type(IBV_SRQT_XRC)),
 *     ('IBV_SRQT_TM', __Pyx_PyLong_From_enum__ibv_srq_type(IBV_SRQT_TM)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_srq_type(IBV_SRQT_TM); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_SRQT_TM, __pyx_t_11); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_srq_type = __Pyx_FlagBase('ibv_srq_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_SRQT_BASIC', __Pyx_PyLong_From_enum__ibv_srq_type(IBV_SRQT_BASIC)),
 *     ('IBV_SRQT_XRC', __Pyx_PyLong_From_enum__ibv_srq_type(IBV_SRQT_XRC)),
*/
  __pyx_t_11 = __Pyx_PyList_Pack(3, __pyx_t_4, __pyx_t_13, __pyx_t_22); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":28
 *     ('IBV_SRQT_TM', __Pyx_PyLong_From_enum__ibv_srq_type(IBV_SRQT_TM)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_22 = __Pyx_Globals(); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_7, __pyx_mstate_global->__pyx_n_u_ibv_srq_type, __pyx_t_11};
    __pyx_t_13 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_22, __pyx_t_13, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_15 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_13);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_srq_type, __pyx_t_15) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":35
 * 
 * 
 *     ibv_srq_type._member_names_ = list(ibv_srq_type.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_ibv_srq_type); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __pyx_t_15 = __Pyx_PySequence_ListKeepNew(__pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_ibv_srq_type); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_15) < (0)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_tm_cap_flags = __Pyx_FlagBase('ibv_tm_cap_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_TM_CAP_RC', __Pyx_PyLong_From_enum__ibv_tm_cap_flags(IBV_TM_CAP_RC)),
 * 
*/
  __pyx_t_15 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_13 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_tm_cap_flags = __Pyx_FlagBase('ibv_tm_cap_flags', [
 *     ('IBV_TM_CAP_RC', __Pyx_PyLong_From_enum__ibv_tm_cap_flags(IBV_TM_CAP_RC)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_22 = __Pyx_PyLong_From_enum__ibv_tm_cap_flags(IBV_TM_CAP_RC); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_TM_CAP_RC, __pyx_t_22); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_tm_cap_flags = __Pyx_FlagBase('ibv_tm_cap_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_TM_CAP_RC', __Pyx_PyLong_From_enum__ibv_tm_cap_flags(IBV_TM_CAP_RC)),
 * 
*/
  __pyx_t_22 = __Pyx_PyList_Pack(1, __pyx_t_11); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":26
 *     ('IBV_TM_CAP_RC', __Pyx_PyLong_From_enum__ibv_tm_cap_flags(IBV_TM_CAP_RC)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_11 = __Pyx_Globals(); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_15, __pyx_mstate_global->__pyx_n_u_ibv_tm_cap_flags, __pyx_t_22};
    __pyx_t_7 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_11, __pyx_t_7, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_6 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_13, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_tm_cap_flags, __pyx_t_6) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":33
 * 
 * 
 *     ibv_tm_cap_flags._member_names_ = list(ibv_tm_cap_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_ibv_tm_cap_flags); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PySequence_ListKeepNew(__pyx_t_13); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_ibv_tm_cap_flags); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_6) < (0)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

    /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_tmh_op = __Pyx_FlagBase('ibv_tmh_op', [             # <<<<<<<<<<<<<<
 *     ('IBV_TMH_NO_TAG', __Pyx_PyLong_From_enum__ibv_tmh_op(IBV_TMH_NO_TAG)),
 *     ('IBV_TMH_RNDV', __Pyx_PyLong_From_enum__ibv_tmh_op(IBV_TMH_RNDV)),
*/
  __pyx_t_6 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_7 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_tmh_op = __Pyx_FlagBase('ibv_tmh_op', [
 *     ('IBV_TMH_NO_TAG', __Pyx_PyLong_From_enum__ibv_tmh_op(IBV_TMH_NO_TAG)),             # <<<<<<<<<<<<<<
 *     ('IBV_TMH_RNDV', __Pyx_PyLong_From_enum__ibv_tmh_op(IBV_TMH_RNDV)),
 *     ('IBV_TMH_FIN', __Pyx_PyLong_From_enum__ibv_tmh_op(IBV_TMH_FIN)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_tmh_op(IBV_TMH_NO_TAG); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_TMH_NO_TAG, __pyx_t_11); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":25
 * ibv_tmh_op = __Pyx_FlagBase('ibv_tmh_op', [
 *     ('IBV_TMH_NO_TAG', __Pyx_PyLong_From_enum__ibv_tmh_op(IBV_TMH_NO_TAG)),
 *     ('IBV_TMH_RNDV', __Pyx_PyLong_From_enum__ibv_tmh_op(IBV_TMH_RNDV)),             # <<<<<<<<<<<<<<
 *     ('IBV_TMH_FIN', __Pyx_PyLong_From_enum__ibv_tmh_op(IBV_TMH_FIN)),
 *     ('IBV_TMH_EAGER', __Pyx_PyLong_From_enum__ibv_tmh_op(IBV_TMH_EAGER)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_tmh_op(IBV_TMH_RNDV); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_TMH_RNDV, __pyx_t_11); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":26
 *     ('IBV_TMH_NO_TAG', __Pyx_PyLong_From_enum__ibv_tmh_op(IBV_TMH_NO_TAG)),
 *     ('IBV_TMH_RNDV', __Pyx_PyLong_From_enum__ibv_tmh_op(IBV_TMH_RNDV)),
 *     ('IBV_TMH_FIN', __Pyx_PyLong_From_enum__ibv_tmh_op(IBV_TMH_FIN)),             # <<<<<<<<<<<<<<
 *     ('IBV_TMH_EAGER', __Pyx_PyLong_From_enum__ibv_tmh_op(IBV_TMH_EAGER)),
 * 
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_tmh_op(IBV_TMH_FIN); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_TMH_FIN, __pyx_t_11); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":27
 *     ('IBV_TMH_RNDV', __Pyx_PyLong_From_enum__ibv_tmh_op(IBV_TMH_RNDV)),
 *     ('IBV_TMH_FIN', __Pyx_PyLong_From_enum__ibv_tmh_op(IBV_TMH_FIN)),
 *     ('IBV_TMH_EAGER', __Pyx_PyLong_From_enum__ibv_tmh_op(IBV_TMH_EAGER)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__ibv_tmh_op(IBV_TMH_EAGER); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_14 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_TMH_EAGER, __pyx_t_11); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_tmh_op = __Pyx_FlagBase('ibv_tmh_op', [             # <<<<<<<<<<<<<<
 *     ('IBV_TMH_NO_TAG', __Pyx_PyLong_From_enum__ibv_tmh_op(IBV_TMH_NO_TAG)),
 *     ('IBV_TMH_RNDV', __Pyx_PyLong_From_enum__ibv_tmh_op(IBV_TMH_RNDV)),
*/
  __pyx_t_11 = __Pyx_PyList_Pack(4, __pyx_t_22, __pyx_t_15, __pyx_t_4, __pyx_t_14); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

  /* "EnumType":29
 *     ('IBV_TMH_EAGER', __Pyx_PyLong_From_enum__ibv_tmh_op(IBV_TMH_EAGER)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_14 = __Pyx_Globals(); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_6, __pyx_mstate_global->__pyx_n_u_ibv_tmh_op, __pyx_t_11};
    __pyx_t_4 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_14, __pyx_t_4, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_13 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_tmh_op, __pyx_t_13) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "EnumType":31
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":36
 * 
 * 
 *     ibv_tmh_op._member_names_ = list(ibv_tmh_op.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_ibv_tmh_op); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_PySequence_ListKeepNew(__pyx_t_7); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_ibv_tmh_op); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_13) < (0)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "EnumType":31
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_tph_mem_type = __Pyx_FlagBase('ibv_tph_mem_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_TPH_MEM_TYPE_VM', __Pyx_PyLong_From_enum__ibv_tph_mem_type(IBV_TPH_MEM_TYPE_VM)),
 *     ('IBV_TPH_MEM_TYPE_PM', __Pyx_PyLong_From_enum__ibv_tph_mem_type(IBV_TPH_MEM_TYPE_PM)),
*/
  __pyx_t_13 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_4 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_tph_mem_type = __Pyx_FlagBase('ibv_tph_mem_type', [
 *     ('IBV_TPH_MEM_TYPE_VM', __Pyx_PyLong_From_enum__ibv_tph_mem_type(IBV_TPH_MEM_TYPE_VM)),             # <<<<<<<<<<<<<<
 *     ('IBV_TPH_MEM_TYPE_PM', __Pyx_PyLong_From_enum__ibv_tph_mem_type(IBV_TPH_MEM_TYPE_PM)),
 * 
*/
  __pyx_t_14 = __Pyx_PyLong_From_enum__ibv_tph_mem_type(IBV_TPH_MEM_TYPE_VM); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_TPH_MEM_TYPE_VM, __pyx_t_14); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

  /* "EnumType":25
 * ibv_tph_mem_type = __Pyx_FlagBase('ibv_tph_mem_type', [
 *     ('IBV_TPH_MEM_TYPE_VM', __Pyx_PyLong_From_enum__ibv_tph_mem_type(IBV_TPH_MEM_TYPE_VM)),
 *     ('IBV_TPH_MEM_TYPE_PM', __Pyx_PyLong_From_enum__ibv_tph_mem_type(IBV_TPH_MEM_TYPE_PM)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_14 = __Pyx_PyLong_From_enum__ibv_tph_mem_type(IBV_TPH_MEM_TYPE_PM); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_6 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_TPH_MEM_TYPE_PM, __pyx_t_14); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_tph_mem_type = __Pyx_FlagBase('ibv_tph_mem_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_TPH_MEM_TYPE_VM', __Pyx_PyLong_From_enum__ibv_tph_mem_type(IBV_TPH_MEM_TYPE_VM)),
 *     ('IBV_TPH_MEM_TYPE_PM', __Pyx_PyLong_From_enum__ibv_tph_mem_type(IBV_TPH_MEM_TYPE_PM)),
*/
  __pyx_t_14 = __Pyx_PyList_Pack(2, __pyx_t_11, __pyx_t_6); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":27
 *     ('IBV_TPH_MEM_TYPE_PM', __Pyx_PyLong_From_enum__ibv_tph_mem_type(IBV_TPH_MEM_TYPE_PM)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_6 = __Pyx_Globals(); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_13, __pyx_mstate_global->__pyx_n_u_ibv_tph_mem_type, __pyx_t_14};
    __pyx_t_11 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_6, __pyx_t_11, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_7 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_11);
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_tph_mem_type, __pyx_t_7) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":34
 * 
 * 
 *     ibv_tph_mem_type._member_names_ = list(ibv_tph_mem_type.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_ibv_tph_mem_type); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PySequence_ListKeepNew(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_ibv_tph_mem_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_7) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_transport_type = __Pyx_FlagBase('ibv_transport_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_TRANSPORT_UNKNOWN', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_UNKNOWN)),
 *     ('IBV_TRANSPORT_IB', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_IB)),
*/
  __pyx_t_7 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_11 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_transport_type = __Pyx_FlagBase('ibv_transport_type', [
 *     ('IBV_TRANSPORT_UNKNOWN', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_UNKNOWN)),             # <<<<<<<<<<<<<<
 *     ('IBV_TRANSPORT_IB', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_IB)),
 *     ('IBV_TRANSPORT_IWARP', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_IWARP)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_UNKNOWN); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_14 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_TRANSPORT_UNKNOWN, __pyx_t_6); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":25
 * ibv_transport_type = __Pyx_FlagBase('ibv_transport_type', [
 *     ('IBV_TRANSPORT_UNKNOWN', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_UNKNOWN)),
 *     ('IBV_TRANSPORT_IB', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_IB)),             # <<<<<<<<<<<<<<
 *     ('IBV_TRANSPORT_IWARP', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_IWARP)),
 *     ('IBV_TRANSPORT_USNIC', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_USNIC)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_IB); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_13 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_TRANSPORT_IB, __pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":26
 *     ('IBV_TRANSPORT_UNKNOWN', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_UNKNOWN)),
 *     ('IBV_TRANSPORT_IB', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_IB)),
 *     ('IBV_TRANSPORT_IWARP', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_IWARP)),             # <<<<<<<<<<<<<<
 *     ('IBV_TRANSPORT_USNIC', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_USNIC)),
 *     ('IBV_TRANSPORT_USNIC_UDP', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_USNIC_UDP)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_IWARP); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_TRANSPORT_IWARP, __pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":27
 *     ('IBV_TRANSPORT_IB', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_IB)),
 *     ('IBV_TRANSPORT_IWARP', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_IWARP)),
 *     ('IBV_TRANSPORT_USNIC', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_USNIC)),             # <<<<<<<<<<<<<<
 *     ('IBV_TRANSPORT_USNIC_UDP', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_USNIC_UDP)),
 * 
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_USNIC); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_TRANSPORT_USNIC, __pyx_t_6); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":28
 *     ('IBV_TRANSPORT_IWARP', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_IWARP)),
 *     ('IBV_TRANSPORT_USNIC', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_USNIC)),
 *     ('IBV_TRANSPORT_USNIC_UDP', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_USNIC_UDP)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_USNIC_UDP); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_TRANSPORT_USNIC_UDP, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_transport_type = __Pyx_FlagBase('ibv_transport_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_TRANSPORT_UNKNOWN', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_UNKNOWN)),
 *     ('IBV_TRANSPORT_IB', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_IB)),
*/
  __pyx_t_6 = __Pyx_PyList_Pack(5, __pyx_t_14, __pyx_t_13, __pyx_t_15, __pyx_t_22, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumType":30
 *     ('IBV_TRANSPORT_USNIC_UDP', __Pyx_PyLong_From_enum__ibv_transport_type(IBV_TRANSPORT_USNIC_UDP)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_5 = __Pyx_Globals(); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_22 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_22, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_7, __pyx_mstate_global->__pyx_n_u_ibv_transport_type, __pyx_t_6};
    __pyx_t_22 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_5, __pyx_t_22, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_4 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_11, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_22);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_transport_type, __pyx_t_4) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":32
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":37
 * 
 * 
 *     ibv_transport_type._member_names_ = list(ibv_transport_type.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_ibv_transport_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PySequence_ListKeepNew(__pyx_t_11); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_ibv_transport_type); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_4) < (0)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "EnumType":32
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_values_mask = __Pyx_FlagBase('ibv_values_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_VALUES_MASK_RAW_CLOCK', __Pyx_PyLong_From_enum__ibv_values_mask(IBV_VALUES_MASK_RAW_CLOCK)),
 * 
*/
  __pyx_t_4 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_22 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_values_mask = __Pyx_FlagBase('ibv_values_mask', [
 *     ('IBV_VALUES_MASK_RAW_CLOCK', __Pyx_PyLong_From_enum__ibv_values_mask(IBV_VALUES_MASK_RAW_CLOCK)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_5 = __Pyx_PyLong_From_enum__ibv_values_mask(IBV_VALUES_MASK_RAW_CLOCK); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_VALUES_MASK_RAW_CLOCK, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_values_mask = __Pyx_FlagBase('ibv_values_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_VALUES_MASK_RAW_CLOCK', __Pyx_PyLong_From_enum__ibv_values_mask(IBV_VALUES_MASK_RAW_CLOCK)),
 * 
*/
  __pyx_t_5 = __Pyx_PyList_Pack(1, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":26
 *     ('IBV_VALUES_MASK_RAW_CLOCK', __Pyx_PyLong_From_enum__ibv_values_mask(IBV_VALUES_MASK_RAW_CLOCK)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_6 = __Pyx_Globals(); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_4, __pyx_mstate_global->__pyx_n_u_ibv_values_mask, __pyx_t_5};
    __pyx_t_7 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_6, __pyx_t_7, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_11 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_22, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_values_mask, __pyx_t_11) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":33
 * 
 * 
 *     ibv_values_mask._member_names_ = list(ibv_values_mask.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_ibv_values_mask); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_22 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PySequence_ListKeepNew(__pyx_t_22); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_ibv_values_mask); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_11) < (0)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

    /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_wc_flags = __Pyx_FlagBase('ibv_wc_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_WC_GRH', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_GRH)),
 *     ('IBV_WC_WITH_IMM', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_WITH_IMM)),
*/
  __pyx_t_11 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_7 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_wc_flags = __Pyx_FlagBase('ibv_wc_flags', [
 *     ('IBV_WC_GRH', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_GRH)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_WITH_IMM', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_WITH_IMM)),
 *     ('IBV_WC_IP_CSUM_OK', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_IP_CSUM_OK)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_GRH); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_GRH, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":25
 * ibv_wc_flags = __Pyx_FlagBase('ibv_wc_flags', [
 *     ('IBV_WC_GRH', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_GRH)),
 *     ('IBV_WC_WITH_IMM', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_WITH_IMM)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_IP_CSUM_OK', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_IP_CSUM_OK)),
 *     ('IBV_WC_WITH_INV', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_WITH_INV)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_WITH_IMM); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_WITH_IMM, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":26
 *     ('IBV_WC_GRH', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_GRH)),
 *     ('IBV_WC_WITH_IMM', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_WITH_IMM)),
 *     ('IBV_WC_IP_CSUM_OK', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_IP_CSUM_OK)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_WITH_INV', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_WITH_INV)),
 *     ('IBV_WC_TM_SYNC_REQ', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_TM_SYNC_REQ)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_IP_CSUM_OK); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_IP_CSUM_OK, __pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":27
 *     ('IBV_WC_WITH_IMM', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_WITH_IMM)),
 *     ('IBV_WC_IP_CSUM_OK', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_IP_CSUM_OK)),
 *     ('IBV_WC_WITH_INV', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_WITH_INV)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_TM_SYNC_REQ', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_TM_SYNC_REQ)),
 *     ('IBV_WC_TM_MATCH', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_TM_MATCH)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_WITH_INV); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_13 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_WITH_INV, __pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":28
 *     ('IBV_WC_IP_CSUM_OK', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_IP_CSUM_OK)),
 *     ('IBV_WC_WITH_INV', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_WITH_INV)),
 *     ('IBV_WC_TM_SYNC_REQ', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_TM_SYNC_REQ)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_TM_MATCH', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_TM_MATCH)),
 *     ('IBV_WC_TM_DATA_VALID', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_TM_DATA_VALID)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_TM_SYNC_REQ); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_14 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_SYNC_REQ, __pyx_t_6); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":29
 *     ('IBV_WC_WITH_INV', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_WITH_INV)),
 *     ('IBV_WC_TM_SYNC_REQ', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_TM_SYNC_REQ)),
 *     ('IBV_WC_TM_MATCH', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_TM_MATCH)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_TM_DATA_VALID', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_TM_DATA_VALID)),
 * 
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_TM_MATCH); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_12 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_MATCH, __pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":30
 *     ('IBV_WC_TM_SYNC_REQ', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_TM_SYNC_REQ)),
 *     ('IBV_WC_TM_MATCH', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_TM_MATCH)),
 *     ('IBV_WC_TM_DATA_VALID', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_TM_DATA_VALID)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_TM_DATA_VALID); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_30 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_DATA_VALID, __pyx_t_6); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_wc_flags = __Pyx_FlagBase('ibv_wc_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_WC_GRH', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_GRH)),
 *     ('IBV_WC_WITH_IMM', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_WITH_IMM)),
*/
  __pyx_t_6 = __Pyx_PyList_Pack(7, __pyx_t_5, __pyx_t_4, __pyx_t_15, __pyx_t_13, __pyx_t_14, __pyx_t_12, __pyx_t_30); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":32
 *     ('IBV_WC_TM_DATA_VALID', __Pyx_PyLong_From_enum__ibv_wc_flags(IBV_WC_TM_DATA_VALID)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_30 = __Pyx_Globals(); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_30, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __pyx_t_30 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_11, __pyx_mstate_global->__pyx_n_u_ibv_wc_flags, __pyx_t_6};
    __pyx_t_12 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_30, __pyx_t_12, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_22 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_12);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_wc_flags, __pyx_t_22) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":34
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":39
 * 
 * 
 *     ibv_wc_flags._member_names_ = list(ibv_wc_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_ibv_wc_flags); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __pyx_t_22 = __Pyx_PySequence_ListKeepNew(__pyx_t_7); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_ibv_wc_flags); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_22) < (0)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "EnumType":34
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_wc_opcode = __Pyx_FlagBase('ibv_wc_opcode', [             # <<<<<<<<<<<<<<
 *     ('IBV_WC_SEND', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_SEND)),
 *     ('IBV_WC_RDMA_WRITE', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RDMA_WRITE)),
*/
  __pyx_t_22 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_12 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_wc_opcode = __Pyx_FlagBase('ibv_wc_opcode', [
 *     ('IBV_WC_SEND', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_SEND)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_RDMA_WRITE', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RDMA_WRITE)),
 *     ('IBV_WC_RDMA_READ', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RDMA_READ)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_SEND); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_6 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_SEND, __pyx_t_30); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":25
 * ibv_wc_opcode = __Pyx_FlagBase('ibv_wc_opcode', [
 *     ('IBV_WC_SEND', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_SEND)),
 *     ('IBV_WC_RDMA_WRITE', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RDMA_WRITE)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_RDMA_READ', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RDMA_READ)),
 *     ('IBV_WC_COMP_SWAP', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_COMP_SWAP)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RDMA_WRITE); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_RDMA_WRITE, __pyx_t_30); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":26
 *     ('IBV_WC_SEND', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_SEND)),
 *     ('IBV_WC_RDMA_WRITE', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RDMA_WRITE)),
 *     ('IBV_WC_RDMA_READ', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RDMA_READ)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_COMP_SWAP', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_COMP_SWAP)),
 *     ('IBV_WC_FETCH_ADD', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_FETCH_ADD)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RDMA_READ); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_14 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_RDMA_READ, __pyx_t_30); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":27
 *     ('IBV_WC_RDMA_WRITE', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RDMA_WRITE)),
 *     ('IBV_WC_RDMA_READ', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RDMA_READ)),
 *     ('IBV_WC_COMP_SWAP', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_COMP_SWAP)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_FETCH_ADD', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_FETCH_ADD)),
 *     ('IBV_WC_BIND_MW', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_BIND_MW)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_COMP_SWAP); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_13 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_COMP_SWAP, __pyx_t_30); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":28
 *     ('IBV_WC_RDMA_READ', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RDMA_READ)),
 *     ('IBV_WC_COMP_SWAP', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_COMP_SWAP)),
 *     ('IBV_WC_FETCH_ADD', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_FETCH_ADD)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_BIND_MW', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_BIND_MW)),
 *     ('IBV_WC_LOCAL_INV', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_LOCAL_INV)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_FETCH_ADD); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_FETCH_ADD, __pyx_t_30); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":29
 *     ('IBV_WC_COMP_SWAP', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_COMP_SWAP)),
 *     ('IBV_WC_FETCH_ADD', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_FETCH_ADD)),
 *     ('IBV_WC_BIND_MW', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_BIND_MW)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_LOCAL_INV', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_LOCAL_INV)),
 *     ('IBV_WC_TSO', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TSO)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_BIND_MW); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_BIND_MW, __pyx_t_30); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":30
 *     ('IBV_WC_FETCH_ADD', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_FETCH_ADD)),
 *     ('IBV_WC_BIND_MW', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_BIND_MW)),
 *     ('IBV_WC_LOCAL_INV', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_LOCAL_INV)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_TSO', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TSO)),
 *     ('IBV_WC_FLUSH', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_FLUSH)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_LOCAL_INV); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_LOCAL_INV, __pyx_t_30); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":31
 *     ('IBV_WC_BIND_MW', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_BIND_MW)),
 *     ('IBV_WC_LOCAL_INV', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_LOCAL_INV)),
 *     ('IBV_WC_TSO', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TSO)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_FLUSH', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_FLUSH)),
 *     ('IBV_WC_ATOMIC_WRITE', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_ATOMIC_WRITE)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TSO); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_8 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_TSO, __pyx_t_30); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":32
 *     ('IBV_WC_LOCAL_INV', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_LOCAL_INV)),
 *     ('IBV_WC_TSO', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TSO)),
 *     ('IBV_WC_FLUSH', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_FLUSH)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_ATOMIC_WRITE', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_ATOMIC_WRITE)),
 *     ('IBV_WC_RECV', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RECV)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_FLUSH); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_21 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_FLUSH, __pyx_t_30); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":33
 *     ('IBV_WC_TSO', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TSO)),
 *     ('IBV_WC_FLUSH', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_FLUSH)),
 *     ('IBV_WC_ATOMIC_WRITE', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_ATOMIC_WRITE)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_RECV', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RECV)),
 *     ('IBV_WC_RECV_RDMA_WITH_IMM', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RECV_RDMA_WITH_IMM)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_ATOMIC_WRITE); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_17 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_ATOMIC_WRITE, __pyx_t_30); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":34
 *     ('IBV_WC_FLUSH', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_FLUSH)),
 *     ('IBV_WC_ATOMIC_WRITE', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_ATOMIC_WRITE)),
 *     ('IBV_WC_RECV', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RECV)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_RECV_RDMA_WITH_IMM', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RECV_RDMA_WITH_IMM)),
 *     ('IBV_WC_TM_ADD', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_ADD)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RECV); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_27 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_RECV, __pyx_t_30); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":35
 *     ('IBV_WC_ATOMIC_WRITE', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_ATOMIC_WRITE)),
 *     ('IBV_WC_RECV', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RECV)),
 *     ('IBV_WC_RECV_RDMA_WITH_IMM', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RECV_RDMA_WITH_IMM)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_TM_ADD', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_ADD)),
 *     ('IBV_WC_TM_DEL', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_DEL)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RECV_RDMA_WITH_IMM); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_2 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_RECV_RDMA_WITH_IMM, __pyx_t_30); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":36
 *     ('IBV_WC_RECV', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RECV)),
 *     ('IBV_WC_RECV_RDMA_WITH_IMM', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RECV_RDMA_WITH_IMM)),
 *     ('IBV_WC_TM_ADD', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_ADD)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_TM_DEL', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_DEL)),
 *     ('IBV_WC_TM_SYNC', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_SYNC)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_ADD); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_10 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_ADD, __pyx_t_30); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":37
 *     ('IBV_WC_RECV_RDMA_WITH_IMM', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RECV_RDMA_WITH_IMM)),
 *     ('IBV_WC_TM_ADD', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_ADD)),
 *     ('IBV_WC_TM_DEL', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_DEL)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_TM_SYNC', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_SYNC)),
 *     ('IBV_WC_TM_RECV', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_RECV)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_DEL); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_28 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_DEL, __pyx_t_30); if (unlikely(!__pyx_t_28)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":38
 *     ('IBV_WC_TM_ADD', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_ADD)),
 *     ('IBV_WC_TM_DEL', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_DEL)),
 *     ('IBV_WC_TM_SYNC', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_SYNC)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_TM_RECV', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_RECV)),
 *     ('IBV_WC_TM_NO_TAG', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_NO_TAG)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_SYNC); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_29 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_SYNC, __pyx_t_30); if (unlikely(!__pyx_t_29)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":39
 *     ('IBV_WC_TM_DEL', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_DEL)),
 *     ('IBV_WC_TM_SYNC', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_SYNC)),
 *     ('IBV_WC_TM_RECV', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_RECV)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_TM_NO_TAG', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_NO_TAG)),
 *     ('IBV_WC_DRIVER2', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_DRIVER2)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_RECV); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_26 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_RECV, __pyx_t_30); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":40
 *     ('IBV_WC_TM_SYNC', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_SYNC)),
 *     ('IBV_WC_TM_RECV', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_RECV)),
 *     ('IBV_WC_TM_NO_TAG', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_NO_TAG)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_DRIVER2', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_DRIVER2)),
 *     ('IBV_WC_DRIVER3', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_DRIVER3)),
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_NO_TAG); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_NO_TAG, __pyx_t_30); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":41
 *     ('IBV_WC_TM_RECV', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_RECV)),
 *     ('IBV_WC_TM_NO_TAG', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_NO_TAG)),
 *     ('IBV_WC_DRIVER2', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_DRIVER2)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_DRIVER3', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_DRIVER3)),
 * 
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_DRIVER2); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_23 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_DRIVER2, __pyx_t_30); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":42
 *     ('IBV_WC_TM_NO_TAG', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_TM_NO_TAG)),
 *     ('IBV_WC_DRIVER2', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_DRIVER2)),
 *     ('IBV_WC_DRIVER3', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_DRIVER3)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_30 = __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_DRIVER3); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __pyx_t_32 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_DRIVER3, __pyx_t_30); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_wc_opcode = __Pyx_FlagBase('ibv_wc_opcode', [             # <<<<<<<<<<<<<<
 *     ('IBV_WC_SEND', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_SEND)),
 *     ('IBV_WC_RDMA_WRITE', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_RDMA_WRITE)),
*/
  __pyx_t_30 = __Pyx_PyList_Pack(19, __pyx_t_6, __pyx_t_11, __pyx_t_14, __pyx_t_13, __pyx_t_15, __pyx_t_4, __pyx_t_5, __pyx_t_8, __pyx_t_21, __pyx_t_17, __pyx_t_27, __pyx_t_2, __pyx_t_10, __pyx_t_28, __pyx_t_29, __pyx_t_26, __pyx_t_16, __pyx_t_23, __pyx_t_32); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":44
 *     ('IBV_WC_DRIVER3', __Pyx_PyLong_From_enum__ibv_wc_opcode(IBV_WC_DRIVER3)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_32 = __Pyx_Globals(); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __pyx_t_32 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_22, __pyx_mstate_global->__pyx_n_u_ibv_wc_opcode, __pyx_t_30};
    __pyx_t_23 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_32, __pyx_t_23, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_7 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_12, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_23);
    __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_wc_opcode, __pyx_t_7) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "EnumType":46
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":51
 * 
 * 
 *     ibv_wc_opcode._member_names_ = list(ibv_wc_opcode.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_ibv_wc_opcode); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PySequence_ListKeepNew(__pyx_t_12); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_ibv_wc_opcode); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_7) < (0)) __PYX_ERR(1, 51, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

    /* "EnumType":46
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_wc_status = __Pyx_FlagBase('ibv_wc_status', [             # <<<<<<<<<<<<<<
 *     ('IBV_WC_SUCCESS', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_SUCCESS)),
 *     ('IBV_WC_LOC_LEN_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_LEN_ERR)),
*/
  __pyx_t_7 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_23 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_wc_status = __Pyx_FlagBase('ibv_wc_status', [
 *     ('IBV_WC_SUCCESS', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_SUCCESS)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_LOC_LEN_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_LEN_ERR)),
 *     ('IBV_WC_LOC_QP_OP_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_QP_OP_ERR)),
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_SUCCESS); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_30 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_SUCCESS, __pyx_t_32); if (unlikely(!__pyx_t_30)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_30);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":25
 * ibv_wc_status = __Pyx_FlagBase('ibv_wc_status', [
 *     ('IBV_WC_SUCCESS', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_SUCCESS)),
 *     ('IBV_WC_LOC_LEN_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_LEN_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_LOC_QP_OP_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_QP_OP_ERR)),
 *     ('IBV_WC_LOC_EEC_OP_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_EEC_OP_ERR)),
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_LEN_ERR); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_LOC_LEN_ERR, __pyx_t_32); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":26
 *     ('IBV_WC_SUCCESS', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_SUCCESS)),
 *     ('IBV_WC_LOC_LEN_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_LEN_ERR)),
 *     ('IBV_WC_LOC_QP_OP_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_QP_OP_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_LOC_EEC_OP_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_EEC_OP_ERR)),
 *     ('IBV_WC_LOC_PROT_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_PROT_ERR)),
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_QP_OP_ERR); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_LOC_QP_OP_ERR, __pyx_t_32); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":27
 *     ('IBV_WC_LOC_LEN_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_LEN_ERR)),
 *     ('IBV_WC_LOC_QP_OP_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_QP_OP_ERR)),
 *     ('IBV_WC_LOC_EEC_OP_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_EEC_OP_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_LOC_PROT_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_PROT_ERR)),
 *     ('IBV_WC_WR_FLUSH_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_WR_FLUSH_ERR)),
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_EEC_OP_ERR); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_26 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_LOC_EEC_OP_ERR, __pyx_t_32); if (unlikely(!__pyx_t_26)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":28
 *     ('IBV_WC_LOC_QP_OP_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_QP_OP_ERR)),
 *     ('IBV_WC_LOC_EEC_OP_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_EEC_OP_ERR)),
 *     ('IBV_WC_LOC_PROT_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_PROT_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_WR_FLUSH_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_WR_FLUSH_ERR)),
 *     ('IBV_WC_MW_BIND_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_MW_BIND_ERR)),
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_PROT_ERR); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_29 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_LOC_PROT_ERR, __pyx_t_32); if (unlikely(!__pyx_t_29)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":29
 *     ('IBV_WC_LOC_EEC_OP_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_EEC_OP_ERR)),
 *     ('IBV_WC_LOC_PROT_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_PROT_ERR)),
 *     ('IBV_WC_WR_FLUSH_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_WR_FLUSH_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_MW_BIND_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_MW_BIND_ERR)),
 *     ('IBV_WC_BAD_RESP_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_BAD_RESP_ERR)),
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_WR_FLUSH_ERR); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_28 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_WR_FLUSH_ERR, __pyx_t_32); if (unlikely(!__pyx_t_28)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":30
 *     ('IBV_WC_LOC_PROT_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_PROT_ERR)),
 *     ('IBV_WC_WR_FLUSH_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_WR_FLUSH_ERR)),
 *     ('IBV_WC_MW_BIND_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_MW_BIND_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_BAD_RESP_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_BAD_RESP_ERR)),
 *     ('IBV_WC_LOC_ACCESS_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_ACCESS_ERR)),
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_MW_BIND_ERR); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_10 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_MW_BIND_ERR, __pyx_t_32); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":31
 *     ('IBV_WC_WR_FLUSH_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_WR_FLUSH_ERR)),
 *     ('IBV_WC_MW_BIND_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_MW_BIND_ERR)),
 *     ('IBV_WC_BAD_RESP_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_BAD_RESP_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_LOC_ACCESS_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_ACCESS_ERR)),
 *     ('IBV_WC_REM_INV_REQ_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_INV_REQ_ERR)),
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_BAD_RESP_ERR); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_2 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_BAD_RESP_ERR, __pyx_t_32); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":32
 *     ('IBV_WC_MW_BIND_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_MW_BIND_ERR)),
 *     ('IBV_WC_BAD_RESP_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_BAD_RESP_ERR)),
 *     ('IBV_WC_LOC_ACCESS_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_ACCESS_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_REM_INV_REQ_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_INV_REQ_ERR)),
 *     ('IBV_WC_REM_ACCESS_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_ACCESS_ERR)),
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_ACCESS_ERR); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_27 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_LOC_ACCESS_ERR, __pyx_t_32); if (unlikely(!__pyx_t_27)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_27);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":33
 *     ('IBV_WC_BAD_RESP_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_BAD_RESP_ERR)),
 *     ('IBV_WC_LOC_ACCESS_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_ACCESS_ERR)),
 *     ('IBV_WC_REM_INV_REQ_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_INV_REQ_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_REM_ACCESS_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_ACCESS_ERR)),
 *     ('IBV_WC_REM_OP_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_OP_ERR)),
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_INV_REQ_ERR); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_17 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_REM_INV_REQ_ERR, __pyx_t_32); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":34
 *     ('IBV_WC_LOC_ACCESS_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_ACCESS_ERR)),
 *     ('IBV_WC_REM_INV_REQ_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_INV_REQ_ERR)),
 *     ('IBV_WC_REM_ACCESS_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_ACCESS_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_REM_OP_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_OP_ERR)),
 *     ('IBV_WC_RETRY_EXC_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_RETRY_EXC_ERR)),
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_ACCESS_ERR); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_21 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_REM_ACCESS_ERR, __pyx_t_32); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":35
 *     ('IBV_WC_REM_INV_REQ_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_INV_REQ_ERR)),
 *     ('IBV_WC_REM_ACCESS_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_ACCESS_ERR)),
 *     ('IBV_WC_REM_OP_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_OP_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_RETRY_EXC_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_RETRY_EXC_ERR)),
 *     ('IBV_WC_RNR_RETRY_EXC_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_RNR_RETRY_EXC_ERR)),
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_OP_ERR); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_8 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_REM_OP_ERR, __pyx_t_32); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":36
 *     ('IBV_WC_REM_ACCESS_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_ACCESS_ERR)),
 *     ('IBV_WC_REM_OP_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_OP_ERR)),
 *     ('IBV_WC_RETRY_EXC_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_RETRY_EXC_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_RNR_RETRY_EXC_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_RNR_RETRY_EXC_ERR)),
 *     ('IBV_WC_LOC_RDD_VIOL_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_RDD_VIOL_ERR)),
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_RETRY_EXC_ERR); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_RETRY_EXC_ERR, __pyx_t_32); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":37
 *     ('IBV_WC_REM_OP_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_OP_ERR)),
 *     ('IBV_WC_RETRY_EXC_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_RETRY_EXC_ERR)),
 *     ('IBV_WC_RNR_RETRY_EXC_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_RNR_RETRY_EXC_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_LOC_RDD_VIOL_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_RDD_VIOL_ERR)),
 *     ('IBV_WC_REM_INV_RD_REQ_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_INV_RD_REQ_ERR)),
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_RNR_RETRY_EXC_ERR); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_RNR_RETRY_EXC_ERR, __pyx_t_32); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":38
 *     ('IBV_WC_RETRY_EXC_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_RETRY_EXC_ERR)),
 *     ('IBV_WC_RNR_RETRY_EXC_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_RNR_RETRY_EXC_ERR)),
 *     ('IBV_WC_LOC_RDD_VIOL_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_RDD_VIOL_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_REM_INV_RD_REQ_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_INV_RD_REQ_ERR)),
 *     ('IBV_WC_REM_ABORT_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_ABORT_ERR)),
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_RDD_VIOL_ERR); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_LOC_RDD_VIOL_ERR, __pyx_t_32); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":39
 *     ('IBV_WC_RNR_RETRY_EXC_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_RNR_RETRY_EXC_ERR)),
 *     ('IBV_WC_LOC_RDD_VIOL_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_RDD_VIOL_ERR)),
 *     ('IBV_WC_REM_INV_RD_REQ_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_INV_RD_REQ_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_REM_ABORT_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_ABORT_ERR)),
 *     ('IBV_WC_INV_EECN_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_INV_EECN_ERR)),
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_INV_RD_REQ_ERR); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_13 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_REM_INV_RD_REQ_ERR, __pyx_t_32); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":40
 *     ('IBV_WC_LOC_RDD_VIOL_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_RDD_VIOL_ERR)),
 *     ('IBV_WC_REM_INV_RD_REQ_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_INV_RD_REQ_ERR)),
 *     ('IBV_WC_REM_ABORT_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_ABORT_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_INV_EECN_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_INV_EECN_ERR)),
 *     ('IBV_WC_INV_EEC_STATE_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_INV_EEC_STATE_ERR)),
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_ABORT_ERR); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_14 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_REM_ABORT_ERR, __pyx_t_32); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":41
 *     ('IBV_WC_REM_INV_RD_REQ_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_INV_RD_REQ_ERR)),
 *     ('IBV_WC_REM_ABORT_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_ABORT_ERR)),
 *     ('IBV_WC_INV_EECN_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_INV_EECN_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_INV_EEC_STATE_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_INV_EEC_STATE_ERR)),
 *     ('IBV_WC_FATAL_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_FATAL_ERR)),
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_INV_EECN_ERR); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_INV_EECN_ERR, __pyx_t_32); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":42
 *     ('IBV_WC_REM_ABORT_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_REM_ABORT_ERR)),
 *     ('IBV_WC_INV_EECN_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_INV_EECN_ERR)),
 *     ('IBV_WC_INV_EEC_STATE_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_INV_EEC_STATE_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_FATAL_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_FATAL_ERR)),
 *     ('IBV_WC_RESP_TIMEOUT_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_RESP_TIMEOUT_ERR)),
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_INV_EEC_STATE_ERR); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_6 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_INV_EEC_STATE_ERR, __pyx_t_32); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":43
 *     ('IBV_WC_INV_EECN_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_INV_EECN_ERR)),
 *     ('IBV_WC_INV_EEC_STATE_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_INV_EEC_STATE_ERR)),
 *     ('IBV_WC_FATAL_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_FATAL_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_RESP_TIMEOUT_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_RESP_TIMEOUT_ERR)),
 *     ('IBV_WC_GENERAL_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_GENERAL_ERR)),
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_FATAL_ERR); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_9 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_FATAL_ERR, __pyx_t_32); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":44
 *     ('IBV_WC_INV_EEC_STATE_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_INV_EEC_STATE_ERR)),
 *     ('IBV_WC_FATAL_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_FATAL_ERR)),
 *     ('IBV_WC_RESP_TIMEOUT_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_RESP_TIMEOUT_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_GENERAL_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_GENERAL_ERR)),
 *     ('IBV_WC_TM_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_TM_ERR)),
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_RESP_TIMEOUT_ERR); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_31 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_RESP_TIMEOUT_ERR, __pyx_t_32); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":45
 *     ('IBV_WC_FATAL_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_FATAL_ERR)),
 *     ('IBV_WC_RESP_TIMEOUT_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_RESP_TIMEOUT_ERR)),
 *     ('IBV_WC_GENERAL_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_GENERAL_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_TM_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_TM_ERR)),
 *     ('IBV_WC_TM_RNDV_INCOMPLETE', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_TM_RNDV_INCOMPLETE)),
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_GENERAL_ERR); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_20 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_GENERAL_ERR, __pyx_t_32); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":46
 *     ('IBV_WC_RESP_TIMEOUT_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_RESP_TIMEOUT_ERR)),
 *     ('IBV_WC_GENERAL_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_GENERAL_ERR)),
 *     ('IBV_WC_TM_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_TM_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_WC_TM_RNDV_INCOMPLETE', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_TM_RNDV_INCOMPLETE)),
 * 
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_TM_ERR); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_24 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_ERR, __pyx_t_32); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":47
 *     ('IBV_WC_GENERAL_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_GENERAL_ERR)),
 *     ('IBV_WC_TM_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_TM_ERR)),
 *     ('IBV_WC_TM_RNDV_INCOMPLETE', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_TM_RNDV_INCOMPLETE)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_32 = __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_TM_RNDV_INCOMPLETE); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_25 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_RNDV_INCOMPLETE, __pyx_t_32); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_wc_status = __Pyx_FlagBase('ibv_wc_status', [             # <<<<<<<<<<<<<<
 *     ('IBV_WC_SUCCESS', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_SUCCESS)),
 *     ('IBV_WC_LOC_LEN_ERR', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_LOC_LEN_ERR)),
*/
  __pyx_t_32 = __Pyx_PyList_Pack(24, __pyx_t_30, __pyx_t_22, __pyx_t_16, __pyx_t_26, __pyx_t_29, __pyx_t_28, __pyx_t_10, __pyx_t_2, __pyx_t_27, __pyx_t_17, __pyx_t_21, __pyx_t_8, __pyx_t_5, __pyx_t_4, __pyx_t_15, __pyx_t_13, __pyx_t_14, __pyx_t_11, __pyx_t_6, __pyx_t_9, __pyx_t_31, __pyx_t_20, __pyx_t_24, __pyx_t_25); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":49
 *     ('IBV_WC_TM_RNDV_INCOMPLETE', __Pyx_PyLong_From_enum__ibv_wc_status(IBV_WC_TM_RNDV_INCOMPLETE)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_25 = __Pyx_Globals(); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_25, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __pyx_t_25 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_7, __pyx_mstate_global->__pyx_n_u_ibv_wc_status, __pyx_t_32};
    __pyx_t_24 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_25, __pyx_t_24, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_12 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_23, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_24);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_wc_status, __pyx_t_12) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "EnumType":51
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":56
 * 
 * 
 *     ibv_wc_status._member_names_ = list(ibv_wc_status.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_ibv_wc_status); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = __Pyx_PySequence_ListKeepNew(__pyx_t_23); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_ibv_wc_status); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_12) < (0)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

    /* "EnumType":51
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_wq_attr_mask = __Pyx_FlagBase('ibv_wq_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_WQ_ATTR_STATE', __Pyx_PyLong_From_enum__ibv_wq_attr_mask(IBV_WQ_ATTR_STATE)),
 *     ('IBV_WQ_ATTR_CURR_STATE', __Pyx_PyLong_From_enum__ibv_wq_attr_mask(IBV_WQ_ATTR_CURR_STATE)),
*/
  __pyx_t_12 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_24 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_wq_attr_mask = __Pyx_FlagBase('ibv_wq_attr_mask', [
 *     ('IBV_WQ_ATTR_STATE', __Pyx_PyLong_From_enum__ibv_wq_attr_mask(IBV_WQ_ATTR_STATE)),             # <<<<<<<<<<<<<<
 *     ('IBV_WQ_ATTR_CURR_STATE', __Pyx_PyLong_From_enum__ibv_wq_attr_mask(IBV_WQ_ATTR_CURR_STATE)),
 *     ('IBV_WQ_ATTR_FLAGS', __Pyx_PyLong_From_enum__ibv_wq_attr_mask(IBV_WQ_ATTR_FLAGS)),
*/
  __pyx_t_25 = __Pyx_PyLong_From_enum__ibv_wq_attr_mask(IBV_WQ_ATTR_STATE); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_32 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WQ_ATTR_STATE, __pyx_t_25); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":25
 * ibv_wq_attr_mask = __Pyx_FlagBase('ibv_wq_attr_mask', [
 *     ('IBV_WQ_ATTR_STATE', __Pyx_PyLong_From_enum__ibv_wq_attr_mask(IBV_WQ_ATTR_STATE)),
 *     ('IBV_WQ_ATTR_CURR_STATE', __Pyx_PyLong_From_enum__ibv_wq_attr_mask(IBV_WQ_ATTR_CURR_STATE)),             # <<<<<<<<<<<<<<
 *     ('IBV_WQ_ATTR_FLAGS', __Pyx_PyLong_From_enum__ibv_wq_attr_mask(IBV_WQ_ATTR_FLAGS)),
 * 
*/
  __pyx_t_25 = __Pyx_PyLong_From_enum__ibv_wq_attr_mask(IBV_WQ_ATTR_CURR_STATE); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_7 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WQ_ATTR_CURR_STATE, __pyx_t_25); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":26
 *     ('IBV_WQ_ATTR_STATE', __Pyx_PyLong_From_enum__ibv_wq_attr_mask(IBV_WQ_ATTR_STATE)),
 *     ('IBV_WQ_ATTR_CURR_STATE', __Pyx_PyLong_From_enum__ibv_wq_attr_mask(IBV_WQ_ATTR_CURR_STATE)),
 *     ('IBV_WQ_ATTR_FLAGS', __Pyx_PyLong_From_enum__ibv_wq_attr_mask(IBV_WQ_ATTR_FLAGS)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_25 = __Pyx_PyLong_From_enum__ibv_wq_attr_mask(IBV_WQ_ATTR_FLAGS); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_20 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WQ_ATTR_FLAGS, __pyx_t_25); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_wq_attr_mask = __Pyx_FlagBase('ibv_wq_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_WQ_ATTR_STATE', __Pyx_PyLong_From_enum__ibv_wq_attr_mask(IBV_WQ_ATTR_STATE)),
 *     ('IBV_WQ_ATTR_CURR_STATE', __Pyx_PyLong_From_enum__ibv_wq_attr_mask(IBV_WQ_ATTR_CURR_STATE)),
*/
  __pyx_t_25 = __Pyx_PyList_Pack(3, __pyx_t_32, __pyx_t_7, __pyx_t_20); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":28
 *     ('IBV_WQ_ATTR_FLAGS', __Pyx_PyLong_From_enum__ibv_wq_attr_mask(IBV_WQ_ATTR_FLAGS)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_20 = __Pyx_Globals(); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_20, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_12, __pyx_mstate_global->__pyx_n_u_ibv_wq_attr_mask, __pyx_t_25};
    __pyx_t_7 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_20, __pyx_t_7, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_23 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_24, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_wq_attr_mask, __pyx_t_23) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":35
 * 
 * 
 *     ibv_wq_attr_mask._member_names_ = list(ibv_wq_attr_mask.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_ibv_wq_attr_mask); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __pyx_t_23 = __Pyx_PySequence_ListKeepNew(__pyx_t_24); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_mstate_global->__pyx_n_u_ibv_wq_attr_mask); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_24, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_23) < (0)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_wq_flags = __Pyx_FlagBase('ibv_wq_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_WQ_FLAGS_CVLAN_STRIPPING', __Pyx_PyLong_From_enum__ibv_wq_flags(IBV_WQ_FLAGS_CVLAN_STRIPPING)),
 *     ('IBV_WQ_FLAGS_SCATTER_FCS', __Pyx_PyLong_From_enum__ibv_wq_flags(IBV_WQ_FLAGS_SCATTER_FCS)),
*/
  __pyx_t_23 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_7 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_wq_flags = __Pyx_FlagBase('ibv_wq_flags', [
 *     ('IBV_WQ_FLAGS_CVLAN_STRIPPING', __Pyx_PyLong_From_enum__ibv_wq_flags(IBV_WQ_FLAGS_CVLAN_STRIPPING)),             # <<<<<<<<<<<<<<
 *     ('IBV_WQ_FLAGS_SCATTER_FCS', __Pyx_PyLong_From_enum__ibv_wq_flags(IBV_WQ_FLAGS_SCATTER_FCS)),
 *     ('IBV_WQ_FLAGS_DELAY_DROP', __Pyx_PyLong_From_enum__ibv_wq_flags(IBV_WQ_FLAGS_DELAY_DROP)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_wq_flags(IBV_WQ_FLAGS_CVLAN_STRIPPING); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_25 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WQ_FLAGS_CVLAN_STRIPPING, __pyx_t_20); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":25
 * ibv_wq_flags = __Pyx_FlagBase('ibv_wq_flags', [
 *     ('IBV_WQ_FLAGS_CVLAN_STRIPPING', __Pyx_PyLong_From_enum__ibv_wq_flags(IBV_WQ_FLAGS_CVLAN_STRIPPING)),
 *     ('IBV_WQ_FLAGS_SCATTER_FCS', __Pyx_PyLong_From_enum__ibv_wq_flags(IBV_WQ_FLAGS_SCATTER_FCS)),             # <<<<<<<<<<<<<<
 *     ('IBV_WQ_FLAGS_DELAY_DROP', __Pyx_PyLong_From_enum__ibv_wq_flags(IBV_WQ_FLAGS_DELAY_DROP)),
 *     ('IBV_WQ_FLAGS_PCI_WRITE_END_PADDING', __Pyx_PyLong_From_enum__ibv_wq_flags(IBV_WQ_FLAGS_PCI_WRITE_END_PADDING)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_wq_flags(IBV_WQ_FLAGS_SCATTER_FCS); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_12 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WQ_FLAGS_SCATTER_FCS, __pyx_t_20); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":26
 *     ('IBV_WQ_FLAGS_CVLAN_STRIPPING', __Pyx_PyLong_From_enum__ibv_wq_flags(IBV_WQ_FLAGS_CVLAN_STRIPPING)),
 *     ('IBV_WQ_FLAGS_SCATTER_FCS', __Pyx_PyLong_From_enum__ibv_wq_flags(IBV_WQ_FLAGS_SCATTER_FCS)),
 *     ('IBV_WQ_FLAGS_DELAY_DROP', __Pyx_PyLong_From_enum__ibv_wq_flags(IBV_WQ_FLAGS_DELAY_DROP)),             # <<<<<<<<<<<<<<
 *     ('IBV_WQ_FLAGS_PCI_WRITE_END_PADDING', __Pyx_PyLong_From_enum__ibv_wq_flags(IBV_WQ_FLAGS_PCI_WRITE_END_PADDING)),
 * 
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_wq_flags(IBV_WQ_FLAGS_DELAY_DROP); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_32 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WQ_FLAGS_DELAY_DROP, __pyx_t_20); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":27
 *     ('IBV_WQ_FLAGS_SCATTER_FCS', __Pyx_PyLong_From_enum__ibv_wq_flags(IBV_WQ_FLAGS_SCATTER_FCS)),
 *     ('IBV_WQ_FLAGS_DELAY_DROP', __Pyx_PyLong_From_enum__ibv_wq_flags(IBV_WQ_FLAGS_DELAY_DROP)),
 *     ('IBV_WQ_FLAGS_PCI_WRITE_END_PADDING', __Pyx_PyLong_From_enum__ibv_wq_flags(IBV_WQ_FLAGS_PCI_WRITE_END_PADDING)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_wq_flags(IBV_WQ_FLAGS_PCI_WRITE_END_PADDING); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_31 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WQ_FLAGS_PCI_WRITE_END_PADDI, __pyx_t_20); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_wq_flags = __Pyx_FlagBase('ibv_wq_flags', [             # <<<<<<<<<<<<<<
 *     ('IBV_WQ_FLAGS_CVLAN_STRIPPING', __Pyx_PyLong_From_enum__ibv_wq_flags(IBV_WQ_FLAGS_CVLAN_STRIPPING)),
 *     ('IBV_WQ_FLAGS_SCATTER_FCS', __Pyx_PyLong_From_enum__ibv_wq_flags(IBV_WQ_FLAGS_SCATTER_FCS)),
*/
  __pyx_t_20 = __Pyx_PyList_Pack(4, __pyx_t_25, __pyx_t_12, __pyx_t_32, __pyx_t_31); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":29
 *     ('IBV_WQ_FLAGS_PCI_WRITE_END_PADDING', __Pyx_PyLong_From_enum__ibv_wq_flags(IBV_WQ_FLAGS_PCI_WRITE_END_PADDING)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_31 = __Pyx_Globals(); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_32 = __Pyx_PyObject_GetAttrStr(__pyx_t_31, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __pyx_t_31 = __Pyx_PyObject_Call(__pyx_t_32, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_23, __pyx_mstate_global->__pyx_n_u_ibv_wq_flags, __pyx_t_20};
    __pyx_t_32 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_31, __pyx_t_32, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_24 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_32);
    __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_wq_flags, __pyx_t_24) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

  /* "EnumType":31
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":36
 * 
 * 
 *     ibv_wq_flags._member_names_ = list(ibv_wq_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_mstate_global->__pyx_n_u_ibv_wq_flags); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PySequence_ListKeepNew(__pyx_t_7); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_ibv_wq_flags); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_24) < (0)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "EnumType":31
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_wq_init_attr_mask = __Pyx_FlagBase('ibv_wq_init_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_WQ_INIT_ATTR_FLAGS', __Pyx_PyLong_From_enum__ibv_wq_init_attr_mask(IBV_WQ_INIT_ATTR_FLAGS)),
 * 
*/
  __pyx_t_24 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_32 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_wq_init_attr_mask = __Pyx_FlagBase('ibv_wq_init_attr_mask', [
 *     ('IBV_WQ_INIT_ATTR_FLAGS', __Pyx_PyLong_From_enum__ibv_wq_init_attr_mask(IBV_WQ_INIT_ATTR_FLAGS)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_31 = __Pyx_PyLong_From_enum__ibv_wq_init_attr_mask(IBV_WQ_INIT_ATTR_FLAGS); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __pyx_t_20 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WQ_INIT_ATTR_FLAGS, __pyx_t_31); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_wq_init_attr_mask = __Pyx_FlagBase('ibv_wq_init_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_WQ_INIT_ATTR_FLAGS', __Pyx_PyLong_From_enum__ibv_wq_init_attr_mask(IBV_WQ_INIT_ATTR_FLAGS)),
 * 
*/
  __pyx_t_31 = __Pyx_PyList_Pack(1, __pyx_t_20); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":26
 *     ('IBV_WQ_INIT_ATTR_FLAGS', __Pyx_PyLong_From_enum__ibv_wq_init_attr_mask(IBV_WQ_INIT_ATTR_FLAGS)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_20 = __Pyx_Globals(); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_20, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_24, __pyx_mstate_global->__pyx_n_u_ibv_wq_init_attr_mask, __pyx_t_31};
    __pyx_t_23 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_20, __pyx_t_23, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_7 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_32, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_23);
    __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_wq_init_attr_mask, __pyx_t_7) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":33
 * 
 * 
 *     ibv_wq_init_attr_mask._member_names_ = list(ibv_wq_init_attr_mask.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_ibv_wq_init_attr_mask); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_32 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PySequence_ListKeepNew(__pyx_t_32); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_ibv_wq_init_attr_mask); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_7) < (0)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

    /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_wq_state = __Pyx_FlagBase('ibv_wq_state', [             # <<<<<<<<<<<<<<
 *     ('IBV_WQS_RESET', __Pyx_PyLong_From_enum__ibv_wq_state(IBV_WQS_RESET)),
 *     ('IBV_WQS_RDY', __Pyx_PyLong_From_enum__ibv_wq_state(IBV_WQS_RDY)),
*/
  __pyx_t_7 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_23 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_wq_state = __Pyx_FlagBase('ibv_wq_state', [
 *     ('IBV_WQS_RESET', __Pyx_PyLong_From_enum__ibv_wq_state(IBV_WQS_RESET)),             # <<<<<<<<<<<<<<
 *     ('IBV_WQS_RDY', __Pyx_PyLong_From_enum__ibv_wq_state(IBV_WQS_RDY)),
 *     ('IBV_WQS_ERR', __Pyx_PyLong_From_enum__ibv_wq_state(IBV_WQS_ERR)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_wq_state(IBV_WQS_RESET); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_31 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WQS_RESET, __pyx_t_20); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":25
 * ibv_wq_state = __Pyx_FlagBase('ibv_wq_state', [
 *     ('IBV_WQS_RESET', __Pyx_PyLong_From_enum__ibv_wq_state(IBV_WQS_RESET)),
 *     ('IBV_WQS_RDY', __Pyx_PyLong_From_enum__ibv_wq_state(IBV_WQS_RDY)),             # <<<<<<<<<<<<<<
 *     ('IBV_WQS_ERR', __Pyx_PyLong_From_enum__ibv_wq_state(IBV_WQS_ERR)),
 *     ('IBV_WQS_UNKNOWN', __Pyx_PyLong_From_enum__ibv_wq_state(IBV_WQS_UNKNOWN)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_wq_state(IBV_WQS_RDY); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_24 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WQS_RDY, __pyx_t_20); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":26
 *     ('IBV_WQS_RESET', __Pyx_PyLong_From_enum__ibv_wq_state(IBV_WQS_RESET)),
 *     ('IBV_WQS_RDY', __Pyx_PyLong_From_enum__ibv_wq_state(IBV_WQS_RDY)),
 *     ('IBV_WQS_ERR', __Pyx_PyLong_From_enum__ibv_wq_state(IBV_WQS_ERR)),             # <<<<<<<<<<<<<<
 *     ('IBV_WQS_UNKNOWN', __Pyx_PyLong_From_enum__ibv_wq_state(IBV_WQS_UNKNOWN)),
 * 
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_wq_state(IBV_WQS_ERR); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_12 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WQS_ERR, __pyx_t_20); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":27
 *     ('IBV_WQS_RDY', __Pyx_PyLong_From_enum__ibv_wq_state(IBV_WQS_RDY)),
 *     ('IBV_WQS_ERR', __Pyx_PyLong_From_enum__ibv_wq_state(IBV_WQS_ERR)),
 *     ('IBV_WQS_UNKNOWN', __Pyx_PyLong_From_enum__ibv_wq_state(IBV_WQS_UNKNOWN)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_wq_state(IBV_WQS_UNKNOWN); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_25 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WQS_UNKNOWN, __pyx_t_20); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_wq_state = __Pyx_FlagBase('ibv_wq_state', [             # <<<<<<<<<<<<<<
 *     ('IBV_WQS_RESET', __Pyx_PyLong_From_enum__ibv_wq_state(IBV_WQS_RESET)),
 *     ('IBV_WQS_RDY', __Pyx_PyLong_From_enum__ibv_wq_state(IBV_WQS_RDY)),
*/
  __pyx_t_20 = __Pyx_PyList_Pack(4, __pyx_t_31, __pyx_t_24, __pyx_t_12, __pyx_t_25); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":29
 *     ('IBV_WQS_UNKNOWN', __Pyx_PyLong_From_enum__ibv_wq_state(IBV_WQS_UNKNOWN)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_25 = __Pyx_Globals(); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_25, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __pyx_t_25 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_7, __pyx_mstate_global->__pyx_n_u_ibv_wq_state, __pyx_t_20};
    __pyx_t_12 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_25, __pyx_t_12, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_32 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_23, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_12);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_wq_state, __pyx_t_32) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;

  /* "EnumType":31
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":36
 * 
 * 
 *     ibv_wq_state._member_names_ = list(ibv_wq_state.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_ibv_wq_state); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
    __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_32, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    __pyx_t_32 = __Pyx_PySequence_ListKeepNew(__pyx_t_23); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_32);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_ibv_wq_state); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_32) < (0)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

    /* "EnumType":31
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_wq_type = __Pyx_FlagBase('ibv_wq_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_WQT_RQ', __Pyx_PyLong_From_enum__ibv_wq_type(IBV_WQT_RQ)),
 * 
*/
  __pyx_t_32 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_12 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_wq_type = __Pyx_FlagBase('ibv_wq_type', [
 *     ('IBV_WQT_RQ', __Pyx_PyLong_From_enum__ibv_wq_type(IBV_WQT_RQ)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_25 = __Pyx_PyLong_From_enum__ibv_wq_type(IBV_WQT_RQ); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __pyx_t_20 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WQT_RQ, __pyx_t_25); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_wq_type = __Pyx_FlagBase('ibv_wq_type', [             # <<<<<<<<<<<<<<
 *     ('IBV_WQT_RQ', __Pyx_PyLong_From_enum__ibv_wq_type(IBV_WQT_RQ)),
 * 
*/
  __pyx_t_25 = __Pyx_PyList_Pack(1, __pyx_t_20); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":26
 *     ('IBV_WQT_RQ', __Pyx_PyLong_From_enum__ibv_wq_type(IBV_WQT_RQ)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_20 = __Pyx_Globals(); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_20, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_32, __pyx_mstate_global->__pyx_n_u_ibv_wq_type, __pyx_t_25};
    __pyx_t_7 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_20, __pyx_t_7, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_23 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_12, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_32); __pyx_t_32 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_wq_type, __pyx_t_23) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":33
 * 
 * 
 *     ibv_wq_type._member_names_ = list(ibv_wq_type.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_ibv_wq_type); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __pyx_t_23 = __Pyx_PySequence_ListKeepNew(__pyx_t_12); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_ibv_wq_type); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_23) < (0)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

    /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_wr_opcode = __Pyx_FlagBase('ibv_wr_opcode', [             # <<<<<<<<<<<<<<
 *     ('IBV_WR_RDMA_WRITE', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_RDMA_WRITE)),
 *     ('IBV_WR_RDMA_WRITE_WITH_IMM', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_RDMA_WRITE_WITH_IMM)),
*/
  __pyx_t_23 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_7 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_wr_opcode = __Pyx_FlagBase('ibv_wr_opcode', [
 *     ('IBV_WR_RDMA_WRITE', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_RDMA_WRITE)),             # <<<<<<<<<<<<<<
 *     ('IBV_WR_RDMA_WRITE_WITH_IMM', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_RDMA_WRITE_WITH_IMM)),
 *     ('IBV_WR_SEND', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_SEND)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_RDMA_WRITE); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_25 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WR_RDMA_WRITE, __pyx_t_20); if (unlikely(!__pyx_t_25)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":25
 * ibv_wr_opcode = __Pyx_FlagBase('ibv_wr_opcode', [
 *     ('IBV_WR_RDMA_WRITE', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_RDMA_WRITE)),
 *     ('IBV_WR_RDMA_WRITE_WITH_IMM', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_RDMA_WRITE_WITH_IMM)),             # <<<<<<<<<<<<<<
 *     ('IBV_WR_SEND', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_SEND)),
 *     ('IBV_WR_SEND_WITH_IMM', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_SEND_WITH_IMM)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_RDMA_WRITE_WITH_IMM); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_32 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WR_RDMA_WRITE_WITH_IMM, __pyx_t_20); if (unlikely(!__pyx_t_32)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":26
 *     ('IBV_WR_RDMA_WRITE', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_RDMA_WRITE)),
 *     ('IBV_WR_RDMA_WRITE_WITH_IMM', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_RDMA_WRITE_WITH_IMM)),
 *     ('IBV_WR_SEND', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_SEND)),             # <<<<<<<<<<<<<<
 *     ('IBV_WR_SEND_WITH_IMM', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_SEND_WITH_IMM)),
 *     ('IBV_WR_RDMA_READ', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_RDMA_READ)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_SEND); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_24 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WR_SEND, __pyx_t_20); if (unlikely(!__pyx_t_24)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":27
 *     ('IBV_WR_RDMA_WRITE_WITH_IMM', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_RDMA_WRITE_WITH_IMM)),
 *     ('IBV_WR_SEND', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_SEND)),
 *     ('IBV_WR_SEND_WITH_IMM', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_SEND_WITH_IMM)),             # <<<<<<<<<<<<<<
 *     ('IBV_WR_RDMA_READ', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_RDMA_READ)),
 *     ('IBV_WR_ATOMIC_CMP_AND_SWP', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_ATOMIC_CMP_AND_SWP)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_SEND_WITH_IMM); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_31 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WR_SEND_WITH_IMM, __pyx_t_20); if (unlikely(!__pyx_t_31)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_31);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":28
 *     ('IBV_WR_SEND', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_SEND)),
 *     ('IBV_WR_SEND_WITH_IMM', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_SEND_WITH_IMM)),
 *     ('IBV_WR_RDMA_READ', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_RDMA_READ)),             # <<<<<<<<<<<<<<
 *     ('IBV_WR_ATOMIC_CMP_AND_SWP', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_ATOMIC_CMP_AND_SWP)),
 *     ('IBV_WR_ATOMIC_FETCH_AND_ADD', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_ATOMIC_FETCH_AND_ADD)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_RDMA_READ); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_9 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WR_RDMA_READ, __pyx_t_20); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":29
 *     ('IBV_WR_SEND_WITH_IMM', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_SEND_WITH_IMM)),
 *     ('IBV_WR_RDMA_READ', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_RDMA_READ)),
 *     ('IBV_WR_ATOMIC_CMP_AND_SWP', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_ATOMIC_CMP_AND_SWP)),             # <<<<<<<<<<<<<<
 *     ('IBV_WR_ATOMIC_FETCH_AND_ADD', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_ATOMIC_FETCH_AND_ADD)),
 *     ('IBV_WR_LOCAL_INV', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_LOCAL_INV)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_ATOMIC_CMP_AND_SWP); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_6 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WR_ATOMIC_CMP_AND_SWP, __pyx_t_20); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":30
 *     ('IBV_WR_RDMA_READ', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_RDMA_READ)),
 *     ('IBV_WR_ATOMIC_CMP_AND_SWP', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_ATOMIC_CMP_AND_SWP)),
 *     ('IBV_WR_ATOMIC_FETCH_AND_ADD', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_ATOMIC_FETCH_AND_ADD)),             # <<<<<<<<<<<<<<
 *     ('IBV_WR_LOCAL_INV', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_LOCAL_INV)),
 *     ('IBV_WR_BIND_MW', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_BIND_MW)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_ATOMIC_FETCH_AND_ADD); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WR_ATOMIC_FETCH_AND_ADD, __pyx_t_20); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":31
 *     ('IBV_WR_ATOMIC_CMP_AND_SWP', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_ATOMIC_CMP_AND_SWP)),
 *     ('IBV_WR_ATOMIC_FETCH_AND_ADD', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_ATOMIC_FETCH_AND_ADD)),
 *     ('IBV_WR_LOCAL_INV', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_LOCAL_INV)),             # <<<<<<<<<<<<<<
 *     ('IBV_WR_BIND_MW', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_BIND_MW)),
 *     ('IBV_WR_SEND_WITH_INV', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_SEND_WITH_INV)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_LOCAL_INV); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_14 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WR_LOCAL_INV, __pyx_t_20); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":32
 *     ('IBV_WR_ATOMIC_FETCH_AND_ADD', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_ATOMIC_FETCH_AND_ADD)),
 *     ('IBV_WR_LOCAL_INV', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_LOCAL_INV)),
 *     ('IBV_WR_BIND_MW', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_BIND_MW)),             # <<<<<<<<<<<<<<
 *     ('IBV_WR_SEND_WITH_INV', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_SEND_WITH_INV)),
 *     ('IBV_WR_TSO', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_TSO)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_BIND_MW); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_13 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WR_BIND_MW, __pyx_t_20); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":33
 *     ('IBV_WR_LOCAL_INV', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_LOCAL_INV)),
 *     ('IBV_WR_BIND_MW', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_BIND_MW)),
 *     ('IBV_WR_SEND_WITH_INV', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_SEND_WITH_INV)),             # <<<<<<<<<<<<<<
 *     ('IBV_WR_TSO', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_TSO)),
 *     ('IBV_WR_FLUSH', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_FLUSH)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_SEND_WITH_INV); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WR_SEND_WITH_INV, __pyx_t_20); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":34
 *     ('IBV_WR_BIND_MW', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_BIND_MW)),
 *     ('IBV_WR_SEND_WITH_INV', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_SEND_WITH_INV)),
 *     ('IBV_WR_TSO', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_TSO)),             # <<<<<<<<<<<<<<
 *     ('IBV_WR_FLUSH', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_FLUSH)),
 *     ('IBV_WR_ATOMIC_WRITE', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_ATOMIC_WRITE)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_TSO); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WR_TSO, __pyx_t_20); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":35
 *     ('IBV_WR_SEND_WITH_INV', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_SEND_WITH_INV)),
 *     ('IBV_WR_TSO', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_TSO)),
 *     ('IBV_WR_FLUSH', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_FLUSH)),             # <<<<<<<<<<<<<<
 *     ('IBV_WR_ATOMIC_WRITE', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_ATOMIC_WRITE)),
 * 
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_FLUSH); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WR_FLUSH, __pyx_t_20); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":36
 *     ('IBV_WR_TSO', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_TSO)),
 *     ('IBV_WR_FLUSH', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_FLUSH)),
 *     ('IBV_WR_ATOMIC_WRITE', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_ATOMIC_WRITE)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_ATOMIC_WRITE); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_8 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_WR_ATOMIC_WRITE, __pyx_t_20); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_wr_opcode = __Pyx_FlagBase('ibv_wr_opcode', [             # <<<<<<<<<<<<<<
 *     ('IBV_WR_RDMA_WRITE', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_RDMA_WRITE)),
 *     ('IBV_WR_RDMA_WRITE_WITH_IMM', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_RDMA_WRITE_WITH_IMM)),
*/
  __pyx_t_20 = __Pyx_PyList_Pack(13, __pyx_t_25, __pyx_t_32, __pyx_t_24, __pyx_t_31, __pyx_t_9, __pyx_t_6, __pyx_t_11, __pyx_t_14, __pyx_t_13, __pyx_t_15, __pyx_t_4, __pyx_t_5, __pyx_t_8); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "EnumType":38
 *     ('IBV_WR_ATOMIC_WRITE', __Pyx_PyLong_From_enum__ibv_wr_opcode(IBV_WR_ATOMIC_WRITE)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_8 = __Pyx_Globals(); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_23, __pyx_mstate_global->__pyx_n_u_ibv_wr_opcode, __pyx_t_20};
    __pyx_t_5 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_8, __pyx_t_5, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_12 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_wr_opcode, __pyx_t_12) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "EnumType":40
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":45
 * 
 * 
 *     ibv_wr_opcode._member_names_ = list(ibv_wr_opcode.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_ibv_wr_opcode); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = __Pyx_PySequence_ListKeepNew(__pyx_t_7); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_ibv_wr_opcode); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_12) < (0)) __PYX_ERR(1, 45, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "EnumType":40
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * ibv_xrcd_init_attr_mask = __Pyx_FlagBase('ibv_xrcd_init_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_XRCD_INIT_ATTR_FD', __Pyx_PyLong_From_enum__ibv_xrcd_init_attr_mask(IBV_XRCD_INIT_ATTR_FD)),
 *     ('IBV_XRCD_INIT_ATTR_OFLAGS', __Pyx_PyLong_From_enum__ibv_xrcd_init_attr_mask(IBV_XRCD_INIT_ATTR_OFLAGS)),
*/
  __pyx_t_12 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_5 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * ibv_xrcd_init_attr_mask = __Pyx_FlagBase('ibv_xrcd_init_attr_mask', [
 *     ('IBV_XRCD_INIT_ATTR_FD', __Pyx_PyLong_From_enum__ibv_xrcd_init_attr_mask(IBV_XRCD_INIT_ATTR_FD)),             # <<<<<<<<<<<<<<
 *     ('IBV_XRCD_INIT_ATTR_OFLAGS', __Pyx_PyLong_From_enum__ibv_xrcd_init_attr_mask(IBV_XRCD_INIT_ATTR_OFLAGS)),
 *     ('IBV_XRCD_INIT_ATTR_RESERVED', __Pyx_PyLong_From_enum__ibv_xrcd_init_attr_mask(IBV_XRCD_INIT_ATTR_RESERVED)),
*/
  __pyx_t_8 = __Pyx_PyLong_From_enum__ibv_xrcd_init_attr_mask(IBV_XRCD_INIT_ATTR_FD); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_20 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_XRCD_INIT_ATTR_FD, __pyx_t_8); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "EnumType":25
 * ibv_xrcd_init_attr_mask = __Pyx_FlagBase('ibv_xrcd_init_attr_mask', [
 *     ('IBV_XRCD_INIT_ATTR_FD', __Pyx_PyLong_From_enum__ibv_xrcd_init_attr_mask(IBV_XRCD_INIT_ATTR_FD)),
 *     ('IBV_XRCD_INIT_ATTR_OFLAGS', __Pyx_PyLong_From_enum__ibv_xrcd_init_attr_mask(IBV_XRCD_INIT_ATTR_OFLAGS)),             # <<<<<<<<<<<<<<
 *     ('IBV_XRCD_INIT_ATTR_RESERVED', __Pyx_PyLong_From_enum__ibv_xrcd_init_attr_mask(IBV_XRCD_INIT_ATTR_RESERVED)),
 * 
*/
  __pyx_t_8 = __Pyx_PyLong_From_enum__ibv_xrcd_init_attr_mask(IBV_XRCD_INIT_ATTR_OFLAGS); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_23 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_XRCD_INIT_ATTR_OFLAGS, __pyx_t_8); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "EnumType":26
 *     ('IBV_XRCD_INIT_ATTR_FD', __Pyx_PyLong_From_enum__ibv_xrcd_init_attr_mask(IBV_XRCD_INIT_ATTR_FD)),
 *     ('IBV_XRCD_INIT_ATTR_OFLAGS', __Pyx_PyLong_From_enum__ibv_xrcd_init_attr_mask(IBV_XRCD_INIT_ATTR_OFLAGS)),
 *     ('IBV_XRCD_INIT_ATTR_RESERVED', __Pyx_PyLong_From_enum__ibv_xrcd_init_attr_mask(IBV_XRCD_INIT_ATTR_RESERVED)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
*/
  __pyx_t_8 = __Pyx_PyLong_From_enum__ibv_xrcd_init_attr_mask(IBV_XRCD_INIT_ATTR_RESERVED); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_IBV_XRCD_INIT_ATTR_RESERVED, __pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "EnumType":23
 * 
 * 
 * ibv_xrcd_init_attr_mask = __Pyx_FlagBase('ibv_xrcd_init_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('IBV_XRCD_INIT_ATTR_FD', __Pyx_PyLong_From_enum__ibv_xrcd_init_attr_mask(IBV_XRCD_INIT_ATTR_FD)),
 *     ('IBV_XRCD_INIT_ATTR_OFLAGS', __Pyx_PyLong_From_enum__ibv_xrcd_init_attr_mask(IBV_XRCD_INIT_ATTR_OFLAGS)),
*/
  __pyx_t_8 = __Pyx_PyList_Pack(3, __pyx_t_20, __pyx_t_23, __pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":28
 *     ('IBV_XRCD_INIT_ATTR_RESERVED', __Pyx_PyLong_From_enum__ibv_xrcd_init_attr_mask(IBV_XRCD_INIT_ATTR_RESERVED)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_4 = __Pyx_Globals(); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_18 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_12, __pyx_mstate_global->__pyx_n_u_ibv_xrcd_init_attr_mask, __pyx_t_8};
    __pyx_t_23 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_4, __pyx_t_23, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_7 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_18, (3-__pyx_t_18) | (__pyx_t_18*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_23);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ibv_xrcd_init_attr_mask, __pyx_t_7) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_19 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_19) {

    /* "EnumType":35
 * 
 * 
 *     ibv_xrcd_init_attr_mask._member_names_ = list(ibv_xrcd_init_attr_mask.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_ibv_xrcd_init_attr_mask); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PySequence_ListKeepNew(__pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_ibv_xrcd_init_attr_mask); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_7) < (0)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "pyverbs/libibverbs_enums.pyx":13
 * from libc.stddef cimport size_t
 * 
 * _IBV_DEVICE_RAW_SCATTER_FCS = IBV_DEVICE_RAW_SCATTER_FCS             # <<<<<<<<<<<<<<
 * _IBV_DEVICE_PCI_WRITE_END_PADDING = IBV_DEVICE_PCI_WRITE_END_PADDING
 * _IBV_ALLOCATOR_USE_DEFAULT = <size_t>IBV_ALLOCATOR_USE_DEFAULT
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_PY_LONG_LONG(IBV_DEVICE_RAW_SCATTER_FCS); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_RAW_SCATTER_FCS, __pyx_t_5) < (0)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/libibverbs_enums.pyx":14
 * 
 * _IBV_DEVICE_RAW_SCATTER_FCS = IBV_DEVICE_RAW_SCATTER_FCS
 * _IBV_DEVICE_PCI_WRITE_END_PADDING = IBV_DEVICE_PCI_WRITE_END_PADDING             # <<<<<<<<<<<<<<
 * _IBV_ALLOCATOR_USE_DEFAULT = <size_t>IBV_ALLOCATOR_USE_DEFAULT
 * _IBV_ADVISE_MR_ADVICE_PREFETCH = IBV_ADVISE_MR_ADVICE_PREFETCH
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_PY_LONG_LONG(IBV_DEVICE_PCI_WRITE_END_PADDING); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_IBV_DEVICE_PCI_WRITE_END_PADDIN, __pyx_t_5) < (0)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/libibverbs_enums.pyx":15
 * _IBV_DEVICE_RAW_SCATTER_FCS = IBV_DEVICE_RAW_SCATTER_FCS
 * _IBV_DEVICE_PCI_WRITE_END_PADDING = IBV_DEVICE_PCI_WRITE_END_PADDING
 * _IBV_ALLOCATOR_USE_DEFAULT = <size_t>IBV_ALLOCATOR_USE_DEFAULT             # <<<<<<<<<<<<<<
 * _IBV_ADVISE_MR_ADVICE_PREFETCH = IBV_ADVISE_MR_ADVICE_PREFETCH
 * _IBV_ADVISE_MR_ADVICE_PREFETCH_WRITE = IBV_ADVISE_MR_ADVICE_PREFETCH_WRITE
*/
  __pyx_t_5 = __Pyx_PyLong_FromSize_t(((size_t)IBV_ALLOCATOR_USE_DEFAULT)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_IBV_ALLOCATOR_USE_DEFAULT, __pyx_t_5) < (0)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/libibverbs_enums.pyx":16
 * _IBV_DEVICE_PCI_WRITE_END_PADDING = IBV_DEVICE_PCI_WRITE_END_PADDING
 * _IBV_ALLOCATOR_USE_DEFAULT = <size_t>IBV_ALLOCATOR_USE_DEFAULT
 * _IBV_ADVISE_MR_ADVICE_PREFETCH = IBV_ADVISE_MR_ADVICE_PREFETCH             # <<<<<<<<<<<<<<
 * _IBV_ADVISE_MR_ADVICE_PREFETCH_WRITE = IBV_ADVISE_MR_ADVICE_PREFETCH_WRITE
 * _IBV_ADVISE_MR_FLAG_FLUSH = IBV_ADVISE_MR_FLAG_FLUSH
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_PY_LONG_LONG(IBV_ADVISE_MR_ADVICE_PREFETCH); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_IBV_ADVISE_MR_ADVICE_PREFETCH, __pyx_t_5) < (0)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/libibverbs_enums.pyx":17
 * _IBV_ALLOCATOR_USE_DEFAULT = <size_t>IBV_ALLOCATOR_USE_DEFAULT
 * _IBV_ADVISE_MR_ADVICE_PREFETCH = IBV_ADVISE_MR_ADVICE_PREFETCH
 * _IBV_ADVISE_MR_ADVICE_PREFETCH_WRITE = IBV_ADVISE_MR_ADVICE_PREFETCH_WRITE             # <<<<<<<<<<<<<<
 * _IBV_ADVISE_MR_FLAG_FLUSH = IBV_ADVISE_MR_FLAG_FLUSH
 * _IBV_ADVISE_MR_ADVICE_PREFETCH_NO_FAULT = IBV_ADVISE_MR_ADVICE_PREFETCH_NO_FAULT
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_PY_LONG_LONG(IBV_ADVISE_MR_ADVICE_PREFETCH_WRITE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_IBV_ADVISE_MR_ADVICE_PREFETCH_W, __pyx_t_5) < (0)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/libibverbs_enums.pyx":18
 * _IBV_ADVISE_MR_ADVICE_PREFETCH = IBV_ADVISE_MR_ADVICE_PREFETCH
 * _IBV_ADVISE_MR_ADVICE_PREFETCH_WRITE = IBV_ADVISE_MR_ADVICE_PREFETCH_WRITE
 * _IBV_ADVISE_MR_FLAG_FLUSH = IBV_ADVISE_MR_FLAG_FLUSH             # <<<<<<<<<<<<<<
 * _IBV_ADVISE_MR_ADVICE_PREFETCH_NO_FAULT = IBV_ADVISE_MR_ADVICE_PREFETCH_NO_FAULT
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_PY_LONG_LONG(IBV_ADVISE_MR_FLAG_FLUSH); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_IBV_ADVISE_MR_FLAG_FLUSH, __pyx_t_5) < (0)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/libibverbs_enums.pyx":19
 * _IBV_ADVISE_MR_ADVICE_PREFETCH_WRITE = IBV_ADVISE_MR_ADVICE_PREFETCH_WRITE
 * _IBV_ADVISE_MR_FLAG_FLUSH = IBV_ADVISE_MR_FLAG_FLUSH
 * _IBV_ADVISE_MR_ADVICE_PREFETCH_NO_FAULT = IBV_ADVISE_MR_ADVICE_PREFETCH_NO_FAULT             # <<<<<<<<<<<<<<
*/
  __pyx_t_5 = __Pyx_PyLong_From_unsigned_PY_LONG_LONG(IBV_ADVISE_MR_ADVICE_PREFETCH_NO_FAULT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_IBV_ADVISE_MR_ADVICE_PREFETCH_N, __pyx_t_5) < (0)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/libibverbs_enums.pyx":1
 * from pyverbs.libibverbs_enums cimport (             # <<<<<<<<<<<<<<
 *     IBV_DEVICE_RAW_SCATTER_FCS,
 *     IBV_DEVICE_PCI_WRITE_END_PADDING,
*/
  __pyx_t_5 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_test, __pyx_t_5) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /*--- Wrapped vars code ---*/
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(IBV_LINK_LAYER_ETHERNET);
    if (unlikely(!wrapped)) __PYX_ERR(2, 29, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "IBV_LINK_LAYER_ETHERNET", wrapped) < 0) __PYX_ERR(2, 29, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(IBV_LINK_LAYER_INFINIBAND);
    if (unlikely(!wrapped)) __PYX_ERR(2, 28, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "IBV_LINK_LAYER_INFINIBAND", wrapped) < 0) __PYX_ERR(2, 28, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(IBV_LINK_LAYER_UNSPECIFIED);
    if (unlikely(!wrapped)) __PYX_ERR(2, 27, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "IBV_LINK_LAYER_UNSPECIFIED", wrapped) < 0) __PYX_ERR(2, 27, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(IBV_QPF_GRH_REQUIRED);
    if (unlikely(!wrapped)) __PYX_ERR(2, 392, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "IBV_QPF_GRH_REQUIRED", wrapped) < 0) __PYX_ERR(2, 392, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(IBV_WC_STANDARD_FLAGS);
    if (unlikely(!wrapped)) __PYX_ERR(2, 461, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "IBV_WC_STANDARD_FLAGS", wrapped) < 0) __PYX_ERR(2, 461, __pyx_L1_error);
  }

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_XDECREF(__pyx_t_27);
  __Pyx_XDECREF(__pyx_t_28);
  __Pyx_XDECREF(__pyx_t_29);
  __Pyx_XDECREF(__pyx_t_30);
  __Pyx_XDECREF(__pyx_t_31);
  __Pyx_XDECREF(__pyx_t_32);
  if (__pyx_m) {
    if (__pyx_mstate->__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init pyverbs.libibverbs_enums", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init pyverbs.libibverbs_enums");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #else
  return __pyx_m;
  #endif
}
/* #### Code section: pystring_table ### */
/* #### Code section: cached_builtins ### */

static int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  return 0;
}
/* #### Code section: cached_constants ### */

static int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "EnumType":36
 *     ('IBV_ACCESS_RELAXED_ORDERING', __Pyx_PyLong_From_enum__ibv_access_flags(IBV_ACCESS_RELAXED_ORDERING)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.libibverbs_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_mstate_global->__pyx_tuple[0] = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_module_2, __pyx_mstate_global->__pyx_kp_u_pyverbs_libibverbs_enums); if (unlikely(!__pyx_mstate_global->__pyx_tuple[0])) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[0]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[0]);
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_tuple;
    for (Py_ssize_t i=0; i<1; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  {
    const struct { const unsigned int length: 6; } index[] = {{1},{1},{24},{23},{27},{18},{22},{18},{20},{27},{24},{22},{23},{21},{30},{39},{36},{25},{26},{15},{14},{15},{17},{19},{27},{24},{33},{34},{24},{24},{24},{26},{28},{20},{32},{29},{21},{29},{29},{22},{33},{28},{22},{20},{27},{21},{25},{24},{24},{21},{25},{29},{21},{14},{30},{26},{36},{27},{18},{16},{22},{20},{20},{18},{22},{21},{21},{18},{23},{18},{29},{20},{19},{20},{17},{27},{18},{25},{29},{26},{24},{20},{21},{26},{25},{27},{24},{17},{17},{17},{19},{18},{22},{18},{18},{17},{17},{26},{16},{12},{20},{15},{17},{16},{17},{15},{20},{20},{29},{26},{13},{16},{13},{12},{12},{11},{12},{11},{13},{13},{11},{13},{15},{15},{16},{20},{14},{18},{15},{22},{28},{21},{24},{20},{20},{20},{24},{21},{16},{15},{38},{38},{15},{21},{14},{22},{22},{22},{28},{23},{15},{24},{13},{22},{28},{21},{13},{22},{21},{25},{27},{27},{26},{19},{12},{19},{20},{19},{34},{19},{26},{19},{11},{24},{36},{27},{17},{25},{17},{11},{12},{13},{11},{11},{11},{11},{15},{14},{18},{10},{10},{10},{16},{16},{19},{15},{9},{10},{33},{29},{35},{25},{24},{16},{15},{26},{33},{35},{27},{22},{20},{24},{24},{25},{34},{19},{28},{28},{18},{29},{26},{31},{19},{24},{31},{21},{25},{23},{20},{21},{15},{17},{11},{11},{17},{16},{16},{13},{13},{12},{14},{44},{38},{36},{34},{29},{26},{30},{36},{20},{20},{18},{25},{20},{26},{22},{31},{20},{36},{30},{28},{22},{20},{20},{24},{24},{25},{20},{20},{24},{24},{14},{15},{16},{17},{18},{14},{11},{12},{20},{20},{20},{22},{22},{13},{14},{13},{11},{14},{12},{13},{19},{19},{16},{19},{21},{19},{23},{25},{19},{19},{14},{16},{14},{14},{23},{35},{45},{20},{29},{23},{18},{21},{17},{19},{21},{22},{16},{16},{12},{18},{10},{19},{24},{17},{16},{21},{21},{18},{19},{20},{23},{18},{16},{17},{11},{25},{20},{21},{25},{22},{17},{23},{20},{24},{11},{14},{13},{20},{13},{13},{15},{16},{14},{25},{14},{18},{10},{15},{15},{19},{11},{11},{13},{15},{10},{22},{17},{17},{28},{23},{34},{24},{22},{25},{27},{19},{14},{12},{16},{16},{17},{26},{11},{20},{20},{14},{14},{15},{10},{21},{25},{27},{7},{7},{18},{4},{3},{16},{14},{23},{21},{24},{22},{20},{23},{14},{18},{14},{18},{15},{12},{18},{13},{21},{7},{11},{13},{20},{26},{13},{17},{32},{18},{18},{19},{14},{16},{19},{28},{21},{12},{11},{31},{32},{19},{23},{21},{18},{18},{26},{21},{14},{17},{22},{12},{16},{10},{16},{18},{15},{12},{13},{13},{16},{12},{21},{12},{11},{13},{23},{8},{14},{11},{6},{10},{8},{12},{10},{8}};
    #if (CYTHON_COMPRESS_STRINGS) == 3 && __PYX_LIMITED_VERSION_HEX >= 0x030e0000 /* compression: zstd (2728 bytes) */
const char* const cstring = "(\265/\375`F#\365T\000\372o\214\022& K\265\r\377\241I\245\277\363=\356O%\346x3\023\021\333\225}\216\254-ER\275\024\317\316>\200\020\200\020\300\321\r\001\020\001\007\001m}'\034\336\310\245?Ms\364\010l\257\tV\353*\363\013\270\317\313\252\325\362\254\356\250\372[{Z\375[\037!\373*\244\257\320r\205\237\211\357su\251\227X\226\273\375=\240\302G.wE\275n\365\252#\364\235i\337T\275?rl\371T\006\027\233\276\204\247k\257\216\336\377\234g~\234\335\373D\363^\231\305F9\354\253>\241\013\357\023\017\314j\3378\276w\376\240\246&*7\376\312\230]\275\017\263B\356\225\3352K\035\225\363-4\266gw|_\273\336q]P\370Z=qxV\337\307\251u\213\252\321|\355\347\3319c\220-\316\036\\\336m\243\\\027\345\367\252\270\275\366\202\362S\033\017\357X\023$*\253\325\234\343\253\273\306\354bMv\237\212@\246\214M\363\327\345\256\303GM}q\364\014\225c\007\346\312<\215\257z\257f}\014\017\2161\337\364\303\226\357V\245\360\241\303\242<*\020\351\031\032\014\r\007\000\n\303Dr\020A\202\000@\332\312\017\331\243~\371\234\333\206\014mC(|\343\271\242\345l>\241V\035\331/\354b\251\254s1\317\227[vu\252\276\376\324\344\306\250Rx\250\2741\355\335\271\203\363\207\325\327\251\317\327\273\2609\306\267\211\376\346\251\014\227Y\375\375\034\316\325\206\224/\265\330k\313\331\275\367\364\277t\235\272\226\265\221\333\205}\032\345+jV\177\234Z\313\250^\323mB\002\201\003\001\006\024 ``@\000\205\001\n\001\022\376\361\306\250\354\310\251\215\361\260\276\027\205\004\350U9>\254\324\325Tv\300^\371\013\354\026n\272\373\370/2\004\224\217\362f\257\025\030\000=\326\335\245\3144\376Z\031te\271\330<\307\271N\351k\326eY\333\215\335\252\326\255\314\350C\366\367u\371\214\241\023\371j\213q\352.\344|\247[+\2476\312\211\336\033\331sX.\206\241\343\302\372\216\267~k\216\325{1\312\312\213\251o\2367/\327?O5\330\024#T\261[V!\374|r=\314+\203\001l\225\275zB\200\302\303\250\3775\247\230\365_+\177\342AE\252\365a\010=\306\350\000\265\367P.\273\335\265(\203\372M\363\346\371\034F4W\253\361\360\356\260\333\305v\251Z]\024w\002_\354\251\363W\373,\371<U\300;\377\355J\346\"]""\227\331\3243\353\272\345\310z\221=\037Bn\271\366\262\270\317\2053\261\251g]\224;\201L\263\333+sM\364P\375\362\343\312\006\024\022>u\231\305\360w=:&X\037\177\\o\277wxW}\202]\325\221\035g\207\320\334\365x[\276\3539\311\212\021\375\\\323\330<\202\363\261&j\363\256=w\271\366\326\026\266dhW\324>w\0352\361E\375\271\370\257\251\306W\364\010\374UoM~V\025?\216\331\365\336b\357_d\220:\260\367F\325~a\332\277\240\034;\240z\246\030\221\217\212\220Y\241\305\371\256\351t\263k+?\207W72\365\354r\025b\007\364\327\266=l\317Tkz\377\023E\232\376'GE\241@bL\022\305\241\2504x\311QaD(\024\351\241\200\374Db\370\177\306b\002\201\"U\034\n\007\017\351\361\013$\013\036S\203\211\303\361\021\246\206\003\311!Q\026\220\n\303!5\024\374\002\212b\301C\370\311Dj\220&\214I\362\017\024\223\211\304\230D\024\223\202\021a\"U\030\020\211\"i\240\360p@\032\212G\2041QH\026\223\007\222\251\3010M\"\022\n\204\342PL$\210\324\360\022FDbD \017*\323Djh<\024P\"P\236\321h\250L\212\211\362\014&\221\205$\241<\210H\222\010\323O\020\311\201\264@\361\220\026\r\007\213\007\322\244\220&\014(\212\307O\246\251\321\000\022iL\222\306dA\221\030?\221&H%\262\220(\026\277h<\"\023E\301S(\322b\241\370Ij\220\030,\024\014\322B1\231\036\322x\t\322p,\032\024\216\345Q\201\350\004Y\251\020\273d\246\311\24033\363\343\360\273kk\"\271\035v\275\353\241\334F96\246\213!\324\374\341U\010\355\347\034\345O>\3407\346\026\364\231\374\034w\374V\331s\356W\251>\264\213YW\231\314\346\317y\305\311\372\331\177N;\276\317\305\2221\214\3765\364^\267\\\003\202\371\250\342k\032\205\014\325\310L\222d\030\003\202(\010\207\2539*\251\324\003R0|\212\305X\n\211\220\242\206HH\222J\2269\037\237\305`\2147E\004\240\260\313\343\336[g\366.\255r*\347\212\235\017U\210\27130\035LI\320\001#\324&\221\370-l\336_\353\225\316\220ul\311\222\224\330\037F\306\036^\0213\030\304\242\311(8\215YbG\344\374Y\222\311${\374\030\217\217\250\200\262a\tQ'@84\022\314\nR}\235\361\210\205\314\302\020\335\243\027KU\271 \225\274\221T\357CBG\210>8\337\267X\025\032R\255\267\235\303g\0146\372\315\264j""\214\005H\322\0169'\352k\374\320\016N\372\311d\313Q=\247\007\002\352\357g\353\342\3352@\327]\314\303WA\207%\352\352\260\023%,\330\362\330\207\274\317\353\375D\276f\376V#\320\3622\326G\246\023=}N>2_\243\244\314\364\373x\325JBO\211:\005\307\200\224\232j\036\002r\222\n\3533\231\370jN\001\033M\240\030'Ng\021\003\342\362/ \316U\377\207`\224\255\n1\007oL\\\007.1\335\367\374\360|\344\000i1pQ\337\357\236\355\367_^\230\246j\020\270D\277Z\270\010\236\265\356\202\362'n\234\\\214x\342\253\375J\n\027\360\036\261\213\037\004.*\031\250\353\222n\346\351\377DJ\265,\314\007j\010\246=\201{I\022W\331\234 L\226Q\260!\032\256\336*T\354\374\024Z\332\010^\360\032)\237\036}cB\026~\240\227\002M\247\203\265\220\006\022\032\246\300xT\30146i\262\005\270M,PT\356Yx\216P\022U1L\307\3419\372\272\006\003\254ZF`\326\223e\3678\257\330\326\037\204\364@\312\351U\225\250:\032g\222\221\254\261X\210|\350~\036\031\307Q\226(,\032\244\255n\210\365\002+\230@b\037\352'\242\321++\3420\345R\021\223kF3\036iD\317\260\374\313\014\t\221x|j\261~\250vfW\373P\020`\2579x\007\001\312\034\334\260\362)\314dk\261\2603\322\372\315}\rf\211\213:\206%\216\215\233\254\217\272\204\335*\364%kY)x\027\014\264]\211{+\306\246\267@2B\352h\240\247d\220\225\200-\251\301\306>\254\350\213|\267\351\343\327P/\377\2155Y\304\372C#\007\017\236\241\033;>\253r\245 G5;n\006\235n*G\265Y\202^l\036\330C\004\303\205\241\311s\237\260\r\252\r\363\300=\263\300,:.4^\314\274\215\223\357a\274\247G\272\217\0367\315\302xHI\r\335\007~\330X\372\233\235\221\rd\032\026X9\275\220\340Hz\365\264`\246 M\237\242\\%\343H\227'}\323C}x\366\373S\t\334m\335X%1HUQ\346\374c\207\251y\025O\004tHJf\013\026\021n\327\314\376B\223\356\345\033\323\016\036\020%\362\320\000\2120:\261<\210,\2265\310\004\357\000\023\240\335\212\343\244\355\221\307\200\221\274\016\032\353\250\274\222R8\030\274\210MDYn\002\333i\307\351\344\235%\367\333@\335\027\374\020d\r\027-z\027\004F\252\023+\354n\354$,Z\006\300\274\340\275\372\036\tc\325\"\264\362~\222\337\334\024lP*\007f\342\355\306\354\247\322\367""\326\276\032\264\317\re\013)v,\366S\361\031\004\334\233u\tl\341>D\253\0204l@2T7\225\211i\252\313\275\245[\034\032\330\332\205GFp;\376\301\200f8\351\371\261\374\255#|\243\370\261w\216y\214vU.G\346p\023\324\022\360\236\021\002\263jq\222[\2013\254\264\317\023\216\271\226\253P\315$\001\347W\0140j\025\257\220_\263\000g\031\331d.s\037\\6\035\3429\327!9D2\335\225\233\275\267*8 \235\306b\277\264t0\254 N\022RA\333\216JJtz\256\211\330A\344BD\373\221\347\311@\333\261\337\352\232\000W\3127\354\252>S\276\353\377\270\002\215-L,\364\267K\346\346\222>\300/\005\351\237\033\334;\022k+\272\303\270\257\342\202\006\231\303}\203\027\177S'\020j4\014\221\025\243\201\242$\255\245&\270\313\307\367=H~\010\"\023\230q\335\027V*\340\215]\206\304\237\177\rTK\334tE\322\\\360(\274_\037l:\364)'\317\231\211=\372\022jN\013\276\347\302\271&\316%H\364\374^f\254w\003\217;-\357I\326y \02358s\310\355\225I\360\263\301\2775T{\007\033\240\206bl\310\243\234\\b0\215\340\267t\327\2101\034\333B\031\226\205\2368e\242\324\311\315-%\034\204\351\010i\272bK\r\230\0322\025\216\340\254\022\255\343\277\271\036*\3362\"\007\035\034\255\037\245j\215\315\323\r\307\271\377^U\262W\273\341\201\251A5k\345\225\031X\355\335!e\230-\365\224^\266\005\256\242\277\\\371O\231\300\220#K\245\030\267\255\335\010\024\211\257\343\307\020\277\201\366\331\335SA\221X\335\307\344o\272\357\344\221 a\300\226>\013z'\3704\314K\377\244\300\346oh\206\207`!\305\371\006\220\261\300/H\n\355&a\207\270\324EN\366\315\326h!\224\207Y\364C/Z/\246\300\030R\244\256\274\266\000\300\302\364\263\"$\232\201\214Wc\010]\376n\030\004\214\005\373\006\354\023\207\323\031`\237\3670Z\253@\000On\267\303*f~\203\033\007\031\035vbj\247\370\311\322\230R\353\275\264.8=\224\364\035}\326\257\320D\013@\031\312k2\005\363|\002L%\312\265cX\253\366GUO\244<\365\3066\204\274N\357\034\252W\262\250\177y\364N--\203\007\341";
    PyObject *data = __Pyx_DecompressString(cstring, 2728, 3);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) == 2 /* compression: bz2 (2826 bytes) */
const char* const cstring = "BZh91AY&SY\200\277\014\365\000\000\001_\200@\000\000\001\177`\277\357\377\360\277\357\377\340`\r\327\317Pu\201H\347\007\000\342\010$\321\252U\032\354v\335\335\212\347o7\247z\324\245\360\323A\000L\210\232\232jh\310\000\000\000h\032\t\221\211\005M\352\206\217I\223M\000\000\r\000\rOI\245=\004\000\000\000\000\000\036\246\200HH&\202\002\246i&\2154\000\000\000\001\t\0014\323\r&FL&&\021\204i\204\300\230H\204\020&S&\211G\246\247\246\240a\000\0006\247\234\277\014\370\260\216\300\202\r\314(D\361\237\206\025]\340\311Q1\262\t\263\347\257\247\310$\200\370\036\377h\245\t\247)4\237\301^>f\003\346\204\306\3403\273\355\351FV\230S\026\327\3650\213c \3100|\356\206\024\303h\023\007|\260&\037FN\261\301\241\017(Z\323\317\332\037h_\216Q\216\002:\221>R\266RZ\t\244\006\200\303\010+(f\024&\305K\3505Q\327\322]\013\356\0068$\000\242o\273\3004\031\207\260\307\344i.\005\365v@n\220S\022\310=\001\205\322\n\273Y\233R\203\243\304<,\213KA\331\201\207\226\006\351\200x\006,\200\325\320\301\3370=\255\307\335\226\374\245\312\022\270\307$\220}1\222\355*v\315\\\273e\310\322\263#n\317\304\202,\346\007\010\017A\272\014\273\342\310*\301\022Dj \004?\022z\005@\036\024|\333n\360\264\352{m\364\341\242\365\255U\2132\351\256\266\264Gd\014 \375B\035#\270\211\226\227\273\361A\251\\$\367\2509\352\215\260s\234,\243\225\027a<\3045\213\231rVm)\243\210\013\256\030\270h/\210\024\002\201\247{]\364On\265\251\264W\266s\372\375e\336<\355\331c\205\255\007dY\211Ik~\337\021\272\341\357\230\311^\260\254\314i\231K&\307k@\334\245\323b\343\306\014#v\024\230\270\207\275\2136\031\032Gq\242\257\241\346\341cxA\235\246\"\267\306\321\337\323a%v\216M9\362\263;4B\341@\3145\350\200\340\233\030\300\272\255>l\223\235B\233\n\262\021\220K\301\010\327-!\\\016\203kt\236$\033\333\265\250%\245\314Jc\345\2420\033\316\356\260bDx\250\326\2072\330bh\232\253\311`ak(\377|\007t\320u\004&\275}\\\245\014B;\214L\267$\330\343>\371E\311\365T\030a\216C.&]-\344_#\010\226r\350V\242Sm\331\t\216W 4&\273v\232\214\273\365i\364\276k5t\304Pn\334\"\247\022\3310\272aN\327""\242Tj\232 5X>\356\366\3733\300\356\341\027\004\035\363i'\334\343a\205{>\330\211lj\267J\246\2326\312\362\263\271\352\242\225\001\205tTCV\336\221Y,\367L+u?\232. ;^Kn\316\2436\032\307\317\221\022=\271\306\372\273\357K\2514|$\272\307P\365\212\"T\244IPl\241 \022\022N\013\312sx\240\212\231\331\253\317a\026\271\372eW\361k;\360\245\275\226\215Jn\274\007\217Ow<\020t!\25267\246\305\362\263\352\032\307\220c\325\342\343\332\314\303\201\231\216\224\373Y\235k\241\215\371V\203\305\344\353\356\275v\376\177E\032k\3338\036\315\317\323\r\352RE\271TG\323U\300\017r,\324\274\351\276\330\274Py\367\334\353\301\276\313\005\323\366\326*\270\2071\327]N\277+\244.S\014\3625h\320\224\240a\361\305\034\366pT#\351&8\330\320\324\266\262\223)\022\240G\276\311\204|j\027\223\237\r\217^,\201wx\006\205\221\255\3758\313\025\033&%Z\260\261\327c\234\000@\200\216\215N\270\224\221,\334\302f\372\221\225\277\366\251\027\236\343z\262O]\202\346\315\265\350\346HmZH\230\360%\334>W\344\307\252\352}7\356pWy\203\225i\365\323\010oX-y\254gber\302\362\032)\r\213\233H\271I\030\362\214\222\013\020n\326\374G\320\354\026;\355v\010\244\014\340\242\022\272\032\3507U\241\257\202\213\035\274U0\341%\240\215SiDp\265\225U\311^\257d\013\320\300\311\311\323[\272\033|Hn\324Kl\354W\321s!\264\243\021\r\305\216\345\225Y4\305\t\001\347\236\252\026y*\231\204\r\006\325\360\274\035\315b\327\314\343\010\272B\330\032$SI\0341}`\324\3345\327\\\26452C\3451K@7\214\t\2141J\326\212,\326\316\262\254\237\204\031\006\203\262\017M\004z3\237VA\371\243\177%\322\\\243J\"\210\303G\266\367e%\253\t\201<\372\211\027i\010\314\356\254\214\363\260o=\034q\256[O\224\220\313\366\2513\271T`B\315\222\017*F\220\313\257N-\306\247\204\2250V\3616\263|;u\360\276;\255\321`\356\342\343,\304\306\t\362\342\215\"\r\260m\004L\235\327B)\033\263\033\260lll^;e*[\21611<\264\331\276TB)\261\2446\206\323]\010\305^\ru\267]\271\022A`\215\221B\303\274\275\331dUH@,\313\033?\203>\020|A\037Dv\363)\275\306\324\215\330\032K\341\245\310\014^\303\204\020Zb<\214\021\0323\033\034\337\2213\246EB""\352\321I!\224\244\002\226\022\321h\227,)m\031\264r\t\262A\005\314BaC\000]\330\226\325B6\010\\\247\016B\024\261\202\351\224\213\005\"\\\205\245SM\264{}<0\306\024\271iM\214\336\363\266K\254\332=m\0168\n\350\200\020\3719\314\001\3448a2!@RQ\010\022\246q\316\374s\314\347\")\323R\0141\026\302q\271\242\310Udh\344!\272c\264\020b!\220\004\320\030j \356\244\\\226e\022\270\023\022Dh\337\200\347\226|\334k\317\222\253vc\020-1\0135\305\341y\244\033\204\020\303\227\244Ge\336\326\273\010c\031h ~JV\344\350\037a\370\300e1\324)jH\322u\342\341L@\320\301g\327\230\371\242\272v\2753EKugS\204p\215\375\005\0145\212\004R\003\004\002\311\246\312\341\331 \034\355p\312H=\302\361\020J\327I\244\207\341\3351\2756\233\037\226\273>_\247q\366k\035\2107f\031{5\213(2\376\031\003$\342\345\233\342\004m3\203\026\224Z\320N\010K\025\026\242\270A+O=v:\\<\243g\353\276\310\330~\002\350@\353\322\267 \230ss\267\004\312*\340\322\250k\271<\241\020\251\232\254\360xr5p\tL\026\010\n\010\241\373\211\n\006\204+0H\303\022H\275\240\233H@_\266\321$\226\274n2\202\343\020\026\024 \201AQ\2530)\001Lw)\374a\004\025r!\375\216\220&z\226\355\272<\t0\351^\354\306J)y\213m\254\033*x\304()E\032n\252\313\264@\300T\nRK\337\327\224\240\004\023\\\230\2066\255\344/\206\362\255[\332\276\266|6\341\"Pa\355\252\247\021\025(0\275\346\270\226\224\273/]\222\330\347r\033\235\016Kqs\014)\005\005r9\rA\241niaJ\310`\203\005\000\205[2\354\246[\327\225x\216\263\021\351t\031!\tDN0U\030f\353\246\355\300|l-\220V\rm\224\226m\303I\201}\214M\335\244\261\303D\3045\306\363- w7\340\004\032\024\037\027]\360\005\002\022I\023\007\227\304\326\301\335\276is<\350\356dc\037d#8\341\351c\027\356\320\300\253\232^\365\016\304\0301P\346\340\020[\304\274[\010\251\247;\367\366@\304W\323\264w{\231\352\375B\255\224\020\254Y\331\203\002r\332\267\225\t%\205\307\232v\313\226\233\177)\256\\\265^\034ZNH\033\333\236\320X\353WI\262\357\252%\365\365\314\350\232\325\342Oo>%\217\\c\212\205\200\033C\213\243\256\227lQ\275tL\231\023#X\321y\240k\247\003I\004*^\263f\031\314\335\262S\245""\030\026(\020\235\341\243'\034\352\231\330\252\306\331\205v\346\326\340\273H\214\021\263'g\034\206\367U\344\316\365l\345$\330\215\323IF\223kW\202\371\332#v\027\327:1p\222\320\333\0178\252\250\372\316q\232o\352U\352\273nzwE\215\005\005\0333\030\002\234b\237\r\204A\231{1\356\274\216O.\2422@p\354\333\336r\365\244\331Q\273m\236\255r\352:\276\351)\243<\274q\244n\251\211\267\255\365%\030X\305\221\341\376\nSN\330^\375\350;\313\236'\205\361\020\326\222\242\007W\025\270\301\204\210^3\243\302fbW\0171\260!\244\3059 pS\247\025(&\222\211\363\246\267\355n\003\264\260tdc\331q/\211\313\246\330y\355::\316\326\022K;n\272H\316\246\372\265\210\354\205\276\200m\020\365\214]\345\266\376\213\3046AK\030'\202\307\275\224o\014\244p\035\363\366P\255.\230A\030 P\223\361\250\034tu\204\312c\360\364$D\002\240t\255l\034\253\033eT\211\034M)\034\372\022\001\"\240\372<\233g\030\327](\372\372Y\007\014\304\233\263\376\220\251\247\331\332v{\034\300\032\215\002\322l\245;\353\257\032\304`\252\351\230\327\013\230k\327\n\220f\000\334\302\214\327\037\250\231H5V\372\351\312\223\224\031\360`\027F\241\300p$\317\375\245\\\374\257D \003\270Z\220{Z\320n\317<]\215\007\216t\274a\344\271\365\354\303bB\036\330\3737\266\026\317\237\032E\3536\261AvS\203\2159&\375z\335i\206\253\021\343\212\334\371q1A'\227\220\251\246\036\330\306\255>\225\021\004\025\032\351\260B\210.?6\261\267\030\277\236i=\310\3069\332\024\235I\373\235P\341\001\265\227$\306 \231\207\r\302\353n\375\036g\223\225\263\003\223\307@k\255qo\004H\327\177\343\320\203\277\241\365\376f\266\302\304|_\345\267\315\202\243\004Rc\027\270\251\236\322\303\362\361V\311gx\002\262~\302\312\347\333\007\236\301\203\330\355\333b\220s\335\0309_g,\030i\003x:_/,\026\370\253Y\247\nV\377]\275\376DD\r\315\200\371\263\335\337\256d\"D\311\010\313\303\360\362V\202\3074U\272VP\212 \370\375\233N\365\217Rj\036\003\323cD|=\261\315\0046\317,,\356\013>\343Z\215(7\367\366\003\262\275\241\217\005\305P\005\030\236\373\271,\372BkT\205\345\222\220\276\0278\356c\205\343\331s\332I-v=\324\347\006""\232\210\210\201@\211\030\004\020\022\227qw$S\205\t\010\013\360\317P";
    PyObject *data = __Pyx_DecompressString(cstring, 2826, 2);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) != 0 /* compression: zlib (2811 bytes) */
const char* const cstring = "x\332\215Y[s\333\266\022\376G\235\306'\315\364<\235\241\010H\342\230\004)\000\224\354\274`d\211I4\225dG\222\343\346\337\237]\334\001\322i\037\022a\277\005\201\305b\257\360o\377{\371\371c\270<]\177;\036\236\016Oz\250\206\363\353\351Z\315\326\252(K*\204\232\327\275X\252E\335\316\212z\004w\224\213JH\312d\304Z\366\013*\353Y\204\324mY\324j\303+I#\264\331\250Y\305H\204\264L\021\332\024\t\306i]<P\242ZN(\257\330\"a5\255\244\252\220mS\225c\234\323\202\214\321\\\212\317\224\267jV\010\330B\243d]\t\252\032\256G%U\035\247s*\313\345?pY\253\346E_\313\177\230\246w\317\346\314\353ba\365\251\0315\252K\266\\\365\300'T/\253\031\372\234\372*\"rY\026\021\305Z\246\017W\266=\223\224\253\331\243\244\"\006\272\242\274\247\322@+U\261J\302\247\222\253\246\020\367Z\222wX\235Ve\t:\005\371\201\2559\325\202\265\234\252vM9\357\331\304\004\001\027VS%\227x\027T/A\250\326H\321\313\026\204\221K\325T\213\010\237\025Du\367\364Q\225L\362\014_M\340\345\262`\013P\360\362Qu-\2271\247\347\\\255:%$J\324\264\361\356\372p\362\261\243\021\006vW,\300\n\346u\273\201\217\250\2676\307\247\215j\026R\321\007\260wQ\265Ld\314\r\030s\273\231\004\365^\352\256\370\025s\0261\031\\\244\254\346\217\372\"\002\334\225\225\265 \312@M\005!\251\210\250\001\260kY\255a\306\332z\245\345\361b\243\252N\225\242o2\264A\373R\031(\300\004\321^\346e|L^N\255Q*\316\270b\305\275ZP\0263\300\364>\243b\037P\352\200\213e/\341\334,\2771\301W\366\233\030|\024\252j\340f\324\242\257\342;\354\211*\326\346\310\224\315[^\322\2249\026\364\201\353(A\232b\231\233w\331\365\312\256>\301\355\226\357qd\267\3247\351lIk]\225u\205?\234r\272\300\360\310#V\3334\212\n\031!+\005\356\023h+\354\274\220&\340\032tQ\021k\353\001\253'\260\330\245R$\335\305\370X\376q\260\237\014L\276\005\247\262\3013\2073\241\001\251\013!\325f\245\243q\2714!\30039\315\316.\232\221Lb\245\010/*\026\177\212\206\222~\007@]5\225\034o\263Y\005\251\264k\353\313\203 \033\207\326\300\320!P\221\026\276\224\274\350\246\230t\001F*RNSN/\007\361\261\311\367\026\254\232\317\215M\230X\323\321Rk\035\322\237""\016\323\223\034\302\333n\222\001\372\"5\235d\311b\221\342TdkP\271L\0018[\nT\214\345\262V\335\372\343\030\301\3008B?\245H\323\325\"Ed\231I\324\223\014X?\324\005\034\245\0079\254\"\323\202\304\320\r\017\343\264*1\030wV\005\221\342^\221J\024\263\332\230\211\006(Ki\334\215\332\214\205\256\247Ct5KH\336\202\233\256?L`w\t\006\021l\0161l6\371\201aF\237\241\247\026\2741[#\001\3778d0O\203\205s\035\325\032\331\253\017\277\337}t\343\273\337?\376\351\307\177|r\303\217\277\377\327\217\377\370`\366\2609\347CLh\016k\td\316\302\01793\225\225!\332\336\3062M\212M\005%\215'{v\317 \250G4^_5\257\214\350\006\023\361z\232r\027\336\022H\325}\347RBDF\005\336\030\r\365\\\314\323w\236\203\025X\037,3Z\336\325\211)V\256sLP6\232gR\326xn\220\252\023P\005-Xa\255\013i\331\300\2653}\240\256\340:\346\267\r\204\270(\271\370\"P\374jV\007\361\037\"\225u\274,z\307\305\000\304&sq\006t\346e(\254\304\320\310Px\217\316\332VB\271\323\310\004-\213N\347\275\273IT\027-\230\262c\236K\205\213\024n\022\322\325D\371~\304\032\224!\370\324\006\272\034\003O\305`A\211Hx\025\324\0058!\003\253\240.(\022L\355\017\016)<Zc3\200\337\306\337\325\025\273\207t\006\273\225\217c\206\336]-\t\177\217\305\246X\233\212@r\276{H8\r\246f\266\346E\343!\326v\321x\244\203\266\303\223\220\004\353\306\253hHl\264F4\200\031.\371\010\202?\032W\014\t*\215\267\022*\312\224S\303\255\247\337\213\260\231`Mg\243v:EG\r;\326\245\271\304\320\233\316Ij\276\2045\022\031\222<\201^\251\204B#\275\374u\305\375IV\235\257Up\350L\000\307\220\313i $\017C\341\206bE\302\220\272a\024\357V\035Z'8\032w\024\226\320\246\325\362H\351F}\030\0217z\300\"\332\306\021G\273x\023\312-\337\237!T\233\322\316\221\366StF72\275\330\014\"\t\030!\255\347P\245`5fZ\3630\241\\c\202\025p\364\256\263\375D`\276\333r\204)Y\253\0201\332\036\252r\000\230\303{n\356\334\322`R2bS\206\252V\005\006G\333\0009\016t\020\025x\212\215\371%\230\026T=`K\335\364\004\323l\343\024\024xr\212\217\320\021\013\237# \037f\250O&\021f^4*\266\316p\016M\202O(9>\265e`\030\272j\232lFd\0061\364\253\351\2267\222N\212\326""\"!\207\330\273\212M+0Q\037\332=G\034l\352`9\265\304\306\236\217\330\035\031A\374\001JU\257\310\200k\2011-N\213\000\236\340\326\302=\265\311p\022\325\004\226\201\355D\016C\262\304\316TV\215\027\321wC\261\035\032P\366\216\3048\tG\247\017\016\260E\t\014\361\r\302\016\261*3]\207\003\250\344\372\201\302\321\260\267\306\034\275R\235p\306.\022\"\022\006\245\205BKS=\205\005\321Q\240\203\001\211\314#\030\270>\024\024\220R>\334\375\031\036x\336\231\013\373\367\234aT\370\325\254\315\262\205\020\334\010\355\331!v\351\314>\021\036\262\031\204\326\305\243oP2f\324\206g\234,n\240\327\270g*a\255\010\237\315\300D}K\207eDc[o\360l\236c\330\242\347\330\234\214\021\325\316\3476\354\307\214\252]\027\006\263\250}]2\341w\212\323\221)\024\222\024\023P*@\017\226\260!\005A\354\"S\230\016V\2445\335\007\243\233\321\034\335\222\376\202kxm=^\275b\2351'\353\201\212\010\351{\270\014\373\224c&\351\232&m\204\333\322\335\341\363\036\342\266l)\224\331\362s\314\020\274\034mh\261O96\265\241\307\355\206:d\314\241\020\243\236\252\030\024T\021\031\254N\323q\tn\200V\367\002\026\341+\211u\240\211\034\232\222\215\037\332g#,\365\243\260\271\032c\035\031c~\235\030\263\217E)\352\002\235\177\313pDx?\223\315RQ(\213<1\257\230\033\262\326\265\373HqF\326f\254]\315\234 ~\254R]3\202\326\006B\3435\305\361,\2437\005\357R(*\202\"\320uy\031\344\256o]\324=\025\306\3450&\224X\241 cS\216\2623@\370\364\013UZ\347*8\204B\242\006\252l\241\036\020\033S\370\000mJ\261\273\204\372\217\245|\252\177\304\340m\236+#\030\227\252\251y<\2018\014q\271\351\376\305\024\265\201\206->\206\237\214\3013\303\010\276\333\351\2243\253L\221\0311\365\273\207\275\310\010\266y>B \206\263>\007E=\006H\016\2013\255\362C\201\235`\303cQ\375`\026\251\333\226S\304-\345+\"\030/(\243<\231\275\340\216\007\025\210\242\264d\021\323B\266\364\217p\343\260\252u\032L*,Cg\257\216\026\304\325\332.\003\341f3\244\343\255\314 P\341\350CN\010t\016m|\036\373w\252\010I\252<G\307&\353Jy;\264\023\322\213\344\340v\305,z\\uX~H\004Qm\234\304\017\2461c\204&\307\322\276c\253\212\004""\306j\205>\2241\346*\226\014w5(\016{\237\014\201\002\273\tf\001\204.+\300\277\253\030\243u \302\232@4\205}@2d\010a\206\216\264\210\024\20448\256u?\032\030\356-%P\250\021\207\010g\324\231\372\343\022\031In\037\n\235\200+\021\0179y\364C\3270\"\021\305\277\r\244\nn\266\265\177\375\322\177\201\362\245\235C}\231\353\200x\206ya\236(\271</-\266<\374n\237\346gd\345\326&\316=A&\376N\217\025\030\243\336*\260\202\013\360$D\363(b\360\314\263y\346M<\367&\376^\217\264\341\301.\371\270'J1\277\033XX$8R\316>\r\345\255\211;\333\301\304\034\353\212L\200\255\327`\306@k\341k(3\3167z~=\301\317\374\270\375\272;\036\316\203:\234\325\355\262\335\rO\333\335_\370g\367\257\303\355\360\364Cmw\273\341zU_`\336U\323\267\347\323a\247v\333\027\244v\317\257\347\333pQ\373\341\272\273\034^n\207\347\263\206\277\303r\207\033L\276]\324i{\375K\203\227a{\033\220\247a\277b\300\337v\001\335\017?\016\273\001\367\211\260\323\366\333\304\302\303\217\341|S\267\237/\003R_\216\317o\206\237 ~\021M]_\206]\340?_\376R\327\333\366\366\252'|=\354=\313\215\325\365\347\365\213\346\236\016_\365\\\315>]&\3049\335^\365\317\233_\345\374\274\037<\361\274\177Q_\207\363p\331\036\361tW\207\201\362\317\327\227\347\313M\237\371\351p3\234\227H\367H\274]\340g\007\013\"\360\262\275\340\321\367\317\247-\\\337X\224\227#\\\350)\326\216\337\300/\232\"w\036\362\207\374\376\222.\n\264\2752\277F\200\256\303y\237\n\r\274\261`\000\306\353;\351\276\277\016\227\237\010\354\267\267-Z\344\363e\017\326\345\324\364\016\333ou\331\276\201J\300|o\376\023\220j\357\2544\022\3522\\\206\257x}\303\005\226\267\332\364`\230\366\267\372\266\275~S_\016\303q\237\"\257\347\035Z{\230{\035\216\003@?\016\267\237\352\010&y4 \250#L\271|O\325\200\300X9\210:\215\334N\351m\335N\337@\275z\364\362M\235\206S\230\351\355\307!?\266\307\327\341\352\227\215\275\013\306\301\212\200\010\326\377\226\211\010t\370jJZ\000\375U\276\005\301\023;\375\373\262\333g\237*\245mV\341\031\236\340\022\317\333\023\310\352H\030\235\236\367\257\307!\374*=\005\177\277\277n\217f|\035n\373\341\313""\366\365xS\3526\\\341\377\377\003}\344\303\236";
    PyObject *data = __Pyx_DecompressString(cstring, 2811, 1);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #else /* compression: none (9286 bytes) */
const char* const bytes = ".?pyverbs.libibverbs_enumsIBV_ACCESS_FLUSH_GLOBALIBV_ACCESS_FLUSH_PERSISTENTIBV_ACCESS_HUGETLBIBV_ACCESS_LOCAL_WRITEIBV_ACCESS_MW_BINDIBV_ACCESS_ON_DEMANDIBV_ACCESS_RELAXED_ORDERINGIBV_ACCESS_REMOTE_ATOMICIBV_ACCESS_REMOTE_READIBV_ACCESS_REMOTE_WRITEIBV_ACCESS_ZERO_BASED_IBV_ADVISE_MR_ADVICE_PREFETCH_IBV_ADVISE_MR_ADVICE_PREFETCH_NO_FAULT_IBV_ADVISE_MR_ADVICE_PREFETCH_WRITE_IBV_ADVISE_MR_FLAG_FLUSH_IBV_ALLOCATOR_USE_DEFAULTIBV_ATOMIC_GLOBIBV_ATOMIC_HCAIBV_ATOMIC_NONEIBV_COUNTER_BYTESIBV_COUNTER_PACKETSIBV_CQ_INIT_ATTR_MASK_FLAGSIBV_CQ_INIT_ATTR_MASK_PDIBV_CREATE_CQ_ATTR_IGNORE_OVERRUNIBV_CREATE_CQ_ATTR_SINGLE_THREADEDIBV_DEVICE_AUTO_PATH_MIGIBV_DEVICE_BAD_PKEY_CNTRIBV_DEVICE_BAD_QKEY_CNTRIBV_DEVICE_CHANGE_PHY_PORTIBV_DEVICE_CURR_QP_STATE_MODIBV_DEVICE_INIT_TYPEIBV_DEVICE_MANAGED_FLOW_STEERINGIBV_DEVICE_MEM_MGT_EXTENSIONSIBV_DEVICE_MEM_WINDOWIBV_DEVICE_MEM_WINDOW_TYPE_2AIBV_DEVICE_MEM_WINDOW_TYPE_2BIBV_DEVICE_N_NOTIFY_CQ_IBV_DEVICE_PCI_WRITE_END_PADDINGIBV_DEVICE_PORT_ACTIVE_EVENTIBV_DEVICE_RAW_IP_CSUMIBV_DEVICE_RAW_MULTI_IBV_DEVICE_RAW_SCATTER_FCSIBV_DEVICE_RC_IP_CSUMIBV_DEVICE_RC_RNR_NAK_GENIBV_DEVICE_RESIZE_MAX_WRIBV_DEVICE_SHUTDOWN_PORTIBV_DEVICE_SRQ_RESIZEIBV_DEVICE_SYS_IMAGE_GUIDIBV_DEVICE_UD_AV_PORT_ENFORCEIBV_DEVICE_UD_IP_CSUMIBV_DEVICE_XRCIBV_DMAH_INIT_ATTR_MASK_CPU_IDIBV_DMAH_INIT_ATTR_MASK_PHIBV_DMAH_INIT_ATTR_MASK_TPH_MEM_TYPEIBV_EVENT_CLIENT_REREGISTERIBV_EVENT_COMM_ESTIBV_EVENT_CQ_ERRIBV_EVENT_DEVICE_FATALIBV_EVENT_GID_CHANGEIBV_EVENT_LID_CHANGEIBV_EVENT_PATH_MIGIBV_EVENT_PATH_MIG_ERRIBV_EVENT_PKEY_CHANGEIBV_EVENT_PORT_ACTIVEIBV_EVENT_PORT_ERRIBV_EVENT_QP_ACCESS_ERRIBV_EVENT_QP_FATALIBV_EVENT_QP_LAST_WQE_REACHEDIBV_EVENT_QP_REQ_ERRIBV_EVENT_SM_CHANGEIBV_EVENT_SQ_DRAINEDIBV_EVENT_SRQ_ERRIBV_EVENT_SRQ_LIMIT_REACHEDIBV_EVENT_WQ_FATALIBV_FLOW_ATTR_ALL_DEFAULTIBV_FLOW_ATTR_FLAGS_DONT_TRAPIBV_FLOW_ATTR_FLAGS_EGRESSIBV_FLOW_ATTR_MC_DEFAULTIBV_FLOW_ATTR_NORMALIBV_FLOW_ATTR_SNIFFERIBV_FLOW_SPEC_ACTION_COUNTIBV_FLOW_SPEC_ACTION_DROPIBV_FLOW_SPEC_ACTION_HANDLEIBV_""FLOW_SPEC_ACTION_TAGIBV_FLOW_SPEC_ESPIBV_FLOW_SPEC_ETHIBV_FLOW_SPEC_GREIBV_FLOW_SPEC_INNERIBV_FLOW_SPEC_IPV4IBV_FLOW_SPEC_IPV4_EXTIBV_FLOW_SPEC_IPV6IBV_FLOW_SPEC_MPLSIBV_FLOW_SPEC_TCPIBV_FLOW_SPEC_UDPIBV_FLOW_SPEC_VXLAN_TUNNELIBV_FLUSH_GLOBALIBV_FLUSH_MRIBV_FLUSH_PERSISTENTIBV_FLUSH_RANGEIBV_FORK_DISABLEDIBV_FORK_ENABLEDIBV_FORK_UNNEEDEDIBV_GID_TYPE_IBIBV_GID_TYPE_ROCE_V1IBV_GID_TYPE_ROCE_V2IBV_GID_TYPE_SYSFS_IB_ROCE_V1IBV_GID_TYPE_SYSFS_ROCE_V2IBV_MIG_ARMEDIBV_MIG_MIGRATEDIBV_MIG_REARMIBV_MTU_1024IBV_MTU_2048IBV_MTU_256IBV_MTU_4096IBV_MTU_512IBV_MW_TYPE_1IBV_MW_TYPE_2IBV_NODE_CAIBV_NODE_RNICIBV_NODE_ROUTERIBV_NODE_SWITCHIBV_NODE_UNKNOWNIBV_NODE_UNSPECIFIEDIBV_NODE_USNICIBV_NODE_USNIC_UDPIBV_ODP_SUPPORTIBV_ODP_SUPPORT_ATOMICIBV_ODP_SUPPORT_ATOMIC_WRITEIBV_ODP_SUPPORT_FLUSHIBV_ODP_SUPPORT_IMPLICITIBV_ODP_SUPPORT_READIBV_ODP_SUPPORT_RECVIBV_ODP_SUPPORT_SENDIBV_ODP_SUPPORT_SRQ_RECVIBV_ODP_SUPPORT_WRITEIBV_OPS_SIGNALEDIBV_OPS_TM_SYNCIBV_PARENT_DOMAIN_INIT_ATTR_ALLOCATORSIBV_PARENT_DOMAIN_INIT_ATTR_PD_CONTEXTIBV_PORT_ACTIVEIBV_PORT_ACTIVE_DEFERIBV_PORT_ARMEDIBV_PORT_AUTO_MIGR_SUPIBV_PORT_BOOT_MGMT_SUPIBV_PORT_CAP_MASK2_SUPIBV_PORT_CAP_MASK_NOTICE_SUPIBV_PORT_CLIENT_REG_SUPIBV_PORT_CM_SUPIBV_PORT_DEVICE_MGMT_SUPIBV_PORT_DOWNIBV_PORT_DR_NOTICE_SUPIBV_PORT_EXTENDED_SPEEDS_SUPIBV_PORT_INFO_EXT_SUPIBV_PORT_INITIBV_PORT_IP_BASED_GIDSIBV_PORT_LED_INFO_SUPIBV_PORT_LINK_LATENCY_SUPIBV_PORT_LINK_SPEED_HDR_SUPIBV_PORT_LINK_SPEED_NDR_SUPIBV_PORT_LINK_WIDTH_2X_SUPIBV_PORT_MKEY_NVRAMIBV_PORT_NOPIBV_PORT_NOTICE_SUPIBV_PORT_OPT_IPD_SUPIBV_PORT_PKEY_NVRAMIBV_PORT_PKEY_SW_EXT_PORT_TRAP_SUPIBV_PORT_REINIT_SUPIBV_PORT_SET_NODE_DESC_SUPIBV_PORT_SL_MAP_SUPIBV_PORT_SMIBV_PORT_SNMP_TUNNEL_SUPIBV_PORT_SWITCH_PORT_STATE_TABLE_SUPIBV_PORT_SYS_IMAGE_GUID_SUPIBV_PORT_TRAP_SUPIBV_PORT_VENDOR_CLASS_SUPIBV_PORT_VIRT_SUPIBV_QPS_ERRIBV_QPS_INITIBV_QPS_RESETIBV_QPS_RTRIBV_QPS_RTSIBV_QPS_SQDIBV_QPS_SQEIBV_QPS_UNKNOWNIBV_QPT_DRIVERIBV_QPT_RAW_PACKETIBV_QPT_RCIBV_QPT_UCIBV_QPT_UDIBV_QPT_XRC_RECVIBV_QPT_XRC_S""ENDIBV_QP_ACCESS_FLAGSIBV_QP_ALT_PATHIBV_QP_AVIBV_QP_CAPIBV_QP_CREATE_BLOCK_SELF_MCAST_LBIBV_QP_CREATE_CVLAN_STRIPPINGIBV_QP_CREATE_PCI_WRITE_END_PADDINGIBV_QP_CREATE_SCATTER_FCSIBV_QP_CREATE_SOURCE_QPNIBV_QP_CUR_STATEIBV_QP_DEST_QPNIBV_QP_EN_SQD_ASYNC_NOTIFYIBV_QP_EX_WITH_ATOMIC_CMP_AND_SWPIBV_QP_EX_WITH_ATOMIC_FETCH_AND_ADDIBV_QP_EX_WITH_ATOMIC_WRITEIBV_QP_EX_WITH_BIND_MWIBV_QP_EX_WITH_FLUSHIBV_QP_EX_WITH_LOCAL_INVIBV_QP_EX_WITH_RDMA_READIBV_QP_EX_WITH_RDMA_WRITEIBV_QP_EX_WITH_RDMA_WRITE_WITH_IMMIBV_QP_EX_WITH_SENDIBV_QP_EX_WITH_SEND_WITH_IMMIBV_QP_EX_WITH_SEND_WITH_INVIBV_QP_EX_WITH_TSOIBV_QP_INIT_ATTR_CREATE_FLAGSIBV_QP_INIT_ATTR_IND_TABLEIBV_QP_INIT_ATTR_MAX_TSO_HEADERIBV_QP_INIT_ATTR_PDIBV_QP_INIT_ATTR_RX_HASHIBV_QP_INIT_ATTR_SEND_OPS_FLAGSIBV_QP_INIT_ATTR_XRCDIBV_QP_MAX_DEST_RD_ATOMICIBV_QP_MAX_QP_RD_ATOMICIBV_QP_MIN_RNR_TIMERIBV_QP_PATH_MIG_STATEIBV_QP_PATH_MTUIBV_QP_PKEY_INDEXIBV_QP_PORTIBV_QP_QKEYIBV_QP_RATE_LIMITIBV_QP_RETRY_CNTIBV_QP_RNR_RETRYIBV_QP_RQ_PSNIBV_QP_SQ_PSNIBV_QP_STATEIBV_QP_TIMEOUTIBV_QUERY_QP_DATA_IN_ORDER_ALIGNED_128_BYTESIBV_QUERY_QP_DATA_IN_ORDER_RETURN_CAPSIBV_QUERY_QP_DATA_IN_ORDER_WHOLE_MSGIBV_RAW_PACKET_CAP_CVLAN_STRIPPINGIBV_RAW_PACKET_CAP_DELAY_DROPIBV_RAW_PACKET_CAP_IP_CSUMIBV_RAW_PACKET_CAP_SCATTER_FCSIBV_READ_COUNTERS_ATTR_PREFER_CACHEDIBV_REG_MR_MASK_ADDRIBV_REG_MR_MASK_DMAHIBV_REG_MR_MASK_FDIBV_REG_MR_MASK_FD_OFFSETIBV_REG_MR_MASK_IOVAIBV_REREG_MR_CHANGE_ACCESSIBV_REREG_MR_CHANGE_PDIBV_REREG_MR_CHANGE_TRANSLATIONIBV_REREG_MR_ERR_CMDIBV_REREG_MR_ERR_CMD_AND_DO_FORK_NEWIBV_REREG_MR_ERR_DONT_FORK_NEWIBV_REREG_MR_ERR_DO_FORK_OLDIBV_REREG_MR_ERR_INPUTIBV_RX_HASH_DST_IPV4IBV_RX_HASH_DST_IPV6IBV_RX_HASH_DST_PORT_TCPIBV_RX_HASH_DST_PORT_UDPIBV_RX_HASH_FUNC_TOEPLITZIBV_RX_HASH_SRC_IPV4IBV_RX_HASH_SRC_IPV6IBV_RX_HASH_SRC_PORT_TCPIBV_RX_HASH_SRC_PORT_UDPIBV_SEND_FENCEIBV_SEND_INLINEIBV_SEND_IP_CSUMIBV_SEND_SIGNALEDIBV_SEND_SOLICITEDIBV_SRQT_BASICIBV_SRQT_TMIBV_SRQT_XRCIBV_SRQ_INIT_ATTR_CQIBV_SRQ_INIT_ATTR_PDIBV_SRQ_INIT_ATTR_TMIBV_SRQ_IN""IT_ATTR_TYPEIBV_SRQ_INIT_ATTR_XRCDIBV_SRQ_LIMITIBV_SRQ_MAX_WRIBV_TMH_EAGERIBV_TMH_FINIBV_TMH_NO_TAGIBV_TMH_RNDVIBV_TM_CAP_RCIBV_TPH_MEM_TYPE_PMIBV_TPH_MEM_TYPE_VMIBV_TRANSPORT_IBIBV_TRANSPORT_IWARPIBV_TRANSPORT_UNKNOWNIBV_TRANSPORT_USNICIBV_TRANSPORT_USNIC_UDPIBV_VALUES_MASK_RAW_CLOCKIBV_WC_ATOMIC_WRITEIBV_WC_BAD_RESP_ERRIBV_WC_BIND_MWIBV_WC_COMP_SWAPIBV_WC_DRIVER2IBV_WC_DRIVER3IBV_WC_EX_WITH_BYTE_LENIBV_WC_EX_WITH_COMPLETION_TIMESTAMPIBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCKIBV_WC_EX_WITH_CVLANIBV_WC_EX_WITH_DLID_PATH_BITSIBV_WC_EX_WITH_FLOW_TAGIBV_WC_EX_WITH_IMMIBV_WC_EX_WITH_QP_NUMIBV_WC_EX_WITH_SLIBV_WC_EX_WITH_SLIDIBV_WC_EX_WITH_SRC_QPIBV_WC_EX_WITH_TM_INFOIBV_WC_FATAL_ERRIBV_WC_FETCH_ADDIBV_WC_FLUSHIBV_WC_GENERAL_ERRIBV_WC_GRHIBV_WC_INV_EECN_ERRIBV_WC_INV_EEC_STATE_ERRIBV_WC_IP_CSUM_OKIBV_WC_LOCAL_INVIBV_WC_LOC_ACCESS_ERRIBV_WC_LOC_EEC_OP_ERRIBV_WC_LOC_LEN_ERRIBV_WC_LOC_PROT_ERRIBV_WC_LOC_QP_OP_ERRIBV_WC_LOC_RDD_VIOL_ERRIBV_WC_MW_BIND_ERRIBV_WC_RDMA_READIBV_WC_RDMA_WRITEIBV_WC_RECVIBV_WC_RECV_RDMA_WITH_IMMIBV_WC_REM_ABORT_ERRIBV_WC_REM_ACCESS_ERRIBV_WC_REM_INV_RD_REQ_ERRIBV_WC_REM_INV_REQ_ERRIBV_WC_REM_OP_ERRIBV_WC_RESP_TIMEOUT_ERRIBV_WC_RETRY_EXC_ERRIBV_WC_RNR_RETRY_EXC_ERRIBV_WC_SENDIBV_WC_SUCCESSIBV_WC_TM_ADDIBV_WC_TM_DATA_VALIDIBV_WC_TM_DELIBV_WC_TM_ERRIBV_WC_TM_MATCHIBV_WC_TM_NO_TAGIBV_WC_TM_RECVIBV_WC_TM_RNDV_INCOMPLETEIBV_WC_TM_SYNCIBV_WC_TM_SYNC_REQIBV_WC_TSOIBV_WC_WITH_IMMIBV_WC_WITH_INVIBV_WC_WR_FLUSH_ERRIBV_WQS_ERRIBV_WQS_RDYIBV_WQS_RESETIBV_WQS_UNKNOWNIBV_WQT_RQIBV_WQ_ATTR_CURR_STATEIBV_WQ_ATTR_FLAGSIBV_WQ_ATTR_STATEIBV_WQ_FLAGS_CVLAN_STRIPPINGIBV_WQ_FLAGS_DELAY_DROPIBV_WQ_FLAGS_PCI_WRITE_END_PADDINGIBV_WQ_FLAGS_SCATTER_FCSIBV_WQ_INIT_ATTR_FLAGSIBV_WR_ATOMIC_CMP_AND_SWPIBV_WR_ATOMIC_FETCH_AND_ADDIBV_WR_ATOMIC_WRITEIBV_WR_BIND_MWIBV_WR_FLUSHIBV_WR_LOCAL_INVIBV_WR_RDMA_READIBV_WR_RDMA_WRITEIBV_WR_RDMA_WRITE_WITH_IMMIBV_WR_SENDIBV_WR_SEND_WITH_IMMIBV_WR_SEND_WITH_INVIBV_WR_TAG_ADDIBV_WR_TAG_DELIBV_WR_TAG_SYNCIBV_WR_TSOIBV_XRCD_INIT_ATTR_FDIBV_XRCD""_INIT_ATTR_OFLAGSIBV_XRCD_INIT_ATTR_RESERVEDIntEnumIntFlagcline_in_tracebackenumgetibv_access_flagsibv_atomic_capibv_counter_descriptionibv_cq_init_attr_maskibv_create_cq_attr_flagsibv_create_cq_wc_flagsibv_device_cap_flagsibv_dmah_init_attr_maskibv_event_typeibv_flow_attr_typeibv_flow_flagsibv_flow_spec_typeibv_fork_statusibv_gid_typeibv_gid_type_sysfsibv_mig_stateibv_mr_init_attr_maskibv_mtuibv_mw_typeibv_node_typeibv_odp_general_capsibv_odp_transport_cap_bitsibv_ops_flagsibv_ops_wr_opcodeibv_parent_domain_init_attr_maskibv_placement_typeibv_port_cap_flagsibv_port_cap_flags2ibv_port_stateibv_qp_attr_maskibv_qp_create_flagsibv_qp_create_send_ops_flagsibv_qp_init_attr_maskibv_qp_stateibv_qp_typeibv_query_qp_data_in_order_capsibv_query_qp_data_in_order_flagsibv_raw_packet_capsibv_read_counters_flagsibv_rereg_mr_err_codeibv_rereg_mr_flagsibv_rx_hash_fieldsibv_rx_hash_function_flagsibv_selectivity_levelibv_send_flagsibv_srq_attr_maskibv_srq_init_attr_maskibv_srq_typeibv_tm_cap_flagsibv_tmh_opibv_tph_mem_typeibv_transport_typeibv_values_maskibv_wc_flagsibv_wc_opcodeibv_wc_statusibv_wq_attr_maskibv_wq_flagsibv_wq_init_attr_maskibv_wq_stateibv_wq_typeibv_wr_opcodeibv_xrcd_init_attr_mask__main___member_names___members__module__module____name____qualname__setdefault__test__";
    PyObject *data = NULL;
    CYTHON_UNUSED_VAR(__Pyx_DecompressString);
    #endif
    PyObject **stringtab = __pyx_mstate->__pyx_string_tab;
    Py_ssize_t pos = 0;
    for (int i = 0; i < 470; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyUnicode_DecodeUTF8(bytes + pos, bytes_length, NULL);
      if (likely(string) && i >= 3) PyUnicode_InternInPlace(&string);
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
      stringtab[i] = string;
      pos += bytes_length;
    }
    Py_XDECREF(data);
    for (Py_ssize_t i = 0; i < 470; i++) {
      if (unlikely(PyObject_Hash(stringtab[i]) == -1)) {
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
  }
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_codeobjects ### */

static int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  return 0;
}
/* #### Code section: init_globals ### */

static int __Pyx_InitGlobals(void) {
  /* PythonCompatibility.init */
  if (likely(__Pyx_init_co_variables() == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError (used by HasAttr) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* HasAttr (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!PyUnicode_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_PyObject_GetAttrStrNoError(o, n);
    if (!r) {
        return (unlikely(PyErr_Occurred())) ? -1 : 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}
#endif

/* PyObjectCall (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall (used by PyObjectCallOneArg) */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) != (0)) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
#if CYTHON_VECTORCALL && !CYTHON_COMPILING_IN_LIMITED_API
  #if PY_VERSION_HEX < 0x03090000
    #define __Pyx_PyVectorcall_Function(callable) _PyVectorcall_Function(callable)
  #elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE vectorcallfunc __Pyx_PyVectorcall_Function(PyObject *callable) {
    PyTypeObject *tp = Py_TYPE(callable);
    #if defined(__Pyx_CyFunction_USED)
    if (__Pyx_CyFunction_CheckExact(callable)) {
        return __Pyx_CyFunction_func_vectorcall(callable);
    }
    #endif
    if (!PyType_HasFeature(tp, Py_TPFLAGS_HAVE_VECTORCALL)) {
        return NULL;
    }
    assert(PyCallable_Check(callable));
    Py_ssize_t offset = tp->tp_vectorcall_offset;
    assert(offset > 0);
    vectorcallfunc ptr;
    memcpy(&ptr, (char *) callable + offset, sizeof(ptr));
    return ptr;
}
  #else
    #define __Pyx_PyVectorcall_Function(callable) PyVectorcall_Function(callable)
  #endif
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject *const *args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
          #if CYTHON_COMPILING_IN_LIMITED_API
            return PyObject_Vectorcall(func, args, _nargs, NULL);
          #else
            vectorcallfunc f = __Pyx_PyVectorcall_Function(func);
            if (f) {
                return f(func, args, _nargs, NULL);
            }
          #endif
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_mstate_global->__pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* PyObjectCallOneArg (used by ImportImpl) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* TupleAndListFromArray (used by ImportImpl) */
#if !CYTHON_COMPILING_IN_CPYTHON && CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    Py_ssize_t i;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    for (i = 0; i < n; i++) {
        if (unlikely(__Pyx_PyTuple_SET_ITEM(res, i, src[i]) < (0))) {
            Py_DECREF(res);
            return NULL;
        }
        Py_INCREF(src[i]);
    }
    return res;
}
#elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* ImportImpl (used by Import) */
static int __Pyx__Import_GetModule(PyObject *qualname, PyObject **module) {
    PyObject *imported_module = PyImport_GetModule(qualname);
    if (unlikely(!imported_module)) {
        *module = NULL;
        if (PyErr_Occurred()) {
            return -1;
        }
        return 0;
    }
    *module = imported_module;
    return 1;
}
static int __Pyx__Import_Lookup(PyObject *qualname, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject **module) {
    PyObject *imported_module;
    PyObject *top_level_package_name;
    Py_ssize_t i;
    int status, module_found;
    Py_ssize_t dot_index;
    module_found = __Pyx__Import_GetModule(qualname, &imported_module);
    if (unlikely(!module_found || module_found == -1)) {
        *module = NULL;
        return module_found;
    }
    if (imported_names) {
        for (i = 0; i < len_imported_names; i++) {
            PyObject *imported_name = imported_names[i];
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
            int has_imported_attribute = PyObject_HasAttr(imported_module, imported_name);
#else
            int has_imported_attribute = PyObject_HasAttrWithError(imported_module, imported_name);
            if (unlikely(has_imported_attribute == -1)) goto error;
#endif
            if (!has_imported_attribute) {
                goto not_found;
            }
        }
        *module = imported_module;
        return 1;
    }
    dot_index = PyUnicode_FindChar(qualname, '.', 0, PY_SSIZE_T_MAX, 1);
    if (dot_index == -1) {
        *module = imported_module;
        return 1;
    }
    if (unlikely(dot_index == -2)) goto error;
    top_level_package_name = PyUnicode_Substring(qualname, 0, dot_index);
    if (unlikely(!top_level_package_name)) goto error;
    Py_DECREF(imported_module);
    status = __Pyx__Import_GetModule(top_level_package_name, module);
    Py_DECREF(top_level_package_name);
    return status;
error:
    Py_DECREF(imported_module);
    *module = NULL;
    return -1;
not_found:
    Py_DECREF(imported_module);
    *module = NULL;
    return 0;
}
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *from_list = 0;
    int module_found;
    if (!qualname) {
        qualname = name;
    }
    module_found = __Pyx__Import_Lookup(qualname, imported_names, len_imported_names, &module);
    if (likely(module_found == 1)) {
        return module;
    } else if (unlikely(module_found == -1)) {
        return NULL;
    }
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    if (imported_names) {
#if CYTHON_COMPILING_IN_CPYTHON
        from_list = __Pyx_PyList_FromArray(imported_names, len_imported_names);
        if (unlikely(!from_list))
            goto bad;
#else
        from_list = PyList_New(len_imported_names);
        if (unlikely(!from_list)) goto bad;
        for (Py_ssize_t i=0; i<len_imported_names; ++i) {
            if (PyList_SetItem(from_list, i, __Pyx_NewRef(imported_names[i])) < 0) goto bad;
        }
#endif
    }
    if (level == -1) {
        const char* package_sep = strchr(__Pyx_MODULE_NAME, '.');
        if (package_sep != (0)) {
            module = PyImport_ImportModuleLevelObject(
                name, moddict, empty_dict, from_list, 1);
            if (unlikely(!module)) {
                if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                    goto bad;
                PyErr_Clear();
            }
        }
        level = 0;
    }
    if (!module) {
        module = PyImport_ImportModuleLevelObject(
            name, moddict, empty_dict, from_list, level);
    }
bad:
    Py_XDECREF(from_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level) {
    return __Pyx__Import(name, imported_names, len_imported_names, qualname, __pyx_mstate_global->__pyx_d, level);
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_mstate_global->__pyx_kp_u_);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400) ||\
                CYTHON_COMPILING_IN_GRAAL
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError, "cannot import name %S", name);
    }
    return value;
}

/* ListPack */
static PyObject *__Pyx_PyList_Pack(Py_ssize_t n, ...) {
    va_list va;
    PyObject *l = PyList_New(n);
    va_start(va, n);
    if (unlikely(!l)) goto end;
    for (Py_ssize_t i=0; i<n; ++i) {
        PyObject *arg = va_arg(va, PyObject*);
        Py_INCREF(arg);
        if (__Pyx_PyList_SET_ITEM(l, i, arg) != (0)) {
            Py_CLEAR(l);
            goto end;
        }
    }
    end:
    va_end(va);
    return l;
}

/* GetAttr (used by Globals) */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
    if (likely(PyUnicode_Check(n)))
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* Globals */
static PyObject* __Pyx_Globals(void) {
    return __Pyx_NewRef(__pyx_mstate_global->__pyx_d);
}

/* PyObjectVectorCallKwBuilder */
#if CYTHON_VECTORCALL
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_PyObject_FastCallDict;
    if (__Pyx_PyTuple_SET_ITEM(builder, n, key) != (0)) return -1;
    Py_INCREF(key);
    args[n] = value;
    return 0;
}
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_VectorcallBuilder_AddArgStr;
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n);
}
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    PyObject *pyKey = PyUnicode_FromString(key);
    if (!pyKey) return -1;
    return __Pyx_VectorcallBuilder_AddArg(pyKey, value, builder, args, n);
}
#else // CYTHON_VECTORCALL
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, CYTHON_UNUSED PyObject **args, CYTHON_UNUSED int n) {
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return PyDict_SetItem(builder, key, value);
}
#endif

/* GetBuiltinName (used by GetModuleGlobalName) */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_mstate_global->__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
            "name '%U' is not defined", name);
    }
    return result;
}

/* PyDictVersioning (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        if (!PyErr_Occurred())
            PyErr_SetNone(PyExc_NameError);
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
    PyErr_Clear();
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    if (unlikely(__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, name, &result) == -1)) PyErr_Clear();
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return result;
    }
#else
    result = _PyDict_GetItem_KnownHash(__pyx_mstate_global->__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* dict_setdefault (used by CLineInTraceback) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value) {
    PyObject* value;
#if __PYX_LIMITED_VERSION_HEX >= 0x030F0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4)
    PyDict_SetDefaultRef(d, key, default_value, &value);
#elif CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    PyObject *args[] = {d, key, default_value};
    value = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_n_u_setdefault, args, 3 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#elif CYTHON_COMPILING_IN_LIMITED_API
    value = PyObject_CallMethodObjArgs(d, __pyx_mstate_global->__pyx_n_u_setdefault, key, default_value, NULL);
#else
    value = PyDict_SetDefault(d, key, default_value);
    if (unlikely(!value)) return NULL;
    Py_INCREF(value);
#endif
    return value;
}

/* CLineInTraceback (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
#define __Pyx_PyProbablyModule_GetDict(o) __Pyx_XNewRef(PyModule_GetDict(o))
#elif !CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyProbablyModule_GetDict(o) PyObject_GenericGetDict(o, NULL);
#else
PyObject* __Pyx_PyProbablyModule_GetDict(PyObject *o) {
    PyObject **dict_ptr = _PyObject_GetDictPtr(o);
    return dict_ptr ? __Pyx_XNewRef(*dict_ptr) : NULL;
}
#endif
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline = NULL;
    PyObject *ptype, *pvalue, *ptraceback;
    PyObject *cython_runtime_dict;
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_mstate_global->__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
    cython_runtime_dict = __Pyx_PyProbablyModule_GetDict(__pyx_mstate_global->__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, cython_runtime_dict,
            __Pyx_PyDict_SetDefault(cython_runtime_dict, __pyx_mstate_global->__pyx_n_u_cline_in_traceback, Py_False))
    }
    if (use_cline == NULL || use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    Py_XDECREF(use_cline);
    Py_XDECREF(cython_runtime_dict);
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache (used by AddTraceback) */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static __Pyx_CachedCodeObjectType *__pyx__find_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line) {
    __Pyx_CachedCodeObjectType* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!code_cache->entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if (unlikely(pos >= code_cache->count) || unlikely(code_cache->entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = code_cache->entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__find_code_object;
    return NULL; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just miss.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type old_count = __pyx_atomic_incr_acq_rel(&code_cache->accessor_count);
    if (old_count < 0) {
        __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
        return NULL;
    }
#endif
    __Pyx_CachedCodeObjectType *result = __pyx__find_code_object(code_cache, code_line);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
#endif
    return result;
#endif
}
static void __pyx__insert_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line, __Pyx_CachedCodeObjectType* code_object)
{
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = code_cache->entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            code_cache->entries = entries;
            code_cache->max_count = 64;
            code_cache->count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if ((pos < code_cache->count) && unlikely(code_cache->entries[pos].code_line == code_line)) {
        __Pyx_CachedCodeObjectType* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_INCREF(code_object);
        Py_DECREF(tmp);
        return;
    }
    if (code_cache->count == code_cache->max_count) {
        int new_max = code_cache->max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            code_cache->entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        code_cache->entries = entries;
        code_cache->max_count = new_max;
    }
    for (i=code_cache->count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    code_cache->count++;
    Py_INCREF(code_object);
}
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__insert_code_object;
    return; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just fail.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type expected = 0;
    if (!__pyx_atomic_int_cmp_exchange(&code_cache->accessor_count, &expected, INT_MIN)) {
        return;
    }
#endif
    __pyx__insert_code_object(code_cache, code_line, code_object);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_sub(&code_cache->accessor_count, INT_MIN);
#endif
#endif
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result = PyObject_Call(replace, __pyx_mstate_global->__pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!code_object) {
        code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
        if (unlikely(!code_object)) goto bad;
        py_py_line = PyLong_FromLong(py_line);
        if (unlikely(!py_py_line)) goto bad;
        if (c_line) {
            py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        } else {
            py_funcname = PyUnicode_FromString(funcname);
        }
        if (unlikely(!py_funcname)) goto bad;
        dict = PyDict_New();
        if (unlikely(!dict)) goto bad;
        {
            PyObject *old_code_object = code_object;
            code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
            Py_DECREF(old_code_object);
        }
        if (unlikely(!code_object)) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, code_object);
    } else {
        dict = PyDict_New();
    }
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    if (c_line) {
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
    }
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_mstate_global->__pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_PY_LONG_LONG(unsigned PY_LONG_LONG value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned PY_LONG_LONG neg_one = (unsigned PY_LONG_LONG) -1, const_zero = (unsigned PY_LONG_LONG) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned PY_LONG_LONG) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned PY_LONG_LONG) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned PY_LONG_LONG),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(unsigned PY_LONG_LONG));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_access_flags(enum ibv_access_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_access_flags neg_one = (enum ibv_access_flags) -1, const_zero = (enum ibv_access_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_access_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_access_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_access_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_access_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_access_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_access_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_access_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_atomic_cap(enum ibv_atomic_cap value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_atomic_cap neg_one = (enum ibv_atomic_cap) -1, const_zero = (enum ibv_atomic_cap) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_atomic_cap) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_atomic_cap) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_atomic_cap) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_atomic_cap) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_atomic_cap) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_atomic_cap),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_atomic_cap));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_counter_description(enum ibv_counter_description value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_counter_description neg_one = (enum ibv_counter_description) -1, const_zero = (enum ibv_counter_description) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_counter_description) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_counter_description) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_counter_description) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_counter_description) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_counter_description) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_counter_description),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_counter_description));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_cq_init_attr_mask(enum ibv_cq_init_attr_mask value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_cq_init_attr_mask neg_one = (enum ibv_cq_init_attr_mask) -1, const_zero = (enum ibv_cq_init_attr_mask) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_cq_init_attr_mask) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_cq_init_attr_mask) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_cq_init_attr_mask) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_cq_init_attr_mask) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_cq_init_attr_mask) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_cq_init_attr_mask),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_cq_init_attr_mask));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_create_cq_attr_flags(enum ibv_create_cq_attr_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_create_cq_attr_flags neg_one = (enum ibv_create_cq_attr_flags) -1, const_zero = (enum ibv_create_cq_attr_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_create_cq_attr_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_create_cq_attr_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_create_cq_attr_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_create_cq_attr_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_create_cq_attr_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_create_cq_attr_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_create_cq_attr_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_create_cq_wc_flags(enum ibv_create_cq_wc_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_create_cq_wc_flags neg_one = (enum ibv_create_cq_wc_flags) -1, const_zero = (enum ibv_create_cq_wc_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_create_cq_wc_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_create_cq_wc_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_create_cq_wc_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_create_cq_wc_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_create_cq_wc_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_create_cq_wc_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_create_cq_wc_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_device_cap_flags(enum ibv_device_cap_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_device_cap_flags neg_one = (enum ibv_device_cap_flags) -1, const_zero = (enum ibv_device_cap_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_device_cap_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_device_cap_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_device_cap_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_device_cap_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_device_cap_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_device_cap_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_device_cap_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_dmah_init_attr_mask(enum ibv_dmah_init_attr_mask value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_dmah_init_attr_mask neg_one = (enum ibv_dmah_init_attr_mask) -1, const_zero = (enum ibv_dmah_init_attr_mask) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_dmah_init_attr_mask) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_dmah_init_attr_mask) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_dmah_init_attr_mask) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_dmah_init_attr_mask) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_dmah_init_attr_mask) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_dmah_init_attr_mask),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_dmah_init_attr_mask));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_event_type(enum ibv_event_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_event_type neg_one = (enum ibv_event_type) -1, const_zero = (enum ibv_event_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_event_type) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_event_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_event_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_event_type) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_event_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_event_type),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_event_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_flow_attr_type(enum ibv_flow_attr_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_flow_attr_type neg_one = (enum ibv_flow_attr_type) -1, const_zero = (enum ibv_flow_attr_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_flow_attr_type) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_flow_attr_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_flow_attr_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_flow_attr_type) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_flow_attr_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_flow_attr_type),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_flow_attr_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_flow_flags(enum ibv_flow_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_flow_flags neg_one = (enum ibv_flow_flags) -1, const_zero = (enum ibv_flow_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_flow_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_flow_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_flow_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_flow_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_flow_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_flow_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_flow_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_flow_spec_type(enum ibv_flow_spec_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_flow_spec_type neg_one = (enum ibv_flow_spec_type) -1, const_zero = (enum ibv_flow_spec_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_flow_spec_type) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_flow_spec_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_flow_spec_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_flow_spec_type) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_flow_spec_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_flow_spec_type),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_flow_spec_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_fork_status(enum ibv_fork_status value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_fork_status neg_one = (enum ibv_fork_status) -1, const_zero = (enum ibv_fork_status) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_fork_status) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_fork_status) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_fork_status) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_fork_status) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_fork_status) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_fork_status),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_fork_status));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_gid_type(enum ibv_gid_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_gid_type neg_one = (enum ibv_gid_type) -1, const_zero = (enum ibv_gid_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_gid_type) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_gid_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_gid_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_gid_type) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_gid_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_gid_type),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_gid_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_gid_type_sysfs(enum ibv_gid_type_sysfs value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_gid_type_sysfs neg_one = (enum ibv_gid_type_sysfs) -1, const_zero = (enum ibv_gid_type_sysfs) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_gid_type_sysfs) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_gid_type_sysfs) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_gid_type_sysfs) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_gid_type_sysfs) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_gid_type_sysfs) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_gid_type_sysfs),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_gid_type_sysfs));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_mig_state(enum ibv_mig_state value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_mig_state neg_one = (enum ibv_mig_state) -1, const_zero = (enum ibv_mig_state) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_mig_state) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_mig_state) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_mig_state) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_mig_state) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_mig_state) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_mig_state),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_mig_state));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_mr_init_attr_mask(enum ibv_mr_init_attr_mask value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_mr_init_attr_mask neg_one = (enum ibv_mr_init_attr_mask) -1, const_zero = (enum ibv_mr_init_attr_mask) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_mr_init_attr_mask) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_mr_init_attr_mask) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_mr_init_attr_mask) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_mr_init_attr_mask) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_mr_init_attr_mask) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_mr_init_attr_mask),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_mr_init_attr_mask));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_mtu(enum ibv_mtu value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_mtu neg_one = (enum ibv_mtu) -1, const_zero = (enum ibv_mtu) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_mtu) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_mtu) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_mtu) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_mtu) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_mtu) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_mtu),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_mtu));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_mw_type(enum ibv_mw_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_mw_type neg_one = (enum ibv_mw_type) -1, const_zero = (enum ibv_mw_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_mw_type) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_mw_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_mw_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_mw_type) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_mw_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_mw_type),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_mw_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_node_type(enum ibv_node_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_node_type neg_one = (enum ibv_node_type) -1, const_zero = (enum ibv_node_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_node_type) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_node_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_node_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_node_type) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_node_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_node_type),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_node_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_odp_general_caps(enum ibv_odp_general_caps value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_odp_general_caps neg_one = (enum ibv_odp_general_caps) -1, const_zero = (enum ibv_odp_general_caps) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_odp_general_caps) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_odp_general_caps) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_odp_general_caps) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_odp_general_caps) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_odp_general_caps) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_odp_general_caps),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_odp_general_caps));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_odp_transport_cap_bits(enum ibv_odp_transport_cap_bits value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_odp_transport_cap_bits neg_one = (enum ibv_odp_transport_cap_bits) -1, const_zero = (enum ibv_odp_transport_cap_bits) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_odp_transport_cap_bits) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_odp_transport_cap_bits) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_odp_transport_cap_bits) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_odp_transport_cap_bits) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_odp_transport_cap_bits) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_odp_transport_cap_bits),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_odp_transport_cap_bits));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_ops_flags(enum ibv_ops_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_ops_flags neg_one = (enum ibv_ops_flags) -1, const_zero = (enum ibv_ops_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_ops_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_ops_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_ops_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_ops_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_ops_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_ops_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_ops_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_ops_wr_opcode(enum ibv_ops_wr_opcode value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_ops_wr_opcode neg_one = (enum ibv_ops_wr_opcode) -1, const_zero = (enum ibv_ops_wr_opcode) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_ops_wr_opcode) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_ops_wr_opcode) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_ops_wr_opcode) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_ops_wr_opcode) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_ops_wr_opcode) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_ops_wr_opcode),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_ops_wr_opcode));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_parent_domain_init_attr_mask(enum ibv_parent_domain_init_attr_mask value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_parent_domain_init_attr_mask neg_one = (enum ibv_parent_domain_init_attr_mask) -1, const_zero = (enum ibv_parent_domain_init_attr_mask) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_parent_domain_init_attr_mask) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_parent_domain_init_attr_mask) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_parent_domain_init_attr_mask) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_parent_domain_init_attr_mask) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_parent_domain_init_attr_mask) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_parent_domain_init_attr_mask),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_parent_domain_init_attr_mask));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_placement_type(enum ibv_placement_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_placement_type neg_one = (enum ibv_placement_type) -1, const_zero = (enum ibv_placement_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_placement_type) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_placement_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_placement_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_placement_type) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_placement_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_placement_type),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_placement_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_port_cap_flags(enum ibv_port_cap_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_port_cap_flags neg_one = (enum ibv_port_cap_flags) -1, const_zero = (enum ibv_port_cap_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_port_cap_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_port_cap_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_port_cap_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_port_cap_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_port_cap_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_port_cap_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_port_cap_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_port_cap_flags2(enum ibv_port_cap_flags2 value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_port_cap_flags2 neg_one = (enum ibv_port_cap_flags2) -1, const_zero = (enum ibv_port_cap_flags2) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_port_cap_flags2) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_port_cap_flags2) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_port_cap_flags2) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_port_cap_flags2) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_port_cap_flags2) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_port_cap_flags2),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_port_cap_flags2));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_port_state(enum ibv_port_state value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_port_state neg_one = (enum ibv_port_state) -1, const_zero = (enum ibv_port_state) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_port_state) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_port_state) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_port_state) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_port_state) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_port_state) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_port_state),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_port_state));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_qp_attr_mask(enum ibv_qp_attr_mask value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_qp_attr_mask neg_one = (enum ibv_qp_attr_mask) -1, const_zero = (enum ibv_qp_attr_mask) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_qp_attr_mask) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_qp_attr_mask) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_qp_attr_mask) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_qp_attr_mask) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_qp_attr_mask) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_qp_attr_mask),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_qp_attr_mask));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_qp_create_flags(enum ibv_qp_create_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_qp_create_flags neg_one = (enum ibv_qp_create_flags) -1, const_zero = (enum ibv_qp_create_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_qp_create_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_qp_create_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_qp_create_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_qp_create_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_qp_create_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_qp_create_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_qp_create_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_qp_create_send_ops_flags(enum ibv_qp_create_send_ops_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_qp_create_send_ops_flags neg_one = (enum ibv_qp_create_send_ops_flags) -1, const_zero = (enum ibv_qp_create_send_ops_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_qp_create_send_ops_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_qp_create_send_ops_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_qp_create_send_ops_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_qp_create_send_ops_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_qp_create_send_ops_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_qp_create_send_ops_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_qp_create_send_ops_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_qp_init_attr_mask(enum ibv_qp_init_attr_mask value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_qp_init_attr_mask neg_one = (enum ibv_qp_init_attr_mask) -1, const_zero = (enum ibv_qp_init_attr_mask) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_qp_init_attr_mask) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_qp_init_attr_mask) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_qp_init_attr_mask) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_qp_init_attr_mask) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_qp_init_attr_mask) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_qp_init_attr_mask),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_qp_init_attr_mask));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_qp_state(enum ibv_qp_state value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_qp_state neg_one = (enum ibv_qp_state) -1, const_zero = (enum ibv_qp_state) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_qp_state) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_qp_state) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_qp_state) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_qp_state) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_qp_state) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_qp_state),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_qp_state));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_qp_type(enum ibv_qp_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_qp_type neg_one = (enum ibv_qp_type) -1, const_zero = (enum ibv_qp_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_qp_type) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_qp_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_qp_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_qp_type) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_qp_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_qp_type),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_qp_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_query_qp_data_in_order_caps(enum ibv_query_qp_data_in_order_caps value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_query_qp_data_in_order_caps neg_one = (enum ibv_query_qp_data_in_order_caps) -1, const_zero = (enum ibv_query_qp_data_in_order_caps) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_query_qp_data_in_order_caps) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_query_qp_data_in_order_caps) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_query_qp_data_in_order_caps) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_query_qp_data_in_order_caps) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_query_qp_data_in_order_caps) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_query_qp_data_in_order_caps),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_query_qp_data_in_order_caps));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_query_qp_data_in_order_flags(enum ibv_query_qp_data_in_order_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_query_qp_data_in_order_flags neg_one = (enum ibv_query_qp_data_in_order_flags) -1, const_zero = (enum ibv_query_qp_data_in_order_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_query_qp_data_in_order_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_query_qp_data_in_order_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_query_qp_data_in_order_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_query_qp_data_in_order_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_query_qp_data_in_order_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_query_qp_data_in_order_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_query_qp_data_in_order_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_raw_packet_caps(enum ibv_raw_packet_caps value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_raw_packet_caps neg_one = (enum ibv_raw_packet_caps) -1, const_zero = (enum ibv_raw_packet_caps) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_raw_packet_caps) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_raw_packet_caps) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_raw_packet_caps) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_raw_packet_caps) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_raw_packet_caps) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_raw_packet_caps),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_raw_packet_caps));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_read_counters_flags(enum ibv_read_counters_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_read_counters_flags neg_one = (enum ibv_read_counters_flags) -1, const_zero = (enum ibv_read_counters_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_read_counters_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_read_counters_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_read_counters_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_read_counters_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_read_counters_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_read_counters_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_read_counters_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_rereg_mr_err_code(enum ibv_rereg_mr_err_code value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_rereg_mr_err_code neg_one = (enum ibv_rereg_mr_err_code) -1, const_zero = (enum ibv_rereg_mr_err_code) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_rereg_mr_err_code) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_rereg_mr_err_code) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_rereg_mr_err_code) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_rereg_mr_err_code) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_rereg_mr_err_code) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_rereg_mr_err_code),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_rereg_mr_err_code));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_rereg_mr_flags(enum ibv_rereg_mr_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_rereg_mr_flags neg_one = (enum ibv_rereg_mr_flags) -1, const_zero = (enum ibv_rereg_mr_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_rereg_mr_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_rereg_mr_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_rereg_mr_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_rereg_mr_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_rereg_mr_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_rereg_mr_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_rereg_mr_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_rx_hash_fields(enum ibv_rx_hash_fields value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_rx_hash_fields neg_one = (enum ibv_rx_hash_fields) -1, const_zero = (enum ibv_rx_hash_fields) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_rx_hash_fields) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_rx_hash_fields) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_rx_hash_fields) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_rx_hash_fields) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_rx_hash_fields) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_rx_hash_fields),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_rx_hash_fields));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_rx_hash_function_flags(enum ibv_rx_hash_function_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_rx_hash_function_flags neg_one = (enum ibv_rx_hash_function_flags) -1, const_zero = (enum ibv_rx_hash_function_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_rx_hash_function_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_rx_hash_function_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_rx_hash_function_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_rx_hash_function_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_rx_hash_function_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_rx_hash_function_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_rx_hash_function_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_selectivity_level(enum ibv_selectivity_level value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_selectivity_level neg_one = (enum ibv_selectivity_level) -1, const_zero = (enum ibv_selectivity_level) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_selectivity_level) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_selectivity_level) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_selectivity_level) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_selectivity_level) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_selectivity_level) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_selectivity_level),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_selectivity_level));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_send_flags(enum ibv_send_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_send_flags neg_one = (enum ibv_send_flags) -1, const_zero = (enum ibv_send_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_send_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_send_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_send_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_send_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_send_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_send_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_send_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_srq_attr_mask(enum ibv_srq_attr_mask value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_srq_attr_mask neg_one = (enum ibv_srq_attr_mask) -1, const_zero = (enum ibv_srq_attr_mask) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_srq_attr_mask) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_srq_attr_mask) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_srq_attr_mask) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_srq_attr_mask) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_srq_attr_mask) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_srq_attr_mask),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_srq_attr_mask));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_srq_init_attr_mask(enum ibv_srq_init_attr_mask value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_srq_init_attr_mask neg_one = (enum ibv_srq_init_attr_mask) -1, const_zero = (enum ibv_srq_init_attr_mask) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_srq_init_attr_mask) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_srq_init_attr_mask) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_srq_init_attr_mask) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_srq_init_attr_mask) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_srq_init_attr_mask) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_srq_init_attr_mask),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_srq_init_attr_mask));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_srq_type(enum ibv_srq_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_srq_type neg_one = (enum ibv_srq_type) -1, const_zero = (enum ibv_srq_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_srq_type) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_srq_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_srq_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_srq_type) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_srq_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_srq_type),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_srq_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_tm_cap_flags(enum ibv_tm_cap_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_tm_cap_flags neg_one = (enum ibv_tm_cap_flags) -1, const_zero = (enum ibv_tm_cap_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_tm_cap_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_tm_cap_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_tm_cap_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_tm_cap_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_tm_cap_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_tm_cap_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_tm_cap_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_tmh_op(enum ibv_tmh_op value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_tmh_op neg_one = (enum ibv_tmh_op) -1, const_zero = (enum ibv_tmh_op) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_tmh_op) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_tmh_op) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_tmh_op) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_tmh_op) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_tmh_op) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_tmh_op),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_tmh_op));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_tph_mem_type(enum ibv_tph_mem_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_tph_mem_type neg_one = (enum ibv_tph_mem_type) -1, const_zero = (enum ibv_tph_mem_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_tph_mem_type) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_tph_mem_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_tph_mem_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_tph_mem_type) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_tph_mem_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_tph_mem_type),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_tph_mem_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_transport_type(enum ibv_transport_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_transport_type neg_one = (enum ibv_transport_type) -1, const_zero = (enum ibv_transport_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_transport_type) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_transport_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_transport_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_transport_type) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_transport_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_transport_type),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_transport_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_values_mask(enum ibv_values_mask value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_values_mask neg_one = (enum ibv_values_mask) -1, const_zero = (enum ibv_values_mask) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_values_mask) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_values_mask) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_values_mask) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_values_mask) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_values_mask) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_values_mask),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_values_mask));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_wc_flags(enum ibv_wc_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_wc_flags neg_one = (enum ibv_wc_flags) -1, const_zero = (enum ibv_wc_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_wc_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_wc_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_wc_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_wc_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_wc_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_wc_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_wc_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_wc_opcode(enum ibv_wc_opcode value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_wc_opcode neg_one = (enum ibv_wc_opcode) -1, const_zero = (enum ibv_wc_opcode) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_wc_opcode) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_wc_opcode) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_wc_opcode) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_wc_opcode) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_wc_opcode) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_wc_opcode),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_wc_opcode));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_wc_status(enum ibv_wc_status value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_wc_status neg_one = (enum ibv_wc_status) -1, const_zero = (enum ibv_wc_status) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_wc_status) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_wc_status) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_wc_status) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_wc_status) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_wc_status) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_wc_status),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_wc_status));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_wq_attr_mask(enum ibv_wq_attr_mask value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_wq_attr_mask neg_one = (enum ibv_wq_attr_mask) -1, const_zero = (enum ibv_wq_attr_mask) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_wq_attr_mask) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_wq_attr_mask) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_wq_attr_mask) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_wq_attr_mask) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_wq_attr_mask) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_wq_attr_mask),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_wq_attr_mask));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_wq_flags(enum ibv_wq_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_wq_flags neg_one = (enum ibv_wq_flags) -1, const_zero = (enum ibv_wq_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_wq_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_wq_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_wq_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_wq_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_wq_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_wq_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_wq_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_wq_init_attr_mask(enum ibv_wq_init_attr_mask value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_wq_init_attr_mask neg_one = (enum ibv_wq_init_attr_mask) -1, const_zero = (enum ibv_wq_init_attr_mask) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_wq_init_attr_mask) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_wq_init_attr_mask) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_wq_init_attr_mask) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_wq_init_attr_mask) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_wq_init_attr_mask) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_wq_init_attr_mask),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_wq_init_attr_mask));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_wq_state(enum ibv_wq_state value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_wq_state neg_one = (enum ibv_wq_state) -1, const_zero = (enum ibv_wq_state) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_wq_state) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_wq_state) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_wq_state) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_wq_state) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_wq_state) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_wq_state),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_wq_state));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_wq_type(enum ibv_wq_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_wq_type neg_one = (enum ibv_wq_type) -1, const_zero = (enum ibv_wq_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_wq_type) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_wq_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_wq_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_wq_type) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_wq_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_wq_type),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_wq_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_wr_opcode(enum ibv_wr_opcode value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_wr_opcode neg_one = (enum ibv_wr_opcode) -1, const_zero = (enum ibv_wr_opcode) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_wr_opcode) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_wr_opcode) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_wr_opcode) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_wr_opcode) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_wr_opcode) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_wr_opcode),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_wr_opcode));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__ibv_xrcd_init_attr_mask(enum ibv_xrcd_init_attr_mask value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_xrcd_init_attr_mask neg_one = (enum ibv_xrcd_init_attr_mask) -1, const_zero = (enum ibv_xrcd_init_attr_mask) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum ibv_xrcd_init_attr_mask) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_xrcd_init_attr_mask) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum ibv_xrcd_init_attr_mask) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum ibv_xrcd_init_attr_mask) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum ibv_xrcd_init_attr_mask) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum ibv_xrcd_init_attr_mask),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum ibv_xrcd_init_attr_mask));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
static __Pyx_TypeName
__Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp)
{
    PyObject *module = NULL, *name = NULL, *result = NULL;
    #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
    name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_qualname);
    #else
    name = PyType_GetQualName(tp);
    #endif
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) goto bad;
    module = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_module_2);
    if (unlikely(module == NULL) || unlikely(!PyUnicode_Check(module))) goto bad;
    if (PyUnicode_CompareWithASCIIString(module, "builtins") == 0) {
        result = name;
        name = NULL;
        goto done;
    }
    result = PyUnicode_FromFormat("%U.%U", module, name);
    if (unlikely(result == NULL)) goto bad;
  done:
    Py_XDECREF(name);
    Py_XDECREF(module);
    return result;
  bad:
    PyErr_Clear();
    if (name) {
        result = name;
        name = NULL;
    } else {
        result = __Pyx_NewRef(__pyx_mstate_global->__pyx_kp_u__2);
    }
    goto done;
}
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPyVerify (used by CIntFromPy) */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyLong_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyLong_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* GetRuntimeVersion */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
void __Pyx_init_runtime_version(void) {
    if (__Pyx_cached_runtime_version == 0) {
        const char* rt_version = Py_GetVersion();
        unsigned long version = 0;
        unsigned long factor = 0x01000000UL;
        unsigned int digit = 0;
        int i = 0;
        while (factor) {
            while ('0' <= rt_version[i] && rt_version[i] <= '9') {
                digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
                ++i;
            }
            version += factor * digit;
            if (rt_version[i] != '.')
                break;
            digit = 0;
            factor >>= 8;
            ++i;
        }
        __Pyx_cached_runtime_version = version;
    }
}
#endif
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    return Py_Version & ~0xFFUL;
#else
    return __Pyx_cached_runtime_version;
#endif
}

/* AddModuleRef */
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  static PyObject *__Pyx_PyImport_AddModuleObjectRef(PyObject *name) {
      PyObject *module_dict = PyImport_GetModuleDict();
      PyObject *m;
      if (PyMapping_GetOptionalItem(module_dict, name, &m) < 0) {
          return NULL;
      }
      if (m != NULL && PyModule_Check(m)) {
          return m;
      }
      Py_XDECREF(m);
      m = PyModule_NewObject(name);
      if (m == NULL)
          return NULL;
      if (PyDict_CheckExact(module_dict)) {
          PyObject *new_m;
          (void)PyDict_SetDefaultRef(module_dict, name, m, &new_m);
          Py_DECREF(m);
          return new_m;
      } else {
           if (PyObject_SetItem(module_dict, name, m) != 0) {
                Py_DECREF(m);
                return NULL;
            }
            return m;
      }
  }
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *py_name = PyUnicode_FromString(name);
      if (!py_name) return NULL;
      PyObject *module = __Pyx_PyImport_AddModuleObjectRef(py_name);
      Py_DECREF(py_name);
      return module;
  }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From___pyx_anon_enum(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* DecompressString */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo) {
    PyObject *module = NULL, *decompress, *compressed_bytes, *decompressed;
    const char* module_name = algo == 3 ? "compression.zstd" : algo == 2 ? "bz2" : "zlib";
    PyObject *methodname = PyUnicode_FromString("decompress");
    if (unlikely(!methodname)) return NULL;
    #if __PYX_LIMITED_VERSION_HEX >= 0x030e0000
    if (algo == 3) {
        PyObject *fromlist = Py_BuildValue("[O]", methodname);
        if (unlikely(!fromlist)) goto bad;
        module = PyImport_ImportModuleLevel("compression.zstd", NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
    } else
    #endif
        module = PyImport_ImportModule(module_name);
    if (unlikely(!module)) goto import_failed;
    decompress = PyObject_GetAttr(module, methodname);
    if (unlikely(!decompress)) goto import_failed;
    {
        #ifdef __cplusplus
            char *memview_bytes = const_cast<char*>(s);
        #else
            #if defined(__clang__)
              #pragma clang diagnostic push
              #pragma clang diagnostic ignored "-Wcast-qual"
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic push
              #pragma GCC diagnostic ignored "-Wcast-qual"
            #endif
            char *memview_bytes = (char*) s;
            #if defined(__clang__)
              #pragma clang diagnostic pop
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic pop
            #endif
        #endif
        #if CYTHON_COMPILING_IN_LIMITED_API && !defined(PyBUF_READ)
        int memview_flags = 0x100;
        #else
        int memview_flags = PyBUF_READ;
        #endif
        compressed_bytes = PyMemoryView_FromMemory(memview_bytes, length, memview_flags);
    }
    if (unlikely(!compressed_bytes)) {
        Py_DECREF(decompress);
        goto bad;
    }
    decompressed = PyObject_CallFunctionObjArgs(decompress, compressed_bytes, NULL);
    Py_DECREF(compressed_bytes);
    Py_DECREF(decompress);
    Py_DECREF(module);
    Py_DECREF(methodname);
    return decompressed;
import_failed:
    PyErr_Format(PyExc_ImportError,
        "Failed to import '%.20s.decompress' - cannot initialise module strings. "
        "String compression was configured with the C macro 'CYTHON_COMPRESS_STRINGS=%d'.",
        module_name, algo);
bad:
    Py_XDECREF(module);
    Py_DECREF(methodname);
    return NULL;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    {
        const char* result;
        Py_ssize_t unicode_length;
        CYTHON_MAYBE_UNUSED_VAR(unicode_length); // only for __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (unlikely(PyArg_Parse(o, "s#", &result, length) < 0)) return NULL;
        #else
        result = PyUnicode_AsUTF8AndSize(o, length);
        #endif
        #if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        unicode_length = PyUnicode_GetLength(o);
        if (unlikely(unicode_length < 0)) return NULL;
        if (unlikely(unicode_length != *length)) {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
        #endif
        return result;
    }
#else
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
}
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
    if (PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
    if (PyByteArray_Check(o)) {
#if (CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS) || (CYTHON_COMPILING_IN_PYPY && (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE)))
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
#else
        *length = PyByteArray_Size(o);
        if (*length == -1) return NULL;
        return PyByteArray_AsString(o);
#endif
    } else
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_LongWrongResultType(PyObject* result) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(result));
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
    PyErr_Format(PyExc_TypeError,
                 "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ")",
                 result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  PyObject *res = NULL;
  if (likely(PyLong_Check(x)))
      return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  if (likely(m && m->nb_int)) {
      res = m->nb_int(x);
  }
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
      res = PyNumber_Long(x);
  }
#endif
  if (likely(res)) {
      if (unlikely(!PyLong_CheckExact(res))) {
          return __Pyx_PyNumber_LongWrongResultType(res);
      }
  }
  else if (!PyErr_Occurred()) {
      PyErr_SetString(PyExc_TypeError,
                      "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyLong_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyLong_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b) {
    CYTHON_UNUSED_VAR(b);
    return __Pyx_NewRef(Py_None);
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return __Pyx_NewRef(b ? Py_True: Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t ival) {
    return PyLong_FromSize_t(ival);
}


/* MultiPhaseInitModuleState */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
#ifndef CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#if (CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX >= 0x030C0000)
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 1
#else
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 0
#endif
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE && !CYTHON_ATOMICS
#error "Module state with PEP489 requires atomics. Currently that's one of\
 C11, C++11, gcc atomic intrinsics or MSVC atomic intrinsics"
#endif
#if !CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#define __Pyx_ModuleStateLookup_Lock()
#define __Pyx_ModuleStateLookup_Unlock()
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
static PyMutex __Pyx_ModuleStateLookup_mutex = {0};
#define __Pyx_ModuleStateLookup_Lock() PyMutex_Lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() PyMutex_Unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(__cplusplus) && __cplusplus >= 201103L
#include <mutex>
static std::mutex __Pyx_ModuleStateLookup_mutex;
#define __Pyx_ModuleStateLookup_Lock() __Pyx_ModuleStateLookup_mutex.lock()
#define __Pyx_ModuleStateLookup_Unlock() __Pyx_ModuleStateLookup_mutex.unlock()
#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ > 201112L) && !defined(__STDC_NO_THREADS__)
#include <threads.h>
static mtx_t __Pyx_ModuleStateLookup_mutex;
static once_flag __Pyx_ModuleStateLookup_mutex_once_flag = ONCE_FLAG_INIT;
static void __Pyx_ModuleStateLookup_initialize_mutex(void) {
    mtx_init(&__Pyx_ModuleStateLookup_mutex, mtx_plain);
}
#define __Pyx_ModuleStateLookup_Lock()\
  call_once(&__Pyx_ModuleStateLookup_mutex_once_flag, __Pyx_ModuleStateLookup_initialize_mutex);\
  mtx_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() mtx_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(HAVE_PTHREAD_H)
#include <pthread.h>
static pthread_mutex_t __Pyx_ModuleStateLookup_mutex = PTHREAD_MUTEX_INITIALIZER;
#define __Pyx_ModuleStateLookup_Lock() pthread_mutex_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() pthread_mutex_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(_WIN32)
#include <Windows.h>  // synchapi.h on its own doesn't work
static SRWLOCK __Pyx_ModuleStateLookup_mutex = SRWLOCK_INIT;
#define __Pyx_ModuleStateLookup_Lock() AcquireSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() ReleaseSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#else
#error "No suitable lock available for CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE.\
 Requires C standard >= C11, or C++ standard >= C++11,\
 or pthreads, or the Windows 32 API, or Python >= 3.13."
#endif
typedef struct {
    int64_t id;
    PyObject *module;
} __Pyx_InterpreterIdAndModule;
typedef struct {
    char interpreter_id_as_index;
    Py_ssize_t count;
    Py_ssize_t allocated;
    __Pyx_InterpreterIdAndModule table[1];
} __Pyx_ModuleStateLookupData;
#define __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE 32
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_int_type __Pyx_ModuleStateLookup_read_counter = 0;
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_ptr_type __Pyx_ModuleStateLookup_data = 0;
#else
static __Pyx_ModuleStateLookupData* __Pyx_ModuleStateLookup_data = NULL;
#endif
static __Pyx_InterpreterIdAndModule* __Pyx_State_FindModuleStateLookupTableLowerBound(
        __Pyx_InterpreterIdAndModule* table,
        Py_ssize_t count,
        int64_t interpreterId) {
    __Pyx_InterpreterIdAndModule* begin = table;
    __Pyx_InterpreterIdAndModule* end = begin + count;
    if (begin->id == interpreterId) {
        return begin;
    }
    while ((end - begin) > __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
        __Pyx_InterpreterIdAndModule* halfway = begin + (end - begin)/2;
        if (halfway->id == interpreterId) {
            return halfway;
        }
        if (halfway->id < interpreterId) {
            begin = halfway;
        } else {
            end = halfway;
        }
    }
    for (; begin < end; ++begin) {
        if (begin->id >= interpreterId) return begin;
    }
    return begin;
}
static PyObject *__Pyx_State_FindModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return NULL;
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData* data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
    {
        __pyx_atomic_incr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        if (likely(data)) {
            __Pyx_ModuleStateLookupData* new_data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_acquire(&__Pyx_ModuleStateLookup_data);
            if (likely(data == new_data)) {
                goto read_finished;
            }
        }
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        __Pyx_ModuleStateLookup_Lock();
        __pyx_atomic_incr_relaxed(&__Pyx_ModuleStateLookup_read_counter);
        data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
        __Pyx_ModuleStateLookup_Unlock();
    }
  read_finished:;
#else
    __Pyx_ModuleStateLookupData* data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_InterpreterIdAndModule* found = NULL;
    if (unlikely(!data)) goto end;
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            found = data->table+interpreter_id;
        }
    } else {
        found = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
    }
  end:
    {
        PyObject *result=NULL;
        if (found && found->id == interpreter_id) {
            result = found->module;
        }
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
#endif
        return result;
    }
}
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static void __Pyx_ModuleStateLookup_wait_until_no_readers(void) {
    while (__pyx_atomic_load(&__Pyx_ModuleStateLookup_read_counter) != 0);
}
#else
#define __Pyx_ModuleStateLookup_wait_until_no_readers()
#endif
static int __Pyx_State_AddModuleInterpIdAsIndex(__Pyx_ModuleStateLookupData **old_data, PyObject* module, int64_t interpreter_id) {
    Py_ssize_t to_allocate = (*old_data)->allocated;
    while (to_allocate <= interpreter_id) {
        if (to_allocate == 0) to_allocate = 1;
        else to_allocate *= 2;
    }
    __Pyx_ModuleStateLookupData *new_data = *old_data;
    if (to_allocate != (*old_data)->allocated) {
         new_data = (__Pyx_ModuleStateLookupData *)realloc(
            *old_data,
            sizeof(__Pyx_ModuleStateLookupData)+(to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
        if (!new_data) {
            PyErr_NoMemory();
            return -1;
        }
        for (Py_ssize_t i = new_data->allocated; i < to_allocate; ++i) {
            new_data->table[i].id = i;
            new_data->table[i].module = NULL;
        }
        new_data->allocated = to_allocate;
    }
    new_data->table[interpreter_id].module = module;
    if (new_data->count < interpreter_id+1) {
        new_data->count = interpreter_id+1;
    }
    *old_data = new_data;
    return 0;
}
static void __Pyx_State_ConvertFromInterpIdAsIndex(__Pyx_ModuleStateLookupData *data) {
    __Pyx_InterpreterIdAndModule *read = data->table;
    __Pyx_InterpreterIdAndModule *write = data->table;
    __Pyx_InterpreterIdAndModule *end = read + data->count;
    for (; read<end; ++read) {
        if (read->module) {
            write->id = read->id;
            write->module = read->module;
            ++write;
        }
    }
    data->count = write - data->table;
    for (; write<end; ++write) {
        write->id = 0;
        write->module = NULL;
    }
    data->interpreter_id_as_index = 0;
}
static int __Pyx_State_AddModule(PyObject* module, CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    int result = 0;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *old_data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *old_data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_ModuleStateLookupData *new_data = old_data;
    if (!new_data) {
        new_data = (__Pyx_ModuleStateLookupData *)calloc(1, sizeof(__Pyx_ModuleStateLookupData));
        if (!new_data) {
            result = -1;
            PyErr_NoMemory();
            goto end;
        }
        new_data->allocated = 1;
        new_data->interpreter_id_as_index = 1;
    }
    __Pyx_ModuleStateLookup_wait_until_no_readers();
    if (new_data->interpreter_id_as_index) {
        if (interpreter_id < __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
            result = __Pyx_State_AddModuleInterpIdAsIndex(&new_data, module, interpreter_id);
            goto end;
        }
        __Pyx_State_ConvertFromInterpIdAsIndex(new_data);
    }
    {
        Py_ssize_t insert_at = 0;
        {
            __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
                new_data->table, new_data->count, interpreter_id);
            assert(lower_bound);
            insert_at = lower_bound - new_data->table;
            if (unlikely(insert_at < new_data->count && lower_bound->id == interpreter_id)) {
                lower_bound->module = module;
                goto end;  // already in table, nothing more to do
            }
        }
        if (new_data->count+1 >= new_data->allocated) {
            Py_ssize_t to_allocate = (new_data->count+1)*2;
            new_data =
                (__Pyx_ModuleStateLookupData*)realloc(
                    new_data,
                    sizeof(__Pyx_ModuleStateLookupData) +
                    (to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
            if (!new_data) {
                result = -1;
                new_data = old_data;
                PyErr_NoMemory();
                goto end;
            }
            new_data->allocated = to_allocate;
        }
        ++new_data->count;
        int64_t last_id = interpreter_id;
        PyObject *last_module = module;
        for (Py_ssize_t i=insert_at; i<new_data->count; ++i) {
            int64_t current_id = new_data->table[i].id;
            new_data->table[i].id = last_id;
            last_id = current_id;
            PyObject *current_module = new_data->table[i].module;
            new_data->table[i].module = last_module;
            last_module = current_module;
        }
    }
  end:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, new_data);
#else
    __Pyx_ModuleStateLookup_data = new_data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return result;
}
static int __Pyx_State_RemoveModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *data = __Pyx_ModuleStateLookup_data;
#endif
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            data->table[interpreter_id].module = NULL;
        }
        goto done;
    }
    {
        __Pyx_ModuleStateLookup_wait_until_no_readers();
        __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
        if (!lower_bound) goto done;
        if (lower_bound->id != interpreter_id) goto done;
        __Pyx_InterpreterIdAndModule *end = data->table+data->count;
        for (;lower_bound<end-1; ++lower_bound) {
            lower_bound->id = (lower_bound+1)->id;
            lower_bound->module = (lower_bound+1)->module;
        }
    }
    --data->count;
    if (data->count == 0) {
        free(data);
        data = NULL;
    }
  done:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, data);
#else
    __Pyx_ModuleStateLookup_data = data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return 0;
}
#endif

/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
