/* Generated by Cython 3.2.4 */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
/* InitLimitedAPI */
#if defined(Py_LIMITED_API)
  #if !defined(CYTHON_LIMITED_API)
  #define CYTHON_LIMITED_API 1
  #endif
#elif defined(CYTHON_LIMITED_API)
  #ifdef _MSC_VER
  #pragma message ("Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.")
  #else
  #warning Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x03080000
    #error Cython requires Python 3.8+.
#else
#define __PYX_ABI_VERSION "3_2_4"
#define CYTHON_HEX_VERSION 0x030204F0
#define CYTHON_FUTURE_DIVISION 1
/* CModulePreamble */
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 1
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 1
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PYPY_VERSION_NUM >= 0x07030C00)
  #endif
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_NUM >= 0x07031100)
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND (__PYX_LIMITED_VERSION_HEX >= 0x030A0000)
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #ifndef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 1
  #endif
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #ifdef Py_GIL_DISABLED
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 1
  #else
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #endif
  #if PY_VERSION_HEX < 0x030A0000
    #undef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #elif !defined(CYTHON_USE_TYPE_SLOTS)
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLIST_INTERNALS)
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #elif !defined(CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL 0
  #elif !defined(CYTHON_FAST_GIL)
    #define CYTHON_FAST_GIL (PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_SYS_MONITORING
    #define CYTHON_USE_SYS_MONITORING (PY_VERSION_HEX >= 0x030d00B1)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5 && !CYTHON_USE_MODULE_STATE)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS (!CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
  #if defined(CYTHON_IMMORTAL_CONSTANTS) && PY_VERSION_HEX < 0x030C0000
    #undef CYTHON_IMMORTAL_CONSTANTS
    #define CYTHON_IMMORTAL_CONSTANTS 0  // definitely won't work
  #elif !defined(CYTHON_IMMORTAL_CONSTANTS)
    #define CYTHON_IMMORTAL_CONSTANTS (PY_VERSION_HEX >= 0x030C0000 && !CYTHON_USE_MODULE_STATE && CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
#endif
#ifndef CYTHON_COMPRESS_STRINGS
  #define CYTHON_COMPRESS_STRINGS 1
#endif
#ifndef CYTHON_FAST_PYCCALL
#define CYTHON_FAST_PYCCALL  CYTHON_FAST_PYCALL
#endif
#ifndef CYTHON_VECTORCALL
#if CYTHON_COMPILING_IN_LIMITED_API
#define CYTHON_VECTORCALL  (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
#else
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL)
#endif
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#include <stdint.h>
typedef uintptr_t  __pyx_uintptr_t;
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifndef Py_UNREACHABLE
  #define Py_UNREACHABLE()  assert(0); abort()
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

/* CInitCode */
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

/* PythonCompatibility */
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#define __Pyx_BUILTIN_MODULE_NAME "builtins"
#define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    #ifndef CO_OPTIMIZED
    static int CO_OPTIMIZED;
    #endif
    #ifndef CO_NEWLOCALS
    static int CO_NEWLOCALS;
    #endif
    #ifndef CO_VARARGS
    static int CO_VARARGS;
    #endif
    #ifndef CO_VARKEYWORDS
    static int CO_VARKEYWORDS;
    #endif
    #ifndef CO_ASYNC_GENERATOR
    static int CO_ASYNC_GENERATOR;
    #endif
    #ifndef CO_GENERATOR
    static int CO_GENERATOR;
    #endif
    #ifndef CO_COROUTINE
    static int CO_COROUTINE;
    #endif
#else
    #ifndef CO_COROUTINE
      #define CO_COROUTINE 0x80
    #endif
    #ifndef CO_ASYNC_GENERATOR
      #define CO_ASYNC_GENERATOR 0x200
    #endif
#endif
static int __Pyx_init_co_variables(void);
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef Py_TPFLAGS_IMMUTABLETYPE
  #define Py_TPFLAGS_IMMUTABLETYPE (1UL << 8)
#endif
#ifndef Py_TPFLAGS_DISALLOW_INSTANTIATION
  #define Py_TPFLAGS_DISALLOW_INSTANTIATION (1UL << 7)
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#ifndef METH_FASTCALL
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void (*cfunc)(void)) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if PY_VERSION_HEX < 0x03090000 || (CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000)
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#elif CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) GraalPyFrame_SetLineNumber((frame), (lineno))
#elif CYTHON_COMPILING_IN_GRAAL
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) _PyFrame_SetLineNumber((frame), (lineno))
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#endif
#if CYTHON_USE_MODULE_STATE
static CYTHON_INLINE void *__Pyx__PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#define __Pyx_PyModule_GetState(o) (__pyx_mstatetype *)__Pyx__PyModule_GetState(o)
#else
#define __Pyx_PyModule_GetState(op) ((void)op,__pyx_mstate_global)
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE((PyObject *) obj), name, func_ctype)
#define __Pyx_PyObject_TryGetSlot(obj, name, func_ctype) __Pyx_PyType_TryGetSlot(Py_TYPE(obj), name, func_ctype)
#define __Pyx_PyObject_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#define __Pyx_PyObject_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype) __Pyx_PyType_GetSlot(type, name, func_ctype)
  #define __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype) (((type)->sub) ? ((type)->sub->name) : NULL)
  #define __Pyx_PyType_TryGetSubSlot(type, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype)
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype)\
    ((__PYX_LIMITED_VERSION_HEX >= 0x030A0000 ||\
     (PyType_GetFlags(type) & Py_TPFLAGS_HEAPTYPE) || __Pyx_get_runtime_version() >= 0x030A0000) ?\
     __Pyx_PyType_GetSlot(type, name, func_ctype) : NULL)
  #define __Pyx_PyType_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSlot(obj, name, func_ctype)
  #define __Pyx_PyType_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSlot(obj, name, func_ctype)
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
#define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
#endif
#define __Pyx_PyObject_GetIterNextFunc(iterator)  __Pyx_PyObject_GetSlot(iterator, tp_iternext, iternextfunc)
#if CYTHON_USE_TYPE_SPECS
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#else
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains)
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && PyUnstable_Object_IsUniquelyReferenced(obj)) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#elif CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
enum __Pyx_ReferenceSharing {
  __Pyx_ReferenceSharing_DefinitelyUnique, // We created it so we know it's unshared - no need to check
  __Pyx_ReferenceSharing_OwnStrongReference,
  __Pyx_ReferenceSharing_FunctionArgument,
  __Pyx_ReferenceSharing_SharedReference, // Never trust it to be unshared because it's a global or similar
};
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && PY_VERSION_HEX >= 0x030E0000
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing)\
    (sharing == __Pyx_ReferenceSharing_DefinitelyUnique ? 1 :\
      (sharing == __Pyx_ReferenceSharing_FunctionArgument ? PyUnstable_Object_IsUniqueReferencedTemporary(o) :\
      (sharing == __Pyx_ReferenceSharing_OwnStrongReference ? PyUnstable_Object_IsUniquelyReferenced(o) : 0)))
#elif (CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING) || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)sharing), Py_REFCNT(o) == 1)
#else
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)o), ((void)sharing), 0)
#endif
#if CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyList_GetItemRef(o, i) (likely((i) >= 0) ? PySequence_GetItem(o, i) : (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
  #else
    #define __Pyx_PyList_GetItemRef(o, i) PySequence_ITEM(o, i)
  #endif
#elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #else
    #define __Pyx_PyList_GetItemRef(o, i) __Pyx_XNewRef(PyList_GetItem(o, i))
  #endif
#else
  #define __Pyx_PyList_GetItemRef(o, i) __Pyx_NewRef(PyList_GET_ITEM(o, i))
#endif
#if CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS && !CYTHON_COMPILING_IN_LIMITED_API && CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) (__Pyx_IS_UNIQUELY_REFERENCED(o, unsafe_shared) ?\
    __Pyx_NewRef(PyList_GET_ITEM(o, i)) : __Pyx_PyList_GetItemRef(o, i))
#else
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) __Pyx_PyList_GetItemRef(o, i)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyDict_GetItemRef(dict, key, result) PyDict_GetItemRef(dict, key, result)
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyObject_GetItem(dict, key);
  if (*result == NULL) {
    if (PyErr_ExceptionMatches(PyExc_KeyError)) {
      PyErr_Clear();
      return 0;
    }
    return -1;
  }
  return 1;
}
#else
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyDict_GetItemWithError(dict, key);
  if (*result == NULL) {
    return PyErr_Occurred() ? -1 : 0;
  }
  Py_INCREF(*result);
  return 1;
}
#endif
#if defined(CYTHON_DEBUG_VISIT_CONST) && CYTHON_DEBUG_VISIT_CONST
  #define __Pyx_VISIT_CONST(obj)  Py_VISIT(obj)
#else
  #define __Pyx_VISIT_CONST(obj)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GET_ITEM(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GET_ITEM(o, i)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GetItem(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GetItem(o, i)
#endif
#if CYTHON_ASSUME_SAFE_SIZE
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GET_LENGTH(o)
#else
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GetLength(o)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_InternFromString)
  #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
#endif
#define __Pyx_PyLong_FromHash_t PyLong_FromSsize_t
#define __Pyx_PyLong_AsHash_t   __Pyx_PyIndex_AsSsize_t
#if __PYX_LIMITED_VERSION_HEX >= 0x030A0000
    #define __Pyx_PySendResult PySendResult
#else
    typedef enum {
        PYGEN_RETURN = 0,
        PYGEN_ERROR = -1,
        PYGEN_NEXT = 1,
    } __Pyx_PySendResult;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030A00A3
  typedef __Pyx_PySendResult (*__Pyx_pyiter_sendfunc)(PyObject *iter, PyObject *value, PyObject **result);
#else
  #define __Pyx_pyiter_sendfunc sendfunc
#endif
#if !CYTHON_USE_AM_SEND
#define __PYX_HAS_PY_AM_SEND 0
#elif __PYX_LIMITED_VERSION_HEX >= 0x030A0000
#define __PYX_HAS_PY_AM_SEND 1
#else
#define __PYX_HAS_PY_AM_SEND 2  // our own backported implementation
#endif
#if __PYX_HAS_PY_AM_SEND < 2
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
        __Pyx_pyiter_sendfunc am_send;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_SlotTpAsAsync(s) ((PyAsyncMethods*)(s))
#endif
#if CYTHON_USE_AM_SEND && PY_VERSION_HEX < 0x030A00F0
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (1UL << 21)
#else
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (0)
#endif
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyInterpreterState_Get() PyInterpreterState_Get()
#else
#define __Pyx_PyInterpreterState_Get() PyThreadState_Get()->interp
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030A0000
#ifdef __cplusplus
extern "C"
#endif
PyAPI_FUNC(void *) PyMem_Calloc(size_t nelem, size_t elsize);
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_init_co_variable(PyObject *inspect, const char* name, int *write_to) {
    int value;
    PyObject *py_value = PyObject_GetAttrString(inspect, name);
    if (!py_value) return 0;
    value = (int) PyLong_AsLong(py_value);
    Py_DECREF(py_value);
    *write_to = value;
    return value != -1 || !PyErr_Occurred();
}
static int __Pyx_init_co_variables(void) {
    PyObject *inspect;
    int result;
    inspect = PyImport_ImportModule("inspect");
    result =
#if !defined(CO_OPTIMIZED)
        __Pyx_init_co_variable(inspect, "CO_OPTIMIZED", &CO_OPTIMIZED) &&
#endif
#if !defined(CO_NEWLOCALS)
        __Pyx_init_co_variable(inspect, "CO_NEWLOCALS", &CO_NEWLOCALS) &&
#endif
#if !defined(CO_VARARGS)
        __Pyx_init_co_variable(inspect, "CO_VARARGS", &CO_VARARGS) &&
#endif
#if !defined(CO_VARKEYWORDS)
        __Pyx_init_co_variable(inspect, "CO_VARKEYWORDS", &CO_VARKEYWORDS) &&
#endif
#if !defined(CO_ASYNC_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_ASYNC_GENERATOR", &CO_ASYNC_GENERATOR) &&
#endif
#if !defined(CO_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_GENERATOR", &CO_GENERATOR) &&
#endif
#if !defined(CO_COROUTINE)
        __Pyx_init_co_variable(inspect, "CO_COROUTINE", &CO_COROUTINE) &&
#endif
        1;
    Py_DECREF(inspect);
    return result ? 0 : -1;
}
#else
static int __Pyx_init_co_variables(void) {
    return 0;  // It's a limited API-only feature
}
#endif

/* MathInitCode */
#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #ifndef _USE_MATH_DEFINES
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#ifndef CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#define CYTHON_CLINE_IN_TRACEBACK_RUNTIME 0
#endif
#ifndef CYTHON_CLINE_IN_TRACEBACK
#define CYTHON_CLINE_IN_TRACEBACK CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#endif
#if CYTHON_CLINE_IN_TRACEBACK
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; __pyx_clineno = __LINE__; (void) __pyx_clineno; }
#else
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; (void) __pyx_clineno; }
#endif
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__pyverbs__cq
#define __PYX_HAVE_API__pyverbs__cq
/* Early includes */
#include <stdint.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/time.h>
#include <infiniband/verbs.h>
#include <infiniband/verbs_api.h>
#include <infiniband/driver.h>
#include <infiniband/tm_types.h>
#include "infiniband/verbs.h"
#include "infiniband/driver.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifdef CYTHON_FREETHREADING_COMPATIBLE
#if CYTHON_FREETHREADING_COMPATIBLE
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_NOT_USED
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AS_STRING(s)
#else
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AsString(s)
#endif
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
static CYTHON_INLINE PyObject *__Pyx_NewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_NewRef)
    return Py_NewRef(obj);
#else
    Py_INCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_XNewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_XNewRef)
    return Py_XNewRef(obj);
#else
    Py_XINCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b);
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __Pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AS_DOUBLE(x)
#else
#define __Pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AsDouble(x)
#endif
#define __Pyx_PyFloat_AsFloat(x) ((float) __Pyx_PyFloat_AsDouble(x))
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#elif __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeASCII(c_str, size, NULL)
#else
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
/* PretendToInitialize */
#ifdef __cplusplus
#if __cplusplus > 201103L
#include <type_traits>
#endif
template <typename T>
static void __Pyx_pretend_to_initialize(T* ptr) {
#if __cplusplus > 201103L
    if ((std::is_trivially_default_constructible<T>::value))
#endif
        *ptr = T();
    (void)ptr;
}
#else
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }
#endif


#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * const __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char* const __pyx_f[] = {
  "pyverbs/cq.pyx",
  "EnumTypeToPy",
  "pyverbs/device.pxd",
  "pyverbs/pd.pxd",
  "pyverbs/srq.pxd",
  "pyverbs/qp.pxd",
  "pyverbs/wq.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* Atomics.proto (used by UnpackUnboundCMethod) */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __PYX_GET_CYTHON_COMPILING_IN_CPYTHON_FREETHREADING() CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __pyx_atomic_int_type int
#define __pyx_nonatomic_int_type int
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__))
    #include <stdatomic.h>
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)))
    #include <atomic>
#endif
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__) &&\
                       ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type atomic_int
    #define __pyx_atomic_ptr_type atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) atomic_fetch_add_explicit(value, 1, memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) atomic_fetch_add_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) atomic_fetch_sub_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) atomic_load(value)
    #define __pyx_atomic_store(value, new_value) atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) atomic_load_explicit(value, memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) atomic_load_explicit(value, memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C atomics"
    #endif
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)) &&\
                    ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type std::atomic_int
    #define __pyx_atomic_ptr_type std::atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) std::atomic_fetch_sub_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) std::atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) std::atomic_load(value)
    #define __pyx_atomic_store(value, new_value) std::atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) std::atomic_load_explicit(value, std::memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) std::atomic_load_explicit(value, std::memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) std::atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C++ atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C++ atomics"
    #endif
#elif CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_ptr_type void*
    #define __pyx_nonatomic_ptr_type void*
    #define __pyx_atomic_incr_relaxed(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) __sync_fetch_and_sub(value, 1)
    #define __pyx_atomic_sub(value, arg) __sync_fetch_and_sub(value, arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_store(value, new_value) __sync_lock_test_and_set(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_load_acquire(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) __sync_lock_test_and_set(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_nonatomic_ptr_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER)
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #define __pyx_atomic_ptr_type void*
    #undef __pyx_nonatomic_int_type
    #define __pyx_nonatomic_int_type long
    #define __pyx_nonatomic_ptr_type void*
    #pragma intrinsic (_InterlockedExchangeAdd, _InterlockedExchange, _InterlockedCompareExchange, _InterlockedCompareExchangePointer, _InterlockedExchangePointer)
    #define __pyx_atomic_incr_relaxed(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) _InterlockedExchangeAdd(value, -1)
    #define __pyx_atomic_sub(value, arg) _InterlockedExchangeAdd(value, -arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = _InterlockedCompareExchange(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) _InterlockedExchangeAdd(value, 0)
    #define __pyx_atomic_store(value, new_value) _InterlockedExchange(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) *(void * volatile *)value
    #define __pyx_atomic_pointer_load_acquire(value) _InterlockedCompareExchangePointer(value, 0, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) _InterlockedExchangePointer(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_atomic_ptr_type old = _InterlockedCompareExchangePointer(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif

/* CriticalSectionsDefinition.proto (used by CriticalSections) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection void*
#define __Pyx_PyCriticalSection2 void*
#define __Pyx_PyCriticalSection_End(cs)
#define __Pyx_PyCriticalSection2_End(cs)
#else
#define __Pyx_PyCriticalSection PyCriticalSection
#define __Pyx_PyCriticalSection2 PyCriticalSection2
#define __Pyx_PyCriticalSection_End PyCriticalSection_End
#define __Pyx_PyCriticalSection2_End PyCriticalSection2_End
#endif

/* CriticalSections.proto (used by ParseKeywordsImpl) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection_Begin(cs, arg) (void)(cs)
#define __Pyx_PyCriticalSection2_Begin(cs, arg1, arg2) (void)(cs)
#else
#define __Pyx_PyCriticalSection_Begin PyCriticalSection_Begin
#define __Pyx_PyCriticalSection2_Begin PyCriticalSection2_Begin
#endif
#if PY_VERSION_HEX < 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_BEGIN_CRITICAL_SECTION(o) {
#define __Pyx_END_CRITICAL_SECTION() }
#else
#define __Pyx_BEGIN_CRITICAL_SECTION Py_BEGIN_CRITICAL_SECTION
#define __Pyx_END_CRITICAL_SECTION Py_END_CRITICAL_SECTION
#endif

/* IncludeStructmemberH.proto (used by FixUpExtensionType) */
#include <structmember.h>

/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_7pyverbs_4base_PyverbsObject;
struct __pyx_obj_7pyverbs_4base_PyverbsCM;
struct __pyx_obj_7pyverbs_6device_Context;
struct __pyx_obj_7pyverbs_6device_DeviceAttr;
struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput;
struct __pyx_obj_7pyverbs_6device_ODPCaps;
struct __pyx_obj_7pyverbs_6device_RSSCaps;
struct __pyx_obj_7pyverbs_6device_PacketPacingCaps;
struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps;
struct __pyx_obj_7pyverbs_6device_TMCaps;
struct __pyx_obj_7pyverbs_6device_CQModerationCaps;
struct __pyx_obj_7pyverbs_6device_TSOCaps;
struct __pyx_obj_7pyverbs_6device_DeviceAttrEx;
struct __pyx_obj_7pyverbs_6device_AllocDmAttr;
struct __pyx_obj_7pyverbs_6device_DM;
struct __pyx_obj_7pyverbs_6device_PortAttr;
struct __pyx_obj_7pyverbs_6device_GIDEntry;
struct __pyx_obj_7pyverbs_6device_AsyncEvent;
struct __pyx_obj_7pyverbs_6device_FdArr;
struct __pyx_obj_7pyverbs_2pd_PD;
struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr;
struct __pyx_obj_7pyverbs_2pd_ParentDomain;
struct __pyx_obj_7pyverbs_2pd_ParentDomainContext;
struct __pyx_obj_7pyverbs_3srq_SrqAttr;
struct __pyx_obj_7pyverbs_3srq_SrqInitAttr;
struct __pyx_obj_7pyverbs_3srq_SrqInitAttrEx;
struct __pyx_obj_7pyverbs_3srq_OpsWr;
struct __pyx_obj_7pyverbs_3srq_SRQ;
struct __pyx_obj_7pyverbs_2qp_QPCap;
struct __pyx_obj_7pyverbs_2qp_QPInitAttr;
struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx;
struct __pyx_obj_7pyverbs_2qp_QPAttr;
struct __pyx_obj_7pyverbs_2qp_QP;
struct __pyx_obj_7pyverbs_2qp_DataBuffer;
struct __pyx_obj_7pyverbs_2qp_QPEx;
struct __pyx_obj_7pyverbs_2qp_ECE;
struct __pyx_obj_7pyverbs_2wq_WQInitAttr;
struct __pyx_obj_7pyverbs_2wq_WQAttr;
struct __pyx_obj_7pyverbs_2wq_WQ;
struct __pyx_obj_7pyverbs_2wq_RwqIndTableInitAttr;
struct __pyx_obj_7pyverbs_2wq_RwqIndTable;
struct __pyx_obj_7pyverbs_2wq_RxHashConf;
struct __pyx_obj_7pyverbs_2cq_CompChannel;
struct __pyx_obj_7pyverbs_2cq_CQ;
struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx;
struct __pyx_obj_7pyverbs_2cq_CQEX;
struct __pyx_obj_7pyverbs_2cq_WC;
struct __pyx_obj_7pyverbs_2cq_PollCqAttr;
struct __pyx_obj_7pyverbs_2cq_WcTmInfo;

/* "pyverbs/base.pxd":6
 * #cython: language_level=3
 * 
 * cdef class PyverbsObject(object):             # <<<<<<<<<<<<<<
 *     cdef object __weakref__
 *     cdef object logger
*/
struct __pyx_obj_7pyverbs_4base_PyverbsObject {
  PyObject_HEAD
  PyObject *__weakref__;
  PyObject *logger;
};


/* "pyverbs/base.pxd":10
 *     cdef object logger
 * 
 * cdef class PyverbsCM(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cpdef close(self)
 * 
*/
struct __pyx_obj_7pyverbs_4base_PyverbsCM {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM *__pyx_vtab;
};


/* "pyverbs/device.pxd":10
 * 
 * 
 * cdef class Context(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_context *context
 *     cdef v.ibv_device *device
*/
struct __pyx_obj_7pyverbs_6device_Context {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_context *context;
  struct ibv_device *device;
  PyObject *name;
  PyObject *pds;
  PyObject *dms;
  PyObject *ccs;
  PyObject *cqs;
  PyObject *qps;
  PyObject *xrcds;
  PyObject *vars;
  PyObject *uars;
  PyObject *pps;
  PyObject *sched_nodes;
  PyObject *sched_leafs;
  PyObject *dr_domains;
  PyObject *wqs;
  PyObject *rwq_ind_tbls;
  PyObject *crypto_logins;
  PyObject *event_channels;
  PyObject *dmahs;
};


/* "pyverbs/device.pxd":33
 *     cdef object dmahs
 * 
 * cdef class DeviceAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_device_attr dev_attr
 * 
*/
struct __pyx_obj_7pyverbs_6device_DeviceAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_device_attr dev_attr;
};


/* "pyverbs/device.pxd":36
 *     cdef v.ibv_device_attr dev_attr
 * 
 * cdef class QueryDeviceExInput(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_query_device_ex_input input
 * 
*/
struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_query_device_ex_input input;
};


/* "pyverbs/device.pxd":39
 *     cdef v.ibv_query_device_ex_input input
 * 
 * cdef class ODPCaps(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_odp_caps odp_caps
 *     cdef object xrc_odp_caps
*/
struct __pyx_obj_7pyverbs_6device_ODPCaps {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_odp_caps odp_caps;
  PyObject *xrc_odp_caps;
};


/* "pyverbs/device.pxd":43
 *     cdef object xrc_odp_caps
 * 
 * cdef class RSSCaps(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_rss_caps rss_caps
 * 
*/
struct __pyx_obj_7pyverbs_6device_RSSCaps {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_rss_caps rss_caps;
};


/* "pyverbs/device.pxd":46
 *     cdef v.ibv_rss_caps rss_caps
 * 
 * cdef class PacketPacingCaps(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_packet_pacing_caps packet_pacing_caps
 * 
*/
struct __pyx_obj_7pyverbs_6device_PacketPacingCaps {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_packet_pacing_caps packet_pacing_caps;
};


/* "pyverbs/device.pxd":49
 *     cdef v.ibv_packet_pacing_caps packet_pacing_caps
 * 
 * cdef class PCIAtomicCaps(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_pci_atomic_caps caps
 * 
*/
struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_pci_atomic_caps caps;
};


/* "pyverbs/device.pxd":52
 *     cdef v.ibv_pci_atomic_caps caps
 * 
 * cdef class TMCaps(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_tm_caps tm_caps
 * 
*/
struct __pyx_obj_7pyverbs_6device_TMCaps {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_tm_caps tm_caps;
};


/* "pyverbs/device.pxd":55
 *     cdef v.ibv_tm_caps tm_caps
 * 
 * cdef class CQModerationCaps(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_cq_moderation_caps cq_mod_caps
 * 
*/
struct __pyx_obj_7pyverbs_6device_CQModerationCaps {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_cq_moderation_caps cq_mod_caps;
};


/* "pyverbs/device.pxd":58
 *     cdef v.ibv_cq_moderation_caps cq_mod_caps
 * 
 * cdef class TSOCaps(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_tso_caps tso_caps
 * 
*/
struct __pyx_obj_7pyverbs_6device_TSOCaps {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_tso_caps tso_caps;
};


/* "pyverbs/device.pxd":61
 *     cdef v.ibv_tso_caps tso_caps
 * 
 * cdef class DeviceAttrEx(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_device_attr_ex dev_attr
 * 
*/
struct __pyx_obj_7pyverbs_6device_DeviceAttrEx {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_device_attr_ex dev_attr;
};


/* "pyverbs/device.pxd":64
 *     cdef v.ibv_device_attr_ex dev_attr
 * 
 * cdef class AllocDmAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_alloc_dm_attr alloc_dm_attr
 * 
*/
struct __pyx_obj_7pyverbs_6device_AllocDmAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_alloc_dm_attr alloc_dm_attr;
};


/* "pyverbs/device.pxd":67
 *     cdef v.ibv_alloc_dm_attr alloc_dm_attr
 * 
 * cdef class DM(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_dm *dm
 *     cdef object dm_mrs
*/
struct __pyx_obj_7pyverbs_6device_DM {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_dm *dm;
  PyObject *dm_mrs;
  PyObject *context;
  PyObject *_is_imported;
};


/* "pyverbs/device.pxd":74
 *     cdef add_ref(self, obj)
 * 
 * cdef class PortAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_port_attr attr
 * 
*/
struct __pyx_obj_7pyverbs_6device_PortAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_port_attr attr;
};


/* "pyverbs/device.pxd":77
 *     cdef v.ibv_port_attr attr
 * 
 * cdef class GIDEntry(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_gid_entry entry
 * 
*/
struct __pyx_obj_7pyverbs_6device_GIDEntry {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_gid_entry entry;
};


/* "pyverbs/device.pxd":80
 *     cdef v.ibv_gid_entry entry
 * 
 * cdef class AsyncEvent(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_async_event event
 * 
*/
struct __pyx_obj_7pyverbs_6device_AsyncEvent {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_async_event event;
};


/* "pyverbs/device.pxd":83
 *     cdef v.ibv_async_event event
 * 
 * cdef class FdArr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_fd_arr attr
 *     cpdef close(self)
*/
struct __pyx_obj_7pyverbs_6device_FdArr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct __pyx_vtabstruct_7pyverbs_6device_FdArr *__pyx_vtab;
  struct ibv_fd_arr attr;
};


/* "pyverbs/pd.pxd":12
 * 
 * 
 * cdef class PD(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_pd *pd
 *     cdef Context ctx
*/
struct __pyx_obj_7pyverbs_2pd_PD {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_pd *pd;
  struct __pyx_obj_7pyverbs_6device_Context *ctx;
  PyObject *srqs;
  PyObject *mrs;
  PyObject *mws;
  PyObject *ahs;
  PyObject *qps;
  PyObject *parent_domains;
  PyObject *wqs;
  PyObject *mkeys;
  PyObject *deks;
  PyObject *_is_imported;
};


/* "pyverbs/pd.pxd":28
 *     cdef object _is_imported
 * 
 * cdef class ParentDomainInitAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_parent_domain_init_attr init_attr
 *     cdef object pd
*/
struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_parent_domain_init_attr init_attr;
  PyObject *pd;
  PyObject *alloc;
  PyObject *dealloc;
};


/* "pyverbs/pd.pxd":34
 *     cdef object dealloc
 * 
 * cdef class ParentDomain(PD):             # <<<<<<<<<<<<<<
 *     cdef add_ref(self, obj)
 *     cdef object protection_domain
*/
struct __pyx_obj_7pyverbs_2pd_ParentDomain {
  struct __pyx_obj_7pyverbs_2pd_PD __pyx_base;
  PyObject *protection_domain;
  PyObject *cqs;
};


/* "pyverbs/pd.pxd":39
 *     cdef object cqs
 * 
 * cdef class ParentDomainContext(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef object p_alloc
 *     cdef object p_free
*/
struct __pyx_obj_7pyverbs_2pd_ParentDomainContext {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  PyObject *p_alloc;
  PyObject *p_free;
  PyObject *pd;
  PyObject *user_data;
};


/* "pyverbs/srq.pxd":9
 * from . cimport libibverbs as v
 * 
 * cdef class SrqAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_srq_attr attr
 * 
*/
struct __pyx_obj_7pyverbs_3srq_SrqAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_srq_attr attr;
};


/* "pyverbs/srq.pxd":12
 *     cdef v.ibv_srq_attr attr
 * 
 * cdef class SrqInitAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_srq_init_attr    attr
 * 
*/
struct __pyx_obj_7pyverbs_3srq_SrqInitAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_srq_init_attr attr;
};


/* "pyverbs/srq.pxd":15
 *     cdef v.ibv_srq_init_attr    attr
 * 
 * cdef class SrqInitAttrEx(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_srq_init_attr_ex attr
 *     cdef object _cq
*/
struct __pyx_obj_7pyverbs_3srq_SrqInitAttrEx {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_srq_init_attr_ex attr;
  PyObject *_cq;
  PyObject *_pd;
  PyObject *_xrcd;
};


/* "pyverbs/srq.pxd":21
 *     cdef object _xrcd
 * 
 * cdef class OpsWr(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_ops_wr ops_wr
 * 
*/
struct __pyx_obj_7pyverbs_3srq_OpsWr {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_ops_wr ops_wr;
};


/* "pyverbs/srq.pxd":24
 *     cdef v.ibv_ops_wr ops_wr
 * 
 * cdef class SRQ(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_srq *srq
 *     cdef object cq
*/
struct __pyx_obj_7pyverbs_3srq_SRQ {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_srq *srq;
  PyObject *cq;
  PyObject *qps;
};


/* "pyverbs/qp.pxd":9
 * cimport pyverbs.libibverbs as v
 * 
 * cdef class QPCap(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_qp_cap cap
 * 
*/
struct __pyx_obj_7pyverbs_2qp_QPCap {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_qp_cap cap;
};


/* "pyverbs/qp.pxd":12
 *     cdef v.ibv_qp_cap cap
 * 
 * cdef class QPInitAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_qp_init_attr attr
 *     cdef object scq
*/
struct __pyx_obj_7pyverbs_2qp_QPInitAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_qp_init_attr attr;
  PyObject *scq;
  PyObject *rcq;
  PyObject *srq;
};


/* "pyverbs/qp.pxd":18
 *     cdef object srq
 * 
 * cdef class QPInitAttrEx(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_qp_init_attr_ex attr
 *     cdef object scq
*/
struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_qp_init_attr_ex attr;
  PyObject *scq;
  PyObject *rcq;
  PyObject *_pd;
  PyObject *xrcd;
  PyObject *srq;
  PyObject *ind_table;
};


/* "pyverbs/qp.pxd":27
 *     cdef object ind_table
 * 
 * cdef class QPAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_qp_attr attr
 * 
*/
struct __pyx_obj_7pyverbs_2qp_QPAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_qp_attr attr;
};


/* "pyverbs/qp.pxd":30
 *     cdef v.ibv_qp_attr attr
 * 
 * cdef class QP(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_qp *qp
 *     cdef int type
*/
struct __pyx_obj_7pyverbs_2qp_QP {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_qp *qp;
  int type;
  int state;
  PyObject *pd;
  PyObject *context;
  PyObject *xrcd;
  PyObject *scq;
  PyObject *rcq;
  PyObject *mws;
  PyObject *srq;
  PyObject *flows;
  PyObject *dr_actions;
};


/* "pyverbs/qp.pxd":47
 *     cdef add_ref(self, obj)
 * 
 * cdef class DataBuffer(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_data_buf data
 * 
*/
struct __pyx_obj_7pyverbs_2qp_DataBuffer {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_data_buf data;
};


/* "pyverbs/qp.pxd":50
 *     cdef v.ibv_data_buf data
 * 
 * cdef class QPEx(QP):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_qp_ex *qp_ex
 *     cdef object ind_table
*/
struct __pyx_obj_7pyverbs_2qp_QPEx {
  struct __pyx_obj_7pyverbs_2qp_QP __pyx_base;
  struct ibv_qp_ex *qp_ex;
  PyObject *ind_table;
};


/* "pyverbs/qp.pxd":54
 *     cdef object ind_table
 * 
 * cdef class ECE(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_ece ece
*/
struct __pyx_obj_7pyverbs_2qp_ECE {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_ece ece;
};


/* "pyverbs/wq.pxd":13
 * 
 * 
 * cdef class WQInitAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_wq_init_attr attr
 *     cdef PD pd
*/
struct __pyx_obj_7pyverbs_2wq_WQInitAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_wq_init_attr attr;
  struct __pyx_obj_7pyverbs_2pd_PD *pd;
  PyObject *cq;
};


/* "pyverbs/wq.pxd":18
 *     cdef object cq
 * 
 * cdef class WQAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_wq_attr attr
 * 
*/
struct __pyx_obj_7pyverbs_2wq_WQAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_wq_attr attr;
};


/* "pyverbs/wq.pxd":21
 *     cdef v.ibv_wq_attr attr
 * 
 * cdef class WQ(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_wq *wq
 *     cdef Context context
*/
struct __pyx_obj_7pyverbs_2wq_WQ {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_wq *wq;
  struct __pyx_obj_7pyverbs_6device_Context *context;
  struct __pyx_obj_7pyverbs_2pd_PD *pd;
  PyObject *cq;
  PyObject *rwq_ind_tables;
};


/* "pyverbs/wq.pxd":29
 *     cpdef add_ref(self, obj)
 * 
 * cdef class RwqIndTableInitAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_rwq_ind_table_init_attr attr
 *     cdef object wqs_list
*/
struct __pyx_obj_7pyverbs_2wq_RwqIndTableInitAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_rwq_ind_table_init_attr attr;
  PyObject *wqs_list;
};


/* "pyverbs/wq.pxd":33
 *     cdef object wqs_list
 * 
 * cdef class RwqIndTable(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_rwq_ind_table *rwq_ind_table
 *     cdef Context context
*/
struct __pyx_obj_7pyverbs_2wq_RwqIndTable {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_rwq_ind_table *rwq_ind_table;
  struct __pyx_obj_7pyverbs_6device_Context *context;
  PyObject *wqs;
  PyObject *qps;
};


/* "pyverbs/wq.pxd":40
 *     cpdef add_ref(self, obj)
 * 
 * cdef class RxHashConf(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_rx_hash_conf rx_hash_conf
*/
struct __pyx_obj_7pyverbs_2wq_RxHashConf {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_rx_hash_conf rx_hash_conf;
};


/* "pyverbs/cq.pxd":9
 * cimport pyverbs.libibverbs as v
 * 
 * cdef class CompChannel(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_comp_channel *cc
 *     cpdef close(self)
*/
struct __pyx_obj_7pyverbs_2cq_CompChannel {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_comp_channel *cc;
  PyObject *context;
  PyObject *cqs;
};


/* "pyverbs/cq.pxd":16
 *     cdef object cqs
 * 
 * cdef class CQ(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_cq *cq
 *     cpdef close(self)
*/
struct __pyx_obj_7pyverbs_2cq_CQ {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_cq *cq;
  PyObject *context;
  PyObject *qps;
  PyObject *srqs;
  PyObject *channel;
  PyObject *num_events;
  PyObject *wqs;
};


/* "pyverbs/cq.pxd":27
 *     cdef object wqs
 * 
 * cdef class CqInitAttrEx(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_cq_init_attr_ex attr
 *     cdef object channel
*/
struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_cq_init_attr_ex attr;
  PyObject *channel;
  PyObject *parent_domain;
};


/* "pyverbs/cq.pxd":32
 *     cdef object parent_domain
 * 
 * cdef class CQEX(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_cq_ex *cq
 *     cdef v.ibv_cq *ibv_cq
*/
struct __pyx_obj_7pyverbs_2cq_CQEX {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_cq_ex *cq;
  struct ibv_cq *ibv_cq;
  PyObject *context;
  PyObject *qps;
  PyObject *srqs;
  PyObject *wqs;
};


/* "pyverbs/cq.pxd":42
 *     cdef object wqs
 * 
 * cdef class WC(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_wc wc
 * 
*/
struct __pyx_obj_7pyverbs_2cq_WC {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_wc wc;
};


/* "pyverbs/cq.pxd":45
 *     cdef v.ibv_wc wc
 * 
 * cdef class PollCqAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_poll_cq_attr attr
 * 
*/
struct __pyx_obj_7pyverbs_2cq_PollCqAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_poll_cq_attr attr;
};


/* "pyverbs/cq.pxd":48
 *     cdef v.ibv_poll_cq_attr attr
 * 
 * cdef class WcTmInfo(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_wc_tm_info info
*/
struct __pyx_obj_7pyverbs_2cq_WcTmInfo {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_wc_tm_info info;
};



/* "pyverbs/base.pxd":10
 *     cdef object logger
 * 
 * cdef class PyverbsCM(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cpdef close(self)
 * 
*/

struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM {
  PyObject *(*close)(struct __pyx_obj_7pyverbs_4base_PyverbsCM *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM *__pyx_vtabptr_7pyverbs_4base_PyverbsCM;


/* "pyverbs/device.pxd":10
 * 
 * 
 * cdef class Context(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_context *context
 *     cdef v.ibv_device *device
*/

struct __pyx_vtabstruct_7pyverbs_6device_Context {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_6device_Context *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_6device_Context *__pyx_vtabptr_7pyverbs_6device_Context;


/* "pyverbs/device.pxd":67
 *     cdef v.ibv_alloc_dm_attr alloc_dm_attr
 * 
 * cdef class DM(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_dm *dm
 *     cdef object dm_mrs
*/

struct __pyx_vtabstruct_7pyverbs_6device_DM {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_6device_DM *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_6device_DM *__pyx_vtabptr_7pyverbs_6device_DM;


/* "pyverbs/device.pxd":83
 *     cdef v.ibv_async_event event
 * 
 * cdef class FdArr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_fd_arr attr
 *     cpdef close(self)
*/

struct __pyx_vtabstruct_7pyverbs_6device_FdArr {
  PyObject *(*close)(struct __pyx_obj_7pyverbs_6device_FdArr *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyverbs_6device_FdArr *__pyx_vtabptr_7pyverbs_6device_FdArr;


/* "pyverbs/pd.pxd":12
 * 
 * 
 * cdef class PD(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_pd *pd
 *     cdef Context ctx
*/

struct __pyx_vtabstruct_7pyverbs_2pd_PD {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_2pd_PD *, PyObject *);
  PyObject *(*remove_ref)(struct __pyx_obj_7pyverbs_2pd_PD *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_2pd_PD *__pyx_vtabptr_7pyverbs_2pd_PD;


/* "pyverbs/pd.pxd":34
 *     cdef object dealloc
 * 
 * cdef class ParentDomain(PD):             # <<<<<<<<<<<<<<
 *     cdef add_ref(self, obj)
 *     cdef object protection_domain
*/

struct __pyx_vtabstruct_7pyverbs_2pd_ParentDomain {
  struct __pyx_vtabstruct_7pyverbs_2pd_PD __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2pd_ParentDomain *__pyx_vtabptr_7pyverbs_2pd_ParentDomain;


/* "pyverbs/srq.pxd":21
 *     cdef object _xrcd
 * 
 * cdef class OpsWr(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_ops_wr ops_wr
 * 
*/

struct __pyx_vtabstruct_7pyverbs_3srq_OpsWr {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_3srq_OpsWr *__pyx_vtabptr_7pyverbs_3srq_OpsWr;


/* "pyverbs/srq.pxd":24
 *     cdef v.ibv_ops_wr ops_wr
 * 
 * cdef class SRQ(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_srq *srq
 *     cdef object cq
*/

struct __pyx_vtabstruct_7pyverbs_3srq_SRQ {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_3srq_SRQ *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_3srq_SRQ *__pyx_vtabptr_7pyverbs_3srq_SRQ;


/* "pyverbs/qp.pxd":30
 *     cdef v.ibv_qp_attr attr
 * 
 * cdef class QP(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_qp *qp
 *     cdef int type
*/

struct __pyx_vtabstruct_7pyverbs_2qp_QP {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*update_cqs)(struct __pyx_obj_7pyverbs_2qp_QP *, PyObject *);
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_2qp_QP *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_2qp_QP *__pyx_vtabptr_7pyverbs_2qp_QP;


/* "pyverbs/qp.pxd":47
 *     cdef add_ref(self, obj)
 * 
 * cdef class DataBuffer(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_data_buf data
 * 
*/

struct __pyx_vtabstruct_7pyverbs_2qp_DataBuffer {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2qp_DataBuffer *__pyx_vtabptr_7pyverbs_2qp_DataBuffer;


/* "pyverbs/qp.pxd":50
 *     cdef v.ibv_data_buf data
 * 
 * cdef class QPEx(QP):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_qp_ex *qp_ex
 *     cdef object ind_table
*/

struct __pyx_vtabstruct_7pyverbs_2qp_QPEx {
  struct __pyx_vtabstruct_7pyverbs_2qp_QP __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2qp_QPEx *__pyx_vtabptr_7pyverbs_2qp_QPEx;


/* "pyverbs/qp.pxd":54
 *     cdef object ind_table
 * 
 * cdef class ECE(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_ece ece
*/

struct __pyx_vtabstruct_7pyverbs_2qp_ECE {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2qp_ECE *__pyx_vtabptr_7pyverbs_2qp_ECE;


/* "pyverbs/wq.pxd":21
 *     cdef v.ibv_wq_attr attr
 * 
 * cdef class WQ(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_wq *wq
 *     cdef Context context
*/

struct __pyx_vtabstruct_7pyverbs_2wq_WQ {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_2wq_WQ *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyverbs_2wq_WQ *__pyx_vtabptr_7pyverbs_2wq_WQ;


/* "pyverbs/wq.pxd":33
 *     cdef object wqs_list
 * 
 * cdef class RwqIndTable(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_rwq_ind_table *rwq_ind_table
 *     cdef Context context
*/

struct __pyx_vtabstruct_7pyverbs_2wq_RwqIndTable {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_2wq_RwqIndTable *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyverbs_2wq_RwqIndTable *__pyx_vtabptr_7pyverbs_2wq_RwqIndTable;


/* "pyverbs/cq.pyx":17
 * 
 * 
 * cdef class CompChannel(PyverbsCM):             # <<<<<<<<<<<<<<
 *     """
 *     A completion channel is a file descriptor used to deliver completion
*/

struct __pyx_vtabstruct_7pyverbs_2cq_CompChannel {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_2cq_CompChannel *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_2cq_CompChannel *__pyx_vtabptr_7pyverbs_2cq_CompChannel;


/* "pyverbs/cq.pyx":78
 * 
 * 
 * cdef class CQ(PyverbsCM):             # <<<<<<<<<<<<<<
 *     """
 *     A Completion Queue is the notification mechanism for work request
*/

struct __pyx_vtabstruct_7pyverbs_2cq_CQ {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_2cq_CQ *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_2cq_CQ *__pyx_vtabptr_7pyverbs_2cq_CQ;


/* "pyverbs/cq.pyx":315
 * 
 * 
 * cdef class CQEX(PyverbsCM):             # <<<<<<<<<<<<<<
 *     def __init__(self, Context context not None, CqInitAttrEx init_attr):
 *         """
*/

struct __pyx_vtabstruct_7pyverbs_2cq_CQEX {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_2cq_CQEX *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_2cq_CQEX *__pyx_vtabptr_7pyverbs_2cq_CQEX;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto (used by PyErrFetchRestore) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto (used by GetBuiltinName) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* HasAttr.proto (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_HasAttr(o, n)  PyObject_HasAttrWithError(o, n)
#else
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);
#endif

/* PyObjectCall.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto (used by PyObjectCallOneArg) */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs);

/* PyObjectCallOneArg.proto (used by ImportImpl) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* TupleAndListFromArray.proto (used by ImportImpl) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
#endif
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto (used by ImportImpl) */
#include <string.h>

/* ImportImpl.export */
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level);

/* Import.proto */
static CYTHON_INLINE PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* GetTopmostException.proto (used by SaveResetException) */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyImportError_Check.proto */
#define __Pyx_PyExc_ImportError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_ImportError)

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectFastCallMethod.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyObject_FastCallMethod(name, args, nargsf) PyObject_VectorcallMethod(name, args, nargsf, NULL)
#else
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf);
#endif

/* BytesEquals.proto (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_PySequence_ITEM(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_NewRef(__Pyx_PyTuple_GET_ITEM(args, i))
#else
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_XNewRef(PyTuple_GetItem(args, i))
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_ArgRef_FASTCALL(args, i) __Pyx_NewRef(args[i])
    #define __Pyx_NumKwargs_FASTCALL(kwds) __Pyx_PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
#else
    #define __Pyx_ArgRef_FASTCALL __Pyx_ArgRef_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
#endif
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#if CYTHON_METH_FASTCALL || (CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(args + start, stop - start)
#else
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* py_dict_items.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d);

/* CallCFunction.proto (used by CallUnboundCMethod0) */
#define __Pyx_CallCFunction(cfunc, self, args)\
    ((PyCFunction)(void(*)(void))(cfunc)->func)(self, args)
#define __Pyx_CallCFunctionWithKeywords(cfunc, self, args, kwargs)\
    ((PyCFunctionWithKeywords)(void(*)(void))(cfunc)->func)(self, args, kwargs)
#define __Pyx_CallCFunctionFast(cfunc, self, args, nargs)\
    ((__Pyx_PyCFunctionFast)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs)
#define __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, nargs, kwnames)\
    ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs, kwnames)

/* UnpackUnboundCMethod.proto (used by CallUnboundCMethod0) */
typedef struct {
    PyObject *type;
    PyObject **method_name;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && CYTHON_ATOMICS
    __pyx_atomic_int_type initialized;
#endif
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
static CYTHON_INLINE int __Pyx_CachedCFunction_GetAndSetInitializing(__Pyx_CachedCFunction *cfunc) {
#if !CYTHON_ATOMICS
    return 1;
#else
    __pyx_nonatomic_int_type expected = 0;
    if (__pyx_atomic_int_cmp_exchange(&cfunc->initialized, &expected, 1)) {
        return 0;
    }
    return expected;
#endif
}
static CYTHON_INLINE void __Pyx_CachedCFunction_SetFinishedInitializing(__Pyx_CachedCFunction *cfunc) {
#if CYTHON_ATOMICS
    __pyx_atomic_store(&cfunc->initialized, 2);
#endif
}
#else
#define __Pyx_CachedCFunction_GetAndSetInitializing(cfunc) 2
#define __Pyx_CachedCFunction_SetFinishedInitializing(cfunc)
#endif

/* CallUnboundCMethod0.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* py_dict_values.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d);

/* OwnedDictNext.proto (used by ParseKeywordsImpl) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue);
#else
CYTHON_INLINE
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue);
#endif

/* RaiseDoubleKeywords.proto (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywordsImpl.export */
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name
);
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* CallUnboundCMethod2.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* ParseKeywords.proto */
static CYTHON_INLINE int __Pyx_ParseKeywords(
    PyObject *kwds, PyObject *const *kwvalues, PyObject ** const argnames[],
    PyObject *kwds2, PyObject *values[],
    Py_ssize_t num_pos_args, Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* ArgTypeTestFunc.export */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))

/* PyDictVersioning.proto (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __Pyx_XNewRef(__pyx_dict_cached_value);\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_mstate_global->__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* RaiseException.export */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* RejectKeywords.export */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds);

/* PyLongBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static CYTHON_INLINE PyObject* __Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyLong_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyTypeError_Check.proto */
#define __Pyx_PyExc_TypeError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_TypeError)

/* PyLongCompare.proto */
static CYTHON_INLINE int __Pyx_PyLong_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectVectorCallKwBuilder.proto */
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#if CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_Vectorcall_CallFromBuilder PyObject_Vectorcall
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder _PyObject_Vectorcall
#endif
#define __Pyx_MakeVectorcallBuilderKwds(n) PyTuple_New(n)
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder __Pyx_PyObject_FastCallDict
#define __Pyx_MakeVectorcallBuilderKwds(n) __Pyx_PyDict_NewPresized(n)
#define __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n) PyDict_SetItem(builder, key, value)
#define __Pyx_VectorcallBuilder_AddArgStr(key, value, builder, args, n) PyDict_SetItemString(builder, key, value)
#endif

/* PyObjectVectorCallMethodKwBuilder.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_VectorcallMethod_CallFromBuilder PyObject_VectorcallMethod
#else
static PyObject *__Pyx_Object_VectorcallMethod_CallFromBuilder(PyObject *name, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#endif

/* UnicodeConcatInPlace.proto */
# if CYTHON_COMPILING_IN_CPYTHON
    #if CYTHON_REFNANNY
        #define __Pyx_PyUnicode_ConcatInPlace(left, right, unsafe_shared) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right, unsafe_shared, __pyx_refnanny)
    #else
        #define __Pyx_PyUnicode_ConcatInPlace(left, right, unsafe_shared) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right, unsafe_shared)
    #endif
    #define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_DefinitelyUniqueInPlace(left, right) __Pyx_PyUnicode_ConcatInPlace(left, right, __Pyx_ReferenceSharing_DefinitelyUnique)
    #define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(left, right) __Pyx_PyUnicode_ConcatInPlace(left, right, __Pyx_ReferenceSharing_OwnStrongReference)
    #define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_FunctionArgumentInPlace(left, right) __Pyx_PyUnicode_ConcatInPlace(left, right, __Pyx_ReferenceSharing_DefinitelyUnique)
    #define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_SharedReferenceInPlace(left, right) __Pyx_PyUnicode_ConcatInPlace(left, right, __Pyx_ReferenceSharing_SharedReference)
    static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right, int unsafe_shared
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    );
#else
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_DefinitelyUniqueInPlace __Pyx_PyUnicode_Concat
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace __Pyx_PyUnicode_Concat
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_FunctionArgumentInPlace __Pyx_PyUnicode_Concat
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_SharedReferenceInPlace __Pyx_PyUnicode_Concat
#endif
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_DefinitelyUniqueInPlaceSafe(left, right)\
    ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_DefinitelyUniqueInPlace(left, right))
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlaceSafe(left, right)\
    ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(left, right))
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_FunctionArgumentInPlaceSafe(left, right)\
    ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_FunctionArgumentInPlace(left, right))
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_SharedReferenceInPlaceSafe(left, right)\
    ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_SharedReferenceInPlace(left, right))

/* DictGetItem.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PyKeyError_Check.proto */
#define __Pyx_PyExc_KeyError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_KeyError)

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck, unsafe_shared) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck, int unsafe_shared);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* CallNextTpDealloc.proto */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallTypeTraverse.proto */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#define __Pyx_call_type_traverse(o, always_call, visit, arg) 0
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg);
#endif

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_2_4
#define __PYX_HAVE_RT_ImportType_proto_3_2_4
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_2_4 {
   __Pyx_ImportType_CheckSize_Error_3_2_4 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_2_4 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_2_4 = 2
};
static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* LimitedApiGetTypeDict.proto (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp);
#endif

/* SetItemOnTypeDict.proto (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v);
#define __Pyx_SetItemOnTypeDict(tp, k, v) __Pyx__SetItemOnTypeDict((PyTypeObject*)tp, k, v)

/* FixUpExtensionType.proto */
static CYTHON_INLINE int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);

/* PyObjectCallNoArg.proto (used by PyObjectCallMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectGetMethod.proto (used by PyObjectCallMethod0) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);
#endif

/* PyObjectCallMethod0.proto (used by PyType_Ready) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* ValidateBasesTuple.proto (used by PyType_Ready) */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* MergeVTables.proto */
static int __Pyx_MergeVtables(PyTypeObject *type);

/* DelItemOnTypeDict.proto (used by SetupReduce) */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k);
#define __Pyx_DelItemOnTypeDict(tp, k) __Pyx__DelItemOnTypeDict((PyTypeObject*)tp, k)

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* FunctionImport.proto */
static int __Pyx_ImportFunction_3_2_4(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* dict_setdefault.proto (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value);

/* AddModuleRef.proto (used by FetchSharedCythonModule) */
#if ((CYTHON_COMPILING_IN_CPYTHON_FREETHREADING ) ||\
     __PYX_LIMITED_VERSION_HEX < 0x030d0000)
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name);
#else
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#endif

/* FetchSharedCythonModule.proto (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto (used by CommonTypesMetaclass) */
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases);

/* CommonTypesMetaclass.proto (used by CythonFunctionShared) */
static int __pyx_CommonTypesMetaclass_init(PyObject *module);
#define __Pyx_CommonTypesMetaclass_USED

/* PyMethodNew.proto (used by CythonFunctionShared) */
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ);

/* PyVectorcallFastCallDict.proto (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto (used by CythonFunction) */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && CYTHON_METH_FASTCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_dict;
#endif
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    PyObject *defaults;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_mstate_global->__pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_mstate_global->__pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_mstate_global->__pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void));
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func,
                                                         PyTypeObject *defaults_type);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* CLineInTraceback.proto (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#else
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#endif

/* CodeObjectCache.proto (used by AddTraceback) */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject __Pyx_CachedCodeObjectType;
#else
typedef PyCodeObject __Pyx_CachedCodeObjectType;
#endif
typedef struct {
    __Pyx_CachedCodeObjectType* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_int_type accessor_count;
  #endif
};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_int(unsigned int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_long(unsigned long value);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned int __Pyx_PyLong_As_unsigned_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned long __Pyx_PyLong_As_unsigned_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_char(unsigned char value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_short(unsigned short value);

/* CIntFromPy.proto */
static CYTHON_INLINE enum ibv_wc_status __Pyx_PyLong_As_enum__ibv_wc_status(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE enum ibv_wc_opcode __Pyx_PyLong_As_enum__ibv_wc_opcode(PyObject *);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#else
static __Pyx_TypeName __Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp);
#endif
#else  // !LIMITED_API
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetFullyQualifiedName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2) {
    return PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2);
}
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#ifdef PyExceptionInstance_Check
  #define __Pyx_PyBaseException_Check(obj) PyExceptionInstance_Check(obj)
#else
  #define __Pyx_PyBaseException_Check(obj) __Pyx_TypeCheck(obj, PyExc_BaseException)
#endif

/* GetRuntimeVersion.proto */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
static unsigned long __Pyx_cached_runtime_version = 0;
static void __Pyx_init_runtime_version(void);
#else
#define __Pyx_init_runtime_version()
#endif
static unsigned long __Pyx_get_runtime_version(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* DecompressString.proto */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo);

/* MultiPhaseInitModuleState.proto */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
static PyObject *__Pyx_State_FindModule(void*);
static int __Pyx_State_AddModule(PyObject* module, void*);
static int __Pyx_State_RemoveModule(void*);
#elif CYTHON_USE_MODULE_STATE
#define __Pyx_State_FindModule PyState_FindModule
#define __Pyx_State_AddModule PyState_AddModule
#define __Pyx_State_RemoveModule PyState_RemoveModule
#endif

/* #### Code section: module_declarations ### */
/* CythonABIVersion.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
    #if CYTHON_METH_FASTCALL
        #define __PYX_FASTCALL_ABI_SUFFIX  "_fastcall"
    #else
        #define __PYX_FASTCALL_ABI_SUFFIX
    #endif
    #define __PYX_LIMITED_ABI_SUFFIX "limited" __PYX_FASTCALL_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#else
    #define __PYX_LIMITED_ABI_SUFFIX
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    #define __PYX_AM_SEND_ABI_SUFFIX
#elif __PYX_HAS_PY_AM_SEND == 2
    #define __PYX_AM_SEND_ABI_SUFFIX "amsendbackport"
#else
    #define __PYX_AM_SEND_ABI_SUFFIX "noamsend"
#endif
#ifndef __PYX_MONITORING_ABI_SUFFIX
    #define __PYX_MONITORING_ABI_SUFFIX
#endif
#if CYTHON_USE_TP_FINALIZE
    #define __PYX_TP_FINALIZE_ABI_SUFFIX
#else
    #define __PYX_TP_FINALIZE_ABI_SUFFIX "nofinalize"
#endif
#if CYTHON_USE_FREELISTS || !defined(__Pyx_AsyncGen_USED)
    #define __PYX_FREELISTS_ABI_SUFFIX
#else
    #define __PYX_FREELISTS_ABI_SUFFIX "nofreelists"
#endif
#define CYTHON_ABI  __PYX_ABI_VERSION __PYX_LIMITED_ABI_SUFFIX __PYX_MONITORING_ABI_SUFFIX __PYX_TP_FINALIZE_ABI_SUFFIX __PYX_FREELISTS_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."

static PyObject *__pyx_f_7pyverbs_2cq_11CompChannel_close(struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_7pyverbs_2cq_11CompChannel_add_ref(struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_self, PyObject *__pyx_v_obj); /* proto*/
static PyObject *__pyx_f_7pyverbs_2cq_2CQ_add_ref(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self, PyObject *__pyx_v_obj); /* proto*/
static PyObject *__pyx_f_7pyverbs_2cq_2CQ_close(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_7pyverbs_2cq_4CQEX_add_ref(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self, PyObject *__pyx_v_obj); /* proto*/
static PyObject *__pyx_f_7pyverbs_2cq_4CQEX_close(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/

/* Module declarations from "pyverbs.base" */
static PyObject *(*__pyx_f_7pyverbs_4base_close_weakrefs)(PyObject *); /*proto*/

/* Module declarations from "libc.stdint" */

/* Module declarations from "posix.types" */

/* Module declarations from "posix.signal" */

/* Module declarations from "posix.time" */

/* Module declarations from "pyverbs.libibverbs_enums" */

/* Module declarations from "pyverbs.libibverbs" */

/* Module declarations from "pyverbs.device" */

/* Module declarations from "pyverbs.pd" */

/* Module declarations from "pyverbs" */

/* Module declarations from "pyverbs.srq" */

/* Module declarations from "pyverbs.qp" */

/* Module declarations from "pyverbs.wq" */

/* Module declarations from "pyverbs.cq" */
static PyObject *__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py = 0;
static PyObject *__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py = 0;
static PyObject *__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py = 0;
static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(enum ibv_wc_status); /*proto*/
static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py(enum ibv_wc_opcode); /*proto*/
static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py(enum ibv_create_cq_wc_flags); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "pyverbs.cq"
extern int __pyx_module_is_main_pyverbs__cq;
int __pyx_module_is_main_pyverbs__cq = 0;

/* Implementation of "pyverbs.cq" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_super;
/* #### Code section: string_decls ### */
/* #### Code section: decls ### */
static int __pyx_pf_7pyverbs_2cq_11CompChannel___init__(struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_self, struct __pyx_obj_7pyverbs_6device_Context *__pyx_v_context); /* proto */
static void __pyx_pf_7pyverbs_2cq_11CompChannel_2__dealloc__(struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_11CompChannel_4close(struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_11CompChannel_6get_cq_event(struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_self, struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_expected_cq); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_11CompChannel_2fd___get__(struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_11CompChannel_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_11CompChannel_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7pyverbs_2cq_2CQ___init__(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self, struct __pyx_obj_7pyverbs_6device_Context *__pyx_v_context, PyObject *__pyx_v_cqe, PyObject *__pyx_v_cq_context, struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_channel, PyObject *__pyx_v_comp_vector); /* proto */
static void __pyx_pf_7pyverbs_2cq_2CQ_2__dealloc__(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2CQ_4close(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2CQ_6resize(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self, PyObject *__pyx_v_cqe); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2CQ_8poll(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self, PyObject *__pyx_v_num_entries); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2CQ_10req_notify(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self, PyObject *__pyx_v_solicited_only); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2CQ_12ack_events(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self, PyObject *__pyx_v_num_events); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2CQ_14__str__(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2CQ_12comp_channel___get__(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2CQ_3cqe___get__(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2CQ_2cq___get__(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2CQ_16__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2CQ_18__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7pyverbs_2cq_12CqInitAttrEx___init__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self, PyObject *__pyx_v_cqe, struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_channel, PyObject *__pyx_v_comp_vector, PyObject *__pyx_v_wc_flags, PyObject *__pyx_v_comp_mask, PyObject *__pyx_v_flags, struct __pyx_obj_7pyverbs_2pd_PD *__pyx_v_parent_domain); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_12CqInitAttrEx_3cqe___get__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_3cqe_2__set__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static int __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_10cq_context___set__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_12CqInitAttrEx_13parent_domain___get__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_13parent_domain_2__set__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self, struct __pyx_obj_7pyverbs_2pd_PD *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_12CqInitAttrEx_12comp_channel___get__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_12comp_channel_2__set__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self, struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_12CqInitAttrEx_11comp_vector___get__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_11comp_vector_2__set__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_12CqInitAttrEx_8wc_flags___get__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_8wc_flags_2__set__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_12CqInitAttrEx_9comp_mask___get__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_9comp_mask_2__set__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_12CqInitAttrEx_5flags___get__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_5flags_2__set__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_12CqInitAttrEx_2__str__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_12CqInitAttrEx_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_12CqInitAttrEx_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7pyverbs_2cq_4CQEX___init__(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self, struct __pyx_obj_7pyverbs_6device_Context *__pyx_v_context, struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_init_attr); /* proto */
static void __pyx_pf_7pyverbs_2cq_4CQEX_2__dealloc__(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_4close(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_6start_poll(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self, struct __pyx_obj_7pyverbs_2cq_PollCqAttr *__pyx_v_attr); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_8poll_next(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_10end_poll(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_12read_opcode(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_14read_vendor_err(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_16read_byte_len(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_18read_imm_data(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_20read_qp_num(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_22read_src_qp(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_24read_wc_flags(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_26read_slid(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_28read_sl(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_30read_dlid_path_bits(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_32read_timestamp(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_34read_cvlan(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_36read_flow_tag(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_38read_tm_info(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_40read_completion_wallclock_ns(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_6status___get__(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_4CQEX_6status_2__set__(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_5wr_id___get__(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_4CQEX_5wr_id_2__set__(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_42__str__(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_44__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_46__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7pyverbs_2cq_2WC___init__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_wr_id, PyObject *__pyx_v_status, PyObject *__pyx_v_opcode, PyObject *__pyx_v_vendor_err, PyObject *__pyx_v_byte_len, PyObject *__pyx_v_qp_num, PyObject *__pyx_v_src_qp, PyObject *__pyx_v_imm_data, PyObject *__pyx_v_wc_flags, PyObject *__pyx_v_pkey_index, PyObject *__pyx_v_slid, PyObject *__pyx_v_sl, PyObject *__pyx_v_dlid_path_bits); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2WC_5wr_id___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_2WC_5wr_id_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2WC_6status___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_2WC_6status_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2WC_6opcode___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_2WC_6opcode_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2WC_10vendor_err___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_2WC_10vendor_err_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2WC_8byte_len___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_2WC_8byte_len_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2WC_6qp_num___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_2WC_6qp_num_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2WC_6src_qp___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_2WC_6src_qp_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2WC_8wc_flags___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_2WC_8wc_flags_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2WC_10pkey_index___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_2WC_10pkey_index_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2WC_4slid___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_2WC_4slid_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2WC_2sl___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_2WC_2sl_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2WC_8imm_data___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_2WC_8imm_data_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2WC_14dlid_path_bits___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_2WC_14dlid_path_bits_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2WC_2__str__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2WC_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2WC_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_10PollCqAttr_9comp_mask___get__(struct __pyx_obj_7pyverbs_2cq_PollCqAttr *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_10PollCqAttr_9comp_mask_2__set__(struct __pyx_obj_7pyverbs_2cq_PollCqAttr *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_10PollCqAttr___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_PollCqAttr *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_10PollCqAttr_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_PollCqAttr *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_8WcTmInfo_3tag___get__(struct __pyx_obj_7pyverbs_2cq_WcTmInfo *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_8WcTmInfo_3tag_2__set__(struct __pyx_obj_7pyverbs_2cq_WcTmInfo *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_8WcTmInfo_4priv___get__(struct __pyx_obj_7pyverbs_2cq_WcTmInfo *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_2cq_8WcTmInfo_4priv_2__set__(struct __pyx_obj_7pyverbs_2cq_WcTmInfo *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_8WcTmInfo___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_WcTmInfo *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_8WcTmInfo_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_WcTmInfo *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_cqe_status_to_str(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_status); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_2cqe_opcode_to_str(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_opcode); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_4flags_to_str(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_flags, PyObject *__pyx_v_dictionary); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_6cqe_flags_to_str(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_flags); /* proto */
static PyObject *__pyx_pf_7pyverbs_2cq_8create_wc_flags_to_str(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_flags); /* proto */
static PyObject *__pyx_tp_new_7pyverbs_2cq_CompChannel(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7pyverbs_2cq_CQ(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7pyverbs_2cq_CqInitAttrEx(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7pyverbs_2cq_CQEX(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7pyverbs_2cq_WC(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7pyverbs_2cq_PollCqAttr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7pyverbs_2cq_WcTmInfo(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
/* SmallCodeConfig */
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  PyTypeObject *__pyx_ptype_7pyverbs_4base_PyverbsObject;
  PyTypeObject *__pyx_ptype_7pyverbs_4base_PyverbsCM;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_Context;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_DeviceAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_QueryDeviceExInput;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_ODPCaps;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_RSSCaps;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_PacketPacingCaps;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_PCIAtomicCaps;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_TMCaps;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_CQModerationCaps;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_TSOCaps;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_DeviceAttrEx;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_AllocDmAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_DM;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_PortAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_GIDEntry;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_AsyncEvent;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_FdArr;
  PyTypeObject *__pyx_ptype_7pyverbs_2pd_PD;
  PyTypeObject *__pyx_ptype_7pyverbs_2pd_ParentDomainInitAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_2pd_ParentDomain;
  PyTypeObject *__pyx_ptype_7pyverbs_2pd_ParentDomainContext;
  PyTypeObject *__pyx_ptype_7pyverbs_3srq_SrqAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_3srq_SrqInitAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_3srq_SrqInitAttrEx;
  PyTypeObject *__pyx_ptype_7pyverbs_3srq_OpsWr;
  PyTypeObject *__pyx_ptype_7pyverbs_3srq_SRQ;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_QPCap;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_QPInitAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_QPInitAttrEx;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_QPAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_QP;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_DataBuffer;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_QPEx;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_ECE;
  PyTypeObject *__pyx_ptype_7pyverbs_2wq_WQInitAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_2wq_WQAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_2wq_WQ;
  PyTypeObject *__pyx_ptype_7pyverbs_2wq_RwqIndTableInitAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_2wq_RwqIndTable;
  PyTypeObject *__pyx_ptype_7pyverbs_2wq_RxHashConf;
  PyObject *__pyx_type_7pyverbs_2cq_CompChannel;
  PyObject *__pyx_type_7pyverbs_2cq_CQ;
  PyObject *__pyx_type_7pyverbs_2cq_CqInitAttrEx;
  PyObject *__pyx_type_7pyverbs_2cq_CQEX;
  PyObject *__pyx_type_7pyverbs_2cq_WC;
  PyObject *__pyx_type_7pyverbs_2cq_PollCqAttr;
  PyObject *__pyx_type_7pyverbs_2cq_WcTmInfo;
  PyTypeObject *__pyx_ptype_7pyverbs_2cq_CompChannel;
  PyTypeObject *__pyx_ptype_7pyverbs_2cq_CQ;
  PyTypeObject *__pyx_ptype_7pyverbs_2cq_CqInitAttrEx;
  PyTypeObject *__pyx_ptype_7pyverbs_2cq_CQEX;
  PyTypeObject *__pyx_ptype_7pyverbs_2cq_WC;
  PyTypeObject *__pyx_ptype_7pyverbs_2cq_PollCqAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_2cq_WcTmInfo;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_items;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_values;
  __Pyx_CachedCFunction __pyx_umethod_PyUnicode_Type__format;
  PyObject *__pyx_tuple[2];
  PyObject *__pyx_codeobj_tab[45];
  PyObject *__pyx_string_tab[346];
  PyObject *__pyx_number_tab[21];
/* #### Code section: module_state_contents ### */
/* CommonTypesMetaclass.module_state_decls */
PyTypeObject *__pyx_CommonTypesMetaclassType;

/* CachedMethodType.module_state_decls */
#if CYTHON_COMPILING_IN_LIMITED_API
PyObject *__Pyx_CachedMethodType;
#endif

/* CythonFunctionShared.module_state_decls */
PyTypeObject *__pyx_CyFunctionType;

/* CodeObjectCache.module_state_decls */
struct __Pyx_CodeObjectCache __pyx_code_cache;

/* #### Code section: module_state_end ### */
} __pyx_mstatetype;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate_global (__Pyx_PyModule_GetState(__Pyx_State_FindModule(&__pyx_moduledef)))

#define __pyx_m (__Pyx_State_FindModule(&__pyx_moduledef))
#else
static __pyx_mstatetype __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstatetype * const __pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: constant_name_defines ### */
#define __pyx_kp_u_ __pyx_string_tab[0]
#define __pyx_kp_u_22_20 __pyx_string_tab[1]
#define __pyx_kp_u_22_20_2 __pyx_string_tab[2]
#define __pyx_kp_u_Bind_Memory_window __pyx_string_tab[3]
#define __pyx_kp_u_Closing_CQ __pyx_string_tab[4]
#define __pyx_kp_u_Closing_CQEx __pyx_string_tab[5]
#define __pyx_kp_u_Closing_completion_channel __pyx_string_tab[6]
#define __pyx_kp_u_Compare_and_swap __pyx_string_tab[7]
#define __pyx_kp_u_Created_a_CQ __pyx_string_tab[8]
#define __pyx_kp_u_Created_a_Completion_Channel __pyx_string_tab[9]
#define __pyx_kp_u_Driver_WR __pyx_string_tab[10]
#define __pyx_kp_u_Extended_CQ __pyx_string_tab[11]
#define __pyx_kp_u_Failed_to_close_CQ __pyx_string_tab[12]
#define __pyx_kp_u_Failed_to_create_a_CQ __pyx_string_tab[13]
#define __pyx_kp_u_Failed_to_create_a_completion_ch __pyx_string_tab[14]
#define __pyx_kp_u_Failed_to_create_extended_CQ __pyx_string_tab[15]
#define __pyx_kp_u_Failed_to_destroy_CQEX __pyx_string_tab[16]
#define __pyx_kp_u_Failed_to_destroy_a_completion_c __pyx_string_tab[17]
#define __pyx_kp_u_Failed_to_get_CQ_event __pyx_string_tab[18]
#define __pyx_kp_u_Failed_to_poll_CQ __pyx_string_tab[19]
#define __pyx_kp_u_Failed_to_resize_CQ __pyx_string_tab[20]
#define __pyx_kp_u_Fetch_and_add __pyx_string_tab[21]
#define __pyx_kp_u_IP_csum_OK __pyx_string_tab[22]
#define __pyx_kp_u_Local_invalidate __pyx_string_tab[23]
#define __pyx_kp_u_Note_that_Cython_is_deliberately __pyx_string_tab[24]
#define __pyx_kp_u_Number_of_CQEs __pyx_string_tab[25]
#define __pyx_kp_u_Pickling_of_struct_members_such __pyx_string_tab[26]
#define __pyx_kp_u_Pickling_of_struct_members_such_2 __pyx_string_tab[27]
#define __pyx_kp_u_Pickling_of_struct_members_such_3 __pyx_string_tab[28]
#define __pyx_kp_u_QP_num __pyx_string_tab[29]
#define __pyx_kp_u_RDMA_read __pyx_string_tab[30]
#define __pyx_kp_u_RDMA_write __pyx_string_tab[31]
#define __pyx_kp_u_RNR_retry_exceeded __pyx_string_tab[32]
#define __pyx_kp_u_Receive_RDMA_with_immediate __pyx_string_tab[33]
#define __pyx_kp_u_Received_event_on_an_unexpected __pyx_string_tab[34]
#define __pyx_kp_u_Request_notify_CQ_returned_rc __pyx_string_tab[35]
#define __pyx_kp_u_TM_data_valid __pyx_string_tab[36]
#define __pyx_kp_u_TM_match __pyx_string_tab[37]
#define __pyx_kp_u_TM_sync_request __pyx_string_tab[38]
#define __pyx_kp_u_Tag_matching_add __pyx_string_tab[39]
#define __pyx_kp_u_Tag_matching_delete __pyx_string_tab[40]
#define __pyx_kp_u_Tag_matching_no_tag __pyx_string_tab[41]
#define __pyx_kp_u_Tag_matching_receive __pyx_string_tab[42]
#define __pyx_kp_u_Tag_matching_sync __pyx_string_tab[43]
#define __pyx_kp_u_Unknown_CQE_opcode_op __pyx_string_tab[44]
#define __pyx_kp_u_Unknown_CQE_status __pyx_string_tab[45]
#define __pyx_kp_u_Unrecognized_object_type __pyx_string_tab[46]
#define __pyx_kp_u_WC_fatal_error __pyx_string_tab[47]
#define __pyx_kp_u_WC_flags __pyx_string_tab[48]
#define __pyx_kp_u_WR_ID __pyx_string_tab[49]
#define __pyx_kp_u_WR_flush_error __pyx_string_tab[50]
#define __pyx_kp_u_With_immediate __pyx_string_tab[51]
#define __pyx_kp_u_With_invalidate __pyx_string_tab[52]
#define __pyx_kp_u__2 __pyx_string_tab[53]
#define __pyx_kp_u__3 __pyx_string_tab[54]
#define __pyx_kp_u__4 __pyx_string_tab[55]
#define __pyx_kp_u_add_note __pyx_string_tab[56]
#define __pyx_kp_u_bad_response_error __pyx_string_tab[57]
#define __pyx_kp_u_byte_length __pyx_string_tab[58]
#define __pyx_kp_u_comp_mask_2 __pyx_string_tab[59]
#define __pyx_kp_u_disable __pyx_string_tab[60]
#define __pyx_kp_u_dlid_path_bits_2 __pyx_string_tab[61]
#define __pyx_kp_u_enable __pyx_string_tab[62]
#define __pyx_kp_u_enum_class_ibv_create_cq_wc_flag __pyx_string_tab[63]
#define __pyx_kp_u_enum_class_ibv_wc_opcode_not_imp __pyx_string_tab[64]
#define __pyx_kp_u_enum_class_ibv_wc_status_not_imp __pyx_string_tab[65]
#define __pyx_kp_u_gc __pyx_string_tab[66]
#define __pyx_kp_u_general_error __pyx_string_tab[67]
#define __pyx_kp_u_invalidate_EECN_error __pyx_string_tab[68]
#define __pyx_kp_u_invalidate_EEC_state_error __pyx_string_tab[69]
#define __pyx_kp_u_isenabled __pyx_string_tab[70]
#define __pyx_kp_u_local_EEC_op_error __pyx_string_tab[71]
#define __pyx_kp_u_local_QP_op_error __pyx_string_tab[72]
#define __pyx_kp_u_local_RDD_violation_error __pyx_string_tab[73]
#define __pyx_kp_u_local_access_error __pyx_string_tab[74]
#define __pyx_kp_u_local_length_error __pyx_string_tab[75]
#define __pyx_kp_u_local_protection_error __pyx_string_tab[76]
#define __pyx_kp_u_memory_window_bind_error __pyx_string_tab[77]
#define __pyx_kp_u_pkey_index_2 __pyx_string_tab[78]
#define __pyx_kp_u_pyverbs_cq_pyx __pyx_string_tab[79]
#define __pyx_kp_u_remote_abort_error __pyx_string_tab[80]
#define __pyx_kp_u_remote_access_error __pyx_string_tab[81]
#define __pyx_kp_u_remote_invalidate_RD_request_err __pyx_string_tab[82]
#define __pyx_kp_u_remote_invalidate_request_error __pyx_string_tab[83]
#define __pyx_kp_u_remote_op_error __pyx_string_tab[84]
#define __pyx_kp_u_response_timeout_error __pyx_string_tab[85]
#define __pyx_kp_u_retry_exceeded_error __pyx_string_tab[86]
#define __pyx_kp_u_self_attr_cannot_be_converted_to __pyx_string_tab[87]
#define __pyx_kp_u_self_cc_cannot_be_converted_to_a __pyx_string_tab[88]
#define __pyx_kp_u_self_cq_cannot_be_converted_to_a __pyx_string_tab[89]
#define __pyx_kp_u_self_cq_self_ibv_cq_cannot_be_co __pyx_string_tab[90]
#define __pyx_kp_u_source_QP __pyx_string_tab[91]
#define __pyx_kp_u_tree_fragment __pyx_string_tab[92]
#define __pyx_kp_u_vendor_error __pyx_string_tab[93]
#define __pyx_n_u_CQ __pyx_string_tab[94]
#define __pyx_n_u_CQEX __pyx_string_tab[95]
#define __pyx_n_u_CQEX___reduce_cython __pyx_string_tab[96]
#define __pyx_n_u_CQEX___setstate_cython __pyx_string_tab[97]
#define __pyx_n_u_CQEX_close __pyx_string_tab[98]
#define __pyx_n_u_CQEX_end_poll __pyx_string_tab[99]
#define __pyx_n_u_CQEX_poll_next __pyx_string_tab[100]
#define __pyx_n_u_CQEX_read_byte_len __pyx_string_tab[101]
#define __pyx_n_u_CQEX_read_completion_wallclock_n __pyx_string_tab[102]
#define __pyx_n_u_CQEX_read_cvlan __pyx_string_tab[103]
#define __pyx_n_u_CQEX_read_dlid_path_bits __pyx_string_tab[104]
#define __pyx_n_u_CQEX_read_flow_tag __pyx_string_tab[105]
#define __pyx_n_u_CQEX_read_imm_data __pyx_string_tab[106]
#define __pyx_n_u_CQEX_read_opcode __pyx_string_tab[107]
#define __pyx_n_u_CQEX_read_qp_num __pyx_string_tab[108]
#define __pyx_n_u_CQEX_read_sl __pyx_string_tab[109]
#define __pyx_n_u_CQEX_read_slid __pyx_string_tab[110]
#define __pyx_n_u_CQEX_read_src_qp __pyx_string_tab[111]
#define __pyx_n_u_CQEX_read_timestamp __pyx_string_tab[112]
#define __pyx_n_u_CQEX_read_tm_info __pyx_string_tab[113]
#define __pyx_n_u_CQEX_read_vendor_err __pyx_string_tab[114]
#define __pyx_n_u_CQEX_read_wc_flags __pyx_string_tab[115]
#define __pyx_n_u_CQEX_start_poll __pyx_string_tab[116]
#define __pyx_n_u_CQEs __pyx_string_tab[117]
#define __pyx_n_u_CQ_2 __pyx_string_tab[118]
#define __pyx_n_u_CQ___reduce_cython __pyx_string_tab[119]
#define __pyx_n_u_CQ___setstate_cython __pyx_string_tab[120]
#define __pyx_n_u_CQ_ack_events __pyx_string_tab[121]
#define __pyx_n_u_CQ_close __pyx_string_tab[122]
#define __pyx_n_u_CQ_poll __pyx_string_tab[123]
#define __pyx_n_u_CQ_req_notify __pyx_string_tab[124]
#define __pyx_n_u_CQ_resize __pyx_string_tab[125]
#define __pyx_n_u_CompChannel __pyx_string_tab[126]
#define __pyx_n_u_CompChannel___reduce_cython __pyx_string_tab[127]
#define __pyx_n_u_CompChannel___setstate_cython __pyx_string_tab[128]
#define __pyx_n_u_CompChannel_close __pyx_string_tab[129]
#define __pyx_n_u_CompChannel_get_cq_event __pyx_string_tab[130]
#define __pyx_n_u_CqInitAttrEx __pyx_string_tab[131]
#define __pyx_n_u_CqInitAttrEx___reduce_cython __pyx_string_tab[132]
#define __pyx_n_u_CqInitAttrEx___setstate_cython __pyx_string_tab[133]
#define __pyx_n_u_GRH __pyx_string_tab[134]
#define __pyx_n_u_Handle __pyx_string_tab[135]
#define __pyx_n_u_IBV_WC_ATOMIC_WRITE __pyx_string_tab[136]
#define __pyx_n_u_IBV_WC_BAD_RESP_ERR __pyx_string_tab[137]
#define __pyx_n_u_IBV_WC_BIND_MW __pyx_string_tab[138]
#define __pyx_n_u_IBV_WC_COMP_SWAP __pyx_string_tab[139]
#define __pyx_n_u_IBV_WC_DRIVER2 __pyx_string_tab[140]
#define __pyx_n_u_IBV_WC_DRIVER3 __pyx_string_tab[141]
#define __pyx_n_u_IBV_WC_EX_WITH_BYTE_LEN __pyx_string_tab[142]
#define __pyx_n_u_IBV_WC_EX_WITH_COMPLETION_TIMEST __pyx_string_tab[143]
#define __pyx_n_u_IBV_WC_EX_WITH_COMPLETION_TIMEST_2 __pyx_string_tab[144]
#define __pyx_n_u_IBV_WC_EX_WITH_CVLAN __pyx_string_tab[145]
#define __pyx_n_u_IBV_WC_EX_WITH_DLID_PATH_BITS __pyx_string_tab[146]
#define __pyx_n_u_IBV_WC_EX_WITH_FLOW_TAG __pyx_string_tab[147]
#define __pyx_n_u_IBV_WC_EX_WITH_IMM __pyx_string_tab[148]
#define __pyx_n_u_IBV_WC_EX_WITH_QP_NUM __pyx_string_tab[149]
#define __pyx_n_u_IBV_WC_EX_WITH_SL __pyx_string_tab[150]
#define __pyx_n_u_IBV_WC_EX_WITH_SLID __pyx_string_tab[151]
#define __pyx_n_u_IBV_WC_EX_WITH_SRC_QP __pyx_string_tab[152]
#define __pyx_n_u_IBV_WC_EX_WITH_TM_INFO __pyx_string_tab[153]
#define __pyx_n_u_IBV_WC_FATAL_ERR __pyx_string_tab[154]
#define __pyx_n_u_IBV_WC_FETCH_ADD __pyx_string_tab[155]
#define __pyx_n_u_IBV_WC_FLUSH __pyx_string_tab[156]
#define __pyx_n_u_IBV_WC_GENERAL_ERR __pyx_string_tab[157]
#define __pyx_n_u_IBV_WC_INV_EECN_ERR __pyx_string_tab[158]
#define __pyx_n_u_IBV_WC_INV_EEC_STATE_ERR __pyx_string_tab[159]
#define __pyx_n_u_IBV_WC_LOCAL_INV __pyx_string_tab[160]
#define __pyx_n_u_IBV_WC_LOC_ACCESS_ERR __pyx_string_tab[161]
#define __pyx_n_u_IBV_WC_LOC_EEC_OP_ERR __pyx_string_tab[162]
#define __pyx_n_u_IBV_WC_LOC_LEN_ERR __pyx_string_tab[163]
#define __pyx_n_u_IBV_WC_LOC_PROT_ERR __pyx_string_tab[164]
#define __pyx_n_u_IBV_WC_LOC_QP_OP_ERR __pyx_string_tab[165]
#define __pyx_n_u_IBV_WC_LOC_RDD_VIOL_ERR __pyx_string_tab[166]
#define __pyx_n_u_IBV_WC_MW_BIND_ERR __pyx_string_tab[167]
#define __pyx_n_u_IBV_WC_RDMA_READ __pyx_string_tab[168]
#define __pyx_n_u_IBV_WC_RDMA_WRITE __pyx_string_tab[169]
#define __pyx_n_u_IBV_WC_RECV __pyx_string_tab[170]
#define __pyx_n_u_IBV_WC_RECV_RDMA_WITH_IMM __pyx_string_tab[171]
#define __pyx_n_u_IBV_WC_REM_ABORT_ERR __pyx_string_tab[172]
#define __pyx_n_u_IBV_WC_REM_ACCESS_ERR __pyx_string_tab[173]
#define __pyx_n_u_IBV_WC_REM_INV_RD_REQ_ERR __pyx_string_tab[174]
#define __pyx_n_u_IBV_WC_REM_INV_REQ_ERR __pyx_string_tab[175]
#define __pyx_n_u_IBV_WC_REM_OP_ERR __pyx_string_tab[176]
#define __pyx_n_u_IBV_WC_RESP_TIMEOUT_ERR __pyx_string_tab[177]
#define __pyx_n_u_IBV_WC_RETRY_EXC_ERR __pyx_string_tab[178]
#define __pyx_n_u_IBV_WC_RNR_RETRY_EXC_ERR __pyx_string_tab[179]
#define __pyx_n_u_IBV_WC_SEND __pyx_string_tab[180]
#define __pyx_n_u_IBV_WC_SUCCESS __pyx_string_tab[181]
#define __pyx_n_u_IBV_WC_TM_ADD __pyx_string_tab[182]
#define __pyx_n_u_IBV_WC_TM_DEL __pyx_string_tab[183]
#define __pyx_n_u_IBV_WC_TM_ERR __pyx_string_tab[184]
#define __pyx_n_u_IBV_WC_TM_NO_TAG __pyx_string_tab[185]
#define __pyx_n_u_IBV_WC_TM_RECV __pyx_string_tab[186]
#define __pyx_n_u_IBV_WC_TM_RNDV_INCOMPLETE __pyx_string_tab[187]
#define __pyx_n_u_IBV_WC_TM_SYNC __pyx_string_tab[188]
#define __pyx_n_u_IBV_WC_TSO __pyx_string_tab[189]
#define __pyx_n_u_IBV_WC_WR_FLUSH_ERR __pyx_string_tab[190]
#define __pyx_n_u_PollCqAttr __pyx_string_tab[191]
#define __pyx_n_u_PollCqAttr___reduce_cython __pyx_string_tab[192]
#define __pyx_n_u_PollCqAttr___setstate_cython __pyx_string_tab[193]
#define __pyx_n_u_PyverbsError __pyx_string_tab[194]
#define __pyx_n_u_PyverbsRDMAErrno __pyx_string_tab[195]
#define __pyx_n_u_PyverbsRDMAError __pyx_string_tab[196]
#define __pyx_n_u_Pyx_PyDict_NextRef __pyx_string_tab[197]
#define __pyx_n_u_Receive __pyx_string_tab[198]
#define __pyx_n_u_Send __pyx_string_tab[199]
#define __pyx_n_u_TSO __pyx_string_tab[200]
#define __pyx_n_u_WC __pyx_string_tab[201]
#define __pyx_n_u_WC___reduce_cython __pyx_string_tab[202]
#define __pyx_n_u_WC___setstate_cython __pyx_string_tab[203]
#define __pyx_n_u_WcTmInfo __pyx_string_tab[204]
#define __pyx_n_u_WcTmInfo___reduce_cython __pyx_string_tab[205]
#define __pyx_n_u_WcTmInfo___setstate_cython __pyx_string_tab[206]
#define __pyx_n_u_WeakSet __pyx_string_tab[207]
#define __pyx_n_u_ack_events __pyx_string_tab[208]
#define __pyx_n_u_add __pyx_string_tab[209]
#define __pyx_n_u_add_ref __pyx_string_tab[210]
#define __pyx_n_u_asyncio_coroutines __pyx_string_tab[211]
#define __pyx_n_u_attr __pyx_string_tab[212]
#define __pyx_n_u_byte_len __pyx_string_tab[213]
#define __pyx_n_u_channel __pyx_string_tab[214]
#define __pyx_n_u_class_getitem __pyx_string_tab[215]
#define __pyx_n_u_cline_in_traceback __pyx_string_tab[216]
#define __pyx_n_u_close __pyx_string_tab[217]
#define __pyx_n_u_comp_channel __pyx_string_tab[218]
#define __pyx_n_u_comp_mask __pyx_string_tab[219]
#define __pyx_n_u_comp_vector __pyx_string_tab[220]
#define __pyx_n_u_context __pyx_string_tab[221]
#define __pyx_n_u_cq __pyx_string_tab[222]
#define __pyx_n_u_cq_context __pyx_string_tab[223]
#define __pyx_n_u_cqe __pyx_string_tab[224]
#define __pyx_n_u_cqe_flags __pyx_string_tab[225]
#define __pyx_n_u_cqe_flags_to_str __pyx_string_tab[226]
#define __pyx_n_u_cqe_opcode_to_str __pyx_string_tab[227]
#define __pyx_n_u_cqe_status_to_str __pyx_string_tab[228]
#define __pyx_n_u_create_wc_flags_to_str __pyx_string_tab[229]
#define __pyx_n_u_ctx __pyx_string_tab[230]
#define __pyx_n_u_debug __pyx_string_tab[231]
#define __pyx_n_u_dictionary __pyx_string_tab[232]
#define __pyx_n_u_dlid_path_bits __pyx_string_tab[233]
#define __pyx_n_u_end_poll __pyx_string_tab[234]
#define __pyx_n_u_expected_cq __pyx_string_tab[235]
#define __pyx_n_u_f __pyx_string_tab[236]
#define __pyx_n_u_flags __pyx_string_tab[237]
#define __pyx_n_u_flags_str __pyx_string_tab[238]
#define __pyx_n_u_flags_to_str __pyx_string_tab[239]
#define __pyx_n_u_format __pyx_string_tab[240]
#define __pyx_n_u_func __pyx_string_tab[241]
#define __pyx_n_u_get_cq_event __pyx_string_tab[242]
#define __pyx_n_u_getstate __pyx_string_tab[243]
#define __pyx_n_u_ibv_create_cq_wc_flags __pyx_string_tab[244]
#define __pyx_n_u_ibv_wc_opcode __pyx_string_tab[245]
#define __pyx_n_u_ibv_wc_status __pyx_string_tab[246]
#define __pyx_n_u_imm_data __pyx_string_tab[247]
#define __pyx_n_u_info __pyx_string_tab[248]
#define __pyx_n_u_init __pyx_string_tab[249]
#define __pyx_n_u_init_attr __pyx_string_tab[250]
#define __pyx_n_u_is_coroutine __pyx_string_tab[251]
#define __pyx_n_u_items __pyx_string_tab[252]
#define __pyx_n_u_main __pyx_string_tab[253]
#define __pyx_n_u_module __pyx_string_tab[254]
#define __pyx_n_u_name __pyx_string_tab[255]
#define __pyx_n_u_npolled __pyx_string_tab[256]
#define __pyx_n_u_num_entries __pyx_string_tab[257]
#define __pyx_n_u_num_events __pyx_string_tab[258]
#define __pyx_n_u_op __pyx_string_tab[259]
#define __pyx_n_u_opcode __pyx_string_tab[260]
#define __pyx_n_u_parent_domain __pyx_string_tab[261]
#define __pyx_n_u_pkey_index __pyx_string_tab[262]
#define __pyx_n_u_poll __pyx_string_tab[263]
#define __pyx_n_u_poll_next __pyx_string_tab[264]
#define __pyx_n_u_pop __pyx_string_tab[265]
#define __pyx_n_u_pyverbs_base __pyx_string_tab[266]
#define __pyx_n_u_pyverbs_cq __pyx_string_tab[267]
#define __pyx_n_u_pyverbs_libibverbs_enums __pyx_string_tab[268]
#define __pyx_n_u_pyverbs_pyverbs_error __pyx_string_tab[269]
#define __pyx_n_u_pyx_state __pyx_string_tab[270]
#define __pyx_n_u_pyx_vtable __pyx_string_tab[271]
#define __pyx_n_u_qp_num __pyx_string_tab[272]
#define __pyx_n_u_qualname __pyx_string_tab[273]
#define __pyx_n_u_rc __pyx_string_tab[274]
#define __pyx_n_u_read_byte_len __pyx_string_tab[275]
#define __pyx_n_u_read_completion_wallclock_ns __pyx_string_tab[276]
#define __pyx_n_u_read_cvlan __pyx_string_tab[277]
#define __pyx_n_u_read_dlid_path_bits __pyx_string_tab[278]
#define __pyx_n_u_read_flow_tag __pyx_string_tab[279]
#define __pyx_n_u_read_imm_data __pyx_string_tab[280]
#define __pyx_n_u_read_opcode __pyx_string_tab[281]
#define __pyx_n_u_read_qp_num __pyx_string_tab[282]
#define __pyx_n_u_read_sl __pyx_string_tab[283]
#define __pyx_n_u_read_slid __pyx_string_tab[284]
#define __pyx_n_u_read_src_qp __pyx_string_tab[285]
#define __pyx_n_u_read_timestamp __pyx_string_tab[286]
#define __pyx_n_u_read_tm_info __pyx_string_tab[287]
#define __pyx_n_u_read_vendor_err __pyx_string_tab[288]
#define __pyx_n_u_read_wc_flags __pyx_string_tab[289]
#define __pyx_n_u_reduce __pyx_string_tab[290]
#define __pyx_n_u_reduce_cython __pyx_string_tab[291]
#define __pyx_n_u_reduce_ex __pyx_string_tab[292]
#define __pyx_n_u_req_notify __pyx_string_tab[293]
#define __pyx_n_u_resize __pyx_string_tab[294]
#define __pyx_n_u_self __pyx_string_tab[295]
#define __pyx_n_u_set_name __pyx_string_tab[296]
#define __pyx_n_u_setdefault __pyx_string_tab[297]
#define __pyx_n_u_setstate __pyx_string_tab[298]
#define __pyx_n_u_setstate_cython __pyx_string_tab[299]
#define __pyx_n_u_sl __pyx_string_tab[300]
#define __pyx_n_u_slid __pyx_string_tab[301]
#define __pyx_n_u_solicited_only __pyx_string_tab[302]
#define __pyx_n_u_src_qp __pyx_string_tab[303]
#define __pyx_n_u_start_poll __pyx_string_tab[304]
#define __pyx_n_u_status __pyx_string_tab[305]
#define __pyx_n_u_success __pyx_string_tab[306]
#define __pyx_n_u_super __pyx_string_tab[307]
#define __pyx_n_u_test __pyx_string_tab[308]
#define __pyx_n_u_values __pyx_string_tab[309]
#define __pyx_n_u_vendor_err __pyx_string_tab[310]
#define __pyx_n_u_warn __pyx_string_tab[311]
#define __pyx_n_u_warnings __pyx_string_tab[312]
#define __pyx_n_u_wc __pyx_string_tab[313]
#define __pyx_n_u_wc_flags __pyx_string_tab[314]
#define __pyx_n_u_wcs __pyx_string_tab[315]
#define __pyx_n_u_weakref __pyx_string_tab[316]
#define __pyx_n_u_wr_id __pyx_string_tab[317]
#define __pyx_kp_b_PyObject_PyObject_close_weakrefs __pyx_string_tab[318]
#define __pyx_kp_b_iso88591_A_1D __pyx_string_tab[319]
#define __pyx_kp_b_iso88591_A_1D_2 __pyx_string_tab[320]
#define __pyx_kp_b_iso88591_A_1D_Q_O1 __pyx_string_tab[321]
#define __pyx_kp_b_iso88591_A_4AT __pyx_string_tab[322]
#define __pyx_kp_b_iso88591_A_4q __pyx_string_tab[323]
#define __pyx_kp_b_iso88591_A_4q_2 __pyx_string_tab[324]
#define __pyx_kp_b_iso88591_A_4t3a_t1_G6_1D_1D_s_Q_aq_a __pyx_string_tab[325]
#define __pyx_kp_b_iso88591_A_4t3a_t1_G6_1D_d_Q_t1_Kq_A_a_s __pyx_string_tab[326]
#define __pyx_kp_b_iso88591_A_4t3a_t1_G6_1D_t6_Q_T_s_Q_a_A_a __pyx_string_tab[327]
#define __pyx_kp_b_iso88591_A_5_1_Q_q_E_a __pyx_string_tab[328]
#define __pyx_kp_b_iso88591_A_Qd __pyx_string_tab[329]
#define __pyx_kp_b_iso88591_A_Qd_2 __pyx_string_tab[330]
#define __pyx_kp_b_iso88591_A_a __pyx_string_tab[331]
#define __pyx_kp_b_iso88591_A_at1 __pyx_string_tab[332]
#define __pyx_kp_b_iso88591_A_nAT_a_1 __pyx_string_tab[333]
#define __pyx_kp_b_iso88591_A_q_A __pyx_string_tab[334]
#define __pyx_kp_b_iso88591_A_q_A_2 __pyx_string_tab[335]
#define __pyx_kp_b_iso88591_A_q_E_aq_3c_1_3c_A_1 __pyx_string_tab[336]
#define __pyx_kp_b_iso88591_A_q_a_2_8_A_B_q_q __pyx_string_tab[337]
#define __pyx_kp_b_iso88591_A_xq_Qd_q_A_q __pyx_string_tab[338]
#define __pyx_kp_b_iso88591_G3_0_1_t_d_Q_q_q __pyx_string_tab[339]
#define __pyx_kp_b_iso88591_Q __pyx_string_tab[340]
#define __pyx_kp_b_iso88591_Q_6_1_1A_1 __pyx_string_tab[341]
#define __pyx_kp_b_iso88591_U_3c_DA_U __pyx_string_tab[342]
#define __pyx_kp_b_iso88591_XT_uA_a_5_WF_a_F_V1_v_q_1 __pyx_string_tab[343]
#define __pyx_kp_b_iso88591_a_q_q_a_q_Q_A_1_Q_q_A_A_1_6aq_q __pyx_string_tab[344]
#define __pyx_kp_b_iso88591_q_a_hb_at5_1A_s_A_a_AQ_s_Q_q_wa __pyx_string_tab[345]
#define __pyx_int_0 __pyx_number_tab[0]
#define __pyx_int_1 __pyx_number_tab[1]
#define __pyx_int_2 __pyx_number_tab[2]
#define __pyx_int_3 __pyx_number_tab[3]
#define __pyx_int_4 __pyx_number_tab[4]
#define __pyx_int_5 __pyx_number_tab[5]
#define __pyx_int_6 __pyx_number_tab[6]
#define __pyx_int_7 __pyx_number_tab[7]
#define __pyx_int_8 __pyx_number_tab[8]
#define __pyx_int_16 __pyx_number_tab[9]
#define __pyx_int_32 __pyx_number_tab[10]
#define __pyx_int_64 __pyx_number_tab[11]
#define __pyx_int_100 __pyx_number_tab[12]
#define __pyx_int_128 __pyx_number_tab[13]
#define __pyx_int_129 __pyx_number_tab[14]
#define __pyx_int_130 __pyx_number_tab[15]
#define __pyx_int_131 __pyx_number_tab[16]
#define __pyx_int_132 __pyx_number_tab[17]
#define __pyx_int_133 __pyx_number_tab[18]
#define __pyx_int_134 __pyx_number_tab[19]
#define __pyx_int_135 __pyx_number_tab[20]
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_clear(PyObject *m) {
  __pyx_mstatetype *clear_module_state = __Pyx_PyModule_GetState(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __Pyx_State_RemoveModule(NULL);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4base_PyverbsObject);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4base_PyverbsCM);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_Context);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_DeviceAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_QueryDeviceExInput);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_ODPCaps);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_RSSCaps);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_PacketPacingCaps);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_PCIAtomicCaps);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_TMCaps);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_CQModerationCaps);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_TSOCaps);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_DeviceAttrEx);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_AllocDmAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_DM);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_PortAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_GIDEntry);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_AsyncEvent);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_FdArr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2pd_PD);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2pd_ParentDomainInitAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2pd_ParentDomain);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2pd_ParentDomainContext);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_3srq_SrqAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_3srq_SrqInitAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_3srq_SrqInitAttrEx);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_3srq_OpsWr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_3srq_SRQ);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_QPCap);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_QPInitAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_QPAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_QP);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_DataBuffer);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_QPEx);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_ECE);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2wq_WQInitAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2wq_WQAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2wq_WQ);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2wq_RwqIndTableInitAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2wq_RwqIndTable);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2wq_RxHashConf);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2cq_CompChannel);
  Py_CLEAR(clear_module_state->__pyx_type_7pyverbs_2cq_CompChannel);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2cq_CQ);
  Py_CLEAR(clear_module_state->__pyx_type_7pyverbs_2cq_CQ);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx);
  Py_CLEAR(clear_module_state->__pyx_type_7pyverbs_2cq_CqInitAttrEx);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2cq_CQEX);
  Py_CLEAR(clear_module_state->__pyx_type_7pyverbs_2cq_CQEX);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2cq_WC);
  Py_CLEAR(clear_module_state->__pyx_type_7pyverbs_2cq_WC);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2cq_PollCqAttr);
  Py_CLEAR(clear_module_state->__pyx_type_7pyverbs_2cq_PollCqAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2cq_WcTmInfo);
  Py_CLEAR(clear_module_state->__pyx_type_7pyverbs_2cq_WcTmInfo);
  for (int i=0; i<2; ++i) { Py_CLEAR(clear_module_state->__pyx_tuple[i]); }
  for (int i=0; i<45; ++i) { Py_CLEAR(clear_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<346; ++i) { Py_CLEAR(clear_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<21; ++i) { Py_CLEAR(clear_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_clear_contents ### */
/* CommonTypesMetaclass.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CyFunctionType);

/* #### Code section: module_state_clear_end ### */
return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstatetype *traverse_module_state = __Pyx_PyModule_GetState(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_tuple);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_bytes);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_unicode);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4base_PyverbsObject);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4base_PyverbsCM);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_Context);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_DeviceAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_QueryDeviceExInput);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_ODPCaps);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_RSSCaps);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_PacketPacingCaps);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_PCIAtomicCaps);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_TMCaps);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_CQModerationCaps);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_TSOCaps);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_DeviceAttrEx);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_AllocDmAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_DM);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_PortAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_GIDEntry);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_AsyncEvent);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_FdArr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2pd_PD);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2pd_ParentDomainInitAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2pd_ParentDomain);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2pd_ParentDomainContext);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_3srq_SrqAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_3srq_SrqInitAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_3srq_SrqInitAttrEx);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_3srq_OpsWr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_3srq_SRQ);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_QPCap);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_QPInitAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_QPAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_QP);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_DataBuffer);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_QPEx);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_ECE);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2wq_WQInitAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2wq_WQAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2wq_WQ);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2wq_RwqIndTableInitAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2wq_RwqIndTable);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2wq_RxHashConf);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2cq_CompChannel);
  Py_VISIT(traverse_module_state->__pyx_type_7pyverbs_2cq_CompChannel);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2cq_CQ);
  Py_VISIT(traverse_module_state->__pyx_type_7pyverbs_2cq_CQ);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx);
  Py_VISIT(traverse_module_state->__pyx_type_7pyverbs_2cq_CqInitAttrEx);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2cq_CQEX);
  Py_VISIT(traverse_module_state->__pyx_type_7pyverbs_2cq_CQEX);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2cq_WC);
  Py_VISIT(traverse_module_state->__pyx_type_7pyverbs_2cq_WC);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2cq_PollCqAttr);
  Py_VISIT(traverse_module_state->__pyx_type_7pyverbs_2cq_PollCqAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2cq_WcTmInfo);
  Py_VISIT(traverse_module_state->__pyx_type_7pyverbs_2cq_WcTmInfo);
  for (int i=0; i<2; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_tuple[i]); }
  for (int i=0; i<45; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<346; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<21; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_traverse_contents ### */
/* CommonTypesMetaclass.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CyFunctionType);

/* #### Code section: module_state_traverse_end ### */
return 0;
}
#endif
/* #### Code section: module_code ### */

/* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(ibv_wc_status c_val):
 *     cdef object __pyx_enum
*/

static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(enum ibv_wc_status __pyx_v_c_val) {
  PyObject *__pyx_v___pyx_enum = 0;
  PyObject *__pyx_v_warnings = NULL;
  int __pyx_v_underlying_c_val;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  size_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py", 0);

  /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_wc_status as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py == Py_None);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_wc_status as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py
 *         except ImportError:
*/
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "EnumTypeToPy":13
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py is None:
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_wc_status as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py             # <<<<<<<<<<<<<<
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py = False  # False indicates "don't try again"
*/
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_wc_status};
          __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyverbs_libibverbs_enums, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L4_error)
        }
        __pyx_t_5 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_5);
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_wc_status};
          __pyx_t_7 = 0; {
            __pyx_t_8 = __Pyx_ImportFrom(__pyx_t_5, __pyx_imported_names[__pyx_t_7]); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 13, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_8);
            switch (__pyx_t_7) {
              case 0:
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py);
              __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py, __pyx_t_8);
              __Pyx_GIVEREF(__pyx_t_8);
              break;
              default:;
            }
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_wc_status as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py
 *         except ImportError:
*/
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "EnumTypeToPy":14
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_wc_status as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py
 *         except ImportError:             # <<<<<<<<<<<<<<
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py = False  # False indicates "don't try again"
 *             import warnings
*/
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_ImportError))));
      if (__pyx_t_9) {
        __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_8, &__pyx_t_10) < 0) __PYX_ERR(1, 14, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_10);

        /* "EnumTypeToPy":15
 *             from pyverbs.libibverbs_enums import ibv_wc_status as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py = False  # False indicates "don't try again"             # <<<<<<<<<<<<<<
 *             import warnings
 *             warnings.warn(
*/
        __Pyx_INCREF(Py_False);
        __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py);
        __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py, Py_False);
        __Pyx_GIVEREF(Py_False);

        /* "EnumTypeToPy":16
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py = False  # False indicates "don't try again"
 *             import warnings             # <<<<<<<<<<<<<<
 *             warnings.warn(
 *                 f"enum class ibv_wc_status not importable from pyverbs.libibverbs_enums. "
*/
        __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_warnings, 0, 0, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 16, __pyx_L6_except_error)
        __pyx_t_11 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_v_warnings = __pyx_t_11;
        __pyx_t_11 = 0;

        /* "EnumTypeToPy":17
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py = False  # False indicates "don't try again"
 *             import warnings
 *             warnings.warn(             # <<<<<<<<<<<<<<
 *                 f"enum class ibv_wc_status not importable from pyverbs.libibverbs_enums. "
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
*/
        __pyx_t_12 = __pyx_v_warnings;
        __Pyx_INCREF(__pyx_t_12);
        __pyx_t_13 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_mstate_global->__pyx_kp_u_enum_class_ibv_wc_status_not_imp};
          __pyx_t_11 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_warn, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 17, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_11);
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L5_exception_handled;
      }
      goto __pyx_L6_except_error;

      /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_wc_status as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py
 *         except ImportError:
*/
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L5_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_L9_try_end:;
    }

    /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_wc_status as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py
*/
  }

  /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py == Py_False);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":24
 * 
 * 
 *         return <int>c_val             # <<<<<<<<<<<<<<
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyLong_From_int(((int)__pyx_v_c_val)); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumTypeToPy":25
 * 
 *         return <int>c_val
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py             # <<<<<<<<<<<<<<
 * 
 *     if 0:
*/
  __Pyx_INCREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py);
  __pyx_v___pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py;

  /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_wc_status.IBV_WC_SUCCESS:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_SUCCESS
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_LEN_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_SUCCESS);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":30
 *         pass
 *     elif c_val == ibv_wc_status.IBV_WC_SUCCESS:
 *         return __pyx_enum.IBV_WC_SUCCESS             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_LEN_ERR:
 *         return __pyx_enum.IBV_WC_LOC_LEN_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_SUCCESS); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_wc_status.IBV_WC_SUCCESS:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_SUCCESS
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_LEN_ERR:
*/
  }

  /* "EnumTypeToPy":31
 *     elif c_val == ibv_wc_status.IBV_WC_SUCCESS:
 *         return __pyx_enum.IBV_WC_SUCCESS
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_LEN_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_LEN_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_QP_OP_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_LOC_LEN_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":32
 *         return __pyx_enum.IBV_WC_SUCCESS
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_LEN_ERR:
 *         return __pyx_enum.IBV_WC_LOC_LEN_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_QP_OP_ERR:
 *         return __pyx_enum.IBV_WC_LOC_QP_OP_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_LOC_LEN_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 32, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":31
 *     elif c_val == ibv_wc_status.IBV_WC_SUCCESS:
 *         return __pyx_enum.IBV_WC_SUCCESS
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_LEN_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_LEN_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_QP_OP_ERR:
*/
  }

  /* "EnumTypeToPy":33
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_LEN_ERR:
 *         return __pyx_enum.IBV_WC_LOC_LEN_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_QP_OP_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_QP_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_EEC_OP_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_LOC_QP_OP_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":34
 *         return __pyx_enum.IBV_WC_LOC_LEN_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_QP_OP_ERR:
 *         return __pyx_enum.IBV_WC_LOC_QP_OP_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_EEC_OP_ERR:
 *         return __pyx_enum.IBV_WC_LOC_EEC_OP_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_LOC_QP_OP_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":33
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_LEN_ERR:
 *         return __pyx_enum.IBV_WC_LOC_LEN_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_QP_OP_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_QP_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_EEC_OP_ERR:
*/
  }

  /* "EnumTypeToPy":35
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_QP_OP_ERR:
 *         return __pyx_enum.IBV_WC_LOC_QP_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_EEC_OP_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_EEC_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_PROT_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_LOC_EEC_OP_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":36
 *         return __pyx_enum.IBV_WC_LOC_QP_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_EEC_OP_ERR:
 *         return __pyx_enum.IBV_WC_LOC_EEC_OP_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_PROT_ERR:
 *         return __pyx_enum.IBV_WC_LOC_PROT_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_LOC_EEC_OP_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":35
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_QP_OP_ERR:
 *         return __pyx_enum.IBV_WC_LOC_QP_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_EEC_OP_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_EEC_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_PROT_ERR:
*/
  }

  /* "EnumTypeToPy":37
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_EEC_OP_ERR:
 *         return __pyx_enum.IBV_WC_LOC_EEC_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_PROT_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_PROT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_WR_FLUSH_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_LOC_PROT_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":38
 *         return __pyx_enum.IBV_WC_LOC_EEC_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_PROT_ERR:
 *         return __pyx_enum.IBV_WC_LOC_PROT_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_WR_FLUSH_ERR:
 *         return __pyx_enum.IBV_WC_WR_FLUSH_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_LOC_PROT_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":37
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_EEC_OP_ERR:
 *         return __pyx_enum.IBV_WC_LOC_EEC_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_PROT_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_PROT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_WR_FLUSH_ERR:
*/
  }

  /* "EnumTypeToPy":39
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_PROT_ERR:
 *         return __pyx_enum.IBV_WC_LOC_PROT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_WR_FLUSH_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_WR_FLUSH_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_MW_BIND_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_WR_FLUSH_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":40
 *         return __pyx_enum.IBV_WC_LOC_PROT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_WR_FLUSH_ERR:
 *         return __pyx_enum.IBV_WC_WR_FLUSH_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_MW_BIND_ERR:
 *         return __pyx_enum.IBV_WC_MW_BIND_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_WR_FLUSH_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":39
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_PROT_ERR:
 *         return __pyx_enum.IBV_WC_LOC_PROT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_WR_FLUSH_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_WR_FLUSH_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_MW_BIND_ERR:
*/
  }

  /* "EnumTypeToPy":41
 *     elif c_val == ibv_wc_status.IBV_WC_WR_FLUSH_ERR:
 *         return __pyx_enum.IBV_WC_WR_FLUSH_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_MW_BIND_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_MW_BIND_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_BAD_RESP_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_MW_BIND_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":42
 *         return __pyx_enum.IBV_WC_WR_FLUSH_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_MW_BIND_ERR:
 *         return __pyx_enum.IBV_WC_MW_BIND_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_BAD_RESP_ERR:
 *         return __pyx_enum.IBV_WC_BAD_RESP_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_MW_BIND_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":41
 *     elif c_val == ibv_wc_status.IBV_WC_WR_FLUSH_ERR:
 *         return __pyx_enum.IBV_WC_WR_FLUSH_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_MW_BIND_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_MW_BIND_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_BAD_RESP_ERR:
*/
  }

  /* "EnumTypeToPy":43
 *     elif c_val == ibv_wc_status.IBV_WC_MW_BIND_ERR:
 *         return __pyx_enum.IBV_WC_MW_BIND_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_BAD_RESP_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_BAD_RESP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_ACCESS_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_BAD_RESP_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":44
 *         return __pyx_enum.IBV_WC_MW_BIND_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_BAD_RESP_ERR:
 *         return __pyx_enum.IBV_WC_BAD_RESP_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_ACCESS_ERR:
 *         return __pyx_enum.IBV_WC_LOC_ACCESS_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_BAD_RESP_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":43
 *     elif c_val == ibv_wc_status.IBV_WC_MW_BIND_ERR:
 *         return __pyx_enum.IBV_WC_MW_BIND_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_BAD_RESP_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_BAD_RESP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_ACCESS_ERR:
*/
  }

  /* "EnumTypeToPy":45
 *     elif c_val == ibv_wc_status.IBV_WC_BAD_RESP_ERR:
 *         return __pyx_enum.IBV_WC_BAD_RESP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_ACCESS_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_ACCESS_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_REQ_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_LOC_ACCESS_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":46
 *         return __pyx_enum.IBV_WC_BAD_RESP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_ACCESS_ERR:
 *         return __pyx_enum.IBV_WC_LOC_ACCESS_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_REQ_ERR:
 *         return __pyx_enum.IBV_WC_REM_INV_REQ_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_LOC_ACCESS_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 46, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":45
 *     elif c_val == ibv_wc_status.IBV_WC_BAD_RESP_ERR:
 *         return __pyx_enum.IBV_WC_BAD_RESP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_ACCESS_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_ACCESS_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_REQ_ERR:
*/
  }

  /* "EnumTypeToPy":47
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_ACCESS_ERR:
 *         return __pyx_enum.IBV_WC_LOC_ACCESS_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_REQ_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_REM_INV_REQ_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ACCESS_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_REM_INV_REQ_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":48
 *         return __pyx_enum.IBV_WC_LOC_ACCESS_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_REQ_ERR:
 *         return __pyx_enum.IBV_WC_REM_INV_REQ_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ACCESS_ERR:
 *         return __pyx_enum.IBV_WC_REM_ACCESS_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_REM_INV_REQ_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":47
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_ACCESS_ERR:
 *         return __pyx_enum.IBV_WC_LOC_ACCESS_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_REQ_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_REM_INV_REQ_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ACCESS_ERR:
*/
  }

  /* "EnumTypeToPy":49
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_REQ_ERR:
 *         return __pyx_enum.IBV_WC_REM_INV_REQ_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ACCESS_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_REM_ACCESS_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_OP_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_REM_ACCESS_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":50
 *         return __pyx_enum.IBV_WC_REM_INV_REQ_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ACCESS_ERR:
 *         return __pyx_enum.IBV_WC_REM_ACCESS_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_REM_OP_ERR:
 *         return __pyx_enum.IBV_WC_REM_OP_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_REM_ACCESS_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":49
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_REQ_ERR:
 *         return __pyx_enum.IBV_WC_REM_INV_REQ_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ACCESS_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_REM_ACCESS_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_OP_ERR:
*/
  }

  /* "EnumTypeToPy":51
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ACCESS_ERR:
 *         return __pyx_enum.IBV_WC_REM_ACCESS_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_OP_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_REM_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RETRY_EXC_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_REM_OP_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":52
 *         return __pyx_enum.IBV_WC_REM_ACCESS_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_OP_ERR:
 *         return __pyx_enum.IBV_WC_REM_OP_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_RETRY_EXC_ERR:
 *         return __pyx_enum.IBV_WC_RETRY_EXC_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_REM_OP_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":51
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ACCESS_ERR:
 *         return __pyx_enum.IBV_WC_REM_ACCESS_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_OP_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_REM_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RETRY_EXC_ERR:
*/
  }

  /* "EnumTypeToPy":53
 *     elif c_val == ibv_wc_status.IBV_WC_REM_OP_ERR:
 *         return __pyx_enum.IBV_WC_REM_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RETRY_EXC_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RETRY_EXC_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RNR_RETRY_EXC_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_RETRY_EXC_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":54
 *         return __pyx_enum.IBV_WC_REM_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RETRY_EXC_ERR:
 *         return __pyx_enum.IBV_WC_RETRY_EXC_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_RNR_RETRY_EXC_ERR:
 *         return __pyx_enum.IBV_WC_RNR_RETRY_EXC_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_RETRY_EXC_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":53
 *     elif c_val == ibv_wc_status.IBV_WC_REM_OP_ERR:
 *         return __pyx_enum.IBV_WC_REM_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RETRY_EXC_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RETRY_EXC_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RNR_RETRY_EXC_ERR:
*/
  }

  /* "EnumTypeToPy":55
 *     elif c_val == ibv_wc_status.IBV_WC_RETRY_EXC_ERR:
 *         return __pyx_enum.IBV_WC_RETRY_EXC_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RNR_RETRY_EXC_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RNR_RETRY_EXC_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_RDD_VIOL_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_RNR_RETRY_EXC_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":56
 *         return __pyx_enum.IBV_WC_RETRY_EXC_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RNR_RETRY_EXC_ERR:
 *         return __pyx_enum.IBV_WC_RNR_RETRY_EXC_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_RDD_VIOL_ERR:
 *         return __pyx_enum.IBV_WC_LOC_RDD_VIOL_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_RNR_RETRY_EXC_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":55
 *     elif c_val == ibv_wc_status.IBV_WC_RETRY_EXC_ERR:
 *         return __pyx_enum.IBV_WC_RETRY_EXC_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RNR_RETRY_EXC_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RNR_RETRY_EXC_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_RDD_VIOL_ERR:
*/
  }

  /* "EnumTypeToPy":57
 *     elif c_val == ibv_wc_status.IBV_WC_RNR_RETRY_EXC_ERR:
 *         return __pyx_enum.IBV_WC_RNR_RETRY_EXC_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_RDD_VIOL_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_RDD_VIOL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_RD_REQ_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_LOC_RDD_VIOL_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":58
 *         return __pyx_enum.IBV_WC_RNR_RETRY_EXC_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_RDD_VIOL_ERR:
 *         return __pyx_enum.IBV_WC_LOC_RDD_VIOL_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_RD_REQ_ERR:
 *         return __pyx_enum.IBV_WC_REM_INV_RD_REQ_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_LOC_RDD_VIOL_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":57
 *     elif c_val == ibv_wc_status.IBV_WC_RNR_RETRY_EXC_ERR:
 *         return __pyx_enum.IBV_WC_RNR_RETRY_EXC_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_RDD_VIOL_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_RDD_VIOL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_RD_REQ_ERR:
*/
  }

  /* "EnumTypeToPy":59
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_RDD_VIOL_ERR:
 *         return __pyx_enum.IBV_WC_LOC_RDD_VIOL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_RD_REQ_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_REM_INV_RD_REQ_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ABORT_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_REM_INV_RD_REQ_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":60
 *         return __pyx_enum.IBV_WC_LOC_RDD_VIOL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_RD_REQ_ERR:
 *         return __pyx_enum.IBV_WC_REM_INV_RD_REQ_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ABORT_ERR:
 *         return __pyx_enum.IBV_WC_REM_ABORT_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_REM_INV_RD_REQ_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":59
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_RDD_VIOL_ERR:
 *         return __pyx_enum.IBV_WC_LOC_RDD_VIOL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_RD_REQ_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_REM_INV_RD_REQ_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ABORT_ERR:
*/
  }

  /* "EnumTypeToPy":61
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_RD_REQ_ERR:
 *         return __pyx_enum.IBV_WC_REM_INV_RD_REQ_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ABORT_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_REM_ABORT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EECN_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_REM_ABORT_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":62
 *         return __pyx_enum.IBV_WC_REM_INV_RD_REQ_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ABORT_ERR:
 *         return __pyx_enum.IBV_WC_REM_ABORT_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EECN_ERR:
 *         return __pyx_enum.IBV_WC_INV_EECN_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_REM_ABORT_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":61
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_RD_REQ_ERR:
 *         return __pyx_enum.IBV_WC_REM_INV_RD_REQ_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ABORT_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_REM_ABORT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EECN_ERR:
*/
  }

  /* "EnumTypeToPy":63
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ABORT_ERR:
 *         return __pyx_enum.IBV_WC_REM_ABORT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EECN_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_INV_EECN_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EEC_STATE_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_INV_EECN_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":64
 *         return __pyx_enum.IBV_WC_REM_ABORT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EECN_ERR:
 *         return __pyx_enum.IBV_WC_INV_EECN_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EEC_STATE_ERR:
 *         return __pyx_enum.IBV_WC_INV_EEC_STATE_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_INV_EECN_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":63
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ABORT_ERR:
 *         return __pyx_enum.IBV_WC_REM_ABORT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EECN_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_INV_EECN_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EEC_STATE_ERR:
*/
  }

  /* "EnumTypeToPy":65
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EECN_ERR:
 *         return __pyx_enum.IBV_WC_INV_EECN_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EEC_STATE_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_INV_EEC_STATE_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_FATAL_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_INV_EEC_STATE_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":66
 *         return __pyx_enum.IBV_WC_INV_EECN_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EEC_STATE_ERR:
 *         return __pyx_enum.IBV_WC_INV_EEC_STATE_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_FATAL_ERR:
 *         return __pyx_enum.IBV_WC_FATAL_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_INV_EEC_STATE_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":65
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EECN_ERR:
 *         return __pyx_enum.IBV_WC_INV_EECN_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EEC_STATE_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_INV_EEC_STATE_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_FATAL_ERR:
*/
  }

  /* "EnumTypeToPy":67
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EEC_STATE_ERR:
 *         return __pyx_enum.IBV_WC_INV_EEC_STATE_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_FATAL_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_FATAL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RESP_TIMEOUT_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_FATAL_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":68
 *         return __pyx_enum.IBV_WC_INV_EEC_STATE_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_FATAL_ERR:
 *         return __pyx_enum.IBV_WC_FATAL_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_RESP_TIMEOUT_ERR:
 *         return __pyx_enum.IBV_WC_RESP_TIMEOUT_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_FATAL_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":67
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EEC_STATE_ERR:
 *         return __pyx_enum.IBV_WC_INV_EEC_STATE_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_FATAL_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_FATAL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RESP_TIMEOUT_ERR:
*/
  }

  /* "EnumTypeToPy":69
 *     elif c_val == ibv_wc_status.IBV_WC_FATAL_ERR:
 *         return __pyx_enum.IBV_WC_FATAL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RESP_TIMEOUT_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RESP_TIMEOUT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_GENERAL_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_RESP_TIMEOUT_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":70
 *         return __pyx_enum.IBV_WC_FATAL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RESP_TIMEOUT_ERR:
 *         return __pyx_enum.IBV_WC_RESP_TIMEOUT_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_GENERAL_ERR:
 *         return __pyx_enum.IBV_WC_GENERAL_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_RESP_TIMEOUT_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":69
 *     elif c_val == ibv_wc_status.IBV_WC_FATAL_ERR:
 *         return __pyx_enum.IBV_WC_FATAL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RESP_TIMEOUT_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RESP_TIMEOUT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_GENERAL_ERR:
*/
  }

  /* "EnumTypeToPy":71
 *     elif c_val == ibv_wc_status.IBV_WC_RESP_TIMEOUT_ERR:
 *         return __pyx_enum.IBV_WC_RESP_TIMEOUT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_GENERAL_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_GENERAL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_TM_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_GENERAL_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":72
 *         return __pyx_enum.IBV_WC_RESP_TIMEOUT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_GENERAL_ERR:
 *         return __pyx_enum.IBV_WC_GENERAL_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_TM_ERR:
 *         return __pyx_enum.IBV_WC_TM_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_GENERAL_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":71
 *     elif c_val == ibv_wc_status.IBV_WC_RESP_TIMEOUT_ERR:
 *         return __pyx_enum.IBV_WC_RESP_TIMEOUT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_GENERAL_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_GENERAL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_TM_ERR:
*/
  }

  /* "EnumTypeToPy":73
 *     elif c_val == ibv_wc_status.IBV_WC_GENERAL_ERR:
 *         return __pyx_enum.IBV_WC_GENERAL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_TM_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_TM_RNDV_INCOMPLETE:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_TM_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":74
 *         return __pyx_enum.IBV_WC_GENERAL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_TM_ERR:
 *         return __pyx_enum.IBV_WC_TM_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_TM_RNDV_INCOMPLETE:
 *         return __pyx_enum.IBV_WC_TM_RNDV_INCOMPLETE
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":73
 *     elif c_val == ibv_wc_status.IBV_WC_GENERAL_ERR:
 *         return __pyx_enum.IBV_WC_GENERAL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_TM_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_TM_RNDV_INCOMPLETE:
*/
  }

  /* "EnumTypeToPy":75
 *     elif c_val == ibv_wc_status.IBV_WC_TM_ERR:
 *         return __pyx_enum.IBV_WC_TM_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_TM_RNDV_INCOMPLETE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_RNDV_INCOMPLETE
 *     else:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_TM_RNDV_INCOMPLETE);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":76
 *         return __pyx_enum.IBV_WC_TM_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_TM_RNDV_INCOMPLETE:
 *         return __pyx_enum.IBV_WC_TM_RNDV_INCOMPLETE             # <<<<<<<<<<<<<<
 *     else:
 *         underlying_c_val = <int>c_val
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_RNDV_INCOMPLETE); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":75
 *     elif c_val == ibv_wc_status.IBV_WC_TM_ERR:
 *         return __pyx_enum.IBV_WC_TM_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_TM_RNDV_INCOMPLETE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_RNDV_INCOMPLETE
 *     else:
*/
  }

  /* "EnumTypeToPy":78
 *         return __pyx_enum.IBV_WC_TM_RNDV_INCOMPLETE
 *     else:
 *         underlying_c_val = <int>c_val             # <<<<<<<<<<<<<<
 *         return __pyx_enum(underlying_c_val)
*/
  /*else*/ {
    __pyx_v_underlying_c_val = ((int)__pyx_v_c_val);

    /* "EnumTypeToPy":79
 *     else:
 *         underlying_c_val = <int>c_val
 *         return __pyx_enum(underlying_c_val)             # <<<<<<<<<<<<<<
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = NULL;
    __Pyx_INCREF(__pyx_v___pyx_enum);
    __pyx_t_5 = __pyx_v___pyx_enum; 
    __pyx_t_11 = __Pyx_PyLong_From_int(__pyx_v_underlying_c_val); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_13 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_8);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_13 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_11};
      __pyx_t_10 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 79, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
    }
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;
  }

  /* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(ibv_wc_status c_val):
 *     cdef object __pyx_enum
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_enum);
  __Pyx_XDECREF(__pyx_v_warnings);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py(enum ibv_wc_opcode __pyx_v_c_val) {
  PyObject *__pyx_v___pyx_enum = 0;
  PyObject *__pyx_v_warnings = NULL;
  int __pyx_v_underlying_c_val;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  size_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py", 0);

  /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_wc_opcode as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py == Py_None);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_wc_opcode as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py
 *         except ImportError:
*/
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "EnumTypeToPy":13
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py is None:
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_wc_opcode as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py             # <<<<<<<<<<<<<<
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py = False  # False indicates "don't try again"
*/
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_wc_opcode};
          __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyverbs_libibverbs_enums, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L4_error)
        }
        __pyx_t_5 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_5);
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_wc_opcode};
          __pyx_t_7 = 0; {
            __pyx_t_8 = __Pyx_ImportFrom(__pyx_t_5, __pyx_imported_names[__pyx_t_7]); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 13, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_8);
            switch (__pyx_t_7) {
              case 0:
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py);
              __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py, __pyx_t_8);
              __Pyx_GIVEREF(__pyx_t_8);
              break;
              default:;
            }
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_wc_opcode as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py
 *         except ImportError:
*/
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "EnumTypeToPy":14
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_wc_opcode as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py
 *         except ImportError:             # <<<<<<<<<<<<<<
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py = False  # False indicates "don't try again"
 *             import warnings
*/
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_ImportError))));
      if (__pyx_t_9) {
        __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_8, &__pyx_t_10) < 0) __PYX_ERR(1, 14, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_10);

        /* "EnumTypeToPy":15
 *             from pyverbs.libibverbs_enums import ibv_wc_opcode as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py = False  # False indicates "don't try again"             # <<<<<<<<<<<<<<
 *             import warnings
 *             warnings.warn(
*/
        __Pyx_INCREF(Py_False);
        __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py);
        __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py, Py_False);
        __Pyx_GIVEREF(Py_False);

        /* "EnumTypeToPy":16
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py = False  # False indicates "don't try again"
 *             import warnings             # <<<<<<<<<<<<<<
 *             warnings.warn(
 *                 f"enum class ibv_wc_opcode not importable from pyverbs.libibverbs_enums. "
*/
        __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_warnings, 0, 0, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 16, __pyx_L6_except_error)
        __pyx_t_11 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_v_warnings = __pyx_t_11;
        __pyx_t_11 = 0;

        /* "EnumTypeToPy":17
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py = False  # False indicates "don't try again"
 *             import warnings
 *             warnings.warn(             # <<<<<<<<<<<<<<
 *                 f"enum class ibv_wc_opcode not importable from pyverbs.libibverbs_enums. "
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
*/
        __pyx_t_12 = __pyx_v_warnings;
        __Pyx_INCREF(__pyx_t_12);
        __pyx_t_13 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_mstate_global->__pyx_kp_u_enum_class_ibv_wc_opcode_not_imp};
          __pyx_t_11 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_warn, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 17, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_11);
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L5_exception_handled;
      }
      goto __pyx_L6_except_error;

      /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_wc_opcode as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py
 *         except ImportError:
*/
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L5_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_L9_try_end:;
    }

    /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_wc_opcode as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py
*/
  }

  /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py == Py_False);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":24
 * 
 * 
 *         return <int>c_val             # <<<<<<<<<<<<<<
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyLong_From_int(((int)__pyx_v_c_val)); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumTypeToPy":25
 * 
 *         return <int>c_val
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py             # <<<<<<<<<<<<<<
 * 
 *     if 0:
*/
  __Pyx_INCREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py);
  __pyx_v___pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py;

  /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_wc_opcode.IBV_WC_SEND:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_SEND
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_WRITE:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_SEND);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":30
 *         pass
 *     elif c_val == ibv_wc_opcode.IBV_WC_SEND:
 *         return __pyx_enum.IBV_WC_SEND             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_WRITE:
 *         return __pyx_enum.IBV_WC_RDMA_WRITE
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_SEND); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_wc_opcode.IBV_WC_SEND:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_SEND
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_WRITE:
*/
  }

  /* "EnumTypeToPy":31
 *     elif c_val == ibv_wc_opcode.IBV_WC_SEND:
 *         return __pyx_enum.IBV_WC_SEND
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_WRITE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RDMA_WRITE
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_READ:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_RDMA_WRITE);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":32
 *         return __pyx_enum.IBV_WC_SEND
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_WRITE:
 *         return __pyx_enum.IBV_WC_RDMA_WRITE             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_READ:
 *         return __pyx_enum.IBV_WC_RDMA_READ
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_RDMA_WRITE); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 32, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":31
 *     elif c_val == ibv_wc_opcode.IBV_WC_SEND:
 *         return __pyx_enum.IBV_WC_SEND
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_WRITE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RDMA_WRITE
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_READ:
*/
  }

  /* "EnumTypeToPy":33
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_WRITE:
 *         return __pyx_enum.IBV_WC_RDMA_WRITE
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_READ:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RDMA_READ
 *     elif c_val == ibv_wc_opcode.IBV_WC_COMP_SWAP:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_RDMA_READ);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":34
 *         return __pyx_enum.IBV_WC_RDMA_WRITE
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_READ:
 *         return __pyx_enum.IBV_WC_RDMA_READ             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_COMP_SWAP:
 *         return __pyx_enum.IBV_WC_COMP_SWAP
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_RDMA_READ); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":33
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_WRITE:
 *         return __pyx_enum.IBV_WC_RDMA_WRITE
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_READ:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RDMA_READ
 *     elif c_val == ibv_wc_opcode.IBV_WC_COMP_SWAP:
*/
  }

  /* "EnumTypeToPy":35
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_READ:
 *         return __pyx_enum.IBV_WC_RDMA_READ
 *     elif c_val == ibv_wc_opcode.IBV_WC_COMP_SWAP:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_COMP_SWAP
 *     elif c_val == ibv_wc_opcode.IBV_WC_FETCH_ADD:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_COMP_SWAP);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":36
 *         return __pyx_enum.IBV_WC_RDMA_READ
 *     elif c_val == ibv_wc_opcode.IBV_WC_COMP_SWAP:
 *         return __pyx_enum.IBV_WC_COMP_SWAP             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_FETCH_ADD:
 *         return __pyx_enum.IBV_WC_FETCH_ADD
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_COMP_SWAP); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":35
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_READ:
 *         return __pyx_enum.IBV_WC_RDMA_READ
 *     elif c_val == ibv_wc_opcode.IBV_WC_COMP_SWAP:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_COMP_SWAP
 *     elif c_val == ibv_wc_opcode.IBV_WC_FETCH_ADD:
*/
  }

  /* "EnumTypeToPy":37
 *     elif c_val == ibv_wc_opcode.IBV_WC_COMP_SWAP:
 *         return __pyx_enum.IBV_WC_COMP_SWAP
 *     elif c_val == ibv_wc_opcode.IBV_WC_FETCH_ADD:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_FETCH_ADD
 *     elif c_val == ibv_wc_opcode.IBV_WC_BIND_MW:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_FETCH_ADD);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":38
 *         return __pyx_enum.IBV_WC_COMP_SWAP
 *     elif c_val == ibv_wc_opcode.IBV_WC_FETCH_ADD:
 *         return __pyx_enum.IBV_WC_FETCH_ADD             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_BIND_MW:
 *         return __pyx_enum.IBV_WC_BIND_MW
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_FETCH_ADD); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":37
 *     elif c_val == ibv_wc_opcode.IBV_WC_COMP_SWAP:
 *         return __pyx_enum.IBV_WC_COMP_SWAP
 *     elif c_val == ibv_wc_opcode.IBV_WC_FETCH_ADD:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_FETCH_ADD
 *     elif c_val == ibv_wc_opcode.IBV_WC_BIND_MW:
*/
  }

  /* "EnumTypeToPy":39
 *     elif c_val == ibv_wc_opcode.IBV_WC_FETCH_ADD:
 *         return __pyx_enum.IBV_WC_FETCH_ADD
 *     elif c_val == ibv_wc_opcode.IBV_WC_BIND_MW:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_BIND_MW
 *     elif c_val == ibv_wc_opcode.IBV_WC_LOCAL_INV:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_BIND_MW);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":40
 *         return __pyx_enum.IBV_WC_FETCH_ADD
 *     elif c_val == ibv_wc_opcode.IBV_WC_BIND_MW:
 *         return __pyx_enum.IBV_WC_BIND_MW             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_LOCAL_INV:
 *         return __pyx_enum.IBV_WC_LOCAL_INV
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_BIND_MW); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":39
 *     elif c_val == ibv_wc_opcode.IBV_WC_FETCH_ADD:
 *         return __pyx_enum.IBV_WC_FETCH_ADD
 *     elif c_val == ibv_wc_opcode.IBV_WC_BIND_MW:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_BIND_MW
 *     elif c_val == ibv_wc_opcode.IBV_WC_LOCAL_INV:
*/
  }

  /* "EnumTypeToPy":41
 *     elif c_val == ibv_wc_opcode.IBV_WC_BIND_MW:
 *         return __pyx_enum.IBV_WC_BIND_MW
 *     elif c_val == ibv_wc_opcode.IBV_WC_LOCAL_INV:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOCAL_INV
 *     elif c_val == ibv_wc_opcode.IBV_WC_TSO:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_LOCAL_INV);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":42
 *         return __pyx_enum.IBV_WC_BIND_MW
 *     elif c_val == ibv_wc_opcode.IBV_WC_LOCAL_INV:
 *         return __pyx_enum.IBV_WC_LOCAL_INV             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_TSO:
 *         return __pyx_enum.IBV_WC_TSO
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_LOCAL_INV); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":41
 *     elif c_val == ibv_wc_opcode.IBV_WC_BIND_MW:
 *         return __pyx_enum.IBV_WC_BIND_MW
 *     elif c_val == ibv_wc_opcode.IBV_WC_LOCAL_INV:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOCAL_INV
 *     elif c_val == ibv_wc_opcode.IBV_WC_TSO:
*/
  }

  /* "EnumTypeToPy":43
 *     elif c_val == ibv_wc_opcode.IBV_WC_LOCAL_INV:
 *         return __pyx_enum.IBV_WC_LOCAL_INV
 *     elif c_val == ibv_wc_opcode.IBV_WC_TSO:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TSO
 *     elif c_val == ibv_wc_opcode.IBV_WC_FLUSH:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_TSO);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":44
 *         return __pyx_enum.IBV_WC_LOCAL_INV
 *     elif c_val == ibv_wc_opcode.IBV_WC_TSO:
 *         return __pyx_enum.IBV_WC_TSO             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_FLUSH:
 *         return __pyx_enum.IBV_WC_FLUSH
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_TSO); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":43
 *     elif c_val == ibv_wc_opcode.IBV_WC_LOCAL_INV:
 *         return __pyx_enum.IBV_WC_LOCAL_INV
 *     elif c_val == ibv_wc_opcode.IBV_WC_TSO:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TSO
 *     elif c_val == ibv_wc_opcode.IBV_WC_FLUSH:
*/
  }

  /* "EnumTypeToPy":45
 *     elif c_val == ibv_wc_opcode.IBV_WC_TSO:
 *         return __pyx_enum.IBV_WC_TSO
 *     elif c_val == ibv_wc_opcode.IBV_WC_FLUSH:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_FLUSH
 *     elif c_val == ibv_wc_opcode.IBV_WC_ATOMIC_WRITE:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_FLUSH);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":46
 *         return __pyx_enum.IBV_WC_TSO
 *     elif c_val == ibv_wc_opcode.IBV_WC_FLUSH:
 *         return __pyx_enum.IBV_WC_FLUSH             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_ATOMIC_WRITE:
 *         return __pyx_enum.IBV_WC_ATOMIC_WRITE
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_FLUSH); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 46, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":45
 *     elif c_val == ibv_wc_opcode.IBV_WC_TSO:
 *         return __pyx_enum.IBV_WC_TSO
 *     elif c_val == ibv_wc_opcode.IBV_WC_FLUSH:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_FLUSH
 *     elif c_val == ibv_wc_opcode.IBV_WC_ATOMIC_WRITE:
*/
  }

  /* "EnumTypeToPy":47
 *     elif c_val == ibv_wc_opcode.IBV_WC_FLUSH:
 *         return __pyx_enum.IBV_WC_FLUSH
 *     elif c_val == ibv_wc_opcode.IBV_WC_ATOMIC_WRITE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_ATOMIC_WRITE
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_ATOMIC_WRITE);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":48
 *         return __pyx_enum.IBV_WC_FLUSH
 *     elif c_val == ibv_wc_opcode.IBV_WC_ATOMIC_WRITE:
 *         return __pyx_enum.IBV_WC_ATOMIC_WRITE             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV:
 *         return __pyx_enum.IBV_WC_RECV
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_ATOMIC_WRITE); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":47
 *     elif c_val == ibv_wc_opcode.IBV_WC_FLUSH:
 *         return __pyx_enum.IBV_WC_FLUSH
 *     elif c_val == ibv_wc_opcode.IBV_WC_ATOMIC_WRITE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_ATOMIC_WRITE
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV:
*/
  }

  /* "EnumTypeToPy":49
 *     elif c_val == ibv_wc_opcode.IBV_WC_ATOMIC_WRITE:
 *         return __pyx_enum.IBV_WC_ATOMIC_WRITE
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RECV
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV_RDMA_WITH_IMM:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_RECV);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":50
 *         return __pyx_enum.IBV_WC_ATOMIC_WRITE
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV:
 *         return __pyx_enum.IBV_WC_RECV             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV_RDMA_WITH_IMM:
 *         return __pyx_enum.IBV_WC_RECV_RDMA_WITH_IMM
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_RECV); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":49
 *     elif c_val == ibv_wc_opcode.IBV_WC_ATOMIC_WRITE:
 *         return __pyx_enum.IBV_WC_ATOMIC_WRITE
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RECV
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV_RDMA_WITH_IMM:
*/
  }

  /* "EnumTypeToPy":51
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV:
 *         return __pyx_enum.IBV_WC_RECV
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV_RDMA_WITH_IMM:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RECV_RDMA_WITH_IMM
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_ADD:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_RECV_RDMA_WITH_IMM);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":52
 *         return __pyx_enum.IBV_WC_RECV
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV_RDMA_WITH_IMM:
 *         return __pyx_enum.IBV_WC_RECV_RDMA_WITH_IMM             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_ADD:
 *         return __pyx_enum.IBV_WC_TM_ADD
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_RECV_RDMA_WITH_IMM); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":51
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV:
 *         return __pyx_enum.IBV_WC_RECV
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV_RDMA_WITH_IMM:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RECV_RDMA_WITH_IMM
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_ADD:
*/
  }

  /* "EnumTypeToPy":53
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV_RDMA_WITH_IMM:
 *         return __pyx_enum.IBV_WC_RECV_RDMA_WITH_IMM
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_ADD:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_ADD
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_DEL:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_TM_ADD);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":54
 *         return __pyx_enum.IBV_WC_RECV_RDMA_WITH_IMM
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_ADD:
 *         return __pyx_enum.IBV_WC_TM_ADD             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_DEL:
 *         return __pyx_enum.IBV_WC_TM_DEL
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_ADD); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":53
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV_RDMA_WITH_IMM:
 *         return __pyx_enum.IBV_WC_RECV_RDMA_WITH_IMM
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_ADD:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_ADD
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_DEL:
*/
  }

  /* "EnumTypeToPy":55
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_ADD:
 *         return __pyx_enum.IBV_WC_TM_ADD
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_DEL:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_DEL
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_SYNC:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_TM_DEL);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":56
 *         return __pyx_enum.IBV_WC_TM_ADD
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_DEL:
 *         return __pyx_enum.IBV_WC_TM_DEL             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_SYNC:
 *         return __pyx_enum.IBV_WC_TM_SYNC
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_DEL); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":55
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_ADD:
 *         return __pyx_enum.IBV_WC_TM_ADD
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_DEL:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_DEL
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_SYNC:
*/
  }

  /* "EnumTypeToPy":57
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_DEL:
 *         return __pyx_enum.IBV_WC_TM_DEL
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_SYNC:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_SYNC
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_RECV:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_TM_SYNC);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":58
 *         return __pyx_enum.IBV_WC_TM_DEL
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_SYNC:
 *         return __pyx_enum.IBV_WC_TM_SYNC             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_RECV:
 *         return __pyx_enum.IBV_WC_TM_RECV
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_SYNC); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":57
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_DEL:
 *         return __pyx_enum.IBV_WC_TM_DEL
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_SYNC:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_SYNC
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_RECV:
*/
  }

  /* "EnumTypeToPy":59
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_SYNC:
 *         return __pyx_enum.IBV_WC_TM_SYNC
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_RECV:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_RECV
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_NO_TAG:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_TM_RECV);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":60
 *         return __pyx_enum.IBV_WC_TM_SYNC
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_RECV:
 *         return __pyx_enum.IBV_WC_TM_RECV             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_NO_TAG:
 *         return __pyx_enum.IBV_WC_TM_NO_TAG
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_RECV); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":59
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_SYNC:
 *         return __pyx_enum.IBV_WC_TM_SYNC
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_RECV:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_RECV
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_NO_TAG:
*/
  }

  /* "EnumTypeToPy":61
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_RECV:
 *         return __pyx_enum.IBV_WC_TM_RECV
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_NO_TAG:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_NO_TAG
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER2:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_TM_NO_TAG);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":62
 *         return __pyx_enum.IBV_WC_TM_RECV
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_NO_TAG:
 *         return __pyx_enum.IBV_WC_TM_NO_TAG             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER2:
 *         return __pyx_enum.IBV_WC_DRIVER2
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_NO_TAG); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":61
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_RECV:
 *         return __pyx_enum.IBV_WC_TM_RECV
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_NO_TAG:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_NO_TAG
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER2:
*/
  }

  /* "EnumTypeToPy":63
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_NO_TAG:
 *         return __pyx_enum.IBV_WC_TM_NO_TAG
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER2:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_DRIVER2
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER3:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_DRIVER2);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":64
 *         return __pyx_enum.IBV_WC_TM_NO_TAG
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER2:
 *         return __pyx_enum.IBV_WC_DRIVER2             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER3:
 *         return __pyx_enum.IBV_WC_DRIVER3
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_DRIVER2); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":63
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_NO_TAG:
 *         return __pyx_enum.IBV_WC_TM_NO_TAG
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER2:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_DRIVER2
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER3:
*/
  }

  /* "EnumTypeToPy":65
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER2:
 *         return __pyx_enum.IBV_WC_DRIVER2
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER3:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_DRIVER3
 *     else:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_DRIVER3);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":66
 *         return __pyx_enum.IBV_WC_DRIVER2
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER3:
 *         return __pyx_enum.IBV_WC_DRIVER3             # <<<<<<<<<<<<<<
 *     else:
 *         underlying_c_val = <int>c_val
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_DRIVER3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":65
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER2:
 *         return __pyx_enum.IBV_WC_DRIVER2
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER3:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_DRIVER3
 *     else:
*/
  }

  /* "EnumTypeToPy":68
 *         return __pyx_enum.IBV_WC_DRIVER3
 *     else:
 *         underlying_c_val = <int>c_val             # <<<<<<<<<<<<<<
 *         return __pyx_enum(underlying_c_val)
*/
  /*else*/ {
    __pyx_v_underlying_c_val = ((int)__pyx_v_c_val);

    /* "EnumTypeToPy":69
 *     else:
 *         underlying_c_val = <int>c_val
 *         return __pyx_enum(underlying_c_val)             # <<<<<<<<<<<<<<
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = NULL;
    __Pyx_INCREF(__pyx_v___pyx_enum);
    __pyx_t_5 = __pyx_v___pyx_enum; 
    __pyx_t_11 = __Pyx_PyLong_From_int(__pyx_v_underlying_c_val); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_13 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_8);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_13 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_11};
      __pyx_t_10 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 69, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
    }
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;
  }

  /* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py(ibv_wc_opcode c_val):
 *     cdef object __pyx_enum
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_enum);
  __Pyx_XDECREF(__pyx_v_warnings);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py(enum ibv_create_cq_wc_flags __pyx_v_c_val) {
  PyObject *__pyx_v___pyx_enum = 0;
  PyObject *__pyx_v_warnings = NULL;
  int __pyx_v_underlying_c_val;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  size_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py", 0);

  /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_create_cq_wc_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py == Py_None);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_create_cq_wc_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py
 *         except ImportError:
*/
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "EnumTypeToPy":13
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py is None:
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_create_cq_wc_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py             # <<<<<<<<<<<<<<
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py = False  # False indicates "don't try again"
*/
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_create_cq_wc_flags};
          __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyverbs_libibverbs_enums, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L4_error)
        }
        __pyx_t_5 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_5);
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_create_cq_wc_flags};
          __pyx_t_7 = 0; {
            __pyx_t_8 = __Pyx_ImportFrom(__pyx_t_5, __pyx_imported_names[__pyx_t_7]); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 13, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_8);
            switch (__pyx_t_7) {
              case 0:
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py);
              __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py, __pyx_t_8);
              __Pyx_GIVEREF(__pyx_t_8);
              break;
              default:;
            }
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_create_cq_wc_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py
 *         except ImportError:
*/
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "EnumTypeToPy":14
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_create_cq_wc_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py
 *         except ImportError:             # <<<<<<<<<<<<<<
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py = False  # False indicates "don't try again"
 *             import warnings
*/
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_ImportError))));
      if (__pyx_t_9) {
        __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_8, &__pyx_t_10) < 0) __PYX_ERR(1, 14, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_10);

        /* "EnumTypeToPy":15
 *             from pyverbs.libibverbs_enums import ibv_create_cq_wc_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py = False  # False indicates "don't try again"             # <<<<<<<<<<<<<<
 *             import warnings
 *             warnings.warn(
*/
        __Pyx_INCREF(Py_False);
        __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py);
        __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py, Py_False);
        __Pyx_GIVEREF(Py_False);

        /* "EnumTypeToPy":16
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py = False  # False indicates "don't try again"
 *             import warnings             # <<<<<<<<<<<<<<
 *             warnings.warn(
 *                 f"enum class ibv_create_cq_wc_flags not importable from pyverbs.libibverbs_enums. "
*/
        __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_warnings, 0, 0, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 16, __pyx_L6_except_error)
        __pyx_t_11 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_v_warnings = __pyx_t_11;
        __pyx_t_11 = 0;

        /* "EnumTypeToPy":17
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py = False  # False indicates "don't try again"
 *             import warnings
 *             warnings.warn(             # <<<<<<<<<<<<<<
 *                 f"enum class ibv_create_cq_wc_flags not importable from pyverbs.libibverbs_enums. "
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
*/
        __pyx_t_12 = __pyx_v_warnings;
        __Pyx_INCREF(__pyx_t_12);
        __pyx_t_13 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_mstate_global->__pyx_kp_u_enum_class_ibv_create_cq_wc_flag};
          __pyx_t_11 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_warn, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 17, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_11);
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L5_exception_handled;
      }
      goto __pyx_L6_except_error;

      /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_create_cq_wc_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py
 *         except ImportError:
*/
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L5_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_L9_try_end:;
    }

    /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_create_cq_wc_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py
*/
  }

  /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py == Py_False);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":24
 * 
 * 
 *         return <int>c_val             # <<<<<<<<<<<<<<
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyLong_From_int(((int)__pyx_v_c_val)); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumTypeToPy":25
 * 
 *         return <int>c_val
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py             # <<<<<<<<<<<<<<
 * 
 *     if 0:
*/
  __Pyx_INCREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py);
  __pyx_v___pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py;

  /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_BYTE_LEN:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_BYTE_LEN
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_IMM:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_EX_WITH_BYTE_LEN);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":30
 *         pass
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_BYTE_LEN:
 *         return __pyx_enum.IBV_WC_EX_WITH_BYTE_LEN             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_IMM:
 *         return __pyx_enum.IBV_WC_EX_WITH_IMM
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_BYTE_LEN); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_BYTE_LEN:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_BYTE_LEN
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_IMM:
*/
  }

  /* "EnumTypeToPy":31
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_BYTE_LEN:
 *         return __pyx_enum.IBV_WC_EX_WITH_BYTE_LEN
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_IMM:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_IMM
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_QP_NUM:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_EX_WITH_IMM);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":32
 *         return __pyx_enum.IBV_WC_EX_WITH_BYTE_LEN
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_IMM:
 *         return __pyx_enum.IBV_WC_EX_WITH_IMM             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_QP_NUM:
 *         return __pyx_enum.IBV_WC_EX_WITH_QP_NUM
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_IMM); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 32, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":31
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_BYTE_LEN:
 *         return __pyx_enum.IBV_WC_EX_WITH_BYTE_LEN
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_IMM:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_IMM
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_QP_NUM:
*/
  }

  /* "EnumTypeToPy":33
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_IMM:
 *         return __pyx_enum.IBV_WC_EX_WITH_IMM
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_QP_NUM:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_QP_NUM
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SRC_QP:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_EX_WITH_QP_NUM);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":34
 *         return __pyx_enum.IBV_WC_EX_WITH_IMM
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_QP_NUM:
 *         return __pyx_enum.IBV_WC_EX_WITH_QP_NUM             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SRC_QP:
 *         return __pyx_enum.IBV_WC_EX_WITH_SRC_QP
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_QP_NUM); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":33
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_IMM:
 *         return __pyx_enum.IBV_WC_EX_WITH_IMM
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_QP_NUM:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_QP_NUM
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SRC_QP:
*/
  }

  /* "EnumTypeToPy":35
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_QP_NUM:
 *         return __pyx_enum.IBV_WC_EX_WITH_QP_NUM
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SRC_QP:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_SRC_QP
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SLID:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_EX_WITH_SRC_QP);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":36
 *         return __pyx_enum.IBV_WC_EX_WITH_QP_NUM
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SRC_QP:
 *         return __pyx_enum.IBV_WC_EX_WITH_SRC_QP             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SLID:
 *         return __pyx_enum.IBV_WC_EX_WITH_SLID
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_SRC_QP); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":35
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_QP_NUM:
 *         return __pyx_enum.IBV_WC_EX_WITH_QP_NUM
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SRC_QP:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_SRC_QP
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SLID:
*/
  }

  /* "EnumTypeToPy":37
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SRC_QP:
 *         return __pyx_enum.IBV_WC_EX_WITH_SRC_QP
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SLID:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_SLID
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SL:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_EX_WITH_SLID);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":38
 *         return __pyx_enum.IBV_WC_EX_WITH_SRC_QP
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SLID:
 *         return __pyx_enum.IBV_WC_EX_WITH_SLID             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SL:
 *         return __pyx_enum.IBV_WC_EX_WITH_SL
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_SLID); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":37
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SRC_QP:
 *         return __pyx_enum.IBV_WC_EX_WITH_SRC_QP
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SLID:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_SLID
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SL:
*/
  }

  /* "EnumTypeToPy":39
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SLID:
 *         return __pyx_enum.IBV_WC_EX_WITH_SLID
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SL:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_SL
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_DLID_PATH_BITS:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_EX_WITH_SL);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":40
 *         return __pyx_enum.IBV_WC_EX_WITH_SLID
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SL:
 *         return __pyx_enum.IBV_WC_EX_WITH_SL             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_DLID_PATH_BITS:
 *         return __pyx_enum.IBV_WC_EX_WITH_DLID_PATH_BITS
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_SL); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":39
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SLID:
 *         return __pyx_enum.IBV_WC_EX_WITH_SLID
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SL:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_SL
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_DLID_PATH_BITS:
*/
  }

  /* "EnumTypeToPy":41
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SL:
 *         return __pyx_enum.IBV_WC_EX_WITH_SL
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_DLID_PATH_BITS:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_DLID_PATH_BITS
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_EX_WITH_DLID_PATH_BITS);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":42
 *         return __pyx_enum.IBV_WC_EX_WITH_SL
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_DLID_PATH_BITS:
 *         return __pyx_enum.IBV_WC_EX_WITH_DLID_PATH_BITS             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP:
 *         return __pyx_enum.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_DLID_PATH_BITS); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":41
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_SL:
 *         return __pyx_enum.IBV_WC_EX_WITH_SL
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_DLID_PATH_BITS:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_DLID_PATH_BITS
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP:
*/
  }

  /* "EnumTypeToPy":43
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_DLID_PATH_BITS:
 *         return __pyx_enum.IBV_WC_EX_WITH_DLID_PATH_BITS
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_CVLAN:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_EX_WITH_COMPLETION_TIMESTAMP);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":44
 *         return __pyx_enum.IBV_WC_EX_WITH_DLID_PATH_BITS
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP:
 *         return __pyx_enum.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_CVLAN:
 *         return __pyx_enum.IBV_WC_EX_WITH_CVLAN
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_COMPLETION_TIMEST); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":43
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_DLID_PATH_BITS:
 *         return __pyx_enum.IBV_WC_EX_WITH_DLID_PATH_BITS
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_CVLAN:
*/
  }

  /* "EnumTypeToPy":45
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP:
 *         return __pyx_enum.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_CVLAN:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_CVLAN
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_FLOW_TAG:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_EX_WITH_CVLAN);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":46
 *         return __pyx_enum.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_CVLAN:
 *         return __pyx_enum.IBV_WC_EX_WITH_CVLAN             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_FLOW_TAG:
 *         return __pyx_enum.IBV_WC_EX_WITH_FLOW_TAG
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_CVLAN); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 46, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":45
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP:
 *         return __pyx_enum.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_CVLAN:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_CVLAN
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_FLOW_TAG:
*/
  }

  /* "EnumTypeToPy":47
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_CVLAN:
 *         return __pyx_enum.IBV_WC_EX_WITH_CVLAN
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_FLOW_TAG:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_FLOW_TAG
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_TM_INFO:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_EX_WITH_FLOW_TAG);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":48
 *         return __pyx_enum.IBV_WC_EX_WITH_CVLAN
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_FLOW_TAG:
 *         return __pyx_enum.IBV_WC_EX_WITH_FLOW_TAG             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_TM_INFO:
 *         return __pyx_enum.IBV_WC_EX_WITH_TM_INFO
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_FLOW_TAG); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":47
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_CVLAN:
 *         return __pyx_enum.IBV_WC_EX_WITH_CVLAN
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_FLOW_TAG:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_FLOW_TAG
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_TM_INFO:
*/
  }

  /* "EnumTypeToPy":49
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_FLOW_TAG:
 *         return __pyx_enum.IBV_WC_EX_WITH_FLOW_TAG
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_TM_INFO:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_TM_INFO
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_EX_WITH_TM_INFO);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":50
 *         return __pyx_enum.IBV_WC_EX_WITH_FLOW_TAG
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_TM_INFO:
 *         return __pyx_enum.IBV_WC_EX_WITH_TM_INFO             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK:
 *         return __pyx_enum.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_TM_INFO); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":49
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_FLOW_TAG:
 *         return __pyx_enum.IBV_WC_EX_WITH_FLOW_TAG
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_TM_INFO:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_TM_INFO
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK:
*/
  }

  /* "EnumTypeToPy":51
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_TM_INFO:
 *         return __pyx_enum.IBV_WC_EX_WITH_TM_INFO
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK
 *     else:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":52
 *         return __pyx_enum.IBV_WC_EX_WITH_TM_INFO
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK:
 *         return __pyx_enum.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK             # <<<<<<<<<<<<<<
 *     else:
 *         underlying_c_val = <int>c_val
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_COMPLETION_TIMEST_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":51
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_TM_INFO:
 *         return __pyx_enum.IBV_WC_EX_WITH_TM_INFO
 *     elif c_val == ibv_create_cq_wc_flags.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK
 *     else:
*/
  }

  /* "EnumTypeToPy":54
 *         return __pyx_enum.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK
 *     else:
 *         underlying_c_val = <int>c_val             # <<<<<<<<<<<<<<
 *         return __pyx_enum(underlying_c_val)
*/
  /*else*/ {
    __pyx_v_underlying_c_val = ((int)__pyx_v_c_val);

    /* "EnumTypeToPy":55
 *     else:
 *         underlying_c_val = <int>c_val
 *         return __pyx_enum(underlying_c_val)             # <<<<<<<<<<<<<<
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = NULL;
    __Pyx_INCREF(__pyx_v___pyx_enum);
    __pyx_t_5 = __pyx_v___pyx_enum; 
    __pyx_t_11 = __Pyx_PyLong_From_int(__pyx_v_underlying_c_val); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_13 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_8);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_13 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_11};
      __pyx_t_10 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 55, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
    }
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;
  }

  /* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py(ibv_create_cq_wc_flags c_val):
 *     cdef object __pyx_enum
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_enum);
  __Pyx_XDECREF(__pyx_v_warnings);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":24
 *     CQ.
 *     """
 *     def __init__(self, Context context not None):             # <<<<<<<<<<<<<<
 *         """
 *         Initializes a completion channel object on the given device.
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_11CompChannel_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2cq_11CompChannel___init__, "\n        Initializes a completion channel object on the given device.\n        :param context: The device's context to use\n        :return: A CompChannel object on success\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_7pyverbs_2cq_11CompChannel___init__;
#endif
static int __pyx_pw_7pyverbs_2cq_11CompChannel_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7pyverbs_6device_Context *__pyx_v_context = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_context,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 24, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 24, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 24, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, i); __PYX_ERR(0, 24, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 24, __pyx_L3_error)
    }
    __pyx_v_context = ((struct __pyx_obj_7pyverbs_6device_Context *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 24, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cq.CompChannel.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_context), __pyx_mstate_global->__pyx_ptype_7pyverbs_6device_Context, 0, "context", 0))) __PYX_ERR(0, 24, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2cq_11CompChannel___init__(((struct __pyx_obj_7pyverbs_2cq_CompChannel *)__pyx_v_self), __pyx_v_context);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_11CompChannel___init__(struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_self, struct __pyx_obj_7pyverbs_6device_Context *__pyx_v_context) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyverbs/cq.pyx":30
 *         :return: A CompChannel object on success
 *         """
 *         super().__init__()             # <<<<<<<<<<<<<<
 *         self.cc = v.ibv_create_comp_channel(context.context)
 *         if self.cc == NULL:
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CompChannel), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":31
 *         """
 *         super().__init__()
 *         self.cc = v.ibv_create_comp_channel(context.context)             # <<<<<<<<<<<<<<
 *         if self.cc == NULL:
 *             raise PyverbsRDMAErrno('Failed to create a completion channel')
*/
  __pyx_v_self->cc = ibv_create_comp_channel(__pyx_v_context->context);

  /* "pyverbs/cq.pyx":32
 *         super().__init__()
 *         self.cc = v.ibv_create_comp_channel(context.context)
 *         if self.cc == NULL:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to create a completion channel')
 *         self.context = context
*/
  __pyx_t_6 = (__pyx_v_self->cc == NULL);
  if (unlikely(__pyx_t_6)) {

    /* "pyverbs/cq.pyx":33
 *         self.cc = v.ibv_create_comp_channel(context.context)
 *         if self.cc == NULL:
 *             raise PyverbsRDMAErrno('Failed to create a completion channel')             # <<<<<<<<<<<<<<
 *         self.context = context
 *         context.add_ref(self)
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Failed_to_create_a_completion_ch};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 33, __pyx_L1_error)

    /* "pyverbs/cq.pyx":32
 *         super().__init__()
 *         self.cc = v.ibv_create_comp_channel(context.context)
 *         if self.cc == NULL:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to create a completion channel')
 *         self.context = context
*/
  }

  /* "pyverbs/cq.pyx":34
 *         if self.cc == NULL:
 *             raise PyverbsRDMAErrno('Failed to create a completion channel')
 *         self.context = context             # <<<<<<<<<<<<<<
 *         context.add_ref(self)
 *         self.cqs = weakref.WeakSet()
*/
  __Pyx_INCREF((PyObject *)__pyx_v_context);
  __Pyx_GIVEREF((PyObject *)__pyx_v_context);
  __Pyx_GOTREF(__pyx_v_self->context);
  __Pyx_DECREF(__pyx_v_self->context);
  __pyx_v_self->context = ((PyObject *)__pyx_v_context);

  /* "pyverbs/cq.pyx":35
 *             raise PyverbsRDMAErrno('Failed to create a completion channel')
 *         self.context = context
 *         context.add_ref(self)             # <<<<<<<<<<<<<<
 *         self.cqs = weakref.WeakSet()
 *         self.logger.debug('Created a Completion Channel')
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_7pyverbs_6device_Context *)__pyx_v_context->__pyx_base.__pyx_vtab)->add_ref(__pyx_v_context, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":36
 *         self.context = context
 *         context.add_ref(self)
 *         self.cqs = weakref.WeakSet()             # <<<<<<<<<<<<<<
 *         self.logger.debug('Created a Completion Channel')
 * 
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_weakref); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_WeakSet); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->cqs);
  __Pyx_DECREF(__pyx_v_self->cqs);
  __pyx_v_self->cqs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":37
 *         context.add_ref(self)
 *         self.cqs = weakref.WeakSet()
 *         self.logger.debug('Created a Completion Channel')             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
*/
  __pyx_t_4 = __pyx_v_self->__pyx_base.__pyx_base.logger;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_Created_a_Completion_Channel};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_debug, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":24
 *     CQ.
 *     """
 *     def __init__(self, Context context not None):             # <<<<<<<<<<<<<<
 *         """
 *         Initializes a completion channel object on the given device.
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cq.CompChannel.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":39
 *         self.logger.debug('Created a Completion Channel')
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         self.close()
 * 
*/

/* Python wrapper */
static void __pyx_pw_7pyverbs_2cq_11CompChannel_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7pyverbs_2cq_11CompChannel_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_7pyverbs_2cq_11CompChannel_2__dealloc__(((struct __pyx_obj_7pyverbs_2cq_CompChannel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7pyverbs_2cq_11CompChannel_2__dealloc__(struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "pyverbs/cq.pyx":40
 * 
 *     def __dealloc__(self):
 *         self.close()             # <<<<<<<<<<<<<<
 * 
 *     cpdef close(self):
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_7pyverbs_2cq_CompChannel *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.close(((struct __pyx_obj_7pyverbs_4base_PyverbsCM *)__pyx_v_self), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":39
 *         self.logger.debug('Created a Completion Channel')
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         self.close()
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("pyverbs.cq.CompChannel.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "pyverbs/cq.pyx":42
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.cc != NULL:
 *             if self.logger:
*/

static PyObject *__pyx_pw_7pyverbs_2cq_11CompChannel_5close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7pyverbs_2cq_11CompChannel_close(struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_self, int __pyx_skip_dispatch) {
  int __pyx_v_rc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CompChannel &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_close); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 42, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyverbs_2cq_11CompChannel_5close)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 42, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyverbs/cq.pyx":43
 * 
 *     cpdef close(self):
 *         if self.cc != NULL:             # <<<<<<<<<<<<<<
 *             if self.logger:
 *                 self.logger.debug('Closing completion channel')
*/
  __pyx_t_6 = (__pyx_v_self->cc != NULL);
  if (__pyx_t_6) {

    /* "pyverbs/cq.pyx":44
 *     cpdef close(self):
 *         if self.cc != NULL:
 *             if self.logger:             # <<<<<<<<<<<<<<
 *                 self.logger.debug('Closing completion channel')
 *             close_weakrefs([self.cqs])
*/
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_self->__pyx_base.__pyx_base.logger); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 44, __pyx_L1_error)
    if (__pyx_t_6) {

      /* "pyverbs/cq.pyx":45
 *         if self.cc != NULL:
 *             if self.logger:
 *                 self.logger.debug('Closing completion channel')             # <<<<<<<<<<<<<<
 *             close_weakrefs([self.cqs])
 *             rc = v.ibv_destroy_comp_channel(self.cc)
*/
      __pyx_t_2 = __pyx_v_self->__pyx_base.__pyx_base.logger;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_5 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Closing_completion_channel};
        __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_debug, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 45, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyverbs/cq.pyx":44
 *     cpdef close(self):
 *         if self.cc != NULL:
 *             if self.logger:             # <<<<<<<<<<<<<<
 *                 self.logger.debug('Closing completion channel')
 *             close_weakrefs([self.cqs])
*/
    }

    /* "pyverbs/cq.pyx":46
 *             if self.logger:
 *                 self.logger.debug('Closing completion channel')
 *             close_weakrefs([self.cqs])             # <<<<<<<<<<<<<<
 *             rc = v.ibv_destroy_comp_channel(self.cc)
 *             if rc != 0:
*/
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_self->cqs);
    __Pyx_GIVEREF(__pyx_v_self->cqs);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->cqs) != (0)) __PYX_ERR(0, 46, __pyx_L1_error);
    __pyx_t_2 = __pyx_f_7pyverbs_4base_close_weakrefs(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 46, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyverbs/cq.pyx":47
 *                 self.logger.debug('Closing completion channel')
 *             close_weakrefs([self.cqs])
 *             rc = v.ibv_destroy_comp_channel(self.cc)             # <<<<<<<<<<<<<<
 *             if rc != 0:
 *                 raise PyverbsRDMAError('Failed to destroy a completion channel',
*/
    __pyx_v_rc = ibv_destroy_comp_channel(__pyx_v_self->cc);

    /* "pyverbs/cq.pyx":48
 *             close_weakrefs([self.cqs])
 *             rc = v.ibv_destroy_comp_channel(self.cc)
 *             if rc != 0:             # <<<<<<<<<<<<<<
 *                 raise PyverbsRDMAError('Failed to destroy a completion channel',
 *                                        rc)
*/
    __pyx_t_6 = (__pyx_v_rc != 0);
    if (unlikely(__pyx_t_6)) {

      /* "pyverbs/cq.pyx":49
 *             rc = v.ibv_destroy_comp_channel(self.cc)
 *             if rc != 0:
 *                 raise PyverbsRDMAError('Failed to destroy a completion channel',             # <<<<<<<<<<<<<<
 *                                        rc)
 *             self.cc = NULL
*/
      __pyx_t_1 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 49, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "pyverbs/cq.pyx":50
 *             if rc != 0:
 *                 raise PyverbsRDMAError('Failed to destroy a completion channel',
 *                                        rc)             # <<<<<<<<<<<<<<
 *             self.cc = NULL
 * 
*/
      __pyx_t_3 = __Pyx_PyLong_From_int(__pyx_v_rc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 50, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
        assert(__pyx_t_1);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
        __pyx_t_5 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_mstate_global->__pyx_kp_u_Failed_to_destroy_a_completion_c, __pyx_t_3};
        __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 49, __pyx_L1_error)

      /* "pyverbs/cq.pyx":48
 *             close_weakrefs([self.cqs])
 *             rc = v.ibv_destroy_comp_channel(self.cc)
 *             if rc != 0:             # <<<<<<<<<<<<<<
 *                 raise PyverbsRDMAError('Failed to destroy a completion channel',
 *                                        rc)
*/
    }

    /* "pyverbs/cq.pyx":51
 *                 raise PyverbsRDMAError('Failed to destroy a completion channel',
 *                                        rc)
 *             self.cc = NULL             # <<<<<<<<<<<<<<
 * 
 *     def get_cq_event(self, CQ expected_cq):
*/
    __pyx_v_self->cc = NULL;

    /* "pyverbs/cq.pyx":43
 * 
 *     cpdef close(self):
 *         if self.cc != NULL:             # <<<<<<<<<<<<<<
 *             if self.logger:
 *                 self.logger.debug('Closing completion channel')
*/
  }

  /* "pyverbs/cq.pyx":42
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.cc != NULL:
 *             if self.logger:
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cq.CompChannel.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_11CompChannel_5close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_11CompChannel_5close = {"close", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_11CompChannel_5close, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_11CompChannel_5close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("close", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("close", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_11CompChannel_4close(((struct __pyx_obj_7pyverbs_2cq_CompChannel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_11CompChannel_4close(struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7pyverbs_2cq_11CompChannel_close(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CompChannel.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":53
 *             self.cc = NULL
 * 
 *     def get_cq_event(self, CQ expected_cq):             # <<<<<<<<<<<<<<
 *         """
 *         Waits for the next completion event in the completion event channel
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_11CompChannel_7get_cq_event(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2cq_11CompChannel_6get_cq_event, "\n        Waits for the next completion event in the completion event channel\n        :param expected_cq: The CQ that is expected to get the event\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_2cq_11CompChannel_7get_cq_event = {"get_cq_event", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_11CompChannel_7get_cq_event, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2cq_11CompChannel_6get_cq_event};
static PyObject *__pyx_pw_7pyverbs_2cq_11CompChannel_7get_cq_event(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_expected_cq = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_cq_event (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_expected_cq,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 53, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 53, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_cq_event", 0) < (0)) __PYX_ERR(0, 53, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get_cq_event", 1, 1, 1, i); __PYX_ERR(0, 53, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 53, __pyx_L3_error)
    }
    __pyx_v_expected_cq = ((struct __pyx_obj_7pyverbs_2cq_CQ *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_cq_event", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 53, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cq.CompChannel.get_cq_event", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_expected_cq), __pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQ, 1, "expected_cq", 0))) __PYX_ERR(0, 53, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2cq_11CompChannel_6get_cq_event(((struct __pyx_obj_7pyverbs_2cq_CompChannel *)__pyx_v_self), __pyx_v_expected_cq);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_11CompChannel_6get_cq_event(struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_self, struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_expected_cq) {
  struct ibv_cq *__pyx_v_cq;
  void *__pyx_v_ctx;
  int __pyx_v_rc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_cq_event", 0);

  /* "pyverbs/cq.pyx":62
 *         cdef void *ctx
 * 
 *         rc = v.ibv_get_cq_event(self.cc, &cq, &ctx)             # <<<<<<<<<<<<<<
 *         if rc != 0:
 *             raise PyverbsRDMAErrno('Failed to get CQ event')
*/
  __pyx_v_rc = ibv_get_cq_event(__pyx_v_self->cc, (&__pyx_v_cq), (&__pyx_v_ctx));

  /* "pyverbs/cq.pyx":63
 * 
 *         rc = v.ibv_get_cq_event(self.cc, &cq, &ctx)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to get CQ event')
 *         if cq != expected_cq.cq:
*/
  __pyx_t_1 = (__pyx_v_rc != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pyverbs/cq.pyx":64
 *         rc = v.ibv_get_cq_event(self.cc, &cq, &ctx)
 *         if rc != 0:
 *             raise PyverbsRDMAErrno('Failed to get CQ event')             # <<<<<<<<<<<<<<
 *         if cq != expected_cq.cq:
 *             raise PyverbsRDMAErrno('Received event on an unexpected CQ')
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Failed_to_get_CQ_event};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 64, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 64, __pyx_L1_error)

    /* "pyverbs/cq.pyx":63
 * 
 *         rc = v.ibv_get_cq_event(self.cc, &cq, &ctx)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to get CQ event')
 *         if cq != expected_cq.cq:
*/
  }

  /* "pyverbs/cq.pyx":65
 *         if rc != 0:
 *             raise PyverbsRDMAErrno('Failed to get CQ event')
 *         if cq != expected_cq.cq:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Received event on an unexpected CQ')
 *         expected_cq.num_events += 1
*/
  __pyx_t_1 = (__pyx_v_cq != __pyx_v_expected_cq->cq);
  if (unlikely(__pyx_t_1)) {

    /* "pyverbs/cq.pyx":66
 *             raise PyverbsRDMAErrno('Failed to get CQ event')
 *         if cq != expected_cq.cq:
 *             raise PyverbsRDMAErrno('Received event on an unexpected CQ')             # <<<<<<<<<<<<<<
 *         expected_cq.num_events += 1
 * 
*/
    __pyx_t_4 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_Received_event_on_an_unexpected};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 66, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 66, __pyx_L1_error)

    /* "pyverbs/cq.pyx":65
 *         if rc != 0:
 *             raise PyverbsRDMAErrno('Failed to get CQ event')
 *         if cq != expected_cq.cq:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Received event on an unexpected CQ')
 *         expected_cq.num_events += 1
*/
  }

  /* "pyverbs/cq.pyx":67
 *         if cq != expected_cq.cq:
 *             raise PyverbsRDMAErrno('Received event on an unexpected CQ')
 *         expected_cq.num_events += 1             # <<<<<<<<<<<<<<
 * 
 *     cdef add_ref(self, obj):
*/
  __pyx_t_2 = __Pyx_PyLong_AddObjC(__pyx_v_expected_cq->num_events, __pyx_mstate_global->__pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_expected_cq->num_events);
  __Pyx_DECREF(__pyx_v_expected_cq->num_events);
  __pyx_v_expected_cq->num_events = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":53
 *             self.cc = NULL
 * 
 *     def get_cq_event(self, CQ expected_cq):             # <<<<<<<<<<<<<<
 *         """
 *         Waits for the next completion event in the completion event channel
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cq.CompChannel.get_cq_event", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":69
 *         expected_cq.num_events += 1
 * 
 *     cdef add_ref(self, obj):             # <<<<<<<<<<<<<<
 *         if isinstance(obj, CQ) or isinstance(obj, CQEX):
 *             self.cqs.add(obj)
*/

static PyObject *__pyx_f_7pyverbs_2cq_11CompChannel_add_ref(struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_ref", 0);

  /* "pyverbs/cq.pyx":70
 * 
 *     cdef add_ref(self, obj):
 *         if isinstance(obj, CQ) or isinstance(obj, CQEX):             # <<<<<<<<<<<<<<
 *             self.cqs.add(obj)
 * 
*/
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQ); 
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX); 
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "pyverbs/cq.pyx":71
 *     cdef add_ref(self, obj):
 *         if isinstance(obj, CQ) or isinstance(obj, CQEX):
 *             self.cqs.add(obj)             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
    __pyx_t_4 = __pyx_v_self->cqs;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_obj};
      __pyx_t_3 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_add, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 71, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyverbs/cq.pyx":70
 * 
 *     cdef add_ref(self, obj):
 *         if isinstance(obj, CQ) or isinstance(obj, CQEX):             # <<<<<<<<<<<<<<
 *             self.cqs.add(obj)
 * 
*/
  }

  /* "pyverbs/cq.pyx":69
 *         expected_cq.num_events += 1
 * 
 *     cdef add_ref(self, obj):             # <<<<<<<<<<<<<<
 *         if isinstance(obj, CQ) or isinstance(obj, CQEX):
 *             self.cqs.add(obj)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cq.CompChannel.add_ref", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":73
 *             self.cqs.add(obj)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def fd(self):
 *         return self.cc.fd
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_11CompChannel_2fd_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_11CompChannel_2fd_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_11CompChannel_2fd___get__(((struct __pyx_obj_7pyverbs_2cq_CompChannel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_11CompChannel_2fd___get__(struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":75
 *     @property
 *     def fd(self):
 *         return self.cc.fd             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->cc->fd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":73
 *             self.cqs.add(obj)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def fd(self):
 *         return self.cc.fd
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CompChannel.fd.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.cc cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_11CompChannel_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_11CompChannel_9__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_11CompChannel_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_11CompChannel_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_11CompChannel_8__reduce_cython__(((struct __pyx_obj_7pyverbs_2cq_CompChannel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_11CompChannel_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.cc cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.cc cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_cc_cannot_be_converted_to_a, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.cc cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.CompChannel.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.cc cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.cc cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_11CompChannel_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_11CompChannel_11__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_11CompChannel_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_11CompChannel_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cq.CompChannel.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2cq_11CompChannel_10__setstate_cython__(((struct __pyx_obj_7pyverbs_2cq_CompChannel *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_11CompChannel_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.cc cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.cc cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_cc_cannot_be_converted_to_a, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.cc cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.cc cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.CompChannel.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":83
 *     completions. A CQ can have 0 or more associated QPs.
 *     """
 *     def __init__(self, Context context not None, cqe, cq_context=None,             # <<<<<<<<<<<<<<
 *                  CompChannel channel=None, comp_vector=0):
 *         """
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_2CQ_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2cq_2CQ___init__, "\n        Initializes a CQ object with the given parameters.\n        :param context: The device's context on which to open the CQ\n        :param cqe: CQ's capacity\n        :param cq_context: User context's pointer\n        :param channel: If set, will be used to return completion events\n        :param comp_vector: Will be used for signaling completion events.\n                            Must be larger than 0 and smaller than the\n                            context's num_comp_vectors\n        :return: The newly created CQ\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_7pyverbs_2cq_2CQ___init__;
#endif
static int __pyx_pw_7pyverbs_2cq_2CQ_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7pyverbs_6device_Context *__pyx_v_context = 0;
  PyObject *__pyx_v_cqe = 0;
  PyObject *__pyx_v_cq_context = 0;
  struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_channel = 0;
  PyObject *__pyx_v_comp_vector = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_context,&__pyx_mstate_global->__pyx_n_u_cqe,&__pyx_mstate_global->__pyx_n_u_cq_context,&__pyx_mstate_global->__pyx_n_u_channel,&__pyx_mstate_global->__pyx_n_u_comp_vector,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 83, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 83, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 83, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 83, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 83, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 83, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 83, __pyx_L3_error)
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "pyverbs/cq.pyx":84
 *     """
 *     def __init__(self, Context context not None, cqe, cq_context=None,
 *                  CompChannel channel=None, comp_vector=0):             # <<<<<<<<<<<<<<
 *         """
 *         Initializes a CQ object with the given parameters.
*/
      if (!values[3]) values[3] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2cq_CompChannel *)Py_None));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 5, i); __PYX_ERR(0, 83, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 83, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 83, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 83, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 83, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 83, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }

      /* "pyverbs/cq.pyx":83
 *     completions. A CQ can have 0 or more associated QPs.
 *     """
 *     def __init__(self, Context context not None, cqe, cq_context=None,             # <<<<<<<<<<<<<<
 *                  CompChannel channel=None, comp_vector=0):
 *         """
*/
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));

      /* "pyverbs/cq.pyx":84
 *     """
 *     def __init__(self, Context context not None, cqe, cq_context=None,
 *                  CompChannel channel=None, comp_vector=0):             # <<<<<<<<<<<<<<
 *         """
 *         Initializes a CQ object with the given parameters.
*/
      if (!values[3]) values[3] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2cq_CompChannel *)Py_None));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    }
    __pyx_v_context = ((struct __pyx_obj_7pyverbs_6device_Context *)values[0]);
    __pyx_v_cqe = values[1];
    __pyx_v_cq_context = values[2];
    __pyx_v_channel = ((struct __pyx_obj_7pyverbs_2cq_CompChannel *)values[3]);
    __pyx_v_comp_vector = values[4];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 2, 5, __pyx_nargs); __PYX_ERR(0, 83, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cq.CQ.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_context), __pyx_mstate_global->__pyx_ptype_7pyverbs_6device_Context, 0, "context", 0))) __PYX_ERR(0, 83, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_channel), __pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CompChannel, 1, "channel", 0))) __PYX_ERR(0, 84, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2cq_2CQ___init__(((struct __pyx_obj_7pyverbs_2cq_CQ *)__pyx_v_self), __pyx_v_context, __pyx_v_cqe, __pyx_v_cq_context, __pyx_v_channel, __pyx_v_comp_vector);

  /* "pyverbs/cq.pyx":83
 *     completions. A CQ can have 0 or more associated QPs.
 *     """
 *     def __init__(self, Context context not None, cqe, cq_context=None,             # <<<<<<<<<<<<<<
 *                  CompChannel channel=None, comp_vector=0):
 *         """
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_2CQ___init__(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self, struct __pyx_obj_7pyverbs_6device_Context *__pyx_v_context, PyObject *__pyx_v_cqe, PyObject *__pyx_v_cq_context, struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_channel, PyObject *__pyx_v_comp_vector) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyverbs/cq.pyx":96
 *         :return: The newly created CQ
 *         """
 *         super().__init__()             # <<<<<<<<<<<<<<
 *         if channel is not None:
 *             self.cq = v.ibv_create_cq(context.context, cqe, <void*>cq_context,
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQ), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":97
 *         """
 *         super().__init__()
 *         if channel is not None:             # <<<<<<<<<<<<<<
 *             self.cq = v.ibv_create_cq(context.context, cqe, <void*>cq_context,
 *                                       channel.cc, comp_vector)
*/
  __pyx_t_6 = (((PyObject *)__pyx_v_channel) != Py_None);
  if (__pyx_t_6) {

    /* "pyverbs/cq.pyx":98
 *         super().__init__()
 *         if channel is not None:
 *             self.cq = v.ibv_create_cq(context.context, cqe, <void*>cq_context,             # <<<<<<<<<<<<<<
 *                                       channel.cc, comp_vector)
 *             channel.add_ref(self)
*/
    __pyx_t_7 = __Pyx_PyLong_As_int(__pyx_v_cqe); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 98, __pyx_L1_error)

    /* "pyverbs/cq.pyx":99
 *         if channel is not None:
 *             self.cq = v.ibv_create_cq(context.context, cqe, <void*>cq_context,
 *                                       channel.cc, comp_vector)             # <<<<<<<<<<<<<<
 *             channel.add_ref(self)
 *             self.channel = channel
*/
    __pyx_t_8 = __Pyx_PyLong_As_int(__pyx_v_comp_vector); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 99, __pyx_L1_error)

    /* "pyverbs/cq.pyx":98
 *         super().__init__()
 *         if channel is not None:
 *             self.cq = v.ibv_create_cq(context.context, cqe, <void*>cq_context,             # <<<<<<<<<<<<<<
 *                                       channel.cc, comp_vector)
 *             channel.add_ref(self)
*/
    __pyx_v_self->cq = ibv_create_cq(__pyx_v_context->context, __pyx_t_7, ((void *)__pyx_v_cq_context), __pyx_v_channel->cc, __pyx_t_8);

    /* "pyverbs/cq.pyx":100
 *             self.cq = v.ibv_create_cq(context.context, cqe, <void*>cq_context,
 *                                       channel.cc, comp_vector)
 *             channel.add_ref(self)             # <<<<<<<<<<<<<<
 *             self.channel = channel
 *         else:
*/
    __pyx_t_1 = ((struct __pyx_vtabstruct_7pyverbs_2cq_CompChannel *)__pyx_v_channel->__pyx_base.__pyx_vtab)->add_ref(__pyx_v_channel, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyverbs/cq.pyx":101
 *                                       channel.cc, comp_vector)
 *             channel.add_ref(self)
 *             self.channel = channel             # <<<<<<<<<<<<<<
 *         else:
 *             self.cq = v.ibv_create_cq(context.context, cqe, <void*>cq_context,
*/
    __Pyx_INCREF((PyObject *)__pyx_v_channel);
    __Pyx_GIVEREF((PyObject *)__pyx_v_channel);
    __Pyx_GOTREF(__pyx_v_self->channel);
    __Pyx_DECREF(__pyx_v_self->channel);
    __pyx_v_self->channel = ((PyObject *)__pyx_v_channel);

    /* "pyverbs/cq.pyx":97
 *         """
 *         super().__init__()
 *         if channel is not None:             # <<<<<<<<<<<<<<
 *             self.cq = v.ibv_create_cq(context.context, cqe, <void*>cq_context,
 *                                       channel.cc, comp_vector)
*/
    goto __pyx_L3;
  }

  /* "pyverbs/cq.pyx":103
 *             self.channel = channel
 *         else:
 *             self.cq = v.ibv_create_cq(context.context, cqe, <void*>cq_context,             # <<<<<<<<<<<<<<
 *                                       NULL, comp_vector)
 *             self.channel = None
*/
  /*else*/ {
    __pyx_t_8 = __Pyx_PyLong_As_int(__pyx_v_cqe); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 103, __pyx_L1_error)

    /* "pyverbs/cq.pyx":104
 *         else:
 *             self.cq = v.ibv_create_cq(context.context, cqe, <void*>cq_context,
 *                                       NULL, comp_vector)             # <<<<<<<<<<<<<<
 *             self.channel = None
 *         if self.cq == NULL:
*/
    __pyx_t_7 = __Pyx_PyLong_As_int(__pyx_v_comp_vector); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 104, __pyx_L1_error)

    /* "pyverbs/cq.pyx":103
 *             self.channel = channel
 *         else:
 *             self.cq = v.ibv_create_cq(context.context, cqe, <void*>cq_context,             # <<<<<<<<<<<<<<
 *                                       NULL, comp_vector)
 *             self.channel = None
*/
    __pyx_v_self->cq = ibv_create_cq(__pyx_v_context->context, __pyx_t_8, ((void *)__pyx_v_cq_context), NULL, __pyx_t_7);

    /* "pyverbs/cq.pyx":105
 *             self.cq = v.ibv_create_cq(context.context, cqe, <void*>cq_context,
 *                                       NULL, comp_vector)
 *             self.channel = None             # <<<<<<<<<<<<<<
 *         if self.cq == NULL:
 *             raise PyverbsRDMAErrno('Failed to create a CQ')
*/
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->channel);
    __Pyx_DECREF(__pyx_v_self->channel);
    __pyx_v_self->channel = Py_None;
  }
  __pyx_L3:;

  /* "pyverbs/cq.pyx":106
 *                                       NULL, comp_vector)
 *             self.channel = None
 *         if self.cq == NULL:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to create a CQ')
 *         self.context = context
*/
  __pyx_t_6 = (__pyx_v_self->cq == NULL);
  if (unlikely(__pyx_t_6)) {

    /* "pyverbs/cq.pyx":107
 *             self.channel = None
 *         if self.cq == NULL:
 *             raise PyverbsRDMAErrno('Failed to create a CQ')             # <<<<<<<<<<<<<<
 *         self.context = context
 *         context.add_ref(self)
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Failed_to_create_a_CQ};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 107, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 107, __pyx_L1_error)

    /* "pyverbs/cq.pyx":106
 *                                       NULL, comp_vector)
 *             self.channel = None
 *         if self.cq == NULL:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to create a CQ')
 *         self.context = context
*/
  }

  /* "pyverbs/cq.pyx":108
 *         if self.cq == NULL:
 *             raise PyverbsRDMAErrno('Failed to create a CQ')
 *         self.context = context             # <<<<<<<<<<<<<<
 *         context.add_ref(self)
 *         self.qps = weakref.WeakSet()
*/
  __Pyx_INCREF((PyObject *)__pyx_v_context);
  __Pyx_GIVEREF((PyObject *)__pyx_v_context);
  __Pyx_GOTREF(__pyx_v_self->context);
  __Pyx_DECREF(__pyx_v_self->context);
  __pyx_v_self->context = ((PyObject *)__pyx_v_context);

  /* "pyverbs/cq.pyx":109
 *             raise PyverbsRDMAErrno('Failed to create a CQ')
 *         self.context = context
 *         context.add_ref(self)             # <<<<<<<<<<<<<<
 *         self.qps = weakref.WeakSet()
 *         self.srqs = weakref.WeakSet()
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_7pyverbs_6device_Context *)__pyx_v_context->__pyx_base.__pyx_vtab)->add_ref(__pyx_v_context, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":110
 *         self.context = context
 *         context.add_ref(self)
 *         self.qps = weakref.WeakSet()             # <<<<<<<<<<<<<<
 *         self.srqs = weakref.WeakSet()
 *         self.wqs = weakref.WeakSet()
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_weakref); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_WeakSet); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->qps);
  __Pyx_DECREF(__pyx_v_self->qps);
  __pyx_v_self->qps = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":111
 *         context.add_ref(self)
 *         self.qps = weakref.WeakSet()
 *         self.srqs = weakref.WeakSet()             # <<<<<<<<<<<<<<
 *         self.wqs = weakref.WeakSet()
 *         self.num_events = 0
*/
  __pyx_t_4 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_weakref); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_WeakSet); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    assert(__pyx_t_4);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->srqs);
  __Pyx_DECREF(__pyx_v_self->srqs);
  __pyx_v_self->srqs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":112
 *         self.qps = weakref.WeakSet()
 *         self.srqs = weakref.WeakSet()
 *         self.wqs = weakref.WeakSet()             # <<<<<<<<<<<<<<
 *         self.num_events = 0
 *         self.logger.debug('Created a CQ')
*/
  __pyx_t_3 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_weakref); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_WeakSet); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    assert(__pyx_t_3);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->wqs);
  __Pyx_DECREF(__pyx_v_self->wqs);
  __pyx_v_self->wqs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":113
 *         self.srqs = weakref.WeakSet()
 *         self.wqs = weakref.WeakSet()
 *         self.num_events = 0             # <<<<<<<<<<<<<<
 *         self.logger.debug('Created a CQ')
 * 
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_int_0);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_0);
  __Pyx_GOTREF(__pyx_v_self->num_events);
  __Pyx_DECREF(__pyx_v_self->num_events);
  __pyx_v_self->num_events = __pyx_mstate_global->__pyx_int_0;

  /* "pyverbs/cq.pyx":114
 *         self.wqs = weakref.WeakSet()
 *         self.num_events = 0
 *         self.logger.debug('Created a CQ')             # <<<<<<<<<<<<<<
 * 
 *     cdef add_ref(self, obj):
*/
  __pyx_t_2 = __pyx_v_self->__pyx_base.__pyx_base.logger;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Created_a_CQ};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_debug, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":83
 *     completions. A CQ can have 0 or more associated QPs.
 *     """
 *     def __init__(self, Context context not None, cqe, cq_context=None,             # <<<<<<<<<<<<<<
 *                  CompChannel channel=None, comp_vector=0):
 *         """
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cq.CQ.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":116
 *         self.logger.debug('Created a CQ')
 * 
 *     cdef add_ref(self, obj):             # <<<<<<<<<<<<<<
 *         if isinstance(obj, QP):
 *             self.qps.add(obj)
*/

static PyObject *__pyx_f_7pyverbs_2cq_2CQ_add_ref(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_ref", 0);

  /* "pyverbs/cq.pyx":117
 * 
 *     cdef add_ref(self, obj):
 *         if isinstance(obj, QP):             # <<<<<<<<<<<<<<
 *             self.qps.add(obj)
 *         elif isinstance(obj, SRQ):
*/
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QP); 
  if (__pyx_t_1) {

    /* "pyverbs/cq.pyx":118
 *     cdef add_ref(self, obj):
 *         if isinstance(obj, QP):
 *             self.qps.add(obj)             # <<<<<<<<<<<<<<
 *         elif isinstance(obj, SRQ):
 *             self.srqs.add(obj)
*/
    __pyx_t_3 = __pyx_v_self->qps;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_obj};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_add, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyverbs/cq.pyx":117
 * 
 *     cdef add_ref(self, obj):
 *         if isinstance(obj, QP):             # <<<<<<<<<<<<<<
 *             self.qps.add(obj)
 *         elif isinstance(obj, SRQ):
*/
    goto __pyx_L3;
  }

  /* "pyverbs/cq.pyx":119
 *         if isinstance(obj, QP):
 *             self.qps.add(obj)
 *         elif isinstance(obj, SRQ):             # <<<<<<<<<<<<<<
 *             self.srqs.add(obj)
 *         elif isinstance(obj, WQ):
*/
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_mstate_global->__pyx_ptype_7pyverbs_3srq_SRQ); 
  if (__pyx_t_1) {

    /* "pyverbs/cq.pyx":120
 *             self.qps.add(obj)
 *         elif isinstance(obj, SRQ):
 *             self.srqs.add(obj)             # <<<<<<<<<<<<<<
 *         elif isinstance(obj, WQ):
 *             self.wqs.add(obj)
*/
    __pyx_t_3 = __pyx_v_self->srqs;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_obj};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_add, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 120, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyverbs/cq.pyx":119
 *         if isinstance(obj, QP):
 *             self.qps.add(obj)
 *         elif isinstance(obj, SRQ):             # <<<<<<<<<<<<<<
 *             self.srqs.add(obj)
 *         elif isinstance(obj, WQ):
*/
    goto __pyx_L3;
  }

  /* "pyverbs/cq.pyx":121
 *         elif isinstance(obj, SRQ):
 *             self.srqs.add(obj)
 *         elif isinstance(obj, WQ):             # <<<<<<<<<<<<<<
 *             self.wqs.add(obj)
 *         else:
*/
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_mstate_global->__pyx_ptype_7pyverbs_2wq_WQ); 
  if (likely(__pyx_t_1)) {

    /* "pyverbs/cq.pyx":122
 *             self.srqs.add(obj)
 *         elif isinstance(obj, WQ):
 *             self.wqs.add(obj)             # <<<<<<<<<<<<<<
 *         else:
 *             raise PyverbsError('Unrecognized object type')
*/
    __pyx_t_3 = __pyx_v_self->wqs;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_obj};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_add, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 122, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyverbs/cq.pyx":121
 *         elif isinstance(obj, SRQ):
 *             self.srqs.add(obj)
 *         elif isinstance(obj, WQ):             # <<<<<<<<<<<<<<
 *             self.wqs.add(obj)
 *         else:
*/
    goto __pyx_L3;
  }

  /* "pyverbs/cq.pyx":124
 *             self.wqs.add(obj)
 *         else:
 *             raise PyverbsError('Unrecognized object type')             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
*/
  /*else*/ {
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PyverbsError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_4 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Unrecognized_object_type};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 124, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 124, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "pyverbs/cq.pyx":116
 *         self.logger.debug('Created a CQ')
 * 
 *     cdef add_ref(self, obj):             # <<<<<<<<<<<<<<
 *         if isinstance(obj, QP):
 *             self.qps.add(obj)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyverbs.cq.CQ.add_ref", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":126
 *             raise PyverbsError('Unrecognized object type')
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         self.close()
 * 
*/

/* Python wrapper */
static void __pyx_pw_7pyverbs_2cq_2CQ_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7pyverbs_2cq_2CQ_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_7pyverbs_2cq_2CQ_2__dealloc__(((struct __pyx_obj_7pyverbs_2cq_CQ *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7pyverbs_2cq_2CQ_2__dealloc__(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "pyverbs/cq.pyx":127
 * 
 *     def __dealloc__(self):
 *         self.close()             # <<<<<<<<<<<<<<
 * 
 *     cpdef close(self):
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_7pyverbs_2cq_CQ *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.close(((struct __pyx_obj_7pyverbs_4base_PyverbsCM *)__pyx_v_self), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":126
 *             raise PyverbsError('Unrecognized object type')
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         self.close()
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("pyverbs.cq.CQ.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "pyverbs/cq.pyx":129
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.cq != NULL:
 *             if self.logger:
*/

static PyObject *__pyx_pw_7pyverbs_2cq_2CQ_5close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7pyverbs_2cq_2CQ_close(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self, int __pyx_skip_dispatch) {
  int __pyx_v_rc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQ &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_close); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 129, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyverbs_2cq_2CQ_5close)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 129, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyverbs/cq.pyx":130
 * 
 *     cpdef close(self):
 *         if self.cq != NULL:             # <<<<<<<<<<<<<<
 *             if self.logger:
 *                 self.logger.debug('Closing CQ')
*/
  __pyx_t_6 = (__pyx_v_self->cq != NULL);
  if (__pyx_t_6) {

    /* "pyverbs/cq.pyx":131
 *     cpdef close(self):
 *         if self.cq != NULL:
 *             if self.logger:             # <<<<<<<<<<<<<<
 *                 self.logger.debug('Closing CQ')
 *             close_weakrefs([self.qps, self.srqs, self.wqs])
*/
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_self->__pyx_base.__pyx_base.logger); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 131, __pyx_L1_error)
    if (__pyx_t_6) {

      /* "pyverbs/cq.pyx":132
 *         if self.cq != NULL:
 *             if self.logger:
 *                 self.logger.debug('Closing CQ')             # <<<<<<<<<<<<<<
 *             close_weakrefs([self.qps, self.srqs, self.wqs])
 *             if self.num_events:
*/
      __pyx_t_2 = __pyx_v_self->__pyx_base.__pyx_base.logger;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_5 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Closing_CQ};
        __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_debug, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 132, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyverbs/cq.pyx":131
 *     cpdef close(self):
 *         if self.cq != NULL:
 *             if self.logger:             # <<<<<<<<<<<<<<
 *                 self.logger.debug('Closing CQ')
 *             close_weakrefs([self.qps, self.srqs, self.wqs])
*/
    }

    /* "pyverbs/cq.pyx":133
 *             if self.logger:
 *                 self.logger.debug('Closing CQ')
 *             close_weakrefs([self.qps, self.srqs, self.wqs])             # <<<<<<<<<<<<<<
 *             if self.num_events:
 *                 self.ack_events(self.num_events)
*/
    __pyx_t_1 = PyList_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_self->qps);
    __Pyx_GIVEREF(__pyx_v_self->qps);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->qps) != (0)) __PYX_ERR(0, 133, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_self->srqs);
    __Pyx_GIVEREF(__pyx_v_self->srqs);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->srqs) != (0)) __PYX_ERR(0, 133, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_self->wqs);
    __Pyx_GIVEREF(__pyx_v_self->wqs);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->wqs) != (0)) __PYX_ERR(0, 133, __pyx_L1_error);
    __pyx_t_2 = __pyx_f_7pyverbs_4base_close_weakrefs(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyverbs/cq.pyx":134
 *                 self.logger.debug('Closing CQ')
 *             close_weakrefs([self.qps, self.srqs, self.wqs])
 *             if self.num_events:             # <<<<<<<<<<<<<<
 *                 self.ack_events(self.num_events)
 *             rc = v.ibv_destroy_cq(self.cq)
*/
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_self->num_events); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 134, __pyx_L1_error)
    if (__pyx_t_6) {

      /* "pyverbs/cq.pyx":135
 *             close_weakrefs([self.qps, self.srqs, self.wqs])
 *             if self.num_events:
 *                 self.ack_events(self.num_events)             # <<<<<<<<<<<<<<
 *             rc = v.ibv_destroy_cq(self.cq)
 *             if rc != 0:
*/
      __pyx_t_1 = ((PyObject *)__pyx_v_self);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_5 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_self->num_events};
        __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_ack_events, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 135, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pyverbs/cq.pyx":134
 *                 self.logger.debug('Closing CQ')
 *             close_weakrefs([self.qps, self.srqs, self.wqs])
 *             if self.num_events:             # <<<<<<<<<<<<<<
 *                 self.ack_events(self.num_events)
 *             rc = v.ibv_destroy_cq(self.cq)
*/
    }

    /* "pyverbs/cq.pyx":136
 *             if self.num_events:
 *                 self.ack_events(self.num_events)
 *             rc = v.ibv_destroy_cq(self.cq)             # <<<<<<<<<<<<<<
 *             if rc != 0:
 *                 raise PyverbsRDMAError('Failed to close CQ', rc)
*/
    __pyx_v_rc = ibv_destroy_cq(__pyx_v_self->cq);

    /* "pyverbs/cq.pyx":137
 *                 self.ack_events(self.num_events)
 *             rc = v.ibv_destroy_cq(self.cq)
 *             if rc != 0:             # <<<<<<<<<<<<<<
 *                 raise PyverbsRDMAError('Failed to close CQ', rc)
 *             self.cq = NULL
*/
    __pyx_t_6 = (__pyx_v_rc != 0);
    if (unlikely(__pyx_t_6)) {

      /* "pyverbs/cq.pyx":138
 *             rc = v.ibv_destroy_cq(self.cq)
 *             if rc != 0:
 *                 raise PyverbsRDMAError('Failed to close CQ', rc)             # <<<<<<<<<<<<<<
 *             self.cq = NULL
 *             self.context = None
*/
      __pyx_t_1 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyLong_From_int(__pyx_v_rc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
        assert(__pyx_t_1);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
        __pyx_t_5 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_mstate_global->__pyx_kp_u_Failed_to_close_CQ, __pyx_t_3};
        __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 138, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 138, __pyx_L1_error)

      /* "pyverbs/cq.pyx":137
 *                 self.ack_events(self.num_events)
 *             rc = v.ibv_destroy_cq(self.cq)
 *             if rc != 0:             # <<<<<<<<<<<<<<
 *                 raise PyverbsRDMAError('Failed to close CQ', rc)
 *             self.cq = NULL
*/
    }

    /* "pyverbs/cq.pyx":139
 *             if rc != 0:
 *                 raise PyverbsRDMAError('Failed to close CQ', rc)
 *             self.cq = NULL             # <<<<<<<<<<<<<<
 *             self.context = None
 *             self.channel = None
*/
    __pyx_v_self->cq = NULL;

    /* "pyverbs/cq.pyx":140
 *                 raise PyverbsRDMAError('Failed to close CQ', rc)
 *             self.cq = NULL
 *             self.context = None             # <<<<<<<<<<<<<<
 *             self.channel = None
 * 
*/
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->context);
    __Pyx_DECREF(__pyx_v_self->context);
    __pyx_v_self->context = Py_None;

    /* "pyverbs/cq.pyx":141
 *             self.cq = NULL
 *             self.context = None
 *             self.channel = None             # <<<<<<<<<<<<<<
 * 
 *     def resize(self, cqe):
*/
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->channel);
    __Pyx_DECREF(__pyx_v_self->channel);
    __pyx_v_self->channel = Py_None;

    /* "pyverbs/cq.pyx":130
 * 
 *     cpdef close(self):
 *         if self.cq != NULL:             # <<<<<<<<<<<<<<
 *             if self.logger:
 *                 self.logger.debug('Closing CQ')
*/
  }

  /* "pyverbs/cq.pyx":129
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.cq != NULL:
 *             if self.logger:
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cq.CQ.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2CQ_5close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_2CQ_5close = {"close", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_2CQ_5close, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_2CQ_5close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("close", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("close", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_2CQ_4close(((struct __pyx_obj_7pyverbs_2cq_CQ *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2CQ_4close(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7pyverbs_2cq_2CQ_close(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CQ.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":143
 *             self.channel = None
 * 
 *     def resize(self, cqe):             # <<<<<<<<<<<<<<
 *         """
 *         Resizes the completion queue (CQ) to have at least cqe entries.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2CQ_7resize(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2cq_2CQ_6resize, "\n        Resizes the completion queue (CQ) to have at least cqe entries.\n        :param cqe: The requested CQ depth.\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_2cq_2CQ_7resize = {"resize", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_2CQ_7resize, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2cq_2CQ_6resize};
static PyObject *__pyx_pw_7pyverbs_2cq_2CQ_7resize(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_cqe = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("resize (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_cqe,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 143, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 143, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "resize", 0) < (0)) __PYX_ERR(0, 143, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("resize", 1, 1, 1, i); __PYX_ERR(0, 143, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 143, __pyx_L3_error)
    }
    __pyx_v_cqe = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("resize", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 143, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cq.CQ.resize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2cq_2CQ_6resize(((struct __pyx_obj_7pyverbs_2cq_CQ *)__pyx_v_self), __pyx_v_cqe);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2CQ_6resize(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self, PyObject *__pyx_v_cqe) {
  int __pyx_v_rc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("resize", 0);

  /* "pyverbs/cq.pyx":149
 *         :return: None
 *         """
 *         rc = v.ibv_resize_cq(self.cq, cqe)             # <<<<<<<<<<<<<<
 *         if rc:
 *             raise PyverbsRDMAError('Failed to resize CQ', rc)
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_cqe); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 149, __pyx_L1_error)
  __pyx_v_rc = ibv_resize_cq(__pyx_v_self->cq, __pyx_t_1);

  /* "pyverbs/cq.pyx":150
 *         """
 *         rc = v.ibv_resize_cq(self.cq, cqe)
 *         if rc:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Failed to resize CQ', rc)
 * 
*/
  __pyx_t_2 = (__pyx_v_rc != 0);
  if (unlikely(__pyx_t_2)) {

    /* "pyverbs/cq.pyx":151
 *         rc = v.ibv_resize_cq(self.cq, cqe)
 *         if rc:
 *             raise PyverbsRDMAError('Failed to resize CQ', rc)             # <<<<<<<<<<<<<<
 * 
 *     def poll(self, num_entries=1):
*/
    __pyx_t_4 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyLong_From_int(__pyx_v_rc); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_Failed_to_resize_CQ, __pyx_t_6};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_7, (3-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 151, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 151, __pyx_L1_error)

    /* "pyverbs/cq.pyx":150
 *         """
 *         rc = v.ibv_resize_cq(self.cq, cqe)
 *         if rc:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Failed to resize CQ', rc)
 * 
*/
  }

  /* "pyverbs/cq.pyx":143
 *             self.channel = None
 * 
 *     def resize(self, cqe):             # <<<<<<<<<<<<<<
 *         """
 *         Resizes the completion queue (CQ) to have at least cqe entries.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyverbs.cq.CQ.resize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":153
 *             raise PyverbsRDMAError('Failed to resize CQ', rc)
 * 
 *     def poll(self, num_entries=1):             # <<<<<<<<<<<<<<
 *         """
 *         Polls the CQ for completions.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2CQ_9poll(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2cq_2CQ_8poll, "\n        Polls the CQ for completions.\n        :param num_entries: number of completions to pull\n        :return: (npolled, wcs): The number of polled completions and an array\n                 of the polled completions\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_2cq_2CQ_9poll = {"poll", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_2CQ_9poll, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2cq_2CQ_8poll};
static PyObject *__pyx_pw_7pyverbs_2cq_2CQ_9poll(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_num_entries = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("poll (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_num_entries,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 153, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 153, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "poll", 0) < (0)) __PYX_ERR(0, 153, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_1));
    } else {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 153, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_1));
    }
    __pyx_v_num_entries = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("poll", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 153, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cq.CQ.poll", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2cq_2CQ_8poll(((struct __pyx_obj_7pyverbs_2cq_CQ *)__pyx_v_self), __pyx_v_num_entries);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2CQ_8poll(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self, PyObject *__pyx_v_num_entries) {
  struct ibv_wc __pyx_v_wc;
  PyObject *__pyx_v_wcs = NULL;
  PyObject *__pyx_v_npolled = NULL;
  PyObject *__pyx_v_rc = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_t_19;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("poll", 0);

  /* "pyverbs/cq.pyx":161
 *         """
 *         cdef v.ibv_wc wc
 *         wcs = []             # <<<<<<<<<<<<<<
 *         npolled = 0
 * 
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_wcs = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":162
 *         cdef v.ibv_wc wc
 *         wcs = []
 *         npolled = 0             # <<<<<<<<<<<<<<
 * 
 *         while npolled < num_entries:
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_int_0);
  __pyx_v_npolled = __pyx_mstate_global->__pyx_int_0;

  /* "pyverbs/cq.pyx":164
 *         npolled = 0
 * 
 *         while npolled < num_entries:             # <<<<<<<<<<<<<<
 *             rc = v.ibv_poll_cq(self.cq, 1, &wc)
 *             if rc < 0:
*/
  while (1) {
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_npolled, __pyx_v_num_entries, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 164, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 164, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_2) break;

    /* "pyverbs/cq.pyx":165
 * 
 *         while npolled < num_entries:
 *             rc = v.ibv_poll_cq(self.cq, 1, &wc)             # <<<<<<<<<<<<<<
 *             if rc < 0:
 *                 raise PyverbsRDMAError('Failed to poll CQ', -rc)
*/
    __pyx_t_1 = __Pyx_PyLong_From_int(ibv_poll_cq(__pyx_v_self->cq, 1, (&__pyx_v_wc))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_rc, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "pyverbs/cq.pyx":166
 *         while npolled < num_entries:
 *             rc = v.ibv_poll_cq(self.cq, 1, &wc)
 *             if rc < 0:             # <<<<<<<<<<<<<<
 *                 raise PyverbsRDMAError('Failed to poll CQ', -rc)
 *             if rc == 0:
*/
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_rc, __pyx_mstate_global->__pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 166, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(__pyx_t_2)) {

      /* "pyverbs/cq.pyx":167
 *             rc = v.ibv_poll_cq(self.cq, 1, &wc)
 *             if rc < 0:
 *                 raise PyverbsRDMAError('Failed to poll CQ', -rc)             # <<<<<<<<<<<<<<
 *             if rc == 0:
 *                 break;
*/
      __pyx_t_3 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 167, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PyNumber_Negative(__pyx_v_rc); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 167, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
        assert(__pyx_t_3);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
        __pyx_t_6 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Failed_to_poll_CQ, __pyx_t_5};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_6, (3-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 167, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 167, __pyx_L1_error)

      /* "pyverbs/cq.pyx":166
 *         while npolled < num_entries:
 *             rc = v.ibv_poll_cq(self.cq, 1, &wc)
 *             if rc < 0:             # <<<<<<<<<<<<<<
 *                 raise PyverbsRDMAError('Failed to poll CQ', -rc)
 *             if rc == 0:
*/
    }

    /* "pyverbs/cq.pyx":168
 *             if rc < 0:
 *                 raise PyverbsRDMAError('Failed to poll CQ', -rc)
 *             if rc == 0:             # <<<<<<<<<<<<<<
 *                 break;
 *             npolled += 1
*/
    __pyx_t_2 = (__Pyx_PyLong_BoolEqObjC(__pyx_v_rc, __pyx_mstate_global->__pyx_int_0, 0, 0)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 168, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "pyverbs/cq.pyx":169
 *                 raise PyverbsRDMAError('Failed to poll CQ', -rc)
 *             if rc == 0:
 *                 break;             # <<<<<<<<<<<<<<
 *             npolled += 1
 *             wcs.append(WC(wr_id=wc.wr_id, status=wc.status, opcode=wc.opcode,
*/
      goto __pyx_L4_break;

      /* "pyverbs/cq.pyx":168
 *             if rc < 0:
 *                 raise PyverbsRDMAError('Failed to poll CQ', -rc)
 *             if rc == 0:             # <<<<<<<<<<<<<<
 *                 break;
 *             npolled += 1
*/
    }

    /* "pyverbs/cq.pyx":170
 *             if rc == 0:
 *                 break;
 *             npolled += 1             # <<<<<<<<<<<<<<
 *             wcs.append(WC(wr_id=wc.wr_id, status=wc.status, opcode=wc.opcode,
 *                           vendor_err=wc.vendor_err, byte_len=wc.byte_len,
*/
    __pyx_t_1 = __Pyx_PyLong_AddObjC(__pyx_v_npolled, __pyx_mstate_global->__pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 170, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_npolled, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "pyverbs/cq.pyx":171
 *                 break;
 *             npolled += 1
 *             wcs.append(WC(wr_id=wc.wr_id, status=wc.status, opcode=wc.opcode,             # <<<<<<<<<<<<<<
 *                           vendor_err=wc.vendor_err, byte_len=wc.byte_len,
 *                           qp_num=wc.qp_num, src_qp=wc.src_qp,
*/
    __pyx_t_4 = NULL;
    __pyx_t_5 = __Pyx_PyLong_From_unsigned_long(__pyx_v_wc.wr_id); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(__pyx_v_wc.status); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py(__pyx_v_wc.opcode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);

    /* "pyverbs/cq.pyx":172
 *             npolled += 1
 *             wcs.append(WC(wr_id=wc.wr_id, status=wc.status, opcode=wc.opcode,
 *                           vendor_err=wc.vendor_err, byte_len=wc.byte_len,             # <<<<<<<<<<<<<<
 *                           qp_num=wc.qp_num, src_qp=wc.src_qp,
 *                           imm_data=wc.imm_data, wc_flags=wc.wc_flags,
*/
    __pyx_t_8 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.vendor_err); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 172, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.byte_len); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 172, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);

    /* "pyverbs/cq.pyx":173
 *             wcs.append(WC(wr_id=wc.wr_id, status=wc.status, opcode=wc.opcode,
 *                           vendor_err=wc.vendor_err, byte_len=wc.byte_len,
 *                           qp_num=wc.qp_num, src_qp=wc.src_qp,             # <<<<<<<<<<<<<<
 *                           imm_data=wc.imm_data, wc_flags=wc.wc_flags,
 *                           pkey_index=wc.pkey_index, slid=wc.slid, sl=wc.sl,
*/
    __pyx_t_10 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.qp_num); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.src_qp); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);

    /* "pyverbs/cq.pyx":174
 *                           vendor_err=wc.vendor_err, byte_len=wc.byte_len,
 *                           qp_num=wc.qp_num, src_qp=wc.src_qp,
 *                           imm_data=wc.imm_data, wc_flags=wc.wc_flags,             # <<<<<<<<<<<<<<
 *                           pkey_index=wc.pkey_index, slid=wc.slid, sl=wc.sl,
 *                           dlid_path_bits=wc.dlid_path_bits))
*/
    __pyx_t_12 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.imm_data); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_13 = __Pyx_PyLong_From_int(__pyx_v_wc.wc_flags); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);

    /* "pyverbs/cq.pyx":175
 *                           qp_num=wc.qp_num, src_qp=wc.src_qp,
 *                           imm_data=wc.imm_data, wc_flags=wc.wc_flags,
 *                           pkey_index=wc.pkey_index, slid=wc.slid, sl=wc.sl,             # <<<<<<<<<<<<<<
 *                           dlid_path_bits=wc.dlid_path_bits))
 *         return npolled, wcs
*/
    __pyx_t_14 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.pkey_index); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.slid); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_16 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.sl); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);

    /* "pyverbs/cq.pyx":176
 *                           imm_data=wc.imm_data, wc_flags=wc.wc_flags,
 *                           pkey_index=wc.pkey_index, slid=wc.slid, sl=wc.sl,
 *                           dlid_path_bits=wc.dlid_path_bits))             # <<<<<<<<<<<<<<
 *         return npolled, wcs
 * 
*/
    __pyx_t_17 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.dlid_path_bits); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 176, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_6 = 1;
    {
      PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 13 : 0)] = {__pyx_t_4, NULL};
      __pyx_t_18 = __Pyx_MakeVectorcallBuilderKwds(13); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 171, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_wr_id, __pyx_t_5, __pyx_t_18, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 171, __pyx_L1_error)
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_status, __pyx_t_3, __pyx_t_18, __pyx_callargs+1, 1) < (0)) __PYX_ERR(0, 171, __pyx_L1_error)
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_opcode, __pyx_t_7, __pyx_t_18, __pyx_callargs+1, 2) < (0)) __PYX_ERR(0, 171, __pyx_L1_error)
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_vendor_err, __pyx_t_8, __pyx_t_18, __pyx_callargs+1, 3) < (0)) __PYX_ERR(0, 171, __pyx_L1_error)
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_byte_len, __pyx_t_9, __pyx_t_18, __pyx_callargs+1, 4) < (0)) __PYX_ERR(0, 171, __pyx_L1_error)
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_qp_num, __pyx_t_10, __pyx_t_18, __pyx_callargs+1, 5) < (0)) __PYX_ERR(0, 171, __pyx_L1_error)
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_src_qp, __pyx_t_11, __pyx_t_18, __pyx_callargs+1, 6) < (0)) __PYX_ERR(0, 171, __pyx_L1_error)
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_imm_data, __pyx_t_12, __pyx_t_18, __pyx_callargs+1, 7) < (0)) __PYX_ERR(0, 171, __pyx_L1_error)
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_wc_flags, __pyx_t_13, __pyx_t_18, __pyx_callargs+1, 8) < (0)) __PYX_ERR(0, 171, __pyx_L1_error)
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_pkey_index, __pyx_t_14, __pyx_t_18, __pyx_callargs+1, 9) < (0)) __PYX_ERR(0, 171, __pyx_L1_error)
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_slid, __pyx_t_15, __pyx_t_18, __pyx_callargs+1, 10) < (0)) __PYX_ERR(0, 171, __pyx_L1_error)
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_sl, __pyx_t_16, __pyx_t_18, __pyx_callargs+1, 11) < (0)) __PYX_ERR(0, 171, __pyx_L1_error)
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_dlid_path_bits, __pyx_t_17, __pyx_t_18, __pyx_callargs+1, 12) < (0)) __PYX_ERR(0, 171, __pyx_L1_error)
      __pyx_t_1 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_WC, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_18);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 171, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_1);
    }

    /* "pyverbs/cq.pyx":171
 *                 break;
 *             npolled += 1
 *             wcs.append(WC(wr_id=wc.wr_id, status=wc.status, opcode=wc.opcode,             # <<<<<<<<<<<<<<
 *                           vendor_err=wc.vendor_err, byte_len=wc.byte_len,
 *                           qp_num=wc.qp_num, src_qp=wc.src_qp,
*/
    __pyx_t_19 = __Pyx_PyList_Append(__pyx_v_wcs, ((PyObject *)__pyx_t_1)); if (unlikely(__pyx_t_19 == ((int)-1))) __PYX_ERR(0, 171, __pyx_L1_error)
    __Pyx_DECREF((PyObject *)__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L4_break:;

  /* "pyverbs/cq.pyx":177
 *                           pkey_index=wc.pkey_index, slid=wc.slid, sl=wc.sl,
 *                           dlid_path_bits=wc.dlid_path_bits))
 *         return npolled, wcs             # <<<<<<<<<<<<<<
 * 
 *     def req_notify(self, solicited_only = False):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_npolled);
  __Pyx_GIVEREF(__pyx_v_npolled);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_npolled) != (0)) __PYX_ERR(0, 177, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_wcs);
  __Pyx_GIVEREF(__pyx_v_wcs);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_wcs) != (0)) __PYX_ERR(0, 177, __pyx_L1_error);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":153
 *             raise PyverbsRDMAError('Failed to resize CQ', rc)
 * 
 *     def poll(self, num_entries=1):             # <<<<<<<<<<<<<<
 *         """
 *         Polls the CQ for completions.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("pyverbs.cq.CQ.poll", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_wcs);
  __Pyx_XDECREF(__pyx_v_npolled);
  __Pyx_XDECREF(__pyx_v_rc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":179
 *         return npolled, wcs
 * 
 *     def req_notify(self, solicited_only = False):             # <<<<<<<<<<<<<<
 *         """
 *         Request completion notification on the completion queue.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2CQ_11req_notify(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2cq_2CQ_10req_notify, "\n        Request completion notification on the completion queue.\n        :param solicited_only: If non-zero, notifications will be created only\n                               for incoming send / RDMA write WRs with\n                               immediate data that have the solicited bit set in\n                               their send flags.\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_2cq_2CQ_11req_notify = {"req_notify", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_2CQ_11req_notify, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2cq_2CQ_10req_notify};
static PyObject *__pyx_pw_7pyverbs_2cq_2CQ_11req_notify(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_solicited_only = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("req_notify (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_solicited_only,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 179, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 179, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "req_notify", 0) < (0)) __PYX_ERR(0, 179, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_False));
    } else {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 179, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_False));
    }
    __pyx_v_solicited_only = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("req_notify", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 179, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cq.CQ.req_notify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2cq_2CQ_10req_notify(((struct __pyx_obj_7pyverbs_2cq_CQ *)__pyx_v_self), __pyx_v_solicited_only);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2CQ_10req_notify(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self, PyObject *__pyx_v_solicited_only) {
  int __pyx_v_rc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  size_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("req_notify", 0);

  /* "pyverbs/cq.pyx":188
 *         :return: None
 *         """
 *         rc = v.ibv_req_notify_cq(self.cq, solicited_only)             # <<<<<<<<<<<<<<
 *         if rc != 0:
 *             raise PyverbsRDMAError('Request notify CQ returned {rc}'.
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_solicited_only); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 188, __pyx_L1_error)
  __pyx_v_rc = ibv_req_notify_cq(__pyx_v_self->cq, __pyx_t_1);

  /* "pyverbs/cq.pyx":189
 *         """
 *         rc = v.ibv_req_notify_cq(self.cq, solicited_only)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Request notify CQ returned {rc}'.
 *                                    format(rc=rc), rc)
*/
  __pyx_t_2 = (__pyx_v_rc != 0);
  if (unlikely(__pyx_t_2)) {

    /* "pyverbs/cq.pyx":190
 *         rc = v.ibv_req_notify_cq(self.cq, solicited_only)
 *         if rc != 0:
 *             raise PyverbsRDMAError('Request notify CQ returned {rc}'.             # <<<<<<<<<<<<<<
 *                                    format(rc=rc), rc)
 * 
*/
    __pyx_t_4 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "pyverbs/cq.pyx":191
 *         if rc != 0:
 *             raise PyverbsRDMAError('Request notify CQ returned {rc}'.
 *                                    format(rc=rc), rc)             # <<<<<<<<<<<<<<
 * 
 *     def ack_events(self, num_events):
*/
    __pyx_t_7 = __pyx_mstate_global->__pyx_kp_u_Request_notify_CQ_returned_rc;
    __Pyx_INCREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyLong_From_int(__pyx_v_rc); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = 0;
    {
      PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_7, NULL};
      __pyx_t_10 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 191, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_rc, __pyx_t_8, __pyx_t_10, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 191, __pyx_L1_error)
      __pyx_t_6 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_format, __pyx_callargs+__pyx_t_9, (1-__pyx_t_9) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_10);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 191, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
    }
    __pyx_t_10 = __Pyx_PyLong_From_int(__pyx_v_rc); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_9 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_9 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_6, __pyx_t_10};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_9, (3-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 190, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 190, __pyx_L1_error)

    /* "pyverbs/cq.pyx":189
 *         """
 *         rc = v.ibv_req_notify_cq(self.cq, solicited_only)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Request notify CQ returned {rc}'.
 *                                    format(rc=rc), rc)
*/
  }

  /* "pyverbs/cq.pyx":179
 *         return npolled, wcs
 * 
 *     def req_notify(self, solicited_only = False):             # <<<<<<<<<<<<<<
 *         """
 *         Request completion notification on the completion queue.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pyverbs.cq.CQ.req_notify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":193
 *                                    format(rc=rc), rc)
 * 
 *     def ack_events(self, num_events):             # <<<<<<<<<<<<<<
 *         """
 *         Get and acknowledge CQ events
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2CQ_13ack_events(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2cq_2CQ_12ack_events, "\n        Get and acknowledge CQ events\n        :param num_events: Number of events to acknowledge\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_2cq_2CQ_13ack_events = {"ack_events", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_2CQ_13ack_events, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2cq_2CQ_12ack_events};
static PyObject *__pyx_pw_7pyverbs_2cq_2CQ_13ack_events(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_num_events = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ack_events (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_num_events,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 193, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 193, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "ack_events", 0) < (0)) __PYX_ERR(0, 193, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("ack_events", 1, 1, 1, i); __PYX_ERR(0, 193, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 193, __pyx_L3_error)
    }
    __pyx_v_num_events = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ack_events", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 193, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cq.CQ.ack_events", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2cq_2CQ_12ack_events(((struct __pyx_obj_7pyverbs_2cq_CQ *)__pyx_v_self), __pyx_v_num_events);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2CQ_12ack_events(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self, PyObject *__pyx_v_num_events) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ack_events", 0);

  /* "pyverbs/cq.pyx":199
 *         :return: None
 *         """
 *         v.ibv_ack_cq_events(self.cq, num_events)             # <<<<<<<<<<<<<<
 *         self.num_events -= num_events
 * 
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_num_events); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 199, __pyx_L1_error)
  ibv_ack_cq_events(__pyx_v_self->cq, __pyx_t_1);

  /* "pyverbs/cq.pyx":200
 *         """
 *         v.ibv_ack_cq_events(self.cq, num_events)
 *         self.num_events -= num_events             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
*/
  __pyx_t_2 = PyNumber_InPlaceSubtract(__pyx_v_self->num_events, __pyx_v_num_events); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->num_events);
  __Pyx_DECREF(__pyx_v_self->num_events);
  __pyx_v_self->num_events = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":193
 *                                    format(rc=rc), rc)
 * 
 *     def ack_events(self, num_events):             # <<<<<<<<<<<<<<
 *         """
 *         Get and acknowledge CQ events
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyverbs.cq.CQ.ack_events", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":202
 *         self.num_events -= num_events
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         print_format = '{:22}: {:<20}\n'
 *         return 'CQ\n' +\
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2CQ_15__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_2CQ_15__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2CQ_14__str__(((struct __pyx_obj_7pyverbs_2cq_CQ *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2CQ_14__str__(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self) {
  PyObject *__pyx_v_print_format = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "pyverbs/cq.pyx":203
 * 
 *     def __str__(self):
 *         print_format = '{:22}: {:<20}\n'             # <<<<<<<<<<<<<<
 *         return 'CQ\n' +\
 *                print_format.format('Handle', self.cq.handle) +\
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_22_20);
  __pyx_v_print_format = __pyx_mstate_global->__pyx_kp_u_22_20;

  /* "pyverbs/cq.pyx":204
 *     def __str__(self):
 *         print_format = '{:22}: {:<20}\n'
 *         return 'CQ\n' +\             # <<<<<<<<<<<<<<
 *                print_format.format('Handle', self.cq.handle) +\
 *                print_format.format('CQEs', self.cq.cqe)
*/
  __Pyx_XDECREF(__pyx_r);

  /* "pyverbs/cq.pyx":205
 *         print_format = '{:22}: {:<20}\n'
 *         return 'CQ\n' +\
 *                print_format.format('Handle', self.cq.handle) +\             # <<<<<<<<<<<<<<
 *                print_format.format('CQEs', self.cq.cqe)
 * 
*/
  __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_self->cq->handle); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_n_u_Handle, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":204
 *     def __str__(self):
 *         print_format = '{:22}: {:<20}\n'
 *         return 'CQ\n' +\             # <<<<<<<<<<<<<<
 *                print_format.format('Handle', self.cq.handle) +\
 *                print_format.format('CQEs', self.cq.cqe)
*/
  __pyx_t_1 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_n_u_CQ, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":206
 *         return 'CQ\n' +\
 *                print_format.format('Handle', self.cq.handle) +\
 *                print_format.format('CQEs', self.cq.cqe)             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_v_self->cq->cqe); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_n_u_CQEs, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":205
 *         print_format = '{:22}: {:<20}\n'
 *         return 'CQ\n' +\
 *                print_format.format('Handle', self.cq.handle) +\             # <<<<<<<<<<<<<<
 *                print_format.format('CQEs', self.cq.cqe)
 * 
*/
  __pyx_t_2 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":202
 *         self.num_events -= num_events
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         print_format = '{:22}: {:<20}\n'
 *         return 'CQ\n' +\
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyverbs.cq.CQ.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_print_format);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":208
 *                print_format.format('CQEs', self.cq.cqe)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def comp_channel(self):
 *         return self.channel
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2CQ_12comp_channel_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_2CQ_12comp_channel_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2CQ_12comp_channel___get__(((struct __pyx_obj_7pyverbs_2cq_CQ *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2CQ_12comp_channel___get__(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":210
 *     @property
 *     def comp_channel(self):
 *         return self.channel             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->channel);
  __pyx_r = __pyx_v_self->channel;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":208
 *                print_format.format('CQEs', self.cq.cqe)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def comp_channel(self):
 *         return self.channel
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":212
 *         return self.channel
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def cqe(self):
 *         return self.cq.cqe
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2CQ_3cqe_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_2CQ_3cqe_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2CQ_3cqe___get__(((struct __pyx_obj_7pyverbs_2cq_CQ *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2CQ_3cqe___get__(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":214
 *     @property
 *     def cqe(self):
 *         return self.cq.cqe             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_self->cq->cqe); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":212
 *         return self.channel
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def cqe(self):
 *         return self.cq.cqe
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CQ.cqe.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":216
 *         return self.cq.cqe
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def cq(self):
 *        return <uintptr_t>self.cq
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2CQ_2cq_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_2CQ_2cq_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2CQ_2cq___get__(((struct __pyx_obj_7pyverbs_2cq_CQ *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2CQ_2cq___get__(struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":218
 *     @property
 *     def cq(self):
 *        return <uintptr_t>self.cq             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_FromSize_t(((uintptr_t)__pyx_v_self->cq)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":216
 *         return self.cq.cqe
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def cq(self):
 *        return <uintptr_t>self.cq
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CQ.cq.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.cq cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2CQ_17__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_2CQ_17__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_2CQ_17__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_2CQ_17__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_2CQ_16__reduce_cython__(((struct __pyx_obj_7pyverbs_2cq_CQ *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2CQ_16__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.cq cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.cq cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_cq_cannot_be_converted_to_a, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.cq cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.CQ.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.cq cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.cq cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2CQ_19__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_2CQ_19__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_2CQ_19__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_2CQ_19__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cq.CQ.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2cq_2CQ_18__setstate_cython__(((struct __pyx_obj_7pyverbs_2cq_CQ *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2CQ_18__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_CQ *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.cq cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.cq cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_cq_cannot_be_converted_to_a, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.cq cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.cq cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.CQ.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":222
 * 
 * cdef class CqInitAttrEx(PyverbsObject):
 *     def __init__(self, cqe = 100, CompChannel channel = None, comp_vector = 0,             # <<<<<<<<<<<<<<
 *                  wc_flags = 0, comp_mask = 0, flags = 0, PD parent_domain = None):
 *         """
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2cq_12CqInitAttrEx___init__, "\n        Initializes a CqInitAttrEx object with the given parameters.\n        :param cqe: CQ's capacity\n        :param channel: If set, will be used to return completion events\n        :param comp_vector: Will be used for signaling completion events.\n                            Must be larger than 0 and smaller than the\n                            context's num_comp_vectors\n        :param wc_flags: The wc_flags that should be returned in ibv_poll_cq_ex.\n                         Or'ed bit of enum ibv_wc_flags_ex.\n        :param comp_mask: compatibility mask (extended verb)\n        :param flags: create cq attr flags - one or more flags from\n                      ibv_create_cq_attr_flags enum\n        :param parent_domain: If set, will be used to custom alloc cq buffers.\n        :return:\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_7pyverbs_2cq_12CqInitAttrEx___init__;
#endif
static int __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_cqe = 0;
  struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_channel = 0;
  PyObject *__pyx_v_comp_vector = 0;
  PyObject *__pyx_v_wc_flags = 0;
  PyObject *__pyx_v_comp_mask = 0;
  PyObject *__pyx_v_flags = 0;
  struct __pyx_obj_7pyverbs_2pd_PD *__pyx_v_parent_domain = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[7] = {0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_cqe,&__pyx_mstate_global->__pyx_n_u_channel,&__pyx_mstate_global->__pyx_n_u_comp_vector,&__pyx_mstate_global->__pyx_n_u_wc_flags,&__pyx_mstate_global->__pyx_n_u_comp_mask,&__pyx_mstate_global->__pyx_n_u_flags,&__pyx_mstate_global->__pyx_n_u_parent_domain,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 222, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  7:
        values[6] = __Pyx_ArgRef_VARARGS(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 222, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 222, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 222, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 222, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 222, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 222, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 222, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 222, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_100));
      if (!values[1]) values[1] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2cq_CompChannel *)Py_None));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[5]) values[5] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));

      /* "pyverbs/cq.pyx":223
 * cdef class CqInitAttrEx(PyverbsObject):
 *     def __init__(self, cqe = 100, CompChannel channel = None, comp_vector = 0,
 *                  wc_flags = 0, comp_mask = 0, flags = 0, PD parent_domain = None):             # <<<<<<<<<<<<<<
 *         """
 *         Initializes a CqInitAttrEx object with the given parameters.
*/
      if (!values[6]) values[6] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2pd_PD *)Py_None));
    } else {
      switch (__pyx_nargs) {
        case  7:
        values[6] = __Pyx_ArgRef_VARARGS(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 222, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 222, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 222, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 222, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 222, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 222, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 222, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_100));

      /* "pyverbs/cq.pyx":222
 * 
 * cdef class CqInitAttrEx(PyverbsObject):
 *     def __init__(self, cqe = 100, CompChannel channel = None, comp_vector = 0,             # <<<<<<<<<<<<<<
 *                  wc_flags = 0, comp_mask = 0, flags = 0, PD parent_domain = None):
 *         """
*/
      if (!values[1]) values[1] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2cq_CompChannel *)Py_None));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[5]) values[5] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));

      /* "pyverbs/cq.pyx":223
 * cdef class CqInitAttrEx(PyverbsObject):
 *     def __init__(self, cqe = 100, CompChannel channel = None, comp_vector = 0,
 *                  wc_flags = 0, comp_mask = 0, flags = 0, PD parent_domain = None):             # <<<<<<<<<<<<<<
 *         """
 *         Initializes a CqInitAttrEx object with the given parameters.
*/
      if (!values[6]) values[6] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2pd_PD *)Py_None));
    }
    __pyx_v_cqe = values[0];
    __pyx_v_channel = ((struct __pyx_obj_7pyverbs_2cq_CompChannel *)values[1]);
    __pyx_v_comp_vector = values[2];
    __pyx_v_wc_flags = values[3];
    __pyx_v_comp_mask = values[4];
    __pyx_v_flags = values[5];
    __pyx_v_parent_domain = ((struct __pyx_obj_7pyverbs_2pd_PD *)values[6]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 7, __pyx_nargs); __PYX_ERR(0, 222, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cq.CqInitAttrEx.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_channel), __pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CompChannel, 1, "channel", 0))) __PYX_ERR(0, 222, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_parent_domain), __pyx_mstate_global->__pyx_ptype_7pyverbs_2pd_PD, 1, "parent_domain", 0))) __PYX_ERR(0, 223, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2cq_12CqInitAttrEx___init__(((struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)__pyx_v_self), __pyx_v_cqe, __pyx_v_channel, __pyx_v_comp_vector, __pyx_v_wc_flags, __pyx_v_comp_mask, __pyx_v_flags, __pyx_v_parent_domain);

  /* "pyverbs/cq.pyx":222
 * 
 * cdef class CqInitAttrEx(PyverbsObject):
 *     def __init__(self, cqe = 100, CompChannel channel = None, comp_vector = 0,             # <<<<<<<<<<<<<<
 *                  wc_flags = 0, comp_mask = 0, flags = 0, PD parent_domain = None):
 *         """
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_12CqInitAttrEx___init__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self, PyObject *__pyx_v_cqe, struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_channel, PyObject *__pyx_v_comp_vector, PyObject *__pyx_v_wc_flags, PyObject *__pyx_v_comp_mask, PyObject *__pyx_v_flags, struct __pyx_obj_7pyverbs_2pd_PD *__pyx_v_parent_domain) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  unsigned int __pyx_t_6;
  void *__pyx_t_7;
  int __pyx_t_8;
  unsigned long __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyverbs/cq.pyx":239
 *         :return:
 *         """
 *         super().__init__()             # <<<<<<<<<<<<<<
 *         self.attr.cqe = cqe
 *         self.attr.cq_context = NULL
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 239, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 239, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":240
 *         """
 *         super().__init__()
 *         self.attr.cqe = cqe             # <<<<<<<<<<<<<<
 *         self.attr.cq_context = NULL
 *         self.attr.channel = NULL if channel is None else channel.cc
*/
  __pyx_t_6 = __Pyx_PyLong_As_unsigned_int(__pyx_v_cqe); if (unlikely((__pyx_t_6 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 240, __pyx_L1_error)
  __pyx_v_self->attr.cqe = __pyx_t_6;

  /* "pyverbs/cq.pyx":241
 *         super().__init__()
 *         self.attr.cqe = cqe
 *         self.attr.cq_context = NULL             # <<<<<<<<<<<<<<
 *         self.attr.channel = NULL if channel is None else channel.cc
 *         self.attr.comp_vector = comp_vector
*/
  __pyx_v_self->attr.cq_context = NULL;

  /* "pyverbs/cq.pyx":242
 *         self.attr.cqe = cqe
 *         self.attr.cq_context = NULL
 *         self.attr.channel = NULL if channel is None else channel.cc             # <<<<<<<<<<<<<<
 *         self.attr.comp_vector = comp_vector
 *         self.attr.wc_flags = wc_flags
*/
  __pyx_t_8 = (((PyObject *)__pyx_v_channel) == Py_None);
  if (__pyx_t_8) {
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_7 = __pyx_v_channel->cc;
  }
  __pyx_v_self->attr.channel = __pyx_t_7;

  /* "pyverbs/cq.pyx":243
 *         self.attr.cq_context = NULL
 *         self.attr.channel = NULL if channel is None else channel.cc
 *         self.attr.comp_vector = comp_vector             # <<<<<<<<<<<<<<
 *         self.attr.wc_flags = wc_flags
 *         self.attr.comp_mask = comp_mask
*/
  __pyx_t_6 = __Pyx_PyLong_As_unsigned_int(__pyx_v_comp_vector); if (unlikely((__pyx_t_6 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 243, __pyx_L1_error)
  __pyx_v_self->attr.comp_vector = __pyx_t_6;

  /* "pyverbs/cq.pyx":244
 *         self.attr.channel = NULL if channel is None else channel.cc
 *         self.attr.comp_vector = comp_vector
 *         self.attr.wc_flags = wc_flags             # <<<<<<<<<<<<<<
 *         self.attr.comp_mask = comp_mask
 *         self.attr.flags = flags
*/
  __pyx_t_9 = __Pyx_PyLong_As_unsigned_long(__pyx_v_wc_flags); if (unlikely((__pyx_t_9 == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 244, __pyx_L1_error)
  __pyx_v_self->attr.wc_flags = __pyx_t_9;

  /* "pyverbs/cq.pyx":245
 *         self.attr.comp_vector = comp_vector
 *         self.attr.wc_flags = wc_flags
 *         self.attr.comp_mask = comp_mask             # <<<<<<<<<<<<<<
 *         self.attr.flags = flags
 *         self.attr.parent_domain = NULL if parent_domain is None else parent_domain.pd
*/
  __pyx_t_6 = __Pyx_PyLong_As_unsigned_int(__pyx_v_comp_mask); if (unlikely((__pyx_t_6 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 245, __pyx_L1_error)
  __pyx_v_self->attr.comp_mask = __pyx_t_6;

  /* "pyverbs/cq.pyx":246
 *         self.attr.wc_flags = wc_flags
 *         self.attr.comp_mask = comp_mask
 *         self.attr.flags = flags             # <<<<<<<<<<<<<<
 *         self.attr.parent_domain = NULL if parent_domain is None else parent_domain.pd
 *         self.channel = channel
*/
  __pyx_t_6 = __Pyx_PyLong_As_unsigned_int(__pyx_v_flags); if (unlikely((__pyx_t_6 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 246, __pyx_L1_error)
  __pyx_v_self->attr.flags = __pyx_t_6;

  /* "pyverbs/cq.pyx":247
 *         self.attr.comp_mask = comp_mask
 *         self.attr.flags = flags
 *         self.attr.parent_domain = NULL if parent_domain is None else parent_domain.pd             # <<<<<<<<<<<<<<
 *         self.channel = channel
 *         self.parent_domain = parent_domain
*/
  __pyx_t_8 = (((PyObject *)__pyx_v_parent_domain) == Py_None);
  if (__pyx_t_8) {
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_7 = __pyx_v_parent_domain->pd;
  }
  __pyx_v_self->attr.parent_domain = __pyx_t_7;

  /* "pyverbs/cq.pyx":248
 *         self.attr.flags = flags
 *         self.attr.parent_domain = NULL if parent_domain is None else parent_domain.pd
 *         self.channel = channel             # <<<<<<<<<<<<<<
 *         self.parent_domain = parent_domain
 * 
*/
  __Pyx_INCREF((PyObject *)__pyx_v_channel);
  __Pyx_GIVEREF((PyObject *)__pyx_v_channel);
  __Pyx_GOTREF(__pyx_v_self->channel);
  __Pyx_DECREF(__pyx_v_self->channel);
  __pyx_v_self->channel = ((PyObject *)__pyx_v_channel);

  /* "pyverbs/cq.pyx":249
 *         self.attr.parent_domain = NULL if parent_domain is None else parent_domain.pd
 *         self.channel = channel
 *         self.parent_domain = parent_domain             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_INCREF((PyObject *)__pyx_v_parent_domain);
  __Pyx_GIVEREF((PyObject *)__pyx_v_parent_domain);
  __Pyx_GOTREF(__pyx_v_self->parent_domain);
  __Pyx_DECREF(__pyx_v_self->parent_domain);
  __pyx_v_self->parent_domain = ((PyObject *)__pyx_v_parent_domain);

  /* "pyverbs/cq.pyx":222
 * 
 * cdef class CqInitAttrEx(PyverbsObject):
 *     def __init__(self, cqe = 100, CompChannel channel = None, comp_vector = 0,             # <<<<<<<<<<<<<<
 *                  wc_flags = 0, comp_mask = 0, flags = 0, PD parent_domain = None):
 *         """
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cq.CqInitAttrEx.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":251
 *         self.parent_domain = parent_domain
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def cqe(self):
 *         return self.attr.cqe
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_3cqe_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_3cqe_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_3cqe___get__(((struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_12CqInitAttrEx_3cqe___get__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":253
 *     @property
 *     def cqe(self):
 *         return self.attr.cqe             # <<<<<<<<<<<<<<
 *     @cqe.setter
 *     def cqe(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.cqe); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":251
 *         self.parent_domain = parent_domain
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def cqe(self):
 *         return self.attr.cqe
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CqInitAttrEx.cqe.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":254
 *     def cqe(self):
 *         return self.attr.cqe
 *     @cqe.setter             # <<<<<<<<<<<<<<
 *     def cqe(self, val):
 *         self.attr.cqe = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_3cqe_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_3cqe_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_3cqe_2__set__(((struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_3cqe_2__set__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cq.pyx":256
 *     @cqe.setter
 *     def cqe(self, val):
 *         self.attr.cqe = val             # <<<<<<<<<<<<<<
 * 
 *     # Setter-only properties require the older syntax
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 256, __pyx_L1_error)
  __pyx_v_self->attr.cqe = __pyx_t_1;

  /* "pyverbs/cq.pyx":254
 *     def cqe(self):
 *         return self.attr.cqe
 *     @cqe.setter             # <<<<<<<<<<<<<<
 *     def cqe(self, val):
 *         self.attr.cqe = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.CqInitAttrEx.cqe.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/cq.pyx":260
 *     # Setter-only properties require the older syntax
 *     property cq_context:
 *         def __set__(self, val):             # <<<<<<<<<<<<<<
 *             self.attr.cq_context = <void*>val
 * 
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_10cq_context_1__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_10cq_context_1__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_10cq_context___set__(((struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_10cq_context___set__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;

  /* "pyverbs/cq.pyx":261
 *     property cq_context:
 *         def __set__(self, val):
 *             self.attr.cq_context = <void*>val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_v_self->attr.cq_context = ((void *)__pyx_v_val);

  /* "pyverbs/cq.pyx":260
 *     # Setter-only properties require the older syntax
 *     property cq_context:
 *         def __set__(self, val):             # <<<<<<<<<<<<<<
 *             self.attr.cq_context = <void*>val
 * 
*/

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "pyverbs/cq.pyx":263
 *             self.attr.cq_context = <void*>val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def parent_domain(self):
 *         return self.parent_domain
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_13parent_domain_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_13parent_domain_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_13parent_domain___get__(((struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_12CqInitAttrEx_13parent_domain___get__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":265
 *     @property
 *     def parent_domain(self):
 *         return self.parent_domain             # <<<<<<<<<<<<<<
 *     @parent_domain.setter
 *     def parent_domain(self, PD val):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->parent_domain);
  __pyx_r = __pyx_v_self->parent_domain;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":263
 *             self.attr.cq_context = <void*>val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def parent_domain(self):
 *         return self.parent_domain
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":266
 *     def parent_domain(self):
 *         return self.parent_domain
 *     @parent_domain.setter             # <<<<<<<<<<<<<<
 *     def parent_domain(self, PD val):
 *         self.parent_domain = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_13parent_domain_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_13parent_domain_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_val), __pyx_mstate_global->__pyx_ptype_7pyverbs_2pd_PD, 1, "val", 0))) __PYX_ERR(0, 267, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_13parent_domain_2__set__(((struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)__pyx_v_self), ((struct __pyx_obj_7pyverbs_2pd_PD *)__pyx_v_val));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  goto __pyx_L5_cleaned_up;
  __pyx_L0:;
  __pyx_L5_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_13parent_domain_2__set__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self, struct __pyx_obj_7pyverbs_2pd_PD *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  struct ibv_pd *__pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "pyverbs/cq.pyx":268
 *     @parent_domain.setter
 *     def parent_domain(self, PD val):
 *         self.parent_domain = val             # <<<<<<<<<<<<<<
 *         self.attr.parent_domain = val.pd
 * 
*/
  __Pyx_INCREF((PyObject *)__pyx_v_val);
  __Pyx_GIVEREF((PyObject *)__pyx_v_val);
  __Pyx_GOTREF(__pyx_v_self->parent_domain);
  __Pyx_DECREF(__pyx_v_self->parent_domain);
  __pyx_v_self->parent_domain = ((PyObject *)__pyx_v_val);

  /* "pyverbs/cq.pyx":269
 *     def parent_domain(self, PD val):
 *         self.parent_domain = val
 *         self.attr.parent_domain = val.pd             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __pyx_v_val->pd;
  __pyx_v_self->attr.parent_domain = __pyx_t_1;

  /* "pyverbs/cq.pyx":266
 *     def parent_domain(self):
 *         return self.parent_domain
 *     @parent_domain.setter             # <<<<<<<<<<<<<<
 *     def parent_domain(self, PD val):
 *         self.parent_domain = val
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":271
 *         self.attr.parent_domain = val.pd
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def comp_channel(self):
 *         return self.channel
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_12comp_channel_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_12comp_channel_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_12comp_channel___get__(((struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_12CqInitAttrEx_12comp_channel___get__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":273
 *     @property
 *     def comp_channel(self):
 *         return self.channel             # <<<<<<<<<<<<<<
 *     @comp_channel.setter
 *     def comp_channel(self, CompChannel val):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->channel);
  __pyx_r = __pyx_v_self->channel;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":271
 *         self.attr.parent_domain = val.pd
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def comp_channel(self):
 *         return self.channel
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":274
 *     def comp_channel(self):
 *         return self.channel
 *     @comp_channel.setter             # <<<<<<<<<<<<<<
 *     def comp_channel(self, CompChannel val):
 *         self.channel = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_12comp_channel_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_12comp_channel_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_val), __pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CompChannel, 1, "val", 0))) __PYX_ERR(0, 275, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_12comp_channel_2__set__(((struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)__pyx_v_self), ((struct __pyx_obj_7pyverbs_2cq_CompChannel *)__pyx_v_val));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  goto __pyx_L5_cleaned_up;
  __pyx_L0:;
  __pyx_L5_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_12comp_channel_2__set__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self, struct __pyx_obj_7pyverbs_2cq_CompChannel *__pyx_v_val) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  struct ibv_comp_channel *__pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "pyverbs/cq.pyx":276
 *     @comp_channel.setter
 *     def comp_channel(self, CompChannel val):
 *         self.channel = val             # <<<<<<<<<<<<<<
 *         self.attr.channel = val.cc
 * 
*/
  __Pyx_INCREF((PyObject *)__pyx_v_val);
  __Pyx_GIVEREF((PyObject *)__pyx_v_val);
  __Pyx_GOTREF(__pyx_v_self->channel);
  __Pyx_DECREF(__pyx_v_self->channel);
  __pyx_v_self->channel = ((PyObject *)__pyx_v_val);

  /* "pyverbs/cq.pyx":277
 *     def comp_channel(self, CompChannel val):
 *         self.channel = val
 *         self.attr.channel = val.cc             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __pyx_v_val->cc;
  __pyx_v_self->attr.channel = __pyx_t_1;

  /* "pyverbs/cq.pyx":274
 *     def comp_channel(self):
 *         return self.channel
 *     @comp_channel.setter             # <<<<<<<<<<<<<<
 *     def comp_channel(self, CompChannel val):
 *         self.channel = val
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":279
 *         self.attr.channel = val.cc
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def comp_vector(self):
 *         return self.attr.comp_vector
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_11comp_vector_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_11comp_vector_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_11comp_vector___get__(((struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_12CqInitAttrEx_11comp_vector___get__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":281
 *     @property
 *     def comp_vector(self):
 *         return self.attr.comp_vector             # <<<<<<<<<<<<<<
 *     @comp_vector.setter
 *     def comp_vector(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.comp_vector); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":279
 *         self.attr.channel = val.cc
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def comp_vector(self):
 *         return self.attr.comp_vector
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CqInitAttrEx.comp_vector.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":282
 *     def comp_vector(self):
 *         return self.attr.comp_vector
 *     @comp_vector.setter             # <<<<<<<<<<<<<<
 *     def comp_vector(self, val):
 *         self.attr.comp_vector = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_11comp_vector_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_11comp_vector_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_11comp_vector_2__set__(((struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_11comp_vector_2__set__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cq.pyx":284
 *     @comp_vector.setter
 *     def comp_vector(self, val):
 *         self.attr.comp_vector = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 284, __pyx_L1_error)
  __pyx_v_self->attr.comp_vector = __pyx_t_1;

  /* "pyverbs/cq.pyx":282
 *     def comp_vector(self):
 *         return self.attr.comp_vector
 *     @comp_vector.setter             # <<<<<<<<<<<<<<
 *     def comp_vector(self, val):
 *         self.attr.comp_vector = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.CqInitAttrEx.comp_vector.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/cq.pyx":286
 *         self.attr.comp_vector = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def wc_flags(self):
 *         return self.attr.wc_flags
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_8wc_flags_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_8wc_flags_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_8wc_flags___get__(((struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_12CqInitAttrEx_8wc_flags___get__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":288
 *     @property
 *     def wc_flags(self):
 *         return self.attr.wc_flags             # <<<<<<<<<<<<<<
 *     @wc_flags.setter
 *     def wc_flags(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_long(__pyx_v_self->attr.wc_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":286
 *         self.attr.comp_vector = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def wc_flags(self):
 *         return self.attr.wc_flags
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CqInitAttrEx.wc_flags.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":289
 *     def wc_flags(self):
 *         return self.attr.wc_flags
 *     @wc_flags.setter             # <<<<<<<<<<<<<<
 *     def wc_flags(self, val):
 *         self.attr.wc_flags = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_8wc_flags_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_8wc_flags_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_8wc_flags_2__set__(((struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_8wc_flags_2__set__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned long __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cq.pyx":291
 *     @wc_flags.setter
 *     def wc_flags(self, val):
 *         self.attr.wc_flags = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_long(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 291, __pyx_L1_error)
  __pyx_v_self->attr.wc_flags = __pyx_t_1;

  /* "pyverbs/cq.pyx":289
 *     def wc_flags(self):
 *         return self.attr.wc_flags
 *     @wc_flags.setter             # <<<<<<<<<<<<<<
 *     def wc_flags(self, val):
 *         self.attr.wc_flags = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.CqInitAttrEx.wc_flags.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/cq.pyx":293
 *         self.attr.wc_flags = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def comp_mask(self):
 *         return self.attr.comp_mask
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_9comp_mask_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_9comp_mask_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_9comp_mask___get__(((struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_12CqInitAttrEx_9comp_mask___get__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":295
 *     @property
 *     def comp_mask(self):
 *         return self.attr.comp_mask             # <<<<<<<<<<<<<<
 *     @comp_mask.setter
 *     def comp_mask(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.comp_mask); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":293
 *         self.attr.wc_flags = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def comp_mask(self):
 *         return self.attr.comp_mask
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CqInitAttrEx.comp_mask.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":296
 *     def comp_mask(self):
 *         return self.attr.comp_mask
 *     @comp_mask.setter             # <<<<<<<<<<<<<<
 *     def comp_mask(self, val):
 *         self.attr.comp_mask = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_9comp_mask_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_9comp_mask_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_9comp_mask_2__set__(((struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_9comp_mask_2__set__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cq.pyx":298
 *     @comp_mask.setter
 *     def comp_mask(self, val):
 *         self.attr.comp_mask = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 298, __pyx_L1_error)
  __pyx_v_self->attr.comp_mask = __pyx_t_1;

  /* "pyverbs/cq.pyx":296
 *     def comp_mask(self):
 *         return self.attr.comp_mask
 *     @comp_mask.setter             # <<<<<<<<<<<<<<
 *     def comp_mask(self, val):
 *         self.attr.comp_mask = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.CqInitAttrEx.comp_mask.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/cq.pyx":300
 *         self.attr.comp_mask = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def flags(self):
 *         return self.attr.flags
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_5flags_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_5flags_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_5flags___get__(((struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_12CqInitAttrEx_5flags___get__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":302
 *     @property
 *     def flags(self):
 *         return self.attr.flags             # <<<<<<<<<<<<<<
 *     @flags.setter
 *     def flags(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":300
 *         self.attr.comp_mask = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def flags(self):
 *         return self.attr.flags
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CqInitAttrEx.flags.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":303
 *     def flags(self):
 *         return self.attr.flags
 *     @flags.setter             # <<<<<<<<<<<<<<
 *     def flags(self, val):
 *         self.attr.flags = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_5flags_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_5flags_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_5flags_2__set__(((struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_5flags_2__set__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cq.pyx":305
 *     @flags.setter
 *     def flags(self, val):
 *         self.attr.flags = val             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 305, __pyx_L1_error)
  __pyx_v_self->attr.flags = __pyx_t_1;

  /* "pyverbs/cq.pyx":303
 *     def flags(self):
 *         return self.attr.flags
 *     @flags.setter             # <<<<<<<<<<<<<<
 *     def flags(self, val):
 *         self.attr.flags = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.CqInitAttrEx.flags.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/cq.pyx":307
 *         self.attr.flags = val
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         print_format = '{:22}: {:<20}\n'
 *         return print_format.format('Number of CQEs', self.cqe) +\
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_3__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_3__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_2__str__(((struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_12CqInitAttrEx_2__str__(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_v_print_format = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "pyverbs/cq.pyx":308
 * 
 *     def __str__(self):
 *         print_format = '{:22}: {:<20}\n'             # <<<<<<<<<<<<<<
 *         return print_format.format('Number of CQEs', self.cqe) +\
 *             print_format.format('WC flags', create_wc_flags_to_str(self.wc_flags)) +\
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_22_20);
  __pyx_v_print_format = __pyx_mstate_global->__pyx_kp_u_22_20;

  /* "pyverbs/cq.pyx":309
 *     def __str__(self):
 *         print_format = '{:22}: {:<20}\n'
 *         return print_format.format('Number of CQEs', self.cqe) +\             # <<<<<<<<<<<<<<
 *             print_format.format('WC flags', create_wc_flags_to_str(self.wc_flags)) +\
 *             print_format.format('comp mask', self.comp_mask) +\
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_cqe); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_Number_of_CQEs, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":310
 *         print_format = '{:22}: {:<20}\n'
 *         return print_format.format('Number of CQEs', self.cqe) +\
 *             print_format.format('WC flags', create_wc_flags_to_str(self.wc_flags)) +\             # <<<<<<<<<<<<<<
 *             print_format.format('comp mask', self.comp_mask) +\
 *             print_format.format('flags', self.flags)
*/
  __pyx_t_3 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_create_wc_flags_to_str); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_wc_flags); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_3);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_6 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 310, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_WC_flags, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":309
 *     def __str__(self):
 *         print_format = '{:22}: {:<20}\n'
 *         return print_format.format('Number of CQEs', self.cqe) +\             # <<<<<<<<<<<<<<
 *             print_format.format('WC flags', create_wc_flags_to_str(self.wc_flags)) +\
 *             print_format.format('comp mask', self.comp_mask) +\
*/
  __pyx_t_1 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/cq.pyx":311
 *         return print_format.format('Number of CQEs', self.cqe) +\
 *             print_format.format('WC flags', create_wc_flags_to_str(self.wc_flags)) +\
 *             print_format.format('comp mask', self.comp_mask) +\             # <<<<<<<<<<<<<<
 *             print_format.format('flags', self.flags)
 * 
*/
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_comp_mask); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_comp_mask_2, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/cq.pyx":310
 *         print_format = '{:22}: {:<20}\n'
 *         return print_format.format('Number of CQEs', self.cqe) +\
 *             print_format.format('WC flags', create_wc_flags_to_str(self.wc_flags)) +\             # <<<<<<<<<<<<<<
 *             print_format.format('comp mask', self.comp_mask) +\
 *             print_format.format('flags', self.flags)
*/
  __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":312
 *             print_format.format('WC flags', create_wc_flags_to_str(self.wc_flags)) +\
 *             print_format.format('comp mask', self.comp_mask) +\
 *             print_format.format('flags', self.flags)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_flags); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_n_u_flags, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":311
 *         return print_format.format('Number of CQEs', self.cqe) +\
 *             print_format.format('WC flags', create_wc_flags_to_str(self.wc_flags)) +\
 *             print_format.format('comp mask', self.comp_mask) +\             # <<<<<<<<<<<<<<
 *             print_format.format('flags', self.flags)
 * 
*/
  __pyx_t_2 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":307
 *         self.attr.flags = val
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         print_format = '{:22}: {:<20}\n'
 *         return print_format.format('Number of CQEs', self.cqe) +\
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyverbs.cq.CqInitAttrEx.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_print_format);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_12CqInitAttrEx_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_4__reduce_cython__(((struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_12CqInitAttrEx_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_attr_cannot_be_converted_to, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.CqInitAttrEx.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_12CqInitAttrEx_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cq.CqInitAttrEx.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2cq_12CqInitAttrEx_6__setstate_cython__(((struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_12CqInitAttrEx_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_attr_cannot_be_converted_to, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.CqInitAttrEx.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":316
 * 
 * cdef class CQEX(PyverbsCM):
 *     def __init__(self, Context context not None, CqInitAttrEx init_attr):             # <<<<<<<<<<<<<<
 *         """
 *         Initializes a CQEX object on the given device's context with the given
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_4CQEX_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2cq_4CQEX___init__, "\n        Initializes a CQEX object on the given device's context with the given\n        attributes.\n        :param context: The device's context on which to open the CQ\n        :param init_attr: Initial attributes that describe the CQ\n        :return: The newly created CQEX on success\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_7pyverbs_2cq_4CQEX___init__;
#endif
static int __pyx_pw_7pyverbs_2cq_4CQEX_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7pyverbs_6device_Context *__pyx_v_context = 0;
  struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_init_attr = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_context,&__pyx_mstate_global->__pyx_n_u_init_attr,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 316, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 316, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 316, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 316, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, i); __PYX_ERR(0, 316, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 316, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 316, __pyx_L3_error)
    }
    __pyx_v_context = ((struct __pyx_obj_7pyverbs_6device_Context *)values[0]);
    __pyx_v_init_attr = ((struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 316, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cq.CQEX.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_context), __pyx_mstate_global->__pyx_ptype_7pyverbs_6device_Context, 0, "context", 0))) __PYX_ERR(0, 316, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_init_attr), __pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx, 1, "init_attr", 0))) __PYX_ERR(0, 316, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX___init__(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self), __pyx_v_context, __pyx_v_init_attr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_4CQEX___init__(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self, struct __pyx_obj_7pyverbs_6device_Context *__pyx_v_context, struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *__pyx_v_init_attr) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_init_attr);

  /* "pyverbs/cq.pyx":324
 *         :return: The newly created CQEX on success
 *         """
 *         super().__init__()             # <<<<<<<<<<<<<<
 *         self.qps = weakref.WeakSet()
 *         self.srqs = weakref.WeakSet()
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 324, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 324, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":325
 *         """
 *         super().__init__()
 *         self.qps = weakref.WeakSet()             # <<<<<<<<<<<<<<
 *         self.srqs = weakref.WeakSet()
 *         self.wqs = weakref.WeakSet()
*/
  __pyx_t_3 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_weakref); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_WeakSet); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_3);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->qps);
  __Pyx_DECREF(__pyx_v_self->qps);
  __pyx_v_self->qps = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":326
 *         super().__init__()
 *         self.qps = weakref.WeakSet()
 *         self.srqs = weakref.WeakSet()             # <<<<<<<<<<<<<<
 *         self.wqs = weakref.WeakSet()
 *         if self.cq != NULL:
*/
  __pyx_t_4 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_weakref); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_WeakSet); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    assert(__pyx_t_4);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->srqs);
  __Pyx_DECREF(__pyx_v_self->srqs);
  __pyx_v_self->srqs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":327
 *         self.qps = weakref.WeakSet()
 *         self.srqs = weakref.WeakSet()
 *         self.wqs = weakref.WeakSet()             # <<<<<<<<<<<<<<
 *         if self.cq != NULL:
 *             # Leave CQ initialization to the provider
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_weakref); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_WeakSet); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->wqs);
  __Pyx_DECREF(__pyx_v_self->wqs);
  __pyx_v_self->wqs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":328
 *         self.srqs = weakref.WeakSet()
 *         self.wqs = weakref.WeakSet()
 *         if self.cq != NULL:             # <<<<<<<<<<<<<<
 *             # Leave CQ initialization to the provider
 *             return
*/
  __pyx_t_6 = (__pyx_v_self->cq != NULL);
  if (__pyx_t_6) {

    /* "pyverbs/cq.pyx":330
 *         if self.cq != NULL:
 *             # Leave CQ initialization to the provider
 *             return             # <<<<<<<<<<<<<<
 *         if init_attr is None:
 *             init_attr = CqInitAttrEx()
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "pyverbs/cq.pyx":328
 *         self.srqs = weakref.WeakSet()
 *         self.wqs = weakref.WeakSet()
 *         if self.cq != NULL:             # <<<<<<<<<<<<<<
 *             # Leave CQ initialization to the provider
 *             return
*/
  }

  /* "pyverbs/cq.pyx":331
 *             # Leave CQ initialization to the provider
 *             return
 *         if init_attr is None:             # <<<<<<<<<<<<<<
 *             init_attr = CqInitAttrEx()
 *         self.cq = v.ibv_create_cq_ex(context.context, &init_attr.attr)
*/
  __pyx_t_6 = (((PyObject *)__pyx_v_init_attr) == Py_None);
  if (__pyx_t_6) {

    /* "pyverbs/cq.pyx":332
 *             return
 *         if init_attr is None:
 *             init_attr = CqInitAttrEx()             # <<<<<<<<<<<<<<
 *         self.cq = v.ibv_create_cq_ex(context.context, &init_attr.attr)
 *         if init_attr.comp_channel:
*/
    __pyx_t_3 = NULL;
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 332, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_1);
    }
    __Pyx_DECREF_SET(__pyx_v_init_attr, ((struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "pyverbs/cq.pyx":331
 *             # Leave CQ initialization to the provider
 *             return
 *         if init_attr is None:             # <<<<<<<<<<<<<<
 *             init_attr = CqInitAttrEx()
 *         self.cq = v.ibv_create_cq_ex(context.context, &init_attr.attr)
*/
  }

  /* "pyverbs/cq.pyx":333
 *         if init_attr is None:
 *             init_attr = CqInitAttrEx()
 *         self.cq = v.ibv_create_cq_ex(context.context, &init_attr.attr)             # <<<<<<<<<<<<<<
 *         if init_attr.comp_channel:
 *             init_attr.comp_channel.add_ref(self)
*/
  __pyx_v_self->cq = ibv_create_cq_ex(__pyx_v_context->context, (&__pyx_v_init_attr->attr));

  /* "pyverbs/cq.pyx":334
 *             init_attr = CqInitAttrEx()
 *         self.cq = v.ibv_create_cq_ex(context.context, &init_attr.attr)
 *         if init_attr.comp_channel:             # <<<<<<<<<<<<<<
 *             init_attr.comp_channel.add_ref(self)
 *         if init_attr.parent_domain:
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_init_attr), __pyx_mstate_global->__pyx_n_u_comp_channel); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "pyverbs/cq.pyx":335
 *         self.cq = v.ibv_create_cq_ex(context.context, &init_attr.attr)
 *         if init_attr.comp_channel:
 *             init_attr.comp_channel.add_ref(self)             # <<<<<<<<<<<<<<
 *         if init_attr.parent_domain:
 *             (<ParentDomain>init_attr.parent_domain).add_ref(self)
*/
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_init_attr), __pyx_mstate_global->__pyx_n_u_comp_channel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_t_2;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_v_self)};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_add_ref, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 335, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyverbs/cq.pyx":334
 *             init_attr = CqInitAttrEx()
 *         self.cq = v.ibv_create_cq_ex(context.context, &init_attr.attr)
 *         if init_attr.comp_channel:             # <<<<<<<<<<<<<<
 *             init_attr.comp_channel.add_ref(self)
 *         if init_attr.parent_domain:
*/
  }

  /* "pyverbs/cq.pyx":336
 *         if init_attr.comp_channel:
 *             init_attr.comp_channel.add_ref(self)
 *         if init_attr.parent_domain:             # <<<<<<<<<<<<<<
 *             (<ParentDomain>init_attr.parent_domain).add_ref(self)
 *         if self.cq == NULL:
*/
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_init_attr->parent_domain); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 336, __pyx_L1_error)
  if (__pyx_t_6) {

    /* "pyverbs/cq.pyx":337
 *             init_attr.comp_channel.add_ref(self)
 *         if init_attr.parent_domain:
 *             (<ParentDomain>init_attr.parent_domain).add_ref(self)             # <<<<<<<<<<<<<<
 *         if self.cq == NULL:
 *             raise PyverbsRDMAErrno('Failed to create extended CQ')
*/
    __pyx_t_1 = ((struct __pyx_vtabstruct_7pyverbs_2pd_ParentDomain *)((struct __pyx_obj_7pyverbs_2pd_ParentDomain *)__pyx_v_init_attr->parent_domain)->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.add_ref(((struct __pyx_obj_7pyverbs_2pd_PD *)((struct __pyx_obj_7pyverbs_2pd_ParentDomain *)__pyx_v_init_attr->parent_domain)), ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyverbs/cq.pyx":336
 *         if init_attr.comp_channel:
 *             init_attr.comp_channel.add_ref(self)
 *         if init_attr.parent_domain:             # <<<<<<<<<<<<<<
 *             (<ParentDomain>init_attr.parent_domain).add_ref(self)
 *         if self.cq == NULL:
*/
  }

  /* "pyverbs/cq.pyx":338
 *         if init_attr.parent_domain:
 *             (<ParentDomain>init_attr.parent_domain).add_ref(self)
 *         if self.cq == NULL:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to create extended CQ')
 *         self.ibv_cq = v.ibv_cq_ex_to_cq(self.cq)
*/
  __pyx_t_6 = (__pyx_v_self->cq == NULL);
  if (unlikely(__pyx_t_6)) {

    /* "pyverbs/cq.pyx":339
 *             (<ParentDomain>init_attr.parent_domain).add_ref(self)
 *         if self.cq == NULL:
 *             raise PyverbsRDMAErrno('Failed to create extended CQ')             # <<<<<<<<<<<<<<
 *         self.ibv_cq = v.ibv_cq_ex_to_cq(self.cq)
 *         self.context = context
*/
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 339, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Failed_to_create_extended_CQ};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 339, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 339, __pyx_L1_error)

    /* "pyverbs/cq.pyx":338
 *         if init_attr.parent_domain:
 *             (<ParentDomain>init_attr.parent_domain).add_ref(self)
 *         if self.cq == NULL:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to create extended CQ')
 *         self.ibv_cq = v.ibv_cq_ex_to_cq(self.cq)
*/
  }

  /* "pyverbs/cq.pyx":340
 *         if self.cq == NULL:
 *             raise PyverbsRDMAErrno('Failed to create extended CQ')
 *         self.ibv_cq = v.ibv_cq_ex_to_cq(self.cq)             # <<<<<<<<<<<<<<
 *         self.context = context
 *         context.add_ref(self)
*/
  __pyx_v_self->ibv_cq = ibv_cq_ex_to_cq(__pyx_v_self->cq);

  /* "pyverbs/cq.pyx":341
 *             raise PyverbsRDMAErrno('Failed to create extended CQ')
 *         self.ibv_cq = v.ibv_cq_ex_to_cq(self.cq)
 *         self.context = context             # <<<<<<<<<<<<<<
 *         context.add_ref(self)
 * 
*/
  __Pyx_INCREF((PyObject *)__pyx_v_context);
  __Pyx_GIVEREF((PyObject *)__pyx_v_context);
  __Pyx_GOTREF(__pyx_v_self->context);
  __Pyx_DECREF(__pyx_v_self->context);
  __pyx_v_self->context = ((PyObject *)__pyx_v_context);

  /* "pyverbs/cq.pyx":342
 *         self.ibv_cq = v.ibv_cq_ex_to_cq(self.cq)
 *         self.context = context
 *         context.add_ref(self)             # <<<<<<<<<<<<<<
 * 
 *     cdef add_ref(self, obj):
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_7pyverbs_6device_Context *)__pyx_v_context->__pyx_base.__pyx_vtab)->add_ref(__pyx_v_context, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":316
 * 
 * cdef class CQEX(PyverbsCM):
 *     def __init__(self, Context context not None, CqInitAttrEx init_attr):             # <<<<<<<<<<<<<<
 *         """
 *         Initializes a CQEX object on the given device's context with the given
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_init_attr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":344
 *         context.add_ref(self)
 * 
 *     cdef add_ref(self, obj):             # <<<<<<<<<<<<<<
 *         if isinstance(obj, QP):
 *             self.qps.add(obj)
*/

static PyObject *__pyx_f_7pyverbs_2cq_4CQEX_add_ref(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_ref", 0);

  /* "pyverbs/cq.pyx":345
 * 
 *     cdef add_ref(self, obj):
 *         if isinstance(obj, QP):             # <<<<<<<<<<<<<<
 *             self.qps.add(obj)
 *         elif isinstance(obj, SRQ):
*/
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QP); 
  if (__pyx_t_1) {

    /* "pyverbs/cq.pyx":346
 *     cdef add_ref(self, obj):
 *         if isinstance(obj, QP):
 *             self.qps.add(obj)             # <<<<<<<<<<<<<<
 *         elif isinstance(obj, SRQ):
 *             self.srqs.add(obj)
*/
    __pyx_t_3 = __pyx_v_self->qps;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_obj};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_add, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 346, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyverbs/cq.pyx":345
 * 
 *     cdef add_ref(self, obj):
 *         if isinstance(obj, QP):             # <<<<<<<<<<<<<<
 *             self.qps.add(obj)
 *         elif isinstance(obj, SRQ):
*/
    goto __pyx_L3;
  }

  /* "pyverbs/cq.pyx":347
 *         if isinstance(obj, QP):
 *             self.qps.add(obj)
 *         elif isinstance(obj, SRQ):             # <<<<<<<<<<<<<<
 *             self.srqs.add(obj)
 *         elif isinstance(obj, WQ):
*/
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_mstate_global->__pyx_ptype_7pyverbs_3srq_SRQ); 
  if (__pyx_t_1) {

    /* "pyverbs/cq.pyx":348
 *             self.qps.add(obj)
 *         elif isinstance(obj, SRQ):
 *             self.srqs.add(obj)             # <<<<<<<<<<<<<<
 *         elif isinstance(obj, WQ):
 *             self.wqs.add(obj)
*/
    __pyx_t_3 = __pyx_v_self->srqs;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_obj};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_add, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 348, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyverbs/cq.pyx":347
 *         if isinstance(obj, QP):
 *             self.qps.add(obj)
 *         elif isinstance(obj, SRQ):             # <<<<<<<<<<<<<<
 *             self.srqs.add(obj)
 *         elif isinstance(obj, WQ):
*/
    goto __pyx_L3;
  }

  /* "pyverbs/cq.pyx":349
 *         elif isinstance(obj, SRQ):
 *             self.srqs.add(obj)
 *         elif isinstance(obj, WQ):             # <<<<<<<<<<<<<<
 *             self.wqs.add(obj)
 *         else:
*/
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_mstate_global->__pyx_ptype_7pyverbs_2wq_WQ); 
  if (likely(__pyx_t_1)) {

    /* "pyverbs/cq.pyx":350
 *             self.srqs.add(obj)
 *         elif isinstance(obj, WQ):
 *             self.wqs.add(obj)             # <<<<<<<<<<<<<<
 *         else:
 *             raise PyverbsError('Unrecognized object type')
*/
    __pyx_t_3 = __pyx_v_self->wqs;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_obj};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_add, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 350, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyverbs/cq.pyx":349
 *         elif isinstance(obj, SRQ):
 *             self.srqs.add(obj)
 *         elif isinstance(obj, WQ):             # <<<<<<<<<<<<<<
 *             self.wqs.add(obj)
 *         else:
*/
    goto __pyx_L3;
  }

  /* "pyverbs/cq.pyx":352
 *             self.wqs.add(obj)
 *         else:
 *             raise PyverbsError('Unrecognized object type')             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
*/
  /*else*/ {
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PyverbsError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 352, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_4 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Unrecognized_object_type};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 352, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 352, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "pyverbs/cq.pyx":344
 *         context.add_ref(self)
 * 
 *     cdef add_ref(self, obj):             # <<<<<<<<<<<<<<
 *         if isinstance(obj, QP):
 *             self.qps.add(obj)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.add_ref", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":354
 *             raise PyverbsError('Unrecognized object type')
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         self.close()
 * 
*/

/* Python wrapper */
static void __pyx_pw_7pyverbs_2cq_4CQEX_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7pyverbs_2cq_4CQEX_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_7pyverbs_2cq_4CQEX_2__dealloc__(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7pyverbs_2cq_4CQEX_2__dealloc__(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "pyverbs/cq.pyx":355
 * 
 *     def __dealloc__(self):
 *         self.close()             # <<<<<<<<<<<<<<
 * 
 *     cpdef close(self):
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_7pyverbs_2cq_CQEX *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.close(((struct __pyx_obj_7pyverbs_4base_PyverbsCM *)__pyx_v_self), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":354
 *             raise PyverbsError('Unrecognized object type')
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         self.close()
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("pyverbs.cq.CQEX.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "pyverbs/cq.pyx":357
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.cq != NULL:
 *             if self.logger:
*/

static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_5close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7pyverbs_2cq_4CQEX_close(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self, int __pyx_skip_dispatch) {
  int __pyx_v_rc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_close); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 357, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyverbs_2cq_4CQEX_5close)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 357, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyverbs/cq.pyx":358
 * 
 *     cpdef close(self):
 *         if self.cq != NULL:             # <<<<<<<<<<<<<<
 *             if self.logger:
 *                 self.logger.debug('Closing CQEx')
*/
  __pyx_t_6 = (__pyx_v_self->cq != NULL);
  if (__pyx_t_6) {

    /* "pyverbs/cq.pyx":359
 *     cpdef close(self):
 *         if self.cq != NULL:
 *             if self.logger:             # <<<<<<<<<<<<<<
 *                 self.logger.debug('Closing CQEx')
 *             close_weakrefs([self.srqs, self.qps, self.wqs])
*/
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_self->__pyx_base.__pyx_base.logger); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 359, __pyx_L1_error)
    if (__pyx_t_6) {

      /* "pyverbs/cq.pyx":360
 *         if self.cq != NULL:
 *             if self.logger:
 *                 self.logger.debug('Closing CQEx')             # <<<<<<<<<<<<<<
 *             close_weakrefs([self.srqs, self.qps, self.wqs])
 *             rc = v.ibv_destroy_cq(<v.ibv_cq*>self.cq)
*/
      __pyx_t_2 = __pyx_v_self->__pyx_base.__pyx_base.logger;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_5 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Closing_CQEx};
        __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_debug, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 360, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyverbs/cq.pyx":359
 *     cpdef close(self):
 *         if self.cq != NULL:
 *             if self.logger:             # <<<<<<<<<<<<<<
 *                 self.logger.debug('Closing CQEx')
 *             close_weakrefs([self.srqs, self.qps, self.wqs])
*/
    }

    /* "pyverbs/cq.pyx":361
 *             if self.logger:
 *                 self.logger.debug('Closing CQEx')
 *             close_weakrefs([self.srqs, self.qps, self.wqs])             # <<<<<<<<<<<<<<
 *             rc = v.ibv_destroy_cq(<v.ibv_cq*>self.cq)
 *             if rc != 0:
*/
    __pyx_t_1 = PyList_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_self->srqs);
    __Pyx_GIVEREF(__pyx_v_self->srqs);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->srqs) != (0)) __PYX_ERR(0, 361, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_self->qps);
    __Pyx_GIVEREF(__pyx_v_self->qps);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->qps) != (0)) __PYX_ERR(0, 361, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_self->wqs);
    __Pyx_GIVEREF(__pyx_v_self->wqs);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->wqs) != (0)) __PYX_ERR(0, 361, __pyx_L1_error);
    __pyx_t_2 = __pyx_f_7pyverbs_4base_close_weakrefs(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyverbs/cq.pyx":362
 *                 self.logger.debug('Closing CQEx')
 *             close_weakrefs([self.srqs, self.qps, self.wqs])
 *             rc = v.ibv_destroy_cq(<v.ibv_cq*>self.cq)             # <<<<<<<<<<<<<<
 *             if rc != 0:
 *                 raise PyverbsRDMAError('Failed to destroy CQEX', rc)
*/
    __pyx_v_rc = ibv_destroy_cq(((struct ibv_cq *)__pyx_v_self->cq));

    /* "pyverbs/cq.pyx":363
 *             close_weakrefs([self.srqs, self.qps, self.wqs])
 *             rc = v.ibv_destroy_cq(<v.ibv_cq*>self.cq)
 *             if rc != 0:             # <<<<<<<<<<<<<<
 *                 raise PyverbsRDMAError('Failed to destroy CQEX', rc)
 *             self.cq = NULL
*/
    __pyx_t_6 = (__pyx_v_rc != 0);
    if (unlikely(__pyx_t_6)) {

      /* "pyverbs/cq.pyx":364
 *             rc = v.ibv_destroy_cq(<v.ibv_cq*>self.cq)
 *             if rc != 0:
 *                 raise PyverbsRDMAError('Failed to destroy CQEX', rc)             # <<<<<<<<<<<<<<
 *             self.cq = NULL
 *             self.context = None
*/
      __pyx_t_1 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 364, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyLong_From_int(__pyx_v_rc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 364, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
        assert(__pyx_t_1);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
        __pyx_t_5 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_mstate_global->__pyx_kp_u_Failed_to_destroy_CQEX, __pyx_t_3};
        __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 364, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 364, __pyx_L1_error)

      /* "pyverbs/cq.pyx":363
 *             close_weakrefs([self.srqs, self.qps, self.wqs])
 *             rc = v.ibv_destroy_cq(<v.ibv_cq*>self.cq)
 *             if rc != 0:             # <<<<<<<<<<<<<<
 *                 raise PyverbsRDMAError('Failed to destroy CQEX', rc)
 *             self.cq = NULL
*/
    }

    /* "pyverbs/cq.pyx":365
 *             if rc != 0:
 *                 raise PyverbsRDMAError('Failed to destroy CQEX', rc)
 *             self.cq = NULL             # <<<<<<<<<<<<<<
 *             self.context = None
 * 
*/
    __pyx_v_self->cq = NULL;

    /* "pyverbs/cq.pyx":366
 *                 raise PyverbsRDMAError('Failed to destroy CQEX', rc)
 *             self.cq = NULL
 *             self.context = None             # <<<<<<<<<<<<<<
 * 
 *     def start_poll(self, PollCqAttr attr):
*/
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->context);
    __Pyx_DECREF(__pyx_v_self->context);
    __pyx_v_self->context = Py_None;

    /* "pyverbs/cq.pyx":358
 * 
 *     cpdef close(self):
 *         if self.cq != NULL:             # <<<<<<<<<<<<<<
 *             if self.logger:
 *                 self.logger.debug('Closing CQEx')
*/
  }

  /* "pyverbs/cq.pyx":357
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.cq != NULL:
 *             if self.logger:
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_5close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_4CQEX_5close = {"close", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_5close, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_5close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("close", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("close", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_4close(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_4close(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7pyverbs_2cq_4CQEX_close(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":368
 *             self.context = None
 * 
 *     def start_poll(self, PollCqAttr attr):             # <<<<<<<<<<<<<<
 *         """
 *         Start polling a batch of work completions.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_7start_poll(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2cq_4CQEX_6start_poll, "\n        Start polling a batch of work completions.\n        :param attr: For easy future extensions\n        :return: 0 on success, ENOENT when no completions are available\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_2cq_4CQEX_7start_poll = {"start_poll", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_7start_poll, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2cq_4CQEX_6start_poll};
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_7start_poll(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_2cq_PollCqAttr *__pyx_v_attr = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("start_poll (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_attr,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 368, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 368, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "start_poll", 0) < (0)) __PYX_ERR(0, 368, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("start_poll", 1, 1, 1, i); __PYX_ERR(0, 368, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 368, __pyx_L3_error)
    }
    __pyx_v_attr = ((struct __pyx_obj_7pyverbs_2cq_PollCqAttr *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("start_poll", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 368, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cq.CQEX.start_poll", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_attr), __pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_PollCqAttr, 1, "attr", 0))) __PYX_ERR(0, 368, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_6start_poll(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self), __pyx_v_attr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_6start_poll(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self, struct __pyx_obj_7pyverbs_2cq_PollCqAttr *__pyx_v_attr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("start_poll", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_attr);

  /* "pyverbs/cq.pyx":374
 *         :return: 0 on success, ENOENT when no completions are available
 *         """
 *         if attr is None:             # <<<<<<<<<<<<<<
 *             attr = PollCqAttr()
 *         return v.ibv_start_poll(self.cq, &attr.attr)
*/
  __pyx_t_1 = (((PyObject *)__pyx_v_attr) == Py_None);
  if (__pyx_t_1) {

    /* "pyverbs/cq.pyx":375
 *         """
 *         if attr is None:
 *             attr = PollCqAttr()             # <<<<<<<<<<<<<<
 *         return v.ibv_start_poll(self.cq, &attr.attr)
 * 
*/
    __pyx_t_3 = NULL;
    __pyx_t_4 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_PollCqAttr, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 375, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_2);
    }
    __Pyx_DECREF_SET(__pyx_v_attr, ((struct __pyx_obj_7pyverbs_2cq_PollCqAttr *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "pyverbs/cq.pyx":374
 *         :return: 0 on success, ENOENT when no completions are available
 *         """
 *         if attr is None:             # <<<<<<<<<<<<<<
 *             attr = PollCqAttr()
 *         return v.ibv_start_poll(self.cq, &attr.attr)
*/
  }

  /* "pyverbs/cq.pyx":376
 *         if attr is None:
 *             attr = PollCqAttr()
 *         return v.ibv_start_poll(self.cq, &attr.attr)             # <<<<<<<<<<<<<<
 * 
 *     def poll_next(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyLong_From_int(ibv_start_poll(__pyx_v_self->cq, (&__pyx_v_attr->attr))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":368
 *             self.context = None
 * 
 *     def start_poll(self, PollCqAttr attr):             # <<<<<<<<<<<<<<
 *         """
 *         Start polling a batch of work completions.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.start_poll", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_attr);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":378
 *         return v.ibv_start_poll(self.cq, &attr.attr)
 * 
 *     def poll_next(self):             # <<<<<<<<<<<<<<
 *         """
 *         Get the next work completion.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_9poll_next(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2cq_4CQEX_8poll_next, "\n        Get the next work completion.\n        :return: 0 on success, ENOENT when no completions are available\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_2cq_4CQEX_9poll_next = {"poll_next", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_9poll_next, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2cq_4CQEX_8poll_next};
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_9poll_next(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("poll_next (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("poll_next", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("poll_next", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_8poll_next(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_8poll_next(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("poll_next", 0);

  /* "pyverbs/cq.pyx":383
 *         :return: 0 on success, ENOENT when no completions are available
 *         """
 *         return v.ibv_next_poll(self.cq)             # <<<<<<<<<<<<<<
 * 
 *     def end_poll(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(ibv_next_poll(__pyx_v_self->cq)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":378
 *         return v.ibv_start_poll(self.cq, &attr.attr)
 * 
 *     def poll_next(self):             # <<<<<<<<<<<<<<
 *         """
 *         Get the next work completion.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.poll_next", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":385
 *         return v.ibv_next_poll(self.cq)
 * 
 *     def end_poll(self):             # <<<<<<<<<<<<<<
 *         """
 *         Indicates the end of polling batch of work completions
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_11end_poll(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_2cq_4CQEX_10end_poll, "\n        Indicates the end of polling batch of work completions\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_2cq_4CQEX_11end_poll = {"end_poll", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_11end_poll, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2cq_4CQEX_10end_poll};
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_11end_poll(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("end_poll (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("end_poll", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("end_poll", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_10end_poll(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_10end_poll(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("end_poll", 0);

  /* "pyverbs/cq.pyx":390
 *         :return: None
 *         """
 *         return v.ibv_end_poll(self.cq)             # <<<<<<<<<<<<<<
 * 
 *     def read_opcode(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(ibv_end_poll(__pyx_v_self->cq)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":385
 *         return v.ibv_next_poll(self.cq)
 * 
 *     def end_poll(self):             # <<<<<<<<<<<<<<
 *         """
 *         Indicates the end of polling batch of work completions
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.end_poll", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":392
 *         return v.ibv_end_poll(self.cq)
 * 
 *     def read_opcode(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_opcode(self.cq)
 *     def read_vendor_err(self):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_13read_opcode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_4CQEX_13read_opcode = {"read_opcode", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_13read_opcode, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_13read_opcode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read_opcode (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("read_opcode", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("read_opcode", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_12read_opcode(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_12read_opcode(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_opcode", 0);

  /* "pyverbs/cq.pyx":393
 * 
 *     def read_opcode(self):
 *         return v.ibv_wc_read_opcode(self.cq)             # <<<<<<<<<<<<<<
 *     def read_vendor_err(self):
 *         return v.ibv_wc_read_vendor_err(self.cq)
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py(ibv_wc_read_opcode(__pyx_v_self->cq)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":392
 *         return v.ibv_end_poll(self.cq)
 * 
 *     def read_opcode(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_opcode(self.cq)
 *     def read_vendor_err(self):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.read_opcode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":394
 *     def read_opcode(self):
 *         return v.ibv_wc_read_opcode(self.cq)
 *     def read_vendor_err(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_vendor_err(self.cq)
 *     def read_byte_len(self):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_15read_vendor_err(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_4CQEX_15read_vendor_err = {"read_vendor_err", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_15read_vendor_err, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_15read_vendor_err(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read_vendor_err (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("read_vendor_err", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("read_vendor_err", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_14read_vendor_err(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_14read_vendor_err(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_vendor_err", 0);

  /* "pyverbs/cq.pyx":395
 *         return v.ibv_wc_read_opcode(self.cq)
 *     def read_vendor_err(self):
 *         return v.ibv_wc_read_vendor_err(self.cq)             # <<<<<<<<<<<<<<
 *     def read_byte_len(self):
 *         return v.ibv_wc_read_byte_len(self.cq)
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(ibv_wc_read_vendor_err(__pyx_v_self->cq)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":394
 *     def read_opcode(self):
 *         return v.ibv_wc_read_opcode(self.cq)
 *     def read_vendor_err(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_vendor_err(self.cq)
 *     def read_byte_len(self):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.read_vendor_err", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":396
 *     def read_vendor_err(self):
 *         return v.ibv_wc_read_vendor_err(self.cq)
 *     def read_byte_len(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_byte_len(self.cq)
 *     def read_imm_data(self):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_17read_byte_len(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_4CQEX_17read_byte_len = {"read_byte_len", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_17read_byte_len, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_17read_byte_len(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read_byte_len (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("read_byte_len", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("read_byte_len", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_16read_byte_len(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_16read_byte_len(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_byte_len", 0);

  /* "pyverbs/cq.pyx":397
 *         return v.ibv_wc_read_vendor_err(self.cq)
 *     def read_byte_len(self):
 *         return v.ibv_wc_read_byte_len(self.cq)             # <<<<<<<<<<<<<<
 *     def read_imm_data(self):
 *         return v.ibv_wc_read_imm_data(self.cq)
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(ibv_wc_read_byte_len(__pyx_v_self->cq)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":396
 *     def read_vendor_err(self):
 *         return v.ibv_wc_read_vendor_err(self.cq)
 *     def read_byte_len(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_byte_len(self.cq)
 *     def read_imm_data(self):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.read_byte_len", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":398
 *     def read_byte_len(self):
 *         return v.ibv_wc_read_byte_len(self.cq)
 *     def read_imm_data(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_imm_data(self.cq)
 *     def read_qp_num(self):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_19read_imm_data(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_4CQEX_19read_imm_data = {"read_imm_data", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_19read_imm_data, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_19read_imm_data(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read_imm_data (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("read_imm_data", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("read_imm_data", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_18read_imm_data(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_18read_imm_data(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_imm_data", 0);

  /* "pyverbs/cq.pyx":399
 *         return v.ibv_wc_read_byte_len(self.cq)
 *     def read_imm_data(self):
 *         return v.ibv_wc_read_imm_data(self.cq)             # <<<<<<<<<<<<<<
 *     def read_qp_num(self):
 *         return v.ibv_wc_read_qp_num(self.cq)
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(ibv_wc_read_imm_data(__pyx_v_self->cq)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":398
 *     def read_byte_len(self):
 *         return v.ibv_wc_read_byte_len(self.cq)
 *     def read_imm_data(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_imm_data(self.cq)
 *     def read_qp_num(self):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.read_imm_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":400
 *     def read_imm_data(self):
 *         return v.ibv_wc_read_imm_data(self.cq)
 *     def read_qp_num(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_qp_num(self.cq)
 *     def read_src_qp(self):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_21read_qp_num(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_4CQEX_21read_qp_num = {"read_qp_num", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_21read_qp_num, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_21read_qp_num(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read_qp_num (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("read_qp_num", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("read_qp_num", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_20read_qp_num(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_20read_qp_num(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_qp_num", 0);

  /* "pyverbs/cq.pyx":401
 *         return v.ibv_wc_read_imm_data(self.cq)
 *     def read_qp_num(self):
 *         return v.ibv_wc_read_qp_num(self.cq)             # <<<<<<<<<<<<<<
 *     def read_src_qp(self):
 *         return v.ibv_wc_read_src_qp(self.cq)
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(ibv_wc_read_qp_num(__pyx_v_self->cq)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":400
 *     def read_imm_data(self):
 *         return v.ibv_wc_read_imm_data(self.cq)
 *     def read_qp_num(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_qp_num(self.cq)
 *     def read_src_qp(self):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.read_qp_num", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":402
 *     def read_qp_num(self):
 *         return v.ibv_wc_read_qp_num(self.cq)
 *     def read_src_qp(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_src_qp(self.cq)
 *     def read_wc_flags(self):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_23read_src_qp(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_4CQEX_23read_src_qp = {"read_src_qp", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_23read_src_qp, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_23read_src_qp(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read_src_qp (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("read_src_qp", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("read_src_qp", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_22read_src_qp(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_22read_src_qp(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_src_qp", 0);

  /* "pyverbs/cq.pyx":403
 *         return v.ibv_wc_read_qp_num(self.cq)
 *     def read_src_qp(self):
 *         return v.ibv_wc_read_src_qp(self.cq)             # <<<<<<<<<<<<<<
 *     def read_wc_flags(self):
 *         return v.ibv_wc_read_wc_flags(self.cq)
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(ibv_wc_read_src_qp(__pyx_v_self->cq)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":402
 *     def read_qp_num(self):
 *         return v.ibv_wc_read_qp_num(self.cq)
 *     def read_src_qp(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_src_qp(self.cq)
 *     def read_wc_flags(self):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.read_src_qp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":404
 *     def read_src_qp(self):
 *         return v.ibv_wc_read_src_qp(self.cq)
 *     def read_wc_flags(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_wc_flags(self.cq)
 *     def read_slid(self):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_25read_wc_flags(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_4CQEX_25read_wc_flags = {"read_wc_flags", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_25read_wc_flags, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_25read_wc_flags(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read_wc_flags (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("read_wc_flags", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("read_wc_flags", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_24read_wc_flags(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_24read_wc_flags(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_wc_flags", 0);

  /* "pyverbs/cq.pyx":405
 *         return v.ibv_wc_read_src_qp(self.cq)
 *     def read_wc_flags(self):
 *         return v.ibv_wc_read_wc_flags(self.cq)             # <<<<<<<<<<<<<<
 *     def read_slid(self):
 *         return v.ibv_wc_read_slid(self.cq)
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(ibv_wc_read_wc_flags(__pyx_v_self->cq)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":404
 *     def read_src_qp(self):
 *         return v.ibv_wc_read_src_qp(self.cq)
 *     def read_wc_flags(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_wc_flags(self.cq)
 *     def read_slid(self):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.read_wc_flags", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":406
 *     def read_wc_flags(self):
 *         return v.ibv_wc_read_wc_flags(self.cq)
 *     def read_slid(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_slid(self.cq)
 *     def read_sl(self):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_27read_slid(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_4CQEX_27read_slid = {"read_slid", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_27read_slid, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_27read_slid(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read_slid (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("read_slid", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("read_slid", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_26read_slid(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_26read_slid(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_slid", 0);

  /* "pyverbs/cq.pyx":407
 *         return v.ibv_wc_read_wc_flags(self.cq)
 *     def read_slid(self):
 *         return v.ibv_wc_read_slid(self.cq)             # <<<<<<<<<<<<<<
 *     def read_sl(self):
 *         return v.ibv_wc_read_sl(self.cq)
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(ibv_wc_read_slid(__pyx_v_self->cq)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":406
 *     def read_wc_flags(self):
 *         return v.ibv_wc_read_wc_flags(self.cq)
 *     def read_slid(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_slid(self.cq)
 *     def read_sl(self):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.read_slid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":408
 *     def read_slid(self):
 *         return v.ibv_wc_read_slid(self.cq)
 *     def read_sl(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_sl(self.cq)
 *     def read_dlid_path_bits(self):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_29read_sl(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_4CQEX_29read_sl = {"read_sl", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_29read_sl, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_29read_sl(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read_sl (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("read_sl", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("read_sl", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_28read_sl(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_28read_sl(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_sl", 0);

  /* "pyverbs/cq.pyx":409
 *         return v.ibv_wc_read_slid(self.cq)
 *     def read_sl(self):
 *         return v.ibv_wc_read_sl(self.cq)             # <<<<<<<<<<<<<<
 *     def read_dlid_path_bits(self):
 *         return v.ibv_wc_read_dlid_path_bits(self.cq)
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_char(ibv_wc_read_sl(__pyx_v_self->cq)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":408
 *     def read_slid(self):
 *         return v.ibv_wc_read_slid(self.cq)
 *     def read_sl(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_sl(self.cq)
 *     def read_dlid_path_bits(self):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.read_sl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":410
 *     def read_sl(self):
 *         return v.ibv_wc_read_sl(self.cq)
 *     def read_dlid_path_bits(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_dlid_path_bits(self.cq)
 *     def read_timestamp(self):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_31read_dlid_path_bits(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_4CQEX_31read_dlid_path_bits = {"read_dlid_path_bits", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_31read_dlid_path_bits, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_31read_dlid_path_bits(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read_dlid_path_bits (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("read_dlid_path_bits", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("read_dlid_path_bits", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_30read_dlid_path_bits(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_30read_dlid_path_bits(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_dlid_path_bits", 0);

  /* "pyverbs/cq.pyx":411
 *         return v.ibv_wc_read_sl(self.cq)
 *     def read_dlid_path_bits(self):
 *         return v.ibv_wc_read_dlid_path_bits(self.cq)             # <<<<<<<<<<<<<<
 *     def read_timestamp(self):
 *         return v.ibv_wc_read_completion_ts(self.cq)
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_char(ibv_wc_read_dlid_path_bits(__pyx_v_self->cq)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":410
 *     def read_sl(self):
 *         return v.ibv_wc_read_sl(self.cq)
 *     def read_dlid_path_bits(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_dlid_path_bits(self.cq)
 *     def read_timestamp(self):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.read_dlid_path_bits", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":412
 *     def read_dlid_path_bits(self):
 *         return v.ibv_wc_read_dlid_path_bits(self.cq)
 *     def read_timestamp(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_completion_ts(self.cq)
 *     def read_cvlan(self):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_33read_timestamp(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_4CQEX_33read_timestamp = {"read_timestamp", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_33read_timestamp, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_33read_timestamp(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read_timestamp (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("read_timestamp", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("read_timestamp", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_32read_timestamp(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_32read_timestamp(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_timestamp", 0);

  /* "pyverbs/cq.pyx":413
 *         return v.ibv_wc_read_dlid_path_bits(self.cq)
 *     def read_timestamp(self):
 *         return v.ibv_wc_read_completion_ts(self.cq)             # <<<<<<<<<<<<<<
 *     def read_cvlan(self):
 *         return v.ibv_wc_read_cvlan(self.cq)
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_long(ibv_wc_read_completion_ts(__pyx_v_self->cq)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":412
 *     def read_dlid_path_bits(self):
 *         return v.ibv_wc_read_dlid_path_bits(self.cq)
 *     def read_timestamp(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_completion_ts(self.cq)
 *     def read_cvlan(self):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.read_timestamp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":414
 *     def read_timestamp(self):
 *         return v.ibv_wc_read_completion_ts(self.cq)
 *     def read_cvlan(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_cvlan(self.cq)
 *     def read_flow_tag(self):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_35read_cvlan(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_4CQEX_35read_cvlan = {"read_cvlan", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_35read_cvlan, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_35read_cvlan(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read_cvlan (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("read_cvlan", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("read_cvlan", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_34read_cvlan(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_34read_cvlan(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_cvlan", 0);

  /* "pyverbs/cq.pyx":415
 *         return v.ibv_wc_read_completion_ts(self.cq)
 *     def read_cvlan(self):
 *         return v.ibv_wc_read_cvlan(self.cq)             # <<<<<<<<<<<<<<
 *     def read_flow_tag(self):
 *         return v.ibv_wc_read_flow_tag(self.cq)
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_short(ibv_wc_read_cvlan(__pyx_v_self->cq)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":414
 *     def read_timestamp(self):
 *         return v.ibv_wc_read_completion_ts(self.cq)
 *     def read_cvlan(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_cvlan(self.cq)
 *     def read_flow_tag(self):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.read_cvlan", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":416
 *     def read_cvlan(self):
 *         return v.ibv_wc_read_cvlan(self.cq)
 *     def read_flow_tag(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_flow_tag(self.cq)
 *     def read_tm_info(self):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_37read_flow_tag(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_4CQEX_37read_flow_tag = {"read_flow_tag", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_37read_flow_tag, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_37read_flow_tag(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read_flow_tag (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("read_flow_tag", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("read_flow_tag", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_36read_flow_tag(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_36read_flow_tag(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_flow_tag", 0);

  /* "pyverbs/cq.pyx":417
 *         return v.ibv_wc_read_cvlan(self.cq)
 *     def read_flow_tag(self):
 *         return v.ibv_wc_read_flow_tag(self.cq)             # <<<<<<<<<<<<<<
 *     def read_tm_info(self):
 *         info = WcTmInfo()
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(ibv_wc_read_flow_tag(__pyx_v_self->cq)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":416
 *     def read_cvlan(self):
 *         return v.ibv_wc_read_cvlan(self.cq)
 *     def read_flow_tag(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_flow_tag(self.cq)
 *     def read_tm_info(self):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.read_flow_tag", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":418
 *     def read_flow_tag(self):
 *         return v.ibv_wc_read_flow_tag(self.cq)
 *     def read_tm_info(self):             # <<<<<<<<<<<<<<
 *         info = WcTmInfo()
 *         v.ibv_wc_read_tm_info(self.cq, &info.info)
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_39read_tm_info(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_4CQEX_39read_tm_info = {"read_tm_info", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_39read_tm_info, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_39read_tm_info(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read_tm_info (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("read_tm_info", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("read_tm_info", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_38read_tm_info(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_38read_tm_info(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self) {
  struct __pyx_obj_7pyverbs_2cq_WcTmInfo *__pyx_v_info = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_tm_info", 0);

  /* "pyverbs/cq.pyx":419
 *         return v.ibv_wc_read_flow_tag(self.cq)
 *     def read_tm_info(self):
 *         info = WcTmInfo()             # <<<<<<<<<<<<<<
 *         v.ibv_wc_read_tm_info(self.cq, &info.info)
 *         return info
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_WcTmInfo, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 419, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_info = ((struct __pyx_obj_7pyverbs_2cq_WcTmInfo *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":420
 *     def read_tm_info(self):
 *         info = WcTmInfo()
 *         v.ibv_wc_read_tm_info(self.cq, &info.info)             # <<<<<<<<<<<<<<
 *         return info
 *     def read_completion_wallclock_ns(self):
*/
  ibv_wc_read_tm_info(__pyx_v_self->cq, (&__pyx_v_info->info));

  /* "pyverbs/cq.pyx":421
 *         info = WcTmInfo()
 *         v.ibv_wc_read_tm_info(self.cq, &info.info)
 *         return info             # <<<<<<<<<<<<<<
 *     def read_completion_wallclock_ns(self):
 *         return v.ibv_wc_read_completion_wallclock_ns(self.cq)
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_info);
  __pyx_r = ((PyObject *)__pyx_v_info);
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":418
 *     def read_flow_tag(self):
 *         return v.ibv_wc_read_flow_tag(self.cq)
 *     def read_tm_info(self):             # <<<<<<<<<<<<<<
 *         info = WcTmInfo()
 *         v.ibv_wc_read_tm_info(self.cq, &info.info)
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.read_tm_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_info);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":422
 *         v.ibv_wc_read_tm_info(self.cq, &info.info)
 *         return info
 *     def read_completion_wallclock_ns(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_completion_wallclock_ns(self.cq)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_41read_completion_wallclock_ns(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_4CQEX_41read_completion_wallclock_ns = {"read_completion_wallclock_ns", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_41read_completion_wallclock_ns, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_41read_completion_wallclock_ns(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read_completion_wallclock_ns (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("read_completion_wallclock_ns", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("read_completion_wallclock_ns", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_40read_completion_wallclock_ns(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_40read_completion_wallclock_ns(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("read_completion_wallclock_ns", 0);

  /* "pyverbs/cq.pyx":423
 *         return info
 *     def read_completion_wallclock_ns(self):
 *         return v.ibv_wc_read_completion_wallclock_ns(self.cq)             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_long(ibv_wc_read_completion_wallclock_ns(__pyx_v_self->cq)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":422
 *         v.ibv_wc_read_tm_info(self.cq, &info.info)
 *         return info
 *     def read_completion_wallclock_ns(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_completion_wallclock_ns(self.cq)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.read_completion_wallclock_ns", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":425
 *         return v.ibv_wc_read_completion_wallclock_ns(self.cq)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def status(self):
 *         return self.cq.status
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_6status_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_6status_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_6status___get__(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_6status___get__(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":427
 *     @property
 *     def status(self):
 *         return self.cq.status             # <<<<<<<<<<<<<<
 *     @status.setter
 *     def status(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(__pyx_v_self->cq->status); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":425
 *         return v.ibv_wc_read_completion_wallclock_ns(self.cq)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def status(self):
 *         return self.cq.status
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.status.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":428
 *     def status(self):
 *         return self.cq.status
 *     @status.setter             # <<<<<<<<<<<<<<
 *     def status(self, val):
 *         self.cq.status = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_4CQEX_6status_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_4CQEX_6status_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_6status_2__set__(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_4CQEX_6status_2__set__(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  enum ibv_wc_status __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cq.pyx":430
 *     @status.setter
 *     def status(self, val):
 *         self.cq.status = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = ((enum ibv_wc_status)__Pyx_PyLong_As_enum__ibv_wc_status(__pyx_v_val)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 430, __pyx_L1_error)
  __pyx_v_self->cq->status = __pyx_t_1;

  /* "pyverbs/cq.pyx":428
 *     def status(self):
 *         return self.cq.status
 *     @status.setter             # <<<<<<<<<<<<<<
 *     def status(self, val):
 *         self.cq.status = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.CQEX.status.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/cq.pyx":432
 *         self.cq.status = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def wr_id(self):
 *         return self.cq.wr_id
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_5wr_id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_5wr_id_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_5wr_id___get__(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_5wr_id___get__(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":434
 *     @property
 *     def wr_id(self):
 *         return self.cq.wr_id             # <<<<<<<<<<<<<<
 *     @wr_id.setter
 *     def wr_id(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_long(__pyx_v_self->cq->wr_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":432
 *         self.cq.status = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def wr_id(self):
 *         return self.cq.wr_id
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.wr_id.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":435
 *     def wr_id(self):
 *         return self.cq.wr_id
 *     @wr_id.setter             # <<<<<<<<<<<<<<
 *     def wr_id(self, val):
 *         self.cq.wr_id = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_4CQEX_5wr_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_4CQEX_5wr_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_5wr_id_2__set__(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_4CQEX_5wr_id_2__set__(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned long __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cq.pyx":437
 *     @wr_id.setter
 *     def wr_id(self, val):
 *         self.cq.wr_id = val             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_long(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 437, __pyx_L1_error)
  __pyx_v_self->cq->wr_id = __pyx_t_1;

  /* "pyverbs/cq.pyx":435
 *     def wr_id(self):
 *         return self.cq.wr_id
 *     @wr_id.setter             # <<<<<<<<<<<<<<
 *     def wr_id(self, val):
 *         self.cq.wr_id = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.CQEX.wr_id.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/cq.pyx":439
 *         self.cq.wr_id = val
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         print_format = '{:<22}: {:<20}\n'
 *         return 'Extended CQ:\n' +\
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_43__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_43__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_42__str__(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_42__str__(struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self) {
  PyObject *__pyx_v_print_format = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "pyverbs/cq.pyx":440
 * 
 *     def __str__(self):
 *         print_format = '{:<22}: {:<20}\n'             # <<<<<<<<<<<<<<
 *         return 'Extended CQ:\n' +\
 *                print_format.format('Handle', self.cq.handle) +\
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_22_20_2);
  __pyx_v_print_format = __pyx_mstate_global->__pyx_kp_u_22_20_2;

  /* "pyverbs/cq.pyx":441
 *     def __str__(self):
 *         print_format = '{:<22}: {:<20}\n'
 *         return 'Extended CQ:\n' +\             # <<<<<<<<<<<<<<
 *                print_format.format('Handle', self.cq.handle) +\
 *                print_format.format('CQEs', self.cq.cqe)
*/
  __Pyx_XDECREF(__pyx_r);

  /* "pyverbs/cq.pyx":442
 *         print_format = '{:<22}: {:<20}\n'
 *         return 'Extended CQ:\n' +\
 *                print_format.format('Handle', self.cq.handle) +\             # <<<<<<<<<<<<<<
 *                print_format.format('CQEs', self.cq.cqe)
 * 
*/
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->cq->handle); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_n_u_Handle, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":441
 *     def __str__(self):
 *         print_format = '{:<22}: {:<20}\n'
 *         return 'Extended CQ:\n' +\             # <<<<<<<<<<<<<<
 *                print_format.format('Handle', self.cq.handle) +\
 *                print_format.format('CQEs', self.cq.cqe)
*/
  __pyx_t_1 = __Pyx_PyUnicode_Concat(__pyx_mstate_global->__pyx_kp_u_Extended_CQ, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":443
 *         return 'Extended CQ:\n' +\
 *                print_format.format('Handle', self.cq.handle) +\
 *                print_format.format('CQEs', self.cq.cqe)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_v_self->cq->cqe); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_n_u_CQEs, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":442
 *         print_format = '{:<22}: {:<20}\n'
 *         return 'Extended CQ:\n' +\
 *                print_format.format('Handle', self.cq.handle) +\             # <<<<<<<<<<<<<<
 *                print_format.format('CQEs', self.cq.cqe)
 * 
*/
  __pyx_t_2 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":439
 *         self.cq.wr_id = val
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         print_format = '{:<22}: {:<20}\n'
 *         return 'Extended CQ:\n' +\
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyverbs.cq.CQEX.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_print_format);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.cq,self.ibv_cq cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_45__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_4CQEX_45__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_45__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_45__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_44__reduce_cython__(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_44__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.cq,self.ibv_cq cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.cq,self.ibv_cq cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_cq_self_ibv_cq_cannot_be_co, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.cq,self.ibv_cq cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.CQEX.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.cq,self.ibv_cq cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.cq,self.ibv_cq cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_47__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_4CQEX_47__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_47__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_4CQEX_47__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cq.CQEX.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2cq_4CQEX_46__setstate_cython__(((struct __pyx_obj_7pyverbs_2cq_CQEX *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4CQEX_46__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_CQEX *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.cq,self.ibv_cq cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.cq,self.ibv_cq cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_cq_self_ibv_cq_cannot_be_co, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.cq,self.ibv_cq cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.cq,self.ibv_cq cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.CQEX.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":447
 * 
 * cdef class WC(PyverbsObject):
 *     def __init__(self, wr_id=0, status=0, opcode=0, vendor_err=0, byte_len=0,             # <<<<<<<<<<<<<<
 *                  qp_num=0, src_qp=0, imm_data=0, wc_flags=0, pkey_index=0,
 *                  slid=0, sl=0, dlid_path_bits=0):
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_2WC_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7pyverbs_2cq_2WC_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_wr_id = 0;
  PyObject *__pyx_v_status = 0;
  PyObject *__pyx_v_opcode = 0;
  PyObject *__pyx_v_vendor_err = 0;
  PyObject *__pyx_v_byte_len = 0;
  PyObject *__pyx_v_qp_num = 0;
  PyObject *__pyx_v_src_qp = 0;
  PyObject *__pyx_v_imm_data = 0;
  PyObject *__pyx_v_wc_flags = 0;
  PyObject *__pyx_v_pkey_index = 0;
  PyObject *__pyx_v_slid = 0;
  PyObject *__pyx_v_sl = 0;
  PyObject *__pyx_v_dlid_path_bits = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_wr_id,&__pyx_mstate_global->__pyx_n_u_status,&__pyx_mstate_global->__pyx_n_u_opcode,&__pyx_mstate_global->__pyx_n_u_vendor_err,&__pyx_mstate_global->__pyx_n_u_byte_len,&__pyx_mstate_global->__pyx_n_u_qp_num,&__pyx_mstate_global->__pyx_n_u_src_qp,&__pyx_mstate_global->__pyx_n_u_imm_data,&__pyx_mstate_global->__pyx_n_u_wc_flags,&__pyx_mstate_global->__pyx_n_u_pkey_index,&__pyx_mstate_global->__pyx_n_u_slid,&__pyx_mstate_global->__pyx_n_u_sl,&__pyx_mstate_global->__pyx_n_u_dlid_path_bits,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 447, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case 13:
        values[12] = __Pyx_ArgRef_VARARGS(__pyx_args, 12);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[12])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 12:
        values[11] = __Pyx_ArgRef_VARARGS(__pyx_args, 11);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[11])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 11:
        values[10] = __Pyx_ArgRef_VARARGS(__pyx_args, 10);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[10])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 10:
        values[9] = __Pyx_ArgRef_VARARGS(__pyx_args, 9);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[9])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  9:
        values[8] = __Pyx_ArgRef_VARARGS(__pyx_args, 8);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  8:
        values[7] = __Pyx_ArgRef_VARARGS(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_VARARGS(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 447, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[5]) values[5] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[6]) values[6] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[7]) values[7] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[8]) values[8] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[9]) values[9] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[10]) values[10] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[11]) values[11] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[12]) values[12] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    } else {
      switch (__pyx_nargs) {
        case 13:
        values[12] = __Pyx_ArgRef_VARARGS(__pyx_args, 12);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[12])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 12:
        values[11] = __Pyx_ArgRef_VARARGS(__pyx_args, 11);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[11])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 11:
        values[10] = __Pyx_ArgRef_VARARGS(__pyx_args, 10);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[10])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 10:
        values[9] = __Pyx_ArgRef_VARARGS(__pyx_args, 9);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[9])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  9:
        values[8] = __Pyx_ArgRef_VARARGS(__pyx_args, 8);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  8:
        values[7] = __Pyx_ArgRef_VARARGS(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_VARARGS(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 447, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[5]) values[5] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[6]) values[6] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[7]) values[7] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[8]) values[8] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[9]) values[9] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[10]) values[10] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[11]) values[11] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[12]) values[12] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    }
    __pyx_v_wr_id = values[0];
    __pyx_v_status = values[1];
    __pyx_v_opcode = values[2];
    __pyx_v_vendor_err = values[3];
    __pyx_v_byte_len = values[4];
    __pyx_v_qp_num = values[5];
    __pyx_v_src_qp = values[6];
    __pyx_v_imm_data = values[7];
    __pyx_v_wc_flags = values[8];
    __pyx_v_pkey_index = values[9];
    __pyx_v_slid = values[10];
    __pyx_v_sl = values[11];
    __pyx_v_dlid_path_bits = values[12];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 13, __pyx_nargs); __PYX_ERR(0, 447, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cq.WC.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC___init__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self), __pyx_v_wr_id, __pyx_v_status, __pyx_v_opcode, __pyx_v_vendor_err, __pyx_v_byte_len, __pyx_v_qp_num, __pyx_v_src_qp, __pyx_v_imm_data, __pyx_v_wc_flags, __pyx_v_pkey_index, __pyx_v_slid, __pyx_v_sl, __pyx_v_dlid_path_bits);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_2WC___init__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_wr_id, PyObject *__pyx_v_status, PyObject *__pyx_v_opcode, PyObject *__pyx_v_vendor_err, PyObject *__pyx_v_byte_len, PyObject *__pyx_v_qp_num, PyObject *__pyx_v_src_qp, PyObject *__pyx_v_imm_data, PyObject *__pyx_v_wc_flags, PyObject *__pyx_v_pkey_index, PyObject *__pyx_v_slid, PyObject *__pyx_v_sl, PyObject *__pyx_v_dlid_path_bits) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  unsigned long __pyx_t_6;
  enum ibv_wc_status __pyx_t_7;
  enum ibv_wc_opcode __pyx_t_8;
  unsigned int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyverbs/cq.pyx":450
 *                  qp_num=0, src_qp=0, imm_data=0, wc_flags=0, pkey_index=0,
 *                  slid=0, sl=0, dlid_path_bits=0):
 *         super().__init__()             # <<<<<<<<<<<<<<
 *         self.wc.wr_id = wr_id
 *         self.wc.status = status
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_WC), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 450, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 450, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":451
 *                  slid=0, sl=0, dlid_path_bits=0):
 *         super().__init__()
 *         self.wc.wr_id = wr_id             # <<<<<<<<<<<<<<
 *         self.wc.status = status
 *         self.wc.opcode = opcode
*/
  __pyx_t_6 = __Pyx_PyLong_As_unsigned_long(__pyx_v_wr_id); if (unlikely((__pyx_t_6 == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 451, __pyx_L1_error)
  __pyx_v_self->wc.wr_id = __pyx_t_6;

  /* "pyverbs/cq.pyx":452
 *         super().__init__()
 *         self.wc.wr_id = wr_id
 *         self.wc.status = status             # <<<<<<<<<<<<<<
 *         self.wc.opcode = opcode
 *         self.wc.vendor_err = vendor_err
*/
  __pyx_t_7 = ((enum ibv_wc_status)__Pyx_PyLong_As_enum__ibv_wc_status(__pyx_v_status)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 452, __pyx_L1_error)
  __pyx_v_self->wc.status = __pyx_t_7;

  /* "pyverbs/cq.pyx":453
 *         self.wc.wr_id = wr_id
 *         self.wc.status = status
 *         self.wc.opcode = opcode             # <<<<<<<<<<<<<<
 *         self.wc.vendor_err = vendor_err
 *         self.wc.byte_len = byte_len
*/
  __pyx_t_8 = ((enum ibv_wc_opcode)__Pyx_PyLong_As_enum__ibv_wc_opcode(__pyx_v_opcode)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 453, __pyx_L1_error)
  __pyx_v_self->wc.opcode = __pyx_t_8;

  /* "pyverbs/cq.pyx":454
 *         self.wc.status = status
 *         self.wc.opcode = opcode
 *         self.wc.vendor_err = vendor_err             # <<<<<<<<<<<<<<
 *         self.wc.byte_len = byte_len
 *         self.wc.qp_num = qp_num
*/
  __pyx_t_9 = __Pyx_PyLong_As_unsigned_int(__pyx_v_vendor_err); if (unlikely((__pyx_t_9 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 454, __pyx_L1_error)
  __pyx_v_self->wc.vendor_err = __pyx_t_9;

  /* "pyverbs/cq.pyx":455
 *         self.wc.opcode = opcode
 *         self.wc.vendor_err = vendor_err
 *         self.wc.byte_len = byte_len             # <<<<<<<<<<<<<<
 *         self.wc.qp_num = qp_num
 *         self.wc.src_qp = src_qp
*/
  __pyx_t_9 = __Pyx_PyLong_As_unsigned_int(__pyx_v_byte_len); if (unlikely((__pyx_t_9 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 455, __pyx_L1_error)
  __pyx_v_self->wc.byte_len = __pyx_t_9;

  /* "pyverbs/cq.pyx":456
 *         self.wc.vendor_err = vendor_err
 *         self.wc.byte_len = byte_len
 *         self.wc.qp_num = qp_num             # <<<<<<<<<<<<<<
 *         self.wc.src_qp = src_qp
 *         self.wc.wc_flags = wc_flags
*/
  __pyx_t_9 = __Pyx_PyLong_As_unsigned_int(__pyx_v_qp_num); if (unlikely((__pyx_t_9 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 456, __pyx_L1_error)
  __pyx_v_self->wc.qp_num = __pyx_t_9;

  /* "pyverbs/cq.pyx":457
 *         self.wc.byte_len = byte_len
 *         self.wc.qp_num = qp_num
 *         self.wc.src_qp = src_qp             # <<<<<<<<<<<<<<
 *         self.wc.wc_flags = wc_flags
 *         self.wc.pkey_index = pkey_index
*/
  __pyx_t_9 = __Pyx_PyLong_As_unsigned_int(__pyx_v_src_qp); if (unlikely((__pyx_t_9 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 457, __pyx_L1_error)
  __pyx_v_self->wc.src_qp = __pyx_t_9;

  /* "pyverbs/cq.pyx":458
 *         self.wc.qp_num = qp_num
 *         self.wc.src_qp = src_qp
 *         self.wc.wc_flags = wc_flags             # <<<<<<<<<<<<<<
 *         self.wc.pkey_index = pkey_index
 *         self.wc.slid = slid
*/
  __pyx_t_10 = __Pyx_PyLong_As_int(__pyx_v_wc_flags); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 458, __pyx_L1_error)
  __pyx_v_self->wc.wc_flags = __pyx_t_10;

  /* "pyverbs/cq.pyx":459
 *         self.wc.src_qp = src_qp
 *         self.wc.wc_flags = wc_flags
 *         self.wc.pkey_index = pkey_index             # <<<<<<<<<<<<<<
 *         self.wc.slid = slid
 *         self.wc.imm_data = imm_data
*/
  __pyx_t_9 = __Pyx_PyLong_As_unsigned_int(__pyx_v_pkey_index); if (unlikely((__pyx_t_9 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 459, __pyx_L1_error)
  __pyx_v_self->wc.pkey_index = __pyx_t_9;

  /* "pyverbs/cq.pyx":460
 *         self.wc.wc_flags = wc_flags
 *         self.wc.pkey_index = pkey_index
 *         self.wc.slid = slid             # <<<<<<<<<<<<<<
 *         self.wc.imm_data = imm_data
 *         self.wc.sl = sl
*/
  __pyx_t_9 = __Pyx_PyLong_As_unsigned_int(__pyx_v_slid); if (unlikely((__pyx_t_9 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 460, __pyx_L1_error)
  __pyx_v_self->wc.slid = __pyx_t_9;

  /* "pyverbs/cq.pyx":461
 *         self.wc.pkey_index = pkey_index
 *         self.wc.slid = slid
 *         self.wc.imm_data = imm_data             # <<<<<<<<<<<<<<
 *         self.wc.sl = sl
 *         self.wc.dlid_path_bits = dlid_path_bits
*/
  __pyx_t_9 = __Pyx_PyLong_As_unsigned_int(__pyx_v_imm_data); if (unlikely((__pyx_t_9 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 461, __pyx_L1_error)
  __pyx_v_self->wc.imm_data = __pyx_t_9;

  /* "pyverbs/cq.pyx":462
 *         self.wc.slid = slid
 *         self.wc.imm_data = imm_data
 *         self.wc.sl = sl             # <<<<<<<<<<<<<<
 *         self.wc.dlid_path_bits = dlid_path_bits
 * 
*/
  __pyx_t_9 = __Pyx_PyLong_As_unsigned_int(__pyx_v_sl); if (unlikely((__pyx_t_9 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 462, __pyx_L1_error)
  __pyx_v_self->wc.sl = __pyx_t_9;

  /* "pyverbs/cq.pyx":463
 *         self.wc.imm_data = imm_data
 *         self.wc.sl = sl
 *         self.wc.dlid_path_bits = dlid_path_bits             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_9 = __Pyx_PyLong_As_unsigned_int(__pyx_v_dlid_path_bits); if (unlikely((__pyx_t_9 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 463, __pyx_L1_error)
  __pyx_v_self->wc.dlid_path_bits = __pyx_t_9;

  /* "pyverbs/cq.pyx":447
 * 
 * cdef class WC(PyverbsObject):
 *     def __init__(self, wr_id=0, status=0, opcode=0, vendor_err=0, byte_len=0,             # <<<<<<<<<<<<<<
 *                  qp_num=0, src_qp=0, imm_data=0, wc_flags=0, pkey_index=0,
 *                  slid=0, sl=0, dlid_path_bits=0):
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cq.WC.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":465
 *         self.wc.dlid_path_bits = dlid_path_bits
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def wr_id(self):
 *         return self.wc.wr_id
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_5wr_id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_5wr_id_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_5wr_id___get__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2WC_5wr_id___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":467
 *     @property
 *     def wr_id(self):
 *         return self.wc.wr_id             # <<<<<<<<<<<<<<
 *     @wr_id.setter
 *     def wr_id(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_long(__pyx_v_self->wc.wr_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":465
 *         self.wc.dlid_path_bits = dlid_path_bits
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def wr_id(self):
 *         return self.wc.wr_id
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.WC.wr_id.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":468
 *     def wr_id(self):
 *         return self.wc.wr_id
 *     @wr_id.setter             # <<<<<<<<<<<<<<
 *     def wr_id(self, val):
 *         self.wc.wr_id = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_2WC_5wr_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_2WC_5wr_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_5wr_id_2__set__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_2WC_5wr_id_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned long __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cq.pyx":470
 *     @wr_id.setter
 *     def wr_id(self, val):
 *         self.wc.wr_id = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_long(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 470, __pyx_L1_error)
  __pyx_v_self->wc.wr_id = __pyx_t_1;

  /* "pyverbs/cq.pyx":468
 *     def wr_id(self):
 *         return self.wc.wr_id
 *     @wr_id.setter             # <<<<<<<<<<<<<<
 *     def wr_id(self, val):
 *         self.wc.wr_id = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.WC.wr_id.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/cq.pyx":472
 *         self.wc.wr_id = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def status(self):
 *         return self.wc.status
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_6status_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_6status_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_6status___get__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2WC_6status___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":474
 *     @property
 *     def status(self):
 *         return self.wc.status             # <<<<<<<<<<<<<<
 *     @status.setter
 *     def status(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(__pyx_v_self->wc.status); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":472
 *         self.wc.wr_id = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def status(self):
 *         return self.wc.status
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.WC.status.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":475
 *     def status(self):
 *         return self.wc.status
 *     @status.setter             # <<<<<<<<<<<<<<
 *     def status(self, val):
 *         self.wc.status = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_2WC_6status_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_2WC_6status_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_6status_2__set__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_2WC_6status_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  enum ibv_wc_status __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cq.pyx":477
 *     @status.setter
 *     def status(self, val):
 *         self.wc.status = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = ((enum ibv_wc_status)__Pyx_PyLong_As_enum__ibv_wc_status(__pyx_v_val)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 477, __pyx_L1_error)
  __pyx_v_self->wc.status = __pyx_t_1;

  /* "pyverbs/cq.pyx":475
 *     def status(self):
 *         return self.wc.status
 *     @status.setter             # <<<<<<<<<<<<<<
 *     def status(self, val):
 *         self.wc.status = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.WC.status.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/cq.pyx":479
 *         self.wc.status = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def opcode(self):
 *         return self.wc.opcode
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_6opcode_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_6opcode_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_6opcode___get__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2WC_6opcode___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":481
 *     @property
 *     def opcode(self):
 *         return self.wc.opcode             # <<<<<<<<<<<<<<
 *     @opcode.setter
 *     def opcode(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py(__pyx_v_self->wc.opcode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":479
 *         self.wc.status = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def opcode(self):
 *         return self.wc.opcode
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.WC.opcode.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":482
 *     def opcode(self):
 *         return self.wc.opcode
 *     @opcode.setter             # <<<<<<<<<<<<<<
 *     def opcode(self, val):
 *         self.wc.opcode = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_2WC_6opcode_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_2WC_6opcode_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_6opcode_2__set__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_2WC_6opcode_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  enum ibv_wc_opcode __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cq.pyx":484
 *     @opcode.setter
 *     def opcode(self, val):
 *         self.wc.opcode = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = ((enum ibv_wc_opcode)__Pyx_PyLong_As_enum__ibv_wc_opcode(__pyx_v_val)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 484, __pyx_L1_error)
  __pyx_v_self->wc.opcode = __pyx_t_1;

  /* "pyverbs/cq.pyx":482
 *     def opcode(self):
 *         return self.wc.opcode
 *     @opcode.setter             # <<<<<<<<<<<<<<
 *     def opcode(self, val):
 *         self.wc.opcode = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.WC.opcode.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/cq.pyx":486
 *         self.wc.opcode = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def vendor_err(self):
 *         return self.wc.vendor_err
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_10vendor_err_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_10vendor_err_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_10vendor_err___get__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2WC_10vendor_err___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":488
 *     @property
 *     def vendor_err(self):
 *         return self.wc.vendor_err             # <<<<<<<<<<<<<<
 *     @vendor_err.setter
 *     def vendor_err(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->wc.vendor_err); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":486
 *         self.wc.opcode = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def vendor_err(self):
 *         return self.wc.vendor_err
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.WC.vendor_err.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":489
 *     def vendor_err(self):
 *         return self.wc.vendor_err
 *     @vendor_err.setter             # <<<<<<<<<<<<<<
 *     def vendor_err(self, val):
 *         self.wc.vendor_err = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_2WC_10vendor_err_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_2WC_10vendor_err_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_10vendor_err_2__set__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_2WC_10vendor_err_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cq.pyx":491
 *     @vendor_err.setter
 *     def vendor_err(self, val):
 *         self.wc.vendor_err = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 491, __pyx_L1_error)
  __pyx_v_self->wc.vendor_err = __pyx_t_1;

  /* "pyverbs/cq.pyx":489
 *     def vendor_err(self):
 *         return self.wc.vendor_err
 *     @vendor_err.setter             # <<<<<<<<<<<<<<
 *     def vendor_err(self, val):
 *         self.wc.vendor_err = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.WC.vendor_err.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/cq.pyx":493
 *         self.wc.vendor_err = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def byte_len(self):
 *         return self.wc.byte_len
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_8byte_len_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_8byte_len_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_8byte_len___get__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2WC_8byte_len___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":495
 *     @property
 *     def byte_len(self):
 *         return self.wc.byte_len             # <<<<<<<<<<<<<<
 *     @byte_len.setter
 *     def byte_len(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->wc.byte_len); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":493
 *         self.wc.vendor_err = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def byte_len(self):
 *         return self.wc.byte_len
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.WC.byte_len.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":496
 *     def byte_len(self):
 *         return self.wc.byte_len
 *     @byte_len.setter             # <<<<<<<<<<<<<<
 *     def byte_len(self, val):
 *         self.wc.byte_len = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_2WC_8byte_len_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_2WC_8byte_len_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_8byte_len_2__set__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_2WC_8byte_len_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cq.pyx":498
 *     @byte_len.setter
 *     def byte_len(self, val):
 *         self.wc.byte_len = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 498, __pyx_L1_error)
  __pyx_v_self->wc.byte_len = __pyx_t_1;

  /* "pyverbs/cq.pyx":496
 *     def byte_len(self):
 *         return self.wc.byte_len
 *     @byte_len.setter             # <<<<<<<<<<<<<<
 *     def byte_len(self, val):
 *         self.wc.byte_len = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.WC.byte_len.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/cq.pyx":500
 *         self.wc.byte_len = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qp_num(self):
 *         return self.wc.qp_num
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_6qp_num_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_6qp_num_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_6qp_num___get__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2WC_6qp_num___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":502
 *     @property
 *     def qp_num(self):
 *         return self.wc.qp_num             # <<<<<<<<<<<<<<
 *     @qp_num.setter
 *     def qp_num(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->wc.qp_num); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":500
 *         self.wc.byte_len = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qp_num(self):
 *         return self.wc.qp_num
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.WC.qp_num.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":503
 *     def qp_num(self):
 *         return self.wc.qp_num
 *     @qp_num.setter             # <<<<<<<<<<<<<<
 *     def qp_num(self, val):
 *         self.wc.qp_num = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_2WC_6qp_num_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_2WC_6qp_num_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_6qp_num_2__set__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_2WC_6qp_num_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cq.pyx":505
 *     @qp_num.setter
 *     def qp_num(self, val):
 *         self.wc.qp_num = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 505, __pyx_L1_error)
  __pyx_v_self->wc.qp_num = __pyx_t_1;

  /* "pyverbs/cq.pyx":503
 *     def qp_num(self):
 *         return self.wc.qp_num
 *     @qp_num.setter             # <<<<<<<<<<<<<<
 *     def qp_num(self, val):
 *         self.wc.qp_num = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.WC.qp_num.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/cq.pyx":507
 *         self.wc.qp_num = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def src_qp(self):
 *         return self.wc.src_qp
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_6src_qp_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_6src_qp_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_6src_qp___get__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2WC_6src_qp___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":509
 *     @property
 *     def src_qp(self):
 *         return self.wc.src_qp             # <<<<<<<<<<<<<<
 *     @src_qp.setter
 *     def src_qp(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->wc.src_qp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":507
 *         self.wc.qp_num = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def src_qp(self):
 *         return self.wc.src_qp
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.WC.src_qp.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":510
 *     def src_qp(self):
 *         return self.wc.src_qp
 *     @src_qp.setter             # <<<<<<<<<<<<<<
 *     def src_qp(self, val):
 *         self.wc.src_qp = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_2WC_6src_qp_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_2WC_6src_qp_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_6src_qp_2__set__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_2WC_6src_qp_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cq.pyx":512
 *     @src_qp.setter
 *     def src_qp(self, val):
 *         self.wc.src_qp = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 512, __pyx_L1_error)
  __pyx_v_self->wc.src_qp = __pyx_t_1;

  /* "pyverbs/cq.pyx":510
 *     def src_qp(self):
 *         return self.wc.src_qp
 *     @src_qp.setter             # <<<<<<<<<<<<<<
 *     def src_qp(self, val):
 *         self.wc.src_qp = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.WC.src_qp.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/cq.pyx":514
 *         self.wc.src_qp = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def wc_flags(self):
 *         return self.wc.wc_flags
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_8wc_flags_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_8wc_flags_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_8wc_flags___get__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2WC_8wc_flags___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":516
 *     @property
 *     def wc_flags(self):
 *         return self.wc.wc_flags             # <<<<<<<<<<<<<<
 *     @wc_flags.setter
 *     def wc_flags(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_self->wc.wc_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":514
 *         self.wc.src_qp = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def wc_flags(self):
 *         return self.wc.wc_flags
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.WC.wc_flags.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":517
 *     def wc_flags(self):
 *         return self.wc.wc_flags
 *     @wc_flags.setter             # <<<<<<<<<<<<<<
 *     def wc_flags(self, val):
 *         self.wc.wc_flags = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_2WC_8wc_flags_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_2WC_8wc_flags_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_8wc_flags_2__set__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_2WC_8wc_flags_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cq.pyx":519
 *     @wc_flags.setter
 *     def wc_flags(self, val):
 *         self.wc.wc_flags = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 519, __pyx_L1_error)
  __pyx_v_self->wc.wc_flags = __pyx_t_1;

  /* "pyverbs/cq.pyx":517
 *     def wc_flags(self):
 *         return self.wc.wc_flags
 *     @wc_flags.setter             # <<<<<<<<<<<<<<
 *     def wc_flags(self, val):
 *         self.wc.wc_flags = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.WC.wc_flags.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/cq.pyx":521
 *         self.wc.wc_flags = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def pkey_index(self):
 *         return self.wc.pkey_index
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_10pkey_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_10pkey_index_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_10pkey_index___get__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2WC_10pkey_index___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":523
 *     @property
 *     def pkey_index(self):
 *         return self.wc.pkey_index             # <<<<<<<<<<<<<<
 *     @pkey_index.setter
 *     def pkey_index(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->wc.pkey_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":521
 *         self.wc.wc_flags = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def pkey_index(self):
 *         return self.wc.pkey_index
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.WC.pkey_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":524
 *     def pkey_index(self):
 *         return self.wc.pkey_index
 *     @pkey_index.setter             # <<<<<<<<<<<<<<
 *     def pkey_index(self, val):
 *         self.wc.pkey_index = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_2WC_10pkey_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_2WC_10pkey_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_10pkey_index_2__set__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_2WC_10pkey_index_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cq.pyx":526
 *     @pkey_index.setter
 *     def pkey_index(self, val):
 *         self.wc.pkey_index = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 526, __pyx_L1_error)
  __pyx_v_self->wc.pkey_index = __pyx_t_1;

  /* "pyverbs/cq.pyx":524
 *     def pkey_index(self):
 *         return self.wc.pkey_index
 *     @pkey_index.setter             # <<<<<<<<<<<<<<
 *     def pkey_index(self, val):
 *         self.wc.pkey_index = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.WC.pkey_index.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/cq.pyx":528
 *         self.wc.pkey_index = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def slid(self):
 *         return self.wc.slid
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_4slid_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_4slid_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_4slid___get__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2WC_4slid___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":530
 *     @property
 *     def slid(self):
 *         return self.wc.slid             # <<<<<<<<<<<<<<
 *     @slid.setter
 *     def slid(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->wc.slid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":528
 *         self.wc.pkey_index = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def slid(self):
 *         return self.wc.slid
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.WC.slid.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":531
 *     def slid(self):
 *         return self.wc.slid
 *     @slid.setter             # <<<<<<<<<<<<<<
 *     def slid(self, val):
 *         self.wc.slid = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_2WC_4slid_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_2WC_4slid_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_4slid_2__set__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_2WC_4slid_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cq.pyx":533
 *     @slid.setter
 *     def slid(self, val):
 *         self.wc.slid = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 533, __pyx_L1_error)
  __pyx_v_self->wc.slid = __pyx_t_1;

  /* "pyverbs/cq.pyx":531
 *     def slid(self):
 *         return self.wc.slid
 *     @slid.setter             # <<<<<<<<<<<<<<
 *     def slid(self, val):
 *         self.wc.slid = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.WC.slid.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/cq.pyx":535
 *         self.wc.slid = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def sl(self):
 *         return self.wc.sl
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_2sl_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_2sl_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_2sl___get__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2WC_2sl___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":537
 *     @property
 *     def sl(self):
 *         return self.wc.sl             # <<<<<<<<<<<<<<
 *     @sl.setter
 *     def sl(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->wc.sl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":535
 *         self.wc.slid = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def sl(self):
 *         return self.wc.sl
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.WC.sl.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":538
 *     def sl(self):
 *         return self.wc.sl
 *     @sl.setter             # <<<<<<<<<<<<<<
 *     def sl(self, val):
 *         self.wc.sl = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_2WC_2sl_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_2WC_2sl_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_2sl_2__set__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_2WC_2sl_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cq.pyx":540
 *     @sl.setter
 *     def sl(self, val):
 *         self.wc.sl = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 540, __pyx_L1_error)
  __pyx_v_self->wc.sl = __pyx_t_1;

  /* "pyverbs/cq.pyx":538
 *     def sl(self):
 *         return self.wc.sl
 *     @sl.setter             # <<<<<<<<<<<<<<
 *     def sl(self, val):
 *         self.wc.sl = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.WC.sl.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/cq.pyx":542
 *         self.wc.sl = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def imm_data(self):
 *         return self.wc.imm_data
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_8imm_data_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_8imm_data_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_8imm_data___get__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2WC_8imm_data___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":544
 *     @property
 *     def imm_data(self):
 *         return self.wc.imm_data             # <<<<<<<<<<<<<<
 *     @imm_data.setter
 *     def imm_data(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->wc.imm_data); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":542
 *         self.wc.sl = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def imm_data(self):
 *         return self.wc.imm_data
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.WC.imm_data.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":545
 *     def imm_data(self):
 *         return self.wc.imm_data
 *     @imm_data.setter             # <<<<<<<<<<<<<<
 *     def imm_data(self, val):
 *         self.wc.imm_data = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_2WC_8imm_data_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_2WC_8imm_data_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_8imm_data_2__set__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_2WC_8imm_data_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cq.pyx":547
 *     @imm_data.setter
 *     def imm_data(self, val):
 *         self.wc.imm_data = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 547, __pyx_L1_error)
  __pyx_v_self->wc.imm_data = __pyx_t_1;

  /* "pyverbs/cq.pyx":545
 *     def imm_data(self):
 *         return self.wc.imm_data
 *     @imm_data.setter             # <<<<<<<<<<<<<<
 *     def imm_data(self, val):
 *         self.wc.imm_data = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.WC.imm_data.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/cq.pyx":549
 *         self.wc.imm_data = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dlid_path_bits(self):
 *         return self.wc.dlid_path_bits
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_14dlid_path_bits_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_14dlid_path_bits_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_14dlid_path_bits___get__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2WC_14dlid_path_bits___get__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":551
 *     @property
 *     def dlid_path_bits(self):
 *         return self.wc.dlid_path_bits             # <<<<<<<<<<<<<<
 *     @dlid_path_bits.setter
 *     def dlid_path_bits(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->wc.dlid_path_bits); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":549
 *         self.wc.imm_data = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dlid_path_bits(self):
 *         return self.wc.dlid_path_bits
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.WC.dlid_path_bits.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":552
 *     def dlid_path_bits(self):
 *         return self.wc.dlid_path_bits
 *     @dlid_path_bits.setter             # <<<<<<<<<<<<<<
 *     def dlid_path_bits(self, val):
 *         self.wc.dlid_path_bits = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_2WC_14dlid_path_bits_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_2WC_14dlid_path_bits_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_14dlid_path_bits_2__set__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_2WC_14dlid_path_bits_2__set__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cq.pyx":554
 *     @dlid_path_bits.setter
 *     def dlid_path_bits(self, val):
 *         self.wc.dlid_path_bits = val             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 554, __pyx_L1_error)
  __pyx_v_self->wc.dlid_path_bits = __pyx_t_1;

  /* "pyverbs/cq.pyx":552
 *     def dlid_path_bits(self):
 *         return self.wc.dlid_path_bits
 *     @dlid_path_bits.setter             # <<<<<<<<<<<<<<
 *     def dlid_path_bits(self, val):
 *         self.wc.dlid_path_bits = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.WC.dlid_path_bits.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/cq.pyx":556
 *         self.wc.dlid_path_bits = val
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         print_format = '{:22}: {:<20}\n'
 *         return print_format.format('WR ID', self.wr_id) +\
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_3__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_3__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_2__str__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2WC_2__str__(struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self) {
  PyObject *__pyx_v_print_format = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "pyverbs/cq.pyx":557
 * 
 *     def __str__(self):
 *         print_format = '{:22}: {:<20}\n'             # <<<<<<<<<<<<<<
 *         return print_format.format('WR ID', self.wr_id) +\
 *             print_format.format('status', cqe_status_to_str(self.status)) +\
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_22_20);
  __pyx_v_print_format = __pyx_mstate_global->__pyx_kp_u_22_20;

  /* "pyverbs/cq.pyx":558
 *     def __str__(self):
 *         print_format = '{:22}: {:<20}\n'
 *         return print_format.format('WR ID', self.wr_id) +\             # <<<<<<<<<<<<<<
 *             print_format.format('status', cqe_status_to_str(self.status)) +\
 *             print_format.format('opcode', cqe_opcode_to_str(self.opcode)) +\
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_wr_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_WR_ID, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":559
 *         print_format = '{:22}: {:<20}\n'
 *         return print_format.format('WR ID', self.wr_id) +\
 *             print_format.format('status', cqe_status_to_str(self.status)) +\             # <<<<<<<<<<<<<<
 *             print_format.format('opcode', cqe_opcode_to_str(self.opcode)) +\
 *             print_format.format('vendor error', self.vendor_err) +\
*/
  __pyx_t_3 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_cqe_status_to_str); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_status); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_3);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_6 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_n_u_status, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":558
 *     def __str__(self):
 *         print_format = '{:22}: {:<20}\n'
 *         return print_format.format('WR ID', self.wr_id) +\             # <<<<<<<<<<<<<<
 *             print_format.format('status', cqe_status_to_str(self.status)) +\
 *             print_format.format('opcode', cqe_opcode_to_str(self.opcode)) +\
*/
  __pyx_t_1 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/cq.pyx":560
 *         return print_format.format('WR ID', self.wr_id) +\
 *             print_format.format('status', cqe_status_to_str(self.status)) +\
 *             print_format.format('opcode', cqe_opcode_to_str(self.opcode)) +\             # <<<<<<<<<<<<<<
 *             print_format.format('vendor error', self.vendor_err) +\
 *             print_format.format('byte length', self.byte_len) +\
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_cqe_opcode_to_str); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_opcode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
    __pyx_t_6 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_3};
    __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __pyx_t_5 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_n_u_opcode, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/cq.pyx":559
 *         print_format = '{:22}: {:<20}\n'
 *         return print_format.format('WR ID', self.wr_id) +\
 *             print_format.format('status', cqe_status_to_str(self.status)) +\             # <<<<<<<<<<<<<<
 *             print_format.format('opcode', cqe_opcode_to_str(self.opcode)) +\
 *             print_format.format('vendor error', self.vendor_err) +\
*/
  __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/cq.pyx":561
 *             print_format.format('status', cqe_status_to_str(self.status)) +\
 *             print_format.format('opcode', cqe_opcode_to_str(self.opcode)) +\
 *             print_format.format('vendor error', self.vendor_err) +\             # <<<<<<<<<<<<<<
 *             print_format.format('byte length', self.byte_len) +\
 *             print_format.format('QP num', self.qp_num) +\
*/
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_vendor_err); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_vendor_error, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/cq.pyx":560
 *         return print_format.format('WR ID', self.wr_id) +\
 *             print_format.format('status', cqe_status_to_str(self.status)) +\
 *             print_format.format('opcode', cqe_opcode_to_str(self.opcode)) +\             # <<<<<<<<<<<<<<
 *             print_format.format('vendor error', self.vendor_err) +\
 *             print_format.format('byte length', self.byte_len) +\
*/
  __pyx_t_5 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":562
 *             print_format.format('opcode', cqe_opcode_to_str(self.opcode)) +\
 *             print_format.format('vendor error', self.vendor_err) +\
 *             print_format.format('byte length', self.byte_len) +\             # <<<<<<<<<<<<<<
 *             print_format.format('QP num', self.qp_num) +\
 *             print_format.format('source QP', self.src_qp) +\
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_byte_len); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 562, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_byte_length, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 562, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":561
 *             print_format.format('status', cqe_status_to_str(self.status)) +\
 *             print_format.format('opcode', cqe_opcode_to_str(self.opcode)) +\
 *             print_format.format('vendor error', self.vendor_err) +\             # <<<<<<<<<<<<<<
 *             print_format.format('byte length', self.byte_len) +\
 *             print_format.format('QP num', self.qp_num) +\
*/
  __pyx_t_1 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/cq.pyx":563
 *             print_format.format('vendor error', self.vendor_err) +\
 *             print_format.format('byte length', self.byte_len) +\
 *             print_format.format('QP num', self.qp_num) +\             # <<<<<<<<<<<<<<
 *             print_format.format('source QP', self.src_qp) +\
 *             print_format.format('WC flags', cqe_flags_to_str(self.wc_flags)) +\
*/
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_qp_num); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_QP_num, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyverbs/cq.pyx":562
 *             print_format.format('opcode', cqe_opcode_to_str(self.opcode)) +\
 *             print_format.format('vendor error', self.vendor_err) +\
 *             print_format.format('byte length', self.byte_len) +\             # <<<<<<<<<<<<<<
 *             print_format.format('QP num', self.qp_num) +\
 *             print_format.format('source QP', self.src_qp) +\
*/
  __pyx_t_4 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 562, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/cq.pyx":564
 *             print_format.format('byte length', self.byte_len) +\
 *             print_format.format('QP num', self.qp_num) +\
 *             print_format.format('source QP', self.src_qp) +\             # <<<<<<<<<<<<<<
 *             print_format.format('WC flags', cqe_flags_to_str(self.wc_flags)) +\
 *             print_format.format('pkey index', self.pkey_index) +\
*/
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_src_qp); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_source_QP, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/cq.pyx":563
 *             print_format.format('vendor error', self.vendor_err) +\
 *             print_format.format('byte length', self.byte_len) +\
 *             print_format.format('QP num', self.qp_num) +\             # <<<<<<<<<<<<<<
 *             print_format.format('source QP', self.src_qp) +\
 *             print_format.format('WC flags', cqe_flags_to_str(self.wc_flags)) +\
*/
  __pyx_t_5 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":565
 *             print_format.format('QP num', self.qp_num) +\
 *             print_format.format('source QP', self.src_qp) +\
 *             print_format.format('WC flags', cqe_flags_to_str(self.wc_flags)) +\             # <<<<<<<<<<<<<<
 *             print_format.format('pkey index', self.pkey_index) +\
 *             print_format.format('slid', self.slid) +\
*/
  __pyx_t_4 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_cqe_flags_to_str); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_wc_flags); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    assert(__pyx_t_4);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
    __pyx_t_6 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_WC_flags, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":564
 *             print_format.format('byte length', self.byte_len) +\
 *             print_format.format('QP num', self.qp_num) +\
 *             print_format.format('source QP', self.src_qp) +\             # <<<<<<<<<<<<<<
 *             print_format.format('WC flags', cqe_flags_to_str(self.wc_flags)) +\
 *             print_format.format('pkey index', self.pkey_index) +\
*/
  __pyx_t_1 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/cq.pyx":566
 *             print_format.format('source QP', self.src_qp) +\
 *             print_format.format('WC flags', cqe_flags_to_str(self.wc_flags)) +\
 *             print_format.format('pkey index', self.pkey_index) +\             # <<<<<<<<<<<<<<
 *             print_format.format('slid', self.slid) +\
 *             print_format.format('sl', self.sl) +\
*/
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_pkey_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_pkey_index_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/cq.pyx":565
 *             print_format.format('QP num', self.qp_num) +\
 *             print_format.format('source QP', self.src_qp) +\
 *             print_format.format('WC flags', cqe_flags_to_str(self.wc_flags)) +\             # <<<<<<<<<<<<<<
 *             print_format.format('pkey index', self.pkey_index) +\
 *             print_format.format('slid', self.slid) +\
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/cq.pyx":567
 *             print_format.format('WC flags', cqe_flags_to_str(self.wc_flags)) +\
 *             print_format.format('pkey index', self.pkey_index) +\
 *             print_format.format('slid', self.slid) +\             # <<<<<<<<<<<<<<
 *             print_format.format('sl', self.sl) +\
 *             print_format.format('imm_data', self.imm_data) +\
*/
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_slid); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_n_u_slid, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/cq.pyx":566
 *             print_format.format('source QP', self.src_qp) +\
 *             print_format.format('WC flags', cqe_flags_to_str(self.wc_flags)) +\
 *             print_format.format('pkey index', self.pkey_index) +\             # <<<<<<<<<<<<<<
 *             print_format.format('slid', self.slid) +\
 *             print_format.format('sl', self.sl) +\
*/
  __pyx_t_5 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":568
 *             print_format.format('pkey index', self.pkey_index) +\
 *             print_format.format('slid', self.slid) +\
 *             print_format.format('sl', self.sl) +\             # <<<<<<<<<<<<<<
 *             print_format.format('imm_data', self.imm_data) +\
 *             print_format.format('dlid path bits', self.dlid_path_bits)
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_sl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_n_u_sl, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":567
 *             print_format.format('WC flags', cqe_flags_to_str(self.wc_flags)) +\
 *             print_format.format('pkey index', self.pkey_index) +\
 *             print_format.format('slid', self.slid) +\             # <<<<<<<<<<<<<<
 *             print_format.format('sl', self.sl) +\
 *             print_format.format('imm_data', self.imm_data) +\
*/
  __pyx_t_1 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/cq.pyx":569
 *             print_format.format('slid', self.slid) +\
 *             print_format.format('sl', self.sl) +\
 *             print_format.format('imm_data', self.imm_data) +\             # <<<<<<<<<<<<<<
 *             print_format.format('dlid path bits', self.dlid_path_bits)
 * 
*/
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_imm_data); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_n_u_imm_data, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/cq.pyx":568
 *             print_format.format('pkey index', self.pkey_index) +\
 *             print_format.format('slid', self.slid) +\
 *             print_format.format('sl', self.sl) +\             # <<<<<<<<<<<<<<
 *             print_format.format('imm_data', self.imm_data) +\
 *             print_format.format('dlid path bits', self.dlid_path_bits)
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_1, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/cq.pyx":570
 *             print_format.format('sl', self.sl) +\
 *             print_format.format('imm_data', self.imm_data) +\
 *             print_format.format('dlid path bits', self.dlid_path_bits)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dlid_path_bits); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_dlid_path_bits_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyverbs/cq.pyx":569
 *             print_format.format('slid', self.slid) +\
 *             print_format.format('sl', self.sl) +\
 *             print_format.format('imm_data', self.imm_data) +\             # <<<<<<<<<<<<<<
 *             print_format.format('dlid path bits', self.dlid_path_bits)
 * 
*/
  __pyx_t_5 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":556
 *         self.wc.dlid_path_bits = val
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         print_format = '{:22}: {:<20}\n'
 *         return print_format.format('WR ID', self.wr_id) +\
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyverbs.cq.WC.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_print_format);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.wc must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_2WC_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_2WC_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_4__reduce_cython__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2WC_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.wc must be explicitly requested with @auto_pickle(True)"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "Pickling of struct members such as self.wc must be explicitly requested with @auto_pickle(True)"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_Pickling_of_struct_members_such, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.wc must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.WC.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.wc must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.wc must be explicitly requested with @auto_pickle(True)"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_2WC_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_2WC_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_2WC_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cq.WC.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2cq_2WC_6__setstate_cython__(((struct __pyx_obj_7pyverbs_2cq_WC *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2WC_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_WC *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "Pickling of struct members such as self.wc must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "Pickling of struct members such as self.wc must be explicitly requested with @auto_pickle(True)"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_Pickling_of_struct_members_such, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.wc must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.wc must be explicitly requested with @auto_pickle(True)"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.WC.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":574
 * 
 * cdef class PollCqAttr(PyverbsObject):
 *     @property             # <<<<<<<<<<<<<<
 *     def comp_mask(self):
 *         return self.attr.comp_mask
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_10PollCqAttr_9comp_mask_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_10PollCqAttr_9comp_mask_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_10PollCqAttr_9comp_mask___get__(((struct __pyx_obj_7pyverbs_2cq_PollCqAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_10PollCqAttr_9comp_mask___get__(struct __pyx_obj_7pyverbs_2cq_PollCqAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":576
 *     @property
 *     def comp_mask(self):
 *         return self.attr.comp_mask             # <<<<<<<<<<<<<<
 *     @comp_mask.setter
 *     def comp_mask(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->attr.comp_mask); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":574
 * 
 * cdef class PollCqAttr(PyverbsObject):
 *     @property             # <<<<<<<<<<<<<<
 *     def comp_mask(self):
 *         return self.attr.comp_mask
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.PollCqAttr.comp_mask.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":577
 *     def comp_mask(self):
 *         return self.attr.comp_mask
 *     @comp_mask.setter             # <<<<<<<<<<<<<<
 *     def comp_mask(self, val):
 *         self.attr.comp_mask = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_10PollCqAttr_9comp_mask_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_10PollCqAttr_9comp_mask_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_10PollCqAttr_9comp_mask_2__set__(((struct __pyx_obj_7pyverbs_2cq_PollCqAttr *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_10PollCqAttr_9comp_mask_2__set__(struct __pyx_obj_7pyverbs_2cq_PollCqAttr *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cq.pyx":579
 *     @comp_mask.setter
 *     def comp_mask(self, val):
 *         self.attr.comp_mask = val             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 579, __pyx_L1_error)
  __pyx_v_self->attr.comp_mask = __pyx_t_1;

  /* "pyverbs/cq.pyx":577
 *     def comp_mask(self):
 *         return self.attr.comp_mask
 *     @comp_mask.setter             # <<<<<<<<<<<<<<
 *     def comp_mask(self, val):
 *         self.attr.comp_mask = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.PollCqAttr.comp_mask.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_10PollCqAttr_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_10PollCqAttr_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_10PollCqAttr_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_10PollCqAttr_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_10PollCqAttr___reduce_cython__(((struct __pyx_obj_7pyverbs_2cq_PollCqAttr *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_10PollCqAttr___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_PollCqAttr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_Pickling_of_struct_members_such_2, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.PollCqAttr.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_10PollCqAttr_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_10PollCqAttr_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_10PollCqAttr_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_10PollCqAttr_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cq.PollCqAttr.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2cq_10PollCqAttr_2__setstate_cython__(((struct __pyx_obj_7pyverbs_2cq_PollCqAttr *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_10PollCqAttr_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_PollCqAttr *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_Pickling_of_struct_members_such_2, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.PollCqAttr.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":583
 * 
 * cdef class WcTmInfo(PyverbsObject):
 *     @property             # <<<<<<<<<<<<<<
 *     def tag(self):
 *         return self.info.tag
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_8WcTmInfo_3tag_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_8WcTmInfo_3tag_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_8WcTmInfo_3tag___get__(((struct __pyx_obj_7pyverbs_2cq_WcTmInfo *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_8WcTmInfo_3tag___get__(struct __pyx_obj_7pyverbs_2cq_WcTmInfo *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":585
 *     @property
 *     def tag(self):
 *         return self.info.tag             # <<<<<<<<<<<<<<
 *     @tag.setter
 *     def tag(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_long(__pyx_v_self->info.tag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":583
 * 
 * cdef class WcTmInfo(PyverbsObject):
 *     @property             # <<<<<<<<<<<<<<
 *     def tag(self):
 *         return self.info.tag
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.WcTmInfo.tag.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":586
 *     def tag(self):
 *         return self.info.tag
 *     @tag.setter             # <<<<<<<<<<<<<<
 *     def tag(self, val):
 *         self.info.tag = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_8WcTmInfo_3tag_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_8WcTmInfo_3tag_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_8WcTmInfo_3tag_2__set__(((struct __pyx_obj_7pyverbs_2cq_WcTmInfo *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_8WcTmInfo_3tag_2__set__(struct __pyx_obj_7pyverbs_2cq_WcTmInfo *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned long __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cq.pyx":588
 *     @tag.setter
 *     def tag(self, val):
 *         self.info.tag = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_long(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 588, __pyx_L1_error)
  __pyx_v_self->info.tag = __pyx_t_1;

  /* "pyverbs/cq.pyx":586
 *     def tag(self):
 *         return self.info.tag
 *     @tag.setter             # <<<<<<<<<<<<<<
 *     def tag(self, val):
 *         self.info.tag = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.WcTmInfo.tag.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/cq.pyx":590
 *         self.info.tag = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def priv(self):
 *         return self.info.priv
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_8WcTmInfo_4priv_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_2cq_8WcTmInfo_4priv_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_8WcTmInfo_4priv___get__(((struct __pyx_obj_7pyverbs_2cq_WcTmInfo *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_8WcTmInfo_4priv___get__(struct __pyx_obj_7pyverbs_2cq_WcTmInfo *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cq.pyx":592
 *     @property
 *     def priv(self):
 *         return self.info.priv             # <<<<<<<<<<<<<<
 *     @priv.setter
 *     def priv(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->info.priv); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":590
 *         self.info.tag = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def priv(self):
 *         return self.info.priv
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cq.WcTmInfo.priv.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":593
 *     def priv(self):
 *         return self.info.priv
 *     @priv.setter             # <<<<<<<<<<<<<<
 *     def priv(self, val):
 *         self.info.priv = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_2cq_8WcTmInfo_4priv_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_2cq_8WcTmInfo_4priv_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_2cq_8WcTmInfo_4priv_2__set__(((struct __pyx_obj_7pyverbs_2cq_WcTmInfo *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_2cq_8WcTmInfo_4priv_2__set__(struct __pyx_obj_7pyverbs_2cq_WcTmInfo *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cq.pyx":595
 *     @priv.setter
 *     def priv(self, val):
 *         self.info.priv = val             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __Pyx_PyLong_As_unsigned_int(__pyx_v_val); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 595, __pyx_L1_error)
  __pyx_v_self->info.priv = __pyx_t_1;

  /* "pyverbs/cq.pyx":593
 *     def priv(self):
 *         return self.info.priv
 *     @priv.setter             # <<<<<<<<<<<<<<
 *     def priv(self, val):
 *         self.info.priv = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.WcTmInfo.priv.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.info must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_8WcTmInfo_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_8WcTmInfo_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_8WcTmInfo_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_8WcTmInfo_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_2cq_8WcTmInfo___reduce_cython__(((struct __pyx_obj_7pyverbs_2cq_WcTmInfo *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_8WcTmInfo___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_WcTmInfo *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.info must be explicitly requested with @auto_pickle(True)"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "Pickling of struct members such as self.info must be explicitly requested with @auto_pickle(True)"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_Pickling_of_struct_members_such_3, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.info must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.WcTmInfo.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.info must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.info must be explicitly requested with @auto_pickle(True)"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_8WcTmInfo_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_8WcTmInfo_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_8WcTmInfo_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_8WcTmInfo_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cq.WcTmInfo.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2cq_8WcTmInfo_2__setstate_cython__(((struct __pyx_obj_7pyverbs_2cq_WcTmInfo *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_8WcTmInfo_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_2cq_WcTmInfo *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "Pickling of struct members such as self.info must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "Pickling of struct members such as self.info must be explicitly requested with @auto_pickle(True)"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_Pickling_of_struct_members_such_3, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.info must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.info must be explicitly requested with @auto_pickle(True)"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cq.WcTmInfo.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":598
 * 
 * 
 * def cqe_status_to_str(status):             # <<<<<<<<<<<<<<
 *     try:
 *         return {e.IBV_WC_SUCCESS: "success",
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_1cqe_status_to_str(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_1cqe_status_to_str = {"cqe_status_to_str", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_1cqe_status_to_str, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_1cqe_status_to_str(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_status = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cqe_status_to_str (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_status,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 598, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 598, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "cqe_status_to_str", 0) < (0)) __PYX_ERR(0, 598, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("cqe_status_to_str", 1, 1, 1, i); __PYX_ERR(0, 598, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 598, __pyx_L3_error)
    }
    __pyx_v_status = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cqe_status_to_str", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 598, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cq.cqe_status_to_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2cq_cqe_status_to_str(__pyx_self, __pyx_v_status);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_cqe_status_to_str(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_status) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cqe_status_to_str", 0);

  /* "pyverbs/cq.pyx":599
 * 
 * def cqe_status_to_str(status):
 *     try:             # <<<<<<<<<<<<<<
 *         return {e.IBV_WC_SUCCESS: "success",
 *                 e.IBV_WC_LOC_LEN_ERR: "local length error",
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "pyverbs/cq.pyx":600
 * def cqe_status_to_str(status):
 *     try:
 *         return {e.IBV_WC_SUCCESS: "success",             # <<<<<<<<<<<<<<
 *                 e.IBV_WC_LOC_LEN_ERR: "local length error",
 *                 e.IBV_WC_LOC_QP_OP_ERR: "local QP op error",
*/
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_PyDict_NewPresized(22); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(IBV_WC_SUCCESS); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_mstate_global->__pyx_n_u_success) < (0)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyverbs/cq.pyx":601
 *     try:
 *         return {e.IBV_WC_SUCCESS: "success",
 *                 e.IBV_WC_LOC_LEN_ERR: "local length error",             # <<<<<<<<<<<<<<
 *                 e.IBV_WC_LOC_QP_OP_ERR: "local QP op error",
 *                 e.IBV_WC_LOC_EEC_OP_ERR: "local EEC op error",
*/
      __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(IBV_WC_LOC_LEN_ERR); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 601, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_mstate_global->__pyx_kp_u_local_length_error) < (0)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyverbs/cq.pyx":602
 *         return {e.IBV_WC_SUCCESS: "success",
 *                 e.IBV_WC_LOC_LEN_ERR: "local length error",
 *                 e.IBV_WC_LOC_QP_OP_ERR: "local QP op error",             # <<<<<<<<<<<<<<
 *                 e.IBV_WC_LOC_EEC_OP_ERR: "local EEC op error",
 *                 e.IBV_WC_LOC_PROT_ERR: "local protection error",
*/
      __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(IBV_WC_LOC_QP_OP_ERR); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 602, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_mstate_global->__pyx_kp_u_local_QP_op_error) < (0)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyverbs/cq.pyx":603
 *                 e.IBV_WC_LOC_LEN_ERR: "local length error",
 *                 e.IBV_WC_LOC_QP_OP_ERR: "local QP op error",
 *                 e.IBV_WC_LOC_EEC_OP_ERR: "local EEC op error",             # <<<<<<<<<<<<<<
 *                 e.IBV_WC_LOC_PROT_ERR: "local protection error",
 *                 e.IBV_WC_WR_FLUSH_ERR: "WR flush error",
*/
      __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(IBV_WC_LOC_EEC_OP_ERR); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 603, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_mstate_global->__pyx_kp_u_local_EEC_op_error) < (0)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyverbs/cq.pyx":604
 *                 e.IBV_WC_LOC_QP_OP_ERR: "local QP op error",
 *                 e.IBV_WC_LOC_EEC_OP_ERR: "local EEC op error",
 *                 e.IBV_WC_LOC_PROT_ERR: "local protection error",             # <<<<<<<<<<<<<<
 *                 e.IBV_WC_WR_FLUSH_ERR: "WR flush error",
 *                 e.IBV_WC_MW_BIND_ERR: "memory window bind error",
*/
      __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(IBV_WC_LOC_PROT_ERR); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 604, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_mstate_global->__pyx_kp_u_local_protection_error) < (0)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyverbs/cq.pyx":605
 *                 e.IBV_WC_LOC_EEC_OP_ERR: "local EEC op error",
 *                 e.IBV_WC_LOC_PROT_ERR: "local protection error",
 *                 e.IBV_WC_WR_FLUSH_ERR: "WR flush error",             # <<<<<<<<<<<<<<
 *                 e.IBV_WC_MW_BIND_ERR: "memory window bind error",
 *                 e.IBV_WC_BAD_RESP_ERR: "bad response error",
*/
      __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(IBV_WC_WR_FLUSH_ERR); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 605, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_mstate_global->__pyx_kp_u_WR_flush_error) < (0)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyverbs/cq.pyx":606
 *                 e.IBV_WC_LOC_PROT_ERR: "local protection error",
 *                 e.IBV_WC_WR_FLUSH_ERR: "WR flush error",
 *                 e.IBV_WC_MW_BIND_ERR: "memory window bind error",             # <<<<<<<<<<<<<<
 *                 e.IBV_WC_BAD_RESP_ERR: "bad response error",
 *                 e.IBV_WC_LOC_ACCESS_ERR: "local access error",
*/
      __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(IBV_WC_MW_BIND_ERR); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 606, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_mstate_global->__pyx_kp_u_memory_window_bind_error) < (0)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyverbs/cq.pyx":607
 *                 e.IBV_WC_WR_FLUSH_ERR: "WR flush error",
 *                 e.IBV_WC_MW_BIND_ERR: "memory window bind error",
 *                 e.IBV_WC_BAD_RESP_ERR: "bad response error",             # <<<<<<<<<<<<<<
 *                 e.IBV_WC_LOC_ACCESS_ERR: "local access error",
 *                 e.IBV_WC_REM_INV_REQ_ERR: "remote invalidate request error",
*/
      __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(IBV_WC_BAD_RESP_ERR); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 607, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_mstate_global->__pyx_kp_u_bad_response_error) < (0)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyverbs/cq.pyx":608
 *                 e.IBV_WC_MW_BIND_ERR: "memory window bind error",
 *                 e.IBV_WC_BAD_RESP_ERR: "bad response error",
 *                 e.IBV_WC_LOC_ACCESS_ERR: "local access error",             # <<<<<<<<<<<<<<
 *                 e.IBV_WC_REM_INV_REQ_ERR: "remote invalidate request error",
 *                 e.IBV_WC_REM_ACCESS_ERR: "remote access error",
*/
      __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(IBV_WC_LOC_ACCESS_ERR); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 608, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_mstate_global->__pyx_kp_u_local_access_error) < (0)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyverbs/cq.pyx":609
 *                 e.IBV_WC_BAD_RESP_ERR: "bad response error",
 *                 e.IBV_WC_LOC_ACCESS_ERR: "local access error",
 *                 e.IBV_WC_REM_INV_REQ_ERR: "remote invalidate request error",             # <<<<<<<<<<<<<<
 *                 e.IBV_WC_REM_ACCESS_ERR: "remote access error",
 *                 e.IBV_WC_REM_OP_ERR: "remote op error",
*/
      __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(IBV_WC_REM_INV_REQ_ERR); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 609, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_mstate_global->__pyx_kp_u_remote_invalidate_request_error) < (0)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyverbs/cq.pyx":610
 *                 e.IBV_WC_LOC_ACCESS_ERR: "local access error",
 *                 e.IBV_WC_REM_INV_REQ_ERR: "remote invalidate request error",
 *                 e.IBV_WC_REM_ACCESS_ERR: "remote access error",             # <<<<<<<<<<<<<<
 *                 e.IBV_WC_REM_OP_ERR: "remote op error",
 *                 e.IBV_WC_RETRY_EXC_ERR: "retry exceeded error",
*/
      __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(IBV_WC_REM_ACCESS_ERR); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 610, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_mstate_global->__pyx_kp_u_remote_access_error) < (0)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyverbs/cq.pyx":611
 *                 e.IBV_WC_REM_INV_REQ_ERR: "remote invalidate request error",
 *                 e.IBV_WC_REM_ACCESS_ERR: "remote access error",
 *                 e.IBV_WC_REM_OP_ERR: "remote op error",             # <<<<<<<<<<<<<<
 *                 e.IBV_WC_RETRY_EXC_ERR: "retry exceeded error",
 *                 e.IBV_WC_RNR_RETRY_EXC_ERR: "RNR retry exceeded",
*/
      __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(IBV_WC_REM_OP_ERR); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 611, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_mstate_global->__pyx_kp_u_remote_op_error) < (0)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyverbs/cq.pyx":612
 *                 e.IBV_WC_REM_ACCESS_ERR: "remote access error",
 *                 e.IBV_WC_REM_OP_ERR: "remote op error",
 *                 e.IBV_WC_RETRY_EXC_ERR: "retry exceeded error",             # <<<<<<<<<<<<<<
 *                 e.IBV_WC_RNR_RETRY_EXC_ERR: "RNR retry exceeded",
 *                 e.IBV_WC_LOC_RDD_VIOL_ERR: "local RDD violation error",
*/
      __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(IBV_WC_RETRY_EXC_ERR); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 612, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_mstate_global->__pyx_kp_u_retry_exceeded_error) < (0)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyverbs/cq.pyx":613
 *                 e.IBV_WC_REM_OP_ERR: "remote op error",
 *                 e.IBV_WC_RETRY_EXC_ERR: "retry exceeded error",
 *                 e.IBV_WC_RNR_RETRY_EXC_ERR: "RNR retry exceeded",             # <<<<<<<<<<<<<<
 *                 e.IBV_WC_LOC_RDD_VIOL_ERR: "local RDD violation error",
 *                 e.IBV_WC_REM_INV_RD_REQ_ERR: "remote invalidate RD request error",
*/
      __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(IBV_WC_RNR_RETRY_EXC_ERR); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 613, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_mstate_global->__pyx_kp_u_RNR_retry_exceeded) < (0)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyverbs/cq.pyx":614
 *                 e.IBV_WC_RETRY_EXC_ERR: "retry exceeded error",
 *                 e.IBV_WC_RNR_RETRY_EXC_ERR: "RNR retry exceeded",
 *                 e.IBV_WC_LOC_RDD_VIOL_ERR: "local RDD violation error",             # <<<<<<<<<<<<<<
 *                 e.IBV_WC_REM_INV_RD_REQ_ERR: "remote invalidate RD request error",
 *                 e.IBV_WC_REM_ABORT_ERR: "remote abort error",
*/
      __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(IBV_WC_LOC_RDD_VIOL_ERR); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 614, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_mstate_global->__pyx_kp_u_local_RDD_violation_error) < (0)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyverbs/cq.pyx":615
 *                 e.IBV_WC_RNR_RETRY_EXC_ERR: "RNR retry exceeded",
 *                 e.IBV_WC_LOC_RDD_VIOL_ERR: "local RDD violation error",
 *                 e.IBV_WC_REM_INV_RD_REQ_ERR: "remote invalidate RD request error",             # <<<<<<<<<<<<<<
 *                 e.IBV_WC_REM_ABORT_ERR: "remote abort error",
 *                 e.IBV_WC_INV_EECN_ERR: "invalidate EECN error",
*/
      __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(IBV_WC_REM_INV_RD_REQ_ERR); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 615, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_mstate_global->__pyx_kp_u_remote_invalidate_RD_request_err) < (0)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyverbs/cq.pyx":616
 *                 e.IBV_WC_LOC_RDD_VIOL_ERR: "local RDD violation error",
 *                 e.IBV_WC_REM_INV_RD_REQ_ERR: "remote invalidate RD request error",
 *                 e.IBV_WC_REM_ABORT_ERR: "remote abort error",             # <<<<<<<<<<<<<<
 *                 e.IBV_WC_INV_EECN_ERR: "invalidate EECN error",
 *                 e.IBV_WC_INV_EEC_STATE_ERR: "invalidate EEC state error",
*/
      __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(IBV_WC_REM_ABORT_ERR); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 616, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_mstate_global->__pyx_kp_u_remote_abort_error) < (0)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyverbs/cq.pyx":617
 *                 e.IBV_WC_REM_INV_RD_REQ_ERR: "remote invalidate RD request error",
 *                 e.IBV_WC_REM_ABORT_ERR: "remote abort error",
 *                 e.IBV_WC_INV_EECN_ERR: "invalidate EECN error",             # <<<<<<<<<<<<<<
 *                 e.IBV_WC_INV_EEC_STATE_ERR: "invalidate EEC state error",
 *                 e.IBV_WC_FATAL_ERR: "WC fatal error",
*/
      __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(IBV_WC_INV_EECN_ERR); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 617, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_mstate_global->__pyx_kp_u_invalidate_EECN_error) < (0)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyverbs/cq.pyx":618
 *                 e.IBV_WC_REM_ABORT_ERR: "remote abort error",
 *                 e.IBV_WC_INV_EECN_ERR: "invalidate EECN error",
 *                 e.IBV_WC_INV_EEC_STATE_ERR: "invalidate EEC state error",             # <<<<<<<<<<<<<<
 *                 e.IBV_WC_FATAL_ERR: "WC fatal error",
 *                 e.IBV_WC_RESP_TIMEOUT_ERR: "response timeout error",
*/
      __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(IBV_WC_INV_EEC_STATE_ERR); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 618, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_mstate_global->__pyx_kp_u_invalidate_EEC_state_error) < (0)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyverbs/cq.pyx":619
 *                 e.IBV_WC_INV_EECN_ERR: "invalidate EECN error",
 *                 e.IBV_WC_INV_EEC_STATE_ERR: "invalidate EEC state error",
 *                 e.IBV_WC_FATAL_ERR: "WC fatal error",             # <<<<<<<<<<<<<<
 *                 e.IBV_WC_RESP_TIMEOUT_ERR: "response timeout error",
 *                 e.IBV_WC_GENERAL_ERR: "general error"}[status]
*/
      __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(IBV_WC_FATAL_ERR); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 619, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_mstate_global->__pyx_kp_u_WC_fatal_error) < (0)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyverbs/cq.pyx":620
 *                 e.IBV_WC_INV_EEC_STATE_ERR: "invalidate EEC state error",
 *                 e.IBV_WC_FATAL_ERR: "WC fatal error",
 *                 e.IBV_WC_RESP_TIMEOUT_ERR: "response timeout error",             # <<<<<<<<<<<<<<
 *                 e.IBV_WC_GENERAL_ERR: "general error"}[status]
 *     except KeyError:
*/
      __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(IBV_WC_RESP_TIMEOUT_ERR); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 620, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_mstate_global->__pyx_kp_u_response_timeout_error) < (0)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyverbs/cq.pyx":621
 *                 e.IBV_WC_FATAL_ERR: "WC fatal error",
 *                 e.IBV_WC_RESP_TIMEOUT_ERR: "response timeout error",
 *                 e.IBV_WC_GENERAL_ERR: "general error"}[status]             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         return "Unknown CQE status"
*/
      __pyx_t_5 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(IBV_WC_GENERAL_ERR); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 621, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_4, __pyx_t_5, __pyx_mstate_global->__pyx_kp_u_general_error) < (0)) __PYX_ERR(0, 600, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_t_4, __pyx_v_status); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 621, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_5;
      __pyx_t_5 = 0;
      goto __pyx_L7_try_return;

      /* "pyverbs/cq.pyx":599
 * 
 * def cqe_status_to_str(status):
 *     try:             # <<<<<<<<<<<<<<
 *         return {e.IBV_WC_SUCCESS: "success",
 *                 e.IBV_WC_LOC_LEN_ERR: "local length error",
*/
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "pyverbs/cq.pyx":622
 *                 e.IBV_WC_RESP_TIMEOUT_ERR: "response timeout error",
 *                 e.IBV_WC_GENERAL_ERR: "general error"}[status]
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         return "Unknown CQE status"
 * 
*/
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_KeyError))));
    if (__pyx_t_6) {
      __Pyx_ErrRestore(0,0,0);

      /* "pyverbs/cq.pyx":623
 *                 e.IBV_WC_GENERAL_ERR: "general error"}[status]
 *     except KeyError:
 *         return "Unknown CQE status"             # <<<<<<<<<<<<<<
 * 
 * def cqe_opcode_to_str(opcode):
*/
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_Unknown_CQE_status);
      __pyx_r = __pyx_mstate_global->__pyx_kp_u_Unknown_CQE_status;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;

    /* "pyverbs/cq.pyx":599
 * 
 * def cqe_status_to_str(status):
 *     try:             # <<<<<<<<<<<<<<
 *         return {e.IBV_WC_SUCCESS: "success",
 *                 e.IBV_WC_LOC_LEN_ERR: "local length error",
*/
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "pyverbs/cq.pyx":598
 * 
 * 
 * def cqe_status_to_str(status):             # <<<<<<<<<<<<<<
 *     try:
 *         return {e.IBV_WC_SUCCESS: "success",
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyverbs.cq.cqe_status_to_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":625
 *         return "Unknown CQE status"
 * 
 * def cqe_opcode_to_str(opcode):             # <<<<<<<<<<<<<<
 *     try:
 *         return {0x0: "Send", 0x1:"RDMA write", 0x2: "RDMA read",
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_3cqe_opcode_to_str(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_3cqe_opcode_to_str = {"cqe_opcode_to_str", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_3cqe_opcode_to_str, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_3cqe_opcode_to_str(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_opcode = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cqe_opcode_to_str (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_opcode,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 625, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 625, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "cqe_opcode_to_str", 0) < (0)) __PYX_ERR(0, 625, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("cqe_opcode_to_str", 1, 1, 1, i); __PYX_ERR(0, 625, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 625, __pyx_L3_error)
    }
    __pyx_v_opcode = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cqe_opcode_to_str", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 625, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cq.cqe_opcode_to_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2cq_2cqe_opcode_to_str(__pyx_self, __pyx_v_opcode);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_2cqe_opcode_to_str(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_opcode) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cqe_opcode_to_str", 0);

  /* "pyverbs/cq.pyx":626
 * 
 * def cqe_opcode_to_str(opcode):
 *     try:             # <<<<<<<<<<<<<<
 *         return {0x0: "Send", 0x1:"RDMA write", 0x2: "RDMA read",
 *                 0x3: "Compare and swap", 0x4: "Fetch and add",
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "pyverbs/cq.pyx":627
 * def cqe_opcode_to_str(opcode):
 *     try:
 *         return {0x0: "Send", 0x1:"RDMA write", 0x2: "RDMA read",             # <<<<<<<<<<<<<<
 *                 0x3: "Compare and swap", 0x4: "Fetch and add",
 *                 0x5: "Bind Memory window", 0x6: "Local invalidate",
*/
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_PyDict_NewPresized(16); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 627, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_int_0, __pyx_mstate_global->__pyx_n_u_Send) < (0)) __PYX_ERR(0, 627, __pyx_L3_error)
      if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_int_1, __pyx_mstate_global->__pyx_kp_u_RDMA_write) < (0)) __PYX_ERR(0, 627, __pyx_L3_error)
      if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_int_2, __pyx_mstate_global->__pyx_kp_u_RDMA_read) < (0)) __PYX_ERR(0, 627, __pyx_L3_error)
      if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_int_3, __pyx_mstate_global->__pyx_kp_u_Compare_and_swap) < (0)) __PYX_ERR(0, 627, __pyx_L3_error)
      if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_int_4, __pyx_mstate_global->__pyx_kp_u_Fetch_and_add) < (0)) __PYX_ERR(0, 627, __pyx_L3_error)
      if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_int_5, __pyx_mstate_global->__pyx_kp_u_Bind_Memory_window) < (0)) __PYX_ERR(0, 627, __pyx_L3_error)
      if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_int_6, __pyx_mstate_global->__pyx_kp_u_Local_invalidate) < (0)) __PYX_ERR(0, 627, __pyx_L3_error)
      if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_int_7, __pyx_mstate_global->__pyx_n_u_TSO) < (0)) __PYX_ERR(0, 627, __pyx_L3_error)
      if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_int_128, __pyx_mstate_global->__pyx_n_u_Receive) < (0)) __PYX_ERR(0, 627, __pyx_L3_error)
      if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_int_129, __pyx_mstate_global->__pyx_kp_u_Receive_RDMA_with_immediate) < (0)) __PYX_ERR(0, 627, __pyx_L3_error)
      if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_int_130, __pyx_mstate_global->__pyx_kp_u_Tag_matching_add) < (0)) __PYX_ERR(0, 627, __pyx_L3_error)
      if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_int_131, __pyx_mstate_global->__pyx_kp_u_Tag_matching_delete) < (0)) __PYX_ERR(0, 627, __pyx_L3_error)
      if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_int_132, __pyx_mstate_global->__pyx_kp_u_Tag_matching_sync) < (0)) __PYX_ERR(0, 627, __pyx_L3_error)
      if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_int_133, __pyx_mstate_global->__pyx_kp_u_Tag_matching_receive) < (0)) __PYX_ERR(0, 627, __pyx_L3_error)
      if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_int_134, __pyx_mstate_global->__pyx_kp_u_Tag_matching_no_tag) < (0)) __PYX_ERR(0, 627, __pyx_L3_error)
      if (PyDict_SetItem(__pyx_t_4, __pyx_mstate_global->__pyx_int_135, __pyx_mstate_global->__pyx_kp_u_Driver_WR) < (0)) __PYX_ERR(0, 627, __pyx_L3_error)

      /* "pyverbs/cq.pyx":634
 *                 0x82: "Tag matching - add", 0x83: "Tag matching - delete",
 *                 0x84: "Tag matching - sync", 0x85: "Tag matching - receive",
 *                 0x86: "Tag matching - no tag", 0x87: "Driver WR"}[opcode]             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         return "Unknown CQE opcode {op}".format(op=opcode)
*/
      __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_t_4, __pyx_v_opcode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 634, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_5;
      __pyx_t_5 = 0;
      goto __pyx_L7_try_return;

      /* "pyverbs/cq.pyx":626
 * 
 * def cqe_opcode_to_str(opcode):
 *     try:             # <<<<<<<<<<<<<<
 *         return {0x0: "Send", 0x1:"RDMA write", 0x2: "RDMA read",
 *                 0x3: "Compare and swap", 0x4: "Fetch and add",
*/
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "pyverbs/cq.pyx":635
 *                 0x84: "Tag matching - sync", 0x85: "Tag matching - receive",
 *                 0x86: "Tag matching - no tag", 0x87: "Driver WR"}[opcode]
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         return "Unknown CQE opcode {op}".format(op=opcode)
 * 
*/
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_KeyError))));
    if (__pyx_t_6) {
      __Pyx_AddTraceback("pyverbs.cq.cqe_opcode_to_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_7) < 0) __PYX_ERR(0, 635, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "pyverbs/cq.pyx":636
 *                 0x86: "Tag matching - no tag", 0x87: "Driver WR"}[opcode]
 *     except KeyError:
 *         return "Unknown CQE opcode {op}".format(op=opcode)             # <<<<<<<<<<<<<<
 * 
 * def flags_to_str(flags, dictionary):
*/
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_9 = __pyx_mstate_global->__pyx_kp_u_Unknown_CQE_opcode_op;
      __Pyx_INCREF(__pyx_t_9);
      __pyx_t_10 = 0;
      {
        PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_9, NULL};
        __pyx_t_11 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 636, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_op, __pyx_v_opcode, __pyx_t_11, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 636, __pyx_L5_except_error)
        __pyx_t_8 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_format, __pyx_callargs+__pyx_t_10, (1-__pyx_t_10) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_11);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 636, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_8);
      }
      __pyx_r = __pyx_t_8;
      __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;

    /* "pyverbs/cq.pyx":626
 * 
 * def cqe_opcode_to_str(opcode):
 *     try:             # <<<<<<<<<<<<<<
 *         return {0x0: "Send", 0x1:"RDMA write", 0x2: "RDMA read",
 *                 0x3: "Compare and swap", 0x4: "Fetch and add",
*/
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "pyverbs/cq.pyx":625
 *         return "Unknown CQE status"
 * 
 * def cqe_opcode_to_str(opcode):             # <<<<<<<<<<<<<<
 *     try:
 *         return {0x0: "Send", 0x1:"RDMA write", 0x2: "RDMA read",
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("pyverbs.cq.cqe_opcode_to_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":638
 *         return "Unknown CQE opcode {op}".format(op=opcode)
 * 
 * def flags_to_str(flags, dictionary):             # <<<<<<<<<<<<<<
 *     flags_str = ""
 *     for f in dictionary:
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_5flags_to_str(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_5flags_to_str = {"flags_to_str", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_5flags_to_str, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_5flags_to_str(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_flags = 0;
  PyObject *__pyx_v_dictionary = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("flags_to_str (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_flags,&__pyx_mstate_global->__pyx_n_u_dictionary,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 638, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 638, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 638, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "flags_to_str", 0) < (0)) __PYX_ERR(0, 638, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("flags_to_str", 1, 2, 2, i); __PYX_ERR(0, 638, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 638, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 638, __pyx_L3_error)
    }
    __pyx_v_flags = values[0];
    __pyx_v_dictionary = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("flags_to_str", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 638, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cq.flags_to_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2cq_4flags_to_str(__pyx_self, __pyx_v_flags, __pyx_v_dictionary);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_4flags_to_str(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_flags, PyObject *__pyx_v_dictionary) {
  PyObject *__pyx_v_flags_str = NULL;
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("flags_to_str", 0);

  /* "pyverbs/cq.pyx":639
 * 
 * def flags_to_str(flags, dictionary):
 *     flags_str = ""             # <<<<<<<<<<<<<<
 *     for f in dictionary:
 *         if flags & f:
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u__2);
  __pyx_v_flags_str = __pyx_mstate_global->__pyx_kp_u__2;

  /* "pyverbs/cq.pyx":640
 * def flags_to_str(flags, dictionary):
 *     flags_str = ""
 *     for f in dictionary:             # <<<<<<<<<<<<<<
 *         if flags & f:
 *             flags_str += dictionary[f]
*/
  if (likely(PyList_CheckExact(__pyx_v_dictionary)) || PyTuple_CheckExact(__pyx_v_dictionary)) {
    __pyx_t_1 = __pyx_v_dictionary; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_dictionary); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 640, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 640, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 640, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        __pyx_t_4 = __Pyx_PyList_GetItemRefFast(__pyx_t_1, __pyx_t_2, __Pyx_ReferenceSharing_OwnStrongReference);
        ++__pyx_t_2;
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 640, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2));
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2);
        #endif
        ++__pyx_t_2;
      }
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 640, __pyx_L1_error)
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 640, __pyx_L1_error)
          PyErr_Clear();
        }
        break;
      }
    }
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_f, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "pyverbs/cq.pyx":641
 *     flags_str = ""
 *     for f in dictionary:
 *         if flags & f:             # <<<<<<<<<<<<<<
 *             flags_str += dictionary[f]
 *             flags_str += " "
*/
    __pyx_t_4 = PyNumber_And(__pyx_v_flags, __pyx_v_f); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 641, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 641, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_5) {

      /* "pyverbs/cq.pyx":642
 *     for f in dictionary:
 *         if flags & f:
 *             flags_str += dictionary[f]             # <<<<<<<<<<<<<<
 *             flags_str += " "
 *     return flags_str
*/
      __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_dictionary, __pyx_v_f); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 642, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = PyNumber_InPlaceAdd(__pyx_v_flags_str, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 642, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v_flags_str, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "pyverbs/cq.pyx":643
 *         if flags & f:
 *             flags_str += dictionary[f]
 *             flags_str += " "             # <<<<<<<<<<<<<<
 *     return flags_str
 * 
*/
      __pyx_t_6 = PyNumber_InPlaceAdd(__pyx_v_flags_str, __pyx_mstate_global->__pyx_kp_u__3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 643, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_flags_str, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "pyverbs/cq.pyx":641
 *     flags_str = ""
 *     for f in dictionary:
 *         if flags & f:             # <<<<<<<<<<<<<<
 *             flags_str += dictionary[f]
 *             flags_str += " "
*/
    }

    /* "pyverbs/cq.pyx":640
 * def flags_to_str(flags, dictionary):
 *     flags_str = ""
 *     for f in dictionary:             # <<<<<<<<<<<<<<
 *         if flags & f:
 *             flags_str += dictionary[f]
*/
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":644
 *             flags_str += dictionary[f]
 *             flags_str += " "
 *     return flags_str             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_flags_str);
  __pyx_r = __pyx_v_flags_str;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":638
 *         return "Unknown CQE opcode {op}".format(op=opcode)
 * 
 * def flags_to_str(flags, dictionary):             # <<<<<<<<<<<<<<
 *     flags_str = ""
 *     for f in dictionary:
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyverbs.cq.flags_to_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_flags_str);
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":647
 * 
 * 
 * def cqe_flags_to_str(flags):             # <<<<<<<<<<<<<<
 *     cqe_flags = {1: "GRH", 2: "With immediate", 4: "IP csum OK",
 *                  8: "With invalidate", 16: "TM sync request", 32: "TM match",
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_7cqe_flags_to_str(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_7cqe_flags_to_str = {"cqe_flags_to_str", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_7cqe_flags_to_str, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_7cqe_flags_to_str(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_flags = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cqe_flags_to_str (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_flags,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 647, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 647, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "cqe_flags_to_str", 0) < (0)) __PYX_ERR(0, 647, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("cqe_flags_to_str", 1, 1, 1, i); __PYX_ERR(0, 647, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 647, __pyx_L3_error)
    }
    __pyx_v_flags = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cqe_flags_to_str", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 647, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cq.cqe_flags_to_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2cq_6cqe_flags_to_str(__pyx_self, __pyx_v_flags);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_6cqe_flags_to_str(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_flags) {
  PyObject *__pyx_v_cqe_flags = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cqe_flags_to_str", 0);

  /* "pyverbs/cq.pyx":648
 * 
 * def cqe_flags_to_str(flags):
 *     cqe_flags = {1: "GRH", 2: "With immediate", 4: "IP csum OK",             # <<<<<<<<<<<<<<
 *                  8: "With invalidate", 16: "TM sync request", 32: "TM match",
 *                  64: "TM data valid"}
*/
  __pyx_t_1 = __Pyx_PyDict_NewPresized(7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_int_1, __pyx_mstate_global->__pyx_n_u_GRH) < (0)) __PYX_ERR(0, 648, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_int_2, __pyx_mstate_global->__pyx_kp_u_With_immediate) < (0)) __PYX_ERR(0, 648, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_int_4, __pyx_mstate_global->__pyx_kp_u_IP_csum_OK) < (0)) __PYX_ERR(0, 648, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_int_8, __pyx_mstate_global->__pyx_kp_u_With_invalidate) < (0)) __PYX_ERR(0, 648, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_int_16, __pyx_mstate_global->__pyx_kp_u_TM_sync_request) < (0)) __PYX_ERR(0, 648, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_int_32, __pyx_mstate_global->__pyx_kp_u_TM_match) < (0)) __PYX_ERR(0, 648, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_mstate_global->__pyx_int_64, __pyx_mstate_global->__pyx_kp_u_TM_data_valid) < (0)) __PYX_ERR(0, 648, __pyx_L1_error)
  __pyx_v_cqe_flags = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":651
 *                  8: "With invalidate", 16: "TM sync request", 32: "TM match",
 *                  64: "TM data valid"}
 *     return flags_to_str(flags, cqe_flags)             # <<<<<<<<<<<<<<
 * 
 * def create_wc_flags_to_str(flags):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_flags_to_str); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
    __pyx_t_4 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_v_flags, __pyx_v_cqe_flags};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_4, (3-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 651, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":647
 * 
 * 
 * def cqe_flags_to_str(flags):             # <<<<<<<<<<<<<<
 *     cqe_flags = {1: "GRH", 2: "With immediate", 4: "IP csum OK",
 *                  8: "With invalidate", 16: "TM sync request", 32: "TM match",
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyverbs.cq.cqe_flags_to_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_cqe_flags);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cq.pyx":653
 *     return flags_to_str(flags, cqe_flags)
 * 
 * def create_wc_flags_to_str(flags):             # <<<<<<<<<<<<<<
 *     cqe_flags = {e.IBV_WC_EX_WITH_BYTE_LEN: 'IBV_WC_EX_WITH_BYTE_LEN',
 *                  e.IBV_WC_EX_WITH_IMM: 'IBV_WC_EX_WITH_IMM',
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_2cq_9create_wc_flags_to_str(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_2cq_9create_wc_flags_to_str = {"create_wc_flags_to_str", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_9create_wc_flags_to_str, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_2cq_9create_wc_flags_to_str(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_flags = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_wc_flags_to_str (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_flags,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 653, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 653, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "create_wc_flags_to_str", 0) < (0)) __PYX_ERR(0, 653, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("create_wc_flags_to_str", 1, 1, 1, i); __PYX_ERR(0, 653, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 653, __pyx_L3_error)
    }
    __pyx_v_flags = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_wc_flags_to_str", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 653, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cq.create_wc_flags_to_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_2cq_8create_wc_flags_to_str(__pyx_self, __pyx_v_flags);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_2cq_8create_wc_flags_to_str(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_flags) {
  PyObject *__pyx_v_cqe_flags = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create_wc_flags_to_str", 0);

  /* "pyverbs/cq.pyx":654
 * 
 * def create_wc_flags_to_str(flags):
 *     cqe_flags = {e.IBV_WC_EX_WITH_BYTE_LEN: 'IBV_WC_EX_WITH_BYTE_LEN',             # <<<<<<<<<<<<<<
 *                  e.IBV_WC_EX_WITH_IMM: 'IBV_WC_EX_WITH_IMM',
 *                  e.IBV_WC_EX_WITH_QP_NUM: 'IBV_WC_EX_WITH_QP_NUM',
*/
  __pyx_t_1 = __Pyx_PyDict_NewPresized(11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py(IBV_WC_EX_WITH_BYTE_LEN); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_BYTE_LEN) < (0)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":655
 * def create_wc_flags_to_str(flags):
 *     cqe_flags = {e.IBV_WC_EX_WITH_BYTE_LEN: 'IBV_WC_EX_WITH_BYTE_LEN',
 *                  e.IBV_WC_EX_WITH_IMM: 'IBV_WC_EX_WITH_IMM',             # <<<<<<<<<<<<<<
 *                  e.IBV_WC_EX_WITH_QP_NUM: 'IBV_WC_EX_WITH_QP_NUM',
 *                  e.IBV_WC_EX_WITH_SRC_QP: 'IBV_WC_EX_WITH_SRC_QP',
*/
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py(IBV_WC_EX_WITH_IMM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_IMM) < (0)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":656
 *     cqe_flags = {e.IBV_WC_EX_WITH_BYTE_LEN: 'IBV_WC_EX_WITH_BYTE_LEN',
 *                  e.IBV_WC_EX_WITH_IMM: 'IBV_WC_EX_WITH_IMM',
 *                  e.IBV_WC_EX_WITH_QP_NUM: 'IBV_WC_EX_WITH_QP_NUM',             # <<<<<<<<<<<<<<
 *                  e.IBV_WC_EX_WITH_SRC_QP: 'IBV_WC_EX_WITH_SRC_QP',
 *                  e.IBV_WC_EX_WITH_SLID: 'IBV_WC_EX_WITH_SLID',
*/
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py(IBV_WC_EX_WITH_QP_NUM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_QP_NUM) < (0)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":657
 *                  e.IBV_WC_EX_WITH_IMM: 'IBV_WC_EX_WITH_IMM',
 *                  e.IBV_WC_EX_WITH_QP_NUM: 'IBV_WC_EX_WITH_QP_NUM',
 *                  e.IBV_WC_EX_WITH_SRC_QP: 'IBV_WC_EX_WITH_SRC_QP',             # <<<<<<<<<<<<<<
 *                  e.IBV_WC_EX_WITH_SLID: 'IBV_WC_EX_WITH_SLID',
 *                  e.IBV_WC_EX_WITH_SL: 'IBV_WC_EX_WITH_SL',
*/
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py(IBV_WC_EX_WITH_SRC_QP); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_SRC_QP) < (0)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":658
 *                  e.IBV_WC_EX_WITH_QP_NUM: 'IBV_WC_EX_WITH_QP_NUM',
 *                  e.IBV_WC_EX_WITH_SRC_QP: 'IBV_WC_EX_WITH_SRC_QP',
 *                  e.IBV_WC_EX_WITH_SLID: 'IBV_WC_EX_WITH_SLID',             # <<<<<<<<<<<<<<
 *                  e.IBV_WC_EX_WITH_SL: 'IBV_WC_EX_WITH_SL',
 *                  e.IBV_WC_EX_WITH_DLID_PATH_BITS: 'IBV_WC_EX_WITH_DLID_PATH_BITS',
*/
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py(IBV_WC_EX_WITH_SLID); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_SLID) < (0)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":659
 *                  e.IBV_WC_EX_WITH_SRC_QP: 'IBV_WC_EX_WITH_SRC_QP',
 *                  e.IBV_WC_EX_WITH_SLID: 'IBV_WC_EX_WITH_SLID',
 *                  e.IBV_WC_EX_WITH_SL: 'IBV_WC_EX_WITH_SL',             # <<<<<<<<<<<<<<
 *                  e.IBV_WC_EX_WITH_DLID_PATH_BITS: 'IBV_WC_EX_WITH_DLID_PATH_BITS',
 *                  e.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP: 'IBV_WC_EX_WITH_COMPLETION_TIMESTAMP',
*/
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py(IBV_WC_EX_WITH_SL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_SL) < (0)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":660
 *                  e.IBV_WC_EX_WITH_SLID: 'IBV_WC_EX_WITH_SLID',
 *                  e.IBV_WC_EX_WITH_SL: 'IBV_WC_EX_WITH_SL',
 *                  e.IBV_WC_EX_WITH_DLID_PATH_BITS: 'IBV_WC_EX_WITH_DLID_PATH_BITS',             # <<<<<<<<<<<<<<
 *                  e.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP: 'IBV_WC_EX_WITH_COMPLETION_TIMESTAMP',
 *                  e.IBV_WC_EX_WITH_CVLAN: 'IBV_WC_EX_WITH_CVLAN',
*/
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py(IBV_WC_EX_WITH_DLID_PATH_BITS); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_DLID_PATH_BITS) < (0)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":661
 *                  e.IBV_WC_EX_WITH_SL: 'IBV_WC_EX_WITH_SL',
 *                  e.IBV_WC_EX_WITH_DLID_PATH_BITS: 'IBV_WC_EX_WITH_DLID_PATH_BITS',
 *                  e.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP: 'IBV_WC_EX_WITH_COMPLETION_TIMESTAMP',             # <<<<<<<<<<<<<<
 *                  e.IBV_WC_EX_WITH_CVLAN: 'IBV_WC_EX_WITH_CVLAN',
 *                  e.IBV_WC_EX_WITH_FLOW_TAG: 'IBV_WC_EX_WITH_FLOW_TAG',
*/
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py(IBV_WC_EX_WITH_COMPLETION_TIMESTAMP); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_COMPLETION_TIMEST) < (0)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":662
 *                  e.IBV_WC_EX_WITH_DLID_PATH_BITS: 'IBV_WC_EX_WITH_DLID_PATH_BITS',
 *                  e.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP: 'IBV_WC_EX_WITH_COMPLETION_TIMESTAMP',
 *                  e.IBV_WC_EX_WITH_CVLAN: 'IBV_WC_EX_WITH_CVLAN',             # <<<<<<<<<<<<<<
 *                  e.IBV_WC_EX_WITH_FLOW_TAG: 'IBV_WC_EX_WITH_FLOW_TAG',
 *                  e.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK: 'IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK'}
*/
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py(IBV_WC_EX_WITH_CVLAN); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_CVLAN) < (0)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":663
 *                  e.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP: 'IBV_WC_EX_WITH_COMPLETION_TIMESTAMP',
 *                  e.IBV_WC_EX_WITH_CVLAN: 'IBV_WC_EX_WITH_CVLAN',
 *                  e.IBV_WC_EX_WITH_FLOW_TAG: 'IBV_WC_EX_WITH_FLOW_TAG',             # <<<<<<<<<<<<<<
 *                  e.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK: 'IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK'}
 *     return flags_to_str(flags, cqe_flags)
*/
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py(IBV_WC_EX_WITH_FLOW_TAG); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_FLOW_TAG) < (0)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":664
 *                  e.IBV_WC_EX_WITH_CVLAN: 'IBV_WC_EX_WITH_CVLAN',
 *                  e.IBV_WC_EX_WITH_FLOW_TAG: 'IBV_WC_EX_WITH_FLOW_TAG',
 *                  e.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK: 'IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK'}             # <<<<<<<<<<<<<<
 *     return flags_to_str(flags, cqe_flags)
*/
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py(IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_2, __pyx_mstate_global->__pyx_n_u_IBV_WC_EX_WITH_COMPLETION_TIMEST_2) < (0)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_cqe_flags = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyverbs/cq.pyx":665
 *                  e.IBV_WC_EX_WITH_FLOW_TAG: 'IBV_WC_EX_WITH_FLOW_TAG',
 *                  e.IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK: 'IBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCK'}
 *     return flags_to_str(flags, cqe_flags)             # <<<<<<<<<<<<<<
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_flags_to_str); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
    __pyx_t_4 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_v_flags, __pyx_v_cqe_flags};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_4, (3-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 665, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cq.pyx":653
 *     return flags_to_str(flags, cqe_flags)
 * 
 * def create_wc_flags_to_str(flags):             # <<<<<<<<<<<<<<
 *     cqe_flags = {e.IBV_WC_EX_WITH_BYTE_LEN: 'IBV_WC_EX_WITH_BYTE_LEN',
 *                  e.IBV_WC_EX_WITH_IMM: 'IBV_WC_EX_WITH_IMM',
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyverbs.cq.create_wc_flags_to_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_cqe_flags);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
/* #### Code section: module_exttypes ### */
static struct __pyx_vtabstruct_7pyverbs_2cq_CompChannel __pyx_vtable_7pyverbs_2cq_CompChannel;

static PyObject *__pyx_tp_new_7pyverbs_2cq_CompChannel(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7pyverbs_2cq_CompChannel *p;
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7pyverbs_2cq_CompChannel *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM*)__pyx_vtabptr_7pyverbs_2cq_CompChannel;
  p->context = Py_None; Py_INCREF(Py_None);
  p->cqs = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7pyverbs_2cq_CompChannel(PyObject *o) {
  struct __pyx_obj_7pyverbs_2cq_CompChannel *p = (struct __pyx_obj_7pyverbs_2cq_CompChannel *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyverbs_2cq_CompChannel) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_7pyverbs_2cq_11CompChannel_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->context);
  Py_CLEAR(p->cqs);
  if (PyType_IS_GC(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) PyObject_GC_Track(o);
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_dealloc, destructor)(o); else
  #endif
  __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_7pyverbs_2cq_CompChannel);
}

static int __pyx_tp_traverse_7pyverbs_2cq_CompChannel(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7pyverbs_2cq_CompChannel *p = (struct __pyx_obj_7pyverbs_2cq_CompChannel *)o;
  #if !CYTHON_USE_MODULE_STATE
  e = 0;
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) {
    traverseproc traverse = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_traverse, traverseproc);
    if (traverse) { e = traverse(o, v, a); }
  } else
  #endif
  { e = __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7pyverbs_2cq_CompChannel); }
  if (e) return e;
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  if (p->context) {
    e = (*v)(p->context, a); if (e) return e;
  }
  if (p->cqs) {
    e = (*v)(p->cqs, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyverbs_2cq_CompChannel(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7pyverbs_2cq_CompChannel *p = (struct __pyx_obj_7pyverbs_2cq_CompChannel *)o;
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) {
    inquiry clear = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_clear, inquiry);
    if (clear) clear(o);
  } else
  #endif
  { __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7pyverbs_2cq_CompChannel); }
  tmp = ((PyObject*)p->context);
  p->context = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->cqs);
  p->cqs = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7pyverbs_2cq_11CompChannel_fd(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_11CompChannel_2fd_1__get__(o);
}

static PyMethodDef __pyx_methods_7pyverbs_2cq_CompChannel[] = {
  {"get_cq_event", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_11CompChannel_7get_cq_event, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2cq_11CompChannel_6get_cq_event},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_11CompChannel_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_11CompChannel_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7pyverbs_2cq_CompChannel[] = {
  {"fd", __pyx_getprop_7pyverbs_2cq_11CompChannel_fd, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyverbs_2cq_CompChannel_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyverbs_2cq_CompChannel},
  {Py_tp_doc, (void *)PyDoc_STR("\n    A completion channel is a file descriptor used to deliver completion\n    notifications to a userspace process. When a completion event is generated\n    for a CQ, the event is delivered via the completion channel attached to the\n    CQ.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyverbs_2cq_CompChannel},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyverbs_2cq_CompChannel},
  {Py_tp_methods, (void *)__pyx_methods_7pyverbs_2cq_CompChannel},
  {Py_tp_getset, (void *)__pyx_getsets_7pyverbs_2cq_CompChannel},
  {Py_tp_init, (void *)__pyx_pw_7pyverbs_2cq_11CompChannel_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7pyverbs_2cq_CompChannel},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyverbs_2cq_CompChannel_spec = {
  "pyverbs.cq.CompChannel",
  sizeof(struct __pyx_obj_7pyverbs_2cq_CompChannel),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyverbs_2cq_CompChannel_slots,
};
#else

static PyTypeObject __pyx_type_7pyverbs_2cq_CompChannel = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyverbs.cq.""CompChannel", /*tp_name*/
  sizeof(struct __pyx_obj_7pyverbs_2cq_CompChannel), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyverbs_2cq_CompChannel, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    A completion channel is a file descriptor used to deliver completion\n    notifications to a userspace process. When a completion event is generated\n    for a CQ, the event is delivered via the completion channel attached to the\n    CQ.\n    "), /*tp_doc*/
  __pyx_tp_traverse_7pyverbs_2cq_CompChannel, /*tp_traverse*/
  __pyx_tp_clear_7pyverbs_2cq_CompChannel, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyverbs_2cq_CompChannel, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7pyverbs_2cq_CompChannel, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7pyverbs_2cq_11CompChannel_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyverbs_2cq_CompChannel, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7pyverbs_2cq_CQ __pyx_vtable_7pyverbs_2cq_CQ;

static PyObject *__pyx_tp_new_7pyverbs_2cq_CQ(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7pyverbs_2cq_CQ *p;
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7pyverbs_2cq_CQ *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM*)__pyx_vtabptr_7pyverbs_2cq_CQ;
  p->context = Py_None; Py_INCREF(Py_None);
  p->qps = Py_None; Py_INCREF(Py_None);
  p->srqs = Py_None; Py_INCREF(Py_None);
  p->channel = Py_None; Py_INCREF(Py_None);
  p->num_events = Py_None; Py_INCREF(Py_None);
  p->wqs = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7pyverbs_2cq_CQ(PyObject *o) {
  struct __pyx_obj_7pyverbs_2cq_CQ *p = (struct __pyx_obj_7pyverbs_2cq_CQ *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyverbs_2cq_CQ) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_7pyverbs_2cq_2CQ_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->context);
  Py_CLEAR(p->qps);
  Py_CLEAR(p->srqs);
  Py_CLEAR(p->channel);
  Py_CLEAR(p->num_events);
  Py_CLEAR(p->wqs);
  if (PyType_IS_GC(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) PyObject_GC_Track(o);
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_dealloc, destructor)(o); else
  #endif
  __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_7pyverbs_2cq_CQ);
}

static int __pyx_tp_traverse_7pyverbs_2cq_CQ(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7pyverbs_2cq_CQ *p = (struct __pyx_obj_7pyverbs_2cq_CQ *)o;
  #if !CYTHON_USE_MODULE_STATE
  e = 0;
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) {
    traverseproc traverse = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_traverse, traverseproc);
    if (traverse) { e = traverse(o, v, a); }
  } else
  #endif
  { e = __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7pyverbs_2cq_CQ); }
  if (e) return e;
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  if (p->context) {
    e = (*v)(p->context, a); if (e) return e;
  }
  if (p->qps) {
    e = (*v)(p->qps, a); if (e) return e;
  }
  if (p->srqs) {
    e = (*v)(p->srqs, a); if (e) return e;
  }
  if (p->channel) {
    e = (*v)(p->channel, a); if (e) return e;
  }
  if (p->num_events) {
    e = (*v)(p->num_events, a); if (e) return e;
  }
  if (p->wqs) {
    e = (*v)(p->wqs, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyverbs_2cq_CQ(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7pyverbs_2cq_CQ *p = (struct __pyx_obj_7pyverbs_2cq_CQ *)o;
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) {
    inquiry clear = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_clear, inquiry);
    if (clear) clear(o);
  } else
  #endif
  { __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7pyverbs_2cq_CQ); }
  tmp = ((PyObject*)p->context);
  p->context = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->qps);
  p->qps = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->srqs);
  p->srqs = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->channel);
  p->channel = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->num_events);
  p->num_events = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->wqs);
  p->wqs = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7pyverbs_2cq_2CQ_comp_channel(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_2CQ_12comp_channel_1__get__(o);
}

static PyObject *__pyx_getprop_7pyverbs_2cq_2CQ_cqe(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_2CQ_3cqe_1__get__(o);
}

static PyObject *__pyx_getprop_7pyverbs_2cq_2CQ_cq(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_2CQ_2cq_1__get__(o);
}

static PyMethodDef __pyx_methods_7pyverbs_2cq_CQ[] = {
  {"resize", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_2CQ_7resize, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2cq_2CQ_6resize},
  {"poll", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_2CQ_9poll, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2cq_2CQ_8poll},
  {"req_notify", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_2CQ_11req_notify, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2cq_2CQ_10req_notify},
  {"ack_events", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_2CQ_13ack_events, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2cq_2CQ_12ack_events},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_2CQ_17__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_2CQ_19__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7pyverbs_2cq_CQ[] = {
  {"comp_channel", __pyx_getprop_7pyverbs_2cq_2CQ_comp_channel, 0, 0, 0},
  {"cqe", __pyx_getprop_7pyverbs_2cq_2CQ_cqe, 0, 0, 0},
  {"cq", __pyx_getprop_7pyverbs_2cq_2CQ_cq, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyverbs_2cq_CQ_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyverbs_2cq_CQ},
  {Py_tp_str, (void *)__pyx_pw_7pyverbs_2cq_2CQ_15__str__},
  {Py_tp_doc, (void *)PyDoc_STR("\n    A Completion Queue is the notification mechanism for work request\n    completions. A CQ can have 0 or more associated QPs.\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyverbs_2cq_CQ},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyverbs_2cq_CQ},
  {Py_tp_methods, (void *)__pyx_methods_7pyverbs_2cq_CQ},
  {Py_tp_getset, (void *)__pyx_getsets_7pyverbs_2cq_CQ},
  {Py_tp_init, (void *)__pyx_pw_7pyverbs_2cq_2CQ_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7pyverbs_2cq_CQ},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyverbs_2cq_CQ_spec = {
  "pyverbs.cq.CQ",
  sizeof(struct __pyx_obj_7pyverbs_2cq_CQ),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyverbs_2cq_CQ_slots,
};
#else

static PyTypeObject __pyx_type_7pyverbs_2cq_CQ = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyverbs.cq.""CQ", /*tp_name*/
  sizeof(struct __pyx_obj_7pyverbs_2cq_CQ), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyverbs_2cq_CQ, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7pyverbs_2cq_2CQ_15__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    A Completion Queue is the notification mechanism for work request\n    completions. A CQ can have 0 or more associated QPs.\n    "), /*tp_doc*/
  __pyx_tp_traverse_7pyverbs_2cq_CQ, /*tp_traverse*/
  __pyx_tp_clear_7pyverbs_2cq_CQ, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyverbs_2cq_CQ, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7pyverbs_2cq_CQ, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7pyverbs_2cq_2CQ_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyverbs_2cq_CQ, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7pyverbs_2cq_CqInitAttrEx(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *p;
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)o);
  p->channel = Py_None; Py_INCREF(Py_None);
  p->parent_domain = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7pyverbs_2cq_CqInitAttrEx(PyObject *o) {
  struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *p = (struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyverbs_2cq_CqInitAttrEx) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->channel);
  Py_CLEAR(p->parent_domain);
  if (PyType_IS_GC(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) PyObject_GC_Track(o);
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_dealloc, destructor)(o); else
  #endif
  __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_7pyverbs_2cq_CqInitAttrEx);
}

static int __pyx_tp_traverse_7pyverbs_2cq_CqInitAttrEx(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *p = (struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)o;
  #if !CYTHON_USE_MODULE_STATE
  e = 0;
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    traverseproc traverse = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_traverse, traverseproc);
    if (traverse) { e = traverse(o, v, a); }
  } else
  #endif
  { e = __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7pyverbs_2cq_CqInitAttrEx); }
  if (e) return e;
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  if (p->channel) {
    e = (*v)(p->channel, a); if (e) return e;
  }
  if (p->parent_domain) {
    e = (*v)(p->parent_domain, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyverbs_2cq_CqInitAttrEx(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *p = (struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx *)o;
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    inquiry clear = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_clear, inquiry);
    if (clear) clear(o);
  } else
  #endif
  { __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7pyverbs_2cq_CqInitAttrEx); }
  tmp = ((PyObject*)p->channel);
  p->channel = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->parent_domain);
  p->parent_domain = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7pyverbs_2cq_12CqInitAttrEx_cqe(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_3cqe_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_12CqInitAttrEx_cqe(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_3cqe_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static int __pyx_setprop_7pyverbs_2cq_12CqInitAttrEx_cq_context(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_10cq_context_1__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2cq_12CqInitAttrEx_parent_domain(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_13parent_domain_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_12CqInitAttrEx_parent_domain(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_13parent_domain_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2cq_12CqInitAttrEx_comp_channel(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_12comp_channel_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_12CqInitAttrEx_comp_channel(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_12comp_channel_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2cq_12CqInitAttrEx_comp_vector(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_11comp_vector_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_12CqInitAttrEx_comp_vector(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_11comp_vector_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2cq_12CqInitAttrEx_wc_flags(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_8wc_flags_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_12CqInitAttrEx_wc_flags(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_8wc_flags_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2cq_12CqInitAttrEx_comp_mask(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_9comp_mask_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_12CqInitAttrEx_comp_mask(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_9comp_mask_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2cq_12CqInitAttrEx_flags(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_5flags_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_12CqInitAttrEx_flags(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_5flags_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_7pyverbs_2cq_CqInitAttrEx[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7pyverbs_2cq_CqInitAttrEx[] = {
  {"cqe", __pyx_getprop_7pyverbs_2cq_12CqInitAttrEx_cqe, __pyx_setprop_7pyverbs_2cq_12CqInitAttrEx_cqe, 0, 0},
  {"cq_context", 0, __pyx_setprop_7pyverbs_2cq_12CqInitAttrEx_cq_context, 0, 0},
  {"parent_domain", __pyx_getprop_7pyverbs_2cq_12CqInitAttrEx_parent_domain, __pyx_setprop_7pyverbs_2cq_12CqInitAttrEx_parent_domain, 0, 0},
  {"comp_channel", __pyx_getprop_7pyverbs_2cq_12CqInitAttrEx_comp_channel, __pyx_setprop_7pyverbs_2cq_12CqInitAttrEx_comp_channel, 0, 0},
  {"comp_vector", __pyx_getprop_7pyverbs_2cq_12CqInitAttrEx_comp_vector, __pyx_setprop_7pyverbs_2cq_12CqInitAttrEx_comp_vector, 0, 0},
  {"wc_flags", __pyx_getprop_7pyverbs_2cq_12CqInitAttrEx_wc_flags, __pyx_setprop_7pyverbs_2cq_12CqInitAttrEx_wc_flags, 0, 0},
  {"comp_mask", __pyx_getprop_7pyverbs_2cq_12CqInitAttrEx_comp_mask, __pyx_setprop_7pyverbs_2cq_12CqInitAttrEx_comp_mask, 0, 0},
  {"flags", __pyx_getprop_7pyverbs_2cq_12CqInitAttrEx_flags, __pyx_setprop_7pyverbs_2cq_12CqInitAttrEx_flags, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyverbs_2cq_CqInitAttrEx_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyverbs_2cq_CqInitAttrEx},
  {Py_tp_str, (void *)__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_3__str__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyverbs_2cq_CqInitAttrEx},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyverbs_2cq_CqInitAttrEx},
  {Py_tp_methods, (void *)__pyx_methods_7pyverbs_2cq_CqInitAttrEx},
  {Py_tp_getset, (void *)__pyx_getsets_7pyverbs_2cq_CqInitAttrEx},
  {Py_tp_init, (void *)__pyx_pw_7pyverbs_2cq_12CqInitAttrEx_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7pyverbs_2cq_CqInitAttrEx},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyverbs_2cq_CqInitAttrEx_spec = {
  "pyverbs.cq.CqInitAttrEx",
  sizeof(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyverbs_2cq_CqInitAttrEx_slots,
};
#else

static PyTypeObject __pyx_type_7pyverbs_2cq_CqInitAttrEx = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyverbs.cq.""CqInitAttrEx", /*tp_name*/
  sizeof(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyverbs_2cq_CqInitAttrEx, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_3__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7pyverbs_2cq_CqInitAttrEx, /*tp_traverse*/
  __pyx_tp_clear_7pyverbs_2cq_CqInitAttrEx, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyverbs_2cq_CqInitAttrEx, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7pyverbs_2cq_CqInitAttrEx, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7pyverbs_2cq_12CqInitAttrEx_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyverbs_2cq_CqInitAttrEx, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7pyverbs_2cq_CQEX __pyx_vtable_7pyverbs_2cq_CQEX;

static PyObject *__pyx_tp_new_7pyverbs_2cq_CQEX(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7pyverbs_2cq_CQEX *p;
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7pyverbs_2cq_CQEX *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM*)__pyx_vtabptr_7pyverbs_2cq_CQEX;
  p->context = Py_None; Py_INCREF(Py_None);
  p->qps = Py_None; Py_INCREF(Py_None);
  p->srqs = Py_None; Py_INCREF(Py_None);
  p->wqs = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7pyverbs_2cq_CQEX(PyObject *o) {
  struct __pyx_obj_7pyverbs_2cq_CQEX *p = (struct __pyx_obj_7pyverbs_2cq_CQEX *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyverbs_2cq_CQEX) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_7pyverbs_2cq_4CQEX_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->context);
  Py_CLEAR(p->qps);
  Py_CLEAR(p->srqs);
  Py_CLEAR(p->wqs);
  if (PyType_IS_GC(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) PyObject_GC_Track(o);
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_dealloc, destructor)(o); else
  #endif
  __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_7pyverbs_2cq_CQEX);
}

static int __pyx_tp_traverse_7pyverbs_2cq_CQEX(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7pyverbs_2cq_CQEX *p = (struct __pyx_obj_7pyverbs_2cq_CQEX *)o;
  #if !CYTHON_USE_MODULE_STATE
  e = 0;
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) {
    traverseproc traverse = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_traverse, traverseproc);
    if (traverse) { e = traverse(o, v, a); }
  } else
  #endif
  { e = __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7pyverbs_2cq_CQEX); }
  if (e) return e;
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  if (p->context) {
    e = (*v)(p->context, a); if (e) return e;
  }
  if (p->qps) {
    e = (*v)(p->qps, a); if (e) return e;
  }
  if (p->srqs) {
    e = (*v)(p->srqs, a); if (e) return e;
  }
  if (p->wqs) {
    e = (*v)(p->wqs, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyverbs_2cq_CQEX(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7pyverbs_2cq_CQEX *p = (struct __pyx_obj_7pyverbs_2cq_CQEX *)o;
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) {
    inquiry clear = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_clear, inquiry);
    if (clear) clear(o);
  } else
  #endif
  { __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7pyverbs_2cq_CQEX); }
  tmp = ((PyObject*)p->context);
  p->context = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->qps);
  p->qps = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->srqs);
  p->srqs = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->wqs);
  p->wqs = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7pyverbs_2cq_4CQEX_status(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_4CQEX_6status_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_4CQEX_status(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_4CQEX_6status_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2cq_4CQEX_wr_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_4CQEX_5wr_id_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_4CQEX_wr_id(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_4CQEX_5wr_id_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_7pyverbs_2cq_CQEX[] = {
  {"start_poll", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_7start_poll, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2cq_4CQEX_6start_poll},
  {"poll_next", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_9poll_next, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2cq_4CQEX_8poll_next},
  {"end_poll", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_11end_poll, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_2cq_4CQEX_10end_poll},
  {"read_opcode", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_13read_opcode, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"read_vendor_err", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_15read_vendor_err, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"read_byte_len", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_17read_byte_len, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"read_imm_data", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_19read_imm_data, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"read_qp_num", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_21read_qp_num, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"read_src_qp", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_23read_src_qp, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"read_wc_flags", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_25read_wc_flags, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"read_slid", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_27read_slid, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"read_sl", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_29read_sl, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"read_dlid_path_bits", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_31read_dlid_path_bits, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"read_timestamp", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_33read_timestamp, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"read_cvlan", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_35read_cvlan, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"read_flow_tag", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_37read_flow_tag, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"read_tm_info", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_39read_tm_info, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"read_completion_wallclock_ns", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_41read_completion_wallclock_ns, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_45__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_4CQEX_47__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7pyverbs_2cq_CQEX[] = {
  {"status", __pyx_getprop_7pyverbs_2cq_4CQEX_status, __pyx_setprop_7pyverbs_2cq_4CQEX_status, 0, 0},
  {"wr_id", __pyx_getprop_7pyverbs_2cq_4CQEX_wr_id, __pyx_setprop_7pyverbs_2cq_4CQEX_wr_id, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyverbs_2cq_CQEX_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyverbs_2cq_CQEX},
  {Py_tp_str, (void *)__pyx_pw_7pyverbs_2cq_4CQEX_43__str__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyverbs_2cq_CQEX},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyverbs_2cq_CQEX},
  {Py_tp_methods, (void *)__pyx_methods_7pyverbs_2cq_CQEX},
  {Py_tp_getset, (void *)__pyx_getsets_7pyverbs_2cq_CQEX},
  {Py_tp_init, (void *)__pyx_pw_7pyverbs_2cq_4CQEX_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7pyverbs_2cq_CQEX},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyverbs_2cq_CQEX_spec = {
  "pyverbs.cq.CQEX",
  sizeof(struct __pyx_obj_7pyverbs_2cq_CQEX),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyverbs_2cq_CQEX_slots,
};
#else

static PyTypeObject __pyx_type_7pyverbs_2cq_CQEX = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyverbs.cq.""CQEX", /*tp_name*/
  sizeof(struct __pyx_obj_7pyverbs_2cq_CQEX), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyverbs_2cq_CQEX, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7pyverbs_2cq_4CQEX_43__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7pyverbs_2cq_CQEX, /*tp_traverse*/
  __pyx_tp_clear_7pyverbs_2cq_CQEX, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyverbs_2cq_CQEX, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7pyverbs_2cq_CQEX, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7pyverbs_2cq_4CQEX_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyverbs_2cq_CQEX, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7pyverbs_2cq_WC(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc_7pyverbs_2cq_WC(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyverbs_2cq_WC) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  if (PyType_IS_GC(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) PyObject_GC_Track(o);
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_dealloc, destructor)(o); else
  #endif
  __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_7pyverbs_2cq_WC);
}

static int __pyx_tp_traverse_7pyverbs_2cq_WC(PyObject *o, visitproc v, void *a) {
  int e;
  #if !CYTHON_USE_MODULE_STATE
  e = 0;
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    traverseproc traverse = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_traverse, traverseproc);
    if (traverse) { e = traverse(o, v, a); }
  } else
  #endif
  { e = __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7pyverbs_2cq_WC); }
  if (e) return e;
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyverbs_2cq_WC(PyObject *o) {
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    inquiry clear = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_clear, inquiry);
    if (clear) clear(o);
  } else
  #endif
  { __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7pyverbs_2cq_WC); }
  return 0;
}

static PyObject *__pyx_getprop_7pyverbs_2cq_2WC_wr_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_2WC_5wr_id_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_2WC_wr_id(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_2WC_5wr_id_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2cq_2WC_status(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_2WC_6status_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_2WC_status(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_2WC_6status_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2cq_2WC_opcode(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_2WC_6opcode_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_2WC_opcode(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_2WC_6opcode_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2cq_2WC_vendor_err(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_2WC_10vendor_err_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_2WC_vendor_err(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_2WC_10vendor_err_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2cq_2WC_byte_len(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_2WC_8byte_len_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_2WC_byte_len(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_2WC_8byte_len_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2cq_2WC_qp_num(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_2WC_6qp_num_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_2WC_qp_num(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_2WC_6qp_num_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2cq_2WC_src_qp(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_2WC_6src_qp_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_2WC_src_qp(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_2WC_6src_qp_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2cq_2WC_wc_flags(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_2WC_8wc_flags_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_2WC_wc_flags(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_2WC_8wc_flags_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2cq_2WC_pkey_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_2WC_10pkey_index_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_2WC_pkey_index(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_2WC_10pkey_index_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2cq_2WC_slid(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_2WC_4slid_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_2WC_slid(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_2WC_4slid_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2cq_2WC_sl(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_2WC_2sl_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_2WC_sl(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_2WC_2sl_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2cq_2WC_imm_data(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_2WC_8imm_data_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_2WC_imm_data(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_2WC_8imm_data_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2cq_2WC_dlid_path_bits(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_2WC_14dlid_path_bits_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_2WC_dlid_path_bits(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_2WC_14dlid_path_bits_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_7pyverbs_2cq_WC[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_2WC_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_2WC_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7pyverbs_2cq_WC[] = {
  {"wr_id", __pyx_getprop_7pyverbs_2cq_2WC_wr_id, __pyx_setprop_7pyverbs_2cq_2WC_wr_id, 0, 0},
  {"status", __pyx_getprop_7pyverbs_2cq_2WC_status, __pyx_setprop_7pyverbs_2cq_2WC_status, 0, 0},
  {"opcode", __pyx_getprop_7pyverbs_2cq_2WC_opcode, __pyx_setprop_7pyverbs_2cq_2WC_opcode, 0, 0},
  {"vendor_err", __pyx_getprop_7pyverbs_2cq_2WC_vendor_err, __pyx_setprop_7pyverbs_2cq_2WC_vendor_err, 0, 0},
  {"byte_len", __pyx_getprop_7pyverbs_2cq_2WC_byte_len, __pyx_setprop_7pyverbs_2cq_2WC_byte_len, 0, 0},
  {"qp_num", __pyx_getprop_7pyverbs_2cq_2WC_qp_num, __pyx_setprop_7pyverbs_2cq_2WC_qp_num, 0, 0},
  {"src_qp", __pyx_getprop_7pyverbs_2cq_2WC_src_qp, __pyx_setprop_7pyverbs_2cq_2WC_src_qp, 0, 0},
  {"wc_flags", __pyx_getprop_7pyverbs_2cq_2WC_wc_flags, __pyx_setprop_7pyverbs_2cq_2WC_wc_flags, 0, 0},
  {"pkey_index", __pyx_getprop_7pyverbs_2cq_2WC_pkey_index, __pyx_setprop_7pyverbs_2cq_2WC_pkey_index, 0, 0},
  {"slid", __pyx_getprop_7pyverbs_2cq_2WC_slid, __pyx_setprop_7pyverbs_2cq_2WC_slid, 0, 0},
  {"sl", __pyx_getprop_7pyverbs_2cq_2WC_sl, __pyx_setprop_7pyverbs_2cq_2WC_sl, 0, 0},
  {"imm_data", __pyx_getprop_7pyverbs_2cq_2WC_imm_data, __pyx_setprop_7pyverbs_2cq_2WC_imm_data, 0, 0},
  {"dlid_path_bits", __pyx_getprop_7pyverbs_2cq_2WC_dlid_path_bits, __pyx_setprop_7pyverbs_2cq_2WC_dlid_path_bits, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyverbs_2cq_WC_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyverbs_2cq_WC},
  {Py_tp_str, (void *)__pyx_pw_7pyverbs_2cq_2WC_3__str__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyverbs_2cq_WC},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyverbs_2cq_WC},
  {Py_tp_methods, (void *)__pyx_methods_7pyverbs_2cq_WC},
  {Py_tp_getset, (void *)__pyx_getsets_7pyverbs_2cq_WC},
  {Py_tp_init, (void *)__pyx_pw_7pyverbs_2cq_2WC_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7pyverbs_2cq_WC},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyverbs_2cq_WC_spec = {
  "pyverbs.cq.WC",
  sizeof(struct __pyx_obj_7pyverbs_2cq_WC),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyverbs_2cq_WC_slots,
};
#else

static PyTypeObject __pyx_type_7pyverbs_2cq_WC = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyverbs.cq.""WC", /*tp_name*/
  sizeof(struct __pyx_obj_7pyverbs_2cq_WC), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyverbs_2cq_WC, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7pyverbs_2cq_2WC_3__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7pyverbs_2cq_WC, /*tp_traverse*/
  __pyx_tp_clear_7pyverbs_2cq_WC, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyverbs_2cq_WC, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7pyverbs_2cq_WC, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7pyverbs_2cq_2WC_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyverbs_2cq_WC, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7pyverbs_2cq_PollCqAttr(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc_7pyverbs_2cq_PollCqAttr(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyverbs_2cq_PollCqAttr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  if (PyType_IS_GC(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) PyObject_GC_Track(o);
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_dealloc, destructor)(o); else
  #endif
  __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_7pyverbs_2cq_PollCqAttr);
}

static int __pyx_tp_traverse_7pyverbs_2cq_PollCqAttr(PyObject *o, visitproc v, void *a) {
  int e;
  #if !CYTHON_USE_MODULE_STATE
  e = 0;
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    traverseproc traverse = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_traverse, traverseproc);
    if (traverse) { e = traverse(o, v, a); }
  } else
  #endif
  { e = __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7pyverbs_2cq_PollCqAttr); }
  if (e) return e;
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyverbs_2cq_PollCqAttr(PyObject *o) {
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    inquiry clear = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_clear, inquiry);
    if (clear) clear(o);
  } else
  #endif
  { __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7pyverbs_2cq_PollCqAttr); }
  return 0;
}

static PyObject *__pyx_getprop_7pyverbs_2cq_10PollCqAttr_comp_mask(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_10PollCqAttr_9comp_mask_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_10PollCqAttr_comp_mask(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_10PollCqAttr_9comp_mask_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_7pyverbs_2cq_PollCqAttr[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_10PollCqAttr_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_10PollCqAttr_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7pyverbs_2cq_PollCqAttr[] = {
  {"comp_mask", __pyx_getprop_7pyverbs_2cq_10PollCqAttr_comp_mask, __pyx_setprop_7pyverbs_2cq_10PollCqAttr_comp_mask, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyverbs_2cq_PollCqAttr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyverbs_2cq_PollCqAttr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyverbs_2cq_PollCqAttr},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyverbs_2cq_PollCqAttr},
  {Py_tp_methods, (void *)__pyx_methods_7pyverbs_2cq_PollCqAttr},
  {Py_tp_getset, (void *)__pyx_getsets_7pyverbs_2cq_PollCqAttr},
  {Py_tp_new, (void *)__pyx_tp_new_7pyverbs_2cq_PollCqAttr},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyverbs_2cq_PollCqAttr_spec = {
  "pyverbs.cq.PollCqAttr",
  sizeof(struct __pyx_obj_7pyverbs_2cq_PollCqAttr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyverbs_2cq_PollCqAttr_slots,
};
#else

static PyTypeObject __pyx_type_7pyverbs_2cq_PollCqAttr = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyverbs.cq.""PollCqAttr", /*tp_name*/
  sizeof(struct __pyx_obj_7pyverbs_2cq_PollCqAttr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyverbs_2cq_PollCqAttr, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7pyverbs_2cq_PollCqAttr, /*tp_traverse*/
  __pyx_tp_clear_7pyverbs_2cq_PollCqAttr, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyverbs_2cq_PollCqAttr, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7pyverbs_2cq_PollCqAttr, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyverbs_2cq_PollCqAttr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7pyverbs_2cq_WcTmInfo(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc_7pyverbs_2cq_WcTmInfo(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyverbs_2cq_WcTmInfo) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  if (PyType_IS_GC(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) PyObject_GC_Track(o);
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_dealloc, destructor)(o); else
  #endif
  __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_7pyverbs_2cq_WcTmInfo);
}

static int __pyx_tp_traverse_7pyverbs_2cq_WcTmInfo(PyObject *o, visitproc v, void *a) {
  int e;
  #if !CYTHON_USE_MODULE_STATE
  e = 0;
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    traverseproc traverse = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_traverse, traverseproc);
    if (traverse) { e = traverse(o, v, a); }
  } else
  #endif
  { e = __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7pyverbs_2cq_WcTmInfo); }
  if (e) return e;
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyverbs_2cq_WcTmInfo(PyObject *o) {
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    inquiry clear = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_clear, inquiry);
    if (clear) clear(o);
  } else
  #endif
  { __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7pyverbs_2cq_WcTmInfo); }
  return 0;
}

static PyObject *__pyx_getprop_7pyverbs_2cq_8WcTmInfo_tag(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_8WcTmInfo_3tag_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_8WcTmInfo_tag(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_8WcTmInfo_3tag_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_2cq_8WcTmInfo_priv(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_2cq_8WcTmInfo_4priv_1__get__(o);
}

static int __pyx_setprop_7pyverbs_2cq_8WcTmInfo_priv(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_2cq_8WcTmInfo_4priv_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_7pyverbs_2cq_WcTmInfo[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_8WcTmInfo_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_2cq_8WcTmInfo_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7pyverbs_2cq_WcTmInfo[] = {
  {"tag", __pyx_getprop_7pyverbs_2cq_8WcTmInfo_tag, __pyx_setprop_7pyverbs_2cq_8WcTmInfo_tag, 0, 0},
  {"priv", __pyx_getprop_7pyverbs_2cq_8WcTmInfo_priv, __pyx_setprop_7pyverbs_2cq_8WcTmInfo_priv, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyverbs_2cq_WcTmInfo_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyverbs_2cq_WcTmInfo},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyverbs_2cq_WcTmInfo},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyverbs_2cq_WcTmInfo},
  {Py_tp_methods, (void *)__pyx_methods_7pyverbs_2cq_WcTmInfo},
  {Py_tp_getset, (void *)__pyx_getsets_7pyverbs_2cq_WcTmInfo},
  {Py_tp_new, (void *)__pyx_tp_new_7pyverbs_2cq_WcTmInfo},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyverbs_2cq_WcTmInfo_spec = {
  "pyverbs.cq.WcTmInfo",
  sizeof(struct __pyx_obj_7pyverbs_2cq_WcTmInfo),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyverbs_2cq_WcTmInfo_slots,
};
#else

static PyTypeObject __pyx_type_7pyverbs_2cq_WcTmInfo = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyverbs.cq.""WcTmInfo", /*tp_name*/
  sizeof(struct __pyx_obj_7pyverbs_2cq_WcTmInfo), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyverbs_2cq_WcTmInfo, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7pyverbs_2cq_WcTmInfo, /*tp_traverse*/
  __pyx_tp_clear_7pyverbs_2cq_WcTmInfo, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyverbs_2cq_WcTmInfo, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7pyverbs_2cq_WcTmInfo, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyverbs_2cq_WcTmInfo, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
/* #### Code section: initfunc_declarations ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate); /*proto*/
/* #### Code section: init_module ### */

static int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py = Py_None; Py_INCREF(Py_None);
  __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py = Py_None; Py_INCREF(Py_None);
  __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_t_1 = PyImport_ImportModule("pyverbs.base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_4base_PyverbsCM = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.base", "PyverbsCM",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4base_PyverbsCM), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4base_PyverbsCM),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4base_PyverbsCM), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4base_PyverbsCM),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4base_PyverbsCM), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4base_PyverbsCM),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4base_PyverbsCM) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_4base_PyverbsCM = (struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_4base_PyverbsCM); if (unlikely(!__pyx_vtabptr_7pyverbs_4base_PyverbsCM)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2cq_CompChannel = &__pyx_vtable_7pyverbs_2cq_CompChannel;
  __pyx_vtable_7pyverbs_2cq_CompChannel.__pyx_base = *__pyx_vtabptr_7pyverbs_4base_PyverbsCM;
  __pyx_vtable_7pyverbs_2cq_CompChannel.__pyx_base.close = (PyObject *(*)(struct __pyx_obj_7pyverbs_4base_PyverbsCM *, int __pyx_skip_dispatch))__pyx_f_7pyverbs_2cq_11CompChannel_close;
  __pyx_vtable_7pyverbs_2cq_CompChannel.add_ref = (PyObject *(*)(struct __pyx_obj_7pyverbs_2cq_CompChannel *, PyObject *))__pyx_f_7pyverbs_2cq_11CompChannel_add_ref;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CompChannel = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyverbs_2cq_CompChannel_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CompChannel)) __PYX_ERR(0, 17, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyverbs_2cq_CompChannel_spec, __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CompChannel) < (0)) __PYX_ERR(0, 17, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CompChannel = &__pyx_type_7pyverbs_2cq_CompChannel;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CompChannel->tp_base = __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CompChannel) < (0)) __PYX_ERR(0, 17, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CompChannel);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CompChannel->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CompChannel->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CompChannel->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CompChannel, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 17, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_7pyverbs_2cq_11CompChannel___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_7pyverbs_2cq_11CompChannel___init__.doc = __pyx_doc_7pyverbs_2cq_11CompChannel___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_7pyverbs_2cq_11CompChannel___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CompChannel, __pyx_vtabptr_7pyverbs_2cq_CompChannel) < (0)) __PYX_ERR(0, 17, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CompChannel) < (0)) __PYX_ERR(0, 17, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_CompChannel, (PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CompChannel) < (0)) __PYX_ERR(0, 17, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SLOTS
  if (__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CompChannel->tp_weaklistoffset == 0) __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CompChannel->tp_weaklistoffset = offsetof(struct __pyx_obj_7pyverbs_2cq_CompChannel, __pyx_base.__pyx_base.__weakref__);
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CompChannel) < (0)) __PYX_ERR(0, 17, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2cq_CQ = &__pyx_vtable_7pyverbs_2cq_CQ;
  __pyx_vtable_7pyverbs_2cq_CQ.__pyx_base = *__pyx_vtabptr_7pyverbs_4base_PyverbsCM;
  __pyx_vtable_7pyverbs_2cq_CQ.__pyx_base.close = (PyObject *(*)(struct __pyx_obj_7pyverbs_4base_PyverbsCM *, int __pyx_skip_dispatch))__pyx_f_7pyverbs_2cq_2CQ_close;
  __pyx_vtable_7pyverbs_2cq_CQ.add_ref = (PyObject *(*)(struct __pyx_obj_7pyverbs_2cq_CQ *, PyObject *))__pyx_f_7pyverbs_2cq_2CQ_add_ref;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQ = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyverbs_2cq_CQ_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQ)) __PYX_ERR(0, 78, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyverbs_2cq_CQ_spec, __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQ) < (0)) __PYX_ERR(0, 78, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQ = &__pyx_type_7pyverbs_2cq_CQ;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQ->tp_base = __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQ) < (0)) __PYX_ERR(0, 78, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQ);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQ->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQ->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQ->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQ, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 78, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_7pyverbs_2cq_2CQ___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_7pyverbs_2cq_2CQ___init__.doc = __pyx_doc_7pyverbs_2cq_2CQ___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_7pyverbs_2cq_2CQ___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQ, __pyx_vtabptr_7pyverbs_2cq_CQ) < (0)) __PYX_ERR(0, 78, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQ) < (0)) __PYX_ERR(0, 78, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_CQ_2, (PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQ) < (0)) __PYX_ERR(0, 78, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SLOTS
  if (__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQ->tp_weaklistoffset == 0) __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQ->tp_weaklistoffset = offsetof(struct __pyx_obj_7pyverbs_2cq_CQ, __pyx_base.__pyx_base.__weakref__);
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQ) < (0)) __PYX_ERR(0, 78, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_4base_PyverbsObject = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.base", "PyverbsObject",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4base_PyverbsObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4base_PyverbsObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4base_PyverbsObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4base_PyverbsObject),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4base_PyverbsObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4base_PyverbsObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4base_PyverbsObject) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyverbs_2cq_CqInitAttrEx_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx)) __PYX_ERR(0, 221, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyverbs_2cq_CqInitAttrEx_spec, __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx) < (0)) __PYX_ERR(0, 221, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx = &__pyx_type_7pyverbs_2cq_CqInitAttrEx;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx->tp_base = __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx) < (0)) __PYX_ERR(0, 221, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 221, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_7pyverbs_2cq_12CqInitAttrEx___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_7pyverbs_2cq_12CqInitAttrEx___init__.doc = __pyx_doc_7pyverbs_2cq_12CqInitAttrEx___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_7pyverbs_2cq_12CqInitAttrEx___init__;
    }
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_CqInitAttrEx, (PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx) < (0)) __PYX_ERR(0, 221, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SLOTS
  if (__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx->tp_weaklistoffset == 0) __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx->tp_weaklistoffset = offsetof(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx, __pyx_base.__weakref__);
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx) < (0)) __PYX_ERR(0, 221, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2cq_CQEX = &__pyx_vtable_7pyverbs_2cq_CQEX;
  __pyx_vtable_7pyverbs_2cq_CQEX.__pyx_base = *__pyx_vtabptr_7pyverbs_4base_PyverbsCM;
  __pyx_vtable_7pyverbs_2cq_CQEX.__pyx_base.close = (PyObject *(*)(struct __pyx_obj_7pyverbs_4base_PyverbsCM *, int __pyx_skip_dispatch))__pyx_f_7pyverbs_2cq_4CQEX_close;
  __pyx_vtable_7pyverbs_2cq_CQEX.add_ref = (PyObject *(*)(struct __pyx_obj_7pyverbs_2cq_CQEX *, PyObject *))__pyx_f_7pyverbs_2cq_4CQEX_add_ref;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQEX = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyverbs_2cq_CQEX_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQEX)) __PYX_ERR(0, 315, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyverbs_2cq_CQEX_spec, __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQEX) < (0)) __PYX_ERR(0, 315, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQEX = &__pyx_type_7pyverbs_2cq_CQEX;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX->tp_base = __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQEX) < (0)) __PYX_ERR(0, 315, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQEX);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQEX->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQEX->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQEX->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQEX, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 315, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_7pyverbs_2cq_4CQEX___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_7pyverbs_2cq_4CQEX___init__.doc = __pyx_doc_7pyverbs_2cq_4CQEX___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_7pyverbs_2cq_4CQEX___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQEX, __pyx_vtabptr_7pyverbs_2cq_CQEX) < (0)) __PYX_ERR(0, 315, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQEX) < (0)) __PYX_ERR(0, 315, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_CQEX, (PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQEX) < (0)) __PYX_ERR(0, 315, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SLOTS
  if (__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQEX->tp_weaklistoffset == 0) __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQEX->tp_weaklistoffset = offsetof(struct __pyx_obj_7pyverbs_2cq_CQEX, __pyx_base.__pyx_base.__weakref__);
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQEX) < (0)) __PYX_ERR(0, 315, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_WC = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyverbs_2cq_WC_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_WC)) __PYX_ERR(0, 446, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyverbs_2cq_WC_spec, __pyx_mstate->__pyx_ptype_7pyverbs_2cq_WC) < (0)) __PYX_ERR(0, 446, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_WC = &__pyx_type_7pyverbs_2cq_WC;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_WC->tp_base = __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyverbs_2cq_WC) < (0)) __PYX_ERR(0, 446, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7pyverbs_2cq_WC);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_WC->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyverbs_2cq_WC->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyverbs_2cq_WC->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_WC, (PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2cq_WC) < (0)) __PYX_ERR(0, 446, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SLOTS
  if (__pyx_mstate->__pyx_ptype_7pyverbs_2cq_WC->tp_weaklistoffset == 0) __pyx_mstate->__pyx_ptype_7pyverbs_2cq_WC->tp_weaklistoffset = offsetof(struct __pyx_obj_7pyverbs_2cq_WC, __pyx_base.__weakref__);
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2cq_WC) < (0)) __PYX_ERR(0, 446, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_PollCqAttr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyverbs_2cq_PollCqAttr_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_PollCqAttr)) __PYX_ERR(0, 573, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyverbs_2cq_PollCqAttr_spec, __pyx_mstate->__pyx_ptype_7pyverbs_2cq_PollCqAttr) < (0)) __PYX_ERR(0, 573, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_PollCqAttr = &__pyx_type_7pyverbs_2cq_PollCqAttr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_PollCqAttr->tp_base = __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyverbs_2cq_PollCqAttr) < (0)) __PYX_ERR(0, 573, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7pyverbs_2cq_PollCqAttr);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_PollCqAttr->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyverbs_2cq_PollCqAttr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyverbs_2cq_PollCqAttr->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_PollCqAttr, (PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2cq_PollCqAttr) < (0)) __PYX_ERR(0, 573, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SLOTS
  if (__pyx_mstate->__pyx_ptype_7pyverbs_2cq_PollCqAttr->tp_weaklistoffset == 0) __pyx_mstate->__pyx_ptype_7pyverbs_2cq_PollCqAttr->tp_weaklistoffset = offsetof(struct __pyx_obj_7pyverbs_2cq_PollCqAttr, __pyx_base.__weakref__);
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2cq_PollCqAttr) < (0)) __PYX_ERR(0, 573, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_WcTmInfo = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyverbs_2cq_WcTmInfo_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_WcTmInfo)) __PYX_ERR(0, 582, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyverbs_2cq_WcTmInfo_spec, __pyx_mstate->__pyx_ptype_7pyverbs_2cq_WcTmInfo) < (0)) __PYX_ERR(0, 582, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_WcTmInfo = &__pyx_type_7pyverbs_2cq_WcTmInfo;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_WcTmInfo->tp_base = __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyverbs_2cq_WcTmInfo) < (0)) __PYX_ERR(0, 582, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7pyverbs_2cq_WcTmInfo);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_WcTmInfo->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyverbs_2cq_WcTmInfo->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyverbs_2cq_WcTmInfo->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_WcTmInfo, (PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2cq_WcTmInfo) < (0)) __PYX_ERR(0, 582, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SLOTS
  if (__pyx_mstate->__pyx_ptype_7pyverbs_2cq_WcTmInfo->tp_weaklistoffset == 0) __pyx_mstate->__pyx_ptype_7pyverbs_2cq_WcTmInfo->tp_weaklistoffset = offsetof(struct __pyx_obj_7pyverbs_2cq_WcTmInfo, __pyx_base.__weakref__);
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_2cq_WcTmInfo) < (0)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("pyverbs.device"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_Context = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "Context",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_Context), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_Context),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_Context), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_Context),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_Context), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_Context),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_Context) __PYX_ERR(2, 10, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_6device_Context = (struct __pyx_vtabstruct_7pyverbs_6device_Context*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_6device_Context); if (unlikely(!__pyx_vtabptr_7pyverbs_6device_Context)) __PYX_ERR(2, 10, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_DeviceAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "DeviceAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_DeviceAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DeviceAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_DeviceAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DeviceAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_DeviceAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DeviceAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_DeviceAttr) __PYX_ERR(2, 33, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_QueryDeviceExInput = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "QueryDeviceExInput",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_QueryDeviceExInput) __PYX_ERR(2, 36, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_ODPCaps = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "ODPCaps",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_ODPCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_ODPCaps),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_ODPCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_ODPCaps),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_ODPCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_ODPCaps),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_ODPCaps) __PYX_ERR(2, 39, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_RSSCaps = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "RSSCaps",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_RSSCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_RSSCaps),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_RSSCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_RSSCaps),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_RSSCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_RSSCaps),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_RSSCaps) __PYX_ERR(2, 43, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_PacketPacingCaps = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "PacketPacingCaps",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_PacketPacingCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PacketPacingCaps),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_PacketPacingCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PacketPacingCaps),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_PacketPacingCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PacketPacingCaps),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_PacketPacingCaps) __PYX_ERR(2, 46, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_PCIAtomicCaps = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "PCIAtomicCaps",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_PCIAtomicCaps) __PYX_ERR(2, 49, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_TMCaps = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "TMCaps",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_TMCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_TMCaps),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_TMCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_TMCaps),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_TMCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_TMCaps),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_TMCaps) __PYX_ERR(2, 52, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_CQModerationCaps = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "CQModerationCaps",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_CQModerationCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_CQModerationCaps),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_CQModerationCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_CQModerationCaps),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_CQModerationCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_CQModerationCaps),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_CQModerationCaps) __PYX_ERR(2, 55, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_TSOCaps = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "TSOCaps",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_TSOCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_TSOCaps),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_TSOCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_TSOCaps),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_TSOCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_TSOCaps),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_TSOCaps) __PYX_ERR(2, 58, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_DeviceAttrEx = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "DeviceAttrEx",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_DeviceAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DeviceAttrEx),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_DeviceAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DeviceAttrEx),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_DeviceAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DeviceAttrEx),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_DeviceAttrEx) __PYX_ERR(2, 61, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_AllocDmAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "AllocDmAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_AllocDmAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_AllocDmAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_AllocDmAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_AllocDmAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_AllocDmAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_AllocDmAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_AllocDmAttr) __PYX_ERR(2, 64, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_DM = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "DM",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_DM), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DM),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_DM), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DM),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_DM), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DM),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_DM) __PYX_ERR(2, 67, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_6device_DM = (struct __pyx_vtabstruct_7pyverbs_6device_DM*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_6device_DM); if (unlikely(!__pyx_vtabptr_7pyverbs_6device_DM)) __PYX_ERR(2, 67, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_PortAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "PortAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_PortAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PortAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_PortAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PortAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_PortAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PortAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_PortAttr) __PYX_ERR(2, 74, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_GIDEntry = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "GIDEntry",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_GIDEntry), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_GIDEntry),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_GIDEntry), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_GIDEntry),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_GIDEntry), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_GIDEntry),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_GIDEntry) __PYX_ERR(2, 77, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_AsyncEvent = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "AsyncEvent",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_AsyncEvent), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_AsyncEvent),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_AsyncEvent), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_AsyncEvent),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_AsyncEvent), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_AsyncEvent),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_AsyncEvent) __PYX_ERR(2, 80, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_FdArr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "FdArr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_FdArr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_FdArr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_FdArr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_FdArr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_FdArr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_FdArr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_FdArr) __PYX_ERR(2, 83, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_6device_FdArr = (struct __pyx_vtabstruct_7pyverbs_6device_FdArr*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_6device_FdArr); if (unlikely(!__pyx_vtabptr_7pyverbs_6device_FdArr)) __PYX_ERR(2, 83, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyverbs.pd"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_2pd_PD = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.pd", "PD",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2pd_PD), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_PD),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2pd_PD), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_PD),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2pd_PD), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_PD),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2pd_PD) __PYX_ERR(3, 12, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2pd_PD = (struct __pyx_vtabstruct_7pyverbs_2pd_PD*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2pd_PD); if (unlikely(!__pyx_vtabptr_7pyverbs_2pd_PD)) __PYX_ERR(3, 12, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2pd_ParentDomainInitAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.pd", "ParentDomainInitAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2pd_ParentDomainInitAttr) __PYX_ERR(3, 28, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2pd_ParentDomain = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.pd", "ParentDomain",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomain), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomain),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomain), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomain),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomain), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomain),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2pd_ParentDomain) __PYX_ERR(3, 34, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2pd_ParentDomain = (struct __pyx_vtabstruct_7pyverbs_2pd_ParentDomain*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2pd_ParentDomain); if (unlikely(!__pyx_vtabptr_7pyverbs_2pd_ParentDomain)) __PYX_ERR(3, 34, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2pd_ParentDomainContext = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.pd", "ParentDomainContext",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomainContext), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomainContext),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomainContext), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomainContext),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomainContext), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomainContext),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2pd_ParentDomainContext) __PYX_ERR(3, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyverbs.srq"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_3srq_SrqAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.srq", "SrqAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_3srq_SrqAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SrqAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_3srq_SrqAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SrqAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_3srq_SrqAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SrqAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_3srq_SrqAttr) __PYX_ERR(4, 9, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_3srq_SrqInitAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.srq", "SrqInitAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_3srq_SrqInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SrqInitAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_3srq_SrqInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SrqInitAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_3srq_SrqInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SrqInitAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_3srq_SrqInitAttr) __PYX_ERR(4, 12, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_3srq_SrqInitAttrEx = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.srq", "SrqInitAttrEx",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_3srq_SrqInitAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SrqInitAttrEx),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_3srq_SrqInitAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SrqInitAttrEx),
  #else
  sizeof(struct __pyx_obj_7pyverbs_3srq_SrqInitAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SrqInitAttrEx),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_3srq_SrqInitAttrEx) __PYX_ERR(4, 15, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_3srq_OpsWr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.srq", "OpsWr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_3srq_OpsWr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_OpsWr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_3srq_OpsWr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_OpsWr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_3srq_OpsWr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_OpsWr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_3srq_OpsWr) __PYX_ERR(4, 21, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_3srq_OpsWr = (struct __pyx_vtabstruct_7pyverbs_3srq_OpsWr*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_3srq_OpsWr); if (unlikely(!__pyx_vtabptr_7pyverbs_3srq_OpsWr)) __PYX_ERR(4, 21, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_3srq_SRQ = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.srq", "SRQ",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_3srq_SRQ), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SRQ),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_3srq_SRQ), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SRQ),
  #else
  sizeof(struct __pyx_obj_7pyverbs_3srq_SRQ), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_3srq_SRQ),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_3srq_SRQ) __PYX_ERR(4, 24, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_3srq_SRQ = (struct __pyx_vtabstruct_7pyverbs_3srq_SRQ*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_3srq_SRQ); if (unlikely(!__pyx_vtabptr_7pyverbs_3srq_SRQ)) __PYX_ERR(4, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyverbs.qp"); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPCap = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.qp", "QPCap",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPCap), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPCap),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPCap), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPCap),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPCap), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPCap),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPCap) __PYX_ERR(5, 9, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.qp", "QPInitAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPInitAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPInitAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPInitAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttr) __PYX_ERR(5, 12, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.qp", "QPInitAttrEx",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx) __PYX_ERR(5, 18, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.qp", "QPAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPAttr) __PYX_ERR(5, 27, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QP = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.qp", "QP",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2qp_QP), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QP),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2qp_QP), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QP),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2qp_QP), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QP),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QP) __PYX_ERR(5, 30, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2qp_QP = (struct __pyx_vtabstruct_7pyverbs_2qp_QP*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QP); if (unlikely(!__pyx_vtabptr_7pyverbs_2qp_QP)) __PYX_ERR(5, 30, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_DataBuffer = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.qp", "DataBuffer",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2qp_DataBuffer), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_DataBuffer),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2qp_DataBuffer), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_DataBuffer),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2qp_DataBuffer), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_DataBuffer),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_DataBuffer) __PYX_ERR(5, 47, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2qp_DataBuffer = (struct __pyx_vtabstruct_7pyverbs_2qp_DataBuffer*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_DataBuffer); if (unlikely(!__pyx_vtabptr_7pyverbs_2qp_DataBuffer)) __PYX_ERR(5, 47, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPEx = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.qp", "QPEx",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPEx),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPEx),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPEx),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPEx) __PYX_ERR(5, 50, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2qp_QPEx = (struct __pyx_vtabstruct_7pyverbs_2qp_QPEx*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPEx); if (unlikely(!__pyx_vtabptr_7pyverbs_2qp_QPEx)) __PYX_ERR(5, 50, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_ECE = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.qp", "ECE",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2qp_ECE), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_ECE),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2qp_ECE), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_ECE),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2qp_ECE), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_ECE),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_ECE) __PYX_ERR(5, 54, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2qp_ECE = (struct __pyx_vtabstruct_7pyverbs_2qp_ECE*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_ECE); if (unlikely(!__pyx_vtabptr_7pyverbs_2qp_ECE)) __PYX_ERR(5, 54, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyverbs.wq"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_2wq_WQInitAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.wq", "WQInitAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2wq_WQInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_WQInitAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2wq_WQInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_WQInitAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2wq_WQInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_WQInitAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2wq_WQInitAttr) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2wq_WQAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.wq", "WQAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2wq_WQAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_WQAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2wq_WQAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_WQAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2wq_WQAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_WQAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2wq_WQAttr) __PYX_ERR(6, 18, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2wq_WQ = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.wq", "WQ",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2wq_WQ), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_WQ),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2wq_WQ), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_WQ),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2wq_WQ), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_WQ),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2wq_WQ) __PYX_ERR(6, 21, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2wq_WQ = (struct __pyx_vtabstruct_7pyverbs_2wq_WQ*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2wq_WQ); if (unlikely(!__pyx_vtabptr_7pyverbs_2wq_WQ)) __PYX_ERR(6, 21, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2wq_RwqIndTableInitAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.wq", "RwqIndTableInitAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2wq_RwqIndTableInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_RwqIndTableInitAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2wq_RwqIndTableInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_RwqIndTableInitAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2wq_RwqIndTableInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_RwqIndTableInitAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2wq_RwqIndTableInitAttr) __PYX_ERR(6, 29, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2wq_RwqIndTable = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.wq", "RwqIndTable",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2wq_RwqIndTable), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_RwqIndTable),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2wq_RwqIndTable), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_RwqIndTable),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2wq_RwqIndTable), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_RwqIndTable),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2wq_RwqIndTable) __PYX_ERR(6, 33, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2wq_RwqIndTable = (struct __pyx_vtabstruct_7pyverbs_2wq_RwqIndTable*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2wq_RwqIndTable); if (unlikely(!__pyx_vtabptr_7pyverbs_2wq_RwqIndTable)) __PYX_ERR(6, 33, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2wq_RxHashConf = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.wq", "RxHashConf",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2wq_RxHashConf), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_RxHashConf),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2wq_RxHashConf), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_RxHashConf),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2wq_RxHashConf), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2wq_RxHashConf),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2wq_RxHashConf) __PYX_ERR(6, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  {
    __pyx_t_1 = PyImport_ImportModule("pyverbs.base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    const char * __pyx_import_signature = __Pyx_PyBytes_AsString(__pyx_mstate_global->__pyx_kp_b_PyObject_PyObject_close_weakrefs);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (unlikely(!__pyx_import_signature)) __PYX_ERR(0, 1, __pyx_L1_error)
    #endif
    const char * __pyx_import_name = __pyx_import_signature + 23;
    void (**const __pyx_import_pointers[])(void) = {(void (**)(void))&__pyx_f_7pyverbs_4base_close_weakrefs, (void (**)(void)) NULL};
    void (**const *__pyx_import_pointer)(void) = __pyx_import_pointers;
    const char *__pyx_import_current_signature = __pyx_import_signature;
    while (*__pyx_import_pointer) {
      if (__Pyx_ImportFunction_3_2_4(__pyx_t_1, __pyx_import_name, *__pyx_import_pointer, __pyx_import_current_signature) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
      ++__pyx_import_pointer;
      __pyx_import_name = strchr(__pyx_import_name, '\0') + 1;
      __pyx_import_signature = strchr(__pyx_import_signature, '\0') + 1;
      if (*__pyx_import_signature != '\0') __pyx_import_current_signature = __pyx_import_signature;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_cq(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_cq},
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  {Py_mod_gil, __Pyx_FREETHREADING_COMPATIBLE},
  #endif
  #if PY_VERSION_HEX >= 0x030C0000 && CYTHON_USE_MODULE_STATE
  {Py_mod_multiple_interpreters, Py_MOD_MULTIPLE_INTERPRETERS_NOT_SUPPORTED},
  #endif
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "cq",
      0, /* m_doc */
    #if CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstatetype), /* m_size */
    #else
      (CYTHON_PEP489_MULTI_PHASE_INIT) ? 0 : -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif

/* PyModInitFuncType */
#ifndef CYTHON_NO_PYINIT_EXPORT
  #define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#else
  #ifdef __cplusplus
  #define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
  #else
  #define __Pyx_PyMODINIT_FUNC PyObject *
  #endif
#endif

__Pyx_PyMODINIT_FUNC PyInit_cq(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_cq(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
/* ModuleCreationPEP489 */
#if CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
static PY_INT64_T __Pyx_GetCurrentInterpreterId(void) {
    {
        PyObject *module = PyImport_ImportModule("_interpreters"); // 3.13+ I think
        if (!module) {
            PyErr_Clear(); // just try the 3.8-3.12 version
            module = PyImport_ImportModule("_xxsubinterpreters");
            if (!module) goto bad;
        }
        PyObject *current = PyObject_CallMethod(module, "get_current", NULL);
        Py_DECREF(module);
        if (!current) goto bad;
        if (PyTuple_Check(current)) {
            PyObject *new_current = PySequence_GetItem(current, 0);
            Py_DECREF(current);
            current = new_current;
            if (!new_current) goto bad;
        }
        long long as_c_int = PyLong_AsLongLong(current);
        Py_DECREF(current);
        return as_c_int;
    }
  bad:
    PySys_WriteStderr("__Pyx_GetCurrentInterpreterId failed. Try setting the C define CYTHON_PEP489_MULTI_PHASE_INIT=0\n");
    return -1;
}
#endif
#if !CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    static PY_INT64_T main_interpreter_id = -1;
#if CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
    PY_INT64_T current_id = GraalPyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_GRAAL
    PY_INT64_T current_id = PyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
    PY_INT64_T current_id = __Pyx_GetCurrentInterpreterId();
#elif CYTHON_COMPILING_IN_LIMITED_API
    PY_INT64_T current_id = PyInterpreterState_GetID(PyInterpreterState_Get());
#else
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
#endif
    if (unlikely(current_id == -1)) {
        return -1;
    }
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return 0;
    } else if (unlikely(main_interpreter_id != current_id)) {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#endif
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    #if !CYTHON_USE_MODULE_STATE
    if (__Pyx_check_single_interpreter())
        return NULL;
    #endif
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_cq(PyObject *__pyx_pyinit_module)
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  __pyx_mstatetype *__pyx_mstate = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'cq' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #else
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_t_1 = __pyx_pyinit_module;
  Py_INCREF(__pyx_t_1);
  #else
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #if CYTHON_USE_MODULE_STATE
  {
    int add_module_result = __Pyx_State_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "cq" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = __pyx_t_1;
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  PyUnstable_Module_SetGIL(__pyx_m, Py_MOD_GIL_USED);
  #endif
  __pyx_mstate = __pyx_mstate_global;
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_mstate->__pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_mstate->__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_mstate->__pyx_d);
  __pyx_mstate->__pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_mstate->__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_cython_runtime = __Pyx_PyImport_AddModuleRef("cython_runtime"); if (unlikely(!__pyx_mstate->__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_mstate->__pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /* ImportRefnannyAPI */
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
    PyErr_Clear();
    __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
    if (!__Pyx_RefNanny)
        Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  
__Pyx_RefNannySetupContext("PyInit_cq", 0);
  __Pyx_init_runtime_version();
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_mstate->__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Library function declarations ---*/
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__pyx_module_is_main_pyverbs__cq) {
    if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_name, __pyx_mstate_global->__pyx_n_u_main) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "pyverbs.cq")) {
      if (unlikely((PyDict_SetItemString(modules, "pyverbs.cq", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_CreateCodeObjects(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_export_code(__pyx_mstate);
  (void)__Pyx_modinit_function_export_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_type_init_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_function_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/

  /* "EnumTypeToPy":1
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py = None             # <<<<<<<<<<<<<<
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py")
*/
  __Pyx_INCREF(Py_None);
  __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py);
  __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py, Py_None);
  __Pyx_GIVEREF(Py_None);

  /* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(ibv_wc_status c_val):
 *     cdef object __pyx_enum
*/
  __Pyx_INCREF(Py_None);
  __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py);
  __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py, Py_None);
  __Pyx_GIVEREF(Py_None);

  /* "EnumTypeToPy":1
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py = None             # <<<<<<<<<<<<<<
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py")
*/
  __Pyx_INCREF(Py_None);
  __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py);
  __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_create_cq_wc_flags_to_py, Py_None);
  __Pyx_GIVEREF(Py_None);

  /* "pyverbs/cq.pyx":3
 * # SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)
 * # Copyright (c) 2019, Mellanox Technologies. All rights reserved.
 * import weakref             # <<<<<<<<<<<<<<
 * from libc.stdint cimport uintptr_t
 * 
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_weakref, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_weakref, __pyx_t_2) < (0)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":6
 * from libc.stdint cimport uintptr_t
 * 
 * from pyverbs.pyverbs_error import PyverbsError, PyverbsRDMAError             # <<<<<<<<<<<<<<
 * from pyverbs.base import PyverbsRDMAErrno
 * from pyverbs.pd cimport PD, ParentDomain
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_PyverbsError,__pyx_mstate_global->__pyx_n_u_PyverbsRDMAError};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyverbs_pyverbs_error, __pyx_imported_names, 2, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_PyverbsError,__pyx_mstate_global->__pyx_n_u_PyverbsRDMAError};
    for (__pyx_t_3=0; __pyx_t_3 < 2; __pyx_t_3++) {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 6, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":7
 * 
 * from pyverbs.pyverbs_error import PyverbsError, PyverbsRDMAError
 * from pyverbs.base import PyverbsRDMAErrno             # <<<<<<<<<<<<<<
 * from pyverbs.pd cimport PD, ParentDomain
 * from pyverbs.base cimport close_weakrefs
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyverbs_base, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno};
    __pyx_t_3 = 0; {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 7, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 7, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":42
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.cc != NULL:
 *             if self.logger:
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_11CompChannel_5close, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CompChannel_close, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CompChannel, __pyx_mstate_global->__pyx_n_u_close, __pyx_t_2) < (0)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":53
 *             self.cc = NULL
 * 
 *     def get_cq_event(self, CQ expected_cq):             # <<<<<<<<<<<<<<
 *         """
 *         Waits for the next completion event in the completion event channel
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_11CompChannel_7get_cq_event, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CompChannel_get_cq_event, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[1])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CompChannel, __pyx_mstate_global->__pyx_n_u_get_cq_event, __pyx_t_2) < (0)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.cc cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_11CompChannel_9__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CompChannel___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[2])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.cc cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.cc cannot be converted to a Python object for pickling"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_11CompChannel_11__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CompChannel___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[3])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":129
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.cq != NULL:
 *             if self.logger:
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_2CQ_5close, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQ_close, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[4])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQ, __pyx_mstate_global->__pyx_n_u_close, __pyx_t_2) < (0)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":143
 *             self.channel = None
 * 
 *     def resize(self, cqe):             # <<<<<<<<<<<<<<
 *         """
 *         Resizes the completion queue (CQ) to have at least cqe entries.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_2CQ_7resize, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQ_resize, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[5])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQ, __pyx_mstate_global->__pyx_n_u_resize, __pyx_t_2) < (0)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":153
 *             raise PyverbsRDMAError('Failed to resize CQ', rc)
 * 
 *     def poll(self, num_entries=1):             # <<<<<<<<<<<<<<
 *         """
 *         Polls the CQ for completions.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_2CQ_9poll, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQ_poll, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[6])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[0]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQ, __pyx_mstate_global->__pyx_n_u_poll, __pyx_t_2) < (0)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":179
 *         return npolled, wcs
 * 
 *     def req_notify(self, solicited_only = False):             # <<<<<<<<<<<<<<
 *         """
 *         Request completion notification on the completion queue.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_2CQ_11req_notify, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQ_req_notify, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[7])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[1]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQ, __pyx_mstate_global->__pyx_n_u_req_notify, __pyx_t_2) < (0)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":193
 *                                    format(rc=rc), rc)
 * 
 *     def ack_events(self, num_events):             # <<<<<<<<<<<<<<
 *         """
 *         Get and acknowledge CQ events
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_2CQ_13ack_events, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQ_ack_events, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[8])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQ, __pyx_mstate_global->__pyx_n_u_ack_events, __pyx_t_2) < (0)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.cq cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_2CQ_17__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQ___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[9])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.cq cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.cq cannot be converted to a Python object for pickling"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_2CQ_19__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQ___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[10])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_12CqInitAttrEx_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CqInitAttrEx___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[11])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.attr cannot be converted to a Python object for pickling"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_12CqInitAttrEx_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CqInitAttrEx___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[12])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":357
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.cq != NULL:
 *             if self.logger:
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_4CQEX_5close, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQEX_close, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[13])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX, __pyx_mstate_global->__pyx_n_u_close, __pyx_t_2) < (0)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":368
 *             self.context = None
 * 
 *     def start_poll(self, PollCqAttr attr):             # <<<<<<<<<<<<<<
 *         """
 *         Start polling a batch of work completions.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_4CQEX_7start_poll, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQEX_start_poll, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[14])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX, __pyx_mstate_global->__pyx_n_u_start_poll, __pyx_t_2) < (0)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":378
 *         return v.ibv_start_poll(self.cq, &attr.attr)
 * 
 *     def poll_next(self):             # <<<<<<<<<<<<<<
 *         """
 *         Get the next work completion.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_4CQEX_9poll_next, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQEX_poll_next, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[15])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX, __pyx_mstate_global->__pyx_n_u_poll_next, __pyx_t_2) < (0)) __PYX_ERR(0, 378, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":385
 *         return v.ibv_next_poll(self.cq)
 * 
 *     def end_poll(self):             # <<<<<<<<<<<<<<
 *         """
 *         Indicates the end of polling batch of work completions
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_4CQEX_11end_poll, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQEX_end_poll, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[16])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX, __pyx_mstate_global->__pyx_n_u_end_poll, __pyx_t_2) < (0)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":392
 *         return v.ibv_end_poll(self.cq)
 * 
 *     def read_opcode(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_opcode(self.cq)
 *     def read_vendor_err(self):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_4CQEX_13read_opcode, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQEX_read_opcode, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[17])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX, __pyx_mstate_global->__pyx_n_u_read_opcode, __pyx_t_2) < (0)) __PYX_ERR(0, 392, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":394
 *     def read_opcode(self):
 *         return v.ibv_wc_read_opcode(self.cq)
 *     def read_vendor_err(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_vendor_err(self.cq)
 *     def read_byte_len(self):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_4CQEX_15read_vendor_err, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQEX_read_vendor_err, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[18])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX, __pyx_mstate_global->__pyx_n_u_read_vendor_err, __pyx_t_2) < (0)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":396
 *     def read_vendor_err(self):
 *         return v.ibv_wc_read_vendor_err(self.cq)
 *     def read_byte_len(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_byte_len(self.cq)
 *     def read_imm_data(self):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_4CQEX_17read_byte_len, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQEX_read_byte_len, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[19])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX, __pyx_mstate_global->__pyx_n_u_read_byte_len, __pyx_t_2) < (0)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":398
 *     def read_byte_len(self):
 *         return v.ibv_wc_read_byte_len(self.cq)
 *     def read_imm_data(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_imm_data(self.cq)
 *     def read_qp_num(self):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_4CQEX_19read_imm_data, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQEX_read_imm_data, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[20])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX, __pyx_mstate_global->__pyx_n_u_read_imm_data, __pyx_t_2) < (0)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":400
 *     def read_imm_data(self):
 *         return v.ibv_wc_read_imm_data(self.cq)
 *     def read_qp_num(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_qp_num(self.cq)
 *     def read_src_qp(self):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_4CQEX_21read_qp_num, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQEX_read_qp_num, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[21])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX, __pyx_mstate_global->__pyx_n_u_read_qp_num, __pyx_t_2) < (0)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":402
 *     def read_qp_num(self):
 *         return v.ibv_wc_read_qp_num(self.cq)
 *     def read_src_qp(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_src_qp(self.cq)
 *     def read_wc_flags(self):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_4CQEX_23read_src_qp, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQEX_read_src_qp, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[22])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX, __pyx_mstate_global->__pyx_n_u_read_src_qp, __pyx_t_2) < (0)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":404
 *     def read_src_qp(self):
 *         return v.ibv_wc_read_src_qp(self.cq)
 *     def read_wc_flags(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_wc_flags(self.cq)
 *     def read_slid(self):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_4CQEX_25read_wc_flags, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQEX_read_wc_flags, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[23])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX, __pyx_mstate_global->__pyx_n_u_read_wc_flags, __pyx_t_2) < (0)) __PYX_ERR(0, 404, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":406
 *     def read_wc_flags(self):
 *         return v.ibv_wc_read_wc_flags(self.cq)
 *     def read_slid(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_slid(self.cq)
 *     def read_sl(self):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_4CQEX_27read_slid, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQEX_read_slid, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[24])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX, __pyx_mstate_global->__pyx_n_u_read_slid, __pyx_t_2) < (0)) __PYX_ERR(0, 406, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":408
 *     def read_slid(self):
 *         return v.ibv_wc_read_slid(self.cq)
 *     def read_sl(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_sl(self.cq)
 *     def read_dlid_path_bits(self):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_4CQEX_29read_sl, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQEX_read_sl, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[25])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX, __pyx_mstate_global->__pyx_n_u_read_sl, __pyx_t_2) < (0)) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":410
 *     def read_sl(self):
 *         return v.ibv_wc_read_sl(self.cq)
 *     def read_dlid_path_bits(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_dlid_path_bits(self.cq)
 *     def read_timestamp(self):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_4CQEX_31read_dlid_path_bits, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQEX_read_dlid_path_bits, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[26])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX, __pyx_mstate_global->__pyx_n_u_read_dlid_path_bits, __pyx_t_2) < (0)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":412
 *     def read_dlid_path_bits(self):
 *         return v.ibv_wc_read_dlid_path_bits(self.cq)
 *     def read_timestamp(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_completion_ts(self.cq)
 *     def read_cvlan(self):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_4CQEX_33read_timestamp, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQEX_read_timestamp, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[27])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX, __pyx_mstate_global->__pyx_n_u_read_timestamp, __pyx_t_2) < (0)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":414
 *     def read_timestamp(self):
 *         return v.ibv_wc_read_completion_ts(self.cq)
 *     def read_cvlan(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_cvlan(self.cq)
 *     def read_flow_tag(self):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_4CQEX_35read_cvlan, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQEX_read_cvlan, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[28])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX, __pyx_mstate_global->__pyx_n_u_read_cvlan, __pyx_t_2) < (0)) __PYX_ERR(0, 414, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":416
 *     def read_cvlan(self):
 *         return v.ibv_wc_read_cvlan(self.cq)
 *     def read_flow_tag(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_flow_tag(self.cq)
 *     def read_tm_info(self):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_4CQEX_37read_flow_tag, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQEX_read_flow_tag, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[29])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX, __pyx_mstate_global->__pyx_n_u_read_flow_tag, __pyx_t_2) < (0)) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":418
 *     def read_flow_tag(self):
 *         return v.ibv_wc_read_flow_tag(self.cq)
 *     def read_tm_info(self):             # <<<<<<<<<<<<<<
 *         info = WcTmInfo()
 *         v.ibv_wc_read_tm_info(self.cq, &info.info)
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_4CQEX_39read_tm_info, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQEX_read_tm_info, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[30])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX, __pyx_mstate_global->__pyx_n_u_read_tm_info, __pyx_t_2) < (0)) __PYX_ERR(0, 418, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":422
 *         v.ibv_wc_read_tm_info(self.cq, &info.info)
 *         return info
 *     def read_completion_wallclock_ns(self):             # <<<<<<<<<<<<<<
 *         return v.ibv_wc_read_completion_wallclock_ns(self.cq)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_4CQEX_41read_completion_wallclock_ns, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQEX_read_completion_wallclock_n, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[31])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_CQEX, __pyx_mstate_global->__pyx_n_u_read_completion_wallclock_ns, __pyx_t_2) < (0)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.cq,self.ibv_cq cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_4CQEX_45__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQEX___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[32])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.cq,self.ibv_cq cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.cq,self.ibv_cq cannot be converted to a Python object for pickling"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_4CQEX_47__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CQEX___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[33])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.wc must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_2WC_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_WC___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[34])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.wc must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.wc must be explicitly requested with @auto_pickle(True)"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_2WC_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_WC___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[35])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_10PollCqAttr_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_PollCqAttr___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[36])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_10PollCqAttr_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_PollCqAttr___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[37])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.info must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_8WcTmInfo_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_WcTmInfo___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[38])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self.info must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self.info must be explicitly requested with @auto_pickle(True)"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_8WcTmInfo_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_WcTmInfo___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[39])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":598
 * 
 * 
 * def cqe_status_to_str(status):             # <<<<<<<<<<<<<<
 *     try:
 *         return {e.IBV_WC_SUCCESS: "success",
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_1cqe_status_to_str, 0, __pyx_mstate_global->__pyx_n_u_cqe_status_to_str, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[40])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_cqe_status_to_str, __pyx_t_2) < (0)) __PYX_ERR(0, 598, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":625
 *         return "Unknown CQE status"
 * 
 * def cqe_opcode_to_str(opcode):             # <<<<<<<<<<<<<<
 *     try:
 *         return {0x0: "Send", 0x1:"RDMA write", 0x2: "RDMA read",
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_3cqe_opcode_to_str, 0, __pyx_mstate_global->__pyx_n_u_cqe_opcode_to_str, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[41])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_cqe_opcode_to_str, __pyx_t_2) < (0)) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":638
 *         return "Unknown CQE opcode {op}".format(op=opcode)
 * 
 * def flags_to_str(flags, dictionary):             # <<<<<<<<<<<<<<
 *     flags_str = ""
 *     for f in dictionary:
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_5flags_to_str, 0, __pyx_mstate_global->__pyx_n_u_flags_to_str, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[42])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_flags_to_str, __pyx_t_2) < (0)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":647
 * 
 * 
 * def cqe_flags_to_str(flags):             # <<<<<<<<<<<<<<
 *     cqe_flags = {1: "GRH", 2: "With immediate", 4: "IP csum OK",
 *                  8: "With invalidate", 16: "TM sync request", 32: "TM match",
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_7cqe_flags_to_str, 0, __pyx_mstate_global->__pyx_n_u_cqe_flags_to_str, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[43])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_cqe_flags_to_str, __pyx_t_2) < (0)) __PYX_ERR(0, 647, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":653
 *     return flags_to_str(flags, cqe_flags)
 * 
 * def create_wc_flags_to_str(flags):             # <<<<<<<<<<<<<<
 *     cqe_flags = {e.IBV_WC_EX_WITH_BYTE_LEN: 'IBV_WC_EX_WITH_BYTE_LEN',
 *                  e.IBV_WC_EX_WITH_IMM: 'IBV_WC_EX_WITH_IMM',
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_2cq_9create_wc_flags_to_str, 0, __pyx_mstate_global->__pyx_n_u_create_wc_flags_to_str, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cq, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[44])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_create_wc_flags_to_str, __pyx_t_2) < (0)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cq.pyx":1
 * # SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)             # <<<<<<<<<<<<<<
 * # Copyright (c) 2019, Mellanox Technologies. All rights reserved.
 * import weakref
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_test, __pyx_t_2) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  if (__pyx_m) {
    if (__pyx_mstate->__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init pyverbs.cq", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init pyverbs.cq");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #else
  return __pyx_m;
  #endif
}
/* #### Code section: pystring_table ### */
/* #### Code section: cached_builtins ### */

static int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_super); if (!__pyx_builtin_super) __PYX_ERR(0, 30, __pyx_L1_error)

  /* Cached unbound methods */
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.method_name = &__pyx_mstate->__pyx_n_u_items;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.method_name = &__pyx_mstate->__pyx_n_u_pop;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.method_name = &__pyx_mstate->__pyx_n_u_values;
  __pyx_mstate->__pyx_umethod_PyUnicode_Type__format.type = (PyObject*)(&PyUnicode_Type);
  __pyx_mstate->__pyx_umethod_PyUnicode_Type__format.method_name = &__pyx_mstate->__pyx_n_u_format;
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "pyverbs/cq.pyx":153
 *             raise PyverbsRDMAError('Failed to resize CQ', rc)
 * 
 *     def poll(self, num_entries=1):             # <<<<<<<<<<<<<<
 *         """
 *         Polls the CQ for completions.
*/
  __pyx_mstate_global->__pyx_tuple[0] = PyTuple_Pack(1, __pyx_mstate_global->__pyx_int_1); if (unlikely(!__pyx_mstate_global->__pyx_tuple[0])) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[0]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[0]);

  /* "pyverbs/cq.pyx":179
 *         return npolled, wcs
 * 
 *     def req_notify(self, solicited_only = False):             # <<<<<<<<<<<<<<
 *         """
 *         Request completion notification on the completion queue.
*/
  __pyx_mstate_global->__pyx_tuple[1] = PyTuple_Pack(1, Py_False); if (unlikely(!__pyx_mstate_global->__pyx_tuple[1])) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[1]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[1]);
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_tuple;
    for (Py_ssize_t i=0; i<2; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  {
    const struct { const unsigned int length: 8; } index[] = {{1},{14},{15},{18},{10},{12},{26},{16},{12},{28},{9},{13},{18},{21},{37},{28},{22},{38},{22},{17},{19},{13},{10},{16},{179},{14},{95},{97},{97},{6},{9},{10},{18},{27},{34},{31},{13},{8},{15},{18},{21},{21},{22},{19},{23},{18},{24},{14},{8},{5},{14},{14},{15},{0},{1},{1},{8},{18},{11},{9},{7},{14},{6},{179},{170},{170},{2},{13},{21},{26},{9},{18},{17},{25},{18},{18},{22},{24},{10},{14},{18},{19},{34},{31},{15},{22},{20},{61},{59},{59},{71},{9},{15},{12},{3},{4},{22},{24},{10},{13},{14},{18},{33},{15},{24},{18},{18},{16},{16},{12},{14},{16},{19},{17},{20},{18},{15},{4},{2},{20},{22},{13},{8},{7},{13},{9},{11},{29},{31},{17},{24},{12},{30},{32},{3},{6},{19},{19},{14},{16},{14},{14},{23},{35},{45},{20},{29},{23},{18},{21},{17},{19},{21},{22},{16},{16},{12},{18},{19},{24},{16},{21},{21},{18},{19},{20},{23},{18},{16},{17},{11},{25},{20},{21},{25},{22},{17},{23},{20},{24},{11},{14},{13},{13},{13},{16},{14},{25},{14},{10},{19},{10},{28},{30},{12},{16},{16},{20},{7},{4},{3},{2},{20},{22},{8},{26},{28},{7},{10},{3},{7},{18},{4},{8},{7},{17},{18},{5},{12},{9},{11},{7},{2},{10},{3},{9},{16},{17},{17},{22},{3},{5},{10},{14},{8},{11},{1},{5},{9},{12},{6},{8},{12},{12},{22},{13},{13},{8},{4},{8},{9},{13},{5},{8},{10},{8},{7},{11},{10},{2},{6},{13},{10},{4},{9},{3},{12},{10},{24},{21},{11},{14},{6},{12},{2},{13},{28},{10},{19},{13},{13},{11},{11},{7},{9},{11},{14},{12},{15},{13},{10},{17},{13},{10},{6},{4},{12},{10},{12},{19},{2},{4},{14},{6},{10},{6},{7},{5},{8},{6},{10},{4},{8},{2},{8},{3},{7},{5},{37},{14},{14},{22},{14},{14},{14},{83},{122},{99},{37},{15},{14},{14},{15},{35},{14},{13},{71},{71},{29},{49},{9},{42},{52},{105},{131},{185}};
    #if (CYTHON_COMPRESS_STRINGS) == 3 && __PYX_LIMITED_VERSION_HEX >= 0x030e0000 /* compression: zstd (2898 bytes) */
const char* const cstring = "(\265/\375`Y\032EZ\000\352qx\027G\000\2218\007\374\277\253\365[\351\237\357\335\377\036\375\352W\277\372\255\344,\371\337\215H\221\326Dv8\206\355DIB\332\335\002\373\333\376\267l).\267\345\337Hs\014\301\214H \226\030H\362\027*\357\212\021aN(P\016H\001R\001_\001:\277p\262\337\316i\240\364\214\372\240\306(]\363Jg\231\374\262O\267W\235\277\177\207\323\375pn\361;\326\333/X=\037\r\274g\344M\232\302\352e\262\346[\263\256\340\367\033\202?\216\341\343\367\300\007\377qkw\357\343}\017\367\n\274\232\207\363F2\327\220H\233r\270{\277&O\356\257\344QVw\362\252\2501=\\\361'S\277{\210C\021U\252\235\223c\305+\024\223\370?\326\216IQ\006\207!>\222\037\224\347\342m\331!XI\354\355\232\244\366\3475\377\347J\352V5q*\n\177T\253\327\034v1$K\325\244+\177\316\257\210K$\257\274\343\217\361[\205\204J\207P\027\252y\177\307am\211\373K\241\310\303\230\336\2721]\177\276_0\024\257*\207U\274XwX\200\352Hr\\C\034\020\2741\270\377\313\010\371oQ\324\250\236_\365\265\377\251EK\357\316\037\374\307Uw1\274-\036+\340\203\371;\253\363Hu\277*\334\334v+\370\375\376\336Pw\305o\375}7\361s\2761\270f\035\366\027Q\243\306\254N\264\325S\016\271\026\267\225^\325\326\326\343{*\230C\026\017\032\037RL\254\314\250+4RRVf\240\004\226\004\352\264\214>\216\026\262\034\247\303i\035\030\026\033\250\263u:\032\311\206<\3327\352<\234\317~@\234\257\353|\272\216\355>\027,v\305\307!\331\026\035\210\024\257\203\341x\\H#\215l\310\003\332H\036\313}8\236\327\331\320\250\363,\247\261t\037\307#\331\215\343\001}\232\335H\026\206\205\363\"\230\357\212\025\236\013\253\333\275k=\314\234\324\326\377d\257h\353\333\332q\304\033\014EnQ);\216\341\351v\273\322\242\311\332+\252x\2662X\024\357\312\266J\266\263>[\333\237[\213\355c\022\374m-\212\207h\333\2572\351\371\307\323\375\335K\036oRsHa\214\252[\233B\365\366\2541&\365\217a\357[\227v\n{`}m\275\014\252TG\374\235\363\307\370\212:TU\216\363B\"\224\026\301\377l]\267?\377\t\237PX[\327\332z\325\215y\r\212\241\251\202,\024\n(\342\007\253\351A\024\256\227\254\255\222\336\250Jj\027\026\205\227\366\253Z\250:""\352m\252}k\224\305+\233\302\326\252Ja\255\237\237\367\260T\272<\004exL?\345\027[\245\347Y\243\300[\372\256\263MQc8)\2328N\365H\374t\305\255\367\376*\336~Mu#\225t\005\363)\251\250\034)\271&\246\371H9D\314\010\037\251\310\231\235\"L\227\222AH\272\001\371Ho\t\242\354.\025\027\202S\345\0249/\244P\025T\004\335\220\301\276\227\366R\331\230 \327\207\201Tt\01689scN\360\000\360q\343\305\241\033\364\204\003\345\316\3159\022\022\0131\220j5\270\005`R\262\001\324\n\201\241\335\253\025\307\253\274\357\212?)^\025\377n\355~\024\311[b\326!\016E\315En\273H\202\326\262\336\342-^\322\252P)\374\374\226\254\006\367E\2414\377p\333\221\314\340\2177\277H\202\332\332\321Znyh{\230\255\2658\223&k\373\035M\255}C\360D\226\036du\274\332\235[\335\355\333\024N\341\316\372\2365\336\272\343WY\\z|w\2679\264\343\367\260\377\311Z\313?\207\235\324[c\024\035^\023\010\336\376#\231\373\332\2249\343\207I\033i#\201\371G\246\367:\353?\266\343\220c\351\334\205R\241\321\366q1\352`\270\017\350\323\3518\317\367\302}\272\221\367\261.\274\357\005$\362@,\232\365\300l\234F\362\20080\366\003y\241\020h4\352\266M\263\326\302x\335\247sY1z\351p\234fWl_\327\322\321>\255\005\324\362\341X\317\267y\\g\364\362\261!\221\307\303q\321\276\034f\313\005\202jP\223\031ofdfD\267\201i`\007^/\316\224s\344\225\362rAK\234\232S$B\232\202y0\316\213\306\351\235\031\"\244\242\231\224\031\027X\211\227\246\344\014pj03s\227\267\\\245\312\007\037i\246\301Lh\345\005M\000\032\202\003\207Y*c\\\n\014\330\020)5\006\322\033r\243#\226z\335\227\010\r1\r'\006\264\003z\222eY\226\245Y,K\241\005\320\022\220\330\007\202l\006\306K\205z\200A;\264\240\034n)r\266\230\244\222\343\025\002=\271h\003\324\244A\254AH\252\275H\264B\213r\300\200\327\013:\202n0I\265\031\356\345dpr\234\265#\266\311\264ssF0!hJ\021\"\327\000\227\2012DT\236\270<De\203\014\204\2418\211\365\223\204A\214A\226n^\002^\037*\304@,(\263T\006\n\007\322T\311r\201\230\005\314Ti\302\205\240#\025\231aI3\324\004F\342\225\242\321\240I9\320\022M\252I5\224CG\312\221\242\\\240\231\0226gVb\027\210\024jAAL\243ih""\007\232:I`+#\261\315=\033\234\237\251\231\331\005rv\363\232\nDDg\006\221\0234Ef\2022;\000\300\343\242\030\320\014\355@J\221\271R.(\005\224!\\\037\007\262\031-{}\350\211\010\264\237\0056&\002\021\354\004I\354\236\030<\224\332\331q\246L\026\242\027\203\027\t\n\236\004nI\240\014\371HKH\322\324yO\r\016b\016B\002\202\223\250rs\230\205\214)\032\221\221$)\244\003r \010\013\215\3634\315\245\017\342\271\226d9\"\0253R\"2\"\n\212J\212\nR\3110\006\244H\300+\300O\324N\252\273\266\374\023\204\303\246d\305\021\225e\217\n\354\207[\311;j\244pN\336\n\324C\345>\226/\324\270W1=`\213\376\177\365\312\220BD\325\252-\326\266L\000\261z\254\245\264\r{tE\376\241\230\325V~\210\375,R\210\357)\206F\\,'\r\266^$\3025o\302B\202V\352\366\177\305Z\347\003lG-\025\311\267\r\270N\023\362Q\250\226\004\260/=:|E=\220\331\341I\3624\245\277\266i\304\220\357{\211\01046\223\324\327\344\005\220\036>\2524\020}\206*\310\263\001\026c5\322\303\310X\360\202\021\201D\242\307\007%m'\375\"\220\272N\342\023\247\344\360Fp\243/\341\272\355!\321wC2\225\263\276\230c\272s\201-\307\252Y\234\321\233\330\213\310\301\357\366\252\215\364\303\204c\004\320(V\202\225m\356\342\202\326Np\316\317~\275b\314Q\206\353\027_d\333\345W\241\233\310\r\351JH\237\003\230\006\300\002\"P\030\277\241\237\304_\222\316k\325K\032\0004z!q9\3028\355(\241\364[\031/\211\362\274\006\315p\260\301j\367\253.:\204\027P\nP\314\212\3418\236\25616\233\210p\263\022{\343E8\001r\234=\340\370\033\235\230\254\263\272\355\270\004\236(-\317\032\365X\354=Q1}\206\344y\356\330M\361F\255!\243\273\020\344'\037\tV7\263a\t\220\212\340\307(\024\3472m\230\230\236\033\233\347\036D\026e\t\351\273\253\337\265\265--M\324\223\244\035\230\273n\266\356\036\244\354\005E*ppU\201B\177\300\341\"\267\214R\227_\334x\333\326\210\023\031\013\234\022\202\321\275x@\250\003\310\245@&\234\332!<\344\013\304\272V\215\273=\373\246\373S(OB|V\336\340\024\005\347\335\313\241\256\233\2668\022\027t\313%\277\310F\274w\311\\c\213QT\257i\216\220\206\376\022W\325\0029\344\253\202A\317\334\371\n\215*\300v""\304\266\302`\260\364\377\036\026\013#P\260Hd\205\362\355\230\036\030\265OOm\335\365\200\230f\317\370\3346\234_j\337P\250\303$\301\242\237\2352|M\311\355FD\324\016q\270\236(\362g\020\006M\372\361\347\257,\364`\"\257\360\336\177?\030\033\001n\237\316\206\237\220\365\023@_l\022o\222\003i\225IG\340Ak\021a\241\021\315\240\377\261\024\231\307\360\306\376\022D\345\363$7\371s>{\304\224\003W>b\201\"\316i\322\r\300Q+(1\020\253\020&\005,\003\252z\344\232H@\375\247fiy=^\030!\263B\003\226\031$\034Uv\3566\336\241\334\000\262\022\231\002P\316R\211\010\375\230\202J|\312x\340\251\0100\354\244J\014\241\262c\272N[\227:\307`\r\345\315\024\206-nr\260(m\262\003\376u\014\360\2233\273\242\370\222\031\256W0\346)\002 \231c\364\372\336t\004G\350O6\004\224~\240\230\274l\352T\022\036y\277\374\263\252\204d\261\274Mf\2170u\310\250\t\204\n\371\037\370\265w\370EH\255\"\020;\233\030\352n\346\215@\014g \350O&\\^<\204\242\201<t\340\322\247\356\255}\013\306\270u&S \221g\234 ay\325x\217\333_\240.\035\343\303v\240\024\220\212\024\320\225\255\021\272\214\205\312O\307\250U\340\0137\310\224!Z\333}\202\306\257c\224\010\225rw\261F\252\300\017uo\002\\\210\203\360\2708\361\271/\315\360\005\305\346\270M\302V\265\013\335\007\207\243\254\024\204\216\324y\366\307+\220\\^|Cv\003\201\377\030\034\000W\245B\223=\3164-\267\237\365k~\344,s5\353\006\230\321\231\013\210\315\236m\\\341\347\322H\233T\004\267\276\373\200\260\241\304\326PTn.v\024,:u\203Y8\225\316\307[\010h\n\265\2749^\2544bcHua\332\024\210\273A\317\t8\327\257\n=kGf\365dl\246y\325\312\336\236\222!l\221\245\304\373\\\323e\334_\376\352\223\344\013\261FU\355q\262\017\351\360\223\"A\374\350kE\242[ xT,.d\335\341\304\370\361\0051\325\"\225\372l\300Bc5\033\004x\025M\335\025\221\364\272? \2230r\306mf\347\360\"z\235rk\036^\223\211c\233\r\"\333Q\030\265\203~\340S\021\004\200\022\347\000\032\353\246\270\352\020Qed\254\200\220 \370\277\301\307#\273\t&\325P3\242\360?\364Y\215\026J\266\026\025+aM\274#\262\360\365\373<\330\274\300\273bAT g>\267\264\364y\343\020r\014h\341B""\320\353\000\263B\273\001\223vx\361\303\316S\325\025\325=\323\267\2331P\220\217\204a1\312$\217\005|x\303\332\327R\320\214\027sV,\256\220\333DCzg\032\365q\267t\212\320\337\211\035\266n\213\216,\322\001\254\215\316Prr\034L\206~\230\366\021L\201\226d\r\026\225x\037$\316\263\0000t";
    PyObject *data = __Pyx_DecompressString(cstring, 2898, 3);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) == 2 /* compression: bz2 (3096 bytes) */
const char* const cstring = "BZh91AY&SYsj\314r\000\001\325\377\377\377\377\377\277\377\377\377\367\377\357\377\354\277\377\377\372@@@@@@@@@\000@@@\000@\000`\014{\347[\013N\033{v\367]\226\347t:\352D\251@\000\022\024(\270\017\003\"F\246\215I\372SS\362dm\032\rM$\363)\344\325\037\351L\232b\206\233H\306\204zM\251\352y!\352\r\247\244d\3206\247\222i\r\2502D\323F\223\310jh\325?P@\032\r2\032\000\000\000\000\000\000\000\320\320\0002\000\320\004\225\036\325=\032\236\243\324\r\000\006\2326\240\364\200\320\0004\000\000\000\000\000\033Sj\032\002DPL&\220\312m)\345? S\365M\222h\036\221\352\0324\000\000\000\000\000\000\001\352\000\002\014\000\230\000\000\004\300\002`\000\000\023\000&\206\000\000\000\000\000\t\022\002\010\320\231\000 )\351\212i\344\324m\020a\r\r4\000\000\000\000\000\000hbX\302;B\260\233\036\321\013\371\2300~F\003\001]\221\300\375,+\014\320C48\014?\220\266\310Q\200\351d\036\333H+\271(\307\334%\355\375\317\223\026\260\372\200f\245u4\222\353\247`\246\322Q\023$q\210d\311\021k\000T\013\316\302b&M\240\213\355\022\306 \304\177\247\222+x\304\224\010\007\200\360P \023\300\266\206\272\001\204=-l2\014.wf9\317\030&\223\002\353$5\257\016}\226\303\271\257\332_\267\331?\202?`Y\204\3372x\371L\345\206\3725\036?\n\222\324w\027\375yT\277U@R\200\317=$\344\245(\223\200\247\336*\231\031\314m=&g\225\232\211\375\2524\375\263\303\215\363]W\303\300B\263\235\354\224\327\240\301\320\254\322i\301[\363\226\255\001]\022\252\327H\201\210\261\261\270L\007\242C\374*\307h\014\002\265\027\033\301Fc:\345\017z)\366ZD\346\350\277\027\r 9\033\242x\025N\312\014\261l\331\274\362\3678\0050\327\210}4\2676X\3413\262\333\257\233#\\\330\245\204Ml(\214\000\300\325\t\004K`g\013\023\r\003\225\241^@\251\265\267u\370\347\267\332\025\336k\023\230\274\206\200J\"\222\335\207\352\026g\237\347\023\t\341\\+\201\"\222\241;c\362\202\206/\275\345\252\357c@\257\261\231\366m\226\312\n9vn\315t-E\n\226\262\254\213.\343#\213k\207\017\002\226oh\3308\271`\205\266]p\030x\331{k\330\270\370\321\021\021\021\033\001Q\245(\354\035\252'`\300C\326\tL\260\224\210\234""\010\356\303rD\033\343\305\235d\320@\020Fi*\243\030\360.\207GS\314mH\370\231M\355\007_/\234\216\017\245\303.\n\354c\025\245|\235=x%A\227\333\035\000M\336\236`\027\007\357U\250u\017|\215^6\014#\032C\236\243\225\317[.\314x\347\213\345*fEJ\246V\321j\002\021\255F\3458\3041\264\034\2005\375\304\220W\277^x\314\0100\342xf\032;.\322x1\363~!e\202\354\222v\232\342\001}\245k\330\365\354\337\255\353\327\330\267\257\362\317Kf\035\032Y\201\304h\371\370\330\236\342\004\020\300\224b\230R\004\246@b\035d)\316\271\300\261`g\302A\2235\312LW\314=89\255\233J\214\3402ic\232i\334\277\236\032\265\034\230\3133\003L\301`\267\325\364\243\364\346\341;\016\306\223\374\2341\022$y\212\247\222\352f\202EF\301j8\373Ngln\274\017\025\321\2529C1#X\231E!\206\\\001.\000\315\031\264\250\317\005P\001\2128\233f\246\277\252\2131\253\365\353\n\274\025\361\3379!p\264\247$\342\240\n-vM\265W\261m\225\355\274\253:\345\030\254\223\243y\316 \312\2416Q\333\035D!\215z\020\030\r\221K1\330\344\007VV(K\260\026\2005\306\334\304l\213@<\014\313\203\033k~\177\352\364\343\202\033\275\3469\242\264\300\237c\352\361\001\\^\350P\002\204\346\347\276gh\362\206\352|\202\327>61\014\224_2\t\026\234\261\227\226\332\223O\355\310\001\216\310\271D\222%\272\030Y4V\2644Vz\262\326j#\240[u\232\231\342\031\206\204\002\014\370\337\306M \277\207\221\010\203\274O\272\373\364\314H\352\356\027\323?c\037\037g\201\313\303O\001\237\034\2104!\241\204\214}\251\035\311\252wG/\013\222\025\024\317\211\267\353\314\n\320\254\251\302\204\023\234\203\260\233g\3471\213\026'Ch@\364\372\360\313\234\036\316\023\321EuQl\222\013]\351d\350J\222\010I\007%\010/\243!fFd\254\315>\267\030\202\002^\2060\240\307\013\210}I\230\2129\356A\320\020\325\021'f\337\036\331\237[%\023\252)\331\232x\212f\302\353\265\022%\017a\341&\025_I\006\020\00677pLM\023gM\221kV\267\252+b\256\207\005\363\202\251\230KwD\270\212*\264^yb\222\027\242\226m\025\2070=1\035,\nBS\010\232wgf\243\225\226\273%\240j2Bh\335rS\204\276\363\005\263hv&\205\032Y\036\335\214\212Hd\000kPN(\030\266X\363\211\333jm\262""\346%\331y\024\364,P\244\324g\227\021|\312fUO7x\200\342O\006\354NP\340\201\020\322\0306qJ\256\241\255\372e\254\023\253[\211\343C\t\231\310\264m\010\360\0105\013\033\037\371Mm\341Y\226I*\036\316Xy\210\261\203\336\013\340\240\343\223\305p\005\312\243I6\016\324\313\254\264]U\344:e\222\034\007\021\224Z\3234\205r\005\211\330l\2505Y\256\234\202{\324\234\t\206\343\027\007\264IH`5\305\200a>C\237\222H\273\372X\327'K\2229H\236\224\007Q\322\212\034.\005\256a\244\316\226\3179\246\021\260\331\321.\217\214\221\310\001\215l\221+7\237WOO\2112\000\252fI`\032fiV\"5\030\005\313\247\267Rv+\362J#X-1\216\004Z\253U\356\217A9\027\277\305\274\226-\016\272+X\214\256\001\331\t\303\346\343\220\327\340\267\0214W8\224\221\031\352\026d\366\261\266\325*Q\200GZ\232\340\3063\2250\264\2709\010%\334s\224Ns\t\260\" \232\231\004\330\223\313\005AM\034\201z+\0278-+A\214\241C\226[\347<\274.\n(\230\263&\"\375\263\262z\274\372\271=\310:\337\200\030g\335\n\246\023\277J\205\264\352L\225Ut\312\334TBd\273f\245Sy\342n\203\205\314\020Km\266\333c\033zL\322\240\266-\275$\267\364\220\006*j\362\223p\242\327\367\022\324Z\232\347bi\312kI\206\211md--\004\315\205\025\222*\300\232\326\024\303\0354\204\302D\340\201\330b\016Z\2219\255\254}\317\014\322d\016\315Tb\326\035P\266`u\251\205u\345p\332\303\016m\\7J`\027\307$\330\201\223S\016\245\221{\360\311u\263\215t\006\275\204-\007\277&>\036(\343Jx\222\206\004e\312\332\031\222\250V\223\333c(\236\003)\264\232\352X\004\267S\321\002\270oS5?\023\t\344&\016p\026\233\340\250\221\336C\006t\360\344\274\256\251\243\021d\306\332\332\254\017S\010a\357\221\r\204\006\t\250CB;\374\r\310\316\271\274\036\235AP\351^\212a\357\367\345C\026\nZ\227\001nA\276J\272\336\260%Izz\342\236\350\312\306\332]\367\021\210\252\2660\343\241\\t\032\341\333\314F\351\332+t;W\206\201\336&\266\204\350\325t.\350\224\365Y\0234 EI\2324\025\010*\340i\2037\202\330Y\0074\312b\271\327Rko$\344\201\225|F\361\337`\251\0028\303)\254t\331\t\212\rz\322\252\t\214\0357\"\200t-\306xY\306\335\263\246\365\246\345\276\311kO\312\235\037q""2\306\372\276\033X\251e\232\2715N\326\260<VK\033\264\224\324`\023w\\\345iU\t\025\211C\220\224\230\332\221m\271\317\3151\254Z\355\222D\021\001\013<\243bD\342\330\021\333\033bD\313IH\327\360\314\362\000\312\331\004\t\343\014\3460\214\255\204\260\034\305\003\021\026|\334\2210\270\252\031v\370\002\345k\\\025p\311\001E\2228\203\322a\320\227.\212\210\226:t\207q1b\255^\2444\306\213]\243\264\2040\342\271AQI\265'\214|\331\2546\006B\204\366J\000o\001\201\330\272\320\330\203rD#\275\025k\324]zr\320\326`X\221\312\261\363\030\272\031\t\227\334oUKj\325\344;1EP\214\263\304VT\356\330z\353*=\206|x(\3221\010\304[\366MN\370\3241\036\033\003\234M`4\024\023p\227\223\257\311\036_<\200\221)\271\372\016\204v\235\374Wz\26348\200pFx\010D\265\320\223z\201\005\023\225nL\232\314\337\320\033[<\347\270\355p\341\361:\235\354\321\2427^\226\336G\226\361uf\030p\343[\000xN(\360\026\354\344\026\316\264J;qA\227\rk\233\307\tscg\271s\251\000\242\205\307\3118\027\253&\306\300PXj\002\213\336*\3012\235\255\375(\272\351\325\025W\246\ngB\357\201M\\\321\214\346\314eW\224\321\214\2122v\302Z\261:\241\031\357\242\377\020[\014\r3)\312#\243\rm\3464\361W*\2652]y^F\332\034\333kC\004\235\3305\tm\347\326\254\265\200\341\230T\240S\003F&\264\304\316\005\020]\325w\2212\263\334\027P\270\340\322\241A5-\226\224p]P:O\t\323JE\211\274J\215x\345\263GB\343bY\264RB\314\210\024\3063{\206\362sbSNf\2502\013\251Zl\271,3&\251\266\024,\321&d\204d\242\306\275\ra|\234\020\336*\245\222UP\211\347\271\004\236\260\365\232\321\243\022\322\305\261\271\021q\336\t\220\321M\345\271R\353\rF\270t1\255\301\rl\212R\315\244\246\252\355H\363Z9\364\236\\\003\0274\372\224\326\013\335\346\225q\026\356:\246\241\374x3\343J}\t\030\003\270a\310_,\302\014k\277\2614Y\205\006\356+\033\243\227%\333\007\243\031\331[\341)0\233)Ek\232\212pJ\t\370\324\345\232\3175[n\016;\003X\002\360!0mc\314\033\027\0363?S\307,\305\244\346\375\267\274W/M\021\033\213AF\366\363e\004\013+\315\323\341\000`B\234|\036#\201@;\245\017\223\340\204\315\010(\016v.\354\362\020\362%\222}\243\2318""\300\337\301\276\002N\366v\262\212\216\035\234\003\352\345\261\245{8I\204zL1\376\251\241\211\002\360\001\257\212\340\033\230\261\361^\301w\360\2761\213\334I\314)\215m\276J\200)\220\345\340\t\253\2708\266\244\"\271\376\325C9\227J\314\217\204Fi\362\213\037\004Q\317\244\207S\356nX\032\212\246\266\237g\222Z\021V5\342\025w\200\330\260\027\3067\001\251\217\205\010\2105\316\336\300\246\202$\212\240\213\252\002\004\3770\266\302\304*\323\245\016+P\210a\332\371[\254.Eo\251O\nkE;\372\367\213a\356\277\007J\216<\236{Z\320\r\230\337\377\275\277+\335\016\341\275\243\321K6\215\251\231\267>\303\272\000~x5-\217\031;w\306\255c\326\354\2013Y\254/el\350\212\001\367\rS\202\254`\004\227\221/\024\212\214\2267\314\264q\022\316\216\214\210\251Bt\001\332\035\324\302\253\203h\242\262+\314\241\302\273\310E\223WCg\225v\2314\n\306A\r?\342\356H\247\n\022\016mY\216@";
    PyObject *data = __Pyx_DecompressString(cstring, 3096, 2);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) != 0 /* compression: zlib (2900 bytes) */
const char* const cstring = "x\332\325X\315s\333\306\025\217\033;\361\207\032\213\262\235N\3356]E\211%\371C\211d\307M<iS\230\244lN$~\200\224\030O;\263\003\002K\t\025\010\220\000(\212\365x\246G\035q\344\021G\034y\324Q\177\002\217<\346O\360\237\320\337.\000\001\374H\306\223NgZ\217%\274\367\366\355\373\332\267\357\275\325\306\353g[[o\236\221\327\317\276\335\372\362\315u\376I\241\317uS#\273\254e\331}\322\003l\365\262\206\345\350\346\001\311V\022(\177\022\303\252\325j\033\314\325-\223\250\207\212i2#\013\212b3\242@\220\323S\332Y\233).\323\210\302%$p\262/\033\356\313\331\3721\263I]\316\237\270\314\324\300\225\255<\273\276\255\350\006@\327\"*42\320R\024!M\010\236C\234\265l\206\211%\232\2225\2159\256m\365\271\227?\314R\177^\356\001s\261\217\260cf\272\t\265m\031\306\204\n\2339\372?\205/\314U\017E\244\024M+\224\211\352t[\244\364\375\216\245*\006\321\315c\305\3205\030Z\264`\254{\250@x\337=\204b\335\201A\206\336`6V\215>\201m\272\352\"z`2I9_~\364\344\353'B\256\315\376\301T\327!N\267\241\032\212\3430\207XM\322\350\352\206\253\233\304\355\267\231\263A\nM\322\267\272\304d\221\275\340Kop\017\231I\034\270\006\200\254\302e\313U\270\377\024\333\221\003\253D\323m(\301\361\361\335\333\212\341\260\215b\267\005\353\270.\304\321)\353\352\221\301\323\0058l\355\252.i1\316\300\365\360\010\340\313\214\346FO%\255\256\343\222\006?\233\266\241\253\272\013\357l\326\351\"\3740\256\247\273\207\344\257J\327\265h\233\213dk5\273\313\326\337U\274\342\272\366\177U\201n6\255_\244\240R&f\267%\347v%\360*\232\000z\266\3562\271(\203\342\3426\262\023\025\347\3034\231\251\214\207:\344\341\342\364V\213i:2!Z\322\302\004$H\024dC\327\204!8\036\221\346rh\010\301\021\352M\236\343\\x\3276\261\370\332V\337\324v\t\362M!\"\361\200\264\024\344'\276N\337Tc'j\312AH\347\361x\304\023w\212\202\314d.\233\"\232\026q\225\203)\242\035\332;E\345\312\366\314#\323\352\231<y\210\325V-\215\221\327V\373M\232\354 \t\273\316\236\t!\326\201\211\373\244\021\253\301\223]du=K\232\360\304 \314\266-\233c\206r\340\324eR\310\341W\323\350:\207\321\322D\004C\354\342\342\221\357""\340\036E\260XC\341w\311i[&\212\220\330\330\350c\335`\346\201{\310K\002<p\2164\335Q\032\270\344\330\215[\004Q\r\335u\230\311i\014\347K\304\215\"z\343\230\206\025\210\252\035\332S\251\260\215\237\t\014i[\266\313\371I\323\266Z\244\335GMl8\033\270\353\330\305A\312\005\341\316\276\302\205\345e\266m[\r\360\367IW\324c\324\247\266\306\232D\250\323\030\024\332\010\014\256\272B6\332'\032i\242\006\205\245D\263X\250\363P9fb\271O\210e\363\357\211`\333\2302\031\206FG\361\177`i\230\035\377{\226\036\250\007\314D\331\216\0223\225j\371|\2668\217(\022=\3129=J&\315\020\035\202\257Z\355p)\244\240\216L\022\344\\\216\034\353\226!\nvzAQU\206p\245(a.\247)\210\203\313\013{\274\263\225\236\013\220\333\350/\202\336>b}\304CC\241\tc\373\205\332\341.\333\340\347\275\270\201\370\207\2341%\245<\"\245\234\226sq\255\371)\216y\313\261\337\027\267\324\325[\314\352^0\245khHKz\202*\232\032/\332\252e\302\0037l\204\n)\207\3756*,M\034e;j\005b\263\252\376\362\255\235\377`\353\303\260\333\360:\362\313\304X][e\310\2265\327f\374Z(\007-t\214u\264\r\rl\":\331\312u>\002\361\237\rJ\221\361]\025\365J\210\2444\242b(\020\3319I\027\223\232\200 \216\362\351G \034\240hF\256\300x\233\243\274\206R\344]BI\246+\332S\014\003\242\324#j:)\206cCI\361\363RKy\251\245\274\324&\364\246a\365(:NBA\215\247\274\271%\224\260\232%x\247Mq\303\023\3341\322\260\256\2450[\005w\202\363TC ZiR\213\362i !\204\261\245\210mB\213k\277\240@\200\355\306\341\002)[\231\027\367\371Q\337P\020&\321\365\2611>\200\215P\026Tuh\330\360\005\302gO>~g\223y=\002\347\350\233X\234U\234Z\016\225\246\010\030\206y\177\023Ve;\005Sw%\3345<\035R\360\034\215\223\2533*_\310/_b\2605X\341\371>\255g\251T+\355\026\262\264.\027j\371\210\364\\\312Q9_-\323\274,\307\244B1Gw\353\021\226-\355\226i\265.\225#<'\027\366\363\362\326\004\3668\302\362?\320z\241\366\222>\177U\313\323\235|q\212\314E\355\344k\205R\221\326\n\273\371jM\332-\277\003\013\255K;;\331\235R\366\373i\346\375\035iZGn\247\220\243e\211\033Q\250U\247\026\267wJuZ\223^L\221\013\273\273S\224J\231\026\367\246\211\325\235YBn""\232$g\261y\212X\333\245\205\342v)\242nK5i'\025\356\355|-\373\222J\271X\324\366\316^\365e\004\277\310\027\363\362\004w\241\270Oy\367\233%Q\004\013aO\350\210\030\266b5\301\251\224\315\346\253\325I&\261\271T\236\"\342\370\246(e\271T\233\"!N3\033\321E\351~\241\2246z\267\036&UB\341\0039\362N\312\245\361t^\312\371\354~\n\214\030&OK\316\357R\351yIN[%h\323Nr\"\217\222\314s\2752oa\206:\341\226\270 <\037K{\223\272j\362+\234r6M+\312s\351\325|1\366\265\272'\354\2130$Gr\366@r\371\235\004I\366\003)\226R\271\013<\025\"\216\025s\373\360%\272@\371d\241\372\252\230\215\261j\234\204u9\3143\256\241\314+_\207W\221\004\232\2555\023k3\225\246\034\3162y\336\014#\230\237\030p\323\232\304Q\326\301}\202\237\034\236\341\264\210\036'\263f\364\032\253\242\356\303\310z\266\236\235\265@\320f4\327\325Z\253\200\346\021\177\347\354KVfw3\345\250\312\334\244#\340\r\303\23716k*\374m\245[\033\252ec2\322M\346\360\361'\356\302\321\2373(\025\2234E\375\306\003\264\305Qp\242\233Q\327VT\274\204\324#Q\353y\253\246\321\036\001\363\027\220\000\2161oX6\246\021\027\221P;h\002\0270\303\377\260\347]\000\024oa<\2529\036v\344\024!\034\343cB\370f\212\233fLuO4\326\350\036h\272\230T\025\273?9\023\3043H\374\nFOj\212\375\241\020HHK\303\224\204\247(\245\315\256\251R\232na\002\013#M\347\277\340&\036I\023\357\220x\366\340\023\001v\243\301E\277y\370\251\356\320\213\003\341\021w(B\211p\343_\313\322\272\006\343\220\251\264\3605\205'\032\306\024\274Y\\[g\216\000\3051[\355P3\377\353\033\314\325,.\204\217\3464\034\315\261\363b\004\003k\374\014j(\016\213a\265\363S\217\243\230\036})\013\263\036C>\215b\302\301c\361\314\242\321$\205oW1B\303\305\341%\003\337\317\315z\311\2307g\302\233\030\356&\346\272\324H\227\232\346\242\341\355b\206K\215o\223\223[zh\233\232\327&F\265\213\253Hg.\345\005\201\235p0\036\272\302\211\213O\354\342\262F'\t\010oJ\245k\270\251+L\347\\g\307\340v;\226\370[\022\322\3272\215~\350@20\206Y\346t\305\263\312\351\266\031\216\306\205_\224\342\311\204\227R\342KO\261M\376\303\037\001=5v\252\247:=T\rT\210\036\322Q+\367K""\341\233\341\376Z\002\256\277'n=\215\030\235\177I\343\253\213\243\305\025\177\323\317\005\227\"\354A\260\031\344\206\300\336.\274w\375\367\203\315A\316\277\342W\306W\027NK\203\315\210\347\311P\032\326\3162\021v?X\016\236\004\235\010\373\324_\366\237\370\002\273q\372\344\324\365\036{\312x\341&\200\315\361\342m\357\305\340\251\237\361\227\307\013w\007\313B\370\245\361\302\322h)\322\312\031\035o\305\253\214\027?\036}|\317W\374\316xum\274\260\350}\340)\357,\362\003_\013V\207\267\207\225\024\323\367\203\216\177\331\227\240\314\373\002\334\237\371\312\214\256\321\352\237\317*\221.\376\2711\330\214>\357\254\371C\337\r\236\206\232C=\017\202\33203G\223t\236\231\324$\342}m\341\364+\357}\310^\270\345=\033 \350\213\336Maw>\310\004\237\005\210\300\333\353\357]\313x\277\036T\006\232\277\034E\374s\277\002\207cl9\364.f\375h\240\014\\?\022\177\363\324\034H\203\232\177\007\336\303\241M\021z\034\330h\345\331Y\274\177\325\357\004\227\003I`\221v o\227\260{\364\321\037\323\326\004\035.\344\361\251\352ebA\376fL\372\335@J\023oy\337\301\336K\343\313 =\014\244\361\022\324\214\227\2209\321\257\265\340\322x\t\221\031/m\r\201~}\006t=\310\214C\346\347\347\313\343\3137N\277\365:\203\017\007\"\265n\236\236x\235\361\265?\210@|\036\232\314\251\035\256#\343\335\302\321<\036}\362\345\360\375\341\3468s{t{\325wG\017\376r\246AP\346\016\016\"%\r\033\256\237>\360*\034X\364\360\353\352\351\225\323\n\367\343\251\267\304\317\342\267\203\273\270!\022\0072|\343\346\3502l\216\"B\374K\376m\177\017>L\307b\264\222;\227\306\353\367\203\225`o(|\277\302Cz\305\373\001'\360I\320\035J\343\305;\243;\353\301\235@\021\320\203\340+\334\200\305;^}\260\215\244Z\374\030\311\301\023\346a\260=\\\026\320\243`\037\376p\350\213\340x\370\367\363\014\334\271\261\000\277G7\327\202\017\203\016w7R\224\031e`\3038\303c\232\341\321\315\360pfx\330\303_bA@b\201\237B\010!\235\242\035<\366\031\334\314q\346Q\020\323\304\016\261\360\231\037\257\336\033\335{:T\206\235\320\226\323\316\3502\362\344\355\"\3177\344\331\222\267\374\343\325\217N\017""\275\306\200\337u\357\241\310\310\257\202\367q\345\245\360j|\352I\311\325\370\366L:\253\244\256\014\002\372\033\2345\010=O\361\354Afp\317\377\225\277\206\033\2764\374\346\354\305\371\326ye|\367s_\366\337\014\013g\362Yg|w\331_\362\277\t^\014\267p\017\357\256\370[\376\337\206\327\316\226\3166c\266<\220?\235\257\234\313\347`^\017\226\340\003\314\356\017.\375\033ka\036;";
    PyObject *data = __Pyx_DecompressString(cstring, 2900, 1);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #else /* compression: none (7001 bytes) */
const char* const bytes = ".{:22}: {:<20}\n{:<22}: {:<20}\nBind Memory windowClosing CQClosing CQExClosing completion channelCompare and swapCreated a CQCreated a Completion ChannelDriver WRExtended CQ:\nFailed to close CQFailed to create a CQFailed to create a completion channelFailed to create extended CQFailed to destroy CQEXFailed to destroy a completion channelFailed to get CQ eventFailed to poll CQFailed to resize CQFetch and addIP csum OKLocal invalidateNote that Cython is deliberately stricter than PEP-484 and rejects subclasses of builtin types. If you need to pass subclasses then set the 'annotation_typing' directive to False.Number of CQEsPickling of struct members such as self.wc must be explicitly requested with @auto_pickle(True)Pickling of struct members such as self.attr must be explicitly requested with @auto_pickle(True)Pickling of struct members such as self.info must be explicitly requested with @auto_pickle(True)QP numRDMA readRDMA writeRNR retry exceededReceive RDMA with immediateReceived event on an unexpected CQRequest notify CQ returned {rc}TM data validTM matchTM sync requestTag matching - addTag matching - deleteTag matching - no tagTag matching - receiveTag matching - syncUnknown CQE opcode {op}Unknown CQE statusUnrecognized object typeWC fatal errorWC flagsWR IDWR flush errorWith immediateWith invalidate ?add_notebad response errorbyte lengthcomp maskdisabledlid path bitsenableenum class ibv_create_cq_wc_flags not importable from pyverbs.libibverbs_enums. You are probably using a cpdef enum declared in a .pxd file that does not have a .py  or .pyx file.enum class ibv_wc_opcode not importable from pyverbs.libibverbs_enums. You are probably using a cpdef enum declared in a .pxd file that does not have a .py  or .pyx file.enum class ibv_wc_status not importable from pyverbs.libibverbs_enums. You are probably using a cpdef enum declared in a .pxd file that does not have a .py  or .pyx file.gcgeneral errorinvalidate EECN errorinvalidate EEC state errorisenabledlocal EE""C op errorlocal QP op errorlocal RDD violation errorlocal access errorlocal length errorlocal protection errormemory window bind errorpkey indexpyverbs/cq.pyxremote abort errorremote access errorremote invalidate RD request errorremote invalidate request errorremote op errorresponse timeout errorretry exceeded errorself.attr cannot be converted to a Python object for picklingself.cc cannot be converted to a Python object for picklingself.cq cannot be converted to a Python object for picklingself.cq,self.ibv_cq cannot be converted to a Python object for picklingsource QP(tree fragment)vendor errorCQ\nCQEXCQEX.__reduce_cython__CQEX.__setstate_cython__CQEX.closeCQEX.end_pollCQEX.poll_nextCQEX.read_byte_lenCQEX.read_completion_wallclock_nsCQEX.read_cvlanCQEX.read_dlid_path_bitsCQEX.read_flow_tagCQEX.read_imm_dataCQEX.read_opcodeCQEX.read_qp_numCQEX.read_slCQEX.read_slidCQEX.read_src_qpCQEX.read_timestampCQEX.read_tm_infoCQEX.read_vendor_errCQEX.read_wc_flagsCQEX.start_pollCQEsCQCQ.__reduce_cython__CQ.__setstate_cython__CQ.ack_eventsCQ.closeCQ.pollCQ.req_notifyCQ.resizeCompChannelCompChannel.__reduce_cython__CompChannel.__setstate_cython__CompChannel.closeCompChannel.get_cq_eventCqInitAttrExCqInitAttrEx.__reduce_cython__CqInitAttrEx.__setstate_cython__GRHHandleIBV_WC_ATOMIC_WRITEIBV_WC_BAD_RESP_ERRIBV_WC_BIND_MWIBV_WC_COMP_SWAPIBV_WC_DRIVER2IBV_WC_DRIVER3IBV_WC_EX_WITH_BYTE_LENIBV_WC_EX_WITH_COMPLETION_TIMESTAMPIBV_WC_EX_WITH_COMPLETION_TIMESTAMP_WALLCLOCKIBV_WC_EX_WITH_CVLANIBV_WC_EX_WITH_DLID_PATH_BITSIBV_WC_EX_WITH_FLOW_TAGIBV_WC_EX_WITH_IMMIBV_WC_EX_WITH_QP_NUMIBV_WC_EX_WITH_SLIBV_WC_EX_WITH_SLIDIBV_WC_EX_WITH_SRC_QPIBV_WC_EX_WITH_TM_INFOIBV_WC_FATAL_ERRIBV_WC_FETCH_ADDIBV_WC_FLUSHIBV_WC_GENERAL_ERRIBV_WC_INV_EECN_ERRIBV_WC_INV_EEC_STATE_ERRIBV_WC_LOCAL_INVIBV_WC_LOC_ACCESS_ERRIBV_WC_LOC_EEC_OP_ERRIBV_WC_LOC_LEN_ERRIBV_WC_LOC_PROT_ERRIBV_WC_LOC_QP_OP_ERRIBV_WC_LOC_RDD_VIOL_ERRIBV_WC_MW_BIND_ERRIBV_WC_RDMA_READIBV_WC_RDMA_WRITEIBV_WC_RECVIBV_WC_RECV_RDMA_WITH_IMMIBV_W""C_REM_ABORT_ERRIBV_WC_REM_ACCESS_ERRIBV_WC_REM_INV_RD_REQ_ERRIBV_WC_REM_INV_REQ_ERRIBV_WC_REM_OP_ERRIBV_WC_RESP_TIMEOUT_ERRIBV_WC_RETRY_EXC_ERRIBV_WC_RNR_RETRY_EXC_ERRIBV_WC_SENDIBV_WC_SUCCESSIBV_WC_TM_ADDIBV_WC_TM_DELIBV_WC_TM_ERRIBV_WC_TM_NO_TAGIBV_WC_TM_RECVIBV_WC_TM_RNDV_INCOMPLETEIBV_WC_TM_SYNCIBV_WC_TSOIBV_WC_WR_FLUSH_ERRPollCqAttrPollCqAttr.__reduce_cython__PollCqAttr.__setstate_cython__PyverbsErrorPyverbsRDMAErrnoPyverbsRDMAError__Pyx_PyDict_NextRefReceiveSendTSOWCWC.__reduce_cython__WC.__setstate_cython__WcTmInfoWcTmInfo.__reduce_cython__WcTmInfo.__setstate_cython__WeakSetack_eventsaddadd_refasyncio.coroutinesattrbyte_lenchannel__class_getitem__cline_in_tracebackclosecomp_channelcomp_maskcomp_vectorcontextcqcq_contextcqecqe_flagscqe_flags_to_strcqe_opcode_to_strcqe_status_to_strcreate_wc_flags_to_strctxdebugdictionarydlid_path_bitsend_pollexpected_cqfflagsflags_strflags_to_strformat__func__get_cq_event__getstate__ibv_create_cq_wc_flagsibv_wc_opcodeibv_wc_statusimm_datainfo__init__init_attr_is_coroutineitems__main____module____name__npollednum_entriesnum_eventsopopcodeparent_domainpkey_indexpollpoll_nextpoppyverbs.basepyverbs.cqpyverbs.libibverbs_enumspyverbs.pyverbs_error__pyx_state__pyx_vtable__qp_num__qualname__rcread_byte_lenread_completion_wallclock_nsread_cvlanread_dlid_path_bitsread_flow_tagread_imm_dataread_opcoderead_qp_numread_slread_slidread_src_qpread_timestampread_tm_inforead_vendor_errread_wc_flags__reduce____reduce_cython____reduce_ex__req_notifyresizeself__set_name__setdefault__setstate____setstate_cython__slslidsolicited_onlysrc_qpstart_pollstatussuccesssuper__test__valuesvendor_errwarnwarningswcwc_flagswcsweakrefwr_idPyObject *(PyObject *)\000close_weakrefs\200A\330\010\020\320\020#\2401\240D\250\001\200A\330\010\020\320\020+\2501\250D\260\001\200A\360\014\000\n\034\2301\230D\240\005\240Q\330\010\014\210O\2301\200A\330\010\020\320\0204\260A\260T\270\021\200A\330\010\020\320\020*\250!\2504\250q\200A\330\010\020\320\020\"\240!\2404\240q\200A""\330\010\013\2104\210t\2203\220a\330\014\017\210t\2201\330\020\024\220G\2306\240\021\240!\330\014\032\230!\2301\230D\240\001\330\014\022\320\022+\2501\250D\260\001\330\014\017\210s\220#\220Q\330\020\026\320\026&\240a\240q\330'(\330\014\020\220\006\220a\200A\330\010\013\2104\210t\2203\220a\330\014\017\210t\2201\330\020\024\220G\2306\240\021\240!\330\014\032\230!\2301\230D\240\006\240d\250'\260\024\260Q\330\014\017\210t\2201\330\020\024\220K\230q\240\004\240A\330\014\022\220/\240\021\240$\240a\330\014\017\210s\220#\220Q\330\020\026\320\026&\240a\320'=\270Q\330\014\020\220\006\220a\330\014\020\220\013\2301\330\014\020\220\013\2301\200A\330\010\013\2104\210t\2203\220a\330\014\017\210t\2201\330\020\024\220G\2306\240\021\240!\330\014\032\230!\2301\230D\240\007\240t\2506\260\024\260Q\330\014\022\220/\240\021\240+\250T\260\021\330\014\017\210s\220#\220Q\330\020\026\320\026&\240a\320'A\300\021\330\014\020\220\006\220a\330\014\020\220\013\2301\200A\360\014\000\t\014\2105\220\003\2201\330\014\023\220:\230Q\330\010\020\220\017\230q\240\004\240E\250\021\250$\250a\200A\360\n\000\t\021\220\r\230Q\230d\240!\200A\330\010\020\320\020%\240Q\240d\250!\200A\330\010\020\320\020!\240\021\240$\240a\200A\360\n\000\t\021\220\016\230a\230t\2401\200A\360\014\000\t\017\210n\230A\230T\240\025\240a\330\010\013\2101\330\014\022\320\022\"\240!\320#:\270!\200A\330\010\020\320\020'\240q\250\004\250A\200A\330\010\020\220\017\230q\240\004\240A\200A\360\022\000\t\017\320\016\037\230q\240\004\240E\250\021\250$\250a\250q\330\010\013\2103\210c\220\021\330\014\022\320\022\"\240!\2401\330\010\013\2103\210c\220\033\230A\330\014\022\320\022\"\240!\2401\330\010\023\220?\240!\200\001\330\004\022\320\022,\250A\330\022'\240q\330\022*\250!\330\022*\250!\330\022(\250\001\330\022&\240a\330\0222\260!\330\0228\270\001\330\022)\250\021\330\022,\250A\330\022B\300!\330\004\013\210<\220q\230\007\230q\200A\330\010\017\210x\220q\330\t\035\230Q\230d\240%\240q\250\004\250A\330\010\017\210q\200\001\330\004\021\220\023\220G\2303""\320\0360\260\003\2601\330\021\024\320\024'\240t\320+>\270d\300!\330\021\025\220Q\330\004\013\210<\220q\230\007\230q\200\001\330\004\n\210+\220Q\200\001\330\004\020\220\001\330\004\010\210\005\210Q\330\010\013\2106\220\022\2201\330\014\031\230\032\2401\240A\330\014\031\230\021\330\004\013\2101\320\004*\250!\360\022\000\t\017\320\016 \240\001\240\024\240U\250!\330\010\013\2103\210c\220\021\330\014\022\320\022\"\240!\320#D\300A\330)*\250#\250U\260!\200\001\330\004\005\330\010\020\220\005\220X\230T\240\036\250u\260A\330\020\025\320\025)\250\025\250a\330\020\025\320\025+\2505\260\001\330\020\025\220W\230F\240!\330\020\026\220a\330\020\026\320\026,\250F\260!\330\020\026\320\026-\250V\2601\330\020\026\320\026/\250v\260\\\300\021\300!\330\013\014\330\010\017\320\017(\250\007\250q\260\003\2601\200\001\330\004\005\330\010\021\320\021\"\240!\330\021&\240a\330\021(\250\001\330\021)\250\021\330\021'\240q\330\021'\240q\330\021&\240a\330\021'\240q\330\021)\250\021\330\021*\250!\330\021)\250\021\330\021%\240Q\330\021(\250\001\330\021,\250A\330\021+\2501\330\021-\250Q\330\021(\250\001\330\021'\240q\330\021,\250A\330\021$\240A\330\021+\2501\330\021&\320&6\260a\260q\330\013\014\330\010\017\210q\320\004\037\230q\360\020\000\t\017\210a\330\010\022\220!\340\010\016\210h\220b\230\001\330\014\022\220,\230a\230t\2405\250\003\2501\250A\330\014\017\210s\220\"\220A\330\020\026\320\026&\240a\320'<\270A\270Q\330\014\017\210s\220#\220Q\330\020\021\330\014\027\220q\330\014\017\210w\220a\220r\230\021\230&\240\002\240(\250'\260\022\2609\270G\3002\300Q\330\032%\240R\240}\260I\270R\270q\330\032!\240\022\2409\250G\2602\260Q\330\032#\2402\240[\260\t\270\022\2701\330\032%\240R\240}\260E\270\022\2707\300#\300R\300q\330\032)\250\022\2501\330\010\017\210y\230\001";
    PyObject *data = NULL;
    CYTHON_UNUSED_VAR(__Pyx_DecompressString);
    #endif
    PyObject **stringtab = __pyx_mstate->__pyx_string_tab;
    Py_ssize_t pos = 0;
    for (int i = 0; i < 318; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyUnicode_DecodeUTF8(bytes + pos, bytes_length, NULL);
      if (likely(string) && i >= 94) PyUnicode_InternInPlace(&string);
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
      stringtab[i] = string;
      pos += bytes_length;
    }
    for (int i = 318; i < 346; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyBytes_FromStringAndSize(bytes + pos, bytes_length);
      stringtab[i] = string;
      pos += bytes_length;
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    Py_XDECREF(data);
    for (Py_ssize_t i = 0; i < 346; i++) {
      if (unlikely(PyObject_Hash(stringtab[i]) == -1)) {
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    #if CYTHON_IMMORTAL_CONSTANTS
    {
      PyObject **table = stringtab + 318;
      for (Py_ssize_t i=0; i<28; ++i) {
        #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
        #if PY_VERSION_HEX < 0x030E0000
        if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
        #else
        if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
        #endif
        {
          Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
        }
        #else
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
        #endif
      }
    }
    #endif
  }
  {
    PyObject **numbertab = __pyx_mstate->__pyx_number_tab + 0;
    int8_t const cint_constants_1[] = {0,1,2,3,4,5,6,7,8,16,32,64,100};
    int16_t const cint_constants_2[] = {128,129,130,131,132,133,134,135};
    for (int i = 0; i < 21; i++) {
      numbertab[i] = PyLong_FromLong((i < 13 ? cint_constants_1[i - 0] : cint_constants_2[i - 13]));
      if (unlikely(!numbertab[i])) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_number_tab;
    for (Py_ssize_t i=0; i<21; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_codeobjects ### */
typedef struct {
    unsigned int argcount : 2;
    unsigned int num_posonly_args : 1;
    unsigned int num_kwonly_args : 1;
    unsigned int nlocals : 3;
    unsigned int flags : 10;
    unsigned int first_line : 10;
} __Pyx_PyCode_New_function_description;
/* NewCodeObj.proto */
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
);


static int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate) {
  PyObject* tuple_dedup_map = PyDict_New();
  if (unlikely(!tuple_dedup_map)) return -1;
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 42};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[0] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_close, __pyx_mstate->__pyx_kp_b_iso88591_A_4t3a_t1_G6_1D_1D_s_Q_aq_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[0])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 53};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_expected_cq, __pyx_mstate->__pyx_n_u_cq, __pyx_mstate->__pyx_n_u_ctx, __pyx_mstate->__pyx_n_u_rc};
    __pyx_mstate_global->__pyx_codeobj_tab[1] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_get_cq_event, __pyx_mstate->__pyx_kp_b_iso88591_A_q_E_aq_3c_1_3c_A_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[1])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[2] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[2])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[3] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[3])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 129};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[4] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_close, __pyx_mstate->__pyx_kp_b_iso88591_A_4t3a_t1_G6_1D_d_Q_t1_Kq_A_a_s, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[4])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 143};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_cqe, __pyx_mstate->__pyx_n_u_rc};
    __pyx_mstate_global->__pyx_codeobj_tab[5] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_resize, __pyx_mstate->__pyx_kp_b_iso88591_A_nAT_a_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[5])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 153};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_num_entries, __pyx_mstate->__pyx_n_u_wc, __pyx_mstate->__pyx_n_u_wcs, __pyx_mstate->__pyx_n_u_npolled, __pyx_mstate->__pyx_n_u_rc};
    __pyx_mstate_global->__pyx_codeobj_tab[6] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_poll, __pyx_mstate->__pyx_kp_b_iso88591_q_a_hb_at5_1A_s_A_a_AQ_s_Q_q_wa, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[6])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 179};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_solicited_only, __pyx_mstate->__pyx_n_u_rc};
    __pyx_mstate_global->__pyx_codeobj_tab[7] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_req_notify, __pyx_mstate->__pyx_kp_b_iso88591_U_3c_DA_U, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[7])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 193};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_num_events};
    __pyx_mstate_global->__pyx_codeobj_tab[8] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_ack_events, __pyx_mstate->__pyx_kp_b_iso88591_A_1D_Q_O1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[8])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[9] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[9])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[10] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[10])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[11] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[11])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[12] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[12])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 357};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[13] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_close, __pyx_mstate->__pyx_kp_b_iso88591_A_4t3a_t1_G6_1D_t6_Q_T_s_Q_a_A_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[13])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 368};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_attr};
    __pyx_mstate_global->__pyx_codeobj_tab[14] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_start_poll, __pyx_mstate->__pyx_kp_b_iso88591_A_5_1_Q_q_E_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[14])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 378};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[15] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_poll_next, __pyx_mstate->__pyx_kp_b_iso88591_A_at1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[15])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 385};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[16] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_end_poll, __pyx_mstate->__pyx_kp_b_iso88591_A_Qd, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[16])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 392};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[17] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_read_opcode, __pyx_mstate->__pyx_kp_b_iso88591_A_1D, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[17])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 394};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[18] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_read_vendor_err, __pyx_mstate->__pyx_kp_b_iso88591_A_q_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[18])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 396};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[19] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_read_byte_len, __pyx_mstate->__pyx_kp_b_iso88591_A_Qd_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[19])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 398};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[20] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_read_imm_data, __pyx_mstate->__pyx_kp_b_iso88591_A_Qd_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[20])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 400};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[21] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_read_qp_num, __pyx_mstate->__pyx_kp_b_iso88591_A_1D, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[21])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 402};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[22] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_read_src_qp, __pyx_mstate->__pyx_kp_b_iso88591_A_1D, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[22])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 404};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[23] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_read_wc_flags, __pyx_mstate->__pyx_kp_b_iso88591_A_Qd_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[23])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 406};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[24] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_read_slid, __pyx_mstate->__pyx_kp_b_iso88591_A_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[24])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 408};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[25] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_read_sl, __pyx_mstate->__pyx_kp_b_iso88591_A_q_A_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[25])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 410};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[26] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_read_dlid_path_bits, __pyx_mstate->__pyx_kp_b_iso88591_A_1D_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[26])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 412};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[27] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_read_timestamp, __pyx_mstate->__pyx_kp_b_iso88591_A_4q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[27])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 414};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[28] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_read_cvlan, __pyx_mstate->__pyx_kp_b_iso88591_A_4q_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[28])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 416};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[29] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_read_flow_tag, __pyx_mstate->__pyx_kp_b_iso88591_A_Qd_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[29])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 418};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_info};
    __pyx_mstate_global->__pyx_codeobj_tab[30] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_read_tm_info, __pyx_mstate->__pyx_kp_b_iso88591_A_xq_Qd_q_A_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[30])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 422};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[31] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_read_completion_wallclock_ns, __pyx_mstate->__pyx_kp_b_iso88591_A_4AT, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[31])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[32] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[32])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[33] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[33])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[34] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[34])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[35] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[35])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[36] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[36])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[37] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[37])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[38] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[38])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[39] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[39])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 598};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_status};
    __pyx_mstate_global->__pyx_codeobj_tab[40] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_cqe_status_to_str, __pyx_mstate->__pyx_kp_b_iso88591_a_q_q_a_q_Q_A_1_Q_q_A_A_1_6aq_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[40])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 625};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_opcode};
    __pyx_mstate_global->__pyx_codeobj_tab[41] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_cqe_opcode_to_str, __pyx_mstate->__pyx_kp_b_iso88591_XT_uA_a_5_WF_a_F_V1_v_q_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[41])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 638};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_flags, __pyx_mstate->__pyx_n_u_dictionary, __pyx_mstate->__pyx_n_u_flags_str, __pyx_mstate->__pyx_n_u_f};
    __pyx_mstate_global->__pyx_codeobj_tab[42] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_flags_to_str, __pyx_mstate->__pyx_kp_b_iso88591_Q_6_1_1A_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[42])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 647};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_flags, __pyx_mstate->__pyx_n_u_cqe_flags};
    __pyx_mstate_global->__pyx_codeobj_tab[43] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_cqe_flags_to_str, __pyx_mstate->__pyx_kp_b_iso88591_G3_0_1_t_d_Q_q_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[43])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 653};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_flags, __pyx_mstate->__pyx_n_u_cqe_flags};
    __pyx_mstate_global->__pyx_codeobj_tab[44] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cq_pyx, __pyx_mstate->__pyx_n_u_create_wc_flags_to_str, __pyx_mstate->__pyx_kp_b_iso88591_A_q_a_2_8_A_B_q_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[44])) goto bad;
  }
  Py_DECREF(tuple_dedup_map);
  return 0;
  bad:
  Py_DECREF(tuple_dedup_map);
  return -1;
}
/* #### Code section: init_globals ### */

static int __Pyx_InitGlobals(void) {
  /* PythonCompatibility.init */
  if (likely(__Pyx_init_co_variables() == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CommonTypesMetaclass.init */
  if (likely(__pyx_CommonTypesMetaclass_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CachedMethodType.init */
  #if CYTHON_COMPILING_IN_LIMITED_API
  {
      PyObject *typesModule=NULL;
      typesModule = PyImport_ImportModule("types");
      if (typesModule) {
          __pyx_mstate_global->__Pyx_CachedMethodType = PyObject_GetAttrString(typesModule, "MethodType");
          Py_DECREF(typesModule);
      }
  } // error handling follows
  #endif
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CythonFunctionShared.init */
  if (likely(__pyx_CyFunction_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError (used by GetBuiltinName) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_mstate_global->__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
            "name '%U' is not defined", name);
    }
    return result;
}

/* HasAttr (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!PyUnicode_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_PyObject_GetAttrStrNoError(o, n);
    if (!r) {
        return (unlikely(PyErr_Occurred())) ? -1 : 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}
#endif

/* PyObjectCall (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall (used by PyObjectCallOneArg) */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) != (0)) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
#if CYTHON_VECTORCALL && !CYTHON_COMPILING_IN_LIMITED_API
  #if PY_VERSION_HEX < 0x03090000
    #define __Pyx_PyVectorcall_Function(callable) _PyVectorcall_Function(callable)
  #elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE vectorcallfunc __Pyx_PyVectorcall_Function(PyObject *callable) {
    PyTypeObject *tp = Py_TYPE(callable);
    #if defined(__Pyx_CyFunction_USED)
    if (__Pyx_CyFunction_CheckExact(callable)) {
        return __Pyx_CyFunction_func_vectorcall(callable);
    }
    #endif
    if (!PyType_HasFeature(tp, Py_TPFLAGS_HAVE_VECTORCALL)) {
        return NULL;
    }
    assert(PyCallable_Check(callable));
    Py_ssize_t offset = tp->tp_vectorcall_offset;
    assert(offset > 0);
    vectorcallfunc ptr;
    memcpy(&ptr, (char *) callable + offset, sizeof(ptr));
    return ptr;
}
  #else
    #define __Pyx_PyVectorcall_Function(callable) PyVectorcall_Function(callable)
  #endif
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject *const *args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
          #if CYTHON_COMPILING_IN_LIMITED_API
            return PyObject_Vectorcall(func, args, _nargs, NULL);
          #else
            vectorcallfunc f = __Pyx_PyVectorcall_Function(func);
            if (f) {
                return f(func, args, _nargs, NULL);
            }
          #endif
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_mstate_global->__pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* PyObjectCallOneArg (used by ImportImpl) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* TupleAndListFromArray (used by ImportImpl) */
#if !CYTHON_COMPILING_IN_CPYTHON && CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    Py_ssize_t i;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    for (i = 0; i < n; i++) {
        if (unlikely(__Pyx_PyTuple_SET_ITEM(res, i, src[i]) < (0))) {
            Py_DECREF(res);
            return NULL;
        }
        Py_INCREF(src[i]);
    }
    return res;
}
#elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* ImportImpl (used by Import) */
static int __Pyx__Import_GetModule(PyObject *qualname, PyObject **module) {
    PyObject *imported_module = PyImport_GetModule(qualname);
    if (unlikely(!imported_module)) {
        *module = NULL;
        if (PyErr_Occurred()) {
            return -1;
        }
        return 0;
    }
    *module = imported_module;
    return 1;
}
static int __Pyx__Import_Lookup(PyObject *qualname, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject **module) {
    PyObject *imported_module;
    PyObject *top_level_package_name;
    Py_ssize_t i;
    int status, module_found;
    Py_ssize_t dot_index;
    module_found = __Pyx__Import_GetModule(qualname, &imported_module);
    if (unlikely(!module_found || module_found == -1)) {
        *module = NULL;
        return module_found;
    }
    if (imported_names) {
        for (i = 0; i < len_imported_names; i++) {
            PyObject *imported_name = imported_names[i];
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
            int has_imported_attribute = PyObject_HasAttr(imported_module, imported_name);
#else
            int has_imported_attribute = PyObject_HasAttrWithError(imported_module, imported_name);
            if (unlikely(has_imported_attribute == -1)) goto error;
#endif
            if (!has_imported_attribute) {
                goto not_found;
            }
        }
        *module = imported_module;
        return 1;
    }
    dot_index = PyUnicode_FindChar(qualname, '.', 0, PY_SSIZE_T_MAX, 1);
    if (dot_index == -1) {
        *module = imported_module;
        return 1;
    }
    if (unlikely(dot_index == -2)) goto error;
    top_level_package_name = PyUnicode_Substring(qualname, 0, dot_index);
    if (unlikely(!top_level_package_name)) goto error;
    Py_DECREF(imported_module);
    status = __Pyx__Import_GetModule(top_level_package_name, module);
    Py_DECREF(top_level_package_name);
    return status;
error:
    Py_DECREF(imported_module);
    *module = NULL;
    return -1;
not_found:
    Py_DECREF(imported_module);
    *module = NULL;
    return 0;
}
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *from_list = 0;
    int module_found;
    if (!qualname) {
        qualname = name;
    }
    module_found = __Pyx__Import_Lookup(qualname, imported_names, len_imported_names, &module);
    if (likely(module_found == 1)) {
        return module;
    } else if (unlikely(module_found == -1)) {
        return NULL;
    }
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    if (imported_names) {
#if CYTHON_COMPILING_IN_CPYTHON
        from_list = __Pyx_PyList_FromArray(imported_names, len_imported_names);
        if (unlikely(!from_list))
            goto bad;
#else
        from_list = PyList_New(len_imported_names);
        if (unlikely(!from_list)) goto bad;
        for (Py_ssize_t i=0; i<len_imported_names; ++i) {
            if (PyList_SetItem(from_list, i, __Pyx_NewRef(imported_names[i])) < 0) goto bad;
        }
#endif
    }
    if (level == -1) {
        const char* package_sep = strchr(__Pyx_MODULE_NAME, '.');
        if (package_sep != (0)) {
            module = PyImport_ImportModuleLevelObject(
                name, moddict, empty_dict, from_list, 1);
            if (unlikely(!module)) {
                if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                    goto bad;
                PyErr_Clear();
            }
        }
        level = 0;
    }
    if (!module) {
        module = PyImport_ImportModuleLevelObject(
            name, moddict, empty_dict, from_list, level);
    }
bad:
    Py_XDECREF(from_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level) {
    return __Pyx__Import(name, imported_names, len_imported_names, qualname, __pyx_mstate_global->__pyx_d, level);
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_mstate_global->__pyx_kp_u_);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400) ||\
                CYTHON_COMPILING_IN_GRAAL
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError, "cannot import name %S", name);
    }
    return value;
}

/* GetTopmostException (used by SaveResetException) */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C0000
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#elif __PYX_LIMITED_VERSION_HEX > 0x030C0000
    local_value = PyErr_GetRaisedException();
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
#if __PYX_LIMITED_VERSION_HEX > 0x030C0000
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
#else
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
#endif // __PYX_LIMITED_VERSION_HEX > 0x030C0000
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#elif __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    PyErr_SetHandledException(local_value);
    Py_XDECREF(local_value);
    Py_XDECREF(local_type);
    Py_XDECREF(local_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
#if __PYX_LIMITED_VERSION_HEX <= 0x030C0000
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
#endif
}

/* PyObjectFastCallMethod */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf) {
    PyObject *result;
    PyObject *attr = PyObject_GetAttr(args[0], name);
    if (unlikely(!attr))
        return NULL;
    result = __Pyx_PyObject_FastCall(attr, args+1, nargsf - 1);
    Py_DECREF(attr);
    return result;
}
#endif

/* BytesEquals (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL ||\
        !(CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS)
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length, length2;
        int kind;
        void *data1, *data2;
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        #endif
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length < 0)) return -1;
        #endif
        length2 = __Pyx_PyUnicode_GET_LENGTH(s2);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length2 < 0)) return -1;
        #endif
        if (length != length2) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    return (equals == Py_EQ);
return_ne:
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = __Pyx_PyTuple_GET_SIZE(kwnames);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(n == -1)) return NULL;
    #endif
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        if (s == namei) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        int eq = __Pyx_PyUnicode_Equals(s, namei, Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs;
    PyObject *dict;
#if !CYTHON_ASSUME_SAFE_SIZE
    nkwargs = PyTuple_Size(kwnames);
    if (unlikely(nkwargs < 0)) return NULL;
#else
    nkwargs = PyTuple_GET_SIZE(kwnames);
#endif
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
#if !CYTHON_ASSUME_SAFE_MACROS
        PyObject *key = PyTuple_GetItem(kwnames, i);
        if (!key) goto bad;
#else
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
#endif
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* UnpackUnboundCMethod (used by CallUnboundCMethod0) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
#elif CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject **args, Py_ssize_t nargs, PyObject *kwnames) {
        return _PyObject_Vectorcall
            (method, args ? args+1 : NULL, nargs ? nargs-1 : 0, kwnames);
}
#else
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *const *args, Py_ssize_t nargs, PyObject *kwnames) {
    return
#if PY_VERSION_HEX < 0x03090000
    _PyObject_Vectorcall
#else
    PyObject_Vectorcall
#endif
        (method, args ? args+1 : NULL, nargs ? (size_t) nargs-1 : 0, kwnames);
}
#endif
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
     METH_VARARGS | METH_KEYWORDS,
#else
     METH_FASTCALL | METH_KEYWORDS,
#endif
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method, *result=NULL;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    result = method;
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            result = unbound_method;
        }
    }
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    if (unlikely(target->method)) {
        Py_DECREF(result);
    } else
#endif
    target->method = result;
    return 0;
}

/* CallUnboundCMethod0 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        if (likely(cfunc->flag == METH_NOARGS))
            return __Pyx_CallCFunction(cfunc, self, NULL);
        if (likely(cfunc->flag == METH_FASTCALL))
            return __Pyx_CallCFunctionFast(cfunc, self, NULL, 0);
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, NULL, 0, NULL);
        if (likely(cfunc->flag == (METH_VARARGS | METH_KEYWORDS)))
            return __Pyx_CallCFunctionWithKeywords(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple, NULL);
        if (cfunc->flag == METH_VARARGS)
            return __Pyx_CallCFunction(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple);
        return __Pyx__CallUnboundCMethod0(cfunc, self);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod0(&tmp_cfunc, self);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod0(cfunc, self);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *result;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
    result = __Pyx_PyObject_CallOneArg(cfunc->method, self);
    return result;
}

/* py_dict_items (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_items, d);
}

/* py_dict_values (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_values, d);
}

/* OwnedDictNext (used by ParseKeywordsImpl) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue) {
    PyObject *next = NULL;
    if (!*ppos) {
        if (pvalue) {
            PyObject *dictview = pkey ? __Pyx_PyDict_Items(p) : __Pyx_PyDict_Values(p);
            if (unlikely(!dictview)) goto bad;
            *ppos = PyObject_GetIter(dictview);
            Py_DECREF(dictview);
        } else {
            *ppos = PyObject_GetIter(p);
        }
        if (unlikely(!*ppos)) goto bad;
    }
    next = PyIter_Next(*ppos);
    if (!next) {
        if (PyErr_Occurred()) goto bad;
        return 0;
    }
    if (pkey && pvalue) {
        *pkey = __Pyx_PySequence_ITEM(next, 0);
        if (unlikely(*pkey)) goto bad;
        *pvalue = __Pyx_PySequence_ITEM(next, 1);
        if (unlikely(*pvalue)) goto bad;
        Py_DECREF(next);
    } else if (pkey) {
        *pkey = next;
    } else {
        assert(pvalue);
        *pvalue = next;
    }
    return 1;
  bad:
    Py_XDECREF(next);
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
    PyErr_FormatUnraisable("Exception ignored in __Pyx_PyDict_NextRef");
#else
    PyErr_WriteUnraisable(__pyx_mstate_global->__pyx_n_u_Pyx_PyDict_NextRef);
#endif
    if (pkey) *pkey = NULL;
    if (pvalue) *pvalue = NULL;
    return 0;
}
#else // !CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue) {
    int result = PyDict_Next(p, ppos, pkey, pvalue);
    if (likely(result == 1)) {
        if (pkey) Py_INCREF(*pkey);
        if (pvalue) Py_INCREF(*pvalue);
    }
    return result;
}
#endif

/* RaiseDoubleKeywords (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            return __Pyx_CallCFunctionFast(cfunc, self, args, 2);
        }
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, 2, NULL);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod2(&tmp_cfunc, self, arg1, arg2);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        PyObject *result = NULL;
        PyObject *args = PyTuple_New(2);
        if (unlikely(!args)) return NULL;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = __Pyx_CallCFunctionWithKeywords(cfunc, self, args, NULL);
        else
            result = __Pyx_CallCFunction(cfunc, self, args);
        Py_DECREF(args);
        return result;
    }
#endif
    {
        PyObject *args[4] = {NULL, self, arg1, arg2};
        return __Pyx_PyObject_FastCall(cfunc->method, args+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
    }
}

/* ParseKeywordsImpl (used by ParseKeywords) */
static int __Pyx_ValidateDuplicatePosArgs(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char* function_name)
{
    PyObject ** const *name = argnames;
    while (name != first_kw_arg) {
        PyObject *key = **name;
        int found = PyDict_Contains(kwds, key);
        if (unlikely(found)) {
            if (found == 1) __Pyx_RaiseDoubleKeywordsError(function_name, key);
            goto bad;
        }
        name++;
    }
    return 0;
bad:
    return -1;
}
#if CYTHON_USE_UNICODE_INTERNALS
static CYTHON_INLINE int __Pyx_UnicodeKeywordsEqual(PyObject *s1, PyObject *s2) {
    int kind;
    Py_ssize_t len = PyUnicode_GET_LENGTH(s1);
    if (len != PyUnicode_GET_LENGTH(s2)) return 0;
    kind = PyUnicode_KIND(s1);
    if (kind != PyUnicode_KIND(s2)) return 0;
    const void *data1 = PyUnicode_DATA(s1);
    const void *data2 = PyUnicode_DATA(s2);
    return (memcmp(data1, data2, (size_t) len * (size_t) kind) == 0);
}
#endif
static int __Pyx_MatchKeywordArg_str(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    #if CYTHON_USE_UNICODE_INTERNALS
    Py_hash_t key_hash = ((PyASCIIObject*)key)->hash;
    if (unlikely(key_hash == -1)) {
        key_hash = PyObject_Hash(key);
        if (unlikely(key_hash == -1))
            goto bad;
    }
    #endif
    name = first_kw_arg;
    while (*name) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (key_hash == ((PyASCIIObject*)name_str)->hash && __Pyx_UnicodeKeywordsEqual(name_str, key)) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) {
                *index_found = (size_t) (name - argnames);
                return 1;
            }
        }
        #endif
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (unlikely(key_hash == ((PyASCIIObject*)name_str)->hash)) {
            if (__Pyx_UnicodeKeywordsEqual(name_str, key))
                goto arg_passed_twice;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            if (unlikely(name_str == key)) goto arg_passed_twice;
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) goto arg_passed_twice;
        }
        #endif
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
bad:
    return -1;
}
static int __Pyx_MatchKeywordArg_nostr(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    if (unlikely(!PyUnicode_Check(key))) goto invalid_keyword_type;
    name = first_kw_arg;
    while (*name) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (cmp == 1) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        if (unlikely(cmp == -1)) goto bad;
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (unlikely(cmp != 0)) {
            if (cmp == 1) goto arg_passed_twice;
            else goto bad;
        }
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
bad:
    return -1;
}
static CYTHON_INLINE int __Pyx_MatchKeywordArg(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    return likely(PyUnicode_CheckExact(key)) ?
        __Pyx_MatchKeywordArg_str(key, argnames, first_kw_arg, index_found, function_name) :
        __Pyx_MatchKeywordArg_nostr(key, argnames, first_kw_arg, index_found, function_name);
}
static void __Pyx_RejectUnknownKeyword(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char *function_name)
{
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos = NULL;
    #else
    Py_ssize_t pos = 0;
    #endif
    PyObject *key = NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(kwds);
    while (
        #if CYTHON_AVOID_BORROWED_REFS
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL)
        #else
        PyDict_Next(kwds, &pos, &key, NULL)
        #endif
    ) {
        PyObject** const *name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (!*name) {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp != 1) {
                if (cmp == 0) {
                    PyErr_Format(PyExc_TypeError,
                        "%s() got an unexpected keyword argument '%U'",
                        function_name, key);
                }
                #if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(key);
                #endif
                break;
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        #endif
    }
    __Pyx_END_CRITICAL_SECTION();
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(pos);
    #endif
    assert(PyErr_Occurred());
}
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t extracted = 0;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    name = first_kw_arg;
    while (*name && num_kwargs > extracted) {
        PyObject * key = **name;
        PyObject *value;
        int found = 0;
        #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        found = PyDict_GetItemRef(kwds, key, &value);
        #else
        value = PyDict_GetItemWithError(kwds, key);
        if (value) {
            Py_INCREF(value);
            found = 1;
        } else {
            if (unlikely(PyErr_Occurred())) goto bad;
        }
        #endif
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            extracted++;
        }
        name++;
    }
    if (num_kwargs > extracted) {
        if (ignore_unknown_kwargs) {
            if (unlikely(__Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name) == -1))
                goto bad;
        } else {
            __Pyx_RejectUnknownKeyword(kwds, argnames, first_kw_arg, function_name);
            goto bad;
        }
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t len;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    if (PyDict_Update(kwds2, kwds) < 0) goto bad;
    name = first_kw_arg;
    while (*name) {
        PyObject *key = **name;
        PyObject *value;
#if !CYTHON_COMPILING_IN_LIMITED_API && (PY_VERSION_HEX >= 0x030d00A2 || defined(PyDict_Pop))
        int found = PyDict_Pop(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
        }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        int found = PyDict_GetItemRef(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            if (unlikely(PyDict_DelItem(kwds2, key) < 0)) goto bad;
        }
#else
    #if CYTHON_COMPILING_IN_CPYTHON
        value = _PyDict_Pop(kwds2, key, kwds2);
    #else
        value = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_pop, kwds2, key, kwds2);
    #endif
        if (value == kwds2) {
            Py_DECREF(value);
        } else {
            if (unlikely(!value)) goto bad;
            values[name-argnames] = value;
        }
#endif
        name++;
    }
    len = PyDict_Size(kwds2);
    if (len > 0) {
        return __Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name);
    } else if (unlikely(len == -1)) {
        goto bad;
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject *key = NULL;
    PyObject** const * name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    for (Py_ssize_t pos = 0; pos < num_kwargs; pos++) {
#if CYTHON_AVOID_BORROWED_REFS
        key = __Pyx_PySequence_ITEM(kwds, pos);
#else
        key = __Pyx_PyTuple_GET_ITEM(kwds, pos);
#endif
#if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!key)) goto bad;
#endif
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            PyObject *value = kwvalues[pos];
            values[name-argnames] = __Pyx_NewRef(value);
        } else {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp == 1) {
                PyObject *value = kwvalues[pos];
                values[index_found] = __Pyx_NewRef(value);
            } else {
                if (unlikely(cmp == -1)) goto bad;
                if (kwds2) {
                    PyObject *value = kwvalues[pos];
                    if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
                } else if (!ignore_unknown_kwargs) {
                    goto invalid_keyword;
                }
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        key = NULL;
        #endif
    }
    return 0;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    goto bad;
bad:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(key);
    #endif
    return -1;
}

/* ParseKeywords */
static int __Pyx_ParseKeywords(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds)))
        return __Pyx_ParseKeywordsTuple(kwds, kwvalues, argnames, kwds2, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
    else if (kwds2)
        return __Pyx_ParseKeywordDictToDict(kwds, argnames, kwds2, values, num_pos_args, function_name);
    else
        return __Pyx_ParseKeywordDict(kwds, argnames, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* ArgTypeTestFunc (used by ArgTypeTest) */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    PyObject *extra_info = __pyx_mstate_global->__pyx_empty_unicode;
    int from_annotation_subclass = 0;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (!exact) {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    } else if (exact == 2) {
        if (__Pyx_TypeCheck(obj, type)) {
            from_annotation_subclass = 1;
            extra_info = __pyx_mstate_global->__pyx_kp_u_Note_that_Cython_is_deliberately;
        }
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(type);
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")"
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        "%s%U"
#endif
        , name, type_name, obj_type_name
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        , (from_annotation_subclass ? ". " : ""), extra_info
#endif
        );
#if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    if (exact == 2 && from_annotation_subclass) {
        PyObject *res;
        PyObject *vargs[2];
        vargs[0] = PyErr_GetRaisedException();
        vargs[1] = extra_info;
        res = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_kp_u_add_note, vargs, 2, NULL);
        Py_XDECREF(res);
        PyErr_SetRaisedException(vargs[0]);
    }
#endif
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* PyDictVersioning (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        if (!PyErr_Occurred())
            PyErr_SetNone(PyExc_NameError);
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
    PyErr_Clear();
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    if (unlikely(__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, name, &result) == -1)) PyErr_Clear();
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return result;
    }
#else
    result = _PyDict_GetItem_KnownHash(__pyx_mstate_global->__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* RaiseException */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
#elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
    CYTHON_UNUSED_VAR(clineno);
    CYTHON_UNUSED_VAR(lineno);
    CYTHON_UNUSED_VAR(filename);
    CYTHON_MAYBE_UNUSED_VAR(nogil);
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(0);
    }
    ctx = PyUnicode_FromString(name);
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
    if (nogil)
        PyGILState_Release(state);
}

/* RejectKeywords */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds) {
    PyObject *key = NULL;
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds))) {
        key = __Pyx_PySequence_ITEM(kwds, 0);
    } else {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *pos = NULL;
#else
        Py_ssize_t pos = 0;
#endif
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
        if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return;
#endif
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL);
#if CYTHON_AVOID_BORROWED_REFS
        Py_XDECREF(pos);
#endif
    }
    if (likely(key)) {
        PyErr_Format(PyExc_TypeError,
            "%s() got an unexpected keyword argument '%U'",
            function_name, key);
        Py_DECREF(key);
    }
}

/* PyLongBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_Fallback___Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, int inplace) {
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject* __Pyx_Unpacked___Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long b = intval;
    long a;
    const PY_LONG_LONG llb = intval;
    PY_LONG_LONG lla;
    if (unlikely(__Pyx_PyLong_IsZero(op1))) {
        return __Pyx_NewRef(op2);
    }
    const int is_positive = __Pyx_PyLong_IsPos(op1);
    const digit* digits = __Pyx_PyLong_Digits(op1);
    const Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
    if (likely(size == 1)) {
        a = (long) digits[0];
        if (!is_positive) a *= -1;
    } else {
        switch (size) {
            case 2:
                if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                    a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) a *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) lla *= -1;
                    goto calculate_long_long;
                }
                break;
            case 3:
                if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                    a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) a *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) lla *= -1;
                    goto calculate_long_long;
                }
                break;
            case 4:
                if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                    a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    if (!is_positive) a *= -1;
                    goto calculate_long;
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    if (!is_positive) lla *= -1;
                    goto calculate_long_long;
                }
                break;
        }
        return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    calculate_long:
        {
            long x;
            x = a + b;
            return PyLong_FromLong(x);
        }
    calculate_long_long:
        {
            PY_LONG_LONG llx;
            llx = lla + llb;
            return PyLong_FromLongLong(llx);
        }
    
}
#endif
static PyObject* __Pyx_Float___Pyx_PyLong_AddObjC(PyObject *float_val, long intval, int zerodivision_check) {
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long b = intval;
    double a = __Pyx_PyFloat_AS_DOUBLE(float_val);
        double result;
        
        result = ((double)a) + (double)b;
        return PyFloat_FromDouble(result);
}
static CYTHON_INLINE PyObject* __Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        return __Pyx_Unpacked___Pyx_PyLong_AddObjC(op1, op2, intval, inplace, zerodivision_check);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        return __Pyx_Float___Pyx_PyLong_AddObjC(op1, intval, zerodivision_check);
    }
    return __Pyx_Fallback___Pyx_PyLong_AddObjC(op1, op2, inplace);
}
#endif

/* PyLongCompare */
static CYTHON_INLINE int __Pyx_PyLong_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        return 1;
    }
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            return (__Pyx_PyLong_IsZero(op1) == 1);
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                return 0;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                return 0;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        return (unequal == 0);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = __Pyx_PyFloat_AS_DOUBLE(op1);
        return ((double)a == (double)b);
    }
    return __Pyx_PyObject_IsTrueAndDecref(
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* PyObjectVectorCallKwBuilder */
#if CYTHON_VECTORCALL
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_PyObject_FastCallDict;
    if (__Pyx_PyTuple_SET_ITEM(builder, n, key) != (0)) return -1;
    Py_INCREF(key);
    args[n] = value;
    return 0;
}
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_VectorcallBuilder_AddArgStr;
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n);
}
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    PyObject *pyKey = PyUnicode_FromString(key);
    if (!pyKey) return -1;
    return __Pyx_VectorcallBuilder_AddArg(pyKey, value, builder, args, n);
}
#else // CYTHON_VECTORCALL
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, CYTHON_UNUSED PyObject **args, CYTHON_UNUSED int n) {
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return PyDict_SetItem(builder, key, value);
}
#endif

/* PyObjectVectorCallMethodKwBuilder */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_Object_VectorcallMethod_CallFromBuilder(PyObject *name, PyObject *const *args, size_t nargsf, PyObject *kwnames) {
    PyObject *result;
    PyObject *obj = PyObject_GetAttr(args[0], name);
    if (unlikely(!obj))
        return NULL;
    result = __Pyx_Object_Vectorcall_CallFromBuilder(obj, args+1, nargsf-1, kwnames);
    Py_DECREF(obj);
    return result;
}
#endif

/* UnicodeConcatInPlace */
# if CYTHON_COMPILING_IN_CPYTHON
static int
__Pyx_unicode_modifiable(PyObject *unicode, int unsafe_shared)
{
    if (!__Pyx_IS_UNIQUELY_REFERENCED(unicode, unsafe_shared))
        return 0;
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030F0000
    if (PyUnstable_Unicode_GET_CACHED_HASH(unicode) != -1)
        return 0;
#endif
    if (!PyUnicode_CheckExact(unicode))
        return 0;
    if (PyUnicode_CHECK_INTERNED(unicode))
        return 0;
    return 1;
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right, int unsafe_shared
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    ) {
    PyObject *left = *p_left;
    Py_ssize_t left_len, right_len, new_len;
    if (unlikely(__Pyx_PyUnicode_READY(left) == -1))
        return NULL;
    if (unlikely(__Pyx_PyUnicode_READY(right) == -1))
        return NULL;
    left_len = PyUnicode_GET_LENGTH(left);
    if (left_len == 0) {
        Py_INCREF(right);
        return right;
    }
    right_len = PyUnicode_GET_LENGTH(right);
    if (right_len == 0) {
        Py_INCREF(left);
        return left;
    }
    if (unlikely(left_len > PY_SSIZE_T_MAX - right_len)) {
        PyErr_SetString(PyExc_OverflowError,
                        "strings are too large to concat");
        return NULL;
    }
    new_len = left_len + right_len;
    if (__Pyx_unicode_modifiable(left, unsafe_shared)
            && PyUnicode_CheckExact(right)
            && PyUnicode_KIND(right) <= PyUnicode_KIND(left)
            && !(PyUnicode_IS_ASCII(left) && !PyUnicode_IS_ASCII(right))) {
        int ret;
        __Pyx_GIVEREF(*p_left);
        ret = PyUnicode_Resize(p_left, new_len);
        __Pyx_GOTREF(*p_left);
        if (unlikely(ret != 0))
            return NULL;
        #if PY_VERSION_HEX >= 0x030d0000
        if (unlikely(PyUnicode_CopyCharacters(*p_left, left_len, right, 0, right_len) < 0)) return NULL;
        #else
        _PyUnicode_FastCopyCharacters(*p_left, left_len, right, 0, right_len);
        #endif
        __Pyx_INCREF(*p_left);
        __Pyx_GIVEREF(*p_left);
        return *p_left;
    } else {
        return __Pyx_PyUnicode_Concat(left, right);
    }
  }
#endif

/* DictGetItem */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    if (unlikely(__Pyx_PyDict_GetItemRef(d, key, &value) == 0)) { // no value, no error
        if (unlikely(PyTuple_Check(key))) {
            PyObject* args = PyTuple_Pack(1, key);
            if (likely(args)) {
                PyErr_SetObject(PyExc_KeyError, args);
                Py_DECREF(args);
            }
        } else {
            PyErr_SetObject(PyExc_KeyError, key);
        }
    }
    return value;
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS)) {
        return __Pyx_PyList_GetItemRefFast(o, wrapped_i, unsafe_shared);
    } else
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyList_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyTuple_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)) {
            return __Pyx_PyList_GetItemRefFast(o, n, unsafe_shared);
        } else if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            return __Pyx_NewRef(PyList_GET_ITEM(o, n));
        }
    } else
    #if !CYTHON_AVOID_BORROWED_REFS
    if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            return __Pyx_NewRef(PyTuple_GET_ITEM(o, n));
        }
    } else
    #endif
#endif
#if CYTHON_USE_TYPE_SLOTS && !CYTHON_COMPILING_IN_PYPY
    {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (!is_list && mm && mm->mp_subscript) {
            PyObject *r, *key = PyLong_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (is_list || likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    (void)wraparound;
    (void)boundscheck;
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_mstate_global->__pyx_n_u_class_getitem);
        if (!meth) {
            PyErr_Clear();
        } else {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* CallNextTpDealloc */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc) {
    PyTypeObject* type = Py_TYPE(obj);
    destructor tp_dealloc = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_dealloc, destructor) != current_tp_dealloc)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_dealloc = __Pyx_PyType_GetSlot(type, tp_dealloc, destructor)) == current_tp_dealloc)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type)
        tp_dealloc(obj);
}

/* CallNextTpTraverse */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    traverseproc tp_traverse = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc) != current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_traverse = __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc)) == current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_traverse)
        return tp_traverse(obj, v, a);
    return 0;
}

/* CallTypeTraverse */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg) {
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x03090000
    if (__Pyx_get_runtime_version() < 0x03090000) return 0;
    #endif
    if (!always_call) {
        PyTypeObject *base = __Pyx_PyObject_GetSlot(o, tp_base, PyTypeObject*);
        unsigned long flags = PyType_GetFlags(base);
        if (flags & Py_TPFLAGS_HEAPTYPE) {
            return 0;
        }
    }
    Py_VISIT((PyObject*)Py_TYPE(o));
    return 0;
}
#endif

/* CallNextTpClear */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    inquiry tp_clear = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_clear, inquiry) != current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_clear = __Pyx_PyType_GetSlot(type, tp_clear, inquiry)) == current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_clear)
        tp_clear(obj);
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_3_2_4
#define __PYX_HAVE_RT_ImportType_3_2_4
static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size)
{
    PyObject *result = 0;
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API)
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !( defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API) )
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    if (size == 0) {
        return (PyTypeObject *)result;
    }
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_2_4 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_2_4 && (size_t)basicsize > size) {
        if (PyErr_WarnFormat(NULL, 0,
                "%.200s.%.200s size changed, may indicate binary incompatibility. "
                "Expected %zd from C header, got %zd from PyObject",
                module_name, class_name, size, basicsize) < 0) {
            goto bad;
        }
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* LimitedApiGetTypeDict (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static Py_ssize_t __Pyx_GetTypeDictOffset(void) {
    PyObject *tp_dictoffset_o;
    Py_ssize_t tp_dictoffset;
    tp_dictoffset_o = PyObject_GetAttrString((PyObject*)(&PyType_Type), "__dictoffset__");
    if (unlikely(!tp_dictoffset_o)) return -1;
    tp_dictoffset = PyLong_AsSsize_t(tp_dictoffset_o);
    Py_DECREF(tp_dictoffset_o);
    if (unlikely(tp_dictoffset == 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' doesn't have a dictoffset");
        return -1;
    } else if (unlikely(tp_dictoffset < 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' has an unexpected negative dictoffset. "
            "Please report this as Cython bug");
        return -1;
    }
    return tp_dictoffset;
}
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp) {
    static Py_ssize_t tp_dictoffset = 0;
    if (unlikely(tp_dictoffset == 0)) {
        tp_dictoffset = __Pyx_GetTypeDictOffset();
        if (unlikely(tp_dictoffset == -1 && PyErr_Occurred())) {
            tp_dictoffset = 0; // try again next time?
            return NULL;
        }
    }
    return *(PyObject**)((char*)tp + tp_dictoffset);
}
#endif

/* SetItemOnTypeDict (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_SetItem(tp_dict, k, v);
    if (likely(!result)) {
        PyType_Modified(tp);
        if (unlikely(PyObject_HasAttr(v, __pyx_mstate_global->__pyx_n_u_set_name))) {
            PyObject *setNameResult = PyObject_CallMethodObjArgs(v, __pyx_mstate_global->__pyx_n_u_set_name,  (PyObject *) tp, k, NULL);
            if (!setNameResult) return -1;
            Py_DECREF(setNameResult);
        }
    }
    return result;
}

/* FixUpExtensionType */
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if __PYX_LIMITED_VERSION_HEX > 0x030900B1
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#else
    const PyType_Slot *slot = spec->slots;
    int changed = 0;
#if !CYTHON_COMPILING_IN_LIMITED_API
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
#if !CYTHON_COMPILING_IN_CPYTHON
        const
#endif  // !CYTHON_COMPILING_IN_CPYTHON)
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_vectorcall_offset = memb->offset;
                    changed = 1;
                }
#endif  // CYTHON_METH_FASTCALL
#if !CYTHON_COMPILING_IN_PYPY
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                    Py_DECREF(descr);
                    if (unlikely(set_item_result < 0)) {
                        return -1;
                    }
                    changed = 1;
                }
#endif  // !CYTHON_COMPILING_IN_PYPY
            }
            memb++;
        }
    }
#endif  // !CYTHON_COMPILING_IN_LIMITED_API
#if !CYTHON_COMPILING_IN_PYPY
    slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_getset)
        slot++;
    if (slot && slot->slot == Py_tp_getset) {
        PyGetSetDef *getset = (PyGetSetDef*) slot->pfunc;
        while (getset && getset->name) {
            if (getset->name[0] == '_' && getset->name[1] == '_' && strcmp(getset->name, "__module__") == 0) {
                PyObject *descr = PyDescr_NewGetSet(type, getset);
                if (unlikely(!descr))
                    return -1;
                #if CYTHON_COMPILING_IN_LIMITED_API
                PyObject *pyname = PyUnicode_FromString(getset->name);
                if (unlikely(!pyname)) {
                    Py_DECREF(descr);
                    return -1;
                }
                int set_item_result = __Pyx_SetItemOnTypeDict(type, pyname, descr);
                Py_DECREF(pyname);
                #else
                CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
                int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                #endif
                Py_DECREF(descr);
                if (unlikely(set_item_result < 0)) {
                    return -1;
                }
                changed = 1;
            }
            ++getset;
        }
    }
#else
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#endif  // !CYTHON_COMPILING_IN_PYPY
    if (changed)
        PyType_Modified(type);
#endif  // PY_VERSION_HEX > 0x030900B1
    return 0;
}

/* PyObjectCallNoArg (used by PyObjectCallMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod (used by PyObjectCallMethod0) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(tp);
    PyErr_Format(PyExc_AttributeError,
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}
#endif

/* PyObjectCallMethod0 (used by PyType_Ready) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
#if CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000))
    PyObject *args[1] = {obj};
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_CallNoArg;
    return PyObject_VectorcallMethod(method_name, args, 1 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
#endif
}

/* ValidateBasesTuple (used by PyType_Ready) */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_SIZE
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (unlikely(n < 0)) return -1;
#endif
    for (i = 1; i < n; i++)
    {
        PyTypeObject *b;
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !CYTHON_USE_TYPE_SLOTS
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
CYTHON_UNUSED static int __Pyx_PyType_HasMultipleInheritance(PyTypeObject *t) {
    while (t) {
        PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
        if (bases) {
            return 1;
        }
        t = __Pyx_PyType_GetSlot(t, tp_base, PyTypeObject*);
    }
    return 0;
}
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !CYTHON_COMPILING_IN_CPYTHON || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    if (!__Pyx_PyType_HasMultipleInheritance(t)) {
        return PyType_Ready(t);
    }
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400) &&\
                !CYTHON_COMPILING_IN_GRAAL
        gc = PyImport_GetModule(__pyx_mstate_global->__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_mstate_global->__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* SetVTable */
static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* MergeVTables */
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i=0;
    Py_ssize_t size;
    void** base_vtables;
    __Pyx_TypeName tp_base_name = NULL;
    __Pyx_TypeName base_name = NULL;
    void* unknown = (void*)-1;
    PyObject* bases = __Pyx_PyType_GetSlot(type, tp_bases, PyObject*);
    int base_depth = 0;
    {
        PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        while (base) {
            base_depth += 1;
            base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
        }
    }
    base_vtables = (void**) PyMem_Malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
#if CYTHON_COMPILING_IN_LIMITED_API
    size = PyTuple_Size(bases);
    if (size < 0) goto other_failure;
#else
    size = PyTuple_GET_SIZE(bases);
#endif
    for (i = 1; i < size; i++) {
        PyObject *basei;
        void* base_vtable;
#if CYTHON_AVOID_BORROWED_REFS
        basei = PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#else
        basei = PyTuple_GET_ITEM(bases, i);
#endif
        base_vtable = __Pyx_GetVtable((PyTypeObject*)basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
            }
        }
    }
    PyErr_Clear();
    PyMem_Free(base_vtables);
    return 0;
bad:
    {
        PyTypeObject* basei = NULL;
        PyTypeObject* tp_base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        tp_base_name = __Pyx_PyType_GetFullyQualifiedName(tp_base);
#if CYTHON_AVOID_BORROWED_REFS
        basei = (PyTypeObject*)PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = (PyTypeObject*)PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#else
        basei = (PyTypeObject*)PyTuple_GET_ITEM(bases, i);
#endif
        base_name = __Pyx_PyType_GetFullyQualifiedName(basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
    }
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
really_bad: // bad has failed!
#endif
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
other_failure:
#endif
    PyMem_Free(base_vtables);
    return -1;
}

/* DelItemOnTypeDict (used by SetupReduce) */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_DelItem(tp_dict, k);
    if (likely(!result)) PyType_Modified(tp);
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_mstate_global->__pyx_n_u_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_mstate_global->__pyx_n_u_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython);
            if (likely(reduce_cython)) {
                ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_mstate_global->__pyx_n_u_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetFullyQualifiedName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* PxdImportShared (used by FunctionImport) */
#ifndef __PYX_HAVE_RT_ImportFromPxd_3_2_4
#define __PYX_HAVE_RT_ImportFromPxd_3_2_4
static int __Pyx_ImportFromPxd_3_2_4(PyObject *module, const char *name, void **p, const char *sig, const char *what) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, "__pyx_capi__");
    if (!d)
        goto bad;
#if (defined(Py_LIMITED_API) && Py_LIMITED_API >= 0x030d0000) || (!defined(Py_LIMITED_API) && PY_VERSION_HEX >= 0x030d0000)
    PyDict_GetItemStringRef(d, name, &cobj);
#else
    cobj = PyDict_GetItemString(d, name);
    Py_XINCREF(cobj);
#endif
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C %.8s %.200s",
                PyModule_GetName(module), what, name);
        goto bad;
    }
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C %.8s %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             what, PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    Py_DECREF(cobj);
    return 0;
bad:
    Py_XDECREF(d);
    Py_XDECREF(cobj);
    return -1;
}
#endif

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction_3_2_4
#define __PYX_HAVE_RT_ImportFunction_3_2_4
static int __Pyx_ImportFunction_3_2_4(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    int result = __Pyx_ImportFromPxd_3_2_4(module, funcname, &tmp.p, sig, "function");
    if (result == 0) {
        *f = tmp.fp;
    }
    return result;
}
#endif

/* dict_setdefault (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value) {
    PyObject* value;
#if __PYX_LIMITED_VERSION_HEX >= 0x030F0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4)
    PyDict_SetDefaultRef(d, key, default_value, &value);
#elif CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    PyObject *args[] = {d, key, default_value};
    value = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_n_u_setdefault, args, 3 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#elif CYTHON_COMPILING_IN_LIMITED_API
    value = PyObject_CallMethodObjArgs(d, __pyx_mstate_global->__pyx_n_u_setdefault, key, default_value, NULL);
#else
    value = PyDict_SetDefault(d, key, default_value);
    if (unlikely(!value)) return NULL;
    Py_INCREF(value);
#endif
    return value;
}

/* AddModuleRef (used by FetchSharedCythonModule) */
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  static PyObject *__Pyx_PyImport_AddModuleObjectRef(PyObject *name) {
      PyObject *module_dict = PyImport_GetModuleDict();
      PyObject *m;
      if (PyMapping_GetOptionalItem(module_dict, name, &m) < 0) {
          return NULL;
      }
      if (m != NULL && PyModule_Check(m)) {
          return m;
      }
      Py_XDECREF(m);
      m = PyModule_NewObject(name);
      if (m == NULL)
          return NULL;
      if (PyDict_CheckExact(module_dict)) {
          PyObject *new_m;
          (void)PyDict_SetDefaultRef(module_dict, name, m, &new_m);
          Py_DECREF(m);
          return new_m;
      } else {
           if (PyObject_SetItem(module_dict, name, m) != 0) {
                Py_DECREF(m);
                return NULL;
            }
            return m;
      }
  }
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *py_name = PyUnicode_FromString(name);
      if (!py_name) return NULL;
      PyObject *module = __Pyx_PyImport_AddModuleObjectRef(py_name);
      Py_DECREF(py_name);
      return module;
  }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif

/* FetchSharedCythonModule (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef(__PYX_ABI_MODULE_NAME);
}

/* FetchCommonType (used by CommonTypesMetaclass) */
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject* __Pyx_PyType_FromMetaclass(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *result = __Pyx_PyType_FromModuleAndSpec(module, spec, bases);
    if (result && metaclass) {
        PyObject *old_tp = (PyObject*)Py_TYPE(result);
    Py_INCREF((PyObject*)metaclass);
#if __PYX_LIMITED_VERSION_HEX >= 0x03090000
        Py_SET_TYPE(result, metaclass);
#else
        result->ob_type = metaclass;
#endif
        Py_DECREF(old_tp);
    }
    return result;
}
#else
#define __Pyx_PyType_FromMetaclass(me, mo, s, b) PyType_FromMetaclass(me, mo, s, b)
#endif
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t expected_basicsize) {
    Py_ssize_t basicsize;
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (expected_basicsize == 0) {
        return 0; // size is inherited, nothing useful to check
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
    if (unlikely(!py_basicsize)) return -1;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = NULL;
    if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) return -1;
#else
    basicsize = ((PyTypeObject*) cached_type)->tp_basicsize;
#endif
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module = NULL, *cached_type = NULL, *abi_module_dict, *new_cached_type, *py_object_name;
    int get_item_ref_result;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    py_object_name = PyUnicode_FromString(object_name);
    if (!py_object_name) return NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) goto done;
    abi_module_dict = PyModule_GetDict(abi_module);
    if (!abi_module_dict) goto done;
    get_item_ref_result = __Pyx_PyDict_GetItemRef(abi_module_dict, py_object_name, &cached_type);
    if (get_item_ref_result == 1) {
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else if (unlikely(get_item_ref_result == -1)) {
        goto bad;
    }
    cached_type = __Pyx_PyType_FromMetaclass(
        metaclass,
        CYTHON_USE_MODULE_STATE ? module : abi_module,
        spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    new_cached_type = __Pyx_PyDict_SetDefault(abi_module_dict, py_object_name, cached_type);
    if (unlikely(new_cached_type != cached_type)) {
        if (unlikely(!new_cached_type)) goto bad;
        Py_DECREF(cached_type);
        cached_type = new_cached_type;
        if (__Pyx_VerifyCachedType(
                cached_type,
                object_name,
                spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else {
        Py_DECREF(new_cached_type);
    }
done:
    Py_XDECREF(abi_module);
    Py_DECREF(py_object_name);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CommonTypesMetaclass (used by CythonFunctionShared) */
static PyObject* __pyx_CommonTypesMetaclass_get_module(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED void* context) {
    return PyUnicode_FromString(__PYX_ABI_MODULE_NAME);
}
#if __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject* __pyx_CommonTypesMetaclass_call(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *args, CYTHON_UNUSED PyObject *kwds) {
    PyErr_SetString(PyExc_TypeError, "Cannot instantiate Cython internal types");
    return NULL;
}
static int __pyx_CommonTypesMetaclass_setattr(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *attr, CYTHON_UNUSED PyObject *value) {
    PyErr_SetString(PyExc_TypeError, "Cython internal types are immutable");
    return -1;
}
#endif
static PyGetSetDef __pyx_CommonTypesMetaclass_getset[] = {
    {"__module__", __pyx_CommonTypesMetaclass_get_module, NULL, NULL, NULL},
    {0, 0, 0, 0, 0}
};
static PyType_Slot __pyx_CommonTypesMetaclass_slots[] = {
    {Py_tp_getset, (void *)__pyx_CommonTypesMetaclass_getset},
    #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {Py_tp_call, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_new, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_setattro, (void*)__pyx_CommonTypesMetaclass_setattr},
    #endif
    {0, 0}
};
static PyType_Spec __pyx_CommonTypesMetaclass_spec = {
    __PYX_TYPE_MODULE_PREFIX "_common_types_metatype",
    0,
    0,
    Py_TPFLAGS_IMMUTABLETYPE |
    Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT,
    __pyx_CommonTypesMetaclass_slots
};
static int __pyx_CommonTypesMetaclass_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    PyObject *bases = PyTuple_Pack(1, &PyType_Type);
    if (unlikely(!bases)) {
        return -1;
    }
    mstate->__pyx_CommonTypesMetaclassType = __Pyx_FetchCommonTypeFromSpec(NULL, module, &__pyx_CommonTypesMetaclass_spec, bases);
    Py_DECREF(bases);
    if (unlikely(mstate->__pyx_CommonTypesMetaclassType == NULL)) {
        return -1;
    }
    return 0;
}

/* PyMethodNew (used by CythonFunctionShared) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *result;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    #if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    {
        PyObject *args[] = {func, self};
        result = PyObject_Vectorcall(__pyx_mstate_global->__Pyx_CachedMethodType, args, 2, NULL);
    }
    #else
    result = PyObject_CallFunctionObjArgs(__pyx_mstate_global->__Pyx_CachedMethodType, func, self, NULL);
    #endif
    return result;
}
#else
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#endif

/* PyVectorcallFastCallDict (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i;
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos;
    #else
    Py_ssize_t pos;
    #endif
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    #if !CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t nkw = PyDict_Size(kw);
    if (unlikely(nkw == -1)) return NULL;
    #else
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    #endif
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = 0;
    i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (__Pyx_PyDict_NextRef(kw, &pos, &key, &value)) {
        keys_are_strings &=
        #if CYTHON_COMPILING_IN_LIMITED_API
            PyType_GetFlags(Py_TYPE(key));
        #else
            Py_TYPE(key)->tp_flags;
        #endif
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(PyTuple_SetItem(kwnames, i, key) < 0)) goto cleanup;
        #else
        PyTuple_SET_ITEM(kwnames, i, key);
        #endif
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(pos);
    #endif
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    Py_ssize_t kw_size =
        likely(kw == NULL) ?
        0 :
#if !CYTHON_ASSUME_SAFE_SIZE
        PyDict_Size(kw);
#else
        PyDict_GET_SIZE(kw);
#endif
    if (kw_size == 0) {
        return vc(func, args, nargs, NULL);
    }
#if !CYTHON_ASSUME_SAFE_SIZE
    else if (unlikely(kw_size == -1)) {
        return NULL;
    }
#endif
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared (used by CythonFunction) */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunctionNoMethod(PyObject *func, void (*cfunc)(void)) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if ((PyObject*)Py_TYPE(func) == __pyx_mstate_global->__Pyx_CachedMethodType) {
        int result;
        PyObject *newFunc = PyObject_GetAttr(func, __pyx_mstate_global->__pyx_n_u_func);
        if (unlikely(!newFunc)) {
            PyErr_Clear(); // It's only an optimization, so don't throw an error
            return 0;
        }
        result = __Pyx__IsSameCyOrCFunctionNoMethod(newFunc, cfunc);
        Py_DECREF(newFunc);
        return result;
    }
    return __Pyx__IsSameCyOrCFunctionNoMethod(func, cfunc);
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if (PyMethod_Check(func)) {
        func = PyMethod_GET_FUNCTION(func);
    }
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure) {
    PyObject *result;
    CYTHON_UNUSED_VAR(closure);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_doc_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#else
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_name_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    PyObject *result;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    Py_INCREF(op->func_qualname);
    result = op->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
#endif
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_defaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_kwdefaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_annotations_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine_value(__pyx_CyFunctionObject *op) {
    int is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
    if (is_coroutine) {
        PyObject *is_coroutine_value, *module, *fromlist, *marker = __pyx_mstate_global->__pyx_n_u_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_mstate_global->__pyx_n_u_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        is_coroutine_value = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(is_coroutine_value)) {
            return is_coroutine_value;
        }
ignore:
        PyErr_Clear();
    }
    return __Pyx_PyBool_FromLong(is_coroutine);
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    result = __Pyx_CyFunction_get_is_coroutine_value(op);
    if (unlikely(!result))
        return NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    if (op->func_is_coroutine) {
        Py_DECREF(result);
        result = __Pyx_NewRef(op->func_is_coroutine);
    } else {
        op->func_is_coroutine = __Pyx_NewRef(result);
    }
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static void __Pyx_CyFunction_raise_argument_count_error(__pyx_CyFunctionObject *func, const char* message, Py_ssize_t size) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        py_name, message, size);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        name, message, size);
#endif
}
static void __Pyx_CyFunction_raise_type_error(__pyx_CyFunctionObject *func, const char* message) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s",
        py_name, message);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s",
        name, message);
#endif
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {"func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {"func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
#else
    {"func_dict", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
#endif
    {"func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {"__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {"_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    {"__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#endif
#if CYTHON_METH_FASTCALL
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {"__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(args);
    __Pyx_BEGIN_CRITICAL_SECTION(m);
    Py_INCREF(m->func_qualname);
    result = m->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_dict = NULL;
#endif
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func_dict);
#elif PY_VERSION_HEX < 0x030d0000
    _PyObject_ClearManagedDict((PyObject*)m);
#else
    PyObject_ClearManagedDict((PyObject*)m);
#endif
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    Py_CLEAR(m->defaults);
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    {
        int e = __Pyx_call_type_traverse((PyObject*)m, 1, visit, arg);
        if (e) return e;
    }
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func_dict);
#else
    {
        int e =
#if PY_VERSION_HEX < 0x030d0000
            _PyObject_VisitManagedDict
#else
            PyObject_VisitManagedDict
#endif
                ((PyObject*)m, visit, arg);
        if (e != 0) return e;
    }
#endif
    __Pyx_VISIT_CONST(m->func_name);
    __Pyx_VISIT_CONST(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    __Pyx_VISIT_CONST(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    Py_VISIT(m->defaults);
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
    PyObject *repr;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    repr = PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
    __Pyx_END_CRITICAL_SECTION();
    return repr;
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void(*)(void))meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes no arguments", size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes exactly one argument", size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
    __Pyx_CyFunction_raise_type_error(
        (__pyx_CyFunctionObject*)func, "takes no keyword arguments");
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_SIZE
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(argc < 0)) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            __Pyx_CyFunction_raise_type_error(
                cyfunc, "needs an argument");
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(__Pyx_PyTuple_GET_SIZE(kwnames))) {
        __Pyx_CyFunction_raise_type_error(
            cyfunc, "takes no keyword arguments");
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes no arguments", nargs);
        return NULL;
    }
    return meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes exactly one argument", nargs);
        return NULL;
    }
    return meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    #if PY_VERSION_HEX < 0x030e00A6
    size_t nargs_value = (size_t) nargs;
    #else
    Py_ssize_t nargs_value = nargs;
    #endif
    return ((__Pyx_PyCMethod)(void(*)(void))meth)(self, cls, args, nargs_value, kwnames);
}
#endif
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if CYTHON_METH_FASTCALL
#if defined(Py_TPFLAGS_HAVE_VECTORCALL)
    Py_TPFLAGS_HAVE_VECTORCALL |
#elif defined(_Py_TPFLAGS_HAVE_VECTORCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
#endif // CYTHON_METH_FASTCALL
#if PY_VERSION_HEX >= 0x030C0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_TPFLAGS_MANAGED_DICT |
#endif
    Py_TPFLAGS_IMMUTABLETYPE | Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
static int __pyx_CyFunction_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    mstate->__pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(
        mstate->__pyx_CommonTypesMetaclassType, module, &__pyx_CyFunctionType_spec, NULL);
    if (unlikely(mstate->__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func, PyTypeObject *defaults_type) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_CallObject((PyObject*)defaults_type, NULL); // _PyObject_New(defaults_type);
    if (unlikely(!m->defaults))
        return NULL;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_mstate_global->__pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* CLineInTraceback (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
#define __Pyx_PyProbablyModule_GetDict(o) __Pyx_XNewRef(PyModule_GetDict(o))
#elif !CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyProbablyModule_GetDict(o) PyObject_GenericGetDict(o, NULL);
#else
PyObject* __Pyx_PyProbablyModule_GetDict(PyObject *o) {
    PyObject **dict_ptr = _PyObject_GetDictPtr(o);
    return dict_ptr ? __Pyx_XNewRef(*dict_ptr) : NULL;
}
#endif
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline = NULL;
    PyObject *ptype, *pvalue, *ptraceback;
    PyObject *cython_runtime_dict;
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_mstate_global->__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
    cython_runtime_dict = __Pyx_PyProbablyModule_GetDict(__pyx_mstate_global->__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, cython_runtime_dict,
            __Pyx_PyDict_SetDefault(cython_runtime_dict, __pyx_mstate_global->__pyx_n_u_cline_in_traceback, Py_False))
    }
    if (use_cline == NULL || use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    Py_XDECREF(use_cline);
    Py_XDECREF(cython_runtime_dict);
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache (used by AddTraceback) */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static __Pyx_CachedCodeObjectType *__pyx__find_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line) {
    __Pyx_CachedCodeObjectType* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!code_cache->entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if (unlikely(pos >= code_cache->count) || unlikely(code_cache->entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = code_cache->entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__find_code_object;
    return NULL; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just miss.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type old_count = __pyx_atomic_incr_acq_rel(&code_cache->accessor_count);
    if (old_count < 0) {
        __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
        return NULL;
    }
#endif
    __Pyx_CachedCodeObjectType *result = __pyx__find_code_object(code_cache, code_line);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
#endif
    return result;
#endif
}
static void __pyx__insert_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line, __Pyx_CachedCodeObjectType* code_object)
{
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = code_cache->entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            code_cache->entries = entries;
            code_cache->max_count = 64;
            code_cache->count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if ((pos < code_cache->count) && unlikely(code_cache->entries[pos].code_line == code_line)) {
        __Pyx_CachedCodeObjectType* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_INCREF(code_object);
        Py_DECREF(tmp);
        return;
    }
    if (code_cache->count == code_cache->max_count) {
        int new_max = code_cache->max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            code_cache->entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        code_cache->entries = entries;
        code_cache->max_count = new_max;
    }
    for (i=code_cache->count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    code_cache->count++;
    Py_INCREF(code_object);
}
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__insert_code_object;
    return; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just fail.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type expected = 0;
    if (!__pyx_atomic_int_cmp_exchange(&code_cache->accessor_count, &expected, INT_MIN)) {
        return;
    }
#endif
    __pyx__insert_code_object(code_cache, code_line, code_object);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_sub(&code_cache->accessor_count, INT_MIN);
#endif
#endif
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result = PyObject_Call(replace, __pyx_mstate_global->__pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!code_object) {
        code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
        if (unlikely(!code_object)) goto bad;
        py_py_line = PyLong_FromLong(py_line);
        if (unlikely(!py_py_line)) goto bad;
        if (c_line) {
            py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        } else {
            py_funcname = PyUnicode_FromString(funcname);
        }
        if (unlikely(!py_funcname)) goto bad;
        dict = PyDict_New();
        if (unlikely(!dict)) goto bad;
        {
            PyObject *old_code_object = code_object;
            code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
            Py_DECREF(old_code_object);
        }
        if (unlikely(!code_object)) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, code_object);
    } else {
        dict = PyDict_New();
    }
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    if (c_line) {
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
    }
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_mstate_global->__pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_int(unsigned int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned int) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned int) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(unsigned int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyLong_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_long(unsigned long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned long neg_one = (unsigned long) -1, const_zero = (unsigned long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(unsigned long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(unsigned long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE unsigned int __Pyx_PyLong_As_unsigned_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        unsigned int val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (unsigned int) -1;
        val = __Pyx_PyLong_As_unsigned_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(unsigned int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) >= 2 * PyLong_SHIFT)) {
                            return (unsigned int) (((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) >= 3 * PyLong_SHIFT)) {
                            return (unsigned int) (((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) >= 4 * PyLong_SHIFT)) {
                            return (unsigned int) (((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (unsigned int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(unsigned int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(unsigned int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned int) (((unsigned int)-1)*(((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(unsigned int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned int) ((((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned int) (((unsigned int)-1)*(((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned int) ((((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned int) (((unsigned int)-1)*(((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned int) ((((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(unsigned int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, long, PyLong_AsLong(x))
        } else if ((sizeof(unsigned int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        unsigned int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (unsigned int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (unsigned int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (unsigned int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (unsigned int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(unsigned int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((unsigned int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(unsigned int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((unsigned int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((unsigned int) 1) << (sizeof(unsigned int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (unsigned int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned int");
    return (unsigned int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned int");
    return (unsigned int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE unsigned long __Pyx_PyLong_As_unsigned_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned long neg_one = (unsigned long) -1, const_zero = (unsigned long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        unsigned long val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (unsigned long) -1;
        val = __Pyx_PyLong_As_unsigned_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(unsigned long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) >= 2 * PyLong_SHIFT)) {
                            return (unsigned long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) >= 3 * PyLong_SHIFT)) {
                            return (unsigned long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) >= 4 * PyLong_SHIFT)) {
                            return (unsigned long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (unsigned long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(unsigned long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned long, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(unsigned long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(unsigned long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned long) (((unsigned long)-1)*(((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(unsigned long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned long) ((((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned long) (((unsigned long)-1)*(((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned long) ((((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned long) (((unsigned long)-1)*(((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned long) ((((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(unsigned long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned long, long, PyLong_AsLong(x))
        } else if ((sizeof(unsigned long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned long, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        unsigned long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (unsigned long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (unsigned long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (unsigned long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (unsigned long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(unsigned long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((unsigned long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(unsigned long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((unsigned long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((unsigned long) 1) << (sizeof(unsigned long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (unsigned long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned long");
    return (unsigned long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned long");
    return (unsigned long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_char(unsigned char value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned char neg_one = (unsigned char) -1, const_zero = (unsigned char) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned char) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned char) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(unsigned char) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned char) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned char) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned char),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(unsigned char));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_short(unsigned short value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned short neg_one = (unsigned short) -1, const_zero = (unsigned short) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned short) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned short) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(unsigned short) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned short) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned short) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned short),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(unsigned short));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE enum ibv_wc_status __Pyx_PyLong_As_enum__ibv_wc_status(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_wc_status neg_one = (enum ibv_wc_status) -1, const_zero = (enum ibv_wc_status) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        enum ibv_wc_status val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (enum ibv_wc_status) -1;
        val = __Pyx_PyLong_As_enum__ibv_wc_status(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(enum ibv_wc_status, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(enum ibv_wc_status) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wc_status, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wc_status) >= 2 * PyLong_SHIFT)) {
                            return (enum ibv_wc_status) (((((enum ibv_wc_status)digits[1]) << PyLong_SHIFT) | (enum ibv_wc_status)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(enum ibv_wc_status) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wc_status, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wc_status) >= 3 * PyLong_SHIFT)) {
                            return (enum ibv_wc_status) (((((((enum ibv_wc_status)digits[2]) << PyLong_SHIFT) | (enum ibv_wc_status)digits[1]) << PyLong_SHIFT) | (enum ibv_wc_status)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(enum ibv_wc_status) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wc_status, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wc_status) >= 4 * PyLong_SHIFT)) {
                            return (enum ibv_wc_status) (((((((((enum ibv_wc_status)digits[3]) << PyLong_SHIFT) | (enum ibv_wc_status)digits[2]) << PyLong_SHIFT) | (enum ibv_wc_status)digits[1]) << PyLong_SHIFT) | (enum ibv_wc_status)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (enum ibv_wc_status) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(enum ibv_wc_status) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_wc_status, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(enum ibv_wc_status) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_wc_status, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(enum ibv_wc_status, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(enum ibv_wc_status) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wc_status, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wc_status) - 1 > 2 * PyLong_SHIFT)) {
                            return (enum ibv_wc_status) (((enum ibv_wc_status)-1)*(((((enum ibv_wc_status)digits[1]) << PyLong_SHIFT) | (enum ibv_wc_status)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(enum ibv_wc_status) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wc_status, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wc_status) - 1 > 2 * PyLong_SHIFT)) {
                            return (enum ibv_wc_status) ((((((enum ibv_wc_status)digits[1]) << PyLong_SHIFT) | (enum ibv_wc_status)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(enum ibv_wc_status) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wc_status, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wc_status) - 1 > 3 * PyLong_SHIFT)) {
                            return (enum ibv_wc_status) (((enum ibv_wc_status)-1)*(((((((enum ibv_wc_status)digits[2]) << PyLong_SHIFT) | (enum ibv_wc_status)digits[1]) << PyLong_SHIFT) | (enum ibv_wc_status)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(enum ibv_wc_status) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wc_status, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wc_status) - 1 > 3 * PyLong_SHIFT)) {
                            return (enum ibv_wc_status) ((((((((enum ibv_wc_status)digits[2]) << PyLong_SHIFT) | (enum ibv_wc_status)digits[1]) << PyLong_SHIFT) | (enum ibv_wc_status)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(enum ibv_wc_status) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wc_status, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wc_status) - 1 > 4 * PyLong_SHIFT)) {
                            return (enum ibv_wc_status) (((enum ibv_wc_status)-1)*(((((((((enum ibv_wc_status)digits[3]) << PyLong_SHIFT) | (enum ibv_wc_status)digits[2]) << PyLong_SHIFT) | (enum ibv_wc_status)digits[1]) << PyLong_SHIFT) | (enum ibv_wc_status)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(enum ibv_wc_status) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wc_status, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wc_status) - 1 > 4 * PyLong_SHIFT)) {
                            return (enum ibv_wc_status) ((((((((((enum ibv_wc_status)digits[3]) << PyLong_SHIFT) | (enum ibv_wc_status)digits[2]) << PyLong_SHIFT) | (enum ibv_wc_status)digits[1]) << PyLong_SHIFT) | (enum ibv_wc_status)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(enum ibv_wc_status) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_wc_status, long, PyLong_AsLong(x))
        } else if ((sizeof(enum ibv_wc_status) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_wc_status, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        enum ibv_wc_status val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() or PyLong_AsNativeBytes() not available, cannot convert large enums");
        val = (enum ibv_wc_status) -1;
#endif
        if (unlikely(ret))
            return (enum ibv_wc_status) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum ibv_wc_status");
    return (enum ibv_wc_status) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum ibv_wc_status");
    return (enum ibv_wc_status) -1;
}

/* CIntFromPy */
static CYTHON_INLINE enum ibv_wc_opcode __Pyx_PyLong_As_enum__ibv_wc_opcode(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum ibv_wc_opcode neg_one = (enum ibv_wc_opcode) -1, const_zero = (enum ibv_wc_opcode) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        enum ibv_wc_opcode val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (enum ibv_wc_opcode) -1;
        val = __Pyx_PyLong_As_enum__ibv_wc_opcode(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(enum ibv_wc_opcode, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(enum ibv_wc_opcode) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wc_opcode, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wc_opcode) >= 2 * PyLong_SHIFT)) {
                            return (enum ibv_wc_opcode) (((((enum ibv_wc_opcode)digits[1]) << PyLong_SHIFT) | (enum ibv_wc_opcode)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(enum ibv_wc_opcode) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wc_opcode, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wc_opcode) >= 3 * PyLong_SHIFT)) {
                            return (enum ibv_wc_opcode) (((((((enum ibv_wc_opcode)digits[2]) << PyLong_SHIFT) | (enum ibv_wc_opcode)digits[1]) << PyLong_SHIFT) | (enum ibv_wc_opcode)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(enum ibv_wc_opcode) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wc_opcode, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wc_opcode) >= 4 * PyLong_SHIFT)) {
                            return (enum ibv_wc_opcode) (((((((((enum ibv_wc_opcode)digits[3]) << PyLong_SHIFT) | (enum ibv_wc_opcode)digits[2]) << PyLong_SHIFT) | (enum ibv_wc_opcode)digits[1]) << PyLong_SHIFT) | (enum ibv_wc_opcode)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (enum ibv_wc_opcode) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(enum ibv_wc_opcode) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_wc_opcode, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(enum ibv_wc_opcode) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_wc_opcode, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(enum ibv_wc_opcode, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(enum ibv_wc_opcode) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wc_opcode, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wc_opcode) - 1 > 2 * PyLong_SHIFT)) {
                            return (enum ibv_wc_opcode) (((enum ibv_wc_opcode)-1)*(((((enum ibv_wc_opcode)digits[1]) << PyLong_SHIFT) | (enum ibv_wc_opcode)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(enum ibv_wc_opcode) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wc_opcode, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wc_opcode) - 1 > 2 * PyLong_SHIFT)) {
                            return (enum ibv_wc_opcode) ((((((enum ibv_wc_opcode)digits[1]) << PyLong_SHIFT) | (enum ibv_wc_opcode)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(enum ibv_wc_opcode) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wc_opcode, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wc_opcode) - 1 > 3 * PyLong_SHIFT)) {
                            return (enum ibv_wc_opcode) (((enum ibv_wc_opcode)-1)*(((((((enum ibv_wc_opcode)digits[2]) << PyLong_SHIFT) | (enum ibv_wc_opcode)digits[1]) << PyLong_SHIFT) | (enum ibv_wc_opcode)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(enum ibv_wc_opcode) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wc_opcode, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wc_opcode) - 1 > 3 * PyLong_SHIFT)) {
                            return (enum ibv_wc_opcode) ((((((((enum ibv_wc_opcode)digits[2]) << PyLong_SHIFT) | (enum ibv_wc_opcode)digits[1]) << PyLong_SHIFT) | (enum ibv_wc_opcode)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(enum ibv_wc_opcode) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wc_opcode, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wc_opcode) - 1 > 4 * PyLong_SHIFT)) {
                            return (enum ibv_wc_opcode) (((enum ibv_wc_opcode)-1)*(((((((((enum ibv_wc_opcode)digits[3]) << PyLong_SHIFT) | (enum ibv_wc_opcode)digits[2]) << PyLong_SHIFT) | (enum ibv_wc_opcode)digits[1]) << PyLong_SHIFT) | (enum ibv_wc_opcode)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(enum ibv_wc_opcode) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum ibv_wc_opcode, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum ibv_wc_opcode) - 1 > 4 * PyLong_SHIFT)) {
                            return (enum ibv_wc_opcode) ((((((((((enum ibv_wc_opcode)digits[3]) << PyLong_SHIFT) | (enum ibv_wc_opcode)digits[2]) << PyLong_SHIFT) | (enum ibv_wc_opcode)digits[1]) << PyLong_SHIFT) | (enum ibv_wc_opcode)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(enum ibv_wc_opcode) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_wc_opcode, long, PyLong_AsLong(x))
        } else if ((sizeof(enum ibv_wc_opcode) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum ibv_wc_opcode, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        enum ibv_wc_opcode val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() or PyLong_AsNativeBytes() not available, cannot convert large enums");
        val = (enum ibv_wc_opcode) -1;
#endif
        if (unlikely(ret))
            return (enum ibv_wc_opcode) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum ibv_wc_opcode");
    return (enum ibv_wc_opcode) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum ibv_wc_opcode");
    return (enum ibv_wc_opcode) -1;
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
static __Pyx_TypeName
__Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp)
{
    PyObject *module = NULL, *name = NULL, *result = NULL;
    #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
    name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_qualname);
    #else
    name = PyType_GetQualName(tp);
    #endif
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) goto bad;
    module = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_module);
    if (unlikely(module == NULL) || unlikely(!PyUnicode_Check(module))) goto bad;
    if (PyUnicode_CompareWithASCIIString(module, "builtins") == 0) {
        result = name;
        name = NULL;
        goto done;
    }
    result = PyUnicode_FromFormat("%U.%U", module, name);
    if (unlikely(result == NULL)) goto bad;
  done:
    Py_XDECREF(name);
    Py_XDECREF(module);
    return result;
  bad:
    PyErr_Clear();
    if (name) {
        result = name;
        name = NULL;
    } else {
        result = __Pyx_NewRef(__pyx_mstate_global->__pyx_kp_u__4);
    }
    goto done;
}
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyLong_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* GetRuntimeVersion */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
void __Pyx_init_runtime_version(void) {
    if (__Pyx_cached_runtime_version == 0) {
        const char* rt_version = Py_GetVersion();
        unsigned long version = 0;
        unsigned long factor = 0x01000000UL;
        unsigned int digit = 0;
        int i = 0;
        while (factor) {
            while ('0' <= rt_version[i] && rt_version[i] <= '9') {
                digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
                ++i;
            }
            version += factor * digit;
            if (rt_version[i] != '.')
                break;
            digit = 0;
            factor >>= 8;
            ++i;
        }
        __Pyx_cached_runtime_version = version;
    }
}
#endif
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    return Py_Version & ~0xFFUL;
#else
    return __Pyx_cached_runtime_version;
#endif
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* NewCodeObj */
#if CYTHON_COMPILING_IN_LIMITED_API
    static PyObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                       PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                       PyObject *fv, PyObject *cell, PyObject* fn,
                                       PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
#elif PY_VERSION_HEX >= 0x030B0000
  static PyCodeObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                         PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                         PyObject *fv, PyObject *cell, PyObject* fn,
                                         PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, __pyx_mstate_global->__pyx_empty_bytes);
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030c00A1
    if (likely(result))
        result->_co_firsttraceable = 0;
    #endif
    return result;
  }
#elif !CYTHON_COMPILING_IN_PYPY
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
) {
    PyObject *code_obj = NULL, *varnames_tuple_dedup = NULL, *code_bytes = NULL;
    Py_ssize_t var_count = (Py_ssize_t) descr.nlocals;
    PyObject *varnames_tuple = PyTuple_New(var_count);
    if (unlikely(!varnames_tuple)) return NULL;
    for (Py_ssize_t i=0; i < var_count; i++) {
        Py_INCREF(varnames[i]);
        if (__Pyx_PyTuple_SET_ITEM(varnames_tuple, i, varnames[i]) != (0)) goto done;
    }
    #if CYTHON_COMPILING_IN_LIMITED_API
    varnames_tuple_dedup = PyDict_GetItem(tuple_dedup_map, varnames_tuple);
    if (!varnames_tuple_dedup) {
        if (unlikely(PyDict_SetItem(tuple_dedup_map, varnames_tuple, varnames_tuple) < 0)) goto done;
        varnames_tuple_dedup = varnames_tuple;
    }
    #else
    varnames_tuple_dedup = PyDict_SetDefault(tuple_dedup_map, varnames_tuple, varnames_tuple);
    if (unlikely(!varnames_tuple_dedup)) goto done;
    #endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(varnames_tuple_dedup);
    #endif
    if (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table != NULL && !CYTHON_COMPILING_IN_GRAAL) {
        Py_ssize_t line_table_length = __Pyx_PyBytes_GET_SIZE(line_table);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(line_table_length == -1)) goto done;
        #endif
        Py_ssize_t code_len = (line_table_length * 2 + 4) & ~3LL;
        code_bytes = PyBytes_FromStringAndSize(NULL, code_len);
        if (unlikely(!code_bytes)) goto done;
        char* c_code_bytes = PyBytes_AsString(code_bytes);
        if (unlikely(!c_code_bytes)) goto done;
        memset(c_code_bytes, 0, (size_t) code_len);
    }
    code_obj = (PyObject*) __Pyx__PyCode_New(
        (int) descr.argcount,
        (int) descr.num_posonly_args,
        (int) descr.num_kwonly_args,
        (int) descr.nlocals,
        0,
        (int) descr.flags,
        code_bytes ? code_bytes : __pyx_mstate_global->__pyx_empty_bytes,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        varnames_tuple_dedup,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        filename,
        funcname,
        (int) descr.first_line,
        (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table) ? line_table : __pyx_mstate_global->__pyx_empty_bytes
    );
done:
    Py_XDECREF(code_bytes);
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(varnames_tuple_dedup);
    #endif
    Py_DECREF(varnames_tuple);
    return code_obj;
}

/* DecompressString */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo) {
    PyObject *module = NULL, *decompress, *compressed_bytes, *decompressed;
    const char* module_name = algo == 3 ? "compression.zstd" : algo == 2 ? "bz2" : "zlib";
    PyObject *methodname = PyUnicode_FromString("decompress");
    if (unlikely(!methodname)) return NULL;
    #if __PYX_LIMITED_VERSION_HEX >= 0x030e0000
    if (algo == 3) {
        PyObject *fromlist = Py_BuildValue("[O]", methodname);
        if (unlikely(!fromlist)) goto bad;
        module = PyImport_ImportModuleLevel("compression.zstd", NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
    } else
    #endif
        module = PyImport_ImportModule(module_name);
    if (unlikely(!module)) goto import_failed;
    decompress = PyObject_GetAttr(module, methodname);
    if (unlikely(!decompress)) goto import_failed;
    {
        #ifdef __cplusplus
            char *memview_bytes = const_cast<char*>(s);
        #else
            #if defined(__clang__)
              #pragma clang diagnostic push
              #pragma clang diagnostic ignored "-Wcast-qual"
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic push
              #pragma GCC diagnostic ignored "-Wcast-qual"
            #endif
            char *memview_bytes = (char*) s;
            #if defined(__clang__)
              #pragma clang diagnostic pop
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic pop
            #endif
        #endif
        #if CYTHON_COMPILING_IN_LIMITED_API && !defined(PyBUF_READ)
        int memview_flags = 0x100;
        #else
        int memview_flags = PyBUF_READ;
        #endif
        compressed_bytes = PyMemoryView_FromMemory(memview_bytes, length, memview_flags);
    }
    if (unlikely(!compressed_bytes)) {
        Py_DECREF(decompress);
        goto bad;
    }
    decompressed = PyObject_CallFunctionObjArgs(decompress, compressed_bytes, NULL);
    Py_DECREF(compressed_bytes);
    Py_DECREF(decompress);
    Py_DECREF(module);
    Py_DECREF(methodname);
    return decompressed;
import_failed:
    PyErr_Format(PyExc_ImportError,
        "Failed to import '%.20s.decompress' - cannot initialise module strings. "
        "String compression was configured with the C macro 'CYTHON_COMPRESS_STRINGS=%d'.",
        module_name, algo);
bad:
    Py_XDECREF(module);
    Py_DECREF(methodname);
    return NULL;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    {
        const char* result;
        Py_ssize_t unicode_length;
        CYTHON_MAYBE_UNUSED_VAR(unicode_length); // only for __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (unlikely(PyArg_Parse(o, "s#", &result, length) < 0)) return NULL;
        #else
        result = PyUnicode_AsUTF8AndSize(o, length);
        #endif
        #if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        unicode_length = PyUnicode_GetLength(o);
        if (unlikely(unicode_length < 0)) return NULL;
        if (unlikely(unicode_length != *length)) {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
        #endif
        return result;
    }
#else
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
}
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
    if (PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
    if (PyByteArray_Check(o)) {
#if (CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS) || (CYTHON_COMPILING_IN_PYPY && (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE)))
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
#else
        *length = PyByteArray_Size(o);
        if (*length == -1) return NULL;
        return PyByteArray_AsString(o);
#endif
    } else
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_LongWrongResultType(PyObject* result) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(result));
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
    PyErr_Format(PyExc_TypeError,
                 "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ")",
                 result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  PyObject *res = NULL;
  if (likely(PyLong_Check(x)))
      return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  if (likely(m && m->nb_int)) {
      res = m->nb_int(x);
  }
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
      res = PyNumber_Long(x);
  }
#endif
  if (likely(res)) {
      if (unlikely(!PyLong_CheckExact(res))) {
          return __Pyx_PyNumber_LongWrongResultType(res);
      }
  }
  else if (!PyErr_Occurred()) {
      PyErr_SetString(PyExc_TypeError,
                      "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyLong_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyLong_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b) {
    CYTHON_UNUSED_VAR(b);
    return __Pyx_NewRef(Py_None);
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return __Pyx_NewRef(b ? Py_True: Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t ival) {
    return PyLong_FromSize_t(ival);
}


/* MultiPhaseInitModuleState */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
#ifndef CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#if (CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX >= 0x030C0000)
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 1
#else
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 0
#endif
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE && !CYTHON_ATOMICS
#error "Module state with PEP489 requires atomics. Currently that's one of\
 C11, C++11, gcc atomic intrinsics or MSVC atomic intrinsics"
#endif
#if !CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#define __Pyx_ModuleStateLookup_Lock()
#define __Pyx_ModuleStateLookup_Unlock()
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
static PyMutex __Pyx_ModuleStateLookup_mutex = {0};
#define __Pyx_ModuleStateLookup_Lock() PyMutex_Lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() PyMutex_Unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(__cplusplus) && __cplusplus >= 201103L
#include <mutex>
static std::mutex __Pyx_ModuleStateLookup_mutex;
#define __Pyx_ModuleStateLookup_Lock() __Pyx_ModuleStateLookup_mutex.lock()
#define __Pyx_ModuleStateLookup_Unlock() __Pyx_ModuleStateLookup_mutex.unlock()
#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ > 201112L) && !defined(__STDC_NO_THREADS__)
#include <threads.h>
static mtx_t __Pyx_ModuleStateLookup_mutex;
static once_flag __Pyx_ModuleStateLookup_mutex_once_flag = ONCE_FLAG_INIT;
static void __Pyx_ModuleStateLookup_initialize_mutex(void) {
    mtx_init(&__Pyx_ModuleStateLookup_mutex, mtx_plain);
}
#define __Pyx_ModuleStateLookup_Lock()\
  call_once(&__Pyx_ModuleStateLookup_mutex_once_flag, __Pyx_ModuleStateLookup_initialize_mutex);\
  mtx_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() mtx_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(HAVE_PTHREAD_H)
#include <pthread.h>
static pthread_mutex_t __Pyx_ModuleStateLookup_mutex = PTHREAD_MUTEX_INITIALIZER;
#define __Pyx_ModuleStateLookup_Lock() pthread_mutex_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() pthread_mutex_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(_WIN32)
#include <Windows.h>  // synchapi.h on its own doesn't work
static SRWLOCK __Pyx_ModuleStateLookup_mutex = SRWLOCK_INIT;
#define __Pyx_ModuleStateLookup_Lock() AcquireSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() ReleaseSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#else
#error "No suitable lock available for CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE.\
 Requires C standard >= C11, or C++ standard >= C++11,\
 or pthreads, or the Windows 32 API, or Python >= 3.13."
#endif
typedef struct {
    int64_t id;
    PyObject *module;
} __Pyx_InterpreterIdAndModule;
typedef struct {
    char interpreter_id_as_index;
    Py_ssize_t count;
    Py_ssize_t allocated;
    __Pyx_InterpreterIdAndModule table[1];
} __Pyx_ModuleStateLookupData;
#define __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE 32
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_int_type __Pyx_ModuleStateLookup_read_counter = 0;
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_ptr_type __Pyx_ModuleStateLookup_data = 0;
#else
static __Pyx_ModuleStateLookupData* __Pyx_ModuleStateLookup_data = NULL;
#endif
static __Pyx_InterpreterIdAndModule* __Pyx_State_FindModuleStateLookupTableLowerBound(
        __Pyx_InterpreterIdAndModule* table,
        Py_ssize_t count,
        int64_t interpreterId) {
    __Pyx_InterpreterIdAndModule* begin = table;
    __Pyx_InterpreterIdAndModule* end = begin + count;
    if (begin->id == interpreterId) {
        return begin;
    }
    while ((end - begin) > __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
        __Pyx_InterpreterIdAndModule* halfway = begin + (end - begin)/2;
        if (halfway->id == interpreterId) {
            return halfway;
        }
        if (halfway->id < interpreterId) {
            begin = halfway;
        } else {
            end = halfway;
        }
    }
    for (; begin < end; ++begin) {
        if (begin->id >= interpreterId) return begin;
    }
    return begin;
}
static PyObject *__Pyx_State_FindModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return NULL;
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData* data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
    {
        __pyx_atomic_incr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        if (likely(data)) {
            __Pyx_ModuleStateLookupData* new_data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_acquire(&__Pyx_ModuleStateLookup_data);
            if (likely(data == new_data)) {
                goto read_finished;
            }
        }
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        __Pyx_ModuleStateLookup_Lock();
        __pyx_atomic_incr_relaxed(&__Pyx_ModuleStateLookup_read_counter);
        data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
        __Pyx_ModuleStateLookup_Unlock();
    }
  read_finished:;
#else
    __Pyx_ModuleStateLookupData* data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_InterpreterIdAndModule* found = NULL;
    if (unlikely(!data)) goto end;
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            found = data->table+interpreter_id;
        }
    } else {
        found = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
    }
  end:
    {
        PyObject *result=NULL;
        if (found && found->id == interpreter_id) {
            result = found->module;
        }
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
#endif
        return result;
    }
}
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static void __Pyx_ModuleStateLookup_wait_until_no_readers(void) {
    while (__pyx_atomic_load(&__Pyx_ModuleStateLookup_read_counter) != 0);
}
#else
#define __Pyx_ModuleStateLookup_wait_until_no_readers()
#endif
static int __Pyx_State_AddModuleInterpIdAsIndex(__Pyx_ModuleStateLookupData **old_data, PyObject* module, int64_t interpreter_id) {
    Py_ssize_t to_allocate = (*old_data)->allocated;
    while (to_allocate <= interpreter_id) {
        if (to_allocate == 0) to_allocate = 1;
        else to_allocate *= 2;
    }
    __Pyx_ModuleStateLookupData *new_data = *old_data;
    if (to_allocate != (*old_data)->allocated) {
         new_data = (__Pyx_ModuleStateLookupData *)realloc(
            *old_data,
            sizeof(__Pyx_ModuleStateLookupData)+(to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
        if (!new_data) {
            PyErr_NoMemory();
            return -1;
        }
        for (Py_ssize_t i = new_data->allocated; i < to_allocate; ++i) {
            new_data->table[i].id = i;
            new_data->table[i].module = NULL;
        }
        new_data->allocated = to_allocate;
    }
    new_data->table[interpreter_id].module = module;
    if (new_data->count < interpreter_id+1) {
        new_data->count = interpreter_id+1;
    }
    *old_data = new_data;
    return 0;
}
static void __Pyx_State_ConvertFromInterpIdAsIndex(__Pyx_ModuleStateLookupData *data) {
    __Pyx_InterpreterIdAndModule *read = data->table;
    __Pyx_InterpreterIdAndModule *write = data->table;
    __Pyx_InterpreterIdAndModule *end = read + data->count;
    for (; read<end; ++read) {
        if (read->module) {
            write->id = read->id;
            write->module = read->module;
            ++write;
        }
    }
    data->count = write - data->table;
    for (; write<end; ++write) {
        write->id = 0;
        write->module = NULL;
    }
    data->interpreter_id_as_index = 0;
}
static int __Pyx_State_AddModule(PyObject* module, CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    int result = 0;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *old_data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *old_data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_ModuleStateLookupData *new_data = old_data;
    if (!new_data) {
        new_data = (__Pyx_ModuleStateLookupData *)calloc(1, sizeof(__Pyx_ModuleStateLookupData));
        if (!new_data) {
            result = -1;
            PyErr_NoMemory();
            goto end;
        }
        new_data->allocated = 1;
        new_data->interpreter_id_as_index = 1;
    }
    __Pyx_ModuleStateLookup_wait_until_no_readers();
    if (new_data->interpreter_id_as_index) {
        if (interpreter_id < __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
            result = __Pyx_State_AddModuleInterpIdAsIndex(&new_data, module, interpreter_id);
            goto end;
        }
        __Pyx_State_ConvertFromInterpIdAsIndex(new_data);
    }
    {
        Py_ssize_t insert_at = 0;
        {
            __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
                new_data->table, new_data->count, interpreter_id);
            assert(lower_bound);
            insert_at = lower_bound - new_data->table;
            if (unlikely(insert_at < new_data->count && lower_bound->id == interpreter_id)) {
                lower_bound->module = module;
                goto end;  // already in table, nothing more to do
            }
        }
        if (new_data->count+1 >= new_data->allocated) {
            Py_ssize_t to_allocate = (new_data->count+1)*2;
            new_data =
                (__Pyx_ModuleStateLookupData*)realloc(
                    new_data,
                    sizeof(__Pyx_ModuleStateLookupData) +
                    (to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
            if (!new_data) {
                result = -1;
                new_data = old_data;
                PyErr_NoMemory();
                goto end;
            }
            new_data->allocated = to_allocate;
        }
        ++new_data->count;
        int64_t last_id = interpreter_id;
        PyObject *last_module = module;
        for (Py_ssize_t i=insert_at; i<new_data->count; ++i) {
            int64_t current_id = new_data->table[i].id;
            new_data->table[i].id = last_id;
            last_id = current_id;
            PyObject *current_module = new_data->table[i].module;
            new_data->table[i].module = last_module;
            last_module = current_module;
        }
    }
  end:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, new_data);
#else
    __Pyx_ModuleStateLookup_data = new_data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return result;
}
static int __Pyx_State_RemoveModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *data = __Pyx_ModuleStateLookup_data;
#endif
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            data->table[interpreter_id].module = NULL;
        }
        goto done;
    }
    {
        __Pyx_ModuleStateLookup_wait_until_no_readers();
        __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
        if (!lower_bound) goto done;
        if (lower_bound->id != interpreter_id) goto done;
        __Pyx_InterpreterIdAndModule *end = data->table+data->count;
        for (;lower_bound<end-1; ++lower_bound) {
            lower_bound->id = (lower_bound+1)->id;
            lower_bound->module = (lower_bound+1)->module;
        }
    }
    --data->count;
    if (data->count == 0) {
        free(data);
        data = NULL;
    }
  done:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, data);
#else
    __Pyx_ModuleStateLookup_data = data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return 0;
}
#endif

/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
