/* Generated by Cython 3.2.4 */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
/* InitLimitedAPI */
#if defined(Py_LIMITED_API)
  #if !defined(CYTHON_LIMITED_API)
  #define CYTHON_LIMITED_API 1
  #endif
#elif defined(CYTHON_LIMITED_API)
  #ifdef _MSC_VER
  #pragma message ("Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.")
  #else
  #warning Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x03080000
    #error Cython requires Python 3.8+.
#else
#define __PYX_ABI_VERSION "3_2_4"
#define CYTHON_HEX_VERSION 0x030204F0
#define CYTHON_FUTURE_DIVISION 1
/* CModulePreamble */
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 1
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 1
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PYPY_VERSION_NUM >= 0x07030C00)
  #endif
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_NUM >= 0x07031100)
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND (__PYX_LIMITED_VERSION_HEX >= 0x030A0000)
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #ifndef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 1
  #endif
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #ifdef Py_GIL_DISABLED
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 1
  #else
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #endif
  #if PY_VERSION_HEX < 0x030A0000
    #undef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #elif !defined(CYTHON_USE_TYPE_SLOTS)
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLIST_INTERNALS)
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #elif !defined(CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL 0
  #elif !defined(CYTHON_FAST_GIL)
    #define CYTHON_FAST_GIL (PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_SYS_MONITORING
    #define CYTHON_USE_SYS_MONITORING (PY_VERSION_HEX >= 0x030d00B1)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5 && !CYTHON_USE_MODULE_STATE)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS (!CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
  #if defined(CYTHON_IMMORTAL_CONSTANTS) && PY_VERSION_HEX < 0x030C0000
    #undef CYTHON_IMMORTAL_CONSTANTS
    #define CYTHON_IMMORTAL_CONSTANTS 0  // definitely won't work
  #elif !defined(CYTHON_IMMORTAL_CONSTANTS)
    #define CYTHON_IMMORTAL_CONSTANTS (PY_VERSION_HEX >= 0x030C0000 && !CYTHON_USE_MODULE_STATE && CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
#endif
#ifndef CYTHON_COMPRESS_STRINGS
  #define CYTHON_COMPRESS_STRINGS 1
#endif
#ifndef CYTHON_FAST_PYCCALL
#define CYTHON_FAST_PYCCALL  CYTHON_FAST_PYCALL
#endif
#ifndef CYTHON_VECTORCALL
#if CYTHON_COMPILING_IN_LIMITED_API
#define CYTHON_VECTORCALL  (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
#else
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL)
#endif
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#include <stdint.h>
typedef uintptr_t  __pyx_uintptr_t;
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifndef Py_UNREACHABLE
  #define Py_UNREACHABLE()  assert(0); abort()
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

/* CInitCode */
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

/* PythonCompatibility */
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#define __Pyx_BUILTIN_MODULE_NAME "builtins"
#define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    #ifndef CO_OPTIMIZED
    static int CO_OPTIMIZED;
    #endif
    #ifndef CO_NEWLOCALS
    static int CO_NEWLOCALS;
    #endif
    #ifndef CO_VARARGS
    static int CO_VARARGS;
    #endif
    #ifndef CO_VARKEYWORDS
    static int CO_VARKEYWORDS;
    #endif
    #ifndef CO_ASYNC_GENERATOR
    static int CO_ASYNC_GENERATOR;
    #endif
    #ifndef CO_GENERATOR
    static int CO_GENERATOR;
    #endif
    #ifndef CO_COROUTINE
    static int CO_COROUTINE;
    #endif
#else
    #ifndef CO_COROUTINE
      #define CO_COROUTINE 0x80
    #endif
    #ifndef CO_ASYNC_GENERATOR
      #define CO_ASYNC_GENERATOR 0x200
    #endif
#endif
static int __Pyx_init_co_variables(void);
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef Py_TPFLAGS_IMMUTABLETYPE
  #define Py_TPFLAGS_IMMUTABLETYPE (1UL << 8)
#endif
#ifndef Py_TPFLAGS_DISALLOW_INSTANTIATION
  #define Py_TPFLAGS_DISALLOW_INSTANTIATION (1UL << 7)
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#ifndef METH_FASTCALL
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void (*cfunc)(void)) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if PY_VERSION_HEX < 0x03090000 || (CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000)
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#elif CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) GraalPyFrame_SetLineNumber((frame), (lineno))
#elif CYTHON_COMPILING_IN_GRAAL
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) _PyFrame_SetLineNumber((frame), (lineno))
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#endif
#if CYTHON_USE_MODULE_STATE
static CYTHON_INLINE void *__Pyx__PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#define __Pyx_PyModule_GetState(o) (__pyx_mstatetype *)__Pyx__PyModule_GetState(o)
#else
#define __Pyx_PyModule_GetState(op) ((void)op,__pyx_mstate_global)
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE((PyObject *) obj), name, func_ctype)
#define __Pyx_PyObject_TryGetSlot(obj, name, func_ctype) __Pyx_PyType_TryGetSlot(Py_TYPE(obj), name, func_ctype)
#define __Pyx_PyObject_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#define __Pyx_PyObject_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype) __Pyx_PyType_GetSlot(type, name, func_ctype)
  #define __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype) (((type)->sub) ? ((type)->sub->name) : NULL)
  #define __Pyx_PyType_TryGetSubSlot(type, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype)
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype)\
    ((__PYX_LIMITED_VERSION_HEX >= 0x030A0000 ||\
     (PyType_GetFlags(type) & Py_TPFLAGS_HEAPTYPE) || __Pyx_get_runtime_version() >= 0x030A0000) ?\
     __Pyx_PyType_GetSlot(type, name, func_ctype) : NULL)
  #define __Pyx_PyType_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSlot(obj, name, func_ctype)
  #define __Pyx_PyType_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSlot(obj, name, func_ctype)
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
#define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
#endif
#define __Pyx_PyObject_GetIterNextFunc(iterator)  __Pyx_PyObject_GetSlot(iterator, tp_iternext, iternextfunc)
#if CYTHON_USE_TYPE_SPECS
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#else
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains)
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && PyUnstable_Object_IsUniquelyReferenced(obj)) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#elif CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
enum __Pyx_ReferenceSharing {
  __Pyx_ReferenceSharing_DefinitelyUnique, // We created it so we know it's unshared - no need to check
  __Pyx_ReferenceSharing_OwnStrongReference,
  __Pyx_ReferenceSharing_FunctionArgument,
  __Pyx_ReferenceSharing_SharedReference, // Never trust it to be unshared because it's a global or similar
};
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && PY_VERSION_HEX >= 0x030E0000
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing)\
    (sharing == __Pyx_ReferenceSharing_DefinitelyUnique ? 1 :\
      (sharing == __Pyx_ReferenceSharing_FunctionArgument ? PyUnstable_Object_IsUniqueReferencedTemporary(o) :\
      (sharing == __Pyx_ReferenceSharing_OwnStrongReference ? PyUnstable_Object_IsUniquelyReferenced(o) : 0)))
#elif (CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING) || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)sharing), Py_REFCNT(o) == 1)
#else
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)o), ((void)sharing), 0)
#endif
#if CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyList_GetItemRef(o, i) (likely((i) >= 0) ? PySequence_GetItem(o, i) : (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
  #else
    #define __Pyx_PyList_GetItemRef(o, i) PySequence_ITEM(o, i)
  #endif
#elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #else
    #define __Pyx_PyList_GetItemRef(o, i) __Pyx_XNewRef(PyList_GetItem(o, i))
  #endif
#else
  #define __Pyx_PyList_GetItemRef(o, i) __Pyx_NewRef(PyList_GET_ITEM(o, i))
#endif
#if CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS && !CYTHON_COMPILING_IN_LIMITED_API && CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) (__Pyx_IS_UNIQUELY_REFERENCED(o, unsafe_shared) ?\
    __Pyx_NewRef(PyList_GET_ITEM(o, i)) : __Pyx_PyList_GetItemRef(o, i))
#else
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) __Pyx_PyList_GetItemRef(o, i)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyDict_GetItemRef(dict, key, result) PyDict_GetItemRef(dict, key, result)
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyObject_GetItem(dict, key);
  if (*result == NULL) {
    if (PyErr_ExceptionMatches(PyExc_KeyError)) {
      PyErr_Clear();
      return 0;
    }
    return -1;
  }
  return 1;
}
#else
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyDict_GetItemWithError(dict, key);
  if (*result == NULL) {
    return PyErr_Occurred() ? -1 : 0;
  }
  Py_INCREF(*result);
  return 1;
}
#endif
#if defined(CYTHON_DEBUG_VISIT_CONST) && CYTHON_DEBUG_VISIT_CONST
  #define __Pyx_VISIT_CONST(obj)  Py_VISIT(obj)
#else
  #define __Pyx_VISIT_CONST(obj)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GET_ITEM(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GET_ITEM(o, i)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GetItem(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GetItem(o, i)
#endif
#if CYTHON_ASSUME_SAFE_SIZE
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GET_LENGTH(o)
#else
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GetLength(o)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_InternFromString)
  #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
#endif
#define __Pyx_PyLong_FromHash_t PyLong_FromSsize_t
#define __Pyx_PyLong_AsHash_t   __Pyx_PyIndex_AsSsize_t
#if __PYX_LIMITED_VERSION_HEX >= 0x030A0000
    #define __Pyx_PySendResult PySendResult
#else
    typedef enum {
        PYGEN_RETURN = 0,
        PYGEN_ERROR = -1,
        PYGEN_NEXT = 1,
    } __Pyx_PySendResult;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030A00A3
  typedef __Pyx_PySendResult (*__Pyx_pyiter_sendfunc)(PyObject *iter, PyObject *value, PyObject **result);
#else
  #define __Pyx_pyiter_sendfunc sendfunc
#endif
#if !CYTHON_USE_AM_SEND
#define __PYX_HAS_PY_AM_SEND 0
#elif __PYX_LIMITED_VERSION_HEX >= 0x030A0000
#define __PYX_HAS_PY_AM_SEND 1
#else
#define __PYX_HAS_PY_AM_SEND 2  // our own backported implementation
#endif
#if __PYX_HAS_PY_AM_SEND < 2
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
        __Pyx_pyiter_sendfunc am_send;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_SlotTpAsAsync(s) ((PyAsyncMethods*)(s))
#endif
#if CYTHON_USE_AM_SEND && PY_VERSION_HEX < 0x030A00F0
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (1UL << 21)
#else
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (0)
#endif
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyInterpreterState_Get() PyInterpreterState_Get()
#else
#define __Pyx_PyInterpreterState_Get() PyThreadState_Get()->interp
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030A0000
#ifdef __cplusplus
extern "C"
#endif
PyAPI_FUNC(void *) PyMem_Calloc(size_t nelem, size_t elsize);
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_init_co_variable(PyObject *inspect, const char* name, int *write_to) {
    int value;
    PyObject *py_value = PyObject_GetAttrString(inspect, name);
    if (!py_value) return 0;
    value = (int) PyLong_AsLong(py_value);
    Py_DECREF(py_value);
    *write_to = value;
    return value != -1 || !PyErr_Occurred();
}
static int __Pyx_init_co_variables(void) {
    PyObject *inspect;
    int result;
    inspect = PyImport_ImportModule("inspect");
    result =
#if !defined(CO_OPTIMIZED)
        __Pyx_init_co_variable(inspect, "CO_OPTIMIZED", &CO_OPTIMIZED) &&
#endif
#if !defined(CO_NEWLOCALS)
        __Pyx_init_co_variable(inspect, "CO_NEWLOCALS", &CO_NEWLOCALS) &&
#endif
#if !defined(CO_VARARGS)
        __Pyx_init_co_variable(inspect, "CO_VARARGS", &CO_VARARGS) &&
#endif
#if !defined(CO_VARKEYWORDS)
        __Pyx_init_co_variable(inspect, "CO_VARKEYWORDS", &CO_VARKEYWORDS) &&
#endif
#if !defined(CO_ASYNC_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_ASYNC_GENERATOR", &CO_ASYNC_GENERATOR) &&
#endif
#if !defined(CO_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_GENERATOR", &CO_GENERATOR) &&
#endif
#if !defined(CO_COROUTINE)
        __Pyx_init_co_variable(inspect, "CO_COROUTINE", &CO_COROUTINE) &&
#endif
        1;
    Py_DECREF(inspect);
    return result ? 0 : -1;
}
#else
static int __Pyx_init_co_variables(void) {
    return 0;  // It's a limited API-only feature
}
#endif

/* MathInitCode */
#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #ifndef _USE_MATH_DEFINES
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#ifndef CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#define CYTHON_CLINE_IN_TRACEBACK_RUNTIME 0
#endif
#ifndef CYTHON_CLINE_IN_TRACEBACK
#define CYTHON_CLINE_IN_TRACEBACK CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#endif
#if CYTHON_CLINE_IN_TRACEBACK
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; __pyx_clineno = __LINE__; (void) __pyx_clineno; }
#else
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; (void) __pyx_clineno; }
#endif
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__pyverbs__cmid
#define __PYX_HAVE_API__pyverbs__cmid
/* Early includes */
#include <stdint.h>
#include <rdma/rdma_cma.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/time.h>
#include <infiniband/verbs.h>
#include <infiniband/verbs_api.h>
#include <infiniband/driver.h>
#include <infiniband/tm_types.h>
#include "infiniband/verbs.h"
#include "infiniband/driver.h"
#include <infiniband/ib.h>
#include <rdma/rdma_verbs.h>
#include <string.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifdef CYTHON_FREETHREADING_COMPATIBLE
#if CYTHON_FREETHREADING_COMPATIBLE
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_NOT_USED
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AS_STRING(s)
#else
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AsString(s)
#endif
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
static CYTHON_INLINE PyObject *__Pyx_NewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_NewRef)
    return Py_NewRef(obj);
#else
    Py_INCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_XNewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_XNewRef)
    return Py_XNewRef(obj);
#else
    Py_XINCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b);
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __Pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AS_DOUBLE(x)
#else
#define __Pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AsDouble(x)
#endif
#define __Pyx_PyFloat_AsFloat(x) ((float) __Pyx_PyFloat_AsDouble(x))
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#elif __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeASCII(c_str, size, NULL)
#else
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
/* PretendToInitialize */
#ifdef __cplusplus
#if __cplusplus > 201103L
#include <type_traits>
#endif
template <typename T>
static void __Pyx_pretend_to_initialize(T* ptr) {
#if __cplusplus > 201103L
    if ((std::is_trivially_default_constructible<T>::value))
#endif
        *ptr = T();
    (void)ptr;
}
#else
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }
#endif


#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * const __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char* const __pyx_f[] = {
  "pyverbs/cmid.pyx",
  "EnumTypeToPy",
  "pyverbs/qp.pxd",
  "pyverbs/addr.pxd",
  "pyverbs/device.pxd",
  "pyverbs/pd.pxd",
  "pyverbs/mr.pxd",
  "pyverbs/cq.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* Atomics.proto (used by UnpackUnboundCMethod) */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __PYX_GET_CYTHON_COMPILING_IN_CPYTHON_FREETHREADING() CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __pyx_atomic_int_type int
#define __pyx_nonatomic_int_type int
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__))
    #include <stdatomic.h>
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)))
    #include <atomic>
#endif
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__) &&\
                       ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type atomic_int
    #define __pyx_atomic_ptr_type atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) atomic_fetch_add_explicit(value, 1, memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) atomic_fetch_add_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) atomic_fetch_sub_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) atomic_load(value)
    #define __pyx_atomic_store(value, new_value) atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) atomic_load_explicit(value, memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) atomic_load_explicit(value, memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C atomics"
    #endif
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)) &&\
                    ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type std::atomic_int
    #define __pyx_atomic_ptr_type std::atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) std::atomic_fetch_sub_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) std::atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) std::atomic_load(value)
    #define __pyx_atomic_store(value, new_value) std::atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) std::atomic_load_explicit(value, std::memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) std::atomic_load_explicit(value, std::memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) std::atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C++ atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C++ atomics"
    #endif
#elif CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_ptr_type void*
    #define __pyx_nonatomic_ptr_type void*
    #define __pyx_atomic_incr_relaxed(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) __sync_fetch_and_sub(value, 1)
    #define __pyx_atomic_sub(value, arg) __sync_fetch_and_sub(value, arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_store(value, new_value) __sync_lock_test_and_set(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_load_acquire(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) __sync_lock_test_and_set(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_nonatomic_ptr_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER)
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #define __pyx_atomic_ptr_type void*
    #undef __pyx_nonatomic_int_type
    #define __pyx_nonatomic_int_type long
    #define __pyx_nonatomic_ptr_type void*
    #pragma intrinsic (_InterlockedExchangeAdd, _InterlockedExchange, _InterlockedCompareExchange, _InterlockedCompareExchangePointer, _InterlockedExchangePointer)
    #define __pyx_atomic_incr_relaxed(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) _InterlockedExchangeAdd(value, -1)
    #define __pyx_atomic_sub(value, arg) _InterlockedExchangeAdd(value, -arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = _InterlockedCompareExchange(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) _InterlockedExchangeAdd(value, 0)
    #define __pyx_atomic_store(value, new_value) _InterlockedExchange(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) *(void * volatile *)value
    #define __pyx_atomic_pointer_load_acquire(value) _InterlockedCompareExchangePointer(value, 0, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) _InterlockedExchangePointer(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_atomic_ptr_type old = _InterlockedCompareExchangePointer(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif

/* CriticalSectionsDefinition.proto (used by CriticalSections) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection void*
#define __Pyx_PyCriticalSection2 void*
#define __Pyx_PyCriticalSection_End(cs)
#define __Pyx_PyCriticalSection2_End(cs)
#else
#define __Pyx_PyCriticalSection PyCriticalSection
#define __Pyx_PyCriticalSection2 PyCriticalSection2
#define __Pyx_PyCriticalSection_End PyCriticalSection_End
#define __Pyx_PyCriticalSection2_End PyCriticalSection2_End
#endif

/* CriticalSections.proto (used by ParseKeywordsImpl) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection_Begin(cs, arg) (void)(cs)
#define __Pyx_PyCriticalSection2_Begin(cs, arg1, arg2) (void)(cs)
#else
#define __Pyx_PyCriticalSection_Begin PyCriticalSection_Begin
#define __Pyx_PyCriticalSection2_Begin PyCriticalSection2_Begin
#endif
#if PY_VERSION_HEX < 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_BEGIN_CRITICAL_SECTION(o) {
#define __Pyx_END_CRITICAL_SECTION() }
#else
#define __Pyx_BEGIN_CRITICAL_SECTION Py_BEGIN_CRITICAL_SECTION
#define __Pyx_END_CRITICAL_SECTION Py_END_CRITICAL_SECTION
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* IncludeStructmemberH.proto (used by FixUpExtensionType) */
#include <structmember.h>

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_7pyverbs_4base_PyverbsObject;
struct __pyx_obj_7pyverbs_4base_PyverbsCM;
struct __pyx_obj_7pyverbs_2qp_QPCap;
struct __pyx_obj_7pyverbs_2qp_QPInitAttr;
struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx;
struct __pyx_obj_7pyverbs_2qp_QPAttr;
struct __pyx_obj_7pyverbs_2qp_QP;
struct __pyx_obj_7pyverbs_2qp_DataBuffer;
struct __pyx_obj_7pyverbs_2qp_QPEx;
struct __pyx_obj_7pyverbs_2qp_ECE;
struct __pyx_obj_7pyverbs_4addr_GID;
struct __pyx_obj_7pyverbs_4addr_GRH;
struct __pyx_obj_7pyverbs_4addr_GlobalRoute;
struct __pyx_obj_7pyverbs_4addr_AHAttr;
struct __pyx_obj_7pyverbs_4addr_AH;
struct __pyx_obj_7pyverbs_6device_Context;
struct __pyx_obj_7pyverbs_6device_DeviceAttr;
struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput;
struct __pyx_obj_7pyverbs_6device_ODPCaps;
struct __pyx_obj_7pyverbs_6device_RSSCaps;
struct __pyx_obj_7pyverbs_6device_PacketPacingCaps;
struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps;
struct __pyx_obj_7pyverbs_6device_TMCaps;
struct __pyx_obj_7pyverbs_6device_CQModerationCaps;
struct __pyx_obj_7pyverbs_6device_TSOCaps;
struct __pyx_obj_7pyverbs_6device_DeviceAttrEx;
struct __pyx_obj_7pyverbs_6device_AllocDmAttr;
struct __pyx_obj_7pyverbs_6device_DM;
struct __pyx_obj_7pyverbs_6device_PortAttr;
struct __pyx_obj_7pyverbs_6device_GIDEntry;
struct __pyx_obj_7pyverbs_6device_AsyncEvent;
struct __pyx_obj_7pyverbs_6device_FdArr;
struct __pyx_obj_7pyverbs_2pd_PD;
struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr;
struct __pyx_obj_7pyverbs_2pd_ParentDomain;
struct __pyx_obj_7pyverbs_2pd_ParentDomainContext;
struct __pyx_obj_7pyverbs_2mr_MR;
struct __pyx_obj_7pyverbs_2mr_MREx;
struct __pyx_obj_7pyverbs_2mr_MWBindInfo;
struct __pyx_obj_7pyverbs_2mr_MWBind;
struct __pyx_obj_7pyverbs_2mr_MW;
struct __pyx_obj_7pyverbs_2mr_DMMR;
struct __pyx_obj_7pyverbs_2mr_DmaBufMR;
struct __pyx_obj_7pyverbs_2mr_DmaHandleInitAttr;
struct __pyx_obj_7pyverbs_2mr_DMAHandle;
struct __pyx_obj_7pyverbs_2cq_CompChannel;
struct __pyx_obj_7pyverbs_2cq_CQ;
struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx;
struct __pyx_obj_7pyverbs_2cq_CQEX;
struct __pyx_obj_7pyverbs_2cq_WC;
struct __pyx_obj_7pyverbs_2cq_PollCqAttr;
struct __pyx_obj_7pyverbs_2cq_WcTmInfo;
struct __pyx_obj_7pyverbs_4cmid_CMID;
struct __pyx_obj_7pyverbs_4cmid_CMEventChannel;
struct __pyx_obj_7pyverbs_4cmid_CMEvent;
struct __pyx_obj_7pyverbs_4cmid_AddrInfo;
struct __pyx_obj_7pyverbs_4cmid_ConnParam;
struct __pyx_obj_7pyverbs_4cmid_UDParam;
struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx;

/* "pyverbs/base.pxd":6
 * #cython: language_level=3
 * 
 * cdef class PyverbsObject(object):             # <<<<<<<<<<<<<<
 *     cdef object __weakref__
 *     cdef object logger
*/
struct __pyx_obj_7pyverbs_4base_PyverbsObject {
  PyObject_HEAD
  PyObject *__weakref__;
  PyObject *logger;
};


/* "pyverbs/base.pxd":10
 *     cdef object logger
 * 
 * cdef class PyverbsCM(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cpdef close(self)
 * 
*/
struct __pyx_obj_7pyverbs_4base_PyverbsCM {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM *__pyx_vtab;
};


/* "pyverbs/qp.pxd":9
 * cimport pyverbs.libibverbs as v
 * 
 * cdef class QPCap(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_qp_cap cap
 * 
*/
struct __pyx_obj_7pyverbs_2qp_QPCap {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_qp_cap cap;
};


/* "pyverbs/qp.pxd":12
 *     cdef v.ibv_qp_cap cap
 * 
 * cdef class QPInitAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_qp_init_attr attr
 *     cdef object scq
*/
struct __pyx_obj_7pyverbs_2qp_QPInitAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_qp_init_attr attr;
  PyObject *scq;
  PyObject *rcq;
  PyObject *srq;
};


/* "pyverbs/qp.pxd":18
 *     cdef object srq
 * 
 * cdef class QPInitAttrEx(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_qp_init_attr_ex attr
 *     cdef object scq
*/
struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_qp_init_attr_ex attr;
  PyObject *scq;
  PyObject *rcq;
  PyObject *_pd;
  PyObject *xrcd;
  PyObject *srq;
  PyObject *ind_table;
};


/* "pyverbs/qp.pxd":27
 *     cdef object ind_table
 * 
 * cdef class QPAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_qp_attr attr
 * 
*/
struct __pyx_obj_7pyverbs_2qp_QPAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_qp_attr attr;
};


/* "pyverbs/qp.pxd":30
 *     cdef v.ibv_qp_attr attr
 * 
 * cdef class QP(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_qp *qp
 *     cdef int type
*/
struct __pyx_obj_7pyverbs_2qp_QP {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_qp *qp;
  int type;
  int state;
  PyObject *pd;
  PyObject *context;
  PyObject *xrcd;
  PyObject *scq;
  PyObject *rcq;
  PyObject *mws;
  PyObject *srq;
  PyObject *flows;
  PyObject *dr_actions;
};


/* "pyverbs/qp.pxd":47
 *     cdef add_ref(self, obj)
 * 
 * cdef class DataBuffer(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_data_buf data
 * 
*/
struct __pyx_obj_7pyverbs_2qp_DataBuffer {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_data_buf data;
};


/* "pyverbs/qp.pxd":50
 *     cdef v.ibv_data_buf data
 * 
 * cdef class QPEx(QP):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_qp_ex *qp_ex
 *     cdef object ind_table
*/
struct __pyx_obj_7pyverbs_2qp_QPEx {
  struct __pyx_obj_7pyverbs_2qp_QP __pyx_base;
  struct ibv_qp_ex *qp_ex;
  PyObject *ind_table;
};


/* "pyverbs/qp.pxd":54
 *     cdef object ind_table
 * 
 * cdef class ECE(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_ece ece
*/
struct __pyx_obj_7pyverbs_2qp_ECE {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_ece ece;
};


/* "pyverbs/addr.pxd":11
 * 
 * 
 * cdef class GID(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_gid gid
 * 
*/
struct __pyx_obj_7pyverbs_4addr_GID {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  union ibv_gid gid;
};


/* "pyverbs/addr.pxd":14
 *     cdef v.ibv_gid gid
 * 
 * cdef class GRH(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_grh grh
 * 
*/
struct __pyx_obj_7pyverbs_4addr_GRH {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_grh grh;
};


/* "pyverbs/addr.pxd":17
 *     cdef v.ibv_grh grh
 * 
 * cdef class GlobalRoute(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_global_route gr
 * 
*/
struct __pyx_obj_7pyverbs_4addr_GlobalRoute {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_global_route gr;
};


/* "pyverbs/addr.pxd":20
 *     cdef v.ibv_global_route gr
 * 
 * cdef class AHAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_ah_attr ah_attr
 *     cdef init_from_ud_param(self, UDParam udparam)
*/
struct __pyx_obj_7pyverbs_4addr_AHAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct __pyx_vtabstruct_7pyverbs_4addr_AHAttr *__pyx_vtab;
  struct ibv_ah_attr ah_attr;
};


/* "pyverbs/addr.pxd":24
 *     cdef init_from_ud_param(self, UDParam udparam)
 * 
 * cdef class AH(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_ah *ah
 *     cdef object pd
*/
struct __pyx_obj_7pyverbs_4addr_AH {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_ah *ah;
  PyObject *pd;
};


/* "pyverbs/device.pxd":10
 * 
 * 
 * cdef class Context(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_context *context
 *     cdef v.ibv_device *device
*/
struct __pyx_obj_7pyverbs_6device_Context {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_context *context;
  struct ibv_device *device;
  PyObject *name;
  PyObject *pds;
  PyObject *dms;
  PyObject *ccs;
  PyObject *cqs;
  PyObject *qps;
  PyObject *xrcds;
  PyObject *vars;
  PyObject *uars;
  PyObject *pps;
  PyObject *sched_nodes;
  PyObject *sched_leafs;
  PyObject *dr_domains;
  PyObject *wqs;
  PyObject *rwq_ind_tbls;
  PyObject *crypto_logins;
  PyObject *event_channels;
  PyObject *dmahs;
};


/* "pyverbs/device.pxd":33
 *     cdef object dmahs
 * 
 * cdef class DeviceAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_device_attr dev_attr
 * 
*/
struct __pyx_obj_7pyverbs_6device_DeviceAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_device_attr dev_attr;
};


/* "pyverbs/device.pxd":36
 *     cdef v.ibv_device_attr dev_attr
 * 
 * cdef class QueryDeviceExInput(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_query_device_ex_input input
 * 
*/
struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_query_device_ex_input input;
};


/* "pyverbs/device.pxd":39
 *     cdef v.ibv_query_device_ex_input input
 * 
 * cdef class ODPCaps(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_odp_caps odp_caps
 *     cdef object xrc_odp_caps
*/
struct __pyx_obj_7pyverbs_6device_ODPCaps {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_odp_caps odp_caps;
  PyObject *xrc_odp_caps;
};


/* "pyverbs/device.pxd":43
 *     cdef object xrc_odp_caps
 * 
 * cdef class RSSCaps(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_rss_caps rss_caps
 * 
*/
struct __pyx_obj_7pyverbs_6device_RSSCaps {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_rss_caps rss_caps;
};


/* "pyverbs/device.pxd":46
 *     cdef v.ibv_rss_caps rss_caps
 * 
 * cdef class PacketPacingCaps(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_packet_pacing_caps packet_pacing_caps
 * 
*/
struct __pyx_obj_7pyverbs_6device_PacketPacingCaps {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_packet_pacing_caps packet_pacing_caps;
};


/* "pyverbs/device.pxd":49
 *     cdef v.ibv_packet_pacing_caps packet_pacing_caps
 * 
 * cdef class PCIAtomicCaps(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_pci_atomic_caps caps
 * 
*/
struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_pci_atomic_caps caps;
};


/* "pyverbs/device.pxd":52
 *     cdef v.ibv_pci_atomic_caps caps
 * 
 * cdef class TMCaps(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_tm_caps tm_caps
 * 
*/
struct __pyx_obj_7pyverbs_6device_TMCaps {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_tm_caps tm_caps;
};


/* "pyverbs/device.pxd":55
 *     cdef v.ibv_tm_caps tm_caps
 * 
 * cdef class CQModerationCaps(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_cq_moderation_caps cq_mod_caps
 * 
*/
struct __pyx_obj_7pyverbs_6device_CQModerationCaps {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_cq_moderation_caps cq_mod_caps;
};


/* "pyverbs/device.pxd":58
 *     cdef v.ibv_cq_moderation_caps cq_mod_caps
 * 
 * cdef class TSOCaps(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_tso_caps tso_caps
 * 
*/
struct __pyx_obj_7pyverbs_6device_TSOCaps {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_tso_caps tso_caps;
};


/* "pyverbs/device.pxd":61
 *     cdef v.ibv_tso_caps tso_caps
 * 
 * cdef class DeviceAttrEx(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_device_attr_ex dev_attr
 * 
*/
struct __pyx_obj_7pyverbs_6device_DeviceAttrEx {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_device_attr_ex dev_attr;
};


/* "pyverbs/device.pxd":64
 *     cdef v.ibv_device_attr_ex dev_attr
 * 
 * cdef class AllocDmAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_alloc_dm_attr alloc_dm_attr
 * 
*/
struct __pyx_obj_7pyverbs_6device_AllocDmAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_alloc_dm_attr alloc_dm_attr;
};


/* "pyverbs/device.pxd":67
 *     cdef v.ibv_alloc_dm_attr alloc_dm_attr
 * 
 * cdef class DM(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_dm *dm
 *     cdef object dm_mrs
*/
struct __pyx_obj_7pyverbs_6device_DM {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_dm *dm;
  PyObject *dm_mrs;
  PyObject *context;
  PyObject *_is_imported;
};


/* "pyverbs/device.pxd":74
 *     cdef add_ref(self, obj)
 * 
 * cdef class PortAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_port_attr attr
 * 
*/
struct __pyx_obj_7pyverbs_6device_PortAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_port_attr attr;
};


/* "pyverbs/device.pxd":77
 *     cdef v.ibv_port_attr attr
 * 
 * cdef class GIDEntry(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_gid_entry entry
 * 
*/
struct __pyx_obj_7pyverbs_6device_GIDEntry {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_gid_entry entry;
};


/* "pyverbs/device.pxd":80
 *     cdef v.ibv_gid_entry entry
 * 
 * cdef class AsyncEvent(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_async_event event
 * 
*/
struct __pyx_obj_7pyverbs_6device_AsyncEvent {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_async_event event;
};


/* "pyverbs/device.pxd":83
 *     cdef v.ibv_async_event event
 * 
 * cdef class FdArr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_fd_arr attr
 *     cpdef close(self)
*/
struct __pyx_obj_7pyverbs_6device_FdArr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct __pyx_vtabstruct_7pyverbs_6device_FdArr *__pyx_vtab;
  struct ibv_fd_arr attr;
};


/* "pyverbs/pd.pxd":12
 * 
 * 
 * cdef class PD(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_pd *pd
 *     cdef Context ctx
*/
struct __pyx_obj_7pyverbs_2pd_PD {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_pd *pd;
  struct __pyx_obj_7pyverbs_6device_Context *ctx;
  PyObject *srqs;
  PyObject *mrs;
  PyObject *mws;
  PyObject *ahs;
  PyObject *qps;
  PyObject *parent_domains;
  PyObject *wqs;
  PyObject *mkeys;
  PyObject *deks;
  PyObject *_is_imported;
};


/* "pyverbs/pd.pxd":28
 *     cdef object _is_imported
 * 
 * cdef class ParentDomainInitAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_parent_domain_init_attr init_attr
 *     cdef object pd
*/
struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_parent_domain_init_attr init_attr;
  PyObject *pd;
  PyObject *alloc;
  PyObject *dealloc;
};


/* "pyverbs/pd.pxd":34
 *     cdef object dealloc
 * 
 * cdef class ParentDomain(PD):             # <<<<<<<<<<<<<<
 *     cdef add_ref(self, obj)
 *     cdef object protection_domain
*/
struct __pyx_obj_7pyverbs_2pd_ParentDomain {
  struct __pyx_obj_7pyverbs_2pd_PD __pyx_base;
  PyObject *protection_domain;
  PyObject *cqs;
};


/* "pyverbs/pd.pxd":39
 *     cdef object cqs
 * 
 * cdef class ParentDomainContext(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef object p_alloc
 *     cdef object p_free
*/
struct __pyx_obj_7pyverbs_2pd_ParentDomainContext {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  PyObject *p_alloc;
  PyObject *p_free;
  PyObject *pd;
  PyObject *user_data;
};


/* "pyverbs/mr.pxd":12
 * 
 * 
 * cdef class MR(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef object pd
 *     cdef object cmid
*/
struct __pyx_obj_7pyverbs_2mr_MR {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *pd;
  PyObject *cmid;
  struct ibv_mr *mr;
  int mmap_length;
  PyObject *is_huge;
  PyObject *is_user_addr;
  void *buf;
  PyObject *_is_imported;
};


/* "pyverbs/mr.pxd":25
 *     cpdef read(self, length, offset)
 * 
 * cdef class MREx(MR):             # <<<<<<<<<<<<<<
 *     cdef object dmah
 * 
*/
struct __pyx_obj_7pyverbs_2mr_MREx {
  struct __pyx_obj_7pyverbs_2mr_MR __pyx_base;
  PyObject *dmah;
};


/* "pyverbs/mr.pxd":28
 *     cdef object dmah
 * 
 * cdef class MWBindInfo(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_mw_bind_info info
 *     cdef object mr
*/
struct __pyx_obj_7pyverbs_2mr_MWBindInfo {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_mw_bind_info info;
  PyObject *mr;
};


/* "pyverbs/mr.pxd":32
 *     cdef object mr
 * 
 * cdef class MWBind(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_mw_bind mw_bind
 *     cdef object mr
*/
struct __pyx_obj_7pyverbs_2mr_MWBind {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_mw_bind mw_bind;
  PyObject *mr;
};


/* "pyverbs/mr.pxd":36
 *     cdef object mr
 * 
 * cdef class MW(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef object pd
 *     cdef v.ibv_mw *mw
*/
struct __pyx_obj_7pyverbs_2mr_MW {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *pd;
  struct ibv_mw *mw;
};


/* "pyverbs/mr.pxd":40
 *     cdef v.ibv_mw *mw
 * 
 * cdef class DMMR(MR):             # <<<<<<<<<<<<<<
 *     cdef object dm
 * 
*/
struct __pyx_obj_7pyverbs_2mr_DMMR {
  struct __pyx_obj_7pyverbs_2mr_MR __pyx_base;
  PyObject *dm;
};


/* "pyverbs/mr.pxd":43
 *     cdef object dm
 * 
 * cdef class DmaBufMR(MR):             # <<<<<<<<<<<<<<
 *     cdef object dmabuf
 *     cdef unsigned long offset
*/
struct __pyx_obj_7pyverbs_2mr_DmaBufMR {
  struct __pyx_obj_7pyverbs_2mr_MR __pyx_base;
  PyObject *dmabuf;
  unsigned long offset;
  PyObject *is_dmabuf_internal;
};


/* "pyverbs/mr.pxd":48
 *     cdef object is_dmabuf_internal
 * 
 * cdef class DmaHandleInitAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_dmah_init_attr init_attr
 * 
*/
struct __pyx_obj_7pyverbs_2mr_DmaHandleInitAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_dmah_init_attr init_attr;
};


/* "pyverbs/mr.pxd":51
 *     cdef v.ibv_dmah_init_attr init_attr
 * 
 * cdef class DMAHandle(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_dmah *dmah
 *     cdef object mrs
*/
struct __pyx_obj_7pyverbs_2mr_DMAHandle {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_dmah *dmah;
  PyObject *mrs;
  PyObject *ctx;
};


/* "pyverbs/cq.pxd":9
 * cimport pyverbs.libibverbs as v
 * 
 * cdef class CompChannel(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_comp_channel *cc
 *     cpdef close(self)
*/
struct __pyx_obj_7pyverbs_2cq_CompChannel {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_comp_channel *cc;
  PyObject *context;
  PyObject *cqs;
};


/* "pyverbs/cq.pxd":16
 *     cdef object cqs
 * 
 * cdef class CQ(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_cq *cq
 *     cpdef close(self)
*/
struct __pyx_obj_7pyverbs_2cq_CQ {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_cq *cq;
  PyObject *context;
  PyObject *qps;
  PyObject *srqs;
  PyObject *channel;
  PyObject *num_events;
  PyObject *wqs;
};


/* "pyverbs/cq.pxd":27
 *     cdef object wqs
 * 
 * cdef class CqInitAttrEx(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_cq_init_attr_ex attr
 *     cdef object channel
*/
struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_cq_init_attr_ex attr;
  PyObject *channel;
  PyObject *parent_domain;
};


/* "pyverbs/cq.pxd":32
 *     cdef object parent_domain
 * 
 * cdef class CQEX(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_cq_ex *cq
 *     cdef v.ibv_cq *ibv_cq
*/
struct __pyx_obj_7pyverbs_2cq_CQEX {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct ibv_cq_ex *cq;
  struct ibv_cq *ibv_cq;
  PyObject *context;
  PyObject *qps;
  PyObject *srqs;
  PyObject *wqs;
};


/* "pyverbs/cq.pxd":42
 *     cdef object wqs
 * 
 * cdef class WC(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_wc wc
 * 
*/
struct __pyx_obj_7pyverbs_2cq_WC {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_wc wc;
};


/* "pyverbs/cq.pxd":45
 *     cdef v.ibv_wc wc
 * 
 * cdef class PollCqAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_poll_cq_attr attr
 * 
*/
struct __pyx_obj_7pyverbs_2cq_PollCqAttr {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_poll_cq_attr attr;
};


/* "pyverbs/cq.pxd":48
 *     cdef v.ibv_poll_cq_attr attr
 * 
 * cdef class WcTmInfo(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_wc_tm_info info
*/
struct __pyx_obj_7pyverbs_2cq_WcTmInfo {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct ibv_wc_tm_info info;
};


/* "pyverbs/cmid.pxd":10
 * 
 * 
 * cdef class CMID(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef cm.rdma_cm_id *id
 *     cdef object event_channel
*/
struct __pyx_obj_7pyverbs_4cmid_CMID {
  struct __pyx_obj_7pyverbs_4base_PyverbsCM __pyx_base;
  struct rdma_cm_id *id;
  PyObject *event_channel;
  PyObject *ctx;
  PyObject *pd;
  PyObject *mrs;
};


/* "pyverbs/cmid.pxd":20
 * 
 * 
 * cdef class CMEventChannel(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef cm.rdma_event_channel *event_channel
 *     cpdef close(self)
*/
struct __pyx_obj_7pyverbs_4cmid_CMEventChannel {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct __pyx_vtabstruct_7pyverbs_4cmid_CMEventChannel *__pyx_vtab;
  struct rdma_event_channel *event_channel;
};


/* "pyverbs/cmid.pxd":25
 * 
 * 
 * cdef class CMEvent(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef cm.rdma_cm_event *event
 *     cpdef close(self)
*/
struct __pyx_obj_7pyverbs_4cmid_CMEvent {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct __pyx_vtabstruct_7pyverbs_4cmid_CMEvent *__pyx_vtab;
  struct rdma_cm_event *event;
};


/* "pyverbs/cmid.pxd":30
 * 
 * 
 * cdef class AddrInfo(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef cm.rdma_addrinfo *addr_info
 *     cpdef close(self)
*/
struct __pyx_obj_7pyverbs_4cmid_AddrInfo {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct __pyx_vtabstruct_7pyverbs_4cmid_AddrInfo *__pyx_vtab;
  struct rdma_addrinfo *addr_info;
};


/* "pyverbs/cmid.pxd":35
 * 
 * 
 * cdef class ConnParam(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef cm.rdma_conn_param conn_param
 *     cdef object data
*/
struct __pyx_obj_7pyverbs_4cmid_ConnParam {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct rdma_conn_param conn_param;
  PyObject *data;
};


/* "pyverbs/cmid.pxd":40
 * 
 * 
 * cdef class UDParam(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef cm.rdma_ud_param ud_param
 * 
*/
struct __pyx_obj_7pyverbs_4cmid_UDParam {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct rdma_ud_param ud_param;
};


/* "pyverbs/cmid.pxd":43
 *     cdef cm.rdma_ud_param ud_param
 * 
 * cdef class JoinMCAttrEx(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef cm.rdma_cm_join_mc_attr_ex join_mc_attr_ex
*/
struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx {
  struct __pyx_obj_7pyverbs_4base_PyverbsObject __pyx_base;
  struct rdma_cm_join_mc_attr_ex join_mc_attr_ex;
};



/* "pyverbs/base.pxd":10
 *     cdef object logger
 * 
 * cdef class PyverbsCM(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cpdef close(self)
 * 
*/

struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM {
  PyObject *(*close)(struct __pyx_obj_7pyverbs_4base_PyverbsCM *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM *__pyx_vtabptr_7pyverbs_4base_PyverbsCM;


/* "pyverbs/qp.pxd":30
 *     cdef v.ibv_qp_attr attr
 * 
 * cdef class QP(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_qp *qp
 *     cdef int type
*/

struct __pyx_vtabstruct_7pyverbs_2qp_QP {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*update_cqs)(struct __pyx_obj_7pyverbs_2qp_QP *, PyObject *);
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_2qp_QP *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_2qp_QP *__pyx_vtabptr_7pyverbs_2qp_QP;


/* "pyverbs/qp.pxd":47
 *     cdef add_ref(self, obj)
 * 
 * cdef class DataBuffer(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_data_buf data
 * 
*/

struct __pyx_vtabstruct_7pyverbs_2qp_DataBuffer {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2qp_DataBuffer *__pyx_vtabptr_7pyverbs_2qp_DataBuffer;


/* "pyverbs/qp.pxd":50
 *     cdef v.ibv_data_buf data
 * 
 * cdef class QPEx(QP):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_qp_ex *qp_ex
 *     cdef object ind_table
*/

struct __pyx_vtabstruct_7pyverbs_2qp_QPEx {
  struct __pyx_vtabstruct_7pyverbs_2qp_QP __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2qp_QPEx *__pyx_vtabptr_7pyverbs_2qp_QPEx;


/* "pyverbs/qp.pxd":54
 *     cdef object ind_table
 * 
 * cdef class ECE(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_ece ece
*/

struct __pyx_vtabstruct_7pyverbs_2qp_ECE {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2qp_ECE *__pyx_vtabptr_7pyverbs_2qp_ECE;


/* "pyverbs/addr.pxd":20
 *     cdef v.ibv_global_route gr
 * 
 * cdef class AHAttr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_ah_attr ah_attr
 *     cdef init_from_ud_param(self, UDParam udparam)
*/

struct __pyx_vtabstruct_7pyverbs_4addr_AHAttr {
  PyObject *(*init_from_ud_param)(struct __pyx_obj_7pyverbs_4addr_AHAttr *, struct __pyx_obj_7pyverbs_4cmid_UDParam *);
};
static struct __pyx_vtabstruct_7pyverbs_4addr_AHAttr *__pyx_vtabptr_7pyverbs_4addr_AHAttr;


/* "pyverbs/addr.pxd":24
 *     cdef init_from_ud_param(self, UDParam udparam)
 * 
 * cdef class AH(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_ah *ah
 *     cdef object pd
*/

struct __pyx_vtabstruct_7pyverbs_4addr_AH {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_4addr_AH *__pyx_vtabptr_7pyverbs_4addr_AH;


/* "pyverbs/device.pxd":10
 * 
 * 
 * cdef class Context(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_context *context
 *     cdef v.ibv_device *device
*/

struct __pyx_vtabstruct_7pyverbs_6device_Context {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_6device_Context *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_6device_Context *__pyx_vtabptr_7pyverbs_6device_Context;


/* "pyverbs/device.pxd":67
 *     cdef v.ibv_alloc_dm_attr alloc_dm_attr
 * 
 * cdef class DM(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_dm *dm
 *     cdef object dm_mrs
*/

struct __pyx_vtabstruct_7pyverbs_6device_DM {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_6device_DM *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_6device_DM *__pyx_vtabptr_7pyverbs_6device_DM;


/* "pyverbs/device.pxd":83
 *     cdef v.ibv_async_event event
 * 
 * cdef class FdArr(PyverbsObject):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_fd_arr attr
 *     cpdef close(self)
*/

struct __pyx_vtabstruct_7pyverbs_6device_FdArr {
  PyObject *(*close)(struct __pyx_obj_7pyverbs_6device_FdArr *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyverbs_6device_FdArr *__pyx_vtabptr_7pyverbs_6device_FdArr;


/* "pyverbs/pd.pxd":12
 * 
 * 
 * cdef class PD(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_pd *pd
 *     cdef Context ctx
*/

struct __pyx_vtabstruct_7pyverbs_2pd_PD {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_2pd_PD *, PyObject *);
  PyObject *(*remove_ref)(struct __pyx_obj_7pyverbs_2pd_PD *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_2pd_PD *__pyx_vtabptr_7pyverbs_2pd_PD;


/* "pyverbs/pd.pxd":34
 *     cdef object dealloc
 * 
 * cdef class ParentDomain(PD):             # <<<<<<<<<<<<<<
 *     cdef add_ref(self, obj)
 *     cdef object protection_domain
*/

struct __pyx_vtabstruct_7pyverbs_2pd_ParentDomain {
  struct __pyx_vtabstruct_7pyverbs_2pd_PD __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2pd_ParentDomain *__pyx_vtabptr_7pyverbs_2pd_ParentDomain;


/* "pyverbs/mr.pxd":12
 * 
 * 
 * cdef class MR(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef object pd
 *     cdef object cmid
*/

struct __pyx_vtabstruct_7pyverbs_2mr_MR {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  void (*_allocate_buffer)(struct __pyx_obj_7pyverbs_2mr_MR *, size_t, int, int *);
  void (*_free_buffer)(struct __pyx_obj_7pyverbs_2mr_MR *, int, int);
  PyObject *(*read)(struct __pyx_obj_7pyverbs_2mr_MR *, PyObject *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyverbs_2mr_MR *__pyx_vtabptr_7pyverbs_2mr_MR;


/* "pyverbs/mr.pxd":25
 *     cpdef read(self, length, offset)
 * 
 * cdef class MREx(MR):             # <<<<<<<<<<<<<<
 *     cdef object dmah
 * 
*/

struct __pyx_vtabstruct_7pyverbs_2mr_MREx {
  struct __pyx_vtabstruct_7pyverbs_2mr_MR __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2mr_MREx *__pyx_vtabptr_7pyverbs_2mr_MREx;


/* "pyverbs/mr.pxd":28
 *     cdef object dmah
 * 
 * cdef class MWBindInfo(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_mw_bind_info info
 *     cdef object mr
*/

struct __pyx_vtabstruct_7pyverbs_2mr_MWBindInfo {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2mr_MWBindInfo *__pyx_vtabptr_7pyverbs_2mr_MWBindInfo;


/* "pyverbs/mr.pxd":32
 *     cdef object mr
 * 
 * cdef class MWBind(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_mw_bind mw_bind
 *     cdef object mr
*/

struct __pyx_vtabstruct_7pyverbs_2mr_MWBind {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2mr_MWBind *__pyx_vtabptr_7pyverbs_2mr_MWBind;


/* "pyverbs/mr.pxd":36
 *     cdef object mr
 * 
 * cdef class MW(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef object pd
 *     cdef v.ibv_mw *mw
*/

struct __pyx_vtabstruct_7pyverbs_2mr_MW {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2mr_MW *__pyx_vtabptr_7pyverbs_2mr_MW;


/* "pyverbs/mr.pxd":40
 *     cdef v.ibv_mw *mw
 * 
 * cdef class DMMR(MR):             # <<<<<<<<<<<<<<
 *     cdef object dm
 * 
*/

struct __pyx_vtabstruct_7pyverbs_2mr_DMMR {
  struct __pyx_vtabstruct_7pyverbs_2mr_MR __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2mr_DMMR *__pyx_vtabptr_7pyverbs_2mr_DMMR;


/* "pyverbs/mr.pxd":43
 *     cdef object dm
 * 
 * cdef class DmaBufMR(MR):             # <<<<<<<<<<<<<<
 *     cdef object dmabuf
 *     cdef unsigned long offset
*/

struct __pyx_vtabstruct_7pyverbs_2mr_DmaBufMR {
  struct __pyx_vtabstruct_7pyverbs_2mr_MR __pyx_base;
};
static struct __pyx_vtabstruct_7pyverbs_2mr_DmaBufMR *__pyx_vtabptr_7pyverbs_2mr_DmaBufMR;


/* "pyverbs/mr.pxd":51
 *     cdef v.ibv_dmah_init_attr init_attr
 * 
 * cdef class DMAHandle(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_dmah *dmah
 *     cdef object mrs
*/

struct __pyx_vtabstruct_7pyverbs_2mr_DMAHandle {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_2mr_DMAHandle *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_2mr_DMAHandle *__pyx_vtabptr_7pyverbs_2mr_DMAHandle;


/* "pyverbs/cq.pxd":9
 * cimport pyverbs.libibverbs as v
 * 
 * cdef class CompChannel(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_comp_channel *cc
 *     cpdef close(self)
*/

struct __pyx_vtabstruct_7pyverbs_2cq_CompChannel {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_2cq_CompChannel *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_2cq_CompChannel *__pyx_vtabptr_7pyverbs_2cq_CompChannel;


/* "pyverbs/cq.pxd":16
 *     cdef object cqs
 * 
 * cdef class CQ(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_cq *cq
 *     cpdef close(self)
*/

struct __pyx_vtabstruct_7pyverbs_2cq_CQ {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_2cq_CQ *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_2cq_CQ *__pyx_vtabptr_7pyverbs_2cq_CQ;


/* "pyverbs/cq.pxd":32
 *     cdef object parent_domain
 * 
 * cdef class CQEX(PyverbsCM):             # <<<<<<<<<<<<<<
 *     cdef v.ibv_cq_ex *cq
 *     cdef v.ibv_cq *ibv_cq
*/

struct __pyx_vtabstruct_7pyverbs_2cq_CQEX {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_2cq_CQEX *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_2cq_CQEX *__pyx_vtabptr_7pyverbs_2cq_CQEX;


/* "pyverbs/cmid.pyx":347
 * 
 * 
 * cdef class CMID(PyverbsCM):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, object creator=None, QPInitAttr qp_init_attr=None,
*/

struct __pyx_vtabstruct_7pyverbs_4cmid_CMID {
  struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM __pyx_base;
  PyObject *(*add_ref)(struct __pyx_obj_7pyverbs_4cmid_CMID *, PyObject *);
};
static struct __pyx_vtabstruct_7pyverbs_4cmid_CMID *__pyx_vtabptr_7pyverbs_4cmid_CMID;


/* "pyverbs/cmid.pyx":322
 * 
 * 
 * cdef class CMEventChannel(PyverbsObject):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self):
*/

struct __pyx_vtabstruct_7pyverbs_4cmid_CMEventChannel {
  PyObject *(*close)(struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyverbs_4cmid_CMEventChannel *__pyx_vtabptr_7pyverbs_4cmid_CMEventChannel;


/* "pyverbs/cmid.pyx":268
 * 
 * 
 * cdef class CMEvent(PyverbsObject):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, CMEventChannel channel):
*/

struct __pyx_vtabstruct_7pyverbs_4cmid_CMEvent {
  PyObject *(*close)(struct __pyx_obj_7pyverbs_4cmid_CMEvent *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyverbs_4cmid_CMEvent *__pyx_vtabptr_7pyverbs_4cmid_CMEvent;


/* "pyverbs/cmid.pyx":167
 * 
 * 
 * cdef class AddrInfo(PyverbsObject):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, src=None, dst=None, src_service=None, dst_service=None,
*/

struct __pyx_vtabstruct_7pyverbs_4cmid_AddrInfo {
  PyObject *(*close)(struct __pyx_obj_7pyverbs_4cmid_AddrInfo *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyverbs_4cmid_AddrInfo *__pyx_vtabptr_7pyverbs_4cmid_AddrInfo;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto (used by PyErrFetchRestore) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto (used by GetBuiltinName) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* HasAttr.proto (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_HasAttr(o, n)  PyObject_HasAttrWithError(o, n)
#else
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);
#endif

/* PyObjectCall.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto (used by PyObjectCallOneArg) */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs);

/* PyObjectCallOneArg.proto (used by ImportImpl) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* TupleAndListFromArray.proto (used by ImportImpl) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
#endif
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto (used by ImportImpl) */
#include <string.h>

/* ImportImpl.export */
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level);

/* Import.proto */
static CYTHON_INLINE PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* GetTopmostException.proto (used by SaveResetException) */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyImportError_Check.proto */
#define __Pyx_PyExc_ImportError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_ImportError)

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectFastCallMethod.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyObject_FastCallMethod(name, args, nargsf) PyObject_VectorcallMethod(name, args, nargsf, NULL)
#else
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf);
#endif

/* BytesEquals.proto (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_PySequence_ITEM(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_NewRef(__Pyx_PyTuple_GET_ITEM(args, i))
#else
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_XNewRef(PyTuple_GetItem(args, i))
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_ArgRef_FASTCALL(args, i) __Pyx_NewRef(args[i])
    #define __Pyx_NumKwargs_FASTCALL(kwds) __Pyx_PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
#else
    #define __Pyx_ArgRef_FASTCALL __Pyx_ArgRef_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
#endif
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#if CYTHON_METH_FASTCALL || (CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(args + start, stop - start)
#else
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* py_dict_items.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d);

/* CallCFunction.proto (used by CallUnboundCMethod0) */
#define __Pyx_CallCFunction(cfunc, self, args)\
    ((PyCFunction)(void(*)(void))(cfunc)->func)(self, args)
#define __Pyx_CallCFunctionWithKeywords(cfunc, self, args, kwargs)\
    ((PyCFunctionWithKeywords)(void(*)(void))(cfunc)->func)(self, args, kwargs)
#define __Pyx_CallCFunctionFast(cfunc, self, args, nargs)\
    ((__Pyx_PyCFunctionFast)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs)
#define __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, nargs, kwnames)\
    ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs, kwnames)

/* UnpackUnboundCMethod.proto (used by CallUnboundCMethod0) */
typedef struct {
    PyObject *type;
    PyObject **method_name;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && CYTHON_ATOMICS
    __pyx_atomic_int_type initialized;
#endif
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
static CYTHON_INLINE int __Pyx_CachedCFunction_GetAndSetInitializing(__Pyx_CachedCFunction *cfunc) {
#if !CYTHON_ATOMICS
    return 1;
#else
    __pyx_nonatomic_int_type expected = 0;
    if (__pyx_atomic_int_cmp_exchange(&cfunc->initialized, &expected, 1)) {
        return 0;
    }
    return expected;
#endif
}
static CYTHON_INLINE void __Pyx_CachedCFunction_SetFinishedInitializing(__Pyx_CachedCFunction *cfunc) {
#if CYTHON_ATOMICS
    __pyx_atomic_store(&cfunc->initialized, 2);
#endif
}
#else
#define __Pyx_CachedCFunction_GetAndSetInitializing(cfunc) 2
#define __Pyx_CachedCFunction_SetFinishedInitializing(cfunc)
#endif

/* CallUnboundCMethod0.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* py_dict_values.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d);

/* OwnedDictNext.proto (used by ParseKeywordsImpl) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue);
#else
CYTHON_INLINE
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue);
#endif

/* RaiseDoubleKeywords.proto (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywordsImpl.export */
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name
);
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* CallUnboundCMethod2.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* ParseKeywords.proto */
static CYTHON_INLINE int __Pyx_ParseKeywords(
    PyObject *kwds, PyObject *const *kwvalues, PyObject ** const argnames[],
    PyObject *kwds2, PyObject *values[],
    Py_ssize_t num_pos_args, Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* PyDictVersioning.proto (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __Pyx_XNewRef(__pyx_dict_cached_value);\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_mstate_global->__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* SetStringIndexingError.proto (used by SetItemIntByteArray) */
static void __Pyx_SetStringIndexingError(const char* message, int has_gil);

/* SetItemIntByteArray.proto */
#define __Pyx_SetItemInt_ByteArray(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_ByteArray_Fast(o, (Py_ssize_t)i, v, wraparound, boundscheck, has_gil, unsafe_shared) :\
    (__Pyx_SetStringIndexingError("bytearray index out of range", has_gil), -1))
static CYTHON_INLINE int __Pyx_SetItemInt_ByteArray_Fast(PyObject* string, Py_ssize_t i, unsigned char v,
                                                         int wraparound, int boundscheck, int has_gil, int unsafe_shared);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck, unsafe_shared) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck, int unsafe_shared);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck, unsafe_shared) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck, int unsafe_shared);

/* UnicodeConcatInPlace.proto */
# if CYTHON_COMPILING_IN_CPYTHON
    #if CYTHON_REFNANNY
        #define __Pyx_PyUnicode_ConcatInPlace(left, right, unsafe_shared) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right, unsafe_shared, __pyx_refnanny)
    #else
        #define __Pyx_PyUnicode_ConcatInPlace(left, right, unsafe_shared) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right, unsafe_shared)
    #endif
    #define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_DefinitelyUniqueInPlace(left, right) __Pyx_PyUnicode_ConcatInPlace(left, right, __Pyx_ReferenceSharing_DefinitelyUnique)
    #define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(left, right) __Pyx_PyUnicode_ConcatInPlace(left, right, __Pyx_ReferenceSharing_OwnStrongReference)
    #define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_FunctionArgumentInPlace(left, right) __Pyx_PyUnicode_ConcatInPlace(left, right, __Pyx_ReferenceSharing_DefinitelyUnique)
    #define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_SharedReferenceInPlace(left, right) __Pyx_PyUnicode_ConcatInPlace(left, right, __Pyx_ReferenceSharing_SharedReference)
    static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right, int unsafe_shared
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    );
#else
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_DefinitelyUniqueInPlace __Pyx_PyUnicode_Concat
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace __Pyx_PyUnicode_Concat
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_FunctionArgumentInPlace __Pyx_PyUnicode_Concat
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_SharedReferenceInPlace __Pyx_PyUnicode_Concat
#endif
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_DefinitelyUniqueInPlaceSafe(left, right)\
    ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_DefinitelyUniqueInPlace(left, right))
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlaceSafe(left, right)\
    ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(left, right))
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_FunctionArgumentInPlaceSafe(left, right)\
    ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_FunctionArgumentInPlace(left, right))
#define __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_SharedReferenceInPlaceSafe(left, right)\
    ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_SharedReferenceInPlace(left, right))

/* RejectKeywords.export */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds);

/* PyTypeError_Check.proto */
#define __Pyx_PyExc_TypeError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_TypeError)

/* RaiseException.export */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* ArgTypeTestFunc.export */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* decode_c_string_utf16.proto (used by decode_c_bytes) */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_bytes.proto (used by decode_bytes) */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* decode_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_bytes(
         PyObject* string, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    char* as_c_string;
    Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
    as_c_string = PyBytes_AS_STRING(string);
    size = PyBytes_GET_SIZE(string);
#else
    if (PyBytes_AsStringAndSize(string, &as_c_string, &size) < 0) {
        return NULL;
    }
#endif
    return __Pyx_decode_c_bytes(
        as_c_string, size,
        start, stop, encoding, errors, decode_func);
}

/* PyObjectVectorCallKwBuilder.proto */
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#if CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_Vectorcall_CallFromBuilder PyObject_Vectorcall
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder _PyObject_Vectorcall
#endif
#define __Pyx_MakeVectorcallBuilderKwds(n) PyTuple_New(n)
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder __Pyx_PyObject_FastCallDict
#define __Pyx_MakeVectorcallBuilderKwds(n) __Pyx_PyDict_NewPresized(n)
#define __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n) PyDict_SetItem(builder, key, value)
#define __Pyx_VectorcallBuilder_AddArgStr(key, value, builder, args, n) PyDict_SetItemString(builder, key, value)
#endif

/* PyObjectVectorCallMethodKwBuilder.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_VectorcallMethod_CallFromBuilder PyObject_VectorcallMethod
#else
static PyObject *__Pyx_Object_VectorcallMethod_CallFromBuilder(PyObject *name, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#endif

/* decode_c_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* PyObjectDelAttr.proto (used by PyObjectSetAttrStr) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
#define __Pyx_PyObject_DelAttr(o, n) PyObject_SetAttr(o, n, NULL)
#else
#define __Pyx_PyObject_DelAttr(o, n) PyObject_DelAttr(o, n)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   __Pyx_PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* CallNextTpDealloc.proto */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallTypeTraverse.proto */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#define __Pyx_call_type_traverse(o, always_call, visit, arg) 0
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg);
#endif

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_2_4
#define __PYX_HAVE_RT_ImportType_proto_3_2_4
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_2_4 {
   __Pyx_ImportType_CheckSize_Error_3_2_4 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_2_4 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_2_4 = 2
};
static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* LimitedApiGetTypeDict.proto (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp);
#endif

/* SetItemOnTypeDict.proto (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v);
#define __Pyx_SetItemOnTypeDict(tp, k, v) __Pyx__SetItemOnTypeDict((PyTypeObject*)tp, k, v)

/* FixUpExtensionType.proto */
static CYTHON_INLINE int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);

/* PyObjectCallNoArg.proto (used by PyObjectCallMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectGetMethod.proto (used by PyObjectCallMethod0) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);
#endif

/* PyObjectCallMethod0.proto (used by PyType_Ready) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* ValidateBasesTuple.proto (used by PyType_Ready) */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* MergeVTables.proto */
static int __Pyx_MergeVtables(PyTypeObject *type);

/* DelItemOnTypeDict.proto (used by SetupReduce) */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k);
#define __Pyx_DelItemOnTypeDict(tp, k) __Pyx__DelItemOnTypeDict((PyTypeObject*)tp, k)

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* FunctionImport.proto */
static int __Pyx_ImportFunction_3_2_4(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* dict_setdefault.proto (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value);

/* AddModuleRef.proto (used by FetchSharedCythonModule) */
#if ((CYTHON_COMPILING_IN_CPYTHON_FREETHREADING ) ||\
     __PYX_LIMITED_VERSION_HEX < 0x030d0000)
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name);
#else
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#endif

/* FetchSharedCythonModule.proto (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto (used by CommonTypesMetaclass) */
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases);

/* CommonTypesMetaclass.proto (used by CythonFunctionShared) */
static int __pyx_CommonTypesMetaclass_init(PyObject *module);
#define __Pyx_CommonTypesMetaclass_USED

/* PyMethodNew.proto (used by CythonFunctionShared) */
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ);

/* PyVectorcallFastCallDict.proto (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto (used by CythonFunction) */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && CYTHON_METH_FASTCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_dict;
#endif
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    PyObject *defaults;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_mstate_global->__pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_mstate_global->__pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_mstate_global->__pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void));
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func,
                                                         PyTypeObject *defaults_type);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* CLineInTraceback.proto (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#else
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#endif

/* CodeObjectCache.proto (used by AddTraceback) */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject __Pyx_CachedCodeObjectType;
#else
typedef PyCodeObject __Pyx_CachedCodeObjectType;
#endif
typedef struct {
    __Pyx_CachedCodeObjectType* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_int_type accessor_count;
  #endif
};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE uint8_t __Pyx_PyLong_As_uint8_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint32_t __Pyx_PyLong_As_uint32_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_uint32_t(uint32_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_uint8_t(uint8_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyLong_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From___pyx_anon_enum(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE enum rdma_cm_event_type __Pyx_PyLong_As_enum__rdma_cm_event_type(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE enum rdma_port_space __Pyx_PyLong_As_enum__rdma_port_space(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_int(unsigned int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE uint64_t __Pyx_PyLong_As_uint64_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_long(unsigned long value);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#else
static __Pyx_TypeName __Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp);
#endif
#else  // !LIMITED_API
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetFullyQualifiedName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2) {
    return PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2);
}
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#ifdef PyExceptionInstance_Check
  #define __Pyx_PyBaseException_Check(obj) PyExceptionInstance_Check(obj)
#else
  #define __Pyx_PyBaseException_Check(obj) __Pyx_TypeCheck(obj, PyExc_BaseException)
#endif

/* GetRuntimeVersion.proto */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
static unsigned long __Pyx_cached_runtime_version = 0;
static void __Pyx_init_runtime_version(void);
#else
#define __Pyx_init_runtime_version()
#endif
static unsigned long __Pyx_get_runtime_version(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* DecompressString.proto */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo);

/* MultiPhaseInitModuleState.proto */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
static PyObject *__Pyx_State_FindModule(void*);
static int __Pyx_State_AddModule(PyObject* module, void*);
static int __Pyx_State_RemoveModule(void*);
#elif CYTHON_USE_MODULE_STATE
#define __Pyx_State_FindModule PyState_FindModule
#define __Pyx_State_AddModule PyState_AddModule
#define __Pyx_State_RemoveModule PyState_RemoveModule
#endif

/* #### Code section: module_declarations ### */
/* CythonABIVersion.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
    #if CYTHON_METH_FASTCALL
        #define __PYX_FASTCALL_ABI_SUFFIX  "_fastcall"
    #else
        #define __PYX_FASTCALL_ABI_SUFFIX
    #endif
    #define __PYX_LIMITED_ABI_SUFFIX "limited" __PYX_FASTCALL_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#else
    #define __PYX_LIMITED_ABI_SUFFIX
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    #define __PYX_AM_SEND_ABI_SUFFIX
#elif __PYX_HAS_PY_AM_SEND == 2
    #define __PYX_AM_SEND_ABI_SUFFIX "amsendbackport"
#else
    #define __PYX_AM_SEND_ABI_SUFFIX "noamsend"
#endif
#ifndef __PYX_MONITORING_ABI_SUFFIX
    #define __PYX_MONITORING_ABI_SUFFIX
#endif
#if CYTHON_USE_TP_FINALIZE
    #define __PYX_TP_FINALIZE_ABI_SUFFIX
#else
    #define __PYX_TP_FINALIZE_ABI_SUFFIX "nofinalize"
#endif
#if CYTHON_USE_FREELISTS || !defined(__Pyx_AsyncGen_USED)
    #define __PYX_FREELISTS_ABI_SUFFIX
#else
    #define __PYX_FREELISTS_ABI_SUFFIX "nofreelists"
#endif
#define CYTHON_ABI  __PYX_ABI_VERSION __PYX_LIMITED_ABI_SUFFIX __PYX_MONITORING_ABI_SUFFIX __PYX_TP_FINALIZE_ABI_SUFFIX __PYX_FREELISTS_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."

static PyObject *__pyx_f_7pyverbs_4cmid_8AddrInfo_close(struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_7pyverbs_4cmid_7CMEvent_close(struct __pyx_obj_7pyverbs_4cmid_CMEvent *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_7pyverbs_4cmid_14CMEventChannel_close(struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_7pyverbs_4cmid_4CMID_add_ref(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_obj); /* proto*/
static PyObject *__pyx_f_7pyverbs_4cmid_4CMID_close(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/

/* Module declarations from "pyverbs.base" */
static PyObject *(*__pyx_f_7pyverbs_4base_close_weakrefs)(PyObject *); /*proto*/

/* Module declarations from "libc.stdint" */

/* Module declarations from "pyverbs.librdmacm_enums" */

/* Module declarations from "posix.types" */

/* Module declarations from "posix.signal" */

/* Module declarations from "posix.time" */

/* Module declarations from "pyverbs.libibverbs_enums" */

/* Module declarations from "pyverbs.libibverbs" */

/* Module declarations from "pyverbs.librdmacm" */

/* Module declarations from "libc.string" */

/* Module declarations from "pyverbs.qp" */

/* Module declarations from "pyverbs" */

/* Module declarations from "pyverbs.addr" */

/* Module declarations from "pyverbs.device" */

/* Module declarations from "pyverbs.pd" */

/* Module declarations from "pyverbs.mr" */

/* Module declarations from "pyverbs.cq" */

/* Module declarations from "pyverbs.cmid" */
static PyObject *__pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py = 0;
static PyObject *__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py = 0;
static PyObject *__pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py = 0;
static PyObject *__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py = 0;
static PyObject *__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py = 0;
static PyObject *__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py = 0;
static PyObject *__Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py(enum rdma_port_space); /*proto*/
static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py(enum ibv_send_flags); /*proto*/
static PyObject *__Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py(enum rdma_cm_event_type); /*proto*/
static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py(enum ibv_access_flags); /*proto*/
static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(enum ibv_wc_status); /*proto*/
static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py(enum ibv_wc_opcode); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "pyverbs.cmid"
extern int __pyx_module_is_main_pyverbs__cmid;
int __pyx_module_is_main_pyverbs__cmid = 0;

/* Implementation of "pyverbs.cmid" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_super;
/* #### Code section: string_decls ### */
/* #### Code section: decls ### */
static int __pyx_pf_7pyverbs_4cmid_9ConnParam___init__(struct __pyx_obj_7pyverbs_4cmid_ConnParam *__pyx_v_self, PyObject *__pyx_v_resources, PyObject *__pyx_v_depth, PyObject *__pyx_v_flow_control, PyObject *__pyx_v_retry, PyObject *__pyx_v_rnr_retry, PyObject *__pyx_v_srq, PyObject *__pyx_v_qp_num, PyObject *__pyx_v_data_len); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_9ConnParam_3qpn___get__(struct __pyx_obj_7pyverbs_4cmid_ConnParam *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_4cmid_9ConnParam_3qpn_2__set__(struct __pyx_obj_7pyverbs_4cmid_ConnParam *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_9ConnParam_12private_data___get__(struct __pyx_obj_7pyverbs_4cmid_ConnParam *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_9ConnParam_2set_private_data(struct __pyx_obj_7pyverbs_4cmid_ConnParam *__pyx_v_self, PyObject *__pyx_v_data); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_9ConnParam_4__str__(struct __pyx_obj_7pyverbs_4cmid_ConnParam *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_9ConnParam_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_ConnParam *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_9ConnParam_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_ConnParam *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7pyverbs_4cmid_12JoinMCAttrEx___init__(struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *__pyx_v_self, struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_addr, PyObject *__pyx_v_comp_mask, PyObject *__pyx_v_join_flags); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_12JoinMCAttrEx_10join_flags___get__(struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_4cmid_12JoinMCAttrEx_10join_flags_2__set__(struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_12JoinMCAttrEx_9comp_mask___get__(struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *__pyx_v_self); /* proto */
static int __pyx_pf_7pyverbs_4cmid_12JoinMCAttrEx_9comp_mask_2__set__(struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *__pyx_v_self, PyObject *__pyx_v_val); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_12JoinMCAttrEx_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_12JoinMCAttrEx_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7pyverbs_4cmid_7UDParam___init__(struct __pyx_obj_7pyverbs_4cmid_UDParam *__pyx_v_self, struct __pyx_obj_7pyverbs_4cmid_CMEvent *__pyx_v_cm_event); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_7UDParam_6qp_num___get__(struct __pyx_obj_7pyverbs_4cmid_UDParam *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_7UDParam_4qkey___get__(struct __pyx_obj_7pyverbs_4cmid_UDParam *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_7UDParam_7ah_attr___get__(struct __pyx_obj_7pyverbs_4cmid_UDParam *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_7UDParam_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_UDParam *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_7UDParam_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_UDParam *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7pyverbs_4cmid_8AddrInfo___init__(struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_self, PyObject *__pyx_v_src, PyObject *__pyx_v_dst, PyObject *__pyx_v_src_service, PyObject *__pyx_v_dst_service, PyObject *__pyx_v_port_space, PyObject *__pyx_v_flags); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_8AddrInfo_2get_addrinfo(struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_self, PyObject *__pyx_v_src, PyObject *__pyx_v_dst, PyObject *__pyx_v_src_service, PyObject *__pyx_v_dst_service, PyObject *__pyx_v_port_space, PyObject *__pyx_v_flags); /* proto */
static void __pyx_pf_7pyverbs_4cmid_8AddrInfo_4__dealloc__(struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_8AddrInfo_6close(struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_8AddrInfo_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_8AddrInfo_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7pyverbs_4cmid_7CMEvent___init__(struct __pyx_obj_7pyverbs_4cmid_CMEvent *__pyx_v_self, struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *__pyx_v_channel); /* proto */
static void __pyx_pf_7pyverbs_4cmid_7CMEvent_2__dealloc__(struct __pyx_obj_7pyverbs_4cmid_CMEvent *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_7CMEvent_4close(struct __pyx_obj_7pyverbs_4cmid_CMEvent *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_7CMEvent_10event_type___get__(struct __pyx_obj_7pyverbs_4cmid_CMEvent *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_7CMEvent_6ack_cm_event(struct __pyx_obj_7pyverbs_4cmid_CMEvent *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_7CMEvent_8event_str(struct __pyx_obj_7pyverbs_4cmid_CMEvent *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_7CMEvent_12private_data___get__(struct __pyx_obj_7pyverbs_4cmid_CMEvent *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_7CMEvent_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_CMEvent *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_7CMEvent_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_CMEvent *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7pyverbs_4cmid_14CMEventChannel___init__(struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *__pyx_v_self); /* proto */
static void __pyx_pf_7pyverbs_4cmid_14CMEventChannel_2__dealloc__(struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_14CMEventChannel_4close(struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_14CMEventChannel_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_14CMEventChannel_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7pyverbs_4cmid_4CMID___init__(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_creator, struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_qp_init_attr, struct __pyx_obj_7pyverbs_2pd_PD *__pyx_v_pd, PyObject *__pyx_v_port_space, struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_listen_id); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_8dev_name___get__(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_8port_num___get__(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_13event_channel___get__(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_7context___get__(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_2pd___get__(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_3qpn___get__(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self); /* proto */
static void __pyx_pf_7pyverbs_4cmid_4CMID_2__dealloc__(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_4close(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_6get_request(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_8bind_addr(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_lai); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_10resolve_addr(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_rai, PyObject *__pyx_v_timeout_ms); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_12join_multicast(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_addr, struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *__pyx_v_mc_attr, PyObject *__pyx_v_context); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_14leave_multicast(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_addr); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_16resolve_route(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_timeout_ms); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_18listen(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_backlog); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_20connect(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_4cmid_ConnParam *__pyx_v_param); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_22disconnect(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_24accept(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_4cmid_ConnParam *__pyx_v_param); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_26establish(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_28set_local_ece(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_ece); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_30get_remote_ece(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_32create_qp(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_qp_init); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_34query_qp(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_attr_mask); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_36init_qp_attr(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_qp_state); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_38reg_msgs(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_size); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_40reg_read(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_size); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_42reg_write(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_size); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_44post_recv(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_2mr_MR *__pyx_v_mr, PyObject *__pyx_v_length); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_46post_send(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_2mr_MR *__pyx_v_mr, PyObject *__pyx_v_flags, PyObject *__pyx_v_length); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_48post_read(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_2mr_MR *__pyx_v_mr, PyObject *__pyx_v_length, PyObject *__pyx_v_remote_addr, PyObject *__pyx_v_rkey, PyObject *__pyx_v_flags); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_50post_write(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_2mr_MR *__pyx_v_mr, PyObject *__pyx_v_length, PyObject *__pyx_v_remote_addr, PyObject *__pyx_v_rkey, PyObject *__pyx_v_flags); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_52post_ud_send(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_2mr_MR *__pyx_v_mr, struct __pyx_obj_7pyverbs_4addr_AH *__pyx_v_ah, PyObject *__pyx_v_rqpn, PyObject *__pyx_v_flags, PyObject *__pyx_v_length); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_54get_recv_comp(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_56get_send_comp(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_58set_option(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_level, PyObject *__pyx_v_optname, PyObject *__pyx_v_optval, PyObject *__pyx_v_optlen); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_60reject(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_private_data); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_62resolve_addrinfo(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_node, PyObject *__pyx_v_service, PyObject *__pyx_v_port_space, PyObject *__pyx_v_flags); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_64query_addrinfo(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_66__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_68__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_7pyverbs_4cmid_CMID(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7pyverbs_4cmid_CMEventChannel(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7pyverbs_4cmid_CMEvent(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7pyverbs_4cmid_AddrInfo(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7pyverbs_4cmid_ConnParam(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7pyverbs_4cmid_UDParam(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7pyverbs_4cmid_JoinMCAttrEx(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
/* SmallCodeConfig */
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  PyTypeObject *__pyx_ptype_7pyverbs_4base_PyverbsObject;
  PyTypeObject *__pyx_ptype_7pyverbs_4base_PyverbsCM;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_QPCap;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_QPInitAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_QPInitAttrEx;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_QPAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_QP;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_DataBuffer;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_QPEx;
  PyTypeObject *__pyx_ptype_7pyverbs_2qp_ECE;
  PyTypeObject *__pyx_ptype_7pyverbs_4addr_GID;
  PyTypeObject *__pyx_ptype_7pyverbs_4addr_GRH;
  PyTypeObject *__pyx_ptype_7pyverbs_4addr_GlobalRoute;
  PyTypeObject *__pyx_ptype_7pyverbs_4addr_AHAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_4addr_AH;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_Context;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_DeviceAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_QueryDeviceExInput;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_ODPCaps;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_RSSCaps;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_PacketPacingCaps;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_PCIAtomicCaps;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_TMCaps;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_CQModerationCaps;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_TSOCaps;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_DeviceAttrEx;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_AllocDmAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_DM;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_PortAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_GIDEntry;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_AsyncEvent;
  PyTypeObject *__pyx_ptype_7pyverbs_6device_FdArr;
  PyTypeObject *__pyx_ptype_7pyverbs_2pd_PD;
  PyTypeObject *__pyx_ptype_7pyverbs_2pd_ParentDomainInitAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_2pd_ParentDomain;
  PyTypeObject *__pyx_ptype_7pyverbs_2pd_ParentDomainContext;
  PyTypeObject *__pyx_ptype_7pyverbs_2mr_MR;
  PyTypeObject *__pyx_ptype_7pyverbs_2mr_MREx;
  PyTypeObject *__pyx_ptype_7pyverbs_2mr_MWBindInfo;
  PyTypeObject *__pyx_ptype_7pyverbs_2mr_MWBind;
  PyTypeObject *__pyx_ptype_7pyverbs_2mr_MW;
  PyTypeObject *__pyx_ptype_7pyverbs_2mr_DMMR;
  PyTypeObject *__pyx_ptype_7pyverbs_2mr_DmaBufMR;
  PyTypeObject *__pyx_ptype_7pyverbs_2mr_DmaHandleInitAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_2mr_DMAHandle;
  PyTypeObject *__pyx_ptype_7pyverbs_2cq_CompChannel;
  PyTypeObject *__pyx_ptype_7pyverbs_2cq_CQ;
  PyTypeObject *__pyx_ptype_7pyverbs_2cq_CqInitAttrEx;
  PyTypeObject *__pyx_ptype_7pyverbs_2cq_CQEX;
  PyTypeObject *__pyx_ptype_7pyverbs_2cq_WC;
  PyTypeObject *__pyx_ptype_7pyverbs_2cq_PollCqAttr;
  PyTypeObject *__pyx_ptype_7pyverbs_2cq_WcTmInfo;
  PyObject *__pyx_type_7pyverbs_4cmid_CMID;
  PyObject *__pyx_type_7pyverbs_4cmid_CMEventChannel;
  PyObject *__pyx_type_7pyverbs_4cmid_CMEvent;
  PyObject *__pyx_type_7pyverbs_4cmid_AddrInfo;
  PyObject *__pyx_type_7pyverbs_4cmid_ConnParam;
  PyObject *__pyx_type_7pyverbs_4cmid_UDParam;
  PyObject *__pyx_type_7pyverbs_4cmid_JoinMCAttrEx;
  PyTypeObject *__pyx_ptype_7pyverbs_4cmid_CMID;
  PyTypeObject *__pyx_ptype_7pyverbs_4cmid_CMEventChannel;
  PyTypeObject *__pyx_ptype_7pyverbs_4cmid_CMEvent;
  PyTypeObject *__pyx_ptype_7pyverbs_4cmid_AddrInfo;
  PyTypeObject *__pyx_ptype_7pyverbs_4cmid_ConnParam;
  PyTypeObject *__pyx_ptype_7pyverbs_4cmid_UDParam;
  PyTypeObject *__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_items;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_values;
  __Pyx_CachedCFunction __pyx_umethod_PyUnicode_Type__format;
  PyObject *__pyx_k__4;
  PyObject *__pyx_k__5;
  PyObject *__pyx_k__6;
  PyObject *__pyx_tuple[7];
  PyObject *__pyx_codeobj_tab[52];
  PyObject *__pyx_string_tab[432];
  PyObject *__pyx_number_tab[4];
/* #### Code section: module_state_contents ### */
/* CommonTypesMetaclass.module_state_decls */
PyTypeObject *__pyx_CommonTypesMetaclassType;

/* CachedMethodType.module_state_decls */
#if CYTHON_COMPILING_IN_LIMITED_API
PyObject *__Pyx_CachedMethodType;
#endif

/* CythonFunctionShared.module_state_decls */
PyTypeObject *__pyx_CyFunctionType;

/* CodeObjectCache.module_state_decls */
struct __Pyx_CodeObjectCache __pyx_code_cache;

/* #### Code section: module_state_end ### */
} __pyx_mstatetype;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate_global (__Pyx_PyModule_GetState(__Pyx_State_FindModule(&__pyx_moduledef)))

#define __pyx_m (__Pyx_State_FindModule(&__pyx_moduledef))
#else
static __pyx_mstatetype __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstatetype * const __pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: constant_name_defines ### */
#define __pyx_kp_u_ __pyx_string_tab[0]
#define __pyx_kp_u_4_4 __pyx_string_tab[1]
#define __pyx_kp_u_Cannot_create_CM_ID_from_obj __pyx_string_tab[2]
#define __pyx_kp_u_Closing_AddrInfo __pyx_string_tab[3]
#define __pyx_kp_u_Closing_CMEvent __pyx_string_tab[4]
#define __pyx_kp_u_Closing_CMEventChannel __pyx_string_tab[5]
#define __pyx_kp_u_Closing_CMID __pyx_string_tab[6]
#define __pyx_kp_u_Connection_parameters __pyx_string_tab[7]
#define __pyx_kp_u_Created_a_CMEvent __pyx_string_tab[8]
#define __pyx_kp_u_Created_a_CMEventChannel __pyx_string_tab[9]
#define __pyx_kp_u_Currently_only_RDMA_OPTION_ID_AC __pyx_string_tab[10]
#define __pyx_kp_u_Failed_to_Accept_Connection __pyx_string_tab[11]
#define __pyx_kp_u_Failed_to_Acknowledge_Event __pyx_string_tab[12]
#define __pyx_kp_u_Failed_to_Bind_ID __pyx_string_tab[13]
#define __pyx_kp_u_Failed_to_Complete_an_active_con __pyx_string_tab[14]
#define __pyx_kp_u_Failed_to_Connect __pyx_string_tab[15]
#define __pyx_kp_u_Failed_to_Create_QP __pyx_string_tab[16]
#define __pyx_kp_u_Failed_to_Disconnect __pyx_string_tab[17]
#define __pyx_kp_u_Failed_to_Join_multicast __pyx_string_tab[18]
#define __pyx_kp_u_Failed_to_Post_Read __pyx_string_tab[19]
#define __pyx_kp_u_Failed_to_Post_Receive __pyx_string_tab[20]
#define __pyx_kp_u_Failed_to_Post_Send __pyx_string_tab[21]
#define __pyx_kp_u_Failed_to_Post_Write __pyx_string_tab[22]
#define __pyx_kp_u_Failed_to_Query_AddrInfo __pyx_string_tab[23]
#define __pyx_kp_u_Failed_to_Reject_Connection __pyx_string_tab[24]
#define __pyx_kp_u_Failed_to_Resolve_AddrInfo __pyx_string_tab[25]
#define __pyx_kp_u_Failed_to_Resolve_Address __pyx_string_tab[26]
#define __pyx_kp_u_Failed_to_Resolve_Route __pyx_string_tab[27]
#define __pyx_kp_u_Failed_to_close_CMID __pyx_string_tab[28]
#define __pyx_kp_u_Failed_to_create_CMEvent __pyx_string_tab[29]
#define __pyx_kp_u_Failed_to_create_CMEventChannel __pyx_string_tab[30]
#define __pyx_kp_u_Failed_to_create_CM_ID __pyx_string_tab[31]
#define __pyx_kp_u_Failed_to_get_Address_Info __pyx_string_tab[32]
#define __pyx_kp_u_Failed_to_get_QP_attributes __pyx_string_tab[33]
#define __pyx_kp_u_Failed_to_get_remote_ECE __pyx_string_tab[34]
#define __pyx_kp_u_Failed_to_get_request_no_connect __pyx_string_tab[35]
#define __pyx_kp_u_Failed_to_leave_multicast __pyx_string_tab[36]
#define __pyx_kp_u_Failed_to_query_QP __pyx_string_tab[37]
#define __pyx_kp_u_Failed_to_retrieve_receive_compl __pyx_string_tab[38]
#define __pyx_kp_u_Failed_to_retrieve_send_completi __pyx_string_tab[39]
#define __pyx_kp_u_Failed_to_set_local_ECE __pyx_string_tab[40]
#define __pyx_kp_u_Failed_to_set_option __pyx_string_tab[41]
#define __pyx_kp_u_Join_to_multicast_must_have_Addr __pyx_string_tab[42]
#define __pyx_kp_u_Listen_Failed __pyx_string_tab[43]
#define __pyx_kp_u_Note_that_Cython_is_deliberately __pyx_string_tab[44]
#define __pyx_kp_u_Unrecognized_object_type __pyx_string_tab[45]
#define __pyx_kp_u__2 __pyx_string_tab[46]
#define __pyx_kp_u__3 __pyx_string_tab[47]
#define __pyx_kp_u__7 __pyx_string_tab[48]
#define __pyx_kp_u_add_note __pyx_string_tab[49]
#define __pyx_kp_u_disable __pyx_string_tab[50]
#define __pyx_kp_u_enable __pyx_string_tab[51]
#define __pyx_kp_u_enum_class_ibv_access_flags_not __pyx_string_tab[52]
#define __pyx_kp_u_enum_class_ibv_send_flags_not_im __pyx_string_tab[53]
#define __pyx_kp_u_enum_class_ibv_wc_opcode_not_imp __pyx_string_tab[54]
#define __pyx_kp_u_enum_class_ibv_wc_status_not_imp __pyx_string_tab[55]
#define __pyx_kp_u_enum_class_rdma_cm_event_type_no __pyx_string_tab[56]
#define __pyx_kp_u_enum_class_rdma_port_space_not_i __pyx_string_tab[57]
#define __pyx_kp_u_flow_control_2 __pyx_string_tab[58]
#define __pyx_kp_u_gc __pyx_string_tab[59]
#define __pyx_kp_u_initiator_depth __pyx_string_tab[60]
#define __pyx_kp_u_isenabled __pyx_string_tab[61]
#define __pyx_kp_u_listen_ID_not_provided __pyx_string_tab[62]
#define __pyx_kp_u_pyverbs_cmid_pyx __pyx_string_tab[63]
#define __pyx_kp_u_qp_number __pyx_string_tab[64]
#define __pyx_kp_u_responder_resources __pyx_string_tab[65]
#define __pyx_kp_u_retry_count __pyx_string_tab[66]
#define __pyx_kp_u_rnr_retry_count __pyx_string_tab[67]
#define __pyx_kp_u_self_addr_info_cannot_be_convert __pyx_string_tab[68]
#define __pyx_kp_u_self_conn_param_cannot_be_conver __pyx_string_tab[69]
#define __pyx_kp_u_self_event_cannot_be_converted_t __pyx_string_tab[70]
#define __pyx_kp_u_self_event_channel_cannot_be_con __pyx_string_tab[71]
#define __pyx_kp_u_self_id_cannot_be_converted_to_a __pyx_string_tab[72]
#define __pyx_kp_u_self_join_mc_attr_ex_cannot_be_c __pyx_string_tab[73]
#define __pyx_kp_u_self_ud_param_cannot_be_converte __pyx_string_tab[74]
#define __pyx_kp_u_tree_fragment __pyx_string_tab[75]
#define __pyx_kp_u_utf_8 __pyx_string_tab[76]
#define __pyx_n_u_AddrInfo __pyx_string_tab[77]
#define __pyx_n_u_AddrInfo___reduce_cython __pyx_string_tab[78]
#define __pyx_n_u_AddrInfo___setstate_cython __pyx_string_tab[79]
#define __pyx_n_u_AddrInfo_close __pyx_string_tab[80]
#define __pyx_n_u_AddrInfo_get_addrinfo __pyx_string_tab[81]
#define __pyx_n_u_CMEvent __pyx_string_tab[82]
#define __pyx_n_u_CMEventChannel __pyx_string_tab[83]
#define __pyx_n_u_CMEventChannel___reduce_cython __pyx_string_tab[84]
#define __pyx_n_u_CMEventChannel___setstate_cython __pyx_string_tab[85]
#define __pyx_n_u_CMEventChannel_close __pyx_string_tab[86]
#define __pyx_n_u_CMEvent___reduce_cython __pyx_string_tab[87]
#define __pyx_n_u_CMEvent___setstate_cython __pyx_string_tab[88]
#define __pyx_n_u_CMEvent_ack_cm_event __pyx_string_tab[89]
#define __pyx_n_u_CMEvent_close __pyx_string_tab[90]
#define __pyx_n_u_CMEvent_event_str __pyx_string_tab[91]
#define __pyx_n_u_CMID __pyx_string_tab[92]
#define __pyx_n_u_CMID___reduce_cython __pyx_string_tab[93]
#define __pyx_n_u_CMID___setstate_cython __pyx_string_tab[94]
#define __pyx_n_u_CMID_accept __pyx_string_tab[95]
#define __pyx_n_u_CMID_bind_addr __pyx_string_tab[96]
#define __pyx_n_u_CMID_close __pyx_string_tab[97]
#define __pyx_n_u_CMID_connect __pyx_string_tab[98]
#define __pyx_n_u_CMID_create_qp __pyx_string_tab[99]
#define __pyx_n_u_CMID_disconnect __pyx_string_tab[100]
#define __pyx_n_u_CMID_establish __pyx_string_tab[101]
#define __pyx_n_u_CMID_get_recv_comp __pyx_string_tab[102]
#define __pyx_n_u_CMID_get_remote_ece __pyx_string_tab[103]
#define __pyx_n_u_CMID_get_request __pyx_string_tab[104]
#define __pyx_n_u_CMID_get_send_comp __pyx_string_tab[105]
#define __pyx_n_u_CMID_init_qp_attr __pyx_string_tab[106]
#define __pyx_n_u_CMID_join_multicast __pyx_string_tab[107]
#define __pyx_n_u_CMID_leave_multicast __pyx_string_tab[108]
#define __pyx_n_u_CMID_listen __pyx_string_tab[109]
#define __pyx_n_u_CMID_post_read __pyx_string_tab[110]
#define __pyx_n_u_CMID_post_recv __pyx_string_tab[111]
#define __pyx_n_u_CMID_post_send __pyx_string_tab[112]
#define __pyx_n_u_CMID_post_ud_send __pyx_string_tab[113]
#define __pyx_n_u_CMID_post_write __pyx_string_tab[114]
#define __pyx_n_u_CMID_query_addrinfo __pyx_string_tab[115]
#define __pyx_n_u_CMID_query_qp __pyx_string_tab[116]
#define __pyx_n_u_CMID_reg_msgs __pyx_string_tab[117]
#define __pyx_n_u_CMID_reg_read __pyx_string_tab[118]
#define __pyx_n_u_CMID_reg_write __pyx_string_tab[119]
#define __pyx_n_u_CMID_reject __pyx_string_tab[120]
#define __pyx_n_u_CMID_resolve_addr __pyx_string_tab[121]
#define __pyx_n_u_CMID_resolve_addrinfo __pyx_string_tab[122]
#define __pyx_n_u_CMID_resolve_route __pyx_string_tab[123]
#define __pyx_n_u_CMID_set_local_ece __pyx_string_tab[124]
#define __pyx_n_u_CMID_set_option __pyx_string_tab[125]
#define __pyx_n_u_ConnParam __pyx_string_tab[126]
#define __pyx_n_u_ConnParam___reduce_cython __pyx_string_tab[127]
#define __pyx_n_u_ConnParam___setstate_cython __pyx_string_tab[128]
#define __pyx_n_u_ConnParam_set_private_data __pyx_string_tab[129]
#define __pyx_n_u_IBV_ACCESS_FLUSH_GLOBAL __pyx_string_tab[130]
#define __pyx_n_u_IBV_ACCESS_FLUSH_PERSISTENT __pyx_string_tab[131]
#define __pyx_n_u_IBV_ACCESS_HUGETLB __pyx_string_tab[132]
#define __pyx_n_u_IBV_ACCESS_LOCAL_WRITE __pyx_string_tab[133]
#define __pyx_n_u_IBV_ACCESS_MW_BIND __pyx_string_tab[134]
#define __pyx_n_u_IBV_ACCESS_ON_DEMAND __pyx_string_tab[135]
#define __pyx_n_u_IBV_ACCESS_RELAXED_ORDERING __pyx_string_tab[136]
#define __pyx_n_u_IBV_ACCESS_REMOTE_ATOMIC __pyx_string_tab[137]
#define __pyx_n_u_IBV_ACCESS_REMOTE_READ __pyx_string_tab[138]
#define __pyx_n_u_IBV_ACCESS_REMOTE_WRITE __pyx_string_tab[139]
#define __pyx_n_u_IBV_ACCESS_ZERO_BASED __pyx_string_tab[140]
#define __pyx_n_u_IBV_SEND_FENCE __pyx_string_tab[141]
#define __pyx_n_u_IBV_SEND_INLINE __pyx_string_tab[142]
#define __pyx_n_u_IBV_SEND_IP_CSUM __pyx_string_tab[143]
#define __pyx_n_u_IBV_SEND_SIGNALED __pyx_string_tab[144]
#define __pyx_n_u_IBV_SEND_SOLICITED __pyx_string_tab[145]
#define __pyx_n_u_IBV_WC_ATOMIC_WRITE __pyx_string_tab[146]
#define __pyx_n_u_IBV_WC_BAD_RESP_ERR __pyx_string_tab[147]
#define __pyx_n_u_IBV_WC_BIND_MW __pyx_string_tab[148]
#define __pyx_n_u_IBV_WC_COMP_SWAP __pyx_string_tab[149]
#define __pyx_n_u_IBV_WC_DRIVER2 __pyx_string_tab[150]
#define __pyx_n_u_IBV_WC_DRIVER3 __pyx_string_tab[151]
#define __pyx_n_u_IBV_WC_FATAL_ERR __pyx_string_tab[152]
#define __pyx_n_u_IBV_WC_FETCH_ADD __pyx_string_tab[153]
#define __pyx_n_u_IBV_WC_FLUSH __pyx_string_tab[154]
#define __pyx_n_u_IBV_WC_GENERAL_ERR __pyx_string_tab[155]
#define __pyx_n_u_IBV_WC_INV_EECN_ERR __pyx_string_tab[156]
#define __pyx_n_u_IBV_WC_INV_EEC_STATE_ERR __pyx_string_tab[157]
#define __pyx_n_u_IBV_WC_LOCAL_INV __pyx_string_tab[158]
#define __pyx_n_u_IBV_WC_LOC_ACCESS_ERR __pyx_string_tab[159]
#define __pyx_n_u_IBV_WC_LOC_EEC_OP_ERR __pyx_string_tab[160]
#define __pyx_n_u_IBV_WC_LOC_LEN_ERR __pyx_string_tab[161]
#define __pyx_n_u_IBV_WC_LOC_PROT_ERR __pyx_string_tab[162]
#define __pyx_n_u_IBV_WC_LOC_QP_OP_ERR __pyx_string_tab[163]
#define __pyx_n_u_IBV_WC_LOC_RDD_VIOL_ERR __pyx_string_tab[164]
#define __pyx_n_u_IBV_WC_MW_BIND_ERR __pyx_string_tab[165]
#define __pyx_n_u_IBV_WC_RDMA_READ __pyx_string_tab[166]
#define __pyx_n_u_IBV_WC_RDMA_WRITE __pyx_string_tab[167]
#define __pyx_n_u_IBV_WC_RECV __pyx_string_tab[168]
#define __pyx_n_u_IBV_WC_RECV_RDMA_WITH_IMM __pyx_string_tab[169]
#define __pyx_n_u_IBV_WC_REM_ABORT_ERR __pyx_string_tab[170]
#define __pyx_n_u_IBV_WC_REM_ACCESS_ERR __pyx_string_tab[171]
#define __pyx_n_u_IBV_WC_REM_INV_RD_REQ_ERR __pyx_string_tab[172]
#define __pyx_n_u_IBV_WC_REM_INV_REQ_ERR __pyx_string_tab[173]
#define __pyx_n_u_IBV_WC_REM_OP_ERR __pyx_string_tab[174]
#define __pyx_n_u_IBV_WC_RESP_TIMEOUT_ERR __pyx_string_tab[175]
#define __pyx_n_u_IBV_WC_RETRY_EXC_ERR __pyx_string_tab[176]
#define __pyx_n_u_IBV_WC_RNR_RETRY_EXC_ERR __pyx_string_tab[177]
#define __pyx_n_u_IBV_WC_SEND __pyx_string_tab[178]
#define __pyx_n_u_IBV_WC_SUCCESS __pyx_string_tab[179]
#define __pyx_n_u_IBV_WC_TM_ADD __pyx_string_tab[180]
#define __pyx_n_u_IBV_WC_TM_DEL __pyx_string_tab[181]
#define __pyx_n_u_IBV_WC_TM_ERR __pyx_string_tab[182]
#define __pyx_n_u_IBV_WC_TM_NO_TAG __pyx_string_tab[183]
#define __pyx_n_u_IBV_WC_TM_RECV __pyx_string_tab[184]
#define __pyx_n_u_IBV_WC_TM_RNDV_INCOMPLETE __pyx_string_tab[185]
#define __pyx_n_u_IBV_WC_TM_SYNC __pyx_string_tab[186]
#define __pyx_n_u_IBV_WC_TSO __pyx_string_tab[187]
#define __pyx_n_u_IBV_WC_WR_FLUSH_ERR __pyx_string_tab[188]
#define __pyx_n_u_JoinMCAttrEx __pyx_string_tab[189]
#define __pyx_n_u_JoinMCAttrEx___reduce_cython __pyx_string_tab[190]
#define __pyx_n_u_JoinMCAttrEx___setstate_cython __pyx_string_tab[191]
#define __pyx_n_u_PyverbsError __pyx_string_tab[192]
#define __pyx_n_u_PyverbsRDMAErrno __pyx_string_tab[193]
#define __pyx_n_u_PyverbsRDMAError __pyx_string_tab[194]
#define __pyx_n_u_PyverbsUserError __pyx_string_tab[195]
#define __pyx_n_u_Pyx_PyDict_NextRef __pyx_string_tab[196]
#define __pyx_n_u_RDMA_CM_EVENT_ADDRINFO_ERROR __pyx_string_tab[197]
#define __pyx_n_u_RDMA_CM_EVENT_ADDRINFO_RESOLVED __pyx_string_tab[198]
#define __pyx_n_u_RDMA_CM_EVENT_ADDR_CHANGE __pyx_string_tab[199]
#define __pyx_n_u_RDMA_CM_EVENT_ADDR_ERROR __pyx_string_tab[200]
#define __pyx_n_u_RDMA_CM_EVENT_ADDR_RESOLVED __pyx_string_tab[201]
#define __pyx_n_u_RDMA_CM_EVENT_CONNECT_ERROR __pyx_string_tab[202]
#define __pyx_n_u_RDMA_CM_EVENT_CONNECT_REQUEST __pyx_string_tab[203]
#define __pyx_n_u_RDMA_CM_EVENT_CONNECT_RESPONSE __pyx_string_tab[204]
#define __pyx_n_u_RDMA_CM_EVENT_DEVICE_REMOVAL __pyx_string_tab[205]
#define __pyx_n_u_RDMA_CM_EVENT_DISCONNECTED __pyx_string_tab[206]
#define __pyx_n_u_RDMA_CM_EVENT_ESTABLISHED __pyx_string_tab[207]
#define __pyx_n_u_RDMA_CM_EVENT_MULTICAST_ERROR __pyx_string_tab[208]
#define __pyx_n_u_RDMA_CM_EVENT_MULTICAST_JOIN __pyx_string_tab[209]
#define __pyx_n_u_RDMA_CM_EVENT_REJECTED __pyx_string_tab[210]
#define __pyx_n_u_RDMA_CM_EVENT_ROUTE_ERROR __pyx_string_tab[211]
#define __pyx_n_u_RDMA_CM_EVENT_ROUTE_RESOLVED __pyx_string_tab[212]
#define __pyx_n_u_RDMA_CM_EVENT_TIMEWAIT_EXIT __pyx_string_tab[213]
#define __pyx_n_u_RDMA_CM_EVENT_UNREACHABLE __pyx_string_tab[214]
#define __pyx_n_u_RDMA_PS_IB __pyx_string_tab[215]
#define __pyx_n_u_RDMA_PS_IPOIB __pyx_string_tab[216]
#define __pyx_n_u_RDMA_PS_TCP __pyx_string_tab[217]
#define __pyx_n_u_RDMA_PS_UDP __pyx_string_tab[218]
#define __pyx_n_u_UDParam __pyx_string_tab[219]
#define __pyx_n_u_UDParam___reduce_cython __pyx_string_tab[220]
#define __pyx_n_u_UDParam___setstate_cython __pyx_string_tab[221]
#define __pyx_n_u_WeakSet __pyx_string_tab[222]
#define __pyx_n_u_accept __pyx_string_tab[223]
#define __pyx_n_u_ack_cm_event __pyx_string_tab[224]
#define __pyx_n_u_add __pyx_string_tab[225]
#define __pyx_n_u_addr __pyx_string_tab[226]
#define __pyx_n_u_addressof __pyx_string_tab[227]
#define __pyx_n_u_ah __pyx_string_tab[228]
#define __pyx_n_u_ai __pyx_string_tab[229]
#define __pyx_n_u_asyncio_coroutines __pyx_string_tab[230]
#define __pyx_n_u_attr __pyx_string_tab[231]
#define __pyx_n_u_attr_mask __pyx_string_tab[232]
#define __pyx_n_u_backlog __pyx_string_tab[233]
#define __pyx_n_u_bind_addr __pyx_string_tab[234]
#define __pyx_n_u_buffer __pyx_string_tab[235]
#define __pyx_n_u_byte_len __pyx_string_tab[236]
#define __pyx_n_u_c_char __pyx_string_tab[237]
#define __pyx_n_u_channel __pyx_string_tab[238]
#define __pyx_n_u_cline_in_traceback __pyx_string_tab[239]
#define __pyx_n_u_close __pyx_string_tab[240]
#define __pyx_n_u_cm_event __pyx_string_tab[241]
#define __pyx_n_u_cmid __pyx_string_tab[242]
#define __pyx_n_u_comp_mask __pyx_string_tab[243]
#define __pyx_n_u_conn __pyx_string_tab[244]
#define __pyx_n_u_connect __pyx_string_tab[245]
#define __pyx_n_u_context __pyx_string_tab[246]
#define __pyx_n_u_create_qp __pyx_string_tab[247]
#define __pyx_n_u_create_string_buffer __pyx_string_tab[248]
#define __pyx_n_u_creator __pyx_string_tab[249]
#define __pyx_n_u_ctypes __pyx_string_tab[250]
#define __pyx_n_u_data __pyx_string_tab[251]
#define __pyx_n_u_data_len __pyx_string_tab[252]
#define __pyx_n_u_data_ptr __pyx_string_tab[253]
#define __pyx_n_u_debug __pyx_string_tab[254]
#define __pyx_n_u_depth __pyx_string_tab[255]
#define __pyx_n_u_disconnect __pyx_string_tab[256]
#define __pyx_n_u_dlid_path_bits __pyx_string_tab[257]
#define __pyx_n_u_dst __pyx_string_tab[258]
#define __pyx_n_u_dst_addr __pyx_string_tab[259]
#define __pyx_n_u_dst_service __pyx_string_tab[260]
#define __pyx_n_u_dst_srvc __pyx_string_tab[261]
#define __pyx_n_u_ece __pyx_string_tab[262]
#define __pyx_n_u_encode __pyx_string_tab[263]
#define __pyx_n_u_establish __pyx_string_tab[264]
#define __pyx_n_u_event_str __pyx_string_tab[265]
#define __pyx_n_u_event_type __pyx_string_tab[266]
#define __pyx_n_u_flags __pyx_string_tab[267]
#define __pyx_n_u_flow_control __pyx_string_tab[268]
#define __pyx_n_u_format __pyx_string_tab[269]
#define __pyx_n_u_func __pyx_string_tab[270]
#define __pyx_n_u_get_addrinfo __pyx_string_tab[271]
#define __pyx_n_u_get_recv_comp __pyx_string_tab[272]
#define __pyx_n_u_get_remote_ece __pyx_string_tab[273]
#define __pyx_n_u_get_request __pyx_string_tab[274]
#define __pyx_n_u_get_send_comp __pyx_string_tab[275]
#define __pyx_n_u_getstate __pyx_string_tab[276]
#define __pyx_n_u_hints __pyx_string_tab[277]
#define __pyx_n_u_hints_ptr __pyx_string_tab[278]
#define __pyx_n_u_i __pyx_string_tab[279]
#define __pyx_n_u_ibv_access_flags __pyx_string_tab[280]
#define __pyx_n_u_ibv_send_flags __pyx_string_tab[281]
#define __pyx_n_u_ibv_wc_opcode __pyx_string_tab[282]
#define __pyx_n_u_ibv_wc_status __pyx_string_tab[283]
#define __pyx_n_u_imm_data __pyx_string_tab[284]
#define __pyx_n_u_init __pyx_string_tab[285]
#define __pyx_n_u_init_attr __pyx_string_tab[286]
#define __pyx_n_u_init_qp_attr __pyx_string_tab[287]
#define __pyx_n_u_is_coroutine __pyx_string_tab[288]
#define __pyx_n_u_items __pyx_string_tab[289]
#define __pyx_n_u_join_flags __pyx_string_tab[290]
#define __pyx_n_u_join_multicast __pyx_string_tab[291]
#define __pyx_n_u_lai __pyx_string_tab[292]
#define __pyx_n_u_leave_multicast __pyx_string_tab[293]
#define __pyx_n_u_length __pyx_string_tab[294]
#define __pyx_n_u_level __pyx_string_tab[295]
#define __pyx_n_u_listen __pyx_string_tab[296]
#define __pyx_n_u_listen_id __pyx_string_tab[297]
#define __pyx_n_u_main __pyx_string_tab[298]
#define __pyx_n_u_mc_attr __pyx_string_tab[299]
#define __pyx_n_u_mc_join_attr __pyx_string_tab[300]
#define __pyx_n_u_module __pyx_string_tab[301]
#define __pyx_n_u_mr __pyx_string_tab[302]
#define __pyx_n_u_name __pyx_string_tab[303]
#define __pyx_n_u_node __pyx_string_tab[304]
#define __pyx_n_u_node_addr __pyx_string_tab[305]
#define __pyx_n_u_obj __pyx_string_tab[306]
#define __pyx_n_u_opcode __pyx_string_tab[307]
#define __pyx_n_u_optlen __pyx_string_tab[308]
#define __pyx_n_u_optname __pyx_string_tab[309]
#define __pyx_n_u_optval __pyx_string_tab[310]
#define __pyx_n_u_param __pyx_string_tab[311]
#define __pyx_n_u_pd __pyx_string_tab[312]
#define __pyx_n_u_pkey_index __pyx_string_tab[313]
#define __pyx_n_u_pop __pyx_string_tab[314]
#define __pyx_n_u_port_space __pyx_string_tab[315]
#define __pyx_n_u_post_read __pyx_string_tab[316]
#define __pyx_n_u_post_recv __pyx_string_tab[317]
#define __pyx_n_u_post_send __pyx_string_tab[318]
#define __pyx_n_u_post_ud_send __pyx_string_tab[319]
#define __pyx_n_u_post_write __pyx_string_tab[320]
#define __pyx_n_u_private_data __pyx_string_tab[321]
#define __pyx_n_u_ptr __pyx_string_tab[322]
#define __pyx_n_u_pyverbs_base __pyx_string_tab[323]
#define __pyx_n_u_pyverbs_cmid __pyx_string_tab[324]
#define __pyx_n_u_pyverbs_libibverbs_enums __pyx_string_tab[325]
#define __pyx_n_u_pyverbs_librdmacm_enums __pyx_string_tab[326]
#define __pyx_n_u_pyverbs_pyverbs_error __pyx_string_tab[327]
#define __pyx_n_u_pyx_state __pyx_string_tab[328]
#define __pyx_n_u_pyx_vtable __pyx_string_tab[329]
#define __pyx_n_u_qp_attr __pyx_string_tab[330]
#define __pyx_n_u_qp_init __pyx_string_tab[331]
#define __pyx_n_u_qp_init_attr __pyx_string_tab[332]
#define __pyx_n_u_qp_num __pyx_string_tab[333]
#define __pyx_n_u_qp_state __pyx_string_tab[334]
#define __pyx_n_u_qualname __pyx_string_tab[335]
#define __pyx_n_u_query_addrinfo __pyx_string_tab[336]
#define __pyx_n_u_query_qp __pyx_string_tab[337]
#define __pyx_n_u_rai __pyx_string_tab[338]
#define __pyx_n_u_rc __pyx_string_tab[339]
#define __pyx_n_u_rdma_cm_event_type __pyx_string_tab[340]
#define __pyx_n_u_rdma_port_space __pyx_string_tab[341]
#define __pyx_n_u_reduce __pyx_string_tab[342]
#define __pyx_n_u_reduce_cython __pyx_string_tab[343]
#define __pyx_n_u_reduce_ex __pyx_string_tab[344]
#define __pyx_n_u_reg_msgs __pyx_string_tab[345]
#define __pyx_n_u_reg_read __pyx_string_tab[346]
#define __pyx_n_u_reg_write __pyx_string_tab[347]
#define __pyx_n_u_reject __pyx_string_tab[348]
#define __pyx_n_u_remote_addr __pyx_string_tab[349]
#define __pyx_n_u_res __pyx_string_tab[350]
#define __pyx_n_u_resolve_addr __pyx_string_tab[351]
#define __pyx_n_u_resolve_addrinfo __pyx_string_tab[352]
#define __pyx_n_u_resolve_route __pyx_string_tab[353]
#define __pyx_n_u_resources __pyx_string_tab[354]
#define __pyx_n_u_ret __pyx_string_tab[355]
#define __pyx_n_u_retry __pyx_string_tab[356]
#define __pyx_n_u_rkey __pyx_string_tab[357]
#define __pyx_n_u_rnr_retry __pyx_string_tab[358]
#define __pyx_n_u_rqpn __pyx_string_tab[359]
#define __pyx_n_u_self __pyx_string_tab[360]
#define __pyx_n_u_service __pyx_string_tab[361]
#define __pyx_n_u_set_local_ece __pyx_string_tab[362]
#define __pyx_n_u_set_name __pyx_string_tab[363]
#define __pyx_n_u_set_option __pyx_string_tab[364]
#define __pyx_n_u_set_private_data __pyx_string_tab[365]
#define __pyx_n_u_setdefault __pyx_string_tab[366]
#define __pyx_n_u_setstate __pyx_string_tab[367]
#define __pyx_n_u_setstate_cython __pyx_string_tab[368]
#define __pyx_n_u_size __pyx_string_tab[369]
#define __pyx_n_u_sl __pyx_string_tab[370]
#define __pyx_n_u_slid __pyx_string_tab[371]
#define __pyx_n_u_src __pyx_string_tab[372]
#define __pyx_n_u_src_addr __pyx_string_tab[373]
#define __pyx_n_u_src_qp __pyx_string_tab[374]
#define __pyx_n_u_src_service __pyx_string_tab[375]
#define __pyx_n_u_src_srvc __pyx_string_tab[376]
#define __pyx_n_u_srq __pyx_string_tab[377]
#define __pyx_n_u_srvc __pyx_string_tab[378]
#define __pyx_n_u_status __pyx_string_tab[379]
#define __pyx_n_u_super __pyx_string_tab[380]
#define __pyx_n_u_test __pyx_string_tab[381]
#define __pyx_n_u_timeout_ms __pyx_string_tab[382]
#define __pyx_n_u_to_conn __pyx_string_tab[383]
#define __pyx_n_u_value __pyx_string_tab[384]
#define __pyx_n_u_values __pyx_string_tab[385]
#define __pyx_n_u_vendor_err __pyx_string_tab[386]
#define __pyx_n_u_warn __pyx_string_tab[387]
#define __pyx_n_u_warnings __pyx_string_tab[388]
#define __pyx_n_u_wc __pyx_string_tab[389]
#define __pyx_n_u_wc_flags __pyx_string_tab[390]
#define __pyx_n_u_weakref __pyx_string_tab[391]
#define __pyx_n_u_wr_id __pyx_string_tab[392]
#define __pyx_kp_b_PyObject_PyObject_close_weakrefs __pyx_string_tab[393]
#define __pyx_kp_b_iso88591_3a_AQ_A_1_vZq_AT_m_Q_4s_1 __pyx_string_tab[394]
#define __pyx_kp_b_iso88591_7q_4q_Rs_4uF_F_A_E_4s_1 __pyx_string_tab[395]
#define __pyx_kp_b_iso88591_A_1D_1G3a_E_as_4q_Qe1_E_e3at83aq __pyx_string_tab[396]
#define __pyx_kp_b_iso88591_A_1_e1G7_1_3c_1_y __pyx_string_tab[397]
#define __pyx_kp_b_iso88591_A_4_Q_t1_G6_Q_M __pyx_string_tab[398]
#define __pyx_kp_b_iso88591_A_4_c_t1_G6_a __pyx_string_tab[399]
#define __pyx_kp_b_iso88591_A_4q_4s_1_4z_IQ __pyx_string_tab[400]
#define __pyx_kp_b_iso88591_A_4t3a_t1_G6_t_Q_4q_4s_c_at1_Q_4 __pyx_string_tab[401]
#define __pyx_kp_b_iso88591_A_4uAS_3c_1 __pyx_string_tab[402]
#define __pyx_kp_b_iso88591_A_4wc_1 __pyx_string_tab[403]
#define __pyx_kp_b_iso88591_A_4wc_t1_G6_Q __pyx_string_tab[404]
#define __pyx_kp_b_iso88591_A_85_1_Q_U_4s_1 __pyx_string_tab[405]
#define __pyx_kp_b_iso88591_A_Q_4s_1 __pyx_string_tab[406]
#define __pyx_kp_b_iso88591_A_Qd_t_Q_4s_1 __pyx_string_tab[407]
#define __pyx_kp_b_iso88591_A_U_Q_4s_1 __pyx_string_tab[408]
#define __pyx_kp_b_iso88591_A_XQ_AT_b_4s_1_q __pyx_string_tab[409]
#define __pyx_kp_b_iso88591_A_a_Qd_q_q_4s_1_G7_5_F_AQ_q __pyx_string_tab[410]
#define __pyx_kp_b_iso88591_A_c_1D_Qc_3c_1_q __pyx_string_tab[411]
#define __pyx_kp_b_iso88591_A_q_A_4s_1 __pyx_string_tab[412]
#define __pyx_kp_b_iso88591_A_q_E_4r_1_A_1_r_9G2Q_R_IRq_9G2Q __pyx_string_tab[413]
#define __pyx_kp_b_iso88591_A_q_E_Ja_4s_1_4uCq_wauA_4t3a_b __pyx_string_tab[414]
#define __pyx_kp_b_iso88591_A_q_E_r_xr_A_4s_1 __pyx_string_tab[415]
#define __pyx_kp_b_iso88591_A_q_F_d_e1G1_4s_1 __pyx_string_tab[416]
#define __pyx_kp_b_iso88591_A_r_q __pyx_string_tab[417]
#define __pyx_kp_b_iso88591_A_r_q_2 __pyx_string_tab[418]
#define __pyx_kp_b_iso88591_A_vQ_Ja_m1D_5_WKq_3c_9_vQ __pyx_string_tab[419]
#define __pyx_kp_b_iso88591_Fa_4uD_A_1_4q_U_j_0_1_1D_QgQ_3 __pyx_string_tab[420]
#define __pyx_kp_b_iso88591_J_4q_Rs_q_E_r_xr_4s_1 __pyx_string_tab[421]
#define __pyx_kp_b_iso88591_N2C9A_q_1_5_q_z_t7_1_q_87_z_1_7 __pyx_string_tab[422]
#define __pyx_kp_b_iso88591_Q __pyx_string_tab[423]
#define __pyx_kp_b_iso88591_Q_4uCz_j_a_4s_1 __pyx_string_tab[424]
#define __pyx_kp_b_iso88591_Q_AT_a_4s_1 __pyx_string_tab[425]
#define __pyx_kp_b_iso88591_Q_o_AT_a_4s_1 __pyx_string_tab[426]
#define __pyx_kp_b_iso88591_Q_r_q __pyx_string_tab[427]
#define __pyx_kp_b_iso88591_a_o_Qd_q_4s_1 __pyx_string_tab[428]
#define __pyx_kp_b_iso88591_q_1D_Q_4s_1_4uCq_wauA_4t3a_b __pyx_string_tab[429]
#define __pyx_kp_b_iso88591_q_4q_Rs_q_E_r_xr_4s_1 __pyx_string_tab[430]
#define __pyx_kp_b_iso88591_z_MQ_a_a_a_a_1_Q_4wa_z_q_c_wa_4 __pyx_string_tab[431]
#define __pyx_int_0 __pyx_number_tab[0]
#define __pyx_int_1 __pyx_number_tab[1]
#define __pyx_int_5 __pyx_number_tab[2]
#define __pyx_int_2000 __pyx_number_tab[3]
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_clear(PyObject *m) {
  __pyx_mstatetype *clear_module_state = __Pyx_PyModule_GetState(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __Pyx_State_RemoveModule(NULL);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4base_PyverbsObject);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4base_PyverbsCM);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_QPCap);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_QPInitAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_QPAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_QP);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_DataBuffer);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_QPEx);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2qp_ECE);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4addr_GID);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4addr_GRH);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4addr_GlobalRoute);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4addr_AHAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4addr_AH);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_Context);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_DeviceAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_QueryDeviceExInput);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_ODPCaps);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_RSSCaps);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_PacketPacingCaps);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_PCIAtomicCaps);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_TMCaps);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_CQModerationCaps);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_TSOCaps);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_DeviceAttrEx);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_AllocDmAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_DM);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_PortAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_GIDEntry);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_AsyncEvent);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_6device_FdArr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2pd_PD);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2pd_ParentDomainInitAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2pd_ParentDomain);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2pd_ParentDomainContext);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2mr_MR);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2mr_MREx);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2mr_MWBindInfo);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2mr_MWBind);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2mr_MW);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2mr_DMMR);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2mr_DmaBufMR);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2mr_DmaHandleInitAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2mr_DMAHandle);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2cq_CompChannel);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2cq_CQ);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2cq_CQEX);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2cq_WC);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2cq_PollCqAttr);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_2cq_WcTmInfo);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4cmid_CMID);
  Py_CLEAR(clear_module_state->__pyx_type_7pyverbs_4cmid_CMID);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4cmid_CMEventChannel);
  Py_CLEAR(clear_module_state->__pyx_type_7pyverbs_4cmid_CMEventChannel);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4cmid_CMEvent);
  Py_CLEAR(clear_module_state->__pyx_type_7pyverbs_4cmid_CMEvent);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4cmid_AddrInfo);
  Py_CLEAR(clear_module_state->__pyx_type_7pyverbs_4cmid_AddrInfo);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4cmid_ConnParam);
  Py_CLEAR(clear_module_state->__pyx_type_7pyverbs_4cmid_ConnParam);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4cmid_UDParam);
  Py_CLEAR(clear_module_state->__pyx_type_7pyverbs_4cmid_UDParam);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx);
  Py_CLEAR(clear_module_state->__pyx_type_7pyverbs_4cmid_JoinMCAttrEx);
  Py_CLEAR(clear_module_state->__pyx_k__4);
  Py_CLEAR(clear_module_state->__pyx_k__5);
  Py_CLEAR(clear_module_state->__pyx_k__6);
  for (int i=0; i<7; ++i) { Py_CLEAR(clear_module_state->__pyx_tuple[i]); }
  for (int i=0; i<52; ++i) { Py_CLEAR(clear_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<432; ++i) { Py_CLEAR(clear_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<4; ++i) { Py_CLEAR(clear_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_clear_contents ### */
/* CommonTypesMetaclass.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CyFunctionType);

/* #### Code section: module_state_clear_end ### */
return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstatetype *traverse_module_state = __Pyx_PyModule_GetState(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_tuple);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_bytes);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_unicode);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4base_PyverbsObject);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4base_PyverbsCM);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_QPCap);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_QPInitAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_QPAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_QP);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_DataBuffer);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_QPEx);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2qp_ECE);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4addr_GID);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4addr_GRH);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4addr_GlobalRoute);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4addr_AHAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4addr_AH);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_Context);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_DeviceAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_QueryDeviceExInput);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_ODPCaps);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_RSSCaps);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_PacketPacingCaps);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_PCIAtomicCaps);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_TMCaps);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_CQModerationCaps);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_TSOCaps);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_DeviceAttrEx);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_AllocDmAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_DM);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_PortAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_GIDEntry);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_AsyncEvent);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_6device_FdArr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2pd_PD);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2pd_ParentDomainInitAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2pd_ParentDomain);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2pd_ParentDomainContext);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2mr_MR);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2mr_MREx);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2mr_MWBindInfo);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2mr_MWBind);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2mr_MW);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2mr_DMMR);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2mr_DmaBufMR);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2mr_DmaHandleInitAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2mr_DMAHandle);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2cq_CompChannel);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2cq_CQ);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2cq_CQEX);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2cq_WC);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2cq_PollCqAttr);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_2cq_WcTmInfo);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4cmid_CMID);
  Py_VISIT(traverse_module_state->__pyx_type_7pyverbs_4cmid_CMID);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4cmid_CMEventChannel);
  Py_VISIT(traverse_module_state->__pyx_type_7pyverbs_4cmid_CMEventChannel);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4cmid_CMEvent);
  Py_VISIT(traverse_module_state->__pyx_type_7pyverbs_4cmid_CMEvent);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4cmid_AddrInfo);
  Py_VISIT(traverse_module_state->__pyx_type_7pyverbs_4cmid_AddrInfo);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4cmid_ConnParam);
  Py_VISIT(traverse_module_state->__pyx_type_7pyverbs_4cmid_ConnParam);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4cmid_UDParam);
  Py_VISIT(traverse_module_state->__pyx_type_7pyverbs_4cmid_UDParam);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx);
  Py_VISIT(traverse_module_state->__pyx_type_7pyverbs_4cmid_JoinMCAttrEx);
  Py_VISIT(traverse_module_state->__pyx_k__4);
  Py_VISIT(traverse_module_state->__pyx_k__5);
  Py_VISIT(traverse_module_state->__pyx_k__6);
  for (int i=0; i<7; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_tuple[i]); }
  for (int i=0; i<52; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<432; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<4; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_traverse_contents ### */
/* CommonTypesMetaclass.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CyFunctionType);

/* #### Code section: module_state_traverse_end ### */
return 0;
}
#endif
/* #### Code section: module_code ### */

/* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py(rdma_port_space c_val):
 *     cdef object __pyx_enum
*/

static PyObject *__Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py(enum rdma_port_space __pyx_v_c_val) {
  PyObject *__pyx_v___pyx_enum = 0;
  PyObject *__pyx_v_warnings = NULL;
  int __pyx_v_underlying_c_val;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  size_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py", 0);

  /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.librdmacm_enums import rdma_port_space as __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py == Py_None);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.librdmacm_enums import rdma_port_space as __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py
 *         except ImportError:
*/
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "EnumTypeToPy":13
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py is None:
 *         try:
 *             from pyverbs.librdmacm_enums import rdma_port_space as __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py             # <<<<<<<<<<<<<<
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py = False  # False indicates "don't try again"
*/
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_rdma_port_space};
          __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyverbs_librdmacm_enums, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L4_error)
        }
        __pyx_t_5 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_5);
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_rdma_port_space};
          __pyx_t_7 = 0; {
            __pyx_t_8 = __Pyx_ImportFrom(__pyx_t_5, __pyx_imported_names[__pyx_t_7]); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 13, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_8);
            switch (__pyx_t_7) {
              case 0:
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py);
              __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py, __pyx_t_8);
              __Pyx_GIVEREF(__pyx_t_8);
              break;
              default:;
            }
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.librdmacm_enums import rdma_port_space as __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py
 *         except ImportError:
*/
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "EnumTypeToPy":14
 *         try:
 *             from pyverbs.librdmacm_enums import rdma_port_space as __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py
 *         except ImportError:             # <<<<<<<<<<<<<<
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py = False  # False indicates "don't try again"
 *             import warnings
*/
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_ImportError))));
      if (__pyx_t_9) {
        __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_8, &__pyx_t_10) < 0) __PYX_ERR(1, 14, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_10);

        /* "EnumTypeToPy":15
 *             from pyverbs.librdmacm_enums import rdma_port_space as __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py = False  # False indicates "don't try again"             # <<<<<<<<<<<<<<
 *             import warnings
 *             warnings.warn(
*/
        __Pyx_INCREF(Py_False);
        __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py);
        __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py, Py_False);
        __Pyx_GIVEREF(Py_False);

        /* "EnumTypeToPy":16
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py = False  # False indicates "don't try again"
 *             import warnings             # <<<<<<<<<<<<<<
 *             warnings.warn(
 *                 f"enum class rdma_port_space not importable from pyverbs.librdmacm_enums. "
*/
        __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_warnings, 0, 0, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 16, __pyx_L6_except_error)
        __pyx_t_11 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_v_warnings = __pyx_t_11;
        __pyx_t_11 = 0;

        /* "EnumTypeToPy":17
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py = False  # False indicates "don't try again"
 *             import warnings
 *             warnings.warn(             # <<<<<<<<<<<<<<
 *                 f"enum class rdma_port_space not importable from pyverbs.librdmacm_enums. "
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
*/
        __pyx_t_12 = __pyx_v_warnings;
        __Pyx_INCREF(__pyx_t_12);
        __pyx_t_13 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_mstate_global->__pyx_kp_u_enum_class_rdma_port_space_not_i};
          __pyx_t_11 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_warn, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 17, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_11);
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L5_exception_handled;
      }
      goto __pyx_L6_except_error;

      /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.librdmacm_enums import rdma_port_space as __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py
 *         except ImportError:
*/
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L5_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_L9_try_end:;
    }

    /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.librdmacm_enums import rdma_port_space as __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py
*/
  }

  /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py == Py_False);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":24
 * 
 * 
 *         return <int>c_val             # <<<<<<<<<<<<<<
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyLong_From_int(((int)__pyx_v_c_val)); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumTypeToPy":25
 * 
 *         return <int>c_val
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py             # <<<<<<<<<<<<<<
 * 
 *     if 0:
*/
  __Pyx_INCREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py);
  __pyx_v___pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py;

  /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == rdma_port_space.RDMA_PS_IPOIB:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_PS_IPOIB
 *     elif c_val == rdma_port_space.RDMA_PS_TCP:
*/
  __pyx_t_1 = (__pyx_v_c_val == RDMA_PS_IPOIB);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":30
 *         pass
 *     elif c_val == rdma_port_space.RDMA_PS_IPOIB:
 *         return __pyx_enum.RDMA_PS_IPOIB             # <<<<<<<<<<<<<<
 *     elif c_val == rdma_port_space.RDMA_PS_TCP:
 *         return __pyx_enum.RDMA_PS_TCP
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_RDMA_PS_IPOIB); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == rdma_port_space.RDMA_PS_IPOIB:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_PS_IPOIB
 *     elif c_val == rdma_port_space.RDMA_PS_TCP:
*/
  }

  /* "EnumTypeToPy":31
 *     elif c_val == rdma_port_space.RDMA_PS_IPOIB:
 *         return __pyx_enum.RDMA_PS_IPOIB
 *     elif c_val == rdma_port_space.RDMA_PS_TCP:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_PS_TCP
 *     elif c_val == rdma_port_space.RDMA_PS_UDP:
*/
  __pyx_t_1 = (__pyx_v_c_val == RDMA_PS_TCP);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":32
 *         return __pyx_enum.RDMA_PS_IPOIB
 *     elif c_val == rdma_port_space.RDMA_PS_TCP:
 *         return __pyx_enum.RDMA_PS_TCP             # <<<<<<<<<<<<<<
 *     elif c_val == rdma_port_space.RDMA_PS_UDP:
 *         return __pyx_enum.RDMA_PS_UDP
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_RDMA_PS_TCP); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 32, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":31
 *     elif c_val == rdma_port_space.RDMA_PS_IPOIB:
 *         return __pyx_enum.RDMA_PS_IPOIB
 *     elif c_val == rdma_port_space.RDMA_PS_TCP:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_PS_TCP
 *     elif c_val == rdma_port_space.RDMA_PS_UDP:
*/
  }

  /* "EnumTypeToPy":33
 *     elif c_val == rdma_port_space.RDMA_PS_TCP:
 *         return __pyx_enum.RDMA_PS_TCP
 *     elif c_val == rdma_port_space.RDMA_PS_UDP:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_PS_UDP
 *     elif c_val == rdma_port_space.RDMA_PS_IB:
*/
  __pyx_t_1 = (__pyx_v_c_val == RDMA_PS_UDP);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":34
 *         return __pyx_enum.RDMA_PS_TCP
 *     elif c_val == rdma_port_space.RDMA_PS_UDP:
 *         return __pyx_enum.RDMA_PS_UDP             # <<<<<<<<<<<<<<
 *     elif c_val == rdma_port_space.RDMA_PS_IB:
 *         return __pyx_enum.RDMA_PS_IB
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_RDMA_PS_UDP); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":33
 *     elif c_val == rdma_port_space.RDMA_PS_TCP:
 *         return __pyx_enum.RDMA_PS_TCP
 *     elif c_val == rdma_port_space.RDMA_PS_UDP:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_PS_UDP
 *     elif c_val == rdma_port_space.RDMA_PS_IB:
*/
  }

  /* "EnumTypeToPy":35
 *     elif c_val == rdma_port_space.RDMA_PS_UDP:
 *         return __pyx_enum.RDMA_PS_UDP
 *     elif c_val == rdma_port_space.RDMA_PS_IB:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_PS_IB
 *     else:
*/
  __pyx_t_1 = (__pyx_v_c_val == RDMA_PS_IB);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":36
 *         return __pyx_enum.RDMA_PS_UDP
 *     elif c_val == rdma_port_space.RDMA_PS_IB:
 *         return __pyx_enum.RDMA_PS_IB             # <<<<<<<<<<<<<<
 *     else:
 *         underlying_c_val = <int>c_val
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_RDMA_PS_IB); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":35
 *     elif c_val == rdma_port_space.RDMA_PS_UDP:
 *         return __pyx_enum.RDMA_PS_UDP
 *     elif c_val == rdma_port_space.RDMA_PS_IB:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_PS_IB
 *     else:
*/
  }

  /* "EnumTypeToPy":38
 *         return __pyx_enum.RDMA_PS_IB
 *     else:
 *         underlying_c_val = <int>c_val             # <<<<<<<<<<<<<<
 *         return __pyx_enum(underlying_c_val)
*/
  /*else*/ {
    __pyx_v_underlying_c_val = ((int)__pyx_v_c_val);

    /* "EnumTypeToPy":39
 *     else:
 *         underlying_c_val = <int>c_val
 *         return __pyx_enum(underlying_c_val)             # <<<<<<<<<<<<<<
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = NULL;
    __Pyx_INCREF(__pyx_v___pyx_enum);
    __pyx_t_5 = __pyx_v___pyx_enum; 
    __pyx_t_11 = __Pyx_PyLong_From_int(__pyx_v_underlying_c_val); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_13 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_8);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_13 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_11};
      __pyx_t_10 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 39, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
    }
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;
  }

  /* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py(rdma_port_space c_val):
 *     cdef object __pyx_enum
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_enum);
  __Pyx_XDECREF(__pyx_v_warnings);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py(enum ibv_send_flags __pyx_v_c_val) {
  PyObject *__pyx_v___pyx_enum = 0;
  PyObject *__pyx_v_warnings = NULL;
  int __pyx_v_underlying_c_val;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  size_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py", 0);

  /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_send_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py == Py_None);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_send_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py
 *         except ImportError:
*/
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "EnumTypeToPy":13
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py is None:
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_send_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py             # <<<<<<<<<<<<<<
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py = False  # False indicates "don't try again"
*/
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_send_flags};
          __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyverbs_libibverbs_enums, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L4_error)
        }
        __pyx_t_5 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_5);
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_send_flags};
          __pyx_t_7 = 0; {
            __pyx_t_8 = __Pyx_ImportFrom(__pyx_t_5, __pyx_imported_names[__pyx_t_7]); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 13, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_8);
            switch (__pyx_t_7) {
              case 0:
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py);
              __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py, __pyx_t_8);
              __Pyx_GIVEREF(__pyx_t_8);
              break;
              default:;
            }
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_send_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py
 *         except ImportError:
*/
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "EnumTypeToPy":14
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_send_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py
 *         except ImportError:             # <<<<<<<<<<<<<<
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py = False  # False indicates "don't try again"
 *             import warnings
*/
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_ImportError))));
      if (__pyx_t_9) {
        __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_8, &__pyx_t_10) < 0) __PYX_ERR(1, 14, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_10);

        /* "EnumTypeToPy":15
 *             from pyverbs.libibverbs_enums import ibv_send_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py = False  # False indicates "don't try again"             # <<<<<<<<<<<<<<
 *             import warnings
 *             warnings.warn(
*/
        __Pyx_INCREF(Py_False);
        __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py);
        __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py, Py_False);
        __Pyx_GIVEREF(Py_False);

        /* "EnumTypeToPy":16
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py = False  # False indicates "don't try again"
 *             import warnings             # <<<<<<<<<<<<<<
 *             warnings.warn(
 *                 f"enum class ibv_send_flags not importable from pyverbs.libibverbs_enums. "
*/
        __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_warnings, 0, 0, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 16, __pyx_L6_except_error)
        __pyx_t_11 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_v_warnings = __pyx_t_11;
        __pyx_t_11 = 0;

        /* "EnumTypeToPy":17
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py = False  # False indicates "don't try again"
 *             import warnings
 *             warnings.warn(             # <<<<<<<<<<<<<<
 *                 f"enum class ibv_send_flags not importable from pyverbs.libibverbs_enums. "
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
*/
        __pyx_t_12 = __pyx_v_warnings;
        __Pyx_INCREF(__pyx_t_12);
        __pyx_t_13 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_mstate_global->__pyx_kp_u_enum_class_ibv_send_flags_not_im};
          __pyx_t_11 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_warn, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 17, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_11);
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L5_exception_handled;
      }
      goto __pyx_L6_except_error;

      /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_send_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py
 *         except ImportError:
*/
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L5_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_L9_try_end:;
    }

    /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_send_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py
*/
  }

  /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py == Py_False);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":24
 * 
 * 
 *         return <int>c_val             # <<<<<<<<<<<<<<
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyLong_From_int(((int)__pyx_v_c_val)); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumTypeToPy":25
 * 
 *         return <int>c_val
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py             # <<<<<<<<<<<<<<
 * 
 *     if 0:
*/
  __Pyx_INCREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py);
  __pyx_v___pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py;

  /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_send_flags.IBV_SEND_FENCE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_SEND_FENCE
 *     elif c_val == ibv_send_flags.IBV_SEND_SIGNALED:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_SEND_FENCE);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":30
 *         pass
 *     elif c_val == ibv_send_flags.IBV_SEND_FENCE:
 *         return __pyx_enum.IBV_SEND_FENCE             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_send_flags.IBV_SEND_SIGNALED:
 *         return __pyx_enum.IBV_SEND_SIGNALED
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_SEND_FENCE); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_send_flags.IBV_SEND_FENCE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_SEND_FENCE
 *     elif c_val == ibv_send_flags.IBV_SEND_SIGNALED:
*/
  }

  /* "EnumTypeToPy":31
 *     elif c_val == ibv_send_flags.IBV_SEND_FENCE:
 *         return __pyx_enum.IBV_SEND_FENCE
 *     elif c_val == ibv_send_flags.IBV_SEND_SIGNALED:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_SEND_SIGNALED
 *     elif c_val == ibv_send_flags.IBV_SEND_SOLICITED:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_SEND_SIGNALED);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":32
 *         return __pyx_enum.IBV_SEND_FENCE
 *     elif c_val == ibv_send_flags.IBV_SEND_SIGNALED:
 *         return __pyx_enum.IBV_SEND_SIGNALED             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_send_flags.IBV_SEND_SOLICITED:
 *         return __pyx_enum.IBV_SEND_SOLICITED
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_SEND_SIGNALED); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 32, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":31
 *     elif c_val == ibv_send_flags.IBV_SEND_FENCE:
 *         return __pyx_enum.IBV_SEND_FENCE
 *     elif c_val == ibv_send_flags.IBV_SEND_SIGNALED:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_SEND_SIGNALED
 *     elif c_val == ibv_send_flags.IBV_SEND_SOLICITED:
*/
  }

  /* "EnumTypeToPy":33
 *     elif c_val == ibv_send_flags.IBV_SEND_SIGNALED:
 *         return __pyx_enum.IBV_SEND_SIGNALED
 *     elif c_val == ibv_send_flags.IBV_SEND_SOLICITED:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_SEND_SOLICITED
 *     elif c_val == ibv_send_flags.IBV_SEND_INLINE:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_SEND_SOLICITED);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":34
 *         return __pyx_enum.IBV_SEND_SIGNALED
 *     elif c_val == ibv_send_flags.IBV_SEND_SOLICITED:
 *         return __pyx_enum.IBV_SEND_SOLICITED             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_send_flags.IBV_SEND_INLINE:
 *         return __pyx_enum.IBV_SEND_INLINE
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_SEND_SOLICITED); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":33
 *     elif c_val == ibv_send_flags.IBV_SEND_SIGNALED:
 *         return __pyx_enum.IBV_SEND_SIGNALED
 *     elif c_val == ibv_send_flags.IBV_SEND_SOLICITED:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_SEND_SOLICITED
 *     elif c_val == ibv_send_flags.IBV_SEND_INLINE:
*/
  }

  /* "EnumTypeToPy":35
 *     elif c_val == ibv_send_flags.IBV_SEND_SOLICITED:
 *         return __pyx_enum.IBV_SEND_SOLICITED
 *     elif c_val == ibv_send_flags.IBV_SEND_INLINE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_SEND_INLINE
 *     elif c_val == ibv_send_flags.IBV_SEND_IP_CSUM:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_SEND_INLINE);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":36
 *         return __pyx_enum.IBV_SEND_SOLICITED
 *     elif c_val == ibv_send_flags.IBV_SEND_INLINE:
 *         return __pyx_enum.IBV_SEND_INLINE             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_send_flags.IBV_SEND_IP_CSUM:
 *         return __pyx_enum.IBV_SEND_IP_CSUM
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_SEND_INLINE); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":35
 *     elif c_val == ibv_send_flags.IBV_SEND_SOLICITED:
 *         return __pyx_enum.IBV_SEND_SOLICITED
 *     elif c_val == ibv_send_flags.IBV_SEND_INLINE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_SEND_INLINE
 *     elif c_val == ibv_send_flags.IBV_SEND_IP_CSUM:
*/
  }

  /* "EnumTypeToPy":37
 *     elif c_val == ibv_send_flags.IBV_SEND_INLINE:
 *         return __pyx_enum.IBV_SEND_INLINE
 *     elif c_val == ibv_send_flags.IBV_SEND_IP_CSUM:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_SEND_IP_CSUM
 *     else:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_SEND_IP_CSUM);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":38
 *         return __pyx_enum.IBV_SEND_INLINE
 *     elif c_val == ibv_send_flags.IBV_SEND_IP_CSUM:
 *         return __pyx_enum.IBV_SEND_IP_CSUM             # <<<<<<<<<<<<<<
 *     else:
 *         underlying_c_val = <int>c_val
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_SEND_IP_CSUM); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":37
 *     elif c_val == ibv_send_flags.IBV_SEND_INLINE:
 *         return __pyx_enum.IBV_SEND_INLINE
 *     elif c_val == ibv_send_flags.IBV_SEND_IP_CSUM:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_SEND_IP_CSUM
 *     else:
*/
  }

  /* "EnumTypeToPy":40
 *         return __pyx_enum.IBV_SEND_IP_CSUM
 *     else:
 *         underlying_c_val = <int>c_val             # <<<<<<<<<<<<<<
 *         return __pyx_enum(underlying_c_val)
*/
  /*else*/ {
    __pyx_v_underlying_c_val = ((int)__pyx_v_c_val);

    /* "EnumTypeToPy":41
 *     else:
 *         underlying_c_val = <int>c_val
 *         return __pyx_enum(underlying_c_val)             # <<<<<<<<<<<<<<
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = NULL;
    __Pyx_INCREF(__pyx_v___pyx_enum);
    __pyx_t_5 = __pyx_v___pyx_enum; 
    __pyx_t_11 = __Pyx_PyLong_From_int(__pyx_v_underlying_c_val); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_13 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_8);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_13 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_11};
      __pyx_t_10 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 41, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
    }
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;
  }

  /* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py(ibv_send_flags c_val):
 *     cdef object __pyx_enum
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_enum);
  __Pyx_XDECREF(__pyx_v_warnings);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py(enum rdma_cm_event_type __pyx_v_c_val) {
  PyObject *__pyx_v___pyx_enum = 0;
  PyObject *__pyx_v_warnings = NULL;
  int __pyx_v_underlying_c_val;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  size_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py", 0);

  /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.librdmacm_enums import rdma_cm_event_type as __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py == Py_None);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.librdmacm_enums import rdma_cm_event_type as __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py
 *         except ImportError:
*/
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "EnumTypeToPy":13
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py is None:
 *         try:
 *             from pyverbs.librdmacm_enums import rdma_cm_event_type as __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py             # <<<<<<<<<<<<<<
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py = False  # False indicates "don't try again"
*/
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_rdma_cm_event_type};
          __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyverbs_librdmacm_enums, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L4_error)
        }
        __pyx_t_5 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_5);
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_rdma_cm_event_type};
          __pyx_t_7 = 0; {
            __pyx_t_8 = __Pyx_ImportFrom(__pyx_t_5, __pyx_imported_names[__pyx_t_7]); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 13, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_8);
            switch (__pyx_t_7) {
              case 0:
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py);
              __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py, __pyx_t_8);
              __Pyx_GIVEREF(__pyx_t_8);
              break;
              default:;
            }
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.librdmacm_enums import rdma_cm_event_type as __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py
 *         except ImportError:
*/
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "EnumTypeToPy":14
 *         try:
 *             from pyverbs.librdmacm_enums import rdma_cm_event_type as __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py
 *         except ImportError:             # <<<<<<<<<<<<<<
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py = False  # False indicates "don't try again"
 *             import warnings
*/
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_ImportError))));
      if (__pyx_t_9) {
        __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_8, &__pyx_t_10) < 0) __PYX_ERR(1, 14, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_10);

        /* "EnumTypeToPy":15
 *             from pyverbs.librdmacm_enums import rdma_cm_event_type as __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py = False  # False indicates "don't try again"             # <<<<<<<<<<<<<<
 *             import warnings
 *             warnings.warn(
*/
        __Pyx_INCREF(Py_False);
        __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py);
        __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py, Py_False);
        __Pyx_GIVEREF(Py_False);

        /* "EnumTypeToPy":16
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py = False  # False indicates "don't try again"
 *             import warnings             # <<<<<<<<<<<<<<
 *             warnings.warn(
 *                 f"enum class rdma_cm_event_type not importable from pyverbs.librdmacm_enums. "
*/
        __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_warnings, 0, 0, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 16, __pyx_L6_except_error)
        __pyx_t_11 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_v_warnings = __pyx_t_11;
        __pyx_t_11 = 0;

        /* "EnumTypeToPy":17
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py = False  # False indicates "don't try again"
 *             import warnings
 *             warnings.warn(             # <<<<<<<<<<<<<<
 *                 f"enum class rdma_cm_event_type not importable from pyverbs.librdmacm_enums. "
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
*/
        __pyx_t_12 = __pyx_v_warnings;
        __Pyx_INCREF(__pyx_t_12);
        __pyx_t_13 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_mstate_global->__pyx_kp_u_enum_class_rdma_cm_event_type_no};
          __pyx_t_11 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_warn, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 17, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_11);
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L5_exception_handled;
      }
      goto __pyx_L6_except_error;

      /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.librdmacm_enums import rdma_cm_event_type as __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py
 *         except ImportError:
*/
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L5_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_L9_try_end:;
    }

    /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.librdmacm_enums import rdma_cm_event_type as __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py
*/
  }

  /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py == Py_False);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":24
 * 
 * 
 *         return <int>c_val             # <<<<<<<<<<<<<<
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyLong_From_int(((int)__pyx_v_c_val)); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumTypeToPy":25
 * 
 *         return <int>c_val
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py             # <<<<<<<<<<<<<<
 * 
 *     if 0:
*/
  __Pyx_INCREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py);
  __pyx_v___pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py;

  /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDR_RESOLVED:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_ADDR_RESOLVED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDR_ERROR:
*/
  __pyx_t_1 = (__pyx_v_c_val == RDMA_CM_EVENT_ADDR_RESOLVED);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":30
 *         pass
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDR_RESOLVED:
 *         return __pyx_enum.RDMA_CM_EVENT_ADDR_RESOLVED             # <<<<<<<<<<<<<<
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDR_ERROR:
 *         return __pyx_enum.RDMA_CM_EVENT_ADDR_ERROR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_RDMA_CM_EVENT_ADDR_RESOLVED); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDR_RESOLVED:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_ADDR_RESOLVED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDR_ERROR:
*/
  }

  /* "EnumTypeToPy":31
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDR_RESOLVED:
 *         return __pyx_enum.RDMA_CM_EVENT_ADDR_RESOLVED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDR_ERROR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_ADDR_ERROR
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ROUTE_RESOLVED:
*/
  __pyx_t_1 = (__pyx_v_c_val == RDMA_CM_EVENT_ADDR_ERROR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":32
 *         return __pyx_enum.RDMA_CM_EVENT_ADDR_RESOLVED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDR_ERROR:
 *         return __pyx_enum.RDMA_CM_EVENT_ADDR_ERROR             # <<<<<<<<<<<<<<
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ROUTE_RESOLVED:
 *         return __pyx_enum.RDMA_CM_EVENT_ROUTE_RESOLVED
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_RDMA_CM_EVENT_ADDR_ERROR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 32, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":31
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDR_RESOLVED:
 *         return __pyx_enum.RDMA_CM_EVENT_ADDR_RESOLVED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDR_ERROR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_ADDR_ERROR
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ROUTE_RESOLVED:
*/
  }

  /* "EnumTypeToPy":33
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDR_ERROR:
 *         return __pyx_enum.RDMA_CM_EVENT_ADDR_ERROR
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ROUTE_RESOLVED:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_ROUTE_RESOLVED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ROUTE_ERROR:
*/
  __pyx_t_1 = (__pyx_v_c_val == RDMA_CM_EVENT_ROUTE_RESOLVED);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":34
 *         return __pyx_enum.RDMA_CM_EVENT_ADDR_ERROR
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ROUTE_RESOLVED:
 *         return __pyx_enum.RDMA_CM_EVENT_ROUTE_RESOLVED             # <<<<<<<<<<<<<<
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ROUTE_ERROR:
 *         return __pyx_enum.RDMA_CM_EVENT_ROUTE_ERROR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_RDMA_CM_EVENT_ROUTE_RESOLVED); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":33
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDR_ERROR:
 *         return __pyx_enum.RDMA_CM_EVENT_ADDR_ERROR
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ROUTE_RESOLVED:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_ROUTE_RESOLVED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ROUTE_ERROR:
*/
  }

  /* "EnumTypeToPy":35
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ROUTE_RESOLVED:
 *         return __pyx_enum.RDMA_CM_EVENT_ROUTE_RESOLVED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ROUTE_ERROR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_ROUTE_ERROR
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_REQUEST:
*/
  __pyx_t_1 = (__pyx_v_c_val == RDMA_CM_EVENT_ROUTE_ERROR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":36
 *         return __pyx_enum.RDMA_CM_EVENT_ROUTE_RESOLVED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ROUTE_ERROR:
 *         return __pyx_enum.RDMA_CM_EVENT_ROUTE_ERROR             # <<<<<<<<<<<<<<
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_REQUEST:
 *         return __pyx_enum.RDMA_CM_EVENT_CONNECT_REQUEST
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_RDMA_CM_EVENT_ROUTE_ERROR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":35
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ROUTE_RESOLVED:
 *         return __pyx_enum.RDMA_CM_EVENT_ROUTE_RESOLVED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ROUTE_ERROR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_ROUTE_ERROR
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_REQUEST:
*/
  }

  /* "EnumTypeToPy":37
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ROUTE_ERROR:
 *         return __pyx_enum.RDMA_CM_EVENT_ROUTE_ERROR
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_REQUEST:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_CONNECT_REQUEST
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_RESPONSE:
*/
  __pyx_t_1 = (__pyx_v_c_val == RDMA_CM_EVENT_CONNECT_REQUEST);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":38
 *         return __pyx_enum.RDMA_CM_EVENT_ROUTE_ERROR
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_REQUEST:
 *         return __pyx_enum.RDMA_CM_EVENT_CONNECT_REQUEST             # <<<<<<<<<<<<<<
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_RESPONSE:
 *         return __pyx_enum.RDMA_CM_EVENT_CONNECT_RESPONSE
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_RDMA_CM_EVENT_CONNECT_REQUEST); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":37
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ROUTE_ERROR:
 *         return __pyx_enum.RDMA_CM_EVENT_ROUTE_ERROR
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_REQUEST:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_CONNECT_REQUEST
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_RESPONSE:
*/
  }

  /* "EnumTypeToPy":39
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_REQUEST:
 *         return __pyx_enum.RDMA_CM_EVENT_CONNECT_REQUEST
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_RESPONSE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_CONNECT_RESPONSE
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_ERROR:
*/
  __pyx_t_1 = (__pyx_v_c_val == RDMA_CM_EVENT_CONNECT_RESPONSE);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":40
 *         return __pyx_enum.RDMA_CM_EVENT_CONNECT_REQUEST
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_RESPONSE:
 *         return __pyx_enum.RDMA_CM_EVENT_CONNECT_RESPONSE             # <<<<<<<<<<<<<<
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_ERROR:
 *         return __pyx_enum.RDMA_CM_EVENT_CONNECT_ERROR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_RDMA_CM_EVENT_CONNECT_RESPONSE); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":39
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_REQUEST:
 *         return __pyx_enum.RDMA_CM_EVENT_CONNECT_REQUEST
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_RESPONSE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_CONNECT_RESPONSE
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_ERROR:
*/
  }

  /* "EnumTypeToPy":41
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_RESPONSE:
 *         return __pyx_enum.RDMA_CM_EVENT_CONNECT_RESPONSE
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_ERROR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_CONNECT_ERROR
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_UNREACHABLE:
*/
  __pyx_t_1 = (__pyx_v_c_val == RDMA_CM_EVENT_CONNECT_ERROR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":42
 *         return __pyx_enum.RDMA_CM_EVENT_CONNECT_RESPONSE
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_ERROR:
 *         return __pyx_enum.RDMA_CM_EVENT_CONNECT_ERROR             # <<<<<<<<<<<<<<
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_UNREACHABLE:
 *         return __pyx_enum.RDMA_CM_EVENT_UNREACHABLE
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_RDMA_CM_EVENT_CONNECT_ERROR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":41
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_RESPONSE:
 *         return __pyx_enum.RDMA_CM_EVENT_CONNECT_RESPONSE
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_ERROR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_CONNECT_ERROR
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_UNREACHABLE:
*/
  }

  /* "EnumTypeToPy":43
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_ERROR:
 *         return __pyx_enum.RDMA_CM_EVENT_CONNECT_ERROR
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_UNREACHABLE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_UNREACHABLE
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_REJECTED:
*/
  __pyx_t_1 = (__pyx_v_c_val == RDMA_CM_EVENT_UNREACHABLE);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":44
 *         return __pyx_enum.RDMA_CM_EVENT_CONNECT_ERROR
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_UNREACHABLE:
 *         return __pyx_enum.RDMA_CM_EVENT_UNREACHABLE             # <<<<<<<<<<<<<<
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_REJECTED:
 *         return __pyx_enum.RDMA_CM_EVENT_REJECTED
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_RDMA_CM_EVENT_UNREACHABLE); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":43
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_CONNECT_ERROR:
 *         return __pyx_enum.RDMA_CM_EVENT_CONNECT_ERROR
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_UNREACHABLE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_UNREACHABLE
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_REJECTED:
*/
  }

  /* "EnumTypeToPy":45
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_UNREACHABLE:
 *         return __pyx_enum.RDMA_CM_EVENT_UNREACHABLE
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_REJECTED:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_REJECTED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ESTABLISHED:
*/
  __pyx_t_1 = (__pyx_v_c_val == RDMA_CM_EVENT_REJECTED);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":46
 *         return __pyx_enum.RDMA_CM_EVENT_UNREACHABLE
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_REJECTED:
 *         return __pyx_enum.RDMA_CM_EVENT_REJECTED             # <<<<<<<<<<<<<<
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ESTABLISHED:
 *         return __pyx_enum.RDMA_CM_EVENT_ESTABLISHED
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_RDMA_CM_EVENT_REJECTED); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 46, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":45
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_UNREACHABLE:
 *         return __pyx_enum.RDMA_CM_EVENT_UNREACHABLE
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_REJECTED:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_REJECTED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ESTABLISHED:
*/
  }

  /* "EnumTypeToPy":47
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_REJECTED:
 *         return __pyx_enum.RDMA_CM_EVENT_REJECTED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ESTABLISHED:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_ESTABLISHED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_DISCONNECTED:
*/
  __pyx_t_1 = (__pyx_v_c_val == RDMA_CM_EVENT_ESTABLISHED);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":48
 *         return __pyx_enum.RDMA_CM_EVENT_REJECTED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ESTABLISHED:
 *         return __pyx_enum.RDMA_CM_EVENT_ESTABLISHED             # <<<<<<<<<<<<<<
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_DISCONNECTED:
 *         return __pyx_enum.RDMA_CM_EVENT_DISCONNECTED
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_RDMA_CM_EVENT_ESTABLISHED); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":47
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_REJECTED:
 *         return __pyx_enum.RDMA_CM_EVENT_REJECTED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ESTABLISHED:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_ESTABLISHED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_DISCONNECTED:
*/
  }

  /* "EnumTypeToPy":49
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ESTABLISHED:
 *         return __pyx_enum.RDMA_CM_EVENT_ESTABLISHED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_DISCONNECTED:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_DISCONNECTED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_DEVICE_REMOVAL:
*/
  __pyx_t_1 = (__pyx_v_c_val == RDMA_CM_EVENT_DISCONNECTED);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":50
 *         return __pyx_enum.RDMA_CM_EVENT_ESTABLISHED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_DISCONNECTED:
 *         return __pyx_enum.RDMA_CM_EVENT_DISCONNECTED             # <<<<<<<<<<<<<<
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_DEVICE_REMOVAL:
 *         return __pyx_enum.RDMA_CM_EVENT_DEVICE_REMOVAL
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_RDMA_CM_EVENT_DISCONNECTED); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":49
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ESTABLISHED:
 *         return __pyx_enum.RDMA_CM_EVENT_ESTABLISHED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_DISCONNECTED:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_DISCONNECTED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_DEVICE_REMOVAL:
*/
  }

  /* "EnumTypeToPy":51
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_DISCONNECTED:
 *         return __pyx_enum.RDMA_CM_EVENT_DISCONNECTED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_DEVICE_REMOVAL:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_DEVICE_REMOVAL
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_MULTICAST_JOIN:
*/
  __pyx_t_1 = (__pyx_v_c_val == RDMA_CM_EVENT_DEVICE_REMOVAL);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":52
 *         return __pyx_enum.RDMA_CM_EVENT_DISCONNECTED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_DEVICE_REMOVAL:
 *         return __pyx_enum.RDMA_CM_EVENT_DEVICE_REMOVAL             # <<<<<<<<<<<<<<
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_MULTICAST_JOIN:
 *         return __pyx_enum.RDMA_CM_EVENT_MULTICAST_JOIN
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_RDMA_CM_EVENT_DEVICE_REMOVAL); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":51
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_DISCONNECTED:
 *         return __pyx_enum.RDMA_CM_EVENT_DISCONNECTED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_DEVICE_REMOVAL:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_DEVICE_REMOVAL
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_MULTICAST_JOIN:
*/
  }

  /* "EnumTypeToPy":53
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_DEVICE_REMOVAL:
 *         return __pyx_enum.RDMA_CM_EVENT_DEVICE_REMOVAL
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_MULTICAST_JOIN:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_MULTICAST_JOIN
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_MULTICAST_ERROR:
*/
  __pyx_t_1 = (__pyx_v_c_val == RDMA_CM_EVENT_MULTICAST_JOIN);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":54
 *         return __pyx_enum.RDMA_CM_EVENT_DEVICE_REMOVAL
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_MULTICAST_JOIN:
 *         return __pyx_enum.RDMA_CM_EVENT_MULTICAST_JOIN             # <<<<<<<<<<<<<<
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_MULTICAST_ERROR:
 *         return __pyx_enum.RDMA_CM_EVENT_MULTICAST_ERROR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_RDMA_CM_EVENT_MULTICAST_JOIN); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":53
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_DEVICE_REMOVAL:
 *         return __pyx_enum.RDMA_CM_EVENT_DEVICE_REMOVAL
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_MULTICAST_JOIN:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_MULTICAST_JOIN
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_MULTICAST_ERROR:
*/
  }

  /* "EnumTypeToPy":55
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_MULTICAST_JOIN:
 *         return __pyx_enum.RDMA_CM_EVENT_MULTICAST_JOIN
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_MULTICAST_ERROR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_MULTICAST_ERROR
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDR_CHANGE:
*/
  __pyx_t_1 = (__pyx_v_c_val == RDMA_CM_EVENT_MULTICAST_ERROR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":56
 *         return __pyx_enum.RDMA_CM_EVENT_MULTICAST_JOIN
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_MULTICAST_ERROR:
 *         return __pyx_enum.RDMA_CM_EVENT_MULTICAST_ERROR             # <<<<<<<<<<<<<<
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDR_CHANGE:
 *         return __pyx_enum.RDMA_CM_EVENT_ADDR_CHANGE
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_RDMA_CM_EVENT_MULTICAST_ERROR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":55
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_MULTICAST_JOIN:
 *         return __pyx_enum.RDMA_CM_EVENT_MULTICAST_JOIN
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_MULTICAST_ERROR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_MULTICAST_ERROR
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDR_CHANGE:
*/
  }

  /* "EnumTypeToPy":57
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_MULTICAST_ERROR:
 *         return __pyx_enum.RDMA_CM_EVENT_MULTICAST_ERROR
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDR_CHANGE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_ADDR_CHANGE
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_TIMEWAIT_EXIT:
*/
  __pyx_t_1 = (__pyx_v_c_val == RDMA_CM_EVENT_ADDR_CHANGE);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":58
 *         return __pyx_enum.RDMA_CM_EVENT_MULTICAST_ERROR
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDR_CHANGE:
 *         return __pyx_enum.RDMA_CM_EVENT_ADDR_CHANGE             # <<<<<<<<<<<<<<
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_TIMEWAIT_EXIT:
 *         return __pyx_enum.RDMA_CM_EVENT_TIMEWAIT_EXIT
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_RDMA_CM_EVENT_ADDR_CHANGE); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":57
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_MULTICAST_ERROR:
 *         return __pyx_enum.RDMA_CM_EVENT_MULTICAST_ERROR
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDR_CHANGE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_ADDR_CHANGE
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_TIMEWAIT_EXIT:
*/
  }

  /* "EnumTypeToPy":59
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDR_CHANGE:
 *         return __pyx_enum.RDMA_CM_EVENT_ADDR_CHANGE
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_TIMEWAIT_EXIT:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_TIMEWAIT_EXIT
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDRINFO_RESOLVED:
*/
  __pyx_t_1 = (__pyx_v_c_val == RDMA_CM_EVENT_TIMEWAIT_EXIT);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":60
 *         return __pyx_enum.RDMA_CM_EVENT_ADDR_CHANGE
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_TIMEWAIT_EXIT:
 *         return __pyx_enum.RDMA_CM_EVENT_TIMEWAIT_EXIT             # <<<<<<<<<<<<<<
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDRINFO_RESOLVED:
 *         return __pyx_enum.RDMA_CM_EVENT_ADDRINFO_RESOLVED
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_RDMA_CM_EVENT_TIMEWAIT_EXIT); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":59
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDR_CHANGE:
 *         return __pyx_enum.RDMA_CM_EVENT_ADDR_CHANGE
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_TIMEWAIT_EXIT:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_TIMEWAIT_EXIT
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDRINFO_RESOLVED:
*/
  }

  /* "EnumTypeToPy":61
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_TIMEWAIT_EXIT:
 *         return __pyx_enum.RDMA_CM_EVENT_TIMEWAIT_EXIT
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDRINFO_RESOLVED:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_ADDRINFO_RESOLVED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDRINFO_ERROR:
*/
  __pyx_t_1 = (__pyx_v_c_val == RDMA_CM_EVENT_ADDRINFO_RESOLVED);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":62
 *         return __pyx_enum.RDMA_CM_EVENT_TIMEWAIT_EXIT
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDRINFO_RESOLVED:
 *         return __pyx_enum.RDMA_CM_EVENT_ADDRINFO_RESOLVED             # <<<<<<<<<<<<<<
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDRINFO_ERROR:
 *         return __pyx_enum.RDMA_CM_EVENT_ADDRINFO_ERROR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_RDMA_CM_EVENT_ADDRINFO_RESOLVED); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":61
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_TIMEWAIT_EXIT:
 *         return __pyx_enum.RDMA_CM_EVENT_TIMEWAIT_EXIT
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDRINFO_RESOLVED:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_ADDRINFO_RESOLVED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDRINFO_ERROR:
*/
  }

  /* "EnumTypeToPy":63
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDRINFO_RESOLVED:
 *         return __pyx_enum.RDMA_CM_EVENT_ADDRINFO_RESOLVED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDRINFO_ERROR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_ADDRINFO_ERROR
 *     else:
*/
  __pyx_t_1 = (__pyx_v_c_val == RDMA_CM_EVENT_ADDRINFO_ERROR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":64
 *         return __pyx_enum.RDMA_CM_EVENT_ADDRINFO_RESOLVED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDRINFO_ERROR:
 *         return __pyx_enum.RDMA_CM_EVENT_ADDRINFO_ERROR             # <<<<<<<<<<<<<<
 *     else:
 *         underlying_c_val = <int>c_val
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_RDMA_CM_EVENT_ADDRINFO_ERROR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":63
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDRINFO_RESOLVED:
 *         return __pyx_enum.RDMA_CM_EVENT_ADDRINFO_RESOLVED
 *     elif c_val == rdma_cm_event_type.RDMA_CM_EVENT_ADDRINFO_ERROR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.RDMA_CM_EVENT_ADDRINFO_ERROR
 *     else:
*/
  }

  /* "EnumTypeToPy":66
 *         return __pyx_enum.RDMA_CM_EVENT_ADDRINFO_ERROR
 *     else:
 *         underlying_c_val = <int>c_val             # <<<<<<<<<<<<<<
 *         return __pyx_enum(underlying_c_val)
*/
  /*else*/ {
    __pyx_v_underlying_c_val = ((int)__pyx_v_c_val);

    /* "EnumTypeToPy":67
 *     else:
 *         underlying_c_val = <int>c_val
 *         return __pyx_enum(underlying_c_val)             # <<<<<<<<<<<<<<
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = NULL;
    __Pyx_INCREF(__pyx_v___pyx_enum);
    __pyx_t_5 = __pyx_v___pyx_enum; 
    __pyx_t_11 = __Pyx_PyLong_From_int(__pyx_v_underlying_c_val); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_13 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_8);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_13 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_11};
      __pyx_t_10 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 67, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
    }
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;
  }

  /* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py(rdma_cm_event_type c_val):
 *     cdef object __pyx_enum
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_enum);
  __Pyx_XDECREF(__pyx_v_warnings);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py(enum ibv_access_flags __pyx_v_c_val) {
  PyObject *__pyx_v___pyx_enum = 0;
  PyObject *__pyx_v_warnings = NULL;
  int __pyx_v_underlying_c_val;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  size_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py", 0);

  /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_access_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py == Py_None);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_access_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py
 *         except ImportError:
*/
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "EnumTypeToPy":13
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py is None:
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_access_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py             # <<<<<<<<<<<<<<
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py = False  # False indicates "don't try again"
*/
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_access_flags};
          __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyverbs_libibverbs_enums, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L4_error)
        }
        __pyx_t_5 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_5);
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_access_flags};
          __pyx_t_7 = 0; {
            __pyx_t_8 = __Pyx_ImportFrom(__pyx_t_5, __pyx_imported_names[__pyx_t_7]); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 13, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_8);
            switch (__pyx_t_7) {
              case 0:
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py);
              __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py, __pyx_t_8);
              __Pyx_GIVEREF(__pyx_t_8);
              break;
              default:;
            }
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_access_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py
 *         except ImportError:
*/
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "EnumTypeToPy":14
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_access_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py
 *         except ImportError:             # <<<<<<<<<<<<<<
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py = False  # False indicates "don't try again"
 *             import warnings
*/
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_ImportError))));
      if (__pyx_t_9) {
        __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_8, &__pyx_t_10) < 0) __PYX_ERR(1, 14, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_10);

        /* "EnumTypeToPy":15
 *             from pyverbs.libibverbs_enums import ibv_access_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py = False  # False indicates "don't try again"             # <<<<<<<<<<<<<<
 *             import warnings
 *             warnings.warn(
*/
        __Pyx_INCREF(Py_False);
        __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py);
        __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py, Py_False);
        __Pyx_GIVEREF(Py_False);

        /* "EnumTypeToPy":16
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py = False  # False indicates "don't try again"
 *             import warnings             # <<<<<<<<<<<<<<
 *             warnings.warn(
 *                 f"enum class ibv_access_flags not importable from pyverbs.libibverbs_enums. "
*/
        __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_warnings, 0, 0, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 16, __pyx_L6_except_error)
        __pyx_t_11 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_v_warnings = __pyx_t_11;
        __pyx_t_11 = 0;

        /* "EnumTypeToPy":17
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py = False  # False indicates "don't try again"
 *             import warnings
 *             warnings.warn(             # <<<<<<<<<<<<<<
 *                 f"enum class ibv_access_flags not importable from pyverbs.libibverbs_enums. "
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
*/
        __pyx_t_12 = __pyx_v_warnings;
        __Pyx_INCREF(__pyx_t_12);
        __pyx_t_13 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_mstate_global->__pyx_kp_u_enum_class_ibv_access_flags_not};
          __pyx_t_11 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_warn, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 17, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_11);
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L5_exception_handled;
      }
      goto __pyx_L6_except_error;

      /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_access_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py
 *         except ImportError:
*/
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L5_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_L9_try_end:;
    }

    /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_access_flags as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py
*/
  }

  /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py == Py_False);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":24
 * 
 * 
 *         return <int>c_val             # <<<<<<<<<<<<<<
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyLong_From_int(((int)__pyx_v_c_val)); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumTypeToPy":25
 * 
 *         return <int>c_val
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py             # <<<<<<<<<<<<<<
 * 
 *     if 0:
*/
  __Pyx_INCREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py);
  __pyx_v___pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py;

  /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_access_flags.IBV_ACCESS_LOCAL_WRITE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_ACCESS_LOCAL_WRITE
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_WRITE:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_ACCESS_LOCAL_WRITE);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":30
 *         pass
 *     elif c_val == ibv_access_flags.IBV_ACCESS_LOCAL_WRITE:
 *         return __pyx_enum.IBV_ACCESS_LOCAL_WRITE             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_WRITE:
 *         return __pyx_enum.IBV_ACCESS_REMOTE_WRITE
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_ACCESS_LOCAL_WRITE); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_access_flags.IBV_ACCESS_LOCAL_WRITE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_ACCESS_LOCAL_WRITE
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_WRITE:
*/
  }

  /* "EnumTypeToPy":31
 *     elif c_val == ibv_access_flags.IBV_ACCESS_LOCAL_WRITE:
 *         return __pyx_enum.IBV_ACCESS_LOCAL_WRITE
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_WRITE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_ACCESS_REMOTE_WRITE
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_READ:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_ACCESS_REMOTE_WRITE);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":32
 *         return __pyx_enum.IBV_ACCESS_LOCAL_WRITE
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_WRITE:
 *         return __pyx_enum.IBV_ACCESS_REMOTE_WRITE             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_READ:
 *         return __pyx_enum.IBV_ACCESS_REMOTE_READ
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_ACCESS_REMOTE_WRITE); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 32, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":31
 *     elif c_val == ibv_access_flags.IBV_ACCESS_LOCAL_WRITE:
 *         return __pyx_enum.IBV_ACCESS_LOCAL_WRITE
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_WRITE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_ACCESS_REMOTE_WRITE
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_READ:
*/
  }

  /* "EnumTypeToPy":33
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_WRITE:
 *         return __pyx_enum.IBV_ACCESS_REMOTE_WRITE
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_READ:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_ACCESS_REMOTE_READ
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_ATOMIC:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_ACCESS_REMOTE_READ);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":34
 *         return __pyx_enum.IBV_ACCESS_REMOTE_WRITE
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_READ:
 *         return __pyx_enum.IBV_ACCESS_REMOTE_READ             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_ATOMIC:
 *         return __pyx_enum.IBV_ACCESS_REMOTE_ATOMIC
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_ACCESS_REMOTE_READ); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":33
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_WRITE:
 *         return __pyx_enum.IBV_ACCESS_REMOTE_WRITE
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_READ:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_ACCESS_REMOTE_READ
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_ATOMIC:
*/
  }

  /* "EnumTypeToPy":35
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_READ:
 *         return __pyx_enum.IBV_ACCESS_REMOTE_READ
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_ATOMIC:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_ACCESS_REMOTE_ATOMIC
 *     elif c_val == ibv_access_flags.IBV_ACCESS_MW_BIND:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_ACCESS_REMOTE_ATOMIC);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":36
 *         return __pyx_enum.IBV_ACCESS_REMOTE_READ
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_ATOMIC:
 *         return __pyx_enum.IBV_ACCESS_REMOTE_ATOMIC             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_access_flags.IBV_ACCESS_MW_BIND:
 *         return __pyx_enum.IBV_ACCESS_MW_BIND
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_ACCESS_REMOTE_ATOMIC); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":35
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_READ:
 *         return __pyx_enum.IBV_ACCESS_REMOTE_READ
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_ATOMIC:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_ACCESS_REMOTE_ATOMIC
 *     elif c_val == ibv_access_flags.IBV_ACCESS_MW_BIND:
*/
  }

  /* "EnumTypeToPy":37
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_ATOMIC:
 *         return __pyx_enum.IBV_ACCESS_REMOTE_ATOMIC
 *     elif c_val == ibv_access_flags.IBV_ACCESS_MW_BIND:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_ACCESS_MW_BIND
 *     elif c_val == ibv_access_flags.IBV_ACCESS_ZERO_BASED:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_ACCESS_MW_BIND);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":38
 *         return __pyx_enum.IBV_ACCESS_REMOTE_ATOMIC
 *     elif c_val == ibv_access_flags.IBV_ACCESS_MW_BIND:
 *         return __pyx_enum.IBV_ACCESS_MW_BIND             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_access_flags.IBV_ACCESS_ZERO_BASED:
 *         return __pyx_enum.IBV_ACCESS_ZERO_BASED
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_ACCESS_MW_BIND); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":37
 *     elif c_val == ibv_access_flags.IBV_ACCESS_REMOTE_ATOMIC:
 *         return __pyx_enum.IBV_ACCESS_REMOTE_ATOMIC
 *     elif c_val == ibv_access_flags.IBV_ACCESS_MW_BIND:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_ACCESS_MW_BIND
 *     elif c_val == ibv_access_flags.IBV_ACCESS_ZERO_BASED:
*/
  }

  /* "EnumTypeToPy":39
 *     elif c_val == ibv_access_flags.IBV_ACCESS_MW_BIND:
 *         return __pyx_enum.IBV_ACCESS_MW_BIND
 *     elif c_val == ibv_access_flags.IBV_ACCESS_ZERO_BASED:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_ACCESS_ZERO_BASED
 *     elif c_val == ibv_access_flags.IBV_ACCESS_ON_DEMAND:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_ACCESS_ZERO_BASED);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":40
 *         return __pyx_enum.IBV_ACCESS_MW_BIND
 *     elif c_val == ibv_access_flags.IBV_ACCESS_ZERO_BASED:
 *         return __pyx_enum.IBV_ACCESS_ZERO_BASED             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_access_flags.IBV_ACCESS_ON_DEMAND:
 *         return __pyx_enum.IBV_ACCESS_ON_DEMAND
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_ACCESS_ZERO_BASED); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":39
 *     elif c_val == ibv_access_flags.IBV_ACCESS_MW_BIND:
 *         return __pyx_enum.IBV_ACCESS_MW_BIND
 *     elif c_val == ibv_access_flags.IBV_ACCESS_ZERO_BASED:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_ACCESS_ZERO_BASED
 *     elif c_val == ibv_access_flags.IBV_ACCESS_ON_DEMAND:
*/
  }

  /* "EnumTypeToPy":41
 *     elif c_val == ibv_access_flags.IBV_ACCESS_ZERO_BASED:
 *         return __pyx_enum.IBV_ACCESS_ZERO_BASED
 *     elif c_val == ibv_access_flags.IBV_ACCESS_ON_DEMAND:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_ACCESS_ON_DEMAND
 *     elif c_val == ibv_access_flags.IBV_ACCESS_HUGETLB:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_ACCESS_ON_DEMAND);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":42
 *         return __pyx_enum.IBV_ACCESS_ZERO_BASED
 *     elif c_val == ibv_access_flags.IBV_ACCESS_ON_DEMAND:
 *         return __pyx_enum.IBV_ACCESS_ON_DEMAND             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_access_flags.IBV_ACCESS_HUGETLB:
 *         return __pyx_enum.IBV_ACCESS_HUGETLB
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_ACCESS_ON_DEMAND); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":41
 *     elif c_val == ibv_access_flags.IBV_ACCESS_ZERO_BASED:
 *         return __pyx_enum.IBV_ACCESS_ZERO_BASED
 *     elif c_val == ibv_access_flags.IBV_ACCESS_ON_DEMAND:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_ACCESS_ON_DEMAND
 *     elif c_val == ibv_access_flags.IBV_ACCESS_HUGETLB:
*/
  }

  /* "EnumTypeToPy":43
 *     elif c_val == ibv_access_flags.IBV_ACCESS_ON_DEMAND:
 *         return __pyx_enum.IBV_ACCESS_ON_DEMAND
 *     elif c_val == ibv_access_flags.IBV_ACCESS_HUGETLB:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_ACCESS_HUGETLB
 *     elif c_val == ibv_access_flags.IBV_ACCESS_FLUSH_GLOBAL:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_ACCESS_HUGETLB);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":44
 *         return __pyx_enum.IBV_ACCESS_ON_DEMAND
 *     elif c_val == ibv_access_flags.IBV_ACCESS_HUGETLB:
 *         return __pyx_enum.IBV_ACCESS_HUGETLB             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_access_flags.IBV_ACCESS_FLUSH_GLOBAL:
 *         return __pyx_enum.IBV_ACCESS_FLUSH_GLOBAL
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_ACCESS_HUGETLB); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":43
 *     elif c_val == ibv_access_flags.IBV_ACCESS_ON_DEMAND:
 *         return __pyx_enum.IBV_ACCESS_ON_DEMAND
 *     elif c_val == ibv_access_flags.IBV_ACCESS_HUGETLB:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_ACCESS_HUGETLB
 *     elif c_val == ibv_access_flags.IBV_ACCESS_FLUSH_GLOBAL:
*/
  }

  /* "EnumTypeToPy":45
 *     elif c_val == ibv_access_flags.IBV_ACCESS_HUGETLB:
 *         return __pyx_enum.IBV_ACCESS_HUGETLB
 *     elif c_val == ibv_access_flags.IBV_ACCESS_FLUSH_GLOBAL:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_ACCESS_FLUSH_GLOBAL
 *     elif c_val == ibv_access_flags.IBV_ACCESS_FLUSH_PERSISTENT:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_ACCESS_FLUSH_GLOBAL);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":46
 *         return __pyx_enum.IBV_ACCESS_HUGETLB
 *     elif c_val == ibv_access_flags.IBV_ACCESS_FLUSH_GLOBAL:
 *         return __pyx_enum.IBV_ACCESS_FLUSH_GLOBAL             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_access_flags.IBV_ACCESS_FLUSH_PERSISTENT:
 *         return __pyx_enum.IBV_ACCESS_FLUSH_PERSISTENT
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_ACCESS_FLUSH_GLOBAL); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 46, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":45
 *     elif c_val == ibv_access_flags.IBV_ACCESS_HUGETLB:
 *         return __pyx_enum.IBV_ACCESS_HUGETLB
 *     elif c_val == ibv_access_flags.IBV_ACCESS_FLUSH_GLOBAL:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_ACCESS_FLUSH_GLOBAL
 *     elif c_val == ibv_access_flags.IBV_ACCESS_FLUSH_PERSISTENT:
*/
  }

  /* "EnumTypeToPy":47
 *     elif c_val == ibv_access_flags.IBV_ACCESS_FLUSH_GLOBAL:
 *         return __pyx_enum.IBV_ACCESS_FLUSH_GLOBAL
 *     elif c_val == ibv_access_flags.IBV_ACCESS_FLUSH_PERSISTENT:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_ACCESS_FLUSH_PERSISTENT
 *     elif c_val == ibv_access_flags.IBV_ACCESS_RELAXED_ORDERING:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_ACCESS_FLUSH_PERSISTENT);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":48
 *         return __pyx_enum.IBV_ACCESS_FLUSH_GLOBAL
 *     elif c_val == ibv_access_flags.IBV_ACCESS_FLUSH_PERSISTENT:
 *         return __pyx_enum.IBV_ACCESS_FLUSH_PERSISTENT             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_access_flags.IBV_ACCESS_RELAXED_ORDERING:
 *         return __pyx_enum.IBV_ACCESS_RELAXED_ORDERING
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_ACCESS_FLUSH_PERSISTENT); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":47
 *     elif c_val == ibv_access_flags.IBV_ACCESS_FLUSH_GLOBAL:
 *         return __pyx_enum.IBV_ACCESS_FLUSH_GLOBAL
 *     elif c_val == ibv_access_flags.IBV_ACCESS_FLUSH_PERSISTENT:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_ACCESS_FLUSH_PERSISTENT
 *     elif c_val == ibv_access_flags.IBV_ACCESS_RELAXED_ORDERING:
*/
  }

  /* "EnumTypeToPy":49
 *     elif c_val == ibv_access_flags.IBV_ACCESS_FLUSH_PERSISTENT:
 *         return __pyx_enum.IBV_ACCESS_FLUSH_PERSISTENT
 *     elif c_val == ibv_access_flags.IBV_ACCESS_RELAXED_ORDERING:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_ACCESS_RELAXED_ORDERING
 *     else:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_ACCESS_RELAXED_ORDERING);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":50
 *         return __pyx_enum.IBV_ACCESS_FLUSH_PERSISTENT
 *     elif c_val == ibv_access_flags.IBV_ACCESS_RELAXED_ORDERING:
 *         return __pyx_enum.IBV_ACCESS_RELAXED_ORDERING             # <<<<<<<<<<<<<<
 *     else:
 *         underlying_c_val = <int>c_val
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_ACCESS_RELAXED_ORDERING); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":49
 *     elif c_val == ibv_access_flags.IBV_ACCESS_FLUSH_PERSISTENT:
 *         return __pyx_enum.IBV_ACCESS_FLUSH_PERSISTENT
 *     elif c_val == ibv_access_flags.IBV_ACCESS_RELAXED_ORDERING:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_ACCESS_RELAXED_ORDERING
 *     else:
*/
  }

  /* "EnumTypeToPy":52
 *         return __pyx_enum.IBV_ACCESS_RELAXED_ORDERING
 *     else:
 *         underlying_c_val = <int>c_val             # <<<<<<<<<<<<<<
 *         return __pyx_enum(underlying_c_val)
*/
  /*else*/ {
    __pyx_v_underlying_c_val = ((int)__pyx_v_c_val);

    /* "EnumTypeToPy":53
 *     else:
 *         underlying_c_val = <int>c_val
 *         return __pyx_enum(underlying_c_val)             # <<<<<<<<<<<<<<
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = NULL;
    __Pyx_INCREF(__pyx_v___pyx_enum);
    __pyx_t_5 = __pyx_v___pyx_enum; 
    __pyx_t_11 = __Pyx_PyLong_From_int(__pyx_v_underlying_c_val); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_13 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_8);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_13 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_11};
      __pyx_t_10 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 53, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
    }
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;
  }

  /* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py(ibv_access_flags c_val):
 *     cdef object __pyx_enum
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_enum);
  __Pyx_XDECREF(__pyx_v_warnings);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(enum ibv_wc_status __pyx_v_c_val) {
  PyObject *__pyx_v___pyx_enum = 0;
  PyObject *__pyx_v_warnings = NULL;
  int __pyx_v_underlying_c_val;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  size_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py", 0);

  /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_wc_status as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py == Py_None);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_wc_status as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py
 *         except ImportError:
*/
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "EnumTypeToPy":13
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py is None:
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_wc_status as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py             # <<<<<<<<<<<<<<
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py = False  # False indicates "don't try again"
*/
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_wc_status};
          __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyverbs_libibverbs_enums, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L4_error)
        }
        __pyx_t_5 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_5);
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_wc_status};
          __pyx_t_7 = 0; {
            __pyx_t_8 = __Pyx_ImportFrom(__pyx_t_5, __pyx_imported_names[__pyx_t_7]); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 13, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_8);
            switch (__pyx_t_7) {
              case 0:
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py);
              __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py, __pyx_t_8);
              __Pyx_GIVEREF(__pyx_t_8);
              break;
              default:;
            }
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_wc_status as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py
 *         except ImportError:
*/
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "EnumTypeToPy":14
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_wc_status as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py
 *         except ImportError:             # <<<<<<<<<<<<<<
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py = False  # False indicates "don't try again"
 *             import warnings
*/
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_ImportError))));
      if (__pyx_t_9) {
        __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_8, &__pyx_t_10) < 0) __PYX_ERR(1, 14, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_10);

        /* "EnumTypeToPy":15
 *             from pyverbs.libibverbs_enums import ibv_wc_status as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py = False  # False indicates "don't try again"             # <<<<<<<<<<<<<<
 *             import warnings
 *             warnings.warn(
*/
        __Pyx_INCREF(Py_False);
        __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py);
        __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py, Py_False);
        __Pyx_GIVEREF(Py_False);

        /* "EnumTypeToPy":16
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py = False  # False indicates "don't try again"
 *             import warnings             # <<<<<<<<<<<<<<
 *             warnings.warn(
 *                 f"enum class ibv_wc_status not importable from pyverbs.libibverbs_enums. "
*/
        __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_warnings, 0, 0, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 16, __pyx_L6_except_error)
        __pyx_t_11 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_v_warnings = __pyx_t_11;
        __pyx_t_11 = 0;

        /* "EnumTypeToPy":17
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py = False  # False indicates "don't try again"
 *             import warnings
 *             warnings.warn(             # <<<<<<<<<<<<<<
 *                 f"enum class ibv_wc_status not importable from pyverbs.libibverbs_enums. "
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
*/
        __pyx_t_12 = __pyx_v_warnings;
        __Pyx_INCREF(__pyx_t_12);
        __pyx_t_13 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_mstate_global->__pyx_kp_u_enum_class_ibv_wc_status_not_imp};
          __pyx_t_11 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_warn, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 17, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_11);
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L5_exception_handled;
      }
      goto __pyx_L6_except_error;

      /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_wc_status as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py
 *         except ImportError:
*/
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L5_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_L9_try_end:;
    }

    /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_wc_status as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py
*/
  }

  /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py == Py_False);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":24
 * 
 * 
 *         return <int>c_val             # <<<<<<<<<<<<<<
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyLong_From_int(((int)__pyx_v_c_val)); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumTypeToPy":25
 * 
 *         return <int>c_val
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py             # <<<<<<<<<<<<<<
 * 
 *     if 0:
*/
  __Pyx_INCREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py);
  __pyx_v___pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py;

  /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_wc_status.IBV_WC_SUCCESS:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_SUCCESS
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_LEN_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_SUCCESS);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":30
 *         pass
 *     elif c_val == ibv_wc_status.IBV_WC_SUCCESS:
 *         return __pyx_enum.IBV_WC_SUCCESS             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_LEN_ERR:
 *         return __pyx_enum.IBV_WC_LOC_LEN_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_SUCCESS); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_wc_status.IBV_WC_SUCCESS:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_SUCCESS
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_LEN_ERR:
*/
  }

  /* "EnumTypeToPy":31
 *     elif c_val == ibv_wc_status.IBV_WC_SUCCESS:
 *         return __pyx_enum.IBV_WC_SUCCESS
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_LEN_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_LEN_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_QP_OP_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_LOC_LEN_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":32
 *         return __pyx_enum.IBV_WC_SUCCESS
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_LEN_ERR:
 *         return __pyx_enum.IBV_WC_LOC_LEN_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_QP_OP_ERR:
 *         return __pyx_enum.IBV_WC_LOC_QP_OP_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_LOC_LEN_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 32, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":31
 *     elif c_val == ibv_wc_status.IBV_WC_SUCCESS:
 *         return __pyx_enum.IBV_WC_SUCCESS
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_LEN_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_LEN_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_QP_OP_ERR:
*/
  }

  /* "EnumTypeToPy":33
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_LEN_ERR:
 *         return __pyx_enum.IBV_WC_LOC_LEN_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_QP_OP_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_QP_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_EEC_OP_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_LOC_QP_OP_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":34
 *         return __pyx_enum.IBV_WC_LOC_LEN_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_QP_OP_ERR:
 *         return __pyx_enum.IBV_WC_LOC_QP_OP_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_EEC_OP_ERR:
 *         return __pyx_enum.IBV_WC_LOC_EEC_OP_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_LOC_QP_OP_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":33
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_LEN_ERR:
 *         return __pyx_enum.IBV_WC_LOC_LEN_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_QP_OP_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_QP_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_EEC_OP_ERR:
*/
  }

  /* "EnumTypeToPy":35
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_QP_OP_ERR:
 *         return __pyx_enum.IBV_WC_LOC_QP_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_EEC_OP_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_EEC_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_PROT_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_LOC_EEC_OP_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":36
 *         return __pyx_enum.IBV_WC_LOC_QP_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_EEC_OP_ERR:
 *         return __pyx_enum.IBV_WC_LOC_EEC_OP_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_PROT_ERR:
 *         return __pyx_enum.IBV_WC_LOC_PROT_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_LOC_EEC_OP_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":35
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_QP_OP_ERR:
 *         return __pyx_enum.IBV_WC_LOC_QP_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_EEC_OP_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_EEC_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_PROT_ERR:
*/
  }

  /* "EnumTypeToPy":37
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_EEC_OP_ERR:
 *         return __pyx_enum.IBV_WC_LOC_EEC_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_PROT_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_PROT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_WR_FLUSH_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_LOC_PROT_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":38
 *         return __pyx_enum.IBV_WC_LOC_EEC_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_PROT_ERR:
 *         return __pyx_enum.IBV_WC_LOC_PROT_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_WR_FLUSH_ERR:
 *         return __pyx_enum.IBV_WC_WR_FLUSH_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_LOC_PROT_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":37
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_EEC_OP_ERR:
 *         return __pyx_enum.IBV_WC_LOC_EEC_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_PROT_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_PROT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_WR_FLUSH_ERR:
*/
  }

  /* "EnumTypeToPy":39
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_PROT_ERR:
 *         return __pyx_enum.IBV_WC_LOC_PROT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_WR_FLUSH_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_WR_FLUSH_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_MW_BIND_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_WR_FLUSH_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":40
 *         return __pyx_enum.IBV_WC_LOC_PROT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_WR_FLUSH_ERR:
 *         return __pyx_enum.IBV_WC_WR_FLUSH_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_MW_BIND_ERR:
 *         return __pyx_enum.IBV_WC_MW_BIND_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_WR_FLUSH_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":39
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_PROT_ERR:
 *         return __pyx_enum.IBV_WC_LOC_PROT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_WR_FLUSH_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_WR_FLUSH_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_MW_BIND_ERR:
*/
  }

  /* "EnumTypeToPy":41
 *     elif c_val == ibv_wc_status.IBV_WC_WR_FLUSH_ERR:
 *         return __pyx_enum.IBV_WC_WR_FLUSH_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_MW_BIND_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_MW_BIND_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_BAD_RESP_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_MW_BIND_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":42
 *         return __pyx_enum.IBV_WC_WR_FLUSH_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_MW_BIND_ERR:
 *         return __pyx_enum.IBV_WC_MW_BIND_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_BAD_RESP_ERR:
 *         return __pyx_enum.IBV_WC_BAD_RESP_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_MW_BIND_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":41
 *     elif c_val == ibv_wc_status.IBV_WC_WR_FLUSH_ERR:
 *         return __pyx_enum.IBV_WC_WR_FLUSH_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_MW_BIND_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_MW_BIND_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_BAD_RESP_ERR:
*/
  }

  /* "EnumTypeToPy":43
 *     elif c_val == ibv_wc_status.IBV_WC_MW_BIND_ERR:
 *         return __pyx_enum.IBV_WC_MW_BIND_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_BAD_RESP_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_BAD_RESP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_ACCESS_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_BAD_RESP_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":44
 *         return __pyx_enum.IBV_WC_MW_BIND_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_BAD_RESP_ERR:
 *         return __pyx_enum.IBV_WC_BAD_RESP_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_ACCESS_ERR:
 *         return __pyx_enum.IBV_WC_LOC_ACCESS_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_BAD_RESP_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":43
 *     elif c_val == ibv_wc_status.IBV_WC_MW_BIND_ERR:
 *         return __pyx_enum.IBV_WC_MW_BIND_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_BAD_RESP_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_BAD_RESP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_ACCESS_ERR:
*/
  }

  /* "EnumTypeToPy":45
 *     elif c_val == ibv_wc_status.IBV_WC_BAD_RESP_ERR:
 *         return __pyx_enum.IBV_WC_BAD_RESP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_ACCESS_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_ACCESS_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_REQ_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_LOC_ACCESS_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":46
 *         return __pyx_enum.IBV_WC_BAD_RESP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_ACCESS_ERR:
 *         return __pyx_enum.IBV_WC_LOC_ACCESS_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_REQ_ERR:
 *         return __pyx_enum.IBV_WC_REM_INV_REQ_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_LOC_ACCESS_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 46, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":45
 *     elif c_val == ibv_wc_status.IBV_WC_BAD_RESP_ERR:
 *         return __pyx_enum.IBV_WC_BAD_RESP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_ACCESS_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_ACCESS_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_REQ_ERR:
*/
  }

  /* "EnumTypeToPy":47
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_ACCESS_ERR:
 *         return __pyx_enum.IBV_WC_LOC_ACCESS_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_REQ_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_REM_INV_REQ_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ACCESS_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_REM_INV_REQ_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":48
 *         return __pyx_enum.IBV_WC_LOC_ACCESS_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_REQ_ERR:
 *         return __pyx_enum.IBV_WC_REM_INV_REQ_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ACCESS_ERR:
 *         return __pyx_enum.IBV_WC_REM_ACCESS_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_REM_INV_REQ_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":47
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_ACCESS_ERR:
 *         return __pyx_enum.IBV_WC_LOC_ACCESS_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_REQ_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_REM_INV_REQ_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ACCESS_ERR:
*/
  }

  /* "EnumTypeToPy":49
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_REQ_ERR:
 *         return __pyx_enum.IBV_WC_REM_INV_REQ_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ACCESS_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_REM_ACCESS_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_OP_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_REM_ACCESS_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":50
 *         return __pyx_enum.IBV_WC_REM_INV_REQ_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ACCESS_ERR:
 *         return __pyx_enum.IBV_WC_REM_ACCESS_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_REM_OP_ERR:
 *         return __pyx_enum.IBV_WC_REM_OP_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_REM_ACCESS_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":49
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_REQ_ERR:
 *         return __pyx_enum.IBV_WC_REM_INV_REQ_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ACCESS_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_REM_ACCESS_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_OP_ERR:
*/
  }

  /* "EnumTypeToPy":51
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ACCESS_ERR:
 *         return __pyx_enum.IBV_WC_REM_ACCESS_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_OP_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_REM_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RETRY_EXC_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_REM_OP_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":52
 *         return __pyx_enum.IBV_WC_REM_ACCESS_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_OP_ERR:
 *         return __pyx_enum.IBV_WC_REM_OP_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_RETRY_EXC_ERR:
 *         return __pyx_enum.IBV_WC_RETRY_EXC_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_REM_OP_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":51
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ACCESS_ERR:
 *         return __pyx_enum.IBV_WC_REM_ACCESS_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_OP_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_REM_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RETRY_EXC_ERR:
*/
  }

  /* "EnumTypeToPy":53
 *     elif c_val == ibv_wc_status.IBV_WC_REM_OP_ERR:
 *         return __pyx_enum.IBV_WC_REM_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RETRY_EXC_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RETRY_EXC_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RNR_RETRY_EXC_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_RETRY_EXC_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":54
 *         return __pyx_enum.IBV_WC_REM_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RETRY_EXC_ERR:
 *         return __pyx_enum.IBV_WC_RETRY_EXC_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_RNR_RETRY_EXC_ERR:
 *         return __pyx_enum.IBV_WC_RNR_RETRY_EXC_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_RETRY_EXC_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":53
 *     elif c_val == ibv_wc_status.IBV_WC_REM_OP_ERR:
 *         return __pyx_enum.IBV_WC_REM_OP_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RETRY_EXC_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RETRY_EXC_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RNR_RETRY_EXC_ERR:
*/
  }

  /* "EnumTypeToPy":55
 *     elif c_val == ibv_wc_status.IBV_WC_RETRY_EXC_ERR:
 *         return __pyx_enum.IBV_WC_RETRY_EXC_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RNR_RETRY_EXC_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RNR_RETRY_EXC_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_RDD_VIOL_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_RNR_RETRY_EXC_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":56
 *         return __pyx_enum.IBV_WC_RETRY_EXC_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RNR_RETRY_EXC_ERR:
 *         return __pyx_enum.IBV_WC_RNR_RETRY_EXC_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_RDD_VIOL_ERR:
 *         return __pyx_enum.IBV_WC_LOC_RDD_VIOL_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_RNR_RETRY_EXC_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":55
 *     elif c_val == ibv_wc_status.IBV_WC_RETRY_EXC_ERR:
 *         return __pyx_enum.IBV_WC_RETRY_EXC_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RNR_RETRY_EXC_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RNR_RETRY_EXC_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_RDD_VIOL_ERR:
*/
  }

  /* "EnumTypeToPy":57
 *     elif c_val == ibv_wc_status.IBV_WC_RNR_RETRY_EXC_ERR:
 *         return __pyx_enum.IBV_WC_RNR_RETRY_EXC_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_RDD_VIOL_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_RDD_VIOL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_RD_REQ_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_LOC_RDD_VIOL_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":58
 *         return __pyx_enum.IBV_WC_RNR_RETRY_EXC_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_RDD_VIOL_ERR:
 *         return __pyx_enum.IBV_WC_LOC_RDD_VIOL_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_RD_REQ_ERR:
 *         return __pyx_enum.IBV_WC_REM_INV_RD_REQ_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_LOC_RDD_VIOL_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":57
 *     elif c_val == ibv_wc_status.IBV_WC_RNR_RETRY_EXC_ERR:
 *         return __pyx_enum.IBV_WC_RNR_RETRY_EXC_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_RDD_VIOL_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOC_RDD_VIOL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_RD_REQ_ERR:
*/
  }

  /* "EnumTypeToPy":59
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_RDD_VIOL_ERR:
 *         return __pyx_enum.IBV_WC_LOC_RDD_VIOL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_RD_REQ_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_REM_INV_RD_REQ_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ABORT_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_REM_INV_RD_REQ_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":60
 *         return __pyx_enum.IBV_WC_LOC_RDD_VIOL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_RD_REQ_ERR:
 *         return __pyx_enum.IBV_WC_REM_INV_RD_REQ_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ABORT_ERR:
 *         return __pyx_enum.IBV_WC_REM_ABORT_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_REM_INV_RD_REQ_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":59
 *     elif c_val == ibv_wc_status.IBV_WC_LOC_RDD_VIOL_ERR:
 *         return __pyx_enum.IBV_WC_LOC_RDD_VIOL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_RD_REQ_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_REM_INV_RD_REQ_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ABORT_ERR:
*/
  }

  /* "EnumTypeToPy":61
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_RD_REQ_ERR:
 *         return __pyx_enum.IBV_WC_REM_INV_RD_REQ_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ABORT_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_REM_ABORT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EECN_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_REM_ABORT_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":62
 *         return __pyx_enum.IBV_WC_REM_INV_RD_REQ_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ABORT_ERR:
 *         return __pyx_enum.IBV_WC_REM_ABORT_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EECN_ERR:
 *         return __pyx_enum.IBV_WC_INV_EECN_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_REM_ABORT_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":61
 *     elif c_val == ibv_wc_status.IBV_WC_REM_INV_RD_REQ_ERR:
 *         return __pyx_enum.IBV_WC_REM_INV_RD_REQ_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ABORT_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_REM_ABORT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EECN_ERR:
*/
  }

  /* "EnumTypeToPy":63
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ABORT_ERR:
 *         return __pyx_enum.IBV_WC_REM_ABORT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EECN_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_INV_EECN_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EEC_STATE_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_INV_EECN_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":64
 *         return __pyx_enum.IBV_WC_REM_ABORT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EECN_ERR:
 *         return __pyx_enum.IBV_WC_INV_EECN_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EEC_STATE_ERR:
 *         return __pyx_enum.IBV_WC_INV_EEC_STATE_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_INV_EECN_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":63
 *     elif c_val == ibv_wc_status.IBV_WC_REM_ABORT_ERR:
 *         return __pyx_enum.IBV_WC_REM_ABORT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EECN_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_INV_EECN_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EEC_STATE_ERR:
*/
  }

  /* "EnumTypeToPy":65
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EECN_ERR:
 *         return __pyx_enum.IBV_WC_INV_EECN_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EEC_STATE_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_INV_EEC_STATE_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_FATAL_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_INV_EEC_STATE_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":66
 *         return __pyx_enum.IBV_WC_INV_EECN_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EEC_STATE_ERR:
 *         return __pyx_enum.IBV_WC_INV_EEC_STATE_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_FATAL_ERR:
 *         return __pyx_enum.IBV_WC_FATAL_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_INV_EEC_STATE_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":65
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EECN_ERR:
 *         return __pyx_enum.IBV_WC_INV_EECN_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EEC_STATE_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_INV_EEC_STATE_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_FATAL_ERR:
*/
  }

  /* "EnumTypeToPy":67
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EEC_STATE_ERR:
 *         return __pyx_enum.IBV_WC_INV_EEC_STATE_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_FATAL_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_FATAL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RESP_TIMEOUT_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_FATAL_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":68
 *         return __pyx_enum.IBV_WC_INV_EEC_STATE_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_FATAL_ERR:
 *         return __pyx_enum.IBV_WC_FATAL_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_RESP_TIMEOUT_ERR:
 *         return __pyx_enum.IBV_WC_RESP_TIMEOUT_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_FATAL_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":67
 *     elif c_val == ibv_wc_status.IBV_WC_INV_EEC_STATE_ERR:
 *         return __pyx_enum.IBV_WC_INV_EEC_STATE_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_FATAL_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_FATAL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RESP_TIMEOUT_ERR:
*/
  }

  /* "EnumTypeToPy":69
 *     elif c_val == ibv_wc_status.IBV_WC_FATAL_ERR:
 *         return __pyx_enum.IBV_WC_FATAL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RESP_TIMEOUT_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RESP_TIMEOUT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_GENERAL_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_RESP_TIMEOUT_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":70
 *         return __pyx_enum.IBV_WC_FATAL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RESP_TIMEOUT_ERR:
 *         return __pyx_enum.IBV_WC_RESP_TIMEOUT_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_GENERAL_ERR:
 *         return __pyx_enum.IBV_WC_GENERAL_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_RESP_TIMEOUT_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":69
 *     elif c_val == ibv_wc_status.IBV_WC_FATAL_ERR:
 *         return __pyx_enum.IBV_WC_FATAL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_RESP_TIMEOUT_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RESP_TIMEOUT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_GENERAL_ERR:
*/
  }

  /* "EnumTypeToPy":71
 *     elif c_val == ibv_wc_status.IBV_WC_RESP_TIMEOUT_ERR:
 *         return __pyx_enum.IBV_WC_RESP_TIMEOUT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_GENERAL_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_GENERAL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_TM_ERR:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_GENERAL_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":72
 *         return __pyx_enum.IBV_WC_RESP_TIMEOUT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_GENERAL_ERR:
 *         return __pyx_enum.IBV_WC_GENERAL_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_TM_ERR:
 *         return __pyx_enum.IBV_WC_TM_ERR
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_GENERAL_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":71
 *     elif c_val == ibv_wc_status.IBV_WC_RESP_TIMEOUT_ERR:
 *         return __pyx_enum.IBV_WC_RESP_TIMEOUT_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_GENERAL_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_GENERAL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_TM_ERR:
*/
  }

  /* "EnumTypeToPy":73
 *     elif c_val == ibv_wc_status.IBV_WC_GENERAL_ERR:
 *         return __pyx_enum.IBV_WC_GENERAL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_TM_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_TM_RNDV_INCOMPLETE:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_TM_ERR);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":74
 *         return __pyx_enum.IBV_WC_GENERAL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_TM_ERR:
 *         return __pyx_enum.IBV_WC_TM_ERR             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_status.IBV_WC_TM_RNDV_INCOMPLETE:
 *         return __pyx_enum.IBV_WC_TM_RNDV_INCOMPLETE
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_ERR); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":73
 *     elif c_val == ibv_wc_status.IBV_WC_GENERAL_ERR:
 *         return __pyx_enum.IBV_WC_GENERAL_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_TM_ERR:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_TM_RNDV_INCOMPLETE:
*/
  }

  /* "EnumTypeToPy":75
 *     elif c_val == ibv_wc_status.IBV_WC_TM_ERR:
 *         return __pyx_enum.IBV_WC_TM_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_TM_RNDV_INCOMPLETE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_RNDV_INCOMPLETE
 *     else:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_TM_RNDV_INCOMPLETE);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":76
 *         return __pyx_enum.IBV_WC_TM_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_TM_RNDV_INCOMPLETE:
 *         return __pyx_enum.IBV_WC_TM_RNDV_INCOMPLETE             # <<<<<<<<<<<<<<
 *     else:
 *         underlying_c_val = <int>c_val
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_RNDV_INCOMPLETE); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":75
 *     elif c_val == ibv_wc_status.IBV_WC_TM_ERR:
 *         return __pyx_enum.IBV_WC_TM_ERR
 *     elif c_val == ibv_wc_status.IBV_WC_TM_RNDV_INCOMPLETE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_RNDV_INCOMPLETE
 *     else:
*/
  }

  /* "EnumTypeToPy":78
 *         return __pyx_enum.IBV_WC_TM_RNDV_INCOMPLETE
 *     else:
 *         underlying_c_val = <int>c_val             # <<<<<<<<<<<<<<
 *         return __pyx_enum(underlying_c_val)
*/
  /*else*/ {
    __pyx_v_underlying_c_val = ((int)__pyx_v_c_val);

    /* "EnumTypeToPy":79
 *     else:
 *         underlying_c_val = <int>c_val
 *         return __pyx_enum(underlying_c_val)             # <<<<<<<<<<<<<<
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = NULL;
    __Pyx_INCREF(__pyx_v___pyx_enum);
    __pyx_t_5 = __pyx_v___pyx_enum; 
    __pyx_t_11 = __Pyx_PyLong_From_int(__pyx_v_underlying_c_val); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_13 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_8);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_13 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_11};
      __pyx_t_10 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 79, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
    }
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;
  }

  /* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(ibv_wc_status c_val):
 *     cdef object __pyx_enum
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_enum);
  __Pyx_XDECREF(__pyx_v_warnings);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py(enum ibv_wc_opcode __pyx_v_c_val) {
  PyObject *__pyx_v___pyx_enum = 0;
  PyObject *__pyx_v_warnings = NULL;
  int __pyx_v_underlying_c_val;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  size_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py", 0);

  /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_wc_opcode as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py == Py_None);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_wc_opcode as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py
 *         except ImportError:
*/
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      /*try:*/ {

        /* "EnumTypeToPy":13
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py is None:
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_wc_opcode as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py             # <<<<<<<<<<<<<<
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py = False  # False indicates "don't try again"
*/
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_wc_opcode};
          __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyverbs_libibverbs_enums, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L4_error)
        }
        __pyx_t_5 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_5);
        {
          PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_ibv_wc_opcode};
          __pyx_t_7 = 0; {
            __pyx_t_8 = __Pyx_ImportFrom(__pyx_t_5, __pyx_imported_names[__pyx_t_7]); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 13, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_8);
            switch (__pyx_t_7) {
              case 0:
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py);
              __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py, __pyx_t_8);
              __Pyx_GIVEREF(__pyx_t_8);
              break;
              default:;
            }
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_wc_opcode as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py
 *         except ImportError:
*/
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "EnumTypeToPy":14
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_wc_opcode as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py
 *         except ImportError:             # <<<<<<<<<<<<<<
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py = False  # False indicates "don't try again"
 *             import warnings
*/
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_ImportError))));
      if (__pyx_t_9) {
        __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_8, &__pyx_t_10) < 0) __PYX_ERR(1, 14, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_10);

        /* "EnumTypeToPy":15
 *             from pyverbs.libibverbs_enums import ibv_wc_opcode as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py = False  # False indicates "don't try again"             # <<<<<<<<<<<<<<
 *             import warnings
 *             warnings.warn(
*/
        __Pyx_INCREF(Py_False);
        __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py);
        __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py, Py_False);
        __Pyx_GIVEREF(Py_False);

        /* "EnumTypeToPy":16
 *         except ImportError:
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py = False  # False indicates "don't try again"
 *             import warnings             # <<<<<<<<<<<<<<
 *             warnings.warn(
 *                 f"enum class ibv_wc_opcode not importable from pyverbs.libibverbs_enums. "
*/
        __pyx_t_6 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_warnings, 0, 0, NULL, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 16, __pyx_L6_except_error)
        __pyx_t_11 = __pyx_t_6;
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_v_warnings = __pyx_t_11;
        __pyx_t_11 = 0;

        /* "EnumTypeToPy":17
 *             __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py = False  # False indicates "don't try again"
 *             import warnings
 *             warnings.warn(             # <<<<<<<<<<<<<<
 *                 f"enum class ibv_wc_opcode not importable from pyverbs.libibverbs_enums. "
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
*/
        __pyx_t_12 = __pyx_v_warnings;
        __Pyx_INCREF(__pyx_t_12);
        __pyx_t_13 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_mstate_global->__pyx_kp_u_enum_class_ibv_wc_opcode_not_imp};
          __pyx_t_11 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_warn, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 17, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_11);
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L5_exception_handled;
      }
      goto __pyx_L6_except_error;

      /* "EnumTypeToPy":12
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py is None:
 *         try:             # <<<<<<<<<<<<<<
 *             from pyverbs.libibverbs_enums import ibv_wc_opcode as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py
 *         except ImportError:
*/
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      goto __pyx_L1_error;
      __pyx_L5_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
      __pyx_L9_try_end:;
    }

    /* "EnumTypeToPy":11
 * 
 * 
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py is None:             # <<<<<<<<<<<<<<
 *         try:
 *             from pyverbs.libibverbs_enums import ibv_wc_opcode as __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py
*/
  }

  /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = (__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py == Py_False);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":24
 * 
 * 
 *         return <int>c_val             # <<<<<<<<<<<<<<
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyLong_From_int(((int)__pyx_v_c_val)); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 24, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":21
 *                 "You are probably using a cpdef enum declared in a .pxd file that "
 *                 "does not have a .py  or .pyx file.")
 *     if __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py is False:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumTypeToPy":25
 * 
 *         return <int>c_val
 *     __pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py             # <<<<<<<<<<<<<<
 * 
 *     if 0:
*/
  __Pyx_INCREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py);
  __pyx_v___pyx_enum = __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py;

  /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_wc_opcode.IBV_WC_SEND:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_SEND
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_WRITE:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_SEND);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":30
 *         pass
 *     elif c_val == ibv_wc_opcode.IBV_WC_SEND:
 *         return __pyx_enum.IBV_WC_SEND             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_WRITE:
 *         return __pyx_enum.IBV_WC_RDMA_WRITE
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_SEND); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":29
 *     if 0:
 *         pass
 *     elif c_val == ibv_wc_opcode.IBV_WC_SEND:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_SEND
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_WRITE:
*/
  }

  /* "EnumTypeToPy":31
 *     elif c_val == ibv_wc_opcode.IBV_WC_SEND:
 *         return __pyx_enum.IBV_WC_SEND
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_WRITE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RDMA_WRITE
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_READ:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_RDMA_WRITE);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":32
 *         return __pyx_enum.IBV_WC_SEND
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_WRITE:
 *         return __pyx_enum.IBV_WC_RDMA_WRITE             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_READ:
 *         return __pyx_enum.IBV_WC_RDMA_READ
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_RDMA_WRITE); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 32, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":31
 *     elif c_val == ibv_wc_opcode.IBV_WC_SEND:
 *         return __pyx_enum.IBV_WC_SEND
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_WRITE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RDMA_WRITE
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_READ:
*/
  }

  /* "EnumTypeToPy":33
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_WRITE:
 *         return __pyx_enum.IBV_WC_RDMA_WRITE
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_READ:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RDMA_READ
 *     elif c_val == ibv_wc_opcode.IBV_WC_COMP_SWAP:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_RDMA_READ);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":34
 *         return __pyx_enum.IBV_WC_RDMA_WRITE
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_READ:
 *         return __pyx_enum.IBV_WC_RDMA_READ             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_COMP_SWAP:
 *         return __pyx_enum.IBV_WC_COMP_SWAP
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_RDMA_READ); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":33
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_WRITE:
 *         return __pyx_enum.IBV_WC_RDMA_WRITE
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_READ:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RDMA_READ
 *     elif c_val == ibv_wc_opcode.IBV_WC_COMP_SWAP:
*/
  }

  /* "EnumTypeToPy":35
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_READ:
 *         return __pyx_enum.IBV_WC_RDMA_READ
 *     elif c_val == ibv_wc_opcode.IBV_WC_COMP_SWAP:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_COMP_SWAP
 *     elif c_val == ibv_wc_opcode.IBV_WC_FETCH_ADD:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_COMP_SWAP);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":36
 *         return __pyx_enum.IBV_WC_RDMA_READ
 *     elif c_val == ibv_wc_opcode.IBV_WC_COMP_SWAP:
 *         return __pyx_enum.IBV_WC_COMP_SWAP             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_FETCH_ADD:
 *         return __pyx_enum.IBV_WC_FETCH_ADD
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_COMP_SWAP); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":35
 *     elif c_val == ibv_wc_opcode.IBV_WC_RDMA_READ:
 *         return __pyx_enum.IBV_WC_RDMA_READ
 *     elif c_val == ibv_wc_opcode.IBV_WC_COMP_SWAP:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_COMP_SWAP
 *     elif c_val == ibv_wc_opcode.IBV_WC_FETCH_ADD:
*/
  }

  /* "EnumTypeToPy":37
 *     elif c_val == ibv_wc_opcode.IBV_WC_COMP_SWAP:
 *         return __pyx_enum.IBV_WC_COMP_SWAP
 *     elif c_val == ibv_wc_opcode.IBV_WC_FETCH_ADD:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_FETCH_ADD
 *     elif c_val == ibv_wc_opcode.IBV_WC_BIND_MW:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_FETCH_ADD);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":38
 *         return __pyx_enum.IBV_WC_COMP_SWAP
 *     elif c_val == ibv_wc_opcode.IBV_WC_FETCH_ADD:
 *         return __pyx_enum.IBV_WC_FETCH_ADD             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_BIND_MW:
 *         return __pyx_enum.IBV_WC_BIND_MW
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_FETCH_ADD); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":37
 *     elif c_val == ibv_wc_opcode.IBV_WC_COMP_SWAP:
 *         return __pyx_enum.IBV_WC_COMP_SWAP
 *     elif c_val == ibv_wc_opcode.IBV_WC_FETCH_ADD:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_FETCH_ADD
 *     elif c_val == ibv_wc_opcode.IBV_WC_BIND_MW:
*/
  }

  /* "EnumTypeToPy":39
 *     elif c_val == ibv_wc_opcode.IBV_WC_FETCH_ADD:
 *         return __pyx_enum.IBV_WC_FETCH_ADD
 *     elif c_val == ibv_wc_opcode.IBV_WC_BIND_MW:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_BIND_MW
 *     elif c_val == ibv_wc_opcode.IBV_WC_LOCAL_INV:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_BIND_MW);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":40
 *         return __pyx_enum.IBV_WC_FETCH_ADD
 *     elif c_val == ibv_wc_opcode.IBV_WC_BIND_MW:
 *         return __pyx_enum.IBV_WC_BIND_MW             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_LOCAL_INV:
 *         return __pyx_enum.IBV_WC_LOCAL_INV
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_BIND_MW); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":39
 *     elif c_val == ibv_wc_opcode.IBV_WC_FETCH_ADD:
 *         return __pyx_enum.IBV_WC_FETCH_ADD
 *     elif c_val == ibv_wc_opcode.IBV_WC_BIND_MW:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_BIND_MW
 *     elif c_val == ibv_wc_opcode.IBV_WC_LOCAL_INV:
*/
  }

  /* "EnumTypeToPy":41
 *     elif c_val == ibv_wc_opcode.IBV_WC_BIND_MW:
 *         return __pyx_enum.IBV_WC_BIND_MW
 *     elif c_val == ibv_wc_opcode.IBV_WC_LOCAL_INV:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOCAL_INV
 *     elif c_val == ibv_wc_opcode.IBV_WC_TSO:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_LOCAL_INV);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":42
 *         return __pyx_enum.IBV_WC_BIND_MW
 *     elif c_val == ibv_wc_opcode.IBV_WC_LOCAL_INV:
 *         return __pyx_enum.IBV_WC_LOCAL_INV             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_TSO:
 *         return __pyx_enum.IBV_WC_TSO
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_LOCAL_INV); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":41
 *     elif c_val == ibv_wc_opcode.IBV_WC_BIND_MW:
 *         return __pyx_enum.IBV_WC_BIND_MW
 *     elif c_val == ibv_wc_opcode.IBV_WC_LOCAL_INV:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_LOCAL_INV
 *     elif c_val == ibv_wc_opcode.IBV_WC_TSO:
*/
  }

  /* "EnumTypeToPy":43
 *     elif c_val == ibv_wc_opcode.IBV_WC_LOCAL_INV:
 *         return __pyx_enum.IBV_WC_LOCAL_INV
 *     elif c_val == ibv_wc_opcode.IBV_WC_TSO:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TSO
 *     elif c_val == ibv_wc_opcode.IBV_WC_FLUSH:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_TSO);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":44
 *         return __pyx_enum.IBV_WC_LOCAL_INV
 *     elif c_val == ibv_wc_opcode.IBV_WC_TSO:
 *         return __pyx_enum.IBV_WC_TSO             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_FLUSH:
 *         return __pyx_enum.IBV_WC_FLUSH
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_TSO); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":43
 *     elif c_val == ibv_wc_opcode.IBV_WC_LOCAL_INV:
 *         return __pyx_enum.IBV_WC_LOCAL_INV
 *     elif c_val == ibv_wc_opcode.IBV_WC_TSO:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TSO
 *     elif c_val == ibv_wc_opcode.IBV_WC_FLUSH:
*/
  }

  /* "EnumTypeToPy":45
 *     elif c_val == ibv_wc_opcode.IBV_WC_TSO:
 *         return __pyx_enum.IBV_WC_TSO
 *     elif c_val == ibv_wc_opcode.IBV_WC_FLUSH:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_FLUSH
 *     elif c_val == ibv_wc_opcode.IBV_WC_ATOMIC_WRITE:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_FLUSH);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":46
 *         return __pyx_enum.IBV_WC_TSO
 *     elif c_val == ibv_wc_opcode.IBV_WC_FLUSH:
 *         return __pyx_enum.IBV_WC_FLUSH             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_ATOMIC_WRITE:
 *         return __pyx_enum.IBV_WC_ATOMIC_WRITE
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_FLUSH); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 46, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":45
 *     elif c_val == ibv_wc_opcode.IBV_WC_TSO:
 *         return __pyx_enum.IBV_WC_TSO
 *     elif c_val == ibv_wc_opcode.IBV_WC_FLUSH:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_FLUSH
 *     elif c_val == ibv_wc_opcode.IBV_WC_ATOMIC_WRITE:
*/
  }

  /* "EnumTypeToPy":47
 *     elif c_val == ibv_wc_opcode.IBV_WC_FLUSH:
 *         return __pyx_enum.IBV_WC_FLUSH
 *     elif c_val == ibv_wc_opcode.IBV_WC_ATOMIC_WRITE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_ATOMIC_WRITE
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_ATOMIC_WRITE);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":48
 *         return __pyx_enum.IBV_WC_FLUSH
 *     elif c_val == ibv_wc_opcode.IBV_WC_ATOMIC_WRITE:
 *         return __pyx_enum.IBV_WC_ATOMIC_WRITE             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV:
 *         return __pyx_enum.IBV_WC_RECV
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_ATOMIC_WRITE); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 48, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":47
 *     elif c_val == ibv_wc_opcode.IBV_WC_FLUSH:
 *         return __pyx_enum.IBV_WC_FLUSH
 *     elif c_val == ibv_wc_opcode.IBV_WC_ATOMIC_WRITE:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_ATOMIC_WRITE
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV:
*/
  }

  /* "EnumTypeToPy":49
 *     elif c_val == ibv_wc_opcode.IBV_WC_ATOMIC_WRITE:
 *         return __pyx_enum.IBV_WC_ATOMIC_WRITE
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RECV
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV_RDMA_WITH_IMM:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_RECV);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":50
 *         return __pyx_enum.IBV_WC_ATOMIC_WRITE
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV:
 *         return __pyx_enum.IBV_WC_RECV             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV_RDMA_WITH_IMM:
 *         return __pyx_enum.IBV_WC_RECV_RDMA_WITH_IMM
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_RECV); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":49
 *     elif c_val == ibv_wc_opcode.IBV_WC_ATOMIC_WRITE:
 *         return __pyx_enum.IBV_WC_ATOMIC_WRITE
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RECV
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV_RDMA_WITH_IMM:
*/
  }

  /* "EnumTypeToPy":51
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV:
 *         return __pyx_enum.IBV_WC_RECV
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV_RDMA_WITH_IMM:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RECV_RDMA_WITH_IMM
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_ADD:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_RECV_RDMA_WITH_IMM);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":52
 *         return __pyx_enum.IBV_WC_RECV
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV_RDMA_WITH_IMM:
 *         return __pyx_enum.IBV_WC_RECV_RDMA_WITH_IMM             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_ADD:
 *         return __pyx_enum.IBV_WC_TM_ADD
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_RECV_RDMA_WITH_IMM); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":51
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV:
 *         return __pyx_enum.IBV_WC_RECV
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV_RDMA_WITH_IMM:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_RECV_RDMA_WITH_IMM
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_ADD:
*/
  }

  /* "EnumTypeToPy":53
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV_RDMA_WITH_IMM:
 *         return __pyx_enum.IBV_WC_RECV_RDMA_WITH_IMM
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_ADD:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_ADD
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_DEL:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_TM_ADD);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":54
 *         return __pyx_enum.IBV_WC_RECV_RDMA_WITH_IMM
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_ADD:
 *         return __pyx_enum.IBV_WC_TM_ADD             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_DEL:
 *         return __pyx_enum.IBV_WC_TM_DEL
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_ADD); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":53
 *     elif c_val == ibv_wc_opcode.IBV_WC_RECV_RDMA_WITH_IMM:
 *         return __pyx_enum.IBV_WC_RECV_RDMA_WITH_IMM
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_ADD:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_ADD
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_DEL:
*/
  }

  /* "EnumTypeToPy":55
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_ADD:
 *         return __pyx_enum.IBV_WC_TM_ADD
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_DEL:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_DEL
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_SYNC:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_TM_DEL);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":56
 *         return __pyx_enum.IBV_WC_TM_ADD
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_DEL:
 *         return __pyx_enum.IBV_WC_TM_DEL             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_SYNC:
 *         return __pyx_enum.IBV_WC_TM_SYNC
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_DEL); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":55
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_ADD:
 *         return __pyx_enum.IBV_WC_TM_ADD
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_DEL:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_DEL
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_SYNC:
*/
  }

  /* "EnumTypeToPy":57
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_DEL:
 *         return __pyx_enum.IBV_WC_TM_DEL
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_SYNC:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_SYNC
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_RECV:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_TM_SYNC);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":58
 *         return __pyx_enum.IBV_WC_TM_DEL
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_SYNC:
 *         return __pyx_enum.IBV_WC_TM_SYNC             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_RECV:
 *         return __pyx_enum.IBV_WC_TM_RECV
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_SYNC); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":57
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_DEL:
 *         return __pyx_enum.IBV_WC_TM_DEL
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_SYNC:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_SYNC
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_RECV:
*/
  }

  /* "EnumTypeToPy":59
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_SYNC:
 *         return __pyx_enum.IBV_WC_TM_SYNC
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_RECV:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_RECV
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_NO_TAG:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_TM_RECV);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":60
 *         return __pyx_enum.IBV_WC_TM_SYNC
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_RECV:
 *         return __pyx_enum.IBV_WC_TM_RECV             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_NO_TAG:
 *         return __pyx_enum.IBV_WC_TM_NO_TAG
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_RECV); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":59
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_SYNC:
 *         return __pyx_enum.IBV_WC_TM_SYNC
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_RECV:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_RECV
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_NO_TAG:
*/
  }

  /* "EnumTypeToPy":61
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_RECV:
 *         return __pyx_enum.IBV_WC_TM_RECV
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_NO_TAG:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_NO_TAG
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER2:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_TM_NO_TAG);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":62
 *         return __pyx_enum.IBV_WC_TM_RECV
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_NO_TAG:
 *         return __pyx_enum.IBV_WC_TM_NO_TAG             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER2:
 *         return __pyx_enum.IBV_WC_DRIVER2
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_TM_NO_TAG); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":61
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_RECV:
 *         return __pyx_enum.IBV_WC_TM_RECV
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_NO_TAG:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_TM_NO_TAG
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER2:
*/
  }

  /* "EnumTypeToPy":63
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_NO_TAG:
 *         return __pyx_enum.IBV_WC_TM_NO_TAG
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER2:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_DRIVER2
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER3:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_DRIVER2);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":64
 *         return __pyx_enum.IBV_WC_TM_NO_TAG
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER2:
 *         return __pyx_enum.IBV_WC_DRIVER2             # <<<<<<<<<<<<<<
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER3:
 *         return __pyx_enum.IBV_WC_DRIVER3
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_DRIVER2); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":63
 *     elif c_val == ibv_wc_opcode.IBV_WC_TM_NO_TAG:
 *         return __pyx_enum.IBV_WC_TM_NO_TAG
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER2:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_DRIVER2
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER3:
*/
  }

  /* "EnumTypeToPy":65
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER2:
 *         return __pyx_enum.IBV_WC_DRIVER2
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER3:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_DRIVER3
 *     else:
*/
  __pyx_t_1 = (__pyx_v_c_val == IBV_WC_DRIVER3);
  if (__pyx_t_1) {

    /* "EnumTypeToPy":66
 *         return __pyx_enum.IBV_WC_DRIVER2
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER3:
 *         return __pyx_enum.IBV_WC_DRIVER3             # <<<<<<<<<<<<<<
 *     else:
 *         underlying_c_val = <int>c_val
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v___pyx_enum, __pyx_mstate_global->__pyx_n_u_IBV_WC_DRIVER3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "EnumTypeToPy":65
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER2:
 *         return __pyx_enum.IBV_WC_DRIVER2
 *     elif c_val == ibv_wc_opcode.IBV_WC_DRIVER3:             # <<<<<<<<<<<<<<
 *         return __pyx_enum.IBV_WC_DRIVER3
 *     else:
*/
  }

  /* "EnumTypeToPy":68
 *         return __pyx_enum.IBV_WC_DRIVER3
 *     else:
 *         underlying_c_val = <int>c_val             # <<<<<<<<<<<<<<
 *         return __pyx_enum(underlying_c_val)
*/
  /*else*/ {
    __pyx_v_underlying_c_val = ((int)__pyx_v_c_val);

    /* "EnumTypeToPy":69
 *     else:
 *         underlying_c_val = <int>c_val
 *         return __pyx_enum(underlying_c_val)             # <<<<<<<<<<<<<<
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = NULL;
    __Pyx_INCREF(__pyx_v___pyx_enum);
    __pyx_t_5 = __pyx_v___pyx_enum; 
    __pyx_t_11 = __Pyx_PyLong_From_int(__pyx_v_underlying_c_val); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_13 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_8);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_13 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_11};
      __pyx_t_10 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (__pyx_t_13*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 69, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
    }
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;
  }

  /* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py(ibv_wc_opcode c_val):
 *     cdef object __pyx_enum
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("EnumTypeToPy.__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_enum);
  __Pyx_XDECREF(__pyx_v_warnings);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":23
 * cdef class ConnParam(PyverbsObject):
 * 
 *     def __init__(self, resources=1, depth=1, flow_control=0, retry=5,             # <<<<<<<<<<<<<<
 *                  rnr_retry=5, srq=0, qp_num=0, data_len=0):
 *         """
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_4cmid_9ConnParam_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_9ConnParam___init__, "\n        Initialize a ConnParam object over an underlying rdma_conn_param\n        C object which contains connection parameters. There are a few types of\n        port spaces in RDMACM: RDMA_PS_TCP, RDMA_PS_UDP, RDMA_PS_IB and\n        RDMA_PS_IPOIB. RDMA_PS_TCP resembles RC QP connection, which provides\n        reliable, connection-oriented QP communication. This object applies only\n        to RDMA_PS_TCP port space.\n        :param resources: Max outstanding RDMA read and atomic ops that local\n                          side will accept from the remote side.\n        :param depth: Max outstanding RDMA read and atomic ops that local side\n                      will have to the remote side.\n        :param flow_control: Specifies if hardware flow control is available.\n        :param retry: Max number of times that a send, RDMA or atomic op from\n                      the remote peer should be retried.\n        :param rnr_retry: The maximum number of times that a send operation from\n                          the remote peer should be retried on a connection\n                          after receiving a receiver not ready (RNR) error.\n        :param srq: Specifies if the QP using shared receive queue, ignored if\n                    the QP created by CMID.\n        :param qp_num: Specifies the QP number, ignored if the QP created by\n                       CMID.\n        :param data_len: Specifies the private data length.\n                         RDMA_PS_TCP connect: 56\n                                     accept:  196\n                         RDMA_PS_UDP connect: 180\n                                     accept:  136\n        :return: ConnParam object\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_7pyverbs_4cmid_9ConnParam___init__;
#endif
static int __pyx_pw_7pyverbs_4cmid_9ConnParam_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_resources = 0;
  PyObject *__pyx_v_depth = 0;
  PyObject *__pyx_v_flow_control = 0;
  PyObject *__pyx_v_retry = 0;
  PyObject *__pyx_v_rnr_retry = 0;
  PyObject *__pyx_v_srq = 0;
  PyObject *__pyx_v_qp_num = 0;
  PyObject *__pyx_v_data_len = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[8] = {0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_resources,&__pyx_mstate_global->__pyx_n_u_depth,&__pyx_mstate_global->__pyx_n_u_flow_control,&__pyx_mstate_global->__pyx_n_u_retry,&__pyx_mstate_global->__pyx_n_u_rnr_retry,&__pyx_mstate_global->__pyx_n_u_srq,&__pyx_mstate_global->__pyx_n_u_qp_num,&__pyx_mstate_global->__pyx_n_u_data_len,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 23, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  8:
        values[7] = __Pyx_ArgRef_VARARGS(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 23, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_VARARGS(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 23, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 23, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 23, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 23, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 23, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 23, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 23, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 23, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_1));
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_1));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_5));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_5));
      if (!values[5]) values[5] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[6]) values[6] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[7]) values[7] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    } else {
      switch (__pyx_nargs) {
        case  8:
        values[7] = __Pyx_ArgRef_VARARGS(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 23, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_VARARGS(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 23, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 23, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 23, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 23, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 23, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 23, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 23, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_1));
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_1));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_5));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_5));
      if (!values[5]) values[5] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[6]) values[6] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[7]) values[7] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    }
    __pyx_v_resources = values[0];
    __pyx_v_depth = values[1];
    __pyx_v_flow_control = values[2];
    __pyx_v_retry = values[3];
    __pyx_v_rnr_retry = values[4];
    __pyx_v_srq = values[5];
    __pyx_v_qp_num = values[6];
    __pyx_v_data_len = values[7];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 8, __pyx_nargs); __PYX_ERR(0, 23, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.ConnParam.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_4cmid_9ConnParam___init__(((struct __pyx_obj_7pyverbs_4cmid_ConnParam *)__pyx_v_self), __pyx_v_resources, __pyx_v_depth, __pyx_v_flow_control, __pyx_v_retry, __pyx_v_rnr_retry, __pyx_v_srq, __pyx_v_qp_num, __pyx_v_data_len);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_4cmid_9ConnParam___init__(struct __pyx_obj_7pyverbs_4cmid_ConnParam *__pyx_v_self, PyObject *__pyx_v_resources, PyObject *__pyx_v_depth, PyObject *__pyx_v_flow_control, PyObject *__pyx_v_retry, PyObject *__pyx_v_rnr_retry, PyObject *__pyx_v_srq, PyObject *__pyx_v_qp_num, PyObject *__pyx_v_data_len) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  uint8_t __pyx_t_6;
  uint32_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyverbs/cmid.pyx":53
 *         :return: ConnParam object
 *         """
 *         super().__init__()             # <<<<<<<<<<<<<<
 *         memset(&self.conn_param, 0, sizeof(cm.rdma_conn_param))
 *         self.conn_param.responder_resources = resources
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_ConnParam), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":54
 *         """
 *         super().__init__()
 *         memset(&self.conn_param, 0, sizeof(cm.rdma_conn_param))             # <<<<<<<<<<<<<<
 *         self.conn_param.responder_resources = resources
 *         self.conn_param.initiator_depth = depth
*/
  (void)(memset((&__pyx_v_self->conn_param), 0, (sizeof(struct rdma_conn_param))));

  /* "pyverbs/cmid.pyx":55
 *         super().__init__()
 *         memset(&self.conn_param, 0, sizeof(cm.rdma_conn_param))
 *         self.conn_param.responder_resources = resources             # <<<<<<<<<<<<<<
 *         self.conn_param.initiator_depth = depth
 *         self.conn_param.flow_control = flow_control
*/
  __pyx_t_6 = __Pyx_PyLong_As_uint8_t(__pyx_v_resources); if (unlikely((__pyx_t_6 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 55, __pyx_L1_error)
  __pyx_v_self->conn_param.responder_resources = __pyx_t_6;

  /* "pyverbs/cmid.pyx":56
 *         memset(&self.conn_param, 0, sizeof(cm.rdma_conn_param))
 *         self.conn_param.responder_resources = resources
 *         self.conn_param.initiator_depth = depth             # <<<<<<<<<<<<<<
 *         self.conn_param.flow_control = flow_control
 *         self.conn_param.retry_count = retry
*/
  __pyx_t_6 = __Pyx_PyLong_As_uint8_t(__pyx_v_depth); if (unlikely((__pyx_t_6 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 56, __pyx_L1_error)
  __pyx_v_self->conn_param.initiator_depth = __pyx_t_6;

  /* "pyverbs/cmid.pyx":57
 *         self.conn_param.responder_resources = resources
 *         self.conn_param.initiator_depth = depth
 *         self.conn_param.flow_control = flow_control             # <<<<<<<<<<<<<<
 *         self.conn_param.retry_count = retry
 *         self.conn_param.rnr_retry_count = rnr_retry
*/
  __pyx_t_6 = __Pyx_PyLong_As_uint8_t(__pyx_v_flow_control); if (unlikely((__pyx_t_6 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 57, __pyx_L1_error)
  __pyx_v_self->conn_param.flow_control = __pyx_t_6;

  /* "pyverbs/cmid.pyx":58
 *         self.conn_param.initiator_depth = depth
 *         self.conn_param.flow_control = flow_control
 *         self.conn_param.retry_count = retry             # <<<<<<<<<<<<<<
 *         self.conn_param.rnr_retry_count = rnr_retry
 *         self.conn_param.srq = srq
*/
  __pyx_t_6 = __Pyx_PyLong_As_uint8_t(__pyx_v_retry); if (unlikely((__pyx_t_6 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 58, __pyx_L1_error)
  __pyx_v_self->conn_param.retry_count = __pyx_t_6;

  /* "pyverbs/cmid.pyx":59
 *         self.conn_param.flow_control = flow_control
 *         self.conn_param.retry_count = retry
 *         self.conn_param.rnr_retry_count = rnr_retry             # <<<<<<<<<<<<<<
 *         self.conn_param.srq = srq
 *         self.conn_param.qp_num = qp_num
*/
  __pyx_t_6 = __Pyx_PyLong_As_uint8_t(__pyx_v_rnr_retry); if (unlikely((__pyx_t_6 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 59, __pyx_L1_error)
  __pyx_v_self->conn_param.rnr_retry_count = __pyx_t_6;

  /* "pyverbs/cmid.pyx":60
 *         self.conn_param.retry_count = retry
 *         self.conn_param.rnr_retry_count = rnr_retry
 *         self.conn_param.srq = srq             # <<<<<<<<<<<<<<
 *         self.conn_param.qp_num = qp_num
 *         self.data = (ctypes.c_char * data_len)()
*/
  __pyx_t_6 = __Pyx_PyLong_As_uint8_t(__pyx_v_srq); if (unlikely((__pyx_t_6 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 60, __pyx_L1_error)
  __pyx_v_self->conn_param.srq = __pyx_t_6;

  /* "pyverbs/cmid.pyx":61
 *         self.conn_param.rnr_retry_count = rnr_retry
 *         self.conn_param.srq = srq
 *         self.conn_param.qp_num = qp_num             # <<<<<<<<<<<<<<
 *         self.data = (ctypes.c_char * data_len)()
 * 
*/
  __pyx_t_7 = __Pyx_PyLong_As_uint32_t(__pyx_v_qp_num); if (unlikely((__pyx_t_7 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L1_error)
  __pyx_v_self->conn_param.qp_num = __pyx_t_7;

  /* "pyverbs/cmid.pyx":62
 *         self.conn_param.srq = srq
 *         self.conn_param.qp_num = qp_num
 *         self.data = (ctypes.c_char * data_len)()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_3 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_ctypes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_c_char); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Multiply(__pyx_t_4, __pyx_v_data_len); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    assert(__pyx_t_3);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->data);
  __Pyx_DECREF(__pyx_v_self->data);
  __pyx_v_self->data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":23
 * cdef class ConnParam(PyverbsObject):
 * 
 *     def __init__(self, resources=1, depth=1, flow_control=0, retry=5,             # <<<<<<<<<<<<<<
 *                  rnr_retry=5, srq=0, qp_num=0, data_len=0):
 *         """
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cmid.ConnParam.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":64
 *         self.data = (ctypes.c_char * data_len)()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qpn(self):
 *         return self.conn_param.qp_num
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_9ConnParam_3qpn_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_4cmid_9ConnParam_3qpn_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_4cmid_9ConnParam_3qpn___get__(((struct __pyx_obj_7pyverbs_4cmid_ConnParam *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_9ConnParam_3qpn___get__(struct __pyx_obj_7pyverbs_4cmid_ConnParam *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cmid.pyx":66
 *     @property
 *     def qpn(self):
 *         return self.conn_param.qp_num             # <<<<<<<<<<<<<<
 *     @qpn.setter
 *     def qpn(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_uint32_t(__pyx_v_self->conn_param.qp_num); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":64
 *         self.data = (ctypes.c_char * data_len)()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qpn(self):
 *         return self.conn_param.qp_num
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cmid.ConnParam.qpn.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":67
 *     def qpn(self):
 *         return self.conn_param.qp_num
 *     @qpn.setter             # <<<<<<<<<<<<<<
 *     def qpn(self, val):
 *         self.conn_param.qp_num = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_4cmid_9ConnParam_3qpn_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_4cmid_9ConnParam_3qpn_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_4cmid_9ConnParam_3qpn_2__set__(((struct __pyx_obj_7pyverbs_4cmid_ConnParam *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_4cmid_9ConnParam_3qpn_2__set__(struct __pyx_obj_7pyverbs_4cmid_ConnParam *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  uint32_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cmid.pyx":69
 *     @qpn.setter
 *     def qpn(self, val):
 *         self.conn_param.qp_num = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_uint32_t(__pyx_v_val); if (unlikely((__pyx_t_1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 69, __pyx_L1_error)
  __pyx_v_self->conn_param.qp_num = __pyx_t_1;

  /* "pyverbs/cmid.pyx":67
 *     def qpn(self):
 *         return self.conn_param.qp_num
 *     @qpn.setter             # <<<<<<<<<<<<<<
 *     def qpn(self, val):
 *         self.conn_param.qp_num = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cmid.ConnParam.qpn.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":71
 *         self.conn_param.qp_num = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def private_data(self):
 *         data_array = bytearray(self.conn_param.private_data_len)
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_9ConnParam_12private_data_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_4cmid_9ConnParam_12private_data_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_4cmid_9ConnParam_12private_data___get__(((struct __pyx_obj_7pyverbs_4cmid_ConnParam *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_9ConnParam_12private_data___get__(struct __pyx_obj_7pyverbs_4cmid_ConnParam *__pyx_v_self) {
  PyObject *__pyx_v_data_array = NULL;
  unsigned char const *__pyx_v_p;
  uint8_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  uint8_t __pyx_t_5;
  uint8_t __pyx_t_6;
  uint8_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cmid.pyx":73
 *     @property
 *     def private_data(self):
 *         data_array = bytearray(self.conn_param.private_data_len)             # <<<<<<<<<<<<<<
 *         cdef const unsigned char *p = <const unsigned char*>self.conn_param.private_data
 *         for i in range(self.conn_param.private_data_len):
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = __Pyx_PyLong_From_uint8_t(__pyx_v_self->conn_param.private_data_len); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(&PyByteArray_Type), __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_data_array = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":74
 *     def private_data(self):
 *         data_array = bytearray(self.conn_param.private_data_len)
 *         cdef const unsigned char *p = <const unsigned char*>self.conn_param.private_data             # <<<<<<<<<<<<<<
 *         for i in range(self.conn_param.private_data_len):
 *             data_array[i] = p[i]
*/
  __pyx_v_p = ((unsigned char const *)__pyx_v_self->conn_param.private_data);

  /* "pyverbs/cmid.pyx":75
 *         data_array = bytearray(self.conn_param.private_data_len)
 *         cdef const unsigned char *p = <const unsigned char*>self.conn_param.private_data
 *         for i in range(self.conn_param.private_data_len):             # <<<<<<<<<<<<<<
 *             data_array[i] = p[i]
 *         return bytes(data_array)
*/
  __pyx_t_5 = __pyx_v_self->conn_param.private_data_len;
  __pyx_t_6 = __pyx_t_5;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "pyverbs/cmid.pyx":76
 *         cdef const unsigned char *p = <const unsigned char*>self.conn_param.private_data
 *         for i in range(self.conn_param.private_data_len):
 *             data_array[i] = p[i]             # <<<<<<<<<<<<<<
 *         return bytes(data_array)
 * 
*/
    if (unlikely((__pyx_v_p[__pyx_v_i]) > 255)) {
      PyErr_SetString(PyExc_ValueError, "byte must be in range(0, 256)"); __PYX_ERR(0, 76, __pyx_L1_error)
    }
    if (unlikely((__Pyx_SetItemInt_ByteArray(__pyx_v_data_array, __pyx_v_i, ((unsigned char)(__pyx_v_p[__pyx_v_i])), uint8_t, 0, __Pyx_PyLong_From_uint8_t, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference) < 0))) __PYX_ERR(0, 76, __pyx_L1_error)
  }

  /* "pyverbs/cmid.pyx":77
 *         for i in range(self.conn_param.private_data_len):
 *             data_array[i] = p[i]
 *         return bytes(data_array)             # <<<<<<<<<<<<<<
 * 
 *     def set_private_data(self, data):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_data_array};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(&PyBytes_Type), __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":71
 *         self.conn_param.qp_num = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def private_data(self):
 *         data_array = bytearray(self.conn_param.private_data_len)
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyverbs.cmid.ConnParam.private_data.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data_array);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":79
 *         return bytes(data_array)
 * 
 *     def set_private_data(self, data):             # <<<<<<<<<<<<<<
 *         if (min(len(self.data), len(data)) == 0):
 *             return
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_9ConnParam_3set_private_data(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_9ConnParam_3set_private_data = {"set_private_data", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_9ConnParam_3set_private_data, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_4cmid_9ConnParam_3set_private_data(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_private_data (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_data,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 79, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 79, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "set_private_data", 0) < (0)) __PYX_ERR(0, 79, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("set_private_data", 1, 1, 1, i); __PYX_ERR(0, 79, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 79, __pyx_L3_error)
    }
    __pyx_v_data = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_private_data", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 79, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.ConnParam.set_private_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_4cmid_9ConnParam_2set_private_data(((struct __pyx_obj_7pyverbs_4cmid_ConnParam *)__pyx_v_self), __pyx_v_data);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_9ConnParam_2set_private_data(struct __pyx_obj_7pyverbs_4cmid_ConnParam *__pyx_v_self, PyObject *__pyx_v_data) {
  Py_ssize_t __pyx_v_i;
  size_t __pyx_v_ptr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  size_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_private_data", 0);

  /* "pyverbs/cmid.pyx":80
 * 
 *     def set_private_data(self, data):
 *         if (min(len(self.data), len(data)) == 0):             # <<<<<<<<<<<<<<
 *             return
 * 
*/
  __pyx_t_1 = PyObject_Length(__pyx_v_data); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 80, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_self->data;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = (__pyx_t_1 < __pyx_t_3);
  if (__pyx_t_5) {
    __pyx_t_4 = __pyx_t_1;
  } else {
    __pyx_t_4 = __pyx_t_3;
  }
  __pyx_t_5 = (__pyx_t_4 == 0);
  if (__pyx_t_5) {

    /* "pyverbs/cmid.pyx":81
 *     def set_private_data(self, data):
 *         if (min(len(self.data), len(data)) == 0):
 *             return             # <<<<<<<<<<<<<<
 * 
 *         for i in range(len(self.data)):
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "pyverbs/cmid.pyx":80
 * 
 *     def set_private_data(self, data):
 *         if (min(len(self.data), len(data)) == 0):             # <<<<<<<<<<<<<<
 *             return
 * 
*/
  }

  /* "pyverbs/cmid.pyx":83
 *             return
 * 
 *         for i in range(len(self.data)):             # <<<<<<<<<<<<<<
 *             self.data[i] = 0
 * 
*/
  __pyx_t_2 = __pyx_v_self->data;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_4 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 83, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_4;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_1; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "pyverbs/cmid.pyx":84
 * 
 *         for i in range(len(self.data)):
 *             self.data[i] = 0             # <<<<<<<<<<<<<<
 * 
 *         for i in range(min(len(self.data), len(data))):
*/
    if (unlikely((__Pyx_SetItemInt(__pyx_v_self->data, __pyx_v_i, __pyx_mstate_global->__pyx_int_0, Py_ssize_t, 1, PyLong_FromSsize_t, 0, 1, 1, 1, __Pyx_ReferenceSharing_SharedReference) < 0))) __PYX_ERR(0, 84, __pyx_L1_error)
  }

  /* "pyverbs/cmid.pyx":86
 *             self.data[i] = 0
 * 
 *         for i in range(min(len(self.data), len(data))):             # <<<<<<<<<<<<<<
 *             self.data[i] = data[i]
 * 
*/
  __pyx_t_4 = PyObject_Length(__pyx_v_data); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 86, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_self->data;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = (__pyx_t_4 < __pyx_t_1);
  if (__pyx_t_5) {
    __pyx_t_3 = __pyx_t_4;
  } else {
    __pyx_t_3 = __pyx_t_1;
  }
  __pyx_t_4 = __pyx_t_3;
  __pyx_t_3 = __pyx_t_4;
  for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "pyverbs/cmid.pyx":87
 * 
 *         for i in range(min(len(self.data), len(data))):
 *             self.data[i] = data[i]             # <<<<<<<<<<<<<<
 * 
 *         cdef size_t ptr = ctypes.addressof(self.data)
*/
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_data, __pyx_v_i, Py_ssize_t, 1, PyLong_FromSsize_t, 0, 1, 1, 1, __Pyx_ReferenceSharing_FunctionArgument); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely((__Pyx_SetItemInt(__pyx_v_self->data, __pyx_v_i, __pyx_t_2, Py_ssize_t, 1, PyLong_FromSsize_t, 0, 1, 1, 1, __Pyx_ReferenceSharing_SharedReference) < 0))) __PYX_ERR(0, 87, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "pyverbs/cmid.pyx":89
 *             self.data[i] = data[i]
 * 
 *         cdef size_t ptr = ctypes.addressof(self.data)             # <<<<<<<<<<<<<<
 *         self.conn_param.private_data = <const void*>ptr
 *         self.conn_param.private_data_len = min(len(self.data), len(data))
*/
  __pyx_t_6 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_ctypes); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_addressof); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_9 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_8);
    assert(__pyx_t_6);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_8);
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_8, __pyx__function);
    __pyx_t_9 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_v_self->data};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_8, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  __pyx_t_9 = __Pyx_PyLong_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_9 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_ptr = __pyx_t_9;

  /* "pyverbs/cmid.pyx":90
 * 
 *         cdef size_t ptr = ctypes.addressof(self.data)
 *         self.conn_param.private_data = <const void*>ptr             # <<<<<<<<<<<<<<
 *         self.conn_param.private_data_len = min(len(self.data), len(data))
 * 
*/
  __pyx_v_self->conn_param.private_data = ((void const *)__pyx_v_ptr);

  /* "pyverbs/cmid.pyx":91
 *         cdef size_t ptr = ctypes.addressof(self.data)
 *         self.conn_param.private_data = <const void*>ptr
 *         self.conn_param.private_data_len = min(len(self.data), len(data))             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
*/
  __pyx_t_4 = PyObject_Length(__pyx_v_data); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 91, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_self->data;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 91, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = (__pyx_t_4 < __pyx_t_3);
  if (__pyx_t_5) {
    __pyx_t_1 = __pyx_t_4;
  } else {
    __pyx_t_1 = __pyx_t_3;
  }
  __pyx_v_self->conn_param.private_data_len = __pyx_t_1;

  /* "pyverbs/cmid.pyx":79
 *         return bytes(data_array)
 * 
 *     def set_private_data(self, data):             # <<<<<<<<<<<<<<
 *         if (min(len(self.data), len(data)) == 0):
 *             return
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyverbs.cmid.ConnParam.set_private_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":93
 *         self.conn_param.private_data_len = min(len(self.data), len(data))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         print_format  = '{:<4}: {:<4}\n'
 *         return '{}: {}\n'.format('Connection parameters', "") +\
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_9ConnParam_5__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_4cmid_9ConnParam_5__str__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_4cmid_9ConnParam_4__str__(((struct __pyx_obj_7pyverbs_4cmid_ConnParam *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_9ConnParam_4__str__(struct __pyx_obj_7pyverbs_4cmid_ConnParam *__pyx_v_self) {
  PyObject *__pyx_v_print_format = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "pyverbs/cmid.pyx":94
 * 
 *     def __str__(self):
 *         print_format  = '{:<4}: {:<4}\n'             # <<<<<<<<<<<<<<
 *         return '{}: {}\n'.format('Connection parameters', "") +\
 *                print_format.format('responder resources', self.conn_param.responder_resources) +\
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u_4_4);
  __pyx_v_print_format = __pyx_mstate_global->__pyx_kp_u_4_4;

  /* "pyverbs/cmid.pyx":95
 *     def __str__(self):
 *         print_format  = '{:<4}: {:<4}\n'
 *         return '{}: {}\n'.format('Connection parameters', "") +\             # <<<<<<<<<<<<<<
 *                print_format.format('responder resources', self.conn_param.responder_resources) +\
 *                print_format.format('initiator depth', self.conn_param.initiator_depth) +\
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_mstate_global->__pyx_kp_u__2, __pyx_mstate_global->__pyx_n_u_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":96
 *         print_format  = '{:<4}: {:<4}\n'
 *         return '{}: {}\n'.format('Connection parameters', "") +\
 *                print_format.format('responder resources', self.conn_param.responder_resources) +\             # <<<<<<<<<<<<<<
 *                print_format.format('initiator depth', self.conn_param.initiator_depth) +\
 *                print_format.format('flow control', self.conn_param.flow_control) +\
*/
  __pyx_t_1 = __Pyx_PyLong_From_uint8_t(__pyx_v_self->conn_param.responder_resources); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_responder_resources, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":95
 *     def __str__(self):
 *         print_format  = '{:<4}: {:<4}\n'
 *         return '{}: {}\n'.format('Connection parameters', "") +\             # <<<<<<<<<<<<<<
 *                print_format.format('responder resources', self.conn_param.responder_resources) +\
 *                print_format.format('initiator depth', self.conn_param.initiator_depth) +\
*/
  __pyx_t_1 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/cmid.pyx":97
 *         return '{}: {}\n'.format('Connection parameters', "") +\
 *                print_format.format('responder resources', self.conn_param.responder_resources) +\
 *                print_format.format('initiator depth', self.conn_param.initiator_depth) +\             # <<<<<<<<<<<<<<
 *                print_format.format('flow control', self.conn_param.flow_control) +\
 *                print_format.format('retry count', self.conn_param.retry_count) +\
*/
  __pyx_t_3 = __Pyx_PyLong_From_uint8_t(__pyx_v_self->conn_param.initiator_depth); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_initiator_depth, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/cmid.pyx":96
 *         print_format  = '{:<4}: {:<4}\n'
 *         return '{}: {}\n'.format('Connection parameters', "") +\
 *                print_format.format('responder resources', self.conn_param.responder_resources) +\             # <<<<<<<<<<<<<<
 *                print_format.format('initiator depth', self.conn_param.initiator_depth) +\
 *                print_format.format('flow control', self.conn_param.flow_control) +\
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":98
 *                print_format.format('responder resources', self.conn_param.responder_resources) +\
 *                print_format.format('initiator depth', self.conn_param.initiator_depth) +\
 *                print_format.format('flow control', self.conn_param.flow_control) +\             # <<<<<<<<<<<<<<
 *                print_format.format('retry count', self.conn_param.retry_count) +\
 *                print_format.format('rnr retry count', self.conn_param.rnr_retry_count) +\
*/
  __pyx_t_2 = __Pyx_PyLong_From_uint8_t(__pyx_v_self->conn_param.flow_control); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_flow_control_2, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":97
 *         return '{}: {}\n'.format('Connection parameters', "") +\
 *                print_format.format('responder resources', self.conn_param.responder_resources) +\
 *                print_format.format('initiator depth', self.conn_param.initiator_depth) +\             # <<<<<<<<<<<<<<
 *                print_format.format('flow control', self.conn_param.flow_control) +\
 *                print_format.format('retry count', self.conn_param.retry_count) +\
*/
  __pyx_t_2 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":99
 *                print_format.format('initiator depth', self.conn_param.initiator_depth) +\
 *                print_format.format('flow control', self.conn_param.flow_control) +\
 *                print_format.format('retry count', self.conn_param.retry_count) +\             # <<<<<<<<<<<<<<
 *                print_format.format('rnr retry count', self.conn_param.rnr_retry_count) +\
 *                print_format.format('srq', self.conn_param.srq) +\
*/
  __pyx_t_1 = __Pyx_PyLong_From_uint8_t(__pyx_v_self->conn_param.retry_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_retry_count, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":98
 *                print_format.format('responder resources', self.conn_param.responder_resources) +\
 *                print_format.format('initiator depth', self.conn_param.initiator_depth) +\
 *                print_format.format('flow control', self.conn_param.flow_control) +\             # <<<<<<<<<<<<<<
 *                print_format.format('retry count', self.conn_param.retry_count) +\
 *                print_format.format('rnr retry count', self.conn_param.rnr_retry_count) +\
*/
  __pyx_t_1 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/cmid.pyx":100
 *                print_format.format('flow control', self.conn_param.flow_control) +\
 *                print_format.format('retry count', self.conn_param.retry_count) +\
 *                print_format.format('rnr retry count', self.conn_param.rnr_retry_count) +\             # <<<<<<<<<<<<<<
 *                print_format.format('srq', self.conn_param.srq) +\
 *                print_format.format('qp number', self.conn_param.qp_num)
*/
  __pyx_t_3 = __Pyx_PyLong_From_uint8_t(__pyx_v_self->conn_param.rnr_retry_count); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_rnr_retry_count, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyverbs/cmid.pyx":99
 *                print_format.format('initiator depth', self.conn_param.initiator_depth) +\
 *                print_format.format('flow control', self.conn_param.flow_control) +\
 *                print_format.format('retry count', self.conn_param.retry_count) +\             # <<<<<<<<<<<<<<
 *                print_format.format('rnr retry count', self.conn_param.rnr_retry_count) +\
 *                print_format.format('srq', self.conn_param.srq) +\
*/
  __pyx_t_3 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":101
 *                print_format.format('retry count', self.conn_param.retry_count) +\
 *                print_format.format('rnr retry count', self.conn_param.rnr_retry_count) +\
 *                print_format.format('srq', self.conn_param.srq) +\             # <<<<<<<<<<<<<<
 *                print_format.format('qp number', self.conn_param.qp_num)
 * 
*/
  __pyx_t_2 = __Pyx_PyLong_From_uint8_t(__pyx_v_self->conn_param.srq); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_n_u_srq, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":100
 *                print_format.format('flow control', self.conn_param.flow_control) +\
 *                print_format.format('retry count', self.conn_param.retry_count) +\
 *                print_format.format('rnr retry count', self.conn_param.rnr_retry_count) +\             # <<<<<<<<<<<<<<
 *                print_format.format('srq', self.conn_param.srq) +\
 *                print_format.format('qp number', self.conn_param.qp_num)
*/
  __pyx_t_2 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":102
 *                print_format.format('rnr retry count', self.conn_param.rnr_retry_count) +\
 *                print_format.format('srq', self.conn_param.srq) +\
 *                print_format.format('qp number', self.conn_param.qp_num)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __Pyx_PyLong_From_uint32_t(__pyx_v_self->conn_param.qp_num); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyUnicode_Type__format, __pyx_v_print_format, __pyx_mstate_global->__pyx_kp_u_qp_number, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":101
 *                print_format.format('retry count', self.conn_param.retry_count) +\
 *                print_format.format('rnr retry count', self.conn_param.rnr_retry_count) +\
 *                print_format.format('srq', self.conn_param.srq) +\             # <<<<<<<<<<<<<<
 *                print_format.format('qp number', self.conn_param.qp_num)
 * 
*/
  __pyx_t_1 = __Pyx_PyUnicode_Concat__Pyx_ReferenceSharing_OwnStrongReferenceInPlace(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":93
 *         self.conn_param.private_data_len = min(len(self.data), len(data))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         print_format  = '{:<4}: {:<4}\n'
 *         return '{}: {}\n'.format('Connection parameters', "") +\
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyverbs.cmid.ConnParam.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_print_format);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.conn_param cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_9ConnParam_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_9ConnParam_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_9ConnParam_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_4cmid_9ConnParam_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_4cmid_9ConnParam_6__reduce_cython__(((struct __pyx_obj_7pyverbs_4cmid_ConnParam *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_9ConnParam_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_ConnParam *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.conn_param cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.conn_param cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_conn_param_cannot_be_conver, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.conn_param cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cmid.ConnParam.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.conn_param cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.conn_param cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_9ConnParam_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_9ConnParam_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_9ConnParam_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_4cmid_9ConnParam_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.ConnParam.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_4cmid_9ConnParam_8__setstate_cython__(((struct __pyx_obj_7pyverbs_4cmid_ConnParam *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_9ConnParam_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_ConnParam *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.conn_param cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.conn_param cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_conn_param_cannot_be_conver, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.conn_param cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.conn_param cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cmid.ConnParam.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":107
 * cdef class JoinMCAttrEx(PyverbsObject):
 * 
 *     def __init__(self, AddrInfo addr not None, comp_mask=0, join_flags=0):             # <<<<<<<<<<<<<<
 *         """
 *         Initialize a JoinMCAttrEx object over an underlying
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_12JoinMCAttrEx___init__, "\n        Initialize a JoinMCAttrEx object over an underlying\n        rdma_cm_join_mc_attr_ex C object which contains the extended join\n        multicast attributes.\n        :param addr: Multicast address identifying the group to join.\n        :param comp_mask: Bitwise OR between \"rdma_cm_join_mc_attr_mask\" enum.\n        :param join_flags: Single flag from \"rdma_cm_mc_join_flags\" enum.\n                           Indicates the type of the join requests.\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_7pyverbs_4cmid_12JoinMCAttrEx___init__;
#endif
static int __pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_addr = 0;
  PyObject *__pyx_v_comp_mask = 0;
  PyObject *__pyx_v_join_flags = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_addr,&__pyx_mstate_global->__pyx_n_u_comp_mask,&__pyx_mstate_global->__pyx_n_u_join_flags,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 107, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 107, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 107, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 107, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 107, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 3, i); __PYX_ERR(0, 107, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 107, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 107, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 107, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    }
    __pyx_v_addr = ((struct __pyx_obj_7pyverbs_4cmid_AddrInfo *)values[0]);
    __pyx_v_comp_mask = values[1];
    __pyx_v_join_flags = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 3, __pyx_nargs); __PYX_ERR(0, 107, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.JoinMCAttrEx.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_addr), __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_AddrInfo, 0, "addr", 0))) __PYX_ERR(0, 107, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_4cmid_12JoinMCAttrEx___init__(((struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *)__pyx_v_self), __pyx_v_addr, __pyx_v_comp_mask, __pyx_v_join_flags);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_4cmid_12JoinMCAttrEx___init__(struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *__pyx_v_self, struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_addr, PyObject *__pyx_v_comp_mask, PyObject *__pyx_v_join_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  struct sockaddr *__pyx_t_6;
  uint32_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyverbs/cmid.pyx":117
 *                            Indicates the type of the join requests.
 *         """
 *         super().__init__()             # <<<<<<<<<<<<<<
 *         self.join_mc_attr_ex.addr = addr.addr_info.ai_src_addr
 *         self.join_mc_attr_ex.comp_mask = comp_mask
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":118
 *         """
 *         super().__init__()
 *         self.join_mc_attr_ex.addr = addr.addr_info.ai_src_addr             # <<<<<<<<<<<<<<
 *         self.join_mc_attr_ex.comp_mask = comp_mask
 *         self.join_mc_attr_ex.join_flags = join_flags
*/
  __pyx_t_6 = __pyx_v_addr->addr_info->ai_src_addr;
  __pyx_v_self->join_mc_attr_ex.addr = __pyx_t_6;

  /* "pyverbs/cmid.pyx":119
 *         super().__init__()
 *         self.join_mc_attr_ex.addr = addr.addr_info.ai_src_addr
 *         self.join_mc_attr_ex.comp_mask = comp_mask             # <<<<<<<<<<<<<<
 *         self.join_mc_attr_ex.join_flags = join_flags
 * 
*/
  __pyx_t_7 = __Pyx_PyLong_As_uint32_t(__pyx_v_comp_mask); if (unlikely((__pyx_t_7 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 119, __pyx_L1_error)
  __pyx_v_self->join_mc_attr_ex.comp_mask = __pyx_t_7;

  /* "pyverbs/cmid.pyx":120
 *         self.join_mc_attr_ex.addr = addr.addr_info.ai_src_addr
 *         self.join_mc_attr_ex.comp_mask = comp_mask
 *         self.join_mc_attr_ex.join_flags = join_flags             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_7 = __Pyx_PyLong_As_uint32_t(__pyx_v_join_flags); if (unlikely((__pyx_t_7 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 120, __pyx_L1_error)
  __pyx_v_self->join_mc_attr_ex.join_flags = __pyx_t_7;

  /* "pyverbs/cmid.pyx":107
 * cdef class JoinMCAttrEx(PyverbsObject):
 * 
 *     def __init__(self, AddrInfo addr not None, comp_mask=0, join_flags=0):             # <<<<<<<<<<<<<<
 *         """
 *         Initialize a JoinMCAttrEx object over an underlying
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cmid.JoinMCAttrEx.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":122
 *         self.join_mc_attr_ex.join_flags = join_flags
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def join_flags(self):
 *         return self.join_mc_attr_ex.join_flags
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_10join_flags_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_10join_flags_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_4cmid_12JoinMCAttrEx_10join_flags___get__(((struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_12JoinMCAttrEx_10join_flags___get__(struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cmid.pyx":124
 *     @property
 *     def join_flags(self):
 *         return self.join_mc_attr_ex.join_flags             # <<<<<<<<<<<<<<
 *     @join_flags.setter
 *     def join_flags(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_uint32_t(__pyx_v_self->join_mc_attr_ex.join_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":122
 *         self.join_mc_attr_ex.join_flags = join_flags
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def join_flags(self):
 *         return self.join_mc_attr_ex.join_flags
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cmid.JoinMCAttrEx.join_flags.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":125
 *     def join_flags(self):
 *         return self.join_mc_attr_ex.join_flags
 *     @join_flags.setter             # <<<<<<<<<<<<<<
 *     def join_flags(self, val):
 *         self.join_mc_attr_ex.join_flags = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_10join_flags_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_10join_flags_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_4cmid_12JoinMCAttrEx_10join_flags_2__set__(((struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_4cmid_12JoinMCAttrEx_10join_flags_2__set__(struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  uint32_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cmid.pyx":127
 *     @join_flags.setter
 *     def join_flags(self, val):
 *         self.join_mc_attr_ex.join_flags = val             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_1 = __Pyx_PyLong_As_uint32_t(__pyx_v_val); if (unlikely((__pyx_t_1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 127, __pyx_L1_error)
  __pyx_v_self->join_mc_attr_ex.join_flags = __pyx_t_1;

  /* "pyverbs/cmid.pyx":125
 *     def join_flags(self):
 *         return self.join_mc_attr_ex.join_flags
 *     @join_flags.setter             # <<<<<<<<<<<<<<
 *     def join_flags(self, val):
 *         self.join_mc_attr_ex.join_flags = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cmid.JoinMCAttrEx.join_flags.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":129
 *         self.join_mc_attr_ex.join_flags = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def comp_mask(self):
 *         return self.join_mc_attr_ex.comp_mask
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_9comp_mask_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_9comp_mask_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_4cmid_12JoinMCAttrEx_9comp_mask___get__(((struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_12JoinMCAttrEx_9comp_mask___get__(struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cmid.pyx":131
 *     @property
 *     def comp_mask(self):
 *         return self.join_mc_attr_ex.comp_mask             # <<<<<<<<<<<<<<
 *     @comp_mask.setter
 *     def comp_mask(self, val):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_uint32_t(__pyx_v_self->join_mc_attr_ex.comp_mask); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":129
 *         self.join_mc_attr_ex.join_flags = val
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def comp_mask(self):
 *         return self.join_mc_attr_ex.comp_mask
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cmid.JoinMCAttrEx.comp_mask.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":132
 *     def comp_mask(self):
 *         return self.join_mc_attr_ex.comp_mask
 *     @comp_mask.setter             # <<<<<<<<<<<<<<
 *     def comp_mask(self, val):
 *         self.join_mc_attr_ex.comp_mask = val
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_9comp_mask_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val); /*proto*/
static int __pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_9comp_mask_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_val) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_4cmid_12JoinMCAttrEx_9comp_mask_2__set__(((struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *)__pyx_v_self), ((PyObject *)__pyx_v_val));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_4cmid_12JoinMCAttrEx_9comp_mask_2__set__(struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *__pyx_v_self, PyObject *__pyx_v_val) {
  int __pyx_r;
  uint32_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "pyverbs/cmid.pyx":134
 *     @comp_mask.setter
 *     def comp_mask(self, val):
 *         self.join_mc_attr_ex.comp_mask = val             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __Pyx_PyLong_As_uint32_t(__pyx_v_val); if (unlikely((__pyx_t_1 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 134, __pyx_L1_error)
  __pyx_v_self->join_mc_attr_ex.comp_mask = __pyx_t_1;

  /* "pyverbs/cmid.pyx":132
 *     def comp_mask(self):
 *         return self.join_mc_attr_ex.comp_mask
 *     @comp_mask.setter             # <<<<<<<<<<<<<<
 *     def comp_mask(self, val):
 *         self.join_mc_attr_ex.comp_mask = val
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cmid.JoinMCAttrEx.comp_mask.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.join_mc_attr_ex cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_12JoinMCAttrEx_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_4cmid_12JoinMCAttrEx_2__reduce_cython__(((struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_12JoinMCAttrEx_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.join_mc_attr_ex cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.join_mc_attr_ex cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_join_mc_attr_ex_cannot_be_c, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.join_mc_attr_ex cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cmid.JoinMCAttrEx.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.join_mc_attr_ex cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.join_mc_attr_ex cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_12JoinMCAttrEx_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.JoinMCAttrEx.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_4cmid_12JoinMCAttrEx_4__setstate_cython__(((struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_12JoinMCAttrEx_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.join_mc_attr_ex cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.join_mc_attr_ex cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_join_mc_attr_ex_cannot_be_c, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.join_mc_attr_ex cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.join_mc_attr_ex cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cmid.JoinMCAttrEx.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":139
 * cdef class UDParam(PyverbsObject):
 * 
 *     def __init__(self, CMEvent cm_event not None):             # <<<<<<<<<<<<<<
 *         """
 *         Initialize a UDParam object over an underlying rdma_ud_param
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_4cmid_7UDParam_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_7UDParam___init__, "\n        Initialize a UDParam object over an underlying rdma_ud_param\n        C object which contains UD connection parameters.\n        :param cm_event: The creator of UDParam. When the active side gets\n                         connection establishment event, the event contains\n                         UDParam for the passive CMID details.\n        :return: UDParam object\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_7pyverbs_4cmid_7UDParam___init__;
#endif
static int __pyx_pw_7pyverbs_4cmid_7UDParam_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7pyverbs_4cmid_CMEvent *__pyx_v_cm_event = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_cm_event,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 139, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 139, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 139, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, i); __PYX_ERR(0, 139, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 139, __pyx_L3_error)
    }
    __pyx_v_cm_event = ((struct __pyx_obj_7pyverbs_4cmid_CMEvent *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 139, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.UDParam.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_cm_event), __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMEvent, 0, "cm_event", 0))) __PYX_ERR(0, 139, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_4cmid_7UDParam___init__(((struct __pyx_obj_7pyverbs_4cmid_UDParam *)__pyx_v_self), __pyx_v_cm_event);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_4cmid_7UDParam___init__(struct __pyx_obj_7pyverbs_4cmid_UDParam *__pyx_v_self, struct __pyx_obj_7pyverbs_4cmid_CMEvent *__pyx_v_cm_event) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  struct rdma_ud_param __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyverbs/cmid.pyx":148
 *         :return: UDParam object
 *         """
 *         super().__init__()             # <<<<<<<<<<<<<<
 *         memset(&self.ud_param, 0, sizeof(cm.rdma_ud_param))
 *         self.ud_param = (<CMEvent>cm_event).event.param.ud
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_UDParam), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":149
 *         """
 *         super().__init__()
 *         memset(&self.ud_param, 0, sizeof(cm.rdma_ud_param))             # <<<<<<<<<<<<<<
 *         self.ud_param = (<CMEvent>cm_event).event.param.ud
 * 
*/
  (void)(memset((&__pyx_v_self->ud_param), 0, (sizeof(struct rdma_ud_param))));

  /* "pyverbs/cmid.pyx":150
 *         super().__init__()
 *         memset(&self.ud_param, 0, sizeof(cm.rdma_ud_param))
 *         self.ud_param = (<CMEvent>cm_event).event.param.ud             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_6 = __pyx_v_cm_event->event->param.ud;
  __pyx_v_self->ud_param = __pyx_t_6;

  /* "pyverbs/cmid.pyx":139
 * cdef class UDParam(PyverbsObject):
 * 
 *     def __init__(self, CMEvent cm_event not None):             # <<<<<<<<<<<<<<
 *         """
 *         Initialize a UDParam object over an underlying rdma_ud_param
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cmid.UDParam.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":152
 *         self.ud_param = (<CMEvent>cm_event).event.param.ud
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qp_num(self):
 *         return self.ud_param.qp_num
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_7UDParam_6qp_num_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_4cmid_7UDParam_6qp_num_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_4cmid_7UDParam_6qp_num___get__(((struct __pyx_obj_7pyverbs_4cmid_UDParam *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_7UDParam_6qp_num___get__(struct __pyx_obj_7pyverbs_4cmid_UDParam *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cmid.pyx":154
 *     @property
 *     def qp_num(self):
 *         return self.ud_param.qp_num             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_uint32_t(__pyx_v_self->ud_param.qp_num); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":152
 *         self.ud_param = (<CMEvent>cm_event).event.param.ud
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qp_num(self):
 *         return self.ud_param.qp_num
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cmid.UDParam.qp_num.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":156
 *         return self.ud_param.qp_num
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qkey(self):
 *         return self.ud_param.qkey
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_7UDParam_4qkey_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_4cmid_7UDParam_4qkey_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_4cmid_7UDParam_4qkey___get__(((struct __pyx_obj_7pyverbs_4cmid_UDParam *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_7UDParam_4qkey___get__(struct __pyx_obj_7pyverbs_4cmid_UDParam *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cmid.pyx":158
 *     @property
 *     def qkey(self):
 *         return self.ud_param.qkey             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_uint32_t(__pyx_v_self->ud_param.qkey); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":156
 *         return self.ud_param.qp_num
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qkey(self):
 *         return self.ud_param.qkey
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cmid.UDParam.qkey.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":160
 *         return self.ud_param.qkey
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ah_attr(self):
 *         ah_attr = AHAttr()
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_7UDParam_7ah_attr_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_4cmid_7UDParam_7ah_attr_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_4cmid_7UDParam_7ah_attr___get__(((struct __pyx_obj_7pyverbs_4cmid_UDParam *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_7UDParam_7ah_attr___get__(struct __pyx_obj_7pyverbs_4cmid_UDParam *__pyx_v_self) {
  struct __pyx_obj_7pyverbs_4addr_AHAttr *__pyx_v_ah_attr = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cmid.pyx":162
 *     @property
 *     def ah_attr(self):
 *         ah_attr = AHAttr()             # <<<<<<<<<<<<<<
 *         ah_attr.init_from_ud_param(self)
 *         return ah_attr
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_4addr_AHAttr, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 162, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_ah_attr = ((struct __pyx_obj_7pyverbs_4addr_AHAttr *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":163
 *     def ah_attr(self):
 *         ah_attr = AHAttr()
 *         ah_attr.init_from_ud_param(self)             # <<<<<<<<<<<<<<
 *         return ah_attr
 * 
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_7pyverbs_4addr_AHAttr *)__pyx_v_ah_attr->__pyx_vtab)->init_from_ud_param(__pyx_v_ah_attr, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":164
 *         ah_attr = AHAttr()
 *         ah_attr.init_from_ud_param(self)
 *         return ah_attr             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_ah_attr);
  __pyx_r = ((PyObject *)__pyx_v_ah_attr);
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":160
 *         return self.ud_param.qkey
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ah_attr(self):
 *         ah_attr = AHAttr()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyverbs.cmid.UDParam.ah_attr.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_ah_attr);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ud_param cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_7UDParam_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_7UDParam_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_7UDParam_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_4cmid_7UDParam_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_4cmid_7UDParam_2__reduce_cython__(((struct __pyx_obj_7pyverbs_4cmid_UDParam *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_7UDParam_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_UDParam *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ud_param cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.ud_param cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_ud_param_cannot_be_converte, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ud_param cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cmid.UDParam.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ud_param cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ud_param cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_7UDParam_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_7UDParam_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_7UDParam_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_4cmid_7UDParam_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.UDParam.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_4cmid_7UDParam_4__setstate_cython__(((struct __pyx_obj_7pyverbs_4cmid_UDParam *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_7UDParam_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_UDParam *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.ud_param cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.ud_param cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_ud_param_cannot_be_converte, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ud_param cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ud_param cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cmid.UDParam.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":169
 * cdef class AddrInfo(PyverbsObject):
 * 
 *     def __init__(self, src=None, dst=None, src_service=None, dst_service=None,             # <<<<<<<<<<<<<<
 *                  port_space=0, flags=0):
 *         """
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_4cmid_8AddrInfo_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_8AddrInfo___init__, "\n        Initialize an AddrInfo object over an underlying rdma_addrinfo C object.\n        :param src: Name, dotted-decimal IPv4 or IPv6 hex address to bind to.\n        :param dst: Name, dotted-decimal IPv4 or IPv6 hex address to connect to.\n        :param src_service: The service name or port number of the source\n                            address.\n        :param dst_service: The service name or port number of the destination\n                            address.\n        :param port_space: RDMA port space used (RDMA_PS_UDP or RDMA_PS_TCP).\n        :param flags: Hint flags which control the operation.\n        :return: An AddrInfo object which contains information needed to\n        establish communication.\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_7pyverbs_4cmid_8AddrInfo___init__;
#endif
static int __pyx_pw_7pyverbs_4cmid_8AddrInfo_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_src = 0;
  PyObject *__pyx_v_dst = 0;
  PyObject *__pyx_v_src_service = 0;
  PyObject *__pyx_v_dst_service = 0;
  PyObject *__pyx_v_port_space = 0;
  PyObject *__pyx_v_flags = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_dst,&__pyx_mstate_global->__pyx_n_u_src_service,&__pyx_mstate_global->__pyx_n_u_dst_service,&__pyx_mstate_global->__pyx_n_u_port_space,&__pyx_mstate_global->__pyx_n_u_flags,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 169, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 169, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 169, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 169, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 169, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 169, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 169, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 169, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[5]) values[5] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    } else {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 169, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 169, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 169, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 169, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 169, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 169, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[5]) values[5] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    }
    __pyx_v_src = values[0];
    __pyx_v_dst = values[1];
    __pyx_v_src_service = values[2];
    __pyx_v_dst_service = values[3];
    __pyx_v_port_space = values[4];
    __pyx_v_flags = values[5];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 6, __pyx_nargs); __PYX_ERR(0, 169, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.AddrInfo.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_4cmid_8AddrInfo___init__(((struct __pyx_obj_7pyverbs_4cmid_AddrInfo *)__pyx_v_self), __pyx_v_src, __pyx_v_dst, __pyx_v_src_service, __pyx_v_dst_service, __pyx_v_port_space, __pyx_v_flags);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_4cmid_8AddrInfo___init__(struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_self, PyObject *__pyx_v_src, PyObject *__pyx_v_dst, PyObject *__pyx_v_src_service, PyObject *__pyx_v_dst_service, PyObject *__pyx_v_port_space, PyObject *__pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyverbs/cmid.pyx":184
 *         establish communication.
 *         """
 *         super().__init__()             # <<<<<<<<<<<<<<
 *         self.addr_info = NULL
 * 
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_AddrInfo), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":185
 *         """
 *         super().__init__()
 *         self.addr_info = NULL             # <<<<<<<<<<<<<<
 * 
 *         if src or dst or src_service or dst_service:
*/
  __pyx_v_self->addr_info = NULL;

  /* "pyverbs/cmid.pyx":187
 *         self.addr_info = NULL
 * 
 *         if src or dst or src_service or dst_service:             # <<<<<<<<<<<<<<
 *             self.get_addrinfo(src, dst, src_service, dst_service, port_space, flags)
 * 
*/
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_src); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 187, __pyx_L1_error)
  if (!__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_dst); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 187, __pyx_L1_error)
  if (!__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_src_service); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 187, __pyx_L1_error)
  if (!__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_dst_service); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 187, __pyx_L1_error)
  __pyx_t_6 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_6) {

    /* "pyverbs/cmid.pyx":188
 * 
 *         if src or dst or src_service or dst_service:
 *             self.get_addrinfo(src, dst, src_service, dst_service, port_space, flags)             # <<<<<<<<<<<<<<
 * 
 *     def get_addrinfo(self, src=None, dst=None, src_service=None, dst_service=None,
*/
    __pyx_t_3 = ((PyObject *)__pyx_v_self);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[7] = {__pyx_t_3, __pyx_v_src, __pyx_v_dst, __pyx_v_src_service, __pyx_v_dst_service, __pyx_v_port_space, __pyx_v_flags};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_get_addrinfo, __pyx_callargs+__pyx_t_5, (7-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 188, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyverbs/cmid.pyx":187
 *         self.addr_info = NULL
 * 
 *         if src or dst or src_service or dst_service:             # <<<<<<<<<<<<<<
 *             self.get_addrinfo(src, dst, src_service, dst_service, port_space, flags)
 * 
*/
  }

  /* "pyverbs/cmid.pyx":169
 * cdef class AddrInfo(PyverbsObject):
 * 
 *     def __init__(self, src=None, dst=None, src_service=None, dst_service=None,             # <<<<<<<<<<<<<<
 *                  port_space=0, flags=0):
 *         """
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cmid.AddrInfo.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":190
 *             self.get_addrinfo(src, dst, src_service, dst_service, port_space, flags)
 * 
 *     def get_addrinfo(self, src=None, dst=None, src_service=None, dst_service=None,             # <<<<<<<<<<<<<<
 *                      port_space=0, flags=0):
 *         """
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_8AddrInfo_3get_addrinfo(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_8AddrInfo_2get_addrinfo, "\n        Resolve source and destination addresses.\n        :param src: Name, dotted-decimal IPv4 or IPv6 hex address to bind to.\n        :param dst: Name, dotted-decimal IPv4 or IPv6 hex address to connect to.\n        :param src_service: The service name or port number of the source address.\n        :param dst_service: The service name or port number of the destination address.\n        :param port_space: RDMA port space used (RDMA_PS_UDP, RDMA_PS_TCP or RDMA_PS_IB).\n        :param flags: Hint flags which control the operation.\n        :raises PyverbsRDMAErrno: If fails to get Address Info.\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_8AddrInfo_3get_addrinfo = {"get_addrinfo", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_8AddrInfo_3get_addrinfo, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_8AddrInfo_2get_addrinfo};
static PyObject *__pyx_pw_7pyverbs_4cmid_8AddrInfo_3get_addrinfo(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_src = 0;
  PyObject *__pyx_v_dst = 0;
  PyObject *__pyx_v_src_service = 0;
  PyObject *__pyx_v_dst_service = 0;
  PyObject *__pyx_v_port_space = 0;
  PyObject *__pyx_v_flags = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[6] = {0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_addrinfo (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_src,&__pyx_mstate_global->__pyx_n_u_dst,&__pyx_mstate_global->__pyx_n_u_src_service,&__pyx_mstate_global->__pyx_n_u_dst_service,&__pyx_mstate_global->__pyx_n_u_port_space,&__pyx_mstate_global->__pyx_n_u_flags,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 190, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 190, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 190, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 190, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 190, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 190, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 190, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_addrinfo", 0) < (0)) __PYX_ERR(0, 190, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[5]) values[5] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    } else {
      switch (__pyx_nargs) {
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 190, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 190, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 190, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 190, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 190, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 190, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[5]) values[5] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    }
    __pyx_v_src = values[0];
    __pyx_v_dst = values[1];
    __pyx_v_src_service = values[2];
    __pyx_v_dst_service = values[3];
    __pyx_v_port_space = values[4];
    __pyx_v_flags = values[5];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_addrinfo", 0, 0, 6, __pyx_nargs); __PYX_ERR(0, 190, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.AddrInfo.get_addrinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_4cmid_8AddrInfo_2get_addrinfo(((struct __pyx_obj_7pyverbs_4cmid_AddrInfo *)__pyx_v_self), __pyx_v_src, __pyx_v_dst, __pyx_v_src_service, __pyx_v_dst_service, __pyx_v_port_space, __pyx_v_flags);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_8AddrInfo_2get_addrinfo(struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_self, PyObject *__pyx_v_src, PyObject *__pyx_v_dst, PyObject *__pyx_v_src_service, PyObject *__pyx_v_dst_service, PyObject *__pyx_v_port_space, PyObject *__pyx_v_flags) {
  char *__pyx_v_src_srvc;
  char *__pyx_v_dst_srvc;
  char *__pyx_v_src_addr;
  char *__pyx_v_dst_addr;
  struct rdma_addrinfo __pyx_v_hints;
  struct rdma_addrinfo *__pyx_v_hints_ptr;
  struct rdma_addrinfo *__pyx_v_res;
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  char *__pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_addrinfo", 0);
  __Pyx_INCREF(__pyx_v_src);
  __Pyx_INCREF(__pyx_v_dst);
  __Pyx_INCREF(__pyx_v_src_service);
  __Pyx_INCREF(__pyx_v_dst_service);

  /* "pyverbs/cmid.pyx":203
 *         :return: None
 *         """
 *         cdef char* src_srvc = NULL             # <<<<<<<<<<<<<<
 *         cdef char* dst_srvc = NULL
 *         cdef char* src_addr = NULL
*/
  __pyx_v_src_srvc = NULL;

  /* "pyverbs/cmid.pyx":204
 *         """
 *         cdef char* src_srvc = NULL
 *         cdef char* dst_srvc = NULL             # <<<<<<<<<<<<<<
 *         cdef char* src_addr = NULL
 *         cdef char* dst_addr = NULL
*/
  __pyx_v_dst_srvc = NULL;

  /* "pyverbs/cmid.pyx":205
 *         cdef char* src_srvc = NULL
 *         cdef char* dst_srvc = NULL
 *         cdef char* src_addr = NULL             # <<<<<<<<<<<<<<
 *         cdef char* dst_addr = NULL
 *         cdef cm.rdma_addrinfo hints
*/
  __pyx_v_src_addr = NULL;

  /* "pyverbs/cmid.pyx":206
 *         cdef char* dst_srvc = NULL
 *         cdef char* src_addr = NULL
 *         cdef char* dst_addr = NULL             # <<<<<<<<<<<<<<
 *         cdef cm.rdma_addrinfo hints
 *         cdef cm.rdma_addrinfo *hints_ptr = NULL
*/
  __pyx_v_dst_addr = NULL;

  /* "pyverbs/cmid.pyx":208
 *         cdef char* dst_addr = NULL
 *         cdef cm.rdma_addrinfo hints
 *         cdef cm.rdma_addrinfo *hints_ptr = NULL             # <<<<<<<<<<<<<<
 *         cdef cm.rdma_addrinfo *res = NULL
 * 
*/
  __pyx_v_hints_ptr = NULL;

  /* "pyverbs/cmid.pyx":209
 *         cdef cm.rdma_addrinfo hints
 *         cdef cm.rdma_addrinfo *hints_ptr = NULL
 *         cdef cm.rdma_addrinfo *res = NULL             # <<<<<<<<<<<<<<
 * 
 *         if src is not None:
*/
  __pyx_v_res = NULL;

  /* "pyverbs/cmid.pyx":211
 *         cdef cm.rdma_addrinfo *res = NULL
 * 
 *         if src is not None:             # <<<<<<<<<<<<<<
 *             if isinstance(src, str):
 *                 src = src.encode('utf-8')
*/
  __pyx_t_1 = (__pyx_v_src != Py_None);
  if (__pyx_t_1) {

    /* "pyverbs/cmid.pyx":212
 * 
 *         if src is not None:
 *             if isinstance(src, str):             # <<<<<<<<<<<<<<
 *                 src = src.encode('utf-8')
 *             src_addr = <char*>src
*/
    __pyx_t_1 = PyUnicode_Check(__pyx_v_src); 
    if (__pyx_t_1) {

      /* "pyverbs/cmid.pyx":213
 *         if src is not None:
 *             if isinstance(src, str):
 *                 src = src.encode('utf-8')             # <<<<<<<<<<<<<<
 *             src_addr = <char*>src
 *         if dst is not None:
*/
      __pyx_t_3 = __pyx_v_src;
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_utf_8};
        __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_encode, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 213, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_DECREF_SET(__pyx_v_src, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "pyverbs/cmid.pyx":212
 * 
 *         if src is not None:
 *             if isinstance(src, str):             # <<<<<<<<<<<<<<
 *                 src = src.encode('utf-8')
 *             src_addr = <char*>src
*/
    }

    /* "pyverbs/cmid.pyx":214
 *             if isinstance(src, str):
 *                 src = src.encode('utf-8')
 *             src_addr = <char*>src             # <<<<<<<<<<<<<<
 *         if dst is not None:
 *             if isinstance(dst, str):
*/
    __pyx_t_5 = __Pyx_PyObject_AsWritableString(__pyx_v_src); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(0, 214, __pyx_L1_error)
    __pyx_v_src_addr = ((char *)__pyx_t_5);

    /* "pyverbs/cmid.pyx":211
 *         cdef cm.rdma_addrinfo *res = NULL
 * 
 *         if src is not None:             # <<<<<<<<<<<<<<
 *             if isinstance(src, str):
 *                 src = src.encode('utf-8')
*/
  }

  /* "pyverbs/cmid.pyx":215
 *                 src = src.encode('utf-8')
 *             src_addr = <char*>src
 *         if dst is not None:             # <<<<<<<<<<<<<<
 *             if isinstance(dst, str):
 *                 dst = dst.encode('utf-8')
*/
  __pyx_t_1 = (__pyx_v_dst != Py_None);
  if (__pyx_t_1) {

    /* "pyverbs/cmid.pyx":216
 *             src_addr = <char*>src
 *         if dst is not None:
 *             if isinstance(dst, str):             # <<<<<<<<<<<<<<
 *                 dst = dst.encode('utf-8')
 *             dst_addr = <char*>dst
*/
    __pyx_t_1 = PyUnicode_Check(__pyx_v_dst); 
    if (__pyx_t_1) {

      /* "pyverbs/cmid.pyx":217
 *         if dst is not None:
 *             if isinstance(dst, str):
 *                 dst = dst.encode('utf-8')             # <<<<<<<<<<<<<<
 *             dst_addr = <char*>dst
 *         if src_service is not None:
*/
      __pyx_t_3 = __pyx_v_dst;
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_utf_8};
        __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_encode, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 217, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_DECREF_SET(__pyx_v_dst, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "pyverbs/cmid.pyx":216
 *             src_addr = <char*>src
 *         if dst is not None:
 *             if isinstance(dst, str):             # <<<<<<<<<<<<<<
 *                 dst = dst.encode('utf-8')
 *             dst_addr = <char*>dst
*/
    }

    /* "pyverbs/cmid.pyx":218
 *             if isinstance(dst, str):
 *                 dst = dst.encode('utf-8')
 *             dst_addr = <char*>dst             # <<<<<<<<<<<<<<
 *         if src_service is not None:
 *             if isinstance(src_service, str):
*/
    __pyx_t_5 = __Pyx_PyObject_AsWritableString(__pyx_v_dst); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(0, 218, __pyx_L1_error)
    __pyx_v_dst_addr = ((char *)__pyx_t_5);

    /* "pyverbs/cmid.pyx":215
 *                 src = src.encode('utf-8')
 *             src_addr = <char*>src
 *         if dst is not None:             # <<<<<<<<<<<<<<
 *             if isinstance(dst, str):
 *                 dst = dst.encode('utf-8')
*/
  }

  /* "pyverbs/cmid.pyx":219
 *                 dst = dst.encode('utf-8')
 *             dst_addr = <char*>dst
 *         if src_service is not None:             # <<<<<<<<<<<<<<
 *             if isinstance(src_service, str):
 *                 src_service = src_service.encode('utf-8')
*/
  __pyx_t_1 = (__pyx_v_src_service != Py_None);
  if (__pyx_t_1) {

    /* "pyverbs/cmid.pyx":220
 *             dst_addr = <char*>dst
 *         if src_service is not None:
 *             if isinstance(src_service, str):             # <<<<<<<<<<<<<<
 *                 src_service = src_service.encode('utf-8')
 *             src_srvc = <char*>src_service
*/
    __pyx_t_1 = PyUnicode_Check(__pyx_v_src_service); 
    if (__pyx_t_1) {

      /* "pyverbs/cmid.pyx":221
 *         if src_service is not None:
 *             if isinstance(src_service, str):
 *                 src_service = src_service.encode('utf-8')             # <<<<<<<<<<<<<<
 *             src_srvc = <char*>src_service
 *         if dst_service is not None:
*/
      __pyx_t_3 = __pyx_v_src_service;
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_utf_8};
        __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_encode, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 221, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_DECREF_SET(__pyx_v_src_service, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "pyverbs/cmid.pyx":220
 *             dst_addr = <char*>dst
 *         if src_service is not None:
 *             if isinstance(src_service, str):             # <<<<<<<<<<<<<<
 *                 src_service = src_service.encode('utf-8')
 *             src_srvc = <char*>src_service
*/
    }

    /* "pyverbs/cmid.pyx":222
 *             if isinstance(src_service, str):
 *                 src_service = src_service.encode('utf-8')
 *             src_srvc = <char*>src_service             # <<<<<<<<<<<<<<
 *         if dst_service is not None:
 *             if isinstance(dst_service, str):
*/
    __pyx_t_5 = __Pyx_PyObject_AsWritableString(__pyx_v_src_service); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(0, 222, __pyx_L1_error)
    __pyx_v_src_srvc = ((char *)__pyx_t_5);

    /* "pyverbs/cmid.pyx":219
 *                 dst = dst.encode('utf-8')
 *             dst_addr = <char*>dst
 *         if src_service is not None:             # <<<<<<<<<<<<<<
 *             if isinstance(src_service, str):
 *                 src_service = src_service.encode('utf-8')
*/
  }

  /* "pyverbs/cmid.pyx":223
 *                 src_service = src_service.encode('utf-8')
 *             src_srvc = <char*>src_service
 *         if dst_service is not None:             # <<<<<<<<<<<<<<
 *             if isinstance(dst_service, str):
 *                 dst_service = dst_service.encode('utf-8')
*/
  __pyx_t_1 = (__pyx_v_dst_service != Py_None);
  if (__pyx_t_1) {

    /* "pyverbs/cmid.pyx":224
 *             src_srvc = <char*>src_service
 *         if dst_service is not None:
 *             if isinstance(dst_service, str):             # <<<<<<<<<<<<<<
 *                 dst_service = dst_service.encode('utf-8')
 *             dst_srvc = <char*>dst_service
*/
    __pyx_t_1 = PyUnicode_Check(__pyx_v_dst_service); 
    if (__pyx_t_1) {

      /* "pyverbs/cmid.pyx":225
 *         if dst_service is not None:
 *             if isinstance(dst_service, str):
 *                 dst_service = dst_service.encode('utf-8')             # <<<<<<<<<<<<<<
 *             dst_srvc = <char*>dst_service
 * 
*/
      __pyx_t_3 = __pyx_v_dst_service;
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_utf_8};
        __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_encode, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 225, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_DECREF_SET(__pyx_v_dst_service, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "pyverbs/cmid.pyx":224
 *             src_srvc = <char*>src_service
 *         if dst_service is not None:
 *             if isinstance(dst_service, str):             # <<<<<<<<<<<<<<
 *                 dst_service = dst_service.encode('utf-8')
 *             dst_srvc = <char*>dst_service
*/
    }

    /* "pyverbs/cmid.pyx":226
 *             if isinstance(dst_service, str):
 *                 dst_service = dst_service.encode('utf-8')
 *             dst_srvc = <char*>dst_service             # <<<<<<<<<<<<<<
 * 
 *         hints_ptr = &hints
*/
    __pyx_t_5 = __Pyx_PyObject_AsWritableString(__pyx_v_dst_service); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(0, 226, __pyx_L1_error)
    __pyx_v_dst_srvc = ((char *)__pyx_t_5);

    /* "pyverbs/cmid.pyx":223
 *                 src_service = src_service.encode('utf-8')
 *             src_srvc = <char*>src_service
 *         if dst_service is not None:             # <<<<<<<<<<<<<<
 *             if isinstance(dst_service, str):
 *                 dst_service = dst_service.encode('utf-8')
*/
  }

  /* "pyverbs/cmid.pyx":228
 *             dst_srvc = <char*>dst_service
 * 
 *         hints_ptr = &hints             # <<<<<<<<<<<<<<
 *         memset(hints_ptr, 0, sizeof(cm.rdma_addrinfo))
 *         hints.ai_port_space = port_space
*/
  __pyx_v_hints_ptr = (&__pyx_v_hints);

  /* "pyverbs/cmid.pyx":229
 * 
 *         hints_ptr = &hints
 *         memset(hints_ptr, 0, sizeof(cm.rdma_addrinfo))             # <<<<<<<<<<<<<<
 *         hints.ai_port_space = port_space
 *         hints.ai_flags = flags
*/
  (void)(memset(__pyx_v_hints_ptr, 0, (sizeof(struct rdma_addrinfo))));

  /* "pyverbs/cmid.pyx":230
 *         hints_ptr = &hints
 *         memset(hints_ptr, 0, sizeof(cm.rdma_addrinfo))
 *         hints.ai_port_space = port_space             # <<<<<<<<<<<<<<
 *         hints.ai_flags = flags
 * 
*/
  __pyx_t_6 = __Pyx_PyLong_As_int(__pyx_v_port_space); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 230, __pyx_L1_error)
  __pyx_v_hints.ai_port_space = __pyx_t_6;

  /* "pyverbs/cmid.pyx":231
 *         memset(hints_ptr, 0, sizeof(cm.rdma_addrinfo))
 *         hints.ai_port_space = port_space
 *         hints.ai_flags = flags             # <<<<<<<<<<<<<<
 * 
 *         # Set hint fields to support GID address
*/
  __pyx_t_6 = __Pyx_PyLong_As_int(__pyx_v_flags); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 231, __pyx_L1_error)
  __pyx_v_hints.ai_flags = __pyx_t_6;

  /* "pyverbs/cmid.pyx":234
 * 
 *         # Set hint fields to support GID address
 *         if port_space == ce.RDMA_PS_IB:             # <<<<<<<<<<<<<<
 *             hints.ai_flags |= ce.RAI_FAMILY | ce.RAI_NUMERICHOST
 *             hints.ai_family = cm.AF_IB
*/
  __pyx_t_2 = __Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py(RDMA_PS_IB); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_port_space, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "pyverbs/cmid.pyx":235
 *         # Set hint fields to support GID address
 *         if port_space == ce.RDMA_PS_IB:
 *             hints.ai_flags |= ce.RAI_FAMILY | ce.RAI_NUMERICHOST             # <<<<<<<<<<<<<<
 *             hints.ai_family = cm.AF_IB
 * 
*/
    __pyx_v_hints.ai_flags = (__pyx_v_hints.ai_flags | (RAI_FAMILY | RAI_NUMERICHOST));

    /* "pyverbs/cmid.pyx":236
 *         if port_space == ce.RDMA_PS_IB:
 *             hints.ai_flags |= ce.RAI_FAMILY | ce.RAI_NUMERICHOST
 *             hints.ai_family = cm.AF_IB             # <<<<<<<<<<<<<<
 * 
 *         if flags & ce.RAI_PASSIVE:
*/
    __pyx_v_hints.ai_family = AF_IB;

    /* "pyverbs/cmid.pyx":234
 * 
 *         # Set hint fields to support GID address
 *         if port_space == ce.RDMA_PS_IB:             # <<<<<<<<<<<<<<
 *             hints.ai_flags |= ce.RAI_FAMILY | ce.RAI_NUMERICHOST
 *             hints.ai_family = cm.AF_IB
*/
  }

  /* "pyverbs/cmid.pyx":238
 *             hints.ai_family = cm.AF_IB
 * 
 *         if flags & ce.RAI_PASSIVE:             # <<<<<<<<<<<<<<
 *             ret = cm.rdma_getaddrinfo(src_addr, src_srvc, hints_ptr,
 *                                       &self.addr_info)
*/
  __pyx_t_3 = __Pyx_PyLong_From___pyx_anon_enum(RAI_PASSIVE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyNumber_And(__pyx_v_flags, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_1) {

    /* "pyverbs/cmid.pyx":239
 * 
 *         if flags & ce.RAI_PASSIVE:
 *             ret = cm.rdma_getaddrinfo(src_addr, src_srvc, hints_ptr,             # <<<<<<<<<<<<<<
 *                                       &self.addr_info)
 *         else:
*/
    __pyx_v_ret = rdma_getaddrinfo(__pyx_v_src_addr, __pyx_v_src_srvc, __pyx_v_hints_ptr, (&__pyx_v_self->addr_info));

    /* "pyverbs/cmid.pyx":238
 *             hints.ai_family = cm.AF_IB
 * 
 *         if flags & ce.RAI_PASSIVE:             # <<<<<<<<<<<<<<
 *             ret = cm.rdma_getaddrinfo(src_addr, src_srvc, hints_ptr,
 *                                       &self.addr_info)
*/
    goto __pyx_L12;
  }

  /* "pyverbs/cmid.pyx":242
 *                                       &self.addr_info)
 *         else:
 *             if src:             # <<<<<<<<<<<<<<
 *                 hints.ai_flags |= ce.RAI_PASSIVE
 *                 ret = cm.rdma_getaddrinfo(src_addr, src_srvc, hints_ptr, &res)
*/
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_src); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 242, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "pyverbs/cmid.pyx":243
 *         else:
 *             if src:
 *                 hints.ai_flags |= ce.RAI_PASSIVE             # <<<<<<<<<<<<<<
 *                 ret = cm.rdma_getaddrinfo(src_addr, src_srvc, hints_ptr, &res)
 *                 if ret != 0:
*/
      __pyx_v_hints.ai_flags = (__pyx_v_hints.ai_flags | RAI_PASSIVE);

      /* "pyverbs/cmid.pyx":244
 *             if src:
 *                 hints.ai_flags |= ce.RAI_PASSIVE
 *                 ret = cm.rdma_getaddrinfo(src_addr, src_srvc, hints_ptr, &res)             # <<<<<<<<<<<<<<
 *                 if ret != 0:
 *                     raise PyverbsRDMAErrno('Failed to get Address Info')
*/
      __pyx_v_ret = rdma_getaddrinfo(__pyx_v_src_addr, __pyx_v_src_srvc, __pyx_v_hints_ptr, (&__pyx_v_res));

      /* "pyverbs/cmid.pyx":245
 *                 hints.ai_flags |= ce.RAI_PASSIVE
 *                 ret = cm.rdma_getaddrinfo(src_addr, src_srvc, hints_ptr, &res)
 *                 if ret != 0:             # <<<<<<<<<<<<<<
 *                     raise PyverbsRDMAErrno('Failed to get Address Info')
 *                 hints.ai_src_addr = <cm.sockaddr*>res.ai_src_addr
*/
      __pyx_t_1 = (__pyx_v_ret != 0);
      if (unlikely(__pyx_t_1)) {

        /* "pyverbs/cmid.pyx":246
 *                 ret = cm.rdma_getaddrinfo(src_addr, src_srvc, hints_ptr, &res)
 *                 if ret != 0:
 *                     raise PyverbsRDMAErrno('Failed to get Address Info')             # <<<<<<<<<<<<<<
 *                 hints.ai_src_addr = <cm.sockaddr*>res.ai_src_addr
 *                 hints.ai_src_len = res.ai_src_len
*/
        __pyx_t_3 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 246, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_4 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_7, __pyx__function);
          __pyx_t_4 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Failed_to_get_Address_Info};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 246, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_Raise(__pyx_t_2, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __PYX_ERR(0, 246, __pyx_L1_error)

        /* "pyverbs/cmid.pyx":245
 *                 hints.ai_flags |= ce.RAI_PASSIVE
 *                 ret = cm.rdma_getaddrinfo(src_addr, src_srvc, hints_ptr, &res)
 *                 if ret != 0:             # <<<<<<<<<<<<<<
 *                     raise PyverbsRDMAErrno('Failed to get Address Info')
 *                 hints.ai_src_addr = <cm.sockaddr*>res.ai_src_addr
*/
      }

      /* "pyverbs/cmid.pyx":247
 *                 if ret != 0:
 *                     raise PyverbsRDMAErrno('Failed to get Address Info')
 *                 hints.ai_src_addr = <cm.sockaddr*>res.ai_src_addr             # <<<<<<<<<<<<<<
 *                 hints.ai_src_len = res.ai_src_len
 *                 hints.ai_flags &= ~ce.RAI_PASSIVE
*/
      __pyx_v_hints.ai_src_addr = ((struct sockaddr *)__pyx_v_res->ai_src_addr);

      /* "pyverbs/cmid.pyx":248
 *                     raise PyverbsRDMAErrno('Failed to get Address Info')
 *                 hints.ai_src_addr = <cm.sockaddr*>res.ai_src_addr
 *                 hints.ai_src_len = res.ai_src_len             # <<<<<<<<<<<<<<
 *                 hints.ai_flags &= ~ce.RAI_PASSIVE
 *             ret = cm.rdma_getaddrinfo(dst_addr, dst_srvc, hints_ptr,
*/
      __pyx_t_6 = __pyx_v_res->ai_src_len;
      __pyx_v_hints.ai_src_len = __pyx_t_6;

      /* "pyverbs/cmid.pyx":249
 *                 hints.ai_src_addr = <cm.sockaddr*>res.ai_src_addr
 *                 hints.ai_src_len = res.ai_src_len
 *                 hints.ai_flags &= ~ce.RAI_PASSIVE             # <<<<<<<<<<<<<<
 *             ret = cm.rdma_getaddrinfo(dst_addr, dst_srvc, hints_ptr,
 *                                       &self.addr_info)
*/
      __pyx_v_hints.ai_flags = (__pyx_v_hints.ai_flags & (~RAI_PASSIVE));

      /* "pyverbs/cmid.pyx":242
 *                                       &self.addr_info)
 *         else:
 *             if src:             # <<<<<<<<<<<<<<
 *                 hints.ai_flags |= ce.RAI_PASSIVE
 *                 ret = cm.rdma_getaddrinfo(src_addr, src_srvc, hints_ptr, &res)
*/
    }

    /* "pyverbs/cmid.pyx":250
 *                 hints.ai_src_len = res.ai_src_len
 *                 hints.ai_flags &= ~ce.RAI_PASSIVE
 *             ret = cm.rdma_getaddrinfo(dst_addr, dst_srvc, hints_ptr,             # <<<<<<<<<<<<<<
 *                                       &self.addr_info)
 *             if src:
*/
    __pyx_v_ret = rdma_getaddrinfo(__pyx_v_dst_addr, __pyx_v_dst_srvc, __pyx_v_hints_ptr, (&__pyx_v_self->addr_info));

    /* "pyverbs/cmid.pyx":252
 *             ret = cm.rdma_getaddrinfo(dst_addr, dst_srvc, hints_ptr,
 *                                       &self.addr_info)
 *             if src:             # <<<<<<<<<<<<<<
 *                 cm.rdma_freeaddrinfo(res)
 *         if ret != 0:
*/
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_src); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 252, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "pyverbs/cmid.pyx":253
 *                                       &self.addr_info)
 *             if src:
 *                 cm.rdma_freeaddrinfo(res)             # <<<<<<<<<<<<<<
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to get Address Info')
*/
      rdma_freeaddrinfo(__pyx_v_res);

      /* "pyverbs/cmid.pyx":252
 *             ret = cm.rdma_getaddrinfo(dst_addr, dst_srvc, hints_ptr,
 *                                       &self.addr_info)
 *             if src:             # <<<<<<<<<<<<<<
 *                 cm.rdma_freeaddrinfo(res)
 *         if ret != 0:
*/
    }
  }
  __pyx_L12:;

  /* "pyverbs/cmid.pyx":254
 *             if src:
 *                 cm.rdma_freeaddrinfo(res)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to get Address Info')
 * 
*/
  __pyx_t_1 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pyverbs/cmid.pyx":255
 *                 cm.rdma_freeaddrinfo(res)
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to get Address Info')             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
*/
    __pyx_t_7 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
      assert(__pyx_t_7);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
      __pyx_t_4 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_mstate_global->__pyx_kp_u_Failed_to_get_Address_Info};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 255, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 255, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":254
 *             if src:
 *                 cm.rdma_freeaddrinfo(res)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to get Address Info')
 * 
*/
  }

  /* "pyverbs/cmid.pyx":190
 *             self.get_addrinfo(src, dst, src_service, dst_service, port_space, flags)
 * 
 *     def get_addrinfo(self, src=None, dst=None, src_service=None, dst_service=None,             # <<<<<<<<<<<<<<
 *                      port_space=0, flags=0):
 *         """
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyverbs.cmid.AddrInfo.get_addrinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_src);
  __Pyx_XDECREF(__pyx_v_dst);
  __Pyx_XDECREF(__pyx_v_src_service);
  __Pyx_XDECREF(__pyx_v_dst_service);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":257
 *             raise PyverbsRDMAErrno('Failed to get Address Info')
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         self.close()
 * 
*/

/* Python wrapper */
static void __pyx_pw_7pyverbs_4cmid_8AddrInfo_5__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7pyverbs_4cmid_8AddrInfo_5__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_7pyverbs_4cmid_8AddrInfo_4__dealloc__(((struct __pyx_obj_7pyverbs_4cmid_AddrInfo *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7pyverbs_4cmid_8AddrInfo_4__dealloc__(struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "pyverbs/cmid.pyx":258
 * 
 *     def __dealloc__(self):
 *         self.close()             # <<<<<<<<<<<<<<
 * 
 *     cpdef close(self):
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_7pyverbs_4cmid_AddrInfo *)__pyx_v_self->__pyx_vtab)->close(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":257
 *             raise PyverbsRDMAErrno('Failed to get Address Info')
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         self.close()
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("pyverbs.cmid.AddrInfo.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "pyverbs/cmid.pyx":260
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.addr_info != NULL:
 *             if self.logger:
*/

static PyObject *__pyx_pw_7pyverbs_4cmid_8AddrInfo_7close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7pyverbs_4cmid_8AddrInfo_close(struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_AddrInfo &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_close); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 260, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyverbs_4cmid_8AddrInfo_7close)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 260, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyverbs/cmid.pyx":261
 * 
 *     cpdef close(self):
 *         if self.addr_info != NULL:             # <<<<<<<<<<<<<<
 *             if self.logger:
 *                 self.logger.debug('Closing AddrInfo')
*/
  __pyx_t_6 = (__pyx_v_self->addr_info != NULL);
  if (__pyx_t_6) {

    /* "pyverbs/cmid.pyx":262
 *     cpdef close(self):
 *         if self.addr_info != NULL:
 *             if self.logger:             # <<<<<<<<<<<<<<
 *                 self.logger.debug('Closing AddrInfo')
 *             cm.rdma_freeaddrinfo(self.addr_info)
*/
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_self->__pyx_base.logger); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 262, __pyx_L1_error)
    if (__pyx_t_6) {

      /* "pyverbs/cmid.pyx":263
 *         if self.addr_info != NULL:
 *             if self.logger:
 *                 self.logger.debug('Closing AddrInfo')             # <<<<<<<<<<<<<<
 *             cm.rdma_freeaddrinfo(self.addr_info)
 *         self.addr_info = NULL
*/
      __pyx_t_2 = __pyx_v_self->__pyx_base.logger;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_5 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Closing_AddrInfo};
        __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_debug, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 263, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyverbs/cmid.pyx":262
 *     cpdef close(self):
 *         if self.addr_info != NULL:
 *             if self.logger:             # <<<<<<<<<<<<<<
 *                 self.logger.debug('Closing AddrInfo')
 *             cm.rdma_freeaddrinfo(self.addr_info)
*/
    }

    /* "pyverbs/cmid.pyx":264
 *             if self.logger:
 *                 self.logger.debug('Closing AddrInfo')
 *             cm.rdma_freeaddrinfo(self.addr_info)             # <<<<<<<<<<<<<<
 *         self.addr_info = NULL
 * 
*/
    rdma_freeaddrinfo(__pyx_v_self->addr_info);

    /* "pyverbs/cmid.pyx":261
 * 
 *     cpdef close(self):
 *         if self.addr_info != NULL:             # <<<<<<<<<<<<<<
 *             if self.logger:
 *                 self.logger.debug('Closing AddrInfo')
*/
  }

  /* "pyverbs/cmid.pyx":265
 *                 self.logger.debug('Closing AddrInfo')
 *             cm.rdma_freeaddrinfo(self.addr_info)
 *         self.addr_info = NULL             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_v_self->addr_info = NULL;

  /* "pyverbs/cmid.pyx":260
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.addr_info != NULL:
 *             if self.logger:
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cmid.AddrInfo.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_8AddrInfo_7close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_8AddrInfo_7close = {"close", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_8AddrInfo_7close, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_4cmid_8AddrInfo_7close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("close", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("close", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_4cmid_8AddrInfo_6close(((struct __pyx_obj_7pyverbs_4cmid_AddrInfo *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_8AddrInfo_6close(struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7pyverbs_4cmid_8AddrInfo_close(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cmid.AddrInfo.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.addr_info cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_8AddrInfo_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_8AddrInfo_9__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_8AddrInfo_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_4cmid_8AddrInfo_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_4cmid_8AddrInfo_8__reduce_cython__(((struct __pyx_obj_7pyverbs_4cmid_AddrInfo *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_8AddrInfo_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.addr_info cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.addr_info cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_addr_info_cannot_be_convert, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.addr_info cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cmid.AddrInfo.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.addr_info cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.addr_info cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_8AddrInfo_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_8AddrInfo_11__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_8AddrInfo_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_4cmid_8AddrInfo_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.AddrInfo.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_4cmid_8AddrInfo_10__setstate_cython__(((struct __pyx_obj_7pyverbs_4cmid_AddrInfo *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_8AddrInfo_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.addr_info cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.addr_info cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_addr_info_cannot_be_convert, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.addr_info cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.addr_info cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cmid.AddrInfo.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":270
 * cdef class CMEvent(PyverbsObject):
 * 
 *     def __init__(self, CMEventChannel channel):             # <<<<<<<<<<<<<<
 *         """
 *         Initialize a CMEvent object over an underlying rdma_cm_event C object
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_4cmid_7CMEvent_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_7CMEvent___init__, "\n        Initialize a CMEvent object over an underlying rdma_cm_event C object\n        :param channel: Event Channel on which this event has been received\n        :return: CMEvent object\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_7pyverbs_4cmid_7CMEvent___init__;
#endif
static int __pyx_pw_7pyverbs_4cmid_7CMEvent_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *__pyx_v_channel = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_channel,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 270, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 270, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 270, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, i); __PYX_ERR(0, 270, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 270, __pyx_L3_error)
    }
    __pyx_v_channel = ((struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 270, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMEvent.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_channel), __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMEventChannel, 1, "channel", 0))) __PYX_ERR(0, 270, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_4cmid_7CMEvent___init__(((struct __pyx_obj_7pyverbs_4cmid_CMEvent *)__pyx_v_self), __pyx_v_channel);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_4cmid_7CMEvent___init__(struct __pyx_obj_7pyverbs_4cmid_CMEvent *__pyx_v_self, struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *__pyx_v_channel) {
  int __pyx_v_ret;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyverbs/cmid.pyx":276
 *         :return: CMEvent object
 *         """
 *         super().__init__()             # <<<<<<<<<<<<<<
 *         ret = cm.rdma_get_cm_event(channel.event_channel, &self.event)
 *         if ret != 0:
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMEvent), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":277
 *         """
 *         super().__init__()
 *         ret = cm.rdma_get_cm_event(channel.event_channel, &self.event)             # <<<<<<<<<<<<<<
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to create CMEvent')
*/
  __pyx_v_ret = rdma_get_cm_event(__pyx_v_channel->event_channel, (&__pyx_v_self->event));

  /* "pyverbs/cmid.pyx":278
 *         super().__init__()
 *         ret = cm.rdma_get_cm_event(channel.event_channel, &self.event)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to create CMEvent')
 *         self.logger.debug('Created a CMEvent')
*/
  __pyx_t_6 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_6)) {

    /* "pyverbs/cmid.pyx":279
 *         ret = cm.rdma_get_cm_event(channel.event_channel, &self.event)
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to create CMEvent')             # <<<<<<<<<<<<<<
 *         self.logger.debug('Created a CMEvent')
 * 
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Failed_to_create_CMEvent};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 279, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":278
 *         super().__init__()
 *         ret = cm.rdma_get_cm_event(channel.event_channel, &self.event)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to create CMEvent')
 *         self.logger.debug('Created a CMEvent')
*/
  }

  /* "pyverbs/cmid.pyx":280
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to create CMEvent')
 *         self.logger.debug('Created a CMEvent')             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
*/
  __pyx_t_2 = __pyx_v_self->__pyx_base.logger;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Created_a_CMEvent};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_debug, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":270
 * cdef class CMEvent(PyverbsObject):
 * 
 *     def __init__(self, CMEventChannel channel):             # <<<<<<<<<<<<<<
 *         """
 *         Initialize a CMEvent object over an underlying rdma_cm_event C object
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cmid.CMEvent.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":282
 *         self.logger.debug('Created a CMEvent')
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         self.close()
 * 
*/

/* Python wrapper */
static void __pyx_pw_7pyverbs_4cmid_7CMEvent_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7pyverbs_4cmid_7CMEvent_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_7pyverbs_4cmid_7CMEvent_2__dealloc__(((struct __pyx_obj_7pyverbs_4cmid_CMEvent *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7pyverbs_4cmid_7CMEvent_2__dealloc__(struct __pyx_obj_7pyverbs_4cmid_CMEvent *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "pyverbs/cmid.pyx":283
 * 
 *     def __dealloc__(self):
 *         self.close()             # <<<<<<<<<<<<<<
 * 
 *     cpdef close(self):
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_7pyverbs_4cmid_CMEvent *)__pyx_v_self->__pyx_vtab)->close(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":282
 *         self.logger.debug('Created a CMEvent')
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         self.close()
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("pyverbs.cmid.CMEvent.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "pyverbs/cmid.pyx":285
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.event != NULL:
 *             if self.logger:
*/

static PyObject *__pyx_pw_7pyverbs_4cmid_7CMEvent_5close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7pyverbs_4cmid_7CMEvent_close(struct __pyx_obj_7pyverbs_4cmid_CMEvent *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMEvent &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_close); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyverbs_4cmid_7CMEvent_5close)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 285, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyverbs/cmid.pyx":286
 * 
 *     cpdef close(self):
 *         if self.event != NULL:             # <<<<<<<<<<<<<<
 *             if self.logger:
 *                 self.logger.debug('Closing CMEvent')
*/
  __pyx_t_6 = (__pyx_v_self->event != NULL);
  if (__pyx_t_6) {

    /* "pyverbs/cmid.pyx":287
 *     cpdef close(self):
 *         if self.event != NULL:
 *             if self.logger:             # <<<<<<<<<<<<<<
 *                 self.logger.debug('Closing CMEvent')
 *             self.ack_cm_event()
*/
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_self->__pyx_base.logger); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 287, __pyx_L1_error)
    if (__pyx_t_6) {

      /* "pyverbs/cmid.pyx":288
 *         if self.event != NULL:
 *             if self.logger:
 *                 self.logger.debug('Closing CMEvent')             # <<<<<<<<<<<<<<
 *             self.ack_cm_event()
 *             self.event = NULL
*/
      __pyx_t_2 = __pyx_v_self->__pyx_base.logger;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_5 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Closing_CMEvent};
        __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_debug, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 288, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyverbs/cmid.pyx":287
 *     cpdef close(self):
 *         if self.event != NULL:
 *             if self.logger:             # <<<<<<<<<<<<<<
 *                 self.logger.debug('Closing CMEvent')
 *             self.ack_cm_event()
*/
    }

    /* "pyverbs/cmid.pyx":289
 *             if self.logger:
 *                 self.logger.debug('Closing CMEvent')
 *             self.ack_cm_event()             # <<<<<<<<<<<<<<
 *             self.event = NULL
 * 
*/
    __pyx_t_2 = ((PyObject *)__pyx_v_self);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_ack_cm_event, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyverbs/cmid.pyx":290
 *                 self.logger.debug('Closing CMEvent')
 *             self.ack_cm_event()
 *             self.event = NULL             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
    __pyx_v_self->event = NULL;

    /* "pyverbs/cmid.pyx":286
 * 
 *     cpdef close(self):
 *         if self.event != NULL:             # <<<<<<<<<<<<<<
 *             if self.logger:
 *                 self.logger.debug('Closing CMEvent')
*/
  }

  /* "pyverbs/cmid.pyx":285
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.event != NULL:
 *             if self.logger:
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cmid.CMEvent.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_7CMEvent_5close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_7CMEvent_5close = {"close", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_7CMEvent_5close, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_4cmid_7CMEvent_5close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("close", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("close", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_4cmid_7CMEvent_4close(((struct __pyx_obj_7pyverbs_4cmid_CMEvent *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_7CMEvent_4close(struct __pyx_obj_7pyverbs_4cmid_CMEvent *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7pyverbs_4cmid_7CMEvent_close(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cmid.CMEvent.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":292
 *             self.event = NULL
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def event_type(self):
 *         return self.event.event
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_7CMEvent_10event_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_4cmid_7CMEvent_10event_type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_4cmid_7CMEvent_10event_type___get__(((struct __pyx_obj_7pyverbs_4cmid_CMEvent *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_7CMEvent_10event_type___get__(struct __pyx_obj_7pyverbs_4cmid_CMEvent *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cmid.pyx":294
 *     @property
 *     def event_type(self):
 *         return self.event.event             # <<<<<<<<<<<<<<
 * 
 *     def ack_cm_event(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py(__pyx_v_self->event->event); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":292
 *             self.event = NULL
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def event_type(self):
 *         return self.event.event
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cmid.CMEvent.event_type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":296
 *         return self.event.event
 * 
 *     def ack_cm_event(self):             # <<<<<<<<<<<<<<
 *         """
 *         Free a communication event. This call frees the event structure and any
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_7CMEvent_7ack_cm_event(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_7CMEvent_6ack_cm_event, "\n        Free a communication event. This call frees the event structure and any\n        memory that it references.\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_7CMEvent_7ack_cm_event = {"ack_cm_event", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_7CMEvent_7ack_cm_event, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_7CMEvent_6ack_cm_event};
static PyObject *__pyx_pw_7pyverbs_4cmid_7CMEvent_7ack_cm_event(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ack_cm_event (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("ack_cm_event", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("ack_cm_event", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_4cmid_7CMEvent_6ack_cm_event(((struct __pyx_obj_7pyverbs_4cmid_CMEvent *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_7CMEvent_6ack_cm_event(struct __pyx_obj_7pyverbs_4cmid_CMEvent *__pyx_v_self) {
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  size_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ack_cm_event", 0);

  /* "pyverbs/cmid.pyx":302
 *         :return: None
 *         """
 *         ret = cm.rdma_ack_cm_event(self.event)             # <<<<<<<<<<<<<<
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Acknowledge Event - {}'
*/
  __pyx_v_ret = rdma_ack_cm_event(__pyx_v_self->event);

  /* "pyverbs/cmid.pyx":303
 *         """
 *         ret = cm.rdma_ack_cm_event(self.event)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Acknowledge Event - {}'
 *                                    .format(self.event_str()))
*/
  __pyx_t_1 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pyverbs/cmid.pyx":304
 *         ret = cm.rdma_ack_cm_event(self.event)
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Acknowledge Event - {}'             # <<<<<<<<<<<<<<
 *                                    .format(self.event_str()))
 *         self.event = NULL
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "pyverbs/cmid.pyx":305
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Acknowledge Event - {}'
 *                                    .format(self.event_str()))             # <<<<<<<<<<<<<<
 *         self.event = NULL
 * 
*/
    __pyx_t_6 = __pyx_mstate_global->__pyx_kp_u_Failed_to_Acknowledge_Event;
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_8 = ((PyObject *)__pyx_v_self);
    __Pyx_INCREF(__pyx_t_8);
    __pyx_t_9 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, NULL};
      __pyx_t_7 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_event_str, __pyx_callargs+__pyx_t_9, (1-__pyx_t_9) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 305, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
    }
    __pyx_t_9 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_7};
      __pyx_t_5 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_format, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 305, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
    }
    __pyx_t_9 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_9 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 304, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 304, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":303
 *         """
 *         ret = cm.rdma_ack_cm_event(self.event)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Acknowledge Event - {}'
 *                                    .format(self.event_str()))
*/
  }

  /* "pyverbs/cmid.pyx":306
 *             raise PyverbsRDMAErrno('Failed to Acknowledge Event - {}'
 *                                    .format(self.event_str()))
 *         self.event = NULL             # <<<<<<<<<<<<<<
 * 
 *     def event_str(self):
*/
  __pyx_v_self->event = NULL;

  /* "pyverbs/cmid.pyx":296
 *         return self.event.event
 * 
 *     def ack_cm_event(self):             # <<<<<<<<<<<<<<
 *         """
 *         Free a communication event. This call frees the event structure and any
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyverbs.cmid.CMEvent.ack_cm_event", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":308
 *         self.event = NULL
 * 
 *     def event_str(self):             # <<<<<<<<<<<<<<
 *         if self.event == NULL:
 *             return ''
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_7CMEvent_9event_str(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_7CMEvent_9event_str = {"event_str", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_7CMEvent_9event_str, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_4cmid_7CMEvent_9event_str(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("event_str (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("event_str", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("event_str", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_4cmid_7CMEvent_8event_str(((struct __pyx_obj_7pyverbs_4cmid_CMEvent *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_7CMEvent_8event_str(struct __pyx_obj_7pyverbs_4cmid_CMEvent *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  enum rdma_cm_event_type __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("event_str", 0);

  /* "pyverbs/cmid.pyx":309
 * 
 *     def event_str(self):
 *         if self.event == NULL:             # <<<<<<<<<<<<<<
 *             return ''
 *         return (<bytes>cm.rdma_event_str(self.event_type)).decode()
*/
  __pyx_t_1 = (__pyx_v_self->event == NULL);
  if (__pyx_t_1) {

    /* "pyverbs/cmid.pyx":310
 *     def event_str(self):
 *         if self.event == NULL:
 *             return ''             # <<<<<<<<<<<<<<
 *         return (<bytes>cm.rdma_event_str(self.event_type)).decode()
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_kp_u__3);
    __pyx_r = __pyx_mstate_global->__pyx_kp_u__3;
    goto __pyx_L0;

    /* "pyverbs/cmid.pyx":309
 * 
 *     def event_str(self):
 *         if self.event == NULL:             # <<<<<<<<<<<<<<
 *             return ''
 *         return (<bytes>cm.rdma_event_str(self.event_type)).decode()
*/
  }

  /* "pyverbs/cmid.pyx":311
 *         if self.event == NULL:
 *             return ''
 *         return (<bytes>cm.rdma_event_str(self.event_type)).decode()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_event_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((enum rdma_cm_event_type)__Pyx_PyLong_As_enum__rdma_cm_event_type(__pyx_t_2)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBytes_FromString(rdma_event_str(__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(__pyx_t_2 == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "decode");
    __PYX_ERR(0, 311, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_decode_bytes(((PyObject*)__pyx_t_2), 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":308
 *         self.event = NULL
 * 
 *     def event_str(self):             # <<<<<<<<<<<<<<
 *         if self.event == NULL:
 *             return ''
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cmid.CMEvent.event_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":313
 *         return (<bytes>cm.rdma_event_str(self.event_type)).decode()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def private_data(self):
 *         data_array = bytearray(self.event.param.conn.private_data_len)
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_7CMEvent_12private_data_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_4cmid_7CMEvent_12private_data_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_4cmid_7CMEvent_12private_data___get__(((struct __pyx_obj_7pyverbs_4cmid_CMEvent *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_7CMEvent_12private_data___get__(struct __pyx_obj_7pyverbs_4cmid_CMEvent *__pyx_v_self) {
  PyObject *__pyx_v_data_array = NULL;
  unsigned char const *__pyx_v_p;
  uint8_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  uint8_t __pyx_t_5;
  uint8_t __pyx_t_6;
  uint8_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cmid.pyx":315
 *     @property
 *     def private_data(self):
 *         data_array = bytearray(self.event.param.conn.private_data_len)             # <<<<<<<<<<<<<<
 *         cdef const unsigned char *p = <const unsigned char*>self.event.param.conn.private_data
 *         for i in range(self.event.param.conn.private_data_len):
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = __Pyx_PyLong_From_uint8_t(__pyx_v_self->event->param.conn.private_data_len); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(&PyByteArray_Type), __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 315, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_data_array = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":316
 *     def private_data(self):
 *         data_array = bytearray(self.event.param.conn.private_data_len)
 *         cdef const unsigned char *p = <const unsigned char*>self.event.param.conn.private_data             # <<<<<<<<<<<<<<
 *         for i in range(self.event.param.conn.private_data_len):
 *             data_array[i] = p[i]
*/
  __pyx_v_p = ((unsigned char const *)__pyx_v_self->event->param.conn.private_data);

  /* "pyverbs/cmid.pyx":317
 *         data_array = bytearray(self.event.param.conn.private_data_len)
 *         cdef const unsigned char *p = <const unsigned char*>self.event.param.conn.private_data
 *         for i in range(self.event.param.conn.private_data_len):             # <<<<<<<<<<<<<<
 *             data_array[i] = p[i]
 *         return bytes(data_array)
*/
  __pyx_t_5 = __pyx_v_self->event->param.conn.private_data_len;
  __pyx_t_6 = __pyx_t_5;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "pyverbs/cmid.pyx":318
 *         cdef const unsigned char *p = <const unsigned char*>self.event.param.conn.private_data
 *         for i in range(self.event.param.conn.private_data_len):
 *             data_array[i] = p[i]             # <<<<<<<<<<<<<<
 *         return bytes(data_array)
 * 
*/
    if (unlikely((__pyx_v_p[__pyx_v_i]) > 255)) {
      PyErr_SetString(PyExc_ValueError, "byte must be in range(0, 256)"); __PYX_ERR(0, 318, __pyx_L1_error)
    }
    if (unlikely((__Pyx_SetItemInt_ByteArray(__pyx_v_data_array, __pyx_v_i, ((unsigned char)(__pyx_v_p[__pyx_v_i])), uint8_t, 0, __Pyx_PyLong_From_uint8_t, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference) < 0))) __PYX_ERR(0, 318, __pyx_L1_error)
  }

  /* "pyverbs/cmid.pyx":319
 *         for i in range(self.event.param.conn.private_data_len):
 *             data_array[i] = p[i]
 *         return bytes(data_array)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_data_array};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(&PyBytes_Type), __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":313
 *         return (<bytes>cm.rdma_event_str(self.event_type)).decode()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def private_data(self):
 *         data_array = bytearray(self.event.param.conn.private_data_len)
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyverbs.cmid.CMEvent.private_data.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data_array);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.event cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_7CMEvent_11__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_7CMEvent_11__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_7CMEvent_11__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_4cmid_7CMEvent_11__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_4cmid_7CMEvent_10__reduce_cython__(((struct __pyx_obj_7pyverbs_4cmid_CMEvent *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_7CMEvent_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_CMEvent *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.event cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.event cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_event_cannot_be_converted_t, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.event cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cmid.CMEvent.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.event cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.event cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_7CMEvent_13__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_7CMEvent_13__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_7CMEvent_13__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_4cmid_7CMEvent_13__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMEvent.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_4cmid_7CMEvent_12__setstate_cython__(((struct __pyx_obj_7pyverbs_4cmid_CMEvent *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_7CMEvent_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_CMEvent *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.event cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.event cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_event_cannot_be_converted_t, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.event cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.event cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cmid.CMEvent.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":324
 * cdef class CMEventChannel(PyverbsObject):
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         """
 *         Initialize a CMEventChannel object over an underlying rdma_event_channel
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_4cmid_14CMEventChannel_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_14CMEventChannel___init__, "\n        Initialize a CMEventChannel object over an underlying rdma_event_channel\n        C object.\n        :return: EventChannel object\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_7pyverbs_4cmid_14CMEventChannel___init__;
#endif
static int __pyx_pw_7pyverbs_4cmid_14CMEventChannel_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, __pyx_nargs); return -1; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return -1;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__init__", __pyx_kwds); return -1;}
  __pyx_r = __pyx_pf_7pyverbs_4cmid_14CMEventChannel___init__(((struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_4cmid_14CMEventChannel___init__(struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyverbs/cmid.pyx":330
 *         :return: EventChannel object
 *         """
 *         super().__init__()             # <<<<<<<<<<<<<<
 *         self.event_channel = cm.rdma_create_event_channel()
 *         if self.event_channel == NULL:
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMEventChannel), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":331
 *         """
 *         super().__init__()
 *         self.event_channel = cm.rdma_create_event_channel()             # <<<<<<<<<<<<<<
 *         if self.event_channel == NULL:
 *             raise PyverbsRDMAErrno('Failed to create CMEventChannel')
*/
  __pyx_v_self->event_channel = rdma_create_event_channel();

  /* "pyverbs/cmid.pyx":332
 *         super().__init__()
 *         self.event_channel = cm.rdma_create_event_channel()
 *         if self.event_channel == NULL:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to create CMEventChannel')
 *         self.logger.debug('Created a CMEventChannel')
*/
  __pyx_t_6 = (__pyx_v_self->event_channel == NULL);
  if (unlikely(__pyx_t_6)) {

    /* "pyverbs/cmid.pyx":333
 *         self.event_channel = cm.rdma_create_event_channel()
 *         if self.event_channel == NULL:
 *             raise PyverbsRDMAErrno('Failed to create CMEventChannel')             # <<<<<<<<<<<<<<
 *         self.logger.debug('Created a CMEventChannel')
 * 
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Failed_to_create_CMEventChannel};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 333, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 333, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":332
 *         super().__init__()
 *         self.event_channel = cm.rdma_create_event_channel()
 *         if self.event_channel == NULL:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to create CMEventChannel')
 *         self.logger.debug('Created a CMEventChannel')
*/
  }

  /* "pyverbs/cmid.pyx":334
 *         if self.event_channel == NULL:
 *             raise PyverbsRDMAErrno('Failed to create CMEventChannel')
 *         self.logger.debug('Created a CMEventChannel')             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
*/
  __pyx_t_2 = __pyx_v_self->__pyx_base.logger;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Created_a_CMEventChannel};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_debug, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 334, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":324
 * cdef class CMEventChannel(PyverbsObject):
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         """
 *         Initialize a CMEventChannel object over an underlying rdma_event_channel
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cmid.CMEventChannel.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":336
 *         self.logger.debug('Created a CMEventChannel')
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         self.close()
 * 
*/

/* Python wrapper */
static void __pyx_pw_7pyverbs_4cmid_14CMEventChannel_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7pyverbs_4cmid_14CMEventChannel_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_7pyverbs_4cmid_14CMEventChannel_2__dealloc__(((struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7pyverbs_4cmid_14CMEventChannel_2__dealloc__(struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "pyverbs/cmid.pyx":337
 * 
 *     def __dealloc__(self):
 *         self.close()             # <<<<<<<<<<<<<<
 * 
 *     cpdef close(self):
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_7pyverbs_4cmid_CMEventChannel *)__pyx_v_self->__pyx_vtab)->close(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":336
 *         self.logger.debug('Created a CMEventChannel')
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         self.close()
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("pyverbs.cmid.CMEventChannel.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "pyverbs/cmid.pyx":339
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.event_channel != NULL:
 *             if self.logger:
*/

static PyObject *__pyx_pw_7pyverbs_4cmid_14CMEventChannel_5close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7pyverbs_4cmid_14CMEventChannel_close(struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMEventChannel &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_close); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 339, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyverbs_4cmid_14CMEventChannel_5close)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 339, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyverbs/cmid.pyx":340
 * 
 *     cpdef close(self):
 *         if self.event_channel != NULL:             # <<<<<<<<<<<<<<
 *             if self.logger:
 *                 self.logger.debug('Closing CMEventChannel')
*/
  __pyx_t_6 = (__pyx_v_self->event_channel != NULL);
  if (__pyx_t_6) {

    /* "pyverbs/cmid.pyx":341
 *     cpdef close(self):
 *         if self.event_channel != NULL:
 *             if self.logger:             # <<<<<<<<<<<<<<
 *                 self.logger.debug('Closing CMEventChannel')
 *             cm.rdma_destroy_event_channel(self.event_channel)
*/
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_self->__pyx_base.logger); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 341, __pyx_L1_error)
    if (__pyx_t_6) {

      /* "pyverbs/cmid.pyx":342
 *         if self.event_channel != NULL:
 *             if self.logger:
 *                 self.logger.debug('Closing CMEventChannel')             # <<<<<<<<<<<<<<
 *             cm.rdma_destroy_event_channel(self.event_channel)
 *             self.event_channel = NULL
*/
      __pyx_t_2 = __pyx_v_self->__pyx_base.logger;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_5 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Closing_CMEventChannel};
        __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_debug, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 342, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyverbs/cmid.pyx":341
 *     cpdef close(self):
 *         if self.event_channel != NULL:
 *             if self.logger:             # <<<<<<<<<<<<<<
 *                 self.logger.debug('Closing CMEventChannel')
 *             cm.rdma_destroy_event_channel(self.event_channel)
*/
    }

    /* "pyverbs/cmid.pyx":343
 *             if self.logger:
 *                 self.logger.debug('Closing CMEventChannel')
 *             cm.rdma_destroy_event_channel(self.event_channel)             # <<<<<<<<<<<<<<
 *             self.event_channel = NULL
 * 
*/
    rdma_destroy_event_channel(__pyx_v_self->event_channel);

    /* "pyverbs/cmid.pyx":344
 *                 self.logger.debug('Closing CMEventChannel')
 *             cm.rdma_destroy_event_channel(self.event_channel)
 *             self.event_channel = NULL             # <<<<<<<<<<<<<<
 * 
 * 
*/
    __pyx_v_self->event_channel = NULL;

    /* "pyverbs/cmid.pyx":340
 * 
 *     cpdef close(self):
 *         if self.event_channel != NULL:             # <<<<<<<<<<<<<<
 *             if self.logger:
 *                 self.logger.debug('Closing CMEventChannel')
*/
  }

  /* "pyverbs/cmid.pyx":339
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.event_channel != NULL:
 *             if self.logger:
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cmid.CMEventChannel.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_14CMEventChannel_5close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_14CMEventChannel_5close = {"close", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_14CMEventChannel_5close, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_4cmid_14CMEventChannel_5close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("close", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("close", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_4cmid_14CMEventChannel_4close(((struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_14CMEventChannel_4close(struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7pyverbs_4cmid_14CMEventChannel_close(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cmid.CMEventChannel.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.event_channel cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_14CMEventChannel_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_14CMEventChannel_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_14CMEventChannel_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_4cmid_14CMEventChannel_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_4cmid_14CMEventChannel_6__reduce_cython__(((struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_14CMEventChannel_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.event_channel cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.event_channel cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_event_channel_cannot_be_con, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.event_channel cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cmid.CMEventChannel.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.event_channel cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.event_channel cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_14CMEventChannel_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_14CMEventChannel_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_14CMEventChannel_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_4cmid_14CMEventChannel_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMEventChannel.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_4cmid_14CMEventChannel_8__setstate_cython__(((struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_14CMEventChannel_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.event_channel cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.event_channel cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_event_channel_cannot_be_con, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.event_channel cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.event_channel cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cmid.CMEventChannel.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":349
 * cdef class CMID(PyverbsCM):
 * 
 *     def __init__(self, object creator=None, QPInitAttr qp_init_attr=None,             # <<<<<<<<<<<<<<
 *                  PD pd=None, port_space=ce.RDMA_PS_TCP, CMID listen_id=None):
 *         """
*/

/* Python wrapper */
static int __pyx_pw_7pyverbs_4cmid_4CMID_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID___init__, "\n        Initialize a CMID object over an underlying rdma_cm_id C object.\n        This is the main RDMA CM object which provides most of the rdmacm API.\n        Currently only synchronous RDMA_PS_TCP communication supported.\n        Notes: User-specific context, currently not supported.\n        :param creator: For synchronous communication we need AddrInfo object in\n                        order to establish connection. We allow creator to be\n                        None for inner usage, see get_request method.\n        :param qp_init_attr: Optional initial QP attributes of CMID\n                             associated QP.\n        :param pd: Optional parameter, a PD to be associated with this CMID.\n        :param port_space: RDMA port space.\n        :param listen_id: When passive side establishes a connection, it creates\n                          a new CMID. listen_id is used to initialize the new\n                          CMID.\n        :return: CMID object for synchronous communication.\n        ");
#if CYTHON_UPDATE_DESCRIPTOR_DOC
struct wrapperbase __pyx_wrapperbase_7pyverbs_4cmid_4CMID___init__;
#endif
static int __pyx_pw_7pyverbs_4cmid_4CMID_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_creator = 0;
  struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_qp_init_attr = 0;
  struct __pyx_obj_7pyverbs_2pd_PD *__pyx_v_pd = 0;
  PyObject *__pyx_v_port_space = 0;
  struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_listen_id = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_creator,&__pyx_mstate_global->__pyx_n_u_qp_init_attr,&__pyx_mstate_global->__pyx_n_u_pd,&__pyx_mstate_global->__pyx_n_u_port_space,&__pyx_mstate_global->__pyx_n_u_listen_id,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 349, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 349, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 349, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 349, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 349, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 349, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < (0)) __PYX_ERR(0, 349, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[1]) values[1] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)Py_None));

      /* "pyverbs/cmid.pyx":350
 * 
 *     def __init__(self, object creator=None, QPInitAttr qp_init_attr=None,
 *                  PD pd=None, port_space=ce.RDMA_PS_TCP, CMID listen_id=None):             # <<<<<<<<<<<<<<
 *         """
 *         Initialize a CMID object over an underlying rdma_cm_id C object.
*/
      if (!values[2]) values[2] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2pd_PD *)Py_None));
      if (!values[3]) values[3] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__4);
      if (!values[4]) values[4] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_4cmid_CMID *)Py_None));
    } else {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 349, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 349, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 349, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 349, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 349, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }

      /* "pyverbs/cmid.pyx":349
 * cdef class CMID(PyverbsCM):
 * 
 *     def __init__(self, object creator=None, QPInitAttr qp_init_attr=None,             # <<<<<<<<<<<<<<
 *                  PD pd=None, port_space=ce.RDMA_PS_TCP, CMID listen_id=None):
 *         """
*/
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[1]) values[1] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)Py_None));

      /* "pyverbs/cmid.pyx":350
 * 
 *     def __init__(self, object creator=None, QPInitAttr qp_init_attr=None,
 *                  PD pd=None, port_space=ce.RDMA_PS_TCP, CMID listen_id=None):             # <<<<<<<<<<<<<<
 *         """
 *         Initialize a CMID object over an underlying rdma_cm_id C object.
*/
      if (!values[2]) values[2] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_2pd_PD *)Py_None));
      if (!values[3]) values[3] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__4);
      if (!values[4]) values[4] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_4cmid_CMID *)Py_None));
    }
    __pyx_v_creator = values[0];
    __pyx_v_qp_init_attr = ((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)values[1]);
    __pyx_v_pd = ((struct __pyx_obj_7pyverbs_2pd_PD *)values[2]);
    __pyx_v_port_space = values[3];
    __pyx_v_listen_id = ((struct __pyx_obj_7pyverbs_4cmid_CMID *)values[4]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 5, __pyx_nargs); __PYX_ERR(0, 349, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_qp_init_attr), __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPInitAttr, 1, "qp_init_attr", 0))) __PYX_ERR(0, 349, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_pd), __pyx_mstate_global->__pyx_ptype_7pyverbs_2pd_PD, 1, "pd", 0))) __PYX_ERR(0, 350, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_listen_id), __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, 1, "listen_id", 0))) __PYX_ERR(0, 350, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID___init__(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_creator, __pyx_v_qp_init_attr, __pyx_v_pd, __pyx_v_port_space, __pyx_v_listen_id);

  /* "pyverbs/cmid.pyx":349
 * cdef class CMID(PyverbsCM):
 * 
 *     def __init__(self, object creator=None, QPInitAttr qp_init_attr=None,             # <<<<<<<<<<<<<<
 *                  PD pd=None, port_space=ce.RDMA_PS_TCP, CMID listen_id=None):
 *         """
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyverbs_4cmid_4CMID___init__(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_creator, struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_qp_init_attr, struct __pyx_obj_7pyverbs_2pd_PD *__pyx_v_pd, PyObject *__pyx_v_port_space, struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_listen_id) {
  struct ibv_qp_init_attr *__pyx_v_init;
  struct ibv_pd *__pyx_v_in_pd;
  int __pyx_v_ret;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  void *__pyx_t_7;
  struct ibv_pd *__pyx_t_8;
  enum rdma_port_space __pyx_t_9;
  struct rdma_cm_id *__pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyverbs/cmid.pyx":369
 *         """
 *         cdef v.ibv_qp_init_attr *init
 *         cdef v.ibv_pd *in_pd = NULL             # <<<<<<<<<<<<<<
 * 
 *         super().__init__()
*/
  __pyx_v_in_pd = NULL;

  /* "pyverbs/cmid.pyx":371
 *         cdef v.ibv_pd *in_pd = NULL
 * 
 *         super().__init__()             # <<<<<<<<<<<<<<
 *         self.pd = None
 *         self.ctx = None
*/
  __pyx_t_4 = NULL;
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_builtin_super, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 371, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 371, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":372
 * 
 *         super().__init__()
 *         self.pd = None             # <<<<<<<<<<<<<<
 *         self.ctx = None
 *         self.event_channel = None
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->pd);
  __Pyx_DECREF(__pyx_v_self->pd);
  __pyx_v_self->pd = Py_None;

  /* "pyverbs/cmid.pyx":373
 *         super().__init__()
 *         self.pd = None
 *         self.ctx = None             # <<<<<<<<<<<<<<
 *         self.event_channel = None
 *         self.mrs = weakref.WeakSet()
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->ctx);
  __Pyx_DECREF(__pyx_v_self->ctx);
  __pyx_v_self->ctx = Py_None;

  /* "pyverbs/cmid.pyx":374
 *         self.pd = None
 *         self.ctx = None
 *         self.event_channel = None             # <<<<<<<<<<<<<<
 *         self.mrs = weakref.WeakSet()
 *         if creator is None:
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->event_channel);
  __Pyx_DECREF(__pyx_v_self->event_channel);
  __pyx_v_self->event_channel = Py_None;

  /* "pyverbs/cmid.pyx":375
 *         self.ctx = None
 *         self.event_channel = None
 *         self.mrs = weakref.WeakSet()             # <<<<<<<<<<<<<<
 *         if creator is None:
 *             return
*/
  __pyx_t_3 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_weakref); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_WeakSet); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_3);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->mrs);
  __Pyx_DECREF(__pyx_v_self->mrs);
  __pyx_v_self->mrs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":376
 *         self.event_channel = None
 *         self.mrs = weakref.WeakSet()
 *         if creator is None:             # <<<<<<<<<<<<<<
 *             return
 *         elif isinstance(creator, AddrInfo):
*/
  __pyx_t_6 = (__pyx_v_creator == Py_None);
  if (__pyx_t_6) {

    /* "pyverbs/cmid.pyx":377
 *         self.mrs = weakref.WeakSet()
 *         if creator is None:
 *             return             # <<<<<<<<<<<<<<
 *         elif isinstance(creator, AddrInfo):
 *             init = NULL if qp_init_attr is None else &qp_init_attr.attr
*/
    __pyx_r = 0;
    goto __pyx_L0;

    /* "pyverbs/cmid.pyx":376
 *         self.event_channel = None
 *         self.mrs = weakref.WeakSet()
 *         if creator is None:             # <<<<<<<<<<<<<<
 *             return
 *         elif isinstance(creator, AddrInfo):
*/
  }

  /* "pyverbs/cmid.pyx":378
 *         if creator is None:
 *             return
 *         elif isinstance(creator, AddrInfo):             # <<<<<<<<<<<<<<
 *             init = NULL if qp_init_attr is None else &qp_init_attr.attr
 *             if pd is not None:
*/
  __pyx_t_6 = __Pyx_TypeCheck(__pyx_v_creator, __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_AddrInfo); 
  if (__pyx_t_6) {

    /* "pyverbs/cmid.pyx":379
 *             return
 *         elif isinstance(creator, AddrInfo):
 *             init = NULL if qp_init_attr is None else &qp_init_attr.attr             # <<<<<<<<<<<<<<
 *             if pd is not None:
 *                 in_pd = pd.pd
*/
    __pyx_t_6 = (((PyObject *)__pyx_v_qp_init_attr) == Py_None);
    if (__pyx_t_6) {
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_7 = (&__pyx_v_qp_init_attr->attr);
    }
    __pyx_v_init = __pyx_t_7;

    /* "pyverbs/cmid.pyx":380
 *         elif isinstance(creator, AddrInfo):
 *             init = NULL if qp_init_attr is None else &qp_init_attr.attr
 *             if pd is not None:             # <<<<<<<<<<<<<<
 *                 in_pd = pd.pd
 *                 self.pd = pd
*/
    __pyx_t_6 = (((PyObject *)__pyx_v_pd) != Py_None);
    if (__pyx_t_6) {

      /* "pyverbs/cmid.pyx":381
 *             init = NULL if qp_init_attr is None else &qp_init_attr.attr
 *             if pd is not None:
 *                 in_pd = pd.pd             # <<<<<<<<<<<<<<
 *                 self.pd = pd
 *             ret = cm.rdma_create_ep(&self.id, (<AddrInfo>creator).addr_info,
*/
      __pyx_t_8 = __pyx_v_pd->pd;
      __pyx_v_in_pd = __pyx_t_8;

      /* "pyverbs/cmid.pyx":382
 *             if pd is not None:
 *                 in_pd = pd.pd
 *                 self.pd = pd             # <<<<<<<<<<<<<<
 *             ret = cm.rdma_create_ep(&self.id, (<AddrInfo>creator).addr_info,
 *                                     in_pd, init)
*/
      __Pyx_INCREF((PyObject *)__pyx_v_pd);
      __Pyx_GIVEREF((PyObject *)__pyx_v_pd);
      __Pyx_GOTREF(__pyx_v_self->pd);
      __Pyx_DECREF(__pyx_v_self->pd);
      __pyx_v_self->pd = ((PyObject *)__pyx_v_pd);

      /* "pyverbs/cmid.pyx":380
 *         elif isinstance(creator, AddrInfo):
 *             init = NULL if qp_init_attr is None else &qp_init_attr.attr
 *             if pd is not None:             # <<<<<<<<<<<<<<
 *                 in_pd = pd.pd
 *                 self.pd = pd
*/
    }

    /* "pyverbs/cmid.pyx":383
 *                 in_pd = pd.pd
 *                 self.pd = pd
 *             ret = cm.rdma_create_ep(&self.id, (<AddrInfo>creator).addr_info,             # <<<<<<<<<<<<<<
 *                                     in_pd, init)
 *             if ret != 0:
*/
    __pyx_v_ret = rdma_create_ep((&__pyx_v_self->id), ((struct __pyx_obj_7pyverbs_4cmid_AddrInfo *)__pyx_v_creator)->addr_info, __pyx_v_in_pd, __pyx_v_init);

    /* "pyverbs/cmid.pyx":385
 *             ret = cm.rdma_create_ep(&self.id, (<AddrInfo>creator).addr_info,
 *                                     in_pd, init)
 *             if ret != 0:             # <<<<<<<<<<<<<<
 *                 raise PyverbsRDMAErrno('Failed to create CM ID')
 *             if not (<AddrInfo>creator).addr_info.ai_flags & ce.RAI_PASSIVE:
*/
    __pyx_t_6 = (__pyx_v_ret != 0);
    if (unlikely(__pyx_t_6)) {

      /* "pyverbs/cmid.pyx":386
 *                                     in_pd, init)
 *             if ret != 0:
 *                 raise PyverbsRDMAErrno('Failed to create CM ID')             # <<<<<<<<<<<<<<
 *             if not (<AddrInfo>creator).addr_info.ai_flags & ce.RAI_PASSIVE:
 *                 self.ctx = Context(cmid=self)
*/
      __pyx_t_4 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 386, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        assert(__pyx_t_4);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
        __pyx_t_5 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_Failed_to_create_CM_ID};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 386, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 386, __pyx_L1_error)

      /* "pyverbs/cmid.pyx":385
 *             ret = cm.rdma_create_ep(&self.id, (<AddrInfo>creator).addr_info,
 *                                     in_pd, init)
 *             if ret != 0:             # <<<<<<<<<<<<<<
 *                 raise PyverbsRDMAErrno('Failed to create CM ID')
 *             if not (<AddrInfo>creator).addr_info.ai_flags & ce.RAI_PASSIVE:
*/
    }

    /* "pyverbs/cmid.pyx":387
 *             if ret != 0:
 *                 raise PyverbsRDMAErrno('Failed to create CM ID')
 *             if not (<AddrInfo>creator).addr_info.ai_flags & ce.RAI_PASSIVE:             # <<<<<<<<<<<<<<
 *                 self.ctx = Context(cmid=self)
 *                 if self.pd is None:
*/
    __pyx_t_6 = (!((((struct __pyx_obj_7pyverbs_4cmid_AddrInfo *)__pyx_v_creator)->addr_info->ai_flags & RAI_PASSIVE) != 0));
    if (__pyx_t_6) {

      /* "pyverbs/cmid.pyx":388
 *                 raise PyverbsRDMAErrno('Failed to create CM ID')
 *             if not (<AddrInfo>creator).addr_info.ai_flags & ce.RAI_PASSIVE:
 *                 self.ctx = Context(cmid=self)             # <<<<<<<<<<<<<<
 *                 if self.pd is None:
 *                     self.pd = PD(self)
*/
      __pyx_t_3 = NULL;
      __pyx_t_5 = 1;
      {
        PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_3, NULL};
        __pyx_t_4 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 388, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_cmid, ((PyObject *)__pyx_v_self), __pyx_t_4, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 388, __pyx_L1_error)
        __pyx_t_1 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_6device_Context, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_4);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 388, __pyx_L1_error)
        __Pyx_GOTREF((PyObject *)__pyx_t_1);
      }
      __Pyx_GIVEREF((PyObject *)__pyx_t_1);
      __Pyx_GOTREF(__pyx_v_self->ctx);
      __Pyx_DECREF(__pyx_v_self->ctx);
      __pyx_v_self->ctx = ((PyObject *)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "pyverbs/cmid.pyx":389
 *             if not (<AddrInfo>creator).addr_info.ai_flags & ce.RAI_PASSIVE:
 *                 self.ctx = Context(cmid=self)
 *                 if self.pd is None:             # <<<<<<<<<<<<<<
 *                     self.pd = PD(self)
 *         elif isinstance(creator, CMEventChannel):
*/
      __pyx_t_6 = (__pyx_v_self->pd == Py_None);
      if (__pyx_t_6) {

        /* "pyverbs/cmid.pyx":390
 *                 self.ctx = Context(cmid=self)
 *                 if self.pd is None:
 *                     self.pd = PD(self)             # <<<<<<<<<<<<<<
 *         elif isinstance(creator, CMEventChannel):
 *              self.event_channel = <CMEventChannel>creator
*/
        __pyx_t_4 = NULL;
        __pyx_t_5 = 1;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, ((PyObject *)__pyx_v_self)};
          __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_2pd_PD, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 390, __pyx_L1_error)
          __Pyx_GOTREF((PyObject *)__pyx_t_1);
        }
        __Pyx_GIVEREF((PyObject *)__pyx_t_1);
        __Pyx_GOTREF(__pyx_v_self->pd);
        __Pyx_DECREF(__pyx_v_self->pd);
        __pyx_v_self->pd = ((PyObject *)__pyx_t_1);
        __pyx_t_1 = 0;

        /* "pyverbs/cmid.pyx":389
 *             if not (<AddrInfo>creator).addr_info.ai_flags & ce.RAI_PASSIVE:
 *                 self.ctx = Context(cmid=self)
 *                 if self.pd is None:             # <<<<<<<<<<<<<<
 *                     self.pd = PD(self)
 *         elif isinstance(creator, CMEventChannel):
*/
      }

      /* "pyverbs/cmid.pyx":387
 *             if ret != 0:
 *                 raise PyverbsRDMAErrno('Failed to create CM ID')
 *             if not (<AddrInfo>creator).addr_info.ai_flags & ce.RAI_PASSIVE:             # <<<<<<<<<<<<<<
 *                 self.ctx = Context(cmid=self)
 *                 if self.pd is None:
*/
    }

    /* "pyverbs/cmid.pyx":378
 *         if creator is None:
 *             return
 *         elif isinstance(creator, AddrInfo):             # <<<<<<<<<<<<<<
 *             init = NULL if qp_init_attr is None else &qp_init_attr.attr
 *             if pd is not None:
*/
    goto __pyx_L3;
  }

  /* "pyverbs/cmid.pyx":391
 *                 if self.pd is None:
 *                     self.pd = PD(self)
 *         elif isinstance(creator, CMEventChannel):             # <<<<<<<<<<<<<<
 *              self.event_channel = <CMEventChannel>creator
 *              ret = cm.rdma_create_id((<CMEventChannel>creator).event_channel,
*/
  __pyx_t_6 = __Pyx_TypeCheck(__pyx_v_creator, __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMEventChannel); 
  if (__pyx_t_6) {

    /* "pyverbs/cmid.pyx":392
 *                     self.pd = PD(self)
 *         elif isinstance(creator, CMEventChannel):
 *              self.event_channel = <CMEventChannel>creator             # <<<<<<<<<<<<<<
 *              ret = cm.rdma_create_id((<CMEventChannel>creator).event_channel,
 *                                      &self.id, NULL, port_space)
*/
    __pyx_t_1 = __pyx_v_creator;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->event_channel);
    __Pyx_DECREF(__pyx_v_self->event_channel);
    __pyx_v_self->event_channel = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "pyverbs/cmid.pyx":394
 *              self.event_channel = <CMEventChannel>creator
 *              ret = cm.rdma_create_id((<CMEventChannel>creator).event_channel,
 *                                      &self.id, NULL, port_space)             # <<<<<<<<<<<<<<
 *              if ret != 0:
 *                 raise PyverbsRDMAErrno('Failed to create CM ID')
*/
    __pyx_t_9 = ((enum rdma_port_space)__Pyx_PyLong_As_enum__rdma_port_space(__pyx_v_port_space)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 394, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":393
 *         elif isinstance(creator, CMEventChannel):
 *              self.event_channel = <CMEventChannel>creator
 *              ret = cm.rdma_create_id((<CMEventChannel>creator).event_channel,             # <<<<<<<<<<<<<<
 *                                      &self.id, NULL, port_space)
 *              if ret != 0:
*/
    __pyx_v_ret = rdma_create_id(((struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *)__pyx_v_creator)->event_channel, (&__pyx_v_self->id), NULL, __pyx_t_9);

    /* "pyverbs/cmid.pyx":395
 *              ret = cm.rdma_create_id((<CMEventChannel>creator).event_channel,
 *                                      &self.id, NULL, port_space)
 *              if ret != 0:             # <<<<<<<<<<<<<<
 *                 raise PyverbsRDMAErrno('Failed to create CM ID')
 *         elif isinstance(creator, CMEvent):
*/
    __pyx_t_6 = (__pyx_v_ret != 0);
    if (unlikely(__pyx_t_6)) {

      /* "pyverbs/cmid.pyx":396
 *                                      &self.id, NULL, port_space)
 *              if ret != 0:
 *                 raise PyverbsRDMAErrno('Failed to create CM ID')             # <<<<<<<<<<<<<<
 *         elif isinstance(creator, CMEvent):
 *             if listen_id is None:
*/
      __pyx_t_4 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 396, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        assert(__pyx_t_4);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
        __pyx_t_5 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_Failed_to_create_CM_ID};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 396, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 396, __pyx_L1_error)

      /* "pyverbs/cmid.pyx":395
 *              ret = cm.rdma_create_id((<CMEventChannel>creator).event_channel,
 *                                      &self.id, NULL, port_space)
 *              if ret != 0:             # <<<<<<<<<<<<<<
 *                 raise PyverbsRDMAErrno('Failed to create CM ID')
 *         elif isinstance(creator, CMEvent):
*/
    }

    /* "pyverbs/cmid.pyx":391
 *                 if self.pd is None:
 *                     self.pd = PD(self)
 *         elif isinstance(creator, CMEventChannel):             # <<<<<<<<<<<<<<
 *              self.event_channel = <CMEventChannel>creator
 *              ret = cm.rdma_create_id((<CMEventChannel>creator).event_channel,
*/
    goto __pyx_L3;
  }

  /* "pyverbs/cmid.pyx":397
 *              if ret != 0:
 *                 raise PyverbsRDMAErrno('Failed to create CM ID')
 *         elif isinstance(creator, CMEvent):             # <<<<<<<<<<<<<<
 *             if listen_id is None:
 *                 raise PyverbsUserError('listen ID not provided')
*/
  __pyx_t_6 = __Pyx_TypeCheck(__pyx_v_creator, __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMEvent); 
  if (likely(__pyx_t_6)) {

    /* "pyverbs/cmid.pyx":398
 *                 raise PyverbsRDMAErrno('Failed to create CM ID')
 *         elif isinstance(creator, CMEvent):
 *             if listen_id is None:             # <<<<<<<<<<<<<<
 *                 raise PyverbsUserError('listen ID not provided')
 *             self.id = (<CMEvent>creator).event.id
*/
    __pyx_t_6 = (((PyObject *)__pyx_v_listen_id) == Py_None);
    if (unlikely(__pyx_t_6)) {

      /* "pyverbs/cmid.pyx":399
 *         elif isinstance(creator, CMEvent):
 *             if listen_id is None:
 *                 raise PyverbsUserError('listen ID not provided')             # <<<<<<<<<<<<<<
 *             self.id = (<CMEvent>creator).event.id
 *             self.event_channel = listen_id.event_channel
*/
      __pyx_t_3 = NULL;
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsUserError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 399, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = 1;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
        assert(__pyx_t_3);
        PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(__pyx__function);
        __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
        __pyx_t_5 = 0;
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_listen_ID_not_provided};
        __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 399, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 399, __pyx_L1_error)

      /* "pyverbs/cmid.pyx":398
 *                 raise PyverbsRDMAErrno('Failed to create CM ID')
 *         elif isinstance(creator, CMEvent):
 *             if listen_id is None:             # <<<<<<<<<<<<<<
 *                 raise PyverbsUserError('listen ID not provided')
 *             self.id = (<CMEvent>creator).event.id
*/
    }

    /* "pyverbs/cmid.pyx":400
 *             if listen_id is None:
 *                 raise PyverbsUserError('listen ID not provided')
 *             self.id = (<CMEvent>creator).event.id             # <<<<<<<<<<<<<<
 *             self.event_channel = listen_id.event_channel
 *             self.ctx = listen_id.ctx
*/
    __pyx_t_10 = ((struct __pyx_obj_7pyverbs_4cmid_CMEvent *)__pyx_v_creator)->event->id;
    __pyx_v_self->id = __pyx_t_10;

    /* "pyverbs/cmid.pyx":401
 *                 raise PyverbsUserError('listen ID not provided')
 *             self.id = (<CMEvent>creator).event.id
 *             self.event_channel = listen_id.event_channel             # <<<<<<<<<<<<<<
 *             self.ctx = listen_id.ctx
 *             self.pd = listen_id.pd
*/
    __pyx_t_1 = __pyx_v_listen_id->event_channel;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->event_channel);
    __Pyx_DECREF(__pyx_v_self->event_channel);
    __pyx_v_self->event_channel = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "pyverbs/cmid.pyx":402
 *             self.id = (<CMEvent>creator).event.id
 *             self.event_channel = listen_id.event_channel
 *             self.ctx = listen_id.ctx             # <<<<<<<<<<<<<<
 *             self.pd = listen_id.pd
 *         else:
*/
    __pyx_t_1 = __pyx_v_listen_id->ctx;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->ctx);
    __Pyx_DECREF(__pyx_v_self->ctx);
    __pyx_v_self->ctx = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "pyverbs/cmid.pyx":403
 *             self.event_channel = listen_id.event_channel
 *             self.ctx = listen_id.ctx
 *             self.pd = listen_id.pd             # <<<<<<<<<<<<<<
 *         else:
 *             raise PyverbsRDMAErrno('Cannot create CM ID from {obj}'
*/
    __pyx_t_1 = __pyx_v_listen_id->pd;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->pd);
    __Pyx_DECREF(__pyx_v_self->pd);
    __pyx_v_self->pd = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "pyverbs/cmid.pyx":397
 *              if ret != 0:
 *                 raise PyverbsRDMAErrno('Failed to create CM ID')
 *         elif isinstance(creator, CMEvent):             # <<<<<<<<<<<<<<
 *             if listen_id is None:
 *                 raise PyverbsUserError('listen ID not provided')
*/
    goto __pyx_L3;
  }

  /* "pyverbs/cmid.pyx":405
 *             self.pd = listen_id.pd
 *         else:
 *             raise PyverbsRDMAErrno('Cannot create CM ID from {obj}'             # <<<<<<<<<<<<<<
 *                                     .format(obj=type(creator)))
 * 
*/
  /*else*/ {
    __pyx_t_4 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "pyverbs/cmid.pyx":406
 *         else:
 *             raise PyverbsRDMAErrno('Cannot create CM ID from {obj}'
 *                                     .format(obj=type(creator)))             # <<<<<<<<<<<<<<
 * 
 *     cdef add_ref(self, obj):
*/
    __pyx_t_11 = __pyx_mstate_global->__pyx_kp_u_Cannot_create_CM_ID_from_obj;
    __Pyx_INCREF(__pyx_t_11);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_11, NULL};
      __pyx_t_12 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 406, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_obj, ((PyObject *)Py_TYPE(__pyx_v_creator)), __pyx_t_12, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 406, __pyx_L1_error)
      __pyx_t_2 = __Pyx_Object_VectorcallMethod_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_n_u_format, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_12);
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 406, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_3, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 405, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 405, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "pyverbs/cmid.pyx":349
 * cdef class CMID(PyverbsCM):
 * 
 *     def __init__(self, object creator=None, QPInitAttr qp_init_attr=None,             # <<<<<<<<<<<<<<
 *                  PD pd=None, port_space=ce.RDMA_PS_TCP, CMID listen_id=None):
 *         """
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":408
 *                                     .format(obj=type(creator)))
 * 
 *     cdef add_ref(self, obj):             # <<<<<<<<<<<<<<
 *         if isinstance(obj, MR):
 *             self.mrs.add(obj)
*/

static PyObject *__pyx_f_7pyverbs_4cmid_4CMID_add_ref(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_ref", 0);

  /* "pyverbs/cmid.pyx":409
 * 
 *     cdef add_ref(self, obj):
 *         if isinstance(obj, MR):             # <<<<<<<<<<<<<<
 *             self.mrs.add(obj)
 *         else:
*/
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_mstate_global->__pyx_ptype_7pyverbs_2mr_MR); 
  if (likely(__pyx_t_1)) {

    /* "pyverbs/cmid.pyx":410
 *     cdef add_ref(self, obj):
 *         if isinstance(obj, MR):
 *             self.mrs.add(obj)             # <<<<<<<<<<<<<<
 *         else:
 *             raise PyverbsError('Unrecognized object type')
*/
    __pyx_t_3 = __pyx_v_self->mrs;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_4 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_obj};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_add, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 410, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyverbs/cmid.pyx":409
 * 
 *     cdef add_ref(self, obj):
 *         if isinstance(obj, MR):             # <<<<<<<<<<<<<<
 *             self.mrs.add(obj)
 *         else:
*/
    goto __pyx_L3;
  }

  /* "pyverbs/cmid.pyx":412
 *             self.mrs.add(obj)
 *         else:
 *             raise PyverbsError('Unrecognized object type')             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  /*else*/ {
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PyverbsError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 412, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_4 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Unrecognized_object_type};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 412, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 412, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "pyverbs/cmid.pyx":408
 *                                     .format(obj=type(creator)))
 * 
 *     cdef add_ref(self, obj):             # <<<<<<<<<<<<<<
 *         if isinstance(obj, MR):
 *             self.mrs.add(obj)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.add_ref", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":414
 *             raise PyverbsError('Unrecognized object type')
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dev_name(self):
 *         return str(v.ibv_get_device_name(self.id.verbs.device).decode('utf-8'))
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_8dev_name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_8dev_name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_8dev_name___get__(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_8dev_name___get__(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char const *__pyx_t_1;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cmid.pyx":416
 *     @property
 *     def dev_name(self):
 *         return str(v.ibv_get_device_name(self.id.verbs.device).decode('utf-8'))             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ibv_get_device_name(__pyx_v_self->id->verbs->device);
  __pyx_t_2 = __Pyx_ssize_strlen(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 416, __pyx_L1_error)
  __pyx_t_3 = __Pyx_decode_c_string(__pyx_t_1, 0, __pyx_t_2, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":414
 *             raise PyverbsError('Unrecognized object type')
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dev_name(self):
 *         return str(v.ibv_get_device_name(self.id.verbs.device).decode('utf-8'))
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.dev_name.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":418
 *         return str(v.ibv_get_device_name(self.id.verbs.device).decode('utf-8'))
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def port_num(self):
 *         return self.id.port_num
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_8port_num_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_8port_num_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_8port_num___get__(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_8port_num___get__(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cmid.pyx":420
 *     @property
 *     def port_num(self):
 *         return self.id.port_num             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_uint8_t(__pyx_v_self->id->port_num); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":418
 *         return str(v.ibv_get_device_name(self.id.verbs.device).decode('utf-8'))
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def port_num(self):
 *         return self.id.port_num
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.port_num.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":422
 *         return self.id.port_num
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def event_channel(self):
 *         return self.event_channel
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_13event_channel_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_13event_channel_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_13event_channel___get__(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_13event_channel___get__(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cmid.pyx":424
 *     @property
 *     def event_channel(self):
 *         return self.event_channel             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->event_channel);
  __pyx_r = __pyx_v_self->event_channel;
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":422
 *         return self.id.port_num
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def event_channel(self):
 *         return self.event_channel
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":426
 *         return self.event_channel
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def context(self):
 *         return self.ctx
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_7context_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_7context_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_7context___get__(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_7context___get__(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cmid.pyx":428
 *     @property
 *     def context(self):
 *         return self.ctx             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->ctx);
  __pyx_r = __pyx_v_self->ctx;
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":426
 *         return self.event_channel
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def context(self):
 *         return self.ctx
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":430
 *         return self.ctx
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def pd(self):
 *         return self.pd
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_2pd_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_2pd_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_2pd___get__(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_2pd___get__(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cmid.pyx":432
 *     @property
 *     def pd(self):
 *         return self.pd             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->pd);
  __pyx_r = __pyx_v_self->pd;
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":430
 *         return self.ctx
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def pd(self):
 *         return self.pd
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":434
 *         return self.pd
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qpn(self):
 *         if self.id.qp:
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_3qpn_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_3qpn_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_3qpn___get__(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_3qpn___get__(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyverbs/cmid.pyx":436
 *     @property
 *     def qpn(self):
 *         if self.id.qp:             # <<<<<<<<<<<<<<
 *             return self.id.qp.qp_num
 *         return None
*/
  __pyx_t_1 = (__pyx_v_self->id->qp != 0);
  if (__pyx_t_1) {

    /* "pyverbs/cmid.pyx":437
 *     def qpn(self):
 *         if self.id.qp:
 *             return self.id.qp.qp_num             # <<<<<<<<<<<<<<
 *         return None
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyLong_From_unsigned_int(__pyx_v_self->id->qp->qp_num); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pyverbs/cmid.pyx":436
 *     @property
 *     def qpn(self):
 *         if self.id.qp:             # <<<<<<<<<<<<<<
 *             return self.id.qp.qp_num
 *         return None
*/
  }

  /* "pyverbs/cmid.pyx":438
 *         if self.id.qp:
 *             return self.id.qp.qp_num
 *         return None             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":434
 *         return self.pd
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def qpn(self):
 *         if self.id.qp:
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.qpn.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":440
 *         return None
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         self.close()
 * 
*/

/* Python wrapper */
static void __pyx_pw_7pyverbs_4cmid_4CMID_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7pyverbs_4cmid_4CMID_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_7pyverbs_4cmid_4CMID_2__dealloc__(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7pyverbs_4cmid_4CMID_2__dealloc__(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "pyverbs/cmid.pyx":441
 * 
 *     def __dealloc__(self):
 *         self.close()             # <<<<<<<<<<<<<<
 * 
 *     cpdef close(self):
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_7pyverbs_4cmid_CMID *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.close(((struct __pyx_obj_7pyverbs_4base_PyverbsCM *)__pyx_v_self), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":440
 *         return None
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         self.close()
 * 
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("pyverbs.cmid.CMID.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "pyverbs/cmid.pyx":443
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.id != NULL:
 *             if self.logger:
*/

static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_5close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7pyverbs_4cmid_4CMID_close(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, int __pyx_skip_dispatch) {
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_close); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 443, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyverbs_4cmid_4CMID_5close)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 443, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyverbs/cmid.pyx":444
 * 
 *     cpdef close(self):
 *         if self.id != NULL:             # <<<<<<<<<<<<<<
 *             if self.logger:
 *                 self.logger.debug('Closing CMID')
*/
  __pyx_t_6 = (__pyx_v_self->id != NULL);
  if (__pyx_t_6) {

    /* "pyverbs/cmid.pyx":445
 *     cpdef close(self):
 *         if self.id != NULL:
 *             if self.logger:             # <<<<<<<<<<<<<<
 *                 self.logger.debug('Closing CMID')
 *             if self.event_channel is None:
*/
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_self->__pyx_base.__pyx_base.logger); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 445, __pyx_L1_error)
    if (__pyx_t_6) {

      /* "pyverbs/cmid.pyx":446
 *         if self.id != NULL:
 *             if self.logger:
 *                 self.logger.debug('Closing CMID')             # <<<<<<<<<<<<<<
 *             if self.event_channel is None:
 *                 cm.rdma_destroy_ep(self.id)
*/
      __pyx_t_2 = __pyx_v_self->__pyx_base.__pyx_base.logger;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_5 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Closing_CMID};
        __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_debug, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 446, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyverbs/cmid.pyx":445
 *     cpdef close(self):
 *         if self.id != NULL:
 *             if self.logger:             # <<<<<<<<<<<<<<
 *                 self.logger.debug('Closing CMID')
 *             if self.event_channel is None:
*/
    }

    /* "pyverbs/cmid.pyx":447
 *             if self.logger:
 *                 self.logger.debug('Closing CMID')
 *             if self.event_channel is None:             # <<<<<<<<<<<<<<
 *                 cm.rdma_destroy_ep(self.id)
 *             else:
*/
    __pyx_t_6 = (__pyx_v_self->event_channel == Py_None);
    if (__pyx_t_6) {

      /* "pyverbs/cmid.pyx":448
 *                 self.logger.debug('Closing CMID')
 *             if self.event_channel is None:
 *                 cm.rdma_destroy_ep(self.id)             # <<<<<<<<<<<<<<
 *             else:
 *                 if self.id.qp != NULL:
*/
      rdma_destroy_ep(__pyx_v_self->id);

      /* "pyverbs/cmid.pyx":447
 *             if self.logger:
 *                 self.logger.debug('Closing CMID')
 *             if self.event_channel is None:             # <<<<<<<<<<<<<<
 *                 cm.rdma_destroy_ep(self.id)
 *             else:
*/
      goto __pyx_L5;
    }

    /* "pyverbs/cmid.pyx":450
 *                 cm.rdma_destroy_ep(self.id)
 *             else:
 *                 if self.id.qp != NULL:             # <<<<<<<<<<<<<<
 *                     cm.rdma_destroy_qp(self.id)
 *                 ret = cm.rdma_destroy_id(self.id)
*/
    /*else*/ {
      __pyx_t_6 = (__pyx_v_self->id->qp != NULL);
      if (__pyx_t_6) {

        /* "pyverbs/cmid.pyx":451
 *             else:
 *                 if self.id.qp != NULL:
 *                     cm.rdma_destroy_qp(self.id)             # <<<<<<<<<<<<<<
 *                 ret = cm.rdma_destroy_id(self.id)
 *                 if ret != 0:
*/
        rdma_destroy_qp(__pyx_v_self->id);

        /* "pyverbs/cmid.pyx":450
 *                 cm.rdma_destroy_ep(self.id)
 *             else:
 *                 if self.id.qp != NULL:             # <<<<<<<<<<<<<<
 *                     cm.rdma_destroy_qp(self.id)
 *                 ret = cm.rdma_destroy_id(self.id)
*/
      }

      /* "pyverbs/cmid.pyx":452
 *                 if self.id.qp != NULL:
 *                     cm.rdma_destroy_qp(self.id)
 *                 ret = cm.rdma_destroy_id(self.id)             # <<<<<<<<<<<<<<
 *                 if ret != 0:
 *                     raise PyverbsRDMAErrno('Failed to close CMID')
*/
      __pyx_v_ret = rdma_destroy_id(__pyx_v_self->id);

      /* "pyverbs/cmid.pyx":453
 *                     cm.rdma_destroy_qp(self.id)
 *                 ret = cm.rdma_destroy_id(self.id)
 *                 if ret != 0:             # <<<<<<<<<<<<<<
 *                     raise PyverbsRDMAErrno('Failed to close CMID')
 *             if self.ctx:
*/
      __pyx_t_6 = (__pyx_v_ret != 0);
      if (unlikely(__pyx_t_6)) {

        /* "pyverbs/cmid.pyx":454
 *                 ret = cm.rdma_destroy_id(self.id)
 *                 if ret != 0:
 *                     raise PyverbsRDMAErrno('Failed to close CMID')             # <<<<<<<<<<<<<<
 *             if self.ctx:
 *                 (<Context>self.ctx).context = NULL
*/
        __pyx_t_2 = NULL;
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 454, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_2);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Failed_to_close_CMID};
          __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 454, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
        }
        __Pyx_Raise(__pyx_t_1, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __PYX_ERR(0, 454, __pyx_L1_error)

        /* "pyverbs/cmid.pyx":453
 *                     cm.rdma_destroy_qp(self.id)
 *                 ret = cm.rdma_destroy_id(self.id)
 *                 if ret != 0:             # <<<<<<<<<<<<<<
 *                     raise PyverbsRDMAErrno('Failed to close CMID')
 *             if self.ctx:
*/
      }
    }
    __pyx_L5:;

    /* "pyverbs/cmid.pyx":455
 *                 if ret != 0:
 *                     raise PyverbsRDMAErrno('Failed to close CMID')
 *             if self.ctx:             # <<<<<<<<<<<<<<
 *                 (<Context>self.ctx).context = NULL
 *             if self.pd:
*/
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_self->ctx); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 455, __pyx_L1_error)
    if (__pyx_t_6) {

      /* "pyverbs/cmid.pyx":456
 *                     raise PyverbsRDMAErrno('Failed to close CMID')
 *             if self.ctx:
 *                 (<Context>self.ctx).context = NULL             # <<<<<<<<<<<<<<
 *             if self.pd:
 *                 (<PD>self.pd).pd = NULL
*/
      ((struct __pyx_obj_7pyverbs_6device_Context *)__pyx_v_self->ctx)->context = NULL;

      /* "pyverbs/cmid.pyx":455
 *                 if ret != 0:
 *                     raise PyverbsRDMAErrno('Failed to close CMID')
 *             if self.ctx:             # <<<<<<<<<<<<<<
 *                 (<Context>self.ctx).context = NULL
 *             if self.pd:
*/
    }

    /* "pyverbs/cmid.pyx":457
 *             if self.ctx:
 *                 (<Context>self.ctx).context = NULL
 *             if self.pd:             # <<<<<<<<<<<<<<
 *                 (<PD>self.pd).pd = NULL
 *             close_weakrefs([self.mrs])
*/
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_self->pd); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 457, __pyx_L1_error)
    if (__pyx_t_6) {

      /* "pyverbs/cmid.pyx":458
 *                 (<Context>self.ctx).context = NULL
 *             if self.pd:
 *                 (<PD>self.pd).pd = NULL             # <<<<<<<<<<<<<<
 *             close_weakrefs([self.mrs])
 *             self.id = NULL
*/
      ((struct __pyx_obj_7pyverbs_2pd_PD *)__pyx_v_self->pd)->pd = NULL;

      /* "pyverbs/cmid.pyx":457
 *             if self.ctx:
 *                 (<Context>self.ctx).context = NULL
 *             if self.pd:             # <<<<<<<<<<<<<<
 *                 (<PD>self.pd).pd = NULL
 *             close_weakrefs([self.mrs])
*/
    }

    /* "pyverbs/cmid.pyx":459
 *             if self.pd:
 *                 (<PD>self.pd).pd = NULL
 *             close_weakrefs([self.mrs])             # <<<<<<<<<<<<<<
 *             self.id = NULL
 * 
*/
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 459, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_self->mrs);
    __Pyx_GIVEREF(__pyx_v_self->mrs);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->mrs) != (0)) __PYX_ERR(0, 459, __pyx_L1_error);
    __pyx_t_4 = __pyx_f_7pyverbs_4base_close_weakrefs(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 459, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "pyverbs/cmid.pyx":460
 *                 (<PD>self.pd).pd = NULL
 *             close_weakrefs([self.mrs])
 *             self.id = NULL             # <<<<<<<<<<<<<<
 * 
 *     def get_request(self):
*/
    __pyx_v_self->id = NULL;

    /* "pyverbs/cmid.pyx":444
 * 
 *     cpdef close(self):
 *         if self.id != NULL:             # <<<<<<<<<<<<<<
 *             if self.logger:
 *                 self.logger.debug('Closing CMID')
*/
  }

  /* "pyverbs/cmid.pyx":443
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.id != NULL:
 *             if self.logger:
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_5close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_5close = {"close", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_5close, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_5close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("close", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("close", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_4close(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_4close(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7pyverbs_4cmid_4CMID_close(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":462
 *             self.id = NULL
 * 
 *     def get_request(self):             # <<<<<<<<<<<<<<
 *         """
 *         Retrieves the next pending connection request event. The call may only
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_7get_request(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_6get_request, "\n        Retrieves the next pending connection request event. The call may only\n        be used on listening CMIDs operating synchronously. If the call is\n        successful, a new CMID representing the connection request will be\n        returned to the user. The new CMID will reference event information\n        associated with the request until the user calls reject, accept, or\n        close on the newly created identifier.\n        :return: New CMID representing the connection request.\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_7get_request = {"get_request", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_7get_request, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_6get_request};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_7get_request(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_request (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("get_request", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("get_request", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_6get_request(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_6get_request(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self) {
  struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_to_conn = NULL;
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_request", 0);

  /* "pyverbs/cmid.pyx":472
 *         :return: New CMID representing the connection request.
 *         """
 *         to_conn = CMID()             # <<<<<<<<<<<<<<
 *         with nogil:
 *             ret = cm.rdma_get_request(self.id, &to_conn.id)
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_to_conn = ((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":473
 *         """
 *         to_conn = CMID()
 *         with nogil:             # <<<<<<<<<<<<<<
 *             ret = cm.rdma_get_request(self.id, &to_conn.id)
 *         if ret != 0:
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "pyverbs/cmid.pyx":474
 *         to_conn = CMID()
 *         with nogil:
 *             ret = cm.rdma_get_request(self.id, &to_conn.id)             # <<<<<<<<<<<<<<
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to get request, no connection established')
*/
        __pyx_v_ret = rdma_get_request(__pyx_v_self->id, (&__pyx_v_to_conn->id));
      }

      /* "pyverbs/cmid.pyx":473
 *         """
 *         to_conn = CMID()
 *         with nogil:             # <<<<<<<<<<<<<<
 *             ret = cm.rdma_get_request(self.id, &to_conn.id)
 *         if ret != 0:
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "pyverbs/cmid.pyx":475
 *         with nogil:
 *             ret = cm.rdma_get_request(self.id, &to_conn.id)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to get request, no connection established')
 *         self.ctx = Context(cmid=to_conn)
*/
  __pyx_t_4 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_4)) {

    /* "pyverbs/cmid.pyx":476
 *             ret = cm.rdma_get_request(self.id, &to_conn.id)
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to get request, no connection established')             # <<<<<<<<<<<<<<
 *         self.ctx = Context(cmid=to_conn)
 *         self.pd = PD(to_conn)
*/
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 476, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_3 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Failed_to_get_request_no_connect};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 476, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 476, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":475
 *         with nogil:
 *             ret = cm.rdma_get_request(self.id, &to_conn.id)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to get request, no connection established')
 *         self.ctx = Context(cmid=to_conn)
*/
  }

  /* "pyverbs/cmid.pyx":477
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to get request, no connection established')
 *         self.ctx = Context(cmid=to_conn)             # <<<<<<<<<<<<<<
 *         self.pd = PD(to_conn)
 *         return to_conn
*/
  __pyx_t_5 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_5, NULL};
    __pyx_t_2 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 477, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_cmid, ((PyObject *)__pyx_v_to_conn), __pyx_t_2, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 477, __pyx_L1_error)
    __pyx_t_1 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_6device_Context, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 477, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __Pyx_GIVEREF((PyObject *)__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->ctx);
  __Pyx_DECREF(__pyx_v_self->ctx);
  __pyx_v_self->ctx = ((PyObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":478
 *             raise PyverbsRDMAErrno('Failed to get request, no connection established')
 *         self.ctx = Context(cmid=to_conn)
 *         self.pd = PD(to_conn)             # <<<<<<<<<<<<<<
 *         return to_conn
 * 
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, ((PyObject *)__pyx_v_to_conn)};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_2pd_PD, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 478, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __Pyx_GIVEREF((PyObject *)__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->pd);
  __Pyx_DECREF(__pyx_v_self->pd);
  __pyx_v_self->pd = ((PyObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":479
 *         self.ctx = Context(cmid=to_conn)
 *         self.pd = PD(to_conn)
 *         return to_conn             # <<<<<<<<<<<<<<
 * 
 *     def bind_addr(self, AddrInfo lai not None):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_to_conn);
  __pyx_r = ((PyObject *)__pyx_v_to_conn);
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":462
 *             self.id = NULL
 * 
 *     def get_request(self):             # <<<<<<<<<<<<<<
 *         """
 *         Retrieves the next pending connection request event. The call may only
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.get_request", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_to_conn);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":481
 *         return to_conn
 * 
 *     def bind_addr(self, AddrInfo lai not None):             # <<<<<<<<<<<<<<
 *         """
 *         Associate a source address with a CMID. If binding to a specific local
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_9bind_addr(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_8bind_addr, "\n        Associate a source address with a CMID. If binding to a specific local\n        address, the CMID will also be bound to a local RDMA device.\n        :param lai: Local address information\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_9bind_addr = {"bind_addr", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_9bind_addr, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_8bind_addr};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_9bind_addr(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_lai = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bind_addr (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_lai,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 481, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 481, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "bind_addr", 0) < (0)) __PYX_ERR(0, 481, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("bind_addr", 1, 1, 1, i); __PYX_ERR(0, 481, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 481, __pyx_L3_error)
    }
    __pyx_v_lai = ((struct __pyx_obj_7pyverbs_4cmid_AddrInfo *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bind_addr", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 481, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.bind_addr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lai), __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_AddrInfo, 0, "lai", 0))) __PYX_ERR(0, 481, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_8bind_addr(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_lai);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_8bind_addr(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_lai) {
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bind_addr", 0);

  /* "pyverbs/cmid.pyx":488
 *         :return: None
 *         """
 *         ret = cm.rdma_bind_addr(self.id, lai.addr_info.ai_src_addr)             # <<<<<<<<<<<<<<
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Bind ID')
*/
  __pyx_v_ret = rdma_bind_addr(__pyx_v_self->id, __pyx_v_lai->addr_info->ai_src_addr);

  /* "pyverbs/cmid.pyx":489
 *         """
 *         ret = cm.rdma_bind_addr(self.id, lai.addr_info.ai_src_addr)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Bind ID')
 *         # After bind address, cm_id contains ibv_context.
*/
  __pyx_t_1 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pyverbs/cmid.pyx":490
 *         ret = cm.rdma_bind_addr(self.id, lai.addr_info.ai_src_addr)
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Bind ID')             # <<<<<<<<<<<<<<
 *         # After bind address, cm_id contains ibv_context.
 *         # Now we can create Context object.
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 490, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Failed_to_Bind_ID};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 490, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 490, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":489
 *         """
 *         ret = cm.rdma_bind_addr(self.id, lai.addr_info.ai_src_addr)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Bind ID')
 *         # After bind address, cm_id contains ibv_context.
*/
  }

  /* "pyverbs/cmid.pyx":493
 *         # After bind address, cm_id contains ibv_context.
 *         # Now we can create Context object.
 *         if self.ctx is None:             # <<<<<<<<<<<<<<
 *             self.ctx = Context(cmid=self)
 *         if self.pd is None:
*/
  __pyx_t_1 = (__pyx_v_self->ctx == Py_None);
  if (__pyx_t_1) {

    /* "pyverbs/cmid.pyx":494
 *         # Now we can create Context object.
 *         if self.ctx is None:
 *             self.ctx = Context(cmid=self)             # <<<<<<<<<<<<<<
 *         if self.pd is None:
 *             self.pd = PD(self)
*/
    __pyx_t_4 = NULL;
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_4, NULL};
      __pyx_t_3 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 494, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_cmid, ((PyObject *)__pyx_v_self), __pyx_t_3, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 494, __pyx_L1_error)
      __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_6device_Context, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_3);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 494, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_2);
    }
    __Pyx_GIVEREF((PyObject *)__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->ctx);
    __Pyx_DECREF(__pyx_v_self->ctx);
    __pyx_v_self->ctx = ((PyObject *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyverbs/cmid.pyx":493
 *         # After bind address, cm_id contains ibv_context.
 *         # Now we can create Context object.
 *         if self.ctx is None:             # <<<<<<<<<<<<<<
 *             self.ctx = Context(cmid=self)
 *         if self.pd is None:
*/
  }

  /* "pyverbs/cmid.pyx":495
 *         if self.ctx is None:
 *             self.ctx = Context(cmid=self)
 *         if self.pd is None:             # <<<<<<<<<<<<<<
 *             self.pd = PD(self)
 * 
*/
  __pyx_t_1 = (__pyx_v_self->pd == Py_None);
  if (__pyx_t_1) {

    /* "pyverbs/cmid.pyx":496
 *             self.ctx = Context(cmid=self)
 *         if self.pd is None:
 *             self.pd = PD(self)             # <<<<<<<<<<<<<<
 * 
 *     def resolve_addr(self, AddrInfo rai not None, timeout_ms=2000):
*/
    __pyx_t_3 = NULL;
    __pyx_t_5 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_v_self)};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_2pd_PD, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 496, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_2);
    }
    __Pyx_GIVEREF((PyObject *)__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->pd);
    __Pyx_DECREF(__pyx_v_self->pd);
    __pyx_v_self->pd = ((PyObject *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyverbs/cmid.pyx":495
 *         if self.ctx is None:
 *             self.ctx = Context(cmid=self)
 *         if self.pd is None:             # <<<<<<<<<<<<<<
 *             self.pd = PD(self)
 * 
*/
  }

  /* "pyverbs/cmid.pyx":481
 *         return to_conn
 * 
 *     def bind_addr(self, AddrInfo lai not None):             # <<<<<<<<<<<<<<
 *         """
 *         Associate a source address with a CMID. If binding to a specific local
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.bind_addr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":498
 *             self.pd = PD(self)
 * 
 *     def resolve_addr(self, AddrInfo rai not None, timeout_ms=2000):             # <<<<<<<<<<<<<<
 *         """
 *         Resolve destination and optional source addresses from IP addresses to
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_11resolve_addr(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_10resolve_addr, "\n        Resolve destination and optional source addresses from IP addresses to\n        an RDMA address. If successful, the specified rdma_cm_id will be bound\n        to a local device.\n        :param rai: Remote address information.\n        :param timeout_ms: Time to wait for resolution to complete [msec]\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_11resolve_addr = {"resolve_addr", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_11resolve_addr, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_10resolve_addr};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_11resolve_addr(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_rai = 0;
  PyObject *__pyx_v_timeout_ms = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("resolve_addr (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_rai,&__pyx_mstate_global->__pyx_n_u_timeout_ms,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 498, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 498, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 498, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "resolve_addr", 0) < (0)) __PYX_ERR(0, 498, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_2000));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("resolve_addr", 0, 1, 2, i); __PYX_ERR(0, 498, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 498, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 498, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_2000));
    }
    __pyx_v_rai = ((struct __pyx_obj_7pyverbs_4cmid_AddrInfo *)values[0]);
    __pyx_v_timeout_ms = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("resolve_addr", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 498, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.resolve_addr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_rai), __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_AddrInfo, 0, "rai", 0))) __PYX_ERR(0, 498, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_10resolve_addr(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_rai, __pyx_v_timeout_ms);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_10resolve_addr(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_rai, PyObject *__pyx_v_timeout_ms) {
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("resolve_addr", 0);

  /* "pyverbs/cmid.pyx":508
 *         """
 *         ret = cm.rdma_resolve_addr(self.id, rai.addr_info.ai_src_addr,
 *                                    rai.addr_info.ai_dst_addr, timeout_ms)             # <<<<<<<<<<<<<<
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Resolve Address')
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_timeout_ms); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 508, __pyx_L1_error)

  /* "pyverbs/cmid.pyx":507
 *         :return: None
 *         """
 *         ret = cm.rdma_resolve_addr(self.id, rai.addr_info.ai_src_addr,             # <<<<<<<<<<<<<<
 *                                    rai.addr_info.ai_dst_addr, timeout_ms)
 *         if ret != 0:
*/
  __pyx_v_ret = rdma_resolve_addr(__pyx_v_self->id, __pyx_v_rai->addr_info->ai_src_addr, __pyx_v_rai->addr_info->ai_dst_addr, __pyx_t_1);

  /* "pyverbs/cmid.pyx":509
 *         ret = cm.rdma_resolve_addr(self.id, rai.addr_info.ai_src_addr,
 *                                    rai.addr_info.ai_dst_addr, timeout_ms)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Resolve Address')
 * 
*/
  __pyx_t_2 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_2)) {

    /* "pyverbs/cmid.pyx":510
 *                                    rai.addr_info.ai_dst_addr, timeout_ms)
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Resolve Address')             # <<<<<<<<<<<<<<
 * 
 *     def join_multicast(self, AddrInfo addr=None, JoinMCAttrEx mc_attr=None,
*/
    __pyx_t_4 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_6 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_Failed_to_Resolve_Address};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 510, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 510, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":509
 *         ret = cm.rdma_resolve_addr(self.id, rai.addr_info.ai_src_addr,
 *                                    rai.addr_info.ai_dst_addr, timeout_ms)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Resolve Address')
 * 
*/
  }

  /* "pyverbs/cmid.pyx":498
 *             self.pd = PD(self)
 * 
 *     def resolve_addr(self, AddrInfo rai not None, timeout_ms=2000):             # <<<<<<<<<<<<<<
 *         """
 *         Resolve destination and optional source addresses from IP addresses to
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.resolve_addr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":512
 *             raise PyverbsRDMAErrno('Failed to Resolve Address')
 * 
 *     def join_multicast(self, AddrInfo addr=None, JoinMCAttrEx mc_attr=None,             # <<<<<<<<<<<<<<
 *                        context=0):
 *         """
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_13join_multicast(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_12join_multicast, "\n        Joins a multicast group and attaches an associated QP to the group.\n        :param addr: Multicast address identifying the group to join.\n        :param mc_attr: JoinMCAttrEx object is requierd to use\n                        rdma_join_multicast_ex. This object contains the join\n                        flags and the AddrInfo to join.\n        :param context: User-defined context associated with the join request.\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_13join_multicast = {"join_multicast", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_13join_multicast, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_12join_multicast};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_13join_multicast(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_addr = 0;
  struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *__pyx_v_mc_attr = 0;
  PyObject *__pyx_v_context = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("join_multicast (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_addr,&__pyx_mstate_global->__pyx_n_u_mc_attr,&__pyx_mstate_global->__pyx_n_u_context,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 512, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 512, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 512, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 512, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "join_multicast", 0) < (0)) __PYX_ERR(0, 512, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_4cmid_AddrInfo *)Py_None));
      if (!values[1]) values[1] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *)Py_None));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    } else {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 512, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 512, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 512, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_4cmid_AddrInfo *)Py_None));
      if (!values[1]) values[1] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *)Py_None));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    }
    __pyx_v_addr = ((struct __pyx_obj_7pyverbs_4cmid_AddrInfo *)values[0]);
    __pyx_v_mc_attr = ((struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *)values[1]);
    __pyx_v_context = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("join_multicast", 0, 0, 3, __pyx_nargs); __PYX_ERR(0, 512, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.join_multicast", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_addr), __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_AddrInfo, 1, "addr", 0))) __PYX_ERR(0, 512, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mc_attr), __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx, 1, "mc_attr", 0))) __PYX_ERR(0, 512, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_12join_multicast(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_addr, __pyx_v_mc_attr, __pyx_v_context);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_12join_multicast(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_addr, struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx *__pyx_v_mc_attr, PyObject *__pyx_v_context) {
  CYTHON_UNUSED struct rdma_cm_join_mc_attr_ex *__pyx_v_mc_join_attr;
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  uintptr_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("join_multicast", 0);

  /* "pyverbs/cmid.pyx":523
 *         :return: None
 *         """
 *         cdef cm.rdma_cm_join_mc_attr_ex  *mc_join_attr = NULL             # <<<<<<<<<<<<<<
 *         if not addr and not mc_attr:
 *             raise PyverbsUserError('Join to multicast must have AddrInfo or JoinMCAttrEx arguments')
*/
  __pyx_v_mc_join_attr = NULL;

  /* "pyverbs/cmid.pyx":524
 *         """
 *         cdef cm.rdma_cm_join_mc_attr_ex  *mc_join_attr = NULL
 *         if not addr and not mc_attr:             # <<<<<<<<<<<<<<
 *             raise PyverbsUserError('Join to multicast must have AddrInfo or JoinMCAttrEx arguments')
 *         if not mc_attr:
*/
  __pyx_t_2 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_addr)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 524, __pyx_L1_error)
  __pyx_t_3 = (!__pyx_t_2);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_mc_attr)); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 524, __pyx_L1_error)
  __pyx_t_2 = (!__pyx_t_3);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyverbs/cmid.pyx":525
 *         cdef cm.rdma_cm_join_mc_attr_ex  *mc_join_attr = NULL
 *         if not addr and not mc_attr:
 *             raise PyverbsUserError('Join to multicast must have AddrInfo or JoinMCAttrEx arguments')             # <<<<<<<<<<<<<<
 *         if not mc_attr:
 *             ret = cm.rdma_join_multicast(self.id, addr.addr_info.ai_src_addr,
*/
    __pyx_t_5 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_PyverbsUserError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 525, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      assert(__pyx_t_5);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_mstate_global->__pyx_kp_u_Join_to_multicast_must_have_Addr};
      __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 525, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 525, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":524
 *         """
 *         cdef cm.rdma_cm_join_mc_attr_ex  *mc_join_attr = NULL
 *         if not addr and not mc_attr:             # <<<<<<<<<<<<<<
 *             raise PyverbsUserError('Join to multicast must have AddrInfo or JoinMCAttrEx arguments')
 *         if not mc_attr:
*/
  }

  /* "pyverbs/cmid.pyx":526
 *         if not addr and not mc_attr:
 *             raise PyverbsUserError('Join to multicast must have AddrInfo or JoinMCAttrEx arguments')
 *         if not mc_attr:             # <<<<<<<<<<<<<<
 *             ret = cm.rdma_join_multicast(self.id, addr.addr_info.ai_src_addr,
 *                                          <void*><uintptr_t>context)
*/
  __pyx_t_1 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_mc_attr)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 526, __pyx_L1_error)
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "pyverbs/cmid.pyx":528
 *         if not mc_attr:
 *             ret = cm.rdma_join_multicast(self.id, addr.addr_info.ai_src_addr,
 *                                          <void*><uintptr_t>context)             # <<<<<<<<<<<<<<
 *         else:
 *             ret = cm.rdma_join_multicast_ex(self.id, &mc_attr.join_mc_attr_ex,
*/
    __pyx_t_8 = __Pyx_PyLong_As_size_t(__pyx_v_context); if (unlikely((__pyx_t_8 == ((uintptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 528, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":527
 *             raise PyverbsUserError('Join to multicast must have AddrInfo or JoinMCAttrEx arguments')
 *         if not mc_attr:
 *             ret = cm.rdma_join_multicast(self.id, addr.addr_info.ai_src_addr,             # <<<<<<<<<<<<<<
 *                                          <void*><uintptr_t>context)
 *         else:
*/
    __pyx_v_ret = rdma_join_multicast(__pyx_v_self->id, __pyx_v_addr->addr_info->ai_src_addr, ((void *)((uintptr_t)__pyx_t_8)));

    /* "pyverbs/cmid.pyx":526
 *         if not addr and not mc_attr:
 *             raise PyverbsUserError('Join to multicast must have AddrInfo or JoinMCAttrEx arguments')
 *         if not mc_attr:             # <<<<<<<<<<<<<<
 *             ret = cm.rdma_join_multicast(self.id, addr.addr_info.ai_src_addr,
 *                                          <void*><uintptr_t>context)
*/
    goto __pyx_L6;
  }

  /* "pyverbs/cmid.pyx":530
 *                                          <void*><uintptr_t>context)
 *         else:
 *             ret = cm.rdma_join_multicast_ex(self.id, &mc_attr.join_mc_attr_ex,             # <<<<<<<<<<<<<<
 *                                             <void*><uintptr_t>context)
 *         if ret != 0:
*/
  /*else*/ {

    /* "pyverbs/cmid.pyx":531
 *         else:
 *             ret = cm.rdma_join_multicast_ex(self.id, &mc_attr.join_mc_attr_ex,
 *                                             <void*><uintptr_t>context)             # <<<<<<<<<<<<<<
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Join multicast')
*/
    __pyx_t_8 = __Pyx_PyLong_As_size_t(__pyx_v_context); if (unlikely((__pyx_t_8 == ((uintptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 531, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":530
 *                                          <void*><uintptr_t>context)
 *         else:
 *             ret = cm.rdma_join_multicast_ex(self.id, &mc_attr.join_mc_attr_ex,             # <<<<<<<<<<<<<<
 *                                             <void*><uintptr_t>context)
 *         if ret != 0:
*/
    __pyx_v_ret = rdma_join_multicast_ex(__pyx_v_self->id, (&__pyx_v_mc_attr->join_mc_attr_ex), ((void *)((uintptr_t)__pyx_t_8)));
  }
  __pyx_L6:;

  /* "pyverbs/cmid.pyx":532
 *             ret = cm.rdma_join_multicast_ex(self.id, &mc_attr.join_mc_attr_ex,
 *                                             <void*><uintptr_t>context)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Join multicast')
 * 
*/
  __pyx_t_2 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_2)) {

    /* "pyverbs/cmid.pyx":533
 *                                             <void*><uintptr_t>context)
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Join multicast')             # <<<<<<<<<<<<<<
 * 
 *     def leave_multicast(self, AddrInfo addr not None):
*/
    __pyx_t_6 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_6);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_mstate_global->__pyx_kp_u_Failed_to_Join_multicast};
      __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 533, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 533, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":532
 *             ret = cm.rdma_join_multicast_ex(self.id, &mc_attr.join_mc_attr_ex,
 *                                             <void*><uintptr_t>context)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Join multicast')
 * 
*/
  }

  /* "pyverbs/cmid.pyx":512
 *             raise PyverbsRDMAErrno('Failed to Resolve Address')
 * 
 *     def join_multicast(self, AddrInfo addr=None, JoinMCAttrEx mc_attr=None,             # <<<<<<<<<<<<<<
 *                        context=0):
 *         """
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.join_multicast", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":535
 *             raise PyverbsRDMAErrno('Failed to Join multicast')
 * 
 *     def leave_multicast(self, AddrInfo addr not None):             # <<<<<<<<<<<<<<
 *         """
 *         Leaves a multicast group and detaches an associated QP from the group.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_15leave_multicast(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_14leave_multicast, "\n        Leaves a multicast group and detaches an associated QP from the group.\n        :param addr: AddrInfo object, represent the multicast address that\n                     identifies the group to leave.\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_15leave_multicast = {"leave_multicast", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_15leave_multicast, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_14leave_multicast};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_15leave_multicast(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_addr = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("leave_multicast (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_addr,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 535, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 535, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "leave_multicast", 0) < (0)) __PYX_ERR(0, 535, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("leave_multicast", 1, 1, 1, i); __PYX_ERR(0, 535, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 535, __pyx_L3_error)
    }
    __pyx_v_addr = ((struct __pyx_obj_7pyverbs_4cmid_AddrInfo *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("leave_multicast", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 535, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.leave_multicast", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_addr), __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_AddrInfo, 0, "addr", 0))) __PYX_ERR(0, 535, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_14leave_multicast(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_addr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_14leave_multicast(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_addr) {
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("leave_multicast", 0);

  /* "pyverbs/cmid.pyx":542
 *         :return: None
 *         """
 *         ret = cm.rdma_leave_multicast(self.id, addr.addr_info.ai_src_addr)             # <<<<<<<<<<<<<<
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to leave multicast')
*/
  __pyx_v_ret = rdma_leave_multicast(__pyx_v_self->id, __pyx_v_addr->addr_info->ai_src_addr);

  /* "pyverbs/cmid.pyx":543
 *         """
 *         ret = cm.rdma_leave_multicast(self.id, addr.addr_info.ai_src_addr)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to leave multicast')
 * 
*/
  __pyx_t_1 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pyverbs/cmid.pyx":544
 *         ret = cm.rdma_leave_multicast(self.id, addr.addr_info.ai_src_addr)
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to leave multicast')             # <<<<<<<<<<<<<<
 * 
 *     def resolve_route(self, timeout_ms=2000):
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 544, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Failed_to_leave_multicast};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 544, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 544, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":543
 *         """
 *         ret = cm.rdma_leave_multicast(self.id, addr.addr_info.ai_src_addr)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to leave multicast')
 * 
*/
  }

  /* "pyverbs/cmid.pyx":535
 *             raise PyverbsRDMAErrno('Failed to Join multicast')
 * 
 *     def leave_multicast(self, AddrInfo addr not None):             # <<<<<<<<<<<<<<
 *         """
 *         Leaves a multicast group and detaches an associated QP from the group.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.leave_multicast", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":546
 *             raise PyverbsRDMAErrno('Failed to leave multicast')
 * 
 *     def resolve_route(self, timeout_ms=2000):             # <<<<<<<<<<<<<<
 *         """
 *         Resolve an RDMA route to the destination address in order to establish
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_17resolve_route(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_16resolve_route, "\n        Resolve an RDMA route to the destination address in order to establish\n        a connection. The destination must already have been resolved by calling\n        resolve_addr. Thus this function is called on the client side after\n        resolve_addr but before calling connect.\n        :param timeout_ms: Time to wait for resolution to complete\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_17resolve_route = {"resolve_route", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_17resolve_route, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_16resolve_route};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_17resolve_route(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_timeout_ms = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("resolve_route (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_timeout_ms,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 546, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 546, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "resolve_route", 0) < (0)) __PYX_ERR(0, 546, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_2000));
    } else {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 546, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_2000));
    }
    __pyx_v_timeout_ms = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("resolve_route", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 546, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.resolve_route", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_16resolve_route(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_timeout_ms);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_16resolve_route(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_timeout_ms) {
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("resolve_route", 0);

  /* "pyverbs/cmid.pyx":555
 *         :return: None
 *         """
 *         ret = cm.rdma_resolve_route(self.id, timeout_ms)             # <<<<<<<<<<<<<<
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Resolve Route')
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_timeout_ms); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 555, __pyx_L1_error)
  __pyx_v_ret = rdma_resolve_route(__pyx_v_self->id, __pyx_t_1);

  /* "pyverbs/cmid.pyx":556
 *         """
 *         ret = cm.rdma_resolve_route(self.id, timeout_ms)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Resolve Route')
 *         # After resolve route, cm_id contains ibv_context.
*/
  __pyx_t_2 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_2)) {

    /* "pyverbs/cmid.pyx":557
 *         ret = cm.rdma_resolve_route(self.id, timeout_ms)
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Resolve Route')             # <<<<<<<<<<<<<<
 *         # After resolve route, cm_id contains ibv_context.
 *         # Now we can create Context object.
*/
    __pyx_t_4 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 557, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_6 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_Failed_to_Resolve_Route};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 557, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 557, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":556
 *         """
 *         ret = cm.rdma_resolve_route(self.id, timeout_ms)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Resolve Route')
 *         # After resolve route, cm_id contains ibv_context.
*/
  }

  /* "pyverbs/cmid.pyx":560
 *         # After resolve route, cm_id contains ibv_context.
 *         # Now we can create Context object.
 *         if self.ctx is None:             # <<<<<<<<<<<<<<
 *             self.ctx = Context(cmid=self)
 *         if self.pd is None:
*/
  __pyx_t_2 = (__pyx_v_self->ctx == Py_None);
  if (__pyx_t_2) {

    /* "pyverbs/cmid.pyx":561
 *         # Now we can create Context object.
 *         if self.ctx is None:
 *             self.ctx = Context(cmid=self)             # <<<<<<<<<<<<<<
 *         if self.pd is None:
 *             self.pd = PD(self)
*/
    __pyx_t_5 = NULL;
    __pyx_t_6 = 1;
    {
      PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_5, NULL};
      __pyx_t_4 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 561, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_cmid, ((PyObject *)__pyx_v_self), __pyx_t_4, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 561, __pyx_L1_error)
      __pyx_t_3 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_6device_Context, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_4);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 561, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_3);
    }
    __Pyx_GIVEREF((PyObject *)__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->ctx);
    __Pyx_DECREF(__pyx_v_self->ctx);
    __pyx_v_self->ctx = ((PyObject *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyverbs/cmid.pyx":560
 *         # After resolve route, cm_id contains ibv_context.
 *         # Now we can create Context object.
 *         if self.ctx is None:             # <<<<<<<<<<<<<<
 *             self.ctx = Context(cmid=self)
 *         if self.pd is None:
*/
  }

  /* "pyverbs/cmid.pyx":562
 *         if self.ctx is None:
 *             self.ctx = Context(cmid=self)
 *         if self.pd is None:             # <<<<<<<<<<<<<<
 *             self.pd = PD(self)
 * 
*/
  __pyx_t_2 = (__pyx_v_self->pd == Py_None);
  if (__pyx_t_2) {

    /* "pyverbs/cmid.pyx":563
 *             self.ctx = Context(cmid=self)
 *         if self.pd is None:
 *             self.pd = PD(self)             # <<<<<<<<<<<<<<
 * 
 *     def listen(self, backlog=0):
*/
    __pyx_t_4 = NULL;
    __pyx_t_6 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, ((PyObject *)__pyx_v_self)};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_2pd_PD, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 563, __pyx_L1_error)
      __Pyx_GOTREF((PyObject *)__pyx_t_3);
    }
    __Pyx_GIVEREF((PyObject *)__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->pd);
    __Pyx_DECREF(__pyx_v_self->pd);
    __pyx_v_self->pd = ((PyObject *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyverbs/cmid.pyx":562
 *         if self.ctx is None:
 *             self.ctx = Context(cmid=self)
 *         if self.pd is None:             # <<<<<<<<<<<<<<
 *             self.pd = PD(self)
 * 
*/
  }

  /* "pyverbs/cmid.pyx":546
 *             raise PyverbsRDMAErrno('Failed to leave multicast')
 * 
 *     def resolve_route(self, timeout_ms=2000):             # <<<<<<<<<<<<<<
 *         """
 *         Resolve an RDMA route to the destination address in order to establish
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.resolve_route", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":565
 *             self.pd = PD(self)
 * 
 *     def listen(self, backlog=0):             # <<<<<<<<<<<<<<
 *         """
 *         Listen for incoming connection requests or datagram service lookup.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_19listen(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_18listen, "\n        Listen for incoming connection requests or datagram service lookup.\n        The listen is restricted to the locally bound source address.\n        :param backlog: The backlog of incoming connection requests\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_19listen = {"listen", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_19listen, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_18listen};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_19listen(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_backlog = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("listen (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_backlog,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 565, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 565, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "listen", 0) < (0)) __PYX_ERR(0, 565, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    } else {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 565, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    }
    __pyx_v_backlog = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("listen", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 565, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.listen", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_18listen(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_backlog);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_18listen(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_backlog) {
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("listen", 0);

  /* "pyverbs/cmid.pyx":572
 *         :return: None
 *         """
 *         ret = cm.rdma_listen(self.id, backlog)             # <<<<<<<<<<<<<<
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Listen Failed')
*/
  __pyx_t_1 = __Pyx_PyLong_As_int(__pyx_v_backlog); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 572, __pyx_L1_error)
  __pyx_v_ret = rdma_listen(__pyx_v_self->id, __pyx_t_1);

  /* "pyverbs/cmid.pyx":573
 *         """
 *         ret = cm.rdma_listen(self.id, backlog)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Listen Failed')
 * 
*/
  __pyx_t_2 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_2)) {

    /* "pyverbs/cmid.pyx":574
 *         ret = cm.rdma_listen(self.id, backlog)
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Listen Failed')             # <<<<<<<<<<<<<<
 * 
 *     def connect(self, ConnParam param=None):
*/
    __pyx_t_4 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_6 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_Listen_Failed};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 574, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 574, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":573
 *         """
 *         ret = cm.rdma_listen(self.id, backlog)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Listen Failed')
 * 
*/
  }

  /* "pyverbs/cmid.pyx":565
 *             self.pd = PD(self)
 * 
 *     def listen(self, backlog=0):             # <<<<<<<<<<<<<<
 *         """
 *         Listen for incoming connection requests or datagram service lookup.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.listen", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":576
 *             raise PyverbsRDMAErrno('Listen Failed')
 * 
 *     def connect(self, ConnParam param=None):             # <<<<<<<<<<<<<<
 *         """
 *         Initiates an active connection request to a remote destination.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_21connect(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_20connect, "\n        Initiates an active connection request to a remote destination.\n        :param param: Optional connection parameters\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_21connect = {"connect", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_21connect, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_20connect};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_21connect(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_4cmid_ConnParam *__pyx_v_param = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("connect (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_param,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 576, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 576, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "connect", 0) < (0)) __PYX_ERR(0, 576, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_4cmid_ConnParam *)Py_None));
    } else {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 576, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_4cmid_ConnParam *)Py_None));
    }
    __pyx_v_param = ((struct __pyx_obj_7pyverbs_4cmid_ConnParam *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("connect", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 576, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.connect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_param), __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_ConnParam, 1, "param", 0))) __PYX_ERR(0, 576, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_20connect(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_param);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_20connect(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_4cmid_ConnParam *__pyx_v_param) {
  struct rdma_conn_param *__pyx_v_conn;
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct rdma_conn_param *__pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("connect", 0);

  /* "pyverbs/cmid.pyx":582
 *         :return: None
 *         """
 *         cdef cm.rdma_conn_param *conn = &param.conn_param if param else NULL             # <<<<<<<<<<<<<<
 *         ret = cm.rdma_connect(self.id, conn)
 *         if ret != 0:
*/
  __pyx_t_2 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_param)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 582, __pyx_L1_error)
  if (__pyx_t_2) {
    __pyx_t_1 = (&__pyx_v_param->conn_param);
  } else {
    __pyx_t_1 = NULL;
  }
  __pyx_v_conn = __pyx_t_1;

  /* "pyverbs/cmid.pyx":583
 *         """
 *         cdef cm.rdma_conn_param *conn = &param.conn_param if param else NULL
 *         ret = cm.rdma_connect(self.id, conn)             # <<<<<<<<<<<<<<
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Connect')
*/
  __pyx_v_ret = rdma_connect(__pyx_v_self->id, __pyx_v_conn);

  /* "pyverbs/cmid.pyx":584
 *         cdef cm.rdma_conn_param *conn = &param.conn_param if param else NULL
 *         ret = cm.rdma_connect(self.id, conn)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Connect')
 * 
*/
  __pyx_t_2 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_2)) {

    /* "pyverbs/cmid.pyx":585
 *         ret = cm.rdma_connect(self.id, conn)
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Connect')             # <<<<<<<<<<<<<<
 * 
 *     def disconnect(self):
*/
    __pyx_t_4 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 585, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_6 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_Failed_to_Connect};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 585, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 585, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":584
 *         cdef cm.rdma_conn_param *conn = &param.conn_param if param else NULL
 *         ret = cm.rdma_connect(self.id, conn)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Connect')
 * 
*/
  }

  /* "pyverbs/cmid.pyx":576
 *             raise PyverbsRDMAErrno('Listen Failed')
 * 
 *     def connect(self, ConnParam param=None):             # <<<<<<<<<<<<<<
 *         """
 *         Initiates an active connection request to a remote destination.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.connect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":587
 *             raise PyverbsRDMAErrno('Failed to Connect')
 * 
 *     def disconnect(self):             # <<<<<<<<<<<<<<
 *         """
 *         Disconnects a connection and transitions any associated QP to error
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_23disconnect(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_22disconnect, "\n        Disconnects a connection and transitions any associated QP to error\n        state.\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_23disconnect = {"disconnect", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_23disconnect, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_22disconnect};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_23disconnect(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("disconnect (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("disconnect", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("disconnect", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_22disconnect(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_22disconnect(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self) {
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("disconnect", 0);

  /* "pyverbs/cmid.pyx":593
 *         :return: None
 *         """
 *         ret = cm.rdma_disconnect(self.id)             # <<<<<<<<<<<<<<
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Disconnect')
*/
  __pyx_v_ret = rdma_disconnect(__pyx_v_self->id);

  /* "pyverbs/cmid.pyx":594
 *         """
 *         ret = cm.rdma_disconnect(self.id)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Disconnect')
 * 
*/
  __pyx_t_1 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pyverbs/cmid.pyx":595
 *         ret = cm.rdma_disconnect(self.id)
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Disconnect')             # <<<<<<<<<<<<<<
 * 
 *     def accept(self, ConnParam param=None):
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 595, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Failed_to_Disconnect};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 595, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 595, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":594
 *         """
 *         ret = cm.rdma_disconnect(self.id)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Disconnect')
 * 
*/
  }

  /* "pyverbs/cmid.pyx":587
 *             raise PyverbsRDMAErrno('Failed to Connect')
 * 
 *     def disconnect(self):             # <<<<<<<<<<<<<<
 *         """
 *         Disconnects a connection and transitions any associated QP to error
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.disconnect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":597
 *             raise PyverbsRDMAErrno('Failed to Disconnect')
 * 
 *     def accept(self, ConnParam param=None):             # <<<<<<<<<<<<<<
 *         """
 *         Is called from the listening side to accept a connection or datagram
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_25accept(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_24accept, "\n        Is called from the listening side to accept a connection or datagram\n        service lookup request.\n        :param param: Optional connection parameters\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_25accept = {"accept", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_25accept, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_24accept};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_25accept(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_4cmid_ConnParam *__pyx_v_param = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("accept (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_param,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 597, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 597, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "accept", 0) < (0)) __PYX_ERR(0, 597, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_4cmid_ConnParam *)Py_None));
    } else {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 597, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef((PyObject *)((struct __pyx_obj_7pyverbs_4cmid_ConnParam *)Py_None));
    }
    __pyx_v_param = ((struct __pyx_obj_7pyverbs_4cmid_ConnParam *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("accept", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 597, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_param), __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_ConnParam, 1, "param", 0))) __PYX_ERR(0, 597, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_24accept(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_param);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_24accept(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_4cmid_ConnParam *__pyx_v_param) {
  struct rdma_conn_param *__pyx_v_conn;
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct rdma_conn_param *__pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("accept", 0);

  /* "pyverbs/cmid.pyx":604
 *         :return: None
 *         """
 *         cdef cm.rdma_conn_param *conn = &param.conn_param if param else NULL             # <<<<<<<<<<<<<<
 *         ret = cm.rdma_accept(self.id, conn)
 *         if ret != 0:
*/
  __pyx_t_2 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_param)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 604, __pyx_L1_error)
  if (__pyx_t_2) {
    __pyx_t_1 = (&__pyx_v_param->conn_param);
  } else {
    __pyx_t_1 = NULL;
  }
  __pyx_v_conn = __pyx_t_1;

  /* "pyverbs/cmid.pyx":605
 *         """
 *         cdef cm.rdma_conn_param *conn = &param.conn_param if param else NULL
 *         ret = cm.rdma_accept(self.id, conn)             # <<<<<<<<<<<<<<
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Accept Connection')
*/
  __pyx_v_ret = rdma_accept(__pyx_v_self->id, __pyx_v_conn);

  /* "pyverbs/cmid.pyx":606
 *         cdef cm.rdma_conn_param *conn = &param.conn_param if param else NULL
 *         ret = cm.rdma_accept(self.id, conn)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Accept Connection')
 * 
*/
  __pyx_t_2 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_2)) {

    /* "pyverbs/cmid.pyx":607
 *         ret = cm.rdma_accept(self.id, conn)
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Accept Connection')             # <<<<<<<<<<<<<<
 * 
 *     def establish(self):
*/
    __pyx_t_4 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 607, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_6 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_Failed_to_Accept_Connection};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 607, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 607, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":606
 *         cdef cm.rdma_conn_param *conn = &param.conn_param if param else NULL
 *         ret = cm.rdma_accept(self.id, conn)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Accept Connection')
 * 
*/
  }

  /* "pyverbs/cmid.pyx":597
 *             raise PyverbsRDMAErrno('Failed to Disconnect')
 * 
 *     def accept(self, ConnParam param=None):             # <<<<<<<<<<<<<<
 *         """
 *         Is called from the listening side to accept a connection or datagram
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.accept", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":609
 *             raise PyverbsRDMAErrno('Failed to Accept Connection')
 * 
 *     def establish(self):             # <<<<<<<<<<<<<<
 *         """
 *         Complete an active connection request.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_27establish(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_26establish, "\n        Complete an active connection request.\n        If a QP has not been created on the CMID, this method should be\n        called by the active side to complete the connection, after getting\n        connect response event. This will trigger a connection established\n        event on the passive side.\n        This method should not be used on a CMID on which a QP has been\n        created.\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_27establish = {"establish", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_27establish, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_26establish};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_27establish(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("establish (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("establish", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("establish", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_26establish(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_26establish(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self) {
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("establish", 0);

  /* "pyverbs/cmid.pyx":619
 *         created.
 *         """
 *         ret = cm.rdma_establish(self.id)             # <<<<<<<<<<<<<<
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Complete an active connection request')
*/
  __pyx_v_ret = rdma_establish(__pyx_v_self->id);

  /* "pyverbs/cmid.pyx":620
 *         """
 *         ret = cm.rdma_establish(self.id)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Complete an active connection request')
 * 
*/
  __pyx_t_1 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pyverbs/cmid.pyx":621
 *         ret = cm.rdma_establish(self.id)
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Complete an active connection request')             # <<<<<<<<<<<<<<
 * 
 *     def set_local_ece(self, ECE ece):
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Failed_to_Complete_an_active_con};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 621, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 621, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":620
 *         """
 *         ret = cm.rdma_establish(self.id)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Complete an active connection request')
 * 
*/
  }

  /* "pyverbs/cmid.pyx":609
 *             raise PyverbsRDMAErrno('Failed to Accept Connection')
 * 
 *     def establish(self):             # <<<<<<<<<<<<<<
 *         """
 *         Complete an active connection request.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.establish", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":623
 *             raise PyverbsRDMAErrno('Failed to Complete an active connection request')
 * 
 *     def set_local_ece(self, ECE ece):             # <<<<<<<<<<<<<<
 *         """
 *         Set local ECE paraemters to be used for REQ/REP communication.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_29set_local_ece(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_28set_local_ece, "\n        Set local ECE paraemters to be used for REQ/REP communication.\n        :param ece: ECE object with the requested configuration\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_29set_local_ece = {"set_local_ece", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_29set_local_ece, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_28set_local_ece};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_29set_local_ece(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_ece = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_local_ece (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_ece,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 623, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 623, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "set_local_ece", 0) < (0)) __PYX_ERR(0, 623, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("set_local_ece", 1, 1, 1, i); __PYX_ERR(0, 623, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 623, __pyx_L3_error)
    }
    __pyx_v_ece = ((struct __pyx_obj_7pyverbs_2qp_ECE *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_local_ece", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 623, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.set_local_ece", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_ece), __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_ECE, 1, "ece", 0))) __PYX_ERR(0, 623, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_28set_local_ece(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_ece);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_28set_local_ece(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_ece) {
  int __pyx_v_rc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_local_ece", 0);

  /* "pyverbs/cmid.pyx":629
 *         :return: None
 *         """
 *         rc = cm.rdma_set_local_ece(self.id, &ece.ece)             # <<<<<<<<<<<<<<
 *         if rc != 0:
 *             raise PyverbsRDMAErrno('Failed to set local ECE')
*/
  __pyx_v_rc = rdma_set_local_ece(__pyx_v_self->id, (&__pyx_v_ece->ece));

  /* "pyverbs/cmid.pyx":630
 *         """
 *         rc = cm.rdma_set_local_ece(self.id, &ece.ece)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to set local ECE')
 * 
*/
  __pyx_t_1 = (__pyx_v_rc != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pyverbs/cmid.pyx":631
 *         rc = cm.rdma_set_local_ece(self.id, &ece.ece)
 *         if rc != 0:
 *             raise PyverbsRDMAErrno('Failed to set local ECE')             # <<<<<<<<<<<<<<
 * 
 *     def get_remote_ece(self):
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 631, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Failed_to_set_local_ECE};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 631, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 631, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":630
 *         """
 *         rc = cm.rdma_set_local_ece(self.id, &ece.ece)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to set local ECE')
 * 
*/
  }

  /* "pyverbs/cmid.pyx":623
 *             raise PyverbsRDMAErrno('Failed to Complete an active connection request')
 * 
 *     def set_local_ece(self, ECE ece):             # <<<<<<<<<<<<<<
 *         """
 *         Set local ECE paraemters to be used for REQ/REP communication.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.set_local_ece", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":633
 *             raise PyverbsRDMAErrno('Failed to set local ECE')
 * 
 *     def get_remote_ece(self):             # <<<<<<<<<<<<<<
 *         """
 *         Get ECE parameters as were received from the communication peer.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_31get_remote_ece(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_30get_remote_ece, "\n        Get ECE parameters as were received from the communication peer.\n        :return: ECE object with the ece configuration\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_31get_remote_ece = {"get_remote_ece", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_31get_remote_ece, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_30get_remote_ece};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_31get_remote_ece(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_remote_ece (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("get_remote_ece", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("get_remote_ece", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_30get_remote_ece(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_30get_remote_ece(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self) {
  struct __pyx_obj_7pyverbs_2qp_ECE *__pyx_v_ece = NULL;
  int __pyx_v_rc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_remote_ece", 0);

  /* "pyverbs/cmid.pyx":638
 *         :return: ECE object with the ece configuration
 *         """
 *         ece = ECE()             # <<<<<<<<<<<<<<
 *         rc = cm.rdma_get_remote_ece(self.id, &ece.ece)
 *         if rc != 0:
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_ECE, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 638, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_ece = ((struct __pyx_obj_7pyverbs_2qp_ECE *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":639
 *         """
 *         ece = ECE()
 *         rc = cm.rdma_get_remote_ece(self.id, &ece.ece)             # <<<<<<<<<<<<<<
 *         if rc != 0:
 *             raise PyverbsRDMAErrno('Failed to get remote ECE')
*/
  __pyx_v_rc = rdma_get_remote_ece(__pyx_v_self->id, (&__pyx_v_ece->ece));

  /* "pyverbs/cmid.pyx":640
 *         ece = ECE()
 *         rc = cm.rdma_get_remote_ece(self.id, &ece.ece)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to get remote ECE')
 *         return ece
*/
  __pyx_t_4 = (__pyx_v_rc != 0);
  if (unlikely(__pyx_t_4)) {

    /* "pyverbs/cmid.pyx":641
 *         rc = cm.rdma_get_remote_ece(self.id, &ece.ece)
 *         if rc != 0:
 *             raise PyverbsRDMAErrno('Failed to get remote ECE')             # <<<<<<<<<<<<<<
 *         return ece
 * 
*/
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 641, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_3 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Failed_to_get_remote_ECE};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 641, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 641, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":640
 *         ece = ECE()
 *         rc = cm.rdma_get_remote_ece(self.id, &ece.ece)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to get remote ECE')
 *         return ece
*/
  }

  /* "pyverbs/cmid.pyx":642
 *         if rc != 0:
 *             raise PyverbsRDMAErrno('Failed to get remote ECE')
 *         return ece             # <<<<<<<<<<<<<<
 * 
 *     def create_qp(self, QPInitAttr qp_init not None):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_ece);
  __pyx_r = ((PyObject *)__pyx_v_ece);
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":633
 *             raise PyverbsRDMAErrno('Failed to set local ECE')
 * 
 *     def get_remote_ece(self):             # <<<<<<<<<<<<<<
 *         """
 *         Get ECE parameters as were received from the communication peer.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.get_remote_ece", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_ece);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":644
 *         return ece
 * 
 *     def create_qp(self, QPInitAttr qp_init not None):             # <<<<<<<<<<<<<<
 *         """
 *         Create a QP, which is associated with CMID.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_33create_qp(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_32create_qp, "\n        Create a QP, which is associated with CMID.\n        If CMID and qp_init don't hold any CQs, new CQs will be created and\n        associated with CMID.\n        If only qp_init provides CQs, they will not be associated with CMID.\n        If both provide CQs they have to be using the same CQs.\n        :param qp_init: QP init attributes\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_33create_qp = {"create_qp", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_33create_qp, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_32create_qp};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_33create_qp(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_qp_init = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_qp (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_qp_init,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 644, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 644, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "create_qp", 0) < (0)) __PYX_ERR(0, 644, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("create_qp", 1, 1, 1, i); __PYX_ERR(0, 644, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 644, __pyx_L3_error)
    }
    __pyx_v_qp_init = ((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_qp", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 644, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.create_qp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_qp_init), __pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPInitAttr, 0, "qp_init", 0))) __PYX_ERR(0, 644, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_32create_qp(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_qp_init);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_32create_qp(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_qp_init) {
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create_qp", 0);

  /* "pyverbs/cmid.pyx":653
 *         :param qp_init: QP init attributes
 *         """
 *         ret = cm.rdma_create_qp(self.id, (<PD>self.pd).pd, &qp_init.attr)             # <<<<<<<<<<<<<<
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Create QP')
*/
  __pyx_v_ret = rdma_create_qp(__pyx_v_self->id, ((struct __pyx_obj_7pyverbs_2pd_PD *)__pyx_v_self->pd)->pd, (&__pyx_v_qp_init->attr));

  /* "pyverbs/cmid.pyx":654
 *         """
 *         ret = cm.rdma_create_qp(self.id, (<PD>self.pd).pd, &qp_init.attr)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Create QP')
 * 
*/
  __pyx_t_1 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pyverbs/cmid.pyx":655
 *         ret = cm.rdma_create_qp(self.id, (<PD>self.pd).pd, &qp_init.attr)
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Create QP')             # <<<<<<<<<<<<<<
 * 
 *     def query_qp(self, attr_mask):
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 655, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Failed_to_Create_QP};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 655, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 655, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":654
 *         """
 *         ret = cm.rdma_create_qp(self.id, (<PD>self.pd).pd, &qp_init.attr)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Create QP')
 * 
*/
  }

  /* "pyverbs/cmid.pyx":644
 *         return ece
 * 
 *     def create_qp(self, QPInitAttr qp_init not None):             # <<<<<<<<<<<<<<
 *         """
 *         Create a QP, which is associated with CMID.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.create_qp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":657
 *             raise PyverbsRDMAErrno('Failed to Create QP')
 * 
 *     def query_qp(self, attr_mask):             # <<<<<<<<<<<<<<
 *         """
 *         Query QP using ibv_query_qp.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_35query_qp(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_34query_qp, "\n        Query QP using ibv_query_qp.\n        :param attr_mask: Which attributes to query (use <enum name> enum)\n        :return: A (QPAttr, QPInitAttr) tuple, containing the relevant QP info\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_35query_qp = {"query_qp", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_35query_qp, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_34query_qp};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_35query_qp(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_attr_mask = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("query_qp (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_attr_mask,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 657, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 657, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "query_qp", 0) < (0)) __PYX_ERR(0, 657, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("query_qp", 1, 1, 1, i); __PYX_ERR(0, 657, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 657, __pyx_L3_error)
    }
    __pyx_v_attr_mask = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("query_qp", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 657, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.query_qp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_34query_qp(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_attr_mask);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_34query_qp(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_attr_mask) {
  struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_attr = NULL;
  struct __pyx_obj_7pyverbs_2qp_QPInitAttr *__pyx_v_init_attr = NULL;
  int __pyx_v_rc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("query_qp", 0);

  /* "pyverbs/cmid.pyx":663
 *         :return: A (QPAttr, QPInitAttr) tuple, containing the relevant QP info
 *         """
 *         attr = QPAttr()             # <<<<<<<<<<<<<<
 *         init_attr = QPInitAttr()
 *         rc = v.ibv_query_qp(self.id.qp, &attr.attr, attr_mask, &init_attr.attr)
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPAttr, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 663, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_attr = ((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":664
 *         """
 *         attr = QPAttr()
 *         init_attr = QPInitAttr()             # <<<<<<<<<<<<<<
 *         rc = v.ibv_query_qp(self.id.qp, &attr.attr, attr_mask, &init_attr.attr)
 *         if rc != 0:
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPInitAttr, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 664, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_init_attr = ((struct __pyx_obj_7pyverbs_2qp_QPInitAttr *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":665
 *         attr = QPAttr()
 *         init_attr = QPInitAttr()
 *         rc = v.ibv_query_qp(self.id.qp, &attr.attr, attr_mask, &init_attr.attr)             # <<<<<<<<<<<<<<
 *         if rc != 0:
 *             raise PyverbsRDMAError('Failed to query QP', rc)
*/
  __pyx_t_4 = __Pyx_PyLong_As_int(__pyx_v_attr_mask); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 665, __pyx_L1_error)
  __pyx_v_rc = ibv_query_qp(__pyx_v_self->id->qp, (&__pyx_v_attr->attr), __pyx_t_4, (&__pyx_v_init_attr->attr));

  /* "pyverbs/cmid.pyx":666
 *         init_attr = QPInitAttr()
 *         rc = v.ibv_query_qp(self.id.qp, &attr.attr, attr_mask, &init_attr.attr)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Failed to query QP', rc)
 *         return attr, init_attr
*/
  __pyx_t_5 = (__pyx_v_rc != 0);
  if (unlikely(__pyx_t_5)) {

    /* "pyverbs/cmid.pyx":667
 *         rc = v.ibv_query_qp(self.id.qp, &attr.attr, attr_mask, &init_attr.attr)
 *         if rc != 0:
 *             raise PyverbsRDMAError('Failed to query QP', rc)             # <<<<<<<<<<<<<<
 *         return attr, init_attr
 * 
*/
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyLong_From_int(__pyx_v_rc); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
      __pyx_t_3 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Failed_to_query_QP, __pyx_t_7};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_3, (3-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 667, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 667, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":666
 *         init_attr = QPInitAttr()
 *         rc = v.ibv_query_qp(self.id.qp, &attr.attr, attr_mask, &init_attr.attr)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAError('Failed to query QP', rc)
 *         return attr, init_attr
*/
  }

  /* "pyverbs/cmid.pyx":668
 *         if rc != 0:
 *             raise PyverbsRDMAError('Failed to query QP', rc)
 *         return attr, init_attr             # <<<<<<<<<<<<<<
 * 
 *     def init_qp_attr(self, qp_state):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_v_attr);
  __Pyx_GIVEREF((PyObject *)__pyx_v_attr);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_attr)) != (0)) __PYX_ERR(0, 668, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_init_attr);
  __Pyx_GIVEREF((PyObject *)__pyx_v_init_attr);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_init_attr)) != (0)) __PYX_ERR(0, 668, __pyx_L1_error);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":657
 *             raise PyverbsRDMAErrno('Failed to Create QP')
 * 
 *     def query_qp(self, attr_mask):             # <<<<<<<<<<<<<<
 *         """
 *         Query QP using ibv_query_qp.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.query_qp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_attr);
  __Pyx_XDECREF((PyObject *)__pyx_v_init_attr);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":670
 *         return attr, init_attr
 * 
 *     def init_qp_attr(self, qp_state):             # <<<<<<<<<<<<<<
 *         """
 *         Initialize a QPAttr object used for state transitions of an external
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_37init_qp_attr(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_36init_qp_attr, "\n        Initialize a QPAttr object used for state transitions of an external\n        QP (a QP which was not created using CMID).\n        When connecting external QPs using CMIDs both sides must call this\n        method before QP state transition to RTR/RTS in order to obtain\n        relevant QP attributes from CMID.\n        :param qp_state: The QP's destination state\n        :return: A (QPAttr, attr_mask) tuple, where attr_mask defines which\n                 attributes of QPAttr are valid\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_37init_qp_attr = {"init_qp_attr", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_37init_qp_attr, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_36init_qp_attr};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_37init_qp_attr(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_qp_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("init_qp_attr (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_qp_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 670, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 670, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "init_qp_attr", 0) < (0)) __PYX_ERR(0, 670, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("init_qp_attr", 1, 1, 1, i); __PYX_ERR(0, 670, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 670, __pyx_L3_error)
    }
    __pyx_v_qp_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("init_qp_attr", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 670, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.init_qp_attr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_36init_qp_attr(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_qp_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_36init_qp_attr(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_qp_state) {
  int __pyx_v_attr_mask;
  struct __pyx_obj_7pyverbs_2qp_QPAttr *__pyx_v_qp_attr = NULL;
  int __pyx_v_rc;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init_qp_attr", 0);

  /* "pyverbs/cmid.pyx":682
 *         """
 *         cdef int attr_mask
 *         qp_attr = QPAttr()             # <<<<<<<<<<<<<<
 *         qp_attr.qp_state = qp_state
 * 
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_2qp_QPAttr, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 682, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_qp_attr = ((struct __pyx_obj_7pyverbs_2qp_QPAttr *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":683
 *         cdef int attr_mask
 *         qp_attr = QPAttr()
 *         qp_attr.qp_state = qp_state             # <<<<<<<<<<<<<<
 * 
 *         rc = cm.rdma_init_qp_attr(self.id, &qp_attr.attr, &attr_mask)
*/
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_qp_attr), __pyx_mstate_global->__pyx_n_u_qp_state, __pyx_v_qp_state) < (0)) __PYX_ERR(0, 683, __pyx_L1_error)

  /* "pyverbs/cmid.pyx":685
 *         qp_attr.qp_state = qp_state
 * 
 *         rc = cm.rdma_init_qp_attr(self.id, &qp_attr.attr, &attr_mask)             # <<<<<<<<<<<<<<
 *         if rc != 0:
 *             raise PyverbsRDMAErrno('Failed to get QP attributes')
*/
  __pyx_v_rc = rdma_init_qp_attr(__pyx_v_self->id, (&__pyx_v_qp_attr->attr), (&__pyx_v_attr_mask));

  /* "pyverbs/cmid.pyx":686
 * 
 *         rc = cm.rdma_init_qp_attr(self.id, &qp_attr.attr, &attr_mask)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to get QP attributes')
 *         return qp_attr, attr_mask
*/
  __pyx_t_4 = (__pyx_v_rc != 0);
  if (unlikely(__pyx_t_4)) {

    /* "pyverbs/cmid.pyx":687
 *         rc = cm.rdma_init_qp_attr(self.id, &qp_attr.attr, &attr_mask)
 *         if rc != 0:
 *             raise PyverbsRDMAErrno('Failed to get QP attributes')             # <<<<<<<<<<<<<<
 *         return qp_attr, attr_mask
 * 
*/
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 687, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_3 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Failed_to_get_QP_attributes};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 687, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 687, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":686
 * 
 *         rc = cm.rdma_init_qp_attr(self.id, &qp_attr.attr, &attr_mask)
 *         if rc != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to get QP attributes')
 *         return qp_attr, attr_mask
*/
  }

  /* "pyverbs/cmid.pyx":688
 *         if rc != 0:
 *             raise PyverbsRDMAErrno('Failed to get QP attributes')
 *         return qp_attr, attr_mask             # <<<<<<<<<<<<<<
 * 
 *     def reg_msgs(self, size):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyLong_From_int(__pyx_v_attr_mask); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF((PyObject *)__pyx_v_qp_attr);
  __Pyx_GIVEREF((PyObject *)__pyx_v_qp_attr);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_v_qp_attr)) != (0)) __PYX_ERR(0, 688, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1) != (0)) __PYX_ERR(0, 688, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":670
 *         return attr, init_attr
 * 
 *     def init_qp_attr(self, qp_state):             # <<<<<<<<<<<<<<
 *         """
 *         Initialize a QPAttr object used for state transitions of an external
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.init_qp_attr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_qp_attr);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":690
 *         return qp_attr, attr_mask
 * 
 *     def reg_msgs(self, size):             # <<<<<<<<<<<<<<
 *         """
 *         Registers a memory region for sending or receiving messages or for
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_39reg_msgs(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_38reg_msgs, "\n        Registers a memory region for sending or receiving messages or for\n        RDMA operations. The registered memory may then be posted to an CMID\n        using post_send or post_recv methods.\n        :param size: The total length of the memory to register\n        :return: registered MR\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_39reg_msgs = {"reg_msgs", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_39reg_msgs, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_38reg_msgs};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_39reg_msgs(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_size = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reg_msgs (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_size,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 690, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 690, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "reg_msgs", 0) < (0)) __PYX_ERR(0, 690, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("reg_msgs", 1, 1, 1, i); __PYX_ERR(0, 690, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 690, __pyx_L3_error)
    }
    __pyx_v_size = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("reg_msgs", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 690, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.reg_msgs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_38reg_msgs(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_size);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_38reg_msgs(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("reg_msgs", 0);

  /* "pyverbs/cmid.pyx":698
 *         :return: registered MR
 *         """
 *         return MR(self, size, e.IBV_ACCESS_LOCAL_WRITE)             # <<<<<<<<<<<<<<
 * 
 *     def reg_read(self, size=0):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = NULL;
  __pyx_t_3 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py(IBV_ACCESS_LOCAL_WRITE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_2, ((PyObject *)__pyx_v_self), __pyx_v_size, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_2mr_MR, __pyx_callargs+__pyx_t_4, (4-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 698, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_r = ((PyObject *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":690
 *         return qp_attr, attr_mask
 * 
 *     def reg_msgs(self, size):             # <<<<<<<<<<<<<<
 *         """
 *         Registers a memory region for sending or receiving messages or for
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.reg_msgs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":700
 *         return MR(self, size, e.IBV_ACCESS_LOCAL_WRITE)
 * 
 *     def reg_read(self, size=0):             # <<<<<<<<<<<<<<
 *         """
 *         Registers a memory region for sending or receiving messages or for
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_41reg_read(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_40reg_read, "\n        Registers a memory region for sending or receiving messages or for\n        remote read operations.\n        :param size: The total length of the memory to register\n        :return: registered MR\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_41reg_read = {"reg_read", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_41reg_read, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_40reg_read};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_41reg_read(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_size = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reg_read (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_size,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 700, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 700, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "reg_read", 0) < (0)) __PYX_ERR(0, 700, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    } else {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 700, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    }
    __pyx_v_size = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("reg_read", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 700, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.reg_read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_40reg_read(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_size);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_40reg_read(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("reg_read", 0);

  /* "pyverbs/cmid.pyx":707
 *         :return: registered MR
 *         """
 *         return MR(self, size, e.IBV_ACCESS_REMOTE_READ)             # <<<<<<<<<<<<<<
 * 
 *     def reg_write(self, size=0):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = NULL;
  __pyx_t_3 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py(IBV_ACCESS_REMOTE_READ); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_2, ((PyObject *)__pyx_v_self), __pyx_v_size, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_2mr_MR, __pyx_callargs+__pyx_t_4, (4-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_r = ((PyObject *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":700
 *         return MR(self, size, e.IBV_ACCESS_LOCAL_WRITE)
 * 
 *     def reg_read(self, size=0):             # <<<<<<<<<<<<<<
 *         """
 *         Registers a memory region for sending or receiving messages or for
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.reg_read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":709
 *         return MR(self, size, e.IBV_ACCESS_REMOTE_READ)
 * 
 *     def reg_write(self, size=0):             # <<<<<<<<<<<<<<
 *         """
 *         Registers a memory region for sending or receiving messages or for
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_43reg_write(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_42reg_write, "\n        Registers a memory region for sending or receiving messages or for\n        remote write operations.\n        :param size: The total length of the memory to register\n        :return: registered MR\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_43reg_write = {"reg_write", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_43reg_write, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_42reg_write};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_43reg_write(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_size = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reg_write (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_size,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 709, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 709, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "reg_write", 0) < (0)) __PYX_ERR(0, 709, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    } else {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 709, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    }
    __pyx_v_size = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("reg_write", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 709, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.reg_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_42reg_write(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_size);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_42reg_write(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("reg_write", 0);

  /* "pyverbs/cmid.pyx":716
 *         :return: registered MR
 *         """
 *         return MR(self, size, e.IBV_ACCESS_REMOTE_WRITE)             # <<<<<<<<<<<<<<
 * 
 *     def post_recv(self, MR mr not None, length=None):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = NULL;
  __pyx_t_3 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py(IBV_ACCESS_REMOTE_WRITE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_2, ((PyObject *)__pyx_v_self), __pyx_v_size, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_2mr_MR, __pyx_callargs+__pyx_t_4, (4-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 716, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_r = ((PyObject *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":709
 *         return MR(self, size, e.IBV_ACCESS_REMOTE_READ)
 * 
 *     def reg_write(self, size=0):             # <<<<<<<<<<<<<<
 *         """
 *         Registers a memory region for sending or receiving messages or for
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.reg_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":718
 *         return MR(self, size, e.IBV_ACCESS_REMOTE_WRITE)
 * 
 *     def post_recv(self, MR mr not None, length=None):             # <<<<<<<<<<<<<<
 *         """
 *         Posts a recv_wr via QP associated with CMID.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_45post_recv(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_44post_recv, "\n        Posts a recv_wr via QP associated with CMID.\n        Context param of rdma_post_recv C function currently not supported.\n        :param mr: A valid MR object.\n        :param length: length of buffer to recv (default: mr length).\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_45post_recv = {"post_recv", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_45post_recv, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_44post_recv};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_45post_recv(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_2mr_MR *__pyx_v_mr = 0;
  PyObject *__pyx_v_length = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("post_recv (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_mr,&__pyx_mstate_global->__pyx_n_u_length,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 718, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 718, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 718, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "post_recv", 0) < (0)) __PYX_ERR(0, 718, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("post_recv", 0, 1, 2, i); __PYX_ERR(0, 718, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 718, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 718, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_mr = ((struct __pyx_obj_7pyverbs_2mr_MR *)values[0]);
    __pyx_v_length = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("post_recv", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 718, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.post_recv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mr), __pyx_mstate_global->__pyx_ptype_7pyverbs_2mr_MR, 0, "mr", 0))) __PYX_ERR(0, 718, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_44post_recv(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_mr, __pyx_v_length);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_44post_recv(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_2mr_MR *__pyx_v_mr, PyObject *__pyx_v_length) {
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("post_recv", 0);
  __Pyx_INCREF(__pyx_v_length);

  /* "pyverbs/cmid.pyx":726
 *         :return: None
 *         """
 *         if not length:             # <<<<<<<<<<<<<<
 *             length = mr.mr.length
 *         ret = cm.rdma_post_recv(self.id, NULL, mr.buf, length, mr.mr)
*/
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_length); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 726, __pyx_L1_error)
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "pyverbs/cmid.pyx":727
 *         """
 *         if not length:
 *             length = mr.mr.length             # <<<<<<<<<<<<<<
 *         ret = cm.rdma_post_recv(self.id, NULL, mr.buf, length, mr.mr)
 *         if ret != 0:
*/
    __pyx_t_3 = __Pyx_PyLong_FromSize_t(__pyx_v_mr->mr->length); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 727, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_length, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyverbs/cmid.pyx":726
 *         :return: None
 *         """
 *         if not length:             # <<<<<<<<<<<<<<
 *             length = mr.mr.length
 *         ret = cm.rdma_post_recv(self.id, NULL, mr.buf, length, mr.mr)
*/
  }

  /* "pyverbs/cmid.pyx":728
 *         if not length:
 *             length = mr.mr.length
 *         ret = cm.rdma_post_recv(self.id, NULL, mr.buf, length, mr.mr)             # <<<<<<<<<<<<<<
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Post Receive')
*/
  __pyx_t_4 = __Pyx_PyLong_As_size_t(__pyx_v_length); if (unlikely((__pyx_t_4 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 728, __pyx_L1_error)
  __pyx_v_ret = rdma_post_recv(__pyx_v_self->id, NULL, __pyx_v_mr->buf, __pyx_t_4, __pyx_v_mr->mr);

  /* "pyverbs/cmid.pyx":729
 *             length = mr.mr.length
 *         ret = cm.rdma_post_recv(self.id, NULL, mr.buf, length, mr.mr)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Post Receive')
 * 
*/
  __pyx_t_2 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_2)) {

    /* "pyverbs/cmid.pyx":730
 *         ret = cm.rdma_post_recv(self.id, NULL, mr.buf, length, mr.mr)
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Post Receive')             # <<<<<<<<<<<<<<
 * 
 *     def post_send(self, MR mr not None, flags=v.IBV_SEND_SIGNALED, length=None):
*/
    __pyx_t_5 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      assert(__pyx_t_5);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
      __pyx_t_4 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_mstate_global->__pyx_kp_u_Failed_to_Post_Receive};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_6, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 730, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 730, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":729
 *             length = mr.mr.length
 *         ret = cm.rdma_post_recv(self.id, NULL, mr.buf, length, mr.mr)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Post Receive')
 * 
*/
  }

  /* "pyverbs/cmid.pyx":718
 *         return MR(self, size, e.IBV_ACCESS_REMOTE_WRITE)
 * 
 *     def post_recv(self, MR mr not None, length=None):             # <<<<<<<<<<<<<<
 *         """
 *         Posts a recv_wr via QP associated with CMID.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.post_recv", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":732
 *             raise PyverbsRDMAErrno('Failed to Post Receive')
 * 
 *     def post_send(self, MR mr not None, flags=v.IBV_SEND_SIGNALED, length=None):             # <<<<<<<<<<<<<<
 *         """
 *         Posts a message via QP associated with CMID.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_47post_send(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_46post_send, "\n        Posts a message via QP associated with CMID.\n        Context param of rdma_post_send C function currently not supported.\n        :param mr: A valid MR object which contains message to send.\n        :param flags: flags for send work request.\n        :param length: length of buffer to send (default: mr length).\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_47post_send = {"post_send", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_47post_send, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_46post_send};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_47post_send(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_2mr_MR *__pyx_v_mr = 0;
  PyObject *__pyx_v_flags = 0;
  PyObject *__pyx_v_length = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("post_send (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_mr,&__pyx_mstate_global->__pyx_n_u_flags,&__pyx_mstate_global->__pyx_n_u_length,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 732, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 732, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 732, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 732, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "post_send", 0) < (0)) __PYX_ERR(0, 732, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__5);
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("post_send", 0, 1, 3, i); __PYX_ERR(0, 732, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 732, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 732, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 732, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[1]) values[1] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__5);
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_mr = ((struct __pyx_obj_7pyverbs_2mr_MR *)values[0]);
    __pyx_v_flags = values[1];
    __pyx_v_length = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("post_send", 0, 1, 3, __pyx_nargs); __PYX_ERR(0, 732, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.post_send", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mr), __pyx_mstate_global->__pyx_ptype_7pyverbs_2mr_MR, 0, "mr", 0))) __PYX_ERR(0, 732, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_46post_send(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_mr, __pyx_v_flags, __pyx_v_length);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_46post_send(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_2mr_MR *__pyx_v_mr, PyObject *__pyx_v_flags, PyObject *__pyx_v_length) {
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("post_send", 0);
  __Pyx_INCREF(__pyx_v_length);

  /* "pyverbs/cmid.pyx":741
 *         :return: None
 *         """
 *         if not length:             # <<<<<<<<<<<<<<
 *             length = mr.mr.length
 *         ret = cm.rdma_post_send(self.id, NULL, mr.buf, length, mr.mr,
*/
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_length); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 741, __pyx_L1_error)
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "pyverbs/cmid.pyx":742
 *         """
 *         if not length:
 *             length = mr.mr.length             # <<<<<<<<<<<<<<
 *         ret = cm.rdma_post_send(self.id, NULL, mr.buf, length, mr.mr,
 *                                 flags)
*/
    __pyx_t_3 = __Pyx_PyLong_FromSize_t(__pyx_v_mr->mr->length); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 742, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_length, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyverbs/cmid.pyx":741
 *         :return: None
 *         """
 *         if not length:             # <<<<<<<<<<<<<<
 *             length = mr.mr.length
 *         ret = cm.rdma_post_send(self.id, NULL, mr.buf, length, mr.mr,
*/
  }

  /* "pyverbs/cmid.pyx":743
 *         if not length:
 *             length = mr.mr.length
 *         ret = cm.rdma_post_send(self.id, NULL, mr.buf, length, mr.mr,             # <<<<<<<<<<<<<<
 *                                 flags)
 *         if ret != 0:
*/
  __pyx_t_4 = __Pyx_PyLong_As_size_t(__pyx_v_length); if (unlikely((__pyx_t_4 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 743, __pyx_L1_error)

  /* "pyverbs/cmid.pyx":744
 *             length = mr.mr.length
 *         ret = cm.rdma_post_send(self.id, NULL, mr.buf, length, mr.mr,
 *                                 flags)             # <<<<<<<<<<<<<<
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Post Send')
*/
  __pyx_t_5 = __Pyx_PyLong_As_int(__pyx_v_flags); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 744, __pyx_L1_error)

  /* "pyverbs/cmid.pyx":743
 *         if not length:
 *             length = mr.mr.length
 *         ret = cm.rdma_post_send(self.id, NULL, mr.buf, length, mr.mr,             # <<<<<<<<<<<<<<
 *                                 flags)
 *         if ret != 0:
*/
  __pyx_v_ret = rdma_post_send(__pyx_v_self->id, NULL, __pyx_v_mr->buf, __pyx_t_4, __pyx_v_mr->mr, __pyx_t_5);

  /* "pyverbs/cmid.pyx":745
 *         ret = cm.rdma_post_send(self.id, NULL, mr.buf, length, mr.mr,
 *                                 flags)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Post Send')
 * 
*/
  __pyx_t_2 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_2)) {

    /* "pyverbs/cmid.pyx":746
 *                                 flags)
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Post Send')             # <<<<<<<<<<<<<<
 * 
 *     def post_read(self, MR mr not None, length, remote_addr, rkey,
*/
    __pyx_t_6 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 746, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      assert(__pyx_t_6);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_7, __pyx__function);
      __pyx_t_4 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_mstate_global->__pyx_kp_u_Failed_to_Post_Send};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 746, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 746, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":745
 *         ret = cm.rdma_post_send(self.id, NULL, mr.buf, length, mr.mr,
 *                                 flags)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Post Send')
 * 
*/
  }

  /* "pyverbs/cmid.pyx":732
 *             raise PyverbsRDMAErrno('Failed to Post Receive')
 * 
 *     def post_send(self, MR mr not None, flags=v.IBV_SEND_SIGNALED, length=None):             # <<<<<<<<<<<<<<
 *         """
 *         Posts a message via QP associated with CMID.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.post_send", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":748
 *             raise PyverbsRDMAErrno('Failed to Post Send')
 * 
 *     def post_read(self, MR mr not None, length, remote_addr, rkey,             # <<<<<<<<<<<<<<
 *                   flags=0):
 *         """
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_49post_read(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_48post_read, "\n        Post read WR using the CMIDs internal QP.\n        :param mr: A valid MR object.\n        :param length: length of buffer to send.\n        :param remote_addr: The remote MR address.\n        :param rkey: The remote MR rkey.\n        :param flags: flags for send work request.\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_49post_read = {"post_read", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_49post_read, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_48post_read};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_49post_read(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_2mr_MR *__pyx_v_mr = 0;
  PyObject *__pyx_v_length = 0;
  PyObject *__pyx_v_remote_addr = 0;
  PyObject *__pyx_v_rkey = 0;
  PyObject *__pyx_v_flags = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("post_read (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_mr,&__pyx_mstate_global->__pyx_n_u_length,&__pyx_mstate_global->__pyx_n_u_remote_addr,&__pyx_mstate_global->__pyx_n_u_rkey,&__pyx_mstate_global->__pyx_n_u_flags,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 748, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 748, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 748, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 748, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 748, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 748, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "post_read", 0) < (0)) __PYX_ERR(0, 748, __pyx_L3_error)
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      for (Py_ssize_t i = __pyx_nargs; i < 4; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("post_read", 0, 4, 5, i); __PYX_ERR(0, 748, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 748, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 748, __pyx_L3_error)
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 748, __pyx_L3_error)
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 748, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 748, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    }
    __pyx_v_mr = ((struct __pyx_obj_7pyverbs_2mr_MR *)values[0]);
    __pyx_v_length = values[1];
    __pyx_v_remote_addr = values[2];
    __pyx_v_rkey = values[3];
    __pyx_v_flags = values[4];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("post_read", 0, 4, 5, __pyx_nargs); __PYX_ERR(0, 748, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.post_read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mr), __pyx_mstate_global->__pyx_ptype_7pyverbs_2mr_MR, 0, "mr", 0))) __PYX_ERR(0, 748, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_48post_read(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_mr, __pyx_v_length, __pyx_v_remote_addr, __pyx_v_rkey, __pyx_v_flags);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_48post_read(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_2mr_MR *__pyx_v_mr, PyObject *__pyx_v_length, PyObject *__pyx_v_remote_addr, PyObject *__pyx_v_rkey, PyObject *__pyx_v_flags) {
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  int __pyx_t_2;
  uint64_t __pyx_t_3;
  uint32_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("post_read", 0);

  /* "pyverbs/cmid.pyx":759
 *         :return: None
 *         """
 *         ret = cm.rdma_post_read(self.id, NULL, mr.buf, length, mr.mr,             # <<<<<<<<<<<<<<
 *                                 flags, remote_addr, rkey)
 *         if ret != 0:
*/
  __pyx_t_1 = __Pyx_PyLong_As_size_t(__pyx_v_length); if (unlikely((__pyx_t_1 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 759, __pyx_L1_error)

  /* "pyverbs/cmid.pyx":760
 *         """
 *         ret = cm.rdma_post_read(self.id, NULL, mr.buf, length, mr.mr,
 *                                 flags, remote_addr, rkey)             # <<<<<<<<<<<<<<
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Post Read')
*/
  __pyx_t_2 = __Pyx_PyLong_As_int(__pyx_v_flags); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 760, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyLong_As_uint64_t(__pyx_v_remote_addr); if (unlikely((__pyx_t_3 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 760, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyLong_As_uint32_t(__pyx_v_rkey); if (unlikely((__pyx_t_4 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 760, __pyx_L1_error)

  /* "pyverbs/cmid.pyx":759
 *         :return: None
 *         """
 *         ret = cm.rdma_post_read(self.id, NULL, mr.buf, length, mr.mr,             # <<<<<<<<<<<<<<
 *                                 flags, remote_addr, rkey)
 *         if ret != 0:
*/
  __pyx_v_ret = rdma_post_read(__pyx_v_self->id, NULL, __pyx_v_mr->buf, __pyx_t_1, __pyx_v_mr->mr, __pyx_t_2, __pyx_t_3, __pyx_t_4);

  /* "pyverbs/cmid.pyx":761
 *         ret = cm.rdma_post_read(self.id, NULL, mr.buf, length, mr.mr,
 *                                 flags, remote_addr, rkey)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Post Read')
 * 
*/
  __pyx_t_5 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_5)) {

    /* "pyverbs/cmid.pyx":762
 *                                 flags, remote_addr, rkey)
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Post Read')             # <<<<<<<<<<<<<<
 * 
 *     def post_write(self, MR mr not None, length, remote_addr, rkey,
*/
    __pyx_t_7 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 762, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_1 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
      assert(__pyx_t_7);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_8, __pyx__function);
      __pyx_t_1 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_mstate_global->__pyx_kp_u_Failed_to_Post_Read};
      __pyx_t_6 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_8, __pyx_callargs+__pyx_t_1, (2-__pyx_t_1) | (__pyx_t_1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 762, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
    }
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 762, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":761
 *         ret = cm.rdma_post_read(self.id, NULL, mr.buf, length, mr.mr,
 *                                 flags, remote_addr, rkey)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Post Read')
 * 
*/
  }

  /* "pyverbs/cmid.pyx":748
 *             raise PyverbsRDMAErrno('Failed to Post Send')
 * 
 *     def post_read(self, MR mr not None, length, remote_addr, rkey,             # <<<<<<<<<<<<<<
 *                   flags=0):
 *         """
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.post_read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":764
 *             raise PyverbsRDMAErrno('Failed to Post Read')
 * 
 *     def post_write(self, MR mr not None, length, remote_addr, rkey,             # <<<<<<<<<<<<<<
 *                    flags=0):
 *         """
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_51post_write(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_50post_write, "\n        Post write WR using the CMIDs internal QP.\n        :param mr: A valid MR object.\n        :param length: length of buffer to send.\n        :param remote_addr: The remote MR address.\n        :param rkey: The remote MR rkey.\n        :param flags: flags for send work request.\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_51post_write = {"post_write", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_51post_write, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_50post_write};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_51post_write(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_2mr_MR *__pyx_v_mr = 0;
  PyObject *__pyx_v_length = 0;
  PyObject *__pyx_v_remote_addr = 0;
  PyObject *__pyx_v_rkey = 0;
  PyObject *__pyx_v_flags = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("post_write (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_mr,&__pyx_mstate_global->__pyx_n_u_length,&__pyx_mstate_global->__pyx_n_u_remote_addr,&__pyx_mstate_global->__pyx_n_u_rkey,&__pyx_mstate_global->__pyx_n_u_flags,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 764, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 764, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 764, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 764, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 764, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 764, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "post_write", 0) < (0)) __PYX_ERR(0, 764, __pyx_L3_error)
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      for (Py_ssize_t i = __pyx_nargs; i < 4; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("post_write", 0, 4, 5, i); __PYX_ERR(0, 764, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 764, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 764, __pyx_L3_error)
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 764, __pyx_L3_error)
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 764, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 764, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    }
    __pyx_v_mr = ((struct __pyx_obj_7pyverbs_2mr_MR *)values[0]);
    __pyx_v_length = values[1];
    __pyx_v_remote_addr = values[2];
    __pyx_v_rkey = values[3];
    __pyx_v_flags = values[4];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("post_write", 0, 4, 5, __pyx_nargs); __PYX_ERR(0, 764, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.post_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mr), __pyx_mstate_global->__pyx_ptype_7pyverbs_2mr_MR, 0, "mr", 0))) __PYX_ERR(0, 764, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_50post_write(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_mr, __pyx_v_length, __pyx_v_remote_addr, __pyx_v_rkey, __pyx_v_flags);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_50post_write(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_2mr_MR *__pyx_v_mr, PyObject *__pyx_v_length, PyObject *__pyx_v_remote_addr, PyObject *__pyx_v_rkey, PyObject *__pyx_v_flags) {
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  int __pyx_t_2;
  uint64_t __pyx_t_3;
  uint32_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("post_write", 0);

  /* "pyverbs/cmid.pyx":775
 *         :return: None
 *         """
 *         ret = cm.rdma_post_write(self.id, NULL, mr.buf, length, mr.mr,             # <<<<<<<<<<<<<<
 *                                  flags, remote_addr, rkey)
 *         if ret != 0:
*/
  __pyx_t_1 = __Pyx_PyLong_As_size_t(__pyx_v_length); if (unlikely((__pyx_t_1 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 775, __pyx_L1_error)

  /* "pyverbs/cmid.pyx":776
 *         """
 *         ret = cm.rdma_post_write(self.id, NULL, mr.buf, length, mr.mr,
 *                                  flags, remote_addr, rkey)             # <<<<<<<<<<<<<<
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Post Write')
*/
  __pyx_t_2 = __Pyx_PyLong_As_int(__pyx_v_flags); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 776, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyLong_As_uint64_t(__pyx_v_remote_addr); if (unlikely((__pyx_t_3 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 776, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyLong_As_uint32_t(__pyx_v_rkey); if (unlikely((__pyx_t_4 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 776, __pyx_L1_error)

  /* "pyverbs/cmid.pyx":775
 *         :return: None
 *         """
 *         ret = cm.rdma_post_write(self.id, NULL, mr.buf, length, mr.mr,             # <<<<<<<<<<<<<<
 *                                  flags, remote_addr, rkey)
 *         if ret != 0:
*/
  __pyx_v_ret = rdma_post_write(__pyx_v_self->id, NULL, __pyx_v_mr->buf, __pyx_t_1, __pyx_v_mr->mr, __pyx_t_2, __pyx_t_3, __pyx_t_4);

  /* "pyverbs/cmid.pyx":777
 *         ret = cm.rdma_post_write(self.id, NULL, mr.buf, length, mr.mr,
 *                                  flags, remote_addr, rkey)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Post Write')
 * 
*/
  __pyx_t_5 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_5)) {

    /* "pyverbs/cmid.pyx":778
 *                                  flags, remote_addr, rkey)
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Post Write')             # <<<<<<<<<<<<<<
 * 
 *     def post_ud_send(self, MR mr not None, AH ah not None, rqpn=0,
*/
    __pyx_t_7 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 778, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_1 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
      assert(__pyx_t_7);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_8, __pyx__function);
      __pyx_t_1 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_mstate_global->__pyx_kp_u_Failed_to_Post_Write};
      __pyx_t_6 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_8, __pyx_callargs+__pyx_t_1, (2-__pyx_t_1) | (__pyx_t_1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 778, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
    }
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 778, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":777
 *         ret = cm.rdma_post_write(self.id, NULL, mr.buf, length, mr.mr,
 *                                  flags, remote_addr, rkey)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Post Write')
 * 
*/
  }

  /* "pyverbs/cmid.pyx":764
 *             raise PyverbsRDMAErrno('Failed to Post Read')
 * 
 *     def post_write(self, MR mr not None, length, remote_addr, rkey,             # <<<<<<<<<<<<<<
 *                    flags=0):
 *         """
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.post_write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":780
 *             raise PyverbsRDMAErrno('Failed to Post Write')
 * 
 *     def post_ud_send(self, MR mr not None, AH ah not None, rqpn=0,             # <<<<<<<<<<<<<<
 *                      flags=v.IBV_SEND_SIGNALED, length=None):
 *         """
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_53post_ud_send(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_52post_ud_send, "\n        Posts a message via UD QP associated with CMID to another UD QP.\n        :param mr: A valid MR object which contains message to send.\n        :param ah: The destination AH.\n        :param rqpn: The remote QP number.\n        :param flags: flags for send work request.\n        :param length: length of buffer to send.\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_53post_ud_send = {"post_ud_send", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_53post_ud_send, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_52post_ud_send};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_53post_ud_send(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyverbs_2mr_MR *__pyx_v_mr = 0;
  struct __pyx_obj_7pyverbs_4addr_AH *__pyx_v_ah = 0;
  PyObject *__pyx_v_rqpn = 0;
  PyObject *__pyx_v_flags = 0;
  PyObject *__pyx_v_length = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("post_ud_send (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_mr,&__pyx_mstate_global->__pyx_n_u_ah,&__pyx_mstate_global->__pyx_n_u_rqpn,&__pyx_mstate_global->__pyx_n_u_flags,&__pyx_mstate_global->__pyx_n_u_length,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 780, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 780, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 780, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 780, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 780, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 780, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "post_ud_send", 0) < (0)) __PYX_ERR(0, 780, __pyx_L3_error)
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[3]) values[3] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__6);

      /* "pyverbs/cmid.pyx":781
 * 
 *     def post_ud_send(self, MR mr not None, AH ah not None, rqpn=0,
 *                      flags=v.IBV_SEND_SIGNALED, length=None):             # <<<<<<<<<<<<<<
 *         """
 *         Posts a message via UD QP associated with CMID to another UD QP.
*/
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)Py_None));
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("post_ud_send", 0, 2, 5, i); __PYX_ERR(0, 780, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 780, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 780, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 780, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 780, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 780, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[3]) values[3] = __Pyx_NewRef(__pyx_mstate_global->__pyx_k__6);
      if (!values[4]) values[4] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_mr = ((struct __pyx_obj_7pyverbs_2mr_MR *)values[0]);
    __pyx_v_ah = ((struct __pyx_obj_7pyverbs_4addr_AH *)values[1]);
    __pyx_v_rqpn = values[2];
    __pyx_v_flags = values[3];
    __pyx_v_length = values[4];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("post_ud_send", 0, 2, 5, __pyx_nargs); __PYX_ERR(0, 780, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.post_ud_send", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mr), __pyx_mstate_global->__pyx_ptype_7pyverbs_2mr_MR, 0, "mr", 0))) __PYX_ERR(0, 780, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_ah), __pyx_mstate_global->__pyx_ptype_7pyverbs_4addr_AH, 0, "ah", 0))) __PYX_ERR(0, 780, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_52post_ud_send(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_mr, __pyx_v_ah, __pyx_v_rqpn, __pyx_v_flags, __pyx_v_length);

  /* "pyverbs/cmid.pyx":780
 *             raise PyverbsRDMAErrno('Failed to Post Write')
 * 
 *     def post_ud_send(self, MR mr not None, AH ah not None, rqpn=0,             # <<<<<<<<<<<<<<
 *                      flags=v.IBV_SEND_SIGNALED, length=None):
 *         """
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_52post_ud_send(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, struct __pyx_obj_7pyverbs_2mr_MR *__pyx_v_mr, struct __pyx_obj_7pyverbs_4addr_AH *__pyx_v_ah, PyObject *__pyx_v_rqpn, PyObject *__pyx_v_flags, PyObject *__pyx_v_length) {
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  uint32_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("post_ud_send", 0);
  __Pyx_INCREF(__pyx_v_length);

  /* "pyverbs/cmid.pyx":791
 *         :return: None
 *         """
 *         if not length:             # <<<<<<<<<<<<<<
 *             length = mr.mr.length
 *         ret = cm.rdma_post_ud_send(self.id, NULL, mr.buf, length, mr.mr,
*/
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_length); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 791, __pyx_L1_error)
  __pyx_t_2 = (!__pyx_t_1);
  if (__pyx_t_2) {

    /* "pyverbs/cmid.pyx":792
 *         """
 *         if not length:
 *             length = mr.mr.length             # <<<<<<<<<<<<<<
 *         ret = cm.rdma_post_ud_send(self.id, NULL, mr.buf, length, mr.mr,
 *                                    flags, ah.ah, rqpn)
*/
    __pyx_t_3 = __Pyx_PyLong_FromSize_t(__pyx_v_mr->mr->length); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 792, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_length, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyverbs/cmid.pyx":791
 *         :return: None
 *         """
 *         if not length:             # <<<<<<<<<<<<<<
 *             length = mr.mr.length
 *         ret = cm.rdma_post_ud_send(self.id, NULL, mr.buf, length, mr.mr,
*/
  }

  /* "pyverbs/cmid.pyx":793
 *         if not length:
 *             length = mr.mr.length
 *         ret = cm.rdma_post_ud_send(self.id, NULL, mr.buf, length, mr.mr,             # <<<<<<<<<<<<<<
 *                                    flags, ah.ah, rqpn)
 *         if ret != 0:
*/
  __pyx_t_4 = __Pyx_PyLong_As_size_t(__pyx_v_length); if (unlikely((__pyx_t_4 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 793, __pyx_L1_error)

  /* "pyverbs/cmid.pyx":794
 *             length = mr.mr.length
 *         ret = cm.rdma_post_ud_send(self.id, NULL, mr.buf, length, mr.mr,
 *                                    flags, ah.ah, rqpn)             # <<<<<<<<<<<<<<
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Post Send')
*/
  __pyx_t_5 = __Pyx_PyLong_As_int(__pyx_v_flags); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 794, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyLong_As_uint32_t(__pyx_v_rqpn); if (unlikely((__pyx_t_6 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 794, __pyx_L1_error)

  /* "pyverbs/cmid.pyx":793
 *         if not length:
 *             length = mr.mr.length
 *         ret = cm.rdma_post_ud_send(self.id, NULL, mr.buf, length, mr.mr,             # <<<<<<<<<<<<<<
 *                                    flags, ah.ah, rqpn)
 *         if ret != 0:
*/
  __pyx_v_ret = rdma_post_ud_send(__pyx_v_self->id, NULL, __pyx_v_mr->buf, __pyx_t_4, __pyx_v_mr->mr, __pyx_t_5, __pyx_v_ah->ah, __pyx_t_6);

  /* "pyverbs/cmid.pyx":795
 *         ret = cm.rdma_post_ud_send(self.id, NULL, mr.buf, length, mr.mr,
 *                                    flags, ah.ah, rqpn)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Post Send')
 * 
*/
  __pyx_t_2 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_2)) {

    /* "pyverbs/cmid.pyx":796
 *                                    flags, ah.ah, rqpn)
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Post Send')             # <<<<<<<<<<<<<<
 * 
 *     def get_recv_comp(self):
*/
    __pyx_t_7 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_4 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
      assert(__pyx_t_7);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_8, __pyx__function);
      __pyx_t_4 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_mstate_global->__pyx_kp_u_Failed_to_Post_Send};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_8, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 796, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 796, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":795
 *         ret = cm.rdma_post_ud_send(self.id, NULL, mr.buf, length, mr.mr,
 *                                    flags, ah.ah, rqpn)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Post Send')
 * 
*/
  }

  /* "pyverbs/cmid.pyx":780
 *             raise PyverbsRDMAErrno('Failed to Post Write')
 * 
 *     def post_ud_send(self, MR mr not None, AH ah not None, rqpn=0,             # <<<<<<<<<<<<<<
 *                      flags=v.IBV_SEND_SIGNALED, length=None):
 *         """
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.post_ud_send", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":798
 *             raise PyverbsRDMAErrno('Failed to Post Send')
 * 
 *     def get_recv_comp(self):             # <<<<<<<<<<<<<<
 *         """
 *         Polls the receive CQ associated with CMID for a work completion.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_55get_recv_comp(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_54get_recv_comp, "\n        Polls the receive CQ associated with CMID for a work completion.\n        :return: The retrieved WC or None if there is no completions\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_55get_recv_comp = {"get_recv_comp", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_55get_recv_comp, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_54get_recv_comp};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_55get_recv_comp(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_recv_comp (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("get_recv_comp", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("get_recv_comp", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_54get_recv_comp(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_54get_recv_comp(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self) {
  struct ibv_wc __pyx_v_wc;
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_recv_comp", 0);

  /* "pyverbs/cmid.pyx":804
 *         """
 *         cdef v.ibv_wc wc
 *         with nogil:             # <<<<<<<<<<<<<<
 *             ret = cm.rdma_get_recv_comp(self.id, &wc)
 *         if ret < 0:
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "pyverbs/cmid.pyx":805
 *         cdef v.ibv_wc wc
 *         with nogil:
 *             ret = cm.rdma_get_recv_comp(self.id, &wc)             # <<<<<<<<<<<<<<
 *         if ret < 0:
 *             raise PyverbsRDMAErrno('Failed to retrieve receive completion')
*/
        __pyx_v_ret = rdma_get_recv_comp(__pyx_v_self->id, (&__pyx_v_wc));
      }

      /* "pyverbs/cmid.pyx":804
 *         """
 *         cdef v.ibv_wc wc
 *         with nogil:             # <<<<<<<<<<<<<<
 *             ret = cm.rdma_get_recv_comp(self.id, &wc)
 *         if ret < 0:
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "pyverbs/cmid.pyx":806
 *         with nogil:
 *             ret = cm.rdma_get_recv_comp(self.id, &wc)
 *         if ret < 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to retrieve receive completion')
 *         elif ret == 0:
*/
  __pyx_t_1 = (__pyx_v_ret < 0);
  if (unlikely(__pyx_t_1)) {

    /* "pyverbs/cmid.pyx":807
 *             ret = cm.rdma_get_recv_comp(self.id, &wc)
 *         if ret < 0:
 *             raise PyverbsRDMAErrno('Failed to retrieve receive completion')             # <<<<<<<<<<<<<<
 *         elif ret == 0:
 *             return None
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 807, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Failed_to_retrieve_receive_compl};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 807, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 807, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":806
 *         with nogil:
 *             ret = cm.rdma_get_recv_comp(self.id, &wc)
 *         if ret < 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to retrieve receive completion')
 *         elif ret == 0:
*/
  }

  /* "pyverbs/cmid.pyx":808
 *         if ret < 0:
 *             raise PyverbsRDMAErrno('Failed to retrieve receive completion')
 *         elif ret == 0:             # <<<<<<<<<<<<<<
 *             return None
 *         return WC(wr_id=wc.wr_id, status=wc.status, opcode=wc.opcode,
*/
  __pyx_t_1 = (__pyx_v_ret == 0);
  if (__pyx_t_1) {

    /* "pyverbs/cmid.pyx":809
 *             raise PyverbsRDMAErrno('Failed to retrieve receive completion')
 *         elif ret == 0:
 *             return None             # <<<<<<<<<<<<<<
 *         return WC(wr_id=wc.wr_id, status=wc.status, opcode=wc.opcode,
 *                   vendor_err=wc.vendor_err, byte_len=wc.byte_len,
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "pyverbs/cmid.pyx":808
 *         if ret < 0:
 *             raise PyverbsRDMAErrno('Failed to retrieve receive completion')
 *         elif ret == 0:             # <<<<<<<<<<<<<<
 *             return None
 *         return WC(wr_id=wc.wr_id, status=wc.status, opcode=wc.opcode,
*/
  }

  /* "pyverbs/cmid.pyx":810
 *         elif ret == 0:
 *             return None
 *         return WC(wr_id=wc.wr_id, status=wc.status, opcode=wc.opcode,             # <<<<<<<<<<<<<<
 *                   vendor_err=wc.vendor_err, byte_len=wc.byte_len,
 *                   qp_num=wc.qp_num, src_qp=wc.src_qp,
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = NULL;
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_long(__pyx_v_wc.wr_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(__pyx_v_wc.status); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py(__pyx_v_wc.opcode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "pyverbs/cmid.pyx":811
 *             return None
 *         return WC(wr_id=wc.wr_id, status=wc.status, opcode=wc.opcode,
 *                   vendor_err=wc.vendor_err, byte_len=wc.byte_len,             # <<<<<<<<<<<<<<
 *                   qp_num=wc.qp_num, src_qp=wc.src_qp,
 *                   imm_data=wc.imm_data, wc_flags=wc.wc_flags,
*/
  __pyx_t_8 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.vendor_err); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.byte_len); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pyverbs/cmid.pyx":812
 *         return WC(wr_id=wc.wr_id, status=wc.status, opcode=wc.opcode,
 *                   vendor_err=wc.vendor_err, byte_len=wc.byte_len,
 *                   qp_num=wc.qp_num, src_qp=wc.src_qp,             # <<<<<<<<<<<<<<
 *                   imm_data=wc.imm_data, wc_flags=wc.wc_flags,
 *                   pkey_index=wc.pkey_index, slid=wc.slid, sl=wc.sl,
*/
  __pyx_t_10 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.qp_num); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.src_qp); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);

  /* "pyverbs/cmid.pyx":813
 *                   vendor_err=wc.vendor_err, byte_len=wc.byte_len,
 *                   qp_num=wc.qp_num, src_qp=wc.src_qp,
 *                   imm_data=wc.imm_data, wc_flags=wc.wc_flags,             # <<<<<<<<<<<<<<
 *                   pkey_index=wc.pkey_index, slid=wc.slid, sl=wc.sl,
 *                   dlid_path_bits=wc.dlid_path_bits)
*/
  __pyx_t_12 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.imm_data); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __Pyx_PyLong_From_int(__pyx_v_wc.wc_flags); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);

  /* "pyverbs/cmid.pyx":814
 *                   qp_num=wc.qp_num, src_qp=wc.src_qp,
 *                   imm_data=wc.imm_data, wc_flags=wc.wc_flags,
 *                   pkey_index=wc.pkey_index, slid=wc.slid, sl=wc.sl,             # <<<<<<<<<<<<<<
 *                   dlid_path_bits=wc.dlid_path_bits)
 * 
*/
  __pyx_t_14 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.pkey_index); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.slid); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_16 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.sl); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);

  /* "pyverbs/cmid.pyx":815
 *                   imm_data=wc.imm_data, wc_flags=wc.wc_flags,
 *                   pkey_index=wc.pkey_index, slid=wc.slid, sl=wc.sl,
 *                   dlid_path_bits=wc.dlid_path_bits)             # <<<<<<<<<<<<<<
 * 
 *     def get_send_comp(self):
*/
  __pyx_t_17 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.dlid_path_bits); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 13 : 0)] = {__pyx_t_4, NULL};
    __pyx_t_18 = __Pyx_MakeVectorcallBuilderKwds(13); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_wr_id, __pyx_t_3, __pyx_t_18, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 810, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_status, __pyx_t_6, __pyx_t_18, __pyx_callargs+1, 1) < (0)) __PYX_ERR(0, 810, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_opcode, __pyx_t_7, __pyx_t_18, __pyx_callargs+1, 2) < (0)) __PYX_ERR(0, 810, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_vendor_err, __pyx_t_8, __pyx_t_18, __pyx_callargs+1, 3) < (0)) __PYX_ERR(0, 810, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_byte_len, __pyx_t_9, __pyx_t_18, __pyx_callargs+1, 4) < (0)) __PYX_ERR(0, 810, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_qp_num, __pyx_t_10, __pyx_t_18, __pyx_callargs+1, 5) < (0)) __PYX_ERR(0, 810, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_src_qp, __pyx_t_11, __pyx_t_18, __pyx_callargs+1, 6) < (0)) __PYX_ERR(0, 810, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_imm_data, __pyx_t_12, __pyx_t_18, __pyx_callargs+1, 7) < (0)) __PYX_ERR(0, 810, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_wc_flags, __pyx_t_13, __pyx_t_18, __pyx_callargs+1, 8) < (0)) __PYX_ERR(0, 810, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_pkey_index, __pyx_t_14, __pyx_t_18, __pyx_callargs+1, 9) < (0)) __PYX_ERR(0, 810, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_slid, __pyx_t_15, __pyx_t_18, __pyx_callargs+1, 10) < (0)) __PYX_ERR(0, 810, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_sl, __pyx_t_16, __pyx_t_18, __pyx_callargs+1, 11) < (0)) __PYX_ERR(0, 810, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_dlid_path_bits, __pyx_t_17, __pyx_t_18, __pyx_callargs+1, 12) < (0)) __PYX_ERR(0, 810, __pyx_L1_error)
    __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_WC, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_18);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 810, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_2);
  }
  __pyx_r = ((PyObject *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":798
 *             raise PyverbsRDMAErrno('Failed to Post Send')
 * 
 *     def get_recv_comp(self):             # <<<<<<<<<<<<<<
 *         """
 *         Polls the receive CQ associated with CMID for a work completion.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.get_recv_comp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":817
 *                   dlid_path_bits=wc.dlid_path_bits)
 * 
 *     def get_send_comp(self):             # <<<<<<<<<<<<<<
 *         """
 *         Polls the send CQ associated with CMID for a work completion.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_57get_send_comp(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_56get_send_comp, "\n        Polls the send CQ associated with CMID for a work completion.\n        :return: The retrieved WC or None if there is no completions\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_57get_send_comp = {"get_send_comp", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_57get_send_comp, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_56get_send_comp};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_57get_send_comp(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_send_comp (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("get_send_comp", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("get_send_comp", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_56get_send_comp(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_56get_send_comp(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self) {
  struct ibv_wc __pyx_v_wc;
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_send_comp", 0);

  /* "pyverbs/cmid.pyx":823
 *         """
 *         cdef v.ibv_wc wc
 *         with nogil:             # <<<<<<<<<<<<<<
 *             ret = cm.rdma_get_send_comp(self.id, &wc)
 *         if ret < 0:
*/
  {
      PyThreadState * _save;
      _save = PyEval_SaveThread();
      __Pyx_FastGIL_Remember();
      /*try:*/ {

        /* "pyverbs/cmid.pyx":824
 *         cdef v.ibv_wc wc
 *         with nogil:
 *             ret = cm.rdma_get_send_comp(self.id, &wc)             # <<<<<<<<<<<<<<
 *         if ret < 0:
 *             raise PyverbsRDMAErrno('Failed to retrieve send completion')
*/
        __pyx_v_ret = rdma_get_send_comp(__pyx_v_self->id, (&__pyx_v_wc));
      }

      /* "pyverbs/cmid.pyx":823
 *         """
 *         cdef v.ibv_wc wc
 *         with nogil:             # <<<<<<<<<<<<<<
 *             ret = cm.rdma_get_send_comp(self.id, &wc)
 *         if ret < 0:
*/
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_FastGIL_Forget();
          PyEval_RestoreThread(_save);
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "pyverbs/cmid.pyx":825
 *         with nogil:
 *             ret = cm.rdma_get_send_comp(self.id, &wc)
 *         if ret < 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to retrieve send completion')
 *         elif ret == 0:
*/
  __pyx_t_1 = (__pyx_v_ret < 0);
  if (unlikely(__pyx_t_1)) {

    /* "pyverbs/cmid.pyx":826
 *             ret = cm.rdma_get_send_comp(self.id, &wc)
 *         if ret < 0:
 *             raise PyverbsRDMAErrno('Failed to retrieve send completion')             # <<<<<<<<<<<<<<
 *         elif ret == 0:
 *             return None
*/
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 826, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_Failed_to_retrieve_send_completi};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 826, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 826, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":825
 *         with nogil:
 *             ret = cm.rdma_get_send_comp(self.id, &wc)
 *         if ret < 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to retrieve send completion')
 *         elif ret == 0:
*/
  }

  /* "pyverbs/cmid.pyx":827
 *         if ret < 0:
 *             raise PyverbsRDMAErrno('Failed to retrieve send completion')
 *         elif ret == 0:             # <<<<<<<<<<<<<<
 *             return None
 *         return WC(wr_id=wc.wr_id, status=wc.status, opcode=wc.opcode,
*/
  __pyx_t_1 = (__pyx_v_ret == 0);
  if (__pyx_t_1) {

    /* "pyverbs/cmid.pyx":828
 *             raise PyverbsRDMAErrno('Failed to retrieve send completion')
 *         elif ret == 0:
 *             return None             # <<<<<<<<<<<<<<
 *         return WC(wr_id=wc.wr_id, status=wc.status, opcode=wc.opcode,
 *                   vendor_err=wc.vendor_err, byte_len=wc.byte_len,
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "pyverbs/cmid.pyx":827
 *         if ret < 0:
 *             raise PyverbsRDMAErrno('Failed to retrieve send completion')
 *         elif ret == 0:             # <<<<<<<<<<<<<<
 *             return None
 *         return WC(wr_id=wc.wr_id, status=wc.status, opcode=wc.opcode,
*/
  }

  /* "pyverbs/cmid.pyx":829
 *         elif ret == 0:
 *             return None
 *         return WC(wr_id=wc.wr_id, status=wc.status, opcode=wc.opcode,             # <<<<<<<<<<<<<<
 *                   vendor_err=wc.vendor_err, byte_len=wc.byte_len,
 *                   qp_num=wc.qp_num, src_qp=wc.src_qp,
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = NULL;
  __pyx_t_3 = __Pyx_PyLong_From_unsigned_long(__pyx_v_wc.wr_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(__pyx_v_wc.status); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py(__pyx_v_wc.opcode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "pyverbs/cmid.pyx":830
 *             return None
 *         return WC(wr_id=wc.wr_id, status=wc.status, opcode=wc.opcode,
 *                   vendor_err=wc.vendor_err, byte_len=wc.byte_len,             # <<<<<<<<<<<<<<
 *                   qp_num=wc.qp_num, src_qp=wc.src_qp,
 *                   imm_data=wc.imm_data, wc_flags=wc.wc_flags,
*/
  __pyx_t_8 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.vendor_err); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.byte_len); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "pyverbs/cmid.pyx":831
 *         return WC(wr_id=wc.wr_id, status=wc.status, opcode=wc.opcode,
 *                   vendor_err=wc.vendor_err, byte_len=wc.byte_len,
 *                   qp_num=wc.qp_num, src_qp=wc.src_qp,             # <<<<<<<<<<<<<<
 *                   imm_data=wc.imm_data, wc_flags=wc.wc_flags,
 *                   pkey_index=wc.pkey_index, slid=wc.slid, sl=wc.sl,
*/
  __pyx_t_10 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.qp_num); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.src_qp); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);

  /* "pyverbs/cmid.pyx":832
 *                   vendor_err=wc.vendor_err, byte_len=wc.byte_len,
 *                   qp_num=wc.qp_num, src_qp=wc.src_qp,
 *                   imm_data=wc.imm_data, wc_flags=wc.wc_flags,             # <<<<<<<<<<<<<<
 *                   pkey_index=wc.pkey_index, slid=wc.slid, sl=wc.sl,
 *                   dlid_path_bits=wc.dlid_path_bits)
*/
  __pyx_t_12 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.imm_data); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __Pyx_PyLong_From_int(__pyx_v_wc.wc_flags); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);

  /* "pyverbs/cmid.pyx":833
 *                   qp_num=wc.qp_num, src_qp=wc.src_qp,
 *                   imm_data=wc.imm_data, wc_flags=wc.wc_flags,
 *                   pkey_index=wc.pkey_index, slid=wc.slid, sl=wc.sl,             # <<<<<<<<<<<<<<
 *                   dlid_path_bits=wc.dlid_path_bits)
 * 
*/
  __pyx_t_14 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.pkey_index); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.slid); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_16 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.sl); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);

  /* "pyverbs/cmid.pyx":834
 *                   imm_data=wc.imm_data, wc_flags=wc.wc_flags,
 *                   pkey_index=wc.pkey_index, slid=wc.slid, sl=wc.sl,
 *                   dlid_path_bits=wc.dlid_path_bits)             # <<<<<<<<<<<<<<
 * 
 *     def set_option(self, level, optname, optval, optlen):
*/
  __pyx_t_17 = __Pyx_PyLong_From_unsigned_int(__pyx_v_wc.dlid_path_bits); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_5 = 1;
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 13 : 0)] = {__pyx_t_4, NULL};
    __pyx_t_18 = __Pyx_MakeVectorcallBuilderKwds(13); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 829, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_wr_id, __pyx_t_3, __pyx_t_18, __pyx_callargs+1, 0) < (0)) __PYX_ERR(0, 829, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_status, __pyx_t_6, __pyx_t_18, __pyx_callargs+1, 1) < (0)) __PYX_ERR(0, 829, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_opcode, __pyx_t_7, __pyx_t_18, __pyx_callargs+1, 2) < (0)) __PYX_ERR(0, 829, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_vendor_err, __pyx_t_8, __pyx_t_18, __pyx_callargs+1, 3) < (0)) __PYX_ERR(0, 829, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_byte_len, __pyx_t_9, __pyx_t_18, __pyx_callargs+1, 4) < (0)) __PYX_ERR(0, 829, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_qp_num, __pyx_t_10, __pyx_t_18, __pyx_callargs+1, 5) < (0)) __PYX_ERR(0, 829, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_src_qp, __pyx_t_11, __pyx_t_18, __pyx_callargs+1, 6) < (0)) __PYX_ERR(0, 829, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_imm_data, __pyx_t_12, __pyx_t_18, __pyx_callargs+1, 7) < (0)) __PYX_ERR(0, 829, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_wc_flags, __pyx_t_13, __pyx_t_18, __pyx_callargs+1, 8) < (0)) __PYX_ERR(0, 829, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_pkey_index, __pyx_t_14, __pyx_t_18, __pyx_callargs+1, 9) < (0)) __PYX_ERR(0, 829, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_slid, __pyx_t_15, __pyx_t_18, __pyx_callargs+1, 10) < (0)) __PYX_ERR(0, 829, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_sl, __pyx_t_16, __pyx_t_18, __pyx_callargs+1, 11) < (0)) __PYX_ERR(0, 829, __pyx_L1_error)
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_dlid_path_bits, __pyx_t_17, __pyx_t_18, __pyx_callargs+1, 12) < (0)) __PYX_ERR(0, 829, __pyx_L1_error)
    __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_2cq_WC, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_18);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 829, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_2);
  }
  __pyx_r = ((PyObject *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":817
 *                   dlid_path_bits=wc.dlid_path_bits)
 * 
 *     def get_send_comp(self):             # <<<<<<<<<<<<<<
 *         """
 *         Polls the send CQ associated with CMID for a work completion.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.get_send_comp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":836
 *                   dlid_path_bits=wc.dlid_path_bits)
 * 
 *     def set_option(self, level, optname, optval, optlen):             # <<<<<<<<<<<<<<
 *         """
 *         Set communication options for a CMID.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_59set_option(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_58set_option, "\n        Set communication options for a CMID.\n        :param level: The protocol level of the option to set.\n        :param optname: The name of the option to set.\n        :param optval: The option data.\n        :param optlen: The size of the data.\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_59set_option = {"set_option", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_59set_option, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_58set_option};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_59set_option(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_level = 0;
  PyObject *__pyx_v_optname = 0;
  PyObject *__pyx_v_optval = 0;
  PyObject *__pyx_v_optlen = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_option (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_level,&__pyx_mstate_global->__pyx_n_u_optname,&__pyx_mstate_global->__pyx_n_u_optval,&__pyx_mstate_global->__pyx_n_u_optlen,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 836, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 836, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 836, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 836, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 836, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "set_option", 0) < (0)) __PYX_ERR(0, 836, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 4; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("set_option", 1, 4, 4, i); __PYX_ERR(0, 836, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 4)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 836, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 836, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 836, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 836, __pyx_L3_error)
    }
    __pyx_v_level = values[0];
    __pyx_v_optname = values[1];
    __pyx_v_optval = values[2];
    __pyx_v_optlen = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_option", 1, 4, 4, __pyx_nargs); __PYX_ERR(0, 836, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.set_option", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_58set_option(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_level, __pyx_v_optname, __pyx_v_optval, __pyx_v_optlen);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_58set_option(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_level, PyObject *__pyx_v_optname, PyObject *__pyx_v_optval, PyObject *__pyx_v_optlen) {
  uint8_t __pyx_v_value;
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  uint8_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_option", 0);

  /* "pyverbs/cmid.pyx":844
 *         :param optlen: The size of the data.
 *         """
 *         if optname != ce.RDMA_OPTION_ID_ACK_TIMEOUT:             # <<<<<<<<<<<<<<
 *             raise PyverbsUserError('Currently only RDMA_OPTION_ID_ACK_TIMEOUT is supported in Pyverbs.')
 *         cdef uint8_t value = optval
*/
  __pyx_t_1 = __Pyx_PyLong_From___pyx_anon_enum(RDMA_OPTION_ID_ACK_TIMEOUT); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_optname, __pyx_t_1, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "pyverbs/cmid.pyx":845
 *         """
 *         if optname != ce.RDMA_OPTION_ID_ACK_TIMEOUT:
 *             raise PyverbsUserError('Currently only RDMA_OPTION_ID_ACK_TIMEOUT is supported in Pyverbs.')             # <<<<<<<<<<<<<<
 *         cdef uint8_t value = optval
 *         ret = cm.rdma_set_option(self.id, level, optname, <void*>&value, optlen)
*/
    __pyx_t_1 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PyverbsUserError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 845, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_1);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_mstate_global->__pyx_kp_u_Currently_only_RDMA_OPTION_ID_AC};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 845, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 845, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":844
 *         :param optlen: The size of the data.
 *         """
 *         if optname != ce.RDMA_OPTION_ID_ACK_TIMEOUT:             # <<<<<<<<<<<<<<
 *             raise PyverbsUserError('Currently only RDMA_OPTION_ID_ACK_TIMEOUT is supported in Pyverbs.')
 *         cdef uint8_t value = optval
*/
  }

  /* "pyverbs/cmid.pyx":846
 *         if optname != ce.RDMA_OPTION_ID_ACK_TIMEOUT:
 *             raise PyverbsUserError('Currently only RDMA_OPTION_ID_ACK_TIMEOUT is supported in Pyverbs.')
 *         cdef uint8_t value = optval             # <<<<<<<<<<<<<<
 *         ret = cm.rdma_set_option(self.id, level, optname, <void*>&value, optlen)
 *         if ret != 0:
*/
  __pyx_t_6 = __Pyx_PyLong_As_uint8_t(__pyx_v_optval); if (unlikely((__pyx_t_6 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 846, __pyx_L1_error)
  __pyx_v_value = __pyx_t_6;

  /* "pyverbs/cmid.pyx":847
 *             raise PyverbsUserError('Currently only RDMA_OPTION_ID_ACK_TIMEOUT is supported in Pyverbs.')
 *         cdef uint8_t value = optval
 *         ret = cm.rdma_set_option(self.id, level, optname, <void*>&value, optlen)             # <<<<<<<<<<<<<<
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to set option')
*/
  __pyx_t_7 = __Pyx_PyLong_As_int(__pyx_v_level); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 847, __pyx_L1_error)
  __pyx_t_8 = __Pyx_PyLong_As_int(__pyx_v_optname); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 847, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyLong_As_size_t(__pyx_v_optlen); if (unlikely((__pyx_t_5 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 847, __pyx_L1_error)
  __pyx_v_ret = rdma_set_option(__pyx_v_self->id, __pyx_t_7, __pyx_t_8, ((void *)(&__pyx_v_value)), __pyx_t_5);

  /* "pyverbs/cmid.pyx":848
 *         cdef uint8_t value = optval
 *         ret = cm.rdma_set_option(self.id, level, optname, <void*>&value, optlen)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to set option')
 * 
*/
  __pyx_t_3 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_3)) {

    /* "pyverbs/cmid.pyx":849
 *         ret = cm.rdma_set_option(self.id, level, optname, <void*>&value, optlen)
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to set option')             # <<<<<<<<<<<<<<
 * 
 *     def reject(self, private_data=None):
*/
    __pyx_t_4 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 849, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_mstate_global->__pyx_kp_u_Failed_to_set_option};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_1, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 849, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 849, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":848
 *         cdef uint8_t value = optval
 *         ret = cm.rdma_set_option(self.id, level, optname, <void*>&value, optlen)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to set option')
 * 
*/
  }

  /* "pyverbs/cmid.pyx":836
 *                   dlid_path_bits=wc.dlid_path_bits)
 * 
 *     def set_option(self, level, optname, optval, optlen):             # <<<<<<<<<<<<<<
 *         """
 *         Set communication options for a CMID.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.set_option", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":851
 *             raise PyverbsRDMAErrno('Failed to set option')
 * 
 *     def reject(self, private_data=None):             # <<<<<<<<<<<<<<
 *         """
 *         Reject a connection or datagram service lookup request.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_61reject(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_60reject, "\n        Reject a connection or datagram service lookup request.\n        :param private_data: Optional private data to send with the reject message.\n        :param private_data_len: Size (in bytes) of the private data being sent.\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_61reject = {"reject", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_61reject, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_60reject};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_61reject(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_private_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reject (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_private_data,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 851, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 851, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "reject", 0) < (0)) __PYX_ERR(0, 851, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));
    } else {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 851, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    __pyx_v_private_data = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("reject", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 851, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.reject", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_60reject(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_private_data);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_60reject(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_private_data) {
  Py_ssize_t __pyx_v_data_len;
  PyObject *__pyx_v_buffer = NULL;
  size_t __pyx_v_data_ptr;
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  size_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("reject", 0);

  /* "pyverbs/cmid.pyx":857
 *         :param private_data_len: Size (in bytes) of the private data being sent.
 *         """
 *         data_len = len(private_data) if private_data else 0             # <<<<<<<<<<<<<<
 *         buffer = ctypes.create_string_buffer(data_len)
 *         if (private_data):
*/
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_private_data); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 857, __pyx_L1_error)
  if (__pyx_t_2) {
    __pyx_t_3 = PyObject_Length(__pyx_v_private_data); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 857, __pyx_L1_error)
    __pyx_t_1 = __pyx_t_3;
  } else {
    __pyx_t_1 = 0;
  }
  __pyx_v_data_len = __pyx_t_1;

  /* "pyverbs/cmid.pyx":858
 *         """
 *         data_len = len(private_data) if private_data else 0
 *         buffer = ctypes.create_string_buffer(data_len)             # <<<<<<<<<<<<<<
 *         if (private_data):
 *             buffer.value = private_data
*/
  __pyx_t_5 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_ctypes); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_create_string_buffer); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyLong_FromSsize_t(__pyx_v_data_len); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
    assert(__pyx_t_5);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_7);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_7, __pyx__function);
    __pyx_t_8 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_6};
    __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 858, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __pyx_v_buffer = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "pyverbs/cmid.pyx":859
 *         data_len = len(private_data) if private_data else 0
 *         buffer = ctypes.create_string_buffer(data_len)
 *         if (private_data):             # <<<<<<<<<<<<<<
 *             buffer.value = private_data
 *         cdef size_t data_ptr = ctypes.addressof(buffer)
*/
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_private_data); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 859, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "pyverbs/cmid.pyx":860
 *         buffer = ctypes.create_string_buffer(data_len)
 *         if (private_data):
 *             buffer.value = private_data             # <<<<<<<<<<<<<<
 *         cdef size_t data_ptr = ctypes.addressof(buffer)
 * 
*/
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_buffer, __pyx_mstate_global->__pyx_n_u_value, __pyx_v_private_data) < (0)) __PYX_ERR(0, 860, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":859
 *         data_len = len(private_data) if private_data else 0
 *         buffer = ctypes.create_string_buffer(data_len)
 *         if (private_data):             # <<<<<<<<<<<<<<
 *             buffer.value = private_data
 *         cdef size_t data_ptr = ctypes.addressof(buffer)
*/
  }

  /* "pyverbs/cmid.pyx":861
 *         if (private_data):
 *             buffer.value = private_data
 *         cdef size_t data_ptr = ctypes.addressof(buffer)             # <<<<<<<<<<<<<<
 * 
 *         ret = cm.rdma_reject(self.id, <const void*>data_ptr, data_len)
*/
  __pyx_t_7 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_ctypes); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_addressof); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_8 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
    assert(__pyx_t_7);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_7);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
    __pyx_t_8 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_v_buffer};
    __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 861, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __pyx_t_8 = __Pyx_PyLong_As_size_t(__pyx_t_4); if (unlikely((__pyx_t_8 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 861, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_data_ptr = __pyx_t_8;

  /* "pyverbs/cmid.pyx":863
 *         cdef size_t data_ptr = ctypes.addressof(buffer)
 * 
 *         ret = cm.rdma_reject(self.id, <const void*>data_ptr, data_len)             # <<<<<<<<<<<<<<
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Reject Connection')
*/
  __pyx_v_ret = rdma_reject(__pyx_v_self->id, ((void const *)__pyx_v_data_ptr), __pyx_v_data_len);

  /* "pyverbs/cmid.pyx":864
 * 
 *         ret = cm.rdma_reject(self.id, <const void*>data_ptr, data_len)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Reject Connection')
 * 
*/
  __pyx_t_2 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_2)) {

    /* "pyverbs/cmid.pyx":865
 *         ret = cm.rdma_reject(self.id, <const void*>data_ptr, data_len)
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Reject Connection')             # <<<<<<<<<<<<<<
 * 
 *     def resolve_addrinfo(self, node=None, service=None, port_space=0, flags=0):
*/
    __pyx_t_5 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 865, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
      assert(__pyx_t_5);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_7, __pyx__function);
      __pyx_t_8 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_mstate_global->__pyx_kp_u_Failed_to_Reject_Connection};
      __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_8, (2-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 865, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 865, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":864
 * 
 *         ret = cm.rdma_reject(self.id, <const void*>data_ptr, data_len)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Reject Connection')
 * 
*/
  }

  /* "pyverbs/cmid.pyx":851
 *             raise PyverbsRDMAErrno('Failed to set option')
 * 
 *     def reject(self, private_data=None):             # <<<<<<<<<<<<<<
 *         """
 *         Reject a connection or datagram service lookup request.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.reject", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_buffer);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":867
 *             raise PyverbsRDMAErrno('Failed to Reject Connection')
 * 
 *     def resolve_addrinfo(self, node=None, service=None, port_space=0, flags=0):             # <<<<<<<<<<<<<<
 *         """
 *         Resolve source and destination address information using RDMA CM.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_63resolve_addrinfo(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_62resolve_addrinfo, "\n        Resolve source and destination address information using RDMA CM.\n        :param node: Name, dotted-decimal IPv4 or IPv6 hex address to connect to.\n                    If None, will use hints.ai_dst_addr if set.\n        :param service: Service name or port number.\n        :param port_space: RDMA port space (RDMA_PS_TCP, RDMA_PS_UDP, or RDMA_PS_IB).\n        :param flags: Optional flags to control resolution:\n                    RAI_PASSIVE: Get address for use as server\n                    RAI_NUMERICHOST: Node name is numeric\n                    RAI_FAMILY: Use specific address family\n                    RAI_SA: Use IB SA path record lookup\n                    RAI_DNS: Use DNS lookup for IP resolution\n        :return: None\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_63resolve_addrinfo = {"resolve_addrinfo", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_63resolve_addrinfo, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_62resolve_addrinfo};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_63resolve_addrinfo(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_node = 0;
  PyObject *__pyx_v_service = 0;
  PyObject *__pyx_v_port_space = 0;
  PyObject *__pyx_v_flags = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("resolve_addrinfo (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_node,&__pyx_mstate_global->__pyx_n_u_service,&__pyx_mstate_global->__pyx_n_u_port_space,&__pyx_mstate_global->__pyx_n_u_flags,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 867, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 867, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 867, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 867, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 867, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "resolve_addrinfo", 0) < (0)) __PYX_ERR(0, 867, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    } else {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 867, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 867, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 867, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 867, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
      if (!values[3]) values[3] = __Pyx_NewRef(((PyObject *)__pyx_mstate_global->__pyx_int_0));
    }
    __pyx_v_node = values[0];
    __pyx_v_service = values[1];
    __pyx_v_port_space = values[2];
    __pyx_v_flags = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("resolve_addrinfo", 0, 0, 4, __pyx_nargs); __PYX_ERR(0, 867, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.resolve_addrinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_62resolve_addrinfo(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v_node, __pyx_v_service, __pyx_v_port_space, __pyx_v_flags);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_62resolve_addrinfo(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, PyObject *__pyx_v_node, PyObject *__pyx_v_service, PyObject *__pyx_v_port_space, PyObject *__pyx_v_flags) {
  char *__pyx_v_srvc;
  char *__pyx_v_node_addr;
  struct rdma_addrinfo __pyx_v_hints;
  struct rdma_addrinfo *__pyx_v_hints_ptr;
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  char *__pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("resolve_addrinfo", 0);
  __Pyx_INCREF(__pyx_v_node);
  __Pyx_INCREF(__pyx_v_service);

  /* "pyverbs/cmid.pyx":882
 *         :return: None
 *         """
 *         cdef char* srvc = NULL             # <<<<<<<<<<<<<<
 *         cdef char* node_addr = NULL
 *         cdef cm.rdma_addrinfo hints
*/
  __pyx_v_srvc = NULL;

  /* "pyverbs/cmid.pyx":883
 *         """
 *         cdef char* srvc = NULL
 *         cdef char* node_addr = NULL             # <<<<<<<<<<<<<<
 *         cdef cm.rdma_addrinfo hints
 *         cdef cm.rdma_addrinfo *hints_ptr = NULL
*/
  __pyx_v_node_addr = NULL;

  /* "pyverbs/cmid.pyx":885
 *         cdef char* node_addr = NULL
 *         cdef cm.rdma_addrinfo hints
 *         cdef cm.rdma_addrinfo *hints_ptr = NULL             # <<<<<<<<<<<<<<
 * 
 *         if node is not None:
*/
  __pyx_v_hints_ptr = NULL;

  /* "pyverbs/cmid.pyx":887
 *         cdef cm.rdma_addrinfo *hints_ptr = NULL
 * 
 *         if node is not None:             # <<<<<<<<<<<<<<
 *             if isinstance(node, str):
 *                 node = node.encode('utf-8')
*/
  __pyx_t_1 = (__pyx_v_node != Py_None);
  if (__pyx_t_1) {

    /* "pyverbs/cmid.pyx":888
 * 
 *         if node is not None:
 *             if isinstance(node, str):             # <<<<<<<<<<<<<<
 *                 node = node.encode('utf-8')
 *             node_addr = <char*>node
*/
    __pyx_t_1 = PyUnicode_Check(__pyx_v_node); 
    if (__pyx_t_1) {

      /* "pyverbs/cmid.pyx":889
 *         if node is not None:
 *             if isinstance(node, str):
 *                 node = node.encode('utf-8')             # <<<<<<<<<<<<<<
 *             node_addr = <char*>node
 *         if service is not None:
*/
      __pyx_t_3 = __pyx_v_node;
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_utf_8};
        __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_encode, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 889, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_DECREF_SET(__pyx_v_node, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "pyverbs/cmid.pyx":888
 * 
 *         if node is not None:
 *             if isinstance(node, str):             # <<<<<<<<<<<<<<
 *                 node = node.encode('utf-8')
 *             node_addr = <char*>node
*/
    }

    /* "pyverbs/cmid.pyx":890
 *             if isinstance(node, str):
 *                 node = node.encode('utf-8')
 *             node_addr = <char*>node             # <<<<<<<<<<<<<<
 *         if service is not None:
 *             if isinstance(service, str):
*/
    __pyx_t_5 = __Pyx_PyObject_AsWritableString(__pyx_v_node); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(0, 890, __pyx_L1_error)
    __pyx_v_node_addr = ((char *)__pyx_t_5);

    /* "pyverbs/cmid.pyx":887
 *         cdef cm.rdma_addrinfo *hints_ptr = NULL
 * 
 *         if node is not None:             # <<<<<<<<<<<<<<
 *             if isinstance(node, str):
 *                 node = node.encode('utf-8')
*/
  }

  /* "pyverbs/cmid.pyx":891
 *                 node = node.encode('utf-8')
 *             node_addr = <char*>node
 *         if service is not None:             # <<<<<<<<<<<<<<
 *             if isinstance(service, str):
 *                 service = service.encode('utf-8')
*/
  __pyx_t_1 = (__pyx_v_service != Py_None);
  if (__pyx_t_1) {

    /* "pyverbs/cmid.pyx":892
 *             node_addr = <char*>node
 *         if service is not None:
 *             if isinstance(service, str):             # <<<<<<<<<<<<<<
 *                 service = service.encode('utf-8')
 *             srvc = <char*>service
*/
    __pyx_t_1 = PyUnicode_Check(__pyx_v_service); 
    if (__pyx_t_1) {

      /* "pyverbs/cmid.pyx":893
 *         if service is not None:
 *             if isinstance(service, str):
 *                 service = service.encode('utf-8')             # <<<<<<<<<<<<<<
 *             srvc = <char*>service
 * 
*/
      __pyx_t_3 = __pyx_v_service;
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_mstate_global->__pyx_kp_u_utf_8};
        __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_encode, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 893, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_DECREF_SET(__pyx_v_service, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "pyverbs/cmid.pyx":892
 *             node_addr = <char*>node
 *         if service is not None:
 *             if isinstance(service, str):             # <<<<<<<<<<<<<<
 *                 service = service.encode('utf-8')
 *             srvc = <char*>service
*/
    }

    /* "pyverbs/cmid.pyx":894
 *             if isinstance(service, str):
 *                 service = service.encode('utf-8')
 *             srvc = <char*>service             # <<<<<<<<<<<<<<
 * 
 *         hints_ptr = &hints
*/
    __pyx_t_5 = __Pyx_PyObject_AsWritableString(__pyx_v_service); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(0, 894, __pyx_L1_error)
    __pyx_v_srvc = ((char *)__pyx_t_5);

    /* "pyverbs/cmid.pyx":891
 *                 node = node.encode('utf-8')
 *             node_addr = <char*>node
 *         if service is not None:             # <<<<<<<<<<<<<<
 *             if isinstance(service, str):
 *                 service = service.encode('utf-8')
*/
  }

  /* "pyverbs/cmid.pyx":896
 *             srvc = <char*>service
 * 
 *         hints_ptr = &hints             # <<<<<<<<<<<<<<
 *         memset(hints_ptr, 0, sizeof(cm.rdma_addrinfo))
 *         hints.ai_port_space = port_space
*/
  __pyx_v_hints_ptr = (&__pyx_v_hints);

  /* "pyverbs/cmid.pyx":897
 * 
 *         hints_ptr = &hints
 *         memset(hints_ptr, 0, sizeof(cm.rdma_addrinfo))             # <<<<<<<<<<<<<<
 *         hints.ai_port_space = port_space
 *         hints.ai_flags = flags
*/
  (void)(memset(__pyx_v_hints_ptr, 0, (sizeof(struct rdma_addrinfo))));

  /* "pyverbs/cmid.pyx":898
 *         hints_ptr = &hints
 *         memset(hints_ptr, 0, sizeof(cm.rdma_addrinfo))
 *         hints.ai_port_space = port_space             # <<<<<<<<<<<<<<
 *         hints.ai_flags = flags
 * 
*/
  __pyx_t_6 = __Pyx_PyLong_As_int(__pyx_v_port_space); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 898, __pyx_L1_error)
  __pyx_v_hints.ai_port_space = __pyx_t_6;

  /* "pyverbs/cmid.pyx":899
 *         memset(hints_ptr, 0, sizeof(cm.rdma_addrinfo))
 *         hints.ai_port_space = port_space
 *         hints.ai_flags = flags             # <<<<<<<<<<<<<<
 * 
 *         # Set hint fields to support GID address
*/
  __pyx_t_6 = __Pyx_PyLong_As_int(__pyx_v_flags); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 899, __pyx_L1_error)
  __pyx_v_hints.ai_flags = __pyx_t_6;

  /* "pyverbs/cmid.pyx":902
 * 
 *         # Set hint fields to support GID address
 *         if port_space == ce.RDMA_PS_IB:             # <<<<<<<<<<<<<<
 *             hints.ai_flags |= ce.RAI_FAMILY | ce.RAI_NUMERICHOST
 *             hints.ai_family = cm.AF_IB
*/
  __pyx_t_2 = __Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py(RDMA_PS_IB); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 902, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_port_space, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 902, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 902, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "pyverbs/cmid.pyx":903
 *         # Set hint fields to support GID address
 *         if port_space == ce.RDMA_PS_IB:
 *             hints.ai_flags |= ce.RAI_FAMILY | ce.RAI_NUMERICHOST             # <<<<<<<<<<<<<<
 *             hints.ai_family = cm.AF_IB
 * 
*/
    __pyx_v_hints.ai_flags = (__pyx_v_hints.ai_flags | (RAI_FAMILY | RAI_NUMERICHOST));

    /* "pyverbs/cmid.pyx":904
 *         if port_space == ce.RDMA_PS_IB:
 *             hints.ai_flags |= ce.RAI_FAMILY | ce.RAI_NUMERICHOST
 *             hints.ai_family = cm.AF_IB             # <<<<<<<<<<<<<<
 * 
 *         ret = cm.rdma_resolve_addrinfo(self.id, node_addr, srvc, hints_ptr)
*/
    __pyx_v_hints.ai_family = AF_IB;

    /* "pyverbs/cmid.pyx":902
 * 
 *         # Set hint fields to support GID address
 *         if port_space == ce.RDMA_PS_IB:             # <<<<<<<<<<<<<<
 *             hints.ai_flags |= ce.RAI_FAMILY | ce.RAI_NUMERICHOST
 *             hints.ai_family = cm.AF_IB
*/
  }

  /* "pyverbs/cmid.pyx":906
 *             hints.ai_family = cm.AF_IB
 * 
 *         ret = cm.rdma_resolve_addrinfo(self.id, node_addr, srvc, hints_ptr)             # <<<<<<<<<<<<<<
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Resolve AddrInfo')
*/
  __pyx_v_ret = rdma_resolve_addrinfo(__pyx_v_self->id, __pyx_v_node_addr, __pyx_v_srvc, __pyx_v_hints_ptr);

  /* "pyverbs/cmid.pyx":907
 * 
 *         ret = cm.rdma_resolve_addrinfo(self.id, node_addr, srvc, hints_ptr)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Resolve AddrInfo')
 * 
*/
  __pyx_t_1 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pyverbs/cmid.pyx":908
 *         ret = cm.rdma_resolve_addrinfo(self.id, node_addr, srvc, hints_ptr)
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Resolve AddrInfo')             # <<<<<<<<<<<<<<
 * 
 *     def query_addrinfo(self):
*/
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 908, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_7, __pyx__function);
      __pyx_t_4 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Failed_to_Resolve_AddrInfo};
      __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_7, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 908, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 908, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":907
 * 
 *         ret = cm.rdma_resolve_addrinfo(self.id, node_addr, srvc, hints_ptr)
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Resolve AddrInfo')
 * 
*/
  }

  /* "pyverbs/cmid.pyx":867
 *             raise PyverbsRDMAErrno('Failed to Reject Connection')
 * 
 *     def resolve_addrinfo(self, node=None, service=None, port_space=0, flags=0):             # <<<<<<<<<<<<<<
 *         """
 *         Resolve source and destination address information using RDMA CM.
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.resolve_addrinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_node);
  __Pyx_XDECREF(__pyx_v_service);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyverbs/cmid.pyx":910
 *             raise PyverbsRDMAErrno('Failed to Resolve AddrInfo')
 * 
 *     def query_addrinfo(self):             # <<<<<<<<<<<<<<
 *         """
 *         Query the resolved address info from a CMID.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_65query_addrinfo(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyverbs_4cmid_4CMID_64query_addrinfo, "\n        Query the resolved address info from a CMID.\n        :return: AddrInfo object containing the resolved address information.\n        ");
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_65query_addrinfo = {"query_addrinfo", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_65query_addrinfo, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_64query_addrinfo};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_65query_addrinfo(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("query_addrinfo (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("query_addrinfo", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("query_addrinfo", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_64query_addrinfo(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_64query_addrinfo(struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self) {
  struct __pyx_obj_7pyverbs_4cmid_AddrInfo *__pyx_v_ai = NULL;
  int __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("query_addrinfo", 0);

  /* "pyverbs/cmid.pyx":915
 *         :return: AddrInfo object containing the resolved address information.
 *         """
 *         ai = AddrInfo()             # <<<<<<<<<<<<<<
 *         ret = cm.rdma_query_addrinfo(self.id, &(ai.addr_info))
 * 
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_AddrInfo, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 915, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_ai = ((struct __pyx_obj_7pyverbs_4cmid_AddrInfo *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyverbs/cmid.pyx":916
 *         """
 *         ai = AddrInfo()
 *         ret = cm.rdma_query_addrinfo(self.id, &(ai.addr_info))             # <<<<<<<<<<<<<<
 * 
 *         if ret != 0:
*/
  __pyx_v_ret = rdma_query_addrinfo(__pyx_v_self->id, (&__pyx_v_ai->addr_info));

  /* "pyverbs/cmid.pyx":918
 *         ret = cm.rdma_query_addrinfo(self.id, &(ai.addr_info))
 * 
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Query AddrInfo')
 * 
*/
  __pyx_t_4 = (__pyx_v_ret != 0);
  if (unlikely(__pyx_t_4)) {

    /* "pyverbs/cmid.pyx":919
 * 
 *         if ret != 0:
 *             raise PyverbsRDMAErrno('Failed to Query AddrInfo')             # <<<<<<<<<<<<<<
 * 
 *         return ai
*/
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 919, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_3 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Failed_to_Query_AddrInfo};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 919, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 919, __pyx_L1_error)

    /* "pyverbs/cmid.pyx":918
 *         ret = cm.rdma_query_addrinfo(self.id, &(ai.addr_info))
 * 
 *         if ret != 0:             # <<<<<<<<<<<<<<
 *             raise PyverbsRDMAErrno('Failed to Query AddrInfo')
 * 
*/
  }

  /* "pyverbs/cmid.pyx":921
 *             raise PyverbsRDMAErrno('Failed to Query AddrInfo')
 * 
 *         return ai             # <<<<<<<<<<<<<<
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_ai);
  __pyx_r = ((PyObject *)__pyx_v_ai);
  goto __pyx_L0;

  /* "pyverbs/cmid.pyx":910
 *             raise PyverbsRDMAErrno('Failed to Resolve AddrInfo')
 * 
 *     def query_addrinfo(self):             # <<<<<<<<<<<<<<
 *         """
 *         Query the resolved address info from a CMID.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyverbs.cmid.CMID.query_addrinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_ai);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.id cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_67__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_67__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_67__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_67__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_66__reduce_cython__(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_66__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self.id cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.id cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_id_cannot_be_converted_to_a, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.id cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cmid.CMID.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.id cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.id cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_69__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyverbs_4cmid_4CMID_69__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_69__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyverbs_4cmid_4CMID_69__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyverbs.cmid.CMID.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyverbs_4cmid_4CMID_68__setstate_cython__(((struct __pyx_obj_7pyverbs_4cmid_CMID *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyverbs_4cmid_4CMID_68__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyverbs_4cmid_CMID *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "self.id cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self.id cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_self_id_cannot_be_converted_to_a, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.id cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.id cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyverbs.cmid.CMID.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
/* #### Code section: module_exttypes ### */
static struct __pyx_vtabstruct_7pyverbs_4cmid_CMID __pyx_vtable_7pyverbs_4cmid_CMID;

static PyObject *__pyx_tp_new_7pyverbs_4cmid_CMID(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7pyverbs_4cmid_CMID *p;
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7pyverbs_4cmid_CMID *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM*)__pyx_vtabptr_7pyverbs_4cmid_CMID;
  p->event_channel = Py_None; Py_INCREF(Py_None);
  p->ctx = Py_None; Py_INCREF(Py_None);
  p->pd = Py_None; Py_INCREF(Py_None);
  p->mrs = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7pyverbs_4cmid_CMID(PyObject *o) {
  struct __pyx_obj_7pyverbs_4cmid_CMID *p = (struct __pyx_obj_7pyverbs_4cmid_CMID *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyverbs_4cmid_CMID) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_7pyverbs_4cmid_4CMID_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->event_channel);
  Py_CLEAR(p->ctx);
  Py_CLEAR(p->pd);
  Py_CLEAR(p->mrs);
  if (PyType_IS_GC(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) PyObject_GC_Track(o);
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_dealloc, destructor)(o); else
  #endif
  __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_7pyverbs_4cmid_CMID);
}

static int __pyx_tp_traverse_7pyverbs_4cmid_CMID(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7pyverbs_4cmid_CMID *p = (struct __pyx_obj_7pyverbs_4cmid_CMID *)o;
  #if !CYTHON_USE_MODULE_STATE
  e = 0;
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) {
    traverseproc traverse = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_traverse, traverseproc);
    if (traverse) { e = traverse(o, v, a); }
  } else
  #endif
  { e = __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7pyverbs_4cmid_CMID); }
  if (e) return e;
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  if (p->event_channel) {
    e = (*v)(p->event_channel, a); if (e) return e;
  }
  if (p->ctx) {
    e = (*v)(p->ctx, a); if (e) return e;
  }
  if (p->pd) {
    e = (*v)(p->pd, a); if (e) return e;
  }
  if (p->mrs) {
    e = (*v)(p->mrs, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyverbs_4cmid_CMID(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7pyverbs_4cmid_CMID *p = (struct __pyx_obj_7pyverbs_4cmid_CMID *)o;
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM)) {
    inquiry clear = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM, tp_clear, inquiry);
    if (clear) clear(o);
  } else
  #endif
  { __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7pyverbs_4cmid_CMID); }
  tmp = ((PyObject*)p->event_channel);
  p->event_channel = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->ctx);
  p->ctx = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->pd);
  p->pd = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->mrs);
  p->mrs = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7pyverbs_4cmid_4CMID_dev_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_4cmid_4CMID_8dev_name_1__get__(o);
}

static PyObject *__pyx_getprop_7pyverbs_4cmid_4CMID_port_num(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_4cmid_4CMID_8port_num_1__get__(o);
}

static PyObject *__pyx_getprop_7pyverbs_4cmid_4CMID_event_channel(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_4cmid_4CMID_13event_channel_1__get__(o);
}

static PyObject *__pyx_getprop_7pyverbs_4cmid_4CMID_context(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_4cmid_4CMID_7context_1__get__(o);
}

static PyObject *__pyx_getprop_7pyverbs_4cmid_4CMID_pd(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_4cmid_4CMID_2pd_1__get__(o);
}

static PyObject *__pyx_getprop_7pyverbs_4cmid_4CMID_qpn(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_4cmid_4CMID_3qpn_1__get__(o);
}

static PyMethodDef __pyx_methods_7pyverbs_4cmid_CMID[] = {
  {"get_request", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_7get_request, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_6get_request},
  {"bind_addr", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_9bind_addr, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_8bind_addr},
  {"resolve_addr", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_11resolve_addr, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_10resolve_addr},
  {"join_multicast", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_13join_multicast, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_12join_multicast},
  {"leave_multicast", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_15leave_multicast, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_14leave_multicast},
  {"resolve_route", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_17resolve_route, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_16resolve_route},
  {"listen", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_19listen, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_18listen},
  {"connect", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_21connect, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_20connect},
  {"disconnect", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_23disconnect, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_22disconnect},
  {"accept", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_25accept, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_24accept},
  {"establish", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_27establish, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_26establish},
  {"set_local_ece", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_29set_local_ece, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_28set_local_ece},
  {"get_remote_ece", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_31get_remote_ece, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_30get_remote_ece},
  {"create_qp", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_33create_qp, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_32create_qp},
  {"query_qp", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_35query_qp, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_34query_qp},
  {"init_qp_attr", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_37init_qp_attr, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_36init_qp_attr},
  {"reg_msgs", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_39reg_msgs, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_38reg_msgs},
  {"reg_read", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_41reg_read, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_40reg_read},
  {"reg_write", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_43reg_write, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_42reg_write},
  {"post_recv", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_45post_recv, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_44post_recv},
  {"post_send", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_47post_send, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_46post_send},
  {"post_read", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_49post_read, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_48post_read},
  {"post_write", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_51post_write, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_50post_write},
  {"post_ud_send", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_53post_ud_send, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_52post_ud_send},
  {"get_recv_comp", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_55get_recv_comp, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_54get_recv_comp},
  {"get_send_comp", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_57get_send_comp, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_56get_send_comp},
  {"set_option", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_59set_option, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_58set_option},
  {"reject", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_61reject, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_60reject},
  {"resolve_addrinfo", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_63resolve_addrinfo, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_62resolve_addrinfo},
  {"query_addrinfo", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_65query_addrinfo, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_4CMID_64query_addrinfo},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_67__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_4CMID_69__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7pyverbs_4cmid_CMID[] = {
  {"dev_name", __pyx_getprop_7pyverbs_4cmid_4CMID_dev_name, 0, 0, 0},
  {"port_num", __pyx_getprop_7pyverbs_4cmid_4CMID_port_num, 0, 0, 0},
  {"event_channel", __pyx_getprop_7pyverbs_4cmid_4CMID_event_channel, 0, 0, 0},
  {"context", __pyx_getprop_7pyverbs_4cmid_4CMID_context, 0, 0, 0},
  {"pd", __pyx_getprop_7pyverbs_4cmid_4CMID_pd, 0, 0, 0},
  {"qpn", __pyx_getprop_7pyverbs_4cmid_4CMID_qpn, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyverbs_4cmid_CMID_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyverbs_4cmid_CMID},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyverbs_4cmid_CMID},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyverbs_4cmid_CMID},
  {Py_tp_methods, (void *)__pyx_methods_7pyverbs_4cmid_CMID},
  {Py_tp_getset, (void *)__pyx_getsets_7pyverbs_4cmid_CMID},
  {Py_tp_init, (void *)__pyx_pw_7pyverbs_4cmid_4CMID_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7pyverbs_4cmid_CMID},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyverbs_4cmid_CMID_spec = {
  "pyverbs.cmid.CMID",
  sizeof(struct __pyx_obj_7pyverbs_4cmid_CMID),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyverbs_4cmid_CMID_slots,
};
#else

static PyTypeObject __pyx_type_7pyverbs_4cmid_CMID = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyverbs.cmid.""CMID", /*tp_name*/
  sizeof(struct __pyx_obj_7pyverbs_4cmid_CMID), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyverbs_4cmid_CMID, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7pyverbs_4cmid_CMID, /*tp_traverse*/
  __pyx_tp_clear_7pyverbs_4cmid_CMID, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyverbs_4cmid_CMID, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7pyverbs_4cmid_CMID, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7pyverbs_4cmid_4CMID_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyverbs_4cmid_CMID, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7pyverbs_4cmid_CMEventChannel __pyx_vtable_7pyverbs_4cmid_CMEventChannel;

static PyObject *__pyx_tp_new_7pyverbs_4cmid_CMEventChannel(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *p;
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *)o);
  p->__pyx_vtab = __pyx_vtabptr_7pyverbs_4cmid_CMEventChannel;
  return o;
}

static void __pyx_tp_dealloc_7pyverbs_4cmid_CMEventChannel(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyverbs_4cmid_CMEventChannel) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_7pyverbs_4cmid_14CMEventChannel_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  if (PyType_IS_GC(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) PyObject_GC_Track(o);
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_dealloc, destructor)(o); else
  #endif
  __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_7pyverbs_4cmid_CMEventChannel);
}

static int __pyx_tp_traverse_7pyverbs_4cmid_CMEventChannel(PyObject *o, visitproc v, void *a) {
  int e;
  #if !CYTHON_USE_MODULE_STATE
  e = 0;
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    traverseproc traverse = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_traverse, traverseproc);
    if (traverse) { e = traverse(o, v, a); }
  } else
  #endif
  { e = __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7pyverbs_4cmid_CMEventChannel); }
  if (e) return e;
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyverbs_4cmid_CMEventChannel(PyObject *o) {
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    inquiry clear = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_clear, inquiry);
    if (clear) clear(o);
  } else
  #endif
  { __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7pyverbs_4cmid_CMEventChannel); }
  return 0;
}

static PyMethodDef __pyx_methods_7pyverbs_4cmid_CMEventChannel[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_14CMEventChannel_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_14CMEventChannel_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyverbs_4cmid_CMEventChannel_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyverbs_4cmid_CMEventChannel},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyverbs_4cmid_CMEventChannel},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyverbs_4cmid_CMEventChannel},
  {Py_tp_methods, (void *)__pyx_methods_7pyverbs_4cmid_CMEventChannel},
  {Py_tp_init, (void *)__pyx_pw_7pyverbs_4cmid_14CMEventChannel_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7pyverbs_4cmid_CMEventChannel},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyverbs_4cmid_CMEventChannel_spec = {
  "pyverbs.cmid.CMEventChannel",
  sizeof(struct __pyx_obj_7pyverbs_4cmid_CMEventChannel),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyverbs_4cmid_CMEventChannel_slots,
};
#else

static PyTypeObject __pyx_type_7pyverbs_4cmid_CMEventChannel = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyverbs.cmid.""CMEventChannel", /*tp_name*/
  sizeof(struct __pyx_obj_7pyverbs_4cmid_CMEventChannel), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyverbs_4cmid_CMEventChannel, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7pyverbs_4cmid_CMEventChannel, /*tp_traverse*/
  __pyx_tp_clear_7pyverbs_4cmid_CMEventChannel, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyverbs_4cmid_CMEventChannel, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7pyverbs_4cmid_14CMEventChannel_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyverbs_4cmid_CMEventChannel, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7pyverbs_4cmid_CMEvent __pyx_vtable_7pyverbs_4cmid_CMEvent;

static PyObject *__pyx_tp_new_7pyverbs_4cmid_CMEvent(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7pyverbs_4cmid_CMEvent *p;
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7pyverbs_4cmid_CMEvent *)o);
  p->__pyx_vtab = __pyx_vtabptr_7pyverbs_4cmid_CMEvent;
  return o;
}

static void __pyx_tp_dealloc_7pyverbs_4cmid_CMEvent(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyverbs_4cmid_CMEvent) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_7pyverbs_4cmid_7CMEvent_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  if (PyType_IS_GC(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) PyObject_GC_Track(o);
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_dealloc, destructor)(o); else
  #endif
  __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_7pyverbs_4cmid_CMEvent);
}

static int __pyx_tp_traverse_7pyverbs_4cmid_CMEvent(PyObject *o, visitproc v, void *a) {
  int e;
  #if !CYTHON_USE_MODULE_STATE
  e = 0;
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    traverseproc traverse = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_traverse, traverseproc);
    if (traverse) { e = traverse(o, v, a); }
  } else
  #endif
  { e = __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7pyverbs_4cmid_CMEvent); }
  if (e) return e;
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyverbs_4cmid_CMEvent(PyObject *o) {
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    inquiry clear = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_clear, inquiry);
    if (clear) clear(o);
  } else
  #endif
  { __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7pyverbs_4cmid_CMEvent); }
  return 0;
}

static PyObject *__pyx_getprop_7pyverbs_4cmid_7CMEvent_event_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_4cmid_7CMEvent_10event_type_1__get__(o);
}

static PyObject *__pyx_getprop_7pyverbs_4cmid_7CMEvent_private_data(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_4cmid_7CMEvent_12private_data_1__get__(o);
}

static PyMethodDef __pyx_methods_7pyverbs_4cmid_CMEvent[] = {
  {"ack_cm_event", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_7CMEvent_7ack_cm_event, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_7CMEvent_6ack_cm_event},
  {"event_str", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_7CMEvent_9event_str, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_7CMEvent_11__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_7CMEvent_13__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7pyverbs_4cmid_CMEvent[] = {
  {"event_type", __pyx_getprop_7pyverbs_4cmid_7CMEvent_event_type, 0, 0, 0},
  {"private_data", __pyx_getprop_7pyverbs_4cmid_7CMEvent_private_data, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyverbs_4cmid_CMEvent_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyverbs_4cmid_CMEvent},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyverbs_4cmid_CMEvent},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyverbs_4cmid_CMEvent},
  {Py_tp_methods, (void *)__pyx_methods_7pyverbs_4cmid_CMEvent},
  {Py_tp_getset, (void *)__pyx_getsets_7pyverbs_4cmid_CMEvent},
  {Py_tp_init, (void *)__pyx_pw_7pyverbs_4cmid_7CMEvent_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7pyverbs_4cmid_CMEvent},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyverbs_4cmid_CMEvent_spec = {
  "pyverbs.cmid.CMEvent",
  sizeof(struct __pyx_obj_7pyverbs_4cmid_CMEvent),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyverbs_4cmid_CMEvent_slots,
};
#else

static PyTypeObject __pyx_type_7pyverbs_4cmid_CMEvent = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyverbs.cmid.""CMEvent", /*tp_name*/
  sizeof(struct __pyx_obj_7pyverbs_4cmid_CMEvent), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyverbs_4cmid_CMEvent, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7pyverbs_4cmid_CMEvent, /*tp_traverse*/
  __pyx_tp_clear_7pyverbs_4cmid_CMEvent, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyverbs_4cmid_CMEvent, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7pyverbs_4cmid_CMEvent, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7pyverbs_4cmid_7CMEvent_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyverbs_4cmid_CMEvent, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7pyverbs_4cmid_AddrInfo __pyx_vtable_7pyverbs_4cmid_AddrInfo;

static PyObject *__pyx_tp_new_7pyverbs_4cmid_AddrInfo(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7pyverbs_4cmid_AddrInfo *p;
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7pyverbs_4cmid_AddrInfo *)o);
  p->__pyx_vtab = __pyx_vtabptr_7pyverbs_4cmid_AddrInfo;
  return o;
}

static void __pyx_tp_dealloc_7pyverbs_4cmid_AddrInfo(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyverbs_4cmid_AddrInfo) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_7pyverbs_4cmid_8AddrInfo_5__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  if (PyType_IS_GC(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) PyObject_GC_Track(o);
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_dealloc, destructor)(o); else
  #endif
  __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_7pyverbs_4cmid_AddrInfo);
}

static int __pyx_tp_traverse_7pyverbs_4cmid_AddrInfo(PyObject *o, visitproc v, void *a) {
  int e;
  #if !CYTHON_USE_MODULE_STATE
  e = 0;
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    traverseproc traverse = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_traverse, traverseproc);
    if (traverse) { e = traverse(o, v, a); }
  } else
  #endif
  { e = __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7pyverbs_4cmid_AddrInfo); }
  if (e) return e;
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyverbs_4cmid_AddrInfo(PyObject *o) {
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    inquiry clear = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_clear, inquiry);
    if (clear) clear(o);
  } else
  #endif
  { __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7pyverbs_4cmid_AddrInfo); }
  return 0;
}

static PyMethodDef __pyx_methods_7pyverbs_4cmid_AddrInfo[] = {
  {"get_addrinfo", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_8AddrInfo_3get_addrinfo, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyverbs_4cmid_8AddrInfo_2get_addrinfo},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_8AddrInfo_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_8AddrInfo_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyverbs_4cmid_AddrInfo_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyverbs_4cmid_AddrInfo},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyverbs_4cmid_AddrInfo},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyverbs_4cmid_AddrInfo},
  {Py_tp_methods, (void *)__pyx_methods_7pyverbs_4cmid_AddrInfo},
  {Py_tp_init, (void *)__pyx_pw_7pyverbs_4cmid_8AddrInfo_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7pyverbs_4cmid_AddrInfo},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyverbs_4cmid_AddrInfo_spec = {
  "pyverbs.cmid.AddrInfo",
  sizeof(struct __pyx_obj_7pyverbs_4cmid_AddrInfo),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyverbs_4cmid_AddrInfo_slots,
};
#else

static PyTypeObject __pyx_type_7pyverbs_4cmid_AddrInfo = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyverbs.cmid.""AddrInfo", /*tp_name*/
  sizeof(struct __pyx_obj_7pyverbs_4cmid_AddrInfo), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyverbs_4cmid_AddrInfo, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7pyverbs_4cmid_AddrInfo, /*tp_traverse*/
  __pyx_tp_clear_7pyverbs_4cmid_AddrInfo, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyverbs_4cmid_AddrInfo, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7pyverbs_4cmid_8AddrInfo_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyverbs_4cmid_AddrInfo, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7pyverbs_4cmid_ConnParam(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7pyverbs_4cmid_ConnParam *p;
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7pyverbs_4cmid_ConnParam *)o);
  p->data = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7pyverbs_4cmid_ConnParam(PyObject *o) {
  struct __pyx_obj_7pyverbs_4cmid_ConnParam *p = (struct __pyx_obj_7pyverbs_4cmid_ConnParam *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyverbs_4cmid_ConnParam) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->data);
  if (PyType_IS_GC(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) PyObject_GC_Track(o);
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_dealloc, destructor)(o); else
  #endif
  __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_7pyverbs_4cmid_ConnParam);
}

static int __pyx_tp_traverse_7pyverbs_4cmid_ConnParam(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7pyverbs_4cmid_ConnParam *p = (struct __pyx_obj_7pyverbs_4cmid_ConnParam *)o;
  #if !CYTHON_USE_MODULE_STATE
  e = 0;
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    traverseproc traverse = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_traverse, traverseproc);
    if (traverse) { e = traverse(o, v, a); }
  } else
  #endif
  { e = __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7pyverbs_4cmid_ConnParam); }
  if (e) return e;
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  if (p->data) {
    e = (*v)(p->data, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyverbs_4cmid_ConnParam(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7pyverbs_4cmid_ConnParam *p = (struct __pyx_obj_7pyverbs_4cmid_ConnParam *)o;
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    inquiry clear = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_clear, inquiry);
    if (clear) clear(o);
  } else
  #endif
  { __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7pyverbs_4cmid_ConnParam); }
  tmp = ((PyObject*)p->data);
  p->data = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7pyverbs_4cmid_9ConnParam_qpn(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_4cmid_9ConnParam_3qpn_1__get__(o);
}

static int __pyx_setprop_7pyverbs_4cmid_9ConnParam_qpn(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_4cmid_9ConnParam_3qpn_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_4cmid_9ConnParam_private_data(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_4cmid_9ConnParam_12private_data_1__get__(o);
}

static PyMethodDef __pyx_methods_7pyverbs_4cmid_ConnParam[] = {
  {"set_private_data", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_9ConnParam_3set_private_data, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_9ConnParam_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_9ConnParam_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7pyverbs_4cmid_ConnParam[] = {
  {"qpn", __pyx_getprop_7pyverbs_4cmid_9ConnParam_qpn, __pyx_setprop_7pyverbs_4cmid_9ConnParam_qpn, 0, 0},
  {"private_data", __pyx_getprop_7pyverbs_4cmid_9ConnParam_private_data, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyverbs_4cmid_ConnParam_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyverbs_4cmid_ConnParam},
  {Py_tp_str, (void *)__pyx_pw_7pyverbs_4cmid_9ConnParam_5__str__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyverbs_4cmid_ConnParam},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyverbs_4cmid_ConnParam},
  {Py_tp_methods, (void *)__pyx_methods_7pyverbs_4cmid_ConnParam},
  {Py_tp_getset, (void *)__pyx_getsets_7pyverbs_4cmid_ConnParam},
  {Py_tp_init, (void *)__pyx_pw_7pyverbs_4cmid_9ConnParam_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7pyverbs_4cmid_ConnParam},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyverbs_4cmid_ConnParam_spec = {
  "pyverbs.cmid.ConnParam",
  sizeof(struct __pyx_obj_7pyverbs_4cmid_ConnParam),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyverbs_4cmid_ConnParam_slots,
};
#else

static PyTypeObject __pyx_type_7pyverbs_4cmid_ConnParam = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyverbs.cmid.""ConnParam", /*tp_name*/
  sizeof(struct __pyx_obj_7pyverbs_4cmid_ConnParam), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyverbs_4cmid_ConnParam, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_7pyverbs_4cmid_9ConnParam_5__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7pyverbs_4cmid_ConnParam, /*tp_traverse*/
  __pyx_tp_clear_7pyverbs_4cmid_ConnParam, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyverbs_4cmid_ConnParam, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7pyverbs_4cmid_ConnParam, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7pyverbs_4cmid_9ConnParam_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyverbs_4cmid_ConnParam, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7pyverbs_4cmid_UDParam(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc_7pyverbs_4cmid_UDParam(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyverbs_4cmid_UDParam) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  if (PyType_IS_GC(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) PyObject_GC_Track(o);
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_dealloc, destructor)(o); else
  #endif
  __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_7pyverbs_4cmid_UDParam);
}

static int __pyx_tp_traverse_7pyverbs_4cmid_UDParam(PyObject *o, visitproc v, void *a) {
  int e;
  #if !CYTHON_USE_MODULE_STATE
  e = 0;
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    traverseproc traverse = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_traverse, traverseproc);
    if (traverse) { e = traverse(o, v, a); }
  } else
  #endif
  { e = __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7pyverbs_4cmid_UDParam); }
  if (e) return e;
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyverbs_4cmid_UDParam(PyObject *o) {
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    inquiry clear = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_clear, inquiry);
    if (clear) clear(o);
  } else
  #endif
  { __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7pyverbs_4cmid_UDParam); }
  return 0;
}

static PyObject *__pyx_getprop_7pyverbs_4cmid_7UDParam_qp_num(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_4cmid_7UDParam_6qp_num_1__get__(o);
}

static PyObject *__pyx_getprop_7pyverbs_4cmid_7UDParam_qkey(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_4cmid_7UDParam_4qkey_1__get__(o);
}

static PyObject *__pyx_getprop_7pyverbs_4cmid_7UDParam_ah_attr(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_4cmid_7UDParam_7ah_attr_1__get__(o);
}

static PyMethodDef __pyx_methods_7pyverbs_4cmid_UDParam[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_7UDParam_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_7UDParam_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7pyverbs_4cmid_UDParam[] = {
  {"qp_num", __pyx_getprop_7pyverbs_4cmid_7UDParam_qp_num, 0, 0, 0},
  {"qkey", __pyx_getprop_7pyverbs_4cmid_7UDParam_qkey, 0, 0, 0},
  {"ah_attr", __pyx_getprop_7pyverbs_4cmid_7UDParam_ah_attr, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyverbs_4cmid_UDParam_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyverbs_4cmid_UDParam},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyverbs_4cmid_UDParam},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyverbs_4cmid_UDParam},
  {Py_tp_methods, (void *)__pyx_methods_7pyverbs_4cmid_UDParam},
  {Py_tp_getset, (void *)__pyx_getsets_7pyverbs_4cmid_UDParam},
  {Py_tp_init, (void *)__pyx_pw_7pyverbs_4cmid_7UDParam_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7pyverbs_4cmid_UDParam},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyverbs_4cmid_UDParam_spec = {
  "pyverbs.cmid.UDParam",
  sizeof(struct __pyx_obj_7pyverbs_4cmid_UDParam),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyverbs_4cmid_UDParam_slots,
};
#else

static PyTypeObject __pyx_type_7pyverbs_4cmid_UDParam = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyverbs.cmid.""UDParam", /*tp_name*/
  sizeof(struct __pyx_obj_7pyverbs_4cmid_UDParam), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyverbs_4cmid_UDParam, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7pyverbs_4cmid_UDParam, /*tp_traverse*/
  __pyx_tp_clear_7pyverbs_4cmid_UDParam, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyverbs_4cmid_UDParam, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7pyverbs_4cmid_UDParam, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7pyverbs_4cmid_7UDParam_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyverbs_4cmid_UDParam, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_7pyverbs_4cmid_JoinMCAttrEx(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc_7pyverbs_4cmid_JoinMCAttrEx(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyverbs_4cmid_JoinMCAttrEx) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  if (PyType_IS_GC(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) PyObject_GC_Track(o);
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_dealloc, destructor)(o); else
  #endif
  __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_7pyverbs_4cmid_JoinMCAttrEx);
}

static int __pyx_tp_traverse_7pyverbs_4cmid_JoinMCAttrEx(PyObject *o, visitproc v, void *a) {
  int e;
  #if !CYTHON_USE_MODULE_STATE
  e = 0;
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    traverseproc traverse = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_traverse, traverseproc);
    if (traverse) { e = traverse(o, v, a); }
  } else
  #endif
  { e = __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7pyverbs_4cmid_JoinMCAttrEx); }
  if (e) return e;
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyverbs_4cmid_JoinMCAttrEx(PyObject *o) {
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject)) {
    inquiry clear = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject, tp_clear, inquiry);
    if (clear) clear(o);
  } else
  #endif
  { __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7pyverbs_4cmid_JoinMCAttrEx); }
  return 0;
}

static PyObject *__pyx_getprop_7pyverbs_4cmid_12JoinMCAttrEx_join_flags(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_10join_flags_1__get__(o);
}

static int __pyx_setprop_7pyverbs_4cmid_12JoinMCAttrEx_join_flags(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_10join_flags_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7pyverbs_4cmid_12JoinMCAttrEx_comp_mask(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_9comp_mask_1__get__(o);
}

static int __pyx_setprop_7pyverbs_4cmid_12JoinMCAttrEx_comp_mask(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_9comp_mask_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_7pyverbs_4cmid_JoinMCAttrEx[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7pyverbs_4cmid_JoinMCAttrEx[] = {
  {"join_flags", __pyx_getprop_7pyverbs_4cmid_12JoinMCAttrEx_join_flags, __pyx_setprop_7pyverbs_4cmid_12JoinMCAttrEx_join_flags, 0, 0},
  {"comp_mask", __pyx_getprop_7pyverbs_4cmid_12JoinMCAttrEx_comp_mask, __pyx_setprop_7pyverbs_4cmid_12JoinMCAttrEx_comp_mask, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyverbs_4cmid_JoinMCAttrEx_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyverbs_4cmid_JoinMCAttrEx},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyverbs_4cmid_JoinMCAttrEx},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyverbs_4cmid_JoinMCAttrEx},
  {Py_tp_methods, (void *)__pyx_methods_7pyverbs_4cmid_JoinMCAttrEx},
  {Py_tp_getset, (void *)__pyx_getsets_7pyverbs_4cmid_JoinMCAttrEx},
  {Py_tp_init, (void *)__pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7pyverbs_4cmid_JoinMCAttrEx},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyverbs_4cmid_JoinMCAttrEx_spec = {
  "pyverbs.cmid.JoinMCAttrEx",
  sizeof(struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyverbs_4cmid_JoinMCAttrEx_slots,
};
#else

static PyTypeObject __pyx_type_7pyverbs_4cmid_JoinMCAttrEx = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyverbs.cmid.""JoinMCAttrEx", /*tp_name*/
  sizeof(struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyverbs_4cmid_JoinMCAttrEx, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7pyverbs_4cmid_JoinMCAttrEx, /*tp_traverse*/
  __pyx_tp_clear_7pyverbs_4cmid_JoinMCAttrEx, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyverbs_4cmid_JoinMCAttrEx, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7pyverbs_4cmid_JoinMCAttrEx, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7pyverbs_4cmid_12JoinMCAttrEx_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyverbs_4cmid_JoinMCAttrEx, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
/* #### Code section: initfunc_declarations ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate); /*proto*/
/* #### Code section: init_module ### */

static int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py = Py_None; Py_INCREF(Py_None);
  __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py = Py_None; Py_INCREF(Py_None);
  __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py = Py_None; Py_INCREF(Py_None);
  __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py = Py_None; Py_INCREF(Py_None);
  __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py = Py_None; Py_INCREF(Py_None);
  __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_t_1 = PyImport_ImportModule("pyverbs.base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_4base_PyverbsCM = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.base", "PyverbsCM",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4base_PyverbsCM), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4base_PyverbsCM),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4base_PyverbsCM), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4base_PyverbsCM),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4base_PyverbsCM), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4base_PyverbsCM),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4base_PyverbsCM) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_4base_PyverbsCM = (struct __pyx_vtabstruct_7pyverbs_4base_PyverbsCM*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_4base_PyverbsCM); if (unlikely(!__pyx_vtabptr_7pyverbs_4base_PyverbsCM)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_4cmid_CMID = &__pyx_vtable_7pyverbs_4cmid_CMID;
  __pyx_vtable_7pyverbs_4cmid_CMID.__pyx_base = *__pyx_vtabptr_7pyverbs_4base_PyverbsCM;
  __pyx_vtable_7pyverbs_4cmid_CMID.__pyx_base.close = (PyObject *(*)(struct __pyx_obj_7pyverbs_4base_PyverbsCM *, int __pyx_skip_dispatch))__pyx_f_7pyverbs_4cmid_4CMID_close;
  __pyx_vtable_7pyverbs_4cmid_CMID.add_ref = (PyObject *(*)(struct __pyx_obj_7pyverbs_4cmid_CMID *, PyObject *))__pyx_f_7pyverbs_4cmid_4CMID_add_ref;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMID = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyverbs_4cmid_CMID_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMID)) __PYX_ERR(0, 347, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyverbs_4cmid_CMID_spec, __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMID) < (0)) __PYX_ERR(0, 347, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMID = &__pyx_type_7pyverbs_4cmid_CMID;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID->tp_base = __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsCM;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMID) < (0)) __PYX_ERR(0, 347, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMID);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMID->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMID->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMID->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMID, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 347, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_7pyverbs_4cmid_4CMID___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_7pyverbs_4cmid_4CMID___init__.doc = __pyx_doc_7pyverbs_4cmid_4CMID___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_7pyverbs_4cmid_4CMID___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_vtabptr_7pyverbs_4cmid_CMID) < (0)) __PYX_ERR(0, 347, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMID) < (0)) __PYX_ERR(0, 347, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_CMID, (PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMID) < (0)) __PYX_ERR(0, 347, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SLOTS
  if (__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMID->tp_weaklistoffset == 0) __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMID->tp_weaklistoffset = offsetof(struct __pyx_obj_7pyverbs_4cmid_CMID, __pyx_base.__pyx_base.__weakref__);
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMID) < (0)) __PYX_ERR(0, 347, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_4base_PyverbsObject = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.base", "PyverbsObject",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4base_PyverbsObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4base_PyverbsObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4base_PyverbsObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4base_PyverbsObject),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4base_PyverbsObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4base_PyverbsObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4base_PyverbsObject) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_4cmid_CMEventChannel = &__pyx_vtable_7pyverbs_4cmid_CMEventChannel;
  __pyx_vtable_7pyverbs_4cmid_CMEventChannel.close = (PyObject *(*)(struct __pyx_obj_7pyverbs_4cmid_CMEventChannel *, int __pyx_skip_dispatch))__pyx_f_7pyverbs_4cmid_14CMEventChannel_close;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEventChannel = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyverbs_4cmid_CMEventChannel_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEventChannel)) __PYX_ERR(0, 322, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyverbs_4cmid_CMEventChannel_spec, __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEventChannel) < (0)) __PYX_ERR(0, 322, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEventChannel = &__pyx_type_7pyverbs_4cmid_CMEventChannel;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMEventChannel->tp_base = __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEventChannel) < (0)) __PYX_ERR(0, 322, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEventChannel);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEventChannel->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEventChannel->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEventChannel->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEventChannel, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 322, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_7pyverbs_4cmid_14CMEventChannel___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_7pyverbs_4cmid_14CMEventChannel___init__.doc = __pyx_doc_7pyverbs_4cmid_14CMEventChannel___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_7pyverbs_4cmid_14CMEventChannel___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEventChannel, __pyx_vtabptr_7pyverbs_4cmid_CMEventChannel) < (0)) __PYX_ERR(0, 322, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEventChannel) < (0)) __PYX_ERR(0, 322, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_CMEventChannel, (PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEventChannel) < (0)) __PYX_ERR(0, 322, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SLOTS
  if (__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEventChannel->tp_weaklistoffset == 0) __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEventChannel->tp_weaklistoffset = offsetof(struct __pyx_obj_7pyverbs_4cmid_CMEventChannel, __pyx_base.__weakref__);
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEventChannel) < (0)) __PYX_ERR(0, 322, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_4cmid_CMEvent = &__pyx_vtable_7pyverbs_4cmid_CMEvent;
  __pyx_vtable_7pyverbs_4cmid_CMEvent.close = (PyObject *(*)(struct __pyx_obj_7pyverbs_4cmid_CMEvent *, int __pyx_skip_dispatch))__pyx_f_7pyverbs_4cmid_7CMEvent_close;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEvent = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyverbs_4cmid_CMEvent_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEvent)) __PYX_ERR(0, 268, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyverbs_4cmid_CMEvent_spec, __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEvent) < (0)) __PYX_ERR(0, 268, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEvent = &__pyx_type_7pyverbs_4cmid_CMEvent;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMEvent->tp_base = __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEvent) < (0)) __PYX_ERR(0, 268, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEvent);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEvent->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEvent->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEvent->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEvent, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 268, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_7pyverbs_4cmid_7CMEvent___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_7pyverbs_4cmid_7CMEvent___init__.doc = __pyx_doc_7pyverbs_4cmid_7CMEvent___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_7pyverbs_4cmid_7CMEvent___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEvent, __pyx_vtabptr_7pyverbs_4cmid_CMEvent) < (0)) __PYX_ERR(0, 268, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEvent) < (0)) __PYX_ERR(0, 268, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_CMEvent, (PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEvent) < (0)) __PYX_ERR(0, 268, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SLOTS
  if (__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEvent->tp_weaklistoffset == 0) __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEvent->tp_weaklistoffset = offsetof(struct __pyx_obj_7pyverbs_4cmid_CMEvent, __pyx_base.__weakref__);
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_CMEvent) < (0)) __PYX_ERR(0, 268, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_4cmid_AddrInfo = &__pyx_vtable_7pyverbs_4cmid_AddrInfo;
  __pyx_vtable_7pyverbs_4cmid_AddrInfo.close = (PyObject *(*)(struct __pyx_obj_7pyverbs_4cmid_AddrInfo *, int __pyx_skip_dispatch))__pyx_f_7pyverbs_4cmid_8AddrInfo_close;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_AddrInfo = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyverbs_4cmid_AddrInfo_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_AddrInfo)) __PYX_ERR(0, 167, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyverbs_4cmid_AddrInfo_spec, __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_AddrInfo) < (0)) __PYX_ERR(0, 167, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_AddrInfo = &__pyx_type_7pyverbs_4cmid_AddrInfo;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_AddrInfo->tp_base = __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_AddrInfo) < (0)) __PYX_ERR(0, 167, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_AddrInfo);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_AddrInfo->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_AddrInfo->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_AddrInfo->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_AddrInfo, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 167, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_7pyverbs_4cmid_8AddrInfo___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_7pyverbs_4cmid_8AddrInfo___init__.doc = __pyx_doc_7pyverbs_4cmid_8AddrInfo___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_7pyverbs_4cmid_8AddrInfo___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_AddrInfo, __pyx_vtabptr_7pyverbs_4cmid_AddrInfo) < (0)) __PYX_ERR(0, 167, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_AddrInfo) < (0)) __PYX_ERR(0, 167, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_AddrInfo, (PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_AddrInfo) < (0)) __PYX_ERR(0, 167, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SLOTS
  if (__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_AddrInfo->tp_weaklistoffset == 0) __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_AddrInfo->tp_weaklistoffset = offsetof(struct __pyx_obj_7pyverbs_4cmid_AddrInfo, __pyx_base.__weakref__);
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_AddrInfo) < (0)) __PYX_ERR(0, 167, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_ConnParam = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyverbs_4cmid_ConnParam_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_ConnParam)) __PYX_ERR(0, 21, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyverbs_4cmid_ConnParam_spec, __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_ConnParam) < (0)) __PYX_ERR(0, 21, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_ConnParam = &__pyx_type_7pyverbs_4cmid_ConnParam;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_ConnParam->tp_base = __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_ConnParam) < (0)) __PYX_ERR(0, 21, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_ConnParam);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_ConnParam->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_ConnParam->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_ConnParam->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_ConnParam, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 21, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_7pyverbs_4cmid_9ConnParam___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_7pyverbs_4cmid_9ConnParam___init__.doc = __pyx_doc_7pyverbs_4cmid_9ConnParam___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_7pyverbs_4cmid_9ConnParam___init__;
    }
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_ConnParam, (PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_ConnParam) < (0)) __PYX_ERR(0, 21, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SLOTS
  if (__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_ConnParam->tp_weaklistoffset == 0) __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_ConnParam->tp_weaklistoffset = offsetof(struct __pyx_obj_7pyverbs_4cmid_ConnParam, __pyx_base.__weakref__);
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_ConnParam) < (0)) __PYX_ERR(0, 21, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_UDParam = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyverbs_4cmid_UDParam_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_UDParam)) __PYX_ERR(0, 137, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyverbs_4cmid_UDParam_spec, __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_UDParam) < (0)) __PYX_ERR(0, 137, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_UDParam = &__pyx_type_7pyverbs_4cmid_UDParam;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_UDParam->tp_base = __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_UDParam) < (0)) __PYX_ERR(0, 137, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_UDParam);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_UDParam->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_UDParam->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_UDParam->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_UDParam, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 137, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_7pyverbs_4cmid_7UDParam___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_7pyverbs_4cmid_7UDParam___init__.doc = __pyx_doc_7pyverbs_4cmid_7UDParam___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_7pyverbs_4cmid_7UDParam___init__;
    }
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_UDParam, (PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_UDParam) < (0)) __PYX_ERR(0, 137, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SLOTS
  if (__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_UDParam->tp_weaklistoffset == 0) __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_UDParam->tp_weaklistoffset = offsetof(struct __pyx_obj_7pyverbs_4cmid_UDParam, __pyx_base.__weakref__);
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_UDParam) < (0)) __PYX_ERR(0, 137, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyverbs_4cmid_JoinMCAttrEx_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx)) __PYX_ERR(0, 105, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyverbs_4cmid_JoinMCAttrEx_spec, __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx) < (0)) __PYX_ERR(0, 105, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx = &__pyx_type_7pyverbs_4cmid_JoinMCAttrEx;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx->tp_base = __pyx_mstate_global->__pyx_ptype_7pyverbs_4base_PyverbsObject;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx) < (0)) __PYX_ERR(0, 105, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_UPDATE_DESCRIPTOR_DOC
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 105, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_7pyverbs_4cmid_12JoinMCAttrEx___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_7pyverbs_4cmid_12JoinMCAttrEx___init__.doc = __pyx_doc_7pyverbs_4cmid_12JoinMCAttrEx___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_7pyverbs_4cmid_12JoinMCAttrEx___init__;
    }
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_JoinMCAttrEx, (PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx) < (0)) __PYX_ERR(0, 105, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SLOTS
  if (__pyx_mstate->__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx->tp_weaklistoffset == 0) __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx->tp_weaklistoffset = offsetof(struct __pyx_obj_7pyverbs_4cmid_JoinMCAttrEx, __pyx_base.__weakref__);
  #endif
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyverbs_4cmid_JoinMCAttrEx) < (0)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("pyverbs.qp"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPCap = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.qp", "QPCap",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPCap), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPCap),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPCap), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPCap),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPCap), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPCap),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPCap) __PYX_ERR(2, 9, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.qp", "QPInitAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPInitAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPInitAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPInitAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttr) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.qp", "QPInitAttrEx",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPInitAttrEx),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPInitAttrEx) __PYX_ERR(2, 18, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.qp", "QPAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPAttr) __PYX_ERR(2, 27, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QP = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.qp", "QP",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2qp_QP), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QP),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2qp_QP), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QP),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2qp_QP), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QP),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QP) __PYX_ERR(2, 30, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2qp_QP = (struct __pyx_vtabstruct_7pyverbs_2qp_QP*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QP); if (unlikely(!__pyx_vtabptr_7pyverbs_2qp_QP)) __PYX_ERR(2, 30, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_DataBuffer = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.qp", "DataBuffer",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2qp_DataBuffer), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_DataBuffer),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2qp_DataBuffer), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_DataBuffer),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2qp_DataBuffer), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_DataBuffer),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_DataBuffer) __PYX_ERR(2, 47, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2qp_DataBuffer = (struct __pyx_vtabstruct_7pyverbs_2qp_DataBuffer*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_DataBuffer); if (unlikely(!__pyx_vtabptr_7pyverbs_2qp_DataBuffer)) __PYX_ERR(2, 47, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPEx = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.qp", "QPEx",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPEx),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPEx),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2qp_QPEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_QPEx),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPEx) __PYX_ERR(2, 50, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2qp_QPEx = (struct __pyx_vtabstruct_7pyverbs_2qp_QPEx*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_QPEx); if (unlikely(!__pyx_vtabptr_7pyverbs_2qp_QPEx)) __PYX_ERR(2, 50, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2qp_ECE = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.qp", "ECE",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2qp_ECE), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_ECE),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2qp_ECE), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_ECE),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2qp_ECE), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2qp_ECE),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2qp_ECE) __PYX_ERR(2, 54, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2qp_ECE = (struct __pyx_vtabstruct_7pyverbs_2qp_ECE*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2qp_ECE); if (unlikely(!__pyx_vtabptr_7pyverbs_2qp_ECE)) __PYX_ERR(2, 54, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyverbs.addr"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_4addr_GID = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.addr", "GID",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4addr_GID), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_GID),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4addr_GID), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_GID),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4addr_GID), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_GID),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4addr_GID) __PYX_ERR(3, 11, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_4addr_GRH = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.addr", "GRH",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4addr_GRH), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_GRH),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4addr_GRH), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_GRH),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4addr_GRH), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_GRH),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4addr_GRH) __PYX_ERR(3, 14, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_4addr_GlobalRoute = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.addr", "GlobalRoute",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4addr_GlobalRoute), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_GlobalRoute),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4addr_GlobalRoute), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_GlobalRoute),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4addr_GlobalRoute), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_GlobalRoute),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4addr_GlobalRoute) __PYX_ERR(3, 17, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_4addr_AHAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.addr", "AHAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4addr_AHAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_AHAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4addr_AHAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_AHAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4addr_AHAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_AHAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4addr_AHAttr) __PYX_ERR(3, 20, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_4addr_AHAttr = (struct __pyx_vtabstruct_7pyverbs_4addr_AHAttr*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_4addr_AHAttr); if (unlikely(!__pyx_vtabptr_7pyverbs_4addr_AHAttr)) __PYX_ERR(3, 20, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_4addr_AH = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.addr", "AH",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_4addr_AH), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_AH),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_4addr_AH), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_AH),
  #else
  sizeof(struct __pyx_obj_7pyverbs_4addr_AH), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_4addr_AH),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_4addr_AH) __PYX_ERR(3, 24, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_4addr_AH = (struct __pyx_vtabstruct_7pyverbs_4addr_AH*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_4addr_AH); if (unlikely(!__pyx_vtabptr_7pyverbs_4addr_AH)) __PYX_ERR(3, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyverbs.device"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_Context = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "Context",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_Context), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_Context),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_Context), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_Context),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_Context), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_Context),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_Context) __PYX_ERR(4, 10, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_6device_Context = (struct __pyx_vtabstruct_7pyverbs_6device_Context*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_6device_Context); if (unlikely(!__pyx_vtabptr_7pyverbs_6device_Context)) __PYX_ERR(4, 10, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_DeviceAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "DeviceAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_DeviceAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DeviceAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_DeviceAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DeviceAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_DeviceAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DeviceAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_DeviceAttr) __PYX_ERR(4, 33, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_QueryDeviceExInput = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "QueryDeviceExInput",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_QueryDeviceExInput),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_QueryDeviceExInput) __PYX_ERR(4, 36, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_ODPCaps = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "ODPCaps",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_ODPCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_ODPCaps),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_ODPCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_ODPCaps),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_ODPCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_ODPCaps),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_ODPCaps) __PYX_ERR(4, 39, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_RSSCaps = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "RSSCaps",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_RSSCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_RSSCaps),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_RSSCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_RSSCaps),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_RSSCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_RSSCaps),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_RSSCaps) __PYX_ERR(4, 43, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_PacketPacingCaps = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "PacketPacingCaps",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_PacketPacingCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PacketPacingCaps),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_PacketPacingCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PacketPacingCaps),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_PacketPacingCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PacketPacingCaps),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_PacketPacingCaps) __PYX_ERR(4, 46, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_PCIAtomicCaps = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "PCIAtomicCaps",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PCIAtomicCaps),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_PCIAtomicCaps) __PYX_ERR(4, 49, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_TMCaps = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "TMCaps",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_TMCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_TMCaps),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_TMCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_TMCaps),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_TMCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_TMCaps),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_TMCaps) __PYX_ERR(4, 52, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_CQModerationCaps = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "CQModerationCaps",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_CQModerationCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_CQModerationCaps),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_CQModerationCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_CQModerationCaps),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_CQModerationCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_CQModerationCaps),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_CQModerationCaps) __PYX_ERR(4, 55, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_TSOCaps = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "TSOCaps",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_TSOCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_TSOCaps),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_TSOCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_TSOCaps),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_TSOCaps), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_TSOCaps),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_TSOCaps) __PYX_ERR(4, 58, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_DeviceAttrEx = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "DeviceAttrEx",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_DeviceAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DeviceAttrEx),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_DeviceAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DeviceAttrEx),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_DeviceAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DeviceAttrEx),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_DeviceAttrEx) __PYX_ERR(4, 61, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_AllocDmAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "AllocDmAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_AllocDmAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_AllocDmAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_AllocDmAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_AllocDmAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_AllocDmAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_AllocDmAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_AllocDmAttr) __PYX_ERR(4, 64, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_DM = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "DM",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_DM), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DM),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_DM), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DM),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_DM), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_DM),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_DM) __PYX_ERR(4, 67, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_6device_DM = (struct __pyx_vtabstruct_7pyverbs_6device_DM*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_6device_DM); if (unlikely(!__pyx_vtabptr_7pyverbs_6device_DM)) __PYX_ERR(4, 67, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_PortAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "PortAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_PortAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PortAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_PortAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PortAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_PortAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_PortAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_PortAttr) __PYX_ERR(4, 74, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_GIDEntry = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "GIDEntry",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_GIDEntry), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_GIDEntry),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_GIDEntry), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_GIDEntry),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_GIDEntry), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_GIDEntry),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_GIDEntry) __PYX_ERR(4, 77, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_AsyncEvent = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "AsyncEvent",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_AsyncEvent), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_AsyncEvent),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_AsyncEvent), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_AsyncEvent),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_AsyncEvent), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_AsyncEvent),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_AsyncEvent) __PYX_ERR(4, 80, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_6device_FdArr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.device", "FdArr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_6device_FdArr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_FdArr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_6device_FdArr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_FdArr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_6device_FdArr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_6device_FdArr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_6device_FdArr) __PYX_ERR(4, 83, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_6device_FdArr = (struct __pyx_vtabstruct_7pyverbs_6device_FdArr*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_6device_FdArr); if (unlikely(!__pyx_vtabptr_7pyverbs_6device_FdArr)) __PYX_ERR(4, 83, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyverbs.pd"); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_2pd_PD = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.pd", "PD",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2pd_PD), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_PD),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2pd_PD), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_PD),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2pd_PD), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_PD),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2pd_PD) __PYX_ERR(5, 12, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2pd_PD = (struct __pyx_vtabstruct_7pyverbs_2pd_PD*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2pd_PD); if (unlikely(!__pyx_vtabptr_7pyverbs_2pd_PD)) __PYX_ERR(5, 12, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2pd_ParentDomainInitAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.pd", "ParentDomainInitAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomainInitAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2pd_ParentDomainInitAttr) __PYX_ERR(5, 28, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2pd_ParentDomain = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.pd", "ParentDomain",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomain), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomain),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomain), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomain),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomain), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomain),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2pd_ParentDomain) __PYX_ERR(5, 34, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2pd_ParentDomain = (struct __pyx_vtabstruct_7pyverbs_2pd_ParentDomain*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2pd_ParentDomain); if (unlikely(!__pyx_vtabptr_7pyverbs_2pd_ParentDomain)) __PYX_ERR(5, 34, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2pd_ParentDomainContext = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.pd", "ParentDomainContext",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomainContext), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomainContext),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomainContext), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomainContext),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2pd_ParentDomainContext), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2pd_ParentDomainContext),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2pd_ParentDomainContext) __PYX_ERR(5, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyverbs.mr"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_2mr_MR = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.mr", "MR",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2mr_MR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MR),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2mr_MR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MR),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2mr_MR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MR),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2mr_MR) __PYX_ERR(6, 12, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2mr_MR = (struct __pyx_vtabstruct_7pyverbs_2mr_MR*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2mr_MR); if (unlikely(!__pyx_vtabptr_7pyverbs_2mr_MR)) __PYX_ERR(6, 12, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2mr_MREx = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.mr", "MREx",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2mr_MREx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MREx),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2mr_MREx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MREx),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2mr_MREx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MREx),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2mr_MREx) __PYX_ERR(6, 25, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2mr_MREx = (struct __pyx_vtabstruct_7pyverbs_2mr_MREx*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2mr_MREx); if (unlikely(!__pyx_vtabptr_7pyverbs_2mr_MREx)) __PYX_ERR(6, 25, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2mr_MWBindInfo = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.mr", "MWBindInfo",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2mr_MWBindInfo), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MWBindInfo),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2mr_MWBindInfo), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MWBindInfo),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2mr_MWBindInfo), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MWBindInfo),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2mr_MWBindInfo) __PYX_ERR(6, 28, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2mr_MWBindInfo = (struct __pyx_vtabstruct_7pyverbs_2mr_MWBindInfo*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2mr_MWBindInfo); if (unlikely(!__pyx_vtabptr_7pyverbs_2mr_MWBindInfo)) __PYX_ERR(6, 28, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2mr_MWBind = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.mr", "MWBind",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2mr_MWBind), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MWBind),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2mr_MWBind), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MWBind),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2mr_MWBind), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MWBind),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2mr_MWBind) __PYX_ERR(6, 32, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2mr_MWBind = (struct __pyx_vtabstruct_7pyverbs_2mr_MWBind*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2mr_MWBind); if (unlikely(!__pyx_vtabptr_7pyverbs_2mr_MWBind)) __PYX_ERR(6, 32, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2mr_MW = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.mr", "MW",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2mr_MW), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MW),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2mr_MW), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MW),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2mr_MW), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_MW),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2mr_MW) __PYX_ERR(6, 36, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2mr_MW = (struct __pyx_vtabstruct_7pyverbs_2mr_MW*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2mr_MW); if (unlikely(!__pyx_vtabptr_7pyverbs_2mr_MW)) __PYX_ERR(6, 36, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2mr_DMMR = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.mr", "DMMR",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2mr_DMMR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DMMR),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2mr_DMMR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DMMR),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2mr_DMMR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DMMR),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2mr_DMMR) __PYX_ERR(6, 40, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2mr_DMMR = (struct __pyx_vtabstruct_7pyverbs_2mr_DMMR*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2mr_DMMR); if (unlikely(!__pyx_vtabptr_7pyverbs_2mr_DMMR)) __PYX_ERR(6, 40, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2mr_DmaBufMR = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.mr", "DmaBufMR",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2mr_DmaBufMR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DmaBufMR),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2mr_DmaBufMR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DmaBufMR),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2mr_DmaBufMR), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DmaBufMR),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2mr_DmaBufMR) __PYX_ERR(6, 43, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2mr_DmaBufMR = (struct __pyx_vtabstruct_7pyverbs_2mr_DmaBufMR*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2mr_DmaBufMR); if (unlikely(!__pyx_vtabptr_7pyverbs_2mr_DmaBufMR)) __PYX_ERR(6, 43, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2mr_DmaHandleInitAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.mr", "DmaHandleInitAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2mr_DmaHandleInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DmaHandleInitAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2mr_DmaHandleInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DmaHandleInitAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2mr_DmaHandleInitAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DmaHandleInitAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2mr_DmaHandleInitAttr) __PYX_ERR(6, 48, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2mr_DMAHandle = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.mr", "DMAHandle",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2mr_DMAHandle), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DMAHandle),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2mr_DMAHandle), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DMAHandle),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2mr_DMAHandle), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2mr_DMAHandle),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2mr_DMAHandle) __PYX_ERR(6, 51, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2mr_DMAHandle = (struct __pyx_vtabstruct_7pyverbs_2mr_DMAHandle*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2mr_DMAHandle); if (unlikely(!__pyx_vtabptr_7pyverbs_2mr_DMAHandle)) __PYX_ERR(6, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyverbs.cq"); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CompChannel = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.cq", "CompChannel",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2cq_CompChannel), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CompChannel),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2cq_CompChannel), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CompChannel),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2cq_CompChannel), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CompChannel),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CompChannel) __PYX_ERR(7, 9, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2cq_CompChannel = (struct __pyx_vtabstruct_7pyverbs_2cq_CompChannel*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CompChannel); if (unlikely(!__pyx_vtabptr_7pyverbs_2cq_CompChannel)) __PYX_ERR(7, 9, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQ = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.cq", "CQ",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2cq_CQ), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CQ),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2cq_CQ), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CQ),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2cq_CQ), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CQ),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQ) __PYX_ERR(7, 16, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2cq_CQ = (struct __pyx_vtabstruct_7pyverbs_2cq_CQ*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQ); if (unlikely(!__pyx_vtabptr_7pyverbs_2cq_CQ)) __PYX_ERR(7, 16, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.cq", "CqInitAttrEx",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CqInitAttrEx),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CqInitAttrEx) __PYX_ERR(7, 27, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQEX = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.cq", "CQEX",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2cq_CQEX), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CQEX),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2cq_CQEX), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CQEX),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2cq_CQEX), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_CQEX),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQEX) __PYX_ERR(7, 32, __pyx_L1_error)
  __pyx_vtabptr_7pyverbs_2cq_CQEX = (struct __pyx_vtabstruct_7pyverbs_2cq_CQEX*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyverbs_2cq_CQEX); if (unlikely(!__pyx_vtabptr_7pyverbs_2cq_CQEX)) __PYX_ERR(7, 32, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_WC = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.cq", "WC",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2cq_WC), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_WC),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2cq_WC), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_WC),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2cq_WC), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_WC),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_WC) __PYX_ERR(7, 42, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_PollCqAttr = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.cq", "PollCqAttr",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2cq_PollCqAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_PollCqAttr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2cq_PollCqAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_PollCqAttr),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2cq_PollCqAttr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_PollCqAttr),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_PollCqAttr) __PYX_ERR(7, 45, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyverbs_2cq_WcTmInfo = __Pyx_ImportType_3_2_4(__pyx_t_1, "pyverbs.cq", "WcTmInfo",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyverbs_2cq_WcTmInfo), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_WcTmInfo),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyverbs_2cq_WcTmInfo), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_WcTmInfo),
  #else
  sizeof(struct __pyx_obj_7pyverbs_2cq_WcTmInfo), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(struct __pyx_obj_7pyverbs_2cq_WcTmInfo),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7pyverbs_2cq_WcTmInfo) __PYX_ERR(7, 48, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  {
    __pyx_t_1 = PyImport_ImportModule("pyverbs.base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    const char * __pyx_import_signature = __Pyx_PyBytes_AsString(__pyx_mstate_global->__pyx_kp_b_PyObject_PyObject_close_weakrefs);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (unlikely(!__pyx_import_signature)) __PYX_ERR(0, 1, __pyx_L1_error)
    #endif
    const char * __pyx_import_name = __pyx_import_signature + 23;
    void (**const __pyx_import_pointers[])(void) = {(void (**)(void))&__pyx_f_7pyverbs_4base_close_weakrefs, (void (**)(void)) NULL};
    void (**const *__pyx_import_pointer)(void) = __pyx_import_pointers;
    const char *__pyx_import_current_signature = __pyx_import_signature;
    while (*__pyx_import_pointer) {
      if (__Pyx_ImportFunction_3_2_4(__pyx_t_1, __pyx_import_name, *__pyx_import_pointer, __pyx_import_current_signature) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
      ++__pyx_import_pointer;
      __pyx_import_name = strchr(__pyx_import_name, '\0') + 1;
      __pyx_import_signature = strchr(__pyx_import_signature, '\0') + 1;
      if (*__pyx_import_signature != '\0') __pyx_import_current_signature = __pyx_import_signature;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_cmid(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_cmid},
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  {Py_mod_gil, __Pyx_FREETHREADING_COMPATIBLE},
  #endif
  #if PY_VERSION_HEX >= 0x030C0000 && CYTHON_USE_MODULE_STATE
  {Py_mod_multiple_interpreters, Py_MOD_MULTIPLE_INTERPRETERS_NOT_SUPPORTED},
  #endif
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "cmid",
      0, /* m_doc */
    #if CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstatetype), /* m_size */
    #else
      (CYTHON_PEP489_MULTI_PHASE_INIT) ? 0 : -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif

/* PyModInitFuncType */
#ifndef CYTHON_NO_PYINIT_EXPORT
  #define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#else
  #ifdef __cplusplus
  #define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
  #else
  #define __Pyx_PyMODINIT_FUNC PyObject *
  #endif
#endif

__Pyx_PyMODINIT_FUNC PyInit_cmid(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_cmid(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
/* ModuleCreationPEP489 */
#if CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
static PY_INT64_T __Pyx_GetCurrentInterpreterId(void) {
    {
        PyObject *module = PyImport_ImportModule("_interpreters"); // 3.13+ I think
        if (!module) {
            PyErr_Clear(); // just try the 3.8-3.12 version
            module = PyImport_ImportModule("_xxsubinterpreters");
            if (!module) goto bad;
        }
        PyObject *current = PyObject_CallMethod(module, "get_current", NULL);
        Py_DECREF(module);
        if (!current) goto bad;
        if (PyTuple_Check(current)) {
            PyObject *new_current = PySequence_GetItem(current, 0);
            Py_DECREF(current);
            current = new_current;
            if (!new_current) goto bad;
        }
        long long as_c_int = PyLong_AsLongLong(current);
        Py_DECREF(current);
        return as_c_int;
    }
  bad:
    PySys_WriteStderr("__Pyx_GetCurrentInterpreterId failed. Try setting the C define CYTHON_PEP489_MULTI_PHASE_INIT=0\n");
    return -1;
}
#endif
#if !CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    static PY_INT64_T main_interpreter_id = -1;
#if CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
    PY_INT64_T current_id = GraalPyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_GRAAL
    PY_INT64_T current_id = PyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
    PY_INT64_T current_id = __Pyx_GetCurrentInterpreterId();
#elif CYTHON_COMPILING_IN_LIMITED_API
    PY_INT64_T current_id = PyInterpreterState_GetID(PyInterpreterState_Get());
#else
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
#endif
    if (unlikely(current_id == -1)) {
        return -1;
    }
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return 0;
    } else if (unlikely(main_interpreter_id != current_id)) {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#endif
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    #if !CYTHON_USE_MODULE_STATE
    if (__Pyx_check_single_interpreter())
        return NULL;
    #endif
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_cmid(PyObject *__pyx_pyinit_module)
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  __pyx_mstatetype *__pyx_mstate = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'cmid' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #else
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_t_1 = __pyx_pyinit_module;
  Py_INCREF(__pyx_t_1);
  #else
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #if CYTHON_USE_MODULE_STATE
  {
    int add_module_result = __Pyx_State_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "cmid" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = __pyx_t_1;
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  PyUnstable_Module_SetGIL(__pyx_m, Py_MOD_GIL_USED);
  #endif
  __pyx_mstate = __pyx_mstate_global;
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_mstate->__pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_mstate->__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_mstate->__pyx_d);
  __pyx_mstate->__pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_mstate->__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_cython_runtime = __Pyx_PyImport_AddModuleRef("cython_runtime"); if (unlikely(!__pyx_mstate->__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_mstate->__pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /* ImportRefnannyAPI */
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
    PyErr_Clear();
    __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
    if (!__Pyx_RefNanny)
        Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  
__Pyx_RefNannySetupContext("PyInit_cmid", 0);
  __Pyx_init_runtime_version();
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_mstate->__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Library function declarations ---*/
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__pyx_module_is_main_pyverbs__cmid) {
    if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_name, __pyx_mstate_global->__pyx_n_u_main) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "pyverbs.cmid")) {
      if (unlikely((PyDict_SetItemString(modules, "pyverbs.cmid", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_CreateCodeObjects(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_export_code(__pyx_mstate);
  (void)__Pyx_modinit_function_export_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_type_init_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_function_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/

  /* "EnumTypeToPy":1
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py = None             # <<<<<<<<<<<<<<
 * 
 * @cname("__Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py")
*/
  __Pyx_INCREF(Py_None);
  __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py);
  __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py, Py_None);
  __Pyx_GIVEREF(Py_None);

  /* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py(rdma_port_space c_val):
 *     cdef object __pyx_enum
*/
  __Pyx_INCREF(Py_None);
  __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py);
  __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py, Py_None);
  __Pyx_GIVEREF(Py_None);

  /* "EnumTypeToPy":1
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py = None             # <<<<<<<<<<<<<<
 * 
 * @cname("__Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py")
*/
  __Pyx_INCREF(Py_None);
  __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py);
  __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py, Py_None);
  __Pyx_GIVEREF(Py_None);

  /* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_cm_event_type_to_py(rdma_cm_event_type c_val):
 *     cdef object __pyx_enum
*/
  __Pyx_INCREF(Py_None);
  __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py);
  __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_access_flags_to_py, Py_None);
  __Pyx_GIVEREF(Py_None);

  /* "EnumTypeToPy":1
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py = None             # <<<<<<<<<<<<<<
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py")
*/
  __Pyx_INCREF(Py_None);
  __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py);
  __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py, Py_None);
  __Pyx_GIVEREF(Py_None);

  /* "EnumTypeToPy":3
 * cdef object __pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py = None
 * 
 * @cname("__Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py")             # <<<<<<<<<<<<<<
 * cdef __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_status_to_py(ibv_wc_status c_val):
 *     cdef object __pyx_enum
*/
  __Pyx_INCREF(Py_None);
  __Pyx_XGOTREF(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py);
  __Pyx_DECREF_SET(__pyx_imported_enum___Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_wc_opcode_to_py, Py_None);
  __Pyx_GIVEREF(Py_None);

  /* "pyverbs/cmid.pyx":3
 * from libc.stdint cimport uintptr_t, uint8_t
 * from libc.string cimport memset
 * import weakref             # <<<<<<<<<<<<<<
 * 
 * import ctypes
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_weakref, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_weakref, __pyx_t_2) < (0)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":5
 * import weakref
 * 
 * import ctypes             # <<<<<<<<<<<<<<
 * from pyverbs.pyverbs_error import PyverbsUserError, PyverbsError, PyverbsRDMAError
 * from pyverbs.qp cimport QPInitAttr, QPAttr, ECE
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_ctypes, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ctypes, __pyx_t_2) < (0)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":6
 * 
 * import ctypes
 * from pyverbs.pyverbs_error import PyverbsUserError, PyverbsError, PyverbsRDMAError             # <<<<<<<<<<<<<<
 * from pyverbs.qp cimport QPInitAttr, QPAttr, ECE
 * from pyverbs.base import PyverbsRDMAErrno
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_PyverbsUserError,__pyx_mstate_global->__pyx_n_u_PyverbsError,__pyx_mstate_global->__pyx_n_u_PyverbsRDMAError};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyverbs_pyverbs_error, __pyx_imported_names, 3, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_PyverbsUserError,__pyx_mstate_global->__pyx_n_u_PyverbsError,__pyx_mstate_global->__pyx_n_u_PyverbsRDMAError};
    for (__pyx_t_3=0; __pyx_t_3 < 3; __pyx_t_3++) {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 6, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 6, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":8
 * from pyverbs.pyverbs_error import PyverbsUserError, PyverbsError, PyverbsRDMAError
 * from pyverbs.qp cimport QPInitAttr, QPAttr, ECE
 * from pyverbs.base import PyverbsRDMAErrno             # <<<<<<<<<<<<<<
 * from pyverbs.base cimport close_weakrefs
 * cimport pyverbs.libibverbs_enums as e
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyverbs_base, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_PyverbsRDMAErrno};
    __pyx_t_3 = 0; {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 8, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_imported_names[__pyx_t_3], __pyx_t_4) < (0)) __PYX_ERR(0, 8, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":79
 *         return bytes(data_array)
 * 
 *     def set_private_data(self, data):             # <<<<<<<<<<<<<<
 *         if (min(len(self.data), len(data)) == 0):
 *             return
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_9ConnParam_3set_private_data, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ConnParam_set_private_data, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_ConnParam, __pyx_mstate_global->__pyx_n_u_set_private_data, __pyx_t_2) < (0)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.conn_param cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_9ConnParam_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ConnParam___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[1])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.conn_param cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.conn_param cannot be converted to a Python object for pickling"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_9ConnParam_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_ConnParam___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[2])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.join_mc_attr_ex cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_12JoinMCAttrEx_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_JoinMCAttrEx___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[3])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.join_mc_attr_ex cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.join_mc_attr_ex cannot be converted to a Python object for pickling"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_12JoinMCAttrEx_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_JoinMCAttrEx___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[4])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ud_param cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_7UDParam_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_UDParam___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[5])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.ud_param cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.ud_param cannot be converted to a Python object for pickling"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_7UDParam_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_UDParam___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[6])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":190
 *             self.get_addrinfo(src, dst, src_service, dst_service, port_space, flags)
 * 
 *     def get_addrinfo(self, src=None, dst=None, src_service=None, dst_service=None,             # <<<<<<<<<<<<<<
 *                      port_space=0, flags=0):
 *         """
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_8AddrInfo_3get_addrinfo, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AddrInfo_get_addrinfo, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[7])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[1]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_AddrInfo, __pyx_mstate_global->__pyx_n_u_get_addrinfo, __pyx_t_2) < (0)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":260
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.addr_info != NULL:
 *             if self.logger:
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_8AddrInfo_7close, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AddrInfo_close, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[8])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_AddrInfo, __pyx_mstate_global->__pyx_n_u_close, __pyx_t_2) < (0)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.addr_info cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_8AddrInfo_9__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AddrInfo___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[9])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.addr_info cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.addr_info cannot be converted to a Python object for pickling"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_8AddrInfo_11__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_AddrInfo___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[10])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":285
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.event != NULL:
 *             if self.logger:
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_7CMEvent_5close, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMEvent_close, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[11])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMEvent, __pyx_mstate_global->__pyx_n_u_close, __pyx_t_2) < (0)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":296
 *         return self.event.event
 * 
 *     def ack_cm_event(self):             # <<<<<<<<<<<<<<
 *         """
 *         Free a communication event. This call frees the event structure and any
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_7CMEvent_7ack_cm_event, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMEvent_ack_cm_event, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[12])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMEvent, __pyx_mstate_global->__pyx_n_u_ack_cm_event, __pyx_t_2) < (0)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":308
 *         self.event = NULL
 * 
 *     def event_str(self):             # <<<<<<<<<<<<<<
 *         if self.event == NULL:
 *             return ''
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_7CMEvent_9event_str, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMEvent_event_str, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[13])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMEvent, __pyx_mstate_global->__pyx_n_u_event_str, __pyx_t_2) < (0)) __PYX_ERR(0, 308, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.event cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_7CMEvent_11__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMEvent___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[14])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.event cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.event cannot be converted to a Python object for pickling"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_7CMEvent_13__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMEvent___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[15])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":339
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.event_channel != NULL:
 *             if self.logger:
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_14CMEventChannel_5close, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMEventChannel_close, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[16])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMEventChannel, __pyx_mstate_global->__pyx_n_u_close, __pyx_t_2) < (0)) __PYX_ERR(0, 339, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.event_channel cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_14CMEventChannel_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMEventChannel___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[17])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.event_channel cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.event_channel cannot be converted to a Python object for pickling"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_14CMEventChannel_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMEventChannel___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[18])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":350
 * 
 *     def __init__(self, object creator=None, QPInitAttr qp_init_attr=None,
 *                  PD pd=None, port_space=ce.RDMA_PS_TCP, CMID listen_id=None):             # <<<<<<<<<<<<<<
 *         """
 *         Initialize a CMID object over an underlying rdma_cm_id C object.
*/
  __pyx_t_2 = __Pyx_Enum_7pyverbs_15librdmacm_enums_enum__space_rdma_port_space_to_py(RDMA_PS_TCP); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate_global->__pyx_k__4 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":443
 *         self.close()
 * 
 *     cpdef close(self):             # <<<<<<<<<<<<<<
 *         if self.id != NULL:
 *             if self.logger:
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_5close, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_close, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[19])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_close, __pyx_t_2) < (0)) __PYX_ERR(0, 443, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":462
 *             self.id = NULL
 * 
 *     def get_request(self):             # <<<<<<<<<<<<<<
 *         """
 *         Retrieves the next pending connection request event. The call may only
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_7get_request, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_get_request, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[20])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_get_request, __pyx_t_2) < (0)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":481
 *         return to_conn
 * 
 *     def bind_addr(self, AddrInfo lai not None):             # <<<<<<<<<<<<<<
 *         """
 *         Associate a source address with a CMID. If binding to a specific local
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_9bind_addr, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_bind_addr, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[21])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_bind_addr, __pyx_t_2) < (0)) __PYX_ERR(0, 481, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":498
 *             self.pd = PD(self)
 * 
 *     def resolve_addr(self, AddrInfo rai not None, timeout_ms=2000):             # <<<<<<<<<<<<<<
 *         """
 *         Resolve destination and optional source addresses from IP addresses to
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_11resolve_addr, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_resolve_addr, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[22])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[2]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_resolve_addr, __pyx_t_2) < (0)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":512
 *             raise PyverbsRDMAErrno('Failed to Resolve Address')
 * 
 *     def join_multicast(self, AddrInfo addr=None, JoinMCAttrEx mc_attr=None,             # <<<<<<<<<<<<<<
 *                        context=0):
 *         """
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_13join_multicast, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_join_multicast, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[23])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[3]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_join_multicast, __pyx_t_2) < (0)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":535
 *             raise PyverbsRDMAErrno('Failed to Join multicast')
 * 
 *     def leave_multicast(self, AddrInfo addr not None):             # <<<<<<<<<<<<<<
 *         """
 *         Leaves a multicast group and detaches an associated QP from the group.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_15leave_multicast, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_leave_multicast, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[24])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_leave_multicast, __pyx_t_2) < (0)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":546
 *             raise PyverbsRDMAErrno('Failed to leave multicast')
 * 
 *     def resolve_route(self, timeout_ms=2000):             # <<<<<<<<<<<<<<
 *         """
 *         Resolve an RDMA route to the destination address in order to establish
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_17resolve_route, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_resolve_route, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[25])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[2]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_resolve_route, __pyx_t_2) < (0)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":565
 *             self.pd = PD(self)
 * 
 *     def listen(self, backlog=0):             # <<<<<<<<<<<<<<
 *         """
 *         Listen for incoming connection requests or datagram service lookup.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_19listen, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_listen, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[26])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[4]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_listen, __pyx_t_2) < (0)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":576
 *             raise PyverbsRDMAErrno('Listen Failed')
 * 
 *     def connect(self, ConnParam param=None):             # <<<<<<<<<<<<<<
 *         """
 *         Initiates an active connection request to a remote destination.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_21connect, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_connect, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[27])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[5]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_connect, __pyx_t_2) < (0)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":587
 *             raise PyverbsRDMAErrno('Failed to Connect')
 * 
 *     def disconnect(self):             # <<<<<<<<<<<<<<
 *         """
 *         Disconnects a connection and transitions any associated QP to error
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_23disconnect, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_disconnect, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[28])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_disconnect, __pyx_t_2) < (0)) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":597
 *             raise PyverbsRDMAErrno('Failed to Disconnect')
 * 
 *     def accept(self, ConnParam param=None):             # <<<<<<<<<<<<<<
 *         """
 *         Is called from the listening side to accept a connection or datagram
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_25accept, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_accept, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[29])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[5]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_accept, __pyx_t_2) < (0)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":609
 *             raise PyverbsRDMAErrno('Failed to Accept Connection')
 * 
 *     def establish(self):             # <<<<<<<<<<<<<<
 *         """
 *         Complete an active connection request.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_27establish, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_establish, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[30])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_establish, __pyx_t_2) < (0)) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":623
 *             raise PyverbsRDMAErrno('Failed to Complete an active connection request')
 * 
 *     def set_local_ece(self, ECE ece):             # <<<<<<<<<<<<<<
 *         """
 *         Set local ECE paraemters to be used for REQ/REP communication.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_29set_local_ece, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_set_local_ece, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[31])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_set_local_ece, __pyx_t_2) < (0)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":633
 *             raise PyverbsRDMAErrno('Failed to set local ECE')
 * 
 *     def get_remote_ece(self):             # <<<<<<<<<<<<<<
 *         """
 *         Get ECE parameters as were received from the communication peer.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_31get_remote_ece, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_get_remote_ece, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[32])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_get_remote_ece, __pyx_t_2) < (0)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":644
 *         return ece
 * 
 *     def create_qp(self, QPInitAttr qp_init not None):             # <<<<<<<<<<<<<<
 *         """
 *         Create a QP, which is associated with CMID.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_33create_qp, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_create_qp, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[33])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_create_qp, __pyx_t_2) < (0)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":657
 *             raise PyverbsRDMAErrno('Failed to Create QP')
 * 
 *     def query_qp(self, attr_mask):             # <<<<<<<<<<<<<<
 *         """
 *         Query QP using ibv_query_qp.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_35query_qp, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_query_qp, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[34])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_query_qp, __pyx_t_2) < (0)) __PYX_ERR(0, 657, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":670
 *         return attr, init_attr
 * 
 *     def init_qp_attr(self, qp_state):             # <<<<<<<<<<<<<<
 *         """
 *         Initialize a QPAttr object used for state transitions of an external
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_37init_qp_attr, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_init_qp_attr, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[35])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_init_qp_attr, __pyx_t_2) < (0)) __PYX_ERR(0, 670, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":690
 *         return qp_attr, attr_mask
 * 
 *     def reg_msgs(self, size):             # <<<<<<<<<<<<<<
 *         """
 *         Registers a memory region for sending or receiving messages or for
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_39reg_msgs, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_reg_msgs, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[36])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_reg_msgs, __pyx_t_2) < (0)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":700
 *         return MR(self, size, e.IBV_ACCESS_LOCAL_WRITE)
 * 
 *     def reg_read(self, size=0):             # <<<<<<<<<<<<<<
 *         """
 *         Registers a memory region for sending or receiving messages or for
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_41reg_read, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_reg_read, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[37])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[4]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_reg_read, __pyx_t_2) < (0)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":709
 *         return MR(self, size, e.IBV_ACCESS_REMOTE_READ)
 * 
 *     def reg_write(self, size=0):             # <<<<<<<<<<<<<<
 *         """
 *         Registers a memory region for sending or receiving messages or for
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_43reg_write, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_reg_write, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[38])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[4]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_reg_write, __pyx_t_2) < (0)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":718
 *         return MR(self, size, e.IBV_ACCESS_REMOTE_WRITE)
 * 
 *     def post_recv(self, MR mr not None, length=None):             # <<<<<<<<<<<<<<
 *         """
 *         Posts a recv_wr via QP associated with CMID.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_45post_recv, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_post_recv, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[39])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[5]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_post_recv, __pyx_t_2) < (0)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":732
 *             raise PyverbsRDMAErrno('Failed to Post Receive')
 * 
 *     def post_send(self, MR mr not None, flags=v.IBV_SEND_SIGNALED, length=None):             # <<<<<<<<<<<<<<
 *         """
 *         Posts a message via QP associated with CMID.
*/
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py(IBV_SEND_SIGNALED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate_global->__pyx_k__5 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py(IBV_SEND_SIGNALED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_t_2, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_47post_send, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_post_send, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[40])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_post_send, __pyx_t_2) < (0)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":748
 *             raise PyverbsRDMAErrno('Failed to Post Send')
 * 
 *     def post_read(self, MR mr not None, length, remote_addr, rkey,             # <<<<<<<<<<<<<<
 *                   flags=0):
 *         """
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_49post_read, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_post_read, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[41])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 748, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[4]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_post_read, __pyx_t_2) < (0)) __PYX_ERR(0, 748, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":764
 *             raise PyverbsRDMAErrno('Failed to Post Read')
 * 
 *     def post_write(self, MR mr not None, length, remote_addr, rkey,             # <<<<<<<<<<<<<<
 *                    flags=0):
 *         """
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_51post_write, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_post_write, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[42])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[4]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_post_write, __pyx_t_2) < (0)) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":781
 * 
 *     def post_ud_send(self, MR mr not None, AH ah not None, rqpn=0,
 *                      flags=v.IBV_SEND_SIGNALED, length=None):             # <<<<<<<<<<<<<<
 *         """
 *         Posts a message via UD QP associated with CMID to another UD QP.
*/
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py(IBV_SEND_SIGNALED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_mstate_global->__pyx_k__6 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":780
 *             raise PyverbsRDMAErrno('Failed to Post Write')
 * 
 *     def post_ud_send(self, MR mr not None, AH ah not None, rqpn=0,             # <<<<<<<<<<<<<<
 *                      flags=v.IBV_SEND_SIGNALED, length=None):
 *         """
*/
  __pyx_t_2 = __Pyx_Enum_7pyverbs_16libibverbs_enums_enum__space_ibv_send_flags_to_py(IBV_SEND_SIGNALED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "pyverbs/cmid.pyx":781
 * 
 *     def post_ud_send(self, MR mr not None, AH ah not None, rqpn=0,
 *                      flags=v.IBV_SEND_SIGNALED, length=None):             # <<<<<<<<<<<<<<
 *         """
 *         Posts a message via UD QP associated with CMID to another UD QP.
*/
  __pyx_t_4 = PyTuple_Pack(3, __pyx_mstate_global->__pyx_int_0, __pyx_t_2, Py_None); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":780
 *             raise PyverbsRDMAErrno('Failed to Post Write')
 * 
 *     def post_ud_send(self, MR mr not None, AH ah not None, rqpn=0,             # <<<<<<<<<<<<<<
 *                      flags=v.IBV_SEND_SIGNALED, length=None):
 *         """
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_53post_ud_send, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_post_ud_send, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[43])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_post_ud_send, __pyx_t_2) < (0)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":798
 *             raise PyverbsRDMAErrno('Failed to Post Send')
 * 
 *     def get_recv_comp(self):             # <<<<<<<<<<<<<<
 *         """
 *         Polls the receive CQ associated with CMID for a work completion.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_55get_recv_comp, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_get_recv_comp, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[44])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_get_recv_comp, __pyx_t_2) < (0)) __PYX_ERR(0, 798, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":817
 *                   dlid_path_bits=wc.dlid_path_bits)
 * 
 *     def get_send_comp(self):             # <<<<<<<<<<<<<<
 *         """
 *         Polls the send CQ associated with CMID for a work completion.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_57get_send_comp, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_get_send_comp, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[45])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_get_send_comp, __pyx_t_2) < (0)) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":836
 *                   dlid_path_bits=wc.dlid_path_bits)
 * 
 *     def set_option(self, level, optname, optval, optlen):             # <<<<<<<<<<<<<<
 *         """
 *         Set communication options for a CMID.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_59set_option, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_set_option, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[46])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_set_option, __pyx_t_2) < (0)) __PYX_ERR(0, 836, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":851
 *             raise PyverbsRDMAErrno('Failed to set option')
 * 
 *     def reject(self, private_data=None):             # <<<<<<<<<<<<<<
 *         """
 *         Reject a connection or datagram service lookup request.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_61reject, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_reject, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[47])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 851, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[5]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_reject, __pyx_t_2) < (0)) __PYX_ERR(0, 851, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":867
 *             raise PyverbsRDMAErrno('Failed to Reject Connection')
 * 
 *     def resolve_addrinfo(self, node=None, service=None, port_space=0, flags=0):             # <<<<<<<<<<<<<<
 *         """
 *         Resolve source and destination address information using RDMA CM.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_63resolve_addrinfo, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_resolve_addrinfo, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[48])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[6]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_resolve_addrinfo, __pyx_t_2) < (0)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":910
 *             raise PyverbsRDMAErrno('Failed to Resolve AddrInfo')
 * 
 *     def query_addrinfo(self):             # <<<<<<<<<<<<<<
 *         """
 *         Query the resolved address info from a CMID.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_65query_addrinfo, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID_query_addrinfo, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[49])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyverbs_4cmid_CMID, __pyx_mstate_global->__pyx_n_u_query_addrinfo, __pyx_t_2) < (0)) __PYX_ERR(0, 910, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.id cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_67__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[50])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self.id cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self.id cannot be converted to a Python object for pickling"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyverbs_4cmid_4CMID_69__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_CMID___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_pyverbs_cmid, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[51])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyverbs/cmid.pyx":1
 * from libc.stdint cimport uintptr_t, uint8_t             # <<<<<<<<<<<<<<
 * from libc.string cimport memset
 * import weakref
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_test, __pyx_t_2) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  if (__pyx_m) {
    if (__pyx_mstate->__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init pyverbs.cmid", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init pyverbs.cmid");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #else
  return __pyx_m;
  #endif
}
/* #### Code section: pystring_table ### */
/* #### Code section: cached_builtins ### */

static int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_super); if (!__pyx_builtin_super) __PYX_ERR(0, 53, __pyx_L1_error)

  /* Cached unbound methods */
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.method_name = &__pyx_mstate->__pyx_n_u_items;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.method_name = &__pyx_mstate->__pyx_n_u_pop;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.method_name = &__pyx_mstate->__pyx_n_u_values;
  __pyx_mstate->__pyx_umethod_PyUnicode_Type__format.type = (PyObject*)(&PyUnicode_Type);
  __pyx_mstate->__pyx_umethod_PyUnicode_Type__format.method_name = &__pyx_mstate->__pyx_n_u_format;
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "pyverbs/cmid.pyx":95
 *     def __str__(self):
 *         print_format  = '{:<4}: {:<4}\n'
 *         return '{}: {}\n'.format('Connection parameters', "") +\             # <<<<<<<<<<<<<<
 *                print_format.format('responder resources', self.conn_param.responder_resources) +\
 *                print_format.format('initiator depth', self.conn_param.initiator_depth) +\
*/
  __pyx_mstate_global->__pyx_tuple[0] = PyTuple_Pack(2, __pyx_mstate_global->__pyx_kp_u_Connection_parameters, __pyx_mstate_global->__pyx_kp_u__3); if (unlikely(!__pyx_mstate_global->__pyx_tuple[0])) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[0]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[0]);

  /* "pyverbs/cmid.pyx":190
 *             self.get_addrinfo(src, dst, src_service, dst_service, port_space, flags)
 * 
 *     def get_addrinfo(self, src=None, dst=None, src_service=None, dst_service=None,             # <<<<<<<<<<<<<<
 *                      port_space=0, flags=0):
 *         """
*/
  __pyx_mstate_global->__pyx_tuple[1] = PyTuple_Pack(6, Py_None, Py_None, Py_None, Py_None, __pyx_mstate_global->__pyx_int_0, __pyx_mstate_global->__pyx_int_0); if (unlikely(!__pyx_mstate_global->__pyx_tuple[1])) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[1]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[1]);

  /* "pyverbs/cmid.pyx":498
 *             self.pd = PD(self)
 * 
 *     def resolve_addr(self, AddrInfo rai not None, timeout_ms=2000):             # <<<<<<<<<<<<<<
 *         """
 *         Resolve destination and optional source addresses from IP addresses to
*/
  __pyx_mstate_global->__pyx_tuple[2] = PyTuple_Pack(1, __pyx_mstate_global->__pyx_int_2000); if (unlikely(!__pyx_mstate_global->__pyx_tuple[2])) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[2]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[2]);

  /* "pyverbs/cmid.pyx":512
 *             raise PyverbsRDMAErrno('Failed to Resolve Address')
 * 
 *     def join_multicast(self, AddrInfo addr=None, JoinMCAttrEx mc_attr=None,             # <<<<<<<<<<<<<<
 *                        context=0):
 *         """
*/
  __pyx_mstate_global->__pyx_tuple[3] = PyTuple_Pack(3, Py_None, Py_None, __pyx_mstate_global->__pyx_int_0); if (unlikely(!__pyx_mstate_global->__pyx_tuple[3])) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[3]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[3]);

  /* "pyverbs/cmid.pyx":565
 *             self.pd = PD(self)
 * 
 *     def listen(self, backlog=0):             # <<<<<<<<<<<<<<
 *         """
 *         Listen for incoming connection requests or datagram service lookup.
*/
  __pyx_mstate_global->__pyx_tuple[4] = PyTuple_Pack(1, __pyx_mstate_global->__pyx_int_0); if (unlikely(!__pyx_mstate_global->__pyx_tuple[4])) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[4]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[4]);

  /* "pyverbs/cmid.pyx":576
 *             raise PyverbsRDMAErrno('Listen Failed')
 * 
 *     def connect(self, ConnParam param=None):             # <<<<<<<<<<<<<<
 *         """
 *         Initiates an active connection request to a remote destination.
*/
  __pyx_mstate_global->__pyx_tuple[5] = PyTuple_Pack(1, Py_None); if (unlikely(!__pyx_mstate_global->__pyx_tuple[5])) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[5]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[5]);

  /* "pyverbs/cmid.pyx":867
 *             raise PyverbsRDMAErrno('Failed to Reject Connection')
 * 
 *     def resolve_addrinfo(self, node=None, service=None, port_space=0, flags=0):             # <<<<<<<<<<<<<<
 *         """
 *         Resolve source and destination address information using RDMA CM.
*/
  __pyx_mstate_global->__pyx_tuple[6] = PyTuple_Pack(4, Py_None, Py_None, __pyx_mstate_global->__pyx_int_0, __pyx_mstate_global->__pyx_int_0); if (unlikely(!__pyx_mstate_global->__pyx_tuple[6])) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[6]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[6]);
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_tuple;
    for (Py_ssize_t i=0; i<7; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  {
    const struct { const unsigned int length: 9; } index[] = {{1},{13},{30},{16},{15},{22},{12},{21},{17},{24},{66},{27},{32},{17},{47},{17},{19},{20},{24},{19},{22},{19},{20},{24},{27},{26},{25},{23},{20},{24},{31},{22},{26},{27},{24},{48},{25},{18},{37},{34},{23},{20},{62},{13},{179},{24},{7},{0},{1},{8},{7},{6},{173},{171},{170},{170},{174},{171},{12},{2},{15},{9},{22},{16},{9},{19},{11},{15},{66},{67},{62},{70},{59},{72},{65},{15},{5},{8},{26},{28},{14},{21},{7},{14},{32},{34},{20},{25},{27},{20},{13},{17},{4},{22},{24},{11},{14},{10},{12},{14},{15},{14},{18},{19},{16},{18},{17},{19},{20},{11},{14},{14},{14},{17},{15},{19},{13},{13},{13},{14},{11},{17},{21},{18},{18},{15},{9},{27},{29},{26},{23},{27},{18},{22},{18},{20},{27},{24},{22},{23},{21},{14},{15},{16},{17},{18},{19},{19},{14},{16},{14},{14},{16},{16},{12},{18},{19},{24},{16},{21},{21},{18},{19},{20},{23},{18},{16},{17},{11},{25},{20},{21},{25},{22},{17},{23},{20},{24},{11},{14},{13},{13},{13},{16},{14},{25},{14},{10},{19},{12},{30},{32},{12},{16},{16},{16},{20},{28},{31},{25},{24},{27},{27},{29},{30},{28},{26},{25},{29},{28},{22},{25},{28},{27},{25},{10},{13},{11},{11},{7},{25},{27},{7},{6},{12},{3},{4},{9},{2},{2},{18},{4},{9},{7},{9},{6},{8},{6},{7},{18},{5},{8},{4},{9},{4},{7},{7},{9},{20},{7},{6},{4},{8},{8},{5},{5},{10},{14},{3},{8},{11},{8},{3},{6},{9},{9},{10},{5},{12},{6},{8},{12},{13},{14},{11},{13},{12},{5},{9},{1},{16},{14},{13},{13},{8},{8},{9},{12},{13},{5},{10},{14},{3},{15},{6},{5},{6},{9},{8},{7},{12},{10},{2},{8},{4},{9},{3},{6},{6},{7},{6},{5},{2},{10},{3},{10},{9},{9},{9},{12},{10},{12},{3},{12},{12},{24},{23},{21},{11},{14},{7},{7},{12},{6},{8},{12},{14},{8},{3},{2},{18},{15},{10},{17},{13},{8},{8},{9},{6},{11},{3},{12},{16},{13},{9},{3},{5},{4},{9},{4},{4},{7},{13},{12},{10},{16},{10},{12},{19},{4},{2},{4},{3},{8},{6},{11},{8},{3},{4},{6},{5},{8},{10},{7},{5},{6},{10},{4},{8},{2},{8},{7},{5},{37},{93},{80},{141},{66},{47},{48},{53},{158},{41},{35},{45},{71},{35},{41},{57},{53},{80},{53},{34},{144},{88},{56},{48},{17},{20},{74},{117},{71},{204},{9},{53},{39},{52},{20},{52},{88},{65},{424}};
    #if (CYTHON_COMPRESS_STRINGS) == 3 && __PYX_LIMITED_VERSION_HEX >= 0x030e0000 /* compression: zstd (3855 bytes) */
const char* const cstring = "(\265/\375`j&-x\000*\207\214\034H\340\260X\0070\314\3434\306\363\004\316\363D\254\345i\236\346i\236\346i\206\240\337|UD\244\010\270\305\250\035\262p+\361,\317\324\210\376\203\267\322|\220}\3756\346z;c\367\264\256\tu\324o\026\2358\205\252\316\375\333\025\215\001\246\001\272\001\351F\273.{y\226c\313Dw\277AM\326\250S\236?\273\306\254]\344V\373\223\350\230\242y\274a\301\331(\306\261GI\263da\022'e\331Ob\326\223\351\210\265\212c+t\247\370\267F\221SW\211yK\233\254\314\371\217\243\316\260U\211\261K=\377yI~&IM\302\200\014\030+\364Ey\266\320S\305\362\334\374]\2228%\353\024|>\327\351\237XEc\230\365a\026\376i\254\233\245\272\302,\024\363\347\\\233\343jk\354\342\301\345\302\311;\341s\177\037\371\376Ui>y\2425=\246k:\210\317\357\274$\361~\276Kmm\215q\274\215+m\024\353\232\0245\365\210\232\234\357\377\375z\230~\363kxi\2564zb\013\036\327Q\327o\276*w\357\032\343\376\255QEC\316\377a&\312\307w\361al\277\212\352\345\2469?\365#\251?\304\235\243'=\226\335T]\250\314\310\313\356\204\352\374\325\327\357\021?L\343S\277\231\272\310\263|\221\001\003_\300\370[7\350i\203{\372\343\270a\264\233\245\206a\341\253\375\260\330\t\237\345F\323T\345\230\227XT\231\274\322'f\023\334\226\273/\203\246O<\231\362V\352\302\023\3051soUc\222\343\373\373\273\364\326)>\321\262\217(\327}\346\327\364\363a\036\312:\277M\222\033n\235%GMN\206\213\025\n\003\273\320\315\302\335\207K\034\376\256\177\217\270\243\347\2519\003\030\204\014\027\021 `\260\004g\214\353R\327h\026\312\371\367O\331\350\2479u\252\023'\232N/_N\235b8P\365\211\337F\343\2052\340g\222Im\334'\222T\245\230\324Gz\331R\037i\233\361\270o\306\353R\240ILy*\221\363@\320\304y#p\213\341r\347\315xah2\301\314p\254\017;\316\203\271\217e\207\263\237\253\320s\243{{\246\024\007\242Hy\242G\232\351\300\321&f\023H\362\272\222\250\nA\356\204\227Q\204\"*4\275l\271T*}9F\374f8\3237\223\273.<\241DSx\002=\322\311\344\211\233\230\312 \347yy&\364@o\324m9\206\023I\233\270\301\264h\010\275\231/\354N\033M*\237J&.\014\275|Ju\034\215H2\245H\0338\372R%U\212\313'\360#\201\237\307\201\036L""\356@/\345\245\266\227l\n\275\024\335\306}%\223\270eO4\225F\340\027z]h\312\235wb#y\341V\362\272\016\374^2M\311\023\277\220&\2236\316Dz!\211\336H\025#\202\236\230\372R\241w\"m/&\322(\307\244T$\230\234Jy\334\314\313\247\273&a\345NOYE7oqWn\375\276!URs4\375\361s\206\215\277yI\246\257\337\344Y\256\364\266\031u\251s\177\236\363\230\177T\311\3379\263x\270\300\352\352JK\377\356*\244\212\3639\217\256|\226\250\013\262\034s6m\3062e\235\372j,\263fa\231\317O5\306\177\334*\212\367\223'\232\362\3378}\221\245k\304@>Y\302\272\317$yY\323\003(\nN\\V\004\237\035\223\226DG\255\037H\366\342\005:qA\356\n\250\035\331\031 \266\324\216\264\236\026\203\324B\034\330\321:\362\251\211\326\236\023\342QC\261&,\014\017\033\003\255cN\213H\026\222uxG\256\310\375\366\244\005\265\241\020ZH\214\234\020\216\232\207EY\021t\000XPc\026\232q\031\316\353\362\206^\316\262\2752\332\024\022\232\t\0277\344\032hQ\330\222\326\365\320vV\263-\"\366W\324:`\375\260.\254^\023\254\336\316\207GM\273-\356\250\025\265\002Z\016\266[E\253\200\332F\003\364z\256\004\227\202Im\233I\230\251\310`^\t/\214\253\266.\037\332\207\345k\270&\360\250\275<:\000\214\324x\275\035\034\324<\263\022\0334Q+B\013\255<\307%\347\200\014\304\353j\023\264\"\273\261\036\353\207c\307\363\276\334\024-\214\215\300Q\023\261+\253\210Gm\313\361\340\250\241\334\315\265@BS\260\2028j!\031\010\025\232\n\tM\257-\353H\007\000\217\032\241V\302\006\215g\000+*\024L\261\323\276a\375\036o\306\233s\353\233\215\222\374\325\214\366w\225\235\177\377T\345/\330\0367\347&\377\027%\317\232\365\344\231\367\2717\377\2353\316\347q\227z\233Z}\301\322\347z\225\242)\316\347\330\333\244djjX\212\276K\237o\252$\232M\263\246\267Jmr\327\3672\3632\347\354r\225%\357gY\223\263\312\312\271kW~\362\177\374\021\266\177,a\235\223\274\227\253\\\323[\352\352\311\335\323q\243yw\336\177\314\346\270\307\363\315\030\213\245\256Z\250\271\262\341\332\305\377,\321r\014\"5\026\272mi[\240^\223\027\014\351\200tld2/\205\255\315k\202\336\246\240\030\030Y9\274\236\227\304\265a!\264\027D\324H<-\211\202\317""\004\275\003!\002\310]\3015\224N)\004-\304\n\355\005\033\341\345\276\332\t-\311ztT\310P\217\273\002\362\272hQL8\224lD\205\034\003\0335\317\214\303\006\217G\033a\005\261\277\206\000\261+r4\013\341\203\366\241wH\332\024\266\344\263\342a\201G\007\266\257$\037\024\026\304\276xm\340\330y5\350\000\000\241\306\246\363a\306i\240\201\355I(\244#\203\201\206A\313B\274-\026Z/\352\035\360v\025vD\262\003o\206\225\232g\346B\365\332\330\r\266\223\201\300\021jz\006S\201Va\203&\373`B\323\351\230\275d&^\232\r\256\034\020Vf\035\201\261\0166h/\266\300G4\033\311D\262\177O^\022F\004\327@\263\255\340\335\256\002\006\013<\354\252Uo\326\002+\267\252\004j\350W\306E\340z\"mH\026\312Ld\035<\321+\362v\270*X\rl\003W\303\007V\204e \022\276\237+\203\r\331\0072\r\262\321\373\301\025qSZ\0256@\310\266\"XA\037;\364n\270\2576%\004dk7BV\020\350\035]\224;\300\355LB\266U\302\n\343i7tXGo\201\3675B\033\361\250\201|\234l4^9;\241\225\233\300\025\333\255=YO\344\200t\0000\360~\270*\241Zh\026`\306!Cy\003\274(\240\022\006\335\235\340\222-\252\235\000\242\2232c\360\246\270\252\007\036v\303\352\341\363\001\232\245\310XD`;\020\260\343z\343yS\356\026B\373\350\225\200\247\355XQI;\303\216\260\342\232@\273!\332/\001\211Jh%##\256\307UY@[\340\366V\267\001\203\252\250\222\213\024\205P\325\214HA\222\024\n\215\001B\020\010\022\226\005z\240\307\372\302\241&\3430\005\rCdFH\204D\244<J\nR\220\014\007\252s\200\214\333\275&\373\315\250\r\221o\260\367\002A\016k\177\204\010\033\030\033\304{\227\332\227S~\034\330\335l\030\264\347\372P\025\242E\033g2\3402\353H\270\330\350d\301j\331\3622\211\027\016\267\222m(:\022b\263\274\035\236\277ufi0\313\221}&\206\330\264<\231(l\326\323\217\230\355a\010\013p$\202K\376R\024bK\033h)\222\227\330\215\265\356LA \200\213\300\017\345\264\004\311\3738\031\211O\251\230\223\221\tv\37468D\231\322V\321n\341\362i\274\375\340\272\016\233\211\302Y8\225\200\222\203a\r\356G\317\361J\003\316\366\300\271\014\306\3168\236+\371\254\231\315\300a\"\261\024\242E]\331\210\306\336\325:\265\240\2259\002\206\222P8J,\022\247""\n\334\301\277R-\313\371B\304\202\002\313/\365\241a{\035\256~ns\327\010\245\244Q0\260\340\330\001\026\360\333?\273(j\340\365\236\254\314\375;\222\035\317\325\351\361\212n\010\350}\270\215\031\\\206\310\305F\215\014Z/}\202\222\354\265\t\241\217\007&\204PM\362\027\260\360\334\t\206\036\211\014\265\233\256\245Vt\261\241\002\034\345\324\370\036\370\016\0137S\311s\250s`\200\031}\310\027\nS\232\275M 6\302K\311\340E\033\nB\226\265i\002\335\251sZ.T\004a\202\204\377\007l\306\207*\230\026\014\306!%e\240\342\r\303\325\264e\313\374`@\323\3726\351/\317\013\304\271\365l\206 \300\3520:&V7(\226'F\0146\360\253+\337\214\000C\334\314\224\212Q\237r\364\266\246\276Vi\273$\232X(\264\036x\033UmT\322\335\313\256\212\007z\254q\301\252B\313\3143\0055g?\204\247R;\306o\227\2474\002\017\034jC\342^c\203\033\036\231\337\242!V$|\232\035(\027\235\014w+\251\252\226\266nS$\220\202\336\365a\211n\020\353/-\253\320&#\267s\022i\362\364!\236\3212\n\263\264\002yu\231\000\377\354q_\365\024C2V1\354\346\231s=\314\265+\217\227\243~\344\313\017\231T\224\206e\026\311\202ofY\317\250\224\365W&\306\243d\2201\222'\224\257i\333\332\272\021\216=\231\005m\344\\\006\020\031\372\r\242T0\372\201\273\344\311\255l\202\236\346o\375D\316\020OC\027\004b]\322^!\373*e\037\n\0231%Pw\244\256\324\037\276\215p\333\215\361\314\210\266\203\360\214\306\210\3064\375\301\016,Hq\346M\025\332\013\026\007~H6\021\037\003\016\211\007\301jm\000\242\263\"nT\242\221~\342\235\312\025\026\261r\202X\244\243It\023\215\r\330\316\005\352\343\231\340\231\235_BnCqi\026\310cQ\365Nmv2\231\374\336\200\363C\177n\203\353YA\210c\247$\375]F\352\005\201\rO\233=]\022\2005W]\033(6\033\317\266\204m!\344\255p\344\245\210\350\r\302\000=#\257R\007:7\255F\024U\023>\263gm\251O\205gw=t\325\354]\343\007\250_\340\353\030\030\313+\205\211\252\360\000d\327$\335\225\367\024\213\\\324\002\300\366Z\020o\227#\003\262\346.\251j\213\320\362\3318\306P\322\376\266\213\016\246O\210\303|\022@\202\223d\353\344@=\355\373\223\261\205o\235\2066\022\337(\300\237\251\370\206\"\333\267\332""\226\0141\222\023)\352\031\364\261w\307\243l6y\030t\022x\300B\266*\340\231}\022d1W\002\\\017)\335\333\320\234\302s\333\323\005wL\035D*so\244rc\324\340X\021\242\253)@\3378I\361z\341e\242@\301\256\2058l\037\306\310\332\202\026\000\236\241?\024%gyv,\336W\220\320\302\351\214\330O\253\207\2348m\273+\232>-\346\014H\031U\211\201\027C\206\342\366\0009-\006\n\333?t\377\307\340\001J\033\3424\241f\351\230\265up\370\335\224\365\032\002\315S\027\3404*\352\301>J\361\203\366\000e\374\020\307gB0$P\\00\244\224\363#\222\367\350)\307\304\267p\022\025\356\323\233\372\037\341V\241\002\257\242\256\226-\205Z\203\277B\223\202*Y\252Cr\205N\264\033\247N\256l\371wR\260\013n\206\342\36443\235U\202\265\306p\013\343%w \216%\215\020>\263Fi\022]\343\"\374\220\014,\300\334\377}\030\033\365\341%75\335@|\234\n\262\234\343\362oK\033+K\304\027{#\374B\206\312\243C\257\363\214\033\363AdWr\363\334\355\320\206A\377\330%\312\255Q=R\034'ss\th\226R\223\004%\000\004\010M\244\351X\002\255dV\334\2367'\334s\342\336\357\267\346\354A3X\364&\355\233J\026\243\216jT*\272\206U\246\323I(2i\025\373\210\210\362)\344\021w5\020\215\2634DZ\216\017\367ZM\004S\366\024\037\277\365\336<yo^7\315UB\365S\324\253\250\276\233\312\251\362\226s\240a\201\325\001\220\360\340\347|2\\\314/\222]\253\304\306-\200\362<9\r\327T\272zF%,H\251\251\000\033\200 KC\276O\370R\r\261<T\3401\205\212\341\037\351}\225+\266\241\260@\t,\301:\035FN\316\242\\\no8Al\277\374K \033g\233O\266jJ\227\032\346a1\224\307\256\325\021L;IK\302\032'MY%?\263\217S\034\314~h\020\360\023F\215\305.yw\306\372I\3411\307\210\205\001A0\262\333N@\256S\023\233\2313z\030\206\265T\330+\311+\273\004\346\"\013\036I\013\007\031\264U2\364\0036\266\336\215C\254\264E\331\023\376\305\200\313\026G\242\334*\013~\311\016\te\255:\013\250\007e\004wnk\345\263=\236/5|^:[\377\014\202}\302\235Lp\341\225\263\300\244JCmT\036\231\332\255\247\202.\005\314\023m\341\000\r\231\370K\236R=\352\r\010\206\370&\211\323\026b\032\242\350\266~\253\202M\300<\000\256\004\225T\217\264U\232\337\315I\002A\317\033\251""\001l!K\264\016\232\217\224x\270\326\235P\336XN&Teqe0\346iL\220\346\346J\211\317\3173\324jK\222k\305\342\224a))W\207\024\377\263?\336\320\215\r`\306\313D\355V\214%\202K#?\203I\371\252~\003\347\337|l\370-\020\360\177R64\367\021\222\004E8O\227\311\023\315\350r\276\242\246XZ\304\304\353\247\201l\345A\206\242\241\204T\013*\225\2332\203Kc\270(6\202\350@\355h\240\213\232\271\263~\261N\037\370\241\236X^\233\271\214\tR\034\363D\216\221\031#g\014\262e\240U#\r\200\360\n\302\240\362g\031\206\023\245\362\341e\225\360Y\270\263\271\030y\233\322\334>\332(\307\214\214\n0\013\354\270\003\310\034\277\377r\301\314I\337k|[\326\303\n\023\311\377\304\234\0163[\243+\034! a\201MH\241X\275\267\221\343#-\035\203\342\003\026\003\357\0009\305]\332\2061\343\263\206\215cQ\355\374\006SB\021#.\247=NEVo\247\271\233E\240\n\306\302\217#\243\346mY\311S6r\310\372\214\010\021F\223l\014\304(S\020\020b\334w\2434N\315h*\375\301g4kh\251\017p\316\022\366[~\347'\250\300\211\314R\014\037\274\246c\021\032";
    PyObject *data = __Pyx_DecompressString(cstring, 3855, 3);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) == 2 /* compression: bz2 (3963 bytes) */
const char* const cstring = "BZh91AY&SY\207\344\"\201\000\003\033\377\377\377\377\377\377\377\377\377\376\377\377\377\377\277\377\377\376\300@@@@@@@@@@@@\000@\000`\020\233\234\372\032;\331\275\356\367o{:\275\032\256\232#0h\242\265\240QMhPPi\240c\200\007\267l%\010\221O\321Q\2434M<&S\311<2\233i\210\312\247\251\352=\031O\325=G\224\364\006\232z\231M\352L\215=F\3216SOS`\232\233M\004l\240\222C@\232\020\010\215\n~T\375S\322\032=A\352d\003A\2404\r4\320\000\014\203M\001\240\320m#C\324\000\323A\024\324\231\037\245\032bi\240\001\3522bi\221\210\320\0004\001\241\220bhi\2204\310\000\014\200\002BD)<\224l\3254\364\207\244\3654\315 0&\215\003@\000\000\000\310\014\2324h\320\000h\001\246\236\242\014\230\0010\t\211\223&\000\004\300\230\t\211\202`\000\000\000\004\310\300L\010a\000$H&\202\032\002dh\023AS\375\001D\336\241\244h\315M\000\003 \304\304\031\014\203 \r\032h\001\246\205:\233\032\003\021`*\261\244\030&L\245C5\243\353\325\331\316\347\035>\311\242\023I\023\3564M[\242J*tGL\250W\371c\245\322\022\302\205\245q\230\355\005\n\341X\210[LM\241\214\030\322M\203`\330\233cm&V\326\344@L\256V?\341\333\306\016\220Ha\322\3176H\001\231\007P\016\203%BF\244\211\030\201\302\302\322\205\235e\327u)*\214\3110\302k\211\216k\253\032\030\005\222,\025\260\014\033!\303S3\330juFU\2702\n\nh\324\322\333\336\342\224e\010Cz\254\353\004B\261(4\311\270\372\206h\321\t36\036\263\261!%\201\216\204\330{\212\305\305M\n\203S,\202\301t\035\332\004\307N\257\"\013\t\362\tT,V\206\013P\235\010\366\257\3212x\316\275\007\213\262\371s\321\210\231\361$\254N\221)\2707xE?I\217Y~\260_\330^\250\376\235!7f[\t$\216\213\r\271\320\177\255r\347\341<\217\342\371\360\201\231)\266\322\317wD\010\300\250\234Q\025\307K\002\224\274\277\356/\017\243t\275OZ\355\035\330\244\204\367\224!\341d\2656\n\316\250-\333\276k\2131m\300\263\024\222\302^~A\024\256\014)\245OI\344\321\244wy\225\002L\305\277\305\022\273\355\267\224,\246\333\022\270\237h\245RVrs\2200\241\250\232D\"I\246\210\304 I\020\205\344\022\234z\343\211\263\336\221\207\260\330\243Z\306\221\020\t\002Il=\337,)\002\275W""\031\021\005,\270\243\001E\320AB\024\021)\2558\020\361ql\263^\353\n\326\301L\3613\023\3050T\231\356f\307z\224\241\231tT\3716\247\320\246]\355\365\316\236a/+Km8p\363q\217}\237\227\\\251\302\335<o\343\235\254m_8\344n\277\376\333\3319\264\324U\323^\355*\212\300\n\333\252/wu\267\206\"!\340\253\321Q\2478g\205\202* ,!\236a6\331@\030\261\005\014^\271$_O\245\247u\265\237;\310\013\311\331\264\222\344w@\033\273.\333\275\323\"\240\242C{*\321\314(\333}w\275\353\\2\007\244\316\377\336\366\337\204\250\010\252\251TT\303\273\273\312!\223\250y\264>\005\350Zv\353\331\232{\312\232\237\001\314\246Sw6a\347;B\351\274\271Z\002'\221@S\307\232?se\373\217\301\327(\322\344]3\372.\304\376\333\010^\265\354\231\355\363\315\027\206\305\336\271\035\343\351\375\210\030C\345J\322gS%&\003\311\317\367\266\326_\261~\324$\265r\006\360\376.\205\333;{\304\247+sE\313\327\013)\013\345\264\3378x\t|\301\271\323\\3\244\254\312\3529\370?\262%_=\373\274\352I\204\316f\020h\324\237\223Z\035\200\326\331\023\374\310(\300\250\242\024h:\227\332h\227}\265\346\323\310\301\312<\354\251\306\221\353<O\323\252\020v\222\366\270\031\242\271\211oY\3020y\202^\270&)\022\031=\031\022\235\"?\321\306\016\351\354M\025\223\227\267\362e%\023\237\233\257\334?s\216\335\253\316LxB\300\3304\036\334E\3764\300A\227\211\363\254\246aui\177\006i\n\326\026bQ\354\351[\314q\005\263\202\320,\307c\266\336'*\000@=\307 \344\260Kq\364\274\236m^\007\006z}-\372.{%\347\255\212S\272\324\276\362{Mz\365\311\261uu\317^s\261a\304\nE*Yw\306\021\0131\317WV\323\302\202~e\351\320\244\316U0\230n\205~\257g\365\316\276\250^\226\224\255+\346\004\366=R\321[\264k\314\236\026}\203H\006\336\334\252\240\330\253\222\225\355\027\326\r\025\2666\r\236\256r4x\240|\372vh\237{\374\371[\262\356\277\352\325\030\230\355\323\207\016r\247G}:n\002\276C\326\027)\r\236\215'Q\213\020\270\007\225L\002\335\313\021\004\022\240D\311\0206'\020@\2032\333\351\353/\005\215c\317\030&\315\241\020\311\263 \3124\335:\244\023Dh\224W\307\320w\240\350\357\370\2140\332d'\264\231\030Gc\212\206\2144!\3307""w\317\004\372\255\023\215\262\327I\243\207C\274Ui\036\252\266\307\200[\207\303 %\341\355q$\365(\360^\376\240\362\370L\227\262Q\023\"\205\023c$\253\273\333\273\223\001\314\226\226\215\223\314\327\325\302G\261[\273\026\335H\357\251\206\301\327g\0134\205d\"\305}!\004\2464V\nhC\004\236\005K\235\260>+F\365\260\337\263\263\273H\257\227\224;'\263TkB\035b\304\273\331r6\215&=_\203\243\022\235ah\336\343\004e\306t\2033\351\314Ygh%R\324\221*\010\256\267\312\257h\205!\206\335\253\246\230|\017O\207\273\312V\233\0030\371=\270\233\347E\270A\260\327\260\226\035\211\023:\351\004\200\026\332\351\271!\032\230\230\371h\241\305\372\356\344TWk\303\231q\313\027U\226\247\361\0225\364K\255\031\326\341\360\356#\\\305\264A+\227T\313\370L\241\2334\371\307\307)F\220}b\375\003\306\305\366l\325\3143\210=\241\nd\304\362!h\031\303I\3512\345%m\\\223\n$\243\027\352\216\331ly\345'\017or+\014\333\0365ld\017\024F\364\333\262n\363%XHj!\310\031\275m\352\312\263\002X\215l\220\214\373}<\251\343\223]\230.\301\221q\243\321\375\367\265\306\267\003\014\256\333h\337%\255\014)+\336\270\3468:8U$\014\331\262\222\362N\373\330\226\337\364\303\315\230|?\177\177\246\200\3405\307\034\202Co\\!\312Vo<\267\005e]\222\231\260\212\323Vei[Ik)+\233\234\236\025\274\231*\273\311\017\033\213cBR\225\244O\006\330B\006\030\324\213\331\230\341\335T\210\010\335\242\t\027P\321$\331[\005%\203~\203\330\220\244f60-\251\302\377\241tFF\22379\320\262\277%?\365w\227k\203\006\"\244nK\337\312>H\001\225*\267\257\257\276\254\242\r\200\313\363\032\013\222\353\373\032\367\036d\230\235\005\250\233\264r\246\321\305F2K\303d\237\212[\321P[T\021\262\362\264\367\2058\352\220Up\311+\227/\212?\214\275\347\306%\036\366\376\n \245\033 d:~t\310\363\n\377?\177y\255-\334Z'\341\273\334`#\320\220Uq\001S,&%<\315~\022\320Gt2I\230\253%d\274rF&q\020\202:\3510\227j\221\201\003X6\335&\246\223^\242\310]GT8\007\243\312\253\001_\211\323O_\243\0175\326\331\023\301\370\026s){\200\035 \r\214\"xL\334b\001\261\215\243KSa\321\200\007z:\203\260\347\235\306\021\342)\001\r$\022\306C""\034\211\020-\275RdbV\233dVa\311\003\0064\216f{\305Im\026\350\233:\243\023L\033\227]#t\234S\266P2\236\251)2\024kW\325\204g\277\244\341\247:\260U\265\271$\331\326\226\035b\226\235}\232\2052\325\024a\002aI\2478!\223\355\210r\207e\223\265\233m\266\333m\266'\242P\306<r;fw=i\0221W\341\322\241\316\351\212\322\374\t\031\r\240\210\030b\230\277\016\010\t$\320ldob-lQ\"\322\213I,\233l\300x\316\003\311E\006\000}\226pk\006S\034\\/s\t]\004R\3213c@\257\213sG1l\002\3245`\324\214o\010\234\263\337\222H\361%Bx\021D\252\273E\314P]M\261\226\375\353J\t61\216\365v\275{&r2\030\330\322`o\215\373\331\202R\243\002\275e0\332\316\210\263GG\301\253\312n\360Gww\217N\361|\273\363JS}\206\323\210,\311\260\333\277\230\005\022w\300\016\265sBb\276v\333p_gW\255\331\345O\n\263,\205\001\021\017~s\272\217\253\332\223!-\t2\215\230@\205W\232\013\347\330j36\235(\242]+j^=\267w43\205\032\016\271\002\325\000r\313]P\270\222aF\214~^dcR1\025\3423\244\230\270Qg=\322\354\"\364\215\034u\025\203s\273\025\361\330d\214W\006A\234)d9\347BL\006\261\226\241\314\352]9\353(\336\314\244\202g3\335\031\000\316)j\250\224\362\027P.Q\321\234\320\007\236KqcU\312\351vY\232\300s\357\323\245,\224\237\233\236\273\366\352Z\351w\003\2013>\000n\221\232\r\033tk\025\013\201U(K\257 6%\251q\272\010\252\362N4;\224W\020\300(\302b\255\030\231H#c\030\307\352\347B\327$\354CAPpP6\341\022Q\262\232\247Te\270z\231\255\305ye4\232\203E\361\t\315#(\204\014\362\t\246\r$O8K9\\ \260\201\332\262\303X\t!4y\234\362:+PI\233H\020\304Z\000\257h\001y\245mP..m\00415f\202R\230\254\003\371\373\026\372\261\276\216\201]\312\270W\341\027\251\030\225@\322\2329\263O`\014\006\000\367c@\2406\r\261\315\210\332_X\253\333\233\3214\266BY\006r\221\201x\214\322=\341\0020\0226x\265\037\222.\273\033\215Z}l2-\356?\t\210\333Ce\226\323m\316:S\340T\314\350`\205\022\207\030\251\337\277s{\211\273\273\232\255\336h3i\236V\243f\224b\212\"g|X\026\332\\\027\263I\024J\303\323\010\345cT\241\215\332\260\360\300<\336,iy\370\0362%\351q?lD\025D\211\224\254\347:\272Y%n\227\2576\341\204 \232\214\342n^""\252AW\035\334\270$Z\215$Q\360\016%<\223\2367p\346\351\352X^.!P\362\264\307v\254\250\025\023\003\336c\351:\246\365<\236\234K\252s\2247BWFwV\205\225d7k<Z(&n?\025\233\227~T\273\030\224\\U\221\311)\205\312\223u\\r\266\215\357\302\n\025\021\004\2352\273\025\030\031\326\216k\261W\203\n\213\330e\022,\351#\230\271\241]\214V\227I\\\3052R\233N2\362\261E\205\325^;\200\341\354\216\207\221E\323\242p\335\366\220\034\361\326d\264H\331\250\030\256\254\354g\226\226\221T\352u\355\351\237\034!R\343\313,Ri\217-\000\345!\215\216\323\023\030B\322\\0\031jjE.4T\332\331\251\323A^I#\301.\236\236:N\213\367J\371d\033\022\253\230\250\224\344\n\220\341\251^\315,(\244\351\004\203\217{X\352k\\\310\243\030n\032(\262\203`\330\316\204\350\206\027\004\205\346\222\2464\253&\013&\245\216WD\352\025s\214\014[\230I\2443\262\021\220\022\212\\\207gu\236\374b\267o\003C\036\375\000+\354J\355c\n\214`\205\332&D0\330@\210\271~x\260\0314\374\233^tpR$`\245\004\302\223,\311\010\277\343\203B\273\005\215\234q\021\307\321\350\016\235\351u\353:\304v\340|4A\320y\002\272\373\222\336\205\324\320v\3504\014\364t\233\304\3550\307\365\340\340\267\233\327EF1[\247OR\355[7&)\016\204\022\t\014\226\277446\220\300\356@`\207zW\255Yr\314\316ekZ\206-\323|\202\301\314\r\0322\024'\014\353\024C<BB\310\315\214\017b\202\214\321\307zR\233\333\312\344\266\000<\0010i6\265p\030\273N\031v\263'@\341\372\243\341<Mr\376\326\363\235>\231\0336M\334\246\233(~\207\334\216zi\t\215\003lcA\313\371\206s\275\020\221j\203_\"\200(&\233N@\211g$t/\331\243M\2461\233\001\272\006\347\357\276\"7\211#\312\374J\016\177\336\343\024O\032\243'\345\334Vs\317\361\331)z\3643\267N\227{).:\207\240\r\325\300>\341OK\307\036s\300U\242\273\371\006\037U0\323b\001S\\^\236Q\265Um\252\177m\341\233\313pF\210\377M\355W\0246\243\365\252p3\013\230\214\371QN\006OF\226\226rqSb\002p\233#16w'-m\234*\314\023\245\235\231\254!\302x\270\030r/\037\3232\270\222\\\361\031\033\330C:)91\332\0132\331\240u\375Oa\036\363\013@\026;\"\263\004\251\263\234\002XTB\270/\027\235\332\277\265 1\3251\216\263""\010\233KeC\215P|\226>rN\261\372\3129\017\371\235\006\030I!\361\216\034\343\251\363\031\344\255\\\205F}\3748\374\"\270\254\252\310\306=\361\321\306\276*U8\350<\250\260\333pUQ\223\314{\216\362\2255*\236a\326\361\027\026\240\254\271\237\212\274t^\334r\321\321\255c\325\003\332\347t\212+L\311\026\252\352G\024l\307\014qW\337\270\214\367\334\224\337\214\223\273;\235\342=\377CGie4\333\302\223\351\307\334\263\2754\317\257$\255.\236%b;Y\033\261\333\373um\305]ttz\365\322\265Oo\203\224\264\035hn\317\177\0076\377[\001r\351\035=\276+\275\263\370\313\275\365\335\357\007\034\3672l\027&I\244\355Oyy\231^\377H\321\372\2710\374\230\272\334\274\024\345(\313\025\026e\214\357\354,\003\n*\252\256H\210\235Q03\345o!\254^^\313\331\350q\353rE:\242h\261`A\352+\255!\341l\214\224*\302C(\016\241\3479D\243\001\203w.\360\227\2521.\333u\202\250F\360\263\225hc!\014!\005g)\224.\231s\000\207\004\334\003\223=W\034/\t\205\213\335\212\270\001\232\030\212\026\202\300\202\352T\231\t\300\224\330J\026ra(R\355f\250\243\033\311\225\206\314\214\205\205\341\220\307\204,\352\211\305\325\237\003\311\377\342\356H\247\n\022\020\374\204P ";
    PyObject *data = __Pyx_DecompressString(cstring, 3963, 2);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) != 0 /* compression: zlib (3853 bytes) */
const char* const cstring = "x\332\3258\313v\333\310\261\326\214&#{dK\224\345\3078\311\030\264=\226=\017\315\225-\217=v\356I \022\222\351\341K$%\315$7\267\017\0104%\330$A\002\240\036v|r\227Zb\211%\226Xb\311\245\226^r\311\245>\301\237\220\252\306\223$\224\207\223snr\216Dt=\272\252\272\252\272\272\272\227\337<\375\315\352\333\247\034\373\\\310\210\355\266jp\222FE\203r\231\002\227\313r\rMmqo\324\372\313\267\231\246\252+\355]\216\227e-\327n\250\001\234)\010\373\264m\214\203{ \2146#l.\233Q\001#\031\212\332\346:\242&\266\250A5=\303\224\311\234\030N\234@\370\222z\232\006P\363\210S\333\360S\311\026xR*\327r\245\"\311e\t\237\371\221\324r\005\241\264U\343\024\235\323{\235\216\252\241\030\245\315\225\217\366\251V\327\227\327E\245\t\030C\345xI\242\035\203\213\014\212\223^\265\325\003\000v)\307\324s\337ro\336F\3645\245-\203_\"DFmu\232\260\024Nls\"\010\333\247\234\024-T\243\335\036\325\2158;\243\305\020\236\2677\313\021*\253\350\3228\333\013\025V\322\3525\rE\022\343\362\312\252np\025*\312\023(\211\202-c\330*m\2173\356h\212\021c\333\354Q\355(\214q\204\257\320\227`O\242\313*TW\233\260\354\244I\021\205\352\372$\241\242\366\342\312%\310\026\312r%\206\013\262\221E\343,\274\237$\t\344\221`\355R#0\206\0335\025)\233eN4\014M\251\203Q\372(I\243-\025\304\t\031a\034\317\342\373\r\327V\343a\007\224Xo*\372\036\215\271\273IEXrB\014\273\314\347\361\014\320(\230A\201[\363\302\010\262Y\226\215\370=d\322!\252\211\034:X\330T%\2619j8\242\325\016\262\262\254\002Lh\024\214\340gO\214\205\223S5\226}\205\014\017\316\021\0169Q\333\355\265\300\347z^\321\r\332\346<\301E\364\217\261'B\216\034\031{\340\003\330\2062m*u\252A `\307\352`\255\004[\036\231`S\n\345oW\237\254\302\256\221a!\230[\270k\353RS\324u\252sj\203\253\367\0240\n\314;\352P}\231\3135\270#\265\307\265\251\267\206\016\360\305'\030{`\t.\014\006\334\022+d\"\256\220\300t(@K\234\254h\324\333\2360{]l\352ty\253\r(u\267\255\274\006\231P\3420\301Q\333\033\254\210o/\374V\224e\002r\250\254\350\020MJ\333\336o\257\3051\255\234R\337'\"\224\022]'\215\246\270\253sX=\225\026V\036\344\364jg\307""\257>\340\t\340\307!A\021\260\242\237a9\242F\271\216\246\326\201\377\210\353\261R)rRG\246\r\216)\222)\250\322\274:&r\313\235C\231k\200\267=G\313*\365t\262x!\371\210\303h\301\367\220\261-\217\031\213\211\362\037b\352\201D\324\216\244\312\364?\301R\330\355F\357\337\333\247\232\334\022\211\324\"\024\253%n\212\277\351X\234\201\023\376\337\254E\313\210\336\021\245\177?S\033M\365\000\353\275\241\251\315]Ii+\206\"\032\300!C[\261\247\350^\245\220\233^}\204.\n\205\2019\373\212Le\337\356\357\244\226\"\243\304n\207\003\213\240L\302\251\324Q\3332TH\030\251=\r\n\013\226\370#P\324\003Mm\215\213\201:m6\226\241@iD\301\032-y\255[\2355\037 \337\360\252\244\010\315\017+\306~uk\200\215\035Ez\325\004\2070\txf\021\326\215}\260\010\226Q\377\334l\"y\307\367\007KQ\344\017\236\372\022\2166\322\222\010\236\374\204\036~\260\234\236\374\341~\274gh\024\263Z\334\305s\365~\317h|\373$8~\203\3572!\220\257=\211\022\211\311!$F\201\203\017kP\002\215\265T!\004-\013\301\244\301\234\t:\247\321.{\004\232\3249A\237\320<\306\301\364\373\2703\305\375\0259\313\242\364*\254[\001nD\250\227A\320]\260;\006\374'\251a\330\004\035\200\027\331e\200\r\353\320\3343\3770\310\327\202#\257\263\363\306\254\261$\335\016\203\344\260Ug`\330\3671\010\275\r\035\306>\301\336,\206\301V\222@c\027C\261.2\204\3319\035N\302\362\002\372X\2062\204\227\262A\313\306P\254\267\034\307\261\362\303\206\035\350\364A\213(\307!i?\202Pc\004A.\217\"\016\360\222\300@\326\253\306r(D\371\016\321\350.i\351\273z\010\204Z\021\210\344x=\237?d\227\201\310\361qD\250%@jxe`\030\010'a\335m\350K\304x\215-\336S\312\270\035\303ABV\304H\223\251\021\022QhGS\366\221*\213\206\230[\333\206\353fF\250V\311z~\253\372\234l\344Kk|~\002]\026*\325\\\265&\024k1\322\363\255\r\241\226_\213a\362\245\014\237';\225\\M\210a\013;d-W\314\3060p\317\315\n\005~\004W\021\362\374OB\226\224*Y\241\222+n\214\220\n\245\232@\370Z\251\220\313L\342+\002\237\235\304\216[\361{\241R\"k|U`\274U\241\230%\353B1#\204P\256\230\317\025c`\231d\252[\205\020\256\3466\212|>6\273Z\312\20355\017\263\223\361""\315\213\364\002j\215\317\2029\3252\021*\225\000\005\236\000\217\370P\246T(\223\352\016_\366\341l%\267-T\036\214@\017}h\235\257\201s#I\353B-\363\234\360\331@?\213\224?\336\020\212Be\204;W\334&\202\220)N\242H\265\306\203\277\"\274\027E\240Fp\340\304\021&6\271T\036C\346\205\342\030\246\\)\325\306P\233\345\311\211\225l\226l\347Jq\243\375\314\211a\330SI\020\360\000\216\273\274\"d\266cC\237!W{Nr\205BH(\020~\255T\211[\305p\343\213D$z\251\202a\334L\"L`G\226\305b\357?\345\214\240k\225\237\211\360S&\216+V\022\361\230j\301p\213\331\347C\265B,\366\000d\205|\004D\363\001(\226H\215\337\210\340\230\213\020*f\267a-\230\211y!t#\020\252?\0273\001T-\371\243\235\212_\021@C\374\"\035\037OV\2471\352D\201\362_\265\004MS5\177\214q\003\270\255\216\302!}K\247\032\203q\366!\374g\341BN\212\364\320\250\320\006\013z\006\334\260\r\025\013\335\004\005e\275\2046\227*g\320 T\245\374\266\220\235$\223\314s\276\270!$\020\316\220w\206\254L\251X\0242\265\244Y\001\t\322iK\250\326\316\"V\313\245bu\314\220\254\260\235\313\010\254\350m\363\3711Z\256\352O\036\267\005\224\360k\371\\\365\3718\241\260\225\257\3452|5\321\314\210\370\242\224+\216\322*\302\213\004=\025H|!I\224GHv\024n\230\035>\007&\374\224\033\363\305V\021\366>\304c-\357\271\241\\%\271\265pT.E@-S\016\206[\3312\374\341\371\347\177&\0234\"L\344\346\016\025_U\251\341uV\361\006\016Nt<\324E\357\tNm\210{\242\"\352GmI\201NU\305\243]iS\035\033\035\326\216\267D\375U\035\3467\325\335\2605\253\367\032\rH\346#P\327\244m\t\257\016\232\177}\220\240\227\246p\037\"0]\2428\221\265p\201r\274paS\305\304b\327\346wnx\217\203=\0206v\376\000\337\253\332\273\304\323\307p\252&\261\307(\354\002\360\037\r`\337\216\241\311\264\336\333e\327\277\250%\204\013 ^\n\214=RW\014]\326\r\370ck\220Y\307\245\355+\022eCm_\202\026\006V\243\3124z:\014\372\332\350\312\316\236n\360\356I\374\273'\\\"Z\242AH\243\007\236 \361\356~\244\367\034m;c\035\347H\263\311\004x\201${\n\\1\331\017.N\031\177\351\032}L\032y\257\031y\022QZ-\3263\021\302\232X\357\227\275\261\306zZ\242\350$\014>\364\207-\2355""\270L\364h\253\333\304\227\306\221F\027\"\260k\3545\301EM\257\335\365~\t8\036\302\014s\203\253\035|\230,\246\220\264T\271\327\204e\266`\334\026[0j\203\351\370\317\034\010W4o1\320L\202\006\370E&\370\354\213Mv\311\353\310\235W\364\010\226#\323\303\216\332\211^*\302F;\354\261\303\366:\336YGMu\274\263\004O\007\217\032uQ\247\301\030\363\366\254g\2443\036A\002\264\377%\324\253\372\035\250\372~\200q\270\317^SH\020\007\370`X\374O\200\002q\360\353\317\352\366\304\246\347\257\321{@p\005\320D\005\266\312\304S\323\330sNXJ\310DQ\t\021\364\020\207\336]\"\270F\2047\010\357\362\340g4\332\240\341SIti\030\277@\214\334\035F\036W\360AE\203Pjm\215x@\267\323\306\373\274\277=G\356\030\254\322\371\371\022]5\306\357\007\000\313\264!B\206\306*#I\250\222\272\362\232\352M\035\212\204\256I\360\307\314\305o\267\203\277\201\0058\204\002\241k]\366a\373J\357\201S\t1`\013\303\257\002\271\3313\300U\206\212\205\241\rY\332\243\354G\207\020\310\020y\210\377\201\250\265\361\037\337+\016$\330\241l\177\035@\245\326h\343\0006\241\\>*yO\023_\335\213\206\367\317\261\032J|F}0}\313N\277\237=w~\321|h\212\203k\367\006\367\236\366\323\303\231\224yup\355\033\207w6\2073\263\307\374pv\301\274o\255\014gnZ\373\366\357\235\256;u:3o\316Z\274U\263\257\330-\367i\037\370>;^=\326\3154\360\016\026@\252\2752\230\376\335\311\324\360W\217\335\356\373\305s\347g\201\334\035\316^1+\246n\201\212\371\301<r\255\332=g\335\275\345\256\367\357\235\334:\341\207\267\277rn9\202\233J\220\367\177<\3142?6W\314\2545c\177l\257\330\033\316CG\034\316^<\005#\005\363\212)\242hk\325\002=\363\346'\346\246I\255\225\200F\255\207\226h\031\366\023\234\343\3048\200\335\206\365\334\260\356\332\027\234\264\263\n4\230\362\343\340\332\222\375\326\345\375\3612\314\321\335\264\273\352\036\366\365\223\364\t\030\363\376\372\271\363\227\315\273\326\324pf\356\370O\250hn0\227\266S\366\035\233:+\316\206\373\270\237\356\257\340:\036\036Kf*Z\007\362\037YS\270\034\\\343\033\353\266\2659\234\235;6\314\225\341\374\242\271a}\017B\322\303K\234=e/\332,\002\005+\025\260\377""\331\222\354T\022\373}'\345\334Aw\200g\301\0140\020\342\232\n\274\334Mp(:\033\027\374\272\237B%9s3Pb`:$(a\230\337\332\267\301\252\005_\356\351\374es\025\374~\207\331u\365\256-\332\206\003\223\256\017\256\337s\246\234EH!\237#=\\\2741\270\201*W\002\321\251\0330\217\332\317\234PY\352\212Y\263\026\255m\260n\366\006\004s\305\312\332S\030\254_\230\242\267\244\371\301\\\2208\274Su\247\022<\034,\343\300\224,@_\006\271\2209\347\255\233\314\240?\366\227NRc,\223+\2357/bXp\032\372r\0363\370\211\371\010\302\035\013\344\257\201\0053\331\013\325\226\263\344~\216\302O\226\336%'\260\037\2220\262I,\227\030\313\227\366\246-;_:\306\031\233\013\354\377\374\006\356\253T\244\375\256\373\221{\267?s\362\321Iz\230\276\347\\t\317\320p\341\334\371K\307?\231\236\351wl\336\2569W\234:LN\237N\362CZ\037wa\322\"f\373\035\310\212\213\227\206\263\213\203\305\300\300\256\373\251\233\230\\\220Q\033\346c\010\341#\010 \000\353\346-\223G\245\236<0b\216E\r\267\315m\330\314Y\347\023g\323\221\274\255\237\260e|#R\346\034l\330i\233?\313\301\227\346\230}Kv\327\231\306J\342\246=NmD\340\305\224y\331\344\375\334\230C\"l\345\217\254{\366\222\263\340\374\340n\364\037\200\333\027~mU\254\267N\316\255\300\022\027>\267\026\254\037\240\346<\200\224^\370\245\365\300\372\203s\336]pW\0026\001\200\307\375\333\375J\037\230\323\366B\030M\337b\301\271\354\274p\305\004\273\337\377\302\253\217=3c\262\312\364\251y\000\305\252g\217nG\334\004uk\212\325\202\353\237\307\235!8W\035\315\275\n\305I;I\r9\277r%\371g!\232\264\016\345Bv\357\270\264\277\322\3378YI\346\207\274\237\367\275\363\251\325\035L\377\312bk\212\341\374]y\274\217\261]4_X\260\302K\307-\266s?\266\037\261-\267\343\376\330\357\236\234\1777\031\333\301\355\037\260\372\314\341\364\301\364\327\203\257\327O\304\341M\016\027\367\264\357o\242\036\024\374i\213\217\247\303g\336i\002q\366\212\314\026\254\343%\2346\367\377\313\375\254\277r\212\204\257\241\010g\335O\334Mw\027B\371\315C\367Y_L<\245\276\033|\367\342]\032=3qF%\2737\235(\346\216]\034<\310\234\374\360\216\177\377\305\271""\363\277\304\3717\255\356\351\014\330\201\333\352\021\004\265\213\205\346\265\225\202\243fj8\177\3154\254\307l9\263\327\321\221(\365\t\356\230\200\353>\220\346oXK\3665X#\253c@<\005\037\263mt\351X\364\317\216\231\213\203\213_\240\333/\036\377\217\225\362\222\351\031\236}\303\331\224y\323\236\0057LyC<\267\007\363^\215~\004\236\372\376$\261NM\r\247/\034\177\215\001\371\357\376&K\231\350\260\316\260\343\342\366]\373\245{\351\014\207BU\364\362>\350\020\222\331\240\202 \333}X\334\025G\355\377\341\335\3240\326U\374U\321#)\t\313\301\014\034\223\003\325\333\222\355/\023\017\276\3014T\007\177]a\351\371\027\354K\310%\247\353\035\025\177O\"%\031v\323z=\270\377l\360\254\360ns\310\301\221\216]\306M\214\355\027\361\037\226U\303\031\360\340\251\177\212\211A\322|\t\2114\177\025N\265kp*@\322\\C\223\207\1777\327o\330\300\347\372\026\2747\377\205\365\312\271\346N\271\377\030\327?\233\246\237\035\177o.\302\374\340\250y\351^\200\306\364\356\0226\027\247\240\030T^1\t\004\002\033\rl}\276r_\364_\275K\275K'\265\033\214\267\350\336\206\023\021\206\377k?\204\364\202\301\037Y'\223\244b\310T,\340\341\230\224\030\177\001 Q\207:";
    PyObject *data = __Pyx_DecompressString(cstring, 3853, 1);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #else /* compression: none (10090 bytes) */
const char* const bytes = ".{:<4}: {:<4}\nCannot create CM ID from {obj}Closing AddrInfoClosing CMEventClosing CMEventChannelClosing CMIDConnection parametersCreated a CMEventCreated a CMEventChannelCurrently only RDMA_OPTION_ID_ACK_TIMEOUT is supported in Pyverbs.Failed to Accept ConnectionFailed to Acknowledge Event - {}Failed to Bind IDFailed to Complete an active connection requestFailed to ConnectFailed to Create QPFailed to DisconnectFailed to Join multicastFailed to Post ReadFailed to Post ReceiveFailed to Post SendFailed to Post WriteFailed to Query AddrInfoFailed to Reject ConnectionFailed to Resolve AddrInfoFailed to Resolve AddressFailed to Resolve RouteFailed to close CMIDFailed to create CMEventFailed to create CMEventChannelFailed to create CM IDFailed to get Address InfoFailed to get QP attributesFailed to get remote ECEFailed to get request, no connection establishedFailed to leave multicastFailed to query QPFailed to retrieve receive completionFailed to retrieve send completionFailed to set local ECEFailed to set optionJoin to multicast must have AddrInfo or JoinMCAttrEx argumentsListen FailedNote that Cython is deliberately stricter than PEP-484 and rejects subclasses of builtin types. If you need to pass subclasses then set the 'annotation_typing' directive to False.Unrecognized object type{}: {}\n?add_notedisableenableenum class ibv_access_flags not importable from pyverbs.libibverbs_enums. You are probably using a cpdef enum declared in a .pxd file that does not have a .py  or .pyx file.enum class ibv_send_flags not importable from pyverbs.libibverbs_enums. You are probably using a cpdef enum declared in a .pxd file that does not have a .py  or .pyx file.enum class ibv_wc_opcode not importable from pyverbs.libibverbs_enums. You are probably using a cpdef enum declared in a .pxd file that does not have a .py  or .pyx file.enum class ibv_wc_status not importable from pyverbs.libibverbs_enums. You are probably using a cpdef enum declared in a .pxd file that does not have a .""py  or .pyx file.enum class rdma_cm_event_type not importable from pyverbs.librdmacm_enums. You are probably using a cpdef enum declared in a .pxd file that does not have a .py  or .pyx file.enum class rdma_port_space not importable from pyverbs.librdmacm_enums. You are probably using a cpdef enum declared in a .pxd file that does not have a .py  or .pyx file.flow controlgcinitiator depthisenabledlisten ID not providedpyverbs/cmid.pyxqp numberresponder resourcesretry countrnr retry countself.addr_info cannot be converted to a Python object for picklingself.conn_param cannot be converted to a Python object for picklingself.event cannot be converted to a Python object for picklingself.event_channel cannot be converted to a Python object for picklingself.id cannot be converted to a Python object for picklingself.join_mc_attr_ex cannot be converted to a Python object for picklingself.ud_param cannot be converted to a Python object for pickling(tree fragment)utf-8AddrInfoAddrInfo.__reduce_cython__AddrInfo.__setstate_cython__AddrInfo.closeAddrInfo.get_addrinfoCMEventCMEventChannelCMEventChannel.__reduce_cython__CMEventChannel.__setstate_cython__CMEventChannel.closeCMEvent.__reduce_cython__CMEvent.__setstate_cython__CMEvent.ack_cm_eventCMEvent.closeCMEvent.event_strCMIDCMID.__reduce_cython__CMID.__setstate_cython__CMID.acceptCMID.bind_addrCMID.closeCMID.connectCMID.create_qpCMID.disconnectCMID.establishCMID.get_recv_compCMID.get_remote_eceCMID.get_requestCMID.get_send_compCMID.init_qp_attrCMID.join_multicastCMID.leave_multicastCMID.listenCMID.post_readCMID.post_recvCMID.post_sendCMID.post_ud_sendCMID.post_writeCMID.query_addrinfoCMID.query_qpCMID.reg_msgsCMID.reg_readCMID.reg_writeCMID.rejectCMID.resolve_addrCMID.resolve_addrinfoCMID.resolve_routeCMID.set_local_eceCMID.set_optionConnParamConnParam.__reduce_cython__ConnParam.__setstate_cython__ConnParam.set_private_dataIBV_ACCESS_FLUSH_GLOBALIBV_ACCESS_FLUSH_PERSISTENTIBV_ACCESS_HUGETLBIBV_ACCESS_LOCAL_WRITEIBV_ACCESS_MW_BI""NDIBV_ACCESS_ON_DEMANDIBV_ACCESS_RELAXED_ORDERINGIBV_ACCESS_REMOTE_ATOMICIBV_ACCESS_REMOTE_READIBV_ACCESS_REMOTE_WRITEIBV_ACCESS_ZERO_BASEDIBV_SEND_FENCEIBV_SEND_INLINEIBV_SEND_IP_CSUMIBV_SEND_SIGNALEDIBV_SEND_SOLICITEDIBV_WC_ATOMIC_WRITEIBV_WC_BAD_RESP_ERRIBV_WC_BIND_MWIBV_WC_COMP_SWAPIBV_WC_DRIVER2IBV_WC_DRIVER3IBV_WC_FATAL_ERRIBV_WC_FETCH_ADDIBV_WC_FLUSHIBV_WC_GENERAL_ERRIBV_WC_INV_EECN_ERRIBV_WC_INV_EEC_STATE_ERRIBV_WC_LOCAL_INVIBV_WC_LOC_ACCESS_ERRIBV_WC_LOC_EEC_OP_ERRIBV_WC_LOC_LEN_ERRIBV_WC_LOC_PROT_ERRIBV_WC_LOC_QP_OP_ERRIBV_WC_LOC_RDD_VIOL_ERRIBV_WC_MW_BIND_ERRIBV_WC_RDMA_READIBV_WC_RDMA_WRITEIBV_WC_RECVIBV_WC_RECV_RDMA_WITH_IMMIBV_WC_REM_ABORT_ERRIBV_WC_REM_ACCESS_ERRIBV_WC_REM_INV_RD_REQ_ERRIBV_WC_REM_INV_REQ_ERRIBV_WC_REM_OP_ERRIBV_WC_RESP_TIMEOUT_ERRIBV_WC_RETRY_EXC_ERRIBV_WC_RNR_RETRY_EXC_ERRIBV_WC_SENDIBV_WC_SUCCESSIBV_WC_TM_ADDIBV_WC_TM_DELIBV_WC_TM_ERRIBV_WC_TM_NO_TAGIBV_WC_TM_RECVIBV_WC_TM_RNDV_INCOMPLETEIBV_WC_TM_SYNCIBV_WC_TSOIBV_WC_WR_FLUSH_ERRJoinMCAttrExJoinMCAttrEx.__reduce_cython__JoinMCAttrEx.__setstate_cython__PyverbsErrorPyverbsRDMAErrnoPyverbsRDMAErrorPyverbsUserError__Pyx_PyDict_NextRefRDMA_CM_EVENT_ADDRINFO_ERRORRDMA_CM_EVENT_ADDRINFO_RESOLVEDRDMA_CM_EVENT_ADDR_CHANGERDMA_CM_EVENT_ADDR_ERRORRDMA_CM_EVENT_ADDR_RESOLVEDRDMA_CM_EVENT_CONNECT_ERRORRDMA_CM_EVENT_CONNECT_REQUESTRDMA_CM_EVENT_CONNECT_RESPONSERDMA_CM_EVENT_DEVICE_REMOVALRDMA_CM_EVENT_DISCONNECTEDRDMA_CM_EVENT_ESTABLISHEDRDMA_CM_EVENT_MULTICAST_ERRORRDMA_CM_EVENT_MULTICAST_JOINRDMA_CM_EVENT_REJECTEDRDMA_CM_EVENT_ROUTE_ERRORRDMA_CM_EVENT_ROUTE_RESOLVEDRDMA_CM_EVENT_TIMEWAIT_EXITRDMA_CM_EVENT_UNREACHABLERDMA_PS_IBRDMA_PS_IPOIBRDMA_PS_TCPRDMA_PS_UDPUDParamUDParam.__reduce_cython__UDParam.__setstate_cython__WeakSetacceptack_cm_eventaddaddraddressofahaiasyncio.coroutinesattrattr_maskbacklogbind_addrbufferbyte_lenc_charchannelcline_in_tracebackclosecm_eventcmidcomp_maskconnconnectcontextcreate_qpcreate_string_buffercreatorctypesdatadata_lendata_ptrdebugdepthdisconnectdlid_path_bitsds""tdst_addrdst_servicedst_srvceceencodeestablishevent_strevent_typeflagsflow_controlformat__func__get_addrinfoget_recv_compget_remote_eceget_requestget_send_comp__getstate__hintshints_ptriibv_access_flagsibv_send_flagsibv_wc_opcodeibv_wc_statusimm_data__init__init_attrinit_qp_attr_is_coroutineitemsjoin_flagsjoin_multicastlaileave_multicastlengthlevellistenlisten_id__main__mc_attrmc_join_attr__module__mr__name__nodenode_addrobjopcodeoptlenoptnameoptvalparampdpkey_indexpopport_spacepost_readpost_recvpost_sendpost_ud_sendpost_writeprivate_dataptrpyverbs.basepyverbs.cmidpyverbs.libibverbs_enumspyverbs.librdmacm_enumspyverbs.pyverbs_error__pyx_state__pyx_vtable__qp_attrqp_initqp_init_attrqp_numqp_state__qualname__query_addrinfoquery_qpraircrdma_cm_event_typerdma_port_space__reduce____reduce_cython____reduce_ex__reg_msgsreg_readreg_writerejectremote_addrresresolve_addrresolve_addrinforesolve_routeresourcesretretryrkeyrnr_retryrqpnselfserviceset_local_ece__set_name__set_optionset_private_datasetdefault__setstate____setstate_cython__sizeslslidsrcsrc_addrsrc_qpsrc_servicesrc_srvcsrqsrvcstatussuper__test__timeout_msto_connvaluevaluesvendor_errwarnwarningswcwc_flagsweakrefwr_idPyObject *(PyObject *)\000close_weakrefs\320\004\"\240!\360\014\000\t\024\2203\220a\320\027(\320(:\270!\330\010\021\220\026\320\027,\250A\250Q\330\010\014\210A\330\014\022\220)\2301\330\010\037\230v\240Z\250q\260\001\340\010\020\220\014\230A\230T\240\025\240m\260:\270Q\330\010\013\2104\210s\220!\330\014\022\320\022\"\240!\2401\320\004@\300\001\330\0347\260q\360\024\000\t\014\2104\210q\330\014\025\220R\220s\230!\330\010\020\320\020\"\240!\2404\240u\250F\260\"\260F\270(\300\"\300A\330#*\250\"\250E\260\021\330\010\013\2104\210s\220!\330\014\022\320\022\"\240!\2401\200A\330\010\020\220\003\2201\220D\230\010\240\003\2401\240G\2503\250a\330\014\r\340\010\014\210E\220\025\220a\220s\230!\2304\230q\330\014\020\220\005\220Q\220e\2301\340\010\014\210E\220\025\220e\2303\230a\230t\2408\2503\250a\250q\330\014\020\220""\005\220Q\220e\2304\230q\240\001\340\010\032\230&\240\n\250!\2504\250q\330\010\014\210K\320\027'\240}\260A\330\010\014\210K\320\027.\250a\250s\260!\2604\260x\270s\300!\3001\200A\360\030\000\t\023\220&\230\001\330\010\017\210|\2301\340\010\017\320\017!\240\021\240$\240e\2501\250G\2607\270!\2701\330\010\013\2103\210c\220\021\330\014\022\320\022\"\240!\2401\330\010\017\210y\230\001\200A\330\010\013\2104\210{\230#\230Q\330\014\017\210t\2201\330\020\024\220G\2306\240\021\240!\330\016 \240\001\240\024\240Q\330\010\014\210M\230\021\200A\330\010\013\2104\210\177\230c\240\021\330\014\017\210t\2201\330\020\024\220G\2306\240\021\240!\330\016)\250\021\250$\250a\330\014\020\320\020!\240\021\200A\360\014\000\t\021\320\020\"\240!\2404\240q\330\010\013\2104\210s\220!\330\014\022\320\022\"\240!\2401\330#*\250!\2504\250z\270\021\330\010\014\210I\220Q\200A\330\010\013\2104\210t\2203\220a\330\014\017\210t\2201\330\020\024\220G\2306\240\021\240!\330\014\017\210t\220?\240#\240Q\330\022\"\240!\2404\240q\340\020\023\2204\220s\230$\230c\240\021\330\026&\240a\240t\2501\330\020\030\320\030(\250\001\250\024\250Q\330\020\023\2204\220s\230!\330\024\032\320\032*\250!\2501\330\014\017\210t\2201\330\021\032\230$\230e\240;\250a\330\014\017\210t\2201\330\021\025\220T\230\024\230V\2401\330\014\032\230!\2301\230D\240\001\330\014\020\220\006\220a\200A\360\014\000\t\020\320\017\"\240!\2404\240u\250A\250S\260\001\330\010\013\2103\210c\220\021\330\014\022\320\022\"\240!\2401\200A\330\010\013\2104\210w\220c\230\021\330\014\023\2201\330\010\020\220\t\230\037\250\001\250\024\250]\270'\300\021\200A\330\010\013\2104\210w\220c\230\021\330\014\017\210t\2201\330\020\024\220G\2306\240\021\240!\330\014\020\220\r\230Q\330\014\020\220\t\230\021\200A\360\020\000\t\014\2108\2205\230\001\330\014\022\320\022\"\240!\2401\330\010\035\230Q\330\010\020\320\020 \240\001\240\024\240U\250'\260\031\270'\300\021\300'\310\021\330\010\013\2104\210s\220!\330\014\022\320\022\"\240!\2401\200A\360\014\000\t\021\320\020 \240\001\240\024""\240Q\330\010\013\2104\210s\220!\330\014\022\320\022\"\240!\2401\200A\360\016\000\t\021\320\020%\240Q\240d\250%\250t\260:\270Q\330\010\013\2104\210s\220!\330\014\022\320\022\"\240!\2401\200A\330\031\032\360\024\000\t\021\320\020 \240\001\240\024\240U\250&\260\002\260&\270\010\300\002\300!\330!(\250\r\260Q\330\010\013\2104\210s\220!\330\014\022\320\022\"\240!\2401\200A\360\n\000\t\016\210X\220Q\330\010\020\320\020$\240A\240T\250\025\250b\260\002\260!\340\010\013\2104\210s\220!\330\014\022\320\022\"\240!\2401\340\010\017\210q\200A\360\024\000\t\023\220$\220a\330\r\016\330\014\024\320\024%\240Q\240d\250%\250q\260\007\260q\330\010\013\2104\210s\220!\330\014\022\320\022\"\240!\2401\330\010\014\210G\2207\230!\2305\240\001\330\010\014\210F\220\"\220A\220Q\330\010\017\210q\200A\360\n\000\t\017\210c\220\021\330\010\017\320\017#\2401\240D\250\005\250Q\250c\260\021\330\010\013\2103\210c\220\021\330\014\022\320\022\"\240!\2401\330\010\017\210q\200A\360\024\000\t\021\220\017\230q\240\004\240A\330\010\013\2104\210s\220!\330\014\022\320\022\"\240!\2401\200A\360\014\000\016\017\330\014\024\320\024'\240q\250\004\250E\260\021\260!\330\010\013\2104\210r\220\021\330\014\022\320\022\"\240!\2401\330\r\021\220\023\220A\330\014\023\2201\330\010\017\210r\220\021\220&\230\002\230(\240'\250\022\2509\260G\2702\270Q\330\022\035\230R\230}\250I\260R\260q\330\022\031\230\022\2309\240G\2502\250Q\330\022\033\2302\230[\250\t\260\022\2601\330\022\035\230R\230}\250E\260\022\2607\270#\270R\270q\330\022!\240\022\2401\200A\360\016\000\t\021\220\017\230q\240\004\240E\250\023\250J\260a\330\010\013\2104\210s\220!\330\014\022\320\022\"\240!\2401\360\006\000\t\014\2104\210u\220C\220q\330\014\020\220\007\220w\230a\230u\240A\330\010\013\2104\210t\2203\220a\330\014\020\220\006\220b\230\001\230\021\200A\330\030\031\360\024\000\t\021\220\017\230q\240\004\240E\250\026\250r\260\026\260x\270r\300\021\330 '\240}\260A\330\010\013\2104\210s\220!\330\014\022\320\022\"\240!\2401\200A\360\022\000\t\021\220\017\230q\240\004""\240F\250$\250d\260$\260e\2701\270G\3001\330\010\013\2104\210s\220!\330\014\022\320\022\"\240!\2401\200A\360\020\000\t\020\210r\220\021\220&\230\007\230q\320\004\034\230A\360\016\000\t\020\210r\220\021\220&\230\007\230q\200A\360\014\000\t\020\210v\220Q\330\010\024\220J\230a\330\010\016\210m\2301\230D\240\003\2405\250\001\250\024\250W\260K\270q\300\t\310\021\330\010\013\2103\210c\220\021\330\014\022\320\022\"\240!\320#9\270\021\330\010\017\210v\220Q\320\004+\320+F\300a\330\037 \360\024\000\t:\270\021\330\010\013\2104\210u\220D\230\004\230A\330\014\022\320\022\"\240!\2401\330\010\013\2104\210q\330\014\024\320\024(\250\001\250\024\250U\260$\260j\300\001\330)0\260\013\2701\340\014\024\320\024+\2501\250D\260\005\260Q\260g\270Q\330,3\260;\270a\330\010\013\2104\210s\220!\330\014\022\320\022\"\240!\2401\320\004/\320/J\310!\360\022\000\t\014\2104\210q\330\014\025\220R\220s\230!\330\010\020\220\017\230q\240\004\240E\250\026\250r\260\026\260x\270r\300\021\330 !\330\010\013\2104\210s\220!\330\014\022\320\022\"\240!\2401\320\004$\240N\3202C\3009\310A\360\036\000\t\033\230!\330\010\037\230q\340\010+\2501\340\010\013\2105\220\007\220q\330\014\017\210z\230\021\230&\240\001\330\020\027\220t\2307\240!\2401\330\014\030\230\007\230q\330\010\013\2108\2207\230!\330\014\017\210z\230\021\230)\2401\330\020\032\230'\240\027\250\001\250\021\330\014\023\2207\230!\340\010\024\220A\220Q\330\010\016\210a\210{\230#\230Q\330\010\r\320\r\036\230a\330\010\r\210\\\230\021\360\006\000\t\014\210;\220e\2301\330\014\021\220\037\240\014\250D\260\001\330\014\021\220\037\240\001\340\010\020\320\020&\240a\240t\2505\260\013\2706\300\021\330\010\013\2104\210s\220!\330\014\022\320\022\"\240!\2401\200\001\330\004\n\210+\220Q\320\004=\270Q\360\022\000\t\021\320\020\"\240!\2404\240u\250C\250z\270\021\330#&\240j\260\016\270a\330\010\013\2104\210s\220!\330\014\022\320\022\"\240!\2401\320\004\035\230Q\360\016\000\t\021\220\014\230A\230T\240\025\240a\330\010\013\2104\210s\220!\330\014\022\320\022\"\240!\2401\320\004%""\240Q\360\016\000\t)\250\001\250\025\250o\270[\310\001\330\010\020\220\014\230A\230T\240\025\240a\330\010\013\2104\210s\220!\330\014\022\320\022\"\240!\2401\320\004\035\230Q\360\016\000\t\020\210r\220\021\220&\230\007\230q\320\004&\240a\360\014\000\t)\250\001\250\025\250o\270[\310\001\330\010\020\220\r\230Q\230d\240%\240q\330\010\013\2104\210s\220!\330\014\022\320\022\"\240!\2401\320\004'\240q\360\022\000\t\021\320\020#\2401\240D\250\005\250Q\330\010\013\2104\210s\220!\330\014\022\320\022\"\240!\2401\360\006\000\t\014\2104\210u\220C\220q\330\014\020\220\007\220w\230a\230u\240A\330\010\013\2104\210t\2203\220a\330\014\020\220\006\220b\230\001\230\021\320\004/\250q\360\020\000\t\014\2104\210q\330\014\025\220R\220s\230!\330\010\020\220\017\230q\240\004\240E\250\026\250r\260\026\260x\270r\300\021\330\010\013\2104\210s\220!\330\014\022\320\022\"\240!\2401\320\004\037\230z\320);\320;M\310Q\330 )\250\021\360\030\000\t\037\230a\330\010\036\230a\330\010\036\230a\330\010\036\230a\340\010+\2501\330\010%\240Q\340\010\013\2104\210w\220a\330\014\017\210z\230\021\230%\230q\330\020\026\220c\230\027\240\001\240\021\330\014\027\220w\230a\330\010\013\2104\210w\220a\330\014\017\210z\230\021\230%\230q\330\020\026\220c\230\027\240\001\240\021\330\014\027\220w\230a\330\010\013\210<\220w\230a\330\014\017\210z\230\021\230-\240q\330\020\036\230k\250\027\260\001\260\021\330\014\027\220w\230a\330\010\013\210<\220w\230a\330\014\017\210z\230\021\230-\240q\330\020\036\230k\250\027\260\001\260\021\330\014\027\220w\230a\340\010\024\220A\220Q\330\010\016\210a\210{\230#\230Q\330\010\r\320\r\036\230a\330\010\r\210\\\230\021\360\006\000\t\014\210;\220e\2301\330\014\021\220\037\240\014\250D\260\001\330\014\021\220\037\240\001\340\010\013\2106\220\024\220Q\330\014\024\320\024%\240Q\240j\260\n\270!\330&'\240t\2501\340\014\017\210q\330\020\025\220_\240A\330\020\030\320\030)\250\021\250*\260J\270k\310\021\310!\330\020\023\2204\220s\230!\330\024\032\320\032*\250!\2501\330\020\025\220_\240N\260#\260Q\330\020""\025\220^\2403\240a\330\020\025\220]\240#\240Q\330\014\024\320\024%\240Q\240j\260\n\270!\330&'\240t\2501\330\014\017\210q\330\022$\240A\240Q\330\010\013\2104\210s\220!\330\014\022\320\022\"\240!\2401";
    PyObject *data = NULL;
    CYTHON_UNUSED_VAR(__Pyx_DecompressString);
    #endif
    PyObject **stringtab = __pyx_mstate->__pyx_string_tab;
    Py_ssize_t pos = 0;
    for (int i = 0; i < 393; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyUnicode_DecodeUTF8(bytes + pos, bytes_length, NULL);
      if (likely(string) && i >= 77) PyUnicode_InternInPlace(&string);
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
      stringtab[i] = string;
      pos += bytes_length;
    }
    for (int i = 393; i < 432; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyBytes_FromStringAndSize(bytes + pos, bytes_length);
      stringtab[i] = string;
      pos += bytes_length;
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    Py_XDECREF(data);
    for (Py_ssize_t i = 0; i < 432; i++) {
      if (unlikely(PyObject_Hash(stringtab[i]) == -1)) {
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    #if CYTHON_IMMORTAL_CONSTANTS
    {
      PyObject **table = stringtab + 393;
      for (Py_ssize_t i=0; i<39; ++i) {
        #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
        #if PY_VERSION_HEX < 0x030E0000
        if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
        #else
        if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
        #endif
        {
          Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
        }
        #else
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
        #endif
      }
    }
    #endif
  }
  {
    PyObject **numbertab = __pyx_mstate->__pyx_number_tab + 0;
    int8_t const cint_constants_1[] = {0,1,5};
    int16_t const cint_constants_2[] = {2000};
    for (int i = 0; i < 4; i++) {
      numbertab[i] = PyLong_FromLong((i < 3 ? cint_constants_1[i - 0] : cint_constants_2[i - 3]));
      if (unlikely(!numbertab[i])) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_number_tab;
    for (Py_ssize_t i=0; i<4; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_codeobjects ### */
typedef struct {
    unsigned int argcount : 3;
    unsigned int num_posonly_args : 1;
    unsigned int num_kwonly_args : 1;
    unsigned int nlocals : 4;
    unsigned int flags : 10;
    unsigned int first_line : 10;
} __Pyx_PyCode_New_function_description;
/* NewCodeObj.proto */
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
);


static int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate) {
  PyObject* tuple_dedup_map = PyDict_New();
  if (unlikely(!tuple_dedup_map)) return -1;
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 79};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_data, __pyx_mstate->__pyx_n_u_i, __pyx_mstate->__pyx_n_u_ptr};
    __pyx_mstate_global->__pyx_codeobj_tab[0] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_set_private_data, __pyx_mstate->__pyx_kp_b_iso88591_A_1D_1G3a_E_as_4q_Qe1_E_e3at83aq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[0])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[1] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[1])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[2] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[2])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[3] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[3])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[4] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[4])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[5] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[5])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[6] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[6])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {7, 0, 0, 15, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 190};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_src, __pyx_mstate->__pyx_n_u_dst, __pyx_mstate->__pyx_n_u_src_service, __pyx_mstate->__pyx_n_u_dst_service, __pyx_mstate->__pyx_n_u_port_space, __pyx_mstate->__pyx_n_u_flags, __pyx_mstate->__pyx_n_u_src_srvc, __pyx_mstate->__pyx_n_u_dst_srvc, __pyx_mstate->__pyx_n_u_src_addr, __pyx_mstate->__pyx_n_u_dst_addr, __pyx_mstate->__pyx_n_u_hints, __pyx_mstate->__pyx_n_u_hints_ptr, __pyx_mstate->__pyx_n_u_res, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[7] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_get_addrinfo, __pyx_mstate->__pyx_kp_b_iso88591_z_MQ_a_a_a_a_1_Q_4wa_z_q_c_wa_4, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[7])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 260};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[8] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_close, __pyx_mstate->__pyx_kp_b_iso88591_A_4_Q_t1_G6_Q_M, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[8])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[9] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[9])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[10] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[10])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 285};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[11] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_close, __pyx_mstate->__pyx_kp_b_iso88591_A_4wc_t1_G6_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[11])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 296};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[12] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_ack_cm_event, __pyx_mstate->__pyx_kp_b_iso88591_A_4q_4s_1_4z_IQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[12])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 308};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[13] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_event_str, __pyx_mstate->__pyx_kp_b_iso88591_A_4wc_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[13])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[14] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[14])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[15] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[15])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 339};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[16] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_close, __pyx_mstate->__pyx_kp_b_iso88591_A_4_c_t1_G6_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[16])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[17] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[17])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[18] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[18])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 443};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[19] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_close, __pyx_mstate->__pyx_kp_b_iso88591_A_4t3a_t1_G6_t_Q_4q_4s_c_at1_Q_4, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[19])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 462};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_to_conn, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[20] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_get_request, __pyx_mstate->__pyx_kp_b_iso88591_A_a_Qd_q_q_4s_1_G7_5_F_AQ_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[20])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 481};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_lai, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[21] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_bind_addr, __pyx_mstate->__pyx_kp_b_iso88591_A_q_E_Ja_4s_1_4uCq_wauA_4t3a_b, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[21])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 498};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_rai, __pyx_mstate->__pyx_n_u_timeout_ms, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[22] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_resolve_addr, __pyx_mstate->__pyx_kp_b_iso88591_Q_4uCz_j_a_4s_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[22])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 512};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_addr, __pyx_mstate->__pyx_n_u_mc_attr, __pyx_mstate->__pyx_n_u_context, __pyx_mstate->__pyx_n_u_mc_join_attr, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[23] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_join_multicast, __pyx_mstate->__pyx_kp_b_iso88591_Fa_4uD_A_1_4q_U_j_0_1_1D_QgQ_3, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[23])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 535};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_addr, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[24] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_leave_multicast, __pyx_mstate->__pyx_kp_b_iso88591_A_Qd_t_Q_4s_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[24])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 546};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_timeout_ms, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[25] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_resolve_route, __pyx_mstate->__pyx_kp_b_iso88591_q_1D_Q_4s_1_4uCq_wauA_4t3a_b, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[25])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 565};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_backlog, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[26] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_listen, __pyx_mstate->__pyx_kp_b_iso88591_Q_AT_a_4s_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[26])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 576};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_param, __pyx_mstate->__pyx_n_u_conn, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[27] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_connect, __pyx_mstate->__pyx_kp_b_iso88591_a_o_Qd_q_4s_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[27])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 587};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[28] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_disconnect, __pyx_mstate->__pyx_kp_b_iso88591_A_Q_4s_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[28])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 597};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_param, __pyx_mstate->__pyx_n_u_conn, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[29] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_accept, __pyx_mstate->__pyx_kp_b_iso88591_Q_o_AT_a_4s_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[29])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 609};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[30] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_establish, __pyx_mstate->__pyx_kp_b_iso88591_A_q_A_4s_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[30])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 623};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_ece, __pyx_mstate->__pyx_n_u_rc};
    __pyx_mstate_global->__pyx_codeobj_tab[31] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_set_local_ece, __pyx_mstate->__pyx_kp_b_iso88591_A_4uAS_3c_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[31])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 633};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_ece, __pyx_mstate->__pyx_n_u_rc};
    __pyx_mstate_global->__pyx_codeobj_tab[32] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_get_remote_ece, __pyx_mstate->__pyx_kp_b_iso88591_A_c_1D_Qc_3c_1_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[32])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 644};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_qp_init, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[33] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_create_qp, __pyx_mstate->__pyx_kp_b_iso88591_A_q_F_d_e1G1_4s_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[33])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 657};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_attr_mask, __pyx_mstate->__pyx_n_u_attr, __pyx_mstate->__pyx_n_u_init_attr, __pyx_mstate->__pyx_n_u_rc};
    __pyx_mstate_global->__pyx_codeobj_tab[34] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_query_qp, __pyx_mstate->__pyx_kp_b_iso88591_A_vQ_Ja_m1D_5_WKq_3c_9_vQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[34])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 670};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_qp_state, __pyx_mstate->__pyx_n_u_attr_mask, __pyx_mstate->__pyx_n_u_qp_attr, __pyx_mstate->__pyx_n_u_rc};
    __pyx_mstate_global->__pyx_codeobj_tab[35] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_init_qp_attr, __pyx_mstate->__pyx_kp_b_iso88591_A_1_e1G7_1_3c_1_y, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[35])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 690};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_size};
    __pyx_mstate_global->__pyx_codeobj_tab[36] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_reg_msgs, __pyx_mstate->__pyx_kp_b_iso88591_A_r_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[36])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 700};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_size};
    __pyx_mstate_global->__pyx_codeobj_tab[37] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_reg_read, __pyx_mstate->__pyx_kp_b_iso88591_A_r_q_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[37])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 709};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_size};
    __pyx_mstate_global->__pyx_codeobj_tab[38] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_reg_write, __pyx_mstate->__pyx_kp_b_iso88591_Q_r_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[38])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 718};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_mr, __pyx_mstate->__pyx_n_u_length, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[39] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_post_recv, __pyx_mstate->__pyx_kp_b_iso88591_q_4q_Rs_q_E_r_xr_4s_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[39])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 732};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_mr, __pyx_mstate->__pyx_n_u_flags, __pyx_mstate->__pyx_n_u_length, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[40] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_post_send, __pyx_mstate->__pyx_kp_b_iso88591_J_4q_Rs_q_E_r_xr_4s_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[40])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 7, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 748};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_mr, __pyx_mstate->__pyx_n_u_length, __pyx_mstate->__pyx_n_u_remote_addr, __pyx_mstate->__pyx_n_u_rkey, __pyx_mstate->__pyx_n_u_flags, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[41] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_post_read, __pyx_mstate->__pyx_kp_b_iso88591_A_q_E_r_xr_A_4s_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[41])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 7, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 764};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_mr, __pyx_mstate->__pyx_n_u_length, __pyx_mstate->__pyx_n_u_remote_addr, __pyx_mstate->__pyx_n_u_rkey, __pyx_mstate->__pyx_n_u_flags, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[42] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_post_write, __pyx_mstate->__pyx_kp_b_iso88591_A_U_Q_4s_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[42])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {6, 0, 0, 7, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 780};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_mr, __pyx_mstate->__pyx_n_u_ah, __pyx_mstate->__pyx_n_u_rqpn, __pyx_mstate->__pyx_n_u_flags, __pyx_mstate->__pyx_n_u_length, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[43] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_post_ud_send, __pyx_mstate->__pyx_kp_b_iso88591_7q_4q_Rs_4uF_F_A_E_4s_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[43])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 798};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_wc, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[44] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_get_recv_comp, __pyx_mstate->__pyx_kp_b_iso88591_A_q_E_4r_1_A_1_r_9G2Q_R_IRq_9G2Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[44])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 817};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_wc, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[45] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_get_send_comp, __pyx_mstate->__pyx_kp_b_iso88591_A_q_E_4r_1_A_1_r_9G2Q_R_IRq_9G2Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[45])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 7, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 836};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_level, __pyx_mstate->__pyx_n_u_optname, __pyx_mstate->__pyx_n_u_optval, __pyx_mstate->__pyx_n_u_optlen, __pyx_mstate->__pyx_n_u_value, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[46] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_set_option, __pyx_mstate->__pyx_kp_b_iso88591_A_85_1_Q_U_4s_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[46])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 851};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_private_data, __pyx_mstate->__pyx_n_u_data_len, __pyx_mstate->__pyx_n_u_buffer, __pyx_mstate->__pyx_n_u_data_ptr, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[47] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_reject, __pyx_mstate->__pyx_kp_b_iso88591_3a_AQ_A_1_vZq_AT_m_Q_4s_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[47])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 10, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 867};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_node, __pyx_mstate->__pyx_n_u_service, __pyx_mstate->__pyx_n_u_port_space, __pyx_mstate->__pyx_n_u_flags, __pyx_mstate->__pyx_n_u_srvc, __pyx_mstate->__pyx_n_u_node_addr, __pyx_mstate->__pyx_n_u_hints, __pyx_mstate->__pyx_n_u_hints_ptr, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[48] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_resolve_addrinfo, __pyx_mstate->__pyx_kp_b_iso88591_N2C9A_q_1_5_q_z_t7_1_q_87_z_1_7, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[48])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 910};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_ai, __pyx_mstate->__pyx_n_u_ret};
    __pyx_mstate_global->__pyx_codeobj_tab[49] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyverbs_cmid_pyx, __pyx_mstate->__pyx_n_u_query_addrinfo, __pyx_mstate->__pyx_kp_b_iso88591_A_XQ_AT_b_4s_1_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[49])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[50] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[50])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[51] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_tree_fragment, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[51])) goto bad;
  }
  Py_DECREF(tuple_dedup_map);
  return 0;
  bad:
  Py_DECREF(tuple_dedup_map);
  return -1;
}
/* #### Code section: init_globals ### */

static int __Pyx_InitGlobals(void) {
  /* PythonCompatibility.init */
  if (likely(__Pyx_init_co_variables() == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CommonTypesMetaclass.init */
  if (likely(__pyx_CommonTypesMetaclass_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CachedMethodType.init */
  #if CYTHON_COMPILING_IN_LIMITED_API
  {
      PyObject *typesModule=NULL;
      typesModule = PyImport_ImportModule("types");
      if (typesModule) {
          __pyx_mstate_global->__Pyx_CachedMethodType = PyObject_GetAttrString(typesModule, "MethodType");
          Py_DECREF(typesModule);
      }
  } // error handling follows
  #endif
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CythonFunctionShared.init */
  if (likely(__pyx_CyFunction_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError (used by GetBuiltinName) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_mstate_global->__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
            "name '%U' is not defined", name);
    }
    return result;
}

/* HasAttr (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!PyUnicode_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_PyObject_GetAttrStrNoError(o, n);
    if (!r) {
        return (unlikely(PyErr_Occurred())) ? -1 : 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}
#endif

/* PyObjectCall (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall (used by PyObjectCallOneArg) */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) != (0)) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
#if CYTHON_VECTORCALL && !CYTHON_COMPILING_IN_LIMITED_API
  #if PY_VERSION_HEX < 0x03090000
    #define __Pyx_PyVectorcall_Function(callable) _PyVectorcall_Function(callable)
  #elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE vectorcallfunc __Pyx_PyVectorcall_Function(PyObject *callable) {
    PyTypeObject *tp = Py_TYPE(callable);
    #if defined(__Pyx_CyFunction_USED)
    if (__Pyx_CyFunction_CheckExact(callable)) {
        return __Pyx_CyFunction_func_vectorcall(callable);
    }
    #endif
    if (!PyType_HasFeature(tp, Py_TPFLAGS_HAVE_VECTORCALL)) {
        return NULL;
    }
    assert(PyCallable_Check(callable));
    Py_ssize_t offset = tp->tp_vectorcall_offset;
    assert(offset > 0);
    vectorcallfunc ptr;
    memcpy(&ptr, (char *) callable + offset, sizeof(ptr));
    return ptr;
}
  #else
    #define __Pyx_PyVectorcall_Function(callable) PyVectorcall_Function(callable)
  #endif
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject *const *args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
          #if CYTHON_COMPILING_IN_LIMITED_API
            return PyObject_Vectorcall(func, args, _nargs, NULL);
          #else
            vectorcallfunc f = __Pyx_PyVectorcall_Function(func);
            if (f) {
                return f(func, args, _nargs, NULL);
            }
          #endif
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_mstate_global->__pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* PyObjectCallOneArg (used by ImportImpl) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* TupleAndListFromArray (used by ImportImpl) */
#if !CYTHON_COMPILING_IN_CPYTHON && CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    Py_ssize_t i;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    for (i = 0; i < n; i++) {
        if (unlikely(__Pyx_PyTuple_SET_ITEM(res, i, src[i]) < (0))) {
            Py_DECREF(res);
            return NULL;
        }
        Py_INCREF(src[i]);
    }
    return res;
}
#elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* ImportImpl (used by Import) */
static int __Pyx__Import_GetModule(PyObject *qualname, PyObject **module) {
    PyObject *imported_module = PyImport_GetModule(qualname);
    if (unlikely(!imported_module)) {
        *module = NULL;
        if (PyErr_Occurred()) {
            return -1;
        }
        return 0;
    }
    *module = imported_module;
    return 1;
}
static int __Pyx__Import_Lookup(PyObject *qualname, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject **module) {
    PyObject *imported_module;
    PyObject *top_level_package_name;
    Py_ssize_t i;
    int status, module_found;
    Py_ssize_t dot_index;
    module_found = __Pyx__Import_GetModule(qualname, &imported_module);
    if (unlikely(!module_found || module_found == -1)) {
        *module = NULL;
        return module_found;
    }
    if (imported_names) {
        for (i = 0; i < len_imported_names; i++) {
            PyObject *imported_name = imported_names[i];
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
            int has_imported_attribute = PyObject_HasAttr(imported_module, imported_name);
#else
            int has_imported_attribute = PyObject_HasAttrWithError(imported_module, imported_name);
            if (unlikely(has_imported_attribute == -1)) goto error;
#endif
            if (!has_imported_attribute) {
                goto not_found;
            }
        }
        *module = imported_module;
        return 1;
    }
    dot_index = PyUnicode_FindChar(qualname, '.', 0, PY_SSIZE_T_MAX, 1);
    if (dot_index == -1) {
        *module = imported_module;
        return 1;
    }
    if (unlikely(dot_index == -2)) goto error;
    top_level_package_name = PyUnicode_Substring(qualname, 0, dot_index);
    if (unlikely(!top_level_package_name)) goto error;
    Py_DECREF(imported_module);
    status = __Pyx__Import_GetModule(top_level_package_name, module);
    Py_DECREF(top_level_package_name);
    return status;
error:
    Py_DECREF(imported_module);
    *module = NULL;
    return -1;
not_found:
    Py_DECREF(imported_module);
    *module = NULL;
    return 0;
}
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *from_list = 0;
    int module_found;
    if (!qualname) {
        qualname = name;
    }
    module_found = __Pyx__Import_Lookup(qualname, imported_names, len_imported_names, &module);
    if (likely(module_found == 1)) {
        return module;
    } else if (unlikely(module_found == -1)) {
        return NULL;
    }
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    if (imported_names) {
#if CYTHON_COMPILING_IN_CPYTHON
        from_list = __Pyx_PyList_FromArray(imported_names, len_imported_names);
        if (unlikely(!from_list))
            goto bad;
#else
        from_list = PyList_New(len_imported_names);
        if (unlikely(!from_list)) goto bad;
        for (Py_ssize_t i=0; i<len_imported_names; ++i) {
            if (PyList_SetItem(from_list, i, __Pyx_NewRef(imported_names[i])) < 0) goto bad;
        }
#endif
    }
    if (level == -1) {
        const char* package_sep = strchr(__Pyx_MODULE_NAME, '.');
        if (package_sep != (0)) {
            module = PyImport_ImportModuleLevelObject(
                name, moddict, empty_dict, from_list, 1);
            if (unlikely(!module)) {
                if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                    goto bad;
                PyErr_Clear();
            }
        }
        level = 0;
    }
    if (!module) {
        module = PyImport_ImportModuleLevelObject(
            name, moddict, empty_dict, from_list, level);
    }
bad:
    Py_XDECREF(from_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level) {
    return __Pyx__Import(name, imported_names, len_imported_names, qualname, __pyx_mstate_global->__pyx_d, level);
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_mstate_global->__pyx_kp_u_);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400) ||\
                CYTHON_COMPILING_IN_GRAAL
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError, "cannot import name %S", name);
    }
    return value;
}

/* GetTopmostException (used by SaveResetException) */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C0000
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#elif __PYX_LIMITED_VERSION_HEX > 0x030C0000
    local_value = PyErr_GetRaisedException();
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
#if __PYX_LIMITED_VERSION_HEX > 0x030C0000
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
#else
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
#endif // __PYX_LIMITED_VERSION_HEX > 0x030C0000
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#elif __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    PyErr_SetHandledException(local_value);
    Py_XDECREF(local_value);
    Py_XDECREF(local_type);
    Py_XDECREF(local_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
#if __PYX_LIMITED_VERSION_HEX <= 0x030C0000
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
#endif
}

/* PyObjectFastCallMethod */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf) {
    PyObject *result;
    PyObject *attr = PyObject_GetAttr(args[0], name);
    if (unlikely(!attr))
        return NULL;
    result = __Pyx_PyObject_FastCall(attr, args+1, nargsf - 1);
    Py_DECREF(attr);
    return result;
}
#endif

/* BytesEquals (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL ||\
        !(CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS)
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals (used by fastcall) */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length, length2;
        int kind;
        void *data1, *data2;
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        #endif
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length < 0)) return -1;
        #endif
        length2 = __Pyx_PyUnicode_GET_LENGTH(s2);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length2 < 0)) return -1;
        #endif
        if (length != length2) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    return (equals == Py_EQ);
return_ne:
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = __Pyx_PyTuple_GET_SIZE(kwnames);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(n == -1)) return NULL;
    #endif
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        if (s == namei) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        int eq = __Pyx_PyUnicode_Equals(s, namei, Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs;
    PyObject *dict;
#if !CYTHON_ASSUME_SAFE_SIZE
    nkwargs = PyTuple_Size(kwnames);
    if (unlikely(nkwargs < 0)) return NULL;
#else
    nkwargs = PyTuple_GET_SIZE(kwnames);
#endif
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
#if !CYTHON_ASSUME_SAFE_MACROS
        PyObject *key = PyTuple_GetItem(kwnames, i);
        if (!key) goto bad;
#else
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
#endif
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* UnpackUnboundCMethod (used by CallUnboundCMethod0) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
#elif CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject **args, Py_ssize_t nargs, PyObject *kwnames) {
        return _PyObject_Vectorcall
            (method, args ? args+1 : NULL, nargs ? nargs-1 : 0, kwnames);
}
#else
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *const *args, Py_ssize_t nargs, PyObject *kwnames) {
    return
#if PY_VERSION_HEX < 0x03090000
    _PyObject_Vectorcall
#else
    PyObject_Vectorcall
#endif
        (method, args ? args+1 : NULL, nargs ? (size_t) nargs-1 : 0, kwnames);
}
#endif
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
     METH_VARARGS | METH_KEYWORDS,
#else
     METH_FASTCALL | METH_KEYWORDS,
#endif
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method, *result=NULL;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    result = method;
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            result = unbound_method;
        }
    }
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    if (unlikely(target->method)) {
        Py_DECREF(result);
    } else
#endif
    target->method = result;
    return 0;
}

/* CallUnboundCMethod0 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        if (likely(cfunc->flag == METH_NOARGS))
            return __Pyx_CallCFunction(cfunc, self, NULL);
        if (likely(cfunc->flag == METH_FASTCALL))
            return __Pyx_CallCFunctionFast(cfunc, self, NULL, 0);
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, NULL, 0, NULL);
        if (likely(cfunc->flag == (METH_VARARGS | METH_KEYWORDS)))
            return __Pyx_CallCFunctionWithKeywords(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple, NULL);
        if (cfunc->flag == METH_VARARGS)
            return __Pyx_CallCFunction(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple);
        return __Pyx__CallUnboundCMethod0(cfunc, self);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod0(&tmp_cfunc, self);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod0(cfunc, self);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *result;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
    result = __Pyx_PyObject_CallOneArg(cfunc->method, self);
    return result;
}

/* py_dict_items (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_items, d);
}

/* py_dict_values (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_values, d);
}

/* OwnedDictNext (used by ParseKeywordsImpl) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue) {
    PyObject *next = NULL;
    if (!*ppos) {
        if (pvalue) {
            PyObject *dictview = pkey ? __Pyx_PyDict_Items(p) : __Pyx_PyDict_Values(p);
            if (unlikely(!dictview)) goto bad;
            *ppos = PyObject_GetIter(dictview);
            Py_DECREF(dictview);
        } else {
            *ppos = PyObject_GetIter(p);
        }
        if (unlikely(!*ppos)) goto bad;
    }
    next = PyIter_Next(*ppos);
    if (!next) {
        if (PyErr_Occurred()) goto bad;
        return 0;
    }
    if (pkey && pvalue) {
        *pkey = __Pyx_PySequence_ITEM(next, 0);
        if (unlikely(*pkey)) goto bad;
        *pvalue = __Pyx_PySequence_ITEM(next, 1);
        if (unlikely(*pvalue)) goto bad;
        Py_DECREF(next);
    } else if (pkey) {
        *pkey = next;
    } else {
        assert(pvalue);
        *pvalue = next;
    }
    return 1;
  bad:
    Py_XDECREF(next);
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
    PyErr_FormatUnraisable("Exception ignored in __Pyx_PyDict_NextRef");
#else
    PyErr_WriteUnraisable(__pyx_mstate_global->__pyx_n_u_Pyx_PyDict_NextRef);
#endif
    if (pkey) *pkey = NULL;
    if (pvalue) *pvalue = NULL;
    return 0;
}
#else // !CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue) {
    int result = PyDict_Next(p, ppos, pkey, pvalue);
    if (likely(result == 1)) {
        if (pkey) Py_INCREF(*pkey);
        if (pvalue) Py_INCREF(*pvalue);
    }
    return result;
}
#endif

/* RaiseDoubleKeywords (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            return __Pyx_CallCFunctionFast(cfunc, self, args, 2);
        }
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, 2, NULL);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod2(&tmp_cfunc, self, arg1, arg2);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        PyObject *result = NULL;
        PyObject *args = PyTuple_New(2);
        if (unlikely(!args)) return NULL;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = __Pyx_CallCFunctionWithKeywords(cfunc, self, args, NULL);
        else
            result = __Pyx_CallCFunction(cfunc, self, args);
        Py_DECREF(args);
        return result;
    }
#endif
    {
        PyObject *args[4] = {NULL, self, arg1, arg2};
        return __Pyx_PyObject_FastCall(cfunc->method, args+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
    }
}

/* ParseKeywordsImpl (used by ParseKeywords) */
static int __Pyx_ValidateDuplicatePosArgs(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char* function_name)
{
    PyObject ** const *name = argnames;
    while (name != first_kw_arg) {
        PyObject *key = **name;
        int found = PyDict_Contains(kwds, key);
        if (unlikely(found)) {
            if (found == 1) __Pyx_RaiseDoubleKeywordsError(function_name, key);
            goto bad;
        }
        name++;
    }
    return 0;
bad:
    return -1;
}
#if CYTHON_USE_UNICODE_INTERNALS
static CYTHON_INLINE int __Pyx_UnicodeKeywordsEqual(PyObject *s1, PyObject *s2) {
    int kind;
    Py_ssize_t len = PyUnicode_GET_LENGTH(s1);
    if (len != PyUnicode_GET_LENGTH(s2)) return 0;
    kind = PyUnicode_KIND(s1);
    if (kind != PyUnicode_KIND(s2)) return 0;
    const void *data1 = PyUnicode_DATA(s1);
    const void *data2 = PyUnicode_DATA(s2);
    return (memcmp(data1, data2, (size_t) len * (size_t) kind) == 0);
}
#endif
static int __Pyx_MatchKeywordArg_str(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    #if CYTHON_USE_UNICODE_INTERNALS
    Py_hash_t key_hash = ((PyASCIIObject*)key)->hash;
    if (unlikely(key_hash == -1)) {
        key_hash = PyObject_Hash(key);
        if (unlikely(key_hash == -1))
            goto bad;
    }
    #endif
    name = first_kw_arg;
    while (*name) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (key_hash == ((PyASCIIObject*)name_str)->hash && __Pyx_UnicodeKeywordsEqual(name_str, key)) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) {
                *index_found = (size_t) (name - argnames);
                return 1;
            }
        }
        #endif
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (unlikely(key_hash == ((PyASCIIObject*)name_str)->hash)) {
            if (__Pyx_UnicodeKeywordsEqual(name_str, key))
                goto arg_passed_twice;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            if (unlikely(name_str == key)) goto arg_passed_twice;
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) goto arg_passed_twice;
        }
        #endif
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
bad:
    return -1;
}
static int __Pyx_MatchKeywordArg_nostr(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    if (unlikely(!PyUnicode_Check(key))) goto invalid_keyword_type;
    name = first_kw_arg;
    while (*name) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (cmp == 1) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        if (unlikely(cmp == -1)) goto bad;
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (unlikely(cmp != 0)) {
            if (cmp == 1) goto arg_passed_twice;
            else goto bad;
        }
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
bad:
    return -1;
}
static CYTHON_INLINE int __Pyx_MatchKeywordArg(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    return likely(PyUnicode_CheckExact(key)) ?
        __Pyx_MatchKeywordArg_str(key, argnames, first_kw_arg, index_found, function_name) :
        __Pyx_MatchKeywordArg_nostr(key, argnames, first_kw_arg, index_found, function_name);
}
static void __Pyx_RejectUnknownKeyword(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char *function_name)
{
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos = NULL;
    #else
    Py_ssize_t pos = 0;
    #endif
    PyObject *key = NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(kwds);
    while (
        #if CYTHON_AVOID_BORROWED_REFS
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL)
        #else
        PyDict_Next(kwds, &pos, &key, NULL)
        #endif
    ) {
        PyObject** const *name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (!*name) {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp != 1) {
                if (cmp == 0) {
                    PyErr_Format(PyExc_TypeError,
                        "%s() got an unexpected keyword argument '%U'",
                        function_name, key);
                }
                #if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(key);
                #endif
                break;
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        #endif
    }
    __Pyx_END_CRITICAL_SECTION();
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(pos);
    #endif
    assert(PyErr_Occurred());
}
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t extracted = 0;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    name = first_kw_arg;
    while (*name && num_kwargs > extracted) {
        PyObject * key = **name;
        PyObject *value;
        int found = 0;
        #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        found = PyDict_GetItemRef(kwds, key, &value);
        #else
        value = PyDict_GetItemWithError(kwds, key);
        if (value) {
            Py_INCREF(value);
            found = 1;
        } else {
            if (unlikely(PyErr_Occurred())) goto bad;
        }
        #endif
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            extracted++;
        }
        name++;
    }
    if (num_kwargs > extracted) {
        if (ignore_unknown_kwargs) {
            if (unlikely(__Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name) == -1))
                goto bad;
        } else {
            __Pyx_RejectUnknownKeyword(kwds, argnames, first_kw_arg, function_name);
            goto bad;
        }
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t len;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    if (PyDict_Update(kwds2, kwds) < 0) goto bad;
    name = first_kw_arg;
    while (*name) {
        PyObject *key = **name;
        PyObject *value;
#if !CYTHON_COMPILING_IN_LIMITED_API && (PY_VERSION_HEX >= 0x030d00A2 || defined(PyDict_Pop))
        int found = PyDict_Pop(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
        }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        int found = PyDict_GetItemRef(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            if (unlikely(PyDict_DelItem(kwds2, key) < 0)) goto bad;
        }
#else
    #if CYTHON_COMPILING_IN_CPYTHON
        value = _PyDict_Pop(kwds2, key, kwds2);
    #else
        value = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_pop, kwds2, key, kwds2);
    #endif
        if (value == kwds2) {
            Py_DECREF(value);
        } else {
            if (unlikely(!value)) goto bad;
            values[name-argnames] = value;
        }
#endif
        name++;
    }
    len = PyDict_Size(kwds2);
    if (len > 0) {
        return __Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name);
    } else if (unlikely(len == -1)) {
        goto bad;
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject *key = NULL;
    PyObject** const * name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    for (Py_ssize_t pos = 0; pos < num_kwargs; pos++) {
#if CYTHON_AVOID_BORROWED_REFS
        key = __Pyx_PySequence_ITEM(kwds, pos);
#else
        key = __Pyx_PyTuple_GET_ITEM(kwds, pos);
#endif
#if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!key)) goto bad;
#endif
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            PyObject *value = kwvalues[pos];
            values[name-argnames] = __Pyx_NewRef(value);
        } else {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp == 1) {
                PyObject *value = kwvalues[pos];
                values[index_found] = __Pyx_NewRef(value);
            } else {
                if (unlikely(cmp == -1)) goto bad;
                if (kwds2) {
                    PyObject *value = kwvalues[pos];
                    if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
                } else if (!ignore_unknown_kwargs) {
                    goto invalid_keyword;
                }
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        key = NULL;
        #endif
    }
    return 0;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    goto bad;
bad:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(key);
    #endif
    return -1;
}

/* ParseKeywords */
static int __Pyx_ParseKeywords(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds)))
        return __Pyx_ParseKeywordsTuple(kwds, kwvalues, argnames, kwds2, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
    else if (kwds2)
        return __Pyx_ParseKeywordDictToDict(kwds, argnames, kwds2, values, num_pos_args, function_name);
    else
        return __Pyx_ParseKeywordDict(kwds, argnames, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyDictVersioning (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        if (!PyErr_Occurred())
            PyErr_SetNone(PyExc_NameError);
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
    PyErr_Clear();
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    if (unlikely(__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, name, &result) == -1)) PyErr_Clear();
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return result;
    }
#else
    result = _PyDict_GetItem_KnownHash(__pyx_mstate_global->__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* SetStringIndexingError (used by SetItemIntByteArray) */
static void __Pyx_SetStringIndexingError(const char* message, int has_gil) {
    if (!has_gil) {
        PyGILState_STATE gil_state = PyGILState_Ensure();
        PyErr_SetString(PyExc_IndexError, message);
        PyGILState_Release(gil_state);
    } else
        PyErr_SetString(PyExc_IndexError, message);
}

/* SetItemIntByteArray */
static CYTHON_INLINE int __Pyx_SetItemInt_ByteArray_Fast_Locked(PyObject* string, Py_ssize_t i, unsigned char v,
                                                                int wraparound, int boundscheck, int has_gil) {
    Py_ssize_t length = __Pyx_PyByteArray_GET_SIZE(string);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(length < 0)) return -1;
    #endif
    if (wraparound & unlikely(i < 0)) i += length;
    if ((!boundscheck) || likely(__Pyx_is_valid_index(i, length))) {
        #if !CYTHON_ASSUME_SAFE_MACROS
        char *asString = PyByteArray_AsString(string);
        if (unlikely(!asString)) return -1;
        asString[i] = (char)v;
        #else
        PyByteArray_AS_STRING(string)[i] = (char) v;
        #endif
        return 0;
    } else {
        __Pyx_SetStringIndexingError("bytearray index out of range", has_gil);
        return -1;
    }
}
static CYTHON_INLINE int __Pyx_SetItemInt_ByteArray_Fast(PyObject* string, Py_ssize_t i, unsigned char v,
                                                         int wraparound, int boundscheck, int has_gil, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    wraparound = wraparound && i<0;
#endif
    if (wraparound | boundscheck) {
        int result;
        __Pyx_PyCriticalSection cs;
        int lock = CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && has_gil && !__Pyx_IS_UNIQUELY_REFERENCED(string, unsafe_shared);
        if (lock) { __Pyx_PyCriticalSection_Begin(&cs, string); }
        result = __Pyx_SetItemInt_ByteArray_Fast_Locked(string, i, v, wraparound, boundscheck, has_gil);
        if (lock) { __Pyx_PyCriticalSection_End(&cs); }
        return result;
    } else {
        #if !CYTHON_ASSUME_SAFE_MACROS
        char *asString = PyByteArray_AsString(string);
        if (unlikely(!asString)) return -1;
        asString[i] = (char)v;
        #else
        PyByteArray_AS_STRING(string)[i] = (char) v;
        #endif
        return 0;
    }
}

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (unlikely(!j)) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE && !CYTHON_AVOID_BORROWED_REFS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS && !__Pyx_IS_UNIQUELY_REFERENCED(o, unsafe_shared))) {
            Py_INCREF(v);
            return PyList_SetItem(o, n, v);
        } else if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old;
            Py_INCREF(v);
            old = PyList_GET_ITEM(o, n);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 0;
        }
    } else
#endif
#if CYTHON_USE_TYPE_SLOTS && !CYTHON_COMPILING_IN_PYPY
    {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (!is_list && mm && mm->mp_ass_subscript) {
            int r;
            PyObject *key = PyLong_FromSsize_t(i);
            if (unlikely(!key)) return -1;
            r = mm->mp_ass_subscript(o, key, v);
            Py_DECREF(key);
            return r;
        }
        if (is_list || likely(sm && sm->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return sm->sq_ass_item(o, i, v);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_SetItem(o, i, v);
    }
#endif
    (void)wraparound;
    (void)boundscheck;
    return __Pyx_SetItemInt_Generic(o, PyLong_FromSsize_t(i), v);
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS)) {
        return __Pyx_PyList_GetItemRefFast(o, wrapped_i, unsafe_shared);
    } else
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyList_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyTuple_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)) {
            return __Pyx_PyList_GetItemRefFast(o, n, unsafe_shared);
        } else if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            return __Pyx_NewRef(PyList_GET_ITEM(o, n));
        }
    } else
    #if !CYTHON_AVOID_BORROWED_REFS
    if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            return __Pyx_NewRef(PyTuple_GET_ITEM(o, n));
        }
    } else
    #endif
#endif
#if CYTHON_USE_TYPE_SLOTS && !CYTHON_COMPILING_IN_PYPY
    {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (!is_list && mm && mm->mp_subscript) {
            PyObject *r, *key = PyLong_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (is_list || likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    (void)wraparound;
    (void)boundscheck;
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
}

/* UnicodeConcatInPlace */
# if CYTHON_COMPILING_IN_CPYTHON
static int
__Pyx_unicode_modifiable(PyObject *unicode, int unsafe_shared)
{
    if (!__Pyx_IS_UNIQUELY_REFERENCED(unicode, unsafe_shared))
        return 0;
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030F0000
    if (PyUnstable_Unicode_GET_CACHED_HASH(unicode) != -1)
        return 0;
#endif
    if (!PyUnicode_CheckExact(unicode))
        return 0;
    if (PyUnicode_CHECK_INTERNED(unicode))
        return 0;
    return 1;
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right, int unsafe_shared
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    ) {
    PyObject *left = *p_left;
    Py_ssize_t left_len, right_len, new_len;
    if (unlikely(__Pyx_PyUnicode_READY(left) == -1))
        return NULL;
    if (unlikely(__Pyx_PyUnicode_READY(right) == -1))
        return NULL;
    left_len = PyUnicode_GET_LENGTH(left);
    if (left_len == 0) {
        Py_INCREF(right);
        return right;
    }
    right_len = PyUnicode_GET_LENGTH(right);
    if (right_len == 0) {
        Py_INCREF(left);
        return left;
    }
    if (unlikely(left_len > PY_SSIZE_T_MAX - right_len)) {
        PyErr_SetString(PyExc_OverflowError,
                        "strings are too large to concat");
        return NULL;
    }
    new_len = left_len + right_len;
    if (__Pyx_unicode_modifiable(left, unsafe_shared)
            && PyUnicode_CheckExact(right)
            && PyUnicode_KIND(right) <= PyUnicode_KIND(left)
            && !(PyUnicode_IS_ASCII(left) && !PyUnicode_IS_ASCII(right))) {
        int ret;
        __Pyx_GIVEREF(*p_left);
        ret = PyUnicode_Resize(p_left, new_len);
        __Pyx_GOTREF(*p_left);
        if (unlikely(ret != 0))
            return NULL;
        #if PY_VERSION_HEX >= 0x030d0000
        if (unlikely(PyUnicode_CopyCharacters(*p_left, left_len, right, 0, right_len) < 0)) return NULL;
        #else
        _PyUnicode_FastCopyCharacters(*p_left, left_len, right, 0, right_len);
        #endif
        __Pyx_INCREF(*p_left);
        __Pyx_GIVEREF(*p_left);
        return *p_left;
    } else {
        return __Pyx_PyUnicode_Concat(left, right);
    }
  }
#endif

/* RejectKeywords */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds) {
    PyObject *key = NULL;
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds))) {
        key = __Pyx_PySequence_ITEM(kwds, 0);
    } else {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *pos = NULL;
#else
        Py_ssize_t pos = 0;
#endif
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
        if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return;
#endif
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL);
#if CYTHON_AVOID_BORROWED_REFS
        Py_XDECREF(pos);
#endif
    }
    if (likely(key)) {
        PyErr_Format(PyExc_TypeError,
            "%s() got an unexpected keyword argument '%U'",
            function_name, key);
        Py_DECREF(key);
    }
}

/* RaiseException */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
#elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}

/* ArgTypeTestFunc (used by ArgTypeTest) */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    PyObject *extra_info = __pyx_mstate_global->__pyx_empty_unicode;
    int from_annotation_subclass = 0;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (!exact) {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    } else if (exact == 2) {
        if (__Pyx_TypeCheck(obj, type)) {
            from_annotation_subclass = 1;
            extra_info = __pyx_mstate_global->__pyx_kp_u_Note_that_Cython_is_deliberately;
        }
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(type);
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")"
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        "%s%U"
#endif
        , name, type_name, obj_type_name
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        , (from_annotation_subclass ? ". " : ""), extra_info
#endif
        );
#if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    if (exact == 2 && from_annotation_subclass) {
        PyObject *res;
        PyObject *vargs[2];
        vargs[0] = PyErr_GetRaisedException();
        vargs[1] = extra_info;
        res = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_kp_u_add_note, vargs, 2, NULL);
        Py_XDECREF(res);
        PyErr_SetRaisedException(vargs[0]);
    }
#endif
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
    CYTHON_UNUSED_VAR(clineno);
    CYTHON_UNUSED_VAR(lineno);
    CYTHON_UNUSED_VAR(filename);
    CYTHON_MAYBE_UNUSED_VAR(nogil);
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(0);
    }
    ctx = PyUnicode_FromString(name);
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
    if (nogil)
        PyGILState_Release(state);
}

/* decode_c_bytes (used by decode_bytes) */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    if (unlikely((start < 0) | (stop < 0))) {
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (stop > length)
        stop = length;
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* PyObjectVectorCallKwBuilder */
#if CYTHON_VECTORCALL
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_PyObject_FastCallDict;
    if (__Pyx_PyTuple_SET_ITEM(builder, n, key) != (0)) return -1;
    Py_INCREF(key);
    args[n] = value;
    return 0;
}
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_VectorcallBuilder_AddArgStr;
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n);
}
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    PyObject *pyKey = PyUnicode_FromString(key);
    if (!pyKey) return -1;
    return __Pyx_VectorcallBuilder_AddArg(pyKey, value, builder, args, n);
}
#else // CYTHON_VECTORCALL
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, CYTHON_UNUSED PyObject **args, CYTHON_UNUSED int n) {
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return PyDict_SetItem(builder, key, value);
}
#endif

/* PyObjectVectorCallMethodKwBuilder */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_Object_VectorcallMethod_CallFromBuilder(PyObject *name, PyObject *const *args, size_t nargsf, PyObject *kwnames) {
    PyObject *result;
    PyObject *obj = PyObject_GetAttr(args[0], name);
    if (unlikely(!obj))
        return NULL;
    result = __Pyx_Object_Vectorcall_CallFromBuilder(obj, args+1, nargsf-1, kwnames);
    Py_DECREF(obj);
    return result;
}
#endif

/* decode_c_string */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    Py_ssize_t length;
    if (unlikely((start < 0) | (stop < 0))) {
        size_t slen = strlen(cstring);
        if (unlikely(slen > (size_t) PY_SSIZE_T_MAX)) {
            PyErr_SetString(PyExc_OverflowError,
                            "c-string too long to convert to Python");
            return NULL;
        }
        length = (Py_ssize_t) slen;
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* CallNextTpDealloc */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc) {
    PyTypeObject* type = Py_TYPE(obj);
    destructor tp_dealloc = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_dealloc, destructor) != current_tp_dealloc)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_dealloc = __Pyx_PyType_GetSlot(type, tp_dealloc, destructor)) == current_tp_dealloc)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type)
        tp_dealloc(obj);
}

/* CallNextTpTraverse */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    traverseproc tp_traverse = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc) != current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_traverse = __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc)) == current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_traverse)
        return tp_traverse(obj, v, a);
    return 0;
}

/* CallTypeTraverse */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg) {
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x03090000
    if (__Pyx_get_runtime_version() < 0x03090000) return 0;
    #endif
    if (!always_call) {
        PyTypeObject *base = __Pyx_PyObject_GetSlot(o, tp_base, PyTypeObject*);
        unsigned long flags = PyType_GetFlags(base);
        if (flags & Py_TPFLAGS_HEAPTYPE) {
            return 0;
        }
    }
    Py_VISIT((PyObject*)Py_TYPE(o));
    return 0;
}
#endif

/* CallNextTpClear */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    inquiry tp_clear = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_clear, inquiry) != current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_clear = __Pyx_PyType_GetSlot(type, tp_clear, inquiry)) == current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_clear)
        tp_clear(obj);
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_3_2_4
#define __PYX_HAVE_RT_ImportType_3_2_4
static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size)
{
    PyObject *result = 0;
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API)
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !( defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API) )
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    if (size == 0) {
        return (PyTypeObject *)result;
    }
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_2_4 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_2_4 && (size_t)basicsize > size) {
        if (PyErr_WarnFormat(NULL, 0,
                "%.200s.%.200s size changed, may indicate binary incompatibility. "
                "Expected %zd from C header, got %zd from PyObject",
                module_name, class_name, size, basicsize) < 0) {
            goto bad;
        }
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* LimitedApiGetTypeDict (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static Py_ssize_t __Pyx_GetTypeDictOffset(void) {
    PyObject *tp_dictoffset_o;
    Py_ssize_t tp_dictoffset;
    tp_dictoffset_o = PyObject_GetAttrString((PyObject*)(&PyType_Type), "__dictoffset__");
    if (unlikely(!tp_dictoffset_o)) return -1;
    tp_dictoffset = PyLong_AsSsize_t(tp_dictoffset_o);
    Py_DECREF(tp_dictoffset_o);
    if (unlikely(tp_dictoffset == 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' doesn't have a dictoffset");
        return -1;
    } else if (unlikely(tp_dictoffset < 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' has an unexpected negative dictoffset. "
            "Please report this as Cython bug");
        return -1;
    }
    return tp_dictoffset;
}
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp) {
    static Py_ssize_t tp_dictoffset = 0;
    if (unlikely(tp_dictoffset == 0)) {
        tp_dictoffset = __Pyx_GetTypeDictOffset();
        if (unlikely(tp_dictoffset == -1 && PyErr_Occurred())) {
            tp_dictoffset = 0; // try again next time?
            return NULL;
        }
    }
    return *(PyObject**)((char*)tp + tp_dictoffset);
}
#endif

/* SetItemOnTypeDict (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_SetItem(tp_dict, k, v);
    if (likely(!result)) {
        PyType_Modified(tp);
        if (unlikely(PyObject_HasAttr(v, __pyx_mstate_global->__pyx_n_u_set_name))) {
            PyObject *setNameResult = PyObject_CallMethodObjArgs(v, __pyx_mstate_global->__pyx_n_u_set_name,  (PyObject *) tp, k, NULL);
            if (!setNameResult) return -1;
            Py_DECREF(setNameResult);
        }
    }
    return result;
}

/* FixUpExtensionType */
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if __PYX_LIMITED_VERSION_HEX > 0x030900B1
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#else
    const PyType_Slot *slot = spec->slots;
    int changed = 0;
#if !CYTHON_COMPILING_IN_LIMITED_API
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
#if !CYTHON_COMPILING_IN_CPYTHON
        const
#endif  // !CYTHON_COMPILING_IN_CPYTHON)
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_vectorcall_offset = memb->offset;
                    changed = 1;
                }
#endif  // CYTHON_METH_FASTCALL
#if !CYTHON_COMPILING_IN_PYPY
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                    Py_DECREF(descr);
                    if (unlikely(set_item_result < 0)) {
                        return -1;
                    }
                    changed = 1;
                }
#endif  // !CYTHON_COMPILING_IN_PYPY
            }
            memb++;
        }
    }
#endif  // !CYTHON_COMPILING_IN_LIMITED_API
#if !CYTHON_COMPILING_IN_PYPY
    slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_getset)
        slot++;
    if (slot && slot->slot == Py_tp_getset) {
        PyGetSetDef *getset = (PyGetSetDef*) slot->pfunc;
        while (getset && getset->name) {
            if (getset->name[0] == '_' && getset->name[1] == '_' && strcmp(getset->name, "__module__") == 0) {
                PyObject *descr = PyDescr_NewGetSet(type, getset);
                if (unlikely(!descr))
                    return -1;
                #if CYTHON_COMPILING_IN_LIMITED_API
                PyObject *pyname = PyUnicode_FromString(getset->name);
                if (unlikely(!pyname)) {
                    Py_DECREF(descr);
                    return -1;
                }
                int set_item_result = __Pyx_SetItemOnTypeDict(type, pyname, descr);
                Py_DECREF(pyname);
                #else
                CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
                int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                #endif
                Py_DECREF(descr);
                if (unlikely(set_item_result < 0)) {
                    return -1;
                }
                changed = 1;
            }
            ++getset;
        }
    }
#else
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#endif  // !CYTHON_COMPILING_IN_PYPY
    if (changed)
        PyType_Modified(type);
#endif  // PY_VERSION_HEX > 0x030900B1
    return 0;
}

/* PyObjectCallNoArg (used by PyObjectCallMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod (used by PyObjectCallMethod0) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(tp);
    PyErr_Format(PyExc_AttributeError,
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}
#endif

/* PyObjectCallMethod0 (used by PyType_Ready) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
#if CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000))
    PyObject *args[1] = {obj};
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_CallNoArg;
    return PyObject_VectorcallMethod(method_name, args, 1 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
#endif
}

/* ValidateBasesTuple (used by PyType_Ready) */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_SIZE
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (unlikely(n < 0)) return -1;
#endif
    for (i = 1; i < n; i++)
    {
        PyTypeObject *b;
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !CYTHON_USE_TYPE_SLOTS
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
CYTHON_UNUSED static int __Pyx_PyType_HasMultipleInheritance(PyTypeObject *t) {
    while (t) {
        PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
        if (bases) {
            return 1;
        }
        t = __Pyx_PyType_GetSlot(t, tp_base, PyTypeObject*);
    }
    return 0;
}
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !CYTHON_COMPILING_IN_CPYTHON || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    if (!__Pyx_PyType_HasMultipleInheritance(t)) {
        return PyType_Ready(t);
    }
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400) &&\
                !CYTHON_COMPILING_IN_GRAAL
        gc = PyImport_GetModule(__pyx_mstate_global->__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_mstate_global->__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* SetVTable */
static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* MergeVTables */
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i=0;
    Py_ssize_t size;
    void** base_vtables;
    __Pyx_TypeName tp_base_name = NULL;
    __Pyx_TypeName base_name = NULL;
    void* unknown = (void*)-1;
    PyObject* bases = __Pyx_PyType_GetSlot(type, tp_bases, PyObject*);
    int base_depth = 0;
    {
        PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        while (base) {
            base_depth += 1;
            base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
        }
    }
    base_vtables = (void**) PyMem_Malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
#if CYTHON_COMPILING_IN_LIMITED_API
    size = PyTuple_Size(bases);
    if (size < 0) goto other_failure;
#else
    size = PyTuple_GET_SIZE(bases);
#endif
    for (i = 1; i < size; i++) {
        PyObject *basei;
        void* base_vtable;
#if CYTHON_AVOID_BORROWED_REFS
        basei = PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#else
        basei = PyTuple_GET_ITEM(bases, i);
#endif
        base_vtable = __Pyx_GetVtable((PyTypeObject*)basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
            }
        }
    }
    PyErr_Clear();
    PyMem_Free(base_vtables);
    return 0;
bad:
    {
        PyTypeObject* basei = NULL;
        PyTypeObject* tp_base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        tp_base_name = __Pyx_PyType_GetFullyQualifiedName(tp_base);
#if CYTHON_AVOID_BORROWED_REFS
        basei = (PyTypeObject*)PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = (PyTypeObject*)PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#else
        basei = (PyTypeObject*)PyTuple_GET_ITEM(bases, i);
#endif
        base_name = __Pyx_PyType_GetFullyQualifiedName(basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
    }
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
really_bad: // bad has failed!
#endif
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
other_failure:
#endif
    PyMem_Free(base_vtables);
    return -1;
}

/* DelItemOnTypeDict (used by SetupReduce) */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_DelItem(tp_dict, k);
    if (likely(!result)) PyType_Modified(tp);
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_mstate_global->__pyx_n_u_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_mstate_global->__pyx_n_u_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython);
            if (likely(reduce_cython)) {
                ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_mstate_global->__pyx_n_u_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetFullyQualifiedName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* PxdImportShared (used by FunctionImport) */
#ifndef __PYX_HAVE_RT_ImportFromPxd_3_2_4
#define __PYX_HAVE_RT_ImportFromPxd_3_2_4
static int __Pyx_ImportFromPxd_3_2_4(PyObject *module, const char *name, void **p, const char *sig, const char *what) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, "__pyx_capi__");
    if (!d)
        goto bad;
#if (defined(Py_LIMITED_API) && Py_LIMITED_API >= 0x030d0000) || (!defined(Py_LIMITED_API) && PY_VERSION_HEX >= 0x030d0000)
    PyDict_GetItemStringRef(d, name, &cobj);
#else
    cobj = PyDict_GetItemString(d, name);
    Py_XINCREF(cobj);
#endif
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C %.8s %.200s",
                PyModule_GetName(module), what, name);
        goto bad;
    }
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C %.8s %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             what, PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    Py_DECREF(cobj);
    return 0;
bad:
    Py_XDECREF(d);
    Py_XDECREF(cobj);
    return -1;
}
#endif

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction_3_2_4
#define __PYX_HAVE_RT_ImportFunction_3_2_4
static int __Pyx_ImportFunction_3_2_4(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    int result = __Pyx_ImportFromPxd_3_2_4(module, funcname, &tmp.p, sig, "function");
    if (result == 0) {
        *f = tmp.fp;
    }
    return result;
}
#endif

/* dict_setdefault (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value) {
    PyObject* value;
#if __PYX_LIMITED_VERSION_HEX >= 0x030F0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4)
    PyDict_SetDefaultRef(d, key, default_value, &value);
#elif CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    PyObject *args[] = {d, key, default_value};
    value = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_n_u_setdefault, args, 3 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#elif CYTHON_COMPILING_IN_LIMITED_API
    value = PyObject_CallMethodObjArgs(d, __pyx_mstate_global->__pyx_n_u_setdefault, key, default_value, NULL);
#else
    value = PyDict_SetDefault(d, key, default_value);
    if (unlikely(!value)) return NULL;
    Py_INCREF(value);
#endif
    return value;
}

/* AddModuleRef (used by FetchSharedCythonModule) */
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  static PyObject *__Pyx_PyImport_AddModuleObjectRef(PyObject *name) {
      PyObject *module_dict = PyImport_GetModuleDict();
      PyObject *m;
      if (PyMapping_GetOptionalItem(module_dict, name, &m) < 0) {
          return NULL;
      }
      if (m != NULL && PyModule_Check(m)) {
          return m;
      }
      Py_XDECREF(m);
      m = PyModule_NewObject(name);
      if (m == NULL)
          return NULL;
      if (PyDict_CheckExact(module_dict)) {
          PyObject *new_m;
          (void)PyDict_SetDefaultRef(module_dict, name, m, &new_m);
          Py_DECREF(m);
          return new_m;
      } else {
           if (PyObject_SetItem(module_dict, name, m) != 0) {
                Py_DECREF(m);
                return NULL;
            }
            return m;
      }
  }
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *py_name = PyUnicode_FromString(name);
      if (!py_name) return NULL;
      PyObject *module = __Pyx_PyImport_AddModuleObjectRef(py_name);
      Py_DECREF(py_name);
      return module;
  }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif

/* FetchSharedCythonModule (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef(__PYX_ABI_MODULE_NAME);
}

/* FetchCommonType (used by CommonTypesMetaclass) */
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject* __Pyx_PyType_FromMetaclass(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *result = __Pyx_PyType_FromModuleAndSpec(module, spec, bases);
    if (result && metaclass) {
        PyObject *old_tp = (PyObject*)Py_TYPE(result);
    Py_INCREF((PyObject*)metaclass);
#if __PYX_LIMITED_VERSION_HEX >= 0x03090000
        Py_SET_TYPE(result, metaclass);
#else
        result->ob_type = metaclass;
#endif
        Py_DECREF(old_tp);
    }
    return result;
}
#else
#define __Pyx_PyType_FromMetaclass(me, mo, s, b) PyType_FromMetaclass(me, mo, s, b)
#endif
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t expected_basicsize) {
    Py_ssize_t basicsize;
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (expected_basicsize == 0) {
        return 0; // size is inherited, nothing useful to check
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
    if (unlikely(!py_basicsize)) return -1;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = NULL;
    if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) return -1;
#else
    basicsize = ((PyTypeObject*) cached_type)->tp_basicsize;
#endif
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module = NULL, *cached_type = NULL, *abi_module_dict, *new_cached_type, *py_object_name;
    int get_item_ref_result;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    py_object_name = PyUnicode_FromString(object_name);
    if (!py_object_name) return NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) goto done;
    abi_module_dict = PyModule_GetDict(abi_module);
    if (!abi_module_dict) goto done;
    get_item_ref_result = __Pyx_PyDict_GetItemRef(abi_module_dict, py_object_name, &cached_type);
    if (get_item_ref_result == 1) {
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else if (unlikely(get_item_ref_result == -1)) {
        goto bad;
    }
    cached_type = __Pyx_PyType_FromMetaclass(
        metaclass,
        CYTHON_USE_MODULE_STATE ? module : abi_module,
        spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    new_cached_type = __Pyx_PyDict_SetDefault(abi_module_dict, py_object_name, cached_type);
    if (unlikely(new_cached_type != cached_type)) {
        if (unlikely(!new_cached_type)) goto bad;
        Py_DECREF(cached_type);
        cached_type = new_cached_type;
        if (__Pyx_VerifyCachedType(
                cached_type,
                object_name,
                spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else {
        Py_DECREF(new_cached_type);
    }
done:
    Py_XDECREF(abi_module);
    Py_DECREF(py_object_name);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CommonTypesMetaclass (used by CythonFunctionShared) */
static PyObject* __pyx_CommonTypesMetaclass_get_module(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED void* context) {
    return PyUnicode_FromString(__PYX_ABI_MODULE_NAME);
}
#if __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject* __pyx_CommonTypesMetaclass_call(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *args, CYTHON_UNUSED PyObject *kwds) {
    PyErr_SetString(PyExc_TypeError, "Cannot instantiate Cython internal types");
    return NULL;
}
static int __pyx_CommonTypesMetaclass_setattr(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *attr, CYTHON_UNUSED PyObject *value) {
    PyErr_SetString(PyExc_TypeError, "Cython internal types are immutable");
    return -1;
}
#endif
static PyGetSetDef __pyx_CommonTypesMetaclass_getset[] = {
    {"__module__", __pyx_CommonTypesMetaclass_get_module, NULL, NULL, NULL},
    {0, 0, 0, 0, 0}
};
static PyType_Slot __pyx_CommonTypesMetaclass_slots[] = {
    {Py_tp_getset, (void *)__pyx_CommonTypesMetaclass_getset},
    #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {Py_tp_call, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_new, (void*)__pyx_CommonTypesMetaclass_call},
    {Py_tp_setattro, (void*)__pyx_CommonTypesMetaclass_setattr},
    #endif
    {0, 0}
};
static PyType_Spec __pyx_CommonTypesMetaclass_spec = {
    __PYX_TYPE_MODULE_PREFIX "_common_types_metatype",
    0,
    0,
    Py_TPFLAGS_IMMUTABLETYPE |
    Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT,
    __pyx_CommonTypesMetaclass_slots
};
static int __pyx_CommonTypesMetaclass_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    PyObject *bases = PyTuple_Pack(1, &PyType_Type);
    if (unlikely(!bases)) {
        return -1;
    }
    mstate->__pyx_CommonTypesMetaclassType = __Pyx_FetchCommonTypeFromSpec(NULL, module, &__pyx_CommonTypesMetaclass_spec, bases);
    Py_DECREF(bases);
    if (unlikely(mstate->__pyx_CommonTypesMetaclassType == NULL)) {
        return -1;
    }
    return 0;
}

/* PyMethodNew (used by CythonFunctionShared) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *result;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    #if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    {
        PyObject *args[] = {func, self};
        result = PyObject_Vectorcall(__pyx_mstate_global->__Pyx_CachedMethodType, args, 2, NULL);
    }
    #else
    result = PyObject_CallFunctionObjArgs(__pyx_mstate_global->__Pyx_CachedMethodType, func, self, NULL);
    #endif
    return result;
}
#else
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#endif

/* PyVectorcallFastCallDict (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i;
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos;
    #else
    Py_ssize_t pos;
    #endif
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    #if !CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t nkw = PyDict_Size(kw);
    if (unlikely(nkw == -1)) return NULL;
    #else
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    #endif
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = 0;
    i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (__Pyx_PyDict_NextRef(kw, &pos, &key, &value)) {
        keys_are_strings &=
        #if CYTHON_COMPILING_IN_LIMITED_API
            PyType_GetFlags(Py_TYPE(key));
        #else
            Py_TYPE(key)->tp_flags;
        #endif
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(PyTuple_SetItem(kwnames, i, key) < 0)) goto cleanup;
        #else
        PyTuple_SET_ITEM(kwnames, i, key);
        #endif
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(pos);
    #endif
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    Py_ssize_t kw_size =
        likely(kw == NULL) ?
        0 :
#if !CYTHON_ASSUME_SAFE_SIZE
        PyDict_Size(kw);
#else
        PyDict_GET_SIZE(kw);
#endif
    if (kw_size == 0) {
        return vc(func, args, nargs, NULL);
    }
#if !CYTHON_ASSUME_SAFE_SIZE
    else if (unlikely(kw_size == -1)) {
        return NULL;
    }
#endif
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared (used by CythonFunction) */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunctionNoMethod(PyObject *func, void (*cfunc)(void)) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if ((PyObject*)Py_TYPE(func) == __pyx_mstate_global->__Pyx_CachedMethodType) {
        int result;
        PyObject *newFunc = PyObject_GetAttr(func, __pyx_mstate_global->__pyx_n_u_func);
        if (unlikely(!newFunc)) {
            PyErr_Clear(); // It's only an optimization, so don't throw an error
            return 0;
        }
        result = __Pyx__IsSameCyOrCFunctionNoMethod(newFunc, cfunc);
        Py_DECREF(newFunc);
        return result;
    }
    return __Pyx__IsSameCyOrCFunctionNoMethod(func, cfunc);
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if (PyMethod_Check(func)) {
        func = PyMethod_GET_FUNCTION(func);
    }
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure) {
    PyObject *result;
    CYTHON_UNUSED_VAR(closure);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_doc_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#else
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_name_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    PyObject *result;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    Py_INCREF(op->func_qualname);
    result = op->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
#endif
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_defaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_kwdefaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_annotations_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine_value(__pyx_CyFunctionObject *op) {
    int is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
    if (is_coroutine) {
        PyObject *is_coroutine_value, *module, *fromlist, *marker = __pyx_mstate_global->__pyx_n_u_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_mstate_global->__pyx_n_u_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        is_coroutine_value = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(is_coroutine_value)) {
            return is_coroutine_value;
        }
ignore:
        PyErr_Clear();
    }
    return __Pyx_PyBool_FromLong(is_coroutine);
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    result = __Pyx_CyFunction_get_is_coroutine_value(op);
    if (unlikely(!result))
        return NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    if (op->func_is_coroutine) {
        Py_DECREF(result);
        result = __Pyx_NewRef(op->func_is_coroutine);
    } else {
        op->func_is_coroutine = __Pyx_NewRef(result);
    }
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static void __Pyx_CyFunction_raise_argument_count_error(__pyx_CyFunctionObject *func, const char* message, Py_ssize_t size) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        py_name, message, size);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        name, message, size);
#endif
}
static void __Pyx_CyFunction_raise_type_error(__pyx_CyFunctionObject *func, const char* message) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s",
        py_name, message);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s",
        name, message);
#endif
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {"func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
    {"func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
#else
    {"func_dict", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
    {"__dict__", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
#endif
    {"func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {"__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {"_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    {"__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#endif
#if CYTHON_METH_FASTCALL
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {"__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(args);
    __Pyx_BEGIN_CRITICAL_SECTION(m);
    Py_INCREF(m->func_qualname);
    result = m->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_dict = NULL;
#endif
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func_dict);
#elif PY_VERSION_HEX < 0x030d0000
    _PyObject_ClearManagedDict((PyObject*)m);
#else
    PyObject_ClearManagedDict((PyObject*)m);
#endif
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    Py_CLEAR(m->defaults);
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    {
        int e = __Pyx_call_type_traverse((PyObject*)m, 1, visit, arg);
        if (e) return e;
    }
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func_dict);
#else
    {
        int e =
#if PY_VERSION_HEX < 0x030d0000
            _PyObject_VisitManagedDict
#else
            PyObject_VisitManagedDict
#endif
                ((PyObject*)m, visit, arg);
        if (e != 0) return e;
    }
#endif
    __Pyx_VISIT_CONST(m->func_name);
    __Pyx_VISIT_CONST(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    __Pyx_VISIT_CONST(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    Py_VISIT(m->defaults);
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
    PyObject *repr;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    repr = PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
    __Pyx_END_CRITICAL_SECTION();
    return repr;
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void(*)(void))meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes no arguments", size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes exactly one argument", size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
    __Pyx_CyFunction_raise_type_error(
        (__pyx_CyFunctionObject*)func, "takes no keyword arguments");
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_SIZE
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(argc < 0)) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            __Pyx_CyFunction_raise_type_error(
                cyfunc, "needs an argument");
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(__Pyx_PyTuple_GET_SIZE(kwnames))) {
        __Pyx_CyFunction_raise_type_error(
            cyfunc, "takes no keyword arguments");
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes no arguments", nargs);
        return NULL;
    }
    return meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes exactly one argument", nargs);
        return NULL;
    }
    return meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    #if PY_VERSION_HEX < 0x030e00A6
    size_t nargs_value = (size_t) nargs;
    #else
    Py_ssize_t nargs_value = nargs;
    #endif
    return ((__Pyx_PyCMethod)(void(*)(void))meth)(self, cls, args, nargs_value, kwnames);
}
#endif
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if CYTHON_METH_FASTCALL
#if defined(Py_TPFLAGS_HAVE_VECTORCALL)
    Py_TPFLAGS_HAVE_VECTORCALL |
#elif defined(_Py_TPFLAGS_HAVE_VECTORCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
#endif // CYTHON_METH_FASTCALL
#if PY_VERSION_HEX >= 0x030C0000 && !CYTHON_COMPILING_IN_LIMITED_API
    Py_TPFLAGS_MANAGED_DICT |
#endif
    Py_TPFLAGS_IMMUTABLETYPE | Py_TPFLAGS_DISALLOW_INSTANTIATION |
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
static int __pyx_CyFunction_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    mstate->__pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(
        mstate->__pyx_CommonTypesMetaclassType, module, &__pyx_CyFunctionType_spec, NULL);
    if (unlikely(mstate->__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func, PyTypeObject *defaults_type) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_CallObject((PyObject*)defaults_type, NULL); // _PyObject_New(defaults_type);
    if (unlikely(!m->defaults))
        return NULL;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_mstate_global->__pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* CLineInTraceback (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
#define __Pyx_PyProbablyModule_GetDict(o) __Pyx_XNewRef(PyModule_GetDict(o))
#elif !CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyProbablyModule_GetDict(o) PyObject_GenericGetDict(o, NULL);
#else
PyObject* __Pyx_PyProbablyModule_GetDict(PyObject *o) {
    PyObject **dict_ptr = _PyObject_GetDictPtr(o);
    return dict_ptr ? __Pyx_XNewRef(*dict_ptr) : NULL;
}
#endif
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline = NULL;
    PyObject *ptype, *pvalue, *ptraceback;
    PyObject *cython_runtime_dict;
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_mstate_global->__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
    cython_runtime_dict = __Pyx_PyProbablyModule_GetDict(__pyx_mstate_global->__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, cython_runtime_dict,
            __Pyx_PyDict_SetDefault(cython_runtime_dict, __pyx_mstate_global->__pyx_n_u_cline_in_traceback, Py_False))
    }
    if (use_cline == NULL || use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    Py_XDECREF(use_cline);
    Py_XDECREF(cython_runtime_dict);
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache (used by AddTraceback) */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static __Pyx_CachedCodeObjectType *__pyx__find_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line) {
    __Pyx_CachedCodeObjectType* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!code_cache->entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if (unlikely(pos >= code_cache->count) || unlikely(code_cache->entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = code_cache->entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__find_code_object;
    return NULL; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just miss.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type old_count = __pyx_atomic_incr_acq_rel(&code_cache->accessor_count);
    if (old_count < 0) {
        __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
        return NULL;
    }
#endif
    __Pyx_CachedCodeObjectType *result = __pyx__find_code_object(code_cache, code_line);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
#endif
    return result;
#endif
}
static void __pyx__insert_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line, __Pyx_CachedCodeObjectType* code_object)
{
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = code_cache->entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            code_cache->entries = entries;
            code_cache->max_count = 64;
            code_cache->count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if ((pos < code_cache->count) && unlikely(code_cache->entries[pos].code_line == code_line)) {
        __Pyx_CachedCodeObjectType* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_INCREF(code_object);
        Py_DECREF(tmp);
        return;
    }
    if (code_cache->count == code_cache->max_count) {
        int new_max = code_cache->max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            code_cache->entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        code_cache->entries = entries;
        code_cache->max_count = new_max;
    }
    for (i=code_cache->count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    code_cache->count++;
    Py_INCREF(code_object);
}
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__insert_code_object;
    return; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just fail.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type expected = 0;
    if (!__pyx_atomic_int_cmp_exchange(&code_cache->accessor_count, &expected, INT_MIN)) {
        return;
    }
#endif
    __pyx__insert_code_object(code_cache, code_line, code_object);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_sub(&code_cache->accessor_count, INT_MIN);
#endif
#endif
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result = PyObject_Call(replace, __pyx_mstate_global->__pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!code_object) {
        code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
        if (unlikely(!code_object)) goto bad;
        py_py_line = PyLong_FromLong(py_line);
        if (unlikely(!py_py_line)) goto bad;
        if (c_line) {
            py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        } else {
            py_funcname = PyUnicode_FromString(funcname);
        }
        if (unlikely(!py_funcname)) goto bad;
        dict = PyDict_New();
        if (unlikely(!dict)) goto bad;
        {
            PyObject *old_code_object = code_object;
            code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
            Py_DECREF(old_code_object);
        }
        if (unlikely(!code_object)) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, code_object);
    } else {
        dict = PyDict_New();
    }
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    if (c_line) {
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
    }
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_mstate_global->__pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE uint8_t __Pyx_PyLong_As_uint8_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint8_t neg_one = (uint8_t) -1, const_zero = (uint8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        uint8_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (uint8_t) -1;
        val = __Pyx_PyLong_As_uint8_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint8_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(uint8_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) >= 2 * PyLong_SHIFT)) {
                            return (uint8_t) (((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint8_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) >= 3 * PyLong_SHIFT)) {
                            return (uint8_t) (((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint8_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) >= 4 * PyLong_SHIFT)) {
                            return (uint8_t) (((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (uint8_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(uint8_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint8_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(uint8_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint8_t) (((uint8_t)-1)*(((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(uint8_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint8_t) ((((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint8_t) (((uint8_t)-1)*(((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint8_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint8_t) ((((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint8_t) (((uint8_t)-1)*(((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint8_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint8_t) ((((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(uint8_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint8_t, long, PyLong_AsLong(x))
        } else if ((sizeof(uint8_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint8_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        uint8_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (uint8_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (uint8_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (uint8_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (uint8_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(uint8_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((uint8_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(uint8_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((uint8_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((uint8_t) 1) << (sizeof(uint8_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (uint8_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint8_t");
    return (uint8_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint8_t");
    return (uint8_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE uint32_t __Pyx_PyLong_As_uint32_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        uint32_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (uint32_t) -1;
        val = __Pyx_PyLong_As_uint32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint32_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(uint32_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) >= 2 * PyLong_SHIFT)) {
                            return (uint32_t) (((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint32_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) >= 3 * PyLong_SHIFT)) {
                            return (uint32_t) (((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint32_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) >= 4 * PyLong_SHIFT)) {
                            return (uint32_t) (((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (uint32_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(uint32_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint32_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(uint32_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint32_t) (((uint32_t)-1)*(((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(uint32_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint32_t) ((((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint32_t) (((uint32_t)-1)*(((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint32_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint32_t) ((((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint32_t) (((uint32_t)-1)*(((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint32_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint32_t) ((((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(uint32_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, long, PyLong_AsLong(x))
        } else if ((sizeof(uint32_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        uint32_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (uint32_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (uint32_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (uint32_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (uint32_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(uint32_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((uint32_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(uint32_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((uint32_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((uint32_t) 1) << (sizeof(uint32_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (uint32_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint32_t");
    return (uint32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint32_t");
    return (uint32_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_uint32_t(uint32_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint32_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(uint32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint32_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(uint32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(uint32_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint32_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_uint8_t(uint8_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint8_t neg_one = (uint8_t) -1, const_zero = (uint8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint8_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(uint8_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint8_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(uint8_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(uint8_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint8_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyLong_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyLong_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 2 * PyLong_SHIFT)) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 3 * PyLong_SHIFT)) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 4 * PyLong_SHIFT)) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (size_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(size_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(size_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
        } else if ((sizeof(size_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        size_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (size_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (size_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (size_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (size_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(size_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((size_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(size_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((size_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((size_t) 1) << (sizeof(size_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (size_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyLong_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From___pyx_anon_enum(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE enum rdma_cm_event_type __Pyx_PyLong_As_enum__rdma_cm_event_type(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum rdma_cm_event_type neg_one = (enum rdma_cm_event_type) -1, const_zero = (enum rdma_cm_event_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        enum rdma_cm_event_type val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (enum rdma_cm_event_type) -1;
        val = __Pyx_PyLong_As_enum__rdma_cm_event_type(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(enum rdma_cm_event_type, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(enum rdma_cm_event_type) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum rdma_cm_event_type, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum rdma_cm_event_type) >= 2 * PyLong_SHIFT)) {
                            return (enum rdma_cm_event_type) (((((enum rdma_cm_event_type)digits[1]) << PyLong_SHIFT) | (enum rdma_cm_event_type)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(enum rdma_cm_event_type) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum rdma_cm_event_type, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum rdma_cm_event_type) >= 3 * PyLong_SHIFT)) {
                            return (enum rdma_cm_event_type) (((((((enum rdma_cm_event_type)digits[2]) << PyLong_SHIFT) | (enum rdma_cm_event_type)digits[1]) << PyLong_SHIFT) | (enum rdma_cm_event_type)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(enum rdma_cm_event_type) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum rdma_cm_event_type, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum rdma_cm_event_type) >= 4 * PyLong_SHIFT)) {
                            return (enum rdma_cm_event_type) (((((((((enum rdma_cm_event_type)digits[3]) << PyLong_SHIFT) | (enum rdma_cm_event_type)digits[2]) << PyLong_SHIFT) | (enum rdma_cm_event_type)digits[1]) << PyLong_SHIFT) | (enum rdma_cm_event_type)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (enum rdma_cm_event_type) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(enum rdma_cm_event_type) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum rdma_cm_event_type, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(enum rdma_cm_event_type) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum rdma_cm_event_type, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(enum rdma_cm_event_type, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(enum rdma_cm_event_type) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum rdma_cm_event_type, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum rdma_cm_event_type) - 1 > 2 * PyLong_SHIFT)) {
                            return (enum rdma_cm_event_type) (((enum rdma_cm_event_type)-1)*(((((enum rdma_cm_event_type)digits[1]) << PyLong_SHIFT) | (enum rdma_cm_event_type)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(enum rdma_cm_event_type) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum rdma_cm_event_type, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum rdma_cm_event_type) - 1 > 2 * PyLong_SHIFT)) {
                            return (enum rdma_cm_event_type) ((((((enum rdma_cm_event_type)digits[1]) << PyLong_SHIFT) | (enum rdma_cm_event_type)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(enum rdma_cm_event_type) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum rdma_cm_event_type, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum rdma_cm_event_type) - 1 > 3 * PyLong_SHIFT)) {
                            return (enum rdma_cm_event_type) (((enum rdma_cm_event_type)-1)*(((((((enum rdma_cm_event_type)digits[2]) << PyLong_SHIFT) | (enum rdma_cm_event_type)digits[1]) << PyLong_SHIFT) | (enum rdma_cm_event_type)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(enum rdma_cm_event_type) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum rdma_cm_event_type, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum rdma_cm_event_type) - 1 > 3 * PyLong_SHIFT)) {
                            return (enum rdma_cm_event_type) ((((((((enum rdma_cm_event_type)digits[2]) << PyLong_SHIFT) | (enum rdma_cm_event_type)digits[1]) << PyLong_SHIFT) | (enum rdma_cm_event_type)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(enum rdma_cm_event_type) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum rdma_cm_event_type, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum rdma_cm_event_type) - 1 > 4 * PyLong_SHIFT)) {
                            return (enum rdma_cm_event_type) (((enum rdma_cm_event_type)-1)*(((((((((enum rdma_cm_event_type)digits[3]) << PyLong_SHIFT) | (enum rdma_cm_event_type)digits[2]) << PyLong_SHIFT) | (enum rdma_cm_event_type)digits[1]) << PyLong_SHIFT) | (enum rdma_cm_event_type)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(enum rdma_cm_event_type) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum rdma_cm_event_type, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum rdma_cm_event_type) - 1 > 4 * PyLong_SHIFT)) {
                            return (enum rdma_cm_event_type) ((((((((((enum rdma_cm_event_type)digits[3]) << PyLong_SHIFT) | (enum rdma_cm_event_type)digits[2]) << PyLong_SHIFT) | (enum rdma_cm_event_type)digits[1]) << PyLong_SHIFT) | (enum rdma_cm_event_type)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(enum rdma_cm_event_type) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum rdma_cm_event_type, long, PyLong_AsLong(x))
        } else if ((sizeof(enum rdma_cm_event_type) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum rdma_cm_event_type, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        enum rdma_cm_event_type val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() or PyLong_AsNativeBytes() not available, cannot convert large enums");
        val = (enum rdma_cm_event_type) -1;
#endif
        if (unlikely(ret))
            return (enum rdma_cm_event_type) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum rdma_cm_event_type");
    return (enum rdma_cm_event_type) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum rdma_cm_event_type");
    return (enum rdma_cm_event_type) -1;
}

/* CIntFromPy */
static CYTHON_INLINE enum rdma_port_space __Pyx_PyLong_As_enum__rdma_port_space(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum rdma_port_space neg_one = (enum rdma_port_space) -1, const_zero = (enum rdma_port_space) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        enum rdma_port_space val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (enum rdma_port_space) -1;
        val = __Pyx_PyLong_As_enum__rdma_port_space(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(enum rdma_port_space, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(enum rdma_port_space) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum rdma_port_space, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum rdma_port_space) >= 2 * PyLong_SHIFT)) {
                            return (enum rdma_port_space) (((((enum rdma_port_space)digits[1]) << PyLong_SHIFT) | (enum rdma_port_space)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(enum rdma_port_space) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum rdma_port_space, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum rdma_port_space) >= 3 * PyLong_SHIFT)) {
                            return (enum rdma_port_space) (((((((enum rdma_port_space)digits[2]) << PyLong_SHIFT) | (enum rdma_port_space)digits[1]) << PyLong_SHIFT) | (enum rdma_port_space)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(enum rdma_port_space) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum rdma_port_space, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum rdma_port_space) >= 4 * PyLong_SHIFT)) {
                            return (enum rdma_port_space) (((((((((enum rdma_port_space)digits[3]) << PyLong_SHIFT) | (enum rdma_port_space)digits[2]) << PyLong_SHIFT) | (enum rdma_port_space)digits[1]) << PyLong_SHIFT) | (enum rdma_port_space)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (enum rdma_port_space) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(enum rdma_port_space) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum rdma_port_space, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(enum rdma_port_space) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum rdma_port_space, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(enum rdma_port_space, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(enum rdma_port_space) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum rdma_port_space, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum rdma_port_space) - 1 > 2 * PyLong_SHIFT)) {
                            return (enum rdma_port_space) (((enum rdma_port_space)-1)*(((((enum rdma_port_space)digits[1]) << PyLong_SHIFT) | (enum rdma_port_space)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(enum rdma_port_space) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum rdma_port_space, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum rdma_port_space) - 1 > 2 * PyLong_SHIFT)) {
                            return (enum rdma_port_space) ((((((enum rdma_port_space)digits[1]) << PyLong_SHIFT) | (enum rdma_port_space)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(enum rdma_port_space) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum rdma_port_space, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum rdma_port_space) - 1 > 3 * PyLong_SHIFT)) {
                            return (enum rdma_port_space) (((enum rdma_port_space)-1)*(((((((enum rdma_port_space)digits[2]) << PyLong_SHIFT) | (enum rdma_port_space)digits[1]) << PyLong_SHIFT) | (enum rdma_port_space)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(enum rdma_port_space) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum rdma_port_space, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum rdma_port_space) - 1 > 3 * PyLong_SHIFT)) {
                            return (enum rdma_port_space) ((((((((enum rdma_port_space)digits[2]) << PyLong_SHIFT) | (enum rdma_port_space)digits[1]) << PyLong_SHIFT) | (enum rdma_port_space)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(enum rdma_port_space) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum rdma_port_space, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum rdma_port_space) - 1 > 4 * PyLong_SHIFT)) {
                            return (enum rdma_port_space) (((enum rdma_port_space)-1)*(((((((((enum rdma_port_space)digits[3]) << PyLong_SHIFT) | (enum rdma_port_space)digits[2]) << PyLong_SHIFT) | (enum rdma_port_space)digits[1]) << PyLong_SHIFT) | (enum rdma_port_space)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(enum rdma_port_space) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(enum rdma_port_space, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(enum rdma_port_space) - 1 > 4 * PyLong_SHIFT)) {
                            return (enum rdma_port_space) ((((((((((enum rdma_port_space)digits[3]) << PyLong_SHIFT) | (enum rdma_port_space)digits[2]) << PyLong_SHIFT) | (enum rdma_port_space)digits[1]) << PyLong_SHIFT) | (enum rdma_port_space)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(enum rdma_port_space) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum rdma_port_space, long, PyLong_AsLong(x))
        } else if ((sizeof(enum rdma_port_space) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(enum rdma_port_space, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        enum rdma_port_space val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() or PyLong_AsNativeBytes() not available, cannot convert large enums");
        val = (enum rdma_port_space) -1;
#endif
        if (unlikely(ret))
            return (enum rdma_port_space) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum rdma_port_space");
    return (enum rdma_port_space) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum rdma_port_space");
    return (enum rdma_port_space) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_int(unsigned int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned int) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned int) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(unsigned int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE uint64_t __Pyx_PyLong_As_uint64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        uint64_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (uint64_t) -1;
        val = __Pyx_PyLong_As_uint64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint64_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(uint64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) >= 2 * PyLong_SHIFT)) {
                            return (uint64_t) (((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) >= 3 * PyLong_SHIFT)) {
                            return (uint64_t) (((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) >= 4 * PyLong_SHIFT)) {
                            return (uint64_t) (((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (uint64_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(uint64_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint64_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(uint64_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint64_t) (((uint64_t)-1)*(((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(uint64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint64_t) ((((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint64_t) (((uint64_t)-1)*(((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint64_t) ((((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint64_t) (((uint64_t)-1)*(((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint64_t) ((((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(uint64_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, long, PyLong_AsLong(x))
        } else if ((sizeof(uint64_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        uint64_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (uint64_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (uint64_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (uint64_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (uint64_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(uint64_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((uint64_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(uint64_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((uint64_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((uint64_t) 1) << (sizeof(uint64_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (uint64_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint64_t");
    return (uint64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint64_t");
    return (uint64_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_long(unsigned long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned long neg_one = (unsigned long) -1, const_zero = (unsigned long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(unsigned long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(unsigned long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
static __Pyx_TypeName
__Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp)
{
    PyObject *module = NULL, *name = NULL, *result = NULL;
    #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
    name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_qualname);
    #else
    name = PyType_GetQualName(tp);
    #endif
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) goto bad;
    module = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_module);
    if (unlikely(module == NULL) || unlikely(!PyUnicode_Check(module))) goto bad;
    if (PyUnicode_CompareWithASCIIString(module, "builtins") == 0) {
        result = name;
        name = NULL;
        goto done;
    }
    result = PyUnicode_FromFormat("%U.%U", module, name);
    if (unlikely(result == NULL)) goto bad;
  done:
    Py_XDECREF(name);
    Py_XDECREF(module);
    return result;
  bad:
    PyErr_Clear();
    if (name) {
        result = name;
        name = NULL;
    } else {
        result = __Pyx_NewRef(__pyx_mstate_global->__pyx_kp_u__7);
    }
    goto done;
}
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyLong_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* GetRuntimeVersion */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
void __Pyx_init_runtime_version(void) {
    if (__Pyx_cached_runtime_version == 0) {
        const char* rt_version = Py_GetVersion();
        unsigned long version = 0;
        unsigned long factor = 0x01000000UL;
        unsigned int digit = 0;
        int i = 0;
        while (factor) {
            while ('0' <= rt_version[i] && rt_version[i] <= '9') {
                digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
                ++i;
            }
            version += factor * digit;
            if (rt_version[i] != '.')
                break;
            digit = 0;
            factor >>= 8;
            ++i;
        }
        __Pyx_cached_runtime_version = version;
    }
}
#endif
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    return Py_Version & ~0xFFUL;
#else
    return __Pyx_cached_runtime_version;
#endif
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* NewCodeObj */
#if CYTHON_COMPILING_IN_LIMITED_API
    static PyObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                       PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                       PyObject *fv, PyObject *cell, PyObject* fn,
                                       PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
#elif PY_VERSION_HEX >= 0x030B0000
  static PyCodeObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                         PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                         PyObject *fv, PyObject *cell, PyObject* fn,
                                         PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, __pyx_mstate_global->__pyx_empty_bytes);
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030c00A1
    if (likely(result))
        result->_co_firsttraceable = 0;
    #endif
    return result;
  }
#elif !CYTHON_COMPILING_IN_PYPY
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
) {
    PyObject *code_obj = NULL, *varnames_tuple_dedup = NULL, *code_bytes = NULL;
    Py_ssize_t var_count = (Py_ssize_t) descr.nlocals;
    PyObject *varnames_tuple = PyTuple_New(var_count);
    if (unlikely(!varnames_tuple)) return NULL;
    for (Py_ssize_t i=0; i < var_count; i++) {
        Py_INCREF(varnames[i]);
        if (__Pyx_PyTuple_SET_ITEM(varnames_tuple, i, varnames[i]) != (0)) goto done;
    }
    #if CYTHON_COMPILING_IN_LIMITED_API
    varnames_tuple_dedup = PyDict_GetItem(tuple_dedup_map, varnames_tuple);
    if (!varnames_tuple_dedup) {
        if (unlikely(PyDict_SetItem(tuple_dedup_map, varnames_tuple, varnames_tuple) < 0)) goto done;
        varnames_tuple_dedup = varnames_tuple;
    }
    #else
    varnames_tuple_dedup = PyDict_SetDefault(tuple_dedup_map, varnames_tuple, varnames_tuple);
    if (unlikely(!varnames_tuple_dedup)) goto done;
    #endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(varnames_tuple_dedup);
    #endif
    if (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table != NULL && !CYTHON_COMPILING_IN_GRAAL) {
        Py_ssize_t line_table_length = __Pyx_PyBytes_GET_SIZE(line_table);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(line_table_length == -1)) goto done;
        #endif
        Py_ssize_t code_len = (line_table_length * 2 + 4) & ~3LL;
        code_bytes = PyBytes_FromStringAndSize(NULL, code_len);
        if (unlikely(!code_bytes)) goto done;
        char* c_code_bytes = PyBytes_AsString(code_bytes);
        if (unlikely(!c_code_bytes)) goto done;
        memset(c_code_bytes, 0, (size_t) code_len);
    }
    code_obj = (PyObject*) __Pyx__PyCode_New(
        (int) descr.argcount,
        (int) descr.num_posonly_args,
        (int) descr.num_kwonly_args,
        (int) descr.nlocals,
        0,
        (int) descr.flags,
        code_bytes ? code_bytes : __pyx_mstate_global->__pyx_empty_bytes,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        varnames_tuple_dedup,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        filename,
        funcname,
        (int) descr.first_line,
        (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table) ? line_table : __pyx_mstate_global->__pyx_empty_bytes
    );
done:
    Py_XDECREF(code_bytes);
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(varnames_tuple_dedup);
    #endif
    Py_DECREF(varnames_tuple);
    return code_obj;
}

/* DecompressString */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo) {
    PyObject *module = NULL, *decompress, *compressed_bytes, *decompressed;
    const char* module_name = algo == 3 ? "compression.zstd" : algo == 2 ? "bz2" : "zlib";
    PyObject *methodname = PyUnicode_FromString("decompress");
    if (unlikely(!methodname)) return NULL;
    #if __PYX_LIMITED_VERSION_HEX >= 0x030e0000
    if (algo == 3) {
        PyObject *fromlist = Py_BuildValue("[O]", methodname);
        if (unlikely(!fromlist)) goto bad;
        module = PyImport_ImportModuleLevel("compression.zstd", NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
    } else
    #endif
        module = PyImport_ImportModule(module_name);
    if (unlikely(!module)) goto import_failed;
    decompress = PyObject_GetAttr(module, methodname);
    if (unlikely(!decompress)) goto import_failed;
    {
        #ifdef __cplusplus
            char *memview_bytes = const_cast<char*>(s);
        #else
            #if defined(__clang__)
              #pragma clang diagnostic push
              #pragma clang diagnostic ignored "-Wcast-qual"
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic push
              #pragma GCC diagnostic ignored "-Wcast-qual"
            #endif
            char *memview_bytes = (char*) s;
            #if defined(__clang__)
              #pragma clang diagnostic pop
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic pop
            #endif
        #endif
        #if CYTHON_COMPILING_IN_LIMITED_API && !defined(PyBUF_READ)
        int memview_flags = 0x100;
        #else
        int memview_flags = PyBUF_READ;
        #endif
        compressed_bytes = PyMemoryView_FromMemory(memview_bytes, length, memview_flags);
    }
    if (unlikely(!compressed_bytes)) {
        Py_DECREF(decompress);
        goto bad;
    }
    decompressed = PyObject_CallFunctionObjArgs(decompress, compressed_bytes, NULL);
    Py_DECREF(compressed_bytes);
    Py_DECREF(decompress);
    Py_DECREF(module);
    Py_DECREF(methodname);
    return decompressed;
import_failed:
    PyErr_Format(PyExc_ImportError,
        "Failed to import '%.20s.decompress' - cannot initialise module strings. "
        "String compression was configured with the C macro 'CYTHON_COMPRESS_STRINGS=%d'.",
        module_name, algo);
bad:
    Py_XDECREF(module);
    Py_DECREF(methodname);
    return NULL;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    {
        const char* result;
        Py_ssize_t unicode_length;
        CYTHON_MAYBE_UNUSED_VAR(unicode_length); // only for __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (unlikely(PyArg_Parse(o, "s#", &result, length) < 0)) return NULL;
        #else
        result = PyUnicode_AsUTF8AndSize(o, length);
        #endif
        #if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        unicode_length = PyUnicode_GetLength(o);
        if (unlikely(unicode_length < 0)) return NULL;
        if (unlikely(unicode_length != *length)) {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
        #endif
        return result;
    }
#else
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
}
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
    if (PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
    if (PyByteArray_Check(o)) {
#if (CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS) || (CYTHON_COMPILING_IN_PYPY && (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE)))
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
#else
        *length = PyByteArray_Size(o);
        if (*length == -1) return NULL;
        return PyByteArray_AsString(o);
#endif
    } else
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_LongWrongResultType(PyObject* result) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(result));
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
    PyErr_Format(PyExc_TypeError,
                 "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ")",
                 result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  PyObject *res = NULL;
  if (likely(PyLong_Check(x)))
      return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  if (likely(m && m->nb_int)) {
      res = m->nb_int(x);
  }
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
      res = PyNumber_Long(x);
  }
#endif
  if (likely(res)) {
      if (unlikely(!PyLong_CheckExact(res))) {
          return __Pyx_PyNumber_LongWrongResultType(res);
      }
  }
  else if (!PyErr_Occurred()) {
      PyErr_SetString(PyExc_TypeError,
                      "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyLong_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyLong_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b) {
    CYTHON_UNUSED_VAR(b);
    return __Pyx_NewRef(Py_None);
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return __Pyx_NewRef(b ? Py_True: Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t ival) {
    return PyLong_FromSize_t(ival);
}


/* MultiPhaseInitModuleState */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
#ifndef CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#if (CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX >= 0x030C0000)
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 1
#else
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 0
#endif
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE && !CYTHON_ATOMICS
#error "Module state with PEP489 requires atomics. Currently that's one of\
 C11, C++11, gcc atomic intrinsics or MSVC atomic intrinsics"
#endif
#if !CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#define __Pyx_ModuleStateLookup_Lock()
#define __Pyx_ModuleStateLookup_Unlock()
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
static PyMutex __Pyx_ModuleStateLookup_mutex = {0};
#define __Pyx_ModuleStateLookup_Lock() PyMutex_Lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() PyMutex_Unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(__cplusplus) && __cplusplus >= 201103L
#include <mutex>
static std::mutex __Pyx_ModuleStateLookup_mutex;
#define __Pyx_ModuleStateLookup_Lock() __Pyx_ModuleStateLookup_mutex.lock()
#define __Pyx_ModuleStateLookup_Unlock() __Pyx_ModuleStateLookup_mutex.unlock()
#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ > 201112L) && !defined(__STDC_NO_THREADS__)
#include <threads.h>
static mtx_t __Pyx_ModuleStateLookup_mutex;
static once_flag __Pyx_ModuleStateLookup_mutex_once_flag = ONCE_FLAG_INIT;
static void __Pyx_ModuleStateLookup_initialize_mutex(void) {
    mtx_init(&__Pyx_ModuleStateLookup_mutex, mtx_plain);
}
#define __Pyx_ModuleStateLookup_Lock()\
  call_once(&__Pyx_ModuleStateLookup_mutex_once_flag, __Pyx_ModuleStateLookup_initialize_mutex);\
  mtx_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() mtx_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(HAVE_PTHREAD_H)
#include <pthread.h>
static pthread_mutex_t __Pyx_ModuleStateLookup_mutex = PTHREAD_MUTEX_INITIALIZER;
#define __Pyx_ModuleStateLookup_Lock() pthread_mutex_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() pthread_mutex_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(_WIN32)
#include <Windows.h>  // synchapi.h on its own doesn't work
static SRWLOCK __Pyx_ModuleStateLookup_mutex = SRWLOCK_INIT;
#define __Pyx_ModuleStateLookup_Lock() AcquireSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() ReleaseSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#else
#error "No suitable lock available for CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE.\
 Requires C standard >= C11, or C++ standard >= C++11,\
 or pthreads, or the Windows 32 API, or Python >= 3.13."
#endif
typedef struct {
    int64_t id;
    PyObject *module;
} __Pyx_InterpreterIdAndModule;
typedef struct {
    char interpreter_id_as_index;
    Py_ssize_t count;
    Py_ssize_t allocated;
    __Pyx_InterpreterIdAndModule table[1];
} __Pyx_ModuleStateLookupData;
#define __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE 32
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_int_type __Pyx_ModuleStateLookup_read_counter = 0;
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_ptr_type __Pyx_ModuleStateLookup_data = 0;
#else
static __Pyx_ModuleStateLookupData* __Pyx_ModuleStateLookup_data = NULL;
#endif
static __Pyx_InterpreterIdAndModule* __Pyx_State_FindModuleStateLookupTableLowerBound(
        __Pyx_InterpreterIdAndModule* table,
        Py_ssize_t count,
        int64_t interpreterId) {
    __Pyx_InterpreterIdAndModule* begin = table;
    __Pyx_InterpreterIdAndModule* end = begin + count;
    if (begin->id == interpreterId) {
        return begin;
    }
    while ((end - begin) > __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
        __Pyx_InterpreterIdAndModule* halfway = begin + (end - begin)/2;
        if (halfway->id == interpreterId) {
            return halfway;
        }
        if (halfway->id < interpreterId) {
            begin = halfway;
        } else {
            end = halfway;
        }
    }
    for (; begin < end; ++begin) {
        if (begin->id >= interpreterId) return begin;
    }
    return begin;
}
static PyObject *__Pyx_State_FindModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return NULL;
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData* data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
    {
        __pyx_atomic_incr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        if (likely(data)) {
            __Pyx_ModuleStateLookupData* new_data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_acquire(&__Pyx_ModuleStateLookup_data);
            if (likely(data == new_data)) {
                goto read_finished;
            }
        }
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        __Pyx_ModuleStateLookup_Lock();
        __pyx_atomic_incr_relaxed(&__Pyx_ModuleStateLookup_read_counter);
        data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
        __Pyx_ModuleStateLookup_Unlock();
    }
  read_finished:;
#else
    __Pyx_ModuleStateLookupData* data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_InterpreterIdAndModule* found = NULL;
    if (unlikely(!data)) goto end;
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            found = data->table+interpreter_id;
        }
    } else {
        found = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
    }
  end:
    {
        PyObject *result=NULL;
        if (found && found->id == interpreter_id) {
            result = found->module;
        }
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
#endif
        return result;
    }
}
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static void __Pyx_ModuleStateLookup_wait_until_no_readers(void) {
    while (__pyx_atomic_load(&__Pyx_ModuleStateLookup_read_counter) != 0);
}
#else
#define __Pyx_ModuleStateLookup_wait_until_no_readers()
#endif
static int __Pyx_State_AddModuleInterpIdAsIndex(__Pyx_ModuleStateLookupData **old_data, PyObject* module, int64_t interpreter_id) {
    Py_ssize_t to_allocate = (*old_data)->allocated;
    while (to_allocate <= interpreter_id) {
        if (to_allocate == 0) to_allocate = 1;
        else to_allocate *= 2;
    }
    __Pyx_ModuleStateLookupData *new_data = *old_data;
    if (to_allocate != (*old_data)->allocated) {
         new_data = (__Pyx_ModuleStateLookupData *)realloc(
            *old_data,
            sizeof(__Pyx_ModuleStateLookupData)+(to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
        if (!new_data) {
            PyErr_NoMemory();
            return -1;
        }
        for (Py_ssize_t i = new_data->allocated; i < to_allocate; ++i) {
            new_data->table[i].id = i;
            new_data->table[i].module = NULL;
        }
        new_data->allocated = to_allocate;
    }
    new_data->table[interpreter_id].module = module;
    if (new_data->count < interpreter_id+1) {
        new_data->count = interpreter_id+1;
    }
    *old_data = new_data;
    return 0;
}
static void __Pyx_State_ConvertFromInterpIdAsIndex(__Pyx_ModuleStateLookupData *data) {
    __Pyx_InterpreterIdAndModule *read = data->table;
    __Pyx_InterpreterIdAndModule *write = data->table;
    __Pyx_InterpreterIdAndModule *end = read + data->count;
    for (; read<end; ++read) {
        if (read->module) {
            write->id = read->id;
            write->module = read->module;
            ++write;
        }
    }
    data->count = write - data->table;
    for (; write<end; ++write) {
        write->id = 0;
        write->module = NULL;
    }
    data->interpreter_id_as_index = 0;
}
static int __Pyx_State_AddModule(PyObject* module, CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    int result = 0;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *old_data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *old_data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_ModuleStateLookupData *new_data = old_data;
    if (!new_data) {
        new_data = (__Pyx_ModuleStateLookupData *)calloc(1, sizeof(__Pyx_ModuleStateLookupData));
        if (!new_data) {
            result = -1;
            PyErr_NoMemory();
            goto end;
        }
        new_data->allocated = 1;
        new_data->interpreter_id_as_index = 1;
    }
    __Pyx_ModuleStateLookup_wait_until_no_readers();
    if (new_data->interpreter_id_as_index) {
        if (interpreter_id < __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
            result = __Pyx_State_AddModuleInterpIdAsIndex(&new_data, module, interpreter_id);
            goto end;
        }
        __Pyx_State_ConvertFromInterpIdAsIndex(new_data);
    }
    {
        Py_ssize_t insert_at = 0;
        {
            __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
                new_data->table, new_data->count, interpreter_id);
            assert(lower_bound);
            insert_at = lower_bound - new_data->table;
            if (unlikely(insert_at < new_data->count && lower_bound->id == interpreter_id)) {
                lower_bound->module = module;
                goto end;  // already in table, nothing more to do
            }
        }
        if (new_data->count+1 >= new_data->allocated) {
            Py_ssize_t to_allocate = (new_data->count+1)*2;
            new_data =
                (__Pyx_ModuleStateLookupData*)realloc(
                    new_data,
                    sizeof(__Pyx_ModuleStateLookupData) +
                    (to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
            if (!new_data) {
                result = -1;
                new_data = old_data;
                PyErr_NoMemory();
                goto end;
            }
            new_data->allocated = to_allocate;
        }
        ++new_data->count;
        int64_t last_id = interpreter_id;
        PyObject *last_module = module;
        for (Py_ssize_t i=insert_at; i<new_data->count; ++i) {
            int64_t current_id = new_data->table[i].id;
            new_data->table[i].id = last_id;
            last_id = current_id;
            PyObject *current_module = new_data->table[i].module;
            new_data->table[i].module = last_module;
            last_module = current_module;
        }
    }
  end:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, new_data);
#else
    __Pyx_ModuleStateLookup_data = new_data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return result;
}
static int __Pyx_State_RemoveModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *data = __Pyx_ModuleStateLookup_data;
#endif
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            data->table[interpreter_id].module = NULL;
        }
        goto done;
    }
    {
        __Pyx_ModuleStateLookup_wait_until_no_readers();
        __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
        if (!lower_bound) goto done;
        if (lower_bound->id != interpreter_id) goto done;
        __Pyx_InterpreterIdAndModule *end = data->table+data->count;
        for (;lower_bound<end-1; ++lower_bound) {
            lower_bound->id = (lower_bound+1)->id;
            lower_bound->module = (lower_bound+1)->module;
        }
    }
    --data->count;
    if (data->count == 0) {
        free(data);
        data = NULL;
    }
  done:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, data);
#else
    __Pyx_ModuleStateLookup_data = data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return 0;
}
#endif

/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
