/* Generated by Cython 3.2.4 */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
/* InitLimitedAPI */
#if defined(Py_LIMITED_API)
  #if !defined(CYTHON_LIMITED_API)
  #define CYTHON_LIMITED_API 1
  #endif
#elif defined(CYTHON_LIMITED_API)
  #ifdef _MSC_VER
  #pragma message ("Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.")
  #else
  #warning Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x03080000
    #error Cython requires Python 3.8+.
#else
#define __PYX_ABI_VERSION "3_2_4"
#define CYTHON_HEX_VERSION 0x030204F0
#define CYTHON_FUTURE_DIVISION 1
/* CModulePreamble */
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 1
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 1
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PYPY_VERSION_NUM >= 0x07030C00)
  #endif
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_NUM >= 0x07031100)
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND (__PYX_LIMITED_VERSION_HEX >= 0x030A0000)
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #ifndef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 1
  #endif
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #ifdef Py_GIL_DISABLED
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 1
  #else
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #endif
  #if PY_VERSION_HEX < 0x030A0000
    #undef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #elif !defined(CYTHON_USE_TYPE_SLOTS)
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLIST_INTERNALS)
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #elif !defined(CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL 0
  #elif !defined(CYTHON_FAST_GIL)
    #define CYTHON_FAST_GIL (PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_SYS_MONITORING
    #define CYTHON_USE_SYS_MONITORING (PY_VERSION_HEX >= 0x030d00B1)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5 && !CYTHON_USE_MODULE_STATE)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS (!CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
  #if defined(CYTHON_IMMORTAL_CONSTANTS) && PY_VERSION_HEX < 0x030C0000
    #undef CYTHON_IMMORTAL_CONSTANTS
    #define CYTHON_IMMORTAL_CONSTANTS 0  // definitely won't work
  #elif !defined(CYTHON_IMMORTAL_CONSTANTS)
    #define CYTHON_IMMORTAL_CONSTANTS (PY_VERSION_HEX >= 0x030C0000 && !CYTHON_USE_MODULE_STATE && CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
#endif
#ifndef CYTHON_COMPRESS_STRINGS
  #define CYTHON_COMPRESS_STRINGS 1
#endif
#ifndef CYTHON_FAST_PYCCALL
#define CYTHON_FAST_PYCCALL  CYTHON_FAST_PYCALL
#endif
#ifndef CYTHON_VECTORCALL
#if CYTHON_COMPILING_IN_LIMITED_API
#define CYTHON_VECTORCALL  (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
#else
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL)
#endif
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#include <stdint.h>
typedef uintptr_t  __pyx_uintptr_t;
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifndef Py_UNREACHABLE
  #define Py_UNREACHABLE()  assert(0); abort()
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

/* CInitCode */
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

/* PythonCompatibility */
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#define __Pyx_BUILTIN_MODULE_NAME "builtins"
#define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    #ifndef CO_OPTIMIZED
    static int CO_OPTIMIZED;
    #endif
    #ifndef CO_NEWLOCALS
    static int CO_NEWLOCALS;
    #endif
    #ifndef CO_VARARGS
    static int CO_VARARGS;
    #endif
    #ifndef CO_VARKEYWORDS
    static int CO_VARKEYWORDS;
    #endif
    #ifndef CO_ASYNC_GENERATOR
    static int CO_ASYNC_GENERATOR;
    #endif
    #ifndef CO_GENERATOR
    static int CO_GENERATOR;
    #endif
    #ifndef CO_COROUTINE
    static int CO_COROUTINE;
    #endif
#else
    #ifndef CO_COROUTINE
      #define CO_COROUTINE 0x80
    #endif
    #ifndef CO_ASYNC_GENERATOR
      #define CO_ASYNC_GENERATOR 0x200
    #endif
#endif
static int __Pyx_init_co_variables(void);
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef Py_TPFLAGS_IMMUTABLETYPE
  #define Py_TPFLAGS_IMMUTABLETYPE (1UL << 8)
#endif
#ifndef Py_TPFLAGS_DISALLOW_INSTANTIATION
  #define Py_TPFLAGS_DISALLOW_INSTANTIATION (1UL << 7)
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#ifndef METH_FASTCALL
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void (*cfunc)(void)) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if PY_VERSION_HEX < 0x03090000 || (CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000)
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#elif CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) GraalPyFrame_SetLineNumber((frame), (lineno))
#elif CYTHON_COMPILING_IN_GRAAL
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) _PyFrame_SetLineNumber((frame), (lineno))
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#endif
#if CYTHON_USE_MODULE_STATE
static CYTHON_INLINE void *__Pyx__PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#define __Pyx_PyModule_GetState(o) (__pyx_mstatetype *)__Pyx__PyModule_GetState(o)
#else
#define __Pyx_PyModule_GetState(op) ((void)op,__pyx_mstate_global)
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE((PyObject *) obj), name, func_ctype)
#define __Pyx_PyObject_TryGetSlot(obj, name, func_ctype) __Pyx_PyType_TryGetSlot(Py_TYPE(obj), name, func_ctype)
#define __Pyx_PyObject_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#define __Pyx_PyObject_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype) __Pyx_PyType_GetSlot(type, name, func_ctype)
  #define __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype) (((type)->sub) ? ((type)->sub->name) : NULL)
  #define __Pyx_PyType_TryGetSubSlot(type, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype)
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype)\
    ((__PYX_LIMITED_VERSION_HEX >= 0x030A0000 ||\
     (PyType_GetFlags(type) & Py_TPFLAGS_HEAPTYPE) || __Pyx_get_runtime_version() >= 0x030A0000) ?\
     __Pyx_PyType_GetSlot(type, name, func_ctype) : NULL)
  #define __Pyx_PyType_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSlot(obj, name, func_ctype)
  #define __Pyx_PyType_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSlot(obj, name, func_ctype)
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
#define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
#endif
#define __Pyx_PyObject_GetIterNextFunc(iterator)  __Pyx_PyObject_GetSlot(iterator, tp_iternext, iternextfunc)
#if CYTHON_USE_TYPE_SPECS
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#else
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains)
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && PyUnstable_Object_IsUniquelyReferenced(obj)) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#elif CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
enum __Pyx_ReferenceSharing {
  __Pyx_ReferenceSharing_DefinitelyUnique, // We created it so we know it's unshared - no need to check
  __Pyx_ReferenceSharing_OwnStrongReference,
  __Pyx_ReferenceSharing_FunctionArgument,
  __Pyx_ReferenceSharing_SharedReference, // Never trust it to be unshared because it's a global or similar
};
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && PY_VERSION_HEX >= 0x030E0000
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing)\
    (sharing == __Pyx_ReferenceSharing_DefinitelyUnique ? 1 :\
      (sharing == __Pyx_ReferenceSharing_FunctionArgument ? PyUnstable_Object_IsUniqueReferencedTemporary(o) :\
      (sharing == __Pyx_ReferenceSharing_OwnStrongReference ? PyUnstable_Object_IsUniquelyReferenced(o) : 0)))
#elif (CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING) || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)sharing), Py_REFCNT(o) == 1)
#else
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)o), ((void)sharing), 0)
#endif
#if CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyList_GetItemRef(o, i) (likely((i) >= 0) ? PySequence_GetItem(o, i) : (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
  #else
    #define __Pyx_PyList_GetItemRef(o, i) PySequence_ITEM(o, i)
  #endif
#elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #else
    #define __Pyx_PyList_GetItemRef(o, i) __Pyx_XNewRef(PyList_GetItem(o, i))
  #endif
#else
  #define __Pyx_PyList_GetItemRef(o, i) __Pyx_NewRef(PyList_GET_ITEM(o, i))
#endif
#if CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS && !CYTHON_COMPILING_IN_LIMITED_API && CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) (__Pyx_IS_UNIQUELY_REFERENCED(o, unsafe_shared) ?\
    __Pyx_NewRef(PyList_GET_ITEM(o, i)) : __Pyx_PyList_GetItemRef(o, i))
#else
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) __Pyx_PyList_GetItemRef(o, i)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyDict_GetItemRef(dict, key, result) PyDict_GetItemRef(dict, key, result)
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyObject_GetItem(dict, key);
  if (*result == NULL) {
    if (PyErr_ExceptionMatches(PyExc_KeyError)) {
      PyErr_Clear();
      return 0;
    }
    return -1;
  }
  return 1;
}
#else
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyDict_GetItemWithError(dict, key);
  if (*result == NULL) {
    return PyErr_Occurred() ? -1 : 0;
  }
  Py_INCREF(*result);
  return 1;
}
#endif
#if defined(CYTHON_DEBUG_VISIT_CONST) && CYTHON_DEBUG_VISIT_CONST
  #define __Pyx_VISIT_CONST(obj)  Py_VISIT(obj)
#else
  #define __Pyx_VISIT_CONST(obj)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GET_ITEM(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GET_ITEM(o, i)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GetItem(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GetItem(o, i)
#endif
#if CYTHON_ASSUME_SAFE_SIZE
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GET_LENGTH(o)
#else
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GetLength(o)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_InternFromString)
  #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
#endif
#define __Pyx_PyLong_FromHash_t PyLong_FromSsize_t
#define __Pyx_PyLong_AsHash_t   __Pyx_PyIndex_AsSsize_t
#if __PYX_LIMITED_VERSION_HEX >= 0x030A0000
    #define __Pyx_PySendResult PySendResult
#else
    typedef enum {
        PYGEN_RETURN = 0,
        PYGEN_ERROR = -1,
        PYGEN_NEXT = 1,
    } __Pyx_PySendResult;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030A00A3
  typedef __Pyx_PySendResult (*__Pyx_pyiter_sendfunc)(PyObject *iter, PyObject *value, PyObject **result);
#else
  #define __Pyx_pyiter_sendfunc sendfunc
#endif
#if !CYTHON_USE_AM_SEND
#define __PYX_HAS_PY_AM_SEND 0
#elif __PYX_LIMITED_VERSION_HEX >= 0x030A0000
#define __PYX_HAS_PY_AM_SEND 1
#else
#define __PYX_HAS_PY_AM_SEND 2  // our own backported implementation
#endif
#if __PYX_HAS_PY_AM_SEND < 2
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
        __Pyx_pyiter_sendfunc am_send;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_SlotTpAsAsync(s) ((PyAsyncMethods*)(s))
#endif
#if CYTHON_USE_AM_SEND && PY_VERSION_HEX < 0x030A00F0
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (1UL << 21)
#else
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (0)
#endif
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyInterpreterState_Get() PyInterpreterState_Get()
#else
#define __Pyx_PyInterpreterState_Get() PyThreadState_Get()->interp
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030A0000
#ifdef __cplusplus
extern "C"
#endif
PyAPI_FUNC(void *) PyMem_Calloc(size_t nelem, size_t elsize);
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_init_co_variable(PyObject *inspect, const char* name, int *write_to) {
    int value;
    PyObject *py_value = PyObject_GetAttrString(inspect, name);
    if (!py_value) return 0;
    value = (int) PyLong_AsLong(py_value);
    Py_DECREF(py_value);
    *write_to = value;
    return value != -1 || !PyErr_Occurred();
}
static int __Pyx_init_co_variables(void) {
    PyObject *inspect;
    int result;
    inspect = PyImport_ImportModule("inspect");
    result =
#if !defined(CO_OPTIMIZED)
        __Pyx_init_co_variable(inspect, "CO_OPTIMIZED", &CO_OPTIMIZED) &&
#endif
#if !defined(CO_NEWLOCALS)
        __Pyx_init_co_variable(inspect, "CO_NEWLOCALS", &CO_NEWLOCALS) &&
#endif
#if !defined(CO_VARARGS)
        __Pyx_init_co_variable(inspect, "CO_VARARGS", &CO_VARARGS) &&
#endif
#if !defined(CO_VARKEYWORDS)
        __Pyx_init_co_variable(inspect, "CO_VARKEYWORDS", &CO_VARKEYWORDS) &&
#endif
#if !defined(CO_ASYNC_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_ASYNC_GENERATOR", &CO_ASYNC_GENERATOR) &&
#endif
#if !defined(CO_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_GENERATOR", &CO_GENERATOR) &&
#endif
#if !defined(CO_COROUTINE)
        __Pyx_init_co_variable(inspect, "CO_COROUTINE", &CO_COROUTINE) &&
#endif
        1;
    Py_DECREF(inspect);
    return result ? 0 : -1;
}
#else
static int __Pyx_init_co_variables(void) {
    return 0;  // It's a limited API-only feature
}
#endif

/* MathInitCode */
#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #ifndef _USE_MATH_DEFINES
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#ifndef CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#define CYTHON_CLINE_IN_TRACEBACK_RUNTIME 0
#endif
#ifndef CYTHON_CLINE_IN_TRACEBACK
#define CYTHON_CLINE_IN_TRACEBACK CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#endif
#if CYTHON_CLINE_IN_TRACEBACK
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; __pyx_clineno = __LINE__; (void) __pyx_clineno; }
#else
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; (void) __pyx_clineno; }
#endif
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__pyverbs__providers__mlx5__mlx5_enums
#define __PYX_HAVE_API__pyverbs__providers__mlx5__mlx5_enums
/* Early includes */
#include "infiniband/mlx5dv.h"
#include "infiniband/mlx5_user_ioctl_verbs.h"
#include "infiniband/mlx5_api.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifdef CYTHON_FREETHREADING_COMPATIBLE
#if CYTHON_FREETHREADING_COMPATIBLE
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_NOT_USED
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AS_STRING(s)
#else
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AsString(s)
#endif
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
static CYTHON_INLINE PyObject *__Pyx_NewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_NewRef)
    return Py_NewRef(obj);
#else
    Py_INCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_XNewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_XNewRef)
    return Py_XNewRef(obj);
#else
    Py_XINCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b);
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __Pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AS_DOUBLE(x)
#else
#define __Pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AsDouble(x)
#endif
#define __Pyx_PyFloat_AsFloat(x) ((float) __Pyx_PyFloat_AsDouble(x))
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#elif __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeASCII(c_str, size, NULL)
#else
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
/* PretendToInitialize */
#ifdef __cplusplus
#if __cplusplus > 201103L
#include <type_traits>
#endif
template <typename T>
static void __Pyx_pretend_to_initialize(T* ptr) {
#if __cplusplus > 201103L
    if ((std::is_trivially_default_constructible<T>::value))
#endif
        *ptr = T();
    (void)ptr;
}
#else
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }
#endif


#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * const __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char* const __pyx_f[] = {
  "pyverbs/providers/mlx5/mlx5_enums.pyx",
  "EnumBase",
  "pyverbs/providers/mlx5/mlx5_enums.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* Atomics.proto (used by CodeObjectCache) */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __PYX_GET_CYTHON_COMPILING_IN_CPYTHON_FREETHREADING() CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __pyx_atomic_int_type int
#define __pyx_nonatomic_int_type int
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__))
    #include <stdatomic.h>
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)))
    #include <atomic>
#endif
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__) &&\
                       ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type atomic_int
    #define __pyx_atomic_ptr_type atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) atomic_fetch_add_explicit(value, 1, memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) atomic_fetch_add_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) atomic_fetch_sub_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) atomic_load(value)
    #define __pyx_atomic_store(value, new_value) atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) atomic_load_explicit(value, memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) atomic_load_explicit(value, memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C atomics"
    #endif
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)) &&\
                    ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type std::atomic_int
    #define __pyx_atomic_ptr_type std::atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) std::atomic_fetch_sub_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) std::atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) std::atomic_load(value)
    #define __pyx_atomic_store(value, new_value) std::atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) std::atomic_load_explicit(value, std::memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) std::atomic_load_explicit(value, std::memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) std::atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C++ atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C++ atomics"
    #endif
#elif CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_ptr_type void*
    #define __pyx_nonatomic_ptr_type void*
    #define __pyx_atomic_incr_relaxed(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) __sync_fetch_and_sub(value, 1)
    #define __pyx_atomic_sub(value, arg) __sync_fetch_and_sub(value, arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_store(value, new_value) __sync_lock_test_and_set(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_load_acquire(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) __sync_lock_test_and_set(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_nonatomic_ptr_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER)
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #define __pyx_atomic_ptr_type void*
    #undef __pyx_nonatomic_int_type
    #define __pyx_nonatomic_int_type long
    #define __pyx_nonatomic_ptr_type void*
    #pragma intrinsic (_InterlockedExchangeAdd, _InterlockedExchange, _InterlockedCompareExchange, _InterlockedCompareExchangePointer, _InterlockedExchangePointer)
    #define __pyx_atomic_incr_relaxed(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) _InterlockedExchangeAdd(value, -1)
    #define __pyx_atomic_sub(value, arg) _InterlockedExchangeAdd(value, -arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = _InterlockedCompareExchange(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) _InterlockedExchangeAdd(value, 0)
    #define __pyx_atomic_store(value, new_value) _InterlockedExchange(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) *(void * volatile *)value
    #define __pyx_atomic_pointer_load_acquire(value) _InterlockedCompareExchangePointer(value, 0, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) _InterlockedExchangePointer(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_atomic_ptr_type old = _InterlockedCompareExchangePointer(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif

/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto (used by PyErrFetchRestore) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto (used by HasAttr) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* HasAttr.proto (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_HasAttr(o, n)  PyObject_HasAttrWithError(o, n)
#else
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);
#endif

/* PyObjectCall.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto (used by PyObjectCallOneArg) */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs);

/* PyObjectCallOneArg.proto (used by ImportImpl) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* TupleAndListFromArray.proto (used by ImportImpl) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
#endif
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto (used by ImportImpl) */
#include <string.h>

/* ImportImpl.export */
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level);

/* Import.proto */
static CYTHON_INLINE PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* ListPack.proto */
static PyObject *__Pyx_PyList_Pack(Py_ssize_t n, ...);

/* GetAttr.proto (used by Globals) */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* Globals.proto */
static PyObject* __Pyx_Globals(void);

/* PyObjectVectorCallKwBuilder.proto */
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#if CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_Vectorcall_CallFromBuilder PyObject_Vectorcall
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder _PyObject_Vectorcall
#endif
#define __Pyx_MakeVectorcallBuilderKwds(n) PyTuple_New(n)
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder __Pyx_PyObject_FastCallDict
#define __Pyx_MakeVectorcallBuilderKwds(n) __Pyx_PyDict_NewPresized(n)
#define __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n) PyDict_SetItem(builder, key, value)
#define __Pyx_VectorcallBuilder_AddArgStr(key, value, builder, args, n) PyDict_SetItemString(builder, key, value)
#endif

/* GetBuiltinName.proto (used by GetModuleGlobalName) */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyDictVersioning.proto (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __Pyx_XNewRef(__pyx_dict_cached_value);\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_mstate_global->__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectDelAttr.proto (used by PyObjectSetAttrStr) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
#define __Pyx_PyObject_DelAttr(o, n) PyObject_SetAttr(o, n, NULL)
#else
#define __Pyx_PyObject_DelAttr(o, n) PyObject_DelAttr(o, n)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   __Pyx_PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* dict_setdefault.proto (used by CLineInTraceback) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value);

/* CLineInTraceback.proto (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#else
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#endif

/* CodeObjectCache.proto (used by AddTraceback) */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject __Pyx_CachedCodeObjectType;
#else
typedef PyCodeObject __Pyx_CachedCodeObjectType;
#endif
typedef struct {
    __Pyx_CachedCodeObjectType* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_int_type accessor_count;
  #endif
};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_PY_LONG_LONG(unsigned PY_LONG_LONG value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_block_size(enum mlx5dv_block_size value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_context_attr_comp_mask(enum mlx5dv_context_attr_comp_mask value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_context_attr_flags(enum mlx5dv_context_attr_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(enum mlx5dv_context_comp_mask value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_context_flags(enum mlx5dv_context_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_flags(enum mlx5dv_cq_init_attr_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_mask(enum mlx5dv_cq_init_attr_mask value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_cqe_comp_res_format(enum mlx5dv_cqe_comp_res_format value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_crypto_caps_flags(enum mlx5dv_crypto_caps_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_crypto_engines_caps(enum mlx5dv_crypto_engines_caps value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_crypto_key_purpose(enum mlx5dv_crypto_key_purpose value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_crypto_key_size(enum mlx5dv_crypto_key_size value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_crypto_login_state(enum mlx5dv_crypto_login_state value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_crypto_standard(enum mlx5dv_crypto_standard value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_crypto_wrapped_import_method_caps(enum mlx5dv_crypto_wrapped_import_method_caps value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_dc_type(enum mlx5dv_dc_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_dek_init_attr_mask(enum mlx5dv_dek_init_attr_mask value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_dek_state(enum mlx5dv_dek_state value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_devx_create_event_channel_flags(enum mlx5dv_devx_create_event_channel_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_dr_action_dest_type(enum mlx5dv_dr_action_dest_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_dr_action_flags(enum mlx5dv_dr_action_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_dr_domain_sync_flags(enum mlx5dv_dr_domain_sync_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_dr_domain_type(enum mlx5dv_dr_domain_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_dr_matcher_layout_flags(enum mlx5dv_dr_matcher_layout_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(enum mlx5dv_flow_action_cap_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(enum mlx5dv_flow_action_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_flow_matcher_attr_mask(enum mlx5dv_flow_matcher_attr_mask value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(enum mlx5dv_mkey_err_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(enum mlx5dv_mkey_init_attr_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_obj_type(enum mlx5dv_obj_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_qp_comp_mask(enum mlx5dv_qp_comp_mask value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(enum mlx5dv_qp_create_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(enum mlx5dv_qp_create_send_ops_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(enum mlx5dv_qp_init_attr_mask value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_sched_elem_attr_flags(enum mlx5dv_sched_elem_attr_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_sig_block_attr_flags(enum mlx5dv_sig_block_attr_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_sig_crc_type(enum mlx5dv_sig_crc_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_sig_mask(enum mlx5dv_sig_mask value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_bg_type(enum mlx5dv_sig_t10dif_bg_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_flags(enum mlx5dv_sig_t10dif_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_sig_type(enum mlx5dv_sig_type value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_signature_crypto_order(enum mlx5dv_signature_crypto_order value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_srq_comp_mask(enum mlx5dv_srq_comp_mask value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_sw_parsing_offloads(enum mlx5dv_sw_parsing_offloads value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(enum mlx5dv_tunnel_offloads value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_vfio_context_attr_flags(enum mlx5dv_vfio_context_attr_flags value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_wc_opcode(enum mlx5dv_wc_opcode value);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#else
static __Pyx_TypeName __Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp);
#endif
#else  // !LIMITED_API
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetFullyQualifiedName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2) {
    return PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2);
}
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#ifdef PyExceptionInstance_Check
  #define __Pyx_PyBaseException_Check(obj) PyExceptionInstance_Check(obj)
#else
  #define __Pyx_PyBaseException_Check(obj) __Pyx_TypeCheck(obj, PyExc_BaseException)
#endif

/* GetRuntimeVersion.proto */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
static unsigned long __Pyx_cached_runtime_version = 0;
static void __Pyx_init_runtime_version(void);
#else
#define __Pyx_init_runtime_version()
#endif
static unsigned long __Pyx_get_runtime_version(void);

/* AddModuleRef.proto */
#if ((CYTHON_COMPILING_IN_CPYTHON_FREETHREADING ) ||\
     __PYX_LIMITED_VERSION_HEX < 0x030d0000)
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name);
#else
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#endif

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From___pyx_anon_enum(int value);

/* DecompressString.proto */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo);

/* MultiPhaseInitModuleState.proto */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
static PyObject *__Pyx_State_FindModule(void*);
static int __Pyx_State_AddModule(PyObject* module, void*);
static int __Pyx_State_RemoveModule(void*);
#elif CYTHON_USE_MODULE_STATE
#define __Pyx_State_FindModule PyState_FindModule
#define __Pyx_State_AddModule PyState_AddModule
#define __Pyx_State_RemoveModule PyState_RemoveModule
#endif

/* #### Code section: module_declarations ### */
/* CythonABIVersion.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
    #if CYTHON_METH_FASTCALL
        #define __PYX_FASTCALL_ABI_SUFFIX  "_fastcall"
    #else
        #define __PYX_FASTCALL_ABI_SUFFIX
    #endif
    #define __PYX_LIMITED_ABI_SUFFIX "limited" __PYX_FASTCALL_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#else
    #define __PYX_LIMITED_ABI_SUFFIX
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    #define __PYX_AM_SEND_ABI_SUFFIX
#elif __PYX_HAS_PY_AM_SEND == 2
    #define __PYX_AM_SEND_ABI_SUFFIX "amsendbackport"
#else
    #define __PYX_AM_SEND_ABI_SUFFIX "noamsend"
#endif
#ifndef __PYX_MONITORING_ABI_SUFFIX
    #define __PYX_MONITORING_ABI_SUFFIX
#endif
#if CYTHON_USE_TP_FINALIZE
    #define __PYX_TP_FINALIZE_ABI_SUFFIX
#else
    #define __PYX_TP_FINALIZE_ABI_SUFFIX "nofinalize"
#endif
#if CYTHON_USE_FREELISTS || !defined(__Pyx_AsyncGen_USED)
    #define __PYX_FREELISTS_ABI_SUFFIX
#else
    #define __PYX_FREELISTS_ABI_SUFFIX "nofreelists"
#endif
#define CYTHON_ABI  __PYX_ABI_VERSION __PYX_LIMITED_ABI_SUFFIX __PYX_MONITORING_ABI_SUFFIX __PYX_TP_FINALIZE_ABI_SUFFIX __PYX_FREELISTS_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."


/* Module declarations from "pyverbs.providers.mlx5.mlx5_enums" */
static PyObject *__Pyx_EnumBase = 0;
static PyObject *__Pyx_FlagBase = 0;
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "pyverbs.providers.mlx5.mlx5_enums"
extern int __pyx_module_is_main_pyverbs__providers__mlx5__mlx5_enums;
int __pyx_module_is_main_pyverbs__providers__mlx5__mlx5_enums = 0;

/* Implementation of "pyverbs.providers.mlx5.mlx5_enums" */
/* #### Code section: global_var ### */
/* #### Code section: string_decls ### */
/* #### Code section: decls ### */
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
/* SmallCodeConfig */
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  PyObject *__pyx_tuple[1];
  PyObject *__pyx_string_tab[243];
/* #### Code section: module_state_contents ### */
/* CodeObjectCache.module_state_decls */
struct __Pyx_CodeObjectCache __pyx_code_cache;

/* #### Code section: module_state_end ### */
} __pyx_mstatetype;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate_global (__Pyx_PyModule_GetState(__Pyx_State_FindModule(&__pyx_moduledef)))

#define __pyx_m (__Pyx_State_FindModule(&__pyx_moduledef))
#else
static __pyx_mstatetype __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstatetype * const __pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: constant_name_defines ### */
#define __pyx_kp_u_ __pyx_string_tab[0]
#define __pyx_kp_u__2 __pyx_string_tab[1]
#define __pyx_kp_u_pyverbs_providers_mlx5_mlx5_enum __pyx_string_tab[2]
#define __pyx_n_u_IntEnum __pyx_string_tab[3]
#define __pyx_n_u_IntFlag __pyx_string_tab[4]
#define __pyx_n_u_MLX5DV_BLOCK_SIZE_4048 __pyx_string_tab[5]
#define __pyx_n_u_MLX5DV_BLOCK_SIZE_4096 __pyx_string_tab[6]
#define __pyx_n_u_MLX5DV_BLOCK_SIZE_4160 __pyx_string_tab[7]
#define __pyx_n_u_MLX5DV_BLOCK_SIZE_512 __pyx_string_tab[8]
#define __pyx_n_u_MLX5DV_BLOCK_SIZE_520 __pyx_string_tab[9]
#define __pyx_n_u_MLX5DV_CONTEXT_ATTR_MASK_FD_ARRA __pyx_string_tab[10]
#define __pyx_n_u_MLX5DV_CONTEXT_FLAGS_BLUEFLAME __pyx_string_tab[11]
#define __pyx_n_u_MLX5DV_CONTEXT_FLAGS_CQE_128B_CO __pyx_string_tab[12]
#define __pyx_n_u_MLX5DV_CONTEXT_FLAGS_CQE_128B_PA __pyx_string_tab[13]
#define __pyx_n_u_MLX5DV_CONTEXT_FLAGS_CQE_V1 __pyx_string_tab[14]
#define __pyx_n_u_MLX5DV_CONTEXT_FLAGS_DEVX __pyx_string_tab[15]
#define __pyx_n_u_MLX5DV_CONTEXT_FLAGS_ENHANCED_MP __pyx_string_tab[16]
#define __pyx_n_u_MLX5DV_CONTEXT_FLAGS_MPW_ALLOWED __pyx_string_tab[17]
#define __pyx_n_u_MLX5DV_CONTEXT_FLAGS_PACKET_BASE __pyx_string_tab[18]
#define __pyx_n_u_MLX5DV_CONTEXT_FLAGS_REAL_TIME_T __pyx_string_tab[19]
#define __pyx_n_u_MLX5DV_CONTEXT_MASK_CLOCK_INFO_U __pyx_string_tab[20]
#define __pyx_n_u_MLX5DV_CONTEXT_MASK_CQE_COMPRESI __pyx_string_tab[21]
#define __pyx_n_u_MLX5DV_CONTEXT_MASK_CRYPTO_OFFLO __pyx_string_tab[22]
#define __pyx_n_u_MLX5DV_CONTEXT_MASK_DCI_STREAMS __pyx_string_tab[23]
#define __pyx_n_u_MLX5DV_CONTEXT_MASK_DC_ODP_CAPS __pyx_string_tab[24]
#define __pyx_n_u_MLX5DV_CONTEXT_MASK_DYN_BFREGS __pyx_string_tab[25]
#define __pyx_n_u_MLX5DV_CONTEXT_MASK_FLOW_ACTION __pyx_string_tab[26]
#define __pyx_n_u_MLX5DV_CONTEXT_MASK_MAX_DC_RD_AT __pyx_string_tab[27]
#define __pyx_n_u_MLX5DV_CONTEXT_MASK_NUM_LAG_PORT __pyx_string_tab[28]
#define __pyx_n_u_MLX5DV_CONTEXT_MASK_OOO_RECV_WRS __pyx_string_tab[29]
#define __pyx_n_u_MLX5DV_CONTEXT_MASK_SIGNATURE_OF __pyx_string_tab[30]
#define __pyx_n_u_MLX5DV_CONTEXT_MASK_STRIDING_RQ __pyx_string_tab[31]
#define __pyx_n_u_MLX5DV_CONTEXT_MASK_SWP __pyx_string_tab[32]
#define __pyx_n_u_MLX5DV_CONTEXT_MASK_TUNNEL_OFFLO __pyx_string_tab[33]
#define __pyx_n_u_MLX5DV_CONTEXT_MASK_WR_MEMCPY_LE __pyx_string_tab[34]
#define __pyx_n_u_MLX5DV_CQE_RES_FORMAT_CSUM __pyx_string_tab[35]
#define __pyx_n_u_MLX5DV_CQE_RES_FORMAT_CSUM_STRID __pyx_string_tab[36]
#define __pyx_n_u_MLX5DV_CQE_RES_FORMAT_HASH __pyx_string_tab[37]
#define __pyx_n_u_MLX5DV_CQ_INIT_ATTR_FLAGS_CQE_PA __pyx_string_tab[38]
#define __pyx_n_u_MLX5DV_CQ_INIT_ATTR_FLAGS_RESERV __pyx_string_tab[39]
#define __pyx_n_u_MLX5DV_CQ_INIT_ATTR_MASK_COMPRES __pyx_string_tab[40]
#define __pyx_n_u_MLX5DV_CQ_INIT_ATTR_MASK_CQE_SIZ __pyx_string_tab[41]
#define __pyx_n_u_MLX5DV_CQ_INIT_ATTR_MASK_FLAGS __pyx_string_tab[42]
#define __pyx_n_u_MLX5DV_CRYPTO_CAPS_CRYPTO __pyx_string_tab[43]
#define __pyx_n_u_MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPT __pyx_string_tab[44]
#define __pyx_n_u_MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPT_2 __pyx_string_tab[45]
#define __pyx_n_u_MLX5DV_CRYPTO_ENGINES_CAP_AES_XT __pyx_string_tab[46]
#define __pyx_n_u_MLX5DV_CRYPTO_ENGINES_CAP_AES_XT_2 __pyx_string_tab[47]
#define __pyx_n_u_MLX5DV_CRYPTO_ENGINES_CAP_AES_XT_3 __pyx_string_tab[48]
#define __pyx_n_u_MLX5DV_CRYPTO_KEY_PURPOSE_AES_XT __pyx_string_tab[49]
#define __pyx_n_u_MLX5DV_CRYPTO_KEY_SIZE_128 __pyx_string_tab[50]
#define __pyx_n_u_MLX5DV_CRYPTO_KEY_SIZE_256 __pyx_string_tab[51]
#define __pyx_n_u_MLX5DV_CRYPTO_LOGIN_STATE_INVALI __pyx_string_tab[52]
#define __pyx_n_u_MLX5DV_CRYPTO_LOGIN_STATE_NO_LOG __pyx_string_tab[53]
#define __pyx_n_u_MLX5DV_CRYPTO_LOGIN_STATE_VALID __pyx_string_tab[54]
#define __pyx_n_u_MLX5DV_CRYPTO_STANDARD_AES_XTS __pyx_string_tab[55]
#define __pyx_n_u_MLX5DV_CRYPTO_WRAPPED_IMPORT_MET __pyx_string_tab[56]
#define __pyx_n_u_MLX5DV_DCTYPE_DCI __pyx_string_tab[57]
#define __pyx_n_u_MLX5DV_DCTYPE_DCT __pyx_string_tab[58]
#define __pyx_n_u_MLX5DV_DEK_INIT_ATTR_CRYPTO_LOGI __pyx_string_tab[59]
#define __pyx_n_u_MLX5DV_DEK_STATE_ERROR __pyx_string_tab[60]
#define __pyx_n_u_MLX5DV_DEK_STATE_READY __pyx_string_tab[61]
#define __pyx_n_u_MLX5DV_DEVX_CREATE_EVENT_CHANNEL __pyx_string_tab[62]
#define __pyx_n_u_MLX5DV_DR_ACTION_DEST __pyx_string_tab[63]
#define __pyx_n_u_MLX5DV_DR_ACTION_DEST_REFORMAT __pyx_string_tab[64]
#define __pyx_n_u_MLX5DV_DR_ACTION_FLAGS_ROOT_LEVE __pyx_string_tab[65]
#define __pyx_n_u_MLX5DV_DR_DOMAIN_SYNC_FLAGS_HW __pyx_string_tab[66]
#define __pyx_n_u_MLX5DV_DR_DOMAIN_SYNC_FLAGS_MEM __pyx_string_tab[67]
#define __pyx_n_u_MLX5DV_DR_DOMAIN_SYNC_FLAGS_SW __pyx_string_tab[68]
#define __pyx_n_u_MLX5DV_DR_DOMAIN_TYPE_FDB __pyx_string_tab[69]
#define __pyx_n_u_MLX5DV_DR_DOMAIN_TYPE_NIC_RX __pyx_string_tab[70]
#define __pyx_n_u_MLX5DV_DR_DOMAIN_TYPE_NIC_TX __pyx_string_tab[71]
#define __pyx_n_u_MLX5DV_DR_MATCHER_LAYOUT_NUM_RUL __pyx_string_tab[72]
#define __pyx_n_u_MLX5DV_DR_MATCHER_LAYOUT_RESIZAB __pyx_string_tab[73]
#define __pyx_n_u_MLX5DV_FLOW_ACTION_COUNTERS_DEVX __pyx_string_tab[74]
#define __pyx_n_u_MLX5DV_FLOW_ACTION_COUNTERS_DEVX_2 __pyx_string_tab[75]
#define __pyx_n_u_MLX5DV_FLOW_ACTION_DEFAULT_MISS __pyx_string_tab[76]
#define __pyx_n_u_MLX5DV_FLOW_ACTION_DEST_DEVX __pyx_string_tab[77]
#define __pyx_n_u_MLX5DV_FLOW_ACTION_DEST_IBV_QP __pyx_string_tab[78]
#define __pyx_n_u_MLX5DV_FLOW_ACTION_DROP __pyx_string_tab[79]
#define __pyx_n_u_MLX5DV_FLOW_ACTION_FLAGS_ESP_AES __pyx_string_tab[80]
#define __pyx_n_u_MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_2 __pyx_string_tab[81]
#define __pyx_n_u_MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_3 __pyx_string_tab[82]
#define __pyx_n_u_MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_4 __pyx_string_tab[83]
#define __pyx_n_u_MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_5 __pyx_string_tab[84]
#define __pyx_n_u_MLX5DV_FLOW_ACTION_IBV_COUNTER __pyx_string_tab[85]
#define __pyx_n_u_MLX5DV_FLOW_ACTION_IBV_FLOW_ACTI __pyx_string_tab[86]
#define __pyx_n_u_MLX5DV_FLOW_ACTION_PACKET_REFORM __pyx_string_tab[87]
#define __pyx_n_u_MLX5DV_FLOW_ACTION_PACKET_REFORM_2 __pyx_string_tab[88]
#define __pyx_n_u_MLX5DV_FLOW_ACTION_PACKET_REFORM_3 __pyx_string_tab[89]
#define __pyx_n_u_MLX5DV_FLOW_ACTION_PACKET_REFORM_4 __pyx_string_tab[90]
#define __pyx_n_u_MLX5DV_FLOW_ACTION_TAG __pyx_string_tab[91]
#define __pyx_n_u_MLX5DV_FLOW_MATCHER_MASK_FT_TYPE __pyx_string_tab[92]
#define __pyx_n_u_MLX5DV_FLOW_MATCHER_MASK_IB_PORT __pyx_string_tab[93]
#define __pyx_n_u_MLX5DV_FLOW_TABLE_TYPE_FDB __pyx_string_tab[94]
#define __pyx_n_u_MLX5DV_FLOW_TABLE_TYPE_NIC_RX __pyx_string_tab[95]
#define __pyx_n_u_MLX5DV_FLOW_TABLE_TYPE_NIC_TX __pyx_string_tab[96]
#define __pyx_n_u_MLX5DV_FLOW_TABLE_TYPE_RDMA_RX __pyx_string_tab[97]
#define __pyx_n_u_MLX5DV_FLOW_TABLE_TYPE_RDMA_TRAN __pyx_string_tab[98]
#define __pyx_n_u_MLX5DV_FLOW_TABLE_TYPE_RDMA_TRAN_2 __pyx_string_tab[99]
#define __pyx_n_u_MLX5DV_FLOW_TABLE_TYPE_RDMA_TX __pyx_string_tab[100]
#define __pyx_n_u_MLX5DV_MKEY_INIT_ATTR_FLAGS_BLOC __pyx_string_tab[101]
#define __pyx_n_u_MLX5DV_MKEY_INIT_ATTR_FLAGS_CRYP __pyx_string_tab[102]
#define __pyx_n_u_MLX5DV_MKEY_INIT_ATTR_FLAGS_INDI __pyx_string_tab[103]
#define __pyx_n_u_MLX5DV_MKEY_INIT_ATTR_FLAGS_REMO __pyx_string_tab[104]
#define __pyx_n_u_MLX5DV_MKEY_NO_ERR __pyx_string_tab[105]
#define __pyx_n_u_MLX5DV_MKEY_SIG_BLOCK_BAD_APPTAG __pyx_string_tab[106]
#define __pyx_n_u_MLX5DV_MKEY_SIG_BLOCK_BAD_GUARD __pyx_string_tab[107]
#define __pyx_n_u_MLX5DV_MKEY_SIG_BLOCK_BAD_REFTAG __pyx_string_tab[108]
#define __pyx_n_u_MLX5DV_OBJ_AH __pyx_string_tab[109]
#define __pyx_n_u_MLX5DV_OBJ_CQ __pyx_string_tab[110]
#define __pyx_n_u_MLX5DV_OBJ_DM __pyx_string_tab[111]
#define __pyx_n_u_MLX5DV_OBJ_PD __pyx_string_tab[112]
#define __pyx_n_u_MLX5DV_OBJ_QP __pyx_string_tab[113]
#define __pyx_n_u_MLX5DV_OBJ_RWQ __pyx_string_tab[114]
#define __pyx_n_u_MLX5DV_OBJ_SRQ __pyx_string_tab[115]
#define __pyx_n_u_MLX5DV_PP_ALLOC_FLAGS_DEDICATED __pyx_string_tab[116]
#define __pyx_n_u_MLX5DV_QP_CREATE_ALLOW_SCATTER_T __pyx_string_tab[117]
#define __pyx_n_u_MLX5DV_QP_CREATE_DISABLE_SCATTER __pyx_string_tab[118]
#define __pyx_n_u_MLX5DV_QP_CREATE_OOO_DP __pyx_string_tab[119]
#define __pyx_n_u_MLX5DV_QP_CREATE_PACKET_BASED_CR __pyx_string_tab[120]
#define __pyx_n_u_MLX5DV_QP_CREATE_SIG_PIPELINING __pyx_string_tab[121]
#define __pyx_n_u_MLX5DV_QP_CREATE_TIR_ALLOW_SELF __pyx_string_tab[122]
#define __pyx_n_u_MLX5DV_QP_CREATE_TIR_ALLOW_SELF_2 __pyx_string_tab[123]
#define __pyx_n_u_MLX5DV_QP_CREATE_TUNNEL_OFFLOADS __pyx_string_tab[124]
#define __pyx_n_u_MLX5DV_QP_EX_WITH_MEMCPY __pyx_string_tab[125]
#define __pyx_n_u_MLX5DV_QP_EX_WITH_MKEY_CONFIGURE __pyx_string_tab[126]
#define __pyx_n_u_MLX5DV_QP_EX_WITH_MR_INTERLEAVED __pyx_string_tab[127]
#define __pyx_n_u_MLX5DV_QP_EX_WITH_MR_LIST __pyx_string_tab[128]
#define __pyx_n_u_MLX5DV_QP_EX_WITH_RAW_WQE __pyx_string_tab[129]
#define __pyx_n_u_MLX5DV_QP_INIT_ATTR_MASK_DC __pyx_string_tab[130]
#define __pyx_n_u_MLX5DV_QP_INIT_ATTR_MASK_DCI_STR __pyx_string_tab[131]
#define __pyx_n_u_MLX5DV_QP_INIT_ATTR_MASK_QP_CREA __pyx_string_tab[132]
#define __pyx_n_u_MLX5DV_QP_INIT_ATTR_MASK_SEND_OP __pyx_string_tab[133]
#define __pyx_n_u_MLX5DV_QP_MASK_RAW_QP_HANDLES __pyx_string_tab[134]
#define __pyx_n_u_MLX5DV_QP_MASK_RAW_QP_TIR_ADDR __pyx_string_tab[135]
#define __pyx_n_u_MLX5DV_QP_MASK_UAR_MMAP_OFFSET __pyx_string_tab[136]
#define __pyx_n_u_MLX5DV_QUERY_PORT_ESW_OWNER_VHCA __pyx_string_tab[137]
#define __pyx_n_u_MLX5DV_QUERY_PORT_VPORT __pyx_string_tab[138]
#define __pyx_n_u_MLX5DV_QUERY_PORT_VPORT_REG_C0 __pyx_string_tab[139]
#define __pyx_n_u_MLX5DV_QUERY_PORT_VPORT_STEERING __pyx_string_tab[140]
#define __pyx_n_u_MLX5DV_QUERY_PORT_VPORT_STEERING_2 __pyx_string_tab[141]
#define __pyx_n_u_MLX5DV_QUERY_PORT_VPORT_VHCA_ID __pyx_string_tab[142]
#define __pyx_n_u_MLX5DV_RAW_PACKET_CAP_TUNNELED_O __pyx_string_tab[143]
#define __pyx_n_u_MLX5DV_RAW_PACKET_CAP_TUNNELED_O_2 __pyx_string_tab[144]
#define __pyx_n_u_MLX5DV_RAW_PACKET_CAP_TUNNELED_O_3 __pyx_string_tab[145]
#define __pyx_n_u_MLX5DV_RAW_PACKET_CAP_TUNNELED_O_4 __pyx_string_tab[146]
#define __pyx_n_u_MLX5DV_RAW_PACKET_CAP_TUNNELED_O_5 __pyx_string_tab[147]
#define __pyx_n_u_MLX5DV_REG_DMABUF_ACCESS_DATA_DI __pyx_string_tab[148]
#define __pyx_n_u_MLX5DV_RES_TYPE_DBR __pyx_string_tab[149]
#define __pyx_n_u_MLX5DV_RES_TYPE_QP __pyx_string_tab[150]
#define __pyx_n_u_MLX5DV_RES_TYPE_RWQ __pyx_string_tab[151]
#define __pyx_n_u_MLX5DV_RES_TYPE_SRQ __pyx_string_tab[152]
#define __pyx_n_u_MLX5DV_SCHED_ELEM_ATTR_FLAGS_BW __pyx_string_tab[153]
#define __pyx_n_u_MLX5DV_SCHED_ELEM_ATTR_FLAGS_MAX __pyx_string_tab[154]
#define __pyx_n_u_MLX5DV_SIGNATURE_CRYPTO_ORDER_SI __pyx_string_tab[155]
#define __pyx_n_u_MLX5DV_SIGNATURE_CRYPTO_ORDER_SI_2 __pyx_string_tab[156]
#define __pyx_n_u_MLX5DV_SIG_BLOCK_ATTR_FLAG_COPY __pyx_string_tab[157]
#define __pyx_n_u_MLX5DV_SIG_CRC_TYPE_CRC32 __pyx_string_tab[158]
#define __pyx_n_u_MLX5DV_SIG_CRC_TYPE_CRC32C __pyx_string_tab[159]
#define __pyx_n_u_MLX5DV_SIG_CRC_TYPE_CRC64_XP10 __pyx_string_tab[160]
#define __pyx_n_u_MLX5DV_SIG_MASK_CRC32 __pyx_string_tab[161]
#define __pyx_n_u_MLX5DV_SIG_MASK_CRC32C __pyx_string_tab[162]
#define __pyx_n_u_MLX5DV_SIG_MASK_CRC64_XP10 __pyx_string_tab[163]
#define __pyx_n_u_MLX5DV_SIG_MASK_T10DIF_APPTAG __pyx_string_tab[164]
#define __pyx_n_u_MLX5DV_SIG_MASK_T10DIF_GUARD __pyx_string_tab[165]
#define __pyx_n_u_MLX5DV_SIG_MASK_T10DIF_REFTAG __pyx_string_tab[166]
#define __pyx_n_u_MLX5DV_SIG_T10DIF_CRC __pyx_string_tab[167]
#define __pyx_n_u_MLX5DV_SIG_T10DIF_CSUM __pyx_string_tab[168]
#define __pyx_n_u_MLX5DV_SIG_T10DIF_FLAG_APP_ESCAP __pyx_string_tab[169]
#define __pyx_n_u_MLX5DV_SIG_T10DIF_FLAG_APP_REF_E __pyx_string_tab[170]
#define __pyx_n_u_MLX5DV_SIG_T10DIF_FLAG_REF_REMAP __pyx_string_tab[171]
#define __pyx_n_u_MLX5DV_SIG_TYPE_CRC __pyx_string_tab[172]
#define __pyx_n_u_MLX5DV_SIG_TYPE_T10DIF __pyx_string_tab[173]
#define __pyx_n_u_MLX5DV_SRQ_MASK_SRQN __pyx_string_tab[174]
#define __pyx_n_u_MLX5DV_SW_PARSING __pyx_string_tab[175]
#define __pyx_n_u_MLX5DV_SW_PARSING_CSUM __pyx_string_tab[176]
#define __pyx_n_u_MLX5DV_SW_PARSING_LSO __pyx_string_tab[177]
#define __pyx_n_u_MLX5DV_UAR_ALLOC_TYPE_BF __pyx_string_tab[178]
#define __pyx_n_u_MLX5DV_UAR_ALLOC_TYPE_NC __pyx_string_tab[179]
#define __pyx_n_u_MLX5DV_VFIO_CTX_FLAGS_INIT_LINK __pyx_string_tab[180]
#define __pyx_n_u_MLX5DV_WC_MEMCPY __pyx_string_tab[181]
#define __pyx_n_u_MLX5DV_WC_RAW_WQE __pyx_string_tab[182]
#define __pyx_n_u_MLX5DV_WC_UMR __pyx_string_tab[183]
#define __pyx_n_u_cline_in_traceback __pyx_string_tab[184]
#define __pyx_n_u_enum __pyx_string_tab[185]
#define __pyx_n_u_get __pyx_string_tab[186]
#define __pyx_n_u_main __pyx_string_tab[187]
#define __pyx_n_u_member_names __pyx_string_tab[188]
#define __pyx_n_u_members __pyx_string_tab[189]
#define __pyx_n_u_mlx5dv_block_size __pyx_string_tab[190]
#define __pyx_n_u_mlx5dv_context_attr_comp_mask __pyx_string_tab[191]
#define __pyx_n_u_mlx5dv_context_attr_flags __pyx_string_tab[192]
#define __pyx_n_u_mlx5dv_context_comp_mask __pyx_string_tab[193]
#define __pyx_n_u_mlx5dv_context_flags __pyx_string_tab[194]
#define __pyx_n_u_mlx5dv_cq_init_attr_flags __pyx_string_tab[195]
#define __pyx_n_u_mlx5dv_cq_init_attr_mask __pyx_string_tab[196]
#define __pyx_n_u_mlx5dv_cqe_comp_res_format __pyx_string_tab[197]
#define __pyx_n_u_mlx5dv_crypto_caps_flags __pyx_string_tab[198]
#define __pyx_n_u_mlx5dv_crypto_engines_caps __pyx_string_tab[199]
#define __pyx_n_u_mlx5dv_crypto_key_purpose __pyx_string_tab[200]
#define __pyx_n_u_mlx5dv_crypto_key_size __pyx_string_tab[201]
#define __pyx_n_u_mlx5dv_crypto_login_state __pyx_string_tab[202]
#define __pyx_n_u_mlx5dv_crypto_standard __pyx_string_tab[203]
#define __pyx_n_u_mlx5dv_crypto_wrapped_import_met __pyx_string_tab[204]
#define __pyx_n_u_mlx5dv_dc_type __pyx_string_tab[205]
#define __pyx_n_u_mlx5dv_dek_init_attr_mask __pyx_string_tab[206]
#define __pyx_n_u_mlx5dv_dek_state __pyx_string_tab[207]
#define __pyx_n_u_mlx5dv_devx_create_event_channel __pyx_string_tab[208]
#define __pyx_n_u_mlx5dv_dr_action_dest_type __pyx_string_tab[209]
#define __pyx_n_u_mlx5dv_dr_action_flags __pyx_string_tab[210]
#define __pyx_n_u_mlx5dv_dr_domain_sync_flags __pyx_string_tab[211]
#define __pyx_n_u_mlx5dv_dr_domain_type __pyx_string_tab[212]
#define __pyx_n_u_mlx5dv_dr_matcher_layout_flags __pyx_string_tab[213]
#define __pyx_n_u_mlx5dv_flow_action_cap_flags __pyx_string_tab[214]
#define __pyx_n_u_mlx5dv_flow_action_type __pyx_string_tab[215]
#define __pyx_n_u_mlx5dv_flow_matcher_attr_mask __pyx_string_tab[216]
#define __pyx_n_u_mlx5dv_mkey_err_type __pyx_string_tab[217]
#define __pyx_n_u_mlx5dv_mkey_init_attr_flags __pyx_string_tab[218]
#define __pyx_n_u_mlx5dv_obj_type __pyx_string_tab[219]
#define __pyx_n_u_mlx5dv_qp_comp_mask __pyx_string_tab[220]
#define __pyx_n_u_mlx5dv_qp_create_flags __pyx_string_tab[221]
#define __pyx_n_u_mlx5dv_qp_create_send_ops_flags __pyx_string_tab[222]
#define __pyx_n_u_mlx5dv_qp_init_attr_mask __pyx_string_tab[223]
#define __pyx_n_u_mlx5dv_sched_elem_attr_flags __pyx_string_tab[224]
#define __pyx_n_u_mlx5dv_sig_block_attr_flags __pyx_string_tab[225]
#define __pyx_n_u_mlx5dv_sig_crc_type __pyx_string_tab[226]
#define __pyx_n_u_mlx5dv_sig_mask __pyx_string_tab[227]
#define __pyx_n_u_mlx5dv_sig_t10dif_bg_type __pyx_string_tab[228]
#define __pyx_n_u_mlx5dv_sig_t10dif_flags __pyx_string_tab[229]
#define __pyx_n_u_mlx5dv_sig_type __pyx_string_tab[230]
#define __pyx_n_u_mlx5dv_signature_crypto_order __pyx_string_tab[231]
#define __pyx_n_u_mlx5dv_srq_comp_mask __pyx_string_tab[232]
#define __pyx_n_u_mlx5dv_sw_parsing_offloads __pyx_string_tab[233]
#define __pyx_n_u_mlx5dv_tunnel_offloads __pyx_string_tab[234]
#define __pyx_n_u_mlx5dv_vfio_context_attr_flags __pyx_string_tab[235]
#define __pyx_n_u_mlx5dv_wc_opcode __pyx_string_tab[236]
#define __pyx_n_u_module __pyx_string_tab[237]
#define __pyx_n_u_module_2 __pyx_string_tab[238]
#define __pyx_n_u_name __pyx_string_tab[239]
#define __pyx_n_u_qualname __pyx_string_tab[240]
#define __pyx_n_u_setdefault __pyx_string_tab[241]
#define __pyx_n_u_test __pyx_string_tab[242]
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_clear(PyObject *m) {
  __pyx_mstatetype *clear_module_state = __Pyx_PyModule_GetState(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __Pyx_State_RemoveModule(NULL);
  #endif
  for (int i=0; i<1; ++i) { Py_CLEAR(clear_module_state->__pyx_tuple[i]); }
  for (int i=0; i<243; ++i) { Py_CLEAR(clear_module_state->__pyx_string_tab[i]); }
/* #### Code section: module_state_clear_contents ### */
/* #### Code section: module_state_clear_end ### */
return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstatetype *traverse_module_state = __Pyx_PyModule_GetState(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_tuple);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_bytes);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_unicode);
  for (int i=0; i<1; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_tuple[i]); }
  for (int i=0; i<243; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_string_tab[i]); }
/* #### Code section: module_state_traverse_contents ### */
/* #### Code section: module_state_traverse_end ### */
return 0;
}
#endif
/* #### Code section: module_code ### */
/* #### Code section: module_exttypes ### */

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
/* #### Code section: initfunc_declarations ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate); /*proto*/
/* #### Code section: init_module ### */

static int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_EnumBase = Py_None; Py_INCREF(Py_None);
  __Pyx_FlagBase = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_mlx5_enums(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_mlx5_enums},
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  {Py_mod_gil, __Pyx_FREETHREADING_COMPATIBLE},
  #endif
  #if PY_VERSION_HEX >= 0x030C0000 && CYTHON_USE_MODULE_STATE
  {Py_mod_multiple_interpreters, Py_MOD_MULTIPLE_INTERPRETERS_NOT_SUPPORTED},
  #endif
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "mlx5_enums",
      0, /* m_doc */
    #if CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstatetype), /* m_size */
    #else
      (CYTHON_PEP489_MULTI_PHASE_INIT) ? 0 : -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif

/* PyModInitFuncType */
#ifndef CYTHON_NO_PYINIT_EXPORT
  #define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#else
  #ifdef __cplusplus
  #define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
  #else
  #define __Pyx_PyMODINIT_FUNC PyObject *
  #endif
#endif

__Pyx_PyMODINIT_FUNC PyInit_mlx5_enums(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_mlx5_enums(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
/* ModuleCreationPEP489 */
#if CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
static PY_INT64_T __Pyx_GetCurrentInterpreterId(void) {
    {
        PyObject *module = PyImport_ImportModule("_interpreters"); // 3.13+ I think
        if (!module) {
            PyErr_Clear(); // just try the 3.8-3.12 version
            module = PyImport_ImportModule("_xxsubinterpreters");
            if (!module) goto bad;
        }
        PyObject *current = PyObject_CallMethod(module, "get_current", NULL);
        Py_DECREF(module);
        if (!current) goto bad;
        if (PyTuple_Check(current)) {
            PyObject *new_current = PySequence_GetItem(current, 0);
            Py_DECREF(current);
            current = new_current;
            if (!new_current) goto bad;
        }
        long long as_c_int = PyLong_AsLongLong(current);
        Py_DECREF(current);
        return as_c_int;
    }
  bad:
    PySys_WriteStderr("__Pyx_GetCurrentInterpreterId failed. Try setting the C define CYTHON_PEP489_MULTI_PHASE_INIT=0\n");
    return -1;
}
#endif
#if !CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    static PY_INT64_T main_interpreter_id = -1;
#if CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
    PY_INT64_T current_id = GraalPyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_GRAAL
    PY_INT64_T current_id = PyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
    PY_INT64_T current_id = __Pyx_GetCurrentInterpreterId();
#elif CYTHON_COMPILING_IN_LIMITED_API
    PY_INT64_T current_id = PyInterpreterState_GetID(PyInterpreterState_Get());
#else
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
#endif
    if (unlikely(current_id == -1)) {
        return -1;
    }
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return 0;
    } else if (unlikely(main_interpreter_id != current_id)) {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#endif
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    #if !CYTHON_USE_MODULE_STATE
    if (__Pyx_check_single_interpreter())
        return NULL;
    #endif
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_mlx5_enums(PyObject *__pyx_pyinit_module)
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  __pyx_mstatetype *__pyx_mstate = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  size_t __pyx_t_12;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'mlx5_enums' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #else
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_t_1 = __pyx_pyinit_module;
  Py_INCREF(__pyx_t_1);
  #else
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #if CYTHON_USE_MODULE_STATE
  {
    int add_module_result = __Pyx_State_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "mlx5_enums" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = __pyx_t_1;
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  PyUnstable_Module_SetGIL(__pyx_m, Py_MOD_GIL_USED);
  #endif
  __pyx_mstate = __pyx_mstate_global;
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_mstate->__pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_mstate->__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_mstate->__pyx_d);
  __pyx_mstate->__pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_mstate->__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_cython_runtime = __Pyx_PyImport_AddModuleRef("cython_runtime"); if (unlikely(!__pyx_mstate->__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_mstate->__pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /* ImportRefnannyAPI */
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
    PyErr_Clear();
    __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
    if (!__Pyx_RefNanny)
        Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  
__Pyx_RefNannySetupContext("PyInit_mlx5_enums", 0);
  __Pyx_init_runtime_version();
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_mstate->__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Library function declarations ---*/
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__pyx_module_is_main_pyverbs__providers__mlx5__mlx5_enums) {
    if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_name, __pyx_mstate_global->__pyx_n_u_main) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "pyverbs.providers.mlx5.mlx5_enums")) {
      if (unlikely((PyDict_SetItemString(modules, "pyverbs.providers.mlx5.mlx5_enums", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_CreateCodeObjects(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_export_code(__pyx_mstate);
  (void)__Pyx_modinit_function_export_code(__pyx_mstate);
  (void)__Pyx_modinit_type_init_code(__pyx_mstate);
  (void)__Pyx_modinit_type_import_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_import_code(__pyx_mstate);
  (void)__Pyx_modinit_function_import_code(__pyx_mstate);
  /*--- Execution code ---*/

  /* "EnumBase":10
 * 
 * cdef object __Pyx_EnumBase
 * from enum import IntEnum as __Pyx_EnumBase             # <<<<<<<<<<<<<<
 * 
 * cdef object __Pyx_FlagBase
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_IntEnum};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_enum, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 10, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_IntEnum};
    __pyx_t_3 = 0; {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 10, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      switch (__pyx_t_3) {
        case 0:
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_XGOTREF(__Pyx_EnumBase);
        __Pyx_DECREF_SET(__Pyx_EnumBase, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        break;
        default:;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumBase":13
 * 
 * cdef object __Pyx_FlagBase
 * from enum import IntFlag as __Pyx_FlagBase             # <<<<<<<<<<<<<<
*/
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_IntFlag};
    __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_enum, __pyx_imported_names, 1, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  {
    PyObject* const __pyx_imported_names[] = {__pyx_mstate_global->__pyx_n_u_IntFlag};
    __pyx_t_3 = 0; {
      __pyx_t_4 = __Pyx_ImportFrom(__pyx_t_2, __pyx_imported_names[__pyx_t_3]); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      switch (__pyx_t_3) {
        case 0:
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_XGOTREF(__Pyx_FlagBase);
        __Pyx_DECREF_SET(__Pyx_FlagBase, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        break;
        default:;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_block_size = __Pyx_FlagBase('mlx5dv_block_size', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_BLOCK_SIZE_512', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_512)),
 *     ('MLX5DV_BLOCK_SIZE_520', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_520)),
*/
  __pyx_t_4 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_5 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_block_size = __Pyx_FlagBase('mlx5dv_block_size', [
 *     ('MLX5DV_BLOCK_SIZE_512', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_512)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_BLOCK_SIZE_520', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_520)),
 *     ('MLX5DV_BLOCK_SIZE_4048', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_4048)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_512); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_BLOCK_SIZE_512, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":25
 * mlx5dv_block_size = __Pyx_FlagBase('mlx5dv_block_size', [
 *     ('MLX5DV_BLOCK_SIZE_512', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_512)),
 *     ('MLX5DV_BLOCK_SIZE_520', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_520)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_BLOCK_SIZE_4048', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_4048)),
 *     ('MLX5DV_BLOCK_SIZE_4096', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_4096)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_520); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_BLOCK_SIZE_520, __pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":26
 *     ('MLX5DV_BLOCK_SIZE_512', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_512)),
 *     ('MLX5DV_BLOCK_SIZE_520', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_520)),
 *     ('MLX5DV_BLOCK_SIZE_4048', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_4048)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_BLOCK_SIZE_4096', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_4096)),
 *     ('MLX5DV_BLOCK_SIZE_4160', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_4160)),
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_4048); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_9 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_BLOCK_SIZE_4048, __pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":27
 *     ('MLX5DV_BLOCK_SIZE_520', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_520)),
 *     ('MLX5DV_BLOCK_SIZE_4048', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_4048)),
 *     ('MLX5DV_BLOCK_SIZE_4096', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_4096)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_BLOCK_SIZE_4160', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_4160)),
 * 
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_4096); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_10 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_BLOCK_SIZE_4096, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":28
 *     ('MLX5DV_BLOCK_SIZE_4048', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_4048)),
 *     ('MLX5DV_BLOCK_SIZE_4096', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_4096)),
 *     ('MLX5DV_BLOCK_SIZE_4160', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_4160)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_4160); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_BLOCK_SIZE_4160, __pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_block_size = __Pyx_FlagBase('mlx5dv_block_size', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_BLOCK_SIZE_512', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_512)),
 *     ('MLX5DV_BLOCK_SIZE_520', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_520)),
*/
  __pyx_t_6 = __Pyx_PyList_Pack(5, __pyx_t_7, __pyx_t_8, __pyx_t_9, __pyx_t_10, __pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":30
 *     ('MLX5DV_BLOCK_SIZE_4160', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_4160)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_11 = __Pyx_Globals(); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_4, __pyx_mstate_global->__pyx_n_u_mlx5dv_block_size, __pyx_t_6};
    __pyx_t_10 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_11, __pyx_t_10, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_5, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_block_size, __pyx_t_2) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":32
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":37
 * 
 * 
 *     mlx5dv_block_size._member_names_ = list(mlx5dv_block_size.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_block_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PySequence_ListKeepNew(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_mlx5dv_block_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_2) < (0)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "EnumType":32
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_context_attr_comp_mask = __Pyx_FlagBase('mlx5dv_context_attr_comp_mask', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_ATTR_MASK_FD_ARRAY', __Pyx_PyLong_From_enum__mlx5dv_context_attr_comp_mask(MLX5DV_CONTEXT_ATTR_MASK_FD_ARRAY)),
 * 
*/
  __pyx_t_2 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_10 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_context_attr_comp_mask = __Pyx_FlagBase('mlx5dv_context_attr_comp_mask', [
 *     ('MLX5DV_CONTEXT_ATTR_MASK_FD_ARRAY', __Pyx_PyLong_From_enum__mlx5dv_context_attr_comp_mask(MLX5DV_CONTEXT_ATTR_MASK_FD_ARRAY)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__mlx5dv_context_attr_comp_mask(MLX5DV_CONTEXT_ATTR_MASK_FD_ARRAY); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_6 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_ATTR_MASK_FD_ARRA, __pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_context_attr_comp_mask = __Pyx_FlagBase('mlx5dv_context_attr_comp_mask', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_ATTR_MASK_FD_ARRAY', __Pyx_PyLong_From_enum__mlx5dv_context_attr_comp_mask(MLX5DV_CONTEXT_ATTR_MASK_FD_ARRAY)),
 * 
*/
  __pyx_t_11 = __Pyx_PyList_Pack(1, __pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":26
 *     ('MLX5DV_CONTEXT_ATTR_MASK_FD_ARRAY', __Pyx_PyLong_From_enum__mlx5dv_context_attr_comp_mask(MLX5DV_CONTEXT_ATTR_MASK_FD_ARRAY)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_6 = __Pyx_Globals(); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_context_attr_comp_mask, __pyx_t_11};
    __pyx_t_4 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_6, __pyx_t_4, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_5 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_10, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_context_attr_comp_mask, __pyx_t_5) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":33
 * 
 * 
 *     mlx5dv_context_attr_comp_mask._member_names_ = list(mlx5dv_context_attr_comp_mask.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_mlx5dv_context_attr_comp_mask); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PySequence_ListKeepNew(__pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_mlx5dv_context_attr_comp_mask); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_5) < (0)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_context_attr_flags = __Pyx_FlagBase('mlx5dv_context_attr_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_FLAGS_DEVX', __Pyx_PyLong_From_enum__mlx5dv_context_attr_flags(MLX5DV_CONTEXT_FLAGS_DEVX)),
 * 
*/
  __pyx_t_5 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_4 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_context_attr_flags = __Pyx_FlagBase('mlx5dv_context_attr_flags', [
 *     ('MLX5DV_CONTEXT_FLAGS_DEVX', __Pyx_PyLong_From_enum__mlx5dv_context_attr_flags(MLX5DV_CONTEXT_FLAGS_DEVX)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_6 = __Pyx_PyLong_From_enum__mlx5dv_context_attr_flags(MLX5DV_CONTEXT_FLAGS_DEVX); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_FLAGS_DEVX, __pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_context_attr_flags = __Pyx_FlagBase('mlx5dv_context_attr_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_FLAGS_DEVX', __Pyx_PyLong_From_enum__mlx5dv_context_attr_flags(MLX5DV_CONTEXT_FLAGS_DEVX)),
 * 
*/
  __pyx_t_6 = __Pyx_PyList_Pack(1, __pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":26
 *     ('MLX5DV_CONTEXT_FLAGS_DEVX', __Pyx_PyLong_From_enum__mlx5dv_context_attr_flags(MLX5DV_CONTEXT_FLAGS_DEVX)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_11 = __Pyx_Globals(); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_5, __pyx_mstate_global->__pyx_n_u_mlx5dv_context_attr_flags, __pyx_t_6};
    __pyx_t_2 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_11, __pyx_t_2, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_10 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_context_attr_flags, __pyx_t_10) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":33
 * 
 * 
 *     mlx5dv_context_attr_flags._member_names_ = list(mlx5dv_context_attr_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_mlx5dv_context_attr_flags); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PySequence_ListKeepNew(__pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_mlx5dv_context_attr_flags); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_10) < (0)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_context_comp_mask = __Pyx_FlagBase('mlx5dv_context_comp_mask', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_MASK_CQE_COMPRESION', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_CQE_COMPRESION)),
 *     ('MLX5DV_CONTEXT_MASK_SWP', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_SWP)),
*/
  __pyx_t_10 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_2 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_context_comp_mask = __Pyx_FlagBase('mlx5dv_context_comp_mask', [
 *     ('MLX5DV_CONTEXT_MASK_CQE_COMPRESION', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_CQE_COMPRESION)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_MASK_SWP', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_SWP)),
 *     ('MLX5DV_CONTEXT_MASK_STRIDING_RQ', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_STRIDING_RQ)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_CQE_COMPRESION); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_6 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_MASK_CQE_COMPRESI, __pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":25
 * mlx5dv_context_comp_mask = __Pyx_FlagBase('mlx5dv_context_comp_mask', [
 *     ('MLX5DV_CONTEXT_MASK_CQE_COMPRESION', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_CQE_COMPRESION)),
 *     ('MLX5DV_CONTEXT_MASK_SWP', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_SWP)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_MASK_STRIDING_RQ', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_STRIDING_RQ)),
 *     ('MLX5DV_CONTEXT_MASK_TUNNEL_OFFLOADS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_TUNNEL_OFFLOADS)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_SWP); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_5 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_MASK_SWP, __pyx_t_11); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":26
 *     ('MLX5DV_CONTEXT_MASK_CQE_COMPRESION', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_CQE_COMPRESION)),
 *     ('MLX5DV_CONTEXT_MASK_SWP', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_SWP)),
 *     ('MLX5DV_CONTEXT_MASK_STRIDING_RQ', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_STRIDING_RQ)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_MASK_TUNNEL_OFFLOADS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_TUNNEL_OFFLOADS)),
 *     ('MLX5DV_CONTEXT_MASK_DYN_BFREGS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_DYN_BFREGS)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_STRIDING_RQ); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_9 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_MASK_STRIDING_RQ, __pyx_t_11); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":27
 *     ('MLX5DV_CONTEXT_MASK_SWP', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_SWP)),
 *     ('MLX5DV_CONTEXT_MASK_STRIDING_RQ', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_STRIDING_RQ)),
 *     ('MLX5DV_CONTEXT_MASK_TUNNEL_OFFLOADS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_TUNNEL_OFFLOADS)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_MASK_DYN_BFREGS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_DYN_BFREGS)),
 *     ('MLX5DV_CONTEXT_MASK_CLOCK_INFO_UPDATE', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_CLOCK_INFO_UPDATE)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_TUNNEL_OFFLOADS); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_8 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_MASK_TUNNEL_OFFLO, __pyx_t_11); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":28
 *     ('MLX5DV_CONTEXT_MASK_STRIDING_RQ', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_STRIDING_RQ)),
 *     ('MLX5DV_CONTEXT_MASK_TUNNEL_OFFLOADS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_TUNNEL_OFFLOADS)),
 *     ('MLX5DV_CONTEXT_MASK_DYN_BFREGS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_DYN_BFREGS)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_MASK_CLOCK_INFO_UPDATE', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_CLOCK_INFO_UPDATE)),
 *     ('MLX5DV_CONTEXT_MASK_FLOW_ACTION_FLAGS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_FLOW_ACTION_FLAGS)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_DYN_BFREGS); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_7 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_MASK_DYN_BFREGS, __pyx_t_11); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":29
 *     ('MLX5DV_CONTEXT_MASK_TUNNEL_OFFLOADS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_TUNNEL_OFFLOADS)),
 *     ('MLX5DV_CONTEXT_MASK_DYN_BFREGS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_DYN_BFREGS)),
 *     ('MLX5DV_CONTEXT_MASK_CLOCK_INFO_UPDATE', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_CLOCK_INFO_UPDATE)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_MASK_FLOW_ACTION_FLAGS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_FLOW_ACTION_FLAGS)),
 *     ('MLX5DV_CONTEXT_MASK_DC_ODP_CAPS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_DC_ODP_CAPS)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_CLOCK_INFO_UPDATE); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_14 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_MASK_CLOCK_INFO_U, __pyx_t_11); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":30
 *     ('MLX5DV_CONTEXT_MASK_DYN_BFREGS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_DYN_BFREGS)),
 *     ('MLX5DV_CONTEXT_MASK_CLOCK_INFO_UPDATE', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_CLOCK_INFO_UPDATE)),
 *     ('MLX5DV_CONTEXT_MASK_FLOW_ACTION_FLAGS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_FLOW_ACTION_FLAGS)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_MASK_DC_ODP_CAPS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_DC_ODP_CAPS)),
 *     ('MLX5DV_CONTEXT_MASK_NUM_LAG_PORTS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_NUM_LAG_PORTS)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_FLOW_ACTION_FLAGS); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_MASK_FLOW_ACTION, __pyx_t_11); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":31
 *     ('MLX5DV_CONTEXT_MASK_CLOCK_INFO_UPDATE', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_CLOCK_INFO_UPDATE)),
 *     ('MLX5DV_CONTEXT_MASK_FLOW_ACTION_FLAGS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_FLOW_ACTION_FLAGS)),
 *     ('MLX5DV_CONTEXT_MASK_DC_ODP_CAPS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_DC_ODP_CAPS)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_MASK_NUM_LAG_PORTS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_NUM_LAG_PORTS)),
 *     ('MLX5DV_CONTEXT_MASK_SIGNATURE_OFFLOAD', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_SIGNATURE_OFFLOAD)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_DC_ODP_CAPS); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_MASK_DC_ODP_CAPS, __pyx_t_11); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":32
 *     ('MLX5DV_CONTEXT_MASK_FLOW_ACTION_FLAGS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_FLOW_ACTION_FLAGS)),
 *     ('MLX5DV_CONTEXT_MASK_DC_ODP_CAPS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_DC_ODP_CAPS)),
 *     ('MLX5DV_CONTEXT_MASK_NUM_LAG_PORTS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_NUM_LAG_PORTS)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_MASK_SIGNATURE_OFFLOAD', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_SIGNATURE_OFFLOAD)),
 *     ('MLX5DV_CONTEXT_MASK_DCI_STREAMS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_DCI_STREAMS)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_NUM_LAG_PORTS); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_17 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_MASK_NUM_LAG_PORT, __pyx_t_11); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":33
 *     ('MLX5DV_CONTEXT_MASK_DC_ODP_CAPS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_DC_ODP_CAPS)),
 *     ('MLX5DV_CONTEXT_MASK_NUM_LAG_PORTS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_NUM_LAG_PORTS)),
 *     ('MLX5DV_CONTEXT_MASK_SIGNATURE_OFFLOAD', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_SIGNATURE_OFFLOAD)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_MASK_DCI_STREAMS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_DCI_STREAMS)),
 *     ('MLX5DV_CONTEXT_MASK_WR_MEMCPY_LENGTH', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_WR_MEMCPY_LENGTH)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_SIGNATURE_OFFLOAD); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_18 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_MASK_SIGNATURE_OF, __pyx_t_11); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":34
 *     ('MLX5DV_CONTEXT_MASK_NUM_LAG_PORTS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_NUM_LAG_PORTS)),
 *     ('MLX5DV_CONTEXT_MASK_SIGNATURE_OFFLOAD', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_SIGNATURE_OFFLOAD)),
 *     ('MLX5DV_CONTEXT_MASK_DCI_STREAMS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_DCI_STREAMS)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_MASK_WR_MEMCPY_LENGTH', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_WR_MEMCPY_LENGTH)),
 *     ('MLX5DV_CONTEXT_MASK_CRYPTO_OFFLOAD', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_CRYPTO_OFFLOAD)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_DCI_STREAMS); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_19 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_MASK_DCI_STREAMS, __pyx_t_11); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":35
 *     ('MLX5DV_CONTEXT_MASK_SIGNATURE_OFFLOAD', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_SIGNATURE_OFFLOAD)),
 *     ('MLX5DV_CONTEXT_MASK_DCI_STREAMS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_DCI_STREAMS)),
 *     ('MLX5DV_CONTEXT_MASK_WR_MEMCPY_LENGTH', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_WR_MEMCPY_LENGTH)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_MASK_CRYPTO_OFFLOAD', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_CRYPTO_OFFLOAD)),
 *     ('MLX5DV_CONTEXT_MASK_MAX_DC_RD_ATOM', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_MAX_DC_RD_ATOM)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_WR_MEMCPY_LENGTH); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_20 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_MASK_WR_MEMCPY_LE, __pyx_t_11); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":36
 *     ('MLX5DV_CONTEXT_MASK_DCI_STREAMS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_DCI_STREAMS)),
 *     ('MLX5DV_CONTEXT_MASK_WR_MEMCPY_LENGTH', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_WR_MEMCPY_LENGTH)),
 *     ('MLX5DV_CONTEXT_MASK_CRYPTO_OFFLOAD', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_CRYPTO_OFFLOAD)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_MASK_MAX_DC_RD_ATOM', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_MAX_DC_RD_ATOM)),
 *     ('MLX5DV_CONTEXT_MASK_OOO_RECV_WRS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_OOO_RECV_WRS)),
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_CRYPTO_OFFLOAD); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_21 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_MASK_CRYPTO_OFFLO, __pyx_t_11); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":37
 *     ('MLX5DV_CONTEXT_MASK_WR_MEMCPY_LENGTH', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_WR_MEMCPY_LENGTH)),
 *     ('MLX5DV_CONTEXT_MASK_CRYPTO_OFFLOAD', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_CRYPTO_OFFLOAD)),
 *     ('MLX5DV_CONTEXT_MASK_MAX_DC_RD_ATOM', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_MAX_DC_RD_ATOM)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_MASK_OOO_RECV_WRS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_OOO_RECV_WRS)),
 * 
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_MAX_DC_RD_ATOM); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_MASK_MAX_DC_RD_AT, __pyx_t_11); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":38
 *     ('MLX5DV_CONTEXT_MASK_CRYPTO_OFFLOAD', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_CRYPTO_OFFLOAD)),
 *     ('MLX5DV_CONTEXT_MASK_MAX_DC_RD_ATOM', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_MAX_DC_RD_ATOM)),
 *     ('MLX5DV_CONTEXT_MASK_OOO_RECV_WRS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_OOO_RECV_WRS)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_OOO_RECV_WRS); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_23 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_MASK_OOO_RECV_WRS, __pyx_t_11); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_context_comp_mask = __Pyx_FlagBase('mlx5dv_context_comp_mask', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_MASK_CQE_COMPRESION', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_CQE_COMPRESION)),
 *     ('MLX5DV_CONTEXT_MASK_SWP', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_SWP)),
*/
  __pyx_t_11 = __Pyx_PyList_Pack(15, __pyx_t_6, __pyx_t_5, __pyx_t_9, __pyx_t_8, __pyx_t_7, __pyx_t_14, __pyx_t_15, __pyx_t_16, __pyx_t_17, __pyx_t_18, __pyx_t_19, __pyx_t_20, __pyx_t_21, __pyx_t_22, __pyx_t_23); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":40
 *     ('MLX5DV_CONTEXT_MASK_OOO_RECV_WRS', __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(MLX5DV_CONTEXT_MASK_OOO_RECV_WRS)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_23 = __Pyx_Globals(); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_22 = __Pyx_PyObject_GetAttrStr(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_22, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_10, __pyx_mstate_global->__pyx_n_u_mlx5dv_context_comp_mask, __pyx_t_11};
    __pyx_t_22 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_23, __pyx_t_22, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_4 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_22);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_context_comp_mask, __pyx_t_4) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":42
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":47
 * 
 * 
 *     mlx5dv_context_comp_mask._member_names_ = list(mlx5dv_context_comp_mask.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_mlx5dv_context_comp_mask); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PySequence_ListKeepNew(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_context_comp_mask); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_4) < (0)) __PYX_ERR(1, 47, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":42
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_context_flags = __Pyx_FlagBase('mlx5dv_context_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_FLAGS_CQE_V1', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_CQE_V1)),
 *     ('MLX5DV_CONTEXT_FLAGS_MPW_ALLOWED', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_MPW_ALLOWED)),
*/
  __pyx_t_4 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_22 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_context_flags = __Pyx_FlagBase('mlx5dv_context_flags', [
 *     ('MLX5DV_CONTEXT_FLAGS_CQE_V1', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_CQE_V1)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_FLAGS_MPW_ALLOWED', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_MPW_ALLOWED)),
 *     ('MLX5DV_CONTEXT_FLAGS_ENHANCED_MPW', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_ENHANCED_MPW)),
*/
  __pyx_t_23 = __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_CQE_V1); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_FLAGS_CQE_V1, __pyx_t_23); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":25
 * mlx5dv_context_flags = __Pyx_FlagBase('mlx5dv_context_flags', [
 *     ('MLX5DV_CONTEXT_FLAGS_CQE_V1', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_CQE_V1)),
 *     ('MLX5DV_CONTEXT_FLAGS_MPW_ALLOWED', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_MPW_ALLOWED)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_FLAGS_ENHANCED_MPW', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_ENHANCED_MPW)),
 *     ('MLX5DV_CONTEXT_FLAGS_CQE_128B_COMP', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_CQE_128B_COMP)),
*/
  __pyx_t_23 = __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_MPW_ALLOWED); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_10 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_FLAGS_MPW_ALLOWED, __pyx_t_23); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":26
 *     ('MLX5DV_CONTEXT_FLAGS_CQE_V1', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_CQE_V1)),
 *     ('MLX5DV_CONTEXT_FLAGS_MPW_ALLOWED', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_MPW_ALLOWED)),
 *     ('MLX5DV_CONTEXT_FLAGS_ENHANCED_MPW', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_ENHANCED_MPW)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_FLAGS_CQE_128B_COMP', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_CQE_128B_COMP)),
 *     ('MLX5DV_CONTEXT_FLAGS_CQE_128B_PAD', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_CQE_128B_PAD)),
*/
  __pyx_t_23 = __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_ENHANCED_MPW); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_21 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_FLAGS_ENHANCED_MP, __pyx_t_23); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":27
 *     ('MLX5DV_CONTEXT_FLAGS_MPW_ALLOWED', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_MPW_ALLOWED)),
 *     ('MLX5DV_CONTEXT_FLAGS_ENHANCED_MPW', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_ENHANCED_MPW)),
 *     ('MLX5DV_CONTEXT_FLAGS_CQE_128B_COMP', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_CQE_128B_COMP)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_FLAGS_CQE_128B_PAD', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_CQE_128B_PAD)),
 *     ('MLX5DV_CONTEXT_FLAGS_PACKET_BASED_CREDIT_MODE', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_PACKET_BASED_CREDIT_MODE)),
*/
  __pyx_t_23 = __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_CQE_128B_COMP); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_20 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_FLAGS_CQE_128B_CO, __pyx_t_23); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":28
 *     ('MLX5DV_CONTEXT_FLAGS_ENHANCED_MPW', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_ENHANCED_MPW)),
 *     ('MLX5DV_CONTEXT_FLAGS_CQE_128B_COMP', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_CQE_128B_COMP)),
 *     ('MLX5DV_CONTEXT_FLAGS_CQE_128B_PAD', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_CQE_128B_PAD)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_FLAGS_PACKET_BASED_CREDIT_MODE', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_PACKET_BASED_CREDIT_MODE)),
 *     ('MLX5DV_CONTEXT_FLAGS_REAL_TIME_TS', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_REAL_TIME_TS)),
*/
  __pyx_t_23 = __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_CQE_128B_PAD); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_19 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_FLAGS_CQE_128B_PA, __pyx_t_23); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":29
 *     ('MLX5DV_CONTEXT_FLAGS_CQE_128B_COMP', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_CQE_128B_COMP)),
 *     ('MLX5DV_CONTEXT_FLAGS_CQE_128B_PAD', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_CQE_128B_PAD)),
 *     ('MLX5DV_CONTEXT_FLAGS_PACKET_BASED_CREDIT_MODE', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_PACKET_BASED_CREDIT_MODE)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_FLAGS_REAL_TIME_TS', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_REAL_TIME_TS)),
 *     ('MLX5DV_CONTEXT_FLAGS_BLUEFLAME', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_BLUEFLAME)),
*/
  __pyx_t_23 = __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_PACKET_BASED_CREDIT_MODE); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_18 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_FLAGS_PACKET_BASE, __pyx_t_23); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":30
 *     ('MLX5DV_CONTEXT_FLAGS_CQE_128B_PAD', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_CQE_128B_PAD)),
 *     ('MLX5DV_CONTEXT_FLAGS_PACKET_BASED_CREDIT_MODE', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_PACKET_BASED_CREDIT_MODE)),
 *     ('MLX5DV_CONTEXT_FLAGS_REAL_TIME_TS', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_REAL_TIME_TS)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_FLAGS_BLUEFLAME', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_BLUEFLAME)),
 * 
*/
  __pyx_t_23 = __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_REAL_TIME_TS); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_17 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_FLAGS_REAL_TIME_T, __pyx_t_23); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":31
 *     ('MLX5DV_CONTEXT_FLAGS_PACKET_BASED_CREDIT_MODE', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_PACKET_BASED_CREDIT_MODE)),
 *     ('MLX5DV_CONTEXT_FLAGS_REAL_TIME_TS', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_REAL_TIME_TS)),
 *     ('MLX5DV_CONTEXT_FLAGS_BLUEFLAME', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_BLUEFLAME)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_23 = __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_BLUEFLAME); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CONTEXT_FLAGS_BLUEFLAME, __pyx_t_23); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_context_flags = __Pyx_FlagBase('mlx5dv_context_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CONTEXT_FLAGS_CQE_V1', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_CQE_V1)),
 *     ('MLX5DV_CONTEXT_FLAGS_MPW_ALLOWED', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_MPW_ALLOWED)),
*/
  __pyx_t_23 = __Pyx_PyList_Pack(8, __pyx_t_11, __pyx_t_10, __pyx_t_21, __pyx_t_20, __pyx_t_19, __pyx_t_18, __pyx_t_17, __pyx_t_16); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":33
 *     ('MLX5DV_CONTEXT_FLAGS_BLUEFLAME', __Pyx_PyLong_From_enum__mlx5dv_context_flags(MLX5DV_CONTEXT_FLAGS_BLUEFLAME)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_16 = __Pyx_Globals(); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_4, __pyx_mstate_global->__pyx_n_u_mlx5dv_context_flags, __pyx_t_23};
    __pyx_t_17 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_16, __pyx_t_17, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_22, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_17);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_context_flags, __pyx_t_2) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":35
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":40
 * 
 * 
 *     mlx5dv_context_flags._member_names_ = list(mlx5dv_context_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_context_flags); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_22 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PySequence_ListKeepNew(__pyx_t_22); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_mlx5dv_context_flags); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_2) < (0)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

    /* "EnumType":35
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_cq_init_attr_flags = __Pyx_FlagBase('mlx5dv_cq_init_attr_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CQ_INIT_ATTR_FLAGS_CQE_PAD', __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_flags(MLX5DV_CQ_INIT_ATTR_FLAGS_CQE_PAD)),
 *     ('MLX5DV_CQ_INIT_ATTR_FLAGS_RESERVED', __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_flags(MLX5DV_CQ_INIT_ATTR_FLAGS_RESERVED)),
*/
  __pyx_t_2 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_17 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_cq_init_attr_flags = __Pyx_FlagBase('mlx5dv_cq_init_attr_flags', [
 *     ('MLX5DV_CQ_INIT_ATTR_FLAGS_CQE_PAD', __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_flags(MLX5DV_CQ_INIT_ATTR_FLAGS_CQE_PAD)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CQ_INIT_ATTR_FLAGS_RESERVED', __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_flags(MLX5DV_CQ_INIT_ATTR_FLAGS_RESERVED)),
 * 
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_flags(MLX5DV_CQ_INIT_ATTR_FLAGS_CQE_PAD); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_23 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CQ_INIT_ATTR_FLAGS_CQE_PA, __pyx_t_16); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":25
 * mlx5dv_cq_init_attr_flags = __Pyx_FlagBase('mlx5dv_cq_init_attr_flags', [
 *     ('MLX5DV_CQ_INIT_ATTR_FLAGS_CQE_PAD', __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_flags(MLX5DV_CQ_INIT_ATTR_FLAGS_CQE_PAD)),
 *     ('MLX5DV_CQ_INIT_ATTR_FLAGS_RESERVED', __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_flags(MLX5DV_CQ_INIT_ATTR_FLAGS_RESERVED)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_flags(MLX5DV_CQ_INIT_ATTR_FLAGS_RESERVED); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CQ_INIT_ATTR_FLAGS_RESERV, __pyx_t_16); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_cq_init_attr_flags = __Pyx_FlagBase('mlx5dv_cq_init_attr_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CQ_INIT_ATTR_FLAGS_CQE_PAD', __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_flags(MLX5DV_CQ_INIT_ATTR_FLAGS_CQE_PAD)),
 *     ('MLX5DV_CQ_INIT_ATTR_FLAGS_RESERVED', __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_flags(MLX5DV_CQ_INIT_ATTR_FLAGS_RESERVED)),
*/
  __pyx_t_16 = __Pyx_PyList_Pack(2, __pyx_t_23, __pyx_t_4); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":27
 *     ('MLX5DV_CQ_INIT_ATTR_FLAGS_RESERVED', __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_flags(MLX5DV_CQ_INIT_ATTR_FLAGS_RESERVED)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_4 = __Pyx_Globals(); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_cq_init_attr_flags, __pyx_t_16};
    __pyx_t_23 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_4, __pyx_t_23, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_22 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_17, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_23);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_cq_init_attr_flags, __pyx_t_22) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":34
 * 
 * 
 *     mlx5dv_cq_init_attr_flags._member_names_ = list(mlx5dv_cq_init_attr_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_mlx5dv_cq_init_attr_flags); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __pyx_t_22 = __Pyx_PySequence_ListKeepNew(__pyx_t_17); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_cq_init_attr_flags); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_22) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_cq_init_attr_mask = __Pyx_FlagBase('mlx5dv_cq_init_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CQ_INIT_ATTR_MASK_COMPRESSED_CQE', __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_mask(MLX5DV_CQ_INIT_ATTR_MASK_COMPRESSED_CQE)),
 *     ('MLX5DV_CQ_INIT_ATTR_MASK_FLAGS', __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_mask(MLX5DV_CQ_INIT_ATTR_MASK_FLAGS)),
*/
  __pyx_t_22 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_23 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_cq_init_attr_mask = __Pyx_FlagBase('mlx5dv_cq_init_attr_mask', [
 *     ('MLX5DV_CQ_INIT_ATTR_MASK_COMPRESSED_CQE', __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_mask(MLX5DV_CQ_INIT_ATTR_MASK_COMPRESSED_CQE)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CQ_INIT_ATTR_MASK_FLAGS', __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_mask(MLX5DV_CQ_INIT_ATTR_MASK_FLAGS)),
 *     ('MLX5DV_CQ_INIT_ATTR_MASK_CQE_SIZE', __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_mask(MLX5DV_CQ_INIT_ATTR_MASK_CQE_SIZE)),
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_mask(MLX5DV_CQ_INIT_ATTR_MASK_COMPRESSED_CQE); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CQ_INIT_ATTR_MASK_COMPRES, __pyx_t_4); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":25
 * mlx5dv_cq_init_attr_mask = __Pyx_FlagBase('mlx5dv_cq_init_attr_mask', [
 *     ('MLX5DV_CQ_INIT_ATTR_MASK_COMPRESSED_CQE', __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_mask(MLX5DV_CQ_INIT_ATTR_MASK_COMPRESSED_CQE)),
 *     ('MLX5DV_CQ_INIT_ATTR_MASK_FLAGS', __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_mask(MLX5DV_CQ_INIT_ATTR_MASK_FLAGS)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CQ_INIT_ATTR_MASK_CQE_SIZE', __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_mask(MLX5DV_CQ_INIT_ATTR_MASK_CQE_SIZE)),
 * 
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_mask(MLX5DV_CQ_INIT_ATTR_MASK_FLAGS); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CQ_INIT_ATTR_MASK_FLAGS, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":26
 *     ('MLX5DV_CQ_INIT_ATTR_MASK_COMPRESSED_CQE', __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_mask(MLX5DV_CQ_INIT_ATTR_MASK_COMPRESSED_CQE)),
 *     ('MLX5DV_CQ_INIT_ATTR_MASK_FLAGS', __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_mask(MLX5DV_CQ_INIT_ATTR_MASK_FLAGS)),
 *     ('MLX5DV_CQ_INIT_ATTR_MASK_CQE_SIZE', __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_mask(MLX5DV_CQ_INIT_ATTR_MASK_CQE_SIZE)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_mask(MLX5DV_CQ_INIT_ATTR_MASK_CQE_SIZE); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_18 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CQ_INIT_ATTR_MASK_CQE_SIZ, __pyx_t_4); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_cq_init_attr_mask = __Pyx_FlagBase('mlx5dv_cq_init_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CQ_INIT_ATTR_MASK_COMPRESSED_CQE', __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_mask(MLX5DV_CQ_INIT_ATTR_MASK_COMPRESSED_CQE)),
 *     ('MLX5DV_CQ_INIT_ATTR_MASK_FLAGS', __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_mask(MLX5DV_CQ_INIT_ATTR_MASK_FLAGS)),
*/
  __pyx_t_4 = __Pyx_PyList_Pack(3, __pyx_t_16, __pyx_t_2, __pyx_t_18); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "EnumType":28
 *     ('MLX5DV_CQ_INIT_ATTR_MASK_CQE_SIZE', __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_mask(MLX5DV_CQ_INIT_ATTR_MASK_CQE_SIZE)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_18 = __Pyx_Globals(); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __pyx_t_18 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_22, __pyx_mstate_global->__pyx_n_u_mlx5dv_cq_init_attr_mask, __pyx_t_4};
    __pyx_t_2 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_18, __pyx_t_2, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_17 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_23, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_cq_init_attr_mask, __pyx_t_17) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":35
 * 
 * 
 *     mlx5dv_cq_init_attr_mask._member_names_ = list(mlx5dv_cq_init_attr_mask.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_cq_init_attr_mask); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __pyx_t_17 = __Pyx_PySequence_ListKeepNew(__pyx_t_23); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_mlx5dv_cq_init_attr_mask); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_17) < (0)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

    /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_cqe_comp_res_format = __Pyx_FlagBase('mlx5dv_cqe_comp_res_format', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CQE_RES_FORMAT_HASH', __Pyx_PyLong_From_enum__mlx5dv_cqe_comp_res_format(MLX5DV_CQE_RES_FORMAT_HASH)),
 *     ('MLX5DV_CQE_RES_FORMAT_CSUM', __Pyx_PyLong_From_enum__mlx5dv_cqe_comp_res_format(MLX5DV_CQE_RES_FORMAT_CSUM)),
*/
  __pyx_t_17 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_2 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_cqe_comp_res_format = __Pyx_FlagBase('mlx5dv_cqe_comp_res_format', [
 *     ('MLX5DV_CQE_RES_FORMAT_HASH', __Pyx_PyLong_From_enum__mlx5dv_cqe_comp_res_format(MLX5DV_CQE_RES_FORMAT_HASH)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CQE_RES_FORMAT_CSUM', __Pyx_PyLong_From_enum__mlx5dv_cqe_comp_res_format(MLX5DV_CQE_RES_FORMAT_CSUM)),
 *     ('MLX5DV_CQE_RES_FORMAT_CSUM_STRIDX', __Pyx_PyLong_From_enum__mlx5dv_cqe_comp_res_format(MLX5DV_CQE_RES_FORMAT_CSUM_STRIDX)),
*/
  __pyx_t_18 = __Pyx_PyLong_From_enum__mlx5dv_cqe_comp_res_format(MLX5DV_CQE_RES_FORMAT_HASH); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CQE_RES_FORMAT_HASH, __pyx_t_18); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "EnumType":25
 * mlx5dv_cqe_comp_res_format = __Pyx_FlagBase('mlx5dv_cqe_comp_res_format', [
 *     ('MLX5DV_CQE_RES_FORMAT_HASH', __Pyx_PyLong_From_enum__mlx5dv_cqe_comp_res_format(MLX5DV_CQE_RES_FORMAT_HASH)),
 *     ('MLX5DV_CQE_RES_FORMAT_CSUM', __Pyx_PyLong_From_enum__mlx5dv_cqe_comp_res_format(MLX5DV_CQE_RES_FORMAT_CSUM)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CQE_RES_FORMAT_CSUM_STRIDX', __Pyx_PyLong_From_enum__mlx5dv_cqe_comp_res_format(MLX5DV_CQE_RES_FORMAT_CSUM_STRIDX)),
 * 
*/
  __pyx_t_18 = __Pyx_PyLong_From_enum__mlx5dv_cqe_comp_res_format(MLX5DV_CQE_RES_FORMAT_CSUM); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CQE_RES_FORMAT_CSUM, __pyx_t_18); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "EnumType":26
 *     ('MLX5DV_CQE_RES_FORMAT_HASH', __Pyx_PyLong_From_enum__mlx5dv_cqe_comp_res_format(MLX5DV_CQE_RES_FORMAT_HASH)),
 *     ('MLX5DV_CQE_RES_FORMAT_CSUM', __Pyx_PyLong_From_enum__mlx5dv_cqe_comp_res_format(MLX5DV_CQE_RES_FORMAT_CSUM)),
 *     ('MLX5DV_CQE_RES_FORMAT_CSUM_STRIDX', __Pyx_PyLong_From_enum__mlx5dv_cqe_comp_res_format(MLX5DV_CQE_RES_FORMAT_CSUM_STRIDX)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_18 = __Pyx_PyLong_From_enum__mlx5dv_cqe_comp_res_format(MLX5DV_CQE_RES_FORMAT_CSUM_STRIDX); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CQE_RES_FORMAT_CSUM_STRID, __pyx_t_18); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_cqe_comp_res_format = __Pyx_FlagBase('mlx5dv_cqe_comp_res_format', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CQE_RES_FORMAT_HASH', __Pyx_PyLong_From_enum__mlx5dv_cqe_comp_res_format(MLX5DV_CQE_RES_FORMAT_HASH)),
 *     ('MLX5DV_CQE_RES_FORMAT_CSUM', __Pyx_PyLong_From_enum__mlx5dv_cqe_comp_res_format(MLX5DV_CQE_RES_FORMAT_CSUM)),
*/
  __pyx_t_18 = __Pyx_PyList_Pack(3, __pyx_t_4, __pyx_t_22, __pyx_t_16); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":28
 *     ('MLX5DV_CQE_RES_FORMAT_CSUM_STRIDX', __Pyx_PyLong_From_enum__mlx5dv_cqe_comp_res_format(MLX5DV_CQE_RES_FORMAT_CSUM_STRIDX)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_16 = __Pyx_Globals(); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_22 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_22, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_cqe_comp_res_format, __pyx_t_18};
    __pyx_t_22 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_16, __pyx_t_22, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_23 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_22);
    __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_cqe_comp_res_format, __pyx_t_23) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":35
 * 
 * 
 *     mlx5dv_cqe_comp_res_format._member_names_ = list(mlx5dv_cqe_comp_res_format.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_mlx5dv_cqe_comp_res_format); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __pyx_t_23 = __Pyx_PySequence_ListKeepNew(__pyx_t_2); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_cqe_comp_res_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_23) < (0)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_crypto_caps_flags = __Pyx_FlagBase('mlx5dv_crypto_caps_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CRYPTO_CAPS_CRYPTO', __Pyx_PyLong_From_enum__mlx5dv_crypto_caps_flags(MLX5DV_CRYPTO_CAPS_CRYPTO)),
 *     ('MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPTO_OPERATIONAL', __Pyx_PyLong_From_enum__mlx5dv_crypto_caps_flags(MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPTO_OPERATIONAL)),
*/
  __pyx_t_23 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_22 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_crypto_caps_flags = __Pyx_FlagBase('mlx5dv_crypto_caps_flags', [
 *     ('MLX5DV_CRYPTO_CAPS_CRYPTO', __Pyx_PyLong_From_enum__mlx5dv_crypto_caps_flags(MLX5DV_CRYPTO_CAPS_CRYPTO)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPTO_OPERATIONAL', __Pyx_PyLong_From_enum__mlx5dv_crypto_caps_flags(MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPTO_OPERATIONAL)),
 *     ('MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPTO_GOING_TO_COMMISSIONING', __Pyx_PyLong_From_enum__mlx5dv_crypto_caps_flags(MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPTO_GOING_TO_COMMISSIONING)),
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__mlx5dv_crypto_caps_flags(MLX5DV_CRYPTO_CAPS_CRYPTO); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_18 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CRYPTO_CAPS_CRYPTO, __pyx_t_16); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":25
 * mlx5dv_crypto_caps_flags = __Pyx_FlagBase('mlx5dv_crypto_caps_flags', [
 *     ('MLX5DV_CRYPTO_CAPS_CRYPTO', __Pyx_PyLong_From_enum__mlx5dv_crypto_caps_flags(MLX5DV_CRYPTO_CAPS_CRYPTO)),
 *     ('MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPTO_OPERATIONAL', __Pyx_PyLong_From_enum__mlx5dv_crypto_caps_flags(MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPTO_OPERATIONAL)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPTO_GOING_TO_COMMISSIONING', __Pyx_PyLong_From_enum__mlx5dv_crypto_caps_flags(MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPTO_GOING_TO_COMMISSIONING)),
 * 
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__mlx5dv_crypto_caps_flags(MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPTO_OPERATIONAL); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_17 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPT, __pyx_t_16); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":26
 *     ('MLX5DV_CRYPTO_CAPS_CRYPTO', __Pyx_PyLong_From_enum__mlx5dv_crypto_caps_flags(MLX5DV_CRYPTO_CAPS_CRYPTO)),
 *     ('MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPTO_OPERATIONAL', __Pyx_PyLong_From_enum__mlx5dv_crypto_caps_flags(MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPTO_OPERATIONAL)),
 *     ('MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPTO_GOING_TO_COMMISSIONING', __Pyx_PyLong_From_enum__mlx5dv_crypto_caps_flags(MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPTO_GOING_TO_COMMISSIONING)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__mlx5dv_crypto_caps_flags(MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPTO_GOING_TO_COMMISSIONING); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPT_2, __pyx_t_16); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_crypto_caps_flags = __Pyx_FlagBase('mlx5dv_crypto_caps_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CRYPTO_CAPS_CRYPTO', __Pyx_PyLong_From_enum__mlx5dv_crypto_caps_flags(MLX5DV_CRYPTO_CAPS_CRYPTO)),
 *     ('MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPTO_OPERATIONAL', __Pyx_PyLong_From_enum__mlx5dv_crypto_caps_flags(MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPTO_OPERATIONAL)),
*/
  __pyx_t_16 = __Pyx_PyList_Pack(3, __pyx_t_18, __pyx_t_17, __pyx_t_4); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":28
 *     ('MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPTO_GOING_TO_COMMISSIONING', __Pyx_PyLong_From_enum__mlx5dv_crypto_caps_flags(MLX5DV_CRYPTO_CAPS_WRAPPED_CRYPTO_GOING_TO_COMMISSIONING)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_4 = __Pyx_Globals(); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_23, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_caps_flags, __pyx_t_16};
    __pyx_t_17 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_4, __pyx_t_17, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_22, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_17);
    __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_caps_flags, __pyx_t_2) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":35
 * 
 * 
 *     mlx5dv_crypto_caps_flags._member_names_ = list(mlx5dv_crypto_caps_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_caps_flags); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_22 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PySequence_ListKeepNew(__pyx_t_22); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_caps_flags); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_2) < (0)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

    /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_crypto_engines_caps = __Pyx_FlagBase('mlx5dv_crypto_engines_caps', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS', __Pyx_PyLong_From_enum__mlx5dv_crypto_engines_caps(MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS)),
 *     ('MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS_SINGLE_BLOCK', __Pyx_PyLong_From_enum__mlx5dv_crypto_engines_caps(MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS_SINGLE_BLOCK)),
*/
  __pyx_t_2 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_17 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_crypto_engines_caps = __Pyx_FlagBase('mlx5dv_crypto_engines_caps', [
 *     ('MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS', __Pyx_PyLong_From_enum__mlx5dv_crypto_engines_caps(MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS_SINGLE_BLOCK', __Pyx_PyLong_From_enum__mlx5dv_crypto_engines_caps(MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS_SINGLE_BLOCK)),
 *     ('MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS_MULTI_BLOCK', __Pyx_PyLong_From_enum__mlx5dv_crypto_engines_caps(MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS_MULTI_BLOCK)),
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__mlx5dv_crypto_engines_caps(MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CRYPTO_ENGINES_CAP_AES_XT, __pyx_t_4); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":25
 * mlx5dv_crypto_engines_caps = __Pyx_FlagBase('mlx5dv_crypto_engines_caps', [
 *     ('MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS', __Pyx_PyLong_From_enum__mlx5dv_crypto_engines_caps(MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS)),
 *     ('MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS_SINGLE_BLOCK', __Pyx_PyLong_From_enum__mlx5dv_crypto_engines_caps(MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS_SINGLE_BLOCK)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS_MULTI_BLOCK', __Pyx_PyLong_From_enum__mlx5dv_crypto_engines_caps(MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS_MULTI_BLOCK)),
 * 
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__mlx5dv_crypto_engines_caps(MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS_SINGLE_BLOCK); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_23 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CRYPTO_ENGINES_CAP_AES_XT_2, __pyx_t_4); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":26
 *     ('MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS', __Pyx_PyLong_From_enum__mlx5dv_crypto_engines_caps(MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS)),
 *     ('MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS_SINGLE_BLOCK', __Pyx_PyLong_From_enum__mlx5dv_crypto_engines_caps(MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS_SINGLE_BLOCK)),
 *     ('MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS_MULTI_BLOCK', __Pyx_PyLong_From_enum__mlx5dv_crypto_engines_caps(MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS_MULTI_BLOCK)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__mlx5dv_crypto_engines_caps(MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS_MULTI_BLOCK); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_18 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CRYPTO_ENGINES_CAP_AES_XT_3, __pyx_t_4); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_crypto_engines_caps = __Pyx_FlagBase('mlx5dv_crypto_engines_caps', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS', __Pyx_PyLong_From_enum__mlx5dv_crypto_engines_caps(MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS)),
 *     ('MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS_SINGLE_BLOCK', __Pyx_PyLong_From_enum__mlx5dv_crypto_engines_caps(MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS_SINGLE_BLOCK)),
*/
  __pyx_t_4 = __Pyx_PyList_Pack(3, __pyx_t_16, __pyx_t_23, __pyx_t_18); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "EnumType":28
 *     ('MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS_MULTI_BLOCK', __Pyx_PyLong_From_enum__mlx5dv_crypto_engines_caps(MLX5DV_CRYPTO_ENGINES_CAP_AES_XTS_MULTI_BLOCK)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_18 = __Pyx_Globals(); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __pyx_t_18 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_engines_caps, __pyx_t_4};
    __pyx_t_23 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_18, __pyx_t_23, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_22 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_17, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_23);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_engines_caps, __pyx_t_22) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":35
 * 
 * 
 *     mlx5dv_crypto_engines_caps._member_names_ = list(mlx5dv_crypto_engines_caps.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_engines_caps); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __pyx_t_22 = __Pyx_PySequence_ListKeepNew(__pyx_t_17); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_engines_caps); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_22) < (0)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

    /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_crypto_key_purpose = __Pyx_FlagBase('mlx5dv_crypto_key_purpose', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CRYPTO_KEY_PURPOSE_AES_XTS', __Pyx_PyLong_From_enum__mlx5dv_crypto_key_purpose(MLX5DV_CRYPTO_KEY_PURPOSE_AES_XTS)),
 * 
*/
  __pyx_t_22 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_23 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_crypto_key_purpose = __Pyx_FlagBase('mlx5dv_crypto_key_purpose', [
 *     ('MLX5DV_CRYPTO_KEY_PURPOSE_AES_XTS', __Pyx_PyLong_From_enum__mlx5dv_crypto_key_purpose(MLX5DV_CRYPTO_KEY_PURPOSE_AES_XTS)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_18 = __Pyx_PyLong_From_enum__mlx5dv_crypto_key_purpose(MLX5DV_CRYPTO_KEY_PURPOSE_AES_XTS); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CRYPTO_KEY_PURPOSE_AES_XT, __pyx_t_18); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_crypto_key_purpose = __Pyx_FlagBase('mlx5dv_crypto_key_purpose', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CRYPTO_KEY_PURPOSE_AES_XTS', __Pyx_PyLong_From_enum__mlx5dv_crypto_key_purpose(MLX5DV_CRYPTO_KEY_PURPOSE_AES_XTS)),
 * 
*/
  __pyx_t_18 = __Pyx_PyList_Pack(1, __pyx_t_4); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":26
 *     ('MLX5DV_CRYPTO_KEY_PURPOSE_AES_XTS', __Pyx_PyLong_From_enum__mlx5dv_crypto_key_purpose(MLX5DV_CRYPTO_KEY_PURPOSE_AES_XTS)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_4 = __Pyx_Globals(); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_22, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_key_purpose, __pyx_t_18};
    __pyx_t_2 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_4, __pyx_t_2, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_17 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_23, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_key_purpose, __pyx_t_17) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":33
 * 
 * 
 *     mlx5dv_crypto_key_purpose._member_names_ = list(mlx5dv_crypto_key_purpose.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_key_purpose); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __pyx_t_17 = __Pyx_PySequence_ListKeepNew(__pyx_t_23); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_key_purpose); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_17) < (0)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

    /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_crypto_key_size = __Pyx_FlagBase('mlx5dv_crypto_key_size', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CRYPTO_KEY_SIZE_128', __Pyx_PyLong_From_enum__mlx5dv_crypto_key_size(MLX5DV_CRYPTO_KEY_SIZE_128)),
 *     ('MLX5DV_CRYPTO_KEY_SIZE_256', __Pyx_PyLong_From_enum__mlx5dv_crypto_key_size(MLX5DV_CRYPTO_KEY_SIZE_256)),
*/
  __pyx_t_17 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_2 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_crypto_key_size = __Pyx_FlagBase('mlx5dv_crypto_key_size', [
 *     ('MLX5DV_CRYPTO_KEY_SIZE_128', __Pyx_PyLong_From_enum__mlx5dv_crypto_key_size(MLX5DV_CRYPTO_KEY_SIZE_128)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CRYPTO_KEY_SIZE_256', __Pyx_PyLong_From_enum__mlx5dv_crypto_key_size(MLX5DV_CRYPTO_KEY_SIZE_256)),
 * 
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__mlx5dv_crypto_key_size(MLX5DV_CRYPTO_KEY_SIZE_128); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_18 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CRYPTO_KEY_SIZE_128, __pyx_t_4); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":25
 * mlx5dv_crypto_key_size = __Pyx_FlagBase('mlx5dv_crypto_key_size', [
 *     ('MLX5DV_CRYPTO_KEY_SIZE_128', __Pyx_PyLong_From_enum__mlx5dv_crypto_key_size(MLX5DV_CRYPTO_KEY_SIZE_128)),
 *     ('MLX5DV_CRYPTO_KEY_SIZE_256', __Pyx_PyLong_From_enum__mlx5dv_crypto_key_size(MLX5DV_CRYPTO_KEY_SIZE_256)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__mlx5dv_crypto_key_size(MLX5DV_CRYPTO_KEY_SIZE_256); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CRYPTO_KEY_SIZE_256, __pyx_t_4); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_crypto_key_size = __Pyx_FlagBase('mlx5dv_crypto_key_size', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CRYPTO_KEY_SIZE_128', __Pyx_PyLong_From_enum__mlx5dv_crypto_key_size(MLX5DV_CRYPTO_KEY_SIZE_128)),
 *     ('MLX5DV_CRYPTO_KEY_SIZE_256', __Pyx_PyLong_From_enum__mlx5dv_crypto_key_size(MLX5DV_CRYPTO_KEY_SIZE_256)),
*/
  __pyx_t_4 = __Pyx_PyList_Pack(2, __pyx_t_18, __pyx_t_22); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":27
 *     ('MLX5DV_CRYPTO_KEY_SIZE_256', __Pyx_PyLong_From_enum__mlx5dv_crypto_key_size(MLX5DV_CRYPTO_KEY_SIZE_256)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_22 = __Pyx_Globals(); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_18, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_key_size, __pyx_t_4};
    __pyx_t_18 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_22, __pyx_t_18, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_23 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_18);
    __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_key_size, __pyx_t_23) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":34
 * 
 * 
 *     mlx5dv_crypto_key_size._member_names_ = list(mlx5dv_crypto_key_size.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_key_size); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __pyx_t_23 = __Pyx_PySequence_ListKeepNew(__pyx_t_2); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_key_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_23) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_crypto_login_state = __Pyx_FlagBase('mlx5dv_crypto_login_state', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CRYPTO_LOGIN_STATE_VALID', __Pyx_PyLong_From_enum__mlx5dv_crypto_login_state(MLX5DV_CRYPTO_LOGIN_STATE_VALID)),
 *     ('MLX5DV_CRYPTO_LOGIN_STATE_NO_LOGIN', __Pyx_PyLong_From_enum__mlx5dv_crypto_login_state(MLX5DV_CRYPTO_LOGIN_STATE_NO_LOGIN)),
*/
  __pyx_t_23 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_18 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_crypto_login_state = __Pyx_FlagBase('mlx5dv_crypto_login_state', [
 *     ('MLX5DV_CRYPTO_LOGIN_STATE_VALID', __Pyx_PyLong_From_enum__mlx5dv_crypto_login_state(MLX5DV_CRYPTO_LOGIN_STATE_VALID)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CRYPTO_LOGIN_STATE_NO_LOGIN', __Pyx_PyLong_From_enum__mlx5dv_crypto_login_state(MLX5DV_CRYPTO_LOGIN_STATE_NO_LOGIN)),
 *     ('MLX5DV_CRYPTO_LOGIN_STATE_INVALID', __Pyx_PyLong_From_enum__mlx5dv_crypto_login_state(MLX5DV_CRYPTO_LOGIN_STATE_INVALID)),
*/
  __pyx_t_22 = __Pyx_PyLong_From_enum__mlx5dv_crypto_login_state(MLX5DV_CRYPTO_LOGIN_STATE_VALID); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CRYPTO_LOGIN_STATE_VALID, __pyx_t_22); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":25
 * mlx5dv_crypto_login_state = __Pyx_FlagBase('mlx5dv_crypto_login_state', [
 *     ('MLX5DV_CRYPTO_LOGIN_STATE_VALID', __Pyx_PyLong_From_enum__mlx5dv_crypto_login_state(MLX5DV_CRYPTO_LOGIN_STATE_VALID)),
 *     ('MLX5DV_CRYPTO_LOGIN_STATE_NO_LOGIN', __Pyx_PyLong_From_enum__mlx5dv_crypto_login_state(MLX5DV_CRYPTO_LOGIN_STATE_NO_LOGIN)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CRYPTO_LOGIN_STATE_INVALID', __Pyx_PyLong_From_enum__mlx5dv_crypto_login_state(MLX5DV_CRYPTO_LOGIN_STATE_INVALID)),
 * 
*/
  __pyx_t_22 = __Pyx_PyLong_From_enum__mlx5dv_crypto_login_state(MLX5DV_CRYPTO_LOGIN_STATE_NO_LOGIN); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_17 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CRYPTO_LOGIN_STATE_NO_LOG, __pyx_t_22); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":26
 *     ('MLX5DV_CRYPTO_LOGIN_STATE_VALID', __Pyx_PyLong_From_enum__mlx5dv_crypto_login_state(MLX5DV_CRYPTO_LOGIN_STATE_VALID)),
 *     ('MLX5DV_CRYPTO_LOGIN_STATE_NO_LOGIN', __Pyx_PyLong_From_enum__mlx5dv_crypto_login_state(MLX5DV_CRYPTO_LOGIN_STATE_NO_LOGIN)),
 *     ('MLX5DV_CRYPTO_LOGIN_STATE_INVALID', __Pyx_PyLong_From_enum__mlx5dv_crypto_login_state(MLX5DV_CRYPTO_LOGIN_STATE_INVALID)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_22 = __Pyx_PyLong_From_enum__mlx5dv_crypto_login_state(MLX5DV_CRYPTO_LOGIN_STATE_INVALID); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CRYPTO_LOGIN_STATE_INVALI, __pyx_t_22); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_crypto_login_state = __Pyx_FlagBase('mlx5dv_crypto_login_state', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CRYPTO_LOGIN_STATE_VALID', __Pyx_PyLong_From_enum__mlx5dv_crypto_login_state(MLX5DV_CRYPTO_LOGIN_STATE_VALID)),
 *     ('MLX5DV_CRYPTO_LOGIN_STATE_NO_LOGIN', __Pyx_PyLong_From_enum__mlx5dv_crypto_login_state(MLX5DV_CRYPTO_LOGIN_STATE_NO_LOGIN)),
*/
  __pyx_t_22 = __Pyx_PyList_Pack(3, __pyx_t_4, __pyx_t_17, __pyx_t_16); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":28
 *     ('MLX5DV_CRYPTO_LOGIN_STATE_INVALID', __Pyx_PyLong_From_enum__mlx5dv_crypto_login_state(MLX5DV_CRYPTO_LOGIN_STATE_INVALID)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_16 = __Pyx_Globals(); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_23, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_login_state, __pyx_t_22};
    __pyx_t_17 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_16, __pyx_t_17, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_18, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_17);
    __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_login_state, __pyx_t_2) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":35
 * 
 * 
 *     mlx5dv_crypto_login_state._member_names_ = list(mlx5dv_crypto_login_state.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_login_state); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PySequence_ListKeepNew(__pyx_t_18); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_login_state); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_2) < (0)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

    /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_crypto_standard = __Pyx_FlagBase('mlx5dv_crypto_standard', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CRYPTO_STANDARD_AES_XTS', __Pyx_PyLong_From_enum__mlx5dv_crypto_standard(MLX5DV_CRYPTO_STANDARD_AES_XTS)),
 * 
*/
  __pyx_t_2 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_17 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_crypto_standard = __Pyx_FlagBase('mlx5dv_crypto_standard', [
 *     ('MLX5DV_CRYPTO_STANDARD_AES_XTS', __Pyx_PyLong_From_enum__mlx5dv_crypto_standard(MLX5DV_CRYPTO_STANDARD_AES_XTS)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__mlx5dv_crypto_standard(MLX5DV_CRYPTO_STANDARD_AES_XTS); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CRYPTO_STANDARD_AES_XTS, __pyx_t_16); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_crypto_standard = __Pyx_FlagBase('mlx5dv_crypto_standard', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CRYPTO_STANDARD_AES_XTS', __Pyx_PyLong_From_enum__mlx5dv_crypto_standard(MLX5DV_CRYPTO_STANDARD_AES_XTS)),
 * 
*/
  __pyx_t_16 = __Pyx_PyList_Pack(1, __pyx_t_22); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":26
 *     ('MLX5DV_CRYPTO_STANDARD_AES_XTS', __Pyx_PyLong_From_enum__mlx5dv_crypto_standard(MLX5DV_CRYPTO_STANDARD_AES_XTS)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_22 = __Pyx_Globals(); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_standard, __pyx_t_16};
    __pyx_t_23 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_22, __pyx_t_23, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_18 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_17, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_23);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_standard, __pyx_t_18) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":33
 * 
 * 
 *     mlx5dv_crypto_standard._member_names_ = list(mlx5dv_crypto_standard.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_standard); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __pyx_t_18 = __Pyx_PySequence_ListKeepNew(__pyx_t_17); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_standard); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_18) < (0)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

    /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_crypto_wrapped_import_method_caps = __Pyx_FlagBase('mlx5dv_crypto_wrapped_import_method_caps', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CRYPTO_WRAPPED_IMPORT_METHOD_CAP_AES_XTS', __Pyx_PyLong_From_enum__mlx5dv_crypto_wrapped_import_method_caps(MLX5DV_CRYPTO_WRAPPED_IMPORT_METHOD_CAP_AES_XTS)),
 * 
*/
  __pyx_t_18 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_23 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_crypto_wrapped_import_method_caps = __Pyx_FlagBase('mlx5dv_crypto_wrapped_import_method_caps', [
 *     ('MLX5DV_CRYPTO_WRAPPED_IMPORT_METHOD_CAP_AES_XTS', __Pyx_PyLong_From_enum__mlx5dv_crypto_wrapped_import_method_caps(MLX5DV_CRYPTO_WRAPPED_IMPORT_METHOD_CAP_AES_XTS)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_22 = __Pyx_PyLong_From_enum__mlx5dv_crypto_wrapped_import_method_caps(MLX5DV_CRYPTO_WRAPPED_IMPORT_METHOD_CAP_AES_XTS); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_CRYPTO_WRAPPED_IMPORT_MET, __pyx_t_22); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_crypto_wrapped_import_method_caps = __Pyx_FlagBase('mlx5dv_crypto_wrapped_import_method_caps', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_CRYPTO_WRAPPED_IMPORT_METHOD_CAP_AES_XTS', __Pyx_PyLong_From_enum__mlx5dv_crypto_wrapped_import_method_caps(MLX5DV_CRYPTO_WRAPPED_IMPORT_METHOD_CAP_AES_XTS)),
 * 
*/
  __pyx_t_22 = __Pyx_PyList_Pack(1, __pyx_t_16); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":26
 *     ('MLX5DV_CRYPTO_WRAPPED_IMPORT_METHOD_CAP_AES_XTS', __Pyx_PyLong_From_enum__mlx5dv_crypto_wrapped_import_method_caps(MLX5DV_CRYPTO_WRAPPED_IMPORT_METHOD_CAP_AES_XTS)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_16 = __Pyx_Globals(); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_18, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_wrapped_import_met, __pyx_t_22};
    __pyx_t_2 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_16, __pyx_t_2, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_17 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_23, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_wrapped_import_met, __pyx_t_17) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":33
 * 
 * 
 *     mlx5dv_crypto_wrapped_import_method_caps._member_names_ = list(mlx5dv_crypto_wrapped_import_method_caps.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_wrapped_import_met); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __pyx_t_17 = __Pyx_PySequence_ListKeepNew(__pyx_t_23); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_mlx5dv_crypto_wrapped_import_met); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_17) < (0)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

    /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_dc_type = __Pyx_FlagBase('mlx5dv_dc_type', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DCTYPE_DCT', __Pyx_PyLong_From_enum__mlx5dv_dc_type(MLX5DV_DCTYPE_DCT)),
 *     ('MLX5DV_DCTYPE_DCI', __Pyx_PyLong_From_enum__mlx5dv_dc_type(MLX5DV_DCTYPE_DCI)),
*/
  __pyx_t_17 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_2 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_dc_type = __Pyx_FlagBase('mlx5dv_dc_type', [
 *     ('MLX5DV_DCTYPE_DCT', __Pyx_PyLong_From_enum__mlx5dv_dc_type(MLX5DV_DCTYPE_DCT)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DCTYPE_DCI', __Pyx_PyLong_From_enum__mlx5dv_dc_type(MLX5DV_DCTYPE_DCI)),
 * 
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__mlx5dv_dc_type(MLX5DV_DCTYPE_DCT); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_DCTYPE_DCT, __pyx_t_16); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":25
 * mlx5dv_dc_type = __Pyx_FlagBase('mlx5dv_dc_type', [
 *     ('MLX5DV_DCTYPE_DCT', __Pyx_PyLong_From_enum__mlx5dv_dc_type(MLX5DV_DCTYPE_DCT)),
 *     ('MLX5DV_DCTYPE_DCI', __Pyx_PyLong_From_enum__mlx5dv_dc_type(MLX5DV_DCTYPE_DCI)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__mlx5dv_dc_type(MLX5DV_DCTYPE_DCI); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_18 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_DCTYPE_DCI, __pyx_t_16); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_dc_type = __Pyx_FlagBase('mlx5dv_dc_type', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DCTYPE_DCT', __Pyx_PyLong_From_enum__mlx5dv_dc_type(MLX5DV_DCTYPE_DCT)),
 *     ('MLX5DV_DCTYPE_DCI', __Pyx_PyLong_From_enum__mlx5dv_dc_type(MLX5DV_DCTYPE_DCI)),
*/
  __pyx_t_16 = __Pyx_PyList_Pack(2, __pyx_t_22, __pyx_t_18); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "EnumType":27
 *     ('MLX5DV_DCTYPE_DCI', __Pyx_PyLong_From_enum__mlx5dv_dc_type(MLX5DV_DCTYPE_DCI)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_18 = __Pyx_Globals(); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_22 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __pyx_t_18 = __Pyx_PyObject_Call(__pyx_t_22, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_dc_type, __pyx_t_16};
    __pyx_t_22 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_18, __pyx_t_22, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_23 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_22);
    __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_dc_type, __pyx_t_23) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":34
 * 
 * 
 *     mlx5dv_dc_type._member_names_ = list(mlx5dv_dc_type.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_mlx5dv_dc_type); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __pyx_t_23 = __Pyx_PySequence_ListKeepNew(__pyx_t_2); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_dc_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_23) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_dek_init_attr_mask = __Pyx_FlagBase('mlx5dv_dek_init_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DEK_INIT_ATTR_CRYPTO_LOGIN', __Pyx_PyLong_From_enum__mlx5dv_dek_init_attr_mask(MLX5DV_DEK_INIT_ATTR_CRYPTO_LOGIN)),
 * 
*/
  __pyx_t_23 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_22 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_dek_init_attr_mask = __Pyx_FlagBase('mlx5dv_dek_init_attr_mask', [
 *     ('MLX5DV_DEK_INIT_ATTR_CRYPTO_LOGIN', __Pyx_PyLong_From_enum__mlx5dv_dek_init_attr_mask(MLX5DV_DEK_INIT_ATTR_CRYPTO_LOGIN)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_18 = __Pyx_PyLong_From_enum__mlx5dv_dek_init_attr_mask(MLX5DV_DEK_INIT_ATTR_CRYPTO_LOGIN); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_DEK_INIT_ATTR_CRYPTO_LOGI, __pyx_t_18); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_dek_init_attr_mask = __Pyx_FlagBase('mlx5dv_dek_init_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DEK_INIT_ATTR_CRYPTO_LOGIN', __Pyx_PyLong_From_enum__mlx5dv_dek_init_attr_mask(MLX5DV_DEK_INIT_ATTR_CRYPTO_LOGIN)),
 * 
*/
  __pyx_t_18 = __Pyx_PyList_Pack(1, __pyx_t_16); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":26
 *     ('MLX5DV_DEK_INIT_ATTR_CRYPTO_LOGIN', __Pyx_PyLong_From_enum__mlx5dv_dek_init_attr_mask(MLX5DV_DEK_INIT_ATTR_CRYPTO_LOGIN)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_16 = __Pyx_Globals(); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_23, __pyx_mstate_global->__pyx_n_u_mlx5dv_dek_init_attr_mask, __pyx_t_18};
    __pyx_t_17 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_16, __pyx_t_17, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_22, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_17);
    __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_dek_init_attr_mask, __pyx_t_2) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":33
 * 
 * 
 *     mlx5dv_dek_init_attr_mask._member_names_ = list(mlx5dv_dek_init_attr_mask.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_dek_init_attr_mask); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_22 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PySequence_ListKeepNew(__pyx_t_22); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_mlx5dv_dek_init_attr_mask); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_2) < (0)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

    /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_dek_state = __Pyx_FlagBase('mlx5dv_dek_state', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DEK_STATE_READY', __Pyx_PyLong_From_enum__mlx5dv_dek_state(MLX5DV_DEK_STATE_READY)),
 *     ('MLX5DV_DEK_STATE_ERROR', __Pyx_PyLong_From_enum__mlx5dv_dek_state(MLX5DV_DEK_STATE_ERROR)),
*/
  __pyx_t_2 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_17 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_dek_state = __Pyx_FlagBase('mlx5dv_dek_state', [
 *     ('MLX5DV_DEK_STATE_READY', __Pyx_PyLong_From_enum__mlx5dv_dek_state(MLX5DV_DEK_STATE_READY)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DEK_STATE_ERROR', __Pyx_PyLong_From_enum__mlx5dv_dek_state(MLX5DV_DEK_STATE_ERROR)),
 * 
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__mlx5dv_dek_state(MLX5DV_DEK_STATE_READY); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_18 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_DEK_STATE_READY, __pyx_t_16); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":25
 * mlx5dv_dek_state = __Pyx_FlagBase('mlx5dv_dek_state', [
 *     ('MLX5DV_DEK_STATE_READY', __Pyx_PyLong_From_enum__mlx5dv_dek_state(MLX5DV_DEK_STATE_READY)),
 *     ('MLX5DV_DEK_STATE_ERROR', __Pyx_PyLong_From_enum__mlx5dv_dek_state(MLX5DV_DEK_STATE_ERROR)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__mlx5dv_dek_state(MLX5DV_DEK_STATE_ERROR); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_23 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_DEK_STATE_ERROR, __pyx_t_16); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_dek_state = __Pyx_FlagBase('mlx5dv_dek_state', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DEK_STATE_READY', __Pyx_PyLong_From_enum__mlx5dv_dek_state(MLX5DV_DEK_STATE_READY)),
 *     ('MLX5DV_DEK_STATE_ERROR', __Pyx_PyLong_From_enum__mlx5dv_dek_state(MLX5DV_DEK_STATE_ERROR)),
*/
  __pyx_t_16 = __Pyx_PyList_Pack(2, __pyx_t_18, __pyx_t_23); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":27
 *     ('MLX5DV_DEK_STATE_ERROR', __Pyx_PyLong_From_enum__mlx5dv_dek_state(MLX5DV_DEK_STATE_ERROR)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_23 = __Pyx_Globals(); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_18, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_dek_state, __pyx_t_16};
    __pyx_t_18 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_23, __pyx_t_18, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_22 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_17, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_18);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_dek_state, __pyx_t_22) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":34
 * 
 * 
 *     mlx5dv_dek_state._member_names_ = list(mlx5dv_dek_state.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_mlx5dv_dek_state); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __pyx_t_22 = __Pyx_PySequence_ListKeepNew(__pyx_t_17); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_dek_state); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_22) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_devx_create_event_channel_flags = __Pyx_FlagBase('mlx5dv_devx_create_event_channel_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DEVX_CREATE_EVENT_CHANNEL_FLAGS_OMIT_EV_DATA', __Pyx_PyLong_From_enum__mlx5dv_devx_create_event_channel_flags(MLX5DV_DEVX_CREATE_EVENT_CHANNEL_FLAGS_OMIT_EV_DATA)),
 * 
*/
  __pyx_t_22 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_18 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_devx_create_event_channel_flags = __Pyx_FlagBase('mlx5dv_devx_create_event_channel_flags', [
 *     ('MLX5DV_DEVX_CREATE_EVENT_CHANNEL_FLAGS_OMIT_EV_DATA', __Pyx_PyLong_From_enum__mlx5dv_devx_create_event_channel_flags(MLX5DV_DEVX_CREATE_EVENT_CHANNEL_FLAGS_OMIT_EV_DATA)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_23 = __Pyx_PyLong_From_enum__mlx5dv_devx_create_event_channel_flags(MLX5DV_DEVX_CREATE_EVENT_CHANNEL_FLAGS_OMIT_EV_DATA); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_DEVX_CREATE_EVENT_CHANNEL, __pyx_t_23); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_devx_create_event_channel_flags = __Pyx_FlagBase('mlx5dv_devx_create_event_channel_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DEVX_CREATE_EVENT_CHANNEL_FLAGS_OMIT_EV_DATA', __Pyx_PyLong_From_enum__mlx5dv_devx_create_event_channel_flags(MLX5DV_DEVX_CREATE_EVENT_CHANNEL_FLAGS_OMIT_EV_DATA)),
 * 
*/
  __pyx_t_23 = __Pyx_PyList_Pack(1, __pyx_t_16); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":26
 *     ('MLX5DV_DEVX_CREATE_EVENT_CHANNEL_FLAGS_OMIT_EV_DATA', __Pyx_PyLong_From_enum__mlx5dv_devx_create_event_channel_flags(MLX5DV_DEVX_CREATE_EVENT_CHANNEL_FLAGS_OMIT_EV_DATA)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_16 = __Pyx_Globals(); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_22, __pyx_mstate_global->__pyx_n_u_mlx5dv_devx_create_event_channel, __pyx_t_23};
    __pyx_t_2 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_16, __pyx_t_2, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_17 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_18, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_devx_create_event_channel, __pyx_t_17) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":33
 * 
 * 
 *     mlx5dv_devx_create_event_channel_flags._member_names_ = list(mlx5dv_devx_create_event_channel_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_devx_create_event_channel); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __pyx_t_17 = __Pyx_PySequence_ListKeepNew(__pyx_t_18); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_mlx5dv_devx_create_event_channel); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_17) < (0)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

    /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_dr_action_dest_type = __Pyx_FlagBase('mlx5dv_dr_action_dest_type', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DR_ACTION_DEST', __Pyx_PyLong_From_enum__mlx5dv_dr_action_dest_type(MLX5DV_DR_ACTION_DEST)),
 *     ('MLX5DV_DR_ACTION_DEST_REFORMAT', __Pyx_PyLong_From_enum__mlx5dv_dr_action_dest_type(MLX5DV_DR_ACTION_DEST_REFORMAT)),
*/
  __pyx_t_17 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_2 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_dr_action_dest_type = __Pyx_FlagBase('mlx5dv_dr_action_dest_type', [
 *     ('MLX5DV_DR_ACTION_DEST', __Pyx_PyLong_From_enum__mlx5dv_dr_action_dest_type(MLX5DV_DR_ACTION_DEST)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DR_ACTION_DEST_REFORMAT', __Pyx_PyLong_From_enum__mlx5dv_dr_action_dest_type(MLX5DV_DR_ACTION_DEST_REFORMAT)),
 * 
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__mlx5dv_dr_action_dest_type(MLX5DV_DR_ACTION_DEST); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_23 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_DR_ACTION_DEST, __pyx_t_16); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":25
 * mlx5dv_dr_action_dest_type = __Pyx_FlagBase('mlx5dv_dr_action_dest_type', [
 *     ('MLX5DV_DR_ACTION_DEST', __Pyx_PyLong_From_enum__mlx5dv_dr_action_dest_type(MLX5DV_DR_ACTION_DEST)),
 *     ('MLX5DV_DR_ACTION_DEST_REFORMAT', __Pyx_PyLong_From_enum__mlx5dv_dr_action_dest_type(MLX5DV_DR_ACTION_DEST_REFORMAT)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__mlx5dv_dr_action_dest_type(MLX5DV_DR_ACTION_DEST_REFORMAT); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_DR_ACTION_DEST_REFORMAT, __pyx_t_16); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_dr_action_dest_type = __Pyx_FlagBase('mlx5dv_dr_action_dest_type', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DR_ACTION_DEST', __Pyx_PyLong_From_enum__mlx5dv_dr_action_dest_type(MLX5DV_DR_ACTION_DEST)),
 *     ('MLX5DV_DR_ACTION_DEST_REFORMAT', __Pyx_PyLong_From_enum__mlx5dv_dr_action_dest_type(MLX5DV_DR_ACTION_DEST_REFORMAT)),
*/
  __pyx_t_16 = __Pyx_PyList_Pack(2, __pyx_t_23, __pyx_t_22); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":27
 *     ('MLX5DV_DR_ACTION_DEST_REFORMAT', __Pyx_PyLong_From_enum__mlx5dv_dr_action_dest_type(MLX5DV_DR_ACTION_DEST_REFORMAT)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_22 = __Pyx_Globals(); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_23, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_dr_action_dest_type, __pyx_t_16};
    __pyx_t_23 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_22, __pyx_t_23, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_18 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_23);
    __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_dr_action_dest_type, __pyx_t_18) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":34
 * 
 * 
 *     mlx5dv_dr_action_dest_type._member_names_ = list(mlx5dv_dr_action_dest_type.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_mlx5dv_dr_action_dest_type); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __pyx_t_18 = __Pyx_PySequence_ListKeepNew(__pyx_t_2); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_dr_action_dest_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_18) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_dr_action_flags = __Pyx_FlagBase('mlx5dv_dr_action_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DR_ACTION_FLAGS_ROOT_LEVEL', __Pyx_PyLong_From_enum__mlx5dv_dr_action_flags(MLX5DV_DR_ACTION_FLAGS_ROOT_LEVEL)),
 * 
*/
  __pyx_t_18 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_23 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_dr_action_flags = __Pyx_FlagBase('mlx5dv_dr_action_flags', [
 *     ('MLX5DV_DR_ACTION_FLAGS_ROOT_LEVEL', __Pyx_PyLong_From_enum__mlx5dv_dr_action_flags(MLX5DV_DR_ACTION_FLAGS_ROOT_LEVEL)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_22 = __Pyx_PyLong_From_enum__mlx5dv_dr_action_flags(MLX5DV_DR_ACTION_FLAGS_ROOT_LEVEL); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_DR_ACTION_FLAGS_ROOT_LEVE, __pyx_t_22); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_dr_action_flags = __Pyx_FlagBase('mlx5dv_dr_action_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DR_ACTION_FLAGS_ROOT_LEVEL', __Pyx_PyLong_From_enum__mlx5dv_dr_action_flags(MLX5DV_DR_ACTION_FLAGS_ROOT_LEVEL)),
 * 
*/
  __pyx_t_22 = __Pyx_PyList_Pack(1, __pyx_t_16); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":26
 *     ('MLX5DV_DR_ACTION_FLAGS_ROOT_LEVEL', __Pyx_PyLong_From_enum__mlx5dv_dr_action_flags(MLX5DV_DR_ACTION_FLAGS_ROOT_LEVEL)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_16 = __Pyx_Globals(); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_18, __pyx_mstate_global->__pyx_n_u_mlx5dv_dr_action_flags, __pyx_t_22};
    __pyx_t_17 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_16, __pyx_t_17, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_23, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_17);
    __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_dr_action_flags, __pyx_t_2) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":33
 * 
 * 
 *     mlx5dv_dr_action_flags._member_names_ = list(mlx5dv_dr_action_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_dr_action_flags); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PySequence_ListKeepNew(__pyx_t_23); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_mlx5dv_dr_action_flags); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_2) < (0)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

    /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_dr_domain_sync_flags = __Pyx_FlagBase('mlx5dv_dr_domain_sync_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DR_DOMAIN_SYNC_FLAGS_SW', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_sync_flags(MLX5DV_DR_DOMAIN_SYNC_FLAGS_SW)),
 *     ('MLX5DV_DR_DOMAIN_SYNC_FLAGS_HW', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_sync_flags(MLX5DV_DR_DOMAIN_SYNC_FLAGS_HW)),
*/
  __pyx_t_2 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_17 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_dr_domain_sync_flags = __Pyx_FlagBase('mlx5dv_dr_domain_sync_flags', [
 *     ('MLX5DV_DR_DOMAIN_SYNC_FLAGS_SW', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_sync_flags(MLX5DV_DR_DOMAIN_SYNC_FLAGS_SW)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DR_DOMAIN_SYNC_FLAGS_HW', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_sync_flags(MLX5DV_DR_DOMAIN_SYNC_FLAGS_HW)),
 *     ('MLX5DV_DR_DOMAIN_SYNC_FLAGS_MEM', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_sync_flags(MLX5DV_DR_DOMAIN_SYNC_FLAGS_MEM)),
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__mlx5dv_dr_domain_sync_flags(MLX5DV_DR_DOMAIN_SYNC_FLAGS_SW); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_DR_DOMAIN_SYNC_FLAGS_SW, __pyx_t_16); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":25
 * mlx5dv_dr_domain_sync_flags = __Pyx_FlagBase('mlx5dv_dr_domain_sync_flags', [
 *     ('MLX5DV_DR_DOMAIN_SYNC_FLAGS_SW', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_sync_flags(MLX5DV_DR_DOMAIN_SYNC_FLAGS_SW)),
 *     ('MLX5DV_DR_DOMAIN_SYNC_FLAGS_HW', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_sync_flags(MLX5DV_DR_DOMAIN_SYNC_FLAGS_HW)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DR_DOMAIN_SYNC_FLAGS_MEM', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_sync_flags(MLX5DV_DR_DOMAIN_SYNC_FLAGS_MEM)),
 * 
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__mlx5dv_dr_domain_sync_flags(MLX5DV_DR_DOMAIN_SYNC_FLAGS_HW); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_18 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_DR_DOMAIN_SYNC_FLAGS_HW, __pyx_t_16); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":26
 *     ('MLX5DV_DR_DOMAIN_SYNC_FLAGS_SW', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_sync_flags(MLX5DV_DR_DOMAIN_SYNC_FLAGS_SW)),
 *     ('MLX5DV_DR_DOMAIN_SYNC_FLAGS_HW', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_sync_flags(MLX5DV_DR_DOMAIN_SYNC_FLAGS_HW)),
 *     ('MLX5DV_DR_DOMAIN_SYNC_FLAGS_MEM', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_sync_flags(MLX5DV_DR_DOMAIN_SYNC_FLAGS_MEM)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__mlx5dv_dr_domain_sync_flags(MLX5DV_DR_DOMAIN_SYNC_FLAGS_MEM); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_DR_DOMAIN_SYNC_FLAGS_MEM, __pyx_t_16); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_dr_domain_sync_flags = __Pyx_FlagBase('mlx5dv_dr_domain_sync_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DR_DOMAIN_SYNC_FLAGS_SW', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_sync_flags(MLX5DV_DR_DOMAIN_SYNC_FLAGS_SW)),
 *     ('MLX5DV_DR_DOMAIN_SYNC_FLAGS_HW', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_sync_flags(MLX5DV_DR_DOMAIN_SYNC_FLAGS_HW)),
*/
  __pyx_t_16 = __Pyx_PyList_Pack(3, __pyx_t_22, __pyx_t_18, __pyx_t_4); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":28
 *     ('MLX5DV_DR_DOMAIN_SYNC_FLAGS_MEM', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_sync_flags(MLX5DV_DR_DOMAIN_SYNC_FLAGS_MEM)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_4 = __Pyx_Globals(); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_18, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_dr_domain_sync_flags, __pyx_t_16};
    __pyx_t_18 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_4, __pyx_t_18, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_23 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_17, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_18);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_dr_domain_sync_flags, __pyx_t_23) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":35
 * 
 * 
 *     mlx5dv_dr_domain_sync_flags._member_names_ = list(mlx5dv_dr_domain_sync_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_mlx5dv_dr_domain_sync_flags); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __pyx_t_23 = __Pyx_PySequence_ListKeepNew(__pyx_t_17); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_dr_domain_sync_flags); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_23) < (0)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

    /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_dr_domain_type = __Pyx_FlagBase('mlx5dv_dr_domain_type', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DR_DOMAIN_TYPE_NIC_RX', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_type(MLX5DV_DR_DOMAIN_TYPE_NIC_RX)),
 *     ('MLX5DV_DR_DOMAIN_TYPE_NIC_TX', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_type(MLX5DV_DR_DOMAIN_TYPE_NIC_TX)),
*/
  __pyx_t_23 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_18 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_dr_domain_type = __Pyx_FlagBase('mlx5dv_dr_domain_type', [
 *     ('MLX5DV_DR_DOMAIN_TYPE_NIC_RX', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_type(MLX5DV_DR_DOMAIN_TYPE_NIC_RX)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DR_DOMAIN_TYPE_NIC_TX', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_type(MLX5DV_DR_DOMAIN_TYPE_NIC_TX)),
 *     ('MLX5DV_DR_DOMAIN_TYPE_FDB', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_type(MLX5DV_DR_DOMAIN_TYPE_FDB)),
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__mlx5dv_dr_domain_type(MLX5DV_DR_DOMAIN_TYPE_NIC_RX); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_DR_DOMAIN_TYPE_NIC_RX, __pyx_t_4); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":25
 * mlx5dv_dr_domain_type = __Pyx_FlagBase('mlx5dv_dr_domain_type', [
 *     ('MLX5DV_DR_DOMAIN_TYPE_NIC_RX', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_type(MLX5DV_DR_DOMAIN_TYPE_NIC_RX)),
 *     ('MLX5DV_DR_DOMAIN_TYPE_NIC_TX', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_type(MLX5DV_DR_DOMAIN_TYPE_NIC_TX)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DR_DOMAIN_TYPE_FDB', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_type(MLX5DV_DR_DOMAIN_TYPE_FDB)),
 * 
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__mlx5dv_dr_domain_type(MLX5DV_DR_DOMAIN_TYPE_NIC_TX); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_DR_DOMAIN_TYPE_NIC_TX, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":26
 *     ('MLX5DV_DR_DOMAIN_TYPE_NIC_RX', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_type(MLX5DV_DR_DOMAIN_TYPE_NIC_RX)),
 *     ('MLX5DV_DR_DOMAIN_TYPE_NIC_TX', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_type(MLX5DV_DR_DOMAIN_TYPE_NIC_TX)),
 *     ('MLX5DV_DR_DOMAIN_TYPE_FDB', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_type(MLX5DV_DR_DOMAIN_TYPE_FDB)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__mlx5dv_dr_domain_type(MLX5DV_DR_DOMAIN_TYPE_FDB); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_DR_DOMAIN_TYPE_FDB, __pyx_t_4); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_dr_domain_type = __Pyx_FlagBase('mlx5dv_dr_domain_type', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DR_DOMAIN_TYPE_NIC_RX', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_type(MLX5DV_DR_DOMAIN_TYPE_NIC_RX)),
 *     ('MLX5DV_DR_DOMAIN_TYPE_NIC_TX', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_type(MLX5DV_DR_DOMAIN_TYPE_NIC_TX)),
*/
  __pyx_t_4 = __Pyx_PyList_Pack(3, __pyx_t_16, __pyx_t_2, __pyx_t_22); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":28
 *     ('MLX5DV_DR_DOMAIN_TYPE_FDB', __Pyx_PyLong_From_enum__mlx5dv_dr_domain_type(MLX5DV_DR_DOMAIN_TYPE_FDB)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_22 = __Pyx_Globals(); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_23, __pyx_mstate_global->__pyx_n_u_mlx5dv_dr_domain_type, __pyx_t_4};
    __pyx_t_2 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_22, __pyx_t_2, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_17 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_18, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_dr_domain_type, __pyx_t_17) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":35
 * 
 * 
 *     mlx5dv_dr_domain_type._member_names_ = list(mlx5dv_dr_domain_type.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_dr_domain_type); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __pyx_t_17 = __Pyx_PySequence_ListKeepNew(__pyx_t_18); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_mlx5dv_dr_domain_type); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_17) < (0)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

    /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_dr_matcher_layout_flags = __Pyx_FlagBase('mlx5dv_dr_matcher_layout_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DR_MATCHER_LAYOUT_RESIZABLE', __Pyx_PyLong_From_enum__mlx5dv_dr_matcher_layout_flags(MLX5DV_DR_MATCHER_LAYOUT_RESIZABLE)),
 *     ('MLX5DV_DR_MATCHER_LAYOUT_NUM_RULE', __Pyx_PyLong_From_enum__mlx5dv_dr_matcher_layout_flags(MLX5DV_DR_MATCHER_LAYOUT_NUM_RULE)),
*/
  __pyx_t_17 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_2 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_dr_matcher_layout_flags = __Pyx_FlagBase('mlx5dv_dr_matcher_layout_flags', [
 *     ('MLX5DV_DR_MATCHER_LAYOUT_RESIZABLE', __Pyx_PyLong_From_enum__mlx5dv_dr_matcher_layout_flags(MLX5DV_DR_MATCHER_LAYOUT_RESIZABLE)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DR_MATCHER_LAYOUT_NUM_RULE', __Pyx_PyLong_From_enum__mlx5dv_dr_matcher_layout_flags(MLX5DV_DR_MATCHER_LAYOUT_NUM_RULE)),
 * 
*/
  __pyx_t_22 = __Pyx_PyLong_From_enum__mlx5dv_dr_matcher_layout_flags(MLX5DV_DR_MATCHER_LAYOUT_RESIZABLE); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_DR_MATCHER_LAYOUT_RESIZAB, __pyx_t_22); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":25
 * mlx5dv_dr_matcher_layout_flags = __Pyx_FlagBase('mlx5dv_dr_matcher_layout_flags', [
 *     ('MLX5DV_DR_MATCHER_LAYOUT_RESIZABLE', __Pyx_PyLong_From_enum__mlx5dv_dr_matcher_layout_flags(MLX5DV_DR_MATCHER_LAYOUT_RESIZABLE)),
 *     ('MLX5DV_DR_MATCHER_LAYOUT_NUM_RULE', __Pyx_PyLong_From_enum__mlx5dv_dr_matcher_layout_flags(MLX5DV_DR_MATCHER_LAYOUT_NUM_RULE)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_22 = __Pyx_PyLong_From_enum__mlx5dv_dr_matcher_layout_flags(MLX5DV_DR_MATCHER_LAYOUT_NUM_RULE); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_23 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_DR_MATCHER_LAYOUT_NUM_RUL, __pyx_t_22); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_dr_matcher_layout_flags = __Pyx_FlagBase('mlx5dv_dr_matcher_layout_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_DR_MATCHER_LAYOUT_RESIZABLE', __Pyx_PyLong_From_enum__mlx5dv_dr_matcher_layout_flags(MLX5DV_DR_MATCHER_LAYOUT_RESIZABLE)),
 *     ('MLX5DV_DR_MATCHER_LAYOUT_NUM_RULE', __Pyx_PyLong_From_enum__mlx5dv_dr_matcher_layout_flags(MLX5DV_DR_MATCHER_LAYOUT_NUM_RULE)),
*/
  __pyx_t_22 = __Pyx_PyList_Pack(2, __pyx_t_4, __pyx_t_23); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":27
 *     ('MLX5DV_DR_MATCHER_LAYOUT_NUM_RULE', __Pyx_PyLong_From_enum__mlx5dv_dr_matcher_layout_flags(MLX5DV_DR_MATCHER_LAYOUT_NUM_RULE)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_23 = __Pyx_Globals(); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_23, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_dr_matcher_layout_flags, __pyx_t_22};
    __pyx_t_4 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_23, __pyx_t_4, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_18 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_dr_matcher_layout_flags, __pyx_t_18) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":34
 * 
 * 
 *     mlx5dv_dr_matcher_layout_flags._member_names_ = list(mlx5dv_dr_matcher_layout_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_mlx5dv_dr_matcher_layout_flags); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __pyx_t_18 = __Pyx_PySequence_ListKeepNew(__pyx_t_2); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_dr_matcher_layout_flags); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_18) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_flow_action_cap_flags = __Pyx_FlagBase('mlx5dv_flow_action_cap_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM)),
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_REQ_METADATA', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_REQ_METADATA)),
*/
  __pyx_t_18 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_4 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_flow_action_cap_flags = __Pyx_FlagBase('mlx5dv_flow_action_cap_flags', [
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_REQ_METADATA', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_REQ_METADATA)),
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_SPI_STEERING', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_SPI_STEERING)),
*/
  __pyx_t_23 = __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_ACTION_FLAGS_ESP_AES, __pyx_t_23); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":25
 * mlx5dv_flow_action_cap_flags = __Pyx_FlagBase('mlx5dv_flow_action_cap_flags', [
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM)),
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_REQ_METADATA', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_REQ_METADATA)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_SPI_STEERING', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_SPI_STEERING)),
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_FULL_OFFLOAD', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_FULL_OFFLOAD)),
*/
  __pyx_t_23 = __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_REQ_METADATA); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_17 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_2, __pyx_t_23); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":26
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM)),
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_REQ_METADATA', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_REQ_METADATA)),
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_SPI_STEERING', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_SPI_STEERING)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_FULL_OFFLOAD', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_FULL_OFFLOAD)),
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_TX_IV_IS_ESN', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_TX_IV_IS_ESN)),
*/
  __pyx_t_23 = __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_SPI_STEERING); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_3, __pyx_t_23); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":27
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_REQ_METADATA', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_REQ_METADATA)),
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_SPI_STEERING', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_SPI_STEERING)),
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_FULL_OFFLOAD', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_FULL_OFFLOAD)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_TX_IV_IS_ESN', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_TX_IV_IS_ESN)),
 * 
*/
  __pyx_t_23 = __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_FULL_OFFLOAD); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_19 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_4, __pyx_t_23); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":28
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_SPI_STEERING', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_SPI_STEERING)),
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_FULL_OFFLOAD', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_FULL_OFFLOAD)),
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_TX_IV_IS_ESN', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_TX_IV_IS_ESN)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_23 = __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_TX_IV_IS_ESN); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_20 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_5, __pyx_t_23); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_flow_action_cap_flags = __Pyx_FlagBase('mlx5dv_flow_action_cap_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM)),
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_REQ_METADATA', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_REQ_METADATA)),
*/
  __pyx_t_23 = __Pyx_PyList_Pack(5, __pyx_t_22, __pyx_t_17, __pyx_t_16, __pyx_t_19, __pyx_t_20); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":30
 *     ('MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_TX_IV_IS_ESN', __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(MLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_TX_IV_IS_ESN)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_20 = __Pyx_Globals(); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_t_20, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_19, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_18, __pyx_mstate_global->__pyx_n_u_mlx5dv_flow_action_cap_flags, __pyx_t_23};
    __pyx_t_19 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_20, __pyx_t_19, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_19);
    __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_flow_action_cap_flags, __pyx_t_2) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":32
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":37
 * 
 * 
 *     mlx5dv_flow_action_cap_flags._member_names_ = list(mlx5dv_flow_action_cap_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_flow_action_cap_flags); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PySequence_ListKeepNew(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_mlx5dv_flow_action_cap_flags); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_2) < (0)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "EnumType":32
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_flow_action_type = __Pyx_FlagBase('mlx5dv_flow_action_type', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_FLOW_ACTION_DEST_IBV_QP', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DEST_IBV_QP)),
 *     ('MLX5DV_FLOW_ACTION_DROP', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DROP)),
*/
  __pyx_t_2 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_19 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_flow_action_type = __Pyx_FlagBase('mlx5dv_flow_action_type', [
 *     ('MLX5DV_FLOW_ACTION_DEST_IBV_QP', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DEST_IBV_QP)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_FLOW_ACTION_DROP', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DROP)),
 *     ('MLX5DV_FLOW_ACTION_IBV_COUNTER', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_IBV_COUNTER)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DEST_IBV_QP); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_23 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_ACTION_DEST_IBV_QP, __pyx_t_20); if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":25
 * mlx5dv_flow_action_type = __Pyx_FlagBase('mlx5dv_flow_action_type', [
 *     ('MLX5DV_FLOW_ACTION_DEST_IBV_QP', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DEST_IBV_QP)),
 *     ('MLX5DV_FLOW_ACTION_DROP', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DROP)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_FLOW_ACTION_IBV_COUNTER', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_IBV_COUNTER)),
 *     ('MLX5DV_FLOW_ACTION_IBV_FLOW_ACTION', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_IBV_FLOW_ACTION)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DROP); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_18 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_ACTION_DROP, __pyx_t_20); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":26
 *     ('MLX5DV_FLOW_ACTION_DEST_IBV_QP', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DEST_IBV_QP)),
 *     ('MLX5DV_FLOW_ACTION_DROP', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DROP)),
 *     ('MLX5DV_FLOW_ACTION_IBV_COUNTER', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_IBV_COUNTER)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_FLOW_ACTION_IBV_FLOW_ACTION', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_IBV_FLOW_ACTION)),
 *     ('MLX5DV_FLOW_ACTION_TAG', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_TAG)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_IBV_COUNTER); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_ACTION_IBV_COUNTER, __pyx_t_20); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":27
 *     ('MLX5DV_FLOW_ACTION_DROP', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DROP)),
 *     ('MLX5DV_FLOW_ACTION_IBV_COUNTER', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_IBV_COUNTER)),
 *     ('MLX5DV_FLOW_ACTION_IBV_FLOW_ACTION', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_IBV_FLOW_ACTION)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_FLOW_ACTION_TAG', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_TAG)),
 *     ('MLX5DV_FLOW_ACTION_DEST_DEVX', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DEST_DEVX)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_IBV_FLOW_ACTION); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_17 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_ACTION_IBV_FLOW_ACTI, __pyx_t_20); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":28
 *     ('MLX5DV_FLOW_ACTION_IBV_COUNTER', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_IBV_COUNTER)),
 *     ('MLX5DV_FLOW_ACTION_IBV_FLOW_ACTION', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_IBV_FLOW_ACTION)),
 *     ('MLX5DV_FLOW_ACTION_TAG', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_TAG)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_FLOW_ACTION_DEST_DEVX', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DEST_DEVX)),
 *     ('MLX5DV_FLOW_ACTION_COUNTERS_DEVX', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_COUNTERS_DEVX)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_TAG); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_ACTION_TAG, __pyx_t_20); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":29
 *     ('MLX5DV_FLOW_ACTION_IBV_FLOW_ACTION', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_IBV_FLOW_ACTION)),
 *     ('MLX5DV_FLOW_ACTION_TAG', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_TAG)),
 *     ('MLX5DV_FLOW_ACTION_DEST_DEVX', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DEST_DEVX)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_FLOW_ACTION_COUNTERS_DEVX', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_COUNTERS_DEVX)),
 *     ('MLX5DV_FLOW_ACTION_DEFAULT_MISS', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DEFAULT_MISS)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DEST_DEVX); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_21 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_ACTION_DEST_DEVX, __pyx_t_20); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":30
 *     ('MLX5DV_FLOW_ACTION_TAG', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_TAG)),
 *     ('MLX5DV_FLOW_ACTION_DEST_DEVX', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DEST_DEVX)),
 *     ('MLX5DV_FLOW_ACTION_COUNTERS_DEVX', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_COUNTERS_DEVX)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_FLOW_ACTION_DEFAULT_MISS', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DEFAULT_MISS)),
 *     ('MLX5DV_FLOW_ACTION_COUNTERS_DEVX_WITH_OFFSET', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_COUNTERS_DEVX_WITH_OFFSET)),
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_COUNTERS_DEVX); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_10 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_ACTION_COUNTERS_DEVX, __pyx_t_20); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":31
 *     ('MLX5DV_FLOW_ACTION_DEST_DEVX', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DEST_DEVX)),
 *     ('MLX5DV_FLOW_ACTION_COUNTERS_DEVX', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_COUNTERS_DEVX)),
 *     ('MLX5DV_FLOW_ACTION_DEFAULT_MISS', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DEFAULT_MISS)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_FLOW_ACTION_COUNTERS_DEVX_WITH_OFFSET', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_COUNTERS_DEVX_WITH_OFFSET)),
 * 
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DEFAULT_MISS); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_ACTION_DEFAULT_MISS, __pyx_t_20); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":32
 *     ('MLX5DV_FLOW_ACTION_COUNTERS_DEVX', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_COUNTERS_DEVX)),
 *     ('MLX5DV_FLOW_ACTION_DEFAULT_MISS', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DEFAULT_MISS)),
 *     ('MLX5DV_FLOW_ACTION_COUNTERS_DEVX_WITH_OFFSET', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_COUNTERS_DEVX_WITH_OFFSET)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_COUNTERS_DEVX_WITH_OFFSET); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_ACTION_COUNTERS_DEVX_2, __pyx_t_20); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_flow_action_type = __Pyx_FlagBase('mlx5dv_flow_action_type', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_FLOW_ACTION_DEST_IBV_QP', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DEST_IBV_QP)),
 *     ('MLX5DV_FLOW_ACTION_DROP', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_DROP)),
*/
  __pyx_t_20 = __Pyx_PyList_Pack(9, __pyx_t_23, __pyx_t_18, __pyx_t_16, __pyx_t_17, __pyx_t_22, __pyx_t_21, __pyx_t_10, __pyx_t_11, __pyx_t_15); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":34
 *     ('MLX5DV_FLOW_ACTION_COUNTERS_DEVX_WITH_OFFSET', __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(MLX5DV_FLOW_ACTION_COUNTERS_DEVX_WITH_OFFSET)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_15 = __Pyx_Globals(); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_flow_action_type, __pyx_t_20};
    __pyx_t_11 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_15, __pyx_t_11, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_4 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_19, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_11);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_flow_action_type, __pyx_t_4) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":36
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":41
 * 
 * 
 *     mlx5dv_flow_action_type._member_names_ = list(mlx5dv_flow_action_type.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_mlx5dv_flow_action_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PySequence_ListKeepNew(__pyx_t_19); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_19, __pyx_mstate_global->__pyx_n_u_mlx5dv_flow_action_type); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 41, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_19, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_4) < (0)) __PYX_ERR(1, 41, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

    /* "EnumType":36
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_flow_matcher_attr_mask = __Pyx_FlagBase('mlx5dv_flow_matcher_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_FLOW_MATCHER_MASK_FT_TYPE', __Pyx_PyLong_From_enum__mlx5dv_flow_matcher_attr_mask(MLX5DV_FLOW_MATCHER_MASK_FT_TYPE)),
 *     ('MLX5DV_FLOW_MATCHER_MASK_IB_PORT', __Pyx_PyLong_From_enum__mlx5dv_flow_matcher_attr_mask(MLX5DV_FLOW_MATCHER_MASK_IB_PORT)),
*/
  __pyx_t_4 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_11 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_flow_matcher_attr_mask = __Pyx_FlagBase('mlx5dv_flow_matcher_attr_mask', [
 *     ('MLX5DV_FLOW_MATCHER_MASK_FT_TYPE', __Pyx_PyLong_From_enum__mlx5dv_flow_matcher_attr_mask(MLX5DV_FLOW_MATCHER_MASK_FT_TYPE)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_FLOW_MATCHER_MASK_IB_PORT', __Pyx_PyLong_From_enum__mlx5dv_flow_matcher_attr_mask(MLX5DV_FLOW_MATCHER_MASK_IB_PORT)),
 * 
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__mlx5dv_flow_matcher_attr_mask(MLX5DV_FLOW_MATCHER_MASK_FT_TYPE); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_20 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_MATCHER_MASK_FT_TYPE, __pyx_t_15); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":25
 * mlx5dv_flow_matcher_attr_mask = __Pyx_FlagBase('mlx5dv_flow_matcher_attr_mask', [
 *     ('MLX5DV_FLOW_MATCHER_MASK_FT_TYPE', __Pyx_PyLong_From_enum__mlx5dv_flow_matcher_attr_mask(MLX5DV_FLOW_MATCHER_MASK_FT_TYPE)),
 *     ('MLX5DV_FLOW_MATCHER_MASK_IB_PORT', __Pyx_PyLong_From_enum__mlx5dv_flow_matcher_attr_mask(MLX5DV_FLOW_MATCHER_MASK_IB_PORT)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__mlx5dv_flow_matcher_attr_mask(MLX5DV_FLOW_MATCHER_MASK_IB_PORT); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_2 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_MATCHER_MASK_IB_PORT, __pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_flow_matcher_attr_mask = __Pyx_FlagBase('mlx5dv_flow_matcher_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_FLOW_MATCHER_MASK_FT_TYPE', __Pyx_PyLong_From_enum__mlx5dv_flow_matcher_attr_mask(MLX5DV_FLOW_MATCHER_MASK_FT_TYPE)),
 *     ('MLX5DV_FLOW_MATCHER_MASK_IB_PORT', __Pyx_PyLong_From_enum__mlx5dv_flow_matcher_attr_mask(MLX5DV_FLOW_MATCHER_MASK_IB_PORT)),
*/
  __pyx_t_15 = __Pyx_PyList_Pack(2, __pyx_t_20, __pyx_t_2); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":27
 *     ('MLX5DV_FLOW_MATCHER_MASK_IB_PORT', __Pyx_PyLong_From_enum__mlx5dv_flow_matcher_attr_mask(MLX5DV_FLOW_MATCHER_MASK_IB_PORT)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_2 = __Pyx_Globals(); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_20, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_4, __pyx_mstate_global->__pyx_n_u_mlx5dv_flow_matcher_attr_mask, __pyx_t_15};
    __pyx_t_20 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_2, __pyx_t_20, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_19 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_11, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_20);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_flow_matcher_attr_mask, __pyx_t_19) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":34
 * 
 * 
 *     mlx5dv_flow_matcher_attr_mask._member_names_ = list(mlx5dv_flow_matcher_attr_mask.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_19, __pyx_mstate_global->__pyx_n_u_mlx5dv_flow_matcher_attr_mask); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_19, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    __pyx_t_19 = __Pyx_PySequence_ListKeepNew(__pyx_t_11); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_mlx5dv_flow_matcher_attr_mask); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_19) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_mkey_err_type = __Pyx_FlagBase('mlx5dv_mkey_err_type', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_MKEY_NO_ERR', __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(MLX5DV_MKEY_NO_ERR)),
 *     ('MLX5DV_MKEY_SIG_BLOCK_BAD_GUARD', __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(MLX5DV_MKEY_SIG_BLOCK_BAD_GUARD)),
*/
  __pyx_t_19 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_20 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_mkey_err_type = __Pyx_FlagBase('mlx5dv_mkey_err_type', [
 *     ('MLX5DV_MKEY_NO_ERR', __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(MLX5DV_MKEY_NO_ERR)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_MKEY_SIG_BLOCK_BAD_GUARD', __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(MLX5DV_MKEY_SIG_BLOCK_BAD_GUARD)),
 *     ('MLX5DV_MKEY_SIG_BLOCK_BAD_REFTAG', __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(MLX5DV_MKEY_SIG_BLOCK_BAD_REFTAG)),
*/
  __pyx_t_2 = __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(MLX5DV_MKEY_NO_ERR); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_MKEY_NO_ERR, __pyx_t_2); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":25
 * mlx5dv_mkey_err_type = __Pyx_FlagBase('mlx5dv_mkey_err_type', [
 *     ('MLX5DV_MKEY_NO_ERR', __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(MLX5DV_MKEY_NO_ERR)),
 *     ('MLX5DV_MKEY_SIG_BLOCK_BAD_GUARD', __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(MLX5DV_MKEY_SIG_BLOCK_BAD_GUARD)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_MKEY_SIG_BLOCK_BAD_REFTAG', __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(MLX5DV_MKEY_SIG_BLOCK_BAD_REFTAG)),
 *     ('MLX5DV_MKEY_SIG_BLOCK_BAD_APPTAG', __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(MLX5DV_MKEY_SIG_BLOCK_BAD_APPTAG)),
*/
  __pyx_t_2 = __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(MLX5DV_MKEY_SIG_BLOCK_BAD_GUARD); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_MKEY_SIG_BLOCK_BAD_GUARD, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":26
 *     ('MLX5DV_MKEY_NO_ERR', __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(MLX5DV_MKEY_NO_ERR)),
 *     ('MLX5DV_MKEY_SIG_BLOCK_BAD_GUARD', __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(MLX5DV_MKEY_SIG_BLOCK_BAD_GUARD)),
 *     ('MLX5DV_MKEY_SIG_BLOCK_BAD_REFTAG', __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(MLX5DV_MKEY_SIG_BLOCK_BAD_REFTAG)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_MKEY_SIG_BLOCK_BAD_APPTAG', __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(MLX5DV_MKEY_SIG_BLOCK_BAD_APPTAG)),
 * 
*/
  __pyx_t_2 = __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(MLX5DV_MKEY_SIG_BLOCK_BAD_REFTAG); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_10 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_MKEY_SIG_BLOCK_BAD_REFTAG, __pyx_t_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":27
 *     ('MLX5DV_MKEY_SIG_BLOCK_BAD_GUARD', __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(MLX5DV_MKEY_SIG_BLOCK_BAD_GUARD)),
 *     ('MLX5DV_MKEY_SIG_BLOCK_BAD_REFTAG', __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(MLX5DV_MKEY_SIG_BLOCK_BAD_REFTAG)),
 *     ('MLX5DV_MKEY_SIG_BLOCK_BAD_APPTAG', __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(MLX5DV_MKEY_SIG_BLOCK_BAD_APPTAG)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_2 = __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(MLX5DV_MKEY_SIG_BLOCK_BAD_APPTAG); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_21 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_MKEY_SIG_BLOCK_BAD_APPTAG, __pyx_t_2); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_mkey_err_type = __Pyx_FlagBase('mlx5dv_mkey_err_type', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_MKEY_NO_ERR', __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(MLX5DV_MKEY_NO_ERR)),
 *     ('MLX5DV_MKEY_SIG_BLOCK_BAD_GUARD', __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(MLX5DV_MKEY_SIG_BLOCK_BAD_GUARD)),
*/
  __pyx_t_2 = __Pyx_PyList_Pack(4, __pyx_t_15, __pyx_t_4, __pyx_t_10, __pyx_t_21); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":29
 *     ('MLX5DV_MKEY_SIG_BLOCK_BAD_APPTAG', __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(MLX5DV_MKEY_SIG_BLOCK_BAD_APPTAG)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_21 = __Pyx_Globals(); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_21 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_19, __pyx_mstate_global->__pyx_n_u_mlx5dv_mkey_err_type, __pyx_t_2};
    __pyx_t_10 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_21, __pyx_t_10, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_11 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_20, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_mkey_err_type, __pyx_t_11) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":31
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":36
 * 
 * 
 *     mlx5dv_mkey_err_type._member_names_ = list(mlx5dv_mkey_err_type.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_mlx5dv_mkey_err_type); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PySequence_ListKeepNew(__pyx_t_20); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_20, __pyx_mstate_global->__pyx_n_u_mlx5dv_mkey_err_type); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_20, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_11) < (0)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

    /* "EnumType":31
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_mkey_init_attr_flags = __Pyx_FlagBase('mlx5dv_mkey_init_attr_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_MKEY_INIT_ATTR_FLAGS_INDIRECT', __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(MLX5DV_MKEY_INIT_ATTR_FLAGS_INDIRECT)),
 *     ('MLX5DV_MKEY_INIT_ATTR_FLAGS_BLOCK_SIGNATURE', __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(MLX5DV_MKEY_INIT_ATTR_FLAGS_BLOCK_SIGNATURE)),
*/
  __pyx_t_11 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_10 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_mkey_init_attr_flags = __Pyx_FlagBase('mlx5dv_mkey_init_attr_flags', [
 *     ('MLX5DV_MKEY_INIT_ATTR_FLAGS_INDIRECT', __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(MLX5DV_MKEY_INIT_ATTR_FLAGS_INDIRECT)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_MKEY_INIT_ATTR_FLAGS_BLOCK_SIGNATURE', __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(MLX5DV_MKEY_INIT_ATTR_FLAGS_BLOCK_SIGNATURE)),
 *     ('MLX5DV_MKEY_INIT_ATTR_FLAGS_CRYPTO', __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(MLX5DV_MKEY_INIT_ATTR_FLAGS_CRYPTO)),
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(MLX5DV_MKEY_INIT_ATTR_FLAGS_INDIRECT); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_2 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_MKEY_INIT_ATTR_FLAGS_INDI, __pyx_t_21); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":25
 * mlx5dv_mkey_init_attr_flags = __Pyx_FlagBase('mlx5dv_mkey_init_attr_flags', [
 *     ('MLX5DV_MKEY_INIT_ATTR_FLAGS_INDIRECT', __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(MLX5DV_MKEY_INIT_ATTR_FLAGS_INDIRECT)),
 *     ('MLX5DV_MKEY_INIT_ATTR_FLAGS_BLOCK_SIGNATURE', __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(MLX5DV_MKEY_INIT_ATTR_FLAGS_BLOCK_SIGNATURE)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_MKEY_INIT_ATTR_FLAGS_CRYPTO', __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(MLX5DV_MKEY_INIT_ATTR_FLAGS_CRYPTO)),
 *     ('MLX5DV_MKEY_INIT_ATTR_FLAGS_REMOTE_INVALIDATE', __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(MLX5DV_MKEY_INIT_ATTR_FLAGS_REMOTE_INVALIDATE)),
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(MLX5DV_MKEY_INIT_ATTR_FLAGS_BLOCK_SIGNATURE); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_19 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_MKEY_INIT_ATTR_FLAGS_BLOC, __pyx_t_21); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":26
 *     ('MLX5DV_MKEY_INIT_ATTR_FLAGS_INDIRECT', __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(MLX5DV_MKEY_INIT_ATTR_FLAGS_INDIRECT)),
 *     ('MLX5DV_MKEY_INIT_ATTR_FLAGS_BLOCK_SIGNATURE', __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(MLX5DV_MKEY_INIT_ATTR_FLAGS_BLOCK_SIGNATURE)),
 *     ('MLX5DV_MKEY_INIT_ATTR_FLAGS_CRYPTO', __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(MLX5DV_MKEY_INIT_ATTR_FLAGS_CRYPTO)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_MKEY_INIT_ATTR_FLAGS_REMOTE_INVALIDATE', __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(MLX5DV_MKEY_INIT_ATTR_FLAGS_REMOTE_INVALIDATE)),
 * 
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(MLX5DV_MKEY_INIT_ATTR_FLAGS_CRYPTO); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_MKEY_INIT_ATTR_FLAGS_CRYP, __pyx_t_21); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":27
 *     ('MLX5DV_MKEY_INIT_ATTR_FLAGS_BLOCK_SIGNATURE', __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(MLX5DV_MKEY_INIT_ATTR_FLAGS_BLOCK_SIGNATURE)),
 *     ('MLX5DV_MKEY_INIT_ATTR_FLAGS_CRYPTO', __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(MLX5DV_MKEY_INIT_ATTR_FLAGS_CRYPTO)),
 *     ('MLX5DV_MKEY_INIT_ATTR_FLAGS_REMOTE_INVALIDATE', __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(MLX5DV_MKEY_INIT_ATTR_FLAGS_REMOTE_INVALIDATE)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_21 = __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(MLX5DV_MKEY_INIT_ATTR_FLAGS_REMOTE_INVALIDATE); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_MKEY_INIT_ATTR_FLAGS_REMO, __pyx_t_21); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_mkey_init_attr_flags = __Pyx_FlagBase('mlx5dv_mkey_init_attr_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_MKEY_INIT_ATTR_FLAGS_INDIRECT', __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(MLX5DV_MKEY_INIT_ATTR_FLAGS_INDIRECT)),
 *     ('MLX5DV_MKEY_INIT_ATTR_FLAGS_BLOCK_SIGNATURE', __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(MLX5DV_MKEY_INIT_ATTR_FLAGS_BLOCK_SIGNATURE)),
*/
  __pyx_t_21 = __Pyx_PyList_Pack(4, __pyx_t_2, __pyx_t_19, __pyx_t_4, __pyx_t_15); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":29
 *     ('MLX5DV_MKEY_INIT_ATTR_FLAGS_REMOTE_INVALIDATE', __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(MLX5DV_MKEY_INIT_ATTR_FLAGS_REMOTE_INVALIDATE)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_15 = __Pyx_Globals(); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_11, __pyx_mstate_global->__pyx_n_u_mlx5dv_mkey_init_attr_flags, __pyx_t_21};
    __pyx_t_4 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_15, __pyx_t_4, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_20 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_10, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_mkey_init_attr_flags, __pyx_t_20) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":31
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":36
 * 
 * 
 *     mlx5dv_mkey_init_attr_flags._member_names_ = list(mlx5dv_mkey_init_attr_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_20, __pyx_mstate_global->__pyx_n_u_mlx5dv_mkey_init_attr_flags); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_20, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __pyx_t_20 = __Pyx_PySequence_ListKeepNew(__pyx_t_10); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_mlx5dv_mkey_init_attr_flags); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_20) < (0)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "EnumType":31
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_obj_type = __Pyx_FlagBase('mlx5dv_obj_type', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_OBJ_QP', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_QP)),
 *     ('MLX5DV_OBJ_CQ', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_CQ)),
*/
  __pyx_t_20 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_4 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_obj_type = __Pyx_FlagBase('mlx5dv_obj_type', [
 *     ('MLX5DV_OBJ_QP', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_QP)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_OBJ_CQ', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_CQ)),
 *     ('MLX5DV_OBJ_SRQ', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_SRQ)),
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_QP); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_21 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_OBJ_QP, __pyx_t_15); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":25
 * mlx5dv_obj_type = __Pyx_FlagBase('mlx5dv_obj_type', [
 *     ('MLX5DV_OBJ_QP', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_QP)),
 *     ('MLX5DV_OBJ_CQ', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_CQ)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_OBJ_SRQ', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_SRQ)),
 *     ('MLX5DV_OBJ_RWQ', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_RWQ)),
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_CQ); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_OBJ_CQ, __pyx_t_15); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":26
 *     ('MLX5DV_OBJ_QP', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_QP)),
 *     ('MLX5DV_OBJ_CQ', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_CQ)),
 *     ('MLX5DV_OBJ_SRQ', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_SRQ)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_OBJ_RWQ', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_RWQ)),
 *     ('MLX5DV_OBJ_DM', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_DM)),
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_SRQ); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_19 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_OBJ_SRQ, __pyx_t_15); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":27
 *     ('MLX5DV_OBJ_CQ', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_CQ)),
 *     ('MLX5DV_OBJ_SRQ', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_SRQ)),
 *     ('MLX5DV_OBJ_RWQ', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_RWQ)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_OBJ_DM', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_DM)),
 *     ('MLX5DV_OBJ_AH', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_AH)),
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_RWQ); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_2 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_OBJ_RWQ, __pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":28
 *     ('MLX5DV_OBJ_SRQ', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_SRQ)),
 *     ('MLX5DV_OBJ_RWQ', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_RWQ)),
 *     ('MLX5DV_OBJ_DM', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_DM)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_OBJ_AH', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_AH)),
 *     ('MLX5DV_OBJ_PD', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_PD)),
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_DM); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_OBJ_DM, __pyx_t_15); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":29
 *     ('MLX5DV_OBJ_RWQ', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_RWQ)),
 *     ('MLX5DV_OBJ_DM', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_DM)),
 *     ('MLX5DV_OBJ_AH', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_AH)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_OBJ_PD', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_PD)),
 * 
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_AH); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_17 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_OBJ_AH, __pyx_t_15); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":30
 *     ('MLX5DV_OBJ_DM', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_DM)),
 *     ('MLX5DV_OBJ_AH', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_AH)),
 *     ('MLX5DV_OBJ_PD', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_PD)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_PD); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_OBJ_PD, __pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_obj_type = __Pyx_FlagBase('mlx5dv_obj_type', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_OBJ_QP', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_QP)),
 *     ('MLX5DV_OBJ_CQ', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_CQ)),
*/
  __pyx_t_15 = __Pyx_PyList_Pack(7, __pyx_t_21, __pyx_t_11, __pyx_t_19, __pyx_t_2, __pyx_t_22, __pyx_t_17, __pyx_t_16); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":32
 *     ('MLX5DV_OBJ_PD', __Pyx_PyLong_From_enum__mlx5dv_obj_type(MLX5DV_OBJ_PD)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_16 = __Pyx_Globals(); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_20, __pyx_mstate_global->__pyx_n_u_mlx5dv_obj_type, __pyx_t_15};
    __pyx_t_17 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_16, __pyx_t_17, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_10 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_17);
    __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_obj_type, __pyx_t_10) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "EnumType":34
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":39
 * 
 * 
 *     mlx5dv_obj_type._member_names_ = list(mlx5dv_obj_type.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_mlx5dv_obj_type); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PySequence_ListKeepNew(__pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_mlx5dv_obj_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_10) < (0)) __PYX_ERR(1, 39, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "EnumType":34
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_qp_comp_mask = __Pyx_FlagBase('mlx5dv_qp_comp_mask', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_MASK_UAR_MMAP_OFFSET', __Pyx_PyLong_From_enum__mlx5dv_qp_comp_mask(MLX5DV_QP_MASK_UAR_MMAP_OFFSET)),
 *     ('MLX5DV_QP_MASK_RAW_QP_HANDLES', __Pyx_PyLong_From_enum__mlx5dv_qp_comp_mask(MLX5DV_QP_MASK_RAW_QP_HANDLES)),
*/
  __pyx_t_10 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_17 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_qp_comp_mask = __Pyx_FlagBase('mlx5dv_qp_comp_mask', [
 *     ('MLX5DV_QP_MASK_UAR_MMAP_OFFSET', __Pyx_PyLong_From_enum__mlx5dv_qp_comp_mask(MLX5DV_QP_MASK_UAR_MMAP_OFFSET)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_MASK_RAW_QP_HANDLES', __Pyx_PyLong_From_enum__mlx5dv_qp_comp_mask(MLX5DV_QP_MASK_RAW_QP_HANDLES)),
 *     ('MLX5DV_QP_MASK_RAW_QP_TIR_ADDR', __Pyx_PyLong_From_enum__mlx5dv_qp_comp_mask(MLX5DV_QP_MASK_RAW_QP_TIR_ADDR)),
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__mlx5dv_qp_comp_mask(MLX5DV_QP_MASK_UAR_MMAP_OFFSET); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_QP_MASK_UAR_MMAP_OFFSET, __pyx_t_16); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":25
 * mlx5dv_qp_comp_mask = __Pyx_FlagBase('mlx5dv_qp_comp_mask', [
 *     ('MLX5DV_QP_MASK_UAR_MMAP_OFFSET', __Pyx_PyLong_From_enum__mlx5dv_qp_comp_mask(MLX5DV_QP_MASK_UAR_MMAP_OFFSET)),
 *     ('MLX5DV_QP_MASK_RAW_QP_HANDLES', __Pyx_PyLong_From_enum__mlx5dv_qp_comp_mask(MLX5DV_QP_MASK_RAW_QP_HANDLES)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_MASK_RAW_QP_TIR_ADDR', __Pyx_PyLong_From_enum__mlx5dv_qp_comp_mask(MLX5DV_QP_MASK_RAW_QP_TIR_ADDR)),
 * 
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__mlx5dv_qp_comp_mask(MLX5DV_QP_MASK_RAW_QP_HANDLES); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_20 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_QP_MASK_RAW_QP_HANDLES, __pyx_t_16); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":26
 *     ('MLX5DV_QP_MASK_UAR_MMAP_OFFSET', __Pyx_PyLong_From_enum__mlx5dv_qp_comp_mask(MLX5DV_QP_MASK_UAR_MMAP_OFFSET)),
 *     ('MLX5DV_QP_MASK_RAW_QP_HANDLES', __Pyx_PyLong_From_enum__mlx5dv_qp_comp_mask(MLX5DV_QP_MASK_RAW_QP_HANDLES)),
 *     ('MLX5DV_QP_MASK_RAW_QP_TIR_ADDR', __Pyx_PyLong_From_enum__mlx5dv_qp_comp_mask(MLX5DV_QP_MASK_RAW_QP_TIR_ADDR)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_16 = __Pyx_PyLong_From_enum__mlx5dv_qp_comp_mask(MLX5DV_QP_MASK_RAW_QP_TIR_ADDR); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_QP_MASK_RAW_QP_TIR_ADDR, __pyx_t_16); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_qp_comp_mask = __Pyx_FlagBase('mlx5dv_qp_comp_mask', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_MASK_UAR_MMAP_OFFSET', __Pyx_PyLong_From_enum__mlx5dv_qp_comp_mask(MLX5DV_QP_MASK_UAR_MMAP_OFFSET)),
 *     ('MLX5DV_QP_MASK_RAW_QP_HANDLES', __Pyx_PyLong_From_enum__mlx5dv_qp_comp_mask(MLX5DV_QP_MASK_RAW_QP_HANDLES)),
*/
  __pyx_t_16 = __Pyx_PyList_Pack(3, __pyx_t_15, __pyx_t_20, __pyx_t_22); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":28
 *     ('MLX5DV_QP_MASK_RAW_QP_TIR_ADDR', __Pyx_PyLong_From_enum__mlx5dv_qp_comp_mask(MLX5DV_QP_MASK_RAW_QP_TIR_ADDR)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_22 = __Pyx_Globals(); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_20, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_10, __pyx_mstate_global->__pyx_n_u_mlx5dv_qp_comp_mask, __pyx_t_16};
    __pyx_t_20 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_22, __pyx_t_20, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_4 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_17, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_20);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_qp_comp_mask, __pyx_t_4) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":35
 * 
 * 
 *     mlx5dv_qp_comp_mask._member_names_ = list(mlx5dv_qp_comp_mask.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_mlx5dv_qp_comp_mask); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PySequence_ListKeepNew(__pyx_t_17); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_qp_comp_mask); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_4) < (0)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

    /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_qp_create_flags = __Pyx_FlagBase('mlx5dv_qp_create_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_CREATE_TUNNEL_OFFLOADS', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_TUNNEL_OFFLOADS)),
 *     ('MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_UC', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_UC)),
*/
  __pyx_t_4 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_20 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_qp_create_flags = __Pyx_FlagBase('mlx5dv_qp_create_flags', [
 *     ('MLX5DV_QP_CREATE_TUNNEL_OFFLOADS', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_TUNNEL_OFFLOADS)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_UC', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_UC)),
 *     ('MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_MC', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_MC)),
*/
  __pyx_t_22 = __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_TUNNEL_OFFLOADS); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_16 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_QP_CREATE_TUNNEL_OFFLOADS, __pyx_t_22); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":25
 * mlx5dv_qp_create_flags = __Pyx_FlagBase('mlx5dv_qp_create_flags', [
 *     ('MLX5DV_QP_CREATE_TUNNEL_OFFLOADS', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_TUNNEL_OFFLOADS)),
 *     ('MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_UC', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_UC)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_MC', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_MC)),
 *     ('MLX5DV_QP_CREATE_DISABLE_SCATTER_TO_CQE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_DISABLE_SCATTER_TO_CQE)),
*/
  __pyx_t_22 = __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_UC); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_10 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_QP_CREATE_TIR_ALLOW_SELF, __pyx_t_22); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":26
 *     ('MLX5DV_QP_CREATE_TUNNEL_OFFLOADS', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_TUNNEL_OFFLOADS)),
 *     ('MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_UC', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_UC)),
 *     ('MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_MC', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_MC)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_CREATE_DISABLE_SCATTER_TO_CQE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_DISABLE_SCATTER_TO_CQE)),
 *     ('MLX5DV_QP_CREATE_ALLOW_SCATTER_TO_CQE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_ALLOW_SCATTER_TO_CQE)),
*/
  __pyx_t_22 = __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_MC); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_QP_CREATE_TIR_ALLOW_SELF_2, __pyx_t_22); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":27
 *     ('MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_UC', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_UC)),
 *     ('MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_MC', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_MC)),
 *     ('MLX5DV_QP_CREATE_DISABLE_SCATTER_TO_CQE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_DISABLE_SCATTER_TO_CQE)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_CREATE_ALLOW_SCATTER_TO_CQE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_ALLOW_SCATTER_TO_CQE)),
 *     ('MLX5DV_QP_CREATE_PACKET_BASED_CREDIT_MODE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_PACKET_BASED_CREDIT_MODE)),
*/
  __pyx_t_22 = __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_DISABLE_SCATTER_TO_CQE); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_2 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_QP_CREATE_DISABLE_SCATTER, __pyx_t_22); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":28
 *     ('MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_MC', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_MC)),
 *     ('MLX5DV_QP_CREATE_DISABLE_SCATTER_TO_CQE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_DISABLE_SCATTER_TO_CQE)),
 *     ('MLX5DV_QP_CREATE_ALLOW_SCATTER_TO_CQE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_ALLOW_SCATTER_TO_CQE)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_CREATE_PACKET_BASED_CREDIT_MODE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_PACKET_BASED_CREDIT_MODE)),
 *     ('MLX5DV_QP_CREATE_SIG_PIPELINING', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_SIG_PIPELINING)),
*/
  __pyx_t_22 = __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_ALLOW_SCATTER_TO_CQE); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_19 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_QP_CREATE_ALLOW_SCATTER_T, __pyx_t_22); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":29
 *     ('MLX5DV_QP_CREATE_DISABLE_SCATTER_TO_CQE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_DISABLE_SCATTER_TO_CQE)),
 *     ('MLX5DV_QP_CREATE_ALLOW_SCATTER_TO_CQE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_ALLOW_SCATTER_TO_CQE)),
 *     ('MLX5DV_QP_CREATE_PACKET_BASED_CREDIT_MODE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_PACKET_BASED_CREDIT_MODE)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_CREATE_SIG_PIPELINING', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_SIG_PIPELINING)),
 *     ('MLX5DV_QP_CREATE_OOO_DP', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_OOO_DP)),
*/
  __pyx_t_22 = __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_PACKET_BASED_CREDIT_MODE); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_QP_CREATE_PACKET_BASED_CR, __pyx_t_22); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":30
 *     ('MLX5DV_QP_CREATE_ALLOW_SCATTER_TO_CQE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_ALLOW_SCATTER_TO_CQE)),
 *     ('MLX5DV_QP_CREATE_PACKET_BASED_CREDIT_MODE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_PACKET_BASED_CREDIT_MODE)),
 *     ('MLX5DV_QP_CREATE_SIG_PIPELINING', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_SIG_PIPELINING)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_CREATE_OOO_DP', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_OOO_DP)),
 * 
*/
  __pyx_t_22 = __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_SIG_PIPELINING); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_21 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_QP_CREATE_SIG_PIPELINING, __pyx_t_22); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":31
 *     ('MLX5DV_QP_CREATE_PACKET_BASED_CREDIT_MODE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_PACKET_BASED_CREDIT_MODE)),
 *     ('MLX5DV_QP_CREATE_SIG_PIPELINING', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_SIG_PIPELINING)),
 *     ('MLX5DV_QP_CREATE_OOO_DP', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_OOO_DP)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_22 = __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_OOO_DP); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_18 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_QP_CREATE_OOO_DP, __pyx_t_22); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_qp_create_flags = __Pyx_FlagBase('mlx5dv_qp_create_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_CREATE_TUNNEL_OFFLOADS', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_TUNNEL_OFFLOADS)),
 *     ('MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_UC', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_UC)),
*/
  __pyx_t_22 = __Pyx_PyList_Pack(8, __pyx_t_16, __pyx_t_10, __pyx_t_15, __pyx_t_2, __pyx_t_19, __pyx_t_11, __pyx_t_21, __pyx_t_18); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "EnumType":33
 *     ('MLX5DV_QP_CREATE_OOO_DP', __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(MLX5DV_QP_CREATE_OOO_DP)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_18 = __Pyx_Globals(); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __pyx_t_18 = __Pyx_PyObject_Call(__pyx_t_21, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_4, __pyx_mstate_global->__pyx_n_u_mlx5dv_qp_create_flags, __pyx_t_22};
    __pyx_t_21 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_18, __pyx_t_21, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_17 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_20, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_21);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_qp_create_flags, __pyx_t_17) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "EnumType":35
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":40
 * 
 * 
 *     mlx5dv_qp_create_flags._member_names_ = list(mlx5dv_qp_create_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_qp_create_flags); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __pyx_t_17 = __Pyx_PySequence_ListKeepNew(__pyx_t_20); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_20, __pyx_mstate_global->__pyx_n_u_mlx5dv_qp_create_flags); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_20, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_17) < (0)) __PYX_ERR(1, 40, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

    /* "EnumType":35
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_qp_create_send_ops_flags = __Pyx_FlagBase('mlx5dv_qp_create_send_ops_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_EX_WITH_MR_INTERLEAVED', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MR_INTERLEAVED)),
 *     ('MLX5DV_QP_EX_WITH_MR_LIST', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MR_LIST)),
*/
  __pyx_t_17 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_21 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_qp_create_send_ops_flags = __Pyx_FlagBase('mlx5dv_qp_create_send_ops_flags', [
 *     ('MLX5DV_QP_EX_WITH_MR_INTERLEAVED', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MR_INTERLEAVED)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_EX_WITH_MR_LIST', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MR_LIST)),
 *     ('MLX5DV_QP_EX_WITH_MKEY_CONFIGURE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MKEY_CONFIGURE)),
*/
  __pyx_t_18 = __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MR_INTERLEAVED); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_QP_EX_WITH_MR_INTERLEAVED, __pyx_t_18); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "EnumType":25
 * mlx5dv_qp_create_send_ops_flags = __Pyx_FlagBase('mlx5dv_qp_create_send_ops_flags', [
 *     ('MLX5DV_QP_EX_WITH_MR_INTERLEAVED', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MR_INTERLEAVED)),
 *     ('MLX5DV_QP_EX_WITH_MR_LIST', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MR_LIST)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_EX_WITH_MKEY_CONFIGURE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MKEY_CONFIGURE)),
 *     ('MLX5DV_QP_EX_WITH_RAW_WQE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_RAW_WQE)),
*/
  __pyx_t_18 = __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MR_LIST); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_QP_EX_WITH_MR_LIST, __pyx_t_18); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "EnumType":26
 *     ('MLX5DV_QP_EX_WITH_MR_INTERLEAVED', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MR_INTERLEAVED)),
 *     ('MLX5DV_QP_EX_WITH_MR_LIST', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MR_LIST)),
 *     ('MLX5DV_QP_EX_WITH_MKEY_CONFIGURE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MKEY_CONFIGURE)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_EX_WITH_RAW_WQE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_RAW_WQE)),
 *     ('MLX5DV_QP_EX_WITH_MEMCPY', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MEMCPY)),
*/
  __pyx_t_18 = __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MKEY_CONFIGURE); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_QP_EX_WITH_MKEY_CONFIGURE, __pyx_t_18); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "EnumType":27
 *     ('MLX5DV_QP_EX_WITH_MR_LIST', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MR_LIST)),
 *     ('MLX5DV_QP_EX_WITH_MKEY_CONFIGURE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MKEY_CONFIGURE)),
 *     ('MLX5DV_QP_EX_WITH_RAW_WQE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_RAW_WQE)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_EX_WITH_MEMCPY', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MEMCPY)),
 * 
*/
  __pyx_t_18 = __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_RAW_WQE); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_19 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_QP_EX_WITH_RAW_WQE, __pyx_t_18); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "EnumType":28
 *     ('MLX5DV_QP_EX_WITH_MKEY_CONFIGURE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MKEY_CONFIGURE)),
 *     ('MLX5DV_QP_EX_WITH_RAW_WQE', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_RAW_WQE)),
 *     ('MLX5DV_QP_EX_WITH_MEMCPY', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MEMCPY)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_18 = __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MEMCPY); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_2 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_QP_EX_WITH_MEMCPY, __pyx_t_18); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_qp_create_send_ops_flags = __Pyx_FlagBase('mlx5dv_qp_create_send_ops_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_EX_WITH_MR_INTERLEAVED', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MR_INTERLEAVED)),
 *     ('MLX5DV_QP_EX_WITH_MR_LIST', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MR_LIST)),
*/
  __pyx_t_18 = __Pyx_PyList_Pack(5, __pyx_t_22, __pyx_t_4, __pyx_t_11, __pyx_t_19, __pyx_t_2); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":30
 *     ('MLX5DV_QP_EX_WITH_MEMCPY', __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(MLX5DV_QP_EX_WITH_MEMCPY)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_2 = __Pyx_Globals(); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_19, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_qp_create_send_ops_flags, __pyx_t_18};
    __pyx_t_19 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_2, __pyx_t_19, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_20 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_21, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_19);
    __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_qp_create_send_ops_flags, __pyx_t_20) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":32
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":37
 * 
 * 
 *     mlx5dv_qp_create_send_ops_flags._member_names_ = list(mlx5dv_qp_create_send_ops_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_20, __pyx_mstate_global->__pyx_n_u_mlx5dv_qp_create_send_ops_flags); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_20, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __pyx_t_20 = __Pyx_PySequence_ListKeepNew(__pyx_t_21); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_mlx5dv_qp_create_send_ops_flags); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_20) < (0)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

    /* "EnumType":32
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_qp_init_attr_mask = __Pyx_FlagBase('mlx5dv_qp_init_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_INIT_ATTR_MASK_QP_CREATE_FLAGS', __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(MLX5DV_QP_INIT_ATTR_MASK_QP_CREATE_FLAGS)),
 *     ('MLX5DV_QP_INIT_ATTR_MASK_DC', __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(MLX5DV_QP_INIT_ATTR_MASK_DC)),
*/
  __pyx_t_20 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_19 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_qp_init_attr_mask = __Pyx_FlagBase('mlx5dv_qp_init_attr_mask', [
 *     ('MLX5DV_QP_INIT_ATTR_MASK_QP_CREATE_FLAGS', __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(MLX5DV_QP_INIT_ATTR_MASK_QP_CREATE_FLAGS)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_INIT_ATTR_MASK_DC', __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(MLX5DV_QP_INIT_ATTR_MASK_DC)),
 *     ('MLX5DV_QP_INIT_ATTR_MASK_SEND_OPS_FLAGS', __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(MLX5DV_QP_INIT_ATTR_MASK_SEND_OPS_FLAGS)),
*/
  __pyx_t_2 = __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(MLX5DV_QP_INIT_ATTR_MASK_QP_CREATE_FLAGS); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_18 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_QP_INIT_ATTR_MASK_QP_CREA, __pyx_t_2); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":25
 * mlx5dv_qp_init_attr_mask = __Pyx_FlagBase('mlx5dv_qp_init_attr_mask', [
 *     ('MLX5DV_QP_INIT_ATTR_MASK_QP_CREATE_FLAGS', __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(MLX5DV_QP_INIT_ATTR_MASK_QP_CREATE_FLAGS)),
 *     ('MLX5DV_QP_INIT_ATTR_MASK_DC', __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(MLX5DV_QP_INIT_ATTR_MASK_DC)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_INIT_ATTR_MASK_SEND_OPS_FLAGS', __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(MLX5DV_QP_INIT_ATTR_MASK_SEND_OPS_FLAGS)),
 *     ('MLX5DV_QP_INIT_ATTR_MASK_DCI_STREAMS', __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(MLX5DV_QP_INIT_ATTR_MASK_DCI_STREAMS)),
*/
  __pyx_t_2 = __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(MLX5DV_QP_INIT_ATTR_MASK_DC); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_17 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_QP_INIT_ATTR_MASK_DC, __pyx_t_2); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":26
 *     ('MLX5DV_QP_INIT_ATTR_MASK_QP_CREATE_FLAGS', __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(MLX5DV_QP_INIT_ATTR_MASK_QP_CREATE_FLAGS)),
 *     ('MLX5DV_QP_INIT_ATTR_MASK_DC', __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(MLX5DV_QP_INIT_ATTR_MASK_DC)),
 *     ('MLX5DV_QP_INIT_ATTR_MASK_SEND_OPS_FLAGS', __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(MLX5DV_QP_INIT_ATTR_MASK_SEND_OPS_FLAGS)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_INIT_ATTR_MASK_DCI_STREAMS', __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(MLX5DV_QP_INIT_ATTR_MASK_DCI_STREAMS)),
 * 
*/
  __pyx_t_2 = __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(MLX5DV_QP_INIT_ATTR_MASK_SEND_OPS_FLAGS); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_QP_INIT_ATTR_MASK_SEND_OP, __pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":27
 *     ('MLX5DV_QP_INIT_ATTR_MASK_DC', __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(MLX5DV_QP_INIT_ATTR_MASK_DC)),
 *     ('MLX5DV_QP_INIT_ATTR_MASK_SEND_OPS_FLAGS', __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(MLX5DV_QP_INIT_ATTR_MASK_SEND_OPS_FLAGS)),
 *     ('MLX5DV_QP_INIT_ATTR_MASK_DCI_STREAMS', __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(MLX5DV_QP_INIT_ATTR_MASK_DCI_STREAMS)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_2 = __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(MLX5DV_QP_INIT_ATTR_MASK_DCI_STREAMS); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_QP_INIT_ATTR_MASK_DCI_STR, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_qp_init_attr_mask = __Pyx_FlagBase('mlx5dv_qp_init_attr_mask', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_QP_INIT_ATTR_MASK_QP_CREATE_FLAGS', __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(MLX5DV_QP_INIT_ATTR_MASK_QP_CREATE_FLAGS)),
 *     ('MLX5DV_QP_INIT_ATTR_MASK_DC', __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(MLX5DV_QP_INIT_ATTR_MASK_DC)),
*/
  __pyx_t_2 = __Pyx_PyList_Pack(4, __pyx_t_18, __pyx_t_17, __pyx_t_11, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":29
 *     ('MLX5DV_QP_INIT_ATTR_MASK_DCI_STREAMS', __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(MLX5DV_QP_INIT_ATTR_MASK_DCI_STREAMS)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_4 = __Pyx_Globals(); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_20, __pyx_mstate_global->__pyx_n_u_mlx5dv_qp_init_attr_mask, __pyx_t_2};
    __pyx_t_11 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_4, __pyx_t_11, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_21 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_19, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_11);
    __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_qp_init_attr_mask, __pyx_t_21) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":31
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":36
 * 
 * 
 *     mlx5dv_qp_init_attr_mask._member_names_ = list(mlx5dv_qp_init_attr_mask.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_mlx5dv_qp_init_attr_mask); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __pyx_t_21 = __Pyx_PySequence_ListKeepNew(__pyx_t_19); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_19, __pyx_mstate_global->__pyx_n_u_mlx5dv_qp_init_attr_mask); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_19, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_21) < (0)) __PYX_ERR(1, 36, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

    /* "EnumType":31
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_sched_elem_attr_flags = __Pyx_FlagBase('mlx5dv_sched_elem_attr_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SCHED_ELEM_ATTR_FLAGS_BW_SHARE', __Pyx_PyLong_From_enum__mlx5dv_sched_elem_attr_flags(MLX5DV_SCHED_ELEM_ATTR_FLAGS_BW_SHARE)),
 *     ('MLX5DV_SCHED_ELEM_ATTR_FLAGS_MAX_AVG_BW', __Pyx_PyLong_From_enum__mlx5dv_sched_elem_attr_flags(MLX5DV_SCHED_ELEM_ATTR_FLAGS_MAX_AVG_BW)),
*/
  __pyx_t_21 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_11 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_sched_elem_attr_flags = __Pyx_FlagBase('mlx5dv_sched_elem_attr_flags', [
 *     ('MLX5DV_SCHED_ELEM_ATTR_FLAGS_BW_SHARE', __Pyx_PyLong_From_enum__mlx5dv_sched_elem_attr_flags(MLX5DV_SCHED_ELEM_ATTR_FLAGS_BW_SHARE)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SCHED_ELEM_ATTR_FLAGS_MAX_AVG_BW', __Pyx_PyLong_From_enum__mlx5dv_sched_elem_attr_flags(MLX5DV_SCHED_ELEM_ATTR_FLAGS_MAX_AVG_BW)),
 * 
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__mlx5dv_sched_elem_attr_flags(MLX5DV_SCHED_ELEM_ATTR_FLAGS_BW_SHARE); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SCHED_ELEM_ATTR_FLAGS_BW, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":25
 * mlx5dv_sched_elem_attr_flags = __Pyx_FlagBase('mlx5dv_sched_elem_attr_flags', [
 *     ('MLX5DV_SCHED_ELEM_ATTR_FLAGS_BW_SHARE', __Pyx_PyLong_From_enum__mlx5dv_sched_elem_attr_flags(MLX5DV_SCHED_ELEM_ATTR_FLAGS_BW_SHARE)),
 *     ('MLX5DV_SCHED_ELEM_ATTR_FLAGS_MAX_AVG_BW', __Pyx_PyLong_From_enum__mlx5dv_sched_elem_attr_flags(MLX5DV_SCHED_ELEM_ATTR_FLAGS_MAX_AVG_BW)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__mlx5dv_sched_elem_attr_flags(MLX5DV_SCHED_ELEM_ATTR_FLAGS_MAX_AVG_BW); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_20 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SCHED_ELEM_ATTR_FLAGS_MAX, __pyx_t_4); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_sched_elem_attr_flags = __Pyx_FlagBase('mlx5dv_sched_elem_attr_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SCHED_ELEM_ATTR_FLAGS_BW_SHARE', __Pyx_PyLong_From_enum__mlx5dv_sched_elem_attr_flags(MLX5DV_SCHED_ELEM_ATTR_FLAGS_BW_SHARE)),
 *     ('MLX5DV_SCHED_ELEM_ATTR_FLAGS_MAX_AVG_BW', __Pyx_PyLong_From_enum__mlx5dv_sched_elem_attr_flags(MLX5DV_SCHED_ELEM_ATTR_FLAGS_MAX_AVG_BW)),
*/
  __pyx_t_4 = __Pyx_PyList_Pack(2, __pyx_t_2, __pyx_t_20); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":27
 *     ('MLX5DV_SCHED_ELEM_ATTR_FLAGS_MAX_AVG_BW', __Pyx_PyLong_From_enum__mlx5dv_sched_elem_attr_flags(MLX5DV_SCHED_ELEM_ATTR_FLAGS_MAX_AVG_BW)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_20 = __Pyx_Globals(); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_20, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_21, __pyx_mstate_global->__pyx_n_u_mlx5dv_sched_elem_attr_flags, __pyx_t_4};
    __pyx_t_2 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_20, __pyx_t_2, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_19 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_11, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_sched_elem_attr_flags, __pyx_t_19) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":34
 * 
 * 
 *     mlx5dv_sched_elem_attr_flags._member_names_ = list(mlx5dv_sched_elem_attr_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_19, __pyx_mstate_global->__pyx_n_u_mlx5dv_sched_elem_attr_flags); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_19, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    __pyx_t_19 = __Pyx_PySequence_ListKeepNew(__pyx_t_11); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_mlx5dv_sched_elem_attr_flags); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_19) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_sig_block_attr_flags = __Pyx_FlagBase('mlx5dv_sig_block_attr_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIG_BLOCK_ATTR_FLAG_COPY_MASK', __Pyx_PyLong_From_enum__mlx5dv_sig_block_attr_flags(MLX5DV_SIG_BLOCK_ATTR_FLAG_COPY_MASK)),
 * 
*/
  __pyx_t_19 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_2 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_sig_block_attr_flags = __Pyx_FlagBase('mlx5dv_sig_block_attr_flags', [
 *     ('MLX5DV_SIG_BLOCK_ATTR_FLAG_COPY_MASK', __Pyx_PyLong_From_enum__mlx5dv_sig_block_attr_flags(MLX5DV_SIG_BLOCK_ATTR_FLAG_COPY_MASK)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_20 = __Pyx_PyLong_From_enum__mlx5dv_sig_block_attr_flags(MLX5DV_SIG_BLOCK_ATTR_FLAG_COPY_MASK); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SIG_BLOCK_ATTR_FLAG_COPY, __pyx_t_20); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_sig_block_attr_flags = __Pyx_FlagBase('mlx5dv_sig_block_attr_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIG_BLOCK_ATTR_FLAG_COPY_MASK', __Pyx_PyLong_From_enum__mlx5dv_sig_block_attr_flags(MLX5DV_SIG_BLOCK_ATTR_FLAG_COPY_MASK)),
 * 
*/
  __pyx_t_20 = __Pyx_PyList_Pack(1, __pyx_t_4); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":26
 *     ('MLX5DV_SIG_BLOCK_ATTR_FLAG_COPY_MASK', __Pyx_PyLong_From_enum__mlx5dv_sig_block_attr_flags(MLX5DV_SIG_BLOCK_ATTR_FLAG_COPY_MASK)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_4 = __Pyx_Globals(); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_21, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_19, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_block_attr_flags, __pyx_t_20};
    __pyx_t_21 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_4, __pyx_t_21, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_11 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_21);
    __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_block_attr_flags, __pyx_t_11) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":33
 * 
 * 
 *     mlx5dv_sig_block_attr_flags._member_names_ = list(mlx5dv_sig_block_attr_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_block_attr_flags); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PySequence_ListKeepNew(__pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_block_attr_flags); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_11) < (0)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_sig_crc_type = __Pyx_FlagBase('mlx5dv_sig_crc_type', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIG_CRC_TYPE_CRC32', __Pyx_PyLong_From_enum__mlx5dv_sig_crc_type(MLX5DV_SIG_CRC_TYPE_CRC32)),
 *     ('MLX5DV_SIG_CRC_TYPE_CRC32C', __Pyx_PyLong_From_enum__mlx5dv_sig_crc_type(MLX5DV_SIG_CRC_TYPE_CRC32C)),
*/
  __pyx_t_11 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_21 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_sig_crc_type = __Pyx_FlagBase('mlx5dv_sig_crc_type', [
 *     ('MLX5DV_SIG_CRC_TYPE_CRC32', __Pyx_PyLong_From_enum__mlx5dv_sig_crc_type(MLX5DV_SIG_CRC_TYPE_CRC32)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIG_CRC_TYPE_CRC32C', __Pyx_PyLong_From_enum__mlx5dv_sig_crc_type(MLX5DV_SIG_CRC_TYPE_CRC32C)),
 *     ('MLX5DV_SIG_CRC_TYPE_CRC64_XP10', __Pyx_PyLong_From_enum__mlx5dv_sig_crc_type(MLX5DV_SIG_CRC_TYPE_CRC64_XP10)),
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__mlx5dv_sig_crc_type(MLX5DV_SIG_CRC_TYPE_CRC32); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_20 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SIG_CRC_TYPE_CRC32, __pyx_t_4); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":25
 * mlx5dv_sig_crc_type = __Pyx_FlagBase('mlx5dv_sig_crc_type', [
 *     ('MLX5DV_SIG_CRC_TYPE_CRC32', __Pyx_PyLong_From_enum__mlx5dv_sig_crc_type(MLX5DV_SIG_CRC_TYPE_CRC32)),
 *     ('MLX5DV_SIG_CRC_TYPE_CRC32C', __Pyx_PyLong_From_enum__mlx5dv_sig_crc_type(MLX5DV_SIG_CRC_TYPE_CRC32C)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIG_CRC_TYPE_CRC64_XP10', __Pyx_PyLong_From_enum__mlx5dv_sig_crc_type(MLX5DV_SIG_CRC_TYPE_CRC64_XP10)),
 * 
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__mlx5dv_sig_crc_type(MLX5DV_SIG_CRC_TYPE_CRC32C); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_19 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SIG_CRC_TYPE_CRC32C, __pyx_t_4); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":26
 *     ('MLX5DV_SIG_CRC_TYPE_CRC32', __Pyx_PyLong_From_enum__mlx5dv_sig_crc_type(MLX5DV_SIG_CRC_TYPE_CRC32)),
 *     ('MLX5DV_SIG_CRC_TYPE_CRC32C', __Pyx_PyLong_From_enum__mlx5dv_sig_crc_type(MLX5DV_SIG_CRC_TYPE_CRC32C)),
 *     ('MLX5DV_SIG_CRC_TYPE_CRC64_XP10', __Pyx_PyLong_From_enum__mlx5dv_sig_crc_type(MLX5DV_SIG_CRC_TYPE_CRC64_XP10)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_4 = __Pyx_PyLong_From_enum__mlx5dv_sig_crc_type(MLX5DV_SIG_CRC_TYPE_CRC64_XP10); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_17 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SIG_CRC_TYPE_CRC64_XP10, __pyx_t_4); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_sig_crc_type = __Pyx_FlagBase('mlx5dv_sig_crc_type', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIG_CRC_TYPE_CRC32', __Pyx_PyLong_From_enum__mlx5dv_sig_crc_type(MLX5DV_SIG_CRC_TYPE_CRC32)),
 *     ('MLX5DV_SIG_CRC_TYPE_CRC32C', __Pyx_PyLong_From_enum__mlx5dv_sig_crc_type(MLX5DV_SIG_CRC_TYPE_CRC32C)),
*/
  __pyx_t_4 = __Pyx_PyList_Pack(3, __pyx_t_20, __pyx_t_19, __pyx_t_17); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "EnumType":28
 *     ('MLX5DV_SIG_CRC_TYPE_CRC64_XP10', __Pyx_PyLong_From_enum__mlx5dv_sig_crc_type(MLX5DV_SIG_CRC_TYPE_CRC64_XP10)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_17 = __Pyx_Globals(); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_17 = __Pyx_PyObject_Call(__pyx_t_19, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_11, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_crc_type, __pyx_t_4};
    __pyx_t_19 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_17, __pyx_t_19, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_21, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_19);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_crc_type, __pyx_t_2) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":35
 * 
 * 
 *     mlx5dv_sig_crc_type._member_names_ = list(mlx5dv_sig_crc_type.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_crc_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PySequence_ListKeepNew(__pyx_t_21); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_crc_type); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_2) < (0)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

    /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_sig_mask = __Pyx_FlagBase('mlx5dv_sig_mask', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIG_MASK_T10DIF_GUARD', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_T10DIF_GUARD)),
 *     ('MLX5DV_SIG_MASK_T10DIF_APPTAG', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_T10DIF_APPTAG)),
*/
  __pyx_t_2 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_19 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_sig_mask = __Pyx_FlagBase('mlx5dv_sig_mask', [
 *     ('MLX5DV_SIG_MASK_T10DIF_GUARD', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_T10DIF_GUARD)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIG_MASK_T10DIF_APPTAG', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_T10DIF_APPTAG)),
 *     ('MLX5DV_SIG_MASK_T10DIF_REFTAG', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_T10DIF_REFTAG)),
*/
  __pyx_t_17 = __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_T10DIF_GUARD); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_4 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SIG_MASK_T10DIF_GUARD, __pyx_t_17); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "EnumType":25
 * mlx5dv_sig_mask = __Pyx_FlagBase('mlx5dv_sig_mask', [
 *     ('MLX5DV_SIG_MASK_T10DIF_GUARD', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_T10DIF_GUARD)),
 *     ('MLX5DV_SIG_MASK_T10DIF_APPTAG', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_T10DIF_APPTAG)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIG_MASK_T10DIF_REFTAG', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_T10DIF_REFTAG)),
 *     ('MLX5DV_SIG_MASK_CRC32', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_CRC32)),
*/
  __pyx_t_17 = __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_T10DIF_APPTAG); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SIG_MASK_T10DIF_APPTAG, __pyx_t_17); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "EnumType":26
 *     ('MLX5DV_SIG_MASK_T10DIF_GUARD', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_T10DIF_GUARD)),
 *     ('MLX5DV_SIG_MASK_T10DIF_APPTAG', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_T10DIF_APPTAG)),
 *     ('MLX5DV_SIG_MASK_T10DIF_REFTAG', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_T10DIF_REFTAG)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIG_MASK_CRC32', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_CRC32)),
 *     ('MLX5DV_SIG_MASK_CRC32C', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_CRC32C)),
*/
  __pyx_t_17 = __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_T10DIF_REFTAG); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_20 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SIG_MASK_T10DIF_REFTAG, __pyx_t_17); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "EnumType":27
 *     ('MLX5DV_SIG_MASK_T10DIF_APPTAG', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_T10DIF_APPTAG)),
 *     ('MLX5DV_SIG_MASK_T10DIF_REFTAG', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_T10DIF_REFTAG)),
 *     ('MLX5DV_SIG_MASK_CRC32', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_CRC32)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIG_MASK_CRC32C', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_CRC32C)),
 *     ('MLX5DV_SIG_MASK_CRC64_XP10', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_CRC64_XP10)),
*/
  __pyx_t_17 = __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_CRC32); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_18 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SIG_MASK_CRC32, __pyx_t_17); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "EnumType":28
 *     ('MLX5DV_SIG_MASK_T10DIF_REFTAG', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_T10DIF_REFTAG)),
 *     ('MLX5DV_SIG_MASK_CRC32', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_CRC32)),
 *     ('MLX5DV_SIG_MASK_CRC32C', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_CRC32C)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIG_MASK_CRC64_XP10', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_CRC64_XP10)),
 * 
*/
  __pyx_t_17 = __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_CRC32C); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SIG_MASK_CRC32C, __pyx_t_17); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "EnumType":29
 *     ('MLX5DV_SIG_MASK_CRC32', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_CRC32)),
 *     ('MLX5DV_SIG_MASK_CRC32C', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_CRC32C)),
 *     ('MLX5DV_SIG_MASK_CRC64_XP10', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_CRC64_XP10)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_17 = __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_CRC64_XP10); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SIG_MASK_CRC64_XP10, __pyx_t_17); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_sig_mask = __Pyx_FlagBase('mlx5dv_sig_mask', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIG_MASK_T10DIF_GUARD', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_T10DIF_GUARD)),
 *     ('MLX5DV_SIG_MASK_T10DIF_APPTAG', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_T10DIF_APPTAG)),
*/
  __pyx_t_17 = __Pyx_PyList_Pack(6, __pyx_t_4, __pyx_t_11, __pyx_t_20, __pyx_t_18, __pyx_t_22, __pyx_t_15); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":31
 *     ('MLX5DV_SIG_MASK_CRC64_XP10', __Pyx_PyLong_From_enum__mlx5dv_sig_mask(MLX5DV_SIG_MASK_CRC64_XP10)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_15 = __Pyx_Globals(); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_22 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_22, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_mask, __pyx_t_17};
    __pyx_t_22 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_15, __pyx_t_22, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_21 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_19, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_22);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_mask, __pyx_t_21) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":33
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":38
 * 
 * 
 *     mlx5dv_sig_mask._member_names_ = list(mlx5dv_sig_mask.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_mask); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __pyx_t_21 = __Pyx_PySequence_ListKeepNew(__pyx_t_19); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_19, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_mask); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 38, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_19, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_21) < (0)) __PYX_ERR(1, 38, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

    /* "EnumType":33
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_sig_t10dif_bg_type = __Pyx_FlagBase('mlx5dv_sig_t10dif_bg_type', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIG_T10DIF_CRC', __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_bg_type(MLX5DV_SIG_T10DIF_CRC)),
 *     ('MLX5DV_SIG_T10DIF_CSUM', __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_bg_type(MLX5DV_SIG_T10DIF_CSUM)),
*/
  __pyx_t_21 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_22 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_sig_t10dif_bg_type = __Pyx_FlagBase('mlx5dv_sig_t10dif_bg_type', [
 *     ('MLX5DV_SIG_T10DIF_CRC', __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_bg_type(MLX5DV_SIG_T10DIF_CRC)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIG_T10DIF_CSUM', __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_bg_type(MLX5DV_SIG_T10DIF_CSUM)),
 * 
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_bg_type(MLX5DV_SIG_T10DIF_CRC); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_17 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SIG_T10DIF_CRC, __pyx_t_15); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":25
 * mlx5dv_sig_t10dif_bg_type = __Pyx_FlagBase('mlx5dv_sig_t10dif_bg_type', [
 *     ('MLX5DV_SIG_T10DIF_CRC', __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_bg_type(MLX5DV_SIG_T10DIF_CRC)),
 *     ('MLX5DV_SIG_T10DIF_CSUM', __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_bg_type(MLX5DV_SIG_T10DIF_CSUM)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_bg_type(MLX5DV_SIG_T10DIF_CSUM); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_2 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SIG_T10DIF_CSUM, __pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_sig_t10dif_bg_type = __Pyx_FlagBase('mlx5dv_sig_t10dif_bg_type', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIG_T10DIF_CRC', __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_bg_type(MLX5DV_SIG_T10DIF_CRC)),
 *     ('MLX5DV_SIG_T10DIF_CSUM', __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_bg_type(MLX5DV_SIG_T10DIF_CSUM)),
*/
  __pyx_t_15 = __Pyx_PyList_Pack(2, __pyx_t_17, __pyx_t_2); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":27
 *     ('MLX5DV_SIG_T10DIF_CSUM', __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_bg_type(MLX5DV_SIG_T10DIF_CSUM)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_2 = __Pyx_Globals(); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_21, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_t10dif_bg_type, __pyx_t_15};
    __pyx_t_17 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_2, __pyx_t_17, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_19 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_22, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_17);
    __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_t10dif_bg_type, __pyx_t_19) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":34
 * 
 * 
 *     mlx5dv_sig_t10dif_bg_type._member_names_ = list(mlx5dv_sig_t10dif_bg_type.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_19, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_t10dif_bg_type); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    __pyx_t_22 = __Pyx_PyObject_GetAttrStr(__pyx_t_19, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    __pyx_t_19 = __Pyx_PySequence_ListKeepNew(__pyx_t_22); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_t10dif_bg_type); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_19) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_sig_t10dif_flags = __Pyx_FlagBase('mlx5dv_sig_t10dif_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIG_T10DIF_FLAG_REF_REMAP', __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_flags(MLX5DV_SIG_T10DIF_FLAG_REF_REMAP)),
 *     ('MLX5DV_SIG_T10DIF_FLAG_APP_ESCAPE', __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_flags(MLX5DV_SIG_T10DIF_FLAG_APP_ESCAPE)),
*/
  __pyx_t_19 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_17 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_sig_t10dif_flags = __Pyx_FlagBase('mlx5dv_sig_t10dif_flags', [
 *     ('MLX5DV_SIG_T10DIF_FLAG_REF_REMAP', __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_flags(MLX5DV_SIG_T10DIF_FLAG_REF_REMAP)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIG_T10DIF_FLAG_APP_ESCAPE', __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_flags(MLX5DV_SIG_T10DIF_FLAG_APP_ESCAPE)),
 *     ('MLX5DV_SIG_T10DIF_FLAG_APP_REF_ESCAPE', __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_flags(MLX5DV_SIG_T10DIF_FLAG_APP_REF_ESCAPE)),
*/
  __pyx_t_2 = __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_flags(MLX5DV_SIG_T10DIF_FLAG_REF_REMAP); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SIG_T10DIF_FLAG_REF_REMAP, __pyx_t_2); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":25
 * mlx5dv_sig_t10dif_flags = __Pyx_FlagBase('mlx5dv_sig_t10dif_flags', [
 *     ('MLX5DV_SIG_T10DIF_FLAG_REF_REMAP', __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_flags(MLX5DV_SIG_T10DIF_FLAG_REF_REMAP)),
 *     ('MLX5DV_SIG_T10DIF_FLAG_APP_ESCAPE', __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_flags(MLX5DV_SIG_T10DIF_FLAG_APP_ESCAPE)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIG_T10DIF_FLAG_APP_REF_ESCAPE', __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_flags(MLX5DV_SIG_T10DIF_FLAG_APP_REF_ESCAPE)),
 * 
*/
  __pyx_t_2 = __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_flags(MLX5DV_SIG_T10DIF_FLAG_APP_ESCAPE); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_21 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SIG_T10DIF_FLAG_APP_ESCAP, __pyx_t_2); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":26
 *     ('MLX5DV_SIG_T10DIF_FLAG_REF_REMAP', __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_flags(MLX5DV_SIG_T10DIF_FLAG_REF_REMAP)),
 *     ('MLX5DV_SIG_T10DIF_FLAG_APP_ESCAPE', __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_flags(MLX5DV_SIG_T10DIF_FLAG_APP_ESCAPE)),
 *     ('MLX5DV_SIG_T10DIF_FLAG_APP_REF_ESCAPE', __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_flags(MLX5DV_SIG_T10DIF_FLAG_APP_REF_ESCAPE)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_2 = __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_flags(MLX5DV_SIG_T10DIF_FLAG_APP_REF_ESCAPE); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_18 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SIG_T10DIF_FLAG_APP_REF_E, __pyx_t_2); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_sig_t10dif_flags = __Pyx_FlagBase('mlx5dv_sig_t10dif_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIG_T10DIF_FLAG_REF_REMAP', __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_flags(MLX5DV_SIG_T10DIF_FLAG_REF_REMAP)),
 *     ('MLX5DV_SIG_T10DIF_FLAG_APP_ESCAPE', __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_flags(MLX5DV_SIG_T10DIF_FLAG_APP_ESCAPE)),
*/
  __pyx_t_2 = __Pyx_PyList_Pack(3, __pyx_t_15, __pyx_t_21, __pyx_t_18); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "EnumType":28
 *     ('MLX5DV_SIG_T10DIF_FLAG_APP_REF_ESCAPE', __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_flags(MLX5DV_SIG_T10DIF_FLAG_APP_REF_ESCAPE)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_18 = __Pyx_Globals(); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __pyx_t_18 = __Pyx_PyObject_Call(__pyx_t_21, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_19, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_t10dif_flags, __pyx_t_2};
    __pyx_t_21 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_18, __pyx_t_21, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_22 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_17, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_21);
    __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_t10dif_flags, __pyx_t_22) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":35
 * 
 * 
 *     mlx5dv_sig_t10dif_flags._member_names_ = list(mlx5dv_sig_t10dif_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_t10dif_flags); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __pyx_t_22 = __Pyx_PySequence_ListKeepNew(__pyx_t_17); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_t10dif_flags); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_22) < (0)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

    /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_sig_type = __Pyx_FlagBase('mlx5dv_sig_type', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIG_TYPE_T10DIF', __Pyx_PyLong_From_enum__mlx5dv_sig_type(MLX5DV_SIG_TYPE_T10DIF)),
 *     ('MLX5DV_SIG_TYPE_CRC', __Pyx_PyLong_From_enum__mlx5dv_sig_type(MLX5DV_SIG_TYPE_CRC)),
*/
  __pyx_t_22 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_21 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_sig_type = __Pyx_FlagBase('mlx5dv_sig_type', [
 *     ('MLX5DV_SIG_TYPE_T10DIF', __Pyx_PyLong_From_enum__mlx5dv_sig_type(MLX5DV_SIG_TYPE_T10DIF)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIG_TYPE_CRC', __Pyx_PyLong_From_enum__mlx5dv_sig_type(MLX5DV_SIG_TYPE_CRC)),
 * 
*/
  __pyx_t_18 = __Pyx_PyLong_From_enum__mlx5dv_sig_type(MLX5DV_SIG_TYPE_T10DIF); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_2 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SIG_TYPE_T10DIF, __pyx_t_18); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "EnumType":25
 * mlx5dv_sig_type = __Pyx_FlagBase('mlx5dv_sig_type', [
 *     ('MLX5DV_SIG_TYPE_T10DIF', __Pyx_PyLong_From_enum__mlx5dv_sig_type(MLX5DV_SIG_TYPE_T10DIF)),
 *     ('MLX5DV_SIG_TYPE_CRC', __Pyx_PyLong_From_enum__mlx5dv_sig_type(MLX5DV_SIG_TYPE_CRC)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_18 = __Pyx_PyLong_From_enum__mlx5dv_sig_type(MLX5DV_SIG_TYPE_CRC); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_19 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SIG_TYPE_CRC, __pyx_t_18); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_sig_type = __Pyx_FlagBase('mlx5dv_sig_type', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIG_TYPE_T10DIF', __Pyx_PyLong_From_enum__mlx5dv_sig_type(MLX5DV_SIG_TYPE_T10DIF)),
 *     ('MLX5DV_SIG_TYPE_CRC', __Pyx_PyLong_From_enum__mlx5dv_sig_type(MLX5DV_SIG_TYPE_CRC)),
*/
  __pyx_t_18 = __Pyx_PyList_Pack(2, __pyx_t_2, __pyx_t_19); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

  /* "EnumType":27
 *     ('MLX5DV_SIG_TYPE_CRC', __Pyx_PyLong_From_enum__mlx5dv_sig_type(MLX5DV_SIG_TYPE_CRC)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_19 = __Pyx_Globals(); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_19, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __pyx_t_19 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_22, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_type, __pyx_t_18};
    __pyx_t_2 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_19, __pyx_t_2, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_17 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_21, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_2);
    __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_type, __pyx_t_17) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":34
 * 
 * 
 *     mlx5dv_sig_type._member_names_ = list(mlx5dv_sig_type.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_type); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __pyx_t_17 = __Pyx_PySequence_ListKeepNew(__pyx_t_21); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_mlx5dv_sig_type); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_17) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_signature_crypto_order = __Pyx_FlagBase('mlx5dv_signature_crypto_order', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIGNATURE_CRYPTO_ORDER_SIGNATURE_AFTER_CRYPTO_ON_TX', __Pyx_PyLong_From_enum__mlx5dv_signature_crypto_order(MLX5DV_SIGNATURE_CRYPTO_ORDER_SIGNATURE_AFTER_CRYPTO_ON_TX)),
 *     ('MLX5DV_SIGNATURE_CRYPTO_ORDER_SIGNATURE_BEFORE_CRYPTO_ON_TX', __Pyx_PyLong_From_enum__mlx5dv_signature_crypto_order(MLX5DV_SIGNATURE_CRYPTO_ORDER_SIGNATURE_BEFORE_CRYPTO_ON_TX)),
*/
  __pyx_t_17 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_2 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_signature_crypto_order = __Pyx_FlagBase('mlx5dv_signature_crypto_order', [
 *     ('MLX5DV_SIGNATURE_CRYPTO_ORDER_SIGNATURE_AFTER_CRYPTO_ON_TX', __Pyx_PyLong_From_enum__mlx5dv_signature_crypto_order(MLX5DV_SIGNATURE_CRYPTO_ORDER_SIGNATURE_AFTER_CRYPTO_ON_TX)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIGNATURE_CRYPTO_ORDER_SIGNATURE_BEFORE_CRYPTO_ON_TX', __Pyx_PyLong_From_enum__mlx5dv_signature_crypto_order(MLX5DV_SIGNATURE_CRYPTO_ORDER_SIGNATURE_BEFORE_CRYPTO_ON_TX)),
 * 
*/
  __pyx_t_19 = __Pyx_PyLong_From_enum__mlx5dv_signature_crypto_order(MLX5DV_SIGNATURE_CRYPTO_ORDER_SIGNATURE_AFTER_CRYPTO_ON_TX); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_18 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SIGNATURE_CRYPTO_ORDER_SI, __pyx_t_19); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

  /* "EnumType":25
 * mlx5dv_signature_crypto_order = __Pyx_FlagBase('mlx5dv_signature_crypto_order', [
 *     ('MLX5DV_SIGNATURE_CRYPTO_ORDER_SIGNATURE_AFTER_CRYPTO_ON_TX', __Pyx_PyLong_From_enum__mlx5dv_signature_crypto_order(MLX5DV_SIGNATURE_CRYPTO_ORDER_SIGNATURE_AFTER_CRYPTO_ON_TX)),
 *     ('MLX5DV_SIGNATURE_CRYPTO_ORDER_SIGNATURE_BEFORE_CRYPTO_ON_TX', __Pyx_PyLong_From_enum__mlx5dv_signature_crypto_order(MLX5DV_SIGNATURE_CRYPTO_ORDER_SIGNATURE_BEFORE_CRYPTO_ON_TX)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_19 = __Pyx_PyLong_From_enum__mlx5dv_signature_crypto_order(MLX5DV_SIGNATURE_CRYPTO_ORDER_SIGNATURE_BEFORE_CRYPTO_ON_TX); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SIGNATURE_CRYPTO_ORDER_SI_2, __pyx_t_19); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_signature_crypto_order = __Pyx_FlagBase('mlx5dv_signature_crypto_order', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SIGNATURE_CRYPTO_ORDER_SIGNATURE_AFTER_CRYPTO_ON_TX', __Pyx_PyLong_From_enum__mlx5dv_signature_crypto_order(MLX5DV_SIGNATURE_CRYPTO_ORDER_SIGNATURE_AFTER_CRYPTO_ON_TX)),
 *     ('MLX5DV_SIGNATURE_CRYPTO_ORDER_SIGNATURE_BEFORE_CRYPTO_ON_TX', __Pyx_PyLong_From_enum__mlx5dv_signature_crypto_order(MLX5DV_SIGNATURE_CRYPTO_ORDER_SIGNATURE_BEFORE_CRYPTO_ON_TX)),
*/
  __pyx_t_19 = __Pyx_PyList_Pack(2, __pyx_t_18, __pyx_t_22); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":27
 *     ('MLX5DV_SIGNATURE_CRYPTO_ORDER_SIGNATURE_BEFORE_CRYPTO_ON_TX', __Pyx_PyLong_From_enum__mlx5dv_signature_crypto_order(MLX5DV_SIGNATURE_CRYPTO_ORDER_SIGNATURE_BEFORE_CRYPTO_ON_TX)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_22 = __Pyx_Globals(); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_18, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_signature_crypto_order, __pyx_t_19};
    __pyx_t_18 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_22, __pyx_t_18, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_21 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_18);
    __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_signature_crypto_order, __pyx_t_21) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":34
 * 
 * 
 *     mlx5dv_signature_crypto_order._member_names_ = list(mlx5dv_signature_crypto_order.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_mlx5dv_signature_crypto_order); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __pyx_t_21 = __Pyx_PySequence_ListKeepNew(__pyx_t_2); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_signature_crypto_order); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_21) < (0)) __PYX_ERR(1, 34, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":29
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_srq_comp_mask = __Pyx_FlagBase('mlx5dv_srq_comp_mask', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SRQ_MASK_SRQN', __Pyx_PyLong_From_enum__mlx5dv_srq_comp_mask(MLX5DV_SRQ_MASK_SRQN)),
 * 
*/
  __pyx_t_21 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_18 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_srq_comp_mask = __Pyx_FlagBase('mlx5dv_srq_comp_mask', [
 *     ('MLX5DV_SRQ_MASK_SRQN', __Pyx_PyLong_From_enum__mlx5dv_srq_comp_mask(MLX5DV_SRQ_MASK_SRQN)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_22 = __Pyx_PyLong_From_enum__mlx5dv_srq_comp_mask(MLX5DV_SRQ_MASK_SRQN); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_19 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SRQ_MASK_SRQN, __pyx_t_22); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_srq_comp_mask = __Pyx_FlagBase('mlx5dv_srq_comp_mask', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SRQ_MASK_SRQN', __Pyx_PyLong_From_enum__mlx5dv_srq_comp_mask(MLX5DV_SRQ_MASK_SRQN)),
 * 
*/
  __pyx_t_22 = __Pyx_PyList_Pack(1, __pyx_t_19); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

  /* "EnumType":26
 *     ('MLX5DV_SRQ_MASK_SRQN', __Pyx_PyLong_From_enum__mlx5dv_srq_comp_mask(MLX5DV_SRQ_MASK_SRQN)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_19 = __Pyx_Globals(); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_19, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __pyx_t_19 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_21, __pyx_mstate_global->__pyx_n_u_mlx5dv_srq_comp_mask, __pyx_t_22};
    __pyx_t_17 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_19, __pyx_t_17, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_2 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_18, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_17);
    __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_srq_comp_mask, __pyx_t_2) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":33
 * 
 * 
 *     mlx5dv_srq_comp_mask._member_names_ = list(mlx5dv_srq_comp_mask.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_srq_comp_mask); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PySequence_ListKeepNew(__pyx_t_18); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_mlx5dv_srq_comp_mask); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_2) < (0)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

    /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_sw_parsing_offloads = __Pyx_FlagBase('mlx5dv_sw_parsing_offloads', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SW_PARSING', __Pyx_PyLong_From_enum__mlx5dv_sw_parsing_offloads(MLX5DV_SW_PARSING)),
 *     ('MLX5DV_SW_PARSING_CSUM', __Pyx_PyLong_From_enum__mlx5dv_sw_parsing_offloads(MLX5DV_SW_PARSING_CSUM)),
*/
  __pyx_t_2 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_17 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_sw_parsing_offloads = __Pyx_FlagBase('mlx5dv_sw_parsing_offloads', [
 *     ('MLX5DV_SW_PARSING', __Pyx_PyLong_From_enum__mlx5dv_sw_parsing_offloads(MLX5DV_SW_PARSING)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SW_PARSING_CSUM', __Pyx_PyLong_From_enum__mlx5dv_sw_parsing_offloads(MLX5DV_SW_PARSING_CSUM)),
 *     ('MLX5DV_SW_PARSING_LSO', __Pyx_PyLong_From_enum__mlx5dv_sw_parsing_offloads(MLX5DV_SW_PARSING_LSO)),
*/
  __pyx_t_19 = __Pyx_PyLong_From_enum__mlx5dv_sw_parsing_offloads(MLX5DV_SW_PARSING); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SW_PARSING, __pyx_t_19); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

  /* "EnumType":25
 * mlx5dv_sw_parsing_offloads = __Pyx_FlagBase('mlx5dv_sw_parsing_offloads', [
 *     ('MLX5DV_SW_PARSING', __Pyx_PyLong_From_enum__mlx5dv_sw_parsing_offloads(MLX5DV_SW_PARSING)),
 *     ('MLX5DV_SW_PARSING_CSUM', __Pyx_PyLong_From_enum__mlx5dv_sw_parsing_offloads(MLX5DV_SW_PARSING_CSUM)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SW_PARSING_LSO', __Pyx_PyLong_From_enum__mlx5dv_sw_parsing_offloads(MLX5DV_SW_PARSING_LSO)),
 * 
*/
  __pyx_t_19 = __Pyx_PyLong_From_enum__mlx5dv_sw_parsing_offloads(MLX5DV_SW_PARSING_CSUM); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_21 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SW_PARSING_CSUM, __pyx_t_19); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

  /* "EnumType":26
 *     ('MLX5DV_SW_PARSING', __Pyx_PyLong_From_enum__mlx5dv_sw_parsing_offloads(MLX5DV_SW_PARSING)),
 *     ('MLX5DV_SW_PARSING_CSUM', __Pyx_PyLong_From_enum__mlx5dv_sw_parsing_offloads(MLX5DV_SW_PARSING_CSUM)),
 *     ('MLX5DV_SW_PARSING_LSO', __Pyx_PyLong_From_enum__mlx5dv_sw_parsing_offloads(MLX5DV_SW_PARSING_LSO)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_19 = __Pyx_PyLong_From_enum__mlx5dv_sw_parsing_offloads(MLX5DV_SW_PARSING_LSO); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_SW_PARSING_LSO, __pyx_t_19); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_sw_parsing_offloads = __Pyx_FlagBase('mlx5dv_sw_parsing_offloads', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_SW_PARSING', __Pyx_PyLong_From_enum__mlx5dv_sw_parsing_offloads(MLX5DV_SW_PARSING)),
 *     ('MLX5DV_SW_PARSING_CSUM', __Pyx_PyLong_From_enum__mlx5dv_sw_parsing_offloads(MLX5DV_SW_PARSING_CSUM)),
*/
  __pyx_t_19 = __Pyx_PyList_Pack(3, __pyx_t_22, __pyx_t_21, __pyx_t_15); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":28
 *     ('MLX5DV_SW_PARSING_LSO', __Pyx_PyLong_From_enum__mlx5dv_sw_parsing_offloads(MLX5DV_SW_PARSING_LSO)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_15 = __Pyx_Globals(); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_21, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mlx5dv_sw_parsing_offloads, __pyx_t_19};
    __pyx_t_21 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_15, __pyx_t_21, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_18 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_17, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_21);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_sw_parsing_offloads, __pyx_t_18) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":35
 * 
 * 
 *     mlx5dv_sw_parsing_offloads._member_names_ = list(mlx5dv_sw_parsing_offloads.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_mlx5dv_sw_parsing_offloads); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __pyx_t_18 = __Pyx_PySequence_ListKeepNew(__pyx_t_17); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_sw_parsing_offloads); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_18) < (0)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

    /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_tunnel_offloads = __Pyx_FlagBase('mlx5dv_tunnel_offloads', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_VXLAN', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_VXLAN)),
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GRE', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GRE)),
*/
  __pyx_t_18 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_21 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_tunnel_offloads = __Pyx_FlagBase('mlx5dv_tunnel_offloads', [
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_VXLAN', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_VXLAN)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GRE', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GRE)),
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GENEVE', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GENEVE)),
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_VXLAN); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_19 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_RAW_PACKET_CAP_TUNNELED_O, __pyx_t_15); if (unlikely(!__pyx_t_19)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":25
 * mlx5dv_tunnel_offloads = __Pyx_FlagBase('mlx5dv_tunnel_offloads', [
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_VXLAN', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_VXLAN)),
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GRE', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GRE)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GENEVE', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GENEVE)),
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_CW_MPLS_OVER_GRE', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_CW_MPLS_OVER_GRE)),
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GRE); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_2 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_RAW_PACKET_CAP_TUNNELED_O_2, __pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":26
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_VXLAN', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_VXLAN)),
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GRE', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GRE)),
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GENEVE', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GENEVE)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_CW_MPLS_OVER_GRE', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_CW_MPLS_OVER_GRE)),
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_CW_MPLS_OVER_UDP', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_CW_MPLS_OVER_UDP)),
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GENEVE); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_RAW_PACKET_CAP_TUNNELED_O_3, __pyx_t_15); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":27
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GRE', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GRE)),
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GENEVE', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GENEVE)),
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_CW_MPLS_OVER_GRE', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_CW_MPLS_OVER_GRE)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_CW_MPLS_OVER_UDP', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_CW_MPLS_OVER_UDP)),
 * 
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_CW_MPLS_OVER_GRE); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_20 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_RAW_PACKET_CAP_TUNNELED_O_4, __pyx_t_15); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":28
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GENEVE', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GENEVE)),
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_CW_MPLS_OVER_GRE', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_CW_MPLS_OVER_GRE)),
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_CW_MPLS_OVER_UDP', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_CW_MPLS_OVER_UDP)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_CW_MPLS_OVER_UDP); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_RAW_PACKET_CAP_TUNNELED_O_5, __pyx_t_15); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_tunnel_offloads = __Pyx_FlagBase('mlx5dv_tunnel_offloads', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_VXLAN', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_VXLAN)),
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GRE', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GRE)),
*/
  __pyx_t_15 = __Pyx_PyList_Pack(5, __pyx_t_19, __pyx_t_2, __pyx_t_22, __pyx_t_20, __pyx_t_11); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":30
 *     ('MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_CW_MPLS_OVER_UDP', __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_CW_MPLS_OVER_UDP)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_11 = __Pyx_Globals(); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_20, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_18, __pyx_mstate_global->__pyx_n_u_mlx5dv_tunnel_offloads, __pyx_t_15};
    __pyx_t_20 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_11, __pyx_t_20, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_17 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_21, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_20);
    __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_tunnel_offloads, __pyx_t_17) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

  /* "EnumType":32
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":37
 * 
 * 
 *     mlx5dv_tunnel_offloads._member_names_ = list(mlx5dv_tunnel_offloads.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_tunnel_offloads); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __pyx_t_17 = __Pyx_PySequence_ListKeepNew(__pyx_t_21); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_mlx5dv_tunnel_offloads); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_17) < (0)) __PYX_ERR(1, 37, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

    /* "EnumType":32
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_vfio_context_attr_flags = __Pyx_FlagBase('mlx5dv_vfio_context_attr_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_VFIO_CTX_FLAGS_INIT_LINK_DOWN', __Pyx_PyLong_From_enum__mlx5dv_vfio_context_attr_flags(MLX5DV_VFIO_CTX_FLAGS_INIT_LINK_DOWN)),
 * 
*/
  __pyx_t_17 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_20 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_vfio_context_attr_flags = __Pyx_FlagBase('mlx5dv_vfio_context_attr_flags', [
 *     ('MLX5DV_VFIO_CTX_FLAGS_INIT_LINK_DOWN', __Pyx_PyLong_From_enum__mlx5dv_vfio_context_attr_flags(MLX5DV_VFIO_CTX_FLAGS_INIT_LINK_DOWN)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_11 = __Pyx_PyLong_From_enum__mlx5dv_vfio_context_attr_flags(MLX5DV_VFIO_CTX_FLAGS_INIT_LINK_DOWN); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_15 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_VFIO_CTX_FLAGS_INIT_LINK, __pyx_t_11); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_vfio_context_attr_flags = __Pyx_FlagBase('mlx5dv_vfio_context_attr_flags', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_VFIO_CTX_FLAGS_INIT_LINK_DOWN', __Pyx_PyLong_From_enum__mlx5dv_vfio_context_attr_flags(MLX5DV_VFIO_CTX_FLAGS_INIT_LINK_DOWN)),
 * 
*/
  __pyx_t_11 = __Pyx_PyList_Pack(1, __pyx_t_15); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":26
 *     ('MLX5DV_VFIO_CTX_FLAGS_INIT_LINK_DOWN', __Pyx_PyLong_From_enum__mlx5dv_vfio_context_attr_flags(MLX5DV_VFIO_CTX_FLAGS_INIT_LINK_DOWN)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_15 = __Pyx_Globals(); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_18, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_17, __pyx_mstate_global->__pyx_n_u_mlx5dv_vfio_context_attr_flags, __pyx_t_11};
    __pyx_t_18 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_15, __pyx_t_18, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_21 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_20, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_18);
    __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_vfio_context_attr_flags, __pyx_t_21) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

  /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":33
 * 
 * 
 *     mlx5dv_vfio_context_attr_flags._member_names_ = list(mlx5dv_vfio_context_attr_flags.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_mlx5dv_vfio_context_attr_flags); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_t_21, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __pyx_t_21 = __Pyx_PySequence_ListKeepNew(__pyx_t_20); if (unlikely(!__pyx_t_21)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_20, __pyx_mstate_global->__pyx_n_u_mlx5dv_vfio_context_attr_flags); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_20, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_21) < (0)) __PYX_ERR(1, 33, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

    /* "EnumType":28
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "EnumType":23
 * 
 * 
 * mlx5dv_wc_opcode = __Pyx_FlagBase('mlx5dv_wc_opcode', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_WC_UMR', __Pyx_PyLong_From_enum__mlx5dv_wc_opcode(MLX5DV_WC_UMR)),
 *     ('MLX5DV_WC_RAW_WQE', __Pyx_PyLong_From_enum__mlx5dv_wc_opcode(MLX5DV_WC_RAW_WQE)),
*/
  __pyx_t_21 = NULL;
  __Pyx_INCREF(__Pyx_FlagBase);
  __pyx_t_18 = __Pyx_FlagBase; 

  /* "EnumType":24
 * 
 * mlx5dv_wc_opcode = __Pyx_FlagBase('mlx5dv_wc_opcode', [
 *     ('MLX5DV_WC_UMR', __Pyx_PyLong_From_enum__mlx5dv_wc_opcode(MLX5DV_WC_UMR)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_WC_RAW_WQE', __Pyx_PyLong_From_enum__mlx5dv_wc_opcode(MLX5DV_WC_RAW_WQE)),
 *     ('MLX5DV_WC_MEMCPY', __Pyx_PyLong_From_enum__mlx5dv_wc_opcode(MLX5DV_WC_MEMCPY)),
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__mlx5dv_wc_opcode(MLX5DV_WC_UMR); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_11 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_WC_UMR, __pyx_t_15); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":25
 * mlx5dv_wc_opcode = __Pyx_FlagBase('mlx5dv_wc_opcode', [
 *     ('MLX5DV_WC_UMR', __Pyx_PyLong_From_enum__mlx5dv_wc_opcode(MLX5DV_WC_UMR)),
 *     ('MLX5DV_WC_RAW_WQE', __Pyx_PyLong_From_enum__mlx5dv_wc_opcode(MLX5DV_WC_RAW_WQE)),             # <<<<<<<<<<<<<<
 *     ('MLX5DV_WC_MEMCPY', __Pyx_PyLong_From_enum__mlx5dv_wc_opcode(MLX5DV_WC_MEMCPY)),
 * 
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__mlx5dv_wc_opcode(MLX5DV_WC_RAW_WQE); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_17 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_WC_RAW_WQE, __pyx_t_15); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":26
 *     ('MLX5DV_WC_UMR', __Pyx_PyLong_From_enum__mlx5dv_wc_opcode(MLX5DV_WC_UMR)),
 *     ('MLX5DV_WC_RAW_WQE', __Pyx_PyLong_From_enum__mlx5dv_wc_opcode(MLX5DV_WC_RAW_WQE)),
 *     ('MLX5DV_WC_MEMCPY', __Pyx_PyLong_From_enum__mlx5dv_wc_opcode(MLX5DV_WC_MEMCPY)),             # <<<<<<<<<<<<<<
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
*/
  __pyx_t_15 = __Pyx_PyLong_From_enum__mlx5dv_wc_opcode(MLX5DV_WC_MEMCPY); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_22 = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_MLX5DV_WC_MEMCPY, __pyx_t_15); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "EnumType":23
 * 
 * 
 * mlx5dv_wc_opcode = __Pyx_FlagBase('mlx5dv_wc_opcode', [             # <<<<<<<<<<<<<<
 *     ('MLX5DV_WC_UMR', __Pyx_PyLong_From_enum__mlx5dv_wc_opcode(MLX5DV_WC_UMR)),
 *     ('MLX5DV_WC_RAW_WQE', __Pyx_PyLong_From_enum__mlx5dv_wc_opcode(MLX5DV_WC_RAW_WQE)),
*/
  __pyx_t_15 = __Pyx_PyList_Pack(3, __pyx_t_11, __pyx_t_17, __pyx_t_22); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "EnumType":28
 *     ('MLX5DV_WC_MEMCPY', __Pyx_PyLong_From_enum__mlx5dv_wc_opcode(MLX5DV_WC_MEMCPY)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_t_22 = __Pyx_Globals(); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_22, __pyx_mstate_global->__pyx_n_u_get); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_mstate_global->__pyx_tuple[0], NULL); if (unlikely(!__pyx_t_22)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  __pyx_t_12 = 1;
  {
    PyObject *__pyx_callargs[3 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_21, __pyx_mstate_global->__pyx_n_u_mlx5dv_wc_opcode, __pyx_t_15};
    __pyx_t_17 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_module, __pyx_t_22, __pyx_t_17, __pyx_callargs+3, 0) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
    __pyx_t_20 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_18, __pyx_callargs+__pyx_t_12, (3-__pyx_t_12) | (__pyx_t_12*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_17);
    __Pyx_XDECREF(__pyx_t_21); __pyx_t_21 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 23, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
  }
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_mlx5dv_wc_opcode, __pyx_t_20) < (0)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_13 = (PY_VERSION_HEX >= 0x030B0000);
  if (__pyx_t_13) {

    /* "EnumType":35
 * 
 * 
 *     mlx5dv_wc_opcode._member_names_ = list(mlx5dv_wc_opcode.__members__)             # <<<<<<<<<<<<<<
*/
    __Pyx_GetModuleGlobalName(__pyx_t_20, __pyx_mstate_global->__pyx_n_u_mlx5dv_wc_opcode); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_20, __pyx_mstate_global->__pyx_n_u_members); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __pyx_t_20 = __Pyx_PySequence_ListKeepNew(__pyx_t_18); if (unlikely(!__pyx_t_20)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_mlx5dv_wc_opcode); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    if (__Pyx_PyObject_SetAttrStr(__pyx_t_18, __pyx_mstate_global->__pyx_n_u_member_names, __pyx_t_20) < (0)) __PYX_ERR(1, 35, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

    /* "EnumType":30
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))
 * 
 * if PY_VERSION_HEX >= 0x030B0000:             # <<<<<<<<<<<<<<
 * 
 * 
*/
  }

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":6
 * #cython: language_level=3
 * 
 * _MLX5DV_RES_TYPE_QP = MLX5DV_RES_TYPE_QP             # <<<<<<<<<<<<<<
 * _MLX5DV_RES_TYPE_RWQ = MLX5DV_RES_TYPE_RWQ
 * _MLX5DV_RES_TYPE_DBR = MLX5DV_RES_TYPE_DBR
*/
  __pyx_t_18 = __Pyx_PyLong_From_unsigned_PY_LONG_LONG(MLX5DV_RES_TYPE_QP); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_RES_TYPE_QP, __pyx_t_18) < (0)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":7
 * 
 * _MLX5DV_RES_TYPE_QP = MLX5DV_RES_TYPE_QP
 * _MLX5DV_RES_TYPE_RWQ = MLX5DV_RES_TYPE_RWQ             # <<<<<<<<<<<<<<
 * _MLX5DV_RES_TYPE_DBR = MLX5DV_RES_TYPE_DBR
 * _MLX5DV_RES_TYPE_SRQ = MLX5DV_RES_TYPE_SRQ
*/
  __pyx_t_18 = __Pyx_PyLong_From_unsigned_PY_LONG_LONG(MLX5DV_RES_TYPE_RWQ); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_RES_TYPE_RWQ, __pyx_t_18) < (0)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":8
 * _MLX5DV_RES_TYPE_QP = MLX5DV_RES_TYPE_QP
 * _MLX5DV_RES_TYPE_RWQ = MLX5DV_RES_TYPE_RWQ
 * _MLX5DV_RES_TYPE_DBR = MLX5DV_RES_TYPE_DBR             # <<<<<<<<<<<<<<
 * _MLX5DV_RES_TYPE_SRQ = MLX5DV_RES_TYPE_SRQ
 * _MLX5DV_PP_ALLOC_FLAGS_DEDICATED_INDEX = MLX5DV_PP_ALLOC_FLAGS_DEDICATED_INDEX
*/
  __pyx_t_18 = __Pyx_PyLong_From_unsigned_PY_LONG_LONG(MLX5DV_RES_TYPE_DBR); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_RES_TYPE_DBR, __pyx_t_18) < (0)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":9
 * _MLX5DV_RES_TYPE_RWQ = MLX5DV_RES_TYPE_RWQ
 * _MLX5DV_RES_TYPE_DBR = MLX5DV_RES_TYPE_DBR
 * _MLX5DV_RES_TYPE_SRQ = MLX5DV_RES_TYPE_SRQ             # <<<<<<<<<<<<<<
 * _MLX5DV_PP_ALLOC_FLAGS_DEDICATED_INDEX = MLX5DV_PP_ALLOC_FLAGS_DEDICATED_INDEX
 * _MLX5DV_UAR_ALLOC_TYPE_BF = MLX5DV_UAR_ALLOC_TYPE_BF
*/
  __pyx_t_18 = __Pyx_PyLong_From_unsigned_PY_LONG_LONG(MLX5DV_RES_TYPE_SRQ); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_RES_TYPE_SRQ, __pyx_t_18) < (0)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":10
 * _MLX5DV_RES_TYPE_DBR = MLX5DV_RES_TYPE_DBR
 * _MLX5DV_RES_TYPE_SRQ = MLX5DV_RES_TYPE_SRQ
 * _MLX5DV_PP_ALLOC_FLAGS_DEDICATED_INDEX = MLX5DV_PP_ALLOC_FLAGS_DEDICATED_INDEX             # <<<<<<<<<<<<<<
 * _MLX5DV_UAR_ALLOC_TYPE_BF = MLX5DV_UAR_ALLOC_TYPE_BF
 * _MLX5DV_UAR_ALLOC_TYPE_NC = MLX5DV_UAR_ALLOC_TYPE_NC
*/
  __pyx_t_18 = __Pyx_PyLong_From_unsigned_PY_LONG_LONG(MLX5DV_PP_ALLOC_FLAGS_DEDICATED_INDEX); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_PP_ALLOC_FLAGS_DEDICATED, __pyx_t_18) < (0)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":11
 * _MLX5DV_RES_TYPE_SRQ = MLX5DV_RES_TYPE_SRQ
 * _MLX5DV_PP_ALLOC_FLAGS_DEDICATED_INDEX = MLX5DV_PP_ALLOC_FLAGS_DEDICATED_INDEX
 * _MLX5DV_UAR_ALLOC_TYPE_BF = MLX5DV_UAR_ALLOC_TYPE_BF             # <<<<<<<<<<<<<<
 * _MLX5DV_UAR_ALLOC_TYPE_NC = MLX5DV_UAR_ALLOC_TYPE_NC
 * 
*/
  __pyx_t_18 = __Pyx_PyLong_From_unsigned_PY_LONG_LONG(MLX5DV_UAR_ALLOC_TYPE_BF); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_UAR_ALLOC_TYPE_BF, __pyx_t_18) < (0)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":12
 * _MLX5DV_PP_ALLOC_FLAGS_DEDICATED_INDEX = MLX5DV_PP_ALLOC_FLAGS_DEDICATED_INDEX
 * _MLX5DV_UAR_ALLOC_TYPE_BF = MLX5DV_UAR_ALLOC_TYPE_BF
 * _MLX5DV_UAR_ALLOC_TYPE_NC = MLX5DV_UAR_ALLOC_TYPE_NC             # <<<<<<<<<<<<<<
 * 
 * MLX5DV_QUERY_PORT_VPORT_ = MLX5DV_QUERY_PORT_VPORT
*/
  __pyx_t_18 = __Pyx_PyLong_From_unsigned_PY_LONG_LONG(MLX5DV_UAR_ALLOC_TYPE_NC); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_UAR_ALLOC_TYPE_NC, __pyx_t_18) < (0)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":14
 * _MLX5DV_UAR_ALLOC_TYPE_NC = MLX5DV_UAR_ALLOC_TYPE_NC
 * 
 * MLX5DV_QUERY_PORT_VPORT_ = MLX5DV_QUERY_PORT_VPORT             # <<<<<<<<<<<<<<
 * MLX5DV_QUERY_PORT_VPORT_VHCA_ID_ = MLX5DV_QUERY_PORT_VPORT_VHCA_ID
 * MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_RX_ = MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_RX
*/
  __pyx_t_18 = __Pyx_PyLong_From_unsigned_PY_LONG_LONG(MLX5DV_QUERY_PORT_VPORT); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_QUERY_PORT_VPORT, __pyx_t_18) < (0)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":15
 * 
 * MLX5DV_QUERY_PORT_VPORT_ = MLX5DV_QUERY_PORT_VPORT
 * MLX5DV_QUERY_PORT_VPORT_VHCA_ID_ = MLX5DV_QUERY_PORT_VPORT_VHCA_ID             # <<<<<<<<<<<<<<
 * MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_RX_ = MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_RX
 * MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_TX_ = MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_TX
*/
  __pyx_t_18 = __Pyx_PyLong_From_unsigned_PY_LONG_LONG(MLX5DV_QUERY_PORT_VPORT_VHCA_ID); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_QUERY_PORT_VPORT_VHCA_ID, __pyx_t_18) < (0)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":16
 * MLX5DV_QUERY_PORT_VPORT_ = MLX5DV_QUERY_PORT_VPORT
 * MLX5DV_QUERY_PORT_VPORT_VHCA_ID_ = MLX5DV_QUERY_PORT_VPORT_VHCA_ID
 * MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_RX_ = MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_RX             # <<<<<<<<<<<<<<
 * MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_TX_ = MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_TX
 * MLX5DV_QUERY_PORT_VPORT_REG_C0_ = MLX5DV_QUERY_PORT_VPORT_REG_C0
*/
  __pyx_t_18 = __Pyx_PyLong_From_unsigned_PY_LONG_LONG(MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_RX); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_QUERY_PORT_VPORT_STEERING, __pyx_t_18) < (0)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":17
 * MLX5DV_QUERY_PORT_VPORT_VHCA_ID_ = MLX5DV_QUERY_PORT_VPORT_VHCA_ID
 * MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_RX_ = MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_RX
 * MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_TX_ = MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_TX             # <<<<<<<<<<<<<<
 * MLX5DV_QUERY_PORT_VPORT_REG_C0_ = MLX5DV_QUERY_PORT_VPORT_REG_C0
 * MLX5DV_QUERY_PORT_ESW_OWNER_VHCA_ID_ = MLX5DV_QUERY_PORT_ESW_OWNER_VHCA_ID
*/
  __pyx_t_18 = __Pyx_PyLong_From_unsigned_PY_LONG_LONG(MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_TX); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_QUERY_PORT_VPORT_STEERING_2, __pyx_t_18) < (0)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":18
 * MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_RX_ = MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_RX
 * MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_TX_ = MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_TX
 * MLX5DV_QUERY_PORT_VPORT_REG_C0_ = MLX5DV_QUERY_PORT_VPORT_REG_C0             # <<<<<<<<<<<<<<
 * MLX5DV_QUERY_PORT_ESW_OWNER_VHCA_ID_ = MLX5DV_QUERY_PORT_ESW_OWNER_VHCA_ID
 * 
*/
  __pyx_t_18 = __Pyx_PyLong_From_unsigned_PY_LONG_LONG(MLX5DV_QUERY_PORT_VPORT_REG_C0); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_QUERY_PORT_VPORT_REG_C0, __pyx_t_18) < (0)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":19
 * MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_TX_ = MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_TX
 * MLX5DV_QUERY_PORT_VPORT_REG_C0_ = MLX5DV_QUERY_PORT_VPORT_REG_C0
 * MLX5DV_QUERY_PORT_ESW_OWNER_VHCA_ID_ = MLX5DV_QUERY_PORT_ESW_OWNER_VHCA_ID             # <<<<<<<<<<<<<<
 * 
 * MLX5DV_FLOW_TABLE_TYPE_RDMA_RX_ = MLX5DV_FLOW_TABLE_TYPE_RDMA_RX
*/
  __pyx_t_18 = __Pyx_PyLong_From_unsigned_PY_LONG_LONG(MLX5DV_QUERY_PORT_ESW_OWNER_VHCA_ID); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_QUERY_PORT_ESW_OWNER_VHCA, __pyx_t_18) < (0)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":21
 * MLX5DV_QUERY_PORT_ESW_OWNER_VHCA_ID_ = MLX5DV_QUERY_PORT_ESW_OWNER_VHCA_ID
 * 
 * MLX5DV_FLOW_TABLE_TYPE_RDMA_RX_ = MLX5DV_FLOW_TABLE_TYPE_RDMA_RX             # <<<<<<<<<<<<<<
 * MLX5DV_FLOW_TABLE_TYPE_RDMA_TX_ = MLX5DV_FLOW_TABLE_TYPE_RDMA_TX
 * MLX5DV_FLOW_TABLE_TYPE_NIC_RX_ = MLX5DV_FLOW_TABLE_TYPE_NIC_RX
*/
  __pyx_t_18 = __Pyx_PyLong_From_int(MLX5DV_FLOW_TABLE_TYPE_RDMA_RX); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_TABLE_TYPE_RDMA_RX, __pyx_t_18) < (0)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":22
 * 
 * MLX5DV_FLOW_TABLE_TYPE_RDMA_RX_ = MLX5DV_FLOW_TABLE_TYPE_RDMA_RX
 * MLX5DV_FLOW_TABLE_TYPE_RDMA_TX_ = MLX5DV_FLOW_TABLE_TYPE_RDMA_TX             # <<<<<<<<<<<<<<
 * MLX5DV_FLOW_TABLE_TYPE_NIC_RX_ = MLX5DV_FLOW_TABLE_TYPE_NIC_RX
 * MLX5DV_FLOW_TABLE_TYPE_NIC_TX_ = MLX5DV_FLOW_TABLE_TYPE_NIC_TX
*/
  __pyx_t_18 = __Pyx_PyLong_From_int(MLX5DV_FLOW_TABLE_TYPE_RDMA_TX); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_TABLE_TYPE_RDMA_TX, __pyx_t_18) < (0)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":23
 * MLX5DV_FLOW_TABLE_TYPE_RDMA_RX_ = MLX5DV_FLOW_TABLE_TYPE_RDMA_RX
 * MLX5DV_FLOW_TABLE_TYPE_RDMA_TX_ = MLX5DV_FLOW_TABLE_TYPE_RDMA_TX
 * MLX5DV_FLOW_TABLE_TYPE_NIC_RX_ = MLX5DV_FLOW_TABLE_TYPE_NIC_RX             # <<<<<<<<<<<<<<
 * MLX5DV_FLOW_TABLE_TYPE_NIC_TX_ = MLX5DV_FLOW_TABLE_TYPE_NIC_TX
 * MLX5DV_FLOW_TABLE_TYPE_FDB_ = MLX5DV_FLOW_TABLE_TYPE_FDB
*/
  __pyx_t_18 = __Pyx_PyLong_From_int(MLX5DV_FLOW_TABLE_TYPE_NIC_RX); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_TABLE_TYPE_NIC_RX, __pyx_t_18) < (0)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":24
 * MLX5DV_FLOW_TABLE_TYPE_RDMA_TX_ = MLX5DV_FLOW_TABLE_TYPE_RDMA_TX
 * MLX5DV_FLOW_TABLE_TYPE_NIC_RX_ = MLX5DV_FLOW_TABLE_TYPE_NIC_RX
 * MLX5DV_FLOW_TABLE_TYPE_NIC_TX_ = MLX5DV_FLOW_TABLE_TYPE_NIC_TX             # <<<<<<<<<<<<<<
 * MLX5DV_FLOW_TABLE_TYPE_FDB_ = MLX5DV_FLOW_TABLE_TYPE_FDB
 * MLX5DV_FLOW_TABLE_TYPE_RDMA_TRANSPORT_RX_ = MLX5DV_FLOW_TABLE_TYPE_RDMA_TRANSPORT_RX
*/
  __pyx_t_18 = __Pyx_PyLong_From_int(MLX5DV_FLOW_TABLE_TYPE_NIC_TX); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_TABLE_TYPE_NIC_TX, __pyx_t_18) < (0)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":25
 * MLX5DV_FLOW_TABLE_TYPE_NIC_RX_ = MLX5DV_FLOW_TABLE_TYPE_NIC_RX
 * MLX5DV_FLOW_TABLE_TYPE_NIC_TX_ = MLX5DV_FLOW_TABLE_TYPE_NIC_TX
 * MLX5DV_FLOW_TABLE_TYPE_FDB_ = MLX5DV_FLOW_TABLE_TYPE_FDB             # <<<<<<<<<<<<<<
 * MLX5DV_FLOW_TABLE_TYPE_RDMA_TRANSPORT_RX_ = MLX5DV_FLOW_TABLE_TYPE_RDMA_TRANSPORT_RX
 * MLX5DV_FLOW_TABLE_TYPE_RDMA_TRANSPORT_TX_ = MLX5DV_FLOW_TABLE_TYPE_RDMA_TRANSPORT_TX
*/
  __pyx_t_18 = __Pyx_PyLong_From_int(MLX5DV_FLOW_TABLE_TYPE_FDB); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_TABLE_TYPE_FDB, __pyx_t_18) < (0)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":26
 * MLX5DV_FLOW_TABLE_TYPE_NIC_TX_ = MLX5DV_FLOW_TABLE_TYPE_NIC_TX
 * MLX5DV_FLOW_TABLE_TYPE_FDB_ = MLX5DV_FLOW_TABLE_TYPE_FDB
 * MLX5DV_FLOW_TABLE_TYPE_RDMA_TRANSPORT_RX_ = MLX5DV_FLOW_TABLE_TYPE_RDMA_TRANSPORT_RX             # <<<<<<<<<<<<<<
 * MLX5DV_FLOW_TABLE_TYPE_RDMA_TRANSPORT_TX_ = MLX5DV_FLOW_TABLE_TYPE_RDMA_TRANSPORT_TX
 * 
*/
  __pyx_t_18 = __Pyx_PyLong_From_int(MLX5DV_FLOW_TABLE_TYPE_RDMA_TRANSPORT_RX); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_TABLE_TYPE_RDMA_TRAN, __pyx_t_18) < (0)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":27
 * MLX5DV_FLOW_TABLE_TYPE_FDB_ = MLX5DV_FLOW_TABLE_TYPE_FDB
 * MLX5DV_FLOW_TABLE_TYPE_RDMA_TRANSPORT_RX_ = MLX5DV_FLOW_TABLE_TYPE_RDMA_TRANSPORT_RX
 * MLX5DV_FLOW_TABLE_TYPE_RDMA_TRANSPORT_TX_ = MLX5DV_FLOW_TABLE_TYPE_RDMA_TRANSPORT_TX             # <<<<<<<<<<<<<<
 * 
 * MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TUNNEL_TO_L2_ = \
*/
  __pyx_t_18 = __Pyx_PyLong_From_int(MLX5DV_FLOW_TABLE_TYPE_RDMA_TRANSPORT_TX); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_TABLE_TYPE_RDMA_TRAN_2, __pyx_t_18) < (0)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":30
 * 
 * MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TUNNEL_TO_L2_ = \
 *     MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TUNNEL_TO_L2             # <<<<<<<<<<<<<<
 * MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TO_L2_TUNNEL_ = \
 *     MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TO_L2_TUNNEL
*/
  __pyx_t_18 = __Pyx_PyLong_From_int(MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TUNNEL_TO_L2); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_ACTION_PACKET_REFORM, __pyx_t_18) < (0)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":32
 *     MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TUNNEL_TO_L2
 * MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TO_L2_TUNNEL_ = \
 *     MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TO_L2_TUNNEL             # <<<<<<<<<<<<<<
 * MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L3_TUNNEL_TO_L2_ = \
 *     MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L3_TUNNEL_TO_L2
*/
  __pyx_t_18 = __Pyx_PyLong_From_int(MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TO_L2_TUNNEL); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_ACTION_PACKET_REFORM_2, __pyx_t_18) < (0)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":34
 *     MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TO_L2_TUNNEL
 * MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L3_TUNNEL_TO_L2_ = \
 *     MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L3_TUNNEL_TO_L2             # <<<<<<<<<<<<<<
 * MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TO_L3_TUNNEL_ = \
 *     MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TO_L3_TUNNEL
*/
  __pyx_t_18 = __Pyx_PyLong_From_int(MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L3_TUNNEL_TO_L2); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_ACTION_PACKET_REFORM_3, __pyx_t_18) < (0)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":36
 *     MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L3_TUNNEL_TO_L2
 * MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TO_L3_TUNNEL_ = \
 *     MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TO_L3_TUNNEL             # <<<<<<<<<<<<<<
 * 
 * MLX5DV_REG_DMABUF_ACCESS_DATA_DIRECT_ = MLX5DV_REG_DMABUF_ACCESS_DATA_DIRECT
*/
  __pyx_t_18 = __Pyx_PyLong_From_int(MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TO_L3_TUNNEL); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_FLOW_ACTION_PACKET_REFORM_4, __pyx_t_18) < (0)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":38
 *     MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TO_L3_TUNNEL
 * 
 * MLX5DV_REG_DMABUF_ACCESS_DATA_DIRECT_ = MLX5DV_REG_DMABUF_ACCESS_DATA_DIRECT             # <<<<<<<<<<<<<<
*/
  __pyx_t_18 = __Pyx_PyLong_From_int(MLX5DV_REG_DMABUF_ACCESS_DATA_DIRECT); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MLX5DV_REG_DMABUF_ACCESS_DATA_DI, __pyx_t_18) < (0)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "pyverbs/providers/mlx5/mlx5_enums.pyx":1
 * # SPDX-License-Identifier: (GPL-2.0 OR Linux-OpenIB)             # <<<<<<<<<<<<<<
 * # Copyright (c) 2024 Nvidia All rights reserved.
 * 
*/
  __pyx_t_18 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_test, __pyx_t_18) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /*--- Wrapped vars code ---*/
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5DV_UMEM_MASK_DMABUF);
    if (unlikely(!wrapped)) __PYX_ERR(2, 298, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5DV_UMEM_MASK_DMABUF", wrapped) < 0) __PYX_ERR(2, 298, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_CQE_INVALID);
    if (unlikely(!wrapped)) __PYX_ERR(2, 228, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_CQE_INVALID", wrapped) < 0) __PYX_ERR(2, 228, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_CQE_NO_PACKET);
    if (unlikely(!wrapped)) __PYX_ERR(2, 224, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_CQE_NO_PACKET", wrapped) < 0) __PYX_ERR(2, 224, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_CQE_OWNER_MASK);
    if (unlikely(!wrapped)) __PYX_ERR(2, 217, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_CQE_OWNER_MASK", wrapped) < 0) __PYX_ERR(2, 217, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_CQE_REQ);
    if (unlikely(!wrapped)) __PYX_ERR(2, 218, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_CQE_REQ", wrapped) < 0) __PYX_ERR(2, 218, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_CQE_REQ_ERR);
    if (unlikely(!wrapped)) __PYX_ERR(2, 226, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_CQE_REQ_ERR", wrapped) < 0) __PYX_ERR(2, 226, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_CQE_RESIZE_CQ);
    if (unlikely(!wrapped)) __PYX_ERR(2, 223, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_CQE_RESIZE_CQ", wrapped) < 0) __PYX_ERR(2, 223, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_CQE_RESP_ERR);
    if (unlikely(!wrapped)) __PYX_ERR(2, 227, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_CQE_RESP_ERR", wrapped) < 0) __PYX_ERR(2, 227, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_CQE_RESP_SEND);
    if (unlikely(!wrapped)) __PYX_ERR(2, 220, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_CQE_RESP_SEND", wrapped) < 0) __PYX_ERR(2, 220, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_CQE_RESP_SEND_IMM);
    if (unlikely(!wrapped)) __PYX_ERR(2, 221, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_CQE_RESP_SEND_IMM", wrapped) < 0) __PYX_ERR(2, 221, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_CQE_RESP_SEND_INV);
    if (unlikely(!wrapped)) __PYX_ERR(2, 222, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_CQE_RESP_SEND_INV", wrapped) < 0) __PYX_ERR(2, 222, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_CQE_RESP_WR_IMM);
    if (unlikely(!wrapped)) __PYX_ERR(2, 219, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_CQE_RESP_WR_IMM", wrapped) < 0) __PYX_ERR(2, 219, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_CQE_SIG_ERR);
    if (unlikely(!wrapped)) __PYX_ERR(2, 225, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_CQE_SIG_ERR", wrapped) < 0) __PYX_ERR(2, 225, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_OPCODE_ATOMIC_CS);
    if (unlikely(!wrapped)) __PYX_ERR(2, 17, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_OPCODE_ATOMIC_CS", wrapped) < 0) __PYX_ERR(2, 17, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_OPCODE_ATOMIC_FA);
    if (unlikely(!wrapped)) __PYX_ERR(2, 18, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_OPCODE_ATOMIC_FA", wrapped) < 0) __PYX_ERR(2, 18, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_OPCODE_ATOMIC_MASKED_CS);
    if (unlikely(!wrapped)) __PYX_ERR(2, 19, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_OPCODE_ATOMIC_MASKED_CS", wrapped) < 0) __PYX_ERR(2, 19, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_OPCODE_ATOMIC_MASKED_FA);
    if (unlikely(!wrapped)) __PYX_ERR(2, 20, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_OPCODE_ATOMIC_MASKED_FA", wrapped) < 0) __PYX_ERR(2, 20, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_OPCODE_CONFIG_CMD);
    if (unlikely(!wrapped)) __PYX_ERR(2, 23, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_OPCODE_CONFIG_CMD", wrapped) < 0) __PYX_ERR(2, 23, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_OPCODE_FMR);
    if (unlikely(!wrapped)) __PYX_ERR(2, 21, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_OPCODE_FMR", wrapped) < 0) __PYX_ERR(2, 21, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_OPCODE_LOCAL_INVAL);
    if (unlikely(!wrapped)) __PYX_ERR(2, 22, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_OPCODE_LOCAL_INVAL", wrapped) < 0) __PYX_ERR(2, 22, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_OPCODE_MMO);
    if (unlikely(!wrapped)) __PYX_ERR(2, 26, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_OPCODE_MMO", wrapped) < 0) __PYX_ERR(2, 26, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_OPCODE_NOP);
    if (unlikely(!wrapped)) __PYX_ERR(2, 9, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_OPCODE_NOP", wrapped) < 0) __PYX_ERR(2, 9, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_OPCODE_RDMA_READ);
    if (unlikely(!wrapped)) __PYX_ERR(2, 16, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_OPCODE_RDMA_READ", wrapped) < 0) __PYX_ERR(2, 16, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_OPCODE_RDMA_WRITE);
    if (unlikely(!wrapped)) __PYX_ERR(2, 11, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_OPCODE_RDMA_WRITE", wrapped) < 0) __PYX_ERR(2, 11, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_OPCODE_RDMA_WRITE_IMM);
    if (unlikely(!wrapped)) __PYX_ERR(2, 12, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_OPCODE_RDMA_WRITE_IMM", wrapped) < 0) __PYX_ERR(2, 12, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_OPCODE_SEND);
    if (unlikely(!wrapped)) __PYX_ERR(2, 13, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_OPCODE_SEND", wrapped) < 0) __PYX_ERR(2, 13, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_OPCODE_SEND_IMM);
    if (unlikely(!wrapped)) __PYX_ERR(2, 14, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_OPCODE_SEND_IMM", wrapped) < 0) __PYX_ERR(2, 14, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_OPCODE_SEND_INVAL);
    if (unlikely(!wrapped)) __PYX_ERR(2, 10, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_OPCODE_SEND_INVAL", wrapped) < 0) __PYX_ERR(2, 10, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_OPCODE_TAG_MATCHING);
    if (unlikely(!wrapped)) __PYX_ERR(2, 25, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_OPCODE_TAG_MATCHING", wrapped) < 0) __PYX_ERR(2, 25, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_OPCODE_TSO);
    if (unlikely(!wrapped)) __PYX_ERR(2, 15, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_OPCODE_TSO", wrapped) < 0) __PYX_ERR(2, 15, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_OPCODE_UMR);
    if (unlikely(!wrapped)) __PYX_ERR(2, 24, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_OPCODE_UMR", wrapped) < 0) __PYX_ERR(2, 24, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_RCV_DBR);
    if (unlikely(!wrapped)) __PYX_ERR(2, 213, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_RCV_DBR", wrapped) < 0) __PYX_ERR(2, 213, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_SEND_WQE_BB);
    if (unlikely(!wrapped)) __PYX_ERR(2, 231, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_SEND_WQE_BB", wrapped) < 0) __PYX_ERR(2, 231, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_SEND_WQE_SHIFT);
    if (unlikely(!wrapped)) __PYX_ERR(2, 232, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_SEND_WQE_SHIFT", wrapped) < 0) __PYX_ERR(2, 232, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_SND_DBR);
    if (unlikely(!wrapped)) __PYX_ERR(2, 214, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_SND_DBR", wrapped) < 0) __PYX_ERR(2, 214, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_WQE_CTRL_CQ_UPDATE);
    if (unlikely(!wrapped)) __PYX_ERR(2, 29, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_WQE_CTRL_CQ_UPDATE", wrapped) < 0) __PYX_ERR(2, 29, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_WQE_CTRL_FENCE);
    if (unlikely(!wrapped)) __PYX_ERR(2, 31, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_WQE_CTRL_FENCE", wrapped) < 0) __PYX_ERR(2, 31, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_WQE_CTRL_INITIATOR_SMALL_FENCE);
    if (unlikely(!wrapped)) __PYX_ERR(2, 32, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_WQE_CTRL_INITIATOR_SMALL_FENCE", wrapped) < 0) __PYX_ERR(2, 32, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyLong_From___pyx_anon_enum(MLX5_WQE_CTRL_SOLICITED);
    if (unlikely(!wrapped)) __PYX_ERR(2, 30, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "MLX5_WQE_CTRL_SOLICITED", wrapped) < 0) __PYX_ERR(2, 30, __pyx_L1_error);
  }

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  if (__pyx_m) {
    if (__pyx_mstate->__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init pyverbs.providers.mlx5.mlx5_enums", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init pyverbs.providers.mlx5.mlx5_enums");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #else
  return __pyx_m;
  #endif
}
/* #### Code section: pystring_table ### */
/* #### Code section: cached_builtins ### */

static int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  return 0;
}
/* #### Code section: cached_constants ### */

static int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "EnumType":30
 *     ('MLX5DV_BLOCK_SIZE_4160', __Pyx_PyLong_From_enum__mlx5dv_block_size(MLX5DV_BLOCK_SIZE_4160)),
 * 
 * ], module=globals().get("__module__", 'pyverbs.providers.mlx5.mlx5_enums'))             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x030B0000:
*/
  __pyx_mstate_global->__pyx_tuple[0] = PyTuple_Pack(2, __pyx_mstate_global->__pyx_n_u_module_2, __pyx_mstate_global->__pyx_kp_u_pyverbs_providers_mlx5_mlx5_enum); if (unlikely(!__pyx_mstate_global->__pyx_tuple[0])) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[0]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[0]);
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_tuple;
    for (Py_ssize_t i=0; i<1; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  {
    const struct { const unsigned int length: 6; } index[] = {{1},{1},{33},{7},{7},{22},{22},{22},{21},{21},{33},{30},{34},{33},{27},{25},{33},{32},{45},{33},{37},{34},{34},{31},{31},{30},{37},{34},{33},{32},{37},{31},{23},{35},{36},{26},{33},{26},{33},{34},{39},{33},{30},{25},{45},{56},{33},{46},{45},{33},{26},{26},{33},{34},{31},{30},{47},{17},{17},{33},{22},{22},{51},{21},{30},{33},{30},{31},{30},{25},{28},{28},{33},{34},{32},{44},{31},{28},{30},{23},{36},{49},{49},{49},{49},{30},{34},{56},{56},{56},{56},{22},{32},{32},{27},{30},{30},{31},{41},{41},{31},{43},{34},{36},{45},{18},{32},{31},{32},{13},{13},{13},{13},{13},{14},{14},{38},{37},{39},{23},{41},{31},{43},{43},{32},{24},{32},{32},{25},{25},{27},{36},{40},{39},{29},{30},{30},{36},{24},{31},{40},{40},{32},{44},{42},{45},{55},{55},{37},{20},{19},{20},{20},{37},{39},{58},{59},{36},{25},{26},{30},{21},{22},{26},{29},{28},{29},{21},{22},{33},{37},{32},{19},{22},{20},{17},{22},{21},{25},{25},{36},{16},{17},{13},{18},{4},{3},{8},{14},{11},{17},{29},{25},{24},{20},{25},{24},{26},{24},{26},{25},{22},{25},{22},{40},{14},{25},{16},{38},{26},{22},{27},{21},{30},{28},{23},{29},{20},{27},{15},{19},{22},{31},{24},{28},{27},{19},{15},{25},{23},{15},{29},{20},{26},{22},{30},{16},{6},{10},{8},{12},{10},{8}};
    #if (CYTHON_COMPRESS_STRINGS) == 3 && __PYX_LIMITED_VERSION_HEX >= 0x030e0000 /* compression: zstd (1904 bytes) */
const char* const cstring = "(\265/\375`<\0325;\000\372T\300\016)\360\250\210u\327\037\006P\177\213+\250\"\002\032 \330\314@V\300-\367jh\310\222BG\t?\320\2056\360\014<\003\317(\017\336\000\320\000\320\000\313h\274f\256\323}\035\227O\3563\303\240\013\310\356\016\271\036\277\367\343\030\302\206\312\207\351/`c\255\351\344n\350\302E\233\375\334Y\334\2171v\224\377k\340\213t\317\272\261\017\306F\327\376Yp/\212C\207\334u\311\342\347[n:\337\026c\033\335\262\304G*\035\rc\233\334\3140]&\233\307\005\2665\351\3715z\267\334\314\342;m\214{\235\361\\\305)/\254\225\374D\347\366\217|g\274\3402 7\034f\007\016[)\237b\000\302\301\231Y\374\270\231\213u\213c\273\367l\371\302\334\367\246tO\033La\\\010\010\007\t\006\n\007\n\004\006\0100\020p>\320\346\027Vt\212;,\241\341\236\003\024Y\270\2324JR\231\324'(\210H\000PS\215 \021\322D\001@1L5b\202R\000@\221.\356\306`s\360\302\243\023n\323\376>\326\364T\321\226\006\210\221\"8\330\356\2647\250k\226\005l\315L\347\233\234K\371\235\341\376[\213\031[2w\024\215\3169\333;u\315.\314(\317e\211g\320\366\274\231\302&.\305r\263k\2326\321\261\356\315\214E\367\321F\1775\250`\313|\374\036\306\202\216y\330\366E\346\330aw\032#\367\351d\326K\227O\333\213\216!\\\306:\205\354p\2577\034\007[\356\216\335\274\300\264\261t\320\305XdP~\247\035\345k\254\227A\310\232[\007\2663]:\037\235B\002\002\317(\273\221w\360B\243\256y\354\346'm\354'\3036g{/\306\205[\030\033\335f\031\031K\271\3750\303\014;\315\223\233\255l\260}\006\273&_\312}\315\\\270\227\306\014\016\35028\220\215\013PX\010\001V/\207\214\353\334qaE\203\0366\323e\233Y/\234\345^\306\374\363\330\361\0313\326\354\311<-\231V|\376\237\223\327\332/\215b\037\373\202%k\232\235\370\351\344\373\263\034\206\314\275\200\353\032\316\031\313]\3176m\364\337j7X_p;g\355\273\030\316\233\341\023\377L\276\243\271\221\373\336{\233f\214\035\203\313\2302\236A\236\2131\363\374\331\246\2639\306\316\262k\246\260\275\271\233E[`\\\374Vs\205\214\321}q\334\2156\307\262I\311fc\221\276\231\316?\222a\006\333\031\323]\2151\223\362L\317\247%\303\344c\314b\354\r(\342\302\315z\233""\205\224{\266\212n~\204t\003\177E\325\352\257\"\025\250\241FU\245\"P\005\342\177\215T\"IR#\351\361\220\220 \r\021\225*\202\032\244;L \324\004Q.AALS\215&\251N\225\n\23280\274\001\022\031\034T\323\303\001BIj\244>\242(\212\364hR)\nD\002\006i\232\252DB\252H\023P\344SE\360xL&()\252\250\211\004\251\036\320\247(R'\023\010\301\244F\324W\005\021\017\265R&\252?\242HBPU\223D\221@\036U\n\005\232$\242\t\212&\025TQ\227\"\225\222(S\205\362\211\372@9L*E\201\036\227*\312\304TA\222Ij\230*\004\202r\200\"\r\017\024E2U\025\364JA\255&\322\243.\200P\177xH\021\206GDT\225@\223O\225AR!\311_U\207\207\007\304P\037\320T}\002y\235@\"\217~\315z\236\310\276\000\233k\320\365\330\026\007.\212m\036\247[\3006\0332-\367\261\014\201\265\250\221\231dhhF\222\244\220\306Q\010\201\241\211N\231\231\007\002)\230\"\221\206\203\030e\030Q\010\031\212$ \n(,I\222\264\006\227\302kF\272\221\035Wt\206%\360 \355\224\271\253c\027\312%\316H\026@\265%\211\264\276\022\304\357Z\233\004pp3\216\340\341\033\026{\177;\010u\243\270\224\250\305v\212\273\207\321\321\326\317R\253\332\n\034!\316%\205j\345v\016\212-N\347S(\337z\3358\025\321\017\000\341_\205{'\331\364\215\304\210H\204\233\002\266|\363\257\000<{n\324\262\370\014\234\231x\214\007\301\027?\366\215\3044\277CIYj\030\303\361\031\340\364G\310\302\354Fw\032\235\021\231\275\033\240\301{\233\301h\017N\303\316F0\376\021lhe^~J\025s\344\001:F[`\232\010\303*Z\355\272f\202\t\t\241\230\212vY#9\260\t\037\302\315\266C%\001sy\207\234\3406S\207G\223\325,\330\354\327J\227&{\246\325L\217\276\001\224+\255\233\243;\306\300\222J? Z\005\006\022!\2345\020-\235\325O\321\230\257a\353}\316\372\032\034\017\300\344>\366\004\351n\316q\340\3120F$x\214\226\200\337\363\310\376\007\237'G\031\257g\226D\022}s\330\2473X\320Z\0372+#\312R\225p\246\331\240T\020\277\010\313q\363\231\224\241\230!p)\000\276A$\243`Y\202\004\321(\224\257\237pg\234\t\247\345&mxN\230\t$\231 \336\013_50\034\315\213c\330W!V\255\356\211\335\354\034\177\213\276\333\206{\326\205\350\"\226\r\262\364\363\247\t|\234r`\371\226\210\332\277~^F""\356\375nIh\n\271:\330=7\373\203O\375\232]|\27377h5b\340\035I\350\344\334\337x!\221!\n06\r\371\233\035\246\332\211\252^\351\t\316\207\n1\360v\004\306\335\301k\370\317\035\n\017\224\207Fu8,P^Td\211\261\0002\340\0030\220\001\340\203\300l\"8\350]\277\376\211x\207\267\324J\355X\220\030\375D\326G~I\232\237\232\263\317\312\220:<\356\342\360\213\260f+g|\210\347\341\305\327\352\032N\264\017\340\223\310\276 .\005\026U0D\032\241*\035\336\372\036\320\3446\001\005\217_\220\231[\342$\010\250\275\026\200kr\222\260\210f\214\033\310\376bw\212\001wq\177\243\274\207\024!YC\270|6]t8\215e\257\270\224\007\\\221I\021\301E\325S\003\335\013QyE\224\247\3079\3501\301S\033 K\t\013\223\177\t\034\231\030\212\364\313\367(\246\273\244\002@\333\010\371B\t\246$J2[I\004\373\010\373/\016\346\245\223\325c\321C\005\355D\250\267T\247\213)0\002\372\207\366B\304Q5\204C\245\034\022\323<\254\370\334\270\274\230\350T6\n\377Lf)\244\031#\331\244v\032\244\232$\200A\001A\232F\337^h\223\262o\204\002{\266\303\225d\353\215\213\006\351U\000#\317\017\024\330\241\353\311\231^i\313INR\033\310O&o \037eC\233\0253\362\210\207+c#n}\030\n\214S\204m\n\2415\224\246i\361;H\002F\333\364BoF\034\243\014\2764\312@05\017\210HQ\263T\023\375\003\375p\034\014A\345!\362\342<\024\365\345\005\024\223A\025T+?D\322\3106\001#+\256\334\301\327\204a2\2364\322\265\242\200\302\014\023\356[\217\231.\241\325\010\201u\020\266\021\200&\326\331\332]\352AJ\250\002";
    PyObject *data = __Pyx_DecompressString(cstring, 1904, 3);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) == 2 /* compression: bz2 (2032 bytes) */
const char* const cstring = "BZh91AY&SY\036\271U\223\000\000\001_\200@\000\000\001\177`\277\377\377\360\277\377\377\360`\t\235\365\001\323AN 4\000\024\351\255\010.\216\2350\324\321\242`E\037\212\236\247\205\014\232\032\000\001\220\0002\020\324\332\"i\246\212\031\000\006\200\000\320\332z\220s\000\230\0010\000\t\200\000\230\000\t\n\022S\366\225\036\321\242\236\221\243#G\352\200\000\032\000\0070\t\200\023\000\000\230\000\t\200\000\221&\201\001\002h%4\036\211\204\000i\240i\241\345_|^&\304\027\306\033\231\324\203\244\374\027W\307YM\2131\263\240W\006Ut\334\374\250q\366$\t\024\000Y\301\343[\033\275;\375\337_t\276|xH\370!$U\323Q\265\321h\014u\005dq\022\t\260\\\224K^\305\260!\325\035\037\326x3\013\001@(\321\366\037\252\325\341J\340\233\202\013&\027\203SV1\215\203\322\272Xb\r\205D\276w\233 \221\356c\004\206\345\200\345/\232\010B\330\217\251P\371\313\2455\027r\020\251\007B\3756\265M\303Ed\215RlN\275\270\257\251\271\257\324\270c\300\200\272h@e$c\265 \202\rp\276\332\267\200\235\023\232\333\017\265\374Q\264\022\321s\004 \276\236\3501I$\210\250^Zu\"\231\230\252\016s\265\001\\\371$\t\031\212^6\213\345\0221\246Xb\326\205D\274\265\2438\366<.O#\364\363\224B\317om@,q#Aa\027\2214\342\003\307WF\"\2579\2122\325L\303`\316#\277r\317\025\307u\251Q\230\344Q~\241\271B\234\334\321k\002\216\267\nUJ\356\206\014\\\233\037jFm|!\275P(EAu;%\215l\372\351h\n&\252\230\323@\343)DH\245iJ\020\311\325*\315t\311s\002\333\276\210A>\302!\t{P\204\302\276\265(w\251\346\352Xn^)\255\205\225\252\032\312lv\024\253J\000\232\023A\206\265\245\025h\324%B!\r\216\276\235\242\345\005\304\336*\304\303\321\227\r\373\261\320\224\027\027F\204_5^\251\316p\340\201\343\215,\222w!5\250\264\270\322F\2762Q?\006s\354-\304\245\324\234\240\2768?_\341\350\225\0224\026\032d\030`\301h\3058\310\304\333C\203J:\201\264\0206k\016\030#\232\360\256o\205@\223\272rBVt\310\212\213V.\352\3519\022\032iGS\005yH\0220\256\314\263_\002\212\344\004n\003*\016SVk\315\242\355\001\343\247^\2362\353\206\354\265\337\300\245\340X\251@n\020N\207+R\334\226\006\205\337L\365\031\025(\333H\305""\372\017\032\342bU\231eA0L\212j&\327\335m\271\004\220d \244!5RX\273\013z\370\034T\3718\031p\270\256\241\272KR\300\221\275\000\321r^@D\347!\226\021NP\372r\351\327&\324\3728\265\270+=\261>\017\242\226\326\205Ie}\331\001>\244T\274\266\306\312\352+[*0<h\013\264\252b!46'p\245\020\r\236b\376/\262(\305r\013\321Y\002\305I4\224\"\004\323^\342%&\001\311\036\314\025\021\307AS\324\352\332\303i\240LH\240X\366\320)\016\254\356\202\236\020\033Z\313r\346\036\260\000j-o\013\303\001\260\360\305G\273\3662\001'h\341gn\345\241!WNW\217K>\206.\255\372L1\272#\275Q\014\250\3143cP\025jwV\312.\027\323T\020\304\251$F\013\2624\0319\013\200\024D\344\353\274j\226A[$dP.\232\271\"9)\003A\331\256\247\241z-\333iI\226\262j\024@\365\342\316-NC6\032\342\014\000V\032[\200\240Y\\\\f\263\"\204\021\321\216\005\003(g\342\365\025\257\006CZ\220\251\320\332\222\301.\025\254P\t\004\033g\236\2051\305(\031\245\244\311\347O%\021\000\370\034 q\001\013\250A\014VP\230l\364f\303\273D\324\010\026\317A\242\035\002F,\0007\034\267\301\240\230\232f\301L\020\201Q\034\261\031\316\242\322wI\020\266\215f\333\314\323r\255]4\2300L\220\330A\027\362+\002(\220\304\326\030&4\203D\211k)oc8\255\220L\3572\222\256\316\207{d&\2149\310@\266\344\335\032x\027{c\266vX\350GT\216\252cf\216\322M\207\246:*\223F\005\021\334In\304\210h\010i\013\263\000\206\220\007Z\253\261i\353\305\352EMF\212\002,\313\225\325\r2\212\355(;$\t\036\246\272sC\2106\002\020+\313\222\255\214nB\\\333cU\024,\t\\\005dXp&.\020eD\032r\0131\264\233I\340*\343\024\365J\226\254-\207\002\0278\264\331\362`{B\203\036Y\306\003\247T8o\216\352\2333\"\322|iO}\020-\036\306\2304.\352\242;Q\031\226\333\351\031[\234\264\2642`P\327V\263\030\224\022\033\242\026\364\023\253\212\224`\266\2262(\002\214\031ux\313\020b\361RC\\\225\243\317\034H\267\227(\202\371q],\330\333\363\005\221\024\345\304\371e\n\316\205\036\351h\320\214\250\361\010\343\213\301~5\266\244g\241\324\301i;{4\326F\312\345\241\236\231\242\203m\356\261\216\322Y\033\214\355k7E\0101\215\215=G\0332S\227'\223~\017n\365\234\334\257@>G""\241\353\334W\361\217\224\224\370\276o\256\262\250\320a\236a9\021\3055~\004\353y\3600Dn\265\005!h\202B\3515G\260x\245\027\277\254\276\202\345\271f\255\"\315+\304$P\337\232\203z\355\321\034v\351\304\3537\366s\225P\245\253\305u=\361^k\222\016[\360P\207\3254\025im\024%\027\244W^\224\3567\361\260\312cH\302\262\3159C7J};X\267$\237\007R\266E>\355&\224\357\241\304t\315\"\201\2676\315B\264\243r\301\225\344\260\311\367\270\234\252\017\026=\000\350,0\224{\004\201\351\036\230Q\364\035~\271\310,\203m\314\"\201\325\373S$\374\373\266\267e\272\337\004\347\333\225F\010D~(\030\211\241\026J\204\"0\373w=.\230\355\035\223\275\035\324f\230\226\306y\3357\327p\252j\301\212&\213\253He\210P\313M\371\345\025\305\300\262ee'\031\253\357\223-\305\025\255\223\014\266\223\212\n\317\235@a\231\000r=\3267\332\360\207\206\230\0224.\026\344\212\362\275&'\212\013>\266i\001\371?]\214\023\364\361d\n\201;\032\223vN\177\355\017\300\244%\202\247\311$\216\235\010\263\261\342\004\224R\267M\014\022{\307\365\n\r\"\0004\354\3034\300bl2S\343l\027\361\255\2615F\241\350y\360op\017q\207\233.\031\313\333\023J\200x\r\264\334\341-m\313\207\342\223(\343\001\252$\343^[|\230*\202O\277 \023\362&\373\324\311hV\312\001;\317Q\244w\241\202\203ug\374\034\370\335\261\363\005\026\222U\306\306\307\034r\r\2228\361\245\026\025\2728\335\304\301\377|[\307\350\3078\364\031F\212i\367\223\252?l\035n\322\330\331\236\376\276\204\223\340\224\\\032%y\370:\371\302\374\313\263\325\030\351\214 \234_\331\331\225\002\377\372\333*\007\257&\232(@U\000U\243y\231\355nLbS\335?\003\027\233\200#\247}\031*\022\035\010K0\203\340,4\216\0077\032_\006\324\364\372Oa\265X:mp\330\303Fn\212t\t\334\016\311\026\242Cm\007\257%0\252[\245RH\317\234\326\215eCr\001t\331\005r\272\r\240e\274\377\213\271\"\234(H\017\\\252\311\200";
    PyObject *data = __Pyx_DecompressString(cstring, 2032, 2);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) != 0 /* compression: zlib (1924 bytes) */
const char* const cstring = "x\332\245X[s\243\270\022\376GSI\346R\273O[2\022\2666\334,\004\266\367\245\213\030\222\361\3066\016\220\314\344\374\372\323\342b\004\026NN\235\027c\365\367\251%\265\372\006_\376:\275\277e\305C\371\345T\344o\2734+\312/\207\375\357\357\365\017d\307\327C\311\217\025\303'>\354}\362\344:\353\3574\206\231\343[\367\020\362\177\030|\273\371\366\207I\372\347\017\203\364\366\307\315\245\364\373\355\235Ax\3271-\337\223l-\201H)\300%\341=\330\024\210\020d3\"\330\016\231\207\250#b\370\317eF\324Z2\270\275\373c\206b7\270\316\010\010\235$\304\267F\210\262xm\004\230\267 \236\305(\270\301\312H@9\020\307\361W\314\274h@\254{&aFBTb\tF\271\004\327\247\346C\nF\034\220\334e \303\021\241\266\237U\233\231{\266\017Q@\211dF\022\036S\031I\260\220\373\236\221!6\201\364\301\267m\307\277\260U\315\240\026\207P\342v\334\320\014\203O\003\260H`\2067\036\314l\301\346F\024\027E\223Y\0227\327\034\333Dr\311Z\255\"\320a\244\357\232\030^\344\002\316\206\300\027fc\371\276\217\006\265bX\t#\036\362\271Gd$\3305C\240\0218\345\336\034\304\322\010\257\002\223XF\236\307\234N\257q\365\025\206\004s\255`\003\016\363\346r\321q\360\362\360\342\300\366\205K$Xa\344N#\315\356\326f\302\202\204\275R\364\031\336\006b\037\013Z\234\\\022P\023\023132\032'j\\\254v\353%\233\246\341B*/L\022\006.\3208\246\362\253\366\277\001X\t\022\004u05^\0340A\2243\021\347c\362\334Ww\251@t*\036\252\000A\301p\036^\007\367\320\2168\037\010>\3272\374\220\200G\364\346\016k\362\340\307l7r$7\221\357\331\006\202H\004~\310\314k+B\235f1\323M w\337\177\014\021\307\307\035\240\257`\276@\343\307\304\341t\232\340\265\303i\206A\001\002\036%*XM{\356\356\200\273*V\321\355\345\302\247\006\363RKn\002\246r\317X ;\001\273\327\334G\337\233Fhv\311\204\3008\030K1\243\321\315Y\032\257UF\256\3311\3630\2460\327\253\270m\"\300wq!\206<\"I7Et\211\213\262P\032\205\270D\023\200\027h\033W\276/1\344c\346\364\004\352\273D\231w\343Y-k\261\272\206b\342\270\006\207\227\223k;\332tf\006<\216\231v=\215I\r\303\223Y\013&0\365n\374H\326YXD\016\233$\250*\364\017\231\235\031z\366\267\374\0103""\242\030\224\337I\034V\\.TJ\r\2314p)\263\tF\025\250\2506\302x5\023\313\324\030\237\305\260\014L\250\360M\342\266=\010\033'\236[\356g8h\215\245\362\177\242\371\324\365\ta\240\n1c\242OS\327'\330\221\343\214\n\332\365\tr\r<\006\256\204\236a\202\262K{\017\023\250660\332\376\247\213\212\306\251\234\273\256D\252\370\275\203\377e\236\257\315\376\364\274\257\377\327z_\257\254'\311\340^:\357oJ[\243f\022\347\263\272{\321q\251B\345\034\2560\0015\001{\r\225\223\250\240.\2712\271\206\245 ^X'\353O\023?X\260\207]U\250\306\335F\367\376\320vd\327\250\203\276\300\310\340\036\345\330\370\311k\034\\\303\357\313a\337I\327d\254\201X>t\t\356\254\335\343\214`\225\013\202\376\332\r\370<\302Z8\r\243\217\365\323\375\331\337@\026\332\300Zj\003\352j\203\200j\203s\256R\003\261\322'\205b\331\031;\010\352\227\023\353\374\232C\271\205\207\245\312F\254\313\205\313\240+\203\365\213\014\204H\301p\257\333\244sg\327\223(\017\353\313\375\200\246\332o\032\\\210?x\037\352\211\312f\001\017\230\303\265\026\255\207%\027\335v\231cc\023\340\0073T\r\221\365i\252k\240\032\033w\304Y[\177\232\236\335 W\327\214\r\276\315\347\275\377\352\270@\223\243\265\034F\372\246z\210;\374\334Rh\200 +X\351\346\035\265\317\324\272\002\215_\342.9\375\321\365>\374\222\0272\217b\253\035\216i5\250\366\210\377\261{\242\016\233\000\353K\240T\214P\214\024p]l\005\007\205}\0311\261\251s#V\245\025\370+\017\375,^X\0048\205KN\334\364\225\023r|\017\005\353f\022\356\252+pU\237\327\237\343\311i\336h\233\352\370\255\317\253\216\267q0t\375\326\305 ^;\304\373$w~\366\254\017\231\314\303\026\363\223d\013\313R\340`\303\033\243\225?\277\306`Zt\016uen\314\370\263\310\306\022i\341\013b\335@C\223\222\341l\026\354>\232\346~&.d\3129F\"\314p\027\262\360\374N\036bM\245\200\273s\007E\005\003~A\316\3471\223\324\207\006\022cz\356\232\346\376\263@\367\202)(\036\260\027\023[e\275\016\364\372\356\370\303\2313\325\\\260\211\251m\2018o\r\263I\260\251\203D\343X\302j\316\216\177\276\336M\002\326\004\362\343\033\254\203\333\033\rm\277\007\r\225\365B\313 \235P\"oo(\267""\207\265q\014\352\205q\214\r\252\242\002[9\256h\020\366\337E4im6\334\000&\r\364Xv\205\200\213]')\002z\016\tt\274\265\342X\324\314\353\244\252\370\326\031S,\273\270\016U \211\260/c\275`p\222^\352\204~\347\356*C65\274^lfO\000^\267\257\330\346X\2211Cu\335\020&r,\242X\0170\217\266\234\2255\254d8\036\026\032\024D\256\330\356w\307\014vG\250\212d\233=$\333g\365e\371)\253\000\016\t\212\361\221\035\036\262\002\216\311!+\317C\374\247>C\247o\360\260\317\267\317P\356\376\223\265\202m~\254\262\337\025$UU\340\340pBE\345\263\t|\334'O\345\010\230\2320\340\276\340\206w&%:\240+y\311\032\305\005\036\3411/\016I\325!\305\373\251\312a\233\234\312\241\246F\236\035\237\320<e\215\017\221\347\354\035N\257\305)/\263K@\267F#\335\347\250\010\312*\251F\000\212\216iR\244C\351\257\"9\235\262\024v\207S^Th\363\352g\236\352\273H\267P\275\237:]i\366l>\270\002\364E\323\354\3557\256\221\241\004\262\267\354\210\366\376\231\034\217\331~p\370\264\200d[\355\362#\362\313j\260\320\031\031\361\323\274\366\226\362\375\270\235\200\206Z\360\006\266?\321\253\366\311{\376:\274\335\307}\376\253[\005O<\211i\nkq\247rl\203\203\272\221\254(\364\t\265\314\354C\371\303\277:\363\345t\341\222J\324\230P\237\327K\313\354\230B>r(\204\2157T\342\236S\310\366\331\341r+\345\356\251\r/#\264-\006N\240D\272b\034V\2677\351\356\021\036\236\306\274\026\030k\034\262\216I\365Zd\235K\346E\232\025\035X\274\\X\245\374\005\247\244(w\307'\310\037\361B\222\264\323\\\275\326\0366\222\276=\356\362\351t\360k\213\026\334\346iv\310\323\327}\206\331\246}B\235\210\324\363\3455\3317\377\313\254J\263\307\344u\217I\253R\036\013\377\005k?\330U";
    PyObject *data = __Pyx_DecompressString(cstring, 1924, 1);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #else /* compression: none (6972 bytes) */
const char* const bytes = ".?pyverbs.providers.mlx5.mlx5_enumsIntEnumIntFlagMLX5DV_BLOCK_SIZE_4048MLX5DV_BLOCK_SIZE_4096MLX5DV_BLOCK_SIZE_4160MLX5DV_BLOCK_SIZE_512MLX5DV_BLOCK_SIZE_520MLX5DV_CONTEXT_ATTR_MASK_FD_ARRAYMLX5DV_CONTEXT_FLAGS_BLUEFLAMEMLX5DV_CONTEXT_FLAGS_CQE_128B_COMPMLX5DV_CONTEXT_FLAGS_CQE_128B_PADMLX5DV_CONTEXT_FLAGS_CQE_V1MLX5DV_CONTEXT_FLAGS_DEVXMLX5DV_CONTEXT_FLAGS_ENHANCED_MPWMLX5DV_CONTEXT_FLAGS_MPW_ALLOWEDMLX5DV_CONTEXT_FLAGS_PACKET_BASED_CREDIT_MODEMLX5DV_CONTEXT_FLAGS_REAL_TIME_TSMLX5DV_CONTEXT_MASK_CLOCK_INFO_UPDATEMLX5DV_CONTEXT_MASK_CQE_COMPRESIONMLX5DV_CONTEXT_MASK_CRYPTO_OFFLOADMLX5DV_CONTEXT_MASK_DCI_STREAMSMLX5DV_CONTEXT_MASK_DC_ODP_CAPSMLX5DV_CONTEXT_MASK_DYN_BFREGSMLX5DV_CONTEXT_MASK_FLOW_ACTION_FLAGSMLX5DV_CONTEXT_MASK_MAX_DC_RD_ATOMMLX5DV_CONTEXT_MASK_NUM_LAG_PORTSMLX5DV_CONTEXT_MASK_OOO_RECV_WRSMLX5DV_CONTEXT_MASK_SIGNATURE_OFFLOADMLX5DV_CONTEXT_MASK_STRIDING_RQMLX5DV_CONTEXT_MASK_SWPMLX5DV_CONTEXT_MASK_TUNNEL_OFFLOADSMLX5DV_CONTEXT_MASK_WR_MEMCPY_LENGTHMLX5DV_CQE_RES_FORMAT_CSUMMLX5DV_CQE_RES_FORMAT_CSUM_STRIDXMLX5DV_CQE_RES_FORMAT_HASHMLX5DV_CQ_INIT_ATTR_FLAGS_CQE_PADMLX5DV_CQ_INIT_ATTR_FLAGS_RESERVEDMLX5DV_CQ_INIT_ATTR_MASK_COMPRESSED_CQEMLX5DV_CQ_INIT_ATTR_MASK_CQE_SIZEMLX5DV_CQ_INIT_ATTR_MASK_FLAGSMLX5DV_CRYPTO_CAPS_CRYPTOMLX5DV_CRYPTO_CAPS_WRAPPED_CRYPTO_OPERATIONALMLX5DV_CRYPTO_CAPS_WRAPPED_CRYPTO_GOING_TO_COMMISSIONINGMLX5DV_CRYPTO_ENGINES_CAP_AES_XTSMLX5DV_CRYPTO_ENGINES_CAP_AES_XTS_SINGLE_BLOCKMLX5DV_CRYPTO_ENGINES_CAP_AES_XTS_MULTI_BLOCKMLX5DV_CRYPTO_KEY_PURPOSE_AES_XTSMLX5DV_CRYPTO_KEY_SIZE_128MLX5DV_CRYPTO_KEY_SIZE_256MLX5DV_CRYPTO_LOGIN_STATE_INVALIDMLX5DV_CRYPTO_LOGIN_STATE_NO_LOGINMLX5DV_CRYPTO_LOGIN_STATE_VALIDMLX5DV_CRYPTO_STANDARD_AES_XTSMLX5DV_CRYPTO_WRAPPED_IMPORT_METHOD_CAP_AES_XTSMLX5DV_DCTYPE_DCIMLX5DV_DCTYPE_DCTMLX5DV_DEK_INIT_ATTR_CRYPTO_LOGINMLX5DV_DEK_STATE_ERRORMLX5DV_DEK_STATE_READYMLX5DV_DEVX_CREATE_EVENT_CHANNEL_FLAGS_OMIT_EV_DATAMLX5DV_DR_ACTION_DESTMLX5DV_DR_ACTION_DEST_REFORMATMLX5DV_DR_ACTION_FLAGS_ROOT_LEVELMLX5DV_DR_DOM""AIN_SYNC_FLAGS_HWMLX5DV_DR_DOMAIN_SYNC_FLAGS_MEMMLX5DV_DR_DOMAIN_SYNC_FLAGS_SWMLX5DV_DR_DOMAIN_TYPE_FDBMLX5DV_DR_DOMAIN_TYPE_NIC_RXMLX5DV_DR_DOMAIN_TYPE_NIC_TXMLX5DV_DR_MATCHER_LAYOUT_NUM_RULEMLX5DV_DR_MATCHER_LAYOUT_RESIZABLEMLX5DV_FLOW_ACTION_COUNTERS_DEVXMLX5DV_FLOW_ACTION_COUNTERS_DEVX_WITH_OFFSETMLX5DV_FLOW_ACTION_DEFAULT_MISSMLX5DV_FLOW_ACTION_DEST_DEVXMLX5DV_FLOW_ACTION_DEST_IBV_QPMLX5DV_FLOW_ACTION_DROPMLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCMMLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_REQ_METADATAMLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_SPI_STEERINGMLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_FULL_OFFLOADMLX5DV_FLOW_ACTION_FLAGS_ESP_AES_GCM_TX_IV_IS_ESNMLX5DV_FLOW_ACTION_IBV_COUNTERMLX5DV_FLOW_ACTION_IBV_FLOW_ACTIONMLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TUNNEL_TO_L2_MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TO_L2_TUNNEL_MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L3_TUNNEL_TO_L2_MLX5DV_FLOW_ACTION_PACKET_REFORMAT_TYPE_L2_TO_L3_TUNNEL_MLX5DV_FLOW_ACTION_TAGMLX5DV_FLOW_MATCHER_MASK_FT_TYPEMLX5DV_FLOW_MATCHER_MASK_IB_PORTMLX5DV_FLOW_TABLE_TYPE_FDB_MLX5DV_FLOW_TABLE_TYPE_NIC_RX_MLX5DV_FLOW_TABLE_TYPE_NIC_TX_MLX5DV_FLOW_TABLE_TYPE_RDMA_RX_MLX5DV_FLOW_TABLE_TYPE_RDMA_TRANSPORT_RX_MLX5DV_FLOW_TABLE_TYPE_RDMA_TRANSPORT_TX_MLX5DV_FLOW_TABLE_TYPE_RDMA_TX_MLX5DV_MKEY_INIT_ATTR_FLAGS_BLOCK_SIGNATUREMLX5DV_MKEY_INIT_ATTR_FLAGS_CRYPTOMLX5DV_MKEY_INIT_ATTR_FLAGS_INDIRECTMLX5DV_MKEY_INIT_ATTR_FLAGS_REMOTE_INVALIDATEMLX5DV_MKEY_NO_ERRMLX5DV_MKEY_SIG_BLOCK_BAD_APPTAGMLX5DV_MKEY_SIG_BLOCK_BAD_GUARDMLX5DV_MKEY_SIG_BLOCK_BAD_REFTAGMLX5DV_OBJ_AHMLX5DV_OBJ_CQMLX5DV_OBJ_DMMLX5DV_OBJ_PDMLX5DV_OBJ_QPMLX5DV_OBJ_RWQMLX5DV_OBJ_SRQ_MLX5DV_PP_ALLOC_FLAGS_DEDICATED_INDEXMLX5DV_QP_CREATE_ALLOW_SCATTER_TO_CQEMLX5DV_QP_CREATE_DISABLE_SCATTER_TO_CQEMLX5DV_QP_CREATE_OOO_DPMLX5DV_QP_CREATE_PACKET_BASED_CREDIT_MODEMLX5DV_QP_CREATE_SIG_PIPELININGMLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_UCMLX5DV_QP_CREATE_TIR_ALLOW_SELF_LOOPBACK_MCMLX5DV_QP_CREATE_TUNNEL_OFFLOADSMLX5DV_QP_EX_WITH_MEMCPYMLX5DV_QP_EX_WITH_MKEY_CONFIGUREML""X5DV_QP_EX_WITH_MR_INTERLEAVEDMLX5DV_QP_EX_WITH_MR_LISTMLX5DV_QP_EX_WITH_RAW_WQEMLX5DV_QP_INIT_ATTR_MASK_DCMLX5DV_QP_INIT_ATTR_MASK_DCI_STREAMSMLX5DV_QP_INIT_ATTR_MASK_QP_CREATE_FLAGSMLX5DV_QP_INIT_ATTR_MASK_SEND_OPS_FLAGSMLX5DV_QP_MASK_RAW_QP_HANDLESMLX5DV_QP_MASK_RAW_QP_TIR_ADDRMLX5DV_QP_MASK_UAR_MMAP_OFFSETMLX5DV_QUERY_PORT_ESW_OWNER_VHCA_ID_MLX5DV_QUERY_PORT_VPORT_MLX5DV_QUERY_PORT_VPORT_REG_C0_MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_RX_MLX5DV_QUERY_PORT_VPORT_STEERING_ICM_TX_MLX5DV_QUERY_PORT_VPORT_VHCA_ID_MLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_VXLANMLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GREMLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_GENEVEMLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_CW_MPLS_OVER_GREMLX5DV_RAW_PACKET_CAP_TUNNELED_OFFLOAD_CW_MPLS_OVER_UDPMLX5DV_REG_DMABUF_ACCESS_DATA_DIRECT__MLX5DV_RES_TYPE_DBR_MLX5DV_RES_TYPE_QP_MLX5DV_RES_TYPE_RWQ_MLX5DV_RES_TYPE_SRQMLX5DV_SCHED_ELEM_ATTR_FLAGS_BW_SHAREMLX5DV_SCHED_ELEM_ATTR_FLAGS_MAX_AVG_BWMLX5DV_SIGNATURE_CRYPTO_ORDER_SIGNATURE_AFTER_CRYPTO_ON_TXMLX5DV_SIGNATURE_CRYPTO_ORDER_SIGNATURE_BEFORE_CRYPTO_ON_TXMLX5DV_SIG_BLOCK_ATTR_FLAG_COPY_MASKMLX5DV_SIG_CRC_TYPE_CRC32MLX5DV_SIG_CRC_TYPE_CRC32CMLX5DV_SIG_CRC_TYPE_CRC64_XP10MLX5DV_SIG_MASK_CRC32MLX5DV_SIG_MASK_CRC32CMLX5DV_SIG_MASK_CRC64_XP10MLX5DV_SIG_MASK_T10DIF_APPTAGMLX5DV_SIG_MASK_T10DIF_GUARDMLX5DV_SIG_MASK_T10DIF_REFTAGMLX5DV_SIG_T10DIF_CRCMLX5DV_SIG_T10DIF_CSUMMLX5DV_SIG_T10DIF_FLAG_APP_ESCAPEMLX5DV_SIG_T10DIF_FLAG_APP_REF_ESCAPEMLX5DV_SIG_T10DIF_FLAG_REF_REMAPMLX5DV_SIG_TYPE_CRCMLX5DV_SIG_TYPE_T10DIFMLX5DV_SRQ_MASK_SRQNMLX5DV_SW_PARSINGMLX5DV_SW_PARSING_CSUMMLX5DV_SW_PARSING_LSO_MLX5DV_UAR_ALLOC_TYPE_BF_MLX5DV_UAR_ALLOC_TYPE_NCMLX5DV_VFIO_CTX_FLAGS_INIT_LINK_DOWNMLX5DV_WC_MEMCPYMLX5DV_WC_RAW_WQEMLX5DV_WC_UMRcline_in_tracebackenumget__main___member_names___members__mlx5dv_block_sizemlx5dv_context_attr_comp_maskmlx5dv_context_attr_flagsmlx5dv_context_comp_maskmlx5dv_context_flagsmlx5dv_cq_init_attr_flagsmlx5dv_cq_init_attr_maskmlx5dv_cqe_comp_res_formatmlx5dv_crypto_caps""_flagsmlx5dv_crypto_engines_capsmlx5dv_crypto_key_purposemlx5dv_crypto_key_sizemlx5dv_crypto_login_statemlx5dv_crypto_standardmlx5dv_crypto_wrapped_import_method_capsmlx5dv_dc_typemlx5dv_dek_init_attr_maskmlx5dv_dek_statemlx5dv_devx_create_event_channel_flagsmlx5dv_dr_action_dest_typemlx5dv_dr_action_flagsmlx5dv_dr_domain_sync_flagsmlx5dv_dr_domain_typemlx5dv_dr_matcher_layout_flagsmlx5dv_flow_action_cap_flagsmlx5dv_flow_action_typemlx5dv_flow_matcher_attr_maskmlx5dv_mkey_err_typemlx5dv_mkey_init_attr_flagsmlx5dv_obj_typemlx5dv_qp_comp_maskmlx5dv_qp_create_flagsmlx5dv_qp_create_send_ops_flagsmlx5dv_qp_init_attr_maskmlx5dv_sched_elem_attr_flagsmlx5dv_sig_block_attr_flagsmlx5dv_sig_crc_typemlx5dv_sig_maskmlx5dv_sig_t10dif_bg_typemlx5dv_sig_t10dif_flagsmlx5dv_sig_typemlx5dv_signature_crypto_ordermlx5dv_srq_comp_maskmlx5dv_sw_parsing_offloadsmlx5dv_tunnel_offloadsmlx5dv_vfio_context_attr_flagsmlx5dv_wc_opcodemodule__module____name____qualname__setdefault__test__";
    PyObject *data = NULL;
    CYTHON_UNUSED_VAR(__Pyx_DecompressString);
    #endif
    PyObject **stringtab = __pyx_mstate->__pyx_string_tab;
    Py_ssize_t pos = 0;
    for (int i = 0; i < 243; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyUnicode_DecodeUTF8(bytes + pos, bytes_length, NULL);
      if (likely(string) && i >= 3) PyUnicode_InternInPlace(&string);
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
      stringtab[i] = string;
      pos += bytes_length;
    }
    Py_XDECREF(data);
    for (Py_ssize_t i = 0; i < 243; i++) {
      if (unlikely(PyObject_Hash(stringtab[i]) == -1)) {
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
  }
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_codeobjects ### */

static int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  return 0;
}
/* #### Code section: init_globals ### */

static int __Pyx_InitGlobals(void) {
  /* PythonCompatibility.init */
  if (likely(__Pyx_init_co_variables() == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore (used by PyObjectGetAttrStrNoError) */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr (used by PyObjectGetAttrStrNoError) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError (used by HasAttr) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* HasAttr (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!PyUnicode_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_PyObject_GetAttrStrNoError(o, n);
    if (!r) {
        return (unlikely(PyErr_Occurred())) ? -1 : 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}
#endif

/* PyObjectCall (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall (used by PyObjectCallOneArg) */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) != (0)) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
#if CYTHON_VECTORCALL && !CYTHON_COMPILING_IN_LIMITED_API
  #if PY_VERSION_HEX < 0x03090000
    #define __Pyx_PyVectorcall_Function(callable) _PyVectorcall_Function(callable)
  #elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE vectorcallfunc __Pyx_PyVectorcall_Function(PyObject *callable) {
    PyTypeObject *tp = Py_TYPE(callable);
    #if defined(__Pyx_CyFunction_USED)
    if (__Pyx_CyFunction_CheckExact(callable)) {
        return __Pyx_CyFunction_func_vectorcall(callable);
    }
    #endif
    if (!PyType_HasFeature(tp, Py_TPFLAGS_HAVE_VECTORCALL)) {
        return NULL;
    }
    assert(PyCallable_Check(callable));
    Py_ssize_t offset = tp->tp_vectorcall_offset;
    assert(offset > 0);
    vectorcallfunc ptr;
    memcpy(&ptr, (char *) callable + offset, sizeof(ptr));
    return ptr;
}
  #else
    #define __Pyx_PyVectorcall_Function(callable) PyVectorcall_Function(callable)
  #endif
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject *const *args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
          #if CYTHON_COMPILING_IN_LIMITED_API
            return PyObject_Vectorcall(func, args, _nargs, NULL);
          #else
            vectorcallfunc f = __Pyx_PyVectorcall_Function(func);
            if (f) {
                return f(func, args, _nargs, NULL);
            }
          #endif
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_mstate_global->__pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* PyObjectCallOneArg (used by ImportImpl) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* TupleAndListFromArray (used by ImportImpl) */
#if !CYTHON_COMPILING_IN_CPYTHON && CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    Py_ssize_t i;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    for (i = 0; i < n; i++) {
        if (unlikely(__Pyx_PyTuple_SET_ITEM(res, i, src[i]) < (0))) {
            Py_DECREF(res);
            return NULL;
        }
        Py_INCREF(src[i]);
    }
    return res;
}
#elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* ImportImpl (used by Import) */
static int __Pyx__Import_GetModule(PyObject *qualname, PyObject **module) {
    PyObject *imported_module = PyImport_GetModule(qualname);
    if (unlikely(!imported_module)) {
        *module = NULL;
        if (PyErr_Occurred()) {
            return -1;
        }
        return 0;
    }
    *module = imported_module;
    return 1;
}
static int __Pyx__Import_Lookup(PyObject *qualname, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject **module) {
    PyObject *imported_module;
    PyObject *top_level_package_name;
    Py_ssize_t i;
    int status, module_found;
    Py_ssize_t dot_index;
    module_found = __Pyx__Import_GetModule(qualname, &imported_module);
    if (unlikely(!module_found || module_found == -1)) {
        *module = NULL;
        return module_found;
    }
    if (imported_names) {
        for (i = 0; i < len_imported_names; i++) {
            PyObject *imported_name = imported_names[i];
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
            int has_imported_attribute = PyObject_HasAttr(imported_module, imported_name);
#else
            int has_imported_attribute = PyObject_HasAttrWithError(imported_module, imported_name);
            if (unlikely(has_imported_attribute == -1)) goto error;
#endif
            if (!has_imported_attribute) {
                goto not_found;
            }
        }
        *module = imported_module;
        return 1;
    }
    dot_index = PyUnicode_FindChar(qualname, '.', 0, PY_SSIZE_T_MAX, 1);
    if (dot_index == -1) {
        *module = imported_module;
        return 1;
    }
    if (unlikely(dot_index == -2)) goto error;
    top_level_package_name = PyUnicode_Substring(qualname, 0, dot_index);
    if (unlikely(!top_level_package_name)) goto error;
    Py_DECREF(imported_module);
    status = __Pyx__Import_GetModule(top_level_package_name, module);
    Py_DECREF(top_level_package_name);
    return status;
error:
    Py_DECREF(imported_module);
    *module = NULL;
    return -1;
not_found:
    Py_DECREF(imported_module);
    *module = NULL;
    return 0;
}
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *from_list = 0;
    int module_found;
    if (!qualname) {
        qualname = name;
    }
    module_found = __Pyx__Import_Lookup(qualname, imported_names, len_imported_names, &module);
    if (likely(module_found == 1)) {
        return module;
    } else if (unlikely(module_found == -1)) {
        return NULL;
    }
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    if (imported_names) {
#if CYTHON_COMPILING_IN_CPYTHON
        from_list = __Pyx_PyList_FromArray(imported_names, len_imported_names);
        if (unlikely(!from_list))
            goto bad;
#else
        from_list = PyList_New(len_imported_names);
        if (unlikely(!from_list)) goto bad;
        for (Py_ssize_t i=0; i<len_imported_names; ++i) {
            if (PyList_SetItem(from_list, i, __Pyx_NewRef(imported_names[i])) < 0) goto bad;
        }
#endif
    }
    if (level == -1) {
        const char* package_sep = strchr(__Pyx_MODULE_NAME, '.');
        if (package_sep != (0)) {
            module = PyImport_ImportModuleLevelObject(
                name, moddict, empty_dict, from_list, 1);
            if (unlikely(!module)) {
                if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                    goto bad;
                PyErr_Clear();
            }
        }
        level = 0;
    }
    if (!module) {
        module = PyImport_ImportModuleLevelObject(
            name, moddict, empty_dict, from_list, level);
    }
bad:
    Py_XDECREF(from_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level) {
    return __Pyx__Import(name, imported_names, len_imported_names, qualname, __pyx_mstate_global->__pyx_d, level);
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_mstate_global->__pyx_kp_u_);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400) ||\
                CYTHON_COMPILING_IN_GRAAL
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError, "cannot import name %S", name);
    }
    return value;
}

/* ListPack */
static PyObject *__Pyx_PyList_Pack(Py_ssize_t n, ...) {
    va_list va;
    PyObject *l = PyList_New(n);
    va_start(va, n);
    if (unlikely(!l)) goto end;
    for (Py_ssize_t i=0; i<n; ++i) {
        PyObject *arg = va_arg(va, PyObject*);
        Py_INCREF(arg);
        if (__Pyx_PyList_SET_ITEM(l, i, arg) != (0)) {
            Py_CLEAR(l);
            goto end;
        }
    }
    end:
    va_end(va);
    return l;
}

/* GetAttr (used by Globals) */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
    if (likely(PyUnicode_Check(n)))
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* Globals */
static PyObject* __Pyx_Globals(void) {
    return __Pyx_NewRef(__pyx_mstate_global->__pyx_d);
}

/* PyObjectVectorCallKwBuilder */
#if CYTHON_VECTORCALL
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_PyObject_FastCallDict;
    if (__Pyx_PyTuple_SET_ITEM(builder, n, key) != (0)) return -1;
    Py_INCREF(key);
    args[n] = value;
    return 0;
}
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_VectorcallBuilder_AddArgStr;
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n);
}
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    PyObject *pyKey = PyUnicode_FromString(key);
    if (!pyKey) return -1;
    return __Pyx_VectorcallBuilder_AddArg(pyKey, value, builder, args, n);
}
#else // CYTHON_VECTORCALL
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, CYTHON_UNUSED PyObject **args, CYTHON_UNUSED int n) {
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return PyDict_SetItem(builder, key, value);
}
#endif

/* GetBuiltinName (used by GetModuleGlobalName) */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_mstate_global->__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
            "name '%U' is not defined", name);
    }
    return result;
}

/* PyDictVersioning (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        if (!PyErr_Occurred())
            PyErr_SetNone(PyExc_NameError);
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
    PyErr_Clear();
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    if (unlikely(__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, name, &result) == -1)) PyErr_Clear();
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return result;
    }
#else
    result = _PyDict_GetItem_KnownHash(__pyx_mstate_global->__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* dict_setdefault (used by CLineInTraceback) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value) {
    PyObject* value;
#if __PYX_LIMITED_VERSION_HEX >= 0x030F0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4)
    PyDict_SetDefaultRef(d, key, default_value, &value);
#elif CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    PyObject *args[] = {d, key, default_value};
    value = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_n_u_setdefault, args, 3 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#elif CYTHON_COMPILING_IN_LIMITED_API
    value = PyObject_CallMethodObjArgs(d, __pyx_mstate_global->__pyx_n_u_setdefault, key, default_value, NULL);
#else
    value = PyDict_SetDefault(d, key, default_value);
    if (unlikely(!value)) return NULL;
    Py_INCREF(value);
#endif
    return value;
}

/* CLineInTraceback (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
#define __Pyx_PyProbablyModule_GetDict(o) __Pyx_XNewRef(PyModule_GetDict(o))
#elif !CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyProbablyModule_GetDict(o) PyObject_GenericGetDict(o, NULL);
#else
PyObject* __Pyx_PyProbablyModule_GetDict(PyObject *o) {
    PyObject **dict_ptr = _PyObject_GetDictPtr(o);
    return dict_ptr ? __Pyx_XNewRef(*dict_ptr) : NULL;
}
#endif
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline = NULL;
    PyObject *ptype, *pvalue, *ptraceback;
    PyObject *cython_runtime_dict;
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_mstate_global->__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
    cython_runtime_dict = __Pyx_PyProbablyModule_GetDict(__pyx_mstate_global->__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, cython_runtime_dict,
            __Pyx_PyDict_SetDefault(cython_runtime_dict, __pyx_mstate_global->__pyx_n_u_cline_in_traceback, Py_False))
    }
    if (use_cline == NULL || use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    Py_XDECREF(use_cline);
    Py_XDECREF(cython_runtime_dict);
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache (used by AddTraceback) */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static __Pyx_CachedCodeObjectType *__pyx__find_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line) {
    __Pyx_CachedCodeObjectType* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!code_cache->entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if (unlikely(pos >= code_cache->count) || unlikely(code_cache->entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = code_cache->entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__find_code_object;
    return NULL; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just miss.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type old_count = __pyx_atomic_incr_acq_rel(&code_cache->accessor_count);
    if (old_count < 0) {
        __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
        return NULL;
    }
#endif
    __Pyx_CachedCodeObjectType *result = __pyx__find_code_object(code_cache, code_line);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
#endif
    return result;
#endif
}
static void __pyx__insert_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line, __Pyx_CachedCodeObjectType* code_object)
{
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = code_cache->entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            code_cache->entries = entries;
            code_cache->max_count = 64;
            code_cache->count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if ((pos < code_cache->count) && unlikely(code_cache->entries[pos].code_line == code_line)) {
        __Pyx_CachedCodeObjectType* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_INCREF(code_object);
        Py_DECREF(tmp);
        return;
    }
    if (code_cache->count == code_cache->max_count) {
        int new_max = code_cache->max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            code_cache->entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        code_cache->entries = entries;
        code_cache->max_count = new_max;
    }
    for (i=code_cache->count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    code_cache->count++;
    Py_INCREF(code_object);
}
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__insert_code_object;
    return; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just fail.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type expected = 0;
    if (!__pyx_atomic_int_cmp_exchange(&code_cache->accessor_count, &expected, INT_MIN)) {
        return;
    }
#endif
    __pyx__insert_code_object(code_cache, code_line, code_object);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_sub(&code_cache->accessor_count, INT_MIN);
#endif
#endif
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result = PyObject_Call(replace, __pyx_mstate_global->__pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!code_object) {
        code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
        if (unlikely(!code_object)) goto bad;
        py_py_line = PyLong_FromLong(py_line);
        if (unlikely(!py_py_line)) goto bad;
        if (c_line) {
            py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        } else {
            py_funcname = PyUnicode_FromString(funcname);
        }
        if (unlikely(!py_funcname)) goto bad;
        dict = PyDict_New();
        if (unlikely(!dict)) goto bad;
        {
            PyObject *old_code_object = code_object;
            code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
            Py_DECREF(old_code_object);
        }
        if (unlikely(!code_object)) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, code_object);
    } else {
        dict = PyDict_New();
    }
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    if (c_line) {
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
    }
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_mstate_global->__pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_unsigned_PY_LONG_LONG(unsigned PY_LONG_LONG value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned PY_LONG_LONG neg_one = (unsigned PY_LONG_LONG) -1, const_zero = (unsigned PY_LONG_LONG) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned PY_LONG_LONG) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned PY_LONG_LONG) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned PY_LONG_LONG),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(unsigned PY_LONG_LONG));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_block_size(enum mlx5dv_block_size value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_block_size neg_one = (enum mlx5dv_block_size) -1, const_zero = (enum mlx5dv_block_size) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_block_size) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_block_size) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_block_size) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_block_size) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_block_size) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_block_size),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_block_size));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_context_attr_comp_mask(enum mlx5dv_context_attr_comp_mask value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_context_attr_comp_mask neg_one = (enum mlx5dv_context_attr_comp_mask) -1, const_zero = (enum mlx5dv_context_attr_comp_mask) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_context_attr_comp_mask) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_context_attr_comp_mask) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_context_attr_comp_mask) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_context_attr_comp_mask) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_context_attr_comp_mask) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_context_attr_comp_mask),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_context_attr_comp_mask));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_context_attr_flags(enum mlx5dv_context_attr_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_context_attr_flags neg_one = (enum mlx5dv_context_attr_flags) -1, const_zero = (enum mlx5dv_context_attr_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_context_attr_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_context_attr_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_context_attr_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_context_attr_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_context_attr_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_context_attr_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_context_attr_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_context_comp_mask(enum mlx5dv_context_comp_mask value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_context_comp_mask neg_one = (enum mlx5dv_context_comp_mask) -1, const_zero = (enum mlx5dv_context_comp_mask) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_context_comp_mask) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_context_comp_mask) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_context_comp_mask) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_context_comp_mask) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_context_comp_mask) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_context_comp_mask),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_context_comp_mask));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_context_flags(enum mlx5dv_context_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_context_flags neg_one = (enum mlx5dv_context_flags) -1, const_zero = (enum mlx5dv_context_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_context_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_context_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_context_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_context_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_context_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_context_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_context_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_flags(enum mlx5dv_cq_init_attr_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_cq_init_attr_flags neg_one = (enum mlx5dv_cq_init_attr_flags) -1, const_zero = (enum mlx5dv_cq_init_attr_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_cq_init_attr_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_cq_init_attr_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_cq_init_attr_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_cq_init_attr_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_cq_init_attr_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_cq_init_attr_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_cq_init_attr_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_cq_init_attr_mask(enum mlx5dv_cq_init_attr_mask value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_cq_init_attr_mask neg_one = (enum mlx5dv_cq_init_attr_mask) -1, const_zero = (enum mlx5dv_cq_init_attr_mask) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_cq_init_attr_mask) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_cq_init_attr_mask) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_cq_init_attr_mask) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_cq_init_attr_mask) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_cq_init_attr_mask) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_cq_init_attr_mask),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_cq_init_attr_mask));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_cqe_comp_res_format(enum mlx5dv_cqe_comp_res_format value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_cqe_comp_res_format neg_one = (enum mlx5dv_cqe_comp_res_format) -1, const_zero = (enum mlx5dv_cqe_comp_res_format) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_cqe_comp_res_format) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_cqe_comp_res_format) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_cqe_comp_res_format) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_cqe_comp_res_format) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_cqe_comp_res_format) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_cqe_comp_res_format),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_cqe_comp_res_format));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_crypto_caps_flags(enum mlx5dv_crypto_caps_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_crypto_caps_flags neg_one = (enum mlx5dv_crypto_caps_flags) -1, const_zero = (enum mlx5dv_crypto_caps_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_crypto_caps_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_crypto_caps_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_crypto_caps_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_crypto_caps_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_crypto_caps_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_crypto_caps_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_crypto_caps_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_crypto_engines_caps(enum mlx5dv_crypto_engines_caps value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_crypto_engines_caps neg_one = (enum mlx5dv_crypto_engines_caps) -1, const_zero = (enum mlx5dv_crypto_engines_caps) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_crypto_engines_caps) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_crypto_engines_caps) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_crypto_engines_caps) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_crypto_engines_caps) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_crypto_engines_caps) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_crypto_engines_caps),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_crypto_engines_caps));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_crypto_key_purpose(enum mlx5dv_crypto_key_purpose value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_crypto_key_purpose neg_one = (enum mlx5dv_crypto_key_purpose) -1, const_zero = (enum mlx5dv_crypto_key_purpose) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_crypto_key_purpose) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_crypto_key_purpose) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_crypto_key_purpose) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_crypto_key_purpose) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_crypto_key_purpose) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_crypto_key_purpose),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_crypto_key_purpose));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_crypto_key_size(enum mlx5dv_crypto_key_size value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_crypto_key_size neg_one = (enum mlx5dv_crypto_key_size) -1, const_zero = (enum mlx5dv_crypto_key_size) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_crypto_key_size) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_crypto_key_size) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_crypto_key_size) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_crypto_key_size) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_crypto_key_size) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_crypto_key_size),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_crypto_key_size));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_crypto_login_state(enum mlx5dv_crypto_login_state value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_crypto_login_state neg_one = (enum mlx5dv_crypto_login_state) -1, const_zero = (enum mlx5dv_crypto_login_state) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_crypto_login_state) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_crypto_login_state) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_crypto_login_state) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_crypto_login_state) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_crypto_login_state) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_crypto_login_state),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_crypto_login_state));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_crypto_standard(enum mlx5dv_crypto_standard value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_crypto_standard neg_one = (enum mlx5dv_crypto_standard) -1, const_zero = (enum mlx5dv_crypto_standard) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_crypto_standard) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_crypto_standard) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_crypto_standard) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_crypto_standard) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_crypto_standard) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_crypto_standard),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_crypto_standard));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_crypto_wrapped_import_method_caps(enum mlx5dv_crypto_wrapped_import_method_caps value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_crypto_wrapped_import_method_caps neg_one = (enum mlx5dv_crypto_wrapped_import_method_caps) -1, const_zero = (enum mlx5dv_crypto_wrapped_import_method_caps) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_crypto_wrapped_import_method_caps) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_crypto_wrapped_import_method_caps) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_crypto_wrapped_import_method_caps) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_crypto_wrapped_import_method_caps) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_crypto_wrapped_import_method_caps) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_crypto_wrapped_import_method_caps),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_crypto_wrapped_import_method_caps));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_dc_type(enum mlx5dv_dc_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_dc_type neg_one = (enum mlx5dv_dc_type) -1, const_zero = (enum mlx5dv_dc_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_dc_type) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_dc_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_dc_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_dc_type) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_dc_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_dc_type),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_dc_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_dek_init_attr_mask(enum mlx5dv_dek_init_attr_mask value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_dek_init_attr_mask neg_one = (enum mlx5dv_dek_init_attr_mask) -1, const_zero = (enum mlx5dv_dek_init_attr_mask) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_dek_init_attr_mask) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_dek_init_attr_mask) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_dek_init_attr_mask) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_dek_init_attr_mask) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_dek_init_attr_mask) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_dek_init_attr_mask),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_dek_init_attr_mask));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_dek_state(enum mlx5dv_dek_state value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_dek_state neg_one = (enum mlx5dv_dek_state) -1, const_zero = (enum mlx5dv_dek_state) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_dek_state) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_dek_state) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_dek_state) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_dek_state) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_dek_state) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_dek_state),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_dek_state));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_devx_create_event_channel_flags(enum mlx5dv_devx_create_event_channel_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_devx_create_event_channel_flags neg_one = (enum mlx5dv_devx_create_event_channel_flags) -1, const_zero = (enum mlx5dv_devx_create_event_channel_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_devx_create_event_channel_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_devx_create_event_channel_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_devx_create_event_channel_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_devx_create_event_channel_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_devx_create_event_channel_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_devx_create_event_channel_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_devx_create_event_channel_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_dr_action_dest_type(enum mlx5dv_dr_action_dest_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_dr_action_dest_type neg_one = (enum mlx5dv_dr_action_dest_type) -1, const_zero = (enum mlx5dv_dr_action_dest_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_dr_action_dest_type) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_dr_action_dest_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_dr_action_dest_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_dr_action_dest_type) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_dr_action_dest_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_dr_action_dest_type),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_dr_action_dest_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_dr_action_flags(enum mlx5dv_dr_action_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_dr_action_flags neg_one = (enum mlx5dv_dr_action_flags) -1, const_zero = (enum mlx5dv_dr_action_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_dr_action_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_dr_action_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_dr_action_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_dr_action_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_dr_action_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_dr_action_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_dr_action_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_dr_domain_sync_flags(enum mlx5dv_dr_domain_sync_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_dr_domain_sync_flags neg_one = (enum mlx5dv_dr_domain_sync_flags) -1, const_zero = (enum mlx5dv_dr_domain_sync_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_dr_domain_sync_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_dr_domain_sync_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_dr_domain_sync_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_dr_domain_sync_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_dr_domain_sync_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_dr_domain_sync_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_dr_domain_sync_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_dr_domain_type(enum mlx5dv_dr_domain_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_dr_domain_type neg_one = (enum mlx5dv_dr_domain_type) -1, const_zero = (enum mlx5dv_dr_domain_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_dr_domain_type) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_dr_domain_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_dr_domain_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_dr_domain_type) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_dr_domain_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_dr_domain_type),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_dr_domain_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_dr_matcher_layout_flags(enum mlx5dv_dr_matcher_layout_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_dr_matcher_layout_flags neg_one = (enum mlx5dv_dr_matcher_layout_flags) -1, const_zero = (enum mlx5dv_dr_matcher_layout_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_dr_matcher_layout_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_dr_matcher_layout_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_dr_matcher_layout_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_dr_matcher_layout_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_dr_matcher_layout_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_dr_matcher_layout_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_dr_matcher_layout_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_flow_action_cap_flags(enum mlx5dv_flow_action_cap_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_flow_action_cap_flags neg_one = (enum mlx5dv_flow_action_cap_flags) -1, const_zero = (enum mlx5dv_flow_action_cap_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_flow_action_cap_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_flow_action_cap_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_flow_action_cap_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_flow_action_cap_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_flow_action_cap_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_flow_action_cap_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_flow_action_cap_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_flow_action_type(enum mlx5dv_flow_action_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_flow_action_type neg_one = (enum mlx5dv_flow_action_type) -1, const_zero = (enum mlx5dv_flow_action_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_flow_action_type) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_flow_action_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_flow_action_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_flow_action_type) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_flow_action_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_flow_action_type),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_flow_action_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_flow_matcher_attr_mask(enum mlx5dv_flow_matcher_attr_mask value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_flow_matcher_attr_mask neg_one = (enum mlx5dv_flow_matcher_attr_mask) -1, const_zero = (enum mlx5dv_flow_matcher_attr_mask) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_flow_matcher_attr_mask) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_flow_matcher_attr_mask) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_flow_matcher_attr_mask) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_flow_matcher_attr_mask) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_flow_matcher_attr_mask) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_flow_matcher_attr_mask),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_flow_matcher_attr_mask));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_mkey_err_type(enum mlx5dv_mkey_err_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_mkey_err_type neg_one = (enum mlx5dv_mkey_err_type) -1, const_zero = (enum mlx5dv_mkey_err_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_mkey_err_type) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_mkey_err_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_mkey_err_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_mkey_err_type) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_mkey_err_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_mkey_err_type),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_mkey_err_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_mkey_init_attr_flags(enum mlx5dv_mkey_init_attr_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_mkey_init_attr_flags neg_one = (enum mlx5dv_mkey_init_attr_flags) -1, const_zero = (enum mlx5dv_mkey_init_attr_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_mkey_init_attr_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_mkey_init_attr_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_mkey_init_attr_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_mkey_init_attr_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_mkey_init_attr_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_mkey_init_attr_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_mkey_init_attr_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_obj_type(enum mlx5dv_obj_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_obj_type neg_one = (enum mlx5dv_obj_type) -1, const_zero = (enum mlx5dv_obj_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_obj_type) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_obj_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_obj_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_obj_type) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_obj_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_obj_type),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_obj_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_qp_comp_mask(enum mlx5dv_qp_comp_mask value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_qp_comp_mask neg_one = (enum mlx5dv_qp_comp_mask) -1, const_zero = (enum mlx5dv_qp_comp_mask) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_qp_comp_mask) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_qp_comp_mask) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_qp_comp_mask) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_qp_comp_mask) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_qp_comp_mask) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_qp_comp_mask),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_qp_comp_mask));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_qp_create_flags(enum mlx5dv_qp_create_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_qp_create_flags neg_one = (enum mlx5dv_qp_create_flags) -1, const_zero = (enum mlx5dv_qp_create_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_qp_create_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_qp_create_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_qp_create_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_qp_create_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_qp_create_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_qp_create_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_qp_create_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_qp_create_send_ops_flags(enum mlx5dv_qp_create_send_ops_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_qp_create_send_ops_flags neg_one = (enum mlx5dv_qp_create_send_ops_flags) -1, const_zero = (enum mlx5dv_qp_create_send_ops_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_qp_create_send_ops_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_qp_create_send_ops_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_qp_create_send_ops_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_qp_create_send_ops_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_qp_create_send_ops_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_qp_create_send_ops_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_qp_create_send_ops_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_qp_init_attr_mask(enum mlx5dv_qp_init_attr_mask value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_qp_init_attr_mask neg_one = (enum mlx5dv_qp_init_attr_mask) -1, const_zero = (enum mlx5dv_qp_init_attr_mask) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_qp_init_attr_mask) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_qp_init_attr_mask) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_qp_init_attr_mask) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_qp_init_attr_mask) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_qp_init_attr_mask) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_qp_init_attr_mask),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_qp_init_attr_mask));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_sched_elem_attr_flags(enum mlx5dv_sched_elem_attr_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_sched_elem_attr_flags neg_one = (enum mlx5dv_sched_elem_attr_flags) -1, const_zero = (enum mlx5dv_sched_elem_attr_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_sched_elem_attr_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_sched_elem_attr_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_sched_elem_attr_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_sched_elem_attr_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_sched_elem_attr_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_sched_elem_attr_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_sched_elem_attr_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_sig_block_attr_flags(enum mlx5dv_sig_block_attr_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_sig_block_attr_flags neg_one = (enum mlx5dv_sig_block_attr_flags) -1, const_zero = (enum mlx5dv_sig_block_attr_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_sig_block_attr_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_sig_block_attr_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_sig_block_attr_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_sig_block_attr_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_sig_block_attr_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_sig_block_attr_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_sig_block_attr_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_sig_crc_type(enum mlx5dv_sig_crc_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_sig_crc_type neg_one = (enum mlx5dv_sig_crc_type) -1, const_zero = (enum mlx5dv_sig_crc_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_sig_crc_type) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_sig_crc_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_sig_crc_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_sig_crc_type) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_sig_crc_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_sig_crc_type),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_sig_crc_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_sig_mask(enum mlx5dv_sig_mask value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_sig_mask neg_one = (enum mlx5dv_sig_mask) -1, const_zero = (enum mlx5dv_sig_mask) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_sig_mask) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_sig_mask) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_sig_mask) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_sig_mask) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_sig_mask) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_sig_mask),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_sig_mask));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_bg_type(enum mlx5dv_sig_t10dif_bg_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_sig_t10dif_bg_type neg_one = (enum mlx5dv_sig_t10dif_bg_type) -1, const_zero = (enum mlx5dv_sig_t10dif_bg_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_sig_t10dif_bg_type) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_sig_t10dif_bg_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_sig_t10dif_bg_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_sig_t10dif_bg_type) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_sig_t10dif_bg_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_sig_t10dif_bg_type),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_sig_t10dif_bg_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_sig_t10dif_flags(enum mlx5dv_sig_t10dif_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_sig_t10dif_flags neg_one = (enum mlx5dv_sig_t10dif_flags) -1, const_zero = (enum mlx5dv_sig_t10dif_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_sig_t10dif_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_sig_t10dif_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_sig_t10dif_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_sig_t10dif_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_sig_t10dif_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_sig_t10dif_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_sig_t10dif_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_sig_type(enum mlx5dv_sig_type value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_sig_type neg_one = (enum mlx5dv_sig_type) -1, const_zero = (enum mlx5dv_sig_type) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_sig_type) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_sig_type) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_sig_type) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_sig_type) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_sig_type) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_sig_type),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_sig_type));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_signature_crypto_order(enum mlx5dv_signature_crypto_order value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_signature_crypto_order neg_one = (enum mlx5dv_signature_crypto_order) -1, const_zero = (enum mlx5dv_signature_crypto_order) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_signature_crypto_order) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_signature_crypto_order) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_signature_crypto_order) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_signature_crypto_order) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_signature_crypto_order) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_signature_crypto_order),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_signature_crypto_order));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_srq_comp_mask(enum mlx5dv_srq_comp_mask value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_srq_comp_mask neg_one = (enum mlx5dv_srq_comp_mask) -1, const_zero = (enum mlx5dv_srq_comp_mask) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_srq_comp_mask) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_srq_comp_mask) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_srq_comp_mask) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_srq_comp_mask) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_srq_comp_mask) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_srq_comp_mask),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_srq_comp_mask));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_sw_parsing_offloads(enum mlx5dv_sw_parsing_offloads value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_sw_parsing_offloads neg_one = (enum mlx5dv_sw_parsing_offloads) -1, const_zero = (enum mlx5dv_sw_parsing_offloads) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_sw_parsing_offloads) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_sw_parsing_offloads) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_sw_parsing_offloads) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_sw_parsing_offloads) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_sw_parsing_offloads) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_sw_parsing_offloads),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_sw_parsing_offloads));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_tunnel_offloads(enum mlx5dv_tunnel_offloads value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_tunnel_offloads neg_one = (enum mlx5dv_tunnel_offloads) -1, const_zero = (enum mlx5dv_tunnel_offloads) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_tunnel_offloads) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_tunnel_offloads) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_tunnel_offloads) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_tunnel_offloads) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_tunnel_offloads) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_tunnel_offloads),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_tunnel_offloads));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_vfio_context_attr_flags(enum mlx5dv_vfio_context_attr_flags value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_vfio_context_attr_flags neg_one = (enum mlx5dv_vfio_context_attr_flags) -1, const_zero = (enum mlx5dv_vfio_context_attr_flags) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_vfio_context_attr_flags) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_vfio_context_attr_flags) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_vfio_context_attr_flags) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_vfio_context_attr_flags) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_vfio_context_attr_flags) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_vfio_context_attr_flags),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_vfio_context_attr_flags));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__mlx5dv_wc_opcode(enum mlx5dv_wc_opcode value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum mlx5dv_wc_opcode neg_one = (enum mlx5dv_wc_opcode) -1, const_zero = (enum mlx5dv_wc_opcode) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum mlx5dv_wc_opcode) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_wc_opcode) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum mlx5dv_wc_opcode) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum mlx5dv_wc_opcode) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum mlx5dv_wc_opcode) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum mlx5dv_wc_opcode),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum mlx5dv_wc_opcode));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
static __Pyx_TypeName
__Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp)
{
    PyObject *module = NULL, *name = NULL, *result = NULL;
    #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
    name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_qualname);
    #else
    name = PyType_GetQualName(tp);
    #endif
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) goto bad;
    module = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_module_2);
    if (unlikely(module == NULL) || unlikely(!PyUnicode_Check(module))) goto bad;
    if (PyUnicode_CompareWithASCIIString(module, "builtins") == 0) {
        result = name;
        name = NULL;
        goto done;
    }
    result = PyUnicode_FromFormat("%U.%U", module, name);
    if (unlikely(result == NULL)) goto bad;
  done:
    Py_XDECREF(name);
    Py_XDECREF(module);
    return result;
  bad:
    PyErr_Clear();
    if (name) {
        result = name;
        name = NULL;
    } else {
        result = __Pyx_NewRef(__pyx_mstate_global->__pyx_kp_u__2);
    }
    goto done;
}
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPyVerify (used by CIntFromPy) */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyLong_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyLong_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* GetRuntimeVersion */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
void __Pyx_init_runtime_version(void) {
    if (__Pyx_cached_runtime_version == 0) {
        const char* rt_version = Py_GetVersion();
        unsigned long version = 0;
        unsigned long factor = 0x01000000UL;
        unsigned int digit = 0;
        int i = 0;
        while (factor) {
            while ('0' <= rt_version[i] && rt_version[i] <= '9') {
                digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
                ++i;
            }
            version += factor * digit;
            if (rt_version[i] != '.')
                break;
            digit = 0;
            factor >>= 8;
            ++i;
        }
        __Pyx_cached_runtime_version = version;
    }
}
#endif
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    return Py_Version & ~0xFFUL;
#else
    return __Pyx_cached_runtime_version;
#endif
}

/* AddModuleRef */
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  static PyObject *__Pyx_PyImport_AddModuleObjectRef(PyObject *name) {
      PyObject *module_dict = PyImport_GetModuleDict();
      PyObject *m;
      if (PyMapping_GetOptionalItem(module_dict, name, &m) < 0) {
          return NULL;
      }
      if (m != NULL && PyModule_Check(m)) {
          return m;
      }
      Py_XDECREF(m);
      m = PyModule_NewObject(name);
      if (m == NULL)
          return NULL;
      if (PyDict_CheckExact(module_dict)) {
          PyObject *new_m;
          (void)PyDict_SetDefaultRef(module_dict, name, m, &new_m);
          Py_DECREF(m);
          return new_m;
      } else {
           if (PyObject_SetItem(module_dict, name, m) != 0) {
                Py_DECREF(m);
                return NULL;
            }
            return m;
      }
  }
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *py_name = PyUnicode_FromString(name);
      if (!py_name) return NULL;
      PyObject *module = __Pyx_PyImport_AddModuleObjectRef(py_name);
      Py_DECREF(py_name);
      return module;
  }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From___pyx_anon_enum(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* DecompressString */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo) {
    PyObject *module = NULL, *decompress, *compressed_bytes, *decompressed;
    const char* module_name = algo == 3 ? "compression.zstd" : algo == 2 ? "bz2" : "zlib";
    PyObject *methodname = PyUnicode_FromString("decompress");
    if (unlikely(!methodname)) return NULL;
    #if __PYX_LIMITED_VERSION_HEX >= 0x030e0000
    if (algo == 3) {
        PyObject *fromlist = Py_BuildValue("[O]", methodname);
        if (unlikely(!fromlist)) goto bad;
        module = PyImport_ImportModuleLevel("compression.zstd", NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
    } else
    #endif
        module = PyImport_ImportModule(module_name);
    if (unlikely(!module)) goto import_failed;
    decompress = PyObject_GetAttr(module, methodname);
    if (unlikely(!decompress)) goto import_failed;
    {
        #ifdef __cplusplus
            char *memview_bytes = const_cast<char*>(s);
        #else
            #if defined(__clang__)
              #pragma clang diagnostic push
              #pragma clang diagnostic ignored "-Wcast-qual"
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic push
              #pragma GCC diagnostic ignored "-Wcast-qual"
            #endif
            char *memview_bytes = (char*) s;
            #if defined(__clang__)
              #pragma clang diagnostic pop
            #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
              #pragma GCC diagnostic pop
            #endif
        #endif
        #if CYTHON_COMPILING_IN_LIMITED_API && !defined(PyBUF_READ)
        int memview_flags = 0x100;
        #else
        int memview_flags = PyBUF_READ;
        #endif
        compressed_bytes = PyMemoryView_FromMemory(memview_bytes, length, memview_flags);
    }
    if (unlikely(!compressed_bytes)) {
        Py_DECREF(decompress);
        goto bad;
    }
    decompressed = PyObject_CallFunctionObjArgs(decompress, compressed_bytes, NULL);
    Py_DECREF(compressed_bytes);
    Py_DECREF(decompress);
    Py_DECREF(module);
    Py_DECREF(methodname);
    return decompressed;
import_failed:
    PyErr_Format(PyExc_ImportError,
        "Failed to import '%.20s.decompress' - cannot initialise module strings. "
        "String compression was configured with the C macro 'CYTHON_COMPRESS_STRINGS=%d'.",
        module_name, algo);
bad:
    Py_XDECREF(module);
    Py_DECREF(methodname);
    return NULL;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    {
        const char* result;
        Py_ssize_t unicode_length;
        CYTHON_MAYBE_UNUSED_VAR(unicode_length); // only for __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (unlikely(PyArg_Parse(o, "s#", &result, length) < 0)) return NULL;
        #else
        result = PyUnicode_AsUTF8AndSize(o, length);
        #endif
        #if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        unicode_length = PyUnicode_GetLength(o);
        if (unlikely(unicode_length < 0)) return NULL;
        if (unlikely(unicode_length != *length)) {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
        #endif
        return result;
    }
#else
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
}
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
    if (PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
    if (PyByteArray_Check(o)) {
#if (CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS) || (CYTHON_COMPILING_IN_PYPY && (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE)))
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
#else
        *length = PyByteArray_Size(o);
        if (*length == -1) return NULL;
        return PyByteArray_AsString(o);
#endif
    } else
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_LongWrongResultType(PyObject* result) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(result));
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
    PyErr_Format(PyExc_TypeError,
                 "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ")",
                 result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  PyObject *res = NULL;
  if (likely(PyLong_Check(x)))
      return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  if (likely(m && m->nb_int)) {
      res = m->nb_int(x);
  }
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
      res = PyNumber_Long(x);
  }
#endif
  if (likely(res)) {
      if (unlikely(!PyLong_CheckExact(res))) {
          return __Pyx_PyNumber_LongWrongResultType(res);
      }
  }
  else if (!PyErr_Occurred()) {
      PyErr_SetString(PyExc_TypeError,
                      "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyLong_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyLong_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b) {
    CYTHON_UNUSED_VAR(b);
    return __Pyx_NewRef(Py_None);
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return __Pyx_NewRef(b ? Py_True: Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t ival) {
    return PyLong_FromSize_t(ival);
}


/* MultiPhaseInitModuleState */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
#ifndef CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#if (CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX >= 0x030C0000)
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 1
#else
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 0
#endif
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE && !CYTHON_ATOMICS
#error "Module state with PEP489 requires atomics. Currently that's one of\
 C11, C++11, gcc atomic intrinsics or MSVC atomic intrinsics"
#endif
#if !CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#define __Pyx_ModuleStateLookup_Lock()
#define __Pyx_ModuleStateLookup_Unlock()
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
static PyMutex __Pyx_ModuleStateLookup_mutex = {0};
#define __Pyx_ModuleStateLookup_Lock() PyMutex_Lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() PyMutex_Unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(__cplusplus) && __cplusplus >= 201103L
#include <mutex>
static std::mutex __Pyx_ModuleStateLookup_mutex;
#define __Pyx_ModuleStateLookup_Lock() __Pyx_ModuleStateLookup_mutex.lock()
#define __Pyx_ModuleStateLookup_Unlock() __Pyx_ModuleStateLookup_mutex.unlock()
#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ > 201112L) && !defined(__STDC_NO_THREADS__)
#include <threads.h>
static mtx_t __Pyx_ModuleStateLookup_mutex;
static once_flag __Pyx_ModuleStateLookup_mutex_once_flag = ONCE_FLAG_INIT;
static void __Pyx_ModuleStateLookup_initialize_mutex(void) {
    mtx_init(&__Pyx_ModuleStateLookup_mutex, mtx_plain);
}
#define __Pyx_ModuleStateLookup_Lock()\
  call_once(&__Pyx_ModuleStateLookup_mutex_once_flag, __Pyx_ModuleStateLookup_initialize_mutex);\
  mtx_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() mtx_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(HAVE_PTHREAD_H)
#include <pthread.h>
static pthread_mutex_t __Pyx_ModuleStateLookup_mutex = PTHREAD_MUTEX_INITIALIZER;
#define __Pyx_ModuleStateLookup_Lock() pthread_mutex_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() pthread_mutex_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(_WIN32)
#include <Windows.h>  // synchapi.h on its own doesn't work
static SRWLOCK __Pyx_ModuleStateLookup_mutex = SRWLOCK_INIT;
#define __Pyx_ModuleStateLookup_Lock() AcquireSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() ReleaseSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#else
#error "No suitable lock available for CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE.\
 Requires C standard >= C11, or C++ standard >= C++11,\
 or pthreads, or the Windows 32 API, or Python >= 3.13."
#endif
typedef struct {
    int64_t id;
    PyObject *module;
} __Pyx_InterpreterIdAndModule;
typedef struct {
    char interpreter_id_as_index;
    Py_ssize_t count;
    Py_ssize_t allocated;
    __Pyx_InterpreterIdAndModule table[1];
} __Pyx_ModuleStateLookupData;
#define __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE 32
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_int_type __Pyx_ModuleStateLookup_read_counter = 0;
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_ptr_type __Pyx_ModuleStateLookup_data = 0;
#else
static __Pyx_ModuleStateLookupData* __Pyx_ModuleStateLookup_data = NULL;
#endif
static __Pyx_InterpreterIdAndModule* __Pyx_State_FindModuleStateLookupTableLowerBound(
        __Pyx_InterpreterIdAndModule* table,
        Py_ssize_t count,
        int64_t interpreterId) {
    __Pyx_InterpreterIdAndModule* begin = table;
    __Pyx_InterpreterIdAndModule* end = begin + count;
    if (begin->id == interpreterId) {
        return begin;
    }
    while ((end - begin) > __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
        __Pyx_InterpreterIdAndModule* halfway = begin + (end - begin)/2;
        if (halfway->id == interpreterId) {
            return halfway;
        }
        if (halfway->id < interpreterId) {
            begin = halfway;
        } else {
            end = halfway;
        }
    }
    for (; begin < end; ++begin) {
        if (begin->id >= interpreterId) return begin;
    }
    return begin;
}
static PyObject *__Pyx_State_FindModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return NULL;
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData* data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
    {
        __pyx_atomic_incr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        if (likely(data)) {
            __Pyx_ModuleStateLookupData* new_data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_acquire(&__Pyx_ModuleStateLookup_data);
            if (likely(data == new_data)) {
                goto read_finished;
            }
        }
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        __Pyx_ModuleStateLookup_Lock();
        __pyx_atomic_incr_relaxed(&__Pyx_ModuleStateLookup_read_counter);
        data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
        __Pyx_ModuleStateLookup_Unlock();
    }
  read_finished:;
#else
    __Pyx_ModuleStateLookupData* data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_InterpreterIdAndModule* found = NULL;
    if (unlikely(!data)) goto end;
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            found = data->table+interpreter_id;
        }
    } else {
        found = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
    }
  end:
    {
        PyObject *result=NULL;
        if (found && found->id == interpreter_id) {
            result = found->module;
        }
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
#endif
        return result;
    }
}
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static void __Pyx_ModuleStateLookup_wait_until_no_readers(void) {
    while (__pyx_atomic_load(&__Pyx_ModuleStateLookup_read_counter) != 0);
}
#else
#define __Pyx_ModuleStateLookup_wait_until_no_readers()
#endif
static int __Pyx_State_AddModuleInterpIdAsIndex(__Pyx_ModuleStateLookupData **old_data, PyObject* module, int64_t interpreter_id) {
    Py_ssize_t to_allocate = (*old_data)->allocated;
    while (to_allocate <= interpreter_id) {
        if (to_allocate == 0) to_allocate = 1;
        else to_allocate *= 2;
    }
    __Pyx_ModuleStateLookupData *new_data = *old_data;
    if (to_allocate != (*old_data)->allocated) {
         new_data = (__Pyx_ModuleStateLookupData *)realloc(
            *old_data,
            sizeof(__Pyx_ModuleStateLookupData)+(to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
        if (!new_data) {
            PyErr_NoMemory();
            return -1;
        }
        for (Py_ssize_t i = new_data->allocated; i < to_allocate; ++i) {
            new_data->table[i].id = i;
            new_data->table[i].module = NULL;
        }
        new_data->allocated = to_allocate;
    }
    new_data->table[interpreter_id].module = module;
    if (new_data->count < interpreter_id+1) {
        new_data->count = interpreter_id+1;
    }
    *old_data = new_data;
    return 0;
}
static void __Pyx_State_ConvertFromInterpIdAsIndex(__Pyx_ModuleStateLookupData *data) {
    __Pyx_InterpreterIdAndModule *read = data->table;
    __Pyx_InterpreterIdAndModule *write = data->table;
    __Pyx_InterpreterIdAndModule *end = read + data->count;
    for (; read<end; ++read) {
        if (read->module) {
            write->id = read->id;
            write->module = read->module;
            ++write;
        }
    }
    data->count = write - data->table;
    for (; write<end; ++write) {
        write->id = 0;
        write->module = NULL;
    }
    data->interpreter_id_as_index = 0;
}
static int __Pyx_State_AddModule(PyObject* module, CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    int result = 0;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *old_data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *old_data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_ModuleStateLookupData *new_data = old_data;
    if (!new_data) {
        new_data = (__Pyx_ModuleStateLookupData *)calloc(1, sizeof(__Pyx_ModuleStateLookupData));
        if (!new_data) {
            result = -1;
            PyErr_NoMemory();
            goto end;
        }
        new_data->allocated = 1;
        new_data->interpreter_id_as_index = 1;
    }
    __Pyx_ModuleStateLookup_wait_until_no_readers();
    if (new_data->interpreter_id_as_index) {
        if (interpreter_id < __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
            result = __Pyx_State_AddModuleInterpIdAsIndex(&new_data, module, interpreter_id);
            goto end;
        }
        __Pyx_State_ConvertFromInterpIdAsIndex(new_data);
    }
    {
        Py_ssize_t insert_at = 0;
        {
            __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
                new_data->table, new_data->count, interpreter_id);
            assert(lower_bound);
            insert_at = lower_bound - new_data->table;
            if (unlikely(insert_at < new_data->count && lower_bound->id == interpreter_id)) {
                lower_bound->module = module;
                goto end;  // already in table, nothing more to do
            }
        }
        if (new_data->count+1 >= new_data->allocated) {
            Py_ssize_t to_allocate = (new_data->count+1)*2;
            new_data =
                (__Pyx_ModuleStateLookupData*)realloc(
                    new_data,
                    sizeof(__Pyx_ModuleStateLookupData) +
                    (to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
            if (!new_data) {
                result = -1;
                new_data = old_data;
                PyErr_NoMemory();
                goto end;
            }
            new_data->allocated = to_allocate;
        }
        ++new_data->count;
        int64_t last_id = interpreter_id;
        PyObject *last_module = module;
        for (Py_ssize_t i=insert_at; i<new_data->count; ++i) {
            int64_t current_id = new_data->table[i].id;
            new_data->table[i].id = last_id;
            last_id = current_id;
            PyObject *current_module = new_data->table[i].module;
            new_data->table[i].module = last_module;
            last_module = current_module;
        }
    }
  end:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, new_data);
#else
    __Pyx_ModuleStateLookup_data = new_data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return result;
}
static int __Pyx_State_RemoveModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *data = __Pyx_ModuleStateLookup_data;
#endif
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            data->table[interpreter_id].module = NULL;
        }
        goto done;
    }
    {
        __Pyx_ModuleStateLookup_wait_until_no_readers();
        __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
        if (!lower_bound) goto done;
        if (lower_bound->id != interpreter_id) goto done;
        __Pyx_InterpreterIdAndModule *end = data->table+data->count;
        for (;lower_bound<end-1; ++lower_bound) {
            lower_bound->id = (lower_bound+1)->id;
            lower_bound->module = (lower_bound+1)->module;
        }
    }
    --data->count;
    if (data->count == 0) {
        free(data);
        data = NULL;
    }
  done:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, data);
#else
    __Pyx_ModuleStateLookup_data = data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return 0;
}
#endif

/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
